# -*- coding: utf-8 -*-
# pylint: disable=bare-except,chained-comparison,invalid-name,len-as-condition,line-too-long,protected-access,redefined-outer-name,too-many-arguments,too-many-boolean-expressions,too-many-branches,too-many-lines,too-many-locals,too-many-statements,unneeded-not,unused-import
u"""
Interfaces for the NAG Mark 27.0 `complex` Chapter.

``complex`` - Complex Arithmetic

Base interfaces.

We do not intend for these lower-level base interfaces to be called
directly from user code. Please refer to the ``naginterfaces.library``
subpackage instead.

This module provides facilities for arithmetic operations involving
complex numbers.

For full information please refer to the NAG Library document

https://www.nag.com/numeric/nl/nagdoc_27/flhtml/a02/a02intro.html
"""

# NAG Copyright 2017-2020.

import collections as _collections
import logging as _logging
import sys as _sys

from ..base import utils

def sqrt(xr, xi):
    u"""
    Square root of complex number.

    ``sqrt`` evaluates the square root of the complex number x =
    (x_r,x_i).

    For full information please refer to the NAG Library document for
    a02aa

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/a02/a02aaf.html

    Parameters
    ----------
    xr : float
        x_r, the real part of x.

    xi : float
        x_i, the imaginary part of x.

    Returns
    -------
    yr : float
        y_r, the real part of y.

    yi : float
        y_i, the imaginary part of y.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Wilkinson, J H and Reinsch, C, 1971, `Handbook for Automatic
    Computation II, Linear Algebra`, Springer--Verlag
    """

    excs = []
    fun_name = 'naginterfaces.base.complex.sqrt'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('a02aaft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, xr.
        utils._BYREF_FLOAT64, # Input, xi.
        utils._BYREF_FLOAT64, # Output, yr.
        utils._BYREF_FLOAT64, # Output, yi.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _xr = utils._EngineFloat64ScalarType(
        xr, {'fun_name': fun_name, 'entity': 'xr'},
    ).to_ctypes()
    _xi = utils._EngineFloat64ScalarType(
        xi, {'fun_name': fun_name, 'entity': 'xi'},
    ).to_ctypes()
    _yr = utils._EngineFloat64ScalarType.empty_ctypes()
    _yi = utils._EngineFloat64ScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _xr,
        _xi,
        _yr,
        _yi,
    )
    utils._handle_cb_exception(excs)
    _yr_py = utils._EngineFloat64ScalarType.to_py(_yr)
    _yi_py = utils._EngineFloat64ScalarType.to_py(_yi)
    return_data = _collections.namedtuple(
        'nag_return_data',
        ['yr', 'yi'],
    )(
        _yr_py,
        _yi_py,
    )
    return return_data

def abs(xr, xi): # pylint: disable=redefined-builtin
    u"""
    Modulus of complex number.

    ``abs`` returns the value of the modulus of the complex number x =
    (x_r,x_i).

    For full information please refer to the NAG Library document for
    a02ab

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/a02/a02abf.html

    Parameters
    ----------
    xr : float
        x_r, the real part of x.

    xi : float
        x_i, the imaginary part of x.

    Returns
    -------
    xa : float
        The modulus of the complex number x = (x_r,x_i).

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Wilkinson, J H and Reinsch, C, 1971, `Handbook for Automatic
    Computation II, Linear Algebra`, Springer--Verlag
    """

    excs = []
    fun_name = 'naginterfaces.base.complex.abs'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('a02abft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, xr.
        utils._BYREF_FLOAT64, # Input, xi.
        utils._BYREF_FLOAT64, # Output, xa.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _xr = utils._EngineFloat64ScalarType(
        xr, {'fun_name': fun_name, 'entity': 'xr'},
    ).to_ctypes()
    _xi = utils._EngineFloat64ScalarType(
        xi, {'fun_name': fun_name, 'entity': 'xi'},
    ).to_ctypes()
    _xa = utils._EngineFloat64ScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _xr,
        _xi,
        _xa,
    )
    utils._handle_cb_exception(excs)
    _xa_py = utils._EngineFloat64ScalarType.to_py(_xa)
    return_data = _xa_py
    return return_data

def divide(xr, xi, yr, yi):
    u"""
    Quotient of two complex numbers.

    ``divide`` divides one complex number, x = (x_r,x_i), by a second
    complex number, y = (y_r,y_i), returning the result in z =
    (z_r,z_i).

    For full information please refer to the NAG Library document for
    a02ac

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/a02/a02acf.html

    Parameters
    ----------
    xr : float
        x_r, the real part of x.

    xi : float
        x_i, the imaginary part of x.

    yr : float
        y_r, the real part of y.

    yi : float
        y_i, the imaginary part of y.

    Returns
    -------
    zr : float
        z_r, the real part of z.

    zi : float
        z_i, the imaginary part of z.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Li, X S, Demmel, J W, Bailey, D H, Henry, G, Hida, Y, Iskandar, J,
    Kahan, W, Kapur, A, Martin, M C, Tung, T and Yoo, D J, 2002,
    `Design, implementation and testing of extended and mixed precision
    BLAS`, ACM Trans. Math. Soft. (28(2)), 152--205
    """

    excs = []
    fun_name = 'naginterfaces.base.complex.divide'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('a02acft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, xr.
        utils._BYREF_FLOAT64, # Input, xi.
        utils._BYREF_FLOAT64, # Input, yr.
        utils._BYREF_FLOAT64, # Input, yi.
        utils._BYREF_FLOAT64, # Output, zr.
        utils._BYREF_FLOAT64, # Output, zi.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _xr = utils._EngineFloat64ScalarType(
        xr, {'fun_name': fun_name, 'entity': 'xr'},
    ).to_ctypes()
    _xi = utils._EngineFloat64ScalarType(
        xi, {'fun_name': fun_name, 'entity': 'xi'},
    ).to_ctypes()
    _yr = utils._EngineFloat64ScalarType(
        yr, {'fun_name': fun_name, 'entity': 'yr'},
    ).to_ctypes()
    _yi = utils._EngineFloat64ScalarType(
        yi, {'fun_name': fun_name, 'entity': 'yi'},
    ).to_ctypes()
    _zr = utils._EngineFloat64ScalarType.empty_ctypes()
    _zi = utils._EngineFloat64ScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _xr,
        _xi,
        _yr,
        _yi,
        _zr,
        _zi,
    )
    utils._handle_cb_exception(excs)
    _zr_py = utils._EngineFloat64ScalarType.to_py(_zr)
    _zi_py = utils._EngineFloat64ScalarType.to_py(_zi)
    return_data = _collections.namedtuple(
        'nag_return_data',
        ['zr', 'zi'],
    )(
        _zr_py,
        _zi_py,
    )
    return return_data
