# -*- coding: utf-8 -*-
# pylint: disable=bare-except,chained-comparison,invalid-name,len-as-condition,line-too-long,protected-access,redefined-outer-name,too-many-arguments,too-many-boolean-expressions,too-many-branches,too-many-lines,too-many-locals,too-many-statements,unneeded-not,unused-import
u"""
Interfaces for the NAG Mark 27.0 `contab` Chapter.

``contab`` - Contingency Table Analysis

Base interfaces.

We do not intend for these lower-level base interfaces to be called
directly from user code. Please refer to the ``naginterfaces.library``
subpackage instead.

The functions in this module are for the analysis of discrete
multivariate data.
One suite of functions computes tables while other functions are for the
analysis of two-way contingency tables, conditional logistic models and
one-factor analysis of binary data.

Functions in submodule ``correg`` may be used to fit generalized linear
models to discrete data including binary data and contingency tables.

For full information please refer to the NAG Library document

https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g11/g11intro.html
"""

# NAG Copyright 2017-2020.

import collections as _collections
import logging as _logging
import sys as _sys

from ..base import utils

def chisq(nrow, ncol, nobs, expt, chist):
    u"""
    chi^2 statistics for two-way contingency table.

    ``chisq`` computes chi^2 statistics for a two-way contingency table.
    For a 2*2 table with a small number of observations exact
    probabilities are computed.

    For full information please refer to the NAG Library document for
    g11aa

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g11/g11aaf.html

    Parameters
    ----------
    nrow : int
        r, the number of rows in the contingency table.

    ncol : int
        c, the number of columns in the contingency table.

    nobs : int, shape (`nrow`, `ncol`)
        The contingency table `nobs`[i-1,j-1] must contain n_{ij}, for j
        = 1,2,...,c, for i = 1,2,...,r.

    expt : float, shape (`nrow`, `ncol`), modified in place
        `On exit`: the table of expected values. `expt`[i-1,j-1]
        contains f_{ij}, for j = 1,2,...,c, for i = 1,2,...,r.

    chist : float, shape (`nrow`, `ncol`), modified in place
        `On exit`: the table of chi^2 contributions. `chist`[i-1,j-1]
        contains ((n_{ij}-f_{ij})^2)/(f_{ij}), for j = 1,2,...,c, for i
        = 1,2,...,r.

    Returns
    -------
    prob : float
        If c = 2, r = 2 and n <= 40 then `prob` contains the two tail
        significance level for Fisher's exact test, otherwise `prob`
        contains the significance level from the Pearson chi^2
        statistic.

    chi : float
        The Pearson chi^2 statistic.

    g : float
        The likelihood ratio test statistic.

    df : float
        The degrees of freedom for the statistics.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `ncol` = *<value>*.

            Constraint: `ncol` >= 2.

        (`errno` 1)
            On entry, `nrow` = *<value>*.

            Constraint: `nrow` >= 2.

        (`errno` 2)
            On entry, all elements of `nobs` = 0.

        (`errno` 2)
            On entry, i = *<value>*, j = *<value>* and `nobs`[i-1,j-1] =
            *<value>*.

            Constraint: `nobs`[i-1,j-1] >= 0.

        (`errno` 3)
            On entry, a 2*2 table has a row or column with both values
            zero.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 4)
            At least one cell has an expected frequency, f_{{ij}} <=
            0.5.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Everitt, B S, 1977, `The Analysis of Contingency Tables`, Chapman
    and Hall

    Kendall, M G and Stuart, A, 1973, `The Advanced Theory of Statistics
    (Volume 2)`, (3rd Edition), Griffin
    """

    excs = []
    fun_name = 'naginterfaces.base.contab.chisq'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g11aaft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, nrow.
        utils._BYREF_INT, # Input, ncol.
        utils._EngineIntArrayType.c_type, # Input, nobs[ldnobs, ncol].
        utils._BYREF_INT, # Input, ldnobs.
        utils._EngineFloat64ArrayType.c_type, # Output, expt[ldnobs, ncol].
        utils._EngineFloat64ArrayType.c_type, # Output, chist[ldnobs, ncol].
        utils._BYREF_FLOAT64, # Output, prob.
        utils._BYREF_FLOAT64, # Output, chi.
        utils._BYREF_FLOAT64, # Output, g.
        utils._BYREF_FLOAT64, # Output, df.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((nobs, 'nobs'), (expt, 'expt'), (chist, 'chist')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _nrow = utils._EngineIntScalarType(
        nrow, {'fun_name': fun_name, 'entity': 'nrow'},
    ).to_ctypes()
    _ncol = utils._EngineIntScalarType(
        ncol, {'fun_name': fun_name, 'entity': 'ncol'},
    ).to_ctypes()
    _ldnobs_py = utils._EngineIntArrayType.get_shape(nobs, {'fun_name': fun_name, 'entity': 'nobs'}, exp_rank=2)[0]
    _ldnobs = utils._EngineIntScalarType(
        _ldnobs_py, {'fun_name': fun_name, 'entity': '_ldnobs'},
    ).to_ctypes()
    _nrow_py = utils._EngineIntScalarType.to_py(_nrow)
    _ncol_py = utils._EngineIntScalarType.to_py(_ncol)
    _nobs = utils._EngineIntArrayType(
        nobs, {'fun_name': fun_name, 'entity': 'nobs'},
        exp_shape=(_nrow_py, _ncol_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _expt_cls = utils._EngineFloat64ArrayType(
        expt, {'fun_name': fun_name, 'entity': 'expt'},
        exp_shape=(_nrow_py, _ncol_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _expt = _expt_cls.to_ctypes(can_reference_contents=False)
    _chist_cls = utils._EngineFloat64ArrayType(
        chist, {'fun_name': fun_name, 'entity': 'chist'},
        exp_shape=(_nrow_py, _ncol_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _chist = _chist_cls.to_ctypes(can_reference_contents=False)
    _prob = utils._EngineFloat64ScalarType.empty_ctypes()
    _chi = utils._EngineFloat64ScalarType.empty_ctypes()
    _g = utils._EngineFloat64ScalarType.empty_ctypes()
    _df = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _nrow,
        _ncol,
        _nobs,
        _ldnobs,
        _expt,
        _chist,
        _prob,
        _chi,
        _g,
        _df,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ncol = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ncol >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nrow = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nrow >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, all elements of nobs = 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, j = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'nobs[i-1,j-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: nobs[i-1,j-1] >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a 2*2 table has a row or column with both values zero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99994:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'At least one cell has an expected frequency, f_{{ij}} <= 0.5.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _expt_cls.refresh_py(_expt)
        _chist_cls.refresh_py(_chist)
        _prob_py = utils._EngineFloat64ScalarType.to_py(_prob)
        _chi_py = utils._EngineFloat64ScalarType.to_py(_chi)
        _g_py = utils._EngineFloat64ScalarType.to_py(_g)
        _df_py = utils._EngineFloat64ScalarType.to_py(_df)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['prob', 'chi', 'g', 'df'],
        )(
            _prob_py,
            _chi_py,
            _g_py,
            _df_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def tabulate_stat(stat, update, weight, n, nfac, isf, lfac, ifac, y, wt, table, maxt, ncells, idim, icount, auxt):
    u"""
    Computes multiway table from set of classification factors using
    selected statistic.

    ``tabulate_stat`` computes a table from a set of classification
    factors using a selected statistic.

    For full information please refer to the NAG Library document for
    g11ba

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g11/g11baf.html

    Parameters
    ----------
    stat : str, length 1
        Indicates which statistic is to be computed for the table cells.

        `stat` = 'N'

            The number of observations for each cell.

        `stat` = 'T'

            The total for the variable in `y` for each cell.

        `stat` = 'A'

            The average (mean) for the variable in `y` for each cell.

        `stat` = 'V'

            The variance for the variable in `y` for each cell.

        `stat` = 'L'

            The largest value for the variable in `y` for each cell.

        `stat` = 'S'

            The smallest value for the variable in `y` for each cell.

    update : str, length 1
        Indicates if an existing table is to be updated by further
        observation.

        `update` = 'I'

            The table cells will be initialized to zero before
            tabulations take place.

        `update` = 'U'

            The table input in `table` will be updated. The arguments
            `ncells`, `table`, `icount` and `auxt` must remain unchanged
            from the previous call to ``tabulate_stat``.

    weight : str, length 1
        Indicates if weights are to be used.

        `weight` = 'U'

            Weights are not used and unit weights are assumed.

        `weight` = 'W' or 'V'

            Weights are used and must be supplied in `wt`. The only
            difference between `weight` = 'W' and `weight` = 'V' is if
            the variance is computed.

        `weight` = 'W'

            The divisor for the variance is the sum of the weights minus
            one and if `weight` = 'V', the divisor is the number of
            observations with nonzero weights minus one. The former is
            useful if the weights represent the frequency of the
            observed values.

        If `stat` = 'T' or 'A', the weighted total or mean is computed
        respectively.

        If `stat` = 'N', 'L' or 'S', the only effect of weights is to
        eliminate values with zero weights from the computations.

    n : int
        The number of observations.

    nfac : int
        The number of classifying factors in `ifac`.

    isf : int, shape (`nfac`)
        Indicates which factors in `ifac` are to be used in the
        tabulation.

        If `isf`[i-1] > 0 the ith factor in `ifac` is included in the
        tabulation.

        Note that if `isf`[i-1] <= 0, for i = 1,2,...,`nfac` then the
        statistic for the whole sample is calculated and returned in a
        1*1 table.

    lfac : int, shape (`nfac`)
        The number of levels of the classifying factors in `ifac`.

    ifac : int, shape (`n`, `nfac`)
        The `nfac` coded classification factors for the `n`
        observations.

    y : float, shape (`n`)
        The variable to be tabulated. If `stat` = 'N', `y` is not
        referenced.

    wt : float, shape (:)
        Note: the required length for this argument is determined as
        follows: if `weight` in ('W', 'V'): `n`; otherwise: 1.

        If `weight` = 'W' or 'V', `wt` must contain the `n` weights.
        Otherwise `wt` is not referenced.

    table : float, shape (`maxt`), modified in place
        `On entry`: if `update` = 'U', `table` must be unchanged from
        the previous call to ``tabulate_stat``, otherwise `table` need
        not be set.

        `On exit`: the computed table. The `ncells` cells of the table
        are stored so that for any two factors the index relating to the
        factor referred to later in `lfac` and `ifac` changes faster.
        For further details see [Further Comments].

    maxt : int
        The maximum size of the table to be computed.

    ncells : int
        If `update` = 'U', `ncells` must be unchanged from the previous
        call to ``tabulate_stat``, otherwise `ncells` need not be set.

    idim : int, shape (`nfac`), modified in place
        `On exit`: the first `ndim` elements contain the number of
        levels for the factors defining the table.

    icount : int, shape (`maxt`), modified in place
        `On entry`: if `update` = 'U', `icount` must be unchanged from
        the previous call to ``tabulate_stat``, otherwise `icount` need
        not be set.

        `On exit`: a table containing the number of observations
        contributing to each cell of the table, stored identically to
        `table`. Note if `stat` = 'N' this is the same as is returned in
        `table`.

    auxt : float, shape (:), modified in place
        Note: the required length for this argument is determined as
        follows: if `stat`='A': `maxt`; if `stat`='V': 2*`maxt`;
        otherwise: 0.

        `On entry`: if `update` = 'U', `auxt` must be unchanged from the
        previous call to ``tabulate_stat``, otherwise `auxt` need not be
        set.

        `On exit`: if `stat` = 'A' or 'V', the first `ncells` values
        hold the table containing the sum of the weights for the
        observations contributing to each cell, stored identically to
        `table`.

        If `stat` = 'V', the second set of `ncells` values hold the
        table of cell means.

        Otherwise `auxt` is not referenced.

    Returns
    -------
    ncells : int
        The number of cells in the table.

    ndim : int
        The number of factors defining the table.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `weight` = *<value>*.

            Constraint: `weight` = 'U', 'V' or 'W'.

        (`errno` 1)
            On entry, `update` = *<value>*.

            Constraint: `update` = 'I' or 'U'.

        (`errno` 1)
            On entry, `stat` = *<value>*.

            Constraint: `stat` = 'N', 'T', 'A', 'V', 'L' or 'S'.

        (`errno` 1)
            On entry, `nfac` = *<value>*.

            Constraint: `nfac` >= 1.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 2.

        (`errno` 2)
            On entry, `maxt` = *<value>* and minimum value for `maxt` =
            *<value>*.

            Constraint: `maxt` >= product of the levels of the factors
            included in the tabulation.

        (`errno` 2)
            On entry, i = *<value>* and `wt`[i-1] = *<value>*.

            Constraint: `wt`[i-1] >= 0.0.

        (`errno` 2)
            On entry, i = *<value>*, j = *<value>*, `lfac`[j-1] =
            *<value>* and `ifac`[i-1,j-1] = *<value>*.

            Constraint: `ifac`[i-1,j-1] <= `lfac`[j-1].

        (`errno` 2)
            On entry, i = *<value>*, j = *<value>* and `ifac`[i-1,j-1] =
            *<value>*.

            Constraint: `ifac`[i-1,j-1] > 0.

        (`errno` 2)
            On entry, i = *<value>* and `lfac`[i-1] = *<value>*.

            On entry, `lfac`[i-1] > 2.

        (`errno` 3)
            The variance divisor <= 0.0.

        (`errno` 4)
            `auxt` or `table` has changed between calls.

        (`errno` 4)
            `auxt` has changed between calls.

        (`errno` 4)
            `ncells` has changed between calls.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    John, J A and Quenouille, M H, 1977, `Experiments: Design and
    Analysis`, Griffin

    Kendall, M G and Stuart, A, 1969, `The Advanced Theory of Statistics
    (Volume 1)`, (3rd Edition), Griffin

    West, D H D, 1979, `Updating mean and variance estimates: An
    improved method`, Comm. ACM (22), 532--555
    """

    excs = []
    fun_name = 'naginterfaces.base.contab.tabulate_stat'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g11baft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, stat(len=1).
        utils._BYREF_STRING, # Input, update(len=1).
        utils._BYREF_STRING, # Input, weight(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nfac.
        utils._EngineIntArrayType.c_type, # Input, isf[nfac].
        utils._EngineIntArrayType.c_type, # Input, lfac[nfac].
        utils._EngineIntArrayType.c_type, # Input, ifac[ldf, nfac].
        utils._BYREF_INT, # Input, ldf.
        utils._EngineFloat64ArrayType.c_type, # Input, y[n].
        utils._EngineFloat64ArrayType.c_type, # Input, wt[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, table[maxt].
        utils._BYREF_INT, # Input, maxt.
        utils._BYREF_INT, # Input/Output, ncells.
        utils._BYREF_INT, # Output, ndim.
        utils._EngineIntArrayType.c_type, # Output, idim[nfac].
        utils._EngineIntArrayType.c_type, # Input/Output, icount[maxt].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, auxt[*].
        utils._EngineIntArrayType.c_type, # Workspace, iwk.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of stat.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of update.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of weight.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((ifac, 'ifac'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _stat = utils._EngineCharScalarType(
        stat, {'fun_name': fun_name, 'entity': 'stat'}, exp_chrlen=1,
    ).to_ctypes()
    _update = utils._EngineCharScalarType(
        update, {'fun_name': fun_name, 'entity': 'update'}, exp_chrlen=1,
    ).to_ctypes()
    _weight = utils._EngineCharScalarType(
        weight, {'fun_name': fun_name, 'entity': 'weight'}, exp_chrlen=1,
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nfac = utils._EngineIntScalarType(
        nfac, {'fun_name': fun_name, 'entity': 'nfac'},
    ).to_ctypes()
    _nfac_py = utils._EngineIntScalarType.to_py(_nfac)
    _isf = utils._EngineIntArrayType(
        isf, {'fun_name': fun_name, 'entity': 'isf'},
        exp_shape=(_nfac_py,),
    ).to_ctypes()
    _lfac = utils._EngineIntArrayType(
        lfac, {'fun_name': fun_name, 'entity': 'lfac'},
        exp_shape=(_nfac_py,),
    ).to_ctypes()
    _ldf_py = utils._EngineIntArrayType.get_nd(ifac, {'fun_name': fun_name, 'entity': 'ifac'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldf = utils._EngineIntScalarType(
        _ldf_py, {'fun_name': fun_name, 'entity': '_ldf'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _ifac = utils._EngineIntArrayType(
        ifac, {'fun_name': fun_name, 'entity': 'ifac'},
        exp_shape=(_n_py, _nfac_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _weight_py = utils._EngineCharScalarType.to_py(_weight)
    if _weight_py in ('W', 'w', 'V', 'v'):
        _wt_dim1 = _n_py
    else:
        _wt_dim1 = 1
    exp_shape_wt = (_wt_dim1,)
    _wt = utils._EngineFloat64ArrayType(
        wt, {'fun_name': fun_name, 'entity': 'wt'},
        exp_shape=exp_shape_wt,
    ).to_ctypes()
    _maxt = utils._EngineIntScalarType(
        maxt, {'fun_name': fun_name, 'entity': 'maxt'},
    ).to_ctypes()
    _maxt_py = utils._EngineIntScalarType.to_py(_maxt)
    _table_cls = utils._EngineFloat64ArrayType(
        table, {'fun_name': fun_name, 'entity': 'table'},
        exp_shape=(_maxt_py,),
    )
    _table = _table_cls.to_ctypes()
    _ncells = utils._EngineIntScalarType(
        ncells, {'fun_name': fun_name, 'entity': 'ncells'},
    ).to_ctypes()
    _ndim = utils._EngineIntScalarType.empty_ctypes()
    _idim_cls = utils._EngineIntArrayType(
        idim, {'fun_name': fun_name, 'entity': 'idim'},
        exp_shape=(_nfac_py,),
        check_contents=False,
    )
    _idim = _idim_cls.to_ctypes(can_reference_contents=False)
    _icount_cls = utils._EngineIntArrayType(
        icount, {'fun_name': fun_name, 'entity': 'icount'},
        exp_shape=(_maxt_py,),
    )
    _icount = _icount_cls.to_ctypes()
    _stat_py = utils._EngineCharScalarType.to_py(_stat)
    if _stat_py in ('A', 'a'):
        _auxt_dim1 = _maxt_py
    elif _stat_py in ('V', 'v'):
        _auxt_dim1 = 2*_maxt_py
    else:
        _auxt_dim1 = 0
    exp_shape_auxt = (_auxt_dim1,)
    _auxt_cls = utils._EngineFloat64ArrayType(
        auxt, {'fun_name': fun_name, 'entity': 'auxt'},
        exp_shape=exp_shape_auxt,
    )
    _auxt = _auxt_cls.to_ctypes()
    _iwk = utils._EngineIntArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _stat,
        _update,
        _weight,
        _n,
        _nfac,
        _isf,
        _lfac,
        _ifac,
        _ldf,
        _y,
        _wt,
        _table,
        _maxt,
        _ncells,
        _ndim,
        _idim,
        _icount,
        _auxt,
        _iwk,
        _errbuf,
        _ierr,
        _stat._length_,
        _update._length_,
        _weight._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, weight = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: weight = \'U\', \'V\' or \'W\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, update = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: update = \'I\' or \'U\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, stat = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: stat = \'N\', \'T\', \'A\', \'V\', \'L\' or \'S\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nfac = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nfac >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99985:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, maxt = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'minimum value for maxt = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: maxt >= product of the levels of the',
                    'factors included in the tabulation.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s} and wt[i-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: wt[i-1] >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, j = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'lfac[j-1] = {:s} and ifac[i-1,j-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[4]),
                    ),
                    'Constraint: ifac[i-1,j-1] <= lfac[j-1].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, j = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'ifac[i-1,j-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: ifac[i-1,j-1] > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s} and lfac[i-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'On entry, lfac[i-1] > 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The variance divisor <= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99986:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'auxt or table has changed between calls.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99987:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'auxt has changed between calls.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99988:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'ncells has changed between calls.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _table_cls.refresh_py(_table)
        _ncells_py = utils._EngineIntScalarType.to_py(_ncells)
        _idim_cls.refresh_py(_idim)
        _icount_cls.refresh_py(_icount)
        _auxt_cls.refresh_py(_auxt)
        _ndim_py = utils._EngineIntScalarType.to_py(_ndim)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['ncells', 'ndim'],
        )(
            _ncells_py,
            _ndim_py,
        )
    _err_class.manage()
    return return_data

def tabulate_percentile(typ, n, nfac, isf, lfac, ifac, percnt, y, wt, table, maxt, idim, icount):
    u"""
    Computes multiway table from set of classification factors using
    given percentile/quantile.

    ``tabulate_percentile`` computes a table from a set of
    classification factors using a given percentile or quantile, for
    example the median.

    For full information please refer to the NAG Library document for
    g11bb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g11/g11bbf.html

    Parameters
    ----------
    typ : str, length 1
        Indicates if the variable to be tabulated is discrete or
        continuous.

        `typ` = 'D'

            The percentiles are computed for a discrete variable.

        `typ` = 'C'

            The percentiles are computed for a continuous variable using
            linear interpolation.

    n : int
        The number of observations.

    nfac : int
        The number of classifying factors in `ifac`.

    isf : int, shape (`nfac`)
        Indicates which factors in `ifac` are to be used in the
        tabulation.

        If `isf`[i-1] > 0 the ith factor in `ifac` is included in the
        tabulation.

        Note that if `isf`[i-1] <= 0, for i = 1,2,...,`nfac` then the
        statistic for the whole sample is calculated and returned in a
        1*1 table.

    lfac : int, shape (`nfac`)
        The number of levels of the classifying factors in `ifac`.

    ifac : int, shape (`n`, `nfac`)
        The `nfac` coded classification factors for the `n`
        observations.

    percnt : float
        p, the percentile to be tabulated.

    y : float, shape (`n`)
        The variable to be tabulated.

    wt : None or float, shape (`n`)
        If weight = 'W', `wt` must contain the `n` weights. Otherwise
        `wt` is not referenced.

    table : float, shape (`maxt`), modified in place
        `On exit`: the computed table. The `ncells` cells of the table
        are stored so that for any two factors the index relating to the
        factor occurring later in `lfac` and `ifac` changes faster. For
        further details see [Further Comments].

    maxt : int
        The maximum size of the table to be computed.

    idim : int, shape (`nfac`), modified in place
        `On exit`: the first `ndim` elements contain the number of
        levels for the factors defining the table.

    icount : int, shape (`maxt`), modified in place
        `On exit`: a table containing the number of observations
        contributing to each cell of the table, stored identically to
        `table`.

    Returns
    -------
    ncells : int
        The number of cells in the table.

    ndim : int
        The number of factors defining the table.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `percnt` = *<value>*.

            Constraint: 0.0 < p < 100.0.

        (`errno` 1)
            On entry, `typ` = *<value>*.

            Constraint: `typ` = 'D' or 'C'.

        (`errno` 1)
            On entry, `nfac` = *<value>*.

            Constraint: `nfac` >= 1.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 2.

        (`errno` 2)
            On entry, `maxt` = *<value>* and minimum value for `maxt` =
            *<value>*.

            Constraint: `maxt` >= product of the levels of the factors
            included in the tabulation.

        (`errno` 2)
            On entry, i = *<value>* and `wt`[i-1] = *<value>*.

            Constraint: `wt`[i-1] >= 0.0.

        (`errno` 2)
            On entry, i = *<value>*, j = *<value>*, `lfac`[j-1] =
            *<value>* and `ifac`[i-1,j-1] = *<value>*.

            Constraint: `ifac`[i-1,j-1] <= `lfac`[j-1].

        (`errno` 2)
            On entry, i = *<value>*, j = *<value>* and `ifac`[i-1,j-1] =
            *<value>*.

            Constraint: `ifac`[i-1,j-1] > 0.

        (`errno` 2)
            On entry, i = *<value>* and `lfac`[i-1] = *<value>*.

            On entry, `lfac`[i-1] > 2.

        (`errno` 3)
            Some cells of the table are empty.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    John, J A and Quenouille, M H, 1977, `Experiments: Design and
    Analysis`, Griffin

    Kendall, M G and Stuart, A, 1969, `The Advanced Theory of Statistics
    (Volume 1)`, (3rd Edition), Griffin
    """

    excs = []
    fun_name = 'naginterfaces.base.contab.tabulate_percentile'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g11bbft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, typ(len=1).
        utils._BYREF_STRING, # Input, weight(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nfac.
        utils._EngineIntArrayType.c_type, # Input, isf[nfac].
        utils._EngineIntArrayType.c_type, # Input, lfac[nfac].
        utils._EngineIntArrayType.c_type, # Input, ifac[ldf, nfac].
        utils._BYREF_INT, # Input, ldf.
        utils._BYREF_FLOAT64, # Input, percnt.
        utils._EngineFloat64ArrayType.c_type, # Input, y[n].
        utils._EngineFloat64ArrayType.c_type, # Input, wt[n].
        utils._EngineFloat64ArrayType.c_type, # Output, table[maxt].
        utils._BYREF_INT, # Input, maxt.
        utils._BYREF_INT, # Output, ncells.
        utils._BYREF_INT, # Output, ndim.
        utils._EngineIntArrayType.c_type, # Output, idim[nfac].
        utils._EngineIntArrayType.c_type, # Output, icount[maxt].
        utils._EngineIntArrayType.c_type, # Workspace, iwk.
        utils._EngineFloat64ArrayType.c_type, # Workspace, wk.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of typ.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of weight.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((ifac, 'ifac'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _typ = utils._EngineCharScalarType(
        typ, {'fun_name': fun_name, 'entity': 'typ'}, exp_chrlen=1,
    ).to_ctypes()
    if wt is None:
        _weight = utils._EngineCharScalarType(
            'U', {'fun_name': fun_name, 'entity': '_weight'}, exp_chrlen=1,
        ).to_ctypes()
    else:
        _weight = utils._EngineCharScalarType(
            'W', {'fun_name': fun_name, 'entity': '_weight'}, exp_chrlen=1,
        ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nfac = utils._EngineIntScalarType(
        nfac, {'fun_name': fun_name, 'entity': 'nfac'},
    ).to_ctypes()
    _nfac_py = utils._EngineIntScalarType.to_py(_nfac)
    _isf = utils._EngineIntArrayType(
        isf, {'fun_name': fun_name, 'entity': 'isf'},
        exp_shape=(_nfac_py,),
    ).to_ctypes()
    _lfac = utils._EngineIntArrayType(
        lfac, {'fun_name': fun_name, 'entity': 'lfac'},
        exp_shape=(_nfac_py,),
    ).to_ctypes()
    _ldf_py = utils._EngineIntArrayType.get_nd(ifac, {'fun_name': fun_name, 'entity': 'ifac'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldf = utils._EngineIntScalarType(
        _ldf_py, {'fun_name': fun_name, 'entity': '_ldf'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _ifac = utils._EngineIntArrayType(
        ifac, {'fun_name': fun_name, 'entity': 'ifac'},
        exp_shape=(_n_py, _nfac_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _percnt = utils._EngineFloat64ScalarType(
        percnt, {'fun_name': fun_name, 'entity': 'percnt'},
    ).to_ctypes()
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    if wt is None:
        _wt = utils._EngineFloat64ArrayType.empty_ctypes(_n_py)
    else:
        _wt = utils._EngineFloat64ArrayType(
            wt, {'fun_name': fun_name, 'entity': 'wt'},
            exp_shape=(_n_py,),
            allow_none=True,
        ).to_ctypes()
    _maxt = utils._EngineIntScalarType(
        maxt, {'fun_name': fun_name, 'entity': 'maxt'},
    ).to_ctypes()
    _maxt_py = utils._EngineIntScalarType.to_py(_maxt)
    _table_cls = utils._EngineFloat64ArrayType(
        table, {'fun_name': fun_name, 'entity': 'table'},
        exp_shape=(_maxt_py,),
        check_contents=False,
    )
    _table = _table_cls.to_ctypes(can_reference_contents=False)
    _ncells = utils._EngineIntScalarType.empty_ctypes()
    _ndim = utils._EngineIntScalarType.empty_ctypes()
    _idim_cls = utils._EngineIntArrayType(
        idim, {'fun_name': fun_name, 'entity': 'idim'},
        exp_shape=(_nfac_py,),
        check_contents=False,
    )
    _idim = _idim_cls.to_ctypes(can_reference_contents=False)
    _icount_cls = utils._EngineIntArrayType(
        icount, {'fun_name': fun_name, 'entity': 'icount'},
        exp_shape=(_maxt_py,),
        check_contents=False,
    )
    _icount = _icount_cls.to_ctypes(can_reference_contents=False)
    _iwk = utils._EngineIntArrayType.empty_ctypes(0)
    _wk = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _typ,
        _weight,
        _n,
        _nfac,
        _isf,
        _lfac,
        _ifac,
        _ldf,
        _percnt,
        _y,
        _wt,
        _table,
        _maxt,
        _ncells,
        _ndim,
        _idim,
        _icount,
        _iwk,
        _wk,
        _errbuf,
        _ierr,
        _typ._length_,
        _weight._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, percnt = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: 0.0 < p < 100.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, typ = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: typ = \'D\' or \'C\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nfac = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nfac >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, maxt = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'minimum value for maxt = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: maxt >= product of the levels of the',
                    'factors included in the tabulation.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s} and wt[i-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: wt[i-1] >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, j = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'lfac[j-1] = {:s} and ifac[i-1,j-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[4]),
                    ),
                    'Constraint: ifac[i-1,j-1] <= lfac[j-1].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, j = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'ifac[i-1,j-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: ifac[i-1,j-1] > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s} and lfac[i-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'On entry, lfac[i-1] > 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99988:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Some cells of the table are empty.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _table_cls.refresh_py(_table)
        _idim_cls.refresh_py(_idim)
        _icount_cls.refresh_py(_icount)
        _ncells_py = utils._EngineIntScalarType.to_py(_ncells)
        _ndim_py = utils._EngineIntScalarType.to_py(_ndim)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['ncells', 'ndim'],
        )(
            _ncells_py,
            _ndim_py,
        )
    _err_class.manage()
    return return_data

def tabulate_margin(stat, table, ncells, ndim, idim, isdim, stable, maxst, mlevel, auxt):
    u"""
    Computes marginal tables for multiway table computed by
    ``tabulate_stat`` or ``tabulate_percentile``.

    ``tabulate_margin`` computes a marginal table from a table computed
    by ``tabulate_stat`` or ``tabulate_percentile`` using a selected
    statistic.

    For full information please refer to the NAG Library document for
    g11bc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g11/g11bcf.html

    Parameters
    ----------
    stat : str, length 1
        Indicates which statistic is to be used to compute the marginal
        table.

        `stat` = 'T'

            The total.

        `stat` = 'A'

            The average or mean.

        `stat` = 'M'

            The median.

        `stat` = 'V'

            The variance.

        `stat` = 'L'

            The largest value.

        `stat` = 'S'

            The smallest value.

    table : float, shape (`ncells`)
        The table as computed by ``tabulate_stat`` or
        ``tabulate_percentile``.

    ncells : int
        The number of cells in `table` as returned by ``tabulate_stat``
        or ``tabulate_percentile``.

    ndim : int
        The number of dimensions for `table` as returned by
        ``tabulate_stat`` or ``tabulate_percentile``.

    idim : int, shape (`ndim`)
        The number of levels for each dimension of `table` as returned
        by ``tabulate_stat`` or ``tabulate_percentile``.

    isdim : int, shape (`ndim`)
        Indicates which dimensions of `table` are to be included in the
        sub-table. If `isdim`[i-1] > 0 the dimension or factor indicated
        by `idim`[i-1] is to be included in the sub-table, otherwise it
        is excluded.

    stable : float, shape (`maxst`), modified in place
        `On exit`: the first `mcells` elements contain the sub-table
        computed using the statistic indicated by `stat`. The table is
        stored in a similar way to `table` with the `mcells` cells
        stored so that for any two dimensions the index relating to the
        dimension given later in `idim` changes faster. For further
        details see [Further Comments].

    maxst : int
        The maximum size of sub-table to be computed.

    mlevel : int, shape (`ndim`), modified in place
        `On exit`: the first `mdim` elements contain the number of
        levels for the dimensions of the sub-table in `stable`. The
        remaining elements are not referenced.

    auxt : float, shape (:), modified in place
        Note: the required length for this argument is determined as
        follows: if `stat`='V': `maxst`; otherwise: 1.

        `On exit`: if `stat` = 'V' `auxt` contains the sub-table of
        means corresponding to the sub-table of variances in `stable`.
        Otherwise `auxt` is not referenced.

    Returns
    -------
    mcells : int
        The number of cells in the sub-table in `stable`.

    mdim : int
        The number of dimensions to the sub-table in `stable`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `stat` = *<value>*.

            Constraint: `stat` = 'T', 'A', 'M', 'V', 'L' or 'S'.

        (`errno` 1)
            On entry, `ndim` = *<value>*.

            Constraint: `ndim` >= 2.

        (`errno` 2)
            On entry, `maxst` = *<value>* and minimum value for `maxst`
            = *<value>*.

            Constraint: `maxst` >= product of the levels of the
            dimensions of the `table` included in the sub-table,
            `stable`.

        (`errno` 2)
            On entry, all elements of `isdim` > 0.

        (`errno` 2)
            On entry, no elements of `isdim` > 0.

        (`errno` 2)
            On entry, `ncells` is incompatible with `idim`.

        (`errno` 2)
            On entry, i = *<value>* and `idim`[i-1] = *<value>*.

            Constraint: `idim` >= 2.

    References
    ----------
    John, J A and Quenouille, M H, 1977, `Experiments: Design and
    Analysis`, Griffin

    Kendall, M G and Stuart, A, 1969, `The Advanced Theory of Statistics
    (Volume 1)`, (3rd Edition), Griffin

    West, D H D, 1979, `Updating mean and variance estimates: An
    improved method`, Comm. ACM (22), 532--555
    """

    excs = []
    fun_name = 'naginterfaces.base.contab.tabulate_margin'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g11bcft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, stat(len=1).
        utils._EngineFloat64ArrayType.c_type, # Input, table[ncells].
        utils._BYREF_INT, # Input, ncells.
        utils._BYREF_INT, # Input, ndim.
        utils._EngineIntArrayType.c_type, # Input, idim[ndim].
        utils._EngineIntArrayType.c_type, # Input, isdim[ndim].
        utils._EngineFloat64ArrayType.c_type, # Output, stable[maxst].
        utils._BYREF_INT, # Input, maxst.
        utils._BYREF_INT, # Output, mcells.
        utils._BYREF_INT, # Output, mdim.
        utils._EngineIntArrayType.c_type, # Output, mlevel[ndim].
        utils._EngineFloat64ArrayType.c_type, # Output, auxt[*].
        utils._EngineIntArrayType.c_type, # Workspace, iwk.
        utils._EngineFloat64ArrayType.c_type, # Workspace, wk.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of stat.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _stat = utils._EngineCharScalarType(
        stat, {'fun_name': fun_name, 'entity': 'stat'}, exp_chrlen=1,
    ).to_ctypes()
    _ncells = utils._EngineIntScalarType(
        ncells, {'fun_name': fun_name, 'entity': 'ncells'},
    ).to_ctypes()
    _ncells_py = utils._EngineIntScalarType.to_py(_ncells)
    _table = utils._EngineFloat64ArrayType(
        table, {'fun_name': fun_name, 'entity': 'table'},
        exp_shape=(_ncells_py,),
    ).to_ctypes()
    _ndim = utils._EngineIntScalarType(
        ndim, {'fun_name': fun_name, 'entity': 'ndim'},
    ).to_ctypes()
    _ndim_py = utils._EngineIntScalarType.to_py(_ndim)
    _idim = utils._EngineIntArrayType(
        idim, {'fun_name': fun_name, 'entity': 'idim'},
        exp_shape=(_ndim_py,),
    ).to_ctypes()
    _isdim = utils._EngineIntArrayType(
        isdim, {'fun_name': fun_name, 'entity': 'isdim'},
        exp_shape=(_ndim_py,),
    ).to_ctypes()
    _maxst = utils._EngineIntScalarType(
        maxst, {'fun_name': fun_name, 'entity': 'maxst'},
    ).to_ctypes()
    _maxst_py = utils._EngineIntScalarType.to_py(_maxst)
    _stable_cls = utils._EngineFloat64ArrayType(
        stable, {'fun_name': fun_name, 'entity': 'stable'},
        exp_shape=(_maxst_py,),
        check_contents=False,
    )
    _stable = _stable_cls.to_ctypes(can_reference_contents=False)
    _mcells = utils._EngineIntScalarType.empty_ctypes()
    _mdim = utils._EngineIntScalarType.empty_ctypes()
    _mlevel_cls = utils._EngineIntArrayType(
        mlevel, {'fun_name': fun_name, 'entity': 'mlevel'},
        exp_shape=(_ndim_py,),
        check_contents=False,
    )
    _mlevel = _mlevel_cls.to_ctypes(can_reference_contents=False)
    _stat_py = utils._EngineCharScalarType.to_py(_stat)
    if _stat_py in ('V', 'v'):
        _auxt_dim1 = _maxst_py
    else:
        _auxt_dim1 = 1
    exp_shape_auxt = (_auxt_dim1,)
    _auxt_cls = utils._EngineFloat64ArrayType(
        auxt, {'fun_name': fun_name, 'entity': 'auxt'},
        exp_shape=exp_shape_auxt,
        check_contents=False,
    )
    _auxt = _auxt_cls.to_ctypes(can_reference_contents=False)
    _iwk = utils._EngineIntArrayType.empty_ctypes(0)
    _wk = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _stat,
        _table,
        _ncells,
        _ndim,
        _idim,
        _isdim,
        _stable,
        _maxst,
        _mcells,
        _mdim,
        _mlevel,
        _auxt,
        _iwk,
        _wk,
        _errbuf,
        _ierr,
        _stat._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, stat = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: stat = \'T\', \'A\', \'M\', \'V\', \'L\' or \'S\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ndim = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ndim >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, maxst = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'minimum value for maxst = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: maxst >= product of the levels of the',
                    'dimensions of the table included in the sub-table, stable.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, all elements of isdim > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, no elements of isdim > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ncells is incompatible with idim.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s} and idim[i-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: idim >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _stable_cls.refresh_py(_stable)
        _mlevel_cls.refresh_py(_mlevel)
        _auxt_cls.refresh_py(_auxt)
        _mcells_py = utils._EngineIntScalarType.to_py(_mcells)
        _mdim_py = utils._EngineIntScalarType.to_py(_mdim)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['mcells', 'mdim'],
        )(
            _mcells_py,
            _mdim_py,
        )
    _err_class.manage()
    return return_data

def condl_logistic(n, m, ns, z, isz, ip, ic, isi, b, se, sc, cov, nca, nct, tol, maxit, iprint, io_manager=None):
    u"""
    Returns parameter estimates for the conditional analysis of
    stratified data.

    ``condl_logistic`` returns parameter estimates for the conditional
    logistic analysis of stratified data, for example, data from
    case-control studies and survival analyses.

    For full information please refer to the NAG Library document for
    g11ca

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g11/g11caf.html

    Parameters
    ----------
    n : int
        n, the number of observations.

    m : int
        The number of covariates in array `z`.

    ns : int
        The number of strata, n_s.

    z : float, shape (`n`, `m`)
        The ith row must contain the covariates which are associated
        with the ith observation.

    isz : int, shape (`m`)
        Indicates which subset of covariates are to be included in the
        model.

        If `isz`[j-1] >= 1, the jth covariate is included in the model.

        If `isz`[j-1] = 0, the jth covariate is excluded from the model
        and not referenced.

    ip : int
        p, the number of covariates included in the model as indicated
        by `isz`.

    ic : int, shape (`n`)
        Indicates whether the ith observation is a case or a control.

        If `ic`[i-1] = 0, indicates that the ith observation is a case.

        If `ic`[i-1] = 1, indicates that the ith observation is a
        control.

    isi : int, shape (`n`)
        Stratum indicators which also allow data points to be excluded
        from the analysis.

        If `isi`[i-1] = k, indicates that the ith observation is from
        the kth stratum, where k = 1,2,...,`ns`.

        If `isi`[i-1] = 0, indicates that the ith observation is to be
        omitted from the analysis.

    b : float, shape (`ip`), modified in place
        `On entry`: initial estimates of the covariate coefficient
        parameters beta. `b`[j-1] must contain the initial estimate of
        the coefficent of the covariate in `z` corresponding to the jth
        nonzero value of `isz`.

        `On exit`: `b`[j-1] contains the estimate beta^_i of the
        coefficient of the covariate stored in the ith column of `z`
        where i is the jth nonzero value in the array `isz`.

        Suggested value: in many cases an initial value of zero for
        `b`[j-1] may be used. For another suggestion see [Further
        Comments].

    se : float, shape (`ip`), modified in place
        `On exit`: `se`[j-1] is the asymptotic standard error of the
        estimate contained in `b`[j-1] and score function in `sc`[j-1],
        for j = 1,2,...,`ip`.

    sc : float, shape (`ip`), modified in place
        `On exit`: `sc`[j] is the value of the score function U_j(beta)
        for the estimate contained in `b`[j-1].

    cov : float, shape (`ip`*(`ip`+1)/2), modified in place
        `On exit`: the variance-covariance matrix of the parameter
        estimates in `b` stored in packed form by column, i.e., the
        covariance between the parameter estimates given in `b`[i-1] and
        `b`[j-1], j >= i, is given in `cov`[j(j-1)/2+i].

    nca : int, shape (`ns`), modified in place
        `On exit`: `nca`[i-1] contains the number of cases in the ith
        stratum, for i = 1,2,...,`ns`.

    nct : int, shape (`ns`), modified in place
        `On exit`: `nct`[i-1] contains the number of controls in the ith
        stratum, for i = 1,2,...,`ns`.

    tol : float
        Indicates the accuracy required for the estimation. Convergence
        is assumed when the decrease in deviance is less than
        `tol`*(1.0+CurrentDeviance). This corresponds approximately to
        an absolute accuracy if the deviance is small and a relative
        accuracy if the deviance is large.

    maxit : int
        The maximum number of iterations required for computing the
        estimates. If `maxit` is set to 0 then the standard errors, the
        score functions and the variance-covariance matrix are computed
        for the input value of beta in `b` but beta is not updated.

    iprint : int
        Indicates if the printing of information on the iterations is
        required.

        `iprint` <= 0

            No printing.

        `iprint` >= 1

            The deviance and the current estimates are printed every
            `iprint` iterations. When printing occurs the output is
            directed to the file object associated with the advisory I/O
            unit (see
            :class:`~naginterfaces.base.utils.FileObjManager`).

        Suggested value: `iprint` = 0.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    dev : float
        The deviance, that is, minus twice the maximized log-likelihood.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `tol` = *<value>*.

            Constraint: `tol` >= 10*machine precision.

        (`errno` 1)
            On entry, `ip` = *<value>*.

            Constraint: `ip` >= 1.

        (`errno` 1)
            On entry, `maxit` = *<value>*.

            Constraint: `maxit` >= 0.

        (`errno` 1)
            On entry, `ns` = *<value>*.

            Constraint: `ns` >= 1.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 2.

        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 2)
            On entry, there are not `ip` values of `isz` > 0.

        (`errno` 2)
            On entry, i = *<value>* and `isz`[i-1] < *<value>*.

            Constraint: `isz`[i-1] >= 0.

        (`errno` 2)
            On entry, too few observations included in model.

        (`errno` 2)
            On entry, i = *<value>*, `isi`[i-1] = *<value>* and `ns` =
            *<value>*.

            Constraint: 0 <= `isi`[i-1] <= `ns`.

        (`errno` 2)
            On entry, i = *<value>* and `ic`[i-1] = *<value>*.

            Constraint: `ic`[i] = 0 or 1.

        (`errno` 4)
            Overflow in calculations.

        (`errno` 5)
            The matrix of second partial derivatives is singular.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 6)
            Convergence not achieved in *<value>* iterations.

    References
    ----------
    Cox, D R, 1972, `Regression models in life tables (with
    discussion)`, J. Roy. Statist. Soc. Ser. B (34), 187--220

    Cox, D R and Hinkley, D V, 1974, `Theoretical Statistics`, Chapman
    and Hall

    Howard, S, 1972, `Remark on the paper by Cox, D R (1972): Regression
    methods`, J. R. Statist. Soc. (B 34), and life tables, 187--220

    Krailo, M D and Pike, M C, 1984, `Algorithm AS 196. Conditional
    multivariate logistic analysis of stratified case-control studies`,
    Appl. Statist. (33), 95--103

    Smith, P G, Pike, M C, Hill, P, Breslow, N E and Day, N E, 1981,
    `Algorithm AS 162. Multivariate conditional logistic analysis of
    stratum-matched case-control studies`, Appl. Statist. (30), 190--197
    """

    excs = []
    fun_name = 'naginterfaces.base.contab.condl_logistic'

    utils.FileObjManager._check_type_compat(
        io_manager, {'fun_name': fun_name, 'entity': 'io_manager'}
    )

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    def cb_print_rech_py(_, en_data, nout, rec, ifail, rec_len):
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data, utils._EngineIntScalarType.to_py(nout.contents), rec[:rec_len], io_manager, excs, ifail.contents,
                fun_name=fun_name,
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g11caft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._PRINT_REC_FUNCTYPE,
        utils._PRINT_RECH_FUNCTYPE,
        utils._BYREF_INT, # Input, iemode.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, ns.
        utils._EngineFloat64ArrayType.c_type, # Input, z[ldz, m].
        utils._BYREF_INT, # Input, ldz.
        utils._EngineIntArrayType.c_type, # Input, isz[m].
        utils._BYREF_INT, # Input, ip.
        utils._EngineIntArrayType.c_type, # Input, ic[n].
        utils._EngineIntArrayType.c_type, # Input, isi[n].
        utils._BYREF_FLOAT64, # Output, dev.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, b[ip].
        utils._EngineFloat64ArrayType.c_type, # Output, se[ip].
        utils._EngineFloat64ArrayType.c_type, # Output, sc[ip].
        utils._EngineFloat64ArrayType.c_type, # Output, cov[ip*(ip+1)//2].
        utils._EngineIntArrayType.c_type, # Output, nca[ns].
        utils._EngineIntArrayType.c_type, # Output, nct[ns].
        utils._BYREF_FLOAT64, # Input, tol.
        utils._BYREF_INT, # Input, maxit.
        utils._BYREF_INT, # Input, iprint.
        utils._BYREF_INT, # Input, nout.
        utils._EngineFloat64ArrayType.c_type, # Workspace, wk.
        utils._BYREF_INT, # Input, lwk.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((z, 'z'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _print_rech = utils._PRINT_RECH_FUNCTYPE(cb_print_rech_py)
    _iemode = utils._EngineIntScalarType.c_type(0)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _ns = utils._EngineIntScalarType(
        ns, {'fun_name': fun_name, 'entity': 'ns'},
    ).to_ctypes()
    _ldz_py = utils._EngineFloat64ArrayType.get_nd(z, {'fun_name': fun_name, 'entity': 'z'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldz = utils._EngineIntScalarType(
        _ldz_py, {'fun_name': fun_name, 'entity': '_ldz'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _z = utils._EngineFloat64ArrayType(
        z, {'fun_name': fun_name, 'entity': 'z'},
        exp_shape=(_n_py, _m_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _isz = utils._EngineIntArrayType(
        isz, {'fun_name': fun_name, 'entity': 'isz'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _ip = utils._EngineIntScalarType(
        ip, {'fun_name': fun_name, 'entity': 'ip'},
    ).to_ctypes()
    _ic = utils._EngineIntArrayType(
        ic, {'fun_name': fun_name, 'entity': 'ic'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _isi = utils._EngineIntArrayType(
        isi, {'fun_name': fun_name, 'entity': 'isi'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _dev = utils._EngineFloat64ScalarType.empty_ctypes()
    _ip_py = utils._EngineIntScalarType.to_py(_ip)
    _b_cls = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_ip_py,),
    )
    _b = _b_cls.to_ctypes()
    _se_cls = utils._EngineFloat64ArrayType(
        se, {'fun_name': fun_name, 'entity': 'se'},
        exp_shape=(_ip_py,),
        check_contents=False,
    )
    _se = _se_cls.to_ctypes(can_reference_contents=False)
    _sc_cls = utils._EngineFloat64ArrayType(
        sc, {'fun_name': fun_name, 'entity': 'sc'},
        exp_shape=(_ip_py,),
        check_contents=False,
    )
    _sc = _sc_cls.to_ctypes(can_reference_contents=False)
    _cov_cls = utils._EngineFloat64ArrayType(
        cov, {'fun_name': fun_name, 'entity': 'cov'},
        exp_shape=(_ip_py*(_ip_py+1)//2,),
        check_contents=False,
    )
    _cov = _cov_cls.to_ctypes(can_reference_contents=False)
    _ns_py = utils._EngineIntScalarType.to_py(_ns)
    _nca_cls = utils._EngineIntArrayType(
        nca, {'fun_name': fun_name, 'entity': 'nca'},
        exp_shape=(_ns_py,),
        check_contents=False,
    )
    _nca = _nca_cls.to_ctypes(can_reference_contents=False)
    _nct_cls = utils._EngineIntArrayType(
        nct, {'fun_name': fun_name, 'entity': 'nct'},
        exp_shape=(_ns_py,),
        check_contents=False,
    )
    _nct = _nct_cls.to_ctypes(can_reference_contents=False)
    _tol = utils._EngineFloat64ScalarType(
        tol, {'fun_name': fun_name, 'entity': 'tol'},
    ).to_ctypes()
    _maxit = utils._EngineIntScalarType(
        maxit, {'fun_name': fun_name, 'entity': 'maxit'},
    ).to_ctypes()
    _iprint = utils._EngineIntScalarType(
        iprint, {'fun_name': fun_name, 'entity': 'iprint'},
    ).to_ctypes()
    _nout = utils._EngineIntScalarType.c_type(utils.FileObjManager.advunit)
    _wk = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _lwk = utils._EngineIntScalarType.c_type(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        utils._UNUSED_PRINT_REC,
        _print_rech,
        _iemode,
        _n,
        _m,
        _ns,
        _z,
        _ldz,
        _isz,
        _ip,
        _ic,
        _isi,
        _dev,
        _b,
        _se,
        _sc,
        _cov,
        _nca,
        _nct,
        _tol,
        _maxit,
        _iprint,
        _nout,
        _wk,
        _lwk,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, tol = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: tol >= 10*machine precision.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ip >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, maxit = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: maxit >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ns = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ns >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99985:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, there are not ip values of isz > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99986:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s} and isz[i-1] < {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: isz[i-1] >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99987:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, too few observations included in model.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, isi[i-1] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'ns = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: 0 <= isi[i-1] <= ns.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s} and ic[i-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: ic[i] = 0 or 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99984:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Overflow in calculations.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The matrix of second partial derivatives is singular.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99988:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Convergence not achieved in {:s} iterations.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _b_cls.refresh_py(_b)
        _se_cls.refresh_py(_se)
        _sc_cls.refresh_py(_sc)
        _cov_cls.refresh_py(_cov)
        _nca_cls.refresh_py(_nca)
        _nct_cls.refresh_py(_nct)
        _dev_py = utils._EngineFloat64ScalarType.to_py(_dev)
        return_data = _dev_py
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = _collections.namedtuple(
                'nag_return_data',
                ['dev'],
            )(
                _dev_py,
            )
    _err_class.manage()
    return return_data

def binary(ip, n, gprob, ns, x, irl, a, c, iprint, cgetol, maxit, chisqr, ishow, alpha, pigam, cm, g, expp, obs, exf, y, xl, iob, io_manager=None):
    u"""
    Contingency table, latent variable model for binary data.

    ``binary`` fits a latent variable model (with a single factor) to
    data consisting of a set of measurements on individuals in the form
    of binary-valued sequences (generally referred to as score
    patterns).
    Various measures of goodness-of-fit are calculated along with the
    factor (theta) scores.

    For full information please refer to the NAG Library document for
    g11sa

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g11/g11saf.html

    Parameters
    ----------
    ip : int
        p, the number of dichotomous variables.

    n : int
        n, the number of individuals in the sample.

    gprob : bool
        Must be set equal to True if G(z) = Phi^-1(z) and False if G(z)
        = logit(z).

    ns : int
        `ns` must be set equal to the number of different score patterns
        in the sample, s.

    x : bool, shape (`ns`, `ip`), modified in place
        `On entry`: the first s rows of `x` must contain the s different
        score patterns. The lth row of `x` must contain the lth score
        pattern with `x`[l-1,j-1] set equal to True if x_{lj} = 1 and
        False if x_{lj} = 0. All rows of `x` must be distinct.

        `On exit`: given a valid parameter set then the first s rows of
        `x` still contain the s different score patterns. However, the
        following points should be noted:

        (i) If the estimated factor loading for the jth item is negative
            then that item is re-coded, i.e., 0s and 1s (or True and
            False) in the jth column of `x` are interchanged.

        (#) The rows of `x` will be reordered so that the theta scores
            corresponding to rows of `x` are in increasing order of
            magnitude.

    irl : int, shape (`ns`), modified in place
        `On entry`: the ith component of `irl` must be set equal to the
        frequency with which the ith row of `x` occurs.

        `On exit`: given a valid parameter set then the first s
        components of `irl` are reordered as are the rows of `x`.

    a : float, shape (`ip`), modified in place
        `On entry`: `a`[j-1] must be set equal to an initial estimate of
        alpha_{j1}. **In order to avoid divergence problems with the E-M
        algorithm you are strongly advised to set all the a[j-1] to
        0.5.**

        `On exit`: `a`[j-1] contains the latest estimate of alpha_{j1},
        for j = 1,2,...,p. (Because of possible recoding all elements of
        `a` will be positive.)

    c : float, shape (`ip`), modified in place
        `On entry`: `c`[j-1] must be set equal to an initial estimate of
        alpha_{j0}. **In order to avoid divergence problems with the E-M
        algorithm you are strongly advised to set all the c[j-1] to
        0.0.**

        `On exit`: `c`[j-1] contains the latest estimate of alpha_{j0},
        for j = 1,2,...,p.

    iprint : int
        The frequency with which the maximum likelihood search function
        is to be monitored.

        `iprint` > 0

            The search is monitored once every `iprint` iterations, and
            when the number of quadrature points is increased, and again
            at the final solution point.

        `iprint` = 0

            The search is monitored once at the final point.

        `iprint` < 0

            The search is not monitored at all.

        `iprint` should normally be set to a small positive number.

        Suggested value: `iprint` = 1.

    cgetol : float
        The accuracy to which the solution is required.

        If `cgetol` is set to 10^(-l) and on exit the function exits
        successfully or `errno` = 7, then all elements of the gradient
        vector will be smaller than 10^(-l) in absolute value.

        For most practical purposes the value 10^(-4) should suffice.

        You should be wary of setting `cgetol` too small since the
        convergence criterion may then have become too strict for the
        machine to handle.

        If `cgetol` has been set to a value which is less than the
        square root of the machine precision, epsilon, then ``binary``
        will use the value sqrt(epsilon) instead.

    maxit : int
        The maximum number of iterations to be made in the maximum
        likelihood search. There will be an error exit (see
        [Raises/Warns]) if the search function has not converged in
        `maxit` iterations.

        Suggested value: `maxit` = 1000.

    chisqr : bool
        If `chisqr` is set equal to True, a likelihood ratio statistic
        will be calculated (see `chi`).

        If `chisqr` is set equal to False, no such statistic will be
        calculated.

    ishow : int
        Indicates which of the following three quantities are to be
        printed before exit from the function (given a valid parameter
        set):

        (a) Table of maximum likelihood estimates and standard errors
            (as returned in the output arrays `a`, `c`, `alpha`, `pigam`
            and `cm`).

        (#) Table of observed and expected first - and second-order
            margins (as returned in the output arrays `expp` and `obs`).

        (#) Table of observed and expected frequencies of score patterns
            along with theta scores (as returned in the output arrays
            `irl`, `exf`, `y`, `xl` and `iob`) and the likelihood ratio
            statistic (if required).

        `ishow` = 0

            None of the above are printed.

        `ishow` = 1

            (a) only is printed.

        `ishow` = 2

            (b) only is printed.

        `ishow` = 3

            (c) only is printed.

        `ishow` = 4

            (a) and (b) are printed.

        `ishow` = 5

            (a) and (c) are printed.

        `ishow` = 6

            (b) and (c) are printed.

        `ishow` = 7

            (a), (b) and (c) are printed.

    alpha : float, shape (`ip`), modified in place
        `On exit`: given a valid parameter set then `alpha`[j-1]
        contains the latest estimate of alpha_j. (Because of possible
        recoding all elements of `alpha` will be positive.)

    pigam : float, shape (`ip`), modified in place
        `On exit`: given a valid parameter set then `pigam`[j-1]
        contains the latest estimate of either pi_j if `gprob` = False
        (logit model) or gamma_j if `gprob` = True (probit model).

    cm : float, shape (2*`ip`, 2*`ip`), modified in place
        `On exit`: given a valid parameter set then the strict lower
        triangle of `cm` contains the correlation matrix of the
        parameter estimates held in `alpha` and `pigam` on exit. The
        diagonal elements of `cm` contain the standard errors. Thus:

        [table omitted]

        for i = 1,2,...,p;

        [table omitted]

        for j = 1,2,...,i-1.

        If the second derivative matrix cannot be computed then all the
        elements of `cm` are returned as zero.

    g : float, shape (2*`ip`), modified in place
        `On exit`: given a valid parameter set then `g` contains the
        estimated gradient vector corresponding to the final point held
        in the arrays `alpha` and `pigam`. `g`[2*j-2] contains the
        derivative of the log-likelihood with respect to `alpha`[j-1],
        for j = 1,2,...,p. `g`[2*j-1] contains the derivative of the
        log-likelihood with respect to `pigam`[j-1], for j = 1,2,...,p.

    expp : float, shape (`ip`, `ip`), modified in place
        `On exit`: given a valid parameter set then `expp`[i-1,j-1]
        contains the expected percentage of individuals in the sample
        who respond positively to items i and j (j <= i), corresponding
        to the final point held in the arrays `alpha` and `pigam`.

    obs : float, shape (`ip`, `ip`), modified in place
        `On exit`: given a valid parameter set then `obs`[i-1,j-1]
        contains the observed percentage of individuals in the sample
        who responded positively to items i and j (j <= i).

    exf : float, shape (`ns`), modified in place
        `On exit`: given a valid parameter set then `exf`[l-1] contains
        the expected frequency of the lth score pattern (lth row of
        `x`), corresponding to the final point held in the arrays
        `alpha` and `pigam`.

    y : float, shape (`ns`), modified in place
        `On exit`: given a valid parameter set then `y`[l-1] contains
        the estimated theta score corresponding to the lth row of `x`,
        for the final point held in the arrays `alpha` and `pigam`.

    xl : float, shape (:), modified in place
        Note: the required length for this argument is determined as
        follows: if not((`gprob`): `ns`; otherwise: 0.

        `On exit`: if `gprob` has been set equal to False (logit model)
        then, given a valid parameter set, `xl`[l-1] contains the
        estimated component score corresponding to the lth row of `x`
        for the final point held in the arrays `alpha` and `pigam`.

        If `gprob` is set equal to True (probit model), this array is
        not used.

    iob : int, shape (`ns`), modified in place
        `On exit`: given a valid parameter set then `iob`[l-1] contains
        the number of items in the lth row of `x` for which the response
        was positive (True).

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    niter : int
        Given a valid parameter set then `niter` contains the number of
        iterations performed by the maximum likelihood search function.

    rlogl : float
        Given a valid parameter set then `rlogl` contains the value of
        the log-likelihood kernel corresponding to the final point held
        in the arrays `alpha` and `pigam`, namely

            sum_l = 0^s-1`irl`[l]*log(`exf`[l]/n).

    chi : float
        If `chisqr` was set equal to True on entry, then given a valid
        parameter set, `chi` will contain the value of the likelihood
        ratio statistic corresponding to the final parameter estimates
        held in the arrays `alpha` and `pigam`, namely

            2*sum_l = 0^s-1`irl`[l]*log(`exf`[l]/`irl`[l]).

        The summation is over those elements of `irl` which are
        positive. If `exf`[l-1] is less than 5.0, then adjacent score
        patterns are pooled (the score patterns in `x` being first put
        in order of increasing theta score).

        If `chisqr` has been set equal to False, then `chi` is not used.

    idf : int
        If `chisqr` was set equal to True on entry, then given a valid
        parameter set, `idf` will contain the degrees of freedom
        associated with the likelihood ratio statistic, `chi`.

        [table omitted]

        where s_0 denotes the number of terms summed to calculate `chi`
        (s_0 = s only if there is no pooling).

        If `chisqr` has been set equal to False, `idf` is not used.

    siglev : float
        If `chisqr` was set equal to True on entry, then given a valid
        parameter set, `siglev` will contain the significance level of
        `chi` based on `idf` degrees of freedom. If `idf` is zero or
        negative then `siglev` is set to zero.

        If `chisqr` was set equal to False, `siglev` is not used.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, sum_i(`irl`[i]) = *<value>* and `n` = *<value>*.

            Constraint: sum_i(`irl`[i]) = `n`.

        (`errno` 1)
            On entry, i = *<value>* and `irl`[i-1] = *<value>*.

            Constraint: `irl`[i-1] >= 0.

        (`errno` 1)
            On entry, i = *<value>* and j = *<value>*.

            Constraint: rows i and j of `x` should not be identical.

        (`errno` 1)
            On entry, `ns` = *<value>* and `ip` = *<value>*.

            Constraint: `ns` > 2*`ip`.

        (`errno` 1)
            On entry, `ip` = *<value>*.

            Constraint: `ip` >= 3.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 7.

        (`errno` 1)
            On entry, `ishow` = *<value>*.

            Constraint: `ishow` = 0, 1, 2, 3, 4, 5, 6 or 7.

        (`errno` 1)
            On entry, `ns` = *<value>* and `ip` = *<value>*.

            Constraint: `ns` <= 2^`ip`.

        (`errno` 1)
            On entry, `ns` = *<value>* and `n` = *<value>*.

            Constraint: `ns` <= `n`.

        (`errno` 1)
            On entry, `maxit` = *<value>*.

            Constraint: `maxit` >= 1.

        (`errno` 2)
            For at least one of the `ip` items the responses are all at
            the same level.

        (`errno` 3)
            `maxit` iterations have been performed: `maxit` = *<value>*.

        (`errno` 4)
            One of the elements of `a` has exceeded 10 in absolute.

        (`errno` 5)
            Failure to invert Hessian matrix and `maxit` iterations
            made: `maxit` = *<value>*.

        (`errno` 6)
            Failure to invert Hessian matrix plus Heywood case
            encountered.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 7)
            chi^2 statistic has less than one degree of freedom.

    References
    ----------
    Bartholomew, D J, 1980, `Factor analysis for categorical data (with
    Discussion)`, J. Roy. Statist. Soc. Ser. B (42), 293--321

    Bartholomew, D J, 1987, `Latent Variable Models and Factor
    Analysis`, Griffin

    Bock, R D and Aitkin, M, 1981, `Marginal maximum likelihood
    estimation of item parameters: Application of an E-M algorithm`,
    Psychometrika (46), 443--459
    """

    excs = []
    fun_name = 'naginterfaces.base.contab.binary'

    utils.FileObjManager._check_type_compat(
        io_manager, {'fun_name': fun_name, 'entity': 'io_manager'}
    )

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    def cb_print_rech_py(_, en_data, nout, rec, ifail, rec_len):
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data, utils._EngineIntScalarType.to_py(nout.contents), rec[:rec_len], io_manager, excs, ifail.contents,
                fun_name=fun_name,
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g11saft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._PRINT_REC_FUNCTYPE,
        utils._PRINT_RECH_FUNCTYPE,
        utils._BYREF_INT, # Input, iemode.
        utils._BYREF_INT, # Input, ip.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_LOGICAL, # Input, gprob.
        utils._BYREF_INT, # Input, ns.
        utils._EngineLogicalArrayType.c_type, # Input/Output, x[ldx, ip].
        utils._BYREF_INT, # Input, ldx.
        utils._EngineIntArrayType.c_type, # Input/Output, irl[ns].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, a[ip].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, c[ip].
        utils._BYREF_INT, # Input, iprint.
        utils._BYREF_FLOAT64, # Input, cgetol.
        utils._BYREF_INT, # Input, maxit.
        utils._BYREF_LOGICAL, # Input, chisqr.
        utils._BYREF_INT, # Input, ishow.
        utils._BYREF_INT, # Input, lp.
        utils._BYREF_INT, # Output, niter.
        utils._EngineFloat64ArrayType.c_type, # Output, alpha[ip].
        utils._EngineFloat64ArrayType.c_type, # Output, pigam[ip].
        utils._EngineFloat64ArrayType.c_type, # Output, cm[ldcm, 2*ip].
        utils._BYREF_INT, # Input, ldcm.
        utils._EngineFloat64ArrayType.c_type, # Output, g[2*ip].
        utils._EngineFloat64ArrayType.c_type, # Output, expp[ldexpp, ip].
        utils._BYREF_INT, # Input, ldexpp.
        utils._EngineFloat64ArrayType.c_type, # Output, obs[ldobs, ip].
        utils._BYREF_INT, # Input, ldobs.
        utils._EngineFloat64ArrayType.c_type, # Output, exf[ns].
        utils._EngineFloat64ArrayType.c_type, # Output, y[ns].
        utils._EngineFloat64ArrayType.c_type, # Output, xl[*].
        utils._EngineIntArrayType.c_type, # Output, iob[ns].
        utils._BYREF_FLOAT64, # Output, rlogl.
        utils._BYREF_FLOAT64, # Output, chi.
        utils._BYREF_INT, # Output, idf.
        utils._BYREF_FLOAT64, # Output, siglev.
        utils._EngineFloat64ArrayType.c_type, # Workspace, w.
        utils._BYREF_INT, # Input, lw.
        utils._BYREF_STRING, # Input, routine_name(len=6).
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of routine_name.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((x, 'x'), (cm, 'cm'), (expp, 'expp'), (obs, 'obs')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _print_rech = utils._PRINT_RECH_FUNCTYPE(cb_print_rech_py)
    _iemode = utils._EngineIntScalarType.c_type(0)
    _ip = utils._EngineIntScalarType(
        ip, {'fun_name': fun_name, 'entity': 'ip'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _gprob = utils._EngineLogicalScalarType(
        gprob, {'fun_name': fun_name, 'entity': 'gprob'},
    ).to_ctypes()
    _ns = utils._EngineIntScalarType(
        ns, {'fun_name': fun_name, 'entity': 'ns'},
    ).to_ctypes()
    _ldx_py = utils._EngineLogicalArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _ns_py = utils._EngineIntScalarType.to_py(_ns)
    _ip_py = utils._EngineIntScalarType.to_py(_ip)
    _x_cls = utils._EngineLogicalArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_ns_py, _ip_py),
        mutual_sorder=_m_sorder,
    )
    _x = _x_cls.to_ctypes()
    _irl_cls = utils._EngineIntArrayType(
        irl, {'fun_name': fun_name, 'entity': 'irl'},
        exp_shape=(_ns_py,),
    )
    _irl = _irl_cls.to_ctypes()
    _a_cls = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_ip_py,),
    )
    _a = _a_cls.to_ctypes()
    _c_cls = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_ip_py,),
    )
    _c = _c_cls.to_ctypes()
    _iprint = utils._EngineIntScalarType(
        iprint, {'fun_name': fun_name, 'entity': 'iprint'},
    ).to_ctypes()
    _cgetol = utils._EngineFloat64ScalarType(
        cgetol, {'fun_name': fun_name, 'entity': 'cgetol'},
    ).to_ctypes()
    _maxit = utils._EngineIntScalarType(
        maxit, {'fun_name': fun_name, 'entity': 'maxit'},
    ).to_ctypes()
    _chisqr = utils._EngineLogicalScalarType(
        chisqr, {'fun_name': fun_name, 'entity': 'chisqr'},
    ).to_ctypes()
    _ishow = utils._EngineIntScalarType(
        ishow, {'fun_name': fun_name, 'entity': 'ishow'},
    ).to_ctypes()
    _lp = utils._EngineIntScalarType.c_type(utils.FileObjManager.advunit)
    _niter = utils._EngineIntScalarType.empty_ctypes()
    _alpha_cls = utils._EngineFloat64ArrayType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
        exp_shape=(_ip_py,),
        check_contents=False,
    )
    _alpha = _alpha_cls.to_ctypes(can_reference_contents=False)
    _pigam_cls = utils._EngineFloat64ArrayType(
        pigam, {'fun_name': fun_name, 'entity': 'pigam'},
        exp_shape=(_ip_py,),
        check_contents=False,
    )
    _pigam = _pigam_cls.to_ctypes(can_reference_contents=False)
    _ldcm_py = utils._EngineFloat64ArrayType.get_nd(cm, {'fun_name': fun_name, 'entity': 'cm'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldcm = utils._EngineIntScalarType(
        _ldcm_py, {'fun_name': fun_name, 'entity': '_ldcm'},
    ).to_ctypes()
    _cm_cls = utils._EngineFloat64ArrayType(
        cm, {'fun_name': fun_name, 'entity': 'cm'},
        exp_shape=(2*_ip_py, 2*_ip_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _cm = _cm_cls.to_ctypes(can_reference_contents=False)
    _g_cls = utils._EngineFloat64ArrayType(
        g, {'fun_name': fun_name, 'entity': 'g'},
        exp_shape=(2*_ip_py,),
        check_contents=False,
    )
    _g = _g_cls.to_ctypes(can_reference_contents=False)
    _ldexpp_py = utils._EngineFloat64ArrayType.get_nd(expp, {'fun_name': fun_name, 'entity': 'expp'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldexpp = utils._EngineIntScalarType(
        _ldexpp_py, {'fun_name': fun_name, 'entity': '_ldexpp'},
    ).to_ctypes()
    _expp_cls = utils._EngineFloat64ArrayType(
        expp, {'fun_name': fun_name, 'entity': 'expp'},
        exp_shape=(_ip_py, _ip_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _expp = _expp_cls.to_ctypes(can_reference_contents=False)
    _ldobs_py = utils._EngineFloat64ArrayType.get_nd(obs, {'fun_name': fun_name, 'entity': 'obs'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldobs = utils._EngineIntScalarType(
        _ldobs_py, {'fun_name': fun_name, 'entity': '_ldobs'},
    ).to_ctypes()
    _obs_cls = utils._EngineFloat64ArrayType(
        obs, {'fun_name': fun_name, 'entity': 'obs'},
        exp_shape=(_ip_py, _ip_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _obs = _obs_cls.to_ctypes(can_reference_contents=False)
    _exf_cls = utils._EngineFloat64ArrayType(
        exf, {'fun_name': fun_name, 'entity': 'exf'},
        exp_shape=(_ns_py,),
        check_contents=False,
    )
    _exf = _exf_cls.to_ctypes(can_reference_contents=False)
    _y_cls = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_ns_py,),
        check_contents=False,
    )
    _y = _y_cls.to_ctypes(can_reference_contents=False)
    _gprob_py = utils._EngineLogicalScalarType.to_py(_gprob)
    if not _gprob_py:
        _xl_dim1 = _ns_py
    else:
        _xl_dim1 = 0
    exp_shape_xl = (_xl_dim1,)
    _xl_cls = utils._EngineFloat64ArrayType(
        xl, {'fun_name': fun_name, 'entity': 'xl'},
        exp_shape=exp_shape_xl,
        check_contents=False,
    )
    _xl = _xl_cls.to_ctypes(can_reference_contents=False)
    _iob_cls = utils._EngineIntArrayType(
        iob, {'fun_name': fun_name, 'entity': 'iob'},
        exp_shape=(_ns_py,),
        check_contents=False,
    )
    _iob = _iob_cls.to_ctypes(can_reference_contents=False)
    _rlogl = utils._EngineFloat64ScalarType.empty_ctypes()
    _chi = utils._EngineFloat64ScalarType.empty_ctypes()
    _idf = utils._EngineIntScalarType.empty_ctypes()
    _siglev = utils._EngineFloat64ScalarType.empty_ctypes()
    _w = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _lw = utils._EngineIntScalarType.c_type(0)
    _routine_name = utils._EngineRoutineName(*b'g11sa ')
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        utils._UNUSED_PRINT_REC,
        _print_rech,
        _iemode,
        _ip,
        _n,
        _gprob,
        _ns,
        _x,
        _ldx,
        _irl,
        _a,
        _c,
        _iprint,
        _cgetol,
        _maxit,
        _chisqr,
        _ishow,
        _lp,
        _niter,
        _alpha,
        _pigam,
        _cm,
        _ldcm,
        _g,
        _expp,
        _ldexpp,
        _obs,
        _ldobs,
        _exf,
        _y,
        _xl,
        _iob,
        _rlogl,
        _chi,
        _idf,
        _siglev,
        _w,
        _lw,
        _routine_name,
        _errbuf,
        _ierr,
        _routine_name._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99985:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, sum_i(irl[i]) = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: sum_i(irl[i]) = n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99986:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s} and irl[i-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: irl[i-1] >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99987:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s} and j = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: rows i and j of x should not be identical.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ns = {:s} and ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: ns > 2*ip.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ip >= 3.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 7.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ishow = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ishow = 0, 1, 2, 3, 4, 5, 6 or 7.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ns = {:s} and ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: ns <= 2^ip.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ns = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: ns <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, maxit = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: maxit >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99973:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'For at least one of the ip items the responses are all',
                    'at the same level.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99972:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'maxit iterations have been performed: maxit = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99970:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'One of the elements of a has exceeded 10 in absolute.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99969:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Failure to invert Hessian matrix and maxit iterations made:',
                    'maxit = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99968:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Failure to invert Hessian matrix plus Heywood case encountered.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 99971:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'chi^2 statistic has less than one degree of freedom.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _x_cls.refresh_py(_x)
        _irl_cls.refresh_py(_irl)
        _a_cls.refresh_py(_a)
        _c_cls.refresh_py(_c)
        _alpha_cls.refresh_py(_alpha)
        _pigam_cls.refresh_py(_pigam)
        _cm_cls.refresh_py(_cm)
        _g_cls.refresh_py(_g)
        _expp_cls.refresh_py(_expp)
        _obs_cls.refresh_py(_obs)
        _exf_cls.refresh_py(_exf)
        _y_cls.refresh_py(_y)
        _xl_cls.refresh_py(_xl)
        _iob_cls.refresh_py(_iob)
        _niter_py = utils._EngineIntScalarType.to_py(_niter)
        _rlogl_py = utils._EngineFloat64ScalarType.to_py(_rlogl)
        _chi_py = utils._EngineFloat64ScalarType.to_py(_chi)
        _idf_py = utils._EngineIntScalarType.to_py(_idf)
        _siglev_py = utils._EngineFloat64ScalarType.to_py(_siglev)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['niter', 'rlogl', 'chi', 'idf', 'siglev'],
        )(
            _niter_py,
            _rlogl_py,
            _chi_py,
            _idf_py,
            _siglev_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def binary_service(ip, n, x, irl):
    u"""
    Frequency count for ``binary``.

    ``binary_service`` is a service function which may be used prior to
    calling ``binary`` to calculate the frequency distribution of a set
    of dichotomous score patterns.

    For full information please refer to the NAG Library document for
    g11sb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g11/g11sbf.html

    Parameters
    ----------
    ip : int
        p, the number of dichotomous variables.

    n : int
        n, the number of individuals in the sample.

    x : bool, shape (`n`, `ip`), modified in place
        `On entry`: `x`[i-1,j-1] must be set equal to True if x_{ij} =
        1, and False if x_{ij} = 0, for j = 1,2,...,p, for i =
        1,2,...,n.

        `On exit`: the first s rows of `x` contain the s different score
        patterns.

    irl : int, shape (`n`), modified in place
        `On exit`: the frequency with which the lth row of `x` occurs,
        for l = 1,2,...,s.

    Returns
    -------
    ns : int
        The number of different score patterns, s.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 7.

        (`errno` 1)
            On entry, `ip` = *<value>*.

            Constraint: `ip` >= 3.
    """

    excs = []
    fun_name = 'naginterfaces.base.contab.binary_service'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g11sbft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, ip.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Output, ns.
        utils._EngineLogicalArrayType.c_type, # Input/Output, x[ldx, ip].
        utils._BYREF_INT, # Input, ldx.
        utils._EngineIntArrayType.c_type, # Output, irl[n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((x, 'x'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _ip = utils._EngineIntScalarType(
        ip, {'fun_name': fun_name, 'entity': 'ip'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _ns = utils._EngineIntScalarType.empty_ctypes()
    _ldx_py = utils._EngineLogicalArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _ip_py = utils._EngineIntScalarType.to_py(_ip)
    _x_cls = utils._EngineLogicalArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py, _ip_py),
        mutual_sorder=_m_sorder,
    )
    _x = _x_cls.to_ctypes()
    _irl_cls = utils._EngineIntArrayType(
        irl, {'fun_name': fun_name, 'entity': 'irl'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _irl = _irl_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _ip,
        _n,
        _ns,
        _x,
        _ldx,
        _irl,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 7.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ip >= 3.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        _irl_cls.refresh_py(_irl)
        _ns_py = utils._EngineIntScalarType.to_py(_ns)
        return_data = _ns_py
    _err_class.manage()
    return return_data
