# -*- coding: utf-8 -*-
# pylint: disable=bare-except,chained-comparison,invalid-name,len-as-condition,line-too-long,protected-access,redefined-outer-name,too-many-arguments,too-many-boolean-expressions,too-many-branches,too-many-lines,too-many-locals,too-many-statements,unneeded-not,unused-import
u"""
Interfaces for the NAG Mark 27.0 `det` Chapter.

``det`` - Determinants

Base interfaces.

We do not intend for these lower-level base interfaces to be called
directly from user code. Please refer to the ``naginterfaces.library``
subpackage instead.

This module is concerned with the calculation of determinants of square
matrices.

For full information please refer to the NAG Library document

https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f03/f03intro.html
"""

# NAG Copyright 2017-2020.

import collections as _collections
import logging as _logging
import sys as _sys

from ..base import _nag_enum_consts
from ..base import utils

def real_gen(n, a, ipiv):
    u"""
    Determinant of real matrix, matrix already factorized by
    ``lapacklin.dgetrf``.

    ``real_gen`` computes the determinant of a real n by n matrix A.
    ``lapacklin.dgetrf`` must be called first to supply the matrix A in
    factorized form.

    For full information please refer to the NAG Library document for
    f03ba

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f03/f03baf.html

    Parameters
    ----------
    n : int
        n, the order of the matrix A.

    a : float, shape (`n`, `n`)
        The n by n matrix A in factorized form as returned by
        ``lapacklin.dgetrf``.

    ipiv : int, shape (`n`)
        The row interchanges used to factorize matrix A as returned by
        ``lapacklin.dgetrf``.

    Returns
    -------
    d : float
        The determinant of A is given by `d`*2.0^`dexp`. It is given in
        this form to avoid overflow or underflow.

    dexp : int
        The determinant of A is given by `d`*2.0^`dexp`. It is given in
        this form to avoid overflow or underflow.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 4)
            The matrix A is approximately singular.

    References
    ----------
    Wilkinson, J H and Reinsch, C, 1971, `Handbook for Automatic
    Computation II, Linear Algebra`, Springer--Verlag
    """

    excs = []
    fun_name = 'naginterfaces.base.det.real_gen'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f03baft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, a[lda, n].
        utils._BYREF_INT, # Input, lda.
        utils._EngineIntArrayType.c_type, # Input, ipiv[n].
        utils._BYREF_FLOAT64, # Output, d.
        utils._BYREF_INT, # Output, dexp.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _lda_py = utils._EngineFloat64ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _lda = utils._EngineIntScalarType(
        _lda_py, {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _ipiv = utils._EngineIntArrayType(
        ipiv, {'fun_name': fun_name, 'entity': 'ipiv'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _d = utils._EngineFloat64ScalarType.empty_ctypes()
    _dexp = utils._EngineIntScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _a,
        _lda,
        _ipiv,
        _d,
        _dexp,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The matrix A is approximately singular.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _d_py = utils._EngineFloat64ScalarType.to_py(_d)
        _dexp_py = utils._EngineIntScalarType.to_py(_dexp)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['d', 'dexp'],
        )(
            _d_py,
            _dexp_py,
        )
    _err_class.manage()
    return return_data

def real_sym(n, a):
    u"""
    Determinant of real symmetric positive definite matrix.

    ``real_sym`` computes the determinant of a real n by n symmetric
    positive definite matrix A. ``lapacklin.dpotrf`` must be called
    first to supply the symmetric matrix A in Cholesky factorized form.
    The storage (upper or lower triangular) used by ``lapacklin.dpotrf``
    is not relevant to ``real_sym`` since only the diagonal elements of
    the factorized A are referenced.

    For full information please refer to the NAG Library document for
    f03bf

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f03/f03bff.html

    Parameters
    ----------
    n : int
        n, the order of the matrix A.

    a : float, shape (`n`, `n`)
        The lower or upper triangle of the Cholesky factorized form of
        the n by n positive definite symmetric matrix A. Only the
        diagonal elements are referenced.

    Returns
    -------
    d : float
        The determinant of A is given by `d`*2.0^`dexp`. It is given in
        this form to avoid overflow or underflow.

    dexp : int
        The determinant of A is given by `d`*2.0^`dexp`. It is given in
        this form to avoid overflow or underflow.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` > 0.

        (`errno` 4)
            The matrix A is not positive definite.

    References
    ----------
    Wilkinson, J H and Reinsch, C, 1971, `Handbook for Automatic
    Computation II, Linear Algebra`, Springer--Verlag
    """

    excs = []
    fun_name = 'naginterfaces.base.det.real_sym'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f03bfft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, a[lda, n].
        utils._BYREF_INT, # Input, lda.
        utils._BYREF_FLOAT64, # Output, d.
        utils._BYREF_INT, # Output, dexp.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _lda_py = utils._EngineFloat64ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _lda = utils._EngineIntScalarType(
        _lda_py, {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _d = utils._EngineFloat64ScalarType.empty_ctypes()
    _dexp = utils._EngineIntScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _a,
        _lda,
        _d,
        _dexp,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The matrix A is not positive definite.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _d_py = utils._EngineFloat64ScalarType.to_py(_d)
        _dexp_py = utils._EngineIntScalarType.to_py(_dexp)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['d', 'dexp'],
        )(
            _d_py,
            _dexp_py,
        )
    _err_class.manage()
    return return_data

def real_band_sym(uplo, n, kd, ab):
    u"""
    Determinant of real symmetric positive definite banded matrix
    previously factorized by ``lapacklin.dpbtrf``.

    ``real_band_sym`` computes the determinant of an n by n symmetric
    positive definite banded matrix A that has been stored in
    band-symmetric storage. ``lapacklin.dpbtrf`` must be called first to
    supply the Cholesky factorized form.
    The storage (upper or lower triangular) used by ``lapacklin.dpbtrf``
    is relevant as this determines which elements of the stored
    factorized form are referenced.

    For full information please refer to the NAG Library document for
    f03bh

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f03/f03bhf.html

    Parameters
    ----------
    uplo : str, length 1
        Indicates whether the upper or lower triangular part of A was
        stored and how it was factorized. This should not be altered
        following a call to ``lapacklin.dpbtrf``.

        `uplo` = 'U'

            The upper triangular part of A was originally stored and A
            was factorized as U^TU where U is upper triangular.

        `uplo` = 'L'

            The lower triangular part of A was originally stored and A
            was factorized as LL^T where L is lower triangular.

    n : int
        n, the order of the matrix A.

    kd : int
        k_d, the number of superdiagonals or subdiagonals of the matrix
        A.

    ab : float, shape (`kd`+1, `n`)
        The Cholesky factor of A, as returned by ``lapacklin.dpbtrf``.

    Returns
    -------
    d : float
        The determinant of A is given by `d`*2.0^`dexp`. It is given in
        this form to avoid overflow or underflow.

    dexp : int
        The determinant of A is given by `d`*2.0^`dexp`. It is given in
        this form to avoid overflow or underflow.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `uplo` = *<value>*.

            Constraint: `uplo` = 'L' or 'U'.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` > 0.

        (`errno` 3)
            On entry, `kd` = *<value>*.

            Constraint: `kd` >= 0.

        (`errno` 6)
            The matrix A is not positive definite.

    References
    ----------
    Wilkinson, J H and Reinsch, C, 1971, `Handbook for Automatic
    Computation II, Linear Algebra`, Springer--Verlag
    """

    excs = []
    fun_name = 'naginterfaces.base.det.real_band_sym'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f03bhft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, sorder.
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, kd.
        utils._EngineFloat64ArrayType.c_type, # Input, ab[ldab, n].
        utils._BYREF_INT, # Input, ldab.
        utils._BYREF_FLOAT64, # Output, d.
        utils._BYREF_INT, # Output, dexp.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((ab, 'ab'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _sorder = utils._EngineIntScalarType.c_type(_nag_enum_consts.nag_col_major)
    _uplo = utils._EngineCharScalarType(
        uplo, {'fun_name': fun_name, 'entity': 'uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _kd = utils._EngineIntScalarType(
        kd, {'fun_name': fun_name, 'entity': 'kd'},
    ).to_ctypes()
    _ldab_py = utils._EngineFloat64ArrayType.get_nd(ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldab = utils._EngineIntScalarType(
        _ldab_py, {'fun_name': fun_name, 'entity': '_ldab'},
    ).to_ctypes()
    _kd_py = utils._EngineIntScalarType.to_py(_kd)
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _ab = utils._EngineFloat64ArrayType(
        ab, {'fun_name': fun_name, 'entity': 'ab'},
        exp_shape=(_kd_py+1, _n_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _d = utils._EngineFloat64ScalarType.empty_ctypes()
    _dexp = utils._EngineIntScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _sorder,
        _uplo,
        _n,
        _kd,
        _ab,
        _ldab,
        _d,
        _dexp,
        _errbuf,
        _ierr,
        _uplo._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, uplo = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: uplo = \'L\' or \'U\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, kd = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: kd >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The matrix A is not positive definite.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _d_py = utils._EngineFloat64ScalarType.to_py(_d)
        _dexp_py = utils._EngineIntScalarType.to_py(_dexp)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['d', 'dexp'],
        )(
            _d_py,
            _dexp_py,
        )
    _err_class.manage()
    return return_data

def complex_gen(n, a, ipiv, dexp):
    u"""
    Determinant of complex matrix previously LU factorized.

    ``complex_gen`` computes the determinant of a complex n by n matrix
    A. ``lapacklin.zgetrf`` must be called first to supply the matrix A
    in factorized form.

    For full information please refer to the NAG Library document for
    f03bn

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f03/f03bnf.html

    Parameters
    ----------
    n : int
        n, the order of the matrix A.

    a : complex, shape (`n`, `n`)
        The n by n matrix A in factorized form as returned by
        ``lapacklin.zgetrf``.

    ipiv : int, shape (`n`)
        The row interchanges used to factorize matrix A as returned by
        ``lapacklin.zgetrf``.

    dexp : int, shape (2), modified in place
        `On exit`: the exponents for the real and imaginary parts of the
        determinant. The determinant, d = (d_r,d_i), is returned as d_r
        = D_r*2^j and d_i = D_i*2^k, where `d` = (D_r,D_i) and j and k
        are stored in the first and second elements respectively of the
        array `dexp` on successful exit.

    Returns
    -------
    d : complex
        The mantissa of the real and imaginary parts of the determinant.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 4)
            The matrix A is approximately singular.

    References
    ----------
    Wilkinson, J H and Reinsch, C, 1971, `Handbook for Automatic
    Computation II, Linear Algebra`, Springer--Verlag
    """

    excs = []
    fun_name = 'naginterfaces.base.det.complex_gen'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('f03bnft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineComplex128ArrayType.c_type, # Input, a[lda, n].
        utils._BYREF_INT, # Input, lda.
        utils._EngineIntArrayType.c_type, # Input, ipiv[n].
        utils._BYREF_COMPLEX128, # Output, d.
        utils._EngineIntArrayType.c_type, # Output, dexp[2].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _lda_py = utils._EngineComplex128ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _lda = utils._EngineIntScalarType(
        _lda_py, {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _a = utils._EngineComplex128ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _ipiv = utils._EngineIntArrayType(
        ipiv, {'fun_name': fun_name, 'entity': 'ipiv'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _d = utils._EngineComplex128ScalarType.empty_ctypes()
    _dexp_cls = utils._EngineIntArrayType(
        dexp, {'fun_name': fun_name, 'entity': 'dexp'},
        exp_shape=(2,),
        check_contents=False,
    )
    _dexp = _dexp_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _a,
        _lda,
        _ipiv,
        _d,
        _dexp,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The matrix A is approximately singular.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _dexp_cls.refresh_py(_dexp)
        _d_py = utils._EngineComplex128ScalarType.to_py(_d)
        return_data = _d_py
    _err_class.manage()
    return return_data
