# -*- coding: utf-8 -*-
# pylint: disable=bare-except,chained-comparison,invalid-name,len-as-condition,line-too-long,protected-access,redefined-outer-name,too-many-arguments,too-many-boolean-expressions,too-many-branches,too-many-lines,too-many-locals,too-many-statements,unneeded-not,unused-import
# pylint: disable=simplifiable-if-statement
u"""
Interfaces for the NAG Mark 27.0 `info` Chapter.

``info`` - Library Identification

Base interfaces.

We do not intend for these lower-level base interfaces to be called
directly from user code. Please refer to the ``naginterfaces.library``
subpackage instead.

The functions in this module provide information about the NAG Library.

Information about the precise implementation of the NAG Library in use
will be needed when communicating with the NAG Technical Support Service
(see [Support from NAG]).

For full information please refer to the NAG Library document

https://www.nag.com/numeric/nl/nagdoc_27/flhtml/a00/a00intro.html

Notes
-----
The function ``impl_details_separate`` is not supported in this product.
Please use ``impl_details`` instead.
"""

# NAG Copyright 2017-2020.

import logging as _logging
import sys as _sys

from ..base import utils

def impl_details():
    u"""
    NAG Library Engine identification, details of implementation and mark.

    .. _a00aa-py-ft-doc:

    For full information please refer to the NAG Library documents for
    a00aa and a00ad

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/a00/a00aaf.html

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/a00/a00adf.html

    Returns
    -------
    result : dict
        The Engine information dictionary, with the following keys:

        ``'build_id'`` : int or None
            An integer identifier for the implementation, or None
            if not known.

        ``'compilers'`` : dict
            The compilers used for the implementation,
            with the following keys:

            ``'C'`` : dict
                Information on the C compiler used,
                with the following keys:

                ``'title'`` : str
                    Identifier for this compiler.

                ``'flags'`` : str
                    Base set of flags used with this compiler.

            ``'C++'`` : dict
                Information on the C++ compiler used,
                with the following keys:

                ``'title'`` : str
                    Identifier for this compiler.

                ``'flags'`` : str
                    Base set of flags used with this compiler.

            ``'Fortran'`` : dict
                Information on the Fortran compiler used,
                with the following keys:

                ``'title':`` str
                    Identifier for this compiler.

                ``'flags'`` : str
                    Base set of flags used with this compiler.

        ``'hardware'`` : str
            The intended target hardware for the implementation.

        ``'integer_bits'`` : int
            The bit width of integers used in the implementation.

        ``'mark'`` : tuple
            Triplet of ints for the Mark number of the implementation.

        ``'os'`` : str
            The intended target operating system for the implementation.

            (Other OS variants may be compatible.)

        ``'precision'`` : str
            The base floating-point precision model used in the implementation.

        ``'product_code'`` : str
            The product code for the implementation.

        ``'title'`` : str
            The title of the implementation.

        ``'vendor_lib'`` : str or None
            The subsidiary vendor library used by the implementation,
            or None if self contained.

    Notes
    -----
    In the NAG Library the equivalent C functions for this routine are
    ``nag_implementation_details`` and ``nag_implementation_separated_details``.
    """
    fun_name = 'naginterfaces.base.info.impl_details'

    enginea_ctypes = utils._EngineState._get_symbol('a00aazu')
    enginea_ctypes.restype = None
    enginea_ctypes.argtypes = (
        utils._EngineCharArrayType.c_type, # Output, msg(len=*)[51].
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of msg.
    )
    _msg_chrlen = 200
    _msg_arrlen = 51
    _msg_list = ['']*_msg_arrlen
    _msg_cls = utils._EngineCharArrayType(
        _msg_list, {'fun_name': fun_name, 'entity': 'msg'},
        exp_chrlen=_msg_chrlen, exp_shape=(_msg_arrlen,),
        check_contents=False
    )
    _msg = _msg_cls.to_ctypes(can_reference_contents=False)
    enginea_ctypes(_msg, _msg_chrlen)
    _msg_cls.refresh_py(_msg)
    engined_ctypes = utils._EngineState._get_symbol('a00adzu')
    engined_ctypes.restype = None
    engined_ctypes.argtypes = (
        utils._BYREF_STRING, # Output, impl(len=*).
        utils._BYREF_STRING, # Output, prec(len=*).
        utils._BYREF_STRING, # Output, pcode(len=*).
        utils._BYREF_STRING, # Output, mkmaj(len=*).
        utils._BYREF_STRING, # Output, mkmin(len=*).
        utils._BYREF_STRING, # Output, hdware(len=*).
        utils._BYREF_STRING, # Output, opsys(len=*).
        utils._BYREF_STRING, # Output, fcomp(len=*).
        utils._BYREF_STRING, # Output, ccomp(len=*).
        utils._BYREF_STRING, # Output, cxxcomp(len=*).
        utils._BYREF_STRING, # Output, vend(len=*).
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of impl.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of prec.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of pcode.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of mkmaj.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of mkmin.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of hdware.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of opsys.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of fcomp.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of ccomp.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of cxxcomp.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of vend.
    )
    _impl = utils._EngineCharScalarType.empty_ctypes(_msg_chrlen)
    _prec = utils._EngineCharScalarType.empty_ctypes(_msg_chrlen)
    _pcode = utils._EngineCharScalarType.empty_ctypes(_msg_chrlen)
    _mkmaj = utils._EngineCharScalarType.empty_ctypes(_msg_chrlen)
    _mkmin = utils._EngineCharScalarType.empty_ctypes(_msg_chrlen)
    _hdware = utils._EngineCharScalarType.empty_ctypes(_msg_chrlen)
    _opsys = utils._EngineCharScalarType.empty_ctypes(_msg_chrlen)
    _fcomp = utils._EngineCharScalarType.empty_ctypes(_msg_chrlen)
    _ccomp = utils._EngineCharScalarType.empty_ctypes(_msg_chrlen)
    _cxxcomp = utils._EngineCharScalarType.empty_ctypes(_msg_chrlen)
    _vend = utils._EngineCharScalarType.empty_ctypes(_msg_chrlen)
    engined_ctypes(
        _impl,
        _prec,
        _pcode,
        _mkmaj,
        _mkmin,
        _hdware,
        _opsys,
        _fcomp,
        _ccomp,
        _cxxcomp,
        _vend,
        _msg_chrlen,
        _msg_chrlen,
        _msg_chrlen,
        _msg_chrlen,
        _msg_chrlen,
        _msg_chrlen,
        _msg_chrlen,
        _msg_chrlen,
        _msg_chrlen,
        _msg_chrlen,
        _msg_chrlen,
    )
    _msg_dict = {}
    def extract_title():
        """Extract the implementation title."""
        _msg_dict['title'] = utils._EngineCharScalarType.to_py(_impl, _msg_chrlen)
    extract_title()
    def extract_ibits():
        """Extract the integer bit width."""
        if '-bit integers' not in _msg_list[7]:
            _msg_dict['integer_bits'] = 32
            return
        _cut_off = _msg_list[7].index('-bit integers')
        _msg_dict['integer_bits'] = int(_msg_list[7][:_cut_off].split()[-1])
    extract_ibits()
    def extract_precision():
        """Extract the floating-point model."""
        _msg_dict['precision'] = _msg_list[3].split(':')[1].split()[0].strip()
    extract_precision()
    def extract_prodcode():
        """Extract the product code."""
        _msg_dict['product_code'] = utils._EngineCharScalarType.to_py(_pcode, _msg_chrlen)
    extract_prodcode()
    def extract_mark():
        """Extract the Mark number."""
        _msg_dict['mark'] = tuple([(int(val) if val.isdigit() else 0) for val in _msg_list[5].strip().split()[1].split('.')])
    extract_mark()
    def extract_vendlib():
        """Extract the name of the vendor lib used."""
        _vend_lib = utils._EngineCharScalarType.to_py(_vend, _msg_chrlen)
        _msg_dict['vendor_lib'] = None if _vend_lib == '(self-contained)' else _vend_lib
    extract_vendlib()
    def extract_build_id():
        """Import the build ID."""
        _id = utils._ENGINE_BUILD_ID_STR
        _msg_dict['build_id'] = None if (_id is None or not _id.isdigit()) else int(_id)
    extract_build_id()
    def extract_hardware():
        """Extract the hardware."""
        _msg_dict['hardware'] = utils._EngineCharScalarType.to_py(_hdware, _msg_chrlen)
    extract_hardware()
    def extract_os():
        """Extract the OS."""
        _msg_dict['os'] = utils._EngineCharScalarType.to_py(_opsys, _msg_chrlen)
    extract_os()
    def extract_compilers():
        """Extract the compilers used."""
        def extract_flags(sep_i):
            """Extract the compiler flags."""
            return ''.join([_msg_list[26 + sep_i + 1][3:], _msg_list[26 + sep_i + 2][3:]])
        _msg_dict['compilers'] = {
            'C': {
                'title': utils._EngineCharScalarType.to_py(_ccomp, _msg_chrlen),
                'flags': extract_flags(15),
            },
            'C++': {
                'title': utils._EngineCharScalarType.to_py(_cxxcomp, _msg_chrlen),
                'flags': extract_flags(20),
            },
            'Fortran': {
                'title': utils._EngineCharScalarType.to_py(_fcomp, _msg_chrlen),
                'flags': extract_flags(10),
            },
        }
    extract_compilers()
    expected_keys = (
        'build_id',
        'compilers',
        'hardware',
        'integer_bits',
        'mark',
        'os',
        'precision',
        'product_code',
        'title',
        'vendor_lib',
    )
    for key in expected_keys:
        if key not in _msg_dict:
            raise utils.NagException( # pragma: no cover
                {'fun_name': fun_name},
                'missing expected key \'' + key + '\''
            )
        if key == 'compilers':
            for ckey in ('C', 'C++', 'Fortran'):
                if ckey not in _msg_dict[key]:
                    raise utils.NagException( # pragma: no cover
                        {'fun_name': fun_name},
                        'missing expected ' + key + ' key \'' + ckey + '\''
                    )
                for cckey in ('title', 'flags'):
                    if cckey not in _msg_dict[key][ckey]:
                        raise utils.NagException( # pragma: no cover
                            {'fun_name': fun_name},
                            'missing expected ' + key + ':' + ckey + ' key \'' + cckey + '\''
                        )
    return _msg_dict

def licence():
    u"""
    Check availability of a valid licence key.

    ``licence`` provides a convenient means of checking the availability
    of a valid licence key on licence-managed implementations before
    starting computations that will use NAG Library functions.
    In particular, the use of this function is highly recommended in
    programs that call NAG Library functions within multithreaded
    sections (e.g., OpenMP parallel regions).
    The function need only be called once, before the start of the first
    multithreaded section.

    For full information please refer to the NAG Library document for
    a00ac

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/a00/a00acf.html

    Returns
    -------
    lic : bool
        True if a licence is available.
    """

    excs = []
    fun_name = 'naginterfaces.base.info.licence'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('a00acft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Output, ilic.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _ilic = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _ilic,
    )
    utils._handle_cb_exception(excs)
    _ilic_py = utils._EngineIntScalarType.to_py(_ilic)
    if _ilic_py == 0:
        _lic_py = False
    elif _ilic_py == 1:
        _lic_py = True
    else:
        raise utils.NagException(
            {'fun_name': fun_name},
            'Unexpected value for internal ilic.',
        )
    return_data = _lic_py
    return return_data
