# -*- coding: utf-8 -*-
# pylint: disable=bare-except,chained-comparison,invalid-name,len-as-condition,line-too-long,protected-access,redefined-outer-name,too-many-arguments,too-many-boolean-expressions,too-many-branches,too-many-lines,too-many-locals,too-many-statements,unneeded-not,unused-import
u"""
Interfaces for the NAG Mark 27.0 `omp` Chapter.

``omp`` - OpenMP Utilities

Base interfaces.

We do not intend for these lower-level base interfaces to be called
directly from user code. Please refer to the ``naginterfaces.library``
subpackage instead.

This module contains utilities for controlling the OpenMP environment
for your program.
They are based on OpenMP runtime library routines, although their
functionality varies slightly.

For full information please refer to the NAG Library document

https://www.nag.com/numeric/nl/nagdoc_27/flhtml/x06/x06intro.html
"""

# NAG Copyright 2017-2020.

import logging as _logging
import sys as _sys

from ..base import utils

def set_num_threads(num):
    u"""
    Sets the number of threads for OpenMP parallel regions.

    ``set_num_threads`` sets the number of threads requested for
    subsequent OpenMP parallel regions.

    For full information please refer to the NAG Library document for
    x06aa

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/x06/x06aaf.html

    Parameters
    ----------
    num : int
        The number of threads requested for subsequent OpenMP parallel
        regions.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `num` = *<value>*.

            Constraint: `num` >= 1.

    References
    ----------
    The OpenMP API Specification for Parallel Programming

    Chapman, B, Jost, G and van der Pas, R, 2008, `Using OpenMP Portable
    Shared Memory Parallel Programming`, The MIT Press
    """

    excs = []
    fun_name = 'naginterfaces.base.omp.set_num_threads'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('x06aaft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, num.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _num = utils._EngineIntScalarType(
        num, {'fun_name': fun_name, 'entity': 'num'},
    ).to_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _num,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, num = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: num >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        return_data = None
    _err_class.manage()
    return return_data

def get_num_threads():
    u"""
    The number of OpenMP threads in the current team.

    ``get_num_threads`` returns the number of OpenMP threads in the
    current team.

    For full information please refer to the NAG Library document for
    x06ab

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/x06/x06abf.html

    Returns
    -------
    nt : int
        The returned number of OpenMP threads in the current team.

    References
    ----------
    The OpenMP API Specification for Parallel Programming

    Chapman, B, Jost, G and van der Pas, R, 2008, `Using OpenMP Portable
    Shared Memory Parallel Programming`, The MIT Press
    """

    excs = []
    fun_name = 'naginterfaces.base.omp.get_num_threads'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('x06abft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Output, nt.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _nt = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _nt,
    )
    utils._handle_cb_exception(excs)
    _nt_py = utils._EngineIntScalarType.to_py(_nt)
    return_data = _nt_py
    return return_data

def get_max_threads():
    u"""
    An upper bound on the number of threads in the next parallel region.

    ``get_max_threads`` returns an upper bound on the number of threads
    used for subsequent OpenMP parallel regions.

    For full information please refer to the NAG Library document for
    x06ac

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/x06/x06acf.html

    Returns
    -------
    ntmax : int
        The returned upper bound on the number of threads used for
        subsequent OpenMP parallel regions.

    References
    ----------
    The OpenMP API Specification for Parallel Programming

    Chapman, B, Jost, G and van der Pas, R, 2008, `Using OpenMP Portable
    Shared Memory Parallel Programming`, The MIT Press
    """

    excs = []
    fun_name = 'naginterfaces.base.omp.get_max_threads'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('x06acft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Output, ntmax.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _ntmax = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _ntmax,
    )
    utils._handle_cb_exception(excs)
    _ntmax_py = utils._EngineIntScalarType.to_py(_ntmax)
    return_data = _ntmax_py
    return return_data

def get_thread_num():
    u"""
    The OpenMP thread number of the calling thread.

    ``get_thread_num`` returns the OpenMP thread number of the calling
    thread.

    For full information please refer to the NAG Library document for
    x06ad

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/x06/x06adf.html

    Returns
    -------
    tn : int
        The returned OpenMP thread number of the calling thread.

    References
    ----------
    The OpenMP API Specification for Parallel Programming

    Chapman, B, Jost, G and van der Pas, R, 2008, `Using OpenMP Portable
    Shared Memory Parallel Programming`, The MIT Press
    """

    excs = []
    fun_name = 'naginterfaces.base.omp.get_thread_num'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('x06adft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Output, tn.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _tn = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _tn,
    )
    utils._handle_cb_exception(excs)
    _tn_py = utils._EngineIntScalarType.to_py(_tn)
    return_data = _tn_py
    return return_data

def in_parallel():
    u"""
    Tests for an active OpenMP parallel region.

    ``in_parallel`` returns 1 if called from within an active OpenMP
    parallel region, and 0 otherwise.

    For full information please refer to the NAG Library document for
    x06af

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/x06/x06aff.html

    Returns
    -------
    inpar : int
        1 if called from within an active OpenMP parallel region, and 0
        otherwise.

    References
    ----------
    The OpenMP API Specification for Parallel Programming

    Chapman, B, Jost, G and van der Pas, R, 2008, `Using OpenMP Portable
    Shared Memory Parallel Programming`, The MIT Press
    """

    excs = []
    fun_name = 'naginterfaces.base.omp.in_parallel'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('x06afft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Output, inpar.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _inpar = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _inpar,
    )
    utils._handle_cb_exception(excs)
    _inpar_py = utils._EngineIntScalarType.to_py(_inpar)
    return_data = _inpar_py
    return return_data

def set_nested(nesting):
    u"""
    Enables or disables nested OpenMP parallelism.

    ``set_nested`` enables or disables nested OpenMP parallelism.

    For full information please refer to the NAG Library document for
    x06ag

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/x06/x06agf.html

    Parameters
    ----------
    nesting : int
        If `nesting` = 0, nesting of OpenMP parallel regions is
        disabled, otherwise it is enabled.

    References
    ----------
    The OpenMP API Specification for Parallel Programming

    Chapman, B, Jost, G and van der Pas, R, 2008, `Using OpenMP Portable
    Shared Memory Parallel Programming`, The MIT Press
    """

    excs = []
    fun_name = 'naginterfaces.base.omp.set_nested'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('x06agft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, nesting.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _nesting = utils._EngineIntScalarType(
        nesting, {'fun_name': fun_name, 'entity': 'nesting'},
    ).to_ctypes()
    engine_ctypes(
        _en_data,
        _nesting,
    )
    utils._handle_cb_exception(excs)
    return_data = None
    return return_data

def get_nested():
    u"""
    Tests the status of nested OpenMP parallelism.

    ``get_nested`` returns 1 if nested OpenMP parallelism has been
    enabled, and 0 otherwise.

    For full information please refer to the NAG Library document for
    x06ah

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/x06/x06ahf.html

    Returns
    -------
    nesting : int
        1 if nested OpenMP parallelism has been enabled, and 0
        otherwise.

    References
    ----------
    The OpenMP API Specification for Parallel Programming

    Chapman, B, Jost, G and van der Pas, R, 2008, `Using OpenMP Portable
    Shared Memory Parallel Programming`, The MIT Press
    """

    excs = []
    fun_name = 'naginterfaces.base.omp.get_nested'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('x06ahft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Output, nesting.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _nesting = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _nesting,
    )
    utils._handle_cb_exception(excs)
    _nesting_py = utils._EngineIntScalarType.to_py(_nesting)
    return_data = _nesting_py
    return return_data

def using_threaded_impl():
    u"""
    Tests whether a threaded NAG Library is being used.

    ``using_threaded_impl`` returns 1 when using a multithreaded
    implementation of the NAG Library and 0 otherwise.

    For full information please refer to the NAG Library document for
    x06xa

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/x06/x06xaf.html

    Returns
    -------
    smplib : int
        1 if using a multithreaded implementation of the NAG Library and
        0 otherwise.
    """

    excs = []
    fun_name = 'naginterfaces.base.omp.using_threaded_impl'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('x06xaft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Output, smplib.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _smplib = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _smplib,
    )
    utils._handle_cb_exception(excs)
    _smplib_py = utils._EngineIntScalarType.to_py(_smplib)
    return_data = _smplib_py
    return return_data
