# -*- coding: utf-8 -*-
# pylint: disable=bare-except,chained-comparison,invalid-name,len-as-condition,line-too-long,protected-access,redefined-outer-name,too-many-arguments,too-many-boolean-expressions,too-many-branches,too-many-lines,too-many-locals,too-many-statements,unneeded-not,unused-import
u"""
Interfaces for the NAG Mark 27.0 `smooth` Chapter.

``smooth`` - Smoothing in Statistics

Base interfaces.

We do not intend for these lower-level base interfaces to be called
directly from user code. Please refer to the ``naginterfaces.library``
subpackage instead.

This module is concerned with methods for smoothing data.
Included are methods for density estimation, smoothing time series data,
and statistical applications of splines.
These methods may also be viewed as nonparametric modelling.

For full information please refer to the NAG Library document

https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g10/g10intro.html
"""

# NAG Copyright 2017-2020.

import collections as _collections
import logging as _logging
import sys as _sys

from ..base import utils

def fit_spline(mode, n, x, y, wt, rho, yhat, c, res, h, comm):
    u"""
    Fit cubic smoothing spline, smoothing parameter given.

    ``fit_spline`` fits a cubic smoothing spline for a given smoothing
    parameter.

    For full information please refer to the NAG Library document for
    g10ab

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g10/g10abf.html

    Parameters
    ----------
    mode : str, length 1
        Indicates in which mode the function is to be used.

        `mode` = 'P'

            Initialization and fitting is performed. This partial fit
            can be used in an iterative weighted least squares context
            where the weights are changing at each call to
            ``fit_spline`` or when the coefficients are not required.

        `mode` = 'Q'

            Fitting only is performed. Initialization must have been
            performed previously by a call to ``fit_spline`` with `mode`
            = 'P'. This quick fit may be called repeatedly with
            different values of `rho` without re-initialization.

        `mode` = 'F'

            Initialization and full fitting is performed and the
            function coefficients are calculated.

    n : int
        n, the number of distinct observations.

    x : float, shape (`n`)
        The distinct and ordered values x_i, for i = 1,2,...,n.

    y : float, shape (`n`)
        The values y_i, for i = 1,2,...,n.

    wt : None or float, shape (`n`)
        If weight = 'W', `wt` must contain the n weights. Otherwise `wt`
        is not referenced and unit weights are assumed.

    rho : float
        rho, the smoothing parameter.

    yhat : float, shape (`n`), modified in place
        `On exit`: the fitted values, y^_i, for i = 1,2,...,n.

    c : float, shape (`n`-1, 3), modified in place
        `On entry`: if `mode` = 'Q', `c` must be unaltered from the
        previous call to ``fit_spline`` with `mode` = 'P'. Otherwise `c`
        need not be set.

        `On exit`: if `mode` = 'F', `c` contains the spline
        coefficients. More precisely, the value of the spline at t is
        given by ((`c`[i-1,2]*d+`c`[i-1,1])*d+`c`[i-1,0])*d+y^_i, where
        x_i <= t < x_{i+1} and d = t-x_i.

        If `mode` = 'P' or 'Q', `c` contains information that will be
        used in a subsequent call to ``fit_spline`` with `mode` = 'Q'.

    res : float, shape (`n`), modified in place
        `On exit`: the (weighted) residuals, r_i, for i = 1,2,...,n.

    h : float, shape (`n`), modified in place
        `On exit`: the leverages, h_{ii}, for i = 1,2,...,n.

    comm : dict, communication object, modified in place
        Communication structure.

        `On initial entry`: need not be set.

    Returns
    -------
    rss : float
        The (weighted) residual sum of squares.

    df : float
        The residual degrees of freedom.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `mode` = *<value>*.

            Constraint: `mode` = 'P', 'Q' or 'F'.

        (`errno` 1)
            On entry, `rho` = *<value>*.

            Constraint: `rho` >= 0.0.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 3.

        (`errno` 2)
            On entry, at least one element of `wt` <= 0.0.

        (`errno` 3)
            On entry, `x` is not a strictly ordered array.

    References
    ----------
    Hastie, T J and Tibshirani, R J, 1990, `Generalized Additive
    Models`, Chapman and Hall

    Hutchinson, M F, 1986, `Algorithm 642: A fast procedure for
    calculating minimum cross-validation cubic smoothing splines`, ACM
    Trans. Math. Software (12), 150--153

    Hutchinson, M F and de Hoog, F R, 1985, `Smoothing noisy data with
    spline functions`, Numer. Math. (47), 99--106

    Reinsch, C H, 1967, `Smoothing by spline functions`, Numer. Math.
    (10), 177--183
    """

    excs = []
    fun_name = 'naginterfaces.base.smooth.fit_spline'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g10abft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, mode(len=1).
        utils._BYREF_STRING, # Input, weight(len=1).
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[n].
        utils._EngineFloat64ArrayType.c_type, # Input, y[n].
        utils._EngineFloat64ArrayType.c_type, # Input, wt[n].
        utils._BYREF_FLOAT64, # Input, rho.
        utils._EngineFloat64ArrayType.c_type, # Output, yhat[n].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, c[ldc, 3].
        utils._BYREF_INT, # Input, ldc.
        utils._BYREF_FLOAT64, # Output, rss.
        utils._BYREF_FLOAT64, # Output, df.
        utils._EngineFloat64ArrayType.c_type, # Output, res[n].
        utils._EngineFloat64ArrayType.c_type, # Output, h[n].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, comm[9*n+14].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of mode.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of weight.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((c, 'c'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _mode = utils._EngineCharScalarType(
        mode, {'fun_name': fun_name, 'entity': 'mode'}, exp_chrlen=1,
    ).to_ctypes()
    if wt is None:
        _weight = utils._EngineCharScalarType(
            'U', {'fun_name': fun_name, 'entity': '_weight'}, exp_chrlen=1,
        ).to_ctypes()
    else:
        _weight = utils._EngineCharScalarType(
            'W', {'fun_name': fun_name, 'entity': '_weight'}, exp_chrlen=1,
        ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _wt = utils._EngineFloat64ArrayType(
        wt, {'fun_name': fun_name, 'entity': 'wt'},
        exp_shape=(_n_py,),
        allow_none=True,
    ).to_ctypes()
    _rho = utils._EngineFloat64ScalarType(
        rho, {'fun_name': fun_name, 'entity': 'rho'},
    ).to_ctypes()
    _yhat_cls = utils._EngineFloat64ArrayType(
        yhat, {'fun_name': fun_name, 'entity': 'yhat'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _yhat = _yhat_cls.to_ctypes(can_reference_contents=False)
    _ldc_py = utils._EngineFloat64ArrayType.get_nd(c, {'fun_name': fun_name, 'entity': 'c'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldc = utils._EngineIntScalarType(
        _ldc_py, {'fun_name': fun_name, 'entity': '_ldc'},
    ).to_ctypes()
    _c_cls = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_n_py-1, 3),
        mutual_sorder=_m_sorder,
    )
    _c = _c_cls.to_ctypes()
    _rss = utils._EngineFloat64ScalarType.empty_ctypes()
    _df = utils._EngineFloat64ScalarType.empty_ctypes()
    _res_cls = utils._EngineFloat64ArrayType(
        res, {'fun_name': fun_name, 'entity': 'res'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _res = _res_cls.to_ctypes(can_reference_contents=False)
    _h_cls = utils._EngineFloat64ArrayType(
        h, {'fun_name': fun_name, 'entity': 'h'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _h = _h_cls.to_ctypes(can_reference_contents=False)
    _comm_needs_init = utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        opt_keys=('comm',),
    )
    if _comm_needs_init:
        _comm = utils._EngineFloat64ArrayType.empty_ctypes(9*_n_py+14)
        comm['comm'] = _comm
    else:
        _comm = comm['comm']
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _mode,
        _weight,
        _n,
        _x,
        _y,
        _wt,
        _rho,
        _yhat,
        _c,
        _ldc,
        _rss,
        _df,
        _res,
        _h,
        _comm,
        _errbuf,
        _ierr,
        _mode._length_,
        _weight._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mode = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mode = \'P\', \'Q\' or \'F\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, rho = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: rho >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 3.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, at least one element of wt <= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x is not a strictly ordered array.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _yhat_cls.refresh_py(_yhat)
        _c_cls.refresh_py(_c)
        _res_cls.refresh_py(_res)
        _h_cls.refresh_py(_h)
        _rss_py = utils._EngineFloat64ScalarType.to_py(_rss)
        _df_py = utils._EngineFloat64ScalarType.to_py(_df)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['rss', 'df'],
        )(
            _rss_py,
            _df_py,
        )
    _err_class.manage()
    return return_data

def fit_spline_parest(method, n, x, y, wt, yhat, c, res, h, crit, u, tol, maxcal):
    u"""
    Fit cubic smoothing spline, smoothing parameter estimated.

    ``fit_spline_parest`` estimates the values of the smoothing
    parameter and fits a cubic smoothing spline to a set of data.

    For full information please refer to the NAG Library document for
    g10ac

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g10/g10acf.html

    Parameters
    ----------
    method : str, length 1
        Indicates whether the smoothing parameter is to be found by
        minimization of the CV or GCV functions, or by finding the
        smoothing parameter corresponding to a specified degrees of
        freedom value.

        `method` = 'C'

            Cross-validation is used.

        `method` = 'D'

            The degrees of freedom are specified.

        `method` = 'G'

            Generalized cross-validation is used.

    n : int
        n, the number of observations.

    x : float, shape (`n`)
        The distinct and ordered values x_i, for i = 1,2,...,n.

    y : float, shape (`n`)
        The values y_i, for i = 1,2,...,n.

    wt : None or float, shape (`n`)
        If weight = 'W', `wt` must contain the n weights. Otherwise `wt`
        is not referenced and unit weights are assumed.

    yhat : float, shape (`n`), modified in place
        `On exit`: the fitted values, y^_i, for i = 1,2,...,n.

    c : float, shape (`n`-1, 3), modified in place
        `On exit`: the spline coefficients. More precisely, the value of
        the spline approximation at t is given by
        ((`c`[i-1,2]*d+`c`[i-1,1])*d+`c`[i-1,0])*d+y^_i, where x_i <= t
        < x_{i+1} and d = t-x_i.

    res : float, shape (`n`), modified in place
        `On exit`: the (weighted) residuals, r_i, for i = 1,2,...,n.

    h : float, shape (`n`), modified in place
        `On exit`: the leverages, h_{ii}, for i = 1,2,...,n.

    crit : float
        If `method` = 'D', the required degrees of freedom for the
        spline.

        If `method` = 'C' or 'G', `crit` need not be set.

    u : float
        The upper bound on the smoothing parameter. If `u` <= `tol`, `u`
        = 1000.0 will be used instead. See [Further Comments] for
        details on how this argument is used.

    tol : float
        The accuracy to which the smoothing parameter `rho` is required.
        `tol` should preferably be not much less than sqrt(epsilon),
        where epsilon is the machine precision. If `tol` < epsilon,
        `tol` = sqrt(epsilon) will be used instead.

    maxcal : int
        The maximum number of spline evaluations to be used in finding
        the value of rho. If `maxcal` < 3, `maxcal` = 100 will be used
        instead.

    Returns
    -------
    rss : float
        The (weighted) residual sum of squares.

    df : float
        The residual degrees of freedom. If `method` = 'D' this will be
        n-`crit` to the required accuracy.

    crit : float
        If `method` = 'C', the value of the cross-validation, or if
        `method` = 'G', the value of the generalized cross-validation
        function, evaluated at the value of rho returned in `rho`.

    rho : float
        The smoothing parameter, rho.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `crit` = *<value>*.

            Constraint: if `method` = 'D', `crit` <= `n`.

        (`errno` 1)
            On entry, `crit` = *<value>*.

            Constraint: if `method` = 'D', `crit` > 2.0.

        (`errno` 1)
            On entry, `method` is not valid: `method` = *<value>*.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 3.

        (`errno` 2)
            On entry, at least one element of `wt` <= 0.0.

        (`errno` 3)
            On entry, `x` is not a strictly ordered array.

        (`errno` 4)
            For the specified degrees of freedom, `rho` > `u`: `u` =
            *<value>*.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 5)
            Accuracy of `tol` cannot be achieved: `tol` = *<value>*.

        (`errno` 6)
            `maxcal` iterations have been performed.

        (`errno` 7)
            Optimum value of `rho` lies above `u`: `u` = *<value>*.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Hastie, T J and Tibshirani, R J, 1990, `Generalized Additive
    Models`, Chapman and Hall

    Hutchinson, M F, 1986, `Algorithm 642: A fast procedure for
    calculating minimum cross-validation cubic smoothing splines`, ACM
    Trans. Math. Software (12), 150--153

    Hutchinson, M F and de Hoog, F R, 1985, `Smoothing noisy data with
    spline functions`, Numer. Math. (47), 99--106

    Reinsch, C H, 1967, `Smoothing by spline functions`, Numer. Math.
    (10), 177--183
    """

    excs = []
    fun_name = 'naginterfaces.base.smooth.fit_spline_parest'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g10acft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, method(len=1).
        utils._BYREF_STRING, # Input, weight(len=1).
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[n].
        utils._EngineFloat64ArrayType.c_type, # Input, y[n].
        utils._EngineFloat64ArrayType.c_type, # Input, wt[n].
        utils._EngineFloat64ArrayType.c_type, # Output, yhat[n].
        utils._EngineFloat64ArrayType.c_type, # Output, c[ldc, 3].
        utils._BYREF_INT, # Input, ldc.
        utils._BYREF_FLOAT64, # Output, rss.
        utils._BYREF_FLOAT64, # Output, df.
        utils._EngineFloat64ArrayType.c_type, # Output, res[n].
        utils._EngineFloat64ArrayType.c_type, # Output, h[n].
        utils._BYREF_FLOAT64, # Input/Output, crit.
        utils._BYREF_FLOAT64, # Output, rho.
        utils._BYREF_FLOAT64, # Input, u.
        utils._BYREF_FLOAT64, # Input, tol.
        utils._BYREF_INT, # Input, maxcal.
        utils._EngineFloat64ArrayType.c_type, # Workspace, wk.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of method.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of weight.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((c, 'c'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _method = utils._EngineCharScalarType(
        method, {'fun_name': fun_name, 'entity': 'method'}, exp_chrlen=1,
    ).to_ctypes()
    if wt is None:
        _weight = utils._EngineCharScalarType(
            'U', {'fun_name': fun_name, 'entity': '_weight'}, exp_chrlen=1,
        ).to_ctypes()
    else:
        _weight = utils._EngineCharScalarType(
            'W', {'fun_name': fun_name, 'entity': '_weight'}, exp_chrlen=1,
        ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _wt = utils._EngineFloat64ArrayType(
        wt, {'fun_name': fun_name, 'entity': 'wt'},
        exp_shape=(_n_py,),
        allow_none=True,
    ).to_ctypes()
    _yhat_cls = utils._EngineFloat64ArrayType(
        yhat, {'fun_name': fun_name, 'entity': 'yhat'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _yhat = _yhat_cls.to_ctypes(can_reference_contents=False)
    _ldc_py = utils._EngineFloat64ArrayType.get_nd(c, {'fun_name': fun_name, 'entity': 'c'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldc = utils._EngineIntScalarType(
        _ldc_py, {'fun_name': fun_name, 'entity': '_ldc'},
    ).to_ctypes()
    _c_cls = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_n_py-1, 3),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _c = _c_cls.to_ctypes(can_reference_contents=False)
    _rss = utils._EngineFloat64ScalarType.empty_ctypes()
    _df = utils._EngineFloat64ScalarType.empty_ctypes()
    _res_cls = utils._EngineFloat64ArrayType(
        res, {'fun_name': fun_name, 'entity': 'res'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _res = _res_cls.to_ctypes(can_reference_contents=False)
    _h_cls = utils._EngineFloat64ArrayType(
        h, {'fun_name': fun_name, 'entity': 'h'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _h = _h_cls.to_ctypes(can_reference_contents=False)
    _crit = utils._EngineFloat64ScalarType(
        crit, {'fun_name': fun_name, 'entity': 'crit'},
    ).to_ctypes()
    _rho = utils._EngineFloat64ScalarType.empty_ctypes()
    _u = utils._EngineFloat64ScalarType(
        u, {'fun_name': fun_name, 'entity': 'u'},
    ).to_ctypes()
    _tol = utils._EngineFloat64ScalarType(
        tol, {'fun_name': fun_name, 'entity': 'tol'},
    ).to_ctypes()
    _maxcal = utils._EngineIntScalarType(
        maxcal, {'fun_name': fun_name, 'entity': 'maxcal'},
    ).to_ctypes()
    _wk = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _method,
        _weight,
        _n,
        _x,
        _y,
        _wt,
        _yhat,
        _c,
        _ldc,
        _rss,
        _df,
        _res,
        _h,
        _crit,
        _rho,
        _u,
        _tol,
        _maxcal,
        _wk,
        _errbuf,
        _ierr,
        _method._length_,
        _weight._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, crit = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: if method = \'D\', crit <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, crit = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: if method = \'D\', crit > 2.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, method is not valid: method = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 3.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, at least one element of wt <= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x is not a strictly ordered array.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99988:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'For the specified degrees of freedom, rho > u: u = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99987:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Accuracy of tol cannot be achieved: tol = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99986:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'maxcal iterations have been performed.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 99985:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Optimum value of rho lies above u: u = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _yhat_cls.refresh_py(_yhat)
        _c_cls.refresh_py(_c)
        _res_cls.refresh_py(_res)
        _h_cls.refresh_py(_h)
        _crit_py = utils._EngineFloat64ScalarType.to_py(_crit)
        _rss_py = utils._EngineFloat64ScalarType.to_py(_rss)
        _df_py = utils._EngineFloat64ScalarType.to_py(_df)
        _rho_py = utils._EngineFloat64ScalarType.to_py(_rho)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['rss', 'df', 'crit', 'rho'],
        )(
            _rss_py,
            _df_py,
            _crit_py,
            _rho_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def kerndens_gauss(n, x, wtype, window, slo, shi, ns, smooth, t, comm):
    u"""
    Kernel density estimate using Gaussian kernel (thread safe).

    ``kerndens_gauss`` performs kernel density estimation using a
    Gaussian kernel.

    For full information please refer to the NAG Library document for
    g10bb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g10/g10bbf.html

    Parameters
    ----------
    n : int
        n, the number of observations in the sample.

        On a continuation call, `n` must be unchanged since the last
        call to ``kerndens_gauss``.

    x : float, shape (`n`)
        x_i, for i = 1,2,...,n.

        On a continuation call, `x` must be unchanged since the last
        call to ``kerndens_gauss``.

    wtype : int
        How the window width, h, is to be calculated:

        `wtype` = 1

            h is supplied in `window`.

        `wtype` = 2

            h is to be calculated from the data, with

                h = m*(0.9*min(q_75-q_25, sigma)/(n^0.2))

            where q_75-q_25 is the inter-quartile range and sigma the
            standard deviation of the sample, x, and m is a multipler
            supplied in `window`. The 25% and 75% quartiles, q_25 and
            q_75, are calculated using ``stat.quantiles``. This is the
            'rule-of-thumb' suggested by Silverman (1990).

        Suggested value: `wtype` = 2 and `window` = 1.0.

    window : float
        If `wtype` = 1, then h, the window width. Otherwise, m, the
        multiplier used in the calculation of h.

        Suggested value: `window` = 1.0 and `wtype` = 2.

    slo : float
        If `slo` < `shi` then a, the lower limit of the interval on
        which the estimate is calculated. Otherwise, a and b, the lower
        and upper limits of the interval, are calculated as follows:

            [table omitted]

        where h is the window width.

        For most applications a should be at least three window widths
        below the lowest data point.

        On a continuation call, a supplied `slo` will be ignored and the
        appropriate value from the previous call to ``kerndens_gauss``
        will be extracted from `comm`['rcomm'].

        Suggested value: `slo` = 3.0 and `shi` = 0.0 which would cause a
        and b to be set 3 window widths below and above the lowest and
        highest data points respectively.

    shi : float
        If `slo` < `shi` then b, the upper limit of the interval on
        which the estimate is calculated. Otherwise a value for b is
        calculated from the data as stated in the description of `slo`
        and the value supplied in `shi` is not used.

        For most applications b should be at least three window widths
        above the highest data point.

        On a continuation call, a supplied `shi` will be ignored and the
        appropriate value from the previous call to ``kerndens_gauss``
        will be extracted from `comm`['rcomm'].

        Suggested value: `shi` = 0.0 and `slo` = 3.0 which would cause a
        and b to be set 3 window widths below and above the lowest and
        highest data points respectively.

    ns : int
        n_s, the number of points at which the estimate is calculated.

        On a continuation call, `ns` must be unchanged since the last
        call to ``kerndens_gauss``.

        Suggested value: `ns` = 512.

    smooth : float, shape (`ns`), modified in place
        `On exit`: f^(t_l), for l = 1,2,...,n_s, the n_s values of the
        density estimate.

    t : float, shape (`ns`), modified in place
        `On exit`: t_l, for l = 1,2,...,n_s, the points at which the
        estimate is calculated.

    comm : dict, communication object, modified in place
        Communication structure.

        If not initialized on entry then the values of Y_l are to be
        calculated by this call to ``kerndens_gauss``.

        Otherwise, this is a continuation call and it is assumed that
        the values of Y_l were calculated by a previous call to this
        function and the relevant information is stored in
        `comm`['rcomm'].

    Returns
    -------
    window : float
        h, the window width actually used.

    slo : float
        a, the lower limit actually used.

    shi : float
        b, the upper limit actually used.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, `n` = *<value>*.

            Constraint: `n` > 0.

        (`errno` 12)
            On entry, `n` = *<value>*.

            On entry at previous call, `n` = *<value>*.

            Constraint: on a continuation call, `n` must be unchanged
            since previous call.

        (`errno` 31)
            On entry, `wtype` = *<value>*.

            Constraint: `wtype` = 1 or 2.

        (`errno` 41)
            On entry, `window` = *<value>*.

            Constraint: `window` > 0.0.

        (`errno` 51)
            On entry, `slo` = *<value>*.

            On exit from previous call, `slo` = *<value>*.

            Constraint: on a continuation call, `slo` must be the same
            value returned by the previous call.

        (`errno` 62)
            On entry, `shi` = *<value>*.

            On exit from previous call, `shi` = *<value>*.

            Constraint: on a continuation call, `shi` must be the same
            value returned by the previous call.

        (`errno` 71)
            On entry, `ns` = *<value>*.

            Constraint: `ns` >= 2.

        (`errno` 74)
            On entry, `ns` = *<value>*.

            On entry at previous call, `ns` = *<value>*.

            Constraint: on a continuation call, `ns` must be unchanged
            since previous call.

        (`errno` 111)
            `comm`['rcomm'] has been corrupted between calls.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 61)
            On entry, `slo` = *<value>* and `shi` = *<value>*.

            On entry, min(`x`) = *<value>* and max(`x`) = *<value>*.

            Expected values of at least *<value>* and *<value>* for
            `slo` and `shi`.

    References
    ----------
    Jones, M C and Lotwick, H W, 1984, `Remark AS R50. A remark on
    algorithm AS 176. Kernel density estimation using the Fast Fourier
    Transform`, Appl. Statist. (33), 120--122

    Silverman, B W, 1982, `Algorithm AS 176. Kernel density estimation
    using the fast Fourier transform`, Appl. Statist. (31), 93--99

    Silverman, B W, 1990, `Density Estimation`, Chapman and Hall
    """

    excs = []
    fun_name = 'naginterfaces.base.smooth.kerndens_gauss'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g10bbft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[n].
        utils._BYREF_INT, # Input, wtype.
        utils._BYREF_FLOAT64, # Input/Output, window.
        utils._BYREF_FLOAT64, # Input/Output, slo.
        utils._BYREF_FLOAT64, # Input/Output, shi.
        utils._BYREF_INT, # Input, ns.
        utils._EngineFloat64ArrayType.c_type, # Output, smooth[ns].
        utils._EngineFloat64ArrayType.c_type, # Output, t[ns].
        utils._BYREF_INT, # Input, fcall.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, rcomm[ns+20].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _wtype = utils._EngineIntScalarType(
        wtype, {'fun_name': fun_name, 'entity': 'wtype'},
    ).to_ctypes()
    _window = utils._EngineFloat64ScalarType(
        window, {'fun_name': fun_name, 'entity': 'window'},
    ).to_ctypes()
    _slo = utils._EngineFloat64ScalarType(
        slo, {'fun_name': fun_name, 'entity': 'slo'},
    ).to_ctypes()
    _shi = utils._EngineFloat64ScalarType(
        shi, {'fun_name': fun_name, 'entity': 'shi'},
    ).to_ctypes()
    _ns = utils._EngineIntScalarType(
        ns, {'fun_name': fun_name, 'entity': 'ns'},
    ).to_ctypes()
    _ns_py = utils._EngineIntScalarType.to_py(_ns)
    _smooth_cls = utils._EngineFloat64ArrayType(
        smooth, {'fun_name': fun_name, 'entity': 'smooth'},
        exp_shape=(_ns_py,),
        check_contents=False,
    )
    _smooth = _smooth_cls.to_ctypes(can_reference_contents=False)
    _t_cls = utils._EngineFloat64ArrayType(
        t, {'fun_name': fun_name, 'entity': 't'},
        exp_shape=(_ns_py,),
        check_contents=False,
    )
    _t = _t_cls.to_ctypes(can_reference_contents=False)
    _comm_needs_init = utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        opt_keys=('rcomm',),
    )
    if _comm_needs_init:
        _fcall = utils._EngineIntScalarType.c_type(1)
    else:
        _fcall = utils._EngineIntScalarType.c_type(0)
    if _comm_needs_init:
        _rcomm = utils._EngineFloat64ArrayType.empty_ctypes(_ns_py+20)
        comm['rcomm'] = _rcomm
    else:
        _rcomm = comm['rcomm']
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _x,
        _wtype,
        _window,
        _slo,
        _shi,
        _ns,
        _smooth,
        _t,
        _fcall,
        _rcomm,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 11 and _ifmt == 11:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 12 and _ifmt == 12:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'On entry at previous call, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: on a continuation call, n must be unchanged since previous call.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 31 and _ifmt == 31:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, wtype = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: wtype = 1 or 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 41 and _ifmt == 41:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, window = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: window > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 51 and _ifmt == 51:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, slo = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'On exit from previous call, slo = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: on a continuation call, slo must be the same value returned by the previous call.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 61 and _ifmt == 61:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, slo = {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'shi = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'On entry, min(x) = {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'max(x) = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[4]),
                    ),
                    'Expected values of at least {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[5]),
                    ),
                    '{:s} for slo and shi.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[6]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 62 and _ifmt == 62:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, shi = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'On exit from previous call, shi = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: on a continuation call, shi must be the same value returned by the previous call.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 71 and _ifmt == 71:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ns = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ns >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 74 and _ifmt == 74:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ns = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'On entry at previous call, ns = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: on a continuation call, ns must be unchanged since previous call.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 111 and _ifmt == 111:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'comm[\'rcomm\'] has been corrupted between calls.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _window_py = utils._EngineFloat64ScalarType.to_py(_window)
        _slo_py = utils._EngineFloat64ScalarType.to_py(_slo)
        _shi_py = utils._EngineFloat64ScalarType.to_py(_shi)
        _smooth_cls.refresh_py(_smooth)
        _t_cls.refresh_py(_t)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['window', 'slo', 'shi'],
        )(
            _window_py,
            _slo_py,
            _shi_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def data_runningmedian(itype, n, y, smooth, rough):
    u"""
    Compute smoothed data sequence using running median smoothers.

    ``data_runningmedian`` computes a smoothed data sequence using
    running median smoothers.

    For full information please refer to the NAG Library document for
    g10ca

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g10/g10caf.html

    Parameters
    ----------
    itype : int
        Specifies the method to be used.

            If `itype` = 0, 4253H,twice is used.

            If `itype` = 1, 3RSSH,twice is used.

    n : int
        n, the number of observations.

    y : float, shape (`n`)
        The sample observations.

    smooth : float, shape (`n`), modified in place
        `On exit`: contains the smooth.

    rough : float, shape (`n`), modified in place
        `On exit`: contains the rough.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `itype` = *<value>*.

            Constraint: `itype` = 0 or 1.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` > 6.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Tukey, J W, 1977, `Exploratory Data Analysis`, Addison--Wesley

    Velleman, P F and Hoaglin, D C, 1981, `Applications, Basics, and
    Computing of Exploratory Data Analysis`, Duxbury Press, Boston, MA
    """

    excs = []
    fun_name = 'naginterfaces.base.smooth.data_runningmedian'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g10caft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, itype.
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, y[n].
        utils._EngineFloat64ArrayType.c_type, # Output, smooth[n].
        utils._EngineFloat64ArrayType.c_type, # Output, rough[n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _itype = utils._EngineIntScalarType(
        itype, {'fun_name': fun_name, 'entity': 'itype'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _smooth_cls = utils._EngineFloat64ArrayType(
        smooth, {'fun_name': fun_name, 'entity': 'smooth'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _smooth = _smooth_cls.to_ctypes(can_reference_contents=False)
    _rough_cls = utils._EngineFloat64ArrayType(
        rough, {'fun_name': fun_name, 'entity': 'rough'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _rough = _rough_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _itype,
        _n,
        _y,
        _smooth,
        _rough,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, itype = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: itype = 0 or 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n > 6.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _smooth_cls.refresh_py(_smooth)
        _rough_cls.refresh_py(_rough)
        return_data = None
    _err_class.manage()
    return return_data

def data_order(n, x, y, wt, xord, yord, wtord):
    u"""
    Reorder data to give ordered distinct observations.

    ``data_order`` orders and weights data which is entered
    unsequentially, weighted or unweighted.

    For full information please refer to the NAG Library document for
    g10za

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g10/g10zaf.html

    Parameters
    ----------
    n : int
        n, the number of observations.

    x : float, shape (`n`)
        The values, x_i, for i = 1,2,...,n.

    y : float, shape (`n`)
        The values y_i, for i = 1,2,...,n.

    wt : None or float, shape (`n`)
        If weight = 'W', `wt` must contain the n weights. Otherwise `wt`
        is not referenced and unit weights are assumed.

    xord : float, shape (`n`), modified in place
        `On exit`: the first `nord` elements contain the ordered and
        distinct x_i.

    yord : float, shape (`n`), modified in place
        `On exit`: the first `nord` elements contain the values y'
        corresponding to the values in `xord`.

    wtord : float, shape (`n`), modified in place
        `On exit`: the first `nord` elements contain the values w'
        corresponding to the values of `xord` and `yord`.

    Returns
    -------
    nord : int
        The number of distinct observations.

    rss : float
        The within group sum of squares for tied observations.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, weight = *<value>*.

            Constraint: weight = 'W' or 'U'.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 2)
            On entry, all weights are zero.

        (`errno` 2)
            On entry, i = *<value>* and `wt`[i-1] = *<value>*.

            Constraint: `wt`[i-1] >= 0.0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Draper, N R and Smith, H, 1985, `Applied Regression Analysis`, (2nd
    Edition), Wiley

    West, D H D, 1979, `Updating mean and variance estimates: An
    improved method`, Comm. ACM (22), 532--555
    """

    excs = []
    fun_name = 'naginterfaces.base.smooth.data_order'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g10zaft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, weight(len=1).
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[n].
        utils._EngineFloat64ArrayType.c_type, # Input, y[n].
        utils._EngineFloat64ArrayType.c_type, # Input, wt[n].
        utils._BYREF_INT, # Output, nord.
        utils._EngineFloat64ArrayType.c_type, # Output, xord[n].
        utils._EngineFloat64ArrayType.c_type, # Output, yord[n].
        utils._EngineFloat64ArrayType.c_type, # Output, wtord[n].
        utils._BYREF_FLOAT64, # Output, rss.
        utils._EngineIntArrayType.c_type, # Workspace, iwrk.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of weight.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    if wt is None:
        _weight = utils._EngineCharScalarType(
            'U', {'fun_name': fun_name, 'entity': '_weight'}, exp_chrlen=1,
        ).to_ctypes()
    else:
        _weight = utils._EngineCharScalarType(
            'W', {'fun_name': fun_name, 'entity': '_weight'}, exp_chrlen=1,
        ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _wt = utils._EngineFloat64ArrayType(
        wt, {'fun_name': fun_name, 'entity': 'wt'},
        exp_shape=(_n_py,),
        allow_none=True,
    ).to_ctypes()
    _nord = utils._EngineIntScalarType.empty_ctypes()
    _xord_cls = utils._EngineFloat64ArrayType(
        xord, {'fun_name': fun_name, 'entity': 'xord'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _xord = _xord_cls.to_ctypes(can_reference_contents=False)
    _yord_cls = utils._EngineFloat64ArrayType(
        yord, {'fun_name': fun_name, 'entity': 'yord'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _yord = _yord_cls.to_ctypes(can_reference_contents=False)
    _wtord_cls = utils._EngineFloat64ArrayType(
        wtord, {'fun_name': fun_name, 'entity': 'wtord'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _wtord = _wtord_cls.to_ctypes(can_reference_contents=False)
    _rss = utils._EngineFloat64ScalarType.empty_ctypes()
    _iwrk = utils._EngineIntArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _weight,
        _n,
        _x,
        _y,
        _wt,
        _nord,
        _xord,
        _yord,
        _wtord,
        _rss,
        _iwrk,
        _errbuf,
        _ierr,
        _weight._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, weight = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: weight = \'W\' or \'U\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, all weights are zero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s} and wt[i-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: wt[i-1] >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _xord_cls.refresh_py(_xord)
        _yord_cls.refresh_py(_yord)
        _wtord_cls.refresh_py(_wtord)
        _nord_py = utils._EngineIntScalarType.to_py(_nord)
        _rss_py = utils._EngineFloat64ScalarType.to_py(_rss)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['nord', 'rss'],
        )(
            _nord_py,
            _rss_py,
        )
    _err_class.manage()
    return return_data
