# -*- coding: utf-8 -*-
# pylint: disable=chained-comparison,invalid-name,line-too-long,protected-access,too-many-arguments,too-many-boolean-expressions,too-many-branches,too-many-lines,too-many-locals,too-many-statements,unused-import
u"""
Interfaces for the NAG Mark 27.0 `blgm` Chapter.

``blgm`` - Linear Model Specification

The functions in this module provide a mechanism for specifying a linear
model using a text based modelling language and are intended to be used
in conjunction with the model fitting functions from other modules, for
example submodule ``correg``.

For full information please refer to the NAG Library document

https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g22/g22intro.html

See Also
--------
``naginterfaces.library.examples.blgm`` :
    This subpackage contains examples for the ``blgm`` module.
    See also the :ref:`library_blgm_ex` subsection.
"""

# NAG Copyright 2017-2020.

import collections as _collections

from ..base import blgm as _b_blgm
from ..base import utils as _b_utils
from ..library import _utils

def lm_formula(hform, formula):
    u"""
    Specify a linear model via a formula string.

    Note: this function uses optional algorithmic parameters, see also:
    ``optset``, ``optget``.

    ``lm_formula`` parses a text string containing a formula specifying
    a linear model and outputs a G22 handle to an internal data
    structure.
    This G22 handle can then be passed to various functions in submodule
    ``blgm``.
    In particular, the G22 handle can be passed to ``lm_design_matrix``
    to produce a design matrix or ``lm_submodel`` to produce a vector of
    column inclusion flags suitable for use with functions in submodule
    ``correg``.

    For full information please refer to the NAG Library document for
    g22ya

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g22/g22yaf.html

    Parameters
    ----------
    hform : Handle, modified in place
        `On entry`: must be set to a null Handle, alternatively an
        existing G22 handle may be supplied in which case this function
        will destroy the supplied G22 handle as if ``handle_free`` had
        been called.

        `On exit`: holds a G22 handle to the internal data structure
        containing a description of the model M as specified in
        `formula`. You **must not** change the G22 handle other than
        through functions in submodule ``blgm``.

    formula : str
        A string containing the formula specifying M. See [Description]
        for details on the allowed model syntax.

    Other Parameters
    ----------------
    'Contrast' : str
        Default = 'FIRST'

        This argument controls the default contrasts used for the
        categorical independent variables appearing in the model.
        Six types of contrasts and dummy variables are available:

        'FIRST'
            Treatment contrasts relative to the first level of the
            variable will be used.

        'LAST'
            Treatment contrasts relative to the last level of the
            variable will be used.

        'SUM FIRST'
            Sum contrasts relative to the first level of the variable
            will be used.

        'SUM LAST'
            Sum contrasts relative to the last level of the variable
            will be used.

        'HELMERT'
            Helmert contrasts will be used.

        'POLYNOMIAL'
            Polynomial contrasts will be used.

        'DUMMY'
            Dummy variables will be used rather than a contrast.

        See ``lm_design_matrix`` for more information on contrasts,
        their effect on the design matrix and how they are constructed.

        This argument may have an `instance identifier` associated with
        it (see ``optset`` and ``optget``).
        The `instance identifier` must be the name of one of the
        variables appearing in the model supplied in `formula` when the
        G22 handle was created.
        For example, `CONTRAST : VAR1 = HELMERT` would set Helmert
        contrasts for the variable named `VAR1`.

        If no `instance identifier` is specified, the default contrast
        for all categorical variables in the model is changed, otherwise
        only the default contrast for the named variable is changed.

        In some situations it might be necessary for a variable to use a
        different contrast, depending on where it appears in the model
        formula.
        In order to allow contrasts to be specified on a term by term
        basis the @ operator can be used in the model formula.
        The syntax for this operator is V_j@c, where c is one of: `F`,
        `L`, `SF`, `SL`, `H`, `P` or `D`, corresponding to treatment
        contrasts relative to the first and last levels, sum contrasts
        relative to the first and last levels, Helmert contrasts,
        polynomial contrasts or dummy variables respectively.

        If the contrast has not been explicitly specified via the @
        operator, the value obtained from the option 'Contrast' is used.

        For example, setting `formula` to `VAR1 + VAR1@H.VAR2@P +
        VAR2@H.VAR3`, specifies that the variable named `VAR1` should
        use the default contrasts in the first term and Helmert
        contrasts in the second term.
        The variable named `VAR2` should use polynomial contrasts in the
        second term and Helmert contrasts in the third term.
        The variable named `VAR3` should use the default contrasts in
        the third term.

    'Explicit Mean' : str
        Default = 'NO'

        If 'Explicit Mean' = 'YES', any mean effect included in the
        model will be explicitly added to the design matrix, X, as a
        column of 1s.

        If 'Explicit Mean' = 'NO', it is assumed that the function to
        which X will be passed treats the mean effect as a special case,
        see mean in ``correg.linregm_fit`` for example.

    'Formula' : str
        This argument returns a verbose version of the model formula
        specified in `formula`, expanded and simplified to only contain
        variable names, the operators + and . and any contrast
        identifiers present.

    'Storage Order' : str
        Default = 'OBSVAR'

        This option controls how the design matrix, X, should be stored
        in its output array and only has an effect if the design matrix
        is being constructed using ``lm_design_matrix``.

        If 'Storage Order' = 'OBSVAR', X_{ij}, the value for the jth
        variable of the ith observation of the design matrix is stored
        in x[i-1,j-1].

        If 'Storage Order' = 'VAROBS', X_{ij}, the value for the jth
        variable of the ith observation of the design matrix is stored
        in x[j-1,i-1].

        Where x is the output argument of the same name in
        ``lm_design_matrix``.

    'Subject' : str
        This argument gives the subject terms associated with the
        `formula` in a linear mixed effects model.

        The supplied value must consist of a single term, representing
        either a single independent variable, or a single interaction
        term between two or more independent variables.
        All variables in the subject term must not also appear in the
        model formula.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, `hform` is not a null Handle or a recognised G22
            handle.

        (`errno` 21)
            The formula contained a mismatched parenthesis.

            The position in the formula string of the error is
            *<value>*.

        (`errno` 22)
            An operator was missing.

            The position in the formula string of the error is
            *<value>*.

        (`errno` 23)
            Invalid use of an operator.

            The position in the formula string of the error is
            *<value>*.

        (`errno` 24)
            Invalid specification for the power operator.

            The position in the formula string of the error is
            *<value>*.

        (`errno` 25)
            Invalid specification for the colon operator.

            The position in the formula string of the error is
            *<value>*.

        (`errno` 26)
            Invalid specification for the mean.

            The position in the formula string of the error is
            *<value>*.

        (`errno` 27)
            Invalid variable name.

            The position in the formula string of the error is
            *<value>*.

        (`errno` 28)
            Missing variable name.

            The position in the formula string of the error is
            *<value>*.

        (`errno` 29)
            After processing, the model contains no terms.

        (`errno` 30)
            An invalid contrast specifier has been supplied.

            The position in the formula string of the error is
            *<value>*.

        (`errno` 41)
            On entry, an invalid option was supplied in `formula`.

        (`errno` 42)
            On entry, an option was supplied in `formula`, but the
            expected delimiter '=' was not found.

        (`errno` 43)
            On entry, an option was supplied in `formula`, but the
            supplied optval was invalid.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 31)
            A term contained a repeated variable with a different
            contrast specifier.

    See Also
    --------
    :meth:`naginterfaces.library.examples.blgm.lm_formula_ex.main`
    :meth:`naginterfaces.library.examples.correg.glm_binomial_ex.main`
    :meth:`naginterfaces.library.examples.correg.lmm_init_combine_ex.main`
    """

    _b_blgm.lm_formula(
        hform,
        formula,
    )

def lm_describe_data(hddesc, nobs, levels, vnames=None):
    u"""
    Describe a dataset.

    Note: this function uses optional algorithmic parameters, see also:
    ``optset``, ``optget``.

    ``lm_describe_data`` describes a data matrix.

    For full information please refer to the NAG Library document for
    g22yb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g22/g22ybf.html

    Parameters
    ----------
    hddesc : Handle, modified in place
        `On entry`: must be set to a null Handle, alternatively an
        existing G22 handle may be supplied in which case this function
        will destroy the supplied G22 handle as if ``handle_free`` had
        been called.

        `On exit`: holds a G22 handle to the internal data structure
        containing a description of the data matrix, D. You **must not**
        change the G22 handle other than through the functions in
        submodule ``blgm``.

    nobs : int
        n, the number of observations in the data matrix, D.

    levels : int, array-like, shape (nvar)
        `levels`[j-1] contains the number of levels associated with the
        jth variable of the data matrix, for j = 1,2,...,nvar.

        If the jth variable is binary, ordinal or continuous,
        `levels`[j-1] should be set to 1; otherwise `levels`[j-1] should
        be set to the number of levels associated with the jth variable
        and the corresponding column of the data matrix is assumed to
        take the value 1 to `levels`[j-1].

    vnames : None or str, array-like, shape (lvnames), optional
        If `vnames` is not None, `vnames`[j-1] must contain the name of
        the jth variable, for j = 1,2,...,nvar.

        The names supplied in `vnames` should be at most 50 characters
        long and be unique.

        If a name longer than 50 characters is supplied it will be
        truncated.

        Variable names must not contain any of the characters +.*-:^()@.

    Other Parameters
    ----------------
    'Number of Observations' : int
        If queried, this option will return n, the number of
        observations in the data matrix.

    'Number of Variables' : int
        If queried, this option will return m_d, the number of variables
        in the data matrix.

    'Storage Order' : str
        Default = 'OBSVAR'

        This option states how the data matrix, D, will be stored in its
        input array.

        If 'Storage Order' = 'OBSVAR', D_{ij}, the value for the jth
        variable of the ith observation of the data matrix is stored in
        dat[i-1,j-1].

        If 'Storage Order' = 'VAROBS', D_{ij}, the value for the jth
        variable of the ith observation of the data matrix is stored in
        dat[j-1,i-1].

        Where dat is the input argument of the same name in
        ``lm_design_matrix``.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, `hddesc` is not a null Handle or a recognised G22
            handle.

        (`errno` 21)
            On entry, `nobs` = *<value>*.

            Constraint: `nobs` >= 0.

        (`errno` 31)
            On entry, nvar = *<value>*.

            Constraint: nvar >= 0.

        (`errno` 41)
            On entry, j = *<value>* and `levels`[j-1] = *<value>*.

            Constraint: `levels`[i-1] >= 1.

        (`errno` 51)
            On entry, lvnames = *<value>* and nvar = *<value>*.

            Constraint: lvnames = 0 or nvar.

        (`errno` 61)
            On entry, variable name i contains one more invalid
            characters, i = *<value>*.

        (`errno` 62)
            On entry, variable names i and j are not unique, i =
            *<value>* and j = *<value>*.

        (`errno` 63)
            On entry, variable names i and j are not unique (possibly
            due to truncation), i = *<value>* and j = *<value>*.

            Maximum variable name length is 50.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 64)
            At least one variable name was truncated to 50 characters.
            Each truncated name is unique and will be used in all
            output.

    See Also
    --------
    :meth:`naginterfaces.library.examples.blgm.lm_formula_ex.main`
    :meth:`naginterfaces.library.examples.correg.glm_binomial_ex.main`
    :meth:`naginterfaces.library.examples.correg.lmm_init_combine_ex.main`
    """

    fun_name = 'naginterfaces.library.blgm.lm_describe_data'

    _levels = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'levels'},
        levels,
        copy=False,
    )
    _nvar_py = _b_utils._EngineIntArrayType.get_shape(_levels, {'fun_name': fun_name, 'entity': 'levels'}, exp_rank=1)[0]
    _nvar = _b_utils._EngineIntScalarType(
        _nvar_py, {'fun_name': fun_name, 'entity': '_nvar'},
    ).to_ctypes()
    _vnames = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'vnames'},
        vnames,
        copy=False,
        allow_none=True,
    )
    if vnames is None:
        _lvnames = 0
        _lvnames_py = _lvnames
    else:
        _lvnames_py = _b_utils._EngineCharArrayType.get_shape(_vnames, {'fun_name': fun_name, 'entity': 'vnames'}, exp_rank=1, allow_none=True)[0]
        _lvnames = _b_utils._EngineIntScalarType(
            _lvnames_py, {'fun_name': fun_name, 'entity': '_lvnames'},
        ).to_ctypes()
    _b_blgm.lm_describe_data(
        hddesc,
        nobs,
        _nvar,
        _levels,
        _lvnames,
        _vnames,
    )

def lm_design_matrix(hform, hddesc, dat, hxdesc):
    u"""
    Construct a design matrix from a linear model specified using
    ``lm_formula``.

    Note: this function uses optional algorithmic parameters, see also:
    ``optset``, ``optget``.

    ``lm_design_matrix`` generates a design matrix from a data matrix
    and model description.

    For full information please refer to the NAG Library document for
    g22yc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g22/g22ycf.html

    Parameters
    ----------
    hform : Handle
        A G22 handle to the internal data structure containing a
        description of the model M as returned in hform by
        ``lm_formula``.

    hddesc : Handle
        A G22 handle to the internal data structure containing a
        description of the data matrix, D as returned in hddesc by
        ``lm_describe_data``.

    dat : float, array-like, shape (:, :)
        The data matrix, D. By default D_{ij}, the ith value for the jth
        variable, for j = 1,2,...,m_d, for i = 1,2,...,n, should be
        supplied in `dat`[i-1,j-1].

        If the option 'Storage Order', described in
        ``lm_describe_data``, is set to 'VAROBS', D_{ij} should be
        supplied in `dat`[j-1,i-1].

    hxdesc : Handle, modified in place
        `On entry`: must be set to a null Handle, alternatively an
        existing G22 handle may be supplied in which case this function
        will destroy the supplied G22 handle as if ``handle_free`` had
        been called.

        `On exit`: holds a G22 handle to the internal data structure
        containing a description of the design matrix, X. You **must
        not** change the G22 handle other than through the functions in
        submodule ``blgm``.

    Returns
    -------
    x : float, ndarray, shape (:, :)
        The design matrix, X. By default X_{ij}, the ith value for the
        jth column, for j = 1,2,...,m_x, for i = 1,2,...,n, is returned
        in `x`[i-1,j-1].

        If the option 'Storage Order', described in ``lm_formula``, is
        set to 'VAROBS', X_{ij} is returned in `x`[j-1,i-1].

    Other Parameters
    ----------------
    'Formula' : str
        This option returns a verbose formula string describing the
        model, M, used to create the design matrix.
        This formula will only contain variable names, the operators '+'
        and '.' and any contrast identifiers present.

    'Min Number of Columns' : int
        This option returns the minimum number of columns required to
        hold the design matrix, X.
        In most cases 'Min Number of Columns' = 'Number of Columns'.
        The one exception is when `errno` = 71, that is the size of `x`
        was too small but the data matrix given in `dat` can be used as
        the design matrix.
        In this case, 'Number of Columns' = m_x = m_d and 'Min Number of
        Columns' holds the number of columns that would be required if
        only the relevant parts of `dat` were copied into a new array.

    'Number of Columns' : int
        This option returns m_x, the number of columns in the design
        matrix.

    'Number of Observations' : int
        This option returns n, the number of observations in the design
        matrix.

    'Storage Order' : str
        This option returns how the design matrix, X, is stored in `x`.

        If 'Storage Order' = 'OBSVAR', X_{ij}, the value for the jth
        variable of the ith observation of the design matrix is stored
        in `x`[i-1,j-1].

        If 'Storage Order' = 'VAROBS', X_{ij}, the value for the jth
        variable of the ith observation of the design matrix is stored
        in `x`[j-1,i-1].

        It should be noted that 'Storage Order' is not writeable.
        If you wish to change the storage order of the design matrix you
        need to change 'Storage Order' in `hform` as described in [Other
        Parameters for lm_formula] prior to calling
        ``lm_design_matrix``.

    Raises
    ------
    NagValueError
        (`errno` 11)
            `hform` has not been initialized or is corrupt.

        (`errno` 12)
            `hform` is not a G22 handle as generated by ``lm_formula``.

        (`errno` 13)
            A variable name used when creating `hform` is not present in
            `hddesc`.

            Variable name: *<value>*.

        (`errno` 21)
            `hddesc` has not been initialized or is corrupt.

        (`errno` 22)
            `hddesc` is not a G22 handle as generated by
            ``lm_describe_data``.

        (`errno` 31)
            On entry, column j of the data matrix, D, is not consistent
            with information supplied in `hddesc`, j = *<value>*.

        (`errno` 41)
            On entry, n = *<value>* and lddat = *<value>*.

            Constraint: lddat >= n.

        (`errno` 42)
            On entry, m_d = *<value>* and lddat = *<value>*.

            Constraint: lddat >= m_d.

        (`errno` 51)
            On entry, m_d = *<value>* and sddat = *<value>*.

            Constraint: sddat >= m_d.

        (`errno` 52)
            On entry, n = *<value>* and sddat = *<value>*.

            Constraint: sddat >= n.

        (`errno` 61)
            On entry, `hxdesc` is not a null Handle or a recognised G22
            handle.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 14)
            The model contains categorical variables, but no intercept
            or main effects terms have been requested.

            Please check the design matrix returned matches the model
            you require.

        (`errno` 32)
            Column j of the data matrix, D, required rounding more than
            expected when being treated as a categorical variable, j =
            *<value>*.

    References
    ----------
    Chambers, J M and Hastie, T J, 1992, `Statistical Models in S`,
    Wadsworth and Brooks/Cole Computer Science Series

    See Also
    --------
    :meth:`naginterfaces.library.examples.correg.glm_binomial_ex.main`
    """

    fun_name = 'naginterfaces.library.blgm.lm_design_matrix'

    _dat = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'dat'},
        dat,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_dat, 'dat'),),
    )
    _vu_ldx = _b_utils._EngineIntScalarType.c_type(0)
    _vu_ldx_py = _b_utils._EngineIntScalarType.to_py(_vu_ldx)
    _vu_sdx = _b_utils._EngineIntScalarType.c_type(0)
    _vu_sdx_py = _b_utils._EngineIntScalarType.to_py(_vu_sdx)
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_vu_ldx_py, _vu_sdx_py),
        sorder=_m_sorder,
    )
    _xsize = _b_utils._EngineIntArrayType.empty_ndarray(
        (2,),
    )
    try:
        _return_scalar = _b_blgm.lm_design_matrix(
            hform,
            hddesc,
            _dat,
            hxdesc,
            _x,
            _xsize,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _x_py = _x
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x'],
    )(
        _x_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    _vu_ldx_py = _xsize[0]
    _vu_sdx_py = _xsize[1]
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_vu_ldx_py, _vu_sdx_py),
        sorder=_m_sorder,
    )
    try:
        _return_scalar = _b_blgm.lm_design_matrix(
            hform,
            hddesc,
            _dat,
            hxdesc,
            _x,
            _xsize,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _x_py = _x
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x'],
    )(
        _x_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _x_py

def lm_submodel(what, hform, hxdesc, lisx, lplab, lvinfo, lenlab=210):
    u"""
    Construct a vector indicating which columns of a design matrix to
    include in a submodel specified using ``lm_formula``.

    ``lm_submodel`` produces labels for the columns of a design matrix,
    model parameters and a vector of column inclusion flags suitable for
    use with functions in submodule ``correg``.
    Thus allowing for submodels to be fit using the same design matrix.

    For full information please refer to the NAG Library document for
    g22yd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g22/g22ydf.html

    Parameters
    ----------
    what : str
        Controls what labels are to be produced:

        `what` = 'S'

            Labels for a submodel are required. The submodel must be
            supplied in `hform`.

        `what` = 'X'

            Labels for the design matrix X.

            If `hxdesc` was returned by ``correg.lmm_init`` in hlmm then
            X is the design matrix associated with the fixed parameters.

        `what` = 'Z'

            Labels for the design matrix Z.

            If `hxdesc` was returned by ``correg.lmm_init`` in hlmm then
            Z is the part of the design matrix associated with the
            random parameters.

        `what` = 'V'

            Labels for the variance components.

    hform : Handle
        A G22 handle to the internal data structure containing a
        description of the required submodel M_S, as returned in hform
        by ``lm_formula``. If `what` != 'S' `hform` is not referenced
        and need not be set.

    hxdesc : Handle
        A G22 handle to the internal data structure containing a
        description of the design matrix, D.

    lisx : int
        Length of `isx`.

    lplab : int
        The length of `plab`.

        As p <= m_x+1, if labels are required, using `lplab` = m_x+1
        will always be sufficient.

    lvinfo : int
        The length of `vinfo`.

        Let n_T denote the number of terms in M_S, n_{Tt} denote the
        number of variables in the tth term and m_{xt} denote the number
        of columns of X corresponding to the tth term.

        The required size of `vinfo`, denoted a is given by:

            a = sum(m_{xt}*(1+3n_{Tt})).

        If the model includes a mean effect, a should be incremented by
        one.

        The values n_T, n_{Tt} and m_{xt} are not trivial to calculate
        as they require the formula describing the model to be fully
        expanded and the contrast / dummy variable encoding to be known.

        Therefore, if `lisx`, `lplab` or `lvinfo` are too small and
        `lvinfo` >= 3, `errno` = 102 is returned and the required sizes
        for these arrays are returned in `vinfo`[0], `vinfo`[1] and
        `vinfo`[2] respectively.

    lenlab : int, optional
        Length of the strings allocated in `plab`. At most `lenlab`
        characters will be written into each element of `plab`.

    Returns
    -------
    intcpt : str
        If `intcpt` = 'M', in order to fit the model M_S to D using X,
        any analysis function should include an implicit mean effect
        (intercept term).

        `intcpt` = 'Z', if M_S does not include a mean effect or the
        mean effect has been explicitly included in the design matrix.

    ip : int
        p, the number of parameters in the (sub)model, including the
        intercept if one is present. If `what` = 'S', then the submodel
        is the one specified in `hform` otherwise the model is the one
        used when defining the design matrix described in `hxdesc`.

        If `lisx` != 0, if `intcpt` = 'Z', p = sum_i = 1^m_x`isx`[i-1],
        otherwise p = sum_i = 1^m_x`isx`[i-1]+1.

    isx : None or int, ndarray, shape (`lisx`)
        If `lisx` != 0, an array indicating which columns of the design
        matrix from the model specified in `hform` are to be used.

        `isx`[j-1] = 0

            The jth column of the design matrix, X, should not be
            included in the analysis.

        `isx`[j-1] = 1

            The jth column of the design matrix, X, should be included
            in the analysis.

        If `lisx` = 0, `isx` is not referenced.

    plab : None or str, ndarray, shape (min(`ip`, `lplab`))
        If `lplab` != 0, the names associated with the p parameters in
        the model.

        If `intcpt` = 'Z', the labels in `plab` are also the labels for
        the columns of design matrix used in the analysis.

        If `intcpt` = 'M', columns `plab`[1] to `plab`[p-1] are the
        corresponding column labels.

        If a mean effect is present in M_S, the corresponding label is
        always in `plab`[0].

        If `lplab` = 0, `plab` is not referenced.

    vinfo : None or int, ndarray, shape (`lvinfo`)
        If `lvinfo` != 0, information encoding a description of the
        parameters in the model.

        The encoding information can be extracted as follows:

        (i) Set k = 1.

        (#) Iterate j from 1 to p.

            (1) Set b = `vinfo`[k-1].

            (#) Increment k.

            (#) Iterate i from 1 to b.

                (a) Set v_i = `vinfo`[k-1].

                (#) Set l_i = `vinfo`[k].

                (#) Set c_i = `vinfo`[k+1].

                (#) Increment k by 3.

            (#) The jth model parameter corresponds to the interaction
                between the b variables held in columns v_1,v_2,...,v_b
                of D. Therefore, b = 1 indicates a main effect, b = 2 a
                two-way interaction, etc..

                If b = 0, the jth model parameter corresponds to the
                mean effect.

                If l_i = 0, the corresponding variable v_i is binary,
                ordinal or continuous.

                Otherwise, l_i is the level for the corresponding
                variable for model parameter j.

                c_i is a numeric flag indicating the contrast used in
                the case of a categorical variable.

                With c_i = 0 indicating that dummy variables were used
                for variable v_i in this term.

                The remaining six types of contrast; treatment contrasts
                (with respect to the first and last levels), sum
                contrasts (with respect to the first and last levels),
                Helmert contrasts and polynomial contrasts, as described
                in ``lm_design_matrix``, are identified by the integers
                one to six respectively.

        If `lvinfo` = 0, `vinfo` is not referenced.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, `what` = *<value>* was an illegal value.

        (`errno` 12)
            Supplied value of `what` is not valid for the G22 handle
            supplied in `hxdesc`.

        (`errno` 21)
            `hform` has not been initialized or is corrupt.

        (`errno` 22)
            `hform` is not a G22 handle as generated by ``lm_formula``.

        (`errno` 23)
            A variable name used when creating `hform` is not present in
            `hxdesc`.

            Variable name: *<value>*.

        (`errno` 24)
            The model and the design matrix are not consistent. The
            design matrix was constructed in the presence of a mean
            effect and the model does not include a mean effect.

        (`errno` 25)
            The model and the design matrix are not consistent. The
            model includes a term not present in the design matrix.

            Term: *<value>*.

        (`errno` 26)
            The model and the design matrix are not consistent.

            Term: *<value>*.

            This is likely due to the design matrix being constructed in
            the presence of either a mean effect or main effect that is
            not present in the model.

        (`errno` 31)
            `hxdesc` has not been initialized or is corrupt.

        (`errno` 32)
            `hxdesc` is not a G22 handle as generated by
            ``lm_design_matrix``.

        (`errno` 61)
            On entry, `lisx` = *<value>* and m_x = *<value>*.

            Constraint: `lisx` = 0 or `lisx` >= m_x.

        (`errno` 81)
            On entry, `lplab` = *<value>* and p = *<value>*.

            Constraint: `lplab` = 0 or `lplab` >= p.

        (`errno` 91)
            On entry, `plab` is too short to hold the parameter labels.
            Long labels will be truncated.

            The longest parameter label is *<value>*.

        (`errno` 101)
            On entry, `lvinfo` is too small.

            `lvinfo` = *<value>*.

            Constraint: `lvinfo` = 0 or `lvinfo` >= *<value>*.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 27)
            The model and the design matrix are not consistent. The
            model specifies different contrasts to those used when the
            design matrix was constructed. The contrasts specified in
            `hform` will be ignored.

        (`errno` 33)
            `hxdesc` has not passed through the model fitting function.

        (`errno` 102)
            On entry, one or more of `lisx`, `lplab` or `lvinfo` are
            nonzero, but too small.

            Minimum values are zero, or *<value>*, *<value>* and
            *<value>* respectively.

            The minimum values are returned in the first three elements
            of `vinfo`.

    See Also
    --------
    :meth:`naginterfaces.library.examples.correg.glm_binomial_ex.main`
    :meth:`naginterfaces.library.examples.correg.lmm_init_combine_ex.main`
    """

    fun_name = 'naginterfaces.library.blgm.lm_submodel'

    _lisx = _b_utils._EngineIntScalarType(
        lisx, {'fun_name': fun_name, 'entity': 'lisx'},
    ).to_ctypes()
    _lisx_py = _b_utils._EngineIntScalarType.to_py(_lisx)
    if _lisx_py != 0:
        _isx = _b_utils._EngineIntArrayType.empty_ndarray(
            (_lisx_py,),
        )
    else:
        _isx = None
    _lplab = _b_utils._EngineIntScalarType(
        lplab, {'fun_name': fun_name, 'entity': 'lplab'},
    ).to_ctypes()
    _lplab_py = _b_utils._EngineIntScalarType.to_py(_lplab)
    _lenlab = _b_utils._EngineIntScalarType(
        lenlab, {'fun_name': fun_name, 'entity': 'lenlab'},
    ).to_ctypes()
    _lenlab_py = _b_utils._EngineIntScalarType.to_py(_lenlab)
    _vl_lenlab_py = _lenlab_py
    if _lplab_py != 0:
        _plab = _b_utils._EngineCharArrayType.empty_ndarray(
            (_lplab_py,),
            chrlen=_vl_lenlab_py,
        )
    else:
        _plab = None
    _lvinfo = _b_utils._EngineIntScalarType(
        lvinfo, {'fun_name': fun_name, 'entity': 'lvinfo'},
    ).to_ctypes()
    _lvinfo_py = _b_utils._EngineIntScalarType.to_py(_lvinfo)
    if _lvinfo_py != 0:
        _vinfo = _b_utils._EngineIntArrayType.empty_ndarray(
            (_lvinfo_py,),
        )
    else:
        _vinfo = None
    try:
        _return_scalars = _b_blgm.lm_submodel(
            what,
            hform,
            hxdesc,
            _lisx,
            _isx,
            _lplab,
            _plab,
            _lvinfo,
            _vinfo,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _intcpt_py = _return_scalars.intcpt
    _ip_py = _return_scalars.ip
    if _lisx_py != 0:
        _isx_py = _isx
    else:
        _isx_py = None
    if _lplab_py != 0:
        _plab_py = _plab[:min(_ip_py, _lplab_py)]
    else:
        _plab_py = None
    if _lvinfo_py != 0:
        _vinfo_py = _vinfo
    else:
        _vinfo_py = None
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['intcpt', 'ip', 'isx', 'plab', 'vinfo'],
    )(
        _intcpt_py,
        _ip_py,
        _isx_py,
        _plab_py,
        _vinfo_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def handle_free(handle):
    u"""
    Destroy a G22 handle and deallocate all the memory used.

    ``handle_free`` destroys a G22 handle and deallocates all the memory
    used.

    For full information please refer to the NAG Library document for
    g22za

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g22/g22zaf.html

    Parameters
    ----------
    handle : Handle, modified in place
        `On entry`: the G22 handle to be destroyed.

        `On exit`: the handle is destroyed and set to a null Handle.

    Raises
    ------
    NagValueError
        (`errno` 12)
            `handle` has been corrupted.

        (`errno` 13)
            `handle` is a handle to an unknown data structure.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 11)
            `handle` has not been initialized.
    """

    _b_blgm.handle_free(
        handle,
    )

def optset(handle, optstr):
    u"""
    Option setting function for submodule ``blgm``.

    ``optset`` is a general option setting function for functions in
    submodule ``blgm``.
    It can set a single option or reset all of them to their default.

    For full information please refer to the NAG Library document for
    g22zm

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g22/g22zmf.html

    Parameters
    ----------
    handle : Handle
        The G22 handle which **must** have been initialized by one of
        submodule ``blgm``'s initialization functions.

    optstr : str
        A string identifying the option, its value and, where required,
        the instance identifier.

        Defaults

            Resets all options to their default values.

        option = optval

            Sets (all instances) of option to optval.

        option:instance identifier = optval

            Sets a single instance of option to optval.

        option = default

            Resets (all instances) of option to their default value.

        option:instance identifier = default

            Resets a single instance of option to its default value.

        `optstr` is case insensitive and option, `instance identifier`
        and optval may consist of one or more tokens separated by white
        space.

        See the documentation of the individual functions in [the G22
        Introduction] for details of valid values for option, `instance
        identifier` and optval.

    Raises
    ------
    NagValueError
        (`errno` 11)
            `handle` has not been initialized or is corrupt.

        (`errno` 12)
            `handle` is not a G22 handle.

        (`errno` 21)
            On entry, option was not recognized.

            `optstr` = *<value>*.

        (`errno` 22)
            On entry, the expected delimiter '=' was not found.

            `optstr` = *<value>*.

        (`errno` 23)
            On entry, option is read only.

            `optstr` = *<value>*.

        (`errno` 24)
            On entry, could not convert optval to an integer.

            `optstr` = *<value>*.

        (`errno` 25)
            On entry, could not convert optval to a real.

            `optstr` = *<value>*.

        (`errno` 26)
            On entry, optval is not a valid value for option.

            `optstr` = *<value>*.

        (`errno` 121)
            Invalid `instance identifier` for option.

            On entry, `optstr` = *<value>*.

        (`errno` 122)
            Numeric `instance identifier` is out of range.

            On entry, instance identifier = *<value>*.

            Constraint: *<value>* <= instance identifier and instance
            identifier <= *<value>*.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 123)
            On entry, option cannot have an associated `instance
            identifier`. The supplied `instance identifier` was ignored.

            `optstr` = *<value>*.

    See Also
    --------
    :meth:`naginterfaces.library.examples.blgm.lm_formula_ex.main`
    """

    _b_blgm.optset(
        handle,
        optstr,
    )

def optget(handle, optstr):
    u"""
    Option getting function for submodule ``blgm``.

    ``optget`` is a general option getting function for submodule
    ``blgm``.
    It is used to query the value of options.

    For full information please refer to the NAG Library document for
    g22zn

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g22/g22znf.html

    Parameters
    ----------
    handle : Handle
        The G22 handle which **must** have been initialized by one of
        submodule ``blgm``'s initialization functions.

    optstr : str
        A string identifying the option and, where required, the
        instance identifier.

        **identify**

            Returns a string description of the G22 handle supplied in
            `handle`. See [Further Comments] for more details.

        option

            Returns the value of option. If there are multiple instances
            of option, the value of the first is returned. If not all
            instances of option have the same value, `errno` = 124 is
            returned.

        option:instance identifier

            Returns the value of a single instance of option.

        `optstr` is case insensitive and option and `instance
        identifier` may consist of one or more tokens separated by white
        space.

        See the documentation of the individual submodule ``blgm``
        functions for details of valid values for option and `instance
        identifier`.

    Returns
    -------
    optvalue : dict
        The option-value ``dict``, with the following keys:

        ``'value'`` : float, int or str
            The value of the requested option.

        ``'annotation'`` : None or str
            Possible additional information about the option value.

    Raises
    ------
    NagValueError
        (`errno` 11)
            `handle` has not been initialized or is corrupt.

        (`errno` 12)
            `handle` is not a G22 handle.

        (`errno` 21)
            On entry, option was not recognized: `optstr` = *<value>*.

        (`errno` 22)
            On entry, option is not readable: `optstr` = *<value>*.

        (`errno` 121)
            Invalid `instance identifier` for option.

            On entry, `optstr` = *<value>*.

        (`errno` 122)
            Numeric `instance identifier` is out of range.

            On entry, instance identifier = *<value>*.

            Constraint: *<value>* <= instance identifier and instance
            identifier <= *<value>*.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 123)
            On entry, option cannot have an associated `instance
            identifier`. The supplied `instance identifier` was ignored.

            `optstr` = *<value>*.

        (`errno` 124)
            option has multiple instances. Information from the first
            instance has been returned.
    """

    try:
        _return_scalars = _b_blgm.optget(
            handle,
            optstr,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    fun_name = 'naginterfaces.library.blgm.optget'

    _optype_py = _return_scalars.optype
    _ivalue_py = _return_scalars.ivalue
    _rvalue_py = _return_scalars.rvalue
    _cvalue_py = _return_scalars.cvalue
    if _optype_py == 1:
        _optvalue_py = {}
        _optvalue_py['value'] = _ivalue_py
        _optvalue_py['annotation'] = None
    elif _optype_py == 2:
        _optvalue_py = {}
        _optvalue_py['value'] = _rvalue_py
        _optvalue_py['annotation'] = None
    elif _optype_py == 3:
        _optvalue_py = {}
        _optvalue_py['value'] = _cvalue_py
        _optvalue_py['annotation'] = None
    else:
        raise _b_utils.NagException(
            {'fun_name': fun_name},
            'inconceivable option configuration',
        )
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['optvalue'],
    )(
        _optvalue_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _optvalue_py
