# -*- coding: utf-8 -*-
# pylint: disable=chained-comparison,invalid-name,line-too-long,protected-access,too-many-arguments,too-many-boolean-expressions,too-many-branches,too-many-lines,too-many-locals,too-many-statements,unused-import
u"""
Interfaces for the NAG Mark 27.0 `complex` Chapter.

``complex`` - Complex Arithmetic

This module provides facilities for arithmetic operations involving
complex numbers.

For full information please refer to the NAG Library document

https://www.nag.com/numeric/nl/nagdoc_27/flhtml/a02/a02intro.html
"""

# NAG Copyright 2017-2020.

import collections as _collections

from ..base import complex as _b_complex

def sqrt(xr, xi):
    u"""
    Square root of complex number.

    ``sqrt`` evaluates the square root of the complex number x =
    (x_r,x_i).

    For full information please refer to the NAG Library document for
    a02aa

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/a02/a02aaf.html

    Parameters
    ----------
    xr : float
        x_r, the real part of x.

    xi : float
        x_i, the imaginary part of x.

    Returns
    -------
    yr : float
        y_r, the real part of y.

    yi : float
        y_i, the imaginary part of y.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Wilkinson, J H and Reinsch, C, 1971, `Handbook for Automatic
    Computation II, Linear Algebra`, Springer--Verlag
    """

    _return_scalars = _b_complex.sqrt(
        xr,
        xi,
    )
    _yr_py = _return_scalars.yr
    _yi_py = _return_scalars.yi
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['yr', 'yi'],
    )(
        _yr_py,
        _yi_py,
    )
    return _return_data

def abs(xr, xi): # pylint: disable=redefined-builtin
    u"""
    Modulus of complex number.

    ``abs`` returns the value of the modulus of the complex number x =
    (x_r,x_i).

    For full information please refer to the NAG Library document for
    a02ab

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/a02/a02abf.html

    Parameters
    ----------
    xr : float
        x_r, the real part of x.

    xi : float
        x_i, the imaginary part of x.

    Returns
    -------
    xa : float
        The modulus of the complex number x = (x_r,x_i).

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Wilkinson, J H and Reinsch, C, 1971, `Handbook for Automatic
    Computation II, Linear Algebra`, Springer--Verlag
    """

    _return_scalar = _b_complex.abs(
        xr,
        xi,
    )
    return _return_scalar

def divide(xr, xi, yr, yi):
    u"""
    Quotient of two complex numbers.

    ``divide`` divides one complex number, x = (x_r,x_i), by a second
    complex number, y = (y_r,y_i), returning the result in z =
    (z_r,z_i).

    For full information please refer to the NAG Library document for
    a02ac

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/a02/a02acf.html

    Parameters
    ----------
    xr : float
        x_r, the real part of x.

    xi : float
        x_i, the imaginary part of x.

    yr : float
        y_r, the real part of y.

    yi : float
        y_i, the imaginary part of y.

    Returns
    -------
    zr : float
        z_r, the real part of z.

    zi : float
        z_i, the imaginary part of z.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Li, X S, Demmel, J W, Bailey, D H, Henry, G, Hida, Y, Iskandar, J,
    Kahan, W, Kapur, A, Martin, M C, Tung, T and Yoo, D J, 2002,
    `Design, implementation and testing of extended and mixed precision
    BLAS`, ACM Trans. Math. Soft. (28(2)), 152--205
    """

    _return_scalars = _b_complex.divide(
        xr,
        xi,
        yr,
        yi,
    )
    _zr_py = _return_scalars.zr
    _zi_py = _return_scalars.zi
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['zr', 'zi'],
    )(
        _zr_py,
        _zi_py,
    )
    return _return_data
