# -*- coding: utf-8 -*-
# pylint: disable=chained-comparison,invalid-name,line-too-long,protected-access,too-many-arguments,too-many-boolean-expressions,too-many-branches,too-many-lines,too-many-locals,too-many-statements,unused-import
u"""
Interfaces for the NAG Mark 27.0 `det` Chapter.

``det`` - Determinants

This module is concerned with the calculation of determinants of square
matrices.

For full information please refer to the NAG Library document

https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f03/f03intro.html
"""

# NAG Copyright 2017-2020.

import collections as _collections

from ..base import det as _b_det
from ..base import utils as _b_utils
from ..library import _utils

def real_gen(a, ipiv):
    u"""
    Determinant of real matrix, matrix already factorized by
    ``lapacklin.dgetrf``.

    ``real_gen`` computes the determinant of a real n by n matrix A.
    ``lapacklin.dgetrf`` must be called first to supply the matrix A in
    factorized form.

    For full information please refer to the NAG Library document for
    f03ba

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f03/f03baf.html

    Parameters
    ----------
    a : float, array-like, shape (n, n)
        The n by n matrix A in factorized form as returned by
        ``lapacklin.dgetrf``.

    ipiv : int, array-like, shape (n)
        The row interchanges used to factorize matrix A as returned by
        ``lapacklin.dgetrf``.

    Returns
    -------
    d : float
        The determinant of A is given by `d`*2.0^`dexp`. It is given in
        this form to avoid overflow or underflow.

    dexp : int
        The determinant of A is given by `d`*2.0^`dexp`. It is given in
        this form to avoid overflow or underflow.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 4)
            The matrix A is approximately singular.

    References
    ----------
    Wilkinson, J H and Reinsch, C, 1971, `Handbook for Automatic
    Computation II, Linear Algebra`, Springer--Verlag
    """

    fun_name = 'naginterfaces.library.det.real_gen'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    _return_scalars = _b_det.real_gen(
        _n,
        _a,
        _ipiv,
    )
    _d_py = _return_scalars.d
    _dexp_py = _return_scalars.dexp
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['d', 'dexp'],
    )(
        _d_py,
        _dexp_py,
    )
    return _return_data

def real_sym(a):
    u"""
    Determinant of real symmetric positive definite matrix.

    ``real_sym`` computes the determinant of a real n by n symmetric
    positive definite matrix A. ``lapacklin.dpotrf`` must be called
    first to supply the symmetric matrix A in Cholesky factorized form.
    The storage (upper or lower triangular) used by ``lapacklin.dpotrf``
    is not relevant to ``real_sym`` since only the diagonal elements of
    the factorized A are referenced.

    For full information please refer to the NAG Library document for
    f03bf

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f03/f03bff.html

    Parameters
    ----------
    a : float, array-like, shape (n, n)
        The lower or upper triangle of the Cholesky factorized form of
        the n by n positive definite symmetric matrix A. Only the
        diagonal elements are referenced.

    Returns
    -------
    d : float
        The determinant of A is given by `d`*2.0^`dexp`. It is given in
        this form to avoid overflow or underflow.

    dexp : int
        The determinant of A is given by `d`*2.0^`dexp`. It is given in
        this form to avoid overflow or underflow.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n > 0.

        (`errno` 4)
            The matrix A is not positive definite.

    References
    ----------
    Wilkinson, J H and Reinsch, C, 1971, `Handbook for Automatic
    Computation II, Linear Algebra`, Springer--Verlag
    """

    fun_name = 'naginterfaces.library.det.real_sym'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _return_scalars = _b_det.real_sym(
        _n,
        _a,
    )
    _d_py = _return_scalars.d
    _dexp_py = _return_scalars.dexp
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['d', 'dexp'],
    )(
        _d_py,
        _dexp_py,
    )
    return _return_data

def real_band_sym(uplo, kd, ab):
    u"""
    Determinant of real symmetric positive definite banded matrix
    previously factorized by ``lapacklin.dpbtrf``.

    ``real_band_sym`` computes the determinant of an n by n symmetric
    positive definite banded matrix A that has been stored in
    band-symmetric storage. ``lapacklin.dpbtrf`` must be called first to
    supply the Cholesky factorized form.
    The storage (upper or lower triangular) used by ``lapacklin.dpbtrf``
    is relevant as this determines which elements of the stored
    factorized form are referenced.

    For full information please refer to the NAG Library document for
    f03bh

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f03/f03bhf.html

    Parameters
    ----------
    uplo : str, length 1
        Indicates whether the upper or lower triangular part of A was
        stored and how it was factorized. This should not be altered
        following a call to ``lapacklin.dpbtrf``.

        `uplo` = 'U'

            The upper triangular part of A was originally stored and A
            was factorized as U^TU where U is upper triangular.

        `uplo` = 'L'

            The lower triangular part of A was originally stored and A
            was factorized as LL^T where L is lower triangular.

    kd : int
        k_d, the number of superdiagonals or subdiagonals of the matrix
        A.

    ab : float, array-like, shape (`kd`+1, n)
        The Cholesky factor of A, as returned by ``lapacklin.dpbtrf``.

    Returns
    -------
    d : float
        The determinant of A is given by `d`*2.0^`dexp`. It is given in
        this form to avoid overflow or underflow.

    dexp : int
        The determinant of A is given by `d`*2.0^`dexp`. It is given in
        this form to avoid overflow or underflow.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `uplo` = *<value>*.

            Constraint: `uplo` = 'L' or 'U'.

        (`errno` 2)
            On entry, n = *<value>*.

            Constraint: n > 0.

        (`errno` 3)
            On entry, `kd` = *<value>*.

            Constraint: `kd` >= 0.

        (`errno` 6)
            The matrix A is not positive definite.

    References
    ----------
    Wilkinson, J H and Reinsch, C, 1971, `Handbook for Automatic
    Computation II, Linear Algebra`, Springer--Verlag
    """

    fun_name = 'naginterfaces.library.det.real_band_sym'

    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _return_scalars = _b_det.real_band_sym(
        uplo,
        _n,
        kd,
        _ab,
    )
    _d_py = _return_scalars.d
    _dexp_py = _return_scalars.dexp
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['d', 'dexp'],
    )(
        _d_py,
        _dexp_py,
    )
    return _return_data

def complex_gen(a, ipiv):
    u"""
    Determinant of complex matrix previously LU factorized.

    ``complex_gen`` computes the determinant of a complex n by n matrix
    A. ``lapacklin.zgetrf`` must be called first to supply the matrix A
    in factorized form.

    For full information please refer to the NAG Library document for
    f03bn

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f03/f03bnf.html

    Parameters
    ----------
    a : complex, array-like, shape (n, n)
        The n by n matrix A in factorized form as returned by
        ``lapacklin.zgetrf``.

    ipiv : int, array-like, shape (n)
        The row interchanges used to factorize matrix A as returned by
        ``lapacklin.zgetrf``.

    Returns
    -------
    d : complex
        The mantissa of the real and imaginary parts of the determinant.

    dexp : int, ndarray, shape (2)
        The exponents for the real and imaginary parts of the
        determinant. The determinant, d = (d_r,d_i), is returned as d_r
        = D_r*2^j and d_i = D_i*2^k, where `d` = (D_r,D_i) and j and k
        are stored in the first and second elements respectively of the
        array `dexp` on successful exit.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 4)
            The matrix A is approximately singular.

    References
    ----------
    Wilkinson, J H and Reinsch, C, 1971, `Handbook for Automatic
    Computation II, Linear Algebra`, Springer--Verlag
    """

    fun_name = 'naginterfaces.library.det.complex_gen'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    _dexp = _b_utils._EngineIntArrayType.empty_ndarray(
        (2,),
    )
    _return_scalar = _b_det.complex_gen(
        _n,
        _a,
        _ipiv,
        _dexp,
    )
    _d_py = _return_scalar
    _dexp_py = _dexp
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['d', 'dexp'],
    )(
        _d_py,
        _dexp_py,
    )
    return _return_data
