# -*- coding: utf-8 -*-
# pylint: disable=chained-comparison,invalid-name,line-too-long,protected-access,too-many-arguments,too-many-boolean-expressions,too-many-branches,too-many-lines,too-many-locals,too-many-statements,unused-import
u"""
Interfaces for the NAG Mark 27.0 `eigen` Chapter.

``eigen`` - Eigenvalues and Eigenvectors

This module provides functions for various types of matrix eigenvalue
problem:

    standard eigenvalue problems (finding eigenvalues and eigenvectors
    of a square matrix A);

    singular value problems (finding singular values and singular
    vectors of a rectangular matrix A);

    generalized eigenvalue problems (finding eigenvalues and
    eigenvectors of a matrix pencil A-lambda B).

    quadratic eigenvalue problems (finding eigenvalues and eigenvectors
    of the quadratic lambda^2A+lambda B+C).

Functions are provided for both `real` and `complex` data.

The majority of functions for these problems can be found in submodule
``lapackeig`` which contains software derived from LAPACK (see Anderson
`et al.` (1999)).
However, you should read [the F02 Introduction] before turning to
submodule ``lapackeig``, especially if you are a new user. Submodule
``sparseig`` contains functions for large sparse eigenvalue problems,
although one such function is also available in this module.

Submodule ``eigen`` and submodule ``lapackeig`` contain **Black Box**
(or **Driver**) functions that enable many problems to be solved by a
call to a single function, and the decision trees in [Decision Trees]
direct you to the most appropriate functions in submodule ``eigen`` and
submodule ``lapackeig``.
The submodule ``eigen`` functions call functions in submodule
``lapacklin`` and submodule ``lapackeig`` wherever possible to perform
the computations, and there are pointers in [Decision Trees] to the
relevant decision trees in submodule ``lapackeig``.

For full information please refer to the NAG Library document

https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f02/f02intro.html
"""

# NAG Copyright 2017-2020.

import collections as _collections

from ..base import eigen as _b_eigen
from ..base import utils as _b_utils
from ..library import _utils

def real_gen_eigsys(crit, a, wl, wu, mest):
    u"""
    Selected eigenvalues and eigenvectors of real nonsymmetric matrix
    (Black Box).

    ``real_gen_eigsys`` computes selected eigenvalues and eigenvectors
    of a real general matrix.

    For full information please refer to the NAG Library document for
    f02ec

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f02/f02ecf.html

    Parameters
    ----------
    crit : str, length 1
        Indicates the criterion for selecting eigenvalues.

        `crit` = 'M'

            Eigenvalues are selected according to their moduli: w_l <=
            |lambda_i| <= w_u.

        `crit` = 'R'

            Eigenvalues are selected according to their real parts: w_l
            <= Re(lambda_i) <= w_u.

    a : float, array-like, shape (n, n)
        The n by n general matrix A.

    wl : float
        w_l and w_u, the lower and upper bounds on the criterion for the
        selected eigenvalues (see `crit`).

    wu : float
        w_l and w_u, the lower and upper bounds on the criterion for the
        selected eigenvalues (see `crit`).

    mest : int
        `mest` must be an upper bound on m, the number of eigenvalues
        and eigenvectors selected. No eigenvectors are computed if
        `mest` < m.

    Returns
    -------
    a : float, ndarray, shape (n, n)
        Contains the Hessenberg form of the balanced input matrix A'
        (see [Further Comments]).

    m : int
        m, the number of eigenvalues actually selected.

    wr : float, ndarray, shape (n)
        The first `m` elements of `wr` and `wi` hold the real and
        imaginary parts, respectively, of the selected eigenvalues;
        elements `m`+1 to n contain the other eigenvalues. Complex
        conjugate pairs of eigenvalues are stored in consecutive
        elements of the arrays, with the eigenvalue having positive
        imaginary part first. See also [Further Comments].

    wi : float, ndarray, shape (n)
        The first `m` elements of `wr` and `wi` hold the real and
        imaginary parts, respectively, of the selected eigenvalues;
        elements `m`+1 to n contain the other eigenvalues. Complex
        conjugate pairs of eigenvalues are stored in consecutive
        elements of the arrays, with the eigenvalue having positive
        imaginary part first. See also [Further Comments].

    vr : float, ndarray, shape (n, `mest`)
        Contains the real parts of the selected eigenvectors, with the
        ith column holding the real part of the eigenvector associated
        with the eigenvalue lambda_i (stored in `wr`[i-1] and
        `wi`[i-1]).

    vi : float, ndarray, shape (n, `mest`)
        Contains the imaginary parts of the selected eigenvectors, with
        the ith column holding the imaginary part of the eigenvector
        associated with the eigenvalue lambda_i (stored in `wr`[i-1] and
        `wi`[i-1]).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `mest` = *<value>*.

            Constraint: `mest` >= 1.

        (`errno` 1)
            On entry, `wu` = *<value>* and `wl` = *<value>*.

            Constraint: `wu` > `wl`.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 0.

        (`errno` 1)
            On entry, `crit` = *<value>*.

            Constraint: `crit` = 'M' or 'R'.

        (`errno` 2)
            The QR algorithm failed to converge: only *<value>*
            eigenvalues have been computed; no eigenvectors have been
            computed.

        (`errno` 3)
            There are more than `mest` eigenvalues in the specified
            range. `m` (number of eigenvalues in range) = *<value>* and
            `mest` = *<value>*. No eigenvectors have been computed.
            Rerun with second dimension of `vr` and `vi` = `mest` >=
            `m`.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 4)
            Inverse iteration failed to compute all the specified
            eigenvectors. The number of eigenvectors which failed to
            converge is *<value>*. The corresponding columns of `vr` and
            `vi` are set to zero.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.eigen.real_gen_eigsys'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_a, 'a'),),
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _wr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _wi = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _mest = _b_utils._EngineIntScalarType(
        mest, {'fun_name': fun_name, 'entity': 'mest'},
    ).to_ctypes()
    _mest_py = _b_utils._EngineIntScalarType.to_py(_mest)
    _vr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py, _mest_py),
        sorder=_m_sorder,
    )
    _vi = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py, _mest_py),
        sorder=_m_sorder,
    )
    try:
        _return_scalar = _b_eigen.real_gen_eigsys(
            crit,
            _n,
            _a,
            wl,
            wu,
            _mest,
            _wr,
            _wi,
            _vr,
            _vi,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _a_py = _a
    _m_py = _return_scalar
    _wr_py = _wr
    _wi_py = _wi
    _vr_py = _vr
    _vi_py = _vi
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'm', 'wr', 'wi', 'vr', 'vi'],
    )(
        _a_py,
        _m_py,
        _wr_py,
        _wi_py,
        _vr_py,
        _vi_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def real_gen_sparse_arnoldi(a, icolzp, irowix, nev, ncv, sigma, monit=None, option=None, data=None, io_manager=None):
    u"""
    Selected eigenvalues and eigenvectors of a real sparse general
    matrix.

    Note: this function uses optional algorithmic parameters, see also:
    ``sparseig.real_init``, ``sparseig.real_iter``,
    ``sparseig.real_proc``, ``sparseig.real_option``,
    ``sparseig.real_monit``.

    ``real_gen_sparse_arnoldi`` computes selected eigenvalues and
    eigenvectors of a real sparse general matrix.

    For full information please refer to the NAG Library document for
    f02ek

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f02/f02ekf.html

    Parameters
    ----------
    a : float, array-like, shape (nnz)
        The array of nonzero elements (and diagonal elements if a
        nonzero inverse shift is to be applied) of the n by n general
        matrix A.

    icolzp : int, array-like, shape (n+1)
        `icolzp`[i-1] contains the index in `a` of the start of column
        i, for i = 1,2,...,n; `icolzp`[n] must contain the value nnz+1.
        Thus the number of nonzero elements in column i of A is
        `icolzp`[i]-`icolzp`[i-1]; when shifts are applied this includes
        diagonal elements irrespective of value. See [the F11
        Introduction].

    irowix : int, array-like, shape (nnz)
        `irowix`[i-1] contains the row index for each entry in `a`. See
        [the F11 Introduction].

    nev : int
        The number of eigenvalues to be computed.

    ncv : int
        The number of Arnoldi basis vectors to use during the
        computation.

        At present there is no `a priori` analysis to guide the
        selection of `ncv` relative to `nev`.

        However, it is recommended that `ncv` >= 2*`nev`+1.

        If many problems of the same type are to be solved, you should
        experiment with increasing `ncv` while keeping `nev` fixed for a
        given test problem.

        This will usually decrease the required number of matrix-vector
        operations but it also increases the work and storage required
        to maintain the orthogonal basis vectors.

        The optimal 'cross-over' with respect to CPU time is problem
        dependent and must be determined empirically.

    sigma : float
        If the 'Shifted Inverse Real' mode has been selected then
        `sigma` contains the real shift used; otherwise `sigma` is not
        referenced. This mode can be selected by setting the appropriate
        options in the user-supplied function `option`.

    monit : None or callable monit(niter, nconv, w, rzest, data=None),
        optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `monit` is used to monitor the progress of
        ``real_gen_sparse_arnoldi``. `monit` may be None if no
        monitoring is actually required. `monit` is called after the
        solution of each eigenvalue sub-problem and also just prior to
        return from ``real_gen_sparse_arnoldi``.

        Parameters
        ~~~~~~~~~~
        niter : int
            The number of the current Arnoldi iteration.

        nconv : int
            The number of converged eigenvalues so far.

        w : complex, ndarray, shape (ncv)
            The first `nconv` elements of `w` contain the converged
            approximate eigenvalues.

        rzest : float, ndarray, shape (ncv)
            The first `nconv` elements of `rzest` contain the Ritz
            estimates (error bounds) on the converged approximate
            eigenvalues.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

    option : None or callable option(comm, data=None), optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        You can supply non-default options to the Arnoldi eigensolver by
        repeated calls to ``sparseig.real_option`` from within `option`.
        (Please note that it is only necessary to call
        ``sparseig.real_option``; no call to ``sparseig.real_init`` is
        required from within `option`.) For example, you can set the
        mode to 'Shifted Inverse Real', you can increase the 'Iteration
        Limit' beyond its default and you can print varying levels of
        detail on the iterative process using 'Print Level'.

        If only the default options (including that the eigenvalues of
        largest magnitude are sought) are to be used then `option` may
        be None.

        Parameters
        ~~~~~~~~~~
        comm : dict, communication object, modifiable in place
            Communication structure.

            This argument has been initialized to hold the default
            option set.

            If you wish to supply non-default options to the Arnoldi
            eigensolver, this argument may be passed as argument comm in
            a call to ``sparseig.real_option``.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

    data : arbitrary, optional
        User-communication data for callback functions.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    a : float, ndarray, shape (nnz)
        If a nonzero shifted inverse is to be applied then the diagonal
        elements of A have the shift value, as supplied in `sigma`,
        subtracted.

    nconv : int
        The number of converged approximations to the selected
        eigenvalues. On successful exit, this will normally be either
        `nev` or `nev`+1 depending on the number of complex conjugate
        pairs of eigenvalues returned.

    w : complex, ndarray, shape (`ncv`)
        The first `nconv` elements contain the converged approximations
        to the selected eigenvalues. Since complex conjugate pairs of
        eigenvalues appear together, it is possible (given an odd number
        of converged real eigenvalues) for ``real_gen_sparse_arnoldi``
        to return one more eigenvalue than requested.

    v : float, ndarray, shape (n, `ncv`)
        Contains the eigenvectors associated with the eigenvalue
        lambda_i, for i = 1,2,...,`nconv` (stored in `w`). For a real
        eigenvalue, lambda_j, the corresponding eigenvector is real and
        is stored in `v`[i-1,j-1], for i = 1,2,...,n. For complex
        conjugate pairs of eigenvalues, w_{j+1} = w_j_, the real and
        imaginary parts of the corresponding eigenvectors are stored,
        respectively, in `v`[i-1,j-1] and `v`[i-1,j], for i = 1,2,...,n.
        The imaginary parts stored are for the first of the conjugate
        pair of eigenvectors; the other eigenvector in the pair is
        obtained by negating these imaginary parts.

    resid : float, ndarray, shape (`nev`+1)
        The residual ||Aw_i-lambda_i w_i||_2 for the estimates to the
        eigenpair lambda_i and w_i is returned in `resid`[i-1], for i =
        1,2,...,`nconv`.

    Other Parameters
    ----------------
    'Advisory' : int
        Default = 0

        If the option 'List' is set then option specifications are
        listed in a 'List' file by setting the option to a file
        identification (unit) number associated with advisory messages
        (see :class:`~naginterfaces.base.utils.FileObjManager` and the
        ``FileObjManager`` method
        :meth:`~naginterfaces.base.utils.FileObjManager.unit_from_fileobj`).

    'Defaults' : valueless
        This special keyword may be used to reset all options to their
        default values.

    'Iteration Limit' : int
        Default = 300

        The limit on the number of Arnoldi iterations that can be
        performed before ``real_gen_sparse_arnoldi`` exits with `errno`
        = 24.

    'Largest Magnitude' : valueless
        Default

        The Arnoldi iterative method converges on a number of
        eigenvalues with given properties.
        The default is to compute the eigenvalues of largest magnitude
        using 'Largest Magnitude'.
        Alternatively, eigenvalues may be chosen which have 'Largest
        Real' part, 'Largest Imaginary' part, 'Smallest Magnitude',
        'Smallest Real' part or 'Smallest Imaginary' part.

        Note that these options select the eigenvalue properties for
        eigenvalues of op the linear operator determined by the
        computational mode and problem type.

    'Largest Imaginary' : valueless
        The Arnoldi iterative method converges on a number of
        eigenvalues with given properties.
        The default is to compute the eigenvalues of largest magnitude
        using 'Largest Magnitude'.
        Alternatively, eigenvalues may be chosen which have 'Largest
        Real' part, 'Largest Imaginary' part, 'Smallest Magnitude',
        'Smallest Real' part or 'Smallest Imaginary' part.

        Note that these options select the eigenvalue properties for
        eigenvalues of op the linear operator determined by the
        computational mode and problem type.

    'Largest Real' : valueless
        The Arnoldi iterative method converges on a number of
        eigenvalues with given properties.
        The default is to compute the eigenvalues of largest magnitude
        using 'Largest Magnitude'.
        Alternatively, eigenvalues may be chosen which have 'Largest
        Real' part, 'Largest Imaginary' part, 'Smallest Magnitude',
        'Smallest Real' part or 'Smallest Imaginary' part.

        Note that these options select the eigenvalue properties for
        eigenvalues of op the linear operator determined by the
        computational mode and problem type.

    'Smallest Imaginary' : valueless
        The Arnoldi iterative method converges on a number of
        eigenvalues with given properties.
        The default is to compute the eigenvalues of largest magnitude
        using 'Largest Magnitude'.
        Alternatively, eigenvalues may be chosen which have 'Largest
        Real' part, 'Largest Imaginary' part, 'Smallest Magnitude',
        'Smallest Real' part or 'Smallest Imaginary' part.

        Note that these options select the eigenvalue properties for
        eigenvalues of op the linear operator determined by the
        computational mode and problem type.

    'Smallest Magnitude' : valueless
        The Arnoldi iterative method converges on a number of
        eigenvalues with given properties.
        The default is to compute the eigenvalues of largest magnitude
        using 'Largest Magnitude'.
        Alternatively, eigenvalues may be chosen which have 'Largest
        Real' part, 'Largest Imaginary' part, 'Smallest Magnitude',
        'Smallest Real' part or 'Smallest Imaginary' part.

        Note that these options select the eigenvalue properties for
        eigenvalues of op the linear operator determined by the
        computational mode and problem type.

    'Smallest Real' : valueless
        The Arnoldi iterative method converges on a number of
        eigenvalues with given properties.
        The default is to compute the eigenvalues of largest magnitude
        using 'Largest Magnitude'.
        Alternatively, eigenvalues may be chosen which have 'Largest
        Real' part, 'Largest Imaginary' part, 'Smallest Magnitude',
        'Smallest Real' part or 'Smallest Imaginary' part.

        Note that these options select the eigenvalue properties for
        eigenvalues of op the linear operator determined by the
        computational mode and problem type.

    'Nolist' : valueless
        Default

        Option 'List' enables printing of each option specification as
        it is supplied. 'Nolist' suppresses this printing.

    'List' : valueless
        Option 'List' enables printing of each option specification as
        it is supplied. 'Nolist' suppresses this printing.

    'Monitoring' : int
        Default = -1

        If i > 0, monitoring information is output to unit number (see
        :meth:`~naginterfaces.base.utils.FileObjManager.unit_from_fileobj`)
        i during the solution of each problem; this may be the same as
        the 'Advisory' unit number (see
        :meth:`~naginterfaces.base.utils.FileObjManager.unit_from_fileobj`).
        The type of information produced is dependent on the value of
        'Print Level', see the description of the option 'Print Level'
        for details of the information produced.
        Please see the ``FileObjManager`` method
        :meth:`~naginterfaces.base.utils.FileObjManager.unit_from_fileobj`
        to associate a file with a given unit number (see
        :meth:`~naginterfaces.base.utils.FileObjManager.unit_from_fileobj`).

    'Print Level' : int
        Default = 0

        This controls the amount of printing produced by
        ``real_gen_sparse_arnoldi`` as follows.

        [table omitted]

    'Regular' : valueless
        Default

        These options define the computational mode which in turn
        defines the form of operation op(x) to be performed.

        Given a standard eigenvalue problem in the form Ax = lambda x
        then the following modes are available with the appropriate
        operator op(x).

        [table omitted]

    'Shifted Inverse Real' : valueless
        These options define the computational mode which in turn
        defines the form of operation op(x) to be performed.

        Given a standard eigenvalue problem in the form Ax = lambda x
        then the following modes are available with the appropriate
        operator op(x).

        [table omitted]

    'Tolerance' : float
        Default = epsilon

        An approximate eigenvalue has deemed to have converged when the
        corresponding Ritz estimate is within 'Tolerance' relative to
        the magnitude of the eigenvalue.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n > 0.

        (`errno` 2)
            On entry, nnz = *<value>*.

            Constraint: nnz > 0.

        (`errno` 2)
            On entry, nnz = *<value>* and n = *<value>*.

            Constraint: nnz <= n*n.

        (`errno` 3)
            On entry, in shifted inverse mode, the jth diagonal element
            of A is not defined, for j = *<value>*.

        (`errno` 4)
            On entry, `icolzp`[0] = *<value>*.

            Constraint: `icolzp`[0] = 1.

        (`errno` 4)
            On entry, for i = *<value>*, `icolzp`[i-1] = *<value>* and
            `icolzp`[i] = *<value>*.

            Constraint: `icolzp`[i-1] < `icolzp`[i].

        (`errno` 4)
            On entry, `icolzp`[n] = *<value>* and nnz = *<value>*.

            Constraint: `icolzp`[n] = nnz+1.

        (`errno` 5)
            On entry, in specification of column *<value>*, and for j =
            *<value>*, `irowix`[j-1] = *<value>* and `irowix`[j] =
            *<value>*.

            Constraint: `irowix`[j-1] < `irowix`[j].

        (`errno` 6)
            On entry, `nev` = *<value>*.

            Constraint: `nev` > 0.

        (`errno` 7)
            On entry, `ncv` = *<value>* and `nev` = *<value>*.

            Constraint: `ncv` > `nev`+1.

        (`errno` 7)
            On entry, `ncv` = *<value>* and n = *<value>*.

            Constraint: `ncv` <= n.

        (`errno` 8)
            On entry, the matrix A-sigma*I is numerically singular and
            could not be inverted. Try perturbing the value of sigma.

        (`errno` 8)
            On entry, the matrix A-sigma*I is nearly numerically
            singular and could not be inverted. Try perturbing the value
            of sigma. Norm of matrix = *<value>*, Reciprocal condition
            number = *<value>*.

        (`errno` 21)
            The maximum number of iterations <= 0, the option 'Iteration
            Limit' has been set to *<value>*.

        (`errno` 22)
            An unexpected internal error occurred when solving an
            eigenproblem.

            This error should not occur. Please contact NAG.

        (`errno` 23)
            An unexpected internal error occurred when solving an
            eigenproblem.

            This error should not occur. Please contact NAG.

        (`errno` 24)
            The maximum number of iterations has been reached.

            The maximum number of iterations = *<value>*.

            The number of converged eigenvalues = *<value>*.

            See the function document for further details.

        (`errno` 25)
            No shifts could be applied during a cycle of the implicitly
            restarted Arnoldi iteration.

        (`errno` 26)
            Could not build an Arnoldi factorization. The size of the
            current Arnoldi factorization = *<value>*.

        (`errno` 27)
            Error in internal call to compute eigenvalues and
            corresponding error bounds of the current upper Hessenberg
            matrix.

            Please contact NAG.

        (`errno` 32)
            An unexpected internal error occurred when postprocessing an
            eigenproblem.

            This error should not occur. Please contact NAG.

        (`errno` 33)
            The number of eigenvalues found to sufficient accuracy is
            zero.

        (`errno` 34)
            Internal inconsistency in the number of converged Ritz
            values. Number counted = *<value>*, number expected =
            *<value>*.

        (`errno` 35)
            During calculation of a real Schur form, there was a failure
            to compute *<value>* eigenvalues in a total of *<value>*
            iterations.

        (`errno` 36)
            The computed Schur form could not be reordered by an
            internal call.

            This routine returned with errno = *<value>*.

            Please contact NAG.

        (`errno` 37)
            In calculating eigenvectors, an internal call returned with
            an error.

            Please contact NAG.

    Warns
    -----
    NagCallbackTerminateWarning
        (`errno` 9)
            User requested termination in `monit`, `istat` = *<value>*.

        (`errno` 10)
            User requested termination in `option`, `istat` = *<value>*.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Lehoucq, R B, Sorensen, D C and Yang, C, 1998, `ARPACK Users' Guide:
    Solution of Large-scale Eigenvalue Problems with Implicitly
    Restarted Arnoldi Methods`, SIAM, Philadelphia
    """

    def cb_monit(ncv, niter, nconv, w, rzest, data=None):
        """Wrapping user's monit. Called as base monit."""
        if data is not None:
            monit(
                niter,
                nconv,
                w,
                rzest,
                data,
            )
        else:
            monit(
                niter,
                nconv,
                w,
                rzest,
            )

    def cb_option(comm, data=None):
        """Wrapping user's option. Called as base option."""
        if data is not None:
            option(
                comm,
                data,
            )
        else:
            option(
                comm,
            )

    fun_name = 'naginterfaces.library.eigen.real_gen_sparse_arnoldi'

    _icolzp = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'icolzp'},
        icolzp,
        copy=False,
    )
    _n = _b_utils._EngineIntScalarType(
        _b_utils._EngineIntArrayType.get_shape(_icolzp, {'fun_name': fun_name, 'entity': 'icolzp'}, exp_rank=1)[0]-1, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nnz_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=1)[0]
    _nnz = _b_utils._EngineIntScalarType(
        _nnz_py, {'fun_name': fun_name, 'entity': '_nnz'},
    ).to_ctypes()
    _irowix = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'irowix'},
        irowix,
        copy=False,
    )
    _b_utils._check_callable(monit, {'fun_name': fun_name, 'entity': 'monit'}, allow_none=True)
    _monit = None if monit is None else cb_monit
    _b_utils._check_callable(option, {'fun_name': fun_name, 'entity': 'option'}, allow_none=True)
    _option = None if option is None else cb_option
    _ncv = _b_utils._EngineIntScalarType(
        ncv, {'fun_name': fun_name, 'entity': 'ncv'},
    ).to_ctypes()
    _ncv_py = _b_utils._EngineIntScalarType.to_py(_ncv)
    _w = _b_utils._EngineComplex128ArrayType.empty_ndarray(
        (_ncv_py,),
    )
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _v = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py, _ncv_py),
        sorder=_b_utils._EngineData.default_sorder,
    )
    _nev = _b_utils._EngineIntScalarType(
        nev, {'fun_name': fun_name, 'entity': 'nev'},
    ).to_ctypes()
    _nev_py = _b_utils._EngineIntScalarType.to_py(_nev)
    _resid = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nev_py+1,),
    )
    try:
        _return_scalar = _b_eigen.real_gen_sparse_arnoldi(
            _n,
            _nnz,
            _a,
            _icolzp,
            _irowix,
            _nev,
            _ncv,
            sigma,
            _monit,
            _option,
            _w,
            _v,
            _resid,
            data,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _a_py = _a
    _nconv_py = _return_scalar
    _w_py = _w
    _v_py = _v
    _resid_py = _resid
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'nconv', 'w', 'v', 'resid'],
    )(
        _a_py,
        _nconv_py,
        _w_py,
        _v_py,
        _resid_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def real_symm_sparse_eigsys(m, noits, tol, dot, image, novecs, x, monit=None, data=None):
    u"""
    Selected eigenvalues and eigenvectors of sparse symmetric
    eigenproblem (Black Box).

    ``real_symm_sparse_eigsys`` finds eigenvalues and eigenvectors of a
    real sparse symmetric or generalized symmetric eigenvalue problem.

    For full information please refer to the NAG Library document for
    f02fj

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f02/f02fjf.html

    Parameters
    ----------
    m : int
        m, the number of eigenvalues required.

    noits : int
        The maximum number of major iterations (eigenvalue sub-problems)
        to be performed. If `noits` <= 0, the value 100 is used in place
        of `noits`.

    tol : float
        A relative tolerance to be used in accepting eigenvalues and
        eigenvectors. If the eigenvalues are required to about t
        significant figures, `tol` should be set to about 10^(-t). d_i
        is accepted as an eigenvalue as soon as two successive
        approximations to d_i differ by less than (|d~_i|*`tol`)/10,
        where d~_i is the latest approximation to d_i. Once an
        eigenvalue has been accepted, an eigenvector is accepted as soon
        as (d_i f_i)/(d_i-d_k) < `tol`, where f_i is the normalized
        residual of the current approximation to the eigenvector (see
        [Further Comments] for further information). The values of the
        f_i and d_i can be printed from `monit`. If `tol` is supplied
        outside the range (epsilon,1.0), where epsilon is the machine
        precision, the value epsilon is used in place of `tol`.

    dot : callable retval = dot(z, w, data=None)
        `dot` must return the value w^TBz for given vectors w and z.

        For the standard eigenvalue problem, where B = I, `dot` must
        return the dot product w^Tz.

        Parameters
        ~~~~~~~~~~
        z : float, ndarray, shape (n)
            The vector z for which w^TBz is required.

        w : float, ndarray, shape (n)
            The vector w for which w^TBz is required.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        retval : float
            The value w^TBz.

    image : callable w = image(z, data=None)
        `image` must return the vector w = Cz for a given vector z.

        Parameters
        ~~~~~~~~~~
        z : float, ndarray, shape (n)
            The vector z for which Cz is required.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        w : float, array-like, shape (n)
            The vector w = Cz.

    novecs : int
        The number of approximate vectors that are being supplied in
        `x`. If `novecs` is outside the range (0,k), the value 0 is used
        in place of `novecs`.

    x : float, array-like, shape (n, k)
        If 0 < `novecs` <= k, the first `novecs` columns of `x` must
        contain approximations to the eigenvectors corresponding to the
        `novecs` eigenvalues of largest absolute value of C. Supplying
        approximate eigenvectors can be useful when reasonable
        approximations are known, or when ``real_symm_sparse_eigsys`` is
        being restarted with a larger value of k. Otherwise it is not
        necessary to supply approximate vectors, as simultaneous
        iteration vectors will be generated randomly by
        ``real_symm_sparse_eigsys``.

    monit : None or callable monit(istate, nextit, nevals, nevecs, f, d,
        data=None), optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `monit` is used to monitor the progress of
        ``real_symm_sparse_eigsys``. `monit` may be None if no
        monitoring is actually required. `monit` is called after the
        solution of each eigenvalue sub-problem and also just prior to
        return from ``real_symm_sparse_eigsys``.

        The arguments `istate` and `nextit` allow selective printing by
        `monit`.

        Parameters
        ~~~~~~~~~~
        istate : int
            Specifies the state of ``real_symm_sparse_eigsys``.

            `istate` = 0

                No eigenvalue or eigenvector has just been accepted.

            `istate` = 1

                One or more eigenvalues have been accepted since the
                last call to `monit`.

            `istate` = 2

                One or more eigenvectors have been accepted since the
                last call to `monit`.

            `istate` = 3

                One or more eigenvalues and eigenvectors have been
                accepted since the last call to `monit`.

            `istate` = 4

                Return from ``real_symm_sparse_eigsys`` is about to
                occur.

        nextit : int
            The number of the next iteration.

        nevals : int
            The number of eigenvalues accepted so far.

        nevecs : int
            The number of eigenvectors accepted so far.

        f : float, ndarray, shape (k)
            A vector of error quantities measuring the state of
            convergence of the simultaneous iteration vectors. See `tol`
            and [Further Comments] for further details. Each element of
            `f` is initially set to the value 4.0 and an element remains
            at 4.0 until the corresponding vector is tested.

        d : float, ndarray, shape (k)
            `d`[i-1] contains the latest approximation to the absolute
            value of the ith eigenvalue of C.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

    data : arbitrary, optional
        User-communication data for callback functions.

    Returns
    -------
    m : int
        m', the number of eigenvalues actually found. It is equal to m
        if no exception or warning is raised on exit, and is less than m
        if `errno` = 2, 3 or 4. See [Raises/Warns] and [Further
        Comments] for further information.

    noits : int
        The number of iterations actually performed.

    x : float, ndarray, shape (n, k)
        If the function exits successfully or `errno` = 2, 3 or 4, the
        first m' columns contain the eigenvectors corresponding to the
        eigenvalues returned in the first m' elements of `d`; and the
        next k-m'-1 columns contain approximations to the eigenvectors
        corresponding to the approximate eigenvalues returned in the
        next k-m'-1 elements of `d`. Here m' is the value returned in
        `m`, the number of eigenvalues actually found. The kth column is
        used as workspace.

    d : float, ndarray, shape (k)
        If the function exits successfully or `errno` = 2, 3 or 4, the
        first m' elements contain the first m' eigenvalues in decreasing
        order of magnitude; and the next k-m'-1 elements contain
        approximations to the next k-m'-1 eigenvalues. Here m' is the
        value returned in `m`, the number of eigenvalues actually found.
        `d`[k-1] contains the value e where (-e,e) is the latest
        interval over which Chebyshev acceleration is performed.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, k = *<value>* and n = *<value>*.

            Constraint: k <= n.

        (`errno` 1)
            On entry, `m` = *<value>* and k = *<value>*.

            Constraint: `m` < k.

        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 5)
            Not all requested eigenvalues and vectors have been
            obtained.

        (`errno` 5)
            Convergence of eigenvalue sub-problem occurred.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            Not all requested eigenvalues and vectors have been
            obtained.

        (`errno` 3)
            Not all requested eigenvalues and vectors have been
            obtained.

        (`errno` 4)
            Not all requested eigenvalues and vectors have been
            obtained.

    NagCallbackTerminateWarning
        (`errno` i < 0)
            User set `iflag` negative in `dot` or `image`.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Nikolai, P J, 1979, `Algorithm 538: Eigenvectors and eigenvalues of
    real generalized symmetric matrices by simultaneous iteration`, ACM
    Trans. Math. Software (5), 118--125

    Parlett, B N, 1998, `The Symmetric Eigenvalue Problem`, SIAM,
    Philadelphia

    Rutishauser, H, 1969, `Computational aspects of F L Bauer's
    simultaneous iteration method`, Numer. Math. (13), 4--13

    Rutishauser, H, 1970, `Simultaneous iteration method for symmetric
    matrices`, Numer. Math. (16), 205--223
    """

    def cb_dot(n, z, w, data=None):
        """Wrapping user's dot. Called as base dot."""
        if data is not None:
            _retval_dot = dot(
                z,
                w,
                data,
            )
        else:
            _retval_dot = dot(
                z,
                w,
            )
        return _retval_dot

    fun_name = 'naginterfaces.library.eigen.real_symm_sparse_eigsys'

    cb_image_name = fun_name + ':image'

    def cb_image(n, z, w, data=None):
        """Wrapping user's image. Called as base image."""
        if data is not None:
            _w_image = image(
                z,
                data,
            )
        else:
            _w_image = image(
                z,
            )
        w[:] = _utils._to_ndarray(
            {'fun_name': cb_image_name, 'returned': 'w'},
            _w_image,
            copy=False,
            exp_shape=(n,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )

    def cb_monit(istate, nextit, nevals, nevecs, k, f, d, data=None):
        """Wrapping user's monit. Called as base monit."""
        if data is not None:
            monit(
                istate,
                nextit,
                nevals,
                nevecs,
                f,
                d,
                data,
            )
        else:
            monit(
                istate,
                nextit,
                nevals,
                nevecs,
                f,
                d,
            )

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _k_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2)[1]
    _k = _b_utils._EngineIntScalarType(
        _k_py, {'fun_name': fun_name, 'entity': '_k'},
    ).to_ctypes()
    _b_utils._check_callable(dot, {'fun_name': fun_name, 'entity': 'dot'})
    _dot = cb_dot
    _b_utils._check_callable(image, {'fun_name': fun_name, 'entity': 'image'})
    _image = cb_image
    _b_utils._check_callable(monit, {'fun_name': fun_name, 'entity': 'monit'}, allow_none=True)
    _monit = None if monit is None else cb_monit
    _d = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_k_py,),
    )
    try:
        _return_scalars = _b_eigen.real_symm_sparse_eigsys(
            _n,
            m,
            _k,
            noits,
            tol,
            _dot,
            _image,
            _monit,
            novecs,
            _x,
            _d,
            data,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _m_py = _return_scalars.m
    _noits_py = _return_scalars.noits
    _x_py = _x
    _d_py = _d
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['m', 'noits', 'x', 'd'],
    )(
        _m_py,
        _noits_py,
        _x_py,
        _d_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def real_symm_sparse_arnoldi(n, a, irow, icol, nev, ncv, sigma, monit=None, option=None, data=None, io_manager=None):
    u"""
    Selected eigenvalues and eigenvectors of a real symmetric sparse
    matrix.

    Note: this function uses optional algorithmic parameters, see also:
    ``sparseig.real_symm_init``, ``sparseig.real_symm_iter``,
    ``sparseig.real_symm_proc``, ``sparseig.real_symm_option``,
    ``sparseig.real_symm_monit``.

    ``real_symm_sparse_arnoldi`` computes selected eigenvalues and
    eigenvectors of a real sparse symmetric matrix.

    For full information please refer to the NAG Library document for
    f02fk

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f02/f02fkf.html

    Parameters
    ----------
    n : int
        n, the order of the matrix A.

    a : float, array-like, shape (nnz)
        The array of nonzero elements of the lower triangular part of
        the n by n symmetric matrix A.

    irow : int, array-like, shape (nnz)
        The row and column indices of the elements supplied in array
        `a`.

        If `irow`[k-1] = i and `icol`[k-1] = j then A_{ij} is stored in
        `a`[k-1]. `irow` does not need to be ordered: an internal sort
        will force the correct ordering.

    icol : int, array-like, shape (nnz)
        The row and column indices of the elements supplied in array
        `a`.

        If `irow`[k-1] = i and `icol`[k-1] = j then A_{ij} is stored in
        `a`[k-1]. `irow` does not need to be ordered: an internal sort
        will force the correct ordering.

    nev : int
        The number of eigenvalues to be computed.

    ncv : int
        The number of Arnoldi basis vectors to use during the
        computation.

        At present there is no `a priori` analysis to guide the
        selection of `ncv` relative to `nev`.

        However, it is recommended that `ncv` >= 2*`nev`+1.

        If many problems of the same type are to be solved, you should
        experiment with increasing `ncv` while keeping `nev` fixed for a
        given test problem.

        This will usually decrease the required number of matrix-vector
        operations but it also increases the work and storage required
        to maintain the orthogonal basis vectors.

        The optimal 'cross-over' with respect to computation time is
        problem dependent and must be determined empirically.

    sigma : float
        If the 'Shifted Inverse' mode has been selected then `sigma`
        contains the real shift used; otherwise `sigma` is not
        referenced. This mode can be selected by setting the appropriate
        options in the user-supplied function `option`.

    monit : None or callable monit(niter, nconv, w, rzest, data=None),
        optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `monit` is used to monitor the progress of
        ``real_symm_sparse_arnoldi``. `monit` may be None if no
        monitoring is actually required. `monit` is called after the
        solution of each eigenvalue sub-problem and also just prior to
        return from ``real_symm_sparse_arnoldi``.

        Parameters
        ~~~~~~~~~~
        niter : int
            The number of the current Arnoldi iteration.

        nconv : int
            The number of converged eigenvalues so far.

        w : float, ndarray, shape (ncv)
            The first `nconv` elements of `w` contain the converged
            approximate eigenvalues.

        rzest : float, ndarray, shape (ncv)
            The first `nconv` elements of `rzest` contain the Ritz
            estimates (error bounds) on the converged approximate
            eigenvalues.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

    option : None or callable option(comm, data=None), optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        You can supply non-default options to the Arnoldi eigensolver by
        repeated calls to ``sparseig.real_symm_option`` from within
        `option`. (Please note that it is only necessary to call
        ``sparseig.real_symm_option``; no call to
        ``sparseig.real_symm_init`` is required from within `option`.)
        For example, you can set the mode to 'Shifted Inverse', you can
        increase the 'Iteration Limit' beyond its default and you can
        print varying levels of detail on the iterative process using
        'Print Level'.

        If only the default options (including that the eigenvalues of
        largest magnitude are sought) are to be used then `option` may
        be None.

        Parameters
        ~~~~~~~~~~
        comm : dict, communication object, modifiable in place
            Communication structure.

            This argument has been initialized to hold the default
            option set.

            If you wish to supply non-default options to the Arnoldi
            eigensolver, this argument may be passed as argument comm in
            a call to ``sparseig.real_symm_option``.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

    data : arbitrary, optional
        User-communication data for callback functions.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    nconv : int
        The number of converged approximations to the selected
        eigenvalues. On successful exit, this will normally be `nev`.

    w : float, ndarray, shape (`ncv`)
        The first `nconv` elements contain the converged approximations
        to the selected eigenvalues.

    v : float, ndarray, shape (`n`, `ncv`)
        Contains the eigenvectors associated with the eigenvalue
        lambda_i, for i = 1,2,...,`nconv` (stored in `w`). For
        eigenvalue, lambda_j, the corresponding eigenvector is stored in
        `v`[i-1,j-1], for i = 1,2,...,n.

    resid : float, ndarray, shape (`nev`)
        The residual ||Aw_i-lambda_i w_i||_2 for the estimates to the
        eigenpair lambda_i and w_i is returned in `resid`[i-1], for i =
        1,2,...,`nconv`.

    Other Parameters
    ----------------
    'Advisory' : int
        Default = advisory message unit number

        If the option 'List' is set then option specifications are
        listed in a 'List' file by setting the option to a file
        identification (unit) number associated with advisory messages
        (see :class:`~naginterfaces.base.utils.FileObjManager` and the
        ``FileObjManager`` method
        :meth:`~naginterfaces.base.utils.FileObjManager.unit_from_fileobj`).

    'Defaults' : valueless
        This special keyword may be used to reset all options to their
        default values.

    'Iteration Limit' : int
        Default = 300

        The limit on the number of Lanczos iterations that can be
        performed before ``sparseig.real_symm_iter`` exits.
        If not all requested eigenvalues have converged to within
        'Tolerance' and the number of Lanczos iterations has reached
        this limit then ``sparseig.real_symm_iter`` exits with an error;
        ``sparseig.real_symm_proc`` can still be called subsequently to
        return the number of converged eigenvalues, the converged
        eigenvalues and, if requested, the corresponding eigenvectors.

    'Largest Magnitude' : valueless
        Default

        The Lanczos iterative method converges on a number of
        eigenvalues with given properties.
        The default is for ``sparseig.real_symm_iter`` to compute the
        eigenvalues of largest magnitude using 'Largest Magnitude'.
        Alternatively, eigenvalues may be chosen which have 'Largest
        Algebraic' part, 'Smallest Magnitude', or 'Smallest Algebraic'
        part; or eigenvalues which are from 'Both Ends' of the algebraic
        spectrum.

    'Both Ends' : valueless
        The Lanczos iterative method converges on a number of
        eigenvalues with given properties.
        The default is for ``sparseig.real_symm_iter`` to compute the
        eigenvalues of largest magnitude using 'Largest Magnitude'.
        Alternatively, eigenvalues may be chosen which have 'Largest
        Algebraic' part, 'Smallest Magnitude', or 'Smallest Algebraic'
        part; or eigenvalues which are from 'Both Ends' of the algebraic
        spectrum.

    'Largest Algebraic' : valueless
        The Lanczos iterative method converges on a number of
        eigenvalues with given properties.
        The default is for ``sparseig.real_symm_iter`` to compute the
        eigenvalues of largest magnitude using 'Largest Magnitude'.
        Alternatively, eigenvalues may be chosen which have 'Largest
        Algebraic' part, 'Smallest Magnitude', or 'Smallest Algebraic'
        part; or eigenvalues which are from 'Both Ends' of the algebraic
        spectrum.

    'Smallest Algebraic' : valueless
        The Lanczos iterative method converges on a number of
        eigenvalues with given properties.
        The default is for ``sparseig.real_symm_iter`` to compute the
        eigenvalues of largest magnitude using 'Largest Magnitude'.
        Alternatively, eigenvalues may be chosen which have 'Largest
        Algebraic' part, 'Smallest Magnitude', or 'Smallest Algebraic'
        part; or eigenvalues which are from 'Both Ends' of the algebraic
        spectrum.

    'Smallest Magnitude' : valueless
        The Lanczos iterative method converges on a number of
        eigenvalues with given properties.
        The default is for ``sparseig.real_symm_iter`` to compute the
        eigenvalues of largest magnitude using 'Largest Magnitude'.
        Alternatively, eigenvalues may be chosen which have 'Largest
        Algebraic' part, 'Smallest Magnitude', or 'Smallest Algebraic'
        part; or eigenvalues which are from 'Both Ends' of the algebraic
        spectrum.

    'Nolist' : valueless
        Default

        Option 'List' enables printing of each option specification as
        it is supplied. 'Nolist' suppresses this printing.

    'List' : valueless
        Option 'List' enables printing of each option specification as
        it is supplied. 'Nolist' suppresses this printing.

    'Monitoring' : int
        Default = -1

        If i > 0, monitoring information is output to unit number (see
        :meth:`~naginterfaces.base.utils.FileObjManager.unit_from_fileobj`)
        i during the solution of each problem; this may be the same as
        the 'Advisory' unit number (see
        :meth:`~naginterfaces.base.utils.FileObjManager.unit_from_fileobj`).
        The type of information produced is dependent on the value of
        'Print Level', see the description of the option 'Print Level'
        for details of the information produced.
        Please see the ``FileObjManager`` method
        :meth:`~naginterfaces.base.utils.FileObjManager.unit_from_fileobj`
        to associate a file with a given unit number (see
        :meth:`~naginterfaces.base.utils.FileObjManager.unit_from_fileobj`).

    'Print Level' : int
        Default = 0

        This controls the amount of printing produced by
        ``real_symm_sparse_arnoldi`` as follows.

        [table omitted]

        Note that setting 'Print Level' >= 30 can result in very lengthy
        'Monitoring' output.

    'Regular' : valueless
        Default

        These options define the computational mode which in turn
        defines the form of operation op(x) to be performed.

        [table omitted]

    'Regular Inverse' : valueless
        These options define the computational mode which in turn
        defines the form of operation op(x) to be performed.

        [table omitted]

    'Shifted Inverse' : valueless
        These options define the computational mode which in turn
        defines the form of operation op(x) to be performed.

        [table omitted]

    'Tolerance' : float
        Default = epsilon

        An approximate eigenvalue has deemed to have converged when the
        corresponding Ritz estimate is within 'Tolerance' relative to
        the magnitude of the eigenvalue.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` > 0.

        (`errno` 2)
            On entry, nnz = *<value>*.

            Constraint: nnz > 0.

        (`errno` 2)
            On entry, nnz = *<value>* and `n` = *<value>*.

            Constraint: nnz <= `n`*(`n`+1)/2.

        (`errno` 4)
            On entry, for i = *<value>*, `irow`[i-1] = *<value>*.

            Constraint: 1 <= `irow`[i-1] <= `n`.

        (`errno` 5)
            On entry, for i = *<value>*, `icol`[i-1] = *<value>*,
            `irow`[i-1] = *<value>*.

            Constraint: 1 <= `icol`[i-1] <= `irow`[i-1].

        (`errno` 6)
            On entry, `nev` = *<value>*.

            Constraint: `nev` > 0.

        (`errno` 6)
            On entry, `nev` = *<value>* and `n` = *<value>*.

            Constraint: `nev` < (`n`-1).

        (`errno` 7)
            On entry, `ncv` = *<value>* and `nev` = *<value>*.

            Constraint: `ncv` > `nev`.

        (`errno` 7)
            On entry, `ncv` = *<value>* and `n` = *<value>*.

            Constraint: `ncv` <= `n`.

        (`errno` 8)
            On entry, the matrix (A-sigma*I) is numerically singular and
            could not be inverted. Try perturbing the value of sigma.

        (`errno` 20)
            The maximum number of iterations, through the option
            'Iteration Limit', has been set to a non-positive value.

        (`errno` 21)
            The option 'Both Ends' has been set but only 1 eigenvalue is
            requested.

        (`errno` 22)
            The maximum number of iterations has been reached.

            The maximum number of iterations = *<value>*.

            The number of converged eigenvalues = *<value>*.

            See the function document for further details.

        (`errno` 30)
            A serious error, code (*<value>*,*<value>*), has occurred in
            an internal call. Check all function calls and array sizes.
            If the call is correct then please contact NAG for
            assistance.

    Warns
    -----
    NagCallbackTerminateWarning
        (`errno` 9)
            User requested termination in `monit`, `istat` = *<value>*.

        (`errno` 10)
            User requested termination in `option`, `istat` = *<value>*.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    HSL, 2011, `A collection of Fortran codes for large-scale scientific
    computation`, http://www.hsl.rl.ac.uk/

    Lehoucq, R B, Sorensen, D C and Yang, C, 1998, `ARPACK Users' Guide:
    Solution of Large-scale Eigenvalue Problems with Implicitly
    Restarted Arnoldi Methods`, SIAM, Philadelphia
    """

    def cb_monit(ncv, niter, nconv, w, rzest, data=None):
        """Wrapping user's monit. Called as base monit."""
        if data is not None:
            monit(
                niter,
                nconv,
                w,
                rzest,
                data,
            )
        else:
            monit(
                niter,
                nconv,
                w,
                rzest,
            )

    def cb_option(comm, data=None):
        """Wrapping user's option. Called as base option."""
        if data is not None:
            option(
                comm,
                data,
            )
        else:
            option(
                comm,
            )

    fun_name = 'naginterfaces.library.eigen.real_symm_sparse_arnoldi'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nnz_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=1)[0]
    _nnz = _b_utils._EngineIntScalarType(
        _nnz_py, {'fun_name': fun_name, 'entity': '_nnz'},
    ).to_ctypes()
    _irow = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'irow'},
        irow,
        copy=False,
    )
    _icol = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'icol'},
        icol,
        copy=False,
    )
    _b_utils._check_callable(monit, {'fun_name': fun_name, 'entity': 'monit'}, allow_none=True)
    _monit = None if monit is None else cb_monit
    _b_utils._check_callable(option, {'fun_name': fun_name, 'entity': 'option'}, allow_none=True)
    _option = None if option is None else cb_option
    _ncv = _b_utils._EngineIntScalarType(
        ncv, {'fun_name': fun_name, 'entity': 'ncv'},
    ).to_ctypes()
    _ncv_py = _b_utils._EngineIntScalarType.to_py(_ncv)
    _w = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ncv_py,),
    )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _v = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py, _ncv_py),
        sorder=_b_utils._EngineData.default_sorder,
    )
    _nev = _b_utils._EngineIntScalarType(
        nev, {'fun_name': fun_name, 'entity': 'nev'},
    ).to_ctypes()
    _nev_py = _b_utils._EngineIntScalarType.to_py(_nev)
    _resid = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nev_py,),
    )
    try:
        _return_scalar = _b_eigen.real_symm_sparse_arnoldi(
            _n,
            _nnz,
            _a,
            _irow,
            _icol,
            _nev,
            _ncv,
            sigma,
            _monit,
            _option,
            _w,
            _v,
            _resid,
            data,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _nconv_py = _return_scalar
    _w_py = _w
    _v_py = _v
    _resid_py = _resid
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['nconv', 'w', 'v', 'resid'],
    )(
        _nconv_py,
        _w_py,
        _v_py,
        _resid_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def complex_gen_eigsys(crit, n, a, wl, wu, mest):
    u"""
    Selected eigenvalues and eigenvectors of complex nonsymmetric matrix
    (Black Box).

    ``complex_gen_eigsys`` computes selected eigenvalues and
    eigenvectors of a complex general matrix.

    For full information please refer to the NAG Library document for
    f02gc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f02/f02gcf.html

    Parameters
    ----------
    crit : str, length 1
        Indicates the criterion for selecting eigenvalues.

        `crit` = 'M'

            Eigenvalues are selected according to their moduli: w_l <=
            |lambda_i| <= w_u.

        `crit` = 'R'

            Eigenvalues are selected according to their real parts: w_l
            <= Re(lambda_i) <= w_u.

    n : int
        n, the order of the matrix A.

    a : complex, array-like, shape (`n`, `n`)
        The n by n general matrix A.

    wl : float
        w_l and w_u, the lower and upper bounds on the criterion for the
        selected eigenvalues (see `crit`).

    wu : float
        w_l and w_u, the lower and upper bounds on the criterion for the
        selected eigenvalues (see `crit`).

    mest : int
        `mest` must be an upper bound on m, the number of eigenvalues
        and eigenvectors selected. No eigenvectors are computed if
        `mest` < m.

    Returns
    -------
    a : complex, ndarray, shape (`n`, `n`)
        Contains the Hessenberg form of the balanced input matrix A'
        (see [Further Comments]).

    m : int
        m, the number of eigenvalues actually selected.

    w : complex, ndarray, shape (`n`)
        The first `m` elements of `w` hold the selected eigenvalues;
        elements `m`+1 to `n` contain the other eigenvalues.

    v : complex, ndarray, shape (`n`, `mest`)
        Contains the selected eigenvectors, with the ith column holding
        the eigenvector associated with the eigenvalue lambda_i (stored
        in `w`[i-1]).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `mest` = *<value>*.

            Constraint: `mest` >= 1.

        (`errno` 1)
            On entry, `wu` = *<value>* and `wl` = *<value>*.

            Constraint: `wu` > `wl`.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 1)
            On entry, `crit` = *<value>*.

            Constraint: `crit` = 'M' or 'R'.

        (`errno` 2)
            The QR algorithm failed to converge: only *<value>*
            eigenvalues have been computed; no eigenvectors have been
            computed.

        (`errno` 3)
            There are more than `mest` eigenvalues in the specified
            range. `m` (number of eigenvalues in range) = *<value>* and
            `mest` = *<value>*. No eigenvectors have been computed.
            Rerun with second dimension of `v` = `mest` >= `m`.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 4)
            Inverse iteration failed to compute all the specified
            eigenvectors. The number of eigenvectors which failed to
            converge is *<value>*. The corresponding columns of `v` are
            set to zero.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.eigen.complex_gen_eigsys'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_a, 'a'),),
    )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _w = _b_utils._EngineComplex128ArrayType.empty_ndarray(
        (_n_py,),
    )
    _mest = _b_utils._EngineIntScalarType(
        mest, {'fun_name': fun_name, 'entity': 'mest'},
    ).to_ctypes()
    _mest_py = _b_utils._EngineIntScalarType.to_py(_mest)
    _v = _b_utils._EngineComplex128ArrayType.empty_ndarray(
        (_n_py, _mest_py),
        sorder=_m_sorder,
    )
    try:
        _return_scalar = _b_eigen.complex_gen_eigsys(
            crit,
            _n,
            _a,
            wl,
            wu,
            _mest,
            _w,
            _v,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _a_py = _a
    _m_py = _return_scalar
    _w_py = _w
    _v_py = _v
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'm', 'w', 'v'],
    )(
        _a_py,
        _m_py,
        _w_py,
        _v_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def real_gen_quad(scal, jobvl, jobvr, sense, tol, a, b, c):
    u"""
    Solves the quadratic eigenvalue problem for real matrices.

    ``real_gen_quad`` solves the quadratic eigenvalue problem

        (lambda^2*A+lambda*B+C)*x = 0,

    where A, B and C are real n by n matrices.

    The function returns the 2*n eigenvalues, lambda_j, for j =
    1,2,...,2n, and can optionally return the corresponding right
    eigenvectors, x_j and/or left eigenvectors, y_j as well as estimates
    of the condition numbers of the computed eigenvalues and backward
    errors of the computed right and left eigenvectors.
    A left eigenvector satisfies the equation

        y^H*(lambda^2*A+lambda*B+C) = 0,

    where y^H is the complex conjugate transpose of y.

    lambda is represented as the pair (alpha,beta), such that lambda =
    alpha/beta.
    Note that the computation of alpha/beta may overflow and indeed beta
    may be zero.

    For full information please refer to the NAG Library document for
    f02jc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f02/f02jcf.html

    Parameters
    ----------
    scal : int
        Determines the form of scaling to be performed on A, B and C.

        `scal` = 0

            No scaling.

        `scal` = 1 (the recommended value)

            Fan, Lin and Van Dooren scaling if
            (||B||)/(sqrt(||A||*||C||)) < 10 and no scaling otherwise
            where ||Z|| is the Frobenius norm of Z.

        `scal` = 2

            Fan, Lin and Van Dooren scaling.

        `scal` = 3

            Tropical scaling with largest root.

        `scal` = 4

            Tropical scaling with smallest root.

    jobvl : str, length 1
        If `jobvl` = 'N', do not compute left eigenvectors.

        If `jobvl` = 'V', compute the left eigenvectors.

        If `sense` = 1, 2, 4, 5, 6 or 7, `jobvl` must be set to 'V'.

    jobvr : str, length 1
        If `jobvr` = 'N', do not compute right eigenvectors.

        If `jobvr` = 'V', compute the right eigenvectors.

        If `sense` = 1, 3, 4, 5, 6 or 7, `jobvr` must be set to 'V'.

    sense : int
        Determines whether, or not, condition numbers and backward
        errors are computed.

        `sense` = 0

            Do not compute condition numbers, or backward errors.

        `sense` = 1

            Just compute condition numbers for the eigenvalues.

        `sense` = 2

            Just compute backward errors for the left eigenpairs.

        `sense` = 3

            Just compute backward errors for the right eigenpairs.

        `sense` = 4

            Compute backward errors for the left and right eigenpairs.

        `sense` = 5

            Compute condition numbers for the eigenvalues and backward
            errors for the left eigenpairs.

        `sense` = 6

            Compute condition numbers for the eigenvalues and backward
            errors for the right eigenpairs.

        `sense` = 7

            Compute condition numbers for the eigenvalues and backward
            errors for the left and right eigenpairs.

    tol : float
        `tol` is used as the tolerance for making decisions on rank in
        the deflation procedure. If `tol` is zero on entry then
        n*machine precision is used in place of `tol`, where machine
        precision is as returned by function ``machine.precision``. A
        diagonal element of a triangular matrix, R, is regarded as zero
        if |r_{jj}| <= `tol`*size(X), or n*machine precision*size(X)
        when `tol` is zero, where size(X) is based on the size of the
        absolute values of the elements of the matrix X containing the
        matrix R. See Hammarling `et al.` (2013) for the motivation. If
        `tol` is -1.0 on entry then no deflation is attempted. The
        recommended value for `tol` is zero.

    a : float, array-like, shape (n, n)
        The n by n matrix A.

    b : float, array-like, shape (n, n)
        The n by n matrix B.

    c : float, array-like, shape (n, n)
        The n by n matrix C.

    Returns
    -------
    a : float, ndarray, shape (n, n)
        `a` is used as internal workspace, but if `jobvl` = 'V' or
        `jobvr` = 'V', `a` is restored on exit.

    b : float, ndarray, shape (n, n)
        `b` is used as internal workspace, but is restored on exit.

    c : float, ndarray, shape (n, n)
        `c` is used as internal workspace, but if `jobvl` = 'V' or
        `jobvr` = 'V', `c` is restored on exit.

    alphar : float, ndarray, shape (2*n)
        `alphar`[j-1], for j = 1,2,...,2n, contains the real part of
        alpha_j for the jth eigenvalue pair (alpha_j,beta_j) of the
        quadratic eigenvalue problem.

    alphai : float, ndarray, shape (2*n)
        `alphai`[j-1], for j = 1,2,...,2n, contains the imaginary part
        of alpha_j for the jth eigenvalue pair (alpha_j,beta_j) of the
        quadratic eigenvalue problem. If `alphai`[j-1] is zero then the
        jth eigenvalue is real; if `alphai`[j-1] is positive then the
        jth and (j+1)th eigenvalues are a complex conjugate pair, with
        `alphai`[j] negative.

    beta : float, ndarray, shape (2*n)
        `beta`[j-1], for j = 1,2,...,2*n, contains the second part of
        the jth eigenvalue pair (alpha_j,beta_j) of the quadratic
        eigenvalue problem, with beta_j >= 0. Infinite eigenvalues have
        beta_j set to zero.

    vl : None or float, ndarray, shape (n, :)
        If `jobvl` = 'V', the left eigenvectors y_j are stored one after
        another in the columns of `vl`, in the same order as the
        corresponding eigenvalues. If the jth eigenvalue is real, then
        y_j = `vl`[:-1,j-1], the jth column of `vl`. If the jth and
        (j+1)th eigenvalues form a complex conjugate pair, then y_j =
        `vl`[:-1,j-1]+i*`vl`[:-1,j] and y_{j+1} =
        `vl`[:-1,j-1]-i*`vl`[:-1,j]. Each eigenvector will be normalized
        with length unity and with the element of largest modulus real
        and positive.

        If `jobvl` = 'N', `vl` is not referenced.

    vr : None or float, ndarray, shape (n, :)
        If `jobvr` = 'V', the right eigenvectors x_j are stored one
        after another in the columns of `vr`, in the same order as the
        corresponding eigenvalues. If the jth eigenvalue is real, then
        x_j = `vr`[:-1,j-1], the jth column of `vr`. If the jth and
        (j+1)th eigenvalues form a complex conjugate pair, then x_j =
        `vr`[:-1,j-1]+i*`vr`[:-1,j] and x_{j+1} =
        `vr`[:-1,j-1]-i*`vr`[:-1,j]. Each eigenvector will be normalized
        with length unity and with the element of largest modulus real
        and positive.

        If `jobvr` = 'N', `vr` is not referenced.

    s : None or float, ndarray, shape (:)
        If `sense` = 1, 5, 6 or 7, `s`[j-1] contains the condition
        number estimate for the jth eigenvalue (large condition numbers
        imply that the problem is near one with multiple eigenvalues).
        Infinite condition numbers are returned as the largest model
        `float` number (``machine.real_largest``).

        If `sense` = 0, 2, 3 or 4, `s` is not referenced.

    bevl : None or float, ndarray, shape (:)
        If `sense` = 2, 4, 5 or 7, `bevl`[j-1] contains the backward
        error estimate for the computed left eigenpair (lambda_j,y_j).

        If `sense` = 0, 1, 3 or 6, `bevl` is not referenced.

    bevr : None or float, ndarray, shape (:)
        If `sense` = 3, 4, 6 or 7, `bevr`[j-1] contains the backward
        error estimate for the computed right eigenpair (lambda_j,x_j).

        If `sense` = 0, 1, 2 or 5, `bevr` is not referenced.

    iwarn : int
        `iwarn` will be positive if there are warnings, otherwise
        `iwarn` will be 0.

        If no exception or warning is raised then:

            if `iwarn` = 1 then one, or both, of the matrices A and C is
            zero. In this case no scaling is performed, even if `scal` >
            0;

            if `iwarn` = 2 then the matrices A and C are singular, or
            nearly singular, so the problem is potentially ill-posed;

            if `iwarn` = 3 then both the conditions for `iwarn` = 1 and
            `iwarn` = 2 above, apply. If `iwarn` = 4, ||`b`|| >=
            10sqrt(||A||·||C||) and backward stability cannot be
            guaranteed.

        If `errno` = 2, `iwarn` returns the value of errno from
        ``lapackeig.dgges``.

        If `errno` = 3, `iwarn` returns the value of errno from
        ``lapackeig.dggev``.

    Raises
    ------
    NagValueError
        (`errno` -19)
            On entry, ldvr = *<value>*, n = *<value>* and `jobvr` =
            *<value>*.

            Constraint: when `jobvr` = 'V', ldvr >= n.

        (`errno` -17)
            On entry, ldvl = *<value>*, n = *<value>* and `jobvl` =
            *<value>*.

            Constraint: when `jobvl` = 'V', ldvl >= n.

        (`errno` -6)
            On entry, n = *<value>*.

            Constraint: n >= 0.

        (`errno` -4)
            On entry, `sense` = *<value>*.

            Constraint: `sense` = 0, 1, 2, 3, 4, 5, 6 or 7.

        (`errno` -3)
            On entry, `sense` = *<value>* and `jobvr` = *<value>*.

            Constraint: when `jobvr` = 'N', `sense` = 0 or 2,

            when `jobvr` = 'V', `sense` = 1, 3, 4, 5, 6 or 7.

        (`errno` -3)
            On entry, `jobvr` = *<value>*.

            Constraint: `jobvr` = 'N' or 'V'.

        (`errno` -2)
            On entry, `sense` = *<value>* and `jobvl` = *<value>*.

            Constraint: when `jobvl` = 'N', `sense` = 0 or 3,

            when `jobvl` = 'V', `sense` = 1, 2, 4, 5, 6 or 7.

        (`errno` -2)
            On entry, `jobvl` = *<value>*.

            Constraint: `jobvl` = 'N' or 'V'.

        (`errno` -1)
            On entry, `scal` = *<value>*.

            Constraint: `scal` = 0, 1, 2, 3 or 4.

        (`errno` 1)
            The quadratic matrix polynomial is nonregular (singular).

        (`errno` 2)
            The QZ iteration failed in ``lapackeig.dgges``.

            `iwarn` returns the value of errno returned by
            ``lapackeig.dgges``. This failure is unlikely to happen, but
            if it does, please contact NAG.

        (`errno` 3)
            The QZ iteration failed in ``lapackeig.dggev``.

            `iwarn` returns the value of errno returned by
            ``lapackeig.dggev``. This failure is unlikely to happen, but
            if it does, please contact NAG.

    References
    ----------
    Fan, H -Y, Lin, W.-W and Van Dooren, P., 2004, `Normwise scaling of
    second order polynomial matrices`, SIAM J. Matrix Anal. Appl. (26,
    1), 252--256

    Gaubert, S and Sharify, M, 2009, `Tropical scaling of polynomial
    matrices`, Lecture Notes in Control and Information Sciences Series
    (389), 291--303, Springer--Verlag

    Hammarling, S, Munro, C J and Tisseur, F, 2013, `An algorithm for
    the complete solution of quadratic eigenvalue problems`, ACM Trans.
    Math. Software. (39(3):18:1--18:119),
    http://eprints.maths.manchester.ac.uk/2061/
    """

    fun_name = 'naginterfaces.library.eigen.real_gen_quad'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _c = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_a, 'a'), (_b, 'b'), (_c, 'c')),
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _alphar = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (2*_n_py,),
    )
    _alphai = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (2*_n_py,),
    )
    _beta = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (2*_n_py,),
    )
    _jobvl = _b_utils._EngineCharScalarType(
        jobvl, {'fun_name': fun_name, 'entity': 'jobvl'}, exp_chrlen=1,
    ).to_ctypes()
    _jobvl_py = _b_utils._EngineCharScalarType.to_py(_jobvl)
    if _jobvl_py not in ('N', 'n'):
        if _jobvl_py in ('V', 'v'):
            _vl_dim2 = 2*_n_py
        else:
            _vl_dim2 = 0
        exp_shape_vl = (_n_py, _vl_dim2)
        _vl = _b_utils._EngineFloat64ArrayType.empty_ndarray(
            exp_shape_vl,
            sorder=_m_sorder,
        )
    else:
        _vl = None
    _jobvr = _b_utils._EngineCharScalarType(
        jobvr, {'fun_name': fun_name, 'entity': 'jobvr'}, exp_chrlen=1,
    ).to_ctypes()
    _jobvr_py = _b_utils._EngineCharScalarType.to_py(_jobvr)
    if _jobvr_py not in ('N', 'n'):
        if _jobvr_py in ('V', 'v'):
            _vr_dim2 = 2*_n_py
        else:
            _vr_dim2 = 0
        exp_shape_vr = (_n_py, _vr_dim2)
        _vr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
            exp_shape_vr,
            sorder=_m_sorder,
        )
    else:
        _vr = None
    _sense = _b_utils._EngineIntScalarType(
        sense, {'fun_name': fun_name, 'entity': 'sense'},
    ).to_ctypes()
    _sense_py = _b_utils._EngineIntScalarType.to_py(_sense)
    if _sense_py not in (0, 2, 3, 4):
        if _sense_py in (1, 5, 6, 7):
            _s_dim1 = 2*_n_py
        else:
            _s_dim1 = 0
        exp_shape_s = (_s_dim1,)
        _s = _b_utils._EngineFloat64ArrayType.empty_ndarray(
            exp_shape_s,
        )
    else:
        _s = None
    if _sense_py not in (0, 1, 3, 6):
        if _sense_py in (2, 4, 5, 7):
            _bevl_dim1 = 2*_n_py
        else:
            _bevl_dim1 = 0
        exp_shape_bevl = (_bevl_dim1,)
        _bevl = _b_utils._EngineFloat64ArrayType.empty_ndarray(
            exp_shape_bevl,
        )
    else:
        _bevl = None
    if _sense_py not in (0, 1, 2, 5):
        if _sense_py in (3, 4, 6, 7):
            _bevr_dim1 = 2*_n_py
        else:
            _bevr_dim1 = 0
        exp_shape_bevr = (_bevr_dim1,)
        _bevr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
            exp_shape_bevr,
        )
    else:
        _bevr = None
    _return_scalar = _b_eigen.real_gen_quad(
        scal,
        jobvl,
        jobvr,
        _sense,
        tol,
        _n,
        _a,
        _b,
        _c,
        _alphar,
        _alphai,
        _beta,
        _vl,
        _vr,
        _s,
        _bevl,
        _bevr,
    )
    _a_py = _a
    _b_py = _b
    _c_py = _c
    _alphar_py = _alphar
    _alphai_py = _alphai
    _beta_py = _beta
    if _jobvl_py not in ('N', 'n'):
        _vl_py = _vl
    else:
        _vl_py = None
    if _jobvr_py not in ('N', 'n'):
        _vr_py = _vr
    else:
        _vr_py = None
    if _sense_py not in (0, 2, 3, 4):
        _s_py = _s
    else:
        _s_py = None
    if _sense_py not in (0, 1, 3, 6):
        _bevl_py = _bevl
    else:
        _bevl_py = None
    if _sense_py not in (0, 1, 2, 5):
        _bevr_py = _bevr
    else:
        _bevr_py = None
    _iwarn_py = _return_scalar
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'b', 'c', 'alphar', 'alphai', 'beta', 'vl', 'vr', 's', 'bevl', 'bevr', 'iwarn'],
    )(
        _a_py,
        _b_py,
        _c_py,
        _alphar_py,
        _alphai_py,
        _beta_py,
        _vl_py,
        _vr_py,
        _s_py,
        _bevl_py,
        _bevr_py,
        _iwarn_py,
    )
    return _return_data

def complex_gen_quad(scal, jobvl, jobvr, sense, tol, a, b, c):
    u"""
    Solves the quadratic eigenvalue problem for complex matrices.

    ``complex_gen_quad`` solves the quadratic eigenvalue problem

        (lambda^2*A+lambda*B+C)*x = 0,

    where A, B and C are complex n by n matrices.

    The function returns the 2*n eigenvalues, lambda_j, for j =
    1,2,...,2n, and can optionally return the corresponding right
    eigenvectors, x_j and/or left eigenvectors, y_j as well as estimates
    of the condition numbers of the computed eigenvalues and backward
    errors of the computed right and left eigenvectors.
    A left eigenvector satisfies the equation

        y^H*(lambda^2*A+lambda*B+C) = 0,

    where y^H is the complex conjugate transpose of y.

    lambda is represented as the pair (alpha,beta), such that lambda =
    alpha/beta.
    Note that the computation of alpha/beta may overflow and indeed beta
    may be zero.

    For full information please refer to the NAG Library document for
    f02jq

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f02/f02jqf.html

    Parameters
    ----------
    scal : int
        Determines the form of scaling to be performed on A, B and C.

        `scal` = 0

            No scaling.

        `scal` = 1 (the recommended value)

            Fan, Lin and Van Dooren scaling if
            (||B||)/(sqrt(||A||*||C||)) < 10 and no scaling otherwise
            where ||Z|| is the Frobenius norm of Z.

        `scal` = 2

            Fan, Lin and Van Dooren scaling.

        `scal` = 3

            Tropical scaling with largest root.

        `scal` = 4

            Tropical scaling with smallest root.

    jobvl : str, length 1
        If `jobvl` = 'N', do not compute left eigenvectors.

        If `jobvl` = 'V', compute the left eigenvectors.

        If `sense` = 1, 2, 4, 5, 6 or 7, `jobvl` must be set to 'V'.

    jobvr : str, length 1
        If `jobvr` = 'N', do not compute right eigenvectors.

        If `jobvr` = 'V', compute the right eigenvectors.

        If `sense` = 1, 3, 4, 5, 6 or 7, `jobvr` must be set to 'V'.

    sense : int
        Determines whether, or not, condition numbers and backward
        errors are computed.

        `sense` = 0

            Do not compute condition numbers, or backward errors.

        `sense` = 1

            Just compute condition numbers for the eigenvalues.

        `sense` = 2

            Just compute backward errors for the left eigenpairs.

        `sense` = 3

            Just compute backward errors for the right eigenpairs.

        `sense` = 4

            Compute backward errors for the left and right eigenpairs.

        `sense` = 5

            Compute condition numbers for the eigenvalues and backward
            errors for the left eigenpairs.

        `sense` = 6

            Compute condition numbers for the eigenvalues and backward
            errors for the right eigenpairs.

        `sense` = 7

            Compute condition numbers for the eigenvalues and backward
            errors for the left and right eigenpairs.

    tol : float
        `tol` is used as the tolerance for making decisions on rank in
        the deflation procedure. If `tol` is zero on entry then
        n*machine precision is used in place of `tol`, where machine
        precision is as returned by function ``machine.precision``. A
        diagonal element of a triangular matrix, R, is regarded as zero
        if |r_{jj}| <= `tol`*size(X), or n*machine precision*size(X)
        when `tol` is zero, where size(X) is based on the size of the
        absolute values of the elements of the matrix X containing the
        matrix R. See Hammarling `et al.` (2013) for the motivation. If
        `tol` is -1.0 on entry then no deflation is attempted. The
        recommended value for `tol` is zero.

    a : complex, array-like, shape (n, n)
        The n by n matrix A.

    b : complex, array-like, shape (n, n)
        The n by n matrix B.

    c : complex, array-like, shape (n, n)
        The n by n matrix C.

    Returns
    -------
    a : complex, ndarray, shape (n, n)
        `a` is used as internal workspace, but if `jobvl` = 'V' or
        `jobvr` = 'V', `a` is restored on exit.

    b : complex, ndarray, shape (n, n)
        `b` is used as internal workspace, but is restored on exit.

    c : complex, ndarray, shape (n, n)
        `c` is used as internal workspace, but if `jobvl` = 'V' or
        `jobvr` = 'V', `c` is restored on exit.

    alpha : complex, ndarray, shape (2*n)
        `alpha`[j-1], for j = 1,2,...,2n, contains the first part of the
        the jth eigenvalue pair (alpha_j,beta_j) of the quadratic
        eigenvalue problem.

    beta : complex, ndarray, shape (2*n)
        `beta`[j-1], for j = 1,2,...,2*n, contains the second part of
        the jth eigenvalue pair (alpha_j,beta_j) of the quadratic
        eigenvalue problem. Although `beta` is declared complex, it is
        actually real and non-negative. Infinite eigenvalues have beta_j
        set to zero.

    vl : None or complex, ndarray, shape (n, :)
        If `jobvl` = 'V', the left eigenvectors y_j are stored one after
        another in the columns of `vl`, in the same order as the
        corresponding eigenvalues. Each eigenvector will be normalized
        with length unity and with the element of largest modulus real
        and positive.

        If `jobvl` = 'N', `vl` is not referenced.

    vr : None or complex, ndarray, shape (n, :)
        If `jobvr` = 'V', the right eigenvectors x_j are stored one
        after another in the columns of `vr`, in the same order as the
        corresponding eigenvalues. Each eigenvector will be normalized
        with length unity and with the element of largest modulus real
        and positive.

        If `jobvr` = 'N', `vr` is not referenced.

    s : None or float, ndarray, shape (:)
        If `sense` = 1, 5, 6 or 7, `s`[j-1] contains the condition
        number estimate for the jth eigenvalue (large condition numbers
        imply that the problem is near one with multiple eigenvalues).
        Infinite condition numbers are returned as the largest model
        real number (``machine.real_largest``).

        If `sense` = 0, 2, 3 or 4, `s` is not referenced.

    bevl : None or float, ndarray, shape (:)
        If `sense` = 2, 4, 5 or 7, `bevl`[j-1] contains the backward
        error estimate for the computed left eigenpair (lambda_j,y_j).

        If `sense` = 0, 1, 3 or 6, `bevl` is not referenced.

    bevr : None or float, ndarray, shape (:)
        If `sense` = 3, 4, 6 or 7, `bevr`[j-1] contains the backward
        error estimate for the computed right eigenpair (lambda_j,x_j).

        If `sense` = 0, 1, 2 or 5, `bevr` is not referenced.

    iwarn : int
        `iwarn` will be positive if there are warnings, otherwise
        `iwarn` will be 0.

        If no exception or warning is raised then:

            if `iwarn` = 1 then one, or both, of the matrices A and C is
            zero. In this case no scaling is performed, even if `scal` >
            0;

            if `iwarn` = 2 then the matrices A and C are singular, or
            nearly singular, so the problem is potentially ill-posed;

            if `iwarn` = 3 then both the conditions for `iwarn` = 1 and
            `iwarn` = 2 above, apply. If `iwarn` = 4, ||`b`|| >=
            10sqrt(||A||·||C||) and backward stability cannot be
            guaranteed.

        If `errno` = 2, `iwarn` returns the value of errno from
        ``lapackeig.zgges``.

        If `errno` = 3, `iwarn` returns the value of errno from
        ``lapackeig.zggev``.

    Raises
    ------
    NagValueError
        (`errno` -18)
            On entry, ldvr = *<value>*, n = *<value>* and `jobvr` =
            *<value>*.

            Constraint: when `jobvr` = 'V', ldvr >= n.

        (`errno` -16)
            On entry, ldvl = *<value>*, n = *<value>* and `jobvl` =
            *<value>*.

            Constraint: when `jobvl` = 'V', ldvl >= n.

        (`errno` -6)
            On entry, n = *<value>*.

            Constraint: n >= 0.

        (`errno` -4)
            On entry, `sense` = *<value>*.

            Constraint: `sense` = 0, 1, 2, 3, 4, 5, 6 or 7.

        (`errno` -3)
            On entry, `sense` = *<value>* and `jobvr` = *<value>*.

            Constraint: when `jobvr` = 'N', `sense` = 0 or 2,

            when `jobvr` = 'V', `sense` = 1, 3, 4, 5, 6 or 7.

        (`errno` -3)
            On entry, `jobvr` = *<value>*.

            Constraint: `jobvr` = 'N' or 'V'.

        (`errno` -2)
            On entry, `sense` = *<value>* and `jobvl` = *<value>*.

            Constraint: when `jobvl` = 'N', `sense` = 0 or 3,

            when `jobvl` = 'V', `sense` = 1, 2, 4, 5, 6 or 7.

        (`errno` -2)
            On entry, `jobvl` = *<value>*.

            Constraint: `jobvl` = 'N' or 'V'.

        (`errno` -1)
            On entry, `scal` = *<value>*.

            Constraint: `scal` = 0, 1, 2, 3 or 4.

        (`errno` 1)
            The quadratic matrix polynomial is nonregular (singular).

        (`errno` 2)
            The QZ iteration failed in ``lapackeig.zgges``.

            `iwarn` returns the value of errno returned by
            ``lapackeig.zgges``. This failure is unlikely to happen, but
            if it does, please contact NAG.

        (`errno` 3)
            The QZ iteration failed in ``lapackeig.zggev``.

            `iwarn` returns the value of errno returned by
            ``lapackeig.zggev``. This failure is unlikely to happen, but
            if it does, please contact NAG.

    References
    ----------
    Fan, H -Y, Lin, W.-W and Van Dooren, P., 2004, `Normwise scaling of
    second order polynomial matrices`, SIAM J. Matrix Anal. Appl. (26,
    1), 252--256

    Gaubert, S and Sharify, M, 2009, `Tropical scaling of polynomial
    matrices`, Lecture Notes in Control and Information Sciences Series
    (389), 291--303, Springer--Verlag

    Hammarling, S, Munro, C J and Tisseur, F, 2013, `An algorithm for
    the complete solution of quadratic eigenvalue problems`, ACM Trans.
    Math. Software. (39(3):18:1--18:119),
    http://eprints.maths.manchester.ac.uk/2061/
    """

    fun_name = 'naginterfaces.library.eigen.complex_gen_quad'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
    )
    _c = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
        copy=True,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_a, 'a'), (_b, 'b'), (_c, 'c')),
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _alpha = _b_utils._EngineComplex128ArrayType.empty_ndarray(
        (2*_n_py,),
    )
    _beta = _b_utils._EngineComplex128ArrayType.empty_ndarray(
        (2*_n_py,),
    )
    _jobvl = _b_utils._EngineCharScalarType(
        jobvl, {'fun_name': fun_name, 'entity': 'jobvl'}, exp_chrlen=1,
    ).to_ctypes()
    _jobvl_py = _b_utils._EngineCharScalarType.to_py(_jobvl)
    if _jobvl_py not in ('N', 'n'):
        if _jobvl_py in ('V', 'v'):
            _vl_dim2 = 2*_n_py
        else:
            _vl_dim2 = 0
        exp_shape_vl = (_n_py, _vl_dim2)
        _vl = _b_utils._EngineComplex128ArrayType.empty_ndarray(
            exp_shape_vl,
            sorder=_m_sorder,
        )
    else:
        _vl = None
    _jobvr = _b_utils._EngineCharScalarType(
        jobvr, {'fun_name': fun_name, 'entity': 'jobvr'}, exp_chrlen=1,
    ).to_ctypes()
    _jobvr_py = _b_utils._EngineCharScalarType.to_py(_jobvr)
    if _jobvr_py not in ('N', 'n'):
        if _jobvr_py in ('V', 'v'):
            _vr_dim2 = 2*_n_py
        else:
            _vr_dim2 = 0
        exp_shape_vr = (_n_py, _vr_dim2)
        _vr = _b_utils._EngineComplex128ArrayType.empty_ndarray(
            exp_shape_vr,
            sorder=_m_sorder,
        )
    else:
        _vr = None
    _sense = _b_utils._EngineIntScalarType(
        sense, {'fun_name': fun_name, 'entity': 'sense'},
    ).to_ctypes()
    _sense_py = _b_utils._EngineIntScalarType.to_py(_sense)
    if _sense_py not in (0, 2, 3, 4):
        if _sense_py in (1, 5, 6, 7):
            _s_dim1 = 2*_n_py
        else:
            _s_dim1 = 0
        exp_shape_s = (_s_dim1,)
        _s = _b_utils._EngineFloat64ArrayType.empty_ndarray(
            exp_shape_s,
        )
    else:
        _s = None
    if _sense_py not in (0, 1, 3, 6):
        if _sense_py in (2, 4, 5, 7):
            _bevl_dim1 = 2*_n_py
        else:
            _bevl_dim1 = 0
        exp_shape_bevl = (_bevl_dim1,)
        _bevl = _b_utils._EngineFloat64ArrayType.empty_ndarray(
            exp_shape_bevl,
        )
    else:
        _bevl = None
    if _sense_py not in (0, 1, 2, 5):
        if _sense_py in (3, 4, 6, 7):
            _bevr_dim1 = 2*_n_py
        else:
            _bevr_dim1 = 0
        exp_shape_bevr = (_bevr_dim1,)
        _bevr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
            exp_shape_bevr,
        )
    else:
        _bevr = None
    _return_scalar = _b_eigen.complex_gen_quad(
        scal,
        jobvl,
        jobvr,
        _sense,
        tol,
        _n,
        _a,
        _b,
        _c,
        _alpha,
        _beta,
        _vl,
        _vr,
        _s,
        _bevl,
        _bevr,
    )
    _a_py = _a
    _b_py = _b
    _c_py = _c
    _alpha_py = _alpha
    _beta_py = _beta
    if _jobvl_py not in ('N', 'n'):
        _vl_py = _vl
    else:
        _vl_py = None
    if _jobvr_py not in ('N', 'n'):
        _vr_py = _vr
    else:
        _vr_py = None
    if _sense_py not in (0, 2, 3, 4):
        _s_py = _s
    else:
        _s_py = None
    if _sense_py not in (0, 1, 3, 6):
        _bevl_py = _bevl
    else:
        _bevl_py = None
    if _sense_py not in (0, 1, 2, 5):
        _bevr_py = _bevr
    else:
        _bevr_py = None
    _iwarn_py = _return_scalar
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'b', 'c', 'alpha', 'beta', 'vl', 'vr', 's', 'bevl', 'bevr', 'iwarn'],
    )(
        _a_py,
        _b_py,
        _c_py,
        _alpha_py,
        _beta_py,
        _vl_py,
        _vr_py,
        _s_py,
        _bevl_py,
        _bevr_py,
        _iwarn_py,
    )
    return _return_data

def real_gen_partialsvd(m, n, k, ncv, av, data=None, io_manager=None):
    u"""
    Computes leading terms in the singular value decomposition of a real
    general matrix; also computes corresponding left and right singular
    vectors.

    ``real_gen_partialsvd`` returns leading terms in the singular value
    decomposition (SVD) of a real general matrix and computes the
    corresponding left and right singular vectors.

    For full information please refer to the NAG Library document for
    f02wg

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f02/f02wgf.html

    Parameters
    ----------
    m : int
        m, the number of rows of the matrix A.

    n : int
        n, the number of columns of the matrix A.

    k : int
        k, the number of singular values to be computed.

    ncv : int
        This is the number of Lanczos basis vectors to use during the
        computation of the largest eigenvalues of A^TA (m >= n) or AA^T
        (m < n).

        At present there is no `a priori` analysis to guide the
        selection of `ncv` relative to `k`.

        However, it is recommended that `ncv` >= 2*`k`+1.

        If many problems of the same type are to be solved, you should
        experiment with varying `ncv` while keeping `k` fixed for a
        given test problem.

        This will usually decrease the required number of matrix-vector
        operations but it also increases the internal storage required
        to maintain the orthogonal basis vectors.

        The optimal 'cross-over' with respect to CPU time is problem
        dependent and must be determined empirically.

    av : callable (iflag, ax) = av(iflag, m, n, x, data=None)
        `av` must return the vector result of the matrix-vector product
        Ax or A^Tx, as indicated by the input value of `iflag`, for the
        given vector x.

        Parameters
        ~~~~~~~~~~
        iflag : int
            If `iflag` = 1, `ax` must return the m-vector result of the
            matrix-vector product Ax.

            If `iflag` = 2, `ax` must return the n-vector result of the
            matrix-vector product A^Tx.

        m : int
            The number of rows of the matrix A.

        n : int
            The number of columns of the matrix A.

        x : float, ndarray, shape (:)
            The vector to be pre-multiplied by the matrix A or A^T.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        iflag : int
            May be used as a flag to indicate a failure in the
            computation of Ax or A^Tx. If `iflag` is negative on exit
            from `av`, ``real_gen_partialsvd`` will exit immediately
            with errno set to `iflag`.

        ax : float, array-like, shape (:)
            If `iflag` = 1, contains the m-vector result of the
            matrix-vector product Ax.

            If `iflag` = 2, contains the n-vector result of the
            matrix-vector product A^Tx.

    data : arbitrary, optional
        User-communication data for callback functions.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    nconv : int
        The number of converged singular values found.

    sigma : float, ndarray, shape (`ncv`)
        The `nconv` converged singular values are stored in the first
        `nconv` elements of `sigma`.

    u : float, ndarray, shape (`m`, `ncv`)
        The left singular vectors corresponding to the singular values
        stored in `sigma`.

        The ith element of the jth left singular vector u_j is stored in
        `u`[i-1,j-1], for j = 1,2,...,`nconv`, for i = 1,2,...,m.

    v : float, ndarray, shape (`n`, `ncv`)
        The right singular vectors corresponding to the singular values
        stored in `sigma`.

        The ith element of the jth right singular vector v_j is stored
        in `v`[i-1,j-1], for j = 1,2,...,`nconv`, for i = 1,2,...,n.

    resid : float, ndarray, shape (`ncv`)
        The residual ||Av_j-sigma_j u_j||, for m >= n, or
        ||A^Tu_j-sigma_j v_j||, for m < n, for each of the converged
        singular values sigma_j and corresponding left and right
        singular vectors u_j and v_j.

    Raises
    ------
    NagValueError
        (`errno` i < 0)
            On output from user-defined function `av`, `iflag` was set
            to a negative value, `iflag` = *<value>*.

        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 0.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 3)
            On entry, `k` = *<value>*.

            Constraint: `k` > 0.

        (`errno` 4)
            On entry, `k` = *<value>*, `ncv` = *<value>*, `m` =
            *<value>* and `n` = *<value>*.

            Constraint: `k` < `ncv` <= min(`m`, `n`).

        (`errno` 8)
            The maximum number of iterations has been reached. The
            maximum number of iterations = *<value>*. The number of
            converged eigenvalues = *<value>*.

        (`errno` 9)
            No shifts could be applied during a cycle of the implicitly
            restarted Lanczos iteration.

        (`errno` 10)
            Could not build a full Lanczos factorization.

        (`errno` 11)
            The number of eigenvalues found to sufficient accuracy is
            zero.

        (`errno` 20)
            An error occurred during an internal call. Consider
            increasing the size of `ncv` relative to `k`.

    References
    ----------
    Wilkinson, J H, 1978, `Singular Value Decomposition -- Basic
    Aspects`, Numerical Software -- Needs and Availability, (ed D A H
    Jacobs), Academic Press
    """

    fun_name = 'naginterfaces.library.eigen.real_gen_partialsvd'

    cb_av_name = fun_name + ':av'

    def cb_av(iflag, m, n, x, ax, data=None):
        """Wrapping user's av. Called as base av."""
        if data is not None:
            _iflag_av, _ax_av = av(
                iflag,
                m,
                n,
                x,
                data,
            )
        else:
            _iflag_av, _ax_av = av(
                iflag,
                m,
                n,
                x,
            )
        _iflag_av_base = _iflag_av
        _iflag_av_base = _b_utils._EngineIntScalarType(
            _iflag_av, {'fun_name': cb_av_name, 'returned': 'iflag'},
        ).to_ctypes()
        _iflag_av_base = _b_utils._EngineIntScalarType.to_py(_iflag_av_base)
        if _iflag_av_base == 1:
            _ax_av_dim1 = m
        elif _iflag_av_base == 2:
            _ax_av_dim1 = n
        else:
            _ax_av_dim1 = 0
        _exp_shape_ax_av = (_ax_av_dim1,)
        ax[:] = _utils._to_ndarray(
            {'fun_name': cb_av_name, 'returned': 'ax'},
            _ax_av,
            copy=False,
            exp_shape=_exp_shape_ax_av,
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        return _iflag_av_base

    _b_utils._check_callable(av, {'fun_name': fun_name, 'entity': 'av'})
    _av = cb_av
    _ncv = _b_utils._EngineIntScalarType(
        ncv, {'fun_name': fun_name, 'entity': 'ncv'},
    ).to_ctypes()
    _ncv_py = _b_utils._EngineIntScalarType.to_py(_ncv)
    _sigma = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ncv_py,),
    )
    _m = _b_utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = _b_utils._EngineIntScalarType.to_py(_m)
    _u = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_m_py, _ncv_py),
        sorder=_b_utils._EngineData.default_sorder,
    )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _v = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py, _ncv_py),
        sorder=_b_utils._EngineData.default_sorder,
    )
    _resid = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ncv_py,),
    )
    _return_scalar = _b_eigen.real_gen_partialsvd(
        _m,
        _n,
        k,
        _ncv,
        _av,
        _sigma,
        _u,
        _v,
        _resid,
        data,
        io_manager,
    )
    _nconv_py = _return_scalar
    _sigma_py = _sigma
    _u_py = _u
    _v_py = _v
    _resid_py = _resid
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['nconv', 'sigma', 'u', 'v', 'resid'],
    )(
        _nconv_py,
        _sigma_py,
        _u_py,
        _v_py,
        _resid_py,
    )
    return _return_data

def real_triang_svd(n, a, b, wantq, wantp):
    u"""
    SVD of real upper triangular matrix (Black Box).

    ``real_triang_svd`` returns all, or part, of the singular value
    decomposition of a real upper triangular matrix.

    For full information please refer to the NAG Library document for
    f02wu

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f02/f02wuf.html

    Parameters
    ----------
    n : int
        n, the order of the matrix R.

        If `n` = 0, an immediate return is effected.

    a : float, array-like, shape (`n`, `n`)
        The leading n by n upper triangular part of the array `a` must
        contain the upper triangular matrix R.

    b : float, array-like, shape (:, ncolb)
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if ncolb>0: `n`; otherwise: 1.

        With ncolb > 0, the leading n by ncolb part of the array `b`
        must contain the matrix to be transformed.

    wantq : bool
        Must be True if the matrix Q is required.

        If `wantq` = False, the array `q` is not referenced.

    wantp : bool
        Must be True if the matrix P^T is required, in which case P^T is
        overwritten on the array `a`, otherwise `wantp` must be False.

    Returns
    -------
    a : float, ndarray, shape (`n`, `n`)
        If `wantp` = True, the n by n part of `a` will contain the n by
        n orthogonal matrix P^T, otherwise the n by n upper triangular
        part of `a` is used as internal workspace, but the strictly
        lower triangular part of `a` is not referenced.

    b : float, ndarray, shape (:, ncolb)
        The leading n by ncolb part of the array `b` is overwritten by
        the matrix Q^TB.

    q : float, ndarray, shape (:, :)
        With `wantq` = True, the leading n by n part of the array `q`
        will contain the orthogonal matrix Q. Otherwise the array `q` is
        not referenced.

    sv : float, ndarray, shape (`n`)
            If no exception or warning is raised the array `sv` will
            contain the diagonal elements of the matrix.

            If `errno` = 1 the array `sv` will contain the diagonal
            elements of the bidiagonal matrix E in the factorization R =
            QEP^T; the superdiagonal elements of E will be contained in
            the first `n`-1 elements of `work`.

    work : float, ndarray, shape (`n`)
        `work`[0:`n`-1] contains the super-diagonal elements of the
        bidiagonal matrix E computed during the bidiagonalization stage;
        `work`[`n`-1] contains the total number of iterations taken by
        the QR algorithm.

        The rest of the array is used as internal workspace.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, ldq = *<value>* and `n` = *<value>*.

            Constraint: if `wantq` = True, ldq >= `n`.

        (`errno` -1)
            On entry, ldb = *<value>*, `n` = *<value>* and ncolb =
            *<value>*.

            Constraint: if ncolb > 0, ldb >= `n`.

        (`errno` -1)
            On entry, ncolb = *<value>*.

            Constraint: ncolb >= 0.

        (`errno` -1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            The QR algorithm has failed to converge. *<value>* singular
            values have **not** been found.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Chan, T F, 1982, `An improved algorithm for computing the singular
    value decomposition`, ACM Trans. Math. Software (8), 72--83

    Dongarra, J J, Moler, C B, Bunch, J R and Stewart, G W, 1979,
    `LINPACK Users' Guide`, SIAM, Philadelphia

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Hammarling, S, 1985, `The singular value decomposition in
    multivariate statistics`, SIGNUM Newsl. (20(3)), 2--25

    Wilkinson, J H, 1978, `Singular Value Decomposition -- Basic
    Aspects`, Numerical Software -- Needs and Availability, (ed D A H
    Jacobs), Academic Press
    """

    fun_name = 'naginterfaces.library.eigen.real_triang_svd'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_a, 'a'), (_b, 'b')),
    )
    _ncolb_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _ncolb = _b_utils._EngineIntScalarType(
        _ncolb_py, {'fun_name': fun_name, 'entity': '_ncolb'},
    ).to_ctypes()
    _wantq = _b_utils._EngineLogicalScalarType(
        wantq, {'fun_name': fun_name, 'entity': 'wantq'},
    ).to_ctypes()
    _wantq_py = _b_utils._EngineLogicalScalarType.to_py(_wantq)
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    if _wantq_py:
        _q_dim1 = _n_py
    else:
        _q_dim1 = 1
    if _wantq_py:
        _q_dim2 = _n_py
    else:
        _q_dim2 = 1
    exp_shape_q = (_q_dim1, _q_dim2)
    _q = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        exp_shape_q,
        sorder=_m_sorder,
    )
    _sv = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _wantp = _b_utils._EngineLogicalScalarType(
        wantp, {'fun_name': fun_name, 'entity': 'wantp'},
    ).to_ctypes()
    _wantp_py = _b_utils._EngineLogicalScalarType.to_py(_wantp)
    if _ncolb_py == 0 and not _wantq_py and not _wantp_py:
        _work_dim1 = max(1, 2*(_n_py-1))
    elif (_ncolb_py == 0 and not _wantq_py and _wantp_py) or (not _wantp_py and (_ncolb_py > 0 or _wantq_py)):
        _work_dim1 = max(1, 3*(_n_py-1))
    else:
        _work_dim1 = max(1, 5*(_n_py-1))
    exp_shape_work = (_work_dim1,)
    _work = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        exp_shape_work,
    )
    try:
        _b_eigen.real_triang_svd(
            _n,
            _a,
            _ncolb,
            _b,
            wantq,
            _q,
            _sv,
            wantp,
            _work,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _a_py = _a
    _b_py = _b
    _q_py = _q
    _sv_py = _sv
    _work_py = _work[:_n_py]
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'b', 'q', 'sv', 'work'],
    )(
        _a_py,
        _b_py,
        _q_py,
        _sv_py,
        _work_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def complex_triang_svd(n, a, b, wantq, wantp):
    u"""
    SVD of complex upper triangular matrix (Black Box).

    ``complex_triang_svd`` returns all, or part, of the singular value
    decomposition of a complex upper triangular matrix.

    For full information please refer to the NAG Library document for
    f02xu

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f02/f02xuf.html

    Parameters
    ----------
    n : int
        n, the order of the matrix R.

        If `n` = 0, an immediate return is effected.

    a : complex, array-like, shape (`n`, `n`)
        The leading n by n upper triangular part of the array `a` must
        contain the upper triangular matrix R.

    b : complex, array-like, shape (:, ncolb)
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if ncolb>0: `n`; otherwise: 1.

        If ncolb > 0, the leading n by ncolb part of the array `b` must
        contain the matrix to be transformed.

    wantq : bool
        Must be True if the matrix Q is required.

        If `wantq` = False then the array `q` is not referenced.

    wantp : bool
        Must be True if the matrix P^H is required, in which case P^H is
        returned in the array `a`, otherwise `wantp` must be False.

    Returns
    -------
    a : complex, ndarray, shape (`n`, `n`)
        If `wantp` = True, the n by n part of `a` will contain the n by
        n unitary matrix P^H, otherwise the n by n upper triangular part
        of `a` is used as internal workspace, but the strictly lower
        triangular part of `a` is not referenced.

    b : complex, ndarray, shape (:, ncolb)
        Is overwritten by the n by ncolb matrix Q^HB.

    q : complex, ndarray, shape (:, :)
        If `wantq` = True, the leading n by n part of the array `q` will
        contain the unitary matrix Q. Otherwise the array `q` is not
        referenced.

    sv : float, ndarray, shape (`n`)
        The n diagonal elements of the matrix S.

    rwork : float, ndarray, shape (:)
        `rwork`[`n`-1] contains the total number of iterations taken by
        the QR algorithm.

        The rest of the array is used as workspace.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, ldq = *<value>* and `n` = *<value>*.

            Constraint: if `wantq` = True, ldq >= `n`.

        (`errno` -1)
            On entry, ldb = *<value>*, `n` = *<value>* and ncolb =
            *<value>*.

            Constraint: if ncolb > 0, ldb >= `n`.

        (`errno` -1)
            On entry, ncolb = *<value>*.

            Constraint: ncolb >= 0.

        (`errno` -1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            The QR algorithm has failed to converge. *<value>* singular
            values have **not** been found.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Dongarra, J J, Moler, C B, Bunch, J R and Stewart, G W, 1979,
    `LINPACK Users' Guide`, SIAM, Philadelphia

    Hammarling, S, 1985, `The singular value decomposition in
    multivariate statistics`, SIGNUM Newsl. (20(3)), 2--25

    Wilkinson, J H, 1978, `Singular Value Decomposition -- Basic
    Aspects`, Numerical Software -- Needs and Availability, (ed D A H
    Jacobs), Academic Press
    """

    fun_name = 'naginterfaces.library.eigen.complex_triang_svd'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_a, 'a'), (_b, 'b')),
    )
    _ncolb_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _ncolb = _b_utils._EngineIntScalarType(
        _ncolb_py, {'fun_name': fun_name, 'entity': '_ncolb'},
    ).to_ctypes()
    _wantq = _b_utils._EngineLogicalScalarType(
        wantq, {'fun_name': fun_name, 'entity': 'wantq'},
    ).to_ctypes()
    _wantq_py = _b_utils._EngineLogicalScalarType.to_py(_wantq)
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    if _wantq_py:
        _q_dim1 = _n_py
    else:
        _q_dim1 = 1
    if _wantq_py:
        _q_dim2 = _n_py
    else:
        _q_dim2 = 1
    exp_shape_q = (_q_dim1, _q_dim2)
    _q = _b_utils._EngineComplex128ArrayType.empty_ndarray(
        exp_shape_q,
        sorder=_m_sorder,
    )
    _sv = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _wantp = _b_utils._EngineLogicalScalarType(
        wantp, {'fun_name': fun_name, 'entity': 'wantp'},
    ).to_ctypes()
    _wantp_py = _b_utils._EngineLogicalScalarType.to_py(_wantp)
    if _ncolb_py == 0 and not _wantq_py and not _wantp_py:
        _rwork_dim1 = 2*(_n_py-1)
    elif (_ncolb_py == 0 and not _wantq_py and _wantp_py) or (_ncolb_py > 0 and not _wantp_py) or (_wantq_py and not _wantp_py):
        _rwork_dim1 = 3*(_n_py-1)
    else:
        _rwork_dim1 = 5*(_n_py-1)
    exp_shape_rwork = (_rwork_dim1,)
    _rwork = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        exp_shape_rwork,
    )
    try:
        _b_eigen.complex_triang_svd(
            _n,
            _a,
            _ncolb,
            _b,
            wantq,
            _q,
            _sv,
            wantp,
            _rwork,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _a_py = _a
    _b_py = _b
    _q_py = _q
    _sv_py = _sv
    _rwork_py = _rwork
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'b', 'q', 'sv', 'rwork'],
    )(
        _a_py,
        _b_py,
        _q_py,
        _sv_py,
        _rwork_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data
