"""
Runner for ``naginterfaces.library`` example main functions.
"""

# NAG Copyright 2017-2018.

import argparse
import doctest
import importlib
import os
import sys

# Required for relative imports below in Python 3:
import naginterfaces.library.examples # pylint: disable=unused-import

ARG_PARSER = argparse.ArgumentParser(
    description=(
        'Runner (using doctest) for naginterfaces.library examples.'
    ),
    prog='naginterfaces.library.examples',
)

ARG_PARSER.add_argument(
    '--locate',
    action='store_true',
    default=False,
    help=(
        'don\'t run the examples, just print their locations on disk'
        ' (default: %(default)s)'
    )
)
ARG_PARSER.add_argument(
    '--suppress_expected',
    action='store_true',
    default=False,
    help=(
        'if running examples, don\'t run doctest in verbose mode, '
        'thus suppressing display of expected output (default: %(default)s)'
    )
)

ARGS = ARG_PARSER.parse_args()

PACKAGE = 'naginterfaces.library.examples'

EX_SCRIPTS = {
    thedir: [] for thedir in os.listdir(os.path.dirname(__file__))
    if (
        not thedir.startswith('__') and
        os.path.isdir(os.path.join(os.path.dirname(__file__), thedir))
    )
}

for thedir in EX_SCRIPTS:

    for _, _, files in os.walk(
            os.path.join(os.path.dirname(__file__), thedir)
    ):
        EX_SCRIPTS[thedir].extend([
            filename for filename in files
            if (
                filename.endswith('.py') and
                not filename.startswith('__')
            )
        ])

FAIL_LIST = []

for ex_dir in sorted(list(EX_SCRIPTS.keys())):

    if not EX_SCRIPTS[ex_dir]:
        continue

    if not ARGS.locate:
        print('-- BEGIN "' + ex_dir + '" examples')

    for ex_script in sorted(EX_SCRIPTS[ex_dir]):

        if ARGS.locate:
            print(os.path.join(os.path.dirname(__file__), ex_dir, ex_script))
            continue

        ex_name = '_'.join(os.path.splitext(ex_script)[0].split('_')[:-1])
        print('---- BEGIN "' + ex_name + '" example')
        ex_main = importlib.import_module(
            '.' + '.'.join([ex_dir, ex_name + '_ex']),
            package=PACKAGE
        )

        if ex_name == 'impl_details':
            ex_main.main()

        if doctest.testmod(
                ex_main,
                verbose=(
                    ex_name != 'impl_details' and
                    not ARGS.suppress_expected
                ),
                report=False,
                optionflags=doctest.ELLIPSIS | doctest.REPORT_NDIFF,
        ).failed:
            FAIL_LIST.append('.'.join([PACKAGE, ex_dir, ex_name + '_ex']))

        print('---- END "' + ex_name + '" example')

    if not ARGS.locate:
        print('-- END "' + ex_dir + '" examples')

if FAIL_LIST:
    sys.stderr.write('!Error: ' + str(len(FAIL_LIST)) + ' example')
    if len(FAIL_LIST) > 1:
        sys.stderr.write('s')
    sys.stderr.write(' failed:\n')
    sys.stderr.writelines(
        ['!FAIL : ' + fail_ex + '\n' for fail_ex in FAIL_LIST]
    )
elif not ARGS.locate:
    print('ok: all examples ran as expected.')

sys.exit(len(FAIL_LIST))
