# -*- coding: utf-8 -*-
# pylint: disable=chained-comparison,invalid-name,line-too-long,protected-access,too-many-arguments,too-many-boolean-expressions,too-many-branches,too-many-lines,too-many-locals,too-many-statements,unused-import
u"""
Interfaces for the NAG Mark 27.0 `ieee` Chapter.

``ieee`` - IEEE Arithmetic

This module provides functions to handle various aspects of IEEE
floating-point arithmetic behaviour.

For full information please refer to the NAG Library document

https://www.nag.com/numeric/nl/nagdoc_27/flhtml/x07/x07intro.html
"""

# NAG Copyright 2017-2020.

from ..base import ieee as _b_ieee
from ..base import utils as _b_utils
from ..library import _utils

def is_finite(x):
    u"""
    Determines whether its argument has a finite value.

    ``is_finite`` determines whether a floating-point number is finite.

    For full information please refer to the NAG Library document for
    x07aa

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/x07/x07aaf.html

    Parameters
    ----------
    x : float
        The number whose status is to be determined.

    Returns
    -------
    finite : bool
        ifinite = 1 if and only if `x` is finite, and ifinite = 0
        otherwise.

    References
    ----------
    IEEE, 2008, `Standard for Floating-Point Arithmetic` (IEEE Standard
    754-2008), IEEE, New York.
    """

    _return_scalar = _b_ieee.is_finite(
        x,
    )
    return _return_scalar

def is_nan(x):
    u"""
    Determines whether its argument is a NaN (Not A Number).

    ``is_nan`` determines whether a floating-point number is a NaN (Not
    A Number).

    For full information please refer to the NAG Library document for
    x07ab

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/x07/x07abf.html

    Parameters
    ----------
    x : float
        The number whose status is to be determined.

    Returns
    -------
    isnan : bool
        iisnan = 1 if and only if `x` is a NaN, and isnan = 0 otherwise.

    References
    ----------
    IEEE, 2008, `Standard for Floating-Point Arithmetic` (IEEE Standard
    754-2008), IEEE, New York.
    """

    _return_scalar = _b_ieee.is_nan(
        x,
    )
    return _return_scalar

def create_infinity(isign):
    u"""
    Creates a signed infinite value.

    ``create_infinity`` creates a signed infinite value.

    For full information please refer to the NAG Library document for
    x07ba

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/x07/x07baf.html

    Parameters
    ----------
    isign : int
        Determines the sign of the infinity to be created.

        If `isign` is greater than or equal to 0, a positive infinity is
        returned, otherwise a negative infinity is returned.

    Returns
    -------
    x : float
        The required infinite value.

    References
    ----------
    IEEE, 2008, `Standard for Floating-Point Arithmetic` (IEEE Standard
    754-2008), IEEE, New York.
    """

    _return_scalar = _b_ieee.create_infinity(
        isign,
    )
    _x_py = _return_scalar
    return _x_py

def create_nan(quiet):
    u"""
    Creates a NaN (Not A Number).

    ``create_nan`` creates a NaN (Not A Number).

    For full information please refer to the NAG Library document for
    x07bb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/x07/x07bbf.html

    Parameters
    ----------
    quiet : int
        Determines whether a quiet or a signalling NaN is to be created.
        If `quiet` = 1, the returned NaN is quiet, otherwise it is
        signalling. See reference IEEE (2008) for the distinction
        between the two kinds.

    Returns
    -------
    x : float
        The required NaN value.

    References
    ----------
    IEEE, 2008, `Standard for Floating-Point Arithmetic` (IEEE Standard
    754-2008), IEEE, New York.
    """

    _return_scalar = _b_ieee.create_nan(
        quiet,
    )
    _x_py = _return_scalar
    return _x_py

def get_exception_mode():
    u"""
    Gets current behaviour of floating-point exceptions.

    ``get_exception_mode`` gets the current IEEE exception halting mode.

    For full information please refer to the NAG Library document for
    x07ca

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/x07/x07caf.html

    Returns
    -------
    exceptionmode : int, ndarray, shape (3)
        Each of the three elements of `exceptionmode` is set to 1 if the
        corresponding condition will raise an exception, and is set to 0
        otherwise. `exceptionmode`[0] concerns floating-point overflow,
        `exceptionmode`[1] concerns floating-point division by zero, and
        `exceptionmode`[2] concerns floating-point invalid operation.

    References
    ----------
    IEEE, 2008, `Standard for Floating-Point Arithmetic` (IEEE Standard
    754-2008), IEEE, New York.
    """

    _exceptionmode = _b_utils._EngineIntArrayType.empty_ndarray(
        (3,),
    )
    _b_ieee.get_exception_mode(
        _exceptionmode,
    )
    _exceptionmode_py = _exceptionmode
    return _exceptionmode_py

def set_exception_mode(exceptionmode):
    u"""
    Sets behaviour of floating-point exceptions.

    ``set_exception_mode`` sets the current IEEE exception halting mode.

    For full information please refer to the NAG Library document for
    x07cb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/x07/x07cbf.html

    Parameters
    ----------
    exceptionmode : int, array-like, shape (3)
        Each of the three elements of `exceptionmode` must contain any
        nonzero value if the corresponding condition should raise an
        exception, and contain 0 otherwise. `exceptionmode`[0] concerns
        floating-point overflow, `exceptionmode`[1] concerns
        floating-point division by zero, and `exceptionmode`[2] concerns
        floating-point invalid operation.

    References
    ----------
    IEEE, 2008, `Standard for Floating-Point Arithmetic` (IEEE Standard
    754-2008), IEEE, New York.
    """

    fun_name = 'naginterfaces.library.ieee.set_exception_mode'

    _exceptionmode = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'exceptionmode'},
        exceptionmode,
        copy=False,
    )
    _b_ieee.set_exception_mode(
        _exceptionmode,
    )
