# -*- coding: utf-8 -*-
# pylint: disable=chained-comparison,invalid-name,line-too-long,protected-access,too-many-arguments,too-many-boolean-expressions,too-many-branches,too-many-lines,too-many-locals,too-many-statements,unused-import
u"""
Interfaces for the NAG Mark 27.0 `mv` Chapter.

``mv`` - Multivariate Methods

This module is concerned with methods for studying multivariate data.
A multivariate dataset consists of several variables recorded on a
number of objects or individuals.
Multivariate methods can be classified as those that seek to examine the
relationships between the variables (e.g., principal components), known
as variable-directed methods, and those that seek to examine the
relationships between the objects (e.g., cluster analysis), known as
individual-directed methods.

Multiple regression is not included in this module as it involves the
relationship of a single variable, known as the response variable, to
the other variables in the dataset, the explanatory variables.
Routines for multiple regression are provided in submodule ``correg``.

For full information please refer to the NAG Library document

https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g03/g03intro.html

See Also
--------
``naginterfaces.library.examples.mv`` :
    This subpackage contains examples for the ``mv`` module.
    See also the :ref:`library_mv_ex` subsection.
"""

# NAG Copyright 2017-2020.

import collections as _collections

from ..base import mv as _b_mv
from ..base import utils as _b_utils
from ..library import _utils

def prin_comp(matrix, std, x, isx, s, nvar, wt=None):
    u"""
    Performs principal component analysis.

    ``prin_comp`` performs a principal component analysis on a data
    matrix; both the principal component loadings and the principal
    component scores are returned.

    For full information please refer to the NAG Library document for
    g03aa

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g03/g03aaf.html

    Parameters
    ----------
    matrix : str, length 1
        Indicates for which type of matrix the principal component
        analysis is to be carried out.

        `matrix` = 'C'

            It is for the correlation matrix.

        `matrix` = 'S'

            It is for a standardized matrix, with standardizations given
            by `s`.

        `matrix` = 'U'

            It is for the sums of squares and cross-products matrix.

        `matrix` = 'V'

            It is for the variance-covariance matrix.

    std : str, length 1
        Indicates if the principal component scores are to be
        standardized.

        `std` = 'S'

            The principal component scores are standardized so that F'F
            = I, i.e., F = X_s PLambda^-1 = V.

        `std` = 'U'

            The principal component scores are unstandardized, i.e., F =
            X_s P = V Lambda.

        `std` = 'Z'

            The principal component scores are standardized so that they
            have unit variance.

        `std` = 'E'

            The principal component scores are standardized so that they
            have variance equal to the corresponding eigenvalue.

    x : float, array-like, shape (n, m)
        `x`[i-1,j-1] must contain the ith observation for the jth
        variable, for j = 1,2,...,m, for i = 1,2,...,n.

    isx : int, array-like, shape (m)
        `isx`[j-1] indicates whether or not the jth variable is to be
        included in the analysis.

        If `isx`[j-1] > 0, the variable contained in the jth column of
        `x` is included in the principal component analysis, for j =
        1,2,...,m.

    s : float, array-like, shape (m)
        The standardizations to be used, if any.

        If `matrix` = 'S', the first m elements of `s` must contain the
        standardization coefficients, the diagonal elements of sigma.

    nvar : int
        p, the number of variables in the principal component analysis.

    wt : None or float, array-like, shape (n), optional
        If `wt` is not None, the first n elements of `wt` must contain
        the weights to be used in the principal component analysis.

        If `wt`[i-1] = 0.0, the ith observation is not included in the
        analysis.

        The effective number of observations is the sum of the weights.

        If `wt` is None the effective number of observations is n.

    Returns
    -------
    s : float, ndarray, shape (m)
        If `matrix` = 'S', `s` is unchanged on exit.

        If `matrix` = 'C', `s` contains the variances of the selected
        variables. `s`[j-1] contains the variance of the variable in the
        jth column of `x` if `isx`[j-1] > 0.

        If `matrix` = 'U' or 'V', `s` is not referenced.

    e : float, ndarray, shape (`nvar`, 6)
        The statistics of the principal component analysis.

        `e`[i-1,0]

            The eigenvalues associated with the ith principal component,
            lambda_i^2, for i = 1,2,...,p.

        `e`[i-1,1]

            The proportion of variation explained by the ith principal
            component, for i = 1,2,...,p.

        `e`[i-1,2]

            The cumulative proportion of variation explained by the
            first ith principal components, for i = 1,2,...,p.

        `e`[i-1,3]

            The chi^2 statistics, for i = 1,2,...,p.

        `e`[i-1,4]

            The degrees of freedom for the chi^2 statistics, for i =
            1,2,...,p.

        If `matrix` != 'C', `e`[i-1,5] contains significance level for
        the chi^2 statistic, for i = 1,2,...,p.

        If `matrix` = 'C', `e`[i-1,5] is returned as zero.

    p : float, ndarray, shape (`nvar`, `nvar`)
        The first `nvar` columns of `p` contain the principal component
        loadings, a_i. The jth column of `p` contains the `nvar`
        coefficients for the jth principal component.

    v : float, ndarray, shape (n, `nvar`)
        The first `nvar` columns of `v` contain the principal component
        scores. The jth column of `v` contains the n scores for the jth
        principal component.

        If `wt` is not None, any rows for which `wt`[i-1] is zero will
        be set to zero.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `nvar` = *<value>* and m = *<value>*.

            Constraint: `nvar` <= m.

        (`errno` 1)
            On entry, `nvar` = *<value>*.

            Constraint: `nvar` >= 1.

        (`errno` 1)
            On entry, n = *<value>* and `nvar` = *<value>*.

            Constraint: n > `nvar`.

        (`errno` 1)
            On entry, `std` = *<value>*.

            Constraint: `std` = 'E', 'S', 'U' or 'Z'.

        (`errno` 1)
            On entry, `matrix` = *<value>*.

            Constraint: `matrix` = 'C', 'S', 'U' or 'V'.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 2.

        (`errno` 1)
            On entry, m = *<value>*.

            Constraint: m >= 1.

        (`errno` 2)
            On entry, i = *<value>* and `wt`[i-1] < 0.0.

            Constraint: `wt`[i-1] >= 0.0.

        (`errno` 3)
            Number of selected variables >= effective number of
            observations.

        (`errno` 3)
            On entry, `nvar` = *<value>* and *<value>* values of `isx` >
            0.

            Constraint: exactly `nvar` elements of `isx` > 0.

        (`errno` 4)
            On entry, j = *<value>* and `s`[j-1] <= 0.0.

            Constraint: `s`[j-1] > 0.0.

        (`errno` 5)
            The singular value decomposition has failed to converge.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 6)
            All eigenvalues/singular values are zero. This will be
            caused by all the variables being constant.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Chatfield, C and Collins, A J, 1980, `Introduction to Multivariate
    Analysis`, Chapman and Hall

    Cooley, W C and Lohnes, P R, 1971, `Multivariate Data Analysis`,
    Wiley

    Hammarling, S, 1985, `The singular value decomposition in
    multivariate statistics`, SIGNUM Newsl. (20(3)), 2--25

    Kendall, M G and Stuart, A, 1969, `The Advanced Theory of Statistics
    (Volume 1)`, (3rd Edition), Griffin

    Morrison, D F, 1967, `Multivariate Statistical Methods`,
    McGraw--Hill

    See Also
    --------
    :meth:`naginterfaces.library.examples.mv.prin_comp_ex.main`
    """

    fun_name = 'naginterfaces.library.mv.prin_comp'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_x, 'x'),),
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2)[1]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _isx = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'isx'},
        isx,
        copy=False,
    )
    _s = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 's'},
        s,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _wt = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'wt'},
        wt,
        copy=False,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nvar = _b_utils._EngineIntScalarType(
        nvar, {'fun_name': fun_name, 'entity': 'nvar'},
    ).to_ctypes()
    _nvar_py = _b_utils._EngineIntScalarType.to_py(_nvar)
    _e = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nvar_py, 6),
        sorder=_m_sorder,
    )
    _p = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nvar_py, _nvar_py),
        sorder=_m_sorder,
    )
    _v = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py, _nvar_py),
        sorder=_m_sorder,
    )
    try:
        _b_mv.prin_comp(
            matrix,
            std,
            _n,
            _m,
            _x,
            _isx,
            _s,
            _wt,
            _nvar,
            _e,
            _p,
            _v,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _s_py = _s
    _e_py = _e
    _p_py = _p
    _v_py = _v
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['s', 'e', 'p', 'v'],
    )(
        _s_py,
        _e_py,
        _p_py,
        _v_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def canon_var(weight, x, isx, ing, ng, wt, tol):
    u"""
    Performs canonical variate analysis.

    ``canon_var`` performs a canonical variate (canonical
    discrimination) analysis.

    For full information please refer to the NAG Library document for
    g03ac

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g03/g03acf.html

    Parameters
    ----------
    weight : str, length 1
        Indicates if weights are to be used.

        `weight` = 'U'

            No weights are used.

        `weight` = 'W' or 'V'

            Weights are used and must be supplied in `wt`.

        If `weight` = 'W', the weights are treated as frequencies and
        the effective number of observations is the sum of the weights.

        If `weight` = 'V', the weights are treated as being inversely
        proportional to the variance of the observations and the
        effective number of observations is the number of observations
        with nonzero weights.

    x : float, array-like, shape (n, m)
        `x`[i-1,j-1] must contain the ith observation for the jth
        variable, for j = 1,2,...,m, for i = 1,2,...,n.

    isx : int, array-like, shape (m)
        `isx`[j-1] indicates whether or not the jth variable is to be
        included in the analysis.

        If `isx`[j-1] > 0, the variables contained in the jth column of
        `x` is included in the canonical variate analysis, for j =
        1,2,...,m.

    ing : int, array-like, shape (n)
        `ing`[i-1] indicates which group the ith observation is in, for
        i = 1,2,...,n. The effective number of groups is the number of
        groups with nonzero membership.

    ng : int
        The number of groups, n_g.

    wt : float, array-like, shape (:)
        Note: the required length for this argument is determined as
        follows: if `weight` in ('W', 'V'): n; otherwise: 1.

        If `weight` = 'W' or 'V', the first n elements of `wt` must
        contain the weights to be used in the analysis.

        If `wt`[i-1] = 0.0, the ith observation is not included in the
        analysis.

        If `weight` = 'U', `wt` is not referenced.

    tol : float
        The value of `tol` is used to decide if the variables are of
        full rank and, if not, what is the rank of the variables. The
        smaller the value of `tol` the stricter the criterion for
        selecting the singular value decomposition. If a non-negative
        value of `tol` less than machine precision is entered, the
        square root of machine precision is used instead.

    Returns
    -------
    nig : int, ndarray, shape (`ng`)
        `nig`[j-1] gives the number of observations in group j, for j =
        1,2,...,n_g.

    cvm : float, ndarray, shape (`ng`, nx)
        `cvm`[i-1,j-1] contains the mean of the jth canonical variate
        for the ith group, for j = 1,2,...,l, for i = 1,2,...,n_g; the
        remaining columns, if any, are used as workspace.

    e : float, ndarray, shape (min(nx, `ng`-1), 6)
        The statistics of the canonical variate analysis.

        `e`[i-1,0]

            The canonical correlations, delta_i, for i = 1,2,...,l.

        `e`[i-1,1]

            The eigenvalues of the within-group sum of squares matrix,
            lambda_i^2, for i = 1,2,...,l.

        `e`[i-1,2]

            The proportion of variation explained by the ith canonical
            variate, for i = 1,2,...,l.

        `e`[i-1,3]

            The chi^2 statistic for the ith canonical variate, for i =
            1,2,...,l.

        `e`[i-1,4]

            The degrees of freedom for chi^2 statistic for the ith
            canonical variate, for i = 1,2,...,l.

        `e`[i-1,5]

            The significance level for the chi^2 statistic for the ith
            canonical variate, for i = 1,2,...,l.

    ncv : int
        The number of canonical variates, l. This will be the minimum of
        n_g-1 and the rank of `x`.

    cvx : float, ndarray, shape (nx, `ng`-1)
        The canonical variate loadings. `cvx`[i-1,j-1] contains the
        loading coefficient for the ith variable on the jth canonical
        variate, for j = 1,2,...,l, for i = 1,2,...,n_x; the remaining
        columns, if any, are used as workspace.

    irankx : int
        The rank of the dependent variables.

        If the variables are of full rank then `irankx` = nx.

        If the variables are not of full rank then `irankx` is an
        estimate of the rank of the dependent variables. `irankx` is
        calculated as the number of singular values greater than
        `tol`*(largest singular value).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `tol` = *<value>*.

            Constraint: `tol` >= 0.0.

        (`errno` 1)
            On entry, `weight` = *<value>*.

            Constraint: `weight` = 'U', 'W' or 'V'.

        (`errno` 1)
            On entry, lde = *<value>* and min(nx, `ng`-1) = *<value>*.

            Constraint: lde >= min(nx, `ng`-1).

        (`errno` 1)
            On entry, n = *<value>* and nx+`ng` = *<value>*.

            Constraint: n >= nx+`ng`.

        (`errno` 1)
            On entry, m = *<value>* and nx = *<value>*.

            Constraint: m >= nx.

        (`errno` 1)
            On entry, `ng` = *<value>*.

            Constraint: `ng` >= 2.

        (`errno` 1)
            On entry, nx = *<value>*.

            Constraint: nx >= 1.

        (`errno` 2)
            On entry, i = *<value>* and `wt`[i-1] < 0.0.

            Constraint: `wt`[i-1] >= 0.0.

        (`errno` 3)
            On entry, i = *<value>*, `ing`[i-1] = *<value>* and `ng` =
            *<value>*.

            Constraint: 1 <= `ing`[i-1] <= `ng`.

        (`errno` 4)
            On entry, nx = *<value>*, expected value = *<value>*.

            Constraint: nx must be consistent with `isx`.

        (`errno` 5)
            The singular value decomposition has failed to converge.

        (`errno` 7)
            The effective number of observations is less than the
            effective number of groups plus number of variables.

        (`errno` 7)
            Less than 2 groups have nonzero membership.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 6)
            A canonical correlation is equal to 1.0.

        (`errno` 8)
            The rank of `x` is 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Chatfield, C and Collins, A J, 1980, `Introduction to Multivariate
    Analysis`, Chapman and Hall

    Gnanadesikan, R, 1977, `Methods for Statistical Data Analysis of
    Multivariate Observations`, Wiley

    Hammarling, S, 1985, `The singular value decomposition in
    multivariate statistics`, SIGNUM Newsl. (20(3)), 2--25

    Kendall, M G and Stuart, A, 1969, `The Advanced Theory of Statistics
    (Volume 1)`, (3rd Edition), Griffin
    """

    fun_name = 'naginterfaces.library.mv.canon_var'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_x, 'x'),),
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2)[1]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _isx = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'isx'},
        isx,
        copy=False,
    )
    _ing = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ing'},
        ing,
        copy=False,
    )
    _wt = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'wt'},
        wt,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ng = _b_utils._EngineIntScalarType(
        ng, {'fun_name': fun_name, 'entity': 'ng'},
    ).to_ctypes()
    _ng_py = _b_utils._EngineIntScalarType.to_py(_ng)
    _nig = _b_utils._EngineIntArrayType.empty_ndarray(
        (_ng_py,),
    )
    _vl_nx_py = 0
    for _vl_i_py in range(1, _m_py+1):
        if _b_utils._EngineIntArrayType.get_element(_isx, {'fun_name': fun_name, 'entity': 'isx'}, (_vl_i_py-1,)) > 0:
            _vl_nx_py = _vl_nx_py + 1
    _cvm = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ng_py, _vl_nx_py),
        sorder=_m_sorder,
    )
    _e = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (min(_vl_nx_py, _ng_py-1), 6),
        sorder=_m_sorder,
    )
    _cvx = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_vl_nx_py, _ng_py-1),
        sorder=_m_sorder,
    )
    try:
        _return_scalars = _b_mv.canon_var(
            weight,
            _n,
            _m,
            _x,
            _isx,
            _ing,
            _ng,
            _wt,
            _nig,
            _cvm,
            _e,
            _cvx,
            tol,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _nig_py = _nig
    _cvm_py = _cvm
    _e_py = _e
    _ncv_py = _return_scalars.ncv
    _cvx_py = _cvx
    _irankx_py = _return_scalars.irankx
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['nig', 'cvm', 'e', 'ncv', 'cvx', 'irankx'],
    )(
        _nig_py,
        _cvm_py,
        _e_py,
        _ncv_py,
        _cvx_py,
        _irankx_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def canon_corr(z, isz, nx, ny, mcv, tol, wt=None):
    u"""
    Performs canonical correlation analysis.

    ``canon_corr`` performs canonical correlation analysis upon input
    data matrices.

    For full information please refer to the NAG Library document for
    g03ad

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g03/g03adf.html

    Parameters
    ----------
    z : float, array-like, shape (n, m)
        `z`[i-1,j-1] must contain the ith observation for the jth
        variable, for j = 1,2,...,m, for i = 1,2,...,n.

        Both x and y variables are to be included in `z`, the indicator
        array, `isz`, being used to assign the variables in `z` to the x
        or y sets as appropriate.

    isz : int, array-like, shape (m)
        `isz`[j-1] indicates whether or not the jth variable is included
        in the analysis and to which set of variables it belongs.

        `isz`[j-1] > 0

            The variable contained in the jth column of `z` is included
            as an x variable in the analysis.

        `isz`[j-1] < 0

            The variable contained in the jth column of `z` is included
            as a y variable in the analysis.

        `isz`[j-1] = 0

            The variable contained in the jth column of `z` is not
            included in the analysis.

    nx : int
        The number of x variables in the analysis, n_x.

    ny : int
        The number of y variables in the analysis, n_y.

    mcv : int
        An upper limit to the number of canonical variates.

    tol : float
        The value of `tol` is used to decide if the variables are of
        full rank and, if not, what is the rank of the variables. The
        smaller the value of `tol` the stricter the criterion for
        selecting the singular value decomposition. If a non-negative
        value of `tol` less than machine precision is entered, the
        square root of machine precision is used instead.

    wt : None or float, array-like, shape (n), optional
        The elements of `wt` must contain the weights to be used in the
        analysis. The effective number of observations is the sum of the
        weights. If `wt`[i-1] = 0.0 then the ith observation is not
        included in the analysis.

        If weights are not provided then `wt` must be set to None and
        the effective number of observations is n.

    Returns
    -------
    e : float, ndarray, shape (min(`nx`, `ny`), 6)
        The statistics of the canonical variate analysis.

        `e`[i-1,0]

            The canonical correlations, delta_i, for i = 1,2,...,l.

        `e`[i-1,1]

            The eigenvalues of Sigma, lambda_i^2, for i = 1,2,...,l.

        `e`[i-1,2]

            The proportion of variation explained by the ith canonical
            variate, for i = 1,2,...,l.

        `e`[i-1,3]

            The chi^2 statistic for the ith canonical variate, for i =
            1,2,...,l.

        `e`[i-1,4]

            The degrees of freedom for chi^2 statistic for the ith
            canonical variate, for i = 1,2,...,l.

        `e`[i-1,5]

            The significance level for the chi^2 statistic for the ith
            canonical variate, for i = 1,2,...,l.

    ncv : int
        The number of canonical correlations, l. This will be the
        minimum of the rank of X and the rank of Y.

    cvx : float, ndarray, shape (`nx`, `mcv`)
        The canonical variate loadings for the x variables.
        `cvx`[i-1,j-1] contains the loading coefficient for the ith x
        variable on the jth canonical variate.

    cvy : float, ndarray, shape (`ny`, `mcv`)
        The canonical variate loadings for the y variables.
        `cvy`[i-1,j-1] contains the loading coefficient for the ith y
        variable on the jth canonical variate.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `mcv` = *<value>* and min(`nx`, `ny`) = *<value>*.

            Constraint: `mcv` >= min(`nx`, `ny`).

        (`errno` 1)
            On entry, `tol` = *<value>*.

            Constraint: `tol` >= 0.0.

        (`errno` 1)
            On entry, weight = *<value>*.

            Constraint: weight = 'U' or 'W'.

        (`errno` 1)
            On entry, lde = *<value>* and min(`nx`, `ny`) = *<value>*.

            Constraint: lde >= min(`nx`, `ny`).

        (`errno` 1)
            On entry, n = *<value>* and `nx`+`ny` = *<value>*.

            Constraint: n > `nx`+`ny`.

        (`errno` 1)
            On entry, m = *<value>* and `nx`+`ny` = *<value>*.

            Constraint: m >= `nx`+`ny`.

        (`errno` 1)
            On entry, `ny` = *<value>*.

            Constraint: `ny` >= 1.

        (`errno` 1)
            On entry, `nx` = *<value>*.

            Constraint: `nx` >= 1.

        (`errno` 2)
            On entry, i = *<value>* and `wt`[i-1] < 0.0.

            Constraint: `wt`[i-1] >= 0.0.

        (`errno` 3)
            On entry, `ny` = *<value>*, expected value = *<value>*.

            Constraint: `ny` must be consistent with `isz`.

        (`errno` 3)
            On entry, `nx` = *<value>*, expected value = *<value>*.

            Constraint: `nx` must be consistent with `isz`.

        (`errno` 4)
            On entry, the effective number of observations is less than
            `nx`+`ny`+1.

        (`errno` 5)
            The singular value decomposition has failed to converge.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 6)
            A canonical correlation is equal to 1.0.

        (`errno` 7)
            On entry, the rank of the Y matrix is 0.

        (`errno` 7)
            On entry, the rank of the X matrix is 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Hastings, N A J and Peacock, J B, 1975, `Statistical Distributions`,
    Butterworth

    Kendall, M G and Stuart, A, 1976, `The Advanced Theory of Statistics
    (Volume 3)`, (3rd Edition), Griffin

    Morrison, D F, 1967, `Multivariate Statistical Methods`,
    McGraw--Hill
    """

    fun_name = 'naginterfaces.library.mv.canon_corr'

    _z = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'z'},
        z,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_z, 'z'),),
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_z, {'fun_name': fun_name, 'entity': 'z'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_z, {'fun_name': fun_name, 'entity': 'z'}, exp_rank=2)[1]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _isz = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'isz'},
        isz,
        copy=False,
    )
    _wt = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'wt'},
        wt,
        copy=False,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nx = _b_utils._EngineIntScalarType(
        nx, {'fun_name': fun_name, 'entity': 'nx'},
    ).to_ctypes()
    _nx_py = _b_utils._EngineIntScalarType.to_py(_nx)
    _ny = _b_utils._EngineIntScalarType(
        ny, {'fun_name': fun_name, 'entity': 'ny'},
    ).to_ctypes()
    _ny_py = _b_utils._EngineIntScalarType.to_py(_ny)
    _e = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (min(_nx_py, _ny_py), 6),
        sorder=_m_sorder,
    )
    _mcv = _b_utils._EngineIntScalarType(
        mcv, {'fun_name': fun_name, 'entity': 'mcv'},
    ).to_ctypes()
    _mcv_py = _b_utils._EngineIntScalarType.to_py(_mcv)
    _cvx = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nx_py, _mcv_py),
        sorder=_m_sorder,
    )
    _cvy = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ny_py, _mcv_py),
        sorder=_m_sorder,
    )
    try:
        _return_scalar = _b_mv.canon_corr(
            _n,
            _m,
            _z,
            _isz,
            _nx,
            _ny,
            _wt,
            _e,
            _cvx,
            _mcv,
            _cvy,
            tol,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _e_py = _e
    _ncv_py = _return_scalar
    _cvx_py = _cvx
    _cvy_py = _cvy
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['e', 'ncv', 'cvx', 'cvy'],
    )(
        _e_py,
        _ncv_py,
        _cvx_py,
        _cvy_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def rot_orthomax(stand, g, fl, acc=0.00001, maxit=30):
    u"""
    Computes orthogonal rotations for loading matrix, generalized
    orthomax criterion.

    ``rot_orthomax`` computes orthogonal rotations for a matrix of
    loadings using a generalized orthomax criterion.

    For full information please refer to the NAG Library document for
    g03ba

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g03/g03baf.html

    Parameters
    ----------
    stand : str, length 1
        Indicates if the matrix of loadings is to be row standardized
        before rotation.

        `stand` = 'S'

            The loadings are row standardized.

        `stand` = 'U'

            The loadings are left unstandardized.

    g : float
        gamma, the criterion constant with gamma = 1.0 giving varimax
        rotations and gamma = 0.0 giving quartimax rotations.

    fl : float, array-like, shape (nvar, k)
        Lambda, the matrix of loadings. `fl`[i-1,j-1] must contain the
        loading for the ith variable on the jth factor, for j =
        1,2,...,k, for i = 1,2,...,p.

    acc : float, optional
        Indicates the accuracy required. The iterative procedure of
        Cooley and Lohnes (1971) will be stopped and the final
        refinement computed when the change in V is less than
        `acc`*max(1.0, V). If `acc` is greater than or equal to 0.0 but
        less than machine precision or if `acc` is greater than 1.0,
        machine precision will be used instead.

    maxit : int, optional
        The maximum number of iterations.

    Returns
    -------
    fl : float, ndarray, shape (nvar, k)
        If `stand` = 'S', the elements of `fl` are standardized so that
        the sum of squared elements for each row is 1.0 and then after
        the computation of the rotations are rescaled; this may lead to
        slight differences between the input and output values of `fl`.

        If `stand` = 'U', `fl` will be unchanged on exit.

    flr : float, ndarray, shape (nvar, k)
        The rotated matrix of loadings, Lambda*. `flr`[i-1,j-1] will
        contain the rotated loading for the ith variable on the jth
        factor, for j = 1,2,...,k, for i = 1,2,...,p.

    r : float, ndarray, shape (k, k)
        The matrix of rotations, R.

    itera : int
        The number of iterations performed.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `maxit` = *<value>*.

            Constraint: `maxit` > 0.

        (`errno` 1)
            On entry, nvar = *<value>* and k = *<value>*.

            Constraint: nvar >= k.

        (`errno` 1)
            On entry, k = *<value>*.

            Constraint: k >= 2.

        (`errno` 1)
            On entry, `g` = *<value>*.

            Constraint: `g` >= 0.0.

        (`errno` 1)
            On entry, `stand` = *<value>*.

            Constraint: `stand` = 'S' or 'U'.

        (`errno` 1)
            On entry, `acc` = *<value>*.

            Constraint: `acc` >= 0.0.

        (`errno` 2)
            The singular value decomposition has failed to converge.

        (`errno` 3)
            The algorithm to find R has failed to reach the required
            accuracy in `maxit` iterations.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Cooley, W C and Lohnes, P R, 1971, `Multivariate Data Analysis`,
    Wiley

    Lawley, D N and Maxwell, A E, 1971, `Factor Analysis as a
    Statistical Method`, (2nd Edition), Butterworths
    """

    fun_name = 'naginterfaces.library.mv.rot_orthomax'

    _fl = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'fl'},
        fl,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_fl, 'fl'),),
    )
    _nvar_py = _b_utils._EngineFloat64ArrayType.get_shape(_fl, {'fun_name': fun_name, 'entity': 'fl'}, exp_rank=2)[0]
    _nvar = _b_utils._EngineIntScalarType(
        _nvar_py, {'fun_name': fun_name, 'entity': '_nvar'},
    ).to_ctypes()
    _k_py = _b_utils._EngineFloat64ArrayType.get_shape(_fl, {'fun_name': fun_name, 'entity': 'fl'}, exp_rank=2)[1]
    _k = _b_utils._EngineIntScalarType(
        _k_py, {'fun_name': fun_name, 'entity': '_k'},
    ).to_ctypes()
    _flr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nvar_py, _k_py),
        sorder=_m_sorder,
    )
    _r = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_k_py, _k_py),
        sorder=_m_sorder,
    )
    _return_scalar = _b_mv.rot_orthomax(
        stand,
        g,
        _nvar,
        _k,
        _fl,
        _flr,
        _r,
        acc,
        maxit,
    )
    _fl_py = _fl
    _flr_py = _flr
    _r_py = _r
    _itera_py = _return_scalar
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['fl', 'flr', 'r', 'itera'],
    )(
        _fl_py,
        _flr_py,
        _r_py,
        _itera_py,
    )
    return _return_data

def rot_procrustes(stand, pscale, x, y):
    u"""
    Computes Procrustes rotations.

    ``rot_procrustes`` computes Procrustes rotations in which an
    orthogonal rotation is found so that a transformed matrix best
    matches a target matrix.

    For full information please refer to the NAG Library document for
    g03bc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g03/g03bcf.html

    Parameters
    ----------
    stand : str, length 1
        Indicates if translation/normalization is required.

        `stand` = 'N'

            There is no translation or normalization.

        `stand` = 'Z'

            There is translation to the origin (i.e., to zero).

        `stand` = 'C'

            There is translation to origin and then to the Y centroid
            after rotation.

        `stand` = 'U'

            There is unit normalization.

        `stand` = 'S'

            There is translation and normalization (i.e., there is
            standardization).

        `stand` = 'M'

            There is translation and normalization to Y scale, then
            translation to the Y centroid after rotation (i.e., they are
            matched).

    pscale : str, length 1
        Indicates if least squares scaling is to be applied after
        rotation.

        `pscale` = 'S'

            Scaling is applied.

        `pscale` = 'U'

            No scaling is applied.

    x : float, array-like, shape (n, m)
        X, the matrix to be rotated.

    y : float, array-like, shape (n, m)
        The target matrix, `y`.

    Returns
    -------
    x : float, ndarray, shape (n, m)
        If `stand` = 'N', `x` will be unchanged.

        If `stand` = 'Z', 'C', 'S' or 'M', `x` will be translated to
        have zero column means.

        If `stand` = 'U' or 'S', `x` will be scaled to have unit sum of
        squares.

        If `stand` = 'M', `x` will be scaled to have the same sum of
        squares as `y`.

    y : float, ndarray, shape (n, m)
        If `stand` = 'N', `y` will be unchanged.

        If `stand` = 'Z' or 'S', `y` will be translated to have zero
        column means.

        If `stand` = 'U' or 'S', `y` will be scaled to have unit sum of
        squares.

        If `stand` = 'C' or 'M', `y` will be translated and then after
        rotation translated back.

        The output `y` should be the same as the input `y` except for
        rounding errors.

    yhat : float, ndarray, shape (n, m)
        The fitted matrix, Y^.

    r : float, ndarray, shape (m, m)
        The matrix of rotations, R, see [Further Comments].

    alpha : float
        If `pscale` = 'S' the scaling factor, alpha; otherwise `alpha`
        is not set.

    rss : float
        The residual sum of squares.

    res : float, ndarray, shape (n)
        The residuals, r_i, for i = 1,2,...,n.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `pscale` = *<value>*.

            Constraint: `pscale` = 'S' or 'U'.

        (`errno` 1)
            On entry, `stand` = *<value>*.

            Constraint: `stand` = 'N', 'Z', 'C', 'U', 'S' or 'M'.

        (`errno` 1)
            On entry, m = *<value>*.

            Constraint: m > 0.

        (`errno` 1)
            On entry, n = *<value>* and m = *<value>*.

            Constraint: n >= m.

        (`errno` 2)
            Only one distinct point (centred at zero) in `y` array.

        (`errno` 2)
            Only one distinct point (centred at zero) in `x` array.

        (`errno` 3)
            `yhat` contains only zero-points when least squares scaling
            is applied.

        (`errno` 4)
            The singular value decomposition has failed to converge.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Krzanowski, W J, 1990, `Principles of Multivariate Analysis`, Oxford
    University Press

    Lawley, D N and Maxwell, A E, 1971, `Factor Analysis as a
    Statistical Method`, (2nd Edition), Butterworths
    """

    fun_name = 'naginterfaces.library.mv.rot_procrustes'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_x, 'x'), (_y, 'y')),
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2)[1]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _yhat = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py, _m_py),
        sorder=_m_sorder,
    )
    _r = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_m_py, _m_py),
        sorder=_m_sorder,
    )
    _res = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _return_scalars = _b_mv.rot_procrustes(
        stand,
        pscale,
        _n,
        _m,
        _x,
        _y,
        _yhat,
        _r,
        _res,
    )
    _x_py = _x
    _y_py = _y
    _yhat_py = _yhat
    _r_py = _r
    _alpha_py = _return_scalars.alpha
    _rss_py = _return_scalars.rss
    _res_py = _res
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'y', 'yhat', 'r', 'alpha', 'rss', 'res'],
    )(
        _x_py,
        _y_py,
        _yhat_py,
        _r_py,
        _alpha_py,
        _rss_py,
        _res_py,
    )
    return _return_data

def rot_promax(stand, x, ro, power, io_manager=None):
    u"""
    ProMax rotations.

    ``rot_promax`` calculates a ProMax rotation, given information
    following an orthogonal rotation.

    For full information please refer to the NAG Library document for
    g03bd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g03/g03bdf.html

    Parameters
    ----------
    stand : str, length 1
        Indicates how loadings are normalized.

        `stand` = 'S'

            Rows of Y are (Kaiser) normalized by the communalities of
            the loadings.

        `stand` = 'U'

            Rows are not normalized.

    x : float, array-like, shape (n, m)
        The loadings matrix following an orthogonal rotation, X.

    ro : float, array-like, shape (m, m)
        The orthogonal rotation matrix, O.

    power : float
        p, the value of exponent.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    fp : float, ndarray, shape (n, m)
        The factor pattern matrix, P.

    r : float, ndarray, shape (m, m)
        The ProMax rotation matrix, R.

    phi : float, ndarray, shape (m, m)
        The matrix of inter-factor correlations, Phi.

    fs : float, ndarray, shape (n, m)
        The factor structure matrix, S.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `stand` = *<value>*.

            Constraint: `stand` = 'U' or 'S'.

        (`errno` 1)
            On entry, m = *<value>*.

            Constraint: m >= 1.

        (`errno` 1)
            On entry, `power` = *<value>*.

            Constraint: `power` > 1.0.

        (`errno` 2)
            On entry, n = *<value>* and m = *<value>*.

            Constraint: n >= m.

        (`errno` 20)
            The singular value decomposition has failed to converge.

        (`errno` 100)
            An internal error has occurred in this function. Check the
            function call and any array sizes. If the call is correct
            then please contact NAG for assistance.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    fun_name = 'naginterfaces.library.mv.rot_promax'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ro = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ro'},
        ro,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_x, 'x'), (_ro, 'ro')),
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2)[1]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _fp = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py, _m_py),
        sorder=_m_sorder,
    )
    _r = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_m_py, _m_py),
        sorder=_m_sorder,
    )
    _phi = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_m_py, _m_py),
        sorder=_m_sorder,
    )
    _fs = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py, _m_py),
        sorder=_m_sorder,
    )
    _b_mv.rot_promax(
        stand,
        _n,
        _m,
        _x,
        _ro,
        power,
        _fp,
        _r,
        _phi,
        _fs,
        io_manager,
    )
    _fp_py = _fp
    _r_py = _r
    _phi_py = _phi
    _fs_py = _fs
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['fp', 'r', 'phi', 'fs'],
    )(
        _fp_py,
        _r_py,
        _phi_py,
        _fs_py,
    )
    return _return_data

def factor(matrix, n, x, nvar, isx, nfac, iop, wt=None, io_manager=None):
    u"""
    Computes maximum likelihood estimates of the parameters of a factor
    analysis model, factor loadings, communalities and residual
    correlations.

    ``factor`` computes the maximum likelihood estimates of the
    parameters of a factor analysis model.
    Either the data matrix or a correlation/covariance matrix may be
    input.
    Factor loadings, communalities and residual correlations are
    returned.

    For full information please refer to the NAG Library document for
    g03ca

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g03/g03caf.html

    Parameters
    ----------
    matrix : str, length 1
        Selects the type of matrix on which factor analysis is to be
        performed.

        `matrix` = 'D'

            The data matrix will be input in `x` and factor analysis
            will be computed for the correlation matrix.

        `matrix` = 'S'

            The data matrix will be input in `x` and factor analysis
            will be computed for the covariance matrix, i.e., the
            results are scaled as described in [Further Comments].

        `matrix` = 'C'

            The correlation/variance-covariance matrix will be input in
            `x` and factor analysis computed for this matrix.

        See [Further Comments].

    n : int
        If `matrix` = 'D' or 'S' the number of observations in the data
        array `x`.

        If `matrix` = 'C' the (effective) number of observations used in
        computing the (possibly weighted)
        correlation/variance-covariance matrix input in `x`.

    x : float, array-like, shape (:, m)
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `matrix` in ('D', 'S'): `n`; if
        `matrix`='C': m; otherwise: 0.

        The input matrix.

        If `matrix` = 'D' or 'S', `x` must contain the data matrix,
        i.e., `x`[i-1,j-1] must contain the ith observation for the jth
        variable, for j = 1,2,...,m, for i = 1,2,...,n.

        If `matrix` = 'C', `x` must contain the correlation or
        variance-covariance matrix.

        Only the upper triangular part is required.

    nvar : int
        p, the number of variables in the factor analysis.

    isx : int, array-like, shape (m)
        `isx`[j-1] indicates whether or not the jth variable is included
        in the factor analysis. If `isx`[j-1] >= 1, the variable
        represented by the jth column of `x` is included in the
        analysis; otherwise it is excluded, for j = 1,2,...,m.

    nfac : int
        k, the number of factors.

    iop : int, array-like, shape (5)
        Options for the optimization. There are four options to be set:

        [table omitted]

        Let epsilon = machine precision then if `iop`[0] = 0, the
        following default values are used:

            iprint = -1

            maxfun = 100p

            acc = 10sqrt(epsilon)

            eps = epsilon

        If `iop`[0] != 0, then

            iprint = `iop`[1]

            maxfun = `iop`[2]

            acc = 10^(-l) where l = `iop`[3]

            eps = 10^(-l) where l = `iop`[4]

    wt : None or float, array-like, shape (:), optional
        Note: the required length for this argument is determined as
        follows: if weight='W' and `matrix` in ('D', 'S'): `n`;
        otherwise: 1.

        If `wt` is not None and `matrix` = 'D' or 'S', `wt` must contain
        the weights to be used in the factor analysis. The effective
        number of observations in the analysis will then be the sum of
        weights. If `wt`[i-1] = 0.0, the ith observation is not included
        in the analysis.

        If weight = 'U' or `matrix` = 'C', `wt` is not referenced and
        the effective number of observations is n.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    e : float, ndarray, shape (`nvar`)
        The eigenvalues theta_i, for i = 1,2,...,p.

    stat : float, ndarray, shape (4)
        The test statistics.

        `stat`[0]

            Contains the value F(Psi^).

        `stat`[1]

            Contains the test statistic, chi^2.

        `stat`[2]

            Contains the degrees of freedom associated with the test
            statistic.

        `stat`[3]

            Contains the significance level.

    com : float, ndarray, shape (`nvar`)
        The communalities.

    psi : float, ndarray, shape (`nvar`)
        The estimates of psi_i, for i = 1,2,...,p.

    res : float, ndarray, shape (`nvar`*(`nvar`-1)/2)
        The residual correlations. The residual correlation for the ith
        and jth variables is stored in `res`[ (j-1) (j-2) /2+i -1], i <
        j.

    fl : float, ndarray, shape (`nvar`, `nfac`)
        The factor loadings. `fl`[i-1,j-1] contains lambda_{ij}, for j =
        1,2,...,k, for i = 1,2,...,p.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `iop`[0] != 1 and `iop`[4] = *<value>*.

            Constraint: 1 <= eps <= machine precision.

        (`errno` 1)
            On entry, `iop`[0] != 1 and `iop`[3] = *<value>*.

            Constraint: 1 <= acc <= machine precision.

        (`errno` 1)
            On entry, `iop`[0] != 1 and `iop`[2] = *<value>*.

            Constraint: maxfun >= 1.

        (`errno` 1)
            On entry, `nfac` = *<value>* and `nvar` = *<value>*.

            Constraint: `nfac` <= `nvar`.

        (`errno` 1)
            On entry, `nfac` = *<value>*.

            Constraint: `nfac` >= 1.

        (`errno` 1)
            On entry, m = *<value>* and `nvar` = *<value>*.

            Constraint: m >= `nvar`.

        (`errno` 1)
            On entry, `matrix` = *<value>*.

            Constraint: `matrix` = 'D', 'S' or 'C'.

        (`errno` 1)
            On entry, weight = *<value>*.

            Constraint: when `matrix` = 'D' or 'S', weight = 'U' or 'W'.

        (`errno` 1)
            On entry, `n` = *<value>* and `nvar` = *<value>*.

            Constraint: `n` > `nvar`.

        (`errno` 1)
            On entry, `nvar` = *<value>*.

            Constraint: `nvar` > 1.

        (`errno` 2)
            On entry, i = *<value>* and `wt`[i-1] < 0.0.

            Constraint: `wt`[i-1] >= 0.0.

        (`errno` 3)
            The effective number of observations <= 1.

        (`errno` 3)
            The number of variables >= number of included observations.

        (`errno` 3)
            On entry, `nvar` = *<value>* and *<value>* values of `isx` >
            0

            Constraint: exactly `nvar` elements of `isx` > 0.

        (`errno` 4)
            Two eigenvalues of S* are equal.

        (`errno` 4)
            On entry, the data matrix is not of full column rank or the
            input correlation/covariance matrix is not positive
            definite.

        (`errno` 5)
            The singular value decomposition has failed to converge.

        (`errno` 6)
            The estimation procedure has failed to converge in *<value>*
            iterations.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 7)
            The convergence is not certain but a lower point could not
            be found.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Clark, M R B, 1970, `A rapidly convergent method for maximum
    likelihood factor analysis`, British J. Math. Statist. Psych.

    Hammarling, S, 1985, `The singular value decomposition in
    multivariate statistics`, SIGNUM Newsl. (20(3)), 2--25

    Lawley, D N and Maxwell, A E, 1971, `Factor Analysis as a
    Statistical Method`, (2nd Edition), Butterworths

    Morrison, D F, 1967, `Multivariate Statistical Methods`,
    McGraw--Hill
    """

    fun_name = 'naginterfaces.library.mv.factor'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_x, 'x'),),
    )
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2)[1]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _isx = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'isx'},
        isx,
        copy=False,
    )
    _wt = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'wt'},
        wt,
        copy=False,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nvar = _b_utils._EngineIntScalarType(
        nvar, {'fun_name': fun_name, 'entity': 'nvar'},
    ).to_ctypes()
    _nvar_py = _b_utils._EngineIntScalarType.to_py(_nvar)
    _e = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nvar_py,),
    )
    _stat = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (4,),
    )
    _com = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nvar_py,),
    )
    _psi = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nvar_py,),
    )
    _res = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nvar_py*(_nvar_py-1)//2,),
    )
    _nfac = _b_utils._EngineIntScalarType(
        nfac, {'fun_name': fun_name, 'entity': 'nfac'},
    ).to_ctypes()
    _nfac_py = _b_utils._EngineIntScalarType.to_py(_nfac)
    _fl = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nvar_py, _nfac_py),
        sorder=_m_sorder,
    )
    _iop = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'iop'},
        iop,
        copy=False,
    )
    try:
        _b_mv.factor(
            matrix,
            n,
            _m,
            _x,
            _nvar,
            _isx,
            _nfac,
            _wt,
            _e,
            _stat,
            _com,
            _psi,
            _res,
            _fl,
            _iop,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _e_py = _e
    _stat_py = _stat
    _com_py = _com
    _psi_py = _psi
    _res_py = _res
    _fl_py = _fl
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['e', 'stat', 'com', 'psi', 'res', 'fl'],
    )(
        _e_py,
        _stat_py,
        _com_py,
        _psi_py,
        _res_py,
        _fl_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def factor_score(method, rotate, fl, psi, e, r):
    u"""
    Computes factor score coefficients (for use after ``factor``).

    ``factor_score`` computes factor score coefficients from the result
    of fitting a factor analysis model by maximum likelihood as
    performed by ``factor``.

    For full information please refer to the NAG Library document for
    g03cc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g03/g03ccf.html

    Parameters
    ----------
    method : str, length 1
        Indicates which method is to be used to compute the factor score
        coefficients.

        `method` = 'R'

            The regression method is used.

        `method` = 'B'

            Bartlett's method is used.

    rotate : str, length 1
        Indicates whether a rotation is to be applied.

        `rotate` = 'R'

            A rotation will be applied to the coefficients and the
            rotation matrix, R, must be given in `r`.

        `rotate` = 'U'

            No rotation is applied.

    fl : float, array-like, shape (nvar, nfac)
        Lambda, the matrix of unrotated factor loadings as returned by
        ``factor``.

    psi : float, array-like, shape (nvar)
        The diagonal elements of Psi, as returned by ``factor``.

    e : float, array-like, shape (nvar)
        The eigenvalues of the matrix S*, as returned by ``factor``.

    r : float, array-like, shape (:, :)
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `rotate`='R': nfac; otherwise: 1.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `rotate`='U': 1; if `rotate`='R':
        nfac; otherwise: 0.

        If `rotate` = 'R', `r` must contain the orthogonal rotation
        matrix, R, as returned by ``rot_orthomax``.

        If `rotate` = 'U', `r` need not be set.

    Returns
    -------
    fs : float, ndarray, shape (nvar, nfac)
        The matrix of factor score coefficients, Phi. `fs`[i-1,j-1]
        contains the factor score coefficient for the jth factor and the
        ith observed variable, for j = 1,2,...,k, for i = 1,2,...,p.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `rotate` = *<value>*.

            Constraint: `rotate` = 'R' or 'U'.

        (`errno` 1)
            On entry, `method` = *<value>*

            Constraint: `method` = 'B' or 'R'.

        (`errno` 1)
            On entry, nvar = *<value>* and nfac = *<value>*.

            Constraint: nvar >= nfac.

        (`errno` 1)
            On entry, nfac = *<value>*.

            Constraint: nfac >= 1.

        (`errno` 2)
            On entry, i = *<value>* and `psi`[i-1] <= 0.0.

            Constraint: `psi`[i-1] > 0.0.

        (`errno` 2)
            On entry, i = *<value>* and `e`[i-1] <= 1.0.

            Constraint: `e`[i-1] > 1.0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Lawley, D N and Maxwell, A E, 1971, `Factor Analysis as a
    Statistical Method`, (2nd Edition), Butterworths
    """

    fun_name = 'naginterfaces.library.mv.factor_score'

    _fl = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'fl'},
        fl,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _r = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'r'},
        r,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_fl, 'fl'), (_r, 'r')),
    )
    _nvar_py = _b_utils._EngineFloat64ArrayType.get_shape(_fl, {'fun_name': fun_name, 'entity': 'fl'}, exp_rank=2)[0]
    _nvar = _b_utils._EngineIntScalarType(
        _nvar_py, {'fun_name': fun_name, 'entity': '_nvar'},
    ).to_ctypes()
    _nfac_py = _b_utils._EngineFloat64ArrayType.get_shape(_fl, {'fun_name': fun_name, 'entity': 'fl'}, exp_rank=2)[1]
    _nfac = _b_utils._EngineIntScalarType(
        _nfac_py, {'fun_name': fun_name, 'entity': '_nfac'},
    ).to_ctypes()
    _psi = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'psi'},
        psi,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _e = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'e'},
        e,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _fs = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nvar_py, _nfac_py),
        sorder=_m_sorder,
    )
    _b_mv.factor_score(
        method,
        rotate,
        _nvar,
        _nfac,
        _fl,
        _psi,
        _e,
        _r,
        _fs,
    )
    _fs_py = _fs
    return _fs_py

def discrim(x, isx, ing, ng, wt=None):
    u"""
    Computes test statistic for equality of within-group covariance
    matrices and matrices for discriminant analysis.

    ``discrim`` computes a test statistic for the equality of
    within-group covariance matrices and also computes matrices for use
    in discriminant analysis.

    For full information please refer to the NAG Library document for
    g03da

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g03/g03daf.html

    Parameters
    ----------
    x : float, array-like, shape (n, m)
        `x`[k-1,l-1] must contain the kth observation for the lth
        variable, for l = 1,2,...,m, for k = 1,2,...,n.

    isx : int, array-like, shape (m)
        `isx`[l-1] indicates whether or not the lth variable in `x` is
        to be included in the variance-covariance matrices.

        If `isx`[l-1] > 0 the lth variable is included, for l =
        1,2,...,m; otherwise it is not referenced.

    ing : int, array-like, shape (n)
        `ing`[k-1] indicates to which group the kth observation belongs,
        for k = 1,2,...,n.

    ng : int
        The number of groups, n_g.

    wt : None or float, array-like, shape (n), optional
        The elements of `wt` must contain the weights to be used in the
        analysis and the effective number of observations for a group is
        the sum of the weights of the observations in that group. If
        `wt`[k-1] = 0.0 then the kth observation is excluded from the
        calculations.

        If weights are not provided then `wt` must be set to None and
        the effective number of observations for a group is the number
        of observations in that group.

    Returns
    -------
    nig : int, ndarray, shape (`ng`)
        `nig`[j-1] contains the number of observations in the jth group,
        for j = 1,2,...,n_g.

    gmn : float, ndarray, shape (`ng`, nvar)
        The jth row of `gmn` contains the means of the p selected
        variables for the jth group, for j = 1,2,...,n_g.

    det : float, ndarray, shape (`ng`)
        The logarithm of the determinants of the within-group
        variance-covariance matrices.

    gc : float, ndarray, shape ((`ng`+1)*nvar*(nvar+1)/2)
        The first p(p+1)/2 elements of `gc` contain R and the remaining
        n_g blocks of p(p+1)/2 elements contain the R_j matrices. All
        are stored in packed form by columns.

    stat : float
        The likelihood-ratio test statistic, G.

    df : float
        The degrees of freedom for the distribution of G.

    sig : float
        The significance level for G.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, weight = *<value>*.

            Constraint: weight = 'U' or 'W'.

        (`errno` 1)
            On entry, m = *<value>* and nvar = *<value>*.

            Constraint: m >= nvar.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 1)
            On entry, `ng` = *<value>*.

            Constraint: `ng` >= 2.

        (`errno` 1)
            On entry, nvar = *<value>*.

            Constraint: nvar >= 1.

        (`errno` 2)
            On entry, i = *<value>* and `wt`[i-1] < 0.0.

            Constraint: `wt`[i-1] >= 0.0.

        (`errno` 3)
            The number of observations for group *<value>* is less than
            nvar.

        (`errno` 3)
            The effective number of observations for group *<value>* is
            less than 1.

        (`errno` 3)
            On entry, i = *<value>*, `ing`[i-1] = *<value>* and `ng` =
            *<value>*.

            Constraint: 1 <= `ing`[i-1] <= `ng`.

        (`errno` 3)
            On entry, nvar = *<value>* and *<value>* values of `isx` > 0

            Constraint: exactly nvar elements of `isx` > 0.

        (`errno` 4)
            R is not of full rank.

        (`errno` 4)
            R_j is not of full rank for j = *<value>*.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Aitchison, J and Dunsmore, I R, 1975, `Statistical Prediction
    Analysis`, Cambridge

    Kendall, M G and Stuart, A, 1976, `The Advanced Theory of Statistics
    (Volume 3)`, (3rd Edition), Griffin

    Krzanowski, W J, 1990, `Principles of Multivariate Analysis`, Oxford
    University Press

    Morrison, D F, 1967, `Multivariate Statistical Methods`,
    McGraw--Hill

    See Also
    --------
    :meth:`naginterfaces.library.examples.mv.discrim_group_ex.main`
    """

    fun_name = 'naginterfaces.library.mv.discrim'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_x, 'x'),),
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2)[1]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _isx = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'isx'},
        isx,
        copy=False,
    )
    _nvar = 0
    _nvar_py = _nvar
    for _vl_i_py in range(0, _m_py-1+1):
        if _b_utils._EngineIntArrayType.get_element(_isx, {'fun_name': fun_name, 'entity': 'isx'}, (_vl_i_py,)) > 0:
            _nvar = _nvar+1
    _nvar_py = _b_utils._EngineIntScalarType.to_py(_nvar)
    _ing = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ing'},
        ing,
        copy=False,
    )
    _wt = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'wt'},
        wt,
        copy=False,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ng = _b_utils._EngineIntScalarType(
        ng, {'fun_name': fun_name, 'entity': 'ng'},
    ).to_ctypes()
    _ng_py = _b_utils._EngineIntScalarType.to_py(_ng)
    _nig = _b_utils._EngineIntArrayType.empty_ndarray(
        (_ng_py,),
    )
    _gmn = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ng_py, _nvar_py),
        sorder=_m_sorder,
    )
    _det = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ng_py,),
    )
    _gc = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        ((_ng_py+1)*_nvar_py*(_nvar_py+1)//2,),
    )
    _return_scalars = _b_mv.discrim(
        _n,
        _m,
        _x,
        _isx,
        _nvar,
        _ing,
        _ng,
        _wt,
        _nig,
        _gmn,
        _det,
        _gc,
    )
    _nig_py = _nig
    _gmn_py = _gmn
    _det_py = _det
    _gc_py = _gc
    _stat_py = _return_scalars.stat
    _df_py = _return_scalars.df
    _sig_py = _return_scalars.sig
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['nig', 'gmn', 'det', 'gc', 'stat', 'df', 'sig'],
    )(
        _nig_py,
        _gmn_py,
        _det_py,
        _gc_py,
        _stat_py,
        _df_py,
        _sig_py,
    )
    return _return_data

def discrim_mahal(equal, mode, gmn, gc, nobs, isx, x):
    u"""
    Computes Mahalanobis squared distances for group or pooled
    variance-covariance matrices (for use after ``discrim``).

    ``discrim_mahal`` computes Mahalanobis squared distances for group
    or pooled variance-covariance matrices.
    It is intended for use after ``discrim``.

    For full information please refer to the NAG Library document for
    g03db

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g03/g03dbf.html

    Parameters
    ----------
    equal : str, length 1
        Indicates whether or not the within-group variance-covariance
        matrices are assumed to be equal and the pooled
        variance-covariance matrix used.

        `equal` = 'E'

            The within-group variance-covariance matrices are assumed
            equal and the matrix R stored in the first p(p+1)/2 elements
            of `gc` is used.

        `equal` = 'U'

            The within-group variance-covariance matrices are assumed to
            be unequal and the matrices R_j, for j = 1,2,...,n_g, stored
            in the remainder of `gc` are used.

    mode : str, length 1
        Indicates whether distances from sample points are to be
        calculated or distances between the group means.

        `mode` = 'S'

            The distances between the sample points given in `x` and the
            group means are calculated.

        `mode` = 'M'

            The distances between the group means will be calculated.

    gmn : float, array-like, shape (ng, nvar)
        The jth row of `gmn` contains the means of the p selected
        variables for the jth group, for j = 1,2,...,n_g. These are
        returned by ``discrim``.

    gc : float, array-like, shape ((ng+1)*nvar*(nvar+1)/2)
        The first p(p+1)/2 elements of `gc` should contain the upper
        triangular matrix R and the next n_g blocks of p(p+1)/2 elements
        should contain the upper triangular matrices R_j. All matrices
        must be stored packed by column. These matrices are returned by
        ``discrim``. If `equal` = 'E' only the first p(p+1)/2 elements
        are referenced, if `equal` = 'U' only the elements p(p+1)/2+1 to
        (n_g+1)p(p+1)/2 are referenced.

    nobs : int
        If `mode` = 'S', the number of sample points in `x` for which
        distances are to be calculated.

        If `mode` = 'M', `nobs` is not referenced.

    isx : int, array-like, shape (m)
        If `mode` = 'S', `isx`[l-1] indicates if the lth variable in `x`
        is to be included in the distance calculations. If `isx`[l-1] >
        0 the lth variable is included, for l = 1,2,...,m; otherwise the
        lth variable is not referenced.

        If `mode` = 'M', `isx` is not referenced.

    x : float, array-like, shape (:, m)
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `mode`='S': `nobs`; otherwise: 1.

        If `mode` = 'S' the kth row of `x` must contain x_k. That is
        `x`[k-1,l-1] must contain the kth sample value for the lth
        variable, for l = 1,2,...,m, for k = 1,2,...,`nobs`. Otherwise
        `x` is not referenced.

    Returns
    -------
    d : float, ndarray, shape (:, ng)
        The squared distances.

        If `mode` = 'S', `d`[k-1,j-1] contains the squared distance of
        the kth sample point from the jth group mean, D_kj^2, for j =
        1,2,...,n_g, for k = 1,2,...,`nobs`.

        If `mode` = 'M' and `equal` = 'U', `d`[i-1,j-1] contains the
        squared distance between the ith mean and the jth mean, D_ij^2,
        for j = 1,2,...,i-1,i+1,...,n_g, for i = 1,2,...,n_g.

        The elements `d`[i-1,i-1] are not referenced, for i =
        1,2,...,n_g.

        If `mode` = 'M' and `equal` = 'E', `d`[i-1,j-1] contains the
        squared distance between the ith mean and the jth mean, D_ij^2,
        for j = 1,2,...,i-1, for i = 1,2,...,n_g.

        Since D_{ij} = D_{ji} the elements `d`[i-1,j-1] are not
        referenced, for j = i+1,...,n_g, for i = 1,2,...,n_g.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `mode` = *<value>*.

            Constraint: `mode` = 'M' or 'S'.

        (`errno` 1)
            On entry, `equal` = *<value>*.

            Constraint: `equal` = 'E' or 'U'.

        (`errno` 1)
            On entry, m = *<value>* and nvar = *<value>*.

            Constraint: m >= nvar.

        (`errno` 1)
            On entry, `nobs` = *<value>*.

            Constraint: `nobs` >= 1.

        (`errno` 1)
            On entry, ng = *<value>*.

            Constraint: ng >= 2.

        (`errno` 1)
            On entry, nvar = *<value>*.

            Constraint: nvar >= 1.

        (`errno` 2)
            On entry, nvar = *<value>* and *<value>* values of `isx` >
            0.

            Constraint: exactly nvar elements of `isx` > 0.

        (`errno` 2)
            On entry, diagonal element *<value>* of R_j = 0 for j =
            *<value>*.

        (`errno` 2)
            On entry, diagonal element *<value>* of R = 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Aitchison, J and Dunsmore, I R, 1975, `Statistical Prediction
    Analysis`, Cambridge

    Kendall, M G and Stuart, A, 1976, `The Advanced Theory of Statistics
    (Volume 3)`, (3rd Edition), Griffin

    Krzanowski, W J, 1990, `Principles of Multivariate Analysis`, Oxford
    University Press
    """

    fun_name = 'naginterfaces.library.mv.discrim_mahal'

    _gmn = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'gmn'},
        gmn,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_gmn, 'gmn'), (_x, 'x')),
    )
    _nvar_py = _b_utils._EngineFloat64ArrayType.get_shape(_gmn, {'fun_name': fun_name, 'entity': 'gmn'}, exp_rank=2)[1]
    _nvar = _b_utils._EngineIntScalarType(
        _nvar_py, {'fun_name': fun_name, 'entity': '_nvar'},
    ).to_ctypes()
    _ng_py = _b_utils._EngineFloat64ArrayType.get_shape(_gmn, {'fun_name': fun_name, 'entity': 'gmn'}, exp_rank=2)[0]
    _ng = _b_utils._EngineIntScalarType(
        _ng_py, {'fun_name': fun_name, 'entity': '_ng'},
    ).to_ctypes()
    _gc = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'gc'},
        gc,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _isx = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'isx'},
        isx,
        copy=False,
    )
    _m_py = _b_utils._EngineIntArrayType.get_shape(_isx, {'fun_name': fun_name, 'entity': 'isx'}, exp_rank=1)[0]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _mode = _b_utils._EngineCharScalarType(
        mode, {'fun_name': fun_name, 'entity': 'mode'}, exp_chrlen=1,
    ).to_ctypes()
    _mode_py = _b_utils._EngineCharScalarType.to_py(_mode)
    _nobs = _b_utils._EngineIntScalarType(
        nobs, {'fun_name': fun_name, 'entity': 'nobs'},
    ).to_ctypes()
    _nobs_py = _b_utils._EngineIntScalarType.to_py(_nobs)
    if _mode_py in ('S', 's'):
        _d_dim1 = _nobs_py
    elif _mode_py in ('M', 'm'):
        _d_dim1 = _ng_py
    else:
        _d_dim1 = 0
    exp_shape_d = (_d_dim1, _ng_py)
    _d = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        exp_shape_d,
        sorder=_m_sorder,
    )
    _b_mv.discrim_mahal(
        equal,
        mode,
        _nvar,
        _ng,
        _gmn,
        _gc,
        _nobs,
        _m,
        _isx,
        _x,
        _d,
    )
    _d_py = _d
    return _d_py

def discrim_group(typ, equal, priors, nig, gmn, gc, det, isx, x, prior, atiq):
    u"""
    Allocates observations to groups according to selected rules (for
    use after ``discrim``).

    ``discrim_group`` allocates observations to groups according to
    selected rules.
    It is intended for use after ``discrim``.

    For full information please refer to the NAG Library document for
    g03dc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g03/g03dcf.html

    Parameters
    ----------
    typ : str, length 1
        Whether the estimative or predictive approach is used.

        `typ` = 'E'

            The estimative approach is used.

        `typ` = 'P'

            The predictive approach is used.

    equal : str, length 1
        Indicates whether or not the within-group variance-covariance
        matrices are assumed to be equal and the pooled
        variance-covariance matrix used.

        `equal` = 'E'

            The within-group variance-covariance matrices are assumed
            equal and the matrix R stored in the first p(p+1)/2 elements
            of `gc` is used.

        `equal` = 'U'

            The within-group variance-covariance matrices are assumed to
            be unequal and the matrices R_i, for i = 1,2,...,n_g, stored
            in the remainder of `gc` are used.

    priors : str, length 1
        Indicates the form of the prior probabilities to be used.

        `priors` = 'E'

            Equal prior probabilities are used.

        `priors` = 'P'

            Prior probabilities proportional to the group sizes in the
            training set, n_j, are used.

        `priors` = 'I'

            The prior probabilities are input in `prior`.

    nig : int, array-like, shape (ng)
        The number of observations in each group in the training set,
        n_j.

    gmn : float, array-like, shape (ng, nvar)
        The jth row of `gmn` contains the means of the p variables for
        the jth group, for j = 1,2,...,n_j. These are returned by
        ``discrim``.

    gc : float, array-like, shape ((ng+1)*nvar*(nvar+1)/2)
        The first p(p+1)/2 elements of `gc` should contain the upper
        triangular matrix R and the next n_g blocks of p(p+1)/2 elements
        should contain the upper triangular matrices R_j.

        All matrices must be stored packed by column.

        These matrices are returned by ``discrim``.

        If `equal` = 'E' only the first p(p+1)/2 elements are
        referenced, if `equal` = 'U' only the elements p(p+1)/2+1 to
        (n_g+1)p(p+1)/2 are referenced.

    det : float, array-like, shape (ng)
        If `equal` = 'U'. the logarithms of the determinants of the
        within-group variance-covariance matrices as returned by
        ``discrim``. Otherwise `det` is not referenced.

    isx : int, array-like, shape (m)
        `isx`[l-1] indicates if the lth variable in `x` is to be
        included in the distance calculations.

        If `isx`[l-1] > 0, the lth variable is included, for l =
        1,2,...,m; otherwise the lth variable is not referenced.

    x : float, array-like, shape (nobs, m)
        `x`[k-1,l-1] must contain the kth observation for the lth
        variable, for l = 1,2,...,m, for k = 1,2,...,nobs.

    prior : float, array-like, shape (ng)
        If `priors` = 'I', the prior probabilities for the n_g groups.

    atiq : bool
        `atiq` must be True if atypicality indices are required. If
        `atiq` is False the array `ati` is not set.

    Returns
    -------
    prior : float, ndarray, shape (ng)
        If `priors` = 'P', the computed prior probabilities in
        proportion to group sizes for the n_g groups.

        If `priors` = 'I', the input prior probabilities will be
        unchanged.

        If `priors` = 'E', `prior` is not set.

    p : float, ndarray, shape (nobs, ng)
        `p`[k-1,j-1] contains the posterior probability p_{kj} for
        allocating the kth observation to the jth group, for j =
        1,2,...,n_g, for k = 1,2,...,nobs.

    iag : int, ndarray, shape (nobs)
        The groups to which the observations have been allocated.

    ati : float, ndarray, shape (nobs, :)
        If `atiq` is True, `ati`[k-1,j-1] will contain the predictive
        atypicality index for the kth observation with respect to the
        jth group, for j = 1,2,...,n_g, for k = 1,2,...,nobs.

        If `atiq` is False, `ati` is not set.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `typ` = *<value>*.

            Constraint: `typ` = 'E' or 'P'.

        (`errno` 1)
            On entry, `priors` = *<value>*.

            Constraint: `priors` = 'E', 'I' or 'P'.

        (`errno` 1)
            On entry, `equal` = *<value>*.

            Constraint: `equal` = 'E' or 'U'.

        (`errno` 1)
            On entry, m = *<value>* and nvar = *<value>*.

            Constraint: m >= nvar.

        (`errno` 1)
            On entry, nobs = *<value>*.

            Constraint: nobs >= 1.

        (`errno` 1)
            On entry, ng = *<value>*.

            Constraint: ng >= 2.

        (`errno` 1)
            On entry, nvar = *<value>*.

            Constraint: nvar >= 1.

        (`errno` 2)
            On entry, i = *<value>* and `nig`[i-1] = *<value>*.

            Constraint: `nig`[i-1] > 0.

        (`errno` 2)
            On entry, sum_i(`nig`[i]) <= ng+nvar.

        (`errno` 2)
            On entry, i = *<value>*, `nig`[i-1] = *<value>* and nvar =
            *<value>*.

            Constraint: `nig`[i-1] < nvar.

        (`errno` 2)
            On entry, nvar = *<value>* and *<value>* values of `isx` >
            0.

            Constraint: exactly nvar elements of `isx` > 0.

        (`errno` 3)
            On entry, sum_j(`prior`[j]) != 1.0.

        (`errno` 3)
            On entry, j = *<value>* and `prior`[j-1] = *<value>*.

            Constraint: `prior`[j-1] > 0.

        (`errno` 4)
            On entry, a diagonal element of R or R_j is zero.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Aitchison, J and Dunsmore, I R, 1975, `Statistical Prediction
    Analysis`, Cambridge

    Aitchison, J, Habbema, J D F and Kay, J W, 1977, `A critical
    comparison of two methods of statistical discrimination`, Appl.
    Statist. (26), 15--25

    Kendall, M G and Stuart, A, 1976, `The Advanced Theory of Statistics
    (Volume 3)`, (3rd Edition), Griffin

    Krzanowski, W J, 1990, `Principles of Multivariate Analysis`, Oxford
    University Press

    Moran, M A and Murphy, B J, 1979, `A closer look at two alternative
    methods of statistical discrimination`, Appl. Statist. (28),
    223--232

    Morrison, D F, 1967, `Multivariate Statistical Methods`,
    McGraw--Hill

    See Also
    --------
    :meth:`naginterfaces.library.examples.mv.discrim_group_ex.main`
    """

    fun_name = 'naginterfaces.library.mv.discrim_group'

    _gmn = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'gmn'},
        gmn,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_gmn, 'gmn'), (_x, 'x')),
    )
    _nvar_py = _b_utils._EngineFloat64ArrayType.get_shape(_gmn, {'fun_name': fun_name, 'entity': 'gmn'}, exp_rank=2)[1]
    _nvar = _b_utils._EngineIntScalarType(
        _nvar_py, {'fun_name': fun_name, 'entity': '_nvar'},
    ).to_ctypes()
    _nig = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'nig'},
        nig,
        copy=False,
    )
    _ng_py = _b_utils._EngineIntArrayType.get_shape(_nig, {'fun_name': fun_name, 'entity': 'nig'}, exp_rank=1)[0]
    _ng = _b_utils._EngineIntScalarType(
        _ng_py, {'fun_name': fun_name, 'entity': '_ng'},
    ).to_ctypes()
    _gc = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'gc'},
        gc,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _det = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'det'},
        det,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nobs_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2)[0]
    _nobs = _b_utils._EngineIntScalarType(
        _nobs_py, {'fun_name': fun_name, 'entity': '_nobs'},
    ).to_ctypes()
    _isx = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'isx'},
        isx,
        copy=False,
    )
    _m_py = _b_utils._EngineIntArrayType.get_shape(_isx, {'fun_name': fun_name, 'entity': 'isx'}, exp_rank=1)[0]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _prior = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'prior'},
        prior,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _p = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nobs_py, _ng_py),
        sorder=_m_sorder,
    )
    _iag = _b_utils._EngineIntArrayType.empty_ndarray(
        (_nobs_py,),
    )
    _atiq = _b_utils._EngineLogicalScalarType(
        atiq, {'fun_name': fun_name, 'entity': 'atiq'},
    ).to_ctypes()
    _atiq_py = _b_utils._EngineLogicalScalarType.to_py(_atiq)
    if _atiq_py:
        _ati_dim2 = _ng_py
    else:
        _ati_dim2 = 1
    exp_shape_ati = (_nobs_py, _ati_dim2)
    _ati = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        exp_shape_ati,
        sorder=_m_sorder,
    )
    _b_mv.discrim_group(
        typ,
        equal,
        priors,
        _nvar,
        _ng,
        _nig,
        _gmn,
        _gc,
        _det,
        _nobs,
        _m,
        _isx,
        _x,
        _prior,
        _p,
        _iag,
        atiq,
        _ati,
    )
    _prior_py = _prior
    _p_py = _p
    _iag_py = _iag
    _ati_py = _ati
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['prior', 'p', 'iag', 'ati'],
    )(
        _prior_py,
        _p_py,
        _iag_py,
        _ati_py,
    )
    return _return_data

def distance_mat(update, dist, scal, x, isx, s, d):
    u"""
    Computes distance matrix.

    ``distance_mat`` computes a distance (dissimilarity) matrix.

    For full information please refer to the NAG Library document for
    g03ea

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g03/g03eaf.html

    Parameters
    ----------
    update : str, length 1
        Indicates whether or not an existing matrix is to be updated.

        `update` = 'U'

            The matrix D is updated and distances are added to D.

        `update` = 'I'

            The matrix D is initialized to zero before the distances are
            added to D.

    dist : str, length 1
        Indicates which type of distances are computed.

        `dist` = 'A'

            Absolute distances.

        `dist` = 'E'

            Euclidean distances.

        `dist` = 'S'

            Euclidean squared distances.

    scal : str, length 1
        Indicates the standardization of the variables to be used.

        `scal` = 'S'

            Standard deviation.

        `scal` = 'R'

            Range.

        `scal` = 'G'

            Standardizations given in array `s`.

        `scal` = 'U'

            Unscaled.

    x : float, array-like, shape (n, m)
        `x`[i-1,j-1] must contain the value of the jth variable for the
        ith object, for j = 1,2,...,m, for i = 1,2,...,n.

    isx : int, array-like, shape (m)
        `isx`[j-1] indicates whether or not the jth variable in `x` is
        to be included in the distance computations.

        If `isx`[j-1] > 0 the jth variable is included, for j =
        1,2,...,m; otherwise it is not referenced.

    s : float, array-like, shape (m)
        If `scal` = 'G' and `isx`[j-1] > 0 then `s`[j-1] must contain
        the scaling for variable j, for j = 1,2,...,m.

    d : float, array-like, shape (n*(n-1)/2)
        If `update` = 'U', `d` must contain the strictly lower triangle
        of the distance matrix D to be updated. D must be stored packed
        by rows, i.e., `d`[ (i-1) (i-2) /2+j -1], i > j must contain
        d_{ij}.

        If `update` = 'I', `d` need not be set.

    Returns
    -------
    s : float, ndarray, shape (m)
        If `scal` = 'S' and `isx`[j-1] > 0 then `s`[j-1] contains the
        standard deviation of the variable in the jth column of `x`.

        If `scal` = 'R' and `isx`[j-1] > 0, `s`[j-1] contains the range
        of the variable in the jth column of `x`.

        If `scal` = 'U' and `isx`[j-1] > 0, `s`[j-1] = 1.0.

        If `scal` = 'G', `s` is unchanged.

    d : float, ndarray, shape (n*(n-1)/2)
        The strictly lower triangle of the distance matrix D stored
        packed by rows, i.e., d_{ij} is contained in `d`[ (i-1) (i-2)
        /2+j -1], i > j.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `scal` = *<value>*.

            Constraint: `scal` = 'S', 'R', 'G' or 'U'.

        (`errno` 1)
            On entry, `update` = *<value>*.

            Constraint: `update` = 'U' or 'I'.

        (`errno` 1)
            On entry, `dist` = *<value>*.

            Constraint: `dist` = 'A', 'E' or 'S'

        (`errno` 1)
            On entry, m = *<value>*.

            Constraint: m > 0.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 2.

        (`errno` 2)
            On entry, at least one element of `s` <= 0.0.

        (`errno` 2)
            Variable *<value>* is constant.

        (`errno` 2)
            On entry, at least one element of `d` < 0.0.

        (`errno` 2)
            On entry, `isx` does not contain a positive element.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Everitt, B S, 1974, `Cluster Analysis`, Heinemann

    Krzanowski, W J, 1990, `Principles of Multivariate Analysis`, Oxford
    University Press
    """

    fun_name = 'naginterfaces.library.mv.distance_mat'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2)[1]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _isx = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'isx'},
        isx,
        copy=False,
    )
    _s = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 's'},
        s,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _d = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'd'},
        d,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b_mv.distance_mat(
        update,
        dist,
        scal,
        _n,
        _m,
        _x,
        _isx,
        _s,
        _d,
    )
    _s_py = _s
    _d_py = _d
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['s', 'd'],
    )(
        _s_py,
        _d_py,
    )
    return _return_data

def cluster_hier(method, n, d):
    u"""
    Hierarchical cluster analysis.

    ``cluster_hier`` performs hierarchical cluster analysis.

    For full information please refer to the NAG Library document for
    g03ec

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g03/g03ecf.html

    Parameters
    ----------
    method : int
        Indicates which clustering method is used.

        `method` = 1

            Single link.

        `method` = 2

            Complete link.

        `method` = 3

            Group average.

        `method` = 4

            Centroid.

        `method` = 5

            Median.

        `method` = 6

            Minimum variance.

    n : int
        n, the number of objects.

    d : float, array-like, shape (`n`*(`n`-1)/2)
        The strictly lower triangle of the distance matrix. D must be
        stored packed by rows, i.e., `d`[ (i-1) (i-2) /2+j -1], i > j
        must contain d_{ij}.

    Returns
    -------
    d : float, ndarray, shape (`n`*(`n`-1)/2)
        Is overwritten.

    ilc : int, ndarray, shape (`n`-1)
        `ilc`[l-1] contains the number, j, of the cluster merged with
        cluster k (see `iuc`), j < k, at step l, for l = 1,2,...,n-1.

    iuc : int, ndarray, shape (`n`-1)
        `iuc`[l-1] contains the number, k, of the cluster merged with
        cluster j, j < k, at step l, for l = 1,2,...,n-1.

    cd : float, ndarray, shape (`n`-1)
        `cd`[l-1] contains the distance d_{jk}, between clusters j and
        k, j < k, merged at step l, for l = 1,2,...,n-1.

    iord : int, ndarray, shape (`n`)
        The objects in dendrogram order.

    dord : float, ndarray, shape (`n`)
        The clustering distances corresponding to the order in `iord`.
        `dord`[l-1] contains the distance at which cluster `iord`[l-1]
        and `iord`[l] merge, for l = 1,2,...,n-1. `dord`[n-1] contains
        the maximum distance.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 2.

        (`errno` 1)
            On entry, `method` = *<value>*.

            Constraint: `method` = 1, 2, 3, 4, 5 or 6.

        (`errno` 2)
            On entry, at least one element of `d` is negative.

        (`errno` 3)
            Minimum cluster distance not increasing, dendrogram invalid.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Everitt, B S, 1974, `Cluster Analysis`, Heinemann

    Krzanowski, W J, 1990, `Principles of Multivariate Analysis`, Oxford
    University Press
    """

    fun_name = 'naginterfaces.library.mv.cluster_hier'

    _d = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'd'},
        d,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _ilc = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n_py-1,),
    )
    _iuc = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n_py-1,),
    )
    _cd = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py-1,),
    )
    _iord = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n_py,),
    )
    _dord = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _b_mv.cluster_hier(
        method,
        _n,
        _d,
        _ilc,
        _iuc,
        _cd,
        _iord,
        _dord,
    )
    _d_py = _d
    _ilc_py = _ilc
    _iuc_py = _iuc
    _cd_py = _cd
    _iord_py = _iord
    _dord_py = _dord
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['d', 'ilc', 'iuc', 'cd', 'iord', 'dord'],
    )(
        _d_py,
        _ilc_py,
        _iuc_py,
        _cd_py,
        _iord_py,
        _dord_py,
    )
    return _return_data

def cluster_kmeans(x, isx, cmeans, wt=None, maxit=10):
    u"""
    K-means cluster analysis.

    ``cluster_kmeans`` performs K-means cluster analysis.

    For full information please refer to the NAG Library document for
    g03ef

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g03/g03eff.html

    Parameters
    ----------
    x : float, array-like, shape (n, m)
        `x`[i-1,j-1] must contain the value of the jth variable for the
        ith object, for j = 1,2,...,m, for i = 1,2,...,n.

    isx : int, array-like, shape (m)
        `isx`[j-1] indicates whether or not the jth variable is to be
        included in the analysis. If `isx`[j-1] > 0, the variable
        contained in the jth column of `x` is included, for j =
        1,2,...,m.

    cmeans : float, array-like, shape (k, nvar)
        `cmeans`[i-1,j-1] must contain the value of the jth variable for
        the ith initial cluster centre, for j = 1,2,...,p, for i =
        1,2,...,K.

    wt : None or float, array-like, shape (n), optional
        If `wt` is not None, the first n elements of `wt` must contain
        the weights to be used.

        If `wt`[i-1] = 0.0, the ith observation is not included in the
        analysis.

        The effective number of observation is the sum of the weights.

        If `wt` is None, `wt` is not referenced and the effective number
        of observations is n.

    maxit : int, optional
        The maximum number of iterations allowed in the analysis.

    Returns
    -------
    cmeans : float, ndarray, shape (k, nvar)
        `cmeans`[i-1,j-1] contains the value of the jth variable for the
        ith computed cluster centre, for j = 1,2,...,p, for i =
        1,2,...,K.

    inc : int, ndarray, shape (n)
        `inc`[i-1] contains the cluster to which the ith object has been
        allocated, for i = 1,2,...,n.

    nic : int, ndarray, shape (k)
        `nic`[i-1] contains the number of objects in the ith cluster,
        for i = 1,2,...,K.

    css : float, ndarray, shape (k)
        `css`[i-1] contains the within-cluster (weighted) sum of squares
        of the ith cluster, for i = 1,2,...,K.

    csw : float, ndarray, shape (k)
        `csw`[i-1] contains the within-cluster sum of weights of the ith
        cluster, for i = 1,2,...,K. If weight = 'U', the sum of weights
        is the number of objects in the cluster.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `maxit` = *<value>*.

            Constraint: `maxit` > 0.

        (`errno` 1)
            On entry, k = *<value>*.

            Constraint: k >= 2.

        (`errno` 1)
            On entry, m = *<value>* and nvar = *<value>*.

            Constraint: m >= nvar.

        (`errno` 1)
            On entry, nvar = *<value>*.

            Constraint: nvar >= 1.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n > 1.

        (`errno` 1)
            On entry, weight = *<value>*.

            Constraint: weight = 'U' or 'W'.

        (`errno` 2)
            On entry, `wt` has less than two positive values.

        (`errno` 2)
            On entry, i = *<value>* and `wt`[i-1] < 0.0.

            Constraint: `wt`[i-1] >= 0.0.

        (`errno` 3)
            On entry, nvar = *<value>* and *<value>* values of `isx` >
            0.

            Constraint: exactly nvar elements of `isx` > 0.

        (`errno` 4)
            At least one cluster is empty after the initial assignment.

        (`errno` 5)
            Convergence has not been achieved within the maximum number
            of iterations `maxit` = *<value>*.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Everitt, B S, 1974, `Cluster Analysis`, Heinemann

    Hartigan, J A and Wong, M A, 1979, `Algorithm AS 136: A K-means
    clustering algorithm`, Appl. Statist. (28), 100--108

    Kendall, M G and Stuart, A, 1976, `The Advanced Theory of Statistics
    (Volume 3)`, (3rd Edition), Griffin

    Krzanowski, W J, 1990, `Principles of Multivariate Analysis`, Oxford
    University Press

    See Also
    --------
    :meth:`naginterfaces.library.examples.mv.cluster_kmeans_ex.main`
    """

    fun_name = 'naginterfaces.library.mv.cluster_kmeans'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2)[1]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _isx = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'isx'},
        isx,
        copy=False,
    )
    _cmeans = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'cmeans'},
        cmeans,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nvar_py = _b_utils._EngineFloat64ArrayType.get_shape(_cmeans, {'fun_name': fun_name, 'entity': 'cmeans'}, exp_rank=2)[1]
    _nvar = _b_utils._EngineIntScalarType(
        _nvar_py, {'fun_name': fun_name, 'entity': '_nvar'},
    ).to_ctypes()
    _k_py = _b_utils._EngineFloat64ArrayType.get_shape(_cmeans, {'fun_name': fun_name, 'entity': 'cmeans'}, exp_rank=2)[0]
    _k = _b_utils._EngineIntScalarType(
        _k_py, {'fun_name': fun_name, 'entity': '_k'},
    ).to_ctypes()
    _wt = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'wt'},
        wt,
        copy=False,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _inc = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n_py,),
    )
    _nic = _b_utils._EngineIntArrayType.empty_ndarray(
        (_k_py,),
    )
    _css = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_k_py,),
    )
    _csw = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_k_py,),
    )
    _b_mv.cluster_kmeans(
        _n,
        _m,
        _x,
        _isx,
        _nvar,
        _k,
        _cmeans,
        _wt,
        _inc,
        _nic,
        _css,
        _csw,
        maxit,
    )
    _cmeans_py = _cmeans
    _inc_py = _inc
    _nic_py = _nic
    _css_py = _css
    _csw_py = _csw
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['cmeans', 'inc', 'nic', 'css', 'csw'],
    )(
        _cmeans_py,
        _inc_py,
        _nic_py,
        _css_py,
        _csw_py,
    )
    return _return_data

def cluster_hier_dendrogram(orient, dord, dmin, dstep, nsym):
    u"""
    Constructs dendrogram (for use after ``cluster_hier``).

    ``cluster_hier_dendrogram`` produces a dendrogram from the results
    of ``cluster_hier``.

    For full information please refer to the NAG Library document for
    g03eh

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g03/g03ehf.html

    Parameters
    ----------
    orient : str, length 1
        Indicates which orientation the dendrogram is to take.

        `orient` = 'N'

            The end points of the dendrogram are to the north.

        `orient` = 'S'

            The end points of the dendrogram are to the south.

        `orient` = 'E'

            The end points of the dendrogram are to the east.

        `orient` = 'W'

            The end points of the dendrogram are to the west.

    dord : float, array-like, shape (n)
        The array `dord` as output by ``cluster_hier``. `dord` contains
        the distances, in dendrogram order, at which clustering takes
        place.

    dmin : float
        The clustering distance at which the dendrogram begins.

    dstep : float
        The distance represented by one symbol of the dendrogram.

    nsym : int
        The number of character positions used in the dendrogram. Hence
        the clustering distance at which the dendrogram terminates is
        given by `dmin`+`nsym`*`dstep`.

    Returns
    -------
    c : str, ndarray, shape (:)
        The elements of `c` contain consecutive lines of the dendrogram.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, the number of characters that can be stored in
            each element of `c` is insufficient for the requested
            orientation.

        (`errno` 1)
            On entry, `orient` = *<value>*.

            Constraint: `orient` = 'N', 'S', 'E' or 'W'.

        (`errno` 1)
            On entry, `dstep` = *<value>*.

            Constraint: `dstep` > 0.0.

        (`errno` 1)
            On entry, `dmin` = *<value>*.

            Constraint: `dmin` >= 0.0.

        (`errno` 1)
            On entry, `nsym` = *<value>*.

            Constraint: `nsym` >= 1.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n > 2.

        (`errno` 2)
            On entry, `dord`[i-1] > `dord`[n-1], i < n.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Everitt, B S, 1974, `Cluster Analysis`, Heinemann

    Krzanowski, W J, 1990, `Principles of Multivariate Analysis`, Oxford
    University Press
    """

    fun_name = 'naginterfaces.library.mv.cluster_hier_dendrogram'

    _dord = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'dord'},
        dord,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_dord, {'fun_name': fun_name, 'entity': 'dord'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _orient = _b_utils._EngineCharScalarType(
        orient, {'fun_name': fun_name, 'entity': 'orient'}, exp_chrlen=1,
    ).to_ctypes()
    _orient_py = _b_utils._EngineCharScalarType.to_py(_orient)
    _nsym = _b_utils._EngineIntScalarType(
        nsym, {'fun_name': fun_name, 'entity': 'nsym'},
    ).to_ctypes()
    _nsym_py = _b_utils._EngineIntScalarType.to_py(_nsym)
    if _orient_py in ('N', 'n', 'S', 's'):
        _lenc = _b_utils._EngineIntScalarType(
            _nsym_py, {'fun_name': fun_name, 'entity': '_lenc'},
        ).to_ctypes()
        _lenc_py = _b_utils._EngineIntScalarType.to_py(_lenc)
    elif _orient_py in ('E', 'e', 'W', 'w'):
        _lenc = _b_utils._EngineIntScalarType(
            _n_py, {'fun_name': fun_name, 'entity': '_lenc'},
        ).to_ctypes()
        _lenc_py = _b_utils._EngineIntScalarType.to_py(_lenc)
    else:
        _lenc = 0
        _lenc_py = _lenc
    if _orient_py in ('N', 'n', 'S', 's'):
        _vl_c_len_py = 3*_n_py
    elif _orient_py in ('E', 'e', 'W', 'w'):
        _vl_c_len_py = _nsym_py
    else:
        _vl_c_len_py = 0
    if _orient_py in ('N', 'n', 'S', 's'):
        _c_dim1 = _nsym_py
    elif _orient_py in ('E', 'e', 'W', 'w'):
        _c_dim1 = _n_py
    else:
        _c_dim1 = 0
    exp_shape_c = (_c_dim1,)
    _c = _b_utils._EngineCharArrayType.empty_ndarray(
        exp_shape_c,
        chrlen=_vl_c_len_py,
    )
    _b_mv.cluster_hier_dendrogram(
        orient,
        _n,
        _dord,
        dmin,
        dstep,
        _nsym,
        _c,
        _lenc,
    )
    _c_py = _c
    return _c_py

def cluster_hier_indicator(cd, iord, dord, k, dlevel):
    u"""
    Computes cluster indicator variable (for use after
    ``cluster_hier``).

    ``cluster_hier_indicator`` computes a cluster indicator variable
    from the results of ``cluster_hier``.

    For full information please refer to the NAG Library document for
    g03ej

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g03/g03ejf.html

    Parameters
    ----------
    cd : float, array-like, shape (n-1)
        The clustering distances in increasing order as returned by
        ``cluster_hier``.

    iord : int, array-like, shape (n)
        The objects in dendrogram order as returned by ``cluster_hier``.

    dord : float, array-like, shape (n)
        The clustering distances corresponding to the order in `iord`.

    k : int
        Indicates if a specified number of clusters is required.

        If `k` > 0 then ``cluster_hier_indicator`` will attempt to find
        `k` clusters.

        If `k` <= 0 then ``cluster_hier_indicator`` will find the
        clusters based on the distance given in `dlevel`.

    dlevel : float
        If `k` <= 0, `dlevel` must contain the distance at which
        clusters are produced. Otherwise `dlevel` need not be set.

    Returns
    -------
    k : int
        The number of clusters produced, k.

    dlevel : float
        If `k` > 0 on entry, `dlevel` contains the distance at which the
        required number of clusters are found. Otherwise `dlevel`
        remains unchanged.

    ic : int, ndarray, shape (n)
        `ic`[i-1] indicates to which of k clusters the ith object
        belongs, for i = 1,2,...,n.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 2.

        (`errno` 1)
            On entry, `k` < 0 and `dlevel` <= 0.0.

        (`errno` 1)
            On entry, `k` = *<value>* and n = *<value>*.

            Constraint: `k` <= n.

        (`errno` 2)
            On entry the values of `dord` and `cd` are not compatible.

        (`errno` 2)
            On entry the values of `cd` are not in increasing order.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            No clustering is performed when `k` = n.

        (`errno` 3)
            All data is merged when `k` = 1.

        (`errno` 3)
            All data merged into one cluster at `dlevel`, `dlevel` =
            *<value>*.

        (`errno` 3)
            No clustering takes place below `dlevel`, `dlevel` =
            *<value>*.

        (`errno` 4)
            The precise number of clusters requested is not possible
            because of tied clustering distances.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Everitt, B S, 1974, `Cluster Analysis`, Heinemann

    Krzanowski, W J, 1990, `Principles of Multivariate Analysis`, Oxford
    University Press
    """

    fun_name = 'naginterfaces.library.mv.cluster_hier_indicator'

    _cd = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'cd'},
        cd,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n = _b_utils._EngineIntScalarType(
        _b_utils._EngineFloat64ArrayType.get_shape(_cd, {'fun_name': fun_name, 'entity': 'cd'}, exp_rank=1)[0]+1, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _iord = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'iord'},
        iord,
        copy=False,
    )
    _dord = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'dord'},
        dord,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _ic = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _return_scalars = _b_mv.cluster_hier_indicator(
            _n,
            _cd,
            _iord,
            _dord,
            k,
            dlevel,
            _ic,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _k_py = _return_scalars.k
    _dlevel_py = _return_scalars.dlevel
    _ic_py = _ic
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['k', 'dlevel', 'ic'],
    )(
        _k_py,
        _dlevel_py,
        _ic_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def multidimscal_metric(roots, n, d, ndim):
    u"""
    Performs principal coordinate analysis, classical metric scaling.

    ``multidimscal_metric`` performs a principal coordinate analysis
    also known as classical metric scaling.

    For full information please refer to the NAG Library document for
    g03fa

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g03/g03faf.html

    Parameters
    ----------
    roots : str, length 1
        Indicates if all the eigenvalues are to be computed or just the
        `ndim` largest.

        `roots` = 'A'

            All the eigenvalues are computed.

        `roots` = 'L'

            Only the largest `ndim` eigenvalues are computed.

    n : int
        n, the number of objects in the distance matrix.

    d : float, array-like, shape (`n`*(`n`-1)/2)
        The lower triangle of the distance matrix D stored packed by
        rows. That is `d`[ (i-1) * (i-2) /2+j -1] must contain d_{ij}
        for i = 2,3,...,n;j = 1,2,...,i-1.

    ndim : int
        k, the number of dimensions used to represent the data.

    Returns
    -------
    x : float, ndarray, shape (`n`, `ndim`)
        The ith row contains k coordinates for the ith point, i =
        1,2,...,n.

    eigval : float, ndarray, shape (`n`)
        If `roots` = 'A', `eigval` contains the n scaled eigenvalues of
        the matrix E.

        If `roots` = 'L', `eigval` contains the largest k scaled
        eigenvalues of the matrix E.

        In both cases the eigenvalues are divided by the sum of the
        eigenvalues (that is, the trace of E).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `roots` = *<value>*.

            Constraint: `roots` = 'A' or 'L'.

        (`errno` 1)
            On entry, `n` = *<value>* and `ndim` = *<value>*.

            Constraint: `n` > `ndim`.

        (`errno` 1)
            On entry, `ndim` = *<value>*.

            Constraint: `ndim` >= 1.

        (`errno` 2)
            On entry, all the elements of `d` = 0.0.

        (`errno` 2)
            On entry, i = *<value>* and `d`[i-1] < 0.0.

            Constraint: `d`[i-1] >= 0.0.

        (`errno` 3)
            There are less than `ndim` eigenvalues greater than zero.

        (`errno` 4)
            The computation of the eigenvalues or eigenvectors has
            failed.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Chatfield, C and Collins, A J, 1980, `Introduction to Multivariate
    Analysis`, Chapman and Hall

    Gower, J C, 1966, `Some distance properties of latent root and
    vector methods used in multivariate analysis`, Biometrika (53),
    325--338

    Krzanowski, W J, 1990, `Principles of Multivariate Analysis`, Oxford
    University Press
    """

    fun_name = 'naginterfaces.library.mv.multidimscal_metric'

    _d = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'd'},
        d,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _ndim = _b_utils._EngineIntScalarType(
        ndim, {'fun_name': fun_name, 'entity': 'ndim'},
    ).to_ctypes()
    _ndim_py = _b_utils._EngineIntScalarType.to_py(_ndim)
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py, _ndim_py),
        sorder=_b_utils._EngineData.default_sorder,
    )
    _eigval = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _b_mv.multidimscal_metric(
        roots,
        _n,
        _d,
        _ndim,
        _x,
        _eigval,
    )
    _x_py = _x
    _eigval_py = _eigval
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'eigval'],
    )(
        _x_py,
        _eigval_py,
    )
    return _return_data

def multidimscal_ordinal(typ, d, x, itera, iopt, io_manager=None):
    u"""
    Performs non-metric (ordinal) multidimensional scaling.

    ``multidimscal_ordinal`` performs non-metric (ordinal)
    multidimensional scaling.

    For full information please refer to the NAG Library document for
    g03fc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g03/g03fcf.html

    Parameters
    ----------
    typ : str, length 1
        Indicates whether STRESS or SSTRESS is to be used as the
        criterion.

        `typ` = 'T'

            STRESS is used.

        `typ` = 'S'

            SSTRESS is used.

    d : float, array-like, shape (n*(n-1)/2)
        The lower triangle of the distance matrix D stored packed by
        rows. That is `d`[ (i-1) * (i-2) /2+j -1] must contain d_{ij},
        for j = 1,2,...,i-1, for i = 2,3,...,n. If d_{ij} is missing
        then set d_{ij} < 0; for further comments on missing values see
        [Further Comments].

    x : float, array-like, shape (n, ndim)
        The ith row must contain an initial estimate of the coordinates
        for the ith point, for i = 1,2,...,n. One method of computing
        these is to use ``multidimscal_metric``.

    itera : int
        The maximum number of iterations in the optimization process.

        `itera` = 0

            A default value of 50 is used.

        `itera` < 0

            A default value of max(50, 5nm) (the default for
            ``opt.uncon_conjgrd_comp``) is used.

    iopt : int
        Selects the options, other than the number of iterations, that
        control the optimization.

        `iopt` = 0

            The tolerance epsilon is set to 0.00001 ([Accuracy]). All
            other values are set as described in [Further Comments].

        `iopt` > 0

            The tolerance epsilon is set to 10^(-i) where i = `iopt`.
            All other values are set as described in [Further Comments].

        `iopt` < 0

            No values are changed, therefore the default values of
            ``opt.uncon_conjgrd_comp`` are used.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    x : float, ndarray, shape (n, ndim)
        The ith row contains m coordinates for the ith point, for i =
        1,2,...,n.

    stress : float
        The value of STRESS or SSTRESS at the final iteration.

    dfit : float, ndarray, shape (2*n*(n-1))
        Auxiliary outputs.

        If `typ` = 'T', the first n(n-1)/2 elements contain the
        distances, d_{ij}^, for the points returned in `x`, the second
        set of n(n-1)/2 contains the distances d_{ij}^ ordered by the
        input distances, d_{ij}, the third set of n(n-1)/2 elements
        contains the monotonic distances, d_{ij}~, ordered by the input
        distances, d_{ij} and the final set of n(n-1)/2 elements
        contains fitted monotonic distances, d_{ij}~, for the points in
        `x`.

        The d_{ij}~ corresponding to distances which are input as
        missing are set to zero.

        If `typ` = 'S', the results are as above except that the squared
        distances are returned.

        Each distance matrix is stored in lower triangular packed form
        in the same way as the input matrix D.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `typ` = *<value>*.

            Constraint: `typ` = 'S' or 'T'.

        (`errno` 1)
            On entry, n = *<value>* and ndim = *<value>*.

            Constraint: n > ndim.

        (`errno` 1)
            On entry, ndim = *<value>*.

            Constraint: ndim >= 1.

        (`errno` 2)
            On entry, all the elements of `d` <= 0.0.

        (`errno` 3)
            The optimization has failed to converge in `itera` function
            iterations.

        (`errno` 5)
            The optimization cannot begin from the initial
            configuration.

        (`errno` 6)
            The optimization has failed.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 4)
            The conditions for an acceptable solution have not been met
            but a lower point could not be found.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Chatfield, C and Collins, A J, 1980, `Introduction to Multivariate
    Analysis`, Chapman and Hall

    Krzanowski, W J, 1990, `Principles of Multivariate Analysis`, Oxford
    University Press
    """

    fun_name = 'naginterfaces.library.mv.multidimscal_ordinal'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _ndim_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2)[1]
    _ndim = _b_utils._EngineIntScalarType(
        _ndim_py, {'fun_name': fun_name, 'entity': '_ndim'},
    ).to_ctypes()
    _d = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'd'},
        d,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _dfit = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (2*_n_py*(_n_py-1),),
    )
    try:
        _return_scalar = _b_mv.multidimscal_ordinal(
            typ,
            _n,
            _ndim,
            _d,
            _x,
            _dfit,
            itera,
            iopt,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _x_py = _x
    _stress_py = _return_scalar
    _dfit_py = _dfit
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'stress', 'dfit'],
    )(
        _x_py,
        _stress_py,
        _dfit_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def gaussian_mixture(x, nvar, ng, sopt, isx=None, prob=None, niter=15, riter=0, tol=0.0):
    u"""
    Fits a Gaussian mixture model.

    ``gaussian_mixture`` performs a mixture of Normals (Gaussians) for a
    given (co)variance structure.

    For full information please refer to the NAG Library document for
    g03ga

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g03/g03gaf.html

    Parameters
    ----------
    x : float, array-like, shape (n, m)
        `x`[i-1,j-1] must contain the value of the jth variable for the
        ith object, for j = 1,2,...,m, for i = 1,2,...,n.

    nvar : int
        p, the number of variables included in the calculations.

    ng : int
        k, the number of groups in the mixture model.

    sopt : int
        Determines the (co)variance structure:

        `sopt` = 1

            Groupwise covariance matrices.

        `sopt` = 2

            Pooled covariance matrix.

        `sopt` = 3

            Groupwise variances.

        `sopt` = 4

            Pooled variances.

        `sopt` = 5

            Overall variance.

    isx : None or int, array-like, shape (:), optional
        Note: the required length for this argument is determined as
        follows: if `nvar`<m: m; otherwise: 0.

        If `nvar` = m all available variables are included in the model
        and `isx` is not referenced; otherwise the jth variable will be
        included in the analysis if `isx`[j-1] = 1 and excluded if
        `isx`[j-1] = 0, for j = 1,2,...,m.

    prob : None or float, array-like, shape (n, `ng`), optional
        If `prob` is not None, `prob`[i-1,j-1] is the probability that
        the ith object belongs to the jth group. (These probabilities
        are normalised internally.) Otherwise, the initial membership
        probabilities are set internally.

    niter : int, optional
        The maximum number of iterations.

    riter : int, optional
        If `riter` > 0, membership probabilities are rounded to 0.0 or
        1.0 after the completion of the first `riter` iterations.

    tol : float, optional
        Iterations cease the first time an improvement in log-likelihood
        is less than `tol`. If `tol` <= 0 a value of 10^(-3) is used.

    Returns
    -------
    prob : float, ndarray, shape (n, `ng`)
        `prob`[i-1,j-1] is the probability of membership of the ith
        object to the jth group for the fitted model.

    niter : int
        The number of completed iterations.

    w : float, ndarray, shape (`ng`)
        w_j, the mixing probability for the jth group.

    g : float, ndarray, shape (`nvar`, `ng`)
        `g`[i-1,j-1] gives the estimated mean of the ith variable in the
        jth group.

    s : float, ndarray, shape (:, :, :)
        If `sopt` = 1, `s`[i-1,j-1,k-1] gives the (i,j)th element of the
        kth group.

        If `sopt` = 2, `s`[i-1,j-1,0] gives the (i,j)th element of the
        pooled covariance.

        If `sopt` = 3, `s`[j-1,k-1,0] gives the jth variance in the kth
        group.

        If `sopt` = 4, `s`[j-1,0,0] gives the jth pooled variance.

        If `sopt` = 5, `s`[0,0,0] gives the overall variance.

    f : float, ndarray, shape (n, `ng`)
        `f`[i-1,j-1] gives the p-variate Normal (Gaussian) density of
        the ith object in the jth group.

    loglik : float
        The log-likelihood for the fitted mixture model.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, n = *<value>* and p = *<value>*.

            Constraint: n > p, the number of parameters, i.e., too few
            objects have been supplied for the model.

        (`errno` 2)
            On entry, m = *<value>*.

            Constraint: m >= 1.

        (`errno` 5)
            On entry, `nvar` = *<value>* and m = *<value>*.

            Constraint: 1 <= `nvar` <= m.

        (`errno` 6)
            On entry, `nvar` != m and `isx` is invalid.

        (`errno` 7)
            On entry, `ng` = *<value>*.

            Constraint: `ng` >= 1.

        (`errno` 9)
            On entry, row *<value>* of supplied `prob` does not sum to
            1.

        (`errno` 11)
            On entry, `niter` = *<value>*.

            Constraint: `niter` >= 1.

        (`errno` 16)
            On entry, `sopt` != 1, 2, 3, 4 or 5.

        (`errno` 44)
            A covariance matrix is not positive definite, try a
            different initial allocation.

        (`errno` 45)
            An iteration cannot continue due to an empty group, try a
            different initial allocation.

    References
    ----------
    Hartigan, J A, 1975, `Clustering Algorithms`, Wiley

    See Also
    --------
    :meth:`naginterfaces.library.examples.mv.gaussian_mixture_ex.main`
    """

    fun_name = 'naginterfaces.library.mv.gaussian_mixture'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _prob = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'prob'},
        prob,
        copy=True,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_x, 'x'), (_prob, 'prob')),
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2)[1]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _isx = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'isx'},
        isx,
        copy=False,
        allow_none=True,
    )
    if prob is None:
        _popt = 1
        _popt_py = _popt
    else:
        _popt = 2
        _popt_py = _popt
    _ng = _b_utils._EngineIntScalarType(
        ng, {'fun_name': fun_name, 'entity': 'ng'},
    ).to_ctypes()
    _ng_py = _b_utils._EngineIntScalarType.to_py(_ng)
    if prob is None:
        _prob = _b_utils._EngineFloat64ArrayType.empty_ndarray(
            (_n_py, _ng_py),
            sorder=_m_sorder,
        )
    _w = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ng_py,),
    )
    _nvar = _b_utils._EngineIntScalarType(
        nvar, {'fun_name': fun_name, 'entity': 'nvar'},
    ).to_ctypes()
    _nvar_py = _b_utils._EngineIntScalarType.to_py(_nvar)
    _g = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nvar_py, _ng_py),
        sorder=_m_sorder,
    )
    _sopt = _b_utils._EngineIntScalarType(
        sopt, {'fun_name': fun_name, 'entity': 'sopt'},
    ).to_ctypes()
    _sopt_py = _b_utils._EngineIntScalarType.to_py(_sopt)
    if _sopt_py == 5:
        _s_dim1 = 1
    else:
        _s_dim1 = _nvar_py
    if _sopt_py in (1, 2):
        _s_dim2 = _nvar_py
    elif _sopt_py == 3:
        _s_dim2 = _ng_py
    elif _sopt_py in (4, 5):
        _s_dim2 = 1
    else:
        _s_dim2 = 0
    if _sopt_py == 1:
        _s_dim3 = _ng_py
    else:
        _s_dim3 = 1
    exp_shape_s = (_s_dim1, _s_dim2, _s_dim3)
    _s = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        exp_shape_s,
        sorder=_m_sorder,
    )
    _f = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py, _ng_py),
        sorder=_m_sorder,
    )
    _return_scalars = _b_mv.gaussian_mixture(
        _n,
        _m,
        _x,
        _isx,
        _nvar,
        _ng,
        _popt,
        _prob,
        niter,
        riter,
        _w,
        _g,
        _sopt,
        _s,
        _f,
        tol,
    )
    _prob_py = _prob
    _niter_py = _return_scalars.niter
    _w_py = _w
    _g_py = _g
    _s_py = _s
    _f_py = _f
    _loglik_py = _return_scalars.loglik
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['prob', 'niter', 'w', 'g', 's', 'f', 'loglik'],
    )(
        _prob_py,
        _niter_py,
        _w_py,
        _g_py,
        _s_py,
        _f_py,
        _loglik_py,
    )
    return _return_data

def z_scores(x, nvar, isx, s, e):
    u"""
    Produces standardized values (z-scores) for a data matrix.

    ``z_scores`` produces standardized values (z-scores) for a data
    matrix.

    For full information please refer to the NAG Library document for
    g03za

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g03/g03zaf.html

    Parameters
    ----------
    x : float, array-like, shape (n, m)
        `x`[i-1,j-1] must contain the ith sample point for the jth
        variable, x_{ij}, for j = 1,2,...,m, for i = 1,2,...,n.

    nvar : int
        p, the number of variables to be standardized.

    isx : int, array-like, shape (m)
        `isx`[j-1] indicates whether or not the observations on the jth
        variable are included in the matrix of standardized values.

        If `isx`[j-1] != 0, the observations from the jth variable are
        included.

        If `isx`[j-1] = 0, the observations from the jth variable are
        not included.

    s : float, array-like, shape (m)
        If `isx`[j-1] != 0, `s`[j-1] must contain the scaling (standard
        deviation), sigma_j, for the jth variable.

        If `isx`[j-1] = 0, `s`[j-1] is not referenced.

    e : float, array-like, shape (m)
        If `isx`[j-1] != 0, `e`[j-1] must contain the location shift
        (mean), mu_j, for the jth variable.

        If `isx`[j-1] = 0, `e`[j-1] is not referenced.

    Returns
    -------
    z : float, ndarray, shape (n, `nvar`)
        The matrix of standardized values (z-scores), Z.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, m = *<value>* and `nvar` = *<value>*.

            Constraint: m >= `nvar`.

        (`errno` 1)
            On entry, `nvar` = *<value>*.

            Constraint: `nvar` >= 1.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 2)
            On entry, `nvar` = *<value>* and *<value>* values of `isx` >
            0.

            Constraint: exactly `nvar` elements of `isx` > 0.

        (`errno` 3)
            On entry, j = *<value>* and `s`[i-1] <= 0.0.

            Constraint: `s`[j] > 0.0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    fun_name = 'naginterfaces.library.mv.z_scores'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_x, 'x'),),
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2)[1]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _isx = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'isx'},
        isx,
        copy=False,
    )
    _s = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 's'},
        s,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _e = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'e'},
        e,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nvar = _b_utils._EngineIntScalarType(
        nvar, {'fun_name': fun_name, 'entity': 'nvar'},
    ).to_ctypes()
    _nvar_py = _b_utils._EngineIntScalarType.to_py(_nvar)
    _z = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py, _nvar_py),
        sorder=_m_sorder,
    )
    _b_mv.z_scores(
        _n,
        _m,
        _x,
        _nvar,
        _isx,
        _s,
        _e,
        _z,
    )
    _z_py = _z
    return _z_py
