# -*- coding: utf-8 -*-
# pylint: disable=chained-comparison,invalid-name,line-too-long,protected-access,too-many-arguments,too-many-boolean-expressions,too-many-branches,too-many-lines,too-many-locals,too-many-statements,unused-import
u"""
Interfaces for the NAG Mark 27.0 `nonpar` Chapter.

``nonpar`` - Nonparametric Statistics

The functions in this module perform nonparametric statistical tests
which are based on distribution-free methods of analysis.
For convenience, the module contents are divided into five types of
test: tests of location, tests of dispersion, tests of distribution,
tests of association and correlation, and tests of randomness.
There are also functions to fit linear regression models using the ranks
of the observations.

The emphasis in this module is on testing; if you wish to compute
nonparametric correlations you are referred to submodule ``correg``,
which contains several functions for that purpose.

There are a large number of nonparametric tests available.
A selection of some of the more commonly used tests are included in this
module.

For full information please refer to the NAG Library document

https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g08/g08intro.html
"""

# NAG Copyright 2017-2020.

import collections as _collections

from ..base import nonpar as _b_nonpar
from ..base import utils as _b_utils
from ..library import _utils

def test_sign(x, y):
    u"""
    Sign test on two paired samples.

    ``test_sign`` performs the Sign test on two related samples of size
    n.

    For full information please refer to the NAG Library document for
    g08aa

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g08/g08aaf.html

    Parameters
    ----------
    x : float, array-like, shape (n)
        `x`[i-1] and `y`[i-1] must be set to the ith pair of data
        values, {x_i,y_i}, for i = 1,2,...,n.

    y : float, array-like, shape (n)
        `x`[i-1] and `y`[i-1] must be set to the ith pair of data
        values, {x_i,y_i}, for i = 1,2,...,n.

    Returns
    -------
    isgn : int
        The Sign test statistic, S.

    n1 : int
        The number of non-tied pairs, n_1.

    p : float
        The lower tail probability, p, corresponding to S.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 2)
            On entry, the samples are identical, i.e., n_1 = 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Siegel, S, 1956, `Non-parametric Statistics for the Behavioral
    Sciences`, McGraw--Hill
    """

    fun_name = 'naginterfaces.library.nonpar.test_sign'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _return_scalars = _b_nonpar.test_sign(
        _x,
        _y,
        _n,
    )
    _isgn_py = _return_scalars.isgn
    _n1_py = _return_scalars.n1
    _p_py = _return_scalars.p
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['isgn', 'n1', 'p'],
    )(
        _isgn_py,
        _n1_py,
        _p_py,
    )
    return _return_data

def test_median(x, n1):
    u"""
    Median test on two samples of unequal size.

    ``test_median`` performs the Median test on two independent samples
    of possibly unequal size.

    For full information please refer to the NAG Library document for
    g08ac

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g08/g08acf.html

    Parameters
    ----------
    x : float, array-like, shape (n)
        The first n_1 elements of `x` must be set to the data values in
        the first sample, and the next n_2 (= n-n_1) elements to the
        data values in the second sample.

    n1 : int
        The size of the first sample n_1.

    Returns
    -------
    i1 : int
        The number of scores in the first sample which lie below the
        pooled median, i_1.

    i2 : int
        The number of scores in the second sample which lie below the
        pooled median, i_2.

    p : float
        The tail probability p corresponding to the observed dichotomy
        of the two samples.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 2.

        (`errno` 2)
            On entry, `n1` = *<value>* and n = *<value>*.

            Constraint: 1 <= `n1` < n.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Siegel, S, 1956, `Non-parametric Statistics for the Behavioral
    Sciences`, McGraw--Hill
    """

    fun_name = 'naginterfaces.library.nonpar.test_median'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _return_scalars = _b_nonpar.test_median(
        _x,
        _n,
        n1,
    )
    _i1_py = _return_scalars.i1
    _i2_py = _return_scalars.i2
    _p_py = _return_scalars.p
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['i1', 'i2', 'p'],
    )(
        _i1_py,
        _i2_py,
        _p_py,
    )
    return _return_data

def test_friedman(x):
    u"""
    Friedman two-way analysis of variance on k matched samples.

    ``test_friedman`` performs the Friedman two-way analysis of variance
    by ranks on k related samples of size n.

    For full information please refer to the NAG Library document for
    g08ae

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g08/g08aef.html

    Parameters
    ----------
    x : float, array-like, shape (k, n)
        `x`[i-1,j-1] must be set to the value, x_{ij}, of observation j
        in sample i, for j = 1,2,...,n, for i = 1,2,...,k.

    Returns
    -------
    fr : float
        The value of the Friedman test statistic, F.

    p : float
        The approximate significance, p, of the Friedman test statistic.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 3)
            On entry, k = *<value>*.

            Constraint: k >= 2.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Siegel, S, 1956, `Non-parametric Statistics for the Behavioral
    Sciences`, McGraw--Hill
    """

    fun_name = 'naginterfaces.library.nonpar.test_friedman'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _k_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2)[0]
    _k = _b_utils._EngineIntScalarType(
        _k_py, {'fun_name': fun_name, 'entity': '_k'},
    ).to_ctypes()
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _return_scalars = _b_nonpar.test_friedman(
        _x,
        _k,
        _n,
    )
    _fr_py = _return_scalars.fr
    _p_py = _return_scalars.p
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['fr', 'p'],
    )(
        _fr_py,
        _p_py,
    )
    return _return_data

def test_kruskal(x, l):
    u"""
    Kruskal--Wallis one-way analysis of variance on k samples of unequal
    size.

    ``test_kruskal`` performs the Kruskal--Wallis one-way analysis of
    variance by ranks on k independent samples of possibly unequal
    sizes.

    For full information please refer to the NAG Library document for
    g08af

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g08/g08aff.html

    Parameters
    ----------
    x : float, array-like, shape (lx)
        The elements of `x` must contain the observations in the k
        samples. The first l_1 elements must contain the scores in the
        first sample, the next l_2 those in the second sample, and so
        on.

    l : int, array-like, shape (k)
        `l`[i] must contain the number of observations l_i in sample i,
        for i = 1,2,...,k.

    Returns
    -------
    h : float
        The value of the Kruskal--Wallis test statistic, H.

    p : float
        The approximate significance, p, of the Kruskal--Wallis test
        statistic.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, k = *<value>*.

            Constraint: k >= 2.

        (`errno` 2)
            On entry, i = *<value>* and `l`[i-1] = *<value>*.

            Constraint: `l`[i-1] > 0.

        (`errno` 3)
            On entry, sum_i(`l`[i-1]) = *<value>* and lx = *<value>*.

            Constraint: sum_i(`l`[i-1]) = lx.

        (`errno` 4)
            On entry, all the observations were equal.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Moore, P G, Shirley, E A and Edwards, D E, 1972, `Standard
    Statistical Calculations`, Pitman

    Siegel, S, 1956, `Non-parametric Statistics for the Behavioral
    Sciences`, McGraw--Hill
    """

    fun_name = 'naginterfaces.library.nonpar.test_kruskal'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lx_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _lx = _b_utils._EngineIntScalarType(
        _lx_py, {'fun_name': fun_name, 'entity': '_lx'},
    ).to_ctypes()
    _l = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'l'},
        l,
        copy=False,
    )
    _k_py = _b_utils._EngineIntArrayType.get_shape(_l, {'fun_name': fun_name, 'entity': 'l'}, exp_rank=1)[0]
    _k = _b_utils._EngineIntScalarType(
        _k_py, {'fun_name': fun_name, 'entity': '_k'},
    ).to_ctypes()
    _return_scalars = _b_nonpar.test_kruskal(
        _x,
        _lx,
        _l,
        _k,
    )
    _h_py = _return_scalars.h
    _p_py = _return_scalars.p
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['h', 'p'],
    )(
        _h_py,
        _p_py,
    )
    return _return_data

def test_wilcoxon(x, xme, tail, zer):
    u"""
    Performs the Wilcoxon one-sample (matched pairs) signed rank test.

    ``test_wilcoxon`` performs the Wilcoxon signed rank test on a single
    sample of size n.

    For full information please refer to the NAG Library document for
    g08ag

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g08/g08agf.html

    Parameters
    ----------
    x : float, array-like, shape (n)
        The sample observations, x_1,x_2,...,x_n.

    xme : float
        The median test value, X_med.

    tail : str, length 1
        Indicates the choice of tail probability, and hence the
        alternative hypothesis.

        `tail` = 'T'

            A two tailed probability is calculated and the alternative
            hypothesis is H_1: population median != X_med.

        `tail` = 'U'

            An upper tailed probability is calculated and the
            alternative hypothesis is H_1: population median > X_med.

        `tail` = 'L'

            A lower tailed probability is calculated and the alternative
            hypothesis is H_1: population median < X_med.

    zer : str, length 1
        Indicates whether or not to include the cases where d_i = 0.0 in
        the ranking of the d_is.

        `zer` = 'Y'

            All d_i = 0.0 are included when ranking.

        `zer` = 'N'

            All d_i = 0.0, are ignored, that is all cases where d_i =
            0.0 are removed before ranking.

    Returns
    -------
    w : float
        The Wilcoxon rank sum statistic, W, being the sum of the
        positive ranks.

    wnor : float
        The approximate Normal test statistic, z, as described in
        [Description].

    p : float
        The tail probability, p, as specified by the argument `tail`.

    n1 : int
        The number of nonzero d_i's, n_1.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `zer` = *<value>*.

            Constraint: `zer` = 'Y' or 'N'.

        (`errno` 1)
            On entry, `tail` = *<value>*.

            Constraint: `tail` = 'T', 'U' or 'L'.

        (`errno` 2)
            On entry, n = *<value>*.

            Constraint: n >= 1.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            All elements of the sample are equal to `xme`, i.e.,
            variance = 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Conover, W J, 1980, `Practical Nonparametric Statistics`, Wiley

    Neumann, N, 1988, `Some procedures for calculating the distributions
    of elementary nonparametric teststatistics`, Statistical Software
    Newsletter (14(3)), 120--126

    Siegel, S, 1956, `Non-parametric Statistics for the Behavioral
    Sciences`, McGraw--Hill
    """

    fun_name = 'naginterfaces.library.nonpar.test_wilcoxon'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    try:
        _return_scalars = _b_nonpar.test_wilcoxon(
            _n,
            _x,
            xme,
            tail,
            zer,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _w_py = _return_scalars.w
    _wnor_py = _return_scalars.wnor
    _p_py = _return_scalars.p
    _n1_py = _return_scalars.n1
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['w', 'wnor', 'p', 'n1'],
    )(
        _w_py,
        _wnor_py,
        _p_py,
        _n1_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def test_mwu(x, y, tail):
    u"""
    Performs the Mann--Whitney U test on two independent samples.

    ``test_mwu`` performs the Mann--Whitney U test on two independent
    samples of possibly unequal size.

    For full information please refer to the NAG Library document for
    g08ah

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g08/g08ahf.html

    Parameters
    ----------
    x : float, array-like, shape (n1)
        The first vector of observations, x_1,x_2,...,x_n_1.

    y : float, array-like, shape (n2)
        The second vector of observations. y_1,y_2,...,y_n_2.

    tail : str, length 1
        Indicates the choice of tail probability, and hence the
        alternative hypothesis.

        `tail` = 'T'

            A two tailed probability is calculated and the alternative
            hypothesis is H_1:F(x) != G(y).

        `tail` = 'U'

            An upper tailed probability is calculated and the
            alternative hypothesis H_1:F(x) < G(y), i.e., the x's tend
            to be greater than the y's.

        `tail` = 'L'

            A lower tailed probability is calculated and the alternative
            hypothesis H_1:F(x) > G(y), i.e., the x's tend to be less
            than the y's.

    Returns
    -------
    u : float
        The Mann--Whitney rank sum statistic, U.

    unor : float
        The approximate Normal test statistic, z, as described in
        [Description].

    p : float
        The tail probability, p, as specified by the argument `tail`.

    ties : bool
        Indicates whether the pooled sample contained ties or not. This
        will be useful in checking which function to use should one wish
        to calculate an exact tail probability.

        `ties` = False, no ties were present (use ``prob_mwu_noties``
        for an exact probability).

        `ties` = True, ties were present (use ``prob_mwu_ties`` for an
        exact probability).

    ranks : float, ndarray, shape (n1+n2)
        Contains the ranks of the pooled sample. The ranks of the first
        sample are contained in the first n1 elements and those of the
        second sample are contained in the next n2 elements.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, n2 = *<value>*.

            Constraint: n2 >= 1.

        (`errno` 1)
            On entry, n1 = *<value>*.

            Constraint: n1 >= 1.

        (`errno` 2)
            On entry, `tail` = *<value>*.

            Constraint: `tail` = 'T', 'U' or 'L'.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            The pooled sample values are all the same, i.e., the
            variance of `u` = 0.0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Conover, W J, 1980, `Practical Nonparametric Statistics`, Wiley

    Neumann, N, 1988, `Some procedures for calculating the distributions
    of elementary nonparametric teststatistics`, Statistical Software
    Newsletter (14(3)), 120--126

    Siegel, S, 1956, `Non-parametric Statistics for the Behavioral
    Sciences`, McGraw--Hill
    """

    fun_name = 'naginterfaces.library.nonpar.test_mwu'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n1_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n1 = _b_utils._EngineIntScalarType(
        _n1_py, {'fun_name': fun_name, 'entity': '_n1'},
    ).to_ctypes()
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n2_py = _b_utils._EngineFloat64ArrayType.get_shape(_y, {'fun_name': fun_name, 'entity': 'y'}, exp_rank=1)[0]
    _n2 = _b_utils._EngineIntScalarType(
        _n2_py, {'fun_name': fun_name, 'entity': '_n2'},
    ).to_ctypes()
    _ranks = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n1_py+_n2_py,),
    )
    try:
        _return_scalars = _b_nonpar.test_mwu(
            _n1,
            _x,
            _n2,
            _y,
            tail,
            _ranks,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _u_py = _return_scalars.u
    _unor_py = _return_scalars.unor
    _p_py = _return_scalars.p
    _ties_py = _return_scalars.ties
    _ranks_py = _ranks
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['u', 'unor', 'p', 'ties', 'ranks'],
    )(
        _u_py,
        _unor_py,
        _p_py,
        _ties_py,
        _ranks_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def prob_mwu_noties(n1, n2, tail, u):
    u"""
    Computes the exact probabilities for the Mann--Whitney U statistic,
    no ties in pooled sample.

    ``prob_mwu_noties`` calculates the exact tail probability for the
    Mann--Whitney rank sum test statistic for the case where there are
    no ties in the two samples pooled together.

    For full information please refer to the NAG Library document for
    g08aj

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g08/g08ajf.html

    Parameters
    ----------
    n1 : int
        The number of non-tied pairs, n_1.

    n2 : int
        The size of the second sample, n_2.

    tail : str, length 1
        Indicates the choice of tail probability, and hence the
        alternative hypothesis.

        `tail` = 'T'

            A two tailed probability is calculated and the alternative
            hypothesis is H_1:F(x) != G(y).

        `tail` = 'U'

            An upper tailed probability is calculated and the
            alternative hypothesis H_1:F(x) < G(y), i.e., the x's tend
            to be greater than the y's.

        `tail` = 'L'

            A lower tailed probability is calculated and the alternative
            hypothesis H_1:F(x) > G(y), i.e., the x's tend to be less
            than the y's.

    u : float
        U, the value of the Mann--Whitney rank sum test statistic. This
        is the statistic returned through the argument `u` by
        ``test_mwu``.

    Returns
    -------
    p : float
        The exact tail probability, p, as specified by the argument
        `tail`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n2` = *<value>*.

            Constraint: `n2` >= 1.

        (`errno` 1)
            On entry, `n1` = *<value>*.

            Constraint: `n1` >= 1.

        (`errno` 2)
            On entry, `tail` = *<value>*.

            Constraint: `tail` = 'T', 'U' or 'L'.

        (`errno` 3)
            On entry, `u` = *<value>*.

            Constraint: `u` >= 0.0.

        (`errno` 4)
            On entry, lwrk = *<value>* and the minimum workspace is
            *<value>*.

            Constraint: lwrk >= (`n1`*`n2`)/2+1.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Conover, W J, 1980, `Practical Nonparametric Statistics`, Wiley

    Harding, E F, 1983, `An efficient minimal-storage procedure for
    calculating the Mann--Whitney U, generalised U and similar
    distributions`, Appl. Statist. (33), 1--6

    Neumann, N, 1988, `Some procedures for calculating the distributions
    of elementary nonparametric teststatistics`, Statistical Software
    Newsletter (14(3)), 120--126

    Siegel, S, 1956, `Non-parametric Statistics for the Behavioral
    Sciences`, McGraw--Hill
    """

    _return_scalar = _b_nonpar.prob_mwu_noties(
        n1,
        n2,
        tail,
        u,
    )
    _p_py = _return_scalar
    return _p_py

def prob_mwu_ties(n1, n2, tail, ranks, u):
    u"""
    Computes the exact probabilities for the Mann--Whitney U statistic,
    ties in pooled sample.

    ``prob_mwu_ties`` calculates the exact tail probability for the
    Mann--Whitney rank sum test statistic for the case where there are
    ties in the two samples pooled together.

    For full information please refer to the NAG Library document for
    g08ak

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g08/g08akf.html

    Parameters
    ----------
    n1 : int
        The number of non-tied pairs, n_1.

    n2 : int
        The size of the second sample, n_2.

    tail : str, length 1
        Indicates the choice of tail probability, and hence the
        alternative hypothesis.

        `tail` = 'T'

            A two tailed probability is calculated and the alternative
            hypothesis is H_1:F(x) != G(y).

        `tail` = 'U'

            An upper tailed probability is calculated and the
            alternative hypothesis H_1:F(x) < G(y), i.e., the x's tend
            to be greater than the y's.

        `tail` = 'L'

            A lower tailed probability is calculated and the alternative
            hypothesis H_1:F(x) > G(y), i.e., the x's tend to be less
            than the y's.

    ranks : float, array-like, shape (`n1`+`n2`)
        The ranks of the pooled sample. These ranks are output in the
        array `ranks` by ``test_mwu`` and should not be altered in any
        way if you are using the same n_1, n_2 and `u` as used in
        ``test_mwu``.

    u : float
        U, the value of the Mann--Whitney rank sum test statistic. This
        is the statistic returned through the argument `u` by
        ``test_mwu``.

    Returns
    -------
    p : float
        The tail probability, p, as specified by the argument `tail`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n2` = *<value>*.

            Constraint: `n2` >= 1.

        (`errno` 1)
            On entry, `n1` = *<value>*.

            Constraint: `n1` >= 1.

        (`errno` 2)
            On entry, `tail` = *<value>*.

            Constraint: `tail` = 'T', 'U' or 'L'.

        (`errno` 3)
            On entry, `u` = *<value>*.

            Constraint: `u` >= 0.0.

        (`errno` 5)
            On entry, at least one rank, given in `ranks`, was outside
            the expected range.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Conover, W J, 1980, `Practical Nonparametric Statistics`, Wiley

    Neumann, N, 1988, `Some procedures for calculating the distributions
    of elementary nonparametric teststatistics`, Statistical Software
    Newsletter (14(3)), 120--126

    Siegel, S, 1956, `Non-parametric Statistics for the Behavioral
    Sciences`, McGraw--Hill
    """

    fun_name = 'naginterfaces.library.nonpar.prob_mwu_ties'

    _ranks = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ranks'},
        ranks,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _return_scalar = _b_nonpar.prob_mwu_ties(
        n1,
        n2,
        tail,
        _ranks,
        u,
    )
    _p_py = _return_scalar
    return _p_py

def test_cochranq(x):
    u"""
    Performs the Cochran Q test on cross-classified binary data.

    ``test_cochranq`` performs the Cochran Q-test on cross-classified
    binary data.

    For full information please refer to the NAG Library document for
    g08al

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g08/g08alf.html

    Parameters
    ----------
    x : float, array-like, shape (n, k)
        The matrix of observed zero-one data. `x`[i-1,j-1] must contain
        the value x_{ij}, for j = 1,2,...,k, for i = 1,2,...,n.

    Returns
    -------
    q : float
        The value of the Cochran Q-test statistic.

    prob : float
        The upper tail probability, p, associated with the Cochran
        Q-test statistic, that is the probability of obtaining a value
        greater than or equal to the observed value (the output value of
        `q`).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, k = *<value>*.

            Constraint: k >= 2.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 2.

        (`errno` 2)
            On entry, i = *<value>*, j = *<value>* and `x`[i-1,j-1] =
            *<value>*.

            Constraint: `x`[i-1,j-1] = 0.0 or 1.0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            The solution has failed to converge while calculating the
            tail probability.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Conover, W J, 1980, `Practical Nonparametric Statistics`, Wiley

    Siegel, S, 1956, `Non-parametric Statistics for the Behavioral
    Sciences`, McGraw--Hill
    """

    fun_name = 'naginterfaces.library.nonpar.test_cochranq'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _k_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2)[1]
    _k = _b_utils._EngineIntScalarType(
        _k_py, {'fun_name': fun_name, 'entity': '_k'},
    ).to_ctypes()
    try:
        _return_scalars = _b_nonpar.test_cochranq(
            _n,
            _k,
            _x,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _q_py = _return_scalars.q
    _prob_py = _return_scalars.prob
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['q', 'prob'],
    )(
        _q_py,
        _prob_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def test_mooddavid(x, n1, itest):
    u"""
    Mood's and David's tests on two samples of unequal size.

    ``test_mooddavid`` performs Mood's and David's tests for dispersion
    differences between two independent samples of possibly unequal
    size.

    For full information please refer to the NAG Library document for
    g08ba

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g08/g08baf.html

    Parameters
    ----------
    x : float, array-like, shape (n)
        The first n_1 elements of `x` must be set to the data values in
        the first sample, and the next n_2 (= n-n_1) elements to the
        data values in the second sample.

    n1 : int
        The size of the first sample, n_1.

    itest : int
        The test(s) to be carried out.

        `itest` = 0

            Both Mood's and David's tests.

        `itest` = 1

            David's test only.

        `itest` = 2

            Mood's test only.

    Returns
    -------
    r : float, ndarray, shape (n)
        The ranks r_i, assigned to the data values x_i, for i =
        1,2,...,n.

    w : float
        Mood's test statistic, W, if requested.

    v : float
        David's test statistic, V, if requested.

    pw : float
        The lower tail probability, p_w, corresponding to the value of
        W, if Mood's test was requested.

    pv : float
        The lower tail probability, p_v, corresponding to the value of
        V, if David's test was requested.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n > 2.

        (`errno` 2)
            On entry, `n1` = *<value>* and n = *<value>*.

            Constraint: 1 < `n1` < n.

        (`errno` 3)
            On entry, `itest` = *<value>*.

            Constraint: `itest` = 0, 1 or 2.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Cooper, B E, 1975, `Statistics for Experimentalists`, Pergamon Press
    """

    fun_name = 'naginterfaces.library.nonpar.test_mooddavid'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _r = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _return_scalars = _b_nonpar.test_mooddavid(
        _x,
        _n,
        n1,
        _r,
        itest,
    )
    _r_py = _r
    _w_py = _return_scalars.w
    _v_py = _return_scalars.v
    _pw_py = _return_scalars.pw
    _pv_py = _return_scalars.pv
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['r', 'w', 'v', 'pw', 'pv'],
    )(
        _r_py,
        _w_py,
        _v_py,
        _pw_py,
        _pv_py,
    )
    return _return_data

def test_ks_1sample(x, dist, par, estima, ntype):
    u"""
    Performs the one-sample Kolmogorov--Smirnov test for standard
    distributions.

    ``test_ks_1sample`` performs the one sample Kolmogorov--Smirnov
    test, using one of the distributions provided.

    For full information please refer to the NAG Library document for
    g08cb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g08/g08cbf.html

    Parameters
    ----------
    x : float, array-like, shape (n)
        The sample observations x_1,x_2,...,x_n.

    dist : str
        The theoretical (null) distribution from which it is suspected
        the data may arise.

        `dist` = 'U'

            The uniform distribution over (a,b).

        `dist` = 'N'

            The Normal distribution with mean mu and variance sigma^2.

        `dist` = 'G'

            The gamma distribution with shape parameteralpha and scale
            parameter beta, where the mean = alphabeta.

        `dist` = 'BE'

            The beta distribution with shape parameters alpha and beta,
            where the mean = alpha/(alpha+beta).

        `dist` = 'BI'

            The binomial distribution with the number of trials, m, and
            the probability of a success, p.

        `dist` = 'E'

            The exponential distribution with parameter lambda, where
            the mean = 1/lambda.

        `dist` = 'P'

            The Poisson distribution with parameter mu, where the mean =
            mu.

        `dist` = 'NB'

            The negative binomial distribution with the number of
            trials, m, and the probability of success, p.

        `dist` = 'GP'

            The generalized Pareto distribution with shape parameter xi
            and scale beta.

        Any number of characters may be supplied as the actual
        parameter, however only the characters, maximum 2, required to
        uniquely identify the distribution are referenced.

    par : float, array-like, shape (2)
        If `estima` = 'S', `par` must contain the known values of the
        parameter(s) of the null distribution as follows.

        If a uniform distribution is used, `par`[0] and `par`[1] must
        contain the boundaries a and b respectively.

        If a Normal distribution is used, `par`[0] and `par`[1] must
        contain the mean, mu, and the variance, sigma^2, respectively.

        If a gamma distribution is used, `par`[0] and `par`[1] must
        contain the parameters alpha and beta respectively.

        If a beta distribution is used, `par`[0] and `par`[1] must
        contain the parameters alpha and beta respectively.

        If a binomial distribution is used, `par`[0] and `par`[1] must
        contain the parameters m and p respectively.

        If an exponential distribution is used, `par`[0] must contain
        the parameter lambda.

        If a Poisson distribution is used, `par`[0] must contain the
        parameter mu.

        If a negative binomial distribution is used, `par`[0] and
        `par`[1] must contain the parameters m and p respectively.

        If a generalized Pareto distribution is used, `par`[0] and
        `par`[1] must contain the parameters xi and beta respectively.

        If `estima` = 'E', `par` need not be set except when the null
        distribution requested is either the binomial or the negative
        binomial distribution in which case `par`[0] must contain the
        parameter m.

    estima : str, length 1
        `estima` must specify whether values of the parameters of the
        null distribution are known or are to be estimated from the
        data.

        `estima` = 'S'

            Values of the parameters will be supplied in the array `par`
            described above.

        `estima` = 'E'

            Parameters are to be estimated from the data except when the
            null distribution requested is the binomial distribution or
            the negative binomial distribution in which case the first
            parameter, m, must be supplied in `par`[0] and only the
            second parameter, p, is estimated from the data.

    ntype : int
        The test statistic to be calculated, i.e., the choice of
        alternative hypothesis.

        `ntype` = 1

            Computes D_n, to test H_0 against H_1,

        `ntype` = 2

            Computes D_n^+, to test H_0 against H_2,

        `ntype` = 3

            Computes D_n^-, to test H_0 against H_3.

    Returns
    -------
    par : float, ndarray, shape (2)
        If `estima` = 'S', `par` is unchanged; if `estima` = 'E', and
        `dist` = 'BI' or `dist` = 'NB' then `par`[1] is estimated from
        the data; otherwise `par`[0] and `par`[1] are estimated from the
        data.

    d : float
        The Kolmogorov--Smirnov test statistic (D_n, D_n^+ or D_n^-
        according to the value of `ntype`).

    z : float
        A standardized value, Z, of the test statistic, D, without any
        correction for continuity.

    p : float
        The probability, p, associated with the observed value of D
        where D may be D_n,D_n^+ or D_n^- depending on the value of
        `ntype` (see [Description]).

    sx : float, ndarray, shape (n)
        The sample observations, x_1,x_2,...,x_n, sorted in ascending
        order.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 3.

        (`errno` 2)
            On entry, `dist` = *<value>*.

            Constraint: `dist` = 'U', 'N', 'G', 'BE', 'BI', 'E', 'P',
            'NB' or 'GP'.

        (`errno` 3)
            On entry, `ntype` = *<value>*.

            Constraint: `ntype` = 1, 2 or 3.

        (`errno` 4)
            On entry, `estima` = *<value>*.

            Constraint: `estima` = 'E' or 'S'.

        (`errno` 5)
            On entry, `estima` = 'S' and `par`[0] = *<value>*; `par`[1]
            = *<value>*.

            Constraint: for the generalized Pareto distribution with
            `par`[0] < 0, 0 <= `x`[i-1] <= -`par`[1]/`par`[0], for i =
            1,2,...,n.

        (`errno` 5)
            On entry, `estima` = 'S' and `par`[1] = *<value>*.

            Constraint: for the generalized Pareto distribution,
            `par`[1] > 0.

        (`errno` 5)
            On entry, `estima` = 'S' and `par`[1] = *<value>*.

            Constraint: for the negative binomial distribution, 0 <
            `par`[1] < 1.

        (`errno` 5)
            On entry, `dist` = 'NB' and m = `par`[0] = *<value>*.

            Note that m must always be supplied.

            Constraint: for the negative binomial distribution, 1 <=
            `par`[0] < 1/eps, where eps = machine precision, see
            ``machine.precision``.

        (`errno` 5)
            On entry, `estima` = 'S' and `par`[0] = *<value>*.

            Constraint: for the Poisson distribution, 0 < `par`[0] <
            1000000.

        (`errno` 5)
            On entry, `estima` = 'S' and `par`[0] = *<value>*.

            Constraint: for the exponential distribution, `par`[0] > 0.

        (`errno` 5)
            On entry, `estima` = 'S' and `par`[1] = *<value>*.

            Constraint: for the binomial distribution, 0 < `par`[1] < 1.

        (`errno` 5)
            On entry, `dist` = 'BI' and m = `par`[0] = *<value>*.

            Note that m must always be supplied.

            Constraint: for the binomial distribution, 1 <= `par`[0] <
            1/eps, where eps = machine precision, see
            ``machine.precision``.

        (`errno` 5)
            On entry, `estima` = 'S' and `par`[0] = *<value>*; `par`[1]
            = *<value>*.

            Constraint: for the beta distribution, 0 < `par`[0] and
            `par`[1] <= 1000000.

        (`errno` 5)
            On entry, `estima` = 'S' and `par`[0] = *<value>*; `par`[1]
            = *<value>*.

            Constraint: for the gamma distribution, `par`[0] and
            `par`[1] > 0.

        (`errno` 5)
            On entry, `estima` = 'S' and `par`[1] = *<value>*.

            Constraint: for the Normal distribution, `par`[1] > 0.

        (`errno` 5)
            On entry, `estima` = 'S' and `par`[0] = *<value>*; `par`[1]
            = *<value>*.

            Constraint: for the uniform distribution, `par`[0] <
            `par`[1].

        (`errno` 6)
            On entry, `dist` = 'GP' and `estima` = 'E'.

            The parameter estimates are invalid; the data may not be
            from the generalized Pareto distribution.

        (`errno` 6)
            On entry, `dist` = 'E' or 'P' and all observations are zero.

            Constraint: at least one `x`[i-1] > 0, for i = 1,2,...,n.

        (`errno` 6)
            On entry, `dist` = 'BI' and all observations are zero or m.

            Constraint: at least one 0.0 < `x`[i-1] < `par`[0], for i =
            1,2,...,n.

        (`errno` 6)
            On entry, `dist` = 'G', 'E', 'P', 'NB' or 'GP' and at least
            one observation is negative.

            Constraint: `x`[i-1] >= 0, for i = 1,2,...,n.

        (`errno` 6)
            On entry, `dist` = 'BI' and at least one observation is
            illegal.

            Constraint: 0 <= `x`[i-1] <= `par`[0], for i = 1,2,...,n.

        (`errno` 6)
            On entry, `dist` = 'BE' and at least one observation is
            illegal.

            Constraint: 0 <= `x`[i-1] <= 1, for i = 1,2,...,n.

        (`errno` 6)
            On entry, `dist` = 'U' and at least one observation is
            illegal.

            Constraint: `par`[0] <= `x`[i-1] <= `par`[1], for i =
            1,2,...,n.

        (`errno` 7)
            On entry, `dist` = 'U', 'N', 'G', 'BE' or 'GP', `estima` =
            'E' and the whole sample is constant. Thus the variance is
            zero.

        (`errno` 8)
            On entry, `dist` = 'NB', `par`[0] = *<value>*, `par`[1] =
            *<value>*.

            The variance `par`[0]*(1-`par`[1])/(`par`[1]*`par`[1])
            exceeds 1000000.

        (`errno` 8)
            On entry, `dist` = 'BI', `par`[0] = *<value>*, `par`[1] =
            *<value>*.

            The variance `par`[0]*`par`[1]*(1-`par`[1]) exceeds 1000000.

        (`errno` 9)
            On entry, `dist` = 'G' and in the computation of the
            incomplete gamma function by ``specfun.gamma_incomplete``
            the convergence of the Taylor series or Legendre continued
            fraction fails within 600 iterations.

    References
    ----------
    Conover, W J, 1980, `Practical Nonparametric Statistics`, Wiley

    Feller, W, 1948, `On the Kolmogorov--Smirnov limit theorems for
    empirical distributions`, Ann. Math. Statist. (19), 179--181

    Kendall, M G and Stuart, A, 1973, `The Advanced Theory of Statistics
    (Volume 2)`, (3rd Edition), Griffin

    Kolmogorov, A N, 1933, `Sulla determinazione empirica di una legge
    di distribuzione`, Giornale dell' Istituto Italiano degli Attuari
    (4), 83--91

    Siegel, S, 1956, `Non-parametric Statistics for the Behavioral
    Sciences`, McGraw--Hill

    Smirnov, N, 1933, `Estimate of deviation between empirical
    distribution functions in two independent samples`, Bull. Moscow
    Univ. (2(2)), 3--16

    Smirnov, N, 1948, `Table for estimating the goodness of fit of
    empirical distributions`, Ann. Math. Statist. (19), 279--281
    """

    fun_name = 'naginterfaces.library.nonpar.test_ks_1sample'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _par = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'par'},
        par,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _sx = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _return_scalars = _b_nonpar.test_ks_1sample(
        _n,
        _x,
        dist,
        _par,
        estima,
        ntype,
        _sx,
    )
    _par_py = _par
    _d_py = _return_scalars.d
    _z_py = _return_scalars.z
    _p_py = _return_scalars.p
    _sx_py = _sx
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['par', 'd', 'z', 'p', 'sx'],
    )(
        _par_py,
        _d_py,
        _z_py,
        _p_py,
        _sx_py,
    )
    return _return_data

def test_ks_1sample_user(x, cdf, ntype, data=None):
    u"""
    Performs the one-sample Kolmogorov--Smirnov test for a user-supplied
    distribution.

    ``test_ks_1sample_user`` performs the one sample Kolmogorov--Smirnov
    distribution test, using a user-specified distribution.

    For full information please refer to the NAG Library document for
    g08cc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g08/g08ccf.html

    Parameters
    ----------
    x : float, array-like, shape (n)
        The sample observations, x_1,x_2,...,x_n.

    cdf : callable retval = cdf(x, data=None)
        `cdf` must return the value of the theoretical (null) cumulative
        distribution function for a given value of its argument.

        Parameters
        ~~~~~~~~~~
        x : float
            The argument for which `cdf` must be evaluated.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        retval : float
            The value of the theoretical (null) cumulative distribution
            function evaluated at `x`.

    ntype : int
        The statistic to be calculated, i.e., the choice of alternative
        hypothesis.

        `ntype` = 1

            Computes D_n, to test H_0 against H_1.

        `ntype` = 2

            Computes D_n^+, to test H_0 against H_2.

        `ntype` = 3

            Computes D_n^-, to test H_0 against H_3.

    data : arbitrary, optional
        User-communication data for callback functions.

    Returns
    -------
    d : float
        The Kolmogorov--Smirnov test statistic (D_n, D_n^+ or D_n^-
        according to the value of `ntype`).

    z : float
        A standardized value, Z, of the test statistic, D, without the
        continuity correction applied.

    p : float
        The probability, p, associated with the observed value of D,
        where D may D_n, D_n^+ or D_n^- depending on the value of
        `ntype` (see [Description]).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 2)
            On entry, `ntype` = *<value>*.

            Constraint: `ntype` = 1, 2 or 3.

        (`errno` 3)
            On entry, at x = *<value>*, F_0(x) = *<value>*.

            Constraint: 0.0 <= F_0(x) <= 1.

        (`errno` 4)
            On entry, at x = *<value>*, F_0(x) = *<value>* and at y =
            *<value>*, F_0(y) = *<value>*

            Constraint: when x < y, F_0(x) <= F_0(y).

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Conover, W J, 1980, `Practical Nonparametric Statistics`, Wiley

    Feller, W, 1948, `On the Kolmogorov--Smirnov limit theorems for
    empirical distributions`, Ann. Math. Statist. (19), 179--181

    Kendall, M G and Stuart, A, 1973, `The Advanced Theory of Statistics
    (Volume 2)`, (3rd Edition), Griffin

    Kolmogorov, A N, 1933, `Sulla determinazione empirica di una legge
    di distribuzione`, Giornale dell' Istituto Italiano degli Attuari
    (4), 83--91

    Siegel, S, 1956, `Non-parametric Statistics for the Behavioral
    Sciences`, McGraw--Hill

    Smirnov, N, 1933, `Estimate of deviation between empirical
    distribution functions in two independent samples`, Bull. Moscow
    Univ. (2(2)), 3--16

    Smirnov, N, 1948, `Table for estimating the goodness of fit of
    empirical distributions`, Ann. Math. Statist. (19), 279--281
    """

    def cb_cdf(x, data=None):
        """Wrapping user's cdf. Called as base cdf."""
        if data is not None:
            _retval_cdf = cdf(
                x,
                data,
            )
        else:
            _retval_cdf = cdf(
                x,
            )
        return _retval_cdf

    fun_name = 'naginterfaces.library.nonpar.test_ks_1sample_user'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b_utils._check_callable(cdf, {'fun_name': fun_name, 'entity': 'cdf'})
    _cdf = cb_cdf
    _return_scalars = _b_nonpar.test_ks_1sample_user(
        _n,
        _x,
        _cdf,
        ntype,
        data,
    )
    _d_py = _return_scalars.d
    _z_py = _return_scalars.z
    _p_py = _return_scalars.p
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['d', 'z', 'p'],
    )(
        _d_py,
        _z_py,
        _p_py,
    )
    return _return_data

def test_ks_2sample(x, y, ntype):
    u"""
    Performs the two-sample Kolmogorov--Smirnov test.

    ``test_ks_2sample`` performs the two sample Kolmogorov--Smirnov
    distribution test.

    For full information please refer to the NAG Library document for
    g08cd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g08/g08cdf.html

    Parameters
    ----------
    x : float, array-like, shape (n1)
        The observations from the first sample, x_1,x_2,...,x_n_1.

    y : float, array-like, shape (n2)
        The observations from the second sample, y_1,y_2,...,y_n_2.

    ntype : int
        The statistic to be computed, i.e., the choice of alternative
        hypothesis.

        `ntype` = 1

            Computes D_{n_1 n_2}, to test against H_1.

        `ntype` = 2

            Computes D_n_1 n_2^+, to test against H_2.

        `ntype` = 3

            Computes D_n_1 n_2^-, to test against H_3.

    Returns
    -------
    d : float
        The Kolmogorov--Smirnov test statistic (D_{n_1 n_2}, D_n_1 n_2^+
        or D_n_1 n_2^- according to the value of `ntype`).

    z : float
        A standardized value, Z, of the test statistic, D, without any
        correction for continuity.

    p : float
        The tail probability associated with the observed value of D,
        where D may be D_{n_1,n_2},D_n_1,n_2^+ or D_n_1,n_2^- depending
        on the value of `ntype` (see [Description]).

    sx : float, ndarray, shape (n1)
        The observations from the first sample sorted in ascending
        order.

    sy : float, ndarray, shape (n2)
        The observations from the second sample sorted in ascending
        order.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, n2 = *<value>*.

            Constraint: n2 >= 1.

        (`errno` 1)
            On entry, n1 = *<value>*.

            Constraint: n1 >= 1.

        (`errno` 2)
            On entry, `ntype` = *<value>*.

            Constraint: `ntype` = 1, 2 or 3.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            The iterative process used in the approximation of the
            probability for large n_1 and n_2 did not converge. For the
            two sided test p = 1 is returned. For the one-sided test p =
            0.5 is returned.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Conover, W J, 1980, `Practical Nonparametric Statistics`, Wiley

    Feller, W, 1948, `On the Kolmogorov--Smirnov limit theorems for
    empirical distributions`, Ann. Math. Statist. (19), 179--181

    Kendall, M G and Stuart, A, 1973, `The Advanced Theory of Statistics
    (Volume 2)`, (3rd Edition), Griffin

    Kim, P J and Jenrich, R I, 1973, `Tables of exact sampling
    distribution of the two sample Kolmogorov--Smirnov criterion
    D_{mn}(m < n)`, Selected Tables in Mathematical Statistics (1),
    80--129, American Mathematical Society

    Siegel, S, 1956, `Non-parametric Statistics for the Behavioral
    Sciences`, McGraw--Hill

    Smirnov, N, 1933, `Estimate of deviation between empirical
    distribution functions in two independent samples`, Bull. Moscow
    Univ. (2(2)), 3--16

    Smirnov, N, 1948, `Table for estimating the goodness of fit of
    empirical distributions`, Ann. Math. Statist. (19), 279--281
    """

    fun_name = 'naginterfaces.library.nonpar.test_ks_2sample'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n1_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n1 = _b_utils._EngineIntScalarType(
        _n1_py, {'fun_name': fun_name, 'entity': '_n1'},
    ).to_ctypes()
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n2_py = _b_utils._EngineFloat64ArrayType.get_shape(_y, {'fun_name': fun_name, 'entity': 'y'}, exp_rank=1)[0]
    _n2 = _b_utils._EngineIntScalarType(
        _n2_py, {'fun_name': fun_name, 'entity': '_n2'},
    ).to_ctypes()
    _sx = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n1_py,),
    )
    _sy = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n2_py,),
    )
    try:
        _return_scalars = _b_nonpar.test_ks_2sample(
            _n1,
            _x,
            _n2,
            _y,
            ntype,
            _sx,
            _sy,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _d_py = _return_scalars.d
    _z_py = _return_scalars.z
    _p_py = _return_scalars.p
    _sx_py = _sx
    _sy_py = _sy
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['d', 'z', 'p', 'sx', 'sy'],
    )(
        _d_py,
        _z_py,
        _p_py,
        _sx_py,
        _sy_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def test_chisq(ifreq, cb, dist, par, npest, prob):
    u"""
    Performs the chi^2 goodness-of-fit test, for standard continuous
    distributions.

    ``test_chisq`` computes the test statistic for the chi^2
    goodness-of-fit test for data with a chosen number of class
    intervals.

    For full information please refer to the NAG Library document for
    g08cg

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g08/g08cgf.html

    Parameters
    ----------
    ifreq : int, array-like, shape (nclass)
        `ifreq`[i] must specify the frequency of the ith class, O_i, for
        i = 1,2,...,k.

    cb : float, array-like, shape (nclass-1)
        `cb`[i-1] must specify the upper boundary value for the ith
        class, for i = 1,2,...,k-1.

    dist : str, length 1
        Indicates for which distribution the test is to be carried out.

        `dist` = 'N'

            The Normal distribution is used.

        `dist` = 'U'

            The uniform distribution is used.

        `dist` = 'E'

            The exponential distribution is used.

        `dist` = 'C'

            The chi^2-distribution is used.

        `dist` = 'G'

            The gamma distribution is used.

        `dist` = 'A'

            You must supply the class probabilities in the array `prob`.

    par : float, array-like, shape (2)
        Must contain the parameters of the distribution which is being
        tested. If you supply the probabilities (i.e., `dist` = 'A') the
        array `par` is not referenced.

        If a Normal distribution is used then `par`[0] and `par`[1] must
        contain the mean, mu, and the variance, sigma^2, respectively.

        If a uniform distribution is used then `par`[0] and `par`[1]
        must contain the boundaries a and b respectively.

        If an exponential distribution is used then `par`[0] must
        contain the parameter lambda. `par`[1] is not used.

        If a chi^2-distribution is used then `par`[0] must contain the
        number of degrees of freedom. `par`[1] is not used.

        If a gamma distribution is used `par`[0] and `par`[1] must
        contain the parameters alpha and beta respectively.

    npest : int
        The number of estimated parameters of the distribution.

    prob : float, array-like, shape (nclass)
        If you are supplying the probability distribution (i.e., `dist`
        = 'A') then `prob`[i] must contain the probability that X lies
        in the ith class.

        If `dist` != 'A', `prob` is not referenced.

    Returns
    -------
    chisq : float
        The test statistic, X^2, for the chi^2 goodness-of-fit test.

    p : float
        The upper tail probability from the chi^2-distribution
        associated with the test statistic, X^2, and the number of
        degrees of freedom.

    ndf : int
        Contains (nclass-1-`npest`), the degrees of freedom associated
        with the test.

    efreq : float, ndarray, shape (nclass)
        `efreq`[i-1] contains the expected frequency for the ith class,
        E_i, for i = 1,2,...,k.

    chisqi : float, ndarray, shape (nclass)
        `chisqi`[i-1] contains the contribution from the ith class to
        the test statistic, that is, (O_i-E_i)^2/E_i, for i = 1,2,...,k.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, nclass = *<value>*.

            Constraint: nclass >= 2.

        (`errno` 2)
            On entry, `dist` = *<value>*.

            Constraint: `dist` = 'N', 'U', 'E', 'C', 'G' or 'A'.

        (`errno` 3)
            On entry, `npest` = *<value>*.

            Constraint: 0 <= `npest` < nclass-1.

        (`errno` 4)
            On entry, i = *<value>* and `ifreq`[i-1] = *<value>*.

            Constraint: `ifreq`[i-1] >= 0.

        (`errno` 5)
            On entry, i = *<value>*, `cb`[i-2] = *<value>* and `cb`[i-1]
            = *<value>*.

            Constraint: `cb`[i-2] < `cb`[i-1].

        (`errno` 6)
            On entry, `cb`[0] = *<value>*.

            Constraint: `cb`[0] >= 0.0.

        (`errno` 7)
            On entry, `par`[0] = *<value>* and `par`[1] = *<value>*.

            Constraint: for the gamma distribution, `par`[0] > 0.0 and
            `par`[1] > 0.0.

        (`errno` 7)
            On entry, `par`[0] = *<value>*.

            Constraint: for the chi^2 distribution, `par`[0] > 0.0.

        (`errno` 7)
            On entry, `par`[0] = *<value>*.

            Constraint: for the exponential distribution, `par`[0] >
            0.0.

        (`errno` 7)
            On entry, `par`[0] = *<value>* and `par`[1] = *<value>*.

            Constraint: for the uniform distribution, `par`[0] <
            `par`[1], `par`[0] <= `cb`[0] and `par`[1] >=
            `cb`[nclass-2].

        (`errno` 7)
            On entry, `par`[1] = *<value>*.

            Constraint: for the Normal distribution, `par`[1] > 0.0.

        (`errno` 8)
            On entry, sum_i(`prob`[i]) = *<value>*.

            Constraint: sum_i(`prob`[i]) = 1.0.

        (`errno` 8)
            On entry, i = *<value>* and `prob`[i-1] = *<value>*.

            Constraint: `prob` > 0.0

        (`errno` 9)
            An expected frequency equals zero, when the observed
            frequency was not.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 10)
            At least one class has an expected frequency less than 1.

        (`errno` 11)
            The solution has failed to converge whilst computing the
            expected values. The returned solution may be an adequate
            approximation.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Conover, W J, 1980, `Practical Nonparametric Statistics`, Wiley

    Kendall, M G and Stuart, A, 1973, `The Advanced Theory of Statistics
    (Volume 2)`, (3rd Edition), Griffin

    Siegel, S, 1956, `Non-parametric Statistics for the Behavioral
    Sciences`, McGraw--Hill
    """

    fun_name = 'naginterfaces.library.nonpar.test_chisq'

    _ifreq = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ifreq'},
        ifreq,
        copy=False,
    )
    _nclass_py = _b_utils._EngineIntArrayType.get_shape(_ifreq, {'fun_name': fun_name, 'entity': 'ifreq'}, exp_rank=1)[0]
    _nclass = _b_utils._EngineIntScalarType(
        _nclass_py, {'fun_name': fun_name, 'entity': '_nclass'},
    ).to_ctypes()
    _cb = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'cb'},
        cb,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _par = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'par'},
        par,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _prob = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'prob'},
        prob,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _efreq = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nclass_py,),
    )
    _chisqi = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nclass_py,),
    )
    try:
        _return_scalars = _b_nonpar.test_chisq(
            _nclass,
            _ifreq,
            _cb,
            dist,
            _par,
            npest,
            _prob,
            _efreq,
            _chisqi,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _chisq_py = _return_scalars.chisq
    _p_py = _return_scalars.p
    _ndf_py = _return_scalars.ndf
    _efreq_py = _efreq
    _chisqi_py = _chisqi
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['chisq', 'p', 'ndf', 'efreq', 'chisqi'],
    )(
        _chisq_py,
        _p_py,
        _ndf_py,
        _efreq_py,
        _chisqi_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def gofstat_anddar(issort, y):
    u"""
    Calculates the Anderson--Darling goodness-of-fit test statistic.

    ``gofstat_anddar`` calculates the Anderson--Darling goodness-of-fit
    test statistic.

    For full information please refer to the NAG Library document for
    g08ch

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g08/g08chf.html

    Parameters
    ----------
    issort : bool
        Set `issort` = True if the observations are sorted in ascending
        order; otherwise the function will sort the observations.

    y : float, array-like, shape (n)
        y_i, for i = 1,2,...,n, the n observations.

    Returns
    -------
    gofs : float
        The Anderson--Darling goodness-of-fit test statistic.

    y : float, ndarray, shape (n)
        If `issort` = False, the data sorted in ascending order;
        otherwise the array is unchanged.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n > 1.

        (`errno` 3)
            `issort` = True and the data in `y` is not sorted in
            ascending order.

        (`errno` 9)
            The data in `y` must lie in the interval (0,1).

    References
    ----------
    Anderson, T W and Darling, D A, 1952, `Asymptotic theory of certain
    'goodness-of-fit' criteria based on stochastic processes`, Annals of
    Mathematical Statistics (23), 193--212
    """

    fun_name = 'naginterfaces.library.nonpar.gofstat_anddar'

    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_y, {'fun_name': fun_name, 'entity': 'y'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _return_scalar = _b_nonpar.gofstat_anddar(
        _n,
        issort,
        _y,
    )
    _y_py = _y
    _gofs_py = _return_scalar
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['gofs', 'y'],
    )(
        _gofs_py,
        _y_py,
    )
    return _return_data

def gofstat_anddar_unif(issort, y):
    u"""
    Calculates the Anderson--Darling goodness-of-fit test statistic and
    its probability for the case of uniformly distributed data.

    ``gofstat_anddar_unif`` calculates the Anderson--Darling
    goodness-of-fit test statistic and its probability for the case of
    standard uniformly distributed data.

    For full information please refer to the NAG Library document for
    g08cj

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g08/g08cjf.html

    Parameters
    ----------
    issort : bool
        Set `issort` = True if the observations are sorted in ascending
        order; otherwise the function will sort the observations.

    y : float, array-like, shape (n)
        y_i, for i = 1,2,...,n, the n observations.

    Returns
    -------
    y : float, ndarray, shape (n)
        If `issort` = False, the data sorted in ascending order;
        otherwise the array is unchanged.

    a2 : float
        A^2, the Anderson--Darling test statistic.

    p : float
        p, the upper tail probability for A^2.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n > 1.

        (`errno` 3)
            `issort` = True and the data in `y` is not sorted in
            ascending order.

        (`errno` 9)
            The data in `y` must lie in the interval (0,1).

    References
    ----------
    Anderson, T W and Darling, D A, 1952, `Asymptotic theory of certain
    'goodness-of-fit' criteria based on stochastic processes`, Annals of
    Mathematical Statistics (23), 193--212

    Marsaglia, G and Marsaglia, J, 2004, `Evaluating the
    Anderson--Darling distribution`, J. Statist. Software (9(2))
    """

    fun_name = 'naginterfaces.library.nonpar.gofstat_anddar_unif'

    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_y, {'fun_name': fun_name, 'entity': 'y'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _return_scalars = _b_nonpar.gofstat_anddar_unif(
        _n,
        issort,
        _y,
    )
    _y_py = _y
    _a2_py = _return_scalars.a2
    _p_py = _return_scalars.p
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['y', 'a2', 'p'],
    )(
        _y_py,
        _a2_py,
        _p_py,
    )
    return _return_data

def gofstat_anddar_normal(issort, y):
    u"""
    Calculates the Anderson--Darling goodness-of-fit test statistic and
    its probability for the case of a fully-unspecified Normal
    distribution.

    ``gofstat_anddar_normal`` calculates the Anderson--Darling
    goodness-of-fit test statistic and its probability for the case of a
    fully-unspecified Normal distribution.

    For full information please refer to the NAG Library document for
    g08ck

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g08/g08ckf.html

    Parameters
    ----------
    issort : bool
        Set `issort` = True if the observations are sorted in ascending
        order; otherwise the function will sort the observations.

    y : float, array-like, shape (n)
        y_i, for i = 1,2,...,n, the n observations.

    Returns
    -------
    ybar : float
        The maximum likelihood estimate of mean.

    yvar : float
        The maximum likelihood estimate of variance.

    a2 : float
        A^2, the Anderson--Darling test statistic.

    aa2 : float
        The adjusted A^2.

    p : float
        p, the upper tail probability for the adjusted A^2.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n > 1.

        (`errno` 3)
            `issort` = True and the data in `y` is not sorted in
            ascending order.

    References
    ----------
    Anderson, T W and Darling, D A, 1952, `Asymptotic theory of certain
    'goodness-of-fit' criteria based on stochastic processes`, Annals of
    Mathematical Statistics (23), 193--212

    Stephens, M A and D'Agostino, R B, 1986, `Goodness-of-Fit
    Techniques`, Marcel Dekker, New York
    """

    fun_name = 'naginterfaces.library.nonpar.gofstat_anddar_normal'

    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_y, {'fun_name': fun_name, 'entity': 'y'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _return_scalars = _b_nonpar.gofstat_anddar_normal(
        _n,
        issort,
        _y,
    )
    _ybar_py = _return_scalars.ybar
    _yvar_py = _return_scalars.yvar
    _a2_py = _return_scalars.a2
    _aa2_py = _return_scalars.aa2
    _p_py = _return_scalars.p
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['ybar', 'yvar', 'a2', 'aa2', 'p'],
    )(
        _ybar_py,
        _yvar_py,
        _a2_py,
        _aa2_py,
        _p_py,
    )
    return _return_data

def gofstat_anddar_exp(issort, y):
    u"""
    Calculates the Anderson--Darling goodness-of-fit test statistic and
    its probability for the case of an unspecified exponential
    distribution.

    ``gofstat_anddar_exp`` calculates the Anderson--Darling
    goodness-of-fit test statistic and its probability for the case of
    an unspecified exponential distribution.

    For full information please refer to the NAG Library document for
    g08cl

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g08/g08clf.html

    Parameters
    ----------
    issort : bool
        Set `issort` = True if the observations are sorted in ascending
        order; otherwise the function will sort the observations.

    y : float, array-like, shape (n)
        y_i, for i = 1,2,...,n, the n observations.

    Returns
    -------
    ybar : float
        The maximum likelihood estimate of mean.

    a2 : float
        A^2, the Anderson--Darling test statistic.

    aa2 : float
        The adjusted A^2.

    p : float
        p, the upper tail probability for the adjusted A^2.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n > 1.

        (`errno` 3)
            `issort` = True and the data in `y` is not sorted in
            ascending order.

        (`errno` 9)
            The data in `y` must be greater than zero.

    References
    ----------
    Anderson, T W and Darling, D A, 1952, `Asymptotic theory of certain
    'goodness-of-fit' criteria based on stochastic processes`, Annals of
    Mathematical Statistics (23), 193--212

    Stephens, M A and D'Agostino, R B, 1986, `Goodness-of-Fit
    Techniques`, Marcel Dekker, New York
    """

    fun_name = 'naginterfaces.library.nonpar.gofstat_anddar_exp'

    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_y, {'fun_name': fun_name, 'entity': 'y'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _return_scalars = _b_nonpar.gofstat_anddar_exp(
        _n,
        issort,
        _y,
    )
    _ybar_py = _return_scalars.ybar
    _a2_py = _return_scalars.a2
    _aa2_py = _return_scalars.aa2
    _p_py = _return_scalars.p
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['ybar', 'a2', 'aa2', 'p'],
    )(
        _ybar_py,
        _a2_py,
        _aa2_py,
        _p_py,
    )
    return _return_data

def concordance_kendall(x):
    u"""
    Kendall's coefficient of concordance.

    ``concordance_kendall`` calculates Kendall's coefficient of
    concordance on k independent rankings of n objects or individuals.

    For full information please refer to the NAG Library document for
    g08da

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g08/g08daf.html

    Parameters
    ----------
    x : float, array-like, shape (k, n)
        `x`[i-1,j-1] must be set to the value x_{ij} of object j in
        comparison i, for j = 1,2,...,n, for i = 1,2,...,k.

    Returns
    -------
    w : float
        The value of Kendall's coefficient of concordance, W.

    p : float
        The approximate significance, p, of W.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 2.

        (`errno` 3)
            On entry, k = *<value>*.

            Constraint: k >= 2.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Siegel, S, 1956, `Non-parametric Statistics for the Behavioral
    Sciences`, McGraw--Hill
    """

    fun_name = 'naginterfaces.library.nonpar.concordance_kendall'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _k_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2)[0]
    _k = _b_utils._EngineIntScalarType(
        _k_py, {'fun_name': fun_name, 'entity': '_k'},
    ).to_ctypes()
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _return_scalars = _b_nonpar.concordance_kendall(
        _x,
        _k,
        _n,
    )
    _w_py = _return_scalars.w
    _p_py = _return_scalars.p
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['w', 'p'],
    )(
        _w_py,
        _p_py,
    )
    return _return_data

def randtest_runs(cl, x, m, nruns, ncount, comm):
    u"""
    Performs the runs up or runs down test for randomness.

    ``randtest_runs`` performs a runs up (or a runs down) test on a
    sequence of observations.

    For full information please refer to the NAG Library document for
    g08ea

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g08/g08eaf.html

    Parameters
    ----------
    cl : str, length 1
        Must specify the type of call to ``randtest_runs``.

        `cl` = 'S'

            This is the one and only call to ``randtest_runs`` (single
            call mode). All data are to be input at once. All test
            statistics are computed after the counting of runs is
            complete.

        `cl` = 'F'

            This is the first call to the function. All initializations
            are carried out and the counting of runs begins. The final
            test statistics are not computed since further calls will be
            made to ``randtest_runs``.

        `cl` = 'I'

            This is an intermediate call during which the counts of runs
            are updated. The final test statistics are not computed
            since further calls will be made to ``randtest_runs``.

        `cl` = 'L'

            This is the last call to ``randtest_runs``. The test
            statistics are computed after the final counting of runs is
            completed.

    x : float, array-like, shape (n)
        The sequence of observations.

    m : int
        The maximum number of runs to be sought. If `m` <= 0 then no
        limit is placed on the number of runs that are found.

        `m` must not be changed between calls to ``randtest_runs``.

    nruns : int
        If `cl` = 'S' or 'F', `nruns` need not be set.

        If `cl` = 'I' or 'L', `nruns` must contain the value returned by
        the previous call to ``randtest_runs``.

    ncount : int, array-like, shape (maxr)
        If `cl` = 'S' or 'F', `ncount` need not be set.

        If `cl` = 'I' or 'L', `ncount` must contain the values returned
        by the previous call to ``randtest_runs``.

    comm : dict, communication object, modified in place
        Communication structure.

        `On initial entry`: need not be set.

    Returns
    -------
    nruns : int
        The number of runs actually found.

    ncount : int, ndarray, shape (maxr)
        The counts of runs of the different lengths, c_i, for i =
        1,2,...,r.

    ex : float, ndarray, shape (maxr)
        If `cl` = 'S' or 'L', (i.e., if it is the final exit) then `ex`
        contains the expected values of the counts, e_i, for i =
        1,2,...,r.

        Otherwise the elements of `ex` are not set.

    cov : float, ndarray, shape (maxr, maxr)
        If `cl` = 'S' or 'L' (i.e., if it is the final exit) then `cov`
        contains the covariance matrix of the counts, Sigma_c.

        Otherwise the elements of `cov` are not set.

    chi : float
        If `cl` = 'S' or 'L' (i.e., if it is the final exit), `chi`
        contains the approximate chi^2 test statistic, X^2.

        Otherwise `chi` is not set.

    df : float
        If `cl` = 'S' or 'L' (i.e., if it is the final exit) then `df`
        contains the degrees of freedom of the chi^2 statistic.

        Otherwise `df` is not set.

    prob : float
        If `cl` = 'S' or 'L', (i.e., if it is the final exit) then
        `prob` contains the upper tail probability corresponding to the
        chi^2 test statistic, i.e., the significance level.

        Otherwise `prob` is not set.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `cl` = *<value>*.

            Constraint: `cl` = 'S', 'F', 'I' or 'L'.

        (`errno` 2)
            On entry, n = *<value>*.

            Constraint: if `cl` = 'S', n >= 3, otherwise n >= 1.

        (`errno` 3)
            On entry, `m` = *<value>* and n = *<value>*.

            Constraint: if `cl` = 'S', `m` <= n.

        (`errno` 4)
            On entry, maxr = *<value>* and n = *<value>*.

            Constraint: if `cl` = 'S', maxr < n.

        (`errno` 4)
            On entry, maxr = *<value>*.

            Constraint: maxr >= 1.

        (`errno` 7)
            There is a tie in the sequence of observations.

        (`errno` 8)
            The total length of the runs found is less than maxr.

            maxr = *<value>* whereas the total length of all runs is
            *<value>*.

        (`errno` 9)
            The covariance matrix stored in `cov` is not positive
            definite, thus the approximate chi^2 test statistic cannot
            be computed.

            This may be because maxr is too large relative to the length
            of the full sequence.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 10)
            The number of runs requested were not found, only *<value>*
            out of the requested *<value>* where found.

            All statistics are returned and may still be of use.

    References
    ----------
    Dagpunar, J, 1988, `Principles of Random Variate Generation`, Oxford
    University Press

    Knuth, D E, 1981, `The Art of Computer Programming (Volume 2)`, (2nd
    Edition), Addison--Wesley

    Morgan, B J T, 1984, `Elements of Simulation`, Chapman and Hall

    Ripley, B D, 1987, `Stochastic Simulation`, Wiley
    """

    fun_name = 'naginterfaces.library.nonpar.randtest_runs'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _ncount = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ncount'},
        ncount,
        copy=True,
    )
    _maxr_py = _b_utils._EngineIntArrayType.get_shape(_ncount, {'fun_name': fun_name, 'entity': 'ncount'}, exp_rank=1)[0]
    _maxr = _b_utils._EngineIntScalarType(
        _maxr_py, {'fun_name': fun_name, 'entity': '_maxr'},
    ).to_ctypes()
    _ex = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_maxr_py,),
    )
    _cov = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_maxr_py, _maxr_py),
        sorder=_b_utils._EngineData.default_sorder,
    )
    try:
        _return_scalars = _b_nonpar.randtest_runs(
            cl,
            _n,
            _x,
            m,
            _maxr,
            nruns,
            _ncount,
            _ex,
            _cov,
            comm,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _nruns_py = _return_scalars.nruns
    _ncount_py = _ncount
    _ex_py = _ex
    _cov_py = _cov
    _chi_py = _return_scalars.chi
    _df_py = _return_scalars.df
    _prob_py = _return_scalars.prob
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['nruns', 'ncount', 'ex', 'cov', 'chi', 'df', 'prob'],
    )(
        _nruns_py,
        _ncount_py,
        _ex_py,
        _cov_py,
        _chi_py,
        _df_py,
        _prob_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def randtest_pairs(cl, x, lag, ncount, comm):
    u"""
    Performs the pairs (serial) test for randomness.

    ``randtest_pairs`` performs a pairs test on a sequence of
    observations in the interval [0,1].

    For full information please refer to the NAG Library document for
    g08eb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g08/g08ebf.html

    Parameters
    ----------
    cl : str, length 1
        Indicates the type of call to ``randtest_pairs``.

        `cl` = 'S'

            This is the one and only call to ``randtest_pairs`` (single
            call mode). All data are to be input at once. All test
            statistics are computed after the counting of pairs is
            complete.

        `cl` = 'F'

            This is the first call to the function. All initializations
            are carried out and the counting of pairs begins. The final
            test statistics are not computed since further calls will be
            made to ``randtest_pairs``.

        `cl` = 'I'

            This is an intermediate call during which the counts of
            pairs are updated. The final test statistics are not
            computed since further calls will be made to
            ``randtest_pairs``.

        `cl` = 'L'

            This is the last call to ``randtest_pairs``. The test
            statistics are computed after the final counting of runs is
            complete.

    x : float, array-like, shape (n)
        The sequence of observations.

    lag : int
        l, the lag to be used in choosing pairs.

        If `lag` = 1, then we consider the pairs (`x`[i-1],`x`[i]), for
        i = 1,3,...,n-1, where n is the number of observations.

        If `lag` > 1, then we consider the pairs (`x`[i-1],`x`[i+l-1]),
        for i = 1,2,...,l,2l+1,2l+2,...,3l,4l+1,...,n-l, where n is the
        number of observations.

        `lag` must not be changed between calls to ``randtest_pairs``.

    ncount : int, array-like, shape (msize, msize)
        If `cl` = 'S' or 'F', `ncount` need not be set.

        If `cl` = 'I' or 'L', `ncount` must contain the values returned
        by the previous call to ``randtest_pairs``.

    comm : dict, communication object, modified in place
        Communication structure.

        `On initial entry`: need not be set.

    Returns
    -------
    ncount : int, ndarray, shape (msize, msize)
        Is an msize by msize matrix containing the counts of the number
        of pairs in each cell, c_{ij}, for j = 1,2,...,m, for i =
        1,2,...,m.

    ex : float
        If `cl` = 'S' or 'L' (i.e., if it is a final exit) then `ex`
        contains the expected number of counts in each cell, e.

        Otherwise `ex` is not set.

    chi : float
        If `cl` = 'S' or 'L' (i.e., if it is a final exit) then `chi`
        contains the chi^2 test statistic, X^2, for testing the null
        hypothesis of randomness.

        Otherwise `chi` is not set.

    df : float
        If `cl` = 'S' or 'L' (i.e., if it is a final exit) then `df`
        contains the degrees of freedom for the chi^2 statistic.

        Otherwise `df` is not set.

    prob : float
        If `cl` = 'S' or 'L' (i.e., if it is a final exit) then `prob`
        contains the upper tail probability associated with the chi^2
        test statistic, i.e., the significance level.

        Otherwise `prob` is not set.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `cl` = *<value>*.

            Constraint: `cl` = 'S', 'F', 'I' or 'L'.

        (`errno` 2)
            On entry, n = *<value>*.

            Constraint: if `cl` = 'S', n >= 2, otherwise n >= 1.

        (`errno` 3)
            On entry, msize = *<value>*.

            Constraint: msize >= 2

        (`errno` 4)
            On entry, `lag` = *<value>* and n = *<value>*.

            Constraint: `lag` > 0 and if `cl` = 'S', `lag` < n.

        (`errno` 6)
            On entry, at least one element of `x` is out of range.

            Constraint: 0 <= `x`[i-1] <= 1, for i = 1,2,...,n.

        (`errno` 7)
            No pairs were found. This will occur if the value of `lag`
            is greater than or equal to the total number of
            observations.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 8)
            msize is too large relative to the number of pairs,
            therefore the expected value for at least one cell is less
            than or equal to 5.0.

            This implies that the chi^2 distribution may not be a very
            good approximation to the distribution of test statistic.

            msize = *<value>*, number of pairs = *<value>* and expected
            value = *<value>*.

            All statistics are returned and may still be of use.

    References
    ----------
    Dagpunar, J, 1988, `Principles of Random Variate Generation`, Oxford
    University Press

    Knuth, D E, 1981, `The Art of Computer Programming (Volume 2)`, (2nd
    Edition), Addison--Wesley

    Morgan, B J T, 1984, `Elements of Simulation`, Chapman and Hall

    Ripley, B D, 1987, `Stochastic Simulation`, Wiley
    """

    fun_name = 'naginterfaces.library.nonpar.randtest_pairs'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _ncount = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ncount'},
        ncount,
        copy=True,
    )
    _msize_py = _b_utils._EngineIntArrayType.get_shape(_ncount, {'fun_name': fun_name, 'entity': 'ncount'}, exp_rank=2)[0]
    _msize = _b_utils._EngineIntScalarType(
        _msize_py, {'fun_name': fun_name, 'entity': '_msize'},
    ).to_ctypes()
    try:
        _return_scalars = _b_nonpar.randtest_pairs(
            cl,
            _n,
            _x,
            _msize,
            lag,
            _ncount,
            comm,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _ncount_py = _ncount
    _ex_py = _return_scalars.ex
    _chi_py = _return_scalars.chi
    _df_py = _return_scalars.df
    _prob_py = _return_scalars.prob
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['ncount', 'ex', 'chi', 'df', 'prob'],
    )(
        _ncount_py,
        _ex_py,
        _chi_py,
        _df_py,
        _prob_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def randtest_triplets(cl, x, ncount, comm):
    u"""
    Performs the triplets test for randomness.

    ``randtest_triplets`` performs the triplets test on a sequence of
    observations from the interval [0,1].

    For full information please refer to the NAG Library document for
    g08ec

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g08/g08ecf.html

    Parameters
    ----------
    cl : str, length 1
        Indicates the type of call to ``randtest_triplets``.

        `cl` = 'S'

            This is the one and only call to ``randtest_triplets``
            (single call mode). All data are to be input at once. All
            test statistics are computed after counting of the triplets
            is complete.

        `cl` = 'F'

            This is the first call to the function. All initializations
            are carried out and the counting of triplets begins. The
            final test statistics are not computed since further calls
            will be made to ``randtest_triplets``.

        `cl` = 'I'

            This is an intermediate call during which counts of the
            triplets are updated. The final test statistics are not
            computed since further calls will be made to
            ``randtest_triplets``.

        `cl` = 'L'

            This is the last call to ``randtest_triplets``. The test
            statistics are computed after the final counting of the
            triplets is complete.

    x : float, array-like, shape (n)
        The sequence of observations.

    ncount : int, array-like, shape (msize, msize, msize)
        If `cl` = 'S' or 'F', `ncount` need not be set.

        If `cl` = 'I' or 'L', `ncount` must contain the values returned
        by the previous call to ``randtest_triplets``.

    comm : dict, communication object, modified in place
        Communication structure.

        `On initial entry`: need not be set.

    Returns
    -------
    ncount : int, ndarray, shape (msize, msize, msize)
        Is an msize by msize by msize matrix containing the counts of
        the number of triplets, c_{jkl}, for l = 1,2,...,m, for k =
        1,2,...,m, for j = 1,2,...,m.

    ex : float
        If `cl` = 'S' or 'L' (i.e., if it is a final exit) then `ex`
        contains the expected number of counts for each element of the
        count matrix.

        Otherwise `ex` is not set.

    chi : float
        If `cl` = 'S' or 'L' (i.e., if it is a final exit) then `chi`
        contains the chi^2 test statistic, X^2, for testing the null
        hypothesis of randomness.

        Otherwise `chi` is not set.

    df : float
        If `cl` = 'S' or 'L' (i.e., if it is a final exit) then `df`
        contains the degrees of freedom for the chi^2 statistic.

        Otherwise `df` is not set.

    prob : float
        If `cl` = 'S' or 'L' (i.e., if it is a final exit) then `prob`
        contains the upper tail probability associated with the chi^2
        test statistic, i.e., the significance level.

        Otherwise `prob` is not set.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `cl` = *<value>*.

            Constraint: `cl` = 'S', 'F', 'I' or 'L'.

        (`errno` 2)
            On entry, n = *<value>*.

            Constraint: if `cl` = 'S', n >= 3, otherwise n >= 1.

        (`errno` 3)
            On entry, msize = *<value>*.

            Constraint: msize >= 2.

        (`errno` 5)
            On entry, at least one element of `x` is out of range.

            Constraint: 0 <= `x`[i-1] <= 1, for i = 1,2,...,n.

        (`errno` 6)
            No triplets were found because less than 3 observations were
            provided in total.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 7)
            msize is too large relative to the number of triplets,
            therefore the expected value for at least one cell is less
            than or equal to 5.0.

            This implies that the chi^2 distribution may not be a very
            good approximation to the distribution of the test
            statistic.

            msize = *<value>*, number of triplets = *<value>* and
            expected value = *<value>*.

            All statistics are returned and may still be of use.

    References
    ----------
    Dagpunar, J, 1988, `Principles of Random Variate Generation`, Oxford
    University Press

    Knuth, D E, 1981, `The Art of Computer Programming (Volume 2)`, (2nd
    Edition), Addison--Wesley

    Morgan, B J T, 1984, `Elements of Simulation`, Chapman and Hall

    Ripley, B D, 1987, `Stochastic Simulation`, Wiley
    """

    fun_name = 'naginterfaces.library.nonpar.randtest_triplets'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _ncount = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ncount'},
        ncount,
        copy=True,
    )
    _msize_py = _b_utils._EngineIntArrayType.get_shape(_ncount, {'fun_name': fun_name, 'entity': 'ncount'}, exp_rank=3)[0]
    _msize = _b_utils._EngineIntScalarType(
        _msize_py, {'fun_name': fun_name, 'entity': '_msize'},
    ).to_ctypes()
    try:
        _return_scalars = _b_nonpar.randtest_triplets(
            cl,
            _n,
            _x,
            _msize,
            _ncount,
            comm,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _ncount_py = _ncount
    _ex_py = _return_scalars.ex
    _chi_py = _return_scalars.chi
    _df_py = _return_scalars.df
    _prob_py = _return_scalars.prob
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['ncount', 'ex', 'chi', 'df', 'prob'],
    )(
        _ncount_py,
        _ex_py,
        _chi_py,
        _df_py,
        _prob_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def randtest_gaps(cl, x, m, rlo, rup, totlen, ngaps, ncount, comm):
    u"""
    Performs the gaps test for randomness.

    ``randtest_gaps`` performs a gaps test on a sequence of
    observations.

    For full information please refer to the NAG Library document for
    g08ed

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g08/g08edf.html

    Parameters
    ----------
    cl : str, length 1
        Indicates the type of call to ``randtest_gaps``.

        `cl` = 'S'

            This is the one and only call to ``randtest_gaps`` (single
            call mode). All data are to be input at once. All test
            statistics are computed after the counting of gaps is
            complete.

        `cl` = 'F'

            This is the first call to the function. All initializations
            are carried out before the counting of gaps begins. The
            final test statistics are not computed since further calls
            will be made to ``randtest_gaps``.

        `cl` = 'I'

            This is an intermediate call during which the counts of gaps
            are updated. The final test statistics are not computed
            since further calls will be made to ``randtest_gaps``.

        `cl` = 'L'

            This is the last call to ``randtest_gaps``. The test
            statistics are computed after the final counting of gaps is
            complete.

    x : float, array-like, shape (n)
        The sequence of observations.

    m : int
        The maximum number of gaps to be sought. If `m` <= 0 then there
        is no limit placed on the number of gaps that are found.

        `m` should not be changed between calls to ``randtest_gaps``.

    rlo : float
        The lower limit of the interval to be used to define the gaps,
        r_l.

        `rlo` must not be changed between calls to ``randtest_gaps``.

    rup : float
        The upper limit of the interval to be used to define the gaps,
        r_u.

        `rup` must not be changed between calls to ``randtest_gaps``.

    totlen : float
        The total length of the interval which contains all possible
        numbers that may arise in the sequence.

    ngaps : int
        If `cl` = 'S' or 'F', `ngaps` need not be set.

        If `cl` = 'I' or 'L', `ngaps` must contain the value returned by
        the previous call to ``randtest_gaps``.

    ncount : int, array-like, shape (maxg)
        If `cl` = 'S' or 'F', `ncount` need not be set.

        If `cl` = 'I' or 'L', `ncount` must contain the values returned
        by the previous call to ``randtest_gaps``.

    comm : dict, communication object, modified in place
        Communication structure.

        `On initial entry`: need not be set.

    Returns
    -------
    ngaps : int
        The number of gaps actually found, ngaps.

    ncount : int, ndarray, shape (maxg)
        The counts of gaps of the different lengths, c_i, for i =
        1,2,...,k.

    ex : float, ndarray, shape (maxg)
        If `cl` = 'S' or 'L' (i.e., if it is a final exit) then `ex`
        contains the expected values of the counts.

        Otherwise the elements of `ex` are not set.

    chi : float
        If `cl` = 'S' or 'L' (i.e., if it is a final exit) then `chi`
        contains the chi^2 test statistic, X^2, for testing the null
        hypothesis of randomness.

        Otherwise `chi` is not set.

    df : float
        If `cl` = 'S' or 'L' (i.e., if it is a final exit) then `df`
        contains the degrees of freedom for the chi^2 statistic.

        Otherwise `df` is not set.

    prob : float
        If `cl` = 'S' or 'L' (i.e., if it is a final exit) then `prob`
        contains the upper tail probability associated with the chi^2
        test statistic, i.e., the significance level.

        Otherwise `prob` is not set.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `cl` = *<value>*.

            Constraint: `cl` = 'S', 'F', 'I' or 'L'.

        (`errno` 2)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 3)
            On entry, `m` = *<value>* and n = *<value>*.

            Constraint: if `cl` = 'S', `m` <= n.

        (`errno` 4)
            On entry, maxg = *<value>* and n = *<value>*.

            Constraint: if `cl` = 'S', maxg <= n.

        (`errno` 4)
            On entry, maxg = *<value>*.

            Constraint: maxg > 1.

        (`errno` 5)
            On entry, `rlo` = *<value>*, `rup` = *<value>* and `totlen`
            = *<value>*.

            Constraint: `rup`-`rlo` < `totlen`.

        (`errno` 5)
            On entry, `totlen` = *<value>*.

            Constraint: `totlen` > 0.0.

        (`errno` 5)
            On entry, `rlo` = *<value>* and `rup` = *<value>*.

            Constraint: `rup` > `rlo`.

        (`errno` 6)
            No gaps were found. Try using a longer sequence, or increase
            the size of the interval `rup`-`rlo`.

        (`errno` 7)
            The expected frequency in class i = *<value>* is zero. The
            value of (`rup`-`rlo`)/`totlen` may be too close to 0.0 or
            1.0. or maxg is too large relative to the number of gaps
            found.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 8)
            The number of gaps requested were not found, only *<value>*
            out of the requested *<value>* where found.

            All statistics are returned and may still be of use.

        (`errno` 9)
            The expected frequency of at least one class is less than 1.

            This implies that the chi^2 may not be a very good
            approximation to the distribution of the test statistics.

            All statistics are returned and may still be of use.

    References
    ----------
    Dagpunar, J, 1988, `Principles of Random Variate Generation`, Oxford
    University Press

    Knuth, D E, 1981, `The Art of Computer Programming (Volume 2)`, (2nd
    Edition), Addison--Wesley

    Morgan, B J T, 1984, `Elements of Simulation`, Chapman and Hall

    Ripley, B D, 1987, `Stochastic Simulation`, Wiley
    """

    fun_name = 'naginterfaces.library.nonpar.randtest_gaps'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _ncount = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ncount'},
        ncount,
        copy=True,
    )
    _maxg_py = _b_utils._EngineIntArrayType.get_shape(_ncount, {'fun_name': fun_name, 'entity': 'ncount'}, exp_rank=1)[0]
    _maxg = _b_utils._EngineIntScalarType(
        _maxg_py, {'fun_name': fun_name, 'entity': '_maxg'},
    ).to_ctypes()
    _ex = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_maxg_py,),
    )
    try:
        _return_scalars = _b_nonpar.randtest_gaps(
            cl,
            _n,
            _x,
            m,
            _maxg,
            rlo,
            rup,
            totlen,
            ngaps,
            _ncount,
            _ex,
            comm,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _ngaps_py = _return_scalars.ngaps
    _ncount_py = _ncount
    _ex_py = _ex
    _chi_py = _return_scalars.chi
    _df_py = _return_scalars.df
    _prob_py = _return_scalars.prob
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['ngaps', 'ncount', 'ex', 'chi', 'df', 'prob'],
    )(
        _ngaps_py,
        _ncount_py,
        _ex_py,
        _chi_py,
        _df_py,
        _prob_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def rank_regsn(nv, y, x, idist, nmax, tol):
    u"""
    Regression using ranks, uncensored data.

    ``rank_regsn`` calculates the parameter estimates, score statistics
    and their variance-covariance matrices for the linear model using a
    likelihood based on the ranks of the observations.

    For full information please refer to the NAG Library document for
    g08ra

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g08/g08raf.html

    Parameters
    ----------
    nv : int, array-like, shape (ns)
        The number of observations in the ith sample, for i =
        1,2,...,ns.

    y : float, array-like, shape (sum(`nv`))
        The observations in each sample. Specifically, `y`[ sum_k =
        1^i-1`nv`[k-1]+j -1] must contain the jth observation in the ith
        sample.

    x : float, array-like, shape (nsum, ip)
        The design matrices for each sample. Specifically, `x`[ sum_k =
        1^i-1 `nv`[k-1] +j -1,l-1] must contain the value of the lth
        explanatory variable for the jth observation in the ith sample.

    idist : int
        The error distribution to be used in the analysis.

        `idist` = 1

            Normal.

        `idist` = 2

            Logistic.

        `idist` = 3

            Extreme value.

        `idist` = 4

            Double-exponential.

    nmax : int
        The value of the largest sample size.

    tol : float
        The tolerance for judging whether two observations are tied.
        Thus, observations Y_i and Y_j are adjudged to be tied if
        |Y_i-Y_j| < `tol`.

    Returns
    -------
    prvr : float, ndarray, shape (ip+1, ip)
        The variance-covariance matrices of the score statistics and the
        parameter estimates, the former being stored in the upper
        triangle and the latter in the lower triangle. Thus for 1 <= i
        <= j <= ip, `prvr`[i-1,j-1] contains an estimate of the
        covariance between the ith and jth score statistics. For 1 <= j
        <= i <= ip-1, `prvr`[i,j-1] contains an estimate of the
        covariance between the ith and jth parameter estimates.

    irank : int, ndarray, shape (`nmax`)
        For the one sample case, `irank` contains the ranks of the
        observations.

    zin : float, ndarray, shape (`nmax`)
        For the one sample case, `zin` contains the expected values of
        the function g(.) of the order statistics.

    eta : float, ndarray, shape (`nmax`)
        For the one sample case, `eta` contains the expected values of
        the function g'(.) of the order statistics.

    vapvec : float, ndarray, shape (`nmax`*(`nmax`+1)/2)
        For the one sample case, `vapvec` contains the upper triangle of
        the variance-covariance matrix of the function g(.) of the order
        statistics stored column-wise.

    parest : float, ndarray, shape (4*ip+1)
        The statistics calculated by the function.

        The first ip components of `parest` contain the score
        statistics.

        The next ip elements contain the parameter estimates.

        `parest`[2*ip] contains the value of the chi^2 statistic.

        The next ip elements of `parest` contain the standard errors of
        the parameter estimates.

        Finally, the remaining ip elements of `parest` contain the
        z-statistics.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, max_i(`nv`[i]) = *<value>* and `nmax` = *<value>*.

            Constraint: max_i(`nv`[i]) = `nmax`.

        (`errno` 1)
            On entry, sum_i(`nv`[i]) = *<value>* and nsum = *<value>*.

            Constraint: sum_i(`nv`[i]) = nsum.

        (`errno` 1)
            On entry, *<value>* elements of `nv` are < 1.

            Constraint: `nv`[i] >= 1.

        (`errno` 1)
            On entry, `nmax` = *<value>* and ip = *<value>*.

            Constraint: `nmax` > ip.

        (`errno` 1)
            On entry, ip = *<value>*.

            Constraint: ip >= 1.

        (`errno` 1)
            On entry, `tol` = *<value>*.

            Constraint: `tol` > 0.0.

        (`errno` 1)
            On entry, ns = *<value>*.

            Constraint: ns >= 1.

        (`errno` 2)
            On entry, `idist` = *<value>*.

            On entry, `idist` = 1, 2, 3 or 4.

        (`errno` 3)
            On entry, all the observations were adjudged to be tied.

        (`errno` 4)
            The matrix X^T(B-A)X is either ill-conditioned or not
            positive definite.

        (`errno` 5)
            On entry, for j = *<value>*, `x`[i,j-1] = *<value>* for all
            i.

            Constraint: `x`[i,j-1] != `x`[i+1,j-1] for at least one i.

    References
    ----------
    Pettitt, A N, 1982, `Inference for the linear model using a
    likelihood based on ranks`, J. Roy. Statist. Soc. Ser. B (44),
    234--243
    """

    fun_name = 'naginterfaces.library.nonpar.rank_regsn'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_x, 'x'),),
    )
    _nv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'nv'},
        nv,
        copy=True,
    )
    _ns_py = _b_utils._EngineIntArrayType.get_shape(_nv, {'fun_name': fun_name, 'entity': 'nv'}, exp_rank=1)[0]
    _ns = _b_utils._EngineIntScalarType(
        _ns_py, {'fun_name': fun_name, 'entity': '_ns'},
    ).to_ctypes()
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nsum_py = _b_utils._EngineFloat64ArrayType.get_shape(_y, {'fun_name': fun_name, 'entity': 'y'}, exp_rank=1)[0]
    _nsum = _b_utils._EngineIntScalarType(
        _nsum_py, {'fun_name': fun_name, 'entity': '_nsum'},
    ).to_ctypes()
    _ip_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2)[1]
    _ip = _b_utils._EngineIntScalarType(
        _ip_py, {'fun_name': fun_name, 'entity': '_ip'},
    ).to_ctypes()
    _prvr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ip_py+1, _ip_py),
        sorder=_m_sorder,
    )
    _nmax = _b_utils._EngineIntScalarType(
        nmax, {'fun_name': fun_name, 'entity': 'nmax'},
    ).to_ctypes()
    _nmax_py = _b_utils._EngineIntScalarType.to_py(_nmax)
    _irank = _b_utils._EngineIntArrayType.empty_ndarray(
        (_nmax_py,),
    )
    _zin = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nmax_py,),
    )
    _eta = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nmax_py,),
    )
    _vapvec = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nmax_py*(_nmax_py+1)//2,),
    )
    _parest = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (4*_ip_py+1,),
    )
    _b_nonpar.rank_regsn(
        _ns,
        _nv,
        _nsum,
        _y,
        _ip,
        _x,
        idist,
        _nmax,
        tol,
        _prvr,
        _irank,
        _zin,
        _eta,
        _vapvec,
        _parest,
    )
    _prvr_py = _prvr
    _irank_py = _irank
    _zin_py = _zin
    _eta_py = _eta
    _vapvec_py = _vapvec
    _parest_py = _parest
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['prvr', 'irank', 'zin', 'eta', 'vapvec', 'parest'],
    )(
        _prvr_py,
        _irank_py,
        _zin_py,
        _eta_py,
        _vapvec_py,
        _parest_py,
    )
    return _return_data

def rank_regsn_censored(nv, y, x, icen, gamma, nmax, tol):
    u"""
    Regression using ranks, right-censored data.

    ``rank_regsn_censored`` calculates the parameter estimates, score
    statistics and their variance-covariance matrices for the linear
    model using a likelihood based on the ranks of the observations when
    some of the observations may be right-censored.

    For full information please refer to the NAG Library document for
    g08rb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g08/g08rbf.html

    Parameters
    ----------
    nv : int, array-like, shape (ns)
        The number of observations in the ith sample, for i =
        1,2,...,ns.

    y : float, array-like, shape (sum(`nv`))
        The observations in each sample. Specifically, `y`[ sum_k =
        1^i-1`nv`[k-1]+j -1] must contain the jth observation in the ith
        sample.

    x : float, array-like, shape (nsum, ip)
        The design matrices for each sample. Specifically, `x`[ sum_k =
        1^i-1 `nv`[k-1] + j -1,l-1] must contain the value of the lth
        explanatory variable for the jth observations in the ith sample.

    icen : int, array-like, shape (sum(`nv`))
        Defines the censoring variable for the observations in `y`.

        `icen`[i-1] = 0

            If `y`[i-1] is uncensored.

        `icen`[i-1] = 1

            If `y`[i-1] is censored.

    gamma : float
        The value of the parameter defining the generalized logistic
        distribution. For `gamma` <= 0.0001, the limiting extreme value
        distribution is assumed.

    nmax : int
        The value of the largest sample size.

    tol : float
        The tolerance for judging whether two observations are tied.
        Thus, observations Y_i and Y_j are adjudged to be tied if
        |Y_i-Y_j| < `tol`.

    Returns
    -------
    prvr : float, ndarray, shape (ip+1, ip)
        The variance-covariance matrices of the score statistics and the
        parameter estimates, the former being stored in the upper
        triangle and the latter in the lower triangle. Thus for 1 <= i
        <= j <= ip, `prvr`[i-1,j-1] contains an estimate of the
        covariance between the ith and jth score statistics. For 1 <= j
        <= i <= ip-1, `prvr`[i,j-1] contains an estimate of the
        covariance between the ith and jth parameter estimates.

    irank : int, ndarray, shape (`nmax`)
        For the one sample case, `irank` contains the ranks of the
        observations.

    zin : float, ndarray, shape (`nmax`)
        For the one sample case, `zin` contains the expected values of
        the function g(.) of the order statistics.

    eta : float, ndarray, shape (`nmax`)
        For the one sample case, `eta` contains the expected values of
        the function g'(.) of the order statistics.

    vapvec : float, ndarray, shape (`nmax`*(`nmax`+1)/2)
        For the one sample case, `vapvec` contains the upper triangle of
        the variance-covariance matrix of the function g(.) of the order
        statistics stored column-wise.

    parest : float, ndarray, shape (4*ip+1)
        The statistics calculated by the function.

        The first ip components of `parest` contain the score
        statistics.

        The next ip elements contain the parameter estimates.

        `parest`[2*ip] contains the value of the chi^2 statistic.

        The next ip elements of `parest` contain the standard errors of
        the parameter estimates.

        Finally, the remaining ip elements of `parest` contain the
        z-statistics.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `gamma` = *<value>*.

            Constraint: `gamma` >= 0.0.

        (`errno` 1)
            On entry, `nmax` = *<value>* and ip = *<value>*.

            Constraint: `nmax` > ip.

        (`errno` 1)
            On entry, ip = *<value>*.

            Constraint: ip >= 1.

        (`errno` 1)
            On entry, `tol` = *<value>*.

            Constraint: `tol` > 0.0.

        (`errno` 1)
            On entry, ns = *<value>*.

            Constraint: ns >= 1.

        (`errno` 1)
            On entry, max_i(`nv`[i]) = *<value>* and `nmax` = *<value>*.

            Constraint: max_i(`nv`[i]) = `nmax`.

        (`errno` 1)
            On entry, sum_i(`nv`[i]) = *<value>* and nsum = *<value>*.

            Constraint: sum_i(`nv`[i]) = nsum.

        (`errno` 1)
            On entry, *<value>* elements of `nv` are < 1.

            Constraint: `nv`[i] >= 1.

        (`errno` 2)
            On entry, *<value>* elements of `icen` are out of range.

            Constraint: `icen`[i] = 0 or 1, for all i.

        (`errno` 3)
            On entry, all the observations were adjudged to be tied.

        (`errno` 4)
            The matrix X^T(B-A)X is either ill-conditioned or not
            positive definite.

        (`errno` 5)
            On entry, for j = *<value>*, `x`[i,j-1] = *<value>* for all
            i.

            Constraint: `x`[i,j-1] != `x`[i+1,j-1] for at least one i.

    References
    ----------
    Kalbfleisch, J D and Prentice, R L, 1980, `The Statistical Analysis
    of Failure Time Data`, Wiley

    Pettitt, A N, 1982, `Inference for the linear model using a
    likelihood based on ranks`, J. Roy. Statist. Soc. Ser. B (44),
    234--243

    Pettitt, A N, 1983, `Approximate methods using ranks for regression
    with censored data`, Biometrika (70), 121--132
    """

    fun_name = 'naginterfaces.library.nonpar.rank_regsn_censored'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_x, 'x'),),
    )
    _nv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'nv'},
        nv,
        copy=False,
    )
    _ns_py = _b_utils._EngineIntArrayType.get_shape(_nv, {'fun_name': fun_name, 'entity': 'nv'}, exp_rank=1)[0]
    _ns = _b_utils._EngineIntScalarType(
        _ns_py, {'fun_name': fun_name, 'entity': '_ns'},
    ).to_ctypes()
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nsum_py = _b_utils._EngineFloat64ArrayType.get_shape(_y, {'fun_name': fun_name, 'entity': 'y'}, exp_rank=1)[0]
    _nsum = _b_utils._EngineIntScalarType(
        _nsum_py, {'fun_name': fun_name, 'entity': '_nsum'},
    ).to_ctypes()
    _ip_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2)[1]
    _ip = _b_utils._EngineIntScalarType(
        _ip_py, {'fun_name': fun_name, 'entity': '_ip'},
    ).to_ctypes()
    _icen = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'icen'},
        icen,
        copy=False,
    )
    _prvr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ip_py+1, _ip_py),
        sorder=_m_sorder,
    )
    _nmax = _b_utils._EngineIntScalarType(
        nmax, {'fun_name': fun_name, 'entity': 'nmax'},
    ).to_ctypes()
    _nmax_py = _b_utils._EngineIntScalarType.to_py(_nmax)
    _irank = _b_utils._EngineIntArrayType.empty_ndarray(
        (_nmax_py,),
    )
    _zin = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nmax_py,),
    )
    _eta = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nmax_py,),
    )
    _vapvec = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nmax_py*(_nmax_py+1)//2,),
    )
    _parest = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (4*_ip_py+1,),
    )
    _b_nonpar.rank_regsn_censored(
        _ns,
        _nv,
        _nsum,
        _y,
        _ip,
        _x,
        _icen,
        gamma,
        _nmax,
        tol,
        _prvr,
        _irank,
        _zin,
        _eta,
        _vapvec,
        _parest,
    )
    _prvr_py = _prvr
    _irank_py = _irank
    _zin_py = _zin
    _eta_py = _eta
    _vapvec_py = _vapvec
    _parest_py = _parest
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['prvr', 'irank', 'zin', 'eta', 'vapvec', 'parest'],
    )(
        _prvr_py,
        _irank_py,
        _zin_py,
        _eta_py,
        _vapvec_py,
        _parest_py,
    )
    return _return_data
