# -*- coding: utf-8 -*-
# pylint: disable=chained-comparison,invalid-name,line-too-long,protected-access,too-many-arguments,too-many-boolean-expressions,too-many-branches,too-many-lines,too-many-locals,too-many-statements,unused-import
u"""
Interfaces for the NAG Mark 27.0 `roots` Chapter.

``roots`` - Roots of One or More Transcendental Equations

This module is concerned with the calculation of zeros of continuous
functions of one or more variables.
The majority of problems considered are for real-valued functions of
real variables, in which case complex equations must be expressed in
terms of the equivalent larger system of real equations.

Background
----------
The module divides naturally into two parts.

A Single Equation
~~~~~~~~~~~~~~~~~
The first deals with the real zeros of a real function of a single
variable f(x).

There are three functions with simple calling sequences.
The first assumes that you can determine an initial interval [a,b]
within which the desired zero lies, (that is, where f(a)*f(b) < 0), and
outside which all other zeros lie.
The function then systematically subdivides the interval to produce a
final interval containing the zero.
This final interval has a length bounded by your specified error
requirements; the end of the interval where the function has smallest
magnitude is returned as the zero.
This function is guaranteed to converge to a **simple** zero of the
function. (Here we define a simple zero as a zero corresponding to a
sign-change of the function; none of the available functions are capable
of making any finer distinction.) However, as with the other functions
described below, a non-simple zero might be determined and it is left to
you to check for this.
The algorithm used is due to Brent (1973).

The two other functions are both designed for the case where you are
unable to specify an interval containing the simple zero.
One starts from an initial point and performs a search for an interval
containing a simple zero.
If such an interval is computed then the method described above is used
next to determine the zero accurately.
The other method uses a 'continuation' method based on a secant
iteration.
A sequence of subproblems is solved; the first of these is trivial and
the last is the actual problem of finding a zero of f(x).
The intermediate problems employ the solutions of earlier problems to
provide initial guesses for the secant iterations used to calculate
their solutions.

Three other functions are also supplied.
They employ reverse communication and use the same core algorithms as
the functions described above.

Finally, two functions are provided to return values of Lambert's W
function (sometimes known as the 'product log' or 'Omega' function),
which is the inverse function of

    f(w) = we^w for w in C;

that is, if Lambert's W function W(x) = a for x,a in C, then a is a zero
of the function F(w) = we^w-x.
One function uses the iterative method described in Barry `et al.`
(1995) to return values from the real branches of W (restricting x,a in
R).
The second function enforces no such restriction, and uses the approach
described in Corless `et al.` (1996).

Systems of Equations
~~~~~~~~~~~~~~~~~~~~
The functions in the second part of this module are designed to solve a
set of nonlinear equations in n unknowns

    f_i(x) = 0, i = 1,2,...,n, x = (x_1,x_2,...,x_n)^T,

where T stands for transpose.

It is assumed that the functions are continuous and differentiable so
that the matrix of first partial derivatives of the functions, the
**Jacobian** matrix J_{ij}(x) = (df_i/dx_j) evaluated at the point x,
exists, though it may not be possible to calculate it directly.

The functions f_i must be independent, otherwise there will be an
infinity of solutions and the methods will fail.
However, even when the functions are independent the solutions may not
be unique.
Since the methods are iterative, an initial guess at the solution has to
be supplied, and the solution located will usually be the one closest to
this initial guess.

Available Routines
------------------
Zeros of Functions of One Variable
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
The functions can be divided into two classes.
There are three functions (``contfn_interval_rcomm``,
``contfn_cntin_rcomm`` and ``contfn_brent_rcomm``) all written in
reverse communication form and three (``contfn_brent_interval``,
``contfn_cntin`` and ``contfn_brent``) written in direct communication
form (see [Direct and Reverse Communication Routines] for a description
of the difference between these two conventions).
The direct communication functions are designed for inexperienced users
and, in particular, for solving problems where the function f(x) whose
zero is to be calculated, can be coded as a callback function.
These functions find the zero by using the same core algorithms as the
reverse communication functions.
Experienced users are recommended to use the reverse communication
functions directly as they permit you more control of the calculation.
Indeed, if the zero-finding process is embedded in a much larger program
then the reverse communication functions should always be used.

The recommendation as to which function should be used depends mainly on
whether you can supply an interval [a,b] containing the zero; that is,
where f(a)*f(b) < 0.
If the interval can be supplied, then ``contfn_brent`` (or, in reverse
communication, ``contfn_brent_rcomm``) should be used, in general.
This recommendation should be qualified in the case when the only
interval which can be supplied is very long relative to your error
requirements **and** you can also supply a good approximation to the
zero.
In this case ``contfn_cntin`` (or, in reverse communication,
``contfn_cntin_rcomm``) **may** prove more efficient (though these
latter functions will not provide the error bound available from
``contfn_brent_rcomm``).

If an interval containing the zero cannot be supplied then you must
choose between ``contfn_brent_interval`` (or, in reverse communication,
``contfn_interval_rcomm`` followed by ``contfn_brent_rcomm``) and
``contfn_cntin`` (or, in reverse communication, ``contfn_cntin_rcomm``).
``contfn_brent_interval`` first determines an interval containing the
zero, and then proceeds as in ``contfn_brent``; it is particularly
recommended when you do not have a good initial approximation to the
zero.
If a good initial approximation to the zero is available then
``contfn_cntin`` is to be preferred.
Since neither of these latter functions has guaranteed convergence to
the zero, you are recommended to experiment with both in case of
difficulty.

Solution of Sets of Nonlinear Equations
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
The solution of a set of nonlinear equations

    f_i(x_1,x_2,...,x_n) = 0, i = 1,2,...,n

can be regarded as a special case of the problem of finding a minimum of
a sum of squares

    s(x) = sum_i = 1^m[f_i(x_1,x_2,...,x_n)]^2, (m >= n).

So the functions in submodule ``opt`` are relevant as well as the
special nonlinear equations functions.

The functions for solving a set of nonlinear equations can also be
divided into classes.
There are six functions (``sys_func_aa``, ``sys_func_easy``,
``sys_func_expert``, ``sparsys_func_easy``, ``sys_deriv_easy`` and
``sys_deriv_expert``) all written in direct communication form and three
(``sys_func_aa_rcomm``, ``sys_func_rcomm`` and ``sys_deriv_rcomm``)
written in reverse communication form.
The direct communication functions are designed for inexperienced users
and, in particular, these functions require the f_i (and possibly their
derivatives) to be calculated in functions.
These should be set up carefully so the NAG Library functions can work
as efficiently as possible.
Experienced users are recommended to use the reverse communication
functions as they permit you more control of the calculation.
Indeed, if the zero-finding process is embedded in a much larger program
then the reverse communication functions should always be used.

The main decision you have to make is whether to supply the derivatives
df_i/dx_j.
It is advisable to do so if possible, since the results obtained by
algorithms which use derivatives are generally more reliable than those
obtained by algorithms which do not use derivatives.

``sys_deriv_easy``, ``sys_deriv_expert`` and ``sys_deriv_rcomm`` require
you to provide the derivatives, whilst ``sys_func_aa``,
``sys_func_aa_rcomm``, ``sys_func_easy``, ``sys_func_expert``,
``sys_func_rcomm`` and ``sparsys_func_easy`` do not. ``sys_func_easy``,
``sys_deriv_easy`` and ``sparsys_func_easy`` are easy-to-use functions;
greater flexibility may be obtained using ``sys_func_expert`` and
``sys_deriv_expert`` (or, in reverse communication, ``sys_func_rcomm``
and ``sys_deriv_rcomm``), but these have longer argument lists.
``sys_func_easy``, ``sys_func_expert``, ``sys_func_rcomm`` and
``sparsys_func_easy`` approximate the derivatives internally using
finite differences. ``sys_func_aa`` and ``sys_func_aa_rcomm`` do not use
derivatives at all, and may be useful when the cost of evaluating f(x)
is high.

``sparsys_func_easy`` is an easy-to-use function specially adapted for
sparse problems, that is, problems where each function depends on a
small subset of the n variables so that the Jacobian matrix has many
zeros.
It employs sparse linear algebra methods and consequently is expected to
take significantly less time to complete than the other functions,
especially if n is large.

``sys_deriv_check`` is provided for use in conjunction with
``sys_deriv_easy``, ``sys_deriv_expert`` and ``sys_deriv_rcomm`` to
check the user-supplied derivatives for consistency with the functions
themselves.
You are strongly advised to make use of this function whenever
``sys_deriv_easy``, ``sys_deriv_expert`` or ``sys_deriv_rcomm`` is used.

Firstly, the calculation of the functions and their derivatives should
be ordered so that **cancellation errors** are avoided.
This is particularly important in a function that uses these quantities
to build up estimates of higher derivatives.

Secondly, **scaling** of the variables has a considerable effect on the
efficiency of a function.
The problem should be designed so that the elements of x are of similar
magnitude.
The same comment applies to the functions, i.e., all the f_i should be
of comparable size.

The accuracy is usually determined by the accuracy arguments of the
functions, but the following points may be useful.

(i) Greater accuracy in the solution may be requested by choosing
    smaller input values for the accuracy arguments. However, if
    unreasonable accuracy is demanded, rounding errors may become
    important and cause a failure.

(#) Some idea of the accuracies of the x_i may be obtained by monitoring
    the progress of the function to see how many figures remain
    unchanged during the last few iterations.

(#) An approximation to the error in the solution x is given by e where
    e is the solution to the set of linear equations

        J(x)e = -f(x)

    where f(x) = (f_1(x),f_2(x),...,f_n(x))^T.

    Note that the QR decomposition of J is available from
    ``sys_func_expert`` and ``sys_deriv_expert`` (or, in reverse
    communication, ``sys_func_rcomm`` and ``sys_deriv_rcomm``) so that

        Re = -Q^Tf

    and Q^Tf is also provided by these functions.

(#) If the functions f_i(x) are changed by small amounts epsilon_i, for
    i = 1,2,...,n, then the corresponding change in the solution x is
    given approximately by sigma, where sigma is the solution of the set
    of linear equations

        J(x)sigma = -epsilon.

    Thus one can estimate the sensitivity of x to any uncertainties in
    the specification of f_i(x), for i = 1,2,...,n.
    As noted above, the sophisticated functions ``sys_func_expert`` and
    ``sys_deriv_expert`` (or, in reverse communication,
    ``sys_func_rcomm`` and ``sys_deriv_rcomm``) provide the QR
    decomposition of J.

Values of Lambert's W function
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
If you require purely-real values of W, these will be evaluated
marginally more efficiently by ``lambertw_real`` than
``lambertw_complex`` owing to the differing iterative procedures used by
each function.

Tutorial
--------
Zeros of Functions of One Variable
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Consider the problem of finding x such that x = exp(-x); i.e., the value
of x that gives f(x) = 0 when f(x) = x-exp(-x).

>>> from math import exp
>>> f = lambda x: x - exp(-x)

All of the NAG zero-finding routines require you to have some basic
knowledge of the structure of the function you are addressing in order
that the solver can know where to begin.
This knowledge must either be in the form of a rough initial guess for
the zero, or as an interval on which f changes sign (and therefore
passes through zero).

In the worst case a sign-change interval is not known and the initial
guess for the zero may be fairly poor:

>>> x = 1.

If so, use ``contfn_brent_interval``:

>>> from naginterfaces.library import roots
>>> soln = roots.contfn_brent_interval(x, f)

Upon successful exit soln contains the computed zero and the bounding
interval that was used internally:

>>> print('Final value of x is {:1.8f}.'.format(soln.x))
Final value of x is 0.56714329.
>>> print(
...     'Search interval was [' +
...     ', '.join(['{:1.8f}']*2).format(soln.a, soln.b) +
...     '].'
... )
Search interval was [0.50000000, 0.90000000].

For an alternative approach that uses continuation, more suitable for
when a better initial guess is known, use ``contfn_cntin``:

>>> x = 0.6
>>> soln = roots.contfn_cntin(x, f)

The continuation algorithm returns only the final x:

>>> print('Final value of x is {:1.8f}.'.format(soln))
Final value of x is 0.56714329.

Solution of Sets of Nonlinear Equations
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Consider the following tridiagonal system of nine equations in nine
variables:

    [table omitted]

as a vector function

>>> def f(x):
...     fvec = (3. - 2.*x)*x + 1.
...     fvec[1:] = fvec[1:] - x[:n-1]
...     fvec[:n-1] = fvec[:n-1] - 2.*x[1:]
...     return fvec

for which we seek x such that f(x) = 0.

As with scalar root-finding above, the NAG solver needs you to supply a
rough initial guess for x.

>>> import numpy as np
>>> n = 9; x = -1.*np.ones(n)

To solve a system of nonlinear equations, again the NAG routines can
benefit from your knowledge of the structure of the problem as outlined
above.

If you are happy for the Jacobian of the system to be approximated
internally, use, say, ``sys_func_easy``:

>>> from naginterfaces.library import roots
>>> soln = roots.sys_func_easy(f, x)
>>> def show_soln():
...     print(
...         'Final approximate solution is\\n'
...         '(\\n' +
...         '\\n'.join(['  {:.4f},']*n).format(*soln.x) +
...         '\\n)'
...     )
>>> show_soln()
Final approximate solution is
(
  -0.5707,
  -0.6816,
  -0.7017,
  -0.7042,
  -0.7014,
  -0.6919,
  -0.6658,
  -0.5960,
  -0.4164,
)

The exact Jacobian of the system in question can be represented using
the following Python function:

>>> import numpy as np
>>> def fjac(x):
...     fj = np.zeros((len(x), len(x)))
...     fj[0, 0] = 3. - 4.*x[0]
...     fj[0, 1] = -2.
...     for i in range(1, n-2):
...         fj[i, i-1] = -1.
...         fj[i, i] = 3. - 4.*x[i]
...         fj[i, i+1] = -2.
...     fj[n-1, n-2] = -1.
...     fj[n-1, n-1] = 3. - 4.*x[n-1]
...     return fj

Function ``sys_deriv_easy`` can be used to take advantage of these exact
derivatives:

>>> from naginterfaces.library import roots
>>> soln = roots.sys_deriv_easy(f, fjac, x)
>>> show_soln()
Final approximate solution is
(
  -0.5707,
  -0.6816,
  -0.7017,
  -0.7042,
  -0.7014,
  -0.6919,
  -0.6658,
  -0.5960,
  -0.4164,
)

When the Jacobian is known to contain a large number of zero entries,
but isn't available in an exact form, ``sparsys_func_easy`` may be
appropriate.
With this solver we are asked to compute values of the system for a
supplied base-1 index mask. (In the contrived example below we use
precomputed 'full' values for the system, using our definition of f
above, to populate the requested entries of the return vector.
For a real-world sparse problem where memory usage is a concern this
would clearly be counterproductive.) The function vector is updated in
place in the callback, which further optimizes memory usage.

>>> def sparsf(indf, x, fvec):
...     fv = f(x)
...     for ind in indf:
...        fvec[ind-1] = fv[ind-1]

For our system the number of nonzeros is n+2(n-1).
Communicating this in the NAG call can reduce memory requirements.

>>> from naginterfaces.library import roots
>>> soln = roots.sparsys_func_easy(sparsf, x, nnz=3*n-2)
>>> show_soln()
Final approximate solution is
(
  -0.5707,
  -0.6816,
  -0.7017,
  -0.7042,
  -0.7014,
  -0.6919,
  -0.6658,
  -0.5960,
  -0.4164,
)

Computational efficiency may also be enhanced by using the fixed-point
iteration routine ``sys_func_aa`` and its Anderson acceleration
facility.
The potential downsides are that convergence is less reliable than with
the Powell hybrid method used in the solvers shown previously, and that
the system may need to be rearranged into the form g(x)-x = 0 for best
performance:

>>> import numpy as np
>>> def fixedf(x):
...     fvec = np.zeros(n)
...     fvec[0] = (2.*x[1] - 1.)/(3. - 2.*x[0])
...     for i in range(1, n-1):
...         fvec[i] = (2.*x[i+1] - 1. + x[i-1])/(3. - 2.*x[i])
...     fvec[n-1] = (x[n-2] - 1.)/(3. - 2.*x[n-1])
...     return fvec - x

The size of the iteration subspace cache to use is supplied via argument
m:

>>> from naginterfaces.library import roots
>>> soln = roots.sys_func_aa(fixedf, x, m=4)
>>> show_soln()
Final approximate solution is
(
  -0.5707,
  -0.6816,
  -0.7017,
  -0.7042,
  -0.7014,
  -0.6919,
  -0.6658,
  -0.5960,
  -0.4164,
)

For full information please refer to the NAG Library document

https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c05/c05intro.html

See Also
--------
``naginterfaces.library.examples.roots`` :
    This subpackage contains examples for the ``roots`` module.
    See also the :ref:`library_roots_ex` subsection.
"""

# NAG Copyright 2017-2020.

from __future__ import absolute_import

import collections as _collections

from math import sqrt

from ..base import roots as _b_roots
from ..base import utils as _b_utils
from ..library import machine
from ..library import _utils

def contfn_brent_interval(x, f, h=0.1, eps=1000.0*machine.precision(), eta=0.0, data=None):
    u"""
    Zero of continuous function, Brent algorithm, from a given starting
    value, binary search for interval.

    ``contfn_brent_interval`` locates a simple zero of a continuous
    function from a given starting value.
    It uses a binary search to locate an interval containing a zero of
    the function, then Brent's method, which is a combination of
    nonlinear interpolation, linear extrapolation and bisection, to
    locate the zero precisely.

    For full information please refer to the NAG Library document for
    c05au

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c05/c05auf.html

    Parameters
    ----------
    x : float
        An initial approximation to the zero.

    f : callable retval = f(x, data=None)
        `f` must evaluate the function f whose zero is to be determined.

        Parameters
        ~~~~~~~~~~
        x : float
            The point at which the function must be evaluated.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        retval : float
            The value of f at the specified point.

    h : float, optional
        A step length for use in the binary search for an interval
        containing the zero. The maximum interval searched is
        [`x`-256.0*`h`,`x`+256.0*`h`].

    eps : float, optional
        The termination tolerance on x (see [Description]).

    eta : float, optional
        A value such that if |f(x)| <= `eta`, x is accepted as the zero.
        `eta` may be specified as 0.0 (see [Accuracy]).

    data : arbitrary, optional
        User-communication data for callback functions.

    Returns
    -------
    x : float
        If the function exits successfully or `errno` = 4, `x` is the
        final approximation to the zero.

        If `errno` = 3, `x` is likely to be a pole of f(x).

        Otherwise, `x` contains no useful information.

    a : float
        The lower bound of the interval resulting from the binary search

    b : float
        The upper bound of the interval resulting from the binary search

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `eps` = *<value>*.

            Constraint: `eps` > 0.0.

        (`errno` 1)
            On entry, `x` = *<value>* and `h` = *<value>*.

            Constraint: `x`+`h` != `x` (to machine accuracy).

        (`errno` 2)
            An interval containing the zero could not be found.
            Increasing `h` and calling ``contfn_brent_interval`` again
            will increase the range searched for the zero. Decreasing
            `h` and calling ``contfn_brent_interval`` again will refine
            the mesh used in the search for the zero.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            Solution may be a pole rather than a zero.

        (`errno` 4)
            The tolerance `eps` has been set too small for the problem
            being solved. However, the value `x` returned is a good
            approximation to the zero. `eps` = *<value>*.

    References
    ----------
    Brent, R P, 1973, `Algorithms for Minimization Without Derivatives`,
    Prentice--Hall

    See Also
    --------
    :meth:`naginterfaces.library.examples.roots.contfn_brent_interval_ex.main`
    """

    def cb_f(x, data=None):
        """Wrapping user's f. Called as base f."""
        if data is not None:
            _retval_f = f(
                x,
                data,
            )
        else:
            _retval_f = f(
                x,
            )
        return _retval_f

    fun_name = 'naginterfaces.library.roots.contfn_brent_interval'

    _b_utils._check_callable(f, {'fun_name': fun_name, 'entity': 'f'})
    _f = cb_f
    try:
        _return_scalars = _b_roots.contfn_brent_interval(
            x,
            h,
            eps,
            eta,
            _f,
            data,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _x_py = _return_scalars.x
    _a_py = _return_scalars.a
    _b_py = _return_scalars.b
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'a', 'b'],
    )(
        _x_py,
        _a_py,
        _b_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def contfn_interval_rcomm(x, fx, h, boundl, boundu, y, comm, ind):
    u"""
    Binary search for interval containing zero of continuous function
    (reverse communication).

    ``contfn_interval_rcomm`` attempts to locate an interval containing
    a simple zero of a continuous function using a binary search.
    It uses reverse communication for evaluating the function.

    For full information please refer to the NAG Library document for
    c05av

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c05/c05avf.html

    Parameters
    ----------
    x : float
        `On initial entry`: the best available approximation to the
        zero.

    fx : float
        `On initial entry`: if `ind` = 1, `fx` need not be set.

        If `ind` = -1, `fx` must contain f(`x`) for the initial value of
        `x`.

        `On intermediate entry`: must contain f(`x`) for the current
        value of `x`.

    h : float
        `On initial entry`: a basic step size which is used in the
        binary search for an interval containing a zero. The basic step
        sizes `h`,0.1*`h`, 0.01*`h` and 0.001*`h` are used in turn when
        searching for the zero.

    boundl : float
        The lower bound for the interval of search for the zero.

    boundu : float
        The upper bound for the interval of search for the zero.

    y : float
        `On initial entry`: need not be set.

    comm : dict, communication object, modified in place
        Communication structure.

        `On initial entry`: need not be set.

    ind : int
        `On initial entry`: must be set to 1 or -1.

        `ind` = 1

            `fx` need not be set.

        `ind` = -1

            `fx` must contain f(`x`).

    Returns
    -------
    x : float
        `On intermediate exit`: contains the point at which f must be
        evaluated before re-entry to the function.

        `On final exit`: contains one end of an interval containing the
        zero, the other end being in `y`, unless an error has occurred.
        If `errno` = 4, `x` and `y` are the end points of the largest
        interval searched. If a zero is located exactly, its value is
        returned in `x` (and in `y`).

    h : float
        `On final exit`: is undefined.

    y : float
        `On final exit`: contains the closest point found to the final
        value of `x`, such that f(`x`)*f(`y`) <= 0.0. If a value `x` is
        found such that f(`x`) = 0, `y` = `x`. On final exit with
        `errno` = 4, `x` and `y` are the end points of the largest
        interval searched.

    c1 : float
        `On final exit`: if the function exits successfully or `errno` =
        4, `c1` is set to f(`y`).

    ind : int
        `On intermediate exit`: contains 2 or 3. The calling program
        must evaluate f at `x`, storing the result in `fx`, and re-enter
        ``contfn_interval_rcomm`` with all other arguments unchanged.

        `On final exit`: contains 0.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `boundl` = *<value>* and `boundu` = *<value>*.

            Constraint: `boundl` < `boundu`.

        (`errno` 1)
            On entry, `x` = *<value>*, `boundl` = *<value>* and `boundu`
            = *<value>*.

            Constraint: `boundl` <= `x` <= `boundu`.

        (`errno` 1)
            On entry, `x` = *<value>*, `h` = *<value>*, `boundl` =
            *<value>* and `boundu` = *<value>*.

            Constraint: either `x`+`h` or `x`-`h` must lie inside the
            closed interval [`boundl`,`boundu`].

        (`errno` 2)
            On entry, `x` = *<value>* and `h` = *<value>*.

            Constraint: `h` must be sufficiently large that `x`+`h` !=
            `x` on the computer.

        (`errno` 3)
            On entry, `ind` = *<value>*.

            Constraint: `ind` = -1, 1, 2 or 3.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 4)
            An interval containing the zero could not be found. Try
            restarting with modified `x` and `h`.
    """

    try:
        _return_scalars = _b_roots.contfn_interval_rcomm(
            x,
            fx,
            h,
            boundl,
            boundu,
            y,
            comm,
            ind,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _x_py = _return_scalars.x
    _h_py = _return_scalars.h
    _y_py = _return_scalars.y
    _ind_py = _return_scalars.ind
    _c1_py = _return_scalars.c1
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'h', 'y', 'c1', 'ind'],
    )(
        _x_py,
        _h_py,
        _y_py,
        _c1_py,
        _ind_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def contfn_cntin(x, f, eps=1000.0*machine.precision(), eta=0.0, nfmax=1000, data=None):
    u"""
    Zero of continuous function, continuation method, from a given
    starting value.

    ``contfn_cntin`` attempts to locate a zero of a continuous function
    using a continuation method based on a secant iteration.

    For full information please refer to the NAG Library document for
    c05aw

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c05/c05awf.html

    Parameters
    ----------
    x : float
        An initial approximation to the zero.

    f : callable retval = f(x, data=None)
        `f` must evaluate the function f whose zero is to be determined.

        Parameters
        ~~~~~~~~~~
        x : float
            The point at which the function must be evaluated.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        retval : float
            The value of f evaluated at `x`.

    eps : float, optional
        An absolute tolerance to control the accuracy to which the zero
        is determined. In general, the smaller the value of `eps` the
        more accurate `x` will be as an approximation to alpha. Indeed,
        for very small positive values of `eps`, it is likely that the
        final approximation will satisfy |`x`-alpha| < `eps`. You are
        advised to call the function with more than one value for `eps`
        to check the accuracy obtained.

    eta : float, optional
        A value such that if |f(x)| < `eta`, x is accepted as the zero.
        `eta` may be specified as 0.0 (see [Accuracy]).

    nfmax : int, optional
        The maximum permitted number of calls to `f` from
        ``contfn_cntin``. If `f` is inexpensive to evaluate, `nfmax`
        should be given a large value (say > 1000).

    data : arbitrary, optional
        User-communication data for callback functions.

    Returns
    -------
    x : float
        The final approximation to the zero, unless `errno` = 1, 2 or 5,
        in which case it contains no useful information.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `nfmax` = *<value>*.

            Constraint: `nfmax` > 0.

        (`errno` 1)
            On entry, `eps` = *<value>*.

            Constraint: `eps` > 0.0.

        (`errno` 2)
            Internal scale factor invalid for this problem. Consider
            using ``contfn_cntin_rcomm`` instead and setting scal.

        (`errno` 3)
            Either `f` has no zero near `x` or too much accuracy has
            been requested. Check the coding of `f` or increase `eps`.

        (`errno` 4)
            More than `nfmax` calls have been made to `f`.

        (`errno` 5)
            A serious error occurred in an internal call to an auxiliary
            function.
    """

    def cb_f(x, data=None):
        """Wrapping user's f. Called as base f."""
        if data is not None:
            _retval_f = f(
                x,
                data,
            )
        else:
            _retval_f = f(
                x,
            )
        return _retval_f

    fun_name = 'naginterfaces.library.roots.contfn_cntin'

    _b_utils._check_callable(f, {'fun_name': fun_name, 'entity': 'f'})
    _f = cb_f
    _return_scalar = _b_roots.contfn_cntin(
        x,
        eps,
        eta,
        _f,
        nfmax,
        data,
    )
    _x_py = _return_scalar
    return _x_py

def contfn_cntin_rcomm(x, fx, tol, ir, comm, ind, scal=sqrt(machine.precision())):
    u"""
    Zero of continuous function, continuation method, from a given
    starting value (reverse communication).

    ``contfn_cntin_rcomm`` attempts to locate a zero of a continuous
    function using a continuation method based on a secant iteration.
    It uses reverse communication for evaluating the function.

    For full information please refer to the NAG Library document for
    c05ax

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c05/c05axf.html

    Parameters
    ----------
    x : float
        `On initial entry`: an initial approximation to the zero.

    fx : float
        `On initial entry`: if `ind` = 1, `fx` need not be set.

        If `ind` = -1, `fx` must contain f(`x`) for the initial value of
        `x`.

        `On intermediate entry`: must contain f(`x`) for the current
        value of `x`.

    tol : float
        `On initial entry`: a value that controls the accuracy to which
        the zero is determined. `tol` is used in determining the
        convergence of the secant iteration used at each stage of the
        continuation process. It is used directly when solving the last
        problem (theta_m = 0 in [Description]), and sqrt(`tol`) is used
        for the problem defined by theta_r, r < m. Convergence to the
        accuracy specified by `tol` is not guaranteed, and so you are
        recommended to find the zero using at least two values for `tol`
        to check the accuracy obtained.

    ir : int
        `On initial entry`: indicates the type of error test required,
        as follows. Solving the problem defined by theta_r, 1 <= r <= m,
        involves computing a sequence of secant iterates
        x_r^0,x_r^1,.... This sequence will be considered to have
        converged only if:

        for `ir` = 0,

            |x_r^(i+1)-x_r^(i)| <= eps*max(1.0, |x_r^(i)|),

        for `ir` = 1,

            |x_r^(i+1)-x_r^(i)| <= eps,

        for `ir` = 2,

            |x_r^(i+1)-x_r^(i)| <= eps*|x_r^(i)|,

        for some i > 1; here eps is either `tol` or sqrt(`tol`) as
        discussed above.

        Note that there are other subsidiary conditions (not given here)
        which must also be satisfied before the secant iteration is
        considered to have converged.

    comm : dict, communication object, modified in place
        Communication structure.

        `On initial entry`: need not be set.

    ind : int
        `On initial entry`: must be set to 1 or -1.

        `ind` = 1

            `fx` need not be set.

        `ind` = -1

            `fx` must contain f(`x`).

    scal : float, optional
        `On initial entry`: a factor for use in determining a
        significant approximation to the derivative of f(x) at x = x_0,
        the initial value. A number of difference approximations to
        f'(x_0) are calculated using

            f'(x_0)~(f(x_0+h)-f(x_0))/h

        where |h| < |`scal`| and h has the same sign as `scal`. A
        significance (cancellation) check is made on each difference
        approximation and the approximation is rejected if
        insignificant.

    Returns
    -------
    x : float
        `On intermediate exit`: the point at which f must be evaluated
        before re-entry to the function.

        `On final exit`: the final approximation to the zero.

    ind : int
        `On intermediate exit`: contains 2, 3 or 4. The calling program
        must evaluate f at `x`, storing the result in `fx`, and re-enter
        ``contfn_cntin_rcomm`` with all other arguments unchanged.

        `On final exit`: contains 0.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `tol` = *<value>*.

            Constraint: `tol` > 0.0.

        (`errno` 1)
            On entry, `ir` = *<value>*.

            Constraint: `ir` = 0, 1 or 2.

        (`errno` 2)
            On initial entry, `ind` = *<value>*.

            Constraint: `ind` = -1 or 1.

        (`errno` 2)
            On intermediate entry, `ind` = *<value>*.

            Constraint: `ind` = 2, 3 or 4.

        (`errno` 3)
            On entry, `scal` = *<value>*.

            Constraint: `x`+`scal` != `x` (to machine accuracy).

        (`errno` 3)
            Significant derivatives of f cannot be computed. This can
            happen when f is almost constant and nonzero, for any value
            of `scal`.

        (`errno` 4)
            Current problem in the continuation sequence cannot be
            solved. Perhaps the original problem had no solution or the
            continuation path passes through a set of insoluble
            problems: consider refining the initial approximation to the
            zero. Alternatively, `tol` is too small, and the accuracy
            requirement is too stringent, or too large and the initial
            approximation too poor.

        (`errno` 5)
            Continuation away from the initial point is not possible.
            This error exit will usually occur if the problem has not
            been properly posed or the error requirement is extremely
            stringent.

        (`errno` 6)
            Final problem (with theta_m = 0) cannot be solved. It is
            likely that too much accuracy has been requested, or that
            the zero is at alpha = 0 and `ir` = 2.

    References
    ----------
    Swift, A and Lindfield, G R, 1978, `Comparison of a continuation
    method for the numerical solution of a single nonlinear equation`,
    Comput. J. (21), 359--362
    """

    _return_scalars = _b_roots.contfn_cntin_rcomm(
        x,
        fx,
        tol,
        ir,
        scal,
        comm,
        ind,
    )
    _x_py = _return_scalars.x
    _ind_py = _return_scalars.ind
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'ind'],
    )(
        _x_py,
        _ind_py,
    )
    return _return_data

def contfn_brent(a, b, f, eps=1000.0*machine.precision(), eta=0.0, data=None):
    u"""
    Zero of continuous function in a given interval, Brent algorithm.

    ``contfn_brent`` locates a simple zero of a continuous function in a
    given interval using Brent's method, which is a combination of
    nonlinear interpolation, linear extrapolation and bisection.

    For full information please refer to the NAG Library document for
    c05ay

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c05/c05ayf.html

    Parameters
    ----------
    a : float
        a, the lower bound of the interval.

    b : float
        b, the upper bound of the interval.

    f : callable retval = f(x, data=None)
        `f` must evaluate the function f whose zero is to be determined.

        Parameters
        ~~~~~~~~~~
        x : float
            The point at which the function must be evaluated.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        retval : float
            The value of f evaluated at `x`.

    eps : float, optional
        The termination tolerance on x (see [Description]).

    eta : float, optional
        A value such that if |f(x)| <= `eta`, x is accepted as the zero.
        `eta` may be specified as 0.0 (see [Accuracy]).

    data : arbitrary, optional
        User-communication data for callback functions.

    Returns
    -------
    x : float
        If the function exits successfully or `errno` = 2, `x` is the
        final approximation to the zero. If `errno` = 3, `x` is likely
        to be a pole of f(x). Otherwise, `x` contains no useful
        information.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `eps` = *<value>*.

            Constraint: `eps` > 0.0.

        (`errno` 1)
            On entry, `a` = *<value>* and `b` = *<value>*.

            Constraint: `a` != `b`.

        (`errno` 1)
            On entry, `f`(`a`) and `f`(`b`) have the same sign with
            neither equalling 0.0: `f`(`a`) = *<value>* and `f`(`b`) =
            *<value>*.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            No further improvement in the solution is possible. `eps` is
            too small: `eps` = *<value>*. The final value of `x`
            returned is an accurate approximation to the zero.

        (`errno` 3)
            The function values in the interval [`a`,`b`] might contain
            a pole rather than a zero. Reducing `eps` may help in
            distinguishing between a pole and a zero.

    References
    ----------
    Brent, R P, 1973, `Algorithms for Minimization Without Derivatives`,
    Prentice--Hall
    """

    def cb_f(x, data=None):
        """Wrapping user's f. Called as base f."""
        if data is not None:
            _retval_f = f(
                x,
                data,
            )
        else:
            _retval_f = f(
                x,
            )
        return _retval_f

    fun_name = 'naginterfaces.library.roots.contfn_brent'

    _b_utils._check_callable(f, {'fun_name': fun_name, 'entity': 'f'})
    _f = cb_f
    try:
        _return_scalar = _b_roots.contfn_brent(
            a,
            b,
            eps,
            eta,
            _f,
            data,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _x_py = _return_scalar
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x'],
    )(
        _x_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _x_py

def contfn_brent_rcomm(x, y, fx, tolx, c, ind, ir=0):
    u"""
    Zero of continuous function in a given interval, Brent algorithm
    (reverse communication).

    ``contfn_brent_rcomm`` locates a simple zero of a continuous
    function in a given interval by using Brent's method, which is a
    combination of nonlinear interpolation, linear extrapolation and
    bisection.
    It uses reverse communication for evaluating the function.

    For full information please refer to the NAG Library document for
    c05az

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c05/c05azf.html

    Parameters
    ----------
    x : float
        `On initial entry`: `x` and `y` must define an initial interval
        [a,b] containing the zero, such that f(`x`)*f(`y`) <= 0.0. It is
        not necessary that `x` < `y`. Otherwise, `x` and `y` must be the
        same entities output by a previous call to
        ``contfn_brent_rcomm``.

    y : float
        `On initial entry`: `x` and `y` must define an initial interval
        [a,b] containing the zero, such that f(`x`)*f(`y`) <= 0.0. It is
        not necessary that `x` < `y`. Otherwise, `x` and `y` must be the
        same entities output by a previous call to
        ``contfn_brent_rcomm``.

    fx : float
        `On initial entry`: if `ind` = 1, `fx` need not be set.

        If `ind` = -1, `fx` must contain f(`x`) for the initial value of
        `x`.

        `On intermediate entry`: must contain f(`x`) for the current
        value of `x`.

    tolx : float
        `On initial entry`: the accuracy to which the zero is required.
        The type of error test is specified by `ir`.

    c : float, ndarray, shape (17), modified in place
        `On initial entry`: if `ind` = 1, no elements of `c` need be
        set.

        If `ind` = -1, `c`[0] must contain f(`y`), other elements of `c`
        need not be set.

        `On final exit`: is undefined.

    ind : int
        `On initial entry`: must be set to 1 or -1.

        `ind` = 1

            `fx` and `c`[0] need not be set.

        `ind` = -1

            `fx` and `c`[0] must contain f(`x`) and f(`y`) respectively.

    ir : int, optional
        `On initial entry`: indicates the type of error test.

        `ir` = 0

            The test is: |`x`-`y`| <= 2.0*`tolx`*max(1.0, |`x`|).

        `ir` = 1

            The test is: |`x`-`y`| <= 2.0*`tolx`.

        `ir` = 2

            The test is: |`x`-`y`| <= 2.0*`tolx`*|`x`|.

    Returns
    -------
    x : float
        `On intermediate exit`: `x` contains the point at which f must
        be evaluated before re-entry to the function. `y` has no
        external significance but must be communicated back to the next
        call of ``contfn_brent_rcomm`` as argument `y`.

        `On final exit`: `x` and `y` define a smaller interval
        containing the zero, such that f(`x`)*f(`y`) <= 0.0, and
        |`x`-`y`| satisfies the accuracy specified by `tolx` and `ir`,
        unless an error has occurred. If `errno` = 4, `x` and `y`
        generally contain very good approximations to a pole; if `errno`
        = 5, `x` and `y` generally contain very good approximations to
        the zero (see [Raises/Warns]). If a point `x` is found such that
        f(`x`) = 0.0, on final exit `x` = `y` (in this case there is no
        guarantee that `x` is a simple zero). In all cases, the value
        returned in `x` is the better approximation to the zero.

    y : float
        `On intermediate exit`: `x` contains the point at which f must
        be evaluated before re-entry to the function. `y` has no
        external significance but must be communicated back to the next
        call of ``contfn_brent_rcomm`` as argument `y`.

        `On final exit`: `x` and `y` define a smaller interval
        containing the zero, such that f(`x`)*f(`y`) <= 0.0, and
        |`x`-`y`| satisfies the accuracy specified by `tolx` and `ir`,
        unless an error has occurred. If `errno` = 4, `x` and `y`
        generally contain very good approximations to a pole; if `errno`
        = 5, `x` and `y` generally contain very good approximations to
        the zero (see [Raises/Warns]). If a point `x` is found such that
        f(`x`) = 0.0, on final exit `x` = `y` (in this case there is no
        guarantee that `x` is a simple zero). In all cases, the value
        returned in `x` is the better approximation to the zero.

    ind : int
        `On intermediate exit`: contains 2, 3 or 4. The calling program
        must evaluate f at `x`, storing the result in `fx`, and re-enter
        ``contfn_brent_rcomm`` with all other arguments unchanged.

        `On final exit`: contains 0.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, f(`x`) and f(`y`) have the same sign with neither
            equalling 0.0: f(`x`) = *<value>* and f(`y`) = *<value>*.

        (`errno` 2)
            On entry, `ind` = *<value>*.

            Constraint: `ind` = -1, 1, 2, 3 or 4.

        (`errno` 3)
            On entry, `tolx` = *<value>*.

            Constraint: `tolx` > 0.0.

        (`errno` 3)
            On entry, `ir` = *<value>*.

            Constraint: `ir` = 0, 1 or 2.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 4)
            The final interval may contain a pole rather than a zero.
            Note that this error exit is not completely reliable: it may
            be taken in extreme cases when [`x`,`y`] contains a zero, or
            it may not be taken when [`x`,`y`] contains a pole. Both
            these cases occur most frequently when `tolx` is large.

        (`errno` 5)
            The tolerance `tolx` has been set too small for the problem
            being solved. However, the values `x` and `y` returned may
            well be good approximations to the zero. `tolx` = *<value>*.

    References
    ----------
    Brent, R P, 1973, `Algorithms for Minimization Without Derivatives`,
    Prentice--Hall

    Bus, J C P and Dekker, T J, 1975, `Two efficient algorithms with
    guaranteed convergence for finding a zero of a function`, ACM Trans.
    Math. Software (1), 330--345

    Dahlquist, G and Björck, Å, 1974, `Numerical Methods`,
    Prentice--Hall
    """

    fun_name = 'naginterfaces.library.roots.contfn_brent_rcomm'

    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
    )
    try:
        _return_scalars = _b_roots.contfn_brent_rcomm(
            x,
            y,
            fx,
            tolx,
            ir,
            c,
            ind,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _x_py = _return_scalars.x
    _y_py = _return_scalars.y
    _ind_py = _return_scalars.ind
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'y', 'ind'],
    )(
        _x_py,
        _y_py,
        _ind_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def lambertw_real(x, branch, offset):
    u"""
    Real values of Lambert's W function, W(x).

    ``lambertw_real`` returns the real values of Lambert's W function
    W(x).

    For full information please refer to the NAG Library document for
    c05ba

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c05/c05baf.html

    Parameters
    ----------
    x : float
        If `offset` = True, `x` is the offset Delta x from -exp(-1) of
        the intended argument to W; that is, W(beta) is computed, where
        beta = -exp(-1)+Delta x.

        If `offset` = False, `x` is the argument x of the function; that
        is, W(beta) is computed, where beta = x.

    branch : int
        The real branch required.

        `branch` = 0

            The branch W_0 is selected.

        `branch` = -1

            The branch W_{-1} is selected.

    offset : bool
        Controls whether or not `x` is being specified as an offset from
        -exp(-1).

    Returns
    -------
    lambw : float
        The real values of Lambert's W function W(x).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `branch` = *<value>*.

            Constraint: `branch` = 0 or -1.

        (`errno` 1)
            On entry, `offset` = True and `x` = *<value>*.

            Constraint: if `offset` = True then `x` >= 0.0.

        (`errno` 1)
            On entry, `offset` = False and `x` = *<value>*.

            Constraint: if `offset` = False then `x` >= -exp(-1.0).

        (`errno` 1)
            On entry, `branch` = -1, `offset` = True and `x` =
            *<value>*.

            Constraint: if `branch` = -1 and `offset` = True then `x` <
            exp(-1.0).

        (`errno` 1)
            On entry, `branch` = -1, `offset` = False and `x` =
            *<value>*.

            Constraint: if `branch` = -1 and `offset` = False then `x` <
            0.0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            For the given offset `x`, W is negligibly different from -1:
            `x` = *<value>*.

        (`errno` 2)
            `x` is close to -exp(-1). Enter `x` as an offset to -exp(-1)
            for greater accuracy: `x` = *<value>*.

    References
    ----------
    Barry, D J, Culligan--Hensley, P J, and Barry, S J, 1995, `Real
    values of the W-function`, ACM Trans. Math. Software (21(2)),
    161--171
    """

    try:
        _return_scalar = _b_roots.lambertw_real(
            x,
            branch,
            offset,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['lambw'],
    )(
        _return_scalar,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_scalar

def lambertw_complex(branch, offset, z):
    u"""
    Values of Lambert's W function, W(z).

    ``lambertw_complex`` computes the values of Lambert's W function
    W(z).

    For full information please refer to the NAG Library document for
    c05bb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c05/c05bbf.html

    Parameters
    ----------
    branch : int
        The branch required.

    offset : bool
        Controls whether or not `z` is being specified as an offset from
        -exp(-1).

    z : complex
        If `offset` = True, `z` is the offset Delta z from -exp(-1) of
        the intended argument to W; that is, W(beta) is computed, where
        beta = -exp(-1)+Delta z.

        If `offset` = False, `z` is the argument z of the function; that
        is, W(beta) is computed, where beta = z.

    Returns
    -------
    w : complex
        The value W(beta): see also the description of `z`.

    resid : float
        The residual |W(beta)exp(W(beta))-beta|: see also the
        description of `z`.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            For the given offset `z`, W is negligibly different from -1:
            Re(`z`) = *<value>* and Im(`z`) = *<value>*.

        (`errno` 1)
            `z` is close to -exp(-1). Enter `z` as an offset to -exp(-1)
            for greater accuracy: Re(`z`) = *<value>* and Im(`z`) =
            *<value>*.

        (`errno` 2)
            The iterative procedure used internally did not converge in
            *<value>* iterations. Check the value of `resid` for the
            accuracy of `w`.

    References
    ----------
    Corless, R M, Gonnet, G H, Hare, D E G, Jeffrey, D J and Knuth, D E,
    1996, `On the Lambert W function`, Advances in Comp. Math. (3),
    329--359
    """

    try:
        _return_scalars = _b_roots.lambertw_complex(
            branch,
            offset,
            z,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _w_py = _return_scalars.w
    _resid_py = _return_scalars.resid
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['w', 'resid'],
    )(
        _w_py,
        _resid_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def sys_func_aa(fcn, x, atol=sqrt(machine.precision()), rtol=sqrt(machine.precision()), m=0, cndtol=0.0, astart=0, data=None):
    u"""
    Solution of a system of nonlinear equations using Anderson
    acceleration.

    ``sys_func_aa`` finds a solution of a system of nonlinear equations
    by fixed-point iteration using Anderson acceleration.

    For full information please refer to the NAG Library document for
    c05mb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c05/c05mbf.html

    Parameters
    ----------
    fcn : callable fvec = fcn(x, data=None)
        `fcn` must return the values of the functions f_k at a point x.

        Parameters
        ~~~~~~~~~~
        x : float, ndarray, shape (n)
            The components of the point x at which the functions must be
            evaluated.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        fvec : float, array-like, shape (n)
            The function values f_k(x).

    x : float, array-like, shape (n)
        An initial guess at the solution vector, x^_0.

    atol : float, optional
        The absolute convergence criterion; see `rtol`.

    rtol : float, optional
        The relative convergence criterion. At each iteration
        ||f(x^_i)|| is computed. The iteration is deemed to have
        converged if ||f(x^_i)|| <= max(`atol`, `rtol`*||f(x^_0)||).

    m : int, optional
        m, the number of previous iterates to use in Anderson
        acceleration. If m = 0, Anderson acceleration is not used.

    cndtol : float, optional
        The maximum allowable condition number for the triangular QR
        factor generated during Anderson acceleration. At each
        iteration, if the condition number exceeds `cndtol`, columns are
        deleted until it is sufficiently small.

        If `cndtol` = 0.0, no condition number tests are performed.

    astart : int, optional
        The number of iterations by which to delay the start of Anderson
        acceleration.

    data : arbitrary, optional
        User-communication data for callback functions.

    Returns
    -------
    x : float, ndarray, shape (n)
        The final estimate of the solution vector.

    fvec : float, ndarray, shape (n)
        The function values at the final point, `x`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n > 0.

        (`errno` 2)
            On entry, `atol` = *<value>*.

            Constraint: `atol` >= 0.0.

        (`errno` 3)
            On entry, `rtol` = *<value>*.

            Constraint: `rtol` >= 0.0.

        (`errno` 4)
            On entry, `m` = *<value>* and n = *<value>*.

            Constraint: 0 <= `m` <= n.

        (`errno` 5)
            On entry, `cndtol` = *<value>*.

            Constraint: `cndtol` >= 0.0.

        (`errno` 6)
            On entry, `astart` = *<value>*.

            Constraint: `astart` >= 0.

        (`errno` 7)
            An error occurred in evaluating the QR decomposition during
            Anderson acceleration. This may be due to slow convergence
            of the iteration. Try setting the value of `cndtol`. If
            condition number tests are already performed, try decreasing
            `cndtol`.

        (`errno` 11)
            The iteration has diverged and subsequent iterates are too
            large to be computed in floating-point arithmetic.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 8)
            The iteration is not making good progress. This failure exit
            may indicate that the system does not have a zero, or that
            the solution is very close to the origin. Rerunning
            ``sys_func_aa`` from a different starting point may avoid
            the region of difficulty.

        (`errno` 9)
            There have been at least 200*(n+1) calls to `fcn`. Consider
            restarting the calculation from the point held in `x`.

    NagCallbackTerminateWarning
        (`errno` 10)
            Termination requested in `fcn`.

    References
    ----------
    Anderson, D G, 1965, `Iterative Procedures for Nonlinear Integral
    Equations`, J. Assoc. Comput. Mach. (12), 547--560
    """

    fun_name = 'naginterfaces.library.roots.sys_func_aa'

    cb_fcn_name = fun_name + ':fcn'

    def cb_fcn(n, x, fvec, data=None):
        """Wrapping user's fcn. Called as base fcn."""
        if data is not None:
            _fvec_fcn = fcn(
                x,
                data,
            )
        else:
            _fvec_fcn = fcn(
                x,
            )
        fvec[:] = _utils._to_ndarray(
            {'fun_name': cb_fcn_name, 'returned': 'fvec'},
            _fvec_fcn,
            copy=False,
            exp_shape=(n,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )

    _b_utils._check_callable(fcn, {'fun_name': fun_name, 'entity': 'fcn'})
    _fcn = cb_fcn
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _fvec = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _b_roots.sys_func_aa(
            _fcn,
            _n,
            _x,
            _fvec,
            atol,
            rtol,
            m,
            cndtol,
            astart,
            data,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _x_py = _x
    _fvec_py = _fvec
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'fvec'],
    )(
        _x_py,
        _fvec_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def sys_func_aa_rcomm(irevcm, x, fvec, comm, atol=sqrt(machine.precision()), rtol=sqrt(machine.precision()), m=0, cndtol=0.0, astart=0):
    u"""
    Solution of a system of nonlinear equations using Anderson
    acceleration (reverse communication).

    ``sys_func_aa_rcomm`` is a comprehensive reverse communication
    function that finds a solution of a system of nonlinear equations by
    fixed-point iteration using Anderson acceleration.

    For full information please refer to the NAG Library document for
    c05md

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c05/c05mdf.html

    Parameters
    ----------
    irevcm : int
        `On initial entry`: must have the value 0.

    x : float, ndarray, shape (n), modified in place
        `On initial entry`: an initial guess at the solution vector,
        x^_0.

        `On intermediate exit`: contains the current point.

        `On final exit`: the final estimate of the solution vector.

    fvec : float, ndarray, shape (n), modified in place
        `On initial entry`: need not be set.

        `On intermediate entry`: if `irevcm` = 1, `fvec` must not be
        changed.

        If `irevcm` = 2, `fvec` must be set to the values of the
        functions computed at the current point `x`, f(x^_i).

        `On final exit`: the function values at the final point, `x`.

    comm : dict, communication object, modified in place
        Communication structure.

        `On initial entry`: need not be set.

    atol : float, optional
        `On initial entry`: the absolute convergence criterion; see
        `rtol`.

    rtol : float, optional
        `On initial entry`: the relative convergence criterion. At each
        iteration ||f(x^_i)|| is computed. The iteration is deemed to
        have converged if ||f(x^_i)|| <= max(`atol`,
        `rtol`*||f(x^_0)||).

    m : int, optional
        `On initial entry`: m, the number of previous iterates to use in
        Anderson acceleration. If m = 0, Anderson acceleration is not
        used.

    cndtol : float, optional
        `On initial entry`: the maximum allowable condition number for
        the triangular QR factor generated during Anderson acceleration.
        At each iteration, if the condition number exceeds `cndtol`,
        columns are deleted until it is sufficiently small.

        If `cndtol` = 0.0, no condition number tests are performed.

    astart : int, optional
        `On initial entry`: the number of iterations by which to delay
        the start of Anderson acceleration.

    Returns
    -------
    irevcm : int
        `On intermediate exit`: specifies what action you must take
        before re-entering ``sys_func_aa_rcomm`` with `irevcm` set to
        this value. The value of `irevcm` should be interpreted as
        follows:

        `irevcm` = 1

            Indicates the start of a new iteration. No action is
            required by you, but `x` and `fvec` are available for
            printing, and a limit on the number of iterations can be
            applied.

        `irevcm` = 2

            Indicates that before re-entry to ``sys_func_aa_rcomm``,
            `fvec` must contain the function values f(x^_i).

        `On final exit`: `irevcm` = 0 and the algorithm has terminated.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On initial entry, `irevcm` = *<value>*.

            Constraint: `irevcm` = 0.

        (`errno` 1)
            On intermediate entry, `irevcm` = *<value>*.

            Constraint: `irevcm` = 1 or 2.

        (`errno` 2)
            On entry, n = *<value>*.

            Constraint: n > 0.

        (`errno` 3)
            On entry, `atol` = *<value>*.

            Constraint: `atol` >= 0.0.

        (`errno` 4)
            On entry, `rtol` = *<value>*.

            Constraint: `rtol` >= 0.0.

        (`errno` 5)
            On entry, `m` = *<value>* and n = *<value>*.

            Constraint: 0 <= `m` <= n.

        (`errno` 6)
            On entry, `cndtol` = *<value>*.

            Constraint: `cndtol` >= 0.0.

        (`errno` 7)
            On entry, `astart` = *<value>*.

            Constraint: `astart` >= 0.

        (`errno` 8)
            An error occurred in evaluating the QR decomposition during
            Anderson acceleration. This may be due to slow convergence
            of the iteration. Try setting the value of `cndtol`. If
            condition number tests are already performed, try decreasing
            `cndtol`.

        (`errno` 9)
            The iteration is not making good progress, as measured by
            the reduction in the norm of f(x) in the last *<value>*
            iterations.

        (`errno` 10)
            The iteration has diverged and subsequent iterates are too
            large to be computed in floating-point arithmetic.

    References
    ----------
    Anderson, D G, 1965, `Iterative Procedures for Nonlinear Integral
    Equations`, J. Assoc. Comput. Mach. (12), 547--560
    """

    fun_name = 'naginterfaces.library.roots.sys_func_aa_rcomm'

    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'fvec'},
        fvec,
    )
    _return_scalar = _b_roots.sys_func_aa_rcomm(
        irevcm,
        _n,
        x,
        fvec,
        atol,
        rtol,
        m,
        cndtol,
        astart,
        comm,
    )
    _irevcm_py = _return_scalar
    return _irevcm_py

def sys_func_easy(fcn, x, xtol=sqrt(machine.precision()), data=None):
    u"""
    Solution of a system of nonlinear equations using function values
    only (easy-to-use).

    ``sys_func_easy`` is an easy-to-use function that finds a solution
    of a system of nonlinear equations by a modification of the Powell
    hybrid method.

    For full information please refer to the NAG Library document for
    c05qb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c05/c05qbf.html

    Parameters
    ----------
    fcn : callable fvec = fcn(x, data=None)
        `fcn` must return the values of the functions f_i at a point x.

        Parameters
        ~~~~~~~~~~
        x : float, ndarray, shape (n)
            The components of the point x at which the functions must be
            evaluated.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        fvec : float, array-like, shape (n)
            The function values f_i(x).

    x : float, array-like, shape (n)
        An initial guess at the solution vector.

    xtol : float, optional
        The accuracy in `x` to which the solution is required.

    data : arbitrary, optional
        User-communication data for callback functions.

    Returns
    -------
    x : float, ndarray, shape (n)
        The final estimate of the solution vector.

    fvec : float, ndarray, shape (n)
        The function values at the final point returned in `x`.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, n = *<value>*.

            Constraint: n > 0.

        (`errno` 12)
            On entry, `xtol` = *<value>*.

            Constraint: `xtol` >= 0.0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            There have been at least 200*(n+1) calls to `fcn`. Consider
            restarting the calculation from the point held in `x`.

        (`errno` 3)
            No further improvement in the solution is possible. `xtol`
            is too small: `xtol` = *<value>*.

        (`errno` 4)
            The iteration is not making good progress. This failure exit
            may indicate that the system does not have a zero, or that
            the solution is very close to the origin (see [Accuracy]).
            Otherwise, rerunning ``sys_func_easy`` from a different
            starting point may avoid the region of difficulty.

    NagCallbackTerminateWarning
        (`errno` 5)
            Termination requested in `fcn`.

    References
    ----------
    Moré, J J, Garbow, B S and Hillstrom, K E, 1980, `User guide for
    MINPACK-1`, Technical Report ANL-80-74, Argonne National Laboratory

    Powell, M J D, 1970, `A hybrid method for nonlinear algebraic
    equations`, Numerical Methods for Nonlinear Algebraic Equations, (ed
    P Rabinowitz), Gordon and Breach
    """

    fun_name = 'naginterfaces.library.roots.sys_func_easy'

    cb_fcn_name = fun_name + ':fcn'

    def cb_fcn(n, x, fvec, data=None):
        """Wrapping user's fcn. Called as base fcn."""
        if data is not None:
            _fvec_fcn = fcn(
                x,
                data,
            )
        else:
            _fvec_fcn = fcn(
                x,
            )
        fvec[:] = _utils._to_ndarray(
            {'fun_name': cb_fcn_name, 'returned': 'fvec'},
            _fvec_fcn,
            copy=False,
            exp_shape=(n,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )

    _b_utils._check_callable(fcn, {'fun_name': fun_name, 'entity': 'fcn'})
    _fcn = cb_fcn
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _fvec = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _b_roots.sys_func_easy(
            _fcn,
            _n,
            _x,
            _fvec,
            xtol,
            data,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _x_py = _x
    _fvec_py = _fvec
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'fvec'],
    )(
        _x_py,
        _fvec_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def sys_func_expert(fcnval, x, ml, mu, mode, diag, nprint, fcnmon=None, xtol=sqrt(machine.precision()), maxfev=None, epsfcn=0.0, factor=100.0, data=None):
    u"""
    Solution of a system of nonlinear equations using function values
    only (comprehensive).

    ``sys_func_expert`` is a comprehensive function that finds a
    solution of a system of nonlinear equations by a modification of the
    Powell hybrid method.

    For full information please refer to the NAG Library document for
    c05qc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c05/c05qcf.html

    Parameters
    ----------
    fcnval : callable fvec = fcnval(x, data=None)
        `fcnval` must return the values of the functions f_i at a point
        x.

        Parameters
        ~~~~~~~~~~
        x : float, ndarray, shape (n)
            The components of the point x at which the functions must be
            evaluated.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        fvec : float, array-like, shape (n)
            `fvec` must contain the function values f_i(x).

    x : float, array-like, shape (n)
        An initial guess at the solution vector.

    ml : int
        The number of subdiagonals within the band of the Jacobian
        matrix. (If the Jacobian is not banded, or you are unsure, set
        `ml` = n-1.)

    mu : int
        The number of superdiagonals within the band of the Jacobian
        matrix. (If the Jacobian is not banded, or you are unsure, set
        `mu` = n-1.)

    mode : int
        Indicates whether or not you have provided scaling factors in
        `diag`.

        If `mode` = 2, the scaling must have been specified in `diag`.

        Otherwise, if `mode` = 1, the variables will be scaled
        internally.

    diag : float, array-like, shape (n)
        If `mode` = 2, `diag` must contain multiplicative scale factors
        for the variables.

        If `mode` = 1, `diag` need not be set.

    nprint : int
        Indicates whether (and how often) calls to `fcnmon` are to be
        made.

        `nprint` <= 0

            No calls are made.

        `nprint` > 0

            `fcnmon` is called at the beginning of the first iteration,
            every `nprint` iterations thereafter and immediately before
            the return from ``sys_func_expert``.

    fcnmon : None or callable fcnmon(x, fvec, data=None), optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `fcnmon` may be used to monitor the progress of the algorithm.

        Parameters
        ~~~~~~~~~~
        x : float, ndarray, shape (n)
            The components of the current evaluation point x.

        fvec : float, ndarray, shape (n)
            `fvec` contains the function values f_i(x).

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

    xtol : float, optional
        The accuracy in `x` to which the solution is required.

    maxfev : None or int, optional
        Note: if this argument is None then a default value will be
        used, determined as follows: 200*(n+1).

        The maximum number of calls to `fcnval`. ``sys_func_expert``
        will exit with `errno` = 2, if, at the end of an iteration, the
        number of calls to `fcnval` exceeds `maxfev`.

    epsfcn : float, optional
        A rough estimate of the largest relative error in the functions.
        It is used in determining a suitable step for a forward
        difference approximation to the Jacobian. If `epsfcn` is less
        than machine precision (returned by ``machine.precision``) then
        machine precision is used. Consequently a value of 0.0 will
        often be suitable.

    factor : float, optional
        A quantity to be used in determining the initial step bound. In
        most cases, `factor` should lie between 0.1 and 100.0. (The step
        bound is `factor`*||`diag`*`x`||_2 if this is nonzero; otherwise
        the bound is `factor`.)

    data : arbitrary, optional
        User-communication data for callback functions.

    Returns
    -------
    x : float, ndarray, shape (n)
        The final estimate of the solution vector.

    fvec : float, ndarray, shape (n)
        The function values at the final point returned in `x`.

    diag : float, ndarray, shape (n)
        The scale factors actually used (computed internally if `mode` =
        1).

    nfev : int
        The number of calls made to `fcnval`.

    fjac : float, ndarray, shape (n, n)
        The orthogonal matrix Q produced by the QR factorization of the
        final approximate Jacobian.

    r : float, ndarray, shape (n*(n+1)/2)
        The upper triangular matrix R produced by the QR factorization
        of the final approximate Jacobian, stored row-wise.

    qtf : float, ndarray, shape (n)
        The vector Q^Tf.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, n = *<value>*.

            Constraint: n > 0.

        (`errno` 12)
            On entry, `xtol` = *<value>*.

            Constraint: `xtol` >= 0.0.

        (`errno` 13)
            On entry, `mode` = *<value>*.

            Constraint: `mode` = 1 or 2.

        (`errno` 14)
            On entry, `factor` = *<value>*.

            Constraint: `factor` > 0.0.

        (`errno` 15)
            On entry, `mode` = 2 and `diag` contained a non-positive
            element.

        (`errno` 16)
            On entry, `ml` = *<value>*.

            Constraint: `ml` >= 0.

        (`errno` 17)
            On entry, `mu` = *<value>*.

            Constraint: `mu` >= 0.

        (`errno` 18)
            On entry, `maxfev` = *<value>*.

            Constraint: `maxfev` > 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            There have been at least `maxfev` calls to `fcnval`:
            `maxfev` = *<value>*. Consider restarting the calculation
            from the final point held in `x`.

        (`errno` 3)
            No further improvement in the solution is possible. `xtol`
            is too small: `xtol` = *<value>*.

        (`errno` 4)
            The iteration is not making good progress, as measured by
            the improvement from the last *<value>* Jacobian
            evaluations.

        (`errno` 5)
            The iteration is not making good progress, as measured by
            the improvement from the last *<value>* iterations.

    NagCallbackTerminateWarning
        (`errno` 6)
            Termination requested in `fcnval` or `fcnmon`.

    References
    ----------
    Moré, J J, Garbow, B S and Hillstrom, K E, 1980, `User guide for
    MINPACK-1`, Technical Report ANL-80-74, Argonne National Laboratory

    Powell, M J D, 1970, `A hybrid method for nonlinear algebraic
    equations`, Numerical Methods for Nonlinear Algebraic Equations, (ed
    P Rabinowitz), Gordon and Breach
    """

    fun_name = 'naginterfaces.library.roots.sys_func_expert'

    cb_fcnval_name = fun_name + ':fcnval'

    def cb_fcnval(n, x, fvec, data=None):
        """Wrapping user's fcnval. Called as base fcnval."""
        if data is not None:
            _fvec_fcnval = fcnval(
                x,
                data,
            )
        else:
            _fvec_fcnval = fcnval(
                x,
            )
        fvec[:] = _utils._to_ndarray(
            {'fun_name': cb_fcnval_name, 'returned': 'fvec'},
            _fvec_fcnval,
            copy=False,
            exp_shape=(n,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )

    def cb_fcnmon(n, x, fvec, data=None):
        """Wrapping user's fcnmon. Called as base fcnmon."""
        if data is not None:
            fcnmon(
                x,
                fvec,
                data,
            )
        else:
            fcnmon(
                x,
                fvec,
            )

    _b_utils._check_callable(fcnval, {'fun_name': fun_name, 'entity': 'fcnval'})
    _fcnval = cb_fcnval
    _b_utils._check_callable(fcnmon, {'fun_name': fun_name, 'entity': 'fcnmon'}, allow_none=True)
    _fcnmon = None if fcnmon is None else cb_fcnmon
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _fvec = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _maxfev = 200*(_n_py+1) if maxfev is None else maxfev
    _diag = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'diag'},
        diag,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _fjac = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py, _n_py),
        sorder=_b_utils._EngineData.default_sorder,
    )
    _r = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py*(_n_py+1)//2,),
    )
    _qtf = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _return_scalar = _b_roots.sys_func_expert(
            _fcnval,
            _fcnmon,
            _n,
            _x,
            _fvec,
            xtol,
            _maxfev,
            ml,
            mu,
            epsfcn,
            mode,
            _diag,
            factor,
            nprint,
            _fjac,
            _r,
            _qtf,
            data,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _x_py = _x
    _diag_py = _diag
    _fvec_py = _fvec
    _nfev_py = _return_scalar
    _fjac_py = _fjac
    _r_py = _r
    _qtf_py = _qtf
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'fvec', 'diag', 'nfev', 'fjac', 'r', 'qtf'],
    )(
        _x_py,
        _fvec_py,
        _diag_py,
        _nfev_py,
        _fjac_py,
        _r_py,
        _qtf_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def sys_func_rcomm(irevcm, x, fvec, ml, mu, mode, diag, fjac, r, qtf, comm, xtol=sqrt(machine.precision()), epsfcn=0.0, factor=100.0):
    u"""
    Solution of a system of nonlinear equations using function values
    only (reverse communication).

    ``sys_func_rcomm`` is a comprehensive reverse communication function
    that finds a solution of a system of nonlinear equations by a
    modification of the Powell hybrid method.

    For full information please refer to the NAG Library document for
    c05qd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c05/c05qdf.html

    Parameters
    ----------
    irevcm : int
        `On initial entry`: must have the value 0.

    x : float, ndarray, shape (n), modified in place
        `On initial entry`: an initial guess at the solution vector.

        `On intermediate exit`: contains the current point.

        `On final exit`: the final estimate of the solution vector.

    fvec : float, ndarray, shape (n), modified in place
        `On initial entry`: need not be set.

        `On intermediate entry`: if `irevcm` = 1, `fvec` must not be
        changed.

        If `irevcm` = 2, `fvec` must be set to the values of the
        functions computed at the current point `x`.

        `On final exit`: the function values at the final point, `x`.

    ml : int
        `On initial entry`: the number of subdiagonals within the band
        of the Jacobian matrix. (If the Jacobian is not banded, or you
        are unsure, set `ml` = n-1.)

    mu : int
        `On initial entry`: the number of superdiagonals within the band
        of the Jacobian matrix. (If the Jacobian is not banded, or you
        are unsure, set `mu` = n-1.)

    mode : int
        `On initial entry`: indicates whether or not you have provided
        scaling factors in `diag`.

        If `mode` = 2, the scaling must have been supplied in `diag`.

        Otherwise, if `mode` = 1, the variables will be scaled
        internally.

    diag : float, ndarray, shape (n), modified in place
        `On entry`: if `mode` = 2, `diag` must contain multiplicative
        scale factors for the variables.

        If `mode` = 1, `diag` need not be set.

        `On exit`: the scale factors actually used (computed internally
        if `mode` = 1).

    fjac : float, ndarray, shape (n, n), modified in place
        `On initial entry`: need not be set.

        `On intermediate exit`: must not be changed.

        `On final exit`: the orthogonal matrix Q produced by the QR
        factorization of the final approximate Jacobian.

    r : float, ndarray, shape (n*(n+1)/2), modified in place
        `On initial entry`: need not be set.

        `On intermediate exit`: must not be changed.

        `On final exit`: the upper triangular matrix R produced by the
        QR factorization of the final approximate Jacobian, stored
        row-wise.

    qtf : float, ndarray, shape (n), modified in place
        `On initial entry`: need not be set.

        `On intermediate exit`: must not be changed.

        `On final exit`: the vector Q^Tf.

    comm : dict, communication object, modified in place
        Communication structure.

        `On initial entry`: need not be set.

    xtol : float, optional
        `On initial entry`: the accuracy in `x` to which the solution is
        required.

    epsfcn : float, optional
        `On initial entry`: the order of the largest relative error in
        the functions. It is used in determining a suitable step for a
        forward difference approximation to the Jacobian. If `epsfcn` is
        less than machine precision (returned by ``machine.precision``)
        then machine precision is used. Consequently a value of 0.0 will
        often be suitable.

    factor : float, optional
        `On initial entry`: a quantity to be used in determining the
        initial step bound. In most cases, `factor` should lie between
        0.1 and 100.0. (The step bound is `factor`*||`diag`*`x`||_2 if
        this is nonzero; otherwise the bound is `factor`.)

    Returns
    -------
    irevcm : int
        `On intermediate exit`: specifies what action you must take
        before re-entering ``sys_func_rcomm`` with `irevcm` set to this
        value. The value of `irevcm` should be interpreted as follows:

        `irevcm` = 1

            Indicates the start of a new iteration. No action is
            required by you, but `x` and `fvec` are available for
            printing.

        `irevcm` = 2

            Indicates that before re-entry to ``sys_func_rcomm``, `fvec`
            must contain the function values f_i(x).

        `On final exit`: `irevcm` = 0 and the algorithm has terminated.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, `irevcm` = *<value>*.

            Constraint: `irevcm` = 0, 1 or 2.

        (`errno` 11)
            On entry, n = *<value>*.

            Constraint: n > 0.

        (`errno` 12)
            On entry, `xtol` = *<value>*.

            Constraint: `xtol` >= 0.0.

        (`errno` 13)
            On entry, `mode` = *<value>*.

            Constraint: `mode` = 1 or 2.

        (`errno` 14)
            On entry, `factor` = *<value>*.

            Constraint: `factor` > 0.0.

        (`errno` 15)
            On entry, `mode` = 2 and `diag` contained a non-positive
            element.

        (`errno` 16)
            On entry, `ml` = *<value>*.

            Constraint: `ml` >= 0.

        (`errno` 17)
            On entry, `mu` = *<value>*.

            Constraint: `mu` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            No further improvement in the solution is possible. `xtol`
            is too small: `xtol` = *<value>*.

        (`errno` 4)
            The iteration is not making good progress, as measured by
            the improvement from the last *<value>* Jacobian
            evaluations.

        (`errno` 5)
            The iteration is not making good progress, as measured by
            the improvement from the last *<value>* iterations.

    References
    ----------
    Moré, J J, Garbow, B S and Hillstrom, K E, 1980, `User guide for
    MINPACK-1`, Technical Report ANL-80-74, Argonne National Laboratory

    Powell, M J D, 1970, `A hybrid method for nonlinear algebraic
    equations`, Numerical Methods for Nonlinear Algebraic Equations, (ed
    P Rabinowitz), Gordon and Breach

    See Also
    --------
    :meth:`naginterfaces.library.examples.roots.sys_func_rcomm_ex.main`
    """

    fun_name = 'naginterfaces.library.roots.sys_func_rcomm'

    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'fvec'},
        fvec,
    )
    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'diag'},
        diag,
    )
    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'fjac'},
        fjac,
    )
    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'r'},
        r,
    )
    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'qtf'},
        qtf,
    )
    try:
        _return_scalar = _b_roots.sys_func_rcomm(
            irevcm,
            _n,
            x,
            fvec,
            xtol,
            ml,
            mu,
            epsfcn,
            mode,
            diag,
            factor,
            fjac,
            r,
            qtf,
            comm,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _irevcm_py = _return_scalar
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['irevcm'],
    )(
        _irevcm_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _irevcm_py

def sparsys_func_easy(fcn, x, xtol=sqrt(machine.precision()), nnz=None, comm=None, data=None):
    u"""
    Solution of a sparse system of nonlinear equations using function
    values only (easy-to-use).

    ``sparsys_func_easy`` is an easy-to-use function that finds a
    solution of a sparse system of nonlinear equations by a modification
    of the Powell hybrid method.

    For full information please refer to the NAG Library document for
    c05qs

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c05/c05qsf.html

    Parameters
    ----------
    fcn : callable fcn(indf, x, fvec, data=None)
        `fcn` must return the values of the functions f_i at a point x.

        Parameters
        ~~~~~~~~~~
        indf : int, ndarray, shape (lindf)
            Note: this argument represents array indices; the values
            supplied will be base-1.

            `indf` specifies the indices i for which values of f_i(x)
            must be computed. The indices are specified in strictly
            ascending order.

        x : float, ndarray, shape (n)
            The components of the point x at which the functions must be
            evaluated. `x`[i-1] contains the coordinate x_i.

        fvec : float, ndarray, shape (n), to be modified in place
            `On exit`: `fvec`[i-1] must contain the function values
            f_i(x), for all indices i in `indf`.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

    x : float, array-like, shape (n)
        An initial guess at the solution vector. `x`[i-1] must contain
        the coordinate x_i.

    xtol : float, optional
        The accuracy in `x` to which the solution is required.

    nnz : None or int, optional
        Note: if this argument is None then a default value will be
        used, determined as follows: n*n.

        If supplied `nnz` should be an upper estimate of the number of
        nonzero entries found in the Jacobian.

        The size of the data that is self communicated to this routine
        via `comm` is a function of `nnz`.

        A suitable overestimate for `nnz` is n^2, but this may be
        suboptimal when n is large.

        If an existing sparsity pattern for the problem is communicated
        via `comm`, `nnz` will be ignored.

        The value that was supplied in the call to ``sparsys_func_easy``
        which computed that sparsity pattern will be extracted from
        `comm` and used instead.

    comm : None or dict, communication object, optional, modified in
    place
        `Optionally, on entry`: an existing sparsity pattern for the
        Jacobian, computed by a subsequent call. Otherwise, supplying an
        uninitialized object indicates it is the first time
        ``sparsys_func_easy`` is called for this specific problem.

        `On exit`, if not None on entry: will be populated with the
        sparsity pattern of the Jacobian. On subsequent calls, the same
        communication object can be passed in again if the problem has a
        Jacobian of the same sparsity pattern.

    data : arbitrary, optional
        User-communication data for callback functions.

    Returns
    -------
    x : float, ndarray, shape (n)
        The final estimate of the solution vector.

    fvec : float, ndarray, shape (n)
        The function values at the final point returned in `x`.
        `fvec`[i-1] contains the function values f_i.

    actual_nnz : int
        The number of nonzero entries found in the Jacobian.

    Raises
    ------
    NagValueError
        (`errno` 6)
            On entry, the supplied `nnz` = *<value>* was too small for
            this problem.

        (`errno` 7)
            On entry, the supplied `nnz` = *<value>* was too small for
            this problem.

        (`errno` 9)
            An internal error has occurred. Code = *<value>*.

        (`errno` 11)
            On entry, n = *<value>*.

            Constraint: n > 0.

        (`errno` 12)
            On entry, `xtol` = *<value>*.

            Constraint: `xtol` >= 0.0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            There have been at least 200*(n+1) calls to `fcn`. Consider
            restarting the calculation from the point held in `x` and
            supplying the populated sparsity pattern from `comm`.

        (`errno` 3)
            No further improvement in the solution is possible. `xtol`
            is too small: `xtol` = *<value>*.

        (`errno` 4)
            The iteration is not making good progress. This failure exit
            may indicate that the system does not have a zero, or that
            the solution is very close to the origin (see [Accuracy]).
            Otherwise, rerunning ``sparsys_func_easy`` from a different
            starting point may avoid the region of difficulty. The
            condition number of the Jacobian is *<value>*.

    NagCallbackTerminateWarning
        (`errno` 5)
            Termination requested in `fcn`.

    References
    ----------
    Broyden, C G, 1965, `A class of methods for solving nonlinear
    simultaneous equations`, Mathematics of Computation (19(92)),
    577--593

    Moré, J J, Garbow, B S and Hillstrom, K E, 1980, `User guide for
    MINPACK-1`, Technical Report ANL-80-74, Argonne National Laboratory

    Powell, M J D, 1970, `A hybrid method for nonlinear algebraic
    equations`, Numerical Methods for Nonlinear Algebraic Equations, (ed
    P Rabinowitz), Gordon and Breach

    Schubert, L K, 1970, `Modification of a quasi-Newton method for
    nonlinear equations with a sparse Jacobian`, Mathematics of
    Computation (24(109)), 27--30
    """

    def cb_fcn(n, lindf, indf, x, fvec, data=None):
        """Wrapping user's fcn. Called as base fcn."""
        if data is not None:
            fcn(
                indf,
                x,
                fvec,
                data,
            )
        else:
            fcn(
                indf,
                x,
                fvec,
            )

    fun_name = 'naginterfaces.library.roots.sparsys_func_easy'

    _b_utils._check_callable(fcn, {'fun_name': fun_name, 'entity': 'fcn'})
    _fcn = cb_fcn
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _fvec = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _nnz = _n_py*_n_py if nnz is None else nnz
    try:
        _return_scalar = _b_roots.sparsys_func_easy(
            _fcn,
            _n,
            _x,
            _fvec,
            xtol,
            _nnz,
            comm,
            data,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _x_py = _x
    _fvec_py = _fvec
    _actual_nnz_py = _return_scalar
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'fvec', 'actual_nnz'],
    )(
        _x_py,
        _fvec_py,
        _actual_nnz_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def sys_deriv_easy(fcnval, fcnjac, x, xtol=sqrt(machine.precision()), data=None, spiked_sorder='C'):
    u"""
    Solution of a system of nonlinear equations using first derivatives
    (easy-to-use).

    ``sys_deriv_easy`` is an easy-to-use function that finds a solution
    of a system of nonlinear equations by a modification of the Powell
    hybrid method.
    You must provide the Jacobian.

    For full information please refer to the NAG Library document for
    c05rb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c05/c05rbf.html

    Parameters
    ----------
    fcnval : callable fvec = fcnval(x, data=None)
        `fcnval` must return the values of the functions f_i at a point
        x.

        Parameters
        ~~~~~~~~~~
        x : float, ndarray, shape (n)
            The components of the point x at which the functions must be
            evaluated.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        fvec : float, array-like, shape (n)
            `fvec` must contain the function values f_i(x).

    fcnjac : callable fjac = fcnjac(x, data=None)
        `fcnjac` must return the Jacobian at x.

        Parameters
        ~~~~~~~~~~
        x : float, ndarray, shape (n)
            The components of the point x at which the Jacobian must be
            evaluated.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        fjac : float, array-like, shape (n, n)
            `fjac`[i-1,j-1] must contain the value of df_i/dx_j at the
            point x, for j = 1,2,...,n, for i = 1,2,...,n.

    x : float, array-like, shape (n)
        An initial guess at the solution vector.

    xtol : float, optional
        The accuracy in `x` to which the solution is required.

    data : arbitrary, optional
        User-communication data for callback functions.

    spiked_sorder : str, optional
        If `fjac` in `fcnjac` is spiked (i.e., has unit extent in all
        but one dimension, or has size 1), `spiked_sorder` selects the
        storage order to associate with it in the NAG Engine:

        spiked_sorder = 'C'
          row-major storage will be used;

        spiked_sorder = 'F'
          column-major storage will be used.

    Returns
    -------
    x : float, ndarray, shape (n)
        The final estimate of the solution vector.

    fvec : float, ndarray, shape (n)
        The function values at the final point returned in `x`.

    fjac : float, ndarray, shape (n, n)
        The orthogonal matrix Q produced by the QR factorization of the
        final approximate Jacobian.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, n = *<value>*.

            Constraint: n > 0.

        (`errno` 12)
            On entry, `xtol` = *<value>*.

            Constraint: `xtol` >= 0.0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            There have been at least 100*(n+1) calls to `fcnval` and
            `fcnjac` combined. Consider restarting the calculation from
            the point held in `x`.

        (`errno` 3)
            No further improvement in the solution is possible. `xtol`
            is too small: `xtol` = *<value>*.

        (`errno` 4)
            The iteration is not making good progress. This failure exit
            may indicate that the system does not have a zero, or that
            the solution is very close to the origin (see [Accuracy]).
            Otherwise, rerunning ``sys_deriv_easy`` from a different
            starting point may avoid the region of difficulty.

    NagCallbackTerminateWarning
        (`errno` 5)
            Termination requested in `fcnval` or `fcnjac`.

    References
    ----------
    Moré, J J, Garbow, B S and Hillstrom, K E, 1980, `User guide for
    MINPACK-1`, Technical Report ANL-80-74, Argonne National Laboratory

    Powell, M J D, 1970, `A hybrid method for nonlinear algebraic
    equations`, Numerical Methods for Nonlinear Algebraic Equations, (ed
    P Rabinowitz), Gordon and Breach
    """

    fun_name = 'naginterfaces.library.roots.sys_deriv_easy'

    cb_fcnval_name = fun_name + ':fcnval'

    def cb_fcnval(n, x, fvec, data=None):
        """Wrapping user's fcnval. Called as base fcnval."""
        if data is not None:
            _fvec_fcnval = fcnval(
                x,
                data,
            )
        else:
            _fvec_fcnval = fcnval(
                x,
            )
        fvec[:] = _utils._to_ndarray(
            {'fun_name': cb_fcnval_name, 'returned': 'fvec'},
            _fvec_fcnval,
            copy=False,
            exp_shape=(n,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )

    cb_fcnjac_name = fun_name + ':fcnjac'

    def cb_fcnjac(n, x, fjac, data=None):
        """Wrapping user's fcnjac. Called as base fcnjac."""
        if data is not None:
            _fjac_fcnjac = fcnjac(
                x,
                data,
            )
        else:
            _fjac_fcnjac = fcnjac(
                x,
            )
        _fjac_fcnjac_np = _utils._to_ndarray(
            {'fun_name': cb_fcnjac_name, 'returned': 'fjac'},
            _fjac_fcnjac,
            copy=False,
            exp_shape=(n, n),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _utils._check_callback_sorder(
            {'fun_name': cb_fcnjac_name},
            ((_fjac_fcnjac_np, 'fjac'),),
            wrapper_sorder=_m_sorder,
            spiked_sorder=spiked_sorder,
        )
        fjac[:, :] = _fjac_fcnjac_np

    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        (),
        spiked_sorder=spiked_sorder,
    )
    _b_utils._check_callable(fcnval, {'fun_name': fun_name, 'entity': 'fcnval'})
    _fcnval = cb_fcnval
    _b_utils._check_callable(fcnjac, {'fun_name': fun_name, 'entity': 'fcnjac'})
    _fcnjac = cb_fcnjac
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _fvec = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _fjac = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py, _n_py),
        sorder=_b_utils._EngineData.default_sorder,
    )
    try:
        _b_roots.sys_deriv_easy(
            _fcnval,
            _fcnjac,
            _n,
            _x,
            _fvec,
            _fjac,
            xtol,
            data,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _x_py = _x
    _fvec_py = _fvec
    _fjac_py = _fjac
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'fvec', 'fjac'],
    )(
        _x_py,
        _fvec_py,
        _fjac_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def sys_deriv_expert(fcnval, fcnjac, fcnmon, x, mode, diag, nprint, xtol=sqrt(machine.precision()), maxfev=None, factor=100.0, data=None, spiked_sorder='C'):
    u"""
    Solution of a system of nonlinear equations using first derivatives
    (comprehensive).

    ``sys_deriv_expert`` is a comprehensive function that finds a
    solution of a system of nonlinear equations by a modification of the
    Powell hybrid method.
    You must provide the Jacobian.

    For full information please refer to the NAG Library document for
    c05rc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c05/c05rcf.html

    Parameters
    ----------
    fcnval : callable fvec = fcnval(x, data=None)
        `fcnval` must return the values of the functions f_i at a point
        x.

        Parameters
        ~~~~~~~~~~
        x : float, ndarray, shape (n)
            The components of the point x at which the functions must be
            evaluated.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        fvec : float, array-like, shape (n)
            `fvec` must contain the function values f_i(x).

    fcnjac : callable fjac = fcnjac(x, data=None)
        `fcnjac` must return the Jacobian at x.

        Parameters
        ~~~~~~~~~~
        x : float, ndarray, shape (n)
            The components of the point x at which the Jacobian must be
            evaluated.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        fjac : float, array-like, shape (n, n)
            `fjac`[i-1,j-1] must contain the value of df_i/dx_j at the
            point x, for j = 1,2,...,n, for i = 1,2,...,n.

    fcnmon : callable fcnmon(x, fvec, fjac, data=None)
        `fcnmon` may be used to monitor the progress of the algorithm.

        Parameters
        ~~~~~~~~~~
        x : float, ndarray, shape (n)
            The components of the current evaluation point x.

        fvec : float, ndarray, shape (n)
            `fvec` contains the function values f_i(x).

        fjac : float, ndarray, shape (n, n)
            `fjac`[i-1,j-1] contains the value of df_i/dx_j at the point
            x, for j = 1,2,...,n, for i = 1,2,...,n.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

    x : float, array-like, shape (n)
        An initial guess at the solution vector.

    mode : int
        Indicates whether or not you have provided scaling factors in
        `diag`.

        If `mode` = 2, the scaling must have been specified in `diag`.

        Otherwise, if `mode` = 1, the variables will be scaled
        internally.

    diag : float, array-like, shape (n)
        If `mode` = 2, `diag` must contain multiplicative scale factors
        for the variables.

        If `mode` = 1, `diag` need not be set.

    nprint : int
        Indicates whether (and how often) calls to `fcnmon` are to be
        made.

        `nprint` <= 0

            No calls are made.

        `nprint` > 0

            `fcnmon` is called at the beginning of the first iteration,
            every `nprint` iterations thereafter and immediately before
            the return from ``sys_deriv_expert``.

    xtol : float, optional
        The accuracy in `x` to which the solution is required.

    maxfev : None or int, optional
        Note: if this argument is None then a default value will be
        used, determined as follows: 100*(n+1).

        The maximum number of calls to `fcnval` and `fcnjac` combined.
        ``sys_deriv_expert`` will exit with `errno` = 2, if, at the end
        of an iteration, the number of calls to `fcnval` and `fcnjac`
        combined exceeds `maxfev`.

    factor : float, optional
        A quantity to be used in determining the initial step bound. In
        most cases, `factor` should lie between 0.1 and 100.0. (The step
        bound is `factor`*||`diag`*`x`||_2 if this is nonzero; otherwise
        the bound is `factor`.)

    data : arbitrary, optional
        User-communication data for callback functions.

    spiked_sorder : str, optional
        If `fjac` in `fcnjac` is spiked (i.e., has unit extent in all
        but one dimension, or has size 1), `spiked_sorder` selects the
        storage order to associate with it in the NAG Engine:

        spiked_sorder = 'C'
          row-major storage will be used;

        spiked_sorder = 'F'
          column-major storage will be used.

    Returns
    -------
    x : float, ndarray, shape (n)
        The final estimate of the solution vector.

    fvec : float, ndarray, shape (n)
        The function values at the final point returned in `x`.

    fjac : float, ndarray, shape (n, n)
        The orthogonal matrix Q produced by the QR factorization of the
        final approximate Jacobian.

    diag : float, ndarray, shape (n)
        The scale factors actually used (computed internally if `mode` =
        1).

    nfev : int
        The number of calls made to `fcnval` and `fcnjac` combined to
        evaluate the functions.

    njev : int
        The number of calls made to `fcnjac`.

    r : float, ndarray, shape (n*(n+1)/2)
        The upper triangular matrix R produced by the QR factorization
        of the final approximate Jacobian, stored row-wise.

    qtf : float, ndarray, shape (n)
        The vector Q^Tf.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, n = *<value>*.

            Constraint: n > 0.

        (`errno` 12)
            On entry, `xtol` = *<value>*.

            Constraint: `xtol` >= 0.0.

        (`errno` 13)
            On entry, `mode` = *<value>*.

            Constraint: `mode` = 1 or 2.

        (`errno` 14)
            On entry, `factor` = *<value>*.

            Constraint: `factor` > 0.0.

        (`errno` 15)
            On entry, `mode` = 2 and `diag` contained a non-positive
            element.

        (`errno` 18)
            On entry, `maxfev` = *<value>*.

            Constraint: `maxfev` > 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            There have been at least `maxfev` calls to `fcnval` and
            `fcnjac` combined: `maxfev` = *<value>*. Consider restarting
            the calculation from the final point held in `x`.

        (`errno` 3)
            No further improvement in the solution is possible. `xtol`
            is too small: `xtol` = *<value>*.

        (`errno` 4)
            The iteration is not making good progress, as measured by
            the improvement from the last *<value>* Jacobian
            evaluations. This failure exit may indicate that the system
            does not have a zero, or that the solution is very close to
            the origin (see [Accuracy]). Otherwise, rerunning
            ``sys_deriv_expert`` from a different starting point may
            avoid the region of difficulty.

        (`errno` 5)
            The iteration is not making good progress, as measured by
            the improvement from the last *<value>* iterations. This
            failure exit may indicate that the system does not have a
            zero, or that the solution is very close to the origin (see
            [Accuracy]). Otherwise, rerunning ``sys_deriv_expert`` from
            a different starting point may avoid the region of
            difficulty.

    NagCallbackTerminateWarning
        (`errno` 6)
            Termination requested in `fcnval`, `fcnjac` or `fcnmon`.

    References
    ----------
    Moré, J J, Garbow, B S and Hillstrom, K E, 1980, `User guide for
    MINPACK-1`, Technical Report ANL-80-74, Argonne National Laboratory

    Powell, M J D, 1970, `A hybrid method for nonlinear algebraic
    equations`, Numerical Methods for Nonlinear Algebraic Equations, (ed
    P Rabinowitz), Gordon and Breach
    """

    fun_name = 'naginterfaces.library.roots.sys_deriv_expert'

    cb_fcnval_name = fun_name + ':fcnval'

    def cb_fcnval(n, x, fvec, data=None):
        """Wrapping user's fcnval. Called as base fcnval."""
        if data is not None:
            _fvec_fcnval = fcnval(
                x,
                data,
            )
        else:
            _fvec_fcnval = fcnval(
                x,
            )
        fvec[:] = _utils._to_ndarray(
            {'fun_name': cb_fcnval_name, 'returned': 'fvec'},
            _fvec_fcnval,
            copy=False,
            exp_shape=(n,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )

    cb_fcnjac_name = fun_name + ':fcnjac'

    def cb_fcnjac(n, x, fjac, data=None):
        """Wrapping user's fcnjac. Called as base fcnjac."""
        if data is not None:
            _fjac_fcnjac = fcnjac(
                x,
                data,
            )
        else:
            _fjac_fcnjac = fcnjac(
                x,
            )
        _fjac_fcnjac_np = _utils._to_ndarray(
            {'fun_name': cb_fcnjac_name, 'returned': 'fjac'},
            _fjac_fcnjac,
            copy=False,
            exp_shape=(n, n),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _utils._check_callback_sorder(
            {'fun_name': cb_fcnjac_name},
            ((_fjac_fcnjac_np, 'fjac'),),
            wrapper_sorder=_m_sorder,
            spiked_sorder=spiked_sorder,
        )
        fjac[:, :] = _fjac_fcnjac_np

    def cb_fcnmon(n, x, fvec, fjac, data=None):
        """Wrapping user's fcnmon. Called as base fcnmon."""
        if data is not None:
            fcnmon(
                x,
                fvec,
                fjac,
                data,
            )
        else:
            fcnmon(
                x,
                fvec,
                fjac,
            )

    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        (),
        spiked_sorder=spiked_sorder,
    )
    _b_utils._check_callable(fcnval, {'fun_name': fun_name, 'entity': 'fcnval'})
    _fcnval = cb_fcnval
    _b_utils._check_callable(fcnjac, {'fun_name': fun_name, 'entity': 'fcnjac'})
    _fcnjac = cb_fcnjac
    _b_utils._check_callable(fcnmon, {'fun_name': fun_name, 'entity': 'fcnmon'})
    _fcnmon = cb_fcnmon
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _fvec = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _fjac = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py, _n_py),
        sorder=_b_utils._EngineData.default_sorder,
    )
    _maxfev = 100*(_n_py+1) if maxfev is None else maxfev
    _diag = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'diag'},
        diag,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _r = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py*(_n_py+1)//2,),
    )
    _qtf = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _return_scalars = _b_roots.sys_deriv_expert(
            _fcnval,
            _fcnjac,
            _fcnmon,
            _n,
            _x,
            _fvec,
            _fjac,
            xtol,
            _maxfev,
            mode,
            _diag,
            factor,
            nprint,
            _r,
            _qtf,
            data,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _x_py = _x
    _diag_py = _diag
    _fvec_py = _fvec
    _fjac_py = _fjac
    _nfev_py = _return_scalars.nfev
    _njev_py = _return_scalars.njev
    _r_py = _r
    _qtf_py = _qtf
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'fvec', 'fjac', 'diag', 'nfev', 'njev', 'r', 'qtf'],
    )(
        _x_py,
        _fvec_py,
        _fjac_py,
        _diag_py,
        _nfev_py,
        _njev_py,
        _r_py,
        _qtf_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def sys_deriv_rcomm(irevcm, x, fvec, fjac, mode, diag, r, qtf, comm, xtol=sqrt(machine.precision()), factor=100.0):
    u"""
    Solution of a system of nonlinear equations using first derivatives
    (reverse communication).

    ``sys_deriv_rcomm`` is a comprehensive reverse communication
    function that finds a solution of a system of nonlinear equations by
    a modification of the Powell hybrid method.
    You must provide the Jacobian.

    For full information please refer to the NAG Library document for
    c05rd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c05/c05rdf.html

    Parameters
    ----------
    irevcm : int
        `On initial entry`: must have the value 0.

    x : float, ndarray, shape (n), modified in place
        `On initial entry`: an initial guess at the solution vector.

        `On intermediate exit`: contains the current point.

        `On final exit`: the final estimate of the solution vector.

    fvec : float, ndarray, shape (n), modified in place
        `On initial entry`: need not be set.

        `On intermediate entry`: if `irevcm` != 2, `fvec` must not be
        changed.

        If `irevcm` = 2, `fvec` must be set to the values of the
        functions computed at the current point `x`.

        `On final exit`: the function values at the final point, `x`.

    fjac : float, ndarray, shape (n, n), modified in place
        `On initial entry`: need not be set.

        `On intermediate entry`: if `irevcm` != 3, `fjac` must not be
        changed.

        If `irevcm` = 3, `fjac`[i-1,j-1] must contain the value of
        df_i/dx_j at the point x, for j = 1,2,...,n, for i = 1,2,...,n.

        `On final exit`: the orthogonal matrix Q produced by the QR
        factorization of the final approximate Jacobian.

    mode : int
        `On initial entry`: indicates whether or not you have provided
        scaling factors in `diag`.

        If `mode` = 2, the scaling must have been supplied in `diag`.

        Otherwise, if `mode` = 1, the variables will be scaled
        internally.

    diag : float, ndarray, shape (n), modified in place
        `On initial entry`: if `mode` = 2, `diag` must contain
        multiplicative scale factors for the variables.

        If `mode` = 1, `diag` need not be set.

        `On intermediate exit`: `diag` must not be changed.

        `On final exit`: the scale factors actually used (computed
        internally if `mode` = 1).

    r : float, ndarray, shape (n*(n+1)/2), modified in place
        `On initial entry`: need not be set.

        `On intermediate exit`: must not be changed.

        `On final exit`: the upper triangular matrix R produced by the
        QR factorization of the final approximate Jacobian, stored
        row-wise.

    qtf : float, ndarray, shape (n), modified in place
        `On initial entry`: need not be set.

        `On intermediate exit`: must not be changed.

        `On final exit`: the vector Q^Tf.

    comm : dict, communication object, modified in place
        Communication structure.

        `On initial entry`: need not be set.

    xtol : float, optional
        `On initial entry`: the accuracy in `x` to which the solution is
        required.

    factor : float, optional
        `On initial entry`: a quantity to be used in determining the
        initial step bound. In most cases, `factor` should lie between
        0.1 and 100.0. (The step bound is `factor`*||`diag`*`x`||_2 if
        this is nonzero; otherwise the bound is `factor`.)

    Returns
    -------
    irevcm : int
        `On intermediate exit`: specifies what action you must take
        before re-entering ``sys_deriv_rcomm`` with `irevcm` set to this
        value. The value of `irevcm` should be interpreted as follows:

        `irevcm` = 1

            Indicates the start of a new iteration. No action is
            required by you, but `x` and `fvec` are available for
            printing.

        `irevcm` = 2

            Indicates that before re-entry to ``sys_deriv_rcomm``,
            `fvec` must contain the function values f_i(x).

        `irevcm` = 3

            Indicates that before re-entry to ``sys_deriv_rcomm``,
            `fjac`[i-1,j-1] must contain the value of df_i/dx_j at the
            point x, for j = 1,2,...,n, for i = 1,2,...,n.

        `On final exit`: `irevcm` = 0 and the algorithm has terminated.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, `irevcm` = *<value>*.

            Constraint: `irevcm` = 0, 1, 2 or 3.

        (`errno` 11)
            On entry, n = *<value>*.

            Constraint: n > 0.

        (`errno` 12)
            On entry, `xtol` = *<value>*.

            Constraint: `xtol` >= 0.0.

        (`errno` 13)
            On entry, `mode` = *<value>*.

            Constraint: `mode` = 1 or 2.

        (`errno` 14)
            On entry, `factor` = *<value>*.

            Constraint: `factor` > 0.0.

        (`errno` 15)
            On entry, `mode` = 2 and `diag` contained a non-positive
            element.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            No further improvement in the solution is possible. `xtol`
            is too small: `xtol` = *<value>*.

        (`errno` 4)
            The iteration is not making good progress, as measured by
            the improvement from the last *<value>* Jacobian
            evaluations. This failure exit may indicate that the system
            does not have a zero, or that the solution is very close to
            the origin (see [Accuracy]). Otherwise, rerunning
            ``sys_deriv_rcomm`` from a different starting point may
            avoid the region of difficulty.

        (`errno` 5)
            The iteration is not making good progress, as measured by
            the improvement from the last *<value>* iterations. This
            failure exit may indicate that the system does not have a
            zero, or that the solution is very close to the origin (see
            [Accuracy]). Otherwise, rerunning ``sys_deriv_rcomm`` from a
            different starting point may avoid the region of difficulty.

    References
    ----------
    Moré, J J, Garbow, B S and Hillstrom, K E, 1980, `User guide for
    MINPACK-1`, Technical Report ANL-80-74, Argonne National Laboratory

    Powell, M J D, 1970, `A hybrid method for nonlinear algebraic
    equations`, Numerical Methods for Nonlinear Algebraic Equations, (ed
    P Rabinowitz), Gordon and Breach
    """

    fun_name = 'naginterfaces.library.roots.sys_deriv_rcomm'

    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'fvec'},
        fvec,
    )
    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'fjac'},
        fjac,
    )
    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'diag'},
        diag,
    )
    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'r'},
        r,
    )
    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'qtf'},
        qtf,
    )
    try:
        _return_scalar = _b_roots.sys_deriv_rcomm(
            irevcm,
            _n,
            x,
            fvec,
            fjac,
            xtol,
            mode,
            diag,
            factor,
            r,
            qtf,
            comm,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _irevcm_py = _return_scalar
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['irevcm'],
    )(
        _irevcm_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _irevcm_py

def sys_deriv_check(mode, x, fvec, fjac, fvecp):
    u"""
    Check user's function for calculating first derivatives of a set of
    nonlinear functions of several variables.

    ``sys_deriv_check`` checks the user-supplied gradients of a set of
    nonlinear functions in several variables, for consistency with the
    functions themselves.
    The function must be called twice.

    For full information please refer to the NAG Library document for
    c05zd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c05/c05zdf.html

    Parameters
    ----------
    mode : int
        The value 1 on the first call and the value 2 on the second call
        of ``sys_deriv_check``.

    x : float, array-like, shape (n)
        The components of a point x, at which the consistency check is
        to be made. (See [Accuracy].)

    fvec : float, array-like, shape (m)
        If `mode` = 2, `fvec` must contain the value of the functions
        evaluated at x. If `mode` = 1, `fvec` is not referenced.

    fjac : float, array-like, shape (m, n)
        If `mode` = 2, `fjac` must contain the value of df_i/dx_j at the
        point x, for j = 1,2,...,n, for i = 1,2,...,m. If `mode` = 1,
        `fjac` is not referenced.

    fvecp : float, array-like, shape (m)
        If `mode` = 2, `fvecp` must contain the value of the functions
        evaluated at `xp` (as output by a preceding call to
        ``sys_deriv_check`` with `mode` = 1). If `mode` = 1, `fvecp` is
        not referenced.

    Returns
    -------
    xp : float, ndarray, shape (n)
        If `mode` = 1, `xp` is set to a point neighbouring `x`. If
        `mode` = 2, `xp` is undefined.

    err : float, ndarray, shape (m)
        If `mode` = 2, `err` contains measures of correctness of the
        respective gradients. If `mode` = 1, `err` is undefined. If
        there is no loss of significance (see [Accuracy]), if `err`[i-1]
        is 1.0 the ith user-supplied gradient df_i/dx_j, for j =
        1,2,...,n is correct, whilst if `err`[i-1] is 0.0 the ith
        gradient is incorrect. For values of `err`[i-1] between 0.0 and
        1.0 the categorisation is less certain. In general, a value of
        `err`[i-1] > 0.5 indicates that the ith gradient is probably
        correct.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `mode` = *<value>*.

            Constraint: `mode` = 1 or 2.

        (`errno` 2)
            On entry, m = *<value>*.

            Constraint: m >= 1.

        (`errno` 3)
            On entry, n = *<value>*.

            Constraint: n >= 1.

    References
    ----------
    Moré, J J, Garbow, B S and Hillstrom, K E, 1980, `User guide for
    MINPACK-1`, Technical Report ANL-80-74, Argonne National Laboratory
    """

    fun_name = 'naginterfaces.library.roots.sys_deriv_check'

    _fvec = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'fvec'},
        fvec,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_fvec, {'fun_name': fun_name, 'entity': 'fvec'}, exp_rank=1)[0]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _fjac = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'fjac'},
        fjac,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _xp = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _fvecp = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'fvecp'},
        fvecp,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _err = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_m_py,),
    )
    _b_roots.sys_deriv_check(
        mode,
        _m,
        _n,
        _x,
        _fvec,
        _fjac,
        _xp,
        _fvecp,
        _err,
    )
    _xp_py = _xp
    _err_py = _err
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['xp', 'err'],
    )(
        _xp_py,
        _err_py,
    )
    return _return_data
