# -*- coding: utf-8 -*-
# pylint: disable=chained-comparison,invalid-name,line-too-long,protected-access,too-many-arguments,too-many-boolean-expressions,too-many-branches,too-many-lines,too-many-locals,too-many-statements,unused-import
u"""
Interfaces for the NAG Mark 27.0 `smooth` Chapter.

``smooth`` - Smoothing in Statistics

This module is concerned with methods for smoothing data.
Included are methods for density estimation, smoothing time series data,
and statistical applications of splines.
These methods may also be viewed as nonparametric modelling.

For full information please refer to the NAG Library document

https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g10/g10intro.html
"""

# NAG Copyright 2017-2020.

import collections as _collections

from ..base import smooth as _b_smooth
from ..base import utils as _b_utils
from ..library import _utils

def fit_spline(mode, x, y, rho, c, comm, wt=None):
    u"""
    Fit cubic smoothing spline, smoothing parameter given.

    ``fit_spline`` fits a cubic smoothing spline for a given smoothing
    parameter.

    For full information please refer to the NAG Library document for
    g10ab

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g10/g10abf.html

    Parameters
    ----------
    mode : str, length 1
        Indicates in which mode the function is to be used.

        `mode` = 'P'

            Initialization and fitting is performed. This partial fit
            can be used in an iterative weighted least squares context
            where the weights are changing at each call to
            ``fit_spline`` or when the coefficients are not required.

        `mode` = 'Q'

            Fitting only is performed. Initialization must have been
            performed previously by a call to ``fit_spline`` with `mode`
            = 'P'. This quick fit may be called repeatedly with
            different values of `rho` without re-initialization.

        `mode` = 'F'

            Initialization and full fitting is performed and the
            function coefficients are calculated.

    x : float, array-like, shape (n)
        The distinct and ordered values x_i, for i = 1,2,...,n.

    y : float, array-like, shape (n)
        The values y_i, for i = 1,2,...,n.

    rho : float
        rho, the smoothing parameter.

    c : float, array-like, shape (n-1, 3)
        If `mode` = 'Q', `c` must be unaltered from the previous call to
        ``fit_spline`` with `mode` = 'P'. Otherwise `c` need not be set.

    comm : dict, communication object, modified in place
        Communication structure.

        `On initial entry`: need not be set.

    wt : None or float, array-like, shape (n), optional
        If weight = 'W', `wt` must contain the n weights. Otherwise `wt`
        is not referenced and unit weights are assumed.

    Returns
    -------
    yhat : float, ndarray, shape (n)
        The fitted values, y^_i, for i = 1,2,...,n.

    c : float, ndarray, shape (n-1, 3)
        If `mode` = 'F', `c` contains the spline coefficients. More
        precisely, the value of the spline at t is given by
        ((`c`[i-1,2]*d+`c`[i-1,1])*d+`c`[i-1,0])*d+y^_i, where x_i <= t
        < x_{i+1} and d = t-x_i.

        If `mode` = 'P' or 'Q', `c` contains information that will be
        used in a subsequent call to ``fit_spline`` with `mode` = 'Q'.

    rss : float
        The (weighted) residual sum of squares.

    df : float
        The residual degrees of freedom.

    res : float, ndarray, shape (n)
        The (weighted) residuals, r_i, for i = 1,2,...,n.

    h : float, ndarray, shape (n)
        The leverages, h_{ii}, for i = 1,2,...,n.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `mode` = *<value>*.

            Constraint: `mode` = 'P', 'Q' or 'F'.

        (`errno` 1)
            On entry, `rho` = *<value>*.

            Constraint: `rho` >= 0.0.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 3.

        (`errno` 2)
            On entry, at least one element of `wt` <= 0.0.

        (`errno` 3)
            On entry, `x` is not a strictly ordered array.

    References
    ----------
    Hastie, T J and Tibshirani, R J, 1990, `Generalized Additive
    Models`, Chapman and Hall

    Hutchinson, M F, 1986, `Algorithm 642: A fast procedure for
    calculating minimum cross-validation cubic smoothing splines`, ACM
    Trans. Math. Software (12), 150--153

    Hutchinson, M F and de Hoog, F R, 1985, `Smoothing noisy data with
    spline functions`, Numer. Math. (47), 99--106

    Reinsch, C H, 1967, `Smoothing by spline functions`, Numer. Math.
    (10), 177--183
    """

    fun_name = 'naginterfaces.library.smooth.fit_spline'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _wt = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'wt'},
        wt,
        copy=False,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _yhat = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _c = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _res = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _h = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _return_scalars = _b_smooth.fit_spline(
        mode,
        _n,
        _x,
        _y,
        _wt,
        rho,
        _yhat,
        _c,
        _res,
        _h,
        comm,
    )
    _c_py = _c
    _yhat_py = _yhat
    _rss_py = _return_scalars.rss
    _df_py = _return_scalars.df
    _res_py = _res
    _h_py = _h
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['yhat', 'c', 'rss', 'df', 'res', 'h'],
    )(
        _yhat_py,
        _c_py,
        _rss_py,
        _df_py,
        _res_py,
        _h_py,
    )
    return _return_data

def fit_spline_parest(method, x, y, crit, wt=None, u=0.0, tol=0.0, maxcal=0):
    u"""
    Fit cubic smoothing spline, smoothing parameter estimated.

    ``fit_spline_parest`` estimates the values of the smoothing
    parameter and fits a cubic smoothing spline to a set of data.

    For full information please refer to the NAG Library document for
    g10ac

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g10/g10acf.html

    Parameters
    ----------
    method : str, length 1
        Indicates whether the smoothing parameter is to be found by
        minimization of the CV or GCV functions, or by finding the
        smoothing parameter corresponding to a specified degrees of
        freedom value.

        `method` = 'C'

            Cross-validation is used.

        `method` = 'D'

            The degrees of freedom are specified.

        `method` = 'G'

            Generalized cross-validation is used.

    x : float, array-like, shape (n)
        The distinct and ordered values x_i, for i = 1,2,...,n.

    y : float, array-like, shape (n)
        The values y_i, for i = 1,2,...,n.

    crit : float
        If `method` = 'D', the required degrees of freedom for the
        spline.

        If `method` = 'C' or 'G', `crit` need not be set.

    wt : None or float, array-like, shape (n), optional
        If weight = 'W', `wt` must contain the n weights. Otherwise `wt`
        is not referenced and unit weights are assumed.

    u : float, optional
        The upper bound on the smoothing parameter. If `u` <= `tol`, `u`
        = 1000.0 will be used instead. See [Further Comments] for
        details on how this argument is used.

    tol : float, optional
        The accuracy to which the smoothing parameter `rho` is required.
        `tol` should preferably be not much less than sqrt(epsilon),
        where epsilon is the machine precision. If `tol` < epsilon,
        `tol` = sqrt(epsilon) will be used instead.

    maxcal : int, optional
        The maximum number of spline evaluations to be used in finding
        the value of rho. If `maxcal` < 3, `maxcal` = 100 will be used
        instead.

    Returns
    -------
    yhat : float, ndarray, shape (n)
        The fitted values, y^_i, for i = 1,2,...,n.

    c : float, ndarray, shape (n-1, 3)
        The spline coefficients. More precisely, the value of the spline
        approximation at t is given by
        ((`c`[i-1,2]*d+`c`[i-1,1])*d+`c`[i-1,0])*d+y^_i, where x_i <= t
        < x_{i+1} and d = t-x_i.

    rss : float
        The (weighted) residual sum of squares.

    df : float
        The residual degrees of freedom. If `method` = 'D' this will be
        n-`crit` to the required accuracy.

    res : float, ndarray, shape (n)
        The (weighted) residuals, r_i, for i = 1,2,...,n.

    h : float, ndarray, shape (n)
        The leverages, h_{ii}, for i = 1,2,...,n.

    crit : float
        If `method` = 'C', the value of the cross-validation, or if
        `method` = 'G', the value of the generalized cross-validation
        function, evaluated at the value of rho returned in `rho`.

    rho : float
        The smoothing parameter, rho.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `crit` = *<value>*.

            Constraint: if `method` = 'D', `crit` <= n.

        (`errno` 1)
            On entry, `crit` = *<value>*.

            Constraint: if `method` = 'D', `crit` > 2.0.

        (`errno` 1)
            On entry, `method` is not valid: `method` = *<value>*.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 3.

        (`errno` 2)
            On entry, at least one element of `wt` <= 0.0.

        (`errno` 3)
            On entry, `x` is not a strictly ordered array.

        (`errno` 4)
            For the specified degrees of freedom, `rho` > `u`: `u` =
            *<value>*.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 5)
            Accuracy of `tol` cannot be achieved: `tol` = *<value>*.

        (`errno` 6)
            `maxcal` iterations have been performed.

        (`errno` 7)
            Optimum value of `rho` lies above `u`: `u` = *<value>*.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Hastie, T J and Tibshirani, R J, 1990, `Generalized Additive
    Models`, Chapman and Hall

    Hutchinson, M F, 1986, `Algorithm 642: A fast procedure for
    calculating minimum cross-validation cubic smoothing splines`, ACM
    Trans. Math. Software (12), 150--153

    Hutchinson, M F and de Hoog, F R, 1985, `Smoothing noisy data with
    spline functions`, Numer. Math. (47), 99--106

    Reinsch, C H, 1967, `Smoothing by spline functions`, Numer. Math.
    (10), 177--183
    """

    fun_name = 'naginterfaces.library.smooth.fit_spline_parest'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _wt = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'wt'},
        wt,
        copy=False,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _yhat = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _c = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py-1, 3),
        sorder=_b_utils._EngineData.default_sorder,
    )
    _res = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _h = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _return_scalars = _b_smooth.fit_spline_parest(
            method,
            _n,
            _x,
            _y,
            _wt,
            _yhat,
            _c,
            _res,
            _h,
            crit,
            u,
            tol,
            maxcal,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _crit_py = _return_scalars.crit
    _yhat_py = _yhat
    _c_py = _c
    _rss_py = _return_scalars.rss
    _df_py = _return_scalars.df
    _res_py = _res
    _h_py = _h
    _rho_py = _return_scalars.rho
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['yhat', 'c', 'rss', 'df', 'res', 'h', 'crit', 'rho'],
    )(
        _yhat_py,
        _c_py,
        _rss_py,
        _df_py,
        _res_py,
        _h_py,
        _crit_py,
        _rho_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def kerndens_gauss(x, comm, wtype=2, window=1.0, slo=None, shi=None, ns=512):
    u"""
    Kernel density estimate using Gaussian kernel (thread safe).

    ``kerndens_gauss`` performs kernel density estimation using a
    Gaussian kernel.

    For full information please refer to the NAG Library document for
    g10bb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g10/g10bbf.html

    Parameters
    ----------
    x : float, array-like, shape (n)
        x_i, for i = 1,2,...,n.

        On a continuation call, `x` must be unchanged since the last
        call to ``kerndens_gauss``.

    comm : dict, communication object, modified in place
        Communication structure.

        If not initialized on entry then the values of Y_l are to be
        calculated by this call to ``kerndens_gauss``.

        Otherwise, this is a continuation call and it is assumed that
        the values of Y_l were calculated by a previous call to this
        function and the relevant information is stored in
        `comm`['rcomm'].

    wtype : int, optional
        How the window width, h, is to be calculated:

        `wtype` = 1

            h is supplied in `window`.

        `wtype` = 2

            h is to be calculated from the data, with

                h = m*(0.9*min(q_75-q_25, sigma)/(n^0.2))

            where q_75-q_25 is the inter-quartile range and sigma the
            standard deviation of the sample, x, and m is a multipler
            supplied in `window`. The 25% and 75% quartiles, q_25 and
            q_75, are calculated using ``stat.quantiles``. This is the
            'rule-of-thumb' suggested by Silverman (1990).

    window : float, optional
        If `wtype` = 1, then h, the window width. Otherwise, m, the
        multiplier used in the calculation of h.

    slo : None or float, optional
        Note: if this argument is None then a default value will be
        used, determined as follows: 3.0.

        If `slo` < `shi` then a, the lower limit of the interval on
        which the estimate is calculated. Otherwise, a and b, the lower
        and upper limits of the interval, are calculated as follows:

            [table omitted]

        where h is the window width.

        For most applications a should be at least three window widths
        below the lowest data point.

        On a continuation call, a supplied `slo` will be ignored and the
        appropriate value from the previous call to ``kerndens_gauss``
        will be extracted from `comm`['rcomm'].

    shi : None or float, optional
        Note: if this argument is None then a default value will be
        used, determined as follows: 0.0.

        If `slo` < `shi` then b, the upper limit of the interval on
        which the estimate is calculated. Otherwise a value for b is
        calculated from the data as stated in the description of `slo`
        and the value supplied in `shi` is not used.

        For most applications b should be at least three window widths
        above the highest data point.

        On a continuation call, a supplied `shi` will be ignored and the
        appropriate value from the previous call to ``kerndens_gauss``
        will be extracted from `comm`['rcomm'].

    ns : int, optional
        n_s, the number of points at which the estimate is calculated.

        On a continuation call, `ns` must be unchanged since the last
        call to ``kerndens_gauss``.

    Returns
    -------
    window : float
        h, the window width actually used.

    slo : float
        a, the lower limit actually used.

    shi : float
        b, the upper limit actually used.

    smooth : float, ndarray, shape (`ns`)
        f^(t_l), for l = 1,2,...,n_s, the n_s values of the density
        estimate.

    t : float, ndarray, shape (`ns`)
        t_l, for l = 1,2,...,n_s, the points at which the estimate is
        calculated.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, n = *<value>*.

            Constraint: n > 0.

        (`errno` 12)
            On entry, n = *<value>*.

            On entry at previous call, n = *<value>*.

            Constraint: on a continuation call, n must be unchanged
            since previous call.

        (`errno` 31)
            On entry, `wtype` = *<value>*.

            Constraint: `wtype` = 1 or 2.

        (`errno` 41)
            On entry, `window` = *<value>*.

            Constraint: `window` > 0.0.

        (`errno` 51)
            On entry, `slo` = *<value>*.

            On exit from previous call, `slo` = *<value>*.

            Constraint: on a continuation call, `slo` must be the same
            value returned by the previous call.

        (`errno` 62)
            On entry, `shi` = *<value>*.

            On exit from previous call, `shi` = *<value>*.

            Constraint: on a continuation call, `shi` must be the same
            value returned by the previous call.

        (`errno` 71)
            On entry, `ns` = *<value>*.

            Constraint: `ns` >= 2.

        (`errno` 74)
            On entry, `ns` = *<value>*.

            On entry at previous call, `ns` = *<value>*.

            Constraint: on a continuation call, `ns` must be unchanged
            since previous call.

        (`errno` 111)
            `comm`['rcomm'] has been corrupted between calls.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 61)
            On entry, `slo` = *<value>* and `shi` = *<value>*.

            On entry, min(`x`) = *<value>* and max(`x`) = *<value>*.

            Expected values of at least *<value>* and *<value>* for
            `slo` and `shi`.

    References
    ----------
    Jones, M C and Lotwick, H W, 1984, `Remark AS R50. A remark on
    algorithm AS 176. Kernel density estimation using the Fast Fourier
    Transform`, Appl. Statist. (33), 120--122

    Silverman, B W, 1982, `Algorithm AS 176. Kernel density estimation
    using the fast Fourier transform`, Appl. Statist. (31), 93--99

    Silverman, B W, 1990, `Density Estimation`, Chapman and Hall
    """

    fun_name = 'naginterfaces.library.smooth.kerndens_gauss'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _comm_needs_init = _b_utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        opt_keys=('rcomm',),
    )
    if _comm_needs_init:
        _slo = 3.0
    else:
        _slo = _b_utils._EngineFloat64ArrayType.get_element(comm['rcomm'], {'fun_name': fun_name, 'entity': 'comm[\'rcomm\']'}, (4,), validate=False)
    _slo = _slo if slo is None else slo
    if _comm_needs_init:
        _shi = 0.0
    else:
        _shi = _b_utils._EngineFloat64ArrayType.get_element(comm['rcomm'], {'fun_name': fun_name, 'entity': 'comm[\'rcomm\']'}, (3,), validate=False)
    _shi = _shi if shi is None else shi
    _ns = _b_utils._EngineIntScalarType(
        ns, {'fun_name': fun_name, 'entity': 'ns'},
    ).to_ctypes()
    _ns_py = _b_utils._EngineIntScalarType.to_py(_ns)
    _smooth = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ns_py,),
    )
    _t = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ns_py,),
    )
    try:
        _return_scalars = _b_smooth.kerndens_gauss(
            _n,
            _x,
            wtype,
            window,
            _slo,
            _shi,
            _ns,
            _smooth,
            _t,
            comm,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _window_py = _return_scalars.window
    _slo_py = _return_scalars.slo
    _shi_py = _return_scalars.shi
    _smooth_py = _smooth
    _t_py = _t
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['window', 'slo', 'shi', 'smooth', 't'],
    )(
        _window_py,
        _slo_py,
        _shi_py,
        _smooth_py,
        _t_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def data_runningmedian(itype, y):
    u"""
    Compute smoothed data sequence using running median smoothers.

    ``data_runningmedian`` computes a smoothed data sequence using
    running median smoothers.

    For full information please refer to the NAG Library document for
    g10ca

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g10/g10caf.html

    Parameters
    ----------
    itype : int
        Specifies the method to be used.

            If `itype` = 0, 4253H,twice is used.

            If `itype` = 1, 3RSSH,twice is used.

    y : float, array-like, shape (n)
        The sample observations.

    Returns
    -------
    smooth : float, ndarray, shape (n)
        Contains the smooth.

    rough : float, ndarray, shape (n)
        Contains the rough.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `itype` = *<value>*.

            Constraint: `itype` = 0 or 1.

        (`errno` 2)
            On entry, n = *<value>*.

            Constraint: n > 6.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Tukey, J W, 1977, `Exploratory Data Analysis`, Addison--Wesley

    Velleman, P F and Hoaglin, D C, 1981, `Applications, Basics, and
    Computing of Exploratory Data Analysis`, Duxbury Press, Boston, MA
    """

    fun_name = 'naginterfaces.library.smooth.data_runningmedian'

    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_y, {'fun_name': fun_name, 'entity': 'y'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _smooth = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _rough = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _b_smooth.data_runningmedian(
        itype,
        _n,
        _y,
        _smooth,
        _rough,
    )
    _smooth_py = _smooth
    _rough_py = _rough
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['smooth', 'rough'],
    )(
        _smooth_py,
        _rough_py,
    )
    return _return_data

def data_order(x, y, wt=None):
    u"""
    Reorder data to give ordered distinct observations.

    ``data_order`` orders and weights data which is entered
    unsequentially, weighted or unweighted.

    For full information please refer to the NAG Library document for
    g10za

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g10/g10zaf.html

    Parameters
    ----------
    x : float, array-like, shape (n)
        The values, x_i, for i = 1,2,...,n.

    y : float, array-like, shape (n)
        The values y_i, for i = 1,2,...,n.

    wt : None or float, array-like, shape (n), optional
        If weight = 'W', `wt` must contain the n weights. Otherwise `wt`
        is not referenced and unit weights are assumed.

    Returns
    -------
    nord : int
        The number of distinct observations.

    xord : float, ndarray, shape (n)
        The first `nord` elements contain the ordered and distinct x_i.

    yord : float, ndarray, shape (n)
        The first `nord` elements contain the values y' corresponding to
        the values in `xord`.

    wtord : float, ndarray, shape (n)
        The first `nord` elements contain the values w' corresponding to
        the values of `xord` and `yord`.

    rss : float
        The within group sum of squares for tied observations.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, weight = *<value>*.

            Constraint: weight = 'W' or 'U'.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 2)
            On entry, all weights are zero.

        (`errno` 2)
            On entry, i = *<value>* and `wt`[i-1] = *<value>*.

            Constraint: `wt`[i-1] >= 0.0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Draper, N R and Smith, H, 1985, `Applied Regression Analysis`, (2nd
    Edition), Wiley

    West, D H D, 1979, `Updating mean and variance estimates: An
    improved method`, Comm. ACM (22), 532--555
    """

    fun_name = 'naginterfaces.library.smooth.data_order'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _wt = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'wt'},
        wt,
        copy=False,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _xord = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _yord = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _wtord = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _return_scalars = _b_smooth.data_order(
        _n,
        _x,
        _y,
        _wt,
        _xord,
        _yord,
        _wtord,
    )
    _nord_py = _return_scalars.nord
    _xord_py = _xord
    _yord_py = _yord
    _wtord_py = _wtord
    _rss_py = _return_scalars.rss
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['nord', 'xord', 'yord', 'wtord', 'rss'],
    )(
        _nord_py,
        _xord_py,
        _yord_py,
        _wtord_py,
        _rss_py,
    )
    return _return_data
