# -*- coding: utf-8 -*-
# pylint: disable=chained-comparison,invalid-name,line-too-long,protected-access,too-many-arguments,too-many-boolean-expressions,too-many-branches,too-many-lines,too-many-locals,too-many-statements,unused-import
u"""
Interfaces for the NAG Mark 27.0 `zeros` Chapter.

``zeros`` - Zeros of Polynomials

This module is concerned with computing the zeros of a polynomial with
real or complex coefficients.

For full information please refer to the NAG Library document

https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c02/c02intro.html

See Also
--------
``naginterfaces.library.examples.zeros`` :
    This subpackage contains examples for the ``zeros`` module.
    See also the :ref:`library_zeros_ex` subsection.
"""

# NAG Copyright 2017-2020.

import collections as _collections

from ..base import utils as _b_utils
from ..base import zeros as _b_zeros
from ..library import _utils

def poly_complex(a, scal=True):
    u"""
    All zeros of complex polynomial, modified Laguerre's method.

    ``poly_complex`` finds all the roots of a complex polynomial
    equation, using a variant of Laguerre's method.

    For full information please refer to the NAG Library document for
    c02af

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c02/c02aff.html

    Parameters
    ----------
    a : complex, array-like, shape (n+1)
        Re(`a`[i]) and Im(`a`[i]) must contain the real and imaginary
        parts of a_i (i.e., the coefficient of z^(n-i)), for i =
        0,1,...,n.

    scal : bool, optional
        Indicates whether or not the polynomial is to be scaled. See
        [Further Comments] for advice on when it may be preferable to
        set `scal` = False and for a description of the scaling
        strategy.

    Returns
    -------
    z : complex, ndarray, shape (n)
        The roots.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `a`[0] = 0.0.

            Constraint: `a`[0] != 0.0.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 2)
            The iterative procedure has failed to converge. This error
            is very unlikely to occur. If it does, please contact NAG
            immediately, as some basic assumption for the arithmetic has
            been violated.

        (`errno` 3)
            ``poly_complex`` cannot evaluate P(z) near some of its zeros
            without underflow. If this message occurs please contact
            NAG.

        (`errno` 3)
            ``poly_complex`` cannot evaluate P(z) near some of its zeros
            without overflow. If this message occurs please contact NAG.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Marden, M, 1966, `Geometry of polynomials`, Mathematical Surveys
    (3), American Mathematical Society, Providence, RI

    Smith, B T, 1967, `ZERPOL: a zero finding algorithm for polynomials
    using Laguerre's method`, Technical Report, Department of Computer
    Science, University of Toronto, Canada

    Thompson, K W, 1991, `Error analysis for polynomial solvers`,
    Fortran Journal (Volume 3) (3), 10--13

    Wilkinson, J H, 1965, `The Algebraic Eigenvalue Problem`, Oxford
    University Press, Oxford

    See Also
    --------
    :meth:`naginterfaces.library.examples.zeros.poly_complex_ex.main`
    """

    fun_name = 'naginterfaces.library.zeros.poly_complex'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineComplex128ArrayType,
    )
    _n = _b_utils._EngineIntScalarType(
        _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=1)[0]-1, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _a = _utils._complex_to_float_array(
        _a,
        sorder=_b_utils._EngineData.default_sorder,
    )
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _z = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (2, _n_py),
        sorder=_b_utils._EngineData.default_sorder,
    )
    _b_zeros.poly_complex(
        _a,
        _n,
        scal,
        _z,
    )
    _z_py = _utils._float_array_to_complex(_z)
    return _z_py

def poly_real(a, scal=True):
    u"""
    All zeros of real polynomial, modified Laguerre's method.

    ``poly_real`` finds all the roots of a real polynomial equation,
    using a variant of Laguerre's method.

    For full information please refer to the NAG Library document for
    c02ag

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c02/c02agf.html

    Parameters
    ----------
    a : float, array-like, shape (n+1)
        `a`[i] must contain a_i (i.e., the coefficient of z^(n-i)), for
        i = 0,1,...,n.

    scal : bool, optional
        Indicates whether or not the polynomial is to be scaled. See
        [Further Comments] for advice on when it may be preferable to
        set `scal` = False and for a description of the scaling
        strategy.

    Returns
    -------
    z : complex, ndarray, shape (n)
        The roots. Complex conjugate pairs of roots are stored in
        consecutive elements of `z`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `a`[0] = 0.0.

            Constraint: `a`[0] != 0.0.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 2)
            The iterative procedure has failed to converge. This error
            is very unlikely to occur. If it does, please contact NAG
            immediately, as some basic assumption for the arithmetic has
            been violated.

        (`errno` 3)
            ``poly_real`` cannot evaluate P(z) near some of its zeros
            without underflow. If this message occurs please contact
            NAG.

        (`errno` 3)
            ``poly_real`` cannot evaluate P(z) near some of its zeros
            without overflow. If this message occurs please contact NAG.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Marden, M, 1966, `Geometry of polynomials`, Mathematical Surveys
    (3), American Mathematical Society, Providence, RI

    Smith, B T, 1967, `ZERPOL: a zero finding algorithm for polynomials
    using Laguerre's method`, Technical Report, Department of Computer
    Science, University of Toronto, Canada

    Thompson, K W, 1991, `Error analysis for polynomial solvers`,
    Fortran Journal (Volume 3) (3), 10--13

    Wilkinson, J H, 1965, `The Algebraic Eigenvalue Problem`, Oxford
    University Press, Oxford
    """

    fun_name = 'naginterfaces.library.zeros.poly_real'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n = _b_utils._EngineIntScalarType(
        _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=1)[0]-1, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _z = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (2, _n_py),
        sorder=_b_utils._EngineData.default_sorder,
    )
    _b_zeros.poly_real(
        _a,
        _n,
        scal,
        _z,
    )
    _z_py = _utils._float_array_to_complex(_z)
    return _z_py

def quadratic_complex(a, b, c):
    u"""
    All zeros of complex quadratic equation.

    ``quadratic_complex`` determines the roots of a quadratic equation
    with complex coefficients.

    For full information please refer to the NAG Library document for
    c02ah

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c02/c02ahf.html

    Parameters
    ----------
    a : complex
        a, the coefficient of z^2.

    b : complex
        b, the coefficient of z.

    c : complex
        c, the constant coefficient.

    Returns
    -------
    z_small : complex
        The smallest root in magnitude.

    z_large : complex
        The largest root in magnitude.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, `a` = 0.0.

        (`errno` 2)
            On entry, `a` = 0.0 and `b` = 0.0.

        (`errno` 3)
            On entry, `a` = 0.0 and the root -`c`/`b` overflows: Re(`a`)
            = *<value>*, Re(`c`) = *<value>*, Re(`b`) = *<value>*,
            Im(`a`) = *<value>*, Im(`c`) = *<value>*, Im(`b`) =
            *<value>*.

        (`errno` 4)
            On entry, `c` = 0.0 and the root -`b`/`a` overflows: Re(`c`)
            = *<value>*, Re(`b`) = *<value>*, Re(`a`) = *<value>*,
            Im(`c`) = *<value>*, Im(`b`) = *<value>*, Im(`a`) =
            *<value>*.

        (`errno` 5)
            On entry, B = max(abs(Re(`b`)), abs(Im(`b`))) = *<value>*, A
            = max(abs(Re(`a`)), abs(Im(`a`))) = *<value>* and C =
            max(abs(Re(`c`)), abs(Im(`c`))) = *<value>*. B is so large
            that B^2 is indistinguishable from (B^2-4*A*C) and the root
            -`b`/`a` overflows. Re(`b`) = *<value>*, Im(`b`) =
            *<value>*, Re(`a`) = *<value>* and Im(`a`) = *<value>*.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Smith, B T, 1967, `ZERPOL: a zero finding algorithm for polynomials
    using Laguerre's method`, Technical Report, Department of Computer
    Science, University of Toronto, Canada
    """

    fun_name = 'naginterfaces.library.zeros.quadratic_complex'

    _a = _b_utils._EngineComplex128ScalarType(
        a, {'fun_name': fun_name, 'entity': 'a'},
    ).to_ctypes()
    _a_py = _b_utils._EngineComplex128ScalarType.to_py(_a)
    _ar = _b_utils._EngineFloat64ScalarType(
        _a_py.real, {'fun_name': fun_name, 'entity': '_ar'},
    ).to_ctypes()
    _ai = _b_utils._EngineFloat64ScalarType(
        _a_py.imag, {'fun_name': fun_name, 'entity': '_ai'},
    ).to_ctypes()
    _b = _b_utils._EngineComplex128ScalarType(
        b, {'fun_name': fun_name, 'entity': 'b'},
    ).to_ctypes()
    _b_py = _b_utils._EngineComplex128ScalarType.to_py(_b)
    _br = _b_utils._EngineFloat64ScalarType(
        _b_py.real, {'fun_name': fun_name, 'entity': '_br'},
    ).to_ctypes()
    _bi = _b_utils._EngineFloat64ScalarType(
        _b_py.imag, {'fun_name': fun_name, 'entity': '_bi'},
    ).to_ctypes()
    _c = _b_utils._EngineComplex128ScalarType(
        c, {'fun_name': fun_name, 'entity': 'c'},
    ).to_ctypes()
    _c_py = _b_utils._EngineComplex128ScalarType.to_py(_c)
    _cr = _b_utils._EngineFloat64ScalarType(
        _c_py.real, {'fun_name': fun_name, 'entity': '_cr'},
    ).to_ctypes()
    _ci = _b_utils._EngineFloat64ScalarType(
        _c_py.imag, {'fun_name': fun_name, 'entity': '_ci'},
    ).to_ctypes()
    _zsm = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (2,),
    )
    _zlg = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (2,),
    )
    try:
        _b_zeros.quadratic_complex(
            _ar,
            _ai,
            _br,
            _bi,
            _cr,
            _ci,
            _zsm,
            _zlg,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _z_small_py = _b_utils._EngineFloat64ArrayType.get_element(_zsm, {'fun_name': fun_name, 'entity': '_zsm'}, (0,), validate=False) + _b_utils._EngineFloat64ArrayType.get_element(_zsm, {'fun_name': fun_name, 'entity': '_zsm'}, (1,), validate=False)*1j
    _z_large_py = _b_utils._EngineFloat64ArrayType.get_element(_zlg, {'fun_name': fun_name, 'entity': '_zlg'}, (0,), validate=False) + _b_utils._EngineFloat64ArrayType.get_element(_zlg, {'fun_name': fun_name, 'entity': '_zlg'}, (1,), validate=False)*1j
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['z_small', 'z_large'],
    )(
        _z_small_py,
        _z_large_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def quadratic_real(a, b, c):
    u"""
    All zeros of real quadratic equation.

    ``quadratic_real`` determines the roots of a quadratic equation with
    real coefficients.

    For full information please refer to the NAG Library document for
    c02aj

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c02/c02ajf.html

    Parameters
    ----------
    a : float
        Must contain a, the coefficient of z^2.

    b : float
        Must contain b, the coefficient of z.

    c : float
        Must contain c, the constant coefficient.

    Returns
    -------
    z_small : complex
        The smallest root in magnitude.

    z_large : complex
        The largest root in magnitude.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, `a` = 0.0.

        (`errno` 2)
            On entry, `a` = 0.0 and `b` = 0.0.

        (`errno` 3)
            On entry, `a` = 0.0 and the root -`c`/`b` overflows: `a` =
            *<value>*, `c` = *<value>* and `b` = *<value>*.

        (`errno` 4)
            On entry, `c` = 0.0 and the root -`b`/`a` overflows: `c` =
            *<value>*, `b` = *<value>* and `a` = *<value>*.

        (`errno` 5)
            On entry, `b` is so large that `b`^2 is indistinguishable
            from (`b`^2-4*`a`*`c`) and the root -`b`/`a` overflows: `b`
            = *<value>*, `a` = *<value>* and `c` = *<value>*.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Smith, B T, 1967, `ZERPOL: a zero finding algorithm for polynomials
    using Laguerre's method`, Technical Report, Department of Computer
    Science, University of Toronto, Canada
    """

    _zsm = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (2,),
    )
    _zlg = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (2,),
    )
    try:
        _b_zeros.quadratic_real(
            a,
            b,
            c,
            _zsm,
            _zlg,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    fun_name = 'naginterfaces.library.zeros.quadratic_real'

    _z_small_py = _b_utils._EngineFloat64ArrayType.get_element(_zsm, {'fun_name': fun_name, 'entity': '_zsm'}, (0,), validate=False) + _b_utils._EngineFloat64ArrayType.get_element(_zsm, {'fun_name': fun_name, 'entity': '_zsm'}, (1,), validate=False)*1j
    _z_large_py = _b_utils._EngineFloat64ArrayType.get_element(_zlg, {'fun_name': fun_name, 'entity': '_zlg'}, (0,), validate=False) + _b_utils._EngineFloat64ArrayType.get_element(_zlg, {'fun_name': fun_name, 'entity': '_zlg'}, (1,), validate=False)*1j
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['z_small', 'z_large'],
    )(
        _z_small_py,
        _z_large_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def cubic_real(u, r, s, t):
    u"""
    All zeros of real cubic equation.

    ``cubic_real`` determines the roots of a cubic equation with real
    coefficients.

    For full information please refer to the NAG Library document for
    c02ak

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c02/c02akf.html

    Parameters
    ----------
    u : float
        u, the coefficient of z^3.

    r : float
        r, the coefficient of z^2.

    s : float
        s, the coefficient of z.

    t : float
        t, the constant coefficient.

    Returns
    -------
    zero : complex, ndarray, shape (3)
        `zero`[i-1] contains the ith root.

    errest : float, ndarray, shape (3)
        `errest`[i-1] contains an approximate error estimate for the ith
        root.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `u` = 0.0.

            Constraint: `u` != 0.0.

        (`errno` 2)
            The companion matrix H cannot be formed without overflow.

        (`errno` 3)
            Failure to converge in ``lapackeig.dhseqr``.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    _zeror = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (3,),
    )
    _zeroi = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (3,),
    )
    _errest = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (3,),
    )
    _b_zeros.cubic_real(
        u,
        r,
        s,
        t,
        _zeror,
        _zeroi,
        _errest,
    )
    _zero_py = _zeror + _zeroi*1j
    _errest_py = _errest
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['zero', 'errest'],
    )(
        _zero_py,
        _errest_py,
    )
    return _return_data

def quartic_real(e, a, b, c, d):
    u"""
    All zeros of real quartic equation.

    ``quartic_real`` determines the roots of a quartic equation with
    real coefficients.

    For full information please refer to the NAG Library document for
    c02al

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c02/c02alf.html

    Parameters
    ----------
    e : float
        e, the coefficient of z^4.

    a : float
        a, the coefficient of z^3.

    b : float
        b, the coefficient of z^2.

    c : float
        c, the coefficient of z.

    d : float
        d, the constant coefficient.

    Returns
    -------
    zero : complex, ndarray, shape (4)
        `zero`[i-1] contains the ith root.

    errest : float, ndarray, shape (4)
        `errest`[i-1] contains an approximate error estimate for the ith
        root.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `e` = 0.0.

            Constraint: `e` != 0.0.

        (`errno` 2)
            The companion matrix H cannot be formed without overflow.

        (`errno` 3)
            Failure to converge in ``lapackeig.dhseqr``.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    _zeror = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (4,),
    )
    _zeroi = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (4,),
    )
    _errest = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (4,),
    )
    _b_zeros.quartic_real(
        e,
        a,
        b,
        c,
        d,
        _zeror,
        _zeroi,
        _errest,
    )
    _zero_py = _zeror + _zeroi*1j
    _errest_py = _errest
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['zero', 'errest'],
    )(
        _zero_py,
        _errest_py,
    )
    return _return_data

def cubic_complex(u, r, s, t):
    u"""
    All zeros of complex cubic equation.

    ``cubic_complex`` determines the roots of a cubic equation with
    complex coefficients.

    For full information please refer to the NAG Library document for
    c02am

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c02/c02amf.html

    Parameters
    ----------
    u : complex
        u, the coefficient of z^3.

    r : complex
        r, the coefficient of z^2.

    s : complex
        s, the coefficient of z.

    t : complex
        t, the constant coefficient.

    Returns
    -------
    zero : complex, ndarray, shape (3)
        `zero`[i-1] contains the ith root.

    errest : float, ndarray, shape (3)
        `errest`[i-1] contains an approximate error estimate for the ith
        root.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `u` = (0.0,0.0).

            Constraint: `u` != (0.0,0.0)

        (`errno` 2)
            The companion matrix H cannot be formed without overflow.

        (`errno` 3)
            Failure to converge in ``lapackeig.zhseqr``.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    _zeror = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (3,),
    )
    _zeroi = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (3,),
    )
    _errest = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (3,),
    )
    _b_zeros.cubic_complex(
        u,
        r,
        s,
        t,
        _zeror,
        _zeroi,
        _errest,
    )
    _zero_py = _zeror + _zeroi*1j
    _errest_py = _errest
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['zero', 'errest'],
    )(
        _zero_py,
        _errest_py,
    )
    return _return_data

def quartic_complex(e, a, b, c, d):
    u"""
    All zeros of complex quartic equation.

    ``quartic_complex`` determines the roots of a quartic equation with
    complex coefficients.

    For full information please refer to the NAG Library document for
    c02an

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c02/c02anf.html

    Parameters
    ----------
    e : complex
        e, the coefficient of z^4.

    a : complex
        a, the coefficient of z^3.

    b : complex
        b, the coefficient of z^2.

    c : complex
        c, the coefficient of z.

    d : complex
        d, the constant coefficient.

    Returns
    -------
    zero : complex, ndarray, shape (4)
        `zero`[i-1] contains the ith root.

    errest : float, ndarray, shape (4)
        `errest`[i-1] contains an approximate error estimate for the ith
        root.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `e` = (0.0,0.0).

            Constraint: `e` != (0.0,0.0)

        (`errno` 2)
            The companion matrix H cannot be formed without overflow.

        (`errno` 3)
            Failure to converge in ``lapackeig.zhseqr``.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    _zeror = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (4,),
    )
    _zeroi = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (4,),
    )
    _errest = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (4,),
    )
    _b_zeros.quartic_complex(
        e,
        a,
        b,
        c,
        d,
        _zeror,
        _zeroi,
        _errest,
    )
    _zero_py = _zeror + _zeroi*1j
    _errest_py = _errest
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['zero', 'errest'],
    )(
        _zero_py,
        _errest_py,
    )
    return _return_data
