"""
Deprecated wrappers for the NAG C Library (Mark 26.0) d06 Chapter.

d06 - Mesh Generation.

https://www.nag.com/numeric/CL/nagdoc_cl26/html/D06/d06conts.html

NAG Copyright 2017.
"""

from ctypes import c_double
from ctypes import POINTER
from ctypes import c_char_p
from .util import get_library
from .util import Integer
from .util import Nag_Boolean
from .util import NagError
from .util import NAG_D06BAC_FBND
from .util import Nag_Comm

STRING = c_char_p

_library_cl = get_library()

d06aac_ctypes = _library_cl.d06aac
d06aac_ctypes.restype = None
d06aac_ctypes.argtypes = [Integer, Integer, Integer, POINTER(Integer), POINTER(Integer), POINTER(Integer), POINTER(c_double), POINTER(Integer), POINTER(c_double), Nag_Boolean, c_double, c_double, Integer, STRING, POINTER(NagError)]


def d06aac(nvb, nvmax, nedge, edge, nv, nelt, coor, conn, bspace, smooth, coef, power, itrace, outfile, fail):
    """
    Generates a two-dimensional mesh using a simple incremental method.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/d06/d06aac.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        edge, nv, nelt, coor, conn, bspace
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        coor, bspace
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        edge, nv, nelt, conn
    """
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _edge = nag_integer_type_check_and_cast('d06aac', edge, 'edge')
    _nv = nag_integer_type_check_and_cast('d06aac', nv, 'nv')
    _nelt = nag_integer_type_check_and_cast('d06aac', nelt, 'nelt')
    _coor = nag_double_type_check_and_cast('d06aac', coor, 'coor')
    _conn = nag_integer_type_check_and_cast('d06aac', conn, 'conn')
    _bspace = nag_double_type_check_and_cast('d06aac', bspace, 'bspace')
    d06aac_ctypes(nvb, nvmax, nedge, _edge, _nv, _nelt, _coor, _conn, _bspace, smooth, coef, power, itrace, outfile, fail)

d06abc_ctypes = _library_cl.d06abc
d06abc_ctypes.restype = None
d06abc_ctypes.argtypes = [Integer, Integer, Integer, Integer, POINTER(Integer), POINTER(Integer), POINTER(Integer), POINTER(c_double), POINTER(Integer), POINTER(c_double), Integer, Integer, STRING, POINTER(NagError)]


def d06abc(nvb, nvint, nvmax, nedge, edge, nv, nelt, coor, conn, weight, npropa, itrace, outfile, fail):
    """
    Generates a two-dimensional mesh using a Delaunay--Voronoi process.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/d06/d06abc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        edge, nv, nelt, coor, conn, weight
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        coor, weight
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        edge, nv, nelt, conn
    """
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _edge = nag_integer_type_check_and_cast('d06abc', edge, 'edge')
    _nv = nag_integer_type_check_and_cast('d06abc', nv, 'nv')
    _nelt = nag_integer_type_check_and_cast('d06abc', nelt, 'nelt')
    _coor = nag_double_type_check_and_cast('d06abc', coor, 'coor')
    _conn = nag_integer_type_check_and_cast('d06abc', conn, 'conn')
    _weight = nag_double_type_check_and_cast('d06abc', weight, 'weight')
    d06abc_ctypes(nvb, nvint, nvmax, nedge, _edge, _nv, _nelt, _coor, _conn, _weight, npropa, itrace, outfile, fail)

d06acc_ctypes = _library_cl.d06acc
d06acc_ctypes.restype = None
d06acc_ctypes.argtypes = [Integer, Integer, Integer, Integer, POINTER(Integer), POINTER(Integer), POINTER(Integer), POINTER(c_double), POINTER(Integer), POINTER(c_double), Integer, STRING, POINTER(NagError)]


def d06acc(nvb, nvint, nvmax, nedge, edge, nv, nelt, coor, conn, weight, itrace, outfile, fail):
    """
    Generates a two-dimensional mesh using an Advancing-front method.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/d06/d06acc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        edge, nv, nelt, coor, conn, weight
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        coor, weight
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        edge, nv, nelt, conn
    """
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _edge = nag_integer_type_check_and_cast('d06acc', edge, 'edge')
    _nv = nag_integer_type_check_and_cast('d06acc', nv, 'nv')
    _nelt = nag_integer_type_check_and_cast('d06acc', nelt, 'nelt')
    _coor = nag_double_type_check_and_cast('d06acc', coor, 'coor')
    _conn = nag_integer_type_check_and_cast('d06acc', conn, 'conn')
    _weight = nag_double_type_check_and_cast('d06acc', weight, 'weight')
    d06acc_ctypes(nvb, nvint, nvmax, nedge, _edge, _nv, _nelt, _coor, _conn, _weight, itrace, outfile, fail)

d06bac_ctypes = _library_cl.d06bac
d06bac_ctypes.restype = None
d06bac_ctypes.argtypes = [Integer, POINTER(c_double), POINTER(Integer), NAG_D06BAC_FBND, POINTER(c_double), Integer, POINTER(c_double), Integer, POINTER(Integer), POINTER(Integer), Integer, Integer, POINTER(Integer), POINTER(c_double), POINTER(Integer), POINTER(Integer), Integer, STRING, POINTER(Nag_Comm), POINTER(NagError)]


def d06bac(nlines, coorch, lined, fbnd, crus, sdcrus, rate, ncomp, nlcomp, lcomp, nvmax, nedmx, nvb, coor, nedge, edge, itrace, outfile, comm, fail):
    """
    Generates a boundary mesh.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/d06/d06bac.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        coorch, lined, crus, rate, nlcomp, lcomp, nvb, coor, nedge, edge
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        coorch, crus, rate, coor
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        lined, nlcomp, lcomp, nvb, nedge, edge
    """
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _coorch = nag_double_type_check_and_cast('d06bac', coorch, 'coorch')
    _lined = nag_integer_type_check_and_cast('d06bac', lined, 'lined')
    _crus = nag_double_type_check_and_cast('d06bac', crus, 'crus')
    _rate = nag_double_type_check_and_cast('d06bac', rate, 'rate')
    _nlcomp = nag_integer_type_check_and_cast('d06bac', nlcomp, 'nlcomp')
    _lcomp = nag_integer_type_check_and_cast('d06bac', lcomp, 'lcomp')
    _nvb = nag_integer_type_check_and_cast('d06bac', nvb, 'nvb')
    _coor = nag_double_type_check_and_cast('d06bac', coor, 'coor')
    _nedge = nag_integer_type_check_and_cast('d06bac', nedge, 'nedge')
    _edge = nag_integer_type_check_and_cast('d06bac', edge, 'edge')
    d06bac_ctypes(nlines, _coorch, _lined, fbnd, _crus, sdcrus, _rate, ncomp, _nlcomp, _lcomp, nvmax, nedmx, _nvb, _coor, _nedge, _edge, itrace, outfile, comm, fail)

d06cac_ctypes = _library_cl.d06cac
d06cac_ctypes.restype = None
d06cac_ctypes.argtypes = [Integer, Integer, Integer, POINTER(c_double), POINTER(Integer), POINTER(Integer), Integer, POINTER(Integer), Integer, STRING, Integer, POINTER(NagError)]


def d06cac(nv, nelt, nedge, coor, edge, conn, nvfix, numfix, itrace, outfile, nqint, fail):
    """
    Uses a barycentering technique to smooth a given mesh.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/d06/d06cac.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        coor, edge, conn, numfix
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        coor
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        edge, conn, numfix
    """
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _coor = nag_double_type_check_and_cast('d06cac', coor, 'coor')
    _edge = nag_integer_type_check_and_cast('d06cac', edge, 'edge')
    _conn = nag_integer_type_check_and_cast('d06cac', conn, 'conn')
    _numfix = nag_integer_type_check_and_cast('d06cac', numfix, 'numfix')
    d06cac_ctypes(nv, nelt, nedge, _coor, _edge, _conn, nvfix, _numfix, itrace, outfile, nqint, fail)

d06cbc_ctypes = _library_cl.d06cbc
d06cbc_ctypes.restype = None
d06cbc_ctypes.argtypes = [Integer, Integer, Integer, POINTER(Integer), POINTER(Integer), POINTER(Integer), POINTER(Integer), POINTER(NagError)]


def d06cbc(nv, nelt, nnzmax, conn, nnz, irow, icol, fail):
    """
    Generates a sparsity pattern of a Finite Element matrix associated with
    a given mesh.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/d06/d06cbc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        conn, nnz, irow, icol
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        conn, nnz, irow, icol
    """
    from .util import nag_integer_type_check_and_cast
    _conn = nag_integer_type_check_and_cast('d06cbc', conn, 'conn')
    _nnz = nag_integer_type_check_and_cast('d06cbc', nnz, 'nnz')
    _irow = nag_integer_type_check_and_cast('d06cbc', irow, 'irow')
    _icol = nag_integer_type_check_and_cast('d06cbc', icol, 'icol')
    d06cbc_ctypes(nv, nelt, nnzmax, _conn, _nnz, _irow, _icol, fail)

d06ccc_ctypes = _library_cl.d06ccc
d06ccc_ctypes.restype = None
d06ccc_ctypes.argtypes = [Integer, Integer, Integer, Integer, POINTER(Integer), POINTER(c_double), POINTER(Integer), POINTER(Integer), POINTER(Integer), POINTER(Integer), Integer, STRING, POINTER(NagError)]


def d06ccc(nv, nelt, nedge, nnzmax, nnz, coor, edge, conn, irow, icol, itrace, outfile, fail):
    """
    Renumbers a given mesh using Gibbs method.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/d06/d06ccc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        nnz, coor, edge, conn, irow, icol
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        coor
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        nnz, edge, conn, irow, icol
    """
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _nnz = nag_integer_type_check_and_cast('d06ccc', nnz, 'nnz')
    _coor = nag_double_type_check_and_cast('d06ccc', coor, 'coor')
    _edge = nag_integer_type_check_and_cast('d06ccc', edge, 'edge')
    _conn = nag_integer_type_check_and_cast('d06ccc', conn, 'conn')
    _irow = nag_integer_type_check_and_cast('d06ccc', irow, 'irow')
    _icol = nag_integer_type_check_and_cast('d06ccc', icol, 'icol')
    d06ccc_ctypes(nv, nelt, nedge, nnzmax, _nnz, _coor, _edge, _conn, _irow, _icol, itrace, outfile, fail)

d06dac_ctypes = _library_cl.d06dac
d06dac_ctypes.restype = None
d06dac_ctypes.argtypes = [Integer, Integer, Integer, Integer, Integer, POINTER(Integer), POINTER(c_double), POINTER(c_double), POINTER(Integer), POINTER(Integer), POINTER(c_double), POINTER(Integer), POINTER(Integer), Integer, STRING, POINTER(NagError)]


def d06dac(mode, nv, nedge, nelt, ntrans, itype, trans, coori, edgei, conni, cooro, edgeo, conno, itrace, outfile, fail):
    """
    Generates a mesh resulting from an affine transformation of a given
    mesh.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/d06/d06dac.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        itype, trans, coori, edgei, conni, cooro, edgeo, conno
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        trans, coori, cooro
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        itype, edgei, conni, edgeo, conno
    """
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _itype = nag_integer_type_check_and_cast('d06dac', itype, 'itype')
    _trans = nag_double_type_check_and_cast('d06dac', trans, 'trans')
    _coori = nag_double_type_check_and_cast('d06dac', coori, 'coori')
    _edgei = nag_integer_type_check_and_cast('d06dac', edgei, 'edgei')
    _conni = nag_integer_type_check_and_cast('d06dac', conni, 'conni')
    _cooro = nag_double_type_check_and_cast('d06dac', cooro, 'cooro')
    _edgeo = nag_integer_type_check_and_cast('d06dac', edgeo, 'edgeo')
    _conno = nag_integer_type_check_and_cast('d06dac', conno, 'conno')
    d06dac_ctypes(mode, nv, nedge, nelt, ntrans, _itype, _trans, _coori, _edgei, _conni, _cooro, _edgeo, _conno, itrace, outfile, fail)

d06dbc_ctypes = _library_cl.d06dbc
d06dbc_ctypes.restype = None
d06dbc_ctypes.argtypes = [c_double, Integer, Integer, Integer, POINTER(c_double), POINTER(Integer), POINTER(Integer), POINTER(Integer), Integer, Integer, Integer, POINTER(c_double), POINTER(Integer), POINTER(Integer), POINTER(Integer), POINTER(Integer), POINTER(Integer), POINTER(Integer), POINTER(c_double), POINTER(Integer), POINTER(Integer), POINTER(Integer), Integer, STRING, POINTER(NagError)]


def d06dbc(eps, nv1, nelt1, nedge1, coor1, edge1, conn1, reft1, nv2, nelt2, nedge2, coor2, edge2, conn2, reft2, nv3, nelt3, nedge3, coor3, edge3, conn3, reft3, itrace, outfile, fail):
    """
    Joins together two given adjacent (possibly overlapping) meshes.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/d06/d06dbc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        coor1, edge1, conn1, reft1, coor2, edge2, conn2, reft2, nv3, nelt3, nedge3, coor3, edge3, conn3, reft3
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        coor1, coor2, coor3
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        edge1, conn1, reft1, edge2, conn2, reft2, nv3, nelt3, nedge3, edge3, conn3, reft3
    """
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _coor1 = nag_double_type_check_and_cast('d06dbc', coor1, 'coor1')
    _edge1 = nag_integer_type_check_and_cast('d06dbc', edge1, 'edge1')
    _conn1 = nag_integer_type_check_and_cast('d06dbc', conn1, 'conn1')
    _reft1 = nag_integer_type_check_and_cast('d06dbc', reft1, 'reft1')
    _coor2 = nag_double_type_check_and_cast('d06dbc', coor2, 'coor2')
    _edge2 = nag_integer_type_check_and_cast('d06dbc', edge2, 'edge2')
    _conn2 = nag_integer_type_check_and_cast('d06dbc', conn2, 'conn2')
    _reft2 = nag_integer_type_check_and_cast('d06dbc', reft2, 'reft2')
    _nv3 = nag_integer_type_check_and_cast('d06dbc', nv3, 'nv3')
    _nelt3 = nag_integer_type_check_and_cast('d06dbc', nelt3, 'nelt3')
    _nedge3 = nag_integer_type_check_and_cast('d06dbc', nedge3, 'nedge3')
    _coor3 = nag_double_type_check_and_cast('d06dbc', coor3, 'coor3')
    _edge3 = nag_integer_type_check_and_cast('d06dbc', edge3, 'edge3')
    _conn3 = nag_integer_type_check_and_cast('d06dbc', conn3, 'conn3')
    _reft3 = nag_integer_type_check_and_cast('d06dbc', reft3, 'reft3')
    d06dbc_ctypes(eps, nv1, nelt1, nedge1, _coor1, _edge1, _conn1, _reft1, nv2, nelt2, nedge2, _coor2, _edge2, _conn2, _reft2, _nv3, _nelt3, _nedge3, _coor3, _edge3, _conn3, _reft3, itrace, outfile, fail)

nag_mesh2d_inc = d06aac
nag_mesh2d_delaunay = d06abc
nag_mesh2d_front = d06acc
nag_mesh2d_bound = d06bac
nag_mesh2d_smooth = d06cac
nag_mesh2d_sparse = d06cbc
nag_mesh2d_renum = d06ccc
nag_mesh2d_trans = d06dac
nag_mesh2d_join = d06dbc
__all__ = ['d06aac', 'd06abc', 'd06acc', 'd06bac', 'd06cac', 'd06cbc',
           'd06ccc', 'd06dac', 'd06dbc', 'nag_mesh2d_bound',
           'nag_mesh2d_delaunay', 'nag_mesh2d_front',
           'nag_mesh2d_inc', 'nag_mesh2d_join', 'nag_mesh2d_renum',
           'nag_mesh2d_smooth', 'nag_mesh2d_sparse',
           'nag_mesh2d_trans']
