"""
Deprecated wrappers for the NAG C Library (Mark 26.0) f03 Chapter.

f03 - Determinants.

https://www.nag.com/numeric/CL/nagdoc_cl26/html/F03/f03conts.html

NAG Copyright 2017.
"""
from ctypes import c_double
from ctypes import POINTER

from .util import get_library

from .util import Integer
from .util import NagError
from .util import Complex
from .util import Nag_OrderType
from .util import Nag_UploType

_library_cl = get_library()


f03bac_ctypes = _library_cl.f03bac
f03bac_ctypes.restype = None
f03bac_ctypes.argtypes = [Nag_OrderType, Integer, POINTER(c_double), Integer, POINTER(Integer), POINTER(c_double), POINTER(Integer), POINTER(NagError)]


def f03bac(order, n, a, pda, ipiv, d, _id, fail):
    """
    Determinant of real matrix, matrix previously LU factorized.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f03/f03bac.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        a, ipiv, d, _id
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        a, d
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv, _id
    """
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _a = nag_double_type_check_and_cast('f03bac', a, 'a')
    _ipiv = nag_integer_type_check_and_cast('f03bac', ipiv, 'ipiv')
    _d = nag_double_type_check_and_cast('f03bac', d, 'd')
    __id = nag_integer_type_check_and_cast('f03bac', _id, '_id')
    f03bac_ctypes(order, n, _a, pda, _ipiv, _d, __id, fail)

f03bfc_ctypes = _library_cl.f03bfc
f03bfc_ctypes.restype = None
f03bfc_ctypes.argtypes = [Nag_OrderType, Integer, POINTER(c_double), Integer, POINTER(c_double), POINTER(Integer), POINTER(NagError)]


def f03bfc(order, n, a, pda, d, _id, fail):
    """
    Determinant of real symmetric positive definite matrix previously LL^T
    factorized.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f03/f03bfc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        a, d, _id
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        a, d
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        _id
    """
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _a = nag_double_type_check_and_cast('f03bfc', a, 'a')
    _d = nag_double_type_check_and_cast('f03bfc', d, 'd')
    __id = nag_integer_type_check_and_cast('f03bfc', _id, '_id')
    f03bfc_ctypes(order, n, _a, pda, _d, __id, fail)

f03bhc_ctypes = _library_cl.f03bhc
f03bhc_ctypes.restype = None
f03bhc_ctypes.argtypes = [Nag_OrderType, Nag_UploType, Integer, Integer, POINTER(c_double), Integer, POINTER(c_double), POINTER(Integer), POINTER(NagError)]


def f03bhc(order, uplo, n, kd, ab, pdab, d, _id, fail):
    """
    Determinant of real symmetric positive definite banded matrix previously
    factorized by f07hdc.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f03/f03bhc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ab, d, _id
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        ab, d
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        _id
    """
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _ab = nag_double_type_check_and_cast('f03bhc', ab, 'ab')
    _d = nag_double_type_check_and_cast('f03bhc', d, 'd')
    __id = nag_integer_type_check_and_cast('f03bhc', _id, '_id')
    f03bhc_ctypes(order, uplo, n, kd, _ab, pdab, _d, __id, fail)

f03bnc_ctypes = _library_cl.f03bnc
f03bnc_ctypes.restype = None
f03bnc_ctypes.argtypes = [Nag_OrderType, Integer, POINTER(Complex), Integer, POINTER(Integer), POINTER(Complex), POINTER(Integer), POINTER(NagError)]


def f03bnc(order, n, a, pda, ipiv, d, _id, fail):
    """
    Determinant of complex matrix previously LU factorized.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f03/f03bnc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        a, ipiv, d, _id
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ipiv, _id
    TypeError will be raised when the following arguments are not of data type
    Complex:
        a, d
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _a = nag_complex_type_check_and_cast('f03bnc', a, 'a')
    _ipiv = nag_integer_type_check_and_cast('f03bnc', ipiv, 'ipiv')
    _d = nag_complex_type_check_and_cast('f03bnc', d, 'd')
    __id = nag_integer_type_check_and_cast('f03bnc', _id, '_id')
    f03bnc_ctypes(order, n, _a, pda, _ipiv, _d, __id, fail)

nag_det_real_gen = f03bac
nag_det_real_sym = f03bfc
nag_det_real_band_sym = f03bhc
nag_det_complex_gen = f03bnc
__all__ = ['f03bac', 'f03bfc', 'f03bhc',
           'f03bnc', 'nag_det_complex_gen',
           'nag_det_real_band_sym', 'nag_det_real_gen',
           'nag_det_real_sym']
