"""
Deprecated wrappers for the NAG C Library (Mark 26.0) f12 Chapter.

f12 - Large Scale Eigenproblems.

https://www.nag.com/numeric/CL/nagdoc_cl26/html/F12/f12conts.html

NAG Copyright 2017.
"""
from ctypes import c_double
from ctypes import POINTER

from ctypes import c_char_p
from .util import get_library

from .util import Integer
from .util import NagError
from .util import Complex
from .util import Nag_Complex_Type

_library_cl = get_library()

STRING = c_char_p


f12aac_ctypes = _library_cl.f12aac
f12aac_ctypes.restype = None
f12aac_ctypes.argtypes = [Integer, Integer, Integer, POINTER(Integer), Integer, POINTER(c_double), Integer, POINTER(NagError)]


def f12aac(n, nev, ncv, icomm, licomm, comm, lcomm, fail):
    """
    Initialization routine for (f12abc) computing selected eigenvalues and,
    optionally, eigenvectors of a real nonsymmetric sparse (standard or
    generalized) eigenproblem.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f12/f12aac.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        icomm, comm
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        comm
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        icomm
    """
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _icomm = nag_integer_type_check_and_cast('f12aac', icomm, 'icomm')
    _comm = nag_double_type_check_and_cast('f12aac', comm, 'comm')
    f12aac_ctypes(n, nev, ncv, _icomm, licomm, _comm, lcomm, fail)

f12abc_ctypes = _library_cl.f12abc
f12abc_ctypes.restype = None
f12abc_ctypes.argtypes = [POINTER(Integer), POINTER(c_double), POINTER(c_double), POINTER(POINTER(c_double)), POINTER(POINTER(c_double)), POINTER(POINTER(c_double)), POINTER(Integer), POINTER(c_double), POINTER(Integer), POINTER(NagError)]


def f12abc(irevcm, resid, v, x, y, mx, nshift, comm, icomm, fail):
    """
    Selected eigenvalues and, optionally, eigenvectors of a real
    nonsymmetric sparse eigenproblem, reverse communication.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f12/f12abc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        irevcm, resid, v, x, y, mx, nshift, comm, icomm
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        resid, v, x, y, mx, comm
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        irevcm, nshift, icomm
    """
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _irevcm = nag_integer_type_check_and_cast('f12abc', irevcm, 'irevcm')
    _resid = nag_double_type_check_and_cast('f12abc', resid, 'resid')
    _v = nag_double_type_check_and_cast('f12abc', v, 'v')
    _x = nag_double_type_check_and_cast('f12abc', x, 'x')
    _y = nag_double_type_check_and_cast('f12abc', y, 'y')
    _mx = nag_double_type_check_and_cast('f12abc', mx, 'mx')
    _nshift = nag_integer_type_check_and_cast('f12abc', nshift, 'nshift')
    _comm = nag_double_type_check_and_cast('f12abc', comm, 'comm')
    _icomm = nag_integer_type_check_and_cast('f12abc', icomm, 'icomm')
    f12abc_ctypes(_irevcm, _resid, _v, _x, _y, _mx, _nshift, _comm, _icomm, fail)

f12acc_ctypes = _library_cl.f12acc
f12acc_ctypes.restype = None
f12acc_ctypes.argtypes = [POINTER(Integer), POINTER(c_double), POINTER(c_double), POINTER(c_double), c_double, c_double, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Integer), POINTER(NagError)]


def f12acc(nconv, dr, di, z, sigmar, sigmai, resid, v, comm, icomm, fail):
    """
    Selected eigenvalues and, optionally, eigenvectors of a real
    nonsymmetric sparse eigenproblem, postprocessing for f12abc.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f12/f12acc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        nconv, dr, di, z, resid, v, comm, icomm
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        dr, di, z, resid, v, comm
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        nconv, icomm
    """
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _nconv = nag_integer_type_check_and_cast('f12acc', nconv, 'nconv')
    _dr = nag_double_type_check_and_cast('f12acc', dr, 'dr')
    _di = nag_double_type_check_and_cast('f12acc', di, 'di')
    _z = nag_double_type_check_and_cast('f12acc', z, 'z')
    _resid = nag_double_type_check_and_cast('f12acc', resid, 'resid')
    _v = nag_double_type_check_and_cast('f12acc', v, 'v')
    _comm = nag_double_type_check_and_cast('f12acc', comm, 'comm')
    _icomm = nag_integer_type_check_and_cast('f12acc', icomm, 'icomm')
    f12acc_ctypes(_nconv, _dr, _di, _z, sigmar, sigmai, _resid, _v, _comm, _icomm, fail)

f12adc_ctypes = _library_cl.f12adc
f12adc_ctypes.restype = None
f12adc_ctypes.argtypes = [STRING, POINTER(Integer), POINTER(c_double), POINTER(NagError)]


def f12adc(_str, icomm, comm, fail):
    """
    Set a single option from a string (f12abc/f12acc/f12agc).
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f12/f12adc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        icomm, comm
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        comm
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        icomm
    """
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _icomm = nag_integer_type_check_and_cast('f12adc', icomm, 'icomm')
    _comm = nag_double_type_check_and_cast('f12adc', comm, 'comm')
    f12adc_ctypes(_str, _icomm, _comm, fail)

f12aec_ctypes = _library_cl.f12aec
f12aec_ctypes.restype = None
f12aec_ctypes.argtypes = [POINTER(Integer), POINTER(Integer), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Integer), POINTER(c_double)]


def f12aec(niter, nconv, ritzr, ritzi, rzest, icomm, comm):
    """
    Provides monitoring information for f12abc.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f12/f12aec.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        niter, nconv, ritzr, ritzi, rzest, icomm, comm
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        ritzr, ritzi, rzest, comm
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        niter, nconv, icomm
    """
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _niter = nag_integer_type_check_and_cast('f12aec', niter, 'niter')
    _nconv = nag_integer_type_check_and_cast('f12aec', nconv, 'nconv')
    _ritzr = nag_double_type_check_and_cast('f12aec', ritzr, 'ritzr')
    _ritzi = nag_double_type_check_and_cast('f12aec', ritzi, 'ritzi')
    _rzest = nag_double_type_check_and_cast('f12aec', rzest, 'rzest')
    _icomm = nag_integer_type_check_and_cast('f12aec', icomm, 'icomm')
    _comm = nag_double_type_check_and_cast('f12aec', comm, 'comm')
    f12aec_ctypes(_niter, _nconv, _ritzr, _ritzi, _rzest, _icomm, _comm)

f12afc_ctypes = _library_cl.f12afc
f12afc_ctypes.restype = None
f12afc_ctypes.argtypes = [Integer, Integer, Integer, POINTER(Integer), Integer, POINTER(c_double), Integer, POINTER(NagError)]


def f12afc(n, nev, ncv, icomm, licomm, comm, lcomm, fail):
    """
    Initialization routine for (f12agc) computing selected eigenvalues and,
    optionally, eigenvectors of a real nonsymmetric banded (standard or
    generalized) eigenproblem.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f12/f12afc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        icomm, comm
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        comm
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        icomm
    """
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _icomm = nag_integer_type_check_and_cast('f12afc', icomm, 'icomm')
    _comm = nag_double_type_check_and_cast('f12afc', comm, 'comm')
    f12afc_ctypes(n, nev, ncv, _icomm, licomm, _comm, lcomm, fail)

f12agc_ctypes = _library_cl.f12agc
f12agc_ctypes.restype = None
f12agc_ctypes.argtypes = [Integer, Integer, POINTER(c_double), POINTER(c_double), c_double, c_double, POINTER(Integer), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Integer), POINTER(NagError)]


def f12agc(kl, ku, ab, mb, sigmar, sigmai, nconv, dr, di, z, resid, v, comm, icomm, fail):
    """
    Selected eigenvalues and, optionally, eigenvectors of a real
    nonsymmetric banded eigenproblem, driver.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f12/f12agc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ab, mb, nconv, dr, di, z, resid, v, comm, icomm
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        ab, mb, dr, di, z, resid, v, comm
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        nconv, icomm
    """
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _ab = nag_double_type_check_and_cast('f12agc', ab, 'ab')
    _mb = nag_double_type_check_and_cast('f12agc', mb, 'mb')
    _nconv = nag_integer_type_check_and_cast('f12agc', nconv, 'nconv')
    _dr = nag_double_type_check_and_cast('f12agc', dr, 'dr')
    _di = nag_double_type_check_and_cast('f12agc', di, 'di')
    _z = nag_double_type_check_and_cast('f12agc', z, 'z')
    _resid = nag_double_type_check_and_cast('f12agc', resid, 'resid')
    _v = nag_double_type_check_and_cast('f12agc', v, 'v')
    _comm = nag_double_type_check_and_cast('f12agc', comm, 'comm')
    _icomm = nag_integer_type_check_and_cast('f12agc', icomm, 'icomm')
    f12agc_ctypes(kl, ku, _ab, _mb, sigmar, sigmai, _nconv, _dr, _di, _z, _resid, _v, _comm, _icomm, fail)

f12anc_ctypes = _library_cl.f12anc
f12anc_ctypes.restype = None
f12anc_ctypes.argtypes = [Integer, Integer, Integer, POINTER(Integer), Integer, POINTER(Complex), Integer, POINTER(NagError)]


def f12anc(n, nev, ncv, icomm, licomm, comm, lcomm, fail):
    """
    Initialization routine for (f12apc) computing selected eigenvalues and,
    optionally, eigenvectors of a complex sparse (standard or generalized)
    eigenproblem.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f12/f12anc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        icomm, comm
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        icomm
    TypeError will be raised when the following arguments are not of data type
    Complex:
        comm
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _icomm = nag_integer_type_check_and_cast('f12anc', icomm, 'icomm')
    _comm = nag_complex_type_check_and_cast('f12anc', comm, 'comm')
    f12anc_ctypes(n, nev, ncv, _icomm, licomm, _comm, lcomm, fail)

f12apc_ctypes = _library_cl.f12apc
f12apc_ctypes.restype = None
f12apc_ctypes.argtypes = [POINTER(Integer), POINTER(Complex), POINTER(Complex), POINTER(POINTER(Complex)), POINTER(POINTER(Complex)), POINTER(POINTER(Complex)), POINTER(Integer), POINTER(Complex), POINTER(Integer), POINTER(NagError)]


def f12apc(irevcm, resid, v, x, y, mx, nshift, comm, icomm, fail):
    """
    Selected eigenvalues and, optionally, eigenvectors of a complex sparse
    eigenproblem, reverse communication.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f12/f12apc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        irevcm, resid, v, x, y, mx, nshift, comm, icomm
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        irevcm, nshift, icomm
    TypeError will be raised when the following arguments are not of data type
    Complex:
        resid, v, x, y, mx, comm
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _irevcm = nag_integer_type_check_and_cast('f12apc', irevcm, 'irevcm')
    _resid = nag_complex_type_check_and_cast('f12apc', resid, 'resid')
    _v = nag_complex_type_check_and_cast('f12apc', v, 'v')
    _x = nag_complex_type_check_and_cast('f12apc', x, 'x')
    _y = nag_complex_type_check_and_cast('f12apc', y, 'y')
    _mx = nag_complex_type_check_and_cast('f12apc', mx, 'mx')
    _nshift = nag_integer_type_check_and_cast('f12apc', nshift, 'nshift')
    _comm = nag_complex_type_check_and_cast('f12apc', comm, 'comm')
    _icomm = nag_integer_type_check_and_cast('f12apc', icomm, 'icomm')
    f12apc_ctypes(_irevcm, _resid, _v, _x, _y, _mx, _nshift, _comm, _icomm, fail)

f12aqc_ctypes = _library_cl.f12aqc
f12aqc_ctypes.restype = None
f12aqc_ctypes.argtypes = [POINTER(Integer), POINTER(Complex), POINTER(Complex), Nag_Complex_Type, POINTER(Complex), POINTER(Complex), POINTER(Complex), POINTER(Integer), POINTER(NagError)]


def f12aqc(nconv, d, z, sigma, resid, v, comm, icomm, fail):
    """
    Selected eigenvalues and, optionally, eigenvectors of a complex sparse
    eigenproblem, postprocessing for f12apc.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f12/f12aqc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        nconv, d, z, resid, v, comm, icomm
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        nconv, icomm
    TypeError will be raised when the following arguments are not of data type
    Complex:
        d, z, resid, v, comm
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _nconv = nag_integer_type_check_and_cast('f12aqc', nconv, 'nconv')
    _d = nag_complex_type_check_and_cast('f12aqc', d, 'd')
    _z = nag_complex_type_check_and_cast('f12aqc', z, 'z')
    _resid = nag_complex_type_check_and_cast('f12aqc', resid, 'resid')
    _v = nag_complex_type_check_and_cast('f12aqc', v, 'v')
    _comm = nag_complex_type_check_and_cast('f12aqc', comm, 'comm')
    _icomm = nag_integer_type_check_and_cast('f12aqc', icomm, 'icomm')
    f12aqc_ctypes(_nconv, _d, _z, sigma, _resid, _v, _comm, _icomm, fail)

f12arc_ctypes = _library_cl.f12arc
f12arc_ctypes.restype = None
f12arc_ctypes.argtypes = [STRING, POINTER(Integer), POINTER(Complex), POINTER(NagError)]


def f12arc(_str, icomm, comm, fail):
    """
    Set a single option from a string (f12apc/f12aqc).
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f12/f12arc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        icomm, comm
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        icomm
    TypeError will be raised when the following arguments are not of data type
    Complex:
        comm
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _icomm = nag_integer_type_check_and_cast('f12arc', icomm, 'icomm')
    _comm = nag_complex_type_check_and_cast('f12arc', comm, 'comm')
    f12arc_ctypes(_str, _icomm, _comm, fail)

f12asc_ctypes = _library_cl.f12asc
f12asc_ctypes.restype = None
f12asc_ctypes.argtypes = [POINTER(Integer), POINTER(Integer), POINTER(Complex), POINTER(Complex), POINTER(Integer), POINTER(Complex)]


def f12asc(niter, nconv, ritz, rzest, icomm, comm):
    """
    Provides monitoring information for f12apc.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f12/f12asc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        niter, nconv, ritz, rzest, icomm, comm
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        niter, nconv, icomm
    TypeError will be raised when the following arguments are not of data type
    Complex:
        ritz, rzest, comm
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _niter = nag_integer_type_check_and_cast('f12asc', niter, 'niter')
    _nconv = nag_integer_type_check_and_cast('f12asc', nconv, 'nconv')
    _ritz = nag_complex_type_check_and_cast('f12asc', ritz, 'ritz')
    _rzest = nag_complex_type_check_and_cast('f12asc', rzest, 'rzest')
    _icomm = nag_integer_type_check_and_cast('f12asc', icomm, 'icomm')
    _comm = nag_complex_type_check_and_cast('f12asc', comm, 'comm')
    f12asc_ctypes(_niter, _nconv, _ritz, _rzest, _icomm, _comm)

f12atc_ctypes = _library_cl.f12atc
f12atc_ctypes.restype = None
f12atc_ctypes.argtypes = [Integer, Integer, Integer, POINTER(Integer), Integer, POINTER(Complex), Integer, POINTER(NagError)]


def f12atc(n, nev, ncv, icomm, licomm, comm, lcomm, fail):
    """
    Initialization function for f12auc computing selected eigenvalues and,
    optionally, eigenvectors of a complex banded (standard or generalized)
    eigenproblem.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f12/f12atc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        icomm, comm
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        icomm
    TypeError will be raised when the following arguments are not of data type
    Complex:
        comm
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _icomm = nag_integer_type_check_and_cast('f12atc', icomm, 'icomm')
    _comm = nag_complex_type_check_and_cast('f12atc', comm, 'comm')
    f12atc_ctypes(n, nev, ncv, _icomm, licomm, _comm, lcomm, fail)

f12auc_ctypes = _library_cl.f12auc
f12auc_ctypes.restype = None
f12auc_ctypes.argtypes = [Integer, Integer, POINTER(Complex), POINTER(Complex), Nag_Complex_Type, POINTER(Integer), POINTER(Complex), POINTER(Complex), POINTER(Complex), POINTER(Complex), POINTER(Complex), POINTER(Integer), POINTER(NagError)]


def f12auc(kl, ku, ab, mb, sigma, nconv, d, z, resid, v, comm, icomm, fail):
    """
    Selected eigenvalues and, optionally, eigenvectors of complex non-
    Hermitian banded eigenproblem, driver.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f12/f12auc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ab, mb, nconv, d, z, resid, v, comm, icomm
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        nconv, icomm
    TypeError will be raised when the following arguments are not of data type
    Complex:
        ab, mb, d, z, resid, v, comm
    """
    from .util import nag_complex_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _ab = nag_complex_type_check_and_cast('f12auc', ab, 'ab')
    _mb = nag_complex_type_check_and_cast('f12auc', mb, 'mb')
    _nconv = nag_integer_type_check_and_cast('f12auc', nconv, 'nconv')
    _d = nag_complex_type_check_and_cast('f12auc', d, 'd')
    _z = nag_complex_type_check_and_cast('f12auc', z, 'z')
    _resid = nag_complex_type_check_and_cast('f12auc', resid, 'resid')
    _v = nag_complex_type_check_and_cast('f12auc', v, 'v')
    _comm = nag_complex_type_check_and_cast('f12auc', comm, 'comm')
    _icomm = nag_integer_type_check_and_cast('f12auc', icomm, 'icomm')
    f12auc_ctypes(kl, ku, _ab, _mb, sigma, _nconv, _d, _z, _resid, _v, _comm, _icomm, fail)

f12fac_ctypes = _library_cl.f12fac
f12fac_ctypes.restype = None
f12fac_ctypes.argtypes = [Integer, Integer, Integer, POINTER(Integer), Integer, POINTER(c_double), Integer, POINTER(NagError)]


def f12fac(n, nev, ncv, icomm, licomm, comm, lcomm, fail):
    """
    Initialization routine for (f12fbc) computing selected eigenvalues and,
    optionally, eigenvectors of a real symmetric sparse (standard or
    generalized) eigenproblem.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f12/f12fac.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        icomm, comm
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        comm
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        icomm
    """
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _icomm = nag_integer_type_check_and_cast('f12fac', icomm, 'icomm')
    _comm = nag_double_type_check_and_cast('f12fac', comm, 'comm')
    f12fac_ctypes(n, nev, ncv, _icomm, licomm, _comm, lcomm, fail)

f12fbc_ctypes = _library_cl.f12fbc
f12fbc_ctypes.restype = None
f12fbc_ctypes.argtypes = [POINTER(Integer), POINTER(c_double), POINTER(c_double), POINTER(POINTER(c_double)), POINTER(POINTER(c_double)), POINTER(POINTER(c_double)), POINTER(Integer), POINTER(c_double), POINTER(Integer), POINTER(NagError)]


def f12fbc(irevcm, resid, v, x, y, mx, nshift, comm, icomm, fail):
    """
    Selected eigenvalues and, optionally, eigenvectors of a real symmetric
    sparse eigenproblem, reverse communication.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f12/f12fbc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        irevcm, resid, v, x, y, mx, nshift, comm, icomm
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        resid, v, x, y, mx, comm
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        irevcm, nshift, icomm
    """
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _irevcm = nag_integer_type_check_and_cast('f12fbc', irevcm, 'irevcm')
    _resid = nag_double_type_check_and_cast('f12fbc', resid, 'resid')
    _v = nag_double_type_check_and_cast('f12fbc', v, 'v')
    _x = nag_double_type_check_and_cast('f12fbc', x, 'x')
    _y = nag_double_type_check_and_cast('f12fbc', y, 'y')
    _mx = nag_double_type_check_and_cast('f12fbc', mx, 'mx')
    _nshift = nag_integer_type_check_and_cast('f12fbc', nshift, 'nshift')
    _comm = nag_double_type_check_and_cast('f12fbc', comm, 'comm')
    _icomm = nag_integer_type_check_and_cast('f12fbc', icomm, 'icomm')
    f12fbc_ctypes(_irevcm, _resid, _v, _x, _y, _mx, _nshift, _comm, _icomm, fail)

f12fcc_ctypes = _library_cl.f12fcc
f12fcc_ctypes.restype = None
f12fcc_ctypes.argtypes = [POINTER(Integer), POINTER(c_double), POINTER(c_double), c_double, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Integer), POINTER(NagError)]


def f12fcc(nconv, d, z, sigma, resid, v, comm, icomm, fail):
    """
    Selected eigenvalues and, optionally, eigenvectors of a real symmetric
    sparse eigenproblem, postprocessing for f12fbc.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f12/f12fcc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        nconv, d, z, resid, v, comm, icomm
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        d, z, resid, v, comm
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        nconv, icomm
    """
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _nconv = nag_integer_type_check_and_cast('f12fcc', nconv, 'nconv')
    _d = nag_double_type_check_and_cast('f12fcc', d, 'd')
    _z = nag_double_type_check_and_cast('f12fcc', z, 'z')
    _resid = nag_double_type_check_and_cast('f12fcc', resid, 'resid')
    _v = nag_double_type_check_and_cast('f12fcc', v, 'v')
    _comm = nag_double_type_check_and_cast('f12fcc', comm, 'comm')
    _icomm = nag_integer_type_check_and_cast('f12fcc', icomm, 'icomm')
    f12fcc_ctypes(_nconv, _d, _z, sigma, _resid, _v, _comm, _icomm, fail)

f12fdc_ctypes = _library_cl.f12fdc
f12fdc_ctypes.restype = None
f12fdc_ctypes.argtypes = [STRING, POINTER(Integer), POINTER(c_double), POINTER(NagError)]


def f12fdc(_str, icomm, comm, fail):
    """
    Set a single option from a string (f12fbc/f12fcc/f12fgc).
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f12/f12fdc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        icomm, comm
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        comm
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        icomm
    """
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _icomm = nag_integer_type_check_and_cast('f12fdc', icomm, 'icomm')
    _comm = nag_double_type_check_and_cast('f12fdc', comm, 'comm')
    f12fdc_ctypes(_str, _icomm, _comm, fail)

f12fec_ctypes = _library_cl.f12fec
f12fec_ctypes.restype = None
f12fec_ctypes.argtypes = [POINTER(Integer), POINTER(Integer), POINTER(c_double), POINTER(c_double), POINTER(Integer), POINTER(c_double)]


def f12fec(niter, nconv, ritz, rzest, icomm, comm):
    """
    Provides monitoring information for f12fbc.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f12/f12fec.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        niter, nconv, ritz, rzest, icomm, comm
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        ritz, rzest, comm
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        niter, nconv, icomm
    """
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _niter = nag_integer_type_check_and_cast('f12fec', niter, 'niter')
    _nconv = nag_integer_type_check_and_cast('f12fec', nconv, 'nconv')
    _ritz = nag_double_type_check_and_cast('f12fec', ritz, 'ritz')
    _rzest = nag_double_type_check_and_cast('f12fec', rzest, 'rzest')
    _icomm = nag_integer_type_check_and_cast('f12fec', icomm, 'icomm')
    _comm = nag_double_type_check_and_cast('f12fec', comm, 'comm')
    f12fec_ctypes(_niter, _nconv, _ritz, _rzest, _icomm, _comm)

f12ffc_ctypes = _library_cl.f12ffc
f12ffc_ctypes.restype = None
f12ffc_ctypes.argtypes = [Integer, Integer, Integer, POINTER(Integer), Integer, POINTER(c_double), Integer, POINTER(NagError)]


def f12ffc(n, nev, ncv, icomm, licomm, comm, lcomm, fail):
    """
    Initialization routine for (f12fgc) computing selected eigenvalues and,
    optionally, eigenvectors of a real symmetric banded (standard or
    generalized) eigenproblem.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f12/f12ffc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        icomm, comm
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        comm
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        icomm
    """
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _icomm = nag_integer_type_check_and_cast('f12ffc', icomm, 'icomm')
    _comm = nag_double_type_check_and_cast('f12ffc', comm, 'comm')
    f12ffc_ctypes(n, nev, ncv, _icomm, licomm, _comm, lcomm, fail)

f12fgc_ctypes = _library_cl.f12fgc
f12fgc_ctypes.restype = None
f12fgc_ctypes.argtypes = [Integer, Integer, POINTER(c_double), POINTER(c_double), c_double, POINTER(Integer), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Integer), POINTER(NagError)]


def f12fgc(kl, ku, ab, mb, sigma, nconv, d, z, resid, v, comm, icomm, fail):
    """
    Selected eigenvalues and, optionally, eigenvectors of a real symmetric
    banded eigenproblem, driver.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/f12/f12fgc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ab, mb, nconv, d, z, resid, v, comm, icomm
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        ab, mb, d, z, resid, v, comm
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        nconv, icomm
    """
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _ab = nag_double_type_check_and_cast('f12fgc', ab, 'ab')
    _mb = nag_double_type_check_and_cast('f12fgc', mb, 'mb')
    _nconv = nag_integer_type_check_and_cast('f12fgc', nconv, 'nconv')
    _d = nag_double_type_check_and_cast('f12fgc', d, 'd')
    _z = nag_double_type_check_and_cast('f12fgc', z, 'z')
    _resid = nag_double_type_check_and_cast('f12fgc', resid, 'resid')
    _v = nag_double_type_check_and_cast('f12fgc', v, 'v')
    _comm = nag_double_type_check_and_cast('f12fgc', comm, 'comm')
    _icomm = nag_integer_type_check_and_cast('f12fgc', icomm, 'icomm')
    f12fgc_ctypes(kl, ku, _ab, _mb, sigma, _nconv, _d, _z, _resid, _v, _comm, _icomm, fail)

nag_real_sparse_eigensystem_init = f12aac
nag_real_sparse_eigensystem_iter = f12abc
nag_real_sparse_eigensystem_sol = f12acc
nag_real_sparse_eigensystem_option = f12adc
nag_real_sparse_eigensystem_monit = f12aec
nag_real_banded_sparse_eigensystem_init = f12afc
nag_real_banded_sparse_eigensystem_sol = f12agc
nag_complex_sparse_eigensystem_init = f12anc
nag_complex_sparse_eigensystem_iter = f12apc
nag_complex_sparse_eigensystem_sol = f12aqc
nag_complex_sparse_eigensystem_option = f12arc
nag_complex_sparse_eigensystem_monit = f12asc
nag_complex_banded_eigensystem_init = f12atc
nag_complex_banded_eigensystem_solve = f12auc
nag_real_symm_sparse_eigensystem_init = f12fac
nag_real_symm_sparse_eigensystem_iter = f12fbc
nag_real_symm_sparse_eigensystem_sol = f12fcc
nag_real_symm_sparse_eigensystem_option = f12fdc
nag_real_symm_sparse_eigensystem_monit = f12fec
nag_real_symm_banded_sparse_eigensystem_init = f12ffc
nag_real_symm_banded_sparse_eigensystem_sol = f12fgc
__all__ = ['f12aac', 'f12abc', 'f12acc', 'f12adc', 'f12aec', 'f12afc',
           'f12agc', 'f12anc', 'f12apc', 'f12aqc', 'f12arc', 'f12asc',
           'f12atc', 'f12auc', 'f12fac', 'f12fbc', 'f12fcc', 'f12fdc',
           'f12fec', 'f12ffc', 'f12fgc',
           'nag_complex_banded_eigensystem_init',
           'nag_complex_banded_eigensystem_solve',
           'nag_complex_sparse_eigensystem_init',
           'nag_complex_sparse_eigensystem_iter',
           'nag_complex_sparse_eigensystem_monit',
           'nag_complex_sparse_eigensystem_option',
           'nag_complex_sparse_eigensystem_sol',
           'nag_real_banded_sparse_eigensystem_init',
           'nag_real_banded_sparse_eigensystem_sol',
           'nag_real_sparse_eigensystem_init',
           'nag_real_sparse_eigensystem_iter',
           'nag_real_sparse_eigensystem_monit',
           'nag_real_sparse_eigensystem_option',
           'nag_real_sparse_eigensystem_sol',
           'nag_real_symm_banded_sparse_eigensystem_init',
           'nag_real_symm_banded_sparse_eigensystem_sol',
           'nag_real_symm_sparse_eigensystem_init',
           'nag_real_symm_sparse_eigensystem_iter',
           'nag_real_symm_sparse_eigensystem_monit',
           'nag_real_symm_sparse_eigensystem_option',
           'nag_real_symm_sparse_eigensystem_sol']
