"""
Deprecated wrappers for the NAG C Library (Mark 26.0) g11 Chapter.

g11 - Contingency Table Analysis.

https://www.nag.com/numeric/CL/nagdoc_cl26/html/G11/g11conts.html

NAG Copyright 2017.
"""
from ctypes import c_char_p
from ctypes import c_double
from ctypes import POINTER

from .util import get_library

from .util import Integer
from .util import NagError
from .util import Nag_TableStats
from .util import Nag_TableUpdate
from .util import Nag_Weightstype
from .util import Nag_TabulateVar
from .util import Nag_OrderType
from .util import Nag_Boolean
_library_cl = get_library()
STRING = c_char_p


g11aac_ctypes = _library_cl.g11aac
g11aac_ctypes.restype = None
g11aac_ctypes.argtypes = [Integer, Integer, POINTER(Integer), Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(NagError)]


def g11aac(nrow, ncol, nobst, tdt, expt, chist, prob, chi, g, df, fail):
    """
    chi**2 statistics for two-way contingency table.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/g11/g11aac.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        nobst, expt, chist, prob, chi, g, df
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        expt, chist, prob, chi, g, df
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        nobst
    """
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _nobst = nag_integer_type_check_and_cast('g11aac', nobst, 'nobst')
    _expt = nag_double_type_check_and_cast('g11aac', expt, 'expt')
    _chist = nag_double_type_check_and_cast('g11aac', chist, 'chist')
    _prob = nag_double_type_check_and_cast('g11aac', prob, 'prob')
    _chi = nag_double_type_check_and_cast('g11aac', chi, 'chi')
    _g = nag_double_type_check_and_cast('g11aac', g, 'g')
    _df = nag_double_type_check_and_cast('g11aac', df, 'df')
    g11aac_ctypes(nrow, ncol, _nobst, tdt, _expt, _chist, _prob, _chi, _g, _df, fail)

g11bac_ctypes = _library_cl.g11bac
g11bac_ctypes.restype = None
g11bac_ctypes.argtypes = [Nag_TableStats, Nag_TableUpdate, Nag_Weightstype, Integer, Integer, POINTER(Integer), POINTER(Integer), POINTER(Integer), Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), Integer, POINTER(Integer), POINTER(Integer), POINTER(Integer), POINTER(Integer), POINTER(c_double), POINTER(NagError)]


def g11bac(stat, update, weight, n, nfac, sf, lfac, factor, tdf, y, wt, table, maxt, ncells, ndim, idim, count, commar, fail):
    """
    Computes multiway table from set of classification factors using
    selected statistic.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/g11/g11bac.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        sf, lfac, factor, y, wt, table, ncells, ndim, idim, count, commar
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        y, wt, table, commar
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        sf, lfac, factor, ncells, ndim, idim, count
    """
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _sf = nag_integer_type_check_and_cast('g11bac', sf, 'sf')
    _lfac = nag_integer_type_check_and_cast('g11bac', lfac, 'lfac')
    _factor = nag_integer_type_check_and_cast('g11bac', factor, 'factor')
    _y = nag_double_type_check_and_cast('g11bac', y, 'y')
    _wt = nag_double_type_check_and_cast('g11bac', wt, 'wt')
    _table = nag_double_type_check_and_cast('g11bac', table, 'table')
    _ncells = nag_integer_type_check_and_cast('g11bac', ncells, 'ncells')
    _ndim = nag_integer_type_check_and_cast('g11bac', ndim, 'ndim')
    _idim = nag_integer_type_check_and_cast('g11bac', idim, 'idim')
    _count = nag_integer_type_check_and_cast('g11bac', count, 'count')
    _commar = nag_double_type_check_and_cast('g11bac', commar, 'commar')
    g11bac_ctypes(stat, update, weight, n, nfac, _sf, _lfac, _factor, tdf, _y, _wt, _table, maxt, _ncells, _ndim, _idim, _count, _commar, fail)

g11bbc_ctypes = _library_cl.g11bbc
g11bbc_ctypes.restype = None
g11bbc_ctypes.argtypes = [Nag_TabulateVar, Integer, Integer, POINTER(Integer), POINTER(Integer), POINTER(Integer), Integer, c_double, POINTER(c_double), POINTER(c_double), POINTER(c_double), Integer, POINTER(Integer), POINTER(Integer), POINTER(Integer), POINTER(Integer), POINTER(NagError)]


def g11bbc(_type, n, nfac, sf, lfac, factor, tdf, percnt, y, wt, table, maxt, ncells, ndim, idim, count, fail):
    """
    Computes multiway table from set of classification factors using given
    percentile/quantile.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/g11/g11bbc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        sf, lfac, factor, y, wt, table, ncells, ndim, idim, count
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        y, wt, table
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        sf, lfac, factor, ncells, ndim, idim, count
    """
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _sf = nag_integer_type_check_and_cast('g11bbc', sf, 'sf')
    _lfac = nag_integer_type_check_and_cast('g11bbc', lfac, 'lfac')
    _factor = nag_integer_type_check_and_cast('g11bbc', factor, 'factor')
    _y = nag_double_type_check_and_cast('g11bbc', y, 'y')
    _wt = nag_double_type_check_and_cast('g11bbc', wt, 'wt')
    _table = nag_double_type_check_and_cast('g11bbc', table, 'table')
    _ncells = nag_integer_type_check_and_cast('g11bbc', ncells, 'ncells')
    _ndim = nag_integer_type_check_and_cast('g11bbc', ndim, 'ndim')
    _idim = nag_integer_type_check_and_cast('g11bbc', idim, 'idim')
    _count = nag_integer_type_check_and_cast('g11bbc', count, 'count')
    g11bbc_ctypes(_type, n, nfac, _sf, _lfac, _factor, tdf, percnt, _y, _wt, _table, maxt, _ncells, _ndim, _idim, _count, fail)

g11bcc_ctypes = _library_cl.g11bcc
g11bcc_ctypes.restype = None
g11bcc_ctypes.argtypes = [Nag_TableStats, POINTER(c_double), Integer, Integer, POINTER(Integer), POINTER(Integer), POINTER(c_double), Integer, POINTER(Integer), POINTER(Integer), POINTER(Integer), POINTER(c_double), POINTER(NagError)]


def g11bcc(stat, table, ncells, ndim, idim, isdim, subtable, maxst, mcells, mdim, mlevel, commar, fail):
    """
    Computes marginal tables for multiway table computed by g11bac or
    g11bbc.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/g11/g11bcc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        table, idim, isdim, subtable, mcells, mdim, mlevel, commar
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        table, subtable, commar
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        idim, isdim, mcells, mdim, mlevel
    """
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _table = nag_double_type_check_and_cast('g11bcc', table, 'table')
    _idim = nag_integer_type_check_and_cast('g11bcc', idim, 'idim')
    _isdim = nag_integer_type_check_and_cast('g11bcc', isdim, 'isdim')
    _subtable = nag_double_type_check_and_cast('g11bcc', subtable, 'subtable')
    _mcells = nag_integer_type_check_and_cast('g11bcc', mcells, 'mcells')
    _mdim = nag_integer_type_check_and_cast('g11bcc', mdim, 'mdim')
    _mlevel = nag_integer_type_check_and_cast('g11bcc', mlevel, 'mlevel')
    _commar = nag_double_type_check_and_cast('g11bcc', commar, 'commar')
    g11bcc_ctypes(stat, _table, ncells, ndim, _idim, _isdim, _subtable, maxst, _mcells, _mdim, _mlevel, _commar, fail)

g11cac_ctypes = _library_cl.g11cac
g11cac_ctypes.restype = None
g11cac_ctypes.argtypes = [Nag_OrderType, Integer, Integer, Integer, POINTER(c_double), Integer, POINTER(Integer), Integer, POINTER(Integer), POINTER(Integer), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Integer), POINTER(Integer), c_double, Integer, Integer, STRING, POINTER(NagError)]


def g11cac(order, n, m, ns, z, pdz, isz, p, ic, isi, dev, b, se, sc, cov, nca, nct, tol, maxit, iprint, outfile, fail):
    """
    Returns parameter estimates for the conditional analysis of stratified
    data.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/g11/g11cac.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        z, isz, ic, isi, dev, b, se, sc, cov, nca, nct
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        z, dev, b, se, sc, cov
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        isz, ic, isi, nca, nct
    """
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _z = nag_double_type_check_and_cast('g11cac', z, 'z')
    _isz = nag_integer_type_check_and_cast('g11cac', isz, 'isz')
    _ic = nag_integer_type_check_and_cast('g11cac', ic, 'ic')
    _isi = nag_integer_type_check_and_cast('g11cac', isi, 'isi')
    _dev = nag_double_type_check_and_cast('g11cac', dev, 'dev')
    _b = nag_double_type_check_and_cast('g11cac', b, 'b')
    _se = nag_double_type_check_and_cast('g11cac', se, 'se')
    _sc = nag_double_type_check_and_cast('g11cac', sc, 'sc')
    _cov = nag_double_type_check_and_cast('g11cac', cov, 'cov')
    _nca = nag_integer_type_check_and_cast('g11cac', nca, 'nca')
    _nct = nag_integer_type_check_and_cast('g11cac', nct, 'nct')
    g11cac_ctypes(order, n, m, ns, _z, pdz, _isz, p, _ic, _isi, _dev, _b, _se, _sc, _cov, _nca, _nct, tol, maxit, iprint, outfile, fail)

g11sac_ctypes = _library_cl.g11sac
g11sac_ctypes.restype = None
g11sac_ctypes.argtypes = [Nag_OrderType, Integer, Integer, Nag_Boolean, Integer, POINTER(Nag_Boolean), Integer, POINTER(Integer), POINTER(c_double), POINTER(c_double), Integer, STRING, c_double, Integer, Nag_Boolean, POINTER(Integer), POINTER(c_double), POINTER(c_double), POINTER(c_double), Integer, POINTER(c_double), POINTER(c_double), Integer, POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Integer), POINTER(c_double), POINTER(c_double), POINTER(Integer), POINTER(c_double), POINTER(NagError)]


def g11sac(order, p, n, gprob, ns, x, pdx, irl, a, c, iprint, outfile, cgetol, maxit, chisqr, niter, alpha, pigam, cm, pdcm, g, expp, pde, obs, exf, y, iob, rlogl, chi, idf, siglev, fail):
    """
    Contingency table, latent variable model for binary data.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/g11/g11sac.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        x, irl, a, c, niter, alpha, pigam, cm, g, expp, obs, exf, y, iob, rlogl, chi, idf, siglev
    TypeError will be raised when the following arguments are not of data type
    Nag_Boolean:
        x
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        a, c, alpha, pigam, cm, g, expp, obs, exf, y, rlogl, chi, siglev
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        irl, niter, iob, idf
    """
    from .util import nag_enum_type_check_and_cast
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _x = nag_enum_type_check_and_cast('g11sac', x, 'x', Nag_Boolean)
    _irl = nag_integer_type_check_and_cast('g11sac', irl, 'irl')
    _a = nag_double_type_check_and_cast('g11sac', a, 'a')
    _c = nag_double_type_check_and_cast('g11sac', c, 'c')
    _niter = nag_integer_type_check_and_cast('g11sac', niter, 'niter')
    _alpha = nag_double_type_check_and_cast('g11sac', alpha, 'alpha')
    _pigam = nag_double_type_check_and_cast('g11sac', pigam, 'pigam')
    _cm = nag_double_type_check_and_cast('g11sac', cm, 'cm')
    _g = nag_double_type_check_and_cast('g11sac', g, 'g')
    _expp = nag_double_type_check_and_cast('g11sac', expp, 'expp')
    _obs = nag_double_type_check_and_cast('g11sac', obs, 'obs')
    _exf = nag_double_type_check_and_cast('g11sac', exf, 'exf')
    _y = nag_double_type_check_and_cast('g11sac', y, 'y')
    _iob = nag_integer_type_check_and_cast('g11sac', iob, 'iob')
    _rlogl = nag_double_type_check_and_cast('g11sac', rlogl, 'rlogl')
    _chi = nag_double_type_check_and_cast('g11sac', chi, 'chi')
    _idf = nag_integer_type_check_and_cast('g11sac', idf, 'idf')
    _siglev = nag_double_type_check_and_cast('g11sac', siglev, 'siglev')
    g11sac_ctypes(order, p, n, gprob, ns, _x, pdx, _irl, _a, _c, iprint, outfile, cgetol, maxit, chisqr, _niter, _alpha, _pigam, _cm, pdcm, _g, _expp, pde, _obs, _exf, _y, _iob, _rlogl, _chi, _idf, _siglev, fail)

g11sbc_ctypes = _library_cl.g11sbc
g11sbc_ctypes.restype = None
g11sbc_ctypes.argtypes = [Nag_OrderType, Integer, Integer, POINTER(Integer), POINTER(Nag_Boolean), Integer, POINTER(Integer), POINTER(NagError)]


def g11sbc(order, p, n, ns, x, pdx, irl, fail):
    """
    Frequency count for g11sac.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/g11/g11sbc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        ns, x, irl
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ns, irl
    TypeError will be raised when the following arguments are not of data type
    Nag_Boolean:
        x
    """
    from .util import nag_enum_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _ns = nag_integer_type_check_and_cast('g11sbc', ns, 'ns')
    _x = nag_enum_type_check_and_cast('g11sbc', x, 'x', Nag_Boolean)
    _irl = nag_integer_type_check_and_cast('g11sbc', irl, 'irl')
    g11sbc_ctypes(order, p, n, _ns, _x, pdx, _irl, fail)

nag_chi_sq_2_way_table = g11aac
nag_tabulate_stats = g11bac
nag_tabulate_percentile = g11bbc
nag_tabulate_margin = g11bcc
nag_condl_logistic = g11cac
nag_binary_factor = g11sac
nag_binary_factor_service = g11sbc
__all__ = ['g11aac', 'g11bac', 'g11bbc', 'g11bcc', 'g11cac', 'g11sac',
           'g11sbc', 'nag_binary_factor', 'nag_binary_factor_service',
           'nag_chi_sq_2_way_table', 'nag_condl_logistic',
           'nag_tabulate_margin', 'nag_tabulate_percentile',
           'nag_tabulate_stats']
