"""
Deprecated wrappers for the NAG C Library (Mark 26.0) g12 Chapter.

g12 - Survival Analysis.

https://www.nag.com/numeric/CL/nagdoc_cl26/html/G12/g12conts.html

NAG Copyright 2017.
"""
from ctypes import c_char_p
from ctypes import c_double
from ctypes import POINTER

from .util import get_library
from .util import NagError
from .util import Integer
from .util import Nag_OrderType
STRING = c_char_p
_library_cl = get_library()

g12aac_ctypes = _library_cl.g12aac
g12aac_ctypes.restype = None
g12aac_ctypes.argtypes = [Integer, POINTER(c_double), POINTER(Integer), POINTER(Integer), POINTER(Integer), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(NagError)]


def g12aac(n, t, ic, freq, nd, tp, p, psig, fail):
    """
    Computes Kaplan--Meier (product-limit) estimates of survival
    probabilities.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/g12/g12aac.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        t, ic, freq, nd, tp, p, psig
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        t, tp, p, psig
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ic, freq, nd
    """
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _t = nag_double_type_check_and_cast('g12aac', t, 't')
    _ic = nag_integer_type_check_and_cast('g12aac', ic, 'ic')
    _freq = nag_integer_type_check_and_cast('g12aac', freq, 'freq')
    _nd = nag_integer_type_check_and_cast('g12aac', nd, 'nd')
    _tp = nag_double_type_check_and_cast('g12aac', tp, 'tp')
    _p = nag_double_type_check_and_cast('g12aac', p, 'p')
    _psig = nag_double_type_check_and_cast('g12aac', psig, 'psig')
    g12aac_ctypes(n, _t, _ic, _freq, _nd, _tp, _p, _psig, fail)

g12abc_ctypes = _library_cl.g12abc
g12abc_ctypes.restype = None
g12abc_ctypes.argtypes = [Integer, POINTER(c_double), POINTER(Integer), POINTER(Integer), Integer, POINTER(Integer), POINTER(c_double), POINTER(c_double), POINTER(Integer), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Integer), POINTER(Integer), POINTER(Integer), Integer, POINTER(NagError)]


def g12abc(n, t, ic, grp, ngrp, ifreq, wt, ts, df, p, obsd, expt, nd, di, ni, ldn, fail):
    """
    Computes rank statistics for comparing survival curves.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/g12/g12abc.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        t, ic, grp, ifreq, wt, ts, df, p, obsd, expt, nd, di, ni
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        t, wt, ts, p, obsd, expt
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        ic, grp, ifreq, df, nd, di, ni
    """
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _t = nag_double_type_check_and_cast('g12abc', t, 't')
    _ic = nag_integer_type_check_and_cast('g12abc', ic, 'ic')
    _grp = nag_integer_type_check_and_cast('g12abc', grp, 'grp')
    _ifreq = nag_integer_type_check_and_cast('g12abc', ifreq, 'ifreq')
    _wt = nag_double_type_check_and_cast('g12abc', wt, 'wt')
    _ts = nag_double_type_check_and_cast('g12abc', ts, 'ts')
    _df = nag_integer_type_check_and_cast('g12abc', df, 'df')
    _p = nag_double_type_check_and_cast('g12abc', p, 'p')
    _obsd = nag_double_type_check_and_cast('g12abc', obsd, 'obsd')
    _expt = nag_double_type_check_and_cast('g12abc', expt, 'expt')
    _nd = nag_integer_type_check_and_cast('g12abc', nd, 'nd')
    _di = nag_integer_type_check_and_cast('g12abc', di, 'di')
    _ni = nag_integer_type_check_and_cast('g12abc', ni, 'ni')
    g12abc_ctypes(n, _t, _ic, _grp, ngrp, _ifreq, _wt, _ts, _df, _p, _obsd, _expt, _nd, _di, _ni, ldn, fail)

g12bac_ctypes = _library_cl.g12bac
g12bac_ctypes.restype = None
g12bac_ctypes.argtypes = [Integer, Integer, Integer, POINTER(c_double), Integer, POINTER(Integer), Integer, POINTER(c_double), POINTER(Integer), POINTER(c_double), POINTER(Integer), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(Integer), POINTER(c_double), POINTER(c_double), Integer, Integer, c_double, Integer, Integer, STRING, POINTER(NagError)]


def g12bac(n, m, ns, z, tdz, sz, ip, t, ic, omega, isi, dev, b, se, sc, cov, res, nd, tp, sur, tdsur, ndmax, tol, maxiter, iprint, outfile, fail):
    """
    Fits Cox's proportional hazard model.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/g12/g12bac.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        z, sz, t, ic, omega, isi, dev, b, se, sc, cov, res, nd, tp, sur
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        z, t, omega, dev, b, se, sc, cov, res, tp, sur
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        sz, ic, isi, nd
    """
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _z = nag_double_type_check_and_cast('g12bac', z, 'z')
    _sz = nag_integer_type_check_and_cast('g12bac', sz, 'sz')
    _t = nag_double_type_check_and_cast('g12bac', t, 't')
    _ic = nag_integer_type_check_and_cast('g12bac', ic, 'ic')
    _omega = nag_double_type_check_and_cast('g12bac', omega, 'omega')
    _isi = nag_integer_type_check_and_cast('g12bac', isi, 'isi')
    _dev = nag_double_type_check_and_cast('g12bac', dev, 'dev')
    _b = nag_double_type_check_and_cast('g12bac', b, 'b')
    _se = nag_double_type_check_and_cast('g12bac', se, 'se')
    _sc = nag_double_type_check_and_cast('g12bac', sc, 'sc')
    _cov = nag_double_type_check_and_cast('g12bac', cov, 'cov')
    _res = nag_double_type_check_and_cast('g12bac', res, 'res')
    _nd = nag_integer_type_check_and_cast('g12bac', nd, 'nd')
    _tp = nag_double_type_check_and_cast('g12bac', tp, 'tp')
    _sur = nag_double_type_check_and_cast('g12bac', sur, 'sur')
    g12bac_ctypes(n, m, ns, _z, tdz, _sz, ip, _t, _ic, _omega, _isi, _dev, _b, _se, _sc, _cov, _res, _nd, _tp, _sur, tdsur, ndmax, tol, maxiter, iprint, outfile, fail)

g12zac_ctypes = _library_cl.g12zac
g12zac_ctypes.restype = None
g12zac_ctypes.argtypes = [Nag_OrderType, Integer, Integer, Integer, POINTER(c_double), Integer, POINTER(Integer), Integer, POINTER(c_double), POINTER(Integer), POINTER(Integer), POINTER(Integer), POINTER(Integer), POINTER(Integer), POINTER(c_double), Integer, POINTER(Integer), POINTER(Integer), POINTER(c_double), POINTER(Integer), POINTER(NagError)]


def g12zac(order, n, m, ns, z, pdz, isz, ip, t, ic, isi, num, ixs, nxs, x, mxn, _id, nd, tp, irs, fail):
    """
    Creates the risk sets associated with the Cox proportional hazards model
    for fixed covariates.
    https://www.nag.com/numeric/CL/nagdoc_cl26/html/g12/g12zac.html

    TypeError will be raised when the following arguments are not instances
    of numpy.ndarray:
        z, isz, t, ic, isi, num, ixs, nxs, x, _id, nd, tp, irs
    TypeError will be raised when the following arguments are not of data type
    numpy.float:
        z, t, x, tp
    TypeError will be raised when the following arguments are not of data type
    numpy.int32 or numpy.int64:
        isz, ic, isi, num, ixs, nxs, _id, nd, irs
    """
    from .util import nag_double_type_check_and_cast
    from .util import nag_integer_type_check_and_cast
    _z = nag_double_type_check_and_cast('g12zac', z, 'z')
    _isz = nag_integer_type_check_and_cast('g12zac', isz, 'isz')
    _t = nag_double_type_check_and_cast('g12zac', t, 't')
    _ic = nag_integer_type_check_and_cast('g12zac', ic, 'ic')
    _isi = nag_integer_type_check_and_cast('g12zac', isi, 'isi')
    _num = nag_integer_type_check_and_cast('g12zac', num, 'num')
    _ixs = nag_integer_type_check_and_cast('g12zac', ixs, 'ixs')
    _nxs = nag_integer_type_check_and_cast('g12zac', nxs, 'nxs')
    _x = nag_double_type_check_and_cast('g12zac', x, 'x')
    __id = nag_integer_type_check_and_cast('g12zac', _id, '_id')
    _nd = nag_integer_type_check_and_cast('g12zac', nd, 'nd')
    _tp = nag_double_type_check_and_cast('g12zac', tp, 'tp')
    _irs = nag_integer_type_check_and_cast('g12zac', irs, 'irs')
    g12zac_ctypes(order, n, m, ns, _z, pdz, _isz, ip, _t, _ic, _isi, _num, _ixs, _nxs, _x, mxn, __id, _nd, _tp, _irs, fail)

nag_prod_limit_surviv_fn = g12aac
nag_surviv_logrank = g12abc
nag_surviv_cox_model = g12bac
nag_surviv_risk_sets = g12zac
__all__ = ['g12aac', 'g12abc', 'g12bac', 'g12zac',
           'nag_prod_limit_surviv_fn', 'nag_surviv_cox_model',
           'nag_surviv_logrank', 'nag_surviv_risk_sets']
