#!/usr/bin/env python
"""
Tests of interf. to the NAG Engine from Python via the naginterfaces package.

The NAG and Intel compilers are required to run these tests.

Public use of this module is not supported by NAG.
"""

# NAG Copyright 2017-2019.

# pylint: disable=protected-access,too-many-lines

import ctypes
import _ctypes
import os
import platform
import shutil
import subprocess
import sys
import tempfile
import unittest
import warnings

import numpy as np

from naginterfaces.base import utils as b_utils
from naginterfaces.library import _utils as np_utils

def destroy_so_lib(
        so_lib,
):
    """
    Test utility to close so_lib.
    """
    if platform.system() == 'Windows':
        _ctypes.FreeLibrary(so_lib._handle) # pylint: disable=no-member
    else:
        _ctypes.dlclose(so_lib._handle) # pylint: disable=no-member

def get_so_lib_from_file(
        so_name,
):
    """
    Test utility for opening so_name.
    """
    return ctypes.CDLL(so_name)

class CompilerState(object):
    """
    A class for managing compilation of code.
    """

    verbose = False

    def __init__(
            self,
            compiler,
            debug=True,
    ):

        if compiler not in ['ifort', 'nagfor']:
            raise ValueError('NYI for compiler ' + compiler) # pragma: no cover

        self.compiler = compiler
        self.debug = debug
        self.linker = (
            'clang' if compiler == 'ifort' and platform.system() == 'Darwin'
            else self.compiler
        )
        self.objext = self.get_objext()
        self.chrlen_ctype = self.get_chrlen_ctype()

    def compilate( # pylint: disable=too-many-branches
            self,
            file_name,
            out_dir,
            options=None,
    ):
        """
        Compile file with base file_name to an object
        in out_dir.
        Add in options list if supplied.
        """
        compile_cmd = [self.compiler]

        if self.debug: # pragma: no cover

            if (
                    self.compiler == 'ifort' and
                    platform.system() == 'Windows'
            ):
                compile_cmd.append('-Z7')
            else:
                compile_cmd.append('-g')

        if self.compiler == 'ifort':
            pic_option = '-fPIC'

            if (
                    self.debug and
                    platform.system() != 'Windows'
            ):
                compile_cmd.extend([
                    '-traceback',
                    '-check', 'all',
                ])

            compile_cmd.append(
                '-nologo',
            )

            if platform.system() == 'Windows':
                compile_cmd.append('-module:' + out_dir)
            else:
                compile_cmd.extend(['-module', out_dir])

        elif self.compiler == 'nagfor':
            pic_option = '-PIC'

            if self.debug: # pragma: no cover
                compile_cmd.extend([
                    '-gline',
                    '-C=all',
                    '-nan',
                ])

            compile_cmd.extend([
                '-I', out_dir,
                '-mdir', out_dir,
            ])

            if b_utils._get_sys_data_model() == 'ILP32': # pragma: no cover
                compile_cmd.append('-abi=32')

        else:
            raise ValueError( # pragma: no cover
                'NYI for compiler ' + self.compiler
            )

        if options is not None:
            compile_cmd.extend(options)

        compile_cmd.append('-c')

        if platform.system() != 'Windows':
            compile_cmd.append(pic_option)

        compile_cmd.extend([
            '-o', os.path.join(out_dir, '.'.join([file_name, self.objext])),
            os.path.join(out_dir, file_name + '.f90')
        ])

        if self.verbose:
            print(
                'Compile command: {:s}'.format(
                    ' '.join(compile_cmd)
                )
            ) # pragma: no cover

        # Discard nagfor banner output:
        subprocess.check_output(compile_cmd, stderr=subprocess.STDOUT)

    def create_so(
            self,
            file_name,
            out_dir,
            options=None,
    ):
        """
        Compile file with base file_name
        to a shared library, all in out_dir.
        Return the full path to the shared object.
        """
        # pylint: disable=too-many-branches

        self.compilate(file_name, out_dir, options=options)

        so_cmd = [self.linker]

        if self.linker == 'clang':

            if platform.system() == 'Windows':
                raise ValueError( # pragma: no cover
                    'NYI for linker ' + self.linker
                    )
            if platform.system() == 'Darwin':
                so_cmd.append('-dynamiclib')
            else:
                raise ValueError( # pragma: no cover
                    'NYI (-shared?) for linker ' + self.linker
                    )

        elif self.linker == 'ifort':
            so_cmd.extend([
                '-nologo',
            ])

            if platform.system() == 'Windows':
                so_cmd.extend([
                    '-dll',
                ])
                def_file = os.path.join(out_dir, 'exports.def')

                if platform.machine().endswith('64'):
                    exp_list = [
                        file_name.upper(),
                        '='.join([
                            b_utils._mangle_sym_name(file_name),
                            file_name.upper(),
                        ]),
                    ]
                else: # pragma: no cover
                    exp_list = [b_utils._mangle_sym_name(file_name)]

                with open(def_file, 'w') as def_fo:
                    def_fo.write('EXPORTS\n')
                    def_fo.writelines(
                        [' ' + exp_line + '\n' for exp_line in exp_list]
                    )

                so_cmd.append(def_file)
            elif platform.system() == 'Darwin':
                so_cmd.append('-dynamiclib') # pragma: no cover
            else:
                so_cmd.append('-shared')

        elif self.linker == 'nagfor':
            so_cmd.extend([
                '-ideclient',
                '-unsharedrts',
                '-Wl,-shared',
            ])

            if b_utils._get_sys_data_model() == 'ILP32': # pragma: no cover
                so_cmd.append('-abi=32')

        else:
            raise ValueError( # pragma: no cover
                'NYI for linker ' + self.linker
            )

        output_so = os.path.join(out_dir, file_name + '.so')
        so_cmd.extend([
            '-o', output_so,
            os.path.join(out_dir, '.'.join([file_name, self.objext]))
        ])

        if self.linker == 'clang':

            if self.compiler == 'ifort':
                so_cmd.extend([
                    '-l' + lib for lib in ['ifcoremt', 'intlc']
                ])
            else:
                raise ValueError( # pragma: no cover
                    'NYI for compiler ' + self.compiler
                    )

        if self.verbose:
            print(
                'SO command: {:s}'.format(
                    ' '.join(so_cmd)
                )
            ) # pragma: no cover

        # Discard output from Microsoft link:
        subprocess.check_output(so_cmd)

        os.remove(os.path.join(out_dir, '.'.join([file_name, self.objext])))

        return output_so

    def get_chrlen_ctype(
            self,
    ):
        """
        Return the C type used by this compiler for hidden string lengths.
        """

        if self.compiler == 'ifort':

            if b_utils._get_sys_data_model() == 'ILP32': # pragma: no cover
                return b_utils._SYS_CTYPES_INTS[32]

            return b_utils._SYS_CTYPES_INTS[64]

        if self.compiler == 'nagfor':
            return b_utils._SYS_CTYPES_INTS[32]

        raise ValueError( # pragma: no cover
            'NYI for compiler ' + self.compiler
        )

    def get_objext(
            self,
    ):
        """
        Return the object-file extension for this platform.
        """
        if (
                self.compiler == 'ifort' and
                platform.system() == 'Windows'
        ):
            return 'obj'

        return 'o'

def save_engine_state():
    """
    Save _EngineState attributes that may vary between compilers.
    """
    return (
        b_utils._EngineState._compilers,
        b_utils._EngineState._chrlen_ctype,
        b_utils._EngineState._complex_fun_rtn,
        b_utils._EngineState._literal_true_int,
    )

def restore_engine_state(old_state):
    """
    Restore _EngineState attributes that may vary between compilers.
    """
    b_utils._EngineState._compilers = old_state[0]
    b_utils._EngineState._chrlen_ctype = old_state[1]
    b_utils._EngineState._complex_fun_rtn = old_state[2]
    b_utils._EngineState._literal_true_int = old_state[3]

class PassTest(object):
    """
    Abstract settings for testing passing data to or from the Engine.
    """

    engine_state = None

    @classmethod
    def set_up_deps(cls):
        """
        Compile dependent modules.
        """
        for dep_file in cls.deps:
            shutil.copy(
                os.path.join(
                    os.path.dirname(__file__),
                    'src',
                    dep_file + '.f90'
                ),
                cls.workspace
            )
            cls.compiler.compilate(
                dep_file,
                cls.workspace, options=cls.compiler_options
            )

    @classmethod
    def set_up(cls, file_name, compiler, deps=False, extra_files=None):
        """
        Create the class workspace.
        Compile the low-level extension.
        """
        cls.workspace = tempfile.mkdtemp(prefix=cls.__name__)

        cls.compiler = CompilerState(compiler)

        if cls.engine_state._int_model['bits'] == 64:
            cls.compiler_options = ['-i8']
        else:
            cls.compiler_options = None

        cls.deps = [
            'nag_engine_data_type',
        ]

        if deps:
            cls.set_up_deps()

        cls.file_name = file_name
        files = [cls.file_name]

        if extra_files is not None:
            files.extend(extra_files)

        for filen in files:
            shutil.copy(
                os.path.join(
                    os.path.dirname(__file__),
                    'src',
                    filen + '.f90'
                ),
                cls.workspace
            )

        cls.so_lib = get_so_lib_from_file(
            cls.compiler.create_so(
                cls.file_name, cls.workspace, options=cls.compiler_options
            )
        )
        cls.engine_fun = b_utils._get_symbol_from_lib_handle(
            cls.so_lib, cls.file_name
        )

        cls.extras = {}

        if extra_files is not None:

            for filen in extra_files:
                cls.extras[filen] = {}
                extra_so = get_so_lib_from_file(
                    cls.compiler.create_so(
                        filen, cls.workspace, options=cls.compiler_options
                    )
                )
                cls.extras[filen]['so_lib'] = extra_so
                cls.extras[filen]['fun'] = b_utils._get_symbol_from_lib_handle(
                    extra_so, filen
                )

    @classmethod
    def destroy_shlib(cls, shlib, file_name, file_exts):
        """
        Destroy a shared lib and associated files.
        """
        if cls.compiler.verbose:
            print('Destroying for ' + file_name) # pragma: no cover
        if shlib is not None:
            destroy_so_lib(shlib)
        for file_ext in file_exts:
            trash_file = os.path.join(
                cls.workspace,
                '.'.join([file_name, file_ext]),
            )
            if os.path.isfile(trash_file):
                os.unlink(trash_file)
        if cls.compiler.verbose:
            print('Destroyed.') # pragma: no cover

    @classmethod
    def tearDownClass(cls): # pylint: disable=invalid-name
        """
        Delete the class workspace.
        """
        primary_file_exts = ['so', 'f90']
        if platform.system() == 'Windows':
            primary_file_exts.extend(['exp', 'lib'])
        cls.destroy_shlib(cls.so_lib, cls.file_name, primary_file_exts)
        for filen in cls.extras:
            cls.destroy_shlib(
                cls.extras[filen]['so_lib'],
                filen,
                primary_file_exts,
            )
        for filen in cls.deps:
            file_exts = []
            file_exts.extend(primary_file_exts)
            file_exts.extend([cls.compiler.objext, 'mod'])
            cls.destroy_shlib(
                None,
                filen,
                file_exts,
            )
        if platform.system() == 'Windows':
            if cls.compiler.verbose:
                print('Removing exports.def?') # pragma: no cover
            e_def = os.path.join(cls.workspace, 'exports.def')
            if os.path.isfile(e_def):
                os.unlink(e_def)
                if cls.compiler.verbose:
                    print('Removed exports.def') # pragma: no cover
        if cls.compiler.verbose:
            print('Removing workspace ' + cls.workspace) # pragma: no cover
        # pylint: disable=bare-except
        try:
            dir_l = os.listdir(cls.workspace)
            os.rmdir(cls.workspace)
        except: # pragma: no cover
            print(
                'Warning: failed to delete ' + cls.workspace
            )
            print('Warning: contents are ' + str(dir_l)) # pragma: no cover
        if cls.compiler.verbose:
            print('Removed workspace.') # pragma: no cover
        restore_engine_state(cls.default_engine_state) # pylint: disable=no-member

class PassTestCharCallback(PassTest):
    """
    Abstract low-level unit tests for char-valued callbacks in the Engine.
    """

    @classmethod
    def set_up_fun(cls, compiler):
        """
        The abstract test setup.
        """
        cls.default_engine_state = save_engine_state()

        cls.engine_state = b_utils._EngineState

        if compiler == 'ifort':
            cls.engine_state._compilers = {'Fortran': 'Intel'}
        else:
            cls.engine_state._compilers = {'Fortran': 'NAG'}

        cls.set_up('char_retval_callback_test', compiler)

        cls.engine_state._chrlen_ctype = cls.compiler.chrlen_ctype

        cls.cb = cls.engine_state._format_char_cfun_proto(
            b_utils._BYREF_FLOAT32, # Input, x.
        )

        cls.engine_fun.restype = None
        cls.engine_fun.argtypes = (
            b_utils._BYREF_FLOAT32, # Input, x.
            cls.cb,
            cls.engine_state._chrlen_ctype, # Input, chrlen of f.
        )

    def test_good_direct(self):
        """
        Test correct operation with a direct call.
        """

        def cb_f(retval, chrlen, x): # pylint: disable=invalid-name,missing-docstring
            b_utils._EngineCharScalarType(
                str(x.contents.value).encode(),
                {'fun_name': 'cb', 'returned': 'retval'},
                chrlen,
            ).refresh_ctypes_buffer(retval)

        self.engine_fun(
            b_utils._EngineFloat32ScalarType.c_type(42.),
            b_utils._cast_callback(self.cb, cb_f, {'fun_name': 'z99zz'}),
            30,
        )

    def test_good_wrapper(self):
        """
        Test correct operation via a Python wrapper.
        """

        def cb_f(x): # pylint: disable=invalid-name
            """
            Callback.
            """
            return str(x)

        # pylint: disable=invalid-name
        def py_fun(x, f):
            """
            Python wrapper for self.engine_fun.
            """

            def cb_h(retval, chrlen, x): # pylint: disable=invalid-name,missing-docstring
                _c = f(x.contents.value)
                b_utils._EngineCharScalarType(
                    _c,
                    {'fun_name': 'cb', 'returned': 'c'},
                    chrlen,
                ).refresh_ctypes_buffer(retval)

            _x = b_utils._EngineFloat32ScalarType(
                x,
                {'fun_name': 'py_fun', 'entity': 'x'}
            ).to_ctypes()
            b_utils._check_callable(
                self.cb, {'fun_name': 'z99zz', 'entity': 'f'},
            )
            self.engine_fun(
                _x,
                b_utils._cast_callback(
                    self.cb, cb_h, {'fun_name': 'z99zz', 'entity': 'f'}
                ),
                30,
            )

        py_fun(
            np.float32(42.), cb_f,
        )

class PassTestCharCallbackNag(PassTestCharCallback, unittest.TestCase):
    """
    Pass-test subsuite for NAG.
    """
    compiler = 'nagfor'

    @classmethod
    def setUpClass(cls):
        cls.set_up_fun(cls.compiler)

class PassTestCharCallbackIntel(PassTestCharCallback, unittest.TestCase):
    """
    Pass-test subsuite for Intel.
    """
    compiler = 'ifort'

    @classmethod
    def setUpClass(cls):
        cls.set_up_fun(cls.compiler)

class PassTestCharRetval(PassTest):
    """
    Abstract low-level unit tests for char-valued functions in the Engine.
    """

    @classmethod
    def set_up_fun(cls, compiler):
        """
        The abstract test setup.
        """
        cls.default_engine_state = save_engine_state()

        cls.engine_state = b_utils._EngineState

        if compiler == 'ifort':
            cls.engine_state._compilers = {'Fortran': 'Intel'}
        else:
            cls.engine_state._compilers = {'Fortran': 'NAG'}

        cls.set_up('char_retval_test', compiler)

        cls.engine_state._chrlen_ctype = cls.compiler.chrlen_ctype

        cls.engine_state._format_char_fun_proto(
            cls.engine_fun,
            b_utils._BYREF_FLOAT32, # Input, x.
        )

    def test_good_direct(self):
        """
        Test correct operation with a direct call.
        """
        engine_c = self.engine_state._call_char_fun(
            self.engine_fun, 30,
            b_utils._EngineFloat32ScalarType.c_type(42.),
        )
        engine_c_py = b_utils._EngineCharScalarType.to_py_from_retval(
            engine_c, chrlen=30
        )
        self.assertIsInstance(engine_c_py, str) # pylint: disable=no-member
        self.assertEqual(engine_c_py, '42.0') # pylint: disable=no-member

    def test_good_wrapper(self):
        """
        Test correct operation via a Python wrapper.
        """

        # pylint: disable=invalid-name
        def py_fun(x):
            """
            Python wrapper for self.engine_fun.
            """
            _x = b_utils._EngineFloat32ScalarType(
                x,
                {'fun_name': 'py_fun', 'entity': 'x'}
            ).to_ctypes()
            _retval = self.engine_state._call_char_fun(
                self.engine_fun, 30,
                _x,
            )
            return b_utils._EngineCharScalarType.to_py_from_retval(
                _retval, chrlen=30
            )

        c = py_fun(
            np.float32(42.),
        )
        self.assertIsInstance(c, str) # pylint: disable=no-member
        self.assertEqual( # pylint: disable=no-member
            c, '42.0',
        )

class PassTestCharRetvalNag(PassTestCharRetval, unittest.TestCase):
    """
    Pass-test subsuite for NAG.
    """
    compiler = 'nagfor'

    @classmethod
    def setUpClass(cls):
        cls.set_up_fun(cls.compiler)

class PassTestCharRetvalIntel(PassTestCharRetval, unittest.TestCase):
    """
    Pass-test subsuite for Intel.
    """
    compiler = 'ifort'

    @classmethod
    def setUpClass(cls):
        cls.set_up_fun(cls.compiler)

class PassTestComplexCallback(PassTest):
    """
    Abstract low-level unit tests for complex-valued callbacks in the Engine.
    """

    @classmethod
    def set_up_fun(cls, compiler):
        """
        The abstract test setup.
        """
        cls.default_engine_state = save_engine_state()

        cls.engine_state = b_utils._EngineState

        if compiler == 'ifort':
            cls.engine_state._compilers = {'Fortran': 'Intel'}
            cls.engine_state._complex_fun_rtn = 'as argument'
        else:
            cls.engine_state._compilers = {'Fortran': 'NAG'}
            if (
                    platform.system() == 'Windows' and
                    b_utils._get_sys_data_model() != 'ILP32'
            ):
                cls.engine_state._complex_fun_rtn = 'as argument'
            else:
                cls.engine_state._complex_fun_rtn = 'as retval'

        cls.set_up('complex_retval_callback_test', compiler)

        cls.engine_state._chrlen_ctype = cls.compiler.chrlen_ctype

        cls.cb = cls.engine_state._format_complex_cfun_proto(
            b_utils._EngineComplex64ScalarType.c_type,
            b_utils._BYREF_FLOAT32, # Input, x.
        )

        cls.engine_fun.restype = None
        cls.engine_fun.argtypes = (
            b_utils._BYREF_FLOAT32, # Input, x.
            cls.cb,
        )

    def test_good_direct(self):
        """
        Test correct operation with a direct call.
        """

        if self.engine_state._complex_fun_rtn == 'as argument':

            def cb_f(retval, x): # pylint: disable=invalid-name,missing-docstring
                b_utils._EngineComplex64ScalarType(
                    np.complex64(complex(
                        x.contents.value + 42., x.contents.value + 43.
                    )),
                    {'fun_name': 'cb', 'returned': 'retval'},
                ).refresh_ctypes_buffer(retval)

        else:

            def cb_f(x): # pylint: disable=invalid-name,missing-docstring
                raise ValueError('inconceivable') # pragma: no cover

        if self.engine_state._complex_fun_rtn == 'as argument':
            self.engine_fun(
                b_utils._EngineFloat32ScalarType.c_type(1.),
                b_utils._cast_callback(self.cb, cb_f, {'fun_name': 'z99zz'}),
            )
        else:
            with self.assertRaises(b_utils.NagException): # pylint: disable=no-member
                self.engine_fun(
                    b_utils._EngineFloat32ScalarType.c_type(1.),
                    b_utils._cast_callback(
                        self.cb, cb_f, {'fun_name': 'z99zz'}
                    ),
                )

    def test_good_wrapper(self):
        """
        Test correct operation via a Python wrapper.
        """

        def cb_f(x): # pylint: disable=invalid-name
            """
            Callback.
            """
            return np.complex64(complex(x + 42., x + 43.))

        # pylint: disable=invalid-name
        def py_fun(x, f):
            """
            Python wrapper for self.engine_fun.
            """

            if self.engine_state._complex_fun_rtn == 'as argument':

                def cb_h(retval, x): # pylint: disable=invalid-name,missing-docstring
                    _z = f(x.contents.value)
                    b_utils._EngineComplex64ScalarType(
                        _z,
                        {'fun_name': 'cb', 'returned': 'z'},
                    ).refresh_ctypes_buffer(retval)

            else:

                def cb_h(x): # pylint: disable=invalid-name,missing-docstring
                    raise ValueError('inconceivable') # pragma: no cover

            _x = b_utils._EngineFloat32ScalarType(
                x,
                {'fun_name': 'py_fun', 'entity': 'x'}
            ).to_ctypes()
            b_utils._check_callable(
                self.cb, {'fun_name': 'z99zz', 'entity': 'f'},
            )
            self.engine_fun(
                _x, b_utils._cast_callback(
                    self.cb, cb_h, {'fun_name': 'z99zz', 'entity': 'f'}
                ),
            )

        if self.engine_state._complex_fun_rtn == 'as argument':
            py_fun(
                np.float32(1.), cb_f,
            )
        else:
            self.assertRaises( # pylint: disable=no-member
                b_utils.NagException,
                py_fun,
                np.float32(1.), cb_f
            )

class PassTestComplexCallbackNag(PassTestComplexCallback, unittest.TestCase):
    """
    Pass-test subsuite for NAG.
    """
    compiler = 'nagfor'

    @classmethod
    def setUpClass(cls):
        cls.set_up_fun(cls.compiler)

class PassTestComplexCallbackIntel(PassTestComplexCallback, unittest.TestCase):
    """
    Pass-test subsuite for Intel.
    """
    compiler = 'ifort'

    @classmethod
    def setUpClass(cls):
        cls.set_up_fun(cls.compiler)

class PassTestComplexRetval(PassTest):
    """
    Abstract low-level unit tests for complex-valued functions in the Engine.
    """

    @classmethod
    def set_up_fun(cls, compiler):
        """
        The abstract test setup.
        """
        cls.default_engine_state = save_engine_state()

        cls.engine_state = b_utils._EngineState

        if compiler == 'ifort':
            cls.engine_state._compilers = {'Fortran': 'Intel'}
            cls.engine_state._complex_fun_rtn = 'as argument'
        else:
            cls.engine_state._compilers = {'Fortran': 'NAG'}
            if (
                    platform.system() == 'Windows' and
                    b_utils._get_sys_data_model() != 'ILP32'
            ):
                cls.engine_state._complex_fun_rtn = 'as argument'
            else:
                cls.engine_state._complex_fun_rtn = 'as retval'

        cls.set_up('complex_retval_test', compiler)

        cls.engine_state._chrlen_ctype = cls.compiler.chrlen_ctype

        cls.engine_state._format_complex_fun_proto(
            cls.engine_fun,
            b_utils._EngineComplex64ScalarType.c_type,
            b_utils._BYREF_FLOAT32, # Input, x.
            b_utils._BYREF_COMPLEX64, # Input, z.
        )

    def test_good_direct(self):
        """
        Test correct operation with a direct call.
        """
        engine_c = self.engine_state._call_complex_fun(
            self.engine_fun,
            b_utils._EngineFloat32ScalarType.c_type(1.),
            b_utils._EngineComplex64ScalarType.c_type(44, 45),
        )
        engine_c_py = b_utils._EngineComplex64ScalarType.to_py_from_retval(
            engine_c
        )
        self.assertIsInstance(engine_c_py, np.complex64) # pylint: disable=no-member
        self.assertEqual(engine_c_py, np.complex64(87 + 89j)) # pylint: disable=no-member

    def test_good_wrapper(self):
        """
        Test correct operation via a Python wrapper.
        """

        # pylint: disable=invalid-name
        def py_fun(x, z):
            """
            Python wrapper for self.engine_fun.
            """
            _x = b_utils._EngineFloat32ScalarType(
                x,
                {'fun_name': 'py_fun', 'entity': 'x'}
            ).to_ctypes()
            _z = b_utils._EngineComplex64ScalarType(
                z,
                {'fun_name': 'py_fun', 'entity': 'z'}
            ).to_ctypes()
            _retval = self.engine_state._call_complex_fun(
                self.engine_fun,
                _x, _z,
            )
            return b_utils._EngineComplex64ScalarType.to_py_from_retval(
                _retval
            )

        c = py_fun(
            np.float32(1.), np.complex64(44 + 45j),
        )
        self.assertIsInstance(c, np.complex64) # pylint: disable=no-member
        self.assertEqual( # pylint: disable=no-member
            c,
            np.complex64(87 + 89j),
        )

class PassTestComplexRetvalNag(PassTestComplexRetval, unittest.TestCase):
    """
    Pass-test subsuite for NAG.
    """
    compiler = 'nagfor'

    @classmethod
    def setUpClass(cls):
        cls.set_up_fun(cls.compiler)

class PassTestComplexRetvalIntel(PassTestComplexRetval, unittest.TestCase):
    """
    Pass-test subsuite for Intel.
    """
    compiler = 'ifort'

    @classmethod
    def setUpClass(cls):
        cls.set_up_fun(cls.compiler)

class PassTestFloatRetval(PassTest):
    """
    Abstract low-level unit tests for real-valued functions in the Engine.
    """
    en_scalar_type = None
    cast_inst = None
    test_file = None

    @classmethod
    def set_up_fun(cls, compiler):
        """
        The abstract test setup.
        """
        cls.default_engine_state = save_engine_state()

        cls.engine_state = b_utils._EngineState

        if compiler == 'ifort':
            cls.engine_state._compilers = {'Fortran': 'Intel'}
        else:
            cls.engine_state._compilers = {'Fortran': 'NAG'}

        cls.set_up(cls.test_file, compiler)

        cls.engine_state._chrlen_ctype = cls.compiler.chrlen_ctype

        cls.engine_fun.restype = cls.en_scalar_type.c_type
        cls.engine_fun.argtypes = (
            ctypes.POINTER(cls.en_scalar_type.c_type),
        )

    def test_good_direct(self):
        """
        Test correct operation with a direct call.
        """
        engine_r = self.engine_fun(
            self.en_scalar_type.c_type(1.),
        )
        engine_r_py = self.en_scalar_type.to_py_from_retval(engine_r)
        self.assertIsInstance(engine_r_py, self.cast_inst) # pylint: disable=no-member
        self.assertEqual(engine_r_py, 43.) # pylint: disable=no-member

    def test_good_wrapper(self):
        """
        Test correct operation via a Python wrapper.
        """

        # pylint: disable=invalid-name
        def py_fun(r):
            """
            Python wrapper for self.engine_fun.
            """
            _r = self.en_scalar_type( # pylint: disable=not-callable
                r,
                {'fun_name': 'py_fun', 'entity': 'r'}
            ).to_ctypes()
            _retval = self.engine_fun(
                _r,
            )
            return self.en_scalar_type.to_py_from_retval(_retval)

        r = py_fun(
            self.cast_inst(1.), # pylint: disable=not-callable
        )
        self.assertEqual( # pylint: disable=no-member
            r,
            43.,
        )

class PassTestFloat32Retval(PassTestFloatRetval):
    """
    Abstract low-level unit tests for float-valued functions in the Engine.
    """
    en_scalar_type = b_utils._EngineFloat32ScalarType
    cast_inst = np.float32
    test_file = 'float_retval_test'

class PassTestFloat32RetvalNag(PassTestFloat32Retval, unittest.TestCase):
    """
    Pass-test subsuite for NAG.
    """
    compiler = 'nagfor'

    @classmethod
    def setUpClass(cls):
        cls.set_up_fun(cls.compiler) # pylint: disable=no-member

class PassTestFloat32RetvalIntel(PassTestFloat32Retval, unittest.TestCase):
    """
    Pass-test subsuite for Intel.
    """
    compiler = 'ifort'

    @classmethod
    def setUpClass(cls):
        cls.set_up_fun(cls.compiler) # pylint: disable=no-member

class PassTestFloat64Retval(PassTestFloatRetval):
    """
    Abstract low-level unit tests for double-valued functions in the Engine.
    """
    en_scalar_type = b_utils._EngineFloat64ScalarType
    cast_inst = float
    test_file = 'double_retval_test'

class PassTestFloat64RetvalNag(PassTestFloat64Retval, unittest.TestCase):
    """
    Pass-test subsuite for NAG.
    """
    compiler = 'nagfor'

    @classmethod
    def setUpClass(cls):
        cls.set_up_fun(cls.compiler) # pylint: disable=no-member

class PassTestFloat64RetvalIntel(PassTestFloat64Retval, unittest.TestCase):
    """
    Pass-test subsuite for Intel.
    """
    compiler = 'ifort'

    @classmethod
    def setUpClass(cls):
        cls.set_up_fun(cls.compiler) # pylint: disable=no-member

class PassTestIntCallback(PassTest):
    """
    Abstract low-level unit tests for integer-valued callbacks in the Engine.
    """

    @classmethod
    def set_up_fun(cls, compiler):
        """
        The abstract test setup.
        """
        cls.default_engine_state = save_engine_state()

        cls.engine_state = b_utils._EngineState

        if compiler == 'ifort':
            cls.engine_state._compilers = {'Fortran': 'Intel'}
        else:
            cls.engine_state._compilers = {'Fortran': 'NAG'}

        cls.set_up(
            'int_retval_callback_test', compiler,
            extra_files=['int_retval_auxil']
        )

        cls.cb = cls.engine_state._sys_functype(
            b_utils._EngineIntScalarType.c_type,
            b_utils._BYREF_INT, # Input, x.
        )

        cls.cb_auxil = cls.extras['int_retval_auxil']['fun']
        cls.cb_auxil.restype = b_utils._EngineIntScalarType.c_type
        cls.cb_auxil.argtypes = (
            b_utils._BYREF_INT, # Input, x.
        )

        cls.engine_fun.restype = None
        cls.engine_fun.argtypes = (
            b_utils._BYREF_INT, # Input, x.
            cls.cb,
        )

    def test_good_direct_auxil(self):
        """
        Test correct operation with a direct call.
        """

        def cb_f(x): # pylint: disable=invalid-name,missing-docstring
            _f = self.cb_auxil(x)
            return _f

        self.engine_fun(
            b_utils._EngineIntScalarType.c_type(1),
            b_utils._cast_callback(self.cb, cb_f, {'fun_name': 'z99zz'}),
        )

class PassTestIntCallbackNag(PassTestIntCallback, unittest.TestCase):
    """
    Pass-test subsuite for NAG.
    """
    compiler = 'nagfor'

    @classmethod
    def setUpClass(cls):
        cls.set_up_fun(cls.compiler)

class PassTestIntCallbackIntel(PassTestIntCallback, unittest.TestCase):
    """
    Pass-test subsuite for Intel.
    """
    compiler = 'ifort'

    @classmethod
    def setUpClass(cls):
        cls.set_up_fun(cls.compiler)

class PassTestIntRetval(PassTest):
    """
    Abstract low-level unit tests for int-valued functions in the Engine.
    """
    en_scalar_type = b_utils._EngineIntScalarType
    cast_inst = int
    test_file = 'int_retval_test'

    @classmethod
    def set_up_fun(cls, compiler):
        """
        The abstract test setup.
        """
        cls.default_engine_state = save_engine_state()

        cls.engine_state = b_utils._EngineState

        if compiler == 'ifort':
            cls.engine_state._compilers = {'Fortran': 'Intel'}
        else:
            cls.engine_state._compilers = {'Fortran': 'NAG'}

        cls.set_up(cls.test_file, compiler)

        cls.engine_state._chrlen_ctype = cls.compiler.chrlen_ctype

        cls.engine_fun.restype = cls.en_scalar_type.c_type
        cls.engine_fun.argtypes = (
            ctypes.POINTER(cls.en_scalar_type.c_type),
        )

    def test_good_direct(self):
        """
        Test correct operation with a direct call.
        """
        engine_i = self.engine_fun(
            self.en_scalar_type.c_type(1),
        )
        engine_i_py = self.en_scalar_type.to_py_from_retval(engine_i)
        self.assertIsInstance(engine_i_py, self.cast_inst) # pylint: disable=no-member
        self.assertEqual(engine_i_py, 43) # pylint: disable=no-member

    def test_good_wrapper(self):
        """
        Test correct operation via a Python wrapper.
        """

        # pylint: disable=invalid-name
        def py_fun(i):
            """
            Python wrapper for self.engine_fun.
            """
            _i = self.en_scalar_type( # pylint: disable=not-callable
                i,
                {'fun_name': 'py_fun', 'entity': 'i'}
            ).to_ctypes()
            _retval = self.engine_fun(
                _i,
            )
            return self.en_scalar_type.to_py_from_retval(_retval)

        i = py_fun(
            self.cast_inst(1), # pylint: disable=not-callable
        )
        self.assertEqual( # pylint: disable=no-member
            i,
            43,
        )

class PassTestIntRetvalNag(PassTestIntRetval, unittest.TestCase):
    """
    Pass-test subsuite for NAG.
    """
    compiler = 'nagfor'

    @classmethod
    def setUpClass(cls):
        cls.set_up_fun(cls.compiler)

class PassTestIntRetvalIntel(PassTestIntRetval, unittest.TestCase):
    """
    Pass-test subsuite for Intel.
    """
    compiler = 'ifort'

    @classmethod
    def setUpClass(cls):
        cls.set_up_fun(cls.compiler)

class PassTestLogicalRetval(PassTest):
    """
    Abstract low-level unit tests for bool-valued functions in the Engine.
    """
    en_scalar_type = b_utils._EngineLogicalScalarType
    cast_inst = bool
    test_file = 'logical_retval_test'

    @classmethod
    def set_up_fun(cls, compiler):
        """
        The abstract test setup.
        """
        cls.default_engine_state = save_engine_state()

        cls.engine_state = b_utils._EngineState

        if compiler == 'ifort':
            cls.engine_state._compilers = {'Fortran': 'Intel'}
        else:
            cls.engine_state._compilers = {'Fortran': 'NAG'}

        cls.set_up(cls.test_file, compiler)

        cls.engine_state._chrlen_ctype = cls.compiler.chrlen_ctype

        cls.engine_fun.restype = cls.en_scalar_type.c_type
        cls.engine_fun.argtypes = (
            ctypes.POINTER(cls.en_scalar_type.c_type),
        )

    def test_good_direct(self):
        """
        Test correct operation with a direct call.
        """
        engine_l = self.engine_fun(
            self.en_scalar_type.c_type(0),
        )
        engine_l_py = self.en_scalar_type.to_py_from_retval(engine_l)
        self.assertIsInstance(engine_l_py, self.cast_inst) # pylint: disable=no-member
        self.assertTrue(engine_l_py) # pylint: disable=no-member

    def test_good_wrapper(self):
        """
        Test correct operation via a Python wrapper.
        """

        # pylint: disable=invalid-name
        def py_fun(l):
            """
            Python wrapper for self.engine_fun.
            """
            _l = self.en_scalar_type( # pylint: disable=not-callable
                l,
                {'fun_name': 'py_fun', 'entity': 'l'}
            ).to_ctypes()
            _retval = self.engine_fun(
                _l,
            )
            return self.en_scalar_type.to_py_from_retval(_retval)

        l = py_fun(
            False,
        )
        self.assertTrue( # pylint: disable=no-member
            l,
        )

class PassTestLogicalRetvalNag(PassTestLogicalRetval, unittest.TestCase):
    """
    Pass-test subsuite for NAG.
    """
    compiler = 'nagfor'

    @classmethod
    def setUpClass(cls):
        cls.set_up_fun(cls.compiler)

class PassTestLogicalRetvalIntel(PassTestLogicalRetval, unittest.TestCase):
    """
    Pass-test subsuite for Intel.
    """
    compiler = 'ifort'

    @classmethod
    def setUpClass(cls):
        cls.set_up_fun(cls.compiler)

class PassTestEnData(PassTest):
    """
    Abstract low-level unit tests for passing engine_data to the Engine.
    """

    file_name = 'en_data_pass_test'

    @classmethod
    def set_up_ed(cls, file_name, compiler):
        """
        The abstract test setup.
        """
        cls.default_engine_state = save_engine_state()

        cls.engine_state = b_utils._EngineState

        cls.set_up(file_name, compiler, deps=True)

        cls.cb = b_utils._EngineState._sys_functype(
            None,
            b_utils._BYREF_ENGINEDATA,
        )

        cls.engine_fun.restype = b_utils._EngineIntScalarType.c_type
        cls.engine_fun.argtypes = (
            b_utils._BYREF_ENGINEDATA,
            ctypes.POINTER(ctypes.c_int), # Input, i.
            b_utils._BYREF_LOGICAL, # Input, have_print.
            cls.cb,
        )

    @staticmethod
    def cb_s_empty(en_data):
        """
        No-op callback.
        """

    def test_good_default(self):
        """
        Test correct operation with a default-initialized call.
        """
        _en_data = b_utils._EngineData()
        _have_print = b_utils._EngineLogicalScalarType(
            False,
            {'fun_name': 'z99zz', 'entity': 'have_print'}
        ).to_ctypes()
        self.engine_fun(
            _en_data,
            b_utils._NAG_ED_ROW_MAJOR,
            _have_print,
            self.cb(self.cb_s_empty),
        )

    def test_good_print(self):
        """
        Test correct operation with a default-initialized call and print_rec
        supplied.
        """

        def cb_py_c_print_rec(en_data, nout, rec, length_rec, ifail):
            # pylint: disable=missing-docstring
            pass # pragma: no cover

        self.engine_state._set_debug_level(self.engine_state._debug_args_short)
        _en_data = b_utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
        _have_print = b_utils._EngineLogicalScalarType(
            True,
            {'fun_name': 'z99zz', 'entity': 'have_print'}
        ).to_ctypes()
        self.engine_fun(
            _en_data,
            b_utils._NAG_ED_ROW_MAJOR,
            _have_print,
            self.cb(self.cb_s_empty),
        )
        self.engine_state._set_debug_level(self.engine_state._debug_off)

    def test_good_col(self):
        """
        Test correct operation with a column-major call.
        """
        _en_data = b_utils._EngineData(sorder='F')
        _have_print = b_utils._EngineLogicalScalarType(
            False,
            {'fun_name': 'z99zz', 'entity': 'have_print'}
        ).to_ctypes()
        self.engine_fun(
            _en_data,
            b_utils._NAG_ED_COL_MAJOR,
            _have_print,
            self.cb(self.cb_s_empty),
        )

    def test_good_row(self):
        """
        Test correct operation with a row-major call.
        """
        _en_data = b_utils._EngineData(sorder='C')
        _have_print = b_utils._EngineLogicalScalarType(
            False,
            {'fun_name': 'z99zz', 'entity': 'have_print'}
        ).to_ctypes()
        self.engine_fun(
            _en_data,
            b_utils._NAG_ED_ROW_MAJOR,
            _have_print,
            self.cb(self.cb_s_empty),
        )

    @staticmethod
    def cb_py(data):
        """
        Callback.
        """
        data['y'], data['x'] = data['x'], data['y']

    def test_good_data(self):
        """
        Test correct operation of user data.
        """

        def cb_s(en_data):
            """
            Helper.
            """
            self.cb_py(data)

        data = {'x': 42, 'y': 43}

        _en_data = b_utils._EngineData()
        _have_print = b_utils._EngineLogicalScalarType(
            False,
            {'fun_name': 'z99zz', 'entity': 'have_print'}
        ).to_ctypes()
        self.engine_fun(
            _en_data,
            b_utils._NAG_ED_ROW_MAJOR,
            _have_print,
            self.cb(cb_s),
        )
        self.assertEqual( # pylint: disable=no-member
            data, {'x': 43, 'y':42}
        )

    def test_good_data_none(self):
        """
        Test correct operation of user data.
        """

        def cb_py():
            """
            Callback.
            """

        def cb_s(en_data):
            """
            Helper.
            """
            if data is not None: # pragma: no cover
                cb_py(data) # pylint: disable=too-many-function-args
            else:
                cb_py()

        data = None

        _en_data = b_utils._EngineData()
        _have_print = b_utils._EngineLogicalScalarType(
            False,
            {'fun_name': 'z99zz', 'entity': 'have_print'}
        ).to_ctypes()
        self.engine_fun(
            _en_data,
            b_utils._NAG_ED_ROW_MAJOR,
            _have_print,
            self.cb(cb_s),
        )

class PassTestEnDataNag(PassTestEnData, unittest.TestCase):
    """
    Pass-test subsuite for NAG.
    """
    compiler = 'nagfor'

    @classmethod
    def setUpClass(cls):
        cls.set_up_ed(cls.file_name, cls.compiler)

class PassTestEnDataIntel(PassTestEnData, unittest.TestCase):
    """
    Pass-test subsuite for Intel.
    """
    compiler = 'ifort'

    @classmethod
    def setUpClass(cls):
        cls.set_up_ed(cls.file_name, cls.compiler)

class PassTestScalar(PassTest):
    """
    Abstract low-level unit tests for the passing scalar data to the Engine.
    """

    @classmethod
    def set_up_scalar(cls, compiler):
        """
        The abstract test setup.
        """
        cls.default_engine_state = save_engine_state()

        cls.engine_state = b_utils._EngineState

        if compiler == 'ifort':
            cls.engine_state._compilers = {'Fortran': 'Intel'}
        else:
            cls.engine_state._compilers = {'Fortran': 'NAG'}

        cls.set_up('scalar_pass_test', compiler)

        cls.engine_state._chrlen_ctype = cls.compiler.chrlen_ctype

        the_args = (
            b_utils._BYREF_STRING, # cstar.
            b_utils._BYREF_STRING, # c2.
            b_utils._BYREF_COMPLEX64, # c.
            b_utils._BYREF_COMPLEX128, # z.
            b_utils._BYREF_HANDLE, # Input/Output, cptr.
            b_utils._BYREF_FLOAT32, # r.
            b_utils._BYREF_FLOAT64, # x.
            b_utils._BYREF_INT, # i.
            b_utils._BYREF_LOGICAL, # l.
            b_utils._BYREF_STRING, # cstar_to_user.
            b_utils._BYREF_STRING, # cN_to_user.
            b_utils._BYREF_COMPLEX64, # c_to_user.
            b_utils._BYREF_COMPLEX128, # z_to_user.
            b_utils._BYREF_FLOAT32, # r_to_user.
            b_utils._BYREF_FLOAT64, # x_to_user.
            b_utils._BYREF_INT, # i_to_user.
            b_utils._BYREF_LOGICAL, # l_to_user.
        )

        cb_args = list(the_args)
        cb_args.extend([
            cls.engine_state._chrlen_ctype, # Input, chrlen of cstar.
            cls.engine_state._chrlen_ctype, # Input, chrlen of c2.
            cls.engine_state._chrlen_ctype, # Input, chrlen of cstar_to_user.
            cls.engine_state._chrlen_ctype, # Input, chrlen of c5_to_user.
        ])

        cls.cb = b_utils._EngineState._sys_functype(
            None,
            *tuple(cb_args)
        )

        cls.engine_fun.restype = b_utils._EngineIntScalarType.c_type
        ef_args = list(the_args)
        ef_args.extend([
            cls.cb,
            cls.engine_state._chrlen_ctype, # Input, chrlen of cstar.
            cls.engine_state._chrlen_ctype, # Input, chrlen of c2.
            cls.engine_state._chrlen_ctype, # Input, chrlen of cstar_to_user.
            cls.engine_state._chrlen_ctype, # Input, chrlen of c4_to_user.
        ])
        cls.engine_fun.argtypes = tuple(ef_args)

    def test_good_direct(self): # pylint: disable=too-many-locals,too-many-statements
        """
        Test correct operation with a direct call.
        """

        def cb_h(cstar, c2, c, z, cptr, r, x, i, l, # pylint: disable=invalid-name,too-many-arguments,too-many-locals
                 cstar_to_user, c5_to_user, c_to_user, z_to_user,
                 r_to_user, x_to_user, i_to_user, l_to_user,
                 cstar_len, c2_len, cstar_to_user_len, c5_to_user_len):
            """
            Callback for testing.
            """
            _cstar_to_user = b_utils._EngineCharScalarType.to_py(
                cstar_to_user, chrlen=cstar_to_user_len
            )
            self.assertEqual( # pylint: disable=no-member
                _cstar_to_user,
                'Hell',
            )
            _c5_to_user = b_utils._EngineCharScalarType.to_py(
                c5_to_user, chrlen=c5_to_user_len
            )
            self.assertEqual( # pylint: disable=no-member
                _c5_to_user,
                'World',
            )
            _c_to_user = b_utils._EngineComplex64ScalarType.to_py(
                c_to_user.contents
            )
            self.assertEqual( # pylint: disable=no-member
                _c_to_user,
                44. + 45.j,
            )
            _z_to_user = b_utils._EngineComplex128ScalarType.to_py(
                z_to_user.contents
            )
            self.assertEqual( # pylint: disable=no-member
                _z_to_user,
                46. + 47.j,
            )
            _cptr_to_user = b_utils._EngineCptrScalarType.to_py(cptr.contents)
            self.assertEqual( # pylint: disable=no-member
                _cptr_to_user.value,
                43,
            )
            _r_to_user = b_utils._EngineFloat32ScalarType.to_py(
                r_to_user.contents
            )
            self.assertEqual( # pylint: disable=no-member
                _r_to_user,
                43.
            )
            _x_to_user = b_utils._EngineFloat64ScalarType.to_py(
                x_to_user.contents
            )
            self.assertEqual( # pylint: disable=no-member
                _x_to_user,
                43.
            )
            _i_to_user = b_utils._EngineIntScalarType.to_py(i_to_user.contents)
            self.assertEqual( # pylint: disable=no-member
                _i_to_user,
                44,
            )
            _l_to_user = b_utils._EngineLogicalScalarType.to_py(
                l_to_user.contents
            )
            self.assertTrue( # pylint: disable=no-member
                _l_to_user
            )
            b_utils._EngineCharScalarType(
                b'420',
                {'fun_name': 'cb', 'returned': 'cstar'},
                cstar_len,
            ).refresh_ctypes_buffer(cstar)
            b_utils._EngineCharScalarType(
                b'42',
                {'fun_name': 'cb', 'returned': 'c2'},
                c2_len,
            ).refresh_ctypes_buffer(c2)
            b_utils._EngineComplex64ScalarType(
                np.complex64(42. + 43.j),
                {'fun_name': 'cb', 'returned': 'c'},
            ).refresh_ctypes_buffer(c)
            b_utils._EngineComplex128ScalarType(
                42. + 43.j,
                {'fun_name': 'cb', 'returned': 'z'},
            ).refresh_ctypes_buffer(z)
            b_utils._EngineCptrScalarType(
                b_utils.Handle(42),
                {'fun_name': 'cb', 'returned': 'cptr'},
            ).refresh_ctypes_buffer(cptr)
            b_utils._EngineFloat32ScalarType(
                ctypes.c_float(42.),
                {'fun_name': 'cb', 'returned': 'r'},
            ).refresh_ctypes_buffer(r)
            b_utils._EngineFloat64ScalarType(
                42.,
                {'fun_name': 'cb', 'returned': 'x'},
            ).refresh_ctypes_buffer(x)
            b_utils._EngineIntScalarType(
                42,
                {'fun_name': 'cb', 'returned': 'i'},
            ).refresh_ctypes_buffer(i)
            b_utils._EngineLogicalScalarType(
                False,
                {'fun_name': 'cb', 'returned': 'l'},
            ).refresh_ctypes_buffer(l)

        _cptr = b_utils._EngineCptrScalarType(
            b_utils.Handle(43),
            {'fun_name': 'z99zz', 'entity': 'cptr'},
        ).to_ctypes()
        _l = b_utils._EngineLogicalScalarType(
            False,
            {'fun_name': 'z99zz', 'entity': 'l'}
        ).to_ctypes()
        _cstar_to_user = b_utils._EngineCharScalarType.empty_ctypes(chrlen=6)
        _c4_to_user = b_utils._EngineCharScalarType.empty_ctypes(chrlen=4)
        _c_to_user = b_utils._EngineComplex64ScalarType.empty_ctypes()
        _z_to_user = b_utils._EngineComplex128ScalarType.empty_ctypes()
        _r_to_user = b_utils._EngineFloat32ScalarType.empty_ctypes()
        _x_to_user = b_utils._EngineFloat64ScalarType.empty_ctypes()
        _i_to_user = b_utils._EngineIntScalarType.empty_ctypes()
        _l_to_user = b_utils._EngineLogicalScalarType.empty_ctypes()
        _retval = self.engine_fun(
            ctypes.create_string_buffer(b'420', size=3),
            ctypes.create_string_buffer(b'420', size=3),
            b_utils._EngineComplex64ScalarType.c_type(42.0, 43.0),
            b_utils._EngineComplex128ScalarType.c_type(42.0, 43.0),
            _cptr,
            b_utils._EngineFloat32ScalarType.c_type(42.0),
            b_utils._EngineFloat64ScalarType.c_type(42.0),
            b_utils._EngineIntScalarType.c_type(42),
            _l,
            _cstar_to_user,
            _c4_to_user,
            _c_to_user,
            _z_to_user,
            _r_to_user,
            _x_to_user,
            _i_to_user,
            _l_to_user,
            self.cb(cb_h),
            3,
            2,
            6,
            4,
        )
        self.assertEqual( # pylint: disable=no-member
            b_utils._EngineIntScalarType.to_py_from_retval(_retval),
            0,
        )
        _cstar_to_user_py = b_utils._EngineCharScalarType.to_py(_cstar_to_user)
        self.assertEqual( # pylint: disable=no-member
            _cstar_to_user_py,
            'Hello!',
        )
        _c4_to_user_py = b_utils._EngineCharScalarType.to_py(_c4_to_user)
        self.assertEqual( # pylint: disable=no-member
            _c4_to_user_py,
            '.rld',
        )
        _c_to_user_py = b_utils._EngineComplex64ScalarType.to_py(_c_to_user)
        self.assertEqual( # pylint: disable=no-member
            _c_to_user_py,
            46. + 47.j,
        )
        _z_to_user_py = b_utils._EngineComplex128ScalarType.to_py(_z_to_user)
        self.assertEqual( # pylint: disable=no-member
            _z_to_user_py,
            44. + 45.j,
        )
        _cptr_py = b_utils._EngineCptrScalarType.to_py(_cptr)
        self.assertEqual( # pylint: disable=no-member
            _cptr_py.value,
            42,
        )
        _r_to_user_py = b_utils._EngineFloat32ScalarType.to_py(_r_to_user)
        self.assertEqual( # pylint: disable=no-member
            _r_to_user_py,
            44.
        )
        _x_to_user_py = b_utils._EngineFloat64ScalarType.to_py(_x_to_user)
        self.assertEqual( # pylint: disable=no-member
            _x_to_user_py,
            44.
        )
        _i_to_user_py = b_utils._EngineIntScalarType.to_py(_i_to_user)
        self.assertEqual( # pylint: disable=no-member
            _i_to_user_py,
            45,
        )
        _l_to_user_py = b_utils._EngineLogicalScalarType.to_py(_l_to_user)
        self.assertTrue( # pylint: disable=no-member
            _l_to_user_py
        )

    def test_good_wrapper(self): # pylint: disable=too-many-locals,too-many-statements
        """
        Test correct operation via a Python wrapper.
        """

        def cb_py(cstar_to_user, c5_to_user, c_to_user, z_to_user, cptr, # pylint: disable=too-many-arguments
                  r_to_user, x_to_user, i_to_user, l_to_user):
            """
            Test callback.
            """
            self.assertEqual( # pylint: disable=no-member
                cstar_to_user,
                'Hell',
            )
            self.assertEqual( # pylint: disable=no-member
                c5_to_user,
                'World',
            )
            self.assertEqual( # pylint: disable=no-member
                c_to_user,
                44. + 45.j,
            )
            self.assertEqual( # pylint: disable=no-member
                z_to_user,
                46. + 47.j,
            )
            self.assertEqual( # pylint: disable=no-member
                cptr.value,
                43,
            )
            cptr.value = 42
            self.assertEqual( # pylint: disable=no-member
                r_to_user,
                43.,
            )
            self.assertEqual( # pylint: disable=no-member
                x_to_user,
                43.,
            )
            self.assertEqual( # pylint: disable=no-member
                i_to_user,
                44,
            )
            self.assertTrue( # pylint: disable=no-member
                l_to_user
            )
            return (
                '420',
                b'42',
                np.complex64(42 + 43j),
                42 + 43j,
                np.float32(42.),
                42.,
                42,
                False,
            )

        # pylint: disable=invalid-name,too-many-arguments,too-many-locals
        def py_fun(cstar, c2, c, z, h, r, x, i, l, cb):
            """
            Python wrapper for self.engine_fun.
            """

            def cb_h(cstar, c2, c, z, cptr, r, x, i, l,
                     cstar_to_user, c5_to_user, c_to_user, z_to_user,
                     r_to_user, x_to_user, i_to_user, l_to_user,
                     cstar_len, c2_len, cstar_to_user_len, c5_to_user_len):
                """
                Test helper.
                """
                _cstar_to_user = b_utils._EngineCharScalarType.to_py(
                    cstar_to_user, chrlen=cstar_to_user_len
                )
                _c5_to_user = b_utils._EngineCharScalarType.to_py(
                    c5_to_user, chrlen=c5_to_user_len
                )
                _c_to_user = b_utils._EngineComplex64ScalarType.to_py(
                    c_to_user.contents
                )
                _z_to_user = b_utils._EngineComplex128ScalarType.to_py(
                    z_to_user.contents
                )
                _cptr = b_utils._EngineCptrScalarType.to_py(cptr.contents)
                _r_to_user = b_utils._EngineFloat32ScalarType.to_py(
                    r_to_user.contents
                )
                _x_to_user = b_utils._EngineFloat64ScalarType.to_py(
                    x_to_user.contents
                )
                _i_to_user = b_utils._EngineIntScalarType.to_py(
                    i_to_user.contents
                )
                _l_to_user = b_utils._EngineLogicalScalarType.to_py(
                    l_to_user.contents
                )
                _cstar, _c2, _c, _z, _r, _x, _i, _l = cb(
                    _cstar_to_user, _c5_to_user, _c_to_user, _z_to_user, _cptr,
                    _r_to_user, _x_to_user, _i_to_user, _l_to_user
                )
                b_utils._EngineCharScalarType(
                    _cstar,
                    {'fun_name': 'py_fun:cb_h', 'returned': 'cstar'},
                    cstar_len,
                ).refresh_ctypes_buffer(cstar)
                b_utils._EngineCharScalarType(
                    _c2,
                    {'fun_name': 'py_fun:cb_h', 'returned': 'c2'},
                    c2_len,
                ).refresh_ctypes_buffer(c2)
                b_utils._EngineComplex64ScalarType(
                    _c,
                    {'fun_name': 'py_fun:cb_h', 'returned': 'c'},
                ).refresh_ctypes_buffer(c)
                b_utils._EngineComplex128ScalarType(
                    _z,
                    {'fun_name': 'py_fun:cb_h', 'returned': 'z'},
                ).refresh_ctypes_buffer(z)
                b_utils._EngineCptrScalarType(
                    _cptr,
                    {'fun_name': 'py_fun:cb_h', 'returned': 'cptr'}
                ).refresh_ctypes_buffer(cptr)
                b_utils._EngineFloat32ScalarType(
                    _r,
                    {'fun_name': 'py_fun:cb_h', 'returned': 'r'}
                ).refresh_ctypes_buffer(r)
                b_utils._EngineFloat64ScalarType(
                    _x,
                    {'fun_name': 'py_fun:cb_h', 'returned': 'x'}
                ).refresh_ctypes_buffer(x)
                b_utils._EngineIntScalarType(
                    _i,
                    {'fun_name': 'py_fun:cb_h', 'returned': 'i'}
                ).refresh_ctypes_buffer(i)
                b_utils._EngineLogicalScalarType(
                    _l,
                    {'fun_name': 'py_fun:cb_h', 'returned': 'l'}
                ).refresh_ctypes_buffer(l)

            _cstar = b_utils._EngineCharScalarType(
                cstar,
                {'fun_name': 'py_fun', 'entity': 'cstar'},
                None,
            ).to_ctypes()
            _c2 = b_utils._EngineCharScalarType(
                c2,
                {'fun_name': 'py_fun', 'entity': 'c2'},
                2,
            ).to_ctypes()
            _c = b_utils._EngineComplex64ScalarType(
                c,
                {'fun_name': 'py_fun', 'entity': 'c'}
            ).to_ctypes()
            _z = b_utils._EngineComplex128ScalarType(
                z,
                {'fun_name': 'py_fun', 'entity': 'z'}
            ).to_ctypes()
            _h = b_utils._EngineCptrScalarType(
                h,
                {'fun_name': 'py_fun', 'entity': 'h'}
            ).to_ctypes()
            _r = b_utils._EngineFloat32ScalarType(
                r,
                {'fun_name': 'py_fun', 'entity': 'r'}
            ).to_ctypes()
            _x = b_utils._EngineFloat64ScalarType(
                x,
                {'fun_name': 'py_fun', 'entity': 'x'}
            ).to_ctypes()
            _i = b_utils._EngineIntScalarType(
                i,
                {'fun_name': 'py_fun', 'entity': 'i'}
            ).to_ctypes()
            _l = b_utils._EngineLogicalScalarType(
                l,
                {'fun_name': 'py_fun', 'entity': 'l'}
            ).to_ctypes()
            _cstar_to_user = b_utils._EngineCharScalarType.empty_ctypes(
                chrlen=6
            )
            _c4_to_user = b_utils._EngineCharScalarType.empty_ctypes(chrlen=4)
            _c_to_user = b_utils._EngineComplex64ScalarType.empty_ctypes()
            _z_to_user = b_utils._EngineComplex128ScalarType.empty_ctypes()
            _r_to_user = b_utils._EngineFloat32ScalarType.empty_ctypes()
            _x_to_user = b_utils._EngineFloat64ScalarType.empty_ctypes()
            _i_to_user = b_utils._EngineIntScalarType.empty_ctypes()
            _l_to_user = b_utils._EngineLogicalScalarType.empty_ctypes()
            _cb = self.cb(cb_h)
            _retval = self.engine_fun(
                _cstar,
                _c2,
                _c,
                _z,
                _h,
                _r,
                _x,
                _i,
                _l,
                _cstar_to_user,
                _c4_to_user,
                _c_to_user,
                _z_to_user,
                _r_to_user,
                _x_to_user,
                _i_to_user,
                _l_to_user,
                _cb,
                len(cstar),
                2,
                6,
                4,
            )
            _h = b_utils._EngineCptrScalarType.to_py(_h)
            return (
                b_utils._EngineIntScalarType.to_py_from_retval(_retval),
                b_utils._EngineCharScalarType.to_py(_cstar_to_user),
                b_utils._EngineCharScalarType.to_py(_c4_to_user),
                b_utils._EngineComplex64ScalarType.to_py(_c_to_user),
                b_utils._EngineComplex128ScalarType.to_py(_z_to_user),
                b_utils._EngineFloat32ScalarType.to_py(_r_to_user),
                b_utils._EngineFloat64ScalarType.to_py(_x_to_user),
                b_utils._EngineIntScalarType.to_py(_i_to_user),
                b_utils._EngineLogicalScalarType.to_py(_l_to_user),
            )

        cptr = b_utils.Handle(43)
        (
            retval, _cstar_to_user, _c4_to_user, _c_to_user, _z_to_user,
            _r_to_user, _x_to_user, _i_to_user, _l_to_user
        ) = py_fun(
            '420',
            '42',
            np.complex64(42.0 + 43.0j),
            42.0 + 43.0j,
            cptr,
            np.float32(42.0),
            42.0,
            42,
            False,
            cb_py,
        )
        self.assertEqual( # pylint: disable=no-member
            retval,
            0,
        )
        self.assertEqual( # pylint: disable=no-member
            _cstar_to_user,
            'Hello!',
        )
        self.assertEqual( # pylint: disable=no-member
            _c4_to_user,
            '.rld',
        )
        self.assertEqual( # pylint: disable=no-member
            _c_to_user,
            46. + 47.j,
        )
        self.assertEqual( # pylint: disable=no-member
            _z_to_user,
            44. + 45.j,
        )
        self.assertEqual( # pylint: disable=no-member
            cptr.value,
            42,
        )
        self.assertEqual( # pylint: disable=no-member
            _r_to_user,
            44.,
        )
        self.assertEqual( # pylint: disable=no-member
            _x_to_user,
            44.,
        )
        self.assertEqual( # pylint: disable=no-member
            _i_to_user,
            45,
        )
        self.assertTrue( # pylint: disable=no-member
            _l_to_user,
        )

class PassTestScalarIntel(PassTestScalar, unittest.TestCase):
    """
    Pass-test subsuite for Intel.
    """
    compiler = 'ifort'

    @classmethod
    def setUpClass(cls):
        cls.set_up_scalar(cls.compiler)

class PassTestScalarNag(PassTestScalar, unittest.TestCase):
    """
    Pass-test subsuite for NAG.
    """
    compiler = 'nagfor'

    @classmethod
    def setUpClass(cls):
        cls.set_up_scalar(cls.compiler)

class PassTestArray1Dim(PassTest): # pylint: disable=too-many-instance-attributes
    """
    Abstract Low-level unit tests for the passing array data to the Engine.
    """

    @classmethod
    def set_up_1dim(cls, compiler):
        """
        The abstract test setup.
        """

        # pylint: disable=line-too-long

        cls.default_engine_state = save_engine_state()

        cls.engine_state = b_utils._EngineState

        if compiler == 'ifort':
            cls.engine_state._compilers = {'Fortran': 'Intel'}
            cls.engine_state._literal_true_int = -1
        else:
            cls.engine_state._compilers = {'Fortran': 'NAG'}
            cls.engine_state._literal_true_int = 1

        cls.set_up('array_pass_test', compiler)

        cls.engine_state._chrlen_ctype = cls.compiler.chrlen_ctype

        the_args = (
            b_utils._EngineCharArrayType.c_type, # Input, cstar(len=*)[*].
            b_utils._EngineCharArrayType.c_type, # Input, c2(len=2)[2].
            b_utils._EngineComplex64ArrayType.c_type, # Input, complex64 c[2].
            b_utils._EngineComplex128ArrayType.c_type, # Input, complex128 z[3].
            b_utils._EngineCptrArrayType.c_type, # Input, c_ptr cptr[3].
            b_utils._EngineFloat32ArrayType.c_type, # Input, r[2].
            b_utils._EngineFloat64ArrayType.c_type, # Input, x[3].
            b_utils._EngineIntArrayType.c_type, # Input, i[2].
            b_utils._EngineLogicalArrayType.c_type, # Input, logical l[2].
            b_utils._EngineCharArrayType.c_type, # Output, cstar_to_user(len=*)[*].
            b_utils._EngineCharArrayType.c_type, # Output, c4_to_user(len=4)[4].
            b_utils._EngineComplex64ArrayType.c_type, # Output, complex64 c_to_user[2].
            b_utils._EngineComplex128ArrayType.c_type, # Output, complex128 z_to_user[3].
            b_utils._EngineFloat32ArrayType.c_type, # Output, r_to_user[2].
            b_utils._EngineFloat64ArrayType.c_type, # Output, x_to_user[3].
            b_utils._EngineIntArrayType.c_type, # Output, i_to_user[2].
            b_utils._EngineLogicalArrayType.c_type, # Output, logical l_to_user[2].
        )

        cb_args = list(the_args)
        del cb_args[4] # No c_ptr array in the callback.
        cb_args.extend([
            cls.engine_state._chrlen_ctype, # Input, chrlen of cstar.
            cls.engine_state._chrlen_ctype, # Input, chrlen of c2.
            cls.engine_state._chrlen_ctype, # Input, chrlen of cstar_to_user.
            cls.engine_state._chrlen_ctype, # Input, chrlen of c4_to_user.
        ])

        cls.cb = b_utils._EngineState._sys_functype(
            None,
            *tuple(cb_args)
        )

        cls.engine_fun.restype = b_utils._EngineIntScalarType.c_type
        ef_args = list(the_args)
        ef_args.extend([
            cls.cb,
            cls.engine_state._chrlen_ctype, # Input, chrlen of cstar.
            cls.engine_state._chrlen_ctype, # Input, chrlen of c2.
            cls.engine_state._chrlen_ctype, # Input, chrlen of cstar_to_user.
            cls.engine_state._chrlen_ctype, # Input, chrlen of c4_to_user.
        ])
        cls.engine_fun.argtypes = tuple(ef_args)

    def set_up_direct_data(
            self,
    ):
        """
        The default 'direct' call output (initialized) data set.
        """

        # pylint: disable=attribute-defined-outside-init

        self.cstar_to_user = ['......', '......']
        self.c4_to_user = ['', '', '', '']
        self.c_to_user = [np.complex64(42.), np.complex64(42.)]
        self.z_to_user = [complex(42.), complex(42.), complex(42.)]
        self.r_to_user = [np.float32(42.), np.float32(42.)]
        self.x_to_user = [42., 42., 42.]
        self.i_to_user = [1, 1]
        self.l_to_user = [False, False]

    def set_up_wrapper_data(
            self,
    ):
        """
        The default 'wrapper' call output (uninitialized) data set.
        """

        # pylint: disable=attribute-defined-outside-init

        self.cstar_to_user = b_utils._EngineCharArrayType.empty_ndarray(
            (2,), chrlen=6
        )
        self.c4_to_user = b_utils._EngineCharArrayType.empty_ndarray(
            (4,), chrlen=4
        )
        self.c_to_user = b_utils._EngineComplex64ArrayType.empty_ndarray((2,))
        self.z_to_user = b_utils._EngineComplex128ArrayType.empty_ndarray((3,))
        self.r_to_user = b_utils._EngineFloat32ArrayType.empty_ndarray((2,))
        self.x_to_user = b_utils._EngineFloat64ArrayType.empty_ndarray((3,))
        self.i_to_user = b_utils._EngineIntArrayType.empty_ndarray((2,))
        self.l_to_user = b_utils._EngineLogicalArrayType.empty_ndarray((2,))

    def cast_output_data(
            self,
    ):
        """
        Cast the output data.
        """

        # pylint: disable=attribute-defined-outside-init

        self._cstar_to_user_cls = b_utils._EngineCharArrayType(
            self.cstar_to_user,
            {'fun_name': 'cast_output_data', 'entity': 'cstar_to_user'},
            exp_chrlen=None, exp_shape=(2,), check_contents=False,
        )
        self._cstar_to_user = self._cstar_to_user_cls.to_ctypes()
        self._c4_to_user_cls = b_utils._EngineCharArrayType(
            self.c4_to_user,
            {'fun_name': 'cast_output_data', 'entity': 'c4_to_user'},
            exp_chrlen=4, exp_shape=(4,), check_contents=False,
        )
        self._c4_to_user = self._c4_to_user_cls.to_ctypes()
        self._c_to_user_cls = b_utils._EngineComplex64ArrayType(
            self.c_to_user,
            {'fun_name': 'cast_output_data', 'entity': 'c_to_user'},
            exp_shape=(2,), check_contents=False,
        )
        self._c_to_user = self._c_to_user_cls.to_ctypes()
        self._z_to_user_cls = b_utils._EngineComplex128ArrayType(
            self.z_to_user,
            {'fun_name': 'cast_output_data', 'entity': 'z_to_user'},
            exp_shape=(3,), check_contents=False,
        )
        self._z_to_user = self._z_to_user_cls.to_ctypes()
        self._r_to_user_cls = b_utils._EngineFloat32ArrayType(
            self.r_to_user,
            {'fun_name': 'cast_output_data', 'entity': 'r_to_user'},
            exp_shape=(2,), check_contents=False,
        )
        self._r_to_user = self._r_to_user_cls.to_ctypes()
        self._x_to_user_cls = b_utils._EngineFloat64ArrayType(
            self.x_to_user,
            {'fun_name': 'cast_output_data', 'entity': 'x_to_user'},
            exp_shape=(3,), check_contents=False,
        )
        self._x_to_user = self._x_to_user_cls.to_ctypes()
        self._i_to_user_cls = b_utils._EngineIntArrayType(
            self.i_to_user,
            {'fun_name': 'cast_output_data', 'entity': 'i_to_user'},
            exp_shape=(2,), check_contents=False,
        )
        self._i_to_user = self._i_to_user_cls.to_ctypes()
        self._l_to_user_cls = b_utils._EngineLogicalArrayType(
            self.l_to_user,
            {'fun_name': 'cast_output_data', 'entity': 'l_to_user'},
            exp_shape=(2,), check_contents=False,
        )
        self._l_to_user = self._l_to_user_cls.to_ctypes()

    def direct_call(
            self,
    ):
        """
        Make the 'direct' call.
        """

        # pylint: disable=attribute-defined-outside-init

        self._retval = self.engine_fun(
            ctypes.create_string_buffer(b'Hello World!', size=12),
            ctypes.create_string_buffer(b'HeWo', size=4),
            (b_utils._EngineComplex64ArrayType.c_type._type_*2)(
                *[(42, 43), (44, 45)]
            ),
            (b_utils._EngineComplex128ArrayType.c_type._type_*3)(
                *[(42, 43), (44, 45), (46, 47)]
            ),
            (b_utils._EngineCptrArrayType.c_type._type_*3)(
                *[b_utils.Handle(42), b_utils.Handle(43), b_utils.Handle(44)]
            ),
            (b_utils._EngineFloat32ArrayType.c_type._type_*2)(*[42, 43]),
            (b_utils._EngineFloat64ArrayType.c_type._type_*3)(*[42, 43, 44]),
            (b_utils._EngineIntArrayType.c_type._type_*2)(*[42, 43]),
            (b_utils._EngineLogicalArrayType.c_type._type_*2)(*[1, 0]),
            self._cstar_to_user,
            self._c4_to_user,
            self._c_to_user,
            self._z_to_user,
            self._r_to_user,
            self._x_to_user,
            self._i_to_user,
            self._l_to_user,
            self.cb(self.cb_h),
            6,
            2,
            self._cstar_to_user_cls.chrlen,
            self._c4_to_user_cls.chrlen,
        )

    def refresh_output_data(
            self,
    ):
        """
        Poke the Engine output data back into the Python equivalents.
        """

        # pylint: disable=attribute-defined-outside-init

        self.retval = b_utils._EngineIntScalarType.to_py_from_retval(
            self._retval
        )
        self._cstar_to_user_cls.refresh_py(self._cstar_to_user)
        self._c4_to_user_cls.refresh_py(self._c4_to_user)
        self._c_to_user_cls.refresh_py(self._c_to_user)
        self._z_to_user_cls.refresh_py(self._z_to_user)
        self._r_to_user_cls.refresh_py(self._r_to_user)
        self._x_to_user_cls.refresh_py(self._x_to_user)
        self._i_to_user_cls.refresh_py(self._i_to_user)
        self._l_to_user_cls.refresh_py(self._l_to_user)

    def check_output_data(
            self,
    ):
        """
        Check the output data.
        """
        self.check_py(
            self.retval,
            self.cstar_to_user, self.c4_to_user, self.c_to_user,
            self.z_to_user, self.r_to_user, self.x_to_user, self.i_to_user,
            self.l_to_user,
        )

    def make_and_check_direct_call(
            self,
    ):
        """
        Set up, invoke and check the 'direct' call.
        """
        self.set_up_direct_data()
        self.cast_output_data()
        self.direct_call()
        self.refresh_output_data()
        self.check_output_data()

    def check_cb_input( # pylint: disable=invalid-name,too-many-arguments
            self,
            cstar, c4, c, z, r, x, i, l,
    ):
        """
        Check input data passed in to the callback.
        """
        self.assertEqual( # pylint: disable=no-member
            list(cstar),
            [b'abcdefg', b'hijklmn'],
        )
        self.assertEqual( # pylint: disable=no-member
            list(c4),
            [b'ZYXW', b'VUTS', b'RQPO'],
        )
        self.assertEqual( # pylint: disable=no-member
            list(c),
            [50. + 51.j, 52. + 53.j],
        )
        self.assertEqual( # pylint: disable=no-member
            list(z),
            [54. + 55.j, 56. + 57.j, 58. + 59.j],
        )
        self.assertEqual( # pylint: disable=no-member
            list(r),
            [42., 43.],
        )
        self.assertEqual( # pylint: disable=no-member
            list(x),
            [46., 47., 48.],
        )
        self.assertEqual( # pylint: disable=no-member
            list(i),
            [44, 45],
        )
        self.assertEqual( # pylint: disable=no-member
            list(l),
            [True, False],
        )

    def cast_cb_input( # pylint: disable=too-many-arguments
            self,
            cstar_to_user, c4_to_user, c_to_user, z_to_user,
            r_to_user, x_to_user, i_to_user, l_to_user,
            cstar_to_user_len, c4_to_user_len
    ):
        """
        Cast the input callback data to Python.
        """

        # pylint: disable=attribute-defined-outside-init

        self._cb_cstar_to_user = b_utils._EngineCharArrayType.from_pointer(
            cstar_to_user, (2,), chrlen=cstar_to_user_len
        )
        self._cb_c4_to_user = b_utils._EngineCharArrayType.from_pointer(
            c4_to_user, (3,), chrlen=c4_to_user_len
        )
        self._cb_c_to_user = b_utils._EngineComplex64ArrayType.from_pointer(
            c_to_user, (2,),
        )
        self._cb_z_to_user = b_utils._EngineComplex128ArrayType.from_pointer(
            z_to_user, (3,),
        )
        self._cb_r_to_user = b_utils._EngineFloat32ArrayType.from_pointer(
            r_to_user, (2,),
        )
        self._cb_x_to_user = b_utils._EngineFloat64ArrayType.from_pointer(
            x_to_user, (3,),
        )
        self._cb_i_to_user = b_utils._EngineIntArrayType.from_pointer(
            i_to_user, (2,),
        )
        self._cb_l_to_user = b_utils._EngineLogicalArrayType.from_pointer(
            l_to_user, (2,),
        )

    def cb_h(self, cstar, c2, c, z, r, x, i, l, # pylint: disable=invalid-name,too-many-arguments
             cstar_to_user, c4_to_user, c_to_user, z_to_user,
             r_to_user, x_to_user, i_to_user, l_to_user,
             cstar_len, c2_len, cstar_to_user_len, c4_to_user_len):
        """
        Callback for testing direct Engine call.
        """

        # pylint: disable=too-many-locals

        self.cast_cb_input(
            cstar_to_user, c4_to_user, c_to_user, z_to_user,
            r_to_user, x_to_user, i_to_user, l_to_user,
            cstar_to_user_len, c4_to_user_len
        )

        self.check_cb_input(
            self._cb_cstar_to_user, self._cb_c4_to_user, self._cb_c_to_user,
            self._cb_z_to_user, self._cb_r_to_user, self._cb_x_to_user,
            self._cb_i_to_user, self._cb_l_to_user,
        )

        _cstar = b_utils._EngineCharArrayType.from_pointer(
            cstar, (2,), can_reference_contents=False, chrlen=cstar_len
        )
        _c2 = b_utils._EngineCharArrayType.from_pointer(
            c2, (2,), can_reference_contents=False, chrlen=c2_len
        )
        _c = b_utils._EngineComplex64ArrayType.from_pointer(
            c, (2,), can_reference_contents=False,
        )
        _z = b_utils._EngineComplex128ArrayType.from_pointer(
            z, (3,), can_reference_contents=False,
        )
        _r = b_utils._EngineFloat32ArrayType.from_pointer(
            r, (2,), can_reference_contents=False,
        )
        _x = b_utils._EngineFloat64ArrayType.from_pointer(
            x, (3,), can_reference_contents=False,
        )
        _i = b_utils._EngineIntArrayType.from_pointer(
            i, (2,), can_reference_contents=False,
        )
        _l = b_utils._EngineLogicalArrayType.from_pointer(
            l, (2,), can_reference_contents=False,
        )
        _cstar[:] = ['Hello', 'World!']
        _c2[:] = [b'He', b'Wo']
        _c[:] = np.array([42. + 43.j, 44. + 45.j], dtype=np.complex64)
        _z[:] = [42. + 43.j, 44. + 45.j, 46. + 47.j]
        _r[:] = np.array([42., 43.], dtype=np.float32)
        _x[:] = [42., 43., 44.]
        _i[:] = [42, 43]
        _l[:] = [self.engine_state._literal_true_int, 0]
        b_utils._EngineCharArrayType.refresh_ctypes_buffer(
            _cstar, {'fun_name': 'cb', 'entity': 'cstar'},
            cstar, chrlen=cstar_len,
        )
        b_utils._EngineCharArrayType.refresh_ctypes_buffer(
            _c2, {'fun_name': 'cb', 'entity': 'c2'}, c2, chrlen=c2_len,
        )
        b_utils._EngineComplex64ArrayType.refresh_ctypes_buffer(
            _c, {'fun_name': 'cb', 'entity': 'c'}, c,
        )
        b_utils._EngineComplex128ArrayType.refresh_ctypes_buffer(
            _z, {'fun_name': 'cb', 'entity': 'z'}, z,
        )
        b_utils._EngineFloat32ArrayType.refresh_ctypes_buffer(
            _r, {'fun_name': 'cb', 'entity': 'r'}, r,
        )
        b_utils._EngineFloat64ArrayType.refresh_ctypes_buffer(
            _x, {'fun_name': 'cb', 'entity': 'x'}, x,
        )
        b_utils._EngineIntArrayType.refresh_ctypes_buffer(
            _i, {'fun_name': 'cb', 'entity': 'i'}, i,
        )
        b_utils._EngineLogicalArrayType.refresh_ctypes_buffer(
            _l, {'fun_name': 'cb', 'entity': 'l'}, l,
        )
        b_utils._unlink_pointer(cstar)
        b_utils._unlink_pointer(c2)
        b_utils._unlink_pointer(c)
        b_utils._unlink_pointer(z)
        b_utils._unlink_pointer(r)
        b_utils._unlink_pointer(x)
        b_utils._unlink_pointer(i)
        b_utils._unlink_pointer(l)
        b_utils._unlink_pointer(cstar_to_user)
        b_utils._unlink_pointer(c4_to_user)
        b_utils._unlink_pointer(c_to_user)
        b_utils._unlink_pointer(z_to_user)
        b_utils._unlink_pointer(r_to_user)
        b_utils._unlink_pointer(x_to_user)
        b_utils._unlink_pointer(i_to_user)
        b_utils._unlink_pointer(l_to_user)

    def test_good_direct_cast(self): # pylint: disable=too-many-locals,too-many-statements
        """
        Test correct operation with a direct call. Using ctypes.cast.
        """
        old_cast = b_utils._EngineState._cast_pointer_to_ctypes_array
        b_utils._EngineState._cast_pointer_to_ctypes_array = True

        self.make_and_check_direct_call()

        b_utils._EngineState._cast_pointer_to_ctypes_array = old_cast

    def test_good_direct_as_array(self): # pylint: disable=too-many-locals,too-many-statements
        """
        Test correct operation with a direct call.
        Using numpy.ctypeslib.as_array.
        """
        old_cast = b_utils._EngineState._cast_pointer_to_ctypes_array
        b_utils._EngineState._cast_pointer_to_ctypes_array = False

        if (
                self.engine_state._int_model['bits'] == 64 and
                sys.version_info < (3,)
        ):
            # Silence warning about Python 2 NumPy problem,
            # from numpy/ctypeslib.py:
            # (A builtin ctypes object gave a PEP3118 format string that does
            # not match its itemsize).
            # Since we do not ship with the '_cast_pointer_to_ctypes_array'
            # run-time facility disabled, this is merely an annoyance
            # during testing.
            warnings.simplefilter('ignore', RuntimeWarning)

        self.make_and_check_direct_call()
        warnings.simplefilter('default', RuntimeWarning)

        b_utils._EngineState._cast_pointer_to_ctypes_array = old_cast

    # pylint: disable=invalid-name,too-many-arguments,too-many-locals,too-many-statements
    def py_fun_np(self, cstar, c2, c, z, cptr, r, x, i, l, cb):
        """
        Python numpy-style wrapper for self.engine_fun.
        """

        # pylint: disable=attribute-defined-outside-init

        excs = []

        def cb_h(cstar, c2, c, z, r, x, i, l, # pylint: disable=invalid-name,too-many-arguments
                 cstar_to_user, c4_to_user, c_to_user, z_to_user,
                 r_to_user, x_to_user, i_to_user, l_to_user,
                 cstar_len, c2_len, cstar_to_user_len, c4_to_user_len):
            """
            Callback for testing.
            """
            # pylint: disable=bare-except
            try:
                _cstar_a = b_utils._EngineCharArrayType.from_pointer(
                    cstar, (2,), can_reference_contents=False,
                    chrlen=cstar_len,
                )
                _c2_a = b_utils._EngineCharArrayType.from_pointer(
                    c2, (2,), can_reference_contents=False, chrlen=c2_len,
                )
                _c_a = b_utils._EngineComplex64ArrayType.from_pointer(
                    c, (2,), can_reference_contents=False,
                )
                _z_a = b_utils._EngineComplex128ArrayType.from_pointer(
                    z, (3,), can_reference_contents=False,
                )
                _r_a = b_utils._EngineFloat32ArrayType.from_pointer(
                    r, (2,), can_reference_contents=False,
                )
                _x_a = b_utils._EngineFloat64ArrayType.from_pointer(
                    x, (3,), can_reference_contents=False,
                )
                _i_a = b_utils._EngineIntArrayType.from_pointer(
                    i, (2,), can_reference_contents=False,
                )
                _l_a = b_utils._EngineLogicalArrayType.from_pointer(
                    l, (2,), can_reference_contents=False,
                )
                self.cast_cb_input(
                    cstar_to_user, c4_to_user, c_to_user, z_to_user,
                    r_to_user, x_to_user, i_to_user, l_to_user,
                    cstar_to_user_len, c4_to_user_len
                )
                _cstar, _c2, _c, _z, _r, _x, _i, _l = cb(
                    self._cb_cstar_to_user, self._cb_c4_to_user,
                    self._cb_c_to_user, self._cb_z_to_user,
                    self._cb_r_to_user, self._cb_x_to_user, self._cb_i_to_user,
                    self._cb_l_to_user
                )
                _cstar_a[:] = np_utils._to_ndarray(
                    {'fun_name': 'cb', 'entity': 'cstar'},
                    _cstar, copy=False, exp_shape=(2,),
                    exp_class=b_utils._EngineCharArrayType,
                )
                _c2_a[:] = np_utils._to_ndarray(
                    {'fun_name': 'cb', 'entity': 'c2'},
                    _c2, copy=False, exp_shape=(2,),
                    exp_class=b_utils._EngineCharArrayType,
                )
                _c_a[:] = np_utils._to_ndarray(
                    {'fun_name': 'cb', 'entity': 'c'},
                    _c, copy=False, exp_shape=(2,),
                    exp_class=b_utils._EngineComplex64ArrayType,
                )
                _z_a[:] = np_utils._to_ndarray(
                    {'fun_name': 'cb', 'entity': 'z'},
                    _z, copy=False, exp_shape=(3,),
                    exp_class=b_utils._EngineComplex128ArrayType,
                )
                _r_a[:] = np_utils._to_ndarray(
                    {'fun_name': 'cb', 'entity': 'r'},
                    _r, copy=False, exp_shape=(2,),
                    exp_class=b_utils._EngineFloat32ArrayType,
                )
                _x_a[:] = np_utils._to_ndarray(
                    {'fun_name': 'cb', 'entity': 'x'},
                    _x, copy=False, exp_shape=(3,),
                    exp_class=b_utils._EngineFloat64ArrayType,
                )
                _i_a[:] = np_utils._to_ndarray(
                    {'fun_name': 'cb', 'entity': 'i'},
                    _i, copy=False, exp_shape=(2,),
                    exp_class=b_utils._EngineIntArrayType,
                )
                _l_a[:] = np_utils._to_ndarray(
                    {'fun_name': 'cb', 'entity': 'l'},
                    _l, copy=False, exp_shape=(2,),
                    exp_class=b_utils._EngineLogicalArrayType,
                )
                b_utils._EngineCharArrayType.refresh_ctypes_buffer(
                    _cstar_a, {'fun_name': 'cb', 'entity': 'cstar'}, cstar,
                    chrlen=cstar_len,
                )
                b_utils._EngineCharArrayType.refresh_ctypes_buffer(
                    _c2_a, {'fun_name': 'cb', 'entity': 'c2'},
                    c2, chrlen=c2_len,
                )
                b_utils._EngineComplex64ArrayType.refresh_ctypes_buffer(
                    _c_a, {'fun_name': 'cb', 'entity': 'c'}, c,
                )
                b_utils._EngineComplex128ArrayType.refresh_ctypes_buffer(
                    _z_a, {'fun_name': 'cb', 'entity': 'z'}, z,
                )
                b_utils._EngineFloat32ArrayType.refresh_ctypes_buffer(
                    _r_a, {'fun_name': 'cb', 'entity': 'r'}, r,
                )
                b_utils._EngineFloat64ArrayType.refresh_ctypes_buffer(
                    _x_a, {'fun_name': 'cb', 'entity': 'x'}, x,
                )
                b_utils._EngineIntArrayType.refresh_ctypes_buffer(
                    _i_a, {'fun_name': 'cb', 'entity': 'i'}, i,
                )
                b_utils._EngineLogicalArrayType.refresh_ctypes_buffer(
                    _l, {'fun_name': 'cb', 'entity': 'l'}, l,
                )
                b_utils._unlink_pointer(cstar)
                b_utils._unlink_pointer(c2)
                b_utils._unlink_pointer(c)
                b_utils._unlink_pointer(z)
                b_utils._unlink_pointer(r)
                b_utils._unlink_pointer(x)
                b_utils._unlink_pointer(i)
                b_utils._unlink_pointer(l)
                b_utils._unlink_pointer(cstar_to_user)
                b_utils._unlink_pointer(c4_to_user)
                b_utils._unlink_pointer(c_to_user)
                b_utils._unlink_pointer(z_to_user)
                b_utils._unlink_pointer(r_to_user)
                b_utils._unlink_pointer(x_to_user)
                b_utils._unlink_pointer(i_to_user)
                b_utils._unlink_pointer(l_to_user)
            except: # pragma: no cover
                excs.append(sys.exc_info())

        _cstar_cls = b_utils._EngineCharArrayType(
            cstar,
            {'fun_name': 'py_fun_np', 'entity': 'cstar'},
            exp_chrlen=None, exp_shape=(2,),
        )
        _cstar = _cstar_cls.to_ctypes()
        _c2_cls = b_utils._EngineCharArrayType(
            c2,
            {'fun_name': 'py_fun_np', 'entity': 'c2'},
            exp_chrlen=2, exp_shape=(2,),
        )
        _c2 = _c2_cls.to_ctypes()
        _c = b_utils._EngineComplex64ArrayType(
            c,
            {'fun_name': 'py_fun_np', 'entity': 'c'},
            exp_shape=(2,),
        ).to_ctypes()
        _z = b_utils._EngineComplex128ArrayType(
            z,
            {'fun_name': 'py_fun_np', 'entity': 'z'},
            exp_shape=(3,),
        ).to_ctypes()
        _cptr = b_utils._EngineCptrArrayType(
            cptr,
            {'fun_name': 'py_fun_np', 'entity': 'cptr'},
            exp_shape=(3,),
        ).to_ctypes()
        _r = b_utils._EngineFloat32ArrayType(
            r,
            {'fun_name': 'py_fun_np', 'entity': 'r'},
            exp_shape=(2,),
        ).to_ctypes()
        _x = b_utils._EngineFloat64ArrayType(
            x,
            {'fun_name': 'py_fun_np', 'entity': 'x'},
            exp_shape=(3,),
        ).to_ctypes()
        _i = b_utils._EngineIntArrayType(
            i,
            {'fun_name': 'py_fun_np', 'entity': 'i'},
            exp_shape=(2,),
        ).to_ctypes()
        _l = b_utils._EngineLogicalArrayType(
            l,
            {'fun_name': 'py_fun_np', 'entity': 'l'},
            exp_shape=(2,),
        ).to_ctypes()
        self.set_up_wrapper_data()
        self.cast_output_data()
        self._retval = self.engine_fun(
            _cstar,
            _c2,
            _c,
            _z,
            _cptr,
            _r,
            _x,
            _i,
            _l,
            self._cstar_to_user,
            self._c4_to_user,
            self._c_to_user,
            self._z_to_user,
            self._r_to_user,
            self._x_to_user,
            self._i_to_user,
            self._l_to_user,
            self.cb(cb_h),
            _cstar_cls.chrlen,
            _c2_cls.chrlen,
            self._cstar_to_user_cls.chrlen,
            self._c4_to_user_cls.chrlen,
        )
        b_utils._handle_cb_exception(excs)
        self.refresh_output_data()
        return (
            self.retval,
            self.cstar_to_user,
            self.c4_to_user,
            self.c_to_user,
            self.z_to_user,
            self.r_to_user,
            self.x_to_user,
            self.i_to_user,
            self.l_to_user,
        )

    def cb_py_np(self, cstar, c4, c, z, r, x, i, l):
        """
        Callback for testing.
        """
        self.check_cb_input(
            cstar, c4, c, z, r, x, i, l,
        )

        return (
            np.array(['Hello', 'World!']),
            np.array([b'He', b'Wo']),
            np.array(
                [np.complex64(42 + 43j), np.complex64(44 + 45j)],
                dtype=np.complex64
            ),
            np.array(
                [complex(42., 43.), complex(44., 45.), complex(46., 47.)]
            ),
            np.array([np.float32(42.), np.float32(43.)], dtype=np.float32),
            np.array([42., 43., 44.]),
            np.array([42, 43]),
            np.array([True, False]),
        )

    def test_good_wrapper_cast(self):
        """
        Test correct operation via a Python wrapper.
        """
        old_as_ctypes_array = (
            b_utils._EngineState._cast_pointer_to_ctypes_array
        )
        b_utils._EngineState._cast_pointer_to_ctypes_array = True

        self.call_and_check_py()

        b_utils._EngineState._cast_pointer_to_ctypes_array = (
            old_as_ctypes_array
        )

    def test_good_wrapper_as_array(self):
        """
        Test correct operation via a Python wrapper.
        """
        old_as_ctypes_array = (
            b_utils._EngineState._cast_pointer_to_ctypes_array
        )
        b_utils._EngineState._cast_pointer_to_ctypes_array = False

        self.call_and_check_py()

        b_utils._EngineState._cast_pointer_to_ctypes_array = (
            old_as_ctypes_array
        )

    def call_and_check_py(self):
        """
        Call the Python wrapper and check the results.
        """
        (
            retval, cstar, c4, c, z, r, x, i, l,
        ) = self.py_fun_np(
            ['Hello', 'World!'],
            [b'He', b'Wo'],
            [np.complex64(42 + 43j), np.complex64(44 + 45j)],
            [
                np.complex128(42 + 43j),
                np.complex128(44 + 45j),
                np.complex128(46 + 47j),
            ],
            [b_utils.Handle(42), b_utils.Handle(43), b_utils.Handle(44)],
            np.array([42, 43], dtype=np.float32),
            [42., 43., 44.],
            [42, 43],
            [True, False],
            self.cb_py_np,
        )

        self.check_py(
            retval, cstar, c4, c, z, r, x, i, l,
        )

    def check_py(
            self,
            retval, cstar, c4, c, z, r, x, i, l,
    ):
        """
        Check the wrapper's return data.
        """
        self.assertEqual( # pylint: disable=no-member
            retval,
            0
        )
        self.assertEqual( # pylint: disable=no-member
            list(cstar),
            ['Hello', 'World!'],
        )
        self.assertEqual( # pylint: disable=no-member
            list(c4),
            ['aBcD', 'eFgH', 'iJkL', 'mNoP'],
        )
        self.assertEqual( # pylint: disable=no-member
            list(c),
            [31. + 32.j, 33. + 34.j],
        )
        self.assertEqual( # pylint: disable=no-member
            list(z),
            [35. + 36.j, 37. + 38.j, 39. + 40.j],
        )
        self.assertEqual( # pylint: disable=no-member
            list(r),
            [50., 51.],
        )
        self.assertEqual( # pylint: disable=no-member
            list(x),
            [52., 53., 54.],
        )
        self.assertEqual( # pylint: disable=no-member
            list(i),
            [46, 47],
        )
        self.assertEqual( # pylint: disable=no-member
            list(l),
            [False, True],
        )

class PassTestArray1DimIntel(PassTestArray1Dim, unittest.TestCase):
    """
    Pass-test subsuite for Intel.
    """
    compiler = 'ifort'

    @classmethod
    def setUpClass(cls):
        cls.set_up_1dim(cls.compiler)

class PassTestArray1DimNag(PassTestArray1Dim, unittest.TestCase):
    """
    Pass-test subsuite for NAG.
    """
    compiler = 'nagfor'

    @classmethod
    def setUpClass(cls):
        cls.set_up_1dim(cls.compiler)

class PassTestArrayMultidim(PassTest): # pylint: disable=too-many-instance-attributes
    """
    Abstract low-level unit tests for the passing array data to the Engine.
    """

    @classmethod
    def set_up_multidim(cls, compiler):
        """
        The abstract test setup.
        """

        # pylint: disable=line-too-long

        cls.default_engine_state = save_engine_state()

        cls.engine_state = b_utils._EngineState

        if compiler == 'ifort':
            cls.engine_state._compilers = {'Fortran': 'Intel'}
            cls.engine_state._literal_true_int = -1
        else:
            cls.engine_state._compilers = {'Fortran': 'NAG'}
            cls.engine_state._literal_true_int = 1

        cls.set_up('array_pass_multidim_test', compiler)

        cls.engine_state._chrlen_ctype = cls.compiler.chrlen_ctype

        cb_args = (
            ctypes.POINTER(ctypes.c_int), # Input, int sorder.
            ctypes.POINTER(ctypes.c_int), # Output, int hlperr.
            b_utils._EngineCharArrayType.c_type, # Output, cstar(len=*)[ldcstar*3].
            b_utils._EngineCharArrayType.c_type, # Output, c2(len=2)[ldc2*3].
            b_utils._EngineComplex64ArrayType.c_type, # Output, complex64 c[2*3].
            b_utils._EngineComplex128ArrayType.c_type, # Output, complex128 z[ldz*2].
            b_utils._EngineFloat32ArrayType.c_type, # Output, r[ldr*2].
            b_utils._EngineFloat64ArrayType.c_type, # Output, x[ldx*2].
            b_utils._EngineIntArrayType.c_type, # Output, i[ldi*2].
            b_utils._EngineLogicalArrayType.c_type, # Output, logical l[ldl*3].
            b_utils._EngineCharArrayType.c_type, # Input, cstar_to_user(len=*)[3*2].
            b_utils._EngineCharArrayType.c_type, # Input, c2_to_user(len=2)[3*2].
            b_utils._EngineComplex64ArrayType.c_type, # Input, c_to_user[3*2].
            b_utils._EngineComplex128ArrayType.c_type, # Input, z_to_user[3*2].
            b_utils._EngineFloat32ArrayType.c_type, # Input, r_to_user[3*2].
            b_utils._EngineFloat64ArrayType.c_type, # Input, x_to_user[3*2].
            b_utils._EngineIntArrayType.c_type, # Input, i_to_user[3*2].
            b_utils._EngineLogicalArrayType.c_type, # Input, l_to_user[3*2].
            cls.engine_state._chrlen_ctype, # Input, chrlen of cstar.
            cls.engine_state._chrlen_ctype, # Input, chrlen of c2.
            cls.engine_state._chrlen_ctype, # Input, chrlen of cstar_to_user.
            cls.engine_state._chrlen_ctype, # Input, chrlen of c2_to_user.
        )
        cls.cb = b_utils._EngineState._sys_functype(
            None,
            *cb_args
        )

        cls.engine_fun.restype = b_utils._EngineIntScalarType.c_type
        cls.engine_fun.argtypes = (
            ctypes.POINTER(ctypes.c_int), # Input, int sorder.
            b_utils._BYREF_INT, # Output, int hlperr.
            b_utils._EngineCharArrayType.c_type, # Input, cstar(len=*)[ldcstar,*].
            b_utils._BYREF_INT, # Input, ldcstar.
            b_utils._EngineCharArrayType.c_type, # Input, c2(len=2)[ldc2,*].
            b_utils._BYREF_INT, # Input, ldc2.
            b_utils._EngineComplex64ArrayType.c_type, # Input, complex64 c[2, 3].
            b_utils._EngineComplex128ArrayType.c_type, # Input, complex128 z[ldz,*].
            b_utils._BYREF_INT, # Input, ldz.
            b_utils._EngineFloat32ArrayType.c_type, # Input, r[ldr,*].
            b_utils._BYREF_INT, # Input, ldr.
            b_utils._EngineFloat64ArrayType.c_type, # Input, x[ldx,*].
            b_utils._BYREF_INT, # Input, ldx.
            b_utils._EngineIntArrayType.c_type, # Input, i[ldi, *].
            b_utils._BYREF_INT, # Input, ldi.
            b_utils._EngineLogicalArrayType.c_type, # Input, logical l[ldl, 3].
            b_utils._BYREF_INT, # Input, ldl.
            b_utils._EngineCharArrayType.c_type, # Output, cstar_to_user(len=*)[3*2].
            b_utils._EngineCharArrayType.c_type, # Output, c2_to_user(len=2)[3*2].
            b_utils._EngineComplex64ArrayType.c_type, # Output, c_to_user[3*2].
            b_utils._EngineComplex128ArrayType.c_type, # Output, z_to_user[3*2].
            b_utils._EngineFloat32ArrayType.c_type, # Output, r_to_user[3*2].
            b_utils._EngineFloat64ArrayType.c_type, # Output, x_to_user[3*2].
            b_utils._EngineIntArrayType.c_type, # Output, i_to_user[3*2].
            b_utils._EngineLogicalArrayType.c_type, # Output, l_to_user[3*2].
            cls.cb,
            cls.engine_state._chrlen_ctype, # Input, chrlen of cstar.
            cls.engine_state._chrlen_ctype, # Input, chrlen of c2.
            cls.engine_state._chrlen_ctype, # Input, chrlen of cstar_to_user.
            cls.engine_state._chrlen_ctype, # Input, chrlen of c2_to_user.
        )

    def set_up_direct_data(
            self,
    ):
        """
        The default 'direct' call output (initialized) Fortran-contiguous
        data set.
        """

        # pylint: disable=attribute-defined-outside-init

        self.cstar_to_user = np.array(['?!']*6).reshape((3, 2), order='F')
        self.c2_to_user = np.array(['?!']*6).reshape((3, 2), order='F')
        self.c_to_user = np.array(
            [complex(42., 42.)]*6, dtype=np.complex64
        ).reshape((3, 2), order='F')
        self.z_to_user = np.array(
            [complex(42., 42.)]*6, dtype=complex
        ).reshape((3, 2), order='F')
        self.r_to_user = np.array([42.]*6, dtype=np.float32).reshape(
            (3, 2), order='F'
        )
        self.x_to_user = np.array([42.]*6, dtype=float).reshape(
            (3, 2), order='F'
        )
        self.i_to_user = np.array([42]*6, dtype=int).reshape((3, 2), order='F')
        self.l_to_user = np.array([True]*6, dtype=bool).reshape(
            (3, 2), order='F'
        )

    def set_up_wrapper_data(
            self, order,
    ):
        """
        The default 'wrapper' call output (uninitialized) data set.
        """

        # pylint: disable=attribute-defined-outside-init

        self.cstar_to_user = b_utils._EngineCharArrayType.empty_ndarray(
            (3, 2), sorder=order, chrlen=2
        )
        self.c2_to_user = b_utils._EngineCharArrayType.empty_ndarray(
            (3, 2), sorder=order, chrlen=2
        )
        self.c_to_user = b_utils._EngineComplex64ArrayType.empty_ndarray(
            (3, 2), sorder=order
        )
        self.z_to_user = b_utils._EngineComplex128ArrayType.empty_ndarray(
            (3, 2), sorder=order
        )
        self.r_to_user = b_utils._EngineFloat32ArrayType.empty_ndarray(
            (3, 2), sorder=order
        )
        self.x_to_user = b_utils._EngineFloat64ArrayType.empty_ndarray(
            (3, 2), sorder=order
        )
        self.i_to_user = b_utils._EngineIntArrayType.empty_ndarray(
            (3, 2), sorder=order
        )
        self.l_to_user = b_utils._EngineLogicalArrayType.empty_ndarray(
            (3, 2), sorder=order
        )

    def cast_output_data(
            self, mutual_sorder,
    ):
        """
        Cast the output data.
        """

        # pylint: disable=attribute-defined-outside-init

        self._cstar_to_user_cls = b_utils._EngineCharArrayType(
            self.cstar_to_user,
            {'fun_name': 'cast_output_data', 'entity': 'cstar_to_user'},
            exp_chrlen=None, exp_shape=(3, 2), check_contents=False,
            mutual_sorder=mutual_sorder,
        )
        self._cstar_to_user = self._cstar_to_user_cls.to_ctypes()
        self._c2_to_user_cls = b_utils._EngineCharArrayType(
            self.c2_to_user,
            {'fun_name': 'cast_output_data', 'entity': 'c2_to_user'},
            exp_chrlen=2, exp_shape=(3, 2), check_contents=False,
            mutual_sorder=mutual_sorder,
        )
        self._c2_to_user = self._c2_to_user_cls.to_ctypes()
        self._c_to_user_cls = b_utils._EngineComplex64ArrayType(
            self.c_to_user,
            {'fun_name': 'cast_output_data', 'entity': 'c_to_user'},
            exp_shape=(3, 2), check_contents=False,
            mutual_sorder=mutual_sorder,
        )
        self._c_to_user = self._c_to_user_cls.to_ctypes()
        self._z_to_user_cls = b_utils._EngineComplex128ArrayType(
            self.z_to_user,
            {'fun_name': 'cast_output_data', 'entity': 'z_to_user'},
            exp_shape=(3, 2), check_contents=False,
            mutual_sorder=mutual_sorder,
        )
        self._z_to_user = self._z_to_user_cls.to_ctypes()
        self._r_to_user_cls = b_utils._EngineFloat32ArrayType(
            self.r_to_user,
            {'fun_name': 'cast_output_data', 'entity': 'r_to_user'},
            exp_shape=(3, 2), check_contents=False,
            mutual_sorder=mutual_sorder,
        )
        self._r_to_user = self._r_to_user_cls.to_ctypes()
        self._x_to_user_cls = b_utils._EngineFloat64ArrayType(
            self.x_to_user,
            {'fun_name': 'cast_output_data', 'entity': 'x_to_user'},
            exp_shape=(3, 2), check_contents=False,
            mutual_sorder=mutual_sorder,
        )
        self._x_to_user = self._x_to_user_cls.to_ctypes()
        self._i_to_user_cls = b_utils._EngineIntArrayType(
            self.i_to_user,
            {'fun_name': 'cast_output_data', 'entity': 'i_to_user'},
            exp_shape=(3, 2), check_contents=False,
            mutual_sorder=mutual_sorder,
        )
        self._i_to_user = self._i_to_user_cls.to_ctypes()
        self._l_to_user_cls = b_utils._EngineLogicalArrayType(
            self.l_to_user,
            {'fun_name': 'cast_output_data', 'entity': 'l_to_user'},
            exp_shape=(3, 2), check_contents=False,
            mutual_sorder=mutual_sorder,
        )
        self._l_to_user = self._l_to_user_cls.to_ctypes()

    def refresh_output_data(
            self,
    ):
        """
        Poke the Engine output data back into order-contiguous Python
        equivalents.
        """

        # pylint: disable=attribute-defined-outside-init

        self.retval = b_utils._EngineIntScalarType.to_py_from_retval(
            self._retval
        )
        self._cstar_to_user_cls.refresh_py(self._cstar_to_user)
        self._c2_to_user_cls.refresh_py(self._c2_to_user)
        self._c_to_user_cls.refresh_py(self._c_to_user)
        self._z_to_user_cls.refresh_py(self._z_to_user)
        self._r_to_user_cls.refresh_py(self._r_to_user)
        self._x_to_user_cls.refresh_py(self._x_to_user)
        self._i_to_user_cls.refresh_py(self._i_to_user)
        self._l_to_user_cls.refresh_py(self._l_to_user)

    def check_output_data(
            self, order,
    ):
        """
        Check the output data.
        """
        self.check_py(
            order, self.retval,
            self.cstar_to_user, self.c2_to_user, self.c_to_user,
            self.z_to_user, self.r_to_user, self.x_to_user, self.i_to_user,
            self.l_to_user,
        )

    @staticmethod
    def check_cb_input( # pylint: disable=invalid-name,too-many-arguments
            order,
            cstar, c2, c, z, r, x, i, l,
    ):
        """
        Check input data passed in to the callback.
        """
        if order == 'F':
            np.testing.assert_equal(
                cstar,
                [
                    [b'He', b'Wo'],
                    [b'll', b'rl'],
                    [b'o', b'd!'],
                ]
            )
            np.testing.assert_equal(
                c2,
                [
                    [b'He', b'Wo'],
                    [b'll', b'rl'],
                    [b'o', b'd!'],
                ]
            )
            np.testing.assert_equal(
                c,
                [
                    [42. + 43.j, 48. + 49.j],
                    [44. + 45.j, 50. + 51.j],
                    [46. + 47.j, 52. + 53.j],
                ]
            )
            np.testing.assert_equal(
                z,
                [
                    [42. + 43.j, 48. + 49.j],
                    [44. + 45.j, 50. + 51.j],
                    [46. + 47.j, 52. + 53.j],
                ]
            )
            np.testing.assert_equal(
                r,
                [
                    [42., 45.],
                    [43., 46.],
                    [44., 47.]
                ]
            )
            np.testing.assert_equal(
                x,
                [
                    [42., 45.],
                    [43., 46.],
                    [44., 47.]
                ]
            )
            np.testing.assert_equal(
                i,
                [
                    [42, 45],
                    [43, 46],
                    [44, 47]
                ]
            )
            np.testing.assert_equal(
                l,
                [
                    [True, False],
                    [True, True],
                    [False, False],
                ]
            )
        else:
            np.testing.assert_equal(
                cstar,
                [
                    [b'He', b'll'],
                    [b'o', b'Wo'],
                    [b'rl', b'd!'],
                ]
            )
            np.testing.assert_equal(
                c2,
                [
                    [b'He', b'll'],
                    [b'o', b'Wo'],
                    [b'rl', b'd!'],
                ]
            )
            np.testing.assert_equal(
                c,
                [
                    [42. + 43.j, 44. + 45.j],
                    [46. + 47.j, 48. + 49.j],
                    [50. + 51.j, 52. + 53.j],
                ]
            )
            np.testing.assert_equal(
                z,
                [
                    [42. + 43.j, 44. + 45.j],
                    [46. + 47.j, 48. + 49.j],
                    [50. + 51.j, 52. + 53.j],
                ]
            )
            np.testing.assert_equal(
                r,
                [
                    [42., 43.],
                    [44., 45.],
                    [46., 47.]
                ]
            )
            np.testing.assert_equal(
                x,
                [
                    [42., 43.],
                    [44., 45.],
                    [46., 47.]
                ]
            )
            np.testing.assert_equal(
                i,
                [
                    [42, 43],
                    [44, 45],
                    [46, 47]
                ]
            )
            np.testing.assert_equal(
                l,
                [
                    [True, True],
                    [False, False],
                    [True, False],
                ]
            )

    def cast_cb_input( # pylint: disable=too-many-arguments
            self, order,
            cstar_to_user, c2_to_user, c_to_user, z_to_user,
            r_to_user, x_to_user, i_to_user, l_to_user, cstar_to_user_len,
            c2_to_user_len,
    ):
        """
        Cast the input callback data to Python.
        """

        # pylint: disable=attribute-defined-outside-init

        self._cb_cstar_to_user = b_utils._EngineCharArrayType.from_pointer(
            cstar_to_user, (3, 2), chrlen=cstar_to_user_len, sorder=order,
        )
        self._cb_c2_to_user = b_utils._EngineCharArrayType.from_pointer(
            c2_to_user, (3, 2), chrlen=c2_to_user_len, sorder=order,
        )
        self._cb_c_to_user = b_utils._EngineComplex64ArrayType.from_pointer(
            c_to_user, (3, 2), sorder=order,
        )
        self._cb_z_to_user = b_utils._EngineComplex128ArrayType.from_pointer(
            z_to_user, (3, 2), sorder=order,
        )
        self._cb_r_to_user = b_utils._EngineFloat32ArrayType.from_pointer(
            r_to_user, (3, 2), sorder=order,
        )
        self._cb_x_to_user = b_utils._EngineFloat64ArrayType.from_pointer(
            x_to_user, (3, 2), sorder=order,
        )
        self._cb_i_to_user = b_utils._EngineIntArrayType.from_pointer(
            i_to_user, (3, 2), sorder=order,
        )
        self._cb_l_to_user = b_utils._EngineLogicalArrayType.from_pointer(
            l_to_user, (3, 2), sorder=order,
        )

    def check_py( # pylint: disable=invalid-name,too-many-arguments
            self, order,
            retval,
            cstar, c2, c, z, r, x, i, l,
    ):
        """
        Check the wrapper's return data.
        """
        self.assertEqual( # pylint: disable=no-member
            retval,
            0
        )
        if order == 'F':
            np.testing.assert_equal(
                cstar,
                [
                    ['He', 'Wo'],
                    ['ll', 'rl'],
                    ['o', 'd!'],
                ]
            )
            np.testing.assert_equal(
                c2,
                [
                    ['He', 'Wo'],
                    ['ll', 'rl'],
                    ['o', 'd!'],
                ]
            )
            np.testing.assert_equal(
                c,
                [
                    [42. + 43.j, 48. + 49.j],
                    [44. + 45.j, 50. + 51.j],
                    [46. + 47.j, 52. + 53.j],
                ]
            )
            np.testing.assert_equal(
                z,
                [
                    [42. + 43.j, 48. + 49.j],
                    [44. + 45.j, 50. + 51.j],
                    [46. + 47.j, 52. + 53.j],
                ]
            )
            np.testing.assert_equal(
                r,
                [
                    [53., 56.],
                    [54., 57.],
                    [55., 58.],
                ]
            )
            np.testing.assert_equal(
                x,
                [
                    [53., 56.],
                    [54., 57.],
                    [55., 58.],
                ]
            )
            np.testing.assert_equal(
                i,
                [
                    [53, 56],
                    [54, 57],
                    [55, 58],
                ]
            )
            np.testing.assert_equal(
                l,
                [
                    [True, False],
                    [True, False],
                    [True, False],
                ]
            )
        else:
            np.testing.assert_equal(
                cstar,
                [
                    ['He', 'll'],
                    ['o', 'Wo'],
                    ['rl', 'd!'],
                ]
            )
            np.testing.assert_equal(
                c2,
                [
                    ['He', 'll'],
                    ['o', 'Wo'],
                    ['rl', 'd!'],
                ]
            )
            np.testing.assert_equal(
                c,
                [
                    [42. + 43.j, 44. + 45.j],
                    [46. + 47.j, 48. + 49.j],
                    [50. + 51.j, 52. + 53.j],
                ]
            )
            np.testing.assert_equal(
                z,
                [
                    [42. + 43.j, 44. + 45.j],
                    [46. + 47.j, 48. + 49.j],
                    [50. + 51.j, 52. + 53.j],
                ]
            )
            np.testing.assert_equal(
                r,
                [
                    [53., 54.],
                    [55., 56.],
                    [57., 58.],
                ]
            )
            np.testing.assert_equal(
                x,
                [
                    [53., 54.],
                    [55., 56.],
                    [57., 58.],
                ]
            )
            np.testing.assert_equal(
                i,
                [
                    [53, 54],
                    [55, 56],
                    [57, 58],
                ]
            )
            np.testing.assert_equal(
                l,
                [
                    [True, True],
                    [True, False],
                    [False, False],
                ]
            )

    def test_good_direct(self): # pylint: disable=too-many-statements
        """
        Test correct operation with a direct call.
        """

        # pylint: disable=attribute-defined-outside-init

        def cb_h( # pylint: disable=invalid-name,too-many-arguments
                sorder, hlperr, cstar, c2, c, z, r, x, i, l,
                cstar_to_user, c2_to_user, c_to_user, z_to_user,
                r_to_user, x_to_user, i_to_user, l_to_user,
                cstar_len, c2_len, cstar_to_user_len, c2_to_user_len,
        ):
            """
            Callback for testing.
            """

            # pylint: disable=too-many-locals

            hlperr.contents.value = 0

            order = 'F'

            self.cast_cb_input(
                order,
                cstar_to_user, c2_to_user, c_to_user, z_to_user,
                r_to_user, x_to_user, i_to_user, l_to_user, cstar_to_user_len,
                c2_to_user_len,
            )
            self.check_cb_input(
                order,
                self._cb_cstar_to_user, self._cb_c2_to_user,
                self._cb_c_to_user, self._cb_z_to_user,
                self._cb_r_to_user, self._cb_x_to_user, self._cb_i_to_user,
                self._cb_l_to_user,
            )

            _cstar = b_utils._EngineCharArrayType.from_pointer(
                cstar, (2, 3), can_reference_contents=False, chrlen=cstar_len,
                sorder=order,
            )
            _c2 = b_utils._EngineCharArrayType.from_pointer(
                c2, (2, 3), can_reference_contents=False, chrlen=c2_len,
                sorder=order,
            )
            _c = b_utils._EngineComplex64ArrayType.from_pointer(
                c, (2, 3), can_reference_contents=False, sorder=order,
            )
            _z = b_utils._EngineComplex128ArrayType.from_pointer(
                z, (3, 2), can_reference_contents=False, sorder=order,
            )
            _r = b_utils._EngineFloat32ArrayType.from_pointer(
                r, (5, 2), can_reference_contents=False, sorder=order,
            )
            _x = b_utils._EngineFloat64ArrayType.from_pointer(
                x, (5, 2), can_reference_contents=False, sorder=order,
            )
            _i = b_utils._EngineIntArrayType.from_pointer(
                i, (5, 2), can_reference_contents=False, sorder=order,
            )
            _l = b_utils._EngineLogicalArrayType.from_pointer(
                l, (2, 3), can_reference_contents=False, sorder=order,
            )
            _cstar[:, :] = [
                ['Hello', 'The', 'World!'],
                ['Allo', 'Le', 'Monde!']
            ]
            _c2[:, :] = [[b'He', b'll', b'o'], [b'Wo', b'rl', b'd!']]
            _c[:, :] = np.array(
                [
                    [42. + 43.j, 44. + 45.j, 46. + 47.j],
                    [48. + 49.j, 50. + 51.j, 52. + 53.j]
                ],
                dtype=np.complex64
            )
            _z[:, :] = [
                [42. + 43.j, 44. + 45.j],
                [46. + 47.j, 48. + 49.j],
                [50. + 51.j, 52. + 53.j]
            ]
            _r[:, :] = np.array(
                [
                    [42., 43.], [44., 45.], [46., 47.], [48., 49.], [50., 51.]
                ],
                dtype=np.float32
            )
            _x[:, :] = [
                [42., 43.], [44., 45.], [46., 47.], [48., 49.], [50., 51.]
            ]
            _i[:, :] = [[42, 43], [44, 45], [46, 47], [48, 49], [50, 51]]
            _l[:, :] = [
                [
                    self.engine_state._literal_true_int, 0,
                    self.engine_state._literal_true_int
                ],
                [
                    self.engine_state._literal_true_int,
                    self.engine_state._literal_true_int, 0
                ]
            ]
            b_utils._EngineCharArrayType.refresh_ctypes_buffer(
                _cstar, {'fun_name': 'cb', 'entity': 'cstar'},
                cstar, chrlen=cstar_len,
            )
            b_utils._EngineCharArrayType.refresh_ctypes_buffer(
                _c2, {'fun_name': 'cb', 'entity': 'c2'}, c2, chrlen=c2_len,
            )
            b_utils._EngineComplex64ArrayType.refresh_ctypes_buffer(
                _c, {'fun_name': 'cb', 'entity': 'c'}, c,
            )
            b_utils._EngineComplex128ArrayType.refresh_ctypes_buffer(
                _z, {'fun_name': 'cb', 'entity': 'z'}, z,
            )
            b_utils._EngineFloat32ArrayType.refresh_ctypes_buffer(
                _r, {'fun_name': 'cb', 'entity': 'r'}, r,
            )
            b_utils._EngineFloat64ArrayType.refresh_ctypes_buffer(
                _x, {'fun_name': 'cb', 'entity': 'x'}, x,
            )
            b_utils._EngineIntArrayType.refresh_ctypes_buffer(
                _i, {'fun_name': 'cb', 'entity': 'i'}, i,
            )
            b_utils._EngineLogicalArrayType.refresh_ctypes_buffer(
                _l, {'fun_name': 'cb', 'entity': 'l'}, l,
            )
            b_utils._unlink_pointer(cstar)
            b_utils._unlink_pointer(c2)
            b_utils._unlink_pointer(c)
            b_utils._unlink_pointer(z)
            b_utils._unlink_pointer(r)
            b_utils._unlink_pointer(x)
            b_utils._unlink_pointer(i)
            b_utils._unlink_pointer(l)
            b_utils._unlink_pointer(cstar_to_user)
            b_utils._unlink_pointer(c2_to_user)
            b_utils._unlink_pointer(c_to_user)
            b_utils._unlink_pointer(z_to_user)
            b_utils._unlink_pointer(r_to_user)
            b_utils._unlink_pointer(x_to_user)
            b_utils._unlink_pointer(i_to_user)
            b_utils._unlink_pointer(l_to_user)

        self.set_up_direct_data()
        self.cast_output_data('F')

        _hlperr = b_utils._EngineIntScalarType.empty_ctypes()

        self._retval = self.engine_fun(
            ctypes.c_int(1),
            _hlperr,
            ctypes.create_string_buffer(b'Hello World!', size=12),
            b_utils._EngineIntScalarType.c_type(2),
            ctypes.create_string_buffer(b'Hello World!', size=12),
            b_utils._EngineIntScalarType.c_type(2),
            (b_utils._EngineComplex64ArrayType.c_type._type_*6)(
                *[(42 + 2*i, 42 + 2*i + 1) for i in range(6)]
            ),
            (b_utils._EngineComplex128ArrayType.c_type._type_*6)(
                *[(42 + 2*i, 42 + 2*i + 1) for i in range(6)]
            ),
            b_utils._EngineIntScalarType.c_type(3),
            (b_utils._EngineFloat32ArrayType.c_type._type_*10)(
                *[42 + i for i in range(10)]
            ),
            b_utils._EngineIntScalarType.c_type(5),
            (b_utils._EngineFloat64ArrayType.c_type._type_*10)(
                *[42 + i for i in range(10)]
            ),
            b_utils._EngineIntScalarType.c_type(5),
            (b_utils._EngineIntArrayType.c_type._type_*10)(
                *[42 + i for i in range(10)]
            ),
            b_utils._EngineIntScalarType.c_type(5),
            (b_utils._EngineLogicalArrayType.c_type._type_*6)(
                *[1, 0, 1, 0, 1, 0]
            ),
            b_utils._EngineIntScalarType.c_type(2),
            self._cstar_to_user,
            self._c2_to_user,
            self._c_to_user,
            self._z_to_user,
            self._r_to_user,
            self._x_to_user,
            self._i_to_user,
            self._l_to_user,
            self.cb(cb_h),
            2,
            2,
            2,
            2,
        )
        self.assertEqual( # pylint: disable=no-member
            _hlperr.value,
            0,
        )

        self.refresh_output_data()
        self.check_output_data('F')

    # pylint: disable=invalid-name,too-many-arguments,too-many-locals,too-many-statements
    def py_fun(
            self,
            cstar,
            c2,
            c,
            z,
            r,
            x,
            i,
            l,
            cb,
    ):
        """
        Python wrapper for self.engine_fun.
        """

        # pylint: disable=attribute-defined-outside-init

        excs = []

        def cb_h( # pylint: disable=invalid-name,too-many-arguments
                sorder, hlperr, cstar, c2, c, z, r, x, i, l,
                cstar_to_user, c2_to_user, c_to_user, z_to_user,
                r_to_user, x_to_user, i_to_user, l_to_user,
                cstar_len, c2_len, cstar_to_user_len, c2_to_user_len,
        ):
            """
            Test helper.
            """
            # pylint: disable=bare-except
            order = (
                'F' if sorder.contents.value == b_utils._NAG_ED_COL_MAJOR.value
                else 'C'
            )

            try:
                _cstar_a = b_utils._EngineCharArrayType.from_pointer(
                    cstar, (2, 3), can_reference_contents=False,
                    chrlen=cstar_len, sorder=order,
                )
                _c2_a = b_utils._EngineCharArrayType.from_pointer(
                    c2, (2, 3), can_reference_contents=False, chrlen=c2_len,
                    sorder=order,
                )
                _c_a = b_utils._EngineComplex64ArrayType.from_pointer(
                    c, (2, 3), can_reference_contents=False, sorder=order,
                )
                _z_a = b_utils._EngineComplex128ArrayType.from_pointer(
                    z, (3, 2), can_reference_contents=False, sorder=order,
                )
                _r_a = b_utils._EngineFloat32ArrayType.from_pointer(
                    r, (5, 2), can_reference_contents=False, sorder=order,
                )
                _x_a = b_utils._EngineFloat64ArrayType.from_pointer(
                    x, (5, 2), can_reference_contents=False, sorder=order,
                )
                _i_a = b_utils._EngineIntArrayType.from_pointer(
                    i, (5, 2), can_reference_contents=False, sorder=order,
                )
                _l_a = b_utils._EngineLogicalArrayType.from_pointer(
                    l, (2, 3), can_reference_contents=False, sorder=order,
                )
                self.cast_cb_input(
                    order,
                    cstar_to_user, c2_to_user, c_to_user, z_to_user,
                    r_to_user, x_to_user, i_to_user, l_to_user,
                    cstar_to_user_len, c2_to_user_len,
                )
                _cstar, _c2, _c, _z, _r, _x, _i, _l = cb(
                    self._cb_cstar_to_user, self._cb_c2_to_user,
                    self._cb_c_to_user, self._cb_z_to_user,
                    self._cb_r_to_user, self._cb_x_to_user,
                    self._cb_i_to_user, self._cb_l_to_user,
                )
                hlperr.contents.value = 0
                np_utils._check_callback_sorder(
                    {'fun_name': 'cb'},
                    (
                        (_cstar, '_cstar'),
                        (_c2, '_c2'),
                        (_c, '_c'),
                        (_z, '_z'),
                        (_r, '_r'),
                        (_x, '_x'),
                        (_i, '_i'),
                        (_l, '_l'),
                    ),
                    wrapper_sorder=_m_order
                )
                _cstar_a[:, :] = np_utils._to_ndarray(
                    {'fun_name': 'cb', 'entity': 'cstar'},
                    _cstar, copy=False, exp_shape=(2, 3),
                    exp_class=b_utils._EngineCharArrayType,
                )
                _c2_a[:, :] = np_utils._to_ndarray(
                    {'fun_name': 'cb', 'entity': 'c2'},
                    _c2, copy=False, exp_shape=(2, 3),
                    exp_class=b_utils._EngineCharArrayType,
                )
                _c_a[:, :] = np_utils._to_ndarray(
                    {'fun_name': 'cb', 'entity': 'c'},
                    _c, copy=False, exp_shape=(2, 3),
                    exp_class=b_utils._EngineComplex64ArrayType,
                )
                _z_a[:, :] = np_utils._to_ndarray(
                    {'fun_name': 'cb', 'entity': 'z'},
                    _z, copy=False, exp_shape=(3, 2),
                    exp_class=b_utils._EngineComplex128ArrayType,
                )
                _r_a[:, :] = np_utils._to_ndarray(
                    {'fun_name': 'cb', 'entity': 'r'},
                    _r, copy=False, exp_shape=(5, 2),
                    exp_class=b_utils._EngineFloat32ArrayType,
                )
                _x_a[:, :] = np_utils._to_ndarray(
                    {'fun_name': 'cb', 'entity': 'x'},
                    _x, copy=False, exp_shape=(5, 2),
                    exp_class=b_utils._EngineFloat64ArrayType,
                )
                _i_a[:, :] = np_utils._to_ndarray(
                    {'fun_name': 'cb', 'entity': 'i'},
                    _i, copy=False, exp_shape=(5, 2),
                    exp_class=b_utils._EngineIntArrayType,
                )
                _l_a[:, :] = np_utils._to_ndarray(
                    {'fun_name': 'cb', 'entity': 'l'},
                    _l, copy=False, exp_shape=(2, 3),
                    exp_class=b_utils._EngineLogicalArrayType,
                )
                b_utils._EngineCharArrayType.refresh_ctypes_buffer(
                    _cstar_a, {'fun_name': 'cb', 'entity': 'cstar'}, cstar,
                    chrlen=cstar_len,
                )
                b_utils._EngineCharArrayType.refresh_ctypes_buffer(
                    _c2_a, {'fun_name': 'cb', 'entity': 'c2'},
                    c2, chrlen=c2_len,
                )
                b_utils._EngineComplex64ArrayType.refresh_ctypes_buffer(
                    _c_a, {'fun_name': 'cb', 'entity': 'c'}, c,
                )
                b_utils._EngineComplex128ArrayType.refresh_ctypes_buffer(
                    _z_a, {'fun_name': 'cb', 'entity': 'z'}, z,
                )
                b_utils._EngineFloat32ArrayType.refresh_ctypes_buffer(
                    _r_a, {'fun_name': 'cb', 'entity': 'r'}, r,
                )
                b_utils._EngineFloat64ArrayType.refresh_ctypes_buffer(
                    _x_a, {'fun_name': 'cb', 'entity': 'x'}, x,
                )
                b_utils._EngineIntArrayType.refresh_ctypes_buffer(
                    _i_a, {'fun_name': 'cb', 'entity': 'i'}, i,
                )
                b_utils._EngineLogicalArrayType.refresh_ctypes_buffer(
                    _l_a, {'fun_name': 'cb', 'entity': 'l'}, l,
                )
                b_utils._unlink_pointer(cstar)
                b_utils._unlink_pointer(c2)
                b_utils._unlink_pointer(c)
                b_utils._unlink_pointer(z)
                b_utils._unlink_pointer(r)
                b_utils._unlink_pointer(x)
                b_utils._unlink_pointer(i)
                b_utils._unlink_pointer(l)
                b_utils._unlink_pointer(cstar_to_user)
                b_utils._unlink_pointer(c2_to_user)
                b_utils._unlink_pointer(c_to_user)
                b_utils._unlink_pointer(z_to_user)
                b_utils._unlink_pointer(r_to_user)
                b_utils._unlink_pointer(x_to_user)
                b_utils._unlink_pointer(i_to_user)
                b_utils._unlink_pointer(l_to_user)
            except: # pragma: no cover
                excs.append(sys.exc_info())
                hlperr.contents.value = b_utils._IERR_HELPER

        _m_order = b_utils._get_mutual_sorder(
            {'fun_name': 'py_fun'},
            (
                (cstar, 'cstar'),
                (c2, 'c2'),
                (c, 'c'),
                (z, 'z'),
                (r, 'r'),
                (x, 'x'),
                (i, 'i'),
                (l, 'l'),
            ),
        )
        _ldcstar = b_utils._EngineIntScalarType(
            b_utils._EngineCharArrayType.get_nd(
                cstar,
                {'fun_name': 'py_fun', 'entity': 'cstar'},
                exp_rank=2, sorder=_m_order, dim_type='l',
            ),
            {'fun_name': 'py_fun', 'entity': '_ldcstar'},
        ).to_ctypes()
        _cstar_cls = b_utils._EngineCharArrayType(
            cstar,
            {'fun_name': 'py_fun', 'entity': 'cstar'},
            exp_chrlen=None, exp_shape=(2, 3), mutual_sorder=_m_order
        )
        _cstar = _cstar_cls.to_ctypes()
        _ldc2 = b_utils._EngineIntScalarType(
            b_utils._EngineCharArrayType.get_nd(
                c2,
                {'fun_name': 'py_fun', 'entity': 'c2'},
                exp_rank=2, sorder=_m_order, dim_type='l',
            ),
            {'fun_name': 'py_fun', 'entity': '_ldc2'},
        ).to_ctypes()
        _c2_cls = b_utils._EngineCharArrayType(
            c2,
            {'fun_name': 'py_fun', 'entity': 'c2'},
            exp_chrlen=2, exp_shape=(2, 3), mutual_sorder=_m_order,
        )
        _c2 = _c2_cls.to_ctypes()
        _ldc = b_utils._EngineIntScalarType(
            b_utils._EngineComplex64ArrayType.get_nd(
                c,
                {'fun_name': 'py_fun', 'entity': 'c'},
                exp_rank=2, sorder=_m_order, dim_type='l',
            ),
            {'fun_name': 'py_fun', 'entity': '_ldc'},
        ).to_ctypes()
        _c = b_utils._EngineComplex64ArrayType(
            c,
            {'fun_name': 'py_fun', 'entity': 'c'},
            exp_shape=(2, 3), mutual_sorder=_m_order,
        ).to_ctypes()
        _ldz = b_utils._EngineIntScalarType(
            b_utils._EngineComplex64ArrayType.get_nd(
                z,
                {'fun_name': 'py_fun', 'entity': 'z'},
                exp_rank=2, sorder=_m_order, dim_type='l',
            ),
            {'fun_name': 'py_fun', 'entity': '_ldz'},
        ).to_ctypes()
        _z_cls = b_utils._EngineComplex128ArrayType(
            z,
            {'fun_name': 'py_fun', 'entity': 'z'},
            exp_shape=(3, 2), mutual_sorder=_m_order,
        )
        _z = _z_cls.to_ctypes()
        _ldr = b_utils._EngineIntScalarType(
            b_utils._EngineFloat32ArrayType.get_nd(
                r,
                {'fun_name': 'py_fun', 'entity': 'r'},
                exp_rank=2, sorder=_m_order, dim_type='l',
            ),
            {'fun_name': 'py_fun', 'entity': '_ldr'},
        ).to_ctypes()
        _r_cls = b_utils._EngineFloat32ArrayType(
            r,
            {'fun_name': 'py_fun', 'entity': 'r'},
            exp_shape=(5, 2), mutual_sorder=_m_order,
        )
        _r = _r_cls.to_ctypes()
        _ldx = b_utils._EngineIntScalarType(
            b_utils._EngineFloat64ArrayType.get_nd(
                x,
                {'fun_name': 'py_fun', 'entity': 'x'},
                exp_rank=2, sorder=_m_order, dim_type='l',
            ),
            {'fun_name': 'py_fun', 'entity': '_ldx'},
        ).to_ctypes()
        _x_cls = b_utils._EngineFloat64ArrayType(
            x,
            {'fun_name': 'py_fun', 'entity': 'x'},
            exp_shape=(5, 2), mutual_sorder=_m_order,
        )
        _x = _x_cls.to_ctypes()
        _ldi = b_utils._EngineIntScalarType(
            b_utils._EngineIntArrayType.get_nd(
                i,
                {'fun_name': 'py_fun', 'entity': 'i'},
                exp_rank=2, sorder=_m_order, dim_type='l',
            ),
            {'fun_name': 'py_fun', 'entity': '_ldi'},
        ).to_ctypes()
        _i_cls = b_utils._EngineIntArrayType(
            i,
            {'fun_name': 'py_fun', 'entity': 'i'},
            exp_shape=(5, 2), mutual_sorder=_m_order,
        )
        _i = _i_cls.to_ctypes()
        _ldl = b_utils._EngineIntScalarType(
            b_utils._EngineLogicalArrayType.get_nd(
                l,
                {'fun_name': 'py_fun', 'entity': 'l'},
                exp_rank=2, sorder=_m_order, dim_type='l',
            ),
            {'fun_name': 'py_fun', 'entity': '_ldl'},
        ).to_ctypes()
        _l_cls = b_utils._EngineLogicalArrayType(
            l,
            {'fun_name': 'py_fun', 'entity': 'l'},
            exp_shape=(2, 3), mutual_sorder=_m_order,
        )
        _l = _l_cls.to_ctypes()
        _cb = self.cb(cb_h)
        if _m_order == 'F':
            _sorder = b_utils._NAG_ED_COL_MAJOR
        elif _m_order == 'C':
            _sorder = b_utils._NAG_ED_ROW_MAJOR
        else:
            raise b_utils.NagException( # pragma: no cover
                {'fun_name': 'py_fun'},
                'Inconceivable sorder ' + str(_m_order)
            )
        _hlperr = b_utils._EngineIntScalarType.empty_ctypes()
        self.set_up_wrapper_data(_m_order)
        self.cast_output_data(_m_order)
        self._retval = self.engine_fun(
            _sorder,
            _hlperr,
            _cstar,
            _ldcstar,
            _c2,
            _ldc2,
            _c,
            _z,
            _ldz,
            _r,
            _ldr,
            _x,
            _ldx,
            _i,
            _ldi,
            _l,
            _ldl,
            self._cstar_to_user,
            self._c2_to_user,
            self._c_to_user,
            self._z_to_user,
            self._r_to_user,
            self._x_to_user,
            self._i_to_user,
            self._l_to_user,
            _cb,
            _cstar_cls.chrlen,
            _c2_cls.chrlen,
            self._cstar_to_user_cls.chrlen,
            self._c2_to_user_cls.chrlen,
        )
        b_utils._handle_cb_exception(excs)
        self.refresh_output_data()
        return (
            self.retval,
            self.cstar_to_user,
            self.c2_to_user,
            self.c_to_user,
            self.z_to_user,
            self.r_to_user,
            self.x_to_user,
            self.i_to_user,
            self.l_to_user,
        )

    def make_and_check_wrapper_call(self, order):
        """
        Call the Python wrappwr with order-contiguous input.
        """
        def cb_py(cstar, c2, c, z, r, x, i, l):
            """
            Test callback.
            """
            self.check_cb_input(
                order, cstar, c2, c, z, r, x, i, l,
            )
            return (
                np.array([
                    ['Hello', 'The', 'World!'], ['Allo', 'Le', 'Monde!']
                ], order=order),
                np.array([[b'He', b'll', b'o'], [b'Wo', b'rl', b'd!']],
                         order=order),
                np.array(
                    [
                        [42. + 43.j, 44. + 45.j, 46. + 47.j],
                        [48. + 49.j, 50. + 51.j, 52. + 53.j]
                    ],
                    dtype=np.complex64, order=order
                ),
                np.array(
                    [
                        [42. + 43.j, 44. + 45.j],
                        [46. + 47.j, 48. + 49.j],
                        [50. + 51.j, 52. + 53.j]
                    ],
                    order=order
                ),
                np.array(
                    [
                        [42., 43.], [44., 45.], [46., 47.], [48., 49.],
                        [50., 51.]
                    ],
                    dtype=np.float32, order=order
                ),
                np.array(
                    [
                        [42., 43.], [44., 45.], [46., 47.], [48., 49.],
                        [50., 51.]
                    ],
                    order=order
                ),
                np.array(
                    [
                        [42, 43], [44, 45], [46, 47], [48, 49], [50, 51]
                    ],
                    dtype=b_utils.EngineIntNumPyType, order=order
                ),
                np.array([[True, False, True], [True, True, False]],
                         order=order),
            )

        (
            retval, cstar, c2, c, z, r, x, i, l,
        ) = self.py_fun(
            np.array(
                ['He', 'll', 'o', 'Wo', 'rl', 'd!'],
            ).reshape((2, 3), order=order),
            np.array(
                [b'He', b'll', b'o', b'Wo', b'rl', b'd!'],
            ).reshape((2, 3), order=order),
            np.array(
                [complex(42 + 2*i, 43 + 2*i) for i in range(6)],
                dtype=np.complex64
            ).reshape((2, 3), order=order),
            np.array(
                [complex(42 + 2*i, 43 + 2*i) for i in range(6)],
                dtype=np.complex128
            ).reshape((3, 2), order=order),
            np.array(
                [42 + i for i in range(10)],
                dtype=np.float32
            ).reshape((5, 2), order=order),
            np.array(
                [42 + i for i in range(10)],
                dtype=np.float64
            ).reshape((5, 2), order=order),
            np.array(
                [42 + i for i in range(10)],
                dtype=np.int32
            ).reshape((5, 2), order=order),
            np.array(
                [True, False, True, False, True, False]
            ).reshape((2, 3), order=order),
            cb_py,
        )

        self.check_py(
            order, retval,
            cstar, c2, c, z, r, x, i, l,
        )

    def test_good_wrapper_c(self):
        """
        Test correct operation for C-contiguous input via a Python wrapper.
        """
        self.make_and_check_wrapper_call('C')

    def test_good_wrapper_f(self):
        """
        Test correct operation for Fortran-contiguous input via a Python
        wrapper.
        """
        self.make_and_check_wrapper_call('F')

class PassTestArrayMultidimIntel(PassTestArrayMultidim, unittest.TestCase):
    """
    Pass-test subsuite for Intel.
    """
    compiler = 'ifort'

    @classmethod
    def setUpClass(cls):
        cls.set_up_multidim(cls.compiler)

class PassTestArrayMultidimNag(PassTestArrayMultidim, unittest.TestCase):
    """
    Pass-test subsuite for NAG.
    """
    compiler = 'nagfor'

    @classmethod
    def setUpClass(cls):
        cls.set_up_multidim(cls.compiler)

class PassTestLogical(PassTest):
    """
    Abstract low-level unit tests for passing Logical values to and from the
    Engine.
    """

    @classmethod
    def set_up_logical(cls, compiler):
        """
        The abstract test setup.
        """
        cls.default_engine_state = save_engine_state()

        cls.engine_state = b_utils._EngineState

        if compiler == 'ifort':
            cls.engine_state._compilers = {'Fortran': 'Intel'}
            cls.engine_state._literal_true_int = -1
        else:
            cls.engine_state._compilers = {'Fortran': 'NAG'}
            cls.engine_state._literal_true_int = 1

        cls.set_up('logical_pass_test', compiler)

        cls.engine_fun.restype = None
        cls.engine_fun.argtypes = (
            b_utils._BYREF_LOGICAL, # Input, l_in.
            b_utils._BYREF_LOGICAL, # Output, not_l_in.
            b_utils._BYREF_LOGICAL, # Output, l_out.
        )

    def test_good_direct_true(self):
        """
        Test correct operation with a direct call supplied 'true'.
        """
        not_l_lin = b_utils._EngineLogicalScalarType.empty_ctypes()
        l_out = b_utils._EngineLogicalScalarType.empty_ctypes()
        self.engine_fun(
            b_utils._EngineLogicalScalarType.c_type(1),
            not_l_lin, l_out,
            )
        self.assertEqual( # pylint: disable=no-member
            not_l_lin.value,
            self.not_one, # pylint: disable=no-member
        )
        self.assertEqual( # pylint: disable=no-member
            l_out.value,
            b_utils._EngineLogicalScalarType(
                False, {'fun_name': 'z99zz', 'entity': 'l'}
            ).to_ctypes().value,
        )

    def test_good_direct_false(self):
        """
        Test correct operation with a direct call supplied 'false'.
        """
        not_l_lin = b_utils._EngineLogicalScalarType.empty_ctypes()
        l_out = b_utils._EngineLogicalScalarType.empty_ctypes()
        self.engine_fun(
            b_utils._EngineLogicalScalarType.c_type(0),
            not_l_lin, l_out,
            )
        self.assertEqual( # pylint: disable=no-member
            not_l_lin.value,
            self.not_zero, # pylint: disable=no-member
        )
        self.assertEqual( # pylint: disable=no-member
            l_out.value,
            b_utils._EngineLogicalScalarType(
                True, {'fun_name': 'z99zz', 'entity': 'l'}
            ).to_ctypes().value,
        )

    # pylint: disable=invalid-name
    def py_fun(self, l_in):
        """
        Python wrapper for self.engine_fun.
        """
        _l_in = b_utils._EngineLogicalScalarType(
            l_in,
            {'fun_name': 'py_fun', 'entity': 'l_in'},
        ).to_ctypes()
        _not_l_in = b_utils._EngineLogicalScalarType.empty_ctypes()
        _l_out = b_utils._EngineLogicalScalarType.empty_ctypes()
        self.engine_fun(
            _l_in,
            _not_l_in, _l_out,
        )
        return (
            b_utils._EngineLogicalScalarType.to_py(_not_l_in),
            b_utils._EngineLogicalScalarType.to_py(_l_out),
        )

    def test_good_wrapper_true(self):
        """
        Test correct operation for 'true' supplied via a Python wrapper.
        """
        not_l_in, l_out = self.py_fun(True)

        self.assertFalse( # pylint: disable=no-member
            not_l_in,
        )
        self.assertFalse( # pylint: disable=no-member
            l_out,
        )

    def test_good_wrapper_false(self):
        """
        Test correct operation for 'false' supplied via a Python wrapper.
        """
        not_l_in, l_out = self.py_fun(False)

        self.assertTrue( # pylint: disable=no-member
            not_l_in,
        )
        self.assertTrue( # pylint: disable=no-member
            l_out,
        )

class PassTestLogicalNag(PassTestLogical, unittest.TestCase):
    """
    Pass-test subsuite for NAG.
    """
    compiler = 'nagfor'

    @classmethod
    def setUpClass(cls):
        cls.set_up_logical(cls.compiler)
        cls.not_one = 0
        cls.not_zero = 1

class PassTestLogicalIntel(PassTestLogical, unittest.TestCase):
    """
    Pass-test subsuite for Intel.
    """
    compiler = 'ifort'

    @classmethod
    def setUpClass(cls):
        cls.set_up_logical(cls.compiler)
        cls.not_one = ~1
        cls.not_zero = ~0

class PassTestPrintRec(PassTest):
    """
    Abstract low-level unit tests for record output.
    """

    @classmethod
    def set_up_prt(cls, compiler):
        """
        The abstract test setup.
        """
        cls.default_engine_state = save_engine_state()

        cls.engine_state = b_utils._EngineState

        if compiler == 'ifort':
            cls.engine_state._compilers = {'Fortran': 'Intel'}
        else:
            cls.engine_state._compilers = {'Fortran': 'NAG'}

        cls.set_up('print_rec_test', compiler, deps=True)

        cls.engine_state._chrlen_ctype = cls.compiler.chrlen_ctype

        print_rec_args = list(b_utils._PRINT_REC_ARGS)
        print_rec_args[-1] = cls.engine_state._chrlen_ctype

        cls.print_rec_functype = b_utils._EngineState._sys_functype(
            None, *tuple(print_rec_args)
        )
        cls.print_rech_functype = b_utils._EngineState._sys_functype(
            None,
            cls.print_rec_functype,
            *tuple(print_rec_args)
        )

        cls.engine_fun.restype = None
        cls.engine_fun.argtypes = (
            cls.print_rech_functype,
            cls.print_rec_functype,
            b_utils._BYREF_ENGINEDATA,
            b_utils._BYREF_INT, # Input, nout.
            b_utils._BYREF_INT, # Output, ierr.
        )

    def py_fun(self, unit, io_manager=None):
        """
        Python wrapper.
        """
        excs = []
        fun_name = 'py_fun'

        def cb_print_rech_py(_, en_data, nout, rec, ifail, rec_len):
            # pylint: disable=bare-except,missing-docstring
            try:
                b_utils._py_print_rec(
                    en_data,
                    b_utils._EngineIntScalarType.to_py(nout.contents),
                    rec[:rec_len],
                    io_manager,
                    excs,
                    ifail.contents,
                    fun_name=fun_name,
                )
            except: # pragma: no cover
                b_utils._print_err(
                    sys.exc_info(),
                    excs,
                    en_data,
                    ifail.contents,
                )

        _print_rec = b_utils._cast_callback(
            self.print_rec_functype, None,
            {'fun_name': fun_name}
        )
        _print_rech = self.print_rech_functype(cb_print_rech_py)
        _unit = b_utils._EngineIntScalarType(
            unit, {'fun_name': fun_name, 'entity': 'unit'}
        ).to_ctypes()
        b_utils.FileObjManager._check_type_compat(
            io_manager,
            {'fun_name': fun_name, 'entity': 'io_manager'},
        )
        _ierr = b_utils._EngineIntScalarType.empty_ctypes()
        _en_data = b_utils._EngineData()
        self.engine_fun(
            _print_rech,
            _print_rec,
            _en_data,
            _unit,
            _ierr,
        )
        if excs: # pragma: no cover
            return excs[0], b_utils._EngineIntScalarType.to_py(_ierr)

        return None, b_utils._EngineIntScalarType.to_py(_ierr)

    def test_good(self):
        """
        Correct operation.
        """
        tf_fo = tempfile.NamedTemporaryFile()
        fom = b_utils.FileObjManager((tf_fo.file,))

        this_exc, ierr = self.py_fun(
            fom.unit_from_fileobj(tf_fo.file), io_manager=fom
        )

        self.assertIsNone(this_exc) # pylint: disable=no-member
        self.assertEqual(ierr, 0) # pylint: disable=no-member

        tf_fo.seek(0)
        self.assertEqual( # pylint: disable=no-member
            tf_fo.readlines(), [b'py_fun: Hello World!\n']
        )

        tf_fo.close()

class PassTestPrintRecNag(PassTestPrintRec, unittest.TestCase):
    """
    Pass-test subsuite for NAG.
    """
    compiler = 'nagfor'

    @classmethod
    def setUpClass(cls):
        cls.set_up_prt(cls.compiler)

class PassTestPrintRecIntel(PassTestPrintRec, unittest.TestCase):
    """
    Pass-test subsuite for Intel.
    """
    compiler = 'ifort'

    @classmethod
    def setUpClass(cls):
        cls.set_up_prt(cls.compiler)

class PassTestReadRec(PassTest):
    """
    Abstract low-level unit tests for record input.
    """

    @classmethod
    def set_up_read(cls, compiler):
        """
        The abstract test setup.
        """
        cls.default_engine_state = save_engine_state()

        cls.engine_state = b_utils._EngineState

        if compiler == 'ifort':
            cls.engine_state._compilers = {'Fortran': 'Intel'}
        else:
            cls.engine_state._compilers = {'Fortran': 'NAG'}

        cls.set_up('read_rec_test', compiler, deps=True)

        cls.engine_state._chrlen_ctype = cls.compiler.chrlen_ctype

        read_rec_args = list(b_utils._READ_REC_ARGS)
        read_rec_args[-2:] = [
            cls.engine_state._chrlen_ctype, cls.engine_state._chrlen_ctype
        ]

        cls.read_rec_functype = b_utils._EngineState._sys_functype(
            None, *tuple(read_rec_args)
        )
        cls.read_rech_functype = b_utils._EngineState._sys_functype(
            None,
            cls.read_rec_functype,
            *tuple(read_rec_args)
        )

        cls.engine_fun.restype = b_utils._EngineIntScalarType.c_type
        cls.engine_fun.argtypes = (
            cls.read_rech_functype,
            cls.read_rec_functype,
            b_utils._BYREF_ENGINEDATA,
            b_utils._BYREF_INT, # Input, nin.
            b_utils._BYREF_INT, # Output, ierr.
        )

    def py_fun(self, unit, io_manager=None):
        """
        Python wrapper.
        """
        excs = []

        def cb_read_rech_py( # pylint: disable=too-many-arguments
                read_rec, en_data, nin, recin, errbuf, ifail, recin_len,
                errbuf_len
        ):
            # pylint: disable=bare-except,missing-docstring
            try:
                b_utils._py_read_rec(
                    en_data,
                    b_utils._EngineIntScalarType.to_py(nin.contents),
                    recin,
                    errbuf,
                    recin_len,
                    io_manager,
                    excs,
                    ifail.contents,
                )
            except: # pragma: no cover
                b_utils._read_err(
                    sys.exc_info(),
                    excs,
                    b_utils.FileObjManager._read_err,
                    en_data,
                    ifail.contents,
                )

        _read_rec = b_utils._cast_callback(
            self.read_rec_functype, None,
            {'fun_name': 'py_fun'}
        )
        _read_rech = self.read_rech_functype(cb_read_rech_py)
        _unit = b_utils._EngineIntScalarType(
            unit, {'fun_name': 'py_fun', 'entity': 'unit'}
        ).to_ctypes()
        b_utils.FileObjManager._check_type_compat(
            io_manager,
            {'fun_name': 'py_fun', 'entity': 'io_manager'},
        )
        _ierr = b_utils._EngineIntScalarType.empty_ctypes()
        _en_data = b_utils._EngineData()
        _rtn = self.engine_fun(
            _read_rech,
            _read_rec,
            _en_data,
            _unit,
            _ierr,
        )
        if excs: # pragma: no cover
            return _rtn, excs[0], b_utils._EngineIntScalarType.to_py(_ierr)

        return _rtn, None, b_utils._EngineIntScalarType.to_py(_ierr)

    def test_good(self):
        """
        Correct operation.
        """
        tf_fo = tempfile.NamedTemporaryFile()
        tf_fo.write(b'42\n')
        tf_fo.seek(0)

        fom = b_utils.FileObjManager((tf_fo.file,))

        read_val, this_exc, ierr = self.py_fun(
            fom.unit_from_fileobj(tf_fo.file), io_manager=fom
        )

        self.assertEqual(read_val, 42) # pylint: disable=no-member
        self.assertIsNone(this_exc) # pylint: disable=no-member
        self.assertEqual(ierr, 0) # pylint: disable=no-member

        tf_fo.close()

class PassTestReadRecNag(PassTestReadRec, unittest.TestCase):
    """
    Pass-test subsuite for NAG.
    """
    compiler = 'nagfor'

    @classmethod
    def setUpClass(cls):
        cls.set_up_read(cls.compiler)

class PassTestReadRecIntel(PassTestReadRec, unittest.TestCase):
    """
    Pass-test subsuite for Intel.
    """
    compiler = 'ifort'

    @classmethod
    def setUpClass(cls):
        cls.set_up_read(cls.compiler)

class PassTestNonadvReadRec(PassTest):
    """
    Abstract low-level unit tests for nonadvancing record input.
    """

    @classmethod
    def set_up_nonadv_read(cls, compiler):
        """
        The abstract test setup.
        """
        cls.default_engine_state = save_engine_state()

        cls.engine_state = b_utils._EngineState

        if compiler == 'ifort':
            cls.engine_state._compilers = {'Fortran': 'Intel'}
        else:
            cls.engine_state._compilers = {'Fortran': 'NAG'}

        cls.set_up('nonadv_read_rec_test', compiler, deps=True)

        cls.engine_state._chrlen_ctype = cls.compiler.chrlen_ctype

        nonadv_read_rec_args = list(b_utils._NONADV_READ_REC_ARGS)
        nonadv_read_rec_args[-1] = cls.engine_state._chrlen_ctype

        cls.nonadv_read_rec_functype = b_utils._EngineState._sys_functype(
            None, *tuple(nonadv_read_rec_args)
        )
        cls.nonadv_read_rech_functype = b_utils._EngineState._sys_functype(
            None,
            cls.nonadv_read_rec_functype,
            *tuple(nonadv_read_rec_args)
        )

        cls.engine_fun.restype = b_utils._EngineIntScalarType.c_type
        cls.engine_fun.argtypes = (
            cls.nonadv_read_rech_functype,
            cls.nonadv_read_rec_functype,
            b_utils._BYREF_ENGINEDATA,
            b_utils._BYREF_INT, # Input, nin.
            b_utils._BYREF_INT, # Output, ierr.
        )

    def py_fun(self, unit, io_manager=None):
        """
        Python wrapper.
        """
        excs = []

        def cb_nonadv_read_rech_py( # pylint: disable=line-too-long,too-many-arguments
                nonadv_read_rec, en_data, nin, recin, readlen, inform,
                recin_len
        ):
            # pylint: disable=bare-except,missing-docstring
            try:
                readlen.contents.value = b_utils._py_nonadv_read_rec(
                    en_data,
                    b_utils._EngineIntScalarType.to_py(nin.contents),
                    recin,
                    recin_len,
                    io_manager,
                    excs,
                    inform.contents,
                )
            except: # pragma: no cover
                readlen.contents.value = b_utils._nonadv_read_err(
                    sys.exc_info(),
                    excs,
                    b_utils.FileObjManager._read_err,
                    en_data,
                    inform.contents,
                )

        _nonadv_read_rec = b_utils._cast_callback(
            self.nonadv_read_rec_functype, None,
            {'fun_name': 'py_fun'}
        )
        _nonadv_read_rech = self.nonadv_read_rech_functype(
            cb_nonadv_read_rech_py
        )
        _unit = b_utils._EngineIntScalarType(
            unit, {'fun_name': 'py_fun', 'entity': 'unit'}
        ).to_ctypes()
        b_utils.FileObjManager._check_type_compat(
            io_manager,
            {'fun_name': 'py_fun', 'entity': 'io_manager'},
        )
        _ierr = b_utils._EngineIntScalarType.empty_ctypes()
        _en_data = b_utils._EngineData()
        _rtn = self.engine_fun(
            _nonadv_read_rech,
            _nonadv_read_rec,
            _en_data,
            _unit,
            _ierr,
        )
        if excs: # pragma: no cover
            return _rtn, excs[0], b_utils._EngineIntScalarType.to_py(_ierr)

        return _rtn, None, b_utils._EngineIntScalarType.to_py(_ierr)

    def test_good(self):
        """
        Correct operation.
        """
        tf_fo = tempfile.NamedTemporaryFile()
        tf_fo.write(b'42\n')
        tf_fo.seek(0)

        fom = b_utils.FileObjManager((tf_fo.file,))

        read_val, this_exc, ierr = self.py_fun(
            fom.unit_from_fileobj(tf_fo.file), io_manager=fom
        )

        self.assertEqual(read_val, 42) # pylint: disable=no-member
        self.assertIsNone(this_exc) # pylint: disable=no-member
        self.assertEqual(ierr, 0) # pylint: disable=no-member

        tf_fo.close()

class PassTestNonadvReadRecNag(PassTestNonadvReadRec, unittest.TestCase):
    """
    Pass-test subsuite for NAG.
    """
    compiler = 'nagfor'

    @classmethod
    def setUpClass(cls):
        cls.set_up_nonadv_read(cls.compiler)

class PassTestNonadvReadRecIntel(PassTestNonadvReadRec, unittest.TestCase):
    """
    Pass-test subsuite for Intel.
    """
    compiler = 'ifort'

    @classmethod
    def setUpClass(cls):
        cls.set_up_nonadv_read(cls.compiler)

class PassTestSubroutine(PassTest):
    """
    Abstract low-level unit tests for calling an Engine subroutine.
    """

    @classmethod
    def set_up_subr(cls, compiler):
        """
        The abstract test setup.
        """
        cls.default_engine_state = save_engine_state()

        cls.engine_state = b_utils._EngineState

        cls.set_up('subr_call_test', compiler)

        cls.cb = b_utils._EngineState._sys_functype(
            None,
            b_utils._BYREF_INT, # Output, j.
        )

        cls.engine_fun.restype = None
        cls.engine_fun.argtypes = (
            b_utils._BYREF_INT, # Input, i.
            cls.cb,
        )

    @staticmethod
    def cb_s(j):
        """
        Callback for testing.
        """
        b_utils._EngineIntScalarType(
            42,
            {'fun_name': 'cb', 'returned': 'j'},
        ).refresh_ctypes_buffer(j)

    def test_good_direct(self):
        """
        Test correct operation with a direct call.
        """

        self.engine_fun(
            b_utils._EngineIntScalarType.c_type(42),
            self.cb(self.cb_s),
        )

    def test_good_wrapper(self):
        """
        Test correct operation via a Python wrapper.
        """

        def cb_py():
            """
            Test callback.
            """
            return 42

        # pylint: disable=invalid-name
        def py_fun(i, cb):
            """
            Python wrapper for self.engine_fun.
            """

            def cb_s(j):
                """
                Test helper.
                """
                _j = cb()
                b_utils._EngineIntScalarType(
                    _j,
                    {'fun_name': 'cb', 'returned': 'j'},
                ).refresh_ctypes_buffer(j)

            _i = b_utils._EngineIntScalarType(
                i,
                {'fun_name': 'py_fun', 'entity': 'i'}
            ).to_ctypes()
            _cb = self.cb(cb_s)
            return self.engine_fun(
                _i,
                _cb,
            )

        py_fun(
            42,
            cb_py,
        )

class PassTestSubroutineNag(PassTestSubroutine, unittest.TestCase):
    """
    Pass-test subsuite for NAG.
    """
    compiler = 'nagfor'

    @classmethod
    def setUpClass(cls):
        cls.set_up_subr(cls.compiler)

class PassTestSubroutineIntel(PassTestSubroutine, unittest.TestCase):
    """
    Pass-test subsuite for Intel.
    """
    compiler = 'ifort'

    @classmethod
    def setUpClass(cls):
        cls.set_up_subr(cls.compiler)

class ReturnTestErrbuf(PassTest):
    """
    Abstract low-level unit tests for calling an Engine subroutine.
    """

    @classmethod
    def set_up_eb(cls, compiler):
        """
        The abstract test setup.
        """
        cls.default_engine_state = save_engine_state()

        cls.engine_state = b_utils._EngineState

        if compiler == 'ifort':
            cls.engine_state._compilers = {'Fortran': 'Intel'}
        else:
            cls.engine_state._compilers = {'Fortran': 'NAG'}

        cls.set_up('errbuf_rtn_test', compiler, deps=True)

        cls.engine_state._chrlen_ctype = cls.compiler.chrlen_ctype

        cls.engine_fun.restype = None
        cls.engine_fun.argtypes = (
            b_utils._BYREF_INT, # Input, i.
            b_utils._EngineErrbuf,
            cls.engine_state._chrlen_ctype, # Input, chrlen of errbuf
        )

    def run_and_split(self, i):
        """
        Return the split errbuf list.
        """
        errbuf = b_utils._EngineErrbuf()
        self.engine_fun(
            b_utils._EngineIntScalarType.c_type(i),
            errbuf,
            errbuf._length_,
        )
        return b_utils._EngineState._split_errbuf(errbuf)

    def test_bad_null(self):
        """
        Splitting null errbuf rejected.
        """
        test_case = 0
        self.assertRaises( # pylint: disable=no-member
            b_utils.NagException,
            self.run_and_split, test_case
        )

    def test_bad_too_short(self):
        """
        Splitting short errbuf rejected.
        """
        test_case = 2
        self.assertRaises( # pylint: disable=no-member
            b_utils.NagException,
            self.run_and_split, test_case
        )

    def test_good_contents(self):
        """
        Correct operation.
        """
        test_case = 1
        exp_eb = ['1', '1005', '1']
        eb_split = self.run_and_split(test_case)
        self.assertEqual( # pylint: disable=no-member
            eb_split[:-1], exp_eb
        )
        self.assertEqual( # pylint: disable=no-member
            float(eb_split[-1]), 0.0
        )

    def test_local(self):
        """
        Correct operation with a local exit flag.
        """
        errbuf = b_utils._EngineErrbuf()
        self.engine_fun(
            b_utils._EngineIntScalarType.c_type(1),
            errbuf,
            errbuf._length_,
        )
        ifmt, eb_data = b_utils._EngineState._handle_global_errors(
            'z99zz', errbuf, 1
        )
        self.assertEqual( # pylint: disable=no-member
            ifmt, 1005,
        )
        self.assertEqual( # pylint: disable=no-member
            len(eb_data), 2,
        )
        self.assertEqual( # pylint: disable=no-member
            eb_data[0], '1',
        )
        self.assertEqual( # pylint: disable=no-member
            float(eb_data[1]), 0.0,
        )

class ReturnTestErrbufNag(ReturnTestErrbuf, unittest.TestCase):
    """
    Pass-test subsuite for NAG.
    """
    compiler = 'nagfor'

    @classmethod
    def setUpClass(cls):
        cls.set_up_eb(cls.compiler)

class ReturnTestErrbufIntel(ReturnTestErrbuf, unittest.TestCase):
    """
    Pass-test subsuite for Intel.
    """
    compiler = 'ifort'

    @classmethod
    def setUpClass(cls):
        cls.set_up_eb(cls.compiler)

if __name__ == '__main__': # pragma: no cover
    unittest.main()
