#!/usr/bin/env python
"""
Integration tests for wrapper functionality in the fit submodule.

Public use of this module is not supported by NAG.
"""

# NAG Copyright 2018-2019.

# pylint: disable=invalid-name

import unittest

import numpy as np

from naginterfaces.base import utils as b_utils
from naginterfaces.base import fit as b_chapter
from naginterfaces.library import fit as chapter
from test_functionality_core import IntegrationTest

class TestE02ag(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the e02ag wrapper.
    """

class TestE02agNp(TestE02ag, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestE02ag.set_up(self)
        self.fun = chapter.dim1_cheb_con

    def test_good(self):
        """
        Good call.
        """
        x = [
            0.5,
            1.,
            2.,
            2.5,
            3.,
        ]
        y = [
            0.03,
            -0.75,
            -1.,
            -0.1,
            1.75,
        ]
        w = [1.]*len(x)
        fit = self.fun(
            4, 0., 4., x, y, w, [0., 4.], [1., -2., 9.], [1, 0],
        )
        np.testing.assert_allclose(
            fit.resid,
            [
                -1.0178817895330017E-03,
                7.8115956260072633E-04,
                2.0048292236949624E-03,
                -3.9464479467876167E-03,
                2.1673870655112646E-03,
            ],
            rtol=1e-5,
        )

class TestE02be(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the e02be wrapper.
    """

class TestE02beNp(TestE02be, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestE02be.set_up(self)
        self.fun = chapter.dim1_spline_auto

    def test_good(self):
        """
        Good call.
        """
        x = [
            0.0, 0.5, 1.0, 1.5, 2.0, 2.5, 3.0, 4.0, 4.5, 5.0, 5.5,
            6.0, 7.0, 7.5, 8.0,
        ]
        y = [
            -1.1, -0.372, 0.431, 1.69, 2.11, 3.1, 4.23, 4.35, 4.81,
            4.61, 4.79, 5.23, 6.35, 7.19, 7.97,
        ]
        w = [
            1.00, 2.00, 1.50, 1.00, 3.00, 1.00, 0.50, 1.00, 2.00,
            2.50, 1.00, 3.00, 1.00, 2.00, 1.00,
        ]
        lamda = np.empty(len(x)+7)
        comm = {}
        s = self.fun('C', x, y, w, 1.0, 42, lamda, comm)
        self.assertEqual(s.n, 9)
        np.testing.assert_equal(
            s.lamda[3:s.n-3], [0., 4., 8.],
        )

class TestE02bf(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the e02bf wrapper.
    """

class TestE02bfNp(TestE02bf, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestE02bf.set_up(self)
        self.fitter = chapter.dim1_spline_auto
        self.fun = chapter.dim1_spline_deriv_vector

    def test_good(self):
        """
        Good call.
        """
        x = [
            0.0, 0.5, 1.0, 1.5, 2.0, 2.5, 3.0, 4.0, 4.5, 5.0, 5.5,
            6.0, 7.0, 7.5, 8.0,
        ]
        y = [
            -1.1, -0.372, 0.431, 1.69, 2.11, 3.1, 4.23, 4.35, 4.81,
            4.61, 4.79, 5.23, 6.35, 7.19, 7.97,
        ]
        w = [
            1.00, 1.00, 1.50, 1.00, 1.00, 1.00, 1.00, 1.00, 1.00,
            1.00, 1.00, 1.00, 1.00, 1.00, 1.00,
        ]
        lamda = np.empty(len(x)+4)
        comm = {}
        s = self.fitter('C', x, y, w, 0.001, 42, lamda, comm)
        x_eval = [
            6.5178, 7.2463, 1.0159, 7.3070,
            5.0589, 0.7803, 2.2280, 4.3751,
            7.6601, 7.7191, 1.2609, 7.7647,
            7.6573, 3.8830, 6.4022, 1.1351,
            3.3741, 7.3259, 6.3377, 7.6759,
        ]
        ixloc = np.zeros(len(x_eval), dtype=int)
        evald = self.fun(
            0, s.lamda[:s.n+7], s.c[:s.n+7], 3, 0, x_eval, ixloc, iwrk=None,
        )
        self.assertEqual(evald.ixloc[0], 14)
        self.assertAlmostEqual(evald.s[0, 0], 5.7418, places=4)
        self.assertIsNotNone(evald.iwrk)

class TestE02da(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the e02da wrapper.
    """

class TestE02daNp(TestE02da, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestE02da.set_up(self)
        self.fun = chapter.dim2_spline_panel
        self.sort = chapter.dim2_spline_sort

    def test_good(self):
        """
        Good call.
        """
        y = [
            -0.52,
            -0.61,
            0.93,
            0.09,
            0.88,
            -0.70,
            1.00,
            1.00,
            0.30,
            -0.77,
            -0.23,
            -1.00,
            -0.26,
            -0.83,
            0.22,
            0.89,
            -0.80,
            -0.88,
            0.68,
            -0.14,
            0.67,
            -0.90,
            -0.84,
            0.84,
            0.15,
            -0.91,
            -0.35,
            -0.16,
            -0.35,
            -1.00
        ]
        x = [
            0.60,
            -0.95,
            0.87,
            0.84,
            0.17,
            -0.87,
            1.00,
            0.10,
            0.24,
            -0.77,
            0.32,
            1.00,
            -0.63,
            -0.66,
            0.93,
            0.15,
            0.99,
            -0.54,
            0.44,
            -0.72,
            0.63,
            -0.40,
            0.20,
            0.43,
            0.28,
            -0.24,
            0.86,
            -0.41,
            -0.05,
            -1.00,
        ]
        f = [
            0.93,
            -1.79,
            0.36,
            0.52,
            0.49,
            -1.76,
            0.33,
            0.48,
            0.65,
            -1.82,
            0.92,
            1.00,
            8.88,
            -2.01,
            0.47,
            0.49,
            0.84,
            -2.42,
            0.47,
            7.15,
            0.44,
            -3.34,
            2.78,
            0.44,
            0.70,
            -6.52,
            0.66,
            2.32,
            1.66,
            -1.00,
        ]
        w = [
            10.,
            10.,
            10.,
            10.,
            10.,
            10.,
            1.,
            1.,
            1.,
            1.,
            1.,
            1.,
            1.,
            1.,
            1.,
            1.,
            1.,
            1.,
            1.,
            1.,
            1.,
            1.,
            1.,
            1.,
            1.,
            1.,
            1.,
            1.,
            1.,
            1.,
        ]
        px = 10
        lamda = [-42.]*px
        lamda[4:px-4] = [-0.5, 0.]
        py = 8
        mu = [-42.]*py
        eps = 0.000001
        fit = self.fun(x, y, f, w, lamda, mu, self.sort(lamda, mu, x, y), eps)
        self.assertAlmostEqual(fit.sigma, 14.66711072)
        self.assertEqual(fit.rank, 22)

class TestE02dd(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the e02dd wrapper.
    """

class TestE02ddBase(TestE02dd, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestE02dd.set_up(self)
        self.fun = chapter.dim2_spline_sctr

    def test_good(self):
        """
        Good call.
        """
        start = 'C'
        x = [
            11.16,
            12.85,
            19.85,
            19.72,
            15.91,
            0.00,
            20.87,
            3.45,
            14.26,
            17.43,
            22.80,
            7.58,
            25.00,
            0.00,
            9.66,
            5.22,
            17.25,
            25.00,
            12.13,
            22.23,
            11.52,
            15.20,
            7.54,
            17.32,
            2.14,
            0.51,
            22.69,
            5.47,
            21.67,
            3.31,
        ]
        y = [
            1.24,
            3.06,
            10.72,
            1.39,
            7.74,
            20.00,
            20.00,
            12.78,
            17.87,
            3.46,
            12.39,
            1.98,
            11.87,
            0.00,
            20.00,
            14.66,
            19.57,
            3.87,
            10.79,
            6.21,
            8.53,
            0.00,
            10.69,
            13.78,
            15.03,
            8.37,
            19.63,
            17.13,
            14.36,
            0.33,
        ]
        f = [
            22.15,
            22.11,
            7.97,
            16.83,
            15.30,
            34.60,
            5.74,
            41.24,
            10.74,
            18.60,
            5.47,
            29.87,
            4.40,
            58.20,
            4.73,
            40.36,
            6.43,
            8.74,
            13.71,
            10.25,
            15.74,
            21.60,
            19.31,
            12.11,
            53.10,
            49.43,
            3.25,
            28.63,
            5.52,
            44.08,
        ]
        w = [1.]*len(f)
        s = 10.
        nx = 42
        lamda = [-42.]*nx
        ny = 42
        mu = [-42.]*ny
        comm = {}
        fit = self.fun(start, x, y, f, w, s, nx, lamda, ny, mu, comm)
        np.testing.assert_allclose(
            fit.lamda[3:7], [0., 9.7575, 18.2582, 25.], rtol=1e-5,
        )

class TestE02jd(IntegrationTest):
    """
    Abstract suite of integration tests for the e02jd wrapper.
    """

    # pylint: disable=too-few-public-methods

    exp_a = [1., 0.5, 0.10714286, 1.1904762e-2, 5.952381e-4]
    exp_b = [1., -0.5, 0.10714286, -1.1904762e-2, 5.952381e-4]

class TestE02jdBase(TestE02jd, unittest.TestCase):
    """
    Concrete base tests.
    """

    def setUp(self):
        TestE02jd.set_up(self)
        self.fun = b_chapter.dim2_spline_ts_sctr
        self.init = b_chapter.opt_set

    def test_bad_n(self):
        """
        Invalid n.

        Coverage for: fl_ifail/ierr_fl.
        """
        comm = {}
        self.init('Initialize = dim2_spline_ts_sctr', comm)
        self.my_assert_raises_regex(
            b_utils.NagValueError,
            '.*code 2:101',
            self.fun,
            1, [0.], [0.], [0.], 1, 1, 1, 1, comm,
        )

class TestE02ra(IntegrationTest):
    """
    Abstract suite of integration tests for the e02ra wrapper.
    """

    # pylint: disable=too-few-public-methods

    exp_a = [1., 0.5, 0.10714286, 1.1904762e-2, 5.952381e-4]
    exp_b = [1., -0.5, 0.10714286, -1.1904762e-2, 5.952381e-4]

class TestE02raBase(TestE02ra, unittest.TestCase):
    """
    Concrete base tests.
    """

    def setUp(self):
        TestE02ra.set_up(self)
        self.fun = b_chapter.pade_app

    def test_good(self):
        """
        Good call.
        """
        l = 4
        m = 4
        c = [1.]
        for i in range(1, (l + 1) + (m + 1) - 1):
            c.append(c[i-1]/i)
        a = [None]*(l + 1)
        b = [None]*(m + 1)
        self.fun(l + 1, m + 1, c, l + m + 1, a, b)
        np.testing.assert_allclose(
            a, self.exp_a,
        )
        np.testing.assert_allclose(
            b, self.exp_b,
        )

class TestE02raNp(TestE02ra, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestE02ra.set_up(self)
        self.fun = chapter.pade_app

    def test_good(self):
        """
        Good call.
        """
        l = 4
        m = 4
        c = [1.]
        for i in range(1, (l + 1) + (m + 1) - 1):
            c.append(c[i-1]/i)
        a, b = self.fun(l + 1, m + 1, c)
        np.testing.assert_allclose(
            a, self.exp_a,
        )
        np.testing.assert_allclose(
            b, self.exp_b,
        )

if __name__ == '__main__': # pragma: no cover
    unittest.main()
