#!/usr/bin/env python
"""
Integration tests for wrapper functionality in the glopt submodule.

Public use of this module is not supported by NAG.
"""

# NAG Copyright 2018-2019.

# pylint: disable=invalid-name

from math import cos, exp, sin
import unittest

import numpy as np

from naginterfaces.base import utils as b_utils
from naginterfaces.library import glopt as chapter
from naginterfaces.library import rand
from test_functionality_core import IntegrationTest

class TestE05jb(IntegrationTest):
    """
    Abstract suite of integration tests for the e05jb wrapper.
    """

    # pylint: disable=too-few-public-methods

    fun = lambda *args: None # pragma: no cover
    init = lambda *args: None # pragma: no cover

    @staticmethod
    def peaks_objfun(x, _nstate):
        # pylint: disable=missing-docstring
        return (
            3.*(1. - x[0])**2*exp(-x[0]**2 - (x[1] + 1.)**2)
            - (10.*(x[0]/5. - x[0]**3 - x[1]**5)*exp(-x[0]**2 - x[1]**2))
            - 1./3.0*exp(-(x[0] + 1.)**2 - x[1]**2)
        )

    @staticmethod
    def peaks_data():
        # pylint: disable=missing-docstring
        ibound = 0
        bl = [-3., -3.]
        bu = [3., 3.]
        return ibound, bl, bu

class TestE05jbNp(TestE05jb, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestE05jb.set_up(self)
        self.fun = chapter.bnd_mcs_solve
        self.init = chapter.bnd_mcs_init
        self.optset = chapter.bnd_mcs_optset_char

    def test_default_init(self):
        """
        iinit = None.
        """
        ibound, bl, bu = self.peaks_data()
        comm = self.init()
        result = self.fun(self.peaks_objfun, ibound, bl, bu, comm)
        self.assertEqual(list(result.numpts), [3, 3])
        self.assertEqual(list(result.initpt), [2, 2])
        self.assertAlmostEqual(result.obj, -6.5511333)

    def test_bndary_init(self):
        """
        iinit = 0.
        """
        ibound, bl, bu = self.peaks_data()
        comm = self.init()
        result = self.fun(self.peaks_objfun, ibound, bl, bu, comm, iinit=0)
        self.assertEqual(list(result.numpts), [3, 3])
        self.assertEqual(list(result.initpt), [2, 2])
        self.assertAlmostEqual(result.obj, -6.5511333)

    def test_offbndary_init(self):
        """
        iinit = 1.
        """
        ibound, bl, bu = self.peaks_data()
        comm = self.init()
        result = self.fun(self.peaks_objfun, ibound, bl, bu, comm, iinit=1)
        self.assertEqual(list(result.numpts), [3, 3])
        self.assertEqual(list(result.initpt), [2, 2])
        self.assertAlmostEqual(result.obj, -3.0498494)

    def test_linesearch_init(self):
        """
        iinit = 2.
        """
        ibound, bl, bu = self.peaks_data()
        comm = self.init()
        result = self.fun(self.peaks_objfun, ibound, bl, bu, comm, iinit=2)
        self.assertEqual(list(result.numpts), [7, 7])
        self.assertEqual(list(result.initpt), [4, 4])
        self.assertAlmostEqual(result.obj, -3.0498494)

    def test_supplied_init_no_inlist(self):
        """
        iinit = 3, inlist not supplied.
        """
        ibound, bl, bu = self.peaks_data()
        comm = self.init()
        self.my_assert_raises_regex(
            b_utils.NagTypeError,
            '.*inlist. must not be None',
            self.fun,
            self.peaks_objfun, ibound, bl, bu, comm, iinit=3,
        )

    def test_supplied_init_no_numpts(self):
        """
        iinit = 3, numpts not supplied.
        """
        ibound, bl, bu = self.peaks_data()
        comm = self.init()
        inlist = [
            [bl[0], 0., bu[0]],
            [bl[1], 0., bu[1]],
        ]
        self.my_assert_raises_regex(
            b_utils.NagTypeError,
            '.*numpts. must not be None',
            self.fun,
            self.peaks_objfun, ibound, bl, bu, comm,
            iinit=3, inlist=inlist,
        )

    def test_supplied_init_no_initpt(self):
        """
        iinit = 3, initpt not supplied.
        """
        ibound, bl, bu = self.peaks_data()
        comm = self.init()
        inlist = [
            [bl[0], 0., bu[0]],
            [bl[1], 0., bu[1]],
        ]
        numpts = [3, 3]
        self.my_assert_raises_regex(
            b_utils.NagTypeError,
            '.*initpt. must not be None',
            self.fun,
            self.peaks_objfun, ibound, bl, bu, comm,
            iinit=3, inlist=inlist, numpts=numpts,
        )

    def supplied_init(self, order):
        """
        iinit = 3 test.
        """
        ibound, bl, bu = self.peaks_data()
        comm = self.init()
        inlist = np.array(
            [
                [bl[0], 0., bu[0]],
                [bl[1], 0., bu[1]],
            ],
            order=order,
        )
        numpts = [3, 3]
        initpt = [2, 2]
        result = self.fun(
            self.peaks_objfun, ibound, bl, bu, comm,
            iinit=3, inlist=inlist, numpts=numpts, initpt=initpt,
        )
        self.assertAlmostEqual(result.obj, -6.5511333)

    def test_supplied_init_c(self):
        """
        iinit = 3, row order.
        """
        self.supplied_init('C')

    def test_supplied_init_f(self):
        """
        iinit = 3, column order.
        """
        self.supplied_init('F')

    def test_random_init(self):
        """
        iinit = 4.
        """
        ibound, bl, bu = self.peaks_data()
        comm = self.init()
        self.optset('Repeatability', 'On', comm)
        result = self.fun(
            self.peaks_objfun, ibound, bl, bu, comm, iinit=4, sdlist=10,
        )
        self.assertEqual(result.inlist.shape[1], 10)
        self.assertAlmostEqual(result.obj, -6.5511333)

class TestE05uc(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the e05uc wrapper.
    """

class TestE05ucNp(TestE05uc, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestE05uc.set_up(self)
        self.fun = chapter.nlp_multistart_sqp

        comm = {}
        chapter.optset('Initialize = nlp_multistart_sqp', comm)

        self.a = [[3., -2.]]
        self.args = [
            2, # n, 0
            2, # ncnln, 1
            [-500., -500., -10000., -1., -0.9], # bl, 2
            [500., 500., 10., 500000., 0.9], # bu, 3
            self.cb_objfun, # objfun, 4
            10, # npts, 5
            True, # repeat, 6
            1, # nb, 7
            comm, # comm, 9
        ]

    @staticmethod
    def cb_confun(x, data):
        # pylint: disable=missing-docstring
        c = [
            x[0]**2 - x[1]**2 + 3.*x[0]*x[1],
            cos((x[0]/200.)**2+(x[1]/100.)),
        ]
        cjsl = np.array(
            [
                [
                    2.*x[0] + 3.*x[1],
                    -2.*x[1] + 3.*x[0],
                ],
                [
                    -sin((x[0]/200.)**2+x[1]/100.)*2.*x[0]/200./200.,
                    -sin((x[0]/200.)**2+x[1]/100.)/100.,
                ],
            ],
            order=data['spiked_sorder'],
        )
        return c, cjsl

    @staticmethod
    def cb_objfun(_mode, x, objgrd, _nstate, data):
        # pylint: disable=missing-docstring
        objf = sum(x*np.sin(np.sqrt(np.abs(x))))
        objgrd = (
            np.sin(np.sqrt(np.abs(x))) +
            0.5*np.sqrt(np.abs(x))*np.cos(np.sqrt(np.abs(x)))
        )
        return objf, objgrd

    @staticmethod
    def cb_start(quas, repeat, bl, bu, data):
        # pylint: disable=missing-docstring
        n, npts = quas.shape
        comm = rand.quasi_init(1, n, 100)
        quas = rand.quasi_uniform(npts, comm)
        for j in range(npts):
            quas[:, j] = (bu - bl)*quas[:, j] + bl
        if data['spiked_sorder'] == 'C':
            return quas
        return np.asfortranarray(quas)

    def test_good_spiked11(self):
        """
        Good call with one linear, one nonlinear, constraint.
        """

        def cb_confun1(_mode, _needc, x, cjsl, _nstate, data):
            c2, cjsl2 = self.cb_confun(x, data)
            return [c2[0]], [cjsl2[0, :]]

        self.args[1] = 1
        self.args[2].pop()
        self.args[3].pop()

        for spiked_sorder in ['C', 'F']:
            data = {'spiked_sorder': spiked_sorder}
            soln = self.fun(
                *self.args,
                a=self.a, confun=cb_confun1,
                data=data, spiked_sorder=spiked_sorder
            )
            self.assertAlmostEqual(
                soln.objf[self.args[7] - 1], -810.416177439603,
            )

    def test_good_spiked12(self):
        """
        Good call with one linear, two nonlinear, constraints.
        """

        def cb_confun2(_mode, _needc, x, cjsl, _nstate, data):
            return self.cb_confun(x, data)

        for spiked_sorder in ['C', 'F']:
            data = {'spiked_sorder': spiked_sorder}
            soln = self.fun(
                *self.args,
                a=self.a, confun=cb_confun2,
                data=data, spiked_sorder=spiked_sorder
            )
            self.assertAlmostEqual(
                soln.objf[self.args[7] - 1], -731.70639282,
            )

    def test_good_start12(self):
        """
        Good call with one linear, two nonlinear, constraints; start supplied.
        """

        def cb_confun2(_mode, _needc, x, cjsl, _nstate, data):
            return self.cb_confun(x, data)

        for spiked_sorder in ['C', 'F']:
            data = {'spiked_sorder': spiked_sorder}
            soln = self.fun(
                *self.args,
                a=self.a, start=self.cb_start, confun=cb_confun2,
                data=data, spiked_sorder=spiked_sorder
            )
            self.assertAlmostEqual(
                soln.objf[self.args[7] - 1], -731.70639282,
            )

if __name__ == '__main__': # pragma: no cover
    unittest.main()
