#!/usr/bin/env python
"""
Integration tests for wrapper functionality in the lapackeig submodule.

Public use of this module is not supported by NAG.
"""

# NAG Copyright 2018-2019.

# pylint: disable=invalid-name,too-many-lines

import unittest

import numpy as np

from naginterfaces.base import utils as b_utils
from naginterfaces.base import lapackeig as b_chapter
from naginterfaces.library import lapackeig as chapter
from test_functionality_core import IntegrationTest

class TestF08aa(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the f08aa wrapper.
    """

class TestF08aaNp(TestF08aa, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestF08aa.set_up(self)
        self.fun = chapter.dgels

    def test_good(self):
        """
        Good call.
        """
        # pylint: disable=invalid-name
        a = np.array(
            [
                [-0.57, -1.28, -0.39, 0.25],
                [-1.93, 1.08, -0.31, -2.14],
                [2.30, 0.24, 0.40, -0.35],
                [-1.93, 0.64, -0.66, 0.08],
                [0.15, 0.30, 0.15, -2.13],
                [-0.02, 1.03, -1.43, 0.50],
            ]
        )
        b = np.array(
            [
                [-2.67],
                [-0.55],
                [3.34],
                [-0.77],
                [0.48],
                [4.10],
            ]
        )
        np.testing.assert_allclose(
            self.fun('N', a, b).b[:min(a.shape)],
            [
                [1.5339],
                [1.8707],
                [-1.5241],
                [0.0392],
            ],
            rtol=1e-3,
        )

class TestF08an(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the f08an wrapper.
    """

class TestF08anNp(TestF08an, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestF08an.set_up(self)
        self.fun = chapter.zgels

    def test_good(self):
        """
        Good call.
        """
        # pylint: disable=invalid-name
        a = np.array(
            [
                [
                    0.96 + -0.81j, -0.03 + 0.96j,
                    -0.91 + 2.06j, -0.05 + 0.41j,
                ],
                [
                    -0.98 + 1.98j, -1.20 + 0.19j,
                    -0.66 + 0.42j, -0.81 + 0.56j,
                ],
                [
                    0.62 + -0.46j, 1.01 + 0.02j,
                    0.63 + -0.17j, -1.11 + 0.60j,
                ],
                [
                    -0.37 + 0.38j, 0.19 + -0.54j,
                    -0.98 + -0.36j, 0.22 + -0.20j,
                ],
                [
                    0.83 + 0.51j, 0.20 + 0.01j,
                    -0.17 + -0.46j, 1.47 + 1.59j,
                ],
                [
                    1.08 + -0.28j, 0.20 + -0.12j,
                    -0.07 + 1.23j, 0.26 + 0.26j,
                ],
            ]
        )
        b = np.array(
            [
                [-2.09 + 1.93j],
                [3.34 + -3.53j],
                [-4.94 + -2.04j],
                [0.17 + 4.23j],
                [-5.19 + 3.63j],
                [0.98 + 2.53j],
            ]
        )
        np.testing.assert_allclose(
            self.fun('N', a, b).b[:min(a.shape)],
            [
                [-0.5044 - 1.2179j],
                [-2.4281 + 2.8574j],
                [1.4872 - 2.1955j],
                [0.4537 + 2.6904j],
            ],
            rtol=1e-3,
        )

class TestF08av(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the f08av wrapper.
    """

class TestF08avNp(TestF08av, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestF08av.set_up(self)
        self.fun = chapter.zgelqf

    def test_null(self):
        """
        Zero-sized problem.
        """
        self.fun(np.empty((0, 0), dtype=complex))

class TestF08fa(IntegrationTest):
    """
    Abstract suite of integration tests for the f08fa wrapper.
    """

    def default_args(self, n=4):
        """
        A default argument set.
        """
        # pylint: disable=attribute-defined-outside-init
        self.args = [
            'V', # jobz, 0
            'U', # uplo, 1
            n, # n, 2
            np.random.random((n, n)), # a, 3
            np.empty(n), # w, 4
        ]

    def normalize_a(self):
        """
        Normalize returned eigenvectors.
        """
        # pylint: disable=no-member,unsubscriptable-object,unsupported-assignment-operation
        assert self.args[0] == 'V'

        for m_i, m_el in enumerate(np.argmax(abs(self.args[3]), axis=0)):
            if self.args[3][m_el, m_i] < 0:
                self.args[3][:, m_i] *= -1

class TestF08faBase(TestF08fa, unittest.TestCase):
    """
    Concrete base tests.
    """

    def setUp(self):
        TestF08fa.set_up(self)
        self.fun = b_chapter.dsyev
        self.default_args()

    def test_invalid_jobz(self):
        """
        Invalid jobz.
        """
        self.args[0] = '?'
        self.my_assert_raises_regex(
            b_utils.NagValueError,
            r'.*jobz in \(\'N\', \'n\', \'V\', \'v\'\).',
            self.fun, *self.args
        )

    def test_invalid_uplo(self):
        """
        Invalid uplo.
        """
        self.args[1] = '?'
        self.my_assert_raises_regex(
            b_utils.NagValueError,
            r'.*uplo in \(\'U\', \'u\', \'L\', \'l\'\).',
            self.fun, *self.args
        )

    def test_valid_f(self):
        """
        Correct operation with Fortran-contiguous input.
        """
        a = np.array(
            [[1.0, 2.0, 3.0, 4.0],
             [-42.0, 2.0, 3.0, 4.0],
             [-43.0, -44.0, 3.0, 4.0],
             [-45.0, -46.0, -47.0, 4.0]],
            order='F',
        )
        self.args[3] = a
        self.fun(*self.args)
        np.testing.assert_allclose(
            self.args[4],
            [-2.0531, -0.5146, -0.2943, 12.8621],
            rtol=1e-4,
        )
        self.normalize_a()
        np.testing.assert_allclose(
            self.args[3],
            np.array(
                [[0.7003, -0.5144, -0.2767, 0.4103],
                 [0.3592, 0.4851, 0.6634, 0.4422],
                 [-0.1569, 0.5420, -0.6504, 0.5085],
                 [-0.5965, -0.4543, 0.2457, 0.6144]]
            ),
            rtol=1e-3,
        )

class TestF08faNp(TestF08fa, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestF08fa.set_up(self)
        self.fun = chapter.dsyev
        self.default_args()

    def test_jobz_n(self):
        """
        Eigenvectors declined on output, forcing bytes jobz.
        """
        self.assertIsNone(self.fun(b'N', self.args[1], self.args[3]).a)

    def test_jobz_v(self):
        """
        Eigenvectors non-None on output.
        """
        self.assertIsNotNone(self.fun(b'V', self.args[1], self.args[3]).a)

class TestF08fe(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the f08fe wrapper.
    """

class TestF08feNp(TestF08fe, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestF08fe.set_up(self)
        self.fun = chapter.dsytrd

    def test_null(self):
        """
        Zero-sized problem.
        """
        self.fun('U', [[]])

    def test_unit(self):
        """
        Unit-sized problem.
        """
        soln = self.fun('U', [[42.]])
        self.assertEqual(len(soln.e), 1)
        self.assertEqual(len(soln.tau), 1)

class TestF08fl(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the f08fl wrapper.
    """

class TestF08flNp(TestF08fl, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestF08fl.set_up(self)
        self.fun = chapter.ddisna

    def test_nonmon_evals(self):
        """
        Non-monotonic d supplied.
        """
        self.my_assert_raises_regex(
            b_utils.NagValueError,
            r'd\[0:m\] must be monotonic',
            self.fun,
            'E', 3, 3, [1., 0., 1.],
        )

    def test_neg_evals(self):
        """
        Negative d supplied.
        """
        self.my_assert_raises_regex(
            b_utils.NagValueError,
            r'd\[0:min\(n, m\)\] must not be negative',
            self.fun,
            'L', 3, 3, [-1., 0., 1.],
        )

    def test_nonmon_svals(self):
        """
        Non-monotonic d supplied.
        """
        self.my_assert_raises_regex(
            b_utils.NagValueError,
            r'd\[0:min\(n, m\)\] must be monotonic',
            self.fun,
            'L', 3, 3, [1., 0., 1.],
        )

class TestF08fr(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the f08fr wrapper.
    """

class TestF08frNp(TestF08fr, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestF08fr.set_up(self)
        self.fun = chapter.zheevr

    def test_good_novec(self):
        """
        Good call.
        """
        jobz = 'N'
        erange = 'A'
        uplo = 'U'
        a = [
            [1.0 + 0.0*1j, 2.0 + -1.0*1j, 3.0 + -1.0*1j, 4.0 + -1.0*1j],
            [42. + 42.*1j, 2.0 + 0.0*1j, 3.0 + -2.0*1j, 4.0 + -2.0*1j],
            [43. + 43.*1j, 44. + 44.*1j, 3.0 + 0.0*1j, 4.0 + -3.0*1j],
            [45. + 45.*1j, 46. + 46.*1j, 47. + 47.*1j, 4.0 + 0.0*1j],
        ]
        vl = -42.
        vu = -43.
        il = -42
        iu = -43
        abstol = 0.
        soln = self.fun(jobz, erange, uplo, a, vl, vu, il, iu, abstol)
        np.testing.assert_allclose(
            soln.w,
            [-4.2443054, -0.68858115, 1.14124852, 13.79163803],
        )

class TestF08fs(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the f08fs wrapper.
    """

class TestF08fsNp(TestF08fs, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestF08fs.set_up(self)
        self.fun = chapter.zhetrd

    def test_null(self):
        """
        Zero-sized problem.
        """
        self.fun('U', np.array([[]], dtype=complex))

    def test_unit(self):
        """
        Unit-sized problem.
        """
        soln = self.fun('U', np.array([[42.]], dtype=complex))
        self.assertEqual(len(soln.e), 1)
        self.assertEqual(len(soln.tau), 1)

class TestF08ge(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the f08ge wrapper.
    """

class TestF08geNp(TestF08ge, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestF08ge.set_up(self)
        self.fun = chapter.dsptrd

    def test_unit(self):
        """
        Unit-sized problem.
        """
        soln = self.fun('U', 1, np.array([42.]))
        self.assertEqual(len(soln.e), 1)
        self.assertEqual(len(soln.tau), 1)

class TestF08gs(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the f08gs wrapper.
    """

class TestF08gsNp(TestF08gs, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestF08gs.set_up(self)
        self.fun = chapter.zhptrd

    def test_unit(self):
        """
        Unit-sized problem.
        """
        soln = self.fun('U', 1, np.array([42.], dtype=complex))
        self.assertEqual(len(soln.e), 1)
        self.assertEqual(len(soln.tau), 1)

class TestF08he(IntegrationTest): # pylint: disable=too-few-public-methods,too-many-instance-attributes
    """
    Abstract suite of integration tests for the f08he wrapper.
    """

    def set_up(self):
        # pylint: disable=attribute-defined-outside-init
        IntegrationTest.set_up(self)
        self.uplo = 'L'
        self.ab = np.array([
            [4.99, 1.05, -2.31, -0.43],
            [0.04, -0.79, -1.3, 42.],
            [0.22, 1.04, 42., 42.],
        ])
        self.kd, self.n = self.ab.shape # pylint: disable=unpacking-non-sequence
        self.kd -= 1
        self.d = [42.]*self.n
        self.d_exp = [
            4.9900000000000002,
            -2.4805600000000005,
            -6.6113837955659865e-02,
            0.8566738379556604,
        ]
        self.e = [42.]*(self.n - 1)
        self.e_exp = [
            0.2236067977499790,
            1.1029754695368343,
            1.4300963620317853,
        ]
        self.q = np.empty((self.n, self.n))
        self.q_exp = [
            [1., 0., 0., 0.],
            [0., 0.1788854381999832, -0.1320894800577692, -0.9749627527542107],
            [
                0., 0.9838699100999075,
                2.4016269101412590e-02, 0.1772659550462201,
            ],
            [0., 0., -0.9909468139494252, 0.1342550256917159],
        ]
        self.q_expu = [
            [1., 0., 0., 0.],
            [0., -0.097959, 0.939335, -0.328715],
            [0., 0.199629, -0.305043, -0.93118],
            [0., -0.974963, -0.156838, -0.157637],
        ]

class TestF08heBase(TestF08he, unittest.TestCase):
    """
    Concrete base tests.
    """

    def setUp(self):
        TestF08he.set_up(self)
        self.fun = b_chapter.dsbtrd

    def test_good_qret(self):
        """
        Good call.
        """
        self.fun(
            'V', self.uplo, self.n, self.kd, self.ab,
            self.d, self.e, self.q,
        )
        np.testing.assert_allclose(self.d, self.d_exp)
        np.testing.assert_allclose(self.e, self.e_exp)
        np.testing.assert_allclose(self.q, self.q_exp)

    def test_good_qupd(self):
        """
        Good call.
        """
        q = np.empty(self.q.shape)
        q[:, :] = self.q_exp
        self.fun(
            'U', self.uplo, self.n, self.kd, self.ab,
            self.d, self.e, q,
        )
        np.testing.assert_allclose(q, self.q_expu, atol=1e-6)

    def test_good_qnot(self):
        """
        Good call.
        """
        q = np.array([[None]], dtype=float)
        self.fun(
            'N', self.uplo, self.n, self.kd, self.ab,
            self.d, self.e, q,
        )

class TestF08heNp(TestF08he, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestF08he.set_up(self)
        self.fun = chapter.dsbtrd

    def test_good_qret(self):
        """
        Good call.
        """
        soln = self.fun(
            'V', self.uplo, self.kd, self.ab,
        )
        np.testing.assert_allclose(soln.d, self.d_exp)
        np.testing.assert_allclose(soln.e, self.e_exp)
        np.testing.assert_allclose(soln.q, self.q_exp)

    def test_good_qupd(self):
        """
        Good call.
        """
        q = np.empty(self.q.shape)
        q[:, :] = self.q_exp
        soln = self.fun(
            'U', self.uplo, self.kd, self.ab, q=q,
        )
        np.testing.assert_allclose(soln.q, self.q_expu, atol=1e-6)

    def test_good_qnot(self):
        """
        Good call.
        """
        soln = self.fun(
            'N', self.uplo, self.kd, self.ab,
        )
        self.assertIsNone(soln.q)

    def test_unit(self):
        """
        Unit-sized problem.
        """
        soln = self.fun('N', 'U', 1, np.array([[42.], [43.]]))
        self.assertEqual(len(soln.e), 1)

class TestF08hs(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the f08hs wrapper.
    """

class TestF08hsNp(TestF08hs, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestF08hs.set_up(self)
        self.fun = chapter.zhbtrd

    def test_unit(self):
        """
        Unit-sized problem.
        """
        soln = self.fun('N', 'U', 1, np.array([[42.], [43.]], dtype=complex))
        self.assertEqual(len(soln.e), 1)

class TestF08jk(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the f08jk wrapper.
    """

class TestF08jkNp(TestF08jk, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestF08jk.set_up(self)
        self.fun = chapter.dstein

    def test_nonincr(self):
        """
        Non-increasing iblock supplied.
        """
        self.my_assert_raises_regex(
            b_utils.NagValueError,
            r'iblock\[i\] <= iblock\[i\+1\]',
            self.fun,
            [42., 43.], [44.], 2, [45., 46.], [1, -1], [0, 0],
        )

class TestF08kb(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the f08kb wrapper.
    """

class TestF08kbNp(TestF08kb, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestF08kb.set_up(self)
        self.fun = chapter.dgesvd

    def test_good(self):
        """
        Good call.
        """
        a = [
            [2.27, -1.54, 1.15, -1.94],
            [0.28, -1.67, 0.94, -0.78],
            [-0.48, -3.09, 0.99, -0.21],
            [1.07, 1.22, 0.79, 0.63],
            [-2.35, 2.93, -1.45, 2.30],
            [0.62, -7.39, 1.03, -2.57],
        ]
        res = self.fun('A', 'S', a)
        self.assertNotEqual(
            res.workb[0], 0,
        )

class TestF08kh(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the f08kh wrapper.
    """

class TestF08khNp(TestF08kh, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestF08kh.set_up(self)
        self.fun = chapter.dgejsv

    def test_good(self):
        """
        Good call.
        """
        a = [
            [2.27, -1.54, 1.15, -1.94],
            [0.28, -1.67, 0.94, -0.78],
            [-0.48, -3.09, 0.99, -0.21],
            [1.07, 1.22, 0.79, 0.63],
            [-2.35, 2.93, -1.45, 2.30],
            [0.62, -7.39, 1.03, -2.57],
        ]
        res = self.fun('E', 'U', 'V', 'R', 'N', 'N', a)
        np.testing.assert_allclose(
            res.workb,
            [
                1., 1.,
                9.0170813116961561, 2.0959189424408677,
                1.2427102399856174, 0.3735055404194461,
                9.9124784958542572E-02,
            ],
            rtol=1e-6,
        )
        np.testing.assert_equal(
            res.iworkb,
            [4, 4, 0],
        )

    def test_good_empty(self):
        """
        Good call.
        """
        self.assertIsNone(
            self.fun('E', 'U', 'V', 'R', 'N', 'N', [[]]).iworkb
        )

class TestF08kj(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the f08kj wrapper.
    """

class TestF08kjNp(TestF08kj, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestF08kj.set_up(self)
        self.fun = chapter.dgesvj

    def test_good(self):
        """
        Good call.
        """
        a = [
            [2.27, -1.54, 1.15, -1.94],
            [0.28, -1.67, 0.94, -0.78],
            [-0.48, -3.09, 0.99, -0.21],
            [1.07, 1.22, 0.79, 0.63],
            [-2.35, 2.93, -1.45, 2.30],
            [0.62, -7.39, 1.03, -2.57],
        ]
        v = np.empty((len(a[0]), len(a[0])))
        res = self.fun('G', 'U', 'V', a, 0, v, 0.)
        np.testing.assert_allclose(
            res.work,
            [1., 4., 4., 5., 1.0884145690747164E-16, 0.],
            atol=1e-6,
        )

class TestF08km(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the f08km wrapper.
    """

class TestF08kmNp(TestF08km, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestF08km.set_up(self)
        self.fun = chapter.dgesvdx

    def test_good(self):
        """
        Good call.
        """
        a = [
            [2.27, -1.54, 1.15, -1.94],
            [0.28, -1.67, 0.94, -0.78],
            [-0.48, -3.09, 0.99, -0.21],
            [1.07, 1.22, 0.79, 0.63],
            [-2.35, 2.93, -1.45, 2.30],
            [0.62, -7.39, 1.03, -2.57],
        ]
        res = self.fun('V', 'V', 'V', a, 1., 4., -42, -42)
        np.testing.assert_equal(
            res.jfail,
            0,
        )
        # workb not checked, as only copied back by MKL when info > 0?

class TestF08kp(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the f08kp wrapper.
    """

class TestF08kpNp(TestF08kp, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestF08kp.set_up(self)
        self.fun = chapter.zgesvd

    def test_good(self):
        """
        Good call.
        """
        a = [
            [0.96 + -0.81j, -0.03 + 0.96j, -0.91 + 2.06j, -0.05 + 0.41j],
            [-0.98 + 1.98j, -1.20 + 0.19j, -0.66 + 0.42j, -0.81 + 0.56j],
            [0.62 + -0.46j, 1.01 + 0.02j, 0.63 + -0.17j, -1.11 + 0.60j],
            [-0.37 + 0.38j, 0.19 + -0.54j, -0.98 + -0.36j, 0.22 + -0.20j],
            [0.83 + 0.51j, 0.20 + 0.01j, -0.17 + -0.46j, 1.47 + 1.59j],
            [1.08 + -0.28j, 0.20 + -0.12j, -0.07 + 1.23j, 0.26 + 0.26j],
        ]
        res = self.fun('A', 'S', a)
        np.testing.assert_equal(
            res.rworkb[:-1],
            [0., 0., 0.],
        )

class TestF08kv(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the f08kv wrapper.
    """

class TestF08kvNp(TestF08kv, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestF08kv.set_up(self)
        self.fun = chapter.zgejsv

    def test_good(self):
        """
        Good call.
        """
        a = [
            [0.96 + -0.81j, -0.03 + 0.96j, -0.91 + 2.06j, -0.05 + 0.41j],
            [-0.98 + 1.98j, -1.20 + 0.19j, -0.66 + 0.42j, -0.81 + 0.56j],
            [0.62 + -0.46j, 1.01 + 0.02j, 0.63 + -0.17j, -1.11 + 0.60j],
            [-0.37 + 0.38j, 0.19 + -0.54j, -0.98 + -0.36j, 0.22 + -0.20j],
            [0.83 + 0.51j, 0.20 + 0.01j, -0.17 + -0.46j, 1.47 + 1.59j],
            [1.08 + -0.28j, 0.20 + -0.12j, -0.07 + 1.23j, 0.26 + 0.26j],
        ]
        res = self.fun('E', 'U', 'V', 'R', 'N', 'N', a)
        np.testing.assert_allclose(
            res.rworkb,
            [
                1., 1.,
                2.6112592306813465, 1.9014639268294895,
                1.3933104266079950, 0.,
                5.3914108979713150E+153,
            ],
            rtol=1e-6,
        )
        np.testing.assert_equal(
            res.iworkb,
            [4, 4, 0],
        )

    def test_good_empty(self):
        """
        Good call.
        """
        a = np.empty((0, 0), dtype=complex)
        self.assertIsNone(
            self.fun('E', 'U', 'V', 'R', 'N', 'N', a).iworkb
        )

class TestF08kw(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the f08kw wrapper.
    """

class TestF08kwNp(TestF08kw, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestF08kw.set_up(self)
        self.fun = chapter.zgesvj

    def test_good(self):
        """
        Good call.
        """
        a = [
            [0.96 + -0.81j, -0.03 + 0.96j, -0.91 + 2.06j, -0.05 + 0.41j],
            [-0.98 + 1.98j, -1.20 + 0.19j, -0.66 + 0.42j, -0.81 + 0.56j],
            [0.62 + -0.46j, 1.01 + 0.02j, 0.63 + -0.17j, -1.11 + 0.60j],
            [-0.37 + 0.38j, 0.19 + -0.54j, -0.98 + -0.36j, 0.22 + -0.20j],
            [0.83 + 0.51j, 0.20 + 0.01j, -0.17 + -0.46j, 1.47 + 1.59j],
            [1.08 + -0.28j, 0.20 + -0.12j, -0.07 + 1.23j, 0.26 + 0.26j],
        ]
        v = np.empty((len(a[0]), len(a[0])), dtype=complex)
        res = self.fun('G', 'U', 'V', a, 0, v, 0.)
        np.testing.assert_allclose(
            res.rwork,
            [1., 4., 4., 5., 1.0884145690747164E-16, 0.],
            atol=1e-6,
        )

class TestF08kz(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the f08kz wrapper.
    """

class TestF08kzNp(TestF08kz, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestF08kz.set_up(self)
        self.fun = chapter.zgesvdx

    def test_good(self):
        """
        Good call.
        """
        a = [
            [0.96 + -0.81j, -0.03 + 0.96j, -0.91 + 2.06j, -0.05 + 0.41j],
            [-0.98 + 1.98j, -1.20 + 0.19j, -0.66 + 0.42j, -0.81 + 0.56j],
            [0.62 + -0.46j, 1.01 + 0.02j, 0.63 + -0.17j, -1.11 + 0.60j],
            [-0.37 + 0.38j, 0.19 + -0.54j, -0.98 + -0.36j, 0.22 + -0.20j],
            [0.83 + 0.51j, 0.20 + 0.01j, -0.17 + -0.46j, 1.47 + 1.59j],
            [1.08 + -0.28j, 0.20 + -0.12j, -0.07 + 1.23j, 0.26 + 0.26j],
        ]
        res = self.fun('V', 'V', 'V', a, 1., 4., -42, -42)
        np.testing.assert_equal(
            res.jfail,
            0,
        )
        self.assertNotEqual(
            res.rworkb[0], 0,
        )

class TestF08mb(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the f08mb wrapper.
    """

class TestF08mbNp(TestF08mb, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestF08mb.set_up(self)
        self.fun = chapter.dbdsvdx

    def test_good(self):
        """
        Good call.
        """
        d = [3.62, -2.41, 1.92, -1.43]
        e = [1.26, -1.53, 1.19]
        vl = 0.
        vu = 4.
        il = -42
        iu = -42
        res = self.fun('U', 'V', 'V', d, e, vl, vu, il, iu)
        self.assertEqual(
            res.ns,
            3,
        )
        np.testing.assert_allclose(
            res.s[:res.ns],
            [3.0006, 1.9960, 0.9998],
            rtol=1e-4,
        )

class TestF08md(IntegrationTest):
    """
    Abstract suite of integration tests for the f08md wrapper.
    """

    def default_args(self, compq='i'):
        """
        A default argument set.
        """
        # pylint: disable=attribute-defined-outside-init
        d = [3.62, -2.41, 1.92, -1.43]
        e = [1.26, -1.53, 1.19]
        n = len(d)
        self.args = [
            'u', # uplo, 0
            compq, # compq, 1
            n, # n, 2
            d, # d, 3
            e, # e, 4
            np.empty((
                n if compq == 'i' else 1,
                n if compq == 'i' else 0,
            )), # u, 5
            np.empty((
                n if compq == 'i' else 1,
                n if compq == 'i' else 0,
            )), # vt, 6
            np.empty(
                b_utils._f08mdan(True, n)[0].value if compq == 'p' # pylint: disable=protected-access
                else 0
            ), # q, 7
            np.empty(
                max(
                    b_utils._f08mdan(True, n)[0].value, # pylint: disable=protected-access
                    b_utils._f08mdan(True, n)[1].value, # pylint: disable=protected-access
                ) if compq == 'p'
                else 0, dtype=int
            ), # iq, 8
        ]

class TestF08mdBase(TestF08md, unittest.TestCase):
    """
    Concrete base tests.
    """

    def setUp(self):
        TestF08md.set_up(self)
        self.fun = b_chapter.dbdsdc
        self.default_args()

    def test_good(self):
        """
        Good call.
        """
        self.fun(*self.args)
        np.testing.assert_allclose(
            self.args[3],
            [4.0001, 3.0006, 1.9960, 0.9998],
            rtol=1e-4,
        )

    def test_good_p(self):
        """
        Good call, packed return data.
        """
        self.default_args(compq='p')
        self.fun(*self.args)
        np.testing.assert_allclose(
            self.args[3],
            [4.0001, 3.0006, 1.9960, 0.9998],
            rtol=1e-4,
        )

class TestF08mdNp(TestF08md, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestF08md.set_up(self)
        self.fun = chapter.dbdsdc
        self.default_args()

    def test_good(self):
        """
        Good call.
        """
        np.testing.assert_allclose(
            self.fun(self.args[0], self.args[1], self.args[3], self.args[4]).d,
            [4.0001, 3.0006, 1.9960, 0.9998],
            rtol=1e-4,
        )

    def test_good_p(self):
        """
        Good call, packed return data.
        """
        return_data = self.fun(self.args[0], 'p', self.args[3], self.args[4])
        np.testing.assert_allclose(
            return_data.d,
            [4.0001, 3.0006, 1.9960, 0.9998],
            rtol=1e-4,
        )

class TestF08ne(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the f08ne wrapper.
    """

class TestF08neNp(TestF08ne, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestF08ne.set_up(self)
        self.fun = chapter.dgehrd

    def test_unit(self):
        """
        Unit-sized problem.
        """
        soln = self.fun(1, 1, np.array([[42.]]))
        self.assertEqual(len(soln.tau), 1)

class TestF08ns(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the f08ns wrapper.
    """

class TestF08nsNp(TestF08ns, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestF08ns.set_up(self)
        self.fun = chapter.zgehrd

    def test_unit(self):
        """
        Unit-sized problem.
        """
        soln = self.fun(1, 1, np.array([[42.]], dtype=complex))
        self.assertEqual(len(soln.tau), 1)

class TestF08xp(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the f08xp wrapper.
    """

class TestF08xpBase(TestF08xp, unittest.TestCase):
    """
    Concrete base tests.
    """

    def setUp(self):
        TestF08xp.set_up(self)
        self.fun = b_chapter.zggesx

    def test_good(self):
        """
        Good call.
        """
        data = {'ncalls': 0}

        def cb_selctg(a, b, data):
            # pylint: disable=missing-docstring
            data['ncalls'] += 1
            return abs(a) < 6.*abs(b)

        a = np.array([
            [-21.1 + -22.5j, 53.5 + -50.5j, -34.5 + 127.5j, 7.5 + 0.5j],
            [-0.46 + -7.78j, -3.5 + -37.5j, -15.5 + 58.5j, -10.5 + -1.5j],
            [4.3 + -5.5j, 39.7 + -17.1j, -68.5 + 12.5j, -7.5 + -3.5j],
            [5.5 + 4.4j, 14.4 + 43.3j, -32.5 + -46j, -19 + -32.5j],
        ])
        n = a.shape[0]
        b = np.array([
            [1. + -5.j, 1.6 + 1.2j, -3. + 0j, 0 + -1j],
            [0.8 + -0.6j, 3 + -5j, -4 + 3j, -2.4 + -3.2j],
            [1 + 0j, 2.4 + 1.8j, -4 + -5j, 0 + -3j],
            [0 + 1j, -1.8 + 2.4j, 0 + -4j, 4 + -5j],
        ])
        alpha = np.empty(n, dtype=complex)
        beta = np.empty(n, dtype=complex)
        vsl = np.empty((n, n), dtype=complex)
        vsr = np.empty((n, n), dtype=complex)
        rconde = np.empty(2)
        rcondv = np.empty(2)
        self.fun(
            'V', 'V', 'S', cb_selctg, 'B', n, a, b, alpha, beta,
            vsl, vsr, rconde, rcondv, data,
        )
        self.assertNotEqual(data['ncalls'], 0)
        np.testing.assert_allclose(
            rconde, [0.120549, 0.162506],
            rtol=1e-5,
        )
        np.testing.assert_allclose(
            rcondv, [0.477028, 0.468981],
            rtol=1e-5,
        )

if __name__ == '__main__': # pragma: no cover
    unittest.main()
