#!/usr/bin/env python
"""
Integration tests for wrapper functionality in the lapacklin submodule.

Public use of this module is not supported by NAG.
"""

# NAG Copyright 2018-2019.

# pylint: disable=invalid-name

import ctypes
import unittest
import warnings

import numpy as np

from naginterfaces.base import utils as b_utils
from naginterfaces.base import lapacklin as b_chapter
from naginterfaces.library import lapacklin as chapter
from test_functionality_core import IntegrationTest

class TestF07aa(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the f07aa wrapper.
    """

class TestF07aaBase(TestF07aa, unittest.TestCase):
    """
    Concrete base tests.
    """

    def setUp(self):
        TestF07aa.set_up(self)
        self.fun = b_chapter.dgesv
        self.default_args()

    def default_args(self, n=4, nrhs=1):
        """
        A default argument set.
        """
        # pylint: disable=attribute-defined-outside-init
        self.args = [
            n, # n, 0
            nrhs, # nrhs, 1
            np.random.random((n, n)), # a, 2
            np.empty(n, dtype=int), # ipiv, 3
            np.random.random((n, nrhs)), # b, 4
        ]

    def test_invalid_n_char(self):
        """
        Invalid type should be rejected before constraints are checked.
        """
        self.args[0] = 'Hello'
        self.my_assert_raises_regex(
            b_utils.NagTypeError,
            '.*n. must be an instance of',
            self.fun, *self.args
        )

    def test_invalid_n_int(self):
        """
        Invalid default-int n.
        """
        self.args[0] = -1
        self.my_assert_raises_regex(
            b_utils.NagValueError,
            '.*constraint was not satisfied: n ',
            self.fun, *self.args
        )

    def test_invalid_n_cint(self):
        """
        Invalid cint n.
        """
        self.args[0] = ctypes.c_int(-1)
        self.my_assert_raises_regex(
            b_utils.NagValueError,
            '.*constraint was not satisfied: n ',
            self.fun, *self.args
        )

    def test_invalid_a_shape_cint(self):
        """
        Invalid shape of a, determined by cint n.
        """
        self.args[0] = ctypes.c_int(self.args[0] + 1)
        self.my_assert_raises_regex(
            b_utils.NagShapeError,
            '.*a. has shape .*, but must have shape ',
            self.fun, *self.args
        )

    def test_valid_a_shape_cint(self):
        """
        Valid shape for a, determined by cint n.
        """
        self.args[0] = ctypes.c_int(self.args[0])
        self.fun(*self.args)

    def test_valid_ints(self):
        """
        Using C and NumPy int types.
        """
        self.args[0] = b_utils.EngineIntCType(self.args[0])
        self.args[3] = np.empty(
            self.args[3].shape,
            dtype=b_utils.EngineIntNumPyType
        )
        self.fun(*self.args)

    def test_valid_f(self):
        """
        Correct operation with Fortran-contiguous input.
        """
        a = np.array(
            [[1.80, 2.88, 2.05, -0.89],
             [5.25, -2.95, -0.95, -3.80],
             [1.58, -2.69, -2.90, -1.04],
             [-1.11, -0.66, -0.59, 0.80]],
            order='F',
        )
        b = np.array(
            [[9.52], [24.35], [0.77], [-6.22]],
            order='F',
        )
        n = a.shape[0]
        ipiv = np.empty(n, dtype=int)
        self.fun(n, b.shape[-1], a, ipiv, b)
        self.assertEqual(
            list(ipiv),
            [2, 2, 3, 4]
        )

class TestF07aaNp(TestF07aa, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestF07aa.set_up(self)
        self.fun = chapter.dgesv

    def test_valid_pivots(self):
        """
        Testing integer-width compatibility in case other MKLs
        present in the environment.
        """
        a = np.array(
            [[1.80, 2.88, 2.05, -0.89],
             [5.25, -2.95, -0.95, -3.80],
             [1.58, -2.69, -2.90, -1.04],
             [-1.11, -0.66, -0.59, 0.80]],
        )
        b = np.array(
            [[9.52], [24.35], [0.77], [-6.22]],
        )
        self.assertEqual(
            list(self.fun(a, b).ipiv),
            [2, 2, 3, 4]
        )

    def test_singular(self):
        """
        A singular system supplied.

        Coverage for: handling info from the Engine.
        Coverage for: _STR_FMT_INT.
        """
        warnings.simplefilter('error', b_utils.NagAlgorithmicWarning)

        with self.assertRaises(b_utils.NagAlgorithmicWarning) as w_ctxt:
            self.fun(
                [[1., 0.], [0., 0.]], [[42.], [42.]],
            )

        self.assertEqual(
            w_ctxt.exception.errno,
            2,
        )
        self.assertEqual(
            w_ctxt.exception._errcodes, # pylint: disable=protected-access
            (2,),
        )
        self.assertIsInstance(w_ctxt.exception.return_data, tuple)
        self.assertEqual(
            w_ctxt.exception.return_data.__class__.__name__,
            'nag_return_data'
        )
        self.assertEqual(len(w_ctxt.exception.return_data), 3)

        warnings.simplefilter('always', b_utils.NagAlgorithmicWarning)

class TestF07ab(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the f07ab wrapper.
    """

class TestF07abBase(TestF07ab, unittest.TestCase):
    """
    Concrete base tests.
    """

    def setUp(self):
        TestF07ab.set_up(self)
        self.fun = b_chapter.dgesvx
        self.default_args()

    def default_args(self):
        """
        A default argument set.
        """
        # pylint: disable=attribute-defined-outside-init
        a = np.array([
            [1.80, 2.88, 2.05, -0.89],
            [525.00, -295.00, -95.00, -380.00],
            [1.58, -2.69, -2.90, -1.04],
            [-1.11, -0.66, -0.59, 0.80]
        ])
        n = a.shape[0]
        b = np.array([
            [9.52, 18.47],
            [2435.00, 225.00],
            [0.77, -13.28],
            [-6.22, -6.21]
        ])
        nrhs = b.shape[1]
        self.args = [
            '?', # fact, 0
            '?', # trans, 1
            n, # n, 2
            nrhs, # nrhs, 3
            a, # a, 4
            np.empty((n, n)), # af, 5
            np.zeros(n, dtype=int), # ipiv, 6
            None, # equed, 7
            np.empty(n), # r, 8
            np.empty(n), # c, 9
            b, # b, 10
            np.empty((n, nrhs)), # x, 11
            np.empty(nrhs), # ferr, 12
            np.empty(nrhs), # berr, 13
        ]

    def test_good(self):
        """
        Good call.
        """
        self.args[0] = 'E'
        self.args[1] = 'N'
        self.assertEqual(
            self.fun(*self.args).equed,
            'R',
        )

    def test_invalid_equed(self):
        """
        Invalid equed.
        """
        self.args[0] = 'F'
        self.args[1] = 'N'
        self.args[7] = '?'
        self.my_assert_raises_regex(
            b_utils.NagValueError,
            '.*constraint was not satisfied: equed ',
            self.fun, *self.args
        )

    def test_invalid_r(self):
        """
        Invalid r.
        """
        self.args[0] = 'F'
        self.args[1] = 'N'
        self.args[7] = 'R'
        self.args[8] = [1., 1., 1., -1.]
        self.my_assert_raises_regex(
            b_utils.NagValueError,
            '(?m).*argument r.*\n.*' +
            'Argument fails constraint check.',
            self.fun, *self.args
        )

class TestF07an(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the f07an wrapper.
    """

class TestF07anNp(TestF07an, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestF07an.set_up(self)
        self.fun = chapter.zgesv

    def test_good(self):
        """
        Good call.
        """
        # pylint: disable=invalid-name
        a = np.array(
            [
                [
                    -1.34 + 2.55j, 0.28 + 3.17j,
                    -6.39 + -2.20j, 0.72 + -0.92j,
                ],
                [
                    -0.17 + -1.41j, 3.31 + -0.15j,
                    -0.15 + 1.34j, 1.29 + 1.38j,
                ],
                [
                    -3.29 + -2.39j, -1.91 + 4.42j,
                    -0.14 + -1.35j, 1.72 + 1.35j,
                ],
                [
                    2.41 + 0.39j, -0.56 + 1.47j,
                    -0.83 + -0.69j, -1.96 + 0.67j,
                ],
            ]
        )
        b = np.array(
            [
                [26.26 + 51.78j],
                [6.43 + -8.68j],
                [-5.75 + 25.31j],
                [1.16 + 2.57j],
            ]
        )
        np.testing.assert_allclose(
            self.fun(a, b).b,
            [
                [1.0000 + 1.0000j],
                [2.0000 - 3.0000j],
                [-4.0000 - 5.0000j],
                [0.0000 + 6.0000j],
            ],
        )

class TestF07fa(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the f07fa wrapper.
    """

class TestF07faNp(TestF07fa, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestF07fa.set_up(self)
        self.fun = chapter.dposv

    def test_good(self):
        """
        Good call.
        """
        # pylint: disable=invalid-name
        a = np.array(
            [
                [4.16, -3.12, 0.56, -0.10],
                [0.00, 5.03, -0.83, 1.18],
                [0.00, 0.00, 0.76, 0.34],
                [0.00, 0.00, 0.00, 1.18],
            ],
        )
        b = np.array(
            [
                [8.70],
                [-13.35],
                [1.89],
                [-4.14],
            ],
        )
        np.testing.assert_allclose(
            self.fun('U', a, b).b,
            [
                [1.],
                [-1.],
                [2.],
                [-3.],
            ]
        )

class TestF07fn(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the f07fn wrapper.
    """

class TestF07fnNp(TestF07fn, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestF07fn.set_up(self)
        self.fun = chapter.zposv

    def test_good(self):
        """
        Good call.
        """
        # pylint: disable=invalid-name
        a = np.array(
            [
                [
                    3.23 + 0.0j, 1.51 + -1.92j,
                    1.90 + 0.84j, 0.42 + 2.50j,
                ],
                [
                    0.0 + 0.0j, 3.58 + 0.0j,
                    -0.23 + 1.11j, -1.18 + 1.37j,
                ],
                [
                    0.0 + 0.0j, 0.0 + 0.0j,
                    4.09 + 0.0j, 2.33 + -0.14j,
                ],
                [
                    0.0 + 0.0j, 0.0 + 0.0j,
                    0.0 + 0.0j, 4.29 + 0.0j,
                ],
            ],
        )
        b = np.array(
            [
                [3.93 + -6.14j],
                [6.17 + 9.42j],
                [-7.17 + -21.83j],
                [1.99 + -14.38j],
            ],
        )
        np.testing.assert_allclose(
            self.fun('U', a, b).b,
            [
                [1.00 - 1.00j],
                [-0.00 + 3.00j],
                [-4.00 - 5.00j],
                [2.00 + 1.00j],
            ]
        )

class TestF07md(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the f07md wrapper.
    """

class TestF07mdNp(TestF07md, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestF07md.set_up(self)
        self.fun = chapter.dsytrf

    def test_null(self):
        """
        Zero-sized problem.
        """
        self.fun('U', np.array([[]]))

class TestF07mr(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the f07mr wrapper.
    """

class TestF07mrNp(TestF07mr, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestF07mr.set_up(self)
        self.fun = chapter.zhetrf

    def test_null(self):
        """
        Zero-sized problem.
        """
        self.fun('U', 0, np.array([[]], dtype=complex))

class TestF07nr(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the f07nr wrapper.
    """

class TestF07nrNp(TestF07nr, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestF07nr.set_up(self)
        self.fun = chapter.zsytrf

    def test_null(self):
        """
        Zero-sized problem.
        """
        self.fun('U', np.array([[]], dtype=complex))

if __name__ == '__main__': # pragma: no cover
    unittest.main()
