#!/usr/bin/env python
"""
Integration tests for wrapper functionality in the mip submodule.

Public use of this module is not supported by NAG.
"""

# NAG Copyright 2018-2019.

# pylint: disable=invalid-name

import unittest
import warnings

import numpy as np

from naginterfaces.base import utils as b_utils
from naginterfaces.library import mip as chapter, opt
from test_functionality_core import IntegrationTest

class TestH02cb(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the h02cb wrapper.
    """

class TestH02cbNp(TestH02cb, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestH02cb.set_up(self)
        self.init = opt.nlp1_init
        self.fun = chapter.iqp_dense

    def run_problem(self, sorder):
        """
        Test the solver.
        """
        a = np.array([
            [1.00, 1.00, 1.00, 1.00, 1.00, 1.00, 1.00],
            [0.15, 0.04, 0.02, 0.04, 0.02, 0.01, 0.03],
            [0.03, 0.05, 0.08, 0.02, 0.06, 0.01, 0.00],
            [0.02, 0.04, 0.01, 0.02, 0.02, 0.00, 0.00],
            [0.02, 0.03, 0.00, 0.00, 0.01, 0.00, 0.00],
            [0.70, 0.75, 0.80, 0.75, 0.80, 0.97, 0.00],
            [0.02, 0.06, 0.08, 0.12, 0.02, 0.01, 0.97],
        ], order=sorder)
        bl = [
            -0.01, -0.10, -0.01, -0.04, -0.10, -0.01, -0.01,
            -0.13, -1.0e+25, -1.0e+25, -1.0e+25, -1.0e+25, -9.92e-02, -3.0e-03
        ]
        bu = [
            0.01, 0.15, 0.03, 0.02, 0.05, 1.0e+25, 1.0e+25,
            -0.13, -4.9e-03, -6.4e-03, -3.7e-03, -1.2e-03, 1.0e+25, 2.0e-03
        ]
        cvec = [-0.02, -0.20, -0.20, -0.20, -0.20, 0.04, 0.04]
        h = np.array([
            [2.00, 0.00, 0.00, 0.00, 0.00, 0.00, 0.00],
            [-42.00, 2.00, 0.00, 0.00, 0.00, 0.00, 0.00],
            [-42.00, -42.00, 2.00, 2.00, 0.00, 0.00, 0.00],
            [-42.00, -42.00, -42.00, 2.00, 0.00, 0.00, 0.00],
            [-42.00, -42.00, -42.00, -42.00, 2.00, 0.00, 0.00],
            [-42.00, -42.00, -42.00, -42.00, -42.00, -2.00, -2.00],
            [-42.00, -42.00, -42.00, -42.00, -42.00, -42.00, -2.00],
        ], order=sorder)
        intvar = 4
        istate = np.zeros(sum(a.shape), dtype=int)
        xs = [-0.01, -0.03, 0.00, -0.01, -0.10, 0.02, 0.01]
        strtgy = 2
        comm = self.init('opt.qp_dense_solve')
        soln = self.fun(
            a, bl, bu, cvec, h, intvar, istate, xs, strtgy, comm,
            mdepth=30,
        )
        self.assertAlmostEqual(soln.obj, 0.03746966203590227)

    def test_good_f(self):
        """
        Good call, column ordered.
        """
        self.run_problem('F')

    def test_good_c(self):
        """
        Good call, row ordered.
        """
        self.run_problem('C')

class TestH05aa(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the h05aa wrapper.
    """

class TestH05aaNp(TestH05aa, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestH05aa.set_up(self)
        self.fun = chapter.best_subset_given_size_revcomm

    def test_m_eq_ip(self):
        """
        Good call, m == ip.
        """
        # pylint: disable=too-many-locals
        irevcm = 0
        mincr = 0
        m = 14
        ip = m
        nbest = 1
        drop = 42
        lz = 42
        z = np.zeros(m-ip, dtype=int)
        la = 42
        a = np.zeros(max(nbest, m), dtype=int)
        bscore = np.empty(max(nbest, m))
        bz = np.zeros((m-ip, nbest), dtype=int)
        mincnt = -1
        gamma = -1.
        acc = [-1., -1.]
        comm = {}
        warnings.simplefilter('error', b_utils.NagAlgorithmicWarning)
        self.fun(
            irevcm, mincr, m, ip, nbest, drop, lz, z, la, a, bscore, bz,
            mincnt, gamma, acc, comm,
        )
        warnings.simplefilter('always', b_utils.NagAlgorithmicWarning)

if __name__ == '__main__': # pragma: no cover
    unittest.main()
