#!/usr/bin/env python
"""
Integration tests for wrapper functionality in the mv submodule.

Public use of this module is not supported by NAG.
"""

# NAG Copyright 2018-2019.

# pylint: disable=invalid-name

import unittest

import numpy as np

from naginterfaces.base import utils as b_utils
from naginterfaces.library import mv as chapter
from test_functionality_core import IntegrationTest

class TestG03e(IntegrationTest):
    """
    Abstract suite of integration tests for the g03e wrappers.
    """

    # pylint: disable=too-few-public-methods

    def set_up_dm(self):
        """
        Distance-matrix arguments.
        """
        # pylint: disable=attribute-defined-outside-init
        x = [
            [1., 1., 1.],
            [2., 1., 2.],
            [3., 6., 3.],
            [4., 8., 2.],
            [5., 8., 0.],
        ]
        self.n = len(x)
        m = len(x[0])
        self.dm_args = [
            'I', # update, 0
            'S', # dist, 1
            'U', # scal, 2
            x, # x, 3
            [0, 1, 1], # isx, 4
            np.empty(m), # s, 5
            np.empty(self.n*(self.n-1)//2), # d, 6
        ]

class TestG03eaNp(TestG03e, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestG03e.set_up(self)
        self.fun = chapter.distance_mat

    def test_good(self):
        """
        Good call.
        """
        self.set_up_dm()
        soln = self.fun(*self.dm_args)
        np.testing.assert_equal(
            soln.d,
            [1., 29., 26., 50., 49., 5., 50., 53., 13., 4.],
        )

class TestG03ecNp(TestG03e, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestG03e.set_up(self)
        self.dm = chapter.distance_mat
        self.fun = chapter.cluster_hier

    def test_good(self):
        """
        Good call.
        """
        self.set_up_dm()
        dm_soln = self.dm(*self.dm_args)
        soln = self.fun(5, self.n, dm_soln.d)
        np.testing.assert_equal(
            soln.dord,
            [1., 36.25, 8., 4., 36.25],
        )

class TestG03ehNp(TestG03e, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestG03e.set_up(self)
        self.dm = chapter.distance_mat
        self.cluster = chapter.cluster_hier
        self.fun = chapter.cluster_hier_dendrogram

    def test_good_e(self):
        """
        Good call, eastward.
        """
        self.set_up_dm()
        dm_soln = self.dm(*self.dm_args)
        cl_soln = self.cluster(5, self.n, dm_soln.d)
        nsym = 40
        c = self.fun(
            'E', cl_soln.dord, 0., 1.1, nsym,
        )
        np.testing.assert_equal(
            c,
            [
                '',
                '        ...............................(',
                '       (                         .......',
                '       (                        (    ...',
                '       (........................(...(...',
            ],
        )

    def test_good_s(self):
        """
        Good call, southerly.
        """
        self.set_up_dm()
        dm_soln = self.dm(*self.dm_args)
        cl_soln = self.cluster(5, self.n, dm_soln.d)
        nsym = 40
        c = self.fun(
            'S', cl_soln.dord, 0., 1.1, nsym,
        )
        np.testing.assert_equal(
            c,
            [
                '',
                '',
                '',
                '',
                '',
                '',
                '',
                '   ----------',
                '   I        I',
                '   I        I',
                '   I        I',
                '   I        I',
                '   I        I',
                '   I        I',
                '   I        I',
                '   I        I',
                '   I        I',
                '   I        I',
                '   I        I',
                '   I        I',
                '   I        I',
                '   I        I',
                '   I        I',
                '   I        I',
                '   I        I',
                '   I        I',
                '   I        I',
                '   I        I',
                '   I        I',
                '   I        I',
                '   I        I',
                '   I        I',
                '   I  ------*',
                '   I  I     I',
                '   I  I     I',
                '   I  I     I',
                '   I  I  ---*',
                '   I  I  I  I',
                '   I  I  I  I',
                '---*  I  I  I',
            ],
        )

    def test_bad_orient(self):
        """
        Bad call, illegal orient.
        """
        self.set_up_dm()
        dm_soln = self.dm(*self.dm_args)
        cl_soln = self.cluster(5, self.n, dm_soln.d)
        nsym = 40
        self.my_assert_raises_regex(
            b_utils.NagValueError,
            'On entry, orient',
            self.fun,
            'Z', cl_soln.dord, 0., 1.1, nsym,
        )

if __name__ == '__main__': # pragma: no cover
    unittest.main()
