#!/usr/bin/env python
"""
Integration tests for wrapper functionality in the ode submodule.

Public use of this module is not supported by NAG.
"""

# NAG Copyright 2018-2019.

# pylint: disable=invalid-name

import math
import os
import tempfile
import unittest

import numpy as np

from naginterfaces.base import utils as b_utils
from naginterfaces.library import ode as chapter
from test_functionality_core import IntegrationTest

class TestD02mw(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the d02mw wrapper.
    """

class TestD02mwNp(TestD02mw, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestD02mw.set_up(self)
        self.fun = chapter.dae_dassl_setup

    def test_good(self):
        """
        Good call.
        """
        comm = self.fun(1, 1, 'N', 0.0, 0.0, 0)
        self.assertEqual(
            len(comm['com']),
            50,
        )
        self.assertEqual(
            len(comm['icom']),
            51,
        )

class TestD02mz(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the d02mz wrapper.
    """

class TestD02mzNp(TestD02mz, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestD02mz.set_up(self)
        self.bdf_setup = chapter.ivp_stiff_bdf
        self.la_setup = chapter.ivp_stiff_fulljac_setup
        self.integ = chapter.ivp_stiff_imp_fulljac
        self.diagnostic = chapter.ivp_stiff_integ_diag
        self.fun = chapter.ivp_stiff_interp

    def test_good(self):
        """
        Good call.
        """
        # pylint: disable=too-many-locals,too-many-statements
        neqmax = 3
        sdysav = 6
        maxord = 5
        method = 'Functional-iteration'
        petzld = False
        con = [0.]*6
        tcrit = 0.
        hmin = 1.e-10
        hmax = 10.
        h0 = 0.
        maxstp = 200
        mxhnil = 5
        norm = 'Average-L2'
        comm = self.bdf_setup(
            neqmax, sdysav, maxord, method, petzld, con, tcrit, hmin, hmax, h0,
            maxstp, mxhnil, norm,
        ).comm
        neq = neqmax
        jceval = 'Numerical'
        nwkjac = neq*(neq+1)
        self.la_setup(neq, neqmax, jceval, nwkjac, comm)
        t = 0.
        tout = 0.1
        y = [1., 0., 0.]
        ydot = np.empty(neq)
        rtol = [1.e-4, 1.e-3, 1.e-4]
        atol = [1.e-7, 1.e-8, 1.e-7]
        itol = 4

        def cb_resid(t, y, ydot, ires):
            """Residual."""
            r = -ydot[:]
            if ires == 1:
                alpha = 0.04
                beta = 1.e4
                gamma = 3.e7
                r = [
                    -alpha*y[0] + beta*y[1]*y[2] + r[0],
                    alpha*y[0] - beta*y[1]*y[2] - gamma*y[1]*y[1] + r[1],
                    gamma*y[1]*y[1] + r[2],
                ]
            return r, ires

        lderiv = [False, False]
        itask = 2
        itrace = 0
        cb_monitr = None

        iout = 1
        tstep = 0.02
        xout = tstep

        while True:
            t, tout, y, ydot, lderiv = self.integ(
                t, tout, y, ydot, comm, rtol, atol, itol, cb_resid,
                lderiv, itask, itrace, monitr=cb_monitr,
            )

            d = self.diagnostic(neq, neqmax, comm)

            step_again = False

            while True:

                if d.tcur - d.hu < xout <= d.tcur:
                    sol = self.fun(xout, neq, comm)
                    iout += 1

                    if iout >= 6:
                        np.testing.assert_allclose(
                            sol,
                            [
                                0.9960780631878472,
                                3.5819770806533026E-05,
                                3.8861170413459550E-03,
                            ],
                            rtol=1e-3,
                        )
                        break

                    xout += tstep
                else:
                    step_again = True
                    break

            if not step_again:
                break

class TestD02nb(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the d02nb wrapper.
    """

class TestD02nbNp(TestD02nb, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestD02nb.set_up(self)
        self.fun = chapter.ivp_stiff_exp_fulljac
        self.bdf_setup = chapter.ivp_stiff_bdf
        self.la_setup = chapter.ivp_stiff_fulljac_setup
        self.interp_nat = chapter.ivp_stiff_nat_interp
        self.interp_c1 = chapter.ivp_stiff_c1_interp

    def with_interp(self):
        """
        Interpolating call.
        """
        # pylint: disable=too-many-arguments,too-many-locals

        neq = 3
        neqmax = 3

        maxord = 5
        sdysav = maxord + 1
        method = 'NewtonMod'
        petzld = False
        con = [0.]*6
        tcrit = 10.
        hmin = 1.e-10
        hmax = 10.
        h0 = 0.
        maxstp = 200
        mxhnil = 5
        norm = 'AveragedL2'
        con, comm = self.bdf_setup(
            neqmax, sdysav, maxord, method, petzld, con, tcrit, hmin, hmax, h0,
            maxstp, mxhnil, norm,
        )

        jceval = 'Numerical'
        nwkjac = neqmax*(neqmax+1)
        self.la_setup(neq, neqmax, jceval, nwkjac, comm)

        y = [1., 0., 0.]
        t = 0.
        tout = tcrit
        itol = 1
        rtol = 1.e-4
        atol = 1.e-7

        alpha = 0.04
        beta = 1.e4
        gamma = 3.e7

        def cb_fcn(t, y, ires):
            """The function for the derivative vector."""
            return (
                [
                    -alpha*y[0] + beta*y[1]*y[2],
                    alpha*y[0] - beta*y[1]*y[2] - gamma*y[1]**2,
                    gamma*y[1]**2,
                ],
                ires
            )

        def cb_monitr(
                t, hlast, hnext, y, ydot, comm, r, acor, imon, hmin,
                hmax, nqu,
        ):
            """monitr."""
            inln = 3

            if imon != 1:
                self.fail('inconceivable') # pragma: no cover

            xout = 2.

            while xout <= t - hlast:
                xout += 2.

            while t-hlast < xout <= t:
                z = self.interp_nat(
                    xout, neq, comm, neq, t, nqu, hlast, hnext,
                )
                z1 = self.interp_c1(
                    xout, neq, comm, acor[:, 1], t, nqu, hlast, hnext,
                )
                xout += 2.
                if xout >= 10.:
                    break

            if xout == 12.:
                np.testing.assert_allclose(
                    z,
                    [8.41357701e-01, 1.62329053e-05, 1.58626066e-01],
                    rtol=1e-6,
                )
                np.testing.assert_allclose(
                    z1,
                    [8.41357701e-01, 1.62329053e-05, 1.58626066e-01],
                    rtol=1e-6,
                )

            return hnext, y, imon, inln, hmin, hmax

        itask = 4
        itrace = -11

        ode_soln = self.fun(
            t, tout, y, comm, rtol, atol, itol, cb_fcn, itask, itrace,
            monitr=cb_monitr,
        )

        np.testing.assert_allclose(
            ode_soln.y,
            [8.413577e-01, 1.623291e-05, 1.586261e-01],
            rtol=1e-6,
        )

    def test_good(self):
        """
        Good call.
        """
        self.with_interp()

class TestD02qf(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the d02qf wrapper.
    """

class TestD02qfNp(TestD02qf, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    # pylint: disable=too-many-instance-attributes

    def setUp(self):
        TestD02qf.set_up(self)
        self.setup = chapter.ivp_adams_setup
        self.fun = chapter.ivp_adams_roots
        self.diagn = chapter.ivp_adams_diag
        self.diagnr = chapter.ivp_adams_rootdiag

        self.neqf = 2
        self.neqg = 2
        self.comm = {}

        self.tstart = 0.
        self.tcrit = 10.
        self.y = [0., 1.]

    def do_setup(self, vectol):
        """
        Set up the integrator.
        """
        self.setup(
            'S', self.neqf, vectol, [1.e-6]*(self.neqf if vectol else 1),
            [1.e-4]*(self.neqf if vectol else 1), False, True,
            self.tcrit, 0., 0, self.neqg, False, True, self.comm,
        )

    def do_integration(self):
        """
        Do the integration.
        """
        t = self.tstart
        tout = self.tcrit

        while True:
            t, self.y, root = self.fun(
                lambda x, y: [y[1], -y[0]],
                t, self.y, tout,
                self.neqg, self.comm,
                g=lambda x, y, yp, k: (yp[0] if k == 1 else y[0]),
            )
            if not root:
                break
            d = self.diagn(self.neqf, self.comm)
            dr = self.diagnr(self.neqg, self.comm)
            if d.tcurr >= tout:
                raise ValueError('NYI for tcurr >= tout') # pragma: no cover
                #break

        self.assertEqual(dr.index, 2)
        self.assertEqual(dr.itype, 4)

    def test_good_vec(self):
        """
        Good call, vectors of tolerances.
        """
        self.do_setup(True)
        self.do_integration()

    def test_good_scal(self):
        """
        Good call, scalars of tolerances.
        """
        self.do_setup(False)
        self.do_integration()

class TestD02ra(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the d02ra wrapper.
    """

class TestD02raNp(TestD02ra, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestD02ra.set_up(self)
        self.fun = chapter.bvp_fd_nonlin_gen
        the_np = 17
        mnp = 40
        x = np.empty(mnp)
        x[0] = 0.
        x[the_np-1] = 10.
        n = 3
        cb_fcn = lambda x, eps, y: [
            y[1], y[2], -y[0]*y[2] - 2.*(1.-y[1]*y[1])*eps,
        ]
        cb_g = lambda eps, ya, yb: [ya[0], ya[1], yb[1] - 1.]
        self.cb_jacobf = lambda x, eps, y: [
            [0., 1., 0.], [0., 0., 1.], [-y[2], 2.*2.*y[1]*eps, -y[0]],
        ]
        self.cb_jacobg = lambda eps, ya, yb: (
            [[1., 0., 0.], [0., 1., 0.], [0., 0., 0.]],
            [[0., 0., 0.], [0., 0., 0.], [0., 1., 0.]],
        )
        self.cb_jaceps = lambda x, eps, y: [0., 0., -2.*(1.-y[1]*y[1])]
        self.cb_jacgep = lambda eps, ya, yb: [0., 0., 0.]
        self.args = [
            the_np, # np, 0
            2, # numbeg, 1
            0, # nummix, 2
            1.e-4, # tol, 3
            0, # init, 4
            x, # x, 5
            np.empty((n, mnp)), # y, 6
            cb_fcn, # fcn, 7
            cb_g, # g, 8
            1, # ijac, 9
            1.e-1, # deleps, 10
            0, # itrace, 11
            {}, # comm, 12
        ]

    def test_good(self):
        """
        Good call.
        """
        soln = self.fun(
            *self.args,
            jacobf=self.cb_jacobf, jacobg=self.cb_jacobg,
            jaceps=self.cb_jaceps, jacgep=self.cb_jacgep
        )
        self.assertEqual(soln.np, 33)
        self.assertAlmostEqual(soln.y[0, soln.np-1], 9.5026263)

    def test_bad_cb_order(self):
        """
        Bad call.
        """

        def cb_jacobg(eps, ya, yb):
            aj, bj = self.cb_jacobg(eps, ya, yb)
            return aj, np.array(bj, order='F')

        self.my_assert_raises_regex(
            b_utils.NagTypeError,
            (
                'jacobg, returned bj.* arrays returned from this callback '
                'must use a consistent storage order'
            ),
            self.fun,
            *self.args,
            jacobf=self.cb_jacobf, jacobg=cb_jacobg,
            jaceps=self.cb_jaceps, jacgep=self.cb_jacgep
        )

class TestD02sa(IntegrationTest):
    """
    Abstract suite of integration tests for the d02sa wrapper.
    """

    @staticmethod
    def cb_brkpts(npoint, p, data):
        """
        Breakpoints.
        """
        return [0., p[2], data['xend']]

    @staticmethod
    def cb_bc(p, n, data):
        """
        Boundary conditions.
        """
        return (
            [0., 0.5, p[0]],
            [0., 0.45, -1.2],
        )

    @staticmethod
    def cb_fcn(x, y, p, i, data):
        """
        Derivative functions.
        """
        f = [math.tan(y[2])]
        if i == 1:
            f.extend([
                (
                    -data['alpha']*math.tan(y[2])/y[1] -
                    data['beta']*y[1]/math.cos(y[2])
                ),
                -data['alpha']/y[1]**2,
            ])
        else:
            f.extend([
                -p[1]*math.tan(y[2])/y[1] - p[3]*y[1]/math.cos(y[2]),
                -p[1]/y[1]**2,
            ])
        return f

    @staticmethod
    def cb_eqn(q, p, data):
        """
        Algebraic equations.
        """
        return [0.02 - p[3] - 1.0e-5*p[2]]

    @staticmethod
    def cb_constr(p, data):
        """
        Constraint satisfaction.
        """
        return all(p >= 0.) and p[2] <= data['xend']

class TestD02saNp(TestD02sa, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestD02sa.set_up(self)
        self.fun = chapter.bvp_shoot_genpar_algeq
        n = 3
        npoint = 3
        m = 4
        self.args = [
            [1.2, 0.032, 2.5, 0.02], # p, 0
            n, # n, 1
            n, # n1, 2
            [1.e-3]*m, # pe, 3
            [1.e-6]*m, # pf, 4
            [0.]*m, # dp, 5
            np.zeros((npoint, 6)), # swp, 6
            0, # icount, 7
            self.cb_brkpts, # brkpts, 8
            self.cb_bc, # bc, 9
            self.cb_fcn, # fcn, 10
            0., # ymax, 11
            0, # monlev, 12
            {}, # comm, 13
        ]
        self.data = {
            'alpha': 0.032,
            'beta': 0.02,
            'xend': 5.,
        }

    def test_good_constr(self):
        """
        Good call, supplied constr.
        """
        soln = self.fun(
            *self.args,
            eqn=self.cb_eqn, constr=self.cb_constr,
            data=self.data
        )
        np.testing.assert_allclose(
            soln.p,
            [1.1753313, 0.03045431, 2.33017326, 0.0199767],
        )

    def test_good_no_constr(self):
        """
        Good call, no constr.
        """
        soln = self.fun(
            *self.args,
            eqn=self.cb_eqn,
            data=self.data
        )
        np.testing.assert_allclose(
            soln.p,
            [1.1753313, 0.03045431, 2.33017326, 0.0199767],
        )

    def test_good_e(self):
        """
        Good call, supplied e.
        """
        soln = self.fun(
            *self.args,
            eqn=self.cb_eqn, constr=self.cb_constr,
            data=self.data, e=[1.e-5]*self.args[1]
        )
        np.testing.assert_allclose(
            soln.p,
            [1.1753313, 0.03045431, 2.33017326, 0.0199767],
        )

    def test_good_nagmon(self):
        """
        Good call, NAG-supplied monitoring.
        """
        tf_fd, tf_path = tempfile.mkstemp()
        tf_fo = open(tf_path, 'w')
        fom = b_utils.FileObjManager()
        fom.register_to_advunit(tf_fo)
        self.args[12] = 1
        soln = self.fun(
            *self.args,
            eqn=self.cb_eqn, constr=self.cb_constr,
            data=self.data, io_manager=fom
        )
        np.testing.assert_allclose(
            soln.p,
            [1.1753313, 0.03045431, 2.33017326, 0.0199767],
        )
        tf_fo.close()
        with open(tf_path, 'r') as test_fo:
            test_str_lines = test_fo.readlines()
        self.assertTrue(
            test_str_lines[0].startswith('naginterfaces.base.ode')
        )
        os.close(tf_fd)
        os.unlink(tf_path)

class TestD02colloc(IntegrationTest):
    """
    Abstract suite of integration tests for the collocation ODE solvers.
    """

    # pylint: disable=too-few-public-methods

    def set_up(self):
        # pylint: disable=attribute-defined-outside-init
        IntegrationTest.set_up(self)
        neq = 3
        nmesh = 11
        mxmesh = 51
        mesh = np.empty(mxmesh)
        mesh[:nmesh] = np.linspace(0., 1., nmesh)
        ipmesh = np.zeros(mxmesh, dtype=int)
        ipmesh[0] = 1
        ipmesh[1:nmesh-1] = 2
        ipmesh[nmesh-1] = 1
        self.init_args = [
            [1, 3, 2], # m, 0
            3, # nlbc, 1
            3, # nrbc, 2
            7, # ncol, 3
            [1.e-4]*neq, # tols, 4
            nmesh, # nmesh, 5
            mesh, # mesh, 6
            ipmesh, # ipmesh, 7
        ]

class TestD02tlNp(TestD02colloc, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestD02colloc.set_up(self)
        self.init = chapter.bvp_coll_nlin_setup
        self.fun = chapter.bvp_coll_nlin_solve
        self.diagnost = chapter.bvp_coll_nlin_diag

    def test_good(self):
        """
        Good call.
        """

        def cb_ffun(x, y, m):
            return [
                y[1, 0],
                -(y[0, 0]*y[1, 2] + y[2, 0]*y[2, 1])*sqrofr,
                (y[1, 0]*y[2, 0] - y[0, 0]*y[2, 1])*sqrofr,
            ]

        def cb_fjac(x, y, m, dfdy):
            return [
                [
                    [0., 0., 0.],
                    [1., 0., 0.],
                    [0., 0., 0.],
                ],
                [
                    [-y[1, 2]*sqrofr, 0., 0.],
                    [0., 0., -y[0, 0]*sqrofr],
                    [-y[2, 1]*sqrofr, -y[2, 0]*sqrofr, 0.],
                ],
                [
                    [-y[2, 1]*sqrofr, 0., 0.],
                    [y[2, 0]*sqrofr, 0., 0],
                    [y[1, 0]*sqrofr, -y[0, 0]*sqrofr, 0.],
                ],
            ]

        def cb_gafun(ya, m, nlbc):
            return [
                ya[0, 0],
                ya[1, 0],
                ya[2, 0] - omega,
            ]

        def cb_gbfun(yb, m, nrbc):
            return [
                yb[0, 0],
                yb[1, 0],
                yb[2, 0] + omega,
            ]

        def cb_gajac(ya, m, dgady):
            my_dgady = np.zeros(dgady.shape)
            my_dgady[0, 0, 0] = 1.
            my_dgady[1, 1, 0] = 1.
            my_dgady[2, 2, 0] = 1.
            return my_dgady

        def cb_gbjac(yb, m, dgbdy):
            my_dgbdy = np.zeros(dgbdy.shape)
            my_dgbdy[0, 0, 0] = 1.
            my_dgbdy[1, 1, 0] = 1.
            my_dgbdy[2, 2, 0] = 1.
            return my_dgbdy

        def cb_guess(x, m, mmax):
            y = [
                [-(x - 0.5)*(x*(x - 1.))**2, 0., 0.],
                [
                    -x*(x - 1.)*(5.*x*(x - 1.)+ 1.),
                    -(2.*x - 1.)*(10.*x*(x - 1.) + 1.),
                    -12.*(5.*x*(x - 1.) + x),
                ],
                [
                    -8.*omega*(x - 0.5)**3,
                    -24.*omega*(x - 0.5)**2,
                    0.,
                ],
            ]
            return (
                y,
                [
                    y[1][0],
                    -120.*(x - 0.5),
                    -56.*omega*(x -0.5),
                ]
            )

        omega = 1.
        r = 1.e6
        sqrofr = math.sqrt(r)
        comm = self.init(*self.init_args)
        self.fun(
            cb_ffun, cb_fjac, cb_gafun, cb_gbfun, cb_gajac, cb_gbjac, cb_guess,
            comm,
        )
        d = self.diagnost(len(self.init_args[6]), comm)
        self.assertEqual(
            d.nmesh, 21,
        )
        self.assertAlmostEqual(
            d.ermx, 6.e-10,
        )
        self.assertEqual(
            d.iermx, 20,
        )
        self.assertEqual(
            d.ijermx, 3,
        )

class TestD02tvNp(TestD02colloc, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestD02colloc.set_up(self)
        self.fun = chapter.bvp_coll_nlin_setup

    def test_good(self):
        """
        Good call.
        """
        comm = self.fun(*self.init_args)
        self.assertEqual(
            comm['icomm'][1], 77,
        )

if __name__ == '__main__': # pragma: no cover
    unittest.main()
