#!/usr/bin/env python
"""
Integration tests for wrapper functionality in the opt submodule.

Public use of this module is not supported by NAG.
"""

# NAG Copyright 2018-2020.

# pylint: disable=invalid-name,too-many-lines

import inspect
from math import cos, exp, sin, sqrt
import os
import tempfile
import unittest
import warnings

import numpy as np

from naginterfaces.base import utils as b_utils
from naginterfaces.base import opt as b_chapter, machine
from naginterfaces.library import opt as chapter
from test_functionality_core import IntegrationTest

class TestE04ab(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the e04ab wrapper.
    """

class TestE04abBase(TestE04ab, unittest.TestCase):
    """
    Concrete base tests.
    """

    def setUp(self):
        TestE04ab.set_up(self)
        self.fun = b_chapter.one_var_func

    def test_good(self):
        """
        Good call.
        """
        self.assertAlmostEqual(
            self.fun(lambda x: sin(x)/x, 0., 0., 3.5, 5., 100).x,
            4.493409453814,
        )

class TestE04bb(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the e04bb wrapper.
    """

class TestE04bbBase(TestE04bb, unittest.TestCase):
    """
    Concrete base tests.
    """

    def setUp(self):
        TestE04bb.set_up(self)
        self.fun = b_chapter.one_var_deriv

    def test_good(self):
        """
        Good call.
        """
        self.assertAlmostEqual(
            self.fun(
                lambda x: (sin(x)/x, (cos(x) - sin(x)/x)/x),
                0., 0., 3.5, 5., 100
            ).x,
            4.493409457909,
        )

class TestE04dg(IntegrationTest):
    """
    Abstract suite of integration tests for the e04dg wrapper.
    """

    # pylint: disable=too-few-public-methods

    fun = lambda *args: None # pragma: no cover
    init = lambda *args: None # pragma: no cover

class TestE04dgBase(TestE04dg, unittest.TestCase):
    """
    Concrete base tests.
    """

    def setUp(self):
        TestE04dg.set_up(self)
        self.fun = b_chapter.uncon_conjgrd_comp
        self.init = b_chapter.nlp1_init
        self.lineno = 0

    def default_args(self, n, objfun):
        """
        A default argument set.
        """
        # pylint: disable=attribute-defined-outside-init
        self.args = [
            n, # n, 0
            objfun, # objfun, 1
            [-42]*n, # objgrd, 2
            [-42]*n, # x, 3
            self.init('uncon_conjgrd_comp'), # comm, 4
        ]

    def badderiv_problem(self):
        """
        Bad derivatives programmed.
        """

        def cb_objfun(mode, n, x, objgrd, nstate):
            # pylint: disable=missing-docstring
            objf = np.linalg.norm(x)
            if mode == 2:
                objgrd[:] = [xi**3 for xi in x]
            return objf

        self.default_args(10, cb_objfun)

        self.lineno = inspect.getframeinfo(inspect.currentframe())[1] + 1
        self.fun(*self.args)

    def test_warn_as_warning(self):
        """
        A warning is displayed.
        """
        warnings.simplefilter('ignore', b_utils.NagDeprecatedWarning)
        with warnings.catch_warnings(record=True) as w_ctxt:
            self.badderiv_problem()
        warnings.simplefilter('always', b_utils.NagDeprecatedWarning)
        self.assertEqual(len(w_ctxt), 1)
        self.assertIs(w_ctxt[0].category, b_utils.NagAlgorithmicWarning)
        self.assertEqual(
            w_ctxt[0].filename,
            inspect.getframeinfo(inspect.currentframe())[0]
        )
        self.assertEqual(
            w_ctxt[0].lineno,
            self.lineno
        )

    def test_warn_as_exception(self):
        """
        A warning is raised.
        """
        warnings.simplefilter('ignore', b_utils.NagDeprecatedWarning)
        warnings.simplefilter('error', b_utils.NagAlgorithmicWarning)
        with self.assertRaises(b_utils.NagAlgorithmicWarning) as w_ctxt:
            self.badderiv_problem()
        warnings.simplefilter('always', b_utils.NagDeprecatedWarning)
        self.assertEqual(
            w_ctxt.exception.errno,
            7,
        )
        self.assertEqual(
            w_ctxt.exception._errcodes, # pylint: disable=protected-access
            (7, 7, 1),
        )
        self.assertIsInstance(w_ctxt.exception.return_data, tuple)
        self.assertEqual(
            w_ctxt.exception.return_data.__class__.__name__,
            'nag_return_data'
        )
        self.assertEqual(len(w_ctxt.exception.return_data), 2)
        self.assertEqual(
            w_ctxt.exception.return_data.itera,
            0
        )
        for objgrd_el in self.args[2]:
            self.assertNotEqual(objgrd_el, -42.)
        warnings.simplefilter('always', b_utils.NagAlgorithmicWarning)

    def test_cb_exception(self):
        """
        Callback raises an exception.

        Coverage for: the user raising an exception in a base callback.
        """

        def cb_objfun(mode, n, x, objgrd, nstate):
            # pylint: disable=missing-docstring
            raise ValueError('Hello')

        self.default_args(10, cb_objfun)

        warnings.simplefilter('ignore', b_utils.NagDeprecatedWarning)
        self.my_assert_raises_regex(
            ValueError,
            'Hello',
            self.fun, *self.args
        )
        warnings.simplefilter('default', b_utils.NagDeprecatedWarning)

    def test_user_data(self):
        """
        User data passed.

        Coverage for: non-None data in base wrappers.
        """

        def cb_objfun(mode, n, x, objgrd, nstate, data=None):
            # pylint: disable=missing-docstring,too-many-arguments
            objf = x[0]**2 + data['alpha']*x[0]
            if mode == 2:
                objgrd[:] = 2.*x[0] + data['alpha']
            else:
                objgrd[:] = np.zeros(x.shape)
            return objf

        self.default_args(1, cb_objfun)
        self.args[3] = [1.0]
        warnings.simplefilter('ignore', b_utils.NagDeprecatedWarning)
        self.fun(*self.args, data={'alpha': 42.})
        warnings.simplefilter('default', b_utils.NagDeprecatedWarning)
        self.assertEqual(self.args[3], [-21.])

class TestE04dgNp(TestE04dg, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestE04dg.set_up(self)
        self.fun = chapter.uncon_conjgrd_comp
        self.init = chapter.nlp1_init
        self.lineno = 0

    def default_args(self, n, objfun):
        """
        A default argument set.
        """
        # pylint: disable=attribute-defined-outside-init
        self.args = [
            objfun, # objfun, 0
            [-42.]*n, # x, 1
            self.init('uncon_conjgrd_comp'), # comm, 2
        ]

    def badderiv_problem(self):
        """
        Bad derivatives programmed.
        """

        def cb_objfun(mode, x, nstate):
            # pylint: disable=missing-docstring
            objf = np.linalg.norm(x)
            if mode == 2:
                objgrd = [xi**3 for xi in x]
            else:
                objgrd = np.zeros(x.shape, dtype=int)
            return objf, objgrd

        self.default_args(10, cb_objfun)

        self.lineno = inspect.getframeinfo(inspect.currentframe())[1] + 1
        self.fun(*self.args)

    def test_warn_as_warning(self):
        """
        A warning is displayed.
        """
        warnings.simplefilter('ignore', b_utils.NagDeprecatedWarning)
        with warnings.catch_warnings(record=True) as w_ctxt:
            self.badderiv_problem()
        warnings.simplefilter('default', b_utils.NagDeprecatedWarning)
        self.assertEqual(len(w_ctxt), 1)
        self.assertIs(w_ctxt[0].category, b_utils.NagAlgorithmicWarning)
        self.assertEqual(
            w_ctxt[0].filename,
            inspect.getframeinfo(inspect.currentframe())[0]
        )
        self.assertEqual(
            w_ctxt[0].lineno,
            self.lineno
        )

    def test_warn_as_exception(self):
        """
        A warning is raised.

        Coverage for: catching NagAlgorithmicWarning in level-2 wrappers and
                      reraising.
        """
        warnings.simplefilter('ignore', b_utils.NagDeprecatedWarning)
        warnings.simplefilter('error', b_utils.NagAlgorithmicWarning)
        with self.assertRaises(b_utils.NagAlgorithmicWarning) as w_ctxt:
            self.badderiv_problem()
        warnings.simplefilter('default', b_utils.NagDeprecatedWarning)
        self.assertEqual(
            w_ctxt.exception.errno,
            7,
        )
        self.assertEqual(
            w_ctxt.exception._errcodes, # pylint: disable=protected-access
            (7, 7, 1),
        )
        self.assertIsInstance(w_ctxt.exception.return_data, tuple)
        self.assertEqual(
            w_ctxt.exception.return_data.__class__.__name__,
            'nag_return_data'
        )
        self.assertEqual(len(w_ctxt.exception.return_data), 4)
        self.assertEqual(
            w_ctxt.exception.return_data.itera,
            0
        )
        for objgrd_el in w_ctxt.exception.return_data.objgrd:
            self.assertNotEqual(objgrd_el, -42.)
        warnings.simplefilter('always', b_utils.NagAlgorithmicWarning)

    def test_cb_warn_as_exception(self):
        """
        A callback warning is raised.
        """
        warnings.simplefilter('error', b_utils.NagCallbackTerminateWarning)

        def cb_objfun(mode, x, nstate):
            # pylint: disable=missing-docstring
            raise b_utils.UserCallbackTerminate

        with self.assertRaises(b_utils.NagCallbackTerminateWarning) as w_ctxt:
            warnings.simplefilter('ignore', b_utils.NagDeprecatedWarning)
            self.default_args(10, cb_objfun)
            self.fun(*self.args)

        self.assertEqual(
            w_ctxt.exception.errno,
            -1,
        )
        self.assertEqual(
            w_ctxt.exception._errcodes, # pylint: disable=protected-access
            (-1, -1, 1),
        )
        self.assertIsInstance(w_ctxt.exception.return_data, tuple)
        self.assertEqual(
            w_ctxt.exception.return_data.__class__.__name__,
            'nag_return_data'
        )
        self.assertEqual(len(w_ctxt.exception.return_data), 4)
        self.assertEqual(
            w_ctxt.exception.return_data.itera,
            0
        )
        self.assertIs(
            w_ctxt.exception.cb_excs[0][0],
            b_utils.UserCallbackTerminate,
        )
        warnings.simplefilter('always', b_utils.NagCallbackTerminateWarning)

    def test_user_data(self):
        """
        User data passed.

        Coverage for: non-None data in level-2 wrappers.
        """

        def cb_objfun(mode, x, nstate, data=None):
            # pylint: disable=missing-docstring,too-many-arguments
            objf = x[0]**2 + data['alpha']*x[0]
            if mode == 2:
                objgrd = 2.*x[0] + data['alpha']
            else:
                objgrd = 0
            return objf, objgrd

        self.default_args(1, cb_objfun)
        self.args[1] = 1.0
        warnings.simplefilter('ignore', b_utils.NagDeprecatedWarning)
        self.assertEqual(
            self.fun(*self.args, data={'alpha': 42.}).x,
            [-21.],
        )
        warnings.simplefilter('default', b_utils.NagDeprecatedWarning)

    def test_bad_cb_return_rank(self):
        """
        Bad data returned from callback.
        """

        def cb_objfun(mode, x, nstate, data=None):
            # pylint: disable=missing-docstring,too-many-arguments
            objf = x**2 + x
            objgrd = [2.*x + 1.]
            return objf, objgrd

        self.default_args(1, cb_objfun)
        self.args[1] = 1.0
        warnings.simplefilter('ignore', b_utils.NagDeprecatedWarning)
        self.my_assert_raises_regex(
            b_utils.NagShapeError,
            '.*returned objgrd.*must have rank',
            self.fun, *self.args
        )
        warnings.simplefilter('default', b_utils.NagDeprecatedWarning)

    def test_bad_cb_return_shape(self):
        """
        Bad data returned from callback.
        """

        def cb_objfun(mode, x, nstate, data=None):
            # pylint: disable=missing-docstring,too-many-arguments
            objf = x[0]**2 + x[0]
            objgrd = [2.*x[0], 1.]
            return objf, objgrd

        self.default_args(1, cb_objfun)
        self.args[1] = 1.0
        warnings.simplefilter('ignore', b_utils.NagDeprecatedWarning)
        self.my_assert_raises_regex(
            b_utils.NagShapeError,
            '.*returned objgrd.*must have length',
            self.fun, *self.args
        )
        warnings.simplefilter('default', b_utils.NagDeprecatedWarning)

class TestE04dk(IntegrationTest):
    """
    Abstract suite of integration tests for the e04dk wrapper.
    """

    # pylint: disable=too-few-public-methods

    fun = lambda *args: None # pragma: no cover

class TestE04dkBase(TestE04dk, unittest.TestCase):
    """
    Concrete base tests.
    """

    def setUp(self):
        TestE04dk.set_up(self)
        self.fun = b_chapter.uncon_conjgrd_option_string
        self.init = b_chapter.nlp1_init

    def test_bad_optstr(self):
        """
        Invalid optstr passed.

        Coverage for: handling inform from the Engine.
        Coverage for: _STR_FMT_STR.
        """
        warnings.simplefilter('ignore', b_utils.NagDeprecatedWarning)
        self.my_assert_raises_regex(
            b_utils.NagValueError,
            'supplied option .* invalid.*\n.*foo',
            self.fun, 'foo', self.init('uncon_conjgrd_comp'),
        )
        warnings.simplefilter('default', b_utils.NagDeprecatedWarning)

class TestE04fc(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the e04fc wrapper.
    """

class TestE04fcBase(TestE04fc, unittest.TestCase):
    """
    Concrete base tests.
    """

    def setUp(self):
        TestE04fc.set_up(self)
        self.fun = b_chapter.lsq_uncon_mod_func_comp

    def test_good(self):
        """
        Good call.
        """
        y = [
            0.14, 0.18, 0.22, 0.25, 0.29, 0.32, 0.35,
            0.39, 0.37, 0.58, 0.73, 0.96, 1.34, 2.10, 4.39,
        ]
        t1 = [
            1, 2, 3, 4, 5, 6, 7,
            8, 9, 10, 11, 12, 13, 14, 15,
        ]
        t2 = [
            15, 14, 13, 12, 11, 10,
            9, 8, 7, 6, 5, 4, 3, 2, 1,
        ]
        t3 = [
            1, 2, 3, 4, 5, 6,
            7, 8, 7, 6, 5, 4, 3, 2, 1,
        ]

        def cb_lsqfun(iflag, m, n, x, fvec):
            # pylint: disable=missing-docstring
            for i in range(m):
                fvec[i] = x[0] + t1[i]/(x[1]*t2[i] + x[2]*t3[i]) - y[i]
            return iflag

        x = [0.5, 1, 1.5]
        fvec = np.empty(len(y))
        fjac = np.empty((len(y), len(x)))
        s = np.empty(len(x))
        v = np.empty((len(x), len(x)))
        self.assertAlmostEqual(
            self.fun(
                len(y), len(x), cb_lsqfun, None, -1, 400*len(x),
                0.5, 10.*sqrt(machine.precision()),
                10., x, fvec, fjac, s, v,
            ).fsumsq,
            8.215e-03,
            places=6,
        )

class TestE04ff(IntegrationTest):
    """
    Abstract suite of integration tests for the e04ff wrapper.
    """

    # pylint: disable=too-few-public-methods

    fun = lambda *args: None # pragma: no cover
    init = lambda *args: None # pragma: no cover

class TestE04ffNp(TestE04ff, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestE04ff.set_up(self)
        self.fun = chapter.handle_solve_dfls
        self.init = chapter.handle_init
        self.optset = chapter.handle_opt_set
        self.fin = chapter.handle_free

    def test_mon(self): # pylint: disable=too-many-locals
        """
        Monitoring function supplied.
        """

        def cb_objfun(x, nres):
            # pylint: disable=missing-docstring
            y = [
                4., 2., 1., 0.5, 0.25, 0.167, 0.125, 0.1, 0.0833, 0.0714,
                0.0625
            ]
            z = [
                0.1957, 0.1947, 0.1735, 0.1600, 0.0844, 0.0627, 0.0456, 0.0342,
                0.0323, 0.0235, 0.0246
            ]
            return z - x[0]*(y*(y+x[1]))/(y*(y+x[2]) + x[3])

        m_fd, m_path = tempfile.mkstemp()
        m_fo = open(m_path, 'w')

        def cb_mon(x, rinfo, stats):
            # pylint: disable=missing-docstring
            m_fo.write('Hello from mon\n')

        x = [0.25, 0.39, 0.415, 0.39]
        handle = self.init(nvar=len(x))
        nres = 11
        chapter.handle_set_nlnls(handle, nres, None, None)
        chapter.handle_set_simplebounds(
            handle,
            bl=[-1.e20, 0.2, -1.e20, 0.3],
            bu=[1.e20, 1., 1.e20, 1.e20],
        )
        self.optset(handle, 'Print File = -1')
        self.optset(handle, 'Print Options = NO')
        self.optset(handle, 'Print Solution = NO')
        self.optset(handle, 'DFLS Print Frequency = 0')
        self.optset(handle, 'Monitoring File = -1')
        self.optset(handle, 'Monitoring Level = 0')
        self.optset(handle, 'DFLS Monitor Frequency = 1')
        self.fun(handle, cb_objfun, x, nres, monit=cb_mon)
        self.fin(handle)
        m_fo.close()
        m_fo = open(m_path, 'r')
        self.assertEqual(
            m_fo.readline(),
            'Hello from mon\n',
        )
        m_fo.close()
        os.close(m_fd)
        os.unlink(m_path)

class TestE04jc(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the e04jc wrapper.
    """

class TestE04jcNp(TestE04jc, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestE04jc.set_up(self)
        self.fun = chapter.bounds_bobyqa_func

    def test_good(self):
        """
        Good call.
        """
        x = [3.0, -1.0, 0.0, 1.0]
        cb_obj_func = lambda x: (
            (x[0] + 10*x[1])**2 +
            5*(x[2] - x[3])**2 +
            (x[1] - 2*x[2])**4 +
            10*(x[0] - x[3])**4
        )
        npt = 2*len(x) + 1
        bl = [1.0, -2.0, -1e10, 1.0]
        bu = [3.0, 0.0, 1e10, 3.0]
        rhobeg = 1e-1
        rhoend = 1e-6
        maxcal = 500
        warnings.simplefilter('ignore', b_utils.NagDeprecatedWarning)
        soln = self.fun(
            cb_obj_func, npt, x, bl, bu, rhobeg, rhoend, maxcal,
        )
        warnings.simplefilter('always', b_utils.NagDeprecatedWarning)
        self.assertAlmostEqual(
            soln.f, 2.4337875,
        )

class TestE04mf(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the e04mf wrapper.
    """

class TestE04mfNp(TestE04mf, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestE04mf.set_up(self)
        self.init = chapter.nlp1_init
        self.fun = chapter.lp_solve

    def test_good(self):
        """
        Good call.
        """
        cvec = [-0.02, -0.20, -0.20, -0.20, -0.20, 0.04, 0.04]
        a = [
            [1.00, 1.00, 1.00, 1.00, 1.00, 1.00, 1.00],
            [0.15, 0.04, 0.02, 0.04, 0.02, 0.01, 0.03],
            [0.03, 0.05, 0.08, 0.02, 0.06, 0.01, 0.00],
            [0.02, 0.04, 0.01, 0.02, 0.02, 0.00, 0.00],
            [0.02, 0.03, 0.00, 0.00, 0.01, 0.00, 0.00],
            [0.70, 0.75, 0.80, 0.75, 0.80, 0.97, 0.00],
            [0.02, 0.06, 0.08, 0.12, 0.02, 0.01, 0.97],
        ]
        bl = [
            -0.01, -0.10, -0.01, -0.04, -0.10, -0.01, -0.01, -0.13,
            -1.0E+25, -1.0E+25, -1.0E+25, -1.0E+25, -9.92E-02, -3.0E-03,
        ]
        bu = [
            0.01, 0.15, 0.03, 0.02, 0.05, 1.0E+25, 1.0E+25, -0.13,
            -4.9E-03, -6.4E-03, -3.7E-03, -1.2E-03, 1.0E+25, 2.0E-03,
        ]
        x = [-0.01, -0.03, 0.00, -0.01, -0.10, 0.02, 0.01]

        comm = self.init('lp_solve')

        istate = np.zeros(len(x) + len(a), dtype=int)

        soln = self.fun(len(a), a, bl, bu, cvec, istate, x, comm)

        self.assertAlmostEqual(soln.obj, 0.02359648208469047)

class TestE04mt(IntegrationTest):
    """
    Abstract suite of integration tests for the e04mt wrapper.
    """

    # pylint: disable=too-few-public-methods

    def l2_solve(self):
        """
        Transcribed Library example, level-2 calls.
        """

        # pylint: disable=attribute-defined-outside-init

        self.set_up()

        self.fun = chapter.handle_solve_lp_ipm
        self.init = chapter.handle_init
        self.optset = chapter.handle_opt_set
        self.fin = chapter.handle_free

        self.n = 7
        self.handle = self.init(self.n)

        chapter.handle_set_quadobj(
            self.handle,
            idxc=list(range(1, self.n+1)),
            c=[-0.02, -0.2, -0.2, -0.2, -0.2, 0.04, 0.04],
            irowh=None, icolh=None, h=None,
        )

        chapter.handle_set_simplebounds(
            self.handle,
            [-0.01, -0.1, -0.01, -0.04, -0.1, -0.01, -0.01],
            [0.01, 0.15, 0.03, 0.02, 0.05, 1.e20, 1.e20],
        )

        chapter.handle_set_linconstr(
            self.handle,
            [-0.13, -1.e20, -1.e20, -1.e20, -1.e20, -0.0992, -0.003],
            [-0.13, -0.0049, -0.0064, -0.0037, -0.0012, 1.e20, 0.002],
            [
                1, 1, 1, 1, 1, 1, 1,
                2, 2, 2, 2, 2, 2, 2,
                3, 3, 3, 3, 3, 3,
                4, 4, 4, 4, 4,
                5, 5, 5,
                6, 6, 6, 6, 6, 6,
                7, 7, 7, 7, 7, 7, 7,
            ],
            [
                1, 2, 3, 4, 5, 6, 7,
                1, 2, 3, 4, 5, 6, 7,
                1, 2, 3, 4, 5, 6,
                1, 2, 3, 4, 5,
                1, 2, 5,
                1, 2, 3, 4, 5, 6,
                1, 2, 3, 4, 5, 6, 7,
            ],
            [
                1., 1., 1., 1., 1., 1., 1.,
                0.15, 0.04, 0.02, 0.04, 0.02, 0.01, 0.03,
                0.03, 0.05, 0.08, 0.02, 0.06, 0.01,
                0.02, 0.04, 0.01, 0.02, 0.02,
                0.02, 0.03, 0.01,
                0.7, 0.75, 0.8, 0.75, 0.8, 0.97,
                0.02, 0.06, 0.08, 0.12, 0.02, 0.01, 0.97,
            ],
        )

        for option in [
                'Print Options = NO',
                'Print Level = 0',
                'LPIPM Stop Tolerance = 1.0e-10',
                'LPIPM Centrality Correctors = -6',
        ]:
            self.optset(self.handle, option)

        self.soln = self.fun(self.handle)

class TestE04mtNp(TestE04mt, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def test_good(self):
        """
        Good call.
        """
        self.l2_solve()
        self.assertAlmostEqual(
            self.soln.rinfo[0],
            2.359648e-02,
        )
        self.assertAlmostEqual(self.soln.x[0], -0.01)
        self.assertEqual(self.soln.u[1], 0.)
        self.fin(self.handle)

class TestE04nc(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the e04nc wrapper.
    """

class TestE04ncNp(TestE04nc, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestE04nc.set_up(self)
        self.init = chapter.nlp1_init
        self.fun = chapter.lsq_lincon_solve

    def test_good(self):
        """
        Good call.
        """
        a = [
            [1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
            [1.0, 2.0, 1.0, 1.0, 1.0, 1.0, 2.0, 0.0, 0.0],
            [1.0, 1.0, 3.0, 1.0, 1.0, 1.0, -1.0, -1.0, -3.0],
            [1.0, 1.0, 1.0, 4.0, 1.0, 1.0, 1.0, 1.0, 1.0],
            [1.0, 1.0, 1.0, 3.0, 1.0, 1.0, 1.0, 1.0, 1.0],
            [1.0, 1.0, 2.0, 1.0, 1.0, 0.0, 0.0, 0.0, -1.0],
            [1.0, 1.0, 1.0, 1.0, 0.0, 1.0, 1.0, 1.0, 1.0],
            [1.0, 1.0, 1.0, 0.0, 1.0, 1.0, 1.0, 1.0, 1.0],
            [1.0, 1.0, 0.0, 1.0, 1.0, 1.0, 2.0, 2.0, 3.0],
            [1.0, 0.0, 1.0, 1.0, 1.0, 1.0, 0.0, 2.0, 2.0],
        ]
        b = np.ones(len(a))
        c = [
            [1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 4.0],
            [1.0, 2.0, 3.0, 4.0, -2.0, 1.0, 1.0, 1.0, 1.0],
            [1.0, -1.0, 1.0, -1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
        ]
        bl = [
            0.0, 0.0, -1.0E+25, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 2.0, -1.0E+25,
            1.0,
        ]
        bu = [2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 1.0E+25, 2.0, 4.0]
        x = [1.0, 0.5, 0.3333, 0.25, 0.2, 0.1667, 0.1428, 0.125, 0.1111]

        comm = self.init('lsq_lincon_solve')

        cvec = np.empty(1)
        istate = np.zeros(len(x) + len(c), dtype=int)
        kx = np.zeros(len(x), dtype=int)

        soln = self.fun(c, bl, bu, cvec, istate, kx, x, a, b, comm)

        self.assertAlmostEqual(soln.obj, 0.08134082317338462)

class TestE04nf(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the e04nf wrapper.
    """

class TestE04nfNp(TestE04nf, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestE04nf.set_up(self)
        self.init = chapter.nlp1_init
        self.fun = chapter.qp_dense_solve

    def run_problem(self, sorder):
        """
        Test the solver.
        """
        a = np.array([
            [1.00, 1.00, 1.00, 1.00, 1.00, 1.00, 1.00],
            [0.15, 0.04, 0.02, 0.04, 0.02, 0.01, 0.03],
            [0.03, 0.05, 0.08, 0.02, 0.06, 0.01, 0.00],
            [0.02, 0.04, 0.01, 0.02, 0.02, 0.00, 0.00],
            [0.02, 0.03, 0.00, 0.00, 0.01, 0.00, 0.00],
            [0.70, 0.75, 0.80, 0.75, 0.80, 0.97, 0.00],
            [0.02, 0.06, 0.08, 0.12, 0.02, 0.01, 0.97],
        ], order=sorder)
        nclin = len(a)
        bl = [
            -0.01, -0.10, -0.01, -0.04, -0.10, -0.01, -0.01,
            -0.13, -1.0e+25, -1.0e+25, -1.0e+25, -1.0e+25, -9.92e-02, -3.0e-03,
        ]
        bu = [
            0.01, 0.15, 0.03, 0.02, 0.05, 1.0e+25, 1.0e+25,
            -0.13, -4.9e-03, -6.4e-03, -3.7e-03, -1.2e-03, 1.0e+25, 2.0e-03,
        ]
        cvec = [-0.02, -0.20, -0.20, -0.20, -0.20, 0.04, 0.04]
        h = np.array([
            [2, 0, 0, 0, 0, 0, 0],
            [-42, 2, 0, 0, 0, 0, 0],
            [-42, -42, 2, 2, 0, 0, 0],
            [-42, -42, -42, 2, 0, 0, 0],
            [-42, -42, -42, -42, 2, 0, 0],
            [-42, -42, -42, -42, -42, -2, -2],
            [-42, -42, -42, -42, -42, -42, -2],
        ], order=sorder)
        x = [-0.01, -0.03, 0.00, -0.01, -0.10, 0.02, 0.01]
        istate = np.zeros(len(x) + nclin, dtype=int)
        comm = self.init('qp_dense_solve')
        soln = self.fun(nclin, a, bl, bu, cvec, h, istate, x, comm)
        self.assertEqual(soln.itera, 7)
        self.assertAlmostEqual(soln.obj, 0.0370316458970542)

    def test_good_c(self):
        """
        Good call, row order.
        """
        self.run_problem('C')

    def test_good_f(self):
        """
        Good call, column order.
        """
        self.run_problem('F')

class TestE04nk(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the e04nk wrapper.
    """

class TestE04nkBase(TestE04nk, unittest.TestCase):
    """
    Concrete base tests.
    """

    def setUp(self):
        TestE04nk.set_up(self)
        self.fun = b_chapter.qpconvex1_sparse_solve
        self.init = b_chapter.nlp1_init

    def test_good(self):
        """
        Good call.
        """
        # pylint: disable=too-many-locals
        n = 7
        m = 8
        nnz = 48
        iobj = 8
        ncolh = 7

        def cb_qphx(_nstate, ncolh, x, hx):
            # pylint: disable=missing-docstring

            hx[0] = x[0] * 2
            hx[1] = x[1] * 2
            hx[2] = (x[2] + x[3]) * 2
            hx[3] = hx[2]
            hx[4] = x[4] * 2
            hx[5] = (x[5] + x[6]) * 2
            hx[6] = hx[5]

        a = [
            0.02, 0.02, 0.03, 1.00, 0.70, 0.02, 0.15, -200.00,
            0.06, 0.75, 0.03, 0.04, 0.05, 0.04, 1.00, -2000.00,
            0.02, 1.00, 0.01, 0.08, 0.08, 0.80, -2000.00,
            1.00, 0.12, 0.02, 0.02, 0.75, 0.04, -2000.00,
            0.01, 0.80, 0.02, 1.00, 0.02, 0.06, 0.02, -2000.00,
            1.00, 0.01, 0.01, 0.97, 0.01, 400.00,
            0.97, 0.03, 1.00, 400.00,
        ]
        ha = [
            7, 5, 3, 1, 6, 4, 2, 8, 7, 6, 5, 4, 3, 2, 1, 8, 2, 1, 4,
            3, 7, 6, 8, 1, 7, 3, 4, 6, 2, 8, 5, 6, 7, 1, 2, 3, 4, 8,
            1, 2, 3, 6, 7, 8, 7, 2, 1, 8,
        ]
        ka = [
            indx + 1 for indx in [
                0, 8, 16, 23, 30, 38, 44, 48,
            ]
        ]
        bl = [
            0.0, 0.0, 4.0E+02, 1.0E+02, 0.0,
            0.0, 0.0, 2.0E+03, -1.0E+20, -1.0E+20,
            -1.0E+20, -1.0E+20, 1.5E+03, 2.5E+02, -1.0E+20,
        ]
        bu = [
            2.0E+02, 2.5E+03, 8.0E+02, 7.0E+02, 1.5E+03,
            1.0E+20, 1.0E+20, 2.0E+03, 6.0E+01, 1.0E+02,
            4.0E+01, 3.0E+01, 1.0E+20, 3.0E+02, 1.0E+20,
        ]
        start = 'C'
        names = [' '*8]*5
        nname = 15
        crname = [
            '...X1...', '...X2...', '...X3...', '...X4...', '...X5...',
            '...X6...', '...X7...', '..ROW1..', '..ROW2..', '..ROW3..',
            '..ROW4..', '..ROW5..', '..ROW6..', '..ROW7..', '..COST..',
        ]
        xs = [0.]*(n+m)
        istate = [0]*(n+m)
        clamda = np.empty(n+m)
        comm = self.init('qpconvex1_sparse_solve')
        self.assertAlmostEqual(
            self.fun(
                n, m, nnz, iobj, ncolh, cb_qphx, a, ha, ka, bl, bu, start,
                names, nname, crname,
                0, xs, istate, clamda, 428, 358, comm,
            ).obj,
            -1847784.6771229,
        )

class TestE04rm(IntegrationTest):
    """
    Abstract suite of integration tests for the e04rm wrapper.
    """

    # pylint: disable=too-few-public-methods

    fun = lambda *args: None # pragma: no cover

class TestE04rmNp(TestE04rm, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestE04rm.set_up(self)
        self.fun = chapter.handle_set_nlnls
        self.init = chapter.handle_init

    def test_bad_none_rows(self):
        """
        None rows, some cols.
        """
        self.my_assert_raises_regex(
            b_utils.NagValueError,
            'All of the following must be provided',
            self.fun, self.init(42), 43, irowrd=None, icolrd=[44],
        )

    def test_bad_none_cols(self):
        """
        None cols, some rows.
        """
        self.my_assert_raises_regex(
            b_utils.NagValueError,
            'All of the following must be provided',
            self.fun, self.init(42), 43, irowrd=[44], icolrd=None,
        )

    def test_good(self):
        """
        Good sparsity pattern supplied.
        """
        self.fun(self.init(42), 3, [1, 2, 1, 2, 1, 2], [1, 1, 2, 2, 3, 3])

class TestE04rx(TestE04mt):
    """
    Abstract suite of integration tests for the e04rx wrapper.
    """

    # pylint: disable=too-few-public-methods

    exp_rarr = [
        -0.009999999997653546,
        -0.09999999999529921,
        0.030000000005037403,
        0.019999999990320518,
        -0.0674853419527556,
        -0.002280130331370696,
        -0.00023452768580103678,
    ]

class TestE04rxBase(TestE04rx, unittest.TestCase):
    """
    Concrete base tests.
    """

    def test_good_read(self):
        """
        Good call.
        """
        self.l2_solve()
        rarr = [0.]*self.n
        b_chapter.handle_set_get_real(
            self.handle, 'Primal Variables',
            ioflag=1,
            lrarr=len(rarr), rarr=rarr,
        )
        np.testing.assert_allclose(
            rarr,
            self.exp_rarr,
        )

    def test_query(self):
        """
        rarr too small.
        """
        self.l2_solve()
        rarr = []
        warnings.simplefilter('ignore', b_utils.NagAlgorithmicWarning)
        lrarr = b_chapter.handle_set_get_real(
            self.handle, 'Primal Variables',
            ioflag=1,
            lrarr=len(rarr), rarr=rarr,
        )
        warnings.simplefilter('always', b_utils.NagAlgorithmicWarning)
        self.assertEqual(lrarr, self.n)

class TestE04rxNp(TestE04rx, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def test_good_read(self):
        """
        Good call.
        """
        self.l2_solve()
        lrarr, rarr = chapter.handle_set_get_real(
            self.handle, 'Primal Variables',
        )
        self.assertEqual(lrarr, len(rarr))
        np.testing.assert_allclose(
            rarr,
            self.exp_rarr,
        )

    def test_nyi_write(self):
        """
        Write not implemented.
        """
        self.l2_solve()
        self.my_assert_raises_regex(
            b_utils.NagValueError,
            'Writing mode is not supported',
            chapter.handle_set_get_real,
            self.handle, 'Primal Variables',
            rarr=[42.],
        )

class TestE04st(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the e04st wrapper.
    """

class TestE04stNp(TestE04st, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestE04st.set_up(self)
        self.init = chapter.handle_init
        self.opt_set = chapter.handle_opt_set
        self.fun = chapter.handle_solve_ipopt
        self.set_nlnobj = chapter.handle_set_nlnobj

    def test_bad_objgrd(self):
        """
        None objgrd, but must be none-None.
        """
        handle = self.init(1)
        self.opt_set(
            handle, 'Print Level = 0',
        )
        self.set_nlnobj(handle, [1])
        x = 42.
        self.my_assert_raises_regex(
            b_utils.NagValueError,
            'provide a proper objgrd',
            self.fun,
            handle, x, 0,
        )

class TestE04uc(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the e04uc wrapper.
    """

class TestE04ucNp(TestE04uc, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestE04uc.set_up(self)
        self.init = chapter.nlp1_init
        self.fun = chapter.nlp1_solve

    def test_bounds(self):
        """
        No linear or nonlinear constraints.
        """
        cb_objfun = lambda _mode, x, _objgrd, _nstate: (
            x[0]**2, 2.*x,
        )
        comm = self.init('nlp1_solve')
        x = [1.]
        n = len(x)
        a = np.zeros((0, n))
        bl = -1.
        bu = 1.
        warnings.simplefilter('ignore', b_utils.NagAlgorithmicWarning)
        self.assertAlmostEqual(
            self.fun(
                a, bl, bu, cb_objfun, x, comm,
            ).objf,
            0.,
        )
        warnings.simplefilter('always', b_utils.NagAlgorithmicWarning)

class TestE04us(IntegrationTest):
    """
    Abstract suite of integration tests for the e04us wrapper.
    """

    # pylint: disable=too-few-public-methods

    fun = lambda *args: None # pragma: no cover
    init = lambda *args: None # pragma: no cover

    @staticmethod
    def osborne_objfun(_mode, _needfi, x, fjac, _nstate):
        # pylint: disable=missing-docstring,too-many-arguments,too-many-locals

        x1, x2, x3, x4, x5 = x

        m = fjac.shape[0]
        f = np.empty(m)

        for i in range(m):
            ti = 10.*i
            e4 = exp(-ti*x4)
            e5 = exp(-ti*x5)
            f[i] = (x1+x2*e4+x3*e5)
            fjac[i, :] = [1., e4, e5, -ti*x2*e4, -ti*x3*e5]

        return f

    @staticmethod
    def osborne_data():
        # pylint: disable=missing-docstring
        x = [0.5, 1.5, -1., 0.01, 0.02]
        bl = [0.2, 1.9, -1.e25, 0.01, 0.01]
        bu = [0.4, 1.e25, 1.e25, 0.05, 0.05]
        y = [
            0.844,
            0.908,
            0.932,
            0.936,
            0.925,
            0.908,
            0.881,
            0.850,
            0.818,
            0.784,
            0.751,
            0.718,
            0.685,
            0.658,
            0.628,
            0.603,
            0.580,
            0.558,
            0.538,
            0.522,
            0.506,
            0.490,
            0.478,
            0.467,
            0.457,
            0.448,
            0.438,
            0.431,
            0.424,
            0.420,
            0.414,
            0.411,
            0.406,
        ]
        return x, bl, bu, y

class TestE04usBase(TestE04us, unittest.TestCase):
    """
    Concrete base tests.
    """

    def setUp(self):
        TestE04us.set_up(self)
        self.fun = b_chapter.lsq_gencon_deriv
        self.init = b_chapter.nlp1_init
        self.optset = b_chapter.lsq_gencon_deriv_option_string

    def test_no_constraint_warm(self): # pylint: disable=too-many-locals
        """
        No constraints, second call with warm start and reduced derivative
        level.
        """

        def cb_objfun(mode, m, n, needfi, x, f, fjac, nstate):
            # pylint: disable=missing-docstring,too-many-arguments
            f[:] = self.osborne_objfun(mode, needfi, x, fjac, nstate)

        x, bl, bu, y = self.osborne_data()
        m = 33
        n = 5
        nclin = 0
        a = np.empty((nclin, 1))
        ncnln = 0
        comm = self.init('lsq_gencon_deriv')
        istate = np.empty(n, dtype=int)
        c = np.empty(ncnln)
        cjac = np.empty((ncnln, 1))
        f = np.empty(m)
        fjac = np.empty((m, n))
        clamda = np.empty(len(bl))
        r = np.empty((n, n))
        for i in range(2):
            if i == 1:
                self.optset('Warm Start', comm)
                self.optset('Derivative Level = 1', comm)
                warnings.simplefilter('ignore', b_utils.NagAlgorithmicWarning)
            _, objf = self.fun(
                m, n, nclin, ncnln, a, bl, bu, y, None, cb_objfun,
                istate, c, cjac, f, fjac, clamda, r, x, comm
            )
            self.assertAlmostEqual(
                objf, 2.73244e-5
            )
        warnings.simplefilter('always', b_utils.NagAlgorithmicWarning)

    def test_no_nonlin_cons(self): # pylint: disable=too-many-locals
        """
        Missing nonlinear constraints.
        """
        def cb_objfun(mode, m, n, needfi, x, f, fjac, nstate):
            # pylint: disable=missing-docstring,too-many-arguments
            pass # pragma: no cover

        x = [42.]
        m = 1
        n = 1
        nclin = 0
        a = np.empty((nclin, 1))
        ncnln = 1
        bl = [42., 43.]
        bu = [42., 43.]
        y = [42.]
        comm = self.init('lsq_gencon_deriv')
        istate = np.empty(n, dtype=int)
        c = np.empty(ncnln)
        cjac = np.empty((1, 1))
        f = np.empty(m)
        fjac = np.empty((m, n))
        clamda = np.empty(len(bl))
        r = np.empty((n, n))
        self.my_assert_raises_regex(
            b_utils.NagTypeError,
            '.*confun. must not be None when ncnln != 0',
            self.fun,
            m, n, nclin, ncnln, a, bl, bu, y, None, cb_objfun,
            istate, c, cjac, f, fjac, clamda, r, x, comm
        )

class TestE04usNp(TestE04us, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestE04us.set_up(self)
        self.fun = chapter.lsq_gencon_deriv
        self.init = chapter.nlp1_init
        self.optset = chapter.lsq_gencon_deriv_option_string

    def test_no_constraint_warm(self): # pylint: disable=too-many-locals
        """
        No constraints, second call with warm start and reduced derivative
        level.
        """
        x, bl, bu, y = self.osborne_data()
        comm = self.init('lsq_gencon_deriv')
        rtn_data = self.fun(
            bl, bu, y, self.osborne_objfun, x, comm
        )
        self.assertAlmostEqual(
            rtn_data.objf, 2.73244e-5
        )
        self.optset('Warm Start', comm)
        self.optset('Derivative Level = 1', comm)
        rtn_data = self.fun(
            bl, bu, y, self.osborne_objfun, x, comm,
            istate=rtn_data.istate, clamda=rtn_data.clamda, r=rtn_data.r,
        )
        self.assertAlmostEqual(
            rtn_data.objf, 2.73244e-5
        )
        self.assertIsNone(rtn_data.c)
        self.assertIsNone(rtn_data.cjac)

    def test_no_istate_warm(self): # pylint: disable=too-many-locals
        """
        Warm start but istate not supplied.
        """
        x, bl, bu, y = self.osborne_data()
        comm = self.init('lsq_gencon_deriv')
        self.optset('Warm Start', comm)
        self.my_assert_raises_regex(
            b_utils.NagTypeError,
            '.*istate. must not be None',
            self.fun,
            bl, bu, y, self.osborne_objfun, x, comm
        )

    def test_no_clamda_warm(self): # pylint: disable=too-many-locals
        """
        Warm start but clamda not supplied.
        """
        x, bl, bu, y = self.osborne_data()
        comm = self.init('lsq_gencon_deriv')
        self.optset('Warm Start', comm)
        self.my_assert_raises_regex(
            b_utils.NagTypeError,
            '.*clamda. must not be None',
            self.fun,
            bl, bu, y, self.osborne_objfun, x, comm,
            istate=[42]*len(x),
        )

    def test_no_r_warm(self): # pylint: disable=too-many-locals
        """
        Warm start but r not supplied.
        """
        x, bl, bu, y = self.osborne_data()
        comm = self.init('lsq_gencon_deriv')
        self.optset('Warm Start', comm)
        self.my_assert_raises_regex(
            b_utils.NagTypeError,
            '.*r. must not be None',
            self.fun,
            bl, bu, y, self.osborne_objfun, x, comm,
            istate=[42]*len(x), clamda=[42.]*len(x),
        )

    def test_vardim_column(self):
        """
        Column-ordered variable-dimensioned problem.
        """

        def cb_objfun(mode, needfi, x, fjac, nstate):
            # pylint: disable=missing-docstring
            f = np.zeros(fjac.shape[0])
            sum_nag = 0.
            for i, xi in enumerate(x):
                xim1 = xi - 1.
                sum_nag += (i+1)*xim1
                f[i] = xim1
            f[n] = sum_nag
            f[n+1] = sum_nag**2

            sum_nag = 0.
            for j, xj in enumerate(x):
                sum_nag += (j+1)*(xj-1.)

            for j in range(len(x)):
                dfj = float(j+1)
                fjac[j, j] = 1.
                fjac[n+1, j] = 2.*sum_nag*dfj

            return f

        comm = self.init('lsq_gencon_deriv')
        self.optset('Derivative Level = 0', comm)
        self.optset('JTJ Initial Hessian', comm)
        self.optset('Reset Frequency = 5', comm)
        m = 12
        n = 10
        x = np.empty(n)
        for i in range(n):
            x[i] = 1. - (i+1)/float(n)
        a = np.array(
            [
                [1, 2, 0, 0, 0, 0, 0, 0, 0, 0],
                [0, 1, 2, 3, 0, 0, 0, 0, 0, 0],
                [0, 0, 1, 2, 3, 4, 0, 0, 0, 0],
                [0, 0, 0, 1, 2, 3, 4, 5, 0, 0],
                [0, 0, 0, 0, 1, 2, 3, 4, 5, 6],
            ],
            order='F',
        )
        bl = [
            0.9, 0.8, 0.7, 0.6, 0.5, 0.4, 0.3, 0.2, 0.1, 0.,
            3., -1e20, 9., -1e20, 20.,
        ]
        bu = [
            1., 1., 1., 1., 1., 1., 1., 1., 1., 1.,
            3., 6., 1e20, 15., 1e20,
        ]
        y = [0.]*m
        rtn_data = self.fun(
            bl, bu, y, cb_objfun, x, comm, a=a,
        )
        self.assertAlmostEqual(
            rtn_data.objf,
            0.,
        )
        self.assertTrue(rtn_data.fjac.flags.f_contiguous)
        self.assertTrue(rtn_data.r.flags.f_contiguous)

if __name__ == '__main__': # pragma: no cover
    unittest.main()
