#!/usr/bin/env python
"""
Integration tests for wrapper functionality in the pde submodule.

Public use of this module is not supported by NAG.
"""

# NAG Copyright 2018-2019.

# pylint: disable=invalid-name,too-many-lines

from math import cos, exp, pi, sin
import os
import unittest
import warnings

import numpy as np

from naginterfaces.base import utils as b_utils
from naginterfaces.base import pde as b_chapter
from naginterfaces.library import pde as chapter, machine
from test_functionality_core import IntegrationTest

class TestD03nc(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the d03nc wrapper.
    """

class TestD03ncNp(TestD03nc, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestD03nc.set_up(self)
        self.fun = chapter.dim1_blackscholes_fd

    def test_good(self):
        """
        Good call.
        """
        # pylint: disable=invalid-name,too-many-locals
        ntkeep = 1
        kopt = 4
        x = 50.
        mesh = 'Uniform'
        ns = 21
        s = np.empty(ns)
        s[0], s[-1] = 0., 100.
        nt = 11
        t = np.empty(nt)
        t[0], t[-1] = 0., 0.4166667
        alpha = 1.
        tdpar = [False]*3
        r = 0.1
        q = 0.
        sigma = 0.4
        soln = self.fun(
            kopt, x, mesh, s, t, tdpar, r, q, sigma, ntkeep, alpha,
        )
        np.testing.assert_allclose(
            soln.f[:, 0],
            [
                50.,
                4.50000000e+01,
                4.00000000e+01,
                3.50000000e+01,
                3.00000000e+01,
                2.50000000e+01,
                2.00000000e+01,
                1.50000000e+01,
                1.01543154e+01,
                6.58481025e+00,
                4.06718565e+00,
                2.42636757e+00,
                1.41741564e+00,
                8.19511724e-01,
                4.72406436e-01,
                2.72567970e-01,
                1.57251560e-01,
                8.96615146e-02,
                4.84494159e-02,
                2.10996111e-02,
                0.00000000e+00,
            ],
        )

class TestD03pc(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the d03pc wrapper.
    """

class TestD03pcBase(TestD03pc, unittest.TestCase):
    """
    Concrete base tests.
    """

    def setUp(self):
        TestD03pc.set_up(self)
        self.fun = b_chapter.dim1_parab_fd
        self.interp = b_chapter.dim1_parab_fd_interp

    def test_good(self):
        """
        Good call.
        """
        # pylint: disable=too-many-locals

        def cb_pdedef(npde, t, x, u, ux, p, q, r, ires):
            # pylint: disable=missing-docstring,too-many-arguments
            q[:] = [4.*(u[1] + x*ux[1]), 0.]
            r[:] = [x*ux[0], ux[1] - u[0]*u[1]]
            p[:, :] = 0.
            p[1, 1] = 1. - x**2
            return ires

        def cb_bndary(npde, t, u, ux, ibnd, beta, gamma, ires):
            # pylint: disable=missing-docstring,too-many-arguments
            if ibnd == 0:
                beta[:] = [0., 1.]
                gamma[:] = [u[0], -u[0]*u[1]]
                return ires
            beta[:] = [1., 0.]
            gamma[:] = [-u[0], u[1]]
            return ires

        npde = 2
        m = 1
        ts = 0.
        tout = 1.e-4
        npts = 20
        x = np.array(
            [sin(t) for t in np.linspace(0, pi/2, num=npts)]
        )
        u = np.empty((npde, npts))
        u[0, :] = 2.*x
        u[1, :] = 1.
        acc = 1.e-3
        comm = {}
        itask = 1
        itrace = -1
        ind = 0
        ts, _ = self.fun(
            npde, m, ts, tout, cb_pdedef, cb_bndary, u, npts,
            x, acc, comm, itask, itrace, ind,
        )
        xp = [0., 0.4, 0.6, 0.8, 0.9, 1.]
        intpts = len(xp)
        itype = 1
        up = np.empty((npde, intpts, itype))
        self.interp(npde, m, u, npts, x, xp, intpts, itype, up)
        np.testing.assert_allclose(
            up[:, :, 0],
            [
                [
                    0.00000000e+00, 8.00754225e-01, 1.19884524e+00,
                    1.59895955e+00, 1.79583378e+00, 1.84852820e+00,
                ],
                [
                    9.99711938e-01, 9.99522480e-01, 9.99372906e-01,
                    9.98836686e-01, 9.66327566e-01, -1.51469781e-16,
                ],
            ]
        )

class TestDAE(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the coupled DAE PDE routines.
    """

    # pylint: disable=too-many-instance-attributes

    npde = 1
    nv = 1
    x_len = None

    def set_up(self):
        # pylint: disable=attribute-defined-outside-init
        IntegrationTest.set_up(self)
        self.m = 0
        self.ts = 1.e-4
        self.tout = 0.2
        self.neqn = self.npde*self.npts + self.nv
        self.x = [float(i)/(self.x_len - 1.) for i in range(self.x_len)]
        self.xi = 1.
        self.itol = 1
        self.norm = 'A'
        self.laopt = 'F'
        self.algopt = np.zeros(30)
        self.comm = {}
        self.itask = 1
        self.itrace = 0
        self.ind = 0
        self.data = {'ts': self.ts}

        self.remesh = True
        self.xfix = np.empty(0)
        self.nrmesh = 3
        self.dxmesh = 0.
        self.trmesh = 42.
        self.ipminf = 0

    @staticmethod
    def cb_uvinit(npde, x, nv, data):
        # pylint: disable=missing-docstring
        u = [[exp(data['ts']*(1.-x_i)) - 1. for x_i in x]]
        v = [data['ts']]
        return u, v

    @staticmethod
    def cb_bndary(t, u, ux, v, vdot, ibnd, ires, data):
        # pylint: disable=missing-docstring,too-many-arguments
        beta = 1.
        gamma = [-v[0]*(exp(t) if ibnd == 0 else vdot[0])]
        return beta, gamma, ires

    @staticmethod
    def cb_odedef(t, v, vdot, xi, ucp, ucpx, rcp, ucpt, ucptx, ires, data):
        # pylint: disable=missing-docstring,too-many-arguments
        if ires == 1:
            f = [vdot[0] - v[0]*ucp[0, 0] - ucpx[0, 0] - 1. - t]
        elif ires == -1:
            f = [vdot[0]]
        return f, ires

class TestD03ph(TestDAE): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the d03ph wrapper.
    """

    npts = 121
    x_len = npts

class TestD03phNp(TestD03ph, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestD03ph.set_up(self)
        self.fun = chapter.dim1_parab_dae_fd
        self.u = np.empty(self.neqn)
        self.u[:self.npts] = self.cb_uvinit(
            self.npde, self.x, self.nv, self.data,
        )[0][0]
        self.u[self.neqn-1] = self.ts
        self.atol = [1e-6]
        self.rtol = [1e-6]

    @staticmethod
    def cb_pdedef(t, x, u, ux, v, vdot, ires, _data):
        # pylint: disable=missing-docstring,too-many-arguments
        p = [[v[0]*v[0]]]
        r = [ux[0]]
        q = [-x*ux[0]*v[0]*vdot[0]]
        return p, q, r, ires

    @unittest.skipIf(
        os.environ.get('NAG_PY_CVG_RUN') is not None,
        'skipping during coverage run, excessive resource requirements',
    )
    def test_good(self):
        """
        Good call.
        """
        soln = self.fun(
            self.npde, self.m, self.ts, self.tout,
            self.cb_pdedef, self.cb_bndary, self.u, self.x, self.nv,
            self.xi,
            self.rtol, self.atol, self.itol, self.norm, self.laopt,
            self.algopt, self.comm, self.itask, self.itrace,
            self.ind, odedef=self.cb_odedef, data=self.data,
        )
        self.assertAlmostEqual(
            soln.u[0],
            0.2214763642192363,
            places=5,
        )

class TestD03pj(TestDAE): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the d03pj wrapper.
    """

    nbkpts = 30
    x_len = nbkpts
    npoly = 3
    npts = (nbkpts - 1)*npoly + 1

class TestD03pjNp(TestD03pj, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestD03pj.set_up(self)
        self.fun = chapter.dim1_parab_dae_coll
        self.u = np.empty(self.neqn)
        self.atol = [1e-5]
        self.rtol = [1e-5]

    @staticmethod
    def cb_pdedef(t, x, u, ux, v, vdot, ires, _data):
        # pylint: disable=missing-docstring,too-many-arguments
        p = np.empty((u.shape[0], u.shape[0], u.shape[1]))
        r = np.empty(u.shape)
        q = np.empty(u.shape)
        for i in range(u.shape[1]):
            p[0, 0, i] = v[0]*v[0]
            r[0, i] = ux[0, i]
            q[0, i] = -x[i]*ux[0, i]*v[0]*vdot[0]
        return p, q, r, ires

    def test_good(self):
        """
        Good call.
        """
        soln = self.fun(
            self.npde, self.m, self.ts,
            self.tout, self.cb_pdedef, self.cb_bndary, self.u, self.x,
            self.npoly, self.npts, self.nv,
            self.xi, self.cb_uvinit,
            self.rtol, self.atol, self.itol, self.norm, self.laopt,
            self.algopt, self.comm, self.itask, self.itrace,
            self.ind, odedef=self.cb_odedef, data=self.data,
        )
        self.assertAlmostEqual(
            soln.u[0],
            0.2214560757464106,
            places=4,
        )

    def test_bad_ind(self):
        """
        Invalid ind supplied;
        should interact correctly with initial iemode call.
        """
        self.my_assert_raises_regex(
            b_utils.NagValueError,
            'Constraint: ind =',
            self.fun,
            self.npde, self.m, self.ts,
            self.tout, self.cb_pdedef, self.cb_bndary, self.u, self.x,
            self.npoly, self.npts, self.nv,
            self.xi, self.cb_uvinit,
            self.rtol, self.atol, self.itol, self.norm, self.laopt,
            self.algopt, self.comm, self.itask, self.itrace,
            42, odedef=self.cb_odedef, data=self.data,
        )

class TestD03pk(TestDAE): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the d03pk wrapper.
    """

    npde = 2
    npts = 21
    x_len = npts

class TestD03pkNp(TestD03pk, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestD03pk.set_up(self)
        self.fun = chapter.dim1_parab_dae_keller
        self.u = np.empty(self.neqn)
        k = 0
        for i in range(self.npts):
            self.u[k] = exp(self.ts*(1.-self.x[i])) - 1.
            self.u[k+1] = -self.ts*exp(self.ts*(1.-self.x[i]))
            k += 2
        self.u[self.neqn-1] = self.ts
        self.nleft = 1
        self.atol = [1e-4]
        self.rtol = [1e-4]
        self.algopt[0] = 1.
        self.algopt[12] = 0.005

    @staticmethod
    def cb_pdedef(t, x, u, ut, ux, v, vdot, ires, _data):
        # pylint: disable=missing-docstring,too-many-arguments
        if ires == -1:
            res = [
                v[0]*v[0]*ut[0] - x*u[1]*v[0]*vdot[0],
                0.,
            ]
        else:
            res = [
                v[0]*v[0]*ut[0] - x*u[1]*v[0]*vdot[0] - ux[1],
                u[1] - ux[0],
            ]
        return res, ires

    @staticmethod
    def cb_pk_bndary(t, ibnd, nobc, u, ut, v, vdot, ires, _data):
        # pylint: disable=missing-docstring,too-many-arguments
        if ibnd == 0:
            if ires == -1:
                res = 0.
            else:
                res = u[1] + v[0]*exp(t)
        else:
            if ires == -1:
                res = v[0]*vdot[0]
            else:
                res = u[1] + v[0]*vdot[0]
        return res, ires

    @staticmethod
    def cb_pk_odedef(t, v, vdot, xi, ucp, ucpx, ucpt, ires, _data):
        # pylint: disable=missing-docstring,too-many-arguments
        if ires == -1:
            r = [vdot[0]]
        else:
            r = [vdot[0] - v[0]*ucp[0, 0] - ucp[1, 0] - 1. - t]
        return r, ires

    def test_good(self):
        """
        Good call, correct 'max'-ing of supplied l*save if too small.
        """
        soln = self.fun(
            self.npde, self.ts,
            self.tout, self.cb_pdedef, self.cb_pk_bndary, self.u, self.x,
            self.nleft, self.nv,
            self.xi,
            self.rtol, self.atol, self.itol, self.norm, self.laopt,
            self.algopt, self.comm, self.itask, self.itrace,
            self.ind, data=self.data,
            odedef=self.cb_pk_odedef, lrsave=42, lisave=42,
        )
        self.assertAlmostEqual(
            soln.u[0],
            0.2216413539224482,
            places=5,
        )

class TestD03pl(TestDAE): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the d03pl wrapper.
    """

    npde = 2
    nv = 2
    npts = 201
    x_len = npts

class TestD03plNp(TestD03pl, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    # pylint: disable=too-many-instance-attributes

    def setUp(self):
        # pylint: disable=protected-access
        TestD03pl.set_up(self)
        self.fun = chapter.dim1_parab_convdiff_dae
        self.ts = 0.
        self.tout = 0.5
        self.u = np.empty(self.neqn)
        self.u[:self.npde*self.npts] = self.exact(
            self.ts, self.npde, self.x,
        ).flatten(order='F')
        self.u[self.neqn-2] = self.u[0] - self.u[1]
        self.u[self.neqn-1] = self.u[self.neqn-3] + self.u[self.neqn-4]
        self.xi = [0., 1.]
        self.atol = 1.e-5
        self.rtol = 0.25e-3
        self.norm = '1'
        self.laopt = 'S'
        self.algopt[0] = 1.
        self.algopt[28] = 0.1
        self.algopt[29] = 1.1
        self.itrace = -1
        self.comm = {}

        self.lisave_min_ok = 22411
        self.lisave_min_lb = 10124
        self.lrsave_min_ok = 15158
        self.lrsave_min_lb = 10638

    @staticmethod
    def exact(t, npde, x):
        # pylint: disable=missing-docstring,too-many-locals
        u = np.empty((npde, len(x)), order='F')
        f = 0.
        pi2 = 2.*pi
        for i, xi in enumerate(x):
            x1 = xi + t
            x3 = xi - 3.*t
            f = exp(pi*x3)*sin(pi2*x3)
            g = exp(-pi2*x1)*cos(pi2*x1)
            u[0, i] = f + g
            u[1, i] = f - g
        return u

    @staticmethod
    def cb_pdedef(t, x, u, ux, v, vdot, ires, data):
        # pylint: disable=missing-docstring,too-many-arguments
        p = np.zeros((len(u), len(u)))
        for i in range(len(u)):
            p[i, i] = 1.
        return p, [1.]*len(u), [0.]*len(u), [0.]*len(u), ires

    @staticmethod
    def cb_numflx(t, x, v, uleft, uright, ires, data):
        # pylint: disable=missing-docstring,too-many-arguments
        tmpl = 3.*(uleft[0]+uleft[1])
        tmpr = uright[0] - uright[1]
        flux = [
            0.5*(tmpl-tmpr),
            0.5*(tmpl+tmpr),
        ]
        return flux, ires

    @staticmethod
    def cb_pl_bndary(t, x, u, v, vdot, ibnd, ires, data):
        # pylint: disable=missing-docstring,too-many-arguments
        if ibnd == 0:
            i = 0
            ue = TestD03plNp.exact(t, u.shape[0], [x[i]])
            dudx = (u[0, i+1]-u[1, i+1]-u[0, i]+u[1, i])/(x[i+1]-x[i])
            g = [
                u[0, i] + u[1, i] - ue[0, 0] - ue[1, 0],
                vdot[0] - dudx,
            ]
        else:
            i = len(x) - 1
            ue = TestD03plNp.exact(t, u.shape[0], [x[i]])
            dudx = (u[0, i]+u[1, i]-u[0, i-1]-u[1, i-1])/(x[i]-x[i-1])
            g = [
                u[0, i] - u[1, i] - ue[0, 0] + ue[1, 0],
                vdot[1] + 3.*dudx,
            ]
        return g, ires

    @staticmethod
    def cb_pl_odedef(t, v, vdot, xi, ucp, ucpx, ucpt, ires, data):
        # pylint: disable=missing-docstring,too-many-arguments
        if ires == -1:
            r = [0., 0.]
        else:
            r = [
                v[0] - ucp[0, 0] + ucp[1, 0],
                v[1] - ucp[0, 1] - ucp[1, 1],
            ]
        return r, ires

    @unittest.skipIf(
        os.environ.get('NAG_PY_CVG_RUN') is not None,
        'skipping during coverage run, excessive resource requirements',
    )
    def test_good_multicall(self):
        """
        Good call.
        """
        args = [
            self.npde, self.ts,
            self.tout, self.cb_numflx, self.cb_pl_bndary,
            self.u, self.x,
            self.nv, self.xi,
            self.rtol, self.atol, self.itol, self.norm, self.laopt,
            self.algopt, self.comm, self.itask, self.itrace,
            self.ind,
        ]
        with warnings.catch_warnings(record=True) as w_ctxt:
            soln = self.fun(
                *args,
                pdedef=self.cb_pdedef, odedef=self.cb_pl_odedef,
                data=self.data
            )
        self.assertEqual(len(w_ctxt), 1)
        self.assertIs(w_ctxt[0].category, b_utils.NagAlgorithmicWarning)
        self.assertEqual(
            w_ctxt[0].message._errcodes, # pylint: disable=protected-access
            (15, 15, 99965),
        )
        with warnings.catch_warnings(record=True) as w_ctxt:
            soln = self.fun(
                *args,
                pdedef=self.cb_pdedef, odedef=self.cb_pl_odedef,
                data=self.data,
                lrsave=soln.lrsave_min, lisave=soln.lisave_min
            )
        self.assertEqual(len(w_ctxt), 1)
        self.assertIs(w_ctxt[0].category, b_utils.NagAlgorithmicWarning)
        self.assertEqual(
            w_ctxt[0].message._errcodes, # pylint: disable=protected-access
            (15, 15, 99965),
        )
        self.assertEqual(soln.lrsave_min, self.lrsave_min_ok)
        self.assertEqual(soln.lisave_min, self.lisave_min_ok)
        soln = self.fun(
            *args,
            pdedef=self.cb_pdedef, odedef=self.cb_pl_odedef,
            data=self.data,
            lrsave=soln.lrsave_min, lisave=soln.lisave_min
        )
        self.assertAlmostEqual(
            soln.u[0],
            -4.3253753226215615e-2,
            places=4,
        )
        self.assertGreaterEqual(soln.lrsave_min, self.lrsave_min_lb)
        self.assertGreaterEqual(soln.lisave_min, self.lisave_min_lb)
        args[1] = soln.ts
        args[2] += 0.01
        args[-1] = soln.ind
        # l*save should be ignored on continuation:
        soln = self.fun(
            *args,
            pdedef=self.cb_pdedef, odedef=self.cb_pl_odedef,
            data=self.data
        )
        self.assertAlmostEqual(
            soln.u[0],
            -0.03906478,
            places=3,
        )
        self.assertGreaterEqual(soln.lrsave_min, self.lrsave_min_lb)
        self.assertGreaterEqual(soln.lisave_min, self.lisave_min_lb)

    @unittest.skipIf(
        os.environ.get('NAG_PY_CVG_RUN') is not None,
        'skipping during coverage run, excessive resource requirements',
    )
    def test_good(self):
        """
        Good call, sufficient l*save supplied from the outset.
        """
        soln = self.fun(
            self.npde, self.ts,
            self.tout, self.cb_numflx, self.cb_pl_bndary,
            self.u, self.x,
            self.nv, self.xi,
            self.rtol, self.atol, self.itol, self.norm, self.laopt,
            self.algopt, self.comm, self.itask, self.itrace,
            self.ind,
            data=self.data,
            pdedef=self.cb_pdedef, odedef=self.cb_pl_odedef,
            lrsave=self.lrsave_min_ok, lisave=self.lisave_min_ok,
        )
        self.assertAlmostEqual(
            soln.u[0],
            -4.3253753226215615e-2,
            places=4,
        )
        self.assertGreaterEqual(soln.lrsave_min, self.lrsave_min_lb)
        self.assertGreaterEqual(soln.lisave_min, self.lisave_min_lb)

class TestD03pp(TestDAE): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the d03pp wrapper.
    """

    npde = 1
    nv = 0
    npts = 61
    x_len = npts

class TestD03ppNp(TestD03pp, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    # pylint: disable=too-many-instance-attributes

    def setUp(self):
        # pylint: disable=protected-access
        TestD03pp.set_up(self)
        self.fun = chapter.dim1_parab_remesh_fd
        self.ts = 0.
        self.u = np.empty(self.neqn)
        self.xi = np.empty(self.nv)
        self.atol = [5e-5]
        self.rtol = [5e-5]
        self.dxmesh = 0.5
        self.data = {'e': 5e-3}

    @staticmethod
    def cb_pp_pdedef(t, x, u, ux, v, vdot, ires, data):
        # pylint: disable=missing-docstring,too-many-arguments
        p = [[1.]]
        r = [data['e']*ux[0]]
        q = [u[0]*ux[0]]
        return p, q, r, ires

    @staticmethod
    def cb_pp_bndary(t, u, ux, v, vdot, ibnd, ires, data):
        # pylint: disable=missing-docstring,too-many-arguments,too-many-locals
        beta = 0.

        if ibnd == 0:
            x = 0.
            a = (x-0.25-0.75*t)/(4.*data['e'])
            b = (0.9*x-0.325-0.495*t)/(2.*data['e'])

            if a > 0. and a > b:
                a = exp(-a)
                c = (0.8*x-0.4-0.24*t)/(4.*data['e'])
                c = exp(c)
                ue = (0.5+0.1*c+a)/(1.+c+a)
            elif b > 0. and  b >= a:
                b = exp(-b)
                c = (-0.8*x+0.4+0.24*t)/(4.*data['e'])
                c = exp(c)
                ue = (0.1+0.5*c+b)/(1.+c+b)
            else:
                a = exp(a)
                b = exp(b)
                ue = (1.+0.5*a+0.1*b)/(1.+a+b)

        else:
            x = 1.
            a = (x-0.25-0.75*t)/(4.*data['e'])
            b = (0.9*x-0.325-0.495*t)/(2.*data['e'])

            if a > 0. and a > b:
                a = exp(-a)
                c = (0.8*x-0.4-0.24*t)/(4.*data['e'])
                c = exp(c)
                ue = (0.5+0.1*c+a)/(1.+c+a)
            elif b > 0. and b >= a:
                b = exp(-b)
                c = (-0.8*x+0.4+0.24*t)/(4.*data['e'])
                c = exp(c)
                ue = (0.1+0.5*c+b)/(1.+c+b)
            else:
                a = exp(a)
                b = exp(b)
                ue = (1.+0.5*a+0.1*b)/(1.+a+b)

        gamma = u[0] - ue
        return beta, gamma, ires

    @staticmethod
    def cb_pp_uvinit(npde, x, xi, nv, data):
        # pylint: disable=missing-docstring,too-many-arguments

        u = np.zeros((npde, len(x)))

        t = 0.

        for i, x_i in enumerate(x):
            a = (x_i-0.25-0.75*t)/(4.*data['e'])
            b = (0.9*x_i-0.325-0.495*t)/(2.*data['e'])

            if a > 0. and a > b:
                a = exp(-a)
                c = (0.8*x_i-0.4-0.24*t)/(4.*data['e'])
                c = exp(c)
                u[0, i] = (0.5+0.1*c+a)/(1.+c+a)
            elif b > 0. and b >= a:
                b = exp(-b)
                c = (-0.8*x_i+0.4+0.24*t)/(4.*data['e'])
                c = exp(c)
                u[0, i] = (0.1+0.5*c+b)/(1.+c+b)
            else:
                a = exp(a)
                b = exp(b)
                u[0, i] = (1.+0.5*a+0.1*b)/(1.+a+b)

        v = np.empty(nv)
        return u, v

    @staticmethod
    def cb_monitf(t, x, u, r, data):
        # pylint: disable=missing-docstring,too-many-arguments

        npts = len(x)
        fmon = np.empty(npts)

        fmon[0] = abs((r[0, 1]-r[0, 0])/((x[1]-x[0])*0.5))

        for i in range(1, npts - 1):
            h = (x[i+1]-x[i-1])*0.5
            drdx = (r[0, i+1]-r[0, i])/h
            fmon[i] = abs(drdx)

        fmon[npts-1] = fmon[npts-2]
        return fmon

    @unittest.skipIf(
        os.environ.get('NAG_PY_CVG_RUN') is not None,
        'skipping during coverage run, excessive resource requirements',
    )
    def test_good(self):
        """
        Good call.
        """
        soln = self.fun(
            self.npde, self.m, self.ts, self.tout,
            self.cb_pp_pdedef, self.cb_pp_bndary, self.cb_pp_uvinit,
            self.u, self.x, self.nv,
            self.xi, self.rtol, self.atol, self.itol,
            self.norm, self.laopt, self.algopt,
            self.remesh, self.xfix, self.nrmesh, self.dxmesh,
            self.trmesh, self.ipminf,
            self.comm, self.itask, self.itrace, self.ind,
            monitf=self.cb_monitf, data=self.data,
        )
        self.assertAlmostEqual(
            soln.u[45],
            0.2619978100413757,
        )

class TestD03pr(TestDAE): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the d03pr wrapper.
    """

    npde = 2
    nv = 0
    npts = 61
    x_len = npts

class TestD03prNp(TestD03pr, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    # pylint: disable=too-many-instance-attributes

    def setUp(self):
        # pylint: disable=protected-access
        TestD03pr.set_up(self)
        self.fun = chapter.dim1_parab_remesh_keller
        self.ts = 0.
        self.tout = 0.05
        self.u = np.empty(self.neqn)
        self.nleft = 1
        self.xi = np.empty(self.nv)
        self.atol = [5e-5]
        self.rtol = [5e-5]

    @staticmethod
    def cb_pdedef(t, x, u, ut, ux, v, vdot, ires):
        # pylint: disable=missing-docstring,too-many-arguments

        if ires == -1:
            res = [ut[0], ut[1]]
        else:
            res = [ut[0] + ux[0] + ux[1], ut[1] + 4.0*ux[0] + ux[1]]

        return res, ires

    @staticmethod
    def cb_pr_bndary(t, ibnd, nobc, u, ut, v, vdot, ires):
        # pylint: disable=missing-docstring,too-many-arguments,too-many-locals

        if ires == -1:
            res = 0.
        else:
            pp = 2.*pi
            t1 = t
            t3 = -3.0*t

            if ibnd == 0:
                ppt3 = sin(pp*t3**2)
                ppt1 = sin(pp*t1**2)
                res = u[0] - 0.5*(exp(t3)+exp(t1)+0.5*(ppt3-ppt1)) - 2.0*t**2
            else:
                t3 = t3 + 1.0
                t1 = t1 + 1.0
                ppt3 = sin(pp*t3**2)
                ppt1 = sin(pp*t1**2)
                res = (
                    u[1] - (exp(t3)-exp(t1)+0.5*(ppt3+ppt1)) -
                    (1.0+5.0*t**2-2.0*t)
                )

        return res, ires

    @staticmethod
    def cb_pr_uvinit(npde, x, xi, nv):
        # pylint: disable=missing-docstring,too-many-arguments

        u = np.zeros((npde, len(x)))

        for i, x_i in enumerate(x):
            u[0, i] = exp(x_i)
            u[1, i] = x_i**2 + sin(2.*pi*x_i**2)

        v = np.empty(nv)
        return u, v

    @staticmethod
    def cb_monitf(t, x, u):
        # pylint: disable=missing-docstring,too-many-arguments

        npts = len(x)
        fmon = np.empty(npts)

        for i in range(1, npts - 1):
            h1 = x[i] - x[i-1]
            h2 = x[i+1] - x[i]
            h3 = 0.5*(x[i+1]-x[i-1])
            d2x1 = abs(((u[0, i+1]-u[0, i])/h2-(u[0, i]-u[0, i-1])/h1)/h3)
            d2x2 = abs(((u[1, i+1]-u[1, i])/h2-(u[1, i]-u[1, i-1])/h1)/h3)
            fmon[i] = max(d2x1, d2x2)

        fmon[0] = fmon[1]
        fmon[npts - 1] = fmon[npts - 2]
        return fmon

    @unittest.skipIf(
        os.environ.get('NAG_PY_CVG_RUN') is not None,
        'skipping during coverage run, excessive resource requirements',
    )
    def test_good(self):
        """
        Good call.
        """
        soln = self.fun(
            self.npde, self.ts, self.tout,
            self.cb_pdedef, self.cb_pr_bndary, self.cb_pr_uvinit,
            self.u, self.x, self.nleft, self.nv,
            self.xi, self.rtol, self.atol, self.itol,
            self.norm, self.laopt, self.algopt, self.remesh,
            self.xfix, self.nrmesh, self.dxmesh, self.trmesh, self.ipminf,
            self.comm, self.itask, self.itrace, self.ind,
            xratio=1.2, con=5./float(self.npts-1),
            monitf=self.cb_monitf,
        )
        self.assertAlmostEqual(
            soln.u[45],
            0.7347143232827940,
        )

class TestD03ps(TestDAE): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the d03ps wrapper.
    """

    npde = 1
    nv = 0
    npts = 61
    x_len = npts

class TestD03psNp(TestD03ps, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    # pylint: disable=too-many-instance-attributes

    def setUp(self):
        # pylint: disable=protected-access
        TestD03ps.set_up(self)
        self.fun = chapter.dim1_parab_convdiff_remesh
        self.ts = 0.
        self.tout = 0.1
        self.u = np.empty(self.neqn)
        self.xi = np.empty(self.nv)
        self.atol = [1e-4]
        self.rtol = [1e-4]
        self.norm = '1'
        self.laopt = 'B'
        self.algopt[0] = 1.
        self.algopt[12] = 0.005

    @staticmethod
    def cb_pdedef(t, x, u, ux, v, vdot, ires):
        # pylint: disable=missing-docstring,too-many-arguments
        p = [[1.]]
        c = 0.002
        d = ux[0]
        s = 0.
        return p, c, d, s, ires

    @staticmethod
    def cb_numflx(t, x, v, uleft, uright, ires):
        # pylint: disable=missing-docstring,too-many-arguments
        flux = [uleft[0]]
        return flux, ires

    @staticmethod
    def cb_ps_bndary(t, x, u, v, vdot, ibnd, ires):
        # pylint: disable=missing-docstring,too-many-arguments
        if ibnd == 0:
            g = [u[0, 0]]
        else:
            g = [u[0, -2]]
        return g, ires

    @staticmethod
    def cb_ps_uvinit(npde, x, xi, nv):
        # pylint: disable=missing-docstring,too-many-arguments
        u = np.zeros((npde, len(x)))
        for i, x_i in enumerate(x):
            if 0.2 < x_i <= 0.4:
                u[0, i] = sin(pi*(5.*x_i-1.))
        v = np.empty(nv)
        return u, v

    @staticmethod
    def cb_monitf(t, x, u):
        # pylint: disable=missing-docstring,too-many-arguments
        npts = len(x)
        fmon = np.empty(npts)
        for i in range(1, npts - 1):
            h1 = x[i] - x[i-1]
            h2 = x[i+1] - x[i]
            h3 = 0.5*(x[i+1]-x[i-1])
            fmon[i] = abs(((u[0, i+1]-u[0, i])/h2-(u[0, i]-u[0, i-1])/h1)/h3)
        fmon[0] = fmon[1]
        fmon[npts-1] = fmon[npts-2]
        return fmon

    @unittest.skipIf(
        os.environ.get('NAG_PY_CVG_RUN') is not None,
        'skipping during coverage run, excessive resource requirements',
    )
    def test_good(self):
        """
        Good call.
        """
        soln = self.fun(
            self.npde, self.ts,
            self.tout,
            self.cb_numflx, self.cb_ps_bndary,
            self.cb_ps_uvinit,
            self.u, self.x,
            self.nv,
            self.xi,
            self.rtol, self.atol, self.itol, self.norm, self.laopt,
            self.algopt,
            self.remesh, self.xfix, self.nrmesh, self.dxmesh, self.trmesh,
            self.ipminf,
            self.comm, self.itask, self.itrace, self.ind,
            pdedef=self.cb_pdedef, monitf=self.cb_monitf,
        )
        self.assertAlmostEqual(
            soln.u[45],
            0.1620556461289043,
        )

class TestD03ra(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the d03ra wrapper.
    """

class TestD03raNp(TestD03ra, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        # pylint: disable=protected-access
        TestD03ra.set_up(self)
        self.fun = chapter.dim2_gen_order2_rectangle

    def test_good_sol(self):
        """
        Good call, sol length correct in monitr.
        """
        # pylint: disable=too-many-locals,too-many-statements

        npde = 2

        ts = 0.
        twant = [0.01, 0.025]
        dt = [0.5E-3, 1.0E-7, 0.]
        xmin = 0.
        xmax = 1.
        ymin = 0.
        ymax = 1.
        nx = 11
        ny = 11
        tols = 0.1
        tolt = 0.1

        def cb_pdedef(t, x, y, u, ut, ux, uy, uxx, uxy, uyy):
            # pylint: disable=too-many-arguments
            alpha = 50.
            beta = 300.
            res = np.empty(u.shape)
            fourpi = 4.*pi
            for i, xi in enumerate(x):
                b1 = (
                    1. + alpha*xi*y[i] +
                    beta*sin(fourpi*xi)*sin(fourpi*y[i])
                )
                res[i, 0] = (
                    ut[i, 0] - (uxx[i, 0]+uyy[i, 0]) -
                    u[i, 0]*(b1-u[i, 0]-0.5E-6*u[i, 1])
                )
                res[i, 1] = (
                    -0.05*(uxx[i, 1]+uyy[i, 1]) -
                    u[i, 1]*(-b1+1.0E4*u[i, 0]-u[i, 1])
                )
            return res

        def cb_bndary(t, x, y, u, ut, ux, uy, lbnd, res):
            # pylint: disable=too-many-arguments
            tol = 10.*machine.precision()
            new_res = np.empty(u.shape)
            new_res[:, :] = res
            for jp in lbnd:
                j = jp - 1
                if abs(x[j]) <= tol or abs(x[j]-1.) <= tol:
                    new_res[j, :] = ux[j, :]
                elif abs(y[j]) <= tol or abs(y[j]-1.) <= tol:
                    new_res[j, :] = uy[j, :]
            return new_res

        def cb_pdeiv(npde, t, x, y):
            alpha = 50.
            beta = 300.
            u = np.empty((len(x), npde))
            fourpi = 4.*pi
            u[:, 0] = 10. + (16.*x*(1.-x)*y*(1.-y))**2
            u[:, 1] = (
                -1. - alpha*x*y -
                beta*np.sin(fourpi*x)*np.sin(fourpi*y) + 1.0E4*u[:, 0]
            )
            return u

        def cb_monitr(
                npde, t, dt, dtnew, tlast, ngpts, xpts, ypts, lsol, sol, ierr,
        ):
            # pylint: disable=too-many-arguments
            self.assertEqual(len(sol), lsol[0]+npde*ngpts[0])
            return ierr

        opti = [4, 0, 0, 0]
        optr = np.ones((3, npde))
        optr[0, :] = [250., 1.5e6]
        comm = {}
        itrace = 0
        ind = 0

        for tout in twant:
            soln = self.fun(
                ts, tout, dt, xmin, xmax, ymin, ymax, nx, ny, tols, tolt,
                cb_pdedef, cb_bndary, cb_pdeiv, cb_monitr,
                opti, optr, comm, itrace, ind,
            )
            ts = soln.ts
            dt = soln.dt
            ind = soln.ind

        np.testing.assert_equal(comm['iwk'][0:2], [29, 0])

if __name__ == '__main__': # pragma: no cover
    unittest.main()
