#!/usr/bin/env python
"""
Integration tests for wrapper functionality in the rand submodule.

Public use of this module is not supported by NAG.
"""

# NAG Copyright 2018-2019.

# pylint: disable=invalid-name

import unittest

import numpy as np

from naginterfaces.base import utils as b_utils
from naginterfaces.base import rand as b_chapter
from naginterfaces.library import rand as chapter
from test_functionality_core import IntegrationTest

class TestG05kf(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the g05kf wrapper.
    """

class TestG05kfBase(TestG05kf, unittest.TestCase):
    """
    Concrete base tests.
    """

    def setUp(self):
        TestG05kf.set_up(self)
        self.fun = b_chapter.init_repeat

    def test_good(self):
        """
        Good call.
        """
        statecomm = self.fun(genid=3, subid=0, seed=[42], lseed=1)
        self.assertEqual(len(statecomm['state']), 1260)

class TestG05kfNp(TestG05kf, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestG05kf.set_up(self)
        self.fun = chapter.init_repeat

    def test_good(self):
        """
        Good call.
        """
        statecomm = self.fun(genid=3, seed=[42])
        self.assertEqual(statecomm['state'][0], 1260)

class TestG05kg(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the g05kg wrapper.
    """

class TestG05kgNp(TestG05kg, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestG05kg.set_up(self)
        self.fun = chapter.init_nonrepeat

    def test_good(self):
        """
        Good call.
        """
        statecomm = self.fun(genid=3)
        self.assertEqual(statecomm['state'][0], 1260)

class TestG05nd(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the g05nd wrapper.
    """

class TestG05ndNp(TestG05nd, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestG05nd.set_up(self)
        self.fun = chapter.sample
        self.statecomm = chapter.init_repeat(genid=1, seed=[1762543])
        self.n = 8

    def test_good_ipop(self):
        """
        Good call, ipop supplied.
        """
        for m in range(1, 4+1):
            isampl = self.fun(m, self.statecomm, ipop=list(range(1, self.n+1)))
        np.testing.assert_equal(
            isampl,
            [2, 6, 7, 8],
        )

    def test_good_noipop(self):
        """
        Good call, ipop inferred.
        """
        for m in range(1, 4+1):
            isampl = self.fun(m, self.statecomm, n=self.n)
        np.testing.assert_equal(
            isampl,
            [2, 6, 7, 8],
        )

    def test_bad(self):
        """
        Bad call, ipop and n omitted.
        """
        self.my_assert_raises_regex(
            b_utils.NagTypeError,
            'ipop.*must not be None when n is None',
            self.fun,
            1, self.statecomm,
        )

class TestG05py(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the g05py wrapper.
    """

class TestG05pyNp(TestG05py, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestG05py.set_up(self)
        self.init = chapter.init_repeat
        self.fun = chapter.matrix_corr

    def test_good(self):
        """
        Good call.
        """
        statecomm = self.init(genid=1, seed=[1762543])
        d = [0.7, 0.9, 1.4]
        np.testing.assert_allclose(
            self.fun(d, statecomm),
            [
                [1., -0.2549414, -0.1004264],
                [-0.2549414, 1., 0.23434893],
                [-0.1004264, 0.23434893, 1.],
            ],
            )

class TestPRNG(IntegrationTest, unittest.TestCase):
    """
    Abstract suite of test cases for the pseudoRNGs.
    """

    def set_up(self):
        # pylint: disable=attribute-defined-outside-init

        self.fun = lambda *args: None # pragma: no cover
        self.args = [None]

        self.statecomm = chapter.init_repeat(
            genid=1, subid=1, seed=[1762543],
        )

        self.mode_i = 0
        self.comm_i = -3
        self.x_i = -1
        self.result = None

    def good_setup_gen(self):
        """
        Good call.
        """
        self.fun(*self.args)
        np.testing.assert_allclose(
            self.args[self.x_i],
            self.result,
            rtol=1e-3,
        )

    def good_setup_then_gen(self):
        """
        Good call.
        """
        self.args[self.mode_i] = 0
        self.fun(*self.args)
        self.args[self.mode_i] = 1
        self.fun(*self.args)
        np.testing.assert_allclose(
            self.args[self.x_i],
            self.result,
            rtol=1e-3,
        )

    def good_setup_no_ref(self):
        """
        Good call.
        """
        self.args[self.mode_i] = 3
        self.args[self.comm_i] = None
        self.fun(*self.args)
        np.testing.assert_allclose(
            self.args[self.x_i],
            self.result,
            rtol=1e-3,
        )
        self.assertIsNone(self.args[self.comm_i])

class TestG05ph(TestPRNG): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the g05ph wrapper.
    """

class TestG05phBase(TestG05ph, unittest.TestCase):
    """
    Concrete base tests.
    """

    def setUp(self):
        TestG05ph.set_up(self)
        self.fun = b_chapter.times_arma

        n = 10
        phi = [0.4, 0.2]
        comm = {}
        self.args = [
            2, # mode, 0
            n, # n, 1
            0., # xmean, 2
            len(phi), # ip, 3
            phi, # phi, 4
            0, # iq, 5
            [], # theta, 6
            1., # avar, 7
            comm, # comm, 8
            self.statecomm, # statecomm, 9
            np.empty(n), # x, 10
        ]

        self.result = [
            -1.7103, -0.4042, -0.1845, -1.5004, -1.1946, -1.8184, -1.0895,
            1.6408, 1.3555, 1.1908,
        ]

    test_good_setup_gen = TestPRNG.good_setup_gen
    test_good_setup_then_gen = TestPRNG.good_setup_then_gen

class TestG05pj(TestPRNG): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the g05pj wrapper.
    """

class TestG05pjBase(TestG05pj, unittest.TestCase):
    """
    Concrete base tests.
    """

    def setUp(self):
        TestG05pj.set_up(self)
        self.fun = b_chapter.times_mv_varma

        n = 48
        xmean = [5., 9.]
        k = len(xmean)
        phi = [0.8, 0., 0.07, 0.58]
        var = np.array([
            [2.97, 0.],
            [0.64, 5.38],
        ])
        comm = {}
        self.args = [
            2, # mode, 0
            n, # n, 1
            k, # k, 2
            xmean, # xmean, 3
            1, # ip, 4
            phi, # phi, 5
            0, # iq, 6
            [], # theta, 7
            var, # var, 8
            comm, # comm, 9
            self.statecomm, # statecomm, 10
            np.empty((k, n)), # x, 11
        ]

        self.result = [
            [
                4.833, 2.813, 3.224, 3.825, 1.023, 1.415, 2.184, 3.005,
                5.547, 4.832, 4.705, 5.484, 9.407, 10.335, 8.495, 7.478,
                6.373, 6.692, 6.698, 6.976, 6.200, 4.458, 2.520, 3.517,
                3.054, 5.439, 5.699, 7.136, 5.750, 8.497, 9.563, 11.604,
                9.020, 10.063, 7.976, 5.927, 4.992, 4.222, 3.982, 7.107,
                3.554, 7.045, 7.025, 4.106, 5.106, 5.954, 8.026, 7.212,
            ],
            [
                8.458, 9.140, 10.866, 10.975, 9.245, 5.054, 5.023, 12.486,
                10.534, 10.590, 11.376, 8.793, 14.445, 13.237, 11.030, 8.405,
                7.187, 8.291, 5.920, 9.390, 10.055, 6.222, 7.751, 10.604,
                12.441, 10.664, 10.960, 8.022, 10.073, 12.870, 12.665, 14.064,
                11.867, 12.894, 10.546, 12.754, 8.594, 9.042, 12.029, 12.557,
                9.746, 5.487, 5.500, 8.629, 9.723, 8.632, 6.383, 12.484
            ]
        ]

    test_good_setup_gen = TestPRNG.good_setup_gen
    test_good_setup_then_gen = TestPRNG.good_setup_then_gen

    def test_good_new(self):
        """
        Good call.
        """
        self.fun(*self.args)
        self.args[self.mode_i] = 3
        self.fun(*self.args)

        self.result = [
            [
                5.396, 4.811, 2.685, 5.824, 2.449, 3.563, 5.663, 6.209,
                3.130, 4.308, 4.333, 4.903, 1.770, 1.278, 1.340, -0.527,
                1.745, 3.211, 4.478, 5.170, 5.365, 4.852, 6.080, 6.464,
                2.765, 2.148, 6.641, 7.224, 10.316, 7.102, 5.604, 3.934,
                4.839, 3.698, 5.210, 5.384, 7.652, 7.315, 7.332, 7.561,
                7.537, 7.788, 6.868, 7.575, 6.108, 6.188, 8.132, 10.310,
            ],
            [
                11.345, 10.070, 13.654, 12.409, 11.329, 13.054, 12.465, 9.867,
                10.263, 13.394, 10.553, 10.331, 7.814, 8.747, 10.025, 11.167,
                10.626, 9.366, 9.607, 9.662, 10.492, 10.766, 11.512, 10.813,
                10.799, 8.780, 9.221, 14.245, 11.575, 10.620, 8.282, 5.447,
                9.935, 9.386, 11.627, 10.066, 11.394, 7.951, 7.907, 12.616,
                15.246, 9.962, 13.216, 11.350, 11.227, 6.021, 6.968, 12.428,
            ]
        ]

        np.testing.assert_allclose(
            self.args[self.x_i],
            self.result,
            rtol=1e-3,
        )

class TestG05pz(TestPRNG): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the g05pz wrapper.
    """

class TestG05pzBase(TestG05pz, unittest.TestCase):
    """
    Concrete base tests.
    """

    def setUp(self):
        TestG05pz.set_up(self)
        self.fun = b_chapter.matrix_2waytable

        totr = [9, 11, 7, 23]
        nrow = len(totr)
        totc = [16, 17, 17]
        ncol = len(totc)
        comm = {}
        self.args = [
            2, # mode, 0
            nrow, # nrow, 1
            ncol, # ncol, 2
            totr, # totr, 3
            totc, # totc, 4
            comm, # comm, 5
            self.statecomm, # statecomm, 6
            np.empty((nrow, ncol), dtype=int), # x, 7
        ]

        self.result = [
            [2, 4, 3],
            [6, 1, 4],
            [2, 4, 1],
            [6, 8, 9],
        ]

    test_good_setup_gen = TestPRNG.good_setup_gen
    test_good_setup_then_gen = TestPRNG.good_setup_then_gen

class TestG05rc(TestPRNG): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the g05rc wrapper.
    """

class TestG05rcBase(TestG05rc, unittest.TestCase):
    """
    Concrete base tests.
    """

    def setUp(self):
        TestG05rc.set_up(self)
        self.fun = b_chapter.copula_students_t

        n = 10
        c = np.array([
            [1.69, 0.39, -1.86, 0.07],
            [0., 98.01, -7.07, -0.71],
            [0., 0., 11.56, 0.03],
            [0., 0., 0., 0.01],
        ])
        comm = {}
        self.args = [
            2, # mode, 0
            n, # n, 1
            10, # df, 2
            c.shape[0], # m, 3
            c, # c, 4
            comm, # comm, 5
            self.statecomm, # statecomm, 6
            np.empty((n, c.shape[0])), # x, 7
        ]

        self.result = np.array([
            [0.6445, 0.0527, 0.4082, 0.8876],
            [0.0701, 0.1988, 0.8471, 0.3521],
            [0.7988, 0.6664, 0.2194, 0.5541],
            [0.8202, 0.0492, 0.7059, 0.9341],
            [0.1786, 0.5594, 0.7810, 0.2836],
            [0.4920, 0.2677, 0.3427, 0.5169],
            [0.4139, 0.2978, 0.8762, 0.7145],
            [0.7437, 0.9714, 0.8931, 0.2487],
            [0.4971, 0.9687, 0.8142, 0.1965],
            [0.6464, 0.5304, 0.5817, 0.4565],
        ])

    test_good_setup_gen = TestPRNG.good_setup_gen
    test_good_setup_then_gen = TestPRNG.good_setup_then_gen

class TestG05rd(TestPRNG): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the g05rd wrapper.
    """

class TestG05rdBase(TestG05rd, unittest.TestCase):
    """
    Concrete base tests.
    """

    def setUp(self):
        TestG05rd.set_up(self)
        self.fun = b_chapter.copula_normal

        n = 10
        c = np.array([
            [1.69, 0.39, -1.86, 0.07],
            [0., 98.01, -7.07, -0.71],
            [0., 0., 11.56, 0.03],
            [0., 0., 0., 0.01],
        ])
        comm = {}
        self.args = [
            2, # mode, 0
            n, # n, 1
            c.shape[0], # m, 2
            c, # c, 3
            comm, # comm, 4
            self.statecomm, # statecomm, 5
            np.empty((n, c.shape[0])), # x, 6
        ]

        self.result = np.array([
            [0.6364, 0.0517, 0.4137, 0.8817],
            [0.1065, 0.2461, 0.7993, 0.3806],
            [0.7460, 0.6313, 0.2708, 0.5421],
            [0.7983, 0.0564, 0.6868, 0.9234],
            [0.1046, 0.5790, 0.8533, 0.2208],
            [0.4925, 0.2784, 0.3513, 0.5158],
            [0.3843, 0.2349, 0.9472, 0.7801],
            [0.7871, 0.9941, 0.9403, 0.2044],
            [0.4982, 0.9015, 0.7176, 0.2914],
            [0.6717, 0.5359, 0.5961, 0.4487],
        ])

    test_good_setup_gen = TestPRNG.good_setup_gen

    def test_good_setup_then_gen(self):
        """
        Good call.
        """
        self.args[self.mode_i] = 0
        self.args[self.x_i] = np.empty((self.args[1], 0))
        self.fun(*self.args)
        self.args[self.mode_i] = 1
        self.args[self.x_i] = np.empty((self.args[1], self.args[2]))
        self.fun(*self.args)
        np.testing.assert_allclose(
            self.args[self.x_i],
            self.result,
            rtol=1e-3,
        )

class TestG05ta(TestPRNG): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the g05ta wrapper.
    """

class TestG05taBase(TestG05ta):
    """
    Concrete base tests.
    """

    def setUp(self):
        TestG05ta.set_up(self)
        self.fun = b_chapter.int_binomial

        n = 20
        comm = {}
        self.args = [
            2, # mode, 0
            n, # n, 1
            6000, # m, 2
            0.8, # p, 3
            comm, # comm, 4
            self.statecomm, # statecomm, 5
            np.empty(n, dtype=int), # x, 6
        ]

        self.result = [
            4811, 4761, 4821, 4826, 4761, 4800, 4791, 4825, 4800, 4814, 4749,
            4780, 4810, 4750, 4807, 4782, 4778, 4877, 4840, 4802,
        ]

    test_good_setup_gen = TestPRNG.good_setup_gen
    test_good_setup_then_gen = TestPRNG.good_setup_then_gen
    test_good_setup_no_ref = TestPRNG.good_setup_no_ref

class TestG05tc(TestPRNG): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the g05tc wrapper.
    """

class TestG05tcBase(TestG05tc):
    """
    Concrete base tests.
    """

    def setUp(self):
        TestG05tc.set_up(self)
        self.fun = b_chapter.int_geom

        n = 10
        comm = {}
        self.args = [
            2, # mode, 0
            n, # n, 1
            0.001, # p, 2
            comm, # comm, 3
            self.statecomm, # statecomm, 4
            np.empty(n, dtype=int), # x, 5
        ]

        self.result = [
            1011, 112, 1369, 1600, 110, 677, 484, 1546, 689, 1113,
        ]

    test_good_setup_gen = TestPRNG.good_setup_gen
    test_good_setup_then_gen = TestPRNG.good_setup_then_gen

    def test_good_setup_no_ref(self):
        """
        Good call.
        """
        self.result = [
            451, 2238, 292, 225, 2256, 708, 955, 239, 696, 397,
        ]
        TestPRNG.good_setup_no_ref(self)

class TestG05td(TestPRNG): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the g05td wrapper.
    """

class TestG05tdBase(TestG05td):
    """
    Concrete base tests.
    """

    def setUp(self):
        TestG05td.set_up(self)
        self.fun = b_chapter.int_general

        p = np.array([
            0.01, 0.02, 0.04, 0.08, 0.2, 0.3, 0.2, 0.08, 0.04, 0.02, 0.01,
        ])
        n = 20
        comm = {}
        self.args = [
            2, # mode, 0
            n, # n, 1
            p, # p, 2
            len(p), # np, 3
            -5, # ip, 4
            1, # itype, 5
            comm, # comm, 6
            self.statecomm, # statecomm, 7
            np.empty(n, dtype=int), # x, 8
        ]

        self.result = [
            0, -2, 1, 1, -2, 0, 0, 1, 0, 1,
            -3, -1, 0, -3, 0, -1, -1, 5, 2, 0,
        ]

    test_good_setup_gen = TestPRNG.good_setup_gen
    test_good_setup_then_gen = TestPRNG.good_setup_then_gen
    test_good_setup_no_ref = TestPRNG.good_setup_no_ref

class TestG05te(TestPRNG): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the g05te wrapper.
    """

class TestG05teBase(TestG05te):
    """
    Concrete base tests.
    """

    def setUp(self):
        TestG05te.set_up(self)
        self.fun = b_chapter.int_hypergeom

        n = 20
        comm = {}
        self.args = [
            2, # mode, 0
            n, # n, 1
            500, # ns, 2
            1000, # np, 3
            900, # m, 4
            comm, # comm, 5
            self.statecomm, # statecomm, 6
            np.empty(n, dtype=int), # x, 7
        ]

        self.result = [
            452, 444, 453, 454, 444, 450, 449, 454, 450, 452, 442, 447, 451,
            442, 451, 447, 447, 462, 456, 450,
        ]

    test_good_setup_gen = TestPRNG.good_setup_gen
    test_good_setup_then_gen = TestPRNG.good_setup_then_gen
    test_good_setup_no_ref = TestPRNG.good_setup_no_ref

class TestG05tf(TestPRNG): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the g05tf wrapper.
    """

class TestG05tfBase(TestG05tf):
    """
    Concrete base tests.
    """

    def setUp(self):
        TestG05tf.set_up(self)
        self.fun = b_chapter.int_log

        n = 10
        comm = {}
        self.args = [
            2, # mode, 0
            n, # n, 1
            0.9999, # a, 2
            comm, # comm, 3
            self.statecomm, # statecomm, 4
            np.empty(n, dtype=int), # x, 5
        ]

        self.result = [
            201, 1, 573, 962, 1, 53, 19, 859, 56, 281,
        ]

    test_good_setup_gen = TestPRNG.good_setup_gen
    test_good_setup_then_gen = TestPRNG.good_setup_then_gen

    def test_good_setup_no_ref(self):
        """
        Good call.
        """
        self.result = [
            6, 23, 2765, 30, 3, 1, 299, 968, 166, 4,
        ]
        TestPRNG.good_setup_no_ref(self)

class TestG05tg(TestPRNG): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the g05tg wrapper.
    """

class TestG05tgBase(TestG05tg):
    """
    Concrete base tests.
    """

    def setUp(self):
        TestG05tg.set_up(self)
        self.fun = b_chapter.int_multinomial

        p = [0.08, 0.1, 0.8, 0.02]
        n = 20
        k = len(p)
        comm = {}
        self.args = [
            2, # mode, 0
            n, # n, 1
            6000, # m, 2
            k, # k, 3
            p, # p, 4
            comm, # comm, 5
            self.statecomm, # statecomm, 6
            np.empty((n, k), dtype=int), # x, 7
        ]

        self.result = [
            [468, 603, 4811, 118],
            [490, 630, 4761, 119],
            [482, 575, 4821, 122],
            [495, 591, 4826, 88],
            [512, 611, 4761, 116],
            [474, 601, 4800, 125],
            [485, 595, 4791, 129],
            [468, 582, 4825, 125],
            [485, 598, 4800, 117],
            [485, 573, 4814, 128],
            [501, 634, 4749, 116],
            [482, 618, 4780, 120],
            [470, 584, 4810, 136],
            [479, 642, 4750, 129],
            [476, 608, 4807, 109],
            [473, 631, 4782, 114],
            [509, 596, 4778, 117],
            [450, 565, 4877, 108],
            [484, 556, 4840, 120],
            [466, 615, 4802, 117],
        ]

    test_good_setup_gen = TestPRNG.good_setup_gen
    test_good_setup_then_gen = TestPRNG.good_setup_then_gen
    test_good_setup_no_ref = TestPRNG.good_setup_no_ref

class TestG05th(TestPRNG): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the g05th wrapper.
    """

class TestG05thBase(TestG05th):
    """
    Concrete base tests.
    """

    def setUp(self):
        TestG05th.set_up(self)
        self.fun = b_chapter.int_negbin

        n = 20
        comm = {}
        self.args = [
            2, # mode, 0
            n, # n, 1
            60, # m, 2
            0.999, # p, 3
            comm, # comm, 4
            self.statecomm, # statecomm, 5
            np.empty(n, dtype=int), # x, 6
        ]

        self.result = [
            62339, 50505, 64863, 66289, 50434, 59461, 57365, 65965, 59572,
            63104, 47833, 54735, 62075, 48018, 61458, 55190, 54263, 80995,
            70129, 60200,
        ]

    test_good_setup_gen = TestPRNG.good_setup_gen
    test_good_setup_then_gen = TestPRNG.good_setup_then_gen
    test_good_setup_no_ref = TestPRNG.good_setup_no_ref

class TestG05tj(TestPRNG): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the g05tj wrapper.
    """

class TestG05tjBase(TestG05tj):
    """
    Concrete base tests.
    """

    def setUp(self):
        TestG05tj.set_up(self)
        self.fun = b_chapter.int_poisson

        n = 10
        comm = {}
        self.args = [
            2, # mode, 0
            n, # n, 1
            20., # lamda, 2
            comm, # comm, 3
            self.statecomm, # statecomm, 4
            np.empty(n, dtype=int), # x, 5
        ]

        self.result = [
            21, 15, 23, 24, 14, 20, 19, 23, 20, 22,
        ]

    test_good_setup_gen = TestPRNG.good_setup_gen
    test_good_setup_then_gen = TestPRNG.good_setup_then_gen
    test_good_setup_no_ref = TestPRNG.good_setup_no_ref

if __name__ == '__main__': # pragma: no cover
    unittest.main()
