#!/usr/bin/env python
"""
Integration tests for wrapper functionality in the sort submodule.

Public use of this module is not supported by NAG.
"""

# NAG Copyright 2018-2019.

# pylint: disable=invalid-name

import unittest

import numpy as np

from naginterfaces.base import utils as b_utils
from naginterfaces.base import sort as b_chapter
from naginterfaces.library import sort as chapter
from test_functionality_core import IntegrationTest

class TestM01na(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the m01na wrapper.
    """

class TestM01naBase(TestM01na, unittest.TestCase):
    """
    Concrete base tests.
    """

    def setUp(self):
        TestM01na.set_up(self)
        self.fun = b_chapter.realvec_search

    def test_bad_m1(self):
        """
        Bad m1 supplied.
        """
        self.my_assert_raises_regex(
            b_utils.NagValueError,
            '.*Constraint: m1 >= 1.',
            self.fun,
            True, [42.], 0, 1, 43.,
        )

class TestM01naNp(TestM01na, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestM01na.set_up(self)
        self.fun = chapter.realvec_search

    def test_good(self):
        """
        Good call.
        """
        self.assertEqual(
            self.fun(True, [0.5, 0.6], 0.4),
            0,
        )

class TestM01nd(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the m01nd wrapper.
    """

class TestM01ndNp(TestM01nd, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestM01nd.set_up(self)
        self.fun = chapter.realvec_vec_search
        self.valid = True
        self.m1 = 1
        self.m2 = 16
        self.rv = [
            0.5, 0.6, 1.1, 1.2, 1.3, 1.4, 2.1, 2.3,
            2.4, 4.1, 5.8, 5.9, 6.5, 6.6, 8.6, 9.9,
        ]
        self.item = [2.1, 0.4, 7.1, 10.0]
        self.idx = [7, 0, 14, 16]

    def test_good_binsearch(self):
        """
        Good call.
        """
        np.testing.assert_equal(
            self.fun(self.valid, self.rv, self.m1, self.m2, self.item),
            self.idx,
        )

    def test_good(self):
        """
        Good call.
        """
        comm = {}
        np.testing.assert_equal(
            self.fun(self.valid, self.rv, self.m1, self.m2, self.item, comm),
            self.idx,
        )

    def test_good_reuse(self):
        """
        Good call.
        """
        comm = {}
        self.fun(self.valid, self.rv, self.m1, self.m2, self.item, comm)
        np.testing.assert_equal(
            self.fun(self.valid, self.rv, self.m1, self.m2, self.item, comm),
            self.idx,
        )

if __name__ == '__main__': # pragma: no cover
    unittest.main()
