#!/usr/bin/env python
"""
Integration tests for wrapper functionality in the sparseig submodule.

Public use of this module is not supported by NAG.
"""

# NAG Copyright 2018-2019.

# pylint: disable=invalid-name

import unittest

import numpy as np

from naginterfaces.library import sparseig as chapter
from test_functionality_core import IntegrationTest

class TestF12aa(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the f12aa wrapper.
    """

class TestF12aaNp(TestF12aa, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestF12aa.set_up(self)
        self.fun = chapter.real_init

    def test_good(self):
        """
        Good call.
        """
        comm = self.fun(100, 10, 30)
        self.assertEqual(
            len(comm['icomm']), 140,
        )
        self.assertEqual(
            len(comm['comm']), 3240,
        )

class TestF12af(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the f12af wrapper.
    """

class TestF12afNp(TestF12af, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestF12af.set_up(self)
        self.fun = chapter.real_band_init

    def test_good(self):
        """
        Good call.
        """
        comm = self.fun(100, 4, 10)
        self.assertEqual(
            len(comm['icomm']), 140,
        )
        self.assertEqual(
            len(comm['comm']), 60,
        )

class TestF12an(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the f12an wrapper.
    """

class TestF12anNp(TestF12an, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestF12an.set_up(self)
        self.fun = chapter.complex_init

    def test_good(self):
        """
        Good call.
        """
        comm = self.fun(100, 4, 20)
        self.assertEqual(
            len(comm['icomm']), 140,
        )
        self.assertEqual(
            len(comm['comm']), 1660,
        )

class TestF12at(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the f12at wrapper.
    """

class TestF12atNp(TestF12at, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestF12at.set_up(self)
        self.fun = chapter.complex_band_init

    def test_good(self):
        """
        Good call.
        """
        comm = self.fun(100, 4, 10)
        self.assertEqual(
            len(comm['icomm']), 140,
        )
        self.assertEqual(
            len(comm['comm']), 60,
        )

class TestF12fa(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the f12fa wrapper.
    """

class TestF12faNp(TestF12fa, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestF12fa.set_up(self)
        self.fun = chapter.real_symm_init

    def test_good(self):
        """
        Good call.
        """
        comm = self.fun(100, 4, 10)
        self.assertEqual(
            len(comm['icomm']), 140,
        )
        self.assertEqual(
            len(comm['comm']), 540,
        )

class TestF12ff(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the f12ff wrapper.
    """

class TestF12ffNp(TestF12ff, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestF12ff.set_up(self)
        self.fun = chapter.real_symm_band_init

    def test_good(self):
        """
        Good call.
        """
        comm = self.fun(100, 4, 10)
        self.assertEqual(
            len(comm['icomm']), 140,
        )
        self.assertEqual(
            len(comm['comm']), 60,
        )

class TestF12fg(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the f12fg wrapper.
    """

class TestF12fgNp(TestF12fg, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestF12fg.set_up(self)
        self.init = chapter.real_symm_band_init
        self.fun = chapter.real_symm_band_solve

    def test_good(self):
        """
        Good call.
        """
        # pylint: disable=too-many-locals

        nx = 10
        n = nx**2
        comm = self.init(n, 4, 10)

        kl = nx
        ku = kl
        ab = np.zeros((2*kl + ku + 1, n))

        h2 = 1./(nx+1)**2
        ab[kl + ku, :] = 4./h2

        isup = kl + ku
        isub = kl + ku + 2

        for i in range(1, nx + 1):
            lo = (i-1)*nx

            for j in range(lo + 1, lo + nx):
                ab[isup-1, j] = -1./h2
                ab[isub-1, j-1] = -1./h2

        isup = kl + 1
        isub = 2*kl + ku + 1

        for i in range(1, nx):
            lo = (i-1)*nx

            for j in range(lo + 1, lo + nx + 1):
                ab[isup-1, nx+j-1] = -1./h2
                ab[isub-1, j-1] = -1./h2

        mb = np.empty(ab.shape)
        sigma = -42.
        resid = np.empty(n)

        soln = self.fun(kl, ku, ab, mb, sigma, resid, comm)
        self.assertEqual(soln.nconv, 4)

if __name__ == '__main__': # pragma: no cover
    unittest.main()
