#!/usr/bin/env python
"""
Integration tests for wrapper functionality in the specfun submodule.

Public use of this module is not supported by NAG.
"""

# NAG Copyright 2018-2019.

# pylint: disable=invalid-name

import unittest

import numpy as np

from naginterfaces.base import utils as b_utils
from naginterfaces.base import specfun as b_chapter
from naginterfaces.library import specfun as chapter
from test_functionality_core import IntegrationTest

class TestS01ba(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the s01ba wrapper.
    """

class TestS01baBase(TestS01ba, unittest.TestCase):
    """
    Concrete base tests.
    """

    def setUp(self):
        TestS01ba.set_up(self)
        self.fun = b_chapter.log_shifted

    def test_good(self):
        """
        Good call.
        """
        self.assertAlmostEqual(
            self.fun(0.2),
            0.182321556794,
        )

class TestS10aa(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the s10aa wrapper.
    """

class TestS10aaBase(TestS10aa, unittest.TestCase):
    """
    Concrete base tests.
    """

    def setUp(self):
        TestS10aa.set_up(self)
        self.fun = b_chapter.tanh

    def test_good(self):
        """
        Good call.
        """
        self.assertAlmostEqual(
            self.fun(0.2),
            0.197375320225,
        )

class TestS10ab(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the s10ab wrapper.
    """

class TestS10abBase(TestS10ab, unittest.TestCase):
    """
    Concrete base tests.
    """

    def setUp(self):
        TestS10ab.set_up(self)
        self.fun = b_chapter.sinh

    def test_good(self):
        """
        Good call.
        """
        self.assertAlmostEqual(
            self.fun(0.2),
            0.201336002541,
        )

class TestS10ac(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the s10ac wrapper.
    """

class TestS10acBase(TestS10ac, unittest.TestCase):
    """
    Concrete base tests.
    """

    def setUp(self):
        TestS10ac.set_up(self)
        self.fun = b_chapter.cosh

    def test_good(self):
        """
        Good call.
        """
        self.assertAlmostEqual(
            self.fun(0.2),
            1.020066755619,
        )

class TestS14aa(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the s14aa wrapper.
    """

class TestS14aaNp(TestS14aa, unittest.TestCase):
    """
    Concrete base tests.
    """

    def setUp(self):
        TestS14aa.set_up(self)
        self.fun = chapter.gamma

    def test_good(self):
        """
        Good call.
        """
        self.assertAlmostEqual(
            self.fun(1.25),
            0.9064,
            places=5,
        )

class TestS14af(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the s14af wrapper.
    """

class TestS14afBase(TestS14af, unittest.TestCase):
    """
    Concrete base tests.
    """

    def setUp(self):
        TestS14af.set_up(self)
        self.fun = b_chapter.psi_deriv_complex

    def test_good(self):
        """
        Good call.
        """
        self.assertAlmostEqual(
            self.fun(-1.5 + 2.5j, 1),
            -0.19737 - 0.24271j,
            places=5,
        )

class TestS17dc(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the s17dc wrapper.
    """

class TestS17dcBase(TestS17dc, unittest.TestCase):
    """
    Concrete base tests.
    """

    def setUp(self):
        TestS17dc.set_up(self)
        self.fun = b_chapter.bessel_y_complex

    def test_good(self):
        """
        Good call.
        """
        n = 2
        cy = np.empty(n, dtype=complex)
        self.fun(2.3, complex(2.), n, 'U', cy)
        np.testing.assert_allclose(
            cy,
            [complex(-0.739759), complex(-1.412003)],
            rtol=1e-6,
        )

class TestS17dcNp(TestS17dc, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestS17dc.set_up(self)
        self.fun = chapter.bessel_y_complex

    def test_good(self):
        """
        Good call.
        """
        # pylint: disable=invalid-name
        n = 2
        np.testing.assert_allclose(
            self.fun(2.3, complex(2.), n, 'U').cy,
            [complex(-0.739759), complex(-1.412003)],
            rtol=1e-6,
        )

class TestS18ac(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the s18ac wrapper.
    """

class TestS18acNp(TestS18ac, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestS18ac.set_up(self)
        self.fun = chapter.bessel_k0_real

    def test_good(self):
        """
        Good call.
        """
        self.assertAlmostEqual(
            self.fun(0.4),
            1.1145291,
        )

class TestS22ca(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the s22ca wrapper.
    """

class TestS22caNp(TestS22ca, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestS22ca.set_up(self)
        self.fun = chapter.mathieu_ang_periodic_real
        self.ordval = 2
        self.q = 2.
        self.parity = 0
        self.f = 1.0488344902527265
        self.f_deriv = 0.
        self.a = 5.172665133358294

    def test_bad_x_none(self):
        """
        Bad call.
        """
        self.my_assert_raises_regex(
            b_utils.NagTypeError,
            '.*must not be None when mode != 3',
            self.fun,
            self.ordval, self.q, self.parity, 2,
        )

    def test_good_char(self):
        """
        Good call.
        """
        soln = self.fun(self.ordval, self.q, self.parity, 3)
        self.assertAlmostEqual(
            soln.a,
            self.a,
        )
        self.assertIsNone(soln.f)
        self.assertIsNone(soln.f_deriv)

    def test_good_fval(self):
        """
        Good call.
        """
        soln = self.fun(self.ordval, self.q, self.parity, 0, [0., 1.])
        self.assertAlmostEqual(
            soln.a,
            self.a,
        )
        self.assertAlmostEqual(
            soln.f[0],
            self.f,
        )
        self.assertIsNone(soln.f_deriv)

    def test_good_fderiv(self):
        """
        Good call.
        """
        soln = self.fun(self.ordval, self.q, self.parity, 1, [0., 1.])
        self.assertAlmostEqual(
            soln.a,
            self.a,
        )
        self.assertAlmostEqual(
            soln.f_deriv[0],
            self.f_deriv,
        )
        self.assertIsNone(soln.f)

    def test_good_all(self):
        """
        Good call.
        """
        soln = self.fun(self.ordval, self.q, self.parity, 2, [0., 1.])
        self.assertAlmostEqual(
            soln.a,
            self.a,
        )
        self.assertAlmostEqual(
            soln.f_deriv[0],
            self.f_deriv,
        )
        self.assertAlmostEqual(
            soln.f[0],
            self.f,
        )

class TestS30aa(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the s30aa wrapper.
    """

class TestS30aaBase(TestS30aa, unittest.TestCase):
    """
    Concrete base tests.
    """

    def setUp(self):
        TestS30aa.set_up(self)
        self.fun = b_chapter.opt_bsm_price

    def test_bad_calput(self):
        """
        Bad calput supplied.
        """
        self.my_assert_raises_regex(
            b_utils.NagValueError,
            '.*calput = \'C\' or \'P\'.',
            self.fun,
            'z', None, None, None, None, None, None, None, None, None,
        )

if __name__ == '__main__': # pragma: no cover
    unittest.main()
