#!/usr/bin/env python
"""
Integration tests for wrapper functionality in the tsa submodule.

Public use of this module is not supported by NAG.
"""

# NAG Copyright 2018-2019.

# pylint: disable=invalid-name

import math
import unittest

import numpy as np

from naginterfaces.base import utils as b_utils
from naginterfaces.library import tsa as chapter
from test_functionality_core import IntegrationTest

class TestG13ek(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the g13ek wrapper.
    """

class TestG13ekNp(TestG13ek, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestG13ek.set_up(self)
        self.fun = chapter.kalman_unscented_state

    def test_bad_order(self):
        """
        Using Fortran-contiguous data except in callback f.
        """
        self.my_assert_raises_regex(
            b_utils.NagTypeError,
            r".*:f\).*must use the same storage order.*"
            r"callback storage order 'C'.*function used storage order 'F'$",
            self.run_one_step,
            'F', 'C'
        )

    def test_good_order_f(self):
        """
        Using Fortran-contiguous data everywhere.
        """
        self.run_one_step('F', 'F')

    @b_utils._break_ref_cycles # pylint: disable=protected-access
    def run_one_step(self, order, f_order): # pylint: disable=too-many-locals
        """
        The example problem with configurable storage order.

        Coverage for: breaking pointer reference cycles.
        """
        phi_rt = 0.4
        phi_lt = 0.1
        r = 3.0
        d = 4.0

        def cb_f(xt):
            "The (three-variable) state function."

            t1 = 0.5*r*(phi_rt + phi_lt)
            t3 = (r/d)*(phi_rt - phi_lt)

            fxt = np.empty(xt.shape, order=f_order)

            for i in range(xt.shape[1]):
                fxt[:, i] = xt[:, i] + [
                    math.cos(xt[2, i])*t1, math.sin(xt[2, i])*t1, t3
                ]

            return fxt

        Delta = 5.814
        A = 0.464

        def cb_h(my, yt):
            "The (three-variable, two-observation) measurement function."

            hyt = np.empty((my, yt.shape[1]), order=order)

            for i in range(yt.shape[1]):
                hyt[:, i] = [
                    Delta - yt[0, i]*math.cos(A) - yt[1, i]*math.sin(A),
                    yt[2, i] - A
                ]
                hyt[1, i] = hyt[1, i] + 2*math.pi

            return hyt

        mx = 3
        x = np.zeros(mx, order=order)
        lx = np.zeros((mx, mx), order=order)
        for i in range(lx.shape[0]): # pylint: disable=unsubscriptable-object
            lx[i, i] = 0.1
        st = np.zeros((mx, mx), order=order)
        for i in range(st.shape[0]): # pylint: disable=unsubscriptable-object
            st[i, i] = 0.1
        y = [5.262, 5.923]
        ly = np.zeros((2, 2), order=order)
        for i in range(ly.shape[0]): # pylint: disable=unsubscriptable-object
            ly[i, i] = 0.01

        x, st = self.fun(
            y, lx, ly, cb_f, cb_h, x, st,
        )
        np.testing.assert_allclose(
            x,
            [0.663776, -0.091915, 0.104341],
            rtol=1e-5,
        )
        np.testing.assert_allclose(
            st,
            [
                [-6.722933e-02, 0., 0.],
                [1.314612e-01, 2.210420e-02, 0.],
                [2.140849e-04, 3.599673e-05, 9.972062e-03]
            ],
            rtol=1e-5,
        )

class TestG13me(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the g13me wrapper.
    """

class TestG13meNp(TestG13me, unittest.TestCase):
    """
    Concrete level-2 tests.
    """
    def setUp(self):
        TestG13me.set_up(self)
        self.fun = chapter.inhom_iema

    def test_blocking_data(self):
        """
        Check that the data can be blocked and get the same
        result as not blocking it
        """
        iema = [
            0.6, 0.6, 0.8, 0.1, 0.2, 0.2, 0.5, 0.7, 0.1, 0.4, 0.7, 0.8,
            0.3, 0.2, 0.5, 0.2, 0.3, 0.8, 0.6, 0.1, 0.7, 0.9, 0.6, 0.3,
            0.1, 0.1, 0.4, 1.0, 1.0, 0.1,
        ]
        times = [
            7.5, 8.2, 18.1, 22.8, 25.8, 26.8, 31.1, 38.4, 45.9, 48.2, 48.9,
            57.9, 58.5, 63.9, 65.2, 66.6, 67.4, 69.3, 69.9, 73.0, 75.6,
            77.0, 84.7, 86.8, 88.0, 88.5, 91.0, 93.0, 93.7, 94.0,
        ]
        tau = 2.
        sinit = [5., 0.5, 0.5, 0.5]
        inter = [3, 2]
        res0 = self.fun(iema, times, tau, inter, sinit=sinit)
        comm = {}
        res1 = self.fun(
            iema[0:5], times[0:5], tau, inter, sinit=sinit, comm=comm,
        )
        res2 = self.fun(
            iema[5:], times[5:], tau, inter, pn=res1.pn, comm=comm,
        )
        for pair in zip(res0.iema[0:5], res1.iema):
            self.assertEqual(pair[0], pair[1])
        for pair in zip(res0.iema[5:], res2.iema):
            self.assertEqual(pair[0], pair[1])

class TestG13nd(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the g13nd wrapper.
    """

class TestG13ndNp(TestG13nd, unittest.TestCase):
    """
    Concrete level-2 tests.
    """
    def setUp(self):
        TestG13nd.set_up(self)
        self.fun = chapter.cp_binary

    def test_good(self):
        """
        Good call.
        """
        y = [
            0.00, 0.78, -0.02, 0.17, 0.04, -1.23, 0.24, 1.70, 0.77, 0.06,
            0.67, 0.94, 1.99, 2.64, 2.26, 3.72, 3.14, 2.28, 3.78, 0.83,
            2.80, 1.66, 1.93, 2.71, 2.97, 3.04, 2.29, 3.71, 1.69, 2.76,
            1.96, 3.17, 1.04, 1.50, 1.12, 1.11, 1.00, 1.84, 1.78, 2.39,
            1.85, 0.62, 2.16, 0.78, 1.70, 0.63, 1.79, 1.21, 2.20, -1.34,
            0.04, -0.14, 2.78, 1.83, 0.98, 0.19, 0.57, -1.41, 2.05, 1.17,
            0.44, 2.32, 0.67, 0.73, 1.17, -0.34, 2.95, 1.08, 2.16, 2.27,
            -0.14, -0.24, 0.27, 1.71, -0.04, -1.03, -0.12, -0.67, 1.15, -1.10,
            -1.37, 0.59, 0.44, 0.63, -0.06, -0.62, 0.39, -2.63, -1.63, -0.42,
            -0.73, 0.85, 0.26, 0.48, -0.26, -1.77, -1.53, -1.39, 1.68, 0.43,
        ]
        ctype = 1
        beta = 4.6
        param = 1.
        sp = self.fun(ctype, y, beta=beta, param=param).sparam
        self.assertEqual(len(sp.shape), 2)
        self.assertTrue(sp.flags.c_contiguous)
        self.assertAlmostEqual(sp[0, 0], 1.03/3.)

if __name__ == '__main__': # pragma: no cover
    unittest.main()
