#!/usr/bin/env python
"""
Integration tests for wrapper functionality in the wav submodule.

Public use of this module is not supported by NAG.
"""

# NAG Copyright 2018-2019.

# pylint: disable=invalid-name

import unittest

import numpy as np

from naginterfaces.library import wav as chap
from test_functionality_core import IntegrationTest

class TestC09ca(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the c09ca wrapper.
    """

class TestC09caNp(TestC09ca, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestC09ca.set_up(self)
        self.init = chap.dim1_init
        self.fun = chap.dim1_sngl_fwd

    def test_good(self):
        """
        Good call.
        """
        x = [
            1.0,
            3.0,
            5.0,
            7.0,
            6.0,
            4.0,
            5.0,
            2.0,
        ]
        wavnam = 'DB4'
        wtrans = 'Single level'
        mode = 'Zero'
        n = len(x)
        init_comm = self.init(wavnam, wtrans, mode, n)
        wvlt = self.fun(x, init_comm.comm)
        np.testing.assert_allclose(
            wvlt.ca,
            [0.0011, -0.0043, -0.0174, 4.4778, 8.9557, 7.3401, 2.5816],
            rtol=1.e-2,
        )
        np.testing.assert_allclose(
            wvlt.cd,
            [0.0237, 0.0410, -0.5966, 1.7763, -0.7517, 0.3332, -0.1188],
            rtol=1.e-3,
        )

class TestC09cc(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the c09cc wrapper.
    """

class TestC09ccNp(TestC09cc, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestC09cc.set_up(self)
        self.init = chap.dim1_init
        self.fun = chap.dim1_multi_fwd

    def test_good(self):
        """
        Good call.
        """
        x = [
            6.5271, 6.512, 6.5016, 6.5237, 6.4625,
            6.3496, 6.4025, 6.4035, 6.4407, 6.4746,
            6.5095, 6.6551, 6.61, 6.5969, 6.6083,
            6.652, 6.7113, 6.7227, 6.7196, 6.7649,
            6.7794, 6.8037, 6.8308, 6.7712, 6.7067,
            6.769, 6.7068, 6.7024, 6.6463, 6.6098,
            6.59, 6.596, 6.5457, 6.547, 6.5797,
            6.5895, 6.6275, 6.6795, 6.6598, 6.6925,
            6.6873, 6.7223, 6.7205, 6.6843, 6.703,
            6.647, 6.6008, 6.6061, 6.6097, 6.6485,
            6.6394, 6.6571, 6.6357, 6.6224, 6.6073,
            6.6075, 6.6379, 6.6294, 6.5906, 6.6258,
            6.6369, 6.6515, 6.6826, 6.7042,
        ]
        wavnam = 'DB4'
        wtrans = 'Multilevel'
        mode = 'Zero'
        n = len(x)
        init_comm = self.init(wavnam, wtrans, mode, n)
        wvlt = self.fun(x, init_comm.nwlmax, init_comm.comm)
        np.testing.assert_equal(
            wvlt.dwtlev,
            [7, 7, 8, 10, 14, 21, 35],
        )
        self.assertAlmostEqual(
            wvlt.c[2], -0.3446,
            places=4,
        )

class TestC09da(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the c09da wrapper.
    """

class TestC09daNp(TestC09da, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestC09da.set_up(self)
        self.init = chap.dim1_init
        self.fun = chap.dim1_mxolap_fwd

    def test_good(self):
        """
        Good call.
        """
        x = [
            1.0,
            3.0,
            5.0,
            7.0,
            6.0,
            4.0,
            5.0,
            2.0,
        ]
        wavnam = 'DB4'
        wtrans = 'Time invariant'
        mode = 'Periodic'
        n = len(x)
        init_comm = self.init(wavnam, wtrans, mode, n)
        wvlt = self.fun(x, init_comm.comm)
        np.testing.assert_allclose(
            wvlt.ca,
            [2.7781, 1.5146, 2.2505, 4.8788, 6.6845, 6.3423, 4.7869, 3.7644],
            rtol=1.e-2,
        )
        np.testing.assert_allclose(
            wvlt.cd,
            [
                -0.6187, 0.6272, 0.1883, -1.1966,
                1.2618, 0.3354, -0.3314, -0.2660,
            ],
            rtol=1.e-3,
        )

class TestC09dc(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the d09dc wrapper.
    """

class TestC09dcNp(TestC09dc, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestC09dc.set_up(self)
        self.init = chap.dim1_init
        self.fun = chap.dim1_mxolap_multi_fwd

    def test_good(self):
        """
        Good call.
        """
        x = [
            6.5271, 6.512, 6.5016, 6.5237, 6.4625,
            6.3496, 6.4025, 6.4035, 6.4407, 6.4746,
            6.5095, 6.6551, 6.61, 6.5969, 6.6083,
            6.652, 6.7113, 6.7227, 6.7196, 6.7649,
            6.7794, 6.8037, 6.8308, 6.7712, 6.7067,
            6.769, 6.7068, 6.7024, 6.6463, 6.6098,
            6.59, 6.596, 6.5457, 6.547, 6.5797,
            6.5895, 6.6275, 6.6795, 6.6598, 6.6925,
            6.6873, 6.7223, 6.7205, 6.6843, 6.703,
            6.647, 6.6008, 6.6061, 6.6097, 6.6485,
            6.6394, 6.6571, 6.6357, 6.6224, 6.6073,
            6.6075, 6.6379, 6.6294, 6.5906, 6.6258,
            6.6369, 6.6515, 6.6826, 6.7042,
        ]
        wavnam = 'DB4'
        wtrans = 'U'
        mode = 'Periodic'
        n = len(x)
        init_comm = self.init(wavnam, wtrans, mode, n)
        keepa = 'All'
        wvlt = self.fun(x, keepa, 2, init_comm.comm)
        np.testing.assert_equal(
            wvlt.na, 64,
        )
        self.assertAlmostEqual(
            wvlt.c[2], 6.6415,
            places=4,
        )

class TestC09ec(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the c09ec wrapper.
    """

class TestC09ecNp(TestC09ec, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestC09ec.set_up(self)
        self.init = chap.dim2_init
        self.fun = chap.dim2_multi_fwd

    def test_good(self):
        """
        Good call.
        """
        a = [
            [3., 7., 9., 1., 9., 9., 1., 0.],
            [9., 9., 3., 3., 4., 1., 2., 4.],
            [7., 8., 1., 3., 8., 9., 3., 3.],
            [1., 1., 1., 1., 2., 8., 4., 0.],
            [1., 2., 4., 6., 5., 6., 5., 4.],
            [2., 2., 5., 7., 3., 6., 6., 8.],
            [7., 9., 3., 1., 3., 4., 7., 2.],
        ]
        wavnam = 'DB2'
        wtrans = 'Multilevel'
        mode = 'Half'
        m = len(a)
        n = len(a[0])
        init_comm = self.init(wavnam, wtrans, mode, m, n)
        wvlt = self.fun(a, init_comm.nwlmax, init_comm.comm)
        np.testing.assert_equal(
            wvlt.dwtlvm,
            [4, 5],
        )
        np.testing.assert_equal(
            wvlt.dwtlvn,
            [4, 5],
        )
        self.assertAlmostEqual(
            wvlt.c[0],
            24.9724052747904395,
        )

class TestC09fa(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the c09fa wrapper.
    """

class TestC09faNp(TestC09fa, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestC09fa.set_up(self)
        self.init = chap.dim3_init
        self.fun = chap.dim3_sngl_fwd

    def test_good(self):
        """
        Good call.
        """
        a = [
            [
                [3., 2., 3.],
                [2., 1., 1.],
                [2., 5., 4.],
                [2., 1., 1.],
            ],
            [
                [2., 2., 1.],
                [9., 9., 1.],
                [1., 5., 2.],
                [2., 2., 1.],
            ],
            [
                [2., 2., 4.],
                [5., 3., 1.],
                [1., 2., 7.],
                [2., 7., 2.],
            ],
            [
                [1., 2., 3.],
                [6., 1., 2.],
                [2., 1., 1.],
                [2., 2., 5.],
            ],
            [
                [5., 2., 1.],
                [3., 1., 1.],
                [2., 2., 2.],
                [2., 8., 2.],
            ],
        ]
        wavnam = 'Haar'
        wtrans = 'Single level'
        mode = 'Half'
        m = len(a)
        n = len(a[0])
        fr = len(a[0][0])
        init_comm = self.init(wavnam, wtrans, mode, m, n, fr)
        wvlt = self.fun(a, init_comm.comm)
        self.assertAlmostEqual(
            wvlt[0],
            10.6066017177982168,
        )

class TestC09fc(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the c09fc wrapper.
    """

class TestC09fcNp(TestC09fc, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestC09fc.set_up(self)
        self.init = chap.dim3_init
        self.fun = chap.dim3_multi_fwd

    def test_good(self):
        """
        Good call.
        """
        a = [
            [
                [3.0, 2.0, 3.0, 5.0, 5.0],
                [2.0, 1.0, 1.0, 8.0, 2.0],
                [2.0, 5.0, 4.0, 1.0, 1.0],
                [2.0, 1.0, 1.0, 2.0, 2.0],
                [1.0, 2.0, 1.0, 2.0, 1.0],
                [1.0, 3.0, 1.0, 1.0, 1.0],
            ],
            [
                [2.0, 2.0, 1.0, 1.0, 3.0],
                [9.0, 9.0, 1.0, 2.0, 1.0],
                [1.0, 5.0, 2.0, 2.0, 9.0],
                [2.0, 2.0, 1.0, 9.0, 1.0],
                [1.0, 1.0, 2.0, 2.0, 2.0],
                [3.0, 2.0, 6.0, 9.0, 1.0],
            ],
            [
                [2.0, 2.0, 4.0, 2.0, 2.0],
                [5.0, 3.0, 1.0, 2.0, 3.0],
                [1.0, 2.0, 7.0, 2.0, 1.0],
                [2.0, 7.0, 2.0, 1.0, 1.0],
                [1.0, 1.0, 5.0, 1.0, 7.0],
                [1.0, 1.0, 6.0, 3.0, 2.0],
            ],
            [
                [1.0, 2.0, 3.0, 1.0, 7.0],
                [6.0, 1.0, 2.0, 1.0, 2.0],
                [2.0, 1.0, 1.0, 1.0, 2.0],
                [2.0, 2.0, 5.0, 5.0, 6.0],
                [7.0, 3.0, 9.0, 1.0, 1.0],
                [2.0, 1.0, 5.0, 2.0, 1.0],
            ],
            [
                [5.0, 2.0, 1.0, 3.0, 5.0],
                [3.0, 1.0, 1.0, 2.0, 1.0],
                [2.0, 2.0, 2.0, 8.0, 7.0],
                [2.0, 8.0, 2.0, 1.0, 2.0],
                [4.0, 3.0, 2.0, 9.0, 1.0],
                [7.0, 3.0, 1.0, 2.0, 1.0],
            ],
            [
                [2.0, 1.0, 2.0, 2.0, 2.0],
                [2.0, 4.0, 6.0, 1.0, 1.0],
                [1.0, 5.0, 3.0, 9.0, 3.0],
                [1.0, 1.0, 9.0, 1.0, 2.0],
                [2.0, 2.0, 5.0, 2.0, 2.0],
                [1.0, 7.0, 1.0, 2.0, 1.0],
            ],
            [
                [6.0, 8.0, 1.0, 3.0, 5.0],
                [2.0, 1.0, 1.0, 6.0, 3.0],
                [1.0, 3.0, 8.0, 5.0, 9.0],
                [3.0, 9.0, 2.0, 3.0, 1.0],
                [6.0, 1.0, 1.0, 2.0, 4.0],
                [9.0, 2.0, 3.0, 2.0, 1.0],
            ],
        ]
        wavnam = 'Bior1.1'
        wtrans = 'Multilevel'
        mode = 'Periodic'
        m = len(a)
        n = len(a[0])
        fr = len(a[0][0])
        init_comm = self.init(wavnam, wtrans, mode, m, n, fr)
        self.assertEqual(
            init_comm.nwct, 316,
        )
        wvlt = self.fun(a, init_comm.nwlmax, init_comm.comm)
        np.testing.assert_equal(
            wvlt.dwtlvm,
            [2, 4],
        )
        np.testing.assert_equal(
            wvlt.dwtlvn,
            [2, 3],
        )
        np.testing.assert_equal(
            wvlt.dwtlvfr,
            [2, 3],
        )
        self.assertEqual(
            len(wvlt.c), init_comm.nwct,
        )
        self.assertAlmostEqual(
            wvlt.c[0],
            21.875,
        )

if __name__ == '__main__': # pragma: no cover
    unittest.main()
