#!/usr/bin/env python
"""
Integration tests for wrapper functionality in the zeros submodule.

Public use of this module is not supported by NAG.
"""

# NAG Copyright 2018-2019.

# pylint: disable=invalid-name

import os
import tempfile
import unittest

import numpy as np

from naginterfaces.base import utils as b_utils
from naginterfaces.base import zeros as b_chapter
from naginterfaces.library import zeros as chapter
from test_functionality_core import IntegrationTest

class TestC02af(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the c02af wrapper.
    """

class TestC02afBase(TestC02af, unittest.TestCase):
    """
    Concrete base tests.
    """

    def setUp(self):
        TestC02af.set_up(self)
        self.fun = b_chapter.poly_complex

    def test_good_debug(self):
        """
        Test correct operation.

        Coverage for: debug tracing.
        """
        t_fd, t_path = tempfile.mkstemp()
        n = 5
        z = np.empty((2, n))
        with b_utils._Debug(filename=t_path): # pylint: disable=protected-access
            self.fun(
                np.array(
                    [
                        [5., 30., -0.2, 50., -2., 10.],
                        [6., 20., -6., 100000., 40., 1.]
                    ]
                ),
                n,
                True,
                z,
            )
        t_fo = open(t_path, 'r')
        self.assertTrue(
            t_fo.readline(),
            'DEBUG:naginterfaces',
        )
        t_fo.close()
        os.close(t_fd)
        os.unlink(t_path)

    def test_bad_n(self):
        """
        Invalid n intercepted in the correct place.
        """
        self.my_assert_raises_regex(
            b_utils.NagShapeError,
            '.*expected extent in dimension 2 .* must not be negative',
            self.fun,
            np.empty((2, 2)), -2, True, np.empty((2, 1)),
        )

class TestC02afNp(TestC02af, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestC02af.set_up(self)
        self.fun = chapter.poly_complex

    def test_good_arraylike(self):
        """
        Test correct operation.
        """
        z_a = self.fun(
            [
                5. + 6.j,
                30. + 20.j,
                -0.2 -6.j,
                50. + 100000.j,
                -2. + 40.j,
                10. + 1.j,
            ]
        )
        self.assertIsInstance(z_a, np.ndarray)
        np.testing.assert_allclose(
            z_a,
            [
                -2.432779e+01 + -4.855474j,
                5.248669e+00 + 22.73587j,
                1.465329e+01 - 16.56890j,
                -6.926386e-03 - 7.443430e-03j,
                6.526396e-03 + 7.423236e-03j
            ],
            rtol=1e-5
        )

    def test_bad_n(self):
        """
        Invalid n intercepted in the correct place.
        """
        self.my_assert_raises_regex(
            b_utils.NagShapeError,
            '.*expected extent in dimension 2 .* must not be negative',
            self.fun,
            np.empty((0,), dtype=complex), True,
        )

class TestC02ag(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the c02ag wrapper.
    """

class TestC02agBase(TestC02ag, unittest.TestCase):
    """
    Concrete base tests.
    """

    def setUp(self):
        TestC02ag.set_up(self)
        self.fun = b_chapter.poly_real

    def test_good(self):
        """
        Good call.
        """
        n = 5
        z = np.empty((2, n))
        self.fun(
            np.array([1., 2., 3., 4., 5., 6.]),
            n, True, z,
        )
        np.testing.assert_allclose(
            z,
            [
                [-1.4918, 0.55169, 0.55169, -0.80579, -0.80579],
                [0., 1.25334886, -1.25334886, 1.22290471, -1.22290471]
            ],
            rtol=1e-5,
        )

class TestC02agNp(TestC02ag, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestC02ag.set_up(self)
        self.fun = chapter.poly_real

    def test_good(self):
        """
        Good call.
        """
        np.testing.assert_allclose(
            self.fun(np.array([1, 2, 3, 4, 5, 6])),
            [
                -1.4918,
                0.55169 + 1.25334886j,
                0.55169 - 1.25334886j,
                -0.80579 + 1.22290471j,
                -0.80579 - 1.22290471j,
            ],
            rtol=1e-5,
        )

class TestC02ah(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the c02ah wrapper.
    """

class TestC02ahNp(TestC02ah, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestC02ah.set_up(self)
        self.fun = chapter.quadratic_complex

    def test_good(self):
        """
        Good call.
        """
        zs = self.fun(
            complex(1.), complex(-3., 1.), complex(8., 1.),
        )
        self.assertAlmostEqual(
            zs.z_small, complex(1., 2.),
        )
        self.assertAlmostEqual(
            zs.z_large, complex(2., -3.),
        )

class TestC02aj(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the c02aj wrapper.
    """

class TestC02ajNp(TestC02aj, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestC02aj.set_up(self)
        self.fun = chapter.quadratic_real

    def test_good(self):
        """
        Good call.
        """
        zs = self.fun(1., 3., -10.)
        self.assertAlmostEqual(
            zs.z_small, complex(2.),
        )
        self.assertAlmostEqual(
            zs.z_large, complex(-5.),
        )

class TestC02ak(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the c02ak wrapper.
    """

class TestC02akNp(TestC02ak, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestC02ak.set_up(self)
        self.fun = chapter.cubic_real

    def test_good(self):
        """
        Good call.
        """
        np.testing.assert_allclose(
            self.fun(1., 3., 9., -13.).zero,
            [
                complex(1.),
                complex(-2., 3.),
                complex(-2., -3.),
            ],
        )

class TestC02al(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the c02al wrapper.
    """

class TestC02alNp(TestC02al, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestC02al.set_up(self)
        self.fun = chapter.quartic_real

    def test_good(self):
        """
        Good call.
        """
        np.testing.assert_allclose(
            self.fun(1., 2., 6., -8., -40.).zero,
            [
                complex(2.),
                complex(-2.),
                complex(-1., 3.),
                complex(-1., -3.),
            ],
        )

class TestC02am(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the c02am wrapper.
    """

class TestC02amNp(TestC02am, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestC02am.set_up(self)
        self.fun = chapter.cubic_complex

    def test_good(self):
        """
        Good call.
        """
        np.testing.assert_allclose(
            self.fun(
                complex(1.),
                complex(-2., 3.),
                complex(5., 14.),
                complex(-40., -5.),
            ).zero,
            [
                complex(-2., 3.),
                complex(1., -2.),
                complex(3., -4.),
            ],
        )

class TestC02an(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the c02an wrapper.
    """

class TestC02anNp(TestC02an, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestC02an.set_up(self)
        self.fun = chapter.quartic_complex

    def test_good(self):
        """
        Good call.
        """
        np.testing.assert_allclose(
            self.fun(
                complex(1.),
                complex(0.),
                complex(0., 16.),
                complex(-8., 8.),
                complex(-65.),
            ).zero,
            [
                complex(3., -2.),
                complex(1., -2.),
                complex(-2., 1.),
                complex(-2., 3.),
            ],
        )

if __name__ == '__main__': # pragma: no cover
    unittest.main()
