# -*- coding: utf-8 -*-
# pylint: disable=bare-except,chained-comparison,invalid-name,len-as-condition,line-too-long,protected-access,redefined-outer-name,too-many-arguments,too-many-boolean-expressions,too-many-branches,too-many-lines,too-many-locals,too-many-statements,unneeded-not,unused-import
u"""
Interfaces for the NAG Mark 27.0 `correg` Chapter.

``correg`` - Correlation and Regression Analysis

Base interfaces.

We do not intend for these lower-level base interfaces to be called
directly from user code. Please refer to the ``naginterfaces.library``
subpackage instead.

This module is concerned with two techniques

(i) correlation analysis and

(#) regression modelling,

both of which are concerned with determining the inter-relationships
among two or more variables.

Other modules of the NAG Library which cover similar problems are
submodule ``fit`` and submodule ``opt``. Submodule ``fit`` functions may
be used to fit linear models by criteria other than least squares, and
also for polynomial regression; submodule ``opt`` functions may be used
to fit nonlinear models and linearly constrained linear models.

For full information please refer to the NAG Library document

https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02intro.html
"""

# NAG Copyright 2017-2020.

import collections as _collections
import logging as _logging
import sys as _sys

from ..base import _nag_e_consts
from ..base import _nag_enum_consts
from ..base import _nag_g_consts
from ..base import utils

def corrmat_nearest(g, n, errtol, maxits, maxit, x):
    u"""
    Computes the nearest correlation matrix to a real square matrix,
    using the method of Qi and Sun.

    ``corrmat_nearest`` computes the nearest correlation matrix, in the
    Frobenius norm, to a given square, input matrix.

    For full information please refer to the NAG Library document for
    g02aa

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02aaf.html

    Parameters
    ----------
    g : float, shape (`n`, `n`), modified in place
        `On entry`: G, the initial matrix.

        `On exit`: a symmetric matrix 1/2(G+G^T) with the diagonal set
        to I.

    n : int
        The size of the matrix G.

    errtol : float
        The termination tolerance for the Newton iteration. If `errtol`
        <= 0.0, `n`*sqrt(machine precision) is used.

    maxits : int
        `maxits` specifies the maximum number of iterations used for the
        iterative scheme used to solve the linear algebraic equations at
        each Newton step.

        If `maxits` <= 0, 100 is used.

    maxit : int
        Specifies the maximum number of Newton iterations.

        If `maxit` <= 0, 200 is used.

    x : float, shape (`n`, `n`), modified in place
        `On exit`: contains the nearest correlation matrix.

    Returns
    -------
    itera : int
        The number of Newton steps taken.

    feval : int
        The number of function evaluations of the dual problem.

    nrmgrd : float
        The norm of the gradient of the last Newton step.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` > 0.

        (`errno` 2)
            Newton iteration fails to converge in *<value>* iterations.

        (`errno` 4)
            An intermediate eigenproblem could not be solved. This
            should not occur. Please contact NAG with details of your
            call.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            Machine precision is limiting convergence.

            The array returned in `x` may still be of interest.

    References
    ----------
    Borsdorf, R and Higham, N J, 2010, `A preconditioned (Newton)
    algorithm for the nearest correlation matrix`, IMA Journal of
    Numerical Analysis (30(1)), 94--107

    Qi, H and Sun, D, 2006, `A quadratically convergent Newton method
    for computing the nearest correlation matrix`, SIAM J. Matrix
    AnalAppl (29(2)), 360--385

    See Also
    --------
    :meth:`naginterfaces.library.examples.correg.corrmat_nearest_ex.main`
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.corrmat_nearest'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02aaft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._EngineFloat64ArrayType.c_type, # Input/Output, g[ldg, n].
        utils._BYREF_INT, # Input, ldg.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_FLOAT64, # Input, errtol.
        utils._BYREF_INT, # Input, maxits.
        utils._BYREF_INT, # Input, maxit.
        utils._EngineFloat64ArrayType.c_type, # Output, x[ldx, n].
        utils._BYREF_INT, # Input, ldx.
        utils._BYREF_INT, # Output, itera.
        utils._BYREF_INT, # Output, feval.
        utils._BYREF_FLOAT64, # Output, nrmgrd.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((g, 'g'), (x, 'x')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _ldg_py = utils._EngineFloat64ArrayType.get_nd(g, {'fun_name': fun_name, 'entity': 'g'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldg = utils._EngineIntScalarType(
        _ldg_py, {'fun_name': fun_name, 'entity': '_ldg'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _g_cls = utils._EngineFloat64ArrayType(
        g, {'fun_name': fun_name, 'entity': 'g'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
    )
    _g = _g_cls.to_ctypes()
    _errtol = utils._EngineFloat64ScalarType(
        errtol, {'fun_name': fun_name, 'entity': 'errtol'},
    ).to_ctypes()
    _maxits = utils._EngineIntScalarType(
        maxits, {'fun_name': fun_name, 'entity': 'maxits'},
    ).to_ctypes()
    _maxit = utils._EngineIntScalarType(
        maxit, {'fun_name': fun_name, 'entity': 'maxit'},
    ).to_ctypes()
    _ldx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _itera = utils._EngineIntScalarType.empty_ctypes()
    _feval = utils._EngineIntScalarType.empty_ctypes()
    _nrmgrd = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _g,
        _ldg,
        _n,
        _errtol,
        _maxits,
        _maxit,
        _x,
        _ldx,
        _itera,
        _feval,
        _nrmgrd,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1001:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 1001:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Newton iteration fails to converge in {:s} iterations.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 1005:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Machine precision is limiting convergence.',
                    'The array returned in x may still be of interest.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 1006:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'An intermediate eigenproblem could not be solved. This should not occur.',
                    'Please contact NAG with details of your call.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _g_cls.refresh_py(_g)
        _x_cls.refresh_py(_x)
        _itera_py = utils._EngineIntScalarType.to_py(_itera)
        _feval_py = utils._EngineIntScalarType.to_py(_feval)
        _nrmgrd_py = utils._EngineFloat64ScalarType.to_py(_nrmgrd)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['itera', 'feval', 'nrmgrd'],
        )(
            _itera_py,
            _feval_py,
            _nrmgrd_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def corrmat_nearest_bounded(g, n, opt, alpha, w, errtol, maxits, maxit, x):
    u"""
    Computes the nearest correlation matrix to a real square matrix,
    augmenting ``corrmat_nearest`` to incorporate weights and bounds.

    ``corrmat_nearest_bounded`` computes the nearest correlation matrix,
    in the Frobenius norm or weighted Frobenius norm, and optionally
    with bounds on the eigenvalues, to a given square, input matrix.

    For full information please refer to the NAG Library document for
    g02ab

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02abf.html

    Parameters
    ----------
    g : float, shape (`n`, `n`), modified in place
        `On entry`: G, the initial matrix.

        `On exit`: G is overwritten.

    n : int
        The order of the matrix G.

    opt : str, length 1
        Indicates the problem to be solved.

        `opt` = 'A'

            The lower bound problem is solved.

        `opt` = 'W'

            The weighted norm problem is solved.

        `opt` = 'B'

            Both problems are solved.

    alpha : None or float
        The value of alpha.

        If `opt` = 'W', `alpha` need not be set.

    w : None or float, shape (`n`), modified in place
        `Optionally, on entry`: the square roots of the diagonal
        elements of W, that is the diagonal of W^(1/2).

        If `opt` = 'A', `w` is not referenced and may be None.

        `On exit`, if not None on entry: if `opt` = 'W' or 'B', the
        array is scaled so 0 < `w`[i-1] <= 1, for i = 1,2,...,n.

    errtol : float
        The termination tolerance for the Newton iteration. If `errtol`
        <= 0.0, `n`*sqrt(machine precision) is used.

    maxits : int
        Specifies the maximum number of iterations to be used by the
        iterative scheme to solve the linear algebraic equations at each
        Newton step.

        If `maxits` <= 0, 2*`n` is used.

    maxit : int
        Specifies the maximum number of Newton iterations.

        If `maxit` <= 0, 200 is used.

    x : float, shape (`n`, `n`), modified in place
        `On exit`: contains the nearest correlation matrix.

    Returns
    -------
    itera : int
        The number of Newton steps taken.

    feval : int
        The number of function evaluations of the dual problem.

    nrmgrd : float
        The norm of the gradient of the last Newton step.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` > 0.

        (`errno` 1)
            On entry, the value of `opt` is invalid.

            Constraint: `opt` = 'A', 'W' or 'B'.

        (`errno` 1)
            On entry, `alpha` = *<value>*.

            Constraint: 0.0 < `alpha` < 1.0.

        (`errno` 1)
            On entry, all elements of `w` were not positive.

            Constraint: `w`[i-1] > 0.0, for all i.

        (`errno` 2)
            Newton iteration fails to converge in *<value>* iterations.
            Increase `maxit` or check the call to the function.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            The machine precision is limiting convergence. In this
            instance the returned value of `x` may be useful.

        (`errno` 4)
            An intermediate eigenproblem could not be solved. This
            should not occur. Please contact NAG with details of your
            call.

    References
    ----------
    Borsdorf, R and Higham, N J, 2010, `A preconditioned (Newton)
    algorithm for the nearest correlation matrix`, IMA Journal of
    Numerical Analysis (30(1)), 94--107

    Qi, H and Sun, D, 2006, `A quadratically convergent Newton method
    for computing the nearest correlation matrix`, SIAM J. Matrix
    AnalAppl (29(2)), 360--385
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.corrmat_nearest_bounded'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02abft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._EngineFloat64ArrayType.c_type, # Input/Output, g[ldg, n].
        utils._BYREF_INT, # Input, ldg.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_STRING, # Input, opt(len=1).
        utils._BYREF_FLOAT64, # Input, alpha.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, w[n].
        utils._BYREF_FLOAT64, # Input, errtol.
        utils._BYREF_INT, # Input, maxits.
        utils._BYREF_INT, # Input, maxit.
        utils._EngineFloat64ArrayType.c_type, # Output, x[ldx, n].
        utils._BYREF_INT, # Input, ldx.
        utils._BYREF_INT, # Output, itera.
        utils._BYREF_INT, # Output, feval.
        utils._BYREF_FLOAT64, # Output, nrmgrd.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of opt.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((g, 'g'), (x, 'x')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _ldg_py = utils._EngineFloat64ArrayType.get_nd(g, {'fun_name': fun_name, 'entity': 'g'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldg = utils._EngineIntScalarType(
        _ldg_py, {'fun_name': fun_name, 'entity': '_ldg'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _g_cls = utils._EngineFloat64ArrayType(
        g, {'fun_name': fun_name, 'entity': 'g'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
    )
    _g = _g_cls.to_ctypes()
    _opt = utils._EngineCharScalarType(
        opt, {'fun_name': fun_name, 'entity': 'opt'}, exp_chrlen=1,
    ).to_ctypes()
    _opt_py = utils._EngineCharScalarType.to_py(_opt)
    if _opt_py not in ('W', 'w'):
        if alpha is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'alpha'},
                'must not be None when opt not in (\'W\', \'w\')',
            )
    if alpha is None:
        _alpha = utils._EngineFloat64ScalarType.c_type(0.0)
    else:
        _alpha = utils._EngineFloat64ScalarType(
            alpha, {'fun_name': fun_name, 'entity': 'alpha'},
            allow_none=True,
        ).to_ctypes()
    if _opt_py not in ('A', 'a'):
        if w is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'w'},
                'must not be None when opt not in (\'A\', \'a\')',
            )
    if _opt_py not in ('A', 'a'):
        if w is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'w'},
                'must not be None when opt not in (\'A\', \'a\')',
            )
    if w is None:
        _w = utils._EngineFloat64ArrayType.empty_ctypes(_n_py)
    else:
        _w_cls = utils._EngineFloat64ArrayType(
            w, {'fun_name': fun_name, 'entity': 'w'},
            exp_shape=(_n_py,),
            check_contents=_opt_py not in ('A', 'a'),
            allow_none=True,
        )
        _w = _w_cls.to_ctypes(can_reference_contents=_opt_py not in ('A', 'a'))
    _errtol = utils._EngineFloat64ScalarType(
        errtol, {'fun_name': fun_name, 'entity': 'errtol'},
    ).to_ctypes()
    _maxits = utils._EngineIntScalarType(
        maxits, {'fun_name': fun_name, 'entity': 'maxits'},
    ).to_ctypes()
    _maxit = utils._EngineIntScalarType(
        maxit, {'fun_name': fun_name, 'entity': 'maxit'},
    ).to_ctypes()
    _ldx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _itera = utils._EngineIntScalarType.empty_ctypes()
    _feval = utils._EngineIntScalarType.empty_ctypes()
    _nrmgrd = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _g,
        _ldg,
        _n,
        _opt,
        _alpha,
        _w,
        _errtol,
        _maxits,
        _maxit,
        _x,
        _ldx,
        _itera,
        _feval,
        _nrmgrd,
        _errbuf,
        _ierr,
        _opt._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1001:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1007:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the value of opt is invalid.',
                    'Constraint: opt = \'A\', \'W\' or \'B\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1008:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, alpha = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: 0.0 < alpha < 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1009:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, all elements of w were not positive.',
                    'Constraint: w[i-1] > 0.0, for all i.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 1001:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Newton iteration fails to converge in {:s} iterations.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Increase maxit or check the call to the function.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 1005:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The machine precision is limiting convergence.',
                    'In this instance the returned value of x may be useful.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 1006:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'An intermediate eigenproblem could not be solved. This should not occur.',
                    'Please contact NAG with details of your call.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _g_cls.refresh_py(_g)
        if w is not None:
            _w_cls.refresh_py(_w)
        _x_cls.refresh_py(_x)
        _itera_py = utils._EngineIntScalarType.to_py(_itera)
        _feval_py = utils._EngineIntScalarType.to_py(_feval)
        _nrmgrd_py = utils._EngineFloat64ScalarType.to_py(_nrmgrd)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['itera', 'feval', 'nrmgrd'],
        )(
            _itera_py,
            _feval_py,
            _nrmgrd_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def corrmat_nearest_kfactor(g, n, k, errtol, maxit, x):
    u"""
    Computes the nearest correlation matrix with k-factor structure to a
    real square matrix.

    ``corrmat_nearest_kfactor`` computes the factor loading matrix
    associated with the nearest correlation matrix with k-factor
    structure, in the Frobenius norm, to a given square, input matrix.

    For full information please refer to the NAG Library document for
    g02ae

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02aef.html

    Parameters
    ----------
    g : float, shape (`n`, `n`), modified in place
        `On entry`: G, the initial matrix.

        `On exit`: a symmetric matrix 1/2(G+G^T) with the diagonal
        elements set to unity.

    n : int
        n, the order of the matrix G.

    k : int
        k, the number of factors and columns of X.

    errtol : float
        The termination tolerance for the projected gradient norm. See
        references for further details. If `errtol` <= 0.0, 0.01 is
        used. This is often a suitable default value.

    maxit : int
        Specifies the maximum number of iterations in the spectral
        projected gradient method.

        If `maxit` <= 0, 40000 is used.

    x : float, shape (`n`, `k`), modified in place
        `On exit`: contains the matrix X.

    Returns
    -------
    itera : int
        The number of steps taken in the spectral projected gradient
        method.

    feval : int
        The number of evaluations of ||G-XX^T+diag(XX^T-I)||_F.

    nrmpgd : float
        The norm of the projected gradient at the final iteration.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` > 0.

        (`errno` 1)
            On entry, `k` = *<value>* and `n` = *<value>*.

            Constraint: 0 < `k` <= `n`.

        (`errno` 2)
            Spectral gradient method fails to converge in *<value>*
            iterations.

    References
    ----------
    Birgin, E G, Martínez, J M and Raydan, M, 2001, `Algorithm 813:
    SPG--software for convex-constrained optimization`, ACM Trans. Math.
    Software (27), 340--349

    Borsdorf, R, Higham, N J and Raydan, M, 2010, `Computing a nearest
    correlation matrix with factor structure`, SIAM J. Matrix Anal.
    Appl. (31(5)), 2603--2622
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.corrmat_nearest_kfactor'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02aeft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._EngineFloat64ArrayType.c_type, # Input/Output, g[ldg, n].
        utils._BYREF_INT, # Input, ldg.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, k.
        utils._BYREF_FLOAT64, # Input, errtol.
        utils._BYREF_INT, # Input, maxit.
        utils._EngineFloat64ArrayType.c_type, # Output, x[ldx, k].
        utils._BYREF_INT, # Input, ldx.
        utils._BYREF_INT, # Output, itera.
        utils._BYREF_INT, # Output, feval.
        utils._BYREF_FLOAT64, # Output, nrmpgd.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((g, 'g'), (x, 'x')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _ldg_py = utils._EngineFloat64ArrayType.get_nd(g, {'fun_name': fun_name, 'entity': 'g'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldg = utils._EngineIntScalarType(
        _ldg_py, {'fun_name': fun_name, 'entity': '_ldg'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _g_cls = utils._EngineFloat64ArrayType(
        g, {'fun_name': fun_name, 'entity': 'g'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
    )
    _g = _g_cls.to_ctypes()
    _k = utils._EngineIntScalarType(
        k, {'fun_name': fun_name, 'entity': 'k'},
    ).to_ctypes()
    _errtol = utils._EngineFloat64ScalarType(
        errtol, {'fun_name': fun_name, 'entity': 'errtol'},
    ).to_ctypes()
    _maxit = utils._EngineIntScalarType(
        maxit, {'fun_name': fun_name, 'entity': 'maxit'},
    ).to_ctypes()
    _ldx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _k_py = utils._EngineIntScalarType.to_py(_k)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py, _k_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _itera = utils._EngineIntScalarType.empty_ctypes()
    _feval = utils._EngineIntScalarType.empty_ctypes()
    _nrmpgd = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _g,
        _ldg,
        _n,
        _k,
        _errtol,
        _maxit,
        _x,
        _ldx,
        _itera,
        _feval,
        _nrmpgd,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1001:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1002:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, k = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: 0 < k <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 1001:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Spectral gradient method fails to converge in {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'iterations.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _g_cls.refresh_py(_g)
        _x_cls.refresh_py(_x)
        _itera_py = utils._EngineIntScalarType.to_py(_itera)
        _feval_py = utils._EngineIntScalarType.to_py(_feval)
        _nrmpgd_py = utils._EngineFloat64ScalarType.to_py(_nrmpgd)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['itera', 'feval', 'nrmpgd'],
        )(
            _itera_py,
            _feval_py,
            _nrmpgd_py,
        )
    _err_class.manage()
    return return_data

def corrmat_h_weight(g, n, alpha, h, errtol, maxit, x):
    u"""
    Computes the nearest correlation matrix to a real square matrix,
    using element-wise weighting.

    ``corrmat_h_weight`` computes the nearest correlation matrix, using
    element-wise weighting in the Frobenius norm and optionally with
    bounds on the eigenvalues, to a given square, input matrix.

    For full information please refer to the NAG Library document for
    g02aj

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02ajf.html

    Parameters
    ----------
    g : float, shape (`n`, `n`), modified in place
        `On entry`: G, the initial matrix.

        `On exit`: G is overwritten.

    n : int
        The order of the matrix G.

    alpha : float
        The value of alpha.

        If `alpha` < 0.0, 0.0 is used.

    h : float, shape (`n`, `n`), modified in place
        `On entry`: the matrix of weights H.

        `On exit`: a symmetric matrix 1/2(H+H^T) with its diagonal
        elements set to zero and the remaining elements scaled so that
        the maximum element is 1.0.

    errtol : float
        The termination tolerance for the iteration. If `errtol` <= 0.0,
        `n`*sqrt(machine precision) is used. See [Accuracy] for further
        details.

    maxit : int
        Specifies the maximum number of iterations to be used.

        If `maxit` <= 0, 200 is used.

    x : float, shape (`n`, `n`), modified in place
        `On exit`: contains the nearest correlation matrix.

    Returns
    -------
    itera : int
        The number of iterations taken.

    norm : float
        The value of ||H.(G-X)||_F after the final iteration.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` > 0.

        (`errno` 5)
            On entry, `alpha` = *<value>*.

            Constraint: `alpha` < 1.0.

        (`errno` 6)
            On entry, one or more of the off-diagonal elements of H were
            negative.

        (`errno` 7)
            Function failed to converge in *<value>* iterations.

            Increase `maxit` or check the call to the function.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 8)
            Failure to solve intermediate eigenproblem. This should not
            occur. Please contact NAG with details of your call.

    References
    ----------
    Borsdorf, R and Higham, N J, 2010, `A preconditioned (Newton)
    algorithm for the nearest correlation matrix`, IMA Journal of
    Numerical Analysis (30(1)), 94--107

    Jiang, K, Sun, D and Toh, K-C, 2012, `An inexact accelerated
    proximal gradient method for large scale linearly constrained convex
    SDP`, SIAM J. Optim. (22(3)), 1042--1064

    Qi, H and Sun, D, 2006, `A quadratically convergent Newton method
    for computing the nearest correlation matrix`, SIAM J. Matrix
    AnalAppl (29(2)), 360--385
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.corrmat_h_weight'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02ajft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._EngineFloat64ArrayType.c_type, # Input/Output, g[ldg, n].
        utils._BYREF_INT, # Input, ldg.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_FLOAT64, # Input, alpha.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, h[ldh, n].
        utils._BYREF_INT, # Input, ldh.
        utils._BYREF_FLOAT64, # Input, errtol.
        utils._BYREF_INT, # Input, maxit.
        utils._EngineFloat64ArrayType.c_type, # Output, x[ldx, n].
        utils._BYREF_INT, # Input, ldx.
        utils._BYREF_INT, # Output, itera.
        utils._BYREF_FLOAT64, # Output, norm.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((g, 'g'), (h, 'h'), (x, 'x')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _ldg_py = utils._EngineFloat64ArrayType.get_nd(g, {'fun_name': fun_name, 'entity': 'g'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldg = utils._EngineIntScalarType(
        _ldg_py, {'fun_name': fun_name, 'entity': '_ldg'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _g_cls = utils._EngineFloat64ArrayType(
        g, {'fun_name': fun_name, 'entity': 'g'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
    )
    _g = _g_cls.to_ctypes()
    _alpha = utils._EngineFloat64ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    _ldh_py = utils._EngineFloat64ArrayType.get_nd(h, {'fun_name': fun_name, 'entity': 'h'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldh = utils._EngineIntScalarType(
        _ldh_py, {'fun_name': fun_name, 'entity': '_ldh'},
    ).to_ctypes()
    _h_cls = utils._EngineFloat64ArrayType(
        h, {'fun_name': fun_name, 'entity': 'h'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
    )
    _h = _h_cls.to_ctypes()
    _errtol = utils._EngineFloat64ScalarType(
        errtol, {'fun_name': fun_name, 'entity': 'errtol'},
    ).to_ctypes()
    _maxit = utils._EngineIntScalarType(
        maxit, {'fun_name': fun_name, 'entity': 'maxit'},
    ).to_ctypes()
    _ldx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _itera = utils._EngineIntScalarType.empty_ctypes()
    _norm = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _g,
        _ldg,
        _n,
        _alpha,
        _h,
        _ldh,
        _errtol,
        _maxit,
        _x,
        _ldx,
        _itera,
        _norm,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1003:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 1004:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, alpha = {:s}.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: alpha < 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 1005:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, one or more of the off-diagonal elements of H were negative.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 1012:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Function failed to converge in {:s} iterations.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Increase maxit or check the call to the function.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 1013:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Failure to solve intermediate eigenproblem. This should not occur.',
                    'Please contact NAG with details of your call.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _g_cls.refresh_py(_g)
        _h_cls.refresh_py(_h)
        _x_cls.refresh_py(_x)
        _itera_py = utils._EngineIntScalarType.to_py(_itera)
        _norm_py = utils._EngineFloat64ScalarType.to_py(_norm)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['itera', 'norm'],
        )(
            _itera_py,
            _norm_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def corrmat_nearest_rank(g, n, rank, errtol, ranktol, maxits, maxit, x):
    u"""
    Computes the rank-constrained nearest correlation matrix to a real
    square matrix, using the method of Qi and Sun.

    ``corrmat_nearest_rank`` computes the nearest correlation matrix of
    maximum prescribed rank, in the Frobenius norm, to a given square,
    input matrix.

    For full information please refer to the NAG Library document for
    g02ak

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02akf.html

    Parameters
    ----------
    g : float, shape (`n`, `n`), modified in place
        `On entry`: G~, the initial matrix.

        `On exit`: a symmetric matrix G = 1/2(G~+G~^T) with diagonal
        elements set to 1.0.

    n : int
        The order of the matrix G.

    rank : int
        r, the upper bound for the rank of X.

    errtol : float
        The termination tolerance for the convergence measure of the
        objective function value.

        If `errtol` <= 0.0, then a value of 1.0e-5 is used.

        See [Algorithmic Details] for further details.

    ranktol : float
        The feasibility tolerance for the rank constraint.

        If `ranktol` <= 0.0, sqrt(machine precision) is used.

        See [Algorithmic Details] for further details.

    maxits : int
        Specifies the maximum number of iterations used for the
        majorization approach to solve penalized problems at each level
        of penalty parameter.

        If `maxits` <= 0, then a value of 100 is used.

    maxit : int
        Specifies the maximum number of iterations for the penalty
        method, i.e., the maximum level of penalty parameter.

        If `maxit` <= 0, then a value of 200 is used.

    x : float, shape (`n`, `n`), modified in place
        `On exit`: X, the nearest correlation matrix of rank r.

    Returns
    -------
    f : float
        The difference between X and G given by 1/2||X-G||_F^2.

    rankerr : float
        The rank error of X, defined as sum_i = r+1^n(lambda_i), given
        that lambda_i denote eigenvalues of X sorted in non-increasing
        order.

    nsub : int
        The total number of majorized problems that have been solved,
        i.e., the total number of calls for ``corrmat_nearest``.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` > 0.

        (`errno` 4)
            On entry, `rank` = *<value>* and `n` = *<value>*.

            Constraint: 0 < `rank` <= `n`.

        (`errno` 5)
            Majorized penalty approach fails to converge in `maxit`
            level of penalty iterations.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 6)
            Convergence is limited by machine precision. The objective
            function value or rank is decreasing very slowly.

            The array returned in `x` may still be of interest.

    References
    ----------
    Bai, S, Qi, H--D and Xiu, N, 2015, `Constrained best Euclidean
    distance embedding on a sphere: A matrix optimization approach`,
    SIAM J. Optim. (25(1)), 439--467

    Gao, Y and Sun, D, 2010, `A majorized penalty approach for
    calibrating rank constrained correlation matrix problems`, Technical
    report, Department of Mathematics, National University of Singapore

    Qi, H--D and Yuan, X, 2014, `Computing the nearest Euclidean
    distance matrix with low embedding dimensions`, Mathematical
    Programming (147(1--2)), 351--389
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.corrmat_nearest_rank'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02akft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._EngineFloat64ArrayType.c_type, # Input/Output, g[ldg, n].
        utils._BYREF_INT, # Input, ldg.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, rank.
        utils._BYREF_FLOAT64, # Input, errtol.
        utils._BYREF_FLOAT64, # Input, ranktol.
        utils._BYREF_INT, # Input, maxits.
        utils._BYREF_INT, # Input, maxit.
        utils._EngineFloat64ArrayType.c_type, # Output, x[ldx, n].
        utils._BYREF_INT, # Input, ldx.
        utils._BYREF_FLOAT64, # Output, f.
        utils._BYREF_FLOAT64, # Output, rankerr.
        utils._BYREF_INT, # Output, nsub.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((g, 'g'), (x, 'x')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _ldg_py = utils._EngineFloat64ArrayType.get_nd(g, {'fun_name': fun_name, 'entity': 'g'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldg = utils._EngineIntScalarType(
        _ldg_py, {'fun_name': fun_name, 'entity': '_ldg'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _g_cls = utils._EngineFloat64ArrayType(
        g, {'fun_name': fun_name, 'entity': 'g'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
    )
    _g = _g_cls.to_ctypes()
    _rank = utils._EngineIntScalarType(
        rank, {'fun_name': fun_name, 'entity': 'rank'},
    ).to_ctypes()
    _errtol = utils._EngineFloat64ScalarType(
        errtol, {'fun_name': fun_name, 'entity': 'errtol'},
    ).to_ctypes()
    _ranktol = utils._EngineFloat64ScalarType(
        ranktol, {'fun_name': fun_name, 'entity': 'ranktol'},
    ).to_ctypes()
    _maxits = utils._EngineIntScalarType(
        maxits, {'fun_name': fun_name, 'entity': 'maxits'},
    ).to_ctypes()
    _maxit = utils._EngineIntScalarType(
        maxit, {'fun_name': fun_name, 'entity': 'maxit'},
    ).to_ctypes()
    _ldx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _f = utils._EngineFloat64ScalarType.empty_ctypes()
    _rankerr = utils._EngineFloat64ScalarType.empty_ctypes()
    _nsub = utils._EngineIntScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _g,
        _ldg,
        _n,
        _rank,
        _errtol,
        _ranktol,
        _maxits,
        _maxit,
        _x,
        _ldx,
        _f,
        _rankerr,
        _nsub,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1001:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4001:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, rank = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: 0 < rank <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 5001:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Majorized penalty approach fails to converge in maxit level of penalty',
                    'iterations.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Convergence is limited by machine precision. The objective function',
                    'value or rank is decreasing very slowly.',
                    'The array returned in x may still be of interest.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _g_cls.refresh_py(_g)
        _x_cls.refresh_py(_x)
        _f_py = utils._EngineFloat64ScalarType.to_py(_f)
        _rankerr_py = utils._EngineFloat64ScalarType.to_py(_rankerr)
        _nsub_py = utils._EngineIntScalarType.to_py(_nsub)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['f', 'rankerr', 'nsub'],
        )(
            _f_py,
            _rankerr_py,
            _nsub_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def corrmat_shrinking(g, n, k, errtol, eigtol, x):
    u"""
    Computes a correlation matrix from an approximate matrix with fixed
    submatrix.

    ``corrmat_shrinking`` computes a correlation matrix, subject to
    preserving a leading principal submatrix and applying the smallest
    relative perturbation to the remainder of the approximate input
    matrix.

    For full information please refer to the NAG Library document for
    g02an

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02anf.html

    Parameters
    ----------
    g : float, shape (`n`, `n`), modified in place
        `On entry`: G, the initial matrix.

        `On exit`: a symmetric matrix 1/2(G+G^T) with the diagonal set
        to I.

    n : int
        The order of the matrix G.

    k : int
        k, the order of the leading principal submatrix A.

    errtol : float
        The termination tolerance for the iteration.

        If `errtol` <= 0.0, sqrt(machine precision) is used.

        See [Accuracy] for further details.

    eigtol : float
        The tolerance used in determining the definiteness of A.

        If lambda_min(A) > `n`*λ_max(A)*`eigtol`, where lambda_min(A)
        and lambda_max(A) denote the minimum and maximum eigenvalues of
        A respectively, A is positive definite.

        If `eigtol` <= 0, machine precision is used.

    x : float, shape (`n`, `n`), modified in place
        `On exit`: contains the matrix X.

    Returns
    -------
    alpha : float
        alpha.

    itera : int
        The number of iterations taken.

    eigmin : float
        The smallest eigenvalue of the leading principal submatrix A.

    norm : float
        The value of ||G-X||_F after the final iteration.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` > 0.

        (`errno` 3)
            On entry, `k` = *<value>* and `n` = *<value>*.

            Constraint: `n` >= `k` > 0.

        (`errno` 5)
            The k by k principal leading submatrix of the initial matrix
            G is not positive definite.

        (`errno` 6)
            Failure to solve intermediate eigenproblem. This should not
            occur. Please contact NAG.

    References
    ----------
    Higham, N J, Strabić, N and Šego, V, 2014, `Restoring definiteness
    via shrinking, with an application to correlation matrices with a
    fixed block`, MIMS EPrint 2014.54, Manchester Institute for
    Mathematical Sciences, The University of Manchester, UK
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.corrmat_shrinking'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02anft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._EngineFloat64ArrayType.c_type, # Input/Output, g[ldg, n].
        utils._BYREF_INT, # Input, ldg.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, k.
        utils._BYREF_FLOAT64, # Input, errtol.
        utils._BYREF_FLOAT64, # Input, eigtol.
        utils._EngineFloat64ArrayType.c_type, # Output, x[ldx, n].
        utils._BYREF_INT, # Input, ldx.
        utils._BYREF_FLOAT64, # Output, alpha.
        utils._BYREF_INT, # Output, itera.
        utils._BYREF_FLOAT64, # Output, eigmin.
        utils._BYREF_FLOAT64, # Output, norm.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((g, 'g'), (x, 'x')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _ldg_py = utils._EngineFloat64ArrayType.get_nd(g, {'fun_name': fun_name, 'entity': 'g'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldg = utils._EngineIntScalarType(
        _ldg_py, {'fun_name': fun_name, 'entity': '_ldg'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _g_cls = utils._EngineFloat64ArrayType(
        g, {'fun_name': fun_name, 'entity': 'g'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
    )
    _g = _g_cls.to_ctypes()
    _k = utils._EngineIntScalarType(
        k, {'fun_name': fun_name, 'entity': 'k'},
    ).to_ctypes()
    _errtol = utils._EngineFloat64ScalarType(
        errtol, {'fun_name': fun_name, 'entity': 'errtol'},
    ).to_ctypes()
    _eigtol = utils._EngineFloat64ScalarType(
        eigtol, {'fun_name': fun_name, 'entity': 'eigtol'},
    ).to_ctypes()
    _ldx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _alpha = utils._EngineFloat64ScalarType.empty_ctypes()
    _itera = utils._EngineIntScalarType.empty_ctypes()
    _eigmin = utils._EngineFloat64ScalarType.empty_ctypes()
    _norm = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _g,
        _ldg,
        _n,
        _k,
        _errtol,
        _eigtol,
        _x,
        _ldx,
        _alpha,
        _itera,
        _eigmin,
        _norm,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1001:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 1003:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, k = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: n >= k > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 1005:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The k by k principal leading',
                    'submatrix of the initial matrix G is not positive definite.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 1006:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Failure to solve intermediate eigenproblem. This should not occur.',
                    'Please contact NAG.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _g_cls.refresh_py(_g)
        _x_cls.refresh_py(_x)
        _alpha_py = utils._EngineFloat64ScalarType.to_py(_alpha)
        _itera_py = utils._EngineIntScalarType.to_py(_itera)
        _eigmin_py = utils._EngineFloat64ScalarType.to_py(_eigmin)
        _norm_py = utils._EngineFloat64ScalarType.to_py(_norm)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['alpha', 'itera', 'eigmin', 'norm'],
        )(
            _alpha_py,
            _itera_py,
            _eigmin_py,
            _norm_py,
        )
    _err_class.manage()
    return return_data

def corrmat_target(g, n, theta, h, errtol, eigtol, x):
    u"""
    Computes a correlation matrix from an approximate one using a
    specified target matrix.

    ``corrmat_target`` computes a correlation matrix, by using a
    positive definite **target** matrix derived from weighting the
    approximate input matrix, with an optional bound on the minimum
    eigenvalue.

    For full information please refer to the NAG Library document for
    g02ap

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02apf.html

    Parameters
    ----------
    g : float, shape (`n`, `n`), modified in place
        `On entry`: G, the initial matrix.

        `On exit`: a symmetric matrix 1/2(G+G^T) with the diagonal
        elements set to 1.0.

    n : int
        The order of the matrix G.

    theta : float
        The value of theta. If `theta` < 0.0, 0.0 is used.

    h : float, shape (`n`, `n`), modified in place
        `On entry`: the matrix of weights H.

        `On exit`: a symmetric matrix 1/2(H+H^T) with its diagonal
        elements set to 1.0.

    errtol : float
        The termination tolerance for the iteration.

        If `errtol` <= 0.0, sqrt(machine precision) is used.

        See [Accuracy] for further details.

    eigtol : float
        The tolerance used in determining the definiteness of the target
        matrix T = H.G.

        If lambda_min(T) > `n`*λ_max(T)*`eigtol`, where lambda_min(T)
        and lambda_max(T) denote the minimum and maximum eigenvalues of
        T respectively, T is positive definite.

        If `eigtol` <= 0, machine precision is used.

    x : float, shape (`n`, `n`), modified in place
        `On exit`: contains the matrix X.

    Returns
    -------
    alpha : float
        The constant alpha used in the formation of X.

    itera : int
        The number of iterations taken.

    eigmin : float
        The smallest eigenvalue of the target matrix T.

    norm : float
        The value of ||G-X||_F after the final iteration.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` > 0.

        (`errno` 3)
            On entry, `theta` = *<value>*.

            Constraint: `theta` < 1.0.

        (`errno` 6)
            The target matrix is not positive definite.

        (`errno` 7)
            Failure to solve intermediate eigenproblem. This should not
            occur. Please contact NAG.

    References
    ----------
    Higham, N J, Strabić, N and Šego, V, 2014, `Restoring definiteness
    via shrinking, with an application to correlation matrices with a
    fixed block`, MIMS EPrint 2014.54, Manchester Institute for
    Mathematical Sciences, The University of Manchester, UK
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.corrmat_target'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02apft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._EngineFloat64ArrayType.c_type, # Input/Output, g[ldg, n].
        utils._BYREF_INT, # Input, ldg.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_FLOAT64, # Input, theta.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, h[ldh, n].
        utils._BYREF_INT, # Input, ldh.
        utils._BYREF_FLOAT64, # Input, errtol.
        utils._BYREF_FLOAT64, # Input, eigtol.
        utils._EngineFloat64ArrayType.c_type, # Output, x[ldx, n].
        utils._BYREF_INT, # Input, ldx.
        utils._BYREF_FLOAT64, # Output, alpha.
        utils._BYREF_INT, # Output, itera.
        utils._BYREF_FLOAT64, # Output, eigmin.
        utils._BYREF_FLOAT64, # Output, norm.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((g, 'g'), (h, 'h'), (x, 'x')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _ldg_py = utils._EngineFloat64ArrayType.get_nd(g, {'fun_name': fun_name, 'entity': 'g'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldg = utils._EngineIntScalarType(
        _ldg_py, {'fun_name': fun_name, 'entity': '_ldg'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _g_cls = utils._EngineFloat64ArrayType(
        g, {'fun_name': fun_name, 'entity': 'g'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
    )
    _g = _g_cls.to_ctypes()
    _theta = utils._EngineFloat64ScalarType(
        theta, {'fun_name': fun_name, 'entity': 'theta'},
    ).to_ctypes()
    _ldh_py = utils._EngineFloat64ArrayType.get_nd(h, {'fun_name': fun_name, 'entity': 'h'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldh = utils._EngineIntScalarType(
        _ldh_py, {'fun_name': fun_name, 'entity': '_ldh'},
    ).to_ctypes()
    _h_cls = utils._EngineFloat64ArrayType(
        h, {'fun_name': fun_name, 'entity': 'h'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
    )
    _h = _h_cls.to_ctypes()
    _errtol = utils._EngineFloat64ScalarType(
        errtol, {'fun_name': fun_name, 'entity': 'errtol'},
    ).to_ctypes()
    _eigtol = utils._EngineFloat64ScalarType(
        eigtol, {'fun_name': fun_name, 'entity': 'eigtol'},
    ).to_ctypes()
    _ldx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _alpha = utils._EngineFloat64ScalarType.empty_ctypes()
    _itera = utils._EngineIntScalarType.empty_ctypes()
    _eigmin = utils._EngineFloat64ScalarType.empty_ctypes()
    _norm = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _g,
        _ldg,
        _n,
        _theta,
        _h,
        _ldh,
        _errtol,
        _eigtol,
        _x,
        _ldx,
        _alpha,
        _itera,
        _eigmin,
        _norm,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1001:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 1003:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, theta = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: theta < 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 1006:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The target matrix is not positive definite.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 1007:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Failure to solve intermediate eigenproblem. This should not occur.',
                    'Please contact NAG.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _g_cls.refresh_py(_g)
        _h_cls.refresh_py(_h)
        _x_cls.refresh_py(_x)
        _alpha_py = utils._EngineFloat64ScalarType.to_py(_alpha)
        _itera_py = utils._EngineIntScalarType.to_py(_itera)
        _eigmin_py = utils._EngineFloat64ScalarType.to_py(_eigmin)
        _norm_py = utils._EngineFloat64ScalarType.to_py(_norm)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['alpha', 'itera', 'eigmin', 'norm'],
        )(
            _alpha_py,
            _itera_py,
            _eigmin_py,
            _norm_py,
        )
    _err_class.manage()
    return return_data

def corrmat_fixed(g, n, alpha, h, errtol, maxit, m, x):
    u"""
    Computes the nearest correlation matrix to a real square matrix,
    with fixed elements.

    ``corrmat_fixed`` computes the nearest correlation matrix, in the
    Frobenius norm, while fixing elements and optionally with bounds on
    the eigenvalues, to a given square input matrix.

    For full information please refer to the NAG Library document for
    g02as

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02asf.html

    Parameters
    ----------
    g : float, shape (`n`, `n`), modified in place
        `On entry`: G~, the initial matrix.

        `On exit`: the symmetric matrix G = 1/2(G~+G~^T) with the
        diagonal elements set to 1.0.

    n : int
        The order of the matrix G.

    alpha : float
        The value of alpha.

        If `alpha` < 0.0, a value of 0.0 is used.

    h : int, shape (`n`, `n`)
        The symmetric matrix H. If an element of H is 1 then the
        corresponding element in G is fixed in the output X. Only the
        strictly lower triangular part of H need be set.

    errtol : float
        The termination tolerance for the iteration.

        If `errtol` <= 0.0, sqrt(machine precision) is used.

        See [Accuracy] for further details.

    maxit : int
        Specifies the maximum number of iterations.

        If `maxit` <= 0, a value of 200 is used.

    m : int
        The number of previous iterates to use in the Anderson
        acceleration. If `m` = 0, Anderson acceleration is not used. See
        [Accuracy] for further details.

        If `m` < 0, a value of 4 is used.

    x : float, shape (`n`, `n`), modified in place
        `On exit`: contains the matrix X.

    Returns
    -------
    its : int
        The number of iterations taken.

    fnorm : float
        The value of ||G-X||_F after the final iteration.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` > 0.

        (`errno` 4)
            On entry, `alpha` = *<value>*.

            Constraint: `alpha` < 1.0.

        (`errno` 5)
            On entry, `m` = *<value>* and `n` = *<value>*.

            Constraint: `m` <= `n`*`n`.

        (`errno` 7)
            Function failed to converge in *<value>* iterations.

            A solution may not exist, however, try increasing `maxit`.

        (`errno` 8)
            Failure during Anderson acceleration.

            Consider setting `m` = 0 and recomputing.

        (`errno` 9)
            The fixed element G_{{ij}}, lies outside the interval
            [-1,1], for i = *<value>* and j = *<value>*.

    References
    ----------
    Anderson, D G, 1965, `Iterative Procedures for Nonlinear Integral
    Equations`, J. Assoc. Comput. Mach. (12), 547--560

    Higham, N J and Strabić, N, 2016, `Anderson acceleration of the
    alternating projections method for computing the nearest correlation
    matrix`, Numer. Algor. (72), 1021--1042
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.corrmat_fixed'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02asft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._EngineFloat64ArrayType.c_type, # Input/Output, g[ldg, n].
        utils._BYREF_INT, # Input, ldg.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_FLOAT64, # Input, alpha.
        utils._EngineIntArrayType.c_type, # Input, h[ldh, n].
        utils._BYREF_INT, # Input, ldh.
        utils._BYREF_FLOAT64, # Input, errtol.
        utils._BYREF_INT, # Input, maxit.
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Output, x[ldx, n].
        utils._BYREF_INT, # Input, ldx.
        utils._BYREF_INT, # Output, its.
        utils._BYREF_FLOAT64, # Output, fnorm.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((g, 'g'), (h, 'h'), (x, 'x')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _ldg_py = utils._EngineFloat64ArrayType.get_nd(g, {'fun_name': fun_name, 'entity': 'g'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldg = utils._EngineIntScalarType(
        _ldg_py, {'fun_name': fun_name, 'entity': '_ldg'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _g_cls = utils._EngineFloat64ArrayType(
        g, {'fun_name': fun_name, 'entity': 'g'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
    )
    _g = _g_cls.to_ctypes()
    _alpha = utils._EngineFloat64ScalarType(
        alpha, {'fun_name': fun_name, 'entity': 'alpha'},
    ).to_ctypes()
    _ldh_py = utils._EngineIntArrayType.get_nd(h, {'fun_name': fun_name, 'entity': 'h'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldh = utils._EngineIntScalarType(
        _ldh_py, {'fun_name': fun_name, 'entity': '_ldh'},
    ).to_ctypes()
    _h = utils._EngineIntArrayType(
        h, {'fun_name': fun_name, 'entity': 'h'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _errtol = utils._EngineFloat64ScalarType(
        errtol, {'fun_name': fun_name, 'entity': 'errtol'},
    ).to_ctypes()
    _maxit = utils._EngineIntScalarType(
        maxit, {'fun_name': fun_name, 'entity': 'maxit'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _ldx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _its = utils._EngineIntScalarType.empty_ctypes()
    _fnorm = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _g,
        _ldg,
        _n,
        _alpha,
        _h,
        _ldh,
        _errtol,
        _maxit,
        _m,
        _x,
        _ldx,
        _its,
        _fnorm,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1001:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 1004:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, alpha = {:s}.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: alpha < 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 1005:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: m <= n*n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 1007:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Function failed to converge in {:s} iterations.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'A solution may not exist, however, try increasing maxit.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 1008:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Failure during Anderson acceleration.',
                    'Consider setting m = 0 and recomputing.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 9 and _ifmt == 1010:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The fixed element G_{{ij}}, lies outside the interval [-1,1],',
                    'for i = {:s} and j = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _g_cls.refresh_py(_g)
        _x_cls.refresh_py(_x)
        _its_py = utils._EngineIntScalarType.to_py(_its)
        _fnorm_py = utils._EngineFloat64ScalarType.to_py(_fnorm)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['its', 'fnorm'],
        )(
            _its_py,
            _fnorm_py,
        )
    _err_class.manage()
    return return_data

def coeffs_pearson(n, m, x, xbar, std, ssp, r):
    u"""
    Pearson product-moment correlation coefficients, all variables, no
    missing values.

    ``coeffs_pearson`` computes means and standard deviations of
    variables, sums of squares and cross-products of deviations from
    means, and Pearson product-moment correlation coefficients for a set
    of data.

    For full information please refer to the NAG Library document for
    g02ba

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02baf.html

    Parameters
    ----------
    n : int
        n, the number of observations or cases.

    m : int
        m, the number of variables.

    x : float, shape (`n`, `m`)
        `x`[i-1,j-1] must be set to x_{ij}, the ith observation on the
        jth variable, for j = 1,2,...,m, for i = 1,2,...,n.

    xbar : float, shape (`m`), modified in place
        `On exit`: the mean value, x__j, of the jth variable, for j =
        1,2,...,m.

    std : float, shape (`m`), modified in place
        `On exit`: the standard deviation, s_j, of the jth variable, for
        j = 1,2,...,m.

    ssp : float, shape (`m`, `m`), modified in place
        `On exit`: `ssp`[j-1,k-1] is the cross-product of deviations
        S_{jk}, for k = 1,2,...,m, for j = 1,2,...,m.

    r : float, shape (`m`, `m`), modified in place
        `On exit`: `r`[j-1,k-1] is the product-moment correlation
        coefficient R_{jk} between the jth and kth variables, for k =
        1,2,...,m, for j = 1,2,...,m.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 2.

        (`errno` 2)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 2.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.coeffs_pearson'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02baft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Input, x[ldx, m].
        utils._BYREF_INT, # Input, ldx.
        utils._EngineFloat64ArrayType.c_type, # Output, xbar[m].
        utils._EngineFloat64ArrayType.c_type, # Output, std[m].
        utils._EngineFloat64ArrayType.c_type, # Output, ssp[ldssp, m].
        utils._BYREF_INT, # Input, ldssp.
        utils._EngineFloat64ArrayType.c_type, # Output, r[ldr, m].
        utils._BYREF_INT, # Input, ldr.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((x, 'x'), (ssp, 'ssp'), (r, 'r')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _ldx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py, _m_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _xbar_cls = utils._EngineFloat64ArrayType(
        xbar, {'fun_name': fun_name, 'entity': 'xbar'},
        exp_shape=(_m_py,),
        check_contents=False,
    )
    _xbar = _xbar_cls.to_ctypes(can_reference_contents=False)
    _std_cls = utils._EngineFloat64ArrayType(
        std, {'fun_name': fun_name, 'entity': 'std'},
        exp_shape=(_m_py,),
        check_contents=False,
    )
    _std = _std_cls.to_ctypes(can_reference_contents=False)
    _ldssp_py = utils._EngineFloat64ArrayType.get_nd(ssp, {'fun_name': fun_name, 'entity': 'ssp'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldssp = utils._EngineIntScalarType(
        _ldssp_py, {'fun_name': fun_name, 'entity': '_ldssp'},
    ).to_ctypes()
    _ssp_cls = utils._EngineFloat64ArrayType(
        ssp, {'fun_name': fun_name, 'entity': 'ssp'},
        exp_shape=(_m_py, _m_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _ssp = _ssp_cls.to_ctypes(can_reference_contents=False)
    _ldr_py = utils._EngineFloat64ArrayType.get_nd(r, {'fun_name': fun_name, 'entity': 'r'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldr = utils._EngineIntScalarType(
        _ldr_py, {'fun_name': fun_name, 'entity': '_ldr'},
    ).to_ctypes()
    _r_cls = utils._EngineFloat64ArrayType(
        r, {'fun_name': fun_name, 'entity': 'r'},
        exp_shape=(_m_py, _m_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _r = _r_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _m,
        _x,
        _ldx,
        _xbar,
        _std,
        _ssp,
        _ldssp,
        _r,
        _ldr,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _xbar_cls.refresh_py(_xbar)
        _std_cls.refresh_py(_std)
        _ssp_cls.refresh_py(_ssp)
        _r_cls.refresh_py(_r)
        return_data = None
    _err_class.manage()
    return return_data

def coeffs_pearson_miss_case(n, m, x, miss, xmiss, xbar, std, ssp, r):
    u"""
    Pearson product-moment correlation coefficients, all variables,
    casewise treatment of missing values.

    ``coeffs_pearson_miss_case`` computes means and standard deviations
    of variables, sums of squares and cross-products of deviations from
    means, and Pearson product-moment correlation coefficients for a set
    of data omitting completely any cases with a missing observation for
    any variable.

    For full information please refer to the NAG Library document for
    g02bb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02bbf.html

    Parameters
    ----------
    n : int
        n, the number of observations or cases.

    m : int
        m, the number of variables.

    x : float, shape (`n`, `m`)
        `x`[i-1,j-1] must be set to x_{ij}, the value of the ith
        observation on the jth variable, for j = 1,2,...,m, for i =
        1,2,...,n.

    miss : int, shape (`m`), modified in place
        `On entry`: `miss`[j-1] must be set equal to 1 if a missing
        value, xm_j, is to be specified for the jth variable in the
        array `x`, or set equal to 0 otherwise. Values of `miss` must be
        given for all m variables in the array `x`.

        `On exit`: the array `miss` is overwritten by the function, and
        the information it contained on entry is lost.

    xmiss : float, shape (`m`), modified in place
        `On entry`: `xmiss`[j-1] must be set to the missing value, xm_j,
        to be associated with the jth variable in the array `x`, for
        those variables for which missing values are specified by means
        of the array `miss` (see [Accuracy]).

        `On exit`: the array `xmiss` is overwritten by the function, and
        the information it contained on entry is lost.

    xbar : float, shape (`m`), modified in place
        `On exit`: the mean value, x__j, of the jth variable, for j =
        1,2,...,m.

    std : float, shape (`m`), modified in place
        `On exit`: the standard deviation, s_j, of the jth variable, for
        j = 1,2,...,m.

    ssp : float, shape (`m`, `m`), modified in place
        `On exit`: `ssp`[j-1,k-1] is the cross-product of deviations
        S_{jk}, for k = 1,2,...,m, for j = 1,2,...,m.

    r : float, shape (`m`, `m`), modified in place
        `On exit`: `r`[j-1,k-1] is the product-moment correlation
        coefficient R_{jk} between the jth and kth variables, for k =
        1,2,...,m, for j = 1,2,...,m.

    Returns
    -------
    ncases : int
        The number of cases actually used in the calculations (when
        cases involving missing values have been eliminated).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 2.

        (`errno` 2)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 2.

        (`errno` 4)
            After observations with missing values were omitted, no
            cases remained.

        (`errno` 5)
            After observations with missing values were omitted, only
            one case remained.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.coeffs_pearson_miss_case'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02bbft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Input, x[ldx, m].
        utils._BYREF_INT, # Input, ldx.
        utils._EngineIntArrayType.c_type, # Input/Output, miss[m].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, xmiss[m].
        utils._EngineFloat64ArrayType.c_type, # Output, xbar[m].
        utils._EngineFloat64ArrayType.c_type, # Output, std[m].
        utils._EngineFloat64ArrayType.c_type, # Output, ssp[ldssp, m].
        utils._BYREF_INT, # Input, ldssp.
        utils._EngineFloat64ArrayType.c_type, # Output, r[ldr, m].
        utils._BYREF_INT, # Input, ldr.
        utils._BYREF_INT, # Output, ncases.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((x, 'x'), (ssp, 'ssp'), (r, 'r')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _ldx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py, _m_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _miss_cls = utils._EngineIntArrayType(
        miss, {'fun_name': fun_name, 'entity': 'miss'},
        exp_shape=(_m_py,),
    )
    _miss = _miss_cls.to_ctypes()
    _xmiss_cls = utils._EngineFloat64ArrayType(
        xmiss, {'fun_name': fun_name, 'entity': 'xmiss'},
        exp_shape=(_m_py,),
    )
    _xmiss = _xmiss_cls.to_ctypes()
    _xbar_cls = utils._EngineFloat64ArrayType(
        xbar, {'fun_name': fun_name, 'entity': 'xbar'},
        exp_shape=(_m_py,),
        check_contents=False,
    )
    _xbar = _xbar_cls.to_ctypes(can_reference_contents=False)
    _std_cls = utils._EngineFloat64ArrayType(
        std, {'fun_name': fun_name, 'entity': 'std'},
        exp_shape=(_m_py,),
        check_contents=False,
    )
    _std = _std_cls.to_ctypes(can_reference_contents=False)
    _ldssp_py = utils._EngineFloat64ArrayType.get_nd(ssp, {'fun_name': fun_name, 'entity': 'ssp'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldssp = utils._EngineIntScalarType(
        _ldssp_py, {'fun_name': fun_name, 'entity': '_ldssp'},
    ).to_ctypes()
    _ssp_cls = utils._EngineFloat64ArrayType(
        ssp, {'fun_name': fun_name, 'entity': 'ssp'},
        exp_shape=(_m_py, _m_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _ssp = _ssp_cls.to_ctypes(can_reference_contents=False)
    _ldr_py = utils._EngineFloat64ArrayType.get_nd(r, {'fun_name': fun_name, 'entity': 'r'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldr = utils._EngineIntScalarType(
        _ldr_py, {'fun_name': fun_name, 'entity': '_ldr'},
    ).to_ctypes()
    _r_cls = utils._EngineFloat64ArrayType(
        r, {'fun_name': fun_name, 'entity': 'r'},
        exp_shape=(_m_py, _m_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _r = _r_cls.to_ctypes(can_reference_contents=False)
    _ncases = utils._EngineIntScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _m,
        _x,
        _ldx,
        _miss,
        _xmiss,
        _xbar,
        _std,
        _ssp,
        _ldssp,
        _r,
        _ldr,
        _ncases,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'After observations with missing values were omitted,',
                    'no cases remained.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'After observations with missing values were omitted,',
                    'only one case remained.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _miss_cls.refresh_py(_miss)
        _xmiss_cls.refresh_py(_xmiss)
        _xbar_cls.refresh_py(_xbar)
        _std_cls.refresh_py(_std)
        _ssp_cls.refresh_py(_ssp)
        _r_cls.refresh_py(_r)
        _ncases_py = utils._EngineIntScalarType.to_py(_ncases)
        return_data = _ncases_py
    _err_class.manage()
    return return_data

def coeffs_pearson_miss_pair(n, m, x, miss, xmiss, xbar, std, ssp, r, cnt):
    u"""
    Pearson product-moment correlation coefficients, all variables,
    pairwise treatment of missing values.

    ``coeffs_pearson_miss_pair`` computes means and standard deviations
    of variables, sums of squares and cross-products of deviations from
    means, and Pearson product-moment correlation coefficients for a set
    of data omitting cases with missing values from only those
    calculations involving the variables for which the values are
    missing.

    For full information please refer to the NAG Library document for
    g02bc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02bcf.html

    Parameters
    ----------
    n : int
        n, the number of observations or cases.

    m : int
        m, the number of variables.

    x : float, shape (`n`, `m`)
        `x`[i-1,j-1] must be set to x_{ij}, the value of the ith
        observation on the jth variable, for j = 1,2,...,m, for i =
        1,2,...,n.

    miss : int, shape (`m`)
        `miss`[j-1] must be set equal to 1 if a missing value, xm_j, is
        to be specified for the jth variable in the array `x`, or set
        equal to 0 otherwise. Values of `miss` must be given for all m
        variables in the array `x`.

    xmiss : float, shape (`m`)
        `xmiss`[j-1] must be set to the missing value, xm_j, to be
        associated with the jth variable in the array `x`, for those
        variables for which missing values are specified by means of the
        array `miss` (see [Accuracy]).

    xbar : float, shape (`m`), modified in place
        `On exit`: the mean value, x__j, of the jth variable, for j =
        1,2,...,m.

    std : float, shape (`m`), modified in place
        `On exit`: the standard deviation, s_j, of the jth variable, for
        j = 1,2,...,m.

    ssp : float, shape (`m`, `m`), modified in place
        `On exit`: `ssp`[j-1,k-1] is the cross-product of deviations
        S_{jk}, for k = 1,2,...,m, for j = 1,2,...,m.

    r : float, shape (`m`, `m`), modified in place
        `On exit`: `r`[j-1,k-1] is the product-moment correlation
        coefficient R_{jk} between the jth and kth variables, for k =
        1,2,...,m, for j = 1,2,...,m.

    cnt : float, shape (`m`, `m`), modified in place
        `On exit`: `cnt`[j-1,k-1] is the number of cases, c_{jk},
        actually used in the calculation of S_{jk}, and R_{jk}, the sum
        of cross-products and correlation coefficient for the jth and
        kth variables, for k = 1,2,...,m, for j = 1,2,...,m.

    Returns
    -------
    ncases : int
        The minimum number of cases used in the calculation of any of
        the sums of squares and cross-products and correlation
        coefficients (when cases involving missing values have been
        eliminated).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 2.

        (`errno` 2)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 2.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 4)
            After observations with missing values were omitted, fewer
            than two cases remained for at least one pair of variables.
            (The pairs of variables involved can be determined by
            examination of the contents of the array `cnt`). All means,
            standard deviations, sums of squares and cross-products, and
            correlation coefficients based on two or more cases are
            returned by the function even if `errno` = 4.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.coeffs_pearson_miss_pair'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02bcft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, sorder.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Input, x[ldx, *].
        utils._BYREF_INT, # Input, ldx.
        utils._EngineIntArrayType.c_type, # Input, miss[m].
        utils._EngineFloat64ArrayType.c_type, # Input, xmiss[m].
        utils._EngineFloat64ArrayType.c_type, # Output, xbar[m].
        utils._EngineFloat64ArrayType.c_type, # Output, std[m].
        utils._EngineFloat64ArrayType.c_type, # Output, ssp[ldssp, m].
        utils._BYREF_INT, # Input, ldssp.
        utils._EngineFloat64ArrayType.c_type, # Output, r[ldr, m].
        utils._BYREF_INT, # Input, ldr.
        utils._BYREF_INT, # Output, ncases.
        utils._EngineFloat64ArrayType.c_type, # Output, cnt[ldcnt, m].
        utils._BYREF_INT, # Input, ldcnt.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((x, 'x'), (ssp, 'ssp'), (r, 'r'), (cnt, 'cnt')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _sorder = utils._EngineIntScalarType.c_type(_nag_enum_consts.nag_col_major)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _ldx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _sorder_py = utils._EngineIntScalarType.to_py(_sorder)
    _m_py = utils._EngineIntScalarType.to_py(_m)
    if _sorder_py == 1:
        _x_dim2 = _m_py
    else:
        _x_dim2 = _n_py
    exp_shape_x = (_n_py, _x_dim2)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=exp_shape_x,
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _miss = utils._EngineIntArrayType(
        miss, {'fun_name': fun_name, 'entity': 'miss'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _xmiss = utils._EngineFloat64ArrayType(
        xmiss, {'fun_name': fun_name, 'entity': 'xmiss'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _xbar_cls = utils._EngineFloat64ArrayType(
        xbar, {'fun_name': fun_name, 'entity': 'xbar'},
        exp_shape=(_m_py,),
        check_contents=False,
    )
    _xbar = _xbar_cls.to_ctypes(can_reference_contents=False)
    _std_cls = utils._EngineFloat64ArrayType(
        std, {'fun_name': fun_name, 'entity': 'std'},
        exp_shape=(_m_py,),
        check_contents=False,
    )
    _std = _std_cls.to_ctypes(can_reference_contents=False)
    _ldssp_py = utils._EngineFloat64ArrayType.get_nd(ssp, {'fun_name': fun_name, 'entity': 'ssp'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldssp = utils._EngineIntScalarType(
        _ldssp_py, {'fun_name': fun_name, 'entity': '_ldssp'},
    ).to_ctypes()
    _ssp_cls = utils._EngineFloat64ArrayType(
        ssp, {'fun_name': fun_name, 'entity': 'ssp'},
        exp_shape=(_m_py, _m_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _ssp = _ssp_cls.to_ctypes(can_reference_contents=False)
    _ldr_py = utils._EngineFloat64ArrayType.get_nd(r, {'fun_name': fun_name, 'entity': 'r'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldr = utils._EngineIntScalarType(
        _ldr_py, {'fun_name': fun_name, 'entity': '_ldr'},
    ).to_ctypes()
    _r_cls = utils._EngineFloat64ArrayType(
        r, {'fun_name': fun_name, 'entity': 'r'},
        exp_shape=(_m_py, _m_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _r = _r_cls.to_ctypes(can_reference_contents=False)
    _ncases = utils._EngineIntScalarType.empty_ctypes()
    _ldcnt_py = utils._EngineFloat64ArrayType.get_nd(cnt, {'fun_name': fun_name, 'entity': 'cnt'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldcnt = utils._EngineIntScalarType(
        _ldcnt_py, {'fun_name': fun_name, 'entity': '_ldcnt'},
    ).to_ctypes()
    _cnt_cls = utils._EngineFloat64ArrayType(
        cnt, {'fun_name': fun_name, 'entity': 'cnt'},
        exp_shape=(_m_py, _m_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _cnt = _cnt_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _sorder,
        _n,
        _m,
        _x,
        _ldx,
        _miss,
        _xmiss,
        _xbar,
        _std,
        _ssp,
        _ldssp,
        _r,
        _ldr,
        _ncases,
        _cnt,
        _ldcnt,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99993:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'After observations with missing values were omitted, fewer than two',
                    'cases remained for at least one pair of variables. (The pairs of',
                    'variables involved can be determined by examination of the contents of',
                    'the array cnt). All means, standard deviations, sums of',
                    'squares and cross-products, and correlation coefficients based on two or',
                    'more cases are returned by the function even if errno = 4.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _xbar_cls.refresh_py(_xbar)
        _std_cls.refresh_py(_std)
        _ssp_cls.refresh_py(_ssp)
        _r_cls.refresh_py(_r)
        _cnt_cls.refresh_py(_cnt)
        _ncases_py = utils._EngineIntScalarType.to_py(_ncases)
        return_data = _ncases_py
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = _collections.namedtuple(
                'nag_return_data',
                ['ncases'],
            )(
                _ncases_py,
            )
    _err_class.manage()
    return return_data

def coeffs_zero(n, m, x, xbar, std, sspz, rz):
    u"""
    Correlation-like coefficients (about zero), all variables, no
    missing values.

    ``coeffs_zero`` computes means and standard deviations of variables,
    sums of squares and cross-products about zero, and correlation-like
    coefficients for a set of data.

    For full information please refer to the NAG Library document for
    g02bd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02bdf.html

    Parameters
    ----------
    n : int
        n, the number of observations or cases.

    m : int
        m, the number of variables.

    x : float, shape (`n`, `m`)
        `x`[i-1,j-1] must be set to the value of x_{ij}, the ith
        observation on the jth variable, for j = 1,2,...,m, for i =
        1,2,...,n.

    xbar : float, shape (`m`), modified in place
        `On exit`: `xbar`[j-1] contains the mean value, x__j, of the jth
        variable, for j = 1,2,...,m.

    std : float, shape (`m`), modified in place
        `On exit`: the standard deviation, s_j, of the jth variable, for
        j = 1,2,...,m.

    sspz : float, shape (`m`, `m`), modified in place
        `On exit`: `sspz`[j-1,k-1] is the cross-product about zero,
        S~_{jk}, for k = 1,2,...,m, for j = 1,2,...,m.

    rz : float, shape (`m`, `m`), modified in place
        `On exit`: `rz`[j-1,k-1] is the correlation-like coefficient,
        R~_{jk}, between the jth and kth variables, for k = 1,2,...,m,
        for j = 1,2,...,m.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 2.

        (`errno` 2)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 2.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.coeffs_zero'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02bdft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Input, x[ldx, m].
        utils._BYREF_INT, # Input, ldx.
        utils._EngineFloat64ArrayType.c_type, # Output, xbar[m].
        utils._EngineFloat64ArrayType.c_type, # Output, std[m].
        utils._EngineFloat64ArrayType.c_type, # Output, sspz[ldsspz, m].
        utils._BYREF_INT, # Input, ldsspz.
        utils._EngineFloat64ArrayType.c_type, # Output, rz[ldrz, m].
        utils._BYREF_INT, # Input, ldrz.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((x, 'x'), (sspz, 'sspz'), (rz, 'rz')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _ldx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py, _m_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _xbar_cls = utils._EngineFloat64ArrayType(
        xbar, {'fun_name': fun_name, 'entity': 'xbar'},
        exp_shape=(_m_py,),
        check_contents=False,
    )
    _xbar = _xbar_cls.to_ctypes(can_reference_contents=False)
    _std_cls = utils._EngineFloat64ArrayType(
        std, {'fun_name': fun_name, 'entity': 'std'},
        exp_shape=(_m_py,),
        check_contents=False,
    )
    _std = _std_cls.to_ctypes(can_reference_contents=False)
    _ldsspz_py = utils._EngineFloat64ArrayType.get_nd(sspz, {'fun_name': fun_name, 'entity': 'sspz'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldsspz = utils._EngineIntScalarType(
        _ldsspz_py, {'fun_name': fun_name, 'entity': '_ldsspz'},
    ).to_ctypes()
    _sspz_cls = utils._EngineFloat64ArrayType(
        sspz, {'fun_name': fun_name, 'entity': 'sspz'},
        exp_shape=(_m_py, _m_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _sspz = _sspz_cls.to_ctypes(can_reference_contents=False)
    _ldrz_py = utils._EngineFloat64ArrayType.get_nd(rz, {'fun_name': fun_name, 'entity': 'rz'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldrz = utils._EngineIntScalarType(
        _ldrz_py, {'fun_name': fun_name, 'entity': '_ldrz'},
    ).to_ctypes()
    _rz_cls = utils._EngineFloat64ArrayType(
        rz, {'fun_name': fun_name, 'entity': 'rz'},
        exp_shape=(_m_py, _m_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _rz = _rz_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _m,
        _x,
        _ldx,
        _xbar,
        _std,
        _sspz,
        _ldsspz,
        _rz,
        _ldrz,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _xbar_cls.refresh_py(_xbar)
        _std_cls.refresh_py(_std)
        _sspz_cls.refresh_py(_sspz)
        _rz_cls.refresh_py(_rz)
        return_data = None
    _err_class.manage()
    return return_data

def coeffs_zero_miss_case(n, m, x, miss, xmiss, xbar, std, sspz, rz):
    u"""
    Correlation-like coefficients (about zero), all variables, casewise
    treatment of missing values.

    ``coeffs_zero_miss_case`` computes means and standard deviations of
    variables, sums of squares and cross-products about zero, and
    correlation-like coefficients for a set of data omitting completely
    any cases with a missing observation for any variable.

    For full information please refer to the NAG Library document for
    g02be

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02bef.html

    Parameters
    ----------
    n : int
        n, the number of observations or cases.

    m : int
        m, the number of variables.

    x : float, shape (`n`, `m`)
        `x`[i-1,j-1] must be set to x_{ij}, the value of the ith
        observation on the jth variable, for j = 1,2,...,m, for i =
        1,2,...,n.

    miss : int, shape (`m`), modified in place
        `On entry`: `miss`[j-1] must be set equal to 1 if a missing
        value, xm_j, is to be specified for the jth variable in the
        array `x`, or set equal to 0 otherwise. Values of `miss` must be
        given for all m variables in the array `x`.

        `On exit`: the array `miss` is overwritten by the function, and
        the information it contained on entry is lost.

    xmiss : float, shape (`m`), modified in place
        `On entry`: `xmiss`[j-1] must be set to the missing value, xm_j,
        to be associated with the jth variable in the array `x`, for
        those variables for which missing values are specified by means
        of the array `miss` (see [Accuracy]).

        `On exit`: the array `xmiss` is overwritten by the function, and
        the information it contained on entry is lost.

    xbar : float, shape (`m`), modified in place
        `On exit`: the mean value, x__j, of the jth variable, for j =
        1,2,...,m.

    std : float, shape (`m`), modified in place
        `On exit`: the standard deviation, s_j, of the jth variable, for
        j = 1,2,...,m.

    sspz : float, shape (`m`, `m`), modified in place
        `On exit`: `sspz`[j-1,k-1] is the cross-product about zero,
        S~_{jk}, for k = 1,2,...,m, for j = 1,2,...,m.

    rz : float, shape (`m`, `m`), modified in place
        `On exit`: `rz`[j-1,k-1] is the correlation-like coefficient,
        R~_{jk}, between the jth and kth variables, for k = 1,2,...,m,
        for j = 1,2,...,m.

    Returns
    -------
    ncases : int
        The number of cases actually used in the calculations (when
        cases involving missing values have been eliminated).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 2.

        (`errno` 2)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 2.

        (`errno` 4)
            After observations with missing values were omitted, no
            cases remained.

        (`errno` 5)
            After observations with missing values were omitted, only
            one case remained.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.coeffs_zero_miss_case'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02beft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Input, x[ldx, m].
        utils._BYREF_INT, # Input, ldx.
        utils._EngineIntArrayType.c_type, # Input/Output, miss[m].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, xmiss[m].
        utils._EngineFloat64ArrayType.c_type, # Output, xbar[m].
        utils._EngineFloat64ArrayType.c_type, # Output, std[m].
        utils._EngineFloat64ArrayType.c_type, # Output, sspz[ldsspz, m].
        utils._BYREF_INT, # Input, ldsspz.
        utils._EngineFloat64ArrayType.c_type, # Output, rz[ldrz, m].
        utils._BYREF_INT, # Input, ldrz.
        utils._BYREF_INT, # Output, ncases.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((x, 'x'), (sspz, 'sspz'), (rz, 'rz')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _ldx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py, _m_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _miss_cls = utils._EngineIntArrayType(
        miss, {'fun_name': fun_name, 'entity': 'miss'},
        exp_shape=(_m_py,),
    )
    _miss = _miss_cls.to_ctypes()
    _xmiss_cls = utils._EngineFloat64ArrayType(
        xmiss, {'fun_name': fun_name, 'entity': 'xmiss'},
        exp_shape=(_m_py,),
    )
    _xmiss = _xmiss_cls.to_ctypes()
    _xbar_cls = utils._EngineFloat64ArrayType(
        xbar, {'fun_name': fun_name, 'entity': 'xbar'},
        exp_shape=(_m_py,),
        check_contents=False,
    )
    _xbar = _xbar_cls.to_ctypes(can_reference_contents=False)
    _std_cls = utils._EngineFloat64ArrayType(
        std, {'fun_name': fun_name, 'entity': 'std'},
        exp_shape=(_m_py,),
        check_contents=False,
    )
    _std = _std_cls.to_ctypes(can_reference_contents=False)
    _ldsspz_py = utils._EngineFloat64ArrayType.get_nd(sspz, {'fun_name': fun_name, 'entity': 'sspz'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldsspz = utils._EngineIntScalarType(
        _ldsspz_py, {'fun_name': fun_name, 'entity': '_ldsspz'},
    ).to_ctypes()
    _sspz_cls = utils._EngineFloat64ArrayType(
        sspz, {'fun_name': fun_name, 'entity': 'sspz'},
        exp_shape=(_m_py, _m_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _sspz = _sspz_cls.to_ctypes(can_reference_contents=False)
    _ldrz_py = utils._EngineFloat64ArrayType.get_nd(rz, {'fun_name': fun_name, 'entity': 'rz'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldrz = utils._EngineIntScalarType(
        _ldrz_py, {'fun_name': fun_name, 'entity': '_ldrz'},
    ).to_ctypes()
    _rz_cls = utils._EngineFloat64ArrayType(
        rz, {'fun_name': fun_name, 'entity': 'rz'},
        exp_shape=(_m_py, _m_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _rz = _rz_cls.to_ctypes(can_reference_contents=False)
    _ncases = utils._EngineIntScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _m,
        _x,
        _ldx,
        _miss,
        _xmiss,
        _xbar,
        _std,
        _sspz,
        _ldsspz,
        _rz,
        _ldrz,
        _ncases,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'After observations with missing values were omitted,',
                    'no cases remained.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'After observations with missing values were omitted,',
                    'only one case remained.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _miss_cls.refresh_py(_miss)
        _xmiss_cls.refresh_py(_xmiss)
        _xbar_cls.refresh_py(_xbar)
        _std_cls.refresh_py(_std)
        _sspz_cls.refresh_py(_sspz)
        _rz_cls.refresh_py(_rz)
        _ncases_py = utils._EngineIntScalarType.to_py(_ncases)
        return_data = _ncases_py
    _err_class.manage()
    return return_data

def coeffs_zero_miss_pair(n, m, x, miss, xmiss, xbar, std, sspz, rz, cnt):
    u"""
    Correlation-like coefficients (about zero), all variables, pairwise
    treatment of missing values.

    ``coeffs_zero_miss_pair`` computes means and standard deviations of
    variables, sums of squares and cross-products about zero and
    correlation-like coefficients for a set of data omitting cases with
    missing values from only those calculations involving the variables
    for which the values are missing.

    For full information please refer to the NAG Library document for
    g02bf

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02bff.html

    Parameters
    ----------
    n : int
        n, the number of observations or cases.

    m : int
        m, the number of variables.

    x : float, shape (`n`, `m`)
        `x`[i-1,j-1] must be set to x_{ij}, the value of the ith
        observation on the jth variable, for j = 1,2,...,m, for i =
        1,2,...,n.

    miss : int, shape (`m`)
        `miss`[j-1] must be set equal to 1 if a missing value, xm_j, is
        to be specified for the jth variable in the array `x`, or set
        equal to 0 otherwise. Values of `miss` must be given for all m
        variables in the array `x`.

    xmiss : float, shape (`m`)
        `xmiss`[j-1] must be set to the missing value, xm_j, to be
        associated with the jth variable in the array `x`, for those
        variables for which missing values are specified by means of the
        array `miss` (see [Accuracy]).

    xbar : float, shape (`m`), modified in place
        `On exit`: the mean value, x__j, of the jth variable, for j =
        1,2,...,m.

    std : float, shape (`m`), modified in place
        `On exit`: the standard deviation, s_j, of the jth variable, for
        j = 1,2,...,m.

    sspz : float, shape (`m`, `m`), modified in place
        `On exit`: `sspz`[j-1,k-1] is the cross-product about zero,
        S~_{jk}, for k = 1,2,...,m, for j = 1,2,...,m.

    rz : float, shape (`m`, `m`), modified in place
        `On exit`: `rz`[j-1,k-1] is the correlation-like coefficient,
        R~_{jk}, between the jth and kth variables, for k = 1,2,...,m,
        for j = 1,2,...,m.

    cnt : float, shape (`m`, `m`), modified in place
        `On exit`: `cnt`[j-1,k-1] is the number of cases, c_{jk},
        actually used in the calculation of S~_{jk}, and R~_{jk}, the
        sum of cross-products and correlation-like coefficient for the
        jth and kth variables, for k = 1,2,...,m, for j = 1,2,...,m.

    Returns
    -------
    ncases : int
        The minimum number of cases used in the calculation of any of
        the sums of squares and cross-products and correlation-like
        coefficients (when cases involving missing values have been
        eliminated).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 2.

        (`errno` 2)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 2.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 4)
            After observations with missing values were omitted, fewer
            than two cases remained for at least one pair of variables.
            (The pairs of variables involved can be determined by
            examination of the contents of the array `cnt`). All means,
            standard deviations, sums of squares and cross-products, and
            correlation-like coefficients based on two or more cases are
            returned by the function even if `errno` = 4.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.coeffs_zero_miss_pair'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02bfft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Input, x[ldx, m].
        utils._BYREF_INT, # Input, ldx.
        utils._EngineIntArrayType.c_type, # Input, miss[m].
        utils._EngineFloat64ArrayType.c_type, # Input, xmiss[m].
        utils._EngineFloat64ArrayType.c_type, # Output, xbar[m].
        utils._EngineFloat64ArrayType.c_type, # Output, std[m].
        utils._EngineFloat64ArrayType.c_type, # Output, sspz[ldsspz, m].
        utils._BYREF_INT, # Input, ldsspz.
        utils._EngineFloat64ArrayType.c_type, # Output, rz[ldrz, m].
        utils._BYREF_INT, # Input, ldrz.
        utils._BYREF_INT, # Output, ncases.
        utils._EngineFloat64ArrayType.c_type, # Output, cnt[ldcnt, m].
        utils._BYREF_INT, # Input, ldcnt.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((x, 'x'), (sspz, 'sspz'), (rz, 'rz'), (cnt, 'cnt')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _ldx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py, _m_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _miss = utils._EngineIntArrayType(
        miss, {'fun_name': fun_name, 'entity': 'miss'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _xmiss = utils._EngineFloat64ArrayType(
        xmiss, {'fun_name': fun_name, 'entity': 'xmiss'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _xbar_cls = utils._EngineFloat64ArrayType(
        xbar, {'fun_name': fun_name, 'entity': 'xbar'},
        exp_shape=(_m_py,),
        check_contents=False,
    )
    _xbar = _xbar_cls.to_ctypes(can_reference_contents=False)
    _std_cls = utils._EngineFloat64ArrayType(
        std, {'fun_name': fun_name, 'entity': 'std'},
        exp_shape=(_m_py,),
        check_contents=False,
    )
    _std = _std_cls.to_ctypes(can_reference_contents=False)
    _ldsspz_py = utils._EngineFloat64ArrayType.get_nd(sspz, {'fun_name': fun_name, 'entity': 'sspz'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldsspz = utils._EngineIntScalarType(
        _ldsspz_py, {'fun_name': fun_name, 'entity': '_ldsspz'},
    ).to_ctypes()
    _sspz_cls = utils._EngineFloat64ArrayType(
        sspz, {'fun_name': fun_name, 'entity': 'sspz'},
        exp_shape=(_m_py, _m_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _sspz = _sspz_cls.to_ctypes(can_reference_contents=False)
    _ldrz_py = utils._EngineFloat64ArrayType.get_nd(rz, {'fun_name': fun_name, 'entity': 'rz'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldrz = utils._EngineIntScalarType(
        _ldrz_py, {'fun_name': fun_name, 'entity': '_ldrz'},
    ).to_ctypes()
    _rz_cls = utils._EngineFloat64ArrayType(
        rz, {'fun_name': fun_name, 'entity': 'rz'},
        exp_shape=(_m_py, _m_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _rz = _rz_cls.to_ctypes(can_reference_contents=False)
    _ncases = utils._EngineIntScalarType.empty_ctypes()
    _ldcnt_py = utils._EngineFloat64ArrayType.get_nd(cnt, {'fun_name': fun_name, 'entity': 'cnt'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldcnt = utils._EngineIntScalarType(
        _ldcnt_py, {'fun_name': fun_name, 'entity': '_ldcnt'},
    ).to_ctypes()
    _cnt_cls = utils._EngineFloat64ArrayType(
        cnt, {'fun_name': fun_name, 'entity': 'cnt'},
        exp_shape=(_m_py, _m_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _cnt = _cnt_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _m,
        _x,
        _ldx,
        _miss,
        _xmiss,
        _xbar,
        _std,
        _sspz,
        _ldsspz,
        _rz,
        _ldrz,
        _ncases,
        _cnt,
        _ldcnt,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99993:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'After observations with missing values were omitted, fewer than two',
                    'cases remained for at least one pair of variables. (The pairs of',
                    'variables involved can be determined by examination of the contents of',
                    'the array cnt). All means, standard deviations, sums of squares and',
                    'cross-products, and correlation-like coefficients based on two or more',
                    'cases are returned by the function even if errno = 4.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _xbar_cls.refresh_py(_xbar)
        _std_cls.refresh_py(_std)
        _sspz_cls.refresh_py(_sspz)
        _rz_cls.refresh_py(_rz)
        _cnt_cls.refresh_py(_cnt)
        _ncases_py = utils._EngineIntScalarType.to_py(_ncases)
        return_data = _ncases_py
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = _collections.namedtuple(
                'nag_return_data',
                ['ncases'],
            )(
                _ncases_py,
            )
    _err_class.manage()
    return return_data

def coeffs_pearson_subset(n, m, x, nvars, kvar, xbar, std, ssp, r):
    u"""
    Pearson product-moment correlation coefficients, subset of
    variables, no missing values.

    ``coeffs_pearson_subset`` computes means and standard deviations,
    sums of squares and cross-products of deviations from means, and
    Pearson product-moment correlation coefficients for selected
    variables.

    For full information please refer to the NAG Library document for
    g02bg

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02bgf.html

    Parameters
    ----------
    n : int
        n, the number of observations or cases.

    m : int
        m, the number of variables.

    x : float, shape (`n`, `m`)
        `x`[i-1,j-1] must be set to x_{ij}, the value of the ith
        observation on the jth variable, for j = 1,2,...,m, for i =
        1,2,...,n.

    nvars : int
        p, the number of variables for which information is required.

    kvar : int, shape (`nvars`)
        `kvar`[j] must be set to the column number in `x` of the jth
        variable for which information is required, for j = 0,...,p-1.

    xbar : float, shape (`nvars`), modified in place
        `On exit`: the mean value, x__j, of the variable specified in
        `kvar`[j-1], for j = 1,2,...,p.

    std : float, shape (`nvars`), modified in place
        `On exit`: the standard deviation, s_j, of the variable
        specified in `kvar`[j-1], for j = 1,2,...,p.

    ssp : float, shape (`nvars`, `nvars`), modified in place
        `On exit`: `ssp`[j-1,k-1] is the cross-product of deviations,
        S_{jk}, for the variables specified in `kvar`[j-1] and
        `kvar`[k-1], for k = 1,2,...,p, for j = 1,2,...,p.

    r : float, shape (`nvars`, `nvars`), modified in place
        `On exit`: `r`[j-1,k-1] is the product-moment correlation
        coefficient, R_{jk}, between the variables specified in
        `kvar`[j-1] and `kvar`[k-1], for k = 1,2,...,p, for j =
        1,2,...,p.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 2.

        (`errno` 2)
            On entry, `nvars` = *<value>* and `m` = *<value>*.

            Constraint: `nvars` >= 2 and `nvars` <= `m`.

        (`errno` 4)
            On entry, i = *<value>*, `kvar`[i-1] = *<value>* and `m` =
            *<value>*.

            Constraint: 1 <= `kvar`[i-1] <= `m`.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.coeffs_pearson_subset'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02bgft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Input, x[ldx, m].
        utils._BYREF_INT, # Input, ldx.
        utils._BYREF_INT, # Input, nvars.
        utils._EngineIntArrayType.c_type, # Input, kvar[nvars].
        utils._EngineFloat64ArrayType.c_type, # Output, xbar[nvars].
        utils._EngineFloat64ArrayType.c_type, # Output, std[nvars].
        utils._EngineFloat64ArrayType.c_type, # Output, ssp[ldssp, nvars].
        utils._BYREF_INT, # Input, ldssp.
        utils._EngineFloat64ArrayType.c_type, # Output, r[ldr, nvars].
        utils._BYREF_INT, # Input, ldr.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((x, 'x'), (ssp, 'ssp'), (r, 'r')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _ldx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py, _m_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _nvars = utils._EngineIntScalarType(
        nvars, {'fun_name': fun_name, 'entity': 'nvars'},
    ).to_ctypes()
    _nvars_py = utils._EngineIntScalarType.to_py(_nvars)
    _kvar = utils._EngineIntArrayType(
        kvar, {'fun_name': fun_name, 'entity': 'kvar'},
        exp_shape=(_nvars_py,),
    ).to_ctypes()
    _xbar_cls = utils._EngineFloat64ArrayType(
        xbar, {'fun_name': fun_name, 'entity': 'xbar'},
        exp_shape=(_nvars_py,),
        check_contents=False,
    )
    _xbar = _xbar_cls.to_ctypes(can_reference_contents=False)
    _std_cls = utils._EngineFloat64ArrayType(
        std, {'fun_name': fun_name, 'entity': 'std'},
        exp_shape=(_nvars_py,),
        check_contents=False,
    )
    _std = _std_cls.to_ctypes(can_reference_contents=False)
    _ldssp_py = utils._EngineFloat64ArrayType.get_nd(ssp, {'fun_name': fun_name, 'entity': 'ssp'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldssp = utils._EngineIntScalarType(
        _ldssp_py, {'fun_name': fun_name, 'entity': '_ldssp'},
    ).to_ctypes()
    _ssp_cls = utils._EngineFloat64ArrayType(
        ssp, {'fun_name': fun_name, 'entity': 'ssp'},
        exp_shape=(_nvars_py, _nvars_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _ssp = _ssp_cls.to_ctypes(can_reference_contents=False)
    _ldr_py = utils._EngineFloat64ArrayType.get_nd(r, {'fun_name': fun_name, 'entity': 'r'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldr = utils._EngineIntScalarType(
        _ldr_py, {'fun_name': fun_name, 'entity': '_ldr'},
    ).to_ctypes()
    _r_cls = utils._EngineFloat64ArrayType(
        r, {'fun_name': fun_name, 'entity': 'r'},
        exp_shape=(_nvars_py, _nvars_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _r = _r_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _m,
        _x,
        _ldx,
        _nvars,
        _kvar,
        _xbar,
        _std,
        _ssp,
        _ldssp,
        _r,
        _ldr,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nvars = {:s} and m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nvars >= 2 and nvars <= m.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, kvar[i-1] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'm = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: 1 <= kvar[i-1] <= m.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _xbar_cls.refresh_py(_xbar)
        _std_cls.refresh_py(_std)
        _ssp_cls.refresh_py(_ssp)
        _r_cls.refresh_py(_r)
        return_data = None
    _err_class.manage()
    return return_data

def coeffs_pearson_subset_miss_case(n, m, x, miss, xmiss, mistyp, nvars, kvar, xbar, std, ssp, r):
    u"""
    Pearson product-moment correlation coefficients, subset of
    variables, casewise treatment of missing values.

    ``coeffs_pearson_subset_miss_case`` computes means and standard
    deviations, sums of squares and cross-products of deviations from
    means, and Pearson product-moment correlation coefficients for
    selected variables omitting completely any cases with a missing
    observation for any variable (either over all variables in the
    dataset or over only those variables in the selected subset).

    For full information please refer to the NAG Library document for
    g02bh

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02bhf.html

    Parameters
    ----------
    n : int
        n, the number of observations or cases.

    m : int
        m, the number of variables.

    x : float, shape (`n`, `m`)
        `x`[i-1,j-1] must be set to x_{ij}, the value of the ith
        observation on the jth variable, for j = 1,2,...,m, for i =
        1,2,...,n.

    miss : int, shape (`m`), modified in place
        `On entry`: `miss`[j-1] must be set equal to 1 if a missing
        value, xm_j, is to be specified for the jth variable in the
        array `x`, or set equal to 0 otherwise. Values of `miss` must be
        given for all m variables in the array `x`.

        `On exit`: the array `miss` is overwritten by the function, and
        the information it contained on entry is lost.

    xmiss : float, shape (`m`), modified in place
        `On entry`: `xmiss`[j-1] must be set to the missing value, xm_j,
        to be associated with the jth variable in the array `x`, for
        those variables for which missing values are specified by means
        of the array `miss` (see [Accuracy]).

        `On exit`: the array `xmiss` is overwritten by the function, and
        the information it contained on entry is lost.

    mistyp : int
        Indicates the manner in which missing observations are to be
        treated.

        `mistyp` = 1

            A case is excluded if it contains a missing value for any of
            the variables 1,2,...,m.

        `mistyp` = 0

            A case is excluded if it contains a missing value for any of
            the p(<= m) variables specified in the array `kvar`.

    nvars : int
        p, the number of variables for which information is required.

    kvar : int, shape (`nvars`)
        `kvar`[j] must be set to the column number in `x` of the jth
        variable for which information is required, for j = 0,...,p-1.

    xbar : float, shape (`nvars`), modified in place
        `On exit`: the mean value, of x__j, of the variable specified in
        `kvar`[j-1], for j = 1,2,...,p.

    std : float, shape (`nvars`), modified in place
        `On exit`: the standard deviation, s_j, of the variable
        specified in `kvar`[j-1], for j = 1,2,...,p.

    ssp : float, shape (`nvars`, `nvars`), modified in place
        `On exit`: `ssp`[j-1,k-1] is the cross-product of deviations,
        S_{jk}, for the variables specified in `kvar`[j-1] and
        `kvar`[k-1], for k = 1,2,...,p, for j = 1,2,...,p.

    r : float, shape (`nvars`, `nvars`), modified in place
        `On exit`: `r`[j-1,k-1] is the product-moment correlation
        coefficient, R_{jk}, between the variables specified in
        `kvar`[j-1] and `kvar`[k-1], for k = 1,2,...,p, for j =
        1,2,...,p.

    Returns
    -------
    ncases : int
        The number of cases actually used in the calculations (when
        cases involving missing values have been eliminated).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 2.

        (`errno` 2)
            On entry, `nvars` = *<value>* and `m` = *<value>*.

            Constraint: `nvars` >= 2 and `nvars` <= `m`.

        (`errno` 4)
            On entry, i = *<value>*, `kvar`[i-1] = *<value>* and `m` =
            *<value>*.

            Constraint: 1 <= `kvar`[i-1] <= `m`.

        (`errno` 5)
            On entry, `mistyp` = *<value>*.

            Constraint: `mistyp` = 0 or 1.

        (`errno` 6)
            After observations with missing values were omitted, no
            cases remained.

        (`errno` 7)
            After observations with missing values were omitted, only
            one case remained.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.coeffs_pearson_subset_miss_case'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02bhft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Input, x[ldx, m].
        utils._BYREF_INT, # Input, ldx.
        utils._EngineIntArrayType.c_type, # Input/Output, miss[m].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, xmiss[m].
        utils._BYREF_INT, # Input, mistyp.
        utils._BYREF_INT, # Input, nvars.
        utils._EngineIntArrayType.c_type, # Input, kvar[nvars].
        utils._EngineFloat64ArrayType.c_type, # Output, xbar[nvars].
        utils._EngineFloat64ArrayType.c_type, # Output, std[nvars].
        utils._EngineFloat64ArrayType.c_type, # Output, ssp[ldssp, nvars].
        utils._BYREF_INT, # Input, ldssp.
        utils._EngineFloat64ArrayType.c_type, # Output, r[ldr, nvars].
        utils._BYREF_INT, # Input, ldr.
        utils._BYREF_INT, # Output, ncases.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((x, 'x'), (ssp, 'ssp'), (r, 'r')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _ldx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py, _m_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _miss_cls = utils._EngineIntArrayType(
        miss, {'fun_name': fun_name, 'entity': 'miss'},
        exp_shape=(_m_py,),
    )
    _miss = _miss_cls.to_ctypes()
    _xmiss_cls = utils._EngineFloat64ArrayType(
        xmiss, {'fun_name': fun_name, 'entity': 'xmiss'},
        exp_shape=(_m_py,),
    )
    _xmiss = _xmiss_cls.to_ctypes()
    _mistyp = utils._EngineIntScalarType(
        mistyp, {'fun_name': fun_name, 'entity': 'mistyp'},
    ).to_ctypes()
    _nvars = utils._EngineIntScalarType(
        nvars, {'fun_name': fun_name, 'entity': 'nvars'},
    ).to_ctypes()
    _nvars_py = utils._EngineIntScalarType.to_py(_nvars)
    _kvar = utils._EngineIntArrayType(
        kvar, {'fun_name': fun_name, 'entity': 'kvar'},
        exp_shape=(_nvars_py,),
    ).to_ctypes()
    _xbar_cls = utils._EngineFloat64ArrayType(
        xbar, {'fun_name': fun_name, 'entity': 'xbar'},
        exp_shape=(_nvars_py,),
        check_contents=False,
    )
    _xbar = _xbar_cls.to_ctypes(can_reference_contents=False)
    _std_cls = utils._EngineFloat64ArrayType(
        std, {'fun_name': fun_name, 'entity': 'std'},
        exp_shape=(_nvars_py,),
        check_contents=False,
    )
    _std = _std_cls.to_ctypes(can_reference_contents=False)
    _ldssp_py = utils._EngineFloat64ArrayType.get_nd(ssp, {'fun_name': fun_name, 'entity': 'ssp'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldssp = utils._EngineIntScalarType(
        _ldssp_py, {'fun_name': fun_name, 'entity': '_ldssp'},
    ).to_ctypes()
    _ssp_cls = utils._EngineFloat64ArrayType(
        ssp, {'fun_name': fun_name, 'entity': 'ssp'},
        exp_shape=(_nvars_py, _nvars_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _ssp = _ssp_cls.to_ctypes(can_reference_contents=False)
    _ldr_py = utils._EngineFloat64ArrayType.get_nd(r, {'fun_name': fun_name, 'entity': 'r'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldr = utils._EngineIntScalarType(
        _ldr_py, {'fun_name': fun_name, 'entity': '_ldr'},
    ).to_ctypes()
    _r_cls = utils._EngineFloat64ArrayType(
        r, {'fun_name': fun_name, 'entity': 'r'},
        exp_shape=(_nvars_py, _nvars_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _r = _r_cls.to_ctypes(can_reference_contents=False)
    _ncases = utils._EngineIntScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _m,
        _x,
        _ldx,
        _miss,
        _xmiss,
        _mistyp,
        _nvars,
        _kvar,
        _xbar,
        _std,
        _ssp,
        _ldssp,
        _r,
        _ldr,
        _ncases,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nvars = {:s} and m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nvars >= 2 and nvars <= m.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, kvar[i-1] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'm = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: 1 <= kvar[i-1] <= m.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mistyp = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mistyp = 0 or 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'After observations with missing values were omitted,',
                    'no cases remained.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'After observations with missing values were omitted,',
                    'only one case remained.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _miss_cls.refresh_py(_miss)
        _xmiss_cls.refresh_py(_xmiss)
        _xbar_cls.refresh_py(_xbar)
        _std_cls.refresh_py(_std)
        _ssp_cls.refresh_py(_ssp)
        _r_cls.refresh_py(_r)
        _ncases_py = utils._EngineIntScalarType.to_py(_ncases)
        return_data = _ncases_py
    _err_class.manage()
    return return_data

def coeffs_pearson_subset_miss_pair(n, m, x, miss, xmiss, nvars, kvar, xbar, std, ssp, r, cnt):
    u"""
    Pearson product-moment correlation coefficients, subset of
    variables, pairwise treatment of missing values.

    ``coeffs_pearson_subset_miss_pair`` computes means and standard
    deviations, sums of squares and cross-products of deviations from
    means, and Pearson product-moment correlation coefficients for
    selected variables omitting cases with missing values from only
    those calculations involving the variables for which the values are
    missing.

    For full information please refer to the NAG Library document for
    g02bj

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02bjf.html

    Parameters
    ----------
    n : int
        n, the number of observations or cases.

    m : int
        m, the number of variables.

    x : float, shape (`n`, `m`)
        `x`[i-1,j-1] must be set to x_{ij}, the value of the ith
        observation on the jth variable, for j = 1,2,...,m, for i =
        1,2,...,n.

    miss : int, shape (`m`)
        `miss`[j-1] must be set equal to 1 if a missing value, xm_j, is
        to be specified for the jth variable in the array `x`, or set
        equal to 0 otherwise. Values of `miss` must be given for all m
        variables in the array `x`.

    xmiss : float, shape (`m`)
        `xmiss`[j-1] must be set to the missing value, xm_j, to be
        associated with the jth variable in the array `x`, for those
        variables for which missing values are specified by means of the
        array `miss` (see [Accuracy]).

    nvars : int
        p, the number of variables for which information is required.

    kvar : int, shape (`nvars`)
        `kvar`[j] must be set to the column number in `x` of the jth
        variable for which information is required, for j = 0,...,p-1.

    xbar : float, shape (`nvars`), modified in place
        `On exit`: the mean value, x__j, of the variable specified in
        `kvar`[j-1], for j = 1,2,...,p.

    std : float, shape (`nvars`), modified in place
        `On exit`: the standard deviation, s_j, of the variable
        specified in `kvar`[j-1], for j = 1,2,...,p.

    ssp : float, shape (`nvars`, `nvars`), modified in place
        `On exit`: `ssp`[j-1,k-1] is the cross-product of deviations,
        S_{jk}, for the variables specified in `kvar`[j-1] and
        `kvar`[k-1], for k = 1,2,...,p, for j = 1,2,...,p.

    r : float, shape (`nvars`, `nvars`), modified in place
        `On exit`: `r`[j-1,k-1] is the product-moment correlation
        coefficient, R_{jk}, between the variables specified in
        `kvar`[j-1] and `kvar`[k-1], for k = 1,2,...,p, for j =
        1,2,...,p.

    cnt : float, shape (`nvars`, `nvars`), modified in place
        `On exit`: `cnt`[j-1,k-1] is the number of cases, c_{jk},
        actually used in the calculation of S_{jk}, and R_{jk}, the sum
        of cross-products and correlation coefficient for the variables
        specified in `kvar`[j-1] and `kvar`[k-1], for k = 1,2,...,p, for
        j = 1,2,...,p.

    Returns
    -------
    ncases : int
        The minimum number of cases used in the calculation of any of
        the sums of squares and cross-products and correlation
        coefficients (when cases involving missing values have been
        eliminated).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 2.

        (`errno` 2)
            On entry, `nvars` = *<value>* and `m` = *<value>*.

            Constraint: `nvars` >= 2 and `nvars` <= `m`.

        (`errno` 4)
            On entry, i = *<value>*, `kvar`[i-1] = *<value>* and `m` =
            *<value>*.

            Constraint: 1 <= `kvar`[i-1] <= `m`.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 5)
            After observations with missing values were omitted, fewer
            than two cases remained.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.coeffs_pearson_subset_miss_pair'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02bjft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Input, x[ldx, m].
        utils._BYREF_INT, # Input, ldx.
        utils._EngineIntArrayType.c_type, # Input, miss[m].
        utils._EngineFloat64ArrayType.c_type, # Input, xmiss[m].
        utils._BYREF_INT, # Input, nvars.
        utils._EngineIntArrayType.c_type, # Input, kvar[nvars].
        utils._EngineFloat64ArrayType.c_type, # Output, xbar[nvars].
        utils._EngineFloat64ArrayType.c_type, # Output, std[nvars].
        utils._EngineFloat64ArrayType.c_type, # Output, ssp[ldssp, nvars].
        utils._BYREF_INT, # Input, ldssp.
        utils._EngineFloat64ArrayType.c_type, # Output, r[ldr, nvars].
        utils._BYREF_INT, # Input, ldr.
        utils._BYREF_INT, # Output, ncases.
        utils._EngineFloat64ArrayType.c_type, # Output, cnt[ldcnt, nvars].
        utils._BYREF_INT, # Input, ldcnt.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((x, 'x'), (ssp, 'ssp'), (r, 'r'), (cnt, 'cnt')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _ldx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py, _m_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _miss = utils._EngineIntArrayType(
        miss, {'fun_name': fun_name, 'entity': 'miss'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _xmiss = utils._EngineFloat64ArrayType(
        xmiss, {'fun_name': fun_name, 'entity': 'xmiss'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _nvars = utils._EngineIntScalarType(
        nvars, {'fun_name': fun_name, 'entity': 'nvars'},
    ).to_ctypes()
    _nvars_py = utils._EngineIntScalarType.to_py(_nvars)
    _kvar = utils._EngineIntArrayType(
        kvar, {'fun_name': fun_name, 'entity': 'kvar'},
        exp_shape=(_nvars_py,),
    ).to_ctypes()
    _xbar_cls = utils._EngineFloat64ArrayType(
        xbar, {'fun_name': fun_name, 'entity': 'xbar'},
        exp_shape=(_nvars_py,),
        check_contents=False,
    )
    _xbar = _xbar_cls.to_ctypes(can_reference_contents=False)
    _std_cls = utils._EngineFloat64ArrayType(
        std, {'fun_name': fun_name, 'entity': 'std'},
        exp_shape=(_nvars_py,),
        check_contents=False,
    )
    _std = _std_cls.to_ctypes(can_reference_contents=False)
    _ldssp_py = utils._EngineFloat64ArrayType.get_nd(ssp, {'fun_name': fun_name, 'entity': 'ssp'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldssp = utils._EngineIntScalarType(
        _ldssp_py, {'fun_name': fun_name, 'entity': '_ldssp'},
    ).to_ctypes()
    _ssp_cls = utils._EngineFloat64ArrayType(
        ssp, {'fun_name': fun_name, 'entity': 'ssp'},
        exp_shape=(_nvars_py, _nvars_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _ssp = _ssp_cls.to_ctypes(can_reference_contents=False)
    _ldr_py = utils._EngineFloat64ArrayType.get_nd(r, {'fun_name': fun_name, 'entity': 'r'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldr = utils._EngineIntScalarType(
        _ldr_py, {'fun_name': fun_name, 'entity': '_ldr'},
    ).to_ctypes()
    _r_cls = utils._EngineFloat64ArrayType(
        r, {'fun_name': fun_name, 'entity': 'r'},
        exp_shape=(_nvars_py, _nvars_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _r = _r_cls.to_ctypes(can_reference_contents=False)
    _ncases = utils._EngineIntScalarType.empty_ctypes()
    _ldcnt_py = utils._EngineFloat64ArrayType.get_nd(cnt, {'fun_name': fun_name, 'entity': 'cnt'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldcnt = utils._EngineIntScalarType(
        _ldcnt_py, {'fun_name': fun_name, 'entity': '_ldcnt'},
    ).to_ctypes()
    _cnt_cls = utils._EngineFloat64ArrayType(
        cnt, {'fun_name': fun_name, 'entity': 'cnt'},
        exp_shape=(_nvars_py, _nvars_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _cnt = _cnt_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _m,
        _x,
        _ldx,
        _miss,
        _xmiss,
        _nvars,
        _kvar,
        _xbar,
        _std,
        _ssp,
        _ldssp,
        _r,
        _ldr,
        _ncases,
        _cnt,
        _ldcnt,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nvars = {:s} and m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nvars >= 2 and nvars <= m.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, kvar[i-1] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'm = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: 1 <= kvar[i-1] <= m.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99992:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'After observations with missing values were omitted,',
                    'fewer than two cases remained.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _xbar_cls.refresh_py(_xbar)
        _std_cls.refresh_py(_std)
        _ssp_cls.refresh_py(_ssp)
        _r_cls.refresh_py(_r)
        _cnt_cls.refresh_py(_cnt)
        _ncases_py = utils._EngineIntScalarType.to_py(_ncases)
        return_data = _ncases_py
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = _collections.namedtuple(
                'nag_return_data',
                ['ncases'],
            )(
                _ncases_py,
            )
    _err_class.manage()
    return return_data

def coeffs_zero_subset(n, m, x, nvars, kvar, xbar, std, sspz, rz):
    u"""
    Correlation-like coefficients (about zero), subset of variables, no
    missing values.

    ``coeffs_zero_subset`` computes means and standard deviations, sums
    of squares and cross-products about zero, and correlation-like
    coefficients for selected variables.

    For full information please refer to the NAG Library document for
    g02bk

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02bkf.html

    Parameters
    ----------
    n : int
        n, the number of observations or cases.

    m : int
        m, the number of variables.

    x : float, shape (`n`, `m`)
        `x`[i-1,j-1] must be set to x_{ij}, the value of the ith
        observation on the jth variable, for j = 1,2,...,m, for i =
        1,2,...,n.

    nvars : int
        p, the number of variables for which information is required.

    kvar : int, shape (`nvars`)
        `kvar`[j] must be set to the column number in `x` of the jth
        variable for which information is required, for j = 0,...,p-1.

    xbar : float, shape (`nvars`), modified in place
        `On exit`: the mean value, x__j, of the variable specified in
        `kvar`[j-1], for j = 1,2,...,p.

    std : float, shape (`nvars`), modified in place
        `On exit`: the standard deviation, s_j, of the variable
        specified in `kvar`[j-1], for j = 1,2,...,p.

    sspz : float, shape (`nvars`, `nvars`), modified in place
        `On exit`: `sspz`[j-1,k-1] is the cross-product about zero,
        S~_{jk}, for the variables specified in `kvar`[j-1] and
        `kvar`[k-1], for k = 1,2,...,p, for j = 1,2,...,p.

    rz : float, shape (`nvars`, `nvars`), modified in place
        `On exit`: `rz`[j-1,k-1] is the correlation-like coefficient,
        R~_{jk}, between the variables specified in `kvar`[j-1] and
        `kvar`[k-1], for k = 1,2,...,p, for j = 1,2,...,p.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 2.

        (`errno` 2)
            On entry, `nvars` = *<value>* and `m` = *<value>*.

            Constraint: `nvars` >= 2 and `nvars` <= `m`.

        (`errno` 4)
            On entry, i = *<value>*, `kvar`[i-1] = *<value>* and `m` =
            *<value>*.

            Constraint: 1 <= `kvar`[i-1] <= `m`.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.coeffs_zero_subset'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02bkft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Input, x[ldx, m].
        utils._BYREF_INT, # Input, ldx.
        utils._BYREF_INT, # Input, nvars.
        utils._EngineIntArrayType.c_type, # Input, kvar[nvars].
        utils._EngineFloat64ArrayType.c_type, # Output, xbar[nvars].
        utils._EngineFloat64ArrayType.c_type, # Output, std[nvars].
        utils._EngineFloat64ArrayType.c_type, # Output, sspz[ldsspz, nvars].
        utils._BYREF_INT, # Input, ldsspz.
        utils._EngineFloat64ArrayType.c_type, # Output, rz[ldrz, nvars].
        utils._BYREF_INT, # Input, ldrz.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((x, 'x'), (sspz, 'sspz'), (rz, 'rz')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _ldx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py, _m_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _nvars = utils._EngineIntScalarType(
        nvars, {'fun_name': fun_name, 'entity': 'nvars'},
    ).to_ctypes()
    _nvars_py = utils._EngineIntScalarType.to_py(_nvars)
    _kvar = utils._EngineIntArrayType(
        kvar, {'fun_name': fun_name, 'entity': 'kvar'},
        exp_shape=(_nvars_py,),
    ).to_ctypes()
    _xbar_cls = utils._EngineFloat64ArrayType(
        xbar, {'fun_name': fun_name, 'entity': 'xbar'},
        exp_shape=(_nvars_py,),
        check_contents=False,
    )
    _xbar = _xbar_cls.to_ctypes(can_reference_contents=False)
    _std_cls = utils._EngineFloat64ArrayType(
        std, {'fun_name': fun_name, 'entity': 'std'},
        exp_shape=(_nvars_py,),
        check_contents=False,
    )
    _std = _std_cls.to_ctypes(can_reference_contents=False)
    _ldsspz_py = utils._EngineFloat64ArrayType.get_nd(sspz, {'fun_name': fun_name, 'entity': 'sspz'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldsspz = utils._EngineIntScalarType(
        _ldsspz_py, {'fun_name': fun_name, 'entity': '_ldsspz'},
    ).to_ctypes()
    _sspz_cls = utils._EngineFloat64ArrayType(
        sspz, {'fun_name': fun_name, 'entity': 'sspz'},
        exp_shape=(_nvars_py, _nvars_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _sspz = _sspz_cls.to_ctypes(can_reference_contents=False)
    _ldrz_py = utils._EngineFloat64ArrayType.get_nd(rz, {'fun_name': fun_name, 'entity': 'rz'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldrz = utils._EngineIntScalarType(
        _ldrz_py, {'fun_name': fun_name, 'entity': '_ldrz'},
    ).to_ctypes()
    _rz_cls = utils._EngineFloat64ArrayType(
        rz, {'fun_name': fun_name, 'entity': 'rz'},
        exp_shape=(_nvars_py, _nvars_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _rz = _rz_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _m,
        _x,
        _ldx,
        _nvars,
        _kvar,
        _xbar,
        _std,
        _sspz,
        _ldsspz,
        _rz,
        _ldrz,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nvars = {:s} and m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nvars >= 2 and nvars <= m.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, kvar[i-1] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'm = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: 1 <= kvar[i-1] <= m.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _xbar_cls.refresh_py(_xbar)
        _std_cls.refresh_py(_std)
        _sspz_cls.refresh_py(_sspz)
        _rz_cls.refresh_py(_rz)
        return_data = None
    _err_class.manage()
    return return_data

def coeffs_zero_subset_miss_case(n, m, x, miss, xmiss, mistyp, nvars, kvar, xbar, std, sspz, rz):
    u"""
    Correlation-like coefficients (about zero), subset of variables,
    casewise treatment of missing values.

    ``coeffs_zero_subset_miss_case`` computes means and standard
    deviations, sums of squares and cross-products about zero, and
    correlation-like coefficients for selected variables omitting
    completely any cases with a missing observation for any variable
    (either over all variables in the dataset or over only those
    variables in the selected subset).

    For full information please refer to the NAG Library document for
    g02bl

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02blf.html

    Parameters
    ----------
    n : int
        n, the number of observations or cases.

    m : int
        m, the number of variables.

    x : float, shape (`n`, `m`)
        `x`[i-1,j-1] must be set to x_{ij}, the value of the ith
        observation on the jth variable, for j = 1,2,...,m, for i =
        1,2,...,n.

    miss : int, shape (`m`), modified in place
        `On entry`: `miss`[j-1] must be set equal to 1 if a missing
        value, xm_j, is to be specified for the jth variable in the
        array `x`, or set equal to 0 otherwise. Values of `miss` must be
        given for all m variables in the array `x`.

        `On exit`: the array `miss` is overwritten by the function, and
        the information it contained on entry is lost.

    xmiss : float, shape (`m`), modified in place
        `On entry`: `xmiss`[j-1] must be set to the missing value, xm_j,
        to be associated with the jth variable in the array `x`, for
        those variables for which missing values are specified by means
        of the array `miss` (see [Accuracy]).

        `On exit`: the array `xmiss` is overwritten by the function, and
        the information it contained on entry is lost.

    mistyp : int
        Indicates the manner in which missing observations are to be
        treated.

        `mistyp` = 1

            A case is excluded if it contains a missing value for any of
            the variables 1,2,...,m.

        `mistyp` = 0

            A case is excluded if it contains a missing value for any of
            the p(<= m) variables specified in the array `kvar`.

    nvars : int
        p, the number of variables for which information is required.

    kvar : int, shape (`nvars`)
        `kvar`[j] must be set to the column number in `x` of the jth
        variable for which information is required, for j = 0,...,p-1.

    xbar : float, shape (`nvars`), modified in place
        `On exit`: the mean value, x__j, of the variable specified in
        `kvar`[j-1], for j = 1,2,...,p.

    std : float, shape (`nvars`), modified in place
        `On exit`: the standard deviation, s_j, of the variable
        specified in `kvar`[j-1], for j = 1,2,...,p.

    sspz : float, shape (`nvars`, `nvars`), modified in place
        `On exit`: `sspz`[j-1,k-1] is the cross-product about zero,
        S~_{jk}, for the variables specified in `kvar`[j-1] and
        `kvar`[k-1], for k = 1,2,...,p, for j = 1,2,...,p.

    rz : float, shape (`nvars`, `nvars`), modified in place
        `On exit`: `rz`[j-1,k-1] is the correlation-like coefficient,
        R~_{jk}, between the variables specified in `kvar`[j-1] and
        `kvar`[k-1], for k = 1,2,...,p, for j = 1,2,...,p.

    Returns
    -------
    ncases : int
        The number of cases actually used in the calculations (when
        cases involving missing values have been eliminated).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 2.

        (`errno` 2)
            On entry, `nvars` = *<value>* and `m` = *<value>*.

            Constraint: `nvars` >= 2 and `nvars` <= `m`.

        (`errno` 4)
            On entry, i = *<value>*, `kvar`[i-1] = *<value>* and `m` =
            *<value>*.

            Constraint: 1 <= `kvar`[i-1] <= `m`.

        (`errno` 5)
            On entry, `mistyp` = *<value>*.

            Constraint: `mistyp` = 0 or 1.

        (`errno` 6)
            After observations with missing values were omitted, no
            cases remained.

        (`errno` 7)
            After observations with missing values were omitted, only
            one case remained.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.coeffs_zero_subset_miss_case'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02blft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Input, x[ldx, m].
        utils._BYREF_INT, # Input, ldx.
        utils._EngineIntArrayType.c_type, # Input/Output, miss[m].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, xmiss[m].
        utils._BYREF_INT, # Input, mistyp.
        utils._BYREF_INT, # Input, nvars.
        utils._EngineIntArrayType.c_type, # Input, kvar[nvars].
        utils._EngineFloat64ArrayType.c_type, # Output, xbar[nvars].
        utils._EngineFloat64ArrayType.c_type, # Output, std[nvars].
        utils._EngineFloat64ArrayType.c_type, # Output, sspz[ldsspz, nvars].
        utils._BYREF_INT, # Input, ldsspz.
        utils._EngineFloat64ArrayType.c_type, # Output, rz[ldrz, nvars].
        utils._BYREF_INT, # Input, ldrz.
        utils._BYREF_INT, # Output, ncases.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((x, 'x'), (sspz, 'sspz'), (rz, 'rz')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _ldx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py, _m_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _miss_cls = utils._EngineIntArrayType(
        miss, {'fun_name': fun_name, 'entity': 'miss'},
        exp_shape=(_m_py,),
    )
    _miss = _miss_cls.to_ctypes()
    _xmiss_cls = utils._EngineFloat64ArrayType(
        xmiss, {'fun_name': fun_name, 'entity': 'xmiss'},
        exp_shape=(_m_py,),
    )
    _xmiss = _xmiss_cls.to_ctypes()
    _mistyp = utils._EngineIntScalarType(
        mistyp, {'fun_name': fun_name, 'entity': 'mistyp'},
    ).to_ctypes()
    _nvars = utils._EngineIntScalarType(
        nvars, {'fun_name': fun_name, 'entity': 'nvars'},
    ).to_ctypes()
    _nvars_py = utils._EngineIntScalarType.to_py(_nvars)
    _kvar = utils._EngineIntArrayType(
        kvar, {'fun_name': fun_name, 'entity': 'kvar'},
        exp_shape=(_nvars_py,),
    ).to_ctypes()
    _xbar_cls = utils._EngineFloat64ArrayType(
        xbar, {'fun_name': fun_name, 'entity': 'xbar'},
        exp_shape=(_nvars_py,),
        check_contents=False,
    )
    _xbar = _xbar_cls.to_ctypes(can_reference_contents=False)
    _std_cls = utils._EngineFloat64ArrayType(
        std, {'fun_name': fun_name, 'entity': 'std'},
        exp_shape=(_nvars_py,),
        check_contents=False,
    )
    _std = _std_cls.to_ctypes(can_reference_contents=False)
    _ldsspz_py = utils._EngineFloat64ArrayType.get_nd(sspz, {'fun_name': fun_name, 'entity': 'sspz'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldsspz = utils._EngineIntScalarType(
        _ldsspz_py, {'fun_name': fun_name, 'entity': '_ldsspz'},
    ).to_ctypes()
    _sspz_cls = utils._EngineFloat64ArrayType(
        sspz, {'fun_name': fun_name, 'entity': 'sspz'},
        exp_shape=(_nvars_py, _nvars_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _sspz = _sspz_cls.to_ctypes(can_reference_contents=False)
    _ldrz_py = utils._EngineFloat64ArrayType.get_nd(rz, {'fun_name': fun_name, 'entity': 'rz'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldrz = utils._EngineIntScalarType(
        _ldrz_py, {'fun_name': fun_name, 'entity': '_ldrz'},
    ).to_ctypes()
    _rz_cls = utils._EngineFloat64ArrayType(
        rz, {'fun_name': fun_name, 'entity': 'rz'},
        exp_shape=(_nvars_py, _nvars_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _rz = _rz_cls.to_ctypes(can_reference_contents=False)
    _ncases = utils._EngineIntScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _m,
        _x,
        _ldx,
        _miss,
        _xmiss,
        _mistyp,
        _nvars,
        _kvar,
        _xbar,
        _std,
        _sspz,
        _ldsspz,
        _rz,
        _ldrz,
        _ncases,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nvars = {:s} and m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nvars >= 2 and nvars <= m.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, kvar[i-1] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'm = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: 1 <= kvar[i-1] <= m.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mistyp = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mistyp = 0 or 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'After observations with missing values were omitted,',
                    'no cases remained.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'After observations with missing values were omitted,',
                    'only one case remained.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _miss_cls.refresh_py(_miss)
        _xmiss_cls.refresh_py(_xmiss)
        _xbar_cls.refresh_py(_xbar)
        _std_cls.refresh_py(_std)
        _sspz_cls.refresh_py(_sspz)
        _rz_cls.refresh_py(_rz)
        _ncases_py = utils._EngineIntScalarType.to_py(_ncases)
        return_data = _ncases_py
    _err_class.manage()
    return return_data

def coeffs_zero_subset_miss_pair(n, m, x, miss, xmiss, nvars, kvar, xbar, std, sspz, rz, cnt):
    u"""
    Correlation-like coefficients (about zero), subset of variables,
    pairwise treatment of missing values.

    ``coeffs_zero_subset_miss_pair`` computes means and standard
    deviations, sums of squares and cross-products about zero, and
    correlation-like coefficients for selected variables omitting cases
    with missing values from only those calculations involving the
    variables for which the values are missing.

    For full information please refer to the NAG Library document for
    g02bm

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02bmf.html

    Parameters
    ----------
    n : int
        n, the number of observations or cases.

    m : int
        m, the number of variables.

    x : float, shape (`n`, `m`)
        `x`[i-1,j-1] must be set to x_{ij}, the value of the ith
        observation on the jth variable, for j = 1,2,...,m, for i =
        1,2,...,n.

    miss : int, shape (`m`)
        `miss`[j-1] must be set equal to 1 if a missing value, xm_j, is
        to be specified for the jth variable in the array `x`, or set
        equal to 0 otherwise. Values of `miss` must be given for all m
        variables in the array `x`.

    xmiss : float, shape (`m`)
        `xmiss`[j-1] must be set to the missing value, xm_j, to be
        associated with the jth variable in the array `x`, for those
        variables for which missing values are specified by means of the
        array `miss` (see [Accuracy]).

    nvars : int
        p, the number of variables for which information is required.

    kvar : int, shape (`nvars`)
        `kvar`[j] must be set to the column number in `x` of the jth
        variable for which information is required, for j = 0,...,p-1.

    xbar : float, shape (`nvars`), modified in place
        `On exit`: the mean value, x__j, of the variable specified in
        `kvar`[j-1], for j = 1,2,...,p.

    std : float, shape (`nvars`), modified in place
        `On exit`: the standard deviation, s_j, of the variable
        specified in `kvar`[j-1], for j = 1,2,...,p.

    sspz : float, shape (`nvars`, `nvars`), modified in place
        `On exit`: `sspz`[j-1,k-1] is the cross-product about zero,
        S~_{jk}, for the variables specified in `kvar`[j-1] and
        `kvar`[k-1], for k = 1,2,...,p, for j = 1,2,...,p.

    rz : float, shape (`nvars`, `nvars`), modified in place
        `On exit`: `rz`[j-1,k-1] is the correlation-like coefficient,
        R~_{jk}, between the variables specified in `kvar`[j-1] and
        `kvar`[k-1], for k = 1,2,...,p, for j = 1,2,...,p.

    cnt : float, shape (`nvars`, `nvars`), modified in place
        `On exit`: `cnt`[j-1,k-1] is the number of cases, c_{jk},
        actually used in the calculation of the sum of cross-product and
        correlation-like coefficient for the variables specified in
        `kvar`[j-1] and `kvar`[k-1], for k = 1,2,...,p, for j =
        1,2,...,p.

    Returns
    -------
    ncases : int
        The minimum number of cases used in the calculation of any of
        the sums of squares and cross-products and correlation-like
        coefficients (when cases involving missing values have been
        eliminated).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 2.

        (`errno` 2)
            On entry, `nvars` = *<value>* and `m` = *<value>*.

            Constraint: `nvars` >= 2 and `nvars` <= `m`.

        (`errno` 4)
            On entry, i = *<value>*, `kvar`[i-1] = *<value>* and `m` =
            *<value>*.

            Constraint: 1 <= `kvar`[i-1] <= `m`.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 5)
            After observations with missing values were omitted, fewer
            than two cases remained for at least one pair of variables.
            (The pairs of variables involved can be determined by
            examination of the contents of the array `cnt`). All means,
            standard deviations, sums of squares and cross-products, and
            correlation-like coefficients based on two or more cases are
            returned by the function even if `errno` = 5.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.coeffs_zero_subset_miss_pair'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02bmft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Input, x[ldx, m].
        utils._BYREF_INT, # Input, ldx.
        utils._EngineIntArrayType.c_type, # Input, miss[m].
        utils._EngineFloat64ArrayType.c_type, # Input, xmiss[m].
        utils._BYREF_INT, # Input, nvars.
        utils._EngineIntArrayType.c_type, # Input, kvar[nvars].
        utils._EngineFloat64ArrayType.c_type, # Output, xbar[nvars].
        utils._EngineFloat64ArrayType.c_type, # Output, std[nvars].
        utils._EngineFloat64ArrayType.c_type, # Output, sspz[ldsspz, nvars].
        utils._BYREF_INT, # Input, ldsspz.
        utils._EngineFloat64ArrayType.c_type, # Output, rz[ldrz, nvars].
        utils._BYREF_INT, # Input, ldrz.
        utils._BYREF_INT, # Output, ncases.
        utils._EngineFloat64ArrayType.c_type, # Output, cnt[ldcnt, nvars].
        utils._BYREF_INT, # Input, ldcnt.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((x, 'x'), (sspz, 'sspz'), (rz, 'rz'), (cnt, 'cnt')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _ldx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py, _m_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _miss = utils._EngineIntArrayType(
        miss, {'fun_name': fun_name, 'entity': 'miss'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _xmiss = utils._EngineFloat64ArrayType(
        xmiss, {'fun_name': fun_name, 'entity': 'xmiss'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _nvars = utils._EngineIntScalarType(
        nvars, {'fun_name': fun_name, 'entity': 'nvars'},
    ).to_ctypes()
    _nvars_py = utils._EngineIntScalarType.to_py(_nvars)
    _kvar = utils._EngineIntArrayType(
        kvar, {'fun_name': fun_name, 'entity': 'kvar'},
        exp_shape=(_nvars_py,),
    ).to_ctypes()
    _xbar_cls = utils._EngineFloat64ArrayType(
        xbar, {'fun_name': fun_name, 'entity': 'xbar'},
        exp_shape=(_nvars_py,),
        check_contents=False,
    )
    _xbar = _xbar_cls.to_ctypes(can_reference_contents=False)
    _std_cls = utils._EngineFloat64ArrayType(
        std, {'fun_name': fun_name, 'entity': 'std'},
        exp_shape=(_nvars_py,),
        check_contents=False,
    )
    _std = _std_cls.to_ctypes(can_reference_contents=False)
    _ldsspz_py = utils._EngineFloat64ArrayType.get_nd(sspz, {'fun_name': fun_name, 'entity': 'sspz'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldsspz = utils._EngineIntScalarType(
        _ldsspz_py, {'fun_name': fun_name, 'entity': '_ldsspz'},
    ).to_ctypes()
    _sspz_cls = utils._EngineFloat64ArrayType(
        sspz, {'fun_name': fun_name, 'entity': 'sspz'},
        exp_shape=(_nvars_py, _nvars_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _sspz = _sspz_cls.to_ctypes(can_reference_contents=False)
    _ldrz_py = utils._EngineFloat64ArrayType.get_nd(rz, {'fun_name': fun_name, 'entity': 'rz'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldrz = utils._EngineIntScalarType(
        _ldrz_py, {'fun_name': fun_name, 'entity': '_ldrz'},
    ).to_ctypes()
    _rz_cls = utils._EngineFloat64ArrayType(
        rz, {'fun_name': fun_name, 'entity': 'rz'},
        exp_shape=(_nvars_py, _nvars_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _rz = _rz_cls.to_ctypes(can_reference_contents=False)
    _ncases = utils._EngineIntScalarType.empty_ctypes()
    _ldcnt_py = utils._EngineFloat64ArrayType.get_nd(cnt, {'fun_name': fun_name, 'entity': 'cnt'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldcnt = utils._EngineIntScalarType(
        _ldcnt_py, {'fun_name': fun_name, 'entity': '_ldcnt'},
    ).to_ctypes()
    _cnt_cls = utils._EngineFloat64ArrayType(
        cnt, {'fun_name': fun_name, 'entity': 'cnt'},
        exp_shape=(_nvars_py, _nvars_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _cnt = _cnt_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _m,
        _x,
        _ldx,
        _miss,
        _xmiss,
        _nvars,
        _kvar,
        _xbar,
        _std,
        _sspz,
        _ldsspz,
        _rz,
        _ldrz,
        _ncases,
        _cnt,
        _ldcnt,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nvars = {:s} and m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nvars >= 2 and nvars <= m.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s}, kvar[i-1] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'm = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: 1 <= kvar[i-1] <= m.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99992:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'After observations with missing values were omitted, fewer than two cases',
                    'remained for at least one pair of variables. (The pairs of variables',
                    'involved can be determined by examination of the contents of the array',
                    'cnt). All means, standard deviations, sums of squares and',
                    'cross-products, and correlation-like coefficients based on two or more cases',
                    'are returned by the function even if errno = 5.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _xbar_cls.refresh_py(_xbar)
        _std_cls.refresh_py(_std)
        _sspz_cls.refresh_py(_sspz)
        _rz_cls.refresh_py(_rz)
        _cnt_cls.refresh_py(_cnt)
        _ncases_py = utils._EngineIntScalarType.to_py(_ncases)
        return_data = _ncases_py
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = _collections.namedtuple(
                'nag_return_data',
                ['ncases'],
            )(
                _ncases_py,
            )
    _err_class.manage()
    return return_data

def coeffs_kspearman_overwrite(n, m, x, itype, rr):
    u"""
    Kendall/Spearman non-parametric rank correlation coefficients, no
    missing values, overwriting input data.

    ``coeffs_kspearman_overwrite`` computes Kendall and/or Spearman
    nonparametric rank correlation coefficients for a set of data; the
    data array is overwritten with the ranks of the observations.

    For full information please refer to the NAG Library document for
    g02bn

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02bnf.html

    Parameters
    ----------
    n : int
        n, the number of observations or cases.

    m : int
        m, the number of variables.

    x : float, shape (`n`, `m`), modified in place
        `On entry`: `x`[i-1,j-1] must be set to x_{ij}, the value of the
        ith observation on the jth variable, for j = 1,2,...,m, for i =
        1,2,...,n.

        `On exit`: `x`[i-1,j-1] contains the rank y_{ij} of the
        observation x_{ij}, for j = 1,2,...,m, for i = 1,2,...,n.

    itype : int
        The type of correlation coefficients which are to be calculated.

        `itype` = -1

            Only Kendall's tau coefficients are calculated.

        `itype` = 0

            Both Kendall's tau and Spearman's coefficients are
            calculated.

        `itype` = 1

            Only Spearman's coefficients are calculated.

    rr : float, shape (`m`, `m`), modified in place
        `On exit`: the requested correlation coefficients.

        If only Kendall's tau coefficients are requested (`itype` = -1),
        `rr`[j-1,k-1] contains Kendall's tau for the jth and kth
        variables.

        If only Spearman's coefficients are requested (`itype` = 1),
        `rr`[j-1,k-1] contains Spearman's rank correlation coefficient
        for the jth and kth variables.

        If both Kendall's tau and Spearman's coefficients are requested
        (`itype` = 0), the upper triangle of `rr` contains the Spearman
        coefficients and the lower triangle the Kendall coefficients.

        That is, for the jth and kth variables, where j is less than k,
        `rr`[j-1,k-1] contains the Spearman rank correlation
        coefficient, and `rr`[k-1,j-1] contains Kendall's tau, for k =
        1,2,...,m, for j = 1,2,...,m.

        (Diagonal terms, `rr`[j-1,j-1], are unity for all three values
        of `itype`.)

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 2.

        (`errno` 2)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 2.

        (`errno` 4)
            On entry, `itype` = *<value>*.

            Constraint: `itype` = -1 or 1.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Siegel, S, 1956, `Non-parametric Statistics for the Behavioral
    Sciences`, McGraw--Hill
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.coeffs_kspearman_overwrite'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02bnft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, x[ldx, m].
        utils._BYREF_INT, # Input, ldx.
        utils._BYREF_INT, # Input, itype.
        utils._EngineFloat64ArrayType.c_type, # Output, rr[ldrr, m].
        utils._BYREF_INT, # Input, ldrr.
        utils._EngineIntArrayType.c_type, # Workspace, kworka.
        utils._EngineIntArrayType.c_type, # Workspace, kworkb.
        utils._EngineFloat64ArrayType.c_type, # Workspace, work1.
        utils._EngineFloat64ArrayType.c_type, # Workspace, work2.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((x, 'x'), (rr, 'rr')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _ldx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py, _m_py),
        mutual_sorder=_m_sorder,
    )
    _x = _x_cls.to_ctypes()
    _itype = utils._EngineIntScalarType(
        itype, {'fun_name': fun_name, 'entity': 'itype'},
    ).to_ctypes()
    _ldrr_py = utils._EngineFloat64ArrayType.get_nd(rr, {'fun_name': fun_name, 'entity': 'rr'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldrr = utils._EngineIntScalarType(
        _ldrr_py, {'fun_name': fun_name, 'entity': '_ldrr'},
    ).to_ctypes()
    _rr_cls = utils._EngineFloat64ArrayType(
        rr, {'fun_name': fun_name, 'entity': 'rr'},
        exp_shape=(_m_py, _m_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _rr = _rr_cls.to_ctypes(can_reference_contents=False)
    _kworka = utils._EngineIntArrayType.empty_ctypes(0)
    _kworkb = utils._EngineIntArrayType.empty_ctypes(0)
    _work1 = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _work2 = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _m,
        _x,
        _ldx,
        _itype,
        _rr,
        _ldrr,
        _kworka,
        _kworkb,
        _work1,
        _work2,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, itype = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: itype = -1 or 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        _rr_cls.refresh_py(_rr)
        return_data = None
    _err_class.manage()
    return return_data

def coeffs_kspearman_miss_case_overwrite(n, m, x, miss, xmiss, itype, rr, incase):
    u"""
    Kendall/Spearman non-parametric rank correlation coefficients,
    casewise treatment of missing values, overwriting input data.

    ``coeffs_kspearman_miss_case_overwrite`` computes Kendall and/or
    Spearman nonparametric rank correlation coefficients for a set of
    data omitting completely any cases with a missing observation for
    any variable; the data array is overwritten with the ranks of the
    observations.

    For full information please refer to the NAG Library document for
    g02bp

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02bpf.html

    Parameters
    ----------
    n : int
        n, the number of observations or cases.

    m : int
        m, the number of variables.

    x : float, shape (`n`, `m`), modified in place
        `On entry`: `x`[i-1,j-1] must be set to x_{ij}, the value of the
        ith observation on the jth variable, for j = 1,2,...,m, for i =
        1,2,...,n.

        `On exit`: `x`[i-1,j-1] contains the rank y_{ij} of the
        observation x_{ij}, for j = 1,2,...,m, for i = 1,2,...,n. (For
        those observations containing missing values, and therefore
        excluded from the calculation, y_{ij} = 0, for j = 1,2,...,m.)

    miss : int, shape (`m`), modified in place
        `On entry`: `miss`[j-1] must be set to 1 if a missing value,
        xm_j, is to be specified for the jth variable in the array `x`,
        or set equal to 0 otherwise. Values of `miss` must be given for
        all m variables in the array `x`.

        `On exit`: the array `miss` is overwritten by the function, and
        the information it contained on entry is lost.

    xmiss : float, shape (`m`), modified in place
        `On entry`: `xmiss`[j-1] must be set to the missing value, xm_j,
        to be associated with the jth variable in the array `x`, for
        those variables for which missing values are specified by means
        of the array `miss` (see [Accuracy]).

        `On exit`: the array `xmiss` is overwritten by the function, and
        the information it contained on entry is lost.

    itype : int
        The type of correlation coefficients which are to be calculated.

        `itype` = -1

            Only Kendall's tau coefficients are calculated.

        `itype` = 0

            Both Kendall's tau and Spearman's coefficients are
            calculated.

        `itype` = 1

            Only Spearman's coefficients are calculated.

    rr : float, shape (`m`, `m`), modified in place
        `On exit`: the requested correlation coefficients.

        If only Kendall's tau coefficients are requested (`itype` = -1),
        `rr`[j-1,k-1] contains Kendall's tau for the jth and kth
        variables.

        If only Spearman's coefficients are requested (`itype` = 1),
        `rr`[j-1,k-1] contains Spearman's rank correlation coefficient
        for the jth and kth variables.

        If both Kendall's tau and Spearman's coefficients are requested
        (`itype` = 0), the upper triangle of `rr` contains the Spearman
        coefficients and the lower triangle the Kendall coefficients.

        That is, for the jth and kth variables, where j is less than k,
        `rr`[j-1,k-1] contains the Spearman rank correlation
        coefficient, and `rr`[k-1,j-1] contains Kendall's tau, for k =
        1,2,...,m, for j = 1,2,...,m.

        (Diagonal terms, `rr`[j-1,j-1], are unity for all three values
        of `itype`.)

    incase : int, shape (`n`), modified in place
        `On exit`: `incase`[i-1] holds the value 1 if the ith case was
        included in the calculations, and the value 0 if the ith case
        contained a missing value for at least one variable. That is,
        `incase`[i-1] = w_i (see [Description]), for i = 1,2,...,n.

    Returns
    -------
    ncases : int
        The number of cases, n_c, actually used in the calculations
        (when cases involving missing values have been eliminated).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 2.

        (`errno` 2)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 2.

        (`errno` 4)
            On entry, `itype` = *<value>*.

            Constraint: `itype` = -1 or 1.

        (`errno` 5)
            After observations with missing values were omitted, fewer
            than two cases remained.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Siegel, S, 1956, `Non-parametric Statistics for the Behavioral
    Sciences`, McGraw--Hill
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.coeffs_kspearman_miss_case_overwrite'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02bpft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, x[ldx, m].
        utils._BYREF_INT, # Input, ldx.
        utils._EngineIntArrayType.c_type, # Input/Output, miss[m].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, xmiss[m].
        utils._BYREF_INT, # Input, itype.
        utils._EngineFloat64ArrayType.c_type, # Output, rr[ldrr, m].
        utils._BYREF_INT, # Input, ldrr.
        utils._BYREF_INT, # Output, ncases.
        utils._EngineIntArrayType.c_type, # Output, incase[n].
        utils._EngineIntArrayType.c_type, # Workspace, kworka.
        utils._EngineIntArrayType.c_type, # Workspace, kworkb.
        utils._EngineIntArrayType.c_type, # Workspace, kworkc.
        utils._EngineFloat64ArrayType.c_type, # Workspace, work1.
        utils._EngineFloat64ArrayType.c_type, # Workspace, work2.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((x, 'x'), (rr, 'rr')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _ldx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py, _m_py),
        mutual_sorder=_m_sorder,
    )
    _x = _x_cls.to_ctypes()
    _miss_cls = utils._EngineIntArrayType(
        miss, {'fun_name': fun_name, 'entity': 'miss'},
        exp_shape=(_m_py,),
    )
    _miss = _miss_cls.to_ctypes()
    _xmiss_cls = utils._EngineFloat64ArrayType(
        xmiss, {'fun_name': fun_name, 'entity': 'xmiss'},
        exp_shape=(_m_py,),
    )
    _xmiss = _xmiss_cls.to_ctypes()
    _itype = utils._EngineIntScalarType(
        itype, {'fun_name': fun_name, 'entity': 'itype'},
    ).to_ctypes()
    _ldrr_py = utils._EngineFloat64ArrayType.get_nd(rr, {'fun_name': fun_name, 'entity': 'rr'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldrr = utils._EngineIntScalarType(
        _ldrr_py, {'fun_name': fun_name, 'entity': '_ldrr'},
    ).to_ctypes()
    _rr_cls = utils._EngineFloat64ArrayType(
        rr, {'fun_name': fun_name, 'entity': 'rr'},
        exp_shape=(_m_py, _m_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _rr = _rr_cls.to_ctypes(can_reference_contents=False)
    _ncases = utils._EngineIntScalarType.empty_ctypes()
    _incase_cls = utils._EngineIntArrayType(
        incase, {'fun_name': fun_name, 'entity': 'incase'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _incase = _incase_cls.to_ctypes(can_reference_contents=False)
    _kworka = utils._EngineIntArrayType.empty_ctypes(0)
    _kworkb = utils._EngineIntArrayType.empty_ctypes(0)
    _kworkc = utils._EngineIntArrayType.empty_ctypes(0)
    _work1 = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _work2 = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _m,
        _x,
        _ldx,
        _miss,
        _xmiss,
        _itype,
        _rr,
        _ldrr,
        _ncases,
        _incase,
        _kworka,
        _kworkb,
        _kworkc,
        _work1,
        _work2,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, itype = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: itype = -1 or 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'After observations with missing values were omitted,',
                    'fewer than two cases remained.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        _miss_cls.refresh_py(_miss)
        _xmiss_cls.refresh_py(_xmiss)
        _rr_cls.refresh_py(_rr)
        _incase_cls.refresh_py(_incase)
        _ncases_py = utils._EngineIntScalarType.to_py(_ncases)
        return_data = _ncases_py
    _err_class.manage()
    return return_data

def coeffs_kspearman(n, m, x, itype, rr):
    u"""
    Kendall/Spearman non-parametric rank correlation coefficients, no
    missing values, preserving input data.

    ``coeffs_kspearman`` computes Kendall and/or Spearman nonparametric
    rank correlation coefficients for a set of data; the data array is
    preserved, and the ranks of the observations are not available on
    exit from the function.

    For full information please refer to the NAG Library document for
    g02bq

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02bqf.html

    Parameters
    ----------
    n : int
        n, the number of observations or cases.

    m : int
        m, the number of variables.

    x : float, shape (`n`, `m`)
        `x`[i-1,j-1] must be set to data value x_{ij}, the value of the
        ith observation on the jth variable, for j = 1,2,...,m, for i =
        1,2,...,n.

    itype : int
        The type of correlation coefficients which are to be calculated.

        `itype` = -1

            Only Kendall's tau coefficients are calculated.

        `itype` = 0

            Both Kendall's tau and Spearman's coefficients are
            calculated.

        `itype` = 1

            Only Spearman's coefficients are calculated.

    rr : float, shape (`m`, `m`), modified in place
        `On exit`: the requested correlation coefficients.

        If only Kendall's tau coefficients are requested (`itype` = -1),
        `rr`[j-1,k-1] contains Kendall's tau for the jth and kth
        variables.

        If only Spearman's coefficients are requested (`itype` = 1),
        `rr`[j-1,k-1] contains Spearman's rank correlation coefficient
        for the jth and kth variables.

        If both Kendall's tau and Spearman's coefficients are requested
        (`itype` = 0), the upper triangle of `rr` contains the Spearman
        coefficients and the lower triangle the Kendall coefficients.

        That is, for the jth and kth variables, where j is less than k,
        `rr`[j-1,k-1] contains the Spearman rank correlation
        coefficient, and `rr`[k-1,j-1] contains Kendall's tau, for k =
        1,2,...,m, for j = 1,2,...,m.

        (Diagonal terms, `rr`[j-1,j-1], are unity for all three values
        of `itype`.)

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 2.

        (`errno` 2)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 2.

        (`errno` 4)
            On entry, `itype` = *<value>*.

            Constraint: `itype` = -1 or 1.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Siegel, S, 1956, `Non-parametric Statistics for the Behavioral
    Sciences`, McGraw--Hill
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.coeffs_kspearman'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02bqft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Input, x[ldx, m].
        utils._BYREF_INT, # Input, ldx.
        utils._BYREF_INT, # Input, itype.
        utils._EngineFloat64ArrayType.c_type, # Output, rr[ldrr, m].
        utils._BYREF_INT, # Input, ldrr.
        utils._EngineIntArrayType.c_type, # Workspace, kworka.
        utils._EngineIntArrayType.c_type, # Workspace, kworkb.
        utils._EngineFloat64ArrayType.c_type, # Workspace, work1.
        utils._EngineFloat64ArrayType.c_type, # Workspace, work2.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((x, 'x'), (rr, 'rr')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _ldx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py, _m_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _itype = utils._EngineIntScalarType(
        itype, {'fun_name': fun_name, 'entity': 'itype'},
    ).to_ctypes()
    _ldrr_py = utils._EngineFloat64ArrayType.get_nd(rr, {'fun_name': fun_name, 'entity': 'rr'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldrr = utils._EngineIntScalarType(
        _ldrr_py, {'fun_name': fun_name, 'entity': '_ldrr'},
    ).to_ctypes()
    _rr_cls = utils._EngineFloat64ArrayType(
        rr, {'fun_name': fun_name, 'entity': 'rr'},
        exp_shape=(_m_py, _m_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _rr = _rr_cls.to_ctypes(can_reference_contents=False)
    _kworka = utils._EngineIntArrayType.empty_ctypes(0)
    _kworkb = utils._EngineIntArrayType.empty_ctypes(0)
    _work1 = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _work2 = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _m,
        _x,
        _ldx,
        _itype,
        _rr,
        _ldrr,
        _kworka,
        _kworkb,
        _work1,
        _work2,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, itype = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: itype = -1 or 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _rr_cls.refresh_py(_rr)
        return_data = None
    _err_class.manage()
    return return_data

def coeffs_kspearman_miss_case(n, m, x, miss, xmiss, itype, rr, incase):
    u"""
    Kendall/Spearman non-parametric rank correlation coefficients,
    casewise treatment of missing values, preserving input data.

    ``coeffs_kspearman_miss_case`` computes Kendall and/or Spearman
    nonparametric rank correlation coefficients for a set of data,
    omitting completely any cases with a missing observation for any
    variable; the data array is preserved, and the ranks of the
    observations are not available on exit from the function.

    For full information please refer to the NAG Library document for
    g02br

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02brf.html

    Parameters
    ----------
    n : int
        n, the number of observations or cases.

    m : int
        m, the number of variables.

    x : float, shape (`n`, `m`)
        `x`[i-1,j-1] must be set to x_{ij}, the value of the ith
        observation on the jth variable, where i = 1,2,...,n and j =
        1,2,...,m.

    miss : int, shape (`m`), modified in place
        `On entry`: `miss`[j-1] must be set equal to 1 if a missing
        value, xm_j, is to be specified for the jth variable in the
        array `x`, or set equal to 0 otherwise. Values of `miss` must be
        given for all m variables in the array `x`.

        `On exit`: the array `miss` is overwritten by the function, and
        the information it contained on entry is lost.

    xmiss : float, shape (`m`), modified in place
        `On entry`: `xmiss`[j-1] must be set to the missing value, xm_j,
        to be associated with the jth variable in the array `x`, for
        those variables for which missing values are specified by means
        of the array `miss` (see [Accuracy]).

        `On exit`: the array `xmiss` is overwritten by the function, and
        the information it contained on entry is lost.

    itype : int
        The type of correlation coefficients which are to be calculated.

        `itype` = -1

            Only Kendall's tau coefficients are calculated.

        `itype` = 0

            Both Kendall's tau and Spearman's coefficients are
            calculated.

        `itype` = 1

            Only Spearman's coefficients are calculated.

    rr : float, shape (`m`, `m`), modified in place
        `On exit`: the requested correlation coefficients.

        If only Kendall's tau coefficients are requested (`itype` = -1),
        `rr`[j-1,k-1] contains Kendall's tau for the jth and kth
        variables.

        If only Spearman's coefficients are requested (`itype` = 1),
        `rr`[j-1,k-1] contains Spearman's rank correlation coefficient
        for the jth and kth variables.

        If both Kendall's tau and Spearman's coefficients are requested
        (`itype` = 0), the upper triangle of `rr` contains the Spearman
        coefficients and the lower triangle the Kendall coefficients.

        That is, for the jth and kth variables, where j is less than k,
        `rr`[j-1,k-1] contains the Spearman rank correlation
        coefficient, and `rr`[k-1,j-1] contains Kendall's tau, for k =
        1,2,...,m, for j = 1,2,...,m.

        (Diagonal terms, `rr`[j-1,j-1], are unity for all three values
        of `itype`.)

    incase : int, shape (`n`), modified in place
        `On exit`: `incase`[i-1] holds the value 1 if the ith case was
        included in the calculations, and the value 0 if the ith case
        contained a missing value for at least one variable. That is,
        `incase`[i-1] = w_i (see [Description]), for i = 1,2,...,n.

    Returns
    -------
    ncases : int
        The number of cases, n_c, actually used in the calculations
        (when cases involving missing values have been eliminated).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 2.

        (`errno` 2)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 2.

        (`errno` 4)
            On entry, `itype` = *<value>*.

            Constraint: `itype` = -1 or 1.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 5)
            After observations with missing values were omitted, fewer
            than two cases remained.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Siegel, S, 1956, `Non-parametric Statistics for the Behavioral
    Sciences`, McGraw--Hill

    See Also
    --------
    :meth:`naginterfaces.library.examples.correg.coeffs_kspearman_miss_case_ex.main`
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.coeffs_kspearman_miss_case'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02brft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Input, x[ldx, m].
        utils._BYREF_INT, # Input, ldx.
        utils._EngineIntArrayType.c_type, # Input/Output, miss[m].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, xmiss[m].
        utils._BYREF_INT, # Input, itype.
        utils._EngineFloat64ArrayType.c_type, # Output, rr[ldrr, m].
        utils._BYREF_INT, # Input, ldrr.
        utils._BYREF_INT, # Output, ncases.
        utils._EngineIntArrayType.c_type, # Output, incase[n].
        utils._EngineIntArrayType.c_type, # Workspace, kworka.
        utils._EngineIntArrayType.c_type, # Workspace, kworkb.
        utils._EngineIntArrayType.c_type, # Workspace, kworkc.
        utils._EngineFloat64ArrayType.c_type, # Workspace, work1.
        utils._EngineFloat64ArrayType.c_type, # Workspace, work2.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((x, 'x'), (rr, 'rr')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _ldx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py, _m_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _miss_cls = utils._EngineIntArrayType(
        miss, {'fun_name': fun_name, 'entity': 'miss'},
        exp_shape=(_m_py,),
    )
    _miss = _miss_cls.to_ctypes()
    _xmiss_cls = utils._EngineFloat64ArrayType(
        xmiss, {'fun_name': fun_name, 'entity': 'xmiss'},
        exp_shape=(_m_py,),
    )
    _xmiss = _xmiss_cls.to_ctypes()
    _itype = utils._EngineIntScalarType(
        itype, {'fun_name': fun_name, 'entity': 'itype'},
    ).to_ctypes()
    _ldrr_py = utils._EngineFloat64ArrayType.get_nd(rr, {'fun_name': fun_name, 'entity': 'rr'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldrr = utils._EngineIntScalarType(
        _ldrr_py, {'fun_name': fun_name, 'entity': '_ldrr'},
    ).to_ctypes()
    _rr_cls = utils._EngineFloat64ArrayType(
        rr, {'fun_name': fun_name, 'entity': 'rr'},
        exp_shape=(_m_py, _m_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _rr = _rr_cls.to_ctypes(can_reference_contents=False)
    _ncases = utils._EngineIntScalarType.empty_ctypes()
    _incase_cls = utils._EngineIntArrayType(
        incase, {'fun_name': fun_name, 'entity': 'incase'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _incase = _incase_cls.to_ctypes(can_reference_contents=False)
    _kworka = utils._EngineIntArrayType.empty_ctypes(0)
    _kworkb = utils._EngineIntArrayType.empty_ctypes(0)
    _kworkc = utils._EngineIntArrayType.empty_ctypes(0)
    _work1 = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _work2 = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _m,
        _x,
        _ldx,
        _miss,
        _xmiss,
        _itype,
        _rr,
        _ldrr,
        _ncases,
        _incase,
        _kworka,
        _kworkb,
        _kworkc,
        _work1,
        _work2,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, itype = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: itype = -1 or 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99994:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'After observations with missing values were omitted,',
                    'fewer than two cases remained.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _miss_cls.refresh_py(_miss)
        _xmiss_cls.refresh_py(_xmiss)
        _rr_cls.refresh_py(_rr)
        _incase_cls.refresh_py(_incase)
        _ncases_py = utils._EngineIntScalarType.to_py(_ncases)
        return_data = _ncases_py
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = _collections.namedtuple(
                'nag_return_data',
                ['ncases'],
            )(
                _ncases_py,
            )
    _err_class.manage()
    return return_data

def coeffs_kspearman_miss_pair(n, m, x, miss, xmiss, itype, rr, cnt):
    u"""
    Kendall/Spearman non-parametric rank correlation coefficients,
    pairwise treatment of missing values.

    ``coeffs_kspearman_miss_pair`` computes Kendall and/or Spearman
    nonparametric rank correlation coefficients for a set of data
    omitting cases with missing values from only those calculations
    involving the variables for which the values are missing; the data
    array is preserved, and the ranks of the observations are not
    available on exit from the function.

    For full information please refer to the NAG Library document for
    g02bs

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02bsf.html

    Parameters
    ----------
    n : int
        n, the number of observations or cases.

    m : int
        m, the number of variables.

    x : float, shape (`n`, `m`)
        `x`[i-1,j-1] must be set to x_{ij}, the value of the ith
        observation on the jth variable, for j = 1,2,...,m, for i =
        1,2,...,n.

    miss : int, shape (`m`)
        `miss`[j-1] must be set equal to 1 if a missing value, xm_j, is
        to be specified for the jth variable in the array `x`, or set
        equal to 0 otherwise. Values of `miss` must be given for all m
        variables in the array `x`.

    xmiss : float, shape (`m`)
        `xmiss`[j-1] must be set to the missing value, xm_j, to be
        associated with the jth variable in the array `x`, for those
        variables for which missing values are specified by means of the
        array `miss` (see [Accuracy]).

    itype : int
        The type of correlation coefficients which are to be calculated.

        `itype` = -1

            Only Kendall's tau coefficients are calculated.

        `itype` = 0

            Both Kendall's tau and Spearman's coefficients are
            calculated.

        `itype` = 1

            Only Spearman's coefficients are calculated.

    rr : float, shape (`m`, `m`), modified in place
        `On exit`: the requested correlation coefficients.

        If only Kendall's tau coefficients are requested (`itype` = -1),
        `rr`[j-1,k-1] contains Kendall's tau for the jth and kth
        variables.

        If only Spearman's coefficients are requested (`itype` = 1),
        `rr`[j-1,k-1] contains Spearman's rank correlation coefficient
        for the jth and kth variables.

        If both Kendall's tau and Spearman's coefficients are requested
        (`itype` = 0), the upper triangle of `rr` contains the Spearman
        coefficients and the lower triangle the Kendall coefficients.

        That is, for the jth and kth variables, where j is less than k,
        `rr`[j-1,k-1] contains the Spearman rank correlation
        coefficient, and `rr`[k-1,j-1] contains Kendall's tau, for k =
        1,2,...,m, for j = 1,2,...,m.

        (Diagonal terms, `rr`[j-1,j-1], are unity for all three values
        of `itype`.)

    cnt : float, shape (`m`, `m`), modified in place
        `On exit`: the number of cases, n_{jk}, actually used in the
        calculation of the rank correlation coefficient for the jth and
        kth variables, for k = 1,2,...,m, for j = 1,2,...,m.

    Returns
    -------
    ncases : int
        The minimum number of cases used in the calculation of any of
        the correlation coefficients (when cases involving missing
        values have been eliminated).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 2.

        (`errno` 2)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 2.

        (`errno` 4)
            On entry, `itype` = *<value>*.

            Constraint: `itype` = -1 or 1.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 5)
            After observations with missing values were omitted, fewer
            than two cases remained for at least one pair of variables.
            (The pairs of variables involved can be determined by
            examination of the contents of the array `cnt`). All means,
            standard deviations, sums of squares and cross-products, and
            correlation-like coefficients based on two or more cases are
            returned by the function even if `errno` = 5.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Siegel, S, 1956, `Non-parametric Statistics for the Behavioral
    Sciences`, McGraw--Hill
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.coeffs_kspearman_miss_pair'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02bsft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Input, x[ldx, m].
        utils._BYREF_INT, # Input, ldx.
        utils._EngineIntArrayType.c_type, # Input, miss[m].
        utils._EngineFloat64ArrayType.c_type, # Input, xmiss[m].
        utils._BYREF_INT, # Input, itype.
        utils._EngineFloat64ArrayType.c_type, # Output, rr[ldrr, m].
        utils._BYREF_INT, # Input, ldrr.
        utils._BYREF_INT, # Output, ncases.
        utils._EngineFloat64ArrayType.c_type, # Output, cnt[ldcnt, m].
        utils._BYREF_INT, # Input, ldcnt.
        utils._EngineIntArrayType.c_type, # Workspace, kworka.
        utils._EngineIntArrayType.c_type, # Workspace, kworkb.
        utils._EngineIntArrayType.c_type, # Workspace, kworkc.
        utils._EngineIntArrayType.c_type, # Workspace, kworkd.
        utils._EngineFloat64ArrayType.c_type, # Workspace, work1.
        utils._EngineFloat64ArrayType.c_type, # Workspace, work2.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((x, 'x'), (rr, 'rr'), (cnt, 'cnt')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _ldx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py, _m_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _miss = utils._EngineIntArrayType(
        miss, {'fun_name': fun_name, 'entity': 'miss'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _xmiss = utils._EngineFloat64ArrayType(
        xmiss, {'fun_name': fun_name, 'entity': 'xmiss'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _itype = utils._EngineIntScalarType(
        itype, {'fun_name': fun_name, 'entity': 'itype'},
    ).to_ctypes()
    _ldrr_py = utils._EngineFloat64ArrayType.get_nd(rr, {'fun_name': fun_name, 'entity': 'rr'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldrr = utils._EngineIntScalarType(
        _ldrr_py, {'fun_name': fun_name, 'entity': '_ldrr'},
    ).to_ctypes()
    _rr_cls = utils._EngineFloat64ArrayType(
        rr, {'fun_name': fun_name, 'entity': 'rr'},
        exp_shape=(_m_py, _m_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _rr = _rr_cls.to_ctypes(can_reference_contents=False)
    _ncases = utils._EngineIntScalarType.empty_ctypes()
    _ldcnt_py = utils._EngineFloat64ArrayType.get_nd(cnt, {'fun_name': fun_name, 'entity': 'cnt'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldcnt = utils._EngineIntScalarType(
        _ldcnt_py, {'fun_name': fun_name, 'entity': '_ldcnt'},
    ).to_ctypes()
    _cnt_cls = utils._EngineFloat64ArrayType(
        cnt, {'fun_name': fun_name, 'entity': 'cnt'},
        exp_shape=(_m_py, _m_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _cnt = _cnt_cls.to_ctypes(can_reference_contents=False)
    _kworka = utils._EngineIntArrayType.empty_ctypes(0)
    _kworkb = utils._EngineIntArrayType.empty_ctypes(0)
    _kworkc = utils._EngineIntArrayType.empty_ctypes(0)
    _kworkd = utils._EngineIntArrayType.empty_ctypes(0)
    _work1 = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _work2 = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _m,
        _x,
        _ldx,
        _miss,
        _xmiss,
        _itype,
        _rr,
        _ldrr,
        _ncases,
        _cnt,
        _ldcnt,
        _kworka,
        _kworkb,
        _kworkc,
        _kworkd,
        _work1,
        _work2,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, itype = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: itype = -1 or 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99993:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'After observations with missing values were omitted, fewer than two cases',
                    'remained for at least one pair of variables. (The pairs of variables',
                    'involved can be determined by examination of the contents of the array',
                    'cnt). All means, standard deviations, sums of squares and',
                    'cross-products, and correlation-like coefficients based on two or more cases',
                    'are returned by the function even if errno = 5.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _rr_cls.refresh_py(_rr)
        _cnt_cls.refresh_py(_cnt)
        _ncases_py = utils._EngineIntScalarType.to_py(_ncases)
        return_data = _ncases_py
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = _collections.namedtuple(
                'nag_return_data',
                ['ncases'],
            )(
                _ncases_py,
            )
    _err_class.manage()
    return return_data

def ssqmat_update(mean, m, wt, x, incx, sw, xbar, c):
    u"""
    Update a weighted sum of squares matrix with a new observation.

    ``ssqmat_update`` updates the sample means and sums of squares and
    cross-products, or sums of squares and cross-products of deviations
    about the mean, for a new observation.
    The data may be weighted.

    For full information please refer to the NAG Library document for
    g02bt

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02btf.html

    Parameters
    ----------
    mean : str, length 1
        Indicates whether ``ssqmat_update`` is to calculate sums of
        squares and cross-products, or sums of squares and
        cross-products of deviations about the mean.

        `mean` = 'M'

            The sums of squares and cross-products of deviations about
            the mean are calculated.

        `mean` = 'Z'

            The sums of squares and cross-products are calculated.

    m : int
        m, the number of variables.

    wt : float
        The weight to use for the current observation, w_i.

        For unweighted means and cross-products set `wt` = 1.0.

        The use of a suitable negative value of `wt`, e.g., -w_i will
        have the effect of deleting the observation.

    x : float, shape (`m`*`incx`)
        `x`[(j-1)*`incx`] must contain the value of the jth variable for
        the current observation, j = 1,2,...,m.

    incx : int
        The increment of `x`. Two situations are common.

        If `incx` = 1, the data values are to be found in consecutive
        locations in `x`, i.e., in a column.

        If `incx` = ldx, for some positive integer ldx, the data values
        are to be found as a row of an array with first dimension ldx.

    sw : float
        The sum of weights for the previous observations, W_{i-1}.

        `sw` = 0.0

            The update procedure is initialized.

        `sw`+`wt` = 0.0

            All elements of `xbar` and `c` are set to zero.

    xbar : float, shape (`m`), modified in place
        `On entry`: if `sw` = 0.0, `xbar` is initialized, otherwise
        `xbar`[j-1] must contain the weighted mean of the jth variable
        for the previous (i-1) observations, x__j(i-1), for j =
        1,2,...,m.

        `On exit`: `xbar`[j-1] contains the weighted mean of the jth
        variable, x__j(i), for j = 1,2,...,m.

    c : float, shape ((`m`*`m`+`m`)/2), modified in place
        `On entry`: if `sw` != 0.0, `c` must contain the upper
        triangular part of the matrix of weighted sums of squares and
        cross-products or weighted sums of squares and cross-products of
        deviations about the mean. It is stored packed form by column,
        i.e., the cross-product between the jth and kth variable, k >=
        j, is stored in `c`[k*(k-1)/2+j-1].

        `On exit`: the update sums of squares and cross-products stored
        as on input.

    Returns
    -------
    sw : float
        Contains the updated sum of weights, W_i.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `incx` = *<value>*.

            Constraint: `incx` >= 1.

        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 2)
            On entry, `sw` = *<value>*.

            Constraint: `sw` >= 0.0.

        (`errno` 3)
            On entry, (`sw`+`wt`) = *<value>*.

            Constraint: (`sw`+`wt`) >= 0.0.

        (`errno` 4)
            On entry, `mean` = *<value>*.

            Constraint: `mean` = 'M' or 'Z'.

    References
    ----------
    Chan, T F, Golub, G H and Leveque, R J, 1982, `Updating Formulae and
    a Pairwise Algorithm for Computing Sample Variances`, Compstat,
    Physica-Verlag

    West, D H D, 1979, `Updating mean and variance estimates: An
    improved method`, Comm. ACM (22), 532--555
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.ssqmat_update'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02btft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, mean(len=1).
        utils._BYREF_INT, # Input, m.
        utils._BYREF_FLOAT64, # Input, wt.
        utils._EngineFloat64ArrayType.c_type, # Input, x[m*incx].
        utils._BYREF_INT, # Input, incx.
        utils._BYREF_FLOAT64, # Input/Output, sw.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, xbar[m].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, c[(m*m+m)//2].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of mean.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _mean = utils._EngineCharScalarType(
        mean, {'fun_name': fun_name, 'entity': 'mean'}, exp_chrlen=1,
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _wt = utils._EngineFloat64ScalarType(
        wt, {'fun_name': fun_name, 'entity': 'wt'},
    ).to_ctypes()
    _incx = utils._EngineIntScalarType(
        incx, {'fun_name': fun_name, 'entity': 'incx'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _incx_py = utils._EngineIntScalarType.to_py(_incx)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_m_py*_incx_py,),
    ).to_ctypes()
    _sw = utils._EngineFloat64ScalarType(
        sw, {'fun_name': fun_name, 'entity': 'sw'},
    ).to_ctypes()
    _xbar_cls = utils._EngineFloat64ArrayType(
        xbar, {'fun_name': fun_name, 'entity': 'xbar'},
        exp_shape=(_m_py,),
    )
    _xbar = _xbar_cls.to_ctypes()
    _c_cls = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=((_m_py*_m_py+_m_py)//2,),
    )
    _c = _c_cls.to_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _mean,
        _m,
        _wt,
        _x,
        _incx,
        _sw,
        _xbar,
        _c,
        _errbuf,
        _ierr,
        _mean._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, incx = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: incx >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, sw = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: sw >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, (sw+wt) = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: (sw+wt) >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mean = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mean = \'M\' or \'Z\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _sw_py = utils._EngineFloat64ScalarType.to_py(_sw)
        _xbar_cls.refresh_py(_xbar)
        _c_cls.refresh_py(_c)
        return_data = _sw_py
    _err_class.manage()
    return return_data

def ssqmat(mean, n, m, x, wt, wmean, c):
    u"""
    Computes a weighted sum of squares matrix.

    ``ssqmat`` calculates the sample means and sums of squares and
    cross-products, or sums of squares and cross-products of deviations
    from the mean, in a single pass for a set of data.
    The data may be weighted.

    For full information please refer to the NAG Library document for
    g02bu

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02buf.html

    Parameters
    ----------
    mean : str, length 1
        Indicates whether ``ssqmat`` is to calculate sums of squares and
        cross-products, or sums of squares and cross-products of
        deviations about the mean.

        `mean` = 'M'

            The sums of squares and cross-products of deviations about
            the mean are calculated.

        `mean` = 'Z'

            The sums of squares and cross-products are calculated.

    n : int
        n, the number of observations in the dataset.

    m : int
        m, the number of variables.

    x : float, shape (`n`, `m`)
        `x`[i-1,j-1] must contain the ith observation on the jth
        variable, for j = 1,2,...,m, for i = 1,2,...,n.

    wt : None or float, shape (`n`)
        The optional weights of each observation. If weights are not
        provided then `wt` must be set to None, otherwise `wt`[i-1] must
        contain the weight for the ith observation.

    wmean : float, shape (`m`), modified in place
        `On exit`: the sample means. `wmean`[j-1] contains the mean for
        the jth variable.

    c : float, shape ((`m`*`m`+`m`)/2), modified in place
        `On exit`: the cross-products.

        If `mean` = 'M', `c` contains the upper triangular part of the
        matrix of (weighted) sums of squares and cross-products of
        deviations about the mean.

        If `mean` = 'Z', `c` contains the upper triangular part of the
        matrix of (weighted) sums of squares and cross-products.

        These are stored packed by columns, i.e., the cross-product
        between the jth and kth variable, k >= j, is stored in
        `c`[k*(k-1)/2+j-1].

    Returns
    -------
    sw : float
        The sum of weights.

        If `wt` is None, `sw` contains the number of observations, n.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 2)
            On entry, `mean` = *<value>*.

            Constraint: `mean` = 'M' or 'Z'.

        (`errno` 3)
            On entry, weight = *<value>*.

            Constraint: weight = 'W' or 'U'.

        (`errno` 4)
            On entry, `wt`[*<value>*] < 0.0.

            Constraint: `wt`[i-1] >= 0.0, for i = 1,2,...,n.

    References
    ----------
    Chan, T F, Golub, G H and Leveque, R J, 1982, `Updating Formulae and
    a Pairwise Algorithm for Computing Sample Variances`, Compstat,
    Physica-Verlag

    West, D H D, 1979, `Updating mean and variance estimates: An
    improved method`, Comm. ACM (22), 532--555

    See Also
    --------
    :meth:`naginterfaces.library.examples.correg.lars_param_ex.main`
    :meth:`naginterfaces.library.examples.correg.linregm_fit_stepwise_ex.main`
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.ssqmat'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02buft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, sorder.
        utils._BYREF_STRING, # Input, mean(len=1).
        utils._BYREF_STRING, # Input, weight(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Input, x[ldx, *].
        utils._BYREF_INT, # Input, ldx.
        utils._EngineFloat64ArrayType.c_type, # Input, wt[n].
        utils._BYREF_FLOAT64, # Output, sw.
        utils._EngineFloat64ArrayType.c_type, # Output, wmean[m].
        utils._EngineFloat64ArrayType.c_type, # Output, c[(m*m+m)//2].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of mean.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of weight.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((x, 'x'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _sorder = utils._EngineIntScalarType.c_type(_nag_enum_consts.nag_col_major)
    _mean = utils._EngineCharScalarType(
        mean, {'fun_name': fun_name, 'entity': 'mean'}, exp_chrlen=1,
    ).to_ctypes()
    if wt is None:
        _weight = utils._EngineCharScalarType(
            'U', {'fun_name': fun_name, 'entity': '_weight'}, exp_chrlen=1,
        ).to_ctypes()
    else:
        _weight = utils._EngineCharScalarType(
            'W', {'fun_name': fun_name, 'entity': '_weight'}, exp_chrlen=1,
        ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _ldx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _sorder_py = utils._EngineIntScalarType.to_py(_sorder)
    _m_py = utils._EngineIntScalarType.to_py(_m)
    if _sorder_py == 1:
        _x_dim2 = _m_py
    elif _sorder_py == 2:
        _x_dim2 = _n_py
    else:
        _x_dim2 = 0
    exp_shape_x = (_n_py, _x_dim2)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=exp_shape_x,
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _wt = utils._EngineFloat64ArrayType(
        wt, {'fun_name': fun_name, 'entity': 'wt'},
        exp_shape=(_n_py,),
        allow_none=True,
    ).to_ctypes()
    _sw = utils._EngineFloat64ScalarType.empty_ctypes()
    _wmean_cls = utils._EngineFloat64ArrayType(
        wmean, {'fun_name': fun_name, 'entity': 'wmean'},
        exp_shape=(_m_py,),
        check_contents=False,
    )
    _wmean = _wmean_cls.to_ctypes(can_reference_contents=False)
    _c_cls = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=((_m_py*_m_py+_m_py)//2,),
        check_contents=False,
    )
    _c = _c_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _sorder,
        _mean,
        _weight,
        _n,
        _m,
        _x,
        _ldx,
        _wt,
        _sw,
        _wmean,
        _c,
        _errbuf,
        _ierr,
        _mean._length_,
        _weight._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mean = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mean = \'M\' or \'Z\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, weight = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: weight = \'W\' or \'U\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, wt[{:s}] < 0.0.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: wt[i-1] >= 0.0, for i = 1,2,...,n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _wmean_cls.refresh_py(_wmean)
        _c_cls.refresh_py(_c)
        _sw_py = utils._EngineFloat64ScalarType.to_py(_sw)
        return_data = _sw_py
    _err_class.manage()
    return return_data

def ssqmat_to_corrmat(m, r):
    u"""
    Computes a correlation matrix from a sum of squares matrix.

    ``ssqmat_to_corrmat`` calculates a matrix of Pearson product-moment
    correlation coefficients from sums of squares and cross-products of
    deviations about the mean.

    For full information please refer to the NAG Library document for
    g02bw

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02bwf.html

    Parameters
    ----------
    m : int
        m, the number of variables.

    r : float, shape ((`m`*`m`+`m`)/2), modified in place
        `On entry`: contains the upper triangular part of the sums of
        squares and cross-products matrix of deviations from the mean.
        These are stored packed by column, i.e., the cross-product
        between variable j and k, k >= j, is stored in
        `r`[(k*(k-1)/2+j)-1].

        `On exit`: the Pearson product-moment correlation coefficients.

        These are stored packed by column corresponding to the input
        cross-products.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            A variable has a zero variance. All correlations involving
            the variable with zero variance will be returned as zero.
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.ssqmat_to_corrmat'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02bwft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, r[(m*m+m)//2].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _r_cls = utils._EngineFloat64ArrayType(
        r, {'fun_name': fun_name, 'entity': 'r'},
        exp_shape=((_m_py*_m_py+_m_py)//2,),
    )
    _r = _r_cls.to_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _m,
        _r,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'A variable has a zero variance. All correlations involving the variable',
                    'with zero variance will be returned as zero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _r_cls.refresh_py(_r)
        return_data = None
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def corrmat(nonzwt, n, m, x, wt, xbar, std, v, r):
    u"""
    Computes (optionally weighted) correlation and covariance matrices.

    ``corrmat`` calculates the sample means, the standard deviations,
    the variance-covariance matrix, and the matrix of Pearson
    product-moment correlation coefficients for a set of data.
    Weights may be used.

    For full information please refer to the NAG Library document for
    g02bx

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02bxf.html

    Parameters
    ----------
    nonzwt : str, length 1
        The variance calculation uses a divisor which is either the
        number of weights or the number of nonzero weights.

    n : int
        The number of data observations in the sample.

    m : int
        The number of variables.

    x : float, shape (`n`, `m`)
        `x`[i-1,j-1] must contain the ith observation for the jth
        variable, for j = 1,2,...,`m`, for i = 1,2,...,`n`.

    wt : None or float, shape (`n`)
        w, the optional frequency weighting for each observation, with
        `wt`[i-1] = w_i. Usually w_i will be an integral value
        corresponding to the number of observations associated with the
        ith data value, or zero if the ith data value is to be ignored.
        If `wt` is None, w_i is set to 1 for all i.

    xbar : float, shape (`m`), modified in place
        `On exit`: the sample means. `xbar`[j-1] contains the mean of
        the jth variable.

    std : float, shape (`m`), modified in place
        `On exit`: the standard deviations. `std`[j-1] contains the
        standard deviation for the jth variable.

    v : float, shape (`m`, `m`), modified in place
        `On exit`: the variance-covariance matrix. `v`[j-1,k-1] contains
        the covariance between variables j and k, for k = 1,2,...,`m`,
        for j = 1,2,...,`m`.

    r : float, shape (`m`, `m`), modified in place
        `On exit`: the matrix of Pearson product-moment correlation
        coefficients. `r`[j-1,k-1] contains the correlation coefficient
        between variables j and k.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` > 1.

        (`errno` 2)
            On entry, weight = *<value>*.

            Constraint: weight = 'U', 'V' or 'W'

        (`errno` 3)
            On entry, at least one value of `wt` is negative.

            Constraint: `wt`[i-1] >= 0, for i = 1,2,...,`n`.

        (`errno` 4)
            On entry, *<value>* observations have nonzero weight.

            Constraint: at least two observations must have a nonzero
            weight.

        (`errno` 4)
            On entry, Sum of the weights is *<value>*.

            Constraint: Sum of the weights must be greater than 1.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 5)
            A variable has a zero variance. In this case `v` and `std`
            are returned as calculated but `r` will contain zero for any
            correlation involving a variable with zero variance.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Chan, T F, Golub, G H and Leveque, R J, 1982, `Updating Formulae and
    a Pairwise Algorithm for Computing Sample Variances`, Compstat,
    Physica-Verlag

    West, D H D, 1979, `Updating mean and variance estimates: An
    improved method`, Comm. ACM (22), 532--555
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.corrmat'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02bxft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, weight(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Input, x[ldx, m].
        utils._BYREF_INT, # Input, ldx.
        utils._EngineFloat64ArrayType.c_type, # Input, wt[n].
        utils._EngineFloat64ArrayType.c_type, # Output, xbar[m].
        utils._EngineFloat64ArrayType.c_type, # Output, std[m].
        utils._EngineFloat64ArrayType.c_type, # Output, v[ldv, m].
        utils._BYREF_INT, # Input, ldv.
        utils._EngineFloat64ArrayType.c_type, # Output, r[ldr, m].
        utils._BYREF_INT, # Input, ldr.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of weight.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((x, 'x'), (v, 'v'), (r, 'r')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _nonzwt = utils._EngineCharScalarType(
        nonzwt, {'fun_name': fun_name, 'entity': 'nonzwt'}, exp_chrlen=1,
    ).to_ctypes()
    _nonzwt_py = utils._EngineCharScalarType.to_py(_nonzwt)
    if wt is None:
        _weight = utils._EngineCharScalarType(
            'U', {'fun_name': fun_name, 'entity': '_weight'}, exp_chrlen=1,
        ).to_ctypes()
    elif _nonzwt_py in ('W', 'w'):
        _weight = utils._EngineCharScalarType(
            'V', {'fun_name': fun_name, 'entity': '_weight'}, exp_chrlen=1,
        ).to_ctypes()
    else:
        _weight = utils._EngineCharScalarType(
            'W', {'fun_name': fun_name, 'entity': '_weight'}, exp_chrlen=1,
        ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _ldx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py, _m_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _wt = utils._EngineFloat64ArrayType(
        wt, {'fun_name': fun_name, 'entity': 'wt'},
        exp_shape=(_n_py,),
        allow_none=True,
    ).to_ctypes()
    _xbar_cls = utils._EngineFloat64ArrayType(
        xbar, {'fun_name': fun_name, 'entity': 'xbar'},
        exp_shape=(_m_py,),
        check_contents=False,
    )
    _xbar = _xbar_cls.to_ctypes(can_reference_contents=False)
    _std_cls = utils._EngineFloat64ArrayType(
        std, {'fun_name': fun_name, 'entity': 'std'},
        exp_shape=(_m_py,),
        check_contents=False,
    )
    _std = _std_cls.to_ctypes(can_reference_contents=False)
    _ldv_py = utils._EngineFloat64ArrayType.get_nd(v, {'fun_name': fun_name, 'entity': 'v'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldv = utils._EngineIntScalarType(
        _ldv_py, {'fun_name': fun_name, 'entity': '_ldv'},
    ).to_ctypes()
    _v_cls = utils._EngineFloat64ArrayType(
        v, {'fun_name': fun_name, 'entity': 'v'},
        exp_shape=(_m_py, _m_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _v = _v_cls.to_ctypes(can_reference_contents=False)
    _ldr_py = utils._EngineFloat64ArrayType.get_nd(r, {'fun_name': fun_name, 'entity': 'r'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldr = utils._EngineIntScalarType(
        _ldr_py, {'fun_name': fun_name, 'entity': '_ldr'},
    ).to_ctypes()
    _r_cls = utils._EngineFloat64ArrayType(
        r, {'fun_name': fun_name, 'entity': 'r'},
        exp_shape=(_m_py, _m_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _r = _r_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _weight,
        _n,
        _m,
        _x,
        _ldx,
        _wt,
        _xbar,
        _std,
        _v,
        _ldv,
        _r,
        _ldr,
        _errbuf,
        _ierr,
        _weight._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n > 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, weight = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: weight = \'U\', \'V\' or \'W\'',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, at least one value of wt is negative.',
                    'Constraint: wt[i-1] >= 0, for i = 1,2,...,n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, {:s} observations have nonzero weight.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: at least two observations must have a nonzero weight.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, Sum of the weights is {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: Sum of the weights must be greater than 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99992:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'A variable has a zero variance. In this case v and std',
                    'are returned as calculated but r will contain zero for any correlation',
                    'involving a variable with zero variance.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _xbar_cls.refresh_py(_xbar)
        _std_cls.refresh_py(_std)
        _v_cls.refresh_py(_v)
        _r_cls.refresh_py(_r)
        return_data = None
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def corrmat_partial(m, ny, nx, isz, r, p):
    u"""
    Computes partial correlation/variance-covariance matrix from
    correlation/variance-covariance matrix computed by ``corrmat``.

    ``corrmat_partial`` computes a partial
    correlation/variance-covariance matrix from a correlation or
    variance-covariance matrix computed by ``corrmat``.

    For full information please refer to the NAG Library document for
    g02by

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02byf.html

    Parameters
    ----------
    m : int
        The number of variables in the variance-covariance/correlation
        matrix given in `r`.

    ny : int
        The number of Y variables, n_y, for which partial correlation
        coefficients are to be computed.

    nx : int
        The number of X variables, n_x, which are to be considered as
        fixed.

    isz : int, shape (`m`)
        Indicates which variables belong to set X and Y.

        `isz`[i-1] < 0

            The ith variable is a Y variable, for i = 1,2,...,`m`.

        `isz`[i-1] > 0

            The ith variable is a X variable.

        `isz`[i-1] = 0

            The ith variable is not included in the computations.

    r : float, shape (`m`, `m`)
        The variance-covariance or correlation matrix for the `m`
        variables as given by ``corrmat``. Only the upper triangle need
        be given.

        **Note:** the matrix must be a full rank variance-covariance or
        correlation matrix and so be positive definite.

        This condition is not directly checked by the function.

    p : float, shape (`ny`, `ny`), modified in place
        `On exit`: the strict upper triangle of `p` contains the strict
        upper triangular part of the n_y by n_y partial correlation
        matrix. The lower triangle contains the lower triangle of the
        n_y by n_y partial variance-covariance matrix if the matrix
        given in `r` is a variance-covariance matrix. If the matrix
        given in `r` is a partial correlation matrix then the
        variance-covariance matrix is for standardized variables.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `ny` = *<value>*, `nx` = *<value>* and `m` =
            *<value>*.

            Constraint: `ny`+`nx` <= `m`.

        (`errno` 1)
            On entry, `nx` = *<value>*.

            Constraint: `nx` >= 1.

        (`errno` 1)
            On entry, `ny` = *<value>*.

            Constraint: `ny` > 1.

        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` > 2.

        (`errno` 2)
            On entry, *<value>* values of `isz` are greater than zero
            and `nx` = *<value>*.

            Constraint: exactly `nx` values of `isz` must be greater
            than zero.

        (`errno` 2)
            On entry, *<value>* values of `isz` are less than zero and
            `ny` = *<value>*.

            Constraint: exactly `ny` values of `isz` must be less than
            zero.

        (`errno` 3)
            On entry, the square root of the correlation matrix of the
            independent variables is singular.

            Try removing some of the X variables by setting the
            appropriate element of `isz` = 0.

        (`errno` 3)
            On entry, the correlation matrix of the independent
            variables is singular.

            Try removing some of the X variables by setting the
            appropriate element of `isz` = 0.

        (`errno` 4)
            On entry, an element of the partial correlation matrix is
            greater than 1.

            Constraint: `r` must be positive definite.

        (`errno` 4)
            On entry, a diagonal element of the partial covariance
            matrix is zero.

            Constraint: `r` must be positive definite.

        (`errno` 4)
            On entry, a diagonal element of the partial covariance
            matrix is zero and an element of the partial correlation
            matrix is greater than 1.

            Constraint: `r` must be positive definite.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Krzanowski, W J, 1990, `Principles of Multivariate Analysis`, Oxford
    University Press

    Morrison, D F, 1967, `Multivariate Statistical Methods`,
    McGraw--Hill

    Osborn, J F, 1979, `Statistical Exercises in Medical Research`,
    Blackwell

    Snedecor, G W and Cochran, W G, 1967, `Statistical Methods`, Iowa
    State University Press
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.corrmat_partial'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02byft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, ny.
        utils._BYREF_INT, # Input, nx.
        utils._EngineIntArrayType.c_type, # Input, isz[m].
        utils._EngineFloat64ArrayType.c_type, # Input, r[ldr, m].
        utils._BYREF_INT, # Input, ldr.
        utils._EngineFloat64ArrayType.c_type, # Output, p[ldp, ny].
        utils._BYREF_INT, # Input, ldp.
        utils._EngineFloat64ArrayType.c_type, # Workspace, wk.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((r, 'r'), (p, 'p')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _ny = utils._EngineIntScalarType(
        ny, {'fun_name': fun_name, 'entity': 'ny'},
    ).to_ctypes()
    _nx = utils._EngineIntScalarType(
        nx, {'fun_name': fun_name, 'entity': 'nx'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _isz = utils._EngineIntArrayType(
        isz, {'fun_name': fun_name, 'entity': 'isz'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _ldr_py = utils._EngineFloat64ArrayType.get_nd(r, {'fun_name': fun_name, 'entity': 'r'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldr = utils._EngineIntScalarType(
        _ldr_py, {'fun_name': fun_name, 'entity': '_ldr'},
    ).to_ctypes()
    _r = utils._EngineFloat64ArrayType(
        r, {'fun_name': fun_name, 'entity': 'r'},
        exp_shape=(_m_py, _m_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _ldp_py = utils._EngineFloat64ArrayType.get_nd(p, {'fun_name': fun_name, 'entity': 'p'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldp = utils._EngineIntScalarType(
        _ldp_py, {'fun_name': fun_name, 'entity': '_ldp'},
    ).to_ctypes()
    _ny_py = utils._EngineIntScalarType.to_py(_ny)
    _p_cls = utils._EngineFloat64ArrayType(
        p, {'fun_name': fun_name, 'entity': 'p'},
        exp_shape=(_ny_py, _ny_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _p = _p_cls.to_ctypes(can_reference_contents=False)
    _wk = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _m,
        _ny,
        _nx,
        _isz,
        _r,
        _ldr,
        _p,
        _ldp,
        _wk,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ny = {:s}, nx = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: ny+nx <= m.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nx = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nx >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ny = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ny > 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m > 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, {:s} values of isz are greater than zero'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'and nx = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: exactly nx values of isz must be greater than zero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, {:s} values of isz are less than zero'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'and ny = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: exactly ny values of isz must be less than zero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the square root of the correlation matrix of the independent',
                    'variables is singular.',
                    'Try removing some of the X variables by setting the appropriate',
                    'element of isz = 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the correlation matrix of the independent variables is singular.',
                    'Try removing some of the X variables by setting the appropriate',
                    'element of isz = 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99987:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, an element of the partial correlation matrix is greater than 1.',
                    'Constraint: r must be positive definite.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99988:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a diagonal element of the partial covariance matrix is zero.',
                    'Constraint: r must be positive definite.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a diagonal element of the partial covariance matrix is zero',
                    'and an element of the partial correlation matrix is greater than 1.',
                    'Constraint: r must be positive definite.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _p_cls.refresh_py(_p)
        return_data = None
    _err_class.manage()
    return return_data

def ssqmat_combine(mean, m, xsw, xmean, xc, ysw, ymean, yc):
    u"""
    Combines two sums of squares matrices, for use after ``ssqmat``.

    ``ssqmat_combine`` combines two sets of sample means and sums of
    squares and cross-products matrices.
    It is designed to be used in conjunction with ``ssqmat`` to allow
    large datasets to be summarised.

    For full information please refer to the NAG Library document for
    g02bz

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02bzf.html

    Parameters
    ----------
    mean : str, length 1
        Indicates whether the matrices supplied in `xc` and `yc` are
        sums of squares and cross-products, or sums of squares and
        cross-products of deviations about the mean.

        `mean` = 'M'

            Sums of squares and cross-products of deviations about the
            mean have been supplied.

        `mean` = 'Z'

            Sums of squares and cross-products have been supplied.

    m : int
        m, the number of variables.

    xsw : float
        W_x, the sum of weights, from the first set of data, X. If the
        data is unweighted then this will be the number of observations
        in the first dataset.

    xmean : float, shape (`m`), modified in place
        `On entry`: mu_x, the sample means for the first set of data, X.

        `On exit`: mu_z, the sample means for the combined data, Z.

    xc : float, shape ((`m`*`m`+`m`)/2), modified in place
        `On entry`: C_x, the sums of squares and cross-products matrix
        for the first set of data, X, as returned by ``ssqmat``.

        ``ssqmat``, returns this matrix packed by columns, i.e., the
        cross-product between the jth and kth variable, k >= j, is
        stored in `xc`[k*(k-1)/2+j-1].

        No check is made that C_x is a valid cross-products matrix.

        `On exit`: C_z, the sums of squares and cross-products matrix
        for the combined dataset, Z.

        This matrix is again stored packed by columns.

    ysw : float
        W_y, the sum of weights, from the second set of data, Y. If the
        data is unweighted then this will be the number of observations
        in the second dataset.

    ymean : float, shape (`m`)
        mu_y, the sample means for the second set of data, Y.

    yc : float, shape ((`m`*`m`+`m`)/2)
        C_y, the sums of squares and cross-products matrix for the
        second set of data, Y, as returned by ``ssqmat``.

        ``ssqmat``, returns this matrix packed by columns, i.e., the
        cross-product between the jth and kth variable, k >= j, is
        stored in `yc`[k*(k-1)/2+j-1].

        No check is made that C_y is a valid cross-products matrix.

    Returns
    -------
    xsw : float
        W_z, the sum of weights, from the combined dataset, Z. If both
        datasets are unweighted then this will be the number of
        observations in the combined dataset.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, `mean` = *<value>*.

            Constraint: `mean` = 'M' or 'Z'.

        (`errno` 21)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 31)
            On entry, `xsw` = *<value>*.

            Constraint: `xsw` >= 0.0.

        (`errno` 61)
            On entry, `ysw` = *<value>*.

            Constraint: `ysw` >= 0.0.

    References
    ----------
    Bennett, J, Pebay, P, Roe, D and Thompson, D, 2009, `Numerically
    stable, single-pass, parallel statistics algorithms`, Proceedings of
    IEEE International Conference on Cluster Computing
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.ssqmat_combine'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02bzft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, imean.
        utils._BYREF_INT, # Input, m.
        utils._BYREF_FLOAT64, # Input/Output, xsw.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, xmean[m].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, xc[lc].
        utils._BYREF_FLOAT64, # Input, ysw.
        utils._EngineFloat64ArrayType.c_type, # Input, ymean[m].
        utils._EngineFloat64ArrayType.c_type, # Input, yc[lc].
        utils._BYREF_INT, # Input, lc.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _mean = utils._EngineCharScalarType(
        mean, {'fun_name': fun_name, 'entity': 'mean'}, exp_chrlen=1,
    ).to_ctypes()
    _mean_py = utils._EngineCharScalarType.to_py(_mean)
    if _mean_py.upper() == 'M':
        _imean = utils._EngineIntScalarType.c_type(1)
    elif _mean_py.upper() == 'Z':
        _imean = utils._EngineIntScalarType.c_type(0)
    else:
        _err_class = utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                'On entry, mean = {:s}.'.format(
                    utils._STR_FMT_STR.format(_mean_py),
                ),
                'Constraint: mean = \'M\' or \'Z\'.',
            ],
            errcodes=(11,),
        )
        _err_class.manage()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _xsw = utils._EngineFloat64ScalarType(
        xsw, {'fun_name': fun_name, 'entity': 'xsw'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _xmean_cls = utils._EngineFloat64ArrayType(
        xmean, {'fun_name': fun_name, 'entity': 'xmean'},
        exp_shape=(_m_py,),
    )
    _xmean = _xmean_cls.to_ctypes()
    _lc_py = utils._EngineFloat64ArrayType.get_shape(xc, {'fun_name': fun_name, 'entity': 'xc'}, exp_rank=1)[0]
    _lc = utils._EngineIntScalarType(
        _lc_py, {'fun_name': fun_name, 'entity': '_lc'},
    ).to_ctypes()
    _xc_cls = utils._EngineFloat64ArrayType(
        xc, {'fun_name': fun_name, 'entity': 'xc'},
        exp_shape=((_m_py*_m_py+_m_py)//2,),
    )
    _xc = _xc_cls.to_ctypes()
    _ysw = utils._EngineFloat64ScalarType(
        ysw, {'fun_name': fun_name, 'entity': 'ysw'},
    ).to_ctypes()
    _ymean = utils._EngineFloat64ArrayType(
        ymean, {'fun_name': fun_name, 'entity': 'ymean'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _yc = utils._EngineFloat64ArrayType(
        yc, {'fun_name': fun_name, 'entity': 'yc'},
        exp_shape=((_m_py*_m_py+_m_py)//2,),
    ).to_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _imean,
        _m,
        _xsw,
        _xmean,
        _xc,
        _ysw,
        _ymean,
        _yc,
        _lc,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 21 and _ifmt == 21:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 31 and _ifmt == 31:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, xsw = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: xsw >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 61 and _ifmt == 61:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ysw = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ysw >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _xsw_py = utils._EngineFloat64ScalarType.to_py(_xsw)
        _xmean_cls.refresh_py(_xmean)
        _xc_cls.refresh_py(_xc)
        return_data = _xsw_py
    _err_class.manage()
    return return_data

def linregs_const(n, x, y, result):
    u"""
    Simple linear regression with constant term, no missing values.

    ``linregs_const`` performs a simple linear regression with dependent
    variable y and independent variable x.

    For full information please refer to the NAG Library document for
    g02ca

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02caf.html

    Parameters
    ----------
    n : int
        n, the number of pairs of observations.

    x : float, shape (`n`)
        `x`[i-1] must contain x_i, for i = 1,2,...,n.

    y : float, shape (`n`)
        `y`[i-1] must contain y_i, for i = 1,2,...,n.

    result : float, shape (20), modified in place
        `On exit`: the following information:

        [table omitted]

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` > 2.

        (`errno` 2)
            On entry, all `n` values of at least one of `x` and `y` are
            identical.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Draper, N R and Smith, H, 1985, `Applied Regression Analysis`, (2nd
    Edition), Wiley
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.linregs_const'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02caft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[n].
        utils._EngineFloat64ArrayType.c_type, # Input, y[n].
        utils._EngineFloat64ArrayType.c_type, # Output, result[20].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _result_cls = utils._EngineFloat64ArrayType(
        result, {'fun_name': fun_name, 'entity': 'result'},
        exp_shape=(20,),
        check_contents=False,
    )
    _result = _result_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _x,
        _y,
        _result,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n > 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, all n values of at least one of x and y are identical.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _result_cls.refresh_py(_result)
        return_data = None
    _err_class.manage()
    return return_data

def linregs_noconst(n, x, y, result):
    u"""
    Simple linear regression without constant term, no missing values.

    ``linregs_noconst`` performs a simple linear regression with no
    constant, with dependent variable y and independent variable x.

    For full information please refer to the NAG Library document for
    g02cb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02cbf.html

    Parameters
    ----------
    n : int
        n, the number of pairs of observations.

    x : float, shape (`n`)
        `x`[i-1] must contain x_i, for i = 1,2,...,n.

    y : float, shape (`n`)
        `y`[i-1] must contain y_i, for i = 1,2,...,n.

    result : float, shape (20), modified in place
        `On exit`: the following information:

        [table omitted]

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` > 2.

        (`errno` 2)
            On entry, all `n` values of at least one of `x` and `y` are
            identical.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Draper, N R and Smith, H, 1985, `Applied Regression Analysis`, (2nd
    Edition), Wiley
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.linregs_noconst'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02cbft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[n].
        utils._EngineFloat64ArrayType.c_type, # Input, y[n].
        utils._EngineFloat64ArrayType.c_type, # Output, result[20].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _result_cls = utils._EngineFloat64ArrayType(
        result, {'fun_name': fun_name, 'entity': 'result'},
        exp_shape=(20,),
        check_contents=False,
    )
    _result = _result_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _x,
        _y,
        _result,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n > 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, all n values of at least one of x and y are identical.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _result_cls.refresh_py(_result)
        return_data = None
    _err_class.manage()
    return return_data

def linregs_const_miss(n, x, y, xmiss, ymiss, result):
    u"""
    Simple linear regression with constant term, missing values.

    ``linregs_const_miss`` performs a simple linear regression with
    dependent variable y and independent variable x, omitting cases
    involving missing values.

    For full information please refer to the NAG Library document for
    g02cc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02ccf.html

    Parameters
    ----------
    n : int
        n, the number of pairs of observations.

    x : float, shape (`n`)
        `x`[i-1] must contain x_i, for i = 1,2,...,n.

    y : float, shape (`n`)
        `y`[i-1] must contain y_i, for i = 1,2,...,n.

    xmiss : float
        The value xm which is to be taken as the missing value for the
        variable x. See [Accuracy].

    ymiss : float
        The value ym which is to be taken as the missing value for the
        variable y. See [Accuracy].

    result : float, shape (21), modified in place
        `On exit`: the following information:

        [table omitted]

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` > 2.

        (`errno` 2)
            After observations with missing values were omitted, two or
            fewer cases remained.

        (`errno` 3)
            After observations with missing values were omitted, all
            remaining values of at least one of `x` and `y` were
            identical.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Draper, N R and Smith, H, 1985, `Applied Regression Analysis`, (2nd
    Edition), Wiley
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.linregs_const_miss'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02ccft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[n].
        utils._EngineFloat64ArrayType.c_type, # Input, y[n].
        utils._BYREF_FLOAT64, # Input, xmiss.
        utils._BYREF_FLOAT64, # Input, ymiss.
        utils._EngineFloat64ArrayType.c_type, # Output, result[21].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _xmiss = utils._EngineFloat64ScalarType(
        xmiss, {'fun_name': fun_name, 'entity': 'xmiss'},
    ).to_ctypes()
    _ymiss = utils._EngineFloat64ScalarType(
        ymiss, {'fun_name': fun_name, 'entity': 'ymiss'},
    ).to_ctypes()
    _result_cls = utils._EngineFloat64ArrayType(
        result, {'fun_name': fun_name, 'entity': 'result'},
        exp_shape=(21,),
        check_contents=False,
    )
    _result = _result_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _x,
        _y,
        _xmiss,
        _ymiss,
        _result,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n > 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'After observations with missing values were omitted,',
                    'two or fewer cases remained.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'After observations with missing values were omitted,',
                    'all remaining values of at least one of x and y were identical.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _result_cls.refresh_py(_result)
        return_data = None
    _err_class.manage()
    return return_data

def linregs_noconst_miss(n, x, y, xmiss, ymiss, result):
    u"""
    Simple linear regression without constant term, missing values.

    ``linregs_noconst_miss`` performs a simple linear regression with no
    constant, with dependent variable y and independent variable x,
    omitting cases involving missing values.

    For full information please refer to the NAG Library document for
    g02cd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02cdf.html

    Parameters
    ----------
    n : int
        n, the number of pairs of observations.

    x : float, shape (`n`)
        `x`[i-1] must contain x_i, for i = 1,2,...,n.

    y : float, shape (`n`)
        `y`[i-1] must contain y_i, for i = 1,2,...,n.

    xmiss : float
        The value xm, which is to be taken as the missing value for the
        variable x (see [Accuracy]).

    ymiss : float
        The value ym, which is to be taken as the missing value for the
        variable y (see [Accuracy]).

    result : float, shape (21), modified in place
        `On exit`: the following information:

        [table omitted]

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 2.

        (`errno` 2)
            After observations with missing values were omitted, fewer
            than two cases remained.

        (`errno` 3)
            After observations with missing values were omitted, all
            remaining values of at least one of `x` and `y` were
            identical.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Draper, N R and Smith, H, 1985, `Applied Regression Analysis`, (2nd
    Edition), Wiley
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.linregs_noconst_miss'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02cdft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[n].
        utils._EngineFloat64ArrayType.c_type, # Input, y[n].
        utils._BYREF_FLOAT64, # Input, xmiss.
        utils._BYREF_FLOAT64, # Input, ymiss.
        utils._EngineFloat64ArrayType.c_type, # Output, result[21].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _xmiss = utils._EngineFloat64ScalarType(
        xmiss, {'fun_name': fun_name, 'entity': 'xmiss'},
    ).to_ctypes()
    _ymiss = utils._EngineFloat64ScalarType(
        ymiss, {'fun_name': fun_name, 'entity': 'ymiss'},
    ).to_ctypes()
    _result_cls = utils._EngineFloat64ArrayType(
        result, {'fun_name': fun_name, 'entity': 'result'},
        exp_shape=(21,),
        check_contents=False,
    )
    _result = _result_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _x,
        _y,
        _xmiss,
        _ymiss,
        _result,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'After observations with missing values were omitted,',
                    'fewer than two cases remained.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'After observations with missing values were omitted,',
                    'all remaining values of at least one of x and y were identical.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _result_cls.refresh_py(_result)
        return_data = None
    _err_class.manage()
    return return_data

def linregm_service_select(n, xbar, std, ssp, r, m, korder, xbar2, std2, ssp2, r2):
    u"""
    Service function for multiple linear regression, select elements
    from vectors and matrices.

    ``linregm_service_select`` takes selected elements from two vectors
    (typically vectors of means and standard deviations) to form two
    smaller vectors, and selected rows and columns from two matrices
    (typically either matrices of sums of squares and cross-products of
    deviations from means and Pearson product-moment correlation
    coefficients, or matrices of sums of squares and cross-products
    about zero and correlation-like coefficients) to form two smaller
    matrices, allowing reordering of elements in the process.

    For full information please refer to the NAG Library document for
    g02ce

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02cef.html

    Parameters
    ----------
    n : int
        n, the number of variables in the input data.

    xbar : float, shape (`n`)
        `xbar`[i-1] must be set to x__i, the mean of variable i, for i =
        1,2,...,n.

    std : float, shape (`n`)
        `std`[i-1] must be set to s_i, the standard deviation of
        variable i, for i = 1,2,...,n.

    ssp : float, shape (`n`, `n`)
        `ssp`[i-1,j-1] must be set to the sum of cross-products of
        deviations from means S_{ij} (or about zero, S~_{ij}) for
        variables i and j, for j = 1,2,...,n, for i = 1,2,...,n.

    r : float, shape (`n`, `n`)
        `r`[i-1,j-1] must be set to the Pearson product-moment
        correlation coefficient R_{ij} (or the correlation-like
        coefficient, R~_{ij}) for variables i and j, for j = 1,2,...,n,
        for i = 1,2,...,n.

    m : int
        The number of variables m, required in the reduced vectors and
        matrices.

    korder : int, shape (`m`)
        `korder`[i-1] must be set to the number of the original variable
        which is to be the ith variable in the output vectors and
        matrices, for i = 1,2,...,m.

    xbar2 : float, shape (`m`), modified in place
        `On exit`: the mean of variable i, `xbar`[i-1], where i =
        `korder`[k-1], for k = 1,2,...,m. (The array `xbar2` must differ
        from `xbar` and `std`.)

    std2 : float, shape (`m`), modified in place
        `On exit`: the standard deviation of variable i, `std`[i-1],
        where i = `korder`[k-1], for k = 1,2,...,m. (The array `std2`
        must differ from both `xbar` and `std`.)

    ssp2 : float, shape (`m`, `m`), modified in place
        `On exit`: `ssp2`[k-1,l-1] contains the value of `ssp`[i-1,j-1],
        where i = `korder`[k-1] and j = `korder`[l-1], for l =
        1,2,...,m, for k = 1,2,...,m. (The array `ssp2` must differ from
        both `ssp` and `r`.)

        That is to say: on exit, `ssp2`[k-1,l-1] contains the sum of
        cross-products of deviations from means S_{ij} (or about zero,
        S~_{ij}).

    r2 : float, shape (`m`, `m`), modified in place
        `On exit`: `r2`[k-1,l-1] contains the value of `r`[i-1,j-1],
        where i = `korder`[k-1] and j = `korder`[l-1], for l =
        1,2,...,m, for k = 1,2,...,m. (The array `r2` must differ from
        both `ssp` and `r`.)

        That is to say: on exit, `r2`[k-1,l-1] contains the Pearson
        product-moment coefficient R_{ij} (or the correlation-like
        coefficient, R~_{ij}).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 2.

        (`errno` 2)
            On entry, `n` = *<value>* and `m` = *<value>*.

            Constratint: `n` >= `m`.

        (`errno` 4)
            On entry, `korder`[*<value>*] = *<value>* and `n` =
            *<value>*.

            Constraint: 1 <= `korder`[i-1] <= `n`, for i = 1,2,...,m.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.linregm_service_select'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02ceft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, xbar[n].
        utils._EngineFloat64ArrayType.c_type, # Input, std[n].
        utils._EngineFloat64ArrayType.c_type, # Input, ssp[ldssp, n].
        utils._BYREF_INT, # Input, ldssp.
        utils._EngineFloat64ArrayType.c_type, # Input, r[ldr, n].
        utils._BYREF_INT, # Input, ldr.
        utils._BYREF_INT, # Input, m.
        utils._EngineIntArrayType.c_type, # Input, korder[m].
        utils._EngineFloat64ArrayType.c_type, # Output, xbar2[m].
        utils._EngineFloat64ArrayType.c_type, # Output, std2[m].
        utils._EngineFloat64ArrayType.c_type, # Output, ssp2[ldssp2, m].
        utils._BYREF_INT, # Input, ldssp2.
        utils._EngineFloat64ArrayType.c_type, # Output, r2[ldr2, m].
        utils._BYREF_INT, # Input, ldr2.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((ssp, 'ssp'), (r, 'r'), (ssp2, 'ssp2'), (r2, 'r2')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _xbar = utils._EngineFloat64ArrayType(
        xbar, {'fun_name': fun_name, 'entity': 'xbar'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _std = utils._EngineFloat64ArrayType(
        std, {'fun_name': fun_name, 'entity': 'std'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _ldssp_py = utils._EngineFloat64ArrayType.get_nd(ssp, {'fun_name': fun_name, 'entity': 'ssp'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldssp = utils._EngineIntScalarType(
        _ldssp_py, {'fun_name': fun_name, 'entity': '_ldssp'},
    ).to_ctypes()
    _ssp = utils._EngineFloat64ArrayType(
        ssp, {'fun_name': fun_name, 'entity': 'ssp'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _ldr_py = utils._EngineFloat64ArrayType.get_nd(r, {'fun_name': fun_name, 'entity': 'r'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldr = utils._EngineIntScalarType(
        _ldr_py, {'fun_name': fun_name, 'entity': '_ldr'},
    ).to_ctypes()
    _r = utils._EngineFloat64ArrayType(
        r, {'fun_name': fun_name, 'entity': 'r'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _korder = utils._EngineIntArrayType(
        korder, {'fun_name': fun_name, 'entity': 'korder'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _xbar2_cls = utils._EngineFloat64ArrayType(
        xbar2, {'fun_name': fun_name, 'entity': 'xbar2'},
        exp_shape=(_m_py,),
        check_contents=False,
    )
    _xbar2 = _xbar2_cls.to_ctypes(can_reference_contents=False)
    _std2_cls = utils._EngineFloat64ArrayType(
        std2, {'fun_name': fun_name, 'entity': 'std2'},
        exp_shape=(_m_py,),
        check_contents=False,
    )
    _std2 = _std2_cls.to_ctypes(can_reference_contents=False)
    _ldssp2_py = utils._EngineFloat64ArrayType.get_nd(ssp2, {'fun_name': fun_name, 'entity': 'ssp2'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldssp2 = utils._EngineIntScalarType(
        _ldssp2_py, {'fun_name': fun_name, 'entity': '_ldssp2'},
    ).to_ctypes()
    _ssp2_cls = utils._EngineFloat64ArrayType(
        ssp2, {'fun_name': fun_name, 'entity': 'ssp2'},
        exp_shape=(_m_py, _m_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _ssp2 = _ssp2_cls.to_ctypes(can_reference_contents=False)
    _ldr2_py = utils._EngineFloat64ArrayType.get_nd(r2, {'fun_name': fun_name, 'entity': 'r2'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldr2 = utils._EngineIntScalarType(
        _ldr2_py, {'fun_name': fun_name, 'entity': '_ldr2'},
    ).to_ctypes()
    _r2_cls = utils._EngineFloat64ArrayType(
        r2, {'fun_name': fun_name, 'entity': 'r2'},
        exp_shape=(_m_py, _m_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _r2 = _r2_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _xbar,
        _std,
        _ssp,
        _ldssp,
        _r,
        _ldr,
        _m,
        _korder,
        _xbar2,
        _std2,
        _ssp2,
        _ldssp2,
        _r2,
        _ldr2,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s} and m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constratint: n >= m.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, korder[{:s}] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: 1 <= korder[i-1] <= n, for i = 1,2,...,m.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _xbar2_cls.refresh_py(_xbar2)
        _std2_cls.refresh_py(_std2)
        _ssp2_cls.refresh_py(_ssp2)
        _r2_cls.refresh_py(_r2)
        return_data = None
    _err_class.manage()
    return return_data

def linregm_service_reorder(n, korder, xbar, std, ssp, r):
    u"""
    Service function for multiple linear regression, reorder elements of
    vectors and matrices.

    ``linregm_service_reorder`` reorders the elements in two vectors
    (typically vectors of means and standard deviations), and the rows
    and columns in two matrices (typically either matrices of sums of
    squares and cross-products of deviations from means and Pearson
    product-moment correlation coefficients, or matrices of sums of
    squares and cross-products about zero and correlation-like
    coefficients).

    For full information please refer to the NAG Library document for
    g02cf

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02cff.html

    Parameters
    ----------
    n : int
        n, the number of variables in the input data.

    korder : int, shape (`n`)
        `korder`[i-1] must be set to the number of the original variable
        which is to be the ith variable in the re-arranged data, for i =
        1,2,...,n.

    xbar : float, shape (`n`), modified in place
        `On entry`: `xbar`[i-1] must be set to the mean of variable i,
        for i = 1,2,...,n.

        `On exit`: `xbar`[i-1] contains the mean of variable k where k =
        `korder`[i-1], for i = 1,2,...,n.

    std : float, shape (`n`), modified in place
        `On entry`: `std`[i-1] must be set to the standard deviation of
        variable i, for i = 1,2,...,n.

        `On exit`: `std`[i-1] contains the standard deviation of
        variable k where k = `korder`[i-1], for i = 1,2,...,n.

    ssp : float, shape (`n`, `n`), modified in place
        `On entry`: `ssp`[i-1,j-1] must be set to the sum of
        cross-products of deviations from means S_{ij} (or about zero
        S~_{ij}) for variables i and j, for j = 1,2,...,n, for i =
        1,2,...,n.

        `On exit`: `ssp`[i-1,j-1] contains the sum of cross-products of
        deviations from means S_{kl} (or about zero S~_{kl}) for
        variables k and l, where k = `korder`[i-1], and l =
        `korder`[j-1], i,j = 1,2,...,n.

    r : float, shape (`n`, `n`), modified in place
        `On entry`: `r`[i-1,j-1] must be set to the Pearson
        product-moment correlation coefficient R_{ij} (or the
        correlation-like coefficient R~_{ij}) for variables i and j, for
        j = 1,2,...,n, for i = 1,2,...,n.

        `On exit`: `r`[i-1,j-1] contains the Pearson product-moment
        correlation coefficient R_{kl} (or the correlation-like
        coefficient R~_{kl}) for variables k and l, where k =
        `korder`[i-1] and l = `korder`[j-1], for j = 1,2,...,n, for i =
        1,2,...,n.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 2.

        (`errno` 3)
            On entry, `korder`[*<value>*] = *<value>* and `n` =
            *<value>*.

            Constraint: 1 <= `korder`[i-1] <= `n`, for i = 1,2,...,m.

        (`errno` 4)
            On entry, there is not a one-to-one correspondence between
            the old variables and the new variables; at least one of the
            original variables is not included in the new set, and
            consequently at least one other variable has been included
            more than once.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.linregm_service_reorder'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02cfft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineIntArrayType.c_type, # Input, korder[n].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, xbar[n].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, std[n].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ssp[ldssp, n].
        utils._BYREF_INT, # Input, ldssp.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, r[ldr, n].
        utils._BYREF_INT, # Input, ldr.
        utils._EngineIntArrayType.c_type, # Workspace, kwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((ssp, 'ssp'), (r, 'r')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _korder = utils._EngineIntArrayType(
        korder, {'fun_name': fun_name, 'entity': 'korder'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _xbar_cls = utils._EngineFloat64ArrayType(
        xbar, {'fun_name': fun_name, 'entity': 'xbar'},
        exp_shape=(_n_py,),
    )
    _xbar = _xbar_cls.to_ctypes()
    _std_cls = utils._EngineFloat64ArrayType(
        std, {'fun_name': fun_name, 'entity': 'std'},
        exp_shape=(_n_py,),
    )
    _std = _std_cls.to_ctypes()
    _ldssp_py = utils._EngineFloat64ArrayType.get_nd(ssp, {'fun_name': fun_name, 'entity': 'ssp'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldssp = utils._EngineIntScalarType(
        _ldssp_py, {'fun_name': fun_name, 'entity': '_ldssp'},
    ).to_ctypes()
    _ssp_cls = utils._EngineFloat64ArrayType(
        ssp, {'fun_name': fun_name, 'entity': 'ssp'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
    )
    _ssp = _ssp_cls.to_ctypes()
    _ldr_py = utils._EngineFloat64ArrayType.get_nd(r, {'fun_name': fun_name, 'entity': 'r'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldr = utils._EngineIntScalarType(
        _ldr_py, {'fun_name': fun_name, 'entity': '_ldr'},
    ).to_ctypes()
    _r_cls = utils._EngineFloat64ArrayType(
        r, {'fun_name': fun_name, 'entity': 'r'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
    )
    _r = _r_cls.to_ctypes()
    _kwork = utils._EngineIntArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _korder,
        _xbar,
        _std,
        _ssp,
        _ldssp,
        _r,
        _ldr,
        _kwork,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, korder[{:s}] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: 1 <= korder[i-1] <= n, for i = 1,2,...,m.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, there is not a one-to-one correspondence between the old variables',
                    'and the new variables; at least one of the original variables is not',
                    'included in the new set, and consequently at least one other variable has',
                    'been included more than once.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _xbar_cls.refresh_py(_xbar)
        _std_cls.refresh_py(_std)
        _ssp_cls.refresh_py(_ssp)
        _r_cls.refresh_py(_r)
        return_data = None
    _err_class.manage()
    return return_data

def linregm_coeffs_const(n, k, xbar, ssp, r, result, coef, con, rinv, c):
    u"""
    Multiple linear regression, from correlation coefficients, with
    constant term.

    ``linregm_coeffs_const`` performs a multiple linear regression on a
    set of variables whose means, sums of squares and cross-products of
    deviations from means, and Pearson product-moment correlation
    coefficients are given.

    For full information please refer to the NAG Library document for
    g02cg

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02cgf.html

    Parameters
    ----------
    n : int
        The number of cases n, used in calculating the sums of squares
        and cross-products and correlation coefficients.

    k : int
        k, the number of independent variables in the regression.

    xbar : float, shape (`k`+1)
        `xbar`[i-1] must be set to x__i, the mean value of the ith
        variable, for i = 1,2,...,k+1; the mean of the dependent
        variable must be contained in `xbar`[k].

    ssp : float, shape (`k`+1, `k`+1)
        `ssp`[i-1,j-1] must be set to S_{ij}, the sum of cross-products
        of deviations from means for the ith and jth variables, for j =
        1,2,...,k+1, for i = 1,2,...,k+1; terms involving the dependent
        variable appear in row k+1 and column k+1.

    r : float, shape (`k`+1, `k`+1), modified in place
        `r`[i-1,j-1] must be set to R_{ij}, the Pearson product-moment
        correlation coefficient for the ith and jth variables, for j =
        1,2,...,k+1, for i = 1,2,...,k+1; terms involving the dependent
        variable appear in row k+1 and column k+1.

    result : float, shape (13), modified in place
        `On exit`: the following information:

        [table omitted]

    coef : float, shape (`k`, 3), modified in place
        `On exit`: for i = 1,2,...,k, the following information:

        `coef`[i-1,0]

            b_i, the regression coefficient for the ith variable.

        `coef`[i-1,1]

            se(b_i), the standard error of the regression coefficient
            for the ith variable.

        `coef`[i-1,2]

            t(b_i), the t value of the regression coefficient for the
            ith variable.

    con : float, shape (3), modified in place
        `On exit`: the following information:

        [table omitted]

    rinv : float, shape (`k`, `k`), modified in place
        `On exit`: the inverse of the matrix of correlation coefficients
        for the independent variables; that is, the inverse of the
        matrix consisting of the first k rows and columns of `r`.

    c : float, shape (`k`, `k`), modified in place
        `On exit`: the modified inverse matrix, where

            `c`[i-1,j-1] = `r`[i-1,j-1]*`rinv`[i-1,j-1]/`ssp`[i-1,j-1],
            for j = 1,2,...,k, for i = 1,2,...,k.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `k` = *<value>*.

            Constraint: `k` >= 1.

        (`errno` 3)
            On entry, `n` = *<value>* and `k` = *<value>*.

            Constraint: `n` > `k`+1.

        (`errno` 5)
            The `k` by `k` partition of `r` which requires inversion is
            not positive definite.

        (`errno` 6)
            The refinement following inversion has failed.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Draper, N R and Smith, H, 1985, `Applied Regression Analysis`, (2nd
    Edition), Wiley
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.linregm_coeffs_const'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02cgft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, iemode.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, k1.
        utils._BYREF_INT, # Input, k.
        utils._EngineFloat64ArrayType.c_type, # Input, xbar[k+1].
        utils._EngineFloat64ArrayType.c_type, # Input, ssp[ldssp, k+1].
        utils._BYREF_INT, # Input, ldssp.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, r[ldr, k+1].
        utils._BYREF_INT, # Input, ldr.
        utils._EngineFloat64ArrayType.c_type, # Output, result[13].
        utils._EngineFloat64ArrayType.c_type, # Output, coef[ldcoef, 3].
        utils._BYREF_INT, # Input, ldcoef.
        utils._EngineFloat64ArrayType.c_type, # Output, con[3].
        utils._EngineFloat64ArrayType.c_type, # Output, rinv[ldrinv, k].
        utils._BYREF_INT, # Input, ldrinv.
        utils._EngineFloat64ArrayType.c_type, # Output, c[ldc, k].
        utils._BYREF_INT, # Input, ldc.
        utils._EngineFloat64ArrayType.c_type, # Workspace, wkz.
        utils._BYREF_INT, # Input, ldwkz.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((ssp, 'ssp'), (r, 'r'), (coef, 'coef'), (rinv, 'rinv'), (c, 'c')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _iemode = utils._EngineIntScalarType.c_type(0)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _k = utils._EngineIntScalarType(
        k, {'fun_name': fun_name, 'entity': 'k'},
    ).to_ctypes()
    _k_py = utils._EngineIntScalarType.to_py(_k)
    _k1 = utils._EngineIntScalarType(
        _k_py+1, {'fun_name': fun_name, 'entity': '_k1'},
    ).to_ctypes()
    _xbar = utils._EngineFloat64ArrayType(
        xbar, {'fun_name': fun_name, 'entity': 'xbar'},
        exp_shape=(_k_py+1,),
    ).to_ctypes()
    _ldssp_py = utils._EngineFloat64ArrayType.get_nd(ssp, {'fun_name': fun_name, 'entity': 'ssp'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldssp = utils._EngineIntScalarType(
        _ldssp_py, {'fun_name': fun_name, 'entity': '_ldssp'},
    ).to_ctypes()
    _ssp = utils._EngineFloat64ArrayType(
        ssp, {'fun_name': fun_name, 'entity': 'ssp'},
        exp_shape=(_k_py+1, _k_py+1),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _ldr_py = utils._EngineFloat64ArrayType.get_nd(r, {'fun_name': fun_name, 'entity': 'r'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldr = utils._EngineIntScalarType(
        _ldr_py, {'fun_name': fun_name, 'entity': '_ldr'},
    ).to_ctypes()
    _r_cls = utils._EngineFloat64ArrayType(
        r, {'fun_name': fun_name, 'entity': 'r'},
        exp_shape=(_k_py+1, _k_py+1),
        mutual_sorder=_m_sorder,
    )
    _r = _r_cls.to_ctypes()
    _result_cls = utils._EngineFloat64ArrayType(
        result, {'fun_name': fun_name, 'entity': 'result'},
        exp_shape=(13,),
        check_contents=False,
    )
    _result = _result_cls.to_ctypes(can_reference_contents=False)
    _ldcoef_py = utils._EngineFloat64ArrayType.get_nd(coef, {'fun_name': fun_name, 'entity': 'coef'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldcoef = utils._EngineIntScalarType(
        _ldcoef_py, {'fun_name': fun_name, 'entity': '_ldcoef'},
    ).to_ctypes()
    _coef_cls = utils._EngineFloat64ArrayType(
        coef, {'fun_name': fun_name, 'entity': 'coef'},
        exp_shape=(_k_py, 3),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _coef = _coef_cls.to_ctypes(can_reference_contents=False)
    _con_cls = utils._EngineFloat64ArrayType(
        con, {'fun_name': fun_name, 'entity': 'con'},
        exp_shape=(3,),
        check_contents=False,
    )
    _con = _con_cls.to_ctypes(can_reference_contents=False)
    _ldrinv_py = utils._EngineFloat64ArrayType.get_nd(rinv, {'fun_name': fun_name, 'entity': 'rinv'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldrinv = utils._EngineIntScalarType(
        _ldrinv_py, {'fun_name': fun_name, 'entity': '_ldrinv'},
    ).to_ctypes()
    _rinv_cls = utils._EngineFloat64ArrayType(
        rinv, {'fun_name': fun_name, 'entity': 'rinv'},
        exp_shape=(_k_py, _k_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _rinv = _rinv_cls.to_ctypes(can_reference_contents=False)
    _ldc_py = utils._EngineFloat64ArrayType.get_nd(c, {'fun_name': fun_name, 'entity': 'c'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldc = utils._EngineIntScalarType(
        _ldc_py, {'fun_name': fun_name, 'entity': '_ldc'},
    ).to_ctypes()
    _c_cls = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_k_py, _k_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _c = _c_cls.to_ctypes(can_reference_contents=False)
    _wkz = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _ldwkz = utils._EngineIntScalarType(
        _k_py, {'fun_name': fun_name, 'entity': '_ldwkz'},
    ).to_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _iemode,
        _n,
        _k1,
        _k,
        _xbar,
        _ssp,
        _ldssp,
        _r,
        _ldr,
        _result,
        _coef,
        _ldcoef,
        _con,
        _rinv,
        _ldrinv,
        _c,
        _ldc,
        _wkz,
        _ldwkz,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, k = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: k >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s} and k = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: n > k+1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The k by k partition of r which requires inversion is',
                    'not positive definite.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The refinement following inversion has failed.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _r_cls.refresh_py(_r)
        _result_cls.refresh_py(_result)
        _coef_cls.refresh_py(_coef)
        _con_cls.refresh_py(_con)
        _rinv_cls.refresh_py(_rinv)
        _c_cls.refresh_py(_c)
        return_data = None
    _err_class.manage()
    return return_data

def linregm_coeffs_noconst(n, k, sspz, rz, result, coef, rznv, cz):
    u"""
    Multiple linear regression, from correlation-like coefficients,
    without constant term.

    ``linregm_coeffs_noconst`` performs a multiple linear regression
    with no constant on a set of variables whose sums of squares and
    cross-products about zero and correlation-like coefficients are
    given.

    For full information please refer to the NAG Library document for
    g02ch

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02chf.html

    Parameters
    ----------
    n : int
        n, the number of cases used in calculating the sums of squares
        and cross-products and correlation-like coefficients.

    k : int
        k, the number of independent variables in the regression.

    sspz : float, shape (`k`+1, `k`+1)
        `sspz`[i-1,j-1] must be set to S~_{ij}, the sum of
        cross-products about zero for the ith and jth variables, for j =
        1,2,...,k+1, for i = 1,2,...,k+1; terms involving the dependent
        variable appear in row k+1 and column k+1.

    rz : float, shape (`k`+1, `k`+1), modified in place
        `rz`[i-1,j-1] must be set to R~_{ij}, the correlation-like
        coefficient for the ith and jth variables, for j = 1,2,...,k+1,
        for i = 1,2,...,k+1; coefficients involving the dependent
        variable appear in row k+1 and column k+1.

    result : float, shape (13), modified in place
        `On exit`: the following information:

        [table omitted]

    coef : float, shape (`k`, 3), modified in place
        `On exit`: for i = 1,2,...,k, the following information:

        `coef`[i-1,0]

            b_i, the regression coefficient for the ith variable.

        `coef`[i-1,1]

            se(b_i), the standard error of the regression coefficient
            for the ith variable.

        `coef`[i-1,2]

            t(b_i), the t value of the regression coefficient for the
            ith variable.

    rznv : float, shape (`k`, `k`), modified in place
        `On exit`: the inverse of the matrix of correlation-like
        coefficients for the independent variables; that is, the inverse
        of the matrix consisting of the first k rows and columns of
        `rz`.

    cz : float, shape (`k`, `k`), modified in place
        `On exit`: the modified inverse matrix, C, where

            `cz`[i-1,j-1] =
            `rz`[i-1,j-1]*`rznv`[i-1,j-1]/`sspz`[i-1,j-1], i,j =
            1,2,...,k.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `k` = *<value>*.

            Constraint: `k` >= 1.

        (`errno` 3)
            On entry, `n` = *<value>* and `k` = *<value>*.

            Constraint: `n` >= `k`+1.

        (`errno` 5)
            The K by K partition of `rz` which requires inversion is not
            positive definite.

        (`errno` 6)
            The refinement following the actual inversion has failed.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Draper, N R and Smith, H, 1985, `Applied Regression Analysis`, (2nd
    Edition), Wiley
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.linregm_coeffs_noconst'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02chft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, iemode.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, k1.
        utils._BYREF_INT, # Input, k.
        utils._EngineFloat64ArrayType.c_type, # Input, sspz[ldsspz, k+1].
        utils._BYREF_INT, # Input, ldsspz.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, rz[ldrz, k+1].
        utils._BYREF_INT, # Input, ldrz.
        utils._EngineFloat64ArrayType.c_type, # Output, result[13].
        utils._EngineFloat64ArrayType.c_type, # Output, coef[ldcoef, 3].
        utils._BYREF_INT, # Input, ldcoef.
        utils._EngineFloat64ArrayType.c_type, # Output, rznv[ldrznv, k].
        utils._BYREF_INT, # Input, ldrznv.
        utils._EngineFloat64ArrayType.c_type, # Output, cz[ldcz, k].
        utils._BYREF_INT, # Input, ldcz.
        utils._EngineFloat64ArrayType.c_type, # Workspace, wkz.
        utils._BYREF_INT, # Input, ldwkz.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((sspz, 'sspz'), (rz, 'rz'), (coef, 'coef'), (rznv, 'rznv'), (cz, 'cz')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _iemode = utils._EngineIntScalarType.c_type(0)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _k = utils._EngineIntScalarType(
        k, {'fun_name': fun_name, 'entity': 'k'},
    ).to_ctypes()
    _k_py = utils._EngineIntScalarType.to_py(_k)
    _k1 = utils._EngineIntScalarType(
        _k_py+1, {'fun_name': fun_name, 'entity': '_k1'},
    ).to_ctypes()
    _ldsspz_py = utils._EngineFloat64ArrayType.get_nd(sspz, {'fun_name': fun_name, 'entity': 'sspz'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldsspz = utils._EngineIntScalarType(
        _ldsspz_py, {'fun_name': fun_name, 'entity': '_ldsspz'},
    ).to_ctypes()
    _sspz = utils._EngineFloat64ArrayType(
        sspz, {'fun_name': fun_name, 'entity': 'sspz'},
        exp_shape=(_k_py+1, _k_py+1),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _ldrz_py = utils._EngineFloat64ArrayType.get_nd(rz, {'fun_name': fun_name, 'entity': 'rz'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldrz = utils._EngineIntScalarType(
        _ldrz_py, {'fun_name': fun_name, 'entity': '_ldrz'},
    ).to_ctypes()
    _rz_cls = utils._EngineFloat64ArrayType(
        rz, {'fun_name': fun_name, 'entity': 'rz'},
        exp_shape=(_k_py+1, _k_py+1),
        mutual_sorder=_m_sorder,
    )
    _rz = _rz_cls.to_ctypes()
    _result_cls = utils._EngineFloat64ArrayType(
        result, {'fun_name': fun_name, 'entity': 'result'},
        exp_shape=(13,),
        check_contents=False,
    )
    _result = _result_cls.to_ctypes(can_reference_contents=False)
    _ldcoef_py = utils._EngineFloat64ArrayType.get_nd(coef, {'fun_name': fun_name, 'entity': 'coef'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldcoef = utils._EngineIntScalarType(
        _ldcoef_py, {'fun_name': fun_name, 'entity': '_ldcoef'},
    ).to_ctypes()
    _coef_cls = utils._EngineFloat64ArrayType(
        coef, {'fun_name': fun_name, 'entity': 'coef'},
        exp_shape=(_k_py, 3),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _coef = _coef_cls.to_ctypes(can_reference_contents=False)
    _ldrznv_py = utils._EngineFloat64ArrayType.get_nd(rznv, {'fun_name': fun_name, 'entity': 'rznv'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldrznv = utils._EngineIntScalarType(
        _ldrznv_py, {'fun_name': fun_name, 'entity': '_ldrznv'},
    ).to_ctypes()
    _rznv_cls = utils._EngineFloat64ArrayType(
        rznv, {'fun_name': fun_name, 'entity': 'rznv'},
        exp_shape=(_k_py, _k_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _rznv = _rznv_cls.to_ctypes(can_reference_contents=False)
    _ldcz_py = utils._EngineFloat64ArrayType.get_nd(cz, {'fun_name': fun_name, 'entity': 'cz'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldcz = utils._EngineIntScalarType(
        _ldcz_py, {'fun_name': fun_name, 'entity': '_ldcz'},
    ).to_ctypes()
    _cz_cls = utils._EngineFloat64ArrayType(
        cz, {'fun_name': fun_name, 'entity': 'cz'},
        exp_shape=(_k_py, _k_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _cz = _cz_cls.to_ctypes(can_reference_contents=False)
    _wkz = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _ldwkz = utils._EngineIntScalarType(
        _k_py, {'fun_name': fun_name, 'entity': '_ldwkz'},
    ).to_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _iemode,
        _n,
        _k1,
        _k,
        _sspz,
        _ldsspz,
        _rz,
        _ldrz,
        _result,
        _coef,
        _ldcoef,
        _rznv,
        _ldrznv,
        _cz,
        _ldcz,
        _wkz,
        _ldwkz,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, k = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: k >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s} and k = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: n >= k+1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The K by K partition of rz which requires inversion is',
                    'not positive definite.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The refinement following the actual inversion has failed.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _rz_cls.refresh_py(_rz)
        _result_cls.refresh_py(_result)
        _coef_cls.refresh_py(_coef)
        _rznv_cls.refresh_py(_rznv)
        _cz_cls.refresh_py(_cz)
        return_data = None
    _err_class.manage()
    return return_data

def linregm_fit(mean, n, x, m, isx, y, wt, b, se, cov, res, h, q, p, tol, wk):
    u"""
    Fits a general (multiple) linear regression model.

    ``linregm_fit`` performs a general multiple linear regression when
    the independent variables may be linearly dependent. Parameter
    estimates, standard errors, residuals and influence statistics are
    computed. ``linregm_fit`` may be used to perform a weighted
    regression.

    For full information please refer to the NAG Library document for
    g02da

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02daf.html

    Parameters
    ----------
    mean : str, length 1
        Indicates if a mean term is to be included.

        `mean` = 'M'

            A mean term, intercept, will be included in the model.

        `mean` = 'Z'

            The model will pass through the origin, zero-point.

    n : int
        n, the number of observations.

    x : float, shape (`n`, `m`)
        `x`[i-1,j-1] must contain the ith observation for the jth
        independent variable, for j = 1,2,...,`m`, for i = 1,2,...,`n`.

    m : int
        m, the total number of independent variables in the dataset.

    isx : int, shape (`m`)
        Indicates which independent variables are to be included in the
        model.

        `isx`[j-1] > 0

            The variable contained in the jth column of `x` is included
            in the regression model.

    y : float, shape (`n`)
        y, the observations on the dependent variable.

    wt : None or float, shape (`n`)
        If provided `wt` must contain the weights to be used with the
        model.

        If `wt`[i-1] = 0.0, the ith observation is not included in the
        model, in which case the effective number of observations is the
        number of observations with nonzero weights.

        The values of `res` and `h` will be set to zero for observations
        with zero weights.

        If `wt` is not provided the effective number of observations is
        n.

    b : float, shape (ip), modified in place
        `On exit`: `b`[i-1], i = 1,2,...,ip contains the least squares
        estimates of the parameters of the regression model, beta^.

        If `mean` = 'M', `b`[0] will contain the estimate of the mean
        parameter and `b`[i] will contain the coefficient of the
        variable contained in column j of `x`, where `isx`[j-1] is the
        ith positive value in the array `isx`.

        If `mean` = 'Z', `b`[i-1] will contain the coefficient of the
        variable contained in column j of `x`, where `isx`[j-1] is the
        ith positive value in the array `isx`.

    se : float, shape (ip), modified in place
        `On exit`: `se`[i-1], i = 1,2,...,ip contains the standard
        errors of the ip parameter estimates given in `b`.

    cov : float, shape (ip*(ip+1)/2), modified in place
        `On exit`: the first ip*(ip+1)/2 elements of `cov` contain the
        upper triangular part of the variance-covariance matrix of the
        ip parameter estimates given in `b`. They are stored packed by
        column, i.e., the covariance between the parameter estimate
        given in `b`[i-1] and the parameter estimate given in `b`[j-1],
        j >= i, is stored in `cov`[j*(j-1)/2+i-1].

    res : float, shape (`n`), modified in place
        `On exit`: the (weighted) residuals, r_i, for i = 1,2,...,n.

    h : float, shape (`n`), modified in place
        `On exit`: the diagonal elements of H, h_i, for i = 1,2,...,n.

    q : float, shape (`n`, ip+1), modified in place
        `On exit`: the results of the QR decomposition:

            the first column of `q` contains c;

            the upper triangular part of columns 2 to ip+1 contain the R
            matrix;

            the strictly lower triangular part of columns 2 to ip+1
            contain details of the Q matrix.

    p : float, shape (2*ip+ip*ip), modified in place
        `On exit`: details of the QR decomposition and SVD if used.

        If `svd` = False, only the first ip elements of `p` are used
        these will contain the zeta values for the QR decomposition (see
        ``lapackeig.dgeqrf`` for details).

        If `svd` = True, the first ip elements of `p` will contain the
        zeta values for the QR decomposition (see ``lapackeig.dgeqrf``
        for details) and the next ip elements of `p` contain singular
        values.

        The following ip by ip elements contain the matrix P* stored by
        columns.

    tol : float
        The value of `tol` is used to decide if the independent
        variables are of full rank and if not what is the rank of the
        independent variables. The smaller the value of `tol` the
        stricter the criterion for selecting the singular value
        decomposition. If `tol` = 0.0, the singular value decomposition
        will never be used; this may cause run time errors or inaccurate
        results if the independent variables are not of full rank.

        Suggested value: `tol` = 0.000001.

    wk : float, shape (max(2, 5*(ip-1)+ip*ip)), modified in place
        `On exit`: if on exit `svd` = True, `wk` contains information
        which is needed by ``linregm_fit_newvar``; otherwise `wk` is
        used as workspace.

    Returns
    -------
    rss : float
        The residual sum of squares for the regression.

    idf : int
        The degrees of freedom associated with the residual sum of
        squares.

    svd : bool
        If a singular value decomposition has been performed then `svd`
        will be True, otherwise `svd` will be False.

    irank : int
        The rank of the independent variables.

        If `svd` = False, `irank` = ip.

        If `svd` = True, `irank` is an estimate of the rank of the
        independent variables.

        `irank` is calculated as the number of singular values greater
        that `tol`* (largest singular value).

        It is possible for the SVD to be carried out but `irank` to be
        returned as ip.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, ip = *<value>*.

            Constraint: ip >= 1.

        (`errno` 1)
            On entry, `tol` = *<value>*.

            Constraint: `tol` >= 0.0.

        (`errno` 1)
            On entry, ip = *<value>* and `n` = *<value>*.

            Constraint: ip <= `n`.

        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 2.

        (`errno` 2)
            On entry, weight = *<value>*.

            Constraint: weight = 'U' or 'W'.

        (`errno` 2)
            On entry, `mean` = *<value>*.

            Constraint: `mean` = 'M' or 'Z'.

        (`errno` 3)
            On entry, `wt`[*<value>*] < 0.0.

            Constraint: `wt`[i-1] >= 0.0, for i = 1,2,...,n.

        (`errno` 4)
            On entry, `isx`[*<value>*] < 0.

            Constraint: `isx`[i-1] >= 0.0, for i = 1,2,...,m.

        (`errno` 4)
            On entry, ip = *<value>*.

            Constraint: ip must be compatible with the number of nonzero
            elements in `isx`.

        (`errno` 6)
            SVD solution failed to converge.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 5)
            The degrees of freedom for the residuals are zero, i.e., the
            designated number of arguments is equal to the effective
            number of observations. In this case the parameter estimates
            will be returned along with the diagonal elements of H, but
            neither standard errors nor the variance-covariance matrix
            will be calculated.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Cook, R D and Weisberg, S, 1982, `Residuals and Influence in
    Regression`, Chapman and Hall

    Draper, N R and Smith, H, 1985, `Applied Regression Analysis`, (2nd
    Edition), Wiley

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Hammarling, S, 1985, `The singular value decomposition in
    multivariate statistics`, SIGNUM Newsl. (20(3)), 2--25

    McCullagh, P and Nelder, J A, 1983, `Generalized Linear Models`,
    Chapman and Hall

    Searle, S R, 1971, `Linear Models`, Wiley

    See Also
    --------
    :meth:`naginterfaces.library.examples.correg.linregm_fit_ex.main`
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.linregm_fit'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02daft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, mean(len=1).
        utils._BYREF_STRING, # Input, weight(len=1).
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[ldx, m].
        utils._BYREF_INT, # Input, ldx.
        utils._BYREF_INT, # Input, m.
        utils._EngineIntArrayType.c_type, # Input, isx[m].
        utils._BYREF_INT, # Input, ip.
        utils._EngineFloat64ArrayType.c_type, # Input, y[n].
        utils._EngineFloat64ArrayType.c_type, # Input, wt[n].
        utils._BYREF_FLOAT64, # Output, rss.
        utils._BYREF_INT, # Output, idf.
        utils._EngineFloat64ArrayType.c_type, # Output, b[ip].
        utils._EngineFloat64ArrayType.c_type, # Output, se[ip].
        utils._EngineFloat64ArrayType.c_type, # Output, cov[ip*(ip+1)//2].
        utils._EngineFloat64ArrayType.c_type, # Output, res[n].
        utils._EngineFloat64ArrayType.c_type, # Output, h[n].
        utils._EngineFloat64ArrayType.c_type, # Output, q[ldq, ip+1].
        utils._BYREF_INT, # Input, ldq.
        utils._BYREF_LOGICAL, # Output, svd.
        utils._BYREF_INT, # Output, irank.
        utils._EngineFloat64ArrayType.c_type, # Output, p[2*ip+ip*ip].
        utils._BYREF_FLOAT64, # Input, tol.
        utils._EngineFloat64ArrayType.c_type, # Output, wk[max(2, 5*(ip-1)+ip*ip)].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of mean.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of weight.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((x, 'x'), (q, 'q')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _mean = utils._EngineCharScalarType(
        mean, {'fun_name': fun_name, 'entity': 'mean'}, exp_chrlen=1,
    ).to_ctypes()
    if wt is None:
        _weight = utils._EngineCharScalarType(
            'U', {'fun_name': fun_name, 'entity': '_weight'}, exp_chrlen=1,
        ).to_ctypes()
    else:
        _weight = utils._EngineCharScalarType(
            'W', {'fun_name': fun_name, 'entity': '_weight'}, exp_chrlen=1,
        ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _ldx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py, _m_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _isx = utils._EngineIntArrayType(
        isx, {'fun_name': fun_name, 'entity': 'isx'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _vl_p_py = 0
    for _vl_i_py in range(1, _m_py+1):
        if utils._EngineIntArrayType.get_element(isx, {'fun_name': fun_name, 'entity': 'isx'}, (_vl_i_py-1,)) > 0:
            _vl_p_py = _vl_p_py + 1
    _mean_py = utils._EngineCharScalarType.to_py(_mean)
    if _mean_py in ('M', 'm'):
        _vl_ip_py = _vl_p_py+1
    else:
        _vl_ip_py = _vl_p_py
    _ip = utils._EngineIntScalarType(
        _vl_ip_py, {'fun_name': fun_name, 'entity': '_ip'},
    ).to_ctypes()
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _wt = utils._EngineFloat64ArrayType(
        wt, {'fun_name': fun_name, 'entity': 'wt'},
        exp_shape=(_n_py,),
        allow_none=True,
    ).to_ctypes()
    _rss = utils._EngineFloat64ScalarType.empty_ctypes()
    _idf = utils._EngineIntScalarType.empty_ctypes()
    _ip_py = utils._EngineIntScalarType.to_py(_ip)
    _b_cls = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_ip_py,),
        check_contents=False,
    )
    _b = _b_cls.to_ctypes(can_reference_contents=False)
    _se_cls = utils._EngineFloat64ArrayType(
        se, {'fun_name': fun_name, 'entity': 'se'},
        exp_shape=(_ip_py,),
        check_contents=False,
    )
    _se = _se_cls.to_ctypes(can_reference_contents=False)
    _cov_cls = utils._EngineFloat64ArrayType(
        cov, {'fun_name': fun_name, 'entity': 'cov'},
        exp_shape=(_ip_py*(_ip_py+1)//2,),
        check_contents=False,
    )
    _cov = _cov_cls.to_ctypes(can_reference_contents=False)
    _res_cls = utils._EngineFloat64ArrayType(
        res, {'fun_name': fun_name, 'entity': 'res'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _res = _res_cls.to_ctypes(can_reference_contents=False)
    _h_cls = utils._EngineFloat64ArrayType(
        h, {'fun_name': fun_name, 'entity': 'h'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _h = _h_cls.to_ctypes(can_reference_contents=False)
    _ldq_py = utils._EngineFloat64ArrayType.get_nd(q, {'fun_name': fun_name, 'entity': 'q'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldq = utils._EngineIntScalarType(
        _ldq_py, {'fun_name': fun_name, 'entity': '_ldq'},
    ).to_ctypes()
    _q_cls = utils._EngineFloat64ArrayType(
        q, {'fun_name': fun_name, 'entity': 'q'},
        exp_shape=(_n_py, _ip_py+1),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _q = _q_cls.to_ctypes(can_reference_contents=False)
    _svd = utils._EngineLogicalScalarType.empty_ctypes()
    _irank = utils._EngineIntScalarType.empty_ctypes()
    _p_cls = utils._EngineFloat64ArrayType(
        p, {'fun_name': fun_name, 'entity': 'p'},
        exp_shape=(2*_ip_py+_ip_py*_ip_py,),
        check_contents=False,
    )
    _p = _p_cls.to_ctypes(can_reference_contents=False)
    _tol = utils._EngineFloat64ScalarType(
        tol, {'fun_name': fun_name, 'entity': 'tol'},
    ).to_ctypes()
    _wk_cls = utils._EngineFloat64ArrayType(
        wk, {'fun_name': fun_name, 'entity': 'wk'},
        exp_shape=(max(2, 5*(_ip_py-1)+_ip_py*_ip_py),),
        check_contents=False,
    )
    _wk = _wk_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _mean,
        _weight,
        _n,
        _x,
        _ldx,
        _m,
        _isx,
        _ip,
        _y,
        _wt,
        _rss,
        _idf,
        _b,
        _se,
        _cov,
        _res,
        _h,
        _q,
        _ldq,
        _svd,
        _irank,
        _p,
        _tol,
        _wk,
        _errbuf,
        _ierr,
        _mean._length_,
        _weight._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99988:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ip >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, tol = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: tol >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ip = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: ip <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, weight = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: weight = \'U\' or \'W\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mean = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mean = \'M\' or \'Z\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, wt[{:s}] < 0.0.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: wt[i-1] >= 0.0, for i = 1,2,...,n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99986:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, isx[{:s}] < 0.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: isx[i-1] >= 0.0, for i = 1,2,...,m.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99987:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ip must be compatible with the number of nonzero elements in',
                    'isx.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99991:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The degrees of freedom for the residuals are zero, i.e., the',
                    'designated number of arguments is equal to the effective number of',
                    'observations. In this case the parameter estimates will be returned along',
                    'with the diagonal elements of H, but neither standard errors nor the',
                    'variance-covariance matrix will be calculated.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'SVD solution failed to converge.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _b_cls.refresh_py(_b)
        _se_cls.refresh_py(_se)
        _cov_cls.refresh_py(_cov)
        _res_cls.refresh_py(_res)
        _h_cls.refresh_py(_h)
        _q_cls.refresh_py(_q)
        _p_cls.refresh_py(_p)
        _wk_cls.refresh_py(_wk)
        _rss_py = utils._EngineFloat64ScalarType.to_py(_rss)
        _idf_py = utils._EngineIntScalarType.to_py(_idf)
        _svd_py = utils._EngineLogicalScalarType.to_py(_svd)
        _irank_py = utils._EngineIntScalarType.to_py(_irank)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['rss', 'idf', 'svd', 'irank'],
        )(
            _rss_py,
            _idf_py,
            _svd_py,
            _irank_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def linregm_obs_edit(update, mean, m, isx, q, x, ix, y, wt, rss):
    u"""
    Add/delete an observation to/from a general linear regression model.

    ``linregm_obs_edit`` adds or deletes an observation from a general
    regression model fitted by ``linregm_fit``.

    For full information please refer to the NAG Library document for
    g02dc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02dcf.html

    Parameters
    ----------
    update : str, length 1
        Indicates if an observation is to be added or deleted.

        `update` = 'A'

            The observation is added.

        `update` = 'D'

            The observation is deleted.

    mean : str, length 1
        Indicates if a mean has been used in the model.

        `mean` = 'M'

            A mean term or intercept will have been included in the
            model by ``linregm_fit``.

        `mean` = 'Z'

            A model with no mean term or intercept will have been fitted
            by ``linregm_fit``.

    m : int
        m, the total number of independent variables in the dataset.

    isx : int, shape (`m`)
        If `isx`[j-1] is greater than 0, the value contained in
        `x`[(j-1)*`ix`] is to be included as a value of x^T, for j =
        1,2,...,`m`.

    q : float, shape (ip, ip+1), modified in place
        `On entry`: must be array `q` as output by ``linregm_fit``,
        ``linregm_var_add``, ``linregm_var_del`` or
        ``linregm_fit_onestep``, or a previous call to
        ``linregm_obs_edit``.

        `On exit`: the first ip elements of the first column of `q` will
        contain c_1^* the upper triangular part of columns 2 to ip+1
        will contain R* the remainder is unchanged.

    x : float, shape ((`m`-1)*`ix`+1)
        The ip values for the dependent variables of the new
        observation, x^T. The positions will depend on the value of
        `ix`.

    ix : int
        The increment for elements of `x`. Two situations are common:

        `ix` = 1

            The values of x are to be chosen from consecutive locations
            in `x`, i.e., `x`[0],`x`[1],...,`x`[`m`-1].

        `ix` = ldx

            The values of x are to be chosen from a row of a
            two-dimensional array with first dimension ldx, i.e.,
            `x`[0],`x`[ldx],...,`x`[(`m`-1)ldx].

    y : float
        The value of the dependent variable for the new observation,
        y_new.

    wt : None or float
        If provided, `wt` must contain the weight to be used with the
        new observation.

        If `wt` is None, the observation is not included in the model.

    rss : float
        The value of the residual sums of squares for the original set
        of observations.

    Returns
    -------
    rss : float
        The updated values of the residual sums of squares.

        **Note:** this will only be valid if the model is of full rank.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, *<value>* elements of `isx` > 0 instead of ip-1
            (for mean) = *<value>*.

        (`errno` 1)
            On entry, `ix` = *<value>*.

            Constraint: `ix` >= 1.

        (`errno` 1)
            On entry, *<value>* elements of `isx` > 0 instead of ip =
            *<value>*.

        (`errno` 1)
            On entry, `mean` = *<value>*.

            Constraint: `mean` = 'M' or 'Z'.

        (`errno` 1)
            On entry, `rss` = *<value>*.

            Constraint: `rss` >= 0.0.

        (`errno` 1)
            On entry, `update` = *<value>*.

            Constraint: `update` = 'A' or 'D'.

        (`errno` 1)
            On entry, weight = *<value>*.

            Constraint: weight = 'U' or 'W'.

        (`errno` 1)
            On entry, ip = *<value>*.

            Constraint: ip >= 1.

        (`errno` 2)
            On entry, `wt` = *<value>*.

            Constraint: `wt` >= 0.0.

        (`errno` 3)
            The R matrix could not be updated.

        (`errno` 4)
            The residual sums of squares cannot be updated.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Hammarling, S, 1985, `The singular value decomposition in
    multivariate statistics`, SIGNUM Newsl. (20(3)), 2--25
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.linregm_obs_edit'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02dcft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, update(len=1).
        utils._BYREF_STRING, # Input, mean(len=1).
        utils._BYREF_STRING, # Input, weight(len=1).
        utils._BYREF_INT, # Input, m.
        utils._EngineIntArrayType.c_type, # Input, isx[m].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, q[ldq, ip+1].
        utils._BYREF_INT, # Input, ldq.
        utils._BYREF_INT, # Input, ip.
        utils._EngineFloat64ArrayType.c_type, # Input, x[(m-1)*ix+1].
        utils._BYREF_INT, # Input, ix.
        utils._BYREF_FLOAT64, # Input, y.
        utils._BYREF_FLOAT64, # Input, wt.
        utils._BYREF_FLOAT64, # Input/Output, rss.
        utils._EngineFloat64ArrayType.c_type, # Workspace, wk.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of update.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of mean.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of weight.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((q, 'q'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _update = utils._EngineCharScalarType(
        update, {'fun_name': fun_name, 'entity': 'update'}, exp_chrlen=1,
    ).to_ctypes()
    _mean = utils._EngineCharScalarType(
        mean, {'fun_name': fun_name, 'entity': 'mean'}, exp_chrlen=1,
    ).to_ctypes()
    if wt is None:
        _weight = utils._EngineCharScalarType(
            'U', {'fun_name': fun_name, 'entity': '_weight'}, exp_chrlen=1,
        ).to_ctypes()
    else:
        _weight = utils._EngineCharScalarType(
            'W', {'fun_name': fun_name, 'entity': '_weight'}, exp_chrlen=1,
        ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _isx = utils._EngineIntArrayType(
        isx, {'fun_name': fun_name, 'entity': 'isx'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _vl_p_py = 0
    for _vl_i_py in range(1, _m_py+1):
        if utils._EngineIntArrayType.get_element(isx, {'fun_name': fun_name, 'entity': 'isx'}, (_vl_i_py-1,)) > 0:
            _vl_p_py = _vl_p_py + 1
    _mean_py = utils._EngineCharScalarType.to_py(_mean)
    if _mean_py in ('M', 'm'):
        _vl_ip_py = _vl_p_py+1
    else:
        _vl_ip_py = _vl_p_py
    _ldq_py = utils._EngineFloat64ArrayType.get_nd(q, {'fun_name': fun_name, 'entity': 'q'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldq = utils._EngineIntScalarType(
        _ldq_py, {'fun_name': fun_name, 'entity': '_ldq'},
    ).to_ctypes()
    _ip = utils._EngineIntScalarType(
        _vl_ip_py, {'fun_name': fun_name, 'entity': '_ip'},
    ).to_ctypes()
    _ip_py = utils._EngineIntScalarType.to_py(_ip)
    _q_cls = utils._EngineFloat64ArrayType(
        q, {'fun_name': fun_name, 'entity': 'q'},
        exp_shape=(_ip_py, _ip_py+1),
        mutual_sorder=_m_sorder,
    )
    _q = _q_cls.to_ctypes()
    _ix = utils._EngineIntScalarType(
        ix, {'fun_name': fun_name, 'entity': 'ix'},
    ).to_ctypes()
    _ix_py = utils._EngineIntScalarType.to_py(_ix)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=((_m_py-1)*_ix_py+1,),
    ).to_ctypes()
    _y = utils._EngineFloat64ScalarType(
        y, {'fun_name': fun_name, 'entity': 'y'},
    ).to_ctypes()
    if wt is not None:
        if wt is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'wt'},
                'must not be None when wt is not None',
            )
    if wt is None:
        _wt = utils._EngineFloat64ScalarType.c_type(0.0)
    else:
        _wt = utils._EngineFloat64ScalarType(
            wt, {'fun_name': fun_name, 'entity': 'wt'},
            allow_none=True,
        ).to_ctypes()
    _rss = utils._EngineFloat64ScalarType(
        rss, {'fun_name': fun_name, 'entity': 'rss'},
    ).to_ctypes()
    _wk = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _update,
        _mean,
        _weight,
        _m,
        _isx,
        _q,
        _ldq,
        _ip,
        _x,
        _ix,
        _y,
        _wt,
        _rss,
        _wk,
        _errbuf,
        _ierr,
        _update._length_,
        _mean._length_,
        _weight._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99988:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, {:s} elements of isx > 0'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'instead of ip-1 (for mean) = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ix = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ix >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, {:s} elements of isx > 0'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'instead of ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mean = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mean = \'M\' or \'Z\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, rss = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: rss >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, update = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: update = \'A\' or \'D\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, weight = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: weight = \'U\' or \'W\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ip >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, wt = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: wt >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The R matrix could not be updated.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The residual sums of squares cannot be updated.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _q_cls.refresh_py(_q)
        _rss_py = utils._EngineFloat64ScalarType.to_py(_rss)
        return_data = _rss_py
    _err_class.manage()
    return return_data

def linregm_update(n, ip, q, rss, b, se, cov, p, tol):
    u"""
    Estimates of linear parameters and general linear regression model
    from updated model.

    ``linregm_update`` calculates the regression parameters for a
    general linear regression model.
    It is intended to be called after ``linregm_obs_edit``,
    ``linregm_var_add`` or ``linregm_var_del``.

    For full information please refer to the NAG Library document for
    g02dd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02ddf.html

    Parameters
    ----------
    n : int
        The number of observations.

    ip : int
        p, the number of terms in the regression model.

    q : float, shape (`n`, `ip`+1)
        Must be the array `q` as output by ``linregm_obs_edit``,
        ``linregm_var_add``, ``linregm_var_del`` or
        ``linregm_fit_onestep``. If on entry `rss` <= 0.0 then all `n`
        elements of c are needed. This is provided by functions
        ``linregm_var_add``, ``linregm_var_del`` or
        ``linregm_fit_onestep``.

    rss : float
        Either the residual sum of squares or a value less than or equal
        to 0.0 to indicate that the residual sum of squares is to be
        calculated by the function.

    b : float, shape (`ip`), modified in place
        `On exit`: the estimates of the p parameters, beta^.

    se : float, shape (`ip`), modified in place
        `On exit`: the standard errors of the p parameters given in `b`.

    cov : float, shape (`ip`*(`ip`+1)/2), modified in place
        `On exit`: the upper triangular part of the variance-covariance
        matrix of the p parameter estimates given in `b`. They are
        stored packed by column, i.e., the covariance between the
        parameter estimate given in `b`[i-1] and the parameter estimate
        given in `b`[j-1], j >= i, is stored in `cov`[j*(j-1)/2+i-1].

    p : float, shape (`ip`*`ip`+2*`ip`), modified in place
        `On entry`: must be array `p` as output by ``linregm_fit``,
        ``linregm_var_add`` or ``linregm_fit_onestep``, or a previous
        call to ``linregm_update``.

        `On exit`: contains details of the singular value decomposition
        if used.

        If `svd` = False, `p` is not referenced.

        If `svd` = True, the first `ip` elements of `p` are unchanged,
        the next `ip` values contain the singular values.

        The following `ip`*`ip` values contain the matrix P* stored by
        columns.

    tol : float
        The value of `tol` is used to decide if the independent
        variables are of full rank and, if not, what is the rank of the
        independent variables. The smaller the value of `tol` the
        stricter the criterion for selecting the singular value
        decomposition. If `tol` = 0.0, the singular value decomposition
        will never be used, this may cause run time errors or
        inaccuracies if the independent variables are not of full rank.

        Suggested value: `tol` = 0.000001.

    Returns
    -------
    rss : float
        If `rss` <= 0.0 on entry, then on exit `rss` will contain the
        residual sum of squares as calculated by ``linregm_update``.

        If `rss` was positive on entry, it will be unchanged.

    idf : int
        The degrees of freedom associated with the residual sum of
        squares.

    svd : bool
        If a singular value decomposition has been performed, `svd` =
        True, otherwise `svd` = False.

    irank : int
        The rank of the independent variables.

        If `svd` = False, `irank` = `ip`.

        If `svd` = True, `irank` is an estimate of the rank of the
        independent variables.

        `irank` is calculated as the number of singular values greater
        than `tol`* (largest singular value).

        It is possible for the SVD to be carried out but `irank` to be
        returned as `ip`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `tol` = *<value>*.

            Constraint: `tol` >= 0.0.

        (`errno` 1)
            On entry, ldq = *<value>* and `ip` = *<value>*.

            Constraint: if `rss` <= 0.0, ldq >= `n`.

        (`errno` 1)
            On entry, `ip` = *<value>*.

            Constraint: `ip` >= 1.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 2)
            The degrees of freedom for error are less than or equal to
            0. In this case the estimates of beta are returned but not
            the standard errors or covariances.

        (`errno` 3)
            SVD solution failed to converge.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Hammarling, S, 1985, `The singular value decomposition in
    multivariate statistics`, SIGNUM Newsl. (20(3)), 2--25

    Searle, S R, 1971, `Linear Models`, Wiley
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.linregm_update'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02ddft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, ip.
        utils._EngineFloat64ArrayType.c_type, # Input, q[ldq, ip+1].
        utils._BYREF_INT, # Input, ldq.
        utils._BYREF_FLOAT64, # Input/Output, rss.
        utils._BYREF_INT, # Output, idf.
        utils._EngineFloat64ArrayType.c_type, # Output, b[ip].
        utils._EngineFloat64ArrayType.c_type, # Output, se[ip].
        utils._EngineFloat64ArrayType.c_type, # Output, cov[ip*(ip+1)//2].
        utils._BYREF_LOGICAL, # Output, svd.
        utils._BYREF_INT, # Output, irank.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, p[ip*ip+2*ip].
        utils._BYREF_FLOAT64, # Input, tol.
        utils._EngineFloat64ArrayType.c_type, # Workspace, wk.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((q, 'q'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _ip = utils._EngineIntScalarType(
        ip, {'fun_name': fun_name, 'entity': 'ip'},
    ).to_ctypes()
    _ldq_py = utils._EngineFloat64ArrayType.get_nd(q, {'fun_name': fun_name, 'entity': 'q'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldq = utils._EngineIntScalarType(
        _ldq_py, {'fun_name': fun_name, 'entity': '_ldq'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _ip_py = utils._EngineIntScalarType.to_py(_ip)
    _q = utils._EngineFloat64ArrayType(
        q, {'fun_name': fun_name, 'entity': 'q'},
        exp_shape=(_n_py, _ip_py+1),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _rss = utils._EngineFloat64ScalarType(
        rss, {'fun_name': fun_name, 'entity': 'rss'},
    ).to_ctypes()
    _idf = utils._EngineIntScalarType.empty_ctypes()
    _b_cls = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_ip_py,),
        check_contents=False,
    )
    _b = _b_cls.to_ctypes(can_reference_contents=False)
    _se_cls = utils._EngineFloat64ArrayType(
        se, {'fun_name': fun_name, 'entity': 'se'},
        exp_shape=(_ip_py,),
        check_contents=False,
    )
    _se = _se_cls.to_ctypes(can_reference_contents=False)
    _cov_cls = utils._EngineFloat64ArrayType(
        cov, {'fun_name': fun_name, 'entity': 'cov'},
        exp_shape=(_ip_py*(_ip_py+1)//2,),
        check_contents=False,
    )
    _cov = _cov_cls.to_ctypes(can_reference_contents=False)
    _svd = utils._EngineLogicalScalarType.empty_ctypes()
    _irank = utils._EngineIntScalarType.empty_ctypes()
    _p_cls = utils._EngineFloat64ArrayType(
        p, {'fun_name': fun_name, 'entity': 'p'},
        exp_shape=(_ip_py*_ip_py+2*_ip_py,),
    )
    _p = _p_cls.to_ctypes()
    _tol = utils._EngineFloat64ScalarType(
        tol, {'fun_name': fun_name, 'entity': 'tol'},
    ).to_ctypes()
    _wk = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _ip,
        _q,
        _ldq,
        _rss,
        _idf,
        _b,
        _se,
        _cov,
        _svd,
        _irank,
        _p,
        _tol,
        _wk,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, tol = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: tol >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ldq = {:s} and ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if rss <= 0.0, ldq >= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ip >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The degrees of freedom for error are less than or equal to 0. In this case',
                    'the estimates of beta are returned but not the standard errors or',
                    'covariances.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'SVD solution failed to converge.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _rss_py = utils._EngineFloat64ScalarType.to_py(_rss)
        _b_cls.refresh_py(_b)
        _se_cls.refresh_py(_se)
        _cov_cls.refresh_py(_cov)
        _p_cls.refresh_py(_p)
        _idf_py = utils._EngineIntScalarType.to_py(_idf)
        _svd_py = utils._EngineLogicalScalarType.to_py(_svd)
        _irank_py = utils._EngineIntScalarType.to_py(_irank)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['rss', 'idf', 'svd', 'irank'],
        )(
            _rss_py,
            _idf_py,
            _svd_py,
            _irank_py,
        )
    _err_class.manage()
    return return_data

def linregm_var_add(n, ip, q, p, wt, x, tol):
    u"""
    Add a new independent variable to a general linear regression model.

    ``linregm_var_add`` adds a new independent variable to a general
    linear regression model.

    For full information please refer to the NAG Library document for
    g02de

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02def.html

    Parameters
    ----------
    n : int
        n, the number of observations.

    ip : int
        p, the number of independent variables already in the model.

    q : float, shape (`n`, `ip`+2), modified in place
        `On entry`: if `ip` != 0, `q` must contain the results of the QR
        decomposition for the model with p parameters as returned by
        ``linregm_fit`` or a previous call to ``linregm_var_add``.

        If `ip` = 0, the first column of `q` should contain the n values
        of the dependent variable, y.

        `On exit`: the results of the QR decomposition for the model
        with p+1 parameters:

            the first column of `q` contains the updated value of c;

            the columns 2 to `ip`+1 are unchanged;

            the first `ip`+1 elements of column `ip`+2 contain the new
            column of R, while the remaining `n`-`ip`-1 elements contain
            details of the matrix Q_{p+1}.

    p : float, shape (`ip`+1), modified in place
        `On entry`: contains further details of the QR decomposition
        used. The first `ip` elements of `p` must contain the zeta
        values for the QR decomposition (see ``lapackeig.dgeqrf`` for
        details).

        The first `ip` elements of array `p` are provided by
        ``linregm_fit`` or by previous calls to ``linregm_var_add``.

        `On exit`: the first `ip` elements of `p` are unchanged and the
        (`ip`+1)th element contains the zeta value for Q_{p+1}.

    wt : None or float, shape (`n`)
        If provided `wt` must contain the weights to be used with the
        model.

        If `wt`[i-1] = 0.0, the ith observation is not included in the
        model, in which case the effective number of observations is the
        number of observations with nonzero weights.

        If `wt` is not provided the effective number of observations is
        n.

    x : float, shape (`n`)
        x, the new independent variable.

    tol : float
        The value of `tol` is used to decide if the new independent
        variable is linearly related to independent variables already
        included in the model. If the new variable is linearly related
        then c is not updated. The smaller the value of `tol` the
        stricter the criterion for deciding if there is a linear
        relationship.

        Suggested value: `tol` = 0.000001.

    Returns
    -------
    rss : float
        The residual sum of squares for the new fitted model.

        **Note:** this will only be valid if the model is of full rank,
        see [Further Comments].

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `tol` = *<value>*.

            Constraint: `tol` > 0.0.

        (`errno` 1)
            On entry, `ip` = *<value>* and `n` = *<value>*.

            Constraint: `ip` < `n`.

        (`errno` 1)
            On entry, weight = *<value>*.

            Constraint: weight = 'W' or 'U'.

        (`errno` 1)
            On entry, `ip` = *<value>*.

            Constraint: `ip` >= 0.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 2)
            On entry, `wt`[*<value>*] < 0.0.

            Constraint: `wt`[i-1] >= 0.0, for i = 1,2,...,n.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            `x` variable is a linear combination of existing model
            terms.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Draper, N R and Smith, H, 1985, `Applied Regression Analysis`, (2nd
    Edition), Wiley

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Hammarling, S, 1985, `The singular value decomposition in
    multivariate statistics`, SIGNUM Newsl. (20(3)), 2--25

    McCullagh, P and Nelder, J A, 1983, `Generalized Linear Models`,
    Chapman and Hall

    Searle, S R, 1971, `Linear Models`, Wiley
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.linregm_var_add'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02deft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, weight(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, ip.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, q[ldq, ip+2].
        utils._BYREF_INT, # Input, ldq.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, p[ip+1].
        utils._EngineFloat64ArrayType.c_type, # Input, wt[n].
        utils._EngineFloat64ArrayType.c_type, # Input, x[n].
        utils._BYREF_FLOAT64, # Output, rss.
        utils._BYREF_FLOAT64, # Input, tol.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of weight.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((q, 'q'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    if wt is None:
        _weight = utils._EngineCharScalarType(
            'U', {'fun_name': fun_name, 'entity': '_weight'}, exp_chrlen=1,
        ).to_ctypes()
    else:
        _weight = utils._EngineCharScalarType(
            'W', {'fun_name': fun_name, 'entity': '_weight'}, exp_chrlen=1,
        ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _ip = utils._EngineIntScalarType(
        ip, {'fun_name': fun_name, 'entity': 'ip'},
    ).to_ctypes()
    _ldq_py = utils._EngineFloat64ArrayType.get_nd(q, {'fun_name': fun_name, 'entity': 'q'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldq = utils._EngineIntScalarType(
        _ldq_py, {'fun_name': fun_name, 'entity': '_ldq'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _ip_py = utils._EngineIntScalarType.to_py(_ip)
    _q_cls = utils._EngineFloat64ArrayType(
        q, {'fun_name': fun_name, 'entity': 'q'},
        exp_shape=(_n_py, _ip_py+2),
        mutual_sorder=_m_sorder,
    )
    _q = _q_cls.to_ctypes()
    _p_cls = utils._EngineFloat64ArrayType(
        p, {'fun_name': fun_name, 'entity': 'p'},
        exp_shape=(_ip_py+1,),
    )
    _p = _p_cls.to_ctypes()
    _wt = utils._EngineFloat64ArrayType(
        wt, {'fun_name': fun_name, 'entity': 'wt'},
        exp_shape=(_n_py,),
        allow_none=True,
    ).to_ctypes()
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _rss = utils._EngineFloat64ScalarType.empty_ctypes()
    _tol = utils._EngineFloat64ScalarType(
        tol, {'fun_name': fun_name, 'entity': 'tol'},
    ).to_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _weight,
        _n,
        _ip,
        _q,
        _ldq,
        _p,
        _wt,
        _x,
        _rss,
        _tol,
        _errbuf,
        _ierr,
        _weight._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, tol = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: tol > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ip = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: ip < n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, weight = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: weight = \'W\' or \'U\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ip >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, wt[{:s}] < 0.0.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: wt[i-1] >= 0.0, for i = 1,2,...,n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99995:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'x variable is a linear combination of existing model terms.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _q_cls.refresh_py(_q)
        _p_cls.refresh_py(_p)
        _rss_py = utils._EngineFloat64ScalarType.to_py(_rss)
        return_data = _rss_py
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = _collections.namedtuple(
                'nag_return_data',
                ['rss'],
            )(
                _rss_py,
            )
    _err_class.manage()
    return return_data

def linregm_var_del(ip, q, indx, rss):
    u"""
    Delete an independent variable from a general linear regression
    model.

    ``linregm_var_del`` deletes an independent variable from a general
    linear regression model.

    For full information please refer to the NAG Library document for
    g02df

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02dff.html

    Parameters
    ----------
    ip : int
        p, the number of independent variables already in the model.

    q : float, shape (`ip`, `ip`+1), modified in place
        `On entry`: the results of the QR decomposition as returned by
        functions ``linregm_fit``, ``linregm_obs_edit``,
        ``linregm_var_add`` or ``linregm_fit_onestep``, or previous
        calls to ``linregm_var_del``.

        `On exit`: the updated QR decomposition.

    indx : int
        Indicates which independent variable is to be deleted from the
        model.

    rss : float
        The residual sum of squares for the full regression.

    Returns
    -------
    rss : float
        The residual sum of squares with the (`indx`)th variable
        removed. Note that the residual sum of squares will only be
        valid if the regression is of full rank, otherwise the residual
        sum of squares should be obtained using ``linregm_update``.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `rss` = *<value>*.

            Constraint: `rss` >= 0.0.

        (`errno` 1)
            On entry, `indx` = *<value>*.

            Constraint: `indx` >= 1 and `indx` <= `ip`.

        (`errno` 1)
            On entry, `ip` = *<value>*.

            Constraint: `ip` >= 1.

        (`errno` 2)
            On entry, `q`[*<value>*,*<value>*] = 0.0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Hammarling, S, 1985, `The singular value decomposition in
    multivariate statistics`, SIGNUM Newsl. (20(3)), 2--25
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.linregm_var_del'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02dfft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, ip.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, q[ldq, ip+1].
        utils._BYREF_INT, # Input, ldq.
        utils._BYREF_INT, # Input, indx.
        utils._BYREF_FLOAT64, # Input/Output, rss.
        utils._EngineFloat64ArrayType.c_type, # Workspace, wk.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((q, 'q'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _ip = utils._EngineIntScalarType(
        ip, {'fun_name': fun_name, 'entity': 'ip'},
    ).to_ctypes()
    _ldq_py = utils._EngineFloat64ArrayType.get_nd(q, {'fun_name': fun_name, 'entity': 'q'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldq = utils._EngineIntScalarType(
        _ldq_py, {'fun_name': fun_name, 'entity': '_ldq'},
    ).to_ctypes()
    _ip_py = utils._EngineIntScalarType.to_py(_ip)
    _q_cls = utils._EngineFloat64ArrayType(
        q, {'fun_name': fun_name, 'entity': 'q'},
        exp_shape=(_ip_py, _ip_py+1),
        mutual_sorder=_m_sorder,
    )
    _q = _q_cls.to_ctypes()
    _indx = utils._EngineIntScalarType(
        indx, {'fun_name': fun_name, 'entity': 'indx'},
    ).to_ctypes()
    _rss = utils._EngineFloat64ScalarType(
        rss, {'fun_name': fun_name, 'entity': 'rss'},
    ).to_ctypes()
    _wk = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _ip,
        _q,
        _ldq,
        _indx,
        _rss,
        _wk,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, rss = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: rss >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, indx = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: indx >= 1 and indx <= ip.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ip >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, q[{:s},{:s}] = 0.0.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _q_cls.refresh_py(_q)
        _rss_py = utils._EngineFloat64ScalarType.to_py(_rss)
        return_data = _rss_py
    _err_class.manage()
    return return_data

def linregm_fit_newvar(n, wt, rss, ip, irank, cov, q, svd, p, y, b, se, res, wk):
    u"""
    Fits a general linear regression model to new dependent variable.

    ``linregm_fit_newvar`` calculates the estimates of the parameters of
    a general linear regression model for a new dependent variable after
    a call to ``linregm_fit``.

    For full information please refer to the NAG Library document for
    g02dg

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02dgf.html

    Parameters
    ----------
    n : int
        n, the number of observations.

    wt : None or float, shape (`n`)
        If provided `wt` must contain the weights to be used with the
        model.

        If `wt`[i-1] = 0.0, the ith observation is not included in the
        model, in which case the effective number of observations is the
        number of observations with nonzero weights.

        If `wt` is not provided the effective number of observations is
        n.

    rss : float
        The residual sum of squares for the original dependent variable.

    ip : int
        p, the number of independent variables (including the mean if
        fitted).

    irank : int
        The rank of the independent variables, as given by
        ``linregm_fit``.

    cov : float, shape (`ip`*(`ip`+1)/2), modified in place
        `On entry`: the covariance matrix of the parameter estimates as
        given by ``linregm_fit``.

        `On exit`: the upper triangular part of the variance-covariance
        matrix of the `ip` parameter estimates given in `b`. They are
        stored packed by column, i.e., the covariance between the
        parameter estimate given in `b`[i-1] and the parameter estimate
        given in `b`[j-1], j >= i, is stored in `cov`[(j*(j-1)/2+i)-1].

    q : float, shape (`n`, `ip`+1), modified in place
        `On entry`: the results of the QR decomposition as returned by
        ``linregm_fit``.

        `On exit`: the first column of `q` contains the new values of c,
        the remainder of `q` will be unchanged.

    svd : bool
        Indicates if a singular value decomposition was used by
        ``linregm_fit``.

        `svd` = True

            A singular value decomposition was used by ``linregm_fit``.

        `svd` = False

            A singular value decomposition was not used by
            ``linregm_fit``.

    p : float, shape (:)
        Note: the required length for this argument is determined as
        follows: if not((`svd`): `ip`; otherwise: `ip`*`ip`+2*`ip`.

        Details of the QR decomposition and SVD, if used, as returned in
        array `p` by ``linregm_fit``.

        If `svd` = False, only the first `ip` elements of `p` are used;
        these contain the zeta values for the QR decomposition (see
        ``lapackeig.dgeqrf`` for details).

        If `svd` = True, the first `ip` elements of `p` contain the zeta
        values for the QR decomposition (see ``lapackeig.dgeqrf`` for
        details) and the next `ip`*`ip`+`ip` elements of `p` contain
        details of the singular value decomposition.

    y : float, shape (`n`)
        The new dependent variable, y_new.

    b : float, shape (`ip`), modified in place
        `On exit`: the least squares estimates of the parameters of the
        regression model, beta^.

    se : float, shape (`ip`), modified in place
        `On exit`: the standard error of the estimates of the
        parameters.

    res : float, shape (`n`), modified in place
        `On exit`: the residuals for the new regression model.

    wk : float, shape (5*(`ip`-1)+`ip`*`ip`)
        If `svd` = True, `wk` must be unaltered from the previous call
        to ``linregm_fit`` or ``linregm_fit_newvar``.

        If `svd` = False, `wk` is used as workspace.

    Returns
    -------
    rss : float
        The residual sum of squares for the new dependent variable.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `irank` = *<value>*.

            Constraint: `irank` > 0.

        (`errno` 1)
            On entry, `irank` = *<value>* and `ip` = *<value>*.

            Constraint: if `svd` = False, `irank` = `ip`.

        (`errno` 1)
            On entry, `irank` = *<value>* and `ip` = *<value>*.

            Constraint: if `svd` = True, `irank` <= `ip`.

        (`errno` 1)
            On entry, weight = *<value>*.

            Constraint: weight = 'W' or 'U'.

        (`errno` 1)
            On entry, `rss` = *<value>*.

            Constraint: `rss` > 0.0.

        (`errno` 1)
            On entry, `n` = *<value>* and `ip` = *<value>*.

            Constraint: `n` >= `ip`.

        (`errno` 1)
            On entry, `ip` = *<value>*.

            Constraint: `ip` >= 1.

        (`errno` 2)
            On entry, `wt`[*<value>*] < 0.0.

            Constraint: `wt`[i-1] >= 0.0, for i = 1,2,...,n.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Hammarling, S, 1985, `The singular value decomposition in
    multivariate statistics`, SIGNUM Newsl. (20(3)), 2--25

    Searle, S R, 1971, `Linear Models`, Wiley
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.linregm_fit_newvar'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02dgft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, weight(len=1).
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, wt[n].
        utils._BYREF_FLOAT64, # Input/Output, rss.
        utils._BYREF_INT, # Input, ip.
        utils._BYREF_INT, # Input, irank.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, cov[ip*(ip+1)//2].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, q[ldq, ip+1].
        utils._BYREF_INT, # Input, ldq.
        utils._BYREF_LOGICAL, # Input, svd.
        utils._EngineFloat64ArrayType.c_type, # Input, p[*].
        utils._EngineFloat64ArrayType.c_type, # Input, y[n].
        utils._EngineFloat64ArrayType.c_type, # Output, b[ip].
        utils._EngineFloat64ArrayType.c_type, # Output, se[ip].
        utils._EngineFloat64ArrayType.c_type, # Output, res[n].
        utils._EngineFloat64ArrayType.c_type, # Input, wk[5*(ip-1)+ip*ip].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of weight.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((q, 'q'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    if wt is None:
        _weight = utils._EngineCharScalarType(
            'U', {'fun_name': fun_name, 'entity': '_weight'}, exp_chrlen=1,
        ).to_ctypes()
    else:
        _weight = utils._EngineCharScalarType(
            'W', {'fun_name': fun_name, 'entity': '_weight'}, exp_chrlen=1,
        ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _wt = utils._EngineFloat64ArrayType(
        wt, {'fun_name': fun_name, 'entity': 'wt'},
        exp_shape=(_n_py,),
        allow_none=True,
    ).to_ctypes()
    _rss = utils._EngineFloat64ScalarType(
        rss, {'fun_name': fun_name, 'entity': 'rss'},
    ).to_ctypes()
    _ip = utils._EngineIntScalarType(
        ip, {'fun_name': fun_name, 'entity': 'ip'},
    ).to_ctypes()
    _irank = utils._EngineIntScalarType(
        irank, {'fun_name': fun_name, 'entity': 'irank'},
    ).to_ctypes()
    _ip_py = utils._EngineIntScalarType.to_py(_ip)
    _cov_cls = utils._EngineFloat64ArrayType(
        cov, {'fun_name': fun_name, 'entity': 'cov'},
        exp_shape=(_ip_py*(_ip_py+1)//2,),
    )
    _cov = _cov_cls.to_ctypes()
    _ldq_py = utils._EngineFloat64ArrayType.get_nd(q, {'fun_name': fun_name, 'entity': 'q'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldq = utils._EngineIntScalarType(
        _ldq_py, {'fun_name': fun_name, 'entity': '_ldq'},
    ).to_ctypes()
    _q_cls = utils._EngineFloat64ArrayType(
        q, {'fun_name': fun_name, 'entity': 'q'},
        exp_shape=(_n_py, _ip_py+1),
        mutual_sorder=_m_sorder,
    )
    _q = _q_cls.to_ctypes()
    _svd = utils._EngineLogicalScalarType(
        svd, {'fun_name': fun_name, 'entity': 'svd'},
    ).to_ctypes()
    _svd_py = utils._EngineLogicalScalarType.to_py(_svd)
    if not _svd_py:
        _p_dim1 = _ip_py
    else:
        _p_dim1 = _ip_py*_ip_py+2*_ip_py
    exp_shape_p = (_p_dim1,)
    _p = utils._EngineFloat64ArrayType(
        p, {'fun_name': fun_name, 'entity': 'p'},
        exp_shape=exp_shape_p,
    ).to_ctypes()
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _b_cls = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_ip_py,),
        check_contents=False,
    )
    _b = _b_cls.to_ctypes(can_reference_contents=False)
    _se_cls = utils._EngineFloat64ArrayType(
        se, {'fun_name': fun_name, 'entity': 'se'},
        exp_shape=(_ip_py,),
        check_contents=False,
    )
    _se = _se_cls.to_ctypes(can_reference_contents=False)
    _res_cls = utils._EngineFloat64ArrayType(
        res, {'fun_name': fun_name, 'entity': 'res'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _res = _res_cls.to_ctypes(can_reference_contents=False)
    _wk = utils._EngineFloat64ArrayType(
        wk, {'fun_name': fun_name, 'entity': 'wk'},
        exp_shape=(5*(_ip_py-1)+_ip_py*_ip_py,),
    ).to_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _weight,
        _n,
        _wt,
        _rss,
        _ip,
        _irank,
        _cov,
        _q,
        _ldq,
        _svd,
        _p,
        _y,
        _b,
        _se,
        _res,
        _wk,
        _errbuf,
        _ierr,
        _weight._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, irank = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: irank > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, irank = {:s} and ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if svd = False, irank = ip.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, irank = {:s} and ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if svd = True, irank <= ip.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, weight = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: weight = \'W\' or \'U\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, rss = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: rss > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s} and ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: n >= ip.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ip >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, wt[{:s}] < 0.0.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: wt[i-1] >= 0.0, for i = 1,2,...,n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _rss_py = utils._EngineFloat64ScalarType.to_py(_rss)
        _cov_cls.refresh_py(_cov)
        _q_cls.refresh_py(_q)
        _b_cls.refresh_py(_b)
        _se_cls.refresh_py(_se)
        _res_cls.refresh_py(_res)
        return_data = _rss_py
    _err_class.manage()
    return return_data

def linregm_constrain(ip, iconst, p, c, b, rss, idf, se, cov):
    u"""
    Estimates and standard errors of parameters of a general linear
    regression model for given constraints.

    ``linregm_constrain`` calculates the estimates of the parameters of
    a general linear regression model for given constraints from the
    singular value decomposition results.

    For full information please refer to the NAG Library document for
    g02dk

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02dkf.html

    Parameters
    ----------
    ip : int
        p, the number of terms in the linear model.

    iconst : int
        The number of constraints to be imposed on the parameters, n_c.

    p : float, shape (`ip`*`ip`+2*`ip`)
        As returned by ``linregm_fit`` and ``linregm_update``.

    c : float, shape (`ip`, `iconst`)
        The `iconst` constraints stored by column, i.e., the ith
        constraint is stored in the ith column of `c`.

    b : float, shape (`ip`), modified in place
        `On entry`: the parameter estimates computed by using the
        singular value decomposition, beta^_svd.

        `On exit`: the parameter estimates of the parameters with the
        constraints imposed, beta^_c.

    rss : float
        The residual sum of squares as returned by ``linregm_fit`` or
        ``linregm_update``.

    idf : int
        The degrees of freedom associated with the residual sum of
        squares as returned by ``linregm_fit`` or ``linregm_update``.

    se : float, shape (`ip`), modified in place
        `On exit`: the standard error of the parameter estimates in `b`.

    cov : float, shape (`ip`*(`ip`+1)/2), modified in place
        `On exit`: the upper triangular part of the variance-covariance
        matrix of the `ip` parameter estimates given in `b`. They are
        stored packed by column, i.e., the covariance between the
        parameter estimate given in `b`[i-1] and the parameter estimate
        given in `b`[j-1], j >= i, is stored in `cov`[(j*(j-1)/2+i)-1].

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `idf` = *<value>*.

            Constraint: `idf` > 0.

        (`errno` 1)
            On entry, `iconst` = *<value>*.

            Constraint: `iconst` > 0.

        (`errno` 1)
            On entry, `rss` = *<value>*.

            Constraint: `rss` > 0.0.

        (`errno` 1)
            On entry, `iconst` = *<value>* and `ip` = *<value>*.

            Constraint: `iconst` < `ip`.

        (`errno` 1)
            On entry, `ip` = *<value>*.

            Constraint: `ip` >= 1.

        (`errno` 2)
            `c` does not give a model of full rank.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Hammarling, S, 1985, `The singular value decomposition in
    multivariate statistics`, SIGNUM Newsl. (20(3)), 2--25

    Searle, S R, 1971, `Linear Models`, Wiley
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.linregm_constrain'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02dkft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, ip.
        utils._BYREF_INT, # Input, iconst.
        utils._EngineFloat64ArrayType.c_type, # Input, p[ip*ip+2*ip].
        utils._EngineFloat64ArrayType.c_type, # Input, c[ldc, iconst].
        utils._BYREF_INT, # Input, ldc.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, b[ip].
        utils._BYREF_FLOAT64, # Input, rss.
        utils._BYREF_INT, # Input, idf.
        utils._EngineFloat64ArrayType.c_type, # Output, se[ip].
        utils._EngineFloat64ArrayType.c_type, # Output, cov[ip*(ip+1)//2].
        utils._EngineFloat64ArrayType.c_type, # Workspace, wk.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((c, 'c'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _ip = utils._EngineIntScalarType(
        ip, {'fun_name': fun_name, 'entity': 'ip'},
    ).to_ctypes()
    _iconst = utils._EngineIntScalarType(
        iconst, {'fun_name': fun_name, 'entity': 'iconst'},
    ).to_ctypes()
    _ip_py = utils._EngineIntScalarType.to_py(_ip)
    _p = utils._EngineFloat64ArrayType(
        p, {'fun_name': fun_name, 'entity': 'p'},
        exp_shape=(_ip_py*_ip_py+2*_ip_py,),
    ).to_ctypes()
    _ldc_py = utils._EngineFloat64ArrayType.get_nd(c, {'fun_name': fun_name, 'entity': 'c'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldc = utils._EngineIntScalarType(
        _ldc_py, {'fun_name': fun_name, 'entity': '_ldc'},
    ).to_ctypes()
    _iconst_py = utils._EngineIntScalarType.to_py(_iconst)
    _c = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_ip_py, _iconst_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _b_cls = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_ip_py,),
    )
    _b = _b_cls.to_ctypes()
    _rss = utils._EngineFloat64ScalarType(
        rss, {'fun_name': fun_name, 'entity': 'rss'},
    ).to_ctypes()
    _idf = utils._EngineIntScalarType(
        idf, {'fun_name': fun_name, 'entity': 'idf'},
    ).to_ctypes()
    _se_cls = utils._EngineFloat64ArrayType(
        se, {'fun_name': fun_name, 'entity': 'se'},
        exp_shape=(_ip_py,),
        check_contents=False,
    )
    _se = _se_cls.to_ctypes(can_reference_contents=False)
    _cov_cls = utils._EngineFloat64ArrayType(
        cov, {'fun_name': fun_name, 'entity': 'cov'},
        exp_shape=(_ip_py*(_ip_py+1)//2,),
        check_contents=False,
    )
    _cov = _cov_cls.to_ctypes(can_reference_contents=False)
    _wk = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _ip,
        _iconst,
        _p,
        _c,
        _ldc,
        _b,
        _rss,
        _idf,
        _se,
        _cov,
        _wk,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, idf = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: idf > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iconst = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: iconst > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, rss = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: rss > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iconst = {:s} and ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: iconst < ip.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ip >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'c does not give a model of full rank.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _b_cls.refresh_py(_b)
        _se_cls.refresh_py(_se)
        _cov_cls.refresh_py(_cov)
        return_data = None
    _err_class.manage()
    return return_data

def linregm_estfunc(ip, irank, b, cov, p, f, tol):
    u"""
    Computes estimable function of a general linear regression model and
    its standard error.

    ``linregm_estfunc`` gives the estimate of an estimable function
    along with its standard error.

    For full information please refer to the NAG Library document for
    g02dn

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02dnf.html

    Parameters
    ----------
    ip : int
        p, the number of terms in the linear model.

    irank : int
        k, the rank of the independent variables.

    b : float, shape (`ip`)
        The `ip` values of the estimates of the parameters of the model,
        beta^.

    cov : float, shape (`ip`*(`ip`+1)/2)
        The upper triangular part of the variance-covariance matrix of
        the `ip` parameter estimates given in `b`. They are stored
        packed by column, i.e., the covariance between the parameter
        estimate given in `b`[i-1] and the parameter estimate given in
        `b`[j-1], j >= i, is stored in `cov`[(j*(j-1)/2+i)-1].

    p : float, shape (`ip`*`ip`+2*`ip`)
        As returned by ``linregm_fit`` and ``linregm_update``.

    f : float, shape (`ip`)
        f, the linear function to be estimated.

    tol : float
        eta, the tolerance value used in the check for estimability.

        If `tol` <= 0.0 then sqrt(epsilon), where epsilon is the machine
        precision, is used instead.

    Returns
    -------
    est : bool
        Indicates if the function was estimable.

        `est` = True

            The function is estimable.

        `est` = False

            The function is not estimable and `stat`, `sestat` and `t`
            are not set.

    stat : float
        If `est` = True, `stat` contains the estimate of the function,
        f^T beta^.

    sestat : float
        If `est` = True, `sestat` contains the standard error of the
        estimate of the function, se(F).

    t : float
        If `est` = True, `t` contains the t-statistic for the test of
        the function being equal to zero.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `irank` = *<value>* and `ip` = *<value>*.

            Constraint: `irank` <= `ip`.

        (`errno` 1)
            On entry, `irank` = *<value>*.

            Constraint: `irank` >= 1.

        (`errno` 1)
            On entry, `ip` = *<value>*.

            Constraint: `ip` >= 1.

        (`errno` 3)
            Standard error of statistic = 0.0; this may be due to
            rounding errors if the standard error is very small or due
            to mis-specified inputs `cov` and `f`.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            On entry, `irank` = `ip`, i.e., model of full rank. In this
            case `est` is returned as true and all statistics are
            calculated.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Hammarling, S, 1985, `The singular value decomposition in
    multivariate statistics`, SIGNUM Newsl. (20(3)), 2--25

    Searle, S R, 1971, `Linear Models`, Wiley
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.linregm_estfunc'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02dnft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, ip.
        utils._BYREF_INT, # Input, irank.
        utils._EngineFloat64ArrayType.c_type, # Input, b[ip].
        utils._EngineFloat64ArrayType.c_type, # Input, cov[ip*(ip+1)//2].
        utils._EngineFloat64ArrayType.c_type, # Input, p[ip*ip+2*ip].
        utils._EngineFloat64ArrayType.c_type, # Input, f[ip].
        utils._BYREF_LOGICAL, # Output, est.
        utils._BYREF_FLOAT64, # Output, stat.
        utils._BYREF_FLOAT64, # Output, sestat.
        utils._BYREF_FLOAT64, # Output, t.
        utils._BYREF_FLOAT64, # Input, tol.
        utils._EngineFloat64ArrayType.c_type, # Workspace, wk.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _ip = utils._EngineIntScalarType(
        ip, {'fun_name': fun_name, 'entity': 'ip'},
    ).to_ctypes()
    _irank = utils._EngineIntScalarType(
        irank, {'fun_name': fun_name, 'entity': 'irank'},
    ).to_ctypes()
    _ip_py = utils._EngineIntScalarType.to_py(_ip)
    _b = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_ip_py,),
    ).to_ctypes()
    _cov = utils._EngineFloat64ArrayType(
        cov, {'fun_name': fun_name, 'entity': 'cov'},
        exp_shape=(_ip_py*(_ip_py+1)//2,),
    ).to_ctypes()
    _p = utils._EngineFloat64ArrayType(
        p, {'fun_name': fun_name, 'entity': 'p'},
        exp_shape=(_ip_py*_ip_py+2*_ip_py,),
    ).to_ctypes()
    _f = utils._EngineFloat64ArrayType(
        f, {'fun_name': fun_name, 'entity': 'f'},
        exp_shape=(_ip_py,),
    ).to_ctypes()
    _est = utils._EngineLogicalScalarType.empty_ctypes()
    _stat = utils._EngineFloat64ScalarType.empty_ctypes()
    _sestat = utils._EngineFloat64ScalarType.empty_ctypes()
    _t = utils._EngineFloat64ScalarType.empty_ctypes()
    _tol = utils._EngineFloat64ScalarType(
        tol, {'fun_name': fun_name, 'entity': 'tol'},
    ).to_ctypes()
    _wk = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _ip,
        _irank,
        _b,
        _cov,
        _p,
        _f,
        _est,
        _stat,
        _sestat,
        _t,
        _tol,
        _wk,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, irank = {:s} and ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: irank <= ip.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, irank = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: irank >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ip >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, irank = ip,',
                    'i.e., model of full rank. In this case est is returned as true and all',
                    'statistics are calculated.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Standard error of statistic = 0.0;',
                    'this may be due to rounding errors if the standard error is very small',
                    'or due to mis-specified inputs cov and f.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _est_py = utils._EngineLogicalScalarType.to_py(_est)
        _stat_py = utils._EngineFloat64ScalarType.to_py(_stat)
        _sestat_py = utils._EngineFloat64ScalarType.to_py(_sestat)
        _t_py = utils._EngineFloat64ScalarType.to_py(_t)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['est', 'stat', 'sestat', 't'],
        )(
            _est_py,
            _stat_py,
            _sestat_py,
            _t_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def linregm_rssq(mean, n, m, x, vname, isx, y, wt, enmod, modl, rss, nterms, mrank):
    u"""
    Computes residual sums of squares for all possible linear
    regressions for a set of independent variables.

    ``linregm_rssq`` calculates the residual sums of squares for all
    possible linear regressions for a given set of independent
    variables.

    For full information please refer to the NAG Library document for
    g02ea

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02eaf.html

    Parameters
    ----------
    mean : str, length 1
        Indicates if a mean term is to be included.

        `mean` = 'M'

            A mean term, intercept, will be included in the model.

        `mean` = 'Z'

            The model will pass through the origin, zero-point.

    n : int
        n, the number of observations.

    m : int
        The number of variables contained in `x`.

    x : float, shape (`n`, `m`)
        `x`[i-1,j-1] must contain the ith observation for the jth
        independent variable, for j = 1,2,...,`m`, for i = 1,2,...,`n`.

    vname : str, shape (`m`)
        `vname`[j-1] must contain the name of the variable in column j
        of `x`, for j = 1,2,...,`m`.

    isx : int, shape (`m`)
        Indicates which independent variables are to be considered in
        the model.

        `isx`[j-1] >= 2

            The variable contained in the jth column of `x` is included
            in all regression models, i.e., is a forced variable.

        `isx`[j-1] = 1

            The variable contained in the jth column of `x` is included
            in the set from which the regression models are chosen,
            i.e., is a free variable.

        `isx`[j-1] = 0

            The variable contained in the jth column of `x` is not
            included in the models.

    y : float, shape (`n`)
        `y`[i-1] must contain the ith observation on the dependent
        variable, y_i, for i = 1,2,...,n.

    wt : None or float, shape (`n`)
        If provided `wt` must contain the weights to be used with the
        model.

        If `wt`[i-1] = 0.0, the ith observation is not included in the
        model, in which case the effective number of observations is the
        number of observations with nonzero weights.

        If `wt` is not provided the effective number of observations is
        n.

    enmod : int
        The dimension of the arrays `rss`, `nterms` and `mrank`.

    modl : str, shape (`enmod`, `m`), modified in place
        `On exit`: the first `nterms`[i-1] elements of the ith row of
        `modl` contain the names of the independent variables, as given
        in `vname`, that are included in the ith model.

    rss : float, shape (`enmod`), modified in place
        `On exit`: `rss`[i-1] contains the residual sum of squares for
        the ith model, for i = 1,2,...,`nmod`.

    nterms : int, shape (`enmod`), modified in place
        `On exit`: `nterms`[i-1] contains the number of independent
        variables in the ith model, not including the mean if one is
        fitted, for i = 1,2,...,`nmod`.

    mrank : int, shape (`enmod`), modified in place
        `On exit`: `mrank`[i-1] contains the rank of the residual sum of
        squares for the ith model.

    Returns
    -------
    nmod : int
        The total number of models for which residual sums of squares
        have been calculated.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `enmod` = *<value>* and `m` = *<value>*.

            Constraint: `enmod` >= `m`.

        (`errno` 1)
            On entry, weight = *<value>*.

            Constraint: weight = 'W' or 'U'.

        (`errno` 1)
            On entry, `mean` = *<value>*.

            Constraint: `mean` = 'M' or 'Z'.

        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 2.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 2.

        (`errno` 2)
            On entry, `wt`[*<value>*] < 0.0.

            Constraint: `wt`[i-1] >= 0.0, for i = 1,2,...,n.

        (`errno` 3)
            On entry, `isx`[*<value>*] < 0.

            Constraint: `isx`[i-1] >= 0, for i = 1,2,...,`m`.

        (`errno` 3)
            There are no free variables, i.e., no element of `isx` = 1.

        (`errno` 4)
            On entry, `enmod` = *<value>* and number of possible models
            is *<value>*.

            Constraint: `enmod` >= the number of possible models.

        (`errno` 5)
            On entry, the number of independent variables to be
            considered (forced plus free plus mean if included) is
            greater or equal to the effective number of observations.

        (`errno` 6)
            The full model is not of full rank, i.e., some of the
            independent variables may be linear combinations of other
            independent variables. Variables must be excluded from the
            model in order to give full rank.

    References
    ----------
    Clark, M R B, 1981, `A Givens algorithm for moving from one linear
    model to another without going back to the data`, Appl. Statist.
    (30), 198--203

    Smith, D M and Bremner, J M, 1989, `All possible subset regressions
    using the QR decomposition`, Comput. Statist. Data Anal. (7),
    217--236

    Weisberg, S, 1985, `Applied Linear Regression`, Wiley
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.linregm_rssq'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02eaft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, mean(len=1).
        utils._BYREF_STRING, # Input, weight(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Input, x[ldx, m].
        utils._BYREF_INT, # Input, ldx.
        utils._EngineCharArrayType.c_type, # Input, vname(len=*)[m].
        utils._EngineIntArrayType.c_type, # Input, isx[m].
        utils._EngineFloat64ArrayType.c_type, # Input, y[n].
        utils._EngineFloat64ArrayType.c_type, # Input, wt[n].
        utils._BYREF_INT, # Input, enmod.
        utils._BYREF_INT, # Output, nmod.
        utils._EngineCharArrayType.c_type, # Output, modl(len=*)[ldmodl, m].
        utils._BYREF_INT, # Input, ldmodl.
        utils._EngineFloat64ArrayType.c_type, # Output, rss[enmod].
        utils._EngineIntArrayType.c_type, # Output, nterms[enmod].
        utils._EngineIntArrayType.c_type, # Output, mrank[enmod].
        utils._EngineFloat64ArrayType.c_type, # Workspace, wk.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of mean.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of weight.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of vname.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of modl.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((x, 'x'), (modl, 'modl')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _mean = utils._EngineCharScalarType(
        mean, {'fun_name': fun_name, 'entity': 'mean'}, exp_chrlen=1,
    ).to_ctypes()
    if wt is None:
        _weight = utils._EngineCharScalarType(
            'U', {'fun_name': fun_name, 'entity': '_weight'}, exp_chrlen=1,
        ).to_ctypes()
    else:
        _weight = utils._EngineCharScalarType(
            'W', {'fun_name': fun_name, 'entity': '_weight'}, exp_chrlen=1,
        ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _ldx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py, _m_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _vname_cls = utils._EngineCharArrayType(
        vname, {'fun_name': fun_name, 'entity': 'vname'}, exp_chrlen=None,
        exp_shape=(_m_py,),
    )
    _vname = _vname_cls.to_ctypes()
    _vname_chrlen = _vname_cls.chrlen
    _isx = utils._EngineIntArrayType(
        isx, {'fun_name': fun_name, 'entity': 'isx'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _wt = utils._EngineFloat64ArrayType(
        wt, {'fun_name': fun_name, 'entity': 'wt'},
        exp_shape=(_n_py,),
        allow_none=True,
    ).to_ctypes()
    _enmod = utils._EngineIntScalarType(
        enmod, {'fun_name': fun_name, 'entity': 'enmod'},
    ).to_ctypes()
    _nmod = utils._EngineIntScalarType.empty_ctypes()
    _ldmodl_py = utils._EngineCharArrayType.get_nd(modl, {'fun_name': fun_name, 'entity': 'modl'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldmodl = utils._EngineIntScalarType(
        _ldmodl_py, {'fun_name': fun_name, 'entity': '_ldmodl'},
    ).to_ctypes()
    _enmod_py = utils._EngineIntScalarType.to_py(_enmod)
    _modl_cls = utils._EngineCharArrayType(
        modl, {'fun_name': fun_name, 'entity': 'modl'}, exp_chrlen=None,
        exp_shape=(_enmod_py, _m_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _modl = _modl_cls.to_ctypes(can_reference_contents=False)
    _modl_chrlen = _modl_cls.chrlen
    _vl_k_py = 0
    for _vl_i_py in range(1, _m_py+1):
        if utils._EngineIntArrayType.get_element(isx, {'fun_name': fun_name, 'entity': 'isx'}, (_vl_i_py-1,)) == 1:
            _vl_k_py = _vl_k_py + 1
    _rss_cls = utils._EngineFloat64ArrayType(
        rss, {'fun_name': fun_name, 'entity': 'rss'},
        exp_shape=(max(2**_vl_k_py, _m_py),),
        check_contents=False,
    )
    _rss = _rss_cls.to_ctypes(can_reference_contents=False)
    _nterms_cls = utils._EngineIntArrayType(
        nterms, {'fun_name': fun_name, 'entity': 'nterms'},
        exp_shape=(max(2**_vl_k_py, _m_py),),
        check_contents=False,
    )
    _nterms = _nterms_cls.to_ctypes(can_reference_contents=False)
    _mrank_cls = utils._EngineIntArrayType(
        mrank, {'fun_name': fun_name, 'entity': 'mrank'},
        exp_shape=(max(2**_vl_k_py, _m_py),),
        check_contents=False,
    )
    _mrank = _mrank_cls.to_ctypes(can_reference_contents=False)
    _wk = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _mean,
        _weight,
        _n,
        _m,
        _x,
        _ldx,
        _vname,
        _isx,
        _y,
        _wt,
        _enmod,
        _nmod,
        _modl,
        _ldmodl,
        _rss,
        _nterms,
        _mrank,
        _wk,
        _errbuf,
        _ierr,
        _mean._length_,
        _weight._length_,
        _vname_chrlen,
        _modl_chrlen,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, enmod = {:s} and m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: enmod >= m.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, weight = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: weight = \'W\' or \'U\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mean = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mean = \'M\' or \'Z\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, wt[{:s}] < 0.0.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: wt[i-1] >= 0.0, for i = 1,2,...,n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99988:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, isx[{:s}] < 0.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: isx[i-1] >= 0, for i = 1,2,...,m.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'There are no free variables, i.e., no element of isx = 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, enmod = {:s} and number of possible models is'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    '{:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: enmod >= the number of possible models.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the number of independent variables to be considered (forced plus',
                    'free plus mean if included) is greater or equal to the effective number of',
                    'observations.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The full model is not of full rank, i.e., some of the independent variables',
                    'may be linear combinations of other independent variables. Variables must',
                    'be excluded from the model in order to give full rank.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _modl_cls.refresh_py(_modl)
        _rss_cls.refresh_py(_rss)
        _nterms_cls.refresh_py(_nterms)
        _mrank_cls.refresh_py(_mrank)
        _nmod_py = utils._EngineIntScalarType.to_py(_nmod)
        return_data = _nmod_py
    _err_class.manage()
    return return_data

def linregm_rssq_stat(mean, n, sigsq, tss, nmod, nterms, rss, rsq, cp):
    u"""
    Calculates R^2 and C_P values from residual sums of squares.

    ``linregm_rssq_stat`` calculates R^2 and C_p-values from the
    residual sums of squares for a series of linear regression models.

    For full information please refer to the NAG Library document for
    g02ec

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02ecf.html

    Parameters
    ----------
    mean : str, length 1
        Indicates if a mean term is to be included.

        `mean` = 'M'

            A mean term, intercept, will be included in the model.

        `mean` = 'Z'

            The model will pass through the origin, zero-point.

    n : int
        n, the number of observations used in the regression model.

    sigsq : float
        The best estimate of true variance of the errors, sigma^^2.

    tss : float
        The total sum of squares for the regression model.

    nmod : int
        The number of regression models.

    nterms : int, shape (`nmod`)
        `nterms`[i-1] must contain the number of independent variables
        (not counting the mean) fitted to the ith model, for i =
        1,2,...,`nmod`.

    rss : float, shape (`nmod`)
        `rss`[i-1] must contain the residual sum of squares for the ith
        model.

    rsq : float, shape (`nmod`), modified in place
        `On exit`: `rsq`[i-1] contains the R^2-value for the ith model,
        for i = 1,2,...,`nmod`.

    cp : float, shape (`nmod`), modified in place
        `On exit`: `cp`[i-1] contains the C_p-value for the ith model,
        for i = 1,2,...,`nmod`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `mean` = *<value>*.

            Constraint: `mean` = 'M' or 'Z'.

        (`errno` 1)
            On entry, `tss` = *<value>*.

            Constraint: `tss` > 0.0.

        (`errno` 1)
            On entry, `sigsq` = *<value>*.

            Constraint: `sigsq` > 0.0.

        (`errno` 1)
            On entry, `nmod` = *<value>*.

            Constraint: `nmod` > 0.

        (`errno` 2)
            On entry: the number of parameters, p, is *<value>* and `n`
            = *<value>*.

            Constraint: `n` >= 2*p.

        (`errno` 3)
            On entry, `rss`[*<value>*] = *<value>* and `tss` =
            *<value>*.

            Constraint: `rss`[i] <= `tss`, for all i.

        (`errno` 4)
            A value of C_p is less than 0.0. This may occur if `sigsq`
            is too large or if `rss`, `n` or IP are incorrect.

    References
    ----------
    Draper, N R and Smith, H, 1985, `Applied Regression Analysis`, (2nd
    Edition), Wiley

    Weisberg, S, 1985, `Applied Linear Regression`, Wiley
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.linregm_rssq_stat'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02ecft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, mean(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_FLOAT64, # Input, sigsq.
        utils._BYREF_FLOAT64, # Input, tss.
        utils._BYREF_INT, # Input, nmod.
        utils._EngineIntArrayType.c_type, # Input, nterms[nmod].
        utils._EngineFloat64ArrayType.c_type, # Input, rss[nmod].
        utils._EngineFloat64ArrayType.c_type, # Output, rsq[nmod].
        utils._EngineFloat64ArrayType.c_type, # Output, cp[nmod].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of mean.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _mean = utils._EngineCharScalarType(
        mean, {'fun_name': fun_name, 'entity': 'mean'}, exp_chrlen=1,
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _sigsq = utils._EngineFloat64ScalarType(
        sigsq, {'fun_name': fun_name, 'entity': 'sigsq'},
    ).to_ctypes()
    _tss = utils._EngineFloat64ScalarType(
        tss, {'fun_name': fun_name, 'entity': 'tss'},
    ).to_ctypes()
    _nmod = utils._EngineIntScalarType(
        nmod, {'fun_name': fun_name, 'entity': 'nmod'},
    ).to_ctypes()
    _nmod_py = utils._EngineIntScalarType.to_py(_nmod)
    _nterms = utils._EngineIntArrayType(
        nterms, {'fun_name': fun_name, 'entity': 'nterms'},
        exp_shape=(_nmod_py,),
    ).to_ctypes()
    _rss = utils._EngineFloat64ArrayType(
        rss, {'fun_name': fun_name, 'entity': 'rss'},
        exp_shape=(_nmod_py,),
    ).to_ctypes()
    _rsq_cls = utils._EngineFloat64ArrayType(
        rsq, {'fun_name': fun_name, 'entity': 'rsq'},
        exp_shape=(_nmod_py,),
        check_contents=False,
    )
    _rsq = _rsq_cls.to_ctypes(can_reference_contents=False)
    _cp_cls = utils._EngineFloat64ArrayType(
        cp, {'fun_name': fun_name, 'entity': 'cp'},
        exp_shape=(_nmod_py,),
        check_contents=False,
    )
    _cp = _cp_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _mean,
        _n,
        _sigsq,
        _tss,
        _nmod,
        _nterms,
        _rss,
        _rsq,
        _cp,
        _errbuf,
        _ierr,
        _mean._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mean = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mean = \'M\' or \'Z\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, tss = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: tss > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, sigsq = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: sigsq > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nmod = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nmod > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry: the number of parameters, p, is {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: n >= 2*p.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, rss[{:s}] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'tss = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: rss[i] <= tss, for all i.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'A value of C_p is less than 0.0.',
                    'This may occur if sigsq is too large or if rss, n or IP are incorrect.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _rsq_cls.refresh_py(_rsq)
        _cp_cls.refresh_py(_cp)
        return_data = None
    _err_class.manage()
    return return_data

def linregm_fit_onestep(istep, mean, n, m, x, vname, isx, maxip, y, wt, fin, model, nterm, rss, idf, ifr, free, exss, q, p):
    u"""
    Fits a linear regression model by forward selection.

    ``linregm_fit_onestep`` carries out one step of a forward selection
    procedure in order to enable the 'best' linear regression model to
    be found.

    For full information please refer to the NAG Library document for
    g02ee

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02eef.html

    Parameters
    ----------
    istep : int
        Indicates which step in the forward selection process is to be
        carried out.

        `istep` = 0

            The process is initialized.

    mean : str, length 1
        Indicates if a mean term is to be included.

        `mean` = 'M'

            A mean term, intercept, will be included in the model.

        `mean` = 'Z'

            The model will pass through the origin, zero-point.

    n : int
        n, the number of observations.

    m : int
        m, the total number of independent variables in the dataset.

    x : float, shape (`n`, `m`)
        `x`[i-1,j-1] must contain the ith observation for the jth
        independent variable, for j = 1,2,...,`m`, for i = 1,2,...,`n`.

    vname : str, shape (`m`)
        `vname`[j-1] must contain the name of the independent variable
        in column j of `x`, for j = 1,2,...,`m`.

    isx : int, shape (`m`)
        Indicates which independent variables could be considered for
        inclusion in the regression.

        `isx`[j-1] >= 2

            The variable contained in the jth column of `x` is
            automatically included in the regression model, for j =
            1,2,...,`m`.

        `isx`[j-1] = 1

            The variable contained in the jth column of `x` is
            considered for inclusion in the regression model, for j =
            1,2,...,`m`.

        `isx`[j-1] = 0

            The variable in the jth column is not considered for
            inclusion in the model, for j = 1,2,...,`m`.

    maxip : int
        The maximum number of independent variables to be included in
        the model.

    y : float, shape (`n`)
        The dependent variable.

    wt : None or float, shape (`n`)
        If provided `wt` must contain the weights to be used with the
        model.

        If `wt`[i-1] = 0.0, the ith observation is not included in the
        model, in which case the effective number of observations is the
        number of observations with nonzero weights.

        If `wt` is not provided the effective number of observations is
        n.

    fin : float
        The critical value of the F statistic for the term to be
        included in the model, F_c.

        Suggested value: 2.0 is a commonly used value in exploratory
        modelling.

    model : str, shape (`maxip`), modified in place
        `On entry`: if `istep` = 0, `model` need not be set.

        If `istep` != 0, `model` must contain the values returned by the
        previous call to ``linregm_fit_onestep``.

        `On exit`: the names of the variables in the current model.

    nterm : int
        If `istep` = 0, `nterm` need not be set.

        If `istep` != 0, `nterm` must contain the value returned by the
        previous call to ``linregm_fit_onestep``.

    rss : float
        If `istep` = 0, `rss` need not be set.

        If `istep` != 0, `rss` must contain the value returned by the
        previous call to ``linregm_fit_onestep``.

    idf : int
        If `istep` = 0, `idf` need not be set.

        If `istep` != 0, `idf` must contain the value returned by the
        previous call to ``linregm_fit_onestep``.

    ifr : int
        If `istep` = 0, `ifr` need not be set.

        If `istep` != 0, `ifr` must contain the value returned by the
        previous call to ``linregm_fit_onestep``.

    free : str, shape (`maxip`), modified in place
        `On entry`: if `istep` = 0, `free` need not be set.

        If `istep` != 0, `free` must contain the values returned by the
        previous call to ``linregm_fit_onestep``.

        `On exit`: the first `ifr` values of `free` contain the names of
        the free variables.

    exss : float, shape (`maxip`), modified in place
        `On exit`: the first `ifr` values of `exss` contain what would
        be the change in regression sum of squares if the free variables
        had been added to the model, i.e., the extra sum of squares for
        the free variables. `exss`[i-1] contains what would be the
        change in regression sum of squares if the variable `free`[i-1]
        had been added to the model.

    q : float, shape (`n`, `maxip`+2), modified in place
        `On entry`: if `istep` = 0, `q` need not be set.

        If `istep` != 0, `q` must contain the values returned by the
        previous call to ``linregm_fit_onestep``.

        `On exit`: the results of the QR decomposition for the current
        model:

            the first column of `q` contains c = Q^Ty (or Q^TW^(1/2)y
            where W is the vector of weights if used);

            the upper triangular part of columns 2 to p+1 contain the R
            matrix;

            the strictly lower triangular part of columns 2 to p+1
            contain details of the Q matrix;

            the remaining p+1 to p+`ifr` columns of contain Q^TX_free
            (or Q^TW^(1/2)X_free),

        where p = `nterm`, or p = `nterm`+1 if `mean` = 'M'.

    p : float, shape (`maxip`+1), modified in place
        `On entry`: if `istep` = 0, `p` need not be set.

        If `istep` != 0, `p` must contain the values returned by the
        previous call to ``linregm_fit_onestep``.

        `On exit`: the first p elements of `p` contain details of the QR
        decomposition, where p = `nterm`, or p = `nterm`+1 if `mean` =
        'M'.

    Returns
    -------
    istep : int
        Is incremented by 1.

    addvar : bool
        Indicates if a variable has been added to the model.

        `addvar` = True

            A variable has been added to the model.

        `addvar` = False

            No variable had an F value greater than F_c and none were
            added to the model.

    newvar : str
        If `addvar` = True, `newvar` contains the name of the variable
        added to the model.

    chrss : float
        If `addvar` = True, `chrss` contains the change in the residual
        sum of squares due to adding variable `newvar`.

    f : float
        If `addvar` = True, `f` contains the F statistic for the
        inclusion of the variable in `newvar`.

    nterm : int
        The number of independent variables in the current model, not
        including the mean, if any.

    rss : float
        The residual sums of squares for the current model.

    idf : int
        The degrees of freedom for the residual sum of squares for the
        current model.

    ifr : int
        The number of free independent variables, i.e., the number of
        variables not in the model that are still being considered for
        selection.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `fin` = *<value>*.

            Constraint: `fin` >= 0.0.

        (`errno` 1)
            On entry, `rss` = *<value>*.

            Constraint: `rss` > 0.0.

        (`errno` 1)
            On entry, `istep` = *<value>* and `nterm` = *<value>*.

            Constraint: if `istep` != 0, `nterm` > 0.

        (`errno` 1)
            On entry, `istep` = *<value>*.

            Constraint: `istep` >= 0.

        (`errno` 1)
            On entry, weight = *<value>*.

            Constraint: weight = 'W' or 'U'.

        (`errno` 1)
            On entry, `mean` = *<value>*.

            Constraint: `mean` = 'M' or 'Z'.

        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 2.

        (`errno` 2)
            On entry, `wt`[*<value>*] < 0.0.

            Constraint: `wt`[i-1] >= 0.0, for i = 1,2,...,n.

        (`errno` 3)
            On entry, number of forced variables >= `n`.

        (`errno` 3)
            Degrees of freedom for error will equal 0 if new variable is
            added, i.e., the number of variables in the model plus 1 is
            equal to the effective number of observations.

        (`errno` 4)
            On entry, `maxip` = *<value>*.

            Constraint: `maxip` must be large enough to accommodate the
            number of terms given by `isx`.

        (`errno` 4)
            On entry, `isx`[*<value>*] < 0.

            Constraint: `isx`[i-1] >= 0, for i = 1,2,...,`m`.

        (`errno` 4)
            On entry, `isx`[i-1] = 0, for all i = 1,2,...,`m`.

            Constraint: at least one value of `isx` must be nonzero.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 5)
            On entry, the variables forced into the model are not of
            full rank, i.e., some of these variables are linear
            combinations of others.

        (`errno` 6)
            There are no free variables, i.e., no element of `isx` = 0.

        (`errno` 7)
            The value of the change in the sum of squares is greater
            than the input value of `rss`. This may occur due to
            rounding errors if the true residual sum of squares for the
            new model is small relative to the residual sum of squares
            for the previous model.

    References
    ----------
    Draper, N R and Smith, H, 1985, `Applied Regression Analysis`, (2nd
    Edition), Wiley

    Weisberg, S, 1985, `Applied Linear Regression`, Wiley
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.linregm_fit_onestep'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02eeft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input/Output, istep.
        utils._BYREF_STRING, # Input, mean(len=1).
        utils._BYREF_STRING, # Input, weight(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Input, x[ldx, m].
        utils._BYREF_INT, # Input, ldx.
        utils._EngineCharArrayType.c_type, # Input, vname(len=*)[m].
        utils._EngineIntArrayType.c_type, # Input, isx[m].
        utils._BYREF_INT, # Input, maxip.
        utils._EngineFloat64ArrayType.c_type, # Input, y[n].
        utils._EngineFloat64ArrayType.c_type, # Input, wt[n].
        utils._BYREF_FLOAT64, # Input, fin.
        utils._BYREF_LOGICAL, # Output, addvar.
        utils._BYREF_STRING, # Output, newvar(len=*).
        utils._BYREF_FLOAT64, # Output, chrss.
        utils._BYREF_FLOAT64, # Output, f.
        utils._EngineCharArrayType.c_type, # Input/Output, model(len=*)[maxip].
        utils._BYREF_INT, # Input/Output, nterm.
        utils._BYREF_FLOAT64, # Input/Output, rss.
        utils._BYREF_INT, # Input/Output, idf.
        utils._BYREF_INT, # Input/Output, ifr.
        utils._EngineCharArrayType.c_type, # Input/Output, free(len=*)[maxip].
        utils._EngineFloat64ArrayType.c_type, # Output, exss[maxip].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, q[ldq, maxip+2].
        utils._BYREF_INT, # Input, ldq.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, p[maxip+1].
        utils._EngineFloat64ArrayType.c_type, # Workspace, wk.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of mean.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of weight.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of vname.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of newvar.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of model.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of free.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((x, 'x'), (q, 'q')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _istep = utils._EngineIntScalarType(
        istep, {'fun_name': fun_name, 'entity': 'istep'},
    ).to_ctypes()
    _mean = utils._EngineCharScalarType(
        mean, {'fun_name': fun_name, 'entity': 'mean'}, exp_chrlen=1,
    ).to_ctypes()
    if wt is None:
        _weight = utils._EngineCharScalarType(
            'U', {'fun_name': fun_name, 'entity': '_weight'}, exp_chrlen=1,
        ).to_ctypes()
    else:
        _weight = utils._EngineCharScalarType(
            'W', {'fun_name': fun_name, 'entity': '_weight'}, exp_chrlen=1,
        ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _ldx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py, _m_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _vname_cls = utils._EngineCharArrayType(
        vname, {'fun_name': fun_name, 'entity': 'vname'}, exp_chrlen=None,
        exp_shape=(_m_py,),
    )
    _vname = _vname_cls.to_ctypes()
    _vname_chrlen = _vname_cls.chrlen
    _isx = utils._EngineIntArrayType(
        isx, {'fun_name': fun_name, 'entity': 'isx'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _maxip = utils._EngineIntScalarType(
        maxip, {'fun_name': fun_name, 'entity': 'maxip'},
    ).to_ctypes()
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _wt = utils._EngineFloat64ArrayType(
        wt, {'fun_name': fun_name, 'entity': 'wt'},
        exp_shape=(_n_py,),
        allow_none=True,
    ).to_ctypes()
    _fin = utils._EngineFloat64ScalarType(
        fin, {'fun_name': fun_name, 'entity': 'fin'},
    ).to_ctypes()
    _addvar = utils._EngineLogicalScalarType.empty_ctypes()
    _vl_max_vname_length_py = _vname_chrlen
    _newvar = utils._EngineCharScalarType.empty_ctypes(chrlen=_vl_max_vname_length_py)
    _chrss = utils._EngineFloat64ScalarType.empty_ctypes()
    _f = utils._EngineFloat64ScalarType.empty_ctypes()
    _maxip_py = utils._EngineIntScalarType.to_py(_maxip)
    _model_cls = utils._EngineCharArrayType(
        model, {'fun_name': fun_name, 'entity': 'model'}, exp_chrlen=None,
        exp_shape=(_maxip_py,),
    )
    _model = _model_cls.to_ctypes()
    _model_chrlen = _model_cls.chrlen
    _nterm = utils._EngineIntScalarType(
        nterm, {'fun_name': fun_name, 'entity': 'nterm'},
    ).to_ctypes()
    _rss = utils._EngineFloat64ScalarType(
        rss, {'fun_name': fun_name, 'entity': 'rss'},
    ).to_ctypes()
    _idf = utils._EngineIntScalarType(
        idf, {'fun_name': fun_name, 'entity': 'idf'},
    ).to_ctypes()
    _ifr = utils._EngineIntScalarType(
        ifr, {'fun_name': fun_name, 'entity': 'ifr'},
    ).to_ctypes()
    _free_cls = utils._EngineCharArrayType(
        free, {'fun_name': fun_name, 'entity': 'free'}, exp_chrlen=None,
        exp_shape=(_maxip_py,),
    )
    _free = _free_cls.to_ctypes()
    _free_chrlen = _free_cls.chrlen
    _exss_cls = utils._EngineFloat64ArrayType(
        exss, {'fun_name': fun_name, 'entity': 'exss'},
        exp_shape=(_maxip_py,),
        check_contents=False,
    )
    _exss = _exss_cls.to_ctypes(can_reference_contents=False)
    _ldq_py = utils._EngineFloat64ArrayType.get_nd(q, {'fun_name': fun_name, 'entity': 'q'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldq = utils._EngineIntScalarType(
        _ldq_py, {'fun_name': fun_name, 'entity': '_ldq'},
    ).to_ctypes()
    _q_cls = utils._EngineFloat64ArrayType(
        q, {'fun_name': fun_name, 'entity': 'q'},
        exp_shape=(_n_py, _maxip_py+2),
        mutual_sorder=_m_sorder,
    )
    _q = _q_cls.to_ctypes()
    _p_cls = utils._EngineFloat64ArrayType(
        p, {'fun_name': fun_name, 'entity': 'p'},
        exp_shape=(_maxip_py+1,),
    )
    _p = _p_cls.to_ctypes()
    _wk = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _istep,
        _mean,
        _weight,
        _n,
        _m,
        _x,
        _ldx,
        _vname,
        _isx,
        _maxip,
        _y,
        _wt,
        _fin,
        _addvar,
        _newvar,
        _chrss,
        _f,
        _model,
        _nterm,
        _rss,
        _idf,
        _ifr,
        _free,
        _exss,
        _q,
        _ldq,
        _p,
        _wk,
        _errbuf,
        _ierr,
        _mean._length_,
        _weight._length_,
        _vname_chrlen,
        _newvar._length_,
        _model_chrlen,
        _free_chrlen,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99983:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, fin = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: fin >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99985:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, rss = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: rss > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, istep = {:s} and nterm = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if istep != 0, nterm > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, istep = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: istep >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, weight = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: weight = \'W\' or \'U\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mean = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mean = \'M\' or \'Z\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, wt[{:s}] < 0.0.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: wt[i-1] >= 0.0, for i = 1,2,...,n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99984:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, number of forced variables >= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Degrees of freedom for error will equal 0 if new',
                    'variable is added, i.e., the number of',
                    'variables in the model plus 1 is equal to the',
                    'effective number of observations.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99981:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, maxip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: maxip must be large enough to accommodate the',
                    'number of terms given by isx.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99982:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, isx[{:s}] < 0.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: isx[i-1] >= 0, for i = 1,2,...,m.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, isx[i-1] = 0, for all i = 1,2,...,m.',
                    'Constraint: at least one value of isx must be nonzero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99989:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the variables forced into the model are not of full rank,',
                    'i.e., some of these variables are linear combinations of others.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99987:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'There are no free variables, i.e., no element of isx = 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 99986:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The value of the change in the sum of squares is greater than the',
                    'input value of rss. This may occur due to rounding errors',
                    'if the true residual sum of squares for the new model is small relative',
                    'to the residual sum of squares for the previous model.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _istep_py = utils._EngineIntScalarType.to_py(_istep)
        _model_cls.refresh_py(_model)
        _nterm_py = utils._EngineIntScalarType.to_py(_nterm)
        _rss_py = utils._EngineFloat64ScalarType.to_py(_rss)
        _idf_py = utils._EngineIntScalarType.to_py(_idf)
        _ifr_py = utils._EngineIntScalarType.to_py(_ifr)
        _free_cls.refresh_py(_free)
        _exss_cls.refresh_py(_exss)
        _q_cls.refresh_py(_q)
        _p_cls.refresh_py(_p)
        _addvar_py = utils._EngineLogicalScalarType.to_py(_addvar)
        _newvar_py = utils._EngineCharScalarType.to_py(_newvar)
        _chrss_py = utils._EngineFloat64ScalarType.to_py(_chrss)
        _f_py = utils._EngineFloat64ScalarType.to_py(_f)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['istep', 'addvar', 'newvar', 'chrss', 'f', 'nterm', 'rss', 'idf', 'ifr'],
        )(
            _istep_py,
            _addvar_py,
            _newvar_py,
            _chrss_py,
            _f_py,
            _nterm_py,
            _rss_py,
            _idf_py,
            _ifr_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def linregm_fit_stepwise(m, n, wmean, c, sw, isx, fin, fout, tau, b, se, monlev, monfun, data=None, io_manager=None):
    u"""
    Stepwise linear regression.

    ``linregm_fit_stepwise`` calculates a full stepwise selection from p
    variables by using Clarke's sweep algorithm on the correlation
    matrix of a design and data matrix, Z.
    The (weighted) variance-covariance, (weighted) means and sum of
    weights of Z must be supplied.

    For full information please refer to the NAG Library document for
    g02ef

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02eff.html

    Parameters
    ----------
    m : int
        The number of explanatory variables available in the design
        matrix, Z.

    n : int
        The number of observations used in the calculations.

    wmean : float, shape (`m`+1)
        The mean of the design matrix, Z.

    c : float, shape ((`m`+1)*(`m`+2)/2)
        The upper-triangular variance-covariance matrix packed by column
        for the design matrix, Z. Because the function computes the
        correlation matrix R from `c`, the variance-covariance matrix
        need only be supplied up to a scaling factor.

    sw : float
        If weights were used to calculate `c` then `sw` is the sum of
        positive weight values; otherwise `sw` is the number of
        observations used to calculate `c`.

    isx : int, shape (`m`), modified in place
        `On entry`: the value of `isx`[j-1] determines the set of
        variables used to perform full stepwise model selection, for j =
        1,2,...,`m`.

        `isx`[j-1] = -1

            To exclude the variable corresponding to the jth column of X
            from the final model.

        `isx`[j-1] = 1

            To consider the variable corresponding to the jth column of
            X for selection in the final model.

        `isx`[j-1] = 2

            To force the inclusion of the variable corresponding to the
            jth column of X in the final model.

        `On exit`: the value of `isx`[j-1] indicates the status of the
        jth explanatory variable in the model.

        `isx`[j-1] = -1

            Forced exclusion.

        `isx`[j-1] = 0

            Excluded.

        `isx`[j-1] = 1

            Selected.

        `isx`[j-1] = 2

            Forced selection.

    fin : float
        The value of the variance ratio which an explanatory variable
        must exceed to be included in a model.

        Suggested value: `fin` = 4.0.

    fout : float
        The explanatory variable in a model with the lowest variance
        ratio value is removed from the model if its value is less than
        `fout`. `fout` is usually set equal to the value of `fin`; a
        value less than `fin` is occasionally preferred.

        Suggested value: `fout` = `fin`.

    tau : float
        The tolerance, tau, for detecting collinearities between
        variables when adding or removing an explanatory variable from a
        model. Explanatory variables deemed to be collinear are excluded
        from the final model.

        Suggested value: `tau` = 1.0*10^(-6).

    b : float, shape (`m`+1), modified in place
        `On exit`: `b`[0] contains the estimate for the intercept term
        in the fitted model. If `isx`[j-1] != 0, then `b`[j] contains
        the estimate for the jth explanatory variable in the fitted
        model; otherwise `b`[j] = 0.

    se : float, shape (`m`+1), modified in place
        `On exit`: `se`[j-1] contains the standard error for the
        estimate of `b`[j-1], for j = 1,2,...,`m`+1.

    monlev : int
        A value of 1 for `monlev` enables monitoring of the model
        selection process; a value of 0 disables it.

    monfun : None or callable monfun(flag, var, val, data=None)
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        The function for monitoring the model selection process.

        Parameters
        ~~~~~~~~~~
        flag : str, length 1
            The value of `flag` indicates the stage of the stepwise
            selection of explanatory variables.

            `flag` = 'A'

                Variable `var` was added to the current model.

            `flag` = 'B'

                Beginning the backward elimination step.

            `flag` = 'C'

                Variable `var` failed the collinearity test and is
                excluded from the model.

            `flag` = 'D'

                Variable `var` was dropped from the current model.

            `flag` = 'F'

                Beginning the forward selection step

            `flag` = 'K'

                Backward elimination did not remove any variables from
                the current model.

            `flag` = 'S'

                Starting stepwise selection procedure.

            `flag` = 'V'

                The variance ratio for variable `var` takes the value
                `val`.

            `flag` = 'X'

                Finished stepwise selection procedure.

        var : int
            The index of the explanatory variable in the design matrix Z
            to which `flag` pertains.

        val : float
            If `flag` = 'V', `val` is the variance ratio value for the
            coefficient associated with explanatory variable index
            `var`.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

    data : arbitrary, optional
        User-communication data for callback functions.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    rsq : float
        The R^2-statistic for the fitted regression model.

    rms : float
        The mean square of residuals for the fitted regression model.

    df : int
        The number of degrees of freedom for the sum of squares of
        residuals.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` > 1.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` > 1.

        (`errno` 1)
            On entry, `sw` = *<value>*.

            Constraint: `sw` > 1.0.

        (`errno` 1)
            On entry, `fin` = *<value>*.

            Constraint: `fin` > 0.0.

        (`errno` 1)
            On entry, `fout` = *<value>*; `fin` = *<value>*.

            Constraint: 0.0 <= `fout` <= `fin`.

        (`errno` 1)
            On entry, `tau` = *<value>*.

            Constraint: `tau` > 0.0.

        (`errno` 1)
            On entry, `monlev` = *<value>*.

            Constraint: `monlev` = 0 or 1.

        (`errno` 2)
            On entry at least one diagonal element of `c` <= 0.0.

            Constraint: `c` must be positive definite.

        (`errno` 2)
            On entry, `isx`[*<value>*] = *<value>*.

            Constraint: `isx`[j-1] = -1, 1 or 2, for j = 1,2,...,`m`.

        (`errno` 2)
            On entry, `isx`[i-1] != 1, for all i = 1,2,...,`m`.

            Constraint: there must be at least one free variable.

        (`errno` 4)
            All variables are collinear, no model to select.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            Matrix not positive definite, results may be inaccurate.

    References
    ----------
    Clarke, M R B, 1981, `Algorithm AS 178: the Gauss--Jordan sweep
    operator with detection of collinearity`, Appl. Statist. (31),
    166--169

    Dempster, A P, 1969, `Elements of Continuous Multivariate Analysis`,
    Addison--Wesley

    Draper, N R and Smith, H, 1985, `Applied Regression Analysis`, (2nd
    Edition), Wiley

    See Also
    --------
    :meth:`naginterfaces.library.examples.correg.linregm_fit_stepwise_ex.main`
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.linregm_fit_stepwise'

    utils.FileObjManager._check_type_compat(
        io_manager, {'fun_name': fun_name, 'entity': 'io_manager'}
    )

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    def cb_print_rech_py(_, en_data, nout, rec, ifail, rec_len):
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data, utils._EngineIntScalarType.to_py(nout.contents), rec[:rec_len], io_manager, excs, ifail.contents,
                fun_name=fun_name,
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    monfunh = utils._EngineState._sys_functype(
        None,
        utils._UNUSED_FUNCTYPE,
        utils._BYREF_ENGINEDATA,
        utils._PRINT_REC_FUNCTYPE,
        utils._PRINT_RECH_FUNCTYPE,
        utils._BYREF_STRING, # Input, flag(len=1).
        utils._BYREF_INT, # Input, var.
        utils._BYREF_FLOAT64, # Input, val.
        utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of flag.
    )

    if monfun is None:
        monfun_auxil_ctypes = utils._EngineState._get_symbol('g02efht')
        monfun_auxil_ctypes.restype = None
        monfun_auxil_ctypes.argtypes = (
            utils._BYREF_ENGINEDATA,
            utils._PRINT_REC_FUNCTYPE,
            utils._PRINT_RECH_FUNCTYPE,
            utils._BYREF_STRING, # Input, flag(len=1).
            utils._BYREF_INT, # Input, var.
            utils._BYREF_FLOAT64, # Input, val.
            utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
            utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
            utils._EngineState._chrlen_ctype, # Input, hidden chrlen of flag.
        )

    def cb_monfunh(_monfun, en_data, print_rec, print_rech, flag, var, val, iuser, ruser, flag_chrlen):
        """Helper for monfun."""
        try:
            if monfun is None:
                monfun_auxil_ctypes(
                    en_data,
                    print_rec,
                    print_rech,
                    flag,
                    var,
                    val,
                    iuser,
                    ruser,
                    1,
                )
                return
            _flag_monfunh = utils._EngineCharScalarType.to_py(flag.contents)
            _var_monfunh = utils._EngineIntScalarType.to_py(var.contents)
            _val_monfunh = utils._EngineFloat64ScalarType.to_py(val.contents)
            if data is not None:
                monfun(
                    _flag_monfunh,
                    _var_monfunh,
                    _val_monfunh,
                    data,
                    1,
                )
            else:
                monfun(
                    _flag_monfunh,
                    _var_monfunh,
                    _val_monfunh,
                    1,
                )
        except:
            excs.append(_sys.exc_info())
            en_data.contents.hlperr = utils._IERR_HELPER

    engine_ctypes = utils._EngineState._get_symbol('g02efft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._PRINT_REC_FUNCTYPE,
        utils._PRINT_RECH_FUNCTYPE,
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, wmean[m+1].
        utils._EngineFloat64ArrayType.c_type, # Input, c[(m+1)*(m+2)//2].
        utils._BYREF_FLOAT64, # Input, sw.
        utils._EngineIntArrayType.c_type, # Input/Output, isx[m].
        utils._BYREF_FLOAT64, # Input, fin.
        utils._BYREF_FLOAT64, # Input, fout.
        utils._BYREF_FLOAT64, # Input, tau.
        utils._EngineFloat64ArrayType.c_type, # Output, b[m+1].
        utils._EngineFloat64ArrayType.c_type, # Output, se[m+1].
        utils._BYREF_FLOAT64, # Output, rsq.
        utils._BYREF_FLOAT64, # Output, rms.
        utils._BYREF_INT, # Output, df.
        utils._BYREF_INT, # Input, monlev.
        utils._UNUSED_FUNCTYPE,
        monfunh,
        utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _print_rech = utils._PRINT_RECH_FUNCTYPE(cb_print_rech_py)
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _wmean = utils._EngineFloat64ArrayType(
        wmean, {'fun_name': fun_name, 'entity': 'wmean'},
        exp_shape=(_m_py+1,),
    ).to_ctypes()
    _c = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=((_m_py+1)*(_m_py+2)//2,),
    ).to_ctypes()
    _sw = utils._EngineFloat64ScalarType(
        sw, {'fun_name': fun_name, 'entity': 'sw'},
    ).to_ctypes()
    _isx_cls = utils._EngineIntArrayType(
        isx, {'fun_name': fun_name, 'entity': 'isx'},
        exp_shape=(_m_py,),
    )
    _isx = _isx_cls.to_ctypes()
    _fin = utils._EngineFloat64ScalarType(
        fin, {'fun_name': fun_name, 'entity': 'fin'},
    ).to_ctypes()
    _fout = utils._EngineFloat64ScalarType(
        fout, {'fun_name': fun_name, 'entity': 'fout'},
    ).to_ctypes()
    _tau = utils._EngineFloat64ScalarType(
        tau, {'fun_name': fun_name, 'entity': 'tau'},
    ).to_ctypes()
    _b_cls = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_m_py+1,),
        check_contents=False,
    )
    _b = _b_cls.to_ctypes(can_reference_contents=False)
    _se_cls = utils._EngineFloat64ArrayType(
        se, {'fun_name': fun_name, 'entity': 'se'},
        exp_shape=(_m_py+1,),
        check_contents=False,
    )
    _se = _se_cls.to_ctypes(can_reference_contents=False)
    _rsq = utils._EngineFloat64ScalarType.empty_ctypes()
    _rms = utils._EngineFloat64ScalarType.empty_ctypes()
    _df = utils._EngineIntScalarType.empty_ctypes()
    _monlev = utils._EngineIntScalarType(
        monlev, {'fun_name': fun_name, 'entity': 'monlev'},
    ).to_ctypes()
    utils._check_callable(monfun, {'fun_name': fun_name, 'entity': 'monfun'}, allow_none=True)
    _monfun = utils._UNUSED_CALLBACK
    _monfunh = utils._cast_callback(
        monfunh, cb_monfunh,
        {'fun_name': fun_name, 'entity': 'monfun'},
    )
    _iuser = None
    _ruser = None
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        utils._UNUSED_PRINT_REC,
        _print_rech,
        _m,
        _n,
        _wmean,
        _c,
        _sw,
        _isx,
        _fin,
        _fout,
        _tau,
        _b,
        _se,
        _rsq,
        _rms,
        _df,
        _monlev,
        _monfun,
        _monfunh,
        _iuser,
        _ruser,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 101:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m > 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 102:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n > 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 103:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, sw = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: sw > 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 107:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, fin = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: fin > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 108:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, fout = {:s}; fin = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: 0.0 <= fout <= fin.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 109:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, tau = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: tau > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 110:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, monlev = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: monlev = 0 or 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 104:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry at least one diagonal element of c <= 0.0.',
                    'Constraint: c must be positive definite.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 105:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, isx[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: isx[j-1] = -1, 1 or 2, for j = 1,2,...,m.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 106:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, isx[i-1] != 1, for all i = 1,2,...,m.',
                    'Constraint: there must be at least one free variable.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 111:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Matrix not positive definite, results may be inaccurate.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 112:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'All variables are collinear, no model to select.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _isx_cls.refresh_py(_isx)
        _b_cls.refresh_py(_b)
        _se_cls.refresh_py(_se)
        _rsq_py = utils._EngineFloat64ScalarType.to_py(_rsq)
        _rms_py = utils._EngineFloat64ScalarType.to_py(_rms)
        _df_py = utils._EngineIntScalarType.to_py(_df)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['rsq', 'rms', 'df'],
        )(
            _rsq_py,
            _rms_py,
            _df_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def linregm_stat_resinf(n, ip, nres, res, h, rms, sres):
    u"""
    Calculates standardized residuals and influence statistics.

    ``linregm_stat_resinf`` calculates two types of standardized
    residuals and two measures of influence for a linear regression.

    For full information please refer to the NAG Library document for
    g02fa

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02faf.html

    Parameters
    ----------
    n : int
        n, the number of observations included in the regression.

    ip : int
        p, the number of linear parameters estimated in the regression
        model.

    nres : int
        The number of residuals.

    res : float, shape (`nres`)
        The residuals, r_i.

    h : float, shape (`nres`)
        The diagonal elements of H, h_i, corresponding to the residuals
        in `res`.

    rms : float
        The estimate of sigma^2 based on all n observations, s^2, i.e.,
        the residual mean square.

    sres : float, shape (`nres`, 4), modified in place
        `On exit`: the standardized residuals and influence statistics.

        For the observation with residual, r_i, given in `res`[i-1].

        `sres`[i-1,0]

            Is the internally standardized residual, RI_i.

        `sres`[i-1,1]

            Is the externally standardized residual, RE_i.

        `sres`[i-1,2]

            Is Cook's D statistic, D_i.

        `sres`[i-1,3]

            Is Atkinson's T statistic, T_i.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `nres` = *<value>* and `n` = *<value>*.

            Constraint: `nres` <= `n`.

        (`errno` 1)
            On entry, `rms` = *<value>*.

            Constraint: `rms` > 0.0.

        (`errno` 1)
            On entry, `ip` = *<value>* and `n` = *<value>*.

            Constraint: `n`-1 > `ip`.

        (`errno` 1)
            On entry, `ip` = *<value>*.

            Constraint: `ip` >= 1.

        (`errno` 1)
            On entry, `nres` = *<value>*.

            Constraint: `nres` >= 1.

        (`errno` 2)
            On entry, `h`[*<value>*] = *<value>*.

            Constraint: 0.0 < `h`[i] < 1.0, for all i.

        (`errno` 3)
            On entry, a value in `res` is too large for given `rms`.
            `res`[*<value>*] = *<value>* and `rms` = *<value>*.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Atkinson, A C, 1981, `Two graphical displays for outlying and
    influential observations in regression`, Biometrika (68), 13--20

    Cook, R D and Weisberg, S, 1982, `Residuals and Influence in
    Regression`, Chapman and Hall
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.linregm_stat_resinf'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02faft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, ip.
        utils._BYREF_INT, # Input, nres.
        utils._EngineFloat64ArrayType.c_type, # Input, res[nres].
        utils._EngineFloat64ArrayType.c_type, # Input, h[nres].
        utils._BYREF_FLOAT64, # Input, rms.
        utils._EngineFloat64ArrayType.c_type, # Output, sres[ldsres, 4].
        utils._BYREF_INT, # Input, ldsres.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((sres, 'sres'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _ip = utils._EngineIntScalarType(
        ip, {'fun_name': fun_name, 'entity': 'ip'},
    ).to_ctypes()
    _nres = utils._EngineIntScalarType(
        nres, {'fun_name': fun_name, 'entity': 'nres'},
    ).to_ctypes()
    _nres_py = utils._EngineIntScalarType.to_py(_nres)
    _res = utils._EngineFloat64ArrayType(
        res, {'fun_name': fun_name, 'entity': 'res'},
        exp_shape=(_nres_py,),
    ).to_ctypes()
    _h = utils._EngineFloat64ArrayType(
        h, {'fun_name': fun_name, 'entity': 'h'},
        exp_shape=(_nres_py,),
    ).to_ctypes()
    _rms = utils._EngineFloat64ScalarType(
        rms, {'fun_name': fun_name, 'entity': 'rms'},
    ).to_ctypes()
    _ldsres_py = utils._EngineFloat64ArrayType.get_nd(sres, {'fun_name': fun_name, 'entity': 'sres'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldsres = utils._EngineIntScalarType(
        _ldsres_py, {'fun_name': fun_name, 'entity': '_ldsres'},
    ).to_ctypes()
    _sres_cls = utils._EngineFloat64ArrayType(
        sres, {'fun_name': fun_name, 'entity': 'sres'},
        exp_shape=(_nres_py, 4),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _sres = _sres_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _ip,
        _nres,
        _res,
        _h,
        _rms,
        _sres,
        _ldsres,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nres = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nres <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, rms = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: rms > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ip = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: n-1 > ip.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ip >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nres = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nres >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, h[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: 0.0 < h[i] < 1.0, for all i.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a value in res is too large for given rms.',
                    'res[{:s}] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'rms = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _sres_cls.refresh_py(_sres)
        return_data = None
    _err_class.manage()
    return return_data

def linregm_stat_durbwat(n, ip, res):
    u"""
    Computes Durbin--Watson test statistic.

    ``linregm_stat_durbwat`` calculates the Durbin--Watson statistic,
    for a set of residuals, and the upper and lower bounds for its
    significance.

    For full information please refer to the NAG Library document for
    g02fc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02fcf.html

    Parameters
    ----------
    n : int
        n, the number of residuals.

    ip : int
        p, the number of independent variables in the regression model,
        including the mean.

    res : float, shape (`n`)
        The residuals, r_1,r_2,...,r_n.

    Returns
    -------
    d : float
        The Durbin--Watson statistic, d.

    pdl : float
        Lower bound for the significance of the Durbin--Watson
        statistic, p_l.

    pdu : float
        Upper bound for the significance of the Durbin--Watson
        statistic, p_u.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `ip` = *<value>*.

            Constraint: `ip` >= 1.

        (`errno` 1)
            On entry, `n` = *<value>* and `ip` = *<value>*.

            Constraint: `n` > `ip`.

        (`errno` 2)
            On entry, mean of `res` = *<value>*.

            Constraint: the mean of the residuals <= sqrt(epsilon),
            where epsilon = machine precision.

        (`errno` 3)
            On entry, all residuals are identical.

    References
    ----------
    Durbin, J and Watson, G S, 1950, `Testing for serial correlation in
    least squares regression. I`, Biometrika (37), 409--428

    Durbin, J and Watson, G S, 1951, `Testing for serial correlation in
    least squares regression. II`, Biometrika (38), 159--178

    Durbin, J and Watson, G S, 1971, `Testing for serial correlation in
    least squares regression. III`, Biometrika (58), 1--19

    Granger, C W J and Newbold, P, 1986, `Forecasting Economic Time
    Series`, (2nd Edition), Academic Press

    Newbold, P, 1988, `Statistics for Business and Economics`,
    Prentice--Hall
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.linregm_stat_durbwat'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02fcft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, ip.
        utils._EngineFloat64ArrayType.c_type, # Input, res[n].
        utils._BYREF_FLOAT64, # Output, d.
        utils._BYREF_FLOAT64, # Output, pdl.
        utils._BYREF_FLOAT64, # Output, pdu.
        utils._EngineFloat64ArrayType.c_type, # Workspace, work.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _ip = utils._EngineIntScalarType(
        ip, {'fun_name': fun_name, 'entity': 'ip'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _res = utils._EngineFloat64ArrayType(
        res, {'fun_name': fun_name, 'entity': 'res'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _d = utils._EngineFloat64ScalarType.empty_ctypes()
    _pdl = utils._EngineFloat64ScalarType.empty_ctypes()
    _pdu = utils._EngineFloat64ScalarType.empty_ctypes()
    _work = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _ip,
        _res,
        _d,
        _pdl,
        _pdu,
        _work,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ip >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s} and ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: n > ip.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mean of res = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: the mean of the residuals <= sqrt(epsilon), where',
                    'epsilon = machine precision.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, all residuals are identical.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _d_py = utils._EngineFloat64ScalarType.to_py(_d)
        _pdl_py = utils._EngineFloat64ScalarType.to_py(_pdl)
        _pdu_py = utils._EngineFloat64ScalarType.to_py(_pdu)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['d', 'pdl', 'pdu'],
        )(
            _d_py,
            _pdl_py,
            _pdu_py,
        )
    _err_class.manage()
    return return_data

def glm_normal(link, mean, offset, n, x, m, isx, y, wt, s, a, b, se, cov, v, tol, maxit, iprint, eps, io_manager=None):
    u"""
    Fits a generalized linear model with Normal errors.

    ``glm_normal`` fits a generalized linear model with normal errors.

    For full information please refer to the NAG Library document for
    g02ga

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02gaf.html

    Parameters
    ----------
    link : str, length 1
        Indicates which link function is to be used.

        `link` = 'E'

            An exponent link is used.

        `link` = 'I'

            An identity link is used.

        `link` = 'L'

            A log link is used.

        `link` = 'S'

            A square root link is used.

        `link` = 'R'

            A reciprocal link is used.

    mean : str, length 1
        Indicates if a mean term is to be included.

        `mean` = 'M'

            A mean term, intercept, will be included in the model.

        `mean` = 'Z'

            The model will pass through the origin, zero-point.

    offset : str, length 1
        Indicates if an offset is required.

        `offset` = 'Y'

            An offset is required and the offsets must be supplied in
            the seventh column of `v`.

        `offset` = 'N'

            No offset is required.

    n : int
        n, the number of observations.

    x : float, shape (`n`, `m`)
        `x`[i-1,j-1] must contain the ith observation for the jth
        independent variable, for j = 1,2,...,`m`, for i = 1,2,...,`n`.

    m : int
        m, the total number of independent variables.

    isx : int, shape (`m`)
        Indicates which independent variables are to be included in the
        model.

        If `isx`[j-1] > 0, the variable contained in the jth column of
        `x` is included in the regression model.

    y : float, shape (`n`)
        The observations on the dependent variable, y_i, for i =
        1,2,...,n.

    wt : None or float, shape (`n`)
        If provided `wt` must contain the weights to be used with the
        model.

        If `wt`[i-1] = 0.0, the ith observation is not included in the
        model, in which case the effective number of observations is the
        number of observations with nonzero weights.

        If `wt` is not provided the effective number of observations is
        n.

    s : float
        The scale parameter for the model, sigma^2.

        If `s` = 0.0, the scale parameter is estimated with the function
        using the residual mean square.

    a : float
        If `link` = 'E', `a` must contain the power of the exponential.

        If `link` != 'E', `a` is not referenced.

    b : float, shape (ip), modified in place
        `On exit`: the estimates of the parameters of the generalized
        linear model, beta^.

        If `mean` = 'M', `b`[0] will contain the estimate of the mean
        parameter and `b`[i] will contain the coefficient of the
        variable contained in column j of `x`, where `isx`[j-1] is the
        ith positive value in the array `isx`.

        If `mean` = 'Z', `b`[i-1] will contain the coefficient of the
        variable contained in column j of `x`, where `isx`[j-1] is the
        ith positive value in the array `isx`.

    se : float, shape (ip), modified in place
        `On exit`: the standard errors of the linear parameters.

        `se`[i-1] contains the standard error of the parameter estimate
        in `b`[i-1], for i = 1,2,...,ip.

    cov : float, shape (ip*(ip+1)/2), modified in place
        `On exit`: the upper triangular part of the variance-covariance
        matrix of the ip parameter estimates given in `b`. They are
        stored packed by column, i.e., the covariance between the
        parameter estimate given in `b`[i-1] and the parameter estimate
        given in `b`[j-1], j >= i, is stored in `cov`[(j*(j-1)/2+i)-1].

    v : float, shape (`n`, ip+7), modified in place
        `On entry`: if `offset` = 'N', `v` need not be set.

        If `offset` = 'Y', `v`[i-1,6], for i = 1,2,...,n, must contain
        the offset values o_i.

        All other values need not be set.

        `On exit`: auxiliary information on the fitted model.

        [table omitted]

        If the model is not of full rank, i.e., `irank` < ip, the first
        ip rows of columns 8 to ip+7 contain the P* matrix.

    tol : float
        Indicates the accuracy required for the fit of the model.

        The iterative weighted least squares procedure is deemed to have
        converged if the absolute change in deviance between
        interactions is less than `tol`*(1.0+current residual sum of
        squares).

        This is approximately an absolute precision if the residual sum
        of squares is small and a relative precision if the residual sum
        of squares is large.

        If 0.0 <= `tol` < machine precision, ``glm_normal`` will use
        10*machine precision.

    maxit : int
        The maximum number of iterations for the iterative weighted
        least squares.

        If `maxit` = 0, a default value of 10 is used.

    iprint : int
        Indicates if the printing of information on the iterations is
        required.

        `iprint` <= 0

            There is no printing.

        `iprint` > 0

            Every `iprint` iteration, the following is printed:

                the deviance,

                the current estimates,

                and if the weighted least squares equations are
                singular, then this is indicated.

        When printing occurs the output is directed to the file object
        associated with the advisory I/O unit (see
        :class:`~naginterfaces.base.utils.FileObjManager`).

    eps : float
        The value of `eps` is used to decide if the independent
        variables are of full rank and, if not, what is the rank of the
        independent variables. The smaller the value of `eps` the
        stricter the criterion for selecting the singular value
        decomposition.

        If 0.0 <= `eps` < machine precision, the function will use
        machine precision instead.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    s : float
        If on input `s` = 0.0, `s` contains the estimated value of the
        scale parameter, sigma^^2.

        If on input `s` != 0.0, `s` is unchanged on exit.

    rss : float
        The residual sum of squares for the fitted model.

    idf : int
        The degrees of freedom associated with the residual sum of
        squares for the fitted model.

    irank : int
        The rank of the independent variables.

        If the model is of full rank, `irank` = ip.

        If the model is not of full rank, `irank` is an estimate of the
        rank of the independent variables. `irank` is calculated as the
        number of singular values greater than `eps`* (largest singular
        value).

        It is possible for the SVD to be carried out but for `irank` to
        be returned as ip.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, ip = *<value>*.

            Constraint: ip >= 1.

        (`errno` 1)
            On entry, `eps` = *<value>*.

            Constraint: `eps` >= 0.0.

        (`errno` 1)
            On entry, `tol` = *<value>*.

            Constraint: `tol` >= 0.0.

        (`errno` 1)
            On entry, `maxit` = *<value>*.

            Constraint: `maxit` >= 0.

        (`errno` 1)
            On entry, `offset` = *<value>*.

            Constraint: `offset` = 'Y' or 'N'.

        (`errno` 1)
            On entry, `link` = *<value>*.

            Constraint: `link` = 'E', 'I', 'L', 'S' or 'R'.

        (`errno` 1)
            On entry, `a` = 0.0 and `link` = 'E'.

            Constraint: if `link` = 'E', `a` != 0.0.

        (`errno` 1)
            On entry, `s` = *<value>*.

            Constraint: `s` >= 0.0.

        (`errno` 1)
            On entry, weight = *<value>*.

            Constraint: weight = 'W' or 'U'.

        (`errno` 1)
            On entry, `mean` = *<value>*.

            Constraint: `mean` = 'M' or 'Z'.

        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 2.

        (`errno` 2)
            On entry, `wt`[*<value>*] < 0.0.

            Constraint: `wt`[i-1] >= 0.0, for i = 1,2,...,n.

        (`errno` 3)
            On entry, `isx`[*<value>*] < 0.

            Constraint: `isx`[j-1] >= 0.0, for j = 1,2,...,`m`.

        (`errno` 3)
            On entry, ip incompatible with number of nonzero values of
            `isx`: ip = *<value>*.

        (`errno` 3)
            Number of requested x-variables greater than `n`.

        (`errno` 4)
            A fitted value is at a boundary. This will only occur with
            `link` = 'L', 'R' or 'E'. This may occur if there are small
            values of y and the model is not suitable for the data. The
            model should be reformulated with, perhaps, some
            observations dropped.

        (`errno` 5)
            SVD solution failed to converge.

        (`errno` 6)
            The iterative weighted least squares has failed to converge
            in `maxit` (or default 10) iterations. The value of `maxit`
            could be increased but it may be advantageous to examine the
            convergence using the `iprint` option. This may indicate
            that the convergence is slow because the solution is at a
            boundary in which case it may be better to reformulate the
            model.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 7)
            The rank of the model has changed during the weighted least
            squares iterations. The estimate for beta returned may be
            reasonable, but you should check how the deviance has
            changed during iterations.

        (`errno` 8)
            The degrees of freedom for error are 0. A saturated model
            has been fitted.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Cook, R D and Weisberg, S, 1982, `Residuals and Influence in
    Regression`, Chapman and Hall

    McCullagh, P and Nelder, J A, 1983, `Generalized Linear Models`,
    Chapman and Hall

    See Also
    --------
    :meth:`naginterfaces.library.examples.correg.glm_normal_ex.main`
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.glm_normal'

    utils.FileObjManager._check_type_compat(
        io_manager, {'fun_name': fun_name, 'entity': 'io_manager'}
    )

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    def cb_print_rech_py(_, en_data, nout, rec, ifail, rec_len):
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data, utils._EngineIntScalarType.to_py(nout.contents), rec[:rec_len], io_manager, excs, ifail.contents,
                fun_name=fun_name,
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02gaft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._PRINT_REC_FUNCTYPE,
        utils._PRINT_RECH_FUNCTYPE,
        utils._BYREF_INT, # Input, sorder.
        utils._BYREF_STRING, # Input, link(len=1).
        utils._BYREF_STRING, # Input, mean(len=1).
        utils._BYREF_STRING, # Input, offset(len=1).
        utils._BYREF_STRING, # Input, weight(len=1).
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[ldx, *].
        utils._BYREF_INT, # Input, ldx.
        utils._BYREF_INT, # Input, m.
        utils._EngineIntArrayType.c_type, # Input, isx[m].
        utils._BYREF_INT, # Input, ip.
        utils._EngineFloat64ArrayType.c_type, # Input, y[n].
        utils._EngineFloat64ArrayType.c_type, # Input, wt[n].
        utils._BYREF_FLOAT64, # Input/Output, s.
        utils._BYREF_FLOAT64, # Input, a.
        utils._BYREF_FLOAT64, # Output, rss.
        utils._BYREF_INT, # Output, idf.
        utils._EngineFloat64ArrayType.c_type, # Output, b[ip].
        utils._BYREF_INT, # Output, irank.
        utils._EngineFloat64ArrayType.c_type, # Output, se[ip].
        utils._EngineFloat64ArrayType.c_type, # Output, cov[ip*(ip+1)//2].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, v[ldv, ip+7].
        utils._BYREF_INT, # Input, ldv.
        utils._BYREF_FLOAT64, # Input, tol.
        utils._BYREF_INT, # Input, maxit.
        utils._BYREF_INT, # Input, iprint.
        utils._BYREF_INT, # Input, nout.
        utils._BYREF_FLOAT64, # Input, eps.
        utils._EngineFloat64ArrayType.c_type, # Workspace, wk.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of link.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of mean.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of offset.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of weight.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((x, 'x'), (v, 'v')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _print_rech = utils._PRINT_RECH_FUNCTYPE(cb_print_rech_py)
    _sorder = utils._EngineIntScalarType.c_type(_nag_enum_consts.nag_col_major)
    _link = utils._EngineCharScalarType(
        link, {'fun_name': fun_name, 'entity': 'link'}, exp_chrlen=1,
    ).to_ctypes()
    _mean = utils._EngineCharScalarType(
        mean, {'fun_name': fun_name, 'entity': 'mean'}, exp_chrlen=1,
    ).to_ctypes()
    _offset = utils._EngineCharScalarType(
        offset, {'fun_name': fun_name, 'entity': 'offset'}, exp_chrlen=1,
    ).to_ctypes()
    if wt is None:
        _weight = utils._EngineCharScalarType(
            'U', {'fun_name': fun_name, 'entity': '_weight'}, exp_chrlen=1,
        ).to_ctypes()
    else:
        _weight = utils._EngineCharScalarType(
            'W', {'fun_name': fun_name, 'entity': '_weight'}, exp_chrlen=1,
        ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _ldx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _sorder_py = utils._EngineIntScalarType.to_py(_sorder)
    _m_py = utils._EngineIntScalarType.to_py(_m)
    if _sorder_py == 1:
        _x_dim2 = _m_py
    elif _sorder_py == 2:
        _x_dim2 = _n_py
    else:
        _x_dim2 = 0
    exp_shape_x = (_n_py, _x_dim2)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=exp_shape_x,
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _isx = utils._EngineIntArrayType(
        isx, {'fun_name': fun_name, 'entity': 'isx'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _vl_p_py = 0
    for _vl_i_py in range(1, _m_py+1):
        if utils._EngineIntArrayType.get_element(isx, {'fun_name': fun_name, 'entity': 'isx'}, (_vl_i_py-1,)) > 0:
            _vl_p_py = _vl_p_py + 1
    _mean_py = utils._EngineCharScalarType.to_py(_mean)
    if _mean_py in ('M', 'm'):
        _vl_ip_py = _vl_p_py+1
    else:
        _vl_ip_py = _vl_p_py
    _ip = utils._EngineIntScalarType(
        _vl_ip_py, {'fun_name': fun_name, 'entity': '_ip'},
    ).to_ctypes()
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _wt = utils._EngineFloat64ArrayType(
        wt, {'fun_name': fun_name, 'entity': 'wt'},
        exp_shape=(_n_py,),
        allow_none=True,
    ).to_ctypes()
    _s = utils._EngineFloat64ScalarType(
        s, {'fun_name': fun_name, 'entity': 's'},
    ).to_ctypes()
    _a = utils._EngineFloat64ScalarType(
        a, {'fun_name': fun_name, 'entity': 'a'},
    ).to_ctypes()
    _rss = utils._EngineFloat64ScalarType.empty_ctypes()
    _idf = utils._EngineIntScalarType.empty_ctypes()
    _ip_py = utils._EngineIntScalarType.to_py(_ip)
    _b_cls = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_ip_py,),
        check_contents=False,
    )
    _b = _b_cls.to_ctypes(can_reference_contents=False)
    _irank = utils._EngineIntScalarType.empty_ctypes()
    _se_cls = utils._EngineFloat64ArrayType(
        se, {'fun_name': fun_name, 'entity': 'se'},
        exp_shape=(_ip_py,),
        check_contents=False,
    )
    _se = _se_cls.to_ctypes(can_reference_contents=False)
    _cov_cls = utils._EngineFloat64ArrayType(
        cov, {'fun_name': fun_name, 'entity': 'cov'},
        exp_shape=(_ip_py*(_ip_py+1)//2,),
        check_contents=False,
    )
    _cov = _cov_cls.to_ctypes(can_reference_contents=False)
    _ldv_py = utils._EngineFloat64ArrayType.get_nd(v, {'fun_name': fun_name, 'entity': 'v'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldv = utils._EngineIntScalarType(
        _ldv_py, {'fun_name': fun_name, 'entity': '_ldv'},
    ).to_ctypes()
    _v_cls = utils._EngineFloat64ArrayType(
        v, {'fun_name': fun_name, 'entity': 'v'},
        exp_shape=(_n_py, _ip_py+7),
        mutual_sorder=_m_sorder,
    )
    _v = _v_cls.to_ctypes()
    _tol = utils._EngineFloat64ScalarType(
        tol, {'fun_name': fun_name, 'entity': 'tol'},
    ).to_ctypes()
    _maxit = utils._EngineIntScalarType(
        maxit, {'fun_name': fun_name, 'entity': 'maxit'},
    ).to_ctypes()
    _iprint = utils._EngineIntScalarType(
        iprint, {'fun_name': fun_name, 'entity': 'iprint'},
    ).to_ctypes()
    _nout = utils._EngineIntScalarType.c_type(utils.FileObjManager.advunit)
    _eps = utils._EngineFloat64ScalarType(
        eps, {'fun_name': fun_name, 'entity': 'eps'},
    ).to_ctypes()
    _wk = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        utils._UNUSED_PRINT_REC,
        _print_rech,
        _sorder,
        _link,
        _mean,
        _offset,
        _weight,
        _n,
        _x,
        _ldx,
        _m,
        _isx,
        _ip,
        _y,
        _wt,
        _s,
        _a,
        _rss,
        _idf,
        _b,
        _irank,
        _se,
        _cov,
        _v,
        _ldv,
        _tol,
        _maxit,
        _iprint,
        _nout,
        _eps,
        _wk,
        _errbuf,
        _ierr,
        _link._length_,
        _mean._length_,
        _offset._length_,
        _weight._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99978:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ip >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99984:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, eps = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: eps >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99985:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, tol = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: tol >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99986:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, maxit = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: maxit >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99987:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, offset = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: offset = \'Y\' or \'N\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, link = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: link = \'E\', \'I\', \'L\', \'S\' or \'R\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a = 0.0 and link = \'E\'.',
                    'Constraint: if link = \'E\', a != 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, s = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: s >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, weight = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: weight = \'W\' or \'U\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mean = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mean = \'M\' or \'Z\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, wt[{:s}] < 0.0.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: wt[i-1] >= 0.0, for i = 1,2,...,n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99977:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, isx[{:s}] < 0.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: isx[j-1] >= 0.0, for j = 1,2,...,m.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ip incompatible with number of nonzero values of isx:',
                    'ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Number of requested x-variables greater than n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99982:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'A fitted value is at a boundary. This will only occur with',
                    'link = \'L\', \'R\' or \'E\'.',
                    'This may occur if there are small values of y',
                    'and the model is not suitable for the data. The model should be',
                    'reformulated with, perhaps, some observations dropped.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99981:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'SVD solution failed to converge.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99980:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The iterative weighted least squares has failed to converge in',
                    'maxit (or default 10) iterations. The',
                    'value of maxit could be increased but it may be advantageous to',
                    'examine the convergence using the iprint option. This may',
                    'indicate that the convergence is slow because the solution is at a',
                    'boundary in which case it may be better to reformulate the model.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 99983:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The rank of the model has changed during the weighted least squares',
                    'iterations. The estimate for beta returned',
                    'may be reasonable, but you should check how the deviance has changed',
                    'during iterations.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 99979:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The degrees of freedom for error are 0. A saturated model has been fitted.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _s_py = utils._EngineFloat64ScalarType.to_py(_s)
        _b_cls.refresh_py(_b)
        _se_cls.refresh_py(_se)
        _cov_cls.refresh_py(_cov)
        _v_cls.refresh_py(_v)
        _rss_py = utils._EngineFloat64ScalarType.to_py(_rss)
        _idf_py = utils._EngineIntScalarType.to_py(_idf)
        _irank_py = utils._EngineIntScalarType.to_py(_irank)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['s', 'rss', 'idf', 'irank'],
        )(
            _s_py,
            _rss_py,
            _idf_py,
            _irank_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def glm_binomial(link, mean, offset, n, x, m, isx, y, t, wt, b, se, cov, v, tol, maxit, iprint, eps, io_manager=None):
    u"""
    Fits a generalized linear model with binomial errors.

    ``glm_binomial`` fits a generalized linear model with binomial
    errors.

    For full information please refer to the NAG Library document for
    g02gb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02gbf.html

    Parameters
    ----------
    link : str, length 1
        Indicates which link function is to be used.

        `link` = 'G'

            A logistic link is used.

        `link` = 'P'

            A probit link is used.

        `link` = 'C'

            A complementary log-log link is used.

    mean : str, length 1
        Indicates if a mean term is to be included.

        `mean` = 'M'

            A mean term, intercept, will be included in the model.

        `mean` = 'Z'

            The model will pass through the origin, zero-point.

    offset : str, length 1
        Indicates if an offset is required.

        `offset` = 'Y'

            An offset is required and the offsets must be supplied in
            the seventh column of `v`.

        `offset` = 'N'

            No offset is required.

    n : int
        n, the number of observations.

    x : float, shape (`n`, `m`)
        `x`[i-1,j-1] must contain the ith observation for the jth
        independent variable, for j = 1,2,...,`m`, for i = 1,2,...,`n`.

    m : int
        m, the total number of independent variables.

    isx : int, shape (`m`)
        Indicates which independent variables are to be included in the
        model.

        If `isx`[j-1] > 0, the variable contained in the jth column of
        `x` is included in the regression model.

    y : float, shape (`n`)
        The observations on the dependent variable, y_i, for i =
        1,2,...,n.

    t : float, shape (`n`)
        t, the binomial denominator.

    wt : None or float, shape (`n`)
        If provided `wt` must contain the weights to be used with the
        model.

        If `wt`[i-1] = 0.0, the ith observation is not included in the
        model, in which case the effective number of observations is the
        number of observations with nonzero weights.

        If `wt` is not provided the effective number of observations is
        n.

    b : float, shape (ip), modified in place
        `On exit`: the estimates of the parameters of the generalized
        linear model, beta^.

        If `mean` = 'M', the first element of `b` will contain the
        estimate of the mean parameter and `b`[i] will contain the
        coefficient of the variable contained in column j of `x`, where
        `isx`[j-1] is the ith positive value in the array `isx`.

        If `mean` = 'Z', `b`[i-1] will contain the coefficient of the
        variable contained in column j of `x`, where `isx`[j-1] is the
        ith positive value in the array `isx`.

    se : float, shape (ip), modified in place
        `On exit`: the standard errors of the linear parameters.

        `se`[i-1] contains the standard error of the parameter estimate
        in `b`[i-1], for i = 1,2,...,ip.

    cov : float, shape (ip*(ip+1)/2), modified in place
        `On exit`: the upper triangular part of the variance-covariance
        matrix of the ip parameter estimates given in `b`. They are
        stored in packed form by column, i.e., the covariance between
        the parameter estimate given in `b`[i-1] and the parameter
        estimate given in `b`[j-1], j >= i, is stored in
        `cov`[(j*(j-1)/2+i)-1].

    v : float, shape (`n`, ip+7), modified in place
        `On entry`: if `offset` = 'N', `v` need not be set.

        If `offset` = 'Y', `v`[i-1,6], for i = 1,2,...,n must contain
        the offset values o_i.

        All other values need not be set.

        `On exit`: auxiliary information on the fitted model.

        [table omitted]

        If the model is not of full rank, i.e., `irank` < ip, the first
        ip rows of columns 8 to ip+7 contain the P* matrix.

    tol : float
        Indicates the accuracy required for the fit of the model.

        The iterative weighted least squares procedure is deemed to have
        converged if the absolute change in deviance between iterations
        is less than `tol`*(1.0+Current Deviance).

        This is approximately an absolute precision if the deviance is
        small and a relative precision if the deviance is large.

        If 0.0 <= `tol` < machine precision, the function will use
        10*machine precision instead.

    maxit : int
        The maximum number of iterations for the iterative weighted
        least squares.

        If `maxit` = 0, a default value of 10 is used.

    iprint : int
        Indicates if the printing of information on the iterations is
        required.

        `iprint` <= 0

            There is no printing.

        `iprint` > 0

            The following is printed every `iprint` iterations:

                the deviance,

                the current estimates,

                and if the weighted least squares equations are
                singular, then this is indicated.

        When printing occurs the output is directed to the file object
        associated with the advisory I/O unit (see
        :class:`~naginterfaces.base.utils.FileObjManager`).

    eps : float
        The value of `eps` is used to decide if the independent
        variables are of full rank and, if not, what is the rank of the
        independent variables. The smaller the value of `eps` the
        stricter the criterion for selecting the singular value
        decomposition.

        If 0.0 <= `eps` < machine precision, the function will use
        machine precision instead.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    dev : float
        The deviance for the fitted model.

    idf : int
        The degrees of freedom associated with the deviance for the
        fitted model.

    irank : int
        The rank of the independent variables.

        If the model is of full rank, `irank` = ip.

        If the model is not of full rank, `irank` is an estimate of the
        rank of the independent variables. `irank` is calculated as the
        number of singular values greater that `eps`*(largest singular
        value).

        It is possible for the SVD to be carried out but for `irank` to
        be returned as ip.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, ip = *<value>*.

            Constraint: ip >= 1.

        (`errno` 1)
            On entry, `eps` = *<value>*.

            Constraint: `eps` >= 0.0.

        (`errno` 1)
            On entry, `tol` = *<value>*.

            Constraint: `tol` >= 0.0.

        (`errno` 1)
            On entry, `maxit` = *<value>*.

            Constraint: `maxit` >= 0.

        (`errno` 1)
            On entry, `offset` = *<value>*.

            Constraint: `offset` = 'Y' or 'N'.

        (`errno` 1)
            On entry, `link` = *<value>*.

            Constraint: `link` = 'G', 'P' or 'C'.

        (`errno` 1)
            On entry, weight = *<value>*.

            Constraint: weight = 'W' or 'U'.

        (`errno` 1)
            On entry, `mean` = *<value>*.

            Constraint: `mean` = 'M' or 'Z'.

        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 2.

        (`errno` 2)
            On entry, `wt`[*<value>*] < 0.0.

            Constraint: `wt`[i-1] >= 0.0, for i = 1,2,...,n.

        (`errno` 3)
            On entry, `isx`[*<value>*] < 0.

            Constraint: `isx`[j-1] >= 0.0, for j = 1,2,...,`m`.

        (`errno` 3)
            On entry, ip incompatible with number of nonzero values of
            `isx`: ip = *<value>*.

        (`errno` 3)
            Number of requested x-variables greater than `n`.

        (`errno` 4)
            On entry, `t`[*<value>*] = *<value>*.

            Constraint: `t`[i-1] >= 0, for i = 1,2,...,n.

        (`errno` 5)
            On entry, `y`[*<value>*] = *<value>* and `t`[*<value>*] =
            *<value>*.

            Constraint: 0.0 <= `y`[i-1] <= `t`[i-1], for i = 1,2,...,n.

        (`errno` 6)
            A fitted value is at the boundary, i.e., 0.0 or 1.0. This
            may occur if there are y values of 0.0 or t and the model is
            too complex for the data. The model should be reformulated
            with, perhaps, some observations dropped.

        (`errno` 7)
            SVD solution failed to converge.

        (`errno` 8)
            The iterative weighted least squares has failed to converge
            in `maxit` (or default 10) iterations. The value of `maxit`
            could be increased but it may be advantageous to examine the
            convergence using the `iprint` option. This may indicate
            that the convergence is slow because the solution is at a
            boundary in which case it may be better to reformulate the
            model.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 9)
            The rank of the model has changed during the weighted least
            squares iterations. The estimate for beta returned may be
            reasonable, but you should check how the deviance has
            changed during iterations.

        (`errno` 10)
            Degrees of freedom for error are 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Cook, R D and Weisberg, S, 1982, `Residuals and Influence in
    Regression`, Chapman and Hall

    Cox, D R, 1983, `Analysis of Binary Data`, Chapman and Hall

    McCullagh, P and Nelder, J A, 1983, `Generalized Linear Models`,
    Chapman and Hall

    See Also
    --------
    :meth:`naginterfaces.library.examples.correg.glm_binomial_ex.main`
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.glm_binomial'

    utils.FileObjManager._check_type_compat(
        io_manager, {'fun_name': fun_name, 'entity': 'io_manager'}
    )

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    def cb_print_rech_py(_, en_data, nout, rec, ifail, rec_len):
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data, utils._EngineIntScalarType.to_py(nout.contents), rec[:rec_len], io_manager, excs, ifail.contents,
                fun_name=fun_name,
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02gbft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._PRINT_REC_FUNCTYPE,
        utils._PRINT_RECH_FUNCTYPE,
        utils._BYREF_INT, # Input, sorder.
        utils._BYREF_STRING, # Input, link(len=1).
        utils._BYREF_STRING, # Input, mean(len=1).
        utils._BYREF_STRING, # Input, offset(len=1).
        utils._BYREF_STRING, # Input, weight(len=1).
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[ldx, *].
        utils._BYREF_INT, # Input, ldx.
        utils._BYREF_INT, # Input, m.
        utils._EngineIntArrayType.c_type, # Input, isx[m].
        utils._BYREF_INT, # Input, ip.
        utils._EngineFloat64ArrayType.c_type, # Input, y[n].
        utils._EngineFloat64ArrayType.c_type, # Input, t[n].
        utils._EngineFloat64ArrayType.c_type, # Input, wt[n].
        utils._BYREF_FLOAT64, # Output, dev.
        utils._BYREF_INT, # Output, idf.
        utils._EngineFloat64ArrayType.c_type, # Output, b[ip].
        utils._BYREF_INT, # Output, irank.
        utils._EngineFloat64ArrayType.c_type, # Output, se[ip].
        utils._EngineFloat64ArrayType.c_type, # Output, cov[ip*(ip+1)//2].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, v[ldv, ip+7].
        utils._BYREF_INT, # Input, ldv.
        utils._BYREF_FLOAT64, # Input, tol.
        utils._BYREF_INT, # Input, maxit.
        utils._BYREF_INT, # Input, iprint.
        utils._BYREF_INT, # Input, nout.
        utils._BYREF_FLOAT64, # Input, eps.
        utils._EngineFloat64ArrayType.c_type, # Workspace, wk.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of link.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of mean.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of offset.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of weight.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((x, 'x'), (v, 'v')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _print_rech = utils._PRINT_RECH_FUNCTYPE(cb_print_rech_py)
    _sorder = utils._EngineIntScalarType.c_type(_nag_enum_consts.nag_col_major)
    _link = utils._EngineCharScalarType(
        link, {'fun_name': fun_name, 'entity': 'link'}, exp_chrlen=1,
    ).to_ctypes()
    _mean = utils._EngineCharScalarType(
        mean, {'fun_name': fun_name, 'entity': 'mean'}, exp_chrlen=1,
    ).to_ctypes()
    _offset = utils._EngineCharScalarType(
        offset, {'fun_name': fun_name, 'entity': 'offset'}, exp_chrlen=1,
    ).to_ctypes()
    if wt is None:
        _weight = utils._EngineCharScalarType(
            'U', {'fun_name': fun_name, 'entity': '_weight'}, exp_chrlen=1,
        ).to_ctypes()
    else:
        _weight = utils._EngineCharScalarType(
            'W', {'fun_name': fun_name, 'entity': '_weight'}, exp_chrlen=1,
        ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _ldx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _sorder_py = utils._EngineIntScalarType.to_py(_sorder)
    _m_py = utils._EngineIntScalarType.to_py(_m)
    if _sorder_py == 1:
        _x_dim2 = _m_py
    elif _sorder_py == 2:
        _x_dim2 = _n_py
    else:
        _x_dim2 = 0
    exp_shape_x = (_n_py, _x_dim2)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=exp_shape_x,
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _isx = utils._EngineIntArrayType(
        isx, {'fun_name': fun_name, 'entity': 'isx'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _vl_p_py = 0
    for _vl_i_py in range(1, _m_py+1):
        if utils._EngineIntArrayType.get_element(isx, {'fun_name': fun_name, 'entity': 'isx'}, (_vl_i_py-1,)) > 0:
            _vl_p_py = _vl_p_py + 1
    _mean_py = utils._EngineCharScalarType.to_py(_mean)
    if _mean_py in ('M', 'm'):
        _vl_ip_py = _vl_p_py+1
    else:
        _vl_ip_py = _vl_p_py
    _ip = utils._EngineIntScalarType(
        _vl_ip_py, {'fun_name': fun_name, 'entity': '_ip'},
    ).to_ctypes()
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _t = utils._EngineFloat64ArrayType(
        t, {'fun_name': fun_name, 'entity': 't'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _wt = utils._EngineFloat64ArrayType(
        wt, {'fun_name': fun_name, 'entity': 'wt'},
        exp_shape=(_n_py,),
        allow_none=True,
    ).to_ctypes()
    _dev = utils._EngineFloat64ScalarType.empty_ctypes()
    _idf = utils._EngineIntScalarType.empty_ctypes()
    _ip_py = utils._EngineIntScalarType.to_py(_ip)
    _b_cls = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_ip_py,),
        check_contents=False,
    )
    _b = _b_cls.to_ctypes(can_reference_contents=False)
    _irank = utils._EngineIntScalarType.empty_ctypes()
    _se_cls = utils._EngineFloat64ArrayType(
        se, {'fun_name': fun_name, 'entity': 'se'},
        exp_shape=(_ip_py,),
        check_contents=False,
    )
    _se = _se_cls.to_ctypes(can_reference_contents=False)
    _cov_cls = utils._EngineFloat64ArrayType(
        cov, {'fun_name': fun_name, 'entity': 'cov'},
        exp_shape=(_ip_py*(_ip_py+1)//2,),
        check_contents=False,
    )
    _cov = _cov_cls.to_ctypes(can_reference_contents=False)
    _ldv_py = utils._EngineFloat64ArrayType.get_nd(v, {'fun_name': fun_name, 'entity': 'v'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldv = utils._EngineIntScalarType(
        _ldv_py, {'fun_name': fun_name, 'entity': '_ldv'},
    ).to_ctypes()
    _v_cls = utils._EngineFloat64ArrayType(
        v, {'fun_name': fun_name, 'entity': 'v'},
        exp_shape=(_n_py, _ip_py+7),
        mutual_sorder=_m_sorder,
    )
    _v = _v_cls.to_ctypes()
    _tol = utils._EngineFloat64ScalarType(
        tol, {'fun_name': fun_name, 'entity': 'tol'},
    ).to_ctypes()
    _maxit = utils._EngineIntScalarType(
        maxit, {'fun_name': fun_name, 'entity': 'maxit'},
    ).to_ctypes()
    _iprint = utils._EngineIntScalarType(
        iprint, {'fun_name': fun_name, 'entity': 'iprint'},
    ).to_ctypes()
    _nout = utils._EngineIntScalarType.c_type(utils.FileObjManager.advunit)
    _eps = utils._EngineFloat64ScalarType(
        eps, {'fun_name': fun_name, 'entity': 'eps'},
    ).to_ctypes()
    _wk = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        utils._UNUSED_PRINT_REC,
        _print_rech,
        _sorder,
        _link,
        _mean,
        _offset,
        _weight,
        _n,
        _x,
        _ldx,
        _m,
        _isx,
        _ip,
        _y,
        _t,
        _wt,
        _dev,
        _idf,
        _b,
        _irank,
        _se,
        _cov,
        _v,
        _ldv,
        _tol,
        _maxit,
        _iprint,
        _nout,
        _eps,
        _wk,
        _errbuf,
        _ierr,
        _link._length_,
        _mean._length_,
        _offset._length_,
        _weight._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99978:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ip >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99984:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, eps = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: eps >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99985:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, tol = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: tol >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99986:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, maxit = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: maxit >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99987:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, offset = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: offset = \'Y\' or \'N\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, link = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: link = \'G\', \'P\' or \'C\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, weight = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: weight = \'W\' or \'U\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mean = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mean = \'M\' or \'Z\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, wt[{:s}] < 0.0.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: wt[i-1] >= 0.0, for i = 1,2,...,n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99977:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, isx[{:s}] < 0.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: isx[j-1] >= 0.0, for j = 1,2,...,m.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ip incompatible with number of nonzero values of isx:',
                    'ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Number of requested x-variables greater than n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99988:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, t[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: t[i-1] >= 0, for i = 1,2,...,n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, y[{:s}] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    't[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: 0.0 <= y[i-1] <= t[i-1], for i = 1,2,...,n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99982:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'A fitted value is at the boundary, i.e., 0.0 or 1.0.',
                    'This may occur if there are y values of 0.0',
                    'or t and the model is too complex for the data.',
                    'The model should be reformulated with, perhaps, some observations dropped.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 99981:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'SVD solution failed to converge.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 99980:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The iterative weighted least squares has failed to converge in',
                    'maxit (or default 10) iterations.',
                    'The value of maxit could be increased but it may be',
                    'advantageous to examine the convergence using the iprint',
                    'option. This may indicate that the convergence is slow because the',
                    'solution is at a boundary in which case it may be better to reformulate',
                    'the model.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 9 and _ifmt == 99983:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The rank of the model has changed during the weighted least squares',
                    'iterations. The estimate for beta returned',
                    'may be reasonable, but you should check how the deviance has changed',
                    'during iterations.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 10 and _ifmt == 99979:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Degrees of freedom for error are 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _b_cls.refresh_py(_b)
        _se_cls.refresh_py(_se)
        _cov_cls.refresh_py(_cov)
        _v_cls.refresh_py(_v)
        _dev_py = utils._EngineFloat64ScalarType.to_py(_dev)
        _idf_py = utils._EngineIntScalarType.to_py(_idf)
        _irank_py = utils._EngineIntScalarType.to_py(_irank)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['dev', 'idf', 'irank'],
        )(
            _dev_py,
            _idf_py,
            _irank_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def glm_poisson(link, mean, offset, n, x, m, isx, y, wt, a, b, se, cov, v, tol, maxit, iprint, eps, io_manager=None):
    u"""
    Fits a generalized linear model with Poisson errors.

    ``glm_poisson`` fits a generalized linear model with Poisson errors.

    For full information please refer to the NAG Library document for
    g02gc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02gcf.html

    Parameters
    ----------
    link : str, length 1
        Indicates which link function is to be used.

        `link` = 'E'

            An exponent link is used.

        `link` = 'I'

            An identity link is used.

        `link` = 'L'

            A log link is used.

        `link` = 'S'

            A square root link is used.

        `link` = 'R'

            A reciprocal link is used.

    mean : str, length 1
        Indicates if a mean term is to be included.

        `mean` = 'M'

            A mean term, intercept, will be included in the model.

        `mean` = 'Z'

            The model will pass through the origin, zero-point.

    offset : str, length 1
        Indicates if an offset is required.

        `offset` = 'Y'

            An offset is required and the offsets must be supplied in
            the seventh column of `v`.

        `offset` = 'N'

            No offset is required.

    n : int
        n, the number of observations.

    x : float, shape (`n`, `m`)
        The matrix of all possible independent variables. `x`[i-1,j-1]
        must contain the ijth element of `x`, for j = 1,2,...,`m`, for i
        = 1,2,...,n.

    m : int
        m, the total number of independent variables.

    isx : int, shape (`m`)
        Indicates which independent variables are to be included in the
        model.

        If `isx`[j-1] > 0, the variable contained in the jth column of
        `x` is included in the regression model.

    y : float, shape (`n`)
        y, observations on the dependent variable.

    wt : None or float, shape (`n`)
        If provided `wt` must contain the weights to be used with the
        model.

        If `wt`[i-1] = 0.0, the ith observation is not included in the
        model, in which case the effective number of observations is the
        number of observations with nonzero weights.

        If `wt` is not provided the effective number of observations is
        n.

    a : float
        If `link` = 'E', `a` must contain the power of the exponential.

        If `link` != 'E', `a` is not referenced.

    b : float, shape (ip), modified in place
        `On exit`: the estimates of the parameters of the generalized
        linear model, beta^.

        If `mean` = 'M', the first element of `b` will contain the
        estimate of the mean parameter and `b`[i] will contain the
        coefficient of the variable contained in column j of `x`, where
        `isx`[j-1] is the ith positive value in the array `isx`.

        If `mean` = 'Z', `b`[i-1] will contain the coefficient of the
        variable contained in column j of `x`, where `isx`[j-1] is the
        ith positive value in the array `isx`.

    se : float, shape (ip), modified in place
        `On exit`: the standard errors of the linear parameters.

        `se`[i-1] contains the standard error of the parameter estimate
        in `b`[i-1], for i = 1,2,...,ip.

    cov : float, shape (ip*(ip+1)/2), modified in place
        `On exit`: the upper triangular part of the variance-covariance
        matrix of the ip parameter estimates given in `b`. They are
        stored packed by column, i.e., the covariance between the
        parameter estimate given in `b`[i-1] and the parameter estimate
        given in `b`[j-1], j >= i, is stored in `cov`[(j*(j-1)/2+i)-1].

    v : float, shape (`n`, ip+7), modified in place
        `On entry`: if `offset` = 'N', `v` need not be set.

        If `offset` = 'Y', `v`[i-1,6], for i = 1,2,...,n must contain
        the offset values o_i.

        All other values need not be set.

        `On exit`: auxiliary information on the fitted model.

        [table omitted]

        If the model is not of full rank, i.e., `irank` < ip, the first
        ip rows of columns 8 to ip+7 contain the P* matrix.

    tol : float
        Indicates the accuracy required for the fit of the model.

        The iterative weighted least squares procedure is deemed to have
        converged if the absolute change in deviance between iterations
        is less than `tol`*(1.0+Current Deviance).

        This is approximately an absolute precision if the deviance is
        small and a relative precision if the deviance is large.

        If 0.0 <= `tol` < machine precision, the function will use
        10*machine precision instead.

    maxit : int
        The maximum number of iterations for the iterative weighted
        least squares.

        If `maxit` = 0, a default value of 10 is used.

    iprint : int
        Indicates if the printing of information on the iterations is
        required.

        `iprint` <= 0

            There is no printing.

        `iprint` > 0

            Every `iprint` iteration, the following are printed:

                the deviance;

                the current estimates;

                and if the weighted least squares equations are
                singular, then this is indicated.

        When printing occurs the output is directed to the file object
        associated with the advisory I/O unit (see
        :class:`~naginterfaces.base.utils.FileObjManager`).

    eps : float
        The value of `eps` is used to decide if the independent
        variables are of full rank and, if not, what is the rank of the
        independent variables. The smaller the value of `eps` the
        stricter the criterion for selecting the singular value
        decomposition.

        If 0.0 <= `eps` < machine precision, the function will use
        machine precision instead.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    dev : float
        The deviance for the fitted model.

    idf : int
        The degrees of freedom asociated with the deviance for the
        fitted model.

    irank : int
        The rank of the independent variables.

        If the model is of full rank, `irank` = ip.

        If the model is not of full rank, `irank` is an estimate of the
        rank of the independent variables. `irank` is calculated as the
        number of singular values greater that `eps`*(largest singular
        value).

        It is possible for the SVD to be carried out but for `irank` to
        be returned as ip.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, ip = *<value>*.

            Constraint: ip >= 1.

        (`errno` 1)
            On entry, `eps` = *<value>*.

            Constraint: `eps` >= 0.0.

        (`errno` 1)
            On entry, `tol` = *<value>*.

            Constraint: `tol` >= 0.0.

        (`errno` 1)
            On entry, `maxit` = *<value>*.

            Constraint: `maxit` >= 0.

        (`errno` 1)
            On entry, `offset` = *<value>*.

            Constraint: `offset` = 'Y' or 'N'.

        (`errno` 1)
            On entry, `link` = *<value>*.

            Constraint: `link` = 'E', 'I', 'L', 'S' or 'R'.

        (`errno` 1)
            On entry, `a` = 0.0 and `link` = 'E'.

            Constraint: if `link` = 'E', `a` != 0.0.

        (`errno` 1)
            On entry, weight = *<value>*.

            Constraint: weight = 'W' or 'U'.

        (`errno` 1)
            On entry, `mean` = *<value>*.

            Constraint: `mean` = 'M' or 'Z'.

        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 2.

        (`errno` 2)
            On entry, `wt`[*<value>*] < 0.0.

            Constraint: `wt`[i-1] >= 0.0, for i = 1,2,...,n.

        (`errno` 3)
            On entry, `isx`[*<value>*] < 0.

            Constraint: `isx`[j-1] >= 0.0, for j = 1,2,...,`m`.

        (`errno` 3)
            On entry, ip incompatible with number of nonzero values of
            `isx`: ip = *<value>*.

        (`errno` 3)
            Number of requested x-variables greater than `n`.

        (`errno` 4)
            On entry, `y`[*<value>*] = *<value>*.

            Constraint: `y`[i-1] <= 0.0, for i = 1,2,...,n.

        (`errno` 5)
            A fitted value is at the boundary, i.e., mu^= 0.0. This may
            occur if there are y values of 0.0 and the model is too
            complex for the data. The model should be reformulated with,
            perhaps, some observations dropped.

        (`errno` 6)
            SVD solution failed to converge.

        (`errno` 7)
            The iterative weighted least squares has failed to converge
            in `maxit` (or default 10) iterations. The value of `maxit`
            could be increased but it may be advantageous to examine the
            convergence using the `iprint` option. This may indicate
            that the convergence is slow because the solution is at a
            boundary in which case it may be better to reformulate the
            model.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 8)
            The rank of the model has changed during the weighted least
            squares iterations. The estimate for beta returned may be
            reasonable, but you should check how the deviance has
            changed during iterations.

        (`errno` 9)
            The degrees of freedom for error are 0. A saturated model
            has been fitted.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Cook, R D and Weisberg, S, 1982, `Residuals and Influence in
    Regression`, Chapman and Hall

    McCullagh, P and Nelder, J A, 1983, `Generalized Linear Models`,
    Chapman and Hall

    Plackett, R L, 1974, `The Analysis of Categorical Data`, Griffin

    See Also
    --------
    :meth:`naginterfaces.library.examples.surviv.coxmodel_ex.main`
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.glm_poisson'

    utils.FileObjManager._check_type_compat(
        io_manager, {'fun_name': fun_name, 'entity': 'io_manager'}
    )

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    def cb_print_rech_py(_, en_data, nout, rec, ifail, rec_len):
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data, utils._EngineIntScalarType.to_py(nout.contents), rec[:rec_len], io_manager, excs, ifail.contents,
                fun_name=fun_name,
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02gcft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._PRINT_REC_FUNCTYPE,
        utils._PRINT_RECH_FUNCTYPE,
        utils._BYREF_INT, # Input, sorder.
        utils._BYREF_STRING, # Input, link(len=1).
        utils._BYREF_STRING, # Input, mean(len=1).
        utils._BYREF_STRING, # Input, offset(len=1).
        utils._BYREF_STRING, # Input, weight(len=1).
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[ldx, *].
        utils._BYREF_INT, # Input, ldx.
        utils._BYREF_INT, # Input, m.
        utils._EngineIntArrayType.c_type, # Input, isx[m].
        utils._BYREF_INT, # Input, ip.
        utils._EngineFloat64ArrayType.c_type, # Input, y[n].
        utils._EngineFloat64ArrayType.c_type, # Input, wt[n].
        utils._BYREF_FLOAT64, # Input, a.
        utils._BYREF_FLOAT64, # Output, dev.
        utils._BYREF_INT, # Output, idf.
        utils._EngineFloat64ArrayType.c_type, # Output, b[ip].
        utils._BYREF_INT, # Output, irank.
        utils._EngineFloat64ArrayType.c_type, # Output, se[ip].
        utils._EngineFloat64ArrayType.c_type, # Output, cov[ip*(ip+1)//2].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, v[ldv, ip+7].
        utils._BYREF_INT, # Input, ldv.
        utils._BYREF_FLOAT64, # Input, tol.
        utils._BYREF_INT, # Input, maxit.
        utils._BYREF_INT, # Input, iprint.
        utils._BYREF_INT, # Input, nout.
        utils._BYREF_FLOAT64, # Input, eps.
        utils._EngineFloat64ArrayType.c_type, # Workspace, wk.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of link.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of mean.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of offset.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of weight.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((x, 'x'), (v, 'v')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _print_rech = utils._PRINT_RECH_FUNCTYPE(cb_print_rech_py)
    _sorder = utils._EngineIntScalarType.c_type(_nag_enum_consts.nag_col_major)
    _link = utils._EngineCharScalarType(
        link, {'fun_name': fun_name, 'entity': 'link'}, exp_chrlen=1,
    ).to_ctypes()
    _mean = utils._EngineCharScalarType(
        mean, {'fun_name': fun_name, 'entity': 'mean'}, exp_chrlen=1,
    ).to_ctypes()
    _offset = utils._EngineCharScalarType(
        offset, {'fun_name': fun_name, 'entity': 'offset'}, exp_chrlen=1,
    ).to_ctypes()
    if wt is None:
        _weight = utils._EngineCharScalarType(
            'U', {'fun_name': fun_name, 'entity': '_weight'}, exp_chrlen=1,
        ).to_ctypes()
    else:
        _weight = utils._EngineCharScalarType(
            'W', {'fun_name': fun_name, 'entity': '_weight'}, exp_chrlen=1,
        ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _ldx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _sorder_py = utils._EngineIntScalarType.to_py(_sorder)
    _m_py = utils._EngineIntScalarType.to_py(_m)
    if _sorder_py == 1:
        _x_dim2 = _m_py
    elif _sorder_py == 2:
        _x_dim2 = _n_py
    else:
        _x_dim2 = 0
    exp_shape_x = (_n_py, _x_dim2)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=exp_shape_x,
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _isx = utils._EngineIntArrayType(
        isx, {'fun_name': fun_name, 'entity': 'isx'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _vl_p_py = 0
    for _vl_i_py in range(1, _m_py+1):
        if utils._EngineIntArrayType.get_element(isx, {'fun_name': fun_name, 'entity': 'isx'}, (_vl_i_py-1,)) > 0:
            _vl_p_py = _vl_p_py + 1
    _mean_py = utils._EngineCharScalarType.to_py(_mean)
    if _mean_py in ('M', 'm'):
        _vl_ip_py = _vl_p_py+1
    else:
        _vl_ip_py = _vl_p_py
    _ip = utils._EngineIntScalarType(
        _vl_ip_py, {'fun_name': fun_name, 'entity': '_ip'},
    ).to_ctypes()
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _wt = utils._EngineFloat64ArrayType(
        wt, {'fun_name': fun_name, 'entity': 'wt'},
        exp_shape=(_n_py,),
        allow_none=True,
    ).to_ctypes()
    _a = utils._EngineFloat64ScalarType(
        a, {'fun_name': fun_name, 'entity': 'a'},
    ).to_ctypes()
    _dev = utils._EngineFloat64ScalarType.empty_ctypes()
    _idf = utils._EngineIntScalarType.empty_ctypes()
    _ip_py = utils._EngineIntScalarType.to_py(_ip)
    _b_cls = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_ip_py,),
        check_contents=False,
    )
    _b = _b_cls.to_ctypes(can_reference_contents=False)
    _irank = utils._EngineIntScalarType.empty_ctypes()
    _se_cls = utils._EngineFloat64ArrayType(
        se, {'fun_name': fun_name, 'entity': 'se'},
        exp_shape=(_ip_py,),
        check_contents=False,
    )
    _se = _se_cls.to_ctypes(can_reference_contents=False)
    _cov_cls = utils._EngineFloat64ArrayType(
        cov, {'fun_name': fun_name, 'entity': 'cov'},
        exp_shape=(_ip_py*(_ip_py+1)//2,),
        check_contents=False,
    )
    _cov = _cov_cls.to_ctypes(can_reference_contents=False)
    _ldv_py = utils._EngineFloat64ArrayType.get_nd(v, {'fun_name': fun_name, 'entity': 'v'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldv = utils._EngineIntScalarType(
        _ldv_py, {'fun_name': fun_name, 'entity': '_ldv'},
    ).to_ctypes()
    _v_cls = utils._EngineFloat64ArrayType(
        v, {'fun_name': fun_name, 'entity': 'v'},
        exp_shape=(_n_py, _ip_py+7),
        mutual_sorder=_m_sorder,
    )
    _v = _v_cls.to_ctypes()
    _tol = utils._EngineFloat64ScalarType(
        tol, {'fun_name': fun_name, 'entity': 'tol'},
    ).to_ctypes()
    _maxit = utils._EngineIntScalarType(
        maxit, {'fun_name': fun_name, 'entity': 'maxit'},
    ).to_ctypes()
    _iprint = utils._EngineIntScalarType(
        iprint, {'fun_name': fun_name, 'entity': 'iprint'},
    ).to_ctypes()
    _nout = utils._EngineIntScalarType.c_type(utils.FileObjManager.advunit)
    _eps = utils._EngineFloat64ScalarType(
        eps, {'fun_name': fun_name, 'entity': 'eps'},
    ).to_ctypes()
    _wk = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        utils._UNUSED_PRINT_REC,
        _print_rech,
        _sorder,
        _link,
        _mean,
        _offset,
        _weight,
        _n,
        _x,
        _ldx,
        _m,
        _isx,
        _ip,
        _y,
        _wt,
        _a,
        _dev,
        _idf,
        _b,
        _irank,
        _se,
        _cov,
        _v,
        _ldv,
        _tol,
        _maxit,
        _iprint,
        _nout,
        _eps,
        _wk,
        _errbuf,
        _ierr,
        _link._length_,
        _mean._length_,
        _offset._length_,
        _weight._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99978:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ip >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99984:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, eps = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: eps >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99985:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, tol = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: tol >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99986:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, maxit = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: maxit >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99987:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, offset = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: offset = \'Y\' or \'N\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, link = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: link = \'E\', \'I\', \'L\', \'S\' or \'R\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a = 0.0 and link = \'E\'.',
                    'Constraint: if link = \'E\', a != 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, weight = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: weight = \'W\' or \'U\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mean = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mean = \'M\' or \'Z\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, wt[{:s}] < 0.0.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: wt[i-1] >= 0.0, for i = 1,2,...,n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99977:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, isx[{:s}] < 0.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: isx[j-1] >= 0.0, for j = 1,2,...,m.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ip incompatible with number of nonzero values of isx:',
                    'ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Number of requested x-variables greater than n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, y[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: y[i-1] <= 0.0, for i = 1,2,...,n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99982:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'A fitted value is at the boundary, i.e., mu^= 0.0.',
                    'This may occur if there are y values of 0.0',
                    'and the model is too complex for the data. The model should be',
                    'reformulated with, perhaps, some observations dropped.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99981:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'SVD solution failed to converge.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 99980:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The iterative weighted least squares has failed to converge in',
                    'maxit (or default 10) iterations.',
                    'The value of maxit could be increased but it may be',
                    'advantageous to examine the convergence using the iprint',
                    'option. This may indicate that the convergence is slow because the',
                    'solution is at a boundary in which case it may be better to reformulate',
                    'the model.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 99983:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The rank of the model has changed during the weighted least squares',
                    'iterations. The estimate for beta returned',
                    'may be reasonable, but you should check how the deviance has changed',
                    'during iterations.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 9 and _ifmt == 99979:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The degrees of freedom for error are 0. A saturated model has been fitted.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _b_cls.refresh_py(_b)
        _se_cls.refresh_py(_se)
        _cov_cls.refresh_py(_cov)
        _v_cls.refresh_py(_v)
        _dev_py = utils._EngineFloat64ScalarType.to_py(_dev)
        _idf_py = utils._EngineIntScalarType.to_py(_idf)
        _irank_py = utils._EngineIntScalarType.to_py(_irank)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['dev', 'idf', 'irank'],
        )(
            _dev_py,
            _idf_py,
            _irank_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def glm_gamma(link, mean, offset, n, x, m, isx, y, wt, s, a, b, se, cov, v, tol, maxit, iprint, eps, io_manager=None):
    u"""
    Fits a generalized linear model with gamma errors.

    ``glm_gamma`` fits a generalized linear model with gamma errors.

    For full information please refer to the NAG Library document for
    g02gd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02gdf.html

    Parameters
    ----------
    link : str, length 1
        Indicates which link function is to be used.

        `link` = 'E'

            An exponential link is used.

        `link` = 'I'

            An identity link is used.

        `link` = 'L'

            A log link is used.

        `link` = 'S'

            A square root link is used.

        `link` = 'R'

            A reciprocal link is used.

    mean : str, length 1
        Indicates if a mean term is to be included.

        `mean` = 'M'

            A mean term, intercept, will be included in the model.

        `mean` = 'Z'

            The model will pass through the origin, zero-point.

    offset : str, length 1
        Indicates if an offset is required.

        `offset` = 'Y'

            An offset is required and the offsets must be supplied in
            the seventh column of `v`.

        `offset` = 'N'

            No offset is required.

    n : int
        n, the number of observations.

    x : float, shape (`n`, `m`)
        `x`[i-1,j-1] must contain the ith observation for the jth
        independent variable, for j = 1,2,...,`m`, for i = 1,2,...,`n`.

    m : int
        m, the total number of independent variables.

    isx : int, shape (`m`)
        Indicates which independent variables are to be included in the
        model.

        If `isx`[j-1] > 0, the variable contained in the jth column of
        `x` is included in the regression model.

    y : float, shape (`n`)
        y, the dependent variable.

    wt : None or float, shape (`n`)
        If provided `wt` must contain the weights to be used with the
        model.

        If `wt`[i-1] = 0.0, the ith observation is not included in the
        model, in which case the effective number of observations is the
        number of observations with nonzero weights.

        If `wt` is not provided the effective number of observations is
        n.

    s : float
        The scale parameter for the gamma model, nu^-1.

        `s` = 0.0

            The scale parameter is estimated with the function using the
            formula described in [Description].

    a : float
        If `link` = 'E', `a` must contain the power of the exponential.

        If `link` != 'E', `a` is not referenced.

    b : float, shape (ip), modified in place
        `On exit`: the estimates of the parameters of the generalized
        linear model, beta^.

        If `mean` = 'M', the first element of `b` will contain the
        estimate of the mean parameter and `b`[i] will contain the
        coefficient of the variable contained in column j of `x`, where
        `isx`[j-1] is the ith positive value in the array `isx`.

        If `mean` = 'Z', `b`[i-1] will contain the coefficient of the
        variable contained in column j of `x`, where `isx`[j-1] is the
        ith positive value in the array `isx`.

    se : float, shape (ip), modified in place
        `On exit`: the standard errors of the linear parameters.

        `se`[i-1] contains the standard error of the parameter estimate
        in `b`[i-1], for i = 1,2,...,ip.

    cov : float, shape (ip*(ip+1)/2), modified in place
        `On exit`: the upper triangular part of the variance-covariance
        matrix of the ip parameter estimates given in `b`. They are
        stored in packed form by column, i.e., the covariance between
        the parameter estimate given in `b`[i-1] and the parameter
        estimate given in `b`[j-1], j >= i, is stored in
        `cov`[(j*(j-1)/2+i)-1].

    v : float, shape (`n`, ip+7), modified in place
        `On entry`: if `offset` = 'Y', `v`[i-1,6], for i = 1,2,...,n,
        must contain the offset values o_i, otherwise `v` need not be
        set and no offset is used.

        `On exit`: auxiliary information on the fitted model.

        [table omitted]

        If the model is not of full rank, i.e., `irank` < ip, the first
        ip rows of columns 8 to ip+7 contain the P* matrix.

    tol : float
        Indicates the accuracy required for the fit of the model.

        The iterative weighted least squares procedure is deemed to have
        converged if the absolute change in deviance between iterations
        is less than `tol`*(1.0+Current Deviance).

        This is approximately an absolute precision if the deviance is
        small and a relative precision if the deviance is large.

        If 0.0 <= `tol` < machine precision then the function will use
        10*machine precision instead.

    maxit : int
        The maximum number of iterations for the iterative weighted
        least squares.

        `maxit` = 0

            A default value of 10 is used.

    iprint : int
        Indicates if the printing of information on the iterations is
        required.

        `iprint` <= 0

            There is no printing.

        `iprint` > 0

            Every `iprint` iteration, the following are printed:

                the deviance;

                the current estimates;

                and if the weighted least squares equations are
                singular, then this is indicated.

        When printing occurs the output is directed to the file object
        associated with the advisory I/O unit (see
        :class:`~naginterfaces.base.utils.FileObjManager`).

    eps : float
        The value of `eps` is used to decide if the independent
        variables are of full rank and, if not, what is the rank of the
        independent variables. The smaller the value of `eps` the
        stricter the criterion for selecting the singular value
        decomposition.

        If 0.0 <= `eps` < machine precision, the function will use
        machine precision instead.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    s : float
        If on input `s` = 0.0, `s` contains the estimated value of the
        scale parameter, nu^^(-1).

        If on input `s` != 0.0, `s` is unchanged on exit.

    dev : float
        The adjusted deviance for the fitted model.

    idf : int
        The degrees of freedom asociated with the deviance for the
        fitted model.

    irank : int
        The rank of the independent variables.

        If the model is of full rank, `irank` = ip.

        If the model is not of full rank, `irank` is an estimate of the
        rank of the independent variables. `irank` is calculated as the
        number of singular values greater that `eps`*(largest singular
        value).

        It is possible for the SVD to be carried out but for `irank` to
        be returned as ip.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, ip = *<value>*.

            Constraint: ip >= 1.

        (`errno` 1)
            On entry, `eps` = *<value>*.

            Constraint: `eps` >= 0.0.

        (`errno` 1)
            On entry, `tol` = *<value>*.

            Constraint: `tol` >= 0.0.

        (`errno` 1)
            On entry, `maxit` = *<value>*.

            Constraint: `maxit` >= 0.

        (`errno` 1)
            On entry, `offset` = *<value>*.

            Constraint: `offset` = 'Y' or 'N'.

        (`errno` 1)
            On entry, `link` = *<value>*.

            Constraint: `link` = 'E', 'I', 'L', 'S' or 'R'.

        (`errno` 1)
            On entry, `a` = 0.0 and `link` = 'E'.

            Constraint: if `link` = 'E', `a` != 0.0.

        (`errno` 1)
            On entry, `s` = *<value>*.

            Constraint: `s` >= 0.0.

        (`errno` 1)
            On entry, weight = *<value>*.

            Constraint: weight = 'W' or 'U'.

        (`errno` 1)
            On entry, `mean` = *<value>*.

            Constraint: `mean` = 'M' or 'Z'.

        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 2.

        (`errno` 2)
            On entry, `wt`[*<value>*] < 0.0.

            Constraint: `wt`[i-1] >= 0.0, for i = 1,2,...,n.

        (`errno` 3)
            On entry, `isx`[*<value>*] < 0.

            Constraint: `isx`[j-1] >= 0.0, for j = 1,2,...,`m`.

        (`errno` 3)
            On entry, ip incompatible with number of nonzero values of
            `isx`: ip = *<value>*.

        (`errno` 3)
            Number of requested x-variables greater than `n`.

        (`errno` 4)
            On entry, `y`[*<value>*] = *<value>*.

            Constraint: `y`[i-1] <= 0.0, for i = 1,2,...,n.

        (`errno` 5)
            A fitted value is at the boundary, i.e., mu^= 0.0. This may
            occur if there are y values of 0.0 and the model is too
            complex for the data. The model should be reformulated with,
            perhaps, some observations dropped.

        (`errno` 6)
            SVD solution failed to converge.

        (`errno` 7)
            The iterative weighted least squares has failed to converge
            in `maxit` (or default 10) iterations. The value of `maxit`
            could be increased but it may be advantageous to examine the
            convergence using the `iprint` option. This may indicate
            that the convergence is slow because the solution is at a
            boundary in which case it may be better to reformulate the
            model.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 8)
            The rank of the model has changed during the weighted least
            squares iterations. The estimate for beta returned may be
            reasonable, but you should check how the deviance has
            changed during iterations.

        (`errno` 9)
            The degrees of freedom for error are 0. A saturated model
            has been fitted.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Cook, R D and Weisberg, S, 1982, `Residuals and Influence in
    Regression`, Chapman and Hall

    McCullagh, P and Nelder, J A, 1983, `Generalized Linear Models`,
    Chapman and Hall
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.glm_gamma'

    utils.FileObjManager._check_type_compat(
        io_manager, {'fun_name': fun_name, 'entity': 'io_manager'}
    )

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    def cb_print_rech_py(_, en_data, nout, rec, ifail, rec_len):
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data, utils._EngineIntScalarType.to_py(nout.contents), rec[:rec_len], io_manager, excs, ifail.contents,
                fun_name=fun_name,
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02gdft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._PRINT_REC_FUNCTYPE,
        utils._PRINT_RECH_FUNCTYPE,
        utils._BYREF_INT, # Input, sorder.
        utils._BYREF_STRING, # Input, link(len=1).
        utils._BYREF_STRING, # Input, mean(len=1).
        utils._BYREF_STRING, # Input, offset(len=1).
        utils._BYREF_STRING, # Input, weight(len=1).
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[ldx, *].
        utils._BYREF_INT, # Input, ldx.
        utils._BYREF_INT, # Input, m.
        utils._EngineIntArrayType.c_type, # Input, isx[m].
        utils._BYREF_INT, # Input, ip.
        utils._EngineFloat64ArrayType.c_type, # Input, y[n].
        utils._EngineFloat64ArrayType.c_type, # Input, wt[n].
        utils._BYREF_FLOAT64, # Input/Output, s.
        utils._BYREF_FLOAT64, # Input, a.
        utils._BYREF_FLOAT64, # Output, dev.
        utils._BYREF_INT, # Output, idf.
        utils._EngineFloat64ArrayType.c_type, # Output, b[ip].
        utils._BYREF_INT, # Output, irank.
        utils._EngineFloat64ArrayType.c_type, # Output, se[ip].
        utils._EngineFloat64ArrayType.c_type, # Output, cov[ip*(ip+1)//2].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, v[ldv, ip+7].
        utils._BYREF_INT, # Input, ldv.
        utils._BYREF_FLOAT64, # Input, tol.
        utils._BYREF_INT, # Input, maxit.
        utils._BYREF_INT, # Input, iprint.
        utils._BYREF_INT, # Input, nout.
        utils._BYREF_FLOAT64, # Input, eps.
        utils._EngineFloat64ArrayType.c_type, # Workspace, wk.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of link.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of mean.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of offset.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of weight.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((x, 'x'), (v, 'v')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _print_rech = utils._PRINT_RECH_FUNCTYPE(cb_print_rech_py)
    _sorder = utils._EngineIntScalarType.c_type(_nag_enum_consts.nag_col_major)
    _link = utils._EngineCharScalarType(
        link, {'fun_name': fun_name, 'entity': 'link'}, exp_chrlen=1,
    ).to_ctypes()
    _mean = utils._EngineCharScalarType(
        mean, {'fun_name': fun_name, 'entity': 'mean'}, exp_chrlen=1,
    ).to_ctypes()
    _offset = utils._EngineCharScalarType(
        offset, {'fun_name': fun_name, 'entity': 'offset'}, exp_chrlen=1,
    ).to_ctypes()
    if wt is None:
        _weight = utils._EngineCharScalarType(
            'U', {'fun_name': fun_name, 'entity': '_weight'}, exp_chrlen=1,
        ).to_ctypes()
    else:
        _weight = utils._EngineCharScalarType(
            'W', {'fun_name': fun_name, 'entity': '_weight'}, exp_chrlen=1,
        ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _ldx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _sorder_py = utils._EngineIntScalarType.to_py(_sorder)
    _m_py = utils._EngineIntScalarType.to_py(_m)
    if _sorder_py == 1:
        _x_dim2 = _m_py
    elif _sorder_py == 2:
        _x_dim2 = _n_py
    else:
        _x_dim2 = 0
    exp_shape_x = (_n_py, _x_dim2)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=exp_shape_x,
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _isx = utils._EngineIntArrayType(
        isx, {'fun_name': fun_name, 'entity': 'isx'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _vl_p_py = 0
    for _vl_i_py in range(1, _m_py+1):
        if utils._EngineIntArrayType.get_element(isx, {'fun_name': fun_name, 'entity': 'isx'}, (_vl_i_py-1,)) > 0:
            _vl_p_py = _vl_p_py + 1
    _mean_py = utils._EngineCharScalarType.to_py(_mean)
    if _mean_py in ('M', 'm'):
        _vl_ip_py = _vl_p_py+1
    else:
        _vl_ip_py = _vl_p_py
    _ip = utils._EngineIntScalarType(
        _vl_ip_py, {'fun_name': fun_name, 'entity': '_ip'},
    ).to_ctypes()
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _wt = utils._EngineFloat64ArrayType(
        wt, {'fun_name': fun_name, 'entity': 'wt'},
        exp_shape=(_n_py,),
        allow_none=True,
    ).to_ctypes()
    _s = utils._EngineFloat64ScalarType(
        s, {'fun_name': fun_name, 'entity': 's'},
    ).to_ctypes()
    _a = utils._EngineFloat64ScalarType(
        a, {'fun_name': fun_name, 'entity': 'a'},
    ).to_ctypes()
    _dev = utils._EngineFloat64ScalarType.empty_ctypes()
    _idf = utils._EngineIntScalarType.empty_ctypes()
    _ip_py = utils._EngineIntScalarType.to_py(_ip)
    _b_cls = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_ip_py,),
        check_contents=False,
    )
    _b = _b_cls.to_ctypes(can_reference_contents=False)
    _irank = utils._EngineIntScalarType.empty_ctypes()
    _se_cls = utils._EngineFloat64ArrayType(
        se, {'fun_name': fun_name, 'entity': 'se'},
        exp_shape=(_ip_py,),
        check_contents=False,
    )
    _se = _se_cls.to_ctypes(can_reference_contents=False)
    _cov_cls = utils._EngineFloat64ArrayType(
        cov, {'fun_name': fun_name, 'entity': 'cov'},
        exp_shape=(_ip_py*(_ip_py+1)//2,),
        check_contents=False,
    )
    _cov = _cov_cls.to_ctypes(can_reference_contents=False)
    _ldv_py = utils._EngineFloat64ArrayType.get_nd(v, {'fun_name': fun_name, 'entity': 'v'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldv = utils._EngineIntScalarType(
        _ldv_py, {'fun_name': fun_name, 'entity': '_ldv'},
    ).to_ctypes()
    _v_cls = utils._EngineFloat64ArrayType(
        v, {'fun_name': fun_name, 'entity': 'v'},
        exp_shape=(_n_py, _ip_py+7),
        mutual_sorder=_m_sorder,
    )
    _v = _v_cls.to_ctypes()
    _tol = utils._EngineFloat64ScalarType(
        tol, {'fun_name': fun_name, 'entity': 'tol'},
    ).to_ctypes()
    _maxit = utils._EngineIntScalarType(
        maxit, {'fun_name': fun_name, 'entity': 'maxit'},
    ).to_ctypes()
    _iprint = utils._EngineIntScalarType(
        iprint, {'fun_name': fun_name, 'entity': 'iprint'},
    ).to_ctypes()
    _nout = utils._EngineIntScalarType.c_type(utils.FileObjManager.advunit)
    _eps = utils._EngineFloat64ScalarType(
        eps, {'fun_name': fun_name, 'entity': 'eps'},
    ).to_ctypes()
    _wk = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        utils._UNUSED_PRINT_REC,
        _print_rech,
        _sorder,
        _link,
        _mean,
        _offset,
        _weight,
        _n,
        _x,
        _ldx,
        _m,
        _isx,
        _ip,
        _y,
        _wt,
        _s,
        _a,
        _dev,
        _idf,
        _b,
        _irank,
        _se,
        _cov,
        _v,
        _ldv,
        _tol,
        _maxit,
        _iprint,
        _nout,
        _eps,
        _wk,
        _errbuf,
        _ierr,
        _link._length_,
        _mean._length_,
        _offset._length_,
        _weight._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99977:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ip >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99983:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, eps = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: eps >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99984:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, tol = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: tol >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99985:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, maxit = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: maxit >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99986:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, offset = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: offset = \'Y\' or \'N\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, link = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: link = \'E\', \'I\', \'L\', \'S\' or \'R\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a = 0.0 and link = \'E\'.',
                    'Constraint: if link = \'E\', a != 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, s = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: s >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, weight = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: weight = \'W\' or \'U\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mean = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mean = \'M\' or \'Z\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, wt[{:s}] < 0.0.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: wt[i-1] >= 0.0, for i = 1,2,...,n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99976:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, isx[{:s}] < 0.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: isx[j-1] >= 0.0, for j = 1,2,...,m.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ip incompatible with number of nonzero values of isx:',
                    'ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Number of requested x-variables greater than n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99988:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, y[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: y[i-1] <= 0.0, for i = 1,2,...,n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99981:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'A fitted value is at the boundary, i.e., mu^= 0.0.',
                    'This may occur if there are y values of 0.0',
                    'and the model is too complex for the data. The model should be',
                    'reformulated with, perhaps, some observations dropped.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99980:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'SVD solution failed to converge.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 99979:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The iterative weighted least squares has failed to converge in',
                    'maxit (or default 10) iterations.',
                    'The value of maxit could be increased but it may be',
                    'advantageous to examine the convergence using the iprint',
                    'option. This may indicate that the convergence is slow because the',
                    'solution is at a boundary in which case it may be better to reformulate',
                    'the model.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 99982:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The rank of the model has changed during the weighted least squares',
                    'iterations. The estimate for beta returned',
                    'may be reasonable, but you should check how the deviance has changed',
                    'during iterations.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 9 and _ifmt == 99978:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The degrees of freedom for error are 0. A saturated model has been fitted.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _s_py = utils._EngineFloat64ScalarType.to_py(_s)
        _b_cls.refresh_py(_b)
        _se_cls.refresh_py(_se)
        _cov_cls.refresh_py(_cov)
        _v_cls.refresh_py(_v)
        _dev_py = utils._EngineFloat64ScalarType.to_py(_dev)
        _idf_py = utils._EngineIntScalarType.to_py(_idf)
        _irank_py = utils._EngineIntScalarType.to_py(_irank)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['s', 'dev', 'idf', 'irank'],
        )(
            _s_py,
            _dev_py,
            _idf_py,
            _irank_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def glm_constrain(ip, iconst, v, c, b, s, se, cov):
    u"""
    Estimates and standard errors of parameters of a general linear
    model for given constraints.

    ``glm_constrain`` calculates the estimates of the parameters of a
    generalized linear model for given constraints from the singular
    value decomposition results.

    For full information please refer to the NAG Library document for
    g02gk

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02gkf.html

    Parameters
    ----------
    ip : int
        p, the number of terms in the linear model.

    iconst : int
        The number of constraints to be imposed on the parameters, n_c.

    v : float, shape (`ip`, `ip`+7)
        The array `v` as returned by ``glm_normal``, ``glm_binomial``,
        ``glm_poisson`` or ``glm_gamma``.

    c : float, shape (`ip`, `iconst`)
        Contains the `iconst` constraints stored by column, i.e., the
        ith constraint is stored in the ith column of `c`.

    b : float, shape (`ip`), modified in place
        `On entry`: the parameter estimates computed by using the
        singular value decomposition, beta^_svd.

        `On exit`: the parameter estimates of the parameters with the
        constraints imposed, beta^_c.

    s : float
        The estimate of the scale parameter.

        For results from ``glm_normal`` and ``glm_gamma`` then `s` is
        the scale parameter for the model.

        For results from ``glm_binomial`` and ``glm_poisson`` then `s`
        should be set to 1.0.

    se : float, shape (`ip`), modified in place
        `On exit`: the standard error of the parameter estimates in `b`.

    cov : float, shape (`ip`*(`ip`+1)/2), modified in place
        `On exit`: the upper triangular part of the variance-covariance
        matrix of the `ip` parameter estimates given in `b`. They are
        stored packed by column, i.e., the covariance between the
        parameter estimate given in `b`[i-1] and the parameter estimate
        given in `b`[j-1], j >= i, is stored in `cov`[(j*(j-1)/2+i)-1].

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `iconst` = *<value>*.

            Constraint: `iconst` > 0.

        (`errno` 1)
            On entry, `s` = *<value>*.

            Constraint: `s` > 0.0.

        (`errno` 1)
            On entry, `iconst` = *<value>* and `ip` = *<value>*.

            Constraint: `iconst` < `ip`.

        (`errno` 1)
            On entry, `ip` = *<value>*.

            Constraint: `ip` >= 1.

        (`errno` 2)
            `c` does not give a model of full rank.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    McCullagh, P and Nelder, J A, 1983, `Generalized Linear Models`,
    Chapman and Hall

    Searle, S R, 1971, `Linear Models`, Wiley
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.glm_constrain'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02gkft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, ip.
        utils._BYREF_INT, # Input, iconst.
        utils._EngineFloat64ArrayType.c_type, # Input, v[ldv, ip+7].
        utils._BYREF_INT, # Input, ldv.
        utils._EngineFloat64ArrayType.c_type, # Input, c[ldc, iconst].
        utils._BYREF_INT, # Input, ldc.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, b[ip].
        utils._BYREF_FLOAT64, # Input, s.
        utils._EngineFloat64ArrayType.c_type, # Output, se[ip].
        utils._EngineFloat64ArrayType.c_type, # Output, cov[ip*(ip+1)//2].
        utils._EngineFloat64ArrayType.c_type, # Workspace, wk.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((v, 'v'), (c, 'c')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _ip = utils._EngineIntScalarType(
        ip, {'fun_name': fun_name, 'entity': 'ip'},
    ).to_ctypes()
    _iconst = utils._EngineIntScalarType(
        iconst, {'fun_name': fun_name, 'entity': 'iconst'},
    ).to_ctypes()
    _ldv_py = utils._EngineFloat64ArrayType.get_nd(v, {'fun_name': fun_name, 'entity': 'v'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldv = utils._EngineIntScalarType(
        _ldv_py, {'fun_name': fun_name, 'entity': '_ldv'},
    ).to_ctypes()
    _ip_py = utils._EngineIntScalarType.to_py(_ip)
    _v = utils._EngineFloat64ArrayType(
        v, {'fun_name': fun_name, 'entity': 'v'},
        exp_shape=(_ip_py, _ip_py+7),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _ldc_py = utils._EngineFloat64ArrayType.get_nd(c, {'fun_name': fun_name, 'entity': 'c'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldc = utils._EngineIntScalarType(
        _ldc_py, {'fun_name': fun_name, 'entity': '_ldc'},
    ).to_ctypes()
    _iconst_py = utils._EngineIntScalarType.to_py(_iconst)
    _c = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_ip_py, _iconst_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _b_cls = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_ip_py,),
    )
    _b = _b_cls.to_ctypes()
    _s = utils._EngineFloat64ScalarType(
        s, {'fun_name': fun_name, 'entity': 's'},
    ).to_ctypes()
    _se_cls = utils._EngineFloat64ArrayType(
        se, {'fun_name': fun_name, 'entity': 'se'},
        exp_shape=(_ip_py,),
        check_contents=False,
    )
    _se = _se_cls.to_ctypes(can_reference_contents=False)
    _cov_cls = utils._EngineFloat64ArrayType(
        cov, {'fun_name': fun_name, 'entity': 'cov'},
        exp_shape=(_ip_py*(_ip_py+1)//2,),
        check_contents=False,
    )
    _cov = _cov_cls.to_ctypes(can_reference_contents=False)
    _wk = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _ip,
        _iconst,
        _v,
        _ldv,
        _c,
        _ldc,
        _b,
        _s,
        _se,
        _cov,
        _wk,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iconst = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: iconst > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, s = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: s > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iconst = {:s} and ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: iconst < ip.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ip >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'c does not give a model of full rank.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _b_cls.refresh_py(_b)
        _se_cls.refresh_py(_se)
        _cov_cls.refresh_py(_cov)
        return_data = None
    _err_class.manage()
    return return_data

def glm_estfunc(ip, irank, b, cov, v, f, tol):
    u"""
    Computes estimable function of a generalized linear model and its
    standard error.

    ``glm_estfunc`` gives the estimate of an estimable function along
    with its standard error from the results from fitting a generalized
    linear model.

    For full information please refer to the NAG Library document for
    g02gn

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02gnf.html

    Parameters
    ----------
    ip : int
        p, the number of terms in the linear model.

    irank : int
        k, the rank of the dependent variables.

    b : float, shape (`ip`)
        The `ip` values of the estimates of the parameters of the model,
        beta^.

    cov : float, shape (`ip`*(`ip`+1)/2)
        The upper triangular part of the variance-covariance matrix of
        the `ip` parameter estimates given in `b`. They are stored
        packed by column, i.e., the covariance between the parameter
        estimate given in `b`[i-1] and the parameter estimate given in
        `b`[j-1], j >= i, is stored in `cov`[(j*(j-1)/2+i)-1].

    v : float, shape (`ip`, `ip`+7)
        As returned by ``glm_normal``, ``glm_binomial``, ``glm_poisson``
        and ``glm_gamma``.

    f : float, shape (`ip`)
        f, the linear function to be estimated.

    tol : float
        The tolerance value used in the check for estimability, eta.

        If `tol` <= 0.0 then sqrt(epsilon), where epsilon is the machine
        precision, is used instead.

    Returns
    -------
    est : bool
        Indicates if the function was estimable.

        `est` = True

            The function is estimable.

        `est` = False

            The function is not estimable and `stat`, `sestat` and `z`
            are not set.

    stat : float
        If `est` = True, `stat` contains the estimate of the function,
        f^T beta^

    sestat : float
        If `est` = True, `sestat` contains the standard error of the
        estimate of the function, se(F).

    z : float
        If `est` = True, `z` contains the z statistic for the test of
        the function being equal to zero.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `irank` = *<value>* and `ip` = *<value>*.

            Constraint: `irank` <= `ip`.

        (`errno` 1)
            On entry, `irank` = *<value>*.

            Constraint: `irank` >= 1.

        (`errno` 1)
            On entry, `ip` = *<value>*.

            Constraint: `ip` >= 1.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            `irank` = `ip`. In this case `est` is returned as true and
            all statistics are calculated.

        (`errno` 3)
            Standard error of statistic = 0.0; this may be due to
            rounding errors if the standard error is very small or due
            to mis-specified inputs `cov` and `f`.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    McCullagh, P and Nelder, J A, 1983, `Generalized Linear Models`,
    Chapman and Hall

    Searle, S R, 1971, `Linear Models`, Wiley
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.glm_estfunc'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02gnft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, ip.
        utils._BYREF_INT, # Input, irank.
        utils._EngineFloat64ArrayType.c_type, # Input, b[ip].
        utils._EngineFloat64ArrayType.c_type, # Input, cov[ip*(ip+1)//2].
        utils._EngineFloat64ArrayType.c_type, # Input, v[ldv, ip+7].
        utils._BYREF_INT, # Input, ldv.
        utils._EngineFloat64ArrayType.c_type, # Input, f[ip].
        utils._BYREF_LOGICAL, # Output, est.
        utils._BYREF_FLOAT64, # Output, stat.
        utils._BYREF_FLOAT64, # Output, sestat.
        utils._BYREF_FLOAT64, # Output, z.
        utils._BYREF_FLOAT64, # Input, tol.
        utils._EngineFloat64ArrayType.c_type, # Workspace, wk.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((v, 'v'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _ip = utils._EngineIntScalarType(
        ip, {'fun_name': fun_name, 'entity': 'ip'},
    ).to_ctypes()
    _irank = utils._EngineIntScalarType(
        irank, {'fun_name': fun_name, 'entity': 'irank'},
    ).to_ctypes()
    _ip_py = utils._EngineIntScalarType.to_py(_ip)
    _b = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_ip_py,),
    ).to_ctypes()
    _cov = utils._EngineFloat64ArrayType(
        cov, {'fun_name': fun_name, 'entity': 'cov'},
        exp_shape=(_ip_py*(_ip_py+1)//2,),
    ).to_ctypes()
    _ldv_py = utils._EngineFloat64ArrayType.get_nd(v, {'fun_name': fun_name, 'entity': 'v'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldv = utils._EngineIntScalarType(
        _ldv_py, {'fun_name': fun_name, 'entity': '_ldv'},
    ).to_ctypes()
    _v = utils._EngineFloat64ArrayType(
        v, {'fun_name': fun_name, 'entity': 'v'},
        exp_shape=(_ip_py, _ip_py+7),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _f = utils._EngineFloat64ArrayType(
        f, {'fun_name': fun_name, 'entity': 'f'},
        exp_shape=(_ip_py,),
    ).to_ctypes()
    _est = utils._EngineLogicalScalarType.empty_ctypes()
    _stat = utils._EngineFloat64ScalarType.empty_ctypes()
    _sestat = utils._EngineFloat64ScalarType.empty_ctypes()
    _z = utils._EngineFloat64ScalarType.empty_ctypes()
    _tol = utils._EngineFloat64ScalarType(
        tol, {'fun_name': fun_name, 'entity': 'tol'},
    ).to_ctypes()
    _wk = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _ip,
        _irank,
        _b,
        _cov,
        _v,
        _ldv,
        _f,
        _est,
        _stat,
        _sestat,
        _z,
        _tol,
        _wk,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, irank = {:s} and ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: irank <= ip.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, irank = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: irank >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ip >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'irank = ip. In this case est is returned as true and all',
                    'statistics are calculated.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99995:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Standard error of statistic = 0.0; this may be due',
                    'to rounding errors if the standard error is very small or due to',
                    'mis-specified inputs cov and f.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _est_py = utils._EngineLogicalScalarType.to_py(_est)
        _stat_py = utils._EngineFloat64ScalarType.to_py(_stat)
        _sestat_py = utils._EngineFloat64ScalarType.to_py(_sestat)
        _z_py = utils._EngineFloat64ScalarType.to_py(_z)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['est', 'stat', 'sestat', 'z'],
        )(
            _est_py,
            _stat_py,
            _sestat_py,
            _z_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def glm_predict(errfn, link, mean, n, x, m, isx, t, off, wt, s, a, b, cov, vfobs, eta, seeta, pred, sepred):
    u"""
    Computes a predicted value and its associated standard error based
    on a previously fitted generalized linear model.

    ``glm_predict`` allows prediction from a generalized linear model
    fit via ``glm_normal``, ``glm_binomial``, ``glm_poisson`` or
    ``glm_gamma``.

    For full information please refer to the NAG Library document for
    g02gp

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02gpf.html

    Parameters
    ----------
    errfn : str, length 1
        Indicates the distribution used to model the dependent variable,
        y.

        `errfn` = 'B'

            The binomial distribution is used.

        `errfn` = 'G'

            The gamma distribution is used.

        `errfn` = 'N'

            The Normal (Gaussian) distribution is used.

        `errfn` = 'P'

            The Poisson distribution is used.

    link : str, length 1
        Indicates which link function is to be used.

        `link` = 'C'

            A complementary log-log link is used.

        `link` = 'E'

            An exponent link is used.

        `link` = 'G'

            A logistic link is used.

        `link` = 'I'

            An identity link is used.

        `link` = 'L'

            A log link is used.

        `link` = 'P'

            A probit link is used.

        `link` = 'R'

            A reciprocal link is used.

        `link` = 'S'

            A square root link is used.

        Details on the functional form of the different links can be
        found in [the G02 Introduction].

    mean : str, length 1
        Indicates if a mean term is to be included.

        `mean` = 'M'

            A mean term, intercept, will be included in the model.

        `mean` = 'Z'

            The model will pass through the origin, zero-point.

    n : int
        n, the number of observations.

    x : float, shape (`n`, `m`)
        `x`[i-1,j-1] must contain the ith observation for the jth
        independent variable, for j = 1,2,...,`m`, for i = 1,2,...,`n`.

    m : int
        m, the total number of independent variables.

    isx : int, shape (`m`)
        Indicates which independent variables are to be included in the
        model.

        If `isx`[j-1] > 0, the variable contained in the jth column of
        `x` is included in the regression model.

    t : None or float, shape (`n`)
        If `errfn` = 'B', `t`[i-1] must contain the binomial
        denominator, t_i, for the ith observation.

        Otherwise `t` is not referenced and may be None.

    off : None or float, shape (`n`)
        If an offset is required, then `off`[i-1] must contain the value
        of the offset o_i, for the ith observation. Otherwise `off` must
        be None.

    wt : None or float, shape (`n`)
        If weighted estimates are required then `wt`[i-1] must contain
        the weight, omega_i for the ith observation. Otherwise `wt` must
        be supplied as None.

        If `wt`[i-1] = 0.0, the ith observation is not included in the
        model, in which case the effective number of observations is the
        number of observations with positive weights.

        If `wt` = None, the effective number of observations is n.

        If the variance of future observations is not included in the
        standard error of the predicted variable, `wt` is not
        referenced.

    s : float
        If `errfn` = 'N' or 'G' and `vfobs` = True, the scale parameter,
        phi.

        Otherwise `s` is not referenced and phi = 1.

    a : float
        If `link` = 'E', `a` must contain the power of the exponential.

        If `link` != 'E', `a` is not referenced.

    b : float, shape (ip)
        The model parameters, beta.

        If `mean` = 'M', `b`[0] must contain the mean parameter and
        `b`[i] the coefficient of the variable contained in the jth
        independent `x`, where `isx`[j-1] is the ith positive value in
        the array `isx`.

        If `mean` = 'Z', `b`[i-1] must contain the coefficient of the
        variable contained in the jth independent `x`, where `isx`[j-1]
        is the ith positive value in the array `isx`.

    cov : float, shape (ip*(ip+1)/2)
        The upper triangular part of the variance-covariance matrix, C,
        of the model parameters. This matrix should be supplied packed
        by column, i.e., the covariance between parameters beta_i and
        beta_j, that is the values stored in `b`[i-1] and `b`[j-1],
        should be supplied in `cov`[j*(j-1)/2+i-1], for j = i,...,ip,
        for i = 1,2,...,ip.

    vfobs : bool
        If `vfobs` = True, the variance of future observations is
        included in the standard error of the predicted variable (i.e.,
        I_fobs = 1), otherwise I_fobs = 0.

    eta : float, shape (`n`), modified in place
        `On exit`: the linear predictor, eta.

    seeta : float, shape (`n`), modified in place
        `On exit`: the standard error of the linear predictor, se(eta).

    pred : float, shape (`n`), modified in place
        `On exit`: the predicted value, y^.

    sepred : float, shape (`n`), modified in place
        `On exit`: the standard error of the predicted value, se(y^). If
        `pred`[i-1] could not be calculated, ``glm_predict`` returns
        `errno` = 22, and `sepred`[i-1] is set to -99.0.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `errfn` = *<value>*.

            Constraint: `errfn` = 'B', 'G', 'N' or 'P'.

        (`errno` 2)
            On entry, `link` = *<value>*.

            Constraint: if `errfn` = 'B', `link` = 'C', 'G' or 'P',

            otherwise, `link` = 'E', 'I', 'L', 'R' or 'S'.

        (`errno` 2)
            On entry, `errfn` = *<value>* and `link` = *<value>*.

            Constraint: if `errfn` = 'B', `link` = 'C', 'G' or 'P',

            otherwise, `link` = 'E', 'I', 'L', 'R' or 'S'.

        (`errno` 3)
            On entry, `mean` = *<value>*.

            Constraint: `mean` = 'M' or 'Z'.

        (`errno` 5)
            On entry, weight = *<value>*.

            Constraint: if `vfobs` = True, weight = 'U' or 'W'.

        (`errno` 6)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 9)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 10)
            On entry, `isx`[*<value>*] < 0.

            Constraint: `isx`[j-1] >= 0.0, for j = 1,2,...,`m`.

        (`errno` 11)
            On entry, ip = *<value>*.

            Constraint: ip > 0.

        (`errno` 12)
            On entry, `t`[*<value>*] = *<value>*.

            Constraint: `t`[i-1] >= 0.0, for all i.

        (`errno` 14)
            On entry, `wt`[*<value>*] = *<value>*.

            Constraint: `wt`[i-1] >= 0.0, for all i.

        (`errno` 15)
            On entry, `s` = *<value>*.

            Constraint: `s` > 0.0.

        (`errno` 16)
            On entry, `a` = 0.0.

            Constraint: if `link` = 'E', `a` != 0.0.

        (`errno` 18)
            On entry, `cov`[*<value>*] = *<value>*.

            Constraint: `cov`[i] >= 0.0 for at least one diagonal
            element.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 22)
            At least one predicted value could not be calculated as
            required. `sepred` is set to -99.0 for affected predicted
            values.

    References
    ----------
    McCullagh, P and Nelder, J A, 1983, `Generalized Linear Models`,
    Chapman and Hall

    See Also
    --------
    :meth:`naginterfaces.library.examples.correg.glm_binomial_ex.main`
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.glm_predict'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02gpft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, sorder.
        utils._BYREF_STRING, # Input, errfn(len=1).
        utils._BYREF_STRING, # Input, link(len=1).
        utils._BYREF_STRING, # Input, mean(len=1).
        utils._BYREF_STRING, # Input, offset(len=1).
        utils._BYREF_STRING, # Input, weight(len=1).
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[ldx, m].
        utils._BYREF_INT, # Input, ldx.
        utils._BYREF_INT, # Input, m.
        utils._EngineIntArrayType.c_type, # Input, isx[m].
        utils._BYREF_INT, # Input, ip.
        utils._BYREF_INT, # Input, lt.
        utils._EngineFloat64ArrayType.c_type, # Input, t[lt].
        utils._BYREF_INT, # Input, loff.
        utils._EngineFloat64ArrayType.c_type, # Input, off[loff].
        utils._BYREF_INT, # Input, lwt.
        utils._EngineFloat64ArrayType.c_type, # Input, wt[lwt].
        utils._BYREF_FLOAT64, # Input, s.
        utils._BYREF_FLOAT64, # Input, a.
        utils._EngineFloat64ArrayType.c_type, # Input, b[ip].
        utils._EngineFloat64ArrayType.c_type, # Input, cov[ip*(ip+1)//2].
        utils._BYREF_LOGICAL, # Input, vfobs.
        utils._EngineFloat64ArrayType.c_type, # Output, eta[n].
        utils._EngineFloat64ArrayType.c_type, # Output, seeta[n].
        utils._EngineFloat64ArrayType.c_type, # Output, pred[n].
        utils._EngineFloat64ArrayType.c_type, # Output, sepred[n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errfn.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of link.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of mean.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of offset.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of weight.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((x, 'x'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _sorder = utils._EngineIntScalarType.c_type(_nag_enum_consts.nag_col_major)
    _errfn = utils._EngineCharScalarType(
        errfn, {'fun_name': fun_name, 'entity': 'errfn'}, exp_chrlen=1,
    ).to_ctypes()
    _link = utils._EngineCharScalarType(
        link, {'fun_name': fun_name, 'entity': 'link'}, exp_chrlen=1,
    ).to_ctypes()
    _mean = utils._EngineCharScalarType(
        mean, {'fun_name': fun_name, 'entity': 'mean'}, exp_chrlen=1,
    ).to_ctypes()
    if off is None:
        _offset = utils._EngineCharScalarType(
            'N', {'fun_name': fun_name, 'entity': '_offset'}, exp_chrlen=1,
        ).to_ctypes()
    else:
        _offset = utils._EngineCharScalarType(
            'Y', {'fun_name': fun_name, 'entity': '_offset'}, exp_chrlen=1,
        ).to_ctypes()
    if wt is None:
        _weight = utils._EngineCharScalarType(
            'U', {'fun_name': fun_name, 'entity': '_weight'}, exp_chrlen=1,
        ).to_ctypes()
    else:
        _weight = utils._EngineCharScalarType(
            'W', {'fun_name': fun_name, 'entity': '_weight'}, exp_chrlen=1,
        ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _ldx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py, _m_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _isx = utils._EngineIntArrayType(
        isx, {'fun_name': fun_name, 'entity': 'isx'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _vl_p_py = 0
    for _vl_i_py in range(1, _m_py+1):
        if utils._EngineIntArrayType.get_element(isx, {'fun_name': fun_name, 'entity': 'isx'}, (_vl_i_py-1,)) > 0:
            _vl_p_py = _vl_p_py + 1
    _mean_py = utils._EngineCharScalarType.to_py(_mean)
    if _mean_py in ('M', 'm'):
        _vl_ip_py = _vl_p_py+1
    else:
        _vl_ip_py = _vl_p_py
    _ip = utils._EngineIntScalarType(
        _vl_ip_py, {'fun_name': fun_name, 'entity': '_ip'},
    ).to_ctypes()
    _lt_py = utils._EngineFloat64ArrayType.get_shape(t, {'fun_name': fun_name, 'entity': 't'}, exp_rank=1, allow_none=True)[0]
    _lt = utils._EngineIntScalarType(
        _lt_py, {'fun_name': fun_name, 'entity': '_lt'},
    ).to_ctypes()
    _errfn_py = utils._EngineCharScalarType.to_py(_errfn)
    if _errfn_py in ('B', 'b'):
        if t is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 't'},
                'must not be None when errfn in (\'B\', \'b\')',
            )
    _t = utils._EngineFloat64ArrayType(
        t, {'fun_name': fun_name, 'entity': 't'},
        exp_shape=(_n_py,),
        check_contents=_errfn_py in ('B', 'b'),
        allow_none=True,
    ).to_ctypes(can_reference_contents=_errfn_py in ('B', 'b'))
    _loff_py = utils._EngineFloat64ArrayType.get_shape(off, {'fun_name': fun_name, 'entity': 'off'}, exp_rank=1, allow_none=True)[0]
    _loff = utils._EngineIntScalarType(
        _loff_py, {'fun_name': fun_name, 'entity': '_loff'},
    ).to_ctypes()
    _off = utils._EngineFloat64ArrayType(
        off, {'fun_name': fun_name, 'entity': 'off'},
        exp_shape=(_n_py,),
        allow_none=True,
    ).to_ctypes()
    _lwt_py = utils._EngineFloat64ArrayType.get_shape(wt, {'fun_name': fun_name, 'entity': 'wt'}, exp_rank=1, allow_none=True)[0]
    _lwt = utils._EngineIntScalarType(
        _lwt_py, {'fun_name': fun_name, 'entity': '_lwt'},
    ).to_ctypes()
    _wt = utils._EngineFloat64ArrayType(
        wt, {'fun_name': fun_name, 'entity': 'wt'},
        exp_shape=(_n_py,),
        allow_none=True,
    ).to_ctypes()
    _s = utils._EngineFloat64ScalarType(
        s, {'fun_name': fun_name, 'entity': 's'},
    ).to_ctypes()
    _a = utils._EngineFloat64ScalarType(
        a, {'fun_name': fun_name, 'entity': 'a'},
    ).to_ctypes()
    _ip_py = utils._EngineIntScalarType.to_py(_ip)
    _b = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_ip_py,),
    ).to_ctypes()
    _cov = utils._EngineFloat64ArrayType(
        cov, {'fun_name': fun_name, 'entity': 'cov'},
        exp_shape=(_ip_py*(_ip_py+1)//2,),
    ).to_ctypes()
    _vfobs = utils._EngineLogicalScalarType(
        vfobs, {'fun_name': fun_name, 'entity': 'vfobs'},
    ).to_ctypes()
    _eta_cls = utils._EngineFloat64ArrayType(
        eta, {'fun_name': fun_name, 'entity': 'eta'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _eta = _eta_cls.to_ctypes(can_reference_contents=False)
    _seeta_cls = utils._EngineFloat64ArrayType(
        seeta, {'fun_name': fun_name, 'entity': 'seeta'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _seeta = _seeta_cls.to_ctypes(can_reference_contents=False)
    _pred_cls = utils._EngineFloat64ArrayType(
        pred, {'fun_name': fun_name, 'entity': 'pred'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _pred = _pred_cls.to_ctypes(can_reference_contents=False)
    _sepred_cls = utils._EngineFloat64ArrayType(
        sepred, {'fun_name': fun_name, 'entity': 'sepred'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _sepred = _sepred_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _sorder,
        _errfn,
        _link,
        _mean,
        _offset,
        _weight,
        _n,
        _x,
        _ldx,
        _m,
        _isx,
        _ip,
        _lt,
        _t,
        _loff,
        _off,
        _lwt,
        _wt,
        _s,
        _a,
        _b,
        _cov,
        _vfobs,
        _eta,
        _seeta,
        _pred,
        _sepred,
        _errbuf,
        _ierr,
        _errfn._length_,
        _link._length_,
        _mean._length_,
        _offset._length_,
        _weight._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, errfn = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: errfn = \'B\', \'G\', \'N\' or \'P\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, link = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: if errfn = \'B\', link = \'C\', \'G\' or \'P\',',
                    'otherwise, link = \'E\', \'I\', \'L\', \'R\' or \'S\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 102 and _ifmt == 102:
            _ierr_fl = 2
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, errfn = \'{:s}\' and link = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                        utils._STR_FMT_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if errfn = \'B\', link = \'C\', \'G\' or \'P\',',
                    'otherwise, link = \'E\', \'I\', \'L\', \'R\' or \'S\'.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mean = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mean = \'M\' or \'Z\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, weight = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: if vfobs = True, weight = \'U\' or \'W\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 6:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 9 and _ifmt == 9:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 10 and _ifmt == 10:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, isx[{:s}] < 0.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: isx[j-1] >= 0.0, for j = 1,2,...,m.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 11 and _ifmt == 11:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ip > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 12 and _ifmt == 12:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, t[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: t[i-1] >= 0.0, for all i.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 14 and _ifmt == 14:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, wt[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: wt[i-1] >= 0.0, for all i.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 15 and _ifmt == 15:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, s = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: s > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 16 and _ifmt == 16:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a = 0.0.',
                    'Constraint: if link = \'E\', a != 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 18 and _ifmt == 18:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, cov[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: cov[i] >= 0.0 for at least one diagonal element.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 22 and _ifmt == 22:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'At least one predicted value could not be calculated as required.',
                    'sepred is set to -99.0 for affected predicted values.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _eta_cls.refresh_py(_eta)
        _seeta_cls.refresh_py(_seeta)
        _pred_cls.refresh_py(_pred)
        _sepred_cls.refresh_py(_sepred)
        return_data = None
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def robustm(indw, ipsi, isigma, indc, n, m, x, y, cpsi, h1, h2, h3, cucv, dchi, theta, sigma, c, rs, wgt, tol, maxit, nitmon, stat, io_manager=None):
    u"""
    Robust regression, standard M-estimates.

    ``robustm`` performs bounded influence regression (M-estimates).
    Several standard methods are available.

    For full information please refer to the NAG Library document for
    g02ha

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02haf.html

    Parameters
    ----------
    indw : int
        Specifies the type of regression to be performed.

        `indw` < 0

            Mallows type regression with Maronna's proposed weights.

        `indw` = 0

            Huber type regression.

        `indw` > 0

            Schweppe type regression with Krasker--Welsch weights.

    ipsi : int
        Specifies which psi function is to be used.

        `ipsi` = 0

            psi(t) = t, i.e., least squares.

        `ipsi` = 1

            Huber's function.

        `ipsi` = 2

            Hampel's piecewise linear function.

        `ipsi` = 3

            Andrew's sine wave.

        `ipsi` = 4

            Tukey's bi-weight.

    isigma : int
        Specifies how sigma is to be estimated.

        `isigma` < 0

            sigma is estimated by median absolute deviation of
            residuals.

        `isigma` = 0

            sigma is held constant at its initial value.

        `isigma` > 0

            sigma is estimated using the chi function.

    indc : int
        If `indw` != 0, `indc` specifies the approximations used in
        estimating the covariance matrix of theta^.

        `indc` = 1

            Averaging over residuals.

        `indc` != 1

            Replacing expected by observed.

        `indw` = 0

            `indc` is not referenced.

    n : int
        n, the number of observations.

    m : int
        m, the number of independent variables.

    x : float, shape (`n`, `m`), modified in place
        `On entry`: the values of the X matrix, i.e., the independent
        variables. `x`[i-1,j-1] must contain the ijth element of X, for
        j = 1,2,...,m, for i = 1,2,...,n.

        If `indw` < 0, then during calculations the elements of `x` will
        be transformed as described in [Description].

        Before exit the inverse transformation will be applied.

        As a result there may be slight differences between the input
        `x` and the output `x`.

        `On exit`: unchanged, except as described above.

    y : float, shape (`n`), modified in place
        `On entry`: the data values of the dependent variable.

        `y`[i-1] must contain the value of y for the ith observation,
        for i = 1,2,...,n.

        If `indw` < 0, then during calculations the elements of `y` will
        be transformed as described in [Description].

        Before exit the inverse transformation will be applied.

        As a result there may be slight differences between the input
        `y` and the output `y`.

        `On exit`: unchanged, except as described above.

    cpsi : float
        If `ipsi` = 1, `cpsi` must specify the parameter, c, of Huber's
        psi function.

        If `ipsi` != 1 on entry, `cpsi` is not referenced.

    h1 : float
        If `ipsi` = 2, `h1`, `h2`, and `h3` must specify the parameters
        h_1, h_2, and h_3, of Hampel's piecewise linear psi function.
        `h1`, `h2`, and `h3` are not referenced if `ipsi` != 2.

    h2 : float
        If `ipsi` = 2, `h1`, `h2`, and `h3` must specify the parameters
        h_1, h_2, and h_3, of Hampel's piecewise linear psi function.
        `h1`, `h2`, and `h3` are not referenced if `ipsi` != 2.

    h3 : float
        If `ipsi` = 2, `h1`, `h2`, and `h3` must specify the parameters
        h_1, h_2, and h_3, of Hampel's piecewise linear psi function.
        `h1`, `h2`, and `h3` are not referenced if `ipsi` != 2.

    cucv : float
        If `indw` < 0, must specify the value of the constant, c, of the
        function u for Maronna's proposed weights.

        If `indw` > 0, must specify the value of the function u for the
        Krasker--Welsch weights.

        If `indw` = 0, is not referenced.

    dchi : float
        d, the constant of the chi function. `dchi` is not referenced if
        `ipsi` = 0, or if `isigma` <= 0.

    theta : float, shape (`m`), modified in place
        `On entry`: starting values of the parameter vector theta. These
        may be obtained from least squares regression. Alternatively if
        `isigma` < 0 and `sigma` = 1 or if `isigma` > 0 and `sigma`
        approximately equals the standard deviation of the dependent
        variable, y, then `theta`[i-1] = 0.0, for i = 1,2,...,m may
        provide reasonable starting values.

        `On exit`: `theta`[i-1] contains the M-estimate of theta_i, for
        i = 1,2,...,m.

    sigma : float
        A starting value for the estimation of sigma. `sigma` should be
        approximately the standard deviation of the residuals from the
        model evaluated at the value of theta given by `theta` on entry.

    c : float, shape (`m`, `m`), modified in place
        `On exit`: the diagonal elements of `c` contain the estimated
        asymptotic standard errors of the estimates of theta, i.e.,
        `c`[i-1,i-1] contains the estimated asymptotic standard error of
        the estimate contained in `theta`[i-1].

        The elements above the diagonal contain the estimated asymptotic
        correlation between the estimates of theta, i.e., `c`[i-1,j-1],
        1 <= i < j <= m contains the asymptotic correlation between the
        estimates contained in `theta`[i-1] and `theta`[j-1].

        The elements below the diagonal contain the estimated asymptotic
        covariance between the estimates of theta, i.e., `c`[i-1,j-1], 1
        <= j < i <= m contains the estimated asymptotic covariance
        between the estimates contained in `theta`[i-1] and
        `theta`[j-1].

    rs : float, shape (`n`), modified in place
        `On exit`: the residuals from the model evaluated at final value
        of `theta`, i.e., `rs` contains the vector (y-X theta^).

    wgt : float, shape (`n`), modified in place
        `On exit`: the vector of weights. `wgt`[i-1] contains the weight
        for the ith observation, for i = 1,2,...,n.

    tol : float
        The relative precision for the calculation of A (if `indw` !=
        0), the estimates of theta and the estimate of sigma (if
        `isigma` != 0). Convergence is assumed when the relative change
        in all elements being considered is less than `tol`.

        If `indw` < 0 and `isigma` < 0, `tol` is also used to determine
        the precision of beta_1.

        It is advisable for `tol` to be greater than 100*machine
        precision.

    maxit : int
        The maximum number of iterations that should be used in the
        calculation of A (if `indw` != 0), and of the estimates of theta
        and sigma, and of beta_1 (if `indw` < 0 and `isigma` < 0).

        A value of `maxit` = 50 should be adequate for most uses.

    nitmon : int
        The amount of information that is printed on each iteration.

        `nitmon` = 0

            No information is printed.

        `nitmon` != 0

            The current estimate of theta, the change in theta during
            the current iteration and the current value of sigma are
            printed on the first and every abs(`nitmon`) iterations.

        Also, if `indw` != 0 and `nitmon` > 0, then information on the
        iterations to calculate A is printed.

        This is the current estimate of A and the maximum value of
        S_{ij} (see [Description]).

        When printing occurs the output is directed to the file object
        associated with the advisory I/O unit (see
        :class:`~naginterfaces.base.utils.FileObjManager`).

    stat : float, shape (4), modified in place
        `On exit`: the following values are assigned to `stat`:

            `stat`[0] = beta_1 if `isigma` < 0, or `stat`[0] = beta_2 if
            `isigma` > 0.

            `stat`[1] = number of iterations used to calculate A.

            `stat`[2] = number of iterations used to calculate final
            estimates of theta and sigma.

            `stat`[3] = k, the rank of the weighted least squares
            equations.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    sigma : float
        Contains the final estimate of sigma if `isigma` != 0 or the
        value assigned on entry if `isigma` = 0.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>* and ldx = *<value>*.

            Constraint: ldx >= `n`.

        (`errno` 1)
            On entry, `n` = *<value>* and `m` = *<value>*.

            Constraint: `n` > `m`.

        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 2.

        (`errno` 2)
            On entry: `ipsi` = *<value>*.

            Constraint: `ipsi` = 0, 1, 2, 3 or 4.

        (`errno` 3)
            On entry, `cucv` = *<value>*.

            Constraint: `cucv` >= `m`.

        (`errno` 3)
            On entry, `cucv` = *<value>*.

            Constraint: `cucv` >= sqrt(`m`).

        (`errno` 3)
            On entry: `h1`, `h2`, `h3` incorrectly set.

            Constraint: 0.0 <= `h1` <= `h2` <= `h3` and `h3` > 0.0.

        (`errno` 3)
            On entry: `cpsi` = *<value>*.

            Constraint: `cpsi` > 0.0.

        (`errno` 3)
            On entry, `dchi` = *<value>*.

            Constraint: `dchi` > 0.0.

        (`errno` 3)
            On entry, `sigma` = *<value>*.

            Constraint: `sigma` > 0.0.

        (`errno` 4)
            On entry, `maxit` = *<value>*.

            Constraint: `maxit` > 0.

        (`errno` 4)
            On entry, `tol` = *<value>*.

            Constraint: `tol` > 0.0.

        (`errno` 5)
            The number of iterations required to calculate the weights
            exceeds `maxit`. (Only if `indw` != 0.)

        (`errno` 6)
            The number of iterations required to calculate beta_1
            exceeds `maxit`. (Only if `indw` < 0 and `isigma` < 0.)

        (`errno` 7)
            Iterations to calculate estimates of `theta` failed to
            converge in `maxit` iterations: `maxit` = *<value>*.

        (`errno` 7)
            The number of iterations required to calculate theta and
            sigma exceeds `maxit`.

        (`errno` 12)
            Error degrees of freedom n-k <= 0, where n = *<value>* and
            the rank of `x`, k = *<value>*.

        (`errno` 12)
            Estimated value of `sigma` is zero.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 8)
            Weighted least squares equations not of full rank: rank =
            *<value>*.

        (`errno` 9)
            Failure to invert matrix while calculating covariance.

        (`errno` 10)
            Factor for covariance matrix = 0, uncorrected (X^TX)^-1
            given.

        (`errno` 11)
            Variance of an element of `theta` <= 0.0, correlations set
            to 0.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Hampel, F R, Ronchetti, E M, Rousseeuw, P J and Stahel, W A, 1986,
    `Robust Statistics. The Approach Based on Influence Functions`,
    Wiley

    Huber, P J, 1981, `Robust Statistics`, Wiley

    Marazzi, A, 1987, `Weights for bounded influence regression in
    ROBETH`, Cah. Rech. Doc. IUMSP, No. 3 ROB 3, Institut Universitaire
    de Médecine Sociale et Préventive, Lausanne

    Marazzi, A, 1987, `Subroutines for robust and bounded influence
    regression in ROBETH`, Cah. Rech. Doc. IUMSP, No. 3 ROB 2, Institut
    Universitaire de Médecine Sociale et Préventive, Lausanne
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.robustm'

    utils.FileObjManager._check_type_compat(
        io_manager, {'fun_name': fun_name, 'entity': 'io_manager'}
    )

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    def cb_print_rech_py(_, en_data, nout, rec, ifail, rec_len):
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data, utils._EngineIntScalarType.to_py(nout.contents), rec[:rec_len], io_manager, excs, ifail.contents,
                fun_name=fun_name,
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02haft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._PRINT_REC_FUNCTYPE,
        utils._PRINT_RECH_FUNCTYPE,
        utils._BYREF_INT, # Input, indw.
        utils._BYREF_INT, # Input, ipsi.
        utils._BYREF_INT, # Input, isigma.
        utils._BYREF_INT, # Input, indc.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, x[ldx, m].
        utils._BYREF_INT, # Input, ldx.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, y[n].
        utils._BYREF_FLOAT64, # Input, cpsi.
        utils._BYREF_FLOAT64, # Input, h1.
        utils._BYREF_FLOAT64, # Input, h2.
        utils._BYREF_FLOAT64, # Input, h3.
        utils._BYREF_FLOAT64, # Input, cucv.
        utils._BYREF_FLOAT64, # Input, dchi.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, theta[m].
        utils._BYREF_FLOAT64, # Input/Output, sigma.
        utils._EngineFloat64ArrayType.c_type, # Output, c[ldc, m].
        utils._BYREF_INT, # Input, ldc.
        utils._EngineFloat64ArrayType.c_type, # Output, rs[n].
        utils._EngineFloat64ArrayType.c_type, # Output, wgt[n].
        utils._BYREF_FLOAT64, # Input, tol.
        utils._BYREF_INT, # Input, maxit.
        utils._BYREF_INT, # Input, nitmon.
        utils._BYREF_INT, # Input, nout.
        utils._EngineFloat64ArrayType.c_type, # Output, stat[4].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((x, 'x'), (c, 'c')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _print_rech = utils._PRINT_RECH_FUNCTYPE(cb_print_rech_py)
    _indw = utils._EngineIntScalarType(
        indw, {'fun_name': fun_name, 'entity': 'indw'},
    ).to_ctypes()
    _ipsi = utils._EngineIntScalarType(
        ipsi, {'fun_name': fun_name, 'entity': 'ipsi'},
    ).to_ctypes()
    _isigma = utils._EngineIntScalarType(
        isigma, {'fun_name': fun_name, 'entity': 'isigma'},
    ).to_ctypes()
    _indc = utils._EngineIntScalarType(
        indc, {'fun_name': fun_name, 'entity': 'indc'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _ldx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py, _m_py),
        mutual_sorder=_m_sorder,
    )
    _x = _x_cls.to_ctypes()
    _y_cls = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py,),
    )
    _y = _y_cls.to_ctypes()
    _cpsi = utils._EngineFloat64ScalarType(
        cpsi, {'fun_name': fun_name, 'entity': 'cpsi'},
    ).to_ctypes()
    _h1 = utils._EngineFloat64ScalarType(
        h1, {'fun_name': fun_name, 'entity': 'h1'},
    ).to_ctypes()
    _h2 = utils._EngineFloat64ScalarType(
        h2, {'fun_name': fun_name, 'entity': 'h2'},
    ).to_ctypes()
    _h3 = utils._EngineFloat64ScalarType(
        h3, {'fun_name': fun_name, 'entity': 'h3'},
    ).to_ctypes()
    _cucv = utils._EngineFloat64ScalarType(
        cucv, {'fun_name': fun_name, 'entity': 'cucv'},
    ).to_ctypes()
    _dchi = utils._EngineFloat64ScalarType(
        dchi, {'fun_name': fun_name, 'entity': 'dchi'},
    ).to_ctypes()
    _theta_cls = utils._EngineFloat64ArrayType(
        theta, {'fun_name': fun_name, 'entity': 'theta'},
        exp_shape=(_m_py,),
    )
    _theta = _theta_cls.to_ctypes()
    _sigma = utils._EngineFloat64ScalarType(
        sigma, {'fun_name': fun_name, 'entity': 'sigma'},
    ).to_ctypes()
    _ldc_py = utils._EngineFloat64ArrayType.get_nd(c, {'fun_name': fun_name, 'entity': 'c'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldc = utils._EngineIntScalarType(
        _ldc_py, {'fun_name': fun_name, 'entity': '_ldc'},
    ).to_ctypes()
    _c_cls = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_m_py, _m_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _c = _c_cls.to_ctypes(can_reference_contents=False)
    _rs_cls = utils._EngineFloat64ArrayType(
        rs, {'fun_name': fun_name, 'entity': 'rs'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _rs = _rs_cls.to_ctypes(can_reference_contents=False)
    _wgt_cls = utils._EngineFloat64ArrayType(
        wgt, {'fun_name': fun_name, 'entity': 'wgt'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _wgt = _wgt_cls.to_ctypes(can_reference_contents=False)
    _tol = utils._EngineFloat64ScalarType(
        tol, {'fun_name': fun_name, 'entity': 'tol'},
    ).to_ctypes()
    _maxit = utils._EngineIntScalarType(
        maxit, {'fun_name': fun_name, 'entity': 'maxit'},
    ).to_ctypes()
    _nitmon = utils._EngineIntScalarType(
        nitmon, {'fun_name': fun_name, 'entity': 'nitmon'},
    ).to_ctypes()
    _nout = utils._EngineIntScalarType.c_type(utils.FileObjManager.advunit)
    _stat_cls = utils._EngineFloat64ArrayType(
        stat, {'fun_name': fun_name, 'entity': 'stat'},
        exp_shape=(4,),
        check_contents=False,
    )
    _stat = _stat_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        utils._UNUSED_PRINT_REC,
        _print_rech,
        _indw,
        _ipsi,
        _isigma,
        _indc,
        _n,
        _m,
        _x,
        _ldx,
        _y,
        _cpsi,
        _h1,
        _h2,
        _h3,
        _cucv,
        _dchi,
        _theta,
        _sigma,
        _c,
        _ldc,
        _rs,
        _wgt,
        _tol,
        _maxit,
        _nitmon,
        _nout,
        _stat,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s} and ldx = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: ldx >= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s} and m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: n > m.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99988:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry: ipsi = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ipsi = 0, 1, 2, 3 or 4.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, cucv = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: cucv >= m.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, cucv = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: cucv >= sqrt(m).',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry: h1, h2, h3 incorrectly set.',
                    'Constraint: 0.0 <= h1 <= h2 <= h3 and h3 > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry: cpsi = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: cpsi > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, dchi = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: dchi > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, sigma = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: sigma > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99986:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, maxit = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: maxit > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99987:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, tol = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: tol > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99985:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The number of iterations required to calculate the weights exceeds maxit.',
                    '(Only if indw != 0.)',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99984:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The number of iterations required to calculate beta_1 exceeds maxit.',
                    '(Only if indw < 0 and isigma < 0.)',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 99981:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Iterations to calculate estimates of theta failed to converge',
                    'in maxit iterations: maxit = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 99982:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The number of iterations required to calculate theta and sigma',
                    'exceeds maxit.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 99980:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Weighted least squares equations not of full rank:',
                    'rank = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 9 and _ifmt == 99979:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Failure to invert matrix while calculating covariance.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 10 and _ifmt == 99978:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Factor for covariance matrix = 0, uncorrected (X^TX)^-1 given.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 11 and _ifmt == 99977:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Variance of an element of theta <= 0.0, correlations set to 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 12 and _ifmt == 99976:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Error degrees of freedom n-k <= 0, where n = {:s}'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'and the rank of x, k = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 12 and _ifmt == 99983:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Estimated value of sigma is zero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _x_cls.refresh_py(_x)
        _y_cls.refresh_py(_y)
        _theta_cls.refresh_py(_theta)
        _sigma_py = utils._EngineFloat64ScalarType.to_py(_sigma)
        _c_cls.refresh_py(_c)
        _rs_cls.refresh_py(_rs)
        _wgt_cls.refresh_py(_wgt)
        _stat_cls.refresh_py(_stat)
        return_data = _sigma_py
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = _collections.namedtuple(
                'nag_return_data',
                ['sigma'],
            )(
                _sigma_py,
            )
    _err_class.manage()
    return return_data

def robustm_wts(ucv, n, m, x, a, z, bl, bd, tol, maxit, nitmon, data=None, io_manager=None):
    u"""
    Robust regression, compute weights for use with ``robustm_user``.

    ``robustm_wts`` finds, for a real matrix X of full column rank, a
    lower triangular matrix A such that (A^TA)^(-1) is proportional to a
    robust estimate of the covariance of the variables. ``robustm_wts``
    is intended for the calculation of weights of bounded influence
    regression using ``robustm_user``.

    For full information please refer to the NAG Library document for
    g02hb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02hbf.html

    Parameters
    ----------
    ucv : callable retval = ucv(t, data=None)
        `ucv` must return the value of the function u for a given value
        of its argument.

        The value of u must be non-negative.

        Parameters
        ~~~~~~~~~~
        t : float
            The argument for which `ucv` must be evaluated.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        retval : float
            The value of u(t) evaluated at `t`.

    n : int
        n, the number of observations.

    m : int
        m, the number of independent variables.

    x : float, shape (`n`, `m`)
        The real matrix X, i.e., the independent variables. `x`[i-1,j-1]
        must contain the ijth element of `x`, for j = 1,2,...,m, for i =
        1,2,...,n.

    a : float, shape (`m`*(`m`+1)/2), modified in place
        `On entry`: an initial estimate of the lower triangular real
        matrix A. Only the lower triangular elements must be given and
        these should be stored row-wise in the array.

        The diagonal elements must be != 0, although in practice will
        usually be > 0.

        If the magnitudes of the columns of X are of the same order the
        identity matrix will often provide a suitable initial value for
        A.

        If the columns of X are of different magnitudes, the diagonal
        elements of the initial value of A should be approximately
        inversely proportional to the magnitude of the columns of X.

        `On exit`: the lower triangular elements of the matrix A, stored
        row-wise.

    z : float, shape (`n`), modified in place
        `On exit`: the value ||z_i||_2, for i = 1,2,...,n.

    bl : float
        The magnitude of the bound for the off-diagonal elements of S_k.

        Suggested value: `bl` = 0.9.

    bd : float
        The magnitude of the bound for the diagonal elements of S_k.

        Suggested value: `bd` = 0.9.

    tol : float
        The relative precision for the final value of A. Iteration will
        stop when the maximum value of |s_{jl}| is less than `tol`.

    maxit : int
        The maximum number of iterations that will be used during the
        calculation of A.

        A value of `maxit` = 50 will often be adequate.

    nitmon : int
        Determines the amount of information that is printed on each
        iteration.

        `nitmon` > 0

            The value of A and the maximum value of |s_{jl}| will be
            printed at the first and every `nitmon` iterations.

        `nitmon` <= 0

            No iteration monitoring is printed.

        When printing occurs the output is directed to the file object
        associated with the advisory I/O unit (see
        :class:`~naginterfaces.base.utils.FileObjManager`).

    data : arbitrary, optional
        User-communication data for callback functions.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    nit : int
        The number of iterations performed.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>* and ldx = *<value>*.

            Constraint: ldx >= `n`.

        (`errno` 1)
            On entry, `n` = *<value>* and `m` = *<value>*.

            Constraint: `n` >= `m`.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 2.

        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 2)
            On entry, `bd` = *<value>*.

            Constraint: `bd` > 0.0.

        (`errno` 2)
            On entry, `bl` = *<value>*.

            Constraint: `bl` > 0.0.

        (`errno` 2)
            On entry, `maxit` = *<value>*.

            Constraint: `maxit` > 0.

        (`errno` 2)
            On entry, `tol` = *<value>*.

            Constraint: `tol` > 0.0.

        (`errno` 2)
            On entry, i = *<value>* and the ith diagonal element of A is
            0.

            Constraint: all diagonal elements of A must be non-zero.

        (`errno` 3)
            Value returned by `ucv` function < 0: u(*<value>*) =
            *<value>*.

            The value of u must be non-negative.

        (`errno` 4)
            Iterations to calculate weights failed to converge in
            `maxit` iterations: `maxit` = *<value>*.

    References
    ----------
    Hampel, F R, Ronchetti, E M, Rousseeuw, P J and Stahel, W A, 1986,
    `Robust Statistics. The Approach Based on Influence Functions`,
    Wiley

    Huber, P J, 1981, `Robust Statistics`, Wiley

    Marazzi, A, 1987, `Weights for bounded influence regression in
    ROBETH`, Cah. Rech. Doc. IUMSP, No. 3 ROB 3, Institut Universitaire
    de Médecine Sociale et Préventive, Lausanne
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.robustm_wts'

    utils.FileObjManager._check_type_compat(
        io_manager, {'fun_name': fun_name, 'entity': 'io_manager'}
    )

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    def cb_print_rech_py(_, en_data, nout, rec, ifail, rec_len):
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data, utils._EngineIntScalarType.to_py(nout.contents), rec[:rec_len], io_manager, excs, ifail.contents,
                fun_name=fun_name,
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    cb_ucv_name = fun_name + ':ucv'

    ucvh = utils._EngineState._sys_functype(
        None,
        utils._UNUSED_FUNCTYPE,
        utils._UNUSED_FUNCTYPE,
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, t.
        utils._BYREF_FLOAT64, # Output, retval.
        utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
    )

    def cb_ucvh(_ucv, _ucvsub, en_data, t, retval, _iuser, _ruser):
        """Helper for ucv."""
        try:
            _t_ucvh = utils._EngineFloat64ScalarType.to_py(t.contents)
            if data is not None:
                _retval_ucvh = ucv(
                    _t_ucvh,
                    data,
                )
            else:
                _retval_ucvh = ucv(
                    _t_ucvh,
                )
            utils._EngineFloat64ScalarType(
                _retval_ucvh, {'fun_name': cb_ucv_name, 'returned': 'retval'},
            ).refresh_ctypes_buffer(retval)
        except:
            excs.append(_sys.exc_info())
            en_data.contents.hlperr = utils._IERR_HELPER

    engine_ctypes = utils._EngineState._get_symbol('g02hbft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._PRINT_REC_FUNCTYPE,
        utils._PRINT_RECH_FUNCTYPE,
        utils._UNUSED_FUNCTYPE,
        utils._UNUSED_FUNCTYPE,
        ucvh,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Input, x[ldx, m].
        utils._BYREF_INT, # Input, ldx.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, a[m*(m+1)//2].
        utils._EngineFloat64ArrayType.c_type, # Output, z[n].
        utils._BYREF_FLOAT64, # Input, bl.
        utils._BYREF_FLOAT64, # Input, bd.
        utils._BYREF_FLOAT64, # Input, tol.
        utils._BYREF_INT, # Input, maxit.
        utils._BYREF_INT, # Input, nitmon.
        utils._BYREF_INT, # Input, nadv.
        utils._BYREF_INT, # Output, nit.
        utils._EngineFloat64ArrayType.c_type, # Workspace, wk.
        utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((x, 'x'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _print_rech = utils._PRINT_RECH_FUNCTYPE(cb_print_rech_py)
    utils._check_callable(ucv, {'fun_name': fun_name, 'entity': 'ucv'})
    _ucv = utils._UNUSED_CALLBACK
    _ucvsub = utils._UNUSED_CALLBACK
    _ucvh = utils._cast_callback(
        ucvh, cb_ucvh,
        {'fun_name': fun_name, 'entity': 'ucv'},
    )
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _ldx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py, _m_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _a_cls = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_m_py*(_m_py+1)//2,),
    )
    _a = _a_cls.to_ctypes()
    _z_cls = utils._EngineFloat64ArrayType(
        z, {'fun_name': fun_name, 'entity': 'z'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _z = _z_cls.to_ctypes(can_reference_contents=False)
    _bl = utils._EngineFloat64ScalarType(
        bl, {'fun_name': fun_name, 'entity': 'bl'},
    ).to_ctypes()
    _bd = utils._EngineFloat64ScalarType(
        bd, {'fun_name': fun_name, 'entity': 'bd'},
    ).to_ctypes()
    _tol = utils._EngineFloat64ScalarType(
        tol, {'fun_name': fun_name, 'entity': 'tol'},
    ).to_ctypes()
    _maxit = utils._EngineIntScalarType(
        maxit, {'fun_name': fun_name, 'entity': 'maxit'},
    ).to_ctypes()
    _nitmon = utils._EngineIntScalarType(
        nitmon, {'fun_name': fun_name, 'entity': 'nitmon'},
    ).to_ctypes()
    _nadv = utils._EngineIntScalarType.c_type(utils.FileObjManager.advunit)
    _nit = utils._EngineIntScalarType.empty_ctypes()
    _wk = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _iuser = None
    _ruser = None
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        utils._UNUSED_PRINT_REC,
        _print_rech,
        _ucv,
        _ucvsub,
        _ucvh,
        _n,
        _m,
        _x,
        _ldx,
        _a,
        _z,
        _bl,
        _bd,
        _tol,
        _maxit,
        _nitmon,
        _nadv,
        _nit,
        _wk,
        _iuser,
        _ruser,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s} and ldx = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: ldx >= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s} and m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: n >= m.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, bd = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: bd > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, bl = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: bl > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, maxit = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: maxit > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, tol = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: tol > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s} and the ith diagonal element of A is 0.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: all diagonal elements of A must be non-zero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Value returned by ucv function < 0:',
                    'u({:s}) = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'The value of u must be non-negative.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Iterations to calculate weights failed to converge',
                    'in maxit iterations: maxit = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _a_cls.refresh_py(_a)
        _z_cls.refresh_py(_z)
        _nit_py = utils._EngineIntScalarType.to_py(_nit)
        return_data = _nit_py
    _err_class.manage()
    return return_data

def robustm_user(chi, psi, psip0, beta, indw, isigma, n, m, x, y, wgt, theta, sigma, rs, tol, eps, maxit, nitmon, data=None, io_manager=None):
    u"""
    Robust regression, compute regression with user-supplied functions
    and weights.

    ``robustm_user`` performs bounded influence regression (M-estimates)
    using an iterative weighted least squares algorithm.

    For full information please refer to the NAG Library document for
    g02hd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02hdf.html

    Parameters
    ----------
    chi : None or callable retval = chi(t, data=None)
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        If `isigma` > 0, `chi` must return the value of the weight
        function chi for a given value of its argument.

        The value of chi must be non-negative.

        Parameters
        ~~~~~~~~~~
        t : float
            The argument for which `chi` must be evaluated.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        retval : float
            The value of the weight function chi evaluated at `t`.

    psi : callable retval = psi(t, data=None)
        `psi` must return the value of the weight function psi for a
        given value of its argument.

        Parameters
        ~~~~~~~~~~
        t : float
            The argument for which `psi` must be evaluated.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        retval : float
            The value of the weight function psi evaluated at `t`.

    psip0 : float
        The value of psi'(0).

    beta : float
        If `isigma` < 0, `beta` must specify the value of beta_1.

        For Huber and Schweppe type regressions, beta_1 is the 75th
        percentile of the standard Normal distribution (see
        ``stat.inv_cdf_normal``).

        For Mallows type regression beta_1 is the solution to

            1/nsum_i = 1^n Phi(beta_1/sqrt(w_i)) = 0.75,

        where Phi is the standard Normal cumulative distribution
        function (see ``specfun.cdf_normal``).

        If `isigma` > 0, `beta` must specify the value of beta_2.

            [table omitted]

        where phi is the standard normal density, i.e., 1/(sqrt(2
        pi))exp(-1/2x^2).

        If `isigma` = 0, `beta` is not referenced.

    indw : int
        Determines the type of regression to be performed.

        `indw` = 0

            Huber type regression.

        `indw` < 0

            Mallows type regression.

        `indw` > 0

            Schweppe type regression.

    isigma : int
        Determines how sigma is to be estimated.

        `isigma` = 0

            sigma is held constant at its initial value.

        `isigma` < 0

            sigma is estimated by median absolute deviation of
            residuals.

        `isigma` > 0

            sigma is estimated using the chi function.

    n : int
        n, the number of observations.

    m : int
        m, the number of independent variables.

    x : float, shape (`n`, `m`), modified in place
        `On entry`: the values of the X matrix, i.e., the independent
        variables. `x`[i-1,j-1] must contain the ijth element of `x`,
        for j = 1,2,...,m, for i = 1,2,...,n.

        If `indw` < 0, during calculations the elements of `x` will be
        transformed as described in [Description].

        Before exit the inverse transformation will be applied.

        As a result there may be slight differences between the input
        `x` and the output `x`.

        `On exit`: unchanged, except as described above.

    y : float, shape (`n`), modified in place
        `On entry`: the data values of the dependent variable.

        `y`[i-1] must contain the value of y for the ith observation,
        for i = 1,2,...,n.

        If `indw` < 0, during calculations the elements of `y` will be
        transformed as described in [Description].

        Before exit the inverse transformation will be applied.

        As a result there may be slight differences between the input
        `y` and the output `y`.

        `On exit`: unchanged, except as described above.

    wgt : float, shape (`n`), modified in place
        `On entry`: the weight for the ith observation, for i =
        1,2,...,n.

        If `indw` < 0, during calculations elements of `wgt` will be
        transformed as described in [Description].

        Before exit the inverse transformation will be applied.

        As a result there may be slight differences between the input
        `wgt` and the output `wgt`.

        If `wgt`[i-1] <= 0, the ith observation is not included in the
        analysis.

        If `indw` = 0, `wgt` is not referenced.

        `On exit`: unchanged, except as described above.

    theta : float, shape (`m`), modified in place
        `On entry`: starting values of the parameter vector theta. These
        may be obtained from least squares regression. Alternatively if
        `isigma` < 0 and `sigma` = 1 or if `isigma` > 0 and `sigma`
        approximately equals the standard deviation of the dependent
        variable, y, then `theta`[i-1] = 0.0, for i = 1,2,...,m may
        provide reasonable starting values.

        `On exit`: the M-estimate of theta_i, for i = 1,2,...,m.

    sigma : float
        A starting value for the estimation of sigma. `sigma` should be
        approximately the standard deviation of the residuals from the
        model evaluated at the value of theta given by `theta` on entry.

    rs : float, shape (`n`), modified in place
        `On exit`: the residuals from the model evaluated at final value
        of `theta`, i.e., `rs` contains the vector (y-X theta^).

    tol : float
        The relative precision for the final estimates. Convergence is
        assumed when both the relative change in the value of `sigma`
        and the relative change in the value of each element of `theta`
        are less than `tol`.

        It is advisable for `tol` to be greater than 100*machine
        precision.

    eps : float
        A relative tolerance to be used to determine the rank of X. See
        ``linsys.real_gen_solve`` for further details.

        If `eps` < machine precision or `eps` > 1.0, machine precision
        will be used in place of `tol`.

        A reasonable value for `eps` is 5.0*10^(-6) where this value is
        possible.

    maxit : int
        The maximum number of iterations that should be used during the
        estimation.

        A value of `maxit` = 50 should be adequate for most uses.

    nitmon : int
        Determines the amount of information that is printed on each
        iteration.

        `nitmon` <= 0

            No information is printed.

        `nitmon` > 0

            On the first and every `nitmon` iterations the values of
            `sigma`, `theta` and the change in `theta` during the
            iteration are printed.

        When printing occurs the output is directed to the file object
        associated with the advisory I/O unit (see
        :class:`~naginterfaces.base.utils.FileObjManager`).

    data : arbitrary, optional
        User-communication data for callback functions.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    k : int
        The column rank of the matrix X.

    sigma : float
        The final estimate of sigma if `isigma` != 0 or the value
        assigned on entry if `isigma` = 0.

    nit : int
        The number of iterations that were used during the estimation.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>* and ldx = *<value>*.

            Constraint: ldx >= `n`.

        (`errno` 1)
            On entry, `n` = *<value>* and `m` = *<value>*.

            Constraint: `n` > `m`.

        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 2.

        (`errno` 2)
            On entry, `beta` = *<value>*.

            Constraint: `beta` > 0.0.

        (`errno` 2)
            On entry, `sigma` = *<value>*.

            Constraint: `sigma` > 0.0.

        (`errno` 3)
            On entry, `maxit` = *<value>*.

            Constraint: `maxit` > 0.

        (`errno` 3)
            On entry, `tol` = *<value>*.

            Constraint: `tol` > 0.0.

        (`errno` 4)
            Value given by `chi` function < 0: `chi`(*<value>*) =
            *<value>*.

            The value of `chi` must be non-negative.

        (`errno` 5)
            Estimated value of `sigma` is zero.

        (`errno` 6)
            Iterations to solve the weighted least squares equations
            failed to converge.

        (`errno` 8)
            The function has failed to converge in `maxit` iterations.

        (`errno` 9)
            Having removed cases with zero weight, the value of `n`-`k`
            <= 0, i.e., no degree of freedom for error. This error will
            only occur if `isigma` > 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 7)
            The weighted least squares equations are not of full rank.
            This may be due to the X matrix not being of full rank, in
            which case the results will be valid. It may also occur if
            some of the G_{{ii}} values become very small or zero, see
            [Further Comments]. The rank of the equations is given by
            `k`. If the matrix just fails the test for nonsingularity
            then the result `errno` = 7 and `k` = `m` is possible (see
            ``linsys.real_gen_solve``).

    References
    ----------
    Hampel, F R, Ronchetti, E M, Rousseeuw, P J and Stahel, W A, 1986,
    `Robust Statistics. The Approach Based on Influence Functions`,
    Wiley

    Huber, P J, 1981, `Robust Statistics`, Wiley

    Marazzi, A, 1987, `Subroutines for robust and bounded influence
    regression in ROBETH`, Cah. Rech. Doc. IUMSP, No. 3 ROB 2, Institut
    Universitaire de Médecine Sociale et Préventive, Lausanne
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.robustm_user'

    utils.FileObjManager._check_type_compat(
        io_manager, {'fun_name': fun_name, 'entity': 'io_manager'}
    )

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    def cb_print_rech_py(_, en_data, nout, rec, ifail, rec_len):
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data, utils._EngineIntScalarType.to_py(nout.contents), rec[:rec_len], io_manager, excs, ifail.contents,
                fun_name=fun_name,
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    cb_chi_name = fun_name + ':chi'

    chih = utils._EngineState._sys_functype(
        None,
        utils._UNUSED_FUNCTYPE,
        utils._UNUSED_FUNCTYPE,
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, t.
        utils._BYREF_FLOAT64, # Output, retval.
        utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
    )

    if chi is None:
        chi_auxil_ctypes = utils._EngineState._get_symbol('g02hdzt')
        chi_auxil_ctypes.restype = None
        chi_auxil_ctypes.argtypes = (
            utils._BYREF_ENGINEDATA,
            utils._BYREF_FLOAT64, # Input, t.
            utils._BYREF_FLOAT64, # Output, retval.
            utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
            utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
        )

    def cb_chih(_chi, _chisub, en_data, t, retval, iuser, ruser):
        """Helper for chi."""
        try:
            if chi is None:
                chi_auxil_ctypes(
                    en_data,
                    t,
                    retval,
                    iuser,
                    ruser,
                )
                return
            _t_chih = utils._EngineFloat64ScalarType.to_py(t.contents)
            if data is not None:
                _retval_chih = chi(
                    _t_chih,
                    data,
                )
            else:
                _retval_chih = chi(
                    _t_chih,
                )
            utils._EngineFloat64ScalarType(
                _retval_chih, {'fun_name': cb_chi_name, 'returned': 'retval'},
            ).refresh_ctypes_buffer(retval)
        except:
            excs.append(_sys.exc_info())
            en_data.contents.hlperr = utils._IERR_HELPER

    cb_psi_name = fun_name + ':psi'

    psih = utils._EngineState._sys_functype(
        None,
        utils._UNUSED_FUNCTYPE,
        utils._UNUSED_FUNCTYPE,
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, t.
        utils._BYREF_FLOAT64, # Output, retval.
        utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
    )

    def cb_psih(_psi, _psisub, en_data, t, retval, _iuser, _ruser):
        """Helper for psi."""
        try:
            _t_psih = utils._EngineFloat64ScalarType.to_py(t.contents)
            if data is not None:
                _retval_psih = psi(
                    _t_psih,
                    data,
                )
            else:
                _retval_psih = psi(
                    _t_psih,
                )
            utils._EngineFloat64ScalarType(
                _retval_psih, {'fun_name': cb_psi_name, 'returned': 'retval'},
            ).refresh_ctypes_buffer(retval)
        except:
            excs.append(_sys.exc_info())
            en_data.contents.hlperr = utils._IERR_HELPER

    engine_ctypes = utils._EngineState._get_symbol('g02hdft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._PRINT_REC_FUNCTYPE,
        utils._PRINT_RECH_FUNCTYPE,
        utils._UNUSED_FUNCTYPE,
        utils._UNUSED_FUNCTYPE,
        chih,
        utils._UNUSED_FUNCTYPE,
        utils._UNUSED_FUNCTYPE,
        psih,
        utils._BYREF_FLOAT64, # Input, psip0.
        utils._BYREF_FLOAT64, # Input, beta.
        utils._BYREF_INT, # Input, indw.
        utils._BYREF_INT, # Input, isigma.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, x[ldx, m].
        utils._BYREF_INT, # Input, ldx.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, y[n].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, wgt[n].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, theta[m].
        utils._BYREF_INT, # Output, k.
        utils._BYREF_FLOAT64, # Input/Output, sigma.
        utils._EngineFloat64ArrayType.c_type, # Output, rs[n].
        utils._BYREF_FLOAT64, # Input, tol.
        utils._BYREF_FLOAT64, # Input, eps.
        utils._BYREF_INT, # Input, maxit.
        utils._BYREF_INT, # Input, nitmon.
        utils._BYREF_INT, # Input, nadv.
        utils._BYREF_INT, # Output, nit.
        utils._EngineFloat64ArrayType.c_type, # Workspace, wk.
        utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((x, 'x'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _print_rech = utils._PRINT_RECH_FUNCTYPE(cb_print_rech_py)
    utils._check_callable(chi, {'fun_name': fun_name, 'entity': 'chi'}, allow_none=True)
    _chi = utils._UNUSED_CALLBACK
    _chisub = utils._UNUSED_CALLBACK
    _chih = utils._cast_callback(
        chih, cb_chih,
        {'fun_name': fun_name, 'entity': 'chi'},
    )
    utils._check_callable(psi, {'fun_name': fun_name, 'entity': 'psi'})
    _psi = utils._UNUSED_CALLBACK
    _psisub = utils._UNUSED_CALLBACK
    _psih = utils._cast_callback(
        psih, cb_psih,
        {'fun_name': fun_name, 'entity': 'psi'},
    )
    _psip0 = utils._EngineFloat64ScalarType(
        psip0, {'fun_name': fun_name, 'entity': 'psip0'},
    ).to_ctypes()
    _beta = utils._EngineFloat64ScalarType(
        beta, {'fun_name': fun_name, 'entity': 'beta'},
    ).to_ctypes()
    _indw = utils._EngineIntScalarType(
        indw, {'fun_name': fun_name, 'entity': 'indw'},
    ).to_ctypes()
    _isigma = utils._EngineIntScalarType(
        isigma, {'fun_name': fun_name, 'entity': 'isigma'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _ldx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py, _m_py),
        mutual_sorder=_m_sorder,
    )
    _x = _x_cls.to_ctypes()
    _y_cls = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py,),
    )
    _y = _y_cls.to_ctypes()
    _wgt_cls = utils._EngineFloat64ArrayType(
        wgt, {'fun_name': fun_name, 'entity': 'wgt'},
        exp_shape=(_n_py,),
    )
    _wgt = _wgt_cls.to_ctypes()
    _theta_cls = utils._EngineFloat64ArrayType(
        theta, {'fun_name': fun_name, 'entity': 'theta'},
        exp_shape=(_m_py,),
    )
    _theta = _theta_cls.to_ctypes()
    _k = utils._EngineIntScalarType.empty_ctypes()
    _sigma = utils._EngineFloat64ScalarType(
        sigma, {'fun_name': fun_name, 'entity': 'sigma'},
    ).to_ctypes()
    _rs_cls = utils._EngineFloat64ArrayType(
        rs, {'fun_name': fun_name, 'entity': 'rs'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _rs = _rs_cls.to_ctypes(can_reference_contents=False)
    _tol = utils._EngineFloat64ScalarType(
        tol, {'fun_name': fun_name, 'entity': 'tol'},
    ).to_ctypes()
    _eps = utils._EngineFloat64ScalarType(
        eps, {'fun_name': fun_name, 'entity': 'eps'},
    ).to_ctypes()
    _maxit = utils._EngineIntScalarType(
        maxit, {'fun_name': fun_name, 'entity': 'maxit'},
    ).to_ctypes()
    _nitmon = utils._EngineIntScalarType(
        nitmon, {'fun_name': fun_name, 'entity': 'nitmon'},
    ).to_ctypes()
    _nadv = utils._EngineIntScalarType.c_type(utils.FileObjManager.advunit)
    _nit = utils._EngineIntScalarType.empty_ctypes()
    _wk = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _iuser = None
    _ruser = None
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        utils._UNUSED_PRINT_REC,
        _print_rech,
        _chi,
        _chisub,
        _chih,
        _psi,
        _psisub,
        _psih,
        _psip0,
        _beta,
        _indw,
        _isigma,
        _n,
        _m,
        _x,
        _ldx,
        _y,
        _wgt,
        _theta,
        _k,
        _sigma,
        _rs,
        _tol,
        _eps,
        _maxit,
        _nitmon,
        _nadv,
        _nit,
        _wk,
        _iuser,
        _ruser,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s} and ldx = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: ldx >= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s} and m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: n > m.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, beta = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: beta > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, sigma = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: sigma > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99988:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, maxit = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: maxit > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, tol = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: tol > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99987:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Value given by chi function < 0:',
                    'chi({:s}) = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'The value of chi must be non-negative.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99986:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Estimated value of sigma is zero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99985:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Iterations to solve the weighted least squares equations failed to converge.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 99983:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The weighted least squares equations are not of full rank. This may be',
                    'due to the X matrix not being of full rank, in',
                    'which case the results will be valid. It may also occur if some of the',
                    'G_{{ii}} values become very small or zero,',
                    'see [Further Comments]. The rank of the equations is',
                    'given by k. If the matrix just fails the test for nonsingularity',
                    'then the result errno = 7 and k = m is possible (see linsys.real_gen_solve).',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 99984:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The function has failed to converge in maxit iterations.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 9 and _ifmt == 99982:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Having removed cases with zero weight, the value of n-k <= 0,',
                    'i.e., no degree of freedom for error. This error will only occur if',
                    'isigma > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _x_cls.refresh_py(_x)
        _y_cls.refresh_py(_y)
        _wgt_cls.refresh_py(_wgt)
        _theta_cls.refresh_py(_theta)
        _sigma_py = utils._EngineFloat64ScalarType.to_py(_sigma)
        _rs_cls.refresh_py(_rs)
        _k_py = utils._EngineIntScalarType.to_py(_k)
        _nit_py = utils._EngineIntScalarType.to_py(_nit)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['k', 'sigma', 'nit'],
        )(
            _k_py,
            _sigma_py,
            _nit_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def robustm_user_varmat(psi, psp, indw, indc, sigma, n, m, x, rs, wgt, c, wk, data=None):
    u"""
    Robust regression, variance-covariance matrix following
    ``robustm_user``.

    ``robustm_user_varmat`` calculates an estimate of the asymptotic
    variance-covariance matrix for the bounded influence regression
    estimates (M-estimates).
    It is intended for use with ``robustm_user``.

    For full information please refer to the NAG Library document for
    g02hf

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02hff.html

    Parameters
    ----------
    psi : callable retval = psi(t, data=None)
        `psi` must return the value of the psi function for a given
        value of its argument.

        Parameters
        ~~~~~~~~~~
        t : float
            The argument for which `psi` must be evaluated.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        retval : float
            The value of the function psi evaluated at `t`.

    psp : callable retval = psp(t, data=None)
        `psp` must return the value of psi'(t) = d/dt psi(t) for a given
        value of its argument.

        Parameters
        ~~~~~~~~~~
        t : float
            The argument for which `psp` must be evaluated.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        retval : float
            The value of psi'(t) evaluated at `t`.

    indw : int
        The type of regression for which the asymptotic
        variance-covariance matrix is to be calculated.

        `indw` = -1

            Mallows type regression.

        `indw` = 0

            Huber type regression.

        `indw` = 1

            Schweppe type regression.

    indc : int
        If `indw` != 0, `indc` must specify the approximation to be
        used.

        If `indc` = 1, averaging over residuals.

        If `indc` = 0, replacing expected by observed.

        If `indw` = 0, `indc` is not referenced.

    sigma : float
        The value of sigma^, as given by ``robustm_user``.

    n : int
        n, the number of observations.

    m : int
        m, the number of independent variables.

    x : float, shape (`n`, `m`)
        The values of the X matrix, i.e., the independent variables.
        `x`[i-1,j-1] must contain the ijth element of X, for j =
        1,2,...,m, for i = 1,2,...,n.

    rs : float, shape (`n`)
        The residuals from the bounded influence regression. These are
        given by ``robustm_user``.

    wgt : float, shape (`n`)
        If `indw` != 0, `wgt` must contain the vector of weights used by
        the bounded influence regression. These should be used with
        ``robustm_user``.

        If `indw` = 0, `wgt` is not referenced.

    c : float, shape (`m`, `m`), modified in place
        `On exit`: the estimate of the variance-covariance matrix.

    wk : float, shape (`m`*(`n`+`m`+1)+2*`n`), modified in place
        `On exit`: if `indw` != 0, `wk`[i-1], for i = 1,2,...,n, will
        contain the diagonal elements of the matrix D and `wk`[i-1], for
        i = n+1,...,2n, will contain the diagonal elements of matrix P.

        The rest of the array is used as workspace.

    data : arbitrary, optional
        User-communication data for callback functions.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m` = *<value>* and ldc = *<value>*.

            Constraint: ldc >= `m`.

        (`errno` 1)
            On entry, `n` = *<value>* and ldx = *<value>*.

            Constraint: ldx >= `n`.

        (`errno` 1)
            On entry, `n` = *<value>* and `m` = *<value>*.

            Constraint: `n` > `m`.

        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 2.

        (`errno` 2)
            On entry, `sigma` = *<value>*.

            Constraint: `sigma` >= 0.0.

        (`errno` 3)
            S_1 matrix is singular or almost singular.

        (`errno` 3)
            X^TX matrix not positive definite.

        (`errno` 4)
            Correction factor = 0 (Huber type regression).

    References
    ----------
    Hampel, F R, Ronchetti, E M, Rousseeuw, P J and Stahel, W A, 1986,
    `Robust Statistics. The Approach Based on Influence Functions`,
    Wiley

    Huber, P J, 1981, `Robust Statistics`, Wiley

    Marazzi, A, 1987, `Subroutines for robust and bounded influence
    regression in ROBETH`, Cah. Rech. Doc. IUMSP, No. 3 ROB 2, Institut
    Universitaire de Médecine Sociale et Préventive, Lausanne
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.robustm_user_varmat'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    cb_psi_name = fun_name + ':psi'

    psih = utils._EngineState._sys_functype(
        None,
        utils._UNUSED_FUNCTYPE,
        utils._UNUSED_FUNCTYPE,
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, t.
        utils._BYREF_FLOAT64, # Output, retval.
        utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
    )

    def cb_psih(_psi, _psisub, en_data, t, retval, _iuser, _ruser):
        """Helper for psi."""
        try:
            _t_psih = utils._EngineFloat64ScalarType.to_py(t.contents)
            if data is not None:
                _retval_psih = psi(
                    _t_psih,
                    data,
                )
            else:
                _retval_psih = psi(
                    _t_psih,
                )
            utils._EngineFloat64ScalarType(
                _retval_psih, {'fun_name': cb_psi_name, 'returned': 'retval'},
            ).refresh_ctypes_buffer(retval)
        except:
            excs.append(_sys.exc_info())
            en_data.contents.hlperr = utils._IERR_HELPER

    cb_psp_name = fun_name + ':psp'

    psph = utils._EngineState._sys_functype(
        None,
        utils._UNUSED_FUNCTYPE,
        utils._UNUSED_FUNCTYPE,
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, t.
        utils._BYREF_FLOAT64, # Output, retval.
        utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
    )

    def cb_psph(_psp, _pspsub, en_data, t, retval, _iuser, _ruser):
        """Helper for psp."""
        try:
            _t_psph = utils._EngineFloat64ScalarType.to_py(t.contents)
            if data is not None:
                _retval_psph = psp(
                    _t_psph,
                    data,
                )
            else:
                _retval_psph = psp(
                    _t_psph,
                )
            utils._EngineFloat64ScalarType(
                _retval_psph, {'fun_name': cb_psp_name, 'returned': 'retval'},
            ).refresh_ctypes_buffer(retval)
        except:
            excs.append(_sys.exc_info())
            en_data.contents.hlperr = utils._IERR_HELPER

    engine_ctypes = utils._EngineState._get_symbol('g02hfft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._UNUSED_FUNCTYPE,
        utils._UNUSED_FUNCTYPE,
        psih,
        utils._UNUSED_FUNCTYPE,
        utils._UNUSED_FUNCTYPE,
        psph,
        utils._BYREF_INT, # Input, indw.
        utils._BYREF_INT, # Input, indc.
        utils._BYREF_FLOAT64, # Input, sigma.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Input, x[ldx, m].
        utils._BYREF_INT, # Input, ldx.
        utils._EngineFloat64ArrayType.c_type, # Input, rs[n].
        utils._EngineFloat64ArrayType.c_type, # Input, wgt[n].
        utils._EngineFloat64ArrayType.c_type, # Output, c[ldc, m].
        utils._BYREF_INT, # Input, ldc.
        utils._EngineFloat64ArrayType.c_type, # Output, wk[m*(n+m+1)+2*n].
        utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((x, 'x'), (c, 'c')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    utils._check_callable(psi, {'fun_name': fun_name, 'entity': 'psi'})
    _psi = utils._UNUSED_CALLBACK
    _psisub = utils._UNUSED_CALLBACK
    _psih = utils._cast_callback(
        psih, cb_psih,
        {'fun_name': fun_name, 'entity': 'psi'},
    )
    utils._check_callable(psp, {'fun_name': fun_name, 'entity': 'psp'})
    _psp = utils._UNUSED_CALLBACK
    _pspsub = utils._UNUSED_CALLBACK
    _psph = utils._cast_callback(
        psph, cb_psph,
        {'fun_name': fun_name, 'entity': 'psp'},
    )
    _indw = utils._EngineIntScalarType(
        indw, {'fun_name': fun_name, 'entity': 'indw'},
    ).to_ctypes()
    _indc = utils._EngineIntScalarType(
        indc, {'fun_name': fun_name, 'entity': 'indc'},
    ).to_ctypes()
    _sigma = utils._EngineFloat64ScalarType(
        sigma, {'fun_name': fun_name, 'entity': 'sigma'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _ldx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py, _m_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _rs = utils._EngineFloat64ArrayType(
        rs, {'fun_name': fun_name, 'entity': 'rs'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _wgt = utils._EngineFloat64ArrayType(
        wgt, {'fun_name': fun_name, 'entity': 'wgt'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _ldc_py = utils._EngineFloat64ArrayType.get_nd(c, {'fun_name': fun_name, 'entity': 'c'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldc = utils._EngineIntScalarType(
        _ldc_py, {'fun_name': fun_name, 'entity': '_ldc'},
    ).to_ctypes()
    _c_cls = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_m_py, _m_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _c = _c_cls.to_ctypes(can_reference_contents=False)
    _wk_cls = utils._EngineFloat64ArrayType(
        wk, {'fun_name': fun_name, 'entity': 'wk'},
        exp_shape=(_m_py*(_n_py+_m_py+1)+2*_n_py,),
        check_contents=False,
    )
    _wk = _wk_cls.to_ctypes(can_reference_contents=False)
    _iuser = None
    _ruser = None
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _psi,
        _psisub,
        _psih,
        _psp,
        _pspsub,
        _psph,
        _indw,
        _indc,
        _sigma,
        _n,
        _m,
        _x,
        _ldx,
        _rs,
        _wgt,
        _c,
        _ldc,
        _wk,
        _iuser,
        _ruser,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s} and ldc = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: ldc >= m.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s} and ldx = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: ldx >= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s} and m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: n > m.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, sigma = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: sigma >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'S_1 matrix is singular or almost singular.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'X^TX matrix not positive definite.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Correction factor = 0 (Huber type regression).',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _c_cls.refresh_py(_c)
        _wk_cls.refresh_py(_wk)
        return_data = None
    _err_class.manage()
    return return_data

def robustm_corr_huber(n, m, x, eps, cov, theta, maxit, nitmon, tol, io_manager=None):
    u"""
    Calculates a robust estimation of a covariance matrix, Huber's
    weight function.

    ``robustm_corr_huber`` computes a robust estimate of the covariance
    matrix for an expected fraction of gross errors.

    For full information please refer to the NAG Library document for
    g02hk

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02hkf.html

    Parameters
    ----------
    n : int
        n, the number of observations.

    m : int
        m, the number of columns of the matrix X, i.e., number of
        independent variables.

    x : float, shape (`n`, `m`)
        `x`[i-1,j-1] must contain the ith observation for the jth
        variable, for j = 1,2,...,`m`, for i = 1,2,...,`n`.

    eps : float
        epsilon, the expected fraction of gross errors expected in the
        sample.

    cov : float, shape (`m`*(`m`+1)/2), modified in place
        `On exit`: a robust estimate of the covariance matrix, C. The
        upper triangular part of the matrix C is stored packed by
        columns. C_{ij} is returned in `cov`[(j*(j-1)/2+i)-1], i <= j.

    theta : float, shape (`m`), modified in place
        `On exit`: the robust estimate of the location parameters
        theta_j, for j = 1,2,...,m.

    maxit : int
        The maximum number of iterations that will be used during the
        calculation of the covariance matrix.

        Suggested value: 150.

    nitmon : int
        Indicates the amount of information on the iteration that is
        printed.

        `nitmon` > 0

            The value of A, theta and delta (see [Accuracy]) will be
            printed at the first and every `nitmon` iterations.

        `nitmon` <= 0

            No iteration monitoring is printed.

        When printing occurs the output is directed to the file object
        associated with the advisory I/O unit (see
        :class:`~naginterfaces.base.utils.FileObjManager`).

    tol : float
        The relative precision for the final estimates of the covariance
        matrix.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    nit : int
        The number of iterations performed.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `maxit` = *<value>*.

            Constraint: `maxit` > 0.

        (`errno` 1)
            On entry, `tol` = *<value>*.

            Constraint: `tol` > 0.0.

        (`errno` 1)
            On entry, `eps` = *<value>*.

            Constraint: 0.0 <= `eps` <= 1.0.

        (`errno` 1)
            On entry, `n` = *<value>* and `m` = *<value>*.

            Constraint: `n` >= `m`.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 2.

        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 2)
            On entry, a variable has a constant value, i.e., all
            elements in column *<value>* of `x` are identical.

        (`errno` 3)
            The iterative procedure to find C has failed to converge in
            `maxit` iterations.

        (`errno` 4)
            The iterative procedure to find C has become unstable. This
            may happen if the value of `eps` is too large for the
            sample.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Huber, P J, 1981, `Robust Statistics`, Wiley

    Marazzi, A, 1987, `Weights for bounded influence regression in
    ROBETH`, Cah. Rech. Doc. IUMSP, No. 3 ROB 3, Institut Universitaire
    de Médecine Sociale et Préventive, Lausanne
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.robustm_corr_huber'

    utils.FileObjManager._check_type_compat(
        io_manager, {'fun_name': fun_name, 'entity': 'io_manager'}
    )

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    def cb_print_rech_py(_, en_data, nout, rec, ifail, rec_len):
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data, utils._EngineIntScalarType.to_py(nout.contents), rec[:rec_len], io_manager, excs, ifail.contents,
                fun_name=fun_name,
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02hkft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._PRINT_REC_FUNCTYPE,
        utils._PRINT_RECH_FUNCTYPE,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Input, x[ldx, m].
        utils._BYREF_INT, # Input, ldx.
        utils._BYREF_FLOAT64, # Input, eps.
        utils._EngineFloat64ArrayType.c_type, # Output, cov[m*(m+1)//2].
        utils._EngineFloat64ArrayType.c_type, # Output, theta[m].
        utils._BYREF_INT, # Input, maxit.
        utils._BYREF_INT, # Input, nitmon.
        utils._BYREF_INT, # Input, nadv.
        utils._BYREF_FLOAT64, # Input, tol.
        utils._BYREF_INT, # Output, nit.
        utils._EngineFloat64ArrayType.c_type, # Workspace, wk.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((x, 'x'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _print_rech = utils._PRINT_RECH_FUNCTYPE(cb_print_rech_py)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _ldx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py, _m_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _eps = utils._EngineFloat64ScalarType(
        eps, {'fun_name': fun_name, 'entity': 'eps'},
    ).to_ctypes()
    _cov_cls = utils._EngineFloat64ArrayType(
        cov, {'fun_name': fun_name, 'entity': 'cov'},
        exp_shape=(_m_py*(_m_py+1)//2,),
        check_contents=False,
    )
    _cov = _cov_cls.to_ctypes(can_reference_contents=False)
    _theta_cls = utils._EngineFloat64ArrayType(
        theta, {'fun_name': fun_name, 'entity': 'theta'},
        exp_shape=(_m_py,),
        check_contents=False,
    )
    _theta = _theta_cls.to_ctypes(can_reference_contents=False)
    _maxit = utils._EngineIntScalarType(
        maxit, {'fun_name': fun_name, 'entity': 'maxit'},
    ).to_ctypes()
    _nitmon = utils._EngineIntScalarType(
        nitmon, {'fun_name': fun_name, 'entity': 'nitmon'},
    ).to_ctypes()
    _nadv = utils._EngineIntScalarType.c_type(utils.FileObjManager.advunit)
    _tol = utils._EngineFloat64ScalarType(
        tol, {'fun_name': fun_name, 'entity': 'tol'},
    ).to_ctypes()
    _nit = utils._EngineIntScalarType.empty_ctypes()
    _wk = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        utils._UNUSED_PRINT_REC,
        _print_rech,
        _n,
        _m,
        _x,
        _ldx,
        _eps,
        _cov,
        _theta,
        _maxit,
        _nitmon,
        _nadv,
        _tol,
        _nit,
        _wk,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, maxit = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: maxit > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, tol = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: tol > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, eps = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: 0.0 <= eps <= 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s} and m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: n >= m.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a variable has a constant value, i.e., all elements in column',
                    '{:s} of x are identical.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The iterative procedure to find C has failed to converge',
                    'in maxit iterations.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The iterative procedure to find C has become unstable.',
                    'This may happen if the value of eps is too large for the sample.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _cov_cls.refresh_py(_cov)
        _theta_cls.refresh_py(_theta)
        _nit_py = utils._EngineIntScalarType.to_py(_nit)
        return_data = _nit_py
    _err_class.manage()
    return return_data

def robustm_corr_user_deriv(ucv, indm, n, m, x, cov, a, wt, theta, bl, bd, maxit, nitmon, tol, data=None, io_manager=None):
    u"""
    Calculates a robust estimation of a covariance matrix, user-supplied
    weight function plus derivatives.

    ``robustm_corr_user_deriv`` calculates a robust estimate of the
    covariance matrix for user-supplied weight functions and their
    derivatives.

    For full information please refer to the NAG Library document for
    g02hl

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02hlf.html

    Parameters
    ----------
    ucv : callable (u, ud, w, wd) = ucv(t, data=None)
        `ucv` must return the values of the functions u and w and their
        derivatives for a given value of its argument.

        Parameters
        ~~~~~~~~~~
        t : float
            The argument for which the functions u and w must be
            evaluated.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        u : float
            The value of the u function at the point `t`.

        ud : float
            The value of the derivative of the u function at the point
            `t`.

        w : float
            The value of the w function at the point `t`.

        wd : float
            The value of the derivative of the w function at the point
            `t`.

    indm : int
        Indicates which form of the function v will be used.

        `indm` = 1

            v = 1.

        `indm` != 1

            v = u.

    n : int
        n, the number of observations.

    m : int
        m, the number of columns of the matrix X, i.e., number of
        independent variables.

    x : float, shape (`n`, `m`)
        `x`[i-1,j-1] must contain the ith observation on the jth
        variable, for j = 1,2,...,m, for i = 1,2,...,n.

    cov : float, shape (`m`*(`m`+1)/2), modified in place
        `On exit`: contains a robust estimate of the covariance matrix,
        C. The upper triangular part of the matrix C is stored packed by
        columns (lower triangular stored by rows), C_{ij} is returned in
        `cov`[(j*(j-1)/2+i)-1], i <= j.

    a : float, shape (`m`*(`m`+1)/2), modified in place
        `On entry`: an initial estimate of the lower triangular real
        matrix A. Only the lower triangular elements must be given and
        these should be stored row-wise in the array.

        The diagonal elements must be != 0, and in practice will usually
        be > 0.

        If the magnitudes of the columns of X are of the same order, the
        identity matrix will often provide a suitable initial value for
        A.

        If the columns of X are of different magnitudes, the diagonal
        elements of the initial value of A should be approximately
        inversely proportional to the magnitude of the columns of X.

        `On exit`: the lower triangular elements of the inverse of the
        matrix A, stored row-wise.

    wt : float, shape (`n`), modified in place
        `On exit`: `wt`[i-1] contains the weights, wt_i = u(||z_i||_2),
        for i = 1,2,...,n.

    theta : float, shape (`m`), modified in place
        `On entry`: an initial estimate of the location parameter,
        theta_j, for j = 1,2,...,m.

        In many cases an initial estimate of theta_j = 0, for j =
        1,2,...,m, will be adequate.

        Alternatively medians may be used as given by
        ``univar.robust_1var_median``.

        `On exit`: contains the robust estimate of the location
        parameter, theta_j, for j = 1,2,...,m.

    bl : float
        The magnitude of the bound for the off-diagonal elements of S_k,
        BL.

        Suggested value: `bl` = 0.9.

    bd : float
        The magnitude of the bound for the diagonal elements of S_k, BD.

        Suggested value: `bd` = 0.9.

    maxit : int
        The maximum number of iterations that will be used during the
        calculation of A.

        Suggested value: `maxit` = 150.

    nitmon : int
        Indicates the amount of information on the iteration that is
        printed.

        `nitmon` > 0

            The value of A, theta and delta (see [Accuracy]) will be
            printed at the first and every `nitmon` iterations.

        `nitmon` <= 0

            No iteration monitoring is printed.

        When printing occurs the output is directed to the file object
        associated with the advisory I/O unit (see
        :class:`~naginterfaces.base.utils.FileObjManager`).

    tol : float
        The relative precision for the final estimates of the covariance
        matrix. Iteration will stop when maximum delta (see [Accuracy])
        is less than `tol`.

    data : arbitrary, optional
        User-communication data for callback functions.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    nit : int
        The number of iterations performed.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>* and `m` = *<value>*.

            Constraint: `n` >= `m`.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 2.

        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 2)
            On entry, `bd` = *<value>*.

            Constraint: `bd` > 0.0.

        (`errno` 2)
            On entry, `bl` = *<value>*.

            Constraint: `bl` > 0.0.

        (`errno` 2)
            On entry, `maxit` = *<value>*.

            Constraint: `maxit` > 0.

        (`errno` 2)
            On entry, `tol` = *<value>*.

            Constraint: `tol` > 0.0.

        (`errno` 2)
            On entry, i = *<value>* and the ith diagonal element of A is
            0.

            Constraint: all diagonal elements of A must be non-zero.

        (`errno` 3)
            On entry, a variable has a constant value, i.e., all
            elements in column *<value>* of `x` are identical.

        (`errno` 4)
            w value returned by `ucv` < 0.0: w(*<value>*) = *<value>*.

            Constraint: `w` >= 0.0.

        (`errno` 4)
            u value returned by `ucv` < 0.0: u(*<value>*) = *<value>*.

            Constraint: `u` >= 0.0.

        (`errno` 5)
            Iterations to calculate weights failed to converge.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 6)
            The sum D_2 is zero. Try either a larger initial estimate of
            A or make u and w less strict.

        (`errno` 6)
            The sum D_3 is zero. Try either a larger initial estimate of
            A or make u and w less strict.

        (`errno` 6)
            The sum D_1 is zero. Try either a larger initial estimate of
            A or make u and w less strict.

    References
    ----------
    Huber, P J, 1981, `Robust Statistics`, Wiley

    Marazzi, A, 1987, `Weights for bounded influence regression in
    ROBETH`, Cah. Rech. Doc. IUMSP, No. 3 ROB 3, Institut Universitaire
    de Médecine Sociale et Préventive, Lausanne
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.robustm_corr_user_deriv'

    utils.FileObjManager._check_type_compat(
        io_manager, {'fun_name': fun_name, 'entity': 'io_manager'}
    )

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    def cb_print_rech_py(_, en_data, nout, rec, ifail, rec_len):
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data, utils._EngineIntScalarType.to_py(nout.contents), rec[:rec_len], io_manager, excs, ifail.contents,
                fun_name=fun_name,
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    cb_ucv_name = fun_name + ':ucv'

    ucvh = utils._EngineState._sys_functype(
        None,
        utils._UNUSED_FUNCTYPE,
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, t.
        utils._BYREF_FLOAT64, # Output, u.
        utils._BYREF_FLOAT64, # Output, ud.
        utils._BYREF_FLOAT64, # Output, w.
        utils._BYREF_FLOAT64, # Output, wd.
        utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
    )

    def cb_ucvh(_ucv, en_data, t, u, ud, w, wd, _iuser, _ruser):
        """Helper for ucv."""
        try:
            _t_ucvh = utils._EngineFloat64ScalarType.to_py(t.contents)
            if data is not None:
                _u_ucvh, _ud_ucvh, _w_ucvh, _wd_ucvh = ucv(
                    _t_ucvh,
                    data,
                )
            else:
                _u_ucvh, _ud_ucvh, _w_ucvh, _wd_ucvh = ucv(
                    _t_ucvh,
                )
            utils._EngineFloat64ScalarType(
                _u_ucvh, {'fun_name': cb_ucv_name, 'returned': 'u'},
            ).refresh_ctypes_buffer(u)
            utils._EngineFloat64ScalarType(
                _ud_ucvh, {'fun_name': cb_ucv_name, 'returned': 'ud'},
            ).refresh_ctypes_buffer(ud)
            utils._EngineFloat64ScalarType(
                _w_ucvh, {'fun_name': cb_ucv_name, 'returned': 'w'},
            ).refresh_ctypes_buffer(w)
            utils._EngineFloat64ScalarType(
                _wd_ucvh, {'fun_name': cb_ucv_name, 'returned': 'wd'},
            ).refresh_ctypes_buffer(wd)
        except:
            excs.append(_sys.exc_info())
            en_data.contents.hlperr = utils._IERR_HELPER

    engine_ctypes = utils._EngineState._get_symbol('g02hlft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._PRINT_REC_FUNCTYPE,
        utils._PRINT_RECH_FUNCTYPE,
        utils._UNUSED_FUNCTYPE,
        ucvh,
        utils._BYREF_INT, # Input, indm.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Input, x[ldx, m].
        utils._BYREF_INT, # Input, ldx.
        utils._EngineFloat64ArrayType.c_type, # Output, cov[m*(m+1)//2].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, a[m*(m+1)//2].
        utils._EngineFloat64ArrayType.c_type, # Output, wt[n].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, theta[m].
        utils._BYREF_FLOAT64, # Input, bl.
        utils._BYREF_FLOAT64, # Input, bd.
        utils._BYREF_INT, # Input, maxit.
        utils._BYREF_INT, # Input, nitmon.
        utils._BYREF_INT, # Input, nadv.
        utils._BYREF_FLOAT64, # Input, tol.
        utils._BYREF_INT, # Output, nit.
        utils._EngineFloat64ArrayType.c_type, # Workspace, wk.
        utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((x, 'x'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _print_rech = utils._PRINT_RECH_FUNCTYPE(cb_print_rech_py)
    utils._check_callable(ucv, {'fun_name': fun_name, 'entity': 'ucv'})
    _ucv = utils._UNUSED_CALLBACK
    _ucvh = utils._cast_callback(
        ucvh, cb_ucvh,
        {'fun_name': fun_name, 'entity': 'ucv'},
    )
    _indm = utils._EngineIntScalarType(
        indm, {'fun_name': fun_name, 'entity': 'indm'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _ldx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py, _m_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _cov_cls = utils._EngineFloat64ArrayType(
        cov, {'fun_name': fun_name, 'entity': 'cov'},
        exp_shape=(_m_py*(_m_py+1)//2,),
        check_contents=False,
    )
    _cov = _cov_cls.to_ctypes(can_reference_contents=False)
    _a_cls = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_m_py*(_m_py+1)//2,),
    )
    _a = _a_cls.to_ctypes()
    _wt_cls = utils._EngineFloat64ArrayType(
        wt, {'fun_name': fun_name, 'entity': 'wt'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _wt = _wt_cls.to_ctypes(can_reference_contents=False)
    _theta_cls = utils._EngineFloat64ArrayType(
        theta, {'fun_name': fun_name, 'entity': 'theta'},
        exp_shape=(_m_py,),
    )
    _theta = _theta_cls.to_ctypes()
    _bl = utils._EngineFloat64ScalarType(
        bl, {'fun_name': fun_name, 'entity': 'bl'},
    ).to_ctypes()
    _bd = utils._EngineFloat64ScalarType(
        bd, {'fun_name': fun_name, 'entity': 'bd'},
    ).to_ctypes()
    _maxit = utils._EngineIntScalarType(
        maxit, {'fun_name': fun_name, 'entity': 'maxit'},
    ).to_ctypes()
    _nitmon = utils._EngineIntScalarType(
        nitmon, {'fun_name': fun_name, 'entity': 'nitmon'},
    ).to_ctypes()
    _nadv = utils._EngineIntScalarType.c_type(utils.FileObjManager.advunit)
    _tol = utils._EngineFloat64ScalarType(
        tol, {'fun_name': fun_name, 'entity': 'tol'},
    ).to_ctypes()
    _nit = utils._EngineIntScalarType.empty_ctypes()
    _wk = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _iuser = None
    _ruser = None
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        utils._UNUSED_PRINT_REC,
        _print_rech,
        _ucv,
        _ucvh,
        _indm,
        _n,
        _m,
        _x,
        _ldx,
        _cov,
        _a,
        _wt,
        _theta,
        _bl,
        _bd,
        _maxit,
        _nitmon,
        _nadv,
        _tol,
        _nit,
        _wk,
        _iuser,
        _ruser,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s} and m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: n >= m.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, bd = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: bd > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, bl = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: bl > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, maxit = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: maxit > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, tol = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: tol > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s} and the ith diagonal element of A is 0.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: all diagonal elements of A must be non-zero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99979:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a variable has a constant value, i.e., all elements in column',
                    '{:s} of x are identical.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99981:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'w value returned by ucv < 0.0:',
                    'w({:s}) = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: w >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'u value returned by ucv < 0.0:',
                    'u({:s}) = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: u >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Iterations to calculate weights failed to converge.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99977:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The sum D_2 is zero.',
                    'Try either a larger initial estimate of A or make u and w less strict.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99978:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The sum D_3 is zero.',
                    'Try either a larger initial estimate of A or make u and w less strict.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99980:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The sum D_1 is zero.',
                    'Try either a larger initial estimate of A or make u and w less strict.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _cov_cls.refresh_py(_cov)
        _a_cls.refresh_py(_a)
        _wt_cls.refresh_py(_wt)
        _theta_cls.refresh_py(_theta)
        _nit_py = utils._EngineIntScalarType.to_py(_nit)
        return_data = _nit_py
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = _collections.namedtuple(
                'nag_return_data',
                ['nit'],
            )(
                _nit_py,
            )
    _err_class.manage()
    return return_data

def robustm_corr_user(ucv, indm, n, m, x, cov, a, wt, theta, bl, bd, maxit, nitmon, tol, data=None, io_manager=None):
    u"""
    Calculates a robust estimation of a covariance matrix, user-supplied
    weight function.

    ``robustm_corr_user`` computes a robust estimate of the covariance
    matrix for user-supplied weight functions.
    The derivatives of the weight functions are not required.

    For full information please refer to the NAG Library document for
    g02hm

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02hmf.html

    Parameters
    ----------
    ucv : callable (u, w) = ucv(t, data=None)
        `ucv` must return the values of the functions u and w for a
        given value of its argument.

        Parameters
        ~~~~~~~~~~
        t : float
            The argument for which the functions u and w must be
            evaluated.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        u : float
            The value of the u function at the point `t`.

        w : float
            The value of the w function at the point `t`.

    indm : int
        Indicates which form of the function v will be used.

        `indm` = 1

            v = 1.

        `indm` != 1

            v = u.

    n : int
        n, the number of observations.

    m : int
        m, the number of columns of the matrix X, i.e., number of
        independent variables.

    x : float, shape (`n`, `m`)
        `x`[i-1,j-1] must contain the ith observation on the jth
        variable, for j = 1,2,...,m, for i = 1,2,...,n.

    cov : float, shape (`m`*(`m`+1)/2), modified in place
        `On exit`: a robust estimate of the covariance matrix, C. The
        upper triangular part of the matrix C is stored packed by
        columns (lower triangular stored by rows), that is C_{ij} is
        returned in `cov`[j*(j-1)/2+i-1], i <= j.

    a : float, shape (`m`*(`m`+1)/2), modified in place
        `On entry`: an initial estimate of the lower triangular real
        matrix A. Only the lower triangular elements must be given and
        these should be stored row-wise in the array.

        The diagonal elements must be != 0, and in practice will usually
        be > 0.

        If the magnitudes of the columns of X are of the same order, the
        identity matrix will often provide a suitable initial value for
        A.

        If the columns of X are of different magnitudes, the diagonal
        elements of the initial value of A should be approximately
        inversely proportional to the magnitude of the columns of X.

        `On exit`: the lower triangular elements of the inverse of the
        matrix A, stored row-wise.

    wt : float, shape (`n`), modified in place
        `On exit`: `wt`[i-1] contains the weights, wt_i = u(||z_i||_2),
        for i = 1,2,...,n.

    theta : float, shape (`m`), modified in place
        `On entry`: an initial estimate of the location parameter,
        theta_j, for j = 1,2,...,m.

        In many cases an initial estimate of theta_j = 0, for j =
        1,2,...,m, will be adequate.

        Alternatively medians may be used as given by
        ``univar.robust_1var_median``.

        `On exit`: contains the robust estimate of the location
        parameter, theta_j, for j = 1,2,...,m.

    bl : float
        The magnitude of the bound for the off-diagonal elements of S_k,
        BL.

        Suggested value: `bl` = 0.9.

    bd : float
        The magnitude of the bound for the diagonal elements of S_k, BD.

        Suggested value: `bd` = 0.9.

    maxit : int
        The maximum number of iterations that will be used during the
        calculation of A.

        Suggested value: `maxit` = 150.

    nitmon : int
        Indicates the amount of information on the iteration that is
        printed.

        `nitmon` > 0

            The value of A, theta and delta (see [Accuracy]) will be
            printed at the first and every `nitmon` iterations.

        `nitmon` <= 0

            No iteration monitoring is printed.

    tol : float
        The relative precision for the final estimate of the covariance
        matrix. Iteration will stop when maximum delta (see [Accuracy])
        is less than `tol`.

    data : arbitrary, optional
        User-communication data for callback functions.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    nit : int
        The number of iterations performed.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>* and `m` = *<value>*.

            Constraint: `n` >= `m`.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 2.

        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 2)
            On entry, `bd` = *<value>*.

            Constraint: `bd` > 0.0.

        (`errno` 2)
            On entry, `bl` = *<value>*.

            Constraint: `bl` > 0.0.

        (`errno` 2)
            On entry, `maxit` = *<value>*.

            Constraint: `maxit` > 0.

        (`errno` 2)
            On entry, `tol` = *<value>*.

            Constraint: `tol` > 0.0.

        (`errno` 2)
            On entry, i = *<value>* and the ith diagonal element of A is
            0.

            Constraint: all diagonal elements of A must be non-zero.

        (`errno` 3)
            On entry, a variable has a constant value, i.e., all
            elements in column *<value>* of `x` are identical.

        (`errno` 4)
            w value returned by `ucv` < 0.0: w(*<value>*) = *<value>*.

            Constraint: `w` >= 0.0.

        (`errno` 4)
            u value returned by `ucv` < 0.0: u(*<value>*) = *<value>*.

            Constraint: `u` >= 0.0.

        (`errno` 5)
            Iterations to calculate weights failed to converge.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 6)
            The sum D_2 is zero. Try either a larger initial estimate of
            A or make u and w less strict.

        (`errno` 6)
            The sum D_1 is zero. Try either a larger initial estimate of
            A or make u and w less strict.

    References
    ----------
    Huber, P J, 1981, `Robust Statistics`, Wiley

    Marazzi, A, 1987, `Weights for bounded influence regression in
    ROBETH`, Cah. Rech. Doc. IUMSP, No. 3 ROB 3, Institut Universitaire
    de Médecine Sociale et Préventive, Lausanne
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.robustm_corr_user'

    utils.FileObjManager._check_type_compat(
        io_manager, {'fun_name': fun_name, 'entity': 'io_manager'}
    )

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    def cb_print_rech_py(_, en_data, nout, rec, ifail, rec_len):
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data, utils._EngineIntScalarType.to_py(nout.contents), rec[:rec_len], io_manager, excs, ifail.contents,
                fun_name=fun_name,
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    cb_ucv_name = fun_name + ':ucv'

    ucvh = utils._EngineState._sys_functype(
        None,
        utils._UNUSED_FUNCTYPE,
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, t.
        utils._BYREF_FLOAT64, # Output, u.
        utils._BYREF_FLOAT64, # Output, w.
        utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
    )

    def cb_ucvh(_ucv, en_data, t, u, w, _iuser, _ruser):
        """Helper for ucv."""
        try:
            _t_ucvh = utils._EngineFloat64ScalarType.to_py(t.contents)
            if data is not None:
                _u_ucvh, _w_ucvh = ucv(
                    _t_ucvh,
                    data,
                )
            else:
                _u_ucvh, _w_ucvh = ucv(
                    _t_ucvh,
                )
            utils._EngineFloat64ScalarType(
                _u_ucvh, {'fun_name': cb_ucv_name, 'returned': 'u'},
            ).refresh_ctypes_buffer(u)
            utils._EngineFloat64ScalarType(
                _w_ucvh, {'fun_name': cb_ucv_name, 'returned': 'w'},
            ).refresh_ctypes_buffer(w)
        except:
            excs.append(_sys.exc_info())
            en_data.contents.hlperr = utils._IERR_HELPER

    engine_ctypes = utils._EngineState._get_symbol('g02hmft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._PRINT_REC_FUNCTYPE,
        utils._PRINT_RECH_FUNCTYPE,
        utils._UNUSED_FUNCTYPE,
        ucvh,
        utils._BYREF_INT, # Input, indm.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Input, x[ldx, m].
        utils._BYREF_INT, # Input, ldx.
        utils._EngineFloat64ArrayType.c_type, # Output, cov[m*(m+1)//2].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, a[m*(m+1)//2].
        utils._EngineFloat64ArrayType.c_type, # Output, wt[n].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, theta[m].
        utils._BYREF_FLOAT64, # Input, bl.
        utils._BYREF_FLOAT64, # Input, bd.
        utils._BYREF_INT, # Input, maxit.
        utils._BYREF_INT, # Input, nitmon.
        utils._BYREF_INT, # Input, nadv.
        utils._BYREF_FLOAT64, # Input, tol.
        utils._BYREF_INT, # Output, nit.
        utils._EngineFloat64ArrayType.c_type, # Workspace, wk.
        utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((x, 'x'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _print_rech = utils._PRINT_RECH_FUNCTYPE(cb_print_rech_py)
    utils._check_callable(ucv, {'fun_name': fun_name, 'entity': 'ucv'})
    _ucv = utils._UNUSED_CALLBACK
    _ucvh = utils._cast_callback(
        ucvh, cb_ucvh,
        {'fun_name': fun_name, 'entity': 'ucv'},
    )
    _indm = utils._EngineIntScalarType(
        indm, {'fun_name': fun_name, 'entity': 'indm'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _ldx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py, _m_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _cov_cls = utils._EngineFloat64ArrayType(
        cov, {'fun_name': fun_name, 'entity': 'cov'},
        exp_shape=(_m_py*(_m_py+1)//2,),
        check_contents=False,
    )
    _cov = _cov_cls.to_ctypes(can_reference_contents=False)
    _a_cls = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_m_py*(_m_py+1)//2,),
    )
    _a = _a_cls.to_ctypes()
    _wt_cls = utils._EngineFloat64ArrayType(
        wt, {'fun_name': fun_name, 'entity': 'wt'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _wt = _wt_cls.to_ctypes(can_reference_contents=False)
    _theta_cls = utils._EngineFloat64ArrayType(
        theta, {'fun_name': fun_name, 'entity': 'theta'},
        exp_shape=(_m_py,),
    )
    _theta = _theta_cls.to_ctypes()
    _bl = utils._EngineFloat64ScalarType(
        bl, {'fun_name': fun_name, 'entity': 'bl'},
    ).to_ctypes()
    _bd = utils._EngineFloat64ScalarType(
        bd, {'fun_name': fun_name, 'entity': 'bd'},
    ).to_ctypes()
    _maxit = utils._EngineIntScalarType(
        maxit, {'fun_name': fun_name, 'entity': 'maxit'},
    ).to_ctypes()
    _nitmon = utils._EngineIntScalarType(
        nitmon, {'fun_name': fun_name, 'entity': 'nitmon'},
    ).to_ctypes()
    _nadv = utils._EngineIntScalarType.c_type(utils.FileObjManager.advunit)
    _tol = utils._EngineFloat64ScalarType(
        tol, {'fun_name': fun_name, 'entity': 'tol'},
    ).to_ctypes()
    _nit = utils._EngineIntScalarType.empty_ctypes()
    _wk = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _iuser = None
    _ruser = None
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        utils._UNUSED_PRINT_REC,
        _print_rech,
        _ucv,
        _ucvh,
        _indm,
        _n,
        _m,
        _x,
        _ldx,
        _cov,
        _a,
        _wt,
        _theta,
        _bl,
        _bd,
        _maxit,
        _nitmon,
        _nadv,
        _tol,
        _nit,
        _wk,
        _iuser,
        _ruser,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s} and m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: n >= m.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, bd = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: bd > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, bl = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: bl > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, maxit = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: maxit > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, tol = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: tol > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s} and the ith diagonal element of A is 0.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: all diagonal elements of A must be non-zero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99979:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a variable has a constant value, i.e., all elements in column',
                    '{:s} of x are identical.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99981:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'w value returned by ucv < 0.0:',
                    'w({:s}) = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: w >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'u value returned by ucv < 0.0:',
                    'u({:s}) = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: u >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Iterations to calculate weights failed to converge.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99978:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The sum D_2 is zero.',
                    'Try either a larger initial estimate of A or',
                    'make u and w less strict.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99980:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The sum D_1 is zero.',
                    'Try either a larger initial estimate of A',
                    'or make u and w less strict.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _cov_cls.refresh_py(_cov)
        _a_cls.refresh_py(_a)
        _wt_cls.refresh_py(_wt)
        _theta_cls.refresh_py(_theta)
        _nit_py = utils._EngineIntScalarType.to_py(_nit)
        return_data = _nit_py
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = _collections.namedtuple(
                'nag_return_data',
                ['nit'],
            )(
                _nit_py,
            )
    _err_class.manage()
    return return_data

@utils._withdrawal(
    'naginterfaces.base.correg.mixeff_reml',
    mkwdrn=None,
    advice=[
        'Please use lmm_init followed by lmm_fit instead.',
        'See also https://www.nag.com/numeric/py/nagdoc_latest/replace.html',
    ]
)
def mixeff_reml(n, ncol, dat, levels, yvid, cwid, nfv, fvid, fint, nrv, rvid, nvpr, vpr, rint, svid, gamma, lb, b, se, maxit, tol, io_manager=None):
    u"""
    Linear mixed effects regression using Restricted Maximum Likelihood
    (REML).

    ``mixeff_reml`` fits a linear mixed effects regression model using
    restricted maximum likelihood (REML).

    Note: ``mixeff_reml`` is deprecated.
    Please use ``lmm_init`` followed by ``lmm_fit`` instead.
    See also https://www.nag.com/numeric/py/nagdoc_latest/replace.html

    For full information please refer to the NAG Library document for
    g02ja

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02jaf.html

    Parameters
    ----------
    n : int
        n, the number of observations.

    ncol : int
        The number of columns in the data matrix, `dat`.

    dat : float, shape (`n`, `ncol`)
        Array containing all of the data. For the ith observation:

            `dat`[i-1,`yvid`-1] holds the dependent variable, y;

            if `cwid` != 0, `dat`[i-1,`cwid`-1] holds the case weights;

            if `svid` != 0, `dat`[i-1,`svid`-1] holds the subject
            variable.

        The remaining columns hold the values of the independent
        variables.

    levels : int, shape (`ncol`)
        `levels`[i-1] contains the number of levels associated with the
        ith variable of the data matrix `dat`. If this variable is
        continuous or binary (i.e., only takes the values zero or one)
        then `levels`[i-1] should be 1; if the variable is discrete then
        `levels`[i-1] is the number of levels associated with it and
        `dat`[j-1,i-1] is assumed to take the values 1 to `levels`[i-1],
        for j = 1,2,...,`n`.

    yvid : int
        The column of `dat` holding the dependent, y, variable.

    cwid : int
        The column of `dat` holding the case weights.

        If `cwid` = 0, no weights are used.

    nfv : int
        The number of independent variables in the model which are to be
        treated as being fixed.

    fvid : int, shape (`nfv`)
        The columns of the data matrix `dat` holding the fixed
        independent variables with `fvid`[i-1] holding the column number
        corresponding to the ith fixed variable.

    fint : int
        Flag indicating whether a fixed intercept is included (`fint` =
        1).

    nrv : int
        The number of independent variables in the model which are to be
        treated as being random.

    rvid : int, shape (`nrv`)
        The columns of the data matrix `dat` holding the random
        independent variables with `rvid`[i-1] holding the column number
        corresponding to the ith random variable.

    nvpr : int
        If `rint` = 1 and `svid` != 0, `nvpr` is the number of variance
        components being estimated-2, (g-1), else `nvpr` = g.

        If `nrv` = 0, `nvpr` is not referenced.

    vpr : int, shape (`nrv`)
        `vpr`[i-1] holds a flag indicating the variance of the ith
        random variable. The variance of the ith random variable is
        sigma_j^2, where j = `vpr`[i-1]+1 if `rint` = 1 and `svid` != 0
        and j = `vpr`[i-1] otherwise. Random variables with the same
        value of j are assumed to be taken from the same distribution.

    rint : int
        Flag indicating whether a random intercept is included (`rint` =
        1).

        If `svid` = 0, `rint` is not referenced.

    svid : int
        The column of `dat` holding the subject variable.

        If `svid` = 0, no subject variable is used.

        Specifying a subject variable is equivalent to specifying the
        interaction between that variable and all of the random-effects.

        Letting the notation Z_1*Z_S denote the interaction between
        variables Z_1 and Z_S, fitting a model with `rint` = 0,
        random-effects Z_1+Z_2 and subject variable Z_S is equivalent to
        fitting a model with random-effects Z_1*Z_S+Z_2*Z_S and no
        subject variable.

        If `rint` = 1 the model is equivalent to fitting
        Z_S+Z_1*Z_S+Z_2*Z_S and no subject variable.

    gamma : float, shape (`nvpr`+2), modified in place
        `On entry`: holds the initial values of the variance components,
        gamma_0, with `gamma`[i-1] the initial value for
        sigma_i^2/sigma_R^2, for i = 1,2,...,g. If `rint` = 1 and `svid`
        != 0, g = `nvpr`+1, else g = `nvpr`.

        If `gamma`[0] = -1.0, the remaining elements of `gamma` are
        ignored and the initial values for the variance components are
        estimated from the data using MIVQUE0.

        `On exit`: `gamma`[i-1], for i = 1,2,...,g, holds the final
        estimate of sigma_i^2 and `gamma`[g] holds the final estimate
        for sigma_R^2.

    lb : int
        The size of the array `b`.

    b : float, shape (`lb`), modified in place
        `On exit`: the parameter estimates, (beta,nu), with the first
        `nff` elements of `b` containing the fixed effect parameter
        estimates, beta and the next `nrf` elements of `b` containing
        the random effect parameter estimates, nu.

        **Fixed effects**

        If `fint` = 1, `b`[0] contains the estimate of the fixed
        intercept.

        Let L_i denote the number of levels associated with the ith
        fixed variable, that is L_i = `levels`[ `fvid`[i-1] -1].

        Define

            if `fint` = 1, F_1 = 2 else if `fint` = 0, F_1 = 1;

            F_{i+1} = F_i+max(L_i-1,1), i >= 1.

        Then for i = 1,2,...,`nfv`:

            if L_i > 1, `b`[ F_i+j-2 -1] contains the parameter estimate
            for the jth level of the ith fixed variable, for j =
            2,3,...,L_i;

            if L_i <= 1, `b`[F_i-1] contains the parameter estimate for
            the ith fixed variable.

        **Random effects**

        Redefining L_i to denote the number of levels associated with
        the ith random variable, that is L_i = `levels`[ `rvid`[i-1]
        -1].

        Define

            if `rint` = 1, R_1 = 2 else if `rint` = 0, R_1 = 1;

            R_{i+1} = R_i+L_i, i >= 1.

        Then for i = 1,2,...,`nrv`:

            if `svid` = 0,

                if L_i > 1, `b`[ `nff` + R_i +j-1 -1] contains the
                parameter estimate for the jth level of the ith random
                variable, for j = 1,2,...,L_i;

                if L_i <= 1, `b`[ `nff` + R_i -1] contains the parameter
                estimate for the ith random variable;

            if `svid` != 0,

                let L_S denote the number of levels associated with the
                subject variable, that is L_S = `levels`[ `svid` -1];

                if L_i > 1, `b`[ `nff` + (s-1) L_S + R_i + j - 1 -1]
                contains the parameter estimate for the interaction
                between the sth level of the subject variable and the
                jth level of the ith random variable, for j =
                1,2,...,L_i, for s = 1,2,...,L_S;

                if L_i <= 1, `b`[ `nff` + (s-1) L_S + R_i -1] contains
                the parameter estimate for the interaction between the
                sth level of the subject variable and the ith random
                variable, for s = 1,2,...,L_S;

                if `rint` = 1, `b`[`nff`] contains the estimate of the
                random intercept.

    se : float, shape (`lb`), modified in place
        `On exit`: the standard errors of the parameter estimates given
        in `b`.

    maxit : int
        The maximum number of iterations.

        If `maxit` < 0, the default value of 100 is used.

        If `maxit` = 0, the parameter estimates (beta,nu) and
        corresponding standard errors are calculated based on the value
        of gamma_0 supplied in `gamma`.

    tol : float
        The tolerance used to assess convergence.

        If `tol` <= 0.0, the default value of epsilon^0.7 is used, where
        epsilon is the machine precision.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    nff : int
        The number of fixed effects estimated (i.e., the number of
        columns, p, in the design matrix X).

    nrf : int
        The number of random effects estimated (i.e., the number of
        columns, q, in the design matrix Z).

    df : int
        The degrees of freedom.

    reml : float
        -2l_R(gamma^) where l_R is the log of the restricted maximum
        likelihood calculated at gamma^, the estimated variance
        components returned in `gamma`.

    warn : int
        Is set to 1 if a variance component was estimated to be a
        negative value during the fitting process. Otherwise `warn` is
        set to 0.

        If `warn` = 1, the negative estimate is set to zero and the
        estimation process allowed to continue.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: number of observations with nonzero weights must
            be greater than one.

        (`errno` 1)
            On entry, `ncol` = *<value>*.

            Constraint: `ncol` >= 1.

        (`errno` 1)
            On entry, `nfv` = *<value>* and `ncol` = *<value>*.

            Constraint: 0 <= `nfv` < `ncol`.

        (`errno` 1)
            On entry, `nrv` = *<value>* and `ncol` = *<value>*.

            Constraint: 0 <= `nrv` < `ncol` and `nrv`+`rint` > 0.

        (`errno` 1)
            On entry, `nvpr` = *<value>* and `nrv` = *<value>*.

            Constraint: 0 <= `nvpr` <= `nrv` and (`nrv` != 0 or `nvpr`
            >= 1).

        (`errno` 1)
            On entry, `yvid` = *<value>* and `ncol` = *<value>*.

            Constraint: 1 <= `yvid` <= `ncol`.

        (`errno` 1)
            On entry, `svid` = *<value>* and `ncol` = *<value>*.

            Constraint: 0 <= `svid` <= `ncol`.

        (`errno` 1)
            On entry, `cwid` = *<value>* and `ncol` = *<value>*.

            Constraint: 0 <= `cwid` <= `ncol` and any supplied weights
            must be >= 0.0.

        (`errno` 1)
            On entry, `fint` = *<value>*.

            Constraint: `fint` = 0 or 1.

        (`errno` 1)
            On entry, `rint` = *<value>*.

            Constraint: `rint` = 0 or 1.

        (`errno` 1)
            On entry, `lb` too small: `lb` = *<value>*.

        (`errno` 2)
            On entry, `levels`[i] < 1, for at least one i.

        (`errno` 2)
            On entry, `ncol` = *<value>*.

            Constraint: 1 <= `fvid`[i] <= `ncol`, for all i.

        (`errno` 2)
            On entry, `ncol` = *<value>*.

            Constraint: 1 <= `rvid`[i] <= `ncol`, for all i.

        (`errno` 2)
            On entry, `nvpr` = *<value>*.

            Constraint: 1 <= `vpr`[i] <= `nvpr`, for all i.

        (`errno` 2)
            On entry, invalid data: categorical variable with value
            greater than that specified in `levels`.

        (`errno` 2)
            On entry, `gamma`[i] < 0.0, for at least one i.

        (`errno` 3)
            Degrees of freedom < 1: `df` = *<value>*.

        (`errno` 4)
            Routine failed to converge to specified tolerance: `tol` =
            *<value>*.

        (`errno` 4)
            Routine failed to converge in `maxit` iterations: `maxit` =
            *<value>*.

    References
    ----------
    Goodnight, J H, 1979, `A tutorial on the SWEEP operator`, The
    American Statistician (33(3)), 149--158

    Harville, D A, 1977, `Maximum likelihood approaches to variance
    component estimation and to related problems`, JASA (72), 320--340

    Rao, C R, 1972, `Estimation of variance and covariance components in
    a linear model`, J. Am. Stat. Assoc. (67), 112--115

    Stroup, W W, 1989, `Predictable functions and prediction space in
    the mixed model procedure`, Applications of Mixed Models in
    Agriculture and Related Disciplines (Southern Cooperative Series
    Bulletin No. 343), 39--48

    Wolfinger, R, Tobias, R and Sall, J, 1994, `Computing Gaussian
    likelihoods and their derivatives for general linear mixed models`,
    SIAM Sci. Statist. Comput. (15), 1294--1310
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.mixeff_reml'

    utils.FileObjManager._check_type_compat(
        io_manager, {'fun_name': fun_name, 'entity': 'io_manager'}
    )

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    def cb_print_rech_py(_, en_data, nout, rec, ifail, rec_len):
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data, utils._EngineIntScalarType.to_py(nout.contents), rec[:rec_len], io_manager, excs, ifail.contents,
                fun_name=fun_name,
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02jaft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._PRINT_REC_FUNCTYPE,
        utils._PRINT_RECH_FUNCTYPE,
        utils._BYREF_INT, # Input, sorder.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, ncol.
        utils._BYREF_INT, # Input, lddat.
        utils._EngineFloat64ArrayType.c_type, # Input, dat[lddat, *].
        utils._EngineIntArrayType.c_type, # Input, levels[ncol].
        utils._BYREF_INT, # Input, yvid.
        utils._BYREF_INT, # Input, cwid.
        utils._BYREF_INT, # Input, nfv.
        utils._EngineIntArrayType.c_type, # Input, fvid[nfv].
        utils._BYREF_INT, # Input, fint.
        utils._BYREF_INT, # Input, nrv.
        utils._EngineIntArrayType.c_type, # Input, rvid[nrv].
        utils._BYREF_INT, # Input, nvpr.
        utils._EngineIntArrayType.c_type, # Input, vpr[nrv].
        utils._BYREF_INT, # Input, rint.
        utils._BYREF_INT, # Input, svid.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, gamma[nvpr+2].
        utils._BYREF_INT, # Output, nff.
        utils._BYREF_INT, # Output, nrf.
        utils._BYREF_INT, # Output, df.
        utils._BYREF_FLOAT64, # Output, reml.
        utils._BYREF_INT, # Input, lb.
        utils._EngineFloat64ArrayType.c_type, # Output, b[lb].
        utils._EngineFloat64ArrayType.c_type, # Output, se[lb].
        utils._BYREF_INT, # Input, maxit.
        utils._BYREF_FLOAT64, # Input, tol.
        utils._BYREF_INT, # Output, warn.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((dat, 'dat'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _print_rech = utils._PRINT_RECH_FUNCTYPE(cb_print_rech_py)
    _sorder = utils._EngineIntScalarType.c_type(_nag_enum_consts.nag_col_major)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _ncol = utils._EngineIntScalarType(
        ncol, {'fun_name': fun_name, 'entity': 'ncol'},
    ).to_ctypes()
    _lddat_py = utils._EngineFloat64ArrayType.get_nd(dat, {'fun_name': fun_name, 'entity': 'dat'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _lddat = utils._EngineIntScalarType(
        _lddat_py, {'fun_name': fun_name, 'entity': '_lddat'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _sorder_py = utils._EngineIntScalarType.to_py(_sorder)
    _ncol_py = utils._EngineIntScalarType.to_py(_ncol)
    if _sorder_py == 1:
        _dat_dim2 = _ncol_py
    else:
        _dat_dim2 = _n_py
    exp_shape_dat = (_n_py, _dat_dim2)
    _dat = utils._EngineFloat64ArrayType(
        dat, {'fun_name': fun_name, 'entity': 'dat'},
        exp_shape=exp_shape_dat,
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _levels = utils._EngineIntArrayType(
        levels, {'fun_name': fun_name, 'entity': 'levels'},
        exp_shape=(_ncol_py,),
    ).to_ctypes()
    _yvid = utils._EngineIntScalarType(
        yvid, {'fun_name': fun_name, 'entity': 'yvid'},
    ).to_ctypes()
    _cwid = utils._EngineIntScalarType(
        cwid, {'fun_name': fun_name, 'entity': 'cwid'},
    ).to_ctypes()
    _nfv = utils._EngineIntScalarType(
        nfv, {'fun_name': fun_name, 'entity': 'nfv'},
    ).to_ctypes()
    _nfv_py = utils._EngineIntScalarType.to_py(_nfv)
    _fvid = utils._EngineIntArrayType(
        fvid, {'fun_name': fun_name, 'entity': 'fvid'},
        exp_shape=(_nfv_py,),
    ).to_ctypes()
    _fint = utils._EngineIntScalarType(
        fint, {'fun_name': fun_name, 'entity': 'fint'},
    ).to_ctypes()
    _nrv = utils._EngineIntScalarType(
        nrv, {'fun_name': fun_name, 'entity': 'nrv'},
    ).to_ctypes()
    _nrv_py = utils._EngineIntScalarType.to_py(_nrv)
    _rvid = utils._EngineIntArrayType(
        rvid, {'fun_name': fun_name, 'entity': 'rvid'},
        exp_shape=(_nrv_py,),
    ).to_ctypes()
    _nvpr = utils._EngineIntScalarType(
        nvpr, {'fun_name': fun_name, 'entity': 'nvpr'},
    ).to_ctypes()
    _vpr = utils._EngineIntArrayType(
        vpr, {'fun_name': fun_name, 'entity': 'vpr'},
        exp_shape=(_nrv_py,),
    ).to_ctypes()
    _rint = utils._EngineIntScalarType(
        rint, {'fun_name': fun_name, 'entity': 'rint'},
    ).to_ctypes()
    _svid = utils._EngineIntScalarType(
        svid, {'fun_name': fun_name, 'entity': 'svid'},
    ).to_ctypes()
    _nvpr_py = utils._EngineIntScalarType.to_py(_nvpr)
    _gamma_cls = utils._EngineFloat64ArrayType(
        gamma, {'fun_name': fun_name, 'entity': 'gamma'},
        exp_shape=(_nvpr_py+2,),
    )
    _gamma = _gamma_cls.to_ctypes()
    _nff = utils._EngineIntScalarType.empty_ctypes()
    _nrf = utils._EngineIntScalarType.empty_ctypes()
    _df = utils._EngineIntScalarType.empty_ctypes()
    _reml = utils._EngineFloat64ScalarType.empty_ctypes()
    _lb = utils._EngineIntScalarType(
        lb, {'fun_name': fun_name, 'entity': 'lb'},
    ).to_ctypes()
    _lb_py = utils._EngineIntScalarType.to_py(_lb)
    _b_cls = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_lb_py,),
        check_contents=False,
    )
    _b = _b_cls.to_ctypes(can_reference_contents=False)
    _se_cls = utils._EngineFloat64ArrayType(
        se, {'fun_name': fun_name, 'entity': 'se'},
        exp_shape=(_lb_py,),
        check_contents=False,
    )
    _se = _se_cls.to_ctypes(can_reference_contents=False)
    _maxit = utils._EngineIntScalarType(
        maxit, {'fun_name': fun_name, 'entity': 'maxit'},
    ).to_ctypes()
    _tol = utils._EngineFloat64ScalarType(
        tol, {'fun_name': fun_name, 'entity': 'tol'},
    ).to_ctypes()
    _warn = utils._EngineIntScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        utils._UNUSED_PRINT_REC,
        _print_rech,
        _sorder,
        _n,
        _ncol,
        _lddat,
        _dat,
        _levels,
        _yvid,
        _cwid,
        _nfv,
        _fvid,
        _fint,
        _nrv,
        _rvid,
        _nvpr,
        _vpr,
        _rint,
        _svid,
        _gamma,
        _nff,
        _nrf,
        _df,
        _reml,
        _lb,
        _b,
        _se,
        _maxit,
        _tol,
        _warn,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: number of observations with nonzero weights must be greater than',
                    'one.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ncol = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ncol >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nfv = {:s} and ncol = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: 0 <= nfv < ncol.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nrv = {:s} and ncol = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: 0 <= nrv < ncol and nrv+rint > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 6:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nvpr = {:s} and nrv = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: 0 <= nvpr <= nrv and (nrv != 0 or nvpr >= 1).',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 8:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, yvid = {:s} and ncol = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: 1 <= yvid <= ncol.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 11:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, svid = {:s} and ncol = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: 0 <= svid <= ncol.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 12:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, cwid = {:s} and ncol = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: 0 <= cwid <= ncol and any supplied weights must be >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 16:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, fint = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: fint = 0 or 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 17:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, rint = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: rint = 0 or 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 19:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lb too small: lb = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 7:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, levels[i] < 1, for at least one i.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 9:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ncol = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: 1 <= fvid[i] <= ncol, for all i.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 10:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ncol = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: 1 <= rvid[i] <= ncol, for all i.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 13:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nvpr = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: 1 <= vpr[i] <= nvpr, for all i.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 14:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, invalid data: categorical variable with value',
                    'greater than that specified in levels.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 18:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, gamma[i] < 0.0, for at least one i.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 20:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Degrees of freedom < 1: df = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 21:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Routine failed to converge to specified tolerance:',
                    'tol = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 22:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Routine failed to converge in maxit iterations:',
                    'maxit = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _gamma_cls.refresh_py(_gamma)
        _b_cls.refresh_py(_b)
        _se_cls.refresh_py(_se)
        _nff_py = utils._EngineIntScalarType.to_py(_nff)
        _nrf_py = utils._EngineIntScalarType.to_py(_nrf)
        _df_py = utils._EngineIntScalarType.to_py(_df)
        _reml_py = utils._EngineFloat64ScalarType.to_py(_reml)
        _warn_py = utils._EngineIntScalarType.to_py(_warn)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['nff', 'nrf', 'df', 'reml', 'warn'],
        )(
            _nff_py,
            _nrf_py,
            _df_py,
            _reml_py,
            _warn_py,
        )
    _err_class.manage()
    return return_data

@utils._withdrawal(
    'naginterfaces.base.correg.mixeff_ml',
    mkwdrn=None,
    advice=[
        'Please use lmm_init followed by lmm_fit instead.',
        'See also https://www.nag.com/numeric/py/nagdoc_latest/replace.html',
    ]
)
def mixeff_ml(n, ncol, dat, levels, yvid, cwid, nfv, fvid, fint, nrv, rvid, nvpr, vpr, rint, svid, gamma, lb, b, se, maxit, tol, io_manager=None):
    u"""
    Linear mixed effects regression using Maximum Likelihood (ML).

    ``mixeff_ml`` fits a linear mixed effects regression model using
    maximum likelihood (ML).

    Note: ``mixeff_ml`` is deprecated.
    Please use ``lmm_init`` followed by ``lmm_fit`` instead.
    See also https://www.nag.com/numeric/py/nagdoc_latest/replace.html

    For full information please refer to the NAG Library document for
    g02jb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02jbf.html

    Parameters
    ----------
    n : int
        n, the number of observations.

    ncol : int
        The number of columns in the data matrix, `dat`.

    dat : float, shape (`n`, `ncol`)
        Array containing all of the data. For the ith observation:

            `dat`[i-1,`yvid`-1] holds the dependent variable, y;

            if `cwid` != 0, `dat`[i-1,`cwid`-1] holds the case weights;

            if `svid` != 0, `dat`[i-1,`svid`-1] holds the subject
            variable.

        The remaining columns hold the values of the independent
        variables.

    levels : int, shape (`ncol`)
        `levels`[i-1] contains the number of levels associated with the
        ith variable of the data matrix `dat`. If this variable is
        continuous or binary (i.e., only takes the values zero or one)
        then `levels`[i-1] should be 1; if the variable is discrete then
        `levels`[i-1] is the number of levels associated with it and
        `dat`[j-1,i-1] is assumed to take the values 1 to `levels`[i-1],
        for j = 1,2,...,`n`.

    yvid : int
        The column of `dat` holding the dependent, y, variable.

    cwid : int
        The column of `dat` holding the case weights.

        If `cwid` = 0, no weights are used.

    nfv : int
        The number of independent variables in the model which are to be
        treated as being fixed.

    fvid : int, shape (`nfv`)
        The columns of the data matrix `dat` holding the fixed
        independent variables with `fvid`[i-1] holding the column number
        corresponding to the ith fixed variable.

    fint : int
        Flag indicating whether a fixed intercept is included (`fint` =
        1).

    nrv : int
        The number of independent variables in the model which are to be
        treated as being random.

    rvid : int, shape (`nrv`)
        The columns of the data matrix `dat` holding the random
        independent variables with `rvid`[i-1] holding the column number
        corresponding to the ith random variable.

    nvpr : int
        If `rint` = 1 and `svid` != 0, `nvpr` is the number of variance
        components being estimated-2, (g-1), else `nvpr` = g.

        If `nrv` = 0, `nvpr` is not referenced.

    vpr : int, shape (`nrv`)
        `vpr`[i-1] holds a flag indicating the variance of the ith
        random variable. The variance of the ith random variable is
        sigma_j^2, where j = `vpr`[i-1]+1 if `rint` = 1 and `svid` != 0
        and j = `vpr`[i-1] otherwise. Random variables with the same
        value of j are assumed to be taken from the same distribution.

    rint : int
        Flag indicating whether a random intercept is included (`rint` =
        1).

        If `svid` = 0, `rint` is not referenced.

    svid : int
        The column of `dat` holding the subject variable.

        If `svid` = 0, no subject variable is used.

        Specifying a subject variable is equivalent to specifying the
        interaction between that variable and all of the random-effects.

        Letting the notation Z_1*Z_S denote the interaction between
        variables Z_1 and Z_S, fitting a model with `rint` = 0,
        random-effects Z_1+Z_2 and subject variable Z_S is equivalent to
        fitting a model with random-effects Z_1*Z_S+Z_2*Z_S and no
        subject variable.

        If `rint` = 1 the model is equivalent to fitting
        Z_S+Z_1*Z_S+Z_2*Z_S and no subject variable.

    gamma : float, shape (`nvpr`+2), modified in place
        `On entry`: holds the initial values of the variance components,
        gamma_0, with `gamma`[i-1] the initial value for
        sigma_i^2/sigma_R^2, for i = 1,2,...,g. If `rint` = 1 and `svid`
        != 0, g = `nvpr`+1, else g = `nvpr`.

        If `gamma`[0] = -1.0, the remaining elements of `gamma` are
        ignored and the initial values for the variance components are
        estimated from the data using MIVQUE0.

        `On exit`: `gamma`[i-1], for i = 1,2,...,g, holds the final
        estimate of sigma_i^2 and `gamma`[g] holds the final estimate
        for sigma_R^2.

    lb : int
        The size of the array `b`.

    b : float, shape (`lb`), modified in place
        `On exit`: the parameter estimates, (beta,nu), with the first
        `nff` elements of `b` containing the fixed effect parameter
        estimates, beta and the next `nrf` elements of `b` containing
        the random effect parameter estimates, nu.

        **Fixed effects**

        If `fint` = 1, `b`[0] contains the estimate of the fixed
        intercept.

        Let L_i denote the number of levels associated with the ith
        fixed variable, that is L_i = `levels`[ `fvid`[i-1] -1].

        Define

            if `fint` = 1, F_1 = 2 else if `fint` = 0, F_1 = 1;

            F_{i+1} = F_i+max(L_i-1,1), i >= 1.

        Then for i = 1,2,...,`nfv`:

            if L_i > 1, `b`[ F_i+j-2 -1] contains the parameter estimate
            for the jth level of the ith fixed variable, for j =
            2,3,...,L_i;

            if L_i <= 1, `b`[F_i-1] contains the parameter estimate for
            the ith fixed variable.

        **Random effects**

        Redefining L_i to denote the number of levels associated with
        the ith random variable, that is L_i = `levels`[ `rvid`[i-1]
        -1].

        Define

            if `rint` = 1, R_1 = 2 else if `rint` = 0, R_1 = 1;

            R_{i+1} = R_i+L_i, i >= 1.

        Then for i = 1,2,...,`nrv`:

            if `svid` = 0,

                if L_i > 1, `b`[ `nff` + R_i +j-1 -1] contains the
                parameter estimate for the jth level of the ith random
                variable, for j = 1,2,...,L_i;

                if L_i <= 1, `b`[ `nff` + R_i -1] contains the parameter
                estimate for the ith random variable;

            if `svid` != 0,

                let L_S denote the number of levels associated with the
                subject variable, that is L_S = `levels`[ `svid` -1];

                if L_i > 1, `b`[ `nff` + (s-1) L_S + R_i + j - 1 -1]
                contains the parameter estimate for the interaction
                between the sth level of the subject variable and the
                jth level of the ith random variable, for j =
                1,2,...,L_i, for s = 1,2,...,L_S;

                if L_i <= 1, `b`[ `nff` + (s-1) L_S + R_i -1] contains
                the parameter estimate for the interaction between the
                sth level of the subject variable and the ith random
                variable, for s = 1,2,...,L_S;

                if `rint` = 1, `b`[`nff`] contains the estimate of the
                random intercept.

    se : float, shape (`lb`), modified in place
        `On exit`: the standard errors of the parameter estimates given
        in `b`.

    maxit : int
        The maximum number of iterations.

        If `maxit` < 0, the default value of 100 is used.

        If `maxit` = 0, the parameter estimates (beta,nu) and
        corresponding standard errors are calculated based on the value
        of gamma_0 supplied in `gamma`.

    tol : float
        The tolerance used to assess convergence.

        If `tol` <= 0.0, the default value of epsilon^0.7 is used, where
        epsilon is the machine precision.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    nff : int
        The number of fixed effects estimated (i.e., the number of
        columns, p, in the design matrix X).

    nrf : int
        The number of random effects estimated (i.e., the number of
        columns, q, in the design matrix Z).

    df : int
        The degrees of freedom.

    ml : float
        -2l_R(gamma^) where l_R is the log of the maximum likelihood
        calculated at gamma^, the estimated variance components returned
        in `gamma`.

    warn : int
        Is set to 1 if a variance component was estimated to be a
        negative value during the fitting process. Otherwise `warn` is
        set to 0.

        If `warn` = 1, the negative estimate is set to zero and the
        estimation process allowed to continue.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: number of observations with nonzero weights must
            be greater than one.

        (`errno` 1)
            On entry, `ncol` = *<value>*.

            Constraint: `ncol` >= 1.

        (`errno` 1)
            On entry, `nfv` = *<value>* and `ncol` = *<value>*.

            Constraint: 0 <= `nfv` < `ncol`.

        (`errno` 1)
            On entry, `nrv` = *<value>* and `ncol` = *<value>*.

            Constraint: 0 <= `nrv` < `ncol` and `nrv`+`rint` > 0.

        (`errno` 1)
            On entry, `nvpr` = *<value>* and `nrv` = *<value>*.

            Constraint: 0 <= `nvpr` <= `nrv` and (`nrv` <= 0 or `nvpr`
            >= 1).

        (`errno` 1)
            On entry, `yvid` = *<value>* and `ncol` = *<value>*.

            Constraint: 1 <= `yvid` <= `ncol`.

        (`errno` 1)
            On entry, `svid` = *<value>* and `ncol` = *<value>*.

            Constraint: 0 <= `svid` <= `ncol`.

        (`errno` 1)
            On entry, `cwid` = *<value>* and `ncol` = *<value>*.

            Constraint: 0 <= `cwid` <= `ncol` and any supplied weights
            must be >= 0.0.

        (`errno` 1)
            On entry, `fint` = *<value>*.

            Constraint: `fint` = 0 or 1.

        (`errno` 1)
            On entry, `rint` = *<value>*.

            Constraint: `rint` = 0 or 1.

        (`errno` 1)
            On entry, `lb` too small: `lb` = *<value>*.

        (`errno` 2)
            On entry, `levels`[i] < 1, for at least one i.

        (`errno` 2)
            On entry, `ncol` = *<value>*.

            Constraint: 1 <= `fvid`[i] <= `ncol`, for all i.

        (`errno` 2)
            On entry, `ncol` = *<value>*.

            Constraint: 1 <= `rvid`[i] <= `ncol`, for all i.

        (`errno` 2)
            On entry, `nvpr` = *<value>*.

            Constraint: 1 <= `vpr`[i] <= `nvpr`, for all i.

        (`errno` 2)
            On entry, invalid data: categorical variable with value
            greater than that specified in `levels`.

        (`errno` 2)
            On entry, `gamma`[i] < 0, for at least one i.

        (`errno` 3)
            Degrees of freedom < 1: `df` = *<value>*.

        (`errno` 4)
            Routine failed to converge to specified tolerance: `tol` =
            *<value>*.

        (`errno` 4)
            Routine failed to converge in `maxit` iterations: `maxit` =
            *<value>*.

    References
    ----------
    Goodnight, J H, 1979, `A tutorial on the SWEEP operator`, The
    American Statistician (33(3)), 149--158

    Harville, D A, 1977, `Maximum likelihood approaches to variance
    component estimation and to related problems`, JASA (72), 320--340

    Rao, C R, 1972, `Estimation of variance and covariance components in
    a linear model`, J. Am. Stat. Assoc. (67), 112--115

    Stroup, W W, 1989, `Predictable functions and prediction space in
    the mixed model procedure`, Applications of Mixed Models in
    Agriculture and Related Disciplines (Southern Cooperative Series
    Bulletin No. 343), 39--48

    Wolfinger, R, Tobias, R and Sall, J, 1994, `Computing Gaussian
    likelihoods and their derivatives for general linear mixed models`,
    SIAM Sci. Statist. Comput. (15), 1294--1310
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.mixeff_ml'

    utils.FileObjManager._check_type_compat(
        io_manager, {'fun_name': fun_name, 'entity': 'io_manager'}
    )

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    def cb_print_rech_py(_, en_data, nout, rec, ifail, rec_len):
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data, utils._EngineIntScalarType.to_py(nout.contents), rec[:rec_len], io_manager, excs, ifail.contents,
                fun_name=fun_name,
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02jbft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._PRINT_REC_FUNCTYPE,
        utils._PRINT_RECH_FUNCTYPE,
        utils._BYREF_INT, # Input, sorder.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, ncol.
        utils._BYREF_INT, # Input, lddat.
        utils._EngineFloat64ArrayType.c_type, # Input, dat[lddat, *].
        utils._EngineIntArrayType.c_type, # Input, levels[ncol].
        utils._BYREF_INT, # Input, yvid.
        utils._BYREF_INT, # Input, cwid.
        utils._BYREF_INT, # Input, nfv.
        utils._EngineIntArrayType.c_type, # Input, fvid[nfv].
        utils._BYREF_INT, # Input, fint.
        utils._BYREF_INT, # Input, nrv.
        utils._EngineIntArrayType.c_type, # Input, rvid[nrv].
        utils._BYREF_INT, # Input, nvpr.
        utils._EngineIntArrayType.c_type, # Input, vpr[nrv].
        utils._BYREF_INT, # Input, rint.
        utils._BYREF_INT, # Input, svid.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, gamma[nvpr+2].
        utils._BYREF_INT, # Output, nff.
        utils._BYREF_INT, # Output, nrf.
        utils._BYREF_INT, # Output, df.
        utils._BYREF_FLOAT64, # Output, ml.
        utils._BYREF_INT, # Input, lb.
        utils._EngineFloat64ArrayType.c_type, # Output, b[lb].
        utils._EngineFloat64ArrayType.c_type, # Output, se[lb].
        utils._BYREF_INT, # Input, maxit.
        utils._BYREF_FLOAT64, # Input, tol.
        utils._BYREF_INT, # Output, warn.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((dat, 'dat'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _print_rech = utils._PRINT_RECH_FUNCTYPE(cb_print_rech_py)
    _sorder = utils._EngineIntScalarType.c_type(_nag_enum_consts.nag_col_major)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _ncol = utils._EngineIntScalarType(
        ncol, {'fun_name': fun_name, 'entity': 'ncol'},
    ).to_ctypes()
    _lddat_py = utils._EngineFloat64ArrayType.get_nd(dat, {'fun_name': fun_name, 'entity': 'dat'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _lddat = utils._EngineIntScalarType(
        _lddat_py, {'fun_name': fun_name, 'entity': '_lddat'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _sorder_py = utils._EngineIntScalarType.to_py(_sorder)
    _ncol_py = utils._EngineIntScalarType.to_py(_ncol)
    if _sorder_py == 1:
        _dat_dim2 = _ncol_py
    else:
        _dat_dim2 = _n_py
    exp_shape_dat = (_n_py, _dat_dim2)
    _dat = utils._EngineFloat64ArrayType(
        dat, {'fun_name': fun_name, 'entity': 'dat'},
        exp_shape=exp_shape_dat,
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _levels = utils._EngineIntArrayType(
        levels, {'fun_name': fun_name, 'entity': 'levels'},
        exp_shape=(_ncol_py,),
    ).to_ctypes()
    _yvid = utils._EngineIntScalarType(
        yvid, {'fun_name': fun_name, 'entity': 'yvid'},
    ).to_ctypes()
    _cwid = utils._EngineIntScalarType(
        cwid, {'fun_name': fun_name, 'entity': 'cwid'},
    ).to_ctypes()
    _nfv = utils._EngineIntScalarType(
        nfv, {'fun_name': fun_name, 'entity': 'nfv'},
    ).to_ctypes()
    _nfv_py = utils._EngineIntScalarType.to_py(_nfv)
    _fvid = utils._EngineIntArrayType(
        fvid, {'fun_name': fun_name, 'entity': 'fvid'},
        exp_shape=(_nfv_py,),
    ).to_ctypes()
    _fint = utils._EngineIntScalarType(
        fint, {'fun_name': fun_name, 'entity': 'fint'},
    ).to_ctypes()
    _nrv = utils._EngineIntScalarType(
        nrv, {'fun_name': fun_name, 'entity': 'nrv'},
    ).to_ctypes()
    _nrv_py = utils._EngineIntScalarType.to_py(_nrv)
    _rvid = utils._EngineIntArrayType(
        rvid, {'fun_name': fun_name, 'entity': 'rvid'},
        exp_shape=(_nrv_py,),
    ).to_ctypes()
    _nvpr = utils._EngineIntScalarType(
        nvpr, {'fun_name': fun_name, 'entity': 'nvpr'},
    ).to_ctypes()
    _vpr = utils._EngineIntArrayType(
        vpr, {'fun_name': fun_name, 'entity': 'vpr'},
        exp_shape=(_nrv_py,),
    ).to_ctypes()
    _rint = utils._EngineIntScalarType(
        rint, {'fun_name': fun_name, 'entity': 'rint'},
    ).to_ctypes()
    _svid = utils._EngineIntScalarType(
        svid, {'fun_name': fun_name, 'entity': 'svid'},
    ).to_ctypes()
    _nvpr_py = utils._EngineIntScalarType.to_py(_nvpr)
    _gamma_cls = utils._EngineFloat64ArrayType(
        gamma, {'fun_name': fun_name, 'entity': 'gamma'},
        exp_shape=(_nvpr_py+2,),
    )
    _gamma = _gamma_cls.to_ctypes()
    _nff = utils._EngineIntScalarType.empty_ctypes()
    _nrf = utils._EngineIntScalarType.empty_ctypes()
    _df = utils._EngineIntScalarType.empty_ctypes()
    _ml = utils._EngineFloat64ScalarType.empty_ctypes()
    _lb = utils._EngineIntScalarType(
        lb, {'fun_name': fun_name, 'entity': 'lb'},
    ).to_ctypes()
    _lb_py = utils._EngineIntScalarType.to_py(_lb)
    _b_cls = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_lb_py,),
        check_contents=False,
    )
    _b = _b_cls.to_ctypes(can_reference_contents=False)
    _se_cls = utils._EngineFloat64ArrayType(
        se, {'fun_name': fun_name, 'entity': 'se'},
        exp_shape=(_lb_py,),
        check_contents=False,
    )
    _se = _se_cls.to_ctypes(can_reference_contents=False)
    _maxit = utils._EngineIntScalarType(
        maxit, {'fun_name': fun_name, 'entity': 'maxit'},
    ).to_ctypes()
    _tol = utils._EngineFloat64ScalarType(
        tol, {'fun_name': fun_name, 'entity': 'tol'},
    ).to_ctypes()
    _warn = utils._EngineIntScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        utils._UNUSED_PRINT_REC,
        _print_rech,
        _sorder,
        _n,
        _ncol,
        _lddat,
        _dat,
        _levels,
        _yvid,
        _cwid,
        _nfv,
        _fvid,
        _fint,
        _nrv,
        _rvid,
        _nvpr,
        _vpr,
        _rint,
        _svid,
        _gamma,
        _nff,
        _nrf,
        _df,
        _ml,
        _lb,
        _b,
        _se,
        _maxit,
        _tol,
        _warn,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: number of observations with nonzero weights must be greater than',
                    'one.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ncol = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ncol >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nfv = {:s} and ncol = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: 0 <= nfv < ncol.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nrv = {:s} and ncol = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: 0 <= nrv < ncol and nrv+rint > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 6:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nvpr = {:s} and nrv = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: 0 <= nvpr <= nrv and (nrv <= 0 or nvpr >= 1).',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 8:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, yvid = {:s} and ncol = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: 1 <= yvid <= ncol.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 11:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, svid = {:s} and ncol = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: 0 <= svid <= ncol.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 12:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, cwid = {:s} and ncol = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: 0 <= cwid <= ncol and any supplied weights must be >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 16:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, fint = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: fint = 0 or 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 17:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, rint = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: rint = 0 or 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 19:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lb too small: lb = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 7:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, levels[i] < 1, for at least one i.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 9:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ncol = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: 1 <= fvid[i] <= ncol, for all i.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 10:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ncol = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: 1 <= rvid[i] <= ncol, for all i.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 13:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nvpr = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: 1 <= vpr[i] <= nvpr, for all i.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 14:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, invalid data: categorical variable with value',
                    'greater than that specified in levels.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 18:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, gamma[i] < 0, for at least one i.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 20:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Degrees of freedom < 1: df = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 21:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Routine failed to converge to specified tolerance:',
                    'tol = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 22:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Routine failed to converge in maxit iterations:',
                    'maxit = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _gamma_cls.refresh_py(_gamma)
        _b_cls.refresh_py(_b)
        _se_cls.refresh_py(_se)
        _nff_py = utils._EngineIntScalarType.to_py(_nff)
        _nrf_py = utils._EngineIntScalarType.to_py(_nrf)
        _df_py = utils._EngineIntScalarType.to_py(_df)
        _ml_py = utils._EngineFloat64ScalarType.to_py(_ml)
        _warn_py = utils._EngineIntScalarType.to_py(_warn)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['nff', 'nrf', 'df', 'ml', 'warn'],
        )(
            _nff_py,
            _nrf_py,
            _df_py,
            _ml_py,
            _warn_py,
        )
    _err_class.manage()
    return return_data

@utils._withdrawal(
    'naginterfaces.base.correg.mixeff_hier_init',
    mkwdrn=None,
    advice=[
        'Please use lmm_init instead.',
        'See also https://www.nag.com/numeric/py/nagdoc_latest/replace.html',
    ]
)
def mixeff_hier_init(n, ncol, dat, levels, y, wt, fixed, lfixed, nrndm, rndm):
    u"""
    Hierarchical mixed effects regression, initialization function for
    ``mixeff_hier_reml`` and ``mixeff_hier_ml``.

    ``mixeff_hier_init`` preprocesses a dataset prior to fitting a
    linear mixed effects regression model of the following form via
    either ``mixeff_hier_reml`` or ``mixeff_hier_ml``.

    Note: ``mixeff_hier_init`` is deprecated.
    Please use ``lmm_init`` instead.
    See also https://www.nag.com/numeric/py/nagdoc_latest/replace.html

    For full information please refer to the NAG Library document for
    g02jc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02jcf.html

    Parameters
    ----------
    n : int
        n, the number of observations.

        The effective number of observations, that is the number of
        observations with nonzero weight (see `wt` for more detail),
        must be greater than the number of fixed effects in the model
        (as returned in `nff`).

    ncol : int
        The number of columns in the data matrix, `dat`.

    dat : float, shape (`n`, `ncol`)
        A matrix of data, with `dat`[i-1,j-1] holding the ith
        observation on the jth variable. The two design matrices X and Z
        are constructed from `dat` and the information given in `fixed`
        (for X) and `rndm` (for Z).

    levels : int, shape (`ncol`)
        `levels`[i-1] contains the number of levels associated with the
        ith variable held in `dat`.

        If the ith variable is continuous or binary (i.e., only takes
        the values zero or one), then `levels`[i-1] must be set to 1.

        Otherwise the ith variable is assumed to take an integer value
        between 1 and `levels`[i-1], (i.e., the ith variable is discrete
        with `levels`[i-1] levels).

    y : float, shape (`n`)
        y, the vector of observations on the dependent variable.

    wt : None or float, shape (:)
        Note: the required length for this argument is determined as
        follows: if `wt` is not None: `n`; otherwise: 0.

        Optionally, the weights to be used in the weighted regression.

        If `wt`[i-1] = 0.0, the ith observation is not included in the
        model, in which case the effective number of observations is the
        number of observations with nonzero weights.

        If weights are not provided then `wt` must be set to None and
        the effective number of observations is `n`.

    fixed : int, shape (`lfixed`)
        Defines the structure of the fixed effects design matrix, X.

        `fixed`[0]

            The number of variables, N_F, to include as fixed effects
            (not including the intercept if present).

        `fixed`[1]

            The fixed intercept flag which must contain 1 if a fixed
            intercept is to be included and 0 otherwise.

        `fixed`[2+i-1]

            The column of `dat` holding the ith fixed variable, for i =
            1,2,...,`fixed`[0].

        See [Construction of the fixed effects design matrix, X] for
        more details on the construction of X.

    lfixed : int
        Length of the vector `fixed`.

    nrndm : int
        The number of columns in `rndm`.

    rndm : int, shape (:, `nrndm`)
        `rndm`[i-1,j-1] defines the structure of the `random effects`
        design matrix, Z. The bth column of `rndm` defines a block of
        columns in the design matrix Z.

        `rndm`[0,b-1]

            The number of variables, N_R_b, to include as random effects
            in the bth block (not including the random intercept if
            present).

        `rndm`[1,b-1]

            The random intercept flag which must contain 1 if block b
            includes a random intercept and 0 otherwise.

        `rndm`[2+i-1,b-1]

            The column of `dat` holding the ith random variable in the
            bth block, for i = 1,2,...,`rndm`[0,b-1].

        `rndm`[3+N_R_b-1,b-1]

            The number of subject variables, N_S_b, for the bth block.
            The subject variables define the nesting structure for this
            block.

        `rndm`[3+N_R_b+i-1,b-1]

            The column of `dat` holding the ith subject variable in the
            bth block, for i = 1,2,...,`rndm`[3+N_R_b-1,b-1].

        See [Construction of random effects design matrix, Z] for more
        details on the construction of Z.

    Returns
    -------
    nff : int
        p, the number of fixed effects estimated, i.e., the number of
        columns in the design matrix X.

    nlsv : int
        The number of levels for the overall subject variable (see
        [Construction of random effects design matrix, Z] for a
        description of what this means). If there is no overall subject
        variable, `nlsv` = 1.

    nrf : int
        The number of random effects estimated in each of the overall
        subject blocks. The number of columns in the design matrix Z is
        given by q = `nrf`*`nlsv`.

    comm : dict, communication object
        Communication structure.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 3)
            On entry, `ncol` = *<value>*.

            Constraint: `ncol` >= 0.

        (`errno` 4)
            On entry, variable j of observation i is less than 1 or
            greater than `levels`[j-1]: i = *<value>*, j = *<value>*,
            value = *<value>*, `levels`[j-1] = *<value>*.

        (`errno` 6)
            On entry, `levels`[*<value>*] = *<value>*.

            Constraint: `levels`[i-1] >= 1.

        (`errno` 8)
            On entry, `wt`[*<value>*] = *<value>*.

            Constraint: `wt`[i-1] >= 0.0.

        (`errno` 9)
            On entry, number of fixed parameters, *<value>* is less than
            zero.

        (`errno` 10)
            On entry, `lfixed` = *<value>*.

            Constraint: `lfixed` >= *<value>*.

        (`errno` 11)
            On entry, `nrndm` = *<value>*.

            Constraint: `nrndm` > 0.

        (`errno` 12)
            On entry, number of random parameters for random statement i
            is less than 0: i = *<value>*, number of parameters =
            *<value>*.

        (`errno` 102)
            On entry, more fixed factors than observations, `n` =
            *<value>*.

            Constraint: `n` >= *<value>*.

        (`errno` 108)
            On entry, no observations due to zero weights.

        (`errno` 109)
            On entry, invalid value for fixed intercept flag: value =
            *<value>*.

        (`errno` 112)
            On entry, invalid value for random intercept flag for random
            statement i: i = *<value>*, value = *<value>*.

        (`errno` 209)
            On entry, index of fixed variable j is less than 1 or
            greater than `ncol`: j = *<value>*, index = *<value>* and
            `ncol` = *<value>*.

        (`errno` 212)
            On entry, must be at least one parameter, or an intercept in
            each random statement i: i = *<value>*.

        (`errno` 312)
            On entry, index of random variable j in random statement i
            is less than 1 or greater than `ncol`: i = *<value>*, j =
            *<value>*, index = *<value>* and `ncol` = *<value>*.

        (`errno` 412)
            On entry, number of subject parameters for random statement
            i is less than 0: i = *<value>*, number of parameters =
            *<value>*.

        (`errno` 512)
            On entry, nesting variable j in random statement i has one
            level: j = *<value>*, i = *<value>*.
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.mixeff_hier_init'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02jcft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, ncol.
        utils._EngineFloat64ArrayType.c_type, # Input, dat[lddat, tddat].
        utils._BYREF_INT, # Input, lddat.
        utils._BYREF_INT, # Input, tddat.
        utils._EngineIntArrayType.c_type, # Input, levels[ncol].
        utils._EngineFloat64ArrayType.c_type, # Input, y[n].
        utils._EngineFloat64ArrayType.c_type, # Input, wt[lwt].
        utils._BYREF_INT, # Input, lwt.
        utils._EngineIntArrayType.c_type, # Input, fixed[lfixed].
        utils._BYREF_INT, # Input, lfixed.
        utils._BYREF_INT, # Input, nrndm.
        utils._EngineIntArrayType.c_type, # Input, rndm[ldrndm, tdrndm].
        utils._BYREF_INT, # Input, ldrndm.
        utils._BYREF_INT, # Input, tdrndm.
        utils._BYREF_INT, # Output, nff.
        utils._BYREF_INT, # Output, nlsv.
        utils._BYREF_INT, # Output, nrf.
        utils._EngineFloat64ArrayType.c_type, # Output, rcomm[lrcomm].
        utils._BYREF_INT, # Input, lrcomm.
        utils._EngineIntArrayType.c_type, # Output, icomm[licomm].
        utils._BYREF_INT, # Input, licomm.
        utils._BYREF_INT, # Input, sorder.
        utils._BYREF_INT, # Input, idfirst.
        utils._BYREF_INT, # Input, irecode.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((dat, 'dat'), (rndm, 'rndm')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _ncol = utils._EngineIntScalarType(
        ncol, {'fun_name': fun_name, 'entity': 'ncol'},
    ).to_ctypes()
    _lddat_py = utils._EngineFloat64ArrayType.get_nd(dat, {'fun_name': fun_name, 'entity': 'dat'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _lddat = utils._EngineIntScalarType(
        _lddat_py, {'fun_name': fun_name, 'entity': '_lddat'},
    ).to_ctypes()
    _tddat_py = utils._EngineFloat64ArrayType.get_nd(dat, {'fun_name': fun_name, 'entity': 'dat'}, exp_rank=2, sorder=_m_sorder, dim_type='s')
    _tddat = utils._EngineIntScalarType(
        _tddat_py, {'fun_name': fun_name, 'entity': '_tddat'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _ncol_py = utils._EngineIntScalarType.to_py(_ncol)
    _dat = utils._EngineFloat64ArrayType(
        dat, {'fun_name': fun_name, 'entity': 'dat'},
        exp_shape=(_n_py, _ncol_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _levels = utils._EngineIntArrayType(
        levels, {'fun_name': fun_name, 'entity': 'levels'},
        exp_shape=(_ncol_py,),
    ).to_ctypes()
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _lwt_py = utils._EngineFloat64ArrayType.get_shape(wt, {'fun_name': fun_name, 'entity': 'wt'}, exp_rank=1, allow_none=True)[0]
    _lwt = utils._EngineIntScalarType(
        _lwt_py, {'fun_name': fun_name, 'entity': '_lwt'},
    ).to_ctypes()
    if wt is not None:
        _wt_dim1 = _n_py
    else:
        _wt_dim1 = 0
    exp_shape_wt = (_wt_dim1,)
    _wt = utils._EngineFloat64ArrayType(
        wt, {'fun_name': fun_name, 'entity': 'wt'},
        exp_shape=exp_shape_wt,
        allow_none=True,
    ).to_ctypes()
    _lfixed = utils._EngineIntScalarType(
        lfixed, {'fun_name': fun_name, 'entity': 'lfixed'},
    ).to_ctypes()
    _lfixed_py = utils._EngineIntScalarType.to_py(_lfixed)
    _fixed = utils._EngineIntArrayType(
        fixed, {'fun_name': fun_name, 'entity': 'fixed'},
        exp_shape=(_lfixed_py,),
    ).to_ctypes()
    _nrndm = utils._EngineIntScalarType(
        nrndm, {'fun_name': fun_name, 'entity': 'nrndm'},
    ).to_ctypes()
    _ldrndm_py = utils._EngineIntArrayType.get_nd(rndm, {'fun_name': fun_name, 'entity': 'rndm'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldrndm = utils._EngineIntScalarType(
        _ldrndm_py, {'fun_name': fun_name, 'entity': '_ldrndm'},
    ).to_ctypes()
    _tdrndm_py = utils._EngineIntArrayType.get_nd(rndm, {'fun_name': fun_name, 'entity': 'rndm'}, exp_rank=2, sorder=_m_sorder, dim_type='s')
    _tdrndm = utils._EngineIntScalarType(
        _tdrndm_py, {'fun_name': fun_name, 'entity': '_tdrndm'},
    ).to_ctypes()
    _nrndm_py = utils._EngineIntScalarType.to_py(_nrndm)
    _rndm = utils._EngineIntArrayType(
        rndm, {'fun_name': fun_name, 'entity': 'rndm'},
        exp_shape=(None, _nrndm_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _nff = utils._EngineIntScalarType.empty_ctypes()
    _nlsv = utils._EngineIntScalarType.empty_ctypes()
    _nrf = utils._EngineIntScalarType.empty_ctypes()
    _lrcomm = utils._EngineIntScalarType.c_type(2)
    _lrcomm_py = utils._EngineIntScalarType.to_py(_lrcomm)
    _rcomm = utils._EngineFloat64ArrayType.empty_ctypes(_lrcomm_py)
    _comm_py = {}
    _licomm = utils._EngineIntScalarType.c_type(2)
    _licomm_py = utils._EngineIntScalarType.to_py(_licomm)
    _icomm = utils._EngineIntArrayType.empty_ctypes(_licomm_py)
    _sorder = utils._EngineIntScalarType.c_type(_nag_enum_consts.nag_col_major)
    _idfirst = utils._EngineIntScalarType.c_type(1)
    _irecode = utils._EngineIntScalarType.c_type(1)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _ncol,
        _dat,
        _lddat,
        _tddat,
        _levels,
        _y,
        _wt,
        _lwt,
        _fixed,
        _lfixed,
        _nrndm,
        _rndm,
        _ldrndm,
        _tdrndm,
        _nff,
        _nlsv,
        _nrf,
        _rcomm,
        _lrcomm,
        _icomm,
        _licomm,
        _sorder,
        _idfirst,
        _irecode,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    if _ierr.value == utils._IERR_SUCCESS:
        _lrcomm = utils._EngineIntScalarType.c_type(_icomm[1])
        _lrcomm_py = utils._EngineIntScalarType.to_py(_lrcomm)
        _rcomm = utils._EngineFloat64ArrayType.empty_ctypes(_lrcomm_py)
        _comm_py['opts'] = _rcomm
        _licomm = utils._EngineIntScalarType.c_type(_icomm[0])
        _licomm_py = utils._EngineIntScalarType.to_py(_licomm)
        _icomm = utils._EngineIntArrayType.empty_ctypes(_licomm_py)
        _comm_py['iopts'] = _icomm
        engine_ctypes(
            _en_data,
            _n,
            _ncol,
            _dat,
            _lddat,
            _tddat,
            _levels,
            _y,
            _wt,
            _lwt,
            _fixed,
            _lfixed,
            _nrndm,
            _rndm,
            _ldrndm,
            _tdrndm,
            _nff,
            _nlsv,
            _nrf,
            _rcomm,
            _lrcomm,
            _icomm,
            _licomm,
            _sorder,
            _idfirst,
            _irecode,
            _errbuf,
            _ierr,
            _errbuf._length_,
        )
        utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _ierr_fl = 2
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 2:
            _ierr_fl = 3
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ncol = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ncol >= 0.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _ierr_fl = 4
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, variable j of observation i is less than 1 or greater than',
                    'levels[j-1]: i = {:s}, j = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'value = {:s}, levels[j-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[4]),
                    ),
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 6:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, levels[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: levels[i-1] >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 8:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, wt[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: wt[i-1] >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 10 and _ifmt == 10:
            _ierr_fl = 9
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, number of fixed parameters, {:s} is less than'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'zero.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 11 and _ifmt == 11:
            _ierr_fl = 10
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lfixed = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: lfixed >= {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 12 and _ifmt == 12:
            _ierr_fl = 11
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nrndm = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nrndm > 0.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 13 and _ifmt == 13:
            _ierr_fl = 12
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, number of random parameters for random statement i is less than',
                    '0: i = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'number of parameters = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 101 and _ifmt == 101:
            _ierr_fl = 102
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, more fixed factors than observations, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: n >= {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 108 and _ifmt == 108:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, no observations due to zero weights.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 110 and _ifmt == 110:
            _ierr_fl = 109
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, invalid value for fixed intercept flag:',
                    'value = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 113 and _ifmt == 113:
            _ierr_fl = 112
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, invalid value for random intercept flag for random statement',
                    'i: i = {:s}, value = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 310 and _ifmt == 310:
            _ierr_fl = 209
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, index of fixed variable j is less than 1 or greater than ncol:',
                    'j = {:s}, index = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'ncol = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 213 and _ifmt == 213:
            _ierr_fl = 212
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, must be at least one parameter, or an intercept in each random',
                    'statement i: i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 313 and _ifmt == 313:
            _ierr_fl = 312
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, index of random variable j in random statement',
                    'i is less than 1 or greater than ncol:',
                    'i = {:s}, j = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[4]),
                    ),
                    'index = {:s} and ncol = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 413 and _ifmt == 413:
            _ierr_fl = 412
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, number of subject parameters for random statement i is less than',
                    '0: i = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'number of parameters = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 513 and _ifmt == 513:
            _ierr_fl = 512
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nesting variable j in random statement i has one level:',
                    'j = {:s}, i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _nff_py = utils._EngineIntScalarType.to_py(_nff)
        _nlsv_py = utils._EngineIntScalarType.to_py(_nlsv)
        _nrf_py = utils._EngineIntScalarType.to_py(_nrf)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['nff', 'nlsv', 'nrf', 'comm'],
        )(
            _nff_py,
            _nlsv_py,
            _nrf_py,
            _comm_py,
        )
    _err_class.manage()
    return return_data

@utils._withdrawal(
    'naginterfaces.base.correg.mixeff_hier_reml',
    mkwdrn=None,
    advice=[
        'Please use lmm_fit instead.',
        'See also https://www.nag.com/numeric/py/nagdoc_latest/replace.html',
    ]
)
def mixeff_hier_reml(lvpr, vpr, nvpr, gamma, lb, estid, ldid, b, se, czz, cxx, cxz, comm, iopt, liopt, ropt, lropt, io_manager=None):
    u"""
    Hierarchical mixed effects regression using Restricted Maximum
    Likelihood (REML).

    ``mixeff_hier_reml`` fits a multi-level linear mixed effects
    regression model using restricted maximum likelihood (REML).
    Prior to calling ``mixeff_hier_reml`` the initialization function
    ``mixeff_hier_init`` must be called.

    Note: ``mixeff_hier_reml`` is deprecated.
    Please use ``lmm_fit`` instead.
    See also https://www.nag.com/numeric/py/nagdoc_latest/replace.html

    For full information please refer to the NAG Library document for
    g02jd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02jdf.html

    Parameters
    ----------
    lvpr : int
        The sum of the number of random parameters and the random
        intercept flags specified in the call to ``mixeff_hier_init``.

    vpr : int, shape (`lvpr`)
        A vector of flags indicating the mapping between the random
        variables specified in rndm and the variance components,
        sigma_i^2. See [Further Comments] for more details.

    nvpr : int
        g, the number of variance components being estimated (excluding
        the overall variance, sigma_R^2).

    gamma : float, shape (`nvpr`+1), modified in place
        `On entry`: holds the initial values of the variance components,
        gamma_0, with `gamma`[i-1] the initial value for
        sigma_i^2/sigma_R^2, for i = 1,2,...,`nvpr`.

        If `gamma`[0] = -1.0, the remaining elements of `gamma` are
        ignored and the initial values for the variance components are
        estimated from the data using MIVQUE0.

        `On exit`: `gamma`[i-1], for i = 1,2,...,`nvpr`, holds the final
        estimate of sigma_i^2 and `gamma`[`nvpr`] holds the final
        estimate for sigma_R^2.

    lb : int
        The dimension of the arrays `b` and `se`.

    estid : int, shape (`ldid`, `lb`), modified in place
        `On exit`: an array describing the parameter estimates returned
        in `b`. The first nlsv*nrf columns of `estid` describe the
        parameter estimates for the random effects and the last nff
        columns the parameter estimates for the fixed effects.

        For fixed effects:

            for l = nrf*nlsv+1,...,nrf*nlsv+nff

                if `b`[l-1] contains the parameter estimate for the
                intercept, then

                        `estid`[0,l-1] = `estid`[1,l-1] = `estid`[2,l-1]
                        = 0;

                if `b`[l-1] contains the parameter estimate for the ith
                level of the jth fixed variable, that is the vector of
                values held in the kth column of dat when fixed[j+1] = k
                then

                        `estid`[0,l-1] = 0, ; `estid`[1,l-1] = j, ;
                        `estid`[2,l-1] = i;

                if the jth variable is continuous or binary, that is
                levels[fixed[j+1]-1] = 1, `estid`[2,l-1] = 0;

                any remaining rows of the lth column of `estid` are set
                to 0.

        For random effects:

            let

                N_R_b denote the number of random variables in the bth
                random statement, that is N_R_b = rndm(1,b);

                R_{jb} denote the jth random variable from the bth
                random statement, that is the vector of values held in
                the kth column of dat when rndm(2+j,b) = k;

                N_S_b denote the number of subject variables in the bth
                random statement, that is N_S_b = rndm(3+N_R_b,b);

                S_{jb} denote the jth subject variable from the bth
                random statement, that is the vector of values held in
                the kth column of dat when rndm(3+N_R_b+j,b) = k;

                L(S_{jb}) denote the number of levels for S_{jb}, that
                is L(S_{jb}) = levels[rndm(3+N_R_b+j,b)-1];

            then

                for l = 1,2,...nrf*nlsv, if `b`[l-1] contains the
                parameter estimate for the ith level of R_{jb} when
                S_{kb} = s_k, for k = 1,2,...,N_S_b and 1 <= s_k <=
                L(S_{jb}), i.e., s_k is a valid value for the kth
                subject variable, then

                        `estid`[0,l-1] = b, ; `estid`[1,l-1] = j, ;
                        `estid`[2,l-1] = i, ; `estid`[3+k-1,l-1] = s_k,
                        k = 1,2,...,N_S_b;

                if the parameter being estimated is for the intercept,
                then `estid`[1,l-1] = `estid`[2,l-1] = 0;

                if the jth variable is continuous, or binary, that is
                L(S_{jb}) = 1, then `estid`[2,l-1] = 0;

                the remaining rows of the lth column of `estid` are set
                to 0.

        In some situations, certain combinations of variables are never
        observed.

        In such circumstances all elements of the lth row of `estid` are
        set to -999.

    ldid : int
        The first dimension of the array `estid`.

    b : float, shape (`lb`), modified in place
        `On exit`: the parameter estimates, with the first nrf*nlsv
        elements of `b` containing the parameter estimates for the
        random effects, nu, and the remaining nff elements containing
        the parameter estimates for the fixed effects, beta. The order
        of these estimates are described by the `estid` argument.

    se : float, shape (`lb`), modified in place
        `On exit`: the standard errors of the parameter estimates given
        in `b`.

    czz : float, shape (nrf, :), modified in place
        Note: the required extent for this argument in dimension 2 is
        determined as follows: if vlicomm>=8:
        `comm`['iopts'][6]*`comm`['iopts'][7]; otherwise: 0.

        `On exit`: if nlsv = 1, then `czz` holds the lower triangular
        portion of the matrix (1/sigma^2)(Z^TR^^-1Z+G^^-1), where R^ and
        G^ are the estimates of R and G respectively. If nlsv > 1, then
        `czz` holds this matrix in compressed form, with the first nrf
        columns holding the part of the matrix corresponding to the
        first level of the overall subject variable, the next nrf
        columns the part corresponding to the second level of the
        overall subject variable etc.

    cxx : float, shape (nff, :), modified in place
        Note: the required extent for this argument in dimension 2 is
        determined as follows: if vlicomm>=6: `comm`['iopts'][5];
        otherwise: 0.

        `On exit`: `cxx` holds the lower triangular portion of the
        matrix (1/sigma^2)X^TV^^-1X, where V^ is the estimated value of
        V.

    cxz : float, shape (nff, :), modified in place
        Note: the required extent for this argument in dimension 2 is
        determined as follows: if vlicomm>=8:
        `comm`['iopts'][6]*`comm`['iopts'][7]; otherwise: 0.

        `On exit`: if nlsv = 1, then `cxz` holds the matrix
        (1/sigma^2)(X^TV^^-1Z)G^, where V^ and G^ are the estimates of V
        and G respectively. If nlsv > 1, then `cxz` holds this matrix in
        compressed form, with the first nrf columns holding the part of
        the matrix corresponding to the first level of the overall
        subject variable, the next nrf columns the part corresponding to
        the second level of the overall subject variable etc.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``mixeff_hier_init``.

    iopt : None or int, shape (`liopt`)
        Options passed to the optimization function.

        By default ``mixeff_hier_reml`` fits the specified model using a
        modified Newton optimization algorithm as implemented in
        ``opt.bounds_mod_deriv2_comp``. In some cases, where the
        calculation of the derivatives is computationally expensive it
        may be more efficient to use a sequential QP algorithm.

        The sequential QP algorithm as implemented in ``opt.nlp1_solve``
        can be chosen by setting `iopt`[4] = 1.

        If `liopt` < 5 or `iopt`[4] != 1, then the modified Newton
        algorithm will be used.

        Different options are available depending on the optimization
        function used.

        In all cases, using a value of -1 will cause the default value
        to be used.

        In addition only the first `liopt` values of `iopt` are used, so
        for example, if only the first element of `iopt` needs changing
        and default values for all other options are sufficient `liopt`
        can be set to 1.

        The following table lists the association between elements of
        `iopt` and arguments in the optimizer when the modified Newton
        algorithm is being used.

        [table omitted]

        If requested, monitoring information is displayed in a similar
        format to that given by the modified Newton optimizer.

        The following table lists the association between elements of
        `iopt` and options in the optimizer when the sequential QP
        algorithm is being used.

        [table omitted]

        If `liopt` <= 0, default values are used for all options and
        `iopt` is not referenced.

    liopt : int
        Length of the options array `iopt`.

    ropt : None or float, shape (`lropt`)
        Options passed to the optimization function.

        Different options are available depending on the optimization
        function used.

        In all cases, using a value of -1.0 will cause the default value
        to be used.

        In addition only the first `lropt` values of `ropt` are used, so
        for example, if only the first element of `ropt` needs changing
        and default values for all other options are sufficient `lropt`
        can be set to 1.

        The following table lists the association between elements of
        `ropt` and arguments in the optimizer when the modified Newton
        algorithm is being used.

        [table omitted]

        The following table lists the association between elements of
        `ropt` and options in the optimizer when the sequential QP
        algorithm is being used.

        [table omitted]

        where eps is the machine precision returned by
        ``machine.precision`` and zz_{ii} denotes the i diagonal element
        of Z^TZ.

        If `lropt` <= 0, then default values are used for all options
        and `ropt` may be set to None.

    lropt : int
        Length of the options array `ropt`.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    effn : int
        Effective number of observations. If no weights were supplied to
        ``mixeff_hier_init`` or all supplied weights were nonzero,
        `effn` = n.

    rnkx : int
        The rank of the design matrix, X, for the fixed effects.

    ncov : int
        Number of variance components not estimated to be zero. If none
        of the variance components are estimated to be zero, then `ncov`
        = `nvpr`.

    lnlike : float
        -2l_R(gamma^) where l_R is the log of the restricted maximum
        likelihood calculated at gamma^, the estimated variance
        components returned in `gamma`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `lvpr` = *<value>*.

            Constraint: `lvpr` >= *<value>*.

        (`errno` 2)
            On entry, `vpr`[*<value>*] = *<value>* and `nvpr` =
            *<value>*.

            Constraint: 1 <= `vpr`[i] <= `nvpr`.

        (`errno` 3)
            On entry, `nvpr` = *<value>*.

            Constraint: 1 <= `nvpr` <= *<value>*.

        (`errno` 4)
            On entry, `gamma`[*<value>*] = *<value>*.

            Constraint: `gamma`[0] = -1.0 or `gamma`[i-1] >= 0.0.

        (`errno` 9)
            On entry, `lb` = *<value>*.

            Constraint: `lb` >= *<value>*.

        (`errno` 11)
            On entry, `ldid` = *<value>*.

            Constraint: `ldid` >= *<value>*.

        (`errno` 21)
            On entry, `comm`['iopts'] has not been initialized
            correctly.

        (`errno` 32)
            On entry, at least one value of i, for i = 1,2,...,`nvpr`,
            does not appear in `vpr`.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 101)
            Optimal solution found, but requested accuracy not achieved.

        (`errno` 102)
            Too many major iterations.

        (`errno` 103)
            Current point cannot be improved upon.

        (`errno` 104)
            At least one negative estimate for `gamma` was obtained. All
            negative estimates have been set to zero.

    References
    ----------
    Goodnight, J H, 1979, `A tutorial on the SWEEP operator`, The
    American Statistician (33(3)), 149--158

    Harville, D A, 1977, `Maximum likelihood approaches to variance
    component estimation and to related problems`, JASA (72), 320--340

    Rao, C R, 1972, `Estimation of variance and covariance components in
    a linear model`, J. Am. Stat. Assoc. (67), 112--115

    Stroup, W W, 1989, `Predictable functions and prediction space in
    the mixed model procedure`, Applications of Mixed Models in
    Agriculture and Related Disciplines (Southern Cooperative Series
    Bulletin No. 343), 39--48

    Wolfinger, R, Tobias, R and Sall, J, 1994, `Computing Gaussian
    likelihoods and their derivatives for general linear mixed models`,
    SIAM Sci. Statist. Comput. (15), 1294--1310
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.mixeff_hier_reml'

    utils.FileObjManager._check_type_compat(
        io_manager, {'fun_name': fun_name, 'entity': 'io_manager'}
    )

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    def cb_print_rech_py(_, en_data, nout, rec, ifail, rec_len):
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data, utils._EngineIntScalarType.to_py(nout.contents), rec[:rec_len], io_manager, excs, ifail.contents,
                fun_name=fun_name,
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02jdft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._PRINT_REC_FUNCTYPE,
        utils._PRINT_RECH_FUNCTYPE,
        utils._BYREF_INT, # Input, lvpr.
        utils._EngineIntArrayType.c_type, # Input, vpr[lvpr].
        utils._BYREF_INT, # Input, nvpr.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, gamma[nvpr+1].
        utils._BYREF_INT, # Output, effn.
        utils._BYREF_INT, # Output, rnkx.
        utils._BYREF_INT, # Output, ncov.
        utils._BYREF_FLOAT64, # Output, lnlike.
        utils._BYREF_INT, # Input, lb.
        utils._EngineIntArrayType.c_type, # Output, estid[ldid, tdid].
        utils._BYREF_INT, # Input, ldid.
        utils._EngineFloat64ArrayType.c_type, # Output, b[lb].
        utils._EngineFloat64ArrayType.c_type, # Output, se[lb].
        utils._EngineFloat64ArrayType.c_type, # Output, czz[ldczz, tdczz].
        utils._BYREF_INT, # Input, ldczz.
        utils._EngineFloat64ArrayType.c_type, # Output, cxx[ldcxx, tdcxx].
        utils._BYREF_INT, # Input, ldcxx.
        utils._EngineFloat64ArrayType.c_type, # Output, cxz[ldcxz, tdcxz].
        utils._BYREF_INT, # Input, ldcxz.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, rcomm[lrcomm].
        utils._EngineIntArrayType.c_type, # Input/Output, icomm[licomm].
        utils._EngineIntArrayType.c_type, # Input, iopt[liopt].
        utils._BYREF_INT, # Input, liopt.
        utils._EngineFloat64ArrayType.c_type, # Input, ropt[lropt].
        utils._BYREF_INT, # Input, lropt.
        utils._BYREF_INT, # Input, licomm.
        utils._BYREF_INT, # Input, lrcomm.
        utils._BYREF_INT, # Input, tdid.
        utils._BYREF_INT, # Input, tdczz.
        utils._BYREF_INT, # Input, tdcxx.
        utils._BYREF_INT, # Input, tdcxz.
        utils._BYREF_STRING, # Input, routine_name(len=6).
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of routine_name.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((estid, 'estid'), (czz, 'czz'), (cxx, 'cxx'), (cxz, 'cxz')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _print_rech = utils._PRINT_RECH_FUNCTYPE(cb_print_rech_py)
    _lvpr = utils._EngineIntScalarType(
        lvpr, {'fun_name': fun_name, 'entity': 'lvpr'},
    ).to_ctypes()
    _lvpr_py = utils._EngineIntScalarType.to_py(_lvpr)
    _vpr = utils._EngineIntArrayType(
        vpr, {'fun_name': fun_name, 'entity': 'vpr'},
        exp_shape=(_lvpr_py,),
    ).to_ctypes()
    _nvpr = utils._EngineIntScalarType(
        nvpr, {'fun_name': fun_name, 'entity': 'nvpr'},
    ).to_ctypes()
    _nvpr_py = utils._EngineIntScalarType.to_py(_nvpr)
    _gamma_cls = utils._EngineFloat64ArrayType(
        gamma, {'fun_name': fun_name, 'entity': 'gamma'},
        exp_shape=(_nvpr_py+1,),
    )
    _gamma = _gamma_cls.to_ctypes()
    _effn = utils._EngineIntScalarType.empty_ctypes()
    _rnkx = utils._EngineIntScalarType.empty_ctypes()
    _ncov = utils._EngineIntScalarType.empty_ctypes()
    _lnlike = utils._EngineFloat64ScalarType.empty_ctypes()
    _lb = utils._EngineIntScalarType(
        lb, {'fun_name': fun_name, 'entity': 'lb'},
    ).to_ctypes()
    _ldid = utils._EngineIntScalarType(
        ldid, {'fun_name': fun_name, 'entity': 'ldid'},
    ).to_ctypes()
    _tdid_py = utils._EngineIntArrayType.get_shape(estid, {'fun_name': fun_name, 'entity': 'estid'}, exp_rank=2)[1]
    _tdid = utils._EngineIntScalarType(
        _tdid_py, {'fun_name': fun_name, 'entity': '_tdid'},
    ).to_ctypes()
    utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        req_keys=('opts', 'iopts'),
    )
    _vl_vlicomm_py = utils._EngineIntArrayType.get_element(comm['iopts'], {'fun_name': fun_name, 'entity': 'comm[\'iopts\']'}, (0,), validate=False)
    _lb_py = utils._EngineIntScalarType.to_py(_lb)
    if _vl_vlicomm_py >= 5:
        _estid_dim1 = utils._EngineIntArrayType.get_element(comm['iopts'], {'fun_name': fun_name, 'entity': 'comm[\'iopts\']'}, (4,), validate=False)
    else:
        _estid_dim1 = 0
    exp_shape_estid = (_estid_dim1, _lb_py)
    _estid_cls = utils._EngineIntArrayType(
        estid, {'fun_name': fun_name, 'entity': 'estid'},
        exp_shape=exp_shape_estid,
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _estid = _estid_cls.to_ctypes(can_reference_contents=False)
    _vl_g02jcf_nff_py = utils._EngineIntArrayType.get_element(comm['iopts'], {'fun_name': fun_name, 'entity': 'comm[\'iopts\']'}, (5,), validate=False)
    _vl_g02jcf_nrf_py = utils._EngineIntArrayType.get_element(comm['iopts'], {'fun_name': fun_name, 'entity': 'comm[\'iopts\']'}, (6,), validate=False)
    _vl_g02jcf_nlsv_py = utils._EngineIntArrayType.get_element(comm['iopts'], {'fun_name': fun_name, 'entity': 'comm[\'iopts\']'}, (7,), validate=False)
    _b_cls = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_vl_g02jcf_nff_py+_vl_g02jcf_nrf_py*_vl_g02jcf_nlsv_py,),
        check_contents=False,
    )
    _b = _b_cls.to_ctypes(can_reference_contents=False)
    _se_cls = utils._EngineFloat64ArrayType(
        se, {'fun_name': fun_name, 'entity': 'se'},
        exp_shape=(_vl_g02jcf_nff_py+_vl_g02jcf_nrf_py*_vl_g02jcf_nlsv_py,),
        check_contents=False,
    )
    _se = _se_cls.to_ctypes(can_reference_contents=False)
    _ldczz_py = utils._EngineFloat64ArrayType.get_nd(czz, {'fun_name': fun_name, 'entity': 'czz'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldczz = utils._EngineIntScalarType(
        _ldczz_py, {'fun_name': fun_name, 'entity': '_ldczz'},
    ).to_ctypes()
    _tdczz_py = utils._EngineFloat64ArrayType.get_nd(czz, {'fun_name': fun_name, 'entity': 'czz'}, exp_rank=2, sorder=_m_sorder, dim_type='s')
    _tdczz = utils._EngineIntScalarType(
        _tdczz_py, {'fun_name': fun_name, 'entity': '_tdczz'},
    ).to_ctypes()
    if _vl_vlicomm_py >= 8:
        _czz_dim2 = utils._EngineIntArrayType.get_element(comm['iopts'], {'fun_name': fun_name, 'entity': 'comm[\'iopts\']'}, (6,), validate=False)*utils._EngineIntArrayType.get_element(comm['iopts'], {'fun_name': fun_name, 'entity': 'comm[\'iopts\']'}, (7,), validate=False)
    else:
        _czz_dim2 = 0
    exp_shape_czz = (_vl_g02jcf_nrf_py, _czz_dim2)
    _czz_cls = utils._EngineFloat64ArrayType(
        czz, {'fun_name': fun_name, 'entity': 'czz'},
        exp_shape=exp_shape_czz,
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _czz = _czz_cls.to_ctypes(can_reference_contents=False)
    _ldcxx_py = utils._EngineFloat64ArrayType.get_nd(cxx, {'fun_name': fun_name, 'entity': 'cxx'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldcxx = utils._EngineIntScalarType(
        _ldcxx_py, {'fun_name': fun_name, 'entity': '_ldcxx'},
    ).to_ctypes()
    _tdcxx_py = utils._EngineFloat64ArrayType.get_nd(cxx, {'fun_name': fun_name, 'entity': 'cxx'}, exp_rank=2, sorder=_m_sorder, dim_type='s')
    _tdcxx = utils._EngineIntScalarType(
        _tdcxx_py, {'fun_name': fun_name, 'entity': '_tdcxx'},
    ).to_ctypes()
    if _vl_vlicomm_py >= 6:
        _cxx_dim2 = utils._EngineIntArrayType.get_element(comm['iopts'], {'fun_name': fun_name, 'entity': 'comm[\'iopts\']'}, (5,), validate=False)
    else:
        _cxx_dim2 = 0
    exp_shape_cxx = (_vl_g02jcf_nff_py, _cxx_dim2)
    _cxx_cls = utils._EngineFloat64ArrayType(
        cxx, {'fun_name': fun_name, 'entity': 'cxx'},
        exp_shape=exp_shape_cxx,
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _cxx = _cxx_cls.to_ctypes(can_reference_contents=False)
    _ldcxz_py = utils._EngineFloat64ArrayType.get_nd(cxz, {'fun_name': fun_name, 'entity': 'cxz'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldcxz = utils._EngineIntScalarType(
        _ldcxz_py, {'fun_name': fun_name, 'entity': '_ldcxz'},
    ).to_ctypes()
    _tdcxz_py = utils._EngineFloat64ArrayType.get_nd(cxz, {'fun_name': fun_name, 'entity': 'cxz'}, exp_rank=2, sorder=_m_sorder, dim_type='s')
    _tdcxz = utils._EngineIntScalarType(
        _tdcxz_py, {'fun_name': fun_name, 'entity': '_tdcxz'},
    ).to_ctypes()
    if _vl_vlicomm_py >= 8:
        _cxz_dim2 = utils._EngineIntArrayType.get_element(comm['iopts'], {'fun_name': fun_name, 'entity': 'comm[\'iopts\']'}, (6,), validate=False)*utils._EngineIntArrayType.get_element(comm['iopts'], {'fun_name': fun_name, 'entity': 'comm[\'iopts\']'}, (7,), validate=False)
    else:
        _cxz_dim2 = 0
    exp_shape_cxz = (_vl_g02jcf_nff_py, _cxz_dim2)
    _cxz_cls = utils._EngineFloat64ArrayType(
        cxz, {'fun_name': fun_name, 'entity': 'cxz'},
        exp_shape=exp_shape_cxz,
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _cxz = _cxz_cls.to_ctypes(can_reference_contents=False)
    _lrcomm_py = utils._EngineFloat64ArrayType.get_shape(comm['opts'], {'fun_name': fun_name, 'entity': 'comm[\'opts\']'}, exp_rank=1, validate=False)[0]
    _lrcomm = utils._EngineIntScalarType(
        _lrcomm_py, {'fun_name': fun_name, 'entity': '_lrcomm'},
    ).to_ctypes()
    _rcomm = comm['opts']
    _licomm_py = utils._EngineIntArrayType.get_shape(comm['iopts'], {'fun_name': fun_name, 'entity': 'comm[\'iopts\']'}, exp_rank=1, validate=False)[0]
    _licomm = utils._EngineIntScalarType(
        _licomm_py, {'fun_name': fun_name, 'entity': '_licomm'},
    ).to_ctypes()
    _icomm = comm['iopts']
    _liopt = utils._EngineIntScalarType(
        liopt, {'fun_name': fun_name, 'entity': 'liopt'},
    ).to_ctypes()
    _liopt_py = utils._EngineIntScalarType.to_py(_liopt)
    if _liopt_py > 0:
        if iopt is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'iopt'},
                'must not be None when liopt > 0',
            )
    _iopt = utils._EngineIntArrayType(
        iopt, {'fun_name': fun_name, 'entity': 'iopt'},
        exp_shape=(_liopt_py,),
        check_contents=_liopt_py > 0,
        allow_none=True,
    ).to_ctypes(can_reference_contents=_liopt_py > 0)
    _lropt = utils._EngineIntScalarType(
        lropt, {'fun_name': fun_name, 'entity': 'lropt'},
    ).to_ctypes()
    _lropt_py = utils._EngineIntScalarType.to_py(_lropt)
    if _lropt_py > 0:
        if ropt is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'ropt'},
                'must not be None when lropt > 0',
            )
    _ropt = utils._EngineFloat64ArrayType(
        ropt, {'fun_name': fun_name, 'entity': 'ropt'},
        exp_shape=(_lropt_py,),
        check_contents=_lropt_py > 0,
        allow_none=True,
    ).to_ctypes(can_reference_contents=_lropt_py > 0)
    _routine_name = utils._EngineRoutineName(*b'g02jd ')
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        utils._UNUSED_PRINT_REC,
        _print_rech,
        _lvpr,
        _vpr,
        _nvpr,
        _gamma,
        _effn,
        _rnkx,
        _ncov,
        _lnlike,
        _lb,
        _estid,
        _ldid,
        _b,
        _se,
        _czz,
        _ldczz,
        _cxx,
        _ldcxx,
        _cxz,
        _ldcxz,
        _rcomm,
        _icomm,
        _iopt,
        _liopt,
        _ropt,
        _lropt,
        _licomm,
        _lrcomm,
        _tdid,
        _tdczz,
        _tdcxx,
        _tdcxz,
        _routine_name,
        _errbuf,
        _ierr,
        _routine_name._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lvpr = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: lvpr >= {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, vpr[{:s}] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'nvpr = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: 1 <= vpr[i] <= nvpr.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nvpr = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: 1 <= nvpr <= {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, gamma[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: gamma[0] = -1.0 or gamma[i-1] >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 10 and _ifmt == 10:
            _ierr_fl = 9
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lb = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: lb >= {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 13 and _ifmt == 13:
            _ierr_fl = 9
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lb = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: lb >= {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 12 and _ifmt == 12:
            _ierr_fl = 11
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ldid = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ldid >= {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 18 and _ifmt == 18:
            _ierr_fl = 21
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, comm[\'iopts\'] has not been initialized correctly.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 21 and _ifmt == 21:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, comm[\'iopts\'] has not been initialized correctly.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 24 and _ifmt == 24:
            _ierr_fl = 21
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, comm[\'iopts\'] has not been initialized correctly.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 26 and _ifmt == 26:
            _ierr_fl = 21
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, comm[\'iopts\'] has not been initialized correctly.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 27 and _ifmt == 27:
            _ierr_fl = 21
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, comm[\'iopts\'] has not been initialized correctly.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 28 and _ifmt == 28:
            _ierr_fl = 21
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, comm[\'iopts\'] has not been initialized correctly.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 32 and _ifmt == 32:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, at least one value of i, for i = 1,2,...,nvpr,',
                    'does not appear in vpr.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 101 and _ifmt == 101:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Optimal solution found, but requested accuracy not achieved.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 102 and _ifmt == 102:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Too many major iterations.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 103 and _ifmt == 103:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Current point cannot be improved upon.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 104 and _ifmt == 104:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'At least one negative estimate for gamma was obtained.',
                    'All negative estimates have been set to zero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _gamma_cls.refresh_py(_gamma)
        _estid_cls.refresh_py(_estid)
        _b_cls.refresh_py(_b)
        _se_cls.refresh_py(_se)
        _czz_cls.refresh_py(_czz)
        _cxx_cls.refresh_py(_cxx)
        _cxz_cls.refresh_py(_cxz)
        _effn_py = utils._EngineIntScalarType.to_py(_effn)
        _rnkx_py = utils._EngineIntScalarType.to_py(_rnkx)
        _ncov_py = utils._EngineIntScalarType.to_py(_ncov)
        _lnlike_py = utils._EngineFloat64ScalarType.to_py(_lnlike)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['effn', 'rnkx', 'ncov', 'lnlike'],
        )(
            _effn_py,
            _rnkx_py,
            _ncov_py,
            _lnlike_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

@utils._withdrawal(
    'naginterfaces.base.correg.mixeff_hier_ml',
    mkwdrn=None,
    advice=[
        'Please use lmm_fit instead.',
        'See also https://www.nag.com/numeric/py/nagdoc_latest/replace.html',
    ]
)
def mixeff_hier_ml(lvpr, vpr, nvpr, gamma, lb, estid, ldid, b, se, czz, cxx, cxz, comm, iopt, liopt, ropt, lropt, io_manager=None):
    u"""
    Hierarchical mixed effects regression using Maximum Likelihood (ML).

    ``mixeff_hier_ml`` fits a multi-level linear mixed effects
    regression model using maximum likelihood (ML).
    Prior to calling ``mixeff_hier_ml`` the initialization function
    ``mixeff_hier_init`` must be called.

    Note: ``mixeff_hier_ml`` is deprecated.
    Please use ``lmm_fit`` instead.
    See also https://www.nag.com/numeric/py/nagdoc_latest/replace.html

    For full information please refer to the NAG Library document for
    g02je

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02jef.html

    Parameters
    ----------
    lvpr : int
        The sum of the number of random parameters and the random
        intercept flags specified in the call to ``mixeff_hier_init``.

    vpr : int, shape (`lvpr`)
        A vector of flags indicating the mapping between the random
        variables specified in rndm and the variance components,
        sigma_i^2. See [Further Comments] for more details.

    nvpr : int
        g, the number of variance components being estimated (excluding
        the overall variance, sigma_R^2).

    gamma : float, shape (`nvpr`+1), modified in place
        `On entry`: holds the initial values of the variance components,
        gamma_0, with `gamma`[i-1] the initial value for
        sigma_i^2/sigma_R^2, for i = 1,2,...,`nvpr`.

        If `gamma`[0] = -1.0, the remaining elements of `gamma` are
        ignored and the initial values for the variance components are
        estimated from the data using MIVQUE0.

        `On exit`: `gamma`[i-1], for i = 1,2,...,`nvpr`, holds the final
        estimate of sigma_i^2 and `gamma`[`nvpr`] holds the final
        estimate for sigma_R^2.

    lb : int
        The dimension of the arrays `b` and `se`.

    estid : int, shape (`ldid`, `lb`), modified in place
        `On exit`: an array describing the parameter estimates returned
        in `b`. The first nlsv*nrf columns of `estid` describe the
        parameter estimates for the random effects and the last nff
        columns the parameter estimates for the fixed effects.

        For fixed effects:

            for l = nrf*nlsv+1,...,nrf*nlsv+nff

                if `b`[l-1] contains the parameter estimate for the
                intercept, then

                        `estid`[0,l-1] = `estid`[1,l-1] = `estid`[2,l-1]
                        = 0;

                if `b`[l-1] contains the parameter estimate for the ith
                level of the jth fixed variable, that is the vector of
                values held in the kth column of dat when fixed[j+1] = k
                then

                        `estid`[0,l-1] = 0, ; `estid`[1,l-1] = j, ;
                        `estid`[2,l-1] = i;

                if the jth variable is continuous or binary, that is
                levels[fixed[j+1]-1] = 1, `estid`[2,l-1] = 0;

                any remaining rows of the lth column of `estid` are set
                to 0.

        For random effects:

            let

                N_R_b denote the number of random variables in the bth
                random statement, that is N_R_b = rndm(1,b);

                R_{jb} denote the jth random variable from the bth
                random statement, that is the vector of values held in
                the kth column of dat when rndm(2+j,b) = k;

                N_S_b denote the number of subject variables in the bth
                random statement, that is N_S_b = rndm(3+N_R_b,b);

                S_{jb} denote the jth subject variable from the bth
                random statement, that is the vector of values held in
                the kth column of dat when rndm(3+N_R_b+j,b) = k;

                L(S_{jb}) denote the number of levels for S_{jb}, that
                is L(S_{jb}) = levels[rndm(3+N_R_b+j,b)-1];

            then

                for l = 1,2,...nrf*nlsv, if `b`[l-1] contains the
                parameter estimate for the ith level of R_{jb} when
                S_{kb} = s_k, for k = 1,2,...,N_S_b and 1 <= s_k <=
                L(S_{jb}), i.e., s_k is a valid value for the kth
                subject variable, then

                        `estid`[0,l-1] = b, ; `estid`[1,l-1] = j, ;
                        `estid`[2,l-1] = i, ; `estid`[3+k-1,l-1] = s_k,
                        k = 1,2,...,N_S_b;

                if the parameter being estimated is for the intercept,
                then `estid`[1,l-1] = `estid`[2,l-1] = 0;

                if the jth variable is continuous, or binary, that is
                L(S_{jb}) = 1, then `estid`[2,l-1] = 0;

                the remaining rows of the lth column of `estid` are set
                to 0.

        In some situations, certain combinations of variables are never
        observed.

        In such circumstances all elements of the lth row of `estid` are
        set to -999.

    ldid : int
        The first dimension of the array `estid`.

    b : float, shape (`lb`), modified in place
        `On exit`: the parameter estimates, with the first nrf*nlsv
        elements of `b` containing the parameter estimates for the
        random effects, nu, and the remaining nff elements containing
        the parameter estimates for the fixed effects, beta. The order
        of these estimates are described by the `estid` argument.

    se : float, shape (`lb`), modified in place
        `On exit`: the standard errors of the parameter estimates given
        in `b`.

    czz : float, shape (nrf, :), modified in place
        Note: the required extent for this argument in dimension 2 is
        determined as follows: if vlicomm>=8:
        `comm`['iopts'][6]*`comm`['iopts'][7]; otherwise: 0.

        `On exit`: if nlsv = 1, then `czz` holds the lower triangular
        portion of the matrix (1/sigma^2)(Z^TR^^-1Z+G^^-1), where R^ and
        G^ are the estimates of R and G respectively. If nlsv > 1, then
        `czz` holds this matrix in compressed form, with the first nrf
        columns holding the part of the matrix corresponding to the
        first level of the overall subject variable, the next nrf
        columns the part corresponding to the second level of the
        overall subject variable etc.

    cxx : float, shape (nff, :), modified in place
        Note: the required extent for this argument in dimension 2 is
        determined as follows: if vlicomm>=6: `comm`['iopts'][5];
        otherwise: 0.

        `On exit`: `cxx` holds the lower triangular portion of the
        matrix (1/sigma^2)X^TV^^-1X, where V^ is the estimated value of
        V.

    cxz : float, shape (nff, :), modified in place
        Note: the required extent for this argument in dimension 2 is
        determined as follows: if vlicomm>=8:
        `comm`['iopts'][6]*`comm`['iopts'][7]; otherwise: 0.

        `On exit`: if nlsv = 1, then `cxz` holds the matrix
        (1/sigma^2)(X^TV^^-1Z)G^, where V^ and G^ are the estimates of V
        and G respectively. If nlsv > 1, then `cxz` holds this matrix in
        compressed form, with the first nrf columns holding the part of
        the matrix corresponding to the first level of the overall
        subject variable, the next nrf columns the part corresponding to
        the second level of the overall subject variable etc.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``mixeff_hier_init``.

    iopt : None or int, shape (`liopt`)
        Options passed to the optimization function.

        By default ``mixeff_hier_ml`` fits the specified model using a
        modified Newton optimization algorithm as implemented in
        ``opt.bounds_mod_deriv2_comp``. In some cases, where the
        calculation of the derivatives is computationally expensive it
        may be more efficient to use a sequential QP algorithm.

        The sequential QP algorithm as implemented in ``opt.nlp1_solve``
        can be chosen by setting `iopt`[4] = 1.

        If `liopt` < 5 or `iopt`[4] != 1, then the modified Newton
        algorithm will be used.

        Different options are available depending on the optimization
        function used.

        In all cases, using a value of -1 will cause the default value
        to be used.

        In addition only the first `liopt` values of `iopt` are used, so
        for example, if only the first element of `iopt` needs changing
        and default values for all other options are sufficient `liopt`
        can be set to 1.

        The following table lists the association between elements of
        `iopt` and arguments in the optimizer when the modified Newton
        algorithm is being used.

        [table omitted]

        If requested, monitoring information is displayed in a similar
        format to that given by the modified Newton optimizer.

        The following table lists the association between elements of
        `iopt` and options in the optimizer when the sequential QP
        algorithm is being used.

        [table omitted]

        If `liopt` <= 0, default values are used for all options and
        `iopt` is not referenced.

    liopt : int
        Length of the options array `iopt`.

    ropt : None or float, shape (`lropt`)
        Options passed to the optimization function.

        Different options are available depending on the optimization
        function used.

        In all cases, using a value of -1.0 will cause the default value
        to be used.

        In addition only the first `lropt` values of `ropt` are used, so
        for example, if only the first element of `ropt` needs changing
        and default values for all other options are sufficient `lropt`
        can be set to 1.

        The following table lists the association between elements of
        `ropt` and arguments in the optimizer when the modified Newton
        algorithm is being used.

        [table omitted]

        The following table lists the association between elements of
        `ropt` and options in the optimizer when the sequential QP
        algorithm is being used.

        [table omitted]

        where eps is the machine precision returned by
        ``machine.precision`` and zz_{ii} denotes the i diagonal element
        of Z^TZ.

        If `lropt` <= 0, then default values are used for all options
        and `ropt` may be set to None.

    lropt : int
        Length of the options array `ropt`.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    effn : int
        Effective number of observations. If no weights were supplied to
        ``mixeff_hier_init`` or all supplied weights were nonzero,
        `effn` = n.

    rnkx : int
        The rank of the design matrix, X, for the fixed effects.

    ncov : int
        Number of variance components not estimated to be zero. If none
        of the variance components are estimated to be zero, then `ncov`
        = `nvpr`.

    lnlike : float
        -2l_R(gamma^) where l_R is the log of the maximum likelihood
        calculated at gamma^, the estimated variance components returned
        in `gamma`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `lvpr` = *<value>*.

            Constraint: `lvpr` >= *<value>*.

        (`errno` 2)
            On entry, `vpr`[*<value>*] = *<value>* and `nvpr` =
            *<value>*.

            Constraint: 1 <= `vpr`[i] <= `nvpr`.

        (`errno` 3)
            On entry, `nvpr` = *<value>*.

            Constraint: 1 <= `nvpr` <= *<value>*.

        (`errno` 4)
            On entry, `gamma`[*<value>*] = *<value>*.

            Constraint: `gamma`[0] = -1.0 or `gamma`[i-1] >= 0.0.

        (`errno` 9)
            On entry, `lb` = *<value>*.

            Constraint: `lb` >= *<value>*.

        (`errno` 11)
            On entry, `ldid` = *<value>*.

            Constraint: `ldid` >= *<value>*.

        (`errno` 21)
            On entry, `comm`['iopts'] has not been initialized
            correctly.

        (`errno` 32)
            On entry, at least one value of i, for i = 1,2,...,`nvpr`,
            does not appear in `vpr`.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 101)
            Optimal solution found, but requested accuracy not achieved.

        (`errno` 102)
            Too many major iterations.

        (`errno` 103)
            Current point cannot be improved upon.

        (`errno` 104)
            At least one negative estimate for `gamma` was obtained. All
            negative estimates have been set to zero.

    References
    ----------
    Goodnight, J H, 1979, `A tutorial on the SWEEP operator`, The
    American Statistician (33(3)), 149--158

    Harville, D A, 1977, `Maximum likelihood approaches to variance
    component estimation and to related problems`, JASA (72), 320--340

    Rao, C R, 1972, `Estimation of variance and covariance components in
    a linear model`, J. Am. Stat. Assoc. (67), 112--115

    Stroup, W W, 1989, `Predictable functions and prediction space in
    the mixed model procedure`, Applications of Mixed Models in
    Agriculture and Related Disciplines (Southern Cooperative Series
    Bulletin No. 343), 39--48

    Wolfinger, R, Tobias, R and Sall, J, 1994, `Computing Gaussian
    likelihoods and their derivatives for general linear mixed models`,
    SIAM Sci. Statist. Comput. (15), 1294--1310
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.mixeff_hier_ml'

    utils.FileObjManager._check_type_compat(
        io_manager, {'fun_name': fun_name, 'entity': 'io_manager'}
    )

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    def cb_print_rech_py(_, en_data, nout, rec, ifail, rec_len):
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data, utils._EngineIntScalarType.to_py(nout.contents), rec[:rec_len], io_manager, excs, ifail.contents,
                fun_name=fun_name,
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02jeft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._PRINT_REC_FUNCTYPE,
        utils._PRINT_RECH_FUNCTYPE,
        utils._BYREF_INT, # Input, lvpr.
        utils._EngineIntArrayType.c_type, # Input, vpr[lvpr].
        utils._BYREF_INT, # Input, nvpr.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, gamma[nvpr+1].
        utils._BYREF_INT, # Output, effn.
        utils._BYREF_INT, # Output, rnkx.
        utils._BYREF_INT, # Output, ncov.
        utils._BYREF_FLOAT64, # Output, lnlike.
        utils._BYREF_INT, # Input, lb.
        utils._EngineIntArrayType.c_type, # Output, estid[ldid, tdid].
        utils._BYREF_INT, # Input, ldid.
        utils._EngineFloat64ArrayType.c_type, # Output, b[lb].
        utils._EngineFloat64ArrayType.c_type, # Output, se[lb].
        utils._EngineFloat64ArrayType.c_type, # Output, czz[ldczz, tdczz].
        utils._BYREF_INT, # Input, ldczz.
        utils._EngineFloat64ArrayType.c_type, # Output, cxx[ldcxx, tdcxx].
        utils._BYREF_INT, # Input, ldcxx.
        utils._EngineFloat64ArrayType.c_type, # Output, cxz[ldcxz, tdcxz].
        utils._BYREF_INT, # Input, ldcxz.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, rcomm[lrcomm].
        utils._EngineIntArrayType.c_type, # Input/Output, icomm[licomm].
        utils._EngineIntArrayType.c_type, # Input, iopt[liopt].
        utils._BYREF_INT, # Input, liopt.
        utils._EngineFloat64ArrayType.c_type, # Input, ropt[lropt].
        utils._BYREF_INT, # Input, lropt.
        utils._BYREF_INT, # Input, licomm.
        utils._BYREF_INT, # Input, lrcomm.
        utils._BYREF_INT, # Input, tdid.
        utils._BYREF_INT, # Input, tdczz.
        utils._BYREF_INT, # Input, tdcxx.
        utils._BYREF_INT, # Input, tdcxz.
        utils._BYREF_STRING, # Input, routine_name(len=6).
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of routine_name.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((estid, 'estid'), (czz, 'czz'), (cxx, 'cxx'), (cxz, 'cxz')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _print_rech = utils._PRINT_RECH_FUNCTYPE(cb_print_rech_py)
    _lvpr = utils._EngineIntScalarType(
        lvpr, {'fun_name': fun_name, 'entity': 'lvpr'},
    ).to_ctypes()
    _lvpr_py = utils._EngineIntScalarType.to_py(_lvpr)
    _vpr = utils._EngineIntArrayType(
        vpr, {'fun_name': fun_name, 'entity': 'vpr'},
        exp_shape=(_lvpr_py,),
    ).to_ctypes()
    _nvpr = utils._EngineIntScalarType(
        nvpr, {'fun_name': fun_name, 'entity': 'nvpr'},
    ).to_ctypes()
    _nvpr_py = utils._EngineIntScalarType.to_py(_nvpr)
    _gamma_cls = utils._EngineFloat64ArrayType(
        gamma, {'fun_name': fun_name, 'entity': 'gamma'},
        exp_shape=(_nvpr_py+1,),
    )
    _gamma = _gamma_cls.to_ctypes()
    _effn = utils._EngineIntScalarType.empty_ctypes()
    _rnkx = utils._EngineIntScalarType.empty_ctypes()
    _ncov = utils._EngineIntScalarType.empty_ctypes()
    _lnlike = utils._EngineFloat64ScalarType.empty_ctypes()
    _lb = utils._EngineIntScalarType(
        lb, {'fun_name': fun_name, 'entity': 'lb'},
    ).to_ctypes()
    _ldid = utils._EngineIntScalarType(
        ldid, {'fun_name': fun_name, 'entity': 'ldid'},
    ).to_ctypes()
    _tdid_py = utils._EngineIntArrayType.get_shape(estid, {'fun_name': fun_name, 'entity': 'estid'}, exp_rank=2)[1]
    _tdid = utils._EngineIntScalarType(
        _tdid_py, {'fun_name': fun_name, 'entity': '_tdid'},
    ).to_ctypes()
    utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        req_keys=('opts', 'iopts'),
    )
    _vl_vlicomm_py = utils._EngineIntArrayType.get_element(comm['iopts'], {'fun_name': fun_name, 'entity': 'comm[\'iopts\']'}, (0,), validate=False)
    _lb_py = utils._EngineIntScalarType.to_py(_lb)
    if _vl_vlicomm_py >= 5:
        _estid_dim1 = utils._EngineIntArrayType.get_element(comm['iopts'], {'fun_name': fun_name, 'entity': 'comm[\'iopts\']'}, (4,), validate=False)
    else:
        _estid_dim1 = 0
    exp_shape_estid = (_estid_dim1, _lb_py)
    _estid_cls = utils._EngineIntArrayType(
        estid, {'fun_name': fun_name, 'entity': 'estid'},
        exp_shape=exp_shape_estid,
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _estid = _estid_cls.to_ctypes(can_reference_contents=False)
    _vl_g02jcf_nff_py = utils._EngineIntArrayType.get_element(comm['iopts'], {'fun_name': fun_name, 'entity': 'comm[\'iopts\']'}, (5,), validate=False)
    _vl_g02jcf_nrf_py = utils._EngineIntArrayType.get_element(comm['iopts'], {'fun_name': fun_name, 'entity': 'comm[\'iopts\']'}, (6,), validate=False)
    _vl_g02jcf_nlsv_py = utils._EngineIntArrayType.get_element(comm['iopts'], {'fun_name': fun_name, 'entity': 'comm[\'iopts\']'}, (7,), validate=False)
    _b_cls = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_vl_g02jcf_nff_py+_vl_g02jcf_nrf_py*_vl_g02jcf_nlsv_py,),
        check_contents=False,
    )
    _b = _b_cls.to_ctypes(can_reference_contents=False)
    _se_cls = utils._EngineFloat64ArrayType(
        se, {'fun_name': fun_name, 'entity': 'se'},
        exp_shape=(_vl_g02jcf_nff_py+_vl_g02jcf_nrf_py*_vl_g02jcf_nlsv_py,),
        check_contents=False,
    )
    _se = _se_cls.to_ctypes(can_reference_contents=False)
    _ldczz_py = utils._EngineFloat64ArrayType.get_nd(czz, {'fun_name': fun_name, 'entity': 'czz'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldczz = utils._EngineIntScalarType(
        _ldczz_py, {'fun_name': fun_name, 'entity': '_ldczz'},
    ).to_ctypes()
    _tdczz_py = utils._EngineFloat64ArrayType.get_nd(czz, {'fun_name': fun_name, 'entity': 'czz'}, exp_rank=2, sorder=_m_sorder, dim_type='s')
    _tdczz = utils._EngineIntScalarType(
        _tdczz_py, {'fun_name': fun_name, 'entity': '_tdczz'},
    ).to_ctypes()
    if _vl_vlicomm_py >= 8:
        _czz_dim2 = utils._EngineIntArrayType.get_element(comm['iopts'], {'fun_name': fun_name, 'entity': 'comm[\'iopts\']'}, (6,), validate=False)*utils._EngineIntArrayType.get_element(comm['iopts'], {'fun_name': fun_name, 'entity': 'comm[\'iopts\']'}, (7,), validate=False)
    else:
        _czz_dim2 = 0
    exp_shape_czz = (_vl_g02jcf_nrf_py, _czz_dim2)
    _czz_cls = utils._EngineFloat64ArrayType(
        czz, {'fun_name': fun_name, 'entity': 'czz'},
        exp_shape=exp_shape_czz,
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _czz = _czz_cls.to_ctypes(can_reference_contents=False)
    _ldcxx_py = utils._EngineFloat64ArrayType.get_nd(cxx, {'fun_name': fun_name, 'entity': 'cxx'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldcxx = utils._EngineIntScalarType(
        _ldcxx_py, {'fun_name': fun_name, 'entity': '_ldcxx'},
    ).to_ctypes()
    _tdcxx_py = utils._EngineFloat64ArrayType.get_nd(cxx, {'fun_name': fun_name, 'entity': 'cxx'}, exp_rank=2, sorder=_m_sorder, dim_type='s')
    _tdcxx = utils._EngineIntScalarType(
        _tdcxx_py, {'fun_name': fun_name, 'entity': '_tdcxx'},
    ).to_ctypes()
    if _vl_vlicomm_py >= 6:
        _cxx_dim2 = utils._EngineIntArrayType.get_element(comm['iopts'], {'fun_name': fun_name, 'entity': 'comm[\'iopts\']'}, (5,), validate=False)
    else:
        _cxx_dim2 = 0
    exp_shape_cxx = (_vl_g02jcf_nff_py, _cxx_dim2)
    _cxx_cls = utils._EngineFloat64ArrayType(
        cxx, {'fun_name': fun_name, 'entity': 'cxx'},
        exp_shape=exp_shape_cxx,
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _cxx = _cxx_cls.to_ctypes(can_reference_contents=False)
    _ldcxz_py = utils._EngineFloat64ArrayType.get_nd(cxz, {'fun_name': fun_name, 'entity': 'cxz'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldcxz = utils._EngineIntScalarType(
        _ldcxz_py, {'fun_name': fun_name, 'entity': '_ldcxz'},
    ).to_ctypes()
    _tdcxz_py = utils._EngineFloat64ArrayType.get_nd(cxz, {'fun_name': fun_name, 'entity': 'cxz'}, exp_rank=2, sorder=_m_sorder, dim_type='s')
    _tdcxz = utils._EngineIntScalarType(
        _tdcxz_py, {'fun_name': fun_name, 'entity': '_tdcxz'},
    ).to_ctypes()
    if _vl_vlicomm_py >= 8:
        _cxz_dim2 = utils._EngineIntArrayType.get_element(comm['iopts'], {'fun_name': fun_name, 'entity': 'comm[\'iopts\']'}, (6,), validate=False)*utils._EngineIntArrayType.get_element(comm['iopts'], {'fun_name': fun_name, 'entity': 'comm[\'iopts\']'}, (7,), validate=False)
    else:
        _cxz_dim2 = 0
    exp_shape_cxz = (_vl_g02jcf_nff_py, _cxz_dim2)
    _cxz_cls = utils._EngineFloat64ArrayType(
        cxz, {'fun_name': fun_name, 'entity': 'cxz'},
        exp_shape=exp_shape_cxz,
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _cxz = _cxz_cls.to_ctypes(can_reference_contents=False)
    _lrcomm_py = utils._EngineFloat64ArrayType.get_shape(comm['opts'], {'fun_name': fun_name, 'entity': 'comm[\'opts\']'}, exp_rank=1, validate=False)[0]
    _lrcomm = utils._EngineIntScalarType(
        _lrcomm_py, {'fun_name': fun_name, 'entity': '_lrcomm'},
    ).to_ctypes()
    _rcomm = comm['opts']
    _licomm_py = utils._EngineIntArrayType.get_shape(comm['iopts'], {'fun_name': fun_name, 'entity': 'comm[\'iopts\']'}, exp_rank=1, validate=False)[0]
    _licomm = utils._EngineIntScalarType(
        _licomm_py, {'fun_name': fun_name, 'entity': '_licomm'},
    ).to_ctypes()
    _icomm = comm['iopts']
    _liopt = utils._EngineIntScalarType(
        liopt, {'fun_name': fun_name, 'entity': 'liopt'},
    ).to_ctypes()
    _liopt_py = utils._EngineIntScalarType.to_py(_liopt)
    if _liopt_py > 0:
        if iopt is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'iopt'},
                'must not be None when liopt > 0',
            )
    _iopt = utils._EngineIntArrayType(
        iopt, {'fun_name': fun_name, 'entity': 'iopt'},
        exp_shape=(_liopt_py,),
        check_contents=_liopt_py > 0,
        allow_none=True,
    ).to_ctypes(can_reference_contents=_liopt_py > 0)
    _lropt = utils._EngineIntScalarType(
        lropt, {'fun_name': fun_name, 'entity': 'lropt'},
    ).to_ctypes()
    _lropt_py = utils._EngineIntScalarType.to_py(_lropt)
    if _lropt_py > 0:
        if ropt is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'ropt'},
                'must not be None when lropt > 0',
            )
    _ropt = utils._EngineFloat64ArrayType(
        ropt, {'fun_name': fun_name, 'entity': 'ropt'},
        exp_shape=(_lropt_py,),
        check_contents=_lropt_py > 0,
        allow_none=True,
    ).to_ctypes(can_reference_contents=_lropt_py > 0)
    _routine_name = utils._EngineRoutineName(*b'g02je ')
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        utils._UNUSED_PRINT_REC,
        _print_rech,
        _lvpr,
        _vpr,
        _nvpr,
        _gamma,
        _effn,
        _rnkx,
        _ncov,
        _lnlike,
        _lb,
        _estid,
        _ldid,
        _b,
        _se,
        _czz,
        _ldczz,
        _cxx,
        _ldcxx,
        _cxz,
        _ldcxz,
        _rcomm,
        _icomm,
        _iopt,
        _liopt,
        _ropt,
        _lropt,
        _licomm,
        _lrcomm,
        _tdid,
        _tdczz,
        _tdcxx,
        _tdcxz,
        _routine_name,
        _errbuf,
        _ierr,
        _routine_name._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lvpr = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: lvpr >= {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, vpr[{:s}] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'nvpr = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: 1 <= vpr[i] <= nvpr.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nvpr = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: 1 <= nvpr <= {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, gamma[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: gamma[0] = -1.0 or gamma[i-1] >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 10 and _ifmt == 10:
            _ierr_fl = 9
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lb = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: lb >= {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 13 and _ifmt == 13:
            _ierr_fl = 9
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lb = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: lb >= {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 12 and _ifmt == 12:
            _ierr_fl = 11
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ldid = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ldid >= {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 18 and _ifmt == 18:
            _ierr_fl = 21
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, comm[\'iopts\'] has not been initialized correctly.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 21 and _ifmt == 21:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, comm[\'iopts\'] has not been initialized correctly.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 24 and _ifmt == 24:
            _ierr_fl = 21
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, comm[\'iopts\'] has not been initialized correctly.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 26 and _ifmt == 26:
            _ierr_fl = 21
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, comm[\'iopts\'] has not been initialized correctly.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 27 and _ifmt == 27:
            _ierr_fl = 21
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, comm[\'iopts\'] has not been initialized correctly.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 28 and _ifmt == 28:
            _ierr_fl = 21
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, comm[\'iopts\'] has not been initialized correctly.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 32 and _ifmt == 32:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, at least one value of i, for i = 1,2,...,nvpr,',
                    'does not appear in vpr.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 101 and _ifmt == 101:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Optimal solution found, but requested accuracy not achieved.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 102 and _ifmt == 102:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Too many major iterations.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 103 and _ifmt == 103:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Current point cannot be improved upon.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 104 and _ifmt == 104:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'At least one negative estimate for gamma was obtained.',
                    'All negative estimates have been set to zero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _gamma_cls.refresh_py(_gamma)
        _estid_cls.refresh_py(_estid)
        _b_cls.refresh_py(_b)
        _se_cls.refresh_py(_se)
        _czz_cls.refresh_py(_czz)
        _cxx_cls.refresh_py(_cxx)
        _cxz_cls.refresh_py(_cxz)
        _effn_py = utils._EngineIntScalarType.to_py(_effn)
        _rnkx_py = utils._EngineIntScalarType.to_py(_rnkx)
        _ncov_py = utils._EngineIntScalarType.to_py(_ncov)
        _lnlike_py = utils._EngineFloat64ScalarType.to_py(_lnlike)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['effn', 'rnkx', 'ncov', 'lnlike'],
        )(
            _effn_py,
            _rnkx_py,
            _ncov_py,
            _lnlike_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def lmm_init(hlmm, hddesc, hfixed, nrndm, hrndm, n, y, wt, dat):
    u"""
    Linear mixed effects regression, initialization function for
    ``lmm_fit``.

    Note: this function uses optional algorithmic parameters, see also:
    ``blgm.optset``, ``blgm.optget``, ``lmm_fit``.

    ``lmm_init`` preprocesses a dataset prior to fitting a linear mixed
    effects regression model via ``lmm_fit``.

    For full information please refer to the NAG Library document for
    g02jf

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02jff.html

    Parameters
    ----------
    hlmm : Handle, modified in place
        `On entry`: must be set to a null Handle or, alternatively, an
        existing G22 handle may be supplied in which case ``lmm_init``
        will destroy the supplied G22 handle as if ``blgm.handle_free``
        had been called.

        `On exit`: holds a G22 handle to the internal data structure
        containing a description of the model. You **must not** change
        the G22 handle other than through the functions in submodule
        ``correg`` or submodule ``blgm``.

    hddesc : Handle
        A G22 handle to the internal data structure containing a
        description of the data matrix, D, as returned in hddesc by
        ``blgm.lm_describe_data``.

    hfixed : Handle
        A G22 handle to the internal data structure containing a
        description of the fixed part of the model M_f as returned in
        hform by ``blgm.lm_formula``.

        If `hfixed` is a null Handle then the model is assumed to not
        have a fixed part.

    nrndm : int
        The number of elements used to describe the random part of the
        model.

    hrndm : None or Handle, shape (`nrndm`)
        A series of G22 handles to internal data structures containing a
        description of the random part of the model M_r as returned in
        hform by ``blgm.lm_formula``.

    n : int
        n, the number of observations in the dataset, D.

    y : float, shape (`n`)
        y, the vector of observations on the dependent variable.

    wt : None or float, shape (`n`)
        Optionally, the diagonal elements of the weight matrix W_c.

        If `wt`[i-1] = 0.0, the ith observation is not included in the
        model and the effective number of observations is the number of
        observations with nonzero weights.

        If weights are not provided then `wt` must be set to None, and
        the effective number of observations is n.

    dat : float, shape (:, :)
        The data matrix, D. By default, D_{ij}, the ith value for the
        jth variable, for j = 1,2,...,m_d, for i = 1,2,...,n, should be
        supplied in `dat`[i-1,j-1].

        If the option 'Storage Order', described in
        ``blgm.lm_describe_data``, is set to 'VAROBS', D_{ij} should be
        supplied in `dat`[j-1,i-1].

        If either y_i, w_i or D_{ij}, for a variable j used in the
        model, is NaN (Not A Number) then that value is treated as
        missing and the whole observation is excluded from the analysis.

    Returns
    -------
    fnlsv : int
        The number of levels for the overall subject variable in M_f. If
        there is no overall subject variable, `fnlsv` = 1.

    nff : int
        The number of fixed effects estimated in each of the `fnlsv`
        subject blocks. The number of columns, p, in the design matrix X
        is given by p = `nff`*`fnlsv`.

    rnlsv : int
        The number of levels for the overall subject variable in M_r. If
        there is no overall subject variable, `rnlsv` = 1.

    nrf : int
        The number of random effects estimated in each of the `rnlsv`
        subject blocks. The number of columns, q, in the design matrix Z
        is given by q = `nrf`*`rnlsv`.

    nvpr : int
        g, the number of variance components being estimated (excluding
        the overall variance, sigma_R^2). This is defined by the number
        of terms in the random part of the model, M_r (see [Algorithmic
        Details] for details).

    comm : dict, communication object
        Communication structure.

    Other Parameters
    ----------------
    'Gamma Lower Bound' : float
        Default = sqrt(machine precision)/100

        A lower bound for the elements of gamma*, where gamma* =
        gamma/sigma_R^2.

    'Gamma Upper Bound' : float
        Default = 10^20

        An upper bound for the elements of gamma*, where gamma* =
        gamma/sigma_R^2.

    'Initial Distance' : float
        Default = 100000.0

        The initial distance from the solution.

            When 'Solver' = 'E04LB', ``lmm_fit`` passes 'Initial
            Distance' to the solver as stepmx.

            When 'Solver' = 'E04UC', this option is ignored.

    'Initial Value Strategy' : int
        Default = special

        Controls how ``lmm_fit`` will choose the initial values for the
        variance components, gamma, if not supplied.

        'Initial Value Strategy' = 0
            The MIVQUE0 estimates of the variance components based on
            the likelihood specified by 'Likelihood' are used.

        'Initial Value Strategy' = 1
            The MIVQUE0 estimates based on the maximum likelihood are
            used, irrespective of the value of 'Likelihood'.

        See Rao (1972) for a description of the minimum variance
        quadratic unbiased estimators (MIVQUE0).

        By default, for small problems, 'Initial Value Strategy' = 0 and
        for large problems 'Initial Value Strategy' = 1.

    'Likelihood' : str
        Default = 'REML'

        'Likelihood' defines whether ``lmm_fit`` will use the restricted
        maximum likelihood (REML) or the maximum likelihood (ML) when
        fitting the model.

    'Linear Minimization Accuracy' : float
        Default = 0.9

        The accuracy of the linear minimizations.

            When 'Solver' = 'E04LB', ``lmm_fit`` passes 'Linear
            Minimization Accuracy' to the solver as eta.

            When 'Solver' = 'E04UC', this option is ignored.

    'Line Search Tolerance' : float
        Default = 0.9

        The line search tolerance.

            When 'Solver' = 'E04LB', this option is ignored.

            When 'Solver' = 'E04UC', ``lmm_fit`` passes 'Line Search
            Tolerance' to the solver as 'Line Search Tolerance'.

    'List' : valueless
        Option 'List' enables printing of each option specification as
        it is supplied. 'NoList' suppresses this printing.

    'NoList' : valueless
        Default

        Option 'List' enables printing of each option specification as
        it is supplied. 'NoList' suppresses this printing.

    'Major Iteration Limit' : int
        Default = special

        The number of major iterations.

            When 'Solver' = 'E04LB', ``lmm_fit`` passes 'Major Iteration
            Limit' to the solver as maxcal. In this case, the default
            value used is 1000.

            When 'Solver' = 'E04UC', ``lmm_fit`` passes 'Major Iteration
            Limit' to the solver as 'Major Iteration Limit'. In this
            case, the default value used is max(50,3*g), where g is the
            number of variance components being estimated (excluding the
            overall variance, sigma_R^2).

    'Major Print Level' : int
        Default = special

        The frequency that monitoring information is output to 'Unit
        Number'.

            When 'Solver' = 'E04LB', ``lmm_fit`` passes 'Major Print
            Level' to the solver as iprint. In this case, the default
            value used is -1 and hence no monitoring information will be
            output.

            When 'Solver' = 'E04UC', ``lmm_fit`` passes 'Major Print
            Level' to the solver as 'Major Print Level'. In this case,
            the default value used is 0 and hence no monitoring
            information will be output.

    'Maximum Number of Threads' : int
        Default = special

        Controls the maximum number of threads used by ``lmm_fit`` in a
        multithreaded library.
        By default, the maximum number of available threads are used.

        In a library that is not multithreaded, this option has no
        effect.

    'Minor Iteration Limit' : int
        Default = max(50,3*g)

        The number of minor iterations.

            When 'Solver' = 'E04LB', this option is ignored.

            When 'Solver' = 'E04UC', ``lmm_fit`` passes 'Minor Iteration
            Limit' to the solver as 'Minor Iteration Limit'. In this
            case, the default value used is max(50,3*g), where g is the
            number of variance components being estimated (excluding the
            overall variance, sigma_R^2).

    'Minor Print Level' : int
        Default = 0

        The frequency that additional monitoring information is output
        to 'Unit Number'.

            When 'Solver' = 'E04LB', this option is ignored.

            When 'Solver' = 'E04UC', ``lmm_fit`` passes 'Minor Print
            Level' to the solver as 'Minor Print Level'. The default
            value of 0 means that no additional monitoring information
            will be output.

    'Optimality Tolerance' : float
        Default = machine precision^0.72

        The optimality tolerance.

            When 'Solver' = 'E04LB', this option is ignored.

            When 'Solver' = 'E04UC', ``lmm_fit`` passes 'Optimality
            Tolerance' to the solver as 'Optimality Tolerance'.

    'Parallelisation Strategy' : int
        Default = special

        If 'Maximum Number of Threads' > 0 then 'Parallelisation
        Strategy' controls how ``lmm_fit`` is parallelised in a
        multithreaded library.

        'Parallelisation Strategy' = 1
            ``lmm_fit`` will attempt to parallelise operations involving
            Z, even if `rnlsv` = 1.

        'Parallelisation Strategy' = 2
            ``lmm_fit`` will only attempt to parallelise operations
            involving Z, if `rnlsv` > 1.

        By default, 'Parallelisation Strategy' = 1, however, for some
        models / datasets, this may be slower than using
        'Parallelisation Strategy' = 2 when `rnlsv` = 1.

        In a library that is not multithreaded, this option has no
        effect.

    'Solution Accuracy' : float
        Default = 0.0

        The accuracy to which the solution is required.

            When 'Solver' = 'E04LB', ``lmm_fit`` passes 'Solution
            Accuracy' to the solver as xtol.

            When 'Solver' = 'E04UC', this option is ignored.

    'Solver' : str
        Default = special

        Controls which solver ``lmm_fit`` will use when fitting the
        model.
        By default, 'Solver' = 'E04LB' is used for small problems and
        'Solver' = 'E04UC', otherwise.

        If 'Solver' = 'E04LB', then the solver used is the one
        implemented in ``opt.bounds_mod_deriv2_comp`` and if 'Solver' =
        'E04UC', then the solver used is the one implemented in
        ``opt.nlp1_solve``.

    'Sweep Tolerance' : float
        Default = special

        The sweep tolerance used by ``lmm_fit`` when performing the
        sweep operation Wolfinger `et al.` (1994).
        The default value used is 'Sweep Tolerance' =
        max(epsilon,epsilon*(max_i((Z^T)_{ii}))), where epsilon =
        sqrt(machine precision).

    'Unit Number' : int
        Default = advisory message unit number

        The monitoring unit number to which ``lmm_fit`` will send any
        monitoring information.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, `hlmm` is not a null Handle or a recognised G22
            handle.

        (`errno` 21)
            `hddesc` has not been initialized or is corrupt.

        (`errno` 22)
            `hddesc` is not a G22 handle as generated by
            ``blgm.lm_describe_data``.

        (`errno` 31)
            `hfixed` has not been initialized or is corrupt.

        (`errno` 32)
            `hfixed` is not a G22 handle as generated by
            ``blgm.lm_formula``.

        (`errno` 33)
            A variable name used when creating `hfixed` is not present
            in `hddesc`.

            Variable name: *<value>*.

        (`errno` 41)
            On entry, `nrndm` = *<value>*.

            Constraint: `nrndm` >= 0.

        (`errno` 51)
            i = *<value>*.

            `hrndm`[i-1] has not been initialized or is corrupt.

        (`errno` 52)
            i = *<value>*.

            `hrndm`[i-1] is not a G22 handle as generated by
            ``blgm.lm_formula``.

        (`errno` 53)
            No model has been specified.

        (`errno` 54)
            A variable name used when creating `hrndm` is not present in
            `hddesc`.

            Variable name: *<value>*.

        (`errno` 61)
            On entry, lwt = *<value>* and `n` = *<value>*.

            Constraint: lwt = 0 or `n`.

        (`errno` 71)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 72)
            On entry, `n` = *<value>* and n_d = *<value>*.

            Constraint: `n` <= n_d.

        (`errno` 73)
            On entry, no observations due to zero weights or missing
            values.

        (`errno` 91)
            On entry, i = *<value>* and `wt`[i-1] = *<value>*.

            Constraint: `wt`[i-1] >= 0.0.

        (`errno` 101)
            On entry, column j of the data matrix, D, is not consistent
            with information supplied in `hddesc`, j = *<value>*.

        (`errno` 111)
            On entry, n = *<value>* and lddat = *<value>*.

            Constraint: lddat >= n.

        (`errno` 112)
            On entry, m_d = *<value>* and lddat = *<value>*.

            Constraint: lddat >= m_d.

        (`errno` 121)
            On entry, m_d = *<value>* and sddat = *<value>*.

            Constraint: sddat >= m_d.

        (`errno` 122)
            On entry, n = *<value>* and sddat = *<value>*.

            Constraint: sddat >= n.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 34)
            The fixed part of the model contains categorical variables,
            but no intercept or main effects terms have been requested.

        (`errno` 102)
            Column j of the data matrix, D, required rounding more than
            expected when being treated as a categorical variable, j =
            *<value>*.

    References
    ----------
    Rao, C R, 1972, `Estimation of variance and covariance components in
    a linear model`, J. Am. Stat. Assoc. (67), 112--115

    Wolfinger, R, Tobias, R and Sall, J, 1994, `Computing Gaussian
    likelihoods and their derivatives for general linear mixed models`,
    SIAM Sci. Statist. Comput. (15), 1294--1310

    See Also
    --------
    :meth:`naginterfaces.library.examples.correg.lmm_init_combine_ex.main`
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.lmm_init'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02jfft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_HANDLE, # Input/Output, hlmm.
        utils._BYREF_HANDLE, # Input, hddesc.
        utils._BYREF_HANDLE, # Input, hfixed.
        utils._BYREF_INT, # Input, nrndm.
        utils._EngineCptrArrayType.c_type, # Input, hrndm[nrndm].
        utils._BYREF_INT, # Input, lwt.
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, y[n].
        utils._EngineFloat64ArrayType.c_type, # Input, wt[lwt].
        utils._EngineFloat64ArrayType.c_type, # Input, dat[lddat, sddat].
        utils._BYREF_INT, # Input, lddat.
        utils._BYREF_INT, # Input, sddat.
        utils._BYREF_INT, # Output, fnlsv.
        utils._BYREF_INT, # Output, nff.
        utils._BYREF_INT, # Output, rnlsv.
        utils._BYREF_INT, # Output, nrf.
        utils._BYREF_INT, # Output, nvpr.
        utils._EngineFloat64ArrayType.c_type, # Output, rcomm[lrcomm].
        utils._BYREF_INT, # Input, lrcomm.
        utils._EngineIntArrayType.c_type, # Output, icomm[licomm].
        utils._BYREF_INT, # Input, licomm.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((dat, 'dat'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _hlmm = utils._EngineCptrScalarType(
        hlmm, {'fun_name': fun_name, 'entity': 'hlmm'},
    ).to_ctypes()
    _hddesc = utils._EngineCptrScalarType(
        hddesc, {'fun_name': fun_name, 'entity': 'hddesc'},
    ).to_ctypes()
    _hfixed = utils._EngineCptrScalarType(
        hfixed, {'fun_name': fun_name, 'entity': 'hfixed'},
    ).to_ctypes()
    _nrndm = utils._EngineIntScalarType(
        nrndm, {'fun_name': fun_name, 'entity': 'nrndm'},
    ).to_ctypes()
    _nrndm_py = utils._EngineIntScalarType.to_py(_nrndm)
    if _nrndm_py > 0:
        if hrndm is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'hrndm'},
                'must not be None when nrndm > 0',
            )
    _hrndm = utils._EngineCptrArrayType(
        hrndm, {'fun_name': fun_name, 'entity': 'hrndm'},
        exp_shape=(_nrndm_py,),
        check_contents=_nrndm_py > 0,
        allow_none=True,
    ).to_ctypes(can_reference_contents=_nrndm_py > 0)
    _lwt_py = utils._EngineFloat64ArrayType.get_shape(wt, {'fun_name': fun_name, 'entity': 'wt'}, exp_rank=1, allow_none=True)[0]
    _lwt = utils._EngineIntScalarType(
        _lwt_py, {'fun_name': fun_name, 'entity': '_lwt'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    if wt is not None:
        _wt_dim1 = _n_py
    else:
        _wt_dim1 = 0
    exp_shape_wt = (_wt_dim1,)
    _wt = utils._EngineFloat64ArrayType(
        wt, {'fun_name': fun_name, 'entity': 'wt'},
        exp_shape=exp_shape_wt,
        allow_none=True,
    ).to_ctypes()
    _lddat_py = utils._EngineFloat64ArrayType.get_nd(dat, {'fun_name': fun_name, 'entity': 'dat'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _lddat = utils._EngineIntScalarType(
        _lddat_py, {'fun_name': fun_name, 'entity': '_lddat'},
    ).to_ctypes()
    _sddat_py = utils._EngineFloat64ArrayType.get_nd(dat, {'fun_name': fun_name, 'entity': 'dat'}, exp_rank=2, sorder=_m_sorder, dim_type='s')
    _sddat = utils._EngineIntScalarType(
        _sddat_py, {'fun_name': fun_name, 'entity': '_sddat'},
    ).to_ctypes()
    _dat = utils._EngineFloat64ArrayType(
        dat, {'fun_name': fun_name, 'entity': 'dat'},
        exp_shape=(None, None),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _fnlsv = utils._EngineIntScalarType.empty_ctypes()
    _nff = utils._EngineIntScalarType.empty_ctypes()
    _rnlsv = utils._EngineIntScalarType.empty_ctypes()
    _nrf = utils._EngineIntScalarType.empty_ctypes()
    _nvpr = utils._EngineIntScalarType.empty_ctypes()
    _lrcomm = utils._EngineIntScalarType.c_type(2)
    _lrcomm_py = utils._EngineIntScalarType.to_py(_lrcomm)
    _rcomm = utils._EngineFloat64ArrayType.empty_ctypes(_lrcomm_py)
    _comm_py = {}
    _licomm = utils._EngineIntScalarType.c_type(2)
    _licomm_py = utils._EngineIntScalarType.to_py(_licomm)
    _icomm = utils._EngineIntArrayType.empty_ctypes(_licomm_py)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _hlmm,
        _hddesc,
        _hfixed,
        _nrndm,
        _hrndm,
        _lwt,
        _n,
        _y,
        _wt,
        _dat,
        _lddat,
        _sddat,
        _fnlsv,
        _nff,
        _rnlsv,
        _nrf,
        _nvpr,
        _rcomm,
        _lrcomm,
        _icomm,
        _licomm,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
    if _ierr_py == 191 and _ifmt == 191:
        _lrcomm = utils._EngineIntScalarType.c_type(_icomm[1])
        _lrcomm_py = utils._EngineIntScalarType.to_py(_lrcomm)
        _rcomm = utils._EngineFloat64ArrayType.empty_ctypes(_lrcomm_py)
        _comm_py['opts'] = _rcomm
        _licomm = utils._EngineIntScalarType.c_type(_icomm[0])
        _licomm_py = utils._EngineIntScalarType.to_py(_licomm)
        _icomm = utils._EngineIntArrayType.empty_ctypes(_licomm_py)
        _comm_py['iopts'] = _icomm
        engine_ctypes(
            _en_data,
            _hlmm,
            _hddesc,
            _hfixed,
            _nrndm,
            _hrndm,
            _lwt,
            _n,
            _y,
            _wt,
            _dat,
            _lddat,
            _sddat,
            _fnlsv,
            _nff,
            _rnlsv,
            _nrf,
            _nvpr,
            _rcomm,
            _lrcomm,
            _icomm,
            _licomm,
            _errbuf,
            _ierr,
            _errbuf._length_,
        )
        utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 11 and _ifmt == 11:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, hlmm is not a null Handle or a recognised G22 handle.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 21 and _ifmt == 21:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'hddesc has not been initialized or is corrupt.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 22 and _ifmt == 22:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'hddesc is not a G22 handle as generated by blgm.lm_describe_data.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 31 and _ifmt == 31:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'hfixed has not been initialized or is corrupt.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 32 and _ifmt == 32:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'hfixed is not a G22 handle as generated by blgm.lm_formula.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 33 and _ifmt == 33:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'A variable name used when creating hfixed is not present in hddesc.',
                    'Variable name: {:s}.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 34 and _ifmt == 34:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The fixed part of the model contains categorical variables, but no intercept',
                    'or main effects terms have been requested.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 41 and _ifmt == 41:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nrndm = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nrndm >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 51 and _ifmt == 51:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'hrndm[i-1] has not been initialized or is corrupt.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 52 and _ifmt == 52:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'i = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'hrndm[i-1] is not a G22 handle as generated by blgm.lm_formula.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 53 and _ifmt == 53:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'No model has been specified.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 54 and _ifmt == 54:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'A variable name used when creating hrndm is not present in hddesc.',
                    'Variable name: {:s}.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 61 and _ifmt == 61:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lwt = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: lwt = 0 or n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 71 and _ifmt == 71:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 72 and _ifmt == 72:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s} and n_d = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: n <= n_d.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 73 and _ifmt == 73:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, no observations due to zero weights or missing values.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 91 and _ifmt == 91:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s} and wt[i-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: wt[i-1] >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 101 and _ifmt == 101:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, column j of the data matrix, D, is not',
                    'consistent with information supplied in hddesc,',
                    'j = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 102 and _ifmt == 102:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Column j of the data matrix, D, required rounding',
                    'more than expected when being treated as a categorical variable,',
                    'j = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 111 and _ifmt == 111:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s} and lddat = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: lddat >= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 112 and _ifmt == 112:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m_d = {:s} and lddat = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: lddat >= m_d.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 121 and _ifmt == 121:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m_d = {:s} and sddat = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: sddat >= m_d.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 122 and _ifmt == 122:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s} and sddat = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: sddat >= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 191 and _ifmt == 191:
            _err_class = utils._EngineNoErrorExit()
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _fnlsv_py = utils._EngineIntScalarType.to_py(_fnlsv)
        _nff_py = utils._EngineIntScalarType.to_py(_nff)
        _rnlsv_py = utils._EngineIntScalarType.to_py(_rnlsv)
        _nrf_py = utils._EngineIntScalarType.to_py(_nrf)
        _nvpr_py = utils._EngineIntScalarType.to_py(_nvpr)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['fnlsv', 'nff', 'rnlsv', 'nrf', 'nvpr', 'comm'],
        )(
            _fnlsv_py,
            _nff_py,
            _rnlsv_py,
            _nrf_py,
            _nvpr_py,
            _comm_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def lmm_init_combine(hlmm, xcomm, ycomm):
    u"""
    Linear mixed effects regression, initialization function for
    ``lmm_init_combine`` and ``lmm_fit``.

    ``lmm_init_combine`` combines output from multiple calls to
    ``lmm_init`` or ``mixeff_hier_init``.

    For full information please refer to the NAG Library document for
    g02jg

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02jgf.html

    Parameters
    ----------
    hlmm : Handle
        If the two sets of communication arrays were generated using
        ``lmm_init``, then a G22 handle as generated by one of the calls
        to ``lmm_init``.

        If the two sets of communication arrays were generated using
        ``mixeff_hier_init``, then this argument is not referenced and
        need not be set.

    xcomm : dict, communication object, modified in place
        Communication structure for D_x.

        This argument must have been initialized by a prior call to
        ``lmm_init`` or ``mixeff_hier_init``.

    ycomm : dict, communication object
        Communication structure for D_y.

        This argument must have been initialized by a prior call to
        ``lmm_init`` or ``mixeff_hier_init``.

    Raises
    ------
    NagValueError
        (`errno` 11)
            `hlmm` has not been initialized or is corrupt.

        (`errno` 12)
            `hlmm` is not a G22 handle as generated by ``lmm_init``.

        (`errno` 41)
            On entry, `xcomm`['iopts'] has not been initialized
            correctly.

        (`errno` 42)
            On entry, the information stored in `xcomm`['iopts'] and
            `ycomm`['iopts'] is not compatible.

        (`errno` 71)
            On entry, `ycomm`['iopts'] has not been initialized
            correctly.

    See Also
    --------
    :meth:`naginterfaces.library.examples.correg.lmm_init_combine_ex.main`
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.lmm_init_combine'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02jgft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_HANDLE, # Input, hlmm.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, xrcomm[xlrcomm].
        utils._BYREF_INT, # Input, xlrcomm.
        utils._EngineIntArrayType.c_type, # Input/Output, xicomm[xlicomm].
        utils._BYREF_INT, # Input, xlicomm.
        utils._EngineFloat64ArrayType.c_type, # Input, yrcomm[ylrcomm].
        utils._BYREF_INT, # Input, ylrcomm.
        utils._EngineIntArrayType.c_type, # Input, yicomm[ylicomm].
        utils._BYREF_INT, # Input, ylicomm.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _hlmm = utils._EngineCptrScalarType(
        hlmm, {'fun_name': fun_name, 'entity': 'hlmm'},
    ).to_ctypes()
    utils._check_comm(
        xcomm, {'fun_name': fun_name, 'entity': 'xcomm'},
        req_keys=('opts', 'iopts'),
    )
    _xlrcomm_py = utils._EngineFloat64ArrayType.get_shape(xcomm['opts'], {'fun_name': fun_name, 'entity': 'xcomm[\'opts\']'}, exp_rank=1, validate=False)[0]
    _xlrcomm = utils._EngineIntScalarType(
        _xlrcomm_py, {'fun_name': fun_name, 'entity': '_xlrcomm'},
    ).to_ctypes()
    _xrcomm = xcomm['opts']
    _xlicomm_py = utils._EngineIntArrayType.get_shape(xcomm['iopts'], {'fun_name': fun_name, 'entity': 'xcomm[\'iopts\']'}, exp_rank=1, validate=False)[0]
    _xlicomm = utils._EngineIntScalarType(
        _xlicomm_py, {'fun_name': fun_name, 'entity': '_xlicomm'},
    ).to_ctypes()
    _xicomm = xcomm['iopts']
    utils._check_comm(
        ycomm, {'fun_name': fun_name, 'entity': 'ycomm'},
        req_keys=('opts', 'iopts'),
    )
    _ylrcomm_py = utils._EngineFloat64ArrayType.get_shape(ycomm['opts'], {'fun_name': fun_name, 'entity': 'ycomm[\'opts\']'}, exp_rank=1, validate=False)[0]
    _ylrcomm = utils._EngineIntScalarType(
        _ylrcomm_py, {'fun_name': fun_name, 'entity': '_ylrcomm'},
    ).to_ctypes()
    _yrcomm = ycomm['opts']
    _ylicomm_py = utils._EngineIntArrayType.get_shape(ycomm['iopts'], {'fun_name': fun_name, 'entity': 'ycomm[\'iopts\']'}, exp_rank=1, validate=False)[0]
    _ylicomm = utils._EngineIntScalarType(
        _ylicomm_py, {'fun_name': fun_name, 'entity': '_ylicomm'},
    ).to_ctypes()
    _yicomm = ycomm['iopts']
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _hlmm,
        _xrcomm,
        _xlrcomm,
        _xicomm,
        _xlicomm,
        _yrcomm,
        _ylrcomm,
        _yicomm,
        _ylicomm,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 11 and _ifmt == 11:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'hlmm has not been initialized or is corrupt.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 12 and _ifmt == 12:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'hlmm is not a G22 handle as generated by lmm_init.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 41 and _ifmt == 41:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, xcomm[\'iopts\'] has not been initialized correctly.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 42 and _ifmt == 42:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the information stored in xcomm[\'iopts\'] and ycomm[\'iopts\'] is not compatible.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 52 and _ifmt == 52:
            _err_class = utils._EngineNoErrorExit()
        elif _ierr_py == 81 and _ifmt == 81:
            _ierr_fl = 71
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ycomm[\'iopts\'] has not been initialized correctly.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 91 and _ifmt == 91:
            _ierr_fl = 71
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ycomm[\'iopts\'] has not been initialized correctly.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        return_data = None
    _err_class.manage()
    return return_data

def lmm_fit(hlmm, nvpr, gamma, lb, b, se, czz, cxx, cxz, comm, io_manager=None):
    u"""
    Linear mixed effects regression using either Restricted Maximum
    Likelihood (REML) or Maximum Likelihood (ML).

    ``lmm_fit`` fits a multi-level linear mixed effects regression model
    using restricted maximum likelihood (REML) or maximum likelihood
    (ML).
    Prior to calling ``lmm_fit`` the initialization function
    ``lmm_init`` must be called.

    For full information please refer to the NAG Library document for
    g02jh

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02jhf.html

    Parameters
    ----------
    hlmm : Handle
        A G22 handle to the internal data structure containing a
        description of the required model as returned in hlmm by
        ``lmm_init``.

    nvpr : int
        g, the number of variance components being estimated (excluding
        the overall variance, sigma_R^2).

        This should be set to the value of nvpr returned by
        ``lmm_init``.

    gamma : float, shape (`nvpr`+1), modified in place
        `On entry`: holds the initial values of the variance components,
        gamma_0, with `gamma`[i-1] the initial value for
        sigma_i^2/sigma_R^2, for i = 1,2,...,`nvpr`.

        If `gamma`[0] = -1.0, the remaining elements of `gamma` are
        ignored and the initial values for the variance components are
        estimated from the data using MIVQUE0.

        `On exit`: `gamma`[i-1], for i = 1,2,...,`nvpr`, holds the final
        estimate of sigma_i^2 and `gamma`[`nvpr`] holds the final
        estimate for sigma_R^2.

        Labels for the variance components can be obtained using
        ``blgm.lm_submodel``.

    lb : int
        The dimension of the arrays `b` and `se`.

    b : float, shape (`lb`), modified in place
        `On exit`: the parameter estimates, with the first nrf*rnlsv
        elements of `b` containing the parameter estimates for the
        random effects, nu, and the remaining nff*fnlsv elements
        containing the parameter estimates for the fixed effects, beta.

        Labels for the parameter estimates can be obtained using
        ``blgm.lm_submodel``.

    se : float, shape (`lb`), modified in place
        `On exit`: the standard errors of the parameter estimates given
        in `b`.

    czz : None or float, shape (:, :), modified in place
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `czz` is not None: nrf; otherwise: 0.

        `On exit`, if not None on entry: if rnlsv = 1, `czz` holds the
        lower triangular portion of the matrix
        (1/sigma^2)(Z^TR^^-1Z+G^^-1), where R^ and G^ are the estimates
        of R and G respectively.

        If rnlsv > 1, then `czz` holds this matrix in compressed form,
        with the first nrf columns holding the part of the matrix
        corresponding to the first level of the overall random subject
        variable, the next nrf columns holding the part of the matrix
        corresponding to the second level of the overall random subject
        variable etc.

        If ldczz = 0, `czz` is not referenced.

    cxx : None or float, shape (:, :), modified in place
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `cxx` is not None: nff; otherwise: 0.

        `On exit`, if not None on entry: if fnlsv = 1, `cxx` holds the
        lower triangular portion of the matrix (1/sigma^2)(X^TV^^-1X),
        where V^ is the estimated value of V.

        If fnlsv > 1, then `cxx` holds this matrix in compressed form,
        with the first nff columns holding the part of the matrix
        corresponding to the first level of the overall fixed subject
        variable, the next nff columns holding the part of the matrix
        corresponding to the second level of the overall fixed subject
        variable, etc.

        If ldcxx = 0, `cxx` is not referenced.

    cxz : None or float, shape (:, :), modified in place
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `cxz` is not None: fnlsv*nff;
        otherwise: 0.

        `On exit`, if not None on entry: `cxz` holds the matrix
        (1/sigma^2)(X^TV^^-1Z)G^, where V^ and G^ are the estimates of V
        and G respectively.

        If ldcxz = 0, `cxz` is not referenced.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``lmm_init``.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    effn : int
        Effective number of observations. If there are no weights, or
        all weights are nonzero, `effn` = n.

    rnkx : int
        The rank of the design matrix, X, for the fixed effects.

    ncov : int
        Number of variance components not estimated to be zero. If none
        of the variance components are estimated to be zero, `ncov` =
        `nvpr`.

    lnlike : float
        -2l_R(gamma^) where l_R is the log of the restricted maximum
        likelihood calculated at gamma^, the estimated variance
        components returned in `gamma`.

    Raises
    ------
    NagValueError
        (`errno` 11)
            `hlmm` has not been initialized or is corrupt.

        (`errno` 12)
            `hlmm` is not a G22 handle as generated by ``lmm_init``.

        (`errno` 21)
            On entry, `nvpr` = *<value>*.

            Constraint: `nvpr` >= *<value>*.

        (`errno` 31)
            On entry, `gamma`[*<value>*] = *<value>*.

            Constraint: `gamma`[0] = -1.0 or `gamma`[i-1] >= 0.0.

        (`errno` 81)
            On entry, `lb` = *<value>*.

            Constraint: `lb` >= *<value>*.

        (`errno` 171)
            On entry, the communication arrays, `comm`['iopts'] and
            `comm`['opts'], have not been initialized correctly.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1001)
            Optimal solution found, but requested accuracy not achieved.

        (`errno` 1002)
            Too many major iterations.

        (`errno` 1003)
            Current point cannot be improved upon.

        (`errno` 1004)
            At least one negative estimate for `gamma` was obtained. All
            negative estimates have been set to zero.

    References
    ----------
    Goodnight, J H, 1979, `A tutorial on the SWEEP operator`, The
    American Statistician (33(3)), 149--158

    Harville, D A, 1977, `Maximum likelihood approaches to variance
    component estimation and to related problems`, JASA (72), 320--340

    Rao, C R, 1972, `Estimation of variance and covariance components in
    a linear model`, J. Am. Stat. Assoc. (67), 112--115

    Stroup, W W, 1989, `Predictable functions and prediction space in
    the mixed model procedure`, Applications of Mixed Models in
    Agriculture and Related Disciplines (Southern Cooperative Series
    Bulletin No. 343), 39--48

    Wolfinger, R, Tobias, R and Sall, J, 1994, `Computing Gaussian
    likelihoods and their derivatives for general linear mixed models`,
    SIAM Sci. Statist. Comput. (15), 1294--1310

    See Also
    --------
    :meth:`naginterfaces.library.examples.correg.lmm_init_combine_ex.main`
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.lmm_fit'

    utils.FileObjManager._check_type_compat(
        io_manager, {'fun_name': fun_name, 'entity': 'io_manager'}
    )

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    def cb_print_rech_py(_, en_data, nout, rec, ifail, rec_len):
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data, utils._EngineIntScalarType.to_py(nout.contents), rec[:rec_len], io_manager, excs, ifail.contents,
                fun_name=fun_name,
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02jhft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._PRINT_REC_FUNCTYPE,
        utils._PRINT_RECH_FUNCTYPE,
        utils._BYREF_HANDLE, # Input, hlmm.
        utils._BYREF_INT, # Input, nvpr.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, gamma[nvpr+1].
        utils._BYREF_INT, # Output, effn.
        utils._BYREF_INT, # Output, rnkx.
        utils._BYREF_INT, # Output, ncov.
        utils._BYREF_FLOAT64, # Output, lnlike.
        utils._BYREF_INT, # Input, lb.
        utils._EngineFloat64ArrayType.c_type, # Output, b[lb].
        utils._EngineFloat64ArrayType.c_type, # Output, se[lb].
        utils._EngineFloat64ArrayType.c_type, # Output, czz[ldczz, tdczz].
        utils._BYREF_INT, # Input, ldczz.
        utils._BYREF_INT, # Input, tdczz.
        utils._EngineFloat64ArrayType.c_type, # Output, cxx[ldcxx, tdcxx].
        utils._BYREF_INT, # Input, ldcxx.
        utils._BYREF_INT, # Input, tdcxx.
        utils._EngineFloat64ArrayType.c_type, # Output, cxz[ldcxz, tdcxz].
        utils._BYREF_INT, # Input, ldcxz.
        utils._BYREF_INT, # Input, tdcxz.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, rcomm[lrcomm].
        utils._BYREF_INT, # Input, lrcomm.
        utils._EngineIntArrayType.c_type, # Input/Output, icomm[licomm].
        utils._BYREF_INT, # Input, licomm.
        utils._BYREF_STRING, # Input, routine_name(len=6).
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of routine_name.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((czz, 'czz'), (cxx, 'cxx'), (cxz, 'cxz')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _print_rech = utils._PRINT_RECH_FUNCTYPE(cb_print_rech_py)
    _hlmm = utils._EngineCptrScalarType(
        hlmm, {'fun_name': fun_name, 'entity': 'hlmm'},
    ).to_ctypes()
    _nvpr = utils._EngineIntScalarType(
        nvpr, {'fun_name': fun_name, 'entity': 'nvpr'},
    ).to_ctypes()
    _nvpr_py = utils._EngineIntScalarType.to_py(_nvpr)
    _gamma_cls = utils._EngineFloat64ArrayType(
        gamma, {'fun_name': fun_name, 'entity': 'gamma'},
        exp_shape=(_nvpr_py+1,),
    )
    _gamma = _gamma_cls.to_ctypes()
    _effn = utils._EngineIntScalarType.empty_ctypes()
    _rnkx = utils._EngineIntScalarType.empty_ctypes()
    _ncov = utils._EngineIntScalarType.empty_ctypes()
    _lnlike = utils._EngineFloat64ScalarType.empty_ctypes()
    _lb = utils._EngineIntScalarType(
        lb, {'fun_name': fun_name, 'entity': 'lb'},
    ).to_ctypes()
    utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        req_keys=('opts', 'iopts'),
    )
    _vl_g02jff_fnlsv_py = utils._EngineIntArrayType.get_element(comm['iopts'], {'fun_name': fun_name, 'entity': 'comm[\'iopts\']'}, (23,), validate=False)
    _vl_g02jff_nff_py = utils._EngineIntArrayType.get_element(comm['iopts'], {'fun_name': fun_name, 'entity': 'comm[\'iopts\']'}, (5,), validate=False)
    _vl_g02jff_rnlsv_py = utils._EngineIntArrayType.get_element(comm['iopts'], {'fun_name': fun_name, 'entity': 'comm[\'iopts\']'}, (7,), validate=False)
    _vl_g02jff_nrf_py = utils._EngineIntArrayType.get_element(comm['iopts'], {'fun_name': fun_name, 'entity': 'comm[\'iopts\']'}, (6,), validate=False)
    _b_cls = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_vl_g02jff_fnlsv_py*_vl_g02jff_nff_py+_vl_g02jff_rnlsv_py*_vl_g02jff_nrf_py,),
        check_contents=False,
    )
    _b = _b_cls.to_ctypes(can_reference_contents=False)
    _se_cls = utils._EngineFloat64ArrayType(
        se, {'fun_name': fun_name, 'entity': 'se'},
        exp_shape=(_vl_g02jff_fnlsv_py*_vl_g02jff_nff_py+_vl_g02jff_rnlsv_py*_vl_g02jff_nrf_py,),
        check_contents=False,
    )
    _se = _se_cls.to_ctypes(can_reference_contents=False)
    _ldczz_py = utils._EngineFloat64ArrayType.get_nd(czz, {'fun_name': fun_name, 'entity': 'czz'}, exp_rank=2, sorder=_m_sorder, dim_type='l', allow_none=True)
    _ldczz = utils._EngineIntScalarType(
        _ldczz_py, {'fun_name': fun_name, 'entity': '_ldczz'},
    ).to_ctypes()
    _tdczz_py = utils._EngineFloat64ArrayType.get_nd(czz, {'fun_name': fun_name, 'entity': 'czz'}, exp_rank=2, sorder=_m_sorder, dim_type='s', allow_none=True)
    _tdczz = utils._EngineIntScalarType(
        _tdczz_py, {'fun_name': fun_name, 'entity': '_tdczz'},
    ).to_ctypes()
    if czz is not None:
        _czz_dim1 = _vl_g02jff_nrf_py
    else:
        _czz_dim1 = 0
    exp_shape_czz = (_czz_dim1, None)
    _czz_cls = utils._EngineFloat64ArrayType(
        czz, {'fun_name': fun_name, 'entity': 'czz'},
        exp_shape=exp_shape_czz,
        mutual_sorder=_m_sorder,
        check_contents=False,
        allow_none=True,
    )
    _czz = _czz_cls.to_ctypes(can_reference_contents=False)
    _ldcxx_py = utils._EngineFloat64ArrayType.get_nd(cxx, {'fun_name': fun_name, 'entity': 'cxx'}, exp_rank=2, sorder=_m_sorder, dim_type='l', allow_none=True)
    _ldcxx = utils._EngineIntScalarType(
        _ldcxx_py, {'fun_name': fun_name, 'entity': '_ldcxx'},
    ).to_ctypes()
    _tdcxx_py = utils._EngineFloat64ArrayType.get_nd(cxx, {'fun_name': fun_name, 'entity': 'cxx'}, exp_rank=2, sorder=_m_sorder, dim_type='s', allow_none=True)
    _tdcxx = utils._EngineIntScalarType(
        _tdcxx_py, {'fun_name': fun_name, 'entity': '_tdcxx'},
    ).to_ctypes()
    if cxx is not None:
        _cxx_dim1 = _vl_g02jff_nff_py
    else:
        _cxx_dim1 = 0
    exp_shape_cxx = (_cxx_dim1, None)
    _cxx_cls = utils._EngineFloat64ArrayType(
        cxx, {'fun_name': fun_name, 'entity': 'cxx'},
        exp_shape=exp_shape_cxx,
        mutual_sorder=_m_sorder,
        check_contents=False,
        allow_none=True,
    )
    _cxx = _cxx_cls.to_ctypes(can_reference_contents=False)
    _ldcxz_py = utils._EngineFloat64ArrayType.get_nd(cxz, {'fun_name': fun_name, 'entity': 'cxz'}, exp_rank=2, sorder=_m_sorder, dim_type='l', allow_none=True)
    _ldcxz = utils._EngineIntScalarType(
        _ldcxz_py, {'fun_name': fun_name, 'entity': '_ldcxz'},
    ).to_ctypes()
    _tdcxz_py = utils._EngineFloat64ArrayType.get_nd(cxz, {'fun_name': fun_name, 'entity': 'cxz'}, exp_rank=2, sorder=_m_sorder, dim_type='s', allow_none=True)
    _tdcxz = utils._EngineIntScalarType(
        _tdcxz_py, {'fun_name': fun_name, 'entity': '_tdcxz'},
    ).to_ctypes()
    if cxz is not None:
        _cxz_dim1 = _vl_g02jff_fnlsv_py*_vl_g02jff_nff_py
    else:
        _cxz_dim1 = 0
    exp_shape_cxz = (_cxz_dim1, None)
    _cxz_cls = utils._EngineFloat64ArrayType(
        cxz, {'fun_name': fun_name, 'entity': 'cxz'},
        exp_shape=exp_shape_cxz,
        mutual_sorder=_m_sorder,
        check_contents=False,
        allow_none=True,
    )
    _cxz = _cxz_cls.to_ctypes(can_reference_contents=False)
    _lrcomm_py = utils._EngineFloat64ArrayType.get_shape(comm['opts'], {'fun_name': fun_name, 'entity': 'comm[\'opts\']'}, exp_rank=1, validate=False)[0]
    _lrcomm = utils._EngineIntScalarType(
        _lrcomm_py, {'fun_name': fun_name, 'entity': '_lrcomm'},
    ).to_ctypes()
    _rcomm = comm['opts']
    _licomm_py = utils._EngineIntArrayType.get_shape(comm['iopts'], {'fun_name': fun_name, 'entity': 'comm[\'iopts\']'}, exp_rank=1, validate=False)[0]
    _licomm = utils._EngineIntScalarType(
        _licomm_py, {'fun_name': fun_name, 'entity': '_licomm'},
    ).to_ctypes()
    _icomm = comm['iopts']
    _routine_name = utils._EngineRoutineName(*b'g02jh ')
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        utils._UNUSED_PRINT_REC,
        _print_rech,
        _hlmm,
        _nvpr,
        _gamma,
        _effn,
        _rnkx,
        _ncov,
        _lnlike,
        _lb,
        _b,
        _se,
        _czz,
        _ldczz,
        _tdczz,
        _cxx,
        _ldcxx,
        _tdcxx,
        _cxz,
        _ldcxz,
        _tdcxz,
        _rcomm,
        _lrcomm,
        _icomm,
        _licomm,
        _routine_name,
        _errbuf,
        _ierr,
        _routine_name._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 11 and _ifmt == 11:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'hlmm has not been initialized or is corrupt.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 12 and _ifmt == 12:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'hlmm is not a G22 handle as generated by lmm_init.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 21 and _ifmt == 21:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nvpr = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nvpr >= {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 31 and _ifmt == 31:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, gamma[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: gamma[0] = -1.0 or gamma[i-1] >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 81 and _ifmt == 81:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lb = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: lb >= {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 201 and _ifmt == 201:
            _ierr_fl = 171
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the communication arrays, comm[\'iopts\'] and',
                    'comm[\'opts\'], have not been initialized correctly.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 211 and _ifmt == 211:
            _ierr_fl = 171
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the communication arrays, comm[\'iopts\'] and',
                    'comm[\'opts\'], have not been initialized correctly.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 231 and _ifmt == 231:
            _ierr_fl = 171
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the communication arrays, comm[\'iopts\'] and',
                    'comm[\'opts\'], have not been initialized correctly.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1001 and _ifmt == 1001:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Optimal solution found, but requested accuracy not achieved.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1002 and _ifmt == 1002:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Too many major iterations.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1003 and _ifmt == 1003:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Current point cannot be improved upon.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1004 and _ifmt == 1004:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'At least one negative estimate for gamma was obtained.',
                    'All negative estimates have been set to zero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _gamma_cls.refresh_py(_gamma)
        _b_cls.refresh_py(_b)
        _se_cls.refresh_py(_se)
        if czz is not None:
            _czz_cls.refresh_py(_czz)
        if cxx is not None:
            _cxx_cls.refresh_py(_cxx)
        if cxz is not None:
            _cxz_cls.refresh_py(_cxz)
        _effn_py = utils._EngineIntScalarType.to_py(_effn)
        _rnkx_py = utils._EngineIntScalarType.to_py(_rnkx)
        _ncov_py = utils._EngineIntScalarType.to_py(_ncov)
        _lnlike_py = utils._EngineFloat64ScalarType.to_py(_lnlike)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['effn', 'rnkx', 'ncov', 'lnlike'],
        )(
            _effn_py,
            _rnkx_py,
            _ncov_py,
            _lnlike_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def ridge_opt(n, m, x, isx, tau, y, h, opt, niter, tol, orig, b, vif, res, optloo, perr):
    u"""
    Ridge regression, optimizing a ridge regression parameter.

    ``ridge_opt`` calculates a ridge regression, optimizing the ridge
    parameter according to one of four prediction error criteria.

    For full information please refer to the NAG Library document for
    g02ka

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02kaf.html

    Parameters
    ----------
    n : int
        n, the number of observations.

    m : int
        The number of independent variables available in the data matrix
        X.

    x : float, shape (`n`, `m`)
        The values of independent variables in the data matrix X.

    isx : int, shape (`m`)
        Indicates which m independent variables are included in the
        model.

        `isx`[j-1] = 1

            The jth variable in `x` will be included in the model.

        `isx`[j-1] = 0

            Variable j is excluded.

    tau : float
        Singular values less than `tau` of the SVD of the data matrix X
        will be set equal to zero.

        Suggested value: `tau` = 0.0.

    y : float, shape (`n`)
        The n values of the dependent variable y.

    h : float
        An initial value for the ridge regression parameter h; used as a
        starting point for the optimization.

    opt : int
        The measure of prediction error used to optimize the ridge
        regression parameter h. The value of `opt` must be set equal to
        one of:

        `opt` = 1

            Generalized cross-validation (GCV);

        `opt` = 2

            Unbiased estimate of variance (UEV)

        `opt` = 3

            Future prediction error (FPE)

        `opt` = 4

            Bayesian information criteron (BIC).

    niter : int
        The maximum number of iterations allowed to optimize the ridge
        regression parameter h.

    tol : float
        Iterations of the ridge regression parameter h will halt when
        consecutive values of h lie within `tol`.

    orig : int
        If `orig` = 1, the parameter estimates b are calculated for the
        original data; otherwise `orig` = 2 and the parameter estimates
        b~ are calculated for the standardized data.

    b : float, shape (ip+1), modified in place
        `On exit`: contains the intercept and parameter estimates for
        the fitted ridge regression model in the order indicated by
        `isx`. The first element of `b` contains the estimate for the
        intercept; `b`[j] contains the parameter estimate for the jth
        independent variable in the model, for j = 1,2,...,ip.

    vif : float, shape (ip), modified in place
        `On exit`: the variance inflation factors in the order indicated
        by `isx`. For the jth independent variable in the model,
        `vif`[j-1] is the value of v_j, for j = 1,2,...,ip.

    res : float, shape (`n`), modified in place
        `On exit`: `res`[i-1] is the value of the ith residual for the
        fitted ridge regression model, for i = 1,2,...,`n`.

    optloo : int
        If `optloo` = 2, the leave-one-out cross-validation estimate of
        prediction error is calculated; otherwise no such estimate is
        calculated and `optloo` = 1.

    perr : float, shape (5), modified in place
        `On exit`: the first four elements contain, in this order, the
        measures of prediction error: GCV, UEV, FPE and BIC.

        If `optloo` = 2, `perr`[4] is the LOOCV estimate of prediction
        error; otherwise `perr`[4] is not referenced.

    Returns
    -------
    h : float
        `h` is the optimized value of the ridge regression parameter h.

    niter : int
        The number of iterations used to optimize the ridge regression
        parameter h within `tol`.

    nep : float
        The number of effective parameters, gamma, in the model.

    rss : float
        The sum of squares of residual values.

    df : int
        The degrees of freedom for the residual sum of squares `rss`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` > 1.

        (`errno` 1)
            On entry, `tau` = *<value>*.

            Constraint: `tau` >= 0.0.

        (`errno` 1)
            On entry, `opt` = *<value>*.

            Constraint: `opt` = 1, 2, 3 or 4.

        (`errno` 1)
            On entry, `h` = *<value>*.

            Constraint: `h` > 0.0.

        (`errno` 1)
            On entry, `optloo` = *<value>*.

            Constraint: `optloo` = 1 or 2.

        (`errno` 1)
            On entry, `tol` = *<value>*.

            Constraint: `tol` > 0.0.

        (`errno` 1)
            On entry, `niter` = *<value>*.

            Constraint: `niter` >= 1.

        (`errno` 1)
            On entry, `orig` = *<value>*.

            Constraint: `orig` = 1 or 2.

        (`errno` 2)
            On entry, `m` = *<value>* and `n` = *<value>*.

            Constraint: `m` <= `n`.

        (`errno` 2)
            On entry, ip = *<value>*; `m` = *<value>*.

            Constraint: 1 <= ip <= `m`.

        (`errno` 2)
            On entry, `isx`[*<value>*] = *<value>*.

            Constraint: `isx`[j-1] = 0 or 1.

        (`errno` 2)
            On entry, ip = *<value>*.

            Constraint: sum(`isx`) = ip.

        (`errno` 3)
            SVD failed to converge.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` -1)
            Maximum number of iterations used.

    References
    ----------
    Hastie, T, Tibshirani, R and Friedman, J, 2003, `The Elements of
    Statistical Learning: Data Mining, Inference and Prediction`,
    Springer Series in Statistics

    Moody, J.E., 1992, `The effective number of parameters: An analysis
    of generalisation and regularisation in nonlinear learning systems`,
    In: Neural Information Processing Systems, (eds J E Moody, S J
    Hanson, and R P Lippmann), 4, 847--854, Morgan Kaufmann San Mateo CA

    See Also
    --------
    :meth:`naginterfaces.library.examples.correg.ridge_opt_ex.main`
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.ridge_opt'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02kaft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, sorder.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Input, x[ldx, *].
        utils._BYREF_INT, # Input, ldx.
        utils._EngineIntArrayType.c_type, # Input, isx[m].
        utils._BYREF_INT, # Input, ip.
        utils._BYREF_FLOAT64, # Input, tau.
        utils._EngineFloat64ArrayType.c_type, # Input, y[n].
        utils._BYREF_FLOAT64, # Input/Output, h.
        utils._BYREF_INT, # Input, opt.
        utils._BYREF_INT, # Input/Output, niter.
        utils._BYREF_FLOAT64, # Input, tol.
        utils._BYREF_FLOAT64, # Output, nep.
        utils._BYREF_INT, # Input, orig.
        utils._EngineFloat64ArrayType.c_type, # Output, b[ip+1].
        utils._EngineFloat64ArrayType.c_type, # Output, vif[ip].
        utils._EngineFloat64ArrayType.c_type, # Output, res[n].
        utils._BYREF_FLOAT64, # Output, rss.
        utils._BYREF_INT, # Output, df.
        utils._BYREF_INT, # Input, optloo.
        utils._EngineFloat64ArrayType.c_type, # Output, perr[5].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((x, 'x'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _sorder = utils._EngineIntScalarType.c_type(_nag_enum_consts.nag_col_major)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _ldx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _sorder_py = utils._EngineIntScalarType.to_py(_sorder)
    _m_py = utils._EngineIntScalarType.to_py(_m)
    if _sorder_py == 1:
        _x_dim2 = _m_py
    elif _sorder_py == 2:
        _x_dim2 = _n_py
    else:
        _x_dim2 = 0
    exp_shape_x = (_n_py, _x_dim2)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=exp_shape_x,
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _isx = utils._EngineIntArrayType(
        isx, {'fun_name': fun_name, 'entity': 'isx'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _vl_ip_py = 0
    for _vl_i_py in range(1, _m_py+1):
        if utils._EngineIntArrayType.get_element(isx, {'fun_name': fun_name, 'entity': 'isx'}, (_vl_i_py-1,)) == 1:
            _vl_ip_py = _vl_ip_py + 1
    _ip = utils._EngineIntScalarType(
        _vl_ip_py, {'fun_name': fun_name, 'entity': '_ip'},
    ).to_ctypes()
    _tau = utils._EngineFloat64ScalarType(
        tau, {'fun_name': fun_name, 'entity': 'tau'},
    ).to_ctypes()
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _h = utils._EngineFloat64ScalarType(
        h, {'fun_name': fun_name, 'entity': 'h'},
    ).to_ctypes()
    _opt = utils._EngineIntScalarType(
        opt, {'fun_name': fun_name, 'entity': 'opt'},
    ).to_ctypes()
    _niter = utils._EngineIntScalarType(
        niter, {'fun_name': fun_name, 'entity': 'niter'},
    ).to_ctypes()
    _tol = utils._EngineFloat64ScalarType(
        tol, {'fun_name': fun_name, 'entity': 'tol'},
    ).to_ctypes()
    _nep = utils._EngineFloat64ScalarType.empty_ctypes()
    _orig = utils._EngineIntScalarType(
        orig, {'fun_name': fun_name, 'entity': 'orig'},
    ).to_ctypes()
    _ip_py = utils._EngineIntScalarType.to_py(_ip)
    _b_cls = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_ip_py+1,),
        check_contents=False,
    )
    _b = _b_cls.to_ctypes(can_reference_contents=False)
    _vif_cls = utils._EngineFloat64ArrayType(
        vif, {'fun_name': fun_name, 'entity': 'vif'},
        exp_shape=(_ip_py,),
        check_contents=False,
    )
    _vif = _vif_cls.to_ctypes(can_reference_contents=False)
    _res_cls = utils._EngineFloat64ArrayType(
        res, {'fun_name': fun_name, 'entity': 'res'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _res = _res_cls.to_ctypes(can_reference_contents=False)
    _rss = utils._EngineFloat64ScalarType.empty_ctypes()
    _df = utils._EngineIntScalarType.empty_ctypes()
    _optloo = utils._EngineIntScalarType(
        optloo, {'fun_name': fun_name, 'entity': 'optloo'},
    ).to_ctypes()
    _perr_cls = utils._EngineFloat64ArrayType(
        perr, {'fun_name': fun_name, 'entity': 'perr'},
        exp_shape=(5,),
        check_contents=False,
    )
    _perr = _perr_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _sorder,
        _n,
        _m,
        _x,
        _ldx,
        _isx,
        _ip,
        _tau,
        _y,
        _h,
        _opt,
        _niter,
        _tol,
        _nep,
        _orig,
        _b,
        _vif,
        _res,
        _rss,
        _df,
        _optloo,
        _perr,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == -1 and _ifmt == 113:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Maximum number of iterations used.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n > 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 104:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, tau = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: tau >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 105:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, opt = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: opt = 1, 2, 3 or 4.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 106:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, h = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: h > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 107:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, optloo = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: optloo = 1 or 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 108:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, tol = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: tol > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 109:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, niter = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: niter >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 110:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, orig = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: orig = 1 or 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 101:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: m <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 103:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ip = {:s}; m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: 1 <= ip <= m.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 111:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, isx[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: isx[j-1] = 0 or 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 112:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: sum(isx) = ip.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 114:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'SVD failed to converge.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _h_py = utils._EngineFloat64ScalarType.to_py(_h)
        _niter_py = utils._EngineIntScalarType.to_py(_niter)
        _b_cls.refresh_py(_b)
        _vif_cls.refresh_py(_vif)
        _res_cls.refresh_py(_res)
        _perr_cls.refresh_py(_perr)
        _nep_py = utils._EngineFloat64ScalarType.to_py(_nep)
        _rss_py = utils._EngineFloat64ScalarType.to_py(_rss)
        _df_py = utils._EngineIntScalarType.to_py(_df)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['h', 'niter', 'nep', 'rss', 'df'],
        )(
            _h_py,
            _niter_py,
            _nep_py,
            _rss_py,
            _df_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def ridge(n, m, x, isx, y, lh, h, nep, wantb, b, wantvf, vf, lpec, pec, pe):
    u"""
    Ridge regression using a number of supplied ridge regression
    parameters.

    ``ridge`` calculates a ridge regression, with ridge parameters
    supplied by you.

    For full information please refer to the NAG Library document for
    g02kb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02kbf.html

    Parameters
    ----------
    n : int
        n, the number of observations.

    m : int
        The number of independent variables available in the data matrix
        X.

    x : float, shape (`n`, `m`)
        The values of independent variables in the data matrix X.

    isx : int, shape (`m`)
        Indicates which m independent variables are included in the
        model.

        `isx`[j-1] = 1

            The jth variable in `x` will be included in the model.

        `isx`[j-1] = 0

            Variable j is excluded.

    y : float, shape (`n`)
        The n values of the dependent variable y.

    lh : int
        The number of supplied ridge parameters.

    h : float, shape (`lh`)
        `h`[j-1] is the value of the jth ridge parameter h.

    nep : float, shape (`lh`), modified in place
        `On exit`: `nep`[j-1] is the number of effective parameters,
        gamma, in the jth model, for j = 1,2,...,`lh`.

    wantb : int
        Defines the options for parameter estimates.

        `wantb` = 0

            Parameter estimates are not calculated and `b` is not
            referenced.

        `wantb` = 1

            Parameter estimates b are calculated for the original data.

        `wantb` = 2

            Parameter estimates b~ are calculated for the standardized
            data.

    b : float, shape (:, :), modified in place
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `wantb`!=0: ip+1; otherwise: 1.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `wantb`!=0: `lh`; otherwise: 0.

        `On exit`: if `wantb` != 0, `b` contains the intercept and
        parameter estimates for the fitted ridge regression model in the
        order indicated by `isx`. `b`[0,j-1], for j = 1,2,...,`lh`,
        contains the estimate for the intercept; `b`[i,j-1] contains the
        parameter estimate for the ith independent variable in the model
        fitted with ridge parameter `h`[j-1], for i = 1,2,...,ip.

    wantvf : int
        Defines the options for variance inflation factors.

        `wantvf` = 0

            Variance inflation factors are not calculated and the array
            `vf` is not referenced.

        `wantvf` = 1

            Variance inflation factors are calculated.

    vf : float, shape (:, :), modified in place
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `wantvf`!=0: ip; otherwise: 1.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `wantvf`!=0: `lh`; otherwise: 0.

        `On exit`: if `wantvf` = 1, the variance inflation factors. For
        the ith independent variable in a model fitted with ridge
        parameter `h`[j-1], `vf`[i-1,j-1] is the value of v_i, for i =
        1,2,...,ip.

    lpec : int
        The number of prediction error statistics to return; set `lpec`
        <= 0 for no prediction error estimates.

    pec : None or str, length 1, shape (`lpec`)
        If `lpec` > 0, `pec`[j-1] defines the jth prediction error, for
        j = 1,2,...,`lpec`; otherwise `pec` is not referenced.

        `pec`[j-1] = 'B'

            Bayesian information criterion (BIC).

        `pec`[j-1] = 'F'

            Future prediction error (FPE).

        `pec`[j-1] = 'G'

            Generalized cross-validation (GCV).

        `pec`[j-1] = 'L'

            Leave-one-out cross-validation (LOOCV).

        `pec`[j-1] = 'U'

            Unbiased estimate of variance (UEV).

    pe : None or float, shape (`lpec`, :), modified in place
        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `lpec`>0: `lh`; otherwise: 0.

        `On exit`, if not None on entry: if `lpec` <= 0, `pe` is not
        referenced; otherwise `pe`[i-1,j-1] contains the prediction
        error of criterion `pec`[i-1] for the model fitted with ridge
        parameter `h`[j-1], for j = 1,2,...,`lh`, for i =
        1,2,...,`lpec`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 1)
            On entry, `m` = *<value>* and `n` = *<value>*.

            Constraint: `m` <= `n`.

        (`errno` 1)
            On entry, `h`[j-1] < 0 for at least one j.

            Constraint: `h`[j-1] <= 0.0, for all j.

        (`errno` 1)
            On entry, `lh` = *<value>*.

            Constraint: `lh` > 0.

        (`errno` 1)
            On entry, `wantb` = *<value>*.

            Constraint: `wantb` = 0, 1 or 2.

        (`errno` 1)
            On entry, `wantvf` = *<value>*.

            Constraint: `wantvf` = 0 or 1.

        (`errno` 1)
            On entry, `pec`[j-1] is invalid for at least one j.

            Constraint: if `pec` is not None, `pec`[j-1] = 'B', 'F',
            'G', 'L' or 'U', for all j.

        (`errno` 2)
            On entry, `isx`[j-1] != 0 or 1 for at least one j.

            Constraint: `isx`[j-1] = 0 or 1, for all j.

        (`errno` 2)
            On entry, ip is not equal to the sum of elements in `isx`.

            Constraint: exactly ip elements of `isx` must be equal to 1.

        (`errno` 2)
            On entry, ldb = *<value>* and ip = *<value>*.

            Constraint: if `wantb` != 0, ldb >= ip+1.

        (`errno` 2)
            On entry, ldvf = *<value>* and ip = *<value>*.

            Constraint: if `wantvf` != 0, ldvf >= ip.

        (`errno` 3)
            On entry, `wantb` = 0 and `wantvf` = 0.

            Constraint: `wantb` = 0, `wantvf` = 1.

    References
    ----------
    Hastie, T, Tibshirani, R and Friedman, J, 2003, `The Elements of
    Statistical Learning: Data Mining, Inference and Prediction`,
    Springer Series in Statistics

    Moody, J.E., 1992, `The effective number of parameters: An analysis
    of generalisation and regularisation in nonlinear learning systems`,
    In: Neural Information Processing Systems, (eds J E Moody, S J
    Hanson, and R P Lippmann), 4, 847--854, Morgan Kaufmann San Mateo CA
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.ridge'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02kbft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Input, x[ldx, m].
        utils._BYREF_INT, # Input, ldx.
        utils._EngineIntArrayType.c_type, # Input, isx[m].
        utils._BYREF_INT, # Input, ip.
        utils._EngineFloat64ArrayType.c_type, # Input, y[n].
        utils._BYREF_INT, # Input, lh.
        utils._EngineFloat64ArrayType.c_type, # Input, h[lh].
        utils._EngineFloat64ArrayType.c_type, # Output, nep[lh].
        utils._BYREF_INT, # Input, wantb.
        utils._EngineFloat64ArrayType.c_type, # Output, b[ldb, tdb].
        utils._BYREF_INT, # Input, ldb.
        utils._BYREF_INT, # Input, wantvf.
        utils._EngineFloat64ArrayType.c_type, # Output, vf[ldvf, tdvf].
        utils._BYREF_INT, # Input, ldvf.
        utils._BYREF_INT, # Input, lpec.
        utils._EngineCharArrayType.c_type, # Input, pec(len=1)[lpec].
        utils._EngineFloat64ArrayType.c_type, # Output, pe[ldpe, tdpe].
        utils._BYREF_INT, # Input, ldpe.
        utils._BYREF_INT, # Input, tdb.
        utils._BYREF_INT, # Input, tdvf.
        utils._BYREF_INT, # Input, tdpe.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of pec.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((x, 'x'), (b, 'b'), (vf, 'vf'), (pe, 'pe')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _ldx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py, _m_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _isx = utils._EngineIntArrayType(
        isx, {'fun_name': fun_name, 'entity': 'isx'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _vl_ip_py = 0
    for _vl_i_py in range(1, _m_py+1):
        if utils._EngineIntArrayType.get_element(isx, {'fun_name': fun_name, 'entity': 'isx'}, (_vl_i_py-1,)) == 1:
            _vl_ip_py = _vl_ip_py + 1
    _ip = utils._EngineIntScalarType(
        _vl_ip_py, {'fun_name': fun_name, 'entity': '_ip'},
    ).to_ctypes()
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _lh = utils._EngineIntScalarType(
        lh, {'fun_name': fun_name, 'entity': 'lh'},
    ).to_ctypes()
    _lh_py = utils._EngineIntScalarType.to_py(_lh)
    _h = utils._EngineFloat64ArrayType(
        h, {'fun_name': fun_name, 'entity': 'h'},
        exp_shape=(_lh_py,),
    ).to_ctypes()
    _nep_cls = utils._EngineFloat64ArrayType(
        nep, {'fun_name': fun_name, 'entity': 'nep'},
        exp_shape=(_lh_py,),
        check_contents=False,
    )
    _nep = _nep_cls.to_ctypes(can_reference_contents=False)
    _wantb = utils._EngineIntScalarType(
        wantb, {'fun_name': fun_name, 'entity': 'wantb'},
    ).to_ctypes()
    _ldb_py = utils._EngineFloat64ArrayType.get_nd(b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldb = utils._EngineIntScalarType(
        _ldb_py, {'fun_name': fun_name, 'entity': '_ldb'},
    ).to_ctypes()
    _tdb_py = utils._EngineFloat64ArrayType.get_nd(b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2, sorder=_m_sorder, dim_type='s')
    _tdb = utils._EngineIntScalarType(
        _tdb_py, {'fun_name': fun_name, 'entity': '_tdb'},
    ).to_ctypes()
    _wantb_py = utils._EngineIntScalarType.to_py(_wantb)
    _ip_py = utils._EngineIntScalarType.to_py(_ip)
    if _wantb_py != 0:
        _b_dim1 = _ip_py+1
    else:
        _b_dim1 = 1
    if _wantb_py != 0:
        _b_dim2 = _lh_py
    else:
        _b_dim2 = 0
    exp_shape_b = (_b_dim1, _b_dim2)
    _b_cls = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=exp_shape_b,
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _b = _b_cls.to_ctypes(can_reference_contents=False)
    _wantvf = utils._EngineIntScalarType(
        wantvf, {'fun_name': fun_name, 'entity': 'wantvf'},
    ).to_ctypes()
    _ldvf_py = utils._EngineFloat64ArrayType.get_nd(vf, {'fun_name': fun_name, 'entity': 'vf'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldvf = utils._EngineIntScalarType(
        _ldvf_py, {'fun_name': fun_name, 'entity': '_ldvf'},
    ).to_ctypes()
    _tdvf_py = utils._EngineFloat64ArrayType.get_nd(vf, {'fun_name': fun_name, 'entity': 'vf'}, exp_rank=2, sorder=_m_sorder, dim_type='s')
    _tdvf = utils._EngineIntScalarType(
        _tdvf_py, {'fun_name': fun_name, 'entity': '_tdvf'},
    ).to_ctypes()
    _wantvf_py = utils._EngineIntScalarType.to_py(_wantvf)
    if _wantvf_py != 0:
        _vf_dim1 = _ip_py
    else:
        _vf_dim1 = 1
    if _wantvf_py != 0:
        _vf_dim2 = _lh_py
    else:
        _vf_dim2 = 0
    exp_shape_vf = (_vf_dim1, _vf_dim2)
    _vf_cls = utils._EngineFloat64ArrayType(
        vf, {'fun_name': fun_name, 'entity': 'vf'},
        exp_shape=exp_shape_vf,
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _vf = _vf_cls.to_ctypes(can_reference_contents=False)
    _lpec = utils._EngineIntScalarType(
        lpec, {'fun_name': fun_name, 'entity': 'lpec'},
    ).to_ctypes()
    _lpec_py = utils._EngineIntScalarType.to_py(_lpec)
    if _lpec_py > 0:
        if pec is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'pec'},
                'must not be None when lpec > 0',
            )
    _pec_cls = utils._EngineCharArrayType(
        pec, {'fun_name': fun_name, 'entity': 'pec'}, exp_chrlen=1,
        exp_shape=(_lpec_py,),
        check_contents=_lpec_py > 0,
        allow_none=True,
    )
    _pec = _pec_cls.to_ctypes(can_reference_contents=_lpec_py > 0)
    _pec_chrlen = _pec_cls.chrlen
    _ldpe_py = utils._EngineFloat64ArrayType.get_nd(pe, {'fun_name': fun_name, 'entity': 'pe'}, exp_rank=2, sorder=_m_sorder, dim_type='l', allow_none=True)
    _ldpe = utils._EngineIntScalarType(
        _ldpe_py, {'fun_name': fun_name, 'entity': '_ldpe'},
    ).to_ctypes()
    _tdpe_py = utils._EngineFloat64ArrayType.get_nd(pe, {'fun_name': fun_name, 'entity': 'pe'}, exp_rank=2, sorder=_m_sorder, dim_type='s', allow_none=True)
    _tdpe = utils._EngineIntScalarType(
        _tdpe_py, {'fun_name': fun_name, 'entity': '_tdpe'},
    ).to_ctypes()
    if _lpec_py > 0:
        if pe is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'pe'},
                'must not be None when lpec > 0',
            )
    if _lpec_py > 0:
        _pe_dim2 = _lh_py
    else:
        _pe_dim2 = 0
    exp_shape_pe = (_lpec_py, _pe_dim2)
    _pe_cls = utils._EngineFloat64ArrayType(
        pe, {'fun_name': fun_name, 'entity': 'pe'},
        exp_shape=exp_shape_pe,
        mutual_sorder=_m_sorder,
        check_contents=False,
        allow_none=True,
    )
    _pe = _pe_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _m,
        _x,
        _ldx,
        _isx,
        _ip,
        _y,
        _lh,
        _h,
        _nep,
        _wantb,
        _b,
        _ldb,
        _wantvf,
        _vf,
        _ldvf,
        _lpec,
        _pec,
        _pe,
        _ldpe,
        _tdb,
        _tdvf,
        _tdpe,
        _errbuf,
        _ierr,
        _pec_chrlen,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: m <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 7:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, h[j-1] < 0 for at least one j.',
                    'Constraint: h[j-1] <= 0.0, for all j.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 8:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lh = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: lh > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 11:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, wantb = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: wantb = 0, 1 or 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 14:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, wantvf = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: wantvf = 0 or 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 18:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, pec[j-1] is invalid for at least one j.',
                    'Constraint: if pec is not None, pec[j-1] = \'B\', \'F\', \'G\', \'L\' or \'U\', for all j.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, isx[j-1] != 0 or 1 for at least one j.',
                    'Constraint: isx[j-1] = 0 or 1, for all j.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 6:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ip is not equal to the sum of elements in isx.',
                    'Constraint: exactly ip elements of isx must be equal to 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 13:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ldb = {:s} and ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if wantb != 0, ldb >= ip+1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 16:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ldvf = {:s} and ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if wantvf != 0, ldvf >= ip.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 110:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, wantb = 0 and wantvf = 0.',
                    'Constraint: wantb = 0, wantvf = 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _nep_cls.refresh_py(_nep)
        _b_cls.refresh_py(_b)
        _vf_cls.refresh_py(_vf)
        if pe is not None:
            _pe_cls.refresh_py(_pe)
        return_data = None
    _err_class.manage()
    return return_data

def pls_svd(n, mx, x, isx, my, y, xbar, ybar, iscale, xstd, ystd, maxfac, xres, yres, w, p, t, c, u, xcv, ycv, io_manager=None):
    u"""
    Partial least squares (PLS) regression using singular value
    decomposition.

    ``pls_svd`` fits an orthogonal scores partial least squares (PLS)
    regression by using singular value decomposition.

    For full information please refer to the NAG Library document for
    g02la

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02laf.html

    Parameters
    ----------
    n : int
        n, the number of observations.

    mx : int
        The number of predictor variables.

    x : float, shape (`n`, `mx`)
        `x`[i-1,j-1] must contain the ith observation on the jth
        predictor variable, for j = 1,2,...,`mx`, for i = 1,2,...,`n`.

    isx : int, shape (`mx`)
        Indicates which predictor variables are to be included in the
        model.

        `isx`[j-1] = 1

            The jth predictor variable (with variates in the jth column
            of X) is included in the model.

        `isx`[j-1] = 0

            Otherwise.

    my : int
        r, the number of response variables.

    y : float, shape (`n`, `my`)
        `y`[i-1,j-1] must contain the ith observation for the jth
        response variable, for j = 1,2,...,`my`, for i = 1,2,...,`n`.

    xbar : float, shape (ip), modified in place
        `On exit`: mean values of predictor variables in the model.

    ybar : float, shape (`my`), modified in place
        `On exit`: the mean value of each response variable.

    iscale : int
        Indicates how predictor variables are scaled.

        `iscale` = 1

            Data are scaled by the standard deviation of variables.

        `iscale` = 2

            Data are scaled by user-supplied scalings.

        `iscale` = -1

            No scaling.

    xstd : float, shape (ip), modified in place
        `On entry`: if `iscale` = 2, `xstd`[j-1] must contain the
        user-supplied scaling for the jth predictor variable in the
        model, for j = 1,2,...,ip. Otherwise `xstd` need not be set.

        `On exit`: if `iscale` = 1, standard deviations of predictor
        variables in the model. Otherwise `xstd` is not changed.

    ystd : float, shape (`my`), modified in place
        `On entry`: if `iscale` = 2, `ystd`[j-1] must contain the
        user-supplied scaling for the jth response variable in the
        model, for j = 1,2,...,`my`. Otherwise `ystd` need not be set.

        `On exit`: if `iscale` = 1, the standard deviation of each
        response variable. Otherwise `ystd` is not changed.

    maxfac : int
        k, the number of latent variables to calculate.

    xres : float, shape (`n`, ip), modified in place
        `On exit`: the predictor variables' residual matrix X_k.

    yres : float, shape (`n`, `my`), modified in place
        `On exit`: the residuals for each response variable, Y_k.

    w : float, shape (ip, `maxfac`), modified in place
        `On exit`: the jth column of W contains the x-weights w_j, for j
        = 1,2,...,`maxfac`.

    p : float, shape (ip, `maxfac`), modified in place
        `On exit`: the jth column of P contains the x-loadings p_j, for
        j = 1,2,...,`maxfac`.

    t : float, shape (`n`, `maxfac`), modified in place
        `On exit`: the jth column of T contains the x-scores t_j, for j
        = 1,2,...,`maxfac`.

    c : float, shape (`my`, `maxfac`), modified in place
        `On exit`: the jth column of C contains the y-loadings c_j, for
        j = 1,2,...,`maxfac`.

    u : float, shape (`n`, `maxfac`), modified in place
        `On exit`: the jth column of U contains the y-scores u_j, for j
        = 1,2,...,`maxfac`.

    xcv : float, shape (`maxfac`), modified in place
        `On exit`: `xcv`[j-1] contains the cumulative percentage of
        variance in the predictor variables explained by the first j
        factors, for j = 1,2,...,`maxfac`.

    ycv : float, shape (`maxfac`, `my`), modified in place
        `On exit`: `ycv`[i-1,j-1] is the cumulative percentage of
        variance of the jth response variable explained by the first i
        factors, for j = 1,2,...,`my`, for i = 1,2,...,`maxfac`.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` > 1.

        (`errno` 1)
            On entry, `mx` = *<value>*.

            Constraint: `mx` > 1.

        (`errno` 1)
            On entry, `isx`[*<value>*] is invalid.

            Constraint: `isx`[j-1] = 0 or 1, for all j.

        (`errno` 1)
            On entry, `my` = *<value>*.

            Constraint: `my` >= 1.

        (`errno` 1)
            On entry, `iscale` = *<value>*.

            Constraint: `iscale` = -1 or 1.

        (`errno` 2)
            On entry, ip = *<value>* and `mx` = *<value>*.

            Constraint: 1 < ip <= `mx`.

        (`errno` 2)
            On entry, `maxfac` = *<value>* and ip = *<value>*.

            Constraint: 1 <= `maxfac` <= ip.

        (`errno` 3)
            On entry, ip = *<value>* and sum(`isx`) = *<value>*.

            Constraint: the sum of elements in `isx` must equal ip.

    See Also
    --------
    :meth:`naginterfaces.library.examples.correg.pls_ex.main`
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.pls_svd'

    utils.FileObjManager._check_type_compat(
        io_manager, {'fun_name': fun_name, 'entity': 'io_manager'}
    )

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    def cb_print_rech_py(_, en_data, nout, rec, ifail, rec_len):
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data, utils._EngineIntScalarType.to_py(nout.contents), rec[:rec_len], io_manager, excs, ifail.contents,
                fun_name=fun_name,
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02laft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._PRINT_REC_FUNCTYPE,
        utils._PRINT_RECH_FUNCTYPE,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, mx.
        utils._EngineFloat64ArrayType.c_type, # Input, x[ldx, mx].
        utils._BYREF_INT, # Input, ldx.
        utils._EngineIntArrayType.c_type, # Input, isx[mx].
        utils._BYREF_INT, # Input, ip.
        utils._BYREF_INT, # Input, my.
        utils._EngineFloat64ArrayType.c_type, # Input, y[ldy, my].
        utils._BYREF_INT, # Input, ldy.
        utils._EngineFloat64ArrayType.c_type, # Output, xbar[ip].
        utils._EngineFloat64ArrayType.c_type, # Output, ybar[my].
        utils._BYREF_INT, # Input, iscale.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, xstd[ip].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ystd[my].
        utils._BYREF_INT, # Input, maxfac.
        utils._EngineFloat64ArrayType.c_type, # Output, xres[ldxres, ip].
        utils._BYREF_INT, # Input, ldxres.
        utils._EngineFloat64ArrayType.c_type, # Output, yres[ldyres, my].
        utils._BYREF_INT, # Input, ldyres.
        utils._EngineFloat64ArrayType.c_type, # Output, w[ldw, maxfac].
        utils._BYREF_INT, # Input, ldw.
        utils._EngineFloat64ArrayType.c_type, # Output, p[ldp, maxfac].
        utils._BYREF_INT, # Input, ldp.
        utils._EngineFloat64ArrayType.c_type, # Output, t[ldt, maxfac].
        utils._BYREF_INT, # Input, ldt.
        utils._EngineFloat64ArrayType.c_type, # Output, c[ldc, maxfac].
        utils._BYREF_INT, # Input, ldc.
        utils._EngineFloat64ArrayType.c_type, # Output, u[ldu, maxfac].
        utils._BYREF_INT, # Input, ldu.
        utils._EngineFloat64ArrayType.c_type, # Output, xcv[maxfac].
        utils._EngineFloat64ArrayType.c_type, # Output, ycv[ldycv, my].
        utils._BYREF_INT, # Input, ldycv.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((x, 'x'), (y, 'y'), (xres, 'xres'), (yres, 'yres'), (w, 'w'), (p, 'p'), (t, 't'), (c, 'c'), (u, 'u'), (ycv, 'ycv')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _print_rech = utils._PRINT_RECH_FUNCTYPE(cb_print_rech_py)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _mx = utils._EngineIntScalarType(
        mx, {'fun_name': fun_name, 'entity': 'mx'},
    ).to_ctypes()
    _ldx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _mx_py = utils._EngineIntScalarType.to_py(_mx)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py, _mx_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _isx = utils._EngineIntArrayType(
        isx, {'fun_name': fun_name, 'entity': 'isx'},
        exp_shape=(_mx_py,),
    ).to_ctypes()
    _vl_ip_py = 0
    for _vl_i_py in range(1, _mx_py+1):
        if utils._EngineIntArrayType.get_element(isx, {'fun_name': fun_name, 'entity': 'isx'}, (_vl_i_py-1,)) == 1:
            _vl_ip_py = _vl_ip_py + 1
    _ip = utils._EngineIntScalarType(
        _vl_ip_py, {'fun_name': fun_name, 'entity': '_ip'},
    ).to_ctypes()
    _my = utils._EngineIntScalarType(
        my, {'fun_name': fun_name, 'entity': 'my'},
    ).to_ctypes()
    _ldy_py = utils._EngineFloat64ArrayType.get_nd(y, {'fun_name': fun_name, 'entity': 'y'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldy = utils._EngineIntScalarType(
        _ldy_py, {'fun_name': fun_name, 'entity': '_ldy'},
    ).to_ctypes()
    _my_py = utils._EngineIntScalarType.to_py(_my)
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py, _my_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _ip_py = utils._EngineIntScalarType.to_py(_ip)
    _xbar_cls = utils._EngineFloat64ArrayType(
        xbar, {'fun_name': fun_name, 'entity': 'xbar'},
        exp_shape=(_ip_py,),
        check_contents=False,
    )
    _xbar = _xbar_cls.to_ctypes(can_reference_contents=False)
    _ybar_cls = utils._EngineFloat64ArrayType(
        ybar, {'fun_name': fun_name, 'entity': 'ybar'},
        exp_shape=(_my_py,),
        check_contents=False,
    )
    _ybar = _ybar_cls.to_ctypes(can_reference_contents=False)
    _iscale = utils._EngineIntScalarType(
        iscale, {'fun_name': fun_name, 'entity': 'iscale'},
    ).to_ctypes()
    _xstd_cls = utils._EngineFloat64ArrayType(
        xstd, {'fun_name': fun_name, 'entity': 'xstd'},
        exp_shape=(_ip_py,),
    )
    _xstd = _xstd_cls.to_ctypes()
    _ystd_cls = utils._EngineFloat64ArrayType(
        ystd, {'fun_name': fun_name, 'entity': 'ystd'},
        exp_shape=(_my_py,),
    )
    _ystd = _ystd_cls.to_ctypes()
    _maxfac = utils._EngineIntScalarType(
        maxfac, {'fun_name': fun_name, 'entity': 'maxfac'},
    ).to_ctypes()
    _ldxres_py = utils._EngineFloat64ArrayType.get_nd(xres, {'fun_name': fun_name, 'entity': 'xres'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldxres = utils._EngineIntScalarType(
        _ldxres_py, {'fun_name': fun_name, 'entity': '_ldxres'},
    ).to_ctypes()
    _xres_cls = utils._EngineFloat64ArrayType(
        xres, {'fun_name': fun_name, 'entity': 'xres'},
        exp_shape=(_n_py, _ip_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _xres = _xres_cls.to_ctypes(can_reference_contents=False)
    _ldyres_py = utils._EngineFloat64ArrayType.get_nd(yres, {'fun_name': fun_name, 'entity': 'yres'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldyres = utils._EngineIntScalarType(
        _ldyres_py, {'fun_name': fun_name, 'entity': '_ldyres'},
    ).to_ctypes()
    _yres_cls = utils._EngineFloat64ArrayType(
        yres, {'fun_name': fun_name, 'entity': 'yres'},
        exp_shape=(_n_py, _my_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _yres = _yres_cls.to_ctypes(can_reference_contents=False)
    _ldw_py = utils._EngineFloat64ArrayType.get_nd(w, {'fun_name': fun_name, 'entity': 'w'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldw = utils._EngineIntScalarType(
        _ldw_py, {'fun_name': fun_name, 'entity': '_ldw'},
    ).to_ctypes()
    _maxfac_py = utils._EngineIntScalarType.to_py(_maxfac)
    _w_cls = utils._EngineFloat64ArrayType(
        w, {'fun_name': fun_name, 'entity': 'w'},
        exp_shape=(_ip_py, _maxfac_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _w = _w_cls.to_ctypes(can_reference_contents=False)
    _ldp_py = utils._EngineFloat64ArrayType.get_nd(p, {'fun_name': fun_name, 'entity': 'p'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldp = utils._EngineIntScalarType(
        _ldp_py, {'fun_name': fun_name, 'entity': '_ldp'},
    ).to_ctypes()
    _p_cls = utils._EngineFloat64ArrayType(
        p, {'fun_name': fun_name, 'entity': 'p'},
        exp_shape=(_ip_py, _maxfac_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _p = _p_cls.to_ctypes(can_reference_contents=False)
    _ldt_py = utils._EngineFloat64ArrayType.get_nd(t, {'fun_name': fun_name, 'entity': 't'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldt = utils._EngineIntScalarType(
        _ldt_py, {'fun_name': fun_name, 'entity': '_ldt'},
    ).to_ctypes()
    _t_cls = utils._EngineFloat64ArrayType(
        t, {'fun_name': fun_name, 'entity': 't'},
        exp_shape=(_n_py, _maxfac_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _t = _t_cls.to_ctypes(can_reference_contents=False)
    _ldc_py = utils._EngineFloat64ArrayType.get_nd(c, {'fun_name': fun_name, 'entity': 'c'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldc = utils._EngineIntScalarType(
        _ldc_py, {'fun_name': fun_name, 'entity': '_ldc'},
    ).to_ctypes()
    _c_cls = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_my_py, _maxfac_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _c = _c_cls.to_ctypes(can_reference_contents=False)
    _ldu_py = utils._EngineFloat64ArrayType.get_nd(u, {'fun_name': fun_name, 'entity': 'u'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldu = utils._EngineIntScalarType(
        _ldu_py, {'fun_name': fun_name, 'entity': '_ldu'},
    ).to_ctypes()
    _u_cls = utils._EngineFloat64ArrayType(
        u, {'fun_name': fun_name, 'entity': 'u'},
        exp_shape=(_n_py, _maxfac_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _u = _u_cls.to_ctypes(can_reference_contents=False)
    _xcv_cls = utils._EngineFloat64ArrayType(
        xcv, {'fun_name': fun_name, 'entity': 'xcv'},
        exp_shape=(_maxfac_py,),
        check_contents=False,
    )
    _xcv = _xcv_cls.to_ctypes(can_reference_contents=False)
    _ldycv_py = utils._EngineFloat64ArrayType.get_nd(ycv, {'fun_name': fun_name, 'entity': 'ycv'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldycv = utils._EngineIntScalarType(
        _ldycv_py, {'fun_name': fun_name, 'entity': '_ldycv'},
    ).to_ctypes()
    _ycv_cls = utils._EngineFloat64ArrayType(
        ycv, {'fun_name': fun_name, 'entity': 'ycv'},
        exp_shape=(_maxfac_py, _my_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _ycv = _ycv_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        utils._UNUSED_PRINT_REC,
        _print_rech,
        _n,
        _mx,
        _x,
        _ldx,
        _isx,
        _ip,
        _my,
        _y,
        _ldy,
        _xbar,
        _ybar,
        _iscale,
        _xstd,
        _ystd,
        _maxfac,
        _xres,
        _ldxres,
        _yres,
        _ldyres,
        _w,
        _ldw,
        _p,
        _ldp,
        _t,
        _ldt,
        _c,
        _ldc,
        _u,
        _ldu,
        _xcv,
        _ycv,
        _ldycv,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n > 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mx = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mx > 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, isx[{:s}] is invalid.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: isx[j-1] = 0 or 1, for all j.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 7:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, my = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: my >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 12:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iscale = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: iscale = -1 or 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 6:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ip = {:s} and mx = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: 1 < ip <= mx.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 15:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, maxfac = {:s} and ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: 1 <= maxfac <= ip.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 66:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ip = {:s} and sum(isx) = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: the sum of elements in isx must equal ip.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _xbar_cls.refresh_py(_xbar)
        _ybar_cls.refresh_py(_ybar)
        _xstd_cls.refresh_py(_xstd)
        _ystd_cls.refresh_py(_ystd)
        _xres_cls.refresh_py(_xres)
        _yres_cls.refresh_py(_yres)
        _w_cls.refresh_py(_w)
        _p_cls.refresh_py(_p)
        _t_cls.refresh_py(_t)
        _c_cls.refresh_py(_c)
        _u_cls.refresh_py(_u)
        _xcv_cls.refresh_py(_xcv)
        _ycv_cls.refresh_py(_ycv)
        return_data = None
    _err_class.manage()
    return return_data

def pls_wold(n, mx, x, isx, my, y, xbar, ybar, iscale, xstd, ystd, maxfac, maxit, tau, xres, yres, w, p, t, c, u, xcv, ycv, io_manager=None):
    u"""
    Partial least squares (PLS) regression using Wold's iterative
    method.

    ``pls_wold`` fits an orthogonal scores partial least squares (PLS)
    regression by using Wold's iterative method.

    For full information please refer to the NAG Library document for
    g02lb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02lbf.html

    Parameters
    ----------
    n : int
        n, the number of observations.

    mx : int
        The number of predictor variables.

    x : float, shape (`n`, `mx`)
        `x`[i-1,j-1] must contain the ith observation on the jth
        predictor variable, for j = 1,2,...,`mx`, for i = 1,2,...,`n`.

    isx : int, shape (`mx`)
        Indicates which predictor variables are to be included in the
        model.

        `isx`[j-1] = 1

            The jth predictor variable (with variates in the jth column
            of X) is included in the model.

        `isx`[j-1] = 0

            Otherwise.

    my : int
        r, the number of response variables.

    y : float, shape (`n`, `my`)
        `y`[i-1,j-1] must contain the ith observation for the jth
        response variable, for j = 1,2,...,`my`, for i = 1,2,...,`n`.

    xbar : float, shape (ip), modified in place
        `On exit`: mean values of predictor variables in the model.

    ybar : float, shape (`my`), modified in place
        `On exit`: the mean value of each response variable.

    iscale : int
        Indicates how predictor variables are scaled.

        `iscale` = 1

            Data are scaled by the standard deviation of variables.

        `iscale` = 2

            Data are scaled by user-supplied scalings.

        `iscale` = -1

            No scaling.

    xstd : float, shape (ip), modified in place
        `On entry`: if `iscale` = 2, `xstd`[j-1] must contain the
        user-supplied scaling for the jth predictor variable in the
        model, for j = 1,2,...,ip. Otherwise `xstd` need not be set.

        `On exit`: if `iscale` = 1, standard deviations of predictor
        variables in the model. Otherwise `xstd` is not changed.

    ystd : float, shape (`my`), modified in place
        `On entry`: if `iscale` = 2, `ystd`[j-1] must contain the
        user-supplied scaling for the jth response variable in the
        model, for j = 1,2,...,`my`. Otherwise `ystd` need not be set.

        `On exit`: if `iscale` = 1, the standard deviation of each
        response variable. Otherwise `ystd` is not changed.

    maxfac : int
        k, the number of latent variables to calculate.

    maxit : int
        If `my` = 1, `maxit` is not referenced; otherwise the maximum
        number of iterations used to calculate the x-weights.

        Suggested value: `maxit` = 200.

    tau : float
        If `my` = 1, `tau` is not referenced; otherwise the iterative
        procedure used to calculate the x-weights will halt if the
        Euclidean distance between two subsequent estimates is less than
        or equal to `tau`.

        Suggested value: `tau` = 1.0e-4.

    xres : float, shape (`n`, ip), modified in place
        `On exit`: the predictor variables' residual matrix X_k.

    yres : float, shape (`n`, `my`), modified in place
        `On exit`: the residuals for each response variable, Y_k.

    w : float, shape (ip, `maxfac`), modified in place
        `On exit`: the jth column of W contains the x-weights w_j, for j
        = 1,2,...,`maxfac`.

    p : float, shape (ip, `maxfac`), modified in place
        `On exit`: the jth column of P contains the x-loadings p_j, for
        j = 1,2,...,`maxfac`.

    t : float, shape (`n`, `maxfac`), modified in place
        `On exit`: the jth column of T contains the x-scores t_j, for j
        = 1,2,...,`maxfac`.

    c : float, shape (`my`, `maxfac`), modified in place
        `On exit`: the jth column of C contains the y-loadings c_j, for
        j = 1,2,...,`maxfac`.

    u : float, shape (`n`, `maxfac`), modified in place
        `On exit`: the jth column of U contains the y-scores u_j, for j
        = 1,2,...,`maxfac`.

    xcv : float, shape (`maxfac`), modified in place
        `On exit`: `xcv`[j-1] contains the cumulative percentage of
        variance in the predictor variables explained by the first j
        factors, for j = 1,2,...,`maxfac`.

    ycv : float, shape (`maxfac`, `my`), modified in place
        `On exit`: `ycv`[i-1,j-1] is the cumulative percentage of
        variance of the jth response variable explained by the first i
        factors, for j = 1,2,...,`my`, for i = 1,2,...,`maxfac`.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` > 1.

        (`errno` 1)
            On entry, `mx` = *<value>*.

            Constraint: `mx` > 1.

        (`errno` 1)
            On entry, `isx`[*<value>*] is invalid.

            Constraint: `isx`[j-1] = 0 or 1, for all j.

        (`errno` 1)
            On entry, `my` = *<value>*.

            Constraint: `my` >= 1.

        (`errno` 1)
            On entry, `iscale` = *<value>*.

            Constraint: `iscale` = -1 or 1.

        (`errno` 2)
            On entry, ip = *<value>* and `mx` = *<value>*.

            Constraint: 1 < ip <= `mx`.

        (`errno` 2)
            On entry, `maxfac` = *<value>* and ip = *<value>*.

            Constraint: 1 <= `maxfac` <= ip.

        (`errno` 2)
            On entry, `my` = *<value>* and `maxit` = *<value>*.

            Constraint: if `my` > 1, `maxit` > 1.

        (`errno` 2)
            On entry, `tau` = *<value>*.

            Constraint: if `my` > 1, `tau` > 0.0.

        (`errno` 3)
            On entry, ip = *<value>* and sum(`isx`) = *<value>*.

            Constraint: the sum of elements in `isx` must equal ip.

    References
    ----------
    Wold, H, 1966, `Estimation of principal components and related
    models by iterative least squares`, In: Multivariate Analysis, (ed P
    R Krishnaiah), 391--420, Academic Press NY
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.pls_wold'

    utils.FileObjManager._check_type_compat(
        io_manager, {'fun_name': fun_name, 'entity': 'io_manager'}
    )

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    def cb_print_rech_py(_, en_data, nout, rec, ifail, rec_len):
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data, utils._EngineIntScalarType.to_py(nout.contents), rec[:rec_len], io_manager, excs, ifail.contents,
                fun_name=fun_name,
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02lbft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._PRINT_REC_FUNCTYPE,
        utils._PRINT_RECH_FUNCTYPE,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, mx.
        utils._EngineFloat64ArrayType.c_type, # Input, x[ldx, mx].
        utils._BYREF_INT, # Input, ldx.
        utils._EngineIntArrayType.c_type, # Input, isx[mx].
        utils._BYREF_INT, # Input, ip.
        utils._BYREF_INT, # Input, my.
        utils._EngineFloat64ArrayType.c_type, # Input, y[ldy, my].
        utils._BYREF_INT, # Input, ldy.
        utils._EngineFloat64ArrayType.c_type, # Output, xbar[ip].
        utils._EngineFloat64ArrayType.c_type, # Output, ybar[my].
        utils._BYREF_INT, # Input, iscale.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, xstd[ip].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ystd[my].
        utils._BYREF_INT, # Input, maxfac.
        utils._BYREF_INT, # Input, maxit.
        utils._BYREF_FLOAT64, # Input, tau.
        utils._EngineFloat64ArrayType.c_type, # Output, xres[ldxres, ip].
        utils._BYREF_INT, # Input, ldxres.
        utils._EngineFloat64ArrayType.c_type, # Output, yres[ldyres, my].
        utils._BYREF_INT, # Input, ldyres.
        utils._EngineFloat64ArrayType.c_type, # Output, w[ldw, maxfac].
        utils._BYREF_INT, # Input, ldw.
        utils._EngineFloat64ArrayType.c_type, # Output, p[ldp, maxfac].
        utils._BYREF_INT, # Input, ldp.
        utils._EngineFloat64ArrayType.c_type, # Output, t[ldt, maxfac].
        utils._BYREF_INT, # Input, ldt.
        utils._EngineFloat64ArrayType.c_type, # Output, c[ldc, maxfac].
        utils._BYREF_INT, # Input, ldc.
        utils._EngineFloat64ArrayType.c_type, # Output, u[ldu, maxfac].
        utils._BYREF_INT, # Input, ldu.
        utils._EngineFloat64ArrayType.c_type, # Output, xcv[maxfac].
        utils._EngineFloat64ArrayType.c_type, # Output, ycv[ldycv, my].
        utils._BYREF_INT, # Input, ldycv.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((x, 'x'), (y, 'y'), (xres, 'xres'), (yres, 'yres'), (w, 'w'), (p, 'p'), (t, 't'), (c, 'c'), (u, 'u'), (ycv, 'ycv')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _print_rech = utils._PRINT_RECH_FUNCTYPE(cb_print_rech_py)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _mx = utils._EngineIntScalarType(
        mx, {'fun_name': fun_name, 'entity': 'mx'},
    ).to_ctypes()
    _ldx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _mx_py = utils._EngineIntScalarType.to_py(_mx)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py, _mx_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _isx = utils._EngineIntArrayType(
        isx, {'fun_name': fun_name, 'entity': 'isx'},
        exp_shape=(_mx_py,),
    ).to_ctypes()
    _vl_ip_py = 0
    for _vl_i_py in range(1, _mx_py+1):
        if utils._EngineIntArrayType.get_element(isx, {'fun_name': fun_name, 'entity': 'isx'}, (_vl_i_py-1,)) == 1:
            _vl_ip_py = _vl_ip_py + 1
    _ip = utils._EngineIntScalarType(
        _vl_ip_py, {'fun_name': fun_name, 'entity': '_ip'},
    ).to_ctypes()
    _my = utils._EngineIntScalarType(
        my, {'fun_name': fun_name, 'entity': 'my'},
    ).to_ctypes()
    _ldy_py = utils._EngineFloat64ArrayType.get_nd(y, {'fun_name': fun_name, 'entity': 'y'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldy = utils._EngineIntScalarType(
        _ldy_py, {'fun_name': fun_name, 'entity': '_ldy'},
    ).to_ctypes()
    _my_py = utils._EngineIntScalarType.to_py(_my)
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py, _my_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _ip_py = utils._EngineIntScalarType.to_py(_ip)
    _xbar_cls = utils._EngineFloat64ArrayType(
        xbar, {'fun_name': fun_name, 'entity': 'xbar'},
        exp_shape=(_ip_py,),
        check_contents=False,
    )
    _xbar = _xbar_cls.to_ctypes(can_reference_contents=False)
    _ybar_cls = utils._EngineFloat64ArrayType(
        ybar, {'fun_name': fun_name, 'entity': 'ybar'},
        exp_shape=(_my_py,),
        check_contents=False,
    )
    _ybar = _ybar_cls.to_ctypes(can_reference_contents=False)
    _iscale = utils._EngineIntScalarType(
        iscale, {'fun_name': fun_name, 'entity': 'iscale'},
    ).to_ctypes()
    _xstd_cls = utils._EngineFloat64ArrayType(
        xstd, {'fun_name': fun_name, 'entity': 'xstd'},
        exp_shape=(_ip_py,),
    )
    _xstd = _xstd_cls.to_ctypes()
    _ystd_cls = utils._EngineFloat64ArrayType(
        ystd, {'fun_name': fun_name, 'entity': 'ystd'},
        exp_shape=(_my_py,),
    )
    _ystd = _ystd_cls.to_ctypes()
    _maxfac = utils._EngineIntScalarType(
        maxfac, {'fun_name': fun_name, 'entity': 'maxfac'},
    ).to_ctypes()
    _maxit = utils._EngineIntScalarType(
        maxit, {'fun_name': fun_name, 'entity': 'maxit'},
    ).to_ctypes()
    _tau = utils._EngineFloat64ScalarType(
        tau, {'fun_name': fun_name, 'entity': 'tau'},
    ).to_ctypes()
    _ldxres_py = utils._EngineFloat64ArrayType.get_nd(xres, {'fun_name': fun_name, 'entity': 'xres'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldxres = utils._EngineIntScalarType(
        _ldxres_py, {'fun_name': fun_name, 'entity': '_ldxres'},
    ).to_ctypes()
    _xres_cls = utils._EngineFloat64ArrayType(
        xres, {'fun_name': fun_name, 'entity': 'xres'},
        exp_shape=(_n_py, _ip_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _xres = _xres_cls.to_ctypes(can_reference_contents=False)
    _ldyres_py = utils._EngineFloat64ArrayType.get_nd(yres, {'fun_name': fun_name, 'entity': 'yres'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldyres = utils._EngineIntScalarType(
        _ldyres_py, {'fun_name': fun_name, 'entity': '_ldyres'},
    ).to_ctypes()
    _yres_cls = utils._EngineFloat64ArrayType(
        yres, {'fun_name': fun_name, 'entity': 'yres'},
        exp_shape=(_n_py, _my_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _yres = _yres_cls.to_ctypes(can_reference_contents=False)
    _ldw_py = utils._EngineFloat64ArrayType.get_nd(w, {'fun_name': fun_name, 'entity': 'w'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldw = utils._EngineIntScalarType(
        _ldw_py, {'fun_name': fun_name, 'entity': '_ldw'},
    ).to_ctypes()
    _maxfac_py = utils._EngineIntScalarType.to_py(_maxfac)
    _w_cls = utils._EngineFloat64ArrayType(
        w, {'fun_name': fun_name, 'entity': 'w'},
        exp_shape=(_ip_py, _maxfac_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _w = _w_cls.to_ctypes(can_reference_contents=False)
    _ldp_py = utils._EngineFloat64ArrayType.get_nd(p, {'fun_name': fun_name, 'entity': 'p'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldp = utils._EngineIntScalarType(
        _ldp_py, {'fun_name': fun_name, 'entity': '_ldp'},
    ).to_ctypes()
    _p_cls = utils._EngineFloat64ArrayType(
        p, {'fun_name': fun_name, 'entity': 'p'},
        exp_shape=(_ip_py, _maxfac_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _p = _p_cls.to_ctypes(can_reference_contents=False)
    _ldt_py = utils._EngineFloat64ArrayType.get_nd(t, {'fun_name': fun_name, 'entity': 't'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldt = utils._EngineIntScalarType(
        _ldt_py, {'fun_name': fun_name, 'entity': '_ldt'},
    ).to_ctypes()
    _t_cls = utils._EngineFloat64ArrayType(
        t, {'fun_name': fun_name, 'entity': 't'},
        exp_shape=(_n_py, _maxfac_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _t = _t_cls.to_ctypes(can_reference_contents=False)
    _ldc_py = utils._EngineFloat64ArrayType.get_nd(c, {'fun_name': fun_name, 'entity': 'c'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldc = utils._EngineIntScalarType(
        _ldc_py, {'fun_name': fun_name, 'entity': '_ldc'},
    ).to_ctypes()
    _c_cls = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_my_py, _maxfac_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _c = _c_cls.to_ctypes(can_reference_contents=False)
    _ldu_py = utils._EngineFloat64ArrayType.get_nd(u, {'fun_name': fun_name, 'entity': 'u'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldu = utils._EngineIntScalarType(
        _ldu_py, {'fun_name': fun_name, 'entity': '_ldu'},
    ).to_ctypes()
    _u_cls = utils._EngineFloat64ArrayType(
        u, {'fun_name': fun_name, 'entity': 'u'},
        exp_shape=(_n_py, _maxfac_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _u = _u_cls.to_ctypes(can_reference_contents=False)
    _xcv_cls = utils._EngineFloat64ArrayType(
        xcv, {'fun_name': fun_name, 'entity': 'xcv'},
        exp_shape=(_maxfac_py,),
        check_contents=False,
    )
    _xcv = _xcv_cls.to_ctypes(can_reference_contents=False)
    _ldycv_py = utils._EngineFloat64ArrayType.get_nd(ycv, {'fun_name': fun_name, 'entity': 'ycv'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldycv = utils._EngineIntScalarType(
        _ldycv_py, {'fun_name': fun_name, 'entity': '_ldycv'},
    ).to_ctypes()
    _ycv_cls = utils._EngineFloat64ArrayType(
        ycv, {'fun_name': fun_name, 'entity': 'ycv'},
        exp_shape=(_maxfac_py, _my_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _ycv = _ycv_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        utils._UNUSED_PRINT_REC,
        _print_rech,
        _n,
        _mx,
        _x,
        _ldx,
        _isx,
        _ip,
        _my,
        _y,
        _ldy,
        _xbar,
        _ybar,
        _iscale,
        _xstd,
        _ystd,
        _maxfac,
        _maxit,
        _tau,
        _xres,
        _ldxres,
        _yres,
        _ldyres,
        _w,
        _ldw,
        _p,
        _ldp,
        _t,
        _ldt,
        _c,
        _ldc,
        _u,
        _ldu,
        _xcv,
        _ycv,
        _ldycv,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n > 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mx = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mx > 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, isx[{:s}] is invalid.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: isx[j-1] = 0 or 1, for all j.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 7:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, my = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: my >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 12:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iscale = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: iscale = -1 or 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 6:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ip = {:s} and mx = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: 1 < ip <= mx.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 15:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, maxfac = {:s} and ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: 1 <= maxfac <= ip.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 16:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, my = {:s} and maxit = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if my > 1, maxit > 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 17:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, tau = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: if my > 1, tau > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 66:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ip = {:s} and sum(isx) = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: the sum of elements in isx must equal ip.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _xbar_cls.refresh_py(_xbar)
        _ybar_cls.refresh_py(_ybar)
        _xstd_cls.refresh_py(_xstd)
        _ystd_cls.refresh_py(_ystd)
        _xres_cls.refresh_py(_xres)
        _yres_cls.refresh_py(_yres)
        _w_cls.refresh_py(_w)
        _p_cls.refresh_py(_p)
        _t_cls.refresh_py(_t)
        _c_cls.refresh_py(_c)
        _u_cls.refresh_py(_u)
        _xcv_cls.refresh_py(_xcv)
        _ycv_cls.refresh_py(_ycv)
        return_data = None
    _err_class.manage()
    return return_data

def pls_fit(ip, my, maxfac, nfact, p, c, w, rcond, b, orig, xbar, ybar, iscale, xstd, ystd, ob, vipopt, ycv, vip):
    u"""
    PLS parameter estimates following partial least squares regression
    by ``pls_svd`` or ``pls_wold``.

    ``pls_fit`` calculates parameter estimates for a given number of
    factors given the output from an orthogonal scores PLS regression
    (``pls_svd`` or ``pls_wold``).

    For full information please refer to the NAG Library document for
    g02lc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02lcf.html

    Parameters
    ----------
    ip : int
        m, the number of predictor variables in the fitted model.

    my : int
        r, the number of response variables.

    maxfac : int
        k, the number of factors available in the PLS model.

    nfact : int
        l, the number of factors to include in the calculation of
        parameter estimates.

    p : float, shape (`ip`, `maxfac`)
        x-loadings as returned from ``pls_svd`` and ``pls_wold``.

    c : float, shape (`my`, `maxfac`)
        y-loadings as returned from ``pls_svd`` and ``pls_wold``.

    w : float, shape (`ip`, `maxfac`)
        x-weights as returned from ``pls_svd`` and ``pls_wold``.

    rcond : float
        Singular values of P^TW less than `rcond` times the maximum
        singular value are treated as zero when calculating parameter
        estimates. If `rcond` is negative, a value of 0.005 is used.

    b : float, shape (`ip`, `my`), modified in place
        `On exit`: `b`[i-1,j-1] contains the parameter estimate for the
        ith predictor variable in the model for the jth response
        variable, for j = 1,2,...,`my`, for i = 1,2,...,`ip`.

    orig : int
        Indicates how parameter estimates are calculated.

        `orig` = -1

            Parameter estimates for the centred, and possibly, scaled
            data.

        `orig` = 1

            Parameter estimates for the original data.

    xbar : float, shape (`ip`)
        If `orig` = 1, mean values of predictor variables in the model;
        otherwise `xbar` is not referenced.

    ybar : float, shape (`my`)
        If `orig` = 1, mean value of each response variable in the
        model; otherwise `ybar` is not referenced.

    iscale : int
        If `orig` = 1, `iscale` must take the value supplied to either
        ``pls_svd`` or ``pls_wold``; otherwise `iscale` is not
        referenced.

    xstd : float, shape (`ip`)
        If `orig` = 1 and `iscale` != -1, the scalings of predictor
        variables in the model as returned from either ``pls_svd`` or
        ``pls_wold``; otherwise `xstd` is not referenced.

    ystd : float, shape (`my`)
        If `orig` = 1 and `iscale` != -1, the scalings of response
        variables as returned from either ``pls_svd`` or ``pls_wold``;
        otherwise `ystd` is not referenced.

    ob : float, shape (:, `my`), modified in place
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `orig`=1: `ip`+1; otherwise: 1.

        `On exit`: if `orig` = 1, `ob`[0,j-1] contains the intercept
        value for the jth response variable, and `ob`[i,j-1] contains
        the parameter estimate on the original scale for the ith
        predictor variable in the model, for j = 1,2,...,`my`, for i =
        1,2,...,`ip`. Otherwise `ob` is not referenced.

    vipopt : int
        A flag that determines variable influence on projections (VIP)
        options.

        `vipopt` = 0

            VIP are not calculated.

        `vipopt` = 1

            VIP are calculated for predictor variables using the mean
            explained variance in responses.

        `vipopt` = `my`

            VIP are calculated for predictor variables for each response
            variable in the model.

        Note that setting `vipopt` = `my` when `my` = 1 gives the same
        result as setting `vipopt` = 1 directly.

    ycv : float, shape (:, `my`)
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `vipopt`!=0: `nfact`; otherwise: 0.

        If `vipopt` != 0, `ycv`[i-1,j-1] is the cumulative percentage of
        variance of the jth response variable explained by the first i
        factors, for j = 1,2,...,`my`, for i = 1,2,...,`nfact`;
        otherwise `ycv` is not referenced.

    vip : float, shape (:, `vipopt`), modified in place
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `vipopt`!=0: `ip`; otherwise: 0.

        `On exit`: if `vipopt` = 1, `vip`[i-1,0] contains the VIP
        statistic for the ith predictor variable in the model for all
        response variables, for i = 1,2,...,`ip`.

        If `vipopt` = `my`, `vip`[i-1,j-1] contains the VIP statistic
        for the ith predictor variable in the model for the jth response
        variable, for j = 1,2,...,`my`, for i = 1,2,...,`ip`.

        Otherwise `vip` is not referenced.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `ip` = *<value>*.

            Constraint: `ip` > 1.

        (`errno` 1)
            On entry, `my` = *<value>*.

            Constraint: `my` >= 1.

        (`errno` 1)
            On entry, `orig` = *<value>*.

            Constraint: `orig` = -1 or 1.

        (`errno` 1)
            On entry, `iscale` = *<value>*.

            Constraint: if `orig` = 1, `iscale` = -1 or 1.

        (`errno` 1)
            On entry, `vipopt` = *<value>* and `my` = *<value>*.

            Constraint: `vipopt` = 0, 1 or `my`.

        (`errno` 2)
            On entry, `maxfac` = *<value>* and `ip` = *<value>*.

            Constraint: 1 <= `maxfac` <= `ip`.

        (`errno` 2)
            On entry, `nfact` = *<value>* and `maxfac` = *<value>*.

            Constraint: 1 <= `nfact` <= `maxfac`.

        (`errno` 2)
            On entry, ldob = *<value>* and `ip` = *<value>*.

            Constraint: if `orig` = 1, ldob >= `ip`+1.

        (`errno` 2)
            On entry, ldycv = *<value>* and `nfact` = *<value>*.

            Constraint: if `vipopt` != 0, ldycv >= `nfact`.

        (`errno` 2)
            On entry, ldvip = *<value>* and `ip` = *<value>*.

            Constraint: if `vipopt` != 0, ldvip >= `ip`.

    References
    ----------
    Wold, S, 1994, `PLS for multivariate linear modelling QSAR:
    chemometric methods in molecular design`, Methods and Principles in
    Medicinal Chemistry, (ed van de Waterbeemd H), Verlag-Chemie

    See Also
    --------
    :meth:`naginterfaces.library.examples.correg.pls_ex.main`
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.pls_fit'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02lcft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, ip.
        utils._BYREF_INT, # Input, my.
        utils._BYREF_INT, # Input, maxfac.
        utils._BYREF_INT, # Input, nfact.
        utils._EngineFloat64ArrayType.c_type, # Input, p[ldp, maxfac].
        utils._BYREF_INT, # Input, ldp.
        utils._EngineFloat64ArrayType.c_type, # Input, c[ldc, maxfac].
        utils._BYREF_INT, # Input, ldc.
        utils._EngineFloat64ArrayType.c_type, # Input, w[ldw, maxfac].
        utils._BYREF_INT, # Input, ldw.
        utils._BYREF_FLOAT64, # Input, rcond.
        utils._EngineFloat64ArrayType.c_type, # Output, b[ldb, my].
        utils._BYREF_INT, # Input, ldb.
        utils._BYREF_INT, # Input, orig.
        utils._EngineFloat64ArrayType.c_type, # Input, xbar[ip].
        utils._EngineFloat64ArrayType.c_type, # Input, ybar[my].
        utils._BYREF_INT, # Input, iscale.
        utils._EngineFloat64ArrayType.c_type, # Input, xstd[ip].
        utils._EngineFloat64ArrayType.c_type, # Input, ystd[my].
        utils._EngineFloat64ArrayType.c_type, # Output, ob[ldob, my].
        utils._BYREF_INT, # Input, ldob.
        utils._BYREF_INT, # Input, vipopt.
        utils._EngineFloat64ArrayType.c_type, # Input, ycv[ldycv, my].
        utils._BYREF_INT, # Input, ldycv.
        utils._EngineFloat64ArrayType.c_type, # Output, vip[ldvip, vipopt].
        utils._BYREF_INT, # Input, ldvip.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((p, 'p'), (c, 'c'), (w, 'w'), (b, 'b'), (ob, 'ob'), (ycv, 'ycv'), (vip, 'vip')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _ip = utils._EngineIntScalarType(
        ip, {'fun_name': fun_name, 'entity': 'ip'},
    ).to_ctypes()
    _my = utils._EngineIntScalarType(
        my, {'fun_name': fun_name, 'entity': 'my'},
    ).to_ctypes()
    _maxfac = utils._EngineIntScalarType(
        maxfac, {'fun_name': fun_name, 'entity': 'maxfac'},
    ).to_ctypes()
    _nfact = utils._EngineIntScalarType(
        nfact, {'fun_name': fun_name, 'entity': 'nfact'},
    ).to_ctypes()
    _ldp_py = utils._EngineFloat64ArrayType.get_nd(p, {'fun_name': fun_name, 'entity': 'p'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldp = utils._EngineIntScalarType(
        _ldp_py, {'fun_name': fun_name, 'entity': '_ldp'},
    ).to_ctypes()
    _ip_py = utils._EngineIntScalarType.to_py(_ip)
    _maxfac_py = utils._EngineIntScalarType.to_py(_maxfac)
    _p = utils._EngineFloat64ArrayType(
        p, {'fun_name': fun_name, 'entity': 'p'},
        exp_shape=(_ip_py, _maxfac_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _ldc_py = utils._EngineFloat64ArrayType.get_nd(c, {'fun_name': fun_name, 'entity': 'c'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldc = utils._EngineIntScalarType(
        _ldc_py, {'fun_name': fun_name, 'entity': '_ldc'},
    ).to_ctypes()
    _my_py = utils._EngineIntScalarType.to_py(_my)
    _c = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_my_py, _maxfac_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _ldw_py = utils._EngineFloat64ArrayType.get_nd(w, {'fun_name': fun_name, 'entity': 'w'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldw = utils._EngineIntScalarType(
        _ldw_py, {'fun_name': fun_name, 'entity': '_ldw'},
    ).to_ctypes()
    _w = utils._EngineFloat64ArrayType(
        w, {'fun_name': fun_name, 'entity': 'w'},
        exp_shape=(_ip_py, _maxfac_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _rcond = utils._EngineFloat64ScalarType(
        rcond, {'fun_name': fun_name, 'entity': 'rcond'},
    ).to_ctypes()
    _ldb_py = utils._EngineFloat64ArrayType.get_nd(b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldb = utils._EngineIntScalarType(
        _ldb_py, {'fun_name': fun_name, 'entity': '_ldb'},
    ).to_ctypes()
    _b_cls = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_ip_py, _my_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _b = _b_cls.to_ctypes(can_reference_contents=False)
    _orig = utils._EngineIntScalarType(
        orig, {'fun_name': fun_name, 'entity': 'orig'},
    ).to_ctypes()
    _xbar = utils._EngineFloat64ArrayType(
        xbar, {'fun_name': fun_name, 'entity': 'xbar'},
        exp_shape=(_ip_py,),
    ).to_ctypes()
    _ybar = utils._EngineFloat64ArrayType(
        ybar, {'fun_name': fun_name, 'entity': 'ybar'},
        exp_shape=(_my_py,),
    ).to_ctypes()
    _iscale = utils._EngineIntScalarType(
        iscale, {'fun_name': fun_name, 'entity': 'iscale'},
    ).to_ctypes()
    _xstd = utils._EngineFloat64ArrayType(
        xstd, {'fun_name': fun_name, 'entity': 'xstd'},
        exp_shape=(_ip_py,),
    ).to_ctypes()
    _ystd = utils._EngineFloat64ArrayType(
        ystd, {'fun_name': fun_name, 'entity': 'ystd'},
        exp_shape=(_my_py,),
    ).to_ctypes()
    _ldob_py = utils._EngineFloat64ArrayType.get_nd(ob, {'fun_name': fun_name, 'entity': 'ob'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldob = utils._EngineIntScalarType(
        _ldob_py, {'fun_name': fun_name, 'entity': '_ldob'},
    ).to_ctypes()
    _orig_py = utils._EngineIntScalarType.to_py(_orig)
    if _orig_py == 1:
        _ob_dim1 = _ip_py+1
    else:
        _ob_dim1 = 1
    exp_shape_ob = (_ob_dim1, _my_py)
    _ob_cls = utils._EngineFloat64ArrayType(
        ob, {'fun_name': fun_name, 'entity': 'ob'},
        exp_shape=exp_shape_ob,
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _ob = _ob_cls.to_ctypes(can_reference_contents=False)
    _vipopt = utils._EngineIntScalarType(
        vipopt, {'fun_name': fun_name, 'entity': 'vipopt'},
    ).to_ctypes()
    _ldycv_py = utils._EngineFloat64ArrayType.get_nd(ycv, {'fun_name': fun_name, 'entity': 'ycv'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldycv = utils._EngineIntScalarType(
        _ldycv_py, {'fun_name': fun_name, 'entity': '_ldycv'},
    ).to_ctypes()
    _vipopt_py = utils._EngineIntScalarType.to_py(_vipopt)
    _nfact_py = utils._EngineIntScalarType.to_py(_nfact)
    if _vipopt_py != 0:
        _ycv_dim1 = _nfact_py
    else:
        _ycv_dim1 = 0
    exp_shape_ycv = (_ycv_dim1, _my_py)
    _ycv = utils._EngineFloat64ArrayType(
        ycv, {'fun_name': fun_name, 'entity': 'ycv'},
        exp_shape=exp_shape_ycv,
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _ldvip_py = utils._EngineFloat64ArrayType.get_nd(vip, {'fun_name': fun_name, 'entity': 'vip'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldvip = utils._EngineIntScalarType(
        _ldvip_py, {'fun_name': fun_name, 'entity': '_ldvip'},
    ).to_ctypes()
    if _vipopt_py != 0:
        _vip_dim1 = _ip_py
    else:
        _vip_dim1 = 0
    exp_shape_vip = (_vip_dim1, _vipopt_py)
    _vip_cls = utils._EngineFloat64ArrayType(
        vip, {'fun_name': fun_name, 'entity': 'vip'},
        exp_shape=exp_shape_vip,
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _vip = _vip_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _ip,
        _my,
        _maxfac,
        _nfact,
        _p,
        _ldp,
        _c,
        _ldc,
        _w,
        _ldw,
        _rcond,
        _b,
        _ldb,
        _orig,
        _xbar,
        _ybar,
        _iscale,
        _xstd,
        _ystd,
        _ob,
        _ldob,
        _vipopt,
        _ycv,
        _ldycv,
        _vip,
        _ldvip,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ip > 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, my = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: my >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 13:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, orig = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: orig = -1 or 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 16:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iscale = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: if orig = 1, iscale = -1 or 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 21:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, vipopt = {:s} and my = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: vipopt = 0, 1 or my.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, maxfac = {:s} and ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: 1 <= maxfac <= ip.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nfact = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'maxfac = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: 1 <= nfact <= maxfac.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 20:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ldob = {:s} and ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if orig = 1, ldob >= ip+1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 23:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ldycv = {:s} and nfact = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if vipopt != 0, ldycv >= nfact.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 25:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ldvip = {:s} and ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if vipopt != 0, ldvip >= ip.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _b_cls.refresh_py(_b)
        _ob_cls.refresh_py(_ob)
        _vip_cls.refresh_py(_vip)
        return_data = None
    _err_class.manage()
    return return_data

def pls_pred(ip, my, orig, xbar, ybar, iscale, xstd, ystd, b, n, mz, isz, z, yhat):
    u"""
    PLS predictions based on parameter estimates from ``pls_fit``.

    ``pls_pred`` calculates predictions given the output from an
    orthogonal scores PLS regression (``pls_svd`` or ``pls_wold``) and
    ``pls_fit``.

    For full information please refer to the NAG Library document for
    g02ld

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02ldf.html

    Parameters
    ----------
    ip : int
        The number of predictor variables in the fitted model. `ip` must
        take the same value as that supplied to ``pls_svd`` or
        ``pls_wold`` to fit the model.

    my : int
        The number of response variables in the fitted model. `my` must
        take the same value as that supplied to ``pls_svd`` or
        ``pls_wold`` to fit the model.

    orig : int
        Indicates how parameter estimates are supplied.

        `orig` = 1

            Parameter estimates are for the original data.

        `orig` = -1

            Parameter estimates are for the centred, and possibly
            scaled, data.

    xbar : float, shape (`ip`)
        If `orig` = -1, `xbar` must contain mean values of predictor
        variables in the model; otherwise `xbar` is not referenced.

    ybar : float, shape (`my`)
        If `orig` = -1, `ybar` must contain the mean value of each
        response variable in the model; otherwise `ybar` is not
        referenced.

    iscale : int
        If `orig` = -1, `iscale` must take the value supplied to either
        ``pls_svd`` or ``pls_wold``; otherwise `iscale` is not
        referenced.

    xstd : float, shape (`ip`)
        If `orig` = -1 and `iscale` != -1, `xstd` must contain the
        scalings of predictor variables in the model as returned from
        either ``pls_svd`` or ``pls_wold``; otherwise `xstd` is not
        referenced.

    ystd : float, shape (`my`)
        If `orig` = -1 and `iscale` != -1, `ystd` must contain the
        scalings of response variables as returned from either
        ``pls_svd`` or ``pls_wold``; otherwise `ystd` is not referenced.

    b : float, shape (:, `my`)
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `orig`=-1: `ip`; if `orig`=1: 1+`ip`;
        otherwise: 0.

        If `orig` = -1, `b` must contain the parameter estimate for the
        centred, and possibly scaled, data as returned by ``pls_fit``;
        otherwise `b` must contain the parameter estimates for the
        original data as returned by ``pls_fit``.

    n : int
        n, the number of observations in the test data Z.

    mz : int
        The number of available predictor variables in the test data.

    isz : int, shape (`mz`)
        Indicates which predictor variables are to be included in the
        model. Predictor variables included from `z` must be in the same
        order as those included in the fitted model.

        If `isz`[j-1] = 1, the jth predictor variable is included in the
        model, for j = 1,2,...,`mz`, otherwise `isz`[j-1] = 0.

    z : float, shape (`n`, `mz`)
        `z`[i-1,j-1] contains the ith observation on the jth available
        predictor variable, for j = 1,2,...,`mz`, for i = 1,2,...,`n`.

    yhat : float, shape (`n`, `my`), modified in place
        `On exit`: `yhat`[i-1,j-1] contains the ith predicted value of
        the jth y-variable in the model.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `ip` = *<value>*.

            Constraint: `ip` > 1.

        (`errno` 1)
            On entry, `my` = *<value>*.

            Constraint: `my` >= 1.

        (`errno` 1)
            On entry, `orig` = *<value>*.

            Constraint: `orig` = -1 or 1.

        (`errno` 1)
            On entry, `iscale` = *<value>*.

            Constraint: if `orig` = -1, `iscale` = -1, 1 or 2.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 1)
            On entry, `isz`[*<value>*] = *<value>*.

            Constraint: `isz`[j-1] = 0 or 1.

        (`errno` 2)
            On entry, ldb = *<value>* and `ip`+1 = *<value>*.

            Constraint: if `orig` = 1, ldb >= 1+`ip`.

        (`errno` 2)
            On entry, ldb = *<value>* and `ip` = *<value>*.

            Constraint: if `orig` = -1, ldb >= `ip`.

        (`errno` 2)
            On entry, `mz` = *<value>* and `ip` = *<value>*.

            Constraint: `mz` >= `ip`.

        (`errno` 3)
            On entry, the number of elements of `isz` equal to 1 is not
            `ip`.

    See Also
    --------
    :meth:`naginterfaces.library.examples.correg.pls_ex.main`
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.pls_pred'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02ldft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, ip.
        utils._BYREF_INT, # Input, my.
        utils._BYREF_INT, # Input, orig.
        utils._EngineFloat64ArrayType.c_type, # Input, xbar[ip].
        utils._EngineFloat64ArrayType.c_type, # Input, ybar[my].
        utils._BYREF_INT, # Input, iscale.
        utils._EngineFloat64ArrayType.c_type, # Input, xstd[ip].
        utils._EngineFloat64ArrayType.c_type, # Input, ystd[my].
        utils._EngineFloat64ArrayType.c_type, # Input, b[ldb, my].
        utils._BYREF_INT, # Input, ldb.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, mz.
        utils._EngineIntArrayType.c_type, # Input, isz[mz].
        utils._EngineFloat64ArrayType.c_type, # Input, z[ldz, mz].
        utils._BYREF_INT, # Input, ldz.
        utils._EngineFloat64ArrayType.c_type, # Output, yhat[ldyhat, my].
        utils._BYREF_INT, # Input, ldyhat.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((b, 'b'), (z, 'z'), (yhat, 'yhat')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _ip = utils._EngineIntScalarType(
        ip, {'fun_name': fun_name, 'entity': 'ip'},
    ).to_ctypes()
    _my = utils._EngineIntScalarType(
        my, {'fun_name': fun_name, 'entity': 'my'},
    ).to_ctypes()
    _orig = utils._EngineIntScalarType(
        orig, {'fun_name': fun_name, 'entity': 'orig'},
    ).to_ctypes()
    _ip_py = utils._EngineIntScalarType.to_py(_ip)
    _xbar = utils._EngineFloat64ArrayType(
        xbar, {'fun_name': fun_name, 'entity': 'xbar'},
        exp_shape=(_ip_py,),
    ).to_ctypes()
    _my_py = utils._EngineIntScalarType.to_py(_my)
    _ybar = utils._EngineFloat64ArrayType(
        ybar, {'fun_name': fun_name, 'entity': 'ybar'},
        exp_shape=(_my_py,),
    ).to_ctypes()
    _iscale = utils._EngineIntScalarType(
        iscale, {'fun_name': fun_name, 'entity': 'iscale'},
    ).to_ctypes()
    _xstd = utils._EngineFloat64ArrayType(
        xstd, {'fun_name': fun_name, 'entity': 'xstd'},
        exp_shape=(_ip_py,),
    ).to_ctypes()
    _ystd = utils._EngineFloat64ArrayType(
        ystd, {'fun_name': fun_name, 'entity': 'ystd'},
        exp_shape=(_my_py,),
    ).to_ctypes()
    _ldb_py = utils._EngineFloat64ArrayType.get_nd(b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldb = utils._EngineIntScalarType(
        _ldb_py, {'fun_name': fun_name, 'entity': '_ldb'},
    ).to_ctypes()
    _orig_py = utils._EngineIntScalarType.to_py(_orig)
    if _orig_py == -1:
        _b_dim1 = _ip_py
    elif _orig_py == 1:
        _b_dim1 = 1+_ip_py
    else:
        _b_dim1 = 0
    exp_shape_b = (_b_dim1, _my_py)
    _b = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=exp_shape_b,
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _mz = utils._EngineIntScalarType(
        mz, {'fun_name': fun_name, 'entity': 'mz'},
    ).to_ctypes()
    _mz_py = utils._EngineIntScalarType.to_py(_mz)
    _isz = utils._EngineIntArrayType(
        isz, {'fun_name': fun_name, 'entity': 'isz'},
        exp_shape=(_mz_py,),
    ).to_ctypes()
    _ldz_py = utils._EngineFloat64ArrayType.get_nd(z, {'fun_name': fun_name, 'entity': 'z'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldz = utils._EngineIntScalarType(
        _ldz_py, {'fun_name': fun_name, 'entity': '_ldz'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _z = utils._EngineFloat64ArrayType(
        z, {'fun_name': fun_name, 'entity': 'z'},
        exp_shape=(_n_py, _mz_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _ldyhat_py = utils._EngineFloat64ArrayType.get_nd(yhat, {'fun_name': fun_name, 'entity': 'yhat'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldyhat = utils._EngineIntScalarType(
        _ldyhat_py, {'fun_name': fun_name, 'entity': '_ldyhat'},
    ).to_ctypes()
    _yhat_cls = utils._EngineFloat64ArrayType(
        yhat, {'fun_name': fun_name, 'entity': 'yhat'},
        exp_shape=(_n_py, _my_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _yhat = _yhat_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _ip,
        _my,
        _orig,
        _xbar,
        _ybar,
        _iscale,
        _xstd,
        _ystd,
        _b,
        _ldb,
        _n,
        _mz,
        _isz,
        _z,
        _ldz,
        _yhat,
        _ldyhat,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ip > 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, my = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: my >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, orig = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: orig = -1 or 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 6:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iscale = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: if orig = -1, iscale = -1, 1 or 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 11:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 14:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, isz[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: isz[j-1] = 0 or 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == -10:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ldb = {:s} and ip+1 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if orig = 1, ldb >= 1+ip.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 10:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ldb = {:s} and ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if orig = -1, ldb >= ip.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 12:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mz = {:s} and ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: mz >= ip.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == -14:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the number of elements of isz equal to 1 is not ip.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _yhat_cls.refresh_py(_yhat)
        return_data = None
    _err_class.manage()
    return return_data

def lars(mtype, pred, prey, n, m, d, isx, y, mnstep, b, fitsum, ropt, lropt, io_manager=None):
    u"""
    Least angle regression (LARS), least absolute shrinkage and
    selection operator (LASSO) and forward stagewise regression.

    ``lars`` performs Least Angle Regression (LARS), forward stagewise
    linear regression or Least Absolute Shrinkage and Selection Operator
    (LASSO).

    For full information please refer to the NAG Library document for
    g02ma

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02maf.html

    Parameters
    ----------
    mtype : int
        Indicates the type of model to fit.

        `mtype` = 1

            LARS is performed.

        `mtype` = 2

            Forward linear stagewise regression is performed.

        `mtype` = 3

            LASSO model is fit.

        `mtype` = 4

            A positive LASSO model is fit.

    pred : int
        Indicates the type of data preprocessing to perform on the
        independent variables supplied in `d` to comply with the
        standardized form of the design matrix.

        `pred` = 0

            No preprocessing is performed.

        `pred` = 1

            Each of the independent variables, x_j, for j = 1,2,...,p,
            are mean centred prior to fitting the model. The means of
            the independent variables, x_, are returned in `b`, with
            x__j = `b`[j-1,`nstep`+1], for j = 1,2,...,p.

        `pred` = 2

            Each independent variable is normalized, with the jth
            variable scaled by 1/sqrt(x_j^Tx_j). The scaling factor used
            by variable j is returned in `b`[j-1,`nstep`].

        `pred` = 3

            As `pred` = 1 or 2, all of the independent variables are
            mean centred prior to being normalized.

        Suggested value: `pred` = 3.

    prey : int
        Indicates the type of data preprocessing to perform on the
        dependent variable supplied in `y`.

        `prey` = 0

            No preprocessing is performed, this is equivalent to setting
            alpha = 0.

        `prey` = 1

            The dependent variable, y, is mean centred prior to fitting
            the model, so alpha = y_. Which is equivalent to fitting a
            non-penalized intercept to the model and the degrees of
            freedom etc. are adjusted accordingly.

        The value of alpha used is returned in `fitsum`[0,`nstep`].

        Suggested value: `prey` = 1.

    n : int
        n, the number of observations.

    m : int
        m, the total number of independent variables.

    d : float, shape (`n`, `m`)
        D, the data, which along with `pred` and `isx`, defines the
        design matrix X. The ith observation for the jth variable must
        be supplied in `d`[i-1,j-1], for j = 1,2,...,`m`, for i =
        1,2,...,`n`.

    isx : None or int, shape (lisx)
        Indicates which independent variables from `d` will be included
        in the design matrix, X.

        If `isx` is None, all variables are included in the design
        matrix.

        Otherwise `isx`[j-1] must be set as follows, for j =
        1,2,...,`m`:

        `isx`[j-1] = 1

            To indicate that the jth variable, as supplied in `d`, is
            included in the design matrix;

        `isx`[j-1] = 0

            To indicated that the jth variable, as supplied in `d`, is
            not included in the design matrix;

        and p = sum(`isx`[j-1]).

    y : float, shape (`n`)
        y, the observations on the dependent variable.

    mnstep : int
        The maximum number of steps to carry out in the model fitting
        process.

        If `mtype` = 1, i.e., a LARS is being performed, the maximum
        number of steps the algorithm will take is min(p,n) if `prey` =
        0, otherwise min(p,n-1).

        If `mtype` = 2, i.e., a forward linear stagewise regression is
        being performed, the maximum number of steps the algorithm will
        take is likely to be several orders of magnitude more and is no
        longer bound by p or n.

        If `mtype` = 3 or 4, i.e., a LASSO or positive LASSO model is
        being fit, the maximum number of steps the algorithm will take
        lies somewhere between that of the LARS and forward linear
        stagewise regression, again it is no longer bound by p or n.

    b : float, shape (p, `mnstep`+2), modified in place
        `On exit`: beta the parameter estimates, with `b`[j-1,k-1] =
        beta_{kj}, the parameter estimate for the jth variable, j =
        1,2,...,p at the kth step of the model fitting process, k =
        1,2,...,`nstep`.

        By default, when `pred` = 2 or 3 the parameter estimates are
        rescaled prior to being returned.

        If the parameter estimates are required on the normalized scale,
        then this can be overridden via `ropt`.

        The values held in the remaining part of `b` depend on the type
        of preprocessing performed.

        If `pred` = 0,

            [table omitted]

        If `pred` = 1,

            [table omitted]

        If `pred` = 2,

            [table omitted]

        If `pred` = 3,

            [table omitted]

        for j = 1,2,...,p.

    fitsum : float, shape (6, `mnstep`+1), modified in place
        `On exit`: summaries of the model fitting process. When k =
        1,2,...,`nstep`,

        `fitsum`[0,k-1]

            ||beta_k||_1, the sum of the absolute values of the
            parameter estimates for the kth step of the modelling
            fitting process. If `pred` = 2 or 3, the scaled parameter
            estimates are used in the summation.

        `fitsum`[1,k-1]

            RSS_k, the residual sums of squares for the kth step, where
            RSS_k = ||y-X^T beta_k||^2.

        `fitsum`[2,k-1]

            nu_k, approximate degrees of freedom for the kth step.

        `fitsum`[3,k-1]

            C_p^(k), a C_p-type statistic for the kth step, where
            C_p^(k) = (RSS_k)/(sigma^2)-n+2*nu_k.

        `fitsum`[4,k-1]

            C^_k, correlation between the residual at step k-1 and the
            most correlated variable not yet in the active set A, where
            the residual at step 0 is y.

        `fitsum`[5,k-1]

            gamma^_k, the step size used at step k.

        In addition

        `fitsum`[0,`nstep`]

            alpha, with alpha = y_ if `prey` = 1 and 0 otherwise.

        `fitsum`[1,`nstep`]

            RSS_0, the residual sums of squares for the null model,
            where RSS_0 = y^Ty when `prey` = 0 and RSS_0 =
            (y-y_)^T(y-y_) otherwise.

        `fitsum`[2,`nstep`]

            nu_0, the degrees of freedom for the null model, where nu_0
            = 0 if `prey` = 0 and nu_0 = 1 otherwise.

        `fitsum`[3,`nstep`]

            C_p^(0), a C_p-type statistic for the null model, where
            C_p^(0) = (RSS_0)/(sigma^2)-n+2*nu_0.

        `fitsum`[4,`nstep`]

            sigma^2, where sigma^2 = (n-RSS_K)/(nu_K) and K = `nstep`.

        Although the C_p statistics described above are returned when
        `errno` = 112 they may not be meaningful due to the estimate
        sigma^2 not being based on the saturated model.

    ropt : None or float, shape (`lropt`)
        Options to control various aspects of the LARS algorithm.

        The default value will be used for `ropt`[i-1] if `lropt` < i,
        therefore setting `lropt` = 0 will use the default values for
        all options and `ropt` need not be set and may be None.

        The default value will also be used if an invalid value is
        supplied for a particular argument, for example, setting
        `ropt`[i-1] = -1 will use the default value for argument i.

        `ropt`[0]

            The minimum step size that will be taken.

            Default is 100*eps, where eps is the machine precision
            returned by ``machine.precision``.

        `ropt`[1]

            General tolerance, used amongst other things, for comparing
            correlations.

            Default is `ropt`[0].

        `ropt`[2]

            If set to 1, parameter estimates are rescaled before being
            returned.

            If set to 0, no rescaling is performed.

            This argument has no effect when `pred` = 0 or 1.

            Default is for the parameter estimates to be rescaled.

        `ropt`[3]

            If set to 1, it is assumed that the model contains an
            intercept during the model fitting process and when
            calculating the degrees of freedom.

            If set to 0, no intercept is assumed.

            This has no effect on the amount of preprocessing performed
            on `y`.

            Default is to treat the model as having an intercept when
            `prey` = 1 and as not having an intercept when `prey` = 0.

        `ropt`[4]

            As implemented, the LARS algorithm can either work directly
            with y and X, or it can work with the cross-product
            matrices, X^Ty and X^TX. In most cases it is more efficient
            to work with the cross-product matrices. This flag allows
            you direct control over which method is used, however, the
            default value will usually be the best choice.

            If `ropt`[4] = 1, y and X are worked with directly.

            If `ropt`[4] = 0, the cross-product matrices are used.

            Default is 1 when p >= 500 and n < p and 0 otherwise.

    lropt : int
        Length of the options array `ropt`.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    ip : int
        p, number of parameter estimates.

        If `isx` is None, p = `m`, i.e., the number of variables in `d`.

        Otherwise p is the number of nonzero values in `isx`.

    nstep : int
        K, the actual number of steps carried out in the model fitting
        process.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, `mtype` = *<value>*.

            Constraint: `mtype` = 1, 2, 3 or 4.

        (`errno` 21)
            On entry, `pred` = *<value>*.

            Constraint: `pred` = 0, 1, 2 or 3.

        (`errno` 31)
            On entry, `prey` = *<value>*.

            Constraint: `prey` = 0 or 1.

        (`errno` 41)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 51)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 81)
            On entry, `isx`[*<value>*] = *<value>*.

            Constraint: `isx`[i] = 0 or 1, for all i.

        (`errno` 82)
            On entry, all values of `isx` are zero.

            Constraint: at least one value of `isx` must be nonzero.

        (`errno` 91)
            On entry, lisx = *<value>* and `m` = *<value>*.

            Constraint: lisx = 0 or `m`.

        (`errno` 111)
            On entry, `mnstep` = *<value>*.

            Constraint: `mnstep` >= 1.

        (`errno` 151)
            On entry, ldb = *<value>* and `m` = *<value>*.

            Constraint: if lisx = 0 then ldb >= `m`.

        (`errno` 152)
            On entry, ldb = *<value>* and p = *<value>*.

            Constraint: if lisx = `m` then ldb >= p.

        (`errno` 181)
            On entry, `lropt` = *<value>*.

            Constraint: 0 <= `lropt` <= 5.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 112)
            Fitting process did not finish in `mnstep` steps. Try
            increasing the size of `mnstep` and supplying larger output
            arrays.

            All output is returned as documented, up to step `mnstep`,
            however, sigma and the C_p statistics may not be meaningful.

        (`errno` 161)
            sigma^2 is approximately zero and hence the C_p-type
            criterion cannot be calculated. All other output is returned
            as documented.

        (`errno` 162)
            nu_K = n, therefore sigma has been set to a large value.
            Output is returned as documented.

        (`errno` 163)
            Degenerate model, no variables added and `nstep` = 0. Output
            is returned as documented.

    References
    ----------
    Efron, B, Hastie, T, Johnstone, I and Tibshirani, R, 2004, `Least
    Angle Regression`, The Annals of Statistics (Volume 32) (2),
    407--499

    Hastie, T, Tibshirani, R and Friedman, J, 2001, `The Elements of
    Statistical Learning: Data Mining, Inference and Prediction`,
    Springer (New York)

    Tibshirani, R, 1996, `Regression Shrinkage and Selection via the
    Lasso`, Journal of the Royal Statistics Society, Series B
    (Methodological) (Volume 58) (1), 267--288

    Weisberg, S, 1985, `Applied Linear Regression`, Wiley

    See Also
    --------
    :meth:`naginterfaces.library.examples.correg.lars_ex.main`
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.lars'

    utils.FileObjManager._check_type_compat(
        io_manager, {'fun_name': fun_name, 'entity': 'io_manager'}
    )

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    def cb_print_rech_py(_, en_data, nout, rec, ifail, rec_len):
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data, utils._EngineIntScalarType.to_py(nout.contents), rec[:rec_len], io_manager, excs, ifail.contents,
                fun_name=fun_name,
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02maft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._PRINT_REC_FUNCTYPE,
        utils._PRINT_RECH_FUNCTYPE,
        utils._BYREF_INT, # Input, sorder.
        utils._BYREF_INT, # Input, mtype.
        utils._BYREF_INT, # Input, pred.
        utils._BYREF_INT, # Input, prey.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Input, d[ldd, m].
        utils._BYREF_INT, # Input, ldd.
        utils._EngineIntArrayType.c_type, # Input, isx[lisx].
        utils._BYREF_INT, # Input, lisx.
        utils._EngineFloat64ArrayType.c_type, # Input, y[n].
        utils._BYREF_INT, # Input, mnstep.
        utils._BYREF_INT, # Output, ip.
        utils._BYREF_INT, # Output, nstep.
        utils._EngineFloat64ArrayType.c_type, # Output, b[ldb, mnstep+2].
        utils._BYREF_INT, # Input, ldb.
        utils._EngineFloat64ArrayType.c_type, # Output, fitsum[6, mnstep+1].
        utils._EngineFloat64ArrayType.c_type, # Input, ropt[lropt].
        utils._BYREF_INT, # Input, lropt.
        utils._BYREF_INT, # Input, monit.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((d, 'd'), (b, 'b'), (fitsum, 'fitsum')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _print_rech = utils._PRINT_RECH_FUNCTYPE(cb_print_rech_py)
    _sorder = utils._EngineIntScalarType.c_type(_nag_enum_consts.nag_col_major)
    _mtype = utils._EngineIntScalarType(
        mtype, {'fun_name': fun_name, 'entity': 'mtype'},
    ).to_ctypes()
    _pred = utils._EngineIntScalarType(
        pred, {'fun_name': fun_name, 'entity': 'pred'},
    ).to_ctypes()
    _prey = utils._EngineIntScalarType(
        prey, {'fun_name': fun_name, 'entity': 'prey'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _ldd_py = utils._EngineFloat64ArrayType.get_nd(d, {'fun_name': fun_name, 'entity': 'd'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldd = utils._EngineIntScalarType(
        _ldd_py, {'fun_name': fun_name, 'entity': '_ldd'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _d = utils._EngineFloat64ArrayType(
        d, {'fun_name': fun_name, 'entity': 'd'},
        exp_shape=(_n_py, _m_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    if isx is not None:
        _lisx = utils._EngineIntScalarType(
            _m_py, {'fun_name': fun_name, 'entity': '_lisx'},
        ).to_ctypes()
    else:
        _lisx = utils._EngineIntScalarType.c_type(0)
    _lisx_py = utils._EngineIntScalarType.to_py(_lisx)
    _isx = utils._EngineIntArrayType(
        isx, {'fun_name': fun_name, 'entity': 'isx'},
        exp_shape=(_lisx_py,),
        allow_none=True,
    ).to_ctypes()
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _mnstep = utils._EngineIntScalarType(
        mnstep, {'fun_name': fun_name, 'entity': 'mnstep'},
    ).to_ctypes()
    _ip = utils._EngineIntScalarType.empty_ctypes()
    _nstep = utils._EngineIntScalarType.empty_ctypes()
    _ldb_py = utils._EngineFloat64ArrayType.get_nd(b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldb = utils._EngineIntScalarType(
        _ldb_py, {'fun_name': fun_name, 'entity': '_ldb'},
    ).to_ctypes()
    if isx is None:
        _vl_p_py = _m_py
    else:
        _vl_p_py = 0
        for _vl_j_py in range(1, _m_py+1):
            if utils._EngineIntArrayType.get_element(isx, {'fun_name': fun_name, 'entity': 'isx'}, (_vl_j_py-1,)) == 1:
                _vl_p_py = _vl_p_py + 1
    _mnstep_py = utils._EngineIntScalarType.to_py(_mnstep)
    _b_cls = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_vl_p_py, _mnstep_py+2),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _b = _b_cls.to_ctypes(can_reference_contents=False)
    _fitsum_cls = utils._EngineFloat64ArrayType(
        fitsum, {'fun_name': fun_name, 'entity': 'fitsum'},
        exp_shape=(6, _mnstep_py+1),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _fitsum = _fitsum_cls.to_ctypes(can_reference_contents=False)
    _lropt = utils._EngineIntScalarType(
        lropt, {'fun_name': fun_name, 'entity': 'lropt'},
    ).to_ctypes()
    _lropt_py = utils._EngineIntScalarType.to_py(_lropt)
    _ropt = utils._EngineFloat64ArrayType(
        ropt, {'fun_name': fun_name, 'entity': 'ropt'},
        exp_shape=(_lropt_py,),
        allow_none=True,
    ).to_ctypes()
    _monit = utils._EngineIntScalarType.c_type(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        utils._UNUSED_PRINT_REC,
        _print_rech,
        _sorder,
        _mtype,
        _pred,
        _prey,
        _n,
        _m,
        _d,
        _ldd,
        _isx,
        _lisx,
        _y,
        _mnstep,
        _ip,
        _nstep,
        _b,
        _ldb,
        _fitsum,
        _ropt,
        _lropt,
        _monit,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 11 and _ifmt == 11:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mtype = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mtype = 1, 2, 3 or 4.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 21 and _ifmt == 21:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, pred = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: pred = 0, 1, 2 or 3.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 31 and _ifmt == 31:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, prey = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: prey = 0 or 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 41 and _ifmt == 41:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 51 and _ifmt == 51:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 81 and _ifmt == 81:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, isx[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: isx[i] = 0 or 1, for all i.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 82 and _ifmt == 82:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, all values of isx are zero.',
                    'Constraint: at least one value of isx must be nonzero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 91 and _ifmt == 91:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lisx = {:s} and m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: lisx = 0 or m.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 111 and _ifmt == 111:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mnstep = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mnstep >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 112 and _ifmt == 112:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Fitting process did not finish in mnstep steps.',
                    'Try increasing the size of mnstep and supplying larger output arrays.',
                    'All output is returned as documented, up to step mnstep, however,',
                    'sigma and the C_p statistics may not be meaningful.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 151 and _ifmt == 151:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ldb = {:s} and m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if lisx = 0 then ldb >= m.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 152 and _ifmt == 152:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ldb = {:s} and p = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if lisx = m then ldb >= p.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 161 and _ifmt == 161:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'sigma^2 is approximately zero and hence the C_p-type criterion',
                    'cannot be calculated. All other output is returned as documented.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 162 and _ifmt == 162:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'nu_K = n, therefore sigma has been set to a large value.',
                    'Output is returned as documented.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 163 and _ifmt == 163:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Degenerate model, no variables added and nstep = 0.',
                    'Output is returned as documented.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 164 and _ifmt == 164:
            _ierr_fl = 163
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Degenerate model, no variables added and nstep = 0.',
                    'Output is returned as documented.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 181 and _ifmt == 181:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lropt = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: 0 <= lropt <= 5.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _b_cls.refresh_py(_b)
        _fitsum_cls.refresh_py(_fitsum)
        _ip_py = utils._EngineIntScalarType.to_py(_ip)
        _nstep_py = utils._EngineIntScalarType.to_py(_nstep)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['ip', 'nstep'],
        )(
            _ip_py,
            _nstep_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def lars_xtx(mtype, pred, intcpt, n, m, dtd, lddtd, isx, dty, yty, mnstep, b, fitsum, ropt, lropt, io_manager=None):
    u"""
    Least Angle Regression (LARS), Least Absolute Shrinkage and
    Selection Operator (LASSO) and forward stagewise regression using
    the cross-products matrix.

    ``lars_xtx`` performs Least Angle Regression (LARS), forward
    stagewise linear regression or Least Absolute Shrinkage and
    Selection Operator (LASSO) using cross-product matrices.

    For full information please refer to the NAG Library document for
    g02mb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02mbf.html

    Parameters
    ----------
    mtype : int
        Indicates the type of model to fit.

        `mtype` = 1

            LARS is performed.

        `mtype` = 2

            Forward linear stagewise regression is performed.

        `mtype` = 3

            LASSO model is fit.

        `mtype` = 4

            A positive LASSO model is fit.

    pred : int
        Indicates the type of preprocessing to perform on the
        cross-products involving the independent variables, i.e., those
        supplied in `dtd` and `dty`.

        `pred` = 0

            No preprocessing is performed.

        `pred` = 2

            Each independent variable is normalized, with the jth
            variable scaled by 1/sqrt(x_j^Tx_j). The scaling factor used
            by variable j is returned in `b`[j-1,`nstep`].

    intcpt : int
        Indicates the type of data preprocessing that was perform on the
        dependent variable, y, prior to calling this function.

        `intcpt` = 0

            No preprocessing was performed.

        `intcpt` = 1

            The dependent variable, y, was mean centred.

    n : int
        n, the number of observations.

    m : int
        m, the total number of independent variables.

    dtd : float, shape (:, :)
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `dtd`.shape[0]=1: 1; if
        `dtd`.shape[1]=1: `m`*(`m`+1)/2; otherwise: `m`.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `dtd`.shape[0]=1: `m`*(`m`+1)/2; if
        `dtd`.shape[1]=1: 1; otherwise: `m`.

        D^TD, the cross-product matrix, which along with `isx`, defines
        the design matrix cross-product X^TX.

        If the cross-product matrix is packed into a single row or
        column of `dtd`, `dtd`[0,i*(i-1)/2+j-1] or
        `dtd`[i*(i-1)/2+j-1,0] must contain the cross-product of the ith
        and jth variable, for j = 1,2,...,`m`, for i = 1,2,...,`m`.

        That is the cross-product stacked by columns as returned by
        ``ssqmat``, for example.

        Otherwise `dtd`[i-1,j-1] must contain the cross-product of the
        ith and jth variable, for j = 1,2,...,`m`, for i = 1,2,...,`m`.

        It should be noted that, even though D^TD is symmetric, the full
        matrix must be supplied.

        The matrix specified in `dtd` must be a valid cross-products
        matrix.

    lddtd : int
        The column separation of elements in `dtd`; i.e., 1 if the
        cross-product matrix is packed into a single row or column of
        `dtd` or `m` otherwise.

    isx : None or int, shape (lisx)
        Indicates which independent variables from `dtd` will be
        included in the design matrix, X.

        If `isx` is None, all variables are included in the design
        matrix.

        Otherwise `isx`[j-1] must be set as follows, for j =
        1,2,...,`m`:

        `isx`[j-1] = 1

            To indicate that the jth variable, as supplied in `dtd`, is
            included in the design matrix;

        `isx`[j-1] = 0

            To indicate that the jth variable, as supplied in `dtd`, is
            not included in the design matrix;

        and p = sum(`isx`[j-1]).

    dty : float, shape (`m`)
        D^Ty, the cross-product between the dependent variable, y, and
        the independent variables D.

    yty : float
        y^Ty, the sums of squares of the dependent variable.

    mnstep : int
        The maximum number of steps to carry out in the model fitting
        process.

        If `mtype` = 1, i.e., a LARS is being performed, the maximum
        number of steps the algorithm will take is min(p,n) if `intcpt`
        = 0, otherwise min(p,n-1).

        If `mtype` = 2, i.e., a forward linear stagewise regression is
        being performed, the maximum number of steps the algorithm will
        take is likely to be several orders of magnitude more and is no
        longer bound by p or n.

        If `mtype` = 3 or 4, i.e., a LASSO or positive LASSO model is
        being fit, the maximum number of steps the algorithm will take
        lies somewhere between that of the LARS and forward linear
        stagewise regression, again it is no longer bound by p or n.

    b : float, shape (p, `mnstep`+1), modified in place
        `On exit`: beta the parameter estimates, with `b`[j-1,k-1] =
        beta_{kj}, the parameter estimate for the jth variable, j =
        1,2,...,p at the kth step of the model fitting process, k =
        1,2,...,`nstep`.

        By default, when `pred` = 2 the parameter estimates are rescaled
        prior to being returned.

        If the parameter estimates are required on the normalized scale,
        then this can be overridden via `ropt`.

        The values held in the remaining part of `b` depend on the type
        of preprocessing performed.

        [table omitted]

        for j = 1,2,...p.

    fitsum : float, shape (6, `mnstep`+1), modified in place
        `On exit`: summaries of the model fitting process. When k =
        1,2,...,`nstep`

        `fitsum`[0,k-1]

            ||beta_k||_1, the sum of the absolute values of the
            parameter estimates for the kth step of the modelling
            fitting process. If `pred` = 2, the scaled parameter
            estimates are used in the summation.

        `fitsum`[1,k-1]

            RSS_k, the residual sums of squares for the kth step, where
            RSS_k = ||y-X^T beta_k||^2.

        `fitsum`[2,k-1]

            nu_k, approximate degrees of freedom for the kth step.

        `fitsum`[3,k-1]

            C_p^(k), a C_p-type statistic for the kth step, where
            C_p^(k) = (RSS_k)/(sigma^2)-n+2*nu_k.

        `fitsum`[4,k-1]

            C^_k, correlation between the residual at step k-1 and the
            most correlated variable not yet in the active set A, where
            the residual at step 0 is y.

        `fitsum`[5,k-1]

            gamma^_k, the step size used at step k.

        In addition

        `fitsum`[0,`nstep`]

            0.

        `fitsum`[1,`nstep`]

            RSS_0, the residual sums of squares for the null model,
            where RSS_0 = y^Ty.

        `fitsum`[2,`nstep`]

            nu_0, the degrees of freedom for the null model, where nu_0
            = 0 if `intcpt` = 0 and nu_0 = 1 otherwise.

        `fitsum`[3,`nstep`]

            C_p^(0), a C_p-type statistic for the null model, where
            C_p^(0) = (RSS_0)/(sigma^2)-n+2*nu_0.

        `fitsum`[4,`nstep`]

            sigma^2, where sigma^2 = (n-RSS_K)/(nu_K) and K = `nstep`.

        Although the C_p statistics described above are returned when
        `errno` = 122 they may not be meaningful due to the estimate
        sigma^2 not being based on the saturated model.

    ropt : None or float, shape (`lropt`)
        Options to control various aspects of the LARS algorithm.

        The default value will be used for `ropt`[i-1] if `lropt` < i,
        therefore setting `lropt` = 0 will use the default values for
        all options and `ropt` need not be set and may be None.

        The default value will also be used if an invalid value is
        supplied for a particular argument, for example, setting
        `ropt`[i-1] = -1 will use the default value for argument i.

        `ropt`[0]

            The minimum step size that will be taken.

            Default is 100*eps is used, where eps is the machine
            precision returned by ``machine.precision``.

        `ropt`[1]

            General tolerance, used amongst other things, for comparing
            correlations.

            Default is `ropt`[0].

        `ropt`[2]

            If set to 1, parameter estimates are rescaled before being
            returned. If set to 0, no rescaling is performed. This
            argument has no effect when `pred` = 0.

            Default is for the parameter estimates to be rescaled.

    lropt : int
        Length of the options array `ropt`.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    ip : int
        p, number of parameter estimates.

        If `isx` is None, p = `m`, i.e., the number of variables in
        `dtd`.

        Otherwise p is the number of nonzero values in `isx`.

    nstep : int
        K, the actual number of steps carried out in the model fitting
        process.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, `mtype` = *<value>*.

            Constraint: `mtype` = 1, 2, 3 or 4.

        (`errno` 21)
            On entry, `pred` = *<value>*.

            Constraint: `pred` = 0 or 2.

        (`errno` 31)
            On entry, `intcpt` = *<value>*.

            Constraint: `intcpt` = 0 or 1.

        (`errno` 41)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 51)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 61)
            The cross-product matrix supplied in `dtd` is not symmetric.

        (`errno` 62)
            On entry, `dtd`[0,*<value>*] = *<value>*.

            Constraint: diagonal elements of D^TD must be positive.

        (`errno` 62)
            On entry, i = *<value>* and `dtd`[i-1,i-1] = *<value>*.

            Constraint: diagonal elements of D^TD must be positive.

        (`errno` 81)
            On entry, `isx`[*<value>*] = *<value>*.

            Constraint: `isx`[i] = 0 or 1, for all i.

        (`errno` 82)
            On entry, all values of `isx` are zero.

            Constraint: at least one value of `isx` must be nonzero.

        (`errno` 91)
            On entry, lisx = *<value>* and `m` = *<value>*.

            Constraint: lisx = 0 or `m`.

        (`errno` 111)
            On entry, `yty` = *<value>*.

            Constraint: `yty` > 0.0.

        (`errno` 112)
            A negative value for the residual sums of squares was
            obtained. Check the values of `dtd`, `dty` and `yty`.

        (`errno` 121)
            On entry, `mnstep` = *<value>*.

            Constraint: `mnstep` >= 1.

        (`errno` 191)
            On entry, `lropt` = *<value>*.

            Constraint: 0 <= `lropt` <= 3.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 122)
            Fitting process did not finished in `mnstep` steps. Try
            increasing the size of `mnstep` and supplying larger output
            arrays.

            All output is returned as documented, up to step `mnstep`,
            however, sigma and the C_p statistics may not be meaningful.

        (`errno` 171)
            sigma^2 is approximately zero and hence the C_p-type
            criterion cannot be calculated. All other output is returned
            as documented.

        (`errno` 172)
            nu_K = n, therefore sigma has been set to a large value.
            Output is returned as documented.

        (`errno` 173)
            Degenerate model, no variables added and `nstep` = 0. Output
            is returned as documented.

    References
    ----------
    Efron, B, Hastie, T, Johnstone, I and Tibshirani, R, 2004, `Least
    Angle Regression`, The Annals of Statistics (Volume 32) (2),
    407--499

    Hastie, T, Tibshirani, R and Friedman, J, 2001, `The Elements of
    Statistical Learning: Data Mining, Inference and Prediction`,
    Springer (New York)

    Tibshirani, R, 1996, `Regression Shrinkage and Selection via the
    Lasso`, Journal of the Royal Statistics Society, Series B
    (Methodological) (Volume 58) (1), 267--288

    Weisberg, S, 1985, `Applied Linear Regression`, Wiley

    See Also
    --------
    :meth:`naginterfaces.library.examples.correg.lars_param_ex.main`
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.lars_xtx'

    utils.FileObjManager._check_type_compat(
        io_manager, {'fun_name': fun_name, 'entity': 'io_manager'}
    )

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    def cb_print_rech_py(_, en_data, nout, rec, ifail, rec_len):
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data, utils._EngineIntScalarType.to_py(nout.contents), rec[:rec_len], io_manager, excs, ifail.contents,
                fun_name=fun_name,
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02mbft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._PRINT_REC_FUNCTYPE,
        utils._PRINT_RECH_FUNCTYPE,
        utils._BYREF_INT, # Input, mtype.
        utils._BYREF_INT, # Input, pred.
        utils._BYREF_INT, # Input, intcpt.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, packed.
        utils._EngineFloat64ArrayType.c_type, # Input, dtd[lddtd, *].
        utils._BYREF_INT, # Input, lddtd.
        utils._EngineIntArrayType.c_type, # Input, isx[lisx].
        utils._BYREF_INT, # Input, lisx.
        utils._EngineFloat64ArrayType.c_type, # Input, dty[m].
        utils._BYREF_FLOAT64, # Input, yty.
        utils._BYREF_INT, # Input, mnstep.
        utils._BYREF_INT, # Output, ip.
        utils._BYREF_INT, # Output, nstep.
        utils._EngineFloat64ArrayType.c_type, # Output, b[ldb, mnstep+1].
        utils._BYREF_INT, # Input, ldb.
        utils._EngineFloat64ArrayType.c_type, # Output, fitsum[6, mnstep+1].
        utils._EngineFloat64ArrayType.c_type, # Input, ropt[lropt].
        utils._BYREF_INT, # Input, lropt.
        utils._BYREF_INT, # Input, monit.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((dtd, 'dtd'), (b, 'b'), (fitsum, 'fitsum')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _print_rech = utils._PRINT_RECH_FUNCTYPE(cb_print_rech_py)
    _mtype = utils._EngineIntScalarType(
        mtype, {'fun_name': fun_name, 'entity': 'mtype'},
    ).to_ctypes()
    _pred = utils._EngineIntScalarType(
        pred, {'fun_name': fun_name, 'entity': 'pred'},
    ).to_ctypes()
    _intcpt = utils._EngineIntScalarType(
        intcpt, {'fun_name': fun_name, 'entity': 'intcpt'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _lddtd = utils._EngineIntScalarType(
        lddtd, {'fun_name': fun_name, 'entity': 'lddtd'},
    ).to_ctypes()
    _lddtd_py = utils._EngineIntScalarType.to_py(_lddtd)
    if _lddtd_py == 1:
        _packed = utils._EngineIntScalarType.c_type(1)
    else:
        _packed = utils._EngineIntScalarType.c_type(0)
    _m_py = utils._EngineIntScalarType.to_py(_m)
    if utils._EngineFloat64ArrayType.get_shape(dtd, {'fun_name': fun_name, 'entity': 'dtd'}, exp_rank=2)[0] == 1:
        _dtd_dim1 = 1
    elif utils._EngineFloat64ArrayType.get_shape(dtd, {'fun_name': fun_name, 'entity': 'dtd'}, exp_rank=2)[1] == 1:
        _dtd_dim1 = _m_py*(_m_py+1)//2
    else:
        _dtd_dim1 = _m_py
    if utils._EngineFloat64ArrayType.get_shape(dtd, {'fun_name': fun_name, 'entity': 'dtd'}, exp_rank=2)[0] == 1:
        _dtd_dim2 = _m_py*(_m_py+1)//2
    elif utils._EngineFloat64ArrayType.get_shape(dtd, {'fun_name': fun_name, 'entity': 'dtd'}, exp_rank=2)[1] == 1:
        _dtd_dim2 = 1
    else:
        _dtd_dim2 = _m_py
    exp_shape_dtd = (_dtd_dim1, _dtd_dim2)
    _dtd = utils._EngineFloat64ArrayType(
        dtd, {'fun_name': fun_name, 'entity': 'dtd'},
        exp_shape=exp_shape_dtd,
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    if isx is not None:
        _lisx = utils._EngineIntScalarType(
            _m_py, {'fun_name': fun_name, 'entity': '_lisx'},
        ).to_ctypes()
    else:
        _lisx = utils._EngineIntScalarType.c_type(0)
    _lisx_py = utils._EngineIntScalarType.to_py(_lisx)
    _isx = utils._EngineIntArrayType(
        isx, {'fun_name': fun_name, 'entity': 'isx'},
        exp_shape=(_lisx_py,),
        allow_none=True,
    ).to_ctypes()
    _dty = utils._EngineFloat64ArrayType(
        dty, {'fun_name': fun_name, 'entity': 'dty'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _yty = utils._EngineFloat64ScalarType(
        yty, {'fun_name': fun_name, 'entity': 'yty'},
    ).to_ctypes()
    _mnstep = utils._EngineIntScalarType(
        mnstep, {'fun_name': fun_name, 'entity': 'mnstep'},
    ).to_ctypes()
    _ip = utils._EngineIntScalarType.empty_ctypes()
    _nstep = utils._EngineIntScalarType.empty_ctypes()
    _ldb_py = utils._EngineFloat64ArrayType.get_nd(b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldb = utils._EngineIntScalarType(
        _ldb_py, {'fun_name': fun_name, 'entity': '_ldb'},
    ).to_ctypes()
    if isx is None:
        _vl_p_py = _m_py
    else:
        _vl_p_py = 0
        for _vl_j_py in range(1, _m_py+1):
            if utils._EngineIntArrayType.get_element(isx, {'fun_name': fun_name, 'entity': 'isx'}, (_vl_j_py-1,)) == 1:
                _vl_p_py = _vl_p_py + 1
    _mnstep_py = utils._EngineIntScalarType.to_py(_mnstep)
    _b_cls = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_vl_p_py, _mnstep_py+1),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _b = _b_cls.to_ctypes(can_reference_contents=False)
    _fitsum_cls = utils._EngineFloat64ArrayType(
        fitsum, {'fun_name': fun_name, 'entity': 'fitsum'},
        exp_shape=(6, _mnstep_py+1),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _fitsum = _fitsum_cls.to_ctypes(can_reference_contents=False)
    _lropt = utils._EngineIntScalarType(
        lropt, {'fun_name': fun_name, 'entity': 'lropt'},
    ).to_ctypes()
    _lropt_py = utils._EngineIntScalarType.to_py(_lropt)
    _ropt = utils._EngineFloat64ArrayType(
        ropt, {'fun_name': fun_name, 'entity': 'ropt'},
        exp_shape=(_lropt_py,),
        allow_none=True,
    ).to_ctypes()
    _monit = utils._EngineIntScalarType.c_type(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        utils._UNUSED_PRINT_REC,
        _print_rech,
        _mtype,
        _pred,
        _intcpt,
        _n,
        _m,
        _packed,
        _dtd,
        _lddtd,
        _isx,
        _lisx,
        _dty,
        _yty,
        _mnstep,
        _ip,
        _nstep,
        _b,
        _ldb,
        _fitsum,
        _ropt,
        _lropt,
        _monit,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 11 and _ifmt == 11:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mtype = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mtype = 1, 2, 3 or 4.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 21 and _ifmt == 21:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, pred = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: pred = 0 or 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 31 and _ifmt == 31:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, intcpt = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: intcpt = 0 or 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 41 and _ifmt == 41:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 51 and _ifmt == 51:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 61 and _ifmt == 61:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The cross-product matrix supplied in dtd is not symmetric.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 62 and _ifmt == 62:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, dtd[0,{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: diagonal elements of D^TD must be positive.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 63 and _ifmt == 63:
            _ierr_fl = 62
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s} and dtd[i-1,i-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: diagonal elements of D^TD must be positive.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 81 and _ifmt == 81:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, isx[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: isx[i] = 0 or 1, for all i.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 82 and _ifmt == 82:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, all values of isx are zero.',
                    'Constraint: at least one value of isx must be nonzero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 91 and _ifmt == 91:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lisx = {:s} and m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: lisx = 0 or m.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 111 and _ifmt == 111:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, yty = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: yty > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 112 and _ifmt == 112:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'A negative value for the residual sums of squares was obtained.',
                    'Check the values of dtd, dty and yty.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 121 and _ifmt == 121:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mnstep = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mnstep >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 122 and _ifmt == 122:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Fitting process did not finished in mnstep steps.',
                    'Try increasing the size of mnstep and supplying larger output arrays.',
                    'All output is returned as documented, up to step mnstep, however,',
                    'sigma and the C_p statistics may not be meaningful.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 171 and _ifmt == 171:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'sigma^2 is approximately zero and hence the C_p-type criterion',
                    'cannot be calculated. All other output is returned as documented.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 172 and _ifmt == 172:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'nu_K = n, therefore sigma has been set to a large value.',
                    'Output is returned as documented.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 173 and _ifmt == 173:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Degenerate model, no variables added and nstep = 0.',
                    'Output is returned as documented.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 174 and _ifmt == 174:
            _ierr_fl = 173
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Degenerate model, no variables added and nstep = 0.',
                    'Output is returned as documented.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 191 and _ifmt == 191:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lropt = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: 0 <= lropt <= 3.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _b_cls.refresh_py(_b)
        _fitsum_cls.refresh_py(_fitsum)
        _ip_py = utils._EngineIntScalarType.to_py(_ip)
        _nstep_py = utils._EngineIntScalarType.to_py(_nstep)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['ip', 'nstep'],
        )(
            _ip_py,
            _nstep_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def lars_param(nstep, ip, b, fitsum, ktype, nk, lnk, nb):
    u"""
    Calculates additional parameter estimates following Least Angle
    Regression (LARS), Least Absolute Shrinkage and Selection Operator
    (LASSO) or forward stagewise regression.

    ``lars_param`` calculates additional parameter estimates following
    Least Angle Regression (LARS), forward stagewise linear regression
    or Least Absolute Shrinkage and Selection Operator (LASSO) as
    performed by ``lars`` and ``lars_xtx``.

    For full information please refer to the NAG Library document for
    g02mc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02mcf.html

    Parameters
    ----------
    nstep : int
        K, the number of steps carried out in the model fitting process,
        as returned by ``lars`` and ``lars_xtx``.

    ip : int
        p, number of parameter estimates, as returned by ``lars`` and
        ``lars_xtx``.

    b : float, shape (`ip`, `nstep`+1)
        beta the parameter estimates, as returned by ``lars`` and
        ``lars_xtx``, with `b`[j-1,k-1] = beta_{kj}, the parameter
        estimate for the jth variable, for j = 1,2,...,p, at the kth
        step of the model fitting process.

    fitsum : float, shape (6, `nstep`+1)
        Summaries of the model fitting process, as returned by ``lars``
        and ``lars_xtx``.

    ktype : int
        Indicates what target values are held in `nk`.

        `ktype` = 1

            `nk` holds (fractional) LARS step numbers.

        `ktype` = 2

            `nk` holds values for L_1 norm of the (scaled) parameters.

        `ktype` = 3

            `nk` holds ratios with respect to the largest (scaled) L_1
            norm.

        `ktype` = 4

            `nk` holds values for the L_1 norm of the (unscaled)
            parameters.

        `ktype` = 5

            `nk` holds ratios with respect to the largest (unscaled) L_1
            norm.

        If ``lars`` was called with pred = 0 or 1 or ``lars_xtx`` was
        called with pred = 0 then the model fitting routine did not
        rescale the independent variables, X, prior to fitting the model
        and therefore there is no difference between `ktype` = 2 or 3
        and `ktype` = 4 or 5.

    nk : float, shape (`lnk`)
        Target values used for predicting the new set of parameter
        estimates.

    lnk : int
        Number of values supplied in `nk`.

    nb : float, shape (`ip`, `lnk`), modified in place
        `On exit`: beta~ the predicted parameter estimates, with
        `b`[j-1,i-1] = beta~_{ij}, the parameter estimate for variable
        j, j = 1,2,...,p at the point in the fitting process associated
        with `nk`[i-1], i = 1,2,...,`lnk`.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, `nstep` = *<value>*.

            Constraint: `nstep` >= 0.

        (`errno` 21)
            On entry, `ip` = *<value>*.

            Constraint: `ip` >= 1.

        (`errno` 31)
            `b` has been corrupted since the last call to ``lars`` or
            ``lars_xtx``.

        (`errno` 51)
            `fitsum` has been corrupted since the last call to ``lars``
            or ``lars_xtx``.

        (`errno` 61)
            On entry, `ktype` = *<value>*.

            Constraint: `ktype` = 1, 2, 3, 4 or 5.

        (`errno` 71)
            On entry, `ktype` = 1, `nk`[*<value>*] = *<value>* and
            `nstep` = *<value>*

            Constraint: 0 <= `nk`[i] <= `nstep`, for all i.

        (`errno` 72)
            On entry, `ktype` = 2, `nk`[*<value>*] = *<value>*, `nstep`
            = *<value>* and `fitsum`[0,`nstep`-1] = *<value>*.

            Constraint: 0 <= `nk`[i] <= `fitsum`[0,`nstep`-1], for all
            i.

        (`errno` 73)
            On entry, `ktype` = 3 or 5, `nk`[*<value>*] = *<value>*.

            Constraint: 0 <= `nk`[i] <= 1, for all i.

        (`errno` 74)
            On entry, `ktype` = 4, `nk`[*<value>*] = *<value>* and
            ||beta_K||_1 = *<value>*

            Constraint: 0 <= `nk`[i] <= ||beta_K||_1, for all i.

        (`errno` 81)
            On entry, `lnk` = *<value>*.

            Constraint: `lnk` >= 1.

    References
    ----------
    Efron, B, Hastie, T, Johnstone, I and Tibshirani, R, 2004, `Least
    Angle Regression`, The Annals of Statistics (Volume 32) (2),
    407--499

    Hastie, T, Tibshirani, R and Friedman, J, 2001, `The Elements of
    Statistical Learning: Data Mining, Inference and Prediction`,
    Springer (New York)

    Tibshirani, R, 1996, `Regression Shrinkage and Selection via the
    Lasso`, Journal of the Royal Statistics Society, Series B
    (Methodological) (Volume 58) (1), 267--288

    Weisberg, S, 1985, `Applied Linear Regression`, Wiley

    See Also
    --------
    :meth:`naginterfaces.library.examples.correg.lars_param_ex.main`
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.lars_param'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02mcft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, nstep.
        utils._BYREF_INT, # Input, ip.
        utils._EngineFloat64ArrayType.c_type, # Input, b[ldb, nstep+1].
        utils._BYREF_INT, # Input, ldb.
        utils._EngineFloat64ArrayType.c_type, # Input, fitsum[6, nstep+1].
        utils._BYREF_INT, # Input, ktype.
        utils._EngineFloat64ArrayType.c_type, # Input, nk[lnk].
        utils._BYREF_INT, # Input, lnk.
        utils._EngineFloat64ArrayType.c_type, # Output, nb[ldnb, lnk].
        utils._BYREF_INT, # Input, ldnb.
        utils._BYREF_INT, # Input, ilnk.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((b, 'b'), (fitsum, 'fitsum'), (nb, 'nb')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _nstep = utils._EngineIntScalarType(
        nstep, {'fun_name': fun_name, 'entity': 'nstep'},
    ).to_ctypes()
    _ip = utils._EngineIntScalarType(
        ip, {'fun_name': fun_name, 'entity': 'ip'},
    ).to_ctypes()
    _ldb_py = utils._EngineFloat64ArrayType.get_nd(b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldb = utils._EngineIntScalarType(
        _ldb_py, {'fun_name': fun_name, 'entity': '_ldb'},
    ).to_ctypes()
    _ip_py = utils._EngineIntScalarType.to_py(_ip)
    _nstep_py = utils._EngineIntScalarType.to_py(_nstep)
    _b = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_ip_py, _nstep_py+1),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _fitsum = utils._EngineFloat64ArrayType(
        fitsum, {'fun_name': fun_name, 'entity': 'fitsum'},
        exp_shape=(6, _nstep_py+1),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _ktype = utils._EngineIntScalarType(
        ktype, {'fun_name': fun_name, 'entity': 'ktype'},
    ).to_ctypes()
    _lnk = utils._EngineIntScalarType(
        lnk, {'fun_name': fun_name, 'entity': 'lnk'},
    ).to_ctypes()
    _lnk_py = utils._EngineIntScalarType.to_py(_lnk)
    _nk = utils._EngineFloat64ArrayType(
        nk, {'fun_name': fun_name, 'entity': 'nk'},
        exp_shape=(_lnk_py,),
    ).to_ctypes()
    _ldnb_py = utils._EngineFloat64ArrayType.get_nd(nb, {'fun_name': fun_name, 'entity': 'nb'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldnb = utils._EngineIntScalarType(
        _ldnb_py, {'fun_name': fun_name, 'entity': '_ldnb'},
    ).to_ctypes()
    _nb_cls = utils._EngineFloat64ArrayType(
        nb, {'fun_name': fun_name, 'entity': 'nb'},
        exp_shape=(_ip_py, _lnk_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _nb = _nb_cls.to_ctypes(can_reference_contents=False)
    _ilnk = utils._EngineIntScalarType.c_type(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _nstep,
        _ip,
        _b,
        _ldb,
        _fitsum,
        _ktype,
        _nk,
        _lnk,
        _nb,
        _ldnb,
        _ilnk,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 11 and _ifmt == 11:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nstep = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nstep >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 21 and _ifmt == 21:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ip >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 31 and _ifmt == 31:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'b has been corrupted since the last call to lars or lars_xtx.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 51 and _ifmt == 51:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'fitsum has been corrupted since the last call to lars or lars_xtx.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 52 and _ifmt == 52:
            _ierr_fl = 51
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'fitsum has been corrupted since the last call to lars or lars_xtx.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 53 and _ifmt == 53:
            _ierr_fl = 51
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'fitsum has been corrupted since the last call to lars or lars_xtx.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 61 and _ifmt == 61:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ktype = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ktype = 1, 2, 3, 4 or 5.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 71 and _ifmt == 71:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ktype = 1, nk[{:s}] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'and nstep = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: 0 <= nk[i] <= nstep, for all i.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 72 and _ifmt == 72:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ktype = 2, nk[{:s}] = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'nstep = {:s} and fitsum[0,nstep-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[4]),
                    ),
                    'Constraint: 0 <= nk[i] <= fitsum[0,nstep-1], for all i.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 73 and _ifmt == 73:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ktype = 3 or 5, nk[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: 0 <= nk[i] <= 1, for all i.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 74 and _ifmt == 74:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ktype = 4, nk[{:s}] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'and ||beta_K||_1 = {:s}'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: 0 <= nk[i] <= ||beta_K||_1, for all i.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 81 and _ifmt == 81:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lnk = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: lnk >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _nb_cls.refresh_py(_nb)
        return_data = None
    _err_class.manage()
    return return_data

def quantile_linreg_easy(n, m, x, y, ntau, tau, b, bl, bu, info, io_manager=None):
    u"""
    Linear quantile regression, simple interface, independent,
    identically distributed (IID) errors.

    ``quantile_linreg_easy`` performs a multiple linear quantile
    regression, returning the parameter estimates and associated
    confidence limits based on an assumption of Normal, independent,
    identically distributed errors. ``quantile_linreg_easy`` is a
    simplified version of ``quantile_linreg``.

    For full information please refer to the NAG Library document for
    g02qf

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02qff.html

    Parameters
    ----------
    n : int
        n, the number of observations in the dataset.

    m : int
        p, the number of variates in the model.

    x : float, shape (`n`, `m`)
        X, the design matrix, with the ith value for the jth variate
        supplied in `x`[i-1,j-1], for j = 1,2,...,`m`, for i =
        1,2,...,`n`.

    y : float, shape (`n`)
        y, the observations on the dependent variable.

    ntau : int
        The number of quantiles of interest.

    tau : float, shape (`ntau`)
        The vector of quantiles of interest. A separate model is fitted
        to each quantile.

    b : float, shape (`m`, `ntau`), modified in place
        `On exit`: beta^, the estimates of the parameters of the
        regression model, with `b`[j-1,l-1] containing the coefficient
        for the variable in column j of `x`, estimated for tau =
        `tau`[l-1].

    bl : float, shape (`m`, `ntau`), modified in place
        `On exit`: beta^_L, the lower limit of a 95% confidence interval
        for beta^, with `bl`[j-1,l-1] holding the lower limit associated
        with `b`[j-1,l-1].

    bu : float, shape (`m`, `ntau`), modified in place
        `On exit`: beta^_U, the upper limit of a 95% confidence interval
        for beta^, with `bu`[j-1,l-1] holding the upper limit associated
        with `b`[j-1,l-1].

    info : int, shape (`ntau`), modified in place
        `On exit`: `info`[l] holds additional information concerning the
        model fitting and confidence limit calculations when tau =
        `tau`[l].

        [table omitted]

        It is possible for multiple warnings to be applicable to a
        single model.

        In these cases the value returned in `info` is the sum of the
        corresponding individual nonzero warning codes.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    df : float
        The degrees of freedom given by n-k, where n is the number of
        observations and k is the rank of the cross-product matrix X^TX.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 2.

        (`errno` 21)
            On entry, `m` = *<value>* and `n` = *<value>*.

            Constraint: 1 <= `m` < `n`.

        (`errno` 51)
            On entry, `ntau` = *<value>*.

            Constraint: `ntau` >= 1.

        (`errno` 61)
            On entry, `tau`[*<value>*] = *<value>*.

            Constraint: sqrt(epsilon) < `tau`[l-1] < 1-sqrt(epsilon)
            where epsilon is the machine precision returned by
            ``machine.precision``, for all `ntau`.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 111)
            A potential problem occurred whilst fitting the model(s).

            Additional information has been returned in `info`.

    References
    ----------
    Koenker, R, 2005, `Quantile Regression`, Econometric Society
    Monographs, Cambridge University Press, New York
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.quantile_linreg_easy'

    utils.FileObjManager._check_type_compat(
        io_manager, {'fun_name': fun_name, 'entity': 'io_manager'}
    )

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    def cb_print_rech_py(_, en_data, nout, rec, ifail, rec_len):
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data, utils._EngineIntScalarType.to_py(nout.contents), rec[:rec_len], io_manager, excs, ifail.contents,
                fun_name=fun_name,
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02qfft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._PRINT_REC_FUNCTYPE,
        utils._PRINT_RECH_FUNCTYPE,
        utils._BYREF_INT, # Input, sorder.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Input, x[ldx, sdx].
        utils._EngineFloat64ArrayType.c_type, # Input, y[n].
        utils._BYREF_INT, # Input, ntau.
        utils._EngineFloat64ArrayType.c_type, # Input, tau[ntau].
        utils._BYREF_FLOAT64, # Output, df.
        utils._EngineFloat64ArrayType.c_type, # Output, b[m, ntau].
        utils._EngineFloat64ArrayType.c_type, # Output, bl[m, ntau].
        utils._EngineFloat64ArrayType.c_type, # Output, bu[m, ntau].
        utils._EngineIntArrayType.c_type, # Output, info[ntau].
        utils._BYREF_INT, # Input, ldx.
        utils._BYREF_INT, # Input, sdx.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((x, 'x'), (b, 'b'), (bl, 'bl'), (bu, 'bu')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _print_rech = utils._PRINT_RECH_FUNCTYPE(cb_print_rech_py)
    _sorder = utils._EngineIntScalarType.c_type(_nag_enum_consts.nag_col_major)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _ldx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _sdx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='s')
    _sdx = utils._EngineIntScalarType(
        _sdx_py, {'fun_name': fun_name, 'entity': '_sdx'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _sorder_py = utils._EngineIntScalarType.to_py(_sorder)
    _m_py = utils._EngineIntScalarType.to_py(_m)
    if _sorder_py == 1:
        _x_dim2 = _m_py
    elif _sorder_py == 2:
        _x_dim2 = _n_py
    else:
        _x_dim2 = 0
    exp_shape_x = (_n_py, _x_dim2)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=exp_shape_x,
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _ntau = utils._EngineIntScalarType(
        ntau, {'fun_name': fun_name, 'entity': 'ntau'},
    ).to_ctypes()
    _ntau_py = utils._EngineIntScalarType.to_py(_ntau)
    _tau = utils._EngineFloat64ArrayType(
        tau, {'fun_name': fun_name, 'entity': 'tau'},
        exp_shape=(_ntau_py,),
    ).to_ctypes()
    _df = utils._EngineFloat64ScalarType.empty_ctypes()
    _b_cls = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_m_py, _ntau_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _b = _b_cls.to_ctypes(can_reference_contents=False)
    _bl_cls = utils._EngineFloat64ArrayType(
        bl, {'fun_name': fun_name, 'entity': 'bl'},
        exp_shape=(_m_py, _ntau_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _bl = _bl_cls.to_ctypes(can_reference_contents=False)
    _bu_cls = utils._EngineFloat64ArrayType(
        bu, {'fun_name': fun_name, 'entity': 'bu'},
        exp_shape=(_m_py, _ntau_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _bu = _bu_cls.to_ctypes(can_reference_contents=False)
    _info_cls = utils._EngineIntArrayType(
        info, {'fun_name': fun_name, 'entity': 'info'},
        exp_shape=(_ntau_py,),
        check_contents=False,
    )
    _info = _info_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        utils._UNUSED_PRINT_REC,
        _print_rech,
        _sorder,
        _n,
        _m,
        _x,
        _y,
        _ntau,
        _tau,
        _df,
        _b,
        _bl,
        _bu,
        _info,
        _ldx,
        _sdx,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ifmt == 41:
            _ierr_fl = 11
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 2.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ifmt == 51:
            _ierr_fl = 21
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: 1 <= m < n.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ifmt == 91:
            _ierr_fl = 21
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: 1 <= m < n.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ifmt == 121:
            _ierr_fl = 51
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ntau = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ntau >= 1.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ifmt == 131:
            _ierr_fl = 61
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, tau[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: sqrt(epsilon) < tau[l-1] < 1-sqrt(epsilon)',
                    'where epsilon is the machine precision returned by machine.precision, for all ntau.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ifmt == 231:
            _ierr_fl = 111
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'A potential problem occurred whilst fitting the model(s).',
                    'Additional information has been returned in info.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _b_cls.refresh_py(_b)
        _bl_cls.refresh_py(_bl)
        _bu_cls.refresh_py(_bu)
        _info_cls.refresh_py(_info)
        _df_py = utils._EngineFloat64ScalarType.to_py(_df)
        return_data = _df_py
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = _collections.namedtuple(
                'nag_return_data',
                ['df'],
            )(
                _df_py,
            )
    _err_class.manage()
    return return_data

def quantile_linreg(sorder, intcpt, n, m, dat, isx, ip, y, wt, ntau, tau, b, bl, bu, ch, res, comm, statecomm, info, io_manager=None):
    u"""
    Linear quantile regression, comprehensive interface.

    Note: this function uses optional algorithmic parameters, see also:
    ``optset``, ``optget``.

    ``quantile_linreg`` performs a multiple linear quantile regression.
    Parameter estimates and, if required, confidence limits, covariance
    matrices and residuals are calculated. ``quantile_linreg`` may be
    used to perform a weighted quantile regression.
    A simplified interface for ``quantile_linreg`` is provided by
    ``quantile_linreg_easy``.

    For full information please refer to the NAG Library document for
    g02qg

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02qgf.html

    Parameters
    ----------
    sorder : int
        Determines the storage order of variates supplied in `dat`.

    intcpt : str, length 1
        Indicates whether an intercept will be included in the model.
        The intercept is included by adding a column of ones as the
        first column in the design matrix, X.

        `intcpt` = 'Y'

            An intercept will be included in the model.

        `intcpt` = 'N'

            An intercept will not be included in the model.

    n : int
        The total number of observations in the dataset. If no weights
        are supplied, or no zero weights are supplied or observations
        with zero weights are included in the model then `n` = n.
        Otherwise `n` = n+ the number of observations with zero weights.

    m : int
        m, the total number of variates in the dataset.

    dat : float, shape (:, :)
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `sorder`=1: `n`; otherwise: `m`.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `sorder`=1: `m`; otherwise: `n`.

        The ith value for the jth variate, for j = 1,2,...,`m`, for i =
        1,2,...,`n`, must be supplied in

            `dat`[i-1,j-1] if `sorder` = 1, and

            `dat`[j-1,i-1] if `sorder` = 2.

        The design matrix X is constructed from `dat`, `isx` and
        `intcpt`.

    isx : int, shape (`m`)
        Indicates which independent variables are to be included in the
        model.

        `isx`[j-1] = 0

            The jth variate, supplied in `dat`, is not included in the
            regression model.

        `isx`[j-1] = 1

            The jth variate, supplied in `dat`, is included in the
            regression model.

    ip : int
        p, the number of independent variables in the model, including
        the intercept, see `intcpt`, if present.

    y : float, shape (`n`)
        y, the observations on the dependent variable.

    wt : None or float, shape (:)
        Note: the required length for this argument is determined as
        follows: if `wt` is not None: `n`; otherwise: 0.

        If not None, `wt` must contain the diagonal elements of the
        weight matrix W.

        If weights are not provided then `wt` must be set to None.

        When

        'Drop Zero Weights' = 'YES'

            If `wt`[i-1] = 0.0, the ith observation is not included in
            the model, in which case the effective number of
            observations, n, is the number of observations with nonzero
            weights. If 'Return Residuals' = 'YES', the values of `res`
            will be set to zero for observations with zero weights.

        'Drop Zero Weights' = 'NO'

            All observations are included in the model and the effective
            number of observations is `n`, i.e., n = `n`.

    ntau : int
        The number of quantiles of interest.

    tau : float, shape (`ntau`)
        The vector of quantiles of interest. A separate model is fitted
        to each quantile.

    b : float, shape (`ip`, `ntau`), modified in place
        `On entry`: if 'Calculate Initial Values' = 'NO', `b`[i-1,l-1]
        must hold an initial estimates for beta^_i, for l =
        1,2,...,`ntau`, for i = 1,2,...,`ip`. If 'Calculate Initial
        Values' = 'YES', `b` need not be set.

        `On exit`: `b`[i-1,l-1], for i = 1,2,...,`ip`, contains the
        estimates of the parameters of the regression model, beta^,
        estimated for tau = `tau`[l-1].

        If `intcpt` = 'Y', `b`[0,l-1] will contain the estimate
        corresponding to the intercept and `b`[i,l-1] will contain the
        coefficient of the jth variate contained in `dat`, where
        `isx`[j-1] is the ith nonzero value in the array `isx`.

        If `intcpt` = 'N', `b`[i-1,l-1] will contain the coefficient of
        the jth variate contained in `dat`, where `isx`[j-1] is the ith
        nonzero value in the array `isx`.

    bl : None or float, shape (`ip`, :), modified in place
        Note: the required extent for this argument in dimension 2 is
        determined as follows: if rip>0: `ntau`; otherwise: 0.

        `On exit`, if not None on entry: if 'Interval Method' != 'NONE',
        `bl`[i-1,l-1] contains the lower limit of an (100*alpha)%
        confidence interval for `b`[i-1,l-1], for l = 1,2,...,`ntau`,
        for i = 1,2,...,`ip`.

        If 'Interval Method' = 'NONE', `bl` is not referenced.

        The method used for calculating the interval is controlled by
        the options 'Interval Method' and 'Bootstrap Interval Method'.

        The size of the interval, alpha, is controlled by the option
        'Significance Level'.

    bu : None or float, shape (`ip`, :), modified in place
        Note: the required extent for this argument in dimension 2 is
        determined as follows: if rip>0: `ntau`; otherwise: 0.

        `On exit`, if not None on entry: if 'Interval Method' != 'NONE',
        `bu`[i-1,l-1] contains the upper limit of an (100*alpha)%
        confidence interval for `b`[i-1,l-1], for l = 1,2,...,`ntau`,
        for i = 1,2,...,`ip`.

        If 'Interval Method' = 'NONE', `bu` is not referenced.

        The method used for calculating the interval is controlled by
        the options 'Interval Method' and 'Bootstrap Interval Method'.

        The size of the interval, alpha is controlled by the option
        'Significance Level'.

    ch : None or float, shape (`ip`, `ip`, :), modified in place
        Note: the required extent for this argument in dimension 3 is
        determined as follows: if rip>0 and tdch>0: `ntau`; otherwise:
        `ntau`+1.

        `On exit`, if not None on entry: depending on the supplied
        options, `ch` will either not be referenced, hold an estimate of
        the upper triangular part of the covariance matrix, Sigma, or an
        estimate of the upper triangular parts of n*J_n and n^-1*H_n^-1.

        If 'Interval Method' = 'NONE' or 'Matrix Returned' = 'NONE',
        `ch` is not referenced.

        If 'Interval Method' = 'BOOTSTRAP XY' or 'IID' and 'Matrix
        Returned' = 'H INVERSE', `ch` is not referenced.

        Otherwise, for i,j = 1,2,…,`ip`,j >= i and l = 1,2,…,`ntau`:

            If 'Matrix Returned' = 'COVARIANCE', `ch`[i-1,j-1,l-1] holds
            an estimate of the covariance between `b`[i-1,l-1] and
            `b`[j-1,l-1].

            If 'Matrix Returned' = 'H INVERSE', `ch`[i-1,j-1,0] holds an
            estimate of the (i,j)th element of n*J_n and `ch`[i-1,j-1,l]
            holds an estimate of the (i,j)th element of n^-1*H_n^-1, for
            tau = `tau`[l-1].

        The method used for calculating Sigma and H_n^-1 is controlled
        by the option 'Interval Method'.

    res : None or float, shape (`n`, :), modified in place
        Note: the required extent for this argument in dimension 2 is
        determined as follows: if sdres>0: `ntau`; otherwise: 0.

        `On exit`, if not None on entry: if 'Return Residuals' = 'YES',
        `res`[i-1,l-1] holds the (weighted) residuals, r_i, for tau =
        `tau`[l-1], for l = 1,2,...,`ntau`, for i = 1,2,...,`n`.

        If `wt` is not None and 'Drop Zero Weights' = 'YES', the value
        of `res` will be set to zero for observations with zero weights.

        If 'Return Residuals' = 'NO', `res` is not referenced and can be
        set to None.

    comm : None or dict, communication object
        Communication structure.

        If not None, this argument must have been initialized by a prior
        call to ``optset``.

    statecomm : None or dict, RNG communication object, modified in
    place
        RNG communication structure.

        If 'Interval Method' = 'BOOTSTRAP XY', this argument must have
        been initialized by a prior call to ``rand.init_repeat`` or
        ``rand.init_nonrepeat``.

    info : int, shape (`ntau`), modified in place
        `On exit`: `info`[i] holds additional information concerning the
        model fitting and confidence limit calculations when tau =
        `tau`[i].

        [table omitted]

        It is possible for multiple warnings to be applicable to a
        single model.

        In these cases the value returned in `info` is the sum of the
        corresponding individual nonzero warning codes.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    df : float
        The degrees of freedom given by n-k, where n is the effective
        number of observations and k is the rank of the cross-product
        matrix X^TX.

    Other Parameters
    ----------------
    'Band Width Alpha' : float
        Default = 1.0

        A multiplier used to construct the parameter alpha_b used when
        calculating the Sheather--Hall bandwidth (see [Description]),
        with alpha_b = (1-alpha)*'Band Width Alpha'.
        Here, alpha is the 'Significance Level'.

    'Band Width Method' : str
        Default = 'SHEATHER HALL'

        The method used to calculate the bandwidth used in the
        calculation of the asymptotic covariance matrix Sigma and H^-1
        if 'Interval Method' = 'HKS', 'KERNEL' or 'IID' (see
        [Description]).

    'Big' : float
        Default = 10.0^20

        This argument should be set to something larger than the biggest
        value supplied in `dat` and `y`.

    'Bootstrap Interval Method' : str
        Default = 'QUANTILE'

        If 'Interval Method' = 'BOOTSTRAP XY', 'Bootstrap Interval
        Method' controls how the confidence intervals are calculated
        from the bootstrap estimates.

        'Bootstrap Interval Method' = 'T'
            t intervals are calculated. That is, the covariance matrix,
            Sigma = {sigma_{ij}:i,j = 1,2,...,p} is calculated from the
            bootstrap estimates and the limits calculated as
            beta_i+/-t_(n-p,(1+alpha)/2)sigma_{ii} where
            t_(n-p,(1+alpha)/2) is the (1+alpha)/2 percentage point from
            a Student's t distribution on n-p degrees of freedom, n is
            the effective number of observations and alpha is given by
            the option 'Significance Level'.

        'Bootstrap Interval Method' = 'QUANTILE'
            Quantile intervals are calculated. That is, the upper and
            lower limits are taken as the (1+alpha)/2 and (1-alpha)/2
            quantiles of the bootstrap estimates, as calculated using
            ``stat.quantiles``.

    'Bootstrap Iterations' : int
        Default = 100

        The number of bootstrap samples used to calculate the confidence
        limits and covariance matrix (if requested) when 'Interval
        Method' = 'BOOTSTRAP XY'.

    'Bootstrap Monitoring' : str
        Default = 'NO'

        If 'Bootstrap Monitoring' = 'YES' and 'Interval Method' =
        'BOOTSTRAP XY', the parameter estimates for each of the
        bootstrap samples are displayed.
        This information is sent to the unit number specified by 'Unit
        Number'.

    'Calculate Initial Values' : str
        Default = 'YES'

        If 'Calculate Initial Values' = 'YES' then the initial values
        for the regression parameters, beta, are calculated from the
        data.
        Otherwise they must be supplied in `b`.

    'Defaults' : valueless
        This special keyword is used to reset all options to their
        default values.

    'Drop Zero Weights' : str
        Default = 'YES'

        If a weighted regression is being performed and 'Drop Zero
        Weights' = 'YES' then observations with zero weight are dropped
        from the analysis.
        Otherwise such observations are included.

    'Epsilon' : float
        Default = sqrt(epsilon)

        epsilon_u, the tolerance used when calculating the covariance
        matrix and the initial values for u and v.
        For additional details see [Calculation of Covariance Matrix]
        and [Additional information] respectively.

    'Interval Method' : str
        Default = 'IID'

        The value of 'Interval Method' controls whether confidence
        limits are returned in `bl` and `bu` and how these limits are
        calculated.
        This argument also controls how the matrices returned in `ch`
        are calculated.

        'Interval Method' = 'NONE'
            No limits are calculated and `bl`, `bu` and `ch` are not
            referenced.

        'Interval Method' = 'KERNEL'
            The Powell Sandwich method with a Gaussian kernel is used.

        'Interval Method' = 'HKS'
            The Hendricks--Koenker Sandwich is used.

        'Interval Method' = 'IID'
            The errors are assumed to be identical, and independently
            distributed.

        'Interval Method' = 'BOOTSTRAP XY'
            A bootstrap method is used, where sampling is done on the
            pair (y_i,x_i). The number of bootstrap samples is
            controlled by the argument 'Bootstrap Iterations' and the
            type of interval constructed from the bootstrap samples is
            controlled by 'Bootstrap Interval Method'.

    'Iteration Limit' : int
        Default = 100

        The maximum number of iterations to be performed by the interior
        point optimization algorithm.

    'Matrix Returned' : str
        Default = 'NONE'

        The value of 'Matrix Returned' controls the type of matrices
        returned in `ch`.
        If 'Interval Method' = 'NONE', this argument is ignored and `ch`
        is not referenced.
        Otherwise:

        'Matrix Returned' = 'NONE'
            No matrices are returned and `ch` is not referenced.

        'Matrix Returned' = 'COVARIANCE'
            The covariance matrices are returned.

        'Matrix Returned' = 'H INVERSE'
            If 'Interval Method' = 'KERNEL' or 'HKS', the matrices J and
            H^-1 are returned. Otherwise no matrices are returned and
            `ch` is not referenced.

        The matrices returned are calculated as described in
        [Description], with the algorithm used specified by 'Interval
        Method'.
        In the case of 'Interval Method' = 'BOOTSTRAP XY' the covariance
        matrix is calculated directly from the bootstrap estimates.

    'Monitoring' : str
        Default = 'NO'

        If 'Monitoring' = 'YES' then the duality gap is displayed at
        each iteration of the interior point optimization algorithm.
        In addition, the final estimates for beta are also displayed.

        The monitoring information is sent to the unit number specified
        by 'Unit Number'.

    'QR Tolerance' : float
        Default = epsilon^0.9

        The tolerance used to calculate the rank, k, of the p*p
        cross-product matrix, X^TX.
        Letting Q be the orthogonal matrix obtained from a QR
        decomposition of X^TX, then the rank is calculated by comparing
        Q_{ii} with Q_{11}*'QR Tolerance'.

        If the cross-product matrix is rank deficient, the parameter
        estimates for the p-k columns with the smallest values of Q_{ii}
        are set to zero, along with the corresponding entries in `bl`,
        `bu` and `ch`, if returned.
        This is equivalent to dropping these variables from the model.
        Details on the QR decomposition used can be found in
        ``lapackeig.dgeqp3``.

    'Return Residuals' : str
        Default = 'NO'

        If 'Return Residuals' = 'YES', the residuals are returned in
        `res`.
        Otherwise `res` is not referenced.

    'Sigma' : float
        Default = 0.99995

        The scaling factor used when calculating the affine scaling step
        size (see equation [equation]).

    'Significance Level' : float
        Default = 0.95

        alpha, the size of the confidence interval whose limits are
        returned in `bl` and `bu`.

    'Tolerance' : float
        Default = sqrt(epsilon)

        Convergence tolerance.
        The optimization is deemed to have converged if the duality gap
        is less than 'Tolerance' (see [Update and convergence]).

    'Unit Number' : int
        Default = advisory message unit number

        The unit number to which any monitoring information is sent.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, `sorder` = *<value>*.

            Constraint: `sorder` = 1 or 2.

        (`errno` 21)
            On entry, `intcpt` = *<value>*.

            Constraint: `intcpt` = 'N' or 'Y'.

        (`errno` 41)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 2.

        (`errno` 51)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 0.

        (`errno` 81)
            On entry, `isx`[*<value>*] = *<value>*.

            Constraint: `isx`[i] = 0 or 1, for all i.

        (`errno` 91)
            On entry, `ip` = *<value>* and `n` = *<value>*.

            Constraint: 1 <= `ip` < `n`.

        (`errno` 92)
            On entry, `ip` is not consistent with `isx` or `intcpt`:
            `ip` = *<value>*, expected value = *<value>*.

        (`errno` 111)
            On entry, `wt`[*<value>*] = *<value>*.

            Constraint: `wt`[i] >= 0.0, for all i.

        (`errno` 112)
            On entry, effective number of observations = *<value>*.

            Constraint: effective number of observations >= *<value>*.

        (`errno` 121)
            On entry, `ntau` = *<value>*.

            Constraint: `ntau` >= 1.

        (`errno` 131)
            On entry, `tau`[*<value>*] = *<value>*.

            Constraint: sqrt(epsilon) < `tau`[l-1] < 1-sqrt(epsilon)
            where epsilon is the machine precision returned by
            ``machine.precision``, for all `ntau`.

        (`errno` 201)
            On entry, either the option arrays have not been initialized
            or they have been corrupted.

        (`errno` 221)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 231)
            A potential problem occurred whilst fitting the model(s).

            Additional information has been returned in `info`.

    References
    ----------
    Koenker, R, 2005, `Quantile Regression`, Econometric Society
    Monographs, Cambridge University Press, New York

    Mehrotra, S, 1992, `On the implementation of a primal-dual interior
    point method`, SIAM J. Optim. (2), 575--601

    Nocedal, J and Wright, S J, 2006, `Numerical Optimization`, (2nd
    Edition), Springer Series in Operations Research, Springer, New York

    Portnoy, S and Koenker, R, 1997, `The Gaussian hare and the
    Laplacian tortoise: computability of squared-error versus absolute
    error estimators`, Statistical Science (4), 279--300

    Powell, J L, 1991, `Estimation of monotonic regression models under
    quantile restrictions`, `Nonparametric and Semiparametric Methods in
    Econometrics`, Cambridge University Press, Cambridge

    See Also
    --------
    :meth:`naginterfaces.library.examples.correg.quantile_linreg_ex.main`
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.quantile_linreg'

    utils.FileObjManager._check_type_compat(
        io_manager, {'fun_name': fun_name, 'entity': 'io_manager'}
    )

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    def cb_print_rech_py(_, en_data, nout, rec, ifail, rec_len):
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data, utils._EngineIntScalarType.to_py(nout.contents), rec[:rec_len], io_manager, excs, ifail.contents,
                fun_name=fun_name,
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02qgft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._PRINT_REC_FUNCTYPE,
        utils._PRINT_RECH_FUNCTYPE,
        utils._BYREF_INT, # Input, sorder.
        utils._BYREF_INT, # Input, ic1.
        utils._BYREF_INT, # Input, lwt.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Input, dat[lddat, sddat].
        utils._BYREF_INT, # Input, lddat.
        utils._EngineIntArrayType.c_type, # Input, isx[m].
        utils._BYREF_INT, # Input, ip.
        utils._EngineFloat64ArrayType.c_type, # Input, y[n].
        utils._EngineFloat64ArrayType.c_type, # Input, wt[lwt].
        utils._BYREF_INT, # Input, ntau.
        utils._EngineFloat64ArrayType.c_type, # Input, tau[ntau].
        utils._BYREF_FLOAT64, # Output, df.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, b[ip, ntau].
        utils._EngineFloat64ArrayType.c_type, # Output, bl[rip, ntau].
        utils._EngineFloat64ArrayType.c_type, # Output, bu[rip, ntau].
        utils._EngineFloat64ArrayType.c_type, # Output, ch[rip, rip, tdch].
        utils._EngineFloat64ArrayType.c_type, # Output, res[n, sdres].
        utils._EngineIntArrayType.c_type, # Input, iopts[liopts].
        utils._EngineFloat64ArrayType.c_type, # Input, opts[lopts].
        utils._EngineIntArrayType.c_type, # Input/Output, state[lstate].
        utils._EngineIntArrayType.c_type, # Output, info[ntau].
        utils._BYREF_INT, # Input, sddat.
        utils._BYREF_INT, # Input, rip.
        utils._BYREF_INT, # Input, tdch.
        utils._BYREF_INT, # Input, sdres.
        utils._BYREF_INT, # Input, liopts.
        utils._BYREF_INT, # Input, lopts.
        utils._BYREF_INT, # Input, lstate.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((dat, 'dat'), (b, 'b'), (bl, 'bl'), (bu, 'bu'), (ch, 'ch'), (res, 'res')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _print_rech = utils._PRINT_RECH_FUNCTYPE(cb_print_rech_py)
    _sorder = utils._EngineIntScalarType(
        sorder, {'fun_name': fun_name, 'entity': 'sorder'},
    ).to_ctypes()
    _intcpt = utils._EngineCharScalarType(
        intcpt, {'fun_name': fun_name, 'entity': 'intcpt'}, exp_chrlen=1,
    ).to_ctypes()
    _intcpt_py = utils._EngineCharScalarType.to_py(_intcpt)
    if _intcpt_py.upper() == 'Y':
        _ic1 = utils._EngineIntScalarType.c_type(1)
    elif _intcpt_py.upper() == 'N':
        _ic1 = utils._EngineIntScalarType.c_type(0)
    else:
        _err_class = utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                'On entry, intcpt = {:s}.'.format(
                    utils._STR_FMT_STR.format(_intcpt_py),
                ),
                'Constraint: intcpt = \'N\' or \'Y\'.',
            ],
            errcodes=(21,),
        )
        _err_class.manage()
    _lwt_py = utils._EngineFloat64ArrayType.get_shape(wt, {'fun_name': fun_name, 'entity': 'wt'}, exp_rank=1, allow_none=True)[0]
    _lwt = utils._EngineIntScalarType(
        _lwt_py, {'fun_name': fun_name, 'entity': '_lwt'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _lddat_py = utils._EngineFloat64ArrayType.get_nd(dat, {'fun_name': fun_name, 'entity': 'dat'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _lddat = utils._EngineIntScalarType(
        _lddat_py, {'fun_name': fun_name, 'entity': '_lddat'},
    ).to_ctypes()
    _sddat_py = utils._EngineFloat64ArrayType.get_nd(dat, {'fun_name': fun_name, 'entity': 'dat'}, exp_rank=2, sorder=_m_sorder, dim_type='s')
    _sddat = utils._EngineIntScalarType(
        _sddat_py, {'fun_name': fun_name, 'entity': '_sddat'},
    ).to_ctypes()
    _sorder_py = utils._EngineIntScalarType.to_py(_sorder)
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _m_py = utils._EngineIntScalarType.to_py(_m)
    if _sorder_py == 1:
        _dat_dim1 = _n_py
    else:
        _dat_dim1 = _m_py
    if _sorder_py == 1:
        _dat_dim2 = _m_py
    else:
        _dat_dim2 = _n_py
    exp_shape_dat = (_dat_dim1, _dat_dim2)
    _dat = utils._EngineFloat64ArrayType(
        dat, {'fun_name': fun_name, 'entity': 'dat'},
        exp_shape=exp_shape_dat,
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _isx = utils._EngineIntArrayType(
        isx, {'fun_name': fun_name, 'entity': 'isx'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _ip = utils._EngineIntScalarType(
        ip, {'fun_name': fun_name, 'entity': 'ip'},
    ).to_ctypes()
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    if wt is not None:
        _wt_dim1 = _n_py
    else:
        _wt_dim1 = 0
    exp_shape_wt = (_wt_dim1,)
    _wt = utils._EngineFloat64ArrayType(
        wt, {'fun_name': fun_name, 'entity': 'wt'},
        exp_shape=exp_shape_wt,
        allow_none=True,
    ).to_ctypes()
    _ntau = utils._EngineIntScalarType(
        ntau, {'fun_name': fun_name, 'entity': 'ntau'},
    ).to_ctypes()
    _ntau_py = utils._EngineIntScalarType.to_py(_ntau)
    _tau = utils._EngineFloat64ArrayType(
        tau, {'fun_name': fun_name, 'entity': 'tau'},
        exp_shape=(_ntau_py,),
    ).to_ctypes()
    _df = utils._EngineFloat64ScalarType.empty_ctypes()
    _ip_py = utils._EngineIntScalarType.to_py(_ip)
    utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        req_keys=('iopts', 'opts'),
        allow_none=True,
    )
    utils._check_comm(
        statecomm, {'fun_name': fun_name, 'entity': 'statecomm'},
        req_keys=('state',),
        allow_none=True,
    )
    _sdres, _rip, _tdch, _liopts, _lopts, _lstate, _calcib = utils._g02zkbn(0 if comm is None else 1, 'G02QGF', _ip, _ntau, comm, statecomm)
    _calcib_py = utils._EngineIntScalarType.to_py(_calcib)
    _vl_g02zkb_calcib_py = _calcib_py
    _b_cls = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_ip_py, _ntau_py),
        mutual_sorder=_m_sorder,
        check_contents=_vl_g02zkb_calcib_py == 0,
    )
    _b = _b_cls.to_ctypes(can_reference_contents=_vl_g02zkb_calcib_py == 0)
    _rip_py = utils._EngineIntScalarType.to_py(_rip)
    _vl_g02zkb_rip_py = _rip_py
    if _vl_g02zkb_rip_py > 0:
        if bl is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'bl'},
                'must not be None when rip > 0.',
            )
    _rip_py = utils._EngineIntScalarType.to_py(_rip)
    if bl is None:
        _bl = utils._EngineFloat64ArrayType.empty_ctypes((_rip_py)*(_ntau_py))
    else:
        _bl_cls = utils._EngineFloat64ArrayType(
            bl, {'fun_name': fun_name, 'entity': 'bl'},
            exp_shape=(_rip_py, _ntau_py),
            mutual_sorder=_m_sorder,
            check_contents=False,
            allow_none=True,
        )
        _bl = _bl_cls.to_ctypes(can_reference_contents=False)
    if _vl_g02zkb_rip_py > 0:
        if bu is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'bu'},
                'must not be None when rip > 0.',
            )
    if bu is None:
        _bu = utils._EngineFloat64ArrayType.empty_ctypes((_rip_py)*(_ntau_py))
    else:
        _bu_cls = utils._EngineFloat64ArrayType(
            bu, {'fun_name': fun_name, 'entity': 'bu'},
            exp_shape=(_rip_py, _ntau_py),
            mutual_sorder=_m_sorder,
            check_contents=False,
            allow_none=True,
        )
        _bu = _bu_cls.to_ctypes(can_reference_contents=False)
    _tdch_py = utils._EngineIntScalarType.to_py(_tdch)
    _vl_g02zkb_tdch_py = _tdch_py
    if _vl_g02zkb_rip_py > 0 and _vl_g02zkb_tdch_py > 0:
        if ch is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'ch'},
                'must not be None when rip > 0 and tdch > 0.',
            )
    _tdch_py = utils._EngineIntScalarType.to_py(_tdch)
    if ch is None:
        _ch = utils._EngineFloat64ArrayType.empty_ctypes((_rip_py)*(_rip_py)*(_tdch_py))
    else:
        _ch_cls = utils._EngineFloat64ArrayType(
            ch, {'fun_name': fun_name, 'entity': 'ch'},
            exp_shape=(_rip_py, _rip_py, _tdch_py),
            mutual_sorder=_m_sorder,
            check_contents=False,
            allow_none=True,
        )
        _ch = _ch_cls.to_ctypes(can_reference_contents=False)
    _sdres_py = utils._EngineIntScalarType.to_py(_sdres)
    _vl_g02zkb_sdres_py = _sdres_py
    if _vl_g02zkb_sdres_py > 0:
        if res is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'res'},
                'must not be None when sdres > 0.',
            )
    _sdres_py = utils._EngineIntScalarType.to_py(_sdres)
    if res is None:
        _res = utils._EngineFloat64ArrayType.empty_ctypes((_n_py)*(_sdres_py))
    else:
        _res_cls = utils._EngineFloat64ArrayType(
            res, {'fun_name': fun_name, 'entity': 'res'},
            exp_shape=(_n_py, _sdres_py),
            mutual_sorder=_m_sorder,
            check_contents=False,
            allow_none=True,
        )
        _res = _res_cls.to_ctypes(can_reference_contents=False)
    _liopts_py = utils._EngineIntScalarType.to_py(_liopts)
    _vl_g02zkb_liopts_py = _liopts_py
    _vl_dummy_iopts = utils._EngineIntArrayType.empty_ctypes(0)
    if comm is None:
        _iopts = _vl_dummy_iopts
    else:
        _iopts = comm['iopts']
    _lopts_py = utils._EngineIntScalarType.to_py(_lopts)
    _vl_g02zkb_lopts_py = _lopts_py
    _vl_dummy_opts = utils._EngineFloat64ArrayType.empty_ctypes(0)
    if comm is None:
        _opts = _vl_dummy_opts
    else:
        _opts = comm['opts']
    _lstate_py = utils._EngineIntScalarType.to_py(_lstate)
    _vl_g02zkb_lstate_py = _lstate_py
    _vl_dummy_state = utils._EngineIntArrayType.empty_ctypes(0)
    if statecomm is None:
        _state = _vl_dummy_state
    else:
        _state = statecomm['state']
    _info_cls = utils._EngineIntArrayType(
        info, {'fun_name': fun_name, 'entity': 'info'},
        exp_shape=(_ntau_py,),
        check_contents=False,
    )
    _info = _info_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        utils._UNUSED_PRINT_REC,
        _print_rech,
        _sorder,
        _ic1,
        _lwt,
        _n,
        _m,
        _dat,
        _lddat,
        _isx,
        _ip,
        _y,
        _wt,
        _ntau,
        _tau,
        _df,
        _b,
        _bl,
        _bu,
        _ch,
        _res,
        _iopts,
        _opts,
        _state,
        _info,
        _sddat,
        _rip,
        _tdch,
        _sdres,
        _liopts,
        _lopts,
        _lstate,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ifmt == 11:
            _ierr_fl = 11
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, sorder = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: sorder = 1 or 2.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ifmt == 41:
            _ierr_fl = 41
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 2.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ifmt == 51:
            _ierr_fl = 51
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 0.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ifmt == 81:
            _ierr_fl = 81
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, isx[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: isx[i] = 0 or 1, for all i.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ifmt == 91:
            _ierr_fl = 91
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ip = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: 1 <= ip < n.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ifmt == 92:
            _ierr_fl = 92
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ip is not consistent with isx or intcpt:',
                    'ip = {:s}, expected value = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ifmt == 111:
            _ierr_fl = 111
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, wt[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: wt[i] >= 0.0, for all i.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ifmt == 112:
            _ierr_fl = 112
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, effective number of observations = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: effective number of observations >= {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ifmt == 121:
            _ierr_fl = 121
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ntau = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ntau >= 1.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ifmt == 131:
            _ierr_fl = 131
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, tau[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: sqrt(epsilon) < tau[l-1] < 1-sqrt(epsilon)',
                    'where epsilon is the machine precision returned by machine.precision, for all ntau.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ifmt == 201:
            _ierr_fl = 201
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, either the option arrays have not been initialized or they have',
                    'been corrupted.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ifmt == 221:
            _ierr_fl = 221
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ifmt == 231:
            _ierr_fl = 231
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'A potential problem occurred whilst fitting the model(s).',
                    'Additional information has been returned in info.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _b_cls.refresh_py(_b)
        _rip_py = utils._EngineIntScalarType.to_py(_rip)
        _vl_g02zkb_rip_py = _rip_py
        if _vl_g02zkb_rip_py > 0:
            _bl_cls.refresh_py(_bl)
        if _vl_g02zkb_rip_py > 0:
            _bu_cls.refresh_py(_bu)
        _tdch_py = utils._EngineIntScalarType.to_py(_tdch)
        _vl_g02zkb_tdch_py = _tdch_py
        if _vl_g02zkb_rip_py > 0 and _vl_g02zkb_tdch_py > 0:
            _ch_cls.refresh_py(_ch)
        _sdres_py = utils._EngineIntScalarType.to_py(_sdres)
        _vl_g02zkb_sdres_py = _sdres_py
        if _vl_g02zkb_sdres_py > 0:
            _res_cls.refresh_py(_res)
        _info_cls.refresh_py(_info)
        _df_py = utils._EngineFloat64ScalarType.to_py(_df)
        return_data = _df_py
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = _collections.namedtuple(
                'nag_return_data',
                ['df'],
            )(
                _df_py,
            )
    _err_class.manage()
    return return_data

def optset(optstr, comm):
    u"""
    Option setting function for ``quantile_linreg``.

    ``optset`` either initializes or resets the option arrays or sets a
    single option for supported problem solving functions in submodule
    ``correg``.
    Currently, only ``quantile_linreg`` is supported.

    For full information please refer to the NAG Library document for
    g02zk

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02zkf.html

    Parameters
    ----------
    optstr : str
        A string identifying the option to be set.

        Initialize = function name

            Initialize the option arrays `comm`['iopts'] and
            `comm`['opts'] for use with function function name, where
            function name is the name of the problem solving function
            you wish to use.

        Defaults

            Resets all options to their default values.

        option = optval

            See [Other Parameters for quantile_linreg] for details of
            valid values for option and optval. The equals sign (=)
            delimiter must be used to separate the option from its
            optval value.

        `optstr` is case insensitive.

        Each token in the option and optval component must be separated
        by at least one space.

    comm : dict, communication object, modified in place
        Communication structure.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, the option supplied in `optstr` was not
            recognized: `optstr` = *<value>*.

        (`errno` 12)
            On entry, the expected delimiter '=' was not found in
            `optstr`: `optstr` = *<value>*.

        (`errno` 13)
            On entry, could not convert the specified optval to an
            integer: `optstr` = *<value>*.

        (`errno` 13)
            On entry, could not convert the specified optval to a real:
            `optstr` = *<value>*.

        (`errno` 14)
            On entry, attempting to initialize the option arrays but
            specified function name was not valid: name = *<value>*.

        (`errno` 15)
            On entry, the optval supplied for the integer option is not
            valid.

            `optstr` = *<value>*.

        (`errno` 16)
            On entry, the optval supplied for the real option is not
            valid.

            `optstr` = *<value>*.

        (`errno` 17)
            On entry, the optval supplied for the character option is
            not valid.

            `optstr` = *<value>*.

        (`errno` 21)
            On entry, either the option arrays have not been initialized
            or they have been corrupted.

    See Also
    --------
    :meth:`naginterfaces.library.examples.correg.quantile_linreg_ex.main`
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.optset'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02zkft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, optstr(len=*).
        utils._EngineIntArrayType.c_type, # Input/Output, iopts[liopts].
        utils._BYREF_INT, # Input, liopts.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, opts[lopts].
        utils._BYREF_INT, # Input, lopts.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of optstr.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _optstr = utils._EngineCharScalarType(
        optstr, {'fun_name': fun_name, 'entity': 'optstr'}, exp_chrlen=None,
    ).to_ctypes()
    _comm_needs_init = utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        opt_keys=('iopts', 'opts'),
    )
    if _comm_needs_init:
        if _comm_needs_init:
            _iopts_dim1 = 100
        else:
            _iopts_dim1 = utils._EngineIntArrayType.get_shape(comm['iopts'], {'fun_name': fun_name, 'entity': 'comm[\'iopts\']'}, exp_rank=1, validate=False)[0]
        _iopts = utils._EngineIntArrayType.empty_ctypes(_iopts_dim1)
        comm['iopts'] = _iopts
    else:
        _iopts = comm['iopts']
    _liopts_py = utils._EngineIntArrayType.get_shape(comm['iopts'], {'fun_name': fun_name, 'entity': 'comm[\'iopts\']'}, exp_rank=1, validate=False)[0]
    _liopts = utils._EngineIntScalarType(
        _liopts_py, {'fun_name': fun_name, 'entity': '_liopts'},
    ).to_ctypes()
    if _comm_needs_init:
        if _comm_needs_init:
            _opts_dim1 = 100
        else:
            _opts_dim1 = utils._EngineFloat64ArrayType.get_shape(comm['opts'], {'fun_name': fun_name, 'entity': 'comm[\'opts\']'}, exp_rank=1, validate=False)[0]
        _opts = utils._EngineFloat64ArrayType.empty_ctypes(_opts_dim1)
        comm['opts'] = _opts
    else:
        _opts = comm['opts']
    _lopts_py = utils._EngineFloat64ArrayType.get_shape(comm['opts'], {'fun_name': fun_name, 'entity': 'comm[\'opts\']'}, exp_rank=1, validate=False)[0]
    _lopts = utils._EngineIntScalarType(
        _lopts_py, {'fun_name': fun_name, 'entity': '_lopts'},
    ).to_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _optstr,
        _iopts,
        _liopts,
        _opts,
        _lopts,
        _errbuf,
        _ierr,
        _optstr._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        _optstr_py = utils._EngineCharScalarType.to_py(_optstr)
        if _ifmt == 11:
            _ierr_fl = 11
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the option supplied in optstr was not recognized:',
                    'optstr = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_optstr_py),
                    ),
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ifmt == 12:
            _ierr_fl = 12
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the expected delimiter \'=\' was not found in optstr:',
                    'optstr = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_optstr_py),
                    ),
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ifmt == 13:
            _ierr_fl = 13
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, could not convert the specified optval to an integer:',
                    'optstr = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_optstr_py),
                    ),
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ifmt == 14:
            _ierr_fl = 13
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, could not convert the specified optval to a real:',
                    'optstr = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_optstr_py),
                    ),
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ifmt == 15:
            _ierr_fl = 14
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, attempting to initialize the option arrays but',
                    'specified function name was not valid:',
                    'name = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ifmt >= 1001 and _ifmt <= 1004:
            _ierr_fl = 15
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the optval supplied for the integer option is not',
                    'valid.',
                    'optstr = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_optstr_py),
                    ),
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ifmt >= 1011 and _ifmt <= 1014:
            _ierr_fl = 16
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the optval supplied for the real option is not valid.',
                    'optstr = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_optstr_py),
                    ),
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ifmt == 1021:
            _ierr_fl = 17
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the optval supplied for the character option is not',
                    'valid.',
                    'optstr = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_optstr_py),
                    ),
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ifmt == 21:
            _ierr_fl = 21
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, either the option arrays have not been initialized or they have',
                    'been corrupted.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        return_data = None
    _err_class.manage()
    return return_data

def optget(optstr, comm):
    u"""
    Option getting function for ``quantile_linreg``.

    ``optget`` is used to query the value of options available to
    supported problem solving functions in submodule ``correg``.
    Currently, only ``quantile_linreg`` is supported.

    For full information please refer to the NAG Library document for
    g02zl

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/g02/g02zlf.html

    Parameters
    ----------
    optstr : str
        A string identifying the option whose current value is required.
        See [Other Parameters for quantile_linreg] for information on
        valid options. In addition, the following is a valid option:

        Identify

            ``optget`` returns in `cvalue` the function name supplied to
            ``optset`` when the option arrays `comm`['iopts'] and
            `comm`['opts'] were initialized.

    comm : dict, communication object
        Communication structure.

        This argument must have been initialized by a prior call to
        ``quantile_linreg`` or ``optset``.

    Returns
    -------
    ivalue : int
        If the option supplied in `optstr` is an integer valued
        parameter, `ivalue` will hold its current value.

    rvalue : float
        If the option supplied in `optstr` is a real valued parameter,
        `rvalue` will hold its current value.

    cvalue : str
        If the option supplied in `optstr` is a character valued
        parameter, `cvalue` will hold its current value, unless Identify
        is specified, see `optstr`.

    optype : int
        Indicates whether the option supplied in `optstr` is an integer,
        real or character valued parameter and hence which of `ivalue`,
        `rvalue` or `cvalue` holds the current value.

        `optype` = 1

            `optstr` is an integer valued option, its current value has
            been returned in `ivalue`.

        `optype` = 2

            `optstr` is a real valued option, its current value has been
            returned in `rvalue`.

        `optype` = 3

            `optstr` is a character valued option, its current value has
            been returned in `cvalue`.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, the option in `optstr` was not recognized:
            `optstr` = *<value>*.

        (`errno` 61)
            On entry, either the option arrays have not been initialized
            or they have been corrupted.

    See Also
    --------
    :meth:`naginterfaces.library.examples.correg.lmm_init_combine_ex.main`
    """

    excs = []
    fun_name = 'naginterfaces.base.correg.optget'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g02zlft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, optstr(len=*).
        utils._BYREF_INT, # Output, ivalue.
        utils._BYREF_FLOAT64, # Output, rvalue.
        utils._BYREF_STRING, # Output, cvalue(len=*).
        utils._BYREF_INT, # Output, optype.
        utils._EngineIntArrayType.c_type, # Input, iopts[liopts].
        utils._BYREF_INT, # Input, liopts.
        utils._EngineFloat64ArrayType.c_type, # Input, opts[lopts].
        utils._BYREF_INT, # Input, lopts.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of optstr.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of cvalue.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _optstr = utils._EngineCharScalarType(
        optstr, {'fun_name': fun_name, 'entity': 'optstr'}, exp_chrlen=None,
    ).to_ctypes()
    _ivalue = utils._EngineIntScalarType.empty_ctypes()
    _rvalue = utils._EngineFloat64ScalarType.empty_ctypes()
    _cvalue = utils._EngineCharScalarType.empty_ctypes(chrlen=39)
    _optype = utils._EngineIntScalarType.empty_ctypes()
    utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        req_keys=('iopts', 'opts'),
    )
    _liopts_py = utils._EngineIntArrayType.get_shape(comm['iopts'], {'fun_name': fun_name, 'entity': 'comm[\'iopts\']'}, exp_rank=1, validate=False)[0]
    _liopts = utils._EngineIntScalarType(
        _liopts_py, {'fun_name': fun_name, 'entity': '_liopts'},
    ).to_ctypes()
    _iopts = comm['iopts']
    _lopts_py = utils._EngineFloat64ArrayType.get_shape(comm['opts'], {'fun_name': fun_name, 'entity': 'comm[\'opts\']'}, exp_rank=1, validate=False)[0]
    _lopts = utils._EngineIntScalarType(
        _lopts_py, {'fun_name': fun_name, 'entity': '_lopts'},
    ).to_ctypes()
    _opts = comm['opts']
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _optstr,
        _ivalue,
        _rvalue,
        _cvalue,
        _optype,
        _iopts,
        _liopts,
        _opts,
        _lopts,
        _errbuf,
        _ierr,
        _optstr._length_,
        _cvalue._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        _optstr_py = utils._EngineCharScalarType.to_py(_optstr)
        if _ierr_py == _nag_e_consts.IERR_OPT_UNKNOWN and _ifmt == _nag_e_consts.IERR_OPT_UNKNOWN:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the option in optstr was not recognized:',
                    'optstr = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_optstr_py),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == _nag_e_consts.IERR_OPTS_CORRUPT_GETTER and _ifmt == _nag_e_consts.IERR_OPTS_CORRUPT_GETTER:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, either the option arrays have not been initialized or they have',
                    'been corrupted.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _ivalue_py = utils._EngineIntScalarType.to_py(_ivalue)
        _rvalue_py = utils._EngineFloat64ScalarType.to_py(_rvalue)
        _optype_py = utils._EngineIntScalarType.to_py(_optype)
        if _optype_py == _nag_e_consts.OPT_TYPE_IDENTIFY:
            _optype_py = _nag_e_consts.OPT_TYPE_CHAR
            if _ivalue_py == _nag_g_consts.QLR_ID:
                _cvalue_py = 'g02qg'
            elif _ivalue_py == _nag_g_consts.JC_MNUM:
                _cvalue_py = 'g02jc'
            else:
                utils._EngineErrorExit(
                    fun_name=fun_name,
                    msg_lines=[
                        'The communication structure comm has either not been initialized,',
                        'has become corrupted, or is not compatible with this option',
                        'setting function.',
                    ],
                    errcodes=(_nag_e_consts.IERR_OPTS_CORRUPT_GETTER,),
                ).manage()
            _ivalue_py = 0
        else:
            _cvalue_py = utils._EngineCharScalarType.to_py(_cvalue)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['ivalue', 'rvalue', 'cvalue', 'optype'],
        )(
            _ivalue_py,
            _rvalue_py,
            _cvalue_py,
            _optype_py,
        )
    _err_class.manage()
    return return_data
