# -*- coding: utf-8 -*-
# pylint: disable=bare-except,chained-comparison,invalid-name,len-as-condition,line-too-long,protected-access,redefined-outer-name,too-many-arguments,too-many-boolean-expressions,too-many-branches,too-many-lines,too-many-locals,too-many-statements,unneeded-not,unused-import
u"""
Interfaces for the NAG Mark 27.0 `interp` Chapter.

``interp`` - Interpolation

Base interfaces.

We do not intend for these lower-level base interfaces to be called
directly from user code. Please refer to the ``naginterfaces.library``
subpackage instead.

This module is concerned with the interpolation of a function of one or
more variables.
When provided with the value of the function (and possibly one or more
of its lowest-order derivatives) at each of a number of values of the
variable(s), the NAG Library functions provide either an interpolating
function or an interpolated value.
For some of the interpolating functions, there are supporting NAG
Library functions to evaluate, differentiate or integrate them.

For full information please refer to the NAG Library document

https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e01/e01intro.html
"""

# NAG Copyright 2017-2020.

from __future__ import absolute_import

import collections as _collections
import logging as _logging
import sys as _sys

from math import sqrt

from ..base import utils

def dim1_aitken(a, b, c, n, x):
    u"""
    Interpolated values, Aitken's technique, unequally spaced data, one
    variable.

    ``dim1_aitken`` interpolates a function of one variable at a given
    point x from a table of function values y_i evaluated at equidistant
    or non-equidistant points x_i, for i = 1,2,...,n+1, using Aitken's
    technique of successive linear interpolations.

    For full information please refer to the NAG Library document for
    e01aa

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e01/e01aaf.html

    Parameters
    ----------
    a : float, shape (`n`+1), modified in place
        `On entry`: `a`[i-1] must contain the x-component of the ith
        data point, x_i, for i = 1,2,...,n+1.

        `On exit`: `a`[i-1] contains the value x_i-x, for i =
        1,2,...,n+1.

    b : float, shape (`n`+1), modified in place
        `On entry`: `b`[i-1] must contain the y-component (function
        value) of the ith data point, y_i, for i = 1,2,...,n+1.

        `On exit`: the contents of `b` are unspecified.

    c : float, shape (`n`*(`n`+1)/2), modified in place
        `On exit`:

            `c`[0],...,`c`[n-1] contain the first set of linear
            interpolations,

            `c`[n],...,`c`[2*n-2] contain the second set of linear
            interpolations,

            `c`[2n-1],...,`c`[3*n-4] contain the third set of linear
            interpolations,

            ...

            `c`[n*(n+1)/2-1] contains the interpolated function value at
            the point x.

    n : int
        The number of intervals which are to be used in interpolating
        the value at x; that is, there are n+1 data points (x_i,y_i).

    x : float
        The point x at which the interpolation is required. Note that
        `x` may lie outside the interval defined by the minimum and
        maximum values in `a`, in which case an extrapolated value will
        be computed; extrapolated results should be treated with
        considerable caution since there is no information on the
        behaviour of the function outside the defined interval.

    Raises
    ------
    NagValueError
        (`errno` 6)
            On entry, error in parameter `n`.

            Constraint: `n` > 0.

    References
    ----------
    Fröberg, C E, 1970, `Introduction to Numerical Analysis`,
    Addison--Wesley
    """

    excs = []
    fun_name = 'naginterfaces.base.interp.dim1_aitken'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('e01aaft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._EngineFloat64ArrayType.c_type, # Input/Output, a[n+1].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, b[n+1].
        utils._EngineFloat64ArrayType.c_type, # Output, c[n*(n+1)//2].
        utils._BYREF_INT, # Input, n.
        utils._BYREF_FLOAT64, # Input, x.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if not _n_py > 0:
        utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                utils._lapack_constraint_err_text('n > 0.'),
            ],
            errcodes=(6,),
            entity='n',
        ).manage()
    _a_cls = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_n_py+1,),
    )
    _a = _a_cls.to_ctypes()
    _b_cls = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=(_n_py+1,),
    )
    _b = _b_cls.to_ctypes()
    _c_cls = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_n_py*(_n_py+1)//2,),
        check_contents=False,
    )
    _c = _c_cls.to_ctypes(can_reference_contents=False)
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    engine_ctypes(
        _en_data,
        _a,
        _b,
        _c,
        _n,
        _x,
    )
    utils._handle_cb_exception(excs)
    _a_cls.refresh_py(_a)
    _b_cls.refresh_py(_b)
    _c_cls.refresh_py(_c)
    return_data = None
    return return_data

def dim1_everett(n, p, a, g):
    u"""
    Interpolated values, Everett's formula, equally spaced data, one
    variable.

    ``dim1_everett`` interpolates a function of one variable at a given
    point x from a table of function values evaluated at equidistant
    points, using Everett's formula.

    For full information please refer to the NAG Library document for
    e01ab

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e01/e01abf.html

    Parameters
    ----------
    n : int
        n, half the number of points to be used in the interpolation.

    p : float
        The point p at which the interpolated function value is
        required, i.e., p = (x-x_0)/h with -1.0 < p < 1.0.

    a : float, shape (2*`n`), modified in place
        `On entry`: `a`[i-1] must be set to the function value y_{i-n},
        for i = 1,2,...,2n.

        `On exit`: the contents of `a` are unspecified.

    g : float, shape (2*`n`+1), modified in place
        `On exit`: the array contains

        [table omitted]

        The interpolated function value y_p is stored in `g`[2n].

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `p` = *<value>*.

            Constraint: `p` <= 1.0.

        (`errno` 1)
            On entry, `p` = *<value>*.

            Constraint: `p` >= -1.0.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` > 0.

    References
    ----------
    Fröberg, C E, 1970, `Introduction to Numerical Analysis`,
    Addison--Wesley
    """

    excs = []
    fun_name = 'naginterfaces.base.interp.dim1_everett'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('e01abft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_FLOAT64, # Input, p.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, a[2*n].
        utils._EngineFloat64ArrayType.c_type, # Output, g[2*n+1].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _p = utils._EngineFloat64ScalarType(
        p, {'fun_name': fun_name, 'entity': 'p'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _a_cls = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(2*_n_py,),
    )
    _a = _a_cls.to_ctypes()
    _g_cls = utils._EngineFloat64ArrayType(
        g, {'fun_name': fun_name, 'entity': 'g'},
        exp_shape=(2*_n_py+1,),
        check_contents=False,
    )
    _g = _g_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _p,
        _a,
        _g,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, p = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: p <= 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, p = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: p >= -1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _a_cls.refresh_py(_a)
        _g_cls.refresh_py(_g)
        return_data = None
    _err_class.manage()
    return return_data

def dim1_cheb(m, xmin, xmax, x, y, ip, n, itmin, itmax, a, wrk, lwrk, iwrk, liwrk):
    u"""
    Interpolating functions, polynomial interpolant, data may include
    derivative values, one variable.

    ``dim1_cheb`` constructs the Chebyshev series representation of a
    polynomial interpolant to a set of data which may contain derivative
    values.

    For full information please refer to the NAG Library document for
    e01ae

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e01/e01aef.html

    Parameters
    ----------
    m : int
        m, the number of given values of the independent variable x.

    xmin : float
        The lower and upper end points, respectively, of the interval
        [x_min,x_max]. If they are not determined by your problem, it is
        recommended that they be set respectively to the smallest and
        largest values among the x_i.

    xmax : float
        The lower and upper end points, respectively, of the interval
        [x_min,x_max]. If they are not determined by your problem, it is
        recommended that they be set respectively to the smallest and
        largest values among the x_i.

    x : float, shape (`m`)
        The value of x_i, for i = 1,2,...,m. The `x`[i-1] need not be
        ordered.

    y : float, shape (`n`)
        The given values of the dependent variable, and derivatives, as
        follows:

        The first p_1+1 elements contain y_1,y_1^(1),...,y_1^(p_1) in
        that order.

        The next p_2+1 elements contain y_2,y_2^(1),...,y_2^(p_2) in
        that order.

        ...

        The last p_m+1 elements contain y_m,y_m^(1),...,y_m^(p_m) in
        that order.

    ip : int, shape (`m`)
        p_i, the order of the highest-order derivative whose value is
        given at x_i, for i = 1,2,...,m. If the value of y only is given
        for some x_i then the corresponding value of `ip`[i-1] must be
        zero.

    n : int
        n, the total number of interpolating conditions.

    itmin : int
        Respectively the minimum and maximum number of iterations to be
        performed by the function (for full details see [Further
        Comments]). Setting `itmin` and/or `itmax` negative or zero
        invokes default value(s) of 2 and/or 10, respectively.

        The default values will be satisfactory for most problems, but
        occasionally significant improvement will result from using
        higher values.

        Suggested value: `itmin` = 0 and `itmax` = 0.

    itmax : int
        Respectively the minimum and maximum number of iterations to be
        performed by the function (for full details see [Further
        Comments]). Setting `itmin` and/or `itmax` negative or zero
        invokes default value(s) of 2 and/or 10, respectively.

        The default values will be satisfactory for most problems, but
        occasionally significant improvement will result from using
        higher values.

        Suggested value: `itmin` = 0 and `itmax` = 0.

    a : float, shape (`n`), modified in place
        `On exit`: `a`[i] contains the coefficient a_i in the Chebyshev
        series representation of q(x), for i = 0,...,n-1.

    wrk : float, shape (`lwrk`), modified in place
        `On exit`: used as workspace, but see also [Further Comments].

    lwrk : int
        The dimension of the array `wrk`.

    iwrk : int, shape (`liwrk`), modified in place
        `On exit`: used as workspace, but see also [Further Comments].

    liwrk : int
        The dimension of the array `iwrk`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `liwrk` is too small. `liwrk` = *<value>*. Minimum
            possible dimension: *<value>*.

        (`errno` 1)
            On entry, `lwrk` is too small. `lwrk` = *<value>*. Minimum
            possible dimension: *<value>*.

        (`errno` 1)
            On entry, `n` = *<value>* and
            `m`+`ip`[0]+`ip`[1]+...+`ip`[`m`-1] = *<value>*.

            Constraint: `n` = `m`+`ip`[0]+`ip`[1]+...+`ip`[`m`-1].

        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 2)
            On entry, I = *<value>* and `ip`[I-1] = *<value>*.

            Constraint: `ip`[I-1] >= 0.

        (`errno` 3)
            On entry, I = *<value>*, J = *<value>* and `x`[I-1] =
            *<value>*.

            Constraint: `x`[I-1] != `x`[J-1].

        (`errno` 3)
            On entry, I = *<value>*, `x`[I-1] = *<value>*, `xmin` =
            *<value>* and `xmax` = *<value>*.

            Constraint: `xmin` <= `x`[I-1] <= `xmax`.

        (`errno` 3)
            On entry, `xmin` = *<value>* and `xmax` = *<value>*.

            Constraint: `xmin` < `xmax`.

        (`errno` 5)
            The computation has been terminated because the iterative
            process appears to be diverging.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 4)
            Not all the performance indices are less than eight times
            the machine precision, although `itmax` iterations have been
            performed. A more accurate solution may possibly be obtained
            by increasing `itmax` and recalling the function.

    References
    ----------
    Huddleston, R E, 1974, `CDC 6600 routines for the interpolation of
    data and of data with derivatives`, SLL-74-0214, Sandia Laboratories
    (Reprint)

    Krogh, F T, 1970, `Efficient algorithms for polynomial interpolation
    and numerical differentiation`, Math. Comput. (24), 185--190
    """

    excs = []
    fun_name = 'naginterfaces.base.interp.dim1_cheb'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('e01aeft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, m.
        utils._BYREF_FLOAT64, # Input, xmin.
        utils._BYREF_FLOAT64, # Input, xmax.
        utils._EngineFloat64ArrayType.c_type, # Input, x[m].
        utils._EngineFloat64ArrayType.c_type, # Input, y[n].
        utils._EngineIntArrayType.c_type, # Input, ip[m].
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, itmin.
        utils._BYREF_INT, # Input, itmax.
        utils._EngineFloat64ArrayType.c_type, # Output, a[n].
        utils._EngineFloat64ArrayType.c_type, # Output, wrk[lwrk].
        utils._BYREF_INT, # Input, lwrk.
        utils._EngineIntArrayType.c_type, # Output, iwrk[liwrk].
        utils._BYREF_INT, # Input, liwrk.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _xmin = utils._EngineFloat64ScalarType(
        xmin, {'fun_name': fun_name, 'entity': 'xmin'},
    ).to_ctypes()
    _xmax = utils._EngineFloat64ScalarType(
        xmax, {'fun_name': fun_name, 'entity': 'xmax'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _ip = utils._EngineIntArrayType(
        ip, {'fun_name': fun_name, 'entity': 'ip'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _itmin = utils._EngineIntScalarType(
        itmin, {'fun_name': fun_name, 'entity': 'itmin'},
    ).to_ctypes()
    _itmax = utils._EngineIntScalarType(
        itmax, {'fun_name': fun_name, 'entity': 'itmax'},
    ).to_ctypes()
    _a_cls = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _a = _a_cls.to_ctypes(can_reference_contents=False)
    _lwrk = utils._EngineIntScalarType(
        lwrk, {'fun_name': fun_name, 'entity': 'lwrk'},
    ).to_ctypes()
    _lwrk_py = utils._EngineIntScalarType.to_py(_lwrk)
    _wrk_cls = utils._EngineFloat64ArrayType(
        wrk, {'fun_name': fun_name, 'entity': 'wrk'},
        exp_shape=(_lwrk_py,),
        check_contents=False,
    )
    _wrk = _wrk_cls.to_ctypes(can_reference_contents=False)
    _liwrk = utils._EngineIntScalarType(
        liwrk, {'fun_name': fun_name, 'entity': 'liwrk'},
    ).to_ctypes()
    _liwrk_py = utils._EngineIntScalarType.to_py(_liwrk)
    _iwrk_cls = utils._EngineIntArrayType(
        iwrk, {'fun_name': fun_name, 'entity': 'iwrk'},
        exp_shape=(_liwrk_py,),
        check_contents=False,
    )
    _iwrk = _iwrk_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _m,
        _xmin,
        _xmax,
        _x,
        _y,
        _ip,
        _n,
        _itmin,
        _itmax,
        _a,
        _wrk,
        _lwrk,
        _iwrk,
        _liwrk,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, liwrk is too small.',
                    'liwrk = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Minimum possible dimension: {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lwrk is too small.',
                    'lwrk = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Minimum possible dimension: {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'and m+ip[0]+ip[1]+...+ip[m-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: n = m+ip[0]+ip[1]+...+ip[m-1].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, I = {:s} and ip[I-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: ip[I-1] >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, I = {:s}, J = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'x[I-1] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: x[I-1] != x[J-1].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, I = {:s}, x[I-1] = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'xmin = {:s} and xmax = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[4]),
                    ),
                    'Constraint: xmin <= x[I-1] <= xmax.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, xmin = {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'xmax = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: xmin < xmax.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99991:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Not all the performance indices are less than eight times the',
                    'machine precision, although itmax iterations have been performed.',
                    'A more accurate solution may possibly be obtained by increasing itmax and',
                    'recalling the function.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The computation has been terminated because the iterative process appears to',
                    'be diverging.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _a_cls.refresh_py(_a)
        _wrk_cls.refresh_py(_wrk)
        _iwrk_cls.refresh_py(_iwrk)
        return_data = None
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def dim1_spline(m, x, y, lamda, c, lck):
    u"""
    Interpolating functions, cubic spline interpolant, one variable.

    ``dim1_spline`` determines a cubic spline interpolant to a given set
    of data.

    For full information please refer to the NAG Library document for
    e01ba

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e01/e01baf.html

    Parameters
    ----------
    m : int
        m, the number of data points.

    x : float, shape (`m`)
        `x`[i-1] must be set to x_i, the ith data value of the
        independent variable x, for i = 1,2,...,m.

    y : float, shape (`m`)
        `y`[i-1] must be set to y_i, the ith data value of the dependent
        variable y, for i = 1,2,...,m.

    lamda : float, shape (`lck`), modified in place
        `On exit`: the value of lambda_i, the ith knot, for i =
        1,2,...,m+4.

    c : float, shape (`lck`), modified in place
        `On exit`: the coefficient c_i of the B-spline N_i(x), for i =
        1,2,...,m. The remaining elements of the array are not used.

    lck : int
        The dimension of the arrays `lamda` and `c`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `lck` = *<value>* and `m` = *<value>*.

            Constraint: `lck` >= `m`+4.

        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 4.

        (`errno` 2)
            On entry, I = *<value>*, `x`[I-1] = *<value>* and `x`[I-2] =
            *<value>*.

            Constraint: `x`[I-1] > `x`[I-2].

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Cox, M G, 1975, `An algorithm for spline interpolation`, J. Inst.
    Math. Appl. (15), 95--108

    Cox, M G, 1977, `A survey of numerical methods for data and function
    approximation`, The State of the Art in Numerical Analysis, (ed D A
    H Jacobs), 627--668, Academic Press
    """

    excs = []
    fun_name = 'naginterfaces.base.interp.dim1_spline'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('e01baft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, iemode.
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Input, x[m].
        utils._EngineFloat64ArrayType.c_type, # Input, y[m].
        utils._EngineFloat64ArrayType.c_type, # Output, lamda[lck].
        utils._EngineFloat64ArrayType.c_type, # Output, c[lck].
        utils._BYREF_INT, # Input, lck.
        utils._EngineFloat64ArrayType.c_type, # Workspace, wrk.
        utils._BYREF_INT, # Input, lwrk.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _iemode = utils._EngineIntScalarType.c_type(0)
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _lck = utils._EngineIntScalarType(
        lck, {'fun_name': fun_name, 'entity': 'lck'},
    ).to_ctypes()
    _lamda_cls = utils._EngineFloat64ArrayType(
        lamda, {'fun_name': fun_name, 'entity': 'lamda'},
        exp_shape=(_m_py+4,),
        check_contents=False,
    )
    _lamda = _lamda_cls.to_ctypes(can_reference_contents=False)
    _c_cls = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_m_py+4,),
        check_contents=False,
    )
    _c = _c_cls.to_ctypes(can_reference_contents=False)
    _wrk = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _lwrk = utils._EngineIntScalarType.c_type(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _iemode,
        _m,
        _x,
        _y,
        _lamda,
        _c,
        _lck,
        _wrk,
        _lwrk,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lck = {:s} and m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: lck >= m+4.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 4.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, I = {:s}, x[I-1] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'and x[I-2] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: x[I-1] > x[I-2].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _lamda_cls.refresh_py(_lamda)
        _c_cls.refresh_py(_c)
        return_data = None
    _err_class.manage()
    return return_data

def dim1_monotonic(n, x, f, d):
    u"""
    Interpolating functions, monotonicity-preserving, piecewise cubic
    Hermite, one variable.

    ``dim1_monotonic`` computes a monotonicity-preserving piecewise
    cubic Hermite interpolant to a set of data points.

    For full information please refer to the NAG Library document for
    e01be

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e01/e01bef.html

    Parameters
    ----------
    n : int
        n, the number of data points.

    x : float, shape (`n`)
        `x`[r-1] must be set to x_r, the rth value of the independent
        variable (abscissa), for r = 1,2,...,n.

    f : float, shape (`n`)
        `f`[r-1] must be set to f_r, the rth value of the dependent
        variable (ordinate), for r = 1,2,...,n.

    d : float, shape (`n`), modified in place
        `On exit`: estimates of derivatives at the data points. `d`[r-1]
        contains the derivative at `x`[r-1].

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 2.

        (`errno` 2)
            On entry, r = *<value>*, `x`[r-2] = *<value>* and `x`[r-1] =
            *<value>*.

            Constraint: `x`[r-2] < `x`[r-1] for all r.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Fritsch, F N, 1982, `PCHIP final specifications`, Report UCID-30194,
    Lawrence Livermore National Laboratory

    Fritsch, F N and Butland, J, 1984, `A method for constructing local
    monotone piecewise cubic interpolants`, SIAM J. Sci. Statist.
    Comput. (5), 300--304
    """

    excs = []
    fun_name = 'naginterfaces.base.interp.dim1_monotonic'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('e01beft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[n].
        utils._EngineFloat64ArrayType.c_type, # Input, f[n].
        utils._EngineFloat64ArrayType.c_type, # Output, d[n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _f = utils._EngineFloat64ArrayType(
        f, {'fun_name': fun_name, 'entity': 'f'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _d_cls = utils._EngineFloat64ArrayType(
        d, {'fun_name': fun_name, 'entity': 'd'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _d = _d_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _x,
        _f,
        _d,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, r = {:s}, x[r-2] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'and x[r-1] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: x[r-2] < x[r-1] for all r.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _d_cls.refresh_py(_d)
        return_data = None
    _err_class.manage()
    return return_data

def dim1_monotonic_eval(n, x, f, d, m, px, pf):
    u"""
    Interpolated values, interpolant computed by ``dim1_monotonic``,
    function only, one variable.

    ``dim1_monotonic_eval`` evaluates a piecewise cubic Hermite
    interpolant at a set of points.

    For full information please refer to the NAG Library document for
    e01bf

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e01/e01bff.html

    Parameters
    ----------
    n : int
        `n`, `x`, `f` and `d` must be unchanged from the previous call
        of ``dim1_monotonic``.

    x : float, shape (`n`)
        `n`, `x`, `f` and `d` must be unchanged from the previous call
        of ``dim1_monotonic``.

    f : float, shape (`n`)
        `n`, `x`, `f` and `d` must be unchanged from the previous call
        of ``dim1_monotonic``.

    d : float, shape (`n`)
        `n`, `x`, `f` and `d` must be unchanged from the previous call
        of ``dim1_monotonic``.

    m : int
        m, the number of points at which the interpolant is to be
        evaluated. If any point lies outside the interval from `x`[0])
        to `x`[`n`-1], a value is extrapolated from the nearest extreme
        cubic, and a warning is returned. The extrapolation simply
        extends the final cubic at each end.

    px : float, shape (`m`)
        The m values of x at which the interpolant is to be evaluated.

    pf : float, shape (`m`), modified in place
        `On exit`: `pf`[i-1] contains the value of the interpolant
        evaluated at the point `px`[i-1], for i = 1,2,...,m.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 2.

        (`errno` 2)
            On entry, r = *<value>*, `x`[r-2] = *<value>* and `x`[r-1] =
            *<value>*.

            Constraint: `x`[r-2] < `x`[r-1] for all r.

        (`errno` 3)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 4)
            Warning -- some points in array `px` lie outside the range
            `x`[0]...`x`[`n`-1]. Values at these points are unreliable
            because computed by extrapolation.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Fritsch, F N, 1982, `PCHIP final specifications`, Report UCID-30194,
    Lawrence Livermore National Laboratory
    """

    excs = []
    fun_name = 'naginterfaces.base.interp.dim1_monotonic_eval'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('e01bfft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[n].
        utils._EngineFloat64ArrayType.c_type, # Input, f[n].
        utils._EngineFloat64ArrayType.c_type, # Input, d[n].
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Input, px[m].
        utils._EngineFloat64ArrayType.c_type, # Output, pf[m].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _f = utils._EngineFloat64ArrayType(
        f, {'fun_name': fun_name, 'entity': 'f'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _d = utils._EngineFloat64ArrayType(
        d, {'fun_name': fun_name, 'entity': 'd'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _px = utils._EngineFloat64ArrayType(
        px, {'fun_name': fun_name, 'entity': 'px'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _pf_cls = utils._EngineFloat64ArrayType(
        pf, {'fun_name': fun_name, 'entity': 'pf'},
        exp_shape=(_m_py,),
        check_contents=False,
    )
    _pf = _pf_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _x,
        _f,
        _d,
        _m,
        _px,
        _pf,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, r = {:s}, x[r-2] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'and x[r-1] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: x[r-2] < x[r-1] for all r.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99996:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Warning -- some points in array px lie outside the range x[0]...x[n-1].',
                    'Values at these points are unreliable because computed by extrapolation.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _pf_cls.refresh_py(_pf)
        return_data = None
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def dim1_monotonic_deriv(n, x, f, d, m, px, pf, pd):
    u"""
    Interpolated values, interpolant computed by ``dim1_monotonic``,
    function and first derivative, one variable.

    ``dim1_monotonic_deriv`` evaluates a piecewise cubic Hermite
    interpolant and its first derivative at a set of points.

    For full information please refer to the NAG Library document for
    e01bg

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e01/e01bgf.html

    Parameters
    ----------
    n : int
        `n`, `x`, `f` and `d` must be unchanged from the previous call
        of ``dim1_monotonic``.

    x : float, shape (`n`)
        `n`, `x`, `f` and `d` must be unchanged from the previous call
        of ``dim1_monotonic``.

    f : float, shape (`n`)
        `n`, `x`, `f` and `d` must be unchanged from the previous call
        of ``dim1_monotonic``.

    d : float, shape (`n`)
        `n`, `x`, `f` and `d` must be unchanged from the previous call
        of ``dim1_monotonic``.

    m : int
        m, the number of points at which the interpolant is to be
        evaluated.

    px : float, shape (`m`)
        The m values of x at which the interpolant is to be evaluated.

    pf : float, shape (`m`), modified in place
        `On exit`: `pf`[i-1] contains the value of the interpolant
        evaluated at the point `px`[i-1], for i = 1,2,...,m.

    pd : float, shape (`m`), modified in place
        `On exit`: `pd`[i-1] contains the first derivative of the
        interpolant evaluated at the point `px`[i-1], for i = 1,2,...,m.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 2.

        (`errno` 2)
            On entry, r = *<value>*, `x`[r-2] = *<value>* and `x`[r-1] =
            *<value>*.

            Constraint: `x`[r-2] < `x`[r-1] for all r.

        (`errno` 3)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 4)
            Warning -- some points in array `px` lie outside the range
            `x`[0]...`x`[`n`-1]. Values at these points are unreliable
            because computed by extrapolation.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Fritsch, F N, 1982, `PCHIP final specifications`, Report UCID-30194,
    Lawrence Livermore National Laboratory
    """

    excs = []
    fun_name = 'naginterfaces.base.interp.dim1_monotonic_deriv'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('e01bgft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[n].
        utils._EngineFloat64ArrayType.c_type, # Input, f[n].
        utils._EngineFloat64ArrayType.c_type, # Input, d[n].
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Input, px[m].
        utils._EngineFloat64ArrayType.c_type, # Output, pf[m].
        utils._EngineFloat64ArrayType.c_type, # Output, pd[m].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _f = utils._EngineFloat64ArrayType(
        f, {'fun_name': fun_name, 'entity': 'f'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _d = utils._EngineFloat64ArrayType(
        d, {'fun_name': fun_name, 'entity': 'd'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _px = utils._EngineFloat64ArrayType(
        px, {'fun_name': fun_name, 'entity': 'px'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _pf_cls = utils._EngineFloat64ArrayType(
        pf, {'fun_name': fun_name, 'entity': 'pf'},
        exp_shape=(_m_py,),
        check_contents=False,
    )
    _pf = _pf_cls.to_ctypes(can_reference_contents=False)
    _pd_cls = utils._EngineFloat64ArrayType(
        pd, {'fun_name': fun_name, 'entity': 'pd'},
        exp_shape=(_m_py,),
        check_contents=False,
    )
    _pd = _pd_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _x,
        _f,
        _d,
        _m,
        _px,
        _pf,
        _pd,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, r = {:s}, x[r-2] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'and x[r-1] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: x[r-2] < x[r-1] for all r.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99996:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Warning -- some points in array px lie outside the range x[0]...x[n-1].',
                    'Values at these points are unreliable because computed by extrapolation.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _pf_cls.refresh_py(_pf)
        _pd_cls.refresh_py(_pd)
        return_data = None
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def dim1_monotonic_intg(n, x, f, d, a, b):
    u"""
    Interpolated values, interpolant computed by ``dim1_monotonic``,
    definite integral, one variable.

    ``dim1_monotonic_intg`` evaluates the definite integral of a
    piecewise cubic Hermite interpolant over the interval [a,b].

    For full information please refer to the NAG Library document for
    e01bh

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e01/e01bhf.html

    Parameters
    ----------
    n : int
        `n`, `x`, `f` and `d` must be unchanged from the previous call
        of ``dim1_monotonic``.

    x : float, shape (`n`)
        `n`, `x`, `f` and `d` must be unchanged from the previous call
        of ``dim1_monotonic``.

    f : float, shape (`n`)
        `n`, `x`, `f` and `d` must be unchanged from the previous call
        of ``dim1_monotonic``.

    d : float, shape (`n`)
        `n`, `x`, `f` and `d` must be unchanged from the previous call
        of ``dim1_monotonic``.

    a : float
        The interval [a,b] over which integration is to be performed.

    b : float
        The interval [a,b] over which integration is to be performed.

    Returns
    -------
    pint : float
        The value of the definite integral of the interpolant over the
        interval [a,b].

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 2.

        (`errno` 2)
            On entry, r = *<value>*, `x`[r-2] = *<value>* and `x`[r-1] =
            *<value>*.

            Constraint: `x`[r-2] < `x`[r-1] for all r.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            Warning -- either `a` or `b` is outside the range
            `x`[0]...`x`[`n`-1]. The result has been computed by
            extrapolation and is unreliable. `a` = *<value>* `b` =
            *<value>*.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Fritsch, F N, 1982, `PCHIP final specifications`, Report UCID-30194,
    Lawrence Livermore National Laboratory
    """

    excs = []
    fun_name = 'naginterfaces.base.interp.dim1_monotonic_intg'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('e01bhft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[n].
        utils._EngineFloat64ArrayType.c_type, # Input, f[n].
        utils._EngineFloat64ArrayType.c_type, # Input, d[n].
        utils._BYREF_FLOAT64, # Input, a.
        utils._BYREF_FLOAT64, # Input, b.
        utils._BYREF_FLOAT64, # Output, pint.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _f = utils._EngineFloat64ArrayType(
        f, {'fun_name': fun_name, 'entity': 'f'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _d = utils._EngineFloat64ArrayType(
        d, {'fun_name': fun_name, 'entity': 'd'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _a = utils._EngineFloat64ScalarType(
        a, {'fun_name': fun_name, 'entity': 'a'},
    ).to_ctypes()
    _b = utils._EngineFloat64ScalarType(
        b, {'fun_name': fun_name, 'entity': 'b'},
    ).to_ctypes()
    _pint = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _x,
        _f,
        _d,
        _a,
        _b,
        _pint,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, r = {:s}, x[r-2] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'and x[r-1] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: x[r-2] < x[r-1] for all r.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99997:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Warning -- either a or b is outside the range x[0]...x[n-1].',
                    'The result has been computed by extrapolation and is unreliable.',
                    'a = {:s} b = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _pint_py = utils._EngineFloat64ScalarType.to_py(_pint)
        return_data = _pint_py
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = _collections.namedtuple(
                'nag_return_data',
                ['pint'],
            )(
                _pint_py,
            )
    _err_class.manage()
    return return_data

def dim1_monconv_disc(n, lam, negfor, yfor, x, y):
    u"""
    Interpolating variables, monotone convex Hagan--West procedure, one
    variable.

    ``dim1_monconv_disc`` computes, for a given set of data points, the
    forward values and other values required for monotone convex
    interpolation as defined in Hagan and West (2008).
    This form of interpolation is particularly suited to the
    construction of yield curves in Financial Mathematics but can be
    applied to any data where it is desirable to preserve both
    monotonicity and convexity.

    For full information please refer to the NAG Library document for
    e01ce

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e01/e01cef.html

    Parameters
    ----------
    n : int
        n, the number of data points.

    lam : float
        lambda, the amelioration (smoothing) parameter. Forward rates
        are first computed using (2) and then, if lambda > 0, a limiting
        filter is applied which depends on neighbouring discrete forward
        values. This filter has a smoothing effect on the curve that
        increases with lambda.

        Suggested value: lambda = 0.2.

    negfor : bool
        Determines whether or not to allow negative forward rates.

        `negfor` = True

            Negative forward rates are permitted.

        `negfor` = False

            Forward rates calculated must be non-negative.

    yfor : bool
        Determines whether the array `y` contains values, y, or discrete
        forward rates f^d.

        `yfor` = True

            `y` contains the discrete forward rates f_i^d, for i =
            1,2,...,n.

        `yfor` = False

            `y` contains the values y_i, for i = 1,2,...,n.

    x : float, shape (`n`)
        x, the (possibly unordered) set of data points.

    y : float, shape (`n`)
        If `yfor` = True, the discrete forward rates f_i^d corresponding
        to the data points x_i, for i = 1,2,...,n.

        If `yfor` = False, the data values y_i corresponding to the data
        points x_i, for i = 1,2,...,n.

    Returns
    -------
    comm : dict, communication object
        Communication structure.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 2.

        (`errno` 2)
            On entry, `lam` = *<value>*.

            Constraint: 0.0 <= `lam` <= 1.0.

        (`errno` 3)
            On entry, `x` contains duplicate data points.

    References
    ----------
    Hagan, P S and West, G, 2006, `Interpolation methods for curve
    construction`, Applied Mathematical Finance (13(2)), 89--129

    Hagan, P S and West, G, 2008, `Methods for constructing a yield
    curve`, WILLMOTT Magazine (May), 70--81

    West, G, 2011, `The monotone convex method of interpolation`,
    Financial Modelling Agency
    """

    excs = []
    fun_name = 'naginterfaces.base.interp.dim1_monconv_disc'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('e01ceft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_FLOAT64, # Input, lam.
        utils._BYREF_LOGICAL, # Input, negfor.
        utils._BYREF_LOGICAL, # Input, yfor.
        utils._EngineFloat64ArrayType.c_type, # Input, x[n].
        utils._EngineFloat64ArrayType.c_type, # Input, y[n].
        utils._EngineFloat64ArrayType.c_type, # Output, comm[4*n+10].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _lam = utils._EngineFloat64ScalarType(
        lam, {'fun_name': fun_name, 'entity': 'lam'},
    ).to_ctypes()
    _negfor = utils._EngineLogicalScalarType(
        negfor, {'fun_name': fun_name, 'entity': 'negfor'},
    ).to_ctypes()
    _yfor = utils._EngineLogicalScalarType(
        yfor, {'fun_name': fun_name, 'entity': 'yfor'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _comm = utils._EngineFloat64ArrayType.empty_ctypes(4*_n_py+10)
    _comm_py = {}
    _comm_py['comm'] = _comm
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _lam,
        _negfor,
        _yfor,
        _x,
        _y,
        _comm,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lam = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: 0.0 <= lam <= 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x contains duplicate data points.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        return_data = _comm_py
    _err_class.manage()
    return return_data

def dim1_monconv_eval(m, x, val, fwd, comm):
    u"""
    Interpolated values, variables computed by ``dim1_monconv_disc``,
    monotone convex Hagan--West procedure, one variable.

    ``dim1_monconv_eval`` evaluates a monotonic convex interpolant at a
    set of points.

    For full information please refer to the NAG Library document for
    e01cf

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e01/e01cff.html

    Parameters
    ----------
    m : int
        m, the number of points at which the interpolant is to be
        evaluated.

    x : float, shape (`m`)
        x, the points at which the interpolant is to be evaluated.

    val : float, shape (`m`), modified in place
        `On exit`: the values of the interpolant at x.

    fwd : float, shape (`m`), modified in place
        `On exit`: the values of the forward rates at x.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``dim1_monconv_disc``.

    Raises
    ------
    NagValueError
        (`errno` 1)
            Either ``dim1_monconv_disc`` was not called first or the
            communication array has become corrupted.

    References
    ----------
    Hagan, P S and West, G, 2006, `Interpolation methods for curve
    construction`, Applied Mathematical Finance (13(2)), 89--129

    Hagan, P S and West, G, 2008, `Methods for constructing a yield
    curve`, WILLMOTT Magazine (May), 70--81

    West, G, 2011, `The monotone convex method of interpolation`,
    Financial Modelling Agency
    """

    excs = []
    fun_name = 'naginterfaces.base.interp.dim1_monconv_eval'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('e01cfft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Input, x[m].
        utils._EngineFloat64ArrayType.c_type, # Output, val[m].
        utils._EngineFloat64ArrayType.c_type, # Output, fwd[m].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, comm[*].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _val_cls = utils._EngineFloat64ArrayType(
        val, {'fun_name': fun_name, 'entity': 'val'},
        exp_shape=(_m_py,),
        check_contents=False,
    )
    _val = _val_cls.to_ctypes(can_reference_contents=False)
    _fwd_cls = utils._EngineFloat64ArrayType(
        fwd, {'fun_name': fun_name, 'entity': 'fwd'},
        exp_shape=(_m_py,),
        check_contents=False,
    )
    _fwd = _fwd_cls.to_ctypes(can_reference_contents=False)
    utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        req_keys=('comm',),
    )
    _vl_lcomm_py = 4*int(utils._EngineFloat64ArrayType.get_element(comm['comm'], {'fun_name': fun_name, 'entity': 'comm[\'comm\']'}, (0,), validate=False))+10
    _comm = comm['comm']
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _m,
        _x,
        _val,
        _fwd,
        _comm,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Either dim1_monconv_disc was not called first or the communication array',
                    'has become corrupted.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _val_cls.refresh_py(_val)
        _fwd_cls.refresh_py(_fwd)
        return_data = None
    _err_class.manage()
    return return_data

def dim2_spline_grid(mx, my, x, y, f, lamda, mu, c):
    u"""
    Interpolating functions, fitting bicubic spline, data on rectangular
    grid.

    ``dim2_spline_grid`` computes a bicubic spline interpolating surface
    through a set of data values, given on a rectangular grid in the x-y
    plane.

    For full information please refer to the NAG Library document for
    e01da

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e01/e01daf.html

    Parameters
    ----------
    mx : int
        `mx` and `my` must specify m_x and m_y respectively, the number
        of points along the x and y axis that define the rectangular
        grid.

    my : int
        `mx` and `my` must specify m_x and m_y respectively, the number
        of points along the x and y axis that define the rectangular
        grid.

    x : float, shape (`mx`)
        `x`[q-1] and `y`[r-1] must contain x_q, for q = 1,2,...,m_x, and
        y_r, for r = 1,2,...,m_y, respectively.

    y : float, shape (`my`)
        `x`[q-1] and `y`[r-1] must contain x_q, for q = 1,2,...,m_x, and
        y_r, for r = 1,2,...,m_y, respectively.

    f : float, shape (`mx`*`my`)
        `f`[m_y*(q-1)+r-1] must contain f_{q,r}, for r = 1,2,...,m_y,
        for q = 1,2,...,m_x.

    lamda : float, shape (`mx`+4), modified in place
        `lamda` contains the complete set of knots lambda_i associated
        with the x variable

    mu : float, shape (`my`+4), modified in place
        `mu` contains the complete set of knots mu_i associated with the
        y variable

    c : float, shape (`mx`*`my`), modified in place
        `On exit`: the coefficients of the spline interpolant.
        `c`[m_y*(i-1)+j-1] contains the coefficient c_{ij} described in
        [Description].

    Returns
    -------
    px : int
        `px` and `py` contain m_x+4 and m_y+4, the total number of knots
        of the computed spline with respect to the x and y variables,
        respectively.

    py : int
        `px` and `py` contain m_x+4 and m_y+4, the total number of knots
        of the computed spline with respect to the x and y variables,
        respectively.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `my` = *<value>*.

            Constraint: `my` >= 4.

        (`errno` 1)
            On entry, `mx` = *<value>*.

            Constraint: `mx` >= 4.

        (`errno` 2)
            On entry, the `x` or the `y` mesh points are not in strictly
            ascending order.

        (`errno` 3)
            An intermediate set of linear equations is singular -- the
            data is too ill-conditioned to compute B-spline
            coefficients.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Anthony, G T, Cox, M G and Hayes, J G, 1982, `DASL -- Data
    Approximation Subroutine Library`, National Physical Laboratory

    Cox, M G, 1975, `An algorithm for spline interpolation`, J. Inst.
    Math. Appl. (15), 95--108

    de Boor, C, 1972, `On calculating with B-splines`, J. Approx. Theory
    (6), 50--62

    Hayes, J G and Halliday, J, 1974, `The least squares fitting of
    cubic spline surfaces to general data sets`, J. Inst. Math. Appl.
    (14), 89--103
    """

    excs = []
    fun_name = 'naginterfaces.base.interp.dim2_spline_grid'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('e01daft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, mx.
        utils._BYREF_INT, # Input, my.
        utils._EngineFloat64ArrayType.c_type, # Input, x[mx].
        utils._EngineFloat64ArrayType.c_type, # Input, y[my].
        utils._EngineFloat64ArrayType.c_type, # Input, f[mx*my].
        utils._BYREF_INT, # Output, px.
        utils._BYREF_INT, # Output, py.
        utils._EngineFloat64ArrayType.c_type, # Output, lamda[mx+4].
        utils._EngineFloat64ArrayType.c_type, # Output, mu[my+4].
        utils._EngineFloat64ArrayType.c_type, # Output, c[mx*my].
        utils._EngineFloat64ArrayType.c_type, # Workspace, wrk.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _mx = utils._EngineIntScalarType(
        mx, {'fun_name': fun_name, 'entity': 'mx'},
    ).to_ctypes()
    _my = utils._EngineIntScalarType(
        my, {'fun_name': fun_name, 'entity': 'my'},
    ).to_ctypes()
    _mx_py = utils._EngineIntScalarType.to_py(_mx)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_mx_py,),
    ).to_ctypes()
    _my_py = utils._EngineIntScalarType.to_py(_my)
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_my_py,),
    ).to_ctypes()
    _f = utils._EngineFloat64ArrayType(
        f, {'fun_name': fun_name, 'entity': 'f'},
        exp_shape=(_mx_py*_my_py,),
    ).to_ctypes()
    _px = utils._EngineIntScalarType.empty_ctypes()
    _py = utils._EngineIntScalarType.empty_ctypes()
    _lamda_cls = utils._EngineFloat64ArrayType(
        lamda, {'fun_name': fun_name, 'entity': 'lamda'},
        exp_shape=(_mx_py+4,),
        check_contents=False,
    )
    _lamda = _lamda_cls.to_ctypes(can_reference_contents=False)
    _mu_cls = utils._EngineFloat64ArrayType(
        mu, {'fun_name': fun_name, 'entity': 'mu'},
        exp_shape=(_my_py+4,),
        check_contents=False,
    )
    _mu = _mu_cls.to_ctypes(can_reference_contents=False)
    _c_cls = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_mx_py*_my_py,),
        check_contents=False,
    )
    _c = _c_cls.to_ctypes(can_reference_contents=False)
    _wrk = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _mx,
        _my,
        _x,
        _y,
        _f,
        _px,
        _py,
        _lamda,
        _mu,
        _c,
        _wrk,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, my = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: my >= 4.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mx = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mx >= 4.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the x or the y mesh points are not in strictly ascending order.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'An intermediate set of linear equations is singular',
                    '-- the data is too ill-conditioned to compute B-spline coefficients.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _lamda_cls.refresh_py(_lamda)
        _mu_cls.refresh_py(_mu)
        _c_cls.refresh_py(_c)
        _px_py = utils._EngineIntScalarType.to_py(_px)
        _py_py = utils._EngineIntScalarType.to_py(_py)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['px', 'py'],
        )(
            _px_py,
            _py_py,
        )
    _err_class.manage()
    return return_data

def dim2_triangulate(n, x, y, triang):
    u"""
    Triangulation of two-dimensional scattered grid, method of Renka and
    Cline.

    ``dim2_triangulate`` generates a triangulation for a given set of
    two-dimensional points using the method of Renka and Cline.

    For full information please refer to the NAG Library document for
    e01ea

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e01/e01eaf.html

    Parameters
    ----------
    n : int
        n, the number of data points.

    x : float, shape (`n`)
        The x coordinates of the n data points.

    y : float, shape (`n`)
        The y coordinates of the n data points.

    triang : int, shape (7*`n`), modified in place
        `On exit`: a data structure defining the computed triangulation,
        in a form suitable for passing to ``dim2_triang_bary_eval``.
        Details of how the triangulation is encoded in `triang` are
        given in [Further Comments]. These details are most likely to be
        of use when plotting the computed triangulation.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 3.

        (`errno` 2)
            On entry, all the (x,y) pairs are collinear.

    References
    ----------
    Cline, A K and Renka, R L, 1984, `A storage-efficient method for
    construction of a Thiessen triangulation`, Rocky Mountain J. Math.
    (14), 119--139

    Lawson, C L, 1977, `Software for C^1 surface interpolation`,
    Mathematical Software III, (ed J R Rice), 161--194, Academic Press

    Renka, R L, 1984, `Algorithm 624: triangulation and interpolation of
    arbitrarily distributed points in the plane`, ACM Trans. Math.
    Software (10), 440--442

    Renka, R L and Cline, A K, 1984, `A triangle-based C^1 interpolation
    method`, Rocky Mountain J. Math. (14), 223--237
    """

    excs = []
    fun_name = 'naginterfaces.base.interp.dim2_triangulate'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('e01eaft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[n].
        utils._EngineFloat64ArrayType.c_type, # Input, y[n].
        utils._EngineIntArrayType.c_type, # Output, triang[7*n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _triang_cls = utils._EngineIntArrayType(
        triang, {'fun_name': fun_name, 'entity': 'triang'},
        exp_shape=(7*_n_py,),
        check_contents=False,
    )
    _triang = _triang_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _x,
        _y,
        _triang,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT.format(_n_py),
                    ),
                    'Constraint: n >= 3.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, all the (x,y) pairs are collinear.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _triang_cls.refresh_py(_triang)
        return_data = None
    _err_class.manage()
    return return_data

def dim2_triang_bary_eval(m, n, x, y, f, triang, px, py, pf):
    u"""
    Barycentric interpolation on function values provided on a
    two-dimensional scattered grid.

    ``dim2_triang_bary_eval`` performs barycentric interpolation, at a
    given set of points, using a set of function values on a scattered
    grid and a triangulation of that grid computed by
    ``dim2_triangulate``.

    For full information please refer to the NAG Library document for
    e01eb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e01/e01ebf.html

    Parameters
    ----------
    m : int
        m, the number of points to interpolate.

    n : int
        n, the number of data points. `n` must be unchanged from the
        previous call of ``dim2_triangulate``.

    x : float, shape (`n`)
        The coordinates of the rth data point, (x_r,y_r), for r =
        1,2,...,n. `x` and `y` must be unchanged from the previous call
        of ``dim2_triangulate``.

    y : float, shape (`n`)
        The coordinates of the rth data point, (x_r,y_r), for r =
        1,2,...,n. `y` and `y` must be unchanged from the previous call
        of ``dim2_triangulate``.

    f : float, shape (`n`)
        The function values f_r at (x_r,y_r), for r = 1,2,...,n.

    triang : int, shape (7*`n`)
        The triangulation computed by the previous call of
        ``dim2_triangulate``. See [Further Comments] for details of how
        the triangulation used is encoded in `triang`.

    px : float, shape (`m`)
        The coordinates (px_i,py_i), for i = 1,2,...,m, at which
        interpolated function values are sought.

    py : float, shape (`m`)
        The coordinates (px_i,py_i), for i = 1,2,...,m, at which
        interpolated function values are sought.

    pf : float, shape (`m`), modified in place
        `On exit`: the interpolated values F(px_i,py_i), for i =
        1,2,...,m.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 3.

        (`errno` 2)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 3)
            On entry, the triangulation information held in the array
            `triang` does not specify a valid triangulation of the data
            points. `triang` has been corrupted since the call to
            ``dim2_triangulate``.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 4)
            At least one evaluation point lies outside the nodal
            triangulation. For each such point the value returned in
            `pf` is that corresponding to a node on the closest boundary
            line segment.

    References
    ----------
    Cline, A K and Renka, R L, 1984, `A storage-efficient method for
    construction of a Thiessen triangulation`, Rocky Mountain J. Math.
    (14), 119--139

    Lawson, C L, 1977, `Software for C^1 surface interpolation`,
    Mathematical Software III, (ed J R Rice), 161--194, Academic Press

    Renka, R L, 1984, `Algorithm 624: triangulation and interpolation of
    arbitrarily distributed points in the plane`, ACM Trans. Math.
    Software (10), 440--442

    Renka, R L and Cline, A K, 1984, `A triangle-based C^1 interpolation
    method`, Rocky Mountain J. Math. (14), 223--237
    """

    excs = []
    fun_name = 'naginterfaces.base.interp.dim2_triang_bary_eval'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('e01ebft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[n].
        utils._EngineFloat64ArrayType.c_type, # Input, y[n].
        utils._EngineFloat64ArrayType.c_type, # Input, f[n].
        utils._EngineIntArrayType.c_type, # Input, triang[7*n].
        utils._EngineFloat64ArrayType.c_type, # Input, px[m].
        utils._EngineFloat64ArrayType.c_type, # Input, py[m].
        utils._EngineFloat64ArrayType.c_type, # Output, pf[m].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _f = utils._EngineFloat64ArrayType(
        f, {'fun_name': fun_name, 'entity': 'f'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _triang = utils._EngineIntArrayType(
        triang, {'fun_name': fun_name, 'entity': 'triang'},
        exp_shape=(7*_n_py,),
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _px = utils._EngineFloat64ArrayType(
        px, {'fun_name': fun_name, 'entity': 'px'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _py = utils._EngineFloat64ArrayType(
        py, {'fun_name': fun_name, 'entity': 'py'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _pf_cls = utils._EngineFloat64ArrayType(
        pf, {'fun_name': fun_name, 'entity': 'pf'},
        exp_shape=(_m_py,),
        check_contents=False,
    )
    _pf = _pf_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _m,
        _n,
        _x,
        _y,
        _f,
        _triang,
        _px,
        _py,
        _pf,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT.format(_n_py),
                    ),
                    'Constraint: n >= 3.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the triangulation information held in the array triang does not',
                    'specify a valid triangulation of the data points. triang has been',
                    'corrupted since the call to dim2_triangulate.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'At least one evaluation point lies outside the nodal triangulation. For each',
                    'such point the value returned in pf is that corresponding',
                    'to a node on the closest boundary line segment.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _pf_cls.refresh_py(_pf)
        return_data = None
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def dim1_ratnl(n, x, f, a, u):
    u"""
    Interpolating functions, rational interpolant, one variable.

    ``dim1_ratnl`` produces, from a set of function values and
    corresponding abscissae, the coefficients of an interpolating
    rational function expressed in continued fraction form.

    For full information please refer to the NAG Library document for
    e01ra

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e01/e01raf.html

    Parameters
    ----------
    n : int
        n, the number of data points.

    x : float, shape (`n`)
        `x`[i-1] must be set to the value of the ith data abscissa, x_i,
        for i = 1,2,...,n.

    f : float, shape (`n`)
        `f`[i-1] must be set to the value of the data ordinate, f_i,
        corresponding to x_i, for i = 1,2,...,n.

    a : float, shape (`n`), modified in place
        `On exit`: `a`[j-1] contains the value of the parameter a_j in
        R(x), for j = 1,2,...,m. The remaining elements of `a`, if any,
        are set to zero.

    u : float, shape (`n`), modified in place
        `On exit`: `u`[j-1] contains the value of the parameter u_j in
        R(x), for j = 1,2,...,m-1. The u_j are a permuted subset of the
        elements of `x`. The remaining n-m+1 locations contain a
        permutation of the remaining x_i, which can be ignored.

    Returns
    -------
    m : int
        m, the number of terms in the continued fraction representation
        of R(x).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` > 0.

        (`errno` 2)
            On entry, `x`[I-1] is very close to `x`[J-1]: I = *<value>*,
            `x`[I-1] = *<value>*, J = *<value>* and `x`[J-1] =
            *<value>*.

        (`errno` 3)
            A continued fraction of the required form does not exist.

    References
    ----------
    Graves--Morris, P R and Hopkins, T R, 1981, `Reliable rational
    interpolation`, Numer. Math. (36), 111--128
    """

    excs = []
    fun_name = 'naginterfaces.base.interp.dim1_ratnl'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('e01raft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[n].
        utils._EngineFloat64ArrayType.c_type, # Input, f[n].
        utils._BYREF_INT, # Output, m.
        utils._EngineFloat64ArrayType.c_type, # Output, a[n].
        utils._EngineFloat64ArrayType.c_type, # Output, u[n].
        utils._EngineIntArrayType.c_type, # Workspace, iw.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _f = utils._EngineFloat64ArrayType(
        f, {'fun_name': fun_name, 'entity': 'f'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _m = utils._EngineIntScalarType.empty_ctypes()
    _a_cls = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _a = _a_cls.to_ctypes(can_reference_contents=False)
    _u_cls = utils._EngineFloat64ArrayType(
        u, {'fun_name': fun_name, 'entity': 'u'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _u = _u_cls.to_ctypes(can_reference_contents=False)
    _iw = utils._EngineIntArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _x,
        _f,
        _m,
        _a,
        _u,
        _iw,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, x[I-1] is very close to x[J-1]:',
                    'I = {:s}, x[I-1] = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'J = {:s} and x[J-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[4]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'A continued fraction of the required form does not exist.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _a_cls.refresh_py(_a)
        _u_cls.refresh_py(_u)
        _m_py = utils._EngineIntScalarType.to_py(_m)
        return_data = _m_py
    _err_class.manage()
    return return_data

def dim1_ratnl_eval(m, a, u, x):
    u"""
    Interpolated values, evaluate rational interpolant computed by
    ``dim1_ratnl``, one variable.

    ``dim1_ratnl_eval`` evaluates continued fractions of the form
    produced by ``dim1_ratnl``.

    For full information please refer to the NAG Library document for
    e01rb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e01/e01rbf.html

    Parameters
    ----------
    m : int
        m, the number of terms in the continued fraction.

    a : float, shape (`m`)
        `a`[j-1] must be set to the value of the parameter a_j in the
        continued fraction, for j = 1,2,...,m.

    u : float, shape (`m`)
        `u`[j-1] must be set to the value of the parameter u_j in the
        continued fraction, for j = 1,2,...,m-1. (The element `u`[m-1]
        is not used).

    x : float
        The value of x at which the continued fraction is to be
        evaluated.

    Returns
    -------
    f : float
        The value of the continued fraction corresponding to the value
        of x.

    Raises
    ------
    NagValueError
        (`errno` 1)
            `x` corresponds to a pole of R(x), or is very close. `x` =
            *<value>*.

    References
    ----------
    Graves--Morris, P R and Hopkins, T R, 1981, `Reliable rational
    interpolation`, Numer. Math. (36), 111--128
    """

    excs = []
    fun_name = 'naginterfaces.base.interp.dim1_ratnl_eval'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('e01rbft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Input, a[m].
        utils._EngineFloat64ArrayType.c_type, # Input, u[m].
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_FLOAT64, # Output, f.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _u = utils._EngineFloat64ArrayType(
        u, {'fun_name': fun_name, 'entity': 'u'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _x = utils._EngineFloat64ScalarType(
        x, {'fun_name': fun_name, 'entity': 'x'},
    ).to_ctypes()
    _f = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _m,
        _a,
        _u,
        _x,
        _f,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'x corresponds to a pole of R(x), or is very close.',
                    'x = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _f_py = utils._EngineFloat64ScalarType.to_py(_f)
        return_data = _f_py
    _err_class.manage()
    return return_data

def dim2_scat(m, x, y, f, triang, grads):
    u"""
    Interpolating functions, method of Renka and Cline, two variables.

    ``dim2_scat`` generates a two-dimensional surface interpolating a
    set of scattered data points, using the method of Renka and Cline.

    For full information please refer to the NAG Library document for
    e01sa

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e01/e01saf.html

    Parameters
    ----------
    m : int
        m, the number of data points.

    x : float, shape (`m`)
        The coordinates of the rth data point, for r = 1,2,...,m. The
        data points are accepted in any order, but see [Further
        Comments].

    y : float, shape (`m`)
        The coordinates of the rth data point, for r = 1,2,...,m. The
        data points are accepted in any order, but see [Further
        Comments].

    f : float, shape (`m`)
        The coordinates of the rth data point, for r = 1,2,...,m. The
        data points are accepted in any order, but see [Further
        Comments].

    triang : int, shape (7*`m`), modified in place
        `On exit`: a data structure defining the computed triangulation,
        in a form suitable for passing to ``dim2_scat_eval``.

    grads : float, shape (2, `m`), modified in place
        `On exit`: the estimated partial derivatives at the nodes, in a
        form suitable for passing to ``dim2_scat_eval``. The derivatives
        at node r with respect to x and y are contained in
        `grads`[0,r-1] and `grads`[1,r-1] respectively, for r =
        1,2,...,m.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 3.

        (`errno` 2)
            All nodes are collinear. There is no unique solution.

        (`errno` 3)
            On entry, (`x`[I-1],`y`[I-1]) = (`x`[J-1],`y`[J-1]), for I,J
            = *<value>**<value>*, `x`[I-1], `y`[I-1] =
            *<value>**<value>*.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Cline, A K and Renka, R L, 1984, `A storage-efficient method for
    construction of a Thiessen triangulation`, Rocky Mountain J. Math.
    (14), 119--139

    Lawson, C L, 1977, `Software for C^1 surface interpolation`,
    Mathematical Software III, (ed J R Rice), 161--194, Academic Press

    Renka, R L, 1984, `Algorithm 624: triangulation and interpolation of
    arbitrarily distributed points in the plane`, ACM Trans. Math.
    Software (10), 440--442

    Renka, R L and Cline, A K, 1984, `A triangle-based C^1 interpolation
    method`, Rocky Mountain J. Math. (14), 223--237
    """

    excs = []
    fun_name = 'naginterfaces.base.interp.dim2_scat'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('e01saft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Input, x[m].
        utils._EngineFloat64ArrayType.c_type, # Input, y[m].
        utils._EngineFloat64ArrayType.c_type, # Input, f[m].
        utils._EngineIntArrayType.c_type, # Output, triang[7*m].
        utils._EngineFloat64ArrayType.c_type, # Output, grads[2, m].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((grads, 'grads'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _f = utils._EngineFloat64ArrayType(
        f, {'fun_name': fun_name, 'entity': 'f'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _triang_cls = utils._EngineIntArrayType(
        triang, {'fun_name': fun_name, 'entity': 'triang'},
        exp_shape=(7*_m_py,),
        check_contents=False,
    )
    _triang = _triang_cls.to_ctypes(can_reference_contents=False)
    _grads_cls = utils._EngineFloat64ArrayType(
        grads, {'fun_name': fun_name, 'entity': 'grads'},
        exp_shape=(2, _m_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _grads = _grads_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _m,
        _x,
        _y,
        _f,
        _triang,
        _grads,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 3.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'All nodes are collinear.',
                    'There is no unique solution.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, (x[I-1],y[I-1]) = (x[J-1],y[J-1]),',
                    'for I,J = {:s}{:s}, x[I-1],'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'y[I-1] = {:s}{:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[4]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _triang_cls.refresh_py(_triang)
        _grads_cls.refresh_py(_grads)
        return_data = None
    _err_class.manage()
    return return_data

def dim2_scat_eval(m, x, y, f, triang, grads, px, py, ist):
    u"""
    Interpolated values, evaluate interpolant computed by ``dim2_scat``,
    two variables.

    ``dim2_scat_eval`` evaluates at a given point the two-dimensional
    interpolant function computed by ``dim2_scat``.

    For full information please refer to the NAG Library document for
    e01sb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e01/e01sbf.html

    Parameters
    ----------
    m : int
        `m` must be unchanged from the previous call of ``dim2_scat``

    x : float, shape (`m`)
        `x` must be unchanged from the previous call of ``dim2_scat``

    y : float, shape (`m`)
        `y` must be unchanged from the previous call of ``dim2_scat``

    f : float, shape (`m`)
        `f` must be unchanged from the previous call of ``dim2_scat``

    triang : int, shape (7*`m`)
        `triang` must be unchanged from the previous call of
        ``dim2_scat``

    grads : float, shape (2, `m`)
        `grads` must be unchanged from the previous call of
        ``dim2_scat``

    px : float
        The point (px,py) at which the interpolant is to be evaluated.

    py : float
        The point (px,py) at which the interpolant is to be evaluated.

    ist : int
        The index of the starting node in the search for a triangle
        containing the point (px,py). On the first call to
        ``dim2_scat_eval``, `ist` must be set to 1. For efficiency on
        subsequent calls to ``dim2_scat_eval`` an updated value of `ist`
        as returned by ``dim2_scat_eval`` may be supplied instead. An
        input value outside the range 1 <= `ist` <= `m` will be treated
        as 1.

    Returns
    -------
    ist : int
        The index of one of the vertices of the triangle containing the
        point (px,py).

    pf : float
        The value of the interpolant evaluated at the point (px,py).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 3.

        (`errno` 2)
            On entry, `triang` does not contain a valid data point
            triangulation; `triang` may have been corrupted since the
            call to ``dim2_scat``.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            Warning -- the evaluation point (*<value>*,*<value>*) lies
            outside the triangulation boundary. The returned value was
            computed by extrapolation.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Renka, R L, 1984, `Algorithm 624: triangulation and interpolation of
    arbitrarily distributed points in the plane`, ACM Trans. Math.
    Software (10), 440--442

    Renka, R L and Cline, A K, 1984, `A triangle-based C^1 interpolation
    method`, Rocky Mountain J. Math. (14), 223--237
    """

    excs = []
    fun_name = 'naginterfaces.base.interp.dim2_scat_eval'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('e01sbft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Input, x[m].
        utils._EngineFloat64ArrayType.c_type, # Input, y[m].
        utils._EngineFloat64ArrayType.c_type, # Input, f[m].
        utils._EngineIntArrayType.c_type, # Input, triang[7*m].
        utils._EngineFloat64ArrayType.c_type, # Input, grads[2, m].
        utils._BYREF_FLOAT64, # Input, px.
        utils._BYREF_FLOAT64, # Input, py.
        utils._BYREF_INT, # Input/Output, ist.
        utils._BYREF_FLOAT64, # Output, pf.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((grads, 'grads'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _f = utils._EngineFloat64ArrayType(
        f, {'fun_name': fun_name, 'entity': 'f'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _triang = utils._EngineIntArrayType(
        triang, {'fun_name': fun_name, 'entity': 'triang'},
        exp_shape=(7*_m_py,),
    ).to_ctypes()
    _grads = utils._EngineFloat64ArrayType(
        grads, {'fun_name': fun_name, 'entity': 'grads'},
        exp_shape=(2, _m_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _px = utils._EngineFloat64ScalarType(
        px, {'fun_name': fun_name, 'entity': 'px'},
    ).to_ctypes()
    _py = utils._EngineFloat64ScalarType(
        py, {'fun_name': fun_name, 'entity': 'py'},
    ).to_ctypes()
    _ist = utils._EngineIntScalarType(
        ist, {'fun_name': fun_name, 'entity': 'ist'},
    ).to_ctypes()
    _pf = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _m,
        _x,
        _y,
        _f,
        _triang,
        _grads,
        _px,
        _py,
        _ist,
        _pf,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 3.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, triang does not contain a valid data point triangulation;',
                    'triang may have been corrupted since the call to dim2_scat.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99997:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Warning -- the evaluation point',
                    '({:s},{:s}) lies outside the'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'triangulation boundary. The returned value was computed by extrapolation.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _ist_py = utils._EngineIntScalarType.to_py(_ist)
        _pf_py = utils._EngineFloat64ScalarType.to_py(_pf)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['ist', 'pf'],
        )(
            _ist_py,
            _pf_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def dim2_scat_shep(m, x, y, f, nw, nq, iq, liq, rq, lrq):
    u"""
    Interpolating functions, modified Shepard's method, two variables.

    ``dim2_scat_shep`` generates a two-dimensional interpolant to a set
    of scattered data points, using a modified Shepard method.

    For full information please refer to the NAG Library document for
    e01sg

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e01/e01sgf.html

    Parameters
    ----------
    m : int
        m, the number of data points.

    x : float, shape (`m`)
        The Cartesian coordinates of the data points (x_r,y_r), for r =
        1,2,...,m.

    y : float, shape (`m`)
        The Cartesian coordinates of the data points (x_r,y_r), for r =
        1,2,...,m.

    f : float, shape (`m`)
        `f`[r-1] must be set to the data value f_r, for r = 1,2,...,m.

    nw : int
        The number N_w of data points that determines each radius of
        influence R_w, appearing in the definition of each of the
        weights w_r, for r = 1,2,...,m (see [Description]). Note that
        R_w is different for each weight. If `nw` <= 0 the default value
        `nw` = min(19, `m`-1) is used instead.

    nq : int
        The number N_q of data points to be used in the least squares
        fit for coefficients defining the nodal functions q_r(x,y) (see
        [Description]). If `nq` <= 0 the default value `nq` = min(13,
        `m`-1) is used instead.

    iq : int, shape (`liq`), modified in place
        `On exit`: integer data defining the interpolant Q(x,y).

    liq : int
        The dimension of the array `iq`.

    rq : float, shape (`lrq`), modified in place
        `On exit`: real data defining the interpolant Q(x,y).

    lrq : int
        The dimension of the array `rq`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `lrq` is too small: `lrq` = *<value>*.

        (`errno` 1)
            On entry, `liq` is too small: `liq` = *<value>*.

        (`errno` 1)
            On entry, `nw` = *<value>* and `m` = *<value>*.

            Constraint: `nw` <= min(40, `m`-1).

        (`errno` 1)
            On entry, `nq` = *<value>* and `m` = *<value>*.

            Constraint: `nq` <= min(40, `m`-1).

        (`errno` 1)
            On entry, `nq` = *<value>*.

            Constraint: `nq` <= 0 or `nq` >= 5.

        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 6.

        (`errno` 2)
            There are duplicate nodes in the dataset.
            (`x`[I-1],`y`[I-1]) = (`x`[J-1],`y`[J-1]), for I = *<value>*
            and J = *<value>*. The interpolant cannot be derived.

        (`errno` 3)
            All nodes are collinear. There is no unique solution.

    References
    ----------
    Bentley, J L and Friedman, J H, 1979, `Data structures for range
    searching`, ACM Comput. Surv. (11), 397--409

    Franke, R and Nielson, G, 1980, `Smooth interpolation of large sets
    of scattered data`, Internat. J. Num. Methods Engrg. (15),
    1691--1704

    Renka, R J, 1988, `Algorithm 660: QSHEP2D: Quadratic Shepard method
    for bivariate interpolation of scattered data`, ACM Trans. Math.
    Software (14), 149--150

    Renka, R J, 1988, `Multivariate interpolation of large sets of
    scattered data`, ACM Trans. Math. Software (14), 139--148

    Shepard, D, 1968, `A two-dimensional interpolation function for
    irregularly spaced data`, Proc. 23rd Nat. Conf. ACM, 517--523,
    Brandon/Systems Press Inc., Princeton
    """

    excs = []
    fun_name = 'naginterfaces.base.interp.dim2_scat_shep'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('e01sgft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Input, x[m].
        utils._EngineFloat64ArrayType.c_type, # Input, y[m].
        utils._EngineFloat64ArrayType.c_type, # Input, f[m].
        utils._BYREF_INT, # Input, nw.
        utils._BYREF_INT, # Input, nq.
        utils._EngineIntArrayType.c_type, # Output, iq[liq].
        utils._BYREF_INT, # Input, liq.
        utils._EngineFloat64ArrayType.c_type, # Output, rq[lrq].
        utils._BYREF_INT, # Input, lrq.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _f = utils._EngineFloat64ArrayType(
        f, {'fun_name': fun_name, 'entity': 'f'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _nw = utils._EngineIntScalarType(
        nw, {'fun_name': fun_name, 'entity': 'nw'},
    ).to_ctypes()
    _nq = utils._EngineIntScalarType(
        nq, {'fun_name': fun_name, 'entity': 'nq'},
    ).to_ctypes()
    _liq = utils._EngineIntScalarType(
        liq, {'fun_name': fun_name, 'entity': 'liq'},
    ).to_ctypes()
    _iq_cls = utils._EngineIntArrayType(
        iq, {'fun_name': fun_name, 'entity': 'iq'},
        exp_shape=(2*_m_py+1,),
        check_contents=False,
    )
    _iq = _iq_cls.to_ctypes(can_reference_contents=False)
    _lrq = utils._EngineIntScalarType(
        lrq, {'fun_name': fun_name, 'entity': 'lrq'},
    ).to_ctypes()
    _rq_cls = utils._EngineFloat64ArrayType(
        rq, {'fun_name': fun_name, 'entity': 'rq'},
        exp_shape=(6*_m_py+5,),
        check_contents=False,
    )
    _rq = _rq_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _m,
        _x,
        _y,
        _f,
        _nw,
        _nq,
        _iq,
        _liq,
        _rq,
        _lrq,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lrq is too small: lrq = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, liq is too small: liq = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nw = {:s} and m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nw <= min(40, m-1).',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nq = {:s} and m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nq <= min(40, m-1).',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nq = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nq <= 0 or nq >= 5.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 6.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'There are duplicate nodes in the dataset.',
                    '(x[I-1],y[I-1]) = (x[J-1],y[J-1]), for',
                    'I = {:s} and J = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'The interpolant cannot be derived.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'All nodes are collinear.',
                    'There is no unique solution.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _iq_cls.refresh_py(_iq)
        _rq_cls.refresh_py(_rq)
        return_data = None
    _err_class.manage()
    return return_data

def dim2_scat_shep_eval(m, x, y, f, iq, liq, rq, lrq, n, u, v, q, qx, qy):
    u"""
    Interpolated values, evaluate interpolant computed by
    ``dim2_scat_shep``, function and first derivatives, two variables.

    ``dim2_scat_shep_eval`` evaluates the two-dimensional interpolating
    function generated by ``dim2_scat_shep`` and its first partial
    derivatives.

    For full information please refer to the NAG Library document for
    e01sh

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e01/e01shf.html

    Parameters
    ----------
    m : int
        `m`, `x`, `y` and `f` must be the same values as were supplied
        in the preceding call to ``dim2_scat_shep``.

    x : float, shape (`m`)
        `m`, `x`, `y` and `f` must be the same values as were supplied
        in the preceding call to ``dim2_scat_shep``.

    y : float, shape (`m`)
        `m`, `x`, `y` and `f` must be the same values as were supplied
        in the preceding call to ``dim2_scat_shep``.

    f : float, shape (`m`)
        `m`, `x`, `y` and `f` must be the same values as were supplied
        in the preceding call to ``dim2_scat_shep``.

    iq : int, shape (`liq`)
        Must be unchanged from the value returned from a previous call
        to ``dim2_scat_shep``.

    liq : int
        The dimension of the array `iq`.

    rq : float, shape (`lrq`)
        Must be unchanged from the value returned from a previous call
        to ``dim2_scat_shep``.

    lrq : int
        The dimension of the array `rq`.

    n : int
        n, the number of evaluation points.

    u : float, shape (`n`)
        The evaluation points (u_i,v_i), for i = 1,2,...,n.

    v : float, shape (`n`)
        The evaluation points (u_i,v_i), for i = 1,2,...,n.

    q : float, shape (`n`), modified in place
        `On exit`: the values of the interpolant at (u_i,v_i), for i =
        1,2,...,n. If any of these evaluation points lie outside the
        region of definition of the interpolant the corresponding
        entries in `q` are set to the largest machine representable
        number (see ``machine.real_largest``), and
        ``dim2_scat_shep_eval`` returns with `errno` = 3.

    qx : float, shape (`n`), modified in place
        `On exit`: the values of the partial derivatives of the
        interpolant Q(x,y) at (u_i,v_i), for i = 1,2,...,n. If any of
        these evaluation points lie outside the region of definition of
        the interpolant, the corresponding entries in `qx` and `qy` are
        set to the largest machine representable number (see
        ``machine.real_largest``), and ``dim2_scat_shep_eval`` returns
        with `errno` = 3.

    qy : float, shape (`n`), modified in place
        `On exit`: the values of the partial derivatives of the
        interpolant Q(x,y) at (u_i,v_i), for i = 1,2,...,n. If any of
        these evaluation points lie outside the region of definition of
        the interpolant, the corresponding entries in `qx` and `qy` are
        set to the largest machine representable number (see
        ``machine.real_largest``), and ``dim2_scat_shep_eval`` returns
        with `errno` = 3.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `lrq` is too small: `lrq` = *<value>*.

        (`errno` 1)
            On entry, `liq` is too small: `liq` = *<value>*.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 6.

        (`errno` 2)
            On entry, values in `rq` appear to be invalid. Check that
            `rq` has not been corrupted between calls to
            ``dim2_scat_shep`` and ``dim2_scat_shep_eval``.

        (`errno` 2)
            On entry, values in `iq` appear to be invalid. Check that
            `iq` has not been corrupted between calls to
            ``dim2_scat_shep`` and ``dim2_scat_shep_eval``.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            On entry, at least one evaluation point lies outside the
            region of definition of the interpolant. At such points the
            corresponding values in `q` and `qx` contain extrapolated
            approximations. Points should be evaluated one by one to
            identify extrapolated values.

    References
    ----------
    Renka, R J, 1988, `Algorithm 660: QSHEP2D: Quadratic Shepard method
    for bivariate interpolation of scattered data`, ACM Trans. Math.
    Software (14), 149--150
    """

    excs = []
    fun_name = 'naginterfaces.base.interp.dim2_scat_shep_eval'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('e01shft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Input, x[m].
        utils._EngineFloat64ArrayType.c_type, # Input, y[m].
        utils._EngineFloat64ArrayType.c_type, # Input, f[m].
        utils._EngineIntArrayType.c_type, # Input, iq[liq].
        utils._BYREF_INT, # Input, liq.
        utils._EngineFloat64ArrayType.c_type, # Input, rq[lrq].
        utils._BYREF_INT, # Input, lrq.
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, u[n].
        utils._EngineFloat64ArrayType.c_type, # Input, v[n].
        utils._EngineFloat64ArrayType.c_type, # Output, q[n].
        utils._EngineFloat64ArrayType.c_type, # Output, qx[n].
        utils._EngineFloat64ArrayType.c_type, # Output, qy[n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _f = utils._EngineFloat64ArrayType(
        f, {'fun_name': fun_name, 'entity': 'f'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _liq = utils._EngineIntScalarType(
        liq, {'fun_name': fun_name, 'entity': 'liq'},
    ).to_ctypes()
    _iq = utils._EngineIntArrayType(
        iq, {'fun_name': fun_name, 'entity': 'iq'},
        exp_shape=(2*_m_py+1,),
    ).to_ctypes()
    _lrq = utils._EngineIntScalarType(
        lrq, {'fun_name': fun_name, 'entity': 'lrq'},
    ).to_ctypes()
    _rq = utils._EngineFloat64ArrayType(
        rq, {'fun_name': fun_name, 'entity': 'rq'},
        exp_shape=(6*_m_py+5,),
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _u = utils._EngineFloat64ArrayType(
        u, {'fun_name': fun_name, 'entity': 'u'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _v = utils._EngineFloat64ArrayType(
        v, {'fun_name': fun_name, 'entity': 'v'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _q_cls = utils._EngineFloat64ArrayType(
        q, {'fun_name': fun_name, 'entity': 'q'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _q = _q_cls.to_ctypes(can_reference_contents=False)
    _qx_cls = utils._EngineFloat64ArrayType(
        qx, {'fun_name': fun_name, 'entity': 'qx'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _qx = _qx_cls.to_ctypes(can_reference_contents=False)
    _qy_cls = utils._EngineFloat64ArrayType(
        qy, {'fun_name': fun_name, 'entity': 'qy'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _qy = _qy_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _m,
        _x,
        _y,
        _f,
        _iq,
        _liq,
        _rq,
        _lrq,
        _n,
        _u,
        _v,
        _q,
        _qx,
        _qy,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lrq is too small: lrq = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, liq is too small: liq = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 6.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, values in rq appear to be invalid.',
                    'Check that rq has not been corrupted between calls to dim2_scat_shep and dim2_scat_shep_eval.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, values in iq appear to be invalid.',
                    'Check that iq has not been corrupted between calls to dim2_scat_shep and dim2_scat_shep_eval.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99993:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, at least one evaluation point lies outside the region of',
                    'definition of the interpolant. At such points the corresponding',
                    'values in q and qx contain extrapolated approximations. Points',
                    'should be evaluated one by one to identify extrapolated values.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _q_cls.refresh_py(_q)
        _qx_cls.refresh_py(_qx)
        _qy_cls.refresh_py(_qy)
        return_data = None
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def dim3_scat_shep(m, x, y, z, f, nw, nq, iq, liq, rq, lrq):
    u"""
    Interpolating functions, modified Shepard's method, three variables.

    ``dim3_scat_shep`` generates a three-dimensional interpolant to a
    set of scattered data points, using a modified Shepard method.

    For full information please refer to the NAG Library document for
    e01tg

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e01/e01tgf.html

    Parameters
    ----------
    m : int
        m, the number of data points.

    x : float, shape (`m`)
        `x`[r-1], `y`[r-1], `z`[r-1] must be set to the Cartesian
        coordinates of the data point (x_r,y_r,z_r), for r = 1,2,...,m.

    y : float, shape (`m`)
        `x`[r-1], `y`[r-1], `z`[r-1] must be set to the Cartesian
        coordinates of the data point (x_r,y_r,z_r), for r = 1,2,...,m.

    z : float, shape (`m`)
        `x`[r-1], `y`[r-1], `z`[r-1] must be set to the Cartesian
        coordinates of the data point (x_r,y_r,z_r), for r = 1,2,...,m.

    f : float, shape (`m`)
        `f`[r-1] must be set to the data value f_r, for r = 1,2,...,m.

    nw : int
        The number N_w of data points that determines each radius of
        influence R_w, appearing in the definition of each of the
        weights w_r, for r = 1,2,...,m (see [Description]). Note that
        R_w is different for each weight. If `nw` <= 0 the default value
        `nw` = min(32, `m`-1) is used instead.

    nq : int
        The number N_q of data points to be used in the least squares
        fit for coefficients defining the nodal functions q_r(x,y,z)
        (see [Description]). If `nq` <= 0 the default value `nq` =
        min(17, `m`-1) is used instead.

    iq : int, shape (`liq`), modified in place
        `On exit`: integer data defining the interpolant Q(x,y,z).

    liq : int
        The dimension of the array `iq`.

    rq : float, shape (`lrq`), modified in place
        `On exit`: real data defining the interpolant Q(x,y,z).

    lrq : int
        The dimension of the array `rq`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `lrq` is too small: `lrq` = *<value>*.

        (`errno` 1)
            On entry, `liq` is too small: `liq` = *<value>*.

        (`errno` 1)
            On entry, `nw` = *<value>* and `m` = *<value>*.

            Constraint: `nw` <= min(40, `m`-1).

        (`errno` 1)
            On entry, `nq` = *<value>* and `m` = *<value>*.

            Constraint: `nq` <= min(40, `m`-1).

        (`errno` 1)
            On entry, `nq` = *<value>*.

            Constraint: `nq` <= 0 or `nq` >= 9.

        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 10.

        (`errno` 2)
            There are duplicate nodes in the dataset.
            (`x`[I-1],`y`[I-1],`z`[I-1]) = (`x`[J-1],`y`[J-1],`z`[J-1])
            for: I = *<value>* and J = *<value>*. The interpolant cannot
            be derived.

        (`errno` 3)
            All nodes are coplanar. There is no unique solution.

    References
    ----------
    Bentley, J L and Friedman, J H, 1979, `Data structures for range
    searching`, ACM Comput. Surv. (11), 397--409

    Franke, R and Nielson, G, 1980, `Smooth interpolation of large sets
    of scattered data`, Internat. J. Num. Methods Engrg. (15),
    1691--1704

    Renka, R J, 1988, `Multivariate interpolation of large sets of
    scattered data`, ACM Trans. Math. Software (14), 139--148

    Renka, R J, 1988, `Algorithm 661: QSHEP3D: Quadratic Shepard method
    for trivariate interpolation of scattered data`, ACM Trans. Math.
    Software (14), 151--152

    Shepard, D, 1968, `A two-dimensional interpolation function for
    irregularly spaced data`, Proc. 23rd Nat. Conf. ACM, 517--523,
    Brandon/Systems Press Inc., Princeton
    """

    excs = []
    fun_name = 'naginterfaces.base.interp.dim3_scat_shep'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('e01tgft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Input, x[m].
        utils._EngineFloat64ArrayType.c_type, # Input, y[m].
        utils._EngineFloat64ArrayType.c_type, # Input, z[m].
        utils._EngineFloat64ArrayType.c_type, # Input, f[m].
        utils._BYREF_INT, # Input, nw.
        utils._BYREF_INT, # Input, nq.
        utils._EngineIntArrayType.c_type, # Output, iq[liq].
        utils._BYREF_INT, # Input, liq.
        utils._EngineFloat64ArrayType.c_type, # Output, rq[lrq].
        utils._BYREF_INT, # Input, lrq.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _z = utils._EngineFloat64ArrayType(
        z, {'fun_name': fun_name, 'entity': 'z'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _f = utils._EngineFloat64ArrayType(
        f, {'fun_name': fun_name, 'entity': 'f'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _nw = utils._EngineIntScalarType(
        nw, {'fun_name': fun_name, 'entity': 'nw'},
    ).to_ctypes()
    _nq = utils._EngineIntScalarType(
        nq, {'fun_name': fun_name, 'entity': 'nq'},
    ).to_ctypes()
    _liq = utils._EngineIntScalarType(
        liq, {'fun_name': fun_name, 'entity': 'liq'},
    ).to_ctypes()
    _iq_cls = utils._EngineIntArrayType(
        iq, {'fun_name': fun_name, 'entity': 'iq'},
        exp_shape=(2*_m_py+1,),
        check_contents=False,
    )
    _iq = _iq_cls.to_ctypes(can_reference_contents=False)
    _lrq = utils._EngineIntScalarType(
        lrq, {'fun_name': fun_name, 'entity': 'lrq'},
    ).to_ctypes()
    _rq_cls = utils._EngineFloat64ArrayType(
        rq, {'fun_name': fun_name, 'entity': 'rq'},
        exp_shape=(10*_m_py+7,),
        check_contents=False,
    )
    _rq = _rq_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _m,
        _x,
        _y,
        _z,
        _f,
        _nw,
        _nq,
        _iq,
        _liq,
        _rq,
        _lrq,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lrq is too small: lrq = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, liq is too small: liq = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nw = {:s} and m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nw <= min(40, m-1).',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nq = {:s} and m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nq <= min(40, m-1).',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nq = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nq <= 0 or nq >= 9.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 10.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'There are duplicate nodes in the dataset.',
                    '(x[I-1],y[I-1],z[I-1]) = (x[J-1],y[J-1],z[J-1]) for:',
                    'I = {:s} and J = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'The interpolant cannot be derived.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'All nodes are coplanar.',
                    'There is no unique solution.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _iq_cls.refresh_py(_iq)
        _rq_cls.refresh_py(_rq)
        return_data = None
    _err_class.manage()
    return return_data

def dim3_scat_shep_eval(m, x, y, z, f, iq, liq, rq, lrq, n, u, v, w, q, qx, qy, qz):
    u"""
    Interpolated values, evaluate interpolant computed by
    ``dim3_scat_shep``, function and first derivatives, three variables.

    ``dim3_scat_shep_eval`` evaluates the three-dimensional
    interpolating function generated by ``dim3_scat_shep`` and its first
    partial derivatives.

    For full information please refer to the NAG Library document for
    e01th

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e01/e01thf.html

    Parameters
    ----------
    m : int
        `m`, `x`, `y`, `z` and `f` must be the same values as were
        supplied in the preceding call to ``dim3_scat_shep``.

    x : float, shape (`m`)
        `m`, `x`, `y`, `z` and `f` must be the same values as were
        supplied in the preceding call to ``dim3_scat_shep``.

    y : float, shape (`m`)
        `m`, `x`, `y`, `z` and `f` must be the same values as were
        supplied in the preceding call to ``dim3_scat_shep``.

    z : float, shape (`m`)
        `m`, `x`, `y`, `z` and `f` must be the same values as were
        supplied in the preceding call to ``dim3_scat_shep``.

    f : float, shape (`m`)
        `m`, `x`, `y`, `z` and `f` must be the same values as were
        supplied in the preceding call to ``dim3_scat_shep``.

    iq : int, shape (`liq`)
        Must be unchanged from the value returned from a previous call
        to ``dim3_scat_shep``.

    liq : int
        The dimension of the array `iq`.

    rq : float, shape (`lrq`)
        Must be unchanged from the value returned from a previous call
        to ``dim3_scat_shep``.

    lrq : int
        The dimension of the array `rq`.

    n : int
        n, the number of evaluation points.

    u : float, shape (`n`)
        `u`[i-1], `v`[i-1], `w`[i-1] must be set to the evaluation point
        (u_i,v_i,w_i), for i = 1,2,...,n.

    v : float, shape (`n`)
        `u`[i-1], `v`[i-1], `w`[i-1] must be set to the evaluation point
        (u_i,v_i,w_i), for i = 1,2,...,n.

    w : float, shape (`n`)
        `u`[i-1], `v`[i-1], `w`[i-1] must be set to the evaluation point
        (u_i,v_i,w_i), for i = 1,2,...,n.

    q : float, shape (`n`), modified in place
        `On exit`: `q`[i-1] contains the value of the interpolant, at
        (u_i,v_i,w_i), for i = 1,2,...,n. If any of these evaluation
        points lie outside the region of definition of the interpolant
        the corresponding entries in `q` are set to the largest machine
        representable number (see ``machine.real_largest``), and
        ``dim3_scat_shep_eval`` returns with `errno` = 3.

    qx : float, shape (`n`), modified in place
        `On exit`: `qx`[i-1], `qy`[i-1], `qz`[i-1] contains the value of
        the partial derivatives of the interpolant Q(x,y,z) at
        (u_i,v_i,w_i), for i = 1,2,...,n. If any of these evaluation
        points lie outside the region of definition of the interpolant,
        the corresponding entries in `qx`, `qy` and `qz` are set to the
        largest machine representable number (see
        ``machine.real_largest``), and ``dim3_scat_shep_eval`` returns
        with `errno` = 3.

    qy : float, shape (`n`), modified in place
        `On exit`: `qx`[i-1], `qy`[i-1], `qz`[i-1] contains the value of
        the partial derivatives of the interpolant Q(x,y,z) at
        (u_i,v_i,w_i), for i = 1,2,...,n. If any of these evaluation
        points lie outside the region of definition of the interpolant,
        the corresponding entries in `qx`, `qy` and `qz` are set to the
        largest machine representable number (see
        ``machine.real_largest``), and ``dim3_scat_shep_eval`` returns
        with `errno` = 3.

    qz : float, shape (`n`), modified in place
        `On exit`: `qx`[i-1], `qy`[i-1], `qz`[i-1] contains the value of
        the partial derivatives of the interpolant Q(x,y,z) at
        (u_i,v_i,w_i), for i = 1,2,...,n. If any of these evaluation
        points lie outside the region of definition of the interpolant,
        the corresponding entries in `qx`, `qy` and `qz` are set to the
        largest machine representable number (see
        ``machine.real_largest``), and ``dim3_scat_shep_eval`` returns
        with `errno` = 3.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `lrq` is too small: `lrq` = *<value>*.

        (`errno` 1)
            On entry, `liq` is too small: `liq` = *<value>*.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 10.

        (`errno` 2)
            On entry, values in `rq` appear to be invalid. Check that
            `rq` has not been corrupted between calls to
            ``dim3_scat_shep`` and ``dim3_scat_shep_eval``.

        (`errno` 2)
            On entry, values in `iq` appear to be invalid. Check that
            `iq` has not been corrupted between calls to
            ``dim3_scat_shep`` and ``dim3_scat_shep_eval``.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            On entry, at least one evaluation point lies outside the
            region of definition of the interpolant. At such points the
            corresponding values in `q` and `qx` contain extrapolated
            approximations. Points should be evaluated one by one to
            identify extrapolated values.

    References
    ----------
    Renka, R J, 1988, `Algorithm 661: QSHEP3D: Quadratic Shepard method
    for trivariate interpolation of scattered data`, ACM Trans. Math.
    Software (14), 151--152
    """

    excs = []
    fun_name = 'naginterfaces.base.interp.dim3_scat_shep_eval'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('e01thft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Input, x[m].
        utils._EngineFloat64ArrayType.c_type, # Input, y[m].
        utils._EngineFloat64ArrayType.c_type, # Input, z[m].
        utils._EngineFloat64ArrayType.c_type, # Input, f[m].
        utils._EngineIntArrayType.c_type, # Input, iq[liq].
        utils._BYREF_INT, # Input, liq.
        utils._EngineFloat64ArrayType.c_type, # Input, rq[lrq].
        utils._BYREF_INT, # Input, lrq.
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, u[n].
        utils._EngineFloat64ArrayType.c_type, # Input, v[n].
        utils._EngineFloat64ArrayType.c_type, # Input, w[n].
        utils._EngineFloat64ArrayType.c_type, # Output, q[n].
        utils._EngineFloat64ArrayType.c_type, # Output, qx[n].
        utils._EngineFloat64ArrayType.c_type, # Output, qy[n].
        utils._EngineFloat64ArrayType.c_type, # Output, qz[n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _y = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _z = utils._EngineFloat64ArrayType(
        z, {'fun_name': fun_name, 'entity': 'z'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _f = utils._EngineFloat64ArrayType(
        f, {'fun_name': fun_name, 'entity': 'f'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _liq = utils._EngineIntScalarType(
        liq, {'fun_name': fun_name, 'entity': 'liq'},
    ).to_ctypes()
    _iq = utils._EngineIntArrayType(
        iq, {'fun_name': fun_name, 'entity': 'iq'},
        exp_shape=(2*_m_py+1,),
    ).to_ctypes()
    _lrq = utils._EngineIntScalarType(
        lrq, {'fun_name': fun_name, 'entity': 'lrq'},
    ).to_ctypes()
    _rq = utils._EngineFloat64ArrayType(
        rq, {'fun_name': fun_name, 'entity': 'rq'},
        exp_shape=(10*_m_py+7,),
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _u = utils._EngineFloat64ArrayType(
        u, {'fun_name': fun_name, 'entity': 'u'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _v = utils._EngineFloat64ArrayType(
        v, {'fun_name': fun_name, 'entity': 'v'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _w = utils._EngineFloat64ArrayType(
        w, {'fun_name': fun_name, 'entity': 'w'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _q_cls = utils._EngineFloat64ArrayType(
        q, {'fun_name': fun_name, 'entity': 'q'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _q = _q_cls.to_ctypes(can_reference_contents=False)
    _qx_cls = utils._EngineFloat64ArrayType(
        qx, {'fun_name': fun_name, 'entity': 'qx'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _qx = _qx_cls.to_ctypes(can_reference_contents=False)
    _qy_cls = utils._EngineFloat64ArrayType(
        qy, {'fun_name': fun_name, 'entity': 'qy'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _qy = _qy_cls.to_ctypes(can_reference_contents=False)
    _qz_cls = utils._EngineFloat64ArrayType(
        qz, {'fun_name': fun_name, 'entity': 'qz'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _qz = _qz_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _m,
        _x,
        _y,
        _z,
        _f,
        _iq,
        _liq,
        _rq,
        _lrq,
        _n,
        _u,
        _v,
        _w,
        _q,
        _qx,
        _qy,
        _qz,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lrq is too small: lrq = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, liq is too small: liq = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 10.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, values in rq appear to be invalid.',
                    'Check that rq has not been corrupted between calls to dim3_scat_shep and dim3_scat_shep_eval.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, values in iq appear to be invalid.',
                    'Check that iq has not been corrupted between calls to dim3_scat_shep and dim3_scat_shep_eval.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99993:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, at least one evaluation point lies outside the region of',
                    'definition of the interpolant. At such points the corresponding',
                    'values in q and qx contain extrapolated approximations. Points',
                    'should be evaluated one by one to identify extrapolated values.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _q_cls.refresh_py(_q)
        _qx_cls.refresh_py(_qx)
        _qy_cls.refresh_py(_qy)
        _qz_cls.refresh_py(_qz)
        return_data = None
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def dim4_scat_shep(m, x, f, nw, nq, iq, rq):
    u"""
    Interpolating functions, modified Shepard's method, four variables.

    ``dim4_scat_shep`` generates a four-dimensional interpolant to a set
    of scattered data points, using a modified Shepard method.

    For full information please refer to the NAG Library document for
    e01tk

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e01/e01tkf.html

    Parameters
    ----------
    m : int
        m, the number of data points.

    x : float, shape (4, `m`)
        `x`[0,r-1],...,`x`[3,r-1] must be set to the Cartesian
        coordinates of the data point x_r, for r = 1,2,...,m.

    f : float, shape (`m`)
        `f`[r-1] must be set to the data value f_r, for r = 1,2,...,m.

    nw : int
        The number N_w of data points that determines each radius of
        influence R_w, appearing in the definition of each of the
        weights w_r, for r = 1,2,...,m (see [Description]). Note that
        R_w is different for each weight. If `nw` <= 0 the default value
        `nw` = min(32, `m`-1) is used instead.

    nq : int
        The number N_q of data points to be used in the least squares
        fit for coefficients defining the quadratic functions q_r(x)
        (see [Description]). If `nq` <= 0 the default value `nq` =
        min(38, `m`-1) is used instead.

    iq : int, shape (2*`m`+1), modified in place
        `On exit`: integer data defining the interpolant Q(x).

    rq : float, shape (15*`m`+9), modified in place
        `On exit`: real data defining the interpolant Q(x).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `nw` = *<value>* and `m` = *<value>*.

            Constraint: `nw` <= min(50, `m`-1).

        (`errno` 1)
            On entry, `nq` = *<value>* and `m` = *<value>*.

            Constraint: `nq` <= min(50, `m`-1).

        (`errno` 1)
            On entry, `nq` = *<value>*.

            Constraint: `nq` <= 0 or `nq` >= 14.

        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 16.

        (`errno` 2)
            There are duplicate nodes in the dataset. `x`[i-1,k-1] =
            `x`[j-1,k-1], for i = *<value>*, j = *<value>* and k =
            1,2,...,4. The interpolant cannot be derived.

        (`errno` 3)
            On entry, all the data points lie on the same
            three-dimensional hypersurface. No unique solution exists.

    References
    ----------
    Bentley, J L and Friedman, J H, 1979, `Data structures for range
    searching`, ACM Comput. Surv. (11), 397--409

    Berry, M W, Minser, K S, 1999, `Algorithm 798: high-dimensional
    interpolation using the modified Shepard method`, ACM Trans. Math.
    Software (25), 353--366

    Franke, R and Nielson, G, 1980, `Smooth interpolation of large sets
    of scattered data`, Internat. J. Num. Methods Engrg. (15),
    1691--1704

    Renka, R J, 1988, `Multivariate interpolation of large sets of
    scattered data`, ACM Trans. Math. Software (14), 139--148

    Renka, R J, 1988, `Algorithm 661: QSHEP3D: Quadratic Shepard method
    for trivariate interpolation of scattered data`, ACM Trans. Math.
    Software (14), 151--152

    Shepard, D, 1968, `A two-dimensional interpolation function for
    irregularly spaced data`, Proc. 23rd Nat. Conf. ACM, 517--523,
    Brandon/Systems Press Inc., Princeton
    """

    excs = []
    fun_name = 'naginterfaces.base.interp.dim4_scat_shep'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('e01tkft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Input, x[4, m].
        utils._EngineFloat64ArrayType.c_type, # Input, f[m].
        utils._BYREF_INT, # Input, nw.
        utils._BYREF_INT, # Input, nq.
        utils._EngineIntArrayType.c_type, # Output, iq[2*m+1].
        utils._EngineFloat64ArrayType.c_type, # Output, rq[15*m+9].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((x, 'x'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(4, _m_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _f = utils._EngineFloat64ArrayType(
        f, {'fun_name': fun_name, 'entity': 'f'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _nw = utils._EngineIntScalarType(
        nw, {'fun_name': fun_name, 'entity': 'nw'},
    ).to_ctypes()
    _nq = utils._EngineIntScalarType(
        nq, {'fun_name': fun_name, 'entity': 'nq'},
    ).to_ctypes()
    _iq_cls = utils._EngineIntArrayType(
        iq, {'fun_name': fun_name, 'entity': 'iq'},
        exp_shape=(2*_m_py+1,),
        check_contents=False,
    )
    _iq = _iq_cls.to_ctypes(can_reference_contents=False)
    _rq_cls = utils._EngineFloat64ArrayType(
        rq, {'fun_name': fun_name, 'entity': 'rq'},
        exp_shape=(15*_m_py+9,),
        check_contents=False,
    )
    _rq = _rq_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _m,
        _x,
        _f,
        _nw,
        _nq,
        _iq,
        _rq,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nw = {:s} and m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nw <= min(50, m-1).',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nq = {:s} and m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nq <= min(50, m-1).',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nq = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nq <= 0 or nq >= 14.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 16.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'There are duplicate nodes in the dataset.',
                    'x[i-1,k-1] = x[j-1,k-1],',
                    'for i = {:s}, j = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'k = 1,2,...,4. The interpolant cannot be derived.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, all the data points lie on the same three-dimensional',
                    'hypersurface. No unique solution exists.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _iq_cls.refresh_py(_iq)
        _rq_cls.refresh_py(_rq)
        return_data = None
    _err_class.manage()
    return return_data

def dim4_scat_shep_eval(m, x, f, iq, rq, n, xe, q, qx):
    u"""
    Interpolated values, evaluate interpolant computed by
    ``dim4_scat_shep``, function and first derivatives, four variables.

    ``dim4_scat_shep_eval`` evaluates the four-dimensional interpolating
    function generated by ``dim4_scat_shep`` and its first partial
    derivatives.

    For full information please refer to the NAG Library document for
    e01tl

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e01/e01tlf.html

    Parameters
    ----------
    m : int
        **Must** be the same value supplied for argument m in the
        preceding call to ``dim4_scat_shep``.

    x : float, shape (4, `m`)
        Note: the coordinates of x_r are stored in
        `x`[0,r-1]...`x`[3,r-1].

        **Must** be the same array supplied as argument x in the
        preceding call to ``dim4_scat_shep``. It **must** remain
        unchanged between calls.

    f : float, shape (`m`)
        **Must** be the same array supplied as argument f in the
        preceding call to ``dim4_scat_shep``. It **must** remain
        unchanged between calls.

    iq : int, shape (2*`m`+1)
        **Must** be the same array returned as argument iq in the
        preceding call to ``dim4_scat_shep``. It **must** remain
        unchanged between calls.

    rq : float, shape (15*`m`+9)
        **Must** be the same array returned as argument rq in the
        preceding call to ``dim4_scat_shep``. It **must** remain
        unchanged between calls.

    n : int
        n, the number of evaluation points.

    xe : float, shape (4, `n`)
        `xe`[0,r-1],...,`xe`[3,r-1] must be set to the evaluation point
        x_i, for i = 1,2,...,n.

    q : float, shape (`n`), modified in place
        `On exit`: `q`[i-1] contains the value of the interpolant, at
        x_i, for i = 1,2,...,n. If any of these evaluation points lie
        outside the region of definition of the interpolant the
        corresponding entries in `q` are set to the largest machine
        representable number (see ``machine.real_largest``), and
        ``dim4_scat_shep_eval`` returns with `errno` = 3.

    qx : float, shape (4, `n`), modified in place
        `On exit`: `qx`[j-1,i-1] contains the value of the partial
        derivatives with respect to x_j of the interpolant Q(x) at x_i,
        for i = 1,2,...,n, and for each of the four partial derivatives
        j = 1,2,3,4. If any of these evaluation points lie outside the
        region of definition of the interpolant, the corresponding
        entries in `qx` are set to the largest machine representable
        number (see ``machine.real_largest``), and
        ``dim4_scat_shep_eval`` returns with `errno` = 3.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 16.

        (`errno` 2)
            On entry, values in `rq` appear to be invalid. Check that
            `rq` has not been corrupted between calls to
            ``dim4_scat_shep`` and ``dim4_scat_shep_eval``.

        (`errno` 2)
            On entry, values in `iq` appear to be invalid. Check that
            `iq` has not been corrupted between calls to
            ``dim4_scat_shep`` and ``dim4_scat_shep_eval``.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            On entry, at least one evaluation point lies outside the
            region of definition of the interpolant. At such points the
            corresponding values in `q` and `qx` contain extrapolated
            approximations. Points should be evaluated one by one to
            identify extrapolated values.

    References
    ----------
    Berry, M W, Minser, K S, 1999, `Algorithm 798: high-dimensional
    interpolation using the modified Shepard method`, ACM Trans. Math.
    Software (25), 353--366

    Renka, R J, 1988, `Algorithm 661: QSHEP3D: Quadratic Shepard method
    for trivariate interpolation of scattered data`, ACM Trans. Math.
    Software (14), 151--152
    """

    excs = []
    fun_name = 'naginterfaces.base.interp.dim4_scat_shep_eval'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('e01tlft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Input, x[4, m].
        utils._EngineFloat64ArrayType.c_type, # Input, f[m].
        utils._EngineIntArrayType.c_type, # Input, iq[2*m+1].
        utils._EngineFloat64ArrayType.c_type, # Input, rq[15*m+9].
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, xe[4, n].
        utils._EngineFloat64ArrayType.c_type, # Output, q[n].
        utils._EngineFloat64ArrayType.c_type, # Output, qx[4, n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((x, 'x'), (xe, 'xe'), (qx, 'qx')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(4, _m_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _f = utils._EngineFloat64ArrayType(
        f, {'fun_name': fun_name, 'entity': 'f'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _iq = utils._EngineIntArrayType(
        iq, {'fun_name': fun_name, 'entity': 'iq'},
        exp_shape=(2*_m_py+1,),
    ).to_ctypes()
    _rq = utils._EngineFloat64ArrayType(
        rq, {'fun_name': fun_name, 'entity': 'rq'},
        exp_shape=(15*_m_py+9,),
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _xe = utils._EngineFloat64ArrayType(
        xe, {'fun_name': fun_name, 'entity': 'xe'},
        exp_shape=(4, _n_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _q_cls = utils._EngineFloat64ArrayType(
        q, {'fun_name': fun_name, 'entity': 'q'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _q = _q_cls.to_ctypes(can_reference_contents=False)
    _qx_cls = utils._EngineFloat64ArrayType(
        qx, {'fun_name': fun_name, 'entity': 'qx'},
        exp_shape=(4, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _qx = _qx_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _m,
        _x,
        _f,
        _iq,
        _rq,
        _n,
        _xe,
        _q,
        _qx,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 16.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, values in rq appear to be invalid.',
                    'Check that rq has not been corrupted between calls to dim4_scat_shep and dim4_scat_shep_eval.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, values in iq appear to be invalid.',
                    'Check that iq has not been corrupted between calls to dim4_scat_shep and dim4_scat_shep_eval.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99993:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, at least one evaluation point lies outside the region of',
                    'definition of the interpolant. At such points the corresponding',
                    'values in q and qx contain extrapolated approximations. Points',
                    'should be evaluated one by one to identify extrapolated values.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _q_cls.refresh_py(_q)
        _qx_cls.refresh_py(_qx)
        return_data = None
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def dim5_scat_shep(m, x, f, nw, nq, iq, rq):
    u"""
    Interpolating functions, modified Shepard's method, five variables.

    ``dim5_scat_shep`` generates a five-dimensional interpolant to a set
    of scattered data points, using a modified Shepard method.

    For full information please refer to the NAG Library document for
    e01tm

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e01/e01tmf.html

    Parameters
    ----------
    m : int
        m, the number of data points.

        Note: on the basis of experimental results reported in Berry and
        Minser (1999), it is recommended to use `m` >= 4000.

    x : float, shape (5, `m`)
        `x`[0,r-1],...,`x`[4,r-1] must be set to the Cartesian
        coordinates of the data point x_r, for r = 1,2,...,m.

    f : float, shape (`m`)
        `f`[r-1] must be set to the data value f_r, for r = 1,2,...,m.

    nw : int
        The number N_w of data points that determines each radius of
        influence R_w, appearing in the definition of each of the
        weights w_r, for r = 1,2,...,m (see [Description]). Note that
        R_w is different for each weight. If `nw` <= 0 the default value
        `nw` = min(32, `m`-1) is used instead.

    nq : int
        The number N_q of data points to be used in the least squares
        fit for coefficients defining the quadratic functions q_r(x)
        (see [Description]). If `nq` <= 0 the default value `nq` =
        min(50, `m`-1) is used instead.

    iq : int, shape (2*`m`+1), modified in place
        `On exit`: integer data defining the interpolant Q(x).

    rq : float, shape (21*`m`+11), modified in place
        `On exit`: real data defining the interpolant Q(x).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `nw` = *<value>* and `m` = *<value>*.

            Constraint: `nw` <= min(50, `m`-1).

        (`errno` 1)
            On entry, `nq` = *<value>* and `m` = *<value>*.

            Constraint: `nq` <= min(70, `m`-1).

        (`errno` 1)
            On entry, `nq` = *<value>*.

            Constraint: `nq` <= 0 or `nq` >= 20.

        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 23.

        (`errno` 2)
            There are duplicate nodes in the dataset. `x`[i-1,k-1] =
            `x`[j-1,k-1], for i = *<value>*, j = *<value>* and k =
            1,2,...,5. The interpolant cannot be derived.

        (`errno` 3)
            On entry, all the data points lie on the same
            four-dimensional hypersurface. No unique solution exists.

    References
    ----------
    Bentley, J L and Friedman, J H, 1979, `Data structures for range
    searching`, ACM Comput. Surv. (11), 397--409

    Berry, M W, Minser, K S, 1999, `Algorithm 798: high-dimensional
    interpolation using the modified Shepard method`, ACM Trans. Math.
    Software (25), 353--366

    Franke, R and Nielson, G, 1980, `Smooth interpolation of large sets
    of scattered data`, Internat. J. Num. Methods Engrg. (15),
    1691--1704

    Renka, R J, 1988, `Multivariate interpolation of large sets of
    scattered data`, ACM Trans. Math. Software (14), 139--148

    Renka, R J, 1988, `Algorithm 661: QSHEP3D: Quadratic Shepard method
    for trivariate interpolation of scattered data`, ACM Trans. Math.
    Software (14), 151--152

    Shepard, D, 1968, `A two-dimensional interpolation function for
    irregularly spaced data`, Proc. 23rd Nat. Conf. ACM, 517--523,
    Brandon/Systems Press Inc., Princeton
    """

    excs = []
    fun_name = 'naginterfaces.base.interp.dim5_scat_shep'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('e01tmft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Input, x[5, m].
        utils._EngineFloat64ArrayType.c_type, # Input, f[m].
        utils._BYREF_INT, # Input, nw.
        utils._BYREF_INT, # Input, nq.
        utils._EngineIntArrayType.c_type, # Output, iq[2*m+1].
        utils._EngineFloat64ArrayType.c_type, # Output, rq[21*m+11].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((x, 'x'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(5, _m_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _f = utils._EngineFloat64ArrayType(
        f, {'fun_name': fun_name, 'entity': 'f'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _nw = utils._EngineIntScalarType(
        nw, {'fun_name': fun_name, 'entity': 'nw'},
    ).to_ctypes()
    _nq = utils._EngineIntScalarType(
        nq, {'fun_name': fun_name, 'entity': 'nq'},
    ).to_ctypes()
    _iq_cls = utils._EngineIntArrayType(
        iq, {'fun_name': fun_name, 'entity': 'iq'},
        exp_shape=(2*_m_py+1,),
        check_contents=False,
    )
    _iq = _iq_cls.to_ctypes(can_reference_contents=False)
    _rq_cls = utils._EngineFloat64ArrayType(
        rq, {'fun_name': fun_name, 'entity': 'rq'},
        exp_shape=(21*_m_py+11,),
        check_contents=False,
    )
    _rq = _rq_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _m,
        _x,
        _f,
        _nw,
        _nq,
        _iq,
        _rq,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nw = {:s} and m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nw <= min(50, m-1).',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nq = {:s} and m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nq <= min(70, m-1).',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nq = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nq <= 0 or nq >= 20.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 23.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'There are duplicate nodes in the dataset.',
                    'x[i-1,k-1] = x[j-1,k-1],',
                    'for i = {:s}, j = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'k = 1,2,...,5. The interpolant cannot be derived.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, all the data points lie on the same four-dimensional hypersurface.',
                    'No unique solution exists.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _iq_cls.refresh_py(_iq)
        _rq_cls.refresh_py(_rq)
        return_data = None
    _err_class.manage()
    return return_data

def dim5_scat_shep_eval(m, x, f, iq, rq, n, xe, q, qx):
    u"""
    Interpolated values, evaluate interpolant computed by
    ``dim5_scat_shep``, function and first derivatives, five variables.

    ``dim5_scat_shep_eval`` evaluates the five-dimensional interpolating
    function generated by ``dim5_scat_shep`` and its first partial
    derivatives.

    For full information please refer to the NAG Library document for
    e01tn

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e01/e01tnf.html

    Parameters
    ----------
    m : int
        **Must** be the same value supplied for argument m in the
        preceding call to ``dim5_scat_shep``.

    x : float, shape (5, `m`)
        **Must** be the same array supplied as argument x in the
        preceding call to ``dim5_scat_shep``. It **must** remain
        unchanged between calls.

    f : float, shape (`m`)
        **Must** be the same array supplied as argument f in the
        preceding call to ``dim5_scat_shep``. It **must** remain
        unchanged between calls.

    iq : int, shape (2*`m`+1)
        **Must** be the same array returned as argument iq in the
        preceding call to ``dim5_scat_shep``. It **must** remain
        unchanged between calls.

    rq : float, shape (21*`m`+11)
        **Must** be the same array returned as argument rq in the
        preceding call to ``dim5_scat_shep``. It **must** remain
        unchanged between calls.

    n : int
        n, the number of evaluation points.

    xe : float, shape (5, `n`)
        `xe`[0,i-1],...,`xe`[4,i-1] must be set to the evaluation point
        x_i, for i = 1,2,...,n.

    q : float, shape (`n`), modified in place
        `On exit`: `q`[i-1] contains the value of the interpolant, at
        x_i, for i = 1,2,...,n. If any of these evaluation points lie
        outside the region of definition of the interpolant the
        corresponding entries in `q` are set to the largest machine
        representable number (see ``machine.real_largest``), and
        ``dim5_scat_shep_eval`` returns with `errno` = 3.

    qx : float, shape (5, `n`), modified in place
        `On exit`: `qx`[j-1,i-1] contains the value of the partial
        derivatives with respect to x_j of the interpolant Q(x) at x_i,
        for i = 1,2,...,n, and for each of the five partial derivatives
        j = 1,2,3,4,5. If any of these evaluation points lie outside the
        region of definition of the interpolant, the corresponding
        entries in `qx` are set to the largest machine representable
        number (see ``machine.real_largest``), and
        ``dim5_scat_shep_eval`` returns with `errno` = 3.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 23.

        (`errno` 2)
            On entry, values in `rq` appear to be invalid. Check that
            `rq` has not been corrupted between calls to
            ``dim5_scat_shep`` and ``dim5_scat_shep_eval``.

        (`errno` 2)
            On entry, values in `iq` appear to be invalid. Check that
            `iq` has not been corrupted between calls to
            ``dim5_scat_shep`` and ``dim5_scat_shep_eval``.

        (`errno` 3)
            On entry, at least one evaluation point lies outside the
            region of definition of the interpolant. At such points the
            corresponding values in `q` and `qx` contain extrapolated
            approximations. Points should be evaluated one by one to
            identify extrapolated values.

    References
    ----------
    Berry, M W, Minser, K S, 1999, `Algorithm 798: high-dimensional
    interpolation using the modified Shepard method`, ACM Trans. Math.
    Software (25), 353--366

    Renka, R J, 1988, `Algorithm 661: QSHEP3D: Quadratic Shepard method
    for trivariate interpolation of scattered data`, ACM Trans. Math.
    Software (14), 151--152
    """

    excs = []
    fun_name = 'naginterfaces.base.interp.dim5_scat_shep_eval'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('e01tnft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Input, x[5, m].
        utils._EngineFloat64ArrayType.c_type, # Input, f[m].
        utils._EngineIntArrayType.c_type, # Input, iq[2*m+1].
        utils._EngineFloat64ArrayType.c_type, # Input, rq[21*m+11].
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, xe[5, n].
        utils._EngineFloat64ArrayType.c_type, # Output, q[n].
        utils._EngineFloat64ArrayType.c_type, # Output, qx[5, n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((x, 'x'), (xe, 'xe'), (qx, 'qx')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(5, _m_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _f = utils._EngineFloat64ArrayType(
        f, {'fun_name': fun_name, 'entity': 'f'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _iq = utils._EngineIntArrayType(
        iq, {'fun_name': fun_name, 'entity': 'iq'},
        exp_shape=(2*_m_py+1,),
    ).to_ctypes()
    _rq = utils._EngineFloat64ArrayType(
        rq, {'fun_name': fun_name, 'entity': 'rq'},
        exp_shape=(21*_m_py+11,),
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _xe = utils._EngineFloat64ArrayType(
        xe, {'fun_name': fun_name, 'entity': 'xe'},
        exp_shape=(5, _n_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _q_cls = utils._EngineFloat64ArrayType(
        q, {'fun_name': fun_name, 'entity': 'q'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _q = _q_cls.to_ctypes(can_reference_contents=False)
    _qx_cls = utils._EngineFloat64ArrayType(
        qx, {'fun_name': fun_name, 'entity': 'qx'},
        exp_shape=(5, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _qx = _qx_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _m,
        _x,
        _f,
        _iq,
        _rq,
        _n,
        _xe,
        _q,
        _qx,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 23.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, values in rq appear to be invalid.',
                    'Check that rq has not been corrupted between calls to dim5_scat_shep and dim5_scat_shep_eval.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, values in iq appear to be invalid.',
                    'Check that iq has not been corrupted between calls to dim5_scat_shep and dim5_scat_shep_eval.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, at least one evaluation point lies outside the region of',
                    'definition of the interpolant. At such points the corresponding',
                    'values in q and qx contain extrapolated approximations. Points',
                    'should be evaluated one by one to identify extrapolated values.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _q_cls.refresh_py(_q)
        _qx_cls.refresh_py(_qx)
        return_data = None
    _err_class.manage()
    return return_data

def dimn_grid(d, narr, uniform, axis, lx, v, point, method, k, wf):
    u"""
    Interpolates an n-dimensional point on grid data using either
    linear, cubic or modified Shepard's method.

    ``dimn_grid`` interpolates data at a point in n-dimensional space,
    that is defined by a set of gridded data points.
    It offers three methods to interpolate the data: Linear
    Interpolation, Cubic Interpolation and Weighted Average.

    For full information please refer to the NAG Library document for
    e01za

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e01/e01zaf.html

    Parameters
    ----------
    d : int
        d, the number of dimensions.

    narr : int, shape (`d`)
        The number of data ordinates in each dimension, with `narr`[i-1]
        = n_i, for i = 1,2,...,`d`.

    uniform : bool
        States whether the data points are uniformly spaced.

        `uniform` = True

            The data points are uniformly spaced.

        `uniform` = False

            The data points are not uniformly spaced.

    axis : float, shape (`lx`)
        Defines the axis. If the data points are uniformly spaced (see
        argument `uniform`) `axis` should contain the start and end of
        each dimension (x_{1 1},x_{1 n_1},...,x_{d 1},x_{d n_d}).

        If the data points are not uniformly spaced, `axis` should
        contain all the data ordinates for each dimension
        (x_{1 1},x_{1 2},...,x_{1 n_1},...,x_{d 1},x_{d 2},...,x_{d n_d}).

    lx : int
        The dimension of the array `axis`.

    v : float, shape (sum(`narr`))
        Holds the values of the data points in such an order that the
        index of a data value with coordinates (z_1,z_2,...,z_d) is

            sum(z_iprod_{n in S_i}n),

        where S_i = {`narr`[l-1]:l = 1,...,i-1} e.g.,
        ((x_{11},x_{21},...,x_{d1}),(x_{12},x_{21},...,x_{d1}),...,(x_{1n_d},x_{21},...,x_{d1}),(x_{11},x_{22},...,x_{d1}),(x_{12},x_{22},...,x_{d1}),...,(x_{1n_d},x_{2n_d},...,x_{dn_d})).

    point : float, shape (`d`)
        x, the point at which the data value is to be interpolated.

    method : int
        The method to be used.

        `method` = 1

            Weighted Average.

        `method` = 2

            Linear Interpolation.

        `method` = 3

            Cubic Interpolation.

    k : int
        If `method` = 1, `k` controls the number of data points used in
        the Weighted Average method, with `k` points used in each
        dimension, either side of the interpolation point. The total
        number of data points used for the interpolation will therefore
        be (2*`k`)^`d`.

        If `method` != 1, then `k` is not referenced and need not be
        set.

    wf : float
        The power used for the weighted average such that a high power
        will cause closer points to be more heavily weighted.

    Returns
    -------
    ans : float
        Holds the result of the interpolation.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `d` = *<value>*.

            Constraint: `d` >= 2.

        (`errno` 2)
            On entry, `narr`[*<value>*] = *<value>*.

            Constraint: `narr`[i-1] >= 2.

        (`errno` 4)
            On entry, `axis` decreases in dimension *<value>*.

            Constraint: `axis` definition must be strictly increasing.

        (`errno` 5)
            On entry, `lx` = *<value>*, sum of `narr`= *<value>*.

            Constraint: if `uniform` = False, `lx` = sum of `narr`.

        (`errno` 5)
            On entry, `lx` = *<value>*, `d` = *<value>*.

            Constraint: if `uniform` = True, `lx` = 2*`d`.

        (`errno` 7)
            On entry, `k` = *<value>*.

            Constraint: if `method` = 1, `k` >= 1.

        (`errno` 8)
            On entry, `point`[*<value>*] = *<value>* and data range =
            [*<value>*,*<value>*].

            Constraint: `point` must be within the data range.

        (`errno` 9)
            On entry, `method` = 3 and `uniform` = False.

            Constraint: if `method` = 3, `uniform` must be True.

        (`errno` 9)
            On entry, `method` = *<value>*.

            Constraint: `method` = 1, 2 or 3.

        (`errno` 10)
            On entry, `wf` = *<value>*.

            Constraint: if `method` = 1, 1.0 <= `wf` <= 15.0.

        (`errno` 101)
            Cubic Interpolation method does not have enough data
            surrounding `point`; interpolation not possible.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 201)
            Warning: the size of `k` has been reduced, due to too few
            data points around `point`.

    References
    ----------
    Keys, R, 1981, `Cubic Convolution Interpolation for Digital Image
    Processing`, IEEE Transactions on Acoutstics, Speech, and Signal
    Processing (Vol ASSP-29 No. 6), 1153--1160,
    http://hmi.stanford.edu/doc/Tech_Notes/HMI-TN-2004-004-Interpolation/Keys_cubic_interp.pdf

    Shepard, D, 1968, `A two-dimensional interpolation function for
    irregularly spaced data`, Proc. 23rd Nat. Conf. ACM, 517--523,
    Brandon/Systems Press Inc., Princeton
    """

    excs = []
    fun_name = 'naginterfaces.base.interp.dimn_grid'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('e01zaft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, d.
        utils._EngineIntArrayType.c_type, # Input, narr[d].
        utils._BYREF_LOGICAL, # Input, uniform.
        utils._EngineFloat64ArrayType.c_type, # Input, axis[lx].
        utils._BYREF_INT, # Input, lx.
        utils._EngineFloat64ArrayType.c_type, # Input, v[lv].
        utils._BYREF_INT, # Input, lv.
        utils._EngineFloat64ArrayType.c_type, # Input, point[d].
        utils._BYREF_INT, # Input, method.
        utils._BYREF_INT, # Input, k.
        utils._BYREF_FLOAT64, # Input, wf.
        utils._BYREF_FLOAT64, # Output, ans.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _d = utils._EngineIntScalarType(
        d, {'fun_name': fun_name, 'entity': 'd'},
    ).to_ctypes()
    _d_py = utils._EngineIntScalarType.to_py(_d)
    _narr_cls = utils._EngineIntArrayType(
        narr, {'fun_name': fun_name, 'entity': 'narr'},
        exp_shape=(_d_py,),
    )
    _narr = _narr_cls.to_ctypes()
    _uniform = utils._EngineLogicalScalarType(
        uniform, {'fun_name': fun_name, 'entity': 'uniform'},
    ).to_ctypes()
    _lx = utils._EngineIntScalarType(
        lx, {'fun_name': fun_name, 'entity': 'lx'},
    ).to_ctypes()
    _lx_py = utils._EngineIntScalarType.to_py(_lx)
    _axis = utils._EngineFloat64ArrayType(
        axis, {'fun_name': fun_name, 'entity': 'axis'},
        exp_shape=(_lx_py,),
    ).to_ctypes()
    _lv_py = utils._EngineFloat64ArrayType.get_shape(v, {'fun_name': fun_name, 'entity': 'v'}, exp_rank=1)[0]
    _lv = utils._EngineIntScalarType(
        _lv_py, {'fun_name': fun_name, 'entity': '_lv'},
    ).to_ctypes()
    _v = utils._EngineFloat64ArrayType(
        v, {'fun_name': fun_name, 'entity': 'v'},
        exp_shape=(_narr_cls.get_product(0, _d_py-1),),
    ).to_ctypes()
    _point = utils._EngineFloat64ArrayType(
        point, {'fun_name': fun_name, 'entity': 'point'},
        exp_shape=(_d_py,),
    ).to_ctypes()
    _method = utils._EngineIntScalarType(
        method, {'fun_name': fun_name, 'entity': 'method'},
    ).to_ctypes()
    _k = utils._EngineIntScalarType(
        k, {'fun_name': fun_name, 'entity': 'k'},
    ).to_ctypes()
    _wf = utils._EngineFloat64ScalarType(
        wf, {'fun_name': fun_name, 'entity': 'wf'},
    ).to_ctypes()
    _ans = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _d,
        _narr,
        _uniform,
        _axis,
        _lx,
        _v,
        _lv,
        _point,
        _method,
        _k,
        _wf,
        _ans,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, d = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: d >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99989:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, narr[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: narr[i-1] >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99987:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, axis decreases in dimension {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: axis definition must be strictly increasing.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lx = {:s}, sum of narr= {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if uniform = False, lx = sum of narr.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lx = {:s}, d = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if uniform = True, lx = 2*d.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, k = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: if method = 1, k >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, point[{:s}] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'data range = [{:s},{:s}].'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[4]),
                    ),
                    'Constraint: point must be within the data range.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 9 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, method = 3 and uniform = False.',
                    'Constraint: if method = 3, uniform must be True.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 9 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, method = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: method = 1, 2 or 3.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 10 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, wf = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: if method = 1, 1.0 <= wf <= 15.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 101 and _ifmt == 99988:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Cubic Interpolation method does not have enough data surrounding point;',
                    'interpolation not possible.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 201 and _ifmt == 99990:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Warning: the size of k has been reduced,',
                    'due to too few data points around point.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _ans_py = utils._EngineFloat64ScalarType.to_py(_ans)
        return_data = _ans_py
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = _collections.namedtuple(
                'nag_return_data',
                ['ans'],
            )(
                _ans_py,
            )
    _err_class.manage()
    return return_data

def dimn_scat_shep(d, m, x, f, nw, nq, iq, rq):
    u"""
    Interpolating function, modified Shepard's method, d dimensions.

    ``dimn_scat_shep`` generates a multidimensional interpolant to a set
    of scattered data points, using a modified Shepard method.
    When the number of dimensions is no more than five, there are
    corresponding functions in submodule ``interp`` which are specific
    to the given dimensionality. ``dim2_scat_shep`` generates the
    two-dimensional interpolant, while ``dim3_scat_shep``,
    ``dim4_scat_shep`` and ``dim5_scat_shep`` generate the three-, four
    - and five-dimensional interpolants respectively.

    For full information please refer to the NAG Library document for
    e01zm

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e01/e01zmf.html

    Parameters
    ----------
    d : int
        d, the number of dimensions.

    m : int
        m, the number of data points.

        Note: on the basis of experimental results reported in Berry and
        Minser (1999), when `d` >= 5 it is recommended to use `m` >=
        4000.

    x : float, shape (`d`, `m`)
        The `d` components of the first data point must be stored in
        elements 0,1,...,`d`-1 of `x`. The second data point must be
        stored in elements `d`,`d`+1,...,2*`d`-1 of `x`, and so on. In
        general, the `m` data points must be stored in `x`[j,i], for j =
        0,1,...,`d`-1, for i = 0,1,...,`m`-1.

    f : float, shape (`m`)
        `f`[r-1] must be set to the data value f_r, for r = 1,2,...,m.

    nw : int
        The number N_w of data points that determines each radius of
        influence R_w, appearing in the definition of each of the
        weights w_r, for r = 1,2,...,m (see [Description]). Note that
        R_w is different for each weight. If `nw` <= 0 the default value
        `nw` = min(2*(`d`+1)*(`d`+2), `m`-1) is used instead.

        Suggested value: `nw` = -1.

    nq : int
        The number N_q of data points to be used in the least squares
        fit for coefficients defining the quadratic functions q_r(x)
        (see [Description]). If `nq` <= 0 the default value `nq` =
        min((`d`+1)*(`d`+2)*6/5, `m`-1) is used instead.

        Suggested value: `nq` = -1.

    iq : int, shape (2*`m`+1), modified in place
        `On exit`: integer data defining the interpolant Q(x).

    rq : float, shape (lrq), modified in place
        `On exit`: real data defining the interpolant Q(x).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `nw` = *<value>* and `m` = *<value>*.

            Constraint: `nw` <= `m`-1.

        (`errno` 1)
            On entry, `nq` = *<value>* and `m` = *<value>*.

            Constraint: `nq` <= `m`-1.

        (`errno` 1)
            On entry, `nq` = *<value>* and `d` = *<value>*.

            Constraint: `nq` <= 0 or `nq` >= (`d`+1)*(`d`+2)/2-1.

        (`errno` 1)
            On entry, `m` = *<value>* and `d` = *<value>*.

            Constraint: `m` >= (`d`+1)*(`d`+2)/2+2.

        (`errno` 1)
            On entry, ((`d`+1)*(`d`+2)/2)*`m`+2*`d`+1 exceeds the
            largest machine integer.

            `d` = *<value>* and `m` = *<value>*.

        (`errno` 1)
            On entry, `d` = *<value>*.

            Constraint: `d` >= 2.

        (`errno` 2)
            There are duplicate nodes in the dataset. `x`[k-1,i-1] =
            `x`[k-1,j-1], for i = *<value>*, j = *<value>* and k =
            1,2,...,`d`. The interpolant cannot be derived.

        (`errno` 3)
            On entry, all the data points lie on the same hypersurface.
            No unique solution exists.

    References
    ----------
    Bentley, J L and Friedman, J H, 1979, `Data structures for range
    searching`, ACM Comput. Surv. (11), 397--409

    Berry, M W, Minser, K S, 1999, `Algorithm 798: high-dimensional
    interpolation using the modified Shepard method`, ACM Trans. Math.
    Software (25), 353--366

    Franke, R and Nielson, G, 1980, `Smooth interpolation of large sets
    of scattered data`, Internat. J. Num. Methods Engrg. (15),
    1691--1704

    Renka, R J, 1988, `Multivariate interpolation of large sets of
    scattered data`, ACM Trans. Math. Software (14), 139--148

    Renka, R J, 1988, `Algorithm 661: QSHEP3D: Quadratic Shepard method
    for trivariate interpolation of scattered data`, ACM Trans. Math.
    Software (14), 151--152

    Shepard, D, 1968, `A two-dimensional interpolation function for
    irregularly spaced data`, Proc. 23rd Nat. Conf. ACM, 517--523,
    Brandon/Systems Press Inc., Princeton
    """

    excs = []
    fun_name = 'naginterfaces.base.interp.dimn_scat_shep'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('e01zmft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, d.
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Input, x[d, m].
        utils._EngineFloat64ArrayType.c_type, # Input, f[m].
        utils._BYREF_INT, # Input, nw.
        utils._BYREF_INT, # Input, nq.
        utils._EngineIntArrayType.c_type, # Output, iq[2*m+1].
        utils._EngineFloat64ArrayType.c_type, # Output, rq[lrq].
        utils._BYREF_INT, # Input, lrq.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((x, 'x'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _d = utils._EngineIntScalarType(
        d, {'fun_name': fun_name, 'entity': 'd'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _d_py = utils._EngineIntScalarType.to_py(_d)
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_d_py, _m_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _f = utils._EngineFloat64ArrayType(
        f, {'fun_name': fun_name, 'entity': 'f'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _nw = utils._EngineIntScalarType(
        nw, {'fun_name': fun_name, 'entity': 'nw'},
    ).to_ctypes()
    _nq = utils._EngineIntScalarType(
        nq, {'fun_name': fun_name, 'entity': 'nq'},
    ).to_ctypes()
    _iq_cls = utils._EngineIntArrayType(
        iq, {'fun_name': fun_name, 'entity': 'iq'},
        exp_shape=(2*_m_py+1,),
        check_contents=False,
    )
    _iq = _iq_cls.to_ctypes(can_reference_contents=False)
    _lrq_py = utils._EngineFloat64ArrayType.get_shape(rq, {'fun_name': fun_name, 'entity': 'rq'}, exp_rank=1)[0]
    _lrq = utils._EngineIntScalarType(
        _lrq_py, {'fun_name': fun_name, 'entity': '_lrq'},
    ).to_ctypes()
    _rq_cls = utils._EngineFloat64ArrayType(
        rq, {'fun_name': fun_name, 'entity': 'rq'},
        exp_shape=(_lrq_py,),
        check_contents=False,
    )
    _rq = _rq_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _d,
        _m,
        _x,
        _f,
        _nw,
        _nq,
        _iq,
        _rq,
        _lrq,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nw = {:s} and m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nw <= m-1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nq = {:s} and m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nq <= m-1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nq = {:s} and d = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: nq <= 0 or nq >= (d+1)*(d+2)/2-1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s} and d = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: m >= (d+1)*(d+2)/2+2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ((d+1)*(d+2)/2)*m+2*d+1 exceeds the largest machine',
                    'integer.',
                    'd = {:s} and m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, d = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: d >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'There are duplicate nodes in the dataset.',
                    'x[k-1,i-1] = x[k-1,j-1],',
                    'for i = {:s}, j = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'k = 1,2,...,d. The interpolant cannot be derived.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, all the data points lie on the same hypersurface.',
                    'No unique solution exists.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _iq_cls.refresh_py(_iq)
        _rq_cls.refresh_py(_rq)
        return_data = None
    _err_class.manage()
    return return_data

def dimn_scat_shep_eval(d, m, x, f, iq, rq, n, xe, q, qx):
    u"""
    Interpolated values, evaluate interpolant computed by
    ``dimn_scat_shep``, function and first derivatives, d dimensions.

    ``dimn_scat_shep_eval`` evaluates the multidimensional interpolating
    function generated by ``dimn_scat_shep`` and its first partial
    derivatives.

    For full information please refer to the NAG Library document for
    e01zn

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e01/e01znf.html

    Parameters
    ----------
    d : int
        **Must** be the same value supplied for argument d in the
        preceding call to ``dimn_scat_shep``.

    m : int
        **Must** be the same value supplied for argument m in the
        preceding call to ``dimn_scat_shep``.

    x : float, shape (`d`, `m`)
        Note: the ith ordinate of the point x_j is stored in
        `x`[i-1,j-1].

        **Must** be the same array supplied as argument x in the
        preceding call to ``dimn_scat_shep``. It **must** remain
        unchanged between calls.

    f : float, shape (`m`)
        **Must** be the same array supplied as argument f in the
        preceding call to ``dimn_scat_shep``. It **must** remain
        unchanged between calls.

    iq : int, shape (2*`m`+1)
        **Must** be the same array returned as argument iq in the
        preceding call to ``dimn_scat_shep``. It **must** remain
        unchanged between calls.

    rq : float, shape (lrq)
        **Must** be the same array returned as argument rq in the
        preceding call to ``dimn_scat_shep``. It **must** remain
        unchanged between calls.

    n : int
        n, the number of evaluation points.

    xe : float, shape (`d`, `n`)
        Note: the ith ordinate of the point x_j is stored in
        `xe`[i-1,j-1].

        `xe`[0,j-1],...,`xe`[`d`-1,j-1] must be set to the evaluation
        point x_j, for j = 1,2,...,n.

    q : float, shape (`n`), modified in place
        `On exit`: `q`[i-1] contains the value of the interpolant, at
        x_i, for i = 1,2,...,n. If any of these evaluation points lie
        outside the region of definition of the interpolant the
        corresponding entries in `q` are set to an extrapolated
        approximation, and ``dimn_scat_shep_eval`` returns with `errno`
        = 3.

    qx : float, shape (`d`, `n`), modified in place
        `On exit`: `qx`[i-1,j-1] contains the value of the partial
        derivatives with respect to the ith independent variable
        (dimension) of the interpolant Q(x) at x_j, for j = 1,2,...,n,
        and for each of the partial derivatives i = 1,2,...,d. If any of
        these evaluation points lie outside the region of definition of
        the interpolant, the corresponding entries in `qx` are set to
        extrapolated approximations to the partial derivatives, and
        ``dimn_scat_shep_eval`` returns with `errno` = 3.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 1)
            On entry, `m` = *<value>* and `d` = *<value>*.

            Constraint: `m` >= (`d`+1)*(`d`+2)/2+2.

        (`errno` 1)
            On entry, ((`d`+1)*(`d`+2)/2)*`m`+2*`d`+1 exceeds the
            largest machine integer.

            `d` = *<value>* and `m` = *<value>*.

        (`errno` 1)
            On entry, `d` = *<value>*.

            Constraint: `d` >= 2.

        (`errno` 2)
            On entry, values in `rq` appear to be invalid. Check that
            `rq` has not been corrupted between calls to
            ``dimn_scat_shep`` and ``dimn_scat_shep_eval``.

        (`errno` 2)
            On entry, values in `iq` appear to be invalid. Check that
            `iq` has not been corrupted between calls to
            ``dimn_scat_shep`` and ``dimn_scat_shep_eval``.

        (`errno` 3)
            On entry, at least one evaluation point lies outside the
            region of definition of the interpolant. At such points the
            corresponding values in `q` and `qx` contain extrapolated
            approximations. Points should be evaluated one by one to
            identify extrapolated values.

    References
    ----------
    Berry, M W, Minser, K S, 1999, `Algorithm 798: high-dimensional
    interpolation using the modified Shepard method`, ACM Trans. Math.
    Software (25), 353--366

    Renka, R J, 1988, `Algorithm 661: QSHEP3D: Quadratic Shepard method
    for trivariate interpolation of scattered data`, ACM Trans. Math.
    Software (14), 151--152
    """

    excs = []
    fun_name = 'naginterfaces.base.interp.dimn_scat_shep_eval'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('e01znft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, d.
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Input, x[d, m].
        utils._EngineFloat64ArrayType.c_type, # Input, f[m].
        utils._EngineIntArrayType.c_type, # Input, iq[2*m+1].
        utils._EngineFloat64ArrayType.c_type, # Input, rq[lrq].
        utils._BYREF_INT, # Input, lrq.
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, xe[d, n].
        utils._EngineFloat64ArrayType.c_type, # Output, q[n].
        utils._EngineFloat64ArrayType.c_type, # Output, qx[d, n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((x, 'x'), (xe, 'xe'), (qx, 'qx')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _d = utils._EngineIntScalarType(
        d, {'fun_name': fun_name, 'entity': 'd'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _d_py = utils._EngineIntScalarType.to_py(_d)
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_d_py, _m_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _f = utils._EngineFloat64ArrayType(
        f, {'fun_name': fun_name, 'entity': 'f'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _iq = utils._EngineIntArrayType(
        iq, {'fun_name': fun_name, 'entity': 'iq'},
        exp_shape=(2*_m_py+1,),
    ).to_ctypes()
    _lrq_py = utils._EngineFloat64ArrayType.get_shape(rq, {'fun_name': fun_name, 'entity': 'rq'}, exp_rank=1)[0]
    _lrq = utils._EngineIntScalarType(
        _lrq_py, {'fun_name': fun_name, 'entity': '_lrq'},
    ).to_ctypes()
    _rq = utils._EngineFloat64ArrayType(
        rq, {'fun_name': fun_name, 'entity': 'rq'},
        exp_shape=(_lrq_py,),
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _xe = utils._EngineFloat64ArrayType(
        xe, {'fun_name': fun_name, 'entity': 'xe'},
        exp_shape=(_d_py, _n_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _q_cls = utils._EngineFloat64ArrayType(
        q, {'fun_name': fun_name, 'entity': 'q'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _q = _q_cls.to_ctypes(can_reference_contents=False)
    _qx_cls = utils._EngineFloat64ArrayType(
        qx, {'fun_name': fun_name, 'entity': 'qx'},
        exp_shape=(_d_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _qx = _qx_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _d,
        _m,
        _x,
        _f,
        _iq,
        _rq,
        _lrq,
        _n,
        _xe,
        _q,
        _qx,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s} and d = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: m >= (d+1)*(d+2)/2+2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ((d+1)*(d+2)/2)*m+2*d+1 exceeds the largest machine',
                    'integer.',
                    'd = {:s} and m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, d = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: d >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, values in rq appear to be invalid.',
                    'Check that rq has not been corrupted between calls to',
                    'dimn_scat_shep and dimn_scat_shep_eval.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, values in iq appear to be invalid.',
                    'Check that iq has not been corrupted between calls to',
                    'dimn_scat_shep and dimn_scat_shep_eval.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, at least one evaluation point lies outside the region of',
                    'definition of the interpolant. At such points the corresponding',
                    'values in q and qx contain extrapolated approximations. Points',
                    'should be evaluated one by one to identify extrapolated values.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _q_cls.refresh_py(_q)
        _qx_cls.refresh_py(_qx)
        return_data = None
    _err_class.manage()
    return return_data
