# -*- coding: utf-8 -*-
# pylint: disable=bare-except,chained-comparison,invalid-name,len-as-condition,line-too-long,protected-access,redefined-outer-name,too-many-arguments,too-many-boolean-expressions,too-many-branches,too-many-lines,too-many-locals,too-many-statements,unneeded-not,unused-import
u"""
Interfaces for the NAG Mark 27.0 `sort` Chapter.

``sort`` - Sorting and Searching

Base interfaces.

We do not intend for these lower-level base interfaces to be called
directly from user code. Please refer to the ``naginterfaces.library``
subpackage instead.

This module is concerned with sorting and searching numeric or character
data.
It handles only the simplest types of data structure and it is concerned
only with **internal** sorting and searching -- that is, sorting and
searching a set of data which can all be stored within the program.

If you have large files of data or complicated data structures to be
sorted or searched you should use a comprehensive sorting or searching
program or package.

For full information please refer to the NAG Library document

https://www.nag.com/numeric/nl/nagdoc_27/flhtml/m01/m01intro.html
"""

# NAG Copyright 2017-2020.

import logging as _logging
import sys as _sys

from ..base import utils

def realvec_sort(rv, m1, m2, order):
    u"""
    Sort a vector, real numbers.

    ``realvec_sort`` rearranges a vector of `float` numbers into
    ascending or descending order.

    For full information please refer to the NAG Library document for
    m01ca

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/m01/m01caf.html

    Parameters
    ----------
    rv : float, shape (`m2`), modified in place
        `On entry`: elements `m1`-1 to `m2`-1 of `rv` must contain
        `float` values to be sorted.

        `On exit`: these values are rearranged into sorted order.

    m1 : int
        Note: this argument represents an array index; the value you
        supply must be base-1 for compatibility with the NAG Engine.

        The index of the first element of `rv` to be sorted.

    m2 : int
        Note: this argument represents an array index; the value you
        supply must be base-1 for compatibility with the NAG Engine.

        The index of the last element of `rv` to be sorted.

    order : str, length 1
        If `order` = 'A', the values will be sorted into ascending
        (i.e., nondecreasing) order.

        If `order` = 'D', into descending order.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m1` = *<value>* and `m2` = *<value>*.

            Constraint: `m1` <= `m2`.

        (`errno` 1)
            On entry, `m1` = *<value>*.

            Constraint: `m1` >= 1.

        (`errno` 1)
            On entry, `m2` = *<value>*.

            Constraint: `m2` >= 1.

        (`errno` 2)
            On entry, `order` has an illegal value: `order` = *<value>*.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Sedgewick, R, 1978, `Implementing Quicksort programs`, Comm. ACM
    (21), 847--857

    Singleton, R C, 1969, `An efficient algorithm for sorting with
    minimal storage: Algorithm 347`, Comm. ACM (12), 185--187
    """

    excs = []
    fun_name = 'naginterfaces.base.sort.realvec_sort'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('m01caft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._EngineFloat64ArrayType.c_type, # Input/Output, rv[m2].
        utils._BYREF_INT, # Input, m1.
        utils._BYREF_INT, # Input, m2.
        utils._BYREF_STRING, # Input, order(len=1).
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of order.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _m2 = utils._EngineIntScalarType(
        m2, {'fun_name': fun_name, 'entity': 'm2'},
    ).to_ctypes()
    _m2_py = utils._EngineIntScalarType.to_py(_m2)
    _rv_cls = utils._EngineFloat64ArrayType(
        rv, {'fun_name': fun_name, 'entity': 'rv'},
        exp_shape=(_m2_py,),
    )
    _rv = _rv_cls.to_ctypes()
    _m1 = utils._EngineIntScalarType(
        m1, {'fun_name': fun_name, 'entity': 'm1'},
    ).to_ctypes()
    _order = utils._EngineCharScalarType(
        order, {'fun_name': fun_name, 'entity': 'order'}, exp_chrlen=1,
    ).to_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _rv,
        _m1,
        _m2,
        _order,
        _errbuf,
        _ierr,
        _order._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m1 = {:s} and m2 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: m1 <= m2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m1 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m1 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m2 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m2 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, order has an illegal value: order = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _rv_cls.refresh_py(_rv)
        return_data = None
    _err_class.manage()
    return return_data

def intvec_sort(iv, m1, m2, order):
    u"""
    Sort a vector, integer numbers.

    ``intvec_sort`` rearranges a vector of integer numbers into
    ascending or descending order.

    For full information please refer to the NAG Library document for
    m01cb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/m01/m01cbf.html

    Parameters
    ----------
    iv : int, shape (`m2`), modified in place
        `On entry`: elements `m1`-1 to `m2`-1 of `iv` must contain
        integer values to be sorted.

        `On exit`: these values are rearranged into sorted order.

    m1 : int
        Note: this argument represents an array index; the value you
        supply must be base-1 for compatibility with the NAG Engine.

        The index of the first element of `iv` to be sorted.

    m2 : int
        Note: this argument represents an array index; the value you
        supply must be base-1 for compatibility with the NAG Engine.

        The index of the last element of `iv` to be sorted.

    order : str, length 1
        If `order` = 'A', the values will be sorted into ascending
        (i.e., nondecreasing) order.

        If `order` = 'D', into descending order.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m1` = *<value>* and `m2` = *<value>*.

            Constraint: `m1` <= `m2`.

        (`errno` 1)
            On entry, `m1` = *<value>*.

            Constraint: `m1` >= 1.

        (`errno` 1)
            On entry, `m2` = *<value>*.

            Constraint: `m2` >= 1.

        (`errno` 2)
            On entry, `order` has an illegal value: `order` = *<value>*.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Sedgewick, R, 1978, `Implementing Quicksort programs`, Comm. ACM
    (21), 847--857

    Singleton, R C, 1969, `An efficient algorithm for sorting with
    minimal storage: Algorithm 347`, Comm. ACM (12), 185--187
    """

    excs = []
    fun_name = 'naginterfaces.base.sort.intvec_sort'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('m01cbft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._EngineIntArrayType.c_type, # Input/Output, iv[m2].
        utils._BYREF_INT, # Input, m1.
        utils._BYREF_INT, # Input, m2.
        utils._BYREF_STRING, # Input, order(len=1).
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of order.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _m2 = utils._EngineIntScalarType(
        m2, {'fun_name': fun_name, 'entity': 'm2'},
    ).to_ctypes()
    _m2_py = utils._EngineIntScalarType.to_py(_m2)
    _iv_cls = utils._EngineIntArrayType(
        iv, {'fun_name': fun_name, 'entity': 'iv'},
        exp_shape=(_m2_py,),
    )
    _iv = _iv_cls.to_ctypes()
    _m1 = utils._EngineIntScalarType(
        m1, {'fun_name': fun_name, 'entity': 'm1'},
    ).to_ctypes()
    _order = utils._EngineCharScalarType(
        order, {'fun_name': fun_name, 'entity': 'order'}, exp_chrlen=1,
    ).to_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _iv,
        _m1,
        _m2,
        _order,
        _errbuf,
        _ierr,
        _order._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m1 = {:s} and m2 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: m1 <= m2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m1 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m1 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m2 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m2 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, order has an illegal value: order = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _iv_cls.refresh_py(_iv)
        return_data = None
    _err_class.manage()
    return return_data

def charvec_sort(ch, m1, m2, l1, l2, order):
    u"""
    Sort a vector, character data.

    ``charvec_sort`` rearranges a vector of character data so that a
    specified substring is in ASCII or reverse ASCII order.

    For full information please refer to the NAG Library document for
    m01cc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/m01/m01ccf.html

    Parameters
    ----------
    ch : str, shape (`m2`), modified in place
        `On entry`: elements `m1`-1 to `m2`-1 of `ch` must contain
        character data to be sorted.

        `On exit`: these values are rearranged into sorted order.

    m1 : int
        Note: this argument represents an array index; the value you
        supply must be base-1 for compatibility with the NAG Engine.

        The index of the first element of `ch` to be sorted.

    m2 : int
        Note: this argument represents an array index; the value you
        supply must be base-1 for compatibility with the NAG Engine.

        The index of the last element of `ch` to be sorted.

    l1 : int
        Only the substring (`l1`:`l2`) of each element of `ch` is to be
        used in determining the sorted order.

    l2 : int
        Only the substring (`l1`:`l2`) of each element of `ch` is to be
        used in determining the sorted order.

    order : str, length 1
        If `order` = 'A', the values will be sorted into ASCII order.

        If `order` = 'R', into reverse ASCII order.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `l2` = *<value>* and len(`ch`[0]) = *<value>*.

            Constraint: `l2` <= len(`ch`[0]).

        (`errno` 1)
            On entry, `l1` = *<value>* and `l2` = *<value>*.

            Constraint: `l1` <= `l2`.

        (`errno` 1)
            On entry, `l1` = *<value>*.

            Constraint: `l1` >= 1.

        (`errno` 1)
            On entry, `l2` = *<value>*.

            Constraint: `l2` >= 1.

        (`errno` 1)
            On entry, `m1` = *<value>* and `m2` = *<value>*.

            Constraint: `m1` <= `m2`.

        (`errno` 1)
            On entry, `m1` = *<value>*.

            Constraint: `m1` >= 1.

        (`errno` 1)
            On entry, `m2` = *<value>*.

            Constraint: `m2` >= 1.

        (`errno` 2)
            On entry, `order` has an illegal value: `order` = *<value>*.

        (`errno` 3)
            On entry, len(`ch`[0]) = *<value>*.

            Constraint: len(`ch`[0]) <= 255.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Sedgewick, R, 1978, `Implementing Quicksort programs`, Comm. ACM
    (21), 847--857

    Singleton, R C, 1969, `An efficient algorithm for sorting with
    minimal storage: Algorithm 347`, Comm. ACM (12), 185--187
    """

    excs = []
    fun_name = 'naginterfaces.base.sort.charvec_sort'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('m01ccft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._EngineCharArrayType.c_type, # Input/Output, ch(len=*)[m2].
        utils._BYREF_INT, # Input, m1.
        utils._BYREF_INT, # Input, m2.
        utils._BYREF_INT, # Input, l1.
        utils._BYREF_INT, # Input, l2.
        utils._BYREF_STRING, # Input, order(len=1).
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of ch.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of order.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _m2 = utils._EngineIntScalarType(
        m2, {'fun_name': fun_name, 'entity': 'm2'},
    ).to_ctypes()
    _m2_py = utils._EngineIntScalarType.to_py(_m2)
    _ch_cls = utils._EngineCharArrayType(
        ch, {'fun_name': fun_name, 'entity': 'ch'}, exp_chrlen=None,
        exp_shape=(_m2_py,),
    )
    _ch = _ch_cls.to_ctypes()
    _ch_chrlen = _ch_cls.chrlen
    _m1 = utils._EngineIntScalarType(
        m1, {'fun_name': fun_name, 'entity': 'm1'},
    ).to_ctypes()
    _l1 = utils._EngineIntScalarType(
        l1, {'fun_name': fun_name, 'entity': 'l1'},
    ).to_ctypes()
    _l2 = utils._EngineIntScalarType(
        l2, {'fun_name': fun_name, 'entity': 'l2'},
    ).to_ctypes()
    _order = utils._EngineCharScalarType(
        order, {'fun_name': fun_name, 'entity': 'order'}, exp_chrlen=1,
    ).to_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _ch,
        _m1,
        _m2,
        _l1,
        _l2,
        _order,
        _errbuf,
        _ierr,
        _ch_chrlen,
        _order._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, l2 = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'len(ch[0]) = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: l2 <= len(ch[0]).',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, l1 = {:s} and l2 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: l1 <= l2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, l1 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: l1 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, l2 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: l2 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m1 = {:s} and m2 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: m1 <= m2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m1 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m1 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m2 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m2 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, order has an illegal value: order = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, len(ch[0]) = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: len(ch[0]) <= 255.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _ch_cls.refresh_py(_ch)
        return_data = None
    _err_class.manage()
    return return_data

def realvec_rank(rv, m1, m2, order, irank):
    u"""
    Rank a vector, real numbers.

    ``realvec_rank`` ranks a vector of `float` numbers in ascending or
    descending order.

    For full information please refer to the NAG Library document for
    m01da

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/m01/m01daf.html

    Parameters
    ----------
    rv : float, shape (`m2`)
        Elements `m1`-1 to `m2`-1 of `rv` must contain `float` values to
        be ranked.

    m1 : int
        Note: this argument represents an array index; the value you
        supply must be base-1 for compatibility with the NAG Engine.

        The index of the first element of `rv` to be ranked.

    m2 : int
        Note: this argument represents an array index; the value you
        supply must be base-1 for compatibility with the NAG Engine.

        The index of the last element of `rv` to be ranked.

    order : str, length 1
        If `order` = 'A', the values will be ranked in ascending (i.e.,
        nondecreasing) order.

        If `order` = 'D', into descending order.

    irank : int, shape (`m2`), modified in place
        `On exit`: elements `m1`-1 to `m2`-1 of `irank` contain the
        ranks of the corresponding elements of `rv`. Note that the ranks
        are in the range `m1` to `m2`: thus, if `rv`[i-1] is the first
        element in the rank order, `irank`[i-1] is set to `m1`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m1` = *<value>* and `m2` = *<value>*.

            Constraint: `m1` <= `m2`.

        (`errno` 1)
            On entry, `m1` = *<value>*.

            Constraint: `m1` >= 1.

        (`errno` 1)
            On entry, `m2` = *<value>*.

            Constraint: `m2` >= 1.

        (`errno` 2)
            On entry, `order` has an illegal value: `order` = *<value>*.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Knuth, D E, 1973, `The Art of Computer Programming (Volume 3)`, (2nd
    Edition), Addison--Wesley
    """

    excs = []
    fun_name = 'naginterfaces.base.sort.realvec_rank'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('m01daft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._EngineFloat64ArrayType.c_type, # Input, rv[m2].
        utils._BYREF_INT, # Input, m1.
        utils._BYREF_INT, # Input, m2.
        utils._BYREF_STRING, # Input, order(len=1).
        utils._EngineIntArrayType.c_type, # Output, irank[m2].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of order.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _m2 = utils._EngineIntScalarType(
        m2, {'fun_name': fun_name, 'entity': 'm2'},
    ).to_ctypes()
    _m2_py = utils._EngineIntScalarType.to_py(_m2)
    _rv = utils._EngineFloat64ArrayType(
        rv, {'fun_name': fun_name, 'entity': 'rv'},
        exp_shape=(_m2_py,),
    ).to_ctypes()
    _m1 = utils._EngineIntScalarType(
        m1, {'fun_name': fun_name, 'entity': 'm1'},
    ).to_ctypes()
    _order = utils._EngineCharScalarType(
        order, {'fun_name': fun_name, 'entity': 'order'}, exp_chrlen=1,
    ).to_ctypes()
    _irank_cls = utils._EngineIntArrayType(
        irank, {'fun_name': fun_name, 'entity': 'irank'},
        exp_shape=(_m2_py,),
        check_contents=False,
    )
    _irank = _irank_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _rv,
        _m1,
        _m2,
        _order,
        _irank,
        _errbuf,
        _ierr,
        _order._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m1 = {:s} and m2 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: m1 <= m2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m1 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m1 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m2 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m2 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, order has an illegal value: order = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _irank_cls.refresh_py(_irank)
        return_data = None
    _err_class.manage()
    return return_data

def intvec_rank(iv, m1, m2, order, irank):
    u"""
    Rank a vector, integer numbers.

    ``intvec_rank`` ranks a vector of integer numbers in ascending or
    descending order.

    For full information please refer to the NAG Library document for
    m01db

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/m01/m01dbf.html

    Parameters
    ----------
    iv : int, shape (`m2`)
        Elements `m1`-1 to `m2`-1 of `iv` must contain integer values to
        be ranked.

    m1 : int
        Note: this argument represents an array index; the value you
        supply must be base-1 for compatibility with the NAG Engine.

        The index of the first element of `iv` to be ranked.

    m2 : int
        Note: this argument represents an array index; the value you
        supply must be base-1 for compatibility with the NAG Engine.

        `m2` must specify the index of the last element of `iv` to be
        ranked.

    order : str, length 1
        If `order` = 'A', the values will be ranked in ascending (i.e.,
        nondecreasing) order.

        If `order` = 'D', into descending order.

    irank : int, shape (`m2`), modified in place
        `On exit`: elements `m1`-1 to `m2`-1 of `irank` contain the
        ranks of the corresponding elements of `iv`. Note that the ranks
        are in the range `m1` to `m2`: thus, if `iv`[i-1] is the first
        element in the rank order, `irank`[i-1] is set to `m1`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m1` = *<value>* and `m2` = *<value>*.

            Constraint: `m1` <= `m2`.

        (`errno` 1)
            On entry, `m1` = *<value>*.

            Constraint: `m1` >= 1.

        (`errno` 1)
            On entry, `m2` = *<value>*.

            Constraint: `m2` >= 1.

        (`errno` 2)
            On entry, `order` has an illegal value: `order` = *<value>*.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Knuth, D E, 1973, `The Art of Computer Programming (Volume 3)`, (2nd
    Edition), Addison--Wesley
    """

    excs = []
    fun_name = 'naginterfaces.base.sort.intvec_rank'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('m01dbft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._EngineIntArrayType.c_type, # Input, iv[m2].
        utils._BYREF_INT, # Input, m1.
        utils._BYREF_INT, # Input, m2.
        utils._BYREF_STRING, # Input, order(len=1).
        utils._EngineIntArrayType.c_type, # Output, irank[m2].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of order.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _m2 = utils._EngineIntScalarType(
        m2, {'fun_name': fun_name, 'entity': 'm2'},
    ).to_ctypes()
    _m2_py = utils._EngineIntScalarType.to_py(_m2)
    _iv = utils._EngineIntArrayType(
        iv, {'fun_name': fun_name, 'entity': 'iv'},
        exp_shape=(_m2_py,),
    ).to_ctypes()
    _m1 = utils._EngineIntScalarType(
        m1, {'fun_name': fun_name, 'entity': 'm1'},
    ).to_ctypes()
    _order = utils._EngineCharScalarType(
        order, {'fun_name': fun_name, 'entity': 'order'}, exp_chrlen=1,
    ).to_ctypes()
    _irank_cls = utils._EngineIntArrayType(
        irank, {'fun_name': fun_name, 'entity': 'irank'},
        exp_shape=(_m2_py,),
        check_contents=False,
    )
    _irank = _irank_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _iv,
        _m1,
        _m2,
        _order,
        _irank,
        _errbuf,
        _ierr,
        _order._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m1 = {:s} and m2 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: m1 <= m2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m1 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m1 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m2 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m2 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, order has an illegal value: order = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _irank_cls.refresh_py(_irank)
        return_data = None
    _err_class.manage()
    return return_data

def charvec_rank(ch, m1, m2, l1, l2, order, irank):
    u"""
    Rank a vector, character data.

    ``charvec_rank`` ranks a vector of character data in ASCII or
    reverse ASCII order of a specified substring.

    For full information please refer to the NAG Library document for
    m01dc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/m01/m01dcf.html

    Parameters
    ----------
    ch : str, shape (`m2`)
        Elements `m1`-1 to `m2`-1 of `ch` must contain character data to
        be ranked.

    m1 : int
        Note: this argument represents an array index; the value you
        supply must be base-1 for compatibility with the NAG Engine.

        The index of the first element of `ch` to be ranked.

    m2 : int
        Note: this argument represents an array index; the value you
        supply must be base-1 for compatibility with the NAG Engine.

        The index of the last element of `ch` to be ranked.

    l1 : int
        Only the substring (`l1`:`l2`) of each element of `ch` is to be
        used in determining the rank order.

    l2 : int
        Only the substring (`l1`:`l2`) of each element of `ch` is to be
        used in determining the rank order.

    order : str, length 1
        If `order` = 'A', the values will be ranked in ASCII order.

        If `order` = 'R', in reverse ASCII order.

    irank : int, shape (`m2`), modified in place
        `On exit`: elements `m1`-1 to `m2`-1 of `irank` contain the
        ranks of the corresponding elements of `ch`. Note that the ranks
        are in the range `m1` to `m2`: thus, if `ch`[i-1] is the first
        element in the rank order, `irank`[i-1] is set to `m1`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `l2` = *<value>* and len(`ch`[0]) = *<value>*.

            Constraint: `l2` <= len(`ch`[0]).

        (`errno` 1)
            On entry, `l1` = *<value>* and `l2` = *<value>*.

            Constraint: `l1` <= `l2`.

        (`errno` 1)
            On entry, `l1` = *<value>*.

            Constraint: `l1` >= 1.

        (`errno` 1)
            On entry, `l2` = *<value>*.

            Constraint: `l2` >= 1.

        (`errno` 1)
            On entry, `m1` = *<value>* and `m2` = *<value>*.

            Constraint: `m1` <= `m2`.

        (`errno` 1)
            On entry, `m1` = *<value>*.

            Constraint: `m1` >= 1.

        (`errno` 1)
            On entry, `m2` = *<value>*.

            Constraint: `m2` >= 1.

        (`errno` 2)
            On entry, `order` has an illegal value: `order` = *<value>*.

        (`errno` 3)
            On entry, len(`ch`[0]) = *<value>*.

            Constraint: len(`ch`[0]) <= 255.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Knuth, D E, 1973, `The Art of Computer Programming (Volume 3)`, (2nd
    Edition), Addison--Wesley
    """

    excs = []
    fun_name = 'naginterfaces.base.sort.charvec_rank'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('m01dcft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._EngineCharArrayType.c_type, # Input, ch(len=*)[m2].
        utils._BYREF_INT, # Input, m1.
        utils._BYREF_INT, # Input, m2.
        utils._BYREF_INT, # Input, l1.
        utils._BYREF_INT, # Input, l2.
        utils._BYREF_STRING, # Input, order(len=1).
        utils._EngineIntArrayType.c_type, # Output, irank[m2].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of ch.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of order.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _m2 = utils._EngineIntScalarType(
        m2, {'fun_name': fun_name, 'entity': 'm2'},
    ).to_ctypes()
    _m2_py = utils._EngineIntScalarType.to_py(_m2)
    _ch_cls = utils._EngineCharArrayType(
        ch, {'fun_name': fun_name, 'entity': 'ch'}, exp_chrlen=None,
        exp_shape=(_m2_py,),
    )
    _ch = _ch_cls.to_ctypes()
    _ch_chrlen = _ch_cls.chrlen
    _m1 = utils._EngineIntScalarType(
        m1, {'fun_name': fun_name, 'entity': 'm1'},
    ).to_ctypes()
    _l1 = utils._EngineIntScalarType(
        l1, {'fun_name': fun_name, 'entity': 'l1'},
    ).to_ctypes()
    _l2 = utils._EngineIntScalarType(
        l2, {'fun_name': fun_name, 'entity': 'l2'},
    ).to_ctypes()
    _order = utils._EngineCharScalarType(
        order, {'fun_name': fun_name, 'entity': 'order'}, exp_chrlen=1,
    ).to_ctypes()
    _irank_cls = utils._EngineIntArrayType(
        irank, {'fun_name': fun_name, 'entity': 'irank'},
        exp_shape=(_m2_py,),
        check_contents=False,
    )
    _irank = _irank_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _ch,
        _m1,
        _m2,
        _l1,
        _l2,
        _order,
        _irank,
        _errbuf,
        _ierr,
        _ch_chrlen,
        _order._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, l2 = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'len(ch[0]) = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: l2 <= len(ch[0]).',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, l1 = {:s} and l2 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: l1 <= l2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, l1 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: l1 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, l2 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: l2 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m1 = {:s} and m2 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: m1 <= m2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m1 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m1 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m2 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m2 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, order has an illegal value: order = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, len(ch[0]) = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: len(ch[0]) <= 255.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _irank_cls.refresh_py(_irank)
        return_data = None
    _err_class.manage()
    return return_data

def realmat_rank_rows(rm, m1, m2, n1, n2, order, irank):
    u"""
    Rank rows of a matrix, real numbers.

    ``realmat_rank_rows`` ranks the rows of a matrix of `float` numbers
    in ascending or descending order.

    For full information please refer to the NAG Library document for
    m01de

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/m01/m01def.html

    Parameters
    ----------
    rm : float, shape (`m2`, `n2`)
        Columns `n1` to `n2` of rows `m1`-1 to `m2`-1 of `rm` must
        contain `float` data to be ranked.

    m1 : int
        Note: this argument represents an array index; the value you
        supply must be base-1 for compatibility with the NAG Engine.

        The index of the first row of `rm` to be ranked.

    m2 : int
        Note: this argument represents an array index; the value you
        supply must be base-1 for compatibility with the NAG Engine.

        The index of the last row of `rm` to be ranked.

    n1 : int
        The index of the first column of `rm` to be used.

    n2 : int
        The index of the last column of `rm` to be used.

    order : str, length 1
        If `order` = 'A', the rows will be ranked in ascending (i.e.,
        nondecreasing) order.

        If `order` = 'D', into descending order.

    irank : int, shape (`m2`), modified in place
        `On exit`: elements `m1`-1 to `m2`-1 of `irank` contain the
        ranks of the corresponding rows of `rm`. Note that the ranks are
        in the range `m1` to `m2`: thus, if the ith row of `rm` is the
        first in the rank order, `irank`[i-1] is set to `m1`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m2` = *<value>* and ldm = *<value>*.

            Constraint: `m2` <= ldm.

        (`errno` 1)
            On entry, `n1` = *<value>* and `n2` = *<value>*.

            Constraint: `n1` <= `n2`.

        (`errno` 1)
            On entry, `n1` = *<value>*.

            Constraint: `n1` >= 1.

        (`errno` 1)
            On entry, `n2` = *<value>*.

            Constraint: `n2` >= 1.

        (`errno` 1)
            On entry, `m1` = *<value>* and `m2` = *<value>*.

            Constratint: `m1` <= `m2`.

        (`errno` 1)
            On entry, `m1` = *<value>*.

            Constraint: `m1` >= 1.

        (`errno` 1)
            On entry, `m2` = *<value>*.

            Constraint: `m2` >= 1.

        (`errno` 2)
            On entry, `order` has an illegal value: `order` = *<value>*.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Knuth, D E, 1973, `The Art of Computer Programming (Volume 3)`, (2nd
    Edition), Addison--Wesley
    """

    excs = []
    fun_name = 'naginterfaces.base.sort.realmat_rank_rows'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('m01deft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._EngineFloat64ArrayType.c_type, # Input, rm[ldm, n2].
        utils._BYREF_INT, # Input, ldm.
        utils._BYREF_INT, # Input, m1.
        utils._BYREF_INT, # Input, m2.
        utils._BYREF_INT, # Input, n1.
        utils._BYREF_INT, # Input, n2.
        utils._BYREF_STRING, # Input, order(len=1).
        utils._EngineIntArrayType.c_type, # Output, irank[m2].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of order.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((rm, 'rm'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _ldm_py = utils._EngineFloat64ArrayType.get_nd(rm, {'fun_name': fun_name, 'entity': 'rm'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldm = utils._EngineIntScalarType(
        _ldm_py, {'fun_name': fun_name, 'entity': '_ldm'},
    ).to_ctypes()
    _n2 = utils._EngineIntScalarType(
        n2, {'fun_name': fun_name, 'entity': 'n2'},
    ).to_ctypes()
    _m2 = utils._EngineIntScalarType(
        m2, {'fun_name': fun_name, 'entity': 'm2'},
    ).to_ctypes()
    _m2_py = utils._EngineIntScalarType.to_py(_m2)
    _n2_py = utils._EngineIntScalarType.to_py(_n2)
    _rm = utils._EngineFloat64ArrayType(
        rm, {'fun_name': fun_name, 'entity': 'rm'},
        exp_shape=(_m2_py, _n2_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _m1 = utils._EngineIntScalarType(
        m1, {'fun_name': fun_name, 'entity': 'm1'},
    ).to_ctypes()
    _n1 = utils._EngineIntScalarType(
        n1, {'fun_name': fun_name, 'entity': 'n1'},
    ).to_ctypes()
    _order = utils._EngineCharScalarType(
        order, {'fun_name': fun_name, 'entity': 'order'}, exp_chrlen=1,
    ).to_ctypes()
    _irank_cls = utils._EngineIntArrayType(
        irank, {'fun_name': fun_name, 'entity': 'irank'},
        exp_shape=(_m2_py,),
        check_contents=False,
    )
    _irank = _irank_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _rm,
        _ldm,
        _m1,
        _m2,
        _n1,
        _n2,
        _order,
        _irank,
        _errbuf,
        _ierr,
        _order._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m2 = {:s} and ldm = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: m2 <= ldm.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n1 = {:s} and n2 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: n1 <= n2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n1 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n1 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n2 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n2 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m1 = {:s} and m2 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constratint: m1 <= m2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m1 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m1 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m2 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m2 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, order has an illegal value: order = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _irank_cls.refresh_py(_irank)
        return_data = None
    _err_class.manage()
    return return_data

def intmat_rank_rows(im, m1, m2, n1, n2, order, irank):
    u"""
    Rank rows of a matrix, integer numbers.

    ``intmat_rank_rows`` ranks the rows of a matrix of integer numbers
    in ascending or descending order.

    For full information please refer to the NAG Library document for
    m01df

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/m01/m01dff.html

    Parameters
    ----------
    im : int, shape (`m2`, `n2`)
        Columns `n1` to `n2` of rows `m1`-1 to `m2`-1 of `im` must
        contain `int` data to be ranked.

    m1 : int
        Note: this argument represents an array index; the value you
        supply must be base-1 for compatibility with the NAG Engine.

        The index of the first row of `im` to be ranked.

    m2 : int
        Note: this argument represents an array index; the value you
        supply must be base-1 for compatibility with the NAG Engine.

        The index of the last row of `im` to be ranked.

    n1 : int
        The index of the first column of `im` to be used.

    n2 : int
        The index of the last column of `im` to be used.

    order : str, length 1
        If `order` = 'A', the rows will be ranked in ascending (i.e.,
        nondecreasing) order.

        If `order` = 'D', into descending order.

    irank : int, shape (`m2`), modified in place
        `On exit`: elements `m1`-1 to `m2`-1 of `irank` contain the
        ranks of the corresponding rows of `im`. Note that the ranks are
        in the range `m1` to `m2`: thus, if the ith row of `im` is the
        first in the rank order, `irank`[i-1] is set to `m1`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m2` = *<value>* and ldm = *<value>*.

            Constraint: `m2` <= ldm.

        (`errno` 1)
            On entry, `n1` = *<value>* and `n2` = *<value>*.

            Constraint: `n1` <= `n2`.

        (`errno` 1)
            On entry, `n1` = *<value>*.

            Constraint: `n1` >= 1.

        (`errno` 1)
            On entry, `n2` = *<value>*.

            Constraint: `n2` >= 1.

        (`errno` 1)
            On entry, `m1` = *<value>* and `m2` = *<value>*.

            Constraint: `m1` <= `m2`.

        (`errno` 1)
            On entry, `m1` = *<value>*.

            Constraint: `m1` >= 1.

        (`errno` 1)
            On entry, `m2` = *<value>*.

            Constraint: `m2` >= 1.

        (`errno` 2)
            On entry, `order` has an illegal value: `order` = *<value>*.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Knuth, D E, 1973, `The Art of Computer Programming (Volume 3)`, (2nd
    Edition), Addison--Wesley
    """

    excs = []
    fun_name = 'naginterfaces.base.sort.intmat_rank_rows'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('m01dfft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._EngineIntArrayType.c_type, # Input, im[ldm, n2].
        utils._BYREF_INT, # Input, ldm.
        utils._BYREF_INT, # Input, m1.
        utils._BYREF_INT, # Input, m2.
        utils._BYREF_INT, # Input, n1.
        utils._BYREF_INT, # Input, n2.
        utils._BYREF_STRING, # Input, order(len=1).
        utils._EngineIntArrayType.c_type, # Output, irank[m2].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of order.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((im, 'im'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _ldm_py = utils._EngineIntArrayType.get_nd(im, {'fun_name': fun_name, 'entity': 'im'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldm = utils._EngineIntScalarType(
        _ldm_py, {'fun_name': fun_name, 'entity': '_ldm'},
    ).to_ctypes()
    _n2 = utils._EngineIntScalarType(
        n2, {'fun_name': fun_name, 'entity': 'n2'},
    ).to_ctypes()
    _m2 = utils._EngineIntScalarType(
        m2, {'fun_name': fun_name, 'entity': 'm2'},
    ).to_ctypes()
    _m2_py = utils._EngineIntScalarType.to_py(_m2)
    _n2_py = utils._EngineIntScalarType.to_py(_n2)
    _im = utils._EngineIntArrayType(
        im, {'fun_name': fun_name, 'entity': 'im'},
        exp_shape=(_m2_py, _n2_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _m1 = utils._EngineIntScalarType(
        m1, {'fun_name': fun_name, 'entity': 'm1'},
    ).to_ctypes()
    _n1 = utils._EngineIntScalarType(
        n1, {'fun_name': fun_name, 'entity': 'n1'},
    ).to_ctypes()
    _order = utils._EngineCharScalarType(
        order, {'fun_name': fun_name, 'entity': 'order'}, exp_chrlen=1,
    ).to_ctypes()
    _irank_cls = utils._EngineIntArrayType(
        irank, {'fun_name': fun_name, 'entity': 'irank'},
        exp_shape=(_m2_py,),
        check_contents=False,
    )
    _irank = _irank_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _im,
        _ldm,
        _m1,
        _m2,
        _n1,
        _n2,
        _order,
        _irank,
        _errbuf,
        _ierr,
        _order._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m2 = {:s} and ldm = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: m2 <= ldm.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n1 = {:s} and n2 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: n1 <= n2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n1 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n1 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n2 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n2 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m1 = {:s} and m2 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: m1 <= m2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m1 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m1 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m2 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m2 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, order has an illegal value: order = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _irank_cls.refresh_py(_irank)
        return_data = None
    _err_class.manage()
    return return_data

def realmat_rank_columns(rm, m1, m2, n1, n2, order, irank):
    u"""
    Rank columns of a matrix, real numbers.

    ``realmat_rank_columns`` ranks the columns of a matrix of `float`
    numbers in ascending or descending order.

    For full information please refer to the NAG Library document for
    m01dj

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/m01/m01djf.html

    Parameters
    ----------
    rm : float, shape (`m2`, `n2`)
        Rows `m1`-1 to `m2`-1 of columns `n1` to `n2` of `rm` must
        contain `float` data to be ranked.

    m1 : int
        Note: this argument represents an array index; the value you
        supply must be base-1 for compatibility with the NAG Engine.

        The index of the first row of `rm` to be used.

    m2 : int
        Note: this argument represents an array index; the value you
        supply must be base-1 for compatibility with the NAG Engine.

        The index of the last row of `rm` to be used.

    n1 : int
        The index of the first column of `rm` to be ranked.

    n2 : int
        The index of the last column of `rm` to be ranked.

    order : str, length 1
        If `order` = 'A', the columns will be ranked in ascending (i.e.,
        nondecreasing) order.

        If `order` = 'D', into descending order.

    irank : int, shape (`n2`), modified in place
        `On exit`: elements `n1` to `n2` of `irank` contain the ranks of
        the corresponding columns of `rm`. Note that the ranks are in
        the range `n1` to `n2`: thus, if the ith column of `rm` is the
        first in the rank order, `irank`[i-1] is set to `n1`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m2` = *<value>* and ldm = *<value>*.

            Constraint: `m2` <= ldm.

        (`errno` 1)
            On entry, `n1` = *<value>* and `n2` = *<value>*.

            Constraint: `n1` <= `n2`.

        (`errno` 1)
            On entry, `n1` = *<value>*.

            Constraint: `n1` >= 1.

        (`errno` 1)
            On entry, `n2` = *<value>*.

            Constraint: `n2` >= 1.

        (`errno` 1)
            On entry, `m1` = *<value>* and `m2` = *<value>*.

            Constraint: `m1` <= `m2`.

        (`errno` 1)
            On entry, `m1` = *<value>*.

            Constraint: `m1` >= 1.

        (`errno` 1)
            On entry, `m2` = *<value>*.

            Constraint: `m2` >= 1.

        (`errno` 2)
            On entry, `order` has an illegal value: `order` = *<value>*.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Knuth, D E, 1973, `The Art of Computer Programming (Volume 3)`, (2nd
    Edition), Addison--Wesley
    """

    excs = []
    fun_name = 'naginterfaces.base.sort.realmat_rank_columns'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('m01djft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._EngineFloat64ArrayType.c_type, # Input, rm[ldm, n2].
        utils._BYREF_INT, # Input, ldm.
        utils._BYREF_INT, # Input, m1.
        utils._BYREF_INT, # Input, m2.
        utils._BYREF_INT, # Input, n1.
        utils._BYREF_INT, # Input, n2.
        utils._BYREF_STRING, # Input, order(len=1).
        utils._EngineIntArrayType.c_type, # Output, irank[n2].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of order.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((rm, 'rm'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _ldm_py = utils._EngineFloat64ArrayType.get_nd(rm, {'fun_name': fun_name, 'entity': 'rm'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldm = utils._EngineIntScalarType(
        _ldm_py, {'fun_name': fun_name, 'entity': '_ldm'},
    ).to_ctypes()
    _n2 = utils._EngineIntScalarType(
        n2, {'fun_name': fun_name, 'entity': 'n2'},
    ).to_ctypes()
    _m2 = utils._EngineIntScalarType(
        m2, {'fun_name': fun_name, 'entity': 'm2'},
    ).to_ctypes()
    _m2_py = utils._EngineIntScalarType.to_py(_m2)
    _n2_py = utils._EngineIntScalarType.to_py(_n2)
    _rm = utils._EngineFloat64ArrayType(
        rm, {'fun_name': fun_name, 'entity': 'rm'},
        exp_shape=(_m2_py, _n2_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _m1 = utils._EngineIntScalarType(
        m1, {'fun_name': fun_name, 'entity': 'm1'},
    ).to_ctypes()
    _n1 = utils._EngineIntScalarType(
        n1, {'fun_name': fun_name, 'entity': 'n1'},
    ).to_ctypes()
    _order = utils._EngineCharScalarType(
        order, {'fun_name': fun_name, 'entity': 'order'}, exp_chrlen=1,
    ).to_ctypes()
    _irank_cls = utils._EngineIntArrayType(
        irank, {'fun_name': fun_name, 'entity': 'irank'},
        exp_shape=(_n2_py,),
        check_contents=False,
    )
    _irank = _irank_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _rm,
        _ldm,
        _m1,
        _m2,
        _n1,
        _n2,
        _order,
        _irank,
        _errbuf,
        _ierr,
        _order._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m2 = {:s} and ldm = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: m2 <= ldm.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n1 = {:s} and n2 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: n1 <= n2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n1 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n1 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n2 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n2 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m1 = {:s} and m2 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: m1 <= m2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m1 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m1 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m2 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m2 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, order has an illegal value: order = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _irank_cls.refresh_py(_irank)
        return_data = None
    _err_class.manage()
    return return_data

def intmat_rank_columns(im, m1, m2, n1, n2, order, irank):
    u"""
    Rank columns of a matrix, integer numbers.

    ``intmat_rank_columns`` ranks the columns of a matrix of integer
    numbers in ascending or descending order.

    For full information please refer to the NAG Library document for
    m01dk

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/m01/m01dkf.html

    Parameters
    ----------
    im : int, shape (`m2`, `n2`)
        Rows `m1`-1 to `m2`-1 of columns `n1` to `n2` of `im` must
        contain integer data to be ranked.

    m1 : int
        Note: this argument represents an array index; the value you
        supply must be base-1 for compatibility with the NAG Engine.

        The index of the first row of `im` to be used.

    m2 : int
        Note: this argument represents an array index; the value you
        supply must be base-1 for compatibility with the NAG Engine.

        The index of the last row of `im` to be used.

    n1 : int
        The index of the first column of `im` to be ranked.

    n2 : int
        The index of the last column of `im` to be ranked.

    order : str, length 1
        If `order` = 'A', the columns will be ranked in ascending (i.e.,
        nondecreasing) order.

        If `order` = 'D', into descending order.

    irank : int, shape (`n2`), modified in place
        `On exit`: elements `n1` to `n2` of `irank` contain the ranks of
        the corresponding columns of `im`. Note that the ranks are in
        the range `n1` to `n2`: thus, if the ith column of `im` is the
        first in the rank order, `irank`[i-1] is set to `n1`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m2` = *<value>* and ldm = *<value>*.

            Constraint: `m2` <= ldm.

        (`errno` 1)
            On entry, `n1` = *<value>* and `n2` = *<value>*.

            Constraint: `n1` <= `n2`.

        (`errno` 1)
            On entry, `n1` = *<value>*.

            Constraint: `n1` >= 1.

        (`errno` 1)
            On entry, `n2` = *<value>*.

            Constraint: `n2` >= 1.

        (`errno` 1)
            On entry, `m1` = *<value>* and `m2` = *<value>*.

            Constraint: `m1` <= `m2`.

        (`errno` 1)
            On entry, `m1` = *<value>*.

            Constraint: `m1` >= 1.

        (`errno` 1)
            On entry, `m2` = *<value>*.

            Constraint: `m2` >= 1.

        (`errno` 2)
            On entry, `order` has an illegal value: `order` = *<value>*.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Knuth, D E, 1973, `The Art of Computer Programming (Volume 3)`, (2nd
    Edition), Addison--Wesley
    """

    excs = []
    fun_name = 'naginterfaces.base.sort.intmat_rank_columns'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('m01dkft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._EngineIntArrayType.c_type, # Input, im[ldm, n2].
        utils._BYREF_INT, # Input, ldm.
        utils._BYREF_INT, # Input, m1.
        utils._BYREF_INT, # Input, m2.
        utils._BYREF_INT, # Input, n1.
        utils._BYREF_INT, # Input, n2.
        utils._BYREF_STRING, # Input, order(len=1).
        utils._EngineIntArrayType.c_type, # Output, irank[n2].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of order.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((im, 'im'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _ldm_py = utils._EngineIntArrayType.get_nd(im, {'fun_name': fun_name, 'entity': 'im'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldm = utils._EngineIntScalarType(
        _ldm_py, {'fun_name': fun_name, 'entity': '_ldm'},
    ).to_ctypes()
    _n2 = utils._EngineIntScalarType(
        n2, {'fun_name': fun_name, 'entity': 'n2'},
    ).to_ctypes()
    _m2 = utils._EngineIntScalarType(
        m2, {'fun_name': fun_name, 'entity': 'm2'},
    ).to_ctypes()
    _m2_py = utils._EngineIntScalarType.to_py(_m2)
    _n2_py = utils._EngineIntScalarType.to_py(_n2)
    _im = utils._EngineIntArrayType(
        im, {'fun_name': fun_name, 'entity': 'im'},
        exp_shape=(_m2_py, _n2_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _m1 = utils._EngineIntScalarType(
        m1, {'fun_name': fun_name, 'entity': 'm1'},
    ).to_ctypes()
    _n1 = utils._EngineIntScalarType(
        n1, {'fun_name': fun_name, 'entity': 'n1'},
    ).to_ctypes()
    _order = utils._EngineCharScalarType(
        order, {'fun_name': fun_name, 'entity': 'order'}, exp_chrlen=1,
    ).to_ctypes()
    _irank_cls = utils._EngineIntArrayType(
        irank, {'fun_name': fun_name, 'entity': 'irank'},
        exp_shape=(_n2_py,),
        check_contents=False,
    )
    _irank = _irank_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _im,
        _ldm,
        _m1,
        _m2,
        _n1,
        _n2,
        _order,
        _irank,
        _errbuf,
        _ierr,
        _order._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m2 = {:s} and ldm = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: m2 <= ldm.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n1 = {:s} and n2 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: n1 <= n2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n1 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n1 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n2 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n2 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m1 = {:s} and m2 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: m1 <= m2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m1 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m1 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m2 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m2 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, order has an illegal value: order = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _irank_cls.refresh_py(_irank)
        return_data = None
    _err_class.manage()
    return return_data

def arbitrary_rank(compar, m1, m2, irank, data=None):
    u"""
    Rank arbitrary data.

    ``arbitrary_rank`` ranks arbitrary data according to a user-supplied
    comparison function.

    For full information please refer to the NAG Library document for
    m01dz

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/m01/m01dzf.html

    Parameters
    ----------
    compar : callable retval = compar(i, j, data=None)
        `compar` must specify the relative ordering of any two data
        items; it must return True if item `i` must come strictly
        **after** item `j` in the rank ordering.

        Parameters
        ~~~~~~~~~~
        i : int
            `i` and `j` identify the data items to be compared.

        j : int
            `i` and `j` identify the data items to be compared.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        retval : bool
            Must be True if item `i` must come strictly **after** item
            `j` in the rank ordering.

    m1 : int
        Note: this argument represents an array index; the value you
        supply must be base-1 for compatibility with the NAG Engine.

        `m1` must specify the lower bound of the range of data items to
        be ranked, and the range of ranks to be assigned. Specifically,
        ``arbitrary_rank`` ranks the data items identified by integers
        in the range `m1` to `m2`, and assigns ranks in the range `m1`
        to `m2` which are stored in elements `m1`-1 to `m2`-1 of
        `irank`.

    m2 : int
        Note: this argument represents an array index; the value you
        supply must be base-1 for compatibility with the NAG Engine.

        `m2` must specify the upper bound of the range of data items to
        be ranked, and the range of ranks to be assigned. Specifically,
        ``arbitrary_rank`` ranks the data items identified by integers
        in the range `m1` to `m2`, and assigns ranks in the range `m1`
        to `m2` which are stored in elements `m1`-1 to `m2`-1 of
        `irank`.

    irank : int, shape (`m2`), modified in place
        `On exit`: elements `m1`-1 to `m2`-1 of `irank` contain the
        ranks of the data items `m1` to `m2`. Note that the ranks are in
        the range `m1` to `m2`: thus, if item i is first in the rank
        ordering, `irank`[i-1] contains `m1`.

    data : arbitrary, optional
        User-communication data for callback functions.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m1` = *<value>* and `m2` = *<value>*.

            Constraint: `m1` <= `m2`.

        (`errno` 1)
            On entry, `m1` = *<value>*.

            Constraint: `m1` >= 1.

        (`errno` 1)
            On entry, `m2` = *<value>*.

            Constraint: `m2` >= 1.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Knuth, D E, 1973, `The Art of Computer Programming (Volume 3)`, (2nd
    Edition), Addison--Wesley
    """

    excs = []
    fun_name = 'naginterfaces.base.sort.arbitrary_rank'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    cb_compar_name = fun_name + ':compar'

    comparh = utils._EngineState._sys_functype(
        None,
        utils._UNUSED_FUNCTYPE,
        utils._UNUSED_FUNCTYPE,
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, i.
        utils._BYREF_INT, # Input, j.
        utils._BYREF_INT, # Output, iretval.
        utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
    )

    def cb_comparh(_compar, _comparsub, en_data, i, j, iretval, _iuser, _ruser):
        """Helper for compar."""
        try:
            _i_comparh = utils._EngineIntScalarType.to_py(i.contents)
            _j_comparh = utils._EngineIntScalarType.to_py(j.contents)
            if data is not None:
                _retval_comparh = compar(
                    _i_comparh,
                    _j_comparh,
                    data,
                )
            else:
                _retval_comparh = compar(
                    _i_comparh,
                    _j_comparh,
                )
            _retval_comparh_ct = utils._EngineLogicalScalarType(
                _retval_comparh, {'fun_name': cb_compar_name, 'returned': 'retval'},
            ).to_ctypes()
            _retval_comparh_py = utils._EngineLogicalScalarType.to_py(_retval_comparh_ct)
            if _retval_comparh_py is False:
                _iretval_comparh_py = 0
                utils._EngineIntScalarType(
                    _iretval_comparh_py, {'fun_name': cb_compar_name, 'returned': '_iretval'},
                ).refresh_ctypes_buffer(iretval)
            elif _retval_comparh_py is True:
                _iretval_comparh_py = 1
                utils._EngineIntScalarType(
                    _iretval_comparh_py, {'fun_name': cb_compar_name, 'returned': '_iretval'},
                ).refresh_ctypes_buffer(iretval)
            else:
                raise utils.NagValueError(
                    {'fun_name': cb_compar_name, 'returned': 'retval'},
                    'Unexpected value for retval.',
                )
        except:
            excs.append(_sys.exc_info())
            en_data.contents.hlperr = utils._IERR_HELPER

    engine_ctypes = utils._EngineState._get_symbol('m01dzft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._UNUSED_FUNCTYPE,
        utils._UNUSED_FUNCTYPE,
        comparh,
        utils._BYREF_INT, # Input, m1.
        utils._BYREF_INT, # Input, m2.
        utils._EngineIntArrayType.c_type, # Output, irank[m2].
        utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    utils._check_callable(compar, {'fun_name': fun_name, 'entity': 'compar'})
    _compar = utils._UNUSED_CALLBACK
    _comparsub = utils._UNUSED_CALLBACK
    _comparh = utils._cast_callback(
        comparh, cb_comparh,
        {'fun_name': fun_name, 'entity': 'compar'},
    )
    _m1 = utils._EngineIntScalarType(
        m1, {'fun_name': fun_name, 'entity': 'm1'},
    ).to_ctypes()
    _m2 = utils._EngineIntScalarType(
        m2, {'fun_name': fun_name, 'entity': 'm2'},
    ).to_ctypes()
    _m2_py = utils._EngineIntScalarType.to_py(_m2)
    _irank_cls = utils._EngineIntArrayType(
        irank, {'fun_name': fun_name, 'entity': 'irank'},
        exp_shape=(_m2_py,),
        check_contents=False,
    )
    _irank = _irank_cls.to_ctypes(can_reference_contents=False)
    _iuser = None
    _ruser = None
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _compar,
        _comparsub,
        _comparh,
        _m1,
        _m2,
        _irank,
        _iuser,
        _ruser,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m1 = {:s} and m2 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: m1 <= m2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m1 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m1 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m2 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m2 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _irank_cls.refresh_py(_irank)
        return_data = None
    _err_class.manage()
    return return_data

def realvec_rank_rearrange(rv, m1, m2, irank):
    u"""
    Rearrange a vector according to given ranks, real numbers.

    ``realvec_rank_rearrange`` rearranges a vector of real numbers into
    the order specified by a vector of ranks.

    For full information please refer to the NAG Library document for
    m01ea

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/m01/m01eaf.html

    Parameters
    ----------
    rv : float, shape (`m2`), modified in place
        `On entry`: elements `m1`-1 to `m2`-1 of `rv` must contain real
        values to be rearranged.

        `On exit`: these values are rearranged into rank order. For
        example, if `irank`[i-1] = `m1`, then the initial value of
        `rv`[i-1] is moved to `rv`[`m1`-1].

    m1 : int
        Note: this argument represents an array index; the value you
        supply must be base-1 for compatibility with the NAG Engine.

        `m1` and `m2` must specify the range of the ranks supplied in
        `irank` and the elements of `rv` to be rearranged.

    m2 : int
        Note: this argument represents an array index; the value you
        supply must be base-1 for compatibility with the NAG Engine.

        `m1` and `m2` must specify the range of the ranks supplied in
        `irank` and the elements of `rv` to be rearranged.

    irank : int, shape (`m2`), modified in place
        `On entry`: elements `m1`-1 to `m2`-1 of `irank` must contain a
        permutation of the integers `m1` to `m2`, which are interpreted
        as a vector of ranks.

        `On exit`: used as internal workspace prior to being restored
        and hence is unchanged.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m1` = *<value>* and `m2` = *<value>*.

            Constraint: `m1` <= `m2`.

        (`errno` 1)
            On entry, `m1` = *<value>*.

            Constraint: `m1` >= 1.

        (`errno` 1)
            On entry, `m2` = *<value>*.

            Constraint: `m2` >= 1.

        (`errno` 2)
            `irank`[`m1`:`m2`-1] does not contain a permutation of the
            integers `m1` to `m2`. `irank`[I-1] contains an out-of-range
            value: I = *<value>*, `irank`[I-1] = *<value>*.

        (`errno` 3)
            `irank`[`m1`:`m2`-1] does not contain a permutation of the
            integers `m1` to `m2`. `irank` contains a repeated value:
            *<value>*.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.sort.realvec_rank_rearrange'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('m01eaft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._EngineFloat64ArrayType.c_type, # Input/Output, rv[m2].
        utils._BYREF_INT, # Input, m1.
        utils._BYREF_INT, # Input, m2.
        utils._EngineIntArrayType.c_type, # Input/Output, irank[m2].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _m2 = utils._EngineIntScalarType(
        m2, {'fun_name': fun_name, 'entity': 'm2'},
    ).to_ctypes()
    _m2_py = utils._EngineIntScalarType.to_py(_m2)
    _rv_cls = utils._EngineFloat64ArrayType(
        rv, {'fun_name': fun_name, 'entity': 'rv'},
        exp_shape=(_m2_py,),
    )
    _rv = _rv_cls.to_ctypes()
    _m1 = utils._EngineIntScalarType(
        m1, {'fun_name': fun_name, 'entity': 'm1'},
    ).to_ctypes()
    _irank_cls = utils._EngineIntArrayType(
        irank, {'fun_name': fun_name, 'entity': 'irank'},
        exp_shape=(_m2_py,),
    )
    _irank = _irank_cls.to_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _rv,
        _m1,
        _m2,
        _irank,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m1 = {:s} and m2 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: m1 <= m2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m1 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m1 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m2 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m2 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'irank[m1:m2-1] does not contain a permutation of the integers m1 to m2.',
                    'irank[I-1] contains an out-of-range value:',
                    'I = {:s}, irank[I-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'irank[m1:m2-1] does not contain a permutation of the integers m1 to m2.',
                    'irank contains a repeated value: {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _rv_cls.refresh_py(_rv)
        _irank_cls.refresh_py(_irank)
        return_data = None
    _err_class.manage()
    return return_data

def intvec_rank_rearrange(iv, m1, m2, irank):
    u"""
    Rearrange a vector according to given ranks, integer numbers.

    ``intvec_rank_rearrange`` rearranges a vector of integer numbers
    into the order specified by a vector of ranks.

    For full information please refer to the NAG Library document for
    m01eb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/m01/m01ebf.html

    Parameters
    ----------
    iv : int, shape (`m2`), modified in place
        `On entry`: elements `m1`-1 to `m2`-1 of `iv` must contain
        integer values to be rearranged.

        `On exit`: these values are rearranged into rank order. For
        example, if `irank`[i-1] = `m1`, then the initial value of
        `iv`[i-1] is moved to `iv`[`m1`-1].

    m1 : int
        Note: this argument represents an array index; the value you
        supply must be base-1 for compatibility with the NAG Engine.

        `m1` and `m2` specify the range of the ranks supplied in `irank`
        and the elements of `iv` to be rearranged.

    m2 : int
        Note: this argument represents an array index; the value you
        supply must be base-1 for compatibility with the NAG Engine.

        `m1` and `m2` specify the range of the ranks supplied in `irank`
        and the elements of `iv` to be rearranged.

    irank : int, shape (`m2`), modified in place
        `On entry`: elements `m1`-1 to `m2`-1 of `irank` must contain a
        permutation of the integers `m1` to `m2`, which are interpreted
        as a vector of ranks.

        `On exit`: used as internal workspace prior to being restored
        and hence is unchanged.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m1` = *<value>* and `m2` = *<value>*.

            Constraint: `m1` <= `m2`.

        (`errno` 1)
            On entry, `m1` = *<value>*.

            Constraint: `m1` >= 1.

        (`errno` 1)
            On entry, `m2` = *<value>*.

            Constraint: `m2` >= 1.

        (`errno` 2)
            `irank`[`m1`:`m2`-1] does not contain a permutation of the
            integers `m1` to `m2`. `irank`[I-1] contains an out-of-range
            value: I = *<value>*, `irank`[I-1] = *<value>*.

        (`errno` 3)
            `irank`[`m1`:`m2`-1] does not contain a permutation of the
            integers `m1` to `m2`. `irank` contains a repeated value:
            *<value>*.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.sort.intvec_rank_rearrange'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('m01ebft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._EngineIntArrayType.c_type, # Input/Output, iv[m2].
        utils._BYREF_INT, # Input, m1.
        utils._BYREF_INT, # Input, m2.
        utils._EngineIntArrayType.c_type, # Input/Output, irank[m2].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _m2 = utils._EngineIntScalarType(
        m2, {'fun_name': fun_name, 'entity': 'm2'},
    ).to_ctypes()
    _m2_py = utils._EngineIntScalarType.to_py(_m2)
    _iv_cls = utils._EngineIntArrayType(
        iv, {'fun_name': fun_name, 'entity': 'iv'},
        exp_shape=(_m2_py,),
    )
    _iv = _iv_cls.to_ctypes()
    _m1 = utils._EngineIntScalarType(
        m1, {'fun_name': fun_name, 'entity': 'm1'},
    ).to_ctypes()
    _irank_cls = utils._EngineIntArrayType(
        irank, {'fun_name': fun_name, 'entity': 'irank'},
        exp_shape=(_m2_py,),
    )
    _irank = _irank_cls.to_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _iv,
        _m1,
        _m2,
        _irank,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m1 = {:s} and m2 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: m1 <= m2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m1 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m1 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m2 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m2 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'irank[m1:m2-1] does not contain a permutation of the integers m1 to m2.',
                    'irank[I-1] contains an out-of-range value:',
                    'I = {:s}, irank[I-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'irank[m1:m2-1] does not contain a permutation of the integers m1 to m2.',
                    'irank contains a repeated value: {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _iv_cls.refresh_py(_iv)
        _irank_cls.refresh_py(_irank)
        return_data = None
    _err_class.manage()
    return return_data

def charvec_rank_rearrange(ch, m1, m2, irank):
    u"""
    Rearrange a vector according to given ranks, character data.

    ``charvec_rank_rearrange`` rearranges a vector of character data
    into the order specified by a vector of ranks.

    For full information please refer to the NAG Library document for
    m01ec

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/m01/m01ecf.html

    Parameters
    ----------
    ch : str, shape (`m2`), modified in place
        `On entry`: elements `m1`-1 to `m2`-1 of `ch` must contain
        character data to be rearranged.

        `On exit`: these values are rearranged into rank order. For
        example, if `irank`[i-1] = `m1`, then the initial value of
        `ch`[i-1] is moved to `ch`[`m1`-1].

    m1 : int
        Note: this argument represents an array index; the value you
        supply must be base-1 for compatibility with the NAG Engine.

        The range of the ranks supplied in `irank` and the elements of
        `ch` to be rearranged.

    m2 : int
        Note: this argument represents an array index; the value you
        supply must be base-1 for compatibility with the NAG Engine.

        The range of the ranks supplied in `irank` and the elements of
        `ch` to be rearranged.

    irank : int, shape (`m2`), modified in place
        `On entry`: elements `m1`-1 to `m2`-1 of `irank` must contain a
        permutation of the integers `m1` to `m2`, which are interpreted
        as a vector of ranks.

        `On exit`: used as internal workspace prior to being restored
        and hence is unchanged.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m1` = *<value>* and `m2` = *<value>*.

            Constraint: `m1` <= `m2`.

        (`errno` 1)
            On entry, `m1` = *<value>*.

            Constraint: `m1` >= 1.

        (`errno` 1)
            On entry, `m2` = *<value>*.

            Constraint: `m2` >= 1.

        (`errno` 2)
            On entry, len(`ch`[0]) = *<value>*.

            Constraint: len(`ch`[0]) <= 255.

        (`errno` 3)
            `irank`[`m1`:`m2`-1] does not contain a permutation of the
            integers `m1` to `m2`. `irank`[I-1] contains an out-of-range
            value: I = *<value>*, `irank`[I-1] = *<value>*.

        (`errno` 4)
            `irank`[`m1`:`m2`-1] does not contain a permutation of the
            integers `m1` to `m2`. `irank` contains a repeated value:
            *<value>*.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.sort.charvec_rank_rearrange'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('m01ecft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._EngineCharArrayType.c_type, # Input/Output, ch(len=*)[m2].
        utils._BYREF_INT, # Input, m1.
        utils._BYREF_INT, # Input, m2.
        utils._EngineIntArrayType.c_type, # Input/Output, irank[m2].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of ch.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _m2 = utils._EngineIntScalarType(
        m2, {'fun_name': fun_name, 'entity': 'm2'},
    ).to_ctypes()
    _m2_py = utils._EngineIntScalarType.to_py(_m2)
    _ch_cls = utils._EngineCharArrayType(
        ch, {'fun_name': fun_name, 'entity': 'ch'}, exp_chrlen=None,
        exp_shape=(_m2_py,),
    )
    _ch = _ch_cls.to_ctypes()
    _ch_chrlen = _ch_cls.chrlen
    _m1 = utils._EngineIntScalarType(
        m1, {'fun_name': fun_name, 'entity': 'm1'},
    ).to_ctypes()
    _irank_cls = utils._EngineIntArrayType(
        irank, {'fun_name': fun_name, 'entity': 'irank'},
        exp_shape=(_m2_py,),
    )
    _irank = _irank_cls.to_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _ch,
        _m1,
        _m2,
        _irank,
        _errbuf,
        _ierr,
        _ch_chrlen,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m1 = {:s} and m2 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: m1 <= m2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m1 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m1 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m2 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m2 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, len(ch[0]) = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: len(ch[0]) <= 255.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'irank[m1:m2-1] does not contain a permutation of the integers m1 to m2.',
                    'irank[I-1] contains an out-of-range value:',
                    'I = {:s}, irank[I-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'irank[m1:m2-1] does not contain a permutation of the integers m1 to m2.',
                    'irank contains a repeated value: {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _ch_cls.refresh_py(_ch)
        _irank_cls.refresh_py(_irank)
        return_data = None
    _err_class.manage()
    return return_data

def cmplxvec_rank_rearrange(cv, m1, m2, irank):
    u"""
    Rearrange a vector according to given ranks, complex numbers.

    ``cmplxvec_rank_rearrange`` rearranges a vector of complex numbers
    into the order specified by a vector of ranks.

    For full information please refer to the NAG Library document for
    m01ed

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/m01/m01edf.html

    Parameters
    ----------
    cv : complex, shape (`m2`), modified in place
        `On entry`: elements `m1`-1 to `m2`-1 of `cv` must contain
        complex values to be rearranged.

        `On exit`: these values are rearranged into rank order. For
        example, if `irank`[i-1] = `m1`, then the initial value of
        `cv`[i-1] is moved to `cv`[`m1`-1].

    m1 : int
        Note: this argument represents an array index; the value you
        supply must be base-1 for compatibility with the NAG Engine.

        `m1` and `m2` must specify the range of the ranks supplied in
        `irank` and the elements of `cv` to be rearranged.

    m2 : int
        Note: this argument represents an array index; the value you
        supply must be base-1 for compatibility with the NAG Engine.

        `m1` and `m2` must specify the range of the ranks supplied in
        `irank` and the elements of `cv` to be rearranged.

    irank : int, shape (`m2`), modified in place
        `On entry`: elements `m1`-1 to `m2`-1 of `irank` must contain a
        permutation of the integers `m1` to `m2`, which are interpreted
        as a vector of ranks.

        `On exit`: used as internal workspace prior to being restored
        and hence is unchanged.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m1` = *<value>* and `m2` = *<value>*.

            Constraint: `m1` <= `m2`.

        (`errno` 1)
            On entry, `m1` = *<value>*.

            Constraint: `m1` >= 1.

        (`errno` 1)
            On entry, `m2` = *<value>*.

            Constraint: `m2` >= 1.

        (`errno` 2)
            `irank`[`m1`:`m2`-1] does not contain a permutation of the
            integers `m1` to `m2`. `irank`[I-1] contains an out-of-range
            value: I = *<value>*, `irank`[I-1] = *<value>*.

        (`errno` 3)
            `irank`[`m1`:`m2`-1] does not contain a permutation of the
            integers `m1` to `m2`. `irank` contains a repeated value:
            *<value>*.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.sort.cmplxvec_rank_rearrange'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('m01edft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._EngineComplex128ArrayType.c_type, # Input/Output, cv[m2].
        utils._BYREF_INT, # Input, m1.
        utils._BYREF_INT, # Input, m2.
        utils._EngineIntArrayType.c_type, # Input/Output, irank[m2].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _m2 = utils._EngineIntScalarType(
        m2, {'fun_name': fun_name, 'entity': 'm2'},
    ).to_ctypes()
    _m2_py = utils._EngineIntScalarType.to_py(_m2)
    _cv_cls = utils._EngineComplex128ArrayType(
        cv, {'fun_name': fun_name, 'entity': 'cv'},
        exp_shape=(_m2_py,),
    )
    _cv = _cv_cls.to_ctypes()
    _m1 = utils._EngineIntScalarType(
        m1, {'fun_name': fun_name, 'entity': 'm1'},
    ).to_ctypes()
    _irank_cls = utils._EngineIntArrayType(
        irank, {'fun_name': fun_name, 'entity': 'irank'},
        exp_shape=(_m2_py,),
    )
    _irank = _irank_cls.to_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _cv,
        _m1,
        _m2,
        _irank,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m1 = {:s} and m2 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: m1 <= m2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m1 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m1 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m2 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m2 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'irank[m1:m2-1] does not contain a permutation of the integers m1 to m2.',
                    'irank[I-1] contains an out-of-range value:',
                    'I = {:s}, irank[I-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'irank[m1:m2-1] does not contain a permutation of the integers m1 to m2.',
                    'irank contains a repeated value: {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _cv_cls.refresh_py(_cv)
        _irank_cls.refresh_py(_irank)
        return_data = None
    _err_class.manage()
    return return_data

def realvec_search(valid, rv, m1, m2, item):
    u"""
    Binary search in set of real numbers.

    ``realvec_search`` searches an ordered vector of `float` numbers and
    returns the index of the first value equal to the sought-after item.

    For full information please refer to the NAG Library document for
    m01na

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/m01/m01naf.html

    Parameters
    ----------
    valid : bool
        If `valid` is set to True argument checking will be performed.
        If `valid` is set to False ``realvec_search`` will be called
        without argument checking (which includes checking that array
        `rv` is sorted in ascending order) and the function will return
        with no exception or warning is raised. See [Further Comments]
        for further details.

    rv : float, shape (`m2`)
        Elements `m1`-1 to `m2`-1 contain `float` values to be searched.

    m1 : int
        Note: this argument represents an array index; the value you
        supply must be base-1 for compatibility with the NAG Engine.

        The index of the first element of `rv` to be searched.

    m2 : int
        Note: this argument represents an array index; the value you
        supply must be base-1 for compatibility with the NAG Engine.

        The index of the last element of `rv` to be searched.

    item : float
        The sought-after item.

    Returns
    -------
    irv : int
        Note: this argument represents an array index; the value
        returned will be base-1.

        The index of the first value equal to the sought-after item.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, `rv` must be sorted in ascending order: `rv`
            element *<value>* > element *<value>*.

        (`errno` 3)
            On entry, `m1` = *<value>*.

            Constraint: `m1` >= 1.

        (`errno` 4)
            On entry, `m1` = *<value>*, `m2` = *<value>*.

            Constraint: `m1` <= `m2`.

    References
    ----------
    Wirth, N, 2004, `Algorithms and Data Structures`, 35--36, Prentice
    Hall
    """

    excs = []
    fun_name = 'naginterfaces.base.sort.realvec_search'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('m01naft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_LOGICAL, # Input, valid.
        utils._EngineFloat64ArrayType.c_type, # Input, rv[m2].
        utils._BYREF_INT, # Input, m1.
        utils._BYREF_INT, # Input, m2.
        utils._BYREF_FLOAT64, # Input, item.
        utils._BYREF_INT, # Output, irv.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _valid = utils._EngineLogicalScalarType(
        valid, {'fun_name': fun_name, 'entity': 'valid'},
    ).to_ctypes()
    _m2 = utils._EngineIntScalarType(
        m2, {'fun_name': fun_name, 'entity': 'm2'},
    ).to_ctypes()
    _m2_py = utils._EngineIntScalarType.to_py(_m2)
    _rv = utils._EngineFloat64ArrayType(
        rv, {'fun_name': fun_name, 'entity': 'rv'},
        exp_shape=(_m2_py,),
    ).to_ctypes()
    _m1 = utils._EngineIntScalarType(
        m1, {'fun_name': fun_name, 'entity': 'm1'},
    ).to_ctypes()
    _item = utils._EngineFloat64ScalarType(
        item, {'fun_name': fun_name, 'entity': 'item'},
    ).to_ctypes()
    _irv = utils._EngineIntScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _valid,
        _rv,
        _m1,
        _m2,
        _item,
        _irv,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, rv must be sorted in ascending order:',
                    'rv element {:s} > element {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m1 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m1 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m1 = {:s}, m2 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: m1 <= m2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _irv_py = utils._EngineIntScalarType.to_py(_irv)
        return_data = _irv_py
    _err_class.manage()
    return return_data

def intvec_search(valid, iv, m1, m2, item):
    u"""
    Binary search in set of integer numbers.

    ``intvec_search`` searches an ordered vector of integer numbers and
    returns the index of the first value equal to the sought-after item.

    For full information please refer to the NAG Library document for
    m01nb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/m01/m01nbf.html

    Parameters
    ----------
    valid : bool
        If `valid` is set to True argument checking will be performed.
        If `valid` is set to False ``intvec_search`` will be called
        without argument checking (which includes checking that array
        `iv` is sorted in ascending order) and the function will return
        with no exception or warning is raised. See [Further Comments]
        for further details.

    iv : int, shape (`m2`)
        Elements `m1`-1 to `m2`-1 contain integer values to be searched.

    m1 : int
        Note: this argument represents an array index; the value you
        supply must be base-1 for compatibility with the NAG Engine.

        The index of the first element of `iv` to be searched.

    m2 : int
        Note: this argument represents an array index; the value you
        supply must be base-1 for compatibility with the NAG Engine.

        The index of the last element of `iv` to be searched.

    item : int
        The sought-after item.

    Returns
    -------
    iiv : int
        Note: this argument represents an array index; the value
        returned will be base-1.

        The index of the first value equal to the sought-after item.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, `iv` must be sorted in ascending order: `iv`
            element *<value>* > element *<value>*.

        (`errno` 3)
            On entry, `m1` = *<value>*.

            Constraint: `m1` >= 1.

        (`errno` 4)
            On entry, `m1` = *<value>*, `m2` = *<value>*.

            Constraint: `m1` <= `m2`.

    References
    ----------
    Wirth, N, 2004, `Algorithms and Data Structures`, 35--36, Prentice
    Hall
    """

    excs = []
    fun_name = 'naginterfaces.base.sort.intvec_search'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('m01nbft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_LOGICAL, # Input, valid.
        utils._EngineIntArrayType.c_type, # Input, iv[m2].
        utils._BYREF_INT, # Input, m1.
        utils._BYREF_INT, # Input, m2.
        utils._BYREF_INT, # Input, item.
        utils._BYREF_INT, # Output, iiv.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _valid = utils._EngineLogicalScalarType(
        valid, {'fun_name': fun_name, 'entity': 'valid'},
    ).to_ctypes()
    _m2 = utils._EngineIntScalarType(
        m2, {'fun_name': fun_name, 'entity': 'm2'},
    ).to_ctypes()
    _m2_py = utils._EngineIntScalarType.to_py(_m2)
    _iv = utils._EngineIntArrayType(
        iv, {'fun_name': fun_name, 'entity': 'iv'},
        exp_shape=(_m2_py,),
    ).to_ctypes()
    _m1 = utils._EngineIntScalarType(
        m1, {'fun_name': fun_name, 'entity': 'm1'},
    ).to_ctypes()
    _item = utils._EngineIntScalarType(
        item, {'fun_name': fun_name, 'entity': 'item'},
    ).to_ctypes()
    _iiv = utils._EngineIntScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _valid,
        _iv,
        _m1,
        _m2,
        _item,
        _iiv,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iv must be sorted in ascending order:',
                    'iv element {:s} > element {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m1 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m1 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m1 = {:s}, m2 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: m1 <= m2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _iiv_py = utils._EngineIntScalarType.to_py(_iiv)
        return_data = _iiv_py
    _err_class.manage()
    return return_data

def charvec_search(valid, ch, m1, m2, item):
    u"""
    Binary search in set of character data.

    ``charvec_search`` examines an ordered vector of strings and returns
    the index of the first value equal to the sought-after item.
    Character items are compared according to the ASCII collating
    sequence.

    For full information please refer to the NAG Library document for
    m01nc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/m01/m01ncf.html

    Parameters
    ----------
    valid : bool
        If `valid` is set to True argument checking will be performed.
        If `valid` is set to False ``charvec_search`` will be called
        without argument checking, which includes checking that array
        `ch` is sorted in ascending order and the function will return
        with no exception or warning is raised. See [Further Comments]
        for further details.

    ch : str, shape (`m2`)
        Elements `m1`-1 to `m2`-1 contain character data to be searched.

    m1 : int
        Note: this argument represents an array index; the value you
        supply must be base-1 for compatibility with the NAG Engine.

        The index of the first element of `ch` to be searched.

    m2 : int
        Note: this argument represents an array index; the value you
        supply must be base-1 for compatibility with the NAG Engine.

        The index of the last element of `ch` to be searched.

    item : str
        The sought-after item. Trailing space characters are ignored.

    Returns
    -------
    ich : int
        Note: this argument represents an array index; the value
        returned will be base-1.

        The index of the first value equal to the sought-after item.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, `ch` must be sorted in ascending order: `ch`
            element *<value>* > element *<value>*.

        (`errno` 3)
            On entry, `m1` = *<value>*.

            Constraint: `m1` >= 1.

        (`errno` 4)
            On entry, `m1` = *<value>*, `m2` = *<value>*.

            Constraint: `m2` >= `m1`.

        (`errno` 5)
            On entry, the length of each element of `ch` must be at most
            255: maximum string length = *<value>*.

    References
    ----------
    Wirth, N, 2004, `Algorithms and Data Structures`, 35--36, Prentice
    Hall
    """

    excs = []
    fun_name = 'naginterfaces.base.sort.charvec_search'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('m01ncft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_LOGICAL, # Input, valid.
        utils._EngineCharArrayType.c_type, # Input, ch(len=*)[m2].
        utils._BYREF_INT, # Input, m1.
        utils._BYREF_INT, # Input, m2.
        utils._BYREF_STRING, # Input, item(len=*).
        utils._BYREF_INT, # Output, ich.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of ch.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of item.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _valid = utils._EngineLogicalScalarType(
        valid, {'fun_name': fun_name, 'entity': 'valid'},
    ).to_ctypes()
    _m2 = utils._EngineIntScalarType(
        m2, {'fun_name': fun_name, 'entity': 'm2'},
    ).to_ctypes()
    _m2_py = utils._EngineIntScalarType.to_py(_m2)
    _ch_cls = utils._EngineCharArrayType(
        ch, {'fun_name': fun_name, 'entity': 'ch'}, exp_chrlen=None,
        exp_shape=(_m2_py,),
    )
    _ch = _ch_cls.to_ctypes()
    _ch_chrlen = _ch_cls.chrlen
    _m1 = utils._EngineIntScalarType(
        m1, {'fun_name': fun_name, 'entity': 'm1'},
    ).to_ctypes()
    _item = utils._EngineCharScalarType(
        item, {'fun_name': fun_name, 'entity': 'item'}, exp_chrlen=None,
    ).to_ctypes()
    _ich = utils._EngineIntScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _valid,
        _ch,
        _m1,
        _m2,
        _item,
        _ich,
        _errbuf,
        _ierr,
        _ch_chrlen,
        _item._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ch must be sorted in ascending order:',
                    'ch element {:s} > element {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m1 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m1 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m1 = {:s}, m2 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: m2 >= m1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the length of each element of ch must be at most 255:',
                    'maximum string length = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _ich_py = utils._EngineIntScalarType.to_py(_ich)
        return_data = _ich_py
    _err_class.manage()
    return return_data

def realvec_vec_search(valid, rv, n, m1, m2, item, m, idx, comm):
    u"""
    Searches an ordered set of real numbers using an O(1) method.

    ``realvec_vec_search`` searches a strictly ordered vector of `float`
    numbers and returns the indices of the largest numbers in the vector
    which are smaller than or equal to the sought-after items.

    For full information please refer to the NAG Library document for
    m01nd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/m01/m01ndf.html

    Parameters
    ----------
    valid : bool
        If `valid` is set to True argument checking will be performed.
        If `valid` is set to False ``realvec_vec_search`` will be called
        without argument checking (which includes checking that array
        `rv` is sorted in strictly ascending order). See [Further
        Comments] for further details.

    rv : float, shape (`n`)
        X, the `float` values to be searched.

    n : int
        n, the number of `float` values to be searched.

    m1 : int
        Note: this argument represents an array index; the value you
        supply must be base-1 for compatibility with the NAG Engine.

        The index of the first element of `rv` to be searched.

    m2 : int
        Note: this argument represents an array index; the value you
        supply must be base-1 for compatibility with the NAG Engine.

        The index of the last element of `rv` to be searched.

    item : float, shape (`m`)
        Z, the sought-after items.

    m : int
        m, the number of items sought.

    idx : int, shape (`m`), modified in place
        Note: this argument represents array indices; the values
        returned will be base-1.

        `On exit`: `idx`[i-1] contains the index of the largest entry in
        `rv` which is smaller than or equal to `item`[i-1].

    comm : None or dict, communication object, modified in place
        Communication structure.

        If `comm` is None then searching is performed using an
        offset-based binary search.

        Otherwise, `comm` may be supplied uninitialized.

        It will then be populated and the routine will perform a direct
        search of `rv`.

        The populated `comm` can then be supplied again to subsequent
        calls for repeated searching of `rv`.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, `rv` must be sorted in strictly ascending order:
            `rv` element *<value>* > element *<value>*.

        (`errno` 3)
            On entry, `m1` = *<value>*.

            Constraint: `m1` >= 0.

        (`errno` 4)
            On entry, `m1` = *<value>*, `m2` = *<value>*, `n` =
            *<value>*.

            Constraint: `m1` <= `m2` <= `n`.

        (`errno` 5)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 2.

        (`errno` 6)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 7)
            On entry, the values in `rv` are such that the required
            value of lk would overflow the current platform's largest
            positive integer value.

        (`errno` 8)
            `comm` may have become corrupted.

    References
    ----------
    Cannizzo, F, 2018, `A fast and vectorizable alternative to binary
    search in O(1) with wide applicability to arrays of floating point
    numbers`, Journal of Parallel and Distributed Computing (113),
    37--54
    """

    excs = []
    fun_name = 'naginterfaces.base.sort.realvec_vec_search'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('m01ndft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_LOGICAL, # Input, valid.
        utils._BYREF_INT, # Input, mode.
        utils._EngineFloat64ArrayType.c_type, # Input, rv[n].
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, m1.
        utils._BYREF_INT, # Input, m2.
        utils._EngineFloat64ArrayType.c_type, # Input, item[m].
        utils._BYREF_INT, # Input, m.
        utils._EngineIntArrayType.c_type, # Output, idx[m].
        utils._BYREF_FLOAT64, # Input/Output, h.
        utils._EngineIntArrayType.c_type, # Input/Output, k[lk].
        utils._BYREF_INT, # Input/Output, lk.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _valid = utils._EngineLogicalScalarType(
        valid, {'fun_name': fun_name, 'entity': 'valid'},
    ).to_ctypes()
    _comm_needs_init = utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        opt_keys=('h', 'k'),
        allow_none=True,
    )
    if comm is None:
        _mode = utils._EngineIntScalarType.c_type(3)
    elif _comm_needs_init:
        _mode = utils._EngineIntScalarType.c_type(0)
    else:
        _mode = utils._EngineIntScalarType.c_type(2)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _rv = utils._EngineFloat64ArrayType(
        rv, {'fun_name': fun_name, 'entity': 'rv'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _m1 = utils._EngineIntScalarType(
        m1, {'fun_name': fun_name, 'entity': 'm1'},
    ).to_ctypes()
    _m2 = utils._EngineIntScalarType(
        m2, {'fun_name': fun_name, 'entity': 'm2'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _item = utils._EngineFloat64ArrayType(
        item, {'fun_name': fun_name, 'entity': 'item'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _idx_cls = utils._EngineIntArrayType(
        idx, {'fun_name': fun_name, 'entity': 'idx'},
        exp_shape=(_m_py,),
        check_contents=False,
    )
    _idx = _idx_cls.to_ctypes(can_reference_contents=False)
    _vl_dummy_h = utils._EngineFloat64ScalarType.empty_ctypes()
    if comm is None:
        _h = _vl_dummy_h
    else:
        if _comm_needs_init:
            _h = utils._EngineFloat64ScalarType.empty_ctypes()
            comm['h'] = _h
        else:
            _h = comm['h']
    _vl_dummy_k = utils._EngineIntArrayType.empty_ctypes(0)
    if comm is None:
        _k = _vl_dummy_k
    else:
        if _comm_needs_init:
            _k = _vl_dummy_k
            comm['k'] = _k
        else:
            _k = comm['k']
    if comm is None:
        _lk_py = utils._EngineIntArrayType.get_shape(_vl_dummy_k, {'fun_name': fun_name, 'entity': '_vl_dummy_k'}, exp_rank=1)[0]
        _lk = utils._EngineIntScalarType(
            _lk_py, {'fun_name': fun_name, 'entity': '_lk'},
        ).to_ctypes()
    else:
        if _comm_needs_init:
            _lk_py = -1
            _lk = utils._EngineIntScalarType(
                _lk_py, {'fun_name': fun_name, 'entity': '_lk'},
            ).to_ctypes()
        else:
            _lk_py = utils._EngineIntArrayType.get_shape(_k, {'fun_name': fun_name, 'entity': '_k'}, exp_rank=1, validate=False)[0]
            _lk = utils._EngineIntScalarType(
                _lk_py, {'fun_name': fun_name, 'entity': '_lk'},
            ).to_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _valid,
        _mode,
        _rv,
        _n,
        _m1,
        _m2,
        _item,
        _m,
        _idx,
        _h,
        _k,
        _lk,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    if _ierr.value == utils._IERR_SUCCESS:
        if comm is not None and _comm_needs_init:
            _lk_py = utils._EngineIntScalarType.to_py(_lk)
            _k = utils._EngineIntArrayType.empty_ctypes(_lk_py)
            comm['k'] = _k
            _valid = utils._EngineIntScalarType.c_type(0)
            _mode = utils._EngineIntScalarType.c_type(1)
            engine_ctypes(
                _en_data,
                _valid,
                _mode,
                _rv,
                _n,
                _m1,
                _m2,
                _item,
                _m,
                _idx,
                _h,
                _k,
                _lk,
                _errbuf,
                _ierr,
                _errbuf._length_,
            )
            utils._handle_cb_exception(excs)
            if _ierr.value == utils._IERR_SUCCESS:
                _mode = utils._EngineIntScalarType.c_type(2)
                engine_ctypes(
                    _en_data,
                    _valid,
                    _mode,
                    _rv,
                    _n,
                    _m1,
                    _m2,
                    _item,
                    _m,
                    _idx,
                    _h,
                    _k,
                    _lk,
                    _errbuf,
                    _ierr,
                    _errbuf._length_,
                )
                utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 2 and _ifmt == 1002:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, rv must be sorted in strictly ascending order:',
                    'rv element {:s} > element {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 1003:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m1 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m1 >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 1004:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m1 = {:s}, m2 = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: m1 <= m2 <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 1005:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 1006:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 1007:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the values in rv are such that the required value of lk would',
                    'overflow the current platform\'s largest positive integer value.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 1008:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'comm may have become corrupted.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _idx_cls.refresh_py(_idx)
        return_data = None
    _err_class.manage()
    return return_data

def permute_invert(iperm, m1, m2):
    u"""
    Invert a permutation.

    ``permute_invert`` inverts a permutation, and hence converts a rank
    vector to an index vector, or vice versa.

    For full information please refer to the NAG Library document for
    m01za

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/m01/m01zaf.html

    Parameters
    ----------
    iperm : int, shape (`m2`), modified in place
        `On entry`: elements `m1`-1 to `m2`-1 of `iperm` must contain a
        permutation of the integers `m1` to `m2`.

        `On exit`: these elements contain the inverse permutation of the
        integers `m1` to `m2`.

    m1 : int
        Note: this argument represents an array index; the value you
        supply must be base-1 for compatibility with the NAG Engine.

        `m1` and `m2` must specify the range of elements used in the
        array `iperm` and the range of values in the permutation, as
        specified under `iperm`.

    m2 : int
        Note: this argument represents an array index; the value you
        supply must be base-1 for compatibility with the NAG Engine.

        `m1` and `m2` must specify the range of elements used in the
        array `iperm` and the range of values in the permutation, as
        specified under `iperm`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m1` = *<value>* and `m2` = *<value>*.

            Constraint: `m1` <= `m2`.

        (`errno` 1)
            On entry, `m1` = *<value>*.

            Constraint: `m1` >= 1.

        (`errno` 1)
            On entry, `m2` = *<value>*.

            Constraint: `m2` >= 1.

        (`errno` 2)
            `iperm`[`m1`:`m2`-1] does not contain a permutation of the
            integers `m1` to `m2`. `iperm`[I-1] contains an out-of-range
            value: I = *<value>*, `iperm`[I-1] = *<value>*.

        (`errno` 3)
            `iperm`[`m1`:`m2`-1] does not contain a permutation of the
            integers `m1` to `m2`. `iperm` contains a repeated value:
            *<value>*.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    excs = []
    fun_name = 'naginterfaces.base.sort.permute_invert'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('m01zaft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._EngineIntArrayType.c_type, # Input/Output, iperm[m2].
        utils._BYREF_INT, # Input, m1.
        utils._BYREF_INT, # Input, m2.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _m2 = utils._EngineIntScalarType(
        m2, {'fun_name': fun_name, 'entity': 'm2'},
    ).to_ctypes()
    _m2_py = utils._EngineIntScalarType.to_py(_m2)
    _iperm_cls = utils._EngineIntArrayType(
        iperm, {'fun_name': fun_name, 'entity': 'iperm'},
        exp_shape=(_m2_py,),
    )
    _iperm = _iperm_cls.to_ctypes()
    _m1 = utils._EngineIntScalarType(
        m1, {'fun_name': fun_name, 'entity': 'm1'},
    ).to_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _iperm,
        _m1,
        _m2,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m1 = {:s} and m2 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: m1 <= m2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m1 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m1 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m2 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m2 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'iperm[m1:m2-1] does not contain a permutation of the integers m1 to m2.',
                    'iperm[I-1] contains an out-of-range value:',
                    'I = {:s}, iperm[I-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'iperm[m1:m2-1] does not contain a permutation of the integers m1 to m2.',
                    'iperm contains a repeated value: {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _iperm_cls.refresh_py(_iperm)
        return_data = None
    _err_class.manage()
    return return_data

def permute_check(iperm, m1, m2):
    u"""
    Check validity of a permutation.

    ``permute_check`` checks the validity of a permutation.

    For full information please refer to the NAG Library document for
    m01zb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/m01/m01zbf.html

    Parameters
    ----------
    iperm : int, shape (`m2`), modified in place
        `On entry`: elements `m1`-1 to `m2`-1 of `iperm` must be set to
        values which are supposed to be a permutation of the integers
        `m1` to `m2`. If they are a valid permutation, the function
        exits with no exception or warning is raised.

        `On exit`: used as internal workpsace prior to being restored
        and hence is unchanged.

    m1 : int
        Note: this argument represents an array index; the value you
        supply must be base-1 for compatibility with the NAG Engine.

        The range of elements used in the array `iperm` and the range of
        values in the permutation, as specified under `iperm`.

    m2 : int
        Note: this argument represents an array index; the value you
        supply must be base-1 for compatibility with the NAG Engine.

        The range of elements used in the array `iperm` and the range of
        values in the permutation, as specified under `iperm`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m1` = *<value>* and `m2` = *<value>*.

            Constraint: `m1` <= `m2`.

        (`errno` 1)
            On entry, `m1` = *<value>*.

            Constraint: `m1` >= 1.

        (`errno` 1)
            On entry, `m2` = *<value>*.

            Constraint: `m2` >= 1.

        (`errno` 2)
            `iperm`[`m1`:`m2`-1] does not contain a permutation of the
            integers `m1` to `m2`. `iperm`[I-1] contains an out-of-range
            value: I = *<value>*, `iperm`[I-1] = *<value>*.

        (`errno` 3)
            `iperm`[`m1`:`m2`-1] does not contain a permutation of the
            integers `m1` to `m2`. `iperm` contains a repeated value:
            *<value>*.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.sort.permute_check'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('m01zbft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._EngineIntArrayType.c_type, # Input/Output, iperm[m2].
        utils._BYREF_INT, # Input, m1.
        utils._BYREF_INT, # Input, m2.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _m2 = utils._EngineIntScalarType(
        m2, {'fun_name': fun_name, 'entity': 'm2'},
    ).to_ctypes()
    _m2_py = utils._EngineIntScalarType.to_py(_m2)
    _iperm_cls = utils._EngineIntArrayType(
        iperm, {'fun_name': fun_name, 'entity': 'iperm'},
        exp_shape=(_m2_py,),
    )
    _iperm = _iperm_cls.to_ctypes()
    _m1 = utils._EngineIntScalarType(
        m1, {'fun_name': fun_name, 'entity': 'm1'},
    ).to_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _iperm,
        _m1,
        _m2,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m1 = {:s} and m2 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: m1 <= m2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m1 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m1 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m2 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m2 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'iperm[m1:m2-1] does not contain a permutation of the integers m1 to m2.',
                    'iperm[I-1] contains an out-of-range value:',
                    'I = {:s}, iperm[I-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'iperm[m1:m2-1] does not contain a permutation of the integers m1 to m2.',
                    'iperm contains a repeated value: {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _iperm_cls.refresh_py(_iperm)
        return_data = None
    _err_class.manage()
    return return_data

def permute_decompose(iperm, m1, m2, icycl):
    u"""
    Decompose a permutation into cycles.

    ``permute_decompose`` decomposes a permutation into cycles, as an
    aid to reordering ranked data.

    For full information please refer to the NAG Library document for
    m01zc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/m01/m01zcf.html

    Parameters
    ----------
    iperm : int, shape (`m2`), modified in place
        `On entry`: elements `m1`-1 to `m2`-1 of `iperm` must contain a
        permutation of the integers `m1` to `m2`.

        `On exit`: is used as internal workpsace prior to being restored
        and hence is unchanged.

    m1 : int
        Note: this argument represents an array index; the value you
        supply must be base-1 for compatibility with the NAG Engine.

        `m1` and `m2` must specify the range of elements used in the
        array `iperm` and the range of values in the permutation, as
        specified under `iperm`.

    m2 : int
        Note: this argument represents an array index; the value you
        supply must be base-1 for compatibility with the NAG Engine.

        `m1` and `m2` must specify the range of elements used in the
        array `iperm` and the range of values in the permutation, as
        specified under `iperm`.

    icycl : int, shape (`m2`), modified in place
        `On exit`: elements `m1`-1 to `m2`-1 of `icycl` contain a
        representation of the permutation as a list of cycles, with the
        first integer in each cycle negated. (See [Description].)

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m1` = *<value>* and `m2` = *<value>*.

            Constraint: `m1` <= `m2`.

        (`errno` 1)
            On entry, `m1` = *<value>*.

            Constraint: `m1` >= 1.

        (`errno` 1)
            On entry, `m2` = *<value>*.

            Constraint: `m2` >= 1.

        (`errno` 2)
            `iperm`[`m1`:`m2`-1] does not contain a permutation of the
            integers `m1` to `m2`. `iperm`[I-1] contains an out-of-range
            value: I = *<value>*, `iperm`[I-1] = *<value>*.

        (`errno` 3)
            `iperm`[`m1`:`m2`-1] does not contain a permutation of the
            integers `m1` to `m2`. `iperm` contains a repeated value:
            *<value>*.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    excs = []
    fun_name = 'naginterfaces.base.sort.permute_decompose'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('m01zcft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._EngineIntArrayType.c_type, # Input/Output, iperm[m2].
        utils._BYREF_INT, # Input, m1.
        utils._BYREF_INT, # Input, m2.
        utils._EngineIntArrayType.c_type, # Output, icycl[m2].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _m2 = utils._EngineIntScalarType(
        m2, {'fun_name': fun_name, 'entity': 'm2'},
    ).to_ctypes()
    _m2_py = utils._EngineIntScalarType.to_py(_m2)
    _iperm_cls = utils._EngineIntArrayType(
        iperm, {'fun_name': fun_name, 'entity': 'iperm'},
        exp_shape=(_m2_py,),
    )
    _iperm = _iperm_cls.to_ctypes()
    _m1 = utils._EngineIntScalarType(
        m1, {'fun_name': fun_name, 'entity': 'm1'},
    ).to_ctypes()
    _icycl_cls = utils._EngineIntArrayType(
        icycl, {'fun_name': fun_name, 'entity': 'icycl'},
        exp_shape=(_m2_py,),
        check_contents=False,
    )
    _icycl = _icycl_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _iperm,
        _m1,
        _m2,
        _icycl,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m1 = {:s} and m2 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: m1 <= m2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m1 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m1 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m2 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m2 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'iperm[m1:m2-1] does not contain a permutation of the integers m1 to m2.',
                    'iperm[I-1] contains an out-of-range value:',
                    'I = {:s}, iperm[I-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'iperm[m1:m2-1] does not contain a permutation of the integers m1 to m2.',
                    'iperm contains a repeated value: {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _iperm_cls.refresh_py(_iperm)
        _icycl_cls.refresh_py(_icycl)
        return_data = None
    _err_class.manage()
    return return_data
