# -*- coding: utf-8 -*-
# pylint: disable=bare-except,chained-comparison,invalid-name,len-as-condition,line-too-long,protected-access,redefined-outer-name,too-many-arguments,too-many-boolean-expressions,too-many-branches,too-many-lines,too-many-locals,too-many-statements,unneeded-not,unused-import
u"""
Interfaces for the NAG Mark 27.0 `sum` Chapter.

``sum`` - Summation of Series

Base interfaces.

We do not intend for these lower-level base interfaces to be called
directly from user code. Please refer to the ``naginterfaces.library``
subpackage instead.

This module is concerned with the following tasks:

(a) calculating the **discrete Fourier transform** of a sequence of real
    or complex data values;

(#) calculating the **discrete convolution** or the **discrete
    correlation** of two sequences of real or complex data values using
    discrete Fourier transforms;

(#) calculating the **inverse Laplace transform** of a user-supplied
    function;

(#) calculating the **fast Gauss transform** approximation to the
    **discrete Gauss transform**;

(#) direct summation of orthogonal series;

(#) acceleration of convergence of a sequence of real values.

For full information please refer to the NAG Library document

https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c06/c06intro.html
"""

# NAG Copyright 2017-2020.

import collections as _collections
import logging as _logging
import sys as _sys

from ..base import _nag_enum_consts
from ..base import utils

def accelerate(seqn, ncall, comm, nterms):
    u"""
    Acceleration of convergence of sequence, Shanks' transformation and
    epsilon algorithm.

    ``accelerate`` accelerates the convergence of a given convergent
    sequence to its limit.

    For full information please refer to the NAG Library document for
    c06ba

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c06/c06baf.html

    Parameters
    ----------
    seqn : float
        The next term of the sequence to be considered.

    ncall : int
        On the first call `ncall` must be set to 0. Thereafter `ncall`
        must not be changed between calls.

    comm : dict, communication object, modified in place
        Communication structure.

        `On initial entry`: need not be set.

    nterms : int
        The maximum number of terms in the sequence.

    Returns
    -------
    ncall : int
        The number of terms in the sequence that have been considered.

    result : float
        The estimate of the limit of the sequence. For the first two
        calls, `result` = `seqn`.

    abserr : float
        An estimate of the absolute error in `result`. For the first
        three calls, `abserr` is set to a large machine-dependent
        constant.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `ncall` = *<value>*.

            Constraint: `ncall` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Shanks, D, 1955, `Nonlinear transformations of divergent and slowly
    convergent sequences`, J. Math. Phys. (34), 1--42

    Wynn, P, 1956, `On a device for computing the e_m(S_n)
    transformation`, Math. Tables Aids Comput. (10), 91--96
    """

    excs = []
    fun_name = 'naginterfaces.base.sum.accelerate'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('c06baft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, seqn.
        utils._BYREF_INT, # Input/Output, ncall.
        utils._BYREF_FLOAT64, # Output, result.
        utils._BYREF_FLOAT64, # Output, abserr.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, work[lwork].
        utils._BYREF_INT, # Input, lwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _seqn = utils._EngineFloat64ScalarType(
        seqn, {'fun_name': fun_name, 'entity': 'seqn'},
    ).to_ctypes()
    _ncall = utils._EngineIntScalarType(
        ncall, {'fun_name': fun_name, 'entity': 'ncall'},
    ).to_ctypes()
    _result = utils._EngineFloat64ScalarType.empty_ctypes()
    _abserr = utils._EngineFloat64ScalarType.empty_ctypes()
    _comm_needs_init = utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        opt_keys=('work',),
    )
    _nterms = utils._EngineIntScalarType(
        nterms, {'fun_name': fun_name, 'entity': 'nterms'},
    ).to_ctypes()
    _nterms_py = utils._EngineIntScalarType.to_py(_nterms)
    if _comm_needs_init:
        _work = utils._EngineFloat64ArrayType.empty_ctypes(max(7, _nterms_py+6))
        comm['work'] = _work
    else:
        _work = comm['work']
    _lwork_py = utils._EngineFloat64ArrayType.get_shape(comm['work'], {'fun_name': fun_name, 'entity': 'comm[\'work\']'}, exp_rank=1, validate=False)[0]
    _lwork = utils._EngineIntScalarType(
        _lwork_py, {'fun_name': fun_name, 'entity': '_lwork'},
    ).to_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _seqn,
        _ncall,
        _result,
        _abserr,
        _work,
        _lwork,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ncall = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ncall >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _ncall_py = utils._EngineIntScalarType.to_py(_ncall)
        _result_py = utils._EngineFloat64ScalarType.to_py(_result)
        _abserr_py = utils._EngineFloat64ScalarType.to_py(_abserr)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['ncall', 'result', 'abserr'],
        )(
            _ncall_py,
            _result_py,
            _abserr_py,
        )
    _err_class.manage()
    return return_data

def chebyshev(x, lx, xmin, xmax, c, n, s, res):
    u"""
    Sum of a Chebyshev series at a set of points.

    ``chebyshev`` evaluates a polynomial from its Chebyshev series
    representation at a set of points.

    For full information please refer to the NAG Library document for
    c06dc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c06/c06dcf.html

    Parameters
    ----------
    x : float, shape (`lx`)
        x in X, the set of arguments of the series.

    lx : int
        The number of evaluation points in X.

    xmin : float
        The lower end point of the interval [x_min,x_max].

    xmax : float
        The upper end point of the interval [x_min,x_max].

    c : float, shape (`n`)
        `c`[j-1] must contain the coefficient c_j of the Chebyshev
        series, for j = 1,2,...,n.

    n : int
        n, the number of terms in the series.

    s : int
        Determines the series (see [Description]).

        `s` = 1

            The series is general.

        `s` = 2

            The series is even.

        `s` = 3

            The series is odd.

    res : float, shape (`lx`), modified in place
        `On exit`: the Chebyshev series evaluated at the set of points
        X.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `lx` = *<value>*.

            Constraint: `lx` >= 1.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 3)
            On entry, `s` = *<value>*.

            Constraint: `s` = 1, 2 or 3.

        (`errno` 4)
            On entry, `xmax` = *<value>* and `xmin` = *<value>*.

            Constraint: `xmin` < `xmax`.

        (`errno` 5)
            On entry, element `x`[*<value>*] = *<value>*, `xmin` =
            *<value>* and `xmax` = *<value>*.

            Constraint: `xmin` <= `x`[i] <= `xmax`, for all i.

    References
    ----------
    Clenshaw, C W, 1962, `Chebyshev Series for Mathematical Functions`,
    Mathematical tables, HMSO
    """

    excs = []
    fun_name = 'naginterfaces.base.sum.chebyshev'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('c06dcft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._EngineFloat64ArrayType.c_type, # Input, x[lx].
        utils._BYREF_INT, # Input, lx.
        utils._BYREF_FLOAT64, # Input, xmin.
        utils._BYREF_FLOAT64, # Input, xmax.
        utils._EngineFloat64ArrayType.c_type, # Input, c[n].
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, s.
        utils._EngineFloat64ArrayType.c_type, # Output, res[lx].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _lx = utils._EngineIntScalarType(
        lx, {'fun_name': fun_name, 'entity': 'lx'},
    ).to_ctypes()
    _lx_py = utils._EngineIntScalarType.to_py(_lx)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_lx_py,),
    ).to_ctypes()
    _xmin = utils._EngineFloat64ScalarType(
        xmin, {'fun_name': fun_name, 'entity': 'xmin'},
    ).to_ctypes()
    _xmax = utils._EngineFloat64ScalarType(
        xmax, {'fun_name': fun_name, 'entity': 'xmax'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _c = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _s = utils._EngineIntScalarType(
        s, {'fun_name': fun_name, 'entity': 's'},
    ).to_ctypes()
    _res_cls = utils._EngineFloat64ArrayType(
        res, {'fun_name': fun_name, 'entity': 'res'},
        exp_shape=(_lx_py,),
        check_contents=False,
    )
    _res = _res_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _x,
        _lx,
        _xmin,
        _xmax,
        _c,
        _n,
        _s,
        _res,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lx = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: lx >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, s = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: s = 1, 2 or 3.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, xmax = {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'xmin = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: xmin < xmax.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, element x[{:s}] = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'xmin = {:s} and xmax = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[4]),
                    ),
                    'Constraint: xmin <= x[i] <= xmax, for all i.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _res_cls.refresh_py(_res)
        return_data = None
    _err_class.manage()
    return return_data

def fft_real_1d_rfmt(x, n):
    u"""
    Single one-dimensional real discrete Fourier transform, extra
    workspace for greater speed.

    ``fft_real_1d_rfmt`` calculates the discrete Fourier transform of a
    sequence of n real data values (using a work array for extra speed).

    For full information please refer to the NAG Library document for
    c06fa

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c06/c06faf.html

    Parameters
    ----------
    x : float, shape (`n`), modified in place
        `On entry`: if `x` is declared with bounds (0:`n`-1) in the
        function from which ``fft_real_1d_rfmt`` is called, `x`[j] must
        contain x_j, for j = 0,1,...,n-1.

        `On exit`: the discrete Fourier transform stored in Hermitian
        form. If the components of the transform z^_k are written as
        a_k+ib_k, and if `x` is declared with bounds (0:`n`-1) in the
        function from which ``fft_real_1d_rfmt`` is called, then for 0
        <= k <= n/2, a_k is contained in `x`[k-1], and for 1 <= k <=
        (n-1)/2, b_k is contained in `x`[n-k]. (See also [the C06
        Introduction].)

    n : int
        n, the number of data values.

    Raises
    ------
    NagValueError
        (`errno` 3)
            On entry, `n` = *<value>*.

            Constraint: `n` > 1.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Brigham, E O, 1974, `The Fast Fourier Transform`, Prentice--Hall
    """

    excs = []
    fun_name = 'naginterfaces.base.sum.fft_real_1d_rfmt'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('c06faft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._EngineFloat64ArrayType.c_type, # Input/Output, x[n].
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Workspace, work.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    )
    _x = _x_cls.to_ctypes()
    _work = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _x,
        _n,
        _work,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 3 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n > 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def fft_hermitian_1d_rfmt(x, n):
    u"""
    Single one-dimensional Hermitian discrete Fourier transform, extra
    workspace for greater speed.

    ``fft_hermitian_1d_rfmt`` calculates the discrete Fourier transform
    of a Hermitian sequence of n complex data values (using a work array
    for extra speed).

    For full information please refer to the NAG Library document for
    c06fb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c06/c06fbf.html

    Parameters
    ----------
    x : float, shape (`n`), modified in place
        `On entry`: the sequence to be transformed stored in Hermitian
        form. If the data values z_j are written as x_j+iy_j, and if `x`
        is declared with bounds (0:`n`-1) in the function from which
        ``fft_hermitian_1d_rfmt`` is called, then for 0 <= j <= n/2, x_j
        is contained in `x`[j-1], and for 1 <= j <= (n-1)/2, y_j is
        contained in `x`[n-j]. (See also [the C06 Introduction].)

        `On exit`: the components of the discrete Fourier transform
        x^_k. If `x` is declared with bounds (0:`n`-1) in the function
        from which ``fft_hermitian_1d_rfmt`` is called, x^_k is stored
        in `x`[k], for k = 0,1,...,n-1.

    n : int
        n, the number of data values.

    Raises
    ------
    NagValueError
        (`errno` 3)
            On entry, `n` = *<value>*.

            Constraint: `n` > 1.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Brigham, E O, 1974, `The Fast Fourier Transform`, Prentice--Hall
    """

    excs = []
    fun_name = 'naginterfaces.base.sum.fft_hermitian_1d_rfmt'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('c06fbft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._EngineFloat64ArrayType.c_type, # Input/Output, x[n].
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Workspace, work.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    )
    _x = _x_cls.to_ctypes()
    _work = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _x,
        _n,
        _work,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 3 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n > 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def fft_complex_1d_sep(x, y, n):
    u"""
    Single one-dimensional complex discrete Fourier transform, extra
    workspace for greater speed.

    ``fft_complex_1d_sep`` calculates the discrete Fourier transform of
    a sequence of n complex data values (using a work array for extra
    speed).

    For full information please refer to the NAG Library document for
    c06fc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c06/c06fcf.html

    Parameters
    ----------
    x : float, shape (`n`), modified in place
        `On entry`: if `x` is declared with bounds (0:`n`-1) in the
        function from which ``fft_complex_1d_sep`` is called, `x`[j-1]
        must contain x_j, the real part of z_j, for j = 0,1,...,n-1.

        `On exit`: the real parts a_k of the components of the discrete
        Fourier transform. If `x` is declared with bounds (0:`n`-1) in
        the function from which ``fft_complex_1d_sep`` is called, for 0
        <= k <= n-1, a_k is contained in `x`[k-1].

    y : float, shape (`n`), modified in place
        `On entry`: if `y` is declared with bounds (0:`n`-1) in the
        function from which ``fft_complex_1d_sep`` is called, `y`[j-1]
        must contain y_j, the imaginary part of z_j, for j =
        0,1,...,n-1.

        `On exit`: the imaginary parts b_k of the components of the
        discrete Fourier transform. If `y` is declared with bounds
        (0:`n`-1) in the function from which ``fft_complex_1d_sep`` is
        called, then for 0 <= k <= n-1, b_k is contained in `y`[k-1].

    n : int
        n, the number of data values.

    Raises
    ------
    NagValueError
        (`errno` 3)
            On entry, `n` = *<value>*.

            Constraint: `n` > 1.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Brigham, E O, 1974, `The Fast Fourier Transform`, Prentice--Hall
    """

    excs = []
    fun_name = 'naginterfaces.base.sum.fft_complex_1d_sep'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('c06fcft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._EngineFloat64ArrayType.c_type, # Input/Output, x[n].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, y[n].
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Workspace, work.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    )
    _x = _x_cls.to_ctypes()
    _y_cls = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py,),
    )
    _y = _y_cls.to_ctypes()
    _work = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _x,
        _y,
        _n,
        _work,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 3 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n > 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        _y_cls.refresh_py(_y)
        return_data = None
    _err_class.manage()
    return return_data

def fft_complex_multid_1d_sep(ndim, l, nd, n, x, y):
    u"""
    One-dimensional complex discrete Fourier transform of
    multidimensional data.

    ``fft_complex_multid_1d_sep`` computes the discrete Fourier
    transform of one variable in a multivariate sequence of complex data
    values.

    For full information please refer to the NAG Library document for
    c06ff

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c06/c06fff.html

    Parameters
    ----------
    ndim : int
        m, the number of dimensions (or variables) in the multivariate
        data.

    l : int
        l, the index of the variable (or dimension) on which the
        discrete Fourier transform is to be performed.

    nd : int, shape (`ndim`)
        `nd`[i-1] must contain n_i (the dimension of the ith variable),
        for i = 1,2,...,m.

    n : int
        n, the total number of data values.

    x : float, shape (`n`), modified in place
        `On entry`: `x`[ 1 + j_1 + n_1 j_2 + n_1 n_2 j_3 + ... -1] must
        contain the real part of the complex data value z_{j_1
        j_2...j_m}, for 0 <= j_1 <= n_1-1,0 <= j_2 <= n_2-1,...; i.e.,
        the values are stored in consecutive elements of the array
        according to the Fortran convention for storing multidimensional
        arrays.

        `On exit`: the real parts of the corresponding elements of the
        computed transform.

    y : float, shape (`n`), modified in place
        `On entry`: the imaginary parts of the complex data values,
        stored in the same way as the real parts in the array `x`.

        `On exit`: the imaginary parts of the corresponding elements of
        the computed transform.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `ndim` = *<value>*.

            Constraint: `ndim` >= 1.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` = `nd`[0]*`nd`[1]*...*`nd`[`ndim`-1].

        (`errno` 3)
            On entry, `l` = *<value>* and `ndim` = *<value>*.

            Constraint: 1 <= `l` <= `ndim`.

        (`errno` 10*`l`+3)
            On entry, l = *<value>*.

            Constraint: `nd`[l-1] >= 1.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Brigham, E O, 1974, `The Fast Fourier Transform`, Prentice--Hall
    """

    excs = []
    fun_name = 'naginterfaces.base.sum.fft_complex_multid_1d_sep'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('c06ffft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, iemode.
        utils._BYREF_INT, # Input, ndim.
        utils._BYREF_INT, # Input, l.
        utils._EngineIntArrayType.c_type, # Input, nd[ndim].
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, x[n].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, y[n].
        utils._EngineFloat64ArrayType.c_type, # Workspace, work.
        utils._BYREF_INT, # Input, lwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _iemode = utils._EngineIntScalarType.c_type(0)
    _ndim = utils._EngineIntScalarType(
        ndim, {'fun_name': fun_name, 'entity': 'ndim'},
    ).to_ctypes()
    _l = utils._EngineIntScalarType(
        l, {'fun_name': fun_name, 'entity': 'l'},
    ).to_ctypes()
    _ndim_py = utils._EngineIntScalarType.to_py(_ndim)
    _nd = utils._EngineIntArrayType(
        nd, {'fun_name': fun_name, 'entity': 'nd'},
        exp_shape=(_ndim_py,),
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    )
    _x = _x_cls.to_ctypes()
    _y_cls = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py,),
    )
    _y = _y_cls.to_ctypes()
    _work = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _lwork = utils._EngineIntScalarType.c_type(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _iemode,
        _ndim,
        _l,
        _nd,
        _n,
        _x,
        _y,
        _work,
        _lwork,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        _l_py = utils._EngineIntScalarType.to_py(_l)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ndim = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ndim >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n = nd[0]*nd[1]*...*nd[ndim-1].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, l = {:s} and ndim = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: 1 <= l <= ndim.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 10*_l_py+3 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, l = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nd[l-1] >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        _y_cls.refresh_py(_y)
        return_data = None
    _err_class.manage()
    return return_data

def fft_complex_multid_sep(ndim, nd, n, x, y):
    u"""
    Multidimensional complex discrete Fourier transform of
    multidimensional data.

    ``fft_complex_multid_sep`` computes the multidimensional discrete
    Fourier transform of a multivariate sequence of complex data values.

    For full information please refer to the NAG Library document for
    c06fj

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c06/c06fjf.html

    Parameters
    ----------
    ndim : int
        m, the number of dimensions (or variables) in the multivariate
        data.

    nd : int, shape (`ndim`)
        `nd`[i-1] must contain n_i (the dimension of the ith variable),
        for i = 1,2,...,m.

    n : int
        n, the total number of data values.

    x : float, shape (`n`), modified in place
        `On entry`: `x`[ 1 + j_1 + n_1 j_2 + n_1 n_2 j_3 + ... -1] must
        contain the real part of the complex data value z_{j_1
        j_2...j_m}, for 0 <= j_1 <= n_1-1,0 <= j_2 <= n_2-1,...; i.e.,
        the values are stored in consecutive elements of the array
        according to the Fortran convention for storing multidimensional
        arrays.

        `On exit`: the real parts of the corresponding elements of the
        computed transform.

    y : float, shape (`n`), modified in place
        `On entry`: the imaginary parts of the complex data values,
        stored in the same way as the real parts in the array `x`.

        `On exit`: the imaginary parts of the corresponding elements of
        the computed transform.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `ndim` = *<value>*.

            Constraint: `ndim` >= 1.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` = `nd`[0]*`nd`[1]*...*`nd`[`ndim`-1].

        (`errno` i mod 10 = 3)
            On entry, l = *<value>*.

            Constraint: `nd`[l-1] >= 1.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Brigham, E O, 1974, `The Fast Fourier Transform`, Prentice--Hall
    """

    excs = []
    fun_name = 'naginterfaces.base.sum.fft_complex_multid_sep'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('c06fjft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, iemode.
        utils._BYREF_INT, # Input, ndim.
        utils._EngineIntArrayType.c_type, # Input, nd[ndim].
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, x[n].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, y[n].
        utils._EngineFloat64ArrayType.c_type, # Workspace, work.
        utils._BYREF_INT, # Input, lwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _iemode = utils._EngineIntScalarType.c_type(0)
    _ndim = utils._EngineIntScalarType(
        ndim, {'fun_name': fun_name, 'entity': 'ndim'},
    ).to_ctypes()
    _ndim_py = utils._EngineIntScalarType.to_py(_ndim)
    _nd = utils._EngineIntArrayType(
        nd, {'fun_name': fun_name, 'entity': 'nd'},
        exp_shape=(_ndim_py,),
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    )
    _x = _x_cls.to_ctypes()
    _y_cls = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py,),
    )
    _y = _y_cls.to_ctypes()
    _work = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _lwork = utils._EngineIntScalarType.c_type(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _iemode,
        _ndim,
        _nd,
        _n,
        _x,
        _y,
        _work,
        _lwork,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ndim = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ndim >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n = nd[0]*nd[1]*...*nd[ndim-1].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py%10 == 3 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, l = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nd[l-1] >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        _y_cls.refresh_py(_y)
        return_data = None
    _err_class.manage()
    return return_data

def convcorr_real(job, x, y, n):
    u"""
    Circular convolution or correlation of two real vectors, no
    restrictions on n.

    ``convcorr_real`` calculates the circular convolution or correlation
    of two real vectors of period n (using a work array for extra
    speed).

    For full information please refer to the NAG Library document for
    c06fk

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c06/c06fkf.html

    Parameters
    ----------
    job : int
        The computation to be performed.

        `job` = 1

            z_k = sum_j = 0^n-1x_j y_{k-j} (convolution);

        `job` = 2

            w_k = sum_j = 0^n-1x_j y_{k+j} (correlation).

    x : float, shape (`n`), modified in place
        `On entry`: the elements of one period of the vector x. If `x`
        is declared with bounds (0:`n`-1) in the function from which
        ``convcorr_real`` is called, `x`[j] must contain x_j, for j =
        0,1,...,n-1.

        `On exit`: the corresponding elements of the discrete
        convolution or correlation.

    y : float, shape (`n`), modified in place
        `On entry`: the elements of one period of the vector y. If `y`
        is declared with bounds (0:`n`-1) in the function from which
        ``convcorr_real`` is called, `y`[j] must contain y_j, for j =
        0,1,...,n-1.

        `On exit`: the discrete Fourier transform of the convolution or
        correlation returned in the array `x`; the transform is stored
        in Hermitian form; if the components of the transform z_k are
        written as a_k+ib_k, then for 0 <= k <= n/2, a_k is contained in
        `y`[k], and for 1 <= k <= n/2-1, b_k is contained in `y`[n-k].
        (See also [the C06 Introduction].)

    n : int
        n, the number of values in one period of the vectors `x` and
        `y`.

    Raises
    ------
    NagValueError
        (`errno` 3)
            On entry, `n` = *<value>*.

            Constraint: `n` > 1.

        (`errno` 4)
            On entry, `job` = *<value>*.

            Constraint: `job` = 1 or 2.

    References
    ----------
    Brigham, E O, 1974, `The Fast Fourier Transform`, Prentice--Hall

    See Also
    --------
    :meth:`naginterfaces.library.examples.sum.convcorr_real_ex.main`
    """

    excs = []
    fun_name = 'naginterfaces.base.sum.convcorr_real'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('c06fkft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, job.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, x[n].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, y[n].
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Workspace, work.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _job = utils._EngineIntScalarType(
        job, {'fun_name': fun_name, 'entity': 'job'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    )
    _x = _x_cls.to_ctypes()
    _y_cls = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py,),
    )
    _y = _y_cls.to_ctypes()
    _work = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _job,
        _x,
        _y,
        _n,
        _work,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 3 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n > 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, job = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: job = 1 or 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        _y_cls.refresh_py(_y)
        return_data = None
    _err_class.manage()
    return return_data

@utils._withdrawal(
    'naginterfaces.base.sum.withdraw_fft_real_1d_multi_rfmt',
    mkwdrn='28',
    advice=[
        'Please use fft_realherm_1d_multi_row or fft_realherm_1d_multi_col instead.',
        'See also https://www.nag.com/numeric/py/nagdoc_latest/replace.html',
    ]
)
def withdraw_fft_real_1d_multi_rfmt(m, n, x, init, trig):
    u"""
    Multiple one-dimensional real discrete Fourier transforms.

    ``withdraw_fft_real_1d_multi_rfmt`` computes the discrete Fourier
    transforms of m sequences, each containing n real data values.
    This function is designed to be particularly efficient on vector
    processors.

    Note: ``withdraw_fft_real_1d_multi_rfmt`` will be removed in
    naginterfaces 28.0.0.
    Please use ``fft_realherm_1d_multi_row`` or
    ``fft_realherm_1d_multi_col`` instead.
    See also https://www.nag.com/numeric/py/nagdoc_latest/replace.html

    For full information please refer to the NAG Library document for
    c06fp

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c06/c06fpf.html

    Parameters
    ----------
    m : int
        m, the number of sequences to be transformed.

    n : int
        n, the number of real values in each sequence.

    x : float, shape (`m`*`n`), modified in place
        `On entry`: the data must be stored in `x` as if in a
        two-dimensional array of dimension (1:`m`,0:`n`-1); each of the
        m sequences is stored in a **row** of the array. In other words,
        if the data values of the pth sequence to be transformed are
        denoted by x_j^p, for j = 0,1,...,n-1, the mn elements of the
        array `x` must contain the values

x_0^1,x_0^2,...,x_0^m,x_1^1,x_1^2,...,x_1^m,...,x_n-1^1,x_n-1^2,...,x_n-1^m.

        `On exit`: the m discrete Fourier transforms stored as if in a
        two-dimensional array of dimension (1:`m`,0:`n`-1). Each of the
        m transforms is stored in a **row** of the array in Hermitian
        form, overwriting the corresponding original sequence. If the n
        components of the discrete Fourier transform z^_k^p are written
        as a_k^p+ib_k^p, then for 0 <= k <= n/2, a_k^p is contained in
        `x`[p-1,k-1], and for 1 <= k <= (n-1)/2, b_k^p is contained in
        `x`[p-1,n-k-1]. (See also [the C06 Introduction].)

    init : str, length 1
        Indicates whether trigonometric coefficients are to be
        calculated.

        `init` = 'I'

            Calculate the required trigonometric coefficients for the
            given value of n, and store in the array `trig`.

        `init` = 'S' or 'R'

            The required trigonometric coefficients are assumed to have
            been calculated and stored in the array `trig` in a prior
            call to one of ``withdraw_fft_real_1d_multi_rfmt`` or
            ``withdraw_fft_hermitian_1d_multi_rfmt``. The function
            performs a simple check that the current value of n is
            consistent with the values stored in `trig`.

    trig : float, shape (2*`n`), modified in place
        `On entry`: if `init` = 'S' or 'R', `trig` must contain the
        required trigonometric coefficients that have been previously
        calculated. Otherwise `trig` need not be set.

        `On exit`: contains the required coefficients (computed by the
        function if `init` = 'I').

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 3)
            On entry, `init` = *<value>*.

            Constraint: `init` = 'I', 'S' or 'R'.

        (`errno` 5)
            On entry, `init` = *<value>* but `n` and `trig` array
            incompatible.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Brigham, E O, 1974, `The Fast Fourier Transform`, Prentice--Hall

    Temperton, C, 1983, `Fast mixed-radix real Fourier transforms`, J.
    Comput. Phys. (52), 340--350
    """

    excs = []
    fun_name = 'naginterfaces.base.sum.withdraw_fft_real_1d_multi_rfmt'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('c06fpft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, x[m*n].
        utils._BYREF_STRING, # Input, init(len=1).
        utils._EngineFloat64ArrayType.c_type, # Input/Output, trig[2*n].
        utils._EngineFloat64ArrayType.c_type, # Workspace, work.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of init.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_m_py*_n_py,),
    )
    _x = _x_cls.to_ctypes()
    _init = utils._EngineCharScalarType(
        init, {'fun_name': fun_name, 'entity': 'init'}, exp_chrlen=1,
    ).to_ctypes()
    _trig_cls = utils._EngineFloat64ArrayType(
        trig, {'fun_name': fun_name, 'entity': 'trig'},
        exp_shape=(2*_n_py,),
    )
    _trig = _trig_cls.to_ctypes()
    _work = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _m,
        _n,
        _x,
        _init,
        _trig,
        _work,
        _errbuf,
        _ierr,
        _init._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, init = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: init = \'I\', \'S\' or \'R\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, init = \'{:s}\' but n and trig array incompatible.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        _trig_cls.refresh_py(_trig)
        return_data = None
    _err_class.manage()
    return return_data

@utils._withdrawal(
    'naginterfaces.base.sum.withdraw_fft_hermitian_1d_multi_rfmt',
    mkwdrn='28',
    advice=[
        'Please use fft_realherm_1d_multi_row or fft_realherm_1d_multi_col instead.',
        'See also https://www.nag.com/numeric/py/nagdoc_latest/replace.html',
    ]
)
def withdraw_fft_hermitian_1d_multi_rfmt(m, n, x, init, trig):
    u"""
    Multiple one-dimensional Hermitian discrete Fourier transforms.

    ``withdraw_fft_hermitian_1d_multi_rfmt`` computes the discrete
    Fourier transforms of m Hermitian sequences, each containing n
    complex data values.
    This function is designed to be particularly efficient on vector
    processors.

    Note: ``withdraw_fft_hermitian_1d_multi_rfmt`` will be removed in
    naginterfaces 28.0.0.
    Please use ``fft_realherm_1d_multi_row`` or
    ``fft_realherm_1d_multi_col`` instead.
    See also https://www.nag.com/numeric/py/nagdoc_latest/replace.html

    For full information please refer to the NAG Library document for
    c06fq

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c06/c06fqf.html

    Parameters
    ----------
    m : int
        m, the number of sequences to be transformed.

    n : int
        n, the number of data values in each sequence.

    x : float, shape (`m`*`n`), modified in place
        `On entry`: the data must be stored in `x` as if in a
        two-dimensional array of dimension (1:`m`,0:`n`-1); each of the
        m sequences is stored in a **row** of the array in Hermitian
        form. If the n data values z_j^p are written as x_j^p+iy_j^p,
        then for 0 <= j <= n/2, x_j^p is contained in `x`[p-1,j], and
        for 1 <= j <= (n-1)/2, y_j^p is contained in `x`[p,n-j]. (See
        also [the C06 Introduction].)

        `On exit`: the components of the m discrete Fourier transforms,
        stored as if in a two-dimensional array of dimension
        (1:`m`,0:`n`-1). Each of the m transforms is stored as a **row**
        of the array, overwriting the corresponding original sequence.
        If the n components of the discrete Fourier transform are
        denoted by x^_k^p, for k = 0,1,...,n-1, the mn elements of the
        array `x` contain the values

x^_0^1,x^_0^2,...,x^_0^m,x^_1^1,x^_1^2,...,x^_1^m,...,x^_n-1^1,x^_n-1^2,...,x^_n-1^m.

    init : str, length 1
        Indicates whether trigonometric coefficients are to be
        calculated.

        `init` = 'I'

            Calculate the required trigonometric coefficients for the
            given value of n, and store in the array `trig`.

        `init` = 'S' or 'R'

            The required trigonometric coefficients are assumed to have
            been calculated and stored in the array `trig` in a prior
            call to one of ``withdraw_fft_real_1d_multi_rfmt`` or
            ``withdraw_fft_hermitian_1d_multi_rfmt``. The function
            performs a simple check that the current value of n is
            consistent with the values stored in `trig`.

    trig : float, shape (2*`n`), modified in place
        `On entry`: if `init` = 'S' or 'R', `trig` must contain the
        required trigonometric coefficients that have been previously
        calculated. Otherwise `trig` need not be set.

        `On exit`: contains the required coefficients (computed by the
        function if `init` = 'I').

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 3)
            On entry, `init` = *<value>*.

            Constraint: `init` = 'I', 'S' or 'R'.

        (`errno` 5)
            On entry, `init` = *<value>* but `n` and `trig` array
            incompatible.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Brigham, E O, 1974, `The Fast Fourier Transform`, Prentice--Hall

    Temperton, C, 1983, `Fast mixed-radix real Fourier transforms`, J.
    Comput. Phys. (52), 340--350
    """

    excs = []
    fun_name = 'naginterfaces.base.sum.withdraw_fft_hermitian_1d_multi_rfmt'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('c06fqft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, x[m*n].
        utils._BYREF_STRING, # Input, init(len=1).
        utils._EngineFloat64ArrayType.c_type, # Input/Output, trig[2*n].
        utils._EngineFloat64ArrayType.c_type, # Workspace, work.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of init.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_m_py*_n_py,),
    )
    _x = _x_cls.to_ctypes()
    _init = utils._EngineCharScalarType(
        init, {'fun_name': fun_name, 'entity': 'init'}, exp_chrlen=1,
    ).to_ctypes()
    _trig_cls = utils._EngineFloat64ArrayType(
        trig, {'fun_name': fun_name, 'entity': 'trig'},
        exp_shape=(2*_n_py,),
    )
    _trig = _trig_cls.to_ctypes()
    _work = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _m,
        _n,
        _x,
        _init,
        _trig,
        _work,
        _errbuf,
        _ierr,
        _init._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, init = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: init = \'I\', \'S\' or \'R\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, init = \'{:s}\' but n and trig array incompatible.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        _trig_cls.refresh_py(_trig)
        return_data = None
    _err_class.manage()
    return return_data

def fft_complex_3d_sep(n1, n2, n3, x, y):
    u"""
    Three-dimensional complex discrete Fourier transform.

    ``fft_complex_3d_sep`` computes the three-dimensional discrete
    Fourier transform of a trivariate sequence of complex data values.
    This function is designed to be particularly efficient on vector
    processors.

    For full information please refer to the NAG Library document for
    c06fx

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c06/c06fxf.html

    Parameters
    ----------
    n1 : int
        n_1, the first dimension of the transform.

    n2 : int
        n_2, the second dimension of the transform.

    n3 : int
        n_3, the third dimension of the transform.

    x : float, shape (`n1`*`n2`*`n3`), modified in place
        `On entry`: the real and imaginary parts of the complex data
        values must be stored in arrays `x` and `y` respectively. If `x`
        and `y` are regarded as three-dimensional arrays of dimension
        (0:`n1`-1,0:`n2`-1,0:`n3`-1), `x`[j_1-1,j_2-1,j_3-1] and
        `y`[j_1-1,j_2-1,j_3-1] must contain the real and imaginary parts
        of z_{j_1 j_2 j_3}.

        `On exit`: the real and imaginary parts respectively of the
        corresponding elements of the computed transform.

    y : float, shape (`n1`*`n2`*`n3`), modified in place
        `On entry`: the real and imaginary parts of the complex data
        values must be stored in arrays `x` and `y` respectively. If `x`
        and `y` are regarded as three-dimensional arrays of dimension
        (0:`n1`-1,0:`n2`-1,0:`n3`-1), `x`[j_1-1,j_2-1,j_3-1] and
        `y`[j_1-1,j_2-1,j_3-1] must contain the real and imaginary parts
        of z_{j_1 j_2 j_3}.

        `On exit`: the real and imaginary parts respectively of the
        corresponding elements of the computed transform.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n1` = *<value>*.

            Constraint: `n1` >= 1.

        (`errno` 2)
            On entry, `n2` = *<value>*.

            Constraint: `n2` >= 1.

        (`errno` 3)
            On entry, `n3` = *<value>*.

            Constraint: `n3` >= 1.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Brigham, E O, 1974, `The Fast Fourier Transform`, Prentice--Hall

    Temperton, C, 1983, `Self-sorting mixed-radix fast Fourier
    transforms`, J. Comput. Phys. (52), 1--23
    """

    excs = []
    fun_name = 'naginterfaces.base.sum.fft_complex_3d_sep'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('c06fxft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n1.
        utils._BYREF_INT, # Input, n2.
        utils._BYREF_INT, # Input, n3.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, x[n1*n2*n3].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, y[n1*n2*n3].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n1 = utils._EngineIntScalarType(
        n1, {'fun_name': fun_name, 'entity': 'n1'},
    ).to_ctypes()
    _n2 = utils._EngineIntScalarType(
        n2, {'fun_name': fun_name, 'entity': 'n2'},
    ).to_ctypes()
    _n3 = utils._EngineIntScalarType(
        n3, {'fun_name': fun_name, 'entity': 'n3'},
    ).to_ctypes()
    _n1_py = utils._EngineIntScalarType.to_py(_n1)
    _n2_py = utils._EngineIntScalarType.to_py(_n2)
    _n3_py = utils._EngineIntScalarType.to_py(_n3)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n1_py*_n2_py*_n3_py,),
    )
    _x = _x_cls.to_ctypes()
    _y_cls = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n1_py*_n2_py*_n3_py,),
    )
    _y = _y_cls.to_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n1,
        _n2,
        _n3,
        _x,
        _y,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n1 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n1 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n2 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n2 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n3 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n3 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        _y_cls.refresh_py(_y)
        return_data = None
    _err_class.manage()
    return return_data

def invlaplace_crump(f, n, t, valinv, errest, relerr, alphab, tfac, mxterm, data=None):
    u"""
    Inverse Laplace transform, Crump's method.

    ``invlaplace_crump`` estimates values of the inverse Laplace
    transform of a given function using a Fourier series approximation.
    A bound on the exponential order of the inverse is required.

    For full information please refer to the NAG Library document for
    c06la

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c06/c06laf.html

    Parameters
    ----------
    f : callable fp = f(p, data=None)
        `f` must evaluate the function F(p) for a given value of p.

        Parameters
        ~~~~~~~~~~
        p : complex
            The argument p.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        fp : complex
            The value F(p).

    n : int
        n, the number of points at which the value of the inverse
        Laplace transform is required.

    t : float, shape (`n`)
        Each `t`[j-1] must specify a point at which the inverse Laplace
        transform is required, for j = 1,2,...,n.

    valinv : float, shape (`n`), modified in place
        `On exit`: an estimate of the value of the inverse Laplace
        transform at t = `t`[j-1], for j = 1,2,...,n.

    errest : float, shape (`n`), modified in place
        `On exit`: an estimate of the error in `valinv`[j]. This is
        usually an estimate of relative error but, if `valinv`[j-1] <
        `relerr`, `errest`[j] estimates the absolute error.
        `errest`[j-1] is unreliable when `valinv`[j-1] is small but
        slightly greater than `relerr`.

    relerr : float
        The required relative error in the values of the inverse Laplace
        transform. If the absolute value of the inverse is less than
        `relerr`, absolute accuracy is used instead. `relerr` must be in
        the range 0.0 <= `relerr` < 1.0. If `relerr` is set too small or
        to 0.0, the function uses a value sufficiently larger than
        machine precision.

    alphab : float
        alpha_b, an upper bound for alpha (see [Description]). Usually,
        alpha_b should be specified equal to, or slightly larger than,
        the value of alpha. If alpha_b < alpha then the prescribed
        accuracy may not be achieved or completely incorrect results may
        be obtained. If alpha_b is too large ``invlaplace_crump`` will
        be inefficient and convergence may not be achieved.

        **Note:** it is as important to specify alpha_b correctly as it
        is to specify the correct function for inversion.

    tfac : float
        t_fac, a factor to be used in calculating the parameter tau.
        Larger values (e.g., 5.0) may be specified for difficult
        problems, but these may require very large values of `mxterm`.

        Suggested value: `tfac` = 0.8.

    mxterm : int
        The maximum number of (complex) terms to be used in the
        evaluation of the Fourier series.

        Suggested value: `mxterm` >= 100, except for very simple
        problems.

    data : arbitrary, optional
        User-communication data for callback functions.

    Returns
    -------
    nterms : int
        The number of (complex) terms actually used.

    na : int
        The number of values of a used by the function. See [Further
        Comments].

    alow : float
        The smallest value of a used in the algorithm. This may be used
        for checking the value of `alphab`- see [Further Comments].

    ahigh : float
        The largest value of a used in the algorithm. This may be used
        for checking the value of `alphab`- see [Further Comments].

    nfeval : int
        The number of calls to `f` made by the function.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `tfac` = *<value>*.

            Constraint: `tfac` > 0.5.

        (`errno` 1)
            On entry, `relerr` = *<value>*.

            Constraint: 0.0 <= `relerr` < 1.0.

        (`errno` 1)
            On entry, `mxterm` = *<value>*.

            Constraint: `mxterm` >= 1.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 2)
            On entry, the elements of `t` are not strictly increasing.

        (`errno` 2)
            On entry, `t`[0] = *<value>*.

            Constraint: 0.0 <= `t`[0] < `t`[1] < ... < `t`[n-1].

        (`errno` 3)
            On entry, `t`[`n`-1] is too large: `t`[`n`-1] = *<value>*.
            If necessary, scale the problem as described in [Further
            Comments].

        (`errno` 4)
            Required accuracy cannot be obtained. Try increasing `tfac`,
            `alphab`, or both. `tfac` = *<value>* and `alphab` =
            *<value>*.

        (`errno` 5)
            Convergence failure in epsilon algorithm: `mxterm` =
            *<value>*. Some values of `valinv`[j-1] may be calculated to
            the desired accuracy; this may be determined by examining
            the values of `errest`[j-1]. Try reducing the range of `t`
            or increasing `mxterm`. If `errno` = 5 still results, try
            reducing `tfac`.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 6)
            All values of `valinv` have been calculated, but are not of
            the required accuracy; the values of `errest`[j-1] should be
            examined carefully. Try reducing the range of `t`, or
            increasing `tfac`, `alphab` or both.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Crump, K S, 1976, `Numerical inversion of Laplace transforms using a
    Fourier series approximation`, J. Assoc. Comput. Mach. (23), 89--96

    Durbin, F, 1974, `Numerical inversion of Laplace transforms: An
    efficient improvement to Dubner and Abate's method`, Comput. J.
    (17), 371--376

    Wynn, P, 1956, `On a device for computing the e_m(S_n)
    transformation`, Math. Tables Aids Comput. (10), 91--96
    """

    excs = []
    fun_name = 'naginterfaces.base.sum.invlaplace_crump'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    cb_f_name = fun_name + ':f'

    fh = utils._EngineState._sys_functype(
        None,
        utils._UNUSED_FUNCTYPE,
        utils._BYREF_ENGINEDATA,
        utils._BYREF_COMPLEX128, # Input, p.
        utils._BYREF_COMPLEX128, # Output, fp.
        utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
    )

    def cb_fh(_f, en_data, p, fp, _iuser, _ruser):
        """Helper for f."""
        try:
            _p_fh = utils._EngineComplex128ScalarType.to_py(p.contents)
            if data is not None:
                _fp_fh = f(
                    _p_fh,
                    data,
                )
            else:
                _fp_fh = f(
                    _p_fh,
                )
            utils._EngineComplex128ScalarType(
                _fp_fh, {'fun_name': cb_f_name, 'returned': 'fp'},
            ).refresh_ctypes_buffer(fp)
        except:
            excs.append(_sys.exc_info())
            en_data.contents.hlperr = utils._IERR_HELPER

    engine_ctypes = utils._EngineState._get_symbol('c06laft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._UNUSED_FUNCTYPE,
        fh,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, t[n].
        utils._EngineFloat64ArrayType.c_type, # Output, valinv[n].
        utils._EngineFloat64ArrayType.c_type, # Output, errest[n].
        utils._BYREF_FLOAT64, # Input, relerr.
        utils._BYREF_FLOAT64, # Input, alphab.
        utils._BYREF_FLOAT64, # Input, tfac.
        utils._BYREF_INT, # Input, mxterm.
        utils._BYREF_INT, # Output, nterms.
        utils._BYREF_INT, # Output, na.
        utils._BYREF_FLOAT64, # Output, alow.
        utils._BYREF_FLOAT64, # Output, ahigh.
        utils._BYREF_INT, # Output, nfeval.
        utils._EngineFloat64ArrayType.c_type, # Workspace, work.
        utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    utils._check_callable(f, {'fun_name': fun_name, 'entity': 'f'})
    _f = utils._UNUSED_CALLBACK
    _fh = utils._cast_callback(
        fh, cb_fh,
        {'fun_name': fun_name, 'entity': 'f'},
    )
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _t = utils._EngineFloat64ArrayType(
        t, {'fun_name': fun_name, 'entity': 't'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _valinv_cls = utils._EngineFloat64ArrayType(
        valinv, {'fun_name': fun_name, 'entity': 'valinv'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _valinv = _valinv_cls.to_ctypes(can_reference_contents=False)
    _errest_cls = utils._EngineFloat64ArrayType(
        errest, {'fun_name': fun_name, 'entity': 'errest'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _errest = _errest_cls.to_ctypes(can_reference_contents=False)
    _relerr = utils._EngineFloat64ScalarType(
        relerr, {'fun_name': fun_name, 'entity': 'relerr'},
    ).to_ctypes()
    _alphab = utils._EngineFloat64ScalarType(
        alphab, {'fun_name': fun_name, 'entity': 'alphab'},
    ).to_ctypes()
    _tfac = utils._EngineFloat64ScalarType(
        tfac, {'fun_name': fun_name, 'entity': 'tfac'},
    ).to_ctypes()
    _mxterm = utils._EngineIntScalarType(
        mxterm, {'fun_name': fun_name, 'entity': 'mxterm'},
    ).to_ctypes()
    _nterms = utils._EngineIntScalarType.empty_ctypes()
    _na = utils._EngineIntScalarType.empty_ctypes()
    _alow = utils._EngineFloat64ScalarType.empty_ctypes()
    _ahigh = utils._EngineFloat64ScalarType.empty_ctypes()
    _nfeval = utils._EngineIntScalarType.empty_ctypes()
    _work = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _iuser = None
    _ruser = None
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _f,
        _fh,
        _n,
        _t,
        _valinv,
        _errest,
        _relerr,
        _alphab,
        _tfac,
        _mxterm,
        _nterms,
        _na,
        _alow,
        _ahigh,
        _nfeval,
        _work,
        _iuser,
        _ruser,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, tfac = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: tfac > 0.5.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, relerr = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: 0.0 <= relerr < 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mxterm = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mxterm >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99990:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the elements of t are not strictly increasing.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, t[0] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: 0.0 <= t[0] < t[1] < ... < t[n-1].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, t[n-1] is too large: t[n-1] = {:s}. If necessary,'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'scale the problem as described in [Further Comments].',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Required accuracy cannot be obtained. Try increasing tfac,',
                    'alphab, or both.',
                    'tfac = {:s} and alphab = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99991:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Convergence failure in epsilon algorithm: mxterm = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Some values of valinv[j-1] may be calculated to the desired accuracy; this may',
                    'be determined by examining the values of errest[j-1]. Try reducing the range',
                    'of t or increasing mxterm. If errno = 5 still results, try reducing tfac.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99993:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'All values of valinv have been calculated, but are not of the',
                    'required accuracy; the values of errest[j-1] should be examined carefully. Try',
                    'reducing the range of t, or increasing tfac, alphab or both.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _valinv_cls.refresh_py(_valinv)
        _errest_cls.refresh_py(_errest)
        _nterms_py = utils._EngineIntScalarType.to_py(_nterms)
        _na_py = utils._EngineIntScalarType.to_py(_na)
        _alow_py = utils._EngineFloat64ScalarType.to_py(_alow)
        _ahigh_py = utils._EngineFloat64ScalarType.to_py(_ahigh)
        _nfeval_py = utils._EngineIntScalarType.to_py(_nfeval)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['nterms', 'na', 'alow', 'ahigh', 'nfeval'],
        )(
            _nterms_py,
            _na_py,
            _alow_py,
            _ahigh_py,
            _nfeval_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def invlaplace_weeks(f, sigma0, sigma, b, epstol, mmax, acoef, errvec, data=None):
    u"""
    Inverse Laplace transform, modified Weeks' method.

    ``invlaplace_weeks`` computes the inverse Laplace transform f(t) of
    a user-supplied function F(s), defined for complex s.
    The function uses a modification of Weeks' method which is suitable
    when f(t) has continuous derivatives of all orders.
    The function returns the coefficients of an expansion which
    approximates f(t) and can be evaluated for given values of t by
    subsequent calls of ``invlaplace_weeks_eval``.

    For full information please refer to the NAG Library document for
    c06lb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c06/c06lbf.html

    Parameters
    ----------
    f : callable retval = f(s, data=None)
        `f` must return the value of the Laplace transform function F(s)
        for a given complex value of s.

        Parameters
        ~~~~~~~~~~
        s : complex
            The value of s for which F(s) must be evaluated. The real
            part of `s` is greater than sigma_0.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        retval : complex
            The value of the Laplace transform function F(s) for the
            given complex value, s.

    sigma0 : float
        The abscissa of convergence of the Laplace transform, sigma_0.

    sigma : float
        The parameter sigma of the Laguerre expansion. If on entry
        `sigma` <= sigma_0, `sigma` is reset to sigma_0+0.7.

    b : float
        The parameter b of the Laguerre expansion. If on entry `b` <
        2*(sigma-sigma_0), `b` is reset to 2.5*(sigma-sigma_0).

    epstol : float
        The required relative pseudo-accuracy, that is, an upper bound
        on |f(t)-f~(t)|e^(-sigma t).

    mmax : int
        An upper bound on the number of Laguerre expansion coefficients
        to be computed. The number of coefficients actually computed is
        always a power of 2, so `mmax` should be a power of 2; if `mmax`
        is not a power of 2 then the maximum number of coefficients
        calculated will be the largest power of 2 less than `mmax`.

        Suggested value: `mmax` = 1024 is sufficient for all but a few
        exceptional cases.

    acoef : float, shape (`mmax`), modified in place
        `On exit`: the first `m` elements contain the computed Laguerre
        expansion coefficients, a_i.

    errvec : float, shape (8), modified in place
        `On exit`: an 8-component vector of diagnostic information.

        `errvec`[0]

            Overall estimate of the pseudo-error |f(t)-f~(t)|e^(-sigma
            t) = `errvec`[1]+`errvec`[2]+`errvec`[3].

        `errvec`[1]

            Estimate of the discretization pseudo-error.

        `errvec`[2]

            Estimate of the truncation pseudo-error.

        `errvec`[3]

            Estimate of the condition pseudo-error on the basis of
            minimal noise levels in function values.

        `errvec`[4]

            K, coefficient of a heuristic decay function for the
            expansion coefficients.

        `errvec`[5]

            R, base of the decay function for the expansion
            coefficients.

        `errvec`[6]

            Logarithm of the largest expansion coefficient.

        `errvec`[7]

            Logarithm of the smallest nonzero expansion coefficient.

        The values K and R returned in `errvec`[4] and `errvec`[5]
        define a decay function KR^(-i) constructed by the function for
        the purposes of error estimation.

        It satisfies

            |a_i| < KR^(-i), i = 1,2,...,m.

    data : arbitrary, optional
        User-communication data for callback functions.

    Returns
    -------
    sigma : float
        The value actually used for sigma, as just described.

    b : float
        The value actually used for b, as just described.

    m : int
        The number of Laguerre expansion coefficients actually computed.
        The number of calls to `f` is `m`/2+2.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `mmax` = *<value>*.

            Constraint: `mmax` >= 8.

        (`errno` 4)
            The decay rate of the coefficients is too small. Increasing
            `mmax` may help. `mmax` = *<value>*. Pseudo-error estimate
            `errvec`[0] = *<value>* and `epstol` = *<value>*.

        (`errno` 5)
            The decay rate of the coefficients is too small and
            round-off error is such that the required accuracy cannot be
            obtained. Increasing `mmax` or `epstol` may help. `mmax` =
            *<value>*. Pseudo-error estimate `errvec`[0] = *<value>* and
            `epstol` = *<value>*.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            The estimate of the pseudo-error is slightly larger than
            `epstol`. Pseudo-error estimate `errvec`[0] = *<value>* and
            `epstol` = *<value>*.

        (`errno` 3)
            The round-off error level is larger than `epstol`.
            Increasing `epstol` may help. Pseudo-error estimate
            `errvec`[0] = *<value>* and `epstol` = *<value>*.

        (`errno` 6)
            Error bounds cannot be predicted. Check `sigma0`. `sigma0` =
            *<value>*.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Garbow, B S, Giunta, G, Lyness, J N and Murli, A, 1988, `Software
    for an implementation of Weeks' method for the inverse laplace
    transform problem`, ACM Trans. Math. Software (14), 163--170

    Garbow, B S, Giunta, G, Lyness, J N and Murli, A, 1988, `Algorithm
    662: A Fortran software package for the numerical inversion of the
    Laplace transform based on Weeks' method`, ACM Trans. Math. Software
    (14), 171--176
    """

    excs = []
    fun_name = 'naginterfaces.base.sum.invlaplace_weeks'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    cb_f_name = fun_name + ':f'

    fh = utils._EngineState._sys_functype(
        None,
        utils._UNUSED_FUNCTYPE,
        utils._UNUSED_FUNCTYPE,
        utils._BYREF_ENGINEDATA,
        utils._BYREF_COMPLEX128, # Input, s.
        utils._BYREF_COMPLEX128, # Output, retval.
        utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
    )

    def cb_fh(_f, _fsub, en_data, s, retval, _iuser, _ruser):
        """Helper for f."""
        try:
            _s_fh = utils._EngineComplex128ScalarType.to_py(s.contents)
            if data is not None:
                _retval_fh = f(
                    _s_fh,
                    data,
                )
            else:
                _retval_fh = f(
                    _s_fh,
                )
            utils._EngineComplex128ScalarType(
                _retval_fh, {'fun_name': cb_f_name, 'returned': 'retval'},
            ).refresh_ctypes_buffer(retval)
        except:
            excs.append(_sys.exc_info())
            en_data.contents.hlperr = utils._IERR_HELPER

    engine_ctypes = utils._EngineState._get_symbol('c06lbft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._UNUSED_FUNCTYPE,
        utils._UNUSED_FUNCTYPE,
        fh,
        utils._BYREF_FLOAT64, # Input, sigma0.
        utils._BYREF_FLOAT64, # Input/Output, sigma.
        utils._BYREF_FLOAT64, # Input/Output, b.
        utils._BYREF_FLOAT64, # Input, epstol.
        utils._BYREF_INT, # Input, mmax.
        utils._BYREF_INT, # Output, m.
        utils._EngineFloat64ArrayType.c_type, # Output, acoef[mmax].
        utils._EngineFloat64ArrayType.c_type, # Output, errvec[8].
        utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    utils._check_callable(f, {'fun_name': fun_name, 'entity': 'f'})
    _f = utils._UNUSED_CALLBACK
    _fsub = utils._UNUSED_CALLBACK
    _fh = utils._cast_callback(
        fh, cb_fh,
        {'fun_name': fun_name, 'entity': 'f'},
    )
    _sigma0 = utils._EngineFloat64ScalarType(
        sigma0, {'fun_name': fun_name, 'entity': 'sigma0'},
    ).to_ctypes()
    _sigma = utils._EngineFloat64ScalarType(
        sigma, {'fun_name': fun_name, 'entity': 'sigma'},
    ).to_ctypes()
    _b = utils._EngineFloat64ScalarType(
        b, {'fun_name': fun_name, 'entity': 'b'},
    ).to_ctypes()
    _epstol = utils._EngineFloat64ScalarType(
        epstol, {'fun_name': fun_name, 'entity': 'epstol'},
    ).to_ctypes()
    _mmax = utils._EngineIntScalarType(
        mmax, {'fun_name': fun_name, 'entity': 'mmax'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType.empty_ctypes()
    _mmax_py = utils._EngineIntScalarType.to_py(_mmax)
    _acoef_cls = utils._EngineFloat64ArrayType(
        acoef, {'fun_name': fun_name, 'entity': 'acoef'},
        exp_shape=(_mmax_py,),
        check_contents=False,
    )
    _acoef = _acoef_cls.to_ctypes(can_reference_contents=False)
    _errvec_cls = utils._EngineFloat64ArrayType(
        errvec, {'fun_name': fun_name, 'entity': 'errvec'},
        exp_shape=(8,),
        check_contents=False,
    )
    _errvec = _errvec_cls.to_ctypes(can_reference_contents=False)
    _iuser = None
    _ruser = None
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _f,
        _fsub,
        _fh,
        _sigma0,
        _sigma,
        _b,
        _epstol,
        _mmax,
        _m,
        _acoef,
        _errvec,
        _iuser,
        _ruser,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mmax = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mmax >= 8.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99997:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The estimate of the pseudo-error is slightly larger than epstol.',
                    'Pseudo-error estimate errvec[0] = {:s}'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'and epstol = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99996:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The round-off error level is larger than epstol. Increasing epstol may help.',
                    'Pseudo-error estimate errvec[0] = {:s}'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'and epstol = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The decay rate of the coefficients is too small. Increasing mmax may help.',
                    'mmax = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Pseudo-error estimate errvec[0] = {:s}'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'and epstol = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The decay rate of the coefficients is too small',
                    'and round-off error is such that the required accuracy cannot be obtained.',
                    'Increasing mmax or epstol may help. mmax = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Pseudo-error estimate errvec[0] = {:s}'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'and epstol = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99999:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'Error bounds cannot be predicted. Check sigma0.',
                    'sigma0 = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _sigma_py = utils._EngineFloat64ScalarType.to_py(_sigma)
        _b_py = utils._EngineFloat64ScalarType.to_py(_b)
        _acoef_cls.refresh_py(_acoef)
        _errvec_cls.refresh_py(_errvec)
        _m_py = utils._EngineIntScalarType.to_py(_m)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['sigma', 'b', 'm'],
        )(
            _sigma_py,
            _b_py,
            _m_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def invlaplace_weeks_eval(t, sigma, b, m, acoef, errvec):
    u"""
    Evaluate inverse Laplace transform as computed by
    ``invlaplace_weeks``.

    ``invlaplace_weeks_eval`` evaluates an inverse Laplace transform at
    a given point, using the expansion coefficients computed by
    ``invlaplace_weeks``.

    For full information please refer to the NAG Library document for
    c06lc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c06/c06lcf.html

    Parameters
    ----------
    t : float
        The value t for which the inverse Laplace transform f(t) must be
        evaluated.

    sigma : float
        `sigma` must be unchanged from the previous call of
        ``invlaplace_weeks``

    b : float
        `b` must be unchanged from the previous call of
        ``invlaplace_weeks``

    m : int
        `m` must be unchanged from the previous call of
        ``invlaplace_weeks``

    acoef : float, shape (`m`)
        `acoef` must be unchanged from the previous call of
        ``invlaplace_weeks``

    errvec : float, shape (8)
        `errvec` must be unchanged from the previous call of
        ``invlaplace_weeks``

    Returns
    -------
    finv : float
        The approximation to the inverse Laplace transform at t.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            The approximation to f(t) is too large to be representable.

        (`errno` 2)
            The approximation to f(t) is too small to be representable.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Garbow, B S, Giunta, G, Lyness, J N and Murli, A, 1988, `Algorithm
    662: A Fortran software package for the numerical inversion of the
    Laplace transform based on Weeks' method`, ACM Trans. Math. Software
    (14), 171--176
    """

    excs = []
    fun_name = 'naginterfaces.base.sum.invlaplace_weeks_eval'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('c06lcft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, t.
        utils._BYREF_FLOAT64, # Input, sigma.
        utils._BYREF_FLOAT64, # Input, b.
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Input, acoef[m].
        utils._EngineFloat64ArrayType.c_type, # Input, errvec[8].
        utils._BYREF_FLOAT64, # Output, finv.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _t = utils._EngineFloat64ScalarType(
        t, {'fun_name': fun_name, 'entity': 't'},
    ).to_ctypes()
    _sigma = utils._EngineFloat64ScalarType(
        sigma, {'fun_name': fun_name, 'entity': 'sigma'},
    ).to_ctypes()
    _b = utils._EngineFloat64ScalarType(
        b, {'fun_name': fun_name, 'entity': 'b'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _acoef = utils._EngineFloat64ArrayType(
        acoef, {'fun_name': fun_name, 'entity': 'acoef'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _errvec = utils._EngineFloat64ArrayType(
        errvec, {'fun_name': fun_name, 'entity': 'errvec'},
        exp_shape=(8,),
    ).to_ctypes()
    _finv = utils._EngineFloat64ScalarType.empty_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _t,
        _sigma,
        _b,
        _m,
        _acoef,
        _errvec,
        _finv,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The approximation to f(t) is too large to be representable.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'The approximation to f(t) is too small to be representable.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _finv_py = utils._EngineFloat64ScalarType.to_py(_finv)
        return_data = _finv_py
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = _collections.namedtuple(
                'nag_return_data',
                ['finv'],
            )(
                _finv_py,
            )
    _err_class.manage()
    return return_data

def fft_realherm_1d(direct, x, n):
    u"""
    Single one-dimensional real and Hermitian complex discrete Fourier
    transform, using complex storage format for Hermitian sequences.

    ``fft_realherm_1d`` calculates the discrete Fourier transform of a
    sequence of n real data values or of a Hermitian sequence of n
    complex data values stored in compact form in a `float` array.

    For full information please refer to the NAG Library document for
    c06pa

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c06/c06paf.html

    Parameters
    ----------
    direct : str, length 1
        If the forward transform as defined in [Description] is to be
        computed, `direct` must be set equal to 'F'.

        If the backward transform is to be computed, `direct` must be
        set equal to 'B'.

    x : float, shape (`n`+2), modified in place
        `On entry`: if `x` is declared with bounds (0:`n`+1) in the
        function from which ``fft_realherm_1d`` is called:

            if `direct` = 'F', `x`[j] must contain x_j, for j =
            0,1,...,n-1;

            if `direct` = 'B', `x`[2*k] and `x`[2*k+1] must contain the
            real and imaginary parts respectively of z_k, for k =
            0,1,...,n/2. (Note that for the sequence z_k to be
            Hermitian, the imaginary part of z_0, and of z_{n/2} for n
            even, must be zero.)

        `On exit`:

            if `direct` = 'F' and `x` is declared with bounds (0:`n`+1),
            `x`[2*k] and `x`[2*k+1] will contain the real and imaginary
            parts respectively of z^_k, for k = 0,1,...,n/2;

            if `direct` = 'B' and `x` is declared with bounds (0:`n`+1),
            `x`[j] will contain x^_j, for j = 0,1,...,n-1.

    n : int
        n, the number of data values.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 2)
            On entry, `direct` = *<value>*.

            Constraint: `direct` = 'F' or 'B'.

        (`errno` 3)
            An internal error has occurred in this function. Check the
            function call and any array sizes. If the call is correct
            then please contact NAG for assistance.

    References
    ----------
    Brigham, E O, 1974, `The Fast Fourier Transform`, Prentice--Hall

    Temperton, C, 1983, `Self-sorting mixed-radix fast Fourier
    transforms`, J. Comput. Phys. (52), 1--23
    """

    excs = []
    fun_name = 'naginterfaces.base.sum.fft_realherm_1d'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('c06paft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, direct(len=1).
        utils._EngineFloat64ArrayType.c_type, # Input/Output, x[n+2].
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Output, work[1].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of direct.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _direct = utils._EngineCharScalarType(
        direct, {'fun_name': fun_name, 'entity': 'direct'}, exp_chrlen=1,
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py+2,),
    )
    _x = _x_cls.to_ctypes()
    _work = utils._EngineFloat64ArrayType.empty_ctypes(1)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _direct,
        _x,
        _n,
        _work,
        _errbuf,
        _ierr,
        _direct._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, direct = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: direct = \'F\' or \'B\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'An internal error has occurred in this function.',
                    'Check the function call and any array sizes.',
                    'If the call is correct then please contact NAG for assistance.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def fft_complex_1d(direct, x, n):
    u"""
    Single one-dimensional complex discrete Fourier transform, complex
    data type.

    ``fft_complex_1d`` calculates the discrete Fourier transform of a
    sequence of n complex data values (using complex data type).

    For full information please refer to the NAG Library document for
    c06pc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c06/c06pcf.html

    Parameters
    ----------
    direct : str, length 1
        If the forward transform as defined in [Description] is to be
        computed, `direct` must be set equal to 'F'.

        If the backward transform is to be computed, `direct` must be
        set equal to 'B'.

    x : complex, shape (`n`), modified in place
        `On entry`: if `x` is declared with bounds (0:`n`-1) in the
        function from which ``fft_complex_1d`` is called, `x`[j] must
        contain z_j, for j = 0,1,...,n-1.

        `On exit`: the components of the discrete Fourier transform. If
        `x` is declared with bounds (0:`n`-1) in the function from which
        ``fft_complex_1d`` is called, z^_k is contained in `x`[k], for 0
        <= k <= n-1.

    n : int
        n, the number of data values.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 2)
            On entry, `direct` = *<value>*.

            Constraint: `direct` = 'F' or 'B'.

        (`errno` 4)
            An internal error has occurred in this function. Check the
            function call and any array sizes. If the call is correct
            then please contact NAG for assistance.

    References
    ----------
    Bluestein, L I, 1968, `A linear filtering approach to the
    computation of the discrete Fourier transform`, Northeast
    Electronics Research and Engineering Meeting Record 10, 218--219

    Brigham, E O, 1974, `The Fast Fourier Transform`, Prentice--Hall

    Temperton, C, 1983, `Self-sorting mixed-radix fast Fourier
    transforms`, J. Comput. Phys. (52), 1--23
    """

    excs = []
    fun_name = 'naginterfaces.base.sum.fft_complex_1d'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('c06pcft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, direct(len=1).
        utils._EngineComplex128ArrayType.c_type, # Input/Output, x[n].
        utils._BYREF_INT, # Input, n.
        utils._EngineComplex128ArrayType.c_type, # Output, work[1].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of direct.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _direct = utils._EngineCharScalarType(
        direct, {'fun_name': fun_name, 'entity': 'direct'}, exp_chrlen=1,
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x_cls = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    )
    _x = _x_cls.to_ctypes()
    _work = utils._EngineComplex128ArrayType.empty_ctypes(1)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _direct,
        _x,
        _n,
        _work,
        _errbuf,
        _ierr,
        _direct._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, direct = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: direct = \'F\' or \'B\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'An internal error has occurred in this function.',
                    'Check the function call and any array sizes.',
                    'If the call is correct then please contact NAG for assistance.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def fft_complex_multid_1d(direct, ndim, l, nd, n, x):
    u"""
    One-dimensional complex discrete Fourier transform of
    multidimensional data (using complex data type).

    ``fft_complex_multid_1d`` computes the discrete Fourier transform of
    one variable in a multivariate sequence of complex data values.

    For full information please refer to the NAG Library document for
    c06pf

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c06/c06pff.html

    Parameters
    ----------
    direct : str, length 1
        If the forward transform as defined in [Description] is to be
        computed, `direct` must be set equal to 'F'.

        If the backward transform is to be computed, `direct` must be
        set equal to 'B'.

    ndim : int
        m, the number of dimensions (or variables) in the multivariate
        data.

    l : int
        l, the index of the variable (or dimension) on which the
        discrete Fourier transform is to be performed.

    nd : int, shape (`ndim`)
        The elements of `nd` must contain the dimensions of the `ndim`
        variables; that is, `nd`[i-1] must contain the dimension of the
        ith variable.

    n : int
        n, the total number of data values.

    x : complex, shape (`n`), modified in place
        `On entry`: the complex data values. Data values are stored in
        `x` using column-major ordering for storing multidimensional
        arrays; that is, z_{j_1 j_2...j_m} is stored in `x`[ 1 + j_1 +
        n_1 j_2 + n_1 n_2 j_3 + ... ].

        `On exit`: the corresponding elements of the computed transform.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `ndim` = *<value>*.

            Constraint: `ndim` >= 1.

        (`errno` 2)
            On entry, `l` = *<value>*.

            Constraint: 1 <= `l` <= `ndim`.

        (`errno` 3)
            On entry, `direct` = *<value>*.

            Constraint: `direct` = 'F' or 'B'.

        (`errno` 4)
            On entry, `nd`[*<value>*] = *<value>*.

            Constraint: `nd`[i-1] >= 1, for all i.

        (`errno` 5)
            On entry, `n` = *<value>*, product of `nd` elements is
            *<value>*.

            Constraint: `n` must equal the product of the dimensions
            held in array `nd`.

        (`errno` 8)
            An internal error has occurred in this function. Check the
            function call and any array sizes. If the call is correct
            then please contact NAG for assistance.

    References
    ----------
    Brigham, E O, 1974, `The Fast Fourier Transform`, Prentice--Hall

    Temperton, C, 1983, `Self-sorting mixed-radix fast Fourier
    transforms`, J. Comput. Phys. (52), 1--23
    """

    excs = []
    fun_name = 'naginterfaces.base.sum.fft_complex_multid_1d'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('c06pfft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, iemode.
        utils._BYREF_STRING, # Input, direct(len=1).
        utils._BYREF_INT, # Input, ndim.
        utils._BYREF_INT, # Input, l.
        utils._EngineIntArrayType.c_type, # Input, nd[ndim].
        utils._BYREF_INT, # Input, n.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, x[n].
        utils._EngineComplex128ArrayType.c_type, # Output, work[1].
        utils._BYREF_INT, # Input, lwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of direct.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _iemode = utils._EngineIntScalarType.c_type(0)
    _direct = utils._EngineCharScalarType(
        direct, {'fun_name': fun_name, 'entity': 'direct'}, exp_chrlen=1,
    ).to_ctypes()
    _ndim = utils._EngineIntScalarType(
        ndim, {'fun_name': fun_name, 'entity': 'ndim'},
    ).to_ctypes()
    _l = utils._EngineIntScalarType(
        l, {'fun_name': fun_name, 'entity': 'l'},
    ).to_ctypes()
    _ndim_py = utils._EngineIntScalarType.to_py(_ndim)
    _nd = utils._EngineIntArrayType(
        nd, {'fun_name': fun_name, 'entity': 'nd'},
        exp_shape=(_ndim_py,),
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x_cls = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    )
    _x = _x_cls.to_ctypes()
    _work = utils._EngineComplex128ArrayType.empty_ctypes(1)
    _lwork = utils._EngineIntScalarType.c_type(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _iemode,
        _direct,
        _ndim,
        _l,
        _nd,
        _n,
        _x,
        _work,
        _lwork,
        _errbuf,
        _ierr,
        _direct._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ndim = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ndim >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, l = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: 1 <= l <= ndim.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, direct = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: direct = \'F\' or \'B\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nd[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nd[i-1] >= 1, for all i.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry,',
                    'n = {:s}, product of nd elements is {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: n must equal the product of the dimensions held in array nd.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'An internal error has occurred in this function.',
                    'Check the function call and any array sizes.',
                    'If the call is correct then please contact NAG for assistance.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def fft_complex_multid(direct, ndim, nd, n, x):
    u"""
    Multidimensional complex discrete Fourier transform of
    multidimensional data (using complex data type).

    ``fft_complex_multid`` computes the multidimensional discrete
    Fourier transform of a multivariate sequence of complex data values.

    For full information please refer to the NAG Library document for
    c06pj

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c06/c06pjf.html

    Parameters
    ----------
    direct : str, length 1
        If the forward transform as defined in [Description] is to be
        computed, `direct` must be set equal to 'F'.

        If the backward transform is to be computed, `direct` must be
        set equal to 'B'.

    ndim : int
        m, the number of dimensions (or variables) in the multivariate
        data.

    nd : int, shape (`ndim`)
        The elements of `nd` must contain the dimensions of the `ndim`
        variables; that is, `nd`[i-1] must contain the dimension of the
        ith variable.

    n : int
        n, the total number of data values.

    x : complex, shape (`n`), modified in place
        `On entry`: the complex data values. Data values are stored in
        `x` using column-major ordering for storing multidimensional
        arrays; that is, z_{j_1 j_2...j_m} is stored in `x`[ 1 + j_1 +
        n_1 j_2 + n_1 n_2 j_3 + ... ].

        `On exit`: the corresponding elements of the computed transform.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `ndim` = *<value>*.

            Constraint: `ndim` >= 1.

        (`errno` 2)
            On entry, `direct` = *<value>*.

            Constraint: `direct` = 'F' or 'B'.

        (`errno` 3)
            On entry, `nd`[*<value>*] = *<value>*.

            Constraint: `nd`[i-1] >= 1, for all i.

        (`errno` 4)
            On entry, `n` = *<value>*, product of `nd` elements is
            *<value>*.

            Constraint: `n` must equal the product of the dimensions
            held in array `nd`.

        (`errno` 7)
            An internal error has occurred in this function. Check the
            function call and any array sizes. If the call is correct
            then please contact NAG for assistance.

    References
    ----------
    Brigham, E O, 1974, `The Fast Fourier Transform`, Prentice--Hall

    Temperton, C, 1983, `Self-sorting mixed-radix fast Fourier
    transforms`, J. Comput. Phys. (52), 1--23
    """

    excs = []
    fun_name = 'naginterfaces.base.sum.fft_complex_multid'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('c06pjft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, iemode.
        utils._BYREF_STRING, # Input, direct(len=1).
        utils._BYREF_INT, # Input, ndim.
        utils._EngineIntArrayType.c_type, # Input, nd[ndim].
        utils._BYREF_INT, # Input, n.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, x[n].
        utils._EngineComplex128ArrayType.c_type, # Output, work[1].
        utils._BYREF_INT, # Input, lwork.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of direct.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _iemode = utils._EngineIntScalarType.c_type(0)
    _direct = utils._EngineCharScalarType(
        direct, {'fun_name': fun_name, 'entity': 'direct'}, exp_chrlen=1,
    ).to_ctypes()
    _ndim = utils._EngineIntScalarType(
        ndim, {'fun_name': fun_name, 'entity': 'ndim'},
    ).to_ctypes()
    _ndim_py = utils._EngineIntScalarType.to_py(_ndim)
    _nd = utils._EngineIntArrayType(
        nd, {'fun_name': fun_name, 'entity': 'nd'},
        exp_shape=(_ndim_py,),
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x_cls = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    )
    _x = _x_cls.to_ctypes()
    _work = utils._EngineComplex128ArrayType.empty_ctypes(1)
    _lwork = utils._EngineIntScalarType.c_type(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _iemode,
        _direct,
        _ndim,
        _nd,
        _n,
        _x,
        _work,
        _lwork,
        _errbuf,
        _ierr,
        _direct._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ndim = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ndim >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, direct = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: direct = \'F\' or \'B\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nd[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nd[i-1] >= 1, for all i.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry,',
                    'n = {:s}, product of nd elements is {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: n must equal the product of the dimensions held in array nd.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 99993:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'An internal error has occurred in this function.',
                    'Check the function call and any array sizes.',
                    'If the call is correct then please contact NAG for assistance.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def convcorr_complex(job, x, y, n):
    u"""
    Circular convolution or correlation of two complex vectors.

    ``convcorr_complex`` calculates the circular convolution or
    correlation of two complex vectors of period n.

    For full information please refer to the NAG Library document for
    c06pk

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c06/c06pkf.html

    Parameters
    ----------
    job : int
        The computation to be performed:

        `job` = 1

            z_k = sum_j = 0^n-1x_j y_{k-j} (convolution);

        `job` = 2

            w_k = sum_j = 0^n-1x__j y_{k+j} (correlation).

    x : complex, shape (`n`), modified in place
        `On entry`: the elements of one period of the vector x. If `x`
        is declared with bounds (0:`n`-1) in the function from which
        ``convcorr_complex`` is called, `x`[j] must contain x_j, for j =
        0,1,...,n-1.

        `On exit`: the corresponding elements of the discrete
        convolution or correlation.

    y : complex, shape (`n`), modified in place
        `On entry`: the elements of one period of the vector y. If `y`
        is declared with bounds (0:`n`-1) in the function from which
        ``convcorr_complex`` is called, `y`[j-1] must contain y_j, for j
        = 0,1,...,n-1.

        `On exit`: the discrete Fourier transform of the convolution or
        correlation returned in the array `x`.

    n : int
        n, the number of values in one period of the vectors `x` and
        `y`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 2)
            On entry, argument `job` had an illegal value.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Brigham, E O, 1974, `The Fast Fourier Transform`, Prentice--Hall
    """

    excs = []
    fun_name = 'naginterfaces.base.sum.convcorr_complex'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('c06pkft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, job.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, x[n].
        utils._EngineComplex128ArrayType.c_type, # Input/Output, y[n].
        utils._BYREF_INT, # Input, n.
        utils._EngineComplex128ArrayType.c_type, # Output, work[1].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _job = utils._EngineIntScalarType(
        job, {'fun_name': fun_name, 'entity': 'job'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x_cls = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
    )
    _x = _x_cls.to_ctypes()
    _y_cls = utils._EngineComplex128ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=(_n_py,),
    )
    _y = _y_cls.to_ctypes()
    _work = utils._EngineComplex128ArrayType.empty_ctypes(1)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _job,
        _x,
        _y,
        _n,
        _work,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, argument job had an illegal value.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        _y_cls.refresh_py(_y)
        return_data = None
    _err_class.manage()
    return return_data

def fft_realherm_1d_multi_row(direct, m, n, x):
    u"""
    Multiple one-dimensional real and Hermitian complex discrete Fourier
    transforms, using row ordered complex storage format for Hermitian
    sequences.

    ``fft_realherm_1d_multi_row`` computes the discrete Fourier
    transforms of m sequences, each containing n real data values or a
    Hermitian complex sequence stored in a complex storage format.

    For full information please refer to the NAG Library document for
    c06pp

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c06/c06ppf.html

    Parameters
    ----------
    direct : str, length 1
        If the forward transform as defined in [Description] is to be
        computed, `direct` must be set equal to 'F'.

        If the backward transform is to be computed, `direct` must be
        set equal to 'B'.

    m : int
        m, the number of sequences to be transformed.

    n : int
        n, the number of real or complex values in each sequence.

    x : float, shape (`m`*(`n`+2)), modified in place
        `On entry`: the data must be stored such that consecutive
        elements of the same sequence are stored with a stride of `m`
        and corresponding elements of different sequences are stored
        consecutively. An additional two spaces are reserved for each
        sequence to allow for the pairwise storage of real and imaginary
        parts in the transformed domain. In other words, if the data
        values of the pth sequence to be transformed are denoted by
        x_j^p, for j = 0,1,...,n-1:

            if `direct` = 'F', `x`[j*`m`+p-1] must contain x_j^p, for p
            = 1,2,...,m, for j = 0,1,...,n-1;

            if `direct` = 'B', `x`[2*k*`m`+p-1] and `x`[ (2*k+1) *
            `m`+p-1] must contain the real and imaginary parts
            respectively of z^_k^p, for p = 1,2,...,m, for k =
            0,1,...,n/2. (Note that for the sequence z^_k^p to be
            Hermitian, the imaginary part of z^_0^p, and of z^_n/2^p for
            n even, must be zero.)

        `On exit`:

            if `direct` = 'F' then `x`[2*k*`m`+p-1] and
            `x`[(2*k+1)*`m`+p-1] will contain the real and imaginary
            parts respectively of z^_k^p, for p = 1,2,...,m, for k =
            0,1,...,n/2;

            if `direct` = 'B' then `x`[j*`m`+p-1] will contain x_j^p,
            for p = 1,2,...,m, for j = 0,1,...,n-1;

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 3)
            On entry, `direct` = *<value>*.

            Constraint: `direct` = 'F' or 'B'.

        (`errno` 4)
            An internal error has occurred in this function. Check the
            function call and any array sizes. If the call is correct
            then please contact NAG for assistance.

    References
    ----------
    Brigham, E O, 1974, `The Fast Fourier Transform`, Prentice--Hall

    Temperton, C, 1983, `Fast mixed-radix real Fourier transforms`, J.
    Comput. Phys. (52), 340--350
    """

    excs = []
    fun_name = 'naginterfaces.base.sum.fft_realherm_1d_multi_row'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('c06ppft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, direct(len=1).
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, x[m*(n+2)].
        utils._EngineFloat64ArrayType.c_type, # Output, work[1].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of direct.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _direct = utils._EngineCharScalarType(
        direct, {'fun_name': fun_name, 'entity': 'direct'}, exp_chrlen=1,
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_m_py*(_n_py+2),),
    )
    _x = _x_cls.to_ctypes()
    _work = utils._EngineFloat64ArrayType.empty_ctypes(1)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _direct,
        _m,
        _n,
        _x,
        _work,
        _errbuf,
        _ierr,
        _direct._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, direct = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: direct = \'F\' or \'B\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'An internal error has occurred in this function.',
                    'Check the function call and any array sizes.',
                    'If the call is correct then please contact NAG for assistance.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def fft_realherm_1d_multi_col(direct, n, m, x):
    u"""
    Multiple one-dimensional real and Hermitian complex discrete Fourier
    transforms, using column ordered complex storage format for
    Hermitian sequences.

    ``fft_realherm_1d_multi_col`` computes the discrete Fourier
    transforms of m sequences, each containing n real data values or a
    Hermitian complex sequence stored column-wise in a complex storage
    format.

    For full information please refer to the NAG Library document for
    c06pq

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c06/c06pqf.html

    Parameters
    ----------
    direct : str, length 1
        If the forward transform as defined in [Description] is to be
        computed, `direct` must be set equal to 'F'.

        If the backward transform is to be computed, `direct` must be
        set equal to 'B'.

    n : int
        n, the number of real or complex values in each sequence.

    m : int
        m, the number of sequences to be transformed.

    x : float, shape ((`n`+2)*`m`), modified in place
        `On entry`: the m real or Hermitian data sequences to be
        transformed.

            if `direct` = 'F', the m real data sequences, x^p =
            (x_1^p,x_2^p,...,x_n^p), for p = 1,2,...,m, should be stored
            sequentially in `x`, with a stride of n+2 between sequences.

            if `direct` = 'B', the m Hermitian data sequences, z^^p =
            (z^_1^p,z^_2^p,...,z^_n/2+1^p) =
            (Re(z^_1^p),Im(z^_1^p),Re(z^_2^p),Im(z^_2^p),...,Re(z^_n/2+1^p),Im(z^_n/2+1^p)),
            for p = 1,2,...,m, should be stored sequentially in `x`,
            with a stride of n+2 between sequences.

        In other words:

            if `direct` = 'F', `x`[(p-1)*(`n`+2)+j-1] must contain
            x_j^p, for p = 1,2,...,m, for j = 1,2,...,n;

            if `direct` = 'B', `x`[(p-1) * (`n`+2) + (2*k-1)-1] and
            `x`[(p-1) * (`n`+2) + (2*k)-1] must contain the real and
            imaginary parts respectively of z^_k^p, for p = 1,2,...,m,
            for k = 1,2,...,n/2+1. (Note that for the sequence z^_k^p to
            be Hermitian, the imaginary part of z^_1^p, and of
            z^_n/2+1^p for n even, must be zero.)

        `On exit`:

            if `direct` = 'F' then the m sequences, z^^p, for p =
            1,2,...,m stored as described on entry for `direct` = 'B'

            if `direct` = 'B' then the m sequences, x^p, for p =
            1,2,...,m stored as described on entry for `direct` = 'F'

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 3)
            On entry, `direct` = *<value>*.

            Constraint: `direct` = 'F' or 'B'.

        (`errno` 4)
            An internal error has occurred in this function. Check the
            function call and any array sizes. If the call is correct
            then please contact NAG for assistance.

    References
    ----------
    Brigham, E O, 1974, `The Fast Fourier Transform`, Prentice--Hall

    Temperton, C, 1983, `Fast mixed-radix real Fourier transforms`, J.
    Comput. Phys. (52), 340--350
    """

    excs = []
    fun_name = 'naginterfaces.base.sum.fft_realherm_1d_multi_col'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('c06pqft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, direct(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, x[(n+2)*m].
        utils._EngineFloat64ArrayType.c_type, # Output, work[1].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of direct.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _direct = utils._EngineCharScalarType(
        direct, {'fun_name': fun_name, 'entity': 'direct'}, exp_chrlen=1,
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=((_n_py+2)*_m_py,),
    )
    _x = _x_cls.to_ctypes()
    _work = utils._EngineFloat64ArrayType.empty_ctypes(1)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _direct,
        _n,
        _m,
        _x,
        _work,
        _errbuf,
        _ierr,
        _direct._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, direct = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: direct = \'F\' or \'B\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'An internal error has occurred in this function.',
                    'Check the function call and any array sizes.',
                    'If the call is correct then please contact NAG for assistance.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def fft_complex_1d_multi_row(direct, m, n, x):
    u"""
    Multiple one-dimensional complex discrete Fourier transforms using
    complex data type.

    ``fft_complex_1d_multi_row`` computes the discrete Fourier
    transforms of m sequences, each containing n complex data values.

    For full information please refer to the NAG Library document for
    c06pr

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c06/c06prf.html

    Parameters
    ----------
    direct : str, length 1
        If the forward transform as defined in [Description] is to be
        computed, `direct` must be set equal to 'F'.

        If the backward transform is to be computed, `direct` must be
        set equal to 'B'.

    m : int
        m, the number of sequences to be transformed.

    n : int
        n, the number of complex values in each sequence.

    x : complex, shape (`m`*`n`), modified in place
        `On entry`: the complex data must be stored in `x` as if in a
        two-dimensional array of dimension (1:`m`,0:`n`-1); each of the
        m sequences is stored in a **row** of each array. In other
        words, if the elements of the pth sequence to be transformed are
        denoted by z_j^p, for j = 0,1,...,n-1, `x`[j*`m`+p-1] must
        contain z_j^p.

        `On exit`: is overwritten by the complex transforms.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 3)
            On entry, `direct` = *<value>*.

            Constraint: `direct` = 'F' or 'B'.

        (`errno` 5)
            An internal error has occurred in this function. Check the
            function call and any array sizes. If the call is correct
            then please contact NAG for assistance.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Brigham, E O, 1974, `The Fast Fourier Transform`, Prentice--Hall

    Temperton, C, 1983, `Self-sorting mixed-radix fast Fourier
    transforms`, J. Comput. Phys. (52), 1--23
    """

    excs = []
    fun_name = 'naginterfaces.base.sum.fft_complex_1d_multi_row'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('c06prft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, direct(len=1).
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, x[m*n].
        utils._EngineComplex128ArrayType.c_type, # Output, work[1].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of direct.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _direct = utils._EngineCharScalarType(
        direct, {'fun_name': fun_name, 'entity': 'direct'}, exp_chrlen=1,
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x_cls = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_m_py*_n_py,),
    )
    _x = _x_cls.to_ctypes()
    _work = utils._EngineComplex128ArrayType.empty_ctypes(1)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _direct,
        _m,
        _n,
        _x,
        _work,
        _errbuf,
        _ierr,
        _direct._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, direct = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: direct = \'F\' or \'B\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'An internal error has occurred in this function.',
                    'Check the function call and any array sizes.',
                    'If the call is correct then please contact NAG for assistance.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def fft_complex_1d_multi_col(direct, n, m, x):
    u"""
    Multiple one-dimensional complex discrete Fourier transforms,
    complex data type.

    ``fft_complex_1d_multi_col`` computes the discrete Fourier
    transforms of m sequences, stored as columns of an array, each
    containing n complex data values.

    For full information please refer to the NAG Library document for
    c06ps

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c06/c06psf.html

    Parameters
    ----------
    direct : str, length 1
        If the forward transform as defined in [Description] is to be
        computed, `direct` must be set equal to 'F'.

        If the backward transform is to be computed, `direct` must be
        set equal to 'B'.

    n : int
        n, the number of complex values in each sequence.

    m : int
        m, the number of sequences to be transformed.

    x : complex, shape (`n`*`m`), modified in place
        `On entry`: the complex data values z_j^p stored in
        `x`[(p-1)*`n`+j], for p = 1,2,...,`m`, for j = 0,1,...,`n`-1.

        `On exit`: is overwritten by the complex transforms.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 3)
            On entry, `direct` = *<value>*.

            Constraint: `direct` = 'F' or 'B'.

        (`errno` 5)
            An internal error has occurred in this function. Check the
            function call and any array sizes. If the call is correct
            then please contact NAG for assistance.

    References
    ----------
    Brigham, E O, 1974, `The Fast Fourier Transform`, Prentice--Hall

    Temperton, C, 1983, `Self-sorting mixed-radix fast Fourier
    transforms`, J. Comput. Phys. (52), 1--23
    """

    excs = []
    fun_name = 'naginterfaces.base.sum.fft_complex_1d_multi_col'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('c06psft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, direct(len=1).
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, m.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, x[n*m].
        utils._EngineComplex128ArrayType.c_type, # Output, work[1].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of direct.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _direct = utils._EngineCharScalarType(
        direct, {'fun_name': fun_name, 'entity': 'direct'}, exp_chrlen=1,
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _x_cls = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py*_m_py,),
    )
    _x = _x_cls.to_ctypes()
    _work = utils._EngineComplex128ArrayType.empty_ctypes(1)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _direct,
        _n,
        _m,
        _x,
        _work,
        _errbuf,
        _ierr,
        _direct._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, direct = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: direct = \'F\' or \'B\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'An internal error has occurred in this function.',
                    'Check the function call and any array sizes.',
                    'If the call is correct then please contact NAG for assistance.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def fft_complex_2d(direct, m, n, x):
    u"""
    Two-dimensional complex discrete Fourier transform, complex data
    type.

    ``fft_complex_2d`` computes the two-dimensional discrete Fourier
    transform of a bivariate sequence of complex data values (using
    complex data type).

    For full information please refer to the NAG Library document for
    c06pu

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c06/c06puf.html

    Parameters
    ----------
    direct : str, length 1
        If the forward transform as defined in [Description] is to be
        computed, `direct` must be set equal to 'F'.

        If the backward transform is to be computed, `direct` must be
        set equal to 'B'.

    m : int
        m, the first dimension of the transform.

    n : int
        n, the second dimension of the transform.

    x : complex, shape (`m`*`n`), modified in place
        `On entry`: the complex data values. `x`[`m`*j_2+j_1] must
        contain z_{j_1 j_2}, for j_2 = 0,...,`n`-1, for j_1 =
        0,...,`m`-1.

        `On exit`: the corresponding elements of the computed transform.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 3)
            On entry, `direct` = *<value>*.

            Constraint: `direct` = 'F' or 'B'.

        (`errno` 6)
            An internal error has occurred in this function. Check the
            function call and any array sizes. If the call is correct
            then please contact NAG for assistance.

    References
    ----------
    Brigham, E O, 1974, `The Fast Fourier Transform`, Prentice--Hall

    Temperton, C, 1983, `Self-sorting mixed-radix fast Fourier
    transforms`, J. Comput. Phys. (52), 1--23
    """

    excs = []
    fun_name = 'naginterfaces.base.sum.fft_complex_2d'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('c06puft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, direct(len=1).
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, x[m*n].
        utils._EngineComplex128ArrayType.c_type, # Output, work[1].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of direct.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _direct = utils._EngineCharScalarType(
        direct, {'fun_name': fun_name, 'entity': 'direct'}, exp_chrlen=1,
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x_cls = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_m_py*_n_py,),
    )
    _x = _x_cls.to_ctypes()
    _work = utils._EngineComplex128ArrayType.empty_ctypes(1)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _direct,
        _m,
        _n,
        _x,
        _work,
        _errbuf,
        _ierr,
        _direct._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, direct = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: direct = \'F\' or \'B\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 99994:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'An internal error has occurred in this function.',
                    'Check the function call and any array sizes.',
                    'If the call is correct then please contact NAG for assistance.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def fft_real_2d(m, n, x, y):
    u"""
    Two-dimensional real-to-complex discrete Fourier transform.

    ``fft_real_2d`` computes the two-dimensional discrete Fourier
    transform of a bivariate sequence of real data values.

    For full information please refer to the NAG Library document for
    c06pv

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c06/c06pvf.html

    Parameters
    ----------
    m : int
        m, the first dimension of the transform.

    n : int
        n, the second dimension of the transform.

    x : float, shape (`m`*`n`)
        The real input dataset x, where x_{j_1j_2} is stored in `x`[ j_2
        * m + j_1], for j_2 = 0,1,...,n-1, for j_1 = 0,1,...,m-1.

    y : complex, shape ((`m`/2+1)*`n`), modified in place
        `On exit`: the complex output dataset z^, where z^_{k_1k_2} is
        stored in `y`[ k_2 * (m/2+1) + k_1], for k_2 = 0,1,...,n-1, for
        k_1 = 0,1,...,m/2. Note the first dimension is cut roughly by
        half to remove the redundant information due to conjugate
        symmetry.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 3)
            An internal error has occurred in this function. Check the
            function call and any array sizes. If the call is correct
            then please contact NAG for assistance.

    References
    ----------
    Brigham, E O, 1974, `The Fast Fourier Transform`, Prentice--Hall

    Temperton, C, 1983, `Fast mixed-radix real Fourier transforms`, J.
    Comput. Phys. (52), 340--350
    """

    excs = []
    fun_name = 'naginterfaces.base.sum.fft_real_2d'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('c06pvft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, x[m*n].
        utils._EngineComplex128ArrayType.c_type, # Output, y[(m//2+1)*n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_m_py*_n_py,),
    ).to_ctypes()
    _y_cls = utils._EngineComplex128ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=((_m_py//2+1)*_n_py,),
        check_contents=False,
    )
    _y = _y_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _m,
        _n,
        _x,
        _y,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'An internal error has occurred in this function.',
                    'Check the function call and any array sizes.',
                    'If the call is correct then please contact NAG for assistance.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _y_cls.refresh_py(_y)
        return_data = None
    _err_class.manage()
    return return_data

def fft_hermitian_2d(m, n, y, x):
    u"""
    Two-dimensional complex-to-real discrete Fourier transform.

    ``fft_hermitian_2d`` computes the two-dimensional inverse discrete
    Fourier transform of a bivariate Hermitian sequence of complex data
    values.

    For full information please refer to the NAG Library document for
    c06pw

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c06/c06pwf.html

    Parameters
    ----------
    m : int
        m, the first dimension of the transform.

    n : int
        n, the second dimension of the transform.

    y : complex, shape ((`m`/2+1)*`n`)
        The Hermitian sequence of complex input dataset z, where
        z_{j_1j_2} is stored in `y`[ j_2 * (m/2+1) + j_1], for j_2 =
        0,1,...,n-1, for j_1 = 0,1,...,m/2.

    x : float, shape (`m`*`n`), modified in place
        `On exit`: the real output dataset x^, where x^_{k_1k_2} is
        stored in `x`[ k_2 * m + k_1], for k_2 = 0,1,...,n-1, for k_1 =
        0,1,...,m-1.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 3)
            An internal error has occurred in this function. Check the
            function call and any array sizes. If the call is correct
            then please contact NAG for assistance.

    References
    ----------
    Brigham, E O, 1974, `The Fast Fourier Transform`, Prentice--Hall

    Temperton, C, 1983, `Fast mixed-radix real Fourier transforms`, J.
    Comput. Phys. (52), 340--350
    """

    excs = []
    fun_name = 'naginterfaces.base.sum.fft_hermitian_2d'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('c06pwft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._EngineComplex128ArrayType.c_type, # Input, y[(m//2+1)*n].
        utils._EngineFloat64ArrayType.c_type, # Output, x[m*n].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _y = utils._EngineComplex128ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=((_m_py//2+1)*_n_py,),
    ).to_ctypes()
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_m_py*_n_py,),
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _m,
        _n,
        _y,
        _x,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'An internal error has occurred in this function.',
                    'Check the function call and any array sizes.',
                    'If the call is correct then please contact NAG for assistance.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def fft_complex_3d(direct, n1, n2, n3, x):
    u"""
    Three-dimensional complex discrete Fourier transform, complex data
    type.

    ``fft_complex_3d`` computes the three-dimensional discrete Fourier
    transform of a trivariate sequence of complex data values (using
    complex data type).

    For full information please refer to the NAG Library document for
    c06px

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c06/c06pxf.html

    Parameters
    ----------
    direct : str, length 1
        If the forward transform as defined in [Description] is to be
        computed, `direct` must be set equal to 'F'.

        If the backward transform is to be computed, `direct` must be
        set equal to 'B'.

    n1 : int
        n_1, the first dimension of the transform.

    n2 : int
        n_2, the second dimension of the transform.

    n3 : int
        n_3, the third dimension of the transform.

    x : complex, shape (`n1`*`n2`*`n3`), modified in place
        `On entry`: the complex data values. Data values are stored in
        `x` using column-major ordering for storing multidimensional
        arrays; that is, z_{j_1 j_2 j_3} is stored in `x`[ 1 + j_1 + n_1
        j_2 + n_1 n_2 j_3 ].

        `On exit`: the corresponding elements of the computed transform.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n1` = *<value>*.

            Constraint: `n1` >= 1.

        (`errno` 2)
            On entry, `n2` = *<value>*.

            Constraint: `n2` >= 1.

        (`errno` 3)
            On entry, `n3` = *<value>*.

            Constraint: `n3` >= 1.

        (`errno` 4)
            On entry, `direct` = *<value>*.

            Constraint: `direct` = 'F' or 'B'.

        (`errno` 8)
            An internal error has occurred in this function. Check the
            function call and any array sizes. If the call is correct
            then please contact NAG for assistance.

    References
    ----------
    Brigham, E O, 1974, `The Fast Fourier Transform`, Prentice--Hall

    Temperton, C, 1983, `Self-sorting mixed-radix fast Fourier
    transforms`, J. Comput. Phys. (52), 1--23
    """

    excs = []
    fun_name = 'naginterfaces.base.sum.fft_complex_3d'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('c06pxft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, direct(len=1).
        utils._BYREF_INT, # Input, n1.
        utils._BYREF_INT, # Input, n2.
        utils._BYREF_INT, # Input, n3.
        utils._EngineComplex128ArrayType.c_type, # Input/Output, x[n1*n2*n3].
        utils._EngineComplex128ArrayType.c_type, # Output, work[1].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of direct.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _direct = utils._EngineCharScalarType(
        direct, {'fun_name': fun_name, 'entity': 'direct'}, exp_chrlen=1,
    ).to_ctypes()
    _n1 = utils._EngineIntScalarType(
        n1, {'fun_name': fun_name, 'entity': 'n1'},
    ).to_ctypes()
    _n2 = utils._EngineIntScalarType(
        n2, {'fun_name': fun_name, 'entity': 'n2'},
    ).to_ctypes()
    _n3 = utils._EngineIntScalarType(
        n3, {'fun_name': fun_name, 'entity': 'n3'},
    ).to_ctypes()
    _n1_py = utils._EngineIntScalarType.to_py(_n1)
    _n2_py = utils._EngineIntScalarType.to_py(_n2)
    _n3_py = utils._EngineIntScalarType.to_py(_n3)
    _x_cls = utils._EngineComplex128ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n1_py*_n2_py*_n3_py,),
    )
    _x = _x_cls.to_ctypes()
    _work = utils._EngineComplex128ArrayType.empty_ctypes(1)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _direct,
        _n1,
        _n2,
        _n3,
        _x,
        _work,
        _errbuf,
        _ierr,
        _direct._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n1 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n1 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n2 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n2 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n3 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n3 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, direct = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: direct = \'F\' or \'B\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 99992:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'An internal error has occurred in this function.',
                    'Check the function call and any array sizes.',
                    'If the call is correct then please contact NAG for assistance.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def fft_real_3d(n1, n2, n3, x, y):
    u"""
    Three-dimensional real-to-complex discrete Fourier transform.

    ``fft_real_3d`` computes the three-dimensional discrete Fourier
    transform of a trivariate sequence of real data values.

    For full information please refer to the NAG Library document for
    c06py

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c06/c06pyf.html

    Parameters
    ----------
    n1 : int
        n_1, the first dimension of the transform.

    n2 : int
        n_2, the second dimension of the transform.

    n3 : int
        n_3, the third dimension of the transform.

    x : float, shape (`n1`*`n2`*`n3`)
        The real input dataset x, where x_{j_1j_2j_3} is stored in `x`[
        j_3 * n_1 n_2 + j_2 * n_1 + j_1 +1 -1], for j_3 = 0,1,...,n_3-1,
        for j_2 = 0,1,...,n_2-1, for j_1 = 0,1,...,n_1-1.

    y : complex, shape ((`n1`/2+1)*`n2`*`n3`), modified in place
        `On exit`: the complex output dataset z^, where z^_{k_1k_2k_3}
        is stored in `y`[ k_3 * (n_1/2+1) n_2 + k_2 * (n_1/2+1) + k_1 +1
        -1], for k_3 = 0,1,...,n_3-1, for k_2 = 0,1,...,n_2-1, for k_1 =
        0,1,...,n_1/2. Note the first dimension is cut roughly by half
        to remove the redundant information due to conjugate symmetry.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n1` = *<value>*.

            Constraint: `n1` >= 1.

        (`errno` 2)
            On entry, `n2` = *<value>*.

            Constraint: `n2` >= 1.

        (`errno` 3)
            On entry, `n3` = *<value>*.

            Constraint: `n3` >= 1.

        (`errno` 4)
            An internal error has occurred in this function. Check the
            function call and any array sizes. If the call is correct
            then please contact NAG for assistance.

    References
    ----------
    Brigham, E O, 1974, `The Fast Fourier Transform`, Prentice--Hall

    Temperton, C, 1983, `Fast mixed-radix real Fourier transforms`, J.
    Comput. Phys. (52), 340--350
    """

    excs = []
    fun_name = 'naginterfaces.base.sum.fft_real_3d'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('c06pyft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n1.
        utils._BYREF_INT, # Input, n2.
        utils._BYREF_INT, # Input, n3.
        utils._EngineFloat64ArrayType.c_type, # Input, x[n1*n2*n3].
        utils._EngineComplex128ArrayType.c_type, # Output, y[(n1//2+1)*n2*n3].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n1 = utils._EngineIntScalarType(
        n1, {'fun_name': fun_name, 'entity': 'n1'},
    ).to_ctypes()
    _n2 = utils._EngineIntScalarType(
        n2, {'fun_name': fun_name, 'entity': 'n2'},
    ).to_ctypes()
    _n3 = utils._EngineIntScalarType(
        n3, {'fun_name': fun_name, 'entity': 'n3'},
    ).to_ctypes()
    _n1_py = utils._EngineIntScalarType.to_py(_n1)
    _n2_py = utils._EngineIntScalarType.to_py(_n2)
    _n3_py = utils._EngineIntScalarType.to_py(_n3)
    _x = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n1_py*_n2_py*_n3_py,),
    ).to_ctypes()
    _y_cls = utils._EngineComplex128ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=((_n1_py//2+1)*_n2_py*_n3_py,),
        check_contents=False,
    )
    _y = _y_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n1,
        _n2,
        _n3,
        _x,
        _y,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n1 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n1 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n2 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n2 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n3 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n3 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'An internal error has occurred in this function.',
                    'Check the function call and any array sizes.',
                    'If the call is correct then please contact NAG for assistance.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _y_cls.refresh_py(_y)
        return_data = None
    _err_class.manage()
    return return_data

def fft_hermitian_3d(n1, n2, n3, y, x):
    u"""
    Three-dimensional complex-to-real discrete Fourier transform.

    ``fft_hermitian_3d`` computes the three-dimensional inverse discrete
    Fourier transform of a trivariate Hermitian sequence of complex data
    values.

    For full information please refer to the NAG Library document for
    c06pz

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c06/c06pzf.html

    Parameters
    ----------
    n1 : int
        n_1, the first dimension of the transform.

    n2 : int
        n_2, the second dimension of the transform.

    n3 : int
        n_3, the third dimension of the transform.

    y : complex, shape ((`n1`/2+1)*`n2`*`n3`)
        The Hermitian sequence of complex input dataset z, where
        z_{j_1j_2j_3} is stored in `y`[ j_3 * (n_1/2+1) n_2 + j_2 *
        (n_1/2+1) + j_1 +1 -1], for j_3 = 0,1,...,n_3-1, for j_2 =
        0,1,...,n_2-1, for j_1 = 0,1,...,n_1/2.

    x : float, shape (`n1`*`n2`*`n3`), modified in place
        `On exit`: the real output dataset x^, where x^_{k_1k_2k_3} is
        stored in `x`[ k_3 * n_1 n_2 + k_2 * n_1 + k_1 +1 -1], for k_3 =
        0,1,...,n_3-1, for k_2 = 0,1,...,n_2-1, for k_1 = 0,1,...,n_1-1.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n1` = *<value>*.

            Constraint: `n1` >= 1.

        (`errno` 2)
            On entry, `n2` = *<value>*.

            Constraint: `n2` >= 1.

        (`errno` 3)
            On entry, `n3` = *<value>*.

            Constraint: `n3` >= 1.

        (`errno` 4)
            An internal error has occurred in this function. Check the
            function call and any array sizes. If the call is correct
            then please contact NAG for assistance.

    References
    ----------
    Brigham, E O, 1974, `The Fast Fourier Transform`, Prentice--Hall

    Temperton, C, 1983, `Fast mixed-radix real Fourier transforms`, J.
    Comput. Phys. (52), 340--350
    """

    excs = []
    fun_name = 'naginterfaces.base.sum.fft_hermitian_3d'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('c06pzft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n1.
        utils._BYREF_INT, # Input, n2.
        utils._BYREF_INT, # Input, n3.
        utils._EngineComplex128ArrayType.c_type, # Input, y[(n1//2+1)*n2*n3].
        utils._EngineFloat64ArrayType.c_type, # Output, x[n1*n2*n3].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n1 = utils._EngineIntScalarType(
        n1, {'fun_name': fun_name, 'entity': 'n1'},
    ).to_ctypes()
    _n2 = utils._EngineIntScalarType(
        n2, {'fun_name': fun_name, 'entity': 'n2'},
    ).to_ctypes()
    _n3 = utils._EngineIntScalarType(
        n3, {'fun_name': fun_name, 'entity': 'n3'},
    ).to_ctypes()
    _n1_py = utils._EngineIntScalarType.to_py(_n1)
    _n2_py = utils._EngineIntScalarType.to_py(_n2)
    _n3_py = utils._EngineIntScalarType.to_py(_n3)
    _y = utils._EngineComplex128ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=((_n1_py//2+1)*_n2_py*_n3_py,),
    ).to_ctypes()
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n1_py*_n2_py*_n3_py,),
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n1,
        _n2,
        _n3,
        _y,
        _x,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n1 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n1 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n2 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n2 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n3 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n3 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99995:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'An internal error has occurred in this function.',
                    'Check the function call and any array sizes.',
                    'If the call is correct then please contact NAG for assistance.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def fft_real_sine_simple(m, n, x):
    u"""
    Discrete sine transform (easy-to-use).

    ``fft_real_sine_simple`` computes the discrete Fourier sine
    transforms of m sequences of real data values.

    For full information please refer to the NAG Library document for
    c06ra

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c06/c06raf.html

    Parameters
    ----------
    m : int
        m, the number of sequences to be transformed.

    n : int
        One more than the number of real values in each sequence, i.e.,
        the number of values in each sequence is n-1.

    x : float, shape (`m`*(`n`+2)), modified in place
        `On entry`: the data must be stored in `x` as if in a
        two-dimensional array of dimension (1:`m`,1:`n`+2); each of the
        m sequences is stored in a **row** of the array. In other words,
        if the n-1 data values of the pth sequence to be transformed are
        denoted by x_j^p, for p = 1,2,...,m, for j = 1,2,...,n-1, the
        first m(n-1) elements of the array `x` must contain the values

x_1^1,x_1^2,...,x_1^m,x_2^1,x_2^2,...,x_2^m,...,x_n-1^1,x_n-1^2,...,x_n-1^m.

        The nth to (n+2)th elements of each row x_n^p,...,x_n+2^p, for p
        = 1,2,...,m, are required as workspace. These 3m elements may
        contain arbitrary values as they are set to zero by the
        function.

        `On exit`: the m Fourier sine transforms stored as if in a
        two-dimensional array of dimension (1:`m`,1:`n`+2). Each of the
        m transforms is stored in a **row** of the array, overwriting
        the corresponding original sequence. If the (n-1) components of
        the pth Fourier sine transform are denoted by x^_k^p, for p =
        1,2,...,m, for k = 1,2,...,n-1, the m(n+2) elements of the array
        `x` contain the values

x^_1^1,x^_1^2,...,x^_1^m,x^_2^1,x^_2^2,...,x^_2^m,...,x^_n-1^1,x^_n-1^2,...,x^_n-1^m,0,0,...,0
            (3m times).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 3)
            An internal error has occurred in this function. Check the
            function call and any array sizes. If the call is correct
            then please contact NAG for assistance.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Brigham, E O, 1974, `The Fast Fourier Transform`, Prentice--Hall

    Swarztrauber, P N, 1977, `The methods of cyclic reduction, Fourier
    analysis and the FACR algorithm for the discrete solution of
    Poisson's equation on a rectangle`, SIAM Rev. (19(3)), 490--501

    Swarztrauber, P N, 1982, `Vectorizing the FFT's`, Parallel
    Computation, (ed G Rodrique), 51--83, Academic Press

    Temperton, C, 1983, `Fast mixed-radix real Fourier transforms`, J.
    Comput. Phys. (52), 340--350
    """

    excs = []
    fun_name = 'naginterfaces.base.sum.fft_real_sine_simple'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('c06raft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, x[m*(n+2)].
        utils._EngineFloat64ArrayType.c_type, # Output, work[1].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_m_py*(_n_py+2),),
    )
    _x = _x_cls.to_ctypes()
    _work = utils._EngineFloat64ArrayType.empty_ctypes(1)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _m,
        _n,
        _x,
        _work,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'An internal error has occurred in this function.',
                    'Check the function call and any array sizes.',
                    'If the call is correct then please contact NAG for assistance.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def fft_real_cosine_simple(m, n, x):
    u"""
    Discrete cosine transform (easy-to-use).

    ``fft_real_cosine_simple`` computes the discrete Fourier cosine
    transforms of m sequences of real data values.

    For full information please refer to the NAG Library document for
    c06rb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c06/c06rbf.html

    Parameters
    ----------
    m : int
        m, the number of sequences to be transformed.

    n : int
        One less than the number of real values in each sequence, i.e.,
        the number of values in each sequence is n+1.

    x : float, shape (`m`*(`n`+3)), modified in place
        `On entry`: the data must be stored in `x` as if in a
        two-dimensional array of dimension (1:`m`,0:`n`+2); each of the
        m sequences is stored in a **row** of the array. In other words,
        if the (n+1) data values of the pth sequence to be transformed
        are denoted by x_j^p, for p = 1,2,...,m, for j = 0,1,...,n, the
        first m(n+1) elements of the array `x` must contain the values

x_0^1,x_0^2,...,x_0^m,x_1^1,x_1^2,...,x_1^m,...,x_n^1,x_n^2,...,x_n^m.

        The (n+2)th and (n+3)th elements of each row x_n+1^p,x_n+2^p,
        for p = 1,2,...,m, are required as workspace. These 2m elements
        may contain arbitrary values as they are set to zero by the
        function.

        `On exit`: the m Fourier cosine transforms stored as if in a
        two-dimensional array of dimension (1:`m`,0:`n`+2). Each of the
        m transforms is stored in a **row** of the array, overwriting
        the corresponding original data. If the (n+1) components of the
        pth Fourier cosine transform are denoted by x^_k^p, for p =
        1,2,...,m, for k = 0,1,...,n, the m(n+3) elements of the array
        `x` contain the values

x^_0^1,x^_0^2,...,x^_0^m,x^_1^1,x^_1^2,...,x^_1^m,...,x^_n^1,x^_n^2,...,x^_n^m,0,0,...,0
            (2m times).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 3)
            An internal error has occurred in this function. Check the
            function call and any array sizes. If the call is correct
            then please contact NAG for assistance.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Brigham, E O, 1974, `The Fast Fourier Transform`, Prentice--Hall

    Swarztrauber, P N, 1977, `The methods of cyclic reduction, Fourier
    analysis and the FACR algorithm for the discrete solution of
    Poisson's equation on a rectangle`, SIAM Rev. (19(3)), 490--501

    Swarztrauber, P N, 1982, `Vectorizing the FFT's`, Parallel
    Computation, (ed G Rodrique), 51--83, Academic Press

    Temperton, C, 1983, `Fast mixed-radix real Fourier transforms`, J.
    Comput. Phys. (52), 340--350
    """

    excs = []
    fun_name = 'naginterfaces.base.sum.fft_real_cosine_simple'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('c06rbft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, x[m*(n+3)].
        utils._EngineFloat64ArrayType.c_type, # Output, work[1].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_m_py*(_n_py+3),),
    )
    _x = _x_cls.to_ctypes()
    _work = utils._EngineFloat64ArrayType.empty_ctypes(1)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _m,
        _n,
        _x,
        _work,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'An internal error has occurred in this function.',
                    'Check the function call and any array sizes.',
                    'If the call is correct then please contact NAG for assistance.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def fft_real_qtrsine_simple(direct, m, n, x):
    u"""
    Discrete quarter-wave sine transform (easy-to-use).

    ``fft_real_qtrsine_simple`` computes the discrete quarter-wave
    Fourier sine transforms of m sequences of real data values.

    For full information please refer to the NAG Library document for
    c06rc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c06/c06rcf.html

    Parameters
    ----------
    direct : str, length 1
        If the forward transform as defined in [Description] is to be
        computed, `direct` must be set equal to 'F'.

        If the backward transform is to be computed, `direct` must be
        set equal to 'B'.

    m : int
        m, the number of sequences to be transformed.

    n : int
        n, the number of real values in each sequence.

    x : float, shape (`m`*(`n`+2)), modified in place
        `On entry`: the data must be stored in `x` as if in a
        two-dimensional array of dimension (1:`m`,1:`n`+2); each of the
        m sequences is stored in a **row** of the array. In other words,
        if the data values of the pth sequence to be transformed are
        denoted by x_j^p, for p = 1,2,...,m, for j = 1,2,...,n, the
        first mn elements of the array `x` must contain the values

x_1^1,x_1^2,...,x_1^m,x_2^1,x_2^2,...,x_2^m,...,x_n^1,x_n^2,...,x_n^m.

        The (n+1)th and (n+2)th elements of each row x_n+1^p,x_n+2^p,
        for p = 1,2,...,m, are required as workspace. These 2m elements
        may contain arbitrary values as they are set to zero by the
        function.

        `On exit`: the m quarter-wave sine transforms stored as if in a
        two-dimensional array of dimension (1:`m`,1:`n`+2). Each of the
        m transforms is stored in a **row** of the array, overwriting
        the corresponding original sequence. If the n components of the
        pth quarter-wave sine transform are denoted by x^_k^p, for p =
        1,2,...,m, for k = 1,2,...,n, the m(n+2) elements of the array
        `x` contain the values

x^_1^1,x^_1^2,...,x^_1^m,x^_2^1,x^_2^2,...,x^_2^m,...,x^_n^1,x^_n^2,...,x^_n^m,0,0,...,0
            (2m times).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 3)
            On entry, `direct` = *<value>*.

            Constraint: `direct` = 'F' or 'B'.

        (`errno` 4)
            An internal error has occurred in this function. Check the
            function call and any array sizes. If the call is correct
            then please contact NAG for assistance.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Brigham, E O, 1974, `The Fast Fourier Transform`, Prentice--Hall

    Swarztrauber, P N, 1977, `The methods of cyclic reduction, Fourier
    analysis and the FACR algorithm for the discrete solution of
    Poisson's equation on a rectangle`, SIAM Rev. (19(3)), 490--501

    Swarztrauber, P N, 1982, `Vectorizing the FFT's`, Parallel
    Computation, (ed G Rodrique), 51--83, Academic Press

    Temperton, C, 1983, `Fast mixed-radix real Fourier transforms`, J.
    Comput. Phys. (52), 340--350
    """

    excs = []
    fun_name = 'naginterfaces.base.sum.fft_real_qtrsine_simple'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('c06rcft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, direct(len=1).
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, x[m*(n+2)].
        utils._EngineFloat64ArrayType.c_type, # Output, work[1].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of direct.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _direct = utils._EngineCharScalarType(
        direct, {'fun_name': fun_name, 'entity': 'direct'}, exp_chrlen=1,
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_m_py*(_n_py+2),),
    )
    _x = _x_cls.to_ctypes()
    _work = utils._EngineFloat64ArrayType.empty_ctypes(1)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _direct,
        _m,
        _n,
        _x,
        _work,
        _errbuf,
        _ierr,
        _direct._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, direct = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: direct = \'F\' or \'B\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'An internal error has occurred in this function.',
                    'Check the function call and any array sizes.',
                    'If the call is correct then please contact NAG for assistance.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def fft_real_qtrcosine_simple(direct, m, n, x):
    u"""
    Discrete quarter-wave cosine transform (easy-to-use).

    ``fft_real_qtrcosine_simple`` computes the discrete quarter-wave
    Fourier cosine transforms of m sequences of real data values.

    For full information please refer to the NAG Library document for
    c06rd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c06/c06rdf.html

    Parameters
    ----------
    direct : str, length 1
        If the forward transform as defined in [Description] is to be
        computed, `direct` must be set equal to 'F'.

        If the backward transform is to be computed, `direct` must be
        set equal to 'B'.

    m : int
        m, the number of sequences to be transformed.

    n : int
        n, the number of real values in each sequence.

    x : float, shape (`m`*(`n`+2)), modified in place
        `On entry`: the data must be stored in `x` as if in a
        two-dimensional array of dimension (1:`m`,0:`n`+1); each of the
        m sequences is stored in a **row** of the array. In other words,
        if the data values of the pth sequence to be transformed are
        denoted by x_j^p, for p = 1,2,...,m, for j = 0,1,...,n-1, the
        first mn elements of the array `x` must contain the values

x_0^1,x_0^2,...,x_0^m,x_1^1,x_1^2,...,x_1^m,...,x_n-1^1,x_n-1^2,...,x_n-1^m.

        The (n+1)th and (n+2)th elements of each row x_n^p,x_n+1^p, for
        p = 1,2,...,m, are required as workspace. These 2m elements may
        contain arbitrary values as they are set to zero by the
        function.

        `On exit`: the m quarter-wave cosine transforms stored as if in
        a two-dimensional array of dimension (1:`m`,0:`n`+1). Each of
        the m transforms is stored in a **row** of the array,
        overwriting the corresponding original sequence. If the n
        components of the pth quarter-wave cosine transform are denoted
        by x^_k^p, for p = 1,2,...,m, for k = 0,1,...,n-1, the m(n+2)
        elements of the array `x` contain the values

x^_0^1,x^_0^2,...,x^_0^m,x^_1^1,x^_1^2,...,x^_1^m,...,x^_n-1^1,x^_n-1^2,...,x^_n-1^m,0,0,...,0
            (2m times).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 3)
            On entry, `direct` = *<value>*.

            Constraint: `direct` = 'F' or 'B'.

        (`errno` 4)
            An internal error has occurred in this function. Check the
            function call and any array sizes. If the call is correct
            then please contact NAG for assistance.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Brigham, E O, 1974, `The Fast Fourier Transform`, Prentice--Hall

    Swarztrauber, P N, 1977, `The methods of cyclic reduction, Fourier
    analysis and the FACR algorithm for the discrete solution of
    Poisson's equation on a rectangle`, SIAM Rev. (19(3)), 490--501

    Swarztrauber, P N, 1982, `Vectorizing the FFT's`, Parallel
    Computation, (ed G Rodrique), 51--83, Academic Press

    Temperton, C, 1983, `Fast mixed-radix real Fourier transforms`, J.
    Comput. Phys. (52), 340--350
    """

    excs = []
    fun_name = 'naginterfaces.base.sum.fft_real_qtrcosine_simple'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('c06rdft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, direct(len=1).
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, x[m*(n+2)].
        utils._EngineFloat64ArrayType.c_type, # Output, work[1].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of direct.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _direct = utils._EngineCharScalarType(
        direct, {'fun_name': fun_name, 'entity': 'direct'}, exp_chrlen=1,
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_m_py*(_n_py+2),),
    )
    _x = _x_cls.to_ctypes()
    _work = utils._EngineFloat64ArrayType.empty_ctypes(1)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _direct,
        _m,
        _n,
        _x,
        _work,
        _errbuf,
        _ierr,
        _direct._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, direct = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'Constraint: direct = \'F\' or \'B\'.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'An internal error has occurred in this function.',
                    'Check the function call and any array sizes.',
                    'If the call is correct then please contact NAG for assistance.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def fft_sine(m, n, x):
    u"""
    Multiple discrete sine transforms, simple.

    ``fft_sine`` computes the discrete Fourier sine transforms of m
    sequences of real data values.
    The elements of each sequence and its transform are stored
    contiguously.

    For full information please refer to the NAG Library document for
    c06re

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c06/c06ref.html

    Parameters
    ----------
    m : int
        m, the number of sequences to be transformed.

    n : int
        One more than the number of real values in each sequence, i.e.,
        the number of values in each sequence is n-1.

    x : float, shape (`n`-1, `m`), modified in place
        `On entry`: the data values of the pth sequence to be
        transformed, denoted by x_j^p, for p = 1,2,...,m, for j =
        1,2,...,n-1, must be stored in `x`[j-1,p-1].

        `On exit`: the (n-1) components of the pth Fourier sine
        transform, denoted by x^_k^p, for p = 1,2,...,m, for k =
        1,2,...,n-1, are stored in `x`[k-1,p-1], overwriting the
        corresponding original values.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 3)
            An internal error has occurred in this function. Check the
            function call and any array sizes. If the call is correct
            then please contact NAG for assistance.

    References
    ----------
    Brigham, E O, 1974, `The Fast Fourier Transform`, Prentice--Hall

    Swarztrauber, P N, 1977, `The methods of cyclic reduction, Fourier
    analysis and the FACR algorithm for the discrete solution of
    Poisson's equation on a rectangle`, SIAM Rev. (19(3)), 490--501

    Swarztrauber, P N, 1982, `Vectorizing the FFT's`, Parallel
    Computation, (ed G Rodrique), 51--83, Academic Press

    Temperton, C, 1983, `Fast mixed-radix real Fourier transforms`, J.
    Comput. Phys. (52), 340--350
    """

    excs = []
    fun_name = 'naginterfaces.base.sum.fft_sine'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('c06reft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, sorder.
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, x[ldx, m].
        utils._BYREF_INT, # Input, ldx.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((x, 'x'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _sorder = utils._EngineIntScalarType.c_type(_nag_enum_consts.nag_col_major)
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _ldx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py-1, _m_py),
        mutual_sorder=_m_sorder,
    )
    _x = _x_cls.to_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _sorder,
        _m,
        _n,
        _x,
        _ldx,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'An internal error has occurred in this function.',
                    'Check the function call and any array sizes.',
                    'If the call is correct then please contact NAG for assistance.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def fft_cosine(m, n, x):
    u"""
    Multiple discrete cosine transforms, simple.

    ``fft_cosine`` computes the discrete Fourier cosine transforms of m
    sequences of real data values.
    The elements of each sequence and its transform are stored
    contiguously.

    For full information please refer to the NAG Library document for
    c06rf

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c06/c06rff.html

    Parameters
    ----------
    m : int
        m, the number of sequences to be transformed.

    n : int
        One less than the number of real values in each sequence, i.e.,
        the number of values in each sequence is n+1.

    x : float, shape (`n`+1, `m`), modified in place
        `On entry`: the data values of the pth sequence to be
        transformed, denoted by x_j^p, for p = 1,2,...,m, for j =
        0,1,...,n, must be stored in `x`[j,p-1].

        `On exit`: the (n+1) components of the pth Fourier cosine
        transform, denoted by x^_k^p, for p = 1,2,...,m, for k =
        0,1,...,n, are stored in `x`[k,p-1], overwriting the
        corresponding original values.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 3)
            An internal error has occurred in this function. Check the
            function call and any array sizes. If the call is correct
            then please contact NAG for assistance.

    References
    ----------
    Brigham, E O, 1974, `The Fast Fourier Transform`, Prentice--Hall

    Swarztrauber, P N, 1977, `The methods of cyclic reduction, Fourier
    analysis and the FACR algorithm for the discrete solution of
    Poisson's equation on a rectangle`, SIAM Rev. (19(3)), 490--501

    Swarztrauber, P N, 1982, `Vectorizing the FFT's`, Parallel
    Computation, (ed G Rodrique), 51--83, Academic Press

    Temperton, C, 1983, `Fast mixed-radix real Fourier transforms`, J.
    Comput. Phys. (52), 340--350
    """

    excs = []
    fun_name = 'naginterfaces.base.sum.fft_cosine'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('c06rfft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, x[n+1, m].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((x, 'x'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py+1, _m_py),
        mutual_sorder=_m_sorder,
    )
    _x = _x_cls.to_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _m,
        _n,
        _x,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'An internal error has occurred in this function.',
                    'Check the function call and any array sizes.',
                    'If the call is correct then please contact NAG for assistance.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def fft_qtrsine(idir, m, n, x):
    u"""
    Multiple discrete quarter-wave sine transforms, simple.

    ``fft_qtrsine`` computes the discrete quarter-wave Fourier sine
    transforms of m sequences of real data values.
    The elements of each sequence and its transform are stored
    contiguously.

    For full information please refer to the NAG Library document for
    c06rg

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c06/c06rgf.html

    Parameters
    ----------
    idir : int
        Indicates the transform, as defined in [Description], to be
        computed.

        `idir` = 1

            Forward transform.

        `idir` = -1

            Inverse transform.

    m : int
        m, the number of sequences to be transformed.

    n : int
        n, the number of real values in each sequence.

    x : float, shape (`n`, `m`), modified in place
        `On entry`: the data values of the pth sequence to be
        transformed, denoted by x_j^p, for p = 1,2,...,m, for j =
        1,2,...,n, must be stored in `x`[j-1,p-1].

        `On exit`: the n components of the pth quarter-wave sine
        transform, denoted by x^_k^p, for p = 1,2,...,m, for k =
        1,2,...,n, are stored in `x`[k-1,p-1], overwriting the
        corresponding original values.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 3)
            On entry, `idir` = *<value>*.

            Constraint: `idir` = -1 or 1.

        (`errno` 4)
            An internal error has occurred in this function. Check the
            function call and any array sizes. If the call is correct
            then please contact NAG for assistance.

    References
    ----------
    Brigham, E O, 1974, `The Fast Fourier Transform`, Prentice--Hall

    Swarztrauber, P N, 1977, `The methods of cyclic reduction, Fourier
    analysis and the FACR algorithm for the discrete solution of
    Poisson's equation on a rectangle`, SIAM Rev. (19(3)), 490--501

    Swarztrauber, P N, 1982, `Vectorizing the FFT's`, Parallel
    Computation, (ed G Rodrique), 51--83, Academic Press

    Temperton, C, 1983, `Fast mixed-radix real Fourier transforms`, J.
    Comput. Phys. (52), 340--350
    """

    excs = []
    fun_name = 'naginterfaces.base.sum.fft_qtrsine'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('c06rgft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, idir.
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, x[n, m].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((x, 'x'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _idir = utils._EngineIntScalarType(
        idir, {'fun_name': fun_name, 'entity': 'idir'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py, _m_py),
        mutual_sorder=_m_sorder,
    )
    _x = _x_cls.to_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _idir,
        _m,
        _n,
        _x,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, idir = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: idir = -1 or 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'An internal error has occurred in this function.',
                    'Check the function call and any array sizes.',
                    'If the call is correct then please contact NAG for assistance.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def fft_qtrcosine(idir, m, n, x):
    u"""
    Multiple discrete quarter-wave cosine transforms, simple.

    ``fft_qtrcosine`` computes the discrete quarter-wave Fourier cosine
    transforms of m sequences of real data values.
    The elements of each sequence and its transform are stored
    contiguously.

    For full information please refer to the NAG Library document for
    c06rh

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c06/c06rhf.html

    Parameters
    ----------
    idir : int
        Indicates the transform, as defined in [Description], to be
        computed.

        `idir` = 1

            Forward transform.

        `idir` = -1

            Inverse transform.

    m : int
        m, the number of sequences to be transformed.

    n : int
        n, the number of real values in each sequence.

    x : float, shape (`n`, `m`), modified in place
        `On entry`: the data values of the pth sequence to be
        transformed, denoted by x_j^p, for p = 1,2,...,m, for j =
        0,1,...,n-1, must be stored in `x`[j,p-1].

        `On exit`: the n components of the pth quarter-wave cosine
        transform, denoted by x^_k^p, for p = 1,2,...,m, for k =
        0,1,...,n-1, are stored in `x`[k,p-1], overwriting the
        corresponding original values.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 3)
            On entry, `idir` = *<value>*.

            Constraint: `idir` = -1 or 1.

        (`errno` 4)
            An internal error has occurred in this function. Check the
            function call and any array sizes. If the call is correct
            then please contact NAG for assistance.

    References
    ----------
    Brigham, E O, 1974, `The Fast Fourier Transform`, Prentice--Hall

    Swarztrauber, P N, 1977, `The methods of cyclic reduction, Fourier
    analysis and the FACR algorithm for the discrete solution of
    Poisson's equation on a rectangle`, SIAM Rev. (19(3)), 490--501

    Swarztrauber, P N, 1982, `Vectorizing the FFT's`, Parallel
    Computation, (ed G Rodrique), 51--83, Academic Press

    Temperton, C, 1983, `Fast mixed-radix real Fourier transforms`, J.
    Comput. Phys. (52), 340--350
    """

    excs = []
    fun_name = 'naginterfaces.base.sum.fft_qtrcosine'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('c06rhft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, idir.
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, x[n, m].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((x, 'x'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _idir = utils._EngineIntScalarType(
        idir, {'fun_name': fun_name, 'entity': 'idir'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py, _m_py),
        mutual_sorder=_m_sorder,
    )
    _x = _x_cls.to_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _idir,
        _m,
        _n,
        _x,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 99999:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 99998:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 99997:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, idir = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: idir = -1 or 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 99996:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'An internal error has occurred in this function.',
                    'Check the function call and any array sizes.',
                    'If the call is correct then please contact NAG for assistance.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def fast_gauss(d, srcs, n, trgs, m, q, p1, p2, k, hin, lhin, tol, v, term):
    u"""
    Multidimensional fast Gauss transform.

    ``fast_gauss`` calculates the multidimensional fast Gauss transform.

    For full information please refer to the NAG Library document for
    c06sa

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/c06/c06saf.html

    Parameters
    ----------
    d : int
        d, the number of dimensions.

    srcs : float, shape (`d`, `n`)
        x, the locations of the Gaussian sources.

    n : int
        n, the number of Gaussian sources.

    trgs : float, shape (`d`, `m`)
        y, the locations of the target points at which the FGT will be
        evaluated.

    m : int
        m, the number of target points.

    q : float, shape (`n`)
        q, the weights of the Gaussian sources.

    p1 : int
        p_1, the number of terms of the first Taylor series to be
        evaluated.

    p2 : int
        p_2, the number of terms of the second Taylor series to be
        evaluated.

    k : int
        k, the number of clusters into which the source points will be
        aggregated.

    hin : float, shape (`lhin`)
        h, the standard deviations of the Gaussian sources. If `lhin` <
        `n`, the array will be expanded automatically by repeating `hin`
        until it is of length `n`. See [the G01 Introduction] for
        further information.

    lhin : int
        The length of the array `hin`.

    tol : float
        epsilon, the desired accuracy of the FGT approximation of the
        DGT. Determines the radius of the source clusters: the
        contribution of a source point to the FGT approximation at a
        target point is disregarded if the source is outside the
        corresponding cluster radius.

    v : float, shape (`m`), modified in place
        `On exit`: G^(y), the value of the FGT evaluated at y.

    term : float, shape (`m`), modified in place
        `On exit`: `term`[j-1] contains the absolute value of the final
        Taylor series term that is largest, relative to the size of the
        sum of the corresponding series, across all clusters that
        contribute to the FGT at target point `v`[j-1].

    Returns
    -------
    p1 : int
        `p1` is unchanged.

    p2 : int
        `p2` is unchanged.

    k : int
        `k` is unchanged.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `d` = *<value>*.

            Constraint: `d` > 0.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` > 0.

        (`errno` 3)
            On entry, `m` = *<value>*.

            Constraint: `m` > 0.

        (`errno` 4)
            On entry, `p1` = *<value>*.

            Constraint: `p1` > 0.

        (`errno` 5)
            On entry, `p2` = *<value>*.

            Constraint: `p2` > 0.

        (`errno` 6)
            On entry, `k` = *<value>* and `n` = *<value>*.

            Constraint: 1 <= `k` <= `n`.

        (`errno` 7)
            On entry, `hin`[*<value>*] = *<value>*.

            Constraint: `hin`[i] > 0.0, for i = 0,...,`lhin`-1.

        (`errno` 8)
            On entry, `lhin` = *<value>*.

            Constraint: 1 <= `lhin` <= `n`.

        (`errno` 9)
            On entry, `tol` = *<value>*.

            Constraint: `tol` > 0.0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 10)
            On exit, `p1` = *<value>*, `p2` = *<value>* and `k` =
            *<value>*.

            `p1`, `p2` or `k` may have been too small to calculate
            `v`[`m`-1] to the required accuracy `tol`.

    References
    ----------
    Greengard, L and Strain, J, 1991, `The Fast Gauss Transform`, SIAM
    J. Sci. Statist. Comput. (12(1)), 79--94

    Raykar, V C and Duraiswami, R, 2005, `Improved Fast Gauss Transform
    With Variable Source Scales`, University of Maryland Technical
    Report CS-TR-4727/UMIACS-TR-2005-34
    """

    excs = []
    fun_name = 'naginterfaces.base.sum.fast_gauss'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('c06saft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, d.
        utils._EngineFloat64ArrayType.c_type, # Input, srcs[d, n].
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, trgs[d, m].
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Input, q[n].
        utils._BYREF_INT, # Input/Output, p1.
        utils._BYREF_INT, # Input/Output, p2.
        utils._BYREF_INT, # Input/Output, k.
        utils._EngineFloat64ArrayType.c_type, # Input, hin[lhin].
        utils._BYREF_INT, # Input, lhin.
        utils._BYREF_FLOAT64, # Input, tol.
        utils._EngineFloat64ArrayType.c_type, # Output, v[m].
        utils._EngineFloat64ArrayType.c_type, # Output, term[m].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((srcs, 'srcs'), (trgs, 'trgs')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _d = utils._EngineIntScalarType(
        d, {'fun_name': fun_name, 'entity': 'd'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _d_py = utils._EngineIntScalarType.to_py(_d)
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _srcs = utils._EngineFloat64ArrayType(
        srcs, {'fun_name': fun_name, 'entity': 'srcs'},
        exp_shape=(_d_py, _n_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _trgs = utils._EngineFloat64ArrayType(
        trgs, {'fun_name': fun_name, 'entity': 'trgs'},
        exp_shape=(_d_py, _m_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _q = utils._EngineFloat64ArrayType(
        q, {'fun_name': fun_name, 'entity': 'q'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _p1 = utils._EngineIntScalarType(
        p1, {'fun_name': fun_name, 'entity': 'p1'},
    ).to_ctypes()
    _p2 = utils._EngineIntScalarType(
        p2, {'fun_name': fun_name, 'entity': 'p2'},
    ).to_ctypes()
    _k = utils._EngineIntScalarType(
        k, {'fun_name': fun_name, 'entity': 'k'},
    ).to_ctypes()
    _lhin = utils._EngineIntScalarType(
        lhin, {'fun_name': fun_name, 'entity': 'lhin'},
    ).to_ctypes()
    _lhin_py = utils._EngineIntScalarType.to_py(_lhin)
    _hin = utils._EngineFloat64ArrayType(
        hin, {'fun_name': fun_name, 'entity': 'hin'},
        exp_shape=(_lhin_py,),
    ).to_ctypes()
    _tol = utils._EngineFloat64ScalarType(
        tol, {'fun_name': fun_name, 'entity': 'tol'},
    ).to_ctypes()
    _v_cls = utils._EngineFloat64ArrayType(
        v, {'fun_name': fun_name, 'entity': 'v'},
        exp_shape=(_m_py,),
        check_contents=False,
    )
    _v = _v_cls.to_ctypes(can_reference_contents=False)
    _term_cls = utils._EngineFloat64ArrayType(
        term, {'fun_name': fun_name, 'entity': 'term'},
        exp_shape=(_m_py,),
        check_contents=False,
    )
    _term = _term_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _d,
        _srcs,
        _n,
        _trgs,
        _m,
        _q,
        _p1,
        _p2,
        _k,
        _hin,
        _lhin,
        _tol,
        _v,
        _term,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1001:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, d = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: d > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 1002:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 1003:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 1004:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, p1 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: p1 > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 1005:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, p2 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: p2 > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 1006:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, k = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: 1 <= k <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 1007:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, hin[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: hin[i] > 0.0, for i = 0,...,lhin-1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 1008:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lhin = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: 1 <= lhin <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 9 and _ifmt == 1009:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, tol = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: tol > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 10 and _ifmt == 1010:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'On exit, p1 = {:s}, p2 = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'and k = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'p1, p2 or k may have been too small to calculate v[m-1] to',
                    'the required accuracy tol.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _p1_py = utils._EngineIntScalarType.to_py(_p1)
        _p2_py = utils._EngineIntScalarType.to_py(_p2)
        _k_py = utils._EngineIntScalarType.to_py(_k)
        _v_cls.refresh_py(_v)
        _term_cls.refresh_py(_term)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['p1', 'p2', 'k'],
        )(
            _p1_py,
            _p2_py,
            _k_py,
        )
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data
