# -*- coding: utf-8 -*-
# pylint: disable=chained-comparison,invalid-name,line-too-long,protected-access,too-many-arguments,too-many-boolean-expressions,too-many-branches,too-many-lines,too-many-locals,too-many-statements,unused-import
u"""
Interfaces for the NAG Mark 27.0 `fit` Chapter.

``fit`` - Curve and Surface Fitting

The main aim of this module is to assist you in finding a function which
approximates a set of data points.
Typically the data contain random errors, as of experimental
measurement, which need to be smoothed out.
To seek an approximation to the data, it is first necessary to specify
for the approximating function a mathematical form (a polynomial, for
example) which contains a number of unspecified coefficients: the
appropriate fitting function then derives for the coefficients the
values which provide the best fit of that particular form.
The module deals mainly with curve and surface fitting (i.e., fitting
with functions of one and of two variables) when a polynomial or a cubic
spline is used as the fitting function, since these cover the most
common needs.
However, fitting with other functions and/or more variables can be
undertaken by means of general linear or nonlinear functions (some of
which are contained in other modules) depending on whether the
coefficients in the function occur linearly or nonlinearly.
Cases where a graph rather than a set of data points is given can be
treated simply by first reading a suitable set of points from the graph.

The module also contains functions for evaluating, differentiating and
integrating polynomial and spline curves and surfaces, once the
numerical values of their coefficients have been determined.

There is also a function for computing a Padé approximant of a
mathematical function (see [Background to the Problems]).

For full information please refer to the NAG Library document

https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e02/e02intro.html

See Also
--------
``naginterfaces.library.examples.fit`` :
    This subpackage contains examples for the ``fit`` module.
    See also the :ref:`library_fit_ex` subsection.
"""

# NAG Copyright 2017-2020.

import collections as _collections

from ..base import fit as _b_fit
from ..base import _nag_e_consts
from ..base import utils as _b_utils
from ..library import _utils

def dim1_cheb_arb(kplus1, x, y, w):
    u"""
    Least squares curve fit, by polynomials, arbitrary data points.

    ``dim1_cheb_arb`` computes weighted least squares polynomial
    approximations to an arbitrary set of data points.

    For full information please refer to the NAG Library document for
    e02ad

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e02/e02adf.html

    Parameters
    ----------
    kplus1 : int
        k+1, where k is the maximum degree required.

    x : float, array-like, shape (m)
        The values x_r of the independent variable, for r = 1,2,...,m.

    y : float, array-like, shape (m)
        The values y_r of the dependent variable, for r = 1,2,...,m.

    w : float, array-like, shape (m)
        The set of weights, w_r, for r = 1,2,...,m. For advice on the
        choice of weights, see [the E02 Introduction].

    Returns
    -------
    a : float, ndarray, shape (`kplus1`, `kplus1`)
        The coefficients of T_j(x_) in the approximating polynomial of
        degree i. `a`[i,j] contains the coefficient a_{i+1,j+1}, for j =
        0,1,...,i, for i = 0,1,...,k.

    s : float, ndarray, shape (`kplus1`)
        `s`[i] contains the root-mean-square residual s_i, for i =
        0,1,...,k, as described in [Description]. For the interpretation
        of the values of the s_i and their use in selecting an
        appropriate degree, see [the E02 Introduction].

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, i = *<value>* and `w`[i-1] = *<value>*.

            Constraint: `w`[i-1] > 0.0.

        (`errno` 2)
            On entry, i = *<value>*, `x`[i-1] = *<value>* and `x`[i-2] =
            *<value>*.

            Constraint: `x`[i-1] >= `x`[i-2].

        (`errno` 3)
            On entry, all `x`[I-1] have the same value: `x`[0] =
            *<value>*.

        (`errno` 4)
            On entry, `kplus1` = *<value>* and mdist = *<value>*.

            Constraint: `kplus1` <= mdist, where mdist is the number of
            distinct `x` values.

        (`errno` 4)
            On entry, `kplus1` = *<value>* and m = *<value>*.

            Constraint: `kplus1` <= m.

        (`errno` 4)
            On entry, `kplus1` = *<value>*.

            Constraint: `kplus1` >= 1.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Clenshaw, C W, 1960, `Curve fitting with a digital computer`,
    Comput. J. (2), 170--173

    Cox, M G, 1974, `A data-fitting package for the non-specialist
    user`, Software for Numerical Mathematics, (ed D J Evans), Academic
    Press

    Cox, M G and Hayes, J G, 1973, `Curve fitting: a guide and suite of
    algorithms for the non-specialist user`, NPL Report NAC26, National
    Physical Laboratory

    Forsythe, G E, 1957, `Generation and use of orthogonal polynomials
    for data fitting with a digital computer`, J. Soc. Indust. Appl.
    Math. (5), 74--88

    Gentleman, W M, 1969, `An error analysis of Goertzel's (Watt's)
    method for computing Fourier coefficients`, Comput. J. (12),
    160--165

    Hayes, J G (ed.), 1970, `Numerical Approximation to Functions and
    Data`, Athlone Press, London
    """

    fun_name = 'naginterfaces.library.fit.dim1_cheb_arb'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _w = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'w'},
        w,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _kplus1 = _b_utils._EngineIntScalarType(
        kplus1, {'fun_name': fun_name, 'entity': 'kplus1'},
    ).to_ctypes()
    _kplus1_py = _b_utils._EngineIntScalarType.to_py(_kplus1)
    _a = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_kplus1_py, _kplus1_py),
        sorder=_b_utils._EngineData.default_sorder,
    )
    _s = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_kplus1_py,),
    )
    _b_fit.dim1_cheb_arb(
        _m,
        _kplus1,
        _x,
        _y,
        _w,
        _a,
        _s,
    )
    _a_py = _a
    _s_py = _s
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 's'],
    )(
        _a_py,
        _s_py,
    )
    return _return_data

def dim1_cheb_eval(a, xcap):
    u"""
    Evaluation of fitted polynomial in one variable from Chebyshev
    series form (simplified parameter list).

    ``dim1_cheb_eval`` evaluates a polynomial from its Chebyshev series
    representation.

    For full information please refer to the NAG Library document for
    e02ae

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e02/e02aef.html

    Parameters
    ----------
    a : float, array-like, shape (nplus1)
        `a`[i-1] must be set to the value of the ith coefficient in the
        series, for i = 1,2,...,n+1.

    xcap : float
        x_, the argument at which the polynomial is to be evaluated. It
        should lie in the range -1 to +1, but a value just outside this
        range is permitted (see [Raises/Warns]) to allow for possible
        rounding errors committed in the transformation from x to x_
        discussed in [Description]. Provided the recommended form of the
        transformation is used, a successful exit is thus assured
        whenever the value of x lies in the range x_min to x_max.

    Returns
    -------
    p : float
        The value of the polynomial.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `xcap` = *<value>* and EPS = *<value>*.

            Constraint: abs(`xcap`) <= 1+4*EPS, where EPS is machine
            precision.

        (`errno` 2)
            On entry, nplus1 = *<value>*.

            Constraint: nplus1 >= 1.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Clenshaw, C W, 1955, `A note on the summation of Chebyshev series`,
    Math. Tables Aids Comput. (9), 118--120

    Cox, M G, 1974, `A data-fitting package for the non-specialist
    user`, Software for Numerical Mathematics, (ed D J Evans), Academic
    Press

    Cox, M G and Hayes, J G, 1973, `Curve fitting: a guide and suite of
    algorithms for the non-specialist user`, NPL Report NAC26, National
    Physical Laboratory

    Gentleman, W M, 1969, `An error analysis of Goertzel's (Watt's)
    method for computing Fourier coefficients`, Comput. J. (12),
    160--165
    """

    fun_name = 'naginterfaces.library.fit.dim1_cheb_eval'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nplus1_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=1)[0]
    _nplus1 = _b_utils._EngineIntScalarType(
        _nplus1_py, {'fun_name': fun_name, 'entity': '_nplus1'},
    ).to_ctypes()
    _return_scalar = _b_fit.dim1_cheb_eval(
        _nplus1,
        _a,
        xcap,
    )
    _p_py = _return_scalar
    return _p_py

def dim1_cheb_glp(f):
    u"""
    Least squares polynomial fit, special data points (including
    interpolation).

    ``dim1_cheb_glp`` computes the coefficients of a polynomial, in its
    Chebyshev series form, which interpolates (passes exactly through)
    data at a special set of points.
    Least squares polynomial approximations can also be obtained.

    For full information please refer to the NAG Library document for
    e02af

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e02/e02aff.html

    Parameters
    ----------
    f : float, array-like, shape (nplus1)
        For r = 1,2,...,n+1, `f`[r-1] must contain f_r the value of the
        dependent variable (ordinate) corresponding to the value

            x__r = cos(pi(r-1)/n)

        of the independent variable (abscissa) x_, or equivalently to
        the value

            x(r) = 1/2(x_max-x_min)cos(pi(r-1)/n)+1/2(x_max+x_min)

        of your original variable x. Here x_max and x_min are
        respectively the upper and lower ends of the range over which
        you wish to interpolate.

    Returns
    -------
    a : float, ndarray, shape (nplus1)
        `a`[j-1] is the coefficient a_j in the interpolating polynomial,
        for j = 1,2,...,n+1.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, nplus1 = *<value>*.

            Constraint: nplus1 >= 2.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Clenshaw, C W, 1955, `A note on the summation of Chebyshev series`,
    Math. Tables Aids Comput. (9), 118--120

    Clenshaw, C W, 1962, `Chebyshev Series for Mathematical Functions`,
    Mathematical tables, HMSO

    Cox, M G, 1974, `A data-fitting package for the non-specialist
    user`, Software for Numerical Mathematics, (ed D J Evans), Academic
    Press

    Cox, M G and Hayes, J G, 1973, `Curve fitting: a guide and suite of
    algorithms for the non-specialist user`, NPL Report NAC26, National
    Physical Laboratory

    Gentleman, W M, 1969, `An error analysis of Goertzel's (Watt's)
    method for computing Fourier coefficients`, Comput. J. (12),
    160--165
    """

    fun_name = 'naginterfaces.library.fit.dim1_cheb_glp'

    _f = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'f'},
        f,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nplus1_py = _b_utils._EngineFloat64ArrayType.get_shape(_f, {'fun_name': fun_name, 'entity': 'f'}, exp_rank=1)[0]
    _nplus1 = _b_utils._EngineIntScalarType(
        _nplus1_py, {'fun_name': fun_name, 'entity': '_nplus1'},
    ).to_ctypes()
    _a = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nplus1_py,),
    )
    _b_fit.dim1_cheb_glp(
        _nplus1,
        _f,
        _a,
    )
    _a_py = _a
    return _a_py

def dim1_cheb_con(k, xmin, xmax, x, y, w, xf, yf, ip):
    u"""
    Least squares polynomial fit, values and derivatives may be
    constrained, arbitrary data points.

    ``dim1_cheb_con`` computes constrained weighted least squares
    polynomial approximations in Chebyshev series form to an arbitrary
    set of data points.
    The values of the approximations and any number of their derivatives
    can be specified at selected points.

    For full information please refer to the NAG Library document for
    e02ag

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e02/e02agf.html

    Parameters
    ----------
    k : int
        k, the maximum degree required.

    xmin : float
        The lower and upper end points, respectively, of the interval
        [x_min,x_max]. Unless there are specific reasons to the
        contrary, it is recommended that `xmin` and `xmax` be set
        respectively to the lowest and highest value among the x_r and
        xf_r. This avoids the danger of extrapolation provided there is
        a constraint point or data point with nonzero weight at each end
        point.

    xmax : float
        The lower and upper end points, respectively, of the interval
        [x_min,x_max]. Unless there are specific reasons to the
        contrary, it is recommended that `xmin` and `xmax` be set
        respectively to the lowest and highest value among the x_r and
        xf_r. This avoids the danger of extrapolation provided there is
        a constraint point or data point with nonzero weight at each end
        point.

    x : float, array-like, shape (m)
        `x`[r-1] must contain the value x_r of the independent variable
        at the rth data point, for r = 1,2,...,m.

    y : float, array-like, shape (m)
        `y`[r-1] must contain y_r, the value of the dependent variable
        at the rth data point, for r = 1,2,...,m.

    w : float, array-like, shape (m)
        `w`[r-1] must contain the weight w_r to be applied to the data
        point x_r, for r = 1,2,...,m. For advice on the choice of
        weights see [the E02 Introduction]. Negative weights are treated
        as positive. A zero weight causes the corresponding data point
        to be ignored. Zero weight should be given to any data point
        whose x and y values both coincide with those of a constraint
        (otherwise the denominators involved in the root mean square
        residuals S_i will be slightly in error).

    xf : float, array-like, shape (mf)
        `xf`[r-1] must contain xf_r, the value of the independent
        variable at which a constraint is specified, for r = 1,2,...,mf.

    yf : float, array-like, shape (mf+sum(`ip`))
        The values which the approximating polynomials and their
        derivatives are required to take at the points specified in
        `xf`. For each value of `xf`[r-1], `yf` contains in successive
        elements the required value of the approximation, its first
        derivative, second derivative, ...,p_rth derivative, for r =
        1,2,...,mf. Thus the value, yf_s, which the kth derivative of
        each approximation (k = 0 referring to the approximation itself)
        is required to take at the point `xf`[r-1] must be contained in
        `yf`[s-1], where

            s = r+k+p_1+p_2+...+p_{r-1},

        where k = 0,1,...,p_r and r = 1,2,...,mf. The derivatives are
        with respect to the independent variable x.

    ip : int, array-like, shape (mf)
        `ip`[r-1] must contain p_r, the order of the highest-order
        derivative specified at `xf`[r-1], for r = 1,2,...,mf. p_r = 0
        implies that the value of the approximation at `xf`[r-1] is
        specified, but not that of any derivative.

    Returns
    -------
    a : float, ndarray, shape (`k`+1, `k`+1)
        `a`[i,j] contains the coefficient a_{ij} in the approximating
        polynomial of degree i, for j = 0,1,...,i, for i = n,...,k.

    s : float, ndarray, shape (`k`+1)
        `s`[i] contains S_i, for i = n,...,k, the root mean square
        residual corresponding to the approximating polynomial of degree
        i. In the case where the number of data points with nonzero
        weight is equal to k+1-n, S_i is indeterminate: the function
        sets it to zero. For the interpretation of the values of S_i and
        their use in selecting an appropriate degree, see [the E02
        Introduction].

    n : int
        Contains the total number of constraint conditions imposed: `n`
        = mf+p_1+p_2+...+p_mf.

    resid : float, ndarray, shape (m)
        Contains weighted residuals of the highest degree of fit
        determined (k). The residual at x_r is in element `resid`[r-1],
        for r = 1,2,...,m.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, m = *<value>*.

            Constraint: m >= 1.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `k` >= `n`.

        (`errno` 1)
            On entry, mf = *<value>*.

            Constraint: mf >= 1.

        (`errno` 2)
            On entry, I = *<value>* and `ip`[I-1] = *<value>*.

            Constraint: `ip`[I-1] >= 0.

        (`errno` 3)
            On entry, I = *<value>*, `xf`[I-1] = *<value>*, J =
            *<value>* and `xf`[J-1] = *<value>*.

            Constraint: `xf`[I-1] != `xf`[J-1].

        (`errno` 3)
            On entry, `xf`[I-1] lies outside interval [`xmin`,`xmax`]: I
            = *<value>*, `xf`[I-1] = *<value>*, `xmin` = *<value>* and
            `xmax` = *<value>*.

        (`errno` 3)
            On entry, `xmin` = *<value>* and `xmax` = *<value>*.

            Constraint: `xmin` < `xmax`.

        (`errno` 4)
            On entry, `x`[I-1] lies outside interval [`xmin`,`xmax`] for
            some I.

        (`errno` 4)
            On entry, `x`[I-1] lies outside interval [`xmin`,`xmax`]: I
            = *<value>*, `x`[I-1] = *<value>*, `xmin` = *<value>* and
            `xmax` = *<value>*.

        (`errno` 5)
            On entry, i = *<value>*, `x`[i-1] = *<value>* and `x`[i-2] =
            *<value>*.

            Constraint: `x`[i-1] >= `x`[i-2].

        (`errno` 6)
            On entry, `k`+1 > m^('')+`n`, where m^('') is the number of
            data points with nonzero weight and distinct abscissae
            different from all `xf`, and `n` is the total number of
            constraints: `k`+1 = *<value>*, m^('') = *<value>* and `n` =
            *<value>*.

        (`errno` 7)
            The polynomials mu(x) and/or nu(x) cannot be found. The
            problem is too ill-conditioned.

    References
    ----------
    Clenshaw, C W, 1960, `Curve fitting with a digital computer`,
    Comput. J. (2), 170--173

    Forsythe, G E, 1957, `Generation and use of orthogonal polynomials
    for data fitting with a digital computer`, J. Soc. Indust. Appl.
    Math. (5), 74--88

    Hayes, J G (ed.), 1970, `Numerical Approximation to Functions and
    Data`, Athlone Press, London
    """

    fun_name = 'naginterfaces.library.fit.dim1_cheb_con'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _w = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'w'},
        w,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _xf = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xf'},
        xf,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _mf_py = _b_utils._EngineFloat64ArrayType.get_shape(_xf, {'fun_name': fun_name, 'entity': 'xf'}, exp_rank=1)[0]
    _mf = _b_utils._EngineIntScalarType(
        _mf_py, {'fun_name': fun_name, 'entity': '_mf'},
    ).to_ctypes()
    _yf = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'yf'},
        yf,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lyf_py = _b_utils._EngineFloat64ArrayType.get_shape(_yf, {'fun_name': fun_name, 'entity': 'yf'}, exp_rank=1)[0]
    _lyf = _b_utils._EngineIntScalarType(
        _lyf_py, {'fun_name': fun_name, 'entity': '_lyf'},
    ).to_ctypes()
    _ip = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ip'},
        ip,
        copy=False,
    )
    _k = _b_utils._EngineIntScalarType(
        k, {'fun_name': fun_name, 'entity': 'k'},
    ).to_ctypes()
    _k_py = _b_utils._EngineIntScalarType.to_py(_k)
    _a = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_k_py+1, _k_py+1),
        sorder=_b_utils._EngineData.default_sorder,
    )
    _s = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_k_py+1,),
    )
    _resid = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_m_py,),
    )
    _return_scalar = _b_fit.dim1_cheb_con(
        _m,
        _k,
        xmin,
        xmax,
        _x,
        _y,
        _w,
        _mf,
        _xf,
        _yf,
        _lyf,
        _ip,
        _a,
        _s,
        _resid,
    )
    _a_py = _a
    _s_py = _s
    _n_py = _return_scalar
    _resid_py = _resid
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 's', 'n', 'resid'],
    )(
        _a_py,
        _s_py,
        _n_py,
        _resid_py,
    )
    return _return_data

def dim1_cheb_deriv(n, xmin, xmax, a, ia1, iadif1):
    u"""
    Derivative of fitted polynomial in Chebyshev series form.

    ``dim1_cheb_deriv`` determines the coefficients in the Chebyshev
    series representation of the derivative of a polynomial given in
    Chebyshev series form.

    For full information please refer to the NAG Library document for
    e02ah

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e02/e02ahf.html

    Parameters
    ----------
    n : int
        n, the degree of the given polynomial p(x).

    xmin : float
        The lower and upper end points respectively of the interval
        [x_min,x_max]. The Chebyshev series representation is in terms
        of the normalized variable x_, where

            x_= (2x-(x_max+x_min))/(x_max-x_min).

    xmax : float
        The lower and upper end points respectively of the interval
        [x_min,x_max]. The Chebyshev series representation is in terms
        of the normalized variable x_, where

            x_= (2x-(x_max+x_min))/(x_max-x_min).

    a : float, array-like, shape (1+`n`*`ia1`)
        The Chebyshev coefficients of the polynomial p(x). Specifically,
        element i*`ia1` of `a` must contain the coefficient a_i, for i =
        0,1,...,n. Only these n+1 elements will be accessed.

        Unchanged on exit, but see `adif`, below.

    ia1 : int
        The index increment of `a`. Most frequently the Chebyshev
        coefficients are stored in adjacent elements of `a`, and `ia1`
        must be set to 1. However, if for example, they are stored in
        `a`[0],`a`[3],`a`[6],..., the value of `ia1` must be 3. See also
        [Further Comments].

    iadif1 : int
        The index increment of `adif`. Most frequently the Chebyshev
        coefficients are required in adjacent elements of `adif`, and
        `iadif1` must be set to 1. However, if, for example, they are to
        be stored in `adif`[0],`adif`[3],`adif`[6],..., the value of
        `iadif1` must be 3. See [Further Comments].

    Returns
    -------
    patm1 : float
        The value of p(x_min). If this value is passed to the
        integration function ``dim1_cheb_integ`` with the coefficients
        of q(x), the original polynomial p(x) is recovered, including
        its constant coefficient.

    adif : float, ndarray, shape (1+`n`*`iadif1`)
        The Chebyshev coefficients of the derived polynomial q(x). (The
        differentiation is with respect to the variable x.)
        Specifically, element i*`iadif1` of `adif` contains the
        coefficient a__i, for i = 0,1,...,n-1. Additionally, element
        n*`iadif1` is set to zero. A call of the function may have the
        array name `adif` the same as `a`, provided that note is taken
        of the order in which elements are overwritten, when choosing
        the starting elements and increments `ia1` and `iadif1`, i.e.,
        the coefficients a_0,a_1,...,a_{i-1} must be intact after
        coefficient a__i is stored. In particular, it is possible to
        overwrite the a_i completely by having `ia1` = `iadif1`, and the
        actual arrays for `a` and `adif` identical.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, ladif = *<value>*, `n` = *<value>* and `iadif1` =
            *<value>*.

            Constraint: ladif > `n`*`iadif1`.

        (`errno` 1)
            On entry, `iadif1` = *<value>*.

            Constraint: `iadif1` >= 1.

        (`errno` 1)
            On entry, la = *<value>*, `n` = *<value>* and `ia1` =
            *<value>*.

            Constraint: la > `n`*`ia1`.

        (`errno` 1)
            On entry, `ia1` = *<value>*.

            Constraint: `ia1` >= 1.

        (`errno` 1)
            On entry, `xmax` = *<value>* and `xmin` = *<value>*.

            Constraint: `xmax` > `xmin`.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

    References
    ----------
    Modern Computing Methods, 1961, `Chebyshev-series`, NPL Notes on
    Applied Science (16), (2nd Edition), HMSO
    """

    fun_name = 'naginterfaces.library.fit.dim1_cheb_deriv'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _la_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=1)[0]
    _la = _b_utils._EngineIntScalarType(
        _la_py, {'fun_name': fun_name, 'entity': '_la'},
    ).to_ctypes()
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _iadif1 = _b_utils._EngineIntScalarType(
        iadif1, {'fun_name': fun_name, 'entity': 'iadif1'},
    ).to_ctypes()
    _iadif1_py = _b_utils._EngineIntScalarType.to_py(_iadif1)
    _ladif = _b_utils._EngineIntScalarType(
        1+_n_py*_iadif1_py, {'fun_name': fun_name, 'entity': '_ladif'},
    ).to_ctypes()
    _adif = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (1+_n_py*_iadif1_py,),
    )
    _return_scalar = _b_fit.dim1_cheb_deriv(
        _n,
        xmin,
        xmax,
        _a,
        ia1,
        _la,
        _adif,
        _iadif1,
        _ladif,
    )
    _patm1_py = _return_scalar
    _adif_py = _adif
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['patm1', 'adif'],
    )(
        _patm1_py,
        _adif_py,
    )
    return _return_data

def dim1_cheb_integ(n, xmin, xmax, a, ia1, qatm1, iaint1):
    u"""
    Integral of fitted polynomial in Chebyshev series form.

    ``dim1_cheb_integ`` determines the coefficients in the Chebyshev
    series representation of the indefinite integral of a polynomial
    given in Chebyshev series form.

    For full information please refer to the NAG Library document for
    e02aj

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e02/e02ajf.html

    Parameters
    ----------
    n : int
        n, the degree of the given polynomial p(x).

    xmin : float
        The lower and upper end points respectively of the interval
        [x_min,x_max]. The Chebyshev series representation is in terms
        of the normalized variable x_, where

            x_= (2x-(x_max+x_min))/(x_max-x_min).

    xmax : float
        The lower and upper end points respectively of the interval
        [x_min,x_max]. The Chebyshev series representation is in terms
        of the normalized variable x_, where

            x_= (2x-(x_max+x_min))/(x_max-x_min).

    a : float, array-like, shape (1+`n`*`ia1`)
        The Chebyshev coefficients of the polynomial p(x). Specifically,
        element i*`ia1` of `a` must contain the coefficient a_i, for i =
        0,1,...,n. Only these n+1 elements will be accessed.

        Unchanged on exit, but see `aintc`, below.

    ia1 : int
        The index increment of `a`. Most frequently the Chebyshev
        coefficients are stored in adjacent elements of `a`, and `ia1`
        must be set to 1. However, if for example, they are stored in
        `a`[0],`a`[3],`a`[6],..., the value of `ia1` must be 3. See also
        [Further Comments].

    qatm1 : float
        The value that the integrated polynomial is required to have at
        the lower end point of its interval of definition, i.e., at x_=
        -1 which corresponds to x = x_min. Thus, `qatm1` is a constant
        of integration and will normally be set to zero by you.

    iaint1 : int
        The index increment of `aintc`. Most frequently the Chebyshev
        coefficients are required in adjacent elements of `aintc`, and
        `iaint1` must be set to 1. However, if, for example, they are to
        be stored in `aintc`[0],`aintc`[3],`aintc`[6],..., the value of
        `iaint1` must be 3. See also [Further Comments].

    Returns
    -------
    aintc : float, ndarray, shape (1+(`n`+1)*`iaint1`)
        The Chebyshev coefficients of the integral q(x). (The
        integration is with respect to the variable x, and the constant
        coefficient is chosen so that q(x_min) equals `qatm1`).
        Specifically, element i*`iaint1` of `aintc` contains the
        coefficient a'_i, for i = 0,1,...,n+1. A call of the function
        may have the array name `aintc` the same as `a`, provided that
        note is taken of the order in which elements are overwritten
        when choosing starting elements and increments `ia1` and
        `iaint1`: i.e., the coefficients, a_0,a_1,...,a_{i-2} must be
        intact after coefficient a'_i is stored. In particular it is
        possible to overwrite the a_i entirely by having `ia1` =
        `iaint1`, and the actual array for `a` and `aintc` identical.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, laint = *<value>*, `n` = *<value>* and `iaint1` =
            *<value>*.

            Constraint: laint > (`n`+1)*`iaint1`.

        (`errno` 1)
            On entry, `iaint1` = *<value>*.

            Constraint: `iaint1` >= 1.

        (`errno` 1)
            On entry, la = *<value>*, `n` = *<value>* and `ia1` =
            *<value>*.

            Constraint: la > `n`*`ia1`.

        (`errno` 1)
            On entry, `ia1` = *<value>*.

            Constraint: `ia1` >= 1.

        (`errno` 1)
            On entry, `xmax` = *<value>* and `xmin` = *<value>*.

            Constraint: `xmax` > `xmin`.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

    References
    ----------
    Modern Computing Methods, 1961, `Chebyshev-series`, NPL Notes on
    Applied Science (16), (2nd Edition), HMSO
    """

    fun_name = 'naginterfaces.library.fit.dim1_cheb_integ'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _la_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=1)[0]
    _la = _b_utils._EngineIntScalarType(
        _la_py, {'fun_name': fun_name, 'entity': '_la'},
    ).to_ctypes()
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _iaint1 = _b_utils._EngineIntScalarType(
        iaint1, {'fun_name': fun_name, 'entity': 'iaint1'},
    ).to_ctypes()
    _iaint1_py = _b_utils._EngineIntScalarType.to_py(_iaint1)
    _laint = _b_utils._EngineIntScalarType(
        1+(_n_py+1)*_iaint1_py, {'fun_name': fun_name, 'entity': '_laint'},
    ).to_ctypes()
    _aintc = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (1+(_n_py+1)*_iaint1_py,),
    )
    _b_fit.dim1_cheb_integ(
        _n,
        xmin,
        xmax,
        _a,
        ia1,
        _la,
        qatm1,
        _aintc,
        _iaint1,
        _laint,
    )
    _aintc_py = _aintc
    return _aintc_py

def dim1_cheb_eval2(n, xmin, xmax, a, ia1, x):
    u"""
    Evaluation of fitted polynomial in one variable from Chebyshev
    series form.

    ``dim1_cheb_eval2`` evaluates a polynomial from its Chebyshev series
    representation, allowing an arbitrary index increment for accessing
    the array of coefficients.

    For full information please refer to the NAG Library document for
    e02ak

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e02/e02akf.html

    Parameters
    ----------
    n : int
        n, the degree of the given polynomial p(x_).

    xmin : float
        The lower and upper end points respectively of the interval
        [x_min,x_max]. The Chebyshev series representation is in terms
        of the normalized variable x_, where

            x_= (2x-(x_max+x_min))/(x_max-x_min).

    xmax : float
        The lower and upper end points respectively of the interval
        [x_min,x_max]. The Chebyshev series representation is in terms
        of the normalized variable x_, where

            x_= (2x-(x_max+x_min))/(x_max-x_min).

    a : float, array-like, shape (`n`*`ia1`+1)
        The Chebyshev coefficients of the polynomial p(x_).
        Specifically, element i*`ia1` must contain the coefficient a_i,
        for i = 0,1,...,n. Only these n+1 elements will be accessed.

    ia1 : int
        The index increment of `a`. Most frequently, the Chebyshev
        coefficients are stored in adjacent elements of `a`, and `ia1`
        must be set to 1. However, if, for example, they are stored in
        `a`[0],`a`[3],`a`[6],..., the value of `ia1` must be 3.

    x : float
        The argument x at which the polynomial is to be evaluated.

    Returns
    -------
    result : float
        The value of the polynomial p(x_).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `xmax` = *<value>* and `xmin` = *<value>*.

            Constraint: `xmax` > `xmin`.

        (`errno` 1)
            On entry, la = *<value>*, `n` = *<value>* and `ia1` =
            *<value>*.

            Constraint: la > `n`*`ia1`.

        (`errno` 1)
            On entry, `ia1` = *<value>*.

            Constraint: `ia1` >= 1.

        (`errno` 1)
            On entry, `n`+1 = *<value>*>.

            Constraint: `n` >= 0.

        (`errno` 2)
            On entry, `x` does not lie in [`xmin`,`xmax`]: `x` =
            *<value>*, `xmin` = *<value>* and `xmax` = *<value>*.

    References
    ----------
    Clenshaw, C W, 1955, `A note on the summation of Chebyshev series`,
    Math. Tables Aids Comput. (9), 118--120

    Cox, M G, 1973, `A data-fitting package for the non-specialist
    user`, NPL Report NAC 40, National Physical Laboratory

    Cox, M G and Hayes, J G, 1973, `Curve fitting: a guide and suite of
    algorithms for the non-specialist user`, NPL Report NAC26, National
    Physical Laboratory

    Gentleman, W M, 1969, `An error analysis of Goertzel's (Watt's)
    method for computing Fourier coefficients`, Comput. J. (12),
    160--165
    """

    fun_name = 'naginterfaces.library.fit.dim1_cheb_eval2'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _la_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=1)[0]
    _la = _b_utils._EngineIntScalarType(
        _la_py, {'fun_name': fun_name, 'entity': '_la'},
    ).to_ctypes()
    _return_scalar = _b_fit.dim1_cheb_eval2(
        n,
        xmin,
        xmax,
        _a,
        ia1,
        _la,
        x,
    )
    _result_py = _return_scalar
    return _result_py

def dim1_minimax_polynomial(x, y, m):
    u"""
    Minimax curve fit by polynomials.

    ``dim1_minimax_polynomial`` calculates a minimax polynomial fit to a
    set of data points.

    For full information please refer to the NAG Library document for
    e02al

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e02/e02alf.html

    Parameters
    ----------
    x : float, array-like, shape (n)
        The values of the x coordinates, x_i, for i = 1,2,...,n.

    y : float, array-like, shape (n)
        The values of the y coordinates, y_i, for i = 1,2,...,n.

    m : int
        m, where m is the degree of the polynomial to be found.

    Returns
    -------
    a : float, ndarray, shape (`m`+1)
        The coefficients a_i of the minimax polynomial, for i =
        0,1,...,m.

    ref : float
        The final reference deviation, i.e., the maximum deviation of
        the computed polynomial evaluated at x_i from the reference
        values y_i, for i = 1,2,...,n. `ref` may return a negative value
        which indicates that the algorithm started to cycle due to
        round-off errors.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 2)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 0.

        (`errno` 2)
            On entry, `m` = *<value>*.

            Constraint: `m` < 100.

        (`errno` 2)
            On entry, `m` = *<value>* and n = *<value>*.

            Constraint: `m` < n-1.

        (`errno` 3)
            On entry, i = *<value>*, `x`[i] = *<value>* and `x`[i-1] =
            *<value>*.

            Constraint: `x`[i] > `x`[i-1].

    References
    ----------
    Boothroyd, J B, 1967, `Algorithm 318`, Comm. ACM (10), 801

    Stieffel, E, 1959, `Numerical methods of Tchebycheff approximation`,
    On Numerical Approximation, (ed R E Langer), 217--232, University of
    Wisconsin Press
    """

    fun_name = 'naginterfaces.library.fit.dim1_minimax_polynomial'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m = _b_utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = _b_utils._EngineIntScalarType.to_py(_m)
    _a = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_m_py+1,),
    )
    _return_scalar = _b_fit.dim1_minimax_polynomial(
        _n,
        _x,
        _y,
        _m,
        _a,
    )
    _a_py = _a
    _ref_py = _return_scalar
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'ref'],
    )(
        _a_py,
        _ref_py,
    )
    return _return_data

def dim1_spline_knots(x, y, w, lamda):
    u"""
    Least squares curve cubic spline fit (including interpolation).

    ``dim1_spline_knots`` computes a weighted least squares
    approximation to an arbitrary set of data points by a cubic spline
    with knots prescribed by you.
    Cubic spline interpolation can also be carried out.

    For full information please refer to the NAG Library document for
    e02ba

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e02/e02baf.html

    Parameters
    ----------
    x : float, array-like, shape (m)
        The values x_r of the independent variable (abscissa), for r =
        1,2,...,m. The values must satisfy the Schoenberg--Whitney
        conditions (see [Further Comments]).

    y : float, array-like, shape (m)
        The values y_r of the dependent variable (ordinate), for r =
        1,2,...,m.

    w : float, array-like, shape (m)
        The values w_r of the weights, for r = 1,2,...,m. For advice on
        the choice of weights, see [the E02 Introduction].

    lamda : float, array-like, shape (ncap7)
        `lamda`[i-1] must be set to the (i-4)th (interior) knot,
        lambda_i, for i = 5,6,...,n_+3.

    Returns
    -------
    lamda : float, ndarray, shape (ncap7)
        The input values are unchanged, and `lamda`[i-1], for i =
        1,2,...,4, ncap7-3, ncap7-2, ncap7-1, ncap7 contains the
        additional (exterior) knots introduced by the function. For
        advice on the choice of knots, see [the E02 Introduction].

    c : float, ndarray, shape (ncap7)
        The coefficient c_i of the B-spline N_i(x), for i =
        1,2,...,n_+3. The remaining elements of the array are not used.

    ss : float
        The residual sum of squares, theta.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, J = *<value>*, `lamda`[J-1] = *<value>* and
            `lamda`[J] = *<value>*.

            Constraint: `lamda`[J-1] <= `lamda`[J].

        (`errno` 1)
            On entry, ncap7 = *<value>*, `lamda`[ncap7-5] = *<value>*, m
            = *<value>* and `x`[m-1] = *<value>*.

            Constraint: `lamda`[ncap7-5] < `x`[m-1].

        (`errno` 1)
            On entry, `lamda`[4] = *<value>* and `x`[0] = *<value>*.

            Constraint: `lamda`[4] > `x`[0].

        (`errno` 2)
            On entry, i = *<value>* and `w`[i-1] = *<value>*.

            Constraint: `w`[i-1] > 0.0.

        (`errno` 3)
            On entry, the `x` values are not in nondecreasing order. I =
            *<value>*, `x`[I-1] = *<value>*, J = *<value>* and
            xdist[J-1] = *<value>*.

            Constraint: `x`[I-1] >= xdist[J-1], where xdist is the set
            of distinct x-values.

        (`errno` 4)
            On entry, ncap7 = *<value>* and mdist = *<value>*.

            Constraint: ncap7 <= mdist+4, where mdist is the number of
            distinct x-values.

        (`errno` 4)
            On entry, ncap7 = *<value>* and m = *<value>*.

            Constraint: ncap7 <= m+4.

        (`errno` 4)
            On entry, ncap7 = *<value>*.

            Constraint: ncap7 >= 8.

        (`errno` 5)
            On entry, the Schoenberg--Whitney conditions fail to hold
            for at least one subset of the distinct data abscissae. I =
            *<value>*, xdist[I-1] = *<value>*, J = *<value>* and
            `lamda`[J-1] = *<value>*.

            Constraint: xdist[I-1] < `lamda`[J-1], where xdist is the
            set of distinct x-values.

        (`errno` 5)
            On entry, the Schoenberg--Whitney conditions fail to hold
            for at least one subset of the distinct data abscissae. L =
            *<value>*, xdist[L-1] = *<value>*, I = *<value>* and
            `lamda`[I-1] = *<value>*.

            Constraint: xdist[L-1] > `lamda`[I-1], where xdist is the
            set of distinct x-values.

        (`errno` 5)
            On entry, the Schoenberg--Whitney conditions fail to hold
            for at least one subset of the distinct data abscissae. J =
            *<value>*, xdist[J-1] = *<value>*, J+4 = *<value>* and
            `lamda`[J+3] = *<value>*.

            Constraint: xdist[J-1] < `lamda`[J+3], where xdist is the
            set of distinct x-values.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Cox, M G, 1972, `The numerical evaluation of B-splines`, J. Inst.
    Math. Appl. (10), 134--149

    Cox, M G, 1974, `A data-fitting package for the non-specialist
    user`, Software for Numerical Mathematics, (ed D J Evans), Academic
    Press

    Cox, M G, 1975, `Numerical methods for the interpolation and
    approximation of data by spline functions`, PhD Thesis, City
    University, London

    Cox, M G and Hayes, J G, 1973, `Curve fitting: a guide and suite of
    algorithms for the non-specialist user`, NPL Report NAC26, National
    Physical Laboratory

    de Boor, C, 1972, `On calculating with B-splines`, J. Approx. Theory
    (6), 50--62

    Gentleman, W M, 1973, `Least squares computations by Givens
    transformations without square roots`, J. Inst. Math. Applic. (12),
    329--336

    Gentleman, W M, 1974, `Algorithm AS 75. Basic procedures for large
    sparse or weighted linear least squares problems`, Appl. Statist.
    (23), 448--454

    Schoenberg, I J and Whitney, A, 1953, `On Polya frequency functions
    III`, Trans. Amer. Math. Soc. (74), 246--259
    """

    fun_name = 'naginterfaces.library.fit.dim1_spline_knots'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _lamda = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'lamda'},
        lamda,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ncap7_py = _b_utils._EngineFloat64ArrayType.get_shape(_lamda, {'fun_name': fun_name, 'entity': 'lamda'}, exp_rank=1)[0]
    _ncap7 = _b_utils._EngineIntScalarType(
        _ncap7_py, {'fun_name': fun_name, 'entity': '_ncap7'},
    ).to_ctypes()
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _w = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'w'},
        w,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _c = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ncap7_py,),
    )
    _return_scalar = _b_fit.dim1_spline_knots(
        _m,
        _ncap7,
        _x,
        _y,
        _w,
        _lamda,
        _c,
    )
    _lamda_py = _lamda
    _c_py = _c
    _ss_py = _return_scalar
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['lamda', 'c', 'ss'],
    )(
        _lamda_py,
        _c_py,
        _ss_py,
    )
    return _return_data

def dim1_spline_eval(lamda, c, x):
    u"""
    Evaluation of fitted cubic spline, function only.

    ``dim1_spline_eval`` evaluates a cubic spline from its B-spline
    representation.

    For full information please refer to the NAG Library document for
    e02bb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e02/e02bbf.html

    Parameters
    ----------
    lamda : float, array-like, shape (ncap7)
        `lamda`[j-1] must be set to the value of the jth member of the
        complete set of knots, lambda_j, for j = 1,2,...,n_+7.

    c : float, array-like, shape (ncap7)
        The coefficient c_i of the B-spline N_i(x), for i =
        1,2,...,n_+3. The remaining elements of the array are not
        referenced.

    x : float
        The argument x at which the cubic spline is to be evaluated.

    Returns
    -------
    s : float
        The value of the spline, s(x).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `x` = *<value>*, ncap7 = *<value>* and
            `lamda`[ncap7-4] = *<value>*.

            Constraint: `x` <= `lamda`[ncap7-4].

        (`errno` 1)
            On entry, `x` = *<value>* and `lamda`[3] = *<value>*.

            Constraint: `x` >= `lamda`[3].

        (`errno` 2)
            On entry, ncap7 = *<value>*.

            Constraint: ncap7 >= 8.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Cox, M G, 1972, `The numerical evaluation of B-splines`, J. Inst.
    Math. Appl. (10), 134--149

    Cox, M G, 1978, `The numerical evaluation of a spline from its
    B-spline representation`, J. Inst. Math. Appl. (21), 135--143

    Cox, M G and Hayes, J G, 1973, `Curve fitting: a guide and suite of
    algorithms for the non-specialist user`, NPL Report NAC26, National
    Physical Laboratory

    de Boor, C, 1972, `On calculating with B-splines`, J. Approx. Theory
    (6), 50--62
    """

    fun_name = 'naginterfaces.library.fit.dim1_spline_eval'

    _lamda = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'lamda'},
        lamda,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ncap7_py = _b_utils._EngineFloat64ArrayType.get_shape(_lamda, {'fun_name': fun_name, 'entity': 'lamda'}, exp_rank=1)[0]
    _ncap7 = _b_utils._EngineIntScalarType(
        _ncap7_py, {'fun_name': fun_name, 'entity': '_ncap7'},
    ).to_ctypes()
    _c = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _return_scalar = _b_fit.dim1_spline_eval(
        _ncap7,
        _lamda,
        _c,
        x,
    )
    _s_py = _return_scalar
    return _s_py

def dim1_spline_deriv(lamda, c, x, left):
    u"""
    Evaluation of fitted cubic spline, function and derivatives.

    ``dim1_spline_deriv`` evaluates a cubic spline and its first three
    derivatives from its B-spline representation.

    For full information please refer to the NAG Library document for
    e02bc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e02/e02bcf.html

    Parameters
    ----------
    lamda : float, array-like, shape (ncap7)
        `lamda`[j-1] must be set to the value of the jth member of the
        complete set of knots, lambda_j, for j = 1,2,...,n_+7.

    c : float, array-like, shape (ncap7)
        The coefficient c_i of the B-spline N_i(x), for i =
        1,2,...,n_+3. The remaining elements of the array are not
        referenced.

    x : float
        The argument x at which the cubic spline and its derivatives are
        to be evaluated.

    left : int
        Specifies whether left- or right-hand values of the spline and
        its derivatives are to be computed (see [Description]). Left -
        or right-hand values are formed according to whether `left` is
        equal or not equal to 1.

        If x does not coincide with a knot, the value of `left` is
        immaterial.

        If x = `lamda`[3], right-hand values are computed.

        If x = `lamda`[ncap7-4], left-hand values are formed, regardless
        of the value of `left`.

    Returns
    -------
    s : float, ndarray, shape (4)
        `s`[j-1] contains the value of the (j-1)th derivative of the
        spline at the argument x, for j = 1,2,...,4. Note that `s`[0]
        contains the value of the spline.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, ncap7 = *<value>*.

            Constraint: ncap7 >= 8.

        (`errno` 2)
            On entry, `x` = *<value>*, ncap7 = *<value>* and
            `lamda`[ncap7-4] = *<value>*.

            Constraint: `x` <= `lamda`[ncap7-4].

        (`errno` 2)
            On entry, `x` = *<value>* and `lamda`[3] = *<value>*.

            Constraint: `x` >= `lamda`[3].

        (`errno` 2)
            On entry, `lamda`[3] = *<value>*, ncap7 = *<value>* and
            `lamda`[ncap7-4] = *<value>*.

            Constraint: `lamda`[3] < `lamda`[ncap7-4].

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Cox, M G, 1972, `The numerical evaluation of B-splines`, J. Inst.
    Math. Appl. (10), 134--149

    Cox, M G, 1978, `The numerical evaluation of a spline from its
    B-spline representation`, J. Inst. Math. Appl. (21), 135--143

    de Boor, C, 1972, `On calculating with B-splines`, J. Approx. Theory
    (6), 50--62
    """

    fun_name = 'naginterfaces.library.fit.dim1_spline_deriv'

    _lamda = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'lamda'},
        lamda,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ncap7_py = _b_utils._EngineFloat64ArrayType.get_shape(_lamda, {'fun_name': fun_name, 'entity': 'lamda'}, exp_rank=1)[0]
    _ncap7 = _b_utils._EngineIntScalarType(
        _ncap7_py, {'fun_name': fun_name, 'entity': '_ncap7'},
    ).to_ctypes()
    _c = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _s = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (4,),
    )
    _b_fit.dim1_spline_deriv(
        _ncap7,
        _lamda,
        _c,
        x,
        left,
        _s,
    )
    _s_py = _s
    return _s_py

def dim1_spline_integ(lamda, c):
    u"""
    Evaluation of fitted cubic spline, definite integral.

    ``dim1_spline_integ`` computes the definite integral of a cubic
    spline from its B-spline representation.

    For full information please refer to the NAG Library document for
    e02bd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e02/e02bdf.html

    Parameters
    ----------
    lamda : float, array-like, shape (ncap7)
        `lamda`[j-1] must be set to the value of the jth member of the
        complete set of knots, lambda_j, for j = 1,2,...,n_+7.

    c : float, array-like, shape (ncap7)
        The coefficient c_i of the B-spline N_i(x), for i =
        1,2,...,n_+3. The remaining elements of the array are not
        referenced.

    Returns
    -------
    dint : float
        The value of the definite integral of s(x) between the limits x
        = a and x = b, where a = lambda_4 and b = lambda_{n_+4}.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, ncap7 = *<value>*.

            Constraint: ncap7 >= 8.

        (`errno` 2)
            On entry, J = *<value>*, `lamda`[J-1] = *<value>* and
            `lamda`[J-2] = *<value>*.

            Constraint: `lamda`[J-1] >= `lamda`[J-2].

        (`errno` 2)
            On entry, ncap7 = *<value>*, `lamda`[ncap7-2] = *<value>*
            and `lamda`[ncap7-1] = *<value>*.

            Constraint: `lamda`[ncap7-2] = `lamda`[ncap7-1].

        (`errno` 2)
            On entry, ncap7 = *<value>*, `lamda`[ncap7-3] = *<value>*
            and `lamda`[ncap7-2] = *<value>*.

            Constraint: `lamda`[ncap7-3] = `lamda`[ncap7-2].

        (`errno` 2)
            On entry, ncap7 = *<value>*, `lamda`[ncap7-4] = *<value>*
            and `lamda`[ncap7-4] = *<value>*.

            Constraint: `lamda`[ncap7-4] = `lamda`[ncap7-3].

        (`errno` 2)
            On entry, `lamda`[2] = *<value>* and `lamda`[3] = *<value>*.

            Constraint: `lamda`[2] = `lamda`[3].

        (`errno` 2)
            On entry, `lamda`[1] = *<value>* and `lamda`[2] = *<value>*.

            Constraint: `lamda`[1] = `lamda`[2].

        (`errno` 2)
            On entry, `lamda`[0] = *<value>* and `lamda`[1] = *<value>*.

            Constraint: `lamda`[0] = `lamda`[1].

        (`errno` 2)
            On entry, `lamda`[3] = *<value>*, NCAP4 = *<value>* and
            `lamda`[NCAP4-1] = *<value>*.

            Constraint: `lamda`[3] < `lamda`[NCAP4-1].

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Cox, M G, 1975, `An algorithm for spline interpolation`, J. Inst.
    Math. Appl. (15), 95--108
    """

    fun_name = 'naginterfaces.library.fit.dim1_spline_integ'

    _lamda = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'lamda'},
        lamda,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ncap7_py = _b_utils._EngineFloat64ArrayType.get_shape(_lamda, {'fun_name': fun_name, 'entity': 'lamda'}, exp_rank=1)[0]
    _ncap7 = _b_utils._EngineIntScalarType(
        _ncap7_py, {'fun_name': fun_name, 'entity': '_ncap7'},
    ).to_ctypes()
    _c = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _return_scalar = _b_fit.dim1_spline_integ(
        _ncap7,
        _lamda,
        _c,
    )
    _dint_py = _return_scalar
    return _dint_py

def dim1_spline_auto(start, x, y, w, s, n, lamda, comm):
    u"""
    Least squares cubic spline curve fit, automatic knot placement.

    ``dim1_spline_auto`` computes a cubic spline approximation to an
    arbitrary set of data points.
    The knots of the spline are located automatically, but a single
    argument must be specified to control the trade-off between
    closeness of fit and smoothness of fit.

    For full information please refer to the NAG Library document for
    e02be

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e02/e02bef.html

    Parameters
    ----------
    start : str, length 1
        Must be set to 'C' or 'W'.

        `start` = 'C'

            The function will build up the knot set starting with no
            interior knots. No values need be assigned to the arguments
            `n`, `lamda`, `comm`['wrk'] or `comm`['iwrk'].

        `start` = 'W'

            The function will restart the knot-placing strategy using
            the knots found in a previous call of the function. In this
            case, the arguments `n`, `lamda`, `comm`['wrk'], and
            `comm`['iwrk'] must be unchanged from that previous call.
            This warm start can save much time in searching for a
            satisfactory value of `s`.

    x : float, array-like, shape (m)
        The values x_r of the independent variable (abscissa) x, for r =
        1,2,...,m.

    y : float, array-like, shape (m)
        The values y_r of the dependent variable (ordinate) y, for r =
        1,2,...,m.

    w : float, array-like, shape (m)
        The values w_r of the weights, for r = 1,2,...,m. For advice on
        the choice of weights, see [the E02 Introduction].

    s : float
        The smoothing factor, S.

        If S = 0.0, the function returns an interpolating spline.

        If S is smaller than machine precision, it is assumed equal to
        zero.

        For advice on the choice of S, see [Description] and [Further
        Comments].

    n : int
        If the warm start option is used, the value of `n` must be left
        unchanged from the previous call.

    lamda : float, array-like, shape (nest)
        If the warm start option is used, the values
        `lamda`[0],`lamda`[1],...,`lamda`[ `n` -1] must be left
        unchanged from the previous call.

    comm : dict, communication object, modified in place
        Communication structure.

        `On initial entry`: need not be set.

    Returns
    -------
    n : int
        The total number, n, of knots of the computed spline.

    lamda : float, ndarray, shape (nest)
        The knots of the spline, i.e., the positions of the interior
        knots `lamda`[4], `lamda`[5],...,`lamda`[ `n`-5] as well as the
        positions of the additional knots

            `lamda`[0] = `lamda`[1] = `lamda`[2] = `lamda`[3] = x_1

        and

            `lamda`[ `n`-4] = `lamda`[ `n`-3] = `lamda`[ `n`-2] =
            `lamda`[ `n` -1] = x_m

        needed for the B-spline representation.

    c : float, ndarray, shape (nest)
        The coefficient c_i of the B-spline N_i(x) in the spline
        approximation s(x), for i = 1,2,...,n-4.

    fp : float
        The sum of the squared weighted residuals, theta, of the
        computed spline approximation. If `fp` = 0.0, this is an
        interpolating spline. `fp` should equal `s` within a relative
        tolerance of 0.001 unless n = 8 when the spline has no interior
        knots and so is simply a cubic polynomial. For knots to be
        inserted, `s` must be set to a value below the value of `fp`
        produced in this case.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, r = *<value>* and `w`[r-1] = *<value>*.

            Constraint: `w`[r-1] > 0.0 for all r.

        (`errno` 1)
            On entry, nest = *<value>*.

            Constraint: nest >= 8.

        (`errno` 1)
            On entry, nest = *<value>*.

            Constraint: nest >= *<value>* when `s` = 0.0.

        (`errno` 1)
            On entry, `s` = *<value>*.

            Constraint: `s` >= 0.0.

        (`errno` 1)
            On entry, m = *<value>*.

            Constraint: m >= 4.

        (`errno` 1)
            On entry, `start` != 'W' or 'C': `start` = *<value>*.

        (`errno` 2)
            On entry, r = *<value>* and `w`[r-1] = *<value>*.

            Constraint: `w`[r-1] > 0.0 for all r.

        (`errno` 3)
            On entry, r = *<value>*, `x`[r-2] = *<value>* and `x`[r-1] =
            *<value>*.

            Constraint: `x`[r-2] < `x`[r-1] for all r.

        (`errno` 4)
            The number of knots needed is greater than nest: nest =
            *<value>*. Possibly `s` is too small: `s` = *<value>*.

        (`errno` 5)
            The iterative process has failed to converge. Possibly `s`
            is too small: `s` = *<value>*.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Dierckx, P, 1975, `An algorithm for smoothing, differentiating and
    integration of experimental data using spline functions`, J. Comput.
    Appl. Math. (1), 165--184

    Dierckx, P, 1981, `An improved algorithm for curve fitting with
    spline functions`, Report TW54, Department of Computer Science,
    Katholieke Univerciteit Leuven

    Dierckx, P, 1982, `A fast algorithm for smoothing data on a
    rectangular grid while using spline functions`, SIAM J. Numer. Anal.
    (19), 1286--1304

    Reinsch, C H, 1967, `Smoothing by spline functions`, Numer. Math.
    (10), 177--183
    """

    fun_name = 'naginterfaces.library.fit.dim1_spline_auto'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _w = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'w'},
        w,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lamda = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'lamda'},
        lamda,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nest_py = _b_utils._EngineFloat64ArrayType.get_shape(_lamda, {'fun_name': fun_name, 'entity': 'lamda'}, exp_rank=1)[0]
    _nest = _b_utils._EngineIntScalarType(
        _nest_py, {'fun_name': fun_name, 'entity': '_nest'},
    ).to_ctypes()
    _c = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nest_py,),
    )
    _return_scalars = _b_fit.dim1_spline_auto(
        start,
        _m,
        _x,
        _y,
        _w,
        s,
        _nest,
        n,
        _lamda,
        _c,
        comm,
    )
    _n_py = _return_scalars.n
    _lamda_py = _lamda
    _c_py = _c
    _fp_py = _return_scalars.fp
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['n', 'lamda', 'c', 'fp'],
    )(
        _n_py,
        _lamda_py,
        _c_py,
        _fp_py,
    )
    return _return_data

def dim1_spline_deriv_vector(start, lamda, c, deriv, xord, x, ixloc, iwrk=None):
    u"""
    Evaluation of fitted cubic spline, function and optionally
    derivatives at a vector of points.

    ``dim1_spline_deriv_vector`` evaluates a cubic spline and up to its
    first three derivatives from its B-spline representation at a vector
    of points. ``dim1_spline_deriv_vector`` can be used to compute the
    values and derivatives of cubic spline fits and interpolants
    produced by reference to ``interp.dim1_spline``,
    ``dim1_spline_knots`` and ``dim1_spline_auto``.

    For full information please refer to the NAG Library document for
    e02bf

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e02/e02bff.html

    Parameters
    ----------
    start : int
        Indicates the completion state of the first phase of the
        algorithm.

        `start` = 0

            The enclosing interval numbers ixloc_j for the abscissae x_j
            contained in `x` have not been determined, and you wish to
            use the sorted mode of vectorization.

        `start` = 1

            The enclosing interval numbers ixloc_j have been determined
            and are provided in `ixloc`, however the required
            permutation and interval related information has not been
            determined and you wish to use the sorted mode of
            vectorization.

        `start` = 2

            You wish to use the sorted mode of vectorization, and the
            entire first phase has been completed, with the enclosing
            interval numbers supplied in `ixloc`, and the required
            permutation and interval related information provided in
            `iwrk` (from a previous call to
            ``dim1_spline_deriv_vector``).

        `start` = 10

            The enclosing interval numbers ixloc_j for the abscissae x_j
            contained in `x` have not been determined, and you wish to
            use the unsorted mode of vectorization.

        `start` = 11

            The enclosing interval numbers ixloc_j for the abscissae x_j
            contained in `x` have been supplied in `ixloc`, and you wish
            to use the unsorted mode of vectorization.

        `Additional`: `start` = 0 or 10 should be used unless you are
        sure that the knot set is unchanged between calls.

    lamda : float, array-like, shape (ncap7)
        `lamda`[j-1] must be set to the value of the jth member of the
        complete set of knots, lambda_j, for j = 1,2,...,n_+7.

    c : float, array-like, shape (ncap7)
        The coefficient c_i of the B-spline N_i(x), for i =
        1,2,...,n_+3. The remaining elements of the array are not
        referenced.

    deriv : int
        Determines the maximum order of derivatives required, D_ord.

        If `deriv` < 0 left derivatives are calculated, otherwise right
        derivatives are calculated. For abscissae satisfying x_j =
        lambda_4 or x_j = lambda_{n_+4} only right-handed or left-handed
        computation will be used respectively.

        For abscissae which do not coincide exactly with a knot, the
        handedness of the computation is immaterial.

        `deriv` = 0

            No derivatives required.

        `deriv` = +/-1

            Only s(x) and its first derivative are required.

        `deriv` = +/-2

            Only s(x) and its first and second derivatives are required.

        `deriv` = +/-3

            s(x) and its first, second and third derivatives are
            required.

        Note: if |`deriv`| is greater than 3 only the derivatives up to
        and including 3 will be returned.

    xord : int
        Indicates whether `x` is supplied in a sufficiently ordered
        manner. If `x` is sufficiently ordered
        ``dim1_spline_deriv_vector`` will complete faster.

        `xord` != 0

            The abscissae in `x` are ordered at least by ascending
            interval, in that any two abscissae contained in the same
            interval are only separated by abscissae in the same
            interval, and the intervals are arranged in ascending order.
            For example, x_j < x_{j+1}, for j = 1,2,...,nx-1.

        `xord` = 0

            The abscissae in `x` are not sufficiently ordered.

    x : float, array-like, shape (nx)
        The abscissae x_j, for j = 1,2,...,n_x. If `start` = 0 or 10
        then evaluations will only be performed for these x_j satisfying
        lambda_4 <= x_j <= lambda_{n_+4}. Otherwise evaluation will be
        performed unless the corresponding element of `ixloc` contains
        an invalid interval number. Please note that if the `ixloc`[j]
        is a valid interval number then no check is made that `x`[j]
        actually lies in that interval.

    ixloc : int, array-like, shape (nx)
        If `start` = 1, 2 or 11, if you wish x_j to be evaluated,
        `ixloc`[j-1] must be the enclosing interval number ixloc_j of
        the abscissae x_j (see [equation]). If you do not wish x_j to be
        evaluated, you may set the interval number to be either less
        than 4 or greater than n_+4.

        Otherwise, `ixloc` need not be set.

    iwrk : None or int, array-like, shape (liwrk), optional
        If `start` = 2, `iwrk` must be unchanged from a previous call to
        ``dim1_spline_deriv_vector`` with `start` = 0 or 1.

        Otherwise, `iwrk` need not be set.

    Returns
    -------
    ixloc : int, ndarray, shape (nx)
        If `start` = 1, 2 or 11, `ixloc` is unchanged on exit.

        Otherwise, `ixloc`[j-1], contains the enclosing interval number
        ixloc_j, for the abscissa supplied in `x`[j-1], for j =
        1,2,...,n_x.

        Evaluations will only be performed for abscissae x_j satisfying
        lambda_4 <= x_j <= lambda_{n_+4}.

        If evaluation is not performed `ixloc`[j-1] is set to 0 if x_j <
        lambda_4 or n_+7 if x_j > lambda_{n_+4}.

    s : float, ndarray, shape (nx, min(|`deriv`|, 3)+1)
        If x_j is valid, `s`[j-1,d-1] will contain the (d-1)th
        derivative of s(x), for j = 1,2,...,n_x, for d =
        1,2,...,D_ord+1. In particular, `s`[j-1,0] will contain the
        approximation of s(x_j) for all legal values in `x`.

    iwrk : None or int, ndarray, shape (liwrk)
        If `start` = 10 or 11, `iwrk` is unchanged on exit.

        Otherwise, `iwrk` contains the required permutation of elements
        of `x`, if any, and information related to the division of the
        abscissae x_j between the intervals derived from `lamda`.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, all elements of `x` had enclosing interval numbers
            in `ixloc` outside the domain allowed by the provided
            spline.

            *<value>* entries of `x` were indexed below the lower bound
            *<value>*.

            *<value>* entries of `x` were indexed above the upper bound
            *<value>*.

        (`errno` 11)
            On entry, `start` = *<value>*.

            Constraint: `start` = 0, 1, 2, 10 or 11.

        (`errno` 12)
            On entry, `start` = 2 and nx is not consistent with the
            previous call to ``dim1_spline_deriv_vector``.

            On entry, nx = *<value>*.

            Constraint: nx = *<value>*.

        (`errno` 21)
            On entry, ncap7 = *<value>*.

            Constraint: ncap7 >= 8.

        (`errno` 31)
            On entry, `lamda`[3] = *<value>*, ncap7 = *<value>* and
            `lamda`[ncap7-4] = *<value>*.

            Constraint: `lamda`[3] < `lamda`[ncap7-4].

        (`errno` 91)
            On entry, nx = *<value>*.

            Constraint: nx >= 1.

        (`errno` 131)
            On entry, liwrk = *<value>*.

            Constraint: liwrk >= 3*nx+3 = *<value>*.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, at least one element of `x` has an enclosing
            interval number in `ixloc` outside the set allowed by the
            provided spline. The spline has been evaluated for all `x`
            with enclosing interval numbers inside the allowable set.

            *<value>* entries of `x` were indexed below the lower bound
            *<value>*.

            *<value>* entries of `x` were indexed above the upper bound
            *<value>*.

    References
    ----------
    Cox, M G, 1972, `The numerical evaluation of B-splines`, J. Inst.
    Math. Appl. (10), 134--149

    Cox, M G, 1978, `The numerical evaluation of a spline from its
    B-spline representation`, J. Inst. Math. Appl. (21), 135--143

    de Boor, C, 1972, `On calculating with B-splines`, J. Approx. Theory
    (6), 50--62
    """

    fun_name = 'naginterfaces.library.fit.dim1_spline_deriv_vector'

    _lamda = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'lamda'},
        lamda,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ncap7_py = _b_utils._EngineFloat64ArrayType.get_shape(_lamda, {'fun_name': fun_name, 'entity': 'lamda'}, exp_rank=1)[0]
    _ncap7 = _b_utils._EngineIntScalarType(
        _ncap7_py, {'fun_name': fun_name, 'entity': '_ncap7'},
    ).to_ctypes()
    _c = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nx_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _nx = _b_utils._EngineIntScalarType(
        _nx_py, {'fun_name': fun_name, 'entity': '_nx'},
    ).to_ctypes()
    _ixloc = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ixloc'},
        ixloc,
        copy=True,
    )
    _deriv = _b_utils._EngineIntScalarType(
        deriv, {'fun_name': fun_name, 'entity': 'deriv'},
    ).to_ctypes()
    _deriv_py = _b_utils._EngineIntScalarType.to_py(_deriv)
    _s = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nx_py, min(abs(_deriv_py), 3)+1),
        sorder=_b_utils._EngineData.default_sorder,
    )
    _start = _b_utils._EngineIntScalarType(
        start, {'fun_name': fun_name, 'entity': 'start'},
    ).to_ctypes()
    _start_py = _b_utils._EngineIntScalarType.to_py(_start)
    _iwrk = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'iwrk'},
        iwrk,
        copy=True,
        allow_none=True,
    )
    if _start_py in (10, 11):
        _liwrk = 0
        _liwrk_py = _liwrk
    elif _start_py in (0, 1):
        _liwrk = _b_utils._EngineIntScalarType(
            3+3*_nx_py, {'fun_name': fun_name, 'entity': '_liwrk'},
        ).to_ctypes()
        _liwrk_py = _b_utils._EngineIntScalarType.to_py(_liwrk)
    else:
        _liwrk_py = _b_utils._EngineIntArrayType.get_shape(_iwrk, {'fun_name': fun_name, 'entity': 'iwrk'}, exp_rank=1, allow_none=True)[0]
        _liwrk = _b_utils._EngineIntScalarType(
            _liwrk_py, {'fun_name': fun_name, 'entity': '_liwrk'},
        ).to_ctypes()
    if _start_py == 2:
        if iwrk is None:
            raise _b_utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'iwrk'},
                'must not be None when start = 2',
            )
    if iwrk is None:
        _iwrk = _b_utils._EngineIntArrayType.empty_ndarray(
            (_liwrk_py,),
        )
    try:
        _b_fit.dim1_spline_deriv_vector(
            _start,
            _ncap7,
            _lamda,
            _c,
            _deriv,
            xord,
            _x,
            _ixloc,
            _nx,
            _s,
            _iwrk,
            _liwrk,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _ixloc_py = _ixloc
    if _start_py not in (10, 11):
        _iwrk_py = _iwrk
    else:
        _iwrk_py = None
    _s_py = _s
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['ixloc', 's', 'iwrk'],
    )(
        _ixloc_py,
        _s_py,
        _iwrk_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def dim2_cheb_lines(m, k, l, x, y, f, w, na, xmin, xmax, nux, nuy):
    u"""
    Least squares surface fit by polynomials, data on lines parallel to
    one independent coordinate axis.

    ``dim2_cheb_lines`` forms an approximation to the weighted, least
    squares Chebyshev series surface fit to data arbitrarily distributed
    on lines parallel to one independent coordinate axis.

    For full information please refer to the NAG Library document for
    e02ca

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e02/e02caf.html

    Parameters
    ----------
    m : int, array-like, shape (n)
        `m`[s-1] must be set to m_s, the number of data x values on the
        line y = y_s, for s = 1,2,...,n.

    k : int
        k, the required degree of x in the fit.

    l : int
        l, the required degree of y in the fit.

    x : float, array-like, shape (sum(`m`))
        The x values of the data points. The sequence must be

            all points on y = y_1, followed by

            all points on y = y_2, followed by

            ...

            all points on y = y_n.

    y : float, array-like, shape (n)
        `y`[s-1] must contain the y value of line y = y_s, for s =
        1,2,...,n, on which data is given.

    f : float, array-like, shape (sum(`m`))
        f, the data values of the dependent variable in the same
        sequence as the x values.

    w : float, array-like, shape (sum(`m`))
        The weights to be assigned to the data points, in the same
        sequence as the x values. These weights should be calculated
        from estimates of the absolute accuracies of the f_r, expressed
        as standard deviations, probable errors or some other measure
        which is of the same dimensions as f_r. Specifically, each w_r
        should be inversely proportional to the accuracy estimate of
        f_r. Often weights all equal to unity will be satisfactory. If a
        particular weight is zero, the corresponding data point is
        omitted from the fit.

    na : int
        The dimension of the array `a`.

    xmin : float, array-like, shape (n)
        `xmin`[s-1] must contain x_min^(s), the lower end of the range
        of x on the line y = y_s, for s = 1,2,...,n. It must not be
        greater than the lowest data value of x on the line. Each
        x_min^(s) is scaled to -1.0 in the fit. (See also [Further
        Comments].)

    xmax : float, array-like, shape (n)
        `xmax`[s-1] must contain x_max^(s), the upper end of the range
        of x on the line y = y_s, for s = 1,2,...,n. It must not be less
        than the highest data value of x on the line. Each x_max^(s) is
        scaled to +1.0 in the fit. (See also [Further Comments].)

    nux : float, array-like, shape (inuxp1)
        `nux`[i-1] must contain the coefficient of the Chebyshev
        polynomial of degree (i-1) in x_, in the Chebyshev series
        representation of the polynomial factor in x_ which you require
        the fit to contain, for i = 1,2,...,inuxp1. These coefficients
        are defined according to the standard convention of
        [Description].

    nuy : float, array-like, shape (inuyp1)
        `nuy`[i-1] must contain the coefficient of the Chebyshev
        polynomial of degree (i-1) in y_, in the Chebyshev series
        representation of the polynomial factor which you require the
        fit to contain, for i = 1,2,...,inuyp1. These coefficients are
        defined according to the standard convention of [Description].

    Returns
    -------
    a : float, ndarray, shape (`na`)
        Contains the Chebyshev coefficients of the fit.
        `a`[i*(`l`+1)+j-1] is the coefficient a_{ij} of [Description]
        defined according to the standard convention. These coefficients
        are used by ``dim2_cheb_eval`` to calculate values of the fitted
        function.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `na` is too small. `na` = *<value>*. Minimum
            possible dimension: *<value>*.

        (`errno` 1)
            On entry, mtot is too small. mtot = *<value>*. Minimum
            possible dimension: *<value>*.

        (`errno` 1)
            On entry, I = *<value>*, `m`[I-1] = *<value>*, `k` =
            *<value>* and inuxp1 = *<value>*.

            Constraint: `m`[I-1] >= `k`-inuxp1+2.

        (`errno` 1)
            On entry, n = *<value>*, `l` = *<value>* and inuyp1 =
            *<value>*.

            Constraint: n >= `l`-inuyp1+2.

        (`errno` 1)
            On entry, inuyp1 = *<value>* and `l` = *<value>*.

            Constraint: inuyp1 <= `l`+1.

        (`errno` 1)
            On entry, inuxp1 = *<value>* and `k` = *<value>*.

            Constraint: inuxp1 <= `k`+1.

        (`errno` 1)
            On entry, inuyp1 = *<value>*.

            Constraint: inuyp1 >= 1.

        (`errno` 1)
            On entry, inuxp1 = *<value>*.

            Constraint: inuxp1 >= 1.

        (`errno` 1)
            On entry, `l` = *<value>*.

            Constraint: `l` >= 0.

        (`errno` 1)
            On entry, `k` = *<value>*.

            Constraint: `k` >= 0.

        (`errno` 2)
            On entry, `xmin`[I-1] and `xmax`[I-1] do not span the data
            `x` values on `y` = `y`[I-1]: I = *<value>*, `xmin`[I-1] =
            *<value>*, `xmax`[I-1] = *<value>* and `y`[I-1] = *<value>*.

        (`errno` 3)
            On entry, the data `x` values are not nondecreasing for `y`
            = `y`[I-1]: I = *<value>* and `y`[I-1] = *<value>*.

        (`errno` 3)
            On entry, I = *<value>*, `y`[I-1] = *<value>* and `y`[I-2] =
            *<value>*.

            Constraint: `y`[I-1] > `y`[I-2].

        (`errno` 4)
            On entry, the number of distinct `x` values with nonzero
            weight on `y` = `y`[I-1] is less than `k`-inuxp1+2: I =
            *<value>*, `y`[I-1] = *<value>*, `k` = *<value>* and inuxp1
            = *<value>*.

        (`errno` 5)
            On entry, inuxp1 = *<value>*, `nux`[inuxp1-1] = *<value>*,
            inuyp1 = *<value>* and `nuy`[inuyp1-1] = *<value>*.

            Constraint: if `nux`[inuxp1-1] = 0.0, inuxp1 = 1; if
            `nuy`[inuyp1-1] = 0.0, inuyp1 = 1.

    References
    ----------
    Clenshaw, C W and Hayes, J G, 1965, `Curve and surface fitting`, J.
    Inst. Math. Appl. (1), 164--183

    Hayes, J G (ed.), 1970, `Numerical Approximation to Functions and
    Data`, Athlone Press, London
    """

    fun_name = 'naginterfaces.library.fit.dim2_cheb_lines'

    _m = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'm'},
        m,
        copy=False,
    )
    _n_py = _b_utils._EngineIntArrayType.get_shape(_m, {'fun_name': fun_name, 'entity': 'm'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _mtot_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _mtot = _b_utils._EngineIntScalarType(
        _mtot_py, {'fun_name': fun_name, 'entity': '_mtot'},
    ).to_ctypes()
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _f = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'f'},
        f,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _w = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'w'},
        w,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _k = _b_utils._EngineIntScalarType(
        k, {'fun_name': fun_name, 'entity': 'k'},
    ).to_ctypes()
    _k_py = _b_utils._EngineIntScalarType.to_py(_k)
    _l = _b_utils._EngineIntScalarType(
        l, {'fun_name': fun_name, 'entity': 'l'},
    ).to_ctypes()
    _l_py = _b_utils._EngineIntScalarType.to_py(_l)
    _a = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        ((_k_py+1)*(_l_py+1),),
    )
    _xmin = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xmin'},
        xmin,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _xmax = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xmax'},
        xmax,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nux = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'nux'},
        nux,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _inuxp1_py = _b_utils._EngineFloat64ArrayType.get_shape(_nux, {'fun_name': fun_name, 'entity': 'nux'}, exp_rank=1)[0]
    _inuxp1 = _b_utils._EngineIntScalarType(
        _inuxp1_py, {'fun_name': fun_name, 'entity': '_inuxp1'},
    ).to_ctypes()
    _nuy = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'nuy'},
        nuy,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _inuyp1_py = _b_utils._EngineFloat64ArrayType.get_shape(_nuy, {'fun_name': fun_name, 'entity': 'nuy'}, exp_rank=1)[0]
    _inuyp1 = _b_utils._EngineIntScalarType(
        _inuyp1_py, {'fun_name': fun_name, 'entity': '_inuyp1'},
    ).to_ctypes()
    _b_fit.dim2_cheb_lines(
        _m,
        _n,
        _k,
        _l,
        _x,
        _y,
        _f,
        _w,
        _mtot,
        _a,
        na,
        _xmin,
        _xmax,
        _nux,
        _inuxp1,
        _nuy,
        _inuyp1,
    )
    _a_py = _a
    return _a_py

def dim2_cheb_eval(mfirst, k, l, x, xmin, xmax, y, ymin, ymax, a):
    u"""
    Evaluation of fitted polynomial in two variables.

    ``dim2_cheb_eval`` evaluates a bivariate polynomial from the
    rectangular array of coefficients in its double Chebyshev series
    representation.

    For full information please refer to the NAG Library document for
    e02cb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e02/e02cbf.html

    Parameters
    ----------
    mfirst : int
        The index of the first and last x value in the array x at which
        the evaluation is required respectively (see [Further
        Comments]).

    k : int
        The degree k of x and l of y, respectively, in the polynomial.

    l : int
        The degree k of x and l of y, respectively, in the polynomial.

    x : float, array-like, shape (mlast)
        `x`[i-1], for i = `mfirst`,...,mlast, must contain the x values
        at which the evaluation is required.

    xmin : float
        The lower and upper ends, x_min and x_max, of the range of the
        variable x (see [Description]).

        The values of `xmin` and `xmax` may depend on the value of y
        (e.g., when the polynomial has been derived using
        ``dim2_cheb_lines``).

    xmax : float
        The lower and upper ends, x_min and x_max, of the range of the
        variable x (see [Description]).

        The values of `xmin` and `xmax` may depend on the value of y
        (e.g., when the polynomial has been derived using
        ``dim2_cheb_lines``).

    y : float
        The value of the y coordinate of all the points at which the
        evaluation is required.

    ymin : float
        The lower and upper ends, y_min and y_max, of the range of the
        variable y (see [Description]).

    ymax : float
        The lower and upper ends, y_min and y_max, of the range of the
        variable y (see [Description]).

    a : float, array-like, shape ((`k`+1)*(`l`+1))
        The Chebyshev coefficients of the polynomial. The coefficient
        a_{ij} defined according to the standard convention (see
        [Description]) must be in `a`[i*(l+1)+j].

    Returns
    -------
    ff : float, ndarray, shape (mlast)
        `ff`[i-1] gives the value of the polynomial at the point
        (x_i,y), for i = `mfirst`,...,mlast.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, na = *<value>*, `k` = *<value>* and `l` =
            *<value>*.

            Constraint: na >= (`k`+1)*(`l`+1).

        (`errno` 1)
            On entry, `l` = *<value>*.

            Constraint: `l` >= 0.

        (`errno` 1)
            On entry, `k` = *<value>*.

            Constraint: `k` >= 0.

        (`errno` 1)
            On entry, `mfirst` = *<value>* and mlast = *<value>*.

            Constraint: `mfirst` <= mlast.

        (`errno` 2)
            On entry, `y` = *<value>* and `ymax` = *<value>*.

            Constraint: `y` <= `ymax`.

        (`errno` 2)
            On entry, `y` = *<value>* and `ymin` = *<value>*.

            Constraint: `y` >= `ymin`.

        (`errno` 2)
            On entry, `ymin` = *<value>* and `ymax` = *<value>*.

            Constraint: `ymin` < `ymax`.

        (`errno` 3)
            On entry, I = *<value>*, `x`[I-1] = *<value>* and `xmax` =
            *<value>*.

            Constraint: `x`[I-1] <= `xmax`.

        (`errno` 3)
            On entry, I = *<value>*, `x`[I-1] = *<value>* and `xmin` =
            *<value>*.

            Constraint: `x`[I-1] >= `xmin`.

        (`errno` 3)
            On entry, `xmin` = *<value>* and `xmax` = *<value>*.

            Constraint: `xmin` < `xmax`.

        (`errno` 4)
            Unexpected internal failure when evaluating the polynomial.

    References
    ----------
    Clenshaw, C W, 1955, `A note on the summation of Chebyshev series`,
    Math. Tables Aids Comput. (9), 118--120

    Gentleman, W M, 1969, `An error analysis of Goertzel's (Watt's)
    method for computing Fourier coefficients`, Comput. J. (12),
    160--165
    """

    fun_name = 'naginterfaces.library.fit.dim2_cheb_eval'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _mlast_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _mlast = _b_utils._EngineIntScalarType(
        _mlast_py, {'fun_name': fun_name, 'entity': '_mlast'},
    ).to_ctypes()
    _ff = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_mlast_py,),
    )
    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _na_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=1)[0]
    _na = _b_utils._EngineIntScalarType(
        _na_py, {'fun_name': fun_name, 'entity': '_na'},
    ).to_ctypes()
    _b_fit.dim2_cheb_eval(
        mfirst,
        _mlast,
        k,
        l,
        _x,
        xmin,
        xmax,
        y,
        ymin,
        ymax,
        _ff,
        _a,
        _na,
    )
    _ff_py = _ff
    return _ff_py

def dim2_spline_panel(x, y, f, w, lamda, mu, point, eps):
    u"""
    Least squares surface fit, bicubic splines.

    ``dim2_spline_panel`` forms a minimal, weighted least squares
    bicubic spline surface fit with prescribed knots to a given set of
    data points.

    For full information please refer to the NAG Library document for
    e02da

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e02/e02daf.html

    Parameters
    ----------
    x : float, array-like, shape (m)
        The coordinates of the data point (x_r,y_r,f_r), for r =
        1,2,...,m. The order of the data points is immaterial, but see
        the array `point`.

    y : float, array-like, shape (m)
        The coordinates of the data point (x_r,y_r,f_r), for r =
        1,2,...,m. The order of the data points is immaterial, but see
        the array `point`.

    f : float, array-like, shape (m)
        The coordinates of the data point (x_r,y_r,f_r), for r =
        1,2,...,m. The order of the data points is immaterial, but see
        the array `point`.

    w : float, array-like, shape (m)
        The weight w_r of the rth data point. It is important to note
        the definition of weight implied by the equation (1) in
        [Description], since it is also common usage to define weight as
        the square of this weight. In this function, each w_r should be
        chosen inversely proportional to the (absolute) accuracy of the
        corresponding f_r, as expressed, for example, by the standard
        deviation or probable error of the f_r. When the f_r are all of
        the same accuracy, all the w_r may be set equal to 1.0.

    lamda : float, array-like, shape (px)
        `lamda`[i+3] must contain the ith interior knot lambda_{i+4}
        associated with the variable x, for i = 1,2,...,px-8. The knots
        must be in nondecreasing order and lie strictly within the range
        covered by the data values of x. A knot is a value of x at which
        the spline is allowed to be discontinuous in the third
        derivative with respect to x, though continuous up to the second
        derivative. This degree of continuity can be reduced, if you
        require, by the use of coincident knots, provided that no more
        than four knots are chosen to coincide at any point. Two, or
        three, coincident knots allow loss of continuity in,
        respectively, the second and first derivative with respect to x
        at the value of x at which they coincide. Four coincident knots
        split the spline surface into two independent parts. For choice
        of knots see [Further Comments].

    mu : float, array-like, shape (py)
        `mu`[i+3] must contain the ith interior knot mu_{i+4} associated
        with the variable y, for i = 1,2,...,py-8.

    point : int, array-like, shape (npoint)
        Indexing information usually provided by ``dim2_spline_sort``
        which enables the data points to be accessed in the order which
        produces the advantageous matrix structure mentioned in
        [Description]. This order is such that, if the (x,y) plane is
        thought of as being divided into rectangular panels by the two
        sets of knots, all data in a panel occur before data in
        succeeding panels, where the panels are numbered from bottom to
        top and then left to right with the usual arrangement of axes,
        as indicated in Figure [label omitted].

        [figure omitted]

        A data point lying exactly on one or more panel sides is
        considered to be in the highest numbered panel adjacent to the
        point. ``dim2_spline_sort`` should be called to obtain the array
        `point`, unless it is provided by other means.

    eps : float
        A threshold epsilon for determining the effective rank of the
        system of linear equations. The rank is determined as the number
        of elements of the array `dl` which are nonzero. An element of
        `dl` is regarded as zero if it is less than epsilon. Machine
        precision is a suitable value for epsilon in most practical
        applications which have only 2 or 3 decimals accurate in data.
        If some coefficients of the fit prove to be very large compared
        with the data ordinates, this suggests that epsilon should be
        increased so as to decrease the rank. The array `dl` will give a
        guide to appropriate values of epsilon to achieve this, as well
        as to the choice of epsilon in other cases where some
        experimentation may be needed to determine a value which leads
        to a satisfactory fit.

    Returns
    -------
    lamda : float, ndarray, shape (px)
        The interior knots `lamda`[4] to `lamda`[px-5] are unchanged,
        and the segments `lamda`[0:4] and `lamda`[px-4:px] contain
        additional (exterior) knots introduced by the function in order
        to define the full set of B-splines required. The four knots in
        the first segment are all set equal to the lowest data value of
        x and the other four additional knots are all set equal to the
        highest value: there is experimental evidence that coincident
        end-knots are best for numerical accuracy. The complete array
        must be left undisturbed if ``dim2_spline_evalv`` or
        ``dim2_spline_evalm`` is to be used subsequently.

    mu : float, ndarray, shape (py)
        The same remarks apply to `mu` as to `lamda` above, with `y`
        replacing `x`, and y replacing x.

    dl : float, ndarray, shape ((px-4)*(py-4))
        Gives the squares of the diagonal elements of the reduced
        triangular matrix, divided by the mean squared weight. It
        includes those elements, less than epsilon, which are treated as
        zero (see [Description]).

    c : float, ndarray, shape ((px-4)*(py-4))
        Gives the coefficients of the fit. `c`[ (py-4)* (i-1) +j -1] is
        the coefficient c_{ij} of [Description] and [Further Comments],
        for j = 1,2,...,py-4, for i = 1,2,...,px-4. These coefficients
        are used by ``dim2_spline_evalv`` or ``dim2_spline_evalm`` to
        calculate values of the fitted function.

    sigma : float
        Sigma, the weighted sum of squares of residuals. This is not
        computed from the individual residuals but from the right-hand
        sides of the orthogonally-transformed linear equations. For
        further details see page 97 of Hayes and Halliday (1974). The
        two methods of computation are theoretically equivalent, but the
        results may differ because of rounding error.

    rank : int
        The rank of the system as determined by the value of the
        threshold epsilon.

        `rank` = (px-4)*(py-4)

            The least squares solution is unique.

        `rank` != (px-4)*(py-4)

            The minimal least squares solution is computed.

    Raises
    ------
    NagValueError
        (`errno` 1)
            At least one set of knots is not in nondecreasing order.

        (`errno` 2)
            More than four knots coincide at a single point.

        (`errno` 3)
            Array `point` does not indicate the data points in panel
            order.

        (`errno` 4)
            On entry, npoint = *<value>*, m = *<value>*, px = *<value>*
            and py = *<value>*.

            Constraint: npoint >= m+(px-7)*(py-7).

        (`errno` 4)
            On entry, nc = *<value>*, px = *<value>* and py = *<value>*.

            Constraint: nc = (px-4)*(py-4).

        (`errno` 4)
            On entry, py = *<value>*.

            Constraint: py >= 8.

        (`errno` 4)
            On entry, px = *<value>*.

            Constraint: px >= 8.

        (`errno` 4)
            On entry, m = *<value>*.

            Constraint: m > 1.

        (`errno` 5)
            All the weights are zero, or rank determined as zero.

    References
    ----------
    Cox, M G, 1972, `The numerical evaluation of B-splines`, J. Inst.
    Math. Appl. (10), 134--149

    de Boor, C, 1972, `On calculating with B-splines`, J. Approx. Theory
    (6), 50--62

    Gentleman, W M, 1973, `Least squares computations by Givens
    transformations without square roots`, J. Inst. Math. Applic. (12),
    329--336

    Hayes, J G and Halliday, J, 1974, `The least squares fitting of
    cubic spline surfaces to general data sets`, J. Inst. Math. Appl.
    (14), 89--103
    """

    fun_name = 'naginterfaces.library.fit.dim2_spline_panel'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _lamda = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'lamda'},
        lamda,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _px_py = _b_utils._EngineFloat64ArrayType.get_shape(_lamda, {'fun_name': fun_name, 'entity': 'lamda'}, exp_rank=1)[0]
    _px = _b_utils._EngineIntScalarType(
        _px_py, {'fun_name': fun_name, 'entity': '_px'},
    ).to_ctypes()
    _mu = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'mu'},
        mu,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _py_py = _b_utils._EngineFloat64ArrayType.get_shape(_mu, {'fun_name': fun_name, 'entity': 'mu'}, exp_rank=1)[0]
    _py = _b_utils._EngineIntScalarType(
        _py_py, {'fun_name': fun_name, 'entity': '_py'},
    ).to_ctypes()
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _f = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'f'},
        f,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _w = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'w'},
        w,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _point = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'point'},
        point,
        copy=False,
    )
    _npoint_py = _b_utils._EngineIntArrayType.get_shape(_point, {'fun_name': fun_name, 'entity': 'point'}, exp_rank=1)[0]
    _npoint = _b_utils._EngineIntScalarType(
        _npoint_py, {'fun_name': fun_name, 'entity': '_npoint'},
    ).to_ctypes()
    _nc = _b_utils._EngineIntScalarType(
        (_px_py-4)*(_py_py-4), {'fun_name': fun_name, 'entity': '_nc'},
    ).to_ctypes()
    _dl = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        ((_px_py-4)*(_py_py-4),),
    )
    _c = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        ((_px_py-4)*(_py_py-4),),
    )
    _return_scalars = _b_fit.dim2_spline_panel(
        _m,
        _px,
        _py,
        _x,
        _y,
        _f,
        _w,
        _lamda,
        _mu,
        _point,
        _npoint,
        _dl,
        _c,
        _nc,
        eps,
    )
    _lamda_py = _lamda
    _mu_py = _mu
    _dl_py = _dl
    _c_py = _c
    _sigma_py = _return_scalars.sigma
    _rank_py = _return_scalars.rank
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['lamda', 'mu', 'dl', 'c', 'sigma', 'rank'],
    )(
        _lamda_py,
        _mu_py,
        _dl_py,
        _c_py,
        _sigma_py,
        _rank_py,
    )
    return _return_data

def dim2_spline_grid(start, x, y, f, s, nx, lamda, ny, mu, comm):
    u"""
    Least squares surface fit by bicubic splines with automatic knot
    placement, data on rectangular grid.

    ``dim2_spline_grid`` computes a bicubic spline approximation to a
    set of data values, given on a rectangular grid in the x-y plane.
    The knots of the spline are located automatically, but a single
    argument must be specified to control the trade-off between
    closeness of fit and smoothness of fit.

    For full information please refer to the NAG Library document for
    e02dc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e02/e02dcf.html

    Parameters
    ----------
    start : str, length 1
        Determines whether calculations are to be performed afresh (Cold
        Start) or whether knots found in previous calls are to be used
        as an initial estimate of knot placement (Warm Start).

        `start` = 'C'

            The function will build up the knot set starting with no
            interior knots. No values need be assigned to the arguments
            `nx`, `ny`, `lamda`, `mu`, `comm`['wrk'] or `comm`['iwrk'].

        `start` = 'W'

            The function will restart the knot-placing strategy using
            the knots found in a previous call of the function. In this
            case, the arguments `nx`, `ny`, `lamda`, `mu`, `comm`['wrk']
            and `comm`['iwrk'] must be unchanged from that previous
            call. This warm start can save much time in determining a
            satisfactory set of knots for the given value of `s`. This
            is particularly useful when different smoothing factors are
            used for the same dataset.

    x : float, array-like, shape (mx)
        `x`[q-1] must be set to x_q, the x coordinate of the qth grid
        point along the x axis, for q = 1,2,...,m_x.

    y : float, array-like, shape (my)
        `y`[r-1] must be set to y_r, the y coordinate of the rth grid
        point along the y axis, for r = 1,2,...,m_y.

    f : float, array-like, shape (mx*my)
        `f`[m_y*(q-1)+r-1] must contain the data value f_{q,r}, for r =
        1,2,...,m_y, for q = 1,2,...,m_x.

    s : float
        The smoothing factor, S.

        If `s` = 0.0, the function returns an interpolating spline.

        If `s` is smaller than machine precision, it is assumed equal to
        zero.

        For advice on the choice of `s`, see [Description] and [Further
        Comments].

    nx : int
        If the warm start option is used, the value of `nx` must be left
        unchanged from the previous call.

    lamda : float, array-like, shape (nxest)
        If the warm start option is used, the values
        `lamda`[0],`lamda`[1],...,`lamda`[`nx`-1] must be left unchanged
        from the previous call.

    ny : int
        If the warm start option is used, the value of `ny` must be left
        unchanged from the previous call.

    mu : float, array-like, shape (nyest)
        If the warm start option is used, the values
        `mu`[0],`mu`[1],...,`mu`[`ny`-1] must be left unchanged from the
        previous call.

    comm : dict, communication object, modified in place
        Communication structure.

        `On initial entry`: need not be set.

    Returns
    -------
    nx : int
        The total number of knots, n_x, of the computed spline with
        respect to the x variable.

    lamda : float, ndarray, shape (nxest)
        Contains the complete set of knots lambda_i associated with the
        x variable, i.e., the interior knots
        `lamda`[4],`lamda`[5],...,`lamda`[`nx`-5] as well as the
        additional knots

            `lamda`[0] = `lamda`[1] = `lamda`[2] = `lamda`[3] = `x`[0]

        and

            `lamda`[`nx`-4] = `lamda`[`nx`-3] = `lamda`[`nx`-2] =
            `lamda`[`nx`-1] = `x`[mx-1]

        needed for the B-spline representation.

    ny : int
        The total number of knots, n_y, of the computed spline with
        respect to the y variable.

    mu : float, ndarray, shape (nyest)
        Contains the complete set of knots mu_i associated with the y
        variable, i.e., the interior knots
        `mu`[4],`mu`[5],...,`mu`[`ny`-5] as well as the additional knots

            `mu`[0] = `mu`[1] = `mu`[2] = `mu`[3] = `y`[0]

        and

            `mu`[`ny`-4] = `mu`[`ny`-3] = `mu`[`ny`-2] = `mu`[`ny`-1] =
            `y`[my-1]

        needed for the B-spline representation.

    c : float, ndarray, shape ((nxest-4)*(nyest-4))
        The coefficients of the spline approximation. `c`[(n_y-4) *
        (i-1) +j -1] is the coefficient c_{ij} defined in [Description].

    fp : float
        The sum of squared residuals, theta, of the computed spline
        approximation. If `fp` = 0.0, this is an interpolating spline.
        `fp` should equal `s` within a relative tolerance of 0.001
        unless `nx` = `ny` = 8, when the spline has no interior knots
        and so is simply a bicubic polynomial. For knots to be inserted,
        `s` must be set to a value below the value of `fp` produced in
        this case.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, nxest = *<value>*, mx = *<value>*, nyest =
            *<value>* and my = *<value>*.

            Constraint: if `s` = 0.0, nxest >= mx+4 and nyest >= my+4.

        (`errno` 1)
            On entry, `s` = *<value>*.

            Constraint: `s` >= 0.0.

        (`errno` 1)
            On entry, nyest = *<value>*.

            Constraint: nyest >= *<value>*.

        (`errno` 1)
            On entry, my = *<value>*.

            Constraint: my >= *<value>*.

        (`errno` 1)
            On entry, nxest = *<value>*.

            Constraint: nxest >= *<value>*.

        (`errno` 1)
            On entry, mx = *<value>*.

            Constraint: mx >= *<value>*.

        (`errno` 1)
            On entry, `start` != 'W' or 'C': `start` = *<value>*.

        (`errno` 2)
            On entry, q = *<value>*, `x`[q-2] = *<value>* and `x`[q-1] =
            *<value>*.

            Constraint: `x`[q-2] < `x`[q-1] for all q.

        (`errno` 3)
            On entry, r = *<value>*, `y`[r-2] = *<value>* and `y`[r-1] =
            *<value>*.

            Constraint: `y`[r-2] < `y`[r-1] for all r.

        (`errno` 4)
            The number of knots needed in one direction is greater than
            nxest or nyest: nxest, nyest = *<value>*,*<value>*. Possibly
            `s` is too small: `s` = *<value>*.

        (`errno` 5)
            The iterative process has failed to converge. Possibly `s`
            is too small: `s` = *<value>*.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    de Boor, C, 1972, `On calculating with B-splines`, J. Approx. Theory
    (6), 50--62

    Dierckx, P, 1981, `An improved algorithm for curve fitting with
    spline functions`, Report TW54, Department of Computer Science,
    Katholieke Univerciteit Leuven

    Dierckx, P, 1982, `A fast algorithm for smoothing data on a
    rectangular grid while using spline functions`, SIAM J. Numer. Anal.
    (19), 1286--1304

    Hayes, J G and Halliday, J, 1974, `The least squares fitting of
    cubic spline surfaces to general data sets`, J. Inst. Math. Appl.
    (14), 89--103

    Reinsch, C H, 1967, `Smoothing by spline functions`, Numer. Math.
    (10), 177--183
    """

    fun_name = 'naginterfaces.library.fit.dim2_spline_grid'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _mx_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _mx = _b_utils._EngineIntScalarType(
        _mx_py, {'fun_name': fun_name, 'entity': '_mx'},
    ).to_ctypes()
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _my_py = _b_utils._EngineFloat64ArrayType.get_shape(_y, {'fun_name': fun_name, 'entity': 'y'}, exp_rank=1)[0]
    _my = _b_utils._EngineIntScalarType(
        _my_py, {'fun_name': fun_name, 'entity': '_my'},
    ).to_ctypes()
    _f = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'f'},
        f,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lamda = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'lamda'},
        lamda,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nxest_py = _b_utils._EngineFloat64ArrayType.get_shape(_lamda, {'fun_name': fun_name, 'entity': 'lamda'}, exp_rank=1)[0]
    _nxest = _b_utils._EngineIntScalarType(
        _nxest_py, {'fun_name': fun_name, 'entity': '_nxest'},
    ).to_ctypes()
    _mu = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'mu'},
        mu,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nyest_py = _b_utils._EngineFloat64ArrayType.get_shape(_mu, {'fun_name': fun_name, 'entity': 'mu'}, exp_rank=1)[0]
    _nyest = _b_utils._EngineIntScalarType(
        _nyest_py, {'fun_name': fun_name, 'entity': '_nyest'},
    ).to_ctypes()
    _c = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        ((_nxest_py-4)*(_nyest_py-4),),
    )
    _return_scalars = _b_fit.dim2_spline_grid(
        start,
        _mx,
        _x,
        _my,
        _y,
        _f,
        s,
        _nxest,
        _nyest,
        nx,
        _lamda,
        ny,
        _mu,
        _c,
        comm,
    )
    _nx_py = _return_scalars.nx
    _lamda_py = _lamda
    _ny_py = _return_scalars.ny
    _mu_py = _mu
    _c_py = _c
    _fp_py = _return_scalars.fp
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['nx', 'lamda', 'ny', 'mu', 'c', 'fp'],
    )(
        _nx_py,
        _lamda_py,
        _ny_py,
        _mu_py,
        _c_py,
        _fp_py,
    )
    return _return_data

def dim2_spline_sctr(start, x, y, f, w, s, nx, lamda, ny, mu, comm):
    u"""
    Least squares surface fit by bicubic splines with automatic knot
    placement, scattered data.

    ``dim2_spline_sctr`` computes a bicubic spline approximation to a
    set of scattered data.
    The knots of the spline are located automatically, but a single
    argument must be specified to control the trade-off between
    closeness of fit and smoothness of fit.

    For full information please refer to the NAG Library document for
    e02dd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e02/e02ddf.html

    Parameters
    ----------
    start : str, length 1
        Determines whether calculations are to be performed afresh (Cold
        Start) or whether knots found in previous calls are to be used
        as an initial estimate of knot placement (Warm Start).

        `start` = 'C'

            The function will build up the knot set starting with no
            interior knots. No values need be assigned to the arguments
            `nx`, `ny`, `lamda`, `mu` or `comm`['wrk'].

        `start` = 'W'

            The function will restart the knot-placing strategy using
            the knots found in a previous call of the function. In this
            case, the arguments `nx`, `ny`, `lamda`, `mu` and
            `comm`['wrk'] must be unchanged from that previous call.
            This warm start can save much time in determining a
            satisfactory set of knots for the given value of `s`. This
            is particularly useful when different smoothing factors are
            used for the same dataset.

    x : float, array-like, shape (m)
        `x`[r-1], `y`[r-1], `f`[r-1] must be set to the coordinates of
        (x_r,y_r,f_r), the rth data point, for r = 1,2,...,m. The order
        of the data points is immaterial.

    y : float, array-like, shape (m)
        `x`[r-1], `y`[r-1], `f`[r-1] must be set to the coordinates of
        (x_r,y_r,f_r), the rth data point, for r = 1,2,...,m. The order
        of the data points is immaterial.

    f : float, array-like, shape (m)
        `x`[r-1], `y`[r-1], `f`[r-1] must be set to the coordinates of
        (x_r,y_r,f_r), the rth data point, for r = 1,2,...,m. The order
        of the data points is immaterial.

    w : float, array-like, shape (m)
        `w`[r-1] must be set to w_r, the rth value in the set of
        weights, for r = 1,2,...,m. Zero weights are permitted and the
        corresponding points are ignored, except when determining x_min,
        x_max, y_min and y_max (see [Further Comments]). For advice on
        the choice of weights, see [the E02 Introduction].

    s : float
        The smoothing factor, `s`.

        For advice on the choice of `s`, see [Description] and [Further
        Comments].

    nx : int
        If the warm start option is used, the value of `nx` must be left
        unchanged from the previous call.

    lamda : float, array-like, shape (nxest)
        If the warm start option is used, the values
        `lamda`[0],`lamda`[1],...,`lamda`[`nx`-1] must be left unchanged
        from the previous call.

    ny : int
        If the warm start option is used, the value of `ny` must be left
        unchanged from the previous call.

    mu : float, array-like, shape (nyest)
        If the warm start option is used, the values
        `mu`[0],`mu`[1],...,`mu`[`ny`-1] must be left unchanged from the
        previous call.

    comm : dict, communication object, modified in place
        Communication structure.

        `On initial entry`: need not be set.

    Returns
    -------
    nx : int
        The total number of knots, n_x, of the computed spline with
        respect to the x variable.

    lamda : float, ndarray, shape (nxest)
        Contains the complete set of knots lambda_i associated with the
        x variable, i.e., the interior knots
        `lamda`[4],`lamda`[5],...,`lamda`[`nx`-5] as well as the
        additional knots

            `lamda`[0] = `lamda`[1] = `lamda`[2] = `lamda`[3] = x_min

        and

            `lamda`[`nx`-4] = `lamda`[`nx`-3] = `lamda`[`nx`-2] =
            `lamda`[`nx`-1] = x_max

        needed for the B-spline representation (where x_min and x_max
        are as described in [Description]).

    ny : int
        The total number of knots, n_y, of the computed spline with
        respect to the y variable.

    mu : float, ndarray, shape (nyest)
        Contains the complete set of knots mu_i associated with the y
        variable, i.e., the interior knots
        `mu`[4],`mu`[5],...,`mu`[`ny`-5] as well as the additional knots

            `mu`[0] = `mu`[1] = `mu`[2] = `mu`[3] = y_min

        and

            `mu`[`ny`-4] = `mu`[`ny`-3] = `mu`[`ny`-2] = `mu`[`ny`-1] =
            y_max

        needed for the B-spline representation (where y_min and y_max
        are as described in [Description]).

    c : float, ndarray, shape ((nxest-4)*(nyest-4))
        The coefficients of the spline approximation. `c`[(n_y-4) *
        (i-1) +j -1] is the coefficient c_{ij} defined in [Description].

    fp : float
        The weighted sum of squared residuals, theta, of the computed
        spline approximation. `fp` should equal `s` within a relative
        tolerance of 0.001 unless `nx` = `ny` = 8, when the spline has
        no interior knots and so is simply a bicubic polynomial. For
        knots to be inserted, `s` must be set to a value below the value
        of `fp` produced in this case.

    rank : int
        Gives the rank of the system of equations used to compute the
        final spline (as determined by a suitable machine-dependent
        threshold). When `rank` = (`nx`-4)*(`ny`-4), the solution is
        unique; otherwise the system is rank-deficient and the
        minimum-norm solution is computed. The latter case may be caused
        by too small a value of `s`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, nyest = *<value>*.

            Constraint: nyest >= *<value>*.

        (`errno` 1)
            On entry, nxest = *<value>*.

            Constraint: nxest >= *<value>*.

        (`errno` 1)
            On entry, `start` != 'W' or 'C': `start` = *<value>*.

        (`errno` 1)
            On entry, `s` = *<value>*.

            Constraint: `s` > 0.0.

        (`errno` 1)
            On entry, MM < 16, where MM is the number of points with
            nonzero weight: MM = *<value>*.

        (`errno` 2)
            On entry, all the values in array `y` are equal.

        (`errno` 2)
            On entry, all the values in array `x` are equal.

        (`errno` 3)
            The number of knots needed in one direction is greater than
            nxest or nyest: nxest, nyest = *<value>*,*<value>*. Possibly
            `s` is too small: `s` = *<value>*.

        (`errno` 4)
            No more knots added; the number of B-spline coefficients
            already exceeds m. Either m or `s` is probably too small: m
            = *<value>* and `s` = *<value>*.

        (`errno` 5)
            No more knots added; the additional knot would coincide with
            an old one. Possibly an inaccurate data point has too large
            a weight, or `s` is too small: `s` = *<value>*.

        (`errno` 6)
            The iterative process has failed to converge. Possibly `s`
            is too small: `s` = *<value>*.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    de Boor, C, 1972, `On calculating with B-splines`, J. Approx. Theory
    (6), 50--62

    Dierckx, P, 1981, `An algorithm for surface fitting with spline
    functions`, IMA J. Numer. Anal. (1), 267--283

    Dierckx, P, 1981, `An improved algorithm for curve fitting with
    spline functions`, Report TW54, Department of Computer Science,
    Katholieke Univerciteit Leuven

    Hayes, J G and Halliday, J, 1974, `The least squares fitting of
    cubic spline surfaces to general data sets`, J. Inst. Math. Appl.
    (14), 89--103

    Peters, G and Wilkinson, J H, 1970, `The least squares problem and
    pseudo-inverses`, Comput. J. (13), 309--316

    Reinsch, C H, 1967, `Smoothing by spline functions`, Numer. Math.
    (10), 177--183
    """

    fun_name = 'naginterfaces.library.fit.dim2_spline_sctr'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _f = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'f'},
        f,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _w = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'w'},
        w,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lamda = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'lamda'},
        lamda,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nxest_py = _b_utils._EngineFloat64ArrayType.get_shape(_lamda, {'fun_name': fun_name, 'entity': 'lamda'}, exp_rank=1)[0]
    _nxest = _b_utils._EngineIntScalarType(
        _nxest_py, {'fun_name': fun_name, 'entity': '_nxest'},
    ).to_ctypes()
    _mu = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'mu'},
        mu,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nyest_py = _b_utils._EngineFloat64ArrayType.get_shape(_mu, {'fun_name': fun_name, 'entity': 'mu'}, exp_rank=1)[0]
    _nyest = _b_utils._EngineIntScalarType(
        _nyest_py, {'fun_name': fun_name, 'entity': '_nyest'},
    ).to_ctypes()
    _c = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        ((_nxest_py-4)*(_nyest_py-4),),
    )
    _return_scalars = _b_fit.dim2_spline_sctr(
        start,
        _m,
        _x,
        _y,
        _f,
        _w,
        s,
        _nxest,
        _nyest,
        nx,
        _lamda,
        ny,
        _mu,
        _c,
        comm,
    )
    _nx_py = _return_scalars.nx
    _lamda_py = _lamda
    _ny_py = _return_scalars.ny
    _mu_py = _mu
    _c_py = _c
    _fp_py = _return_scalars.fp
    _rank_py = _return_scalars.rank
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['nx', 'lamda', 'ny', 'mu', 'c', 'fp', 'rank'],
    )(
        _nx_py,
        _lamda_py,
        _ny_py,
        _mu_py,
        _c_py,
        _fp_py,
        _rank_py,
    )
    return _return_data

def dim2_spline_evalv(x, y, lamda, mu, c):
    u"""
    Evaluation of fitted bicubic spline at a vector of points.

    ``dim2_spline_evalv`` calculates values of a bicubic spline from its
    B-spline representation.

    For full information please refer to the NAG Library document for
    e02de

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e02/e02def.html

    Parameters
    ----------
    x : float, array-like, shape (m)
        `x` and `y` must contain x_r and y_r, for r = 1,2,...,m,
        respectively. These are the coordinates of the points at which
        values of the spline are required. The order of the points is
        immaterial.

    y : float, array-like, shape (m)
        `x` and `y` must contain x_r and y_r, for r = 1,2,...,m,
        respectively. These are the coordinates of the points at which
        values of the spline are required. The order of the points is
        immaterial.

    lamda : float, array-like, shape (px)
        `lamda` and `mu` must contain the complete sets of knots
        {lambda} and {mu} associated with the x and y variables
        respectively.

    mu : float, array-like, shape (py)
        `lamda` and `mu` must contain the complete sets of knots
        {lambda} and {mu} associated with the x and y variables
        respectively.

    c : float, array-like, shape ((px-4)*(py-4))
        `c`[ (py-4) * (i-1) +j -1] must contain the coefficient c_{ij}
        described in [Description], for j = 1,2,...,py-4, for i =
        1,2,...,px-4.

    Returns
    -------
    ff : float, ndarray, shape (m)
        `ff`[r-1] contains the value of the spline at the point
        (x_r,y_r), for r = 1,2,...,m.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, py = *<value>*.

            Constraint: py >= 8.

        (`errno` 1)
            On entry, px = *<value>*.

            Constraint: px >= 8.

        (`errno` 1)
            On entry, m = *<value>*.

            Constraint: m >= 1.

        (`errno` 2)
            On entry, the knots in `mu` are not in nondecreasing order.

        (`errno` 2)
            On entry, the knots in `lamda` are not in nondecreasing
            order.

        (`errno` 3)
            On entry, point (`x`[K-1],`y`[K-1]) lies outside the
            rectangle bounded by `lamda`[3], `lamda`[px-4], `mu`[3],
            `mu`[py-4]: K = *<value>*, `x`[K-1] = *<value>* and `y`[K-1]
            = *<value>*.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Anthony, G T, Cox, M G and Hayes, J G, 1982, `DASL -- Data
    Approximation Subroutine Library`, National Physical Laboratory

    Cox, M G, 1978, `The numerical evaluation of a spline from its
    B-spline representation`, J. Inst. Math. Appl. (21), 135--143
    """

    fun_name = 'naginterfaces.library.fit.dim2_spline_evalv'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _lamda = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'lamda'},
        lamda,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _px_py = _b_utils._EngineFloat64ArrayType.get_shape(_lamda, {'fun_name': fun_name, 'entity': 'lamda'}, exp_rank=1)[0]
    _px = _b_utils._EngineIntScalarType(
        _px_py, {'fun_name': fun_name, 'entity': '_px'},
    ).to_ctypes()
    _mu = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'mu'},
        mu,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _py_py = _b_utils._EngineFloat64ArrayType.get_shape(_mu, {'fun_name': fun_name, 'entity': 'mu'}, exp_rank=1)[0]
    _py = _b_utils._EngineIntScalarType(
        _py_py, {'fun_name': fun_name, 'entity': '_py'},
    ).to_ctypes()
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _c = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ff = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_m_py,),
    )
    _b_fit.dim2_spline_evalv(
        _m,
        _px,
        _py,
        _x,
        _y,
        _lamda,
        _mu,
        _c,
        _ff,
    )
    _ff_py = _ff
    return _ff_py

def dim2_spline_evalm(x, y, lamda, mu, c):
    u"""
    Evaluation of fitted bicubic spline at a mesh of points.

    ``dim2_spline_evalm`` calculates values of a bicubic spline from its
    B-spline representation.
    The spline is evaluated at all points on a rectangular grid.

    For full information please refer to the NAG Library document for
    e02df

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e02/e02dff.html

    Parameters
    ----------
    x : float, array-like, shape (mx)
        `x` and `y` must contain x_q, for q = 1,2,...,m_x, and y_r, for
        r = 1,2,...,m_y, respectively. These are the x and y coordinates
        that define the rectangular grid of points at which values of
        the spline are required.

    y : float, array-like, shape (my)
        `x` and `y` must contain x_q, for q = 1,2,...,m_x, and y_r, for
        r = 1,2,...,m_y, respectively. These are the x and y coordinates
        that define the rectangular grid of points at which values of
        the spline are required.

    lamda : float, array-like, shape (px)
        `lamda` and `mu` must contain the complete sets of knots
        {lambda} and {mu} associated with the x and y variables
        respectively.

    mu : float, array-like, shape (py)
        `lamda` and `mu` must contain the complete sets of knots
        {lambda} and {mu} associated with the x and y variables
        respectively.

    c : float, array-like, shape ((px-4)*(py-4))
        `c`[ (py-4) * (i-1) +j -1] must contain the coefficient c_{ij}
        described in [Description], for j = 1,2,...,py-4, for i =
        1,2,...,px-4.

    Returns
    -------
    ff : float, ndarray, shape (mx*my)
        `ff`[my*(q-1)+r-1] contains the value of the spline at the point
        (x_q,y_r), for r = 1,2,...,m_y, for q = 1,2,...,m_x.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, my = *<value>*.

            Constraint: my >= 1.

        (`errno` 1)
            On entry, mx = *<value>*.

            Constraint: mx >= 1.

        (`errno` 1)
            On entry, py = *<value>*.

            Constraint: py >= 8.

        (`errno` 1)
            On entry, px = *<value>*.

            Constraint: px >= 8.

        (`errno` 3)
            On entry, the knots in `mu` are not in nondecreasing order.

        (`errno` 3)
            On entry, the knots in `lamda` are not in nondecreasing
            order.

        (`errno` 4)
            On entry, `mu`[3] <= `y`[0] < ... < `y`[my-1] <= `mu`[py-4]
            is violated.

        (`errno` 4)
            On entry, `lamda`[3] <= `x`[0] < ... < `x`[mx-1] <=
            `lamda`[px-4] is violated.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Anthony, G T, Cox, M G and Hayes, J G, 1982, `DASL -- Data
    Approximation Subroutine Library`, National Physical Laboratory

    Cox, M G, 1978, `The numerical evaluation of a spline from its
    B-spline representation`, J. Inst. Math. Appl. (21), 135--143
    """

    fun_name = 'naginterfaces.library.fit.dim2_spline_evalm'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _mx_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _mx = _b_utils._EngineIntScalarType(
        _mx_py, {'fun_name': fun_name, 'entity': '_mx'},
    ).to_ctypes()
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _my_py = _b_utils._EngineFloat64ArrayType.get_shape(_y, {'fun_name': fun_name, 'entity': 'y'}, exp_rank=1)[0]
    _my = _b_utils._EngineIntScalarType(
        _my_py, {'fun_name': fun_name, 'entity': '_my'},
    ).to_ctypes()
    _lamda = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'lamda'},
        lamda,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _px_py = _b_utils._EngineFloat64ArrayType.get_shape(_lamda, {'fun_name': fun_name, 'entity': 'lamda'}, exp_rank=1)[0]
    _px = _b_utils._EngineIntScalarType(
        _px_py, {'fun_name': fun_name, 'entity': '_px'},
    ).to_ctypes()
    _mu = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'mu'},
        mu,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _py_py = _b_utils._EngineFloat64ArrayType.get_shape(_mu, {'fun_name': fun_name, 'entity': 'mu'}, exp_rank=1)[0]
    _py = _b_utils._EngineIntScalarType(
        _py_py, {'fun_name': fun_name, 'entity': '_py'},
    ).to_ctypes()
    _c = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ff = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_mx_py*_my_py,),
    )
    _b_fit.dim2_spline_evalm(
        _mx,
        _my,
        _px,
        _py,
        _x,
        _y,
        _lamda,
        _mu,
        _c,
        _ff,
    )
    _ff_py = _ff
    return _ff_py

def dim2_spline_derivm(x, y, lamda, mu, c, nux, nuy):
    u"""
    Evaluation of spline surface at mesh of points with derivatives.

    ``dim2_spline_derivm`` computes the partial derivative (of order
    nu_x, nu_y), of a bicubic spline approximation to a set of data
    values, from its B-spline representation, at points on a rectangular
    grid in the x-y plane.
    This function may be used to calculate derivatives of a bicubic
    spline given in the form produced by ``interp.dim2_spline_grid``,
    ``dim2_spline_panel``, ``dim2_spline_grid`` and
    ``dim2_spline_sctr``.

    For full information please refer to the NAG Library document for
    e02dh

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e02/e02dhf.html

    Parameters
    ----------
    x : float, array-like, shape (mx)
        `x`[q-1] must be set to x_q, the x coordinate of the qth grid
        point along the x axis, for q = 1,2,...,m_x, on which values of
        the partial derivative are sought.

    y : float, array-like, shape (my)
        `y`[r-1] must be set to y_r, the y coordinate of the rth grid
        point along the y axis, for r = 1,2,...,m_y on which values of
        the partial derivative are sought.

    lamda : float, array-like, shape (px)
        Contains the position of the knots in the x-direction of the
        bicubic spline approximation to be differentiated, e.g., lamda
        as returned by ``dim2_spline_grid``.

    mu : float, array-like, shape (py)
        Contains the position of the knots in the y-direction of the
        bicubic spline approximation to be differentiated, e.g., mu as
        returned by ``dim2_spline_grid``.

    c : float, array-like, shape ((px-4)*(py-4))
        The coefficients of the bicubic spline approximation to be
        differentiated, e.g., c as returned by ``dim2_spline_grid``.

    nux : int
        Specifies the order, nu_x of the partial derivative in the
        x-direction.

    nuy : int
        Specifies the order, nu_y of the partial derivative in the
        y-direction.

    Returns
    -------
    z : float, ndarray, shape (mx*my)
        `z`[m_y*(q-1)+r-1] contains the derivative
        (d^(nu_x+nu_y))/dx^(nu_x)dy^(nu_y)s(x_q,y_r), for r =
        1,2,...,m_y, for q = 1,2,...,m_x.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `nux` = *<value>*.

            Constraint: 0 <= `nux` <= 2.

        (`errno` 2)
            On entry, `nuy` = *<value>*.

            Constraint: 0 <= `nuy` <= 2.

        (`errno` 3)
            On entry, mx = *<value>*.

            Constraint: mx >= 1.

        (`errno` 4)
            On entry, my = *<value>*.

            Constraint: my >= 1.

        (`errno` 5)
            On entry, for i = *<value>*, `x`[i-2] = *<value>* and
            `x`[i-1] = *<value>*.

            Constraint: `x`[i-2] <= `x`[i-1], for i = 2,3,...,mx.

        (`errno` 6)
            On entry, for i = *<value>*, `y`[i-2] = *<value>* and
            `y`[i-1] = *<value>*.

            Constraint: `y`[i-2] <= `y`[i-1], for i = 2,3,...,my.

    References
    ----------
    de Boor, C, 1972, `On calculating with B-splines`, J. Approx. Theory
    (6), 50--62

    Dierckx, P, 1981, `An improved algorithm for curve fitting with
    spline functions`, Report TW54, Department of Computer Science,
    Katholieke Univerciteit Leuven

    Dierckx, P, 1982, `A fast algorithm for smoothing data on a
    rectangular grid while using spline functions`, SIAM J. Numer. Anal.
    (19), 1286--1304

    Hayes, J G and Halliday, J, 1974, `The least squares fitting of
    cubic spline surfaces to general data sets`, J. Inst. Math. Appl.
    (14), 89--103

    Reinsch, C H, 1967, `Smoothing by spline functions`, Numer. Math.
    (10), 177--183
    """

    fun_name = 'naginterfaces.library.fit.dim2_spline_derivm'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _mx_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _mx = _b_utils._EngineIntScalarType(
        _mx_py, {'fun_name': fun_name, 'entity': '_mx'},
    ).to_ctypes()
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _my_py = _b_utils._EngineFloat64ArrayType.get_shape(_y, {'fun_name': fun_name, 'entity': 'y'}, exp_rank=1)[0]
    _my = _b_utils._EngineIntScalarType(
        _my_py, {'fun_name': fun_name, 'entity': '_my'},
    ).to_ctypes()
    _lamda = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'lamda'},
        lamda,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _px_py = _b_utils._EngineFloat64ArrayType.get_shape(_lamda, {'fun_name': fun_name, 'entity': 'lamda'}, exp_rank=1)[0]
    _px = _b_utils._EngineIntScalarType(
        _px_py, {'fun_name': fun_name, 'entity': '_px'},
    ).to_ctypes()
    _mu = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'mu'},
        mu,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _py_py = _b_utils._EngineFloat64ArrayType.get_shape(_mu, {'fun_name': fun_name, 'entity': 'mu'}, exp_rank=1)[0]
    _py = _b_utils._EngineIntScalarType(
        _py_py, {'fun_name': fun_name, 'entity': '_py'},
    ).to_ctypes()
    _c = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _z = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_mx_py*_my_py,),
    )
    _b_fit.dim2_spline_derivm(
        _mx,
        _my,
        _px,
        _py,
        _x,
        _y,
        _lamda,
        _mu,
        _c,
        nux,
        nuy,
        _z,
    )
    _z_py = _z
    return _z_py

def glin_l1sol(a, b, toler=0.0):
    u"""
    L_1-approximation by general linear function.

    ``glin_l1sol`` calculates an l_1 solution to an over-determined
    system of linear equations.

    For full information please refer to the NAG Library document for
    e02ga

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e02/e02gaf.html

    Parameters
    ----------
    a : float, array-like, shape (m+2, nplus2)
        `a`[i-1,j-1] must contain a_{ij}, the element in the ith row and
        jth column of the matrix A, for j = 1,2,...,n, for i =
        1,2,...,m. The remaining elements need not be set.

    b : float, array-like, shape (m)
        `b`[i-1] must contain b_i, the ith element of the vector b, for
        i = 1,2,...,m.

    toler : float, optional
        A non-negative value. In general `toler` specifies a threshold
        below which numbers are regarded as zero. The recommended
        threshold value is epsilon^(2/3) where epsilon is the machine
        precision. The recommended value can be computed within the
        function by setting `toler` to zero. If premature termination
        occurs a larger value for `toler` may result in a valid
        solution.

    Returns
    -------
    a : float, ndarray, shape (m+2, nplus2)
        Contains the last simplex tableau generated by the simplex
        method.

    b : float, ndarray, shape (m)
        The ith residual r_i corresponding to the solution vector x, for
        i = 1,2,...,m.

    x : float, ndarray, shape (nplus2)
        `x`[j-1] contains the jth element of the solution vector x, for
        j = 1,2,...,n. The elements `x`[n] and `x`[n+1] are unused.

    resid : float
        The sum of the absolute values of the residuals for the solution
        vector x.

    irank : int
        The computed rank of the matrix A.

    itera : int
        The number of iterations taken by the simplex method.

    Raises
    ------
    NagValueError
        (`errno` 2)
            Premature termination due to rounding errors. Try using
            larger value of `toler`: `toler` = *<value>*.

        (`errno` 3)
            On entry, nplus2 = *<value>* and m = *<value>*.

            Constraint: 3 <= nplus2 <= m+2.

        (`errno` 3)
            On entry, nplus2 = *<value>*.

            Constraint: nplus2 >= 3.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            An optimal solution has been obtained, but may not be
            unique.

        (`errno` 4)
            More than 1000*n iterations were performed. ``glin_l1sol``
            has terminated without calculating a solution. The output
            data from the function is as computed on the last good
            iteration. Consider increasing the value of `toler`.
            Alternatively, A may be ill conditioned---try scaling its
            columns.

    References
    ----------
    Barrodale, I and Roberts, F D K, 1973, `An improved algorithm for
    discrete l_1 linear approximation`, SIAM J. Numer. Anal. (10),
    839--848

    Barrodale, I and Roberts, F D K, 1974, `Solution of an
    overdetermined system of equations in the l_1-norm`, Comm. ACM
    (17(6)), 319--320
    """

    fun_name = 'naginterfaces.library.fit.glin_l1sol'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=1)[0]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nplus2_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _nplus2 = _b_utils._EngineIntScalarType(
        _nplus2_py, {'fun_name': fun_name, 'entity': '_nplus2'},
    ).to_ctypes()
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nplus2_py,),
    )
    try:
        _return_scalars = _b_fit.glin_l1sol(
            _m,
            _a,
            _b,
            _nplus2,
            toler,
            _x,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _a_py = _a
    _b_py = _b
    _x_py = _x
    _resid_py = _return_scalars.resid
    _irank_py = _return_scalars.irank
    _itera_py = _return_scalars.itera
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'b', 'x', 'resid', 'irank', 'itera'],
    )(
        _a_py,
        _b_py,
        _x_py,
        _resid_py,
        _irank_py,
        _itera_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def glinc_l1sol(m, e, f, x, mxs, monit, iprint, data=None):
    u"""
    L_1-approximation by general linear function subject to linear
    inequality constraints.

    ``glinc_l1sol`` calculates an l_1 solution to an over-determined
    system of linear equations, possibly subject to linear inequality
    constraints.

    For full information please refer to the NAG Library document for
    e02gb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e02/e02gbf.html

    Parameters
    ----------
    m : int
        The number of equations in the over-determined system, m (i.e.,
        the number of rows of the matrix A).

    e : float, array-like, shape (n, mpl)
        The equation and constraint matrices stored in the following
        manner.

        The first m columns contain the m rows of the matrix A; element
        `e`[i-1,j-1] specifying the element a_{ji} in the jth row and
        ith column of A (the coefficient of the ith unknown in the jth
        equation), for j = 1,2,...,m, for i = 1,2,...,n.

        The next l columns contain the l rows of the constraint matrix
        C; element `e`[i-1,j+m-1] containing the element c_{ji} in the
        jth row and ith column of C (the coefficient of the ith unknown
        in the jth constraint), for j = 1,2,...,l, for i = 1,2,...,n.

    f : float, array-like, shape (mpl)
        `f`[i-1], for i = 1,2,...,m, must contain b_i (the ith element
        of the right-hand side vector of the over-determined system of
        equations) and `f`[m+i-1], for i = 1,2,...,l, must contain d_i
        (the ith element of the right-hand side vector of the
        constraints), where l is the number of constraints.

    x : float, array-like, shape (n)
        `x`[i-1] must contain an estimate of the ith unknown, for i =
        1,2,...,n. If no better initial estimate for `x`[i-1] is
        available, set `x`[i-1] = 0.0.

    mxs : int
        The maximum number of steps to be allowed for the solution of
        the unconstrained problem. Typically this may be a modest
        multiple of n. If, on entry, `mxs` is zero or negative, the
        value returned by ``machine.integer_max`` is used.

    monit : callable monit(x, niter, k, el1n, data=None)
        `monit` can be used to print out the current values of any
        selection of its arguments.

        The frequency with which `monit` is called in ``glinc_l1sol`` is
        controlled by `iprint`.

        Parameters
        ~~~~~~~~~~
        x : float, ndarray, shape (n)
            The latest estimate of the unknowns.

        niter : int
            The number of iterations so far carried out.

        k : int
            The total number of equations and constraints which are
            currently active (i.e., the number of equations with zero
            residuals plus the number of constraints which are satisfied
            as equations).

        el1n : float
            The l_1-norm of the current residuals of the over-determined
            system of equations.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

    iprint : int
        The frequency of iteration print out.

        `iprint` > 0

            `monit` is called every `iprint` iterations and at the
            solution.

        `iprint` = 0

            Information is printed out at the solution only. Otherwise
            `monit` is not called.

    data : arbitrary, optional
        User-communication data for callback functions.

    Returns
    -------
    e : float, ndarray, shape (n, mpl)
        Unchanged, except possibly to the extent of a small multiple of
        the machine precision. (See [Further Comments].)

    x : float, ndarray, shape (n)
        The latest estimate of the ith unknown, for i = 1,2,...,n. If no
        exception or warning is raised on exit, these are the solution
        values.

    k : int
        The total number of equations and constraints which are then
        active (i.e., the number of equations with zero residuals plus
        the number of constraints which are satisfied as equalities).

    el1n : float
        The l_1-norm (sum of absolute values) of the equation residuals.

    indx : int, ndarray, shape (mpl)
        Specifies which columns of `e` relate to the inactive equations
        and constraints. `indx`[0] up to `indx`[`k`-1] number the active
        columns and `indx`[`k`] up to `indx`[mpl-1] number the inactive
        columns.

    Raises
    ------
    NagValueError
        (`errno` 1)
            The constraints cannot all be satisfied simultaneously: they
            are not compatible with one another. Hence no solution is
            possible.

        (`errno` 2)
            The limit imposed by `mxs` has been reached without finding
            a solution. Consider restarting from the current point by
            simply calling ``glinc_l1sol`` again without changing the
            arguments.

        (`errno` 3)
            The function has failed because of numerical difficulties;
            the problem is too ill-conditioned. Consider rescaling the
            unknowns.

        (`errno` 4)
            Elements 1 to `m` of one of the first mpl columns of the
            array `e` are all zero -- this corresponds to a zero row in
            either of the matrices A or C.

        (`errno` 4)
            On entry, mpl = *<value>* and `m` = *<value>*.

            Constraint: mpl >= `m`.

        (`errno` 4)
            On entry, `m` = *<value>* and n = *<value>*.

            Constraint: `m` >= n.

        (`errno` 4)
            On entry, n = *<value>*.

            Constraint: n >= 2.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Bartels, R H, Conn, A R and Charalambous, C, 1976, `Minimisation
    techniques for piecewise Differentiable functions -- the l_infinity
    solution to an overdetermined linear system`, Technical Report No.
    247, CORR 76/30, Mathematical Sciences Department, The John Hopkins
    University

    Bartels, R H, Conn, A R and Sinclair, J W, 1976, `A Fortran program
    for solving overdetermined systems of linear equations in the l_1
    Sense`, Technical Report No. 236, CORR 76/7, Mathematical Sciences
    Department, The John Hopkins University

    Bartels, R H, Conn, A R and Sinclair, J W, 1978, `Minimisation
    techniques for piecewise differentiable functions -- the l_1
    solution to an overdetermined linear system`, SIAM J. Numer. Anal.
    (15), 224--241

    Conn, A R and Pietrzykowski, T, 1977, `A penalty-function method
    converging directly to a constrained optimum`, SIAM J. Numer. Anal.
    (14), 348--375
    """

    def cb_monit(n, x, niter, k, el1n, data=None):
        """Wrapping user's monit. Called as base monit."""
        if data is not None:
            monit(
                x,
                niter,
                k,
                el1n,
                data,
            )
        else:
            monit(
                x,
                niter,
                k,
                el1n,
            )

    fun_name = 'naginterfaces.library.fit.glinc_l1sol'

    _e = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'e'},
        e,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_e, {'fun_name': fun_name, 'entity': 'e'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _mpl_py = _b_utils._EngineFloat64ArrayType.get_shape(_e, {'fun_name': fun_name, 'entity': 'e'}, exp_rank=2)[1]
    _mpl = _b_utils._EngineIntScalarType(
        _mpl_py, {'fun_name': fun_name, 'entity': '_mpl'},
    ).to_ctypes()
    _f = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'f'},
        f,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b_utils._check_callable(monit, {'fun_name': fun_name, 'entity': 'monit'})
    _monit = cb_monit
    _indx = _b_utils._EngineIntArrayType.empty_ndarray(
        (_mpl_py,),
    )
    _return_scalars = _b_fit.glinc_l1sol(
        m,
        _n,
        _mpl,
        _e,
        _f,
        _x,
        mxs,
        _monit,
        iprint,
        _indx,
        data,
    )
    _e_py = _e
    _x_py = _x
    _k_py = _return_scalars.k
    _el1n_py = _return_scalars.el1n
    _indx_py = _indx
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['e', 'x', 'k', 'el1n', 'indx'],
    )(
        _e_py,
        _x_py,
        _k_py,
        _el1n_py,
        _indx_py,
    )
    return _return_data

def glin_linf(n, a, b, relerr, tol=0.0):
    u"""
    L_infinity-approximation by general linear function.

    ``glin_linf`` calculates an l_infinity solution to an
    over-determined system of linear equations.

    For full information please refer to the NAG Library document for
    e02gc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e02/e02gcf.html

    Parameters
    ----------
    n : int
        The number of unknowns, n (the number of columns of the matrix
        A).

    a : float, array-like, shape (`n`+3, m+1)
        `a`[j-1,i-1] must contain a_{ij}, the element in the ith row and
        jth column of the matrix A, for j = 1,2,...,n, for i =
        1,2,...,m, (that is, the **transpose** of the matrix). The
        remaining elements need not be set. Preferably, the columns of
        the matrix A (rows of the argument `a`) should be scaled before
        entry: see [Accuracy].

    b : float, array-like, shape (m)
        `b`[i-1] must contain b_i, the ith element of the vector b, for
        i = 1,2,...,m.

    relerr : float
        Must be set to a bound on the relative error acceptable in the
        maximum residual at the solution.

        If `relerr` <= 0.0, the l_infinity solution is computed, and
        `relerr` is set to 0.0 on exit.

        If `relerr` > 0.0, the function obtains instead an approximate
        solution for which the largest residual is less than
        1.0+`relerr` times that of the l_infinity solution; on exit,
        `relerr` contains a smaller value such that the above bound
        still applies. (The usual result of this option, say with
        `relerr` = 0.1, is a saving in the number of simplex
        iterations).

    tol : float, optional
        A threshold below which numbers are regarded as zero. The
        recommended threshold value is 10.0*epsilon, where epsilon is
        the machine precision. If `tol` <= 0.0 on entry, the recommended
        value is used within the function. If premature termination
        occurs, a larger value for `tol` may result in a valid solution.

    Returns
    -------
    a : float, ndarray, shape (`n`+3, m+1)
        Contains the last simplex tableau.

    b : float, ndarray, shape (m)
        The ith residual r_i corresponding to the solution vector x, for
        i = 1,2,...,m. Note however that these residuals may contain few
        significant figures, especially when `resmax` is within one or
        two orders of magnitude of `tol`. Indeed if `resmax` <= `tol`,
        the elements `b`[i-1] may all be set to zero. It is therefore
        often advisable to compute the residuals directly.

    relerr : float
        Is altered as described above.

    x : float, ndarray, shape (`n`)
        If the function exits successfully or `errno` = 1, `x`[j-1]
        contains the jth element of the solution vector x, for j =
        1,2,...,n. Whether this is an l_infinity solution or an
        approximation to one, depends on the value of `relerr` on entry.

    resmax : float
        If the function exits successfully or `errno` = 1, `resmax`
        contains the absolute value of the largest residual(s) for the
        solution vector x. (See `b`.)

    irank : int
        If the function exits successfully or `errno` = 1, `irank`
        contains the computed rank of the matrix A.

    itera : int
        If the function exits successfully or `errno` = 1, `itera`
        contains the number of iterations taken by the simplex method.

    Raises
    ------
    NagValueError
        (`errno` 2)
            Premature termination due to rounding errors. Try using
            larger value of `tol`: `tol` = *<value>*.

        (`errno` 3)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 3)
            On entry, m = *<value>* and `n` = *<value>*.

            Constraint: m >= `n`.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            An optimal solution has been obtained, but may not be
            unique.

    References
    ----------
    Barrodale, I and Phillips, C, 1974, `An improved algorithm for
    discrete Chebyshev linear approximation`, Proc. 4th Manitoba Conf.
    Numerical Mathematics, 177--190, University of Manitoba, Canada

    Barrodale, I and Phillips, C, 1975, `Solution of an overdetermined
    system of linear equations in the Chebyshev norm [F4] (Algorithm
    495)`, ACM Trans. Math. Software (1(3)), 264--270
    """

    fun_name = 'naginterfaces.library.fit.glin_linf'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=1)[0]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _return_scalars = _b_fit.glin_linf(
            _m,
            _n,
            _a,
            _b,
            tol,
            relerr,
            _x,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _a_py = _a
    _b_py = _b
    _relerr_py = _return_scalars.relerr
    _x_py = _x
    _resmax_py = _return_scalars.resmax
    _irank_py = _return_scalars.irank
    _itera_py = _return_scalars.itera
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'b', 'relerr', 'x', 'resmax', 'irank', 'itera'],
    )(
        _a_py,
        _b_py,
        _relerr_py,
        _x_py,
        _resmax_py,
        _irank_py,
        _itera_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def dim2_spline_ts_sctr(x, y, f, lsminp, lsmaxp, nxcels, nycels, comm):
    u"""
    Spline approximation to a set of scattered data using a two-stage
    approximation method.

    Note: this function uses optional algorithmic parameters, see also:
    ``opt_set``, ``opt_get``.

    ``dim2_spline_ts_sctr`` computes a spline approximation to a set of
    scattered data using a two-stage approximation method.

    The computational complexity of the method grows linearly with the
    number of data points; hence large datasets are easily accommodated.

    For full information please refer to the NAG Library document for
    e02jd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e02/e02jdf.html

    Parameters
    ----------
    x : float, array-like, shape (n)
        The (x_i,y_i,f_i) data values to be fitted.

    y : float, array-like, shape (n)
        The (x_i,y_i,f_i) data values to be fitted.

    f : float, array-like, shape (n)
        The (x_i,y_i,f_i) data values to be fitted.

    lsminp : int
        `lsminp` and `lsmaxp` are control parameters for the local
        approximations.

        Each local approximation is computed on a local domain
        containing one of the triangles in the discretization of the
        bounding box.

        The size of each local domain will be adaptively chosen such
        that if it contains fewer than `lsminp` sample points it is
        expanded, else if it contains greater than `lsmaxp` sample
        points a thinning method is applied. `lsmaxp` mainly controls
        computational cost (in that working with a thinned set of points
        is cheaper and may be appropriate if the input data is densely
        distributed), while `lsminp` allows handling of different types
        of scattered data.

        Setting `lsmaxp` < `lsminp`, and therefore forcing either
        expansion or thinning, may be useful for computing initial
        coarse approximations.

        In general smaller values for these arguments reduces cost.

        A calibration procedure (experimenting with a small subset of
        the data to be fitted and validating the results) may be needed
        to choose the most appropriate values for `lsminp` and `lsmaxp`.

    lsmaxp : int
        `lsminp` and `lsmaxp` are control parameters for the local
        approximations.

        Each local approximation is computed on a local domain
        containing one of the triangles in the discretization of the
        bounding box.

        The size of each local domain will be adaptively chosen such
        that if it contains fewer than `lsminp` sample points it is
        expanded, else if it contains greater than `lsmaxp` sample
        points a thinning method is applied. `lsmaxp` mainly controls
        computational cost (in that working with a thinned set of points
        is cheaper and may be appropriate if the input data is densely
        distributed), while `lsminp` allows handling of different types
        of scattered data.

        Setting `lsmaxp` < `lsminp`, and therefore forcing either
        expansion or thinning, may be useful for computing initial
        coarse approximations.

        In general smaller values for these arguments reduces cost.

        A calibration procedure (experimenting with a small subset of
        the data to be fitted and validating the results) may be needed
        to choose the most appropriate values for `lsminp` and `lsmaxp`.

    nxcels : int
        `nxcels` (respectively `nycels`) is the number of cells in the x
        (respectively y) direction that will be used to create the
        triangulation of the bounding box of the domain of the function
        to be fitted.

        Greater efficiency generally comes when `nxcels` and `nycels`
        are chosen to be of the same order of magnitude and are such
        that n is O(`nxcels`*`nycels`).

        Thus for a 'square' triangulation --- when `nxcels` = `nycels`
        --- the quantities sqrt(n) and `nxcels` should be of the same
        order of magnitude.

        See also [Further Comments].

    nycels : int
        `nxcels` (respectively `nycels`) is the number of cells in the x
        (respectively y) direction that will be used to create the
        triangulation of the bounding box of the domain of the function
        to be fitted.

        Greater efficiency generally comes when `nxcels` and `nycels`
        are chosen to be of the same order of magnitude and are such
        that n is O(`nxcels`*`nycels`).

        Thus for a 'square' triangulation --- when `nxcels` = `nycels`
        --- the quantities sqrt(n) and `nxcels` should be of the same
        order of magnitude.

        See also [Further Comments].

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``opt_set``.

    Other Parameters
    ----------------
    'Averaged Spline' : str
        Default = 'NO'

        When the bounding box is triangulated there are 8 equivalent
        configurations of the mesh.
        Setting 'Averaged Spline' = 'YES' will use the averaged value of
        the 8 possible local polynomial approximations over each
        triangle in the mesh.
        This usually gives better results but at (about 8 times) higher
        computational cost.

        Constraint: 'Averaged Spline' = 'YES' or 'NO'.

    'Global Smoothing Level' : int
        Default = 1

        The smoothness level for the global spline approximation.

        'Global Smoothing Level' = 1
            Will use C^1 piecewise cubics.

        'Global Smoothing Level' = 2
            Will use C^2 piecewise sextics.

    'Interpolation Only RBF' : str
        Default = 'YES'

        If 'Interpolation Only RBF' = 'YES', each local RBF
        approximation is computed by interpolation.

        If 'Interpolation Only RBF' = 'NO', each local RBF approximation
        is computed by a discrete least squares approach.
        This is likely to be more accurate and more expensive than
        interpolation.

        If 'Local Method' = 'HYBRID' or 'POLYNOMIAL', this option
        setting is ignored.

    'Local Method' : str
        Default = 'POLYNOMIAL'

        The local approximation scheme to use.

        'Local Method' = 'POLYNOMIAL'
            Uses least squares polynomial approximations.

        'Local Method' = 'HYBRID'
            Uses hybrid polynomial and RBF approximations.

        'Local Method' = 'RBF'
            Uses pure RBF approximations.

        In general 'POLYNOMIAL' is less computationally expensive than
        'HYBRID' is less computationally expensive than 'RBF' with the
        reverse ordering holding for accuracy of results.

    'Minimum Singular Value LHA' : float
        Default = 1.0

        A tolerance measure for accepting or rejecting a local hybrid
        approximation (LHA) as reliable.

        The solution of a local least squares problem solved on each
        triangle subdomain is accepted as reliable if the minimum
        singular value sigma of the collocation matrix (of polynomial
        and radial basis function terms) associated with the least
        squares problem satisfies 'Minimum Singular Value LHA' <= sigma.

        In general the approximation power will be reduced as 'Minimum
        Singular Value LHA' is reduced. (A small sigma indicates that
        the local data has hidden redundancies which prevent it from
        carrying enough information for a good approximation to be
        made.) Setting 'Minimum Singular Value LHA' very large may have
        the detrimental effect that only approximations of low degree
        are deemed reliable.

        A calibration procedure (experimenting with a small subset of
        the data to be fitted and validating the results) may be needed
        to choose the most appropriate value for this parameter.

        If 'Local Method' = 'POLYNOMIAL' or 'RBF', this option setting
        is ignored.

    'Minimum Singular Value LPA' : float
        Default = 1.0

        A tolerance measure for accepting or rejecting a local
        polynomial approximation (LPA) as reliable.
        Clearly this setting is relevant when 'Local Method' =
        'POLYNOMIAL', but it also may be used when 'Local Method' =
        'HYBRID' (see [Further Comments].)

        The solution of a local least squares problem solved on each
        triangle subdomain is accepted as reliable if the minimum
        singular value sigma of the matrix (of Bernstein polynomial
        values) associated with the least squares problem satisfies
        'Minimum Singular Value LPA' <= sigma.

        In general the approximation power will be reduced as 'Minimum
        Singular Value LPA' is reduced. (A small sigma indicates that
        the local data has hidden redundancies which prevent it from
        carrying enough information for a good approximation to be
        made.) Setting 'Minimum Singular Value LPA' very large may have
        the detrimental effect that only approximations of low degree
        are deemed reliable.

        'Minimum Singular Value LPA' will have no effect if 'Polynomial
        Starting Degree' = 0, and it will have little effect if the
        input data is 'smooth' (e.g., from a known function).

        A calibration procedure (experimenting with a small subset of
        the data to be fitted and validating the results) may be needed
        to choose the most appropriate value for this parameter.

        If 'Local Method' = 'RBF', this option setting is ignored.

    'Polynomial Starting Degree' : int
        Default = 5 if 'Local Method' = 'HYBRID',
        Default = 1 otherwise

        The degree to be used for the polynomial part in the initial
        step of each local approximation.

        At this initial step the method will attempt to fit with a local
        approximation having polynomial part of degree 'Polynomial
        Starting Degree'.
        If 'Local Method' = 'POLYNOMIAL' and the approximation is deemed
        unreliable (according to 'Minimum Singular Value LPA'), the
        degree will be decremented by one and a new local approximation
        computed, ending with a constant approximation if no other is
        reliable.
        If 'Local Method' = 'HYBRID' and the approximation is deemed
        unreliable (according to 'Minimum Singular Value LHA'), a pure
        polynomial approximation of this degree will be tried instead.
        The method then proceeds as in the 'POLYNOMIAL' case.

        'Polynomial Starting Degree' is bounded from above by the
        maximum possible spline degree, which is 6 (when performing C^2
        global super-smoothing).
        Note that the best-case approximation error (see [Accuracy]) for
        C^2 smoothing with 'Supersmooth C2' = 'NO' is achieved for local
        polynomials of degree 5; that is, for this level of global
        smoothing no further benefit is gained by setting 'Polynomial
        Starting Degree' = 6.

        The default value gives a good compromise between efficiency and
        accuracy.
        In general the best approximation can be obtained by setting:

            If 'Local Method' = 'POLYNOMIAL'

                if 'Global Smoothing Level' = 1, 'Polynomial Starting
                Degree' = 3;

                if 'Global Smoothing Level' = 2;

                    if 'Supersmooth C2' = 'NO', 'Polynomial Starting
                    Degree' = 5;

                    otherwise 'Polynomial Starting Degree' = 6.

            If 'Local Method' = 'HYBRID', 'Polynomial Starting Degree'
            as small as possible.

        If 'Local Method' = 'RBF', this option setting is ignored.

    'Radial Basis Function' : str
        Default = 'MQ'

        'Radial Basis Function' selects the RBF to use in each local RBF
        approximation, while 'Scaling Coefficient RBF' selects the scale
        factor to use in its evaluation, as described below.

        A calibration procedure (experimenting with a small subset of
        the data to be fitted and validating the results) may be needed
        to choose the most appropriate scale factor and RBF.

        If 'Local Method' = 'POLYNOMIAL', these option settings are
        ignored.

        If 'Local Method' = 'HYBRID' or 'RBF', the following
        (conditionally) positive definite functions may be chosen.

        Define R = sqrt(x^2+y^2) and rho = R/r.

        [table omitted]

        If 'Local Method' = 'HYBRID' the following conditionally
        positive definite functions may also be chosen.

        [table omitted]

    'Scaling Coefficient RBF' : float
        Default = 1.0

        'Radial Basis Function' selects the RBF to use in each local RBF
        approximation, while 'Scaling Coefficient RBF' selects the scale
        factor to use in its evaluation, as described below.

        A calibration procedure (experimenting with a small subset of
        the data to be fitted and validating the results) may be needed
        to choose the most appropriate scale factor and RBF.

        If 'Local Method' = 'POLYNOMIAL', these option settings are
        ignored.

        If 'Local Method' = 'HYBRID' or 'RBF', the following
        (conditionally) positive definite functions may be chosen.

        Define R = sqrt(x^2+y^2) and rho = R/r.

        [table omitted]

        If 'Local Method' = 'HYBRID' the following conditionally
        positive definite functions may also be chosen.

        [table omitted]

    'Separation LRBFA' : float
        Default = 16.0/'Scaling Coefficient RBF'

        A knot-separation parameter used to control the condition number
        of the matrix used in each local RBF approximation (LRBFA).
        A smaller value may mean greater numerical stability but fewer
        knots.

        If 'Local Method' = 'HYBRID' or 'POLYNOMIAL', this option
        setting is ignored.

    'Supersmooth C2' : str
        Default = 'NO'

        If 'Supersmooth C2' = 'YES', the C^2 spline is generated using
        additional smoothness constraints.
        This usually gives better results but at higher computational
        cost.

        If 'Global Smoothing Level' = 1 this option setting is ignored.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, n = *<value>*.

            Constraint: n > 1.

        (`errno` 4)
            On entry, `lsminp` = *<value>* and n = *<value>*.

            Constraint: 1 <= `lsminp` <= n.

        (`errno` 5)
            On entry, `lsmaxp` = *<value>*.

            Constraint: `lsmaxp` >= 1.

        (`errno` 6)
            On entry, `nxcels` = *<value>*.

            Constraint: `nxcels` >= 1.

        (`errno` 7)
            On entry, `nycels` = *<value>*.

            Constraint: `nycels` >= 1.

        (`errno` 9)
            Option arrays are not initialized or are corrupted.

        (`errno` 11)
            An unexpected algorithmic failure was encountered. Please
            contact NAG.

        (`errno` 12)
            On entry, all elements of `x` or of `y` are equal.

        (`errno` 20)
            The selected radial basis function cannot be used with the
            RBF local method.

        (`errno` 21)
            The value of option 'Polynomial Starting Degree' was
            invalid.

    References
    ----------
    Davydov, O, Morandi, R and Sestini, A, 2006, `Local hybrid
    approximation for scattered data fitting with bivariate splines`,
    Comput. Aided Geom. Design (23), 703--721

    Davydov, O, Sestini, A and Morandi, R, 2005, `Local RBF
    approximation for scattered data fitting with bivariate splines`,
    `Trends and Applications in Constructive Approximation`, M. G. de
    Bruin, D. H. Mache, and J. Szabados, Eds (ISNM Vol. 151),
    Birkhauser, 91--102

    Davydov, O and Zeilfelder, F, 2004, `Scattered data fitting by
    direct extension of local polynomials to bivariate splines`,
    Advances in Comp. Math. (21), 223--271

    See Also
    --------
    :meth:`naginterfaces.library.examples.fit.dim2_spline_ts_sctr_ex.main`
    """

    fun_name = 'naginterfaces.library.fit.dim2_spline_ts_sctr'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _f = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'f'},
        f,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b_fit.dim2_spline_ts_sctr(
        _n,
        _x,
        _y,
        _f,
        lsminp,
        lsmaxp,
        nxcels,
        nycels,
        comm,
    )

def dim2_spline_ts_evalv(xevalv, yevalv, comm):
    u"""
    Evaluation at a vector of points of a spline computed by
    ``dim2_spline_ts_sctr``.

    ``dim2_spline_ts_evalv`` calculates a vector of values of a spline
    computed by ``dim2_spline_ts_sctr``.

    For full information please refer to the NAG Library document for
    e02je

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e02/e02jef.html

    Parameters
    ----------
    xevalv : float, array-like, shape (nevalv)
        The (x_i) values at which the spline is to be evaluated.

    yevalv : float, array-like, shape (nevalv)
        The (y_i) values at which the spline is to be evaluated.

    comm : dict, communication object
        Communication structure.

        This argument must have been initialized by prior calls to
        ``dim2_spline_ts_sctr`` and ``opt_set``.

    Returns
    -------
    fevalv : float, ndarray, shape (nevalv)
        If no exception or warning is raised on exit `fevalv`[i-1]
        contains the computed spline value at (x_i,y_i).

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, nevalv = *<value>*.

            Constraint: nevalv >= 1.

        (`errno` 9)
            Option arrays are not initialized or are corrupted.

        (`errno` 10)
            The fitting routine has not been called, or the array of
            coefficients has been corrupted.

        (`errno` 13)
            On entry, `xevalv`[*<value>*] = *<value>* was outside the
            bounding box.

            Constraint: *<value>* <= `xevalv`[i-1] <= *<value>* for all
            i.

        (`errno` 14)
            On entry, `yevalv`[*<value>*] = *<value>* was outside the
            bounding box.

            Constraint: *<value>* <= `yevalv`[i-1] <= *<value>* for all
            i.

    References
    ----------
    Davydov, O, Morandi, R and Sestini, A, 2006, `Local hybrid
    approximation for scattered data fitting with bivariate splines`,
    Comput. Aided Geom. Design (23), 703--721

    Davydov, O, Sestini, A and Morandi, R, 2005, `Local RBF
    approximation for scattered data fitting with bivariate splines`,
    `Trends and Applications in Constructive Approximation`, M. G. de
    Bruin, D. H. Mache, and J. Szabados, Eds (ISNM Vol. 151),
    Birkhauser, 91--102

    Davydov, O and Zeilfelder, F, 2004, `Scattered data fitting by
    direct extension of local polynomials to bivariate splines`,
    Advances in Comp. Math. (21), 223--271

    Farin, G and Hansford, D, 2000, `The Essentials of CAGD`, Natic, MA:
    A K Peters, Ltd.

    See Also
    --------
    :meth:`naginterfaces.library.examples.fit.dim2_spline_ts_sctr_ex.main`
    """

    fun_name = 'naginterfaces.library.fit.dim2_spline_ts_evalv'

    _xevalv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xevalv'},
        xevalv,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nevalv_py = _b_utils._EngineFloat64ArrayType.get_shape(_xevalv, {'fun_name': fun_name, 'entity': 'xevalv'}, exp_rank=1)[0]
    _nevalv = _b_utils._EngineIntScalarType(
        _nevalv_py, {'fun_name': fun_name, 'entity': '_nevalv'},
    ).to_ctypes()
    _yevalv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'yevalv'},
        yevalv,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _fevalv = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nevalv_py,),
    )
    _b_fit.dim2_spline_ts_evalv(
        _nevalv,
        _xevalv,
        _yevalv,
        comm,
        _fevalv,
    )
    _fevalv_py = _fevalv
    return _fevalv_py

def dim2_spline_ts_evalm(xevalm, yevalm, comm):
    u"""
    Evaluation at a mesh of points of a spline computed by
    ``dim2_spline_ts_sctr``.

    ``dim2_spline_ts_evalm`` calculates a mesh of values of a spline
    computed by ``dim2_spline_ts_sctr``.

    For full information please refer to the NAG Library document for
    e02jf

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e02/e02jff.html

    Parameters
    ----------
    xevalm : float, array-like, shape (nxeval)
        The (x_i) values forming the mesh on which the spline is to be
        evaluated.

    yevalm : float, array-like, shape (nyeval)
        The (y_j) values forming the mesh on which the spline is to be
        evaluated.

    comm : dict, communication object
        Communication structure.

        This argument must have been initialized by prior calls to
        ``dim2_spline_ts_sctr`` and ``opt_set``.

    Returns
    -------
    fevalm : float, ndarray, shape (nxeval, nyeval)
        If no exception or warning is raised on exit `fevalm`[i-1,j-1]
        contains the computed spline value at (x_i,y_j).

    Raises
    ------
    NagValueError
        (`errno` 6)
            On entry, nxeval = *<value>*.

            Constraint: nxeval >= 1.

        (`errno` 7)
            On entry, nyeval = *<value>*.

            Constraint: nyeval >= 1.

        (`errno` 9)
            Option arrays are not initialized or are corrupted.

        (`errno` 10)
            The fitting routine has not been called, or the array of
            coefficients has been corrupted.

        (`errno` 13)
            On entry, `xevalm`[*<value>*] = *<value>* was outside the
            bounding box.

            Constraint: *<value>* <= `xevalm`[i-1] <= *<value>* for all
            i.

        (`errno` 14)
            On entry, `yevalm`[*<value>*] = *<value>* was outside the
            bounding box.

            Constraint: *<value>* <= `yevalm`[j-1] <= *<value>* for all
            j.

    References
    ----------
    Davydov, O, Morandi, R and Sestini, A, 2006, `Local hybrid
    approximation for scattered data fitting with bivariate splines`,
    Comput. Aided Geom. Design (23), 703--721

    Davydov, O, Sestini, A and Morandi, R, 2005, `Local RBF
    approximation for scattered data fitting with bivariate splines`,
    `Trends and Applications in Constructive Approximation`, M. G. de
    Bruin, D. H. Mache, and J. Szabados, Eds (ISNM Vol. 151),
    Birkhauser, 91--102

    Davydov, O and Zeilfelder, F, 2004, `Scattered data fitting by
    direct extension of local polynomials to bivariate splines`,
    Advances in Comp. Math. (21), 223--271

    Farin, G and Hansford, D, 2000, `The Essentials of CAGD`, Natic, MA:
    A K Peters, Ltd.
    """

    fun_name = 'naginterfaces.library.fit.dim2_spline_ts_evalm'

    _xevalm = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xevalm'},
        xevalm,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nxeval_py = _b_utils._EngineFloat64ArrayType.get_shape(_xevalm, {'fun_name': fun_name, 'entity': 'xevalm'}, exp_rank=1)[0]
    _nxeval = _b_utils._EngineIntScalarType(
        _nxeval_py, {'fun_name': fun_name, 'entity': '_nxeval'},
    ).to_ctypes()
    _yevalm = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'yevalm'},
        yevalm,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nyeval_py = _b_utils._EngineFloat64ArrayType.get_shape(_yevalm, {'fun_name': fun_name, 'entity': 'yevalm'}, exp_rank=1)[0]
    _nyeval = _b_utils._EngineIntScalarType(
        _nyeval_py, {'fun_name': fun_name, 'entity': '_nyeval'},
    ).to_ctypes()
    _fevalm = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nxeval_py, _nyeval_py),
        sorder=_b_utils._EngineData.default_sorder,
    )
    _b_fit.dim2_spline_ts_evalm(
        _nxeval,
        _nyeval,
        _xevalm,
        _yevalm,
        comm,
        _fevalm,
    )
    _fevalm_py = _fevalm
    return _fevalm_py

def pade_app(ia, ib, c):
    u"""
    Padé approximants.

    ``pade_app`` calculates the coefficients in a Padé approximant to a
    function from its user-supplied Maclaurin expansion.

    For full information please refer to the NAG Library document for
    e02ra

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e02/e02raf.html

    Parameters
    ----------
    ia : int
        `ia` must specify l+1 and `ib` must specify m+1, where l and m
        are the degrees of the numerator and denominator of the
        approximant, respectively.

    ib : int
        `ia` must specify l+1 and `ib` must specify m+1, where l and m
        are the degrees of the numerator and denominator of the
        approximant, respectively.

    c : float, array-like, shape (`ia`+`ib`-1)
        `c`[i-1] must specify, for i = 1,2,...,l+m+1, the coefficient of
        x^(i-1) in the given power series.

    Returns
    -------
    a : float, ndarray, shape (`ia`)
        `a`[j], for j = 1,2,...,l+1, contains the coefficient a_j in the
        numerator of the approximant.

    b : float, ndarray, shape (`ib`)
        `b`[k], for k = 1,2,...,m+1, contains the coefficient b_k in the
        denominator of the approximant.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, ic = *<value>*, `ia` = *<value>* and `ib` =
            *<value>*.

            Constraint: ic >= `ia`+`ib`-1.

        (`errno` 1)
            On entry, `ib` = *<value>*.

            Constraint: `ib` >= 1.

        (`errno` 1)
            On entry, `ia` = *<value>*.

            Constraint: `ia` >= 1.

        (`errno` 2)
            The Pade approximant is degenerate.

    References
    ----------
    Baker, G A Jr and Graves--Morris, P R, 1981, `Padé approximants,
    Part 1: Basic theory`, encyclopaedia of Mathematics and its
    Applications, Addison--Wesley

    Graves--Morris, P R, 1979, `The numerical calculation of Padé
    approximants`, Padé Approximation and its Applications. Lecture
    Notes in Mathematics, (ed L Wuytack) (765), 231--245, Adison--Wesley
    """

    fun_name = 'naginterfaces.library.fit.pade_app'

    _c = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ic_py = _b_utils._EngineFloat64ArrayType.get_shape(_c, {'fun_name': fun_name, 'entity': 'c'}, exp_rank=1)[0]
    _ic = _b_utils._EngineIntScalarType(
        _ic_py, {'fun_name': fun_name, 'entity': '_ic'},
    ).to_ctypes()
    _ia = _b_utils._EngineIntScalarType(
        ia, {'fun_name': fun_name, 'entity': 'ia'},
    ).to_ctypes()
    _ia_py = _b_utils._EngineIntScalarType.to_py(_ia)
    _a = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ia_py,),
    )
    _ib = _b_utils._EngineIntScalarType(
        ib, {'fun_name': fun_name, 'entity': 'ib'},
    ).to_ctypes()
    _ib_py = _b_utils._EngineIntScalarType.to_py(_ib)
    _b = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ib_py,),
    )
    _b_fit.pade_app(
        _ia,
        _ib,
        _c,
        _ic,
        _a,
        _b,
    )
    _a_py = _a
    _b_py = _b
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'b'],
    )(
        _a_py,
        _b_py,
    )
    return _return_data

def pade_eval(a, b, x):
    u"""
    Evaluation of fitted rational function as computed by ``pade_app``.

    ``pade_eval`` evaluates a rational function at a user-supplied
    point, given the numerator and denominator coefficients.

    For full information please refer to the NAG Library document for
    e02rb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e02/e02rbf.html

    Parameters
    ----------
    a : float, array-like, shape (ia)
        `a`[j], for j = 1,2,...,l+1, must contain the value of the
        coefficient a_j in the numerator of the rational function.

    b : float, array-like, shape (ib)
        `b`[k], for k = 1,2,...,m+1, must contain the value of the
        coefficient b_k in the denominator of the rational function.

    x : float
        The point x at which the rational function is to be evaluated.

    Returns
    -------
    ans : float
        The result of evaluating the rational function at the given
        point x.

    Raises
    ------
    NagValueError
        (`errno` 1)
            Evaluation at or near a pole.

        (`errno` 2)
            The first ib entries in `b` are zero: ib = *<value>*.

        (`errno` 2)
            On entry, ib = *<value>*.

            Constraint: ib >= 1.

        (`errno` 2)
            On entry, ia = *<value>*.

            Constraint: ia >= 1.

    References
    ----------
    Conte, S D and de Boor, C, 1965, `Elementary Numerical Analysis`,
    McGraw--Hill

    Peters, G and Wilkinson, J H, 1971, `Practical problems arising in
    the solution of polynomial equations`, J. Inst. Maths. Applics. (8),
    16--35
    """

    fun_name = 'naginterfaces.library.fit.pade_eval'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ia_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=1)[0]
    _ia = _b_utils._EngineIntScalarType(
        _ia_py, {'fun_name': fun_name, 'entity': '_ia'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ib_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=1)[0]
    _ib = _b_utils._EngineIntScalarType(
        _ib_py, {'fun_name': fun_name, 'entity': '_ib'},
    ).to_ctypes()
    _return_scalar = _b_fit.pade_eval(
        _a,
        _ia,
        _b,
        _ib,
        x,
    )
    _ans_py = _return_scalar
    return _ans_py

def dim2_spline_sort(lamda, mu, x, y):
    u"""
    Sort two-dimensional data into panels for fitting bicubic splines.

    ``dim2_spline_sort`` sorts two-dimensional data into rectangular
    panels.

    For full information please refer to the NAG Library document for
    e02za

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e02/e02zaf.html

    Parameters
    ----------
    lamda : float, array-like, shape (px)
        `lamda`[4] to `lamda`[px-5] must contain, in nondecreasing
        order, the intercepts on the x axis of the sides of the panels
        parallel to the y axis.

    mu : float, array-like, shape (py)
        `mu`[4] to `mu`[py-5] must contain, in nondecreasing order, the
        intercepts on the y axis of the sides of the panels parallel to
        the x axis.

    x : float, array-like, shape (m)
        The coordinates of the rth data point (x_r,y_r), for r =
        1,2,...,m.

    y : float, array-like, shape (m)
        The coordinates of the rth data point (x_r,y_r), for r =
        1,2,...,m.

    Returns
    -------
    point : int, ndarray, shape (m+(px-7)*(py-7))
        For i = 1,2,...,npoint, `point`[m+i-1] = I1 is the index of the
        first point in panel i, `point`[I1-1] = I2 is the index of the
        second point in panel i and so on.

        `point`[In-1] = 0 indicates that `x`[In-1],`y`[In-1] was the
        last point in the panel.

        The coordinates of points in panel i can be accessed in turn by
        means of the following instructions:

        ::

          i_n = point[m+i-1]
          while i_n != 0:
              xi = x[i_n-1]
              yi = y[i_n-1]
              .
              .
              .
              i_n = point[i_n]
          ...

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, I = *<value>*, `mu`[I-1] = *<value>* and `mu`[I-2]
            = *<value>*.

            Constraint: `mu`[I-1] >= `mu`[I-2].

        (`errno` 1)
            On entry, I = *<value>*, `lamda`[I-1] = *<value>* and
            `lamda`[I-2] = *<value>*.

            Constraint: `lamda`[I-1] >= `lamda`[I-2].

        (`errno` 2)
            On entry, npoint = *<value>*, m = *<value>*, px = *<value>*
            and py = *<value>*.

            Constraint: npoint >= m+(px-7)*(py-7).

        (`errno` 2)
            On entry, py = *<value>*.

            Constraint: py >= 8.

        (`errno` 2)
            On entry, px = *<value>*.

            Constraint: px >= 8.

        (`errno` 2)
            On entry, m = *<value>*.

            Constraint: m > 0.
    """

    fun_name = 'naginterfaces.library.fit.dim2_spline_sort'

    _lamda = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'lamda'},
        lamda,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _px_py = _b_utils._EngineFloat64ArrayType.get_shape(_lamda, {'fun_name': fun_name, 'entity': 'lamda'}, exp_rank=1)[0]
    _px = _b_utils._EngineIntScalarType(
        _px_py, {'fun_name': fun_name, 'entity': '_px'},
    ).to_ctypes()
    _mu = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'mu'},
        mu,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _py_py = _b_utils._EngineFloat64ArrayType.get_shape(_mu, {'fun_name': fun_name, 'entity': 'mu'}, exp_rank=1)[0]
    _py = _b_utils._EngineIntScalarType(
        _py_py, {'fun_name': fun_name, 'entity': '_py'},
    ).to_ctypes()
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _npoint = _b_utils._EngineIntScalarType(
        _m_py+(_px_py-7)*(_py_py-7), {'fun_name': fun_name, 'entity': '_npoint'},
    ).to_ctypes()
    _point = _b_utils._EngineIntArrayType.empty_ndarray(
        (_m_py+(_px_py-7)*(_py_py-7),),
    )
    _b_fit.dim2_spline_sort(
        _px,
        _py,
        _lamda,
        _mu,
        _m,
        _x,
        _y,
        _point,
        _npoint,
    )
    _point_py = _point
    return _point_py

def opt_set(optstr, comm):
    u"""
    Option setting routine.

    ``opt_set`` either initializes or resets the option arrays or sets a
    single option for supported problem solving functions in submodule
    ``fit``.
    Currently, only ``dim2_spline_ts_sctr`` is supported.

    For full information please refer to the NAG Library document for
    e02zk

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e02/e02zkf.html

    Parameters
    ----------
    optstr : str
        A string identifying the option to be set.

        Initialize = function name

            Initialize the option arrays `comm`['iopts'] and
            `comm`['opts'] for use with function function name, where
            function name is the name of the problem solving function
            you wish to use.

        Defaults

            Resets all options to their default values.

        option = optval

            See [Other Parameters for dim2_spline_ts_sctr] for details
            of valid values for option and optval. The equals sign (=)
            delimiter must be used to separate the option from its
            optval.

        The processing of `optstr` does not depend on its case.

        Each token in the option and optval component must be separated
        by at least one space.

    comm : dict, communication object, modified in place
        Communication structure.

        `On initial entry`: need not be set.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, the option in `optstr` was not recognized:
            `optstr` = *<value>*.

        (`errno` 12)
            On entry, the expected delimiter '=' was not found in
            `optstr`: `optstr` = *<value>*.

        (`errno` 13)
            On entry, could not convert the specified optval to an
            integer: `optstr` = *<value>*.

        (`errno` 13)
            On entry, could not convert the specified optval to a real:
            `optstr` = *<value>*.

        (`errno` 14)
            On entry, attempting to initialize the option arrays but
            specified function name was not valid: name = *<value>*.

        (`errno` 15)
            On entry, the optval supplied for the integer option is not
            valid.

            `optstr` = *<value>*.

        (`errno` 16)
            On entry, the optval supplied for the real option is not
            valid.

            `optstr` = *<value>*.

        (`errno` 17)
            On entry, the optval supplied for the character option is
            not valid.

            `optstr` = *<value>*.

        (`errno` 21)
            On entry, either the option arrays have not been initialized
            or they have been corrupted.

    See Also
    --------
    :meth:`naginterfaces.library.examples.fit.dim2_spline_ts_sctr_ex.main`
    """

    _b_fit.opt_set(
        optstr,
        comm,
    )

def opt_get(optstr, comm):
    u"""
    Option getting routine.

    ``opt_get`` is used to query the value of options available to
    supported problem solving functions in submodule ``fit``.
    Currently, only ``dim2_spline_ts_sctr`` is supported.

    For full information please refer to the NAG Library document for
    e02zl

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e02/e02zlf.html

    Parameters
    ----------
    optstr : str
        A string identifying the option whose current value is required.
        See [Other Parameters for dim2_spline_ts_sctr] for information
        on valid options. In addition, the following is a valid option:

        Identify

            ``opt_get`` returns the function name supplied to
            ``opt_set`` when the option arrays `comm`['iopts'] and
            `comm`['opts'] were initialized.

    comm : dict, communication object
        Communication structure.

        This argument must have been initialized by a prior call to
        ``opt_set``.

    Returns
    -------
    optvalue : dict
        The option-value ``dict``, with the following keys:

        ``'value'`` : float, int or str
            The value of the requested option.

        ``'annotation'`` : None or str
            Possible additional information about the option value.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, the option in `optstr` was not recognized:
            `optstr` = *<value>*.

        (`errno` 61)
            On entry, either the option arrays have not been initialized
            or they have been corrupted.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 41)
            On entry, `optstr` indicates a character option, but cvalue
            is too short to hold the stored value. The returned value
            will be truncated.
    """

    try:
        _return_scalars = _b_fit.opt_get(
            optstr,
            comm,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    fun_name = 'naginterfaces.library.fit.opt_get'

    _optype_py = _return_scalars.optype
    _ivalue_py = _return_scalars.ivalue
    _rvalue_py = _return_scalars.rvalue
    _cvalue_py = _return_scalars.cvalue
    if _optype_py == 1:
        _optvalue_py = {}
        _optvalue_py['value'] = _ivalue_py
        _optvalue_py['annotation'] = None
    elif _optype_py == 2:
        _optvalue_py = {}
        _optvalue_py['value'] = _rvalue_py
        _optvalue_py['annotation'] = None
    elif _optype_py == 3:
        _optvalue_py = {}
        _optvalue_py['value'] = _cvalue_py
        _optvalue_py['annotation'] = None
    elif _optype_py == 6:
        _optvalue_py = {}
        if _ivalue_py == _nag_e_consts.TSFIT_SOLVER_ID:
            _optvalue_py = {}
            _optvalue_py['value'] = 'dim2_spline_ts_sctr'
        else:
            raise _b_utils.NagException(
                {'fun_name': fun_name},
                'inconceivable option configuration',
            )
        _optvalue_py['annotation'] = None
    else:
        raise _b_utils.NagException(
            {'fun_name': fun_name},
            'inconceivable option configuration',
        )
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['optvalue'],
    )(
        _optvalue_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _optvalue_py
