# -*- coding: utf-8 -*-
# pylint: disable=chained-comparison,invalid-name,line-too-long,protected-access,too-many-arguments,too-many-boolean-expressions,too-many-branches,too-many-lines,too-many-locals,too-many-statements,unused-import
u"""
Interfaces for the NAG Mark 27.0 `glopt` Chapter.

``glopt`` - Global Optimization of a Function

Global optimization involves finding the absolute maximum or minimum
value of a function (the `objective function`) of several variables,
possibly subject to restrictions (defined by a set of bounds or
`constraint functions`) on the values of the variables.
Such problems can be much harder to solve than local optimization
problems (which are discussed in submodule ``opt``) because it is
difficult to determine whether a potential optimum found is global, and
because of the nonlocal methods required to avoid becoming trapped near
local optima.
Most optimization functions in the NAG Library are concerned with
function **minimization** only, since the problem of maximizing a given
objective function F is equivalent to minimizing -F.
In ``bnd_mcs_solve``, ``bnd_pso`` and ``nlp_pso``, you may specify
whether you are solving a minimization or maximization problem; in the
latter case, the required transformation of the objective function will
be carried out automatically.
In what follows we refer exclusively to minimization problems.

This introduction is a brief guide to the subject of global
optimization, designed for the casual user.
For further details you may find it beneficial to consult a more
detailed text, see Neumaier (2004).
Furthermore, much of the material in [the E04 Introduction] is also
relevant in this context and it is strongly recommended that you read
[the E04 Introduction].

For full information please refer to the NAG Library document

https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e05/e05intro.html

See Also
--------
``naginterfaces.library.examples.glopt`` :
    This subpackage contains examples for the ``glopt`` module.
    See also the :ref:`library_glopt_ex` subsection.
"""

# NAG Copyright 2017-2020.

import collections as _collections

from ..base import glopt as _b_glopt
from ..base import _nag_e_consts
from ..base import utils as _b_utils
from ..library import _utils

def bnd_mcs_init():
    u"""
    Initialization function for ``bnd_mcs_solve``.

    ``bnd_mcs_init`` is used to initialize communication data for the
    suite of multi-level coordinate search functions: ``bnd_mcs_solve``,
    ``bnd_mcs_optset_file``, ``bnd_mcs_optset_string``,
    ``bnd_mcs_optset_char``, ``bnd_mcs_optset_int``,
    ``bnd_mcs_optset_real``, ``bnd_mcs_option_check``,
    ``bnd_mcs_optget_char``, ``bnd_mcs_optget_int`` and
    ``bnd_mcs_optget_real``.

    For full information please refer to the NAG Library document for
    e05ja

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e05/e05jaf.html

    Returns
    -------
    comm : dict, communication object
        Communication structure.

    See Also
    --------
    :meth:`naginterfaces.library.examples.glopt.bnd_mcs_solve_ex.main`
    """

    _return_scalar = _b_glopt.bnd_mcs_init(
    )
    _comm_py = _return_scalar
    return _comm_py

def bnd_mcs_solve(objfun, ibound, bl, bu, comm, iinit=0, sdlist=3, inlist=None, numpts=None, initpt=None, monit=None, data=None, io_manager=None):
    u"""
    Global optimization by multi-level coordinate search, simple bounds,
    using function values only.

    Note: this function uses optional algorithmic parameters, see also:
    ``bnd_mcs_init``, ``bnd_mcs_optset_file``,
    ``bnd_mcs_optset_string``, ``bnd_mcs_optset_char``,
    ``bnd_mcs_optset_int``, ``bnd_mcs_optset_real``,
    ``bnd_mcs_option_check``, ``bnd_mcs_optget_char``,
    ``bnd_mcs_optget_int``, ``bnd_mcs_optget_real``.

    ``bnd_mcs_solve`` is designed to find the global minimum or maximum
    of an arbitrary function, subject to simple bound-constraints using
    a multi-level coordinate search method.
    Derivatives are not required, but convergence is only guaranteed if
    the objective function is continuous in a neighbourhood of a global
    optimum.
    It is not intended for large problems.

    The initialization function ``bnd_mcs_init`` **must** have been
    called before calling ``bnd_mcs_solve``.

    For full information please refer to the NAG Library document for
    e05jb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e05/e05jbf.html

    Parameters
    ----------
    objfun : callable f = objfun(x, nstate, data=None)
        `objfun` must evaluate the objective function F(x) for a
        specified n-vector x.

        Parameters
        ~~~~~~~~~~
        x : float, ndarray, shape (n)
            x, the vector at which the objective function is to be
            evaluated.

        nstate : int
            If `nstate` = 1 then ``bnd_mcs_solve`` is calling `objfun`
            for the first time. This argument setting allows you to save
            computation time if certain data must be read or calculated
            only once.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        f : float
            Must be set to the value of the objective function at x.

    ibound : int
        Indicates whether the facility for dealing with bounds of
        special forms is to be used. `ibound` must be set to one of the
        following values.

        `ibound` = 0

            You will supply l and u individually.

        `ibound` = 1

            There are no bounds on x.

        `ibound` = 2

            There are semi-infinite bounds 0 <= x.

        `ibound` = 3

            There are constant bounds l = l_1 and u = u_1.

        Note that it only makes sense to fix any components of x when
        `ibound` = 0.

    bl : float, array-like, shape (n)
        `bl` is l, the array of lower bounds.

    bu : float, array-like, shape (n)
        `bu` is u, the array of upper bounds.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``bnd_mcs_init``.

    iinit : int, optional
        Selects which initialization method to use.

        `iinit` = 0

            Simple initialization (boundary and midpoint), with

            `numpts`[i-1] = 3, `initpt`[i-1] = 2 and

            `inlist`[i-1,0] = `bl`[i-1], `inlist`[i-1,1] =
            (`bl`[i-1]+`bu`[i-1])/2, `inlist`[i-1,2] = `bu`[i-1],

            for i = 1,2,...,n.

        `iinit` = 1

            Simple initialization (off-boundary and midpoint), with

            `numpts`[i-1] = 3, `initpt`[i-1] = 2 and

            `inlist`[i-1,0] = (5`bl`[i-1]+`bu`[i-1])/6, `inlist`[i-1,1]
            = (`bl`[i-1]+`bu`[i-1])/2, `inlist`[i-1,2] =
            (`bl`[i-1]+5`bu`[i-1])/6,

            for i = 1,2,...,n.

        `iinit` = 2

            Initialization using linesearches.

        `iinit` = 3

            You are providing your own initialization list.

        `iinit` = 4

            Generate a random initialization list.

        For more information on methods `iinit` = 2, 3 or 4 see
        [Initialization and Sweeps].

        If 'infinite' values (as determined by the value of the option
        'Infinite Bound Size') are detected by ``bnd_mcs_solve`` when
        you are using a simple initialization method (`iinit` = 0 or 1),
        a safeguarded initialization procedure will be attempted, to
        avoid overflow.

    sdlist : int, optional
        When `iinit` = 4, this argument is used to determine the maximal
        number of random initialization points to use in any given
        dimension. Otherwise a value is inferred internally.

    inlist : None or float, array-like, shape (n, `sdlist`), optional
        This argument need not be set on entry if you wish to use one of
        the preset initialization methods (`iinit` != 3).

        `inlist` is the 'initialization list': whenever a sub-box in the
        algorithm is split for the first time (either during the
        `initialization procedure` or later), for each non-fixed
        coordinate i the split is done at the values
        `inlist`[i-1,0:`numpts`[i-1]], as well as at some adaptively
        chosen intermediate points.

        The array sections `inlist`[i-1,0:`numpts`[i-1]], for i =
        1,2,...,n, must be in ascending order with each entry being
        distinct.

        In this context, 'distinct' should be taken to mean relative to
        the safe-range parameter (see ``machine.real_safe``).

    numpts : None or int, array-like, shape (n), optional
        This argument need not be set on entry if you wish to use one of
        the preset initialization methods (`iinit` != 3).

        `numpts` encodes the number of splitting points in each
        non-fixed dimension.

    initpt : None or int, array-like, shape (n), optional
        This argument need not be set on entry if you wish to use one of
        the preset initialization methods (`iinit` != 3).

        You must designate a point stored in `inlist` that you wish
        ``bnd_mcs_solve`` to consider as an 'initial point' for the
        purposes of the splitting procedure.

        Call this initial point x*.

        The coordinates of x* correspond to a set of indices J_i, for i
        = 1,2,...,n, such that x_i^* is stored in `inlist`[i-1,J_i-1],
        for i = 1,2,...,n.

        You must set `initpt`[i-1] = J_i, for i = 1,2,...,n.

    monit : None or callable monit(ncall, xbest, icount, inlist, numpts,
        initpt, xbaskt, boxl, boxu, nstate, data=None), optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `monit` may be used to monitor the optimization process.

        It is invoked upon every successful completion of the procedure
        in which a sub-box is considered for splitting.

        It will also be called just before ``bnd_mcs_solve`` exits if
        that splitting procedure was not successful.

        If no monitoring is required, `monit` may be None.

        Parameters
        ~~~~~~~~~~
        ncall : int
            The cumulative number of calls to `objfun`.

        xbest : float, ndarray, shape (n)
            The current best point.

        icount : int, ndarray, shape (6)
            An array of counters.

            `icount`[0]

                nboxes, the current number of sub-boxes.

            `icount`[1]

                ncloc, the cumulative number of calls to `objfun` made
                in local searches.

            `icount`[2]

                nloc, the cumulative number of points used as start
                points for local searches.

            `icount`[3]

                nsweep, the cumulative number of sweeps through levels.

            `icount`[4]

                m, the cumulative number of splits by initialization
                list.

            `icount`[5]

                s, the current lowest level containing non-split boxes.

        inlist : float, ndarray, shape (n, ninit)
            The initialization list.

        numpts : int, ndarray, shape (n)
            The number of points in each coordinate at which to split
            according to the initialization list `inlist`.

        initpt : int, ndarray, shape (n)
            A pointer to the 'initial point' in `inlist`. Element
            `initpt`[i-1] is the column index in `inlist` of the ith
            coordinate of the initial point.

        xbaskt : float, ndarray, shape (n, nbaskt)
            Note: the jth candidate minimum has its ith coordinate
            stored in `xbaskt`[j-1,i-1], for j = 1,2,...,nbaskt, for i =
            1,2,...,n.

            The 'shopping basket' of candidate minima.

        boxl : float, ndarray, shape (n)
            The array of lower bounds of the current search box.

        boxu : float, ndarray, shape (n)
            The array of upper bounds of the current search box.

        nstate : int
            Is set by ``bnd_mcs_solve`` to indicate at what stage of the
            minimization `monit` was called.

            `nstate` = 1

                This is the first time that `monit` has been called.

            `nstate` = -1

                This is the last time `monit` will be called.

            `nstate` = 0

                This is the first `and` last time `monit` will be
                called.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

    data : arbitrary, optional
        User-communication data for callback functions.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    bl : float, ndarray, shape (n)
        `bl` is l, the array of lower bounds.

    bu : float, ndarray, shape (n)
        `bu` is u, the array of upper bounds.

    inlist : float, ndarray, shape (n, `sdlist`)
        Unless `errno` = 1, 2 or ``NagMemoryError`` on exit, the actual
        initialization data used by ``bnd_mcs_solve``. If you wish to
        monitor the contents of `inlist` you are advised to do so solely
        through `monit`, not through the output value here.

    numpts : int, ndarray, shape (n)
        Unless `errno` = 1, 2 or ``NagMemoryError`` on exit, the actual
        initialization data used by ``bnd_mcs_solve``.

    initpt : int, ndarray, shape (n)
        Unless `errno` = 1, 2 or ``NagMemoryError`` on exit, the actual
        initialization data used by ``bnd_mcs_solve``.

    x : float, ndarray, shape (n)
        If no exception or warning is raised, contains an estimate of
        the global optimum (see also [Accuracy]).

    obj : float
        If no exception or warning is raised, contains the function
        value at `x`.

        If you request early termination of ``bnd_mcs_solve`` in
        `objfun` or `monit`, there is no guarantee that the function
        value at `x` equals `obj`.

    Other Parameters
    ----------------
    'Defaults' : valueless
        This special keyword is used to reset all options to their
        default values, and any random state stored in the array comm
        will be destroyed.

        Any option value given with this keyword will be ignored.
        This option cannot be queried or got.

    'Function Evaluations Limit' : int
        Default = 100n_r^2

        This puts an approximate limit on the number of function calls
        allowed.
        The total number of calls made is checked at the top of an
        internal iteration loop, so it is possible that a few calls more
        than nf may be made.

    'Infinite Bound Size' : float
        Default = r_max^1/4

        This defines the 'infinite' bound infbnd in the definition of
        the problem constraints.
        Any upper bound greater than or equal to infbnd will be regarded
        as infinity (and similarly any lower bound less than or equal to
        -infbnd will be regarded as - infinity).

    'Local Searches' : str
        Default = 'ON'

        If you want to try to accelerate convergence of
        ``bnd_mcs_solve`` by starting local searches from candidate
        minima, you will require lcsrch to be 'ON'.

    'Local Searches Limit' : int
        Default = 50

        This defines the maximal number of iterations to be used in the
        trust region loop of the local-search procedure.

    'Local Searches Tolerance' : float
        Default = 2 epsilon

        The value of loctol is the multiplier used during local searches
        as a stopping criterion for when the approximated gradient is
        small, in the sense described in [Local Search].

    'Minimize' : valueless
        Default

        These keywords specify the required direction of optimization.
        Any option value given with these keywords will be ignored.

    'Maximize' : valueless
        These keywords specify the required direction of optimization.
        Any option value given with these keywords will be ignored.

    'Nolist' : valueless
        Default

        Option 'List' enables printing of each option specification as
        it is supplied. 'Nolist' suppresses this printing.

    'List' : valueless
        Option 'List' enables printing of each option specification as
        it is supplied. 'Nolist' suppresses this printing.

    'Repeatability' : str
        Default = 'OFF'

        For use with random initialization lists (`iinit` = 4).
        When set to 'ON', an internally-initialized random state is
        stored in the array comm for use in subsequent calls to
        ``bnd_mcs_solve``.

    'Splits Limit' : int
        Default = floor(d(n_r+2)/3)

        Along with the initialization list `inlist`, this defines a
        limit on the number of times the root box will be split along
        any single coordinate direction.
        If 'Local Searches' is 'OFF' you may find the default value to
        be too small.

    'Static Limit' : int
        Default = 3n_r

        As the default termination criterion, computation stops when the
        best function value is static for stclim sweeps through levels.
        This argument is ignored if you have specified a target value to
        reach in 'Target Objective Value'.

    'Target Objective Error' : float
        Default = epsilon^(1/4)

        If you have given a target objective value to reach in objval
        (the value of the option 'Target Objective Value'), objerr sets
        your desired relative error (from above if 'Minimize' is set,
        from below if 'Maximize' is set) between `obj` and objval, as
        described in [Accuracy].
        See also the description of the option 'Target Objective
        Safeguard'.

    'Target Objective Safeguard' : float
        Default = epsilon^(1/2)

        If you have given a target objective value to reach in objval
        (the value of the option 'Target Objective Value'), objsfg sets
        your desired safeguarded termination tolerance, for when objval
        is close to zero.

    'Target Objective Value' : float
        This argument may be set if you wish ``bnd_mcs_solve`` to use a
        specific value as the target function value to reach during the
        optimization.
        Setting objval overrides the default termination criterion
        determined by the option 'Static Limit'.

    Raises
    ------
    NagValueError
        (`errno` 1)
            Initialization function ``bnd_mcs_init`` has not been
            called.

        (`errno` 2)
            On entry, n = *<value>*.

            Constraint: n > 0.

        (`errno` 2)
            On entry, `ibound` = *<value>*.

            Constraint: `ibound` = 0, 1, 2 or 3.

        (`errno` 2)
            On entry, `ibound` = 0 or 3 and `bl`[i-1] = *<value>*,
            `bu`[i-1] = *<value>* and i = *<value>*.

            Constraint: if `ibound` = 0 then `bl`[i-1] <= `bu`[i-1], for
            i = 1,2,...,n; if `ibound` = 3 then `bl`[0] < `bu`[0].

        (`errno` 2)
            On entry, `ibound` = 3 and `bl`[0] = `bu`[0] = *<value>*.

            Constraint: if `ibound` = 3 then `bl`[0] < `bu`[0].

        (`errno` 2)
            On entry, `iinit` = *<value>*.

            Constraint: `iinit` = 0, 1, 2, 3 or 4.

        (`errno` 2)
            On entry, `iinit` = *<value>* and `sdlist` = *<value>*.

            Constraint: if `iinit` != 2 then `sdlist` >= 3.

        (`errno` 2)
            On entry, user-supplied `numpts`[i-1] = *<value>*, i =
            *<value>*.

            Constraint: if `x`[i-1] is not fixed then `numpts`[i-1] >=
            3, for i = 1,2,...,n.

        (`errno` 2)
            On entry, user-supplied `numpts`[i-1] = *<value>*, i =
            *<value>* and `sdlist` = *<value>*.

            Constraint: if `x`[i-1] is not fixed then `numpts`[i-1] <=
            `sdlist`, for i = 1,2,...,n.

        (`errno` 2)
            On entry, user-supplied `initpt`[i-1] = *<value>*, i =
            *<value>*.

            Constraint: if `x`[i-1] is not fixed then `initpt`[i-1] >=
            1, for i = 1,2,...,n.

        (`errno` 2)
            On entry, user-supplied `initpt`[i-1] = *<value>*, i =
            *<value>* and `sdlist` = *<value>*.

            Constraint: if `x`[i-1] is not fixed then `initpt`[i-1] <=
            `sdlist`, for i = 1,2,...,n.

        (`errno` 2)
            On entry, user-supplied section `inlist`[i,1:`numpts`[i-1]]
            contained ndist distinct elements, and ndist <
            `numpts`[i-1]: ndist = *<value>*, `numpts`[i-1] = *<value>*,
            i = *<value>*.

        (`errno` 2)
            On entry, user-supplied `inlist`[i-1,j-1] = *<value>*, i =
            *<value>*, j = *<value>*, and `bl`[i-1] = *<value>*.

            Constraint: if `x`[i-1] is not fixed then `inlist`[i-1,j-1]
            >= `bl`[i-1], for j = 1,2,...,`numpts`[i-1], for i =
            1,2,...,n.

        (`errno` 2)
            On entry, user-supplied `inlist`[i-1,j-1] = *<value>*, i =
            *<value>*, j = *<value>*, and `bu`[i-1] = *<value>*.

            Constraint: if `x`[i-1] is not fixed then `inlist`[i-1,j-1]
            <= `bu`[i-1], for j = 1,2,...,`numpts`[i-1], for i =
            1,2,...,n.

        (`errno` 2)
            On entry, user-supplied section `inlist`[i,1:`numpts`[i-1]]
            was not in ascending order: `numpts`[i-1] = *<value>*, i =
            *<value>*.

        (`errno` 2)
            The number of non-fixed variables n_r = 0.

            Constraint: n_r > 0.

        (`errno` 2)
            A value of 'Splits Limit' (smax) smaller than n_r+3 was set:
            smax = *<value>*, n_r = *<value>*.

        (`errno` 3)
            A finite initialization list could not be computed
            internally. Consider reformulating the bounds on the
            problem, try providing your own initialization list, use the
            randomization option (`iinit` = 4) or vary the value of
            'Infinite Bound Size'.

        (`errno` 3)
            The user-supplied initialization list contained infinite
            values, as determined by the option 'Infinite Bound Size'.

        (`errno` 7)
            An error occurred during linesearching. It is likely that
            your objective function is badly scaled: try rescaling it.
            Also, try relaxing the bounds or use a different
            initialization method. If the problem persists, please
            contact NAG quoting error code *<value>*.

        (`errno` 7)
            An error occurred during initialization. It is likely that
            points from the initialization list are very close together.
            Try relaxing the bounds on the variables or use a different
            initialization method.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 4)
            The division procedure completed but your target value could
            not be reached.

            Despite every sub-box being processed 'Splits Limit' times,
            the target value you provided in 'Target Objective Value'
            could not be found to the tolerances given in 'Target
            Objective Error' and 'Target Objective Safeguard'. You could
            try reducing 'Splits Limit' or the objective tolerances.

        (`errno` 5)
            The function evaluations limit was exceeded.

            Approximately 'Function Evaluations Limit' function calls
            have been made without your chosen termination criterion
            being satisfied.

    NagCallbackTerminateWarning
        (`errno` 6)
            User-supplied monitoring function requested termination.

        (`errno` 6)
            User-supplied objective function requested termination.

    References
    ----------
    Huyer, W and Neumaier, A, 1999, `Global optimization by multi-level
    coordinate search`, Journal of Global Optimization (14), 331--355

    See Also
    --------
    :meth:`naginterfaces.library.examples.glopt.bnd_mcs_solve_ex.main`
    """

    def cb_objfun(n, x, nstate, data=None):
        """Wrapping user's objfun. Called as base objfun."""
        if data is not None:
            _f_objfun = objfun(
                x,
                nstate,
                data,
            )
        else:
            _f_objfun = objfun(
                x,
                nstate,
            )
        _f_objfun_base = _f_objfun
        return _f_objfun_base

    def cb_monit(n, ncall, xbest, icount, ninit, inlist, numpts, initpt, nbaskt, xbaskt, boxl, boxu, nstate, data=None):
        """Wrapping user's monit. Called as base monit."""
        if data is not None:
            monit(
                ncall,
                xbest,
                icount,
                inlist,
                numpts,
                initpt,
                xbaskt,
                boxl,
                boxu,
                nstate,
                data,
            )
        else:
            monit(
                ncall,
                xbest,
                icount,
                inlist,
                numpts,
                initpt,
                xbaskt,
                boxl,
                boxu,
                nstate,
            )

    fun_name = 'naginterfaces.library.glopt.bnd_mcs_solve'

    _inlist = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'inlist'},
        inlist,
        copy=True,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_inlist, 'inlist'),),
    )
    _bl = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'bl'},
        bl,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_bl, {'fun_name': fun_name, 'entity': 'bl'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b_utils._check_callable(objfun, {'fun_name': fun_name, 'entity': 'objfun'})
    _objfun = cb_objfun
    _bu = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'bu'},
        bu,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _iinit = _b_utils._EngineIntScalarType(
        iinit, {'fun_name': fun_name, 'entity': 'iinit'},
    ).to_ctypes()
    _iinit_py = _b_utils._EngineIntScalarType.to_py(_iinit)
    _sdlist = _b_utils._EngineIntScalarType(
        sdlist, {'fun_name': fun_name, 'entity': 'sdlist'},
    ).to_ctypes()
    _sdlist_py = _b_utils._EngineIntScalarType.to_py(_sdlist)
    if inlist is not None:
        _sdlist_py = _b_utils._EngineFloat64ArrayType.get_shape(_inlist, {'fun_name': fun_name, 'entity': 'inlist'}, exp_rank=2, allow_none=True)[1]
        _sdlist = _b_utils._EngineIntScalarType(
            _sdlist_py, {'fun_name': fun_name, 'entity': '_sdlist'},
        ).to_ctypes()
    else:
        if _iinit_py == 2:
            _sdlist = 192
            _sdlist_py = _sdlist
        elif _iinit_py == 4:
            _sdlist = _b_utils._EngineIntScalarType(
                _sdlist_py, {'fun_name': fun_name, 'entity': '_sdlist'},
            ).to_ctypes()
            _sdlist_py = _b_utils._EngineIntScalarType.to_py(_sdlist)
        else:
            _sdlist = 3
            _sdlist_py = _sdlist
    if _iinit_py == 3:
        if inlist is None:
            raise _b_utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'inlist'},
                'must not be None when iinit = 3',
            )
    if inlist is None:
        _inlist = _b_utils._EngineFloat64ArrayType.empty_ndarray(
            (_n_py, _sdlist_py),
            sorder=_m_sorder,
        )
    if _iinit_py == 3:
        if numpts is None:
            raise _b_utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'numpts'},
                'must not be None when iinit = 3',
            )
    if numpts is None:
        _numpts = _b_utils._EngineIntArrayType.empty_ndarray(
            (_n_py,),
        )
    else:
        _numpts = _utils._to_ndarray(
            {'fun_name': fun_name, 'entity': 'numpts'},
            numpts,
            copy=True,
        )
    if _iinit_py == 3:
        if initpt is None:
            raise _b_utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'initpt'},
                'must not be None when iinit = 3',
            )
    if initpt is None:
        _initpt = _b_utils._EngineIntArrayType.empty_ndarray(
            (_n_py,),
        )
    else:
        _initpt = _utils._to_ndarray(
            {'fun_name': fun_name, 'entity': 'initpt'},
            initpt,
            copy=True,
        )
    _b_utils._check_callable(monit, {'fun_name': fun_name, 'entity': 'monit'}, allow_none=True)
    _monit = None if monit is None else cb_monit
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _return_scalar = _b_glopt.bnd_mcs_solve(
            _n,
            _objfun,
            ibound,
            _iinit,
            _bl,
            _bu,
            _sdlist,
            _inlist,
            _numpts,
            _initpt,
            _monit,
            _x,
            comm,
            data,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _bl_py = _bl
    _bu_py = _bu
    _inlist_py = _inlist
    _numpts_py = _numpts
    _initpt_py = _initpt
    _x_py = _x
    _obj_py = _return_scalar
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['bl', 'bu', 'inlist', 'numpts', 'initpt', 'x', 'obj'],
    )(
        _bl_py,
        _bu_py,
        _inlist_py,
        _numpts_py,
        _initpt_py,
        _x_py,
        _obj_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def bnd_mcs_optset_file(iopts, comm, io_manager=None):
    u"""
    Supply option values for ``bnd_mcs_solve`` from external file.

    ``bnd_mcs_optset_file`` may be used to supply options to
    ``bnd_mcs_solve`` from an external file.
    The initialization function ``bnd_mcs_init`` **must** have been
    called before calling ``bnd_mcs_optset_file``.

    For full information please refer to the NAG Library document for
    e05jc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e05/e05jcf.html

    Parameters
    ----------
    iopts : int
        The unit number (see
        :meth:`~naginterfaces.base.utils.FileObjManager.unit_from_fileobj`)
        of the option file to be read.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``bnd_mcs_init``.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Raises
    ------
    NagValueError
        (`errno` 1)
            Initialization function ``bnd_mcs_init`` has not been
            called.

        (`errno` 2)
            Could not read options file.

        (`errno` 2)
            At least one option from the options file could not be
            recognized.

        (`errno` 2)
            End-of-file found before BEGIN.

        (`errno` 2)
            BEGIN found, but end-of-file found before END.

        (`errno` 3)
            Attempt to assign a non-positive value of 'Function
            Evaluations Limit' (nf): nf = *<value>*.

        (`errno` 3)
            Attempt to assign an out-of-bounds value of 'Infinite Bound
            Size' (infbnd): infbnd = *<value>*.

        (`errno` 3)
            Attempt to assign a non-positive value of 'Static Limit'
            (stclim): stclim = *<value>*.

        (`errno` 3)
            Attempt to assign too small a value of 'Target Objective
            Error' (objerr): objerr = *<value>*.

        (`errno` 3)
            Attempt to assign too small a value of 'Target Objective
            Safeguard' (objsfg): objsfg = *<value>*.

        (`errno` 3)
            Attempt to assign a non-positive value of 'Local Searches
            Limit' (loclim): loclim = *<value>*.

        (`errno` 3)
            Attempt to assign too small a value of 'Local Searches
            Tolerance' (loctol): loctol = *<value>*.

        (`errno` 3)
            Attempt to assign an illegal value of 'Local Searches'
            (lcsrch): lcsrch = *<value>*.

        (`errno` 3)
            Attempt to assign an illegal value of 'Repeatability'
            (repeat): repeat = *<value>*.

        (`errno` 5)
            One of the numeric values to be set could not be parsed.
            Check that all such strings specify valid integer or real
            values.
    """

    _b_glopt.bnd_mcs_optset_file(
        iopts,
        comm,
        io_manager,
    )

def bnd_mcs_optset_string(optstr, comm, io_manager=None):
    u"""
    Set a single option for ``bnd_mcs_solve`` from a character string.

    ``bnd_mcs_optset_string`` may be used to supply individual options
    to ``bnd_mcs_solve``.
    The initialization function ``bnd_mcs_init`` **must** have been
    called before calling ``bnd_mcs_optset_string``.

    For full information please refer to the NAG Library document for
    e05jd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e05/e05jdf.html

    Parameters
    ----------
    optstr : str
        A string defining a single option (as described in [Description]
        and in [Other Parameters for bnd_mcs_solve]). The implied data
        type (character, integer or real) of each value to set **must**
        match that expected by the option.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``bnd_mcs_init``.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Raises
    ------
    NagValueError
        (`errno` 1)
            Initialization function ``bnd_mcs_init`` has not been
            called.

        (`errno` 2)
            The supplied option is invalid. A keyword or keyword
            combination was not recognized.

        (`errno` 3)
            Attempt to assign a non-positive value of 'Function
            Evaluations Limit' (nf): nf = *<value>*.

        (`errno` 3)
            Attempt to assign an out-of-bounds value of 'Infinite Bound
            Size' (infbnd): infbnd = *<value>*.

        (`errno` 3)
            Attempt to assign a non-positive value of 'Static Limit'
            (stclim): stclim = *<value>*.

        (`errno` 3)
            Attempt to assign too small a value of 'Target Objective
            Error' (objerr): objerr = *<value>*.

        (`errno` 3)
            Attempt to assign too small a value of 'Target Objective
            Safeguard' (objsfg): objsfg = *<value>*.

        (`errno` 3)
            Attempt to assign a non-positive value of 'Local Searches
            Limit' (loclim): loclim = *<value>*.

        (`errno` 3)
            Attempt to assign too small a value of 'Local Searches
            Tolerance' (loctol): loctol = *<value>*.

        (`errno` 3)
            Attempt to assign an illegal value of 'Local Searches'
            (lcsrch): lcsrch = *<value>*.

        (`errno` 3)
            Attempt to assign an illegal value of 'Repeatability'
            (repeat): repeat = *<value>*.

        (`errno` 5)
            The value to be set could not be parsed. Check that it
            specifies a valid integer or real value.
    """

    _b_glopt.bnd_mcs_optset_string(
        optstr,
        comm,
        io_manager,
    )

def bnd_mcs_optset_char(optstr, cvalue, comm, io_manager=None):
    u"""
    Set a single option for ``bnd_mcs_solve`` from an 'ON'/'OFF'-valued
    character argument.

    ``bnd_mcs_optset_char`` may be used to supply individual
    'ON'/'OFF'-valued character options to ``bnd_mcs_solve``.
    The initialization function ``bnd_mcs_init`` **must** have been
    called before calling ``bnd_mcs_optset_char``.

    For full information please refer to the NAG Library document for
    e05je

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e05/e05jef.html

    Parameters
    ----------
    optstr : str
        A string defining a single 'ON'/'OFF'-valued character option
        (as described in [Other Parameters for bnd_mcs_solve]).

    cvalue : str
        The 'ON'/'OFF' value associated with the keyword in `optstr`.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``bnd_mcs_init``.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Raises
    ------
    NagValueError
        (`errno` 1)
            Initialization function ``bnd_mcs_init`` has not been
            called.

        (`errno` 2)
            The supplied option is invalid. A keyword or keyword
            combination was not recognized.

        (`errno` 3)
            Attempt to assign an illegal value of 'Local Searches'
            (lcsrch): lcsrch = *<value>*.

        (`errno` 3)
            Attempt to assign an illegal value of 'Repeatability'
            (repeat): repeat = *<value>*.

    See Also
    --------
    :meth:`naginterfaces.library.examples.glopt.bnd_mcs_solve_ex.main`
    """

    _b_glopt.bnd_mcs_optset_char(
        optstr,
        cvalue,
        comm,
        io_manager,
    )

def bnd_mcs_optset_int(optstr, ivalue, comm, io_manager=None):
    u"""
    Set a single option for ``bnd_mcs_solve`` from an integer argument.

    ``bnd_mcs_optset_int`` may be used to supply individual integer
    options to ``bnd_mcs_solve``.
    The initialization function ``bnd_mcs_init`` **must** have been
    called before calling ``bnd_mcs_optset_int``.

    For full information please refer to the NAG Library document for
    e05jf

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e05/e05jff.html

    Parameters
    ----------
    optstr : str
        A string identifying an integer-valued option (as described in
        [Other Parameters for bnd_mcs_solve]).

    ivalue : int
        An integer value associated with the option in `optstr`.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``bnd_mcs_init``.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Raises
    ------
    NagValueError
        (`errno` 1)
            Initialization function ``bnd_mcs_init`` has not been
            called.

        (`errno` 2)
            The supplied option is invalid. A keyword or keyword
            combination was not recognized.

        (`errno` 3)
            Attempt to assign a non-positive value of 'Function
            Evaluations Limit' (nf): nf = *<value>*.

        (`errno` 3)
            Attempt to assign a non-positive value of 'Static Limit'
            (stclim): stclim = *<value>*.

        (`errno` 3)
            Attempt to assign a non-positive value of 'Local Searches
            Limit' (loclim): loclim = *<value>*.

    See Also
    --------
    :meth:`naginterfaces.library.examples.glopt.bnd_mcs_solve_ex.main`
    """

    _b_glopt.bnd_mcs_optset_int(
        optstr,
        ivalue,
        comm,
        io_manager,
    )

def bnd_mcs_optset_real(optstr, rvalue, comm, io_manager=None):
    u"""
    Set a single option for ``bnd_mcs_solve`` from a real argument.

    ``bnd_mcs_optset_real`` may be used to supply individual real
    options to ``bnd_mcs_solve``.
    The initialization function ``bnd_mcs_init`` **must** have been
    called before calling ``bnd_mcs_optset_real``.

    For full information please refer to the NAG Library document for
    e05jg

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e05/e05jgf.html

    Parameters
    ----------
    optstr : str
        A string identifying a real-valued option (as described in
        [Other Parameters for bnd_mcs_solve]).

    rvalue : float
        The value associated with the option in `optstr`.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``bnd_mcs_init``.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Raises
    ------
    NagValueError
        (`errno` 1)
            Initialization function ``bnd_mcs_init`` has not been
            called.

        (`errno` 2)
            The supplied option is invalid. A keyword or keyword
            combination was not recognized.

        (`errno` 3)
            Attempt to assign an out-of-bounds value of 'Infinite Bound
            Size' (infbnd): infbnd = *<value>*.

        (`errno` 3)
            Attempt to assign too small a value of 'Target Objective
            Error' (objerr): objerr = *<value>*.

        (`errno` 3)
            Attempt to assign too small a value of 'Target Objective
            Safeguard' (objsfg): objsfg = *<value>*.

        (`errno` 3)
            Attempt to assign too small a value of 'Local Searches
            Tolerance' (loctol): loctol = *<value>*.

    See Also
    --------
    :meth:`naginterfaces.library.examples.glopt.bnd_mcs_solve_ex.main`
    """

    _b_glopt.bnd_mcs_optset_real(
        optstr,
        rvalue,
        comm,
        io_manager,
    )

def bnd_mcs_option_check(optstr, comm):
    u"""
    Determine whether an option for ``bnd_mcs_solve`` has been set by
    you or not.

    ``bnd_mcs_option_check`` is used to check if you have set an option
    of ``bnd_mcs_solve``.
    The initialization function ``bnd_mcs_init`` **must** have been
    called before calling ``bnd_mcs_option_check``.

    For full information please refer to the NAG Library document for
    e05jh

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e05/e05jhf.html

    Parameters
    ----------
    optstr : str
        A string identifying an option (as described in [Other
        Parameters for bnd_mcs_solve]).

    comm : dict, communication object
        Communication structure.

        This argument must have been initialized by a prior call to
        ``bnd_mcs_init``.

    Returns
    -------
    ichkopt : int
        ichkopt = 1 if you have set the option of ``bnd_mcs_solve``.

    Raises
    ------
    NagValueError
        (`errno` 1)
            Initialization function ``bnd_mcs_init`` has not been
            called.

        (`errno` 2)
            The supplied option is invalid. A keyword or keyword
            combination was not recognized.
    """

    _return_scalar = _b_glopt.bnd_mcs_option_check(
        optstr,
        comm,
    )
    return _return_scalar

def bnd_mcs_optget_char(optstr, comm):
    u"""
    Get the setting of an 'ON'/'OFF'-valued character option of
    ``bnd_mcs_solve``.

    ``bnd_mcs_optget_char`` is used to get the value of an
    'ON'/'OFF'-valued character ``bnd_mcs_solve`` option.
    ``bnd_mcs_optget_char`` can be used before or after calling
    ``bnd_mcs_solve``, but the initialization function ``bnd_mcs_init``
    **must** have been called before calling ``bnd_mcs_optget_char``.

    For full information please refer to the NAG Library document for
    e05jj

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e05/e05jjf.html

    Parameters
    ----------
    optstr : str
        A string defining a single 'ON'/'OFF'-valued character option
        (as described in [Other Parameters for bnd_mcs_solve]).

    comm : dict, communication object
        Communication structure.

        This argument must have been initialized by a prior call to
        ``bnd_mcs_init``.

    Returns
    -------
    cvalue : str
        If no exception or warning is raised on exit, `cvalue` contains
        the 'ON'/'OFF' value associated with the keyword in `optstr`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            Initialization function ``bnd_mcs_init`` has not been
            called.

        (`errno` 2)
            The supplied option is invalid. A keyword or keyword
            combination was not recognized.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    _return_scalar = _b_glopt.bnd_mcs_optget_char(
        optstr,
        comm,
    )
    _cvalue_py = _return_scalar
    return _cvalue_py

def bnd_mcs_optget_int(optstr, comm):
    u"""
    Get the setting of an integer valued option of ``bnd_mcs_solve``.

    ``bnd_mcs_optget_int`` is used to get the value of an integer
    ``bnd_mcs_solve`` option. ``bnd_mcs_optget_int`` can be used before
    or after calling ``bnd_mcs_solve``, but the initialization function
    ``bnd_mcs_init`` **must** have been called before calling
    ``bnd_mcs_optget_int``.

    For full information please refer to the NAG Library document for
    e05jk

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e05/e05jkf.html

    Parameters
    ----------
    optstr : str
        A string identifying an integer-valued option (as described in
        [Other Parameters for bnd_mcs_solve]).

    comm : dict, communication object
        Communication structure.

        This argument must have been initialized by a prior call to
        ``bnd_mcs_init``.

    Returns
    -------
    ivalue : int
        If no exception or warning is raised on exit, `ivalue` contains
        the integer value associated with the option in `optstr`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            Initialization function ``bnd_mcs_init`` has not been
            called.

        (`errno` 2)
            The supplied option is invalid. A keyword or keyword
            combination was not recognized.
    """

    _return_scalar = _b_glopt.bnd_mcs_optget_int(
        optstr,
        comm,
    )
    _ivalue_py = _return_scalar
    return _ivalue_py

def bnd_mcs_optget_real(optstr, comm):
    u"""
    Get the setting of a real valued option of ``bnd_mcs_solve``.

    ``bnd_mcs_optget_real`` is used to get the value of a real
    ``bnd_mcs_solve`` option. ``bnd_mcs_optget_real`` can be used before
    or after calling ``bnd_mcs_solve``, but the initialization function
    ``bnd_mcs_init`` **must** have been called before calling
    ``bnd_mcs_optget_real``.

    For full information please refer to the NAG Library document for
    e05jl

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e05/e05jlf.html

    Parameters
    ----------
    optstr : str
        A string identifying a real-valued option (as described in
        [Other Parameters for bnd_mcs_solve]).

    comm : dict, communication object
        Communication structure.

        This argument must have been initialized by a prior call to
        ``bnd_mcs_init``.

    Returns
    -------
    rvalue : float
        If no exception or warning is raised on exit, `rvalue` contains
        the real value associated with the option in `optstr`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            Initialization function ``bnd_mcs_init`` has not been
            called.

        (`errno` 2)
            The supplied option is invalid. A keyword or keyword
            combination was not recognized.
    """

    _return_scalar = _b_glopt.bnd_mcs_optget_real(
        optstr,
        comm,
    )
    _rvalue_py = _return_scalar
    return _rvalue_py

def bnd_pso(bl, bu, objfun, comm, npar=None, monmod=None, data=None, io_manager=None, spiked_sorder='C'):
    u"""
    Global optimization using particle swarm algorithm (PSO), bound
    constraints only.

    Note: this function uses optional algorithmic parameters, see also:
    ``optset``.

    ``bnd_pso`` is designed to search for the global minimum or maximum
    of an arbitrary function, using Particle Swarm Optimization (PSO).
    Derivatives are not required, although these may be used by an
    accompanying local minimization function if desired. ``bnd_pso`` is
    essentially identical to ``nlp_pso``, but with a simpler interface
    and with various options removed; otherwise most arguments are
    identical.
    In particular, ``bnd_pso`` does not handle general constraints.

    For full information please refer to the NAG Library document for
    e05sa

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e05/e05saf.html

    Parameters
    ----------
    bl : float, array-like, shape (ndim)
        `bl` is l, the array of lower bounds.

    bu : float, array-like, shape (ndim)
        `bu` is u, the array of upper bounds.

    objfun : callable (objf, vecout) = objfun(mode, x, objf, vecout,
        nstate, data=None)
        `objfun` must, depending on the value of `mode`, calculate the
        objective function `and/or` calculate the gradient of the
        objective function for a ndim-variable vector x.

        Gradients are only required if a local minimizer has been chosen
        which requires gradients.

        See the option 'Local Minimizer' for more information.

        Parameters
        ~~~~~~~~~~
        mode : int
            Indicates which functionality is required.

            `mode` = 0

                F(x) should be returned in `objf`. The value of `objf`
                on entry may be used as an upper bound for the
                calculation. Any expected value of F(x) that is greater
                than `objf` may be approximated by this upper bound;
                that is `objf` can remain unaltered.

            `mode` = 1

                'Local Minimizer' = 'opt.nlp1_solve' only

                First derivatives can be evaluated and returned in
                `vecout`. Any unaltered elements of `vecout` will be
                approximated using finite differences.

            `mode` = 2

                'Local Minimizer' = 'opt.nlp1_solve' only

                F(x) `must` be calculated and returned in `objf`, and
                available first derivatives can be evaluated and
                returned in `vecout`. Any unaltered elements of `vecout`
                will be approximated using finite differences.

            `mode` = 5

                F(x) `must` be calculated and returned in `objf`. The
                value of `objf` on entry may not be used as an upper
                bound.

            `mode` = 6

                'Local Minimizer' = 'opt.uncon_conjgrd_comp' or
                'opt.bounds_mod_deriv_easy' only

                `All` first derivatives `must` be evaluated and returned
                in `vecout`.

            `mode` = 7

                'Local Minimizer' = 'opt.uncon_conjgrd_comp' or
                'opt.bounds_mod_deriv_easy' only

                F(x) `must` be calculated and returned in `objf`, and
                `all` first derivatives `must` be evaluated and returned
                in `vecout`.

        x : float, ndarray, shape (ndim)
            x, the point at which the objective function and/or its
            gradient are to be evaluated.

        objf : float
            The value of `objf` passed to `objfun` varies with the
            argument `mode`.

            `mode` = 0

                `objf` is an upper bound for the value of F(x), often
                equal to the best value of F(x) found so far by a given
                particle. Only objective function values less than the
                value of `objf` on entry will be used further. As such
                this upper bound may be used to stop further evaluation
                when this will only increase the objective function
                value above the upper bound.

            `mode` = 1, 2, 5, 6 or 7

                `objf` is meaningless on entry.

        vecout : float, ndarray, shape (ndim)
            If 'Local Minimizer' = 'opt.nlp1_solve', the values of
            `vecout` are used internally to indicate whether a finite
            difference approximation is required. See
            ``opt.nlp1_solve``.

        nstate : int
            `nstate` indicates various stages of initialization
            throughout the function. This allows for permanent global
            arguments to be initialized the least number of times. For
            example, you may initialize a random number generator seed.

            `nstate` = 2

                `objfun` is called for the very first time. You may save
                computational time if certain data must be read or
                calculated only once.

            `nstate` = 1

                `objfun` is called for the first time by a NAG local
                minimization function. You may save computational time
                if certain data required for the local minimizer need
                only be calculated at the initial point of the local
                minimization.

            `nstate` = 0

                Used in all other cases.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        objf : float
            The value of `objf` returned varies with the argument
            `mode`.

            `mode` = 0

                `objf` must be the value of F(x). Only values of F(x)
                strictly less than `objf` on entry need be accurate.

            `mode` = 1 or 6

                Need not be set.

            `mode` = 2, 5 or 7

                F(x) must be calculated and returned in `objf`. The
                entry value of `objf` may not be used as an upper bound.

        vecout : float, array-like, shape (ndim)
            The required values of `vecout` returned to the calling
            function depend on the value of `mode`.

            `mode` = 0 or 5

                The value of `vecout` need not be set.

            `mode` = 1 or 2

                `vecout` can contain components of the gradient of the
                objective function dF/dx_i for some i = 1,2,...`ndim`,
                or acceptable approximations. Any unaltered elements of
                `vecout` will be approximated using finite differences.

            `mode` = 6 or 7

                `vecout` must contain the gradient of the objective
                function dF/dx_i for all i = 1,2,...`ndim`.
                Approximation of the gradient is strongly discouraged,
                and no finite difference approximations will be
                performed internally (see ``opt.uncon_conjgrd_comp`` and
                ``opt.bounds_mod_deriv_easy``).

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``optset``.

    npar : None or int, optional
        Note: if this argument is None then a default value will be
        used, determined as follows: 10*ndim.

        npar, the number of particles to be used in the swarm. Assuming
        all particles remain within bounds, each complete iteration will
        perform at least `npar` function evaluations. Otherwise,
        significantly fewer objective function evaluations may be
        performed.

    monmod : None or callable x = monmod(x, xb, fb, xbest, fbest, itt,
        data=None), optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        A user-specified monitoring and modification function. `monmod`
        is called once every complete iteration after a finalization
        check.

        It may be used to modify the particle locations that will be
        evaluated at the next iteration.

        This permits the incorporation of algorithmic modifications such
        as including additional advection heuristics and genetic
        mutations. `monmod` is only called during the main loop of the
        algorithm, and as such will be unaware of any further
        improvement from the final local minimization.

        If no monitoring and/or modification is required, `monmod` may
        be None.

        Parameters
        ~~~~~~~~~~
        x : float, ndarray, shape (ndim, npar)
            Note: the ith component of the jth particle, x_j(i), is
            stored in `x`[i-1,j-1].

            The `npar` particle locations, x_j, which will currently be
            used during the next iteration unless altered in `monmod`.

        xb : float, ndarray, shape (ndim)
            The location, x~, of the best solution yet found.

        fb : float
            The objective value, f~ = F(x~), of the best solution yet
            found.

        xbest : float, ndarray, shape (ndim, npar)
            Note: the ith component of the position of the jth
            particle's cognitive memory, x^_j(i), is stored in
            `xbest`[i-1,j-1].

            The locations currently in the cognitive memory, x^_j, for j
            = 1,2,...,`npar` (see [Algorithmic Details]).

        fbest : float, ndarray, shape (npar)
            The objective values currently in the cognitive memory,
            F(x^_j), for j = 1,2,...,`npar`.

        itt : int, ndarray, shape (6)
            Iteration and function evaluation counters (see description
            of `itt` below).

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        x : float, array-like, shape (ndim, npar)
            The particle locations to be used during the next iteration.

    data : arbitrary, optional
        User-communication data for callback functions.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    spiked_sorder : str, optional
        If `x` in `monmod` is spiked (i.e., has unit extent in all but
        one dimension, or has size 1), `spiked_sorder` selects the
        storage order to associate with it in the NAG Engine:

        spiked_sorder = 'C'
          row-major storage will be used;

        spiked_sorder = 'F'
          column-major storage will be used.

    Returns
    -------
    xb : float, ndarray, shape (ndim)
        The location of the best solution found, x~, in R^ndim.

    fb : float
        The objective value of the best solution, f~ = F(x~).

    itt : int, ndarray, shape (6)
        Integer iteration counters for ``bnd_pso``.

        `itt`[0]

            Number of complete iterations.

        `itt`[1]

            Number of complete iterations without improvement to the
            current optimum.

        `itt`[2]

            Number of particles converged to the current optimum.

        `itt`[3]

            Number of improvements to the optimum.

        `itt`[4]

            Number of function evaluations performed.

        `itt`[5]

            Number of particles reset.

    inform : int
        Indicates which finalization criterion was reached. The possible
        values of `inform` are:

        [table omitted]

    Other Parameters
    ----------------
    'Advance Cognitive' : float
        Default = 2.0

        The cognitive advance coefficient, C_s.
        When larger than the global advance coefficient, this will cause
        particles to be attracted toward their previous best positions.
        Setting r = 0.0 will cause ``bnd_pso`` to act predominantly as a
        local optimizer.
        Setting r > 2.0 may cause the swarm to diverge, and is generally
        inadvisable.
        At least one of the global and cognitive coefficients must be
        nonzero.

    'Advance Global' : float
        Default = 2.0

        The global advance coefficient, C_g.
        When larger than the cognitive coefficient this will encourage
        convergence toward the best solution yet found.
        Values r in (0,1) will inhibit particles overshooting the
        optimum.
        Values r in [1,2) cause particles to fly over the optimum some
        of the time.
        Larger values can prohibit convergence.
        Setting r = 0.0 will remove any attraction to the current
        optimum, effectively generating a Monte Carlo multi-start
        optimization algorithm.
        At least one of the global and cognitive coefficients must be
        nonzero.

    'Boundary' : str
        Default = 'FLOATING'

        Determines the behaviour if particles leave the domain described
        by the box bounds.
        This only affects the general PSO algorithm, and will not pass
        down to any NAG local minimizers chosen.

        This option is only effective in those dimensions for which
        `bl`[i-1] != `bu`[i-1], i = 1,2,...,ndim.

        IGNORE
            The box bounds are ignored. The objective function is still
            evaluated at the new particle position.

        RESET
            The particle is re-initialized inside the domain. x^_j and
            f^_j are not affected.

        FLOATING
            The particle position remains the same, however the
            objective function will not be evaluated at the next
            iteration. The particle will probably be advected back into
            the domain at the next advance due to attraction by the
            cognitive and global memory.

        HYPERSPHERICAL
            The box bounds are wrapped around an ndim-dimensional
            hypersphere. As such a particle leaving through a lower
            bound will immediately re-enter through the corresponding
            upper bound and vice versa. The standard distance between
            particles is also modified accordingly.

        FIXED
            The particle rests on the boundary, with the corresponding
            dimensional velocity set to 0.0.

    'Distance Scaling' : str
        Default = 'ON'

        Determines whether distances should be scaled by box widths.

        ON
            When a distance is calculated between x and y, a scaled L^2
            norm is used.

                L^2(x,y) = (sum_{i|u_i != l_i,i <= ndim}((x_i-y_i)/(u_i
                - l_i))^2)^(1/2).

        OFF
            Distances are calculated as the standard L^2 norm without
            any rescaling.

                L^2(x,y) = (sum_i = 1^ndim(x_i-y_i)^2)^(1/2).

    'Distance Tolerance' : float
        Default = 10^-4

        This is the distance, dtol between particles and the global
        optimum which must be reached for the particle to be considered
        converged, i.e., that any subsequent movement of such a particle
        cannot significantly alter the global optimum.
        Once achieved the particle is reset into the box bounds to
        continue searching.

    'Function Precision' : float
        Default = epsilon^0.9

        The argument defines epsilon_r, which is intended to be a
        measure of the accuracy with which the problem function F(x) can
        be computed.
        If r < epsilon or r >= 1, the default value is used.

        The value of epsilon_r should reflect the relative precision of
        1+|F(x)|; i.e., epsilon_r acts as a relative precision when |F|
        is large, and as an absolute precision when |F| is small.
        For example, if F(x) is typically of order 1000 and the first
        six significant digits are known to be correct, an appropriate
        value for epsilon_r would be 10^(-6).
        In contrast, if F(x) is typically of order 10^(-4) and the first
        six significant digits are known to be correct, an appropriate
        value for epsilon_r would be 10^(-10).
        The choice of epsilon_r can be quite complicated for badly
        scaled problems; see Module 8 of Gill `et al.` (1981) for a
        discussion of scaling techniques.
        The default value is appropriate for most simple functions that
        are computed with full accuracy.
        However when the accuracy of the computed function values is
        known to be significantly worse than full precision, the value
        of epsilon_r should be large enough so that no attempt will be
        made to distinguish between function values that differ by less
        than the error inherent in the calculation.

    'Local Boundary Restriction' : float
        Default = 0.5

        Contracts the box boundaries used by a box constrained local
        minimizer to, [beta_l,beta_u], containing the start point x,
        where

            d_i = r * (u_i - l_i) ; beta_l^i = max(l_i, x_i-(d_i)/2) ;
            beta_u^i = min(u_i, x_i+(d_i)/2), i = 1,...,ndim.

        Smaller values of r thereby restrict the size of the domain
        exposed to the local minimizer, possibly reducing the amount of
        work done by the local minimizer.

    'Local Interior Iterations' : int
        The maximum number of iterations or function evaluations the
        chosen local minimizer will perform inside (outside) the main
        loop if applicable.
        For the NAG minimizers these correspond to:

        [table omitted]

        Unless set, these are functions of the arguments passed to
        ``bnd_pso``.

        Setting i = 0 will disable the local minimizer in the
        corresponding algorithmic region.
        For example, setting 'Local Interior Iterations' = 0 and 'Local
        Exterior Iterations' = 30 will cause the algorithm to perform no
        local minimizations inside the main loop of the algorithm, and a
        local minimization with upto 30 iterations after the main loop
        has been exited.

        Note: currently ``opt.bounds_quasi_func_easy`` or
        ``opt.bounds_mod_deriv_easy`` are restricted to using 400*ndim
        and 50*ndim as function evaluation limits respectively. This
        applies to both local minimizations inside and outside the main
        loop. They may still be deactivated in either phase by setting i
        = 0, and may subsequently be reactivated in either phase by
        setting i > 0.

    'Local Interior Major Iterations' : int
        The maximum number of iterations or function evaluations the
        chosen local minimizer will perform inside (outside) the main
        loop if applicable.
        For the NAG minimizers these correspond to:

        [table omitted]

        Unless set, these are functions of the arguments passed to
        ``bnd_pso``.

        Setting i = 0 will disable the local minimizer in the
        corresponding algorithmic region.
        For example, setting 'Local Interior Iterations' = 0 and 'Local
        Exterior Iterations' = 30 will cause the algorithm to perform no
        local minimizations inside the main loop of the algorithm, and a
        local minimization with upto 30 iterations after the main loop
        has been exited.

        Note: currently ``opt.bounds_quasi_func_easy`` or
        ``opt.bounds_mod_deriv_easy`` are restricted to using 400*ndim
        and 50*ndim as function evaluation limits respectively. This
        applies to both local minimizations inside and outside the main
        loop. They may still be deactivated in either phase by setting i
        = 0, and may subsequently be reactivated in either phase by
        setting i > 0.

    'Local Exterior Iterations' : int
        The maximum number of iterations or function evaluations the
        chosen local minimizer will perform inside (outside) the main
        loop if applicable.
        For the NAG minimizers these correspond to:

        [table omitted]

        Unless set, these are functions of the arguments passed to
        ``bnd_pso``.

        Setting i = 0 will disable the local minimizer in the
        corresponding algorithmic region.
        For example, setting 'Local Interior Iterations' = 0 and 'Local
        Exterior Iterations' = 30 will cause the algorithm to perform no
        local minimizations inside the main loop of the algorithm, and a
        local minimization with upto 30 iterations after the main loop
        has been exited.

        Note: currently ``opt.bounds_quasi_func_easy`` or
        ``opt.bounds_mod_deriv_easy`` are restricted to using 400*ndim
        and 50*ndim as function evaluation limits respectively. This
        applies to both local minimizations inside and outside the main
        loop. They may still be deactivated in either phase by setting i
        = 0, and may subsequently be reactivated in either phase by
        setting i > 0.

    'Local Exterior Major Iterations' : int
        The maximum number of iterations or function evaluations the
        chosen local minimizer will perform inside (outside) the main
        loop if applicable.
        For the NAG minimizers these correspond to:

        [table omitted]

        Unless set, these are functions of the arguments passed to
        ``bnd_pso``.

        Setting i = 0 will disable the local minimizer in the
        corresponding algorithmic region.
        For example, setting 'Local Interior Iterations' = 0 and 'Local
        Exterior Iterations' = 30 will cause the algorithm to perform no
        local minimizations inside the main loop of the algorithm, and a
        local minimization with upto 30 iterations after the main loop
        has been exited.

        Note: currently ``opt.bounds_quasi_func_easy`` or
        ``opt.bounds_mod_deriv_easy`` are restricted to using 400*ndim
        and 50*ndim as function evaluation limits respectively. This
        applies to both local minimizations inside and outside the main
        loop. They may still be deactivated in either phase by setting i
        = 0, and may subsequently be reactivated in either phase by
        setting i > 0.

    'Local Interior Tolerance' : float
        Default = 10^-4

        This is the tolerance provided to a local minimizer in the
        interior (exterior) of the main loop of the algorithm.

    'Local Exterior Tolerance' : float
        Default = 10^-4

        This is the tolerance provided to a local minimizer in the
        interior (exterior) of the main loop of the algorithm.

    'Local Interior Minor Iterations' : int
        Where applicable, the secondary number of iterations the chosen
        local minimizer will use inside (outside) the main loop.
        Currently the relevant default values are:

        [table omitted]

    'Local Exterior Minor Iterations' : int
        Where applicable, the secondary number of iterations the chosen
        local minimizer will use inside (outside) the main loop.
        Currently the relevant default values are:

        [table omitted]

    'Local Minimizer' : str
        Default = 'OFF'

        Allows for a choice of submodule ``opt`` functions to be used as
        a coupled, dedicated local minimizer.

        'Local Minimizer' = 'OFF'
            No local minimization will be performed in either the
            INTERIOR or EXTERIOR sections of the algorithm.

        'Local Minimizer' = 'opt.uncon_simplex'
            Use ``opt.uncon_simplex`` as the local minimizer. This does
            not require the calculation of derivatives.

            On a call to `objfun` during a local minimization, `mode` =
            5.

        'Local Minimizer' = 'opt.bounds_mod_deriv_easy'
            Use ``opt.bounds_mod_deriv_easy`` as the local minimizer.
            This requires the calculation of derivatives in `objfun`, as
            indicated by `mode`.

            The box bounds forwarded to this function from ``bnd_pso``
            will have been acted upon by 'Local Boundary Restriction'.
            As such, the domain exposed may be greatly smaller than that
            provided to ``bnd_pso``.

            Accurate derivatives must be provided to this function, and
            will not be approximated internally.
            Each iteration of this local minimizer also requires the
            calculation of both the objective function and its
            derivative.
            Hence on a call to `objfun` during a local minimization,
            `mode` = 7.

        'Local Minimizer' = 'opt.bounds_quasi_func_easy'
            Use ``opt.bounds_quasi_func_easy`` as the local minimizer.
            This does not require the calculation of derivatives.

            On a call to `objfun` during a local minimization, `mode` =
            5.

            The box bounds forwarded to this function from ``bnd_pso``
            will have been acted upon by 'Local Boundary Restriction'.
            As such, the domain exposed may be greatly smaller than that
            provided to ``bnd_pso``.

        'Local Minimizer' = 'opt.uncon_conjgrd_comp'
            Use ``opt.uncon_conjgrd_comp`` as the local minimizer.

            Accurate derivatives must be provided, and will not be
            approximated internally.
            Additionally, each call to `objfun` during a local
            minimization will require either the objective to be
            evaluated alone, or both the objective and its gradient to
            be evaluated.
            Hence on a call to `objfun`, `mode` = 5 or 7.

        'Local Minimizer' = 'opt.nlp1_solve'
            Use ``opt.nlp1_solve`` as the local minimizer. This operates
            such that any derivatives of the objective function that you
            cannot supply, will be approximated internally using finite
            differences.

            Either, the objective, objective gradient, or both may be
            requested during a local minimization, and as such on a call
            to `objfun`, `mode` = 1, 2 or 5.

            The box bounds forwarded to this function from ``bnd_pso``
            will have been acted upon by 'Local Boundary Restriction'.
            As such, the domain exposed may be greatly smaller than that
            provided to ``bnd_pso``.

    'Maximum Function Evaluations' : int
        Default = Imax

        The maximum number of evaluations of the objective function.
        When reached this will return `errno` = 1 and `inform` = 6.

    'Maximum Iterations Completed' : int
        Default = 1000*ndim

        The maximum number of complete iterations that may be performed.
        Once exceeded ``bnd_pso`` will exit with `errno` = 1 and
        `inform` = 5.

        Unless set, this adapts to the parameters passed to ``bnd_pso``.

    'Maximum Iterations Static' : int
        Default = 100

        The maximum number of iterations without any improvement to the
        current global optimum.
        If exceeded ``bnd_pso`` will exit with `errno` = 1 and `inform`
        = 4.
        This exit will be hindered by setting 'Maximum Iterations Static
        Particles' to larger values.

    'Maximum Iterations Static Particles' : int
        Default = 0

        The minimum number of particles that must have converged to the
        current optimum before the function may exit due to 'Maximum
        Iterations Static' with `errno` = 1 and `inform` = 4.

    'Maximum Particles Converged' : int
        Default = Imax

        The maximum number of particles that may converge to the current
        optimum.
        When achieved, ``bnd_pso`` will exit with `errno` = 1 and
        `inform` = 3.
        This exit will be hindered by setting '**Repulsion**' options,
        as these cause the swarm to re-expand.

    'Maximum Particles Reset' : int
        Default = Imax

        The maximum number of particles that may be reset after
        converging to the current optimum.
        Once achieved no further particles will be reset, and any
        particles within 'Distance Tolerance' of the global optimum will
        continue to evolve as normal.

    'Maximum Variable Velocity' : float
        Default = 0.25

        Along any dimension j, the absolute velocity is bounded above by
        |v_j| <= r*(u_j - l_j) = V_max.
        Very low values will greatly increase convergence time.
        There is no upper limit, although larger values will allow more
        particles to be advected out of the box bounds, and values
        greater than 4.0 may cause significant and potentially
        unrecoverable swarm divergence.

    'Optimize' : str
        Default = MINIMIZE

        Determines whether to maximize or minimize the objective
        function.

        MINIMIZE
            The objective function will be minimized.

        MAXIMIZE
            The objective function will be maximized. This is
            accomplished by minimizing the negative of the objective.

    'Repeatability' : str
        Default = 'OFF'

        Allows for the same random number generator seed to be used for
        every call to ``bnd_pso``. 'Repeatability' = 'OFF' is
        recommended in general.

        OFF
            The internal generation of random numbers will be
            nonrepeatable.

        ON
            The same seed will be used.

    'Repulsion Finalize' : int
        Default = Imax

        The number of iterations performed in a repulsive phase before
        re-contraction.
        This allows a re-diversified swarm to contract back toward the
        current optimum, allowing for a finer search of the near optimum
        space.

    'Repulsion Initialize' : int
        Default = Imax

        The number of iterations without any improvement to the global
        optimum before the algorithm begins a repulsive phase.
        This phase allows the particle swarm to re-expand away from the
        current optimum, allowing more of the domain to be investigated.
        The repulsive phase is automatically ended if a superior optimum
        is found.

    'Repulsion Particles' : int
        Default = 0

        The number of particles required to have converged to the
        current optimum before any repulsive phase may be initialized.
        This will prevent repulsion before a satisfactory search of the
        near optimum area has been performed, which may happen for large
        dimensional problems.

    'Seed' : int
        Default = 0

        Sets the random number generator seed to be used when
        'Repeatability' = 'ON'.
        If set to 0, the default seed will be used.
        If not, the absolute value of 'Seed' will be used to generate
        the random number generator seed.

    'Swarm Standard Deviation' : float
        Default = 0.1

        The target standard deviation of the particle distances from the
        current optimum. Once the standard deviation is below this
        level, ``bnd_pso`` will exit with `errno` = 1 and `inform` = 2.
        This criterion will be penalized by the use of '**Repulsion**'
        options, as these cause the swarm to re-expand, increasing the
        standard deviation of the particle distances from the best
        point.

    'Target Objective' : str
        Default = 'OFF'

        Activate or deactivate the use of a target value as a
        finalization criterion.
        If active, then once the supplied target value for the objective
        function is found (beyond the first iteration if 'Target
        Warning' is active) ``bnd_pso`` will exit with no exception or
        warning is raised and `inform` = 1.
        Other than checking for feasibility only ('Optimize' =
        CONSTRAINTS), this is the only finalization criterion that
        guarantees that the algorithm has been successful.
        If the target value was achieved at the initialization phase or
        first iteration and 'Target Warning' is active, ``bnd_pso`` will
        exit with `errno` = 2.
        This option may take any real value r, or the character ON/OFF
        as well as DEFAULT.
        If this option is queried using ``optget``, the current value of
        r will be returned in rvalue, and cvalue will indicate whether
        this option is ON or OFF.
        The behaviour of the option is as follows:

        r
            Once a point is found with an objective value within the
            'Target Objective Tolerance' of r, ``bnd_pso`` will exit
            successfully with no exception or warning is raised and
            `inform` = 1.

        OFF
            The current value of r will remain stored, however it will
            not be used as a finalization criterion.

        ON
            The current value of r stored will be used as a finalization
            criterion.

        DEFAULT
            The stored value of r will be reset to its default value
            (0.0), and this finalization criterion will be deactivated.

    'Target Objective Value' : float
        Default = 0.0

        Activate or deactivate the use of a target value as a
        finalization criterion.
        If active, then once the supplied target value for the objective
        function is found (beyond the first iteration if 'Target
        Warning' is active) ``bnd_pso`` will exit with no exception or
        warning is raised and `inform` = 1.
        Other than checking for feasibility only ('Optimize' =
        CONSTRAINTS), this is the only finalization criterion that
        guarantees that the algorithm has been successful.
        If the target value was achieved at the initialization phase or
        first iteration and 'Target Warning' is active, ``bnd_pso`` will
        exit with `errno` = 2.
        This option may take any real value r, or the character ON/OFF
        as well as DEFAULT.
        If this option is queried using ``optget``, the current value of
        r will be returned in rvalue, and cvalue will indicate whether
        this option is ON or OFF.
        The behaviour of the option is as follows:

        r
            Once a point is found with an objective value within the
            'Target Objective Tolerance' of r, ``bnd_pso`` will exit
            successfully with no exception or warning is raised and
            `inform` = 1.

        OFF
            The current value of r will remain stored, however it will
            not be used as a finalization criterion.

        ON
            The current value of r stored will be used as a finalization
            criterion.

        DEFAULT
            The stored value of r will be reset to its default value
            (0.0), and this finalization criterion will be deactivated.

    'Target Objective Safeguard' : float
        Default = 100.0 epsilon

        If you have given a target objective value to reach in objval
        (the value of the option 'Target Objective Value'), objsfg sets
        your desired safeguarded termination tolerance, for when objval
        is close to zero.

    'Target Objective Tolerance' : float
        Default = 0.0

        The optional tolerance to a user-specified target value.

    'Target Warning' : str
        Default = 'OFF'

        Activates or deactivates the error exit associated with the
        target value being achieved before entry into the main loop of
        the algorithm, `errno` = 2.

        OFF
            No error will be returned, and the function will exit
            normally.

        ON
            An error will be returned if the target objective is reached
            prematurely, and the function will exit with `errno` = 2.

    'Verify Gradients' : str
        Default = 'ON'

        Adjusts the level of gradient checking performed when gradients
        are required.
        Gradient checks are only performed on the first call to the
        chosen local minimizer if it requires gradients.
        There is no guarantee that the gradient check will be correct,
        as the finite differences used in the gradient check are
        themselves subject to inaccuracies.

        OFF
            No gradient checking will be performed.

        ON
            A cheap gradient check will be performed on both the
            gradients corresponding to the objective through `objfun`.

        OBJECTIVE FULL
            A more expensive gradient check will be performed on the
            gradients corresponding to the objective `objfun`.

    'Weight Decrease' : str
        Default = INTEREST

        Determines how particle weights decrease.

        OFF
            Weights do not decrease.

        INTEREST
            Weights decrease through compound interest as w_{IT+1} =
            w_IT(1-W_val), where W_val is the 'Weight Value' and IT is
            the current number of iterations.

        LINEAR
            Weights decrease linearly following w_{IT+1} =
            w_IT-IT*(W_max-W_min)/IT_max, where IT is the iteration
            number and IT_max is the maximum number of iterations as set
            by 'Maximum Iterations Completed'.

    'Weight Initial' : float
        Default = W_max

        The initial value of any particle's inertial weight, W_ini, or
        the minimum possible initial value if initial weights are
        randomized.
        When set, this will override 'Weight Initialize' = 'RANDOMIZED'
        or 'MAXIMUM', and as such these must be set afterwards if so
        desired.

    'Weight Initialize' : str
        Default = MAXIMUM

        Determines how the initial weights are distributed.

        INITIAL
            All weights are initialized at the initial weight, W_ini, if
            set. If 'Weight Initial' has not been set, this will be the
            maximum weight, W_max.

        MAXIMUM
            All weights are initialized at the maximum weight, W_max.

        RANDOMIZED
            Weights are uniformly distributed in (W_min,W_max) or
            (W_ini,W_max) if 'Weight Initial' has been set.

    'Weight Maximum' : float
        Default = 1.0

        The maximum particle weight, W_max.

    'Weight Minimum' : float
        Default = 0.1

        The minimum achievable weight of any particle, W_min.
        Once achieved, no further weight reduction is possible.

    'Weight Reset' : str
        Default = MAXIMUM

        Determines how particle weights are re-initialized.

        INITIAL
            Weights are re-initialized at the initial weight if set. If
            'Weight Initial' has not been set, this will be the maximum
            weight.

        MAXIMUM
            Weights are re-initialized at the maximum weight.

        RANDOMIZED
            Weights are uniformly distributed in (W_min,W_max) or
            (W_ini,W_max) if 'Weight Initial' has been set.

    'Weight Value' : float
        Default = 0.01

        The constant W_val used with 'Weight Decrease' = 'INTEREST'.

    'SMP Callback Thread Safe' : str
        Default = WARNING

        Declare that the callback functions you provide are or are not
        thread safe.
        In particular, this indicates that access to the shared memory
        arrays `iuser` and `ruser` from within your provided callbacks
        is done in a thread safe manner. If these arrays are just used
        to pass constant data, then you may assume they are thread safe.
        If these are also used for workspace, or passing variable data
        such as random number generator seeds, then you must ensure
        these are accessed and updated safely. Whilst this can be done
        using OpenMP critical sections, we suggest their use is
        minimized to prevent unnecessary bottlenecks, and that instead
        individual threads have access to independent subsections of the
        provided arrays where possible.

        YES
            The callback functions have been programmed in a thread safe
            way. The algorithm will use ``OMP_NUM_THREADS`` threads.

        NO
            The callback functions are not thread safe. Setting this
            option will force the algorithm to run on a single thread
            only, and is advisable only for debugging purposes, or if
            you wish to parallelize your callback functions.

        WARNING
            This will cause an immediate exit from ``bnd_pso`` with
            `errno` = 51 if multiple threads are detected. This is to
            inform you that you have not declared the callback functions
            either to be thread safe, or that they are thread unsafe and
            you wish the algorithm to run in serial.

    'SMP Local Minimizer External' : str
        Default = ALL

        Determines how many threads will attempt to locally minimize the
        best found solution after the function has exited the main loop.

        MASTER
            Only the master thread will attempt to find any improvement.
            The local minimization will be launched from the best known
            solution. All other threads will remain effectively idle.

        ALL
            The master thread will perform a local minimization from the
            best known solution, while all other threads will perform a
            local minimization from randomly generated perturbations of
            the best known solution, increasing the chance of an
            improvement. Assuming all local minimizations will take
            approximately the same amount of computation, this will be
            effectively free in terms of real time. It will however
            increase the number of function evaluations performed.

    'SMP Monitor' : str
        Default = SINGLE

        Determines whether the user-supplied function `monmod` is
        invoked once every sub-iteration each thread performs, or only
        once by a single thread after all threads have completed at
        least one sub-iteration.

        SINGLE
            Only one thread will invoke `monmod`, after all threads have
            performed at least one sub-iteration.

        ALL
            Each thread will invoke `monmod` each time it completes a
            sub-iteration. If you wish to alter `x` using `monmod` you
            should use this option, as `monmod` will only receive the
            arrays `x`, `xbest` and `fbest` private to the calling
            thread.

    'SMP Monmod' : str
        Default = SINGLE

        Determines whether the user-supplied function `monmod` is
        invoked once every sub-iteration each thread performs, or only
        once by a single thread after all threads have completed at
        least one sub-iteration.

        SINGLE
            Only one thread will invoke `monmod`, after all threads have
            performed at least one sub-iteration.

        ALL
            Each thread will invoke `monmod` each time it completes a
            sub-iteration. If you wish to alter `x` using `monmod` you
            should use this option, as `monmod` will only receive the
            arrays `x`, `xbest` and `fbest` private to the calling
            thread.

    'SMP Subswarm' : int
        Default = 1

        Determines how many threads support a particle subswarm.
        This is an extra collection of particles constrained to search
        only within a hypercube of edge length 10.0*'Distance Tolerance'
        of the best point known to an individual thread.
        This may improve the number of iterations required to find a
        provided target, particularly if no local minimizer is in use.

        If i <= 0, then this will be disabled on all the threads.

        If i >= OMP_NUM_THREADS, then all the threads will support a
        particle subswarm.

    'SMP Thread Overrun' : int
        Default = Imax

        This option provides control over the level of asynchronicity
        present in a simulation.
        In particular, a barrier synchronization between all threads is
        performed if any thread completes i sub-iterations more than the
        slowest thread, causing all threads to be exposed to the current
        best solution.
        Allowing asynchronous behaviour does however allow individual
        threads to focus on different global optimum candidates some of
        the time, which can inhibit convergence to unwanted sub-optima.
        It also allows for threads to continue searching when other
        threads are completing sub-iterations at a slower rate.

        If i < 1, the algorithm will force a synchronization between
        threads at the end of each iteration.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, ndim = *<value>*.

            Constraint: ndim >= 1.

        (`errno` 12)
            On entry, `npar` = *<value>*.

            Constraint: `npar` >= 5*num_threads, where **num_threads**
            is the value returned by the OpenMP environment variable
            ``OMP_NUM_THREADS``, or **num_threads** is 1 for a serial
            version of this function.

        (`errno` 14)
            On entry, `bl`[*<value>*] = *<value>* and `bu`[*<value>*] =
            *<value>*.

            Constraint: `bu`[i] >= `bl`[i] for all i.

        (`errno` 14)
            On entry, `bl`[i] = `bu`[i] for all i.

            Constraint: `bu`[i] > `bl`[i] for at least one i.

        (`errno` 19)
            Error *<value>* occurred whilst adjusting to interior local
            minimizer options.

        (`errno` 19)
            Error *<value>* occurred whilst adjusting to exterior local
            minimizer options.

        (`errno` 21)
            Either the option arrays have not been initialized for
            ``bnd_pso``, or they have become corrupted.

        (`errno` 51)
            Multiple SMP threads have been detected; however, the option
            'SMP Callback Thread Safe' has not been set.

            Set 'SMP Callback Thread Safe' = 'YES' if the provided
            callbacks are thread safe.

            Set 'SMP Callback Thread Safe' = 'NO' if the provided
            callbacks are not thread safe, to force serial execution.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            A finalization criterion was reached that cannot guarantee
            success.

            On exit, `inform` = *<value>*.

        (`errno` 2)
            Target *<value>* achieved after the first iteration.

            `fb` = *<value>*.

        (`errno` 32)
            Derivative checks indicate possible errors in the supplied
            derivatives.

    NagCallbackTerminateWarning
        (`errno` 3)
            User requested exit *<value>* during call to `objfun`.

        (`errno` 3)
            User requested exit *<value>* during call to `monmod`.

    References
    ----------
    Gill, P E, Murray, W and Wright, M H, 1981, `Practical
    Optimization`, Academic Press

    Kennedy, J and Eberhart, R C, 1995, `Particle Swarm Optimization`,
    Proceedings of the 1995 IEEE International Conference on Neural
    Networks, 1942--1948

    Koh, B, George, A D, Haftka, R T and Fregly, B J, 2006, `Parallel
    Asynchronous Particle Swarm Optimization`, International Journal for
    Numerical Methods in Engineering (67(4)), 578--595

    Vaz, A I and Vicente, L N, 2007, `A Particle Swarm Pattern Search
    Method for Bound Constrained Global Optimization`, Journal of Global
    Optimization (39(2)), 197--219, Kluwer Academic Publishers
    """

    fun_name = 'naginterfaces.library.glopt.bnd_pso'

    cb_objfun_name = fun_name + ':objfun'

    def cb_objfun(mode, ndim, x, objf, vecout, nstate, data=None):
        """Wrapping user's objfun. Called as base objfun."""
        if data is not None:
            _objf_objfun, _vecout_objfun = objfun(
                mode,
                x,
                objf,
                vecout,
                nstate,
                data,
            )
        else:
            _objf_objfun, _vecout_objfun = objfun(
                mode,
                x,
                objf,
                vecout,
                nstate,
            )
        _objf_objfun_base = _objf_objfun
        vecout[:] = _utils._to_ndarray(
            {'fun_name': cb_objfun_name, 'returned': 'vecout'},
            _vecout_objfun,
            copy=False,
            exp_shape=(ndim,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        return _objf_objfun_base

    cb_monmod_name = fun_name + ':monmod'

    def cb_monmod(ndim, npar, x, xb, fb, xbest, fbest, itt, data=None):
        """Wrapping user's monmod. Called as base monmod."""
        if data is not None:
            _x_monmod = monmod(
                x,
                xb,
                fb,
                xbest,
                fbest,
                itt,
                data,
            )
        else:
            _x_monmod = monmod(
                x,
                xb,
                fb,
                xbest,
                fbest,
                itt,
            )
        _x_monmod_np = _utils._to_ndarray(
            {'fun_name': cb_monmod_name, 'returned': 'x'},
            _x_monmod,
            copy=False,
            exp_shape=(ndim, npar),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _utils._check_callback_sorder(
            {'fun_name': cb_monmod_name},
            ((_x_monmod_np, 'x'),),
            wrapper_sorder=_m_sorder,
            spiked_sorder=spiked_sorder,
        )
        x[:, :] = _x_monmod_np

    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        (),
        spiked_sorder=spiked_sorder,
    )
    _bl = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'bl'},
        bl,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ndim_py = _b_utils._EngineFloat64ArrayType.get_shape(_bl, {'fun_name': fun_name, 'entity': 'bl'}, exp_rank=1)[0]
    _ndim = _b_utils._EngineIntScalarType(
        _ndim_py, {'fun_name': fun_name, 'entity': '_ndim'},
    ).to_ctypes()
    _npar = 10*_ndim_py if npar is None else npar
    _xb = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ndim_py,),
    )
    _bu = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'bu'},
        bu,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b_utils._check_callable(objfun, {'fun_name': fun_name, 'entity': 'objfun'})
    _objfun = cb_objfun
    _b_utils._check_callable(monmod, {'fun_name': fun_name, 'entity': 'monmod'}, allow_none=True)
    _monmod = None if monmod is None else cb_monmod
    _itt = _b_utils._EngineIntArrayType.empty_ndarray(
        (6,),
    )
    try:
        _return_scalars = _b_glopt.bnd_pso(
            _ndim,
            _npar,
            _xb,
            _bl,
            _bu,
            _objfun,
            _monmod,
            comm,
            _itt,
            data,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _xb_py = _xb
    _fb_py = _return_scalars.fb
    _itt_py = _itt
    _inform_py = _return_scalars.inform
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['xb', 'fb', 'itt', 'inform'],
    )(
        _xb_py,
        _fb_py,
        _itt_py,
        _inform_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def nlp_pso(bl, bu, xbest, fbest, cbest, objfun, comm, confun=None, monmod=None, data=None, io_manager=None, spiked_sorder='C'):
    u"""
    Global optimization using particle swarm algorithm (PSO),
    comprehensive.

    Note: this function uses optional algorithmic parameters, see also:
    ``optset``.

    ``nlp_pso`` is designed to search for the global minimum or maximum
    of an arbitrary function, subject to general nonlinear constraints,
    using Particle Swarm Optimization (PSO).
    Derivatives are not required, although these may be used by an
    accompanying local minimization function if desired. ``nlp_pso`` is
    essentially identical to ``bnd_pso``, with an expert interface and
    various additional arguments added; otherwise most arguments are
    identical.
    In particular, ``bnd_pso`` does not handle general constraints.

    For full information please refer to the NAG Library document for
    e05sb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e05/e05sbf.html

    Parameters
    ----------
    bl : float, array-like, shape (ndim+ncon)
        `bl` is l, the array of lower bounds.

    bu : float, array-like, shape (ndim+ncon)
        `bu` is u, the array of upper bounds.

    xbest : float, array-like, shape (ndim, npar)
        Note: the ith component of the best position of the jth
        particle, x^_j(i), is stored in `xbest`[i-1,j-1].

        If using 'Start' = 'WARM', the initial particle positions,
        x^_j^0.

    fbest : float, array-like, shape (npar)
        If using 'Start' = 'WARM', objective function values, f^_j^0 =
        F(x^_j^0), corresponding to the npar particle locations stored
        in `xbest`.

    cbest : float, array-like, shape (ncon, npar)
        Note: the kth constraint violation of the jth particle is stored
        in `cbest`[k-1,j-1].

        If using 'Start' = 'WARM', the initial constraint violations,
        e^_j^0 = e(x^_j^0), corresponding to the npar particle
        locations.

    objfun : callable (objf, vecout) = objfun(mode, x, objf, vecout,
        nstate, data=None)
        `objfun` must, depending on the value of `mode`, calculate the
        objective function `and/or` calculate the gradient of the
        objective function for a ndim-variable vector x.

        Gradients are only required if a local minimizer has been chosen
        which requires gradients.

        See the option 'Local Minimizer' for more information.

        Parameters
        ~~~~~~~~~~
        mode : int
            Indicates which functionality is required.

            `mode` = 0

                F(x) should be returned in `objf`. The value of `objf`
                on entry may be used as an upper bound for the
                calculation. Any expected value of F(x) that is greater
                than `objf` may be approximated by this upper bound;
                that is `objf` can remain unaltered.

            `mode` = 1

                'Local Minimizer' = 'opt.nlp1_solve' only

                First derivatives can be evaluated and returned in
                `vecout`. Any unaltered elements of `vecout` will be
                approximated using finite differences.

            `mode` = 2

                'Local Minimizer' = 'opt.nlp1_solve' only

                F(x) `must` be calculated and returned in `objf`, and
                available first derivatives can be evaluated and
                returned in `vecout`. Any unaltered elements of `vecout`
                will be approximated using finite differences.

            `mode` = 5

                F(x) `must` be calculated and returned in `objf`. The
                value of `objf` on entry may not be used as an upper
                bound.

            `mode` = 6

                'Local Minimizer' = 'opt.uncon_conjgrd_comp' or
                'opt.bounds_mod_deriv_easy' only

                `All` first derivatives `must` be evaluated and returned
                in `vecout`.

            `mode` = 7

                'Local Minimizer' = 'opt.uncon_conjgrd_comp' or
                'opt.bounds_mod_deriv_easy' only

                F(x) `must` be calculated and returned in `objf`, and
                `all` first derivatives `must` be evaluated and returned
                in `vecout`.

        x : float, ndarray, shape (ndim)
            x, the point at which the objective function and/or its
            gradient are to be evaluated.

        objf : float
            The value of `objf` passed to `objfun` varies with the
            argument `mode`.

            `mode` = 0

                `objf` is an upper bound for the value of F(x), often
                equal to the best constraint penalised value of F(x)
                found so far by a given particle if the objective
                function is strictly positive (see [Algorithmic
                Details]). Only objective function values less than the
                value of `objf` on entry will be used further. As such
                this upper bound may be used to stop further evaluation
                when this will only increase the objective function
                value above the upper bound.

            `mode` = 1, 2, 5, 6 or 7

                `objf` is meaningless on entry.

        vecout : float, ndarray, shape (ndim)
            If 'Local Minimizer' = 'opt.nlp1_solve', the values of
            `vecout` are used internally to indicate whether a finite
            difference approximation is required. See
            ``opt.nlp1_solve``.

        nstate : int
            `nstate` indicates various stages of initialization
            throughout the function. This allows for permanent global
            arguments to be initialized the least number of times. For
            example, you may initialize a random number generator seed.

            `nstate` = 2

                `objfun` is called for the very first time. You may save
                computational time if certain data must be read or
                calculated only once.

            `nstate` = 1

                `objfun` is called for the first time by a NAG local
                minimization function. You may save computational time
                if certain data required for the local minimizer need
                only be calculated at the initial point of the local
                minimization.

            `nstate` = 0

                Used in all other cases.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        objf : float
            The value of `objf` returned varies with the argument
            `mode`.

            `mode` = 0

                `objf` must be the value of F(x). Only values of F(x)
                strictly less than `objf` on entry need be accurate.

            `mode` = 1 or 6

                Need not be set.

            `mode` = 2, 5 or 7

                F(x) must be calculated and returned in `objf`. The
                entry value of `objf` may not be used as an upper bound.

        vecout : float, array-like, shape (ndim)
            The required values of `vecout` returned to the calling
            function depend on the value of `mode`.

            `mode` = 0 or 5

                The value of `vecout` need not be set.

            `mode` = 1 or 2

                `vecout` can contain components of the gradient of the
                objective function dF/dx_i for some i = 1,2,...`ndim`,
                or acceptable approximations. Any unaltered elements of
                `vecout` will be approximated using finite differences.

            `mode` = 6 or 7

                `vecout` must contain the gradient of the objective
                function dF/dx_i for all i = 1,2,...`ndim`.
                Approximation of the gradient is strongly discouraged,
                and no finite difference approximations will be
                performed internally (see ``opt.uncon_conjgrd_comp`` and
                ``opt.bounds_mod_deriv_easy``).

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``optset``.

    confun : None or callable (c, cjac) = confun(mode, needc, x, cjac,
        nstate, data=None), optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `confun` must calculate any constraints other than the box
        constraints.

        If no constraints are required, `confun` may be None. For
        information on how a NAG local minimizer will use `confun` see
        the documentation for ``opt.nlp1_solve``.

        Parameters
        ~~~~~~~~~~
        mode : int
            Indicates which values must be assigned during each call of
            `confun`. Only the following values need be assigned, for
            each value of k in {1,...,ncon} such that `needc`[k-1] > 0:

            `mode` = 0

                the constraint values c_k(x).

            `mode` = 1

                rows of the constraint Jacobian, dc_k/dx_i(x), for i =
                1,2,...,ndim.

            `mode` = 2

                the constraint values c_k(x) and the corresponding rows
                of the constraint Jacobian, dc_k/dx_i(x), for i =
                1,2,...,ndim.

        needc : int, ndarray, shape (ncon)
            The indices of the elements of `c` and/or `cjac` that must
            be evaluated by `confun`. If `needc`[k-1] > 0, the k-1th
            element of `c`, corresponding to the values of the kth
            constraint, and/or the available elements of the k-1th row
            of `cjac`, corresponding to the derivatives of the kth
            constraint, must be evaluated at x (see argument `mode`).

        x : float, ndarray, shape (ndim)
            x, the vector of variables at which the constraint functions
            and/or the available elements of the constraint Jacobian are
            to be evaluated.

        cjac : float, ndarray, shape (ncon, ndim)
            Note: the derivative of the kth constraint with respect to
            the ith component, dc_k/dx_i, is stored in `cjac`[k-1,i-1].

            The elements of `cjac` are set to special values which
            enable ``nlp_pso`` to detect whether they are changed by
            `confun`.

        nstate : int
            `nstate` indicates various stages of initialization
            throughout the function. This allows for permanent global
            arguments to be initialized a minimum number of times. For
            example, you may initialize a random number generator seed.
            Note that unless the option 'Optimize' = 'CONSTRAINTS' has
            been set, `objfun` will be called before `confun`.

            `nstate` = 2

                `confun` is called for the very first time. This
                argument setting allows you to save computational time
                if certain data must be read or calculated only once.

            `nstate` = 1

                `confun` is called for the first time during a NAG local
                minimization function. This argument setting allows you
                to save computational time if certain data required for
                the local minimizer need only be calculated at the
                initial point of the local minimization.

            `nstate` = 0

                Used in all other cases.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        c : float, array-like, shape (ncon)
            If `needc`[k-1] > 0 and `mode` = 0 or 2, `c`[k-1] must
            contain the value of c_k(x). The remaining elements of `c`,
            corresponding to the non-positive elements of `needc`, need
            not be set.

        cjac : float, array-like, shape (ncon, ndim)
            If `needc`[k-1] > 0 and `mode` = 1 or 2, the elements of
            `cjac` corresponding to the kth row of the constraint
            Jacobian should contain the available elements of the vector
            nabla c_k given by

                nabla c_k = (dc_k/dx_1,dc_k/dx_2,...,dc_k/dx_n),

            where dc_k/dx_i is the partial derivative of the kth
            constraint with respect to the ith variable, evaluated at
            the point x; elements of `cjac` that remain unaltered will
            be approximated internally using finite differences. The
            remaining rows of `cjac`, corresponding to non-positive
            elements of `needc`, need not be set.

            It must be emphasized that unassigned elements of `cjac` are
            not treated as constant; they are estimated by finite
            differences, at nontrivial expense.

            An interval for each element of x is computed automatically
            at the start of the optimization.

            The automatic procedure can usually identify constant
            elements of `cjac`, which are then computed once only by
            finite differences.

    monmod : None or callable x = monmod(x, xb, fb, cb, xbest, fbest,
        cbest, itt, data=None), optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        A user-specified monitoring and modification function. `monmod`
        is called once every complete iteration after a finalization
        check.

        It may be used to modify the particle locations that will be
        evaluated at the next iteration.

        This permits the incorporation of algorithmic modifications such
        as including additional advection heuristics and genetic
        mutations. `monmod` is only called during the main loop of the
        algorithm, and as such will be unaware of any further
        improvement from the final local minimization.

        If no monitoring and/or modification is required, `monmod` may
        be None.

        Parameters
        ~~~~~~~~~~
        x : float, ndarray, shape (ndim, npar)
            Note: the ith component of the jth particle, x_j(i), is
            stored in `x`[i-1,j-1].

            The `npar` particle locations, x_j, which will currently be
            used during the next iteration unless altered in `monmod`.

        xb : float, ndarray, shape (ndim)
            The location, x~, of the best solution yet found.

        fb : float
            The objective value, f~ = F(x~), of the best solution yet
            found.

        cb : float, ndarray, shape (ncon)
            The constraint violations, e~ = e(x~), of the best solution
            yet found.

        xbest : float, ndarray, shape (ndim, npar)
            Note: the ith component of the position of the jth
            particle's cognitive memory, x^_j(i), is stored in
            `xbest`[i-1,j-1].

            The locations currently in the cognitive memory, x^_j, for j
            = 1,2,...,npar (see [Algorithmic Details]).

        fbest : float, ndarray, shape (npar)
            The objective values currently in the cognitive memory,
            F(x^_j), for j = 1,2,...,npar.

        cbest : float, ndarray, shape (ncon, npar)
            Note: the kth constraint violation of the jth particle's
            cognitive memory is stored in `cbest`[k-1,j-1].

            The constraint violations currently in the cognitive memory,
            e^= e(x^_j), for j = 1,2,...,npar, evaluated at x^_j.

        itt : int, ndarray, shape (7)
            Iteration and function evaluation counters (see description
            of `itt` below).

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        x : float, array-like, shape (ndim, npar)
            The particle locations to be used during the next iteration.

    data : arbitrary, optional
        User-communication data for callback functions.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    spiked_sorder : str, optional
        If `xbest` and `cbest` are spiked (i.e., have unit extent in all
        but one dimension, or have size 1), `spiked_sorder` selects the
        storage order to associate with them in the NAG Engine:

        spiked_sorder = 'C'
          row-major storage will be used;

        spiked_sorder = 'F'
          column-major storage will be used.

        Two-dimensional arrays returned from callback functions in this
        routine must then use the same storage order.

    Returns
    -------
    xb : float, ndarray, shape (ndim)
        The location of the best solution found, x~, in R^ndim.

    fb : float
        The objective value of the best solution, f~ = F(x~).

    cb : float, ndarray, shape (ncon)
        The constraint violations of the best solution found, e~ =
        e(x~). These may have been deemed to be acceptable given the
        tolerance and scaling of the constraints. See [Algorithmic
        Details] and [Other Parameters].

    xbest : float, ndarray, shape (ndim, npar)
        The best positions found, x^_j, by the npar particles in the
        swarm.

    fbest : float, ndarray, shape (npar)
        Objective function values, f^_j = F(x^_j), corresponding to the
        locations returned in `xbest`.

    cbest : float, ndarray, shape (ncon, npar)
        The final constraint violations, e^_j, corresponding to the
        locations returned in `xbest`.

    itt : int, ndarray, shape (7)
        Integer iteration counters for ``nlp_pso``.

        `itt`[0]

            Number of complete iterations.

        `itt`[1]

            Number of complete iterations without improvement to the
            current optimum.

        `itt`[2]

            Number of particles converged to the current optimum.

        `itt`[3]

            Number of improvements to the optimum.

        `itt`[4]

            Number of function evaluations performed.

        `itt`[5]

            Number of particles reset.

        `itt`[6]

            Number of violated constraints at completion. Note this is
            always calculated using the L^1 norm and a nonzero result
            does not necessarily mean that the algorithm did not find a
            suitably constrained point with respect to the single norm
            used.

    inform : int
        Indicates which finalization criterion was reached. The possible
        values of `inform` are:

        [table omitted]

        If you wish to continue from the final position gained from a
        previous simulation with adjusted options, you may set the
        option 'Start' = 'WARM', and pass back in the returned arrays
        `xbest`, `fbest`, and `cbest`. You should either record the
        returned values of `xb`, `fb` and `cb` for comparison, as these
        will not be re-used by the algorithm, or include them in
        `xbest`, `fbest` and `cbest` respectively by overwriting the
        entries corresponding to one particle with the relevant
        information.

    Other Parameters
    ----------------
    'Advance Cognitive' : float
        Default = 2.0

        The cognitive advance coefficient, C_s.
        When larger than the global advance coefficient, this will cause
        particles to be attracted toward their previous best positions.
        Setting r = 0.0 will cause ``nlp_pso`` to act predominantly as a
        local optimizer.
        Setting r > 2.0 may cause the swarm to diverge, and is generally
        inadvisable.
        At least one of the global and cognitive coefficients must be
        nonzero.

    'Advance Global' : float
        Default = 2.0

        The global advance coefficient, C_g.
        When larger than the cognitive coefficient this will encourage
        convergence toward the best solution yet found.
        Values r in (0,1) will inhibit particles overshooting the
        optimum.
        Values r in [1,2) cause particles to fly over the optimum some
        of the time.
        Larger values can prohibit convergence.
        Setting r = 0.0 will remove any attraction to the current
        optimum, effectively generating a Monte Carlo multi-start
        optimization algorithm.
        At least one of the global and cognitive coefficients must be
        nonzero.

    'Boundary' : str
        Default = 'FLOATING'

        Determines the behaviour if particles leave the domain described
        by the box bounds.
        This only affects the general PSO algorithm, and will not pass
        down to any NAG local minimizers chosen.

        This option is only effective in those dimensions for which
        `bl`[i-1] != `bu`[i-1], i = 1,2,...,ndim.

        IGNORE
            The box bounds are ignored. The objective function is still
            evaluated at the new particle position.

        RESET
            The particle is re-initialized inside the domain. x^_j, f^_j
            and e^_j are not affected.

        FLOATING
            The particle position remains the same, however the
            objective function will not be evaluated at the next
            iteration. The particle will probably be advected back into
            the domain at the next advance due to attraction by the
            cognitive and global memory.

        HYPERSPHERICAL
            The box bounds are wrapped around an ndim-dimensional
            hypersphere. As such a particle leaving through a lower
            bound will immediately re-enter through the corresponding
            upper bound and vice versa. The standard distance between
            particles is also modified accordingly.

        FIXED
            The particle rests on the boundary, with the corresponding
            dimensional velocity set to 0.0.

    'Constraint Norm' : str
        Default = L1

        Determines with respect to which norm the constraint residuals
        should be constructed.
        These are automatically scaled with respect to ncon as stated.
        For the set of (scaled) violations e, these may be,

        L1
            The L^1 norm will be used, ||e||_1 = 1/nconsum_1^ncon|e_k|

        L2
            The L^2 norm will be used, ||e||_2 =
            1/nconsqrt(sum_1^ncone_k^2)

        L2SQ
            The square of the L^2 norm will be used, ||e||_2^2 =
            1/nconsum_1^ncone_k^2

        LMAX
            The L^infinity norm will be used, ||e||_infinity = max_{0 <
            k <= ncon}(|e_k|)

    'Constraint Scale Maximum' : float
        Default = 1.0e6

        Internally, each constraint violation is scaled with respect to
        the maximum violation yet achieved for that constraint.
        This option acts as a ceiling for this scale.

    'Constraint Scaling' : str
        Default = INITIAL

        Determines whether to scale the constraints and objective
        function when constructing the penalty function.

        OFF
            Neither the constraint violations nor the objective will be
            scaled automatically. This should only be used if the
            constraints and objective are similarly scaled everywhere
            throughout the domain.

        INITIAL
            The maximum of the initial cognitive memories, f^_j and
            e^_j, will be used to scale the objective function and
            constraint violations respectively.

        ADAPTIVE
            Initially, the maximum of the initial cognitive memories,
            f^_j and e^_j, will be used to scale the objective function
            and constraint violations respectively. If a significant
            change is detected in the behaviour of the constraints or
            the objective, these will be rescaled with respect to the
            current state of the cognitive memory.

    'Constraint Superiority' : float
        Default = 0.01

        The minimum scaled improvement in the constraint violation for a
        location to be immediately superior to that in memory,
        regardless of the objective value.

    'Constraint Tolerance' : float
        Default = 10^-4

        The maximum scaled violation of the constraints for which a
        sample particle is considered comparable to the current global
        optimum.
        Should this not be exceeded, then the current global optimum
        will be updated if the value of the objective function of the
        sample particle is superior.

    'Constraint Warning' : str
        Default = 'ON'

        Activates or deactivates the error exit associated with the
        inability to completely satisfy all constraints, `errno` = 4.
        It is advisable to deactivate this option if the exit `errno` =
        4 is preferred in such cases.

        OFF
            `errno` = 4 will not be returned.

        ON
            `errno` = 4 will be returned if any constraints are
            sufficiently violated at the end of the simulation.

    'Distance Scaling' : str
        Default = 'ON'

        Determines whether distances should be scaled by box widths.

        ON
            When a distance is calculated between x and y, a scaled L^2
            norm is used.

                L^2(x,y) = (sum_{i|u_i != l_i,i <= ndim}((x_i-y_i)/(u_i
                - l_i))^2)^(1/2).

        OFF
            Distances are calculated as the standard L^2 norm without
            any rescaling.

                L^2(x,y) = (sum_i = 1^ndim(x_i-y_i)^2)^(1/2).

    'Distance Tolerance' : float
        Default = 10^-4

        This is the distance, dtol between particles and the global
        optimum which must be reached for the particle to be considered
        converged, i.e., that any subsequent movement of such a particle
        cannot significantly alter the global optimum.
        Once achieved the particle is reset into the box bounds to
        continue searching.

    'Function Precision' : float
        Default = epsilon^0.9

        The argument defines epsilon_r, which is intended to be a
        measure of the accuracy with which the problem function F(x) can
        be computed.
        If r < epsilon or r >= 1, the default value is used.

        The value of epsilon_r should reflect the relative precision of
        1+|F(x)|; i.e., epsilon_r acts as a relative precision when |F|
        is large, and as an absolute precision when |F| is small.
        For example, if F(x) is typically of order 1000 and the first
        six significant digits are known to be correct, an appropriate
        value for epsilon_r would be 10^(-6).
        In contrast, if F(x) is typically of order 10^(-4) and the first
        six significant digits are known to be correct, an appropriate
        value for epsilon_r would be 10^(-10).
        The choice of epsilon_r can be quite complicated for badly
        scaled problems; see Module 8 of Gill `et al.` (1981) for a
        discussion of scaling techniques.
        The default value is appropriate for most simple functions that
        are computed with full accuracy.
        However when the accuracy of the computed function values is
        known to be significantly worse than full precision, the value
        of epsilon_r should be large enough so that no attempt will be
        made to distinguish between function values that differ by less
        than the error inherent in the calculation.

    'Local Boundary Restriction' : float
        Default = 0.5

        Contracts the box boundaries used by a box constrained local
        minimizer to, [beta_l,beta_u], containing the start point x,
        where

            d_i = r * (u_i - l_i) ; beta_l^i = max(l_i, x_i-(d_i)/2) ;
            beta_u^i = min(u_i, x_i+(d_i)/2), i = 1,...,ndim.

        Smaller values of r thereby restrict the size of the domain
        exposed to the local minimizer, possibly reducing the amount of
        work done by the local minimizer.

    'Local Interior Iterations' : int
        The maximum number of iterations or function evaluations the
        chosen local minimizer will perform inside (outside) the main
        loop if applicable.
        For the NAG minimizers these correspond to:

        [table omitted]

        Unless set, these are functions of the arguments passed to
        ``nlp_pso``.

        Setting i = 0 will disable the local minimizer in the
        corresponding algorithmic region.
        For example, setting 'Local Interior Iterations' = 0 and 'Local
        Exterior Iterations' = 30 will cause the algorithm to perform no
        local minimizations inside the main loop of the algorithm, and a
        local minimization with upto 30 iterations after the main loop
        has been exited.

        Note: currently ``opt.bounds_quasi_func_easy`` or
        ``opt.bounds_mod_deriv_easy`` are restricted to using 400*ndim
        and 50*ndim as function evaluation limits respectively. This
        applies to both local minimizations inside and outside the main
        loop. They may still be deactivated in either phase by setting i
        = 0, and may subsequently be reactivated in either phase by
        setting i > 0.

    'Local Interior Major Iterations' : int
        The maximum number of iterations or function evaluations the
        chosen local minimizer will perform inside (outside) the main
        loop if applicable.
        For the NAG minimizers these correspond to:

        [table omitted]

        Unless set, these are functions of the arguments passed to
        ``nlp_pso``.

        Setting i = 0 will disable the local minimizer in the
        corresponding algorithmic region.
        For example, setting 'Local Interior Iterations' = 0 and 'Local
        Exterior Iterations' = 30 will cause the algorithm to perform no
        local minimizations inside the main loop of the algorithm, and a
        local minimization with upto 30 iterations after the main loop
        has been exited.

        Note: currently ``opt.bounds_quasi_func_easy`` or
        ``opt.bounds_mod_deriv_easy`` are restricted to using 400*ndim
        and 50*ndim as function evaluation limits respectively. This
        applies to both local minimizations inside and outside the main
        loop. They may still be deactivated in either phase by setting i
        = 0, and may subsequently be reactivated in either phase by
        setting i > 0.

    'Local Exterior Iterations' : int
        The maximum number of iterations or function evaluations the
        chosen local minimizer will perform inside (outside) the main
        loop if applicable.
        For the NAG minimizers these correspond to:

        [table omitted]

        Unless set, these are functions of the arguments passed to
        ``nlp_pso``.

        Setting i = 0 will disable the local minimizer in the
        corresponding algorithmic region.
        For example, setting 'Local Interior Iterations' = 0 and 'Local
        Exterior Iterations' = 30 will cause the algorithm to perform no
        local minimizations inside the main loop of the algorithm, and a
        local minimization with upto 30 iterations after the main loop
        has been exited.

        Note: currently ``opt.bounds_quasi_func_easy`` or
        ``opt.bounds_mod_deriv_easy`` are restricted to using 400*ndim
        and 50*ndim as function evaluation limits respectively. This
        applies to both local minimizations inside and outside the main
        loop. They may still be deactivated in either phase by setting i
        = 0, and may subsequently be reactivated in either phase by
        setting i > 0.

    'Local Exterior Major Iterations' : int
        The maximum number of iterations or function evaluations the
        chosen local minimizer will perform inside (outside) the main
        loop if applicable.
        For the NAG minimizers these correspond to:

        [table omitted]

        Unless set, these are functions of the arguments passed to
        ``nlp_pso``.

        Setting i = 0 will disable the local minimizer in the
        corresponding algorithmic region.
        For example, setting 'Local Interior Iterations' = 0 and 'Local
        Exterior Iterations' = 30 will cause the algorithm to perform no
        local minimizations inside the main loop of the algorithm, and a
        local minimization with upto 30 iterations after the main loop
        has been exited.

        Note: currently ``opt.bounds_quasi_func_easy`` or
        ``opt.bounds_mod_deriv_easy`` are restricted to using 400*ndim
        and 50*ndim as function evaluation limits respectively. This
        applies to both local minimizations inside and outside the main
        loop. They may still be deactivated in either phase by setting i
        = 0, and may subsequently be reactivated in either phase by
        setting i > 0.

    'Local Interior Tolerance' : float
        Default = 10^-4

        This is the tolerance provided to a local minimizer in the
        interior (exterior) of the main loop of the algorithm.

    'Local Exterior Tolerance' : float
        Default = 10^-4

        This is the tolerance provided to a local minimizer in the
        interior (exterior) of the main loop of the algorithm.

    'Local Interior Minor Iterations' : int
        Where applicable, the secondary number of iterations the chosen
        local minimizer will use inside (outside) the main loop.
        Currently the relevant default values are:

        [table omitted]

    'Local Exterior Minor Iterations' : int
        Where applicable, the secondary number of iterations the chosen
        local minimizer will use inside (outside) the main loop.
        Currently the relevant default values are:

        [table omitted]

    'Local Minimizer' : str
        Default = 'OFF'

        Allows for a choice of submodule ``opt`` functions to be used as
        a coupled, dedicated local minimizer.

        'Local Minimizer' = 'OFF'
            No local minimization will be performed in either the
            INTERIOR or EXTERIOR sections of the algorithm.

        'Local Minimizer' = 'opt.uncon_simplex'
            Use ``opt.uncon_simplex`` as the local minimizer. This does
            not require the calculation of derivatives.

            On a call to `objfun` during a local minimization, `mode` =
            5.

        'Local Minimizer' = 'opt.bounds_mod_deriv_easy'
            Use ``opt.bounds_mod_deriv_easy`` as the local minimizer.
            This requires the calculation of derivatives in `objfun`, as
            indicated by `mode`.

            The box bounds forwarded to this function from ``nlp_pso``
            will have been acted upon by 'Local Boundary Restriction'.
            As such, the domain exposed may be greatly smaller than that
            provided to ``nlp_pso``.

            Accurate derivatives must be provided to this function, and
            will not be approximated internally.
            Each iteration of this local minimizer also requires the
            calculation of both the objective function and its
            derivative.
            Hence on a call to `objfun` during a local minimization,
            `mode` = 7.

        'Local Minimizer' = 'opt.bounds_quasi_func_easy'
            Use ``opt.bounds_quasi_func_easy`` as the local minimizer.
            This does not require the calculation of derivatives.

            On a call to `objfun` during a local minimization, `mode` =
            5.

            The box bounds forwarded to this function from ``nlp_pso``
            will have been acted upon by 'Local Boundary Restriction'.
            As such, the domain exposed may be greatly smaller than that
            provided to ``nlp_pso``.

        'Local Minimizer' = 'opt.uncon_conjgrd_comp'
            Use ``opt.uncon_conjgrd_comp`` as the local minimizer.

            Accurate derivatives must be provided, and will not be
            approximated internally.
            Additionally, each call to `objfun` during a local
            minimization will require either the objective to be
            evaluated alone, or both the objective and its gradient to
            be evaluated.
            Hence on a call to `objfun`, `mode` = 5 or 7.

        'Local Minimizer' = 'opt.nlp1_solve'
            Use ``opt.nlp1_solve`` as the local minimizer. This operates
            such that any derivatives of either the objective function
            or the constraint Jacobian, which you cannot supply, will be
            approximated internally using finite differences.

            Either, the objective, objective gradient, or both may be
            requested during a local minimization, and as such on a call
            to `objfun`, `mode` = 1, 2 or 5.

            The box bounds forwarded to this function from ``nlp_pso``
            will have been acted upon by 'Local Boundary Restriction'.
            As such, the domain exposed may be greatly smaller than that
            provided to ``nlp_pso``.

    'Maximum Function Evaluations' : int
        Default = Imax

        The maximum number of evaluations of the objective function.
        When reached this will return `errno` = 1 and `inform` = 6.

    'Maximum Iterations Completed' : int
        Default = 1000*ndim

        The maximum number of complete iterations that may be performed.
        Once exceeded ``nlp_pso`` will exit with `errno` = 1 and
        `inform` = 5.

        Unless set, this adapts to the parameters passed to ``nlp_pso``.

    'Maximum Iterations Static' : int
        Default = 100

        The maximum number of iterations without any improvement to the
        current global optimum.
        If exceeded ``nlp_pso`` will exit with `errno` = 1 and `inform`
        = 4.
        This exit will be hindered by setting 'Maximum Iterations Static
        Particles' to larger values.

    'Maximum Iterations Static Particles' : int
        Default = 0

        The minimum number of particles that must have converged to the
        current optimum before the function may exit due to 'Maximum
        Iterations Static' with `errno` = 1 and `inform` = 4.

    'Maximum Particles Converged' : int
        Default = Imax

        The maximum number of particles that may converge to the current
        optimum.
        When achieved, ``nlp_pso`` will exit with `errno` = 1 and
        `inform` = 3.
        This exit will be hindered by setting '**Repulsion**' options,
        as these cause the swarm to re-expand.

    'Maximum Particles Reset' : int
        Default = Imax

        The maximum number of particles that may be reset after
        converging to the current optimum.
        Once achieved no further particles will be reset, and any
        particles within 'Distance Tolerance' of the global optimum will
        continue to evolve as normal.

    'Maximum Variable Velocity' : float
        Default = 0.25

        Along any dimension j, the absolute velocity is bounded above by
        |v_j| <= r*(u_j - l_j) = V_max.
        Very low values will greatly increase convergence time.
        There is no upper limit, although larger values will allow more
        particles to be advected out of the box bounds, and values
        greater than 4.0 may cause significant and potentially
        unrecoverable swarm divergence.

    'Objective Scale' : float
        Default = 1.0

        The initial scale for the objective function.
        This will remain fixed if 'Objective Scaling' = USER is
        selected.

    'Objective Scaling' : str
        Default = MAXIMUM

        The method of (re)scaling applied to the objective function when
        the function detects a significant difference between the scale
        and the global and cognitive memory (f~ and f^_j).
        This only has an effect when ncon > 0 and 'Constraint Scaling'
        is active.

        MAXIMUM
            The objective is rescaled with respect to the maximum
            absolute value of the objective in the cognitive and global
            memory.

        MEAN
            The objective is rescaled with respect to the mean absolute
            value of the objective in the cognitive and global memory.

        USER
            The scale remains fixed at the value set using 'Objective
            Scale'.

    'Optimize' : str
        Default = MINIMIZE

        Determines whether to maximize or minimize the objective
        function, or ignore the objective and search for a constrained
        point.

        MINIMIZE
            The objective function will be minimized.

        MAXIMIZE
            The objective function will be maximized. This is
            accomplished by minimizing the negative of the objective.

        CONSTRAINTS
            The objective function will be ignored, and the algorithm
            will attempt to find a feasible point given the provided
            constraints. The objective function will be evaluated at the
            best point found with regards to constraint violations, and
            the final positions returned in `xbest`. The objective will
            be calculated at the best point found in terms of
            constraints only. Should a constrained point be found,
            ``nlp_pso`` will exit with no exception or warning is raised
            and `inform` = 6.

    'Repeatability' : str
        Default = 'OFF'

        Allows for the same random number generator seed to be used for
        every call to ``nlp_pso``. 'Repeatability' = 'OFF' is
        recommended in general.

        OFF
            The internal generation of random numbers will be
            nonrepeatable.

        ON
            The same seed will be used.

    'Repulsion Finalize' : int
        Default = Imax

        The number of iterations performed in a repulsive phase before
        re-contraction.
        This allows a re-diversified swarm to contract back toward the
        current optimum, allowing for a finer search of the near optimum
        space.

    'Repulsion Initialize' : int
        Default = Imax

        The number of iterations without any improvement to the global
        optimum before the algorithm begins a repulsive phase.
        This phase allows the particle swarm to re-expand away from the
        current optimum, allowing more of the domain to be investigated.
        The repulsive phase is automatically ended if a superior optimum
        is found.

    'Repulsion Particles' : int
        Default = 0

        The number of particles required to have converged to the
        current optimum before any repulsive phase may be initialized.
        This will prevent repulsion before a satisfactory search of the
        near optimum area has been performed, which may happen for large
        dimensional problems.

    'Seed' : int
        Default = 0

        Sets the random number generator seed to be used when
        'Repeatability' = 'ON'.
        If set to 0, the default seed will be used.
        If not, the absolute value of 'Seed' will be used to generate
        the random number generator seed.

    'Start' : str
        Default = COLD

        Used to affect the initialization of the function.

        COLD
            The random number generators and all initialization data
            will be generated internally. The variables `xbest`, `fbest`
            and `cbest` need not be set.

        WARM
            You must supply the initial best location, function and
            constraint violation values `xbest`, `fbest` and `cbest`.
            This option is recommended if you already have a data set
            you wish to improve upon.

    'Swarm Standard Deviation' : float
        Default = 0.1

        The target standard deviation of the particle distances from the
        current optimum. Once the standard deviation is below this
        level, ``nlp_pso`` will exit with `errno` = 1 and `inform` = 2.
        This criterion will be penalized by the use of '**Repulsion**'
        options, as these cause the swarm to re-expand, increasing the
        standard deviation of the particle distances from the best
        point.

    'Target Objective' : str
        Default = 'OFF'

        Activate or deactivate the use of a target value as a
        finalization criterion.
        If active, then once the supplied target value for the objective
        function is found (beyond the first iteration if 'Target
        Warning' is active) ``nlp_pso`` will exit with no exception or
        warning is raised and `inform` = 1.
        Other than checking for feasibility only ('Optimize' =
        CONSTRAINTS), this is the only finalization criterion that
        guarantees that the algorithm has been successful.
        If the target value was achieved at the initialization phase or
        first iteration and 'Target Warning' is active, ``nlp_pso`` will
        exit with `errno` = 2.
        This option may take any real value r, or the character ON/OFF
        as well as DEFAULT.
        If this option is queried using ``optget``, the current value of
        r will be returned in rvalue, and cvalue will indicate whether
        this option is ON or OFF.
        The behaviour of the option is as follows:

        r
            Once a point is found with an objective value within the
            'Target Objective Tolerance' of r, ``nlp_pso`` will exit
            successfully with no exception or warning is raised and
            `inform` = 1.

        OFF
            The current value of r will remain stored, however it will
            not be used as a finalization criterion.

        ON
            The current value of r stored will be used as a finalization
            criterion.

        DEFAULT
            The stored value of r will be reset to its default value
            (0.0), and this finalization criterion will be deactivated.

    'Target Objective Value' : float
        Default = 0.0

        Activate or deactivate the use of a target value as a
        finalization criterion.
        If active, then once the supplied target value for the objective
        function is found (beyond the first iteration if 'Target
        Warning' is active) ``nlp_pso`` will exit with no exception or
        warning is raised and `inform` = 1.
        Other than checking for feasibility only ('Optimize' =
        CONSTRAINTS), this is the only finalization criterion that
        guarantees that the algorithm has been successful.
        If the target value was achieved at the initialization phase or
        first iteration and 'Target Warning' is active, ``nlp_pso`` will
        exit with `errno` = 2.
        This option may take any real value r, or the character ON/OFF
        as well as DEFAULT.
        If this option is queried using ``optget``, the current value of
        r will be returned in rvalue, and cvalue will indicate whether
        this option is ON or OFF.
        The behaviour of the option is as follows:

        r
            Once a point is found with an objective value within the
            'Target Objective Tolerance' of r, ``nlp_pso`` will exit
            successfully with no exception or warning is raised and
            `inform` = 1.

        OFF
            The current value of r will remain stored, however it will
            not be used as a finalization criterion.

        ON
            The current value of r stored will be used as a finalization
            criterion.

        DEFAULT
            The stored value of r will be reset to its default value
            (0.0), and this finalization criterion will be deactivated.

    'Target Objective Safeguard' : float
        Default = 10.0 epsilon

        If you have given a target objective value to reach in objval
        (the value of the option 'Target Objective Value'), objsfg sets
        your desired safeguarded termination tolerance, for when objval
        is close to zero.

    'Target Objective Tolerance' : float
        Default = 0.0

        The optional tolerance to a user-specified target value.

    'Target Warning' : str
        Default = 'OFF'

        Activates or deactivates the error exit associated with the
        target value being achieved before entry into the main loop of
        the algorithm, `errno` = 2.

        OFF
            No error will be returned, and the function will exit
            normally.

        ON
            An error will be returned if the target objective is reached
            prematurely, and the function will exit with `errno` = 2.

    'Verify Gradients' : str
        Default = 'ON'

        Adjusts the level of gradient checking performed when gradients
        are required.
        Gradient checks are only performed on the first call to the
        chosen local minimizer if it requires gradients.
        There is no guarantee that the gradient check will be correct,
        as the finite differences used in the gradient check are
        themselves subject to inaccuracies.

        OFF
            No gradient checking will be performed.

        ON
            A cheap gradient check will be performed on both the
            gradients corresponding to the objective through `objfun`
            and those provided via the constraint Jacobian through
            `confun`.

        OBJECTIVE
            A more expensive gradient check will be performed on the
            gradients corresponding to the objective `objfun`. The
            gradients of the constraints will not be checked.

        CONSTRAINTS
            A more expensive check will be performed on the elements of
            `cjac` provided via `confun`. The objective gradient will
            not be checked.

        FULL
            A more expensive check will be performed on both the
            gradient of the objective and the constraint Jacobian.

    'Weight Decrease' : str
        Default = INTEREST

        Determines how particle weights decrease.

        OFF
            Weights do not decrease.

        INTEREST
            Weights decrease through compound interest as w_{IT+1} =
            w_IT(1-W_val), where W_val is the 'Weight Value' and IT is
            the current number of iterations.

        LINEAR
            Weights decrease linearly following w_{IT+1} =
            w_IT-IT*(W_max-W_min)/IT_max, where IT is the iteration
            number and IT_max is the maximum number of iterations as set
            by 'Maximum Iterations Completed'.

    'Weight Initial' : float
        Default = W_max

        The initial value of any particle's inertial weight, W_ini, or
        the minimum possible initial value if initial weights are
        randomized.
        When set, this will override 'Weight Initialize' = 'RANDOMIZED'
        or 'MAXIMUM', and as such these must be set afterwards if so
        desired.

    'Weight Initialize' : str
        Default = MAXIMUM

        Determines how the initial weights are distributed.

        INITIAL
            All weights are initialized at the initial weight, W_ini, if
            set. If 'Weight Initial' has not been set, this will be the
            maximum weight, W_max.

        MAXIMUM
            All weights are initialized at the maximum weight, W_max.

        RANDOMIZED
            Weights are uniformly distributed in (W_min,W_max) or
            (W_ini,W_max) if 'Weight Initial' has been set.

    'Weight Maximum' : float
        Default = 1.0

        The maximum particle weight, W_max.

    'Weight Minimum' : float
        Default = 0.1

        The minimum achievable weight of any particle, W_min.
        Once achieved, no further weight reduction is possible.

    'Weight Reset' : str
        Default = MAXIMUM

        Determines how particle weights are re-initialized.

        INITIAL
            Weights are re-initialized at the initial weight if set. If
            'Weight Initial' has not been set, this will be the maximum
            weight.

        MAXIMUM
            Weights are re-initialized at the maximum weight.

        RANDOMIZED
            Weights are uniformly distributed in (W_min,W_max) or
            (W_ini,W_max) if 'Weight Initial' has been set.

    'Weight Value' : float
        Default = 0.01

        The constant W_val used with 'Weight Decrease' = 'INTEREST'.

    'SMP Callback Thread Safe' : str
        Default = WARNING

        Declare that the callback functions you provide are or are not
        thread safe.
        In particular, this indicates that access to the shared memory
        arrays `iuser` and `ruser` from within your provided callbacks
        is done in a thread safe manner. If these arrays are just used
        to pass constant data, then you may assume they are thread safe.
        If these are also used for workspace, or passing variable data
        such as random number generator seeds, then you must ensure
        these are accessed and updated safely. Whilst this can be done
        using OpenMP critical sections, we suggest their use is
        minimized to prevent unnecessary bottlenecks, and that instead
        individual threads have access to independent subsections of the
        provided arrays where possible.

        YES
            The callback functions have been programmed in a thread safe
            way. The algorithm will use ``OMP_NUM_THREADS`` threads.

        NO
            The callback functions are not thread safe. Setting this
            option will force the algorithm to run on a single thread
            only, and is advisable only for debugging purposes, or if
            you wish to parallelize your callback functions.

        WARNING
            This will cause an immediate exit from ``nlp_pso`` with
            `errno` = 51 if multiple threads are detected. This is to
            inform you that you have not declared the callback functions
            either to be thread safe, or that they are thread unsafe and
            you wish the algorithm to run in serial.

    'SMP Local Minimizer External' : str
        Default = ALL

        Determines how many threads will attempt to locally minimize the
        best found solution after the function has exited the main loop.

        MASTER
            Only the master thread will attempt to find any improvement.
            The local minimization will be launched from the best known
            solution. All other threads will remain effectively idle.

        ALL
            The master thread will perform a local minimization from the
            best known solution, while all other threads will perform a
            local minimization from randomly generated perturbations of
            the best known solution, increasing the chance of an
            improvement. Assuming all local minimizations will take
            approximately the same amount of computation, this will be
            effectively free in terms of real time. It will however
            increase the number of function evaluations performed.

    'SMP Monitor' : str
        Default = SINGLE

        Determines whether the user-supplied function `monmod` is
        invoked once every sub-iteration each thread performs, or only
        once by a single thread after all threads have completed at
        least one sub-iteration.

        SINGLE
            Only one thread will invoke `monmod`, after all threads have
            performed at least one sub-iteration.

        ALL
            Each thread will invoke `monmod` each time it completes a
            sub-iteration. If you wish to alter `x` using `monmod` you
            should use this option, as `monmod` will only receive the
            arrays `x`, `xbest`, `fbest` and `cbest` private to the
            calling thread.

    'SMP Monmod' : str
        Default = SINGLE

        Determines whether the user-supplied function `monmod` is
        invoked once every sub-iteration each thread performs, or only
        once by a single thread after all threads have completed at
        least one sub-iteration.

        SINGLE
            Only one thread will invoke `monmod`, after all threads have
            performed at least one sub-iteration.

        ALL
            Each thread will invoke `monmod` each time it completes a
            sub-iteration. If you wish to alter `x` using `monmod` you
            should use this option, as `monmod` will only receive the
            arrays `x`, `xbest`, `fbest` and `cbest` private to the
            calling thread.

    'SMP Subswarm' : int
        Default = 1

        Determines how many threads support a particle subswarm.
        This is an extra collection of particles constrained to search
        only within a hypercube of edge length 10.0*'Distance Tolerance'
        of the best point known to an individual thread.
        This may improve the number of iterations required to find a
        provided target, particularly if no local minimizer is in use.

        If i <= 0, then this will be disabled on all the threads.

        If i >= OMP_NUM_THREADS, then all the threads will support a
        particle subswarm.

    'SMP Thread Overrun' : int
        Default = Imax

        This option provides control over the level of asynchronicity
        present in a simulation.
        In particular, a barrier synchronization between all threads is
        performed if any thread completes i sub-iterations more than the
        slowest thread, causing all threads to be exposed to the current
        best solution.
        Allowing asynchronous behaviour does however allow individual
        threads to focus on different global optimum candidates some of
        the time, which can inhibit convergence to unwanted sub-optima.
        It also allows for threads to continue searching when other
        threads are completing sub-iterations at a slower rate.

        If i < 1, the algorithm will force a synchronization between
        threads at the end of each iteration.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, ndim = *<value>*.

            Constraint: ndim >= 1.

        (`errno` 12)
            On entry, npar = *<value>*.

            Constraint: npar >= 5*num_threads, where **num_threads** is
            the value returned by the OpenMP environment variable
            ``OMP_NUM_THREADS``, or **num_threads** is 1 for a serial
            version of this function.

        (`errno` 13)
            On entry, ncon = *<value>*.

            Constraint: ncon >= 0.

        (`errno` 14)
            On entry, `bl`[i] = `bu`[i] for all box bounds i.

            Constraint: `bu`[i] > `bl`[i] for at least one box bound i.

        (`errno` 14)
            On entry, `bl`[*<value>*] = *<value>* and `bu`[*<value>*] =
            *<value>*.

            Constraint: `bu`[i] >= `bl`[i] for all i.

        (`errno` 17)
            ``nlp_pso`` has been called with ncon > 0 and `confun` is
            None. Only use None with ncon = 0.

        (`errno` 18)
            The option 'Optimize' = 'CONSTRAINTS' is active, however
            ncon = 0.

        (`errno` 19)
            Error *<value>* occurred whilst adjusting to interior local
            minimizer options.

        (`errno` 19)
            Error *<value>* occurred whilst adjusting to exterior local
            minimizer options.

        (`errno` 21)
            Either the option arrays have not been initialized for
            ``nlp_pso``, or they have become corrupted.

        (`errno` 51)
            Multiple SMP threads have been detected; however, the option
            'SMP Callback Thread Safe' has not been set.

            Set 'SMP Callback Thread Safe' = 'YES' if the provided
            callbacks are thread safe.

            Set 'SMP Callback Thread Safe' = 'NO' if the provided
            callbacks are not thread safe, to force serial execution.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            A finalization criterion was reached that cannot guarantee
            success.

            On exit, `inform` = *<value>*.

        (`errno` 2)
            On entry, `fbest`[*<value>*] = *<value>* but 'Target
            Objective Value' = *<value>*.

        (`errno` 2)
            After initialization `fb` = *<value>* but 'Target Objective
            Value' = *<value>*.

        (`errno` 2)
            Target *<value>* achieved after the first iteration.

            `fb` = *<value>*.

        (`errno` 2)
            The option 'Optimize' = 'CONSTRAINTS' has been set and the
            first point sampled satisfied all constraints.

        (`errno` 4)
            Unable to locate strictly feasible point. *<value>*
            constraints remain violated. This exit may be suppressed
            using the option 'Constraint Warning'.

        (`errno` 32)
            Derivative checks indicate possible errors in the supplied
            derivatives.

    NagCallbackTerminateWarning
        (`errno` 3)
            User requested exit *<value>* during call to `objfun`.

        (`errno` 3)
            User requested exit *<value>* during call to `confun`.

        (`errno` 3)
            User requested exit *<value>* during call to `monmod`.

    References
    ----------
    Gill, P E, Murray, W and Wright, M H, 1981, `Practical
    Optimization`, Academic Press

    Kennedy, J and Eberhart, R C, 1995, `Particle Swarm Optimization`,
    Proceedings of the 1995 IEEE International Conference on Neural
    Networks, 1942--1948

    Koh, B, George, A D, Haftka, R T and Fregly, B J, 2006, `Parallel
    Asynchronous Particle Swarm Optimization`, International Journal for
    Numerical Methods in Engineering (67(4)), 578--595

    Vaz, A I and Vicente, L N, 2007, `A Particle Swarm Pattern Search
    Method for Bound Constrained Global Optimization`, Journal of Global
    Optimization (39(2)), 197--219, Kluwer Academic Publishers
    """

    fun_name = 'naginterfaces.library.glopt.nlp_pso'

    cb_objfun_name = fun_name + ':objfun'

    def cb_objfun(mode, ndim, x, objf, vecout, nstate, data=None):
        """Wrapping user's objfun. Called as base objfun."""
        if data is not None:
            _objf_objfun, _vecout_objfun = objfun(
                mode,
                x,
                objf,
                vecout,
                nstate,
                data,
            )
        else:
            _objf_objfun, _vecout_objfun = objfun(
                mode,
                x,
                objf,
                vecout,
                nstate,
            )
        _objf_objfun_base = _objf_objfun
        vecout[:] = _utils._to_ndarray(
            {'fun_name': cb_objfun_name, 'returned': 'vecout'},
            _vecout_objfun,
            copy=False,
            exp_shape=(ndim,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        return _objf_objfun_base

    cb_confun_name = fun_name + ':confun'

    def cb_confun(mode, ncon, ndim, needc, x, c, cjac, nstate, data=None):
        """Wrapping user's confun. Called as base confun."""
        if data is not None:
            _c_confun, _cjac_confun = confun(
                mode,
                needc,
                x,
                cjac,
                nstate,
                data,
            )
        else:
            _c_confun, _cjac_confun = confun(
                mode,
                needc,
                x,
                cjac,
                nstate,
            )
        c[:] = _utils._to_ndarray(
            {'fun_name': cb_confun_name, 'returned': 'c'},
            _c_confun,
            copy=False,
            exp_shape=(ncon,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _cjac_confun_np = _utils._to_ndarray(
            {'fun_name': cb_confun_name, 'returned': 'cjac'},
            _cjac_confun,
            copy=False,
            exp_shape=(ncon, ndim),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _utils._check_callback_sorder(
            {'fun_name': cb_confun_name},
            ((_cjac_confun_np, 'cjac'),),
            wrapper_sorder=_m_sorder,
            spiked_sorder=spiked_sorder,
        )
        cjac[:, :] = _cjac_confun_np

    cb_monmod_name = fun_name + ':monmod'

    def cb_monmod(ndim, ncon, npar, x, xb, fb, cb, xbest, fbest, cbest, itt, data=None):
        """Wrapping user's monmod. Called as base monmod."""
        if data is not None:
            _x_monmod = monmod(
                x,
                xb,
                fb,
                cb,
                xbest,
                fbest,
                cbest,
                itt,
                data,
            )
        else:
            _x_monmod = monmod(
                x,
                xb,
                fb,
                cb,
                xbest,
                fbest,
                cbest,
                itt,
            )
        _x_monmod_np = _utils._to_ndarray(
            {'fun_name': cb_monmod_name, 'returned': 'x'},
            _x_monmod,
            copy=False,
            exp_shape=(ndim, npar),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _utils._check_callback_sorder(
            {'fun_name': cb_monmod_name},
            ((_x_monmod_np, 'x'),),
            wrapper_sorder=_m_sorder,
            spiked_sorder=spiked_sorder,
        )
        x[:, :] = _x_monmod_np

    _xbest = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xbest'},
        xbest,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _cbest = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'cbest'},
        cbest,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_xbest, 'xbest'), (_cbest, 'cbest')),
        spiked_sorder=spiked_sorder,
    )
    _ndim_py = _b_utils._EngineFloat64ArrayType.get_shape(_xbest, {'fun_name': fun_name, 'entity': 'xbest'}, exp_rank=2)[0]
    _ndim = _b_utils._EngineIntScalarType(
        _ndim_py, {'fun_name': fun_name, 'entity': '_ndim'},
    ).to_ctypes()
    _ncon_py = _b_utils._EngineFloat64ArrayType.get_shape(_cbest, {'fun_name': fun_name, 'entity': 'cbest'}, exp_rank=2)[0]
    _ncon = _b_utils._EngineIntScalarType(
        _ncon_py, {'fun_name': fun_name, 'entity': '_ncon'},
    ).to_ctypes()
    _npar_py = _b_utils._EngineFloat64ArrayType.get_shape(_xbest, {'fun_name': fun_name, 'entity': 'xbest'}, exp_rank=2)[1]
    _npar = _b_utils._EngineIntScalarType(
        _npar_py, {'fun_name': fun_name, 'entity': '_npar'},
    ).to_ctypes()
    _xb = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ndim_py,),
    )
    _cb = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_ncon_py,),
    )
    _bl = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'bl'},
        bl,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _bu = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'bu'},
        bu,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _fbest = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'fbest'},
        fbest,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b_utils._check_callable(objfun, {'fun_name': fun_name, 'entity': 'objfun'})
    _objfun = cb_objfun
    _b_utils._check_callable(confun, {'fun_name': fun_name, 'entity': 'confun'}, allow_none=True)
    _confun = None if confun is None else cb_confun
    _b_utils._check_callable(monmod, {'fun_name': fun_name, 'entity': 'monmod'}, allow_none=True)
    _monmod = None if monmod is None else cb_monmod
    _itt = _b_utils._EngineIntArrayType.empty_ndarray(
        (7,),
    )
    try:
        _return_scalars = _b_glopt.nlp_pso(
            _ndim,
            _ncon,
            _npar,
            _xb,
            _cb,
            _bl,
            _bu,
            _xbest,
            _fbest,
            _cbest,
            _objfun,
            _confun,
            _monmod,
            comm,
            _itt,
            data,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _xbest_py = _xbest
    _fbest_py = _fbest
    _cbest_py = _cbest
    _xb_py = _xb
    _fb_py = _return_scalars.fb
    _cb_py = _cb
    _itt_py = _itt
    _inform_py = _return_scalars.inform
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['xb', 'fb', 'cb', 'xbest', 'fbest', 'cbest', 'itt', 'inform'],
    )(
        _xb_py,
        _fb_py,
        _cb_py,
        _xbest_py,
        _fbest_py,
        _cbest_py,
        _itt_py,
        _inform_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def nlp_multistart_sqp(n, ncnln, bl, bu, objfun, npts, repeat, nb, comm, a=None, confun=None, start=None, data=None, io_manager=None, spiked_sorder='C'):
    u"""
    Global optimization using multi-start, nonlinear constraints.

    Note: this function uses optional algorithmic parameters, see also:
    ``optset``.

    ``nlp_multistart_sqp`` is designed to find the global minimum of an
    arbitrary smooth function subject to constraints (which may include
    simple bounds on the variables, linear constraints and smooth
    nonlinear constraints) by generating a number of different starting
    points and performing a local search from each using sequential
    quadratic programming.

    For full information please refer to the NAG Library document for
    e05uc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e05/e05ucf.html

    Parameters
    ----------
    n : int
        n, the number of variables.

    ncnln : int
        n_N, the number of nonlinear constraints.

    bl : float, array-like, shape (`n`+nclin+`ncnln`)
        `bl` must contain the lower bounds for all the constraints.

    bu : float, array-like, shape (`n`+nclin+`ncnln`)
        `bu` must contain the upper bounds for all the constraints.

    objfun : callable (objf, objgrd) = objfun(mode, x, objgrd, nstate,
        data=None)
        `objfun` must calculate the objective function F(x) and
        (optionally) its gradient g(x) = dF/dx for a specified n-vector
        x.

        Parameters
        ~~~~~~~~~~
        mode : int
            Indicates which values must be assigned during each call of
            `objfun`. Only the following values need be assigned:

            `mode` = 0

                `objf`.

            `mode` = 1

                All available elements of `objgrd`.

            `mode` = 2

                `objf` and all available elements of `objgrd`.

        x : float, ndarray, shape (n)
            x, the vector of variables at which the objective function
            and/or all available elements of its gradient are to be
            evaluated.

        objgrd : float, ndarray, shape (n)
            The elements of `objgrd` are set to special values which
            enable ``nlp_multistart_sqp`` to detect whether they are
            changed by `objfun`.

        nstate : int
            If `nstate` = 1 then ``nlp_multistart_sqp`` is calling
            `objfun` for the first time on the current local
            optimization problem. This argument setting allows you to
            save computation time if certain data must be calculated
            only once.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        objf : float
            If `mode` = 0 or 2, `objf` must be set to the value of the
            objective function at x.

        objgrd : float, array-like, shape (n)
            If `mode` = 1 or 2, `objgrd` must return the available
            elements of the gradient evaluated at x.

    npts : int
        The number of different starting points to be generated and
        used. The more points used, the more likely that the best
        returned solution will be a global minimum.

    repeat : bool
        Is passed as an argument to `start` and may be used to
        initialize a random number generator to a repeatable, or
        non-repeatable, sequence.

    nb : int
        The number of solutions to be returned. The function saves up to
        `nb` local minima ordered by increasing value of the final
        objective function. If the defining criterion for 'best
        solution' is only that the value of the objective function is as
        small as possible then `nb` should be set to 1. However, if you
        want to look at other solutions that may have desirable
        properties then setting `nb` > 1 will produce `nb` local minima,
        ordered by increasing value of their objective functions at the
        minima.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``optset``.

    a : None or float, array-like, shape (nclin, :), optional
        Note: the required extent for this argument in dimension 2 is
        determined as follows: if nclin>0: `n`; otherwise: 1.

        The matrix A_L of general linear constraints in (1). That is,
        the ith row contains the coefficients of the ith general linear
        constraint, for i = 1,2,...,nclin.

    confun : None or callable (c, cjsl) = confun(mode, needc, x, cjsl,
        nstate, data=None), optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `confun` must calculate the vector c(x) of nonlinear constraint
        functions and (optionally) its Jacobian (= dc/dx) for a
        specified n-element vector x. If there are no nonlinear
        constraints (i.e., `ncnln` = 0), `confun` will never be called
        by ``nlp_multistart_sqp`` and `confun` may be None. If there are
        nonlinear constraints, the first call to `confun` will occur
        before the first call to `objfun`.

        Parameters
        ~~~~~~~~~~
        mode : int
            Indicates which values must be assigned during each call of
            `confun`. Only the following values need be assigned, for
            each value of i such that `needc`[i-1] > 0:

            `mode` = 0

                `c`[i-1].

            `mode` = 1

                All available elements in the ith row of `cjsl`.

            `mode` = 2

                `c`[i-1] and all available elements in the ith row of
                `cjsl`.

        needc : int, ndarray, shape (ncnln)
            The indices of the elements of `c` and/or `cjsl` that must
            be evaluated by `confun`. If `needc`[i-1] > 0, `c`[i-1]
            and/or the available elements of the ith row of `cjsl` (see
            argument `mode`) must be evaluated at x.

        x : float, ndarray, shape (n)
            x, the vector of variables at which the constraint functions
            and/or the available elements of the constraint Jacobian are
            to be evaluated.

        cjsl : float, ndarray, shape (ncnln, n)
            `cjsl` may be regarded as a two-dimensional 'slice' of the
            three-dimensional array `cjac` of ``nlp_multistart_sqp``.

            Unless 'Derivative Level' = 2 or 3 (the default setting is
            'Derivative Level' = 3), the elements of `cjsl` are set to
            special values which enable ``nlp_multistart_sqp`` to detect
            whether they are changed by `confun`.

        nstate : int
            If `nstate` = 1 then ``nlp_multistart_sqp`` is calling
            `confun` for the first time on the current local
            optimization problem. This argument setting allows you to
            save computation time if certain data must be calculated
            only once.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        c : float, array-like, shape (ncnln)
            If `needc`[k-1] > 0 and `mode` = 0 or 2, `c`[k-1] must
            contain the value of c_k(x). The remaining elements of `c`,
            corresponding to the non-positive elements of `needc`, need
            not be set.

        cjsl : float, array-like, shape (ncnln, n)
            If `needc`[k-1] > 0 and `mode` = 1 or 2, the kth row of
            `cjsl` must contain the available elements of the vector
            nabla c_k given by

                nabla c_k = (dc_k/dx_1,dc_k/dx_2,...,dc_k/dx_n)^T,

            where dc_k/dx_j is the partial derivative of the kth
            constraint with respect to the jth variable, evaluated at
            the point x. See also the argument `nstate`. The remaining
            rows of `cjsl`, corresponding to non-positive elements of
            `needc`, need not be set.

            If all elements of the constraint Jacobian are known (i.e.,
            'Derivative Level' = 2 or 3), any constant elements may be
            assigned to `cjsl` one time only at the start of each local
            optimization.

            An element of `cjsl` that is not subsequently assigned in
            `confun` will retain its initial value throughout the local
            optimization.

            Constant elements may be loaded into `cjsl` during the first
            call to `confun` for the local optimization (signalled by
            the value `nstate` = 1).

            The ability to preload constants is useful when many
            Jacobian elements are identically zero, in which case `cjsl`
            may be initialized to zero and nonzero elements may be reset
            by `confun`.

            Note that constant nonzero elements do affect the values of
            the constraints.

            Thus, if `cjsl`[k-1,j-1] is set to a constant value, it need
            not be reset in subsequent calls to `confun`, but the value
            `cjsl`[k-1,j-1]*`x`[j-1] must nonetheless be added to
            `c`[k-1].

            For example, if `cjsl`[0,0] = 2 and `cjsl`[0,1] = -5 then
            the term 2*`x`[0]-5*`x`[1] must be included in the
            definition of `c`[0].

            It must be emphasized that, if 'Derivative Level' = 0 or 1,
            unassigned elements of `cjsl` are not treated as constant;
            they are estimated by finite differences, at nontrivial
            expense.

            If you do not supply a value for the option 'Difference
            Interval', an interval for each element of x is computed
            automatically at the start of each local optimization.

            The automatic procedure can usually identify constant
            elements of `cjsl`, which are then computed once only by
            finite differences.

    start : None or callable quas = start(quas, repeat, bl, bu,
        data=None), optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `start` must calculate the `npts` starting points to be used by
        the local optimizer.

        If you do not wish to write a function specific to your problem
        then None may be used. In this case, a default function uses the
        NAG quasi-random number generators to distribute starting points
        uniformly across the domain.

        It is affected by the value of `repeat`.

        Parameters
        ~~~~~~~~~~
        quas : float, ndarray, shape (n, npts)
            All elements of `quas` will have been set to zero, so only
            nonzero values need be set subsequently.

        repeat : bool
            Specifies whether a repeatable or non-repeatable sequence of
            points are to be generated.

        bl : float, ndarray, shape (n)
            The lower bounds on the variables. These may be used to
            ensure that the starting points generated in some sense
            'cover' the region, but there is no requirement that a
            starting point be feasible.

        bu : float, ndarray, shape (n)
            The upper bounds on the variables. (See `bl`.)

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        quas : float, array-like, shape (n, npts)
            Must contain the starting points for the `npts` local
            minimizations, i.e., `quas`[j-1,i-1] must contain the jth
            component of the ith starting point.

    data : arbitrary, optional
        User-communication data for callback functions.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    spiked_sorder : str, optional
        If `a` is spiked (i.e., has unit extent in all but one
        dimension, or has size 1), `spiked_sorder` selects the storage
        order to associate with it in the NAG Engine:

        spiked_sorder = 'C'
          row-major storage will be used;

        spiked_sorder = 'F'
          column-major storage will be used.

        Two-dimensional arrays returned from callback functions in this
        routine must then use the same storage order.

    Returns
    -------
    x : float, ndarray, shape (`n`, `nb`)
        `x`[j-1,i-1] contains the final estimate of the ith solution,
        for j = 1,2,...,`n`.

    objf : float, ndarray, shape (`nb`)
        `objf`[i-1] contains the value of the objective function at the
        final iterate for the ith solution.

    objgrd : float, ndarray, shape (`n`, `nb`)
        `objgrd`[j-1,i-1] contains the gradient of the objective
        function for the ith solution at the final iterate (or its
        finite difference approximation), for j = 1,2,...,`n`.

    itera : int, ndarray, shape (`nb`)
        `itera`[i-1] contains the number of major iterations performed
        to obtain the ith solution. If less than `nb` solutions are
        returned then `itera`[`nb`-1] contains the number of starting
        points that have resulted in a converged solution. If this is
        close to `npts` then this might be indicative that fewer than
        `nb` local minima exist.

    c : None or float, ndarray, shape (`ncnln`, `nb`)
        If `ncnln` > 0, `c`[j-1,i-1] contains the value of the jth
        nonlinear constraint function c_j at the final iterate, for the
        ith solution, for j = 1,2,...,`ncnln`.

        If `ncnln` = 0, `c` is returned as None.

    cjac : None or float, ndarray, shape (`ncnln`, :, `nb`)
        If `ncnln` > 0, `cjac` contains the Jacobian matrices of the
        nonlinear constraint functions at the final iterate for each of
        the returned solutions, i.e., `cjac`[k-1,j-1,i-1] contains the
        partial derivative of the kth constraint function with respect
        to the jth variable, for j = 1,2,...,`n`, for k =
        1,2,...,`ncnln`, for the ith solution. (See the discussion of
        argument `cjsl` under `confun`.)

        If `ncnln` = 0, `cjac` is returned as None.

    r : float, ndarray, shape (`n`, `n`, `nb`)
        For each of the `nb` solutions `r` will contain a form of the
        Hessian; for the ith returned solution `r`[ldr,sdr,i] contains
        the Hessian that would be returned from the local minimizer. If
        'Hessian' = 'NO', the default, each `r`[ldr,sdr,i] contains the
        upper triangular Cholesky factor R of Q^THQ, an estimate of the
        transformed and reordered Hessian of the Lagrangian at x. If
        'Hessian' = 'YES', `r`[ldr,sdr,i] contains the upper triangular
        Cholesky factor R of H, the approximate (untransformed) Hessian
        of the Lagrangian, with the variables in the natural order.

    clamda : float, ndarray, shape (`n`+nclin+`ncnln`, `nb`)
        The values of the QP multipliers from the last QP subproblem
        solved for the ith solution. `clamda`[j-1,i-1] should be
        non-negative if `istate`[j-1,i-1] = 1 and non-positive if
        `istate`[j-1,i-1] = 2.

    istate : int, ndarray, shape (`n`+nclin+`ncnln`, `nb`)
        `istate`[j-1,i-1] contains the status of the constraints in the
        QP working set for the ith solution. The significance of each
        possible value of `istate`[j-1,i-1] is as follows:

        [table omitted]

    info : int, ndarray, shape (`nb`)
        `info`[i-1] contains one of 0, 1 or 6. Please see the
        description of each corresponding value of errno on exit from
        ``opt.nlp1_solve`` for detailed explanations of these exit
        values. As usual 0 denotes success.

        If `errno` = 8 on exit, then not all `nb` solutions have been
        found, and `info`[`nb`-1] contains the number of solutions
        actually found.

    Other Parameters
    ----------------
    'Punch Unit' : int
        Default = 6

        This option allows you to send information arising from an
        appropriate setting of 'Out_Level' to be sent to the Fortran
        unit number defined by 'Punch Unit'.
        If you wish this file to be different to the standard output
        unit where other output is displayed then this file should be
        attached by calling the ``FileObjManager`` method
        :meth:`~naginterfaces.base.utils.FileObjManager.unit_from_fileobj`
        prior to calling ``nlp_multistart_sqp``.

    'Out_Level' : int
        Default = 0

        This option defines the amount of extra information to be sent
        to the Fortran unit number defined by 'Punch Unit'.
        The possible choices for i are the following:

        [table omitted]

    'Central Difference Interval' : float
        Default values are computed

        If the algorithm switches to central differences because the
        forward-difference approximation is not sufficiently accurate,
        the value of r is used as the difference interval for every
        element of x.
        The switch to central differences is indicated by ``C`` at the
        end of each line of intermediate printout produced by the major
        iterations (see [Further Comments]).
        The use of finite differences is discussed further under the
        option 'Difference Interval'.

        If you supply a value for this option, a small value between 0.0
        and 1.0 is appropriate.

    'Crash Tolerance' : float
        Default = 0.01

        This value is used when ``nlp_multistart_sqp`` selects an
        initial working set.
        If 0 <= r <= 1, the initial working set will include (if
        possible) bounds or general inequality constraints that lie
        within r of their bounds.
        In particular, a constraint of the form a_j^Tx >= l will be
        included in the initial working set if |a_j^Tx-l| <= r(1+|l|).
        If r < 0 or r > 1, the default value is used.

    'Defaults' : valueless
        This special keyword is used to reset all options to their
        default values, and any random state stored in the array comm
        will be destroyed.

        Any option value given with this keyword will be ignored.
        This option cannot be queried or got.

    'Derivative Level' : int
        Default = 3

        This argument indicates which derivatives are provided in
        functions `objfun` and `confun`.
        The possible choices for i are the following.

        [table omitted]

        The value i = 3 should be used whenever possible, since
        ``nlp_multistart_sqp`` is more reliable (and will usually be
        more efficient) when all derivatives are exact.

        If i = 0 or 2, ``nlp_multistart_sqp`` will estimate the
        unspecified elements of the objective gradient, using finite
        differences.
        The computation of finite difference approximations usually
        increases the total run-time, since a call to `objfun` is
        required for each unspecified element.
        Furthermore, less accuracy can be attained in the solution (see
        Module 8 of Gill `et al.` (1981), for a discussion of limiting
        accuracy).

        If i = 0 or 1, ``nlp_multistart_sqp`` will approximate
        unspecified elements of the constraint Jacobian.
        One call to `confun` is needed for each variable for which
        partial derivatives are not available.
        For example, if the Jacobian has the form

            [table omitted]

        where '*' indicates an element provided by you and '?' indicates
        an unspecified element, ``nlp_multistart_sqp`` will call
        `confun` twice: once to estimate the missing element in column
        2, and again to estimate the two missing elements in column 3.
        (Since columns 1 and 4 are known, they require no calls to
        `confun`.)

        At times, central differences are used rather than forward
        differences, in which case twice as many calls to `objfun` and
        `confun` are needed. (The switch to central differences is not
        under your control.)

        If i < 0 or i > 3, the default value is used.

    'Difference Interval' : float
        Default values are computed

        This option defines an interval used to estimate derivatives by
        finite differences in the following circumstances:

        (a) For verifying the objective and/or constraint gradients (see
            the description of the option 'Verify').

        (#) For estimating unspecified elements of the objective
            gradient or the constraint Jacobian.

        In general, a derivative with respect to the jth variable is
        approximated using the interval delta_j, where delta_j =
        r(1+|x^_j|), with x^ the first point feasible with respect to
        the bounds and linear constraints.
        If the functions are well scaled, the resulting derivative
        approximation should be accurate to O(r).
        See Gill `et al.` (1981) for a discussion of the accuracy in
        finite difference approximations.

        If a difference interval is not specified, a finite difference
        interval will be computed automatically for each variable by a
        procedure that requires up to six calls of `confun` and `objfun`
        for each element.
        This option is recommended if the function is badly scaled or
        you wish to have ``nlp_multistart_sqp`` determine constant
        elements in the objective and constraint gradients (see the
        descriptions of `confun` and `objfun` in [Parameters]).

        If you supply a value for this option, a small value between 0.0
        and 1.0 is appropriate.

    'Feasibility Tolerance' : float
        Default = sqrt(epsilon)

        The scalar r defines the maximum acceptable `absolute`
        violations in linear and nonlinear constraints at a 'feasible'
        point; i.e., a constraint is considered satisfied if its
        violation does not exceed r.
        If r < epsilon or r >= 1, the default value is used.
        Using this keyword sets both options 'Linear Feasibility
        Tolerance' and 'Nonlinear Feasibility Tolerance' to r, if
        epsilon <= r < 1. (Additional details are given under the
        descriptions of these options.)

    'Function Precision' : float
        Default = epsilon^0.9

        This argument defines epsilon_r, which is intended to be a
        measure of the accuracy with which the problem functions F(x)
        and c(x) can be computed.
        If r < epsilon or r >= 1, the default value is used.

        The value of epsilon_r should reflect the relative precision of
        1+|F(x)|; i.e., epsilon_r acts as a relative precision when |F|
        is large, and as an absolute precision when |F| is small.
        For example, if F(x) is typically of order 1000 and the first
        six significant digits are known to be correct, an appropriate
        value for epsilon_r would be 10^(-6).
        In contrast, if F(x) is typically of order 10^(-4) and the first
        six significant digits are known to be correct, an appropriate
        value for epsilon_r would be 10^(-10).
        The choice of epsilon_r can be quite complicated for badly
        scaled problems; see Module 8 of Gill `et al.` (1981) for a
        discussion of scaling techniques.
        The default value is appropriate for most simple functions that
        are computed with full accuracy.
        However, when the accuracy of the computed function values is
        known to be significantly worse than full precision, the value
        of epsilon_r should be large enough so that
        ``nlp_multistart_sqp`` will not attempt to distinguish between
        function values that differ by less than the error inherent in
        the calculation.

    'Hessian' : str
        Default = NO

        This option controls the contents of the upper triangular matrix
        R (see [Parameters]). ``nlp_multistart_sqp`` works exclusively
        with the `transformed and reordered` Hessian H_Q (6), and hence
        extra computation is required to form the Hessian itself.
        If 'Hessian' = 'NO', `r` contains the Cholesky factor of the
        transformed and reordered Hessian.
        If 'Hessian' = 'YES', the Cholesky factor of the approximate
        Hessian itself is formed and stored in `r`.

    'Infinite Bound Size' : float
        Default = 10^20

        This defines the 'infinite' bound infbnd in the definition of
        the problem constraints.
        Any upper bound greater than or equal to infbnd will be regarded
        as infinity (and similarly any lower bound less than or equal to
        -infbnd will be regarded as - infinity).

    'Infinite Step Size' : float
        Default = max(bigbnd, 10^20)

        If r > 0, r specifies the magnitude of the change in variables
        that is treated as a step to an unbounded solution.
        If the change in x during an iteration would exceed the value of
        r, the objective function is considered to be unbounded below in
        the feasible region.
        If r <= 0, the default value is used.

    'Line Search Tolerance' : float
        Default = 0.9

        The value r (0 <= r < 1) controls the accuracy with which the
        step alpha taken during each iteration approximates a minimum of
        the merit function along the search direction (the smaller the
        value of r, the more accurate the linesearch).
        The default value r = 0.9 requests an inaccurate search, and is
        appropriate for most problems, particularly those with any
        nonlinear constraints.

        If there are no nonlinear constraints, a more accurate search
        may be appropriate when it is desirable to reduce the number of
        major iterations -- for example, if the objective function is
        cheap to evaluate, or if a substantial number of derivatives are
        unspecified.
        If r < 0 or r >= 1, the default value is used.

    'Linear Feasibility Tolerance' : float
        Default = sqrt(epsilon)

        The default value of r_2 is epsilon^0.33 if 'Derivative Level' =
        0 or 1, and sqrt(epsilon) otherwise.

        The scalars r_1 and r_2 define the maximum acceptable `absolute`
        violations in linear and nonlinear constraints at a 'feasible'
        point; i.e., a linear constraint is considered satisfied if its
        violation does not exceed r_1, and similarly for a nonlinear
        constraint and r_2.
        If r_m < epsilon or r_m >= 1, the default value is used, for m =
        1,2,...,2.

        On entry to ``nlp_multistart_sqp``, an iterative procedure is
        executed in order to find a point that satisfies the linear
        constraints and bounds on the variables to within the tolerance
        r_1.
        All subsequent iterates will satisfy the linear constraints to
        within the same tolerance (unless r_1 is comparable to the
        finite difference interval).

        For nonlinear constraints, the feasibility tolerance r_2 defines
        the largest constraint violation that is acceptable at an
        optimal point.
        Since nonlinear constraints are generally not satisfied until
        the final iterate, the value of option 'Nonlinear Feasibility
        Tolerance' acts as a partial termination criterion for the
        iterative sequence generated by ``nlp_multistart_sqp`` (see the
        discussion of option 'Optimality Tolerance').

        These tolerances should reflect the precision of the
        corresponding constraints.
        For example, if the variables and the coefficients in the linear
        constraints are of order unity, and the latter are correct to
        about 6 decimal digits, it would be appropriate to specify r_1
        as 10^(-6).

    'Nonlinear Feasibility Tolerance' : float
        Default = epsilon^0.33 or sqrt(epsilon)

        The default value of r_2 is epsilon^0.33 if 'Derivative Level' =
        0 or 1, and sqrt(epsilon) otherwise.

        The scalars r_1 and r_2 define the maximum acceptable `absolute`
        violations in linear and nonlinear constraints at a 'feasible'
        point; i.e., a linear constraint is considered satisfied if its
        violation does not exceed r_1, and similarly for a nonlinear
        constraint and r_2.
        If r_m < epsilon or r_m >= 1, the default value is used, for m =
        1,2,...,2.

        On entry to ``nlp_multistart_sqp``, an iterative procedure is
        executed in order to find a point that satisfies the linear
        constraints and bounds on the variables to within the tolerance
        r_1.
        All subsequent iterates will satisfy the linear constraints to
        within the same tolerance (unless r_1 is comparable to the
        finite difference interval).

        For nonlinear constraints, the feasibility tolerance r_2 defines
        the largest constraint violation that is acceptable at an
        optimal point.
        Since nonlinear constraints are generally not satisfied until
        the final iterate, the value of option 'Nonlinear Feasibility
        Tolerance' acts as a partial termination criterion for the
        iterative sequence generated by ``nlp_multistart_sqp`` (see the
        discussion of option 'Optimality Tolerance').

        These tolerances should reflect the precision of the
        corresponding constraints.
        For example, if the variables and the coefficients in the linear
        constraints are of order unity, and the latter are correct to
        about 6 decimal digits, it would be appropriate to specify r_1
        as 10^(-6).

    'List' : valueless
        Option 'List' enables printing of each option specification as
        it is supplied. 'Nolist' suppresses this printing.

    'Nolist' : valueless
        Default

        Option 'List' enables printing of each option specification as
        it is supplied. 'Nolist' suppresses this printing.

    'Major Iteration Limit' : int
        Default = max(50, 3*(n+n_L)+10n_N)

        The value of i specifies the maximum number of major iterations
        allowed before termination of the subproblem.
        Setting i = 0 and 'Major Print Level' > 0 means that the
        workspace needed will be computed and printed, but no iterations
        will be performed.
        If i < 0, the default value is used.

    'Iteration Limit' : int
        Default = max(50, 3*(n+n_L)+10n_N)

        The value of i specifies the maximum number of major iterations
        allowed before termination of the subproblem.
        Setting i = 0 and 'Major Print Level' > 0 means that the
        workspace needed will be computed and printed, but no iterations
        will be performed.
        If i < 0, the default value is used.

    'Iters' : int
        Default = max(50, 3*(n+n_L)+10n_N)

        The value of i specifies the maximum number of major iterations
        allowed before termination of the subproblem.
        Setting i = 0 and 'Major Print Level' > 0 means that the
        workspace needed will be computed and printed, but no iterations
        will be performed.
        If i < 0, the default value is used.

    'Itns' : int
        Default = max(50, 3*(n+n_L)+10n_N)

        The value of i specifies the maximum number of major iterations
        allowed before termination of the subproblem.
        Setting i = 0 and 'Major Print Level' > 0 means that the
        workspace needed will be computed and printed, but no iterations
        will be performed.
        If i < 0, the default value is used.

    'Major Print Level' : int
        Default = 0

        The value of i controls the amount of printout produced by the
        major iterations of ``nlp_multistart_sqp``, as indicated below.
        A detailed description of the printed output is given in
        [Further Comments] (summary output at each major iteration and
        the final solution) and [Monitoring Information] (monitoring
        information at each major iteration). (See also the description
        of the option 'Minor Print Level'.)

        The following printout is sent to the file object associated
        with the advisory I/O unit (see
        :class:`~naginterfaces.base.utils.FileObjManager`):

        [table omitted]

        For the other values described below, the arguments used by
        ``opt.nlp1_solve`` are displayed in addition to intermediate and
        final output.

        [table omitted]

        The following printout is sent to the logical unit number
        defined by the option 'Monitoring File':

        [table omitted]

        If 'Major Print Level' >= 5 and the unit number defined by the
        option 'Monitoring File' is the advisory unit number, then the
        summary output for each major iteration is suppressed.

    'Print Level' : int
        Default = 0

        The value of i controls the amount of printout produced by the
        major iterations of ``nlp_multistart_sqp``, as indicated below.
        A detailed description of the printed output is given in
        [Further Comments] (summary output at each major iteration and
        the final solution) and [Monitoring Information] (monitoring
        information at each major iteration). (See also the description
        of the option 'Minor Print Level'.)

        The following printout is sent to the file object associated
        with the advisory I/O unit (see
        :class:`~naginterfaces.base.utils.FileObjManager`):

        [table omitted]

        For the other values described below, the arguments used by
        ``opt.nlp1_solve`` are displayed in addition to intermediate and
        final output.

        [table omitted]

        The following printout is sent to the logical unit number
        defined by the option 'Monitoring File':

        [table omitted]

        If 'Major Print Level' >= 5 and the unit number defined by the
        option 'Monitoring File' is the advisory unit number, then the
        summary output for each major iteration is suppressed.

    'Minor Iteration Limit' : int
        Default = max(50, 3*(n+n_L+n_N))

        The value of i specifies the maximum number of iterations for
        finding a feasible point with respect to the bounds and linear
        constraints (if any).
        The value of i also specifies the maximum number of minor
        iterations for the optimality phase of each QP subproblem.
        If i <= 0, the default value is used.

    'Minor Print Level' : int
        Default = 0

        The value of i controls the amount of printout produced by the
        minor iterations of ``nlp_multistart_sqp`` (i.e., the iterations
        of the quadratic programming algorithm), as indicated below.
        A detailed description of the printed output is given in
        [Further Comments] (summary output at each minor iteration and
        the final QP solution) and [Monitoring Information] (monitoring
        information at each minor iteration). (See also the description
        of the option 'Major Print Level'.) The following printout is
        sent to the file object associated with the advisory I/O unit
        (see :class:`~naginterfaces.base.utils.FileObjManager`):

        [table omitted]

        The following printout is sent to the logical unit number
        defined by the option 'Monitoring File':

        [table omitted]

        If 'Minor Print Level' >= 5 and the unit number defined by the
        option 'Monitoring File' is the advisory unit number, the
        summary output for each minor iteration is suppressed.

    'Monitoring File' : int
        Default = -1

        If i >= 0 and 'Major Print Level' >= 5 or i >= 0 and 'Minor
        Print Level' >= 5, monitoring information produced by
        ``nlp_multistart_sqp`` at every iteration is sent to a file with
        logical unit number i. If i < 0 and/or 'Major Print Level' < 5
        and 'Minor Print Level' < 5, no monitoring information is
        produced.

    'Optimality Tolerance' : float
        Default = epsilon_R^0.8

        The argument r (epsilon_R <= r < 1) specifies the accuracy to
        which you wish the final iterate to approximate a solution of
        the problem.
        Broadly speaking, r indicates the number of correct figures
        desired in the objective function at the solution.
        For example, if r is 10^(-6) and ``nlp_multistart_sqp``
        terminates successfully, the final value of F should have
        approximately six correct figures.
        If r < epsilon_r or r >= 1, the default value is used.

        ``nlp_multistart_sqp`` will terminate successfully if the
        iterative sequence of x values is judged to have converged and
        the final point satisfies the first-order Kuhn--Tucker
        conditions (see [Overview]).
        The sequence of iterates is considered to have converged at x if

            alpha||p|| <= sqrt(r)(1+||x||),

        where p is the search direction and alpha the step length from
        (3).
        An iterate is considered to satisfy the first-order conditions
        for a minimum if

            || Z^Tg_FR|| <= sqrt(r)(1+max(1+|F(x)|,||g_FR||))

        and

            |res_j| <= ftol for all j,

        where Z^Tg_FR is the projected gradient (see [Overview]), g_FR
        is the gradient of F(x) with respect to the free variables,
        res_j is the violation of the jth active nonlinear constraint,
        and ftol is the 'Nonlinear Feasibility Tolerance'.

    'Start Objective Check At Variable' : int
        Default = 1

        These keywords take effect only if 'Verify Level' > 0.
        They may be used to control the verification of gradient
        elements computed by `objfun` and/or Jacobian elements computed
        by `confun`.
        For example, if the first 30 elements of the objective gradient
        appeared to be correct in an earlier run, so that only element
        31 remains questionable, it is reasonable to specify 'Start
        Objective Check At Variable' = 31.
        If the first 30 variables appear linearly in the objective, so
        that the corresponding gradient elements are constant, the above
        choice would also be appropriate.

        If i_{2m-1} <= 0 or i_{2m-1} > min(n, i_{2m}), the default value
        is used, for m = 1,2,...,2.
        If i_{2m} <= 0 or i_{2m} > n, the default value is used, for m =
        1,2,...,2.

    'Stop Objective Check At Variable' : int
        Default = n

        These keywords take effect only if 'Verify Level' > 0.
        They may be used to control the verification of gradient
        elements computed by `objfun` and/or Jacobian elements computed
        by `confun`.
        For example, if the first 30 elements of the objective gradient
        appeared to be correct in an earlier run, so that only element
        31 remains questionable, it is reasonable to specify 'Start
        Objective Check At Variable' = 31.
        If the first 30 variables appear linearly in the objective, so
        that the corresponding gradient elements are constant, the above
        choice would also be appropriate.

        If i_{2m-1} <= 0 or i_{2m-1} > min(n, i_{2m}), the default value
        is used, for m = 1,2,...,2.
        If i_{2m} <= 0 or i_{2m} > n, the default value is used, for m =
        1,2,...,2.

    'Start Constraint Check At Variable' : int
        Default = 1

        These keywords take effect only if 'Verify Level' > 0.
        They may be used to control the verification of gradient
        elements computed by `objfun` and/or Jacobian elements computed
        by `confun`.
        For example, if the first 30 elements of the objective gradient
        appeared to be correct in an earlier run, so that only element
        31 remains questionable, it is reasonable to specify 'Start
        Objective Check At Variable' = 31.
        If the first 30 variables appear linearly in the objective, so
        that the corresponding gradient elements are constant, the above
        choice would also be appropriate.

        If i_{2m-1} <= 0 or i_{2m-1} > min(n, i_{2m}), the default value
        is used, for m = 1,2,...,2.
        If i_{2m} <= 0 or i_{2m} > n, the default value is used, for m =
        1,2,...,2.

    'Stop Constraint Check At Variable' : int
        Default = n

        These keywords take effect only if 'Verify Level' > 0.
        They may be used to control the verification of gradient
        elements computed by `objfun` and/or Jacobian elements computed
        by `confun`.
        For example, if the first 30 elements of the objective gradient
        appeared to be correct in an earlier run, so that only element
        31 remains questionable, it is reasonable to specify 'Start
        Objective Check At Variable' = 31.
        If the first 30 variables appear linearly in the objective, so
        that the corresponding gradient elements are constant, the above
        choice would also be appropriate.

        If i_{2m-1} <= 0 or i_{2m-1} > min(n, i_{2m}), the default value
        is used, for m = 1,2,...,2.
        If i_{2m} <= 0 or i_{2m} > n, the default value is used, for m =
        1,2,...,2.

    'Step Limit' : float
        Default = 2.0

        If r > 0,r specifies the maximum change in variables at the
        first step of the linesearch.
        In some cases, such as F(x) = ae^(bx) or F(x) = ax^b, even a
        moderate change in the elements of x can lead to floating-point
        overflow.
        The argument r is therefore used to encourage evaluation of the
        problem functions at meaningful points.
        Given any major iterate x, the first point x~ at which F and c
        are evaluated during the linesearch is restricted so that

            ||x~-x||_2 <= r(1+||x||_2).

        The linesearch may go on and evaluate F and c at points further
        from x if this will result in a lower value of the merit
        function (indicated by L at the end of each line of output
        produced by the major iterations; see [Further Comments]). If L
        is printed for most of the iterations, r should be set to a
        larger value.

        Wherever possible, upper and lower bounds on x should be used to
        prevent evaluation of nonlinear functions at wild values.
        The default value 'Step Limit' = 2.0 should not affect progress
        on well-behaved functions, but values such as 0.1 or 0.01 may be
        helpful when rapidly varying functions are present.
        If a small value of 'Step Limit' is selected, a good starting
        point may be required.
        An important application is to the class of nonlinear least
        squares problems.
        If r <= 0, the default value is used.

    'Verify Level' : int
        Default = 0

        These keywords refer to finite difference checks on the gradient
        elements computed by `objfun` and `confun`.
        The possible choices for i are as follows:

        [table omitted]

        It is possible to specify 'Verify Level' = 0 to 3 in several
        ways.
        For example, the nonlinear objective gradient (if any) will be
        verified if either 'Verify Objective Gradients' or 'Verify
        Level' = 1 is specified.
        The constraint gradients will be verified if 'Verify' = 'YES' or
        'Verify Level' = 2 or 'Verify' is specified.
        Similarly, the objective and the constraint gradients will be
        verified if 'Verify' = 'YES' or 'Verify Level' = 3 or 'Verify'
        is specified.

        If 0 <= i <= 3, gradients will be verified at the first point
        that satisfies the linear constraints and bounds.

        If i = 0, only a 'cheap' test will be performed, requiring one
        call to `objfun` and (if appropriate) one call to `confun`.

        If 1 <= i <= 3, a more reliable (but more expensive) check will
        be made on individual gradient elements, within the ranges
        specified by the 'Start Constraint Check At Variable' and 'Stop
        Constraint Check At Variable' keywords.
        A result of the form OK or BAD? is printed by
        ``nlp_multistart_sqp`` to indicate whether or not each element
        appears to be correct.
        If a gradient element is determined to be extremely poor (i.e.,
        if it appears to have no significant digits of accuracy at all)
        then ``nlp_multistart_sqp`` will also exit with an error
        indicator in argument errno.

        If 10 <= i <= 13, the action is the same as for i-10, except
        that it will take place at the user-specified initial value of
        x.

        If i < -1 or 4 <= i <= 9 or i > 13, the default value is used.

        We suggest that 'Verify Level' = 3 be used whenever a new
        function function is being developed.

    'Verify' : valueless
        These keywords refer to finite difference checks on the gradient
        elements computed by `objfun` and `confun`.
        The possible choices for i are as follows:

        [table omitted]

        It is possible to specify 'Verify Level' = 0 to 3 in several
        ways.
        For example, the nonlinear objective gradient (if any) will be
        verified if either 'Verify Objective Gradients' or 'Verify
        Level' = 1 is specified.
        The constraint gradients will be verified if 'Verify' = 'YES' or
        'Verify Level' = 2 or 'Verify' is specified.
        Similarly, the objective and the constraint gradients will be
        verified if 'Verify' = 'YES' or 'Verify Level' = 3 or 'Verify'
        is specified.

        If 0 <= i <= 3, gradients will be verified at the first point
        that satisfies the linear constraints and bounds.

        If i = 0, only a 'cheap' test will be performed, requiring one
        call to `objfun` and (if appropriate) one call to `confun`.

        If 1 <= i <= 3, a more reliable (but more expensive) check will
        be made on individual gradient elements, within the ranges
        specified by the 'Start Constraint Check At Variable' and 'Stop
        Constraint Check At Variable' keywords.
        A result of the form OK or BAD? is printed by
        ``nlp_multistart_sqp`` to indicate whether or not each element
        appears to be correct.
        If a gradient element is determined to be extremely poor (i.e.,
        if it appears to have no significant digits of accuracy at all)
        then ``nlp_multistart_sqp`` will also exit with an error
        indicator in argument errno.

        If 10 <= i <= 13, the action is the same as for i-10, except
        that it will take place at the user-specified initial value of
        x.

        If i < -1 or 4 <= i <= 9 or i > 13, the default value is used.

        We suggest that 'Verify Level' = 3 be used whenever a new
        function function is being developed.

    'Verify Constraint Gradients' : valueless
        These keywords refer to finite difference checks on the gradient
        elements computed by `objfun` and `confun`.
        The possible choices for i are as follows:

        [table omitted]

        It is possible to specify 'Verify Level' = 0 to 3 in several
        ways.
        For example, the nonlinear objective gradient (if any) will be
        verified if either 'Verify Objective Gradients' or 'Verify
        Level' = 1 is specified.
        The constraint gradients will be verified if 'Verify' = 'YES' or
        'Verify Level' = 2 or 'Verify' is specified.
        Similarly, the objective and the constraint gradients will be
        verified if 'Verify' = 'YES' or 'Verify Level' = 3 or 'Verify'
        is specified.

        If 0 <= i <= 3, gradients will be verified at the first point
        that satisfies the linear constraints and bounds.

        If i = 0, only a 'cheap' test will be performed, requiring one
        call to `objfun` and (if appropriate) one call to `confun`.

        If 1 <= i <= 3, a more reliable (but more expensive) check will
        be made on individual gradient elements, within the ranges
        specified by the 'Start Constraint Check At Variable' and 'Stop
        Constraint Check At Variable' keywords.
        A result of the form OK or BAD? is printed by
        ``nlp_multistart_sqp`` to indicate whether or not each element
        appears to be correct.
        If a gradient element is determined to be extremely poor (i.e.,
        if it appears to have no significant digits of accuracy at all)
        then ``nlp_multistart_sqp`` will also exit with an error
        indicator in argument errno.

        If 10 <= i <= 13, the action is the same as for i-10, except
        that it will take place at the user-specified initial value of
        x.

        If i < -1 or 4 <= i <= 9 or i > 13, the default value is used.

        We suggest that 'Verify Level' = 3 be used whenever a new
        function function is being developed.

    'Verify Gradients' : valueless
        These keywords refer to finite difference checks on the gradient
        elements computed by `objfun` and `confun`.
        The possible choices for i are as follows:

        [table omitted]

        It is possible to specify 'Verify Level' = 0 to 3 in several
        ways.
        For example, the nonlinear objective gradient (if any) will be
        verified if either 'Verify Objective Gradients' or 'Verify
        Level' = 1 is specified.
        The constraint gradients will be verified if 'Verify' = 'YES' or
        'Verify Level' = 2 or 'Verify' is specified.
        Similarly, the objective and the constraint gradients will be
        verified if 'Verify' = 'YES' or 'Verify Level' = 3 or 'Verify'
        is specified.

        If 0 <= i <= 3, gradients will be verified at the first point
        that satisfies the linear constraints and bounds.

        If i = 0, only a 'cheap' test will be performed, requiring one
        call to `objfun` and (if appropriate) one call to `confun`.

        If 1 <= i <= 3, a more reliable (but more expensive) check will
        be made on individual gradient elements, within the ranges
        specified by the 'Start Constraint Check At Variable' and 'Stop
        Constraint Check At Variable' keywords.
        A result of the form OK or BAD? is printed by
        ``nlp_multistart_sqp`` to indicate whether or not each element
        appears to be correct.
        If a gradient element is determined to be extremely poor (i.e.,
        if it appears to have no significant digits of accuracy at all)
        then ``nlp_multistart_sqp`` will also exit with an error
        indicator in argument errno.

        If 10 <= i <= 13, the action is the same as for i-10, except
        that it will take place at the user-specified initial value of
        x.

        If i < -1 or 4 <= i <= 9 or i > 13, the default value is used.

        We suggest that 'Verify Level' = 3 be used whenever a new
        function function is being developed.

    'Verify Objective Gradients' : valueless
        These keywords refer to finite difference checks on the gradient
        elements computed by `objfun` and `confun`.
        The possible choices for i are as follows:

        [table omitted]

        It is possible to specify 'Verify Level' = 0 to 3 in several
        ways.
        For example, the nonlinear objective gradient (if any) will be
        verified if either 'Verify Objective Gradients' or 'Verify
        Level' = 1 is specified.
        The constraint gradients will be verified if 'Verify' = 'YES' or
        'Verify Level' = 2 or 'Verify' is specified.
        Similarly, the objective and the constraint gradients will be
        verified if 'Verify' = 'YES' or 'Verify Level' = 3 or 'Verify'
        is specified.

        If 0 <= i <= 3, gradients will be verified at the first point
        that satisfies the linear constraints and bounds.

        If i = 0, only a 'cheap' test will be performed, requiring one
        call to `objfun` and (if appropriate) one call to `confun`.

        If 1 <= i <= 3, a more reliable (but more expensive) check will
        be made on individual gradient elements, within the ranges
        specified by the 'Start Constraint Check At Variable' and 'Stop
        Constraint Check At Variable' keywords.
        A result of the form OK or BAD? is printed by
        ``nlp_multistart_sqp`` to indicate whether or not each element
        appears to be correct.
        If a gradient element is determined to be extremely poor (i.e.,
        if it appears to have no significant digits of accuracy at all)
        then ``nlp_multistart_sqp`` will also exit with an error
        indicator in argument errno.

        If 10 <= i <= 13, the action is the same as for i-10, except
        that it will take place at the user-specified initial value of
        x.

        If i < -1 or 4 <= i <= 9 or i > 13, the default value is used.

        We suggest that 'Verify Level' = 3 be used whenever a new
        function function is being developed.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `nb` = *<value>* and `npts` = *<value>*.

            Constraint: 1 <= `nb` <= `npts`.

        (`errno` 1)
            On entry, `bl`[i-1] > `bu`[i-1]: i = *<value>*.

            Constraint: `bl`[i-1] <= `bu`[i-1], for all i.

        (`errno` 1)
            On entry, `ncnln` = *<value>*.

            Constraint: `ncnln` >= 0.

        (`errno` 1)
            On entry, nclin = *<value>*.

            Constraint: nclin >= 0.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` > 0.

        (`errno` 1)
            On entry, ldclda = *<value>*, `n` = *<value>*, nclin =
            *<value>* and `ncnln` = *<value>*.

            Constraint: ldclda >= `n`+nclin+`ncnln`.

        (`errno` 1)
            On entry, `ncnln` > 0, sdcjac = *<value>* and `n` =
            *<value>*.

            Constraint: if `ncnln` > 0, sdcjac >= `n`.

        (`errno` 1)
            On entry, listat = *<value>*, `n` = *<value>*, nclin =
            *<value>* and `ncnln` = *<value>*.

            Constraint: listat >= `n`+nclin+`ncnln`.

        (`errno` 2)
            No solution obtained. Linear constraints may be infeasible.

        (`errno` 3)
            No solution obtained. Nonlinear constraints may be
            infeasible.

        (`errno` 4)
            No solution obtained. Many potential solutions reach
            iteration limit.

        (`errno` 7)
            User-supplied derivatives probably wrong.

        (`errno` 10)
            Failed to initialize option arrays.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 8)
            Only *<value>* solutions obtained.

    NagCallbackTerminateWarning
        (`errno` 9)
            User terminated computation from `start` procedure: `mode` =
            *<value>*.

    References
    ----------
    Dennis, J E Jr and Moré, J J, 1977, `Quasi-Newton methods,
    motivation and theory`, SIAM Rev. (19), 46--89

    Dennis, J E Jr and Schnabel, R B, 1981, `A new derivation of
    symmetric positive-definite secant updates`, nonlinear programming,
    (eds O L Mangasarian, R R Meyer and S M Robinson) (4), 167--199,
    Academic Press

    Dennis, J E Jr and Schnabel, R B, 1983, `Numerical Methods for
    Unconstrained Optimization and Nonlinear Equations`, Prentice--Hall

    Fletcher, R, 1987, `Practical Methods of Optimization`, (2nd
    Edition), Wiley

    Gill, P E, Hammarling, S, Murray, W, Saunders, M A and Wright, M H,
    1986, `Users' guide for LSSOL (Version 1.0)`, Report SOL 86-1,
    Department of Operations Research, Stanford University

    Gill, P E, Murray, W, Saunders, M A and Wright, M H, 1984, `Users'
    guide for SOL/QPSOL version 3.2`, Report SOL, 84--5, Department of
    Operations Research, Stanford University

    Gill, P E, Murray, W, Saunders, M A and Wright, M H, 1986, `Users'
    guide for NPSOL (Version 4.0): a Fortran package for nonlinear
    programming`, Report SOL 86-2, Department of Operations Research,
    Stanford University

    Gill, P E, Murray, W, Saunders, M A and Wright, M H, 1986, `Some
    theoretical properties of an augmented Lagrangian merit function`,
    Report SOL, 86--6R, Department of Operations Research, Stanford
    University

    Gill, P E, Murray, W and Wright, M H, 1981, `Practical
    Optimization`, Academic Press

    Powell, M J D, 1974, `Introduction to constrained optimization`,
    Numerical Methods for Constrained Optimization, (eds P E Gill and W
    Murray), 1--28, Academic Press

    Powell, M J D, 1983, `Variable metric methods in constrained
    optimization`, Mathematical Programming: the State of the Art, (eds
    A Bachem, M Grötschel and B Korte), 288--311, Springer--Verlag
    """

    fun_name = 'naginterfaces.library.glopt.nlp_multistart_sqp'

    cb_confun_name = fun_name + ':confun'

    def cb_confun(mode, ncnln, n, needc, x, c, cjsl, nstate, data=None):
        """Wrapping user's confun. Called as base confun."""
        if data is not None:
            _c_confun, _cjsl_confun = confun(
                mode,
                needc,
                x,
                cjsl,
                nstate,
                data,
            )
        else:
            _c_confun, _cjsl_confun = confun(
                mode,
                needc,
                x,
                cjsl,
                nstate,
            )
        c[:] = _utils._to_ndarray(
            {'fun_name': cb_confun_name, 'returned': 'c'},
            _c_confun,
            copy=False,
            exp_shape=(ncnln,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _cjsl_confun_np = _utils._to_ndarray(
            {'fun_name': cb_confun_name, 'returned': 'cjsl'},
            _cjsl_confun,
            copy=False,
            exp_shape=(ncnln, n),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _utils._check_callback_sorder(
            {'fun_name': cb_confun_name},
            ((_cjsl_confun_np, 'cjsl'),),
            wrapper_sorder=_m_sorder,
            spiked_sorder=spiked_sorder,
        )
        cjsl[:, :] = _cjsl_confun_np

    cb_objfun_name = fun_name + ':objfun'

    def cb_objfun(mode, n, x, objgrd, nstate, data=None):
        """Wrapping user's objfun. Called as base objfun."""
        if data is not None:
            _objf_objfun, _objgrd_objfun = objfun(
                mode,
                x,
                objgrd,
                nstate,
                data,
            )
        else:
            _objf_objfun, _objgrd_objfun = objfun(
                mode,
                x,
                objgrd,
                nstate,
            )
        objgrd[:] = _utils._to_ndarray(
            {'fun_name': cb_objfun_name, 'returned': 'objgrd'},
            _objgrd_objfun,
            copy=False,
            exp_shape=(n,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _objf_objfun_base = _objf_objfun
        return _objf_objfun_base

    cb_start_name = fun_name + ':start'

    def cb_start(npts, quas, n, repeat, bl, bu, data=None):
        """Wrapping user's start. Called as base start."""
        if data is not None:
            _quas_start = start(
                quas,
                repeat,
                bl,
                bu,
                data,
            )
        else:
            _quas_start = start(
                quas,
                repeat,
                bl,
                bu,
            )
        _quas_start_np = _utils._to_ndarray(
            {'fun_name': cb_start_name, 'returned': 'quas'},
            _quas_start,
            copy=False,
            exp_shape=(n, npts),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _utils._check_callback_sorder(
            {'fun_name': cb_start_name},
            ((_quas_start_np, 'quas'),),
            wrapper_sorder=_m_sorder,
            spiked_sorder=spiked_sorder,
        )
        quas[:, :] = _quas_start_np

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_a, 'a'),),
        spiked_sorder=spiked_sorder,
    )
    if a is None:
        _a = None
    _nclin = _b_utils._EngineIntScalarType(
        0 if a is None else _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, allow_none=True)[0], {'fun_name': fun_name, 'entity': '_nclin'},
    ).to_ctypes()
    _bl = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'bl'},
        bl,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _bu = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'bu'},
        bu,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b_utils._check_callable(confun, {'fun_name': fun_name, 'entity': 'confun'}, allow_none=True)
    _confun = None if confun is None else cb_confun
    _b_utils._check_callable(objfun, {'fun_name': fun_name, 'entity': 'objfun'})
    _objfun = cb_objfun
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _nb = _b_utils._EngineIntScalarType(
        nb, {'fun_name': fun_name, 'entity': 'nb'},
    ).to_ctypes()
    _nb_py = _b_utils._EngineIntScalarType.to_py(_nb)
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py, _nb_py),
        sorder=_m_sorder,
    )
    _b_utils._check_callable(start, {'fun_name': fun_name, 'entity': 'start'}, allow_none=True)
    _start = None if start is None else cb_start
    _objf = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nb_py,),
    )
    _objgrd = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py, _nb_py),
        sorder=_m_sorder,
    )
    _itera = _b_utils._EngineIntArrayType.empty_ndarray(
        (_nb_py,),
    )
    _ncnln = _b_utils._EngineIntScalarType(
        ncnln, {'fun_name': fun_name, 'entity': 'ncnln'},
    ).to_ctypes()
    _ncnln_py = _b_utils._EngineIntScalarType.to_py(_ncnln)
    if _ncnln_py != 0:
        _c = _b_utils._EngineFloat64ArrayType.empty_ndarray(
            (_ncnln_py, _nb_py),
            sorder=_m_sorder,
        )
    else:
        _c = None
    if _ncnln_py != 0:
        if _ncnln_py > 0:
            _cjac_dim2 = _n_py
        else:
            _cjac_dim2 = 0
        exp_shape_cjac = (_ncnln_py, _cjac_dim2, _nb_py)
        _cjac = _b_utils._EngineFloat64ArrayType.empty_ndarray(
            exp_shape_cjac,
            sorder=_m_sorder,
        )
    else:
        _cjac = None
    _r = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py, _n_py, _nb_py),
        sorder=_m_sorder,
    )
    _nclin_py = _b_utils._EngineIntScalarType.to_py(_nclin)
    _clamda = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py+_nclin_py+_ncnln_py, _nb_py),
        sorder=_m_sorder,
    )
    _istate = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n_py+_nclin_py+_ncnln_py, _nb_py),
        sorder=_m_sorder,
    )
    _info = _b_utils._EngineIntArrayType.empty_ndarray(
        (_nb_py,),
    )
    try:
        _b_glopt.nlp_multistart_sqp(
            _n,
            _nclin,
            _ncnln,
            _a,
            _bl,
            _bu,
            _confun,
            _objfun,
            npts,
            _x,
            _start,
            repeat,
            _nb,
            _objf,
            _objgrd,
            _itera,
            _c,
            _cjac,
            _r,
            _clamda,
            _istate,
            comm,
            _info,
            data,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _x_py = _x
    _objf_py = _objf
    _objgrd_py = _objgrd
    _itera_py = _itera
    if _ncnln_py != 0:
        _c_py = _c
    else:
        _c_py = None
    if _ncnln_py != 0:
        _cjac_py = _cjac
    else:
        _cjac_py = None
    _r_py = _r
    _clamda_py = _clamda
    _istate_py = _istate
    _info_py = _info
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'objf', 'objgrd', 'itera', 'c', 'cjac', 'r', 'clamda', 'istate', 'info'],
    )(
        _x_py,
        _objf_py,
        _objgrd_py,
        _itera_py,
        _c_py,
        _cjac_py,
        _r_py,
        _clamda_py,
        _istate_py,
        _info_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def nlp_multistart_sqp_lsq(ncnln, bl, bu, y, objfun, npts, nb, comm, a=None, confun=None, start=None, repeat1=False, data=None, io_manager=None):
    u"""
    Global optimization of a sum of squares problem using multi-start,
    nonlinear constraints.

    Note: this function uses optional algorithmic parameters, see also:
    ``optset``, ``opt.lsq_gencon_deriv``.

    ``nlp_multistart_sqp_lsq`` is designed to find the global minimum of
    an arbitrary smooth sum of squares function subject to constraints
    (which may include simple bounds on the variables, linear
    constraints and smooth nonlinear constraints) by generating a number
    of different starting points and performing a local search from each
    using sequential quadratic programming.

    For full information please refer to the NAG Library document for
    e05us

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e05/e05usf.html

    Parameters
    ----------
    ncnln : int
        n_N, the number of nonlinear constraints.

    bl : float, array-like, shape (n+nclin+`ncnln`)
        `bl` must contain the lower bounds for all the constraints.

    bu : float, array-like, shape (n+nclin+`ncnln`)
        `bu` must contain the upper bounds for all the constraints.

    y : float, array-like, shape (m)
        The coefficients of the constant vector y of the objective
        function.

    objfun : callable f = objfun(mode, needfi, x, fjsl, nstate,
        data=None)
        `objfun` must calculate either the ith element of the vector
        f(x) = (f_1(x),f_2(x),...,f_m(x))^T or all m elements of f(x)
        and (optionally) its Jacobian (= df/dx) for a specified
        n-element vector x.

        Parameters
        ~~~~~~~~~~
        mode : int
            Indicates which values must be assigned during each call of
            `objfun`. Only the following values need be assigned:

            `mode` = 0 and `needfi` = i, where i > 0

                `f`[i-1].

            `mode` = 0 and `needfi` < 0

                `f`.

            `mode` = 1 and `needfi` < 0

                All available elements of `fjsl`.

            `mode` = 2 and `needfi` < 0

                `f` and all available elements of `fjsl`.

        needfi : int
            If `needfi` = i > 0, only the ith element of f(x) needs to
            be evaluated at x; the remaining elements need not be set.
            This can result in significant computational savings when
            m≫n.

        x : float, ndarray, shape (n)
            x, the vector of variables at which the objective function
            and/or all available elements of its gradient are to be
            evaluated.

        fjsl : float, ndarray, shape (m, n), to be modified in place
            `fjsl` may be regarded as a two-dimensional 'slice' of the
            three-dimensional array `fjac` of
            ``nlp_multistart_sqp_lsq``.

            `On entry`: is set to a special value.

            `On exit`: if `mode` = 1 or 2 and `needfi` < 0, the ith row
            of `fjsl` must contain the available elements of the vector
            nabla f_i given by

                nabla f_i = (df_i/dx_1,df_i/dx_2,...,df_i/dx_n)^T,

            evaluated at the point x. See also the argument `nstate`.

        nstate : int
            If `nstate` = 1 then ``nlp_multistart_sqp_lsq`` is calling
            `objfun` for the first time on the current local
            optimization problem. This argument setting allows you to
            save computation time if certain data must be read or
            calculated only once.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        f : float, array-like, shape (m)
            If `mode` = 0 and `needfi` = i > 0, `f`[i-1] must contain
            the value of f_i at x.

            If `mode` = 0 or 2 and `needfi` < 0, `f`[i-1] must contain
            the value of f_i at x, for i = 1,2,...,m.

    npts : int
        The number of different starting points to be generated and
        used. The more points used, the more likely that the best
        returned solution will be a global minimum.

    nb : int
        The number of solutions to be returned. The function saves up to
        `nb` local minima ordered by increasing value of the final
        objective function. If the defining criterion for 'best
        solution' is only that the value of the objective function is as
        small as possible then `nb` should be set to 1. However, if you
        want to look at other solutions that may have desirable
        properties then setting `nb` > 1 will produce `nb` local minima,
        ordered by increasing value of their objective functions at the
        minima.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``optset``.

    a : None or float, array-like, shape (nclin, :), optional
        Note: the required extent for this argument in dimension 2 is
        determined as follows: if nclin>0: n; otherwise: 1.

        The matrix A_L of general linear constraints in (1). That is,
        the ith row contains the coefficients of the ith general linear
        constraint, for i = 1,2,...,nclin.

    confun : None or callable c = confun(mode, needc, x, cjsl, nstate,
        data=None), optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `confun` must calculate the vector c(x) of nonlinear constraint
        functions and (optionally) its Jacobian (= dc/dx) for a
        specified n-element vector x. If there are no nonlinear
        constraints (i.e., `ncnln` = 0), `confun` will never be called
        by ``nlp_multistart_sqp_lsq`` and `confun` may be None. If there
        are nonlinear constraints, the first call to `confun` will occur
        before the first call to `objfun`.

        Parameters
        ~~~~~~~~~~
        mode : int
            Indicates which values must be assigned during each call of
            `confun`. Only the following values need be assigned, for
            each value of i such that `needc`[i-1] > 0:

            `mode` = 0

                `c`[i-1], the ith nonlinear constraint.

            `mode` = 1

                All available elements in the ith row of `cjsl`.

            `mode` = 2

                `c`[i-1] and all available elements in the ith row of
                `cjsl`.

        needc : int, ndarray, shape (ncnln)
            The indices of the elements of `c` and/or `cjsl` that must
            be evaluated by `confun`. If `needc`[i-1] > 0, `c`[i-1]
            and/or the available elements of the ith row of `cjsl` (see
            argument `mode`) must be evaluated at x.

        x : float, ndarray, shape (n)
            x, the vector of variables at which the constraint functions
            and/or the available elements of the constraint Jacobian are
            to be evaluated.

        cjsl : float, ndarray, shape (ncnln, n), to be modified in place
            `cjsl` may be regarded as a two-dimensional 'slice' of the
            three-dimensional array `cjac` of
            ``nlp_multistart_sqp_lsq``.

            `On entry`: unless 'Derivative Level' = 2 or 3, the elements
            of `cjsl` are set to special values which enable
            ``nlp_multistart_sqp_lsq`` to detect whether they are
            changed by `confun`.

            `On exit`: if `needc`[i-1] > 0 and `mode` = 1 or 2, the ith
            row of `cjsl` must contain the available elements of the
            vector nabla c_i given by

                nabla c_i = (dc_i/dx_1,dc_i/dx_2,...,dc_i/dx_n)^T,

            where dc_i/dx_j is the partial derivative of the ith
            constraint with respect to the jth variable, evaluated at
            the point x. See also the argument `nstate`. The remaining
            rows of `cjsl`, corresponding to non-positive elements of
            `needc`, need not be set.

            If all elements of the constraint Jacobian are known (i.e.,
            'Derivative Level' = 2 or 3; note the default is 'Derivative
            Level' = 3), any constant elements may be assigned to `cjsl`
            one time only at the start of each local optimization.

            An element of `cjsl` that is not subsequently assigned in
            `confun` will retain its initial value throughout the local
            optimization.

            Constant elements may be loaded into `cjsl` during the first
            call to `confun` for the local optimization (signalled by
            the value `nstate` = 1).

            The ability to preload constants is useful when many
            Jacobian elements are identically zero, in which case `cjsl`
            may be initialized to zero and nonzero elements may be reset
            by `confun`.

            Note that constant nonzero elements do affect the values of
            the constraints.

            Thus, if `cjsl`[i-1,j-1] is set to a constant value, it need
            not be reset in subsequent calls to `confun`, but the value
            `cjsl`[i-1,j-1]*`x`[j-1] must nonetheless be added to
            `c`[i-1].

            For example, if `cjsl`[0,0] = 2 and `cjsl`[0,1] = -5 then
            the term 2*`x`[0]-5*`x`[1] must be included in the
            definition of `c`[0].

            It must be emphasized that, if 'Derivative Level' = 0 or 1,
            unassigned elements of `cjsl` are not treated as constant;
            they are estimated by finite differences, at nontrivial
            expense.

            If you do not supply a value for the option 'Difference
            Interval', an interval for each element of x is computed
            automatically at the start of each local optimization.

            The automatic procedure can usually identify constant
            elements of `cjsl`, which are then computed once only by
            finite differences.

        nstate : int
            If `nstate` = 1 then ``nlp_multistart_sqp_lsq`` is calling
            `confun` for the first time on the current local
            optimization problem. This argument setting allows you to
            save computation time if certain data must be read or
            calculated only once.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        c : float, array-like, shape (ncnln)
            If `needc`[i-1] > 0 and `mode` = 0 or 2, `c`[i-1] must
            contain the value of c_i(x). The remaining elements of `c`,
            corresponding to the non-positive elements of `needc`, need
            not be set.

    start : None or callable start(quas, repeat1, bl, bu, data=None),
        optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `start` must calculate the `npts` starting points to be used by
        the local optimizer.

        If you do not wish to write a function specific to your problem
        then None may be used. In this case, a default function uses the
        NAG quasi-random number generators to distribute starting points
        uniformly across the domain.

        It is affected by the value of `repeat1`.

        Parameters
        ~~~~~~~~~~
        quas : float, ndarray, shape (n, npts), to be modified in place
            `On entry`: all elements of `quas` will have been set to
            zero, so only nonzero values need be set subsequently.

            `On exit`: must contain the starting points for the npts
            local minimizations, i.e., `quas`[j-1,i-1] must contain the
            jth component of the ith starting point.

        repeat1 : bool
            Specifies whether a repeatable or non-repeatable sequence of
            points are to be generated.

        bl : float, ndarray, shape (n)
            The lower bounds on the variables. These may be used to
            ensure that the starting points generated in some sense
            'cover' the region, but there is no requirement that a
            starting point be feasible.

        bu : float, ndarray, shape (n)
            The upper bounds on the variables. (See `bl`.)

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

    repeat1 : bool, optional
        Is passed as an argument to `start` and may be used to
        initialize a random number generator to a repeatable, or
        non-repeatable, sequence. See [Further Comments] for more
        detail.

    data : arbitrary, optional
        User-communication data for callback functions.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    x : float, ndarray, shape (n, `nb`)
        `x`[j-1,i-1] contains the final estimate of the ith solution,
        for j = 1,2,...,n.

    objf : float, ndarray, shape (`nb`)
        `objf`[i-1] contains the value of the objective function at the
        final iterate for the ith solution.

    f : float, ndarray, shape (m, `nb`)
        `f`[j-1,i-1] contains the value of the jth function f_j at the
        final iterate, for j = 1,2,...,m, for the ith solution, for i =
        1,2,...,`nb`.

    fjac : float, ndarray, shape (m, n, `nb`)
        For the ith returned solution, the Jacobian matrix of the
        functions f_1,f_2,...,f_m at the final iterate, i.e.,
        `fjac`[k-1,j-1,i-1] contains the partial derivative of the kth
        function with respect to the jth variable, for i = 1,2,...,`nb`,
        for j = 1,2,...,n, for k = 1,2,...,m. (See also the discussion
        of argument `fjsl` under `objfun`.)

    itera : int, ndarray, shape (`nb`)
        `itera`[i-1] contains the number of major iterations performed
        to obtain the ith solution. If less than `nb` solutions are
        returned then `itera`[`nb`-1] contains the number of starting
        points that have resulted in a converged solution. If this is
        close to `npts` then this might be indicative that fewer than
        `nb` local minima exist.

    c : None or float, ndarray, shape (`ncnln`, `nb`)
        If `ncnln` > 0, `c`[j-1,i-1] contains the value of the jth
        nonlinear constraint function c_j at the final iterate, for the
        ith solution, for j = 1,2,...,`ncnln`.

        If `ncnln` = 0, `c` is returned as None.

    cjac : None or float, ndarray, shape (`ncnln`, :, `nb`)
        If `ncnln` > 0, `cjac` contains the Jacobian matrices of the
        nonlinear constraint functions at the final iterate for each of
        the returned solutions, i.e., `cjac`[k-1,j-1,i-1] contains the
        partial derivative of the kth constraint function with respect
        to the jth variable, for j = 1,2,...,n, for k = 1,2,...,`ncnln`,
        for the ith solution. (See the discussion of argument `cjsl`
        under `confun`.)

        If `ncnln` = 0, `cjac` is returned as None.

    clamda : float, ndarray, shape (n+nclin+`ncnln`, `nb`)
        The values of the QP multipliers from the last QP subproblem
        solved for the ith solution. `clamda`[j-1,i-1] should be
        non-negative if `istate`[j-1,i-1] = 1 and non-positive if
        `istate`[j-1,i-1] = 2.

    istate : int, ndarray, shape (n+nclin+`ncnln`, `nb`)
        `istate`[j-1,i-1] contains the status of the constraints in the
        QP working set for the ith solution. The significance of each
        possible value of `istate`[j-1,i-1] is as follows:

        [table omitted]

    info : int, ndarray, shape (`nb`)
        If no exception or warning is raised, `info`[i-1] does not
        contain an error value returned by ``opt.lsq_gencon_deriv``.

        If `errno` = 8 on exit, then not all `nb` solutions have been
        found, and `info`[`nb`-1] contains the number of solutions
        actually found.

    Other Parameters
    ----------------
    'Central Difference Interval' : float
        Default values are computed

        If the algorithm switches to central differences because the
        forward-difference approximation is not sufficiently accurate,
        the value of r is used as the difference interval for every
        element of x.
        The switch to central differences is indicated by ``C`` at the
        end of each line of intermediate printout produced by the major
        iterations (see [Further Comments]).
        The use of finite differences is discussed further under the
        option 'Difference Interval'.

        If you supply a value for this option, a small value between 0.0
        and 1.0 is appropriate.

    'Crash Tolerance' : float
        Default = 0.01

        This value is used when ``nlp_multistart_sqp_lsq`` selects an
        initial working set.
        If 0 <= r <= 1, the initial working set will include (if
        possible) bounds or general inequality constraints that lie
        within r of their bounds.
        In particular, a constraint of the form a_j^Tx >= l will be
        included in the initial working set if |a_j^Tx-l| <= r(1+|l|).
        If r < 0 or r > 1, the default value is used.

    'Defaults' : valueless
        This special keyword is used to reset all options to their
        default values, and any random state stored in the array comm
        will be destroyed.

        Any option value given with this keyword will be ignored.
        This option cannot be queried or got.

    'Derivative Level' : int
        Default = 3

        This argument indicates which derivatives are provided in
        functions `objfun` and `confun`.
        The possible choices for i are the following.

        [table omitted]

        The value i = 3 should be used whenever possible, since
        ``nlp_multistart_sqp_lsq`` is more reliable (and will usually be
        more efficient) when all derivatives are exact.

        If i = 0 or 2, ``nlp_multistart_sqp_lsq`` will estimate the
        unspecified elements of the objective gradient, using finite
        differences.
        The computation of finite difference approximations usually
        increases the total run-time, since a call to `objfun` is
        required for each unspecified element.
        Furthermore, less accuracy can be attained in the solution (see
        Module 8 of Gill `et al.` (1981), for a discussion of limiting
        accuracy).

        If i = 0 or 1, ``nlp_multistart_sqp_lsq`` will approximate
        unspecified elements of the constraint Jacobian.
        One call to `confun` is needed for each variable for which
        partial derivatives are not available.
        For example, if the Jacobian has the form

            [table omitted]

        where '*' indicates an element provided by you and '?' indicates
        an unspecified element, ``nlp_multistart_sqp_lsq`` will call
        `confun` twice: once to estimate the missing element in column
        2, and again to estimate the two missing elements in column 3.
        (Since columns 1 and 4 are known, they require no calls to
        `confun`.)

        At times, central differences are used rather than forward
        differences, in which case twice as many calls to `objfun` and
        `confun` are needed. (The switch to central differences is not
        under your control.)

        If i < 0 or i > 3, the default value is used.

    'Difference Interval' : float
        Default values are computed

        This option defines an interval used to estimate derivatives by
        finite differences in the following circumstances:

        (a) For verifying the objective and/or constraint gradients (see
            the description of the option 'Verify').

        (#) For estimating unspecified elements of the objective
            gradient or the constraint Jacobian.

        In general, a derivative with respect to the jth variable is
        approximated using the interval delta_j, where delta_j =
        r(1+|x^_j|), with x^ the first point feasible with respect to
        the bounds and linear constraints.
        If the functions are well scaled, the resulting derivative
        approximation should be accurate to O(r).
        See Gill `et al.` (1981) for a discussion of the accuracy in
        finite difference approximations.

        If a difference interval is not specified, a finite difference
        interval will be computed automatically for each variable by a
        procedure that requires up to six calls of `confun` and `objfun`
        for each element.
        This option is recommended if the function is badly scaled or
        you wish to have ``nlp_multistart_sqp_lsq`` determine constant
        elements in the objective and constraint gradients (see the
        descriptions of `confun` and `objfun` in [Parameters]).

        If you supply a value for this option, a small value between 0.0
        and 1.0 is appropriate.

    'Feasibility Tolerance' : float
        Default = sqrt(epsilon)

        The scalar r defines the maximum acceptable `absolute`
        violations in linear and nonlinear constraints at a 'feasible'
        point; i.e., a constraint is considered satisfied if its
        violation does not exceed r.
        If r < epsilon or r >= 1, the default value is used.
        Using this keyword sets both options 'Linear Feasibility
        Tolerance' and 'Nonlinear Feasibility Tolerance' to r, if
        epsilon <= r < 1. (Additional details are given under the
        descriptions of these options.)

    'Function Precision' : float
        Default = epsilon^0.9

        This argument defines epsilon_r, which is intended to be a
        measure of the accuracy with which the problem functions F(x)
        and c(x) can be computed.
        If r < epsilon or r >= 1, the default value is used.

        The value of epsilon_r should reflect the relative precision of
        1+|F(x)|; i.e., epsilon_r acts as a relative precision when |F|
        is large, and as an absolute precision when |F| is small.
        For example, if F(x) is typically of order 1000 and the first
        six significant digits are known to be correct, an appropriate
        value for epsilon_r would be 10^(-6).
        In contrast, if F(x) is typically of order 10^(-4) and the first
        six significant digits are known to be correct, an appropriate
        value for epsilon_r would be 10^(-10).
        The choice of epsilon_r can be quite complicated for badly
        scaled problems; see Module 8 of Gill `et al.` (1981) for a
        discussion of scaling techniques.
        The default value is appropriate for most simple functions that
        are computed with full accuracy.
        However, when the accuracy of the computed function values is
        known to be significantly worse than full precision, the value
        of epsilon_r should be large enough so that
        ``nlp_multistart_sqp_lsq`` will not attempt to distinguish
        between function values that differ by less than the error
        inherent in the calculation.

    'Hessian' : str
        Default = 'NO'

        This option controls the contents of the upper triangular matrix
        R (see [Parameters]). ``nlp_multistart_sqp_lsq`` works
        exclusively with the `transformed and reordered` Hessian H_Q
        (6), and hence extra computation is required to form the Hessian
        itself.
        If 'Hessian' = 'NO', r contains the Cholesky factor of the
        transformed and reordered Hessian.
        If 'Hessian' = 'YES', the Cholesky factor of the approximate
        Hessian itself is formed and stored in r.

    'Infinite Bound Size' : float
        Default = 10^20

        This defines the 'infinite' bound infbnd in the definition of
        the problem constraints.
        Any upper bound greater than or equal to infbnd will be regarded
        as infinity (and similarly any lower bound less than or equal to
        -infbnd will be regarded as - infinity).

    'Infinite Step Size' : float
        Default = max(bigbnd, 10^20)

        If r > 0, r specifies the magnitude of the change in variables
        that is treated as a step to an unbounded solution.
        If the change in x during an iteration would exceed the value of
        r, the objective function is considered to be unbounded below in
        the feasible region.
        If r <= 0, the default value is used.

    'Line Search Tolerance' : float
        Default = 0.9

        The value r (0 <= r < 1) controls the accuracy with which the
        step alpha taken during each iteration approximates a minimum of
        the merit function along the search direction (the smaller the
        value of r, the more accurate the linesearch).
        The default value r = 0.9 requests an inaccurate search, and is
        appropriate for most problems, particularly those with any
        nonlinear constraints.

        If there are no nonlinear constraints, a more accurate search
        may be appropriate when it is desirable to reduce the number of
        major iterations -- for example, if the objective function is
        cheap to evaluate, or if a substantial number of derivatives are
        unspecified.
        If r < 0 or r >= 1, the default value is used.

    'Linear Feasibility Tolerance' : float
        Default = sqrt(epsilon)

        The default value of r_2 is epsilon^0.33 if 'Derivative Level' =
        0 or 1, and sqrt(epsilon) otherwise.

        The scalars r_1 and r_2 define the maximum acceptable `absolute`
        violations in linear and nonlinear constraints at a 'feasible'
        point; i.e., a linear constraint is considered satisfied if its
        violation does not exceed r_1, and similarly for a nonlinear
        constraint and r_2.
        If r_m < epsilon or r_m >= 1, the default value is used, for m =
        1,2,...,2.

        On entry to ``nlp_multistart_sqp_lsq``, an iterative procedure
        is executed in order to find a point that satisfies the linear
        constraints and bounds on the variables to within the tolerance
        r_1.
        All subsequent iterates will satisfy the linear constraints to
        within the same tolerance (unless r_1 is comparable to the
        finite difference interval).

        For nonlinear constraints, the feasibility tolerance r_2 defines
        the largest constraint violation that is acceptable at an
        optimal point.
        Since nonlinear constraints are generally not satisfied until
        the final iterate, the value of option 'Nonlinear Feasibility
        Tolerance' acts as a partial termination criterion for the
        iterative sequence generated by ``nlp_multistart_sqp_lsq`` (see
        the discussion of option 'Optimality Tolerance').

        These tolerances should reflect the precision of the
        corresponding constraints.
        For example, if the variables and the coefficients in the linear
        constraints are of order unity, and the latter are correct to
        about 6 decimal digits, it would be appropriate to specify r_1
        as 10^(-6).

    'Nonlinear Feasibility Tolerance' : float
        Default = epsilon^0.33 or sqrt(epsilon)

        The default value of r_2 is epsilon^0.33 if 'Derivative Level' =
        0 or 1, and sqrt(epsilon) otherwise.

        The scalars r_1 and r_2 define the maximum acceptable `absolute`
        violations in linear and nonlinear constraints at a 'feasible'
        point; i.e., a linear constraint is considered satisfied if its
        violation does not exceed r_1, and similarly for a nonlinear
        constraint and r_2.
        If r_m < epsilon or r_m >= 1, the default value is used, for m =
        1,2,...,2.

        On entry to ``nlp_multistart_sqp_lsq``, an iterative procedure
        is executed in order to find a point that satisfies the linear
        constraints and bounds on the variables to within the tolerance
        r_1.
        All subsequent iterates will satisfy the linear constraints to
        within the same tolerance (unless r_1 is comparable to the
        finite difference interval).

        For nonlinear constraints, the feasibility tolerance r_2 defines
        the largest constraint violation that is acceptable at an
        optimal point.
        Since nonlinear constraints are generally not satisfied until
        the final iterate, the value of option 'Nonlinear Feasibility
        Tolerance' acts as a partial termination criterion for the
        iterative sequence generated by ``nlp_multistart_sqp_lsq`` (see
        the discussion of option 'Optimality Tolerance').

        These tolerances should reflect the precision of the
        corresponding constraints.
        For example, if the variables and the coefficients in the linear
        constraints are of order unity, and the latter are correct to
        about 6 decimal digits, it would be appropriate to specify r_1
        as 10^(-6).

    'List' : valueless
        Option 'List' enables printing of each option specification as
        it is supplied. 'Nolist' suppresses this printing.

    'Nolist' : valueless
        Default

        Option 'List' enables printing of each option specification as
        it is supplied. 'Nolist' suppresses this printing.

    'Major Iteration Limit' : int
        Default = max(50, 3*(n+n_L)+10n_N)

        The value of i specifies the maximum number of major iterations
        allowed before termination of the subproblem.
        Setting i = 0 and 'Major Print Level' > 0 means that the
        workspace needed will be computed and printed, but no iterations
        will be performed.
        If i < 0, the default value is used.

    'Iteration Limit' : int
        Default = max(50, 3*(n+n_L)+10n_N)

        The value of i specifies the maximum number of major iterations
        allowed before termination of the subproblem.
        Setting i = 0 and 'Major Print Level' > 0 means that the
        workspace needed will be computed and printed, but no iterations
        will be performed.
        If i < 0, the default value is used.

    'Iters' : int
        Default = max(50, 3*(n+n_L)+10n_N)

        The value of i specifies the maximum number of major iterations
        allowed before termination of the subproblem.
        Setting i = 0 and 'Major Print Level' > 0 means that the
        workspace needed will be computed and printed, but no iterations
        will be performed.
        If i < 0, the default value is used.

    'Itns' : int
        Default = max(50, 3*(n+n_L)+10n_N)

        The value of i specifies the maximum number of major iterations
        allowed before termination of the subproblem.
        Setting i = 0 and 'Major Print Level' > 0 means that the
        workspace needed will be computed and printed, but no iterations
        will be performed.
        If i < 0, the default value is used.

    'Major Print Level' : int
        Default = 0

        The value of i controls the amount of printout produced by the
        major iterations of ``nlp_multistart_sqp_lsq``, as indicated
        below.
        A detailed description of the printed output is given in
        [Further Comments] (summary output at each major iteration and
        the final solution) and [Monitoring Information] (monitoring
        information at each major iteration). (See also the description
        of the option 'Minor Print Level'.)

        The following printout is sent to the file object associated
        with the advisory I/O unit (see
        :class:`~naginterfaces.base.utils.FileObjManager`):

        [table omitted]

        For the other values described below, the arguments used by
        ``opt.nlp1_solve`` are displayed in addition to intermediate and
        final output.

        [table omitted]

        The following printout is sent to the logical unit number
        defined by the option 'Monitoring File':

        [table omitted]

        If 'Major Print Level' >= 5 and the unit number defined by the
        option 'Monitoring File' is the advisory unit number, then the
        summary output for each major iteration is suppressed.

    'Print Level' : int
        Default = 0

        The value of i controls the amount of printout produced by the
        major iterations of ``nlp_multistart_sqp_lsq``, as indicated
        below.
        A detailed description of the printed output is given in
        [Further Comments] (summary output at each major iteration and
        the final solution) and [Monitoring Information] (monitoring
        information at each major iteration). (See also the description
        of the option 'Minor Print Level'.)

        The following printout is sent to the file object associated
        with the advisory I/O unit (see
        :class:`~naginterfaces.base.utils.FileObjManager`):

        [table omitted]

        For the other values described below, the arguments used by
        ``opt.nlp1_solve`` are displayed in addition to intermediate and
        final output.

        [table omitted]

        The following printout is sent to the logical unit number
        defined by the option 'Monitoring File':

        [table omitted]

        If 'Major Print Level' >= 5 and the unit number defined by the
        option 'Monitoring File' is the advisory unit number, then the
        summary output for each major iteration is suppressed.

    'Minor Iteration Limit' : int
        Default = max(50, 3*(n+n_L+n_N))

        The value of i specifies the maximum number of iterations for
        finding a feasible point with respect to the bounds and linear
        constraints (if any).
        The value of i also specifies the maximum number of minor
        iterations for the optimality phase of each QP subproblem.
        If i <= 0, the default value is used.

    'Minor Print Level' : int
        Default = 0

        The value of i controls the amount of printout produced by the
        minor iterations of ``nlp_multistart_sqp_lsq`` (i.e., the
        iterations of the quadratic programming algorithm), as indicated
        below.
        A detailed description of the printed output is given in
        [Further Comments] (summary output at each minor iteration and
        the final QP solution) and [Monitoring Information] (monitoring
        information at each minor iteration). (See also the description
        of the option 'Major Print Level'.) The following printout is
        sent to the file object associated with the advisory I/O unit
        (see :class:`~naginterfaces.base.utils.FileObjManager`):

        [table omitted]

        The following printout is sent to the logical unit number
        defined by the option 'Monitoring File':

        [table omitted]

        If 'Minor Print Level' >= 5 and the unit number defined by the
        option 'Monitoring File' is the advisory unit number, then the
        summary output for each minor iteration is suppressed.

    'Monitoring File' : int
        Default = -1

        If i >= 0 and 'Major Print Level' >= 5 or i >= 0 and 'Minor
        Print Level' >= 5, monitoring information produced by
        ``nlp_multistart_sqp_lsq`` at every iteration is sent to a file
        with logical unit number i. If i < 0 and/or 'Major Print Level'
        < 5 and 'Minor Print Level' < 5, no monitoring information is
        produced.

    'Optimality Tolerance' : float
        Default = epsilon_R^0.8

        The argument r (epsilon_R <= r < 1) specifies the accuracy to
        which you wish the final iterate to approximate a solution of
        the problem.
        Broadly speaking, r indicates the number of correct figures
        desired in the objective function at the solution.
        For example, if r is 10^(-6) and ``nlp_multistart_sqp_lsq``
        terminates successfully, the final value of F should have
        approximately six correct figures.
        If r < epsilon_r or r >= 1, the default value is used.

        ``nlp_multistart_sqp_lsq`` will terminate successfully if the
        iterative sequence of x values is judged to have converged and
        the final point satisfies the first-order Kuhn--Tucker
        conditions (see [Overview]). The sequence of iterates is
        considered to have converged at x if

            alpha||p|| <= sqrt(r)(1+||x||),

        where p is the search direction and alpha the step length from
        (3). An iterate is considered to satisfy the first-order
        conditions for a minimum if

            || Z^Tg_FR|| <= sqrt(r)(1+max(1+|F(x)|,||g_FR||))

        and

            |res_j| <= ftol for all j,

        where Z^Tg_FR is the projected gradient (see [Overview]), g_FR
        is the gradient of F(x) with respect to the free variables,
        res_j is the violation of the jth active nonlinear constraint,
        and ftol is the 'Nonlinear Feasibility Tolerance'.

    'Start Objective Check At Variable' : int
        Default = 1

        These keywords take effect only if 'Verify Level' > 0.
        They may be used to control the verification of gradient
        elements computed by `objfun` and/or Jacobian elements computed
        by `confun`.
        For example, if the first 30 elements of the objective gradient
        appeared to be correct in an earlier run, so that only element
        31 remains questionable, it is reasonable to specify 'Start
        Objective Check At Variable' = 31.
        If the first 30 variables appear linearly in the objective, so
        that the corresponding gradient elements are constant, the above
        choice would also be appropriate.

        If i_{2m-1} <= 0 or i_{2m-1} > min(n, i_{2m}), the default value
        is used, for m = 1,2,...,2.
        If i_{2m} <= 0 or i_{2m} > n, the default value is used, for m =
        1,2,...,2.

    'Stop Objective Check At Variable' : int
        Default = n

        These keywords take effect only if 'Verify Level' > 0.
        They may be used to control the verification of gradient
        elements computed by `objfun` and/or Jacobian elements computed
        by `confun`.
        For example, if the first 30 elements of the objective gradient
        appeared to be correct in an earlier run, so that only element
        31 remains questionable, it is reasonable to specify 'Start
        Objective Check At Variable' = 31.
        If the first 30 variables appear linearly in the objective, so
        that the corresponding gradient elements are constant, the above
        choice would also be appropriate.

        If i_{2m-1} <= 0 or i_{2m-1} > min(n, i_{2m}), the default value
        is used, for m = 1,2,...,2.
        If i_{2m} <= 0 or i_{2m} > n, the default value is used, for m =
        1,2,...,2.

    'Start Constraint Check At Variable' : int
        Default = 1

        These keywords take effect only if 'Verify Level' > 0.
        They may be used to control the verification of gradient
        elements computed by `objfun` and/or Jacobian elements computed
        by `confun`.
        For example, if the first 30 elements of the objective gradient
        appeared to be correct in an earlier run, so that only element
        31 remains questionable, it is reasonable to specify 'Start
        Objective Check At Variable' = 31.
        If the first 30 variables appear linearly in the objective, so
        that the corresponding gradient elements are constant, the above
        choice would also be appropriate.

        If i_{2m-1} <= 0 or i_{2m-1} > min(n, i_{2m}), the default value
        is used, for m = 1,2,...,2.
        If i_{2m} <= 0 or i_{2m} > n, the default value is used, for m =
        1,2,...,2.

    'Stop Constraint Check At Variable' : int
        Default = n

        These keywords take effect only if 'Verify Level' > 0.
        They may be used to control the verification of gradient
        elements computed by `objfun` and/or Jacobian elements computed
        by `confun`.
        For example, if the first 30 elements of the objective gradient
        appeared to be correct in an earlier run, so that only element
        31 remains questionable, it is reasonable to specify 'Start
        Objective Check At Variable' = 31.
        If the first 30 variables appear linearly in the objective, so
        that the corresponding gradient elements are constant, the above
        choice would also be appropriate.

        If i_{2m-1} <= 0 or i_{2m-1} > min(n, i_{2m}), the default value
        is used, for m = 1,2,...,2.
        If i_{2m} <= 0 or i_{2m} > n, the default value is used, for m =
        1,2,...,2.

    'Step Limit' : float
        Default = 2.0

        If r > 0,r specifies the maximum change in variables at the
        first step of the linesearch.
        In some cases, such as F(x) = ae^(bx) or F(x) = ax^b, even a
        moderate change in the elements of x can lead to floating-point
        overflow.
        The argument r is therefore used to encourage evaluation of the
        problem functions at meaningful points.
        Given any major iterate x, the first point x~ at which F and c
        are evaluated during the linesearch is restricted so that

            ||x~-x||_2 <= r(1+||x||_2).

        The linesearch may go on and evaluate F and c at points further
        from x if this will result in a lower value of the merit
        function (indicated by L at the end of each line of output
        produced by the major iterations; see [Further Comments]). If L
        is printed for most of the iterations, r should be set to a
        larger value.

        Wherever possible, upper and lower bounds on x should be used to
        prevent evaluation of nonlinear functions at wild values.
        The default value 'Step Limit' = 2.0 should not affect progress
        on well-behaved functions, but values such as 0.1 or 0.01 may be
        helpful when rapidly varying functions are present.
        If a small value of 'Step Limit' is selected, a good starting
        point may be required.
        An important application is to the class of nonlinear least
        squares problems.
        If r <= 0, the default value is used.

    'Verify Level' : int
        Default = 0

        These keywords refer to finite difference checks on the gradient
        elements computed by `objfun` and `confun`.
        The possible choices for i are as follows:

        [table omitted]

        It is possible to specify 'Verify Level' = 0 to 3 in several
        ways.
        For example, the nonlinear objective gradient (if any) will be
        verified if either 'Verify Objective Gradients' or 'Verify
        Level' = 1 is specified.
        The constraint gradients will be verified if 'Verify' = 'YES' or
        'Verify Level' = 2 or 'Verify' is specified.
        Similarly, the objective and the constraint gradients will be
        verified if 'Verify' = 'YES' or 'Verify Level' = 3 or 'Verify'
        is specified.

        If 0 <= i <= 3, gradients will be verified at the first point
        that satisfies the linear constraints and bounds.

        If i = 0, only a 'cheap' test will be performed, requiring one
        call to `objfun` and (if appropriate) one call to `confun`.

        If 1 <= i <= 3, a more reliable (but more expensive) check will
        be made on individual gradient elements, within the ranges
        specified by the 'Start Constraint Check At Variable' and 'Stop
        Constraint Check At Variable' keywords.
        A result of the form OK or BAD? is printed by
        ``nlp_multistart_sqp_lsq`` to indicate whether or not each
        element appears to be correct.
        If a gradient element is determined to be extremely poor (i.e.,
        if it appears to have no significant digits of accuracy at all)
        then ``nlp_multistart_sqp_lsq`` will also exit with an error
        indicator in argument errno.

        If 10 <= i <= 13, the action is the same as for i-10, except
        that it will take place at the user-specified initial value of
        x.

        If i < -1 or 4 <= i <= 9 or i > 13, the default value is used.

        We suggest that 'Verify Level' = 3 be used whenever a new
        function function is being developed.

    'Verify' : valueless
        These keywords refer to finite difference checks on the gradient
        elements computed by `objfun` and `confun`.
        The possible choices for i are as follows:

        [table omitted]

        It is possible to specify 'Verify Level' = 0 to 3 in several
        ways.
        For example, the nonlinear objective gradient (if any) will be
        verified if either 'Verify Objective Gradients' or 'Verify
        Level' = 1 is specified.
        The constraint gradients will be verified if 'Verify' = 'YES' or
        'Verify Level' = 2 or 'Verify' is specified.
        Similarly, the objective and the constraint gradients will be
        verified if 'Verify' = 'YES' or 'Verify Level' = 3 or 'Verify'
        is specified.

        If 0 <= i <= 3, gradients will be verified at the first point
        that satisfies the linear constraints and bounds.

        If i = 0, only a 'cheap' test will be performed, requiring one
        call to `objfun` and (if appropriate) one call to `confun`.

        If 1 <= i <= 3, a more reliable (but more expensive) check will
        be made on individual gradient elements, within the ranges
        specified by the 'Start Constraint Check At Variable' and 'Stop
        Constraint Check At Variable' keywords.
        A result of the form OK or BAD? is printed by
        ``nlp_multistart_sqp_lsq`` to indicate whether or not each
        element appears to be correct.
        If a gradient element is determined to be extremely poor (i.e.,
        if it appears to have no significant digits of accuracy at all)
        then ``nlp_multistart_sqp_lsq`` will also exit with an error
        indicator in argument errno.

        If 10 <= i <= 13, the action is the same as for i-10, except
        that it will take place at the user-specified initial value of
        x.

        If i < -1 or 4 <= i <= 9 or i > 13, the default value is used.

        We suggest that 'Verify Level' = 3 be used whenever a new
        function function is being developed.

    'Verify Constraint Gradients' : valueless
        These keywords refer to finite difference checks on the gradient
        elements computed by `objfun` and `confun`.
        The possible choices for i are as follows:

        [table omitted]

        It is possible to specify 'Verify Level' = 0 to 3 in several
        ways.
        For example, the nonlinear objective gradient (if any) will be
        verified if either 'Verify Objective Gradients' or 'Verify
        Level' = 1 is specified.
        The constraint gradients will be verified if 'Verify' = 'YES' or
        'Verify Level' = 2 or 'Verify' is specified.
        Similarly, the objective and the constraint gradients will be
        verified if 'Verify' = 'YES' or 'Verify Level' = 3 or 'Verify'
        is specified.

        If 0 <= i <= 3, gradients will be verified at the first point
        that satisfies the linear constraints and bounds.

        If i = 0, only a 'cheap' test will be performed, requiring one
        call to `objfun` and (if appropriate) one call to `confun`.

        If 1 <= i <= 3, a more reliable (but more expensive) check will
        be made on individual gradient elements, within the ranges
        specified by the 'Start Constraint Check At Variable' and 'Stop
        Constraint Check At Variable' keywords.
        A result of the form OK or BAD? is printed by
        ``nlp_multistart_sqp_lsq`` to indicate whether or not each
        element appears to be correct.
        If a gradient element is determined to be extremely poor (i.e.,
        if it appears to have no significant digits of accuracy at all)
        then ``nlp_multistart_sqp_lsq`` will also exit with an error
        indicator in argument errno.

        If 10 <= i <= 13, the action is the same as for i-10, except
        that it will take place at the user-specified initial value of
        x.

        If i < -1 or 4 <= i <= 9 or i > 13, the default value is used.

        We suggest that 'Verify Level' = 3 be used whenever a new
        function function is being developed.

    'Verify Gradients' : valueless
        These keywords refer to finite difference checks on the gradient
        elements computed by `objfun` and `confun`.
        The possible choices for i are as follows:

        [table omitted]

        It is possible to specify 'Verify Level' = 0 to 3 in several
        ways.
        For example, the nonlinear objective gradient (if any) will be
        verified if either 'Verify Objective Gradients' or 'Verify
        Level' = 1 is specified.
        The constraint gradients will be verified if 'Verify' = 'YES' or
        'Verify Level' = 2 or 'Verify' is specified.
        Similarly, the objective and the constraint gradients will be
        verified if 'Verify' = 'YES' or 'Verify Level' = 3 or 'Verify'
        is specified.

        If 0 <= i <= 3, gradients will be verified at the first point
        that satisfies the linear constraints and bounds.

        If i = 0, only a 'cheap' test will be performed, requiring one
        call to `objfun` and (if appropriate) one call to `confun`.

        If 1 <= i <= 3, a more reliable (but more expensive) check will
        be made on individual gradient elements, within the ranges
        specified by the 'Start Constraint Check At Variable' and 'Stop
        Constraint Check At Variable' keywords.
        A result of the form OK or BAD? is printed by
        ``nlp_multistart_sqp_lsq`` to indicate whether or not each
        element appears to be correct.
        If a gradient element is determined to be extremely poor (i.e.,
        if it appears to have no significant digits of accuracy at all)
        then ``nlp_multistart_sqp_lsq`` will also exit with an error
        indicator in argument errno.

        If 10 <= i <= 13, the action is the same as for i-10, except
        that it will take place at the user-specified initial value of
        x.

        If i < -1 or 4 <= i <= 9 or i > 13, the default value is used.

        We suggest that 'Verify Level' = 3 be used whenever a new
        function function is being developed.

    'Verify Objective Gradients' : valueless
        These keywords refer to finite difference checks on the gradient
        elements computed by `objfun` and `confun`.
        The possible choices for i are as follows:

        [table omitted]

        It is possible to specify 'Verify Level' = 0 to 3 in several
        ways.
        For example, the nonlinear objective gradient (if any) will be
        verified if either 'Verify Objective Gradients' or 'Verify
        Level' = 1 is specified.
        The constraint gradients will be verified if 'Verify' = 'YES' or
        'Verify Level' = 2 or 'Verify' is specified.
        Similarly, the objective and the constraint gradients will be
        verified if 'Verify' = 'YES' or 'Verify Level' = 3 or 'Verify'
        is specified.

        If 0 <= i <= 3, gradients will be verified at the first point
        that satisfies the linear constraints and bounds.

        If i = 0, only a 'cheap' test will be performed, requiring one
        call to `objfun` and (if appropriate) one call to `confun`.

        If 1 <= i <= 3, a more reliable (but more expensive) check will
        be made on individual gradient elements, within the ranges
        specified by the 'Start Constraint Check At Variable' and 'Stop
        Constraint Check At Variable' keywords.
        A result of the form OK or BAD? is printed by
        ``nlp_multistart_sqp_lsq`` to indicate whether or not each
        element appears to be correct.
        If a gradient element is determined to be extremely poor (i.e.,
        if it appears to have no significant digits of accuracy at all)
        then ``nlp_multistart_sqp_lsq`` will also exit with an error
        indicator in argument errno.

        If 10 <= i <= 13, the action is the same as for i-10, except
        that it will take place at the user-specified initial value of
        x.

        If i < -1 or 4 <= i <= 9 or i > 13, the default value is used.

        We suggest that 'Verify Level' = 3 be used whenever a new
        function function is being developed.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `nb` = *<value>* and `npts` = *<value>*.

            Constraint: 1 <= `nb` <= `npts`.

        (`errno` 1)
            On entry, `bl`[i-1] > `bu`[i-1]: i = *<value>*.

            Constraint: `bl`[i-1] <= `bu`[i-1], for all i.

        (`errno` 1)
            On entry, `ncnln` = *<value>*.

            Constraint: `ncnln` >= 0.

        (`errno` 1)
            On entry, nclin = *<value>*.

            Constraint: nclin >= 0.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n > 0.

        (`errno` 1)
            On entry, ldclda = *<value>*, n = *<value>*, nclin =
            *<value>* and `ncnln` = *<value>*.

            Constraint: ldclda >= n+nclin+`ncnln`.

        (`errno` 1)
            On entry, m = *<value>*.

            Constraint: m > 0.

        (`errno` 1)
            On entry, `ncnln` > 0, sdcjac = *<value>* and n = *<value>*.

            Constraint: if `ncnln` > 0, sdcjac >= n.

        (`errno` 1)
            On entry, listat = *<value>*, n = *<value>*, nclin =
            *<value>* and `ncnln` = *<value>*.

            Constraint: listat >= n+nclin+`ncnln`.

        (`errno` 2)
            No solution obtained. Linear constraints may be infeasible.

        (`errno` 3)
            No solution obtained. Nonlinear constraints may be
            infeasible.

        (`errno` 4)
            No solution obtained. Many potential solutions reach
            iteration limit.

        (`errno` 7)
            User-supplied derivatives probably wrong.

        (`errno` 10)
            Failed to initialize option arrays.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 8)
            Only *<value>* solutions obtained.

    NagCallbackTerminateWarning
        (`errno` 9)
            User terminated computation from `start` procedure: `mode` =
            *<value>*.

    References
    ----------
    Gill, P E, Murray, W and Wright, M H, 1981, `Practical
    Optimization`, Academic Press

    Hock, W and Schittkowski, K, 1981, `Test Examples for Nonlinear
    Programming Codes. Lecture Notes in Economics and Mathematical
    Systems` (187), Springer--Verlag

    See Also
    --------
    :meth:`naginterfaces.library.examples.glopt.nlp_multistart_sqp_lsq_ex.main`
    """

    fun_name = 'naginterfaces.library.glopt.nlp_multistart_sqp_lsq'

    cb_confun_name = fun_name + ':confun'

    def cb_confun(mode, ncnln, n, needc, x, c, cjsl, nstate, data=None):
        """Wrapping user's confun. Called as base confun."""
        if data is not None:
            _c_confun = confun(
                mode,
                needc,
                x,
                cjsl,
                nstate,
                data,
            )
        else:
            _c_confun = confun(
                mode,
                needc,
                x,
                cjsl,
                nstate,
            )
        c[:] = _utils._to_ndarray(
            {'fun_name': cb_confun_name, 'returned': 'c'},
            _c_confun,
            copy=False,
            exp_shape=(ncnln,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )

    cb_objfun_name = fun_name + ':objfun'

    def cb_objfun(mode, m, n, needfi, x, f, fjsl, nstate, data=None):
        """Wrapping user's objfun. Called as base objfun."""
        if data is not None:
            _f_objfun = objfun(
                mode,
                needfi,
                x,
                fjsl,
                nstate,
                data,
            )
        else:
            _f_objfun = objfun(
                mode,
                needfi,
                x,
                fjsl,
                nstate,
            )
        f[:] = _utils._to_ndarray(
            {'fun_name': cb_objfun_name, 'returned': 'f'},
            _f_objfun,
            copy=False,
            exp_shape=(m,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )

    def cb_start(npts, quas, n, repeat1, bl, bu, data=None):
        """Wrapping user's start. Called as base start."""
        if data is not None:
            start(
                quas,
                repeat1,
                bl,
                bu,
                data,
            )
        else:
            start(
                quas,
                repeat1,
                bl,
                bu,
            )

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_a, 'a'),),
    )
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_y, {'fun_name': fun_name, 'entity': 'y'}, exp_rank=1)[0]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    if a is None:
        _a = None
    _nclin = _b_utils._EngineIntScalarType(
        0 if a is None else _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, allow_none=True)[0], {'fun_name': fun_name, 'entity': '_nclin'},
    ).to_ctypes()
    _bl = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'bl'},
        bl,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nclin_py = _b_utils._EngineIntScalarType.to_py(_nclin)
    _ncnln = _b_utils._EngineIntScalarType(
        ncnln, {'fun_name': fun_name, 'entity': 'ncnln'},
    ).to_ctypes()
    _ncnln_py = _b_utils._EngineIntScalarType.to_py(_ncnln)
    _n = _b_utils._EngineIntScalarType(
        _b_utils._EngineFloat64ArrayType.get_shape(_bl, {'fun_name': fun_name, 'entity': 'bl'}, exp_rank=1)[0]-_nclin_py-_ncnln_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _bu = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'bu'},
        bu,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b_utils._check_callable(confun, {'fun_name': fun_name, 'entity': 'confun'}, allow_none=True)
    _confun = None if confun is None else cb_confun
    _b_utils._check_callable(objfun, {'fun_name': fun_name, 'entity': 'objfun'})
    _objfun = cb_objfun
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _nb = _b_utils._EngineIntScalarType(
        nb, {'fun_name': fun_name, 'entity': 'nb'},
    ).to_ctypes()
    _nb_py = _b_utils._EngineIntScalarType.to_py(_nb)
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py, _nb_py),
        sorder=_m_sorder,
    )
    _b_utils._check_callable(start, {'fun_name': fun_name, 'entity': 'start'}, allow_none=True)
    _start = None if start is None else cb_start
    _objf = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nb_py,),
    )
    _f = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_m_py, _nb_py),
        sorder=_m_sorder,
    )
    _fjac = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_m_py, _n_py, _nb_py),
        sorder=_m_sorder,
    )
    _itera = _b_utils._EngineIntArrayType.empty_ndarray(
        (_nb_py,),
    )
    if _ncnln_py != 0:
        _c = _b_utils._EngineFloat64ArrayType.empty_ndarray(
            (_ncnln_py, _nb_py),
            sorder=_m_sorder,
        )
    else:
        _c = None
    if _ncnln_py != 0:
        if _ncnln_py > 0:
            _cjac_dim2 = _n_py
        else:
            _cjac_dim2 = 0
        exp_shape_cjac = (_ncnln_py, _cjac_dim2, _nb_py)
        _cjac = _b_utils._EngineFloat64ArrayType.empty_ndarray(
            exp_shape_cjac,
            sorder=_m_sorder,
        )
    else:
        _cjac = None
    _clamda = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py+_nclin_py+_ncnln_py, _nb_py),
        sorder=_m_sorder,
    )
    _istate = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n_py+_nclin_py+_ncnln_py, _nb_py),
        sorder=_m_sorder,
    )
    _info = _b_utils._EngineIntArrayType.empty_ndarray(
        (_nb_py,),
    )
    try:
        _b_glopt.nlp_multistart_sqp_lsq(
            _m,
            _n,
            _nclin,
            _ncnln,
            _a,
            _bl,
            _bu,
            _y,
            _confun,
            _objfun,
            npts,
            _x,
            _start,
            repeat1,
            _nb,
            _objf,
            _f,
            _fjac,
            _itera,
            _c,
            _cjac,
            _clamda,
            _istate,
            comm,
            _info,
            data,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _x_py = _x
    _objf_py = _objf
    _f_py = _f
    _fjac_py = _fjac
    _itera_py = _itera
    if _ncnln_py != 0:
        _c_py = _c
    else:
        _c_py = None
    if _ncnln_py != 0:
        _cjac_py = _cjac
    else:
        _cjac_py = None
    _clamda_py = _clamda
    _istate_py = _istate
    _info_py = _info
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'objf', 'f', 'fjac', 'itera', 'c', 'cjac', 'clamda', 'istate', 'info'],
    )(
        _x_py,
        _objf_py,
        _f_py,
        _fjac_py,
        _itera_py,
        _c_py,
        _cjac_py,
        _clamda_py,
        _istate_py,
        _info_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def optset(optstr, comm, io_manager=None):
    u"""
    Option setting routine for ``bnd_pso``, ``nlp_pso``,
    ``nlp_multistart_sqp`` and ``nlp_multistart_sqp_lsq``.

    ``optset`` either initializes or resets the option arrays or sets a
    single option for supported problem solving functions in submodule
    ``glopt``.
    The following functions are supported:

        ``bnd_pso``,

        ``nlp_pso``,

        ``nlp_multistart_sqp``,

        ``nlp_multistart_sqp_lsq``.

    For full information please refer to the NAG Library document for
    e05zk

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e05/e05zkf.html

    Parameters
    ----------
    optstr : str
        A string identifying the option to be set.

        Initialize = function name

            Initialize the option arrays `comm`['iopts'] and
            `comm`['opts'] for use with function function name, where
            function name is the name associated with the function of
            interest.

        Defaults

            Resets all options to their default values.

        option = optval

            See [Other Parameters for bnd_pso], [Other Parameters for
            nlp_pso], [Other Parameters for nlp_multistart_sqp] and
            [Other Parameters for nlp_multistart_sqp_lsq] for details of
            valid values for option and optval. The equals sign (=)
            delimiter must be used to separate the option from its
            optval value.

        `optstr` is case insensitive.

        Each token in the option and optval component must be separated
        by at least one space.

    comm : dict, communication object, modified in place
        Communication structure.

        `On initial entry`: need not be set.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, the option in `optstr` was not recognized:
            `optstr` = *<value>*.

        (`errno` 12)
            On entry, the expected delimiter '=' was not found in
            `optstr`: `optstr` = *<value>*.

        (`errno` 13)
            The option in `optstr` is associated with a numerical value.
            However, the optval, *<value>*, present in `optstr` could
            not be fully interpreted.

        (`errno` 13)
            On entry, could not convert the specified optval to an
            integer: `optstr` = *<value>*.

        (`errno` 13)
            On entry, could not convert the specified optval to a real:
            `optstr` = *<value>*.

        (`errno` 14)
            On entry, the option in `optstr` has been detected as
            Initialize, however the optval, *<value>*, associated with
            `optstr` has not been recognized as a valid function name.

        (`errno` 15)
            On entry, the optval supplied for the integer option is not
            valid.

            `optstr` = *<value>*.

        (`errno` 16)
            On entry, the optval supplied for the real option is not
            valid.

            `optstr` = *<value>*.

        (`errno` 17)
            On entry, the optval supplied for the character option is
            not valid.

            `optstr` = *<value>*.

        (`errno` 17)
            On entry, the option in `optstr` was not recognized:
            `optstr` = *<value>*.

        (`errno` 21)
            On entry, either the option arrays have not been initialized
            or they have been corrupted.

        (`errno` 21)
            On entry, the option in `optstr` was not recognized:
            `optstr` = *<value>*.

    See Also
    --------
    :meth:`naginterfaces.library.examples.glopt.nlp_multistart_sqp_lsq_ex.main`
    """

    _b_glopt.optset(
        optstr,
        comm,
        io_manager,
    )

def optget(optstr, comm):
    u"""
    Option getting routine for ``bnd_pso``, ``nlp_pso``,
    ``nlp_multistart_sqp`` and ``nlp_multistart_sqp_lsq``.

    ``optget`` is used to query the value of options available to
    supported problem solving functions in submodule ``glopt``.
    Currently the following functions are supported:

        ``bnd_pso``,

        ``nlp_pso``,

        ``nlp_multistart_sqp``,

        ``nlp_multistart_sqp_lsq``.

    For full information please refer to the NAG Library document for
    e05zl

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e05/e05zlf.html

    Parameters
    ----------
    optstr : str
        A string identifying the option whose current value is required.
        See [Other Parameters for bnd_pso] and [Other Parameters for
        nlp_pso] for information on valid options. In addition, the
        following is a valid option:

        Identify

            ``optget`` returns the function name supplied to ``optset``
            when the option arrays `comm`['iopts'] and `comm`['opts']
            were initialized.

    comm : dict, communication object
        Communication structure.

        This argument must have been initialized by a prior call to
        ``optset``.

    Returns
    -------
    optvalue : dict
        The option-value ``dict``, with the following keys:

        ``'value'`` : float, int or str
            The value of the requested option.

        ``'annotation'`` : None or str
            Possible additional information about the option value.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, the option in `optstr` has not been recognized.

        (`errno` 61)
            The arrays `comm`['iopts'] and `comm`['opts'] have either
            not been initialized, have become corrupted, or are not
            compatible with this option setting function.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 41)
            On entry, `optstr` indicates a character option, but cvalue
            is too short to hold the stored value. The returned value
            will be truncated.
    """

    try:
        _return_scalars = _b_glopt.optget(
            optstr,
            comm,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    fun_name = 'naginterfaces.library.glopt.optget'

    _optype_py = _return_scalars.optype
    _ivalue_py = _return_scalars.ivalue
    _rvalue_py = _return_scalars.rvalue
    _cvalue_py = _return_scalars.cvalue
    if _optype_py == 1:
        _optvalue_py = {}
        _optvalue_py['value'] = _ivalue_py
        _optvalue_py['annotation'] = None
    elif _optype_py == 2:
        _optvalue_py = {}
        _optvalue_py['value'] = _rvalue_py
        _optvalue_py['annotation'] = None
    elif _optype_py == 3:
        _optvalue_py = {}
        _optvalue_py['value'] = _cvalue_py
        _optvalue_py['annotation'] = None
    elif _optype_py == 4:
        _optvalue_py = {}
        _optvalue_py['value'] = _ivalue_py
        _optvalue_py['annotation'] = _cvalue_py
    elif _optype_py == 5:
        _optvalue_py = {}
        _optvalue_py['value'] = _rvalue_py
        _optvalue_py['annotation'] = _cvalue_py
    elif _optype_py == 6:
        _optvalue_py = {}
        if _ivalue_py == _nag_e_consts.PSO_SOLVER_ID_A:
            _optvalue_py = {}
            _optvalue_py['value'] = 'bnd_pso'
        elif _ivalue_py == _nag_e_consts.PSO_SOLVER_ID_AC:
            _optvalue_py = {}
            _optvalue_py['value'] = 'bnd_pso'
        elif _ivalue_py == _nag_e_consts.PSO_SOLVER_ID_B:
            _optvalue_py = {}
            _optvalue_py['value'] = 'nlp_pso'
        elif _ivalue_py == _nag_e_consts.PSO_SOLVER_ID_BC:
            _optvalue_py = {}
            _optvalue_py['value'] = 'nlp_pso'
        elif _ivalue_py == _nag_e_consts.E05UC_SOLVER_ID:
            _optvalue_py = {}
            _optvalue_py['value'] = 'nlp_multistart_sqp'
        elif _ivalue_py == _nag_e_consts.E05US_SOLVER_ID:
            _optvalue_py = {}
            _optvalue_py['value'] = 'nlp_multistart_sqp_lsq'
        else:
            raise _b_utils.NagException(
                {'fun_name': fun_name},
                'inconceivable option configuration',
            )
        _optvalue_py['annotation'] = None
    else:
        raise _b_utils.NagException(
            {'fun_name': fun_name},
            'inconceivable option configuration',
        )
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['optvalue'],
    )(
        _optvalue_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _optvalue_py
