# -*- coding: utf-8 -*-
# pylint: disable=chained-comparison,invalid-name,line-too-long,protected-access,too-many-arguments,too-many-boolean-expressions,too-many-branches,too-many-lines,too-many-locals,too-many-statements,unused-import
u"""
Interfaces for the NAG Mark 27.0 `lapacklin` Chapter.

``lapacklin`` - Linear Equations (LAPACK)

This module provides functions for the solution of systems of
simultaneous linear equations, and associated computations.
It provides functions for

    matrix factorizations;

    solution of linear equations;

    estimating matrix condition numbers;

    computing error bounds for the solution of linear equations;

    matrix inversion;

    computing scaling factors to equilibrate a matrix.

Functions are provided for both `real` and `complex` data.

For a general introduction to the solution of systems of linear
equations, you should turn first to [the F04 Introduction].
The decision trees, in [the F04 Introduction], direct you to the most
appropriate functions in submodule ``linsys`` or submodule ``lapacklin``
for solving your particular problem.
In particular, submodule ``linsys`` and submodule ``lapacklin`` contain
`Black Box` (or `driver`) functions which enable some standard types of
problem to be solved by a call to a single function.
Where possible, functions in submodule ``linsys`` call submodule
``lapacklin`` functions to perform the necessary computational tasks.

There are two types of driver functions in this module: `simple drivers`
which just return the solution to the linear equations; and `expert
drivers` which also return condition and error estimates and, in many
cases, also allow equilibration.
The simple drivers for real matrices have names of the form and for
complex matrices have names of the form The expert drivers for real
matrices have names of the form and for complex matrices have names of
the form

The functions in this module (submodule ``lapacklin``) handle only
`dense` and `band` matrices (not matrices with more specialised
structures, or general sparse matrices).

The functions in this module have all been derived from the LAPACK
project (see Anderson `et al.` (1999)).
They have been designed to be efficient on a wide range of
high-performance computers, without compromising efficiency on
conventional serial machines.

For full information please refer to the NAG Library document

https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07intro.html

See Also
--------
``naginterfaces.library.examples.lapacklin`` :
    This subpackage contains examples for the ``lapacklin`` module.
    See also the :ref:`library_lapacklin_ex` subsection.
"""

# NAG Copyright 2017-2020.

import collections as _collections

from ..base import lapacklin as _b_lapacklin
from ..base import utils as _b_utils
from ..library import _utils

def dgesv(a, b):
    u"""
    Computes the solution to a real system of linear equations.

    ``dgesv`` computes the solution to a real system of linear equations

        AX = B,

    where A is an n by n matrix and X and B are n by r matrices.

    For full information please refer to the NAG Library document for
    f07aa

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07aaf.html

    Parameters
    ----------
    a : float, array-like, shape (n, n)
        The n by n coefficient matrix A.

    b : float, array-like, shape (n, nrhs)
        The n by r right-hand side matrix B.

    Returns
    -------
    a : float, ndarray, shape (n, n)
        The factors L and U from the factorization A = PLU; the unit
        diagonal elements of L are not stored.

    ipiv : int, ndarray, shape (n)
        If no constraints are violated, the pivot indices that define
        the permutation matrix P; at the ith step row i of the matrix
        was interchanged with row `ipiv`[i-1]. `ipiv`[i-1] = i indicates
        a row interchange was not required.

    b : float, ndarray, shape (n, nrhs)
        If no exception or warning is raised, the n by r solution matrix
        X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -2)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. The
            factorization has been completed, but the factor U is
            exactly singular, so the solution could not be computed.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    See Also
    --------
    :meth:`naginterfaces.library.examples.lapacklin.dgesv_ex.main`
    """

    fun_name = 'naginterfaces.library.lapacklin.dgesv'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ipiv = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _b_lapacklin.dgesv(
            _n,
            _nrhs,
            _a,
            _ipiv,
            _b,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _a_py = _a
    _b_py = _b
    _ipiv_py = _ipiv
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'ipiv', 'b'],
    )(
        _a_py,
        _ipiv_py,
        _b_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def dgesvx(fact, trans, n, nrhs, a, equed, b, af=None, ipiv=None, r=None, c=None):
    u"""
    Uses the LU factorization to compute the solution, error-bound and
    condition estimate for a real system of linear equations.

    ``dgesvx`` uses the LU factorization to compute the solution to a
    real system of linear equations

        AX = B or A^TX = B,

    where A is an n by n matrix and X and B are n by r matrices.
    Error bounds on the solution and a condition estimate are also
    provided.

    For full information please refer to the NAG Library document for
    f07ab

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07abf.html

    Parameters
    ----------
    fact : str, length 1
        Specifies whether or not the factorized form of the matrix A is
        supplied on entry, and if not, whether the matrix A should be
        equilibrated before it is factorized.

        `fact` = 'F'

            `af` and `ipiv` contain the factorized form of A. If `equed`
            != 'N', the matrix A has been equilibrated with scaling
            factors given by `r` and `c`. `a`, `af` and `ipiv` are not
            modified.

        `fact` = 'N'

            The matrix A will be copied to `af` and factorized.

        `fact` = 'E'

            The matrix A will be equilibrated if necessary, then copied
            to `af` and factorized.

    trans : str, length 1
        Specifies the form of the system of equations.

        `trans` = 'N'

            AX = B (No transpose).

        `trans` = 'T' or 'C'

            A^TX = B (Transpose).

    n : int
        n, the number of linear equations, i.e., the order of the matrix
        A.

    nrhs : int
        r, the number of right-hand sides, i.e., the number of columns
        of the matrix B.

    a : float, array-like, shape (`n`, `n`)
        The n by n matrix A.

        If `fact` = 'F' and `equed` != 'N', `a` must have been
        equilibrated by the scaling factors in `r` and/or `c`.

    equed : None or str, length 1
        If `fact` = 'N' or 'E', `equed` need not be set.

        If `fact` = 'F', `equed` must specify the form of the
        equilibration that was performed as follows:

            if `equed` = 'N', no equilibration;

            if `equed` = 'R', row equilibration, i.e., A has been
            premultiplied by D_R;

            if `equed` = 'C', column equilibration, i.e., A has been
            postmultiplied by D_C;

            if `equed` = 'B', both row and column equilibration, i.e., A
            has been replaced by D_R AD_C.

    b : float, array-like, shape (`n`, `nrhs`)
        The n by r right-hand side matrix B.

    af : None or float, array-like, shape (`n`, `n`), optional
        If `fact` = 'F', `af` contains the factors L and U from the
        factorization A = PLU as computed by ``dgetrf``. If `equed` !=
        'N', `af` is the factorized form of the equilibrated matrix A.

        If `fact` = 'N' or 'E', `af` need not be set.

    ipiv : None or int, array-like, shape (`n`), optional
        If `fact` = 'F', `ipiv` contains the pivot indices from the
        factorization A = PLU as computed by ``dgetrf``; at the ith step
        row i of the matrix was interchanged with row `ipiv`[i-1].
        `ipiv`[i-1] = i indicates a row interchange was not required.

        If `fact` = 'N' or 'E', `ipiv` need not be set.

    r : None or float, array-like, shape (`n`), optional
        If `fact` = 'N' or 'E', `r` need not be set.

        If `fact` = 'F' and `equed` = 'R' or 'B', `r` must contain the
        row scale factors for A, D_R; each element of `r` must be
        positive.

    c : None or float, array-like, shape (`n`), optional
        If `fact` = 'N' or 'E', `c` need not be set.

        If `fact` = 'F' and `equed` = 'C' or 'B', `c` must contain the
        column scale factors for A, D_C; each element of `c` must be
        positive.

    Returns
    -------
    a : float, ndarray, shape (`n`, `n`)
        If `fact` = 'F' or 'N', or if `fact` = 'E' and `equed` = 'N',
        `a` is not modified.

        If `fact` = 'E' or `equed` != 'N', A is scaled as follows:

            if `equed` = 'R', A = D_R A;

            if `equed` = 'C', A = AD_C;

            if `equed` = 'B', A = D_R AD_C.

    af : None or float, ndarray, shape (`n`, `n`)
        If `fact` = 'N', `af` returns the factors L and U from the
        factorization A = PLU of the original matrix A.

        If `fact` = 'E', `af` returns the factors L and U from the
        factorization A = PLU of the equilibrated matrix A (see the
        description of `a` for the form of the equilibrated matrix).

        If `fact` = 'F', `af` is unchanged from entry.

    ipiv : None or int, ndarray, shape (`n`)
        If `fact` = 'N', `ipiv` contains the pivot indices from the
        factorization A = PLU of the original matrix A.

        If `fact` = 'E', `ipiv` contains the pivot indices from the
        factorization A = PLU of the equilibrated matrix A.

        If `fact` = 'F', `ipiv` is unchanged from entry.

    equed : None or str, length 1
        If `fact` = 'F', `equed` is unchanged from entry.

        Otherwise, if no constraints are violated, `equed` specifies the
        form of equilibration that was performed as specified above.

    r : None or float, ndarray, shape (`n`)
        If `fact` = 'F', `r` is unchanged from entry.

        Otherwise, if no constraints are violated and `equed` = 'R' or
        'B', `r` contains the row scale factors for A, D_R, such that A
        is multiplied on the left by D_R; each element of `r` is
        positive.

    c : None or float, ndarray, shape (`n`)
        If `fact` = 'F', `c` is unchanged from entry.

        Otherwise, if no constraints are violated and `equed` = 'C' or
        'B', `c` contains the row scale factors for A, D_C; each element
        of `c` is positive.

    b : float, ndarray, shape (`n`, `nrhs`)
        If `equed` = 'N', `b` is not modified.

        If `trans` = 'N' and `equed` = 'R' or 'B', `b` is overwritten by
        D_R B.

        If `trans` = 'T' or 'C' and `equed` = 'C' or 'B', `b` is
        overwritten by D_C B.

    x : float, ndarray, shape (`n`, `nrhs`)
        If the function exits successfully or `errno` = `n` + 1, the n
        by r solution matrix X to the original system of equations. Note
        that the arrays A and B are modified on exit if `equed` != 'N',
        and the solution to the equilibrated system is D_C^-1X if
        `trans` = 'N' and `equed` = 'C' or 'B', or D_R^-1X if `trans` =
        'T' or 'C' and `equed` = 'R' or 'B'.

    rcond : float
        If no constraints are violated, an estimate of the reciprocal
        condition number of the matrix A (after equilibration if that is
        performed), computed as `rcond` = 1.0/(||A||_1||A^-1||_1).

    ferr : float, ndarray, shape (`nrhs`)
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the forward error bound for each computed solution
        vector, such that ||x^_j-x_j||_infinity/||x_j||_infinity <=
        `ferr`[j-1] where x^_j is the jth column of the computed
        solution returned in the array `x` and x_j is the corresponding
        column of the exact solution X. The estimate is as reliable as
        the estimate for `rcond`, and is almost always a slight
        overestimate of the true error.

    berr : float, ndarray, shape (`nrhs`)
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the component-wise relative backward error of each
        computed solution vector x^_j (i.e., the smallest relative
        change in any element of A or B that makes x^_j an exact
        solution).

    recip_growth_factor : float
        If no exception or warning is raised, the reciprocal pivot
        growth factor ||A||/||U||, where ||.|| denotes the maximum
        absolute element norm. If `recip_growth_factor`≪1, the stability
        of the LU factorization of (equilibrated) A could be poor. This
        also means that the solution `x`, condition estimate `rcond`,
        and forward error bound `ferr` could be unreliable. If the
        factorization fails with `errno` in 1 ... `n`, then
        `recip_growth_factor` contains the reciprocal pivot growth
        factor for the leading errno columns of A.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `fact`.

            Constraint: `fact` = 'F', 'N' or 'E'.

        (`errno` -2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` -3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -4)
            On entry, error in parameter `nrhs`.

            Constraint: `nrhs` >= 0.

        (`errno` -10)
            On entry, error in parameter `equed`.

            Constraint: `equed` = 'N', 'R', 'C' or 'B'.

        (`errno` -11)
            On entry, error in parameter `r`.

        (`errno` -12)
            On entry, error in parameter `c`.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` (i > 0) and (i<=`n`))
            Element *<value>* of the diagonal is exactly zero. The
            factorization has been completed, but the factor U is
            exactly singular, so the solution and error bounds could not
            be computed. `rcond` = 0.0 is returned.

        (`errno` `n`+1)
            U is nonsingular, but `rcond` is less than machine
            precision, meaning that the matrix is singular to working
            precision. Nevertheless, the solution and error bounds are
            computed because there are a number of situations where the
            computed solution can be more accurate than the value of
            `rcond` would suggest.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    fun_name = 'naginterfaces.library.lapacklin.dgesvx'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _af = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'af'},
        af,
        copy=True,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_a, 'a'), (_b, 'b'), (_af, 'af')),
    )
    _fact = _b_utils._EngineCharScalarType(
        fact, {'fun_name': fun_name, 'entity': 'fact'}, exp_chrlen=1,
    ).to_ctypes()
    _fact_py = _b_utils._EngineCharScalarType.to_py(_fact)
    if _fact_py in ('F', 'f'):
        if af is None:
            raise _b_utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'af'},
                'must not be None when fact in (\'F\', \'f\')',
            )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    if af is None:
        _af = _b_utils._EngineFloat64ArrayType.empty_ndarray(
            (_n_py, _n_py),
            sorder=_m_sorder,
        )
    if _fact_py in ('F', 'f'):
        if ipiv is None:
            raise _b_utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'ipiv'},
                'must not be None when fact in (\'F\', \'f\')',
            )
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=True,
        allow_none=True,
    )
    _equed = _b_utils._EngineCharScalarType(
        equed, {'fun_name': fun_name, 'entity': 'equed'}, exp_chrlen=1,
        allow_none=True,
    ).to_ctypes()
    _equed_py = _b_utils._EngineCharScalarType.to_py(_equed)
    if (_fact_py in ('F', 'f')) and _equed_py in ('R', 'r', 'B', 'b'):
        if r is None:
            raise _b_utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'r'},
                'must not be None when (fact in (\'F\', \'f\')) and equed in (\'R\', \'r\', \'B\', \'b\')',
            )
    _r = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'r'},
        r,
        copy=True,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    if (_fact_py in ('F', 'f')) and _equed_py in ('C', 'c', 'B', 'b'):
        if c is None:
            raise _b_utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'c'},
                'must not be None when (fact in (\'F\', \'f\')) and equed in (\'C\', \'c\', \'B\', \'b\')',
            )
    _c = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
        copy=True,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nrhs = _b_utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineIntScalarType.to_py(_nrhs)
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py, _nrhs_py),
        sorder=_m_sorder,
    )
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    try:
        _return_scalars = _b_lapacklin.dgesvx(
            fact,
            trans,
            _n,
            _nrhs,
            _a,
            _af,
            _ipiv,
            equed,
            _r,
            _c,
            _b,
            _x,
            _ferr,
            _berr,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _a_py = _a
    if _fact_py not in ('F', 'f'):
        _equed_py = _return_scalars.equed
    else:
        _equed_py = None
    _b_py = _b
    if _fact_py not in ('F', 'f'):
        _af_py = _af
    else:
        _af_py = None
    if _fact_py not in ('F', 'f'):
        _ipiv_py = _ipiv
    else:
        _ipiv_py = None
    if _fact_py not in ('F', 'f') and _equed_py in ('R', 'r', 'B', 'b'):
        _r_py = _r
    else:
        _r_py = None
    if _fact_py not in ('F', 'f') and _equed_py in ('C', 'c', 'B', 'b'):
        _c_py = _c
    else:
        _c_py = None
    _x_py = _x
    _rcond_py = _return_scalars.rcond
    _ferr_py = _ferr
    _berr_py = _berr
    _recip_growth_factor_py = _return_scalars.recip_growth_factor
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'af', 'ipiv', 'equed', 'r', 'c', 'b', 'x', 'rcond', 'ferr', 'berr', 'recip_growth_factor'],
    )(
        _a_py,
        _af_py,
        _ipiv_py,
        _equed_py,
        _r_py,
        _c_py,
        _b_py,
        _x_py,
        _rcond_py,
        _ferr_py,
        _berr_py,
        _recip_growth_factor_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def dsgesv(a, b):
    u"""
    Computes the solution to a real system of linear equations using
    mixed precision arithmetic.

    ``dsgesv`` computes the solution to a real system of linear
    equations

        AX = B,

    where A is an n by n matrix and X and B are n by r matrices.

    For full information please refer to the NAG Library document for
    f07ac

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07acf.html

    Parameters
    ----------
    a : float, array-like, shape (n, n)
        The n by n coefficient matrix A.

    b : float, array-like, shape (n, nrhs)
        The n by r right-hand side matrix B.

    Returns
    -------
    a : float, ndarray, shape (n, n)
        If iterative refinement has been successfully used (i.e., if no
        exception or warning is raised and `itera` >= 0), then A is
        unchanged. If double precision factorization has been used (when
        no exception or warning is raised and `itera` < 0), A contains
        the factors L and U from the factorization A = PLU; the unit
        diagonal elements of L are not stored.

    ipiv : int, ndarray, shape (n)
        If no constraints are violated, the pivot indices that define
        the permutation matrix P; at the ith step row i of the matrix
        was interchanged with row `ipiv`[i-1]. `ipiv`[i-1] = i indicates
        a row interchange was not required. `ipiv` corresponds either to
        the single precision factorization (if no exception or warning
        is raised and `itera` >= 0) or to the double precision
        factorization (if no exception or warning is raised and `itera`
        < 0).

    x : float, ndarray, shape (n, nrhs)
        If no exception or warning is raised, the n by r solution matrix
        X.

    itera : int
        If `itera` > 0, iterative refinement has been successfully used
        and `itera` is the number of iterations carried out.

        If `itera` < 0, iterative refinement has failed for one of the
        reasons given below and double precision factorization has been
        carried out instead.

        `itera` = -1

            Taking into account machine parameters, and the values of n
            and nrhs, it is not worth working in single precision.

        `itera` = -2

            Overflow of an entry occurred when moving from double to
            single precision.

        `itera` = -3

            An intermediate single precision factorization failed.

        `itera` = -31

            The maximum permitted number of iterations was exceeded.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -2)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. The
            factorization has been completed, but the factor U is
            exactly singular, so the solution could not be computed.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Buttari, A, Dongarra, J, Langou, J, Langou, J, Luszczek, P and
    Kurzak, J, 2007, `Mixed precision iterative refinement techniques
    for the solution of dense linear systems`, International Journal of
    High Performance Computing Applications

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.dsgesv'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_a, 'a'), (_b, 'b')),
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ipiv = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n_py,),
    )
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py, _nrhs_py),
        sorder=_m_sorder,
    )
    try:
        _return_scalar = _b_lapacklin.dsgesv(
            _n,
            _nrhs,
            _a,
            _ipiv,
            _b,
            _x,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _a_py = _a
    _ipiv_py = _ipiv
    _x_py = _x
    _itera_py = _return_scalar
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'ipiv', 'x', 'itera'],
    )(
        _a_py,
        _ipiv_py,
        _x_py,
        _itera_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def dgetrf(a):
    u"""
    LU factorization of real m by n matrix.

    ``dgetrf`` computes the LU factorization of a real m by n matrix.

    For full information please refer to the NAG Library document for
    f07ad

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07adf.html

    Parameters
    ----------
    a : float, array-like, shape (m, n)
        The m by n matrix A.

    Returns
    -------
    a : float, ndarray, shape (m, n)
        The factors L and U from the factorization A = PLU; the unit
        diagonal elements of L are not stored.

    ipiv : int, ndarray, shape (min(m, n))
        The pivot indices that define the permutation matrix. At the ith
        step, if `ipiv`[i-1] > i then row i of the matrix A was
        interchanged with row `ipiv`[i-1], for i = 1,2,...,min(m, n).
        `ipiv`[i-1] <= i indicates that, at the ith step, a row
        interchange was not required.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter m.

            Constraint: m >= 0.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. The
            factorization has been completed, but the factor U is
            exactly singular, and division by zero will occur if it is
            used to solve a system of equations.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.dgetrf'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _ipiv = _b_utils._EngineIntArrayType.empty_ndarray(
        (min(_m_py, _n_py),),
    )
    try:
        _b_lapacklin.dgetrf(
            _m,
            _n,
            _a,
            _ipiv,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _a_py = _a
    _ipiv_py = _ipiv
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'ipiv'],
    )(
        _a_py,
        _ipiv_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def dgetrs(trans, a, ipiv, b):
    u"""
    Solution of real system of linear equations, multiple right-hand
    sides, matrix already factorized by ``dgetrf``.

    ``dgetrs`` solves a real system of linear equations with multiple
    right-hand sides,

        AX = B or A^TX = B,

    where A has been factorized by ``dgetrf``.

    For full information please refer to the NAG Library document for
    f07ae

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07aef.html

    Parameters
    ----------
    trans : str, length 1
        Indicates the form of the equations.

        `trans` = 'N'

            AX = B is solved for X.

        `trans` = 'T' or 'C'

            A^TX = B is solved for X.

    a : float, array-like, shape (n, n)
        The LU factorization of A, as returned by ``dgetrf``.

    ipiv : int, array-like, shape (n)
        The pivot indices, as returned by ``dgetrf``.

    b : float, array-like, shape (n, nrhs)
        The n by r right-hand side matrix B.

    Returns
    -------
    b : float, ndarray, shape (n, nrhs)
        The n by r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.dgetrs'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nrhs_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    _b_lapacklin.dgetrs(
        trans,
        _n,
        _nrhs,
        _a,
        _ipiv,
        _b,
    )
    _b_py = _b
    return _b_py

def dgeequ(a):
    u"""
    Computes row and column scalings intended to equilibrate a general
    real matrix and reduce its condition number.

    ``dgeequ`` computes diagonal scaling matrices D_R and D_C intended
    to equilibrate a real m by n matrix A and reduce its condition
    number.

    For full information please refer to the NAG Library document for
    f07af

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07aff.html

    Parameters
    ----------
    a : float, array-like, shape (m, n)
        The matrix A whose scaling factors are to be computed.

    Returns
    -------
    r : float, ndarray, shape (m)
        If no exception or warning is raised or `errno` > `m`, `r`
        contains the row scale factors, the diagonal elements of D_R.
        The elements of `r` will be positive.

    c : float, ndarray, shape (n)
        If no exception or warning is raised, `c` contains the column
        scale factors, the diagonal elements of D_C. The elements of `c`
        will be positive.

    rowcnd : float
        If no exception or warning is raised or `errno` > `m`, `rowcnd`
        contains the ratio of the smallest value of `r`[i-1] to the
        largest value of `r`[i-1]. If `rowcnd` >= 0.1 and `amax` is
        neither too large nor too small, it is not worth scaling by D_R.

    colcnd : float
        If no exception or warning is raised, `colcnd` contains the
        ratio of the smallest value of `c`[i-1] to the largest value of
        `c`[i-1].

        If `colcnd` >= 0.1, it is not worth scaling by D_C.

    amax : float
        max|a_{ij}|. If `amax` is very close to overflow or underflow,
        the matrix A should be scaled.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter m.

            Constraint: m >= 0.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` (i > 0) and (i<=m))
            Row *<value>* of A is exactly zero.

        (`errno` i > m)
            Column *<value>* of A is exactly zero.
    """

    fun_name = 'naginterfaces.library.lapacklin.dgeequ'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _r = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_m_py,),
    )
    _c = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _return_scalars = _b_lapacklin.dgeequ(
            _m,
            _n,
            _a,
            _r,
            _c,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _r_py = _r
    _c_py = _c
    _rowcnd_py = _return_scalars.rowcnd
    _colcnd_py = _return_scalars.colcnd
    _amax_py = _return_scalars.amax
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['r', 'c', 'rowcnd', 'colcnd', 'amax'],
    )(
        _r_py,
        _c_py,
        _rowcnd_py,
        _colcnd_py,
        _amax_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def dgecon(norm, n, a, anorm):
    u"""
    Estimate condition number of real matrix, matrix already factorized
    by ``dgetrf``.

    ``dgecon`` estimates the condition number of a real matrix A, where
    A has been factorized by ``dgetrf``.

    For full information please refer to the NAG Library document for
    f07ag

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07agf.html

    Parameters
    ----------
    norm : str, length 1
        Indicates whether kappa_1(A) or kappa_infinity (A) is estimated.

        `norm` = '1' or 'O'

            kappa_1(A) is estimated.

        `norm` = 'I'

            kappa_infinity (A) is estimated.

    n : int
        n, the order of the matrix A.

    a : float, array-like, shape (`n`, `n`)
        The LU factorization of A, as returned by ``dgetrf``.

    anorm : float
        If `norm` = '1' or 'O', the 1-norm of the **original** matrix A.

        If `norm` = 'I', the infinity-norm of the **original** matrix A.

        `anorm` may be computed by calling ``blas.dlange`` with the same
        value for the argument `norm`.

        `anorm` must be computed either **before** calling ``dgetrf`` or
        else from a **copy** of the original matrix A.

    Returns
    -------
    rcond : float
        An estimate of the reciprocal of the condition number of A.
        `rcond` is set to zero if exact singularity is detected or the
        estimate underflows. If `rcond` is less than machine precision,
        A is singular to working precision.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O' or 'I'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -5)
            On entry, error in parameter `anorm`.

            Constraint: `anorm` >= 0.0.

    References
    ----------
    Higham, N J, 1988, `FORTRAN codes for estimating the one-norm of a
    real or complex matrix, with applications to condition estimation`,
    ACM Trans. Math. Software (14), 381--396
    """

    fun_name = 'naginterfaces.library.lapacklin.dgecon'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _return_scalar = _b_lapacklin.dgecon(
        norm,
        n,
        _a,
        anorm,
    )
    _rcond_py = _return_scalar
    return _rcond_py

def dgerfs(trans, n, a, af, ipiv, b, x):
    u"""
    Refined solution with error bounds of real system of linear
    equations, multiple right-hand sides.

    ``dgerfs`` returns error bounds for the solution of a real system of
    linear equations with multiple right-hand sides, AX = B or A^TX = B.
    It improves the solution by iterative refinement, in order to reduce
    the backward error as much as possible.

    For full information please refer to the NAG Library document for
    f07ah

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07ahf.html

    Parameters
    ----------
    trans : str, length 1
        Indicates the form of the linear equations for which X is the
        computed solution.

        `trans` = 'N'

            The linear equations are of the form AX = B.

        `trans` = 'T' or 'C'

            The linear equations are of the form A^TX = B.

    n : int
        n, the order of the matrix A.

    a : float, array-like, shape (`n`, `n`)
        The n by n original matrix A as supplied to ``dgetrf``.

    af : float, array-like, shape (`n`, `n`)
        The LU factorization of A, as returned by ``dgetrf``.

    ipiv : int, array-like, shape (`n`)
        The pivot indices, as returned by ``dgetrf``.

    b : float, array-like, shape (`n`, nrhs)
        The n by r right-hand side matrix B.

    x : float, array-like, shape (`n`, nrhs)
        The n by r solution matrix X, as returned by ``dgetrs``.

    Returns
    -------
    x : float, ndarray, shape (`n`, nrhs)
        The improved solution matrix X.

    ferr : float, ndarray, shape (nrhs)
        `ferr`[j-1] contains an estimated error bound for the jth
        solution vector, that is, the jth column of X, for j =
        1,2,...,r.

    berr : float, ndarray, shape (nrhs)
        `berr`[j-1] contains the component-wise backward error bound
        beta for the jth solution vector, that is, the jth column of X,
        for j = 1,2,...,r.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -3)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.dgerfs'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nrhs_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _af = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'af'},
        af,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _b_lapacklin.dgerfs(
        trans,
        n,
        _nrhs,
        _a,
        _af,
        _ipiv,
        _b,
        _x,
        _ferr,
        _berr,
    )
    _x_py = _x
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'ferr', 'berr'],
    )(
        _x_py,
        _ferr_py,
        _berr_py,
    )
    return _return_data

def dgetri(a, ipiv):
    u"""
    Inverse of real matrix, matrix already factorized by ``dgetrf``.

    ``dgetri`` computes the inverse of a real matrix A, where A has been
    factorized by ``dgetrf``.

    For full information please refer to the NAG Library document for
    f07aj

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07ajf.html

    Parameters
    ----------
    a : float, array-like, shape (n, n)
        The LU factorization of A, as returned by ``dgetrf``.

    ipiv : int, array-like, shape (n)
        The pivot indices, as returned by ``dgetrf``.

    Returns
    -------
    a : float, ndarray, shape (n, n)
        The factorization is overwritten by the n by n matrix A^-1.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter n.

            Constraint: n >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is zero. U is singular,
            and the inverse of A cannot be computed.

    References
    ----------
    Du Croz, J J and Higham, N J, 1992, `Stability of methods for matrix
    inversion`, IMA J. Numer. Anal. (12), 1--19
    """

    fun_name = 'naginterfaces.library.lapacklin.dgetri'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    try:
        _b_lapacklin.dgetri(
            _n,
            _a,
            _ipiv,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _a_py = _a
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a'],
    )(
        _a_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _a_py

def zgesv(a, b):
    u"""
    Computes the solution to a complex system of linear equations.

    ``zgesv`` computes the solution to a complex system of linear
    equations

        AX = B,

    where A is an n by n matrix and X and B are n by r matrices.

    For full information please refer to the NAG Library document for
    f07an

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07anf.html

    Parameters
    ----------
    a : complex, array-like, shape (n, n)
        The n by n coefficient matrix A.

    b : complex, array-like, shape (n, nrhs)
        The n by r right-hand side matrix B.

    Returns
    -------
    a : complex, ndarray, shape (n, n)
        The factors L and U from the factorization A = PLU; the unit
        diagonal elements of L are not stored.

    ipiv : int, ndarray, shape (n)
        If no constraints are violated, the pivot indices that define
        the permutation matrix P; at the ith step row i of the matrix
        was interchanged with row `ipiv`[i-1]. `ipiv`[i-1] = i indicates
        a row interchange was not required.

    b : complex, ndarray, shape (n, nrhs)
        If no exception or warning is raised, the n by r solution matrix
        X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -2)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. The
            factorization has been completed, but the factor U is
            exactly singular, so the solution could not be computed.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.zgesv'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _ipiv = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _b_lapacklin.zgesv(
            _n,
            _nrhs,
            _a,
            _ipiv,
            _b,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _a_py = _a
    _b_py = _b
    _ipiv_py = _ipiv
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'ipiv', 'b'],
    )(
        _a_py,
        _ipiv_py,
        _b_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def zgesvx(fact, trans, nrhs, a, af, ipiv, equed, r, c, b):
    u"""
    Uses the LU factorization to compute the solution, error-bound and
    condition estimate for a complex system of linear equations.

    ``zgesvx`` uses the LU factorization to compute the solution to a
    complex system of linear equations

        AX = B or A^TX = B or A^HX = B,

    where A is an n by n matrix and X and B are n by r matrices.
    Error bounds on the solution and a condition estimate are also
    provided.

    For full information please refer to the NAG Library document for
    f07ap

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07apf.html

    Parameters
    ----------
    fact : str, length 1
        Specifies whether or not the factorized form of the matrix A is
        supplied on entry, and if not, whether the matrix A should be
        equilibrated before it is factorized.

        `fact` = 'F'

            `af` and `ipiv` contain the factorized form of A. If `equed`
            != 'N', the matrix A has been equilibrated with scaling
            factors given by `r` and `c`. `a`, `af` and `ipiv` are not
            modified.

        `fact` = 'N'

            The matrix A will be copied to `af` and factorized.

        `fact` = 'E'

            The matrix A will be equilibrated if necessary, then copied
            to `af` and factorized.

    trans : str, length 1
        Specifies the form of the system of equations.

        `trans` = 'N'

            AX = B (No transpose).

        `trans` = 'T'

            A^TX = B (Transpose).

        `trans` = 'C'

            A^HX = B (Conjugate transpose).

    nrhs : int
        r, the number of right-hand sides, i.e., the number of columns
        of the matrix B.

    a : complex, array-like, shape (n, n)
        The n by n matrix A.

        If `fact` = 'F' and `equed` != 'N', `a` must have been
        equilibrated by the scaling factors in `r` and/or `c`.

    af : complex, array-like, shape (n, n)
        If `fact` = 'F', `af` contains the factors L and U from the
        factorization A = PLU as computed by ``zgetrf``. If `equed` !=
        'N', `af` is the factorized form of the equilibrated matrix A.

        If `fact` = 'N' or 'E', `af` need not be set.

    ipiv : int, array-like, shape (n)
        If `fact` = 'F', `ipiv` contains the pivot indices from the
        factorization A = PLU as computed by ``zgetrf``; at the ith step
        row i of the matrix was interchanged with row `ipiv`[i-1].
        `ipiv`[i-1] = i indicates a row interchange was not required.

        If `fact` = 'N' or 'E', `ipiv` need not be set.

    equed : str, length 1
        If `fact` = 'N' or 'E', `equed` need not be set.

        If `fact` = 'F', `equed` must specify the form of the
        equilibration that was performed as follows:

            if `equed` = 'N', no equilibration;

            if `equed` = 'R', row equilibration, i.e., A has been
            premultiplied by D_R;

            if `equed` = 'C', column equilibration, i.e., A has been
            postmultiplied by D_C;

            if `equed` = 'B', both row and column equilibration, i.e., A
            has been replaced by D_R AD_C.

    r : float, array-like, shape (n)
        If `fact` = 'N' or 'E', `r` need not be set.

        If `fact` = 'F' and `equed` = 'R' or 'B', `r` must contain the
        row scale factors for A, D_R; each element of `r` must be
        positive.

    c : float, array-like, shape (n)
        If `fact` = 'N' or 'E', `c` need not be set.

        If `fact` = 'F' and `equed` = 'C' or 'B', `c` must contain the
        column scale factors for A, D_C; each element of `c` must be
        positive.

    b : complex, array-like, shape (n, `nrhs`)
        The n by r right-hand side matrix B.

    Returns
    -------
    a : complex, ndarray, shape (n, n)
        If `fact` = 'F' or 'N', or if `fact` = 'E' and `equed` = 'N',
        `a` is not modified.

        If `fact` = 'E' or `equed` != 'N', A is scaled as follows:

            if `equed` = 'R', A = D_R A;

            if `equed` = 'C', A = AD_C;

            if `equed` = 'B', A = D_R AD_C.

    af : complex, ndarray, shape (n, n)
        If `fact` = 'N', `af` returns the factors L and U from the
        factorization A = PLU of the original matrix A.

        If `fact` = 'E', `af` returns the factors L and U from the
        factorization A = PLU of the equilibrated matrix A (see the
        description of `a` for the form of the equilibrated matrix).

        If `fact` = 'F', `af` is unchanged from entry.

    ipiv : int, ndarray, shape (n)
        If `fact` = 'N', `ipiv` contains the pivot indices from the
        factorization A = PLU of the original matrix A.

        If `fact` = 'E', `ipiv` contains the pivot indices from the
        factorization A = PLU of the equilibrated matrix A.

        If `fact` = 'F', `ipiv` is unchanged from entry.

    equed : str, length 1
        If `fact` = 'F', `equed` is unchanged from entry.

        Otherwise, if no constraints are violated, `equed` specifies the
        form of equilibration that was performed as specified above.

    r : float, ndarray, shape (n)
        If `fact` = 'F', `r` is unchanged from entry.

        Otherwise, if no constraints are violated and `equed` = 'R' or
        'B', `r` contains the row scale factors for A, D_R, such that A
        is multiplied on the left by D_R; each element of `r` is
        positive.

    c : float, ndarray, shape (n)
        If `fact` = 'F', `c` is unchanged from entry.

        Otherwise, if no constraints are violated and `equed` = 'C' or
        'B', `c` contains the row scale factors for A, D_C; each element
        of `c` is positive.

    b : complex, ndarray, shape (n, `nrhs`)
        If `equed` = 'N', `b` is not modified.

        If `trans` = 'N' and `equed` = 'R' or 'B', `b` is overwritten by
        D_R B.

        If `trans` = 'T' or 'C' and `equed` = 'C' or 'B', `b` is
        overwritten by D_C B.

    x : complex, ndarray, shape (n, `nrhs`)
        If the function exits successfully or `errno` = `n` + 1, the n
        by r solution matrix X to the original system of equations. Note
        that the arrays A and B are modified on exit if `equed` != 'N',
        and the solution to the equilibrated system is D_C^-1X if
        `trans` = 'N' and `equed` = 'C' or 'B', or D_R^-1X if `trans` =
        'T' or 'C' and `equed` = 'R' or 'B'.

    rcond : float
        If no constraints are violated, an estimate of the reciprocal
        condition number of the matrix A (after equilibration if that is
        performed), computed as `rcond` = 1.0/(||A||_1||A^-1||_1).

    ferr : float, ndarray, shape (`nrhs`)
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the forward error bound for each computed solution
        vector, such that ||x^_j-x_j||_infinity/||x_j||_infinity <=
        `ferr`[j-1] where x^_j is the jth column of the computed
        solution returned in the array `x` and x_j is the corresponding
        column of the exact solution X. The estimate is as reliable as
        the estimate for `rcond`, and is almost always a slight
        overestimate of the true error.

    berr : float, ndarray, shape (`nrhs`)
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the component-wise relative backward error of each
        computed solution vector x^_j (i.e., the smallest relative
        change in any element of A or B that makes x^_j an exact
        solution).

    recip_growth_factor : float
        If no exception or warning is raised, the reciprocal pivot
        growth factor ||A||/||U||, where ||.|| denotes the maximum
        absolute element norm. If `recip_growth_factor`≪1, the stability
        of the LU factorization of (equilibrated) A could be poor. This
        also means that the solution `x`, condition estimate `rcond`,
        and forward error bound `ferr` could be unreliable. If the
        factorization fails with `errno` in 1 ... `n`, then
        `recip_growth_factor` contains the reciprocal pivot growth
        factor for the leading errno columns of A.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `fact`.

            Constraint: `fact` = 'F', 'N' or 'E'.

        (`errno` -2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` -3)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -4)
            On entry, error in parameter `nrhs`.

            Constraint: `nrhs` >= 0.

        (`errno` -10)
            On entry, error in parameter `equed`.

            Constraint: `equed` = 'N', 'R', 'C' or 'B'.

        (`errno` -11)
            On entry, error in parameter `r`.

        (`errno` -12)
            On entry, error in parameter `c`.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` (i > 0) and (i<=n))
            Element *<value>* of the diagonal is exactly zero. The
            factorization has been completed, but the factor U is
            exactly singular, so the solution and error bounds could not
            be computed. `rcond` = 0.0 is returned.

        (`errno` n+1)
            U is nonsingular, but `rcond` is less than machine
            precision, meaning that the matrix is singular to working
            precision. Nevertheless, the solution and error bounds are
            computed because there are a number of situations where the
            computed solution can be more accurate than the value of
            `rcond` would suggest.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    fun_name = 'naginterfaces.library.lapacklin.zgesvx'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _af = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'af'},
        af,
        copy=True,
    )
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_a, 'a'), (_af, 'af'), (_b, 'b')),
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=True,
    )
    _r = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'r'},
        r,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _c = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nrhs = _b_utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineIntScalarType.to_py(_nrhs)
    _x = _b_utils._EngineComplex128ArrayType.empty_ndarray(
        (_n_py, _nrhs_py),
        sorder=_m_sorder,
    )
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    try:
        _return_scalars = _b_lapacklin.zgesvx(
            fact,
            trans,
            _n,
            _nrhs,
            _a,
            _af,
            _ipiv,
            equed,
            _r,
            _c,
            _b,
            _x,
            _ferr,
            _berr,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _a_py = _a
    _af_py = _af
    _ipiv_py = _ipiv
    _equed_py = _return_scalars.equed
    _r_py = _r
    _c_py = _c
    _b_py = _b
    _x_py = _x
    _rcond_py = _return_scalars.rcond
    _ferr_py = _ferr
    _berr_py = _berr
    _recip_growth_factor_py = _return_scalars.recip_growth_factor
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'af', 'ipiv', 'equed', 'r', 'c', 'b', 'x', 'rcond', 'ferr', 'berr', 'recip_growth_factor'],
    )(
        _a_py,
        _af_py,
        _ipiv_py,
        _equed_py,
        _r_py,
        _c_py,
        _b_py,
        _x_py,
        _rcond_py,
        _ferr_py,
        _berr_py,
        _recip_growth_factor_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def zcgesv(a, b):
    u"""
    Computes the solution to a complex system of linear equations using
    mixed precision arithmetic.

    ``zcgesv`` computes the solution to a complex system of linear
    equations

        AX = B,

    where A is an n by n matrix and X and B are n by r matrices.

    For full information please refer to the NAG Library document for
    f07aq

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07aqf.html

    Parameters
    ----------
    a : complex, array-like, shape (n, n)
        The n by n coefficient matrix A.

    b : complex, array-like, shape (n, nrhs)
        The n by r right-hand side matrix B.

    Returns
    -------
    a : complex, ndarray, shape (n, n)
        If iterative refinement has been successfully used (i.e., if no
        exception or warning is raised and `itera` >= 0), then A is
        unchanged. If double precision factorization has been used (when
        no exception or warning is raised and `itera` < 0), A contains
        the factors L and U from the factorization A = PLU; the unit
        diagonal elements of L are not stored.

    ipiv : int, ndarray, shape (n)
        If no constraints are violated, the pivot indices that define
        the permutation matrix P; at the ith step row i of the matrix
        was interchanged with row `ipiv`[i-1]. `ipiv`[i-1] = i indicates
        a row interchange was not required. `ipiv` corresponds either to
        the single precision factorization (if no exception or warning
        is raised and `itera` >= 0) or to the double precision
        factorization (if no exception or warning is raised and `itera`
        < 0).

    x : complex, ndarray, shape (n, nrhs)
        If no exception or warning is raised, the n by r solution matrix
        X.

    itera : int
        If `itera` > 0, iterative refinement has been successfully used
        and `itera` is the number of iterations carried out.

        If `itera` < 0, iterative refinement has failed for one of the
        reasons given below and double precision factorization has been
        carried out instead.

        `itera` = -1

            Taking into account machine parameters, and the values of n
            and nrhs, it is not worth working in single precision.

        `itera` = -2

            Overflow of an entry occurred when moving from double to
            single precision.

        `itera` = -3

            An intermediate single precision factorization failed.

        `itera` = -31

            The maximum permitted number of iterations was exceeded.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -2)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. The
            factorization has been completed, but the factor U is
            exactly singular, so the solution could not be computed.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Buttari, A, Dongarra, J, Langou, J, Langou, J, Luszczek, P and
    Kurzak, J, 2007, `Mixed precision iterative refinement techniques
    for the solution of dense linear systems`, International Journal of
    High Performance Computing Applications

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.zcgesv'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_a, 'a'), (_b, 'b')),
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ipiv = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n_py,),
    )
    _x = _b_utils._EngineComplex128ArrayType.empty_ndarray(
        (_n_py, _nrhs_py),
        sorder=_m_sorder,
    )
    try:
        _return_scalar = _b_lapacklin.zcgesv(
            _n,
            _nrhs,
            _a,
            _ipiv,
            _b,
            _x,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _a_py = _a
    _ipiv_py = _ipiv
    _x_py = _x
    _itera_py = _return_scalar
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'ipiv', 'x', 'itera'],
    )(
        _a_py,
        _ipiv_py,
        _x_py,
        _itera_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def zgetrf(a):
    u"""
    LU factorization of complex m by n matrix.

    ``zgetrf`` computes the LU factorization of a complex m by n matrix.

    For full information please refer to the NAG Library document for
    f07ar

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07arf.html

    Parameters
    ----------
    a : complex, array-like, shape (m, n)
        The m by n matrix A.

    Returns
    -------
    a : complex, ndarray, shape (m, n)
        The factors L and U from the factorization A = PLU; the unit
        diagonal elements of L are not stored.

    ipiv : int, ndarray, shape (min(m, n))
        The pivot indices that define the permutation matrix. At the ith
        step, if `ipiv`[i-1] > i then row i of the matrix A was
        interchanged with row `ipiv`[i-1], for i = 1,2,...,min(m, n).
        `ipiv`[i-1] <= i indicates that, at the ith step, a row
        interchange was not required.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter m.

            Constraint: m >= 0.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. The
            factorization has been completed, but the factor U is
            exactly singular, and division by zero will occur if it is
            used to solve a system of equations.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.zgetrf'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _m_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _ipiv = _b_utils._EngineIntArrayType.empty_ndarray(
        (min(_m_py, _n_py),),
    )
    try:
        _b_lapacklin.zgetrf(
            _m,
            _n,
            _a,
            _ipiv,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _a_py = _a
    _ipiv_py = _ipiv
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'ipiv'],
    )(
        _a_py,
        _ipiv_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def zgetrs(trans, a, ipiv, b):
    u"""
    Solution of complex system of linear equations, multiple right-hand
    sides, matrix already factorized by ``zgetrf``.

    ``zgetrs`` solves a complex system of linear equations with multiple
    right-hand sides,

        AX = B, A^TX = B or A^HX = B,

    where A has been factorized by ``zgetrf``.

    For full information please refer to the NAG Library document for
    f07as

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07asf.html

    Parameters
    ----------
    trans : str, length 1
        Indicates the form of the equations.

        `trans` = 'N'

            AX = B is solved for X.

        `trans` = 'T'

            A^TX = B is solved for X.

        `trans` = 'C'

            A^HX = B is solved for X.

    a : complex, array-like, shape (n, n)
        The LU factorization of A, as returned by ``zgetrf``.

    ipiv : int, array-like, shape (n)
        The pivot indices, as returned by ``zgetrf``.

    b : complex, array-like, shape (n, nrhs)
        The n by r right-hand side matrix B.

    Returns
    -------
    b : complex, ndarray, shape (n, nrhs)
        The n by r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.zgetrs'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
    )
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    _b_lapacklin.zgetrs(
        trans,
        _n,
        _nrhs,
        _a,
        _ipiv,
        _b,
    )
    _b_py = _b
    return _b_py

def zgeequ(a):
    u"""
    Computes row and column scalings intended to equilibrate a general
    complex matrix and reduce its condition number.

    ``zgeequ`` computes real diagonal scaling matrices D_R and D_C
    intended to equilibrate a complex m by n matrix A and reduce its
    condition number.

    For full information please refer to the NAG Library document for
    f07at

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07atf.html

    Parameters
    ----------
    a : complex, array-like, shape (m, n)
        The matrix A whose scaling factors are to be computed.

    Returns
    -------
    r : float, ndarray, shape (m)
        If no exception or warning is raised or `errno` > `m`, `r`
        contains the row scale factors, the diagonal elements of D_R.
        The elements of `r` will be positive.

    c : float, ndarray, shape (n)
        If no exception or warning is raised, `c` contains the column
        scale factors, the diagonal elements of D_C. The elements of `c`
        will be positive.

    rowcnd : float
        If no exception or warning is raised or `errno` > `m`, `rowcnd`
        contains the ratio of the smallest value of `r`[i-1] to the
        largest value of `r`[i-1]. If `rowcnd` >= 0.1 and `amax` is
        neither too large nor too small, it is not worth scaling by D_R.

    colcnd : float
        If no exception or warning is raised, `colcnd` contains the
        ratio of the smallest value of `c`[i-1] to the largest value of
        `c`[i-1].

        If `colcnd` >= 0.1, it is not worth scaling by D_C.

    amax : float
        max|a_{ij}|. If `amax` is very close to overflow or underflow,
        the matrix A should be scaled.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter m.

            Constraint: m >= 0.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` (i > 0) and (i<=m))
            Row *<value>* of A is exactly zero.

        (`errno` i > m)
            Column *<value>* of A is exactly zero.
    """

    fun_name = 'naginterfaces.library.lapacklin.zgeequ'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
    )
    _m_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _r = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_m_py,),
    )
    _c = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _return_scalars = _b_lapacklin.zgeequ(
            _m,
            _n,
            _a,
            _r,
            _c,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _r_py = _r
    _c_py = _c
    _rowcnd_py = _return_scalars.rowcnd
    _colcnd_py = _return_scalars.colcnd
    _amax_py = _return_scalars.amax
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['r', 'c', 'rowcnd', 'colcnd', 'amax'],
    )(
        _r_py,
        _c_py,
        _rowcnd_py,
        _colcnd_py,
        _amax_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def zgecon(norm, a, anorm):
    u"""
    Estimate condition number of complex matrix, matrix already
    factorized by ``zgetrf``.

    ``zgecon`` estimates the condition number of a complex matrix A,
    where A has been factorized by ``zgetrf``.

    For full information please refer to the NAG Library document for
    f07au

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07auf.html

    Parameters
    ----------
    norm : str, length 1
        Indicates whether kappa_1(A) or kappa_infinity (A) is estimated.

        `norm` = '1' or 'O'

            kappa_1(A) is estimated.

        `norm` = 'I'

            kappa_infinity (A) is estimated.

    a : complex, array-like, shape (n, n)
        The LU factorization of A, as returned by ``zgetrf``.

    anorm : float
        If `norm` = '1' or 'O', the 1-norm of the **original** matrix A.

        If `norm` = 'I', the infinity-norm of the **original** matrix A.

        `anorm` may be computed by calling ``blas.zlange`` with the same
        value for the argument `norm`.

        `anorm` must be computed either **before** calling ``zgetrf`` or
        else from a **copy** of the original matrix A.

    Returns
    -------
    rcond : float
        An estimate of the reciprocal of the condition number of A.
        `rcond` is set to zero if exact singularity is detected or the
        estimate underflows. If `rcond` is less than machine precision,
        A is singular to working precision.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O' or 'I'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -5)
            On entry, error in parameter `anorm`.

            Constraint: `anorm` >= 0.0.

    References
    ----------
    Higham, N J, 1988, `FORTRAN codes for estimating the one-norm of a
    real or complex matrix, with applications to condition estimation`,
    ACM Trans. Math. Software (14), 381--396
    """

    fun_name = 'naginterfaces.library.lapacklin.zgecon'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _return_scalar = _b_lapacklin.zgecon(
        norm,
        _n,
        _a,
        anorm,
    )
    _rcond_py = _return_scalar
    return _rcond_py

def zgerfs(trans, n, a, af, ipiv, b, x):
    u"""
    Refined solution with error bounds of complex system of linear
    equations, multiple right-hand sides.

    ``zgerfs`` returns error bounds for the solution of a complex system
    of linear equations with multiple right-hand sides, AX = B, A^TX = B
    or A^HX = B.
    It improves the solution by iterative refinement, in order to reduce
    the backward error as much as possible.

    For full information please refer to the NAG Library document for
    f07av

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07avf.html

    Parameters
    ----------
    trans : str, length 1
        Indicates the form of the linear equations for which X is the
        computed solution as follows:

        `trans` = 'N'

            The linear equations are of the form AX = B.

        `trans` = 'T'

            The linear equations are of the form A^TX = B.

        `trans` = 'C'

            The linear equations are of the form A^HX = B.

    n : int
        n, the order of the matrix A.

    a : complex, array-like, shape (`n`, `n`)
        The n by n original matrix A as supplied to ``zgetrf``.

    af : complex, array-like, shape (`n`, `n`)
        The LU factorization of A, as returned by ``zgetrf``.

    ipiv : int, array-like, shape (`n`)
        The pivot indices, as returned by ``zgetrf``.

    b : complex, array-like, shape (`n`, nrhs)
        The n by r right-hand side matrix B.

    x : complex, array-like, shape (`n`, nrhs)
        The n by r solution matrix X, as returned by ``zgetrs``.

    Returns
    -------
    x : complex, ndarray, shape (`n`, nrhs)
        The improved solution matrix X.

    ferr : float, ndarray, shape (nrhs)
        `ferr`[j-1] contains an estimated error bound for the jth
        solution vector, that is, the jth column of X, for j =
        1,2,...,r.

    berr : float, ndarray, shape (nrhs)
        `berr`[j-1] contains the component-wise backward error bound
        beta for the jth solution vector, that is, the jth column of X,
        for j = 1,2,...,r.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -3)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.zgerfs'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
    )
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
    )
    _af = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'af'},
        af,
        copy=False,
    )
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
    )
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _b_lapacklin.zgerfs(
        trans,
        n,
        _nrhs,
        _a,
        _af,
        _ipiv,
        _b,
        _x,
        _ferr,
        _berr,
    )
    _x_py = _x
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'ferr', 'berr'],
    )(
        _x_py,
        _ferr_py,
        _berr_py,
    )
    return _return_data

def zgetri(a, ipiv):
    u"""
    Inverse of complex matrix, matrix already factorized by ``zgetrf``.

    ``zgetri`` computes the inverse of a complex matrix A, where A has
    been factorized by ``zgetrf``.

    For full information please refer to the NAG Library document for
    f07aw

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07awf.html

    Parameters
    ----------
    a : complex, array-like, shape (n, n)
        The LU factorization of A, as returned by ``zgetrf``.

    ipiv : int, array-like, shape (n)
        The pivot indices, as returned by ``zgetrf``.

    Returns
    -------
    a : complex, ndarray, shape (n, n)
        The factorization is overwritten by the n by n matrix A^-1.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter n.

            Constraint: n >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is zero. U is singular,
            and the inverse of A cannot be computed.

    References
    ----------
    Du Croz, J J and Higham, N J, 1992, `Stability of methods for matrix
    inversion`, IMA J. Numer. Anal. (12), 1--19
    """

    fun_name = 'naginterfaces.library.lapacklin.zgetri'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    try:
        _b_lapacklin.zgetri(
            _n,
            _a,
            _ipiv,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _a_py = _a
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a'],
    )(
        _a_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _a_py

def dgbsv(kl, ku, ab, b):
    u"""
    Computes the solution to a real banded system of linear equations.

    ``dgbsv`` computes the solution to a real system of linear equations

        AX = B,

    where A is an n by n band matrix, with k_l subdiagonals and k_u
    superdiagonals, and X and B are n by r matrices.

    For full information please refer to the NAG Library document for
    f07ba

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07baf.html

    Parameters
    ----------
    kl : int
        k_l, the number of subdiagonals within the band of the matrix A.

    ku : int
        k_u, the number of superdiagonals within the band of the matrix
        A.

    ab : float, array-like, shape (2*`kl`+`ku`+1, n)
        The n by n coefficient matrix A.

        See [Further Comments] for further details.

    b : float, array-like, shape (n, nrhs)
        The n by r right-hand side matrix B.

    Returns
    -------
    ab : float, ndarray, shape (2*`kl`+`ku`+1, n)
        If `errno` >= 0, `ab` is overwritten by details of the
        factorization.

    ipiv : int, ndarray, shape (n)
        If no constraints are violated, the pivot indices that define
        the permutation matrix P; at the ith step row i of the matrix
        was interchanged with row `ipiv`[i-1]. `ipiv`[i-1] = i indicates
        a row interchange was not required.

    b : float, ndarray, shape (n, nrhs)
        If no exception or warning is raised, the n by r solution matrix
        X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -2)
            On entry, error in parameter `kl`.

            Constraint: `kl` >= 0.

        (`errno` -3)
            On entry, error in parameter `ku`.

            Constraint: `ku` >= 0.

        (`errno` -4)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. The
            factorization has been completed, but the factor U is
            exactly singular, so the solution could not be computed.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.dgbsv'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ipiv = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _b_lapacklin.dgbsv(
            _n,
            kl,
            ku,
            _nrhs,
            _ab,
            _ipiv,
            _b,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _ab_py = _ab
    _b_py = _b
    _ipiv_py = _ipiv
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['ab', 'ipiv', 'b'],
    )(
        _ab_py,
        _ipiv_py,
        _b_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def dgbsvx(fact, trans, n, kl, ku, nrhs, ab, afb, ipiv, equed, r, c, b):
    u"""
    Uses the LU factorization to compute the solution, error-bound and
    condition estimate for a real banded system of linear equations.

    ``dgbsvx`` uses the LU factorization to compute the solution to a
    real system of linear equations

        AX = B or A^TX = B,

    where A is an n by n band matrix with k_l subdiagonals and k_u
    superdiagonals, and X and B are n by r matrices.
    Error bounds on the solution and a condition estimate are also
    provided.

    For full information please refer to the NAG Library document for
    f07bb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07bbf.html

    Parameters
    ----------
    fact : str, length 1
        Specifies whether or not the factorized form of the matrix A is
        supplied on entry, and if not, whether the matrix A should be
        equilibrated before it is factorized.

        `fact` = 'F'

            `afb` and `ipiv` contain the factorized form of A. If
            `equed` != 'N', the matrix A has been equilibrated with
            scaling factors given by `r` and `c`. `ab`, `afb` and `ipiv`
            are not modified.

        `fact` = 'N'

            The matrix A will be copied to `afb` and factorized.

        `fact` = 'E'

            The matrix A will be equilibrated if necessary, then copied
            to `afb` and factorized.

    trans : str, length 1
        Specifies the form of the system of equations.

        `trans` = 'N'

            AX = B (No transpose).

        `trans` = 'T' or 'C'

            A^TX = B (Transpose).

    n : int
        n, the number of linear equations, i.e., the order of the matrix
        A.

    kl : int
        k_l, the number of subdiagonals within the band of the matrix A.

    ku : int
        k_u, the number of superdiagonals within the band of the matrix
        A.

    nrhs : int
        r, the number of right-hand sides, i.e., the number of columns
        of the matrix B.

    ab : float, array-like, shape (`kl`+`ku`+1, `n`)
        The n by n coefficient matrix A.

        See [Further Comments] for further details.

        If `fact` = 'F' and `equed` != 'N', A must have been
        equilibrated by the scaling factors in `r` and/or `c`.

    afb : float, array-like, shape (2*`kl`+`ku`+1, `n`)
        If `fact` = 'N' or 'E', `afb` need not be set.

        If `fact` = 'F', details of the LU factorization of the n by n
        band matrix A, as computed by ``dgbtrf``.

        If `equed` != 'N', `afb` is the factorized form of the
        equilibrated matrix A.

    ipiv : int, array-like, shape (`n`)
        If `fact` = 'N' or 'E', `ipiv` need not be set.

        If `fact` = 'F', `ipiv` contains the pivot indices from the
        factorization A = LU, as computed by ``dgbtrf``; row i of the
        matrix was interchanged with row `ipiv`[i-1].

    equed : str, length 1
        If `fact` = 'N' or 'E', `equed` need not be set.

        If `fact` = 'F', `equed` must specify the form of the
        equilibration that was performed as follows:

            if `equed` = 'N', no equilibration;

            if `equed` = 'R', row equilibration, i.e., A has been
            premultiplied by D_R;

            if `equed` = 'C', column equilibration, i.e., A has been
            postmultiplied by D_C;

            if `equed` = 'B', both row and column equilibration, i.e., A
            has been replaced by D_R AD_C.

    r : float, array-like, shape (`n`)
        If `fact` = 'N' or 'E', `r` need not be set.

        If `fact` = 'F' and `equed` = 'R' or 'B', `r` must contain the
        row scale factors for A, D_R; each element of `r` must be
        positive.

    c : float, array-like, shape (`n`)
        If `fact` = 'N' or 'E', `c` need not be set.

        If `fact` = 'F' and `equed` = 'C' or 'B', `c` must contain the
        column scale factors for A, D_C; each element of `c` must be
        positive.

    b : float, array-like, shape (`n`, `nrhs`)
        The n by r right-hand side matrix B.

    Returns
    -------
    ab : float, ndarray, shape (`kl`+`ku`+1, `n`)
        If `fact` = 'F' or 'N', or if `fact` = 'E' and `equed` = 'N',
        `ab` is not modified.

        If `equed` != 'N' then, if no constraints are violated, A is
        scaled as follows:

            if `equed` = 'R', A = D_r A;

            if `equed` = 'C', A = AD_c;

            if `equed` = 'B', A = D_r AD_c.

    afb : float, ndarray, shape (2*`kl`+`ku`+1, `n`)
        If `fact` = 'F', `afb` is unchanged from entry.

        Otherwise, if no constraints are violated, then if `fact` = 'N',
        `afb` returns details of the LU factorization of the band matrix
        A, and if `fact` = 'E', `afb` returns details of the LU
        factorization of the equilibrated band matrix A (see the
        description of `ab` for the form of the equilibrated matrix).

    ipiv : int, ndarray, shape (`n`)
        If `fact` = 'F', `ipiv` is unchanged from entry.

        Otherwise, if no constraints are violated, `ipiv` contains the
        pivot indices that define the permutation matrix P; at the ith
        step row i of the matrix was interchanged with row `ipiv`[i-1].
        `ipiv`[i-1] = i indicates a row interchange was not required.

        If `fact` = 'N', the pivot indices are those corresponding to
        the factorization A = LU of the original matrix A.

        If `fact` = 'E', the pivot indices are those corresponding to
        the factorization of A = LU of the equilibrated matrix A.

    equed : str, length 1
        If `fact` = 'F', `equed` is unchanged from entry.

        Otherwise, if no constraints are violated, `equed` specifies the
        form of equilibration that was performed as specified above.

    r : float, ndarray, shape (`n`)
        If `fact` = 'F', `r` is unchanged from entry.

        Otherwise, if no constraints are violated and `equed` = 'R' or
        'B', `r` contains the row scale factors for A, D_R, such that A
        is multiplied on the left by D_R; each element of `r` is
        positive.

    c : float, ndarray, shape (`n`)
        If `fact` = 'F', `c` is unchanged from entry.

        Otherwise, if no constraints are violated and `equed` = 'C' or
        'B', `c` contains the row scale factors for A, D_C; each element
        of `c` is positive.

    b : float, ndarray, shape (`n`, `nrhs`)
        If `equed` = 'N', `b` is not modified.

        If `trans` = 'N' and `equed` = 'R' or 'B', `b` is overwritten by
        D_R B.

        If `trans` = 'T' or 'C' and `equed` = 'C' or 'B', `b` is
        overwritten by D_C B.

    x : float, ndarray, shape (`n`, `nrhs`)
        If the function exits successfully or `errno` = `n` + 1, the n
        by r solution matrix X to the original system of equations. Note
        that the arrays A and B are modified on exit if `equed` != 'N',
        and the solution to the equilibrated system is D_C^-1X if
        `trans` = 'N' and `equed` = 'C' or 'B', or D_R^-1X if `trans` =
        'T' or 'C' and `equed` = 'R' or 'B'.

    rcond : float
        If no constraints are violated, an estimate of the reciprocal
        condition number of the matrix A (after equilibration if that is
        performed), computed as `rcond` = 1.0/(||A||_1||A^-1||_1).

    ferr : float, ndarray, shape (`nrhs`)
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the forward error bound for each computed solution
        vector, such that ||x^_j-x_j||_infinity/||x_j||_infinity <=
        `ferr`[j-1] where x^_j is the jth column of the computed
        solution returned in the array `x` and x_j is the corresponding
        column of the exact solution X. The estimate is as reliable as
        the estimate for `rcond`, and is almost always a slight
        overestimate of the true error.

    berr : float, ndarray, shape (`nrhs`)
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the component-wise relative backward error of each
        computed solution vector x^_j (i.e., the smallest relative
        change in any element of A or B that makes x^_j an exact
        solution).

    recip_growth_factor : float
        If no exception or warning is raised, the reciprocal pivot
        growth factor ||A||/||U||, where ||.|| denotes the maximum
        absolute element norm. If `recip_growth_factor`≪1, the stability
        of the LU factorization of (equilibrated) A could be poor. This
        also means that the solution `x`, condition estimate `rcond`,
        and forward error bound `ferr` could be unreliable. If the
        factorization fails with `errno` in 1 ... `n`, then
        `recip_growth_factor` contains the reciprocal pivot growth
        factor for the leading errno columns of A.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `fact`.

            Constraint: `fact` = 'F', 'N' or 'E'.

        (`errno` -2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` -3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -4)
            On entry, error in parameter `kl`.

            Constraint: `kl` >= 0.

        (`errno` -5)
            On entry, error in parameter `ku`.

            Constraint: `ku` >= 0.

        (`errno` -6)
            On entry, error in parameter `nrhs`.

            Constraint: `nrhs` >= 0.

        (`errno` -12)
            On entry, error in parameter `equed`.

            Constraint: `equed` = 'N', 'R', 'C' or 'B'.

        (`errno` -13)
            On entry, error in parameter `r`.

        (`errno` -14)
            On entry, error in parameter `c`.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` (i > 0) and (i<=`n`))
            Element *<value>* of the diagonal is exactly zero. The
            factorization has been completed, but the factor U is
            exactly singular, so the solution and error bounds could not
            be computed. `rcond` = 0.0 is returned.

        (`errno` `n`+1)
            U is nonsingular, but `rcond` is less than machine
            precision, meaning that the matrix is singular to working
            precision. Nevertheless, the solution and error bounds are
            computed because there are a number of situations where the
            computed solution can be more accurate than the value of
            `rcond` would suggest.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    fun_name = 'naginterfaces.library.lapacklin.dgbsvx'

    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _afb = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'afb'},
        afb,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_ab, 'ab'), (_afb, 'afb'), (_b, 'b')),
    )
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=True,
    )
    _r = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'r'},
        r,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _c = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _nrhs = _b_utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineIntScalarType.to_py(_nrhs)
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py, _nrhs_py),
        sorder=_m_sorder,
    )
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    try:
        _return_scalars = _b_lapacklin.dgbsvx(
            fact,
            trans,
            _n,
            kl,
            ku,
            _nrhs,
            _ab,
            _afb,
            _ipiv,
            equed,
            _r,
            _c,
            _b,
            _x,
            _ferr,
            _berr,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _ab_py = _ab
    _afb_py = _afb
    _ipiv_py = _ipiv
    _equed_py = _return_scalars.equed
    _r_py = _r
    _c_py = _c
    _b_py = _b
    _x_py = _x
    _rcond_py = _return_scalars.rcond
    _ferr_py = _ferr
    _berr_py = _berr
    _recip_growth_factor_py = _return_scalars.recip_growth_factor
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['ab', 'afb', 'ipiv', 'equed', 'r', 'c', 'b', 'x', 'rcond', 'ferr', 'berr', 'recip_growth_factor'],
    )(
        _ab_py,
        _afb_py,
        _ipiv_py,
        _equed_py,
        _r_py,
        _c_py,
        _b_py,
        _x_py,
        _rcond_py,
        _ferr_py,
        _berr_py,
        _recip_growth_factor_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def dgbtrf(m, kl, ku, ab):
    u"""
    LU factorization of real m by n band matrix.

    ``dgbtrf`` computes the LU factorization of a real m by n band
    matrix.

    For full information please refer to the NAG Library document for
    f07bd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07bdf.html

    Parameters
    ----------
    m : int
        m, the number of rows of the matrix A.

    kl : int
        k_l, the number of subdiagonals within the band of the matrix A.

    ku : int
        k_u, the number of superdiagonals within the band of the matrix
        A.

    ab : float, array-like, shape (2*`kl`+`ku`+1, n)
        The m by n matrix A.

        See [Further Comments] for further details.

    Returns
    -------
    ab : float, ndarray, shape (2*`kl`+`ku`+1, n)
        If `errno` >= 0, `ab` is overwritten by details of the
        factorization.

    ipiv : int, ndarray, shape (min(`m`, n))
        The pivot indices that define the permutation matrix. At the ith
        step, if `ipiv`[i-1] > i then row i of the matrix A was
        interchanged with row `ipiv`[i-1], for i = 1,2,...,min(m, n).
        `ipiv`[i-1] <= i indicates that, at the ith step, a row
        interchange was not required.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `m`.

            Constraint: `m` >= 0.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter `kl`.

            Constraint: `kl` >= 0.

        (`errno` -4)
            On entry, error in parameter `ku`.

            Constraint: `ku` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. The
            factorization has been completed, but the factor U is
            exactly singular, and division by zero will occur if it is
            used to solve a system of equations.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.dgbtrf'

    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _m = _b_utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = _b_utils._EngineIntScalarType.to_py(_m)
    _ipiv = _b_utils._EngineIntArrayType.empty_ndarray(
        (min(_m_py, _n_py),),
    )
    try:
        _b_lapacklin.dgbtrf(
            _m,
            _n,
            kl,
            ku,
            _ab,
            _ipiv,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _ab_py = _ab
    _ipiv_py = _ipiv
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['ab', 'ipiv'],
    )(
        _ab_py,
        _ipiv_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def dgbtrs(trans, kl, ku, ab, ipiv, b):
    u"""
    Solution of real band system of linear equations, multiple
    right-hand sides, matrix already factorized by ``dgbtrf``.

    ``dgbtrs`` solves a real band system of linear equations with
    multiple right-hand sides,

        AX = B or A^TX = B,

    where A has been factorized by ``dgbtrf``.

    For full information please refer to the NAG Library document for
    f07be

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07bef.html

    Parameters
    ----------
    trans : str, length 1
        Indicates the form of the equations.

        `trans` = 'N'

            AX = B is solved for X.

        `trans` = 'T' or 'C'

            A^TX = B is solved for X.

    kl : int
        k_l, the number of subdiagonals within the band of the matrix A.

    ku : int
        k_u, the number of superdiagonals within the band of the matrix
        A.

    ab : float, array-like, shape (2*`kl`+`ku`+1, n)
        The LU factorization of A, as returned by ``dgbtrf``.

    ipiv : int, array-like, shape (n)
        The pivot indices, as returned by ``dgbtrf``.

    b : float, array-like, shape (n, nrhs)
        The n by r right-hand side matrix B.

    Returns
    -------
    b : float, ndarray, shape (n, nrhs)
        The n by r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter `kl`.

            Constraint: `kl` >= 0.

        (`errno` -4)
            On entry, error in parameter `ku`.

            Constraint: `ku` >= 0.

        (`errno` -5)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.dgbtrs'

    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nrhs_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    _b_lapacklin.dgbtrs(
        trans,
        _n,
        kl,
        ku,
        _nrhs,
        _ab,
        _ipiv,
        _b,
    )
    _b_py = _b
    return _b_py

def dgbequ(m, kl, ku, ab):
    u"""
    Computes row and column scalings intended to equilibrate a real
    banded matrix and reduce its condition number.

    ``dgbequ`` computes diagonal scaling matrices D_R and D_C intended
    to equilibrate a real m by n band matrix A of band width
    (k_l+k_u+1), and reduce its condition number.

    For full information please refer to the NAG Library document for
    f07bf

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07bff.html

    Parameters
    ----------
    m : int
        m, the number of rows of the matrix A.

    kl : int
        k_l, the number of subdiagonals of the matrix A.

    ku : int
        k_u, the number of superdiagonals of the matrix A.

    ab : float, array-like, shape (`kl`+`ku`+1, n)
        The m by n band matrix A whose scaling factors are to be
        computed.

        See [Further Comments] for further details.

    Returns
    -------
    r : float, ndarray, shape (`m`)
        If no exception or warning is raised or `errno` > `m`, `r`
        contains the row scale factors, the diagonal elements of D_R.
        The elements of `r` will be positive.

    c : float, ndarray, shape (n)
        If no exception or warning is raised, `c` contains the column
        scale factors, the diagonal elements of D_C. The elements of `c`
        will be positive.

    rowcnd : float
        If no exception or warning is raised or `errno` > `m`, `rowcnd`
        contains the ratio of the smallest value of `r`[i-1] to the
        largest value of `r`[i-1]. If `rowcnd` >= 0.1 and `amax` is
        neither too large nor too small, it is not worth scaling by D_R.

    colcnd : float
        If no exception or warning is raised, `colcnd` contains the
        ratio of the smallest value of `c`[i-1] to the largest value of
        `c`[i-1].

        If `colcnd` >= 0.1, it is not worth scaling by D_C.

    amax : float
        max|a_{ij}|. If `amax` is very close to overflow or underflow,
        the matrix A should be scaled.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `m`.

            Constraint: `m` >= 0.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter `kl`.

            Constraint: `kl` >= 0.

        (`errno` -4)
            On entry, error in parameter `ku`.

            Constraint: `ku` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` (i > 0) and (i<=`m`))
            Row *<value>* of A is exactly zero.

        (`errno` i > `m`)
            Column *<value>* of A is exactly zero.
    """

    fun_name = 'naginterfaces.library.lapacklin.dgbequ'

    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _m = _b_utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = _b_utils._EngineIntScalarType.to_py(_m)
    _r = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_m_py,),
    )
    _c = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _return_scalars = _b_lapacklin.dgbequ(
            _m,
            _n,
            kl,
            ku,
            _ab,
            _r,
            _c,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _r_py = _r
    _c_py = _c
    _rowcnd_py = _return_scalars.rowcnd
    _colcnd_py = _return_scalars.colcnd
    _amax_py = _return_scalars.amax
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['r', 'c', 'rowcnd', 'colcnd', 'amax'],
    )(
        _r_py,
        _c_py,
        _rowcnd_py,
        _colcnd_py,
        _amax_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def dgbcon(norm, kl, ku, ab, ipiv, anorm):
    u"""
    Estimate condition number of real band matrix, matrix already
    factorized by ``dgbtrf``.

    ``dgbcon`` estimates the condition number of a real band matrix A,
    where A has been factorized by ``dgbtrf``.

    For full information please refer to the NAG Library document for
    f07bg

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07bgf.html

    Parameters
    ----------
    norm : str, length 1
        Indicates whether kappa_1(A) or kappa_infinity (A) is estimated.

        `norm` = '1' or 'O'

            kappa_1(A) is estimated.

        `norm` = 'I'

            kappa_infinity (A) is estimated.

    kl : int
        k_l, the number of subdiagonals within the band of the matrix A.

    ku : int
        k_u, the number of superdiagonals within the band of the matrix
        A.

    ab : float, array-like, shape (2*`kl`+`ku`+1, n)
        The LU factorization of A, as returned by ``dgbtrf``.

    ipiv : int, array-like, shape (n)
        The pivot indices, as returned by ``dgbtrf``.

    anorm : float
        If `norm` = '1' or 'O', the 1-norm of the **original** matrix A.

        If `norm` = 'I', the infinity-norm of the **original** matrix A.

        `anorm` may be computed by calling ``blas.dlangb`` with the same
        value for the argument `norm`.

        `anorm` must be computed either **before** calling ``dgbtrf`` or
        else from a **copy** of the original matrix A.

    Returns
    -------
    rcond : float
        An estimate of the reciprocal of the condition number of A.
        `rcond` is set to zero if exact singularity is detected or the
        estimate underflows. If `rcond` is less than machine precision,
        A is singular to working precision.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O' or 'I'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter `kl`.

            Constraint: `kl` >= 0.

        (`errno` -4)
            On entry, error in parameter `ku`.

            Constraint: `ku` >= 0.

        (`errno` -8)
            On entry, error in parameter `anorm`.

            Constraint: `anorm` >= 0.0.

    References
    ----------
    Higham, N J, 1988, `FORTRAN codes for estimating the one-norm of a
    real or complex matrix, with applications to condition estimation`,
    ACM Trans. Math. Software (14), 381--396
    """

    fun_name = 'naginterfaces.library.lapacklin.dgbcon'

    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    _return_scalar = _b_lapacklin.dgbcon(
        norm,
        _n,
        kl,
        ku,
        _ab,
        _ipiv,
        anorm,
    )
    _rcond_py = _return_scalar
    return _rcond_py

def dgbrfs(trans, kl, ku, nrhs, ab, afb, ipiv, b, x):
    u"""
    Refined solution with error bounds of real band system of linear
    equations, multiple right-hand sides.

    ``dgbrfs`` returns error bounds for the solution of a real band
    system of linear equations with multiple right-hand sides, AX = B or
    A^TX = B.
    It improves the solution by iterative refinement, in order to reduce
    the backward error as much as possible.

    For full information please refer to the NAG Library document for
    f07bh

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07bhf.html

    Parameters
    ----------
    trans : str, length 1
        Indicates the form of the linear equations for which X is the
        computed solution.

        `trans` = 'N'

            The linear equations are of the form AX = B.

        `trans` = 'T' or 'C'

            The linear equations are of the form A^TX = B.

    kl : int
        k_l, the number of subdiagonals within the band of the matrix A.

    ku : int
        k_u, the number of superdiagonals within the band of the matrix
        A.

    nrhs : int
        r, the number of right-hand sides.

    ab : float, array-like, shape (`kl`+`ku`+1, n)
        The original n by n band matrix A as supplied to ``dgbtrf`` but
        with reduced requirements since the matrix is not factorized.

        See [Further Comments] for further details.

    afb : float, array-like, shape (2*`kl`+`ku`+1, n)
        The LU factorization of A, as returned by ``dgbtrf``.

    ipiv : int, array-like, shape (n)
        The pivot indices, as returned by ``dgbtrf``.

    b : float, array-like, shape (n, `nrhs`)
        The n by r right-hand side matrix B.

    x : float, array-like, shape (n, `nrhs`)
        The n by r solution matrix X, as returned by ``dgbtrs``.

    Returns
    -------
    x : float, ndarray, shape (n, `nrhs`)
        The improved solution matrix X.

    ferr : float, ndarray, shape (`nrhs`)
        `ferr`[j-1] contains an estimated error bound for the jth
        solution vector, that is, the jth column of X, for j =
        1,2,...,r.

    berr : float, ndarray, shape (`nrhs`)
        `berr`[j-1] contains the component-wise backward error bound
        beta for the jth solution vector, that is, the jth column of X,
        for j = 1,2,...,r.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter `kl`.

            Constraint: `kl` >= 0.

        (`errno` -4)
            On entry, error in parameter `ku`.

            Constraint: `ku` >= 0.

        (`errno` -5)
            On entry, error in parameter `nrhs`.

            Constraint: `nrhs` >= 0.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.dgbrfs'

    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _afb = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'afb'},
        afb,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nrhs = _b_utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineIntScalarType.to_py(_nrhs)
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _b_lapacklin.dgbrfs(
        trans,
        _n,
        kl,
        ku,
        _nrhs,
        _ab,
        _afb,
        _ipiv,
        _b,
        _x,
        _ferr,
        _berr,
    )
    _x_py = _x
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'ferr', 'berr'],
    )(
        _x_py,
        _ferr_py,
        _berr_py,
    )
    return _return_data

def zgbsv(kl, ku, ab, b):
    u"""
    Computes the solution to a complex banded system of linear
    equations.

    ``zgbsv`` computes the solution to a complex system of linear
    equations

        AX = B,

    where A is an n by n band matrix, with k_l subdiagonals and k_u
    superdiagonals, and X and B are n by r matrices.

    For full information please refer to the NAG Library document for
    f07bn

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07bnf.html

    Parameters
    ----------
    kl : int
        k_l, the number of subdiagonals within the band of the matrix A.

    ku : int
        k_u, the number of superdiagonals within the band of the matrix
        A.

    ab : complex, array-like, shape (2*`kl`+`ku`+1, n)
        The n by n coefficient matrix A.

        See [Further Comments] for further details.

    b : complex, array-like, shape (n, nrhs)
        The n by r right-hand side matrix B.

    Returns
    -------
    ab : complex, ndarray, shape (2*`kl`+`ku`+1, n)
        If `errno` >= 0, `ab` is overwritten by details of the
        factorization.

    ipiv : int, ndarray, shape (n)
        If no constraints are violated, the pivot indices that define
        the permutation matrix P; at the ith step row i of the matrix
        was interchanged with row `ipiv`[i-1]. `ipiv`[i-1] = i indicates
        a row interchange was not required.

    b : complex, ndarray, shape (n, nrhs)
        If no exception or warning is raised, the n by r solution matrix
        X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -2)
            On entry, error in parameter `kl`.

            Constraint: `kl` >= 0.

        (`errno` -3)
            On entry, error in parameter `ku`.

            Constraint: `ku` >= 0.

        (`errno` -4)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. The
            factorization has been completed, but the factor U is
            exactly singular, so the solution could not be computed.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.zgbsv'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=True,
    )
    _ipiv = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _b_lapacklin.zgbsv(
            _n,
            kl,
            ku,
            _nrhs,
            _ab,
            _ipiv,
            _b,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _ab_py = _ab
    _b_py = _b
    _ipiv_py = _ipiv
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['ab', 'ipiv', 'b'],
    )(
        _ab_py,
        _ipiv_py,
        _b_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def zgbsvx(fact, trans, kl, ku, nrhs, ab, afb, ipiv, equed, r, c, b):
    u"""
    Uses the LU factorization to compute the solution, error-bound and
    condition estimate for a complex banded system of linear equations.

    ``zgbsvx`` uses the LU factorization to compute the solution to a
    complex system of linear equations

        AX = B, A^TX = B or A^HX = B,

    where A is an n by n band matrix with k_l subdiagonals and k_u
    superdiagonals, and X and B are n by r matrices.
    Error bounds on the solution and a condition estimate are also
    provided.

    For full information please refer to the NAG Library document for
    f07bp

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07bpf.html

    Parameters
    ----------
    fact : str, length 1
        Specifies whether or not the factorized form of the matrix A is
        supplied on entry, and if not, whether the matrix A should be
        equilibrated before it is factorized.

        `fact` = 'F'

            `afb` and `ipiv` contain the factorized form of A. If
            `equed` != 'N', the matrix A has been equilibrated with
            scaling factors given by `r` and `c`. `ab`, `afb` and `ipiv`
            are not modified.

        `fact` = 'N'

            The matrix A will be copied to `afb` and factorized.

        `fact` = 'E'

            The matrix A will be equilibrated if necessary, then copied
            to `afb` and factorized.

    trans : str, length 1
        Specifies the form of the system of equations.

        `trans` = 'N'

            AX = B (No transpose).

        `trans` = 'T'

            A^TX = B (Transpose).

        `trans` = 'C'

            A^HX = B (Conjugate transpose).

    kl : int
        k_l, the number of subdiagonals within the band of the matrix A.

    ku : int
        k_u, the number of superdiagonals within the band of the matrix
        A.

    nrhs : int
        r, the number of right-hand sides, i.e., the number of columns
        of the matrix B.

    ab : complex, array-like, shape (`kl`+`ku`+1, n)
        The n by n coefficient matrix A.

        See [Further Comments] for further details.

        If `fact` = 'F' and `equed` != 'N', A must have been
        equilibrated by the scaling factors in `r` and/or `c`.

    afb : complex, array-like, shape (2*`kl`+`ku`+1, n)
        If `fact` = 'N' or 'E', `afb` need not be set.

        If `fact` = 'F', details of the LU factorization of the n by n
        band matrix A, as computed by ``zgbtrf``.

        If `equed` != 'N', `afb` is the factorized form of the
        equilibrated matrix A.

    ipiv : int, array-like, shape (n)
        If `fact` = 'N' or 'E', `ipiv` need not be set.

        If `fact` = 'F', `ipiv` contains the pivot indices from the
        factorization A = LU, as computed by ``dgbtrf``; row i of the
        matrix was interchanged with row `ipiv`[i-1].

    equed : str, length 1
        If `fact` = 'N' or 'E', `equed` need not be set.

        If `fact` = 'F', `equed` must specify the form of the
        equilibration that was performed as follows:

            if `equed` = 'N', no equilibration;

            if `equed` = 'R', row equilibration, i.e., A has been
            premultiplied by D_R;

            if `equed` = 'C', column equilibration, i.e., A has been
            postmultiplied by D_C;

            if `equed` = 'B', both row and column equilibration, i.e., A
            has been replaced by D_R AD_C.

    r : float, array-like, shape (n)
        If `fact` = 'N' or 'E', `r` need not be set.

        If `fact` = 'F' and `equed` = 'R' or 'B', `r` must contain the
        row scale factors for A, D_R; each element of `r` must be
        positive.

    c : float, array-like, shape (n)
        If `fact` = 'N' or 'E', `c` need not be set.

        If `fact` = 'F' and `equed` = 'C' or 'B', `c` must contain the
        column scale factors for A, D_C; each element of `c` must be
        positive.

    b : complex, array-like, shape (n, `nrhs`)
        The n by r right-hand side matrix B.

    Returns
    -------
    ab : complex, ndarray, shape (`kl`+`ku`+1, n)
        If `fact` = 'F' or 'N', or if `fact` = 'E' and `equed` = 'N',
        `ab` is not modified.

        If `equed` != 'N' then, if no constraints are violated, A is
        scaled as follows:

            if `equed` = 'R', A = D_r A;

            if `equed` = 'C', A = AD_c;

            if `equed` = 'B', A = D_r AD_c.

    afb : complex, ndarray, shape (2*`kl`+`ku`+1, n)
        If `fact` = 'F', `afb` is unchanged from entry.

        Otherwise, if no constraints are violated, then if `fact` = 'N',
        `afb` returns details of the LU factorization of the band matrix
        A, and if `fact` = 'E', `afb` returns details of the LU
        factorization of the equilibrated band matrix A (see the
        description of `ab` for the form of the equilibrated matrix).

    ipiv : int, ndarray, shape (n)
        If `fact` = 'F', `ipiv` is unchanged from entry.

        Otherwise, if no constraints are violated, `ipiv` contains the
        pivot indices that define the permutation matrix P; at the ith
        step row i of the matrix was interchanged with row `ipiv`[i-1].
        `ipiv`[i-1] = i indicates a row interchange was not required.

        If `fact` = 'N', the pivot indices are those corresponding to
        the factorization A = LU of the original matrix A.

        If `fact` = 'E', the pivot indices are those corresponding to
        the factorization of A = LU of the equilibrated matrix A.

    equed : str, length 1
        If `fact` = 'F', `equed` is unchanged from entry.

        Otherwise, if no constraints are violated, `equed` specifies the
        form of equilibration that was performed as specified above.

    r : float, ndarray, shape (n)
        If `fact` = 'F', `r` is unchanged from entry.

        Otherwise, if no constraints are violated and `equed` = 'R' or
        'B', `r` contains the row scale factors for A, D_R, such that A
        is multiplied on the left by D_R; each element of `r` is
        positive.

    c : float, ndarray, shape (n)
        If `fact` = 'F', `c` is unchanged from entry.

        Otherwise, if no constraints are violated and `equed` = 'C' or
        'B', `c` contains the row scale factors for A, D_C; each element
        of `c` is positive.

    b : complex, ndarray, shape (n, `nrhs`)
        If `equed` = 'N', `b` is not modified.

        If `trans` = 'N' and `equed` = 'R' or 'B', `b` is overwritten by
        D_R B.

        If `trans` = 'T' or 'C' and `equed` = 'C' or 'B', `b` is
        overwritten by D_C B.

    x : complex, ndarray, shape (n, `nrhs`)
        If the function exits successfully or `errno` = `n` + 1, the n
        by r solution matrix X to the original system of equations. Note
        that the arrays A and B are modified on exit if `equed` != 'N',
        and the solution to the equilibrated system is D_C^-1X if
        `trans` = 'N' and `equed` = 'C' or 'B', or D_R^-1X if `trans` =
        'T' or 'C' and `equed` = 'R' or 'B'.

    rcond : float
        If no constraints are violated, an estimate of the reciprocal
        condition number of the matrix A (after equilibration if that is
        performed), computed as `rcond` = 1.0/(||A||_1||A^-1||_1).

    ferr : float, ndarray, shape (`nrhs`)
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the forward error bound for each computed solution
        vector, such that ||x^_j-x_j||_infinity/||x_j||_infinity <=
        `ferr`[j-1] where x^_j is the jth column of the computed
        solution returned in the array `x` and x_j is the corresponding
        column of the exact solution X. The estimate is as reliable as
        the estimate for `rcond`, and is almost always a slight
        overestimate of the true error.

    berr : float, ndarray, shape (`nrhs`)
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the component-wise relative backward error of each
        computed solution vector x^_j (i.e., the smallest relative
        change in any element of A or B that makes x^_j an exact
        solution).

    recip_growth_factor : float
        If no exception or warning is raised, the reciprocal pivot
        growth factor ||A||/||U||, where ||.|| denotes the maximum
        absolute element norm. If `recip_growth_factor`≪1, the stability
        of the LU factorization of (equilibrated) A could be poor. This
        also means that the solution `x`, condition estimate `rcond`,
        and forward error bound `ferr` could be unreliable. If the
        factorization fails with `errno` in 1 ... `n`, then
        `recip_growth_factor` contains the reciprocal pivot growth
        factor for the leading errno columns of A.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `fact`.

            Constraint: `fact` = 'F', 'N' or 'E'.

        (`errno` -2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` -3)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -4)
            On entry, error in parameter `kl`.

            Constraint: `kl` >= 0.

        (`errno` -5)
            On entry, error in parameter `ku`.

            Constraint: `ku` >= 0.

        (`errno` -6)
            On entry, error in parameter `nrhs`.

            Constraint: `nrhs` >= 0.

        (`errno` -12)
            On entry, error in parameter `equed`.

            Constraint: `equed` = 'N', 'R', 'C' or 'B'.

        (`errno` -13)
            On entry, error in parameter `r`.

        (`errno` -14)
            On entry, error in parameter `c`.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` (i > 0) and (i<=n))
            Element *<value>* of the diagonal is exactly zero. The
            factorization has been completed, but the factor U is
            exactly singular, so the solution and error bounds could not
            be computed. `rcond` = 0.0 is returned.

        (`errno` n+1)
            U is nonsingular, but `rcond` is less than machine
            precision, meaning that the matrix is singular to working
            precision. Nevertheless, the solution and error bounds are
            computed because there are a number of situations where the
            computed solution can be more accurate than the value of
            `rcond` would suggest.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    fun_name = 'naginterfaces.library.lapacklin.zgbsvx'

    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=True,
    )
    _afb = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'afb'},
        afb,
        copy=True,
    )
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_ab, 'ab'), (_afb, 'afb'), (_b, 'b')),
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=True,
    )
    _r = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'r'},
        r,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _c = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nrhs = _b_utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineIntScalarType.to_py(_nrhs)
    _x = _b_utils._EngineComplex128ArrayType.empty_ndarray(
        (_n_py, _nrhs_py),
        sorder=_m_sorder,
    )
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    try:
        _return_scalars = _b_lapacklin.zgbsvx(
            fact,
            trans,
            _n,
            kl,
            ku,
            _nrhs,
            _ab,
            _afb,
            _ipiv,
            equed,
            _r,
            _c,
            _b,
            _x,
            _ferr,
            _berr,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _ab_py = _ab
    _afb_py = _afb
    _ipiv_py = _ipiv
    _equed_py = _return_scalars.equed
    _r_py = _r
    _c_py = _c
    _b_py = _b
    _x_py = _x
    _rcond_py = _return_scalars.rcond
    _ferr_py = _ferr
    _berr_py = _berr
    _recip_growth_factor_py = _return_scalars.recip_growth_factor
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['ab', 'afb', 'ipiv', 'equed', 'r', 'c', 'b', 'x', 'rcond', 'ferr', 'berr', 'recip_growth_factor'],
    )(
        _ab_py,
        _afb_py,
        _ipiv_py,
        _equed_py,
        _r_py,
        _c_py,
        _b_py,
        _x_py,
        _rcond_py,
        _ferr_py,
        _berr_py,
        _recip_growth_factor_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def zgbtrf(m, kl, ku, ab):
    u"""
    LU factorization of complex m by n band matrix.

    ``zgbtrf`` computes the LU factorization of a complex m by n band
    matrix.

    For full information please refer to the NAG Library document for
    f07br

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07brf.html

    Parameters
    ----------
    m : int
        m, the number of rows of the matrix A.

    kl : int
        k_l, the number of subdiagonals within the band of the matrix A.

    ku : int
        k_u, the number of superdiagonals within the band of the matrix
        A.

    ab : complex, array-like, shape (2*`kl`+`ku`+1, n)
        The m by n matrix A.

        See [Further Comments] for further details.

    Returns
    -------
    ab : complex, ndarray, shape (2*`kl`+`ku`+1, n)
        If `errno` >= 0, `ab` is overwritten by details of the
        factorization.

    ipiv : int, ndarray, shape (min(`m`, n))
        The pivot indices that define the permutation matrix. At the ith
        step, if `ipiv`[i-1] > i then row i of the matrix A was
        interchanged with row `ipiv`[i-1], for i = 1,2,...,min(m, n).
        `ipiv`[i-1] <= i indicates that, at the ith step, a row
        interchange was not required.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `m`.

            Constraint: `m` >= 0.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter `kl`.

            Constraint: `kl` >= 0.

        (`errno` -4)
            On entry, error in parameter `ku`.

            Constraint: `ku` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. The
            factorization has been completed, but the factor U is
            exactly singular, and division by zero will occur if it is
            used to solve a system of equations.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.zgbtrf'

    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _m = _b_utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = _b_utils._EngineIntScalarType.to_py(_m)
    _ipiv = _b_utils._EngineIntArrayType.empty_ndarray(
        (min(_m_py, _n_py),),
    )
    try:
        _b_lapacklin.zgbtrf(
            _m,
            _n,
            kl,
            ku,
            _ab,
            _ipiv,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _ab_py = _ab
    _ipiv_py = _ipiv
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['ab', 'ipiv'],
    )(
        _ab_py,
        _ipiv_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def zgbtrs(trans, kl, ku, ab, ipiv, b):
    u"""
    Solution of complex band system of linear equations, multiple
    right-hand sides, matrix already factorized by ``zgbtrf``.

    ``zgbtrs`` solves a complex band system of linear equations with
    multiple right-hand sides,

        AX = B, A^TX = B or A^HX = B,

    where A has been factorized by ``zgbtrf``.

    For full information please refer to the NAG Library document for
    f07bs

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07bsf.html

    Parameters
    ----------
    trans : str, length 1
        Indicates the form of the equations.

        `trans` = 'N'

            AX = B is solved for X.

        `trans` = 'T'

            A^TX = B is solved for X.

        `trans` = 'C'

            A^HX = B is solved for X.

    kl : int
        k_l, the number of subdiagonals within the band of the matrix A.

    ku : int
        k_u, the number of superdiagonals within the band of the matrix
        A.

    ab : complex, array-like, shape (2*`kl`+`ku`+1, n)
        The LU factorization of A, as returned by ``zgbtrf``.

    ipiv : int, array-like, shape (n)
        The pivot indices, as returned by ``zgbtrf``.

    b : complex, array-like, shape (n, nrhs)
        The n by r right-hand side matrix B.

    Returns
    -------
    b : complex, ndarray, shape (n, nrhs)
        The n by r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter `kl`.

            Constraint: `kl` >= 0.

        (`errno` -4)
            On entry, error in parameter `ku`.

            Constraint: `ku` >= 0.

        (`errno` -5)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.zgbtrs'

    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
    )
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    _b_lapacklin.zgbtrs(
        trans,
        _n,
        kl,
        ku,
        _nrhs,
        _ab,
        _ipiv,
        _b,
    )
    _b_py = _b
    return _b_py

def zgbequ(m, kl, ku, ab):
    u"""
    Computes row and column scalings intended to equilibrate a complex
    banded matrix and reduce its condition number.

    ``zgbequ`` computes diagonal scaling matrices D_R and D_C intended
    to equilibrate a complex m by n band matrix A of band width
    (k_l+k_u+1), and reduce its condition number.

    For full information please refer to the NAG Library document for
    f07bt

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07btf.html

    Parameters
    ----------
    m : int
        m, the number of rows of the matrix A.

    kl : int
        k_l, the number of subdiagonals of the matrix A.

    ku : int
        k_u, the number of superdiagonals of the matrix A.

    ab : complex, array-like, shape (`kl`+`ku`+1, n)
        The m by n band matrix A whose scaling factors are to be
        computed.

        See [Further Comments] for further details.

    Returns
    -------
    r : float, ndarray, shape (`m`)
        If no exception or warning is raised or `errno` > `m`, `r`
        contains the row scale factors, the diagonal elements of D_R.
        The elements of `r` will be positive.

    c : float, ndarray, shape (n)
        If no exception or warning is raised, `c` contains the column
        scale factors, the diagonal elements of D_C. The elements of `c`
        will be positive.

    rowcnd : float
        If no exception or warning is raised or `errno` > `m`, `rowcnd`
        contains the ratio of the smallest value of `r`[i-1] to the
        largest value of `r`[i-1]. If `rowcnd` >= 0.1 and `amax` is
        neither too large nor too small, it is not worth scaling by D_R.

    colcnd : float
        If no exception or warning is raised, `colcnd` contains the
        ratio of the smallest value of `c`[i-1] to the largest value of
        `c`[i-1].

        If `colcnd` >= 0.1, it is not worth scaling by D_C.

    amax : float
        max|a_{ij}|. If `amax` is very close to overflow or underflow,
        the matrix A should be scaled.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `m`.

            Constraint: `m` >= 0.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter `kl`.

            Constraint: `kl` >= 0.

        (`errno` -4)
            On entry, error in parameter `ku`.

            Constraint: `ku` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` (i > 0) and (i<=`m`))
            Row *<value>* of A is exactly zero.

        (`errno` i > `m`)
            Column *<value>* of A is exactly zero.
    """

    fun_name = 'naginterfaces.library.lapacklin.zgbequ'

    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _m = _b_utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = _b_utils._EngineIntScalarType.to_py(_m)
    _r = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_m_py,),
    )
    _c = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _return_scalars = _b_lapacklin.zgbequ(
            _m,
            _n,
            kl,
            ku,
            _ab,
            _r,
            _c,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _r_py = _r
    _c_py = _c
    _rowcnd_py = _return_scalars.rowcnd
    _colcnd_py = _return_scalars.colcnd
    _amax_py = _return_scalars.amax
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['r', 'c', 'rowcnd', 'colcnd', 'amax'],
    )(
        _r_py,
        _c_py,
        _rowcnd_py,
        _colcnd_py,
        _amax_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def zgbcon(norm, kl, ku, ab, ipiv, anorm):
    u"""
    Estimate condition number of complex band matrix, matrix already
    factorized by ``zgbtrf``.

    ``zgbcon`` estimates the condition number of a complex band matrix
    A, where A has been factorized by ``zgbtrf``.

    For full information please refer to the NAG Library document for
    f07bu

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07buf.html

    Parameters
    ----------
    norm : str, length 1
        Indicates whether kappa_1(A) or kappa_infinity (A) is estimated.

        `norm` = '1' or 'O'

            kappa_1(A) is estimated.

        `norm` = 'I'

            kappa_infinity (A) is estimated.

    kl : int
        k_l, the number of subdiagonals within the band of the matrix A.

    ku : int
        k_u, the number of superdiagonals within the band of the matrix
        A.

    ab : complex, array-like, shape (2*`kl`+`ku`+1, n)
        The LU factorization of A, as returned by ``zgbtrf``.

    ipiv : int, array-like, shape (n)
        The pivot indices, as returned by ``zgbtrf``.

    anorm : float
        If `norm` = '1' or 'O', the 1-norm of the **original** matrix A.

        If `norm` = 'I', the infinity-norm of the **original** matrix A.

        `anorm` may be computed by calling ``blas.zlangb`` with the same
        value for the argument `norm`.

        `anorm` must be computed either **before** calling ``zgbtrf`` or
        else from a **copy** of the original matrix A.

    Returns
    -------
    rcond : float
        An estimate of the reciprocal of the condition number of A.
        `rcond` is set to zero if exact singularity is detected or the
        estimate underflows. If `rcond` is less than machine precision,
        A is singular to working precision.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O' or 'I'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter `kl`.

            Constraint: `kl` >= 0.

        (`errno` -4)
            On entry, error in parameter `ku`.

            Constraint: `ku` >= 0.

        (`errno` -8)
            On entry, error in parameter `anorm`.

            Constraint: `anorm` >= 0.0.

    References
    ----------
    Higham, N J, 1988, `FORTRAN codes for estimating the one-norm of a
    real or complex matrix, with applications to condition estimation`,
    ACM Trans. Math. Software (14), 381--396
    """

    fun_name = 'naginterfaces.library.lapacklin.zgbcon'

    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    _return_scalar = _b_lapacklin.zgbcon(
        norm,
        _n,
        kl,
        ku,
        _ab,
        _ipiv,
        anorm,
    )
    _rcond_py = _return_scalar
    return _rcond_py

def zgbrfs(trans, kl, ku, nrhs, ab, afb, ipiv, b, x):
    u"""
    Refined solution with error bounds of complex band system of linear
    equations, multiple right-hand sides.

    ``zgbrfs`` returns error bounds for the solution of a complex band
    system of linear equations with multiple right-hand sides, AX = B,
    A^TX = B or A^HX = B.
    It improves the solution by iterative refinement, in order to reduce
    the backward error as much as possible.

    For full information please refer to the NAG Library document for
    f07bv

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07bvf.html

    Parameters
    ----------
    trans : str, length 1
        Indicates the form of the linear equations for which X is the
        computed solution as follows:

        `trans` = 'N'

            The linear equations are of the form AX = B.

        `trans` = 'T'

            The linear equations are of the form A^TX = B.

        `trans` = 'C'

            The linear equations are of the form A^HX = B.

    kl : int
        k_l, the number of subdiagonals within the band of the matrix A.

    ku : int
        k_u, the number of superdiagonals within the band of the matrix
        A.

    nrhs : int
        r, the number of right-hand sides.

    ab : complex, array-like, shape (`kl`+`ku`+1, n)
        The original n by n band matrix A as supplied to ``zgbtrf`` but
        with reduced requirements since the matrix is not factorized.

        See [Further Comments] for further details.

    afb : complex, array-like, shape (2*`kl`+`ku`+1, n)
        The LU factorization of A, as returned by ``zgbtrf``.

    ipiv : int, array-like, shape (n)
        The pivot indices, as returned by ``zgbtrf``.

    b : complex, array-like, shape (n, `nrhs`)
        The n by r right-hand side matrix B.

    x : complex, array-like, shape (n, `nrhs`)
        The n by r solution matrix X, as returned by ``zgbtrs``.

    Returns
    -------
    x : complex, ndarray, shape (n, `nrhs`)
        The improved solution matrix X.

    ferr : float, ndarray, shape (`nrhs`)
        `ferr`[j-1] contains an estimated error bound for the jth
        solution vector, that is, the jth column of X, for j =
        1,2,...,r.

    berr : float, ndarray, shape (`nrhs`)
        `berr`[j-1] contains the component-wise backward error bound
        beta for the jth solution vector, that is, the jth column of X,
        for j = 1,2,...,r.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter `kl`.

            Constraint: `kl` >= 0.

        (`errno` -4)
            On entry, error in parameter `ku`.

            Constraint: `ku` >= 0.

        (`errno` -5)
            On entry, error in parameter `nrhs`.

            Constraint: `nrhs` >= 0.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.zgbrfs'

    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _afb = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'afb'},
        afb,
        copy=False,
    )
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
    )
    _nrhs = _b_utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineIntScalarType.to_py(_nrhs)
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _b_lapacklin.zgbrfs(
        trans,
        _n,
        kl,
        ku,
        _nrhs,
        _ab,
        _afb,
        _ipiv,
        _b,
        _x,
        _ferr,
        _berr,
    )
    _x_py = _x
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'ferr', 'berr'],
    )(
        _x_py,
        _ferr_py,
        _berr_py,
    )
    return _return_data

def dgtsv(dl, d, du, b):
    u"""
    Computes the solution to a real tridiagonal system of linear
    equations.

    ``dgtsv`` computes the solution to a real system of linear equations

        AX = B,

    where A is an n by n tridiagonal matrix and X and B are n by r
    matrices.

    For full information please refer to the NAG Library document for
    f07ca

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07caf.html

    Parameters
    ----------
    dl : float, array-like, shape (n-1)
        Must contain the (n-1) subdiagonal elements of the matrix A.

    d : float, array-like, shape (n)
        Must contain the n diagonal elements of the matrix A.

    du : float, array-like, shape (n-1)
        Must contain the (n-1) superdiagonal elements of the matrix A.

    b : float, array-like, shape (n, nrhs)
        The n by r right-hand side matrix B.

    Returns
    -------
    dl : float, ndarray, shape (n-1)
        If no constraints are violated, `dl` is overwritten by the (n-2)
        elements of the second superdiagonal of the upper triangular
        matrix U from the LU factorization of A, in
        `dl`[0],`dl`[1],...,`dl`[n-3].

    d : float, ndarray, shape (n)
        If no constraints are violated, `d` is overwritten by the n
        diagonal elements of the upper triangular matrix U from the LU
        factorization of A.

    du : float, ndarray, shape (n-1)
        If no constraints are violated, `du` is overwritten by the (n-1)
        elements of the first superdiagonal of U.

    b : float, ndarray, shape (n, nrhs)
        If no exception or warning is raised, the n by r solution matrix
        X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -2)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero, and the
            solution has not been computed. The factorization has not
            been completed unless n = *<value>*.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug
    """

    fun_name = 'naginterfaces.library.lapacklin.dgtsv'

    _d = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'd'},
        d,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_d, {'fun_name': fun_name, 'entity': 'd'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nrhs_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _dl = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'dl'},
        dl,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _du = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'du'},
        du,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    try:
        _b_lapacklin.dgtsv(
            _n,
            _nrhs,
            _dl,
            _d,
            _du,
            _b,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _dl_py = _dl
    _d_py = _d
    _du_py = _du
    _b_py = _b
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['dl', 'd', 'du', 'b'],
    )(
        _dl_py,
        _d_py,
        _du_py,
        _b_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def dgtsvx(fact, trans, n, nrhs, dl, d, du, dlf, df, duf, du2, ipiv, b):
    u"""
    Uses the LU factorization to compute the solution, error-bound and
    condition estimate for a real tridiagonal system of linear
    equations.

    ``dgtsvx`` uses the LU factorization to compute the solution to a
    real system of linear equations

        AX = B or A^TX = B,

    where A is a tridiagonal matrix of order n and X and B are n by r
    matrices.
    Error bounds on the solution and a condition estimate are also
    provided.

    For full information please refer to the NAG Library document for
    f07cb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07cbf.html

    Parameters
    ----------
    fact : str, length 1
        Specifies whether or not the factorized form of the matrix A has
        been supplied.

        `fact` = 'F'

            `dlf`, `df`, `duf`, `du2` and `ipiv` contain the factorized
            form of the matrix A. `dlf`, `df`, `duf`, `du2` and `ipiv`
            will not be modified.

        `fact` = 'N'

            The matrix A will be copied to `dlf`, `df` and `duf` and
            factorized.

    trans : str, length 1
        Specifies the form of the system of equations.

        `trans` = 'N'

            AX = B (No transpose).

        `trans` = 'T' or 'C'

            A^TX = B (Transpose).

    n : int
        n, the order of the matrix A.

    nrhs : int
        r, the number of right-hand sides, i.e., the number of columns
        of the matrix B.

    dl : float, array-like, shape (`n`-1)
        The (n-1) subdiagonal elements of A.

    d : float, array-like, shape (`n`)
        The n diagonal elements of A.

    du : float, array-like, shape (`n`-1)
        The (n-1) superdiagonal elements of A.

    dlf : float, array-like, shape (`n`-1)
        If `fact` = 'F', `dlf` contains the (n-1) multipliers that
        define the matrix L from the LU factorization of A.

    df : float, array-like, shape (`n`)
        If `fact` = 'F', `df` contains the n diagonal elements of the
        upper triangular matrix U from the LU factorization of A.

    duf : float, array-like, shape (`n`-1)
        If `fact` = 'F', `duf` contains the (n-1) elements of the first
        superdiagonal of U.

    du2 : float, array-like, shape (`n`-2)
        If `fact` = 'F', `du2` contains the (n-2) elements of the second
        superdiagonal of U.

    ipiv : int, array-like, shape (`n`)
        If `fact` = 'F', `ipiv` contains the pivot indices from the LU
        factorization of A.

    b : float, array-like, shape (`n`, `nrhs`)
        The n by r right-hand side matrix B.

    Returns
    -------
    dlf : float, ndarray, shape (`n`-1)
        If `fact` = 'N', `dlf` contains the (n-1) multipliers that
        define the matrix L from the LU factorization of A.

    df : float, ndarray, shape (`n`)
        If `fact` = 'N', `df` contains the n diagonal elements of the
        upper triangular matrix U from the LU factorization of A.

    duf : float, ndarray, shape (`n`-1)
        If `fact` = 'N', `duf` contains the (n-1) elements of the first
        superdiagonal of U.

    du2 : float, ndarray, shape (`n`-2)
        If `fact` = 'N', `du2` contains the (n-2) elements of the second
        superdiagonal of U.

    ipiv : int, ndarray, shape (`n`)
        If `fact` = 'N', `ipiv` contains the pivot indices from the LU
        factorization of A; row i of the matrix was interchanged with
        row `ipiv`[i-1]. `ipiv`[i-1] will always be either i or i+1;
        `ipiv`[i-1] = i indicates a row interchange was not required.

    x : float, ndarray, shape (`n`, `nrhs`)
        If the function exits successfully or `errno` = `n` + 1, the n
        by r solution matrix X.

    rcond : float
        The estimate of the reciprocal condition number of the matrix A.
        If `rcond` = 0.0, the matrix may be exactly singular. This
        condition is indicated by `errno` in 1 ... `n`. Otherwise, if
        `rcond` is less than the machine precision, the matrix is
        singular to working precision. This condition is indicated by
        `errno` = `n` + 1.

    ferr : float, ndarray, shape (`nrhs`)
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the forward error bound for each computed solution
        vector, such that ||x^_j-x_j||_infinity/||x_j||_infinity <=
        `ferr`[j-1] where x^_j is the jth column of the computed
        solution returned in the array `x` and x_j is the corresponding
        column of the exact solution X. The estimate is as reliable as
        the estimate for `rcond`, and is almost always a slight
        overestimate of the true error.

    berr : float, ndarray, shape (`nrhs`)
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the component-wise relative backward error of each
        computed solution vector x^_j (i.e., the smallest relative
        change in any element of A or B that makes x^_j an exact
        solution).

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `fact`.

            Constraint: `fact` = 'F' or 'N'.

        (`errno` -2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` -3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -4)
            On entry, error in parameter `nrhs`.

            Constraint: `nrhs` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` (i > 0) and (i < `n`))
            Element *<value>* of the diagonal is exactly zero. The
            factorization has not been completed, but the factor U is
            exactly singular, so the solution and error bounds could not
            be computed. `rcond` = 0.0 is returned.

        (`errno` (i > 0) and (`n`))
            Element *<value>* of the diagonal is exactly zero. The
            factorization has been completed, but the factor U is
            exactly singular, so the solution and error bounds could not
            be computed. `rcond` = 0.0 is returned.

        (`errno` `n`+1)
            U is nonsingular, but `rcond` is less than machine
            precision, meaning that the matrix is singular to working
            precision. Nevertheless, the solution and error bounds are
            computed because there are a number of situations where the
            computed solution can be more accurate than the value of
            `rcond` would suggest.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    fun_name = 'naginterfaces.library.lapacklin.dgtsvx'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_b, 'b'),),
    )
    _dl = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'dl'},
        dl,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _d = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'd'},
        d,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _du = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'du'},
        du,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _dlf = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'dlf'},
        dlf,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _df = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'df'},
        df,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _duf = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'duf'},
        duf,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _du2 = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'du2'},
        du2,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=True,
    )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _nrhs = _b_utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineIntScalarType.to_py(_nrhs)
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py, _nrhs_py),
        sorder=_m_sorder,
    )
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    try:
        _return_scalar = _b_lapacklin.dgtsvx(
            fact,
            trans,
            _n,
            _nrhs,
            _dl,
            _d,
            _du,
            _dlf,
            _df,
            _duf,
            _du2,
            _ipiv,
            _b,
            _x,
            _ferr,
            _berr,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _dlf_py = _dlf
    _df_py = _df
    _duf_py = _duf
    _du2_py = _du2
    _ipiv_py = _ipiv
    _x_py = _x
    _rcond_py = _return_scalar
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['dlf', 'df', 'duf', 'du2', 'ipiv', 'x', 'rcond', 'ferr', 'berr'],
    )(
        _dlf_py,
        _df_py,
        _duf_py,
        _du2_py,
        _ipiv_py,
        _x_py,
        _rcond_py,
        _ferr_py,
        _berr_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def dgttrf(n, dl, d, du):
    u"""
    LU factorization of real tridiagonal matrix.

    ``dgttrf`` computes the LU factorization of a real n by n
    tridiagonal matrix A.

    For full information please refer to the NAG Library document for
    f07cd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07cdf.html

    Parameters
    ----------
    n : int
        n, the order of the matrix A.

    dl : float, array-like, shape (`n`-1)
        Must contain the (n-1) subdiagonal elements of the matrix A.

    d : float, array-like, shape (`n`)
        Must contain the n diagonal elements of the matrix A.

    du : float, array-like, shape (`n`-1)
        Must contain the (n-1) superdiagonal elements of the matrix A.

    Returns
    -------
    dl : float, ndarray, shape (`n`-1)
        Is overwritten by the (n-1) multipliers that define the matrix L
        of the LU factorization of A.

    d : float, ndarray, shape (`n`)
        Is overwritten by the n diagonal elements of the upper
        triangular matrix U from the LU factorization of A.

    du : float, ndarray, shape (`n`-1)
        Is overwritten by the (n-1) elements of the first superdiagonal
        of U.

    du2 : float, ndarray, shape (`n`-2)
        Contains the (n-2) elements of the second superdiagonal of U.

    ipiv : int, ndarray, shape (`n`)
        Contains the n pivot indices that define the permutation matrix
        P. At the ith step, row i of the matrix was interchanged with
        row `ipiv`[i-1]. `ipiv`[i-1] will always be either i or (i+1),
        `ipiv`[i-1] = i indicating that a row interchange was not
        performed.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. The
            factorization has been completed, but the factor U is
            exactly singular, and division by zero will occur if it is
            used to solve a system of equations.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug
    """

    fun_name = 'naginterfaces.library.lapacklin.dgttrf'

    _dl = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'dl'},
        dl,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _d = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'd'},
        d,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _du = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'du'},
        du,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _du2 = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py-2,),
    )
    _ipiv = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _b_lapacklin.dgttrf(
            _n,
            _dl,
            _d,
            _du,
            _du2,
            _ipiv,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _dl_py = _dl
    _d_py = _d
    _du_py = _du
    _du2_py = _du2
    _ipiv_py = _ipiv
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['dl', 'd', 'du', 'du2', 'ipiv'],
    )(
        _dl_py,
        _d_py,
        _du_py,
        _du2_py,
        _ipiv_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def dgttrs(trans, dl, d, du, du2, ipiv, b):
    u"""
    Solves a real tridiagonal system of linear equations using the LU
    factorization computed by ``dgttrf``.

    ``dgttrs`` computes the solution to a real system of linear
    equations AX = B or A^TX = B, where A is an n by n tridiagonal
    matrix and X and B are n by r matrices, using the LU factorization
    returned by ``dgttrf``.

    For full information please refer to the NAG Library document for
    f07ce

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07cef.html

    Parameters
    ----------
    trans : str, length 1
        Specifies the equations to be solved as follows:

        `trans` = 'N'

            Solve AX = B for X.

        `trans` = 'T' or 'C'

            Solve A^TX = B for X.

    dl : float, array-like, shape (n-1)
        Must contain the (n-1) multipliers that define the matrix L of
        the LU factorization of A.

    d : float, array-like, shape (n)
        Must contain the n diagonal elements of the upper triangular
        matrix U from the LU factorization of A.

    du : float, array-like, shape (n-1)
        Must contain the (n-1) elements of the first superdiagonal of U.

    du2 : float, array-like, shape (n-2)
        Must contain the (n-2) elements of the second superdiagonal of
        U.

    ipiv : int, array-like, shape (n)
        Must contain the n pivot indices that define the permutation
        matrix P. At the ith step, row i of the matrix was interchanged
        with row `ipiv`[i-1], and `ipiv`[i-1] must always be either i or
        (i+1), `ipiv`[i-1] = i indicating that a row interchange was not
        performed.

    b : float, array-like, shape (n, nrhs)
        The n by r matrix of right-hand sides B.

    Returns
    -------
    b : float, ndarray, shape (n, nrhs)
        The n by r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug
    """

    fun_name = 'naginterfaces.library.lapacklin.dgttrs'

    _d = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'd'},
        d,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_d, {'fun_name': fun_name, 'entity': 'd'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nrhs_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _dl = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'dl'},
        dl,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _du = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'du'},
        du,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _du2 = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'du2'},
        du2,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    _b_lapacklin.dgttrs(
        trans,
        _n,
        _nrhs,
        _dl,
        _d,
        _du,
        _du2,
        _ipiv,
        _b,
    )
    _b_py = _b
    return _b_py

def dgtcon(norm, n, dl, d, du, du2, ipiv, anorm):
    u"""
    Estimates the reciprocal of the condition number of a real
    tridiagonal matrix using the LU factorization computed by
    ``dgttrf``.

    ``dgtcon`` estimates the reciprocal condition number of a real n by
    n tridiagonal matrix A, using the LU factorization returned by
    ``dgttrf``.

    For full information please refer to the NAG Library document for
    f07cg

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07cgf.html

    Parameters
    ----------
    norm : str, length 1
        Specifies the norm to be used to estimate kappa(A).

        `norm` = '1' or 'O'

            Estimate kappa_1(A).

        `norm` = 'I'

            Estimate kappa_infinity (A).

    n : int
        n, the order of the matrix A.

    dl : float, array-like, shape (`n`-1)
        Must contain the (n-1) multipliers that define the matrix L of
        the LU factorization of A.

    d : float, array-like, shape (`n`)
        Must contain the n diagonal elements of the upper triangular
        matrix U from the LU factorization of A.

    du : float, array-like, shape (`n`-1)
        Must contain the (n-1) elements of the first superdiagonal of U.

    du2 : float, array-like, shape (`n`-2)
        Must contain the (n-2) elements of the second superdiagonal of
        U.

    ipiv : int, array-like, shape (`n`)
        Must contain the n pivot indices that define the permutation
        matrix P. At the ith step, row i of the matrix was interchanged
        with row `ipiv`[i-1], and `ipiv`[i-1] must always be either i or
        (i+1), `ipiv`[i-1] = i indicating that a row interchange was not
        performed.

    anorm : float
        If `norm` = '1' or 'O', the 1-norm of the **original** matrix A.

        If `norm` = 'I', the infinity-norm of the **original** matrix A.

        `anorm` may be computed by calling ``blas.dlangt`` with the same
        value for the argument `norm`.

        `anorm` must be computed either **before** calling ``dgttrf`` or
        else from a **copy** of the original matrix A.

    Returns
    -------
    rcond : float
        Contains an estimate of the reciprocal condition number.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O' or 'I'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -8)
            On entry, error in parameter `anorm`.

            Constraint: `anorm` >= 0.0.

    References
    ----------
    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    fun_name = 'naginterfaces.library.lapacklin.dgtcon'

    _dl = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'dl'},
        dl,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _d = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'd'},
        d,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _du = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'du'},
        du,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _du2 = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'du2'},
        du2,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    _return_scalar = _b_lapacklin.dgtcon(
        norm,
        n,
        _dl,
        _d,
        _du,
        _du2,
        _ipiv,
        anorm,
    )
    _rcond_py = _return_scalar
    return _rcond_py

def dgtrfs(trans, n, nrhs, dl, d, du, dlf, df, duf, du2, ipiv, b, x):
    u"""
    Refined solution with error bounds of real tridiagonal system of
    linear equations, multiple right-hand sides.

    ``dgtrfs`` computes error bounds and refines the solution to a real
    system of linear equations AX = B or A^TX = B, where A is an n by n
    tridiagonal matrix and X and B are n by r matrices, using the LU
    factorization returned by ``dgttrf`` and an initial solution
    returned by ``dgttrs``.
    Iterative refinement is used to reduce the backward error as much as
    possible.

    For full information please refer to the NAG Library document for
    f07ch

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07chf.html

    Parameters
    ----------
    trans : str, length 1
        Specifies the equations to be solved as follows:

        `trans` = 'N'

            Solve AX = B for X.

        `trans` = 'T' or 'C'

            Solve A^TX = B for X.

    n : int
        n, the order of the matrix A.

    nrhs : int
        r, the number of right-hand sides, i.e., the number of columns
        of the matrix B.

    dl : float, array-like, shape (`n`-1)
        Must contain the (n-1) subdiagonal elements of the matrix A.

    d : float, array-like, shape (`n`)
        Must contain the n diagonal elements of the matrix A.

    du : float, array-like, shape (`n`-1)
        Must contain the (n-1) superdiagonal elements of the matrix A.

    dlf : float, array-like, shape (`n`-1)
        Must contain the (n-1) multipliers that define the matrix L of
        the LU factorization of A.

    df : float, array-like, shape (`n`)
        Must contain the n diagonal elements of the upper triangular
        matrix U from the LU factorization of A.

    duf : float, array-like, shape (`n`-1)
        Must contain the (n-1) elements of the first superdiagonal of U.

    du2 : float, array-like, shape (`n`-2)
        Must contain the (n-2) elements of the second superdiagonal of
        U.

    ipiv : int, array-like, shape (`n`)
        Must contain the n pivot indices that define the permutation
        matrix P. At the ith step, row i of the matrix was interchanged
        with row `ipiv`[i-1], and `ipiv`[i-1] must always be either i or
        (i+1), `ipiv`[i-1] = i indicating that a row interchange was not
        performed.

    b : float, array-like, shape (`n`, `nrhs`)
        The n by r matrix of right-hand sides B.

    x : float, array-like, shape (`n`, `nrhs`)
        The n by r initial solution matrix X.

    Returns
    -------
    x : float, ndarray, shape (`n`, `nrhs`)
        The n by r refined solution matrix X.

    ferr : float, ndarray, shape (`nrhs`)
        Estimate of the forward error bound for each computed solution
        vector, such that ||x^_j-x_j||_infinity/||x^_j||_infinity <=
        `ferr`[j-1], where x^_j is the jth column of the computed
        solution returned in the array `x` and x_j is the corresponding
        column of the exact solution X. The estimate is almost always a
        slight overestimate of the true error.

    berr : float, ndarray, shape (`nrhs`)
        Estimate of the component-wise relative backward error of each
        computed solution vector x^_j (i.e., the smallest relative
        change in any element of A or B that makes x^_j an exact
        solution).

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -3)
            On entry, error in parameter `nrhs`.

            Constraint: `nrhs` >= 0.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug
    """

    fun_name = 'naginterfaces.library.lapacklin.dgtrfs'

    _dl = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'dl'},
        dl,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _d = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'd'},
        d,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _du = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'du'},
        du,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _dlf = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'dlf'},
        dlf,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _df = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'df'},
        df,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _duf = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'duf'},
        duf,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _du2 = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'du2'},
        du2,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nrhs = _b_utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineIntScalarType.to_py(_nrhs)
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _b_lapacklin.dgtrfs(
        trans,
        n,
        _nrhs,
        _dl,
        _d,
        _du,
        _dlf,
        _df,
        _duf,
        _du2,
        _ipiv,
        _b,
        _x,
        _ferr,
        _berr,
    )
    _x_py = _x
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'ferr', 'berr'],
    )(
        _x_py,
        _ferr_py,
        _berr_py,
    )
    return _return_data

def zgtsv(dl, d, du, b):
    u"""
    Computes the solution to a complex tridiagonal system of linear
    equations.

    ``zgtsv`` computes the solution to a complex system of linear
    equations

        AX = B,

    where A is an n by n tridiagonal matrix and X and B are n by r
    matrices.

    For full information please refer to the NAG Library document for
    f07cn

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07cnf.html

    Parameters
    ----------
    dl : complex, array-like, shape (n-1)
        Must contain the (n-1) subdiagonal elements of the matrix A.

    d : complex, array-like, shape (n)
        Must contain the n diagonal elements of the matrix A.

    du : complex, array-like, shape (n-1)
        Must contain the (n-1) superdiagonal elements of the matrix A.

    b : complex, array-like, shape (n, nrhs)
        The n by r right-hand side matrix B.

    Returns
    -------
    dl : complex, ndarray, shape (n-1)
        If no constraints are violated, `dl` is overwritten by the (n-2)
        elements of the second superdiagonal of the upper triangular
        matrix U from the LU factorization of A, in
        `dl`[0],`dl`[1],...,`dl`[n-3].

    d : complex, ndarray, shape (n)
        If no constraints are violated, `d` is overwritten by the n
        diagonal elements of the upper triangular matrix U from the LU
        factorization of A.

    du : complex, ndarray, shape (n-1)
        If no constraints are violated, `du` is overwritten by the (n-1)
        elements of the first superdiagonal of U.

    b : complex, ndarray, shape (n, nrhs)
        If no exception or warning is raised, the n by r solution matrix
        X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -2)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero, and the
            solution has not been computed. The factorization has not
            been completed unless n = *<value>*.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug
    """

    fun_name = 'naginterfaces.library.lapacklin.zgtsv'

    _d = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'd'},
        d,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_d, {'fun_name': fun_name, 'entity': 'd'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
    )
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _dl = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'dl'},
        dl,
        copy=True,
    )
    _du = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'du'},
        du,
        copy=True,
    )
    try:
        _b_lapacklin.zgtsv(
            _n,
            _nrhs,
            _dl,
            _d,
            _du,
            _b,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _dl_py = _dl
    _d_py = _d
    _du_py = _du
    _b_py = _b
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['dl', 'd', 'du', 'b'],
    )(
        _dl_py,
        _d_py,
        _du_py,
        _b_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def zgtsvx(fact, trans, n, nrhs, dl, d, du, dlf, df, duf, du2, ipiv, b):
    u"""
    Uses the LU factorization to compute the solution, error-bound and
    condition estimate for a complex tridiagonal system of linear
    equations.

    ``zgtsvx`` uses the LU factorization to compute the solution to a
    complex system of linear equations

        AX = B, A^TX = B or A^HX = B,

    where A is a tridiagonal matrix of order n and X and B are n by r
    matrices.
    Error bounds on the solution and a condition estimate are also
    provided.

    For full information please refer to the NAG Library document for
    f07cp

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07cpf.html

    Parameters
    ----------
    fact : str, length 1
        Specifies whether or not the factorized form of the matrix A has
        been supplied.

        `fact` = 'F'

            `dlf`, `df`, `duf`, `du2` and `ipiv` contain the factorized
            form of the matrix A. `dlf`, `df`, `duf`, `du2` and `ipiv`
            will not be modified.

        `fact` = 'N'

            The matrix A will be copied to `dlf`, `df` and `duf` and
            factorized.

    trans : str, length 1
        Specifies the form of the system of equations.

        `trans` = 'N'

            AX = B (No transpose).

        `trans` = 'T'

            A^TX = B (Transpose).

        `trans` = 'C'

            A^HX = B (Conjugate transpose).

    n : int
        n, the order of the matrix A.

    nrhs : int
        r, the number of right-hand sides, i.e., the number of columns
        of the matrix B.

    dl : complex, array-like, shape (`n`-1)
        The (n-1) subdiagonal elements of A.

    d : complex, array-like, shape (`n`)
        The n diagonal elements of A.

    du : complex, array-like, shape (`n`-1)
        The (n-1) superdiagonal elements of A.

    dlf : complex, array-like, shape (`n`-1)
        If `fact` = 'F', `dlf` contains the (n-1) multipliers that
        define the matrix L from the LU factorization of A.

    df : complex, array-like, shape (`n`)
        If `fact` = 'F', `df` contains the n diagonal elements of the
        upper triangular matrix U from the LU factorization of A.

    duf : complex, array-like, shape (`n`-1)
        If `fact` = 'F', `duf` contains the (n-1) elements of the first
        superdiagonal of U.

    du2 : complex, array-like, shape (`n`-2)
        If `fact` = 'F', `du2` contains the (n-2) elements of the second
        superdiagonal of U.

    ipiv : int, array-like, shape (`n`)
        If `fact` = 'F', `ipiv` contains the pivot indices from the LU
        factorization of A.

    b : complex, array-like, shape (`n`, `nrhs`)
        The n by r right-hand side matrix B.

    Returns
    -------
    dlf : complex, ndarray, shape (`n`-1)
        If `fact` = 'N', `dlf` contains the (n-1) multipliers that
        define the matrix L from the LU factorization of A.

    df : complex, ndarray, shape (`n`)
        If `fact` = 'N', `df` contains the n diagonal elements of the
        upper triangular matrix U from the LU factorization of A.

    duf : complex, ndarray, shape (`n`-1)
        If `fact` = 'N', `duf` contains the (n-1) elements of the first
        superdiagonal of U.

    du2 : complex, ndarray, shape (`n`-2)
        If `fact` = 'N', `du2` contains the (n-2) elements of the second
        superdiagonal of U.

    ipiv : int, ndarray, shape (`n`)
        If `fact` = 'N', `ipiv` contains the pivot indices from the LU
        factorization of A; row i of the matrix was interchanged with
        row `ipiv`[i-1]. `ipiv`[i-1] will always be either i or i+1;
        `ipiv`[i-1] = i indicates a row interchange was not required.

    x : complex, ndarray, shape (`n`, `nrhs`)
        If the function exits successfully or `errno` = `n` + 1, the n
        by r solution matrix X.

    rcond : float
        The estimate of the reciprocal condition number of the matrix A.
        If `rcond` = 0.0, the matrix may be exactly singular. This
        condition is indicated by `errno` in 1 ... `n`. Otherwise, if
        `rcond` is less than the machine precision, the matrix is
        singular to working precision. This condition is indicated by
        `errno` = `n` + 1.

    ferr : float, ndarray, shape (`nrhs`)
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the forward error bound for each computed solution
        vector, such that ||x^_j-x_j||_infinity/||x_j||_infinity <=
        `ferr`[j-1] where x^_j is the jth column of the computed
        solution returned in the array `x` and x_j is the corresponding
        column of the exact solution X. The estimate is as reliable as
        the estimate for `rcond`, and is almost always a slight
        overestimate of the true error.

    berr : float, ndarray, shape (`nrhs`)
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the component-wise relative backward error of each
        computed solution vector x^_j (i.e., the smallest relative
        change in any element of A or B that makes x^_j an exact
        solution).

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `fact`.

            Constraint: `fact` = 'F' or 'N'.

        (`errno` -2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` -3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -4)
            On entry, error in parameter `nrhs`.

            Constraint: `nrhs` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` (i > 0) and (i < `n`))
            Element *<value>* of the diagonal is exactly zero. The
            factorization has not been completed, but the factor U is
            exactly singular, so the solution and error bounds could not
            be computed. `rcond` = 0.0 is returned.

        (`errno` (i > 0) and (`n`))
            Element *<value>* of the diagonal is exactly zero. The
            factorization has been completed, but the factor U is
            exactly singular, so the solution and error bounds could not
            be computed. `rcond` = 0.0 is returned.

        (`errno` `n`+1)
            U is nonsingular, but `rcond` is less than machine
            precision, meaning that the matrix is singular to working
            precision. Nevertheless, the solution and error bounds are
            computed because there are a number of situations where the
            computed solution can be more accurate than the value of
            `rcond` would suggest.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    fun_name = 'naginterfaces.library.lapacklin.zgtsvx'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_b, 'b'),),
    )
    _dl = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'dl'},
        dl,
        copy=False,
    )
    _d = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'd'},
        d,
        copy=False,
    )
    _du = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'du'},
        du,
        copy=False,
    )
    _dlf = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'dlf'},
        dlf,
        copy=True,
    )
    _df = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'df'},
        df,
        copy=True,
    )
    _duf = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'duf'},
        duf,
        copy=True,
    )
    _du2 = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'du2'},
        du2,
        copy=True,
    )
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=True,
    )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _nrhs = _b_utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineIntScalarType.to_py(_nrhs)
    _x = _b_utils._EngineComplex128ArrayType.empty_ndarray(
        (_n_py, _nrhs_py),
        sorder=_m_sorder,
    )
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    try:
        _return_scalar = _b_lapacklin.zgtsvx(
            fact,
            trans,
            _n,
            _nrhs,
            _dl,
            _d,
            _du,
            _dlf,
            _df,
            _duf,
            _du2,
            _ipiv,
            _b,
            _x,
            _ferr,
            _berr,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _dlf_py = _dlf
    _df_py = _df
    _duf_py = _duf
    _du2_py = _du2
    _ipiv_py = _ipiv
    _x_py = _x
    _rcond_py = _return_scalar
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['dlf', 'df', 'duf', 'du2', 'ipiv', 'x', 'rcond', 'ferr', 'berr'],
    )(
        _dlf_py,
        _df_py,
        _duf_py,
        _du2_py,
        _ipiv_py,
        _x_py,
        _rcond_py,
        _ferr_py,
        _berr_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def zgttrf(n, dl, d, du):
    u"""
    LU factorization of complex tridiagonal matrix.

    ``zgttrf`` computes the LU factorization of a complex n by n
    tridiagonal matrix A.

    For full information please refer to the NAG Library document for
    f07cr

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07crf.html

    Parameters
    ----------
    n : int
        n, the order of the matrix A.

    dl : complex, array-like, shape (`n`-1)
        Must contain the (n-1) subdiagonal elements of the matrix A.

    d : complex, array-like, shape (`n`)
        Must contain the n diagonal elements of the matrix A.

    du : complex, array-like, shape (`n`-1)
        Must contain the (n-1) superdiagonal elements of the matrix A.

    Returns
    -------
    dl : complex, ndarray, shape (`n`-1)
        Is overwritten by the (n-1) multipliers that define the matrix L
        of the LU factorization of A.

    d : complex, ndarray, shape (`n`)
        Is overwritten by the n diagonal elements of the upper
        triangular matrix U from the LU factorization of A.

    du : complex, ndarray, shape (`n`-1)
        Is overwritten by the (n-1) elements of the first superdiagonal
        of U.

    du2 : complex, ndarray, shape (`n`-2)
        Contains the (n-2) elements of the second superdiagonal of U.

    ipiv : int, ndarray, shape (`n`)
        Contains the n pivot indices that define the permutation matrix
        P. At the ith step, row i of the matrix was interchanged with
        row `ipiv`[i-1]. `ipiv`[i-1] will always be either i or (i+1),
        `ipiv`[i-1] = i indicating that a row interchange was not
        performed.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. The
            factorization has been completed, but the factor U is
            exactly singular, and division by zero will occur if it is
            used to solve a system of equations.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug
    """

    fun_name = 'naginterfaces.library.lapacklin.zgttrf'

    _dl = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'dl'},
        dl,
        copy=True,
    )
    _d = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'd'},
        d,
        copy=True,
    )
    _du = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'du'},
        du,
        copy=True,
    )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _du2 = _b_utils._EngineComplex128ArrayType.empty_ndarray(
        (_n_py-2,),
    )
    _ipiv = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _b_lapacklin.zgttrf(
            _n,
            _dl,
            _d,
            _du,
            _du2,
            _ipiv,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _dl_py = _dl
    _d_py = _d
    _du_py = _du
    _du2_py = _du2
    _ipiv_py = _ipiv
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['dl', 'd', 'du', 'du2', 'ipiv'],
    )(
        _dl_py,
        _d_py,
        _du_py,
        _du2_py,
        _ipiv_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def zgttrs(trans, dl, d, du, du2, ipiv, b):
    u"""
    Solves a complex tridiagonal system of linear equations using the LU
    factorization computed by ``dgttrf``.

    ``zgttrs`` computes the solution to a complex system of linear
    equations AX = B or A^TX = B or A^HX = B, where A is an n by n
    tridiagonal matrix and X and B are n by r matrices, using the LU
    factorization returned by ``zgttrf``.

    For full information please refer to the NAG Library document for
    f07cs

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07csf.html

    Parameters
    ----------
    trans : str, length 1
        Specifies the equations to be solved as follows:

        `trans` = 'N'

            Solve AX = B for X.

        `trans` = 'T'

            Solve A^TX = B for X.

        `trans` = 'C'

            Solve A^HX = B for X.

    dl : complex, array-like, shape (n-1)
        Must contain the (n-1) multipliers that define the matrix L of
        the LU factorization of A.

    d : complex, array-like, shape (n)
        Must contain the n diagonal elements of the upper triangular
        matrix U from the LU factorization of A.

    du : complex, array-like, shape (n-1)
        Must contain the (n-1) elements of the first superdiagonal of U.

    du2 : complex, array-like, shape (n-2)
        Must contain the (n-2) elements of the second superdiagonal of
        U.

    ipiv : int, array-like, shape (n)
        Must contain the n pivot indices that define the permutation
        matrix P. At the ith step, row i of the matrix was interchanged
        with row `ipiv`[i-1], and `ipiv`[i-1] must always be either i or
        (i+1), `ipiv`[i-1] = i indicating that a row interchange was not
        performed.

    b : complex, array-like, shape (n, nrhs)
        The n by r matrix of right-hand sides B.

    Returns
    -------
    b : complex, ndarray, shape (n, nrhs)
        The n by r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.zgttrs'

    _d = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'd'},
        d,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_d, {'fun_name': fun_name, 'entity': 'd'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
    )
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _dl = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'dl'},
        dl,
        copy=False,
    )
    _du = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'du'},
        du,
        copy=False,
    )
    _du2 = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'du2'},
        du2,
        copy=False,
    )
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    _b_lapacklin.zgttrs(
        trans,
        _n,
        _nrhs,
        _dl,
        _d,
        _du,
        _du2,
        _ipiv,
        _b,
    )
    _b_py = _b
    return _b_py

def zgtcon(norm, dl, d, du, du2, ipiv, anorm):
    u"""
    Estimates the reciprocal of the condition number of a complex
    tridiagonal matrix using the LU factorization computed by
    ``dgttrf``.

    ``zgtcon`` estimates the reciprocal condition number of a complex n
    by n tridiagonal matrix A, using the LU factorization returned by
    ``zgttrf``.

    For full information please refer to the NAG Library document for
    f07cu

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07cuf.html

    Parameters
    ----------
    norm : str, length 1
        Specifies the norm to be used to estimate kappa(A).

        `norm` = '1' or 'O'

            Estimate kappa_1(A).

        `norm` = 'I'

            Estimate kappa_infinity (A).

    dl : complex, array-like, shape (n-1)
        Must contain the (n-1) multipliers that define the matrix L of
        the LU factorization of A.

    d : complex, array-like, shape (n)
        Must contain the n diagonal elements of the upper triangular
        matrix U from the LU factorization of A.

    du : complex, array-like, shape (n-1)
        Must contain the (n-1) elements of the first superdiagonal of U.

    du2 : complex, array-like, shape (n-2)
        Must contain the (n-2) elements of the second superdiagonal of
        U.

    ipiv : int, array-like, shape (n)
        Must contain the n pivot indices that define the permutation
        matrix P. At the ith step, row i of the matrix was interchanged
        with row `ipiv`[i-1], and `ipiv`[i-1] must always be either i or
        (i+1), `ipiv`[i-1] = i indicating that a row interchange was not
        performed.

    anorm : float
        If `norm` = '1' or 'O', the 1-norm of the **original** matrix A.

        If `norm` = 'I', the infinity-norm of the **original** matrix A.

        `anorm` may be computed by calling ``blas.zlangt`` with the same
        value for the argument `norm`.

        `anorm` must be computed either **before** calling ``zgttrf`` or
        else from a **copy** of the original matrix A.

    Returns
    -------
    rcond : float
        Contains an estimate of the reciprocal condition number.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O' or 'I'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -8)
            On entry, error in parameter `anorm`.

            Constraint: `anorm` >= 0.0.

    References
    ----------
    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    fun_name = 'naginterfaces.library.lapacklin.zgtcon'

    _d = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'd'},
        d,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_d, {'fun_name': fun_name, 'entity': 'd'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _dl = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'dl'},
        dl,
        copy=False,
    )
    _du = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'du'},
        du,
        copy=False,
    )
    _du2 = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'du2'},
        du2,
        copy=False,
    )
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    _return_scalar = _b_lapacklin.zgtcon(
        norm,
        _n,
        _dl,
        _d,
        _du,
        _du2,
        _ipiv,
        anorm,
    )
    _rcond_py = _return_scalar
    return _rcond_py

def zgtrfs(trans, n, nrhs, dl, d, du, dlf, df, duf, du2, ipiv, b, x):
    u"""
    Refined solution with error bounds of complex tridiagonal system of
    linear equations, multiple right-hand sides.

    ``zgtrfs`` computes error bounds and refines the solution to a
    complex system of linear equations AX = B or A^TX = B or A^HX = B,
    where A is an n by n tridiagonal matrix and X and B are n by r
    matrices, using the LU factorization returned by ``zgttrf`` and an
    initial solution returned by ``zgttrs``.
    Iterative refinement is used to reduce the backward error as much as
    possible.

    For full information please refer to the NAG Library document for
    f07cv

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07cvf.html

    Parameters
    ----------
    trans : str, length 1
        Specifies the equations to be solved as follows:

        `trans` = 'N'

            Solve AX = B for X.

        `trans` = 'T'

            Solve A^TX = B for X.

        `trans` = 'C'

            Solve A^HX = B for X.

    n : int
        n, the order of the matrix A.

    nrhs : int
        r, the number of right-hand sides, i.e., the number of columns
        of the matrix B.

    dl : complex, array-like, shape (`n`-1)
        Must contain the (n-1) subdiagonal elements of the matrix A.

    d : complex, array-like, shape (`n`)
        Must contain the n diagonal elements of the matrix A.

    du : complex, array-like, shape (`n`-1)
        Must contain the (n-1) superdiagonal elements of the matrix A.

    dlf : complex, array-like, shape (`n`-1)
        Must contain the (n-1) multipliers that define the matrix L of
        the LU factorization of A.

    df : complex, array-like, shape (`n`)
        Must contain the n diagonal elements of the upper triangular
        matrix U from the LU factorization of A.

    duf : complex, array-like, shape (`n`-1)
        Must contain the (n-1) elements of the first superdiagonal of U.

    du2 : complex, array-like, shape (`n`-2)
        Must contain the (n-2) elements of the second superdiagonal of
        U.

    ipiv : int, array-like, shape (`n`)
        Must contain the n pivot indices that define the permutation
        matrix P. At the ith step, row i of the matrix was interchanged
        with row `ipiv`[i-1], and `ipiv`[i-1] must always be either i or
        (i+1), `ipiv`[i-1] = i indicating that a row interchange was not
        performed.

    b : complex, array-like, shape (`n`, `nrhs`)
        The n by r matrix of right-hand sides B.

    x : complex, array-like, shape (`n`, `nrhs`)
        The n by r initial solution matrix X.

    Returns
    -------
    x : complex, ndarray, shape (`n`, `nrhs`)
        The n by r refined solution matrix X.

    ferr : float, ndarray, shape (`nrhs`)
        Estimate of the forward error bound for each computed solution
        vector, such that ||x^_j-x_j||_infinity/||x^_j||_infinity <=
        `ferr`[j-1], where x^_j is the jth column of the computed
        solution returned in the array `x` and x_j is the corresponding
        column of the exact solution X. The estimate is almost always a
        slight overestimate of the true error.

    berr : float, ndarray, shape (`nrhs`)
        Estimate of the component-wise relative backward error of each
        computed solution vector x^_j (i.e., the smallest relative
        change in any element of A or B that makes x^_j an exact
        solution).

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -3)
            On entry, error in parameter `nrhs`.

            Constraint: `nrhs` >= 0.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug
    """

    fun_name = 'naginterfaces.library.lapacklin.zgtrfs'

    _dl = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'dl'},
        dl,
        copy=False,
    )
    _d = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'd'},
        d,
        copy=False,
    )
    _du = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'du'},
        du,
        copy=False,
    )
    _dlf = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'dlf'},
        dlf,
        copy=False,
    )
    _df = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'df'},
        df,
        copy=False,
    )
    _duf = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'duf'},
        duf,
        copy=False,
    )
    _du2 = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'du2'},
        du2,
        copy=False,
    )
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
    )
    _nrhs = _b_utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineIntScalarType.to_py(_nrhs)
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _b_lapacklin.zgtrfs(
        trans,
        n,
        _nrhs,
        _dl,
        _d,
        _du,
        _dlf,
        _df,
        _duf,
        _du2,
        _ipiv,
        _b,
        _x,
        _ferr,
        _berr,
    )
    _x_py = _x
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'ferr', 'berr'],
    )(
        _x_py,
        _ferr_py,
        _berr_py,
    )
    return _return_data

def dposv(uplo, a, b):
    u"""
    Computes the solution to a real symmetric positive definite system
    of linear equations.

    ``dposv`` computes the solution to a real system of linear equations

        AX = B,

    where A is an n by n symmetric positive definite matrix and X and B
    are n by r matrices.

    For full information please refer to the NAG Library document for
    f07fa

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07faf.html

    Parameters
    ----------
    uplo : str, length 1
        If `uplo` = 'U', the upper triangle of A is stored.

        If `uplo` = 'L', the lower triangle of A is stored.

    a : float, array-like, shape (n, n)
        The n by n symmetric matrix A.

    b : float, array-like, shape (n, nrhs)
        The n by r right-hand side matrix B.

    Returns
    -------
    a : float, ndarray, shape (n, n)
        If no exception or warning is raised, the factor U or L from the
        Cholesky factorization A = U^TU or A = LL^T.

    b : float, ndarray, shape (n, nrhs)
        If no exception or warning is raised, the n by r solution matrix
        X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

        (`errno` i > 0)
            The leading minor of order *<value>* of A is not positive
            definite, so the factorization could not be completed, and
            the solution has not been computed.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.dposv'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nrhs_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _b_lapacklin.dposv(
        uplo,
        _n,
        _nrhs,
        _a,
        _b,
    )
    _a_py = _a
    _b_py = _b
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'b'],
    )(
        _a_py,
        _b_py,
    )
    return _return_data

def dposvx(fact, uplo, n, nrhs, a, af, equed, s, b):
    u"""
    Uses the Cholesky factorization to compute the solution, error-bound
    and condition estimate for a real symmetric positive definite system
    of linear equations.

    ``dposvx`` uses the Cholesky factorization

        A = U^TU or A = LL^T

    to compute the solution to a real system of linear equations

        AX = B,

    where A is an n by n symmetric positive definite matrix and X and B
    are n by r matrices.
    Error bounds on the solution and a condition estimate are also
    provided.

    For full information please refer to the NAG Library document for
    f07fb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07fbf.html

    Parameters
    ----------
    fact : str, length 1
        Specifies whether or not the factorized form of the matrix A is
        supplied on entry, and if not, whether the matrix A should be
        equilibrated before it is factorized.

        `fact` = 'F'

            `af` contains the factorized form of A. If `equed` = 'Y',
            the matrix A has been equilibrated with scaling factors
            given by `s`. `a` and `af` will not be modified.

        `fact` = 'N'

            The matrix A will be copied to `af` and factorized.

        `fact` = 'E'

            The matrix A will be equilibrated if necessary, then copied
            to `af` and factorized.

    uplo : str, length 1
        If `uplo` = 'U', the upper triangle of A is stored.

        If `uplo` = 'L', the lower triangle of A is stored.

    n : int
        n, the number of linear equations, i.e., the order of the matrix
        A.

    nrhs : int
        r, the number of right-hand sides, i.e., the number of columns
        of the matrix B.

    a : float, array-like, shape (`n`, `n`)
        The n by n symmetric matrix A.

        If `fact` = 'F' and `equed` = 'Y', `a` must have been
        equilibrated by the scaling factor in `s` as D_S AD_S.

    af : float, array-like, shape (`n`, `n`)
        If `fact` = 'F', `af` contains the triangular factor U or L from
        the Cholesky factorization A = U^TU or A = LL^T, in the same
        storage format as `a`. If `equed` != 'N', `af` is the factorized
        form of the equilibrated matrix D_S AD_S.

    equed : str, length 1
        If `fact` = 'N' or 'E', `equed` need not be set.

        If `fact` = 'F', `equed` must specify the form of the
        equilibration that was performed as follows:

            if `equed` = 'N', no equilibration;

            if `equed` = 'Y', equilibration was performed, i.e., A has
            been replaced by D_S AD_S.

    s : float, array-like, shape (`n`)
        If `fact` = 'N' or 'E', `s` need not be set.

        If `fact` = 'F' and `equed` = 'Y', `s` must contain the scale
        factors, D_S, for A; each element of `s` must be positive.

    b : float, array-like, shape (`n`, `nrhs`)
        The n by r right-hand side matrix B.

    Returns
    -------
    a : float, ndarray, shape (`n`, `n`)
        If `fact` = 'F' or 'N', or if `fact` = 'E' and `equed` = 'N',
        `a` is not modified.

        If `fact` = 'E' and `equed` = 'Y', `a` is overwritten by D_S
        AD_S.

    af : float, ndarray, shape (`n`, `n`)
        If `fact` = 'N', `af` returns the triangular factor U or L from
        the Cholesky factorization A = U^TU or A = LL^T of the original
        matrix A.

        If `fact` = 'E', `af` returns the triangular factor U or L from
        the Cholesky factorization A = U^TU or A = LL^T of the
        equilibrated matrix A (see the description of `a` for the form
        of the equilibrated matrix).

    equed : str, length 1
        If `fact` = 'F', `equed` is unchanged from entry.

        Otherwise, if no constraints are violated, `equed` specifies the
        form of the equilibration that was performed as specified above.

    s : float, ndarray, shape (`n`)
        If `fact` = 'F', `s` is unchanged from entry.

        Otherwise, if no constraints are violated and `equed` = 'Y', `s`
        contains the scale factors, D_S, for A; each element of `s` is
        positive.

    b : float, ndarray, shape (`n`, `nrhs`)
        If `equed` = 'N', `b` is not modified.

        If `equed` = 'Y', `b` is overwritten by D_S B.

    x : float, ndarray, shape (`n`, `nrhs`)
        If the function exits successfully or `errno` = `n` + 1, the n
        by r solution matrix X to the original system of equations. Note
        that the arrays A and B are modified on exit if `equed` = 'Y',
        and the solution to the equilibrated system is D_S^-1X.

    rcond : float
        If no constraints are violated, an estimate of the reciprocal
        condition number of the matrix A (after equilibration if that is
        performed), computed as `rcond` = 1.0/(||A||_1||A^-1||_1).

    ferr : float, ndarray, shape (`nrhs`)
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the forward error bound for each computed solution
        vector, such that ||x^_j-x_j||_infinity/||x_j||_infinity <=
        `ferr`[j-1] where x^_j is the jth column of the computed
        solution returned in the array `x` and x_j is the corresponding
        column of the exact solution X. The estimate is as reliable as
        the estimate for `rcond`, and is almost always a slight
        overestimate of the true error.

    berr : float, ndarray, shape (`nrhs`)
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the component-wise relative backward error of each
        computed solution vector x^_j (i.e., the smallest relative
        change in any element of A or B that makes x^_j an exact
        solution).

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `fact`.

            Constraint: `fact` = 'F', 'N' or 'E'.

        (`errno` -2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -4)
            On entry, error in parameter `nrhs`.

            Constraint: `nrhs` >= 0.

        (`errno` -9)
            On entry, error in parameter `equed`.

            Constraint: `equed` = 'N' or 'Y'.

        (`errno` -10)
            On entry, error in parameter `s`.

        (`errno` (i > 0) and (i<=`n`))
            The leading minor of order *<value>* of A is not positive
            definite, so the factorization could not be completed, and
            the solution has not been computed. `rcond` = 0.0 is
            returned.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` `n`+1)
            U (or L) is nonsingular, but `rcond` is less than machine
            precision, meaning that the matrix is singular to working
            precision. Nevertheless, the solution and error bounds are
            computed because there are a number of situations where the
            computed solution can be more accurate than the value of
            `rcond` would suggest.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    fun_name = 'naginterfaces.library.lapacklin.dposvx'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _af = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'af'},
        af,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_a, 'a'), (_af, 'af'), (_b, 'b')),
    )
    _s = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 's'},
        s,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _nrhs = _b_utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineIntScalarType.to_py(_nrhs)
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py, _nrhs_py),
        sorder=_m_sorder,
    )
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    try:
        _return_scalars = _b_lapacklin.dposvx(
            fact,
            uplo,
            _n,
            _nrhs,
            _a,
            _af,
            equed,
            _s,
            _b,
            _x,
            _ferr,
            _berr,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _a_py = _a
    _af_py = _af
    _equed_py = _return_scalars.equed
    _s_py = _s
    _b_py = _b
    _x_py = _x
    _rcond_py = _return_scalars.rcond
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'af', 'equed', 's', 'b', 'x', 'rcond', 'ferr', 'berr'],
    )(
        _a_py,
        _af_py,
        _equed_py,
        _s_py,
        _b_py,
        _x_py,
        _rcond_py,
        _ferr_py,
        _berr_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def dsposv(uplo, n, nrhs, a, b):
    u"""
    Computes the solution to a real symmetric positive definite system
    of linear equations using mixed precision arithmetic.

    ``dsposv`` uses the Cholesky factorization

        A = U^TU or A = LL^T

    to compute the solution to a real system of linear equations

        AX = B,

    where A is an n by n symmetric positive definite matrix and X and B
    are n by r matrices.

    For full information please refer to the NAG Library document for
    f07fc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07fcf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

            The upper triangular part of A is stored.

        `uplo` = 'L'

            The lower triangular part of A is stored.

    n : int
        n, the number of linear equations, i.e., the order of the matrix
        A.

    nrhs : int
        r, the number of right-hand sides, i.e., the number of columns
        of the matrix B.

    a : float, array-like, shape (`n`, `n`)
        The n by n symmetric positive definite matrix A.

    b : float, array-like, shape (`n`, `nrhs`)
        The right-hand side matrix B.

    Returns
    -------
    a : float, ndarray, shape (`n`, `n`)
        If iterative refinement has been successfully used (no exception
        or warning is raised and `itera` >= 0, see `itera`), then `a` is
        unchanged. If full precision factorization has been used (no
        exception or warning is raised and `itera` < 0, see `itera`),
        then the array A contains the factor U or L from the Cholesky
        factorization A = U^TU or A = LL^T.

    x : float, ndarray, shape (`n`, `nrhs`)
        If no exception or warning is raised, the n by r solution matrix
        X.

    itera : int
        Information on the progress of the interative refinement
        process.

        `itera` < 0

            Iterative refinement has failed for one of the reasons given
            below, full precision factorization has been performed
            instead.

            [table omitted]

        `itera` > 0

            Iterative refinement has been sucessfully used. `itera`
            returns the number of iterations.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -3)
            On entry, error in parameter `nrhs`.

            Constraint: `nrhs` >= 0.

        (`errno` (i > 0) and (i<=`n`))
            The leading minor of order *<value>* of A is not positive
            definite, so the factorization could not be completed, and
            the solution has not been computed.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    fun_name = 'naginterfaces.library.lapacklin.dsposv'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_a, 'a'), (_b, 'b')),
    )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _nrhs = _b_utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineIntScalarType.to_py(_nrhs)
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py, _nrhs_py),
        sorder=_m_sorder,
    )
    _return_scalar = _b_lapacklin.dsposv(
        uplo,
        _n,
        _nrhs,
        _a,
        _b,
        _x,
    )
    _a_py = _a
    _x_py = _x
    _itera_py = _return_scalar
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'x', 'itera'],
    )(
        _a_py,
        _x_py,
        _itera_py,
    )
    return _return_data

def dpotrf(uplo, a):
    u"""
    Cholesky factorization of real symmetric positive definite matrix.

    ``dpotrf`` computes the Cholesky factorization of a real symmetric
    positive definite matrix.

    For full information please refer to the NAG Library document for
    f07fd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07fdf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored and how A is to be factorized.

        `uplo` = 'U'

            The upper triangular part of A is stored and A is factorized
            as U^TU, where U is upper triangular.

        `uplo` = 'L'

            The lower triangular part of A is stored and A is factorized
            as LL^T, where L is lower triangular.

    a : float, array-like, shape (n, n)
        The n by n symmetric positive definite matrix A.

    Returns
    -------
    a : float, ndarray, shape (n, n)
        The upper or lower triangle of A is overwritten by the Cholesky
        factor U or L as specified by `uplo`.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` i > 0)
            The leading minor of order *<value>* is not positive
            definite and the factorization could not be completed. Hence
            A itself is not positive definite. This may indicate an
            error in forming the matrix A. To factorize a symmetric
            matrix which is not positive definite, call ``dsytrf``
            instead.

    References
    ----------
    Demmel, J W, 1989, `On floating-point errors in Cholesky`, LAPACK
    Working Note No. 14, University of Tennessee, Knoxville,
    https://www.netlib.org/lapack/lawnspdf/lawn14.pdf

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.dpotrf'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b_lapacklin.dpotrf(
        uplo,
        _n,
        _a,
    )
    _a_py = _a
    return _a_py

def dpotrs(uplo, a, b):
    u"""
    Solution of real symmetric positive definite system of linear
    equations, multiple right-hand sides, matrix already factorized by
    ``dpotrf``.

    ``dpotrs`` solves a real symmetric positive definite system of
    linear equations with multiple right-hand sides,

        AX = B,

    where A has been factorized by ``dpotrf``.

    For full information please refer to the NAG Library document for
    f07fe

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07fef.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = U^TU, where U is upper triangular.

        `uplo` = 'L'

            A = LL^T, where L is lower triangular.

    a : float, array-like, shape (n, n)
        The Cholesky factor of A, as returned by ``dpotrf``.

    b : float, array-like, shape (n, nrhs)
        The n by r right-hand side matrix B.

    Returns
    -------
    b : float, ndarray, shape (n, nrhs)
        The n by r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.dpotrs'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nrhs_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _b_lapacklin.dpotrs(
        uplo,
        _n,
        _nrhs,
        _a,
        _b,
    )
    _b_py = _b
    return _b_py

def dpoequ(n, a):
    u"""
    Computes row and column scalings intended to equilibrate a real
    symmetric positive definite matrix and reduce its condition number.

    ``dpoequ`` computes a diagonal scaling matrix S intended to
    equilibrate a real n by n symmetric positive definite matrix A and
    reduce its condition number.

    For full information please refer to the NAG Library document for
    f07ff

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07fff.html

    Parameters
    ----------
    n : int
        n, the order of the matrix A.

    a : float, array-like, shape (`n`, `n`)
        The matrix A whose scaling factors are to be computed. Only the
        diagonal elements of the array `a` are referenced.

    Returns
    -------
    s : float, ndarray, shape (`n`)
        If no exception or warning is raised, `s` contains the diagonal
        elements of the scaling matrix S.

    scond : float
        If no exception or warning is raised, `scond` contains the ratio
        of the smallest value of `s` to the largest value of `s`. If
        `scond` >= 0.1 and `amax` is neither too large nor too small, it
        is not worth scaling by S.

    amax : float
        max|a_{ij}|. If `amax` is very close to overflow or underflow,
        the matrix A should be scaled.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` i > 0)
            The *<value>*th diagonal element of A is not positive (and
            hence A cannot be positive definite).

    References
    ----------
    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    fun_name = 'naginterfaces.library.lapacklin.dpoequ'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _s = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _return_scalars = _b_lapacklin.dpoequ(
        _n,
        _a,
        _s,
    )
    _s_py = _s
    _scond_py = _return_scalars.scond
    _amax_py = _return_scalars.amax
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['s', 'scond', 'amax'],
    )(
        _s_py,
        _scond_py,
        _amax_py,
    )
    return _return_data

def dpocon(uplo, n, a, anorm):
    u"""
    Estimate condition number of real symmetric positive definite
    matrix, matrix already factorized by ``dpotrf``.

    ``dpocon`` estimates the condition number of a real symmetric
    positive definite matrix A, where A has been factorized by
    ``dpotrf``.

    For full information please refer to the NAG Library document for
    f07fg

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07fgf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = U^TU, where U is upper triangular.

        `uplo` = 'L'

            A = LL^T, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    a : float, array-like, shape (`n`, `n`)
        The Cholesky factor of A, as returned by ``dpotrf``.

    anorm : float
        The 1-norm of the **original** matrix A, which may be computed
        by calling ``blas.dlansy`` with its argument norm = '1'. `anorm`
        must be computed either **before** calling ``dpotrf`` or else
        from a **copy** of the original matrix A.

    Returns
    -------
    rcond : float
        An estimate of the reciprocal of the condition number of A.
        `rcond` is set to zero if exact singularity is detected or the
        estimate underflows. If `rcond` is less than machine precision,
        A is singular to working precision.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -5)
            On entry, error in parameter `anorm`.

            Constraint: `anorm` >= 0.0.

    References
    ----------
    Higham, N J, 1988, `FORTRAN codes for estimating the one-norm of a
    real or complex matrix, with applications to condition estimation`,
    ACM Trans. Math. Software (14), 381--396
    """

    fun_name = 'naginterfaces.library.lapacklin.dpocon'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _return_scalar = _b_lapacklin.dpocon(
        uplo,
        n,
        _a,
        anorm,
    )
    _rcond_py = _return_scalar
    return _rcond_py

def dporfs(uplo, n, a, af, b, x):
    u"""
    Refined solution with error bounds of real symmetric positive
    definite system of linear equations, multiple right-hand sides.

    ``dporfs`` returns error bounds for the solution of a real symmetric
    positive definite system of linear equations with multiple
    right-hand sides, AX = B.
    It improves the solution by iterative refinement, in order to reduce
    the backward error as much as possible.

    For full information please refer to the NAG Library document for
    f07fh

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07fhf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored and how A is to be factorized.

        `uplo` = 'U'

            The upper triangular part of A is stored and A is factorized
            as U^TU, where U is upper triangular.

        `uplo` = 'L'

            The lower triangular part of A is stored and A is factorized
            as LL^T, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    a : float, array-like, shape (`n`, `n`)
        The n by n original symmetric positive definite matrix A as
        supplied to ``dpotrf``.

    af : float, array-like, shape (`n`, `n`)
        The Cholesky factor of A, as returned by ``dpotrf``.

    b : float, array-like, shape (`n`, nrhs)
        The n by r right-hand side matrix B.

    x : float, array-like, shape (`n`, nrhs)
        The n by r solution matrix X, as returned by ``dpotrs``.

    Returns
    -------
    x : float, ndarray, shape (`n`, nrhs)
        The improved solution matrix X.

    ferr : float, ndarray, shape (nrhs)
        `ferr`[j-1] contains an estimated error bound for the jth
        solution vector, that is, the jth column of X, for j =
        1,2,...,r.

    berr : float, ndarray, shape (nrhs)
        `berr`[j-1] contains the component-wise backward error bound
        beta for the jth solution vector, that is, the jth column of X,
        for j = 1,2,...,r.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -3)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.dporfs'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nrhs_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _af = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'af'},
        af,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _b_lapacklin.dporfs(
        uplo,
        n,
        _nrhs,
        _a,
        _af,
        _b,
        _x,
        _ferr,
        _berr,
    )
    _x_py = _x
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'ferr', 'berr'],
    )(
        _x_py,
        _ferr_py,
        _berr_py,
    )
    return _return_data

def dpotri(uplo, a):
    u"""
    Inverse of real symmetric positive definite matrix, matrix already
    factorized by ``dpotrf``.

    ``dpotri`` computes the inverse of a real symmetric positive
    definite matrix A, where A has been factorized by ``dpotrf``.

    For full information please refer to the NAG Library document for
    f07fj

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07fjf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = U^TU, where U is upper triangular.

        `uplo` = 'L'

            A = LL^T, where L is lower triangular.

    a : float, array-like, shape (n, n)
        The upper triangular matrix U if `uplo` = 'U' or the lower
        triangular matrix L if `uplo` = 'L', as returned by ``dpotrf``.

    Returns
    -------
    a : float, ndarray, shape (n, n)
        U is overwritten by the upper triangle of A^-1 if `uplo` = 'U';
        L is overwritten by the lower triangle of A^-1 if `uplo` = 'L'.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Diagonal element *<value>* of the Cholesky factor is zero;
            the Cholesky factor is singular and the inverse of A cannot
            be computed.

    References
    ----------
    Du Croz, J J and Higham, N J, 1992, `Stability of methods for matrix
    inversion`, IMA J. Numer. Anal. (12), 1--19
    """

    fun_name = 'naginterfaces.library.lapacklin.dpotri'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    try:
        _b_lapacklin.dpotri(
            uplo,
            _n,
            _a,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _a_py = _a
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a'],
    )(
        _a_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _a_py

def zposv(uplo, a, b):
    u"""
    Computes the solution to a complex Hermitian positive definite
    system of linear equations.

    ``zposv`` computes the solution to a complex system of linear
    equations

        AX = B,

    where A is an n by n Hermitian positive definite matrix and X and B
    are n by r matrices.

    For full information please refer to the NAG Library document for
    f07fn

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07fnf.html

    Parameters
    ----------
    uplo : str, length 1
        If `uplo` = 'U', the upper triangle of A is stored.

        If `uplo` = 'L', the lower triangle of A is stored.

    a : complex, array-like, shape (n, n)
        The n by n Hermitian matrix A.

    b : complex, array-like, shape (n, nrhs)
        The n by r right-hand side matrix B.

    Returns
    -------
    a : complex, ndarray, shape (n, n)
        If no exception or warning is raised, the factor U or L from the
        Cholesky factorization A = U^HU or A = LL^H.

    b : complex, ndarray, shape (n, nrhs)
        If no exception or warning is raised, the n by r solution matrix
        X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

        (`errno` i > 0)
            The leading minor of order *<value>* of A is not positive
            definite, so the factorization could not be completed, and
            the solution has not been computed.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.zposv'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
    )
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _b_lapacklin.zposv(
        uplo,
        _n,
        _nrhs,
        _a,
        _b,
    )
    _a_py = _a
    _b_py = _b
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'b'],
    )(
        _a_py,
        _b_py,
    )
    return _return_data

def zposvx(fact, uplo, n, nrhs, a, af, equed, s, b):
    u"""
    Uses the Cholesky factorization to compute the solution, error-bound
    and condition estimate for a complex Hermitian positive definite
    system of linear equations.

    ``zposvx`` uses the Cholesky factorization

        A = U^HU or A = LL^H

    to compute the solution to a complex system of linear equations

        AX = B,

    where A is an n by n Hermitian positive definite matrix and X and B
    are n by r matrices.
    Error bounds on the solution and a condition estimate are also
    provided.

    For full information please refer to the NAG Library document for
    f07fp

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07fpf.html

    Parameters
    ----------
    fact : str, length 1
        Specifies whether or not the factorized form of the matrix A is
        supplied on entry, and if not, whether the matrix A should be
        equilibrated before it is factorized.

        `fact` = 'F'

            `af` contains the factorized form of A. If `equed` = 'Y',
            the matrix A has been equilibrated with scaling factors
            given by `s`. `a` and `af` will not be modified.

        `fact` = 'N'

            The matrix A will be copied to `af` and factorized.

        `fact` = 'E'

            The matrix A will be equilibrated if necessary, then copied
            to `af` and factorized.

    uplo : str, length 1
        If `uplo` = 'U', the upper triangle of A is stored.

        If `uplo` = 'L', the lower triangle of A is stored.

    n : int
        n, the number of linear equations, i.e., the order of the matrix
        A.

    nrhs : int
        r, the number of right-hand sides, i.e., the number of columns
        of the matrix B.

    a : complex, array-like, shape (`n`, `n`)
        The n by n Hermitian matrix A.

        If `fact` = 'F' and `equed` = 'Y', `a` must have been
        equilibrated by the scaling factor in `s` as D_S AD_S.

    af : complex, array-like, shape (`n`, `n`)
        If `fact` = 'F', `af` contains the triangular factor U or L from
        the Cholesky factorization A = U^HU or A = LL^H, in the same
        storage format as `a`. If `equed` != 'N', `af` is the factorized
        form of the equilibrated matrix D_S AD_S.

    equed : str, length 1
        If `fact` = 'N' or 'E', `equed` need not be set.

        If `fact` = 'F', `equed` must specify the form of the
        equilibration that was performed as follows:

            if `equed` = 'N', no equilibration;

            if `equed` = 'Y', equilibration was performed, i.e., A has
            been replaced by D_S AD_S.

    s : float, array-like, shape (`n`)
        If `fact` = 'N' or 'E', `s` need not be set.

        If `fact` = 'F' and `equed` = 'Y', `s` must contain the scale
        factors, D_S, for A; each element of `s` must be positive.

    b : complex, array-like, shape (`n`, `nrhs`)
        The n by r right-hand side matrix B.

    Returns
    -------
    a : complex, ndarray, shape (`n`, `n`)
        If `fact` = 'F' or 'N', or if `fact` = 'E' and `equed` = 'N',
        `a` is not modified.

        If `fact` = 'E' and `equed` = 'Y', `a` is overwritten by D_S
        AD_S.

    af : complex, ndarray, shape (`n`, `n`)
        If `fact` = 'N', `af` returns the triangular factor U or L from
        the Cholesky factorization A = U^HU or A = LL^H of the original
        matrix A.

        If `fact` = 'E', `af` returns the triangular factor U or L from
        the Cholesky factorization A = U^HU or A = LL^H of the
        equilibrated matrix A (see the description of `a` for the form
        of the equilibrated matrix).

    equed : str, length 1
        If `fact` = 'F', `equed` is unchanged from entry.

        Otherwise, if no constraints are violated, `equed` specifies the
        form of the equilibration that was performed as specified above.

    s : float, ndarray, shape (`n`)
        If `fact` = 'F', `s` is unchanged from entry.

        Otherwise, if no constraints are violated and `equed` = 'Y', `s`
        contains the scale factors, D_S, for A; each element of `s` is
        positive.

    b : complex, ndarray, shape (`n`, `nrhs`)
        If `equed` = 'N', `b` is not modified.

        If `equed` = 'Y', `b` is overwritten by D_S B.

    x : complex, ndarray, shape (`n`, `nrhs`)
        If the function exits successfully or `errno` = `n` + 1, the n
        by r solution matrix X to the original system of equations. Note
        that the arrays A and B are modified on exit if `equed` = 'Y',
        and the solution to the equilibrated system is D_S^-1X.

    rcond : float
        If no constraints are violated, an estimate of the reciprocal
        condition number of the matrix A (after equilibration if that is
        performed), computed as `rcond` = 1.0/(||A||_1||A^-1||_1).

    ferr : float, ndarray, shape (`nrhs`)
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the forward error bound for each computed solution
        vector, such that ||x^_j-x_j||_infinity/||x_j||_infinity <=
        `ferr`[j-1] where x^_j is the jth column of the computed
        solution returned in the array `x` and x_j is the corresponding
        column of the exact solution X. The estimate is as reliable as
        the estimate for `rcond`, and is almost always a slight
        overestimate of the true error.

    berr : float, ndarray, shape (`nrhs`)
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the component-wise relative backward error of each
        computed solution vector x^_j (i.e., the smallest relative
        change in any element of A or B that makes x^_j an exact
        solution).

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `fact`.

            Constraint: `fact` = 'F', 'N' or 'E'.

        (`errno` -2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -4)
            On entry, error in parameter `nrhs`.

            Constraint: `nrhs` >= 0.

        (`errno` -9)
            On entry, error in parameter `equed`.

            Constraint: `equed` = 'N' or 'Y'.

        (`errno` -10)
            On entry, error in parameter `s`.

        (`errno` (i > 0) and (i<=`n`))
            The leading minor of order *<value>* of A is not positive
            definite, so the factorization could not be completed, and
            the solution has not been computed. `rcond` = 0.0 is
            returned.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` `n`+1)
            U (or L) is nonsingular, but `rcond` is less than machine
            precision, meaning that the matrix is singular to working
            precision. Nevertheless, the solution and error bounds are
            computed because there are a number of situations where the
            computed solution can be more accurate than the value of
            `rcond` would suggest.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    fun_name = 'naginterfaces.library.lapacklin.zposvx'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _af = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'af'},
        af,
        copy=True,
    )
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_a, 'a'), (_af, 'af'), (_b, 'b')),
    )
    _s = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 's'},
        s,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _nrhs = _b_utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineIntScalarType.to_py(_nrhs)
    _x = _b_utils._EngineComplex128ArrayType.empty_ndarray(
        (_n_py, _nrhs_py),
        sorder=_m_sorder,
    )
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    try:
        _return_scalars = _b_lapacklin.zposvx(
            fact,
            uplo,
            _n,
            _nrhs,
            _a,
            _af,
            equed,
            _s,
            _b,
            _x,
            _ferr,
            _berr,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _a_py = _a
    _af_py = _af
    _equed_py = _return_scalars.equed
    _s_py = _s
    _b_py = _b
    _x_py = _x
    _rcond_py = _return_scalars.rcond
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'af', 'equed', 's', 'b', 'x', 'rcond', 'ferr', 'berr'],
    )(
        _a_py,
        _af_py,
        _equed_py,
        _s_py,
        _b_py,
        _x_py,
        _rcond_py,
        _ferr_py,
        _berr_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def zcposv(uplo, n, nrhs, a, b):
    u"""
    Computes the solution to a complex Hermitian positive definite
    system of linear equations using mixed precision arithmetic.

    ``zcposv`` uses the Cholesky factorization

        A = U^HU or A = LL^H

    to compute the solution to a complex system of linear equations

        AX = B,

    where A is an n by n Hermitian positive definite matrix and X and B
    are n by r matrices.

    For full information please refer to the NAG Library document for
    f07fq

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07fqf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

            The upper triangular part of A is stored.

        `uplo` = 'L'

            The lower triangular part of A is stored.

    n : int
        n, the number of linear equations, i.e., the order of the matrix
        A.

    nrhs : int
        r, the number of right-hand sides, i.e., the number of columns
        of the matrix B.

    a : complex, array-like, shape (`n`, `n`)
        The n by n Hermitian positive definite matrix A.

    b : complex, array-like, shape (`n`, `nrhs`)
        The right-hand side matrix B.

    Returns
    -------
    a : complex, ndarray, shape (`n`, `n`)
        If iterative refinement has been successfully used (no exception
        or warning is raised and `itera` >= 0, see `itera`), then `a` is
        unchanged. If full precision factorization has been used (no
        exception or warning is raised and `itera` < 0, see `itera`),
        then the array A contains the factor U or L from the Cholesky
        factorization A = U^HU or A = LL^H.

    x : complex, ndarray, shape (`n`, `nrhs`)
        If no exception or warning is raised, the n by r solution matrix
        X.

    itera : int
        Information on the progress of the interative refinement
        process.

        `itera` < 0

            Iterative refinement has failed for one of the reasons given
            below, full precision factorization has been performed
            instead.

            [table omitted]

        `itera` > 0

            Iterative refinement has been sucessfully used. `itera`
            returns the number of iterations.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -3)
            On entry, error in parameter `nrhs`.

            Constraint: `nrhs` >= 0.

        (`errno` (i > 0) and (i<=`n`))
            The leading minor of order *<value>* of A is not positive
            definite, so the factorization could not be completed, and
            the solution has not been computed.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    fun_name = 'naginterfaces.library.lapacklin.zcposv'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_a, 'a'), (_b, 'b')),
    )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _nrhs = _b_utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineIntScalarType.to_py(_nrhs)
    _x = _b_utils._EngineComplex128ArrayType.empty_ndarray(
        (_n_py, _nrhs_py),
        sorder=_m_sorder,
    )
    _return_scalar = _b_lapacklin.zcposv(
        uplo,
        _n,
        _nrhs,
        _a,
        _b,
        _x,
    )
    _a_py = _a
    _x_py = _x
    _itera_py = _return_scalar
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'x', 'itera'],
    )(
        _a_py,
        _x_py,
        _itera_py,
    )
    return _return_data

def zpotrf(uplo, a):
    u"""
    Cholesky factorization of complex Hermitian positive definite
    matrix.

    ``zpotrf`` computes the Cholesky factorization of a complex
    Hermitian positive definite matrix.

    For full information please refer to the NAG Library document for
    f07fr

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07frf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored and how A is to be factorized.

        `uplo` = 'U'

            The upper triangular part of A is stored and A is factorized
            as U^HU, where U is upper triangular.

        `uplo` = 'L'

            The lower triangular part of A is stored and A is factorized
            as LL^H, where L is lower triangular.

    a : complex, array-like, shape (n, n)
        The n by n Hermitian positive definite matrix A.

    Returns
    -------
    a : complex, ndarray, shape (n, n)
        The upper or lower triangle of A is overwritten by the Cholesky
        factor U or L as specified by `uplo`.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` i > 0)
            The leading minor of order *<value>* is not positive
            definite and the factorization could not be completed. Hence
            A itself is not positive definite. This may indicate an
            error in forming the matrix A. To factorize a Hermitian
            matrix which is not positive definite, call ``zhetrf``
            instead.

    References
    ----------
    Demmel, J W, 1989, `On floating-point errors in Cholesky`, LAPACK
    Working Note No. 14, University of Tennessee, Knoxville,
    https://www.netlib.org/lapack/lawnspdf/lawn14.pdf

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.zpotrf'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b_lapacklin.zpotrf(
        uplo,
        _n,
        _a,
    )
    _a_py = _a
    return _a_py

def zpotrs(uplo, a, b):
    u"""
    Solution of complex Hermitian positive definite system of linear
    equations, multiple right-hand sides, matrix already factorized by
    ``zpotrf``.

    ``zpotrs`` solves a complex Hermitian positive definite system of
    linear equations with multiple right-hand sides,

        AX = B,

    where A has been factorized by ``zpotrf``.

    For full information please refer to the NAG Library document for
    f07fs

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07fsf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = U^HU, where U is upper triangular.

        `uplo` = 'L'

            A = LL^H, where L is lower triangular.

    a : complex, array-like, shape (n, n)
        The Cholesky factor of A, as returned by ``zpotrf``.

    b : complex, array-like, shape (n, nrhs)
        The n by r right-hand side matrix B.

    Returns
    -------
    b : complex, ndarray, shape (n, nrhs)
        The n by r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.zpotrs'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
    )
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _b_lapacklin.zpotrs(
        uplo,
        _n,
        _nrhs,
        _a,
        _b,
    )
    _b_py = _b
    return _b_py

def zpoequ(n, a):
    u"""
    Computes row and column scalings intended to equilibrate a complex
    Hermitian positive definite matrix and reduce its condition number.

    ``zpoequ`` computes a diagonal scaling matrix S intended to
    equilibrate a complex n by n Hermitian positive definite matrix A
    and reduce its condition number.

    For full information please refer to the NAG Library document for
    f07ft

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07ftf.html

    Parameters
    ----------
    n : int
        n, the order of the matrix A.

    a : complex, array-like, shape (`n`, `n`)
        The matrix A whose scaling factors are to be computed. Only the
        diagonal elements of the array `a` are referenced.

    Returns
    -------
    s : float, ndarray, shape (`n`)
        If no exception or warning is raised, `s` contains the diagonal
        elements of the scaling matrix S.

    scond : float
        If no exception or warning is raised, `scond` contains the ratio
        of the smallest value of `s` to the largest value of `s`. If
        `scond` >= 0.1 and `amax` is neither too large nor too small, it
        is not worth scaling by S.

    amax : float
        max|a_{ij}|. If `amax` is very close to overflow or underflow,
        the matrix A should be scaled.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` i > 0)
            The *<value>*th diagonal element of A is not positive (and
            hence A cannot be positive definite).

    References
    ----------
    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    fun_name = 'naginterfaces.library.lapacklin.zpoequ'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
    )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _s = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _return_scalars = _b_lapacklin.zpoequ(
        _n,
        _a,
        _s,
    )
    _s_py = _s
    _scond_py = _return_scalars.scond
    _amax_py = _return_scalars.amax
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['s', 'scond', 'amax'],
    )(
        _s_py,
        _scond_py,
        _amax_py,
    )
    return _return_data

def zpocon(uplo, n, a, anorm):
    u"""
    Estimate condition number of complex Hermitian positive definite
    matrix, matrix already factorized by ``zpotrf``.

    ``zpocon`` estimates the condition number of a complex Hermitian
    positive definite matrix A, where A has been factorized by
    ``zpotrf``.

    For full information please refer to the NAG Library document for
    f07fu

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07fuf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = U^HU, where U is upper triangular.

        `uplo` = 'L'

            A = LL^H, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    a : complex, array-like, shape (`n`, `n`)
        The Cholesky factor of A, as returned by ``zpotrf``.

    anorm : float
        The 1-norm of the **original** matrix A, which may be computed
        by calling ``blas.zlanhe`` with its argument norm = '1'. `anorm`
        must be computed either **before** calling ``zpotrf`` or else
        from a **copy** of the original matrix A.

    Returns
    -------
    rcond : float
        An estimate of the reciprocal of the condition number of A.
        `rcond` is set to zero if exact singularity is detected or the
        estimate underflows. If `rcond` is less than machine precision,
        A is singular to working precision.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -5)
            On entry, error in parameter `anorm`.

            Constraint: `anorm` >= 0.0.

    References
    ----------
    Higham, N J, 1988, `FORTRAN codes for estimating the one-norm of a
    real or complex matrix, with applications to condition estimation`,
    ACM Trans. Math. Software (14), 381--396
    """

    fun_name = 'naginterfaces.library.lapacklin.zpocon'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
    )
    _return_scalar = _b_lapacklin.zpocon(
        uplo,
        n,
        _a,
        anorm,
    )
    _rcond_py = _return_scalar
    return _rcond_py

def zporfs(uplo, n, a, af, b, x):
    u"""
    Refined solution with error bounds of complex Hermitian positive
    definite system of linear equations, multiple right-hand sides.

    ``zporfs`` returns error bounds for the solution of a complex
    Hermitian positive definite system of linear equations with multiple
    right-hand sides, AX = B.
    It improves the solution by iterative refinement, in order to reduce
    the backward error as much as possible.

    For full information please refer to the NAG Library document for
    f07fv

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07fvf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored and how A is to be factorized.

        `uplo` = 'U'

            The upper triangular part of A is stored and A is factorized
            as U^HU, where U is upper triangular.

        `uplo` = 'L'

            The lower triangular part of A is stored and A is factorized
            as LL^H, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    a : complex, array-like, shape (`n`, `n`)
        The n by n original Hermitian positive definite matrix A as
        supplied to ``zpotrf``.

    af : complex, array-like, shape (`n`, `n`)
        The Cholesky factor of A, as returned by ``zpotrf``.

    b : complex, array-like, shape (`n`, nrhs)
        The n by r right-hand side matrix B.

    x : complex, array-like, shape (`n`, nrhs)
        The n by r solution matrix X, as returned by ``zpotrs``.

    Returns
    -------
    x : complex, ndarray, shape (`n`, nrhs)
        The improved solution matrix X.

    ferr : float, ndarray, shape (nrhs)
        `ferr`[j-1] contains an estimated error bound for the jth
        solution vector, that is, the jth column of X, for j =
        1,2,...,r.

    berr : float, ndarray, shape (nrhs)
        `berr`[j-1] contains the component-wise backward error bound
        beta for the jth solution vector, that is, the jth column of X,
        for j = 1,2,...,r.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -3)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.zporfs'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
    )
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
    )
    _af = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'af'},
        af,
        copy=False,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
    )
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _b_lapacklin.zporfs(
        uplo,
        n,
        _nrhs,
        _a,
        _af,
        _b,
        _x,
        _ferr,
        _berr,
    )
    _x_py = _x
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'ferr', 'berr'],
    )(
        _x_py,
        _ferr_py,
        _berr_py,
    )
    return _return_data

def zpotri(uplo, a):
    u"""
    Inverse of complex Hermitian positive definite matrix, matrix
    already factorized by ``zpotrf``.

    ``zpotri`` computes the inverse of a complex Hermitian positive
    definite matrix A, where A has been factorized by ``zpotrf``.

    For full information please refer to the NAG Library document for
    f07fw

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07fwf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = U^HU, where U is upper triangular.

        `uplo` = 'L'

            A = LL^H, where L is lower triangular.

    a : complex, array-like, shape (n, n)
        The upper triangular matrix U if `uplo` = 'U' or the lower
        triangular matrix L if `uplo` = 'L', as returned by ``zpotrf``.

    Returns
    -------
    a : complex, ndarray, shape (n, n)
        U is overwritten by the upper triangle of A^-1 if `uplo` = 'U';
        L is overwritten by the lower triangle of A^-1 if `uplo` = 'L'.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Diagonal element *<value>* of the Cholesky factor is zero;
            the Cholesky factor is singular and the inverse of A cannot
            be computed.

    References
    ----------
    Du Croz, J J and Higham, N J, 1992, `Stability of methods for matrix
    inversion`, IMA J. Numer. Anal. (12), 1--19
    """

    fun_name = 'naginterfaces.library.lapacklin.zpotri'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    try:
        _b_lapacklin.zpotri(
            uplo,
            _n,
            _a,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _a_py = _a
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a'],
    )(
        _a_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _a_py

def dppsv(uplo, ap, b):
    u"""
    Computes the solution to a real symmetric positive definite system
    of linear equations, packed storage.

    ``dppsv`` computes the solution to a real system of linear equations

        AX = B,

    where A is an n by n symmetric positive definite matrix stored in
    packed format and X and B are n by r matrices.

    For full information please refer to the NAG Library document for
    f07ga

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07gaf.html

    Parameters
    ----------
    uplo : str, length 1
        If `uplo` = 'U', the upper triangle of A is stored.

        If `uplo` = 'L', the lower triangle of A is stored.

    ap : float, array-like, shape (n*(n+1)/2)
        The n by n symmetric matrix A, packed by columns.

    b : float, array-like, shape (n, nrhs)
        The n by r right-hand side matrix B.

    Returns
    -------
    ap : float, ndarray, shape (n*(n+1)/2)
        If no exception or warning is raised, the factor U or L from the
        Cholesky factorization A = U^TU or A = LL^T, in the same storage
        format as A.

    b : float, ndarray, shape (n, nrhs)
        If no exception or warning is raised, the n by r solution matrix
        X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

        (`errno` i > 0)
            The leading minor of order *<value>* of A is not positive
            definite, so the factorization could not be completed, and
            the solution has not been computed.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.dppsv'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b_lapacklin.dppsv(
        uplo,
        _n,
        _nrhs,
        _ap,
        _b,
    )
    _ap_py = _ap
    _b_py = _b
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['ap', 'b'],
    )(
        _ap_py,
        _b_py,
    )
    return _return_data

def dppsvx(fact, uplo, n, nrhs, ap, afp, equed, s, b):
    u"""
    Uses the Cholesky factorization to compute the solution, error-bound
    and condition estimate for a real symmetric positive definite system
    of linear equations, packed storage.

    ``dppsvx`` uses the Cholesky factorization

        A = U^TU or A = LL^T

    to compute the solution to a real system of linear equations

        AX = B,

    where A is an n by n symmetric positive definite matrix stored in
    packed format and X and B are n by r matrices.
    Error bounds on the solution and a condition estimate are also
    provided.

    For full information please refer to the NAG Library document for
    f07gb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07gbf.html

    Parameters
    ----------
    fact : str, length 1
        Specifies whether or not the factorized form of the matrix A is
        supplied on entry, and if not, whether the matrix A should be
        equilibrated before it is factorized.

        `fact` = 'F'

            `afp` contains the factorized form of A. If `equed` = 'Y',
            the matrix A has been equilibrated with scaling factors
            given by `s`. `ap` and `afp` will not be modified.

        `fact` = 'N'

            The matrix A will be copied to `afp` and factorized.

        `fact` = 'E'

            The matrix A will be equilibrated if necessary, then copied
            to `afp` and factorized.

    uplo : str, length 1
        If `uplo` = 'U', the upper triangle of A is stored.

        If `uplo` = 'L', the lower triangle of A is stored.

    n : int
        n, the number of linear equations, i.e., the order of the matrix
        A.

    nrhs : int
        r, the number of right-hand sides, i.e., the number of columns
        of the matrix B.

    ap : float, array-like, shape (`n`*(`n`+1)/2)
        If `fact` = 'F' and `equed` = 'Y', `ap` must contain the
        equilibrated matrix D_S AD_S; otherwise, `ap` must contain the n
        by n symmetric matrix A, packed by columns.

    afp : float, array-like, shape (`n`*(`n`+1)/2)
        If `fact` = 'F', `afp` contains the triangular factor U or L
        from the Cholesky factorization A = U^TU or A = LL^T, in the
        same storage format as `ap`. If `equed` = 'Y', `afp` is the
        factorized form of the equilibrated matrix D_S AD_S.

    equed : str, length 1
        If `fact` = 'N' or 'E', `equed` need not be set.

        If `fact` = 'F', `equed` must specify the form of the
        equilibration that was performed as follows:

            if `equed` = 'N', no equilibration;

            if `equed` = 'Y', equilibration was performed, i.e., A has
            been replaced by D_S AD_S.

    s : float, array-like, shape (`n`)
        If `fact` = 'N' or 'E', `s` need not be set.

        If `fact` = 'F' and `equed` = 'Y', `s` must contain the scale
        factors, D_S, for A; each element of `s` must be positive.

    b : float, array-like, shape (`n`, `nrhs`)
        The n by r right-hand side matrix B.

    Returns
    -------
    ap : float, ndarray, shape (`n`*(`n`+1)/2)
        If `fact` = 'F' or 'N', or if `fact` = 'E' and `equed` = 'N',
        `ap` is not modified.

        If `fact` = 'E' and `equed` = 'Y', `ap` is overwritten by D_S
        AD_S.

    afp : float, ndarray, shape (`n`*(`n`+1)/2)
        If `fact` = 'N' or if `fact` = 'E' and `equed` = 'N', `afp`
        returns the triangular factor U or L from the Cholesky
        factorization A = U^TU or A = LL^T of the original matrix A.

        If `fact` = 'E' and `equed` = 'Y', `afp` returns the triangular
        factor U or L from the Cholesky factorization A = U^TU or A =
        LL^T of the equilibrated matrix A (see the description of `ap`
        for the form of the equilibrated matrix).

    equed : str, length 1
        If `fact` = 'F', `equed` is unchanged from entry.

        Otherwise, if no constraints are violated, `equed` specifies the
        form of the equilibration that was performed as specified above.

    s : float, ndarray, shape (`n`)
        If `fact` = 'F', `s` is unchanged from entry.

        Otherwise, if no constraints are violated and `equed` = 'Y', `s`
        contains the scale factors, D_S, for A; each element of `s` is
        positive.

    b : float, ndarray, shape (`n`, `nrhs`)
        If `equed` = 'N', `b` is not modified.

        If `equed` = 'Y', `b` is overwritten by D_S B.

    x : float, ndarray, shape (`n`, `nrhs`)
        If the function exits successfully or `errno` = `n` + 1, the n
        by r solution matrix X to the original system of equations. Note
        that the arrays A and B are modified on exit if `equed` = 'Y',
        and the solution to the equilibrated system is D_S^-1X.

    rcond : float
        If no constraints are violated, an estimate of the reciprocal
        condition number of the matrix A (after equilibration if that is
        performed), computed as `rcond` = 1.0/(||A||_1||A^-1||_1).

    ferr : float, ndarray, shape (`nrhs`)
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the forward error bound for each computed solution
        vector, such that ||x^_j-x_j||_infinity/||x_j||_infinity <=
        `ferr`[j-1] where x^_j is the jth column of the computed
        solution returned in the array `x` and x_j is the corresponding
        column of the exact solution X. The estimate is as reliable as
        the estimate for `rcond`, and is almost always a slight
        overestimate of the true error.

    berr : float, ndarray, shape (`nrhs`)
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the component-wise relative backward error of each
        computed solution vector x^_j (i.e., the smallest relative
        change in any element of A or B that makes x^_j an exact
        solution).

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `fact`.

            Constraint: `fact` = 'F', 'N' or 'E'.

        (`errno` -2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -4)
            On entry, error in parameter `nrhs`.

            Constraint: `nrhs` >= 0.

        (`errno` -7)
            On entry, error in parameter `equed`.

            Constraint: `equed` = 'N' or 'Y'.

        (`errno` -8)
            On entry, error in parameter `s`.

        (`errno` (i > 0) and (i<=`n`))
            The leading minor of order *<value>* of A is not positive
            definite, so the factorization could not be completed, and
            the solution has not been computed. `rcond` = 0.0 is
            returned.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` `n`+1)
            U (or L) is nonsingular, but `rcond` is less than machine
            precision, meaning that the matrix is singular to working
            precision. Nevertheless, the solution and error bounds are
            computed because there are a number of situations where the
            computed solution can be more accurate than the value of
            `rcond` would suggest.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    fun_name = 'naginterfaces.library.lapacklin.dppsvx'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_b, 'b'),),
    )
    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _afp = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'afp'},
        afp,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _s = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 's'},
        s,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _nrhs = _b_utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineIntScalarType.to_py(_nrhs)
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py, _nrhs_py),
        sorder=_m_sorder,
    )
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    try:
        _return_scalars = _b_lapacklin.dppsvx(
            fact,
            uplo,
            _n,
            _nrhs,
            _ap,
            _afp,
            equed,
            _s,
            _b,
            _x,
            _ferr,
            _berr,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _ap_py = _ap
    _afp_py = _afp
    _equed_py = _return_scalars.equed
    _s_py = _s
    _b_py = _b
    _x_py = _x
    _rcond_py = _return_scalars.rcond
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['ap', 'afp', 'equed', 's', 'b', 'x', 'rcond', 'ferr', 'berr'],
    )(
        _ap_py,
        _afp_py,
        _equed_py,
        _s_py,
        _b_py,
        _x_py,
        _rcond_py,
        _ferr_py,
        _berr_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def dpptrf(uplo, n, ap):
    u"""
    Cholesky factorization of real symmetric positive definite matrix,
    packed storage.

    ``dpptrf`` computes the Cholesky factorization of a real symmetric
    positive definite matrix, using packed storage.

    For full information please refer to the NAG Library document for
    f07gd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07gdf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored and how A is to be factorized.

        `uplo` = 'U'

            The upper triangular part of A is stored and A is factorized
            as U^TU, where U is upper triangular.

        `uplo` = 'L'

            The lower triangular part of A is stored and A is factorized
            as LL^T, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    ap : float, array-like, shape (`n`*(`n`+1)/2)
        The n by n symmetric matrix A, packed by columns.

    Returns
    -------
    ap : float, ndarray, shape (`n`*(`n`+1)/2)
        If no exception or warning is raised, the factor U or L from the
        Cholesky factorization A = U^TU or A = LL^T, in the same storage
        format as A.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` i > 0)
            The leading minor of order *<value>* is not positive
            definite and the factorization could not be completed. Hence
            A itself is not positive definite. This may indicate an
            error in forming the matrix A. To factorize a symmetric
            matrix which is not positive definite, call ``dsptrf``
            instead.

    References
    ----------
    Demmel, J W, 1989, `On floating-point errors in Cholesky`, LAPACK
    Working Note No. 14, University of Tennessee, Knoxville,
    https://www.netlib.org/lapack/lawnspdf/lawn14.pdf

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.dpptrf'

    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b_lapacklin.dpptrf(
        uplo,
        n,
        _ap,
    )
    _ap_py = _ap
    return _ap_py

def dpptrs(uplo, n, ap, b):
    u"""
    Solution of real symmetric positive definite system of linear
    equations, multiple right-hand sides, matrix already factorized by
    ``dpptrf``, packed storage.

    ``dpptrs`` solves a real symmetric positive definite system of
    linear equations with multiple right-hand sides,

        AX = B,

    where A has been factorized by ``dpptrf``, using packed storage.

    For full information please refer to the NAG Library document for
    f07ge

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07gef.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = U^TU, where U is upper triangular.

        `uplo` = 'L'

            A = LL^T, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    ap : float, array-like, shape (`n`*(`n`+1)/2)
        The Cholesky factor of A stored in packed form, as returned by
        ``dpptrf``.

    b : float, array-like, shape (`n`, nrhs)
        The n by r right-hand side matrix B.

    Returns
    -------
    b : float, ndarray, shape (`n`, nrhs)
        The n by r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -3)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.dpptrs'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nrhs_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b_lapacklin.dpptrs(
        uplo,
        n,
        _nrhs,
        _ap,
        _b,
    )
    _b_py = _b
    return _b_py

def dppequ(uplo, n, ap):
    u"""
    Computes row and column scalings intended to equilibrate a real
    symmetric positive definite matrix and reduce its condition number,
    packed storage.

    ``dppequ`` computes a diagonal scaling matrix S intended to
    equilibrate a real n by n symmetric positive definite matrix A,
    stored in packed format, and reduce its condition number.

    For full information please refer to the NAG Library document for
    f07gf

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07gff.html

    Parameters
    ----------
    uplo : str, length 1
        Indicates whether the upper or lower triangular part of A is
        stored in the array `ap`, as follows:

        `uplo` = 'U'

            The upper triangle of A is stored.

        `uplo` = 'L'

            The lower triangle of A is stored.

    n : int
        n, the order of the matrix A.

    ap : float, array-like, shape (`n`*(`n`+1)/2)
        The n by n symmetric matrix A, packed by columns.

        Only the elements of `ap` corresponding to the diagonal elements
        A are referenced.

    Returns
    -------
    s : float, ndarray, shape (`n`)
        If no exception or warning is raised, `s` contains the diagonal
        elements of the scaling matrix S.

    scond : float
        If no exception or warning is raised, `scond` contains the ratio
        of the smallest value of `s` to the largest value of `s`. If
        `scond` >= 0.1 and `amax` is neither too large nor too small, it
        is not worth scaling by S.

    amax : float
        max|a_{ij}|. If `amax` is very close to overflow or underflow,
        the matrix A should be scaled.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` i > 0)
            The *<value>*th diagonal element of A is not positive (and
            hence A cannot be positive definite).

    References
    ----------
    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    fun_name = 'naginterfaces.library.lapacklin.dppequ'

    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _s = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _return_scalars = _b_lapacklin.dppequ(
        uplo,
        _n,
        _ap,
        _s,
    )
    _s_py = _s
    _scond_py = _return_scalars.scond
    _amax_py = _return_scalars.amax
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['s', 'scond', 'amax'],
    )(
        _s_py,
        _scond_py,
        _amax_py,
    )
    return _return_data

def dppcon(uplo, n, ap, anorm):
    u"""
    Estimate condition number of real symmetric positive definite
    matrix, matrix already factorized by ``dpptrf``, packed storage.

    ``dppcon`` estimates the condition number of a real symmetric
    positive definite matrix A, where A has been factorized by
    ``dpptrf``, using packed storage.

    For full information please refer to the NAG Library document for
    f07gg

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07ggf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = U^TU, where U is upper triangular.

        `uplo` = 'L'

            A = LL^T, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    ap : float, array-like, shape (`n`*(`n`+1)/2)
        The Cholesky factor of A stored in packed form, as returned by
        ``dpptrf``.

    anorm : float
        The 1-norm of the **original** matrix A, which may be computed
        by calling ``blas.dlansp`` with its argument norm = '1'. `anorm`
        must be computed either **before** calling ``dpptrf`` or else
        from a **copy** of the original matrix A.

    Returns
    -------
    rcond : float
        An estimate of the reciprocal of the condition number of A.
        `rcond` is set to zero if exact singularity is detected or the
        estimate underflows. If `rcond` is less than machine precision,
        A is singular to working precision.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -4)
            On entry, error in parameter `anorm`.

            Constraint: `anorm` >= 0.0.

    References
    ----------
    Higham, N J, 1988, `FORTRAN codes for estimating the one-norm of a
    real or complex matrix, with applications to condition estimation`,
    ACM Trans. Math. Software (14), 381--396
    """

    fun_name = 'naginterfaces.library.lapacklin.dppcon'

    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _return_scalar = _b_lapacklin.dppcon(
        uplo,
        n,
        _ap,
        anorm,
    )
    _rcond_py = _return_scalar
    return _rcond_py

def dpprfs(uplo, n, ap, afp, b, x):
    u"""
    Refined solution with error bounds of real symmetric positive
    definite system of linear equations, multiple right-hand sides,
    packed storage.

    ``dpprfs`` returns error bounds for the solution of a real symmetric
    positive definite system of linear equations with multiple
    right-hand sides, AX = B, using packed storage.
    It improves the solution by iterative refinement, in order to reduce
    the backward error as much as possible.

    For full information please refer to the NAG Library document for
    f07gh

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07ghf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored and how A is to be factorized.

        `uplo` = 'U'

            The upper triangular part of A is stored and A is factorized
            as U^TU, where U is upper triangular.

        `uplo` = 'L'

            The lower triangular part of A is stored and A is factorized
            as LL^T, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    ap : float, array-like, shape (`n`*(`n`+1)/2)
        The n by n original symmetric positive definite matrix A as
        supplied to ``dpptrf``.

    afp : float, array-like, shape (`n`*(`n`+1)/2)
        The Cholesky factor of A stored in packed form, as returned by
        ``dpptrf``.

    b : float, array-like, shape (`n`, nrhs)
        The n by r right-hand side matrix B.

    x : float, array-like, shape (`n`, nrhs)
        The n by r solution matrix X, as returned by ``dpptrs``.

    Returns
    -------
    x : float, ndarray, shape (`n`, nrhs)
        The improved solution matrix X.

    ferr : float, ndarray, shape (nrhs)
        `ferr`[j-1] contains an estimated error bound for the jth
        solution vector, that is, the jth column of X, for j =
        1,2,...,r.

    berr : float, ndarray, shape (nrhs)
        `berr`[j-1] contains the component-wise backward error bound
        beta for the jth solution vector, that is, the jth column of X,
        for j = 1,2,...,r.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -3)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.dpprfs'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nrhs_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _afp = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'afp'},
        afp,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _b_lapacklin.dpprfs(
        uplo,
        n,
        _nrhs,
        _ap,
        _afp,
        _b,
        _x,
        _ferr,
        _berr,
    )
    _x_py = _x
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'ferr', 'berr'],
    )(
        _x_py,
        _ferr_py,
        _berr_py,
    )
    return _return_data

def dpptri(uplo, n, ap):
    u"""
    Inverse of real symmetric positive definite matrix, matrix already
    factorized by ``dpptrf``, packed storage.

    ``dpptri`` computes the inverse of a real symmetric positive
    definite matrix A, where A has been factorized by ``dpptrf``, using
    packed storage.

    For full information please refer to the NAG Library document for
    f07gj

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07gjf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = U^TU, where U is upper triangular.

        `uplo` = 'L'

            A = LL^T, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    ap : float, array-like, shape (`n`*(`n`+1)/2)
        The Cholesky factor of A stored in packed form, as returned by
        ``dpptrf``.

    Returns
    -------
    ap : float, ndarray, shape (`n`*(`n`+1)/2)
        The factorization is overwritten by the n by n matrix A^-1.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Diagonal element *<value>* of the Cholesky factor is zero;
            the Cholesky factor is singular and the inverse of A cannot
            be computed.

    References
    ----------
    Du Croz, J J and Higham, N J, 1992, `Stability of methods for matrix
    inversion`, IMA J. Numer. Anal. (12), 1--19
    """

    fun_name = 'naginterfaces.library.lapacklin.dpptri'

    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    try:
        _b_lapacklin.dpptri(
            uplo,
            n,
            _ap,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _ap_py = _ap
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['ap'],
    )(
        _ap_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _ap_py

def zppsv(uplo, ap, b):
    u"""
    Computes the solution to a complex Hermitian positive definite
    system of linear equations, packed storage.

    ``zppsv`` computes the solution to a complex system of linear
    equations

        AX = B,

    where A is an n by n Hermitian positive definite matrix stored in
    packed format and X and B are n by r matrices.

    For full information please refer to the NAG Library document for
    f07gn

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07gnf.html

    Parameters
    ----------
    uplo : str, length 1
        If `uplo` = 'U', the upper triangle of A is stored.

        If `uplo` = 'L', the lower triangle of A is stored.

    ap : complex, array-like, shape (n*(n+1)/2)
        The n by n Hermitian matrix A, packed by columns.

    b : complex, array-like, shape (n, nrhs)
        The n by r right-hand side matrix B.

    Returns
    -------
    ap : complex, ndarray, shape (n*(n+1)/2)
        If no exception or warning is raised, the factor U or L from the
        Cholesky factorization A = U^HU or A = LL^H, in the same storage
        format as A.

    b : complex, ndarray, shape (n, nrhs)
        If no exception or warning is raised, the n by r solution matrix
        X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

        (`errno` i > 0)
            The leading minor of order *<value>* of A is not positive
            definite, so the factorization could not be completed, and
            the solution has not been computed.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.zppsv'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=True,
    )
    _b_lapacklin.zppsv(
        uplo,
        _n,
        _nrhs,
        _ap,
        _b,
    )
    _ap_py = _ap
    _b_py = _b
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['ap', 'b'],
    )(
        _ap_py,
        _b_py,
    )
    return _return_data

def zppsvx(fact, uplo, n, nrhs, ap, afp, equed, s, b):
    u"""
    Uses the Cholesky factorization to compute the solution, error-bound
    and condition estimate for a complex Hermitian positive definite
    system of linear equations, packed storage.

    ``zppsvx`` uses the Cholesky factorization

        A = U^HU or A = LL^H

    to compute the solution to a complex system of linear equations

        AX = B,

    where A is an n by n Hermitian positive definite matrix stored in
    packed format and X and B are n by r matrices.
    Error bounds on the solution and a condition estimate are also
    provided.

    For full information please refer to the NAG Library document for
    f07gp

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07gpf.html

    Parameters
    ----------
    fact : str, length 1
        Specifies whether or not the factorized form of the matrix A is
        supplied on entry, and if not, whether the matrix A should be
        equilibrated before it is factorized.

        `fact` = 'F'

            `afp` contains the factorized form of A. If `equed` = 'Y',
            the matrix A has been equilibrated with scaling factors
            given by `s`. `ap` and `afp` will not be modified.

        `fact` = 'N'

            The matrix A will be copied to `afp` and factorized.

        `fact` = 'E'

            The matrix A will be equilibrated if necessary, then copied
            to `afp` and factorized.

    uplo : str, length 1
        If `uplo` = 'U', the upper triangle of A is stored.

        If `uplo` = 'L', the lower triangle of A is stored.

    n : int
        n, the number of linear equations, i.e., the order of the matrix
        A.

    nrhs : int
        r, the number of right-hand sides, i.e., the number of columns
        of the matrix B.

    ap : complex, array-like, shape (`n`*(`n`+1)/2)
        If `fact` = 'F' and `equed` = 'Y', `ap` must contain the
        equilibrated matrix D_S AD_S; otherwise, `ap` must contain the n
        by n Hermitian matrix A, packed by columns.

    afp : complex, array-like, shape (`n`*(`n`+1)/2)
        If `fact` = 'F', `afp` contains the triangular factor U or L
        from the Cholesky factorization A = U^HU or A = LL^H, in the
        same storage format as `ap`. If `equed` = 'Y', `afp` is the
        factorized form of the equilibrated matrix D_S AD_S.

    equed : str, length 1
        If `fact` = 'N' or 'E', `equed` need not be set.

        If `fact` = 'F', `equed` must specify the form of the
        equilibration that was performed as follows:

            if `equed` = 'N', no equilibration;

            if `equed` = 'Y', equilibration was performed, i.e., A has
            been replaced by D_S AD_S.

    s : float, array-like, shape (`n`)
        If `fact` = 'N' or 'E', `s` need not be set.

        If `fact` = 'F' and `equed` = 'Y', `s` must contain the scale
        factors, D_S, for A; each element of `s` must be positive.

    b : complex, array-like, shape (`n`, `nrhs`)
        The n by r right-hand side matrix B.

    Returns
    -------
    ap : complex, ndarray, shape (`n`*(`n`+1)/2)
        If `fact` = 'F' or 'N', or if `fact` = 'E' and `equed` = 'N',
        `ap` is not modified.

        If `fact` = 'E' and `equed` = 'Y', `ap` is overwritten by D_S
        AD_S.

    afp : complex, ndarray, shape (`n`*(`n`+1)/2)
        If `fact` = 'N' or if `fact` = 'E' and `equed` = 'N', `afp`
        returns the triangular factor U or L from the Cholesky
        factorization A = U^HU or A = LL^H of the original matrix A.

        If `fact` = 'E' and `equed` = 'Y', `afp` returns the triangular
        factor U or L from the Cholesky factorization A = U^HU or A =
        LL^H of the equilibrated matrix A (see the description of `ap`
        for the form of the equilibrated matrix).

    equed : str, length 1
        If `fact` = 'F', `equed` is unchanged from entry.

        Otherwise, if no constraints are violated, `equed` specifies the
        form of the equilibration that was performed as specified above.

    s : float, ndarray, shape (`n`)
        If `fact` = 'F', `s` is unchanged from entry.

        Otherwise, if no constraints are violated and `equed` = 'Y', `s`
        contains the scale factors, D_S, for A; each element of `s` is
        positive.

    b : complex, ndarray, shape (`n`, `nrhs`)
        If `equed` = 'N', `b` is not modified.

        If `equed` = 'Y', `b` is overwritten by D_S B.

    x : complex, ndarray, shape (`n`, `nrhs`)
        If the function exits successfully or `errno` = `n` + 1, the n
        by r solution matrix X to the original system of equations. Note
        that the arrays A and B are modified on exit if `equed` = 'Y',
        and the solution to the equilibrated system is D_S^-1X.

    rcond : float
        If no constraints are violated, an estimate of the reciprocal
        condition number of the matrix A (after equilibration if that is
        performed), computed as `rcond` = 1.0/(||A||_1||A^-1||_1).

    ferr : float, ndarray, shape (`nrhs`)
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the forward error bound for each computed solution
        vector, such that ||x^_j-x_j||_infinity/||x_j||_infinity <=
        `ferr`[j-1] where x^_j is the jth column of the computed
        solution returned in the array `x` and x_j is the corresponding
        column of the exact solution X. The estimate is as reliable as
        the estimate for `rcond`, and is almost always a slight
        overestimate of the true error.

    berr : float, ndarray, shape (`nrhs`)
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the component-wise relative backward error of each
        computed solution vector x^_j (i.e., the smallest relative
        change in any element of A or B that makes x^_j an exact
        solution).

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `fact`.

            Constraint: `fact` = 'F', 'N' or 'E'.

        (`errno` -2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -4)
            On entry, error in parameter `nrhs`.

            Constraint: `nrhs` >= 0.

        (`errno` -7)
            On entry, error in parameter `equed`.

            Constraint: `equed` = 'N' or 'Y'.

        (`errno` -8)
            On entry, error in parameter `s`.

        (`errno` (i > 0) and (i<=`n`))
            The leading minor of order *<value>* of A is not positive
            definite, so the factorization could not be completed, and
            the solution has not been computed. `rcond` = 0.0 is
            returned.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` `n`+1)
            U (or L) is nonsingular, but `rcond` is less than machine
            precision, meaning that the matrix is singular to working
            precision. Nevertheless, the solution and error bounds are
            computed because there are a number of situations where the
            computed solution can be more accurate than the value of
            `rcond` would suggest.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    fun_name = 'naginterfaces.library.lapacklin.zppsvx'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_b, 'b'),),
    )
    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=True,
    )
    _afp = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'afp'},
        afp,
        copy=True,
    )
    _s = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 's'},
        s,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _nrhs = _b_utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineIntScalarType.to_py(_nrhs)
    _x = _b_utils._EngineComplex128ArrayType.empty_ndarray(
        (_n_py, _nrhs_py),
        sorder=_m_sorder,
    )
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    try:
        _return_scalars = _b_lapacklin.zppsvx(
            fact,
            uplo,
            _n,
            _nrhs,
            _ap,
            _afp,
            equed,
            _s,
            _b,
            _x,
            _ferr,
            _berr,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _ap_py = _ap
    _afp_py = _afp
    _equed_py = _return_scalars.equed
    _s_py = _s
    _b_py = _b
    _x_py = _x
    _rcond_py = _return_scalars.rcond
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['ap', 'afp', 'equed', 's', 'b', 'x', 'rcond', 'ferr', 'berr'],
    )(
        _ap_py,
        _afp_py,
        _equed_py,
        _s_py,
        _b_py,
        _x_py,
        _rcond_py,
        _ferr_py,
        _berr_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def zpptrf(uplo, n, ap):
    u"""
    Cholesky factorization of complex Hermitian positive definite
    matrix, packed storage.

    ``zpptrf`` computes the Cholesky factorization of a complex
    Hermitian positive definite matrix, using packed storage.

    For full information please refer to the NAG Library document for
    f07gr

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07grf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored and how A is to be factorized.

        `uplo` = 'U'

            The upper triangular part of A is stored and A is factorized
            as U^HU, where U is upper triangular.

        `uplo` = 'L'

            The lower triangular part of A is stored and A is factorized
            as LL^H, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    ap : complex, array-like, shape (`n`*(`n`+1)/2)
        The n by n Hermitian matrix A, packed by columns.

    Returns
    -------
    ap : complex, ndarray, shape (`n`*(`n`+1)/2)
        If no exception or warning is raised, the factor U or L from the
        Cholesky factorization A = U^HU or A = LL^H, in the same storage
        format as A.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` i > 0)
            The leading minor of order *<value>* is not positive
            definite and the factorization could not be completed. Hence
            A itself is not positive definite. This may indicate an
            error in forming the matrix A. To factorize a Hermitian
            matrix which is not positive definite, call ``zhptrf``
            instead.

    References
    ----------
    Demmel, J W, 1989, `On floating-point errors in Cholesky`, LAPACK
    Working Note No. 14, University of Tennessee, Knoxville,
    https://www.netlib.org/lapack/lawnspdf/lawn14.pdf

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.zpptrf'

    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=True,
    )
    _b_lapacklin.zpptrf(
        uplo,
        n,
        _ap,
    )
    _ap_py = _ap
    return _ap_py

def zpptrs(uplo, n, ap, b):
    u"""
    Solution of complex Hermitian positive definite system of linear
    equations, multiple right-hand sides, matrix already factorized by
    ``zpptrf``, packed storage.

    ``zpptrs`` solves a complex Hermitian positive definite system of
    linear equations with multiple right-hand sides,

        AX = B,

    where A has been factorized by ``zpptrf``, using packed storage.

    For full information please refer to the NAG Library document for
    f07gs

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07gsf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = U^HU, where U is upper triangular.

        `uplo` = 'L'

            A = LL^H, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    ap : complex, array-like, shape (`n`*(`n`+1)/2)
        The Cholesky factor of A stored in packed form, as returned by
        ``zpptrf``.

    b : complex, array-like, shape (`n`, nrhs)
        The n by r right-hand side matrix B.

    Returns
    -------
    b : complex, ndarray, shape (`n`, nrhs)
        The n by r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -3)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.zpptrs'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
    )
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=False,
    )
    _b_lapacklin.zpptrs(
        uplo,
        n,
        _nrhs,
        _ap,
        _b,
    )
    _b_py = _b
    return _b_py

def zppequ(uplo, n, ap):
    u"""
    Computes row and column scalings intended to equilibrate a complex
    Hermitian positive definite matrix and reduce its condition number,
    packed storage.

    ``zppequ`` computes a diagonal scaling matrix S intended to
    equilibrate a complex n by n Hermitian positive definite matrix A,
    stored in packed format, and reduce its condition number.

    For full information please refer to the NAG Library document for
    f07gt

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07gtf.html

    Parameters
    ----------
    uplo : str, length 1
        Indicates whether the upper or lower triangular part of A is
        stored in the array `ap`, as follows:

        `uplo` = 'U'

            The upper triangle of A is stored.

        `uplo` = 'L'

            The lower triangle of A is stored.

    n : int
        n, the order of the matrix A.

    ap : complex, array-like, shape (`n`*(`n`+1)/2)
        The n by n Hermitian matrix A, packed by columns.

        Only the elements of `ap` corresponding to the diagonal elements
        A are referenced.

    Returns
    -------
    s : float, ndarray, shape (`n`)
        If no exception or warning is raised, `s` contains the diagonal
        elements of the scaling matrix S.

    scond : float
        If no exception or warning is raised, `scond` contains the ratio
        of the smallest value of `s` to the largest value of `s`. If
        `scond` >= 0.1 and `amax` is neither too large nor too small, it
        is not worth scaling by S.

    amax : float
        max|a_{ij}|. If `amax` is very close to overflow or underflow,
        the matrix A should be scaled.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` i > 0)
            The *<value>*th diagonal element of A is not positive (and
            hence A cannot be positive definite).

    References
    ----------
    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    fun_name = 'naginterfaces.library.lapacklin.zppequ'

    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=False,
    )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _s = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _return_scalars = _b_lapacklin.zppequ(
        uplo,
        _n,
        _ap,
        _s,
    )
    _s_py = _s
    _scond_py = _return_scalars.scond
    _amax_py = _return_scalars.amax
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['s', 'scond', 'amax'],
    )(
        _s_py,
        _scond_py,
        _amax_py,
    )
    return _return_data

def zppcon(uplo, n, ap, anorm):
    u"""
    Estimate condition number of complex Hermitian positive definite
    matrix, matrix already factorized by ``zpptrf``, packed storage.

    ``zppcon`` estimates the condition number of a complex Hermitian
    positive definite matrix A, where A has been factorized by
    ``zpptrf``, using packed storage.

    For full information please refer to the NAG Library document for
    f07gu

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07guf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = U^HU, where U is upper triangular.

        `uplo` = 'L'

            A = LL^H, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    ap : complex, array-like, shape (`n`*(`n`+1)/2)
        The Cholesky factor of A stored in packed form, as returned by
        ``zpptrf``.

    anorm : float
        The 1-norm of the **original** matrix A, which may be computed
        by calling ``blas.zlanhp`` with its argument norm = '1'. `anorm`
        must be computed either **before** calling ``zpptrf`` or else
        from a **copy** of the original matrix A.

    Returns
    -------
    rcond : float
        An estimate of the reciprocal of the condition number of A.
        `rcond` is set to zero if exact singularity is detected or the
        estimate underflows. If `rcond` is less than machine precision,
        A is singular to working precision.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -4)
            On entry, error in parameter `anorm`.

            Constraint: `anorm` >= 0.0.

    References
    ----------
    Higham, N J, 1988, `FORTRAN codes for estimating the one-norm of a
    real or complex matrix, with applications to condition estimation`,
    ACM Trans. Math. Software (14), 381--396
    """

    fun_name = 'naginterfaces.library.lapacklin.zppcon'

    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=False,
    )
    _return_scalar = _b_lapacklin.zppcon(
        uplo,
        n,
        _ap,
        anorm,
    )
    _rcond_py = _return_scalar
    return _rcond_py

def zpprfs(uplo, n, ap, afp, b, x):
    u"""
    Refined solution with error bounds of complex Hermitian positive
    definite system of linear equations, multiple right-hand sides,
    packed storage.

    ``zpprfs`` returns error bounds for the solution of a complex
    Hermitian positive definite system of linear equations with multiple
    right-hand sides, AX = B, using packed storage.
    It improves the solution by iterative refinement, in order to reduce
    the backward error as much as possible.

    For full information please refer to the NAG Library document for
    f07gv

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07gvf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored and how A is to be factorized.

        `uplo` = 'U'

            The upper triangular part of A is stored and A is factorized
            as U^HU, where U is upper triangular.

        `uplo` = 'L'

            The lower triangular part of A is stored and A is factorized
            as LL^H, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    ap : complex, array-like, shape (`n`*(`n`+1)/2)
        The n by n original Hermitian positive definite matrix A as
        supplied to ``zpptrf``.

    afp : complex, array-like, shape (`n`*(`n`+1)/2)
        The Cholesky factor of A stored in packed form, as returned by
        ``zpptrf``.

    b : complex, array-like, shape (`n`, nrhs)
        The n by r right-hand side matrix B.

    x : complex, array-like, shape (`n`, nrhs)
        The n by r solution matrix X, as returned by ``zpptrs``.

    Returns
    -------
    x : complex, ndarray, shape (`n`, nrhs)
        The improved solution matrix X.

    ferr : float, ndarray, shape (nrhs)
        `ferr`[j-1] contains an estimated error bound for the jth
        solution vector, that is, the jth column of X, for j =
        1,2,...,r.

    berr : float, ndarray, shape (nrhs)
        `berr`[j-1] contains the component-wise backward error bound
        beta for the jth solution vector, that is, the jth column of X,
        for j = 1,2,...,r.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -3)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.zpprfs'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
    )
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=False,
    )
    _afp = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'afp'},
        afp,
        copy=False,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
    )
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _b_lapacklin.zpprfs(
        uplo,
        n,
        _nrhs,
        _ap,
        _afp,
        _b,
        _x,
        _ferr,
        _berr,
    )
    _x_py = _x
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'ferr', 'berr'],
    )(
        _x_py,
        _ferr_py,
        _berr_py,
    )
    return _return_data

def zpptri(uplo, n, ap):
    u"""
    Inverse of complex Hermitian positive definite matrix, matrix
    already factorized by ``zpptrf``, packed storage.

    ``zpptri`` computes the inverse of a complex Hermitian positive
    definite matrix A, where A has been factorized by ``zpptrf``, using
    packed storage.

    For full information please refer to the NAG Library document for
    f07gw

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07gwf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = U^HU, where U is upper triangular.

        `uplo` = 'L'

            A = LL^H, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    ap : complex, array-like, shape (`n`*(`n`+1)/2)
        The Cholesky factor of A stored in packed form, as returned by
        ``zpptrf``.

    Returns
    -------
    ap : complex, ndarray, shape (`n`*(`n`+1)/2)
        The factorization is overwritten by the n by n matrix A^-1.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Diagonal element *<value>* of the Cholesky factor is zero;
            the Cholesky factor is singular and the inverse of A cannot
            be computed.

    References
    ----------
    Du Croz, J J and Higham, N J, 1992, `Stability of methods for matrix
    inversion`, IMA J. Numer. Anal. (12), 1--19
    """

    fun_name = 'naginterfaces.library.lapacklin.zpptri'

    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=True,
    )
    try:
        _b_lapacklin.zpptri(
            uplo,
            n,
            _ap,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _ap_py = _ap
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['ap'],
    )(
        _ap_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _ap_py

def dpbsv(uplo, kd, ab, b):
    u"""
    Computes the solution to a real symmetric positive definite banded
    system of linear equations.

    ``dpbsv`` computes the solution to a real system of linear equations

        AX = B,

    where A is an n by n symmetric positive definite band matrix of
    bandwidth (2k_d+1) and X and B are n by r matrices.

    For full information please refer to the NAG Library document for
    f07ha

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07haf.html

    Parameters
    ----------
    uplo : str, length 1
        If `uplo` = 'U', the upper triangle of A is stored.

        If `uplo` = 'L', the lower triangle of A is stored.

    kd : int
        k_d, the number of superdiagonals of the matrix A if `uplo` =
        'U', or the number of subdiagonals if `uplo` = 'L'.

    ab : float, array-like, shape (`kd`+1, n)
        The upper or lower triangle of the symmetric band matrix A.

    b : float, array-like, shape (n, nrhs)
        The n by r right-hand side matrix B.

    Returns
    -------
    ab : float, ndarray, shape (`kd`+1, n)
        If no exception or warning is raised, the triangular factor U or
        L from the Cholesky factorization A = U^TU or A = LL^T of the
        band matrix A, in the same storage format as A.

    b : float, ndarray, shape (n, nrhs)
        If no exception or warning is raised, the n by r solution matrix
        X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter `kd`.

            Constraint: `kd` >= 0.

        (`errno` -4)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

        (`errno` i > 0)
            The leading minor of order *<value>* of A is not positive
            definite, so the factorization could not be completed, and
            the solution has not been computed.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.dpbsv'

    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nrhs_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _b_lapacklin.dpbsv(
        uplo,
        _n,
        kd,
        _nrhs,
        _ab,
        _b,
    )
    _ab_py = _ab
    _b_py = _b
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['ab', 'b'],
    )(
        _ab_py,
        _b_py,
    )
    return _return_data

def dpbsvx(fact, uplo, n, kd, nrhs, ab, afb, equed, s, b):
    u"""
    Uses the Cholesky factorization to compute the solution, error-bound
    and condition estimate for a real symmetric positive definite banded
    system of linear equations.

    ``dpbsvx`` uses the Cholesky factorization

        A = U^TU or A = LL^T

    to compute the solution to a real system of linear equations

        AX = B,

    where A is an n by n symmetric positive definite band matrix of
    bandwidth (2k_d+1) and X and B are n by r matrices.
    Error bounds on the solution and a condition estimate are also
    provided.

    For full information please refer to the NAG Library document for
    f07hb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07hbf.html

    Parameters
    ----------
    fact : str, length 1
        Specifies whether or not the factorized form of the matrix A is
        supplied on entry, and if not, whether the matrix A should be
        equilibrated before it is factorized.

        `fact` = 'F'

            `afb` contains the factorized form of A. If `equed` = 'Y',
            the matrix A has been equilibrated with scaling factors
            given by `s`. `ab` and `afb` will not be modified.

        `fact` = 'N'

            The matrix A will be copied to `afb` and factorized.

        `fact` = 'E'

            The matrix A will be equilibrated if necessary, then copied
            to `afb` and factorized.

    uplo : str, length 1
        If `uplo` = 'U', the upper triangle of A is stored.

        If `uplo` = 'L', the lower triangle of A is stored.

    n : int
        n, the number of linear equations, i.e., the order of the matrix
        A.

    kd : int
        k_d, the number of superdiagonals of the matrix A if `uplo` =
        'U', or the number of subdiagonals if `uplo` = 'L'.

    nrhs : int
        r, the number of right-hand sides, i.e., the number of columns
        of the matrix B.

    ab : float, array-like, shape (`kd`+1, `n`)
        The upper or lower triangle of the symmetric band matrix A,
        except if `fact` = 'F' and `equed` = 'Y', in which case `ab`
        must contain the equilibrated matrix D_S AD_S.

    afb : float, array-like, shape (`kd`+1, `n`)
        If `fact` = 'F', `afb` contains the triangular factor U or L
        from the Cholesky factorization A = U^TU or A = LL^T of the band
        matrix A, in the same storage format as A. If `equed` = 'Y',
        `afb` is the factorized form of the equilibrated matrix A.

    equed : str, length 1
        If `fact` = 'N' or 'E', `equed` need not be set.

        If `fact` = 'F', `equed` must specify the form of the
        equilibration that was performed as follows:

            if `equed` = 'N', no equilibration;

            if `equed` = 'Y', equilibration was performed, i.e., A has
            been replaced by D_S AD_S.

    s : float, array-like, shape (`n`)
        If `fact` = 'N' or 'E', `s` need not be set.

        If `fact` = 'F' and `equed` = 'Y', `s` must contain the scale
        factors, D_S, for A; each element of `s` must be positive.

    b : float, array-like, shape (`n`, `nrhs`)
        The n by r right-hand side matrix B.

    Returns
    -------
    ab : float, ndarray, shape (`kd`+1, `n`)
        If `fact` = 'E' and `equed` = 'Y', `ab` is overwritten by D_S
        AD_S.

    afb : float, ndarray, shape (`kd`+1, `n`)
        If `fact` = 'N', `afb` returns the triangular factor U or L from
        the Cholesky factorization A = U^TU or A = LL^T.

        If `fact` = 'E', `afb` returns the triangular factor U or L from
        the Cholesky factorization A = U^TU or A = LL^T of the
        equilibrated matrix A (see the description of `ab` for the form
        of the equilibrated matrix).

    equed : str, length 1
        If `fact` = 'F', `equed` is unchanged from entry.

        Otherwise, if no constraints are violated, `equed` specifies the
        form of the equilibration that was performed as specified above.

    s : float, ndarray, shape (`n`)
        If `fact` = 'F', `s` is unchanged from entry.

        Otherwise, if no constraints are violated and `equed` = 'Y', `s`
        contains the scale factors, D_S, for A; each element of `s` is
        positive.

    b : float, ndarray, shape (`n`, `nrhs`)
        If `equed` = 'N', `b` is not modified.

        If `equed` = 'Y', `b` is overwritten by D_S B.

    x : float, ndarray, shape (`n`, `nrhs`)
        If the function exits successfully or `errno` = `n` + 1, the n
        by r solution matrix X to the original system of equations. Note
        that the arrays A and B are modified on exit if `equed` = 'Y',
        and the solution to the equilibrated system is D_S^-1X.

    rcond : float
        If no constraints are violated, an estimate of the reciprocal
        condition number of the matrix A (after equilibration if that is
        performed), computed as `rcond` = 1.0/(||A||_1||A^-1||_1).

    ferr : float, ndarray, shape (`nrhs`)
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the forward error bound for each computed solution
        vector, such that ||x^_j-x_j||_infinity/||x_j||_infinity <=
        `ferr`[j-1] where x^_j is the jth column of the computed
        solution returned in the array `x` and x_j is the corresponding
        column of the exact solution X. The estimate is as reliable as
        the estimate for `rcond`, and is almost always a slight
        overestimate of the true error.

    berr : float, ndarray, shape (`nrhs`)
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the component-wise relative backward error of each
        computed solution vector x^_j (i.e., the smallest relative
        change in any element of A or B that makes x^_j an exact
        solution).

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `fact`.

            Constraint: `fact` = 'F', 'N' or 'E'.

        (`errno` -2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -4)
            On entry, error in parameter `kd`.

            Constraint: `kd` >= 0.

        (`errno` -5)
            On entry, error in parameter `nrhs`.

            Constraint: `nrhs` >= 0.

        (`errno` -10)
            On entry, error in parameter `equed`.

            Constraint: `equed` = 'N' or 'Y'.

        (`errno` -11)
            On entry, error in parameter `s`.

        (`errno` (i > 0) and (i<=`n`))
            The leading minor of order *<value>* of A is not positive
            definite, so the factorization could not be completed, and
            the solution has not been computed. `rcond` = 0.0 is
            returned.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` `n`+1)
            U (or L) is nonsingular, but `rcond` is less than machine
            precision, meaning that the matrix is singular to working
            precision. Nevertheless, the solution and error bounds are
            computed because there are a number of situations where the
            computed solution can be more accurate than the value of
            `rcond` would suggest.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    fun_name = 'naginterfaces.library.lapacklin.dpbsvx'

    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _afb = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'afb'},
        afb,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_ab, 'ab'), (_afb, 'afb'), (_b, 'b')),
    )
    _s = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 's'},
        s,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _nrhs = _b_utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineIntScalarType.to_py(_nrhs)
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py, _nrhs_py),
        sorder=_m_sorder,
    )
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    try:
        _return_scalars = _b_lapacklin.dpbsvx(
            fact,
            uplo,
            _n,
            kd,
            _nrhs,
            _ab,
            _afb,
            equed,
            _s,
            _b,
            _x,
            _ferr,
            _berr,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _ab_py = _ab
    _afb_py = _afb
    _equed_py = _return_scalars.equed
    _s_py = _s
    _b_py = _b
    _x_py = _x
    _rcond_py = _return_scalars.rcond
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['ab', 'afb', 'equed', 's', 'b', 'x', 'rcond', 'ferr', 'berr'],
    )(
        _ab_py,
        _afb_py,
        _equed_py,
        _s_py,
        _b_py,
        _x_py,
        _rcond_py,
        _ferr_py,
        _berr_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def dpbtrf(uplo, kd, ab):
    u"""
    Cholesky factorization of real symmetric positive definite band
    matrix.

    ``dpbtrf`` computes the Cholesky factorization of a real symmetric
    positive definite band matrix.

    For full information please refer to the NAG Library document for
    f07hd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07hdf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored and how A is to be factorized.

        `uplo` = 'U'

            The upper triangular part of A is stored and A is factorized
            as U^TU, where U is upper triangular.

        `uplo` = 'L'

            The lower triangular part of A is stored and A is factorized
            as LL^T, where L is lower triangular.

    kd : int
        k_d, the number of superdiagonals or subdiagonals of the matrix
        A.

    ab : float, array-like, shape (`kd`+1, n)
        The n by n symmetric positive definite band matrix A.

    Returns
    -------
    ab : float, ndarray, shape (`kd`+1, n)
        The upper or lower triangle of A is overwritten by the Cholesky
        factor U or L as specified by `uplo`, using the same storage
        format as described above.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter `kd`.

            Constraint: `kd` >= 0.

        (`errno` i > 0)
            The leading minor of order *<value>* is not positive
            definite and the factorization could not be completed. Hence
            A itself is not positive definite. This may indicate an
            error in forming the matrix A. There is no function
            specifically designed to factorize a symmetric band matrix
            which is not positive definite; the matrix must be treated
            either as a nonsymmetric band matrix, by calling ``dgbtrf``
            or as a full symmetric matrix, by calling ``dsytrf``.

    References
    ----------
    Demmel, J W, 1989, `On floating-point errors in Cholesky`, LAPACK
    Working Note No. 14, University of Tennessee, Knoxville,
    https://www.netlib.org/lapack/lawnspdf/lawn14.pdf

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.dpbtrf'

    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b_lapacklin.dpbtrf(
        uplo,
        _n,
        kd,
        _ab,
    )
    _ab_py = _ab
    return _ab_py

def dpbtrs(uplo, kd, ab, b):
    u"""
    Solution of real symmetric positive definite band system of linear
    equations, multiple right-hand sides, matrix already factorized by
    ``dpbtrf``.

    ``dpbtrs`` solves a real symmetric positive definite band system of
    linear equations with multiple right-hand sides,

        AX = B,

    where A has been factorized by ``dpbtrf``.

    For full information please refer to the NAG Library document for
    f07he

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07hef.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = U^TU, where U is upper triangular.

        `uplo` = 'L'

            A = LL^T, where L is lower triangular.

    kd : int
        k_d, the number of superdiagonals or subdiagonals of the matrix
        A.

    ab : float, array-like, shape (`kd`+1, n)
        The Cholesky factor of A, as returned by ``dpbtrf``.

    b : float, array-like, shape (n, nrhs)
        The n by r right-hand side matrix B.

    Returns
    -------
    b : float, ndarray, shape (n, nrhs)
        The n by r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter `kd`.

            Constraint: `kd` >= 0.

        (`errno` -4)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.dpbtrs'

    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nrhs_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _b_lapacklin.dpbtrs(
        uplo,
        _n,
        kd,
        _nrhs,
        _ab,
        _b,
    )
    _b_py = _b
    return _b_py

def dpbequ(uplo, n, kd, ab):
    u"""
    Computes row and column scalings intended to equilibrate a real
    symmetric positive definite banded matrix and reduce its condition
    number.

    ``dpbequ`` computes a diagonal scaling matrix S intended to
    equilibrate a real n by n symmetric positive definite band matrix A,
    with bandwidth (2k_d+1), and reduce its condition number.

    For full information please refer to the NAG Library document for
    f07hf

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07hff.html

    Parameters
    ----------
    uplo : str, length 1
        Indicates whether the upper or lower triangular part of A is
        stored in the array `ab`, as follows:

        `uplo` = 'U'

            The upper triangle of A is stored.

        `uplo` = 'L'

            The lower triangle of A is stored.

    n : int
        n, the order of the matrix A.

    kd : int
        k_d, the number of superdiagonals of the matrix A if `uplo` =
        'U', or the number of subdiagonals if `uplo` = 'L'.

    ab : float, array-like, shape (`kd`+1, `n`)
        The upper or lower triangle of the symmetric positive definite
        band matrix A whose scaling factors are to be computed.

        Only the elements of the array `ab` corresponding to the
        diagonal elements of A are referenced. (Row (k_d+1) of `ab` when
        `uplo` = 'U', row 1 of `ab` when `uplo` = 'L'.)

    Returns
    -------
    s : float, ndarray, shape (`n`)
        If no exception or warning is raised, `s` contains the diagonal
        elements of the scaling matrix S.

    scond : float
        If no exception or warning is raised, `scond` contains the ratio
        of the smallest value of `s` to the largest value of `s`. If
        `scond` >= 0.1 and `amax` is neither too large nor too small, it
        is not worth scaling by S.

    amax : float
        max|a_{ij}|. If `amax` is very close to overflow or underflow,
        the matrix A should be scaled.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -3)
            On entry, error in parameter `kd`.

            Constraint: `kd` >= 0.

        (`errno` i > 0)
            The *<value>*th diagonal element of A is not positive (and
            hence A cannot be positive definite).

    References
    ----------
    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    fun_name = 'naginterfaces.library.lapacklin.dpbequ'

    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _s = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _return_scalars = _b_lapacklin.dpbequ(
        uplo,
        _n,
        kd,
        _ab,
        _s,
    )
    _s_py = _s
    _scond_py = _return_scalars.scond
    _amax_py = _return_scalars.amax
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['s', 'scond', 'amax'],
    )(
        _s_py,
        _scond_py,
        _amax_py,
    )
    return _return_data

def dpbcon(uplo, kd, ab, anorm):
    u"""
    Estimate condition number of real symmetric positive definite band
    matrix, matrix already factorized by ``dpbtrf``.

    ``dpbcon`` estimates the condition number of a real symmetric
    positive definite band matrix A, where A has been factorized by
    ``dpbtrf``.

    For full information please refer to the NAG Library document for
    f07hg

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07hgf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = U^TU, where U is upper triangular.

        `uplo` = 'L'

            A = LL^T, where L is lower triangular.

    kd : int
        k_d, the number of superdiagonals or subdiagonals of the matrix
        A.

    ab : float, array-like, shape (`kd`+1, n)
        The Cholesky factor of A, as returned by ``dpbtrf``.

    anorm : float
        The 1-norm of the **original** matrix A, which may be computed
        by calling ``blas.dlansb`` with its argument norm = '1'. `anorm`
        must be computed either **before** calling ``dpbtrf`` or else
        from a **copy** of the original matrix A.

    Returns
    -------
    rcond : float
        An estimate of the reciprocal of the condition number of A.
        `rcond` is set to zero if exact singularity is detected or the
        estimate underflows. If `rcond` is less than machine precision,
        A is singular to working precision.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter `kd`.

            Constraint: `kd` >= 0.

        (`errno` -6)
            On entry, error in parameter `anorm`.

            Constraint: `anorm` >= 0.0.

    References
    ----------
    Higham, N J, 1988, `FORTRAN codes for estimating the one-norm of a
    real or complex matrix, with applications to condition estimation`,
    ACM Trans. Math. Software (14), 381--396
    """

    fun_name = 'naginterfaces.library.lapacklin.dpbcon'

    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _return_scalar = _b_lapacklin.dpbcon(
        uplo,
        _n,
        kd,
        _ab,
        anorm,
    )
    _rcond_py = _return_scalar
    return _rcond_py

def dpbrfs(uplo, kd, nrhs, ab, afb, b, x):
    u"""
    Refined solution with error bounds of real symmetric positive
    definite band system of linear equations, multiple right-hand sides.

    ``dpbrfs`` returns error bounds for the solution of a real symmetric
    positive definite band system of linear equations with multiple
    right-hand sides, AX = B.
    It improves the solution by iterative refinement, in order to reduce
    the backward error as much as possible.

    For full information please refer to the NAG Library document for
    f07hh

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07hhf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored and how A is to be factorized.

        `uplo` = 'U'

            The upper triangular part of A is stored and A is factorized
            as U^TU, where U is upper triangular.

        `uplo` = 'L'

            The lower triangular part of A is stored and A is factorized
            as LL^T, where L is lower triangular.

    kd : int
        k_d, the number of superdiagonals or subdiagonals of the matrix
        A.

    nrhs : int
        r, the number of right-hand sides.

    ab : float, array-like, shape (`kd`+1, n)
        The n by n original symmetric positive definite band matrix A as
        supplied to ``dpbtrf``.

    afb : float, array-like, shape (`kd`+1, n)
        The Cholesky factor of A, as returned by ``dpbtrf``.

    b : float, array-like, shape (n, `nrhs`)
        The n by r right-hand side matrix B.

    x : float, array-like, shape (n, `nrhs`)
        The n by r solution matrix X, as returned by ``dpbtrs``.

    Returns
    -------
    x : float, ndarray, shape (n, `nrhs`)
        The improved solution matrix X.

    ferr : float, ndarray, shape (`nrhs`)
        `ferr`[j-1] contains an estimated error bound for the jth
        solution vector, that is, the jth column of X, for j =
        1,2,...,r.

    berr : float, ndarray, shape (`nrhs`)
        `berr`[j-1] contains the component-wise backward error bound
        beta for the jth solution vector, that is, the jth column of X,
        for j = 1,2,...,r.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter `kd`.

            Constraint: `kd` >= 0.

        (`errno` -4)
            On entry, error in parameter `nrhs`.

            Constraint: `nrhs` >= 0.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.dpbrfs'

    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _afb = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'afb'},
        afb,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nrhs = _b_utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineIntScalarType.to_py(_nrhs)
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _b_lapacklin.dpbrfs(
        uplo,
        _n,
        kd,
        _nrhs,
        _ab,
        _afb,
        _b,
        _x,
        _ferr,
        _berr,
    )
    _x_py = _x
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'ferr', 'berr'],
    )(
        _x_py,
        _ferr_py,
        _berr_py,
    )
    return _return_data

def zpbsv(uplo, kd, ab, b):
    u"""
    Computes the solution to a complex Hermitian positive definite
    banded system of linear equations.

    ``zpbsv`` computes the solution to a complex system of linear
    equations

        AX = B,

    where A is an n by n Hermitian positive definite band matrix of
    bandwidth (2k_d+1) and X and B are n by r matrices.

    For full information please refer to the NAG Library document for
    f07hn

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07hnf.html

    Parameters
    ----------
    uplo : str, length 1
        If `uplo` = 'U', the upper triangle of A is stored.

        If `uplo` = 'L', the lower triangle of A is stored.

    kd : int
        k_d, the number of superdiagonals of the matrix A if `uplo` =
        'U', or the number of subdiagonals if `uplo` = 'L'.

    ab : complex, array-like, shape (`kd`+1, n)
        The upper or lower triangle of the Hermitian band matrix A.

    b : complex, array-like, shape (n, nrhs)
        The n by r right-hand side matrix B.

    Returns
    -------
    ab : complex, ndarray, shape (`kd`+1, n)
        If no exception or warning is raised, the triangular factor U or
        L from the Cholesky factorization A = U^HU or A = LL^H of the
        band matrix A, in the same storage format as A.

    b : complex, ndarray, shape (n, nrhs)
        If no exception or warning is raised, the n by r solution matrix
        X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter `kd`.

            Constraint: `kd` >= 0.

        (`errno` -4)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

        (`errno` i > 0)
            The leading minor of order *<value>* of A is not positive
            definite, so the factorization could not be completed, and
            the solution has not been computed.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.zpbsv'

    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
    )
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _b_lapacklin.zpbsv(
        uplo,
        _n,
        kd,
        _nrhs,
        _ab,
        _b,
    )
    _ab_py = _ab
    _b_py = _b
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['ab', 'b'],
    )(
        _ab_py,
        _b_py,
    )
    return _return_data

def zpbsvx(fact, uplo, n, kd, nrhs, ab, afb, equed, s, b):
    u"""
    Uses the Cholesky factorization to compute the solution, error-bound
    and condition estimate for a complex Hermitian positive definite
    banded system of linear equations.

    ``zpbsvx`` uses the Cholesky factorization

        A = U^HU or A = LL^H

    to compute the solution to a complex system of linear equations

        AX = B,

    where A is an n by n Hermitian positive definite band matrix of
    bandwidth (2k_d+1) and X and B are n by r matrices.
    Error bounds on the solution and a condition estimate are also
    provided.

    For full information please refer to the NAG Library document for
    f07hp

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07hpf.html

    Parameters
    ----------
    fact : str, length 1
        Specifies whether or not the factorized form of the matrix A is
        supplied on entry, and if not, whether the matrix A should be
        equilibrated before it is factorized.

        `fact` = 'F'

            `afb` contains the factorized form of A. If `equed` = 'Y',
            the matrix A has been equilibrated with scaling factors
            given by `s`. `ab` and `afb` will not be modified.

        `fact` = 'N'

            The matrix A will be copied to `afb` and factorized.

        `fact` = 'E'

            The matrix A will be equilibrated if necessary, then copied
            to `afb` and factorized.

    uplo : str, length 1
        If `uplo` = 'U', the upper triangle of A is stored.

        If `uplo` = 'L', the lower triangle of A is stored.

    n : int
        n, the number of linear equations, i.e., the order of the matrix
        A.

    kd : int
        k_d, the number of superdiagonals of the matrix A if `uplo` =
        'U', or the number of subdiagonals if `uplo` = 'L'.

    nrhs : int
        r, the number of right-hand sides, i.e., the number of columns
        of the matrix B.

    ab : complex, array-like, shape (`kd`+1, `n`)
        The upper or lower triangle of the Hermitian band matrix A,
        except if `fact` = 'F' and `equed` = 'Y', in which case `ab`
        must contain the equilibrated matrix D_S AD_S.

    afb : complex, array-like, shape (`kd`+1, `n`)
        If `fact` = 'F', `afb` contains the triangular factor U or L
        from the Cholesky factorization A = U^HU or A = LL^H of the band
        matrix A, in the same storage format as A. If `equed` = 'Y',
        `afb` is the factorized form of the equilibrated matrix A.

    equed : str, length 1
        If `fact` = 'N' or 'E', `equed` need not be set.

        If `fact` = 'F', `equed` must specify the form of the
        equilibration that was performed as follows:

            if `equed` = 'N', no equilibration;

            if `equed` = 'Y', equilibration was performed, i.e., A has
            been replaced by D_S AD_S.

    s : float, array-like, shape (`n`)
        If `fact` = 'N' or 'E', `s` need not be set.

        If `fact` = 'F' and `equed` = 'Y', `s` must contain the scale
        factors, D_S, for A; each element of `s` must be positive.

    b : complex, array-like, shape (`n`, `nrhs`)
        The n by r right-hand side matrix B.

    Returns
    -------
    ab : complex, ndarray, shape (`kd`+1, `n`)
        If `fact` = 'E' and `equed` = 'Y', `ab` is overwritten by D_S
        AD_S.

    afb : complex, ndarray, shape (`kd`+1, `n`)
        If `fact` = 'N', `afb` returns the triangular factor U or L from
        the Cholesky factorization A = U^HU or A = LL^H.

        If `fact` = 'E', `afb` returns the triangular factor U or L from
        the Cholesky factorization A = U^HU or A = LL^H of the
        equilibrated matrix A (see the description of `ab` for the form
        of the equilibrated matrix).

    equed : str, length 1
        If `fact` = 'F', `equed` is unchanged from entry.

        Otherwise, if no constraints are violated, `equed` specifies the
        form of the equilibration that was performed as specified above.

    s : float, ndarray, shape (`n`)
        If `fact` = 'F', `s` is unchanged from entry.

        Otherwise, if no constraints are violated and `equed` = 'Y', `s`
        contains the scale factors, D_S, for A; each element of `s` is
        positive.

    b : complex, ndarray, shape (`n`, `nrhs`)
        If `equed` = 'N', `b` is not modified.

        If `equed` = 'Y', `b` is overwritten by D_S B.

    x : complex, ndarray, shape (`n`, `nrhs`)
        If the function exits successfully or `errno` = `n` + 1, the n
        by r solution matrix X to the original system of equations. Note
        that the arrays A and B are modified on exit if `equed` = 'Y',
        and the solution to the equilibrated system is D_S^-1X.

    rcond : float
        If no constraints are violated, an estimate of the reciprocal
        condition number of the matrix A (after equilibration if that is
        performed), computed as `rcond` = 1.0/(||A||_1||A^-1||_1).

    ferr : float, ndarray, shape (`nrhs`)
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the forward error bound for each computed solution
        vector, such that ||x^_j-x_j||_infinity/||x_j||_infinity <=
        `ferr`[j-1] where x^_j is the jth column of the computed
        solution returned in the array `x` and x_j is the corresponding
        column of the exact solution X. The estimate is as reliable as
        the estimate for `rcond`, and is almost always a slight
        overestimate of the true error.

    berr : float, ndarray, shape (`nrhs`)
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the component-wise relative backward error of each
        computed solution vector x^_j (i.e., the smallest relative
        change in any element of A or B that makes x^_j an exact
        solution).

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `fact`.

            Constraint: `fact` = 'F', 'N' or 'E'.

        (`errno` -2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -4)
            On entry, error in parameter `kd`.

            Constraint: `kd` >= 0.

        (`errno` -5)
            On entry, error in parameter `nrhs`.

            Constraint: `nrhs` >= 0.

        (`errno` -10)
            On entry, error in parameter `equed`.

            Constraint: `equed` = 'N' or 'Y'.

        (`errno` -11)
            On entry, error in parameter `s`.

        (`errno` (i > 0) and (i<=`n`))
            The leading minor of order *<value>* of A is not positive
            definite, so the factorization could not be completed, and
            the solution has not been computed. `rcond` = 0.0 is
            returned.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` `n`+1)
            U (or L) is nonsingular, but `rcond` is less than machine
            precision, meaning that the matrix is singular to working
            precision. Nevertheless, the solution and error bounds are
            computed because there are a number of situations where the
            computed solution can be more accurate than the value of
            `rcond` would suggest.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    fun_name = 'naginterfaces.library.lapacklin.zpbsvx'

    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=True,
    )
    _afb = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'afb'},
        afb,
        copy=True,
    )
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_ab, 'ab'), (_afb, 'afb'), (_b, 'b')),
    )
    _s = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 's'},
        s,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _nrhs = _b_utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineIntScalarType.to_py(_nrhs)
    _x = _b_utils._EngineComplex128ArrayType.empty_ndarray(
        (_n_py, _nrhs_py),
        sorder=_m_sorder,
    )
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    try:
        _return_scalars = _b_lapacklin.zpbsvx(
            fact,
            uplo,
            _n,
            kd,
            _nrhs,
            _ab,
            _afb,
            equed,
            _s,
            _b,
            _x,
            _ferr,
            _berr,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _ab_py = _ab
    _afb_py = _afb
    _equed_py = _return_scalars.equed
    _s_py = _s
    _b_py = _b
    _x_py = _x
    _rcond_py = _return_scalars.rcond
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['ab', 'afb', 'equed', 's', 'b', 'x', 'rcond', 'ferr', 'berr'],
    )(
        _ab_py,
        _afb_py,
        _equed_py,
        _s_py,
        _b_py,
        _x_py,
        _rcond_py,
        _ferr_py,
        _berr_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def zpbtrf(uplo, kd, ab):
    u"""
    Cholesky factorization of complex Hermitian positive definite band
    matrix.

    ``zpbtrf`` computes the Cholesky factorization of a complex
    Hermitian positive definite band matrix.

    For full information please refer to the NAG Library document for
    f07hr

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07hrf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored and how A is to be factorized.

        `uplo` = 'U'

            The upper triangular part of A is stored and A is factorized
            as U^HU, where U is upper triangular.

        `uplo` = 'L'

            The lower triangular part of A is stored and A is factorized
            as LL^H, where L is lower triangular.

    kd : int
        k_d, the number of superdiagonals or subdiagonals of the matrix
        A.

    ab : complex, array-like, shape (`kd`+1, n)
        The n by n Hermitian positive definite band matrix A.

    Returns
    -------
    ab : complex, ndarray, shape (`kd`+1, n)
        The upper or lower triangle of A is overwritten by the Cholesky
        factor U or L as specified by `uplo`, using the same storage
        format as described above.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter `kd`.

            Constraint: `kd` >= 0.

        (`errno` i > 0)
            The leading minor of order *<value>* is not positive
            definite and the factorization could not be completed. Hence
            A itself is not positive definite. This may indicate an
            error in forming the matrix A. There is no function
            specifically designed to factorize a Hermitian band matrix
            which is not positive definite; the matrix must be treated
            either as a nonsymmetric band matrix, by calling ``zgbtrf``
            or as a full Hermitian matrix, by calling ``zhetrf``.

    References
    ----------
    Demmel, J W, 1989, `On floating-point errors in Cholesky`, LAPACK
    Working Note No. 14, University of Tennessee, Knoxville,
    https://www.netlib.org/lapack/lawnspdf/lawn14.pdf

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.zpbtrf'

    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b_lapacklin.zpbtrf(
        uplo,
        _n,
        kd,
        _ab,
    )
    _ab_py = _ab
    return _ab_py

def zpbtrs(uplo, kd, ab, b):
    u"""
    Solution of complex Hermitian positive definite band system of
    linear equations, multiple right-hand sides, matrix already
    factorized by ``zpbtrf``.

    ``zpbtrs`` solves a complex Hermitian positive definite band system
    of linear equations with multiple right-hand sides,

        AX = B,

    where A has been factorized by ``zpbtrf``.

    For full information please refer to the NAG Library document for
    f07hs

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07hsf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = U^HU, where U is upper triangular.

        `uplo` = 'L'

            A = LL^H, where L is lower triangular.

    kd : int
        k_d, the number of superdiagonals or subdiagonals of the matrix
        A.

    ab : complex, array-like, shape (`kd`+1, n)
        The Cholesky factor of A, as returned by ``zpbtrf``.

    b : complex, array-like, shape (n, nrhs)
        The n by r right-hand side matrix B.

    Returns
    -------
    b : complex, ndarray, shape (n, nrhs)
        The n by r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter `kd`.

            Constraint: `kd` >= 0.

        (`errno` -4)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.zpbtrs'

    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
    )
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _b_lapacklin.zpbtrs(
        uplo,
        _n,
        kd,
        _nrhs,
        _ab,
        _b,
    )
    _b_py = _b
    return _b_py

def zpbequ(uplo, n, kd, ab):
    u"""
    Computes row and column scalings intended to equilibrate a complex
    Hermitian positive definite banded matrix and reduce its condition
    number.

    ``zpbequ`` computes a diagonal scaling matrix S intended to
    equilibrate a complex n by n Hermitian positive definite band matrix
    A, with bandwidth (2k_d+1), and reduce its condition number.

    For full information please refer to the NAG Library document for
    f07ht

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07htf.html

    Parameters
    ----------
    uplo : str, length 1
        Indicates whether the upper or lower triangular part of A is
        stored in the array `ab`, as follows:

        `uplo` = 'U'

            The upper triangle of A is stored.

        `uplo` = 'L'

            The lower triangle of A is stored.

    n : int
        n, the order of the matrix A.

    kd : int
        k_d, the number of superdiagonals of the matrix A if `uplo` =
        'U', or the number of subdiagonals if `uplo` = 'L'.

    ab : complex, array-like, shape (`kd`+1, `n`)
        The upper or lower triangle of the Hermitian positive definite
        band matrix A whose scaling factors are to be computed.

        Only the elements of the array `ab` corresponding to the
        diagonal elements of A are referenced. (Row (k_d+1) of `ab` when
        `uplo` = 'U', row 1 of `ab` when `uplo` = 'L'.)

    Returns
    -------
    s : float, ndarray, shape (`n`)
        If no exception or warning is raised, `s` contains the diagonal
        elements of the scaling matrix S.

    scond : float
        If no exception or warning is raised, `scond` contains the ratio
        of the smallest value of `s` to the largest value of `s`. If
        `scond` >= 0.1 and `amax` is neither too large nor too small, it
        is not worth scaling by S.

    amax : float
        max|a_{ij}|. If `amax` is very close to overflow or underflow,
        the matrix A should be scaled.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -3)
            On entry, error in parameter `kd`.

            Constraint: `kd` >= 0.

        (`errno` i > 0)
            The *<value>*th diagonal element of A is not positive (and
            hence A cannot be positive definite).

    References
    ----------
    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    fun_name = 'naginterfaces.library.lapacklin.zpbequ'

    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=False,
    )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _s = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _return_scalars = _b_lapacklin.zpbequ(
        uplo,
        _n,
        kd,
        _ab,
        _s,
    )
    _s_py = _s
    _scond_py = _return_scalars.scond
    _amax_py = _return_scalars.amax
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['s', 'scond', 'amax'],
    )(
        _s_py,
        _scond_py,
        _amax_py,
    )
    return _return_data

def zpbcon(uplo, kd, ab, anorm):
    u"""
    Estimate condition number of complex Hermitian positive definite
    band matrix, matrix already factorized by ``zpbtrf``.

    ``zpbcon`` estimates the condition number of a complex Hermitian
    positive definite band matrix A, where A has been factorized by
    ``zpbtrf``.

    For full information please refer to the NAG Library document for
    f07hu

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07huf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = U^HU, where U is upper triangular.

        `uplo` = 'L'

            A = LL^H, where L is lower triangular.

    kd : int
        k_d, the number of superdiagonals or subdiagonals of the matrix
        A.

    ab : complex, array-like, shape (`kd`+1, n)
        The Cholesky factor of A, as returned by ``zpbtrf``.

    anorm : float
        The 1-norm of the **original** matrix A, which may be computed
        by calling ``blas.zlanhb`` with its argument norm = '1'. `anorm`
        must be computed either **before** calling ``zpbtrf`` or else
        from a **copy** of the original matrix A.

    Returns
    -------
    rcond : float
        An estimate of the reciprocal of the condition number of A.
        `rcond` is set to zero if exact singularity is detected or the
        estimate underflows. If `rcond` is less than machine precision,
        A is singular to working precision.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter `kd`.

            Constraint: `kd` >= 0.

        (`errno` -6)
            On entry, error in parameter `anorm`.

            Constraint: `anorm` >= 0.0.

    References
    ----------
    Higham, N J, 1988, `FORTRAN codes for estimating the one-norm of a
    real or complex matrix, with applications to condition estimation`,
    ACM Trans. Math. Software (14), 381--396
    """

    fun_name = 'naginterfaces.library.lapacklin.zpbcon'

    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _return_scalar = _b_lapacklin.zpbcon(
        uplo,
        _n,
        kd,
        _ab,
        anorm,
    )
    _rcond_py = _return_scalar
    return _rcond_py

def zpbrfs(uplo, kd, nrhs, ab, afb, b, x):
    u"""
    Refined solution with error bounds of complex Hermitian positive
    definite band system of linear equations, multiple right-hand sides.

    ``zpbrfs`` returns error bounds for the solution of a complex
    Hermitian positive definite band system of linear equations with
    multiple right-hand sides, AX = B.
    It improves the solution by iterative refinement, in order to reduce
    the backward error as much as possible.

    For full information please refer to the NAG Library document for
    f07hv

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07hvf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored and how A is to be factorized.

        `uplo` = 'U'

            The upper triangular part of A is stored and A is factorized
            as U^HU, where U is upper triangular.

        `uplo` = 'L'

            The lower triangular part of A is stored and A is factorized
            as LL^H, where L is lower triangular.

    kd : int
        k_d, the number of superdiagonals or subdiagonals of the matrix
        A.

    nrhs : int
        r, the number of right-hand sides.

    ab : complex, array-like, shape (`kd`+1, n)
        The n by n original Hermitian positive definite band matrix A as
        supplied to ``zpbtrf``.

    afb : complex, array-like, shape (`kd`+1, n)
        The Cholesky factor of A, as returned by ``zpbtrf``.

    b : complex, array-like, shape (n, `nrhs`)
        The n by r right-hand side matrix B.

    x : complex, array-like, shape (n, `nrhs`)
        The n by r solution matrix X, as returned by ``zpbtrs``.

    Returns
    -------
    x : complex, ndarray, shape (n, `nrhs`)
        The improved solution matrix X.

    ferr : float, ndarray, shape (`nrhs`)
        `ferr`[j-1] contains an estimated error bound for the jth
        solution vector, that is, the jth column of X, for j =
        1,2,...,r.

    berr : float, ndarray, shape (`nrhs`)
        `berr`[j-1] contains the component-wise backward error bound
        beta for the jth solution vector, that is, the jth column of X,
        for j = 1,2,...,r.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter `kd`.

            Constraint: `kd` >= 0.

        (`errno` -4)
            On entry, error in parameter `nrhs`.

            Constraint: `nrhs` >= 0.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.zpbrfs'

    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _afb = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'afb'},
        afb,
        copy=False,
    )
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
    )
    _nrhs = _b_utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineIntScalarType.to_py(_nrhs)
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _b_lapacklin.zpbrfs(
        uplo,
        _n,
        kd,
        _nrhs,
        _ab,
        _afb,
        _b,
        _x,
        _ferr,
        _berr,
    )
    _x_py = _x
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'ferr', 'berr'],
    )(
        _x_py,
        _ferr_py,
        _berr_py,
    )
    return _return_data

def dptsv(d, e, b):
    u"""
    Computes the solution to a real symmetric positive definite
    tridiagonal system of linear equations.

    ``dptsv`` computes the solution to a real system of linear equations

        AX = B,

    where A is an n by n symmetric positive definite tridiagonal matrix,
    and X and B are n by r matrices.

    For full information please refer to the NAG Library document for
    f07ja

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07jaf.html

    Parameters
    ----------
    d : float, array-like, shape (n)
        The n diagonal elements of the tridiagonal matrix A.

    e : float, array-like, shape (n-1)
        The (n-1) subdiagonal elements of the tridiagonal matrix A.

    b : float, array-like, shape (n, nrhs)
        The n by r right-hand side matrix B.

    Returns
    -------
    d : float, ndarray, shape (n)
        The n diagonal elements of the diagonal matrix D from the
        factorization A = LDL^T.

    e : float, ndarray, shape (n-1)
        The (n-1) subdiagonal elements of the unit bidiagonal factor L
        from the LDL^T factorization of A. (`e` can also be regarded as
        the superdiagonal of the unit bidiagonal factor U from the U^TDU
        factorization of A.)

    b : float, ndarray, shape (n, nrhs)
        If no exception or warning is raised, the n by r solution matrix
        X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -2)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

        (`errno` i > 0)
            The leading minor of order *<value>* is not positive
            definite, and the solution has not been computed. The
            factorization has not been completed unless n = *<value>*.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.dptsv'

    _d = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'd'},
        d,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_d, {'fun_name': fun_name, 'entity': 'd'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nrhs_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _e = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'e'},
        e,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b_lapacklin.dptsv(
        _n,
        _nrhs,
        _d,
        _e,
        _b,
    )
    _d_py = _d
    _e_py = _e
    _b_py = _b
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['d', 'e', 'b'],
    )(
        _d_py,
        _e_py,
        _b_py,
    )
    return _return_data

def dptsvx(fact, nrhs, d, e, df, ef, b):
    u"""
    Uses the LDL^T factorization to compute the solution, error-bound
    and condition estimate for a real symmetric positive definite
    tridiagonal system of linear equations.

    ``dptsvx`` uses the factorization

        A = LDL^T

    to compute the solution to a real system of linear equations

        AX = B,

    where A is an n by n symmetric positive definite tridiagonal matrix
    and X and B are n by r matrices.
    Error bounds on the solution and a condition estimate are also
    provided.

    For full information please refer to the NAG Library document for
    f07jb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07jbf.html

    Parameters
    ----------
    fact : str, length 1
        Specifies whether or not the factorized form of the matrix A has
        been supplied.

        `fact` = 'F'

            `df` and `ef` contain the factorized form of the matrix A.
            `df` and `ef` will not be modified.

        `fact` = 'N'

            The matrix A will be copied to `df` and `ef` and factorized.

    nrhs : int
        r, the number of right-hand sides, i.e., the number of columns
        of the matrix B.

    d : float, array-like, shape (n)
        The n diagonal elements of the tridiagonal matrix A.

    e : float, array-like, shape (n-1)
        The (n-1) subdiagonal elements of the tridiagonal matrix A.

    df : float, array-like, shape (n)
        If `fact` = 'F', `df` must contain the n diagonal elements of
        the diagonal matrix D from the LDL^T factorization of A.

    ef : float, array-like, shape (n-1)
        If `fact` = 'F', `ef` must contain the (n-1) subdiagonal
        elements of the unit bidiagonal factor L from the LDL^T
        factorization of A.

    b : float, array-like, shape (n, `nrhs`)
        The n by r right-hand side matrix B.

    Returns
    -------
    df : float, ndarray, shape (n)
        If `fact` = 'N', `df` contains the n diagonal elements of the
        diagonal matrix D from the LDL^T factorization of A.

    ef : float, ndarray, shape (n-1)
        If `fact` = 'N', `ef` contains the (n-1) subdiagonal elements of
        the unit bidiagonal factor L from the LDL^T factorization of A.

    x : float, ndarray, shape (n, `nrhs`)
        If the function exits successfully or `errno` = `n` + 1, the n
        by r solution matrix X.

    rcond : float
        The reciprocal condition number of the matrix A. If `rcond` is
        less than the machine precision (in particular, if `rcond` =
        0.0), the matrix is singular to working precision. This
        condition is indicated by a return code of `errno` = `n` + 1.

    ferr : float, ndarray, shape (`nrhs`)
        The forward error bound for each solution vector x^_j (the jth
        column of the solution matrix X). If x_j is the true solution
        corresponding to x^_j, `ferr`[j-1] is an estimated upper bound
        for the magnitude of the largest element in (x^_j-x_j) divided
        by the magnitude of the largest element in x^_j.

    berr : float, ndarray, shape (`nrhs`)
        The component-wise relative backward error of each solution
        vector x^_j (i.e., the smallest relative change in any element
        of A or B that makes x^_j an exact solution).

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `fact`.

            Constraint: `fact` = 'F' or 'N'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter `nrhs`.

            Constraint: `nrhs` >= 0.

        (`errno` (i > 0) and (i<=n))
            The leading minor of order *<value>* of A is not positive
            definite, so the factorization could not be completed, and
            the solution has not been computed. `rcond` = 0.0 is
            returned.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` n+1)
            D is nonsingular, but `rcond` is less than machine
            precision, meaning that the matrix is singular to working
            precision. Nevertheless, the solution and error bounds are
            computed because there are a number of situations where the
            computed solution can be more accurate than the value of
            `rcond` would suggest.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    fun_name = 'naginterfaces.library.lapacklin.dptsvx'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_b, 'b'),),
    )
    _d = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'd'},
        d,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_d, {'fun_name': fun_name, 'entity': 'd'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _e = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'e'},
        e,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _df = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'df'},
        df,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ef = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ef'},
        ef,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nrhs = _b_utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineIntScalarType.to_py(_nrhs)
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py, _nrhs_py),
        sorder=_m_sorder,
    )
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    try:
        _return_scalar = _b_lapacklin.dptsvx(
            fact,
            _n,
            _nrhs,
            _d,
            _e,
            _df,
            _ef,
            _b,
            _x,
            _ferr,
            _berr,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _df_py = _df
    _ef_py = _ef
    _x_py = _x
    _rcond_py = _return_scalar
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['df', 'ef', 'x', 'rcond', 'ferr', 'berr'],
    )(
        _df_py,
        _ef_py,
        _x_py,
        _rcond_py,
        _ferr_py,
        _berr_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def dpttrf(d, e):
    u"""
    Computes the LDL^T factorization of a real symmetric positive
    definite tridiagonal matrix.

    ``dpttrf`` computes the modified Cholesky factorization of a real n
    by n symmetric positive definite tridiagonal matrix A.

    For full information please refer to the NAG Library document for
    f07jd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07jdf.html

    Parameters
    ----------
    d : float, array-like, shape (n)
        Must contain the n diagonal elements of the matrix A.

    e : float, array-like, shape (n-1)
        Must contain the (n-1) subdiagonal elements of the matrix A.

    Returns
    -------
    d : float, ndarray, shape (n)
        Is overwritten by the n diagonal elements of the diagonal matrix
        D from the LDL^T factorization of A.

    e : float, ndarray, shape (n-1)
        Is overwritten by the (n-1) subdiagonal elements of the lower
        bidiagonal matrix L. (`e` can also be regarded as containing the
        (n-1) superdiagonal elements of the upper bidiagonal matrix U.)

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` (i > 0) and (i < n))
            The leading minor of order *<value>* is not positive
            definite, the factorization could not be completed.

        (`errno` (i > 0) and (n))
            The leading minor of order n is not positive definite, the
            factorization was completed, but `d`[n-1] <= 0.
    """

    fun_name = 'naginterfaces.library.lapacklin.dpttrf'

    _d = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'd'},
        d,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_d, {'fun_name': fun_name, 'entity': 'd'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _e = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'e'},
        e,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b_lapacklin.dpttrf(
        _n,
        _d,
        _e,
    )
    _d_py = _d
    _e_py = _e
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['d', 'e'],
    )(
        _d_py,
        _e_py,
    )
    return _return_data

def dpttrs(d, e, b):
    u"""
    Solves a real symmetric positive definite tridiagonal system using
    the LDL^T factorization computed by ``dpttrf``.

    ``dpttrs`` computes the solution to a real system of linear
    equations AX = B, where A is an n by n symmetric positive definite
    tridiagonal matrix and X and B are n by r matrices, using the LDL^T
    factorization returned by ``dpttrf``.

    For full information please refer to the NAG Library document for
    f07je

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07jef.html

    Parameters
    ----------
    d : float, array-like, shape (n)
        Must contain the n diagonal elements of the diagonal matrix D
        from the LDL^T factorization of A.

    e : float, array-like, shape (n-1)
        Must contain the (n-1) subdiagonal elements of the unit lower
        bidiagonal matrix L. (`e` can also be regarded as the
        superdiagonal of the unit upper bidiagonal matrix U from the
        U^TDU factorization of A.)

    b : float, array-like, shape (n, nrhs)
        The n by r matrix of right-hand sides B.

    Returns
    -------
    b : float, ndarray, shape (n, nrhs)
        The n by r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -2)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug
    """

    fun_name = 'naginterfaces.library.lapacklin.dpttrs'

    _d = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'd'},
        d,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_d, {'fun_name': fun_name, 'entity': 'd'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nrhs_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _e = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'e'},
        e,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b_lapacklin.dpttrs(
        _n,
        _nrhs,
        _d,
        _e,
        _b,
    )
    _b_py = _b
    return _b_py

def dptcon(n, d, e, anorm):
    u"""
    Computes the reciprocal of the condition number of a real symmetric
    positive definite tridiagonal system using the LDL^T factorization
    computed by ``dpttrf``.

    ``dptcon`` computes the reciprocal condition number of a real n by n
    symmetric positive definite tridiagonal matrix A, using the LDL^T
    factorization returned by ``dpttrf``.

    For full information please refer to the NAG Library document for
    f07jg

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07jgf.html

    Parameters
    ----------
    n : int
        n, the order of the matrix A.

    d : float, array-like, shape (`n`)
        Must contain the n diagonal elements of the diagonal matrix D
        from the LDL^T factorization of A.

    e : float, array-like, shape (`n`-1)
        Must contain the (n-1) subdiagonal elements of the unit lower
        bidiagonal matrix L. (`e` can also be regarded as the
        superdiagonal of the unit upper bidiagonal matrix U from the
        U^TDU factorization of A.)

    anorm : float
        The 1-norm of the **original** matrix A, which may be computed
        by calling ``blas.dlanst`` with its argument norm = '1'. `anorm`
        must be computed either **before** calling ``dpttrf`` or else
        from a **copy** of the original matrix A.

    Returns
    -------
    rcond : float
        The reciprocal condition number, 1/kappa_1(A) =
        1/(||A||_1||A^-1||_1).

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -4)
            On entry, error in parameter `anorm`.

            Constraint: `anorm` >= 0.0.

    References
    ----------
    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    fun_name = 'naginterfaces.library.lapacklin.dptcon'

    _d = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'd'},
        d,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _e = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'e'},
        e,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _return_scalar = _b_lapacklin.dptcon(
        n,
        _d,
        _e,
        anorm,
    )
    _rcond_py = _return_scalar
    return _rcond_py

def dptrfs(nrhs, d, e, df, ef, b, x):
    u"""
    Refined solution with error bounds of real symmetric positive
    definite tridiagonal system of linear equations, multiple right-hand
    sides.

    ``dptrfs`` computes error bounds and refines the solution to a real
    system of linear equations AX = B, where A is an n by n symmetric
    positive definite tridiagonal matrix and X and B are n by r
    matrices, using the modified Cholesky factorization returned by
    ``dpttrf`` and an initial solution returned by ``dpttrs``.
    Iterative refinement is used to reduce the backward error as much as
    possible.

    For full information please refer to the NAG Library document for
    f07jh

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07jhf.html

    Parameters
    ----------
    nrhs : int
        r, the number of right-hand sides, i.e., the number of columns
        of the matrix B.

    d : float, array-like, shape (n)
        Must contain the n diagonal elements of the matrix of A.

    e : float, array-like, shape (n-1)
        Must contain the (n-1) subdiagonal elements of the matrix A.

    df : float, array-like, shape (n)
        Must contain the n diagonal elements of the diagonal matrix D
        from the LDL^T factorization of A.

    ef : float, array-like, shape (n)
        Must contain the (n-1) subdiagonal elements of the unit
        bidiagonal matrix L from the LDL^T factorization of A.

    b : float, array-like, shape (n, `nrhs`)
        The n by r matrix of right-hand sides B.

    x : float, array-like, shape (n, `nrhs`)
        The n by r initial solution matrix X.

    Returns
    -------
    x : float, ndarray, shape (n, `nrhs`)
        The n by r refined solution matrix X.

    ferr : float, ndarray, shape (`nrhs`)
        Estimate of the forward error bound for each computed solution
        vector, such that ||x^_j-x_j||_infinity/||x^_j||_infinity <=
        `ferr`[j-1], where x^_j is the jth column of the computed
        solution returned in the array `x` and x_j is the corresponding
        column of the exact solution X. The estimate is almost always a
        slight overestimate of the true error.

    berr : float, ndarray, shape (`nrhs`)
        Estimate of the component-wise relative backward error of each
        computed solution vector x^_j (i.e., the smallest relative
        change in any element of A or B that makes x^_j an exact
        solution).

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -2)
            On entry, error in parameter `nrhs`.

            Constraint: `nrhs` >= 0.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug
    """

    fun_name = 'naginterfaces.library.lapacklin.dptrfs'

    _d = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'd'},
        d,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_d, {'fun_name': fun_name, 'entity': 'd'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _e = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'e'},
        e,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _df = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'df'},
        df,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ef = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ef'},
        ef,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nrhs = _b_utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineIntScalarType.to_py(_nrhs)
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _b_lapacklin.dptrfs(
        _n,
        _nrhs,
        _d,
        _e,
        _df,
        _ef,
        _b,
        _x,
        _ferr,
        _berr,
    )
    _x_py = _x
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'ferr', 'berr'],
    )(
        _x_py,
        _ferr_py,
        _berr_py,
    )
    return _return_data

def zptsv(d, e, b):
    u"""
    Computes the solution to a complex Hermitian positive definite
    tridiagonal system of linear equations.

    ``zptsv`` computes the solution to a complex system of linear
    equations

        AX = B,

    where A is an n by n Hermitian positive definite tridiagonal matrix,
    and X and B are n by r matrices.

    For full information please refer to the NAG Library document for
    f07jn

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07jnf.html

    Parameters
    ----------
    d : float, array-like, shape (n)
        The n diagonal elements of the tridiagonal matrix A.

    e : complex, array-like, shape (n-1)
        The (n-1) subdiagonal elements of the tridiagonal matrix A.

    b : complex, array-like, shape (n, nrhs)
        The n by r right-hand side matrix B.

    Returns
    -------
    d : float, ndarray, shape (n)
        The n diagonal elements of the diagonal matrix D from the
        factorization A = LDL^H.

    e : complex, ndarray, shape (n-1)
        The (n-1) subdiagonal elements of the unit bidiagonal factor L
        from the LDL^H factorization of A. (`e` can also be regarded as
        the superdiagonal of the unit bidiagonal factor U from the U^HDU
        factorization of A.)

    b : complex, ndarray, shape (n, nrhs)
        If no exception or warning is raised, the n by r solution matrix
        X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -2)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

        (`errno` i > 0)
            The leading minor of order *<value>* is not positive
            definite, and the solution has not been computed. The
            factorization has not been completed unless n = *<value>*.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.zptsv'

    _d = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'd'},
        d,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_d, {'fun_name': fun_name, 'entity': 'd'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
    )
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _e = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'e'},
        e,
        copy=True,
    )
    _b_lapacklin.zptsv(
        _n,
        _nrhs,
        _d,
        _e,
        _b,
    )
    _d_py = _d
    _e_py = _e
    _b_py = _b
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['d', 'e', 'b'],
    )(
        _d_py,
        _e_py,
        _b_py,
    )
    return _return_data

def zptsvx(fact, n, nrhs, d, e, df, ef, b):
    u"""
    Uses the LDL^T factorization to compute the solution, error-bound
    and condition estimate for a complex Hermitian positive definite
    tridiagonal system of linear equations.

    ``zptsvx`` uses the factorization

        A = LDL^H

    to compute the solution to a complex system of linear equations

        AX = B,

    where A is an n by n Hermitian positive definite tridiagonal matrix
    and X and B are n by r matrices.
    Error bounds on the solution and a condition estimate are also
    provided.

    For full information please refer to the NAG Library document for
    f07jp

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07jpf.html

    Parameters
    ----------
    fact : str, length 1
        Specifies whether or not the factorized form of the matrix A has
        been supplied.

        `fact` = 'F'

            `df` and `ef` contain the factorized form of the matrix A.
            `df` and `ef` will not be modified.

        `fact` = 'N'

            The matrix A will be copied to `df` and `ef` and factorized.

    n : int
        n, the order of the matrix A.

    nrhs : int
        r, the number of right-hand sides, i.e., the number of columns
        of the matrix B.

    d : float, array-like, shape (`n`)
        The n diagonal elements of the tridiagonal matrix A.

    e : complex, array-like, shape (`n`-1)
        The (n-1) subdiagonal elements of the tridiagonal matrix A.

    df : float, array-like, shape (`n`)
        If `fact` = 'F', `df` must contain the n diagonal elements of
        the diagonal matrix D from the LDL^H factorization of A.

    ef : complex, array-like, shape (`n`-1)
        If `fact` = 'F', `ef` must contain the (n-1) subdiagonal
        elements of the unit bidiagonal factor L from the LDL^H
        factorization of A.

    b : complex, array-like, shape (`n`, `nrhs`)
        The n by r right-hand side matrix B.

    Returns
    -------
    df : float, ndarray, shape (`n`)
        If `fact` = 'N', `df` contains the n diagonal elements of the
        diagonal matrix D from the LDL^H factorization of A.

    ef : complex, ndarray, shape (`n`-1)
        If `fact` = 'N', `ef` contains the (n-1) subdiagonal elements of
        the unit bidiagonal factor L from the LDL^H factorization of A.

    x : complex, ndarray, shape (`n`, `nrhs`)
        If the function exits successfully or `errno` = `n` + 1, the n
        by r solution matrix X.

    rcond : float
        The reciprocal condition number of the matrix A. If `rcond` is
        less than the machine precision (in particular, if `rcond` =
        0.0), the matrix is singular to working precision. This
        condition is indicated by a return code of `errno` = `n` + 1.

    ferr : float, ndarray, shape (`nrhs`)
        The forward error bound for each solution vector x^_j (the jth
        column of the solution matrix X). If x_j is the true solution
        corresponding to x^_j, `ferr`[j-1] is an estimated upper bound
        for the magnitude of the largest element in (x^_j-x_j) divided
        by the magnitude of the largest element in x^_j.

    berr : float, ndarray, shape (`nrhs`)
        The component-wise relative backward error of each solution
        vector x^_j (i.e., the smallest relative change in any element
        of A or B that makes x^_j an exact solution).

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `fact`.

            Constraint: `fact` = 'F' or 'N'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -3)
            On entry, error in parameter `nrhs`.

            Constraint: `nrhs` >= 0.

        (`errno` (i > 0) and (i<=`n`))
            The leading minor of order *<value>* of A is not positive
            definite, so the factorization could not be completed, and
            the solution has not been computed. `rcond` = 0.0 is
            returned.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` `n`+1)
            D is nonsingular, but `rcond` is less than machine
            precision, meaning that the matrix is singular to working
            precision. Nevertheless, the solution and error bounds are
            computed because there are a number of situations where the
            computed solution can be more accurate than the value of
            `rcond` would suggest.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    fun_name = 'naginterfaces.library.lapacklin.zptsvx'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_b, 'b'),),
    )
    _d = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'd'},
        d,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _e = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'e'},
        e,
        copy=False,
    )
    _df = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'df'},
        df,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ef = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ef'},
        ef,
        copy=True,
    )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _nrhs = _b_utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineIntScalarType.to_py(_nrhs)
    _x = _b_utils._EngineComplex128ArrayType.empty_ndarray(
        (_n_py, _nrhs_py),
        sorder=_m_sorder,
    )
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    try:
        _return_scalar = _b_lapacklin.zptsvx(
            fact,
            _n,
            _nrhs,
            _d,
            _e,
            _df,
            _ef,
            _b,
            _x,
            _ferr,
            _berr,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _df_py = _df
    _ef_py = _ef
    _x_py = _x
    _rcond_py = _return_scalar
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['df', 'ef', 'x', 'rcond', 'ferr', 'berr'],
    )(
        _df_py,
        _ef_py,
        _x_py,
        _rcond_py,
        _ferr_py,
        _berr_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def zpttrf(d, e):
    u"""
    Computes the LDL^H factorization of a complex Hermitian positive
    definite tridiagonal matrix.

    ``zpttrf`` computes the modified Cholesky factorization of a complex
    n by n Hermitian positive definite tridiagonal matrix A.

    For full information please refer to the NAG Library document for
    f07jr

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07jrf.html

    Parameters
    ----------
    d : float, array-like, shape (n)
        Must contain the n diagonal elements of the matrix A.

    e : complex, array-like, shape (n-1)
        Must contain the (n-1) subdiagonal elements of the matrix A.

    Returns
    -------
    d : float, ndarray, shape (n)
        Is overwritten by the n diagonal elements of the diagonal matrix
        D from the LDL^H factorization of A.

    e : complex, ndarray, shape (n-1)
        Is overwritten by the (n-1) subdiagonal elements of the lower
        bidiagonal matrix L. (`e` can also be regarded as containing the
        (n-1) superdiagonal elements of the upper bidiagonal matrix U.)

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` (i > 0) and (i < n))
            The leading minor of order *<value>* is not positive
            definite, the factorization could not be completed.

        (`errno` (i > 0) and (n))
            The leading minor of order n is not positive definite, the
            factorization was completed, but `d`[n-1] <= 0.
    """

    fun_name = 'naginterfaces.library.lapacklin.zpttrf'

    _d = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'd'},
        d,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_d, {'fun_name': fun_name, 'entity': 'd'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _e = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'e'},
        e,
        copy=True,
    )
    _b_lapacklin.zpttrf(
        _n,
        _d,
        _e,
    )
    _d_py = _d
    _e_py = _e
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['d', 'e'],
    )(
        _d_py,
        _e_py,
    )
    return _return_data

def zpttrs(uplo, d, e, b):
    u"""
    Solves a complex Hermitian positive definite tridiagonal system
    using the LDL^H factorization computed by ``zpttrf``.

    ``zpttrs`` computes the solution to a complex system of linear
    equations AX = B, where A is an n by n Hermitian positive definite
    tridiagonal matrix and X and B are n by r matrices, using the LDL^H
    factorization returned by ``zpttrf``.

    For full information please refer to the NAG Library document for
    f07js

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07jsf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies the form of the factorization as follows:

        `uplo` = 'U'

            A = U^HDU.

        `uplo` = 'L'

            A = LDL^H.

    d : float, array-like, shape (n)
        Must contain the n diagonal elements of the diagonal matrix D
        from the LDL^H or U^HDU factorization of A.

    e : complex, array-like, shape (n-1)
        If `uplo` = 'U', `e` must contain the (n-1) superdiagonal
        elements of the unit upper bidiagonal matrix U from the U^HDU
        factorization of A.

        If `uplo` = 'L', `e` must contain the (n-1) subdiagonal elements
        of the unit lower bidiagonal matrix L from the LDL^H
        factorization of A.

    b : complex, array-like, shape (n, nrhs)
        The n by r matrix of right-hand sides B.

    Returns
    -------
    b : complex, ndarray, shape (n, nrhs)
        The n by r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug
    """

    fun_name = 'naginterfaces.library.lapacklin.zpttrs'

    _d = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'd'},
        d,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_d, {'fun_name': fun_name, 'entity': 'd'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
    )
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _e = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'e'},
        e,
        copy=False,
    )
    _b_lapacklin.zpttrs(
        uplo,
        _n,
        _nrhs,
        _d,
        _e,
        _b,
    )
    _b_py = _b
    return _b_py

def zptcon(n, d, e, anorm):
    u"""
    Computes the reciprocal of the condition number of a complex
    Hermitian positive definite tridiagonal system using the LDL^H
    factorization computed by ``zpttrf``.

    ``zptcon`` computes the reciprocal condition number of a complex n
    by n Hermitian positive definite tridiagonal matrix A, using the
    LDL^H factorization returned by ``zpttrf``.

    For full information please refer to the NAG Library document for
    f07ju

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07juf.html

    Parameters
    ----------
    n : int
        n, the order of the matrix A.

    d : float, array-like, shape (`n`)
        Must contain the n diagonal elements of the diagonal matrix D
        from the LDL^H factorization of A.

    e : complex, array-like, shape (`n`-1)
        Must contain the (n-1) subdiagonal elements of the unit lower
        bidiagonal matrix L. (`e` can also be regarded as the
        superdiagonal of the unit upper bidiagonal matrix U from the
        U^HDU factorization of A.)

    anorm : float
        The 1-norm of the **original** matrix A, which may be computed
        by calling ``blas.zlanht`` with its argument norm = '1'. `anorm`
        must be computed either **before** calling ``zpttrf`` or else
        from a **copy** of the original matrix A.

    Returns
    -------
    rcond : float
        The reciprocal condition number, 1/kappa_1(A) =
        1/(||A||_1||A^-1||_1).

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -4)
            On entry, error in parameter `anorm`.

            Constraint: `anorm` >= 0.0.

    References
    ----------
    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    fun_name = 'naginterfaces.library.lapacklin.zptcon'

    _d = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'd'},
        d,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _e = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'e'},
        e,
        copy=False,
    )
    _return_scalar = _b_lapacklin.zptcon(
        n,
        _d,
        _e,
        anorm,
    )
    _rcond_py = _return_scalar
    return _rcond_py

def zptrfs(uplo, n, nrhs, d, e, df, ef, b, x):
    u"""
    Refined solution with error bounds of complex Hermitian positive
    definite tridiagonal system of linear equations, multiple right-hand
    sides.

    ``zptrfs`` computes error bounds and refines the solution to a
    complex system of linear equations AX = B, where A is an n by n
    Hermitian positive definite tridiagonal matrix and X and B are n by
    r matrices, using the modified Cholesky factorization returned by
    ``zpttrf`` and an initial solution returned by ``zpttrs``.
    Iterative refinement is used to reduce the backward error as much as
    possible.

    For full information please refer to the NAG Library document for
    f07jv

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07jvf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies the form of the factorization as follows:

        `uplo` = 'U'

            A = U^HDU.

        `uplo` = 'L'

            A = LDL^H.

    n : int
        n, the order of the matrix A.

    nrhs : int
        r, the number of right-hand sides, i.e., the number of columns
        of the matrix B.

    d : float, array-like, shape (`n`)
        Must contain the n diagonal elements of the matrix of A.

    e : complex, array-like, shape (`n`-1)
        If `uplo` = 'U', `e` must contain the (n-1) superdiagonal
        elements of the matrix A.

        If `uplo` = 'L', `e` must contain the (n-1) subdiagonal elements
        of the matrix A.

    df : float, array-like, shape (`n`)
        Must contain the n diagonal elements of the diagonal matrix D
        from the LDL^T factorization of A.

    ef : complex, array-like, shape (`n`-1)
        If `uplo` = 'U', `ef` must contain the (n-1) superdiagonal
        elements of the unit upper bidiagonal matrix U from the U^HDU
        factorization of A.

        If `uplo` = 'L', `ef` must contain the (n-1) subdiagonal
        elements of the unit lower bidiagonal matrix L from the LDL^H
        factorization of A.

    b : complex, array-like, shape (`n`, `nrhs`)
        The n by r matrix of right-hand sides B.

    x : complex, array-like, shape (`n`, `nrhs`)
        The n by r initial solution matrix X.

    Returns
    -------
    x : complex, ndarray, shape (`n`, `nrhs`)
        The n by r refined solution matrix X.

    ferr : float, ndarray, shape (`nrhs`)
        Estimate of the forward error bound for each computed solution
        vector, such that ||x^_j-x_j||_infinity/||x^_j||_infinity <=
        `ferr`[j-1], where x^_j is the jth column of the computed
        solution returned in the array `x` and x_j is the corresponding
        column of the exact solution X. The estimate is almost always a
        slight overestimate of the true error.

    berr : float, ndarray, shape (`nrhs`)
        Estimate of the component-wise relative backward error of each
        computed solution vector x^_j (i.e., the smallest relative
        change in any element of A or B that makes x^_j an exact
        solution).

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -3)
            On entry, error in parameter `nrhs`.

            Constraint: `nrhs` >= 0.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug
    """

    fun_name = 'naginterfaces.library.lapacklin.zptrfs'

    _d = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'd'},
        d,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _e = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'e'},
        e,
        copy=False,
    )
    _df = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'df'},
        df,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ef = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ef'},
        ef,
        copy=False,
    )
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
    )
    _nrhs = _b_utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineIntScalarType.to_py(_nrhs)
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _b_lapacklin.zptrfs(
        uplo,
        n,
        _nrhs,
        _d,
        _e,
        _df,
        _ef,
        _b,
        _x,
        _ferr,
        _berr,
    )
    _x_py = _x
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'ferr', 'berr'],
    )(
        _x_py,
        _ferr_py,
        _berr_py,
    )
    return _return_data

def dpstrf(uplo, n, a, tol=-1):
    u"""
    Cholesky factorization, with complete pivoting, of a real,
    symmetric, positive semidefinite matrix.

    ``dpstrf`` computes the Cholesky factorization with complete
    pivoting of a real symmetric positive semidefinite matrix.

    For full information please refer to the NAG Library document for
    f07kd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07kdf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored and how A is to be factorized.

        `uplo` = 'U'

            The upper triangular part of A is stored and A is factorized
            as U^TU, where U is upper triangular.

        `uplo` = 'L'

            The lower triangular part of A is stored and A is factorized
            as LL^T, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    a : float, array-like, shape (`n`, `n`)
        The n by n symmetric positive semidefinite matrix A.

    tol : float, optional
        User defined tolerance. If `tol` < 0, n*max_{k =
        1,n}|A_{kk}|*machine precision will be used. The algorithm
        terminates at the rth step if the (r+1)th step pivot < `tol`.

    Returns
    -------
    a : float, ndarray, shape (`n`, `n`)
        If `uplo` = 'U', the first `rank` rows of the upper triangle of
        A are overwritten with the nonzero elements of the Cholesky
        factor U, and the remaining rows of the triangle are destroyed.

        If `uplo` = 'L', the first `rank` columns of the lower triangle
        of A are overwritten with the nonzero elements of the Cholesky
        factor L, and the remaining columns of the triangle are
        destroyed.

    piv : int, ndarray, shape (`n`)
        `piv` is such that the nonzero entries of P are P(`piv`[k-1],k)
        = 1, for k = 1,2,...,n.

    rank : int
        The computed rank of A given by the number of steps the
        algorithm completed.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            The matrix A is not positive definite. It is either positive
            semidefinite with computed rank as returned in `rank` and
            less than n, or it may be indefinite, see [Further
            Comments].

    References
    ----------
    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia

    Lucas, C, 2004, `LAPACK-style codes for Level 2 and 3 pivoted
    Cholesky factorizations`, LAPACK Working Note No. 161. Technical
    Report CS-04-522, Department of Computer Science, University of
    Tennessee, 107 Ayres Hall, Knoxville, TN 37996-1301, USA,
    https://www.netlib.org/lapack/lawnspdf/lawn161.pdf
    """

    fun_name = 'naginterfaces.library.lapacklin.dpstrf'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _piv = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _return_scalar = _b_lapacklin.dpstrf(
            uplo,
            _n,
            _a,
            _piv,
            tol,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _a_py = _a
    _piv_py = _piv
    _rank_py = _return_scalar
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'piv', 'rank'],
    )(
        _a_py,
        _piv_py,
        _rank_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def zpstrf(uplo, n, a, tol=-1):
    u"""
    Cholesky factorization of complex Hermitian positive semidefinite
    matrix.

    ``zpstrf`` computes the Cholesky factorization with complete
    pivoting of a complex Hermitian positive semidefinite matrix.

    For full information please refer to the NAG Library document for
    f07kr

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07krf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored and how A is to be factorized.

        `uplo` = 'U'

            The upper triangular part of A is stored and A is factorized
            as U^HU, where U is upper triangular.

        `uplo` = 'L'

            The lower triangular part of A is stored and A is factorized
            as LL^H, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    a : complex, array-like, shape (`n`, `n`)
        The n by n Hermitian positive semidefinite matrix A.

    tol : float, optional
        User defined tolerance. If `tol` < 0, n*max_{k =
        1,n}|A_{kk}|*machine precision will be used. The algorithm
        terminates at the rth step if the (r+1)th step pivot < `tol`.

    Returns
    -------
    a : complex, ndarray, shape (`n`, `n`)
        If `uplo` = 'U', the first `rank` rows of the upper triangle of
        A are overwritten with the nonzero elements of the Cholesky
        factor U, and the remaining rows of the triangle are destroyed.

        If `uplo` = 'L', the first `rank` columns of the lower triangle
        of A are overwritten with the nonzero elements of the Cholesky
        factor L, and the remaining columns of the triangle are
        destroyed.

    piv : int, ndarray, shape (`n`)
        `piv` is such that the nonzero entries of P are P(`piv`[k-1],k)
        = 1, for k = 1,2,...,n.

    rank : int
        The computed rank of A given by the number of steps the
        algorithm completed.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            The matrix A is not positive definite. It is either positive
            semidefinite with computed rank as returned in `rank` and
            less than n, or it may be indefinite, see [Further
            Comments].

    References
    ----------
    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia

    Lucas, C, 2004, `LAPACK-style codes for Level 2 and 3 pivoted
    Cholesky factorizations`, LAPACK Working Note No. 161. Technical
    Report CS-04-522, Department of Computer Science, University of
    Tennessee, 107 Ayres Hall, Knoxville, TN 37996-1301, USA,
    https://www.netlib.org/lapack/lawnspdf/lawn161.pdf
    """

    fun_name = 'naginterfaces.library.lapacklin.zpstrf'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _piv = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _return_scalar = _b_lapacklin.zpstrf(
            uplo,
            _n,
            _a,
            _piv,
            tol,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _a_py = _a
    _piv_py = _piv
    _rank_py = _return_scalar
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'piv', 'rank'],
    )(
        _a_py,
        _piv_py,
        _rank_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def dsysv(uplo, a, b):
    u"""
    Computes the solution to a real symmetric system of linear
    equations.

    ``dsysv`` computes the solution to a real system of linear equations

        AX = B,

    where A is an n by n symmetric matrix and X and B are n by r
    matrices.

    For full information please refer to the NAG Library document for
    f07ma

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07maf.html

    Parameters
    ----------
    uplo : str, length 1
        If `uplo` = 'U', the upper triangle of A is stored.

        If `uplo` = 'L', the lower triangle of A is stored.

    a : float, array-like, shape (n, n)
        The n by n symmetric matrix A.

    b : float, array-like, shape (n, nrhs)
        The n by r right-hand side matrix B.

    Returns
    -------
    a : float, ndarray, shape (n, n)
        If no exception or warning is raised, the block diagonal matrix
        D and the multipliers used to obtain the factor U or L from the
        factorization A = UDU^T or A = LDL^T as computed by ``dsytrf``.

    ipiv : int, ndarray, shape (n)
        Details of the interchanges and the block structure of D. More
        precisely,

            if `ipiv`[i-1] = k > 0, d_{ii} is a 1 by 1 pivot block and
            the ith row and column of A were interchanged with the kth
            row and column;

            if `uplo` = 'U' and `ipiv`[i-2] = `ipiv`[i-1] = -l < 0,
            [table omitted] is a 2 by 2 pivot block and the (i-1)th row
            and column of A were interchanged with the lth row and
            column;

            if `uplo` = 'L' and `ipiv`[i-1] = `ipiv`[i] = -m < 0, [table
            omitted] is a 2 by 2 pivot block and the (i+1)th row and
            column of A were interchanged with the mth row and column.

    b : float, ndarray, shape (n, nrhs)
        If no exception or warning is raised, the n by r solution matrix
        X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. The
            factorization has been completed, but the block diagonal
            matrix D is exactly singular, so the solution could not be
            computed.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.dsysv'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nrhs_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ipiv = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _b_lapacklin.dsysv(
            uplo,
            _n,
            _nrhs,
            _a,
            _ipiv,
            _b,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _a_py = _a
    _b_py = _b
    _ipiv_py = _ipiv
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'ipiv', 'b'],
    )(
        _a_py,
        _ipiv_py,
        _b_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def dsysvx(fact, uplo, a, af, ipiv, b):
    u"""
    Uses the diagonal pivoting factorization to compute the solution to
    a real symmetric system of linear equations.

    ``dsysvx`` uses the diagonal pivoting factorization to compute the
    solution to a real system of linear equations

        AX = B,

    where A is an n by n symmetric matrix and X and B are n by r
    matrices.
    Error bounds on the solution and a condition estimate are also
    provided.

    For full information please refer to the NAG Library document for
    f07mb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07mbf.html

    Parameters
    ----------
    fact : str, length 1
        Specifies whether or not the factorized form of the matrix A has
        been supplied.

        `fact` = 'F'

            `af` and `ipiv` contain the factorized form of the matrix A.
            `af` and `ipiv` will not be modified.

        `fact` = 'N'

            The matrix A will be copied to `af` and factorized.

    uplo : str, length 1
        If `uplo` = 'U', the upper triangle of A is stored.

        If `uplo` = 'L', the lower triangle of A is stored.

    a : float, array-like, shape (n, n)
        The n by n symmetric matrix A.

    af : float, array-like, shape (n, n)
        If `fact` = 'F', `af` contains the block diagonal matrix D and
        the multipliers used to obtain the factor U or L from the
        factorization `a` = UDU^T or `a` = LDL^T as computed by
        ``dsytrf``.

    ipiv : int, array-like, shape (n)
        If `fact` = 'F', `ipiv` contains details of the interchanges and
        the block structure of D, as determined by ``dsytrf``.

            if `ipiv`[i-1] = k > 0, d_{ii} is a 1 by 1 pivot block and
            the ith row and column of A were interchanged with the kth
            row and column;

            if `uplo` = 'U' and `ipiv`[i-2] = `ipiv`[i-1] = -l < 0,
            [table omitted] is a 2 by 2 pivot block and the (i-1)th row
            and column of A were interchanged with the lth row and
            column;

            if `uplo` = 'L' and `ipiv`[i-1] = `ipiv`[i] = -m < 0, [table
            omitted] is a 2 by 2 pivot block and the (i+1)th row and
            column of A were interchanged with the mth row and column.

    b : float, array-like, shape (n, nrhs)
        The n by r right-hand side matrix B.

    Returns
    -------
    af : float, ndarray, shape (n, n)
        If `fact` = 'N', `af` returns the block diagonal matrix D and
        the multipliers used to obtain the factor U or L from the
        factorization `a` = UDU^T or `a` = LDL^T.

    ipiv : int, ndarray, shape (n)
        If `fact` = 'N', `ipiv` contains details of the interchanges and
        the block structure of D, as determined by ``dsytrf``, as
        described above.

    x : float, ndarray, shape (n, nrhs)
        If the function exits successfully or `errno` = `n` + 1, the n
        by r solution matrix X.

    rcond : float
        The estimate of the reciprocal condition number of the matrix A.
        If `rcond` = 0.0, the matrix may be exactly singular. This
        condition is indicated by `errno` in 1 ... `n`. Otherwise, if
        `rcond` is less than the machine precision, the matrix is
        singular to working precision. This condition is indicated by
        `errno` = `n` + 1.

    ferr : float, ndarray, shape (nrhs)
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the forward error bound for each computed solution
        vector, such that ||x^_j-x_j||_infinity/||x_j||_infinity <=
        `ferr`[j-1] where x^_j is the jth column of the computed
        solution returned in the array `x` and x_j is the corresponding
        column of the exact solution X. The estimate is as reliable as
        the estimate for `rcond`, and is almost always a slight
        overestimate of the true error.

    berr : float, ndarray, shape (nrhs)
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the component-wise relative backward error of each
        computed solution vector x^_j (i.e., the smallest relative
        change in any element of A or B that makes x^_j an exact
        solution).

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `fact`.

            Constraint: `fact` = 'F' or 'N'.

        (`errno` -2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -3)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -4)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` (i > 0) and (i<=n))
            Element *<value>* of the diagonal is exactly zero. The
            factorization has been completed, but the factor D is
            exactly singular, so the solution and error bounds could not
            be computed. `rcond` = 0.0 is returned.

        (`errno` n+1)
            D is nonsingular, but `rcond` is less than machine
            precision, meaning that the matrix is singular to working
            precision. Nevertheless, the solution and error bounds are
            computed because there are a number of situations where the
            computed solution can be more accurate than the value of
            `rcond` would suggest.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    fun_name = 'naginterfaces.library.lapacklin.dsysvx'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _af = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'af'},
        af,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_a, 'a'), (_af, 'af'), (_b, 'b')),
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=True,
    )
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py, _nrhs_py),
        sorder=_m_sorder,
    )
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    try:
        _return_scalar = _b_lapacklin.dsysvx(
            fact,
            uplo,
            _n,
            _nrhs,
            _a,
            _af,
            _ipiv,
            _b,
            _x,
            _ferr,
            _berr,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _af_py = _af
    _ipiv_py = _ipiv
    _x_py = _x
    _rcond_py = _return_scalar
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['af', 'ipiv', 'x', 'rcond', 'ferr', 'berr'],
    )(
        _af_py,
        _ipiv_py,
        _x_py,
        _rcond_py,
        _ferr_py,
        _berr_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def dsytrf(uplo, a):
    u"""
    Bunch--Kaufman factorization of real symmetric indefinite matrix.

    ``dsytrf`` computes the Bunch--Kaufman factorization of a real
    symmetric indefinite matrix.

    For full information please refer to the NAG Library document for
    f07md

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07mdf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored and how A is to be factorized.

        `uplo` = 'U'

            The upper triangular part of A is stored and A is factorized
            as PUDU^TP^T, where U is upper triangular.

        `uplo` = 'L'

            The lower triangular part of A is stored and A is factorized
            as PLDL^TP^T, where L is lower triangular.

    a : float, array-like, shape (n, n)
        The n by n symmetric indefinite matrix A.

    Returns
    -------
    a : float, ndarray, shape (n, n)
        The upper or lower triangle of A is overwritten by details of
        the block diagonal matrix D and the multipliers used to obtain
        the factor U or L as specified by `uplo`.

    ipiv : int, ndarray, shape (n)
        Details of the interchanges and the block structure of D. More
        precisely,

            if `ipiv`[i-1] = k > 0, d_{ii} is a 1 by 1 pivot block and
            the ith row and column of A were interchanged with the kth
            row and column;

            if `uplo` = 'U' and `ipiv`[i-2] = `ipiv`[i-1] = -l < 0,
            [table omitted] is a 2 by 2 pivot block and the (i-1)th row
            and column of A were interchanged with the lth row and
            column;

            if `uplo` = 'L' and `ipiv`[i-1] = `ipiv`[i] = -m < 0, [table
            omitted] is a 2 by 2 pivot block and the (i+1)th row and
            column of A were interchanged with the mth row and column.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. The
            factorization has been completed, but the block diagonal
            matrix D is exactly singular, and division by zero will
            occur if it is used to solve a system of equations.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.dsytrf'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _ipiv = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _b_lapacklin.dsytrf(
            uplo,
            _n,
            _a,
            _ipiv,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _a_py = _a
    _ipiv_py = _ipiv
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'ipiv'],
    )(
        _a_py,
        _ipiv_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def dsytrs(uplo, a, ipiv, b):
    u"""
    Solution of real symmetric indefinite system of linear equations,
    multiple right-hand sides, matrix already factorized by ``dsytrf``.

    ``dsytrs`` solves a real symmetric indefinite system of linear
    equations with multiple right-hand sides,

        AX = B,

    where A has been factorized by ``dsytrf``.

    For full information please refer to the NAG Library document for
    f07me

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07mef.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = PUDU^TP^T, where U is upper triangular.

        `uplo` = 'L'

            A = PLDL^TP^T, where L is lower triangular.

    a : float, array-like, shape (n, n)
        Details of the factorization of A, as returned by ``dsytrf``.

    ipiv : int, array-like, shape (n)
        Details of the interchanges and the block structure of D, as
        returned by ``dsytrf``.

    b : float, array-like, shape (n, nrhs)
        The n by r right-hand side matrix B.

    Returns
    -------
    b : float, ndarray, shape (n, nrhs)
        The n by r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.dsytrs'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nrhs_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    _b_lapacklin.dsytrs(
        uplo,
        _n,
        _nrhs,
        _a,
        _ipiv,
        _b,
    )
    _b_py = _b
    return _b_py

def dsycon(uplo, n, a, ipiv, anorm):
    u"""
    Estimate condition number of real symmetric indefinite matrix,
    matrix already factorized by ``dsytrf``.

    ``dsycon`` estimates the condition number of a real symmetric
    indefinite matrix A, where A has been factorized by ``dsytrf``.

    For full information please refer to the NAG Library document for
    f07mg

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07mgf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = PUDU^TP^T, where U is upper triangular.

        `uplo` = 'L'

            A = PLDL^TP^T, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    a : float, array-like, shape (`n`, `n`)
        Details of the factorization of A, as returned by ``dsytrf``.

    ipiv : int, array-like, shape (`n`)
        Details of the interchanges and the block structure of D, as
        returned by ``dsytrf``.

    anorm : float
        The 1-norm of the **original** matrix A, which may be computed
        by calling ``blas.dlansy`` with its argument norm = '1'. `anorm`
        must be computed either **before** calling ``dsytrf`` or else
        from a **copy** of the original matrix A.

    Returns
    -------
    rcond : float
        An estimate of the reciprocal of the condition number of A.
        `rcond` is set to zero if exact singularity is detected or the
        estimate underflows. If `rcond` is less than machine precision,
        A is singular to working precision.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -6)
            On entry, error in parameter `anorm`.

            Constraint: `anorm` >= 0.0.

    References
    ----------
    Higham, N J, 1988, `FORTRAN codes for estimating the one-norm of a
    real or complex matrix, with applications to condition estimation`,
    ACM Trans. Math. Software (14), 381--396
    """

    fun_name = 'naginterfaces.library.lapacklin.dsycon'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    _return_scalar = _b_lapacklin.dsycon(
        uplo,
        n,
        _a,
        _ipiv,
        anorm,
    )
    _rcond_py = _return_scalar
    return _rcond_py

def dsyrfs(uplo, n, a, af, ipiv, b, x):
    u"""
    Refined solution with error bounds of real symmetric indefinite
    system of linear equations, multiple right-hand sides.

    ``dsyrfs`` returns error bounds for the solution of a real symmetric
    indefinite system of linear equations with multiple right-hand
    sides, AX = B.
    It improves the solution by iterative refinement, in order to reduce
    the backward error as much as possible.

    For full information please refer to the NAG Library document for
    f07mh

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07mhf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored and how A is to be factorized.

        `uplo` = 'U'

            The upper triangular part of A is stored and A is factorized
            as PUDU^TP^T, where U is upper triangular.

        `uplo` = 'L'

            The lower triangular part of A is stored and A is factorized
            as PLDL^TP^T, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    a : float, array-like, shape (`n`, `n`)
        The n by n original symmetric matrix A as supplied to
        ``dsytrf``.

    af : float, array-like, shape (`n`, `n`)
        Details of the factorization of A, as returned by ``dsytrf``.

    ipiv : int, array-like, shape (`n`)
        Details of the interchanges and the block structure of D, as
        returned by ``dsytrf``.

    b : float, array-like, shape (`n`, nrhs)
        The n by r right-hand side matrix B.

    x : float, array-like, shape (`n`, nrhs)
        The n by r solution matrix X, as returned by ``dsytrs``.

    Returns
    -------
    x : float, ndarray, shape (`n`, nrhs)
        The improved solution matrix X.

    ferr : float, ndarray, shape (nrhs)
        `ferr`[j-1] contains an estimated error bound for the jth
        solution vector, that is, the jth column of X, for j =
        1,2,...,r.

    berr : float, ndarray, shape (nrhs)
        `berr`[j-1] contains the component-wise backward error bound
        beta for the jth solution vector, that is, the jth column of X,
        for j = 1,2,...,r.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -3)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.dsyrfs'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nrhs_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _af = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'af'},
        af,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _b_lapacklin.dsyrfs(
        uplo,
        n,
        _nrhs,
        _a,
        _af,
        _ipiv,
        _b,
        _x,
        _ferr,
        _berr,
    )
    _x_py = _x
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'ferr', 'berr'],
    )(
        _x_py,
        _ferr_py,
        _berr_py,
    )
    return _return_data

def dsytri(uplo, n, a, ipiv):
    u"""
    Inverse of real symmetric indefinite matrix, matrix already
    factorized by ``dsytrf``.

    ``dsytri`` computes the inverse of a real symmetric indefinite
    matrix A, where A has been factorized by ``dsytrf``.

    For full information please refer to the NAG Library document for
    f07mj

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07mjf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = PUDU^TP^T, where U is upper triangular.

        `uplo` = 'L'

            A = PLDL^TP^T, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    a : float, array-like, shape (`n`, `n`)
        Details of the factorization of A, as returned by ``dsytrf``.

    ipiv : int, array-like, shape (`n`)
        Details of the interchanges and the block structure of D, as
        returned by ``dsytrf``.

    Returns
    -------
    a : float, ndarray, shape (`n`, `n`)
        The factorization is overwritten by the n by n symmetric matrix
        A^-1.

        If `uplo` = 'U', the upper triangle of A^-1 is stored in the
        upper triangular part of the array.

        If `uplo` = 'L', the lower triangle of A^-1 is stored in the
        lower triangular part of the array.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. D is
            singular and the inverse of A cannot be computed.

    References
    ----------
    Du Croz, J J and Higham, N J, 1992, `Stability of methods for matrix
    inversion`, IMA J. Numer. Anal. (12), 1--19
    """

    fun_name = 'naginterfaces.library.lapacklin.dsytri'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    try:
        _b_lapacklin.dsytri(
            uplo,
            n,
            _a,
            _ipiv,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _a_py = _a
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a'],
    )(
        _a_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _a_py

def zhesv(uplo, a, b):
    u"""
    Computes the solution to a complex Hermitian system of linear
    equations.

    ``zhesv`` computes the solution to a complex system of linear
    equations

        AX = B,

    where A is an n by n Hermitian matrix and X and B are n by r
    matrices.

    For full information please refer to the NAG Library document for
    f07mn

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07mnf.html

    Parameters
    ----------
    uplo : str, length 1
        If `uplo` = 'U', the upper triangle of A is stored.

        If `uplo` = 'L', the lower triangle of A is stored.

    a : complex, array-like, shape (n, n)
        The n by n Hermitian matrix A.

    b : complex, array-like, shape (n, nrhs)
        The n by r right-hand side matrix B.

    Returns
    -------
    a : complex, ndarray, shape (n, n)
        If no exception or warning is raised, the block diagonal matrix
        D and the multipliers used to obtain the factor U or L from the
        factorization A = UDU^H or A = LDL^H as computed by ``zhetrf``.

    ipiv : int, ndarray, shape (n)
        Details of the interchanges and the block structure of D. More
        precisely,

            if `ipiv`[i-1] = k > 0, d_{ii} is a 1 by 1 pivot block and
            the ith row and column of A were interchanged with the kth
            row and column;

            if `uplo` = 'U' and `ipiv`[i-2] = `ipiv`[i-1] = -l < 0,
            [table omitted] is a 2 by 2 pivot block and the (i-1)th row
            and column of A were interchanged with the lth row and
            column;

            if `uplo` = 'L' and `ipiv`[i-1] = `ipiv`[i] = -m < 0, [table
            omitted] is a 2 by 2 pivot block and the (i+1)th row and
            column of A were interchanged with the mth row and column.

    b : complex, ndarray, shape (n, nrhs)
        If no exception or warning is raised, the n by r solution matrix
        X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. The
            factorization has been completed, but the block diagonal
            matrix D is exactly singular, so the solution could not be
            computed.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.zhesv'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
    )
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ipiv = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _b_lapacklin.zhesv(
            uplo,
            _n,
            _nrhs,
            _a,
            _ipiv,
            _b,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _a_py = _a
    _b_py = _b
    _ipiv_py = _ipiv
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'ipiv', 'b'],
    )(
        _a_py,
        _ipiv_py,
        _b_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def zhesvx(fact, uplo, a, af, ipiv, b):
    u"""
    Uses the diagonal pivoting factorization to compute the solution to
    a complex Hermitian system of linear equations.

    ``zhesvx`` uses the diagonal pivoting factorization to compute the
    solution to a complex system of linear equations

        AX = B,

    where A is an n by n Hermitian matrix and X and B are n by r
    matrices.
    Error bounds on the solution and a condition estimate are also
    provided.

    For full information please refer to the NAG Library document for
    f07mp

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07mpf.html

    Parameters
    ----------
    fact : str, length 1
        Specifies whether or not the factorized form of the matrix A has
        been supplied.

        `fact` = 'F'

            `af` and `ipiv` contain the factorized form of the matrix A.
            `af` and `ipiv` will not be modified.

        `fact` = 'N'

            The matrix A will be copied to `af` and factorized.

    uplo : str, length 1
        If `uplo` = 'U', the upper triangle of A is stored.

        If `uplo` = 'L', the lower triangle of A is stored.

    a : complex, array-like, shape (n, n)
        The n by n Hermitian matrix A.

    af : complex, array-like, shape (n, n)
        If `fact` = 'F', `af` contains the block diagonal matrix D and
        the multipliers used to obtain the factor U or L from the
        factorization `a` = UDU^H or `a` = LDL^H as computed by
        ``zhetrf``.

    ipiv : int, array-like, shape (n)
        If `fact` = 'F', `ipiv` contains details of the interchanges and
        the block structure of D, as determined by ``zhetrf``.

            if `ipiv`[i-1] = k > 0, d_{ii} is a 1 by 1 pivot block and
            the ith row and column of A were interchanged with the kth
            row and column;

            if `uplo` = 'U' and `ipiv`[i-2] = `ipiv`[i-1] = -l < 0,
            [table omitted] is a 2 by 2 pivot block and the (i-1)th row
            and column of A were interchanged with the lth row and
            column;

            if `uplo` = 'L' and `ipiv`[i-1] = `ipiv`[i] = -m < 0, [table
            omitted] is a 2 by 2 pivot block and the (i+1)th row and
            column of A were interchanged with the mth row and column.

    b : complex, array-like, shape (n, nrhs)
        The n by r right-hand side matrix B.

    Returns
    -------
    af : complex, ndarray, shape (n, n)
        If `fact` = 'N', `af` returns the block diagonal matrix D and
        the multipliers used to obtain the factor U or L from the
        factorization `a` = UDU^H or `a` = LDL^H.

    ipiv : int, ndarray, shape (n)
        If `fact` = 'N', `ipiv` contains details of the interchanges and
        the block structure of D, as determined by ``zhetrf``, as
        described above.

    x : complex, ndarray, shape (n, nrhs)
        If the function exits successfully or `errno` = `n` + 1, the n
        by r solution matrix X.

    rcond : float
        The estimate of the reciprocal condition number of the matrix A.
        If `rcond` = 0.0, the matrix may be exactly singular. This
        condition is indicated by `errno` in 1 ... `n`. Otherwise, if
        `rcond` is less than the machine precision, the matrix is
        singular to working precision. This condition is indicated by
        `errno` = `n` + 1.

    ferr : float, ndarray, shape (nrhs)
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the forward error bound for each computed solution
        vector, such that ||x^_j-x_j||_infinity/||x_j||_infinity <=
        `ferr`[j-1] where x^_j is the jth column of the computed
        solution returned in the array `x` and x_j is the corresponding
        column of the exact solution X. The estimate is as reliable as
        the estimate for `rcond`, and is almost always a slight
        overestimate of the true error.

    berr : float, ndarray, shape (nrhs)
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the component-wise relative backward error of each
        computed solution vector x^_j (i.e., the smallest relative
        change in any element of A or B that makes x^_j an exact
        solution).

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `fact`.

            Constraint: `fact` = 'F' or 'N'.

        (`errno` -2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -3)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -4)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` (i > 0) and (i<=n))
            Element *<value>* of the diagonal is exactly zero. The
            factorization has been completed, but the factor D is
            exactly singular, so the solution and error bounds could not
            be computed. `rcond` = 0.0 is returned.

        (`errno` n+1)
            D is nonsingular, but `rcond` is less than machine
            precision, meaning that the matrix is singular to working
            precision. Nevertheless, the solution and error bounds are
            computed because there are a number of situations where the
            computed solution can be more accurate than the value of
            `rcond` would suggest.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    fun_name = 'naginterfaces.library.lapacklin.zhesvx'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
    )
    _af = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'af'},
        af,
        copy=True,
    )
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_a, 'a'), (_af, 'af'), (_b, 'b')),
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=True,
    )
    _x = _b_utils._EngineComplex128ArrayType.empty_ndarray(
        (_n_py, _nrhs_py),
        sorder=_m_sorder,
    )
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    try:
        _return_scalar = _b_lapacklin.zhesvx(
            fact,
            uplo,
            _n,
            _nrhs,
            _a,
            _af,
            _ipiv,
            _b,
            _x,
            _ferr,
            _berr,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _af_py = _af
    _ipiv_py = _ipiv
    _x_py = _x
    _rcond_py = _return_scalar
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['af', 'ipiv', 'x', 'rcond', 'ferr', 'berr'],
    )(
        _af_py,
        _ipiv_py,
        _x_py,
        _rcond_py,
        _ferr_py,
        _berr_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def zhetrf(uplo, n, a):
    u"""
    Bunch--Kaufman factorization of complex Hermitian indefinite matrix.

    ``zhetrf`` computes the Bunch--Kaufman factorization of a complex
    Hermitian indefinite matrix.

    For full information please refer to the NAG Library document for
    f07mr

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07mrf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored and how A is to be factorized.

        `uplo` = 'U'

            The upper triangular part of A is stored and A is factorized
            as PUDU^HP^T, where U is upper triangular.

        `uplo` = 'L'

            The lower triangular part of A is stored and A is factorized
            as PLDL^HP^T, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    a : complex, array-like, shape (`n`, `n`)
        The n by n Hermitian indefinite matrix A.

    Returns
    -------
    a : complex, ndarray, shape (`n`, `n`)
        The upper or lower triangle of A is overwritten by details of
        the block diagonal matrix D and the multipliers used to obtain
        the factor U or L as specified by `uplo`.

    ipiv : int, ndarray, shape (`n`)
        Details of the interchanges and the block structure of D. More
        precisely,

            if `ipiv`[i-1] = k > 0, d_{ii} is a 1 by 1 pivot block and
            the ith row and column of A were interchanged with the kth
            row and column;

            if `uplo` = 'U' and `ipiv`[i-2] = `ipiv`[i-1] = -l < 0,
            [table omitted] is a 2 by 2 pivot block and the (i-1)th row
            and column of A were interchanged with the lth row and
            column;

            if `uplo` = 'L' and `ipiv`[i-1] = `ipiv`[i] = -m < 0, [table
            omitted] is a 2 by 2 pivot block and the (i+1)th row and
            column of A were interchanged with the mth row and column.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. The
            factorization has been completed, but the block diagonal
            matrix D is exactly singular, and division by zero will
            occur if it is used to solve a system of equations.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.zhetrf'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _ipiv = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _b_lapacklin.zhetrf(
            uplo,
            _n,
            _a,
            _ipiv,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _a_py = _a
    _ipiv_py = _ipiv
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'ipiv'],
    )(
        _a_py,
        _ipiv_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def zhetrs(uplo, a, ipiv, b):
    u"""
    Solution of complex Hermitian indefinite system of linear equations,
    multiple right-hand sides, matrix already factorized by ``zhetrf``.

    ``zhetrs`` solves a complex Hermitian indefinite system of linear
    equations with multiple right-hand sides,

        AX = B,

    where A has been factorized by ``zhetrf``.

    For full information please refer to the NAG Library document for
    f07ms

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07msf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = PUDU^HP^T, where U is upper triangular.

        `uplo` = 'L'

            A = PLDL^HP^T, where L is lower triangular.

    a : complex, array-like, shape (n, n)
        Details of the factorization of A, as returned by ``zhetrf``.

    ipiv : int, array-like, shape (n)
        Details of the interchanges and the block structure of D, as
        returned by ``zhetrf``.

    b : complex, array-like, shape (n, nrhs)
        The n by r right-hand side matrix B.

    Returns
    -------
    b : complex, ndarray, shape (n, nrhs)
        The n by r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.zhetrs'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
    )
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    _b_lapacklin.zhetrs(
        uplo,
        _n,
        _nrhs,
        _a,
        _ipiv,
        _b,
    )
    _b_py = _b
    return _b_py

def zhecon(uplo, a, ipiv, anorm):
    u"""
    Estimate condition number of complex Hermitian indefinite matrix,
    matrix already factorized by ``zhetrf``.

    ``zhecon`` estimates the condition number of a complex Hermitian
    indefinite matrix A, where A has been factorized by ``zhetrf``.

    For full information please refer to the NAG Library document for
    f07mu

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07muf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = PUDU^HP^T, where U is upper triangular.

        `uplo` = 'L'

            A = PLDL^HP^T, where L is lower triangular.

    a : complex, array-like, shape (n, n)
        Details of the factorization of A, as returned by ``zhetrf``.

    ipiv : int, array-like, shape (n)
        Details of the interchanges and the block structure of D, as
        returned by ``zhetrf``.

    anorm : float
        The 1-norm of the **original** matrix A, which may be computed
        by calling ``blas.zlanhe`` with its argument norm = '1'. `anorm`
        must be computed either **before** calling ``zhetrf`` or else
        from a **copy** of the original matrix A.

    Returns
    -------
    rcond : float
        An estimate of the reciprocal of the condition number of A.
        `rcond` is set to zero if exact singularity is detected or the
        estimate underflows. If `rcond` is less than machine precision,
        A is singular to working precision.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -6)
            On entry, error in parameter `anorm`.

            Constraint: `anorm` >= 0.0.

    References
    ----------
    Higham, N J, 1988, `FORTRAN codes for estimating the one-norm of a
    real or complex matrix, with applications to condition estimation`,
    ACM Trans. Math. Software (14), 381--396
    """

    fun_name = 'naginterfaces.library.lapacklin.zhecon'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    _return_scalar = _b_lapacklin.zhecon(
        uplo,
        _n,
        _a,
        _ipiv,
        anorm,
    )
    _rcond_py = _return_scalar
    return _rcond_py

def zherfs(uplo, n, a, af, ipiv, b, x):
    u"""
    Refined solution with error bounds of complex Hermitian indefinite
    system of linear equations, multiple right-hand sides.

    ``zherfs`` returns error bounds for the solution of a complex
    Hermitian indefinite system of linear equations with multiple
    right-hand sides, AX = B.
    It improves the solution by iterative refinement, in order to reduce
    the backward error as much as possible.

    For full information please refer to the NAG Library document for
    f07mv

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07mvf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored and how A is to be factorized.

        `uplo` = 'U'

            The upper triangular part of A is stored and A is factorized
            as PUDU^HP^T, where U is upper triangular.

        `uplo` = 'L'

            The lower triangular part of A is stored and A is factorized
            as PLDL^HP^T, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    a : complex, array-like, shape (`n`, `n`)
        The n by n original Hermitian matrix A as supplied to
        ``zhetrf``.

    af : complex, array-like, shape (`n`, `n`)
        Details of the factorization of A, as returned by ``zhetrf``.

    ipiv : int, array-like, shape (`n`)
        Details of the interchanges and the block structure of D, as
        returned by ``zhetrf``.

    b : complex, array-like, shape (`n`, nrhs)
        The n by r right-hand side matrix B.

    x : complex, array-like, shape (`n`, nrhs)
        The n by r solution matrix X, as returned by ``zhetrs``.

    Returns
    -------
    x : complex, ndarray, shape (`n`, nrhs)
        The improved solution matrix X.

    ferr : float, ndarray, shape (nrhs)
        `ferr`[j-1] contains an estimated error bound for the jth
        solution vector, that is, the jth column of X, for j =
        1,2,...,r.

    berr : float, ndarray, shape (nrhs)
        `berr`[j-1] contains the component-wise backward error bound
        beta for the jth solution vector, that is, the jth column of X,
        for j = 1,2,...,r.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -3)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.zherfs'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
    )
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
    )
    _af = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'af'},
        af,
        copy=False,
    )
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
    )
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _b_lapacklin.zherfs(
        uplo,
        n,
        _nrhs,
        _a,
        _af,
        _ipiv,
        _b,
        _x,
        _ferr,
        _berr,
    )
    _x_py = _x
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'ferr', 'berr'],
    )(
        _x_py,
        _ferr_py,
        _berr_py,
    )
    return _return_data

def zhetri(uplo, n, a, ipiv):
    u"""
    Inverse of complex Hermitian indefinite matrix, matrix already
    factorized by ``zhetrf``.

    ``zhetri`` computes the inverse of a complex Hermitian indefinite
    matrix A, where A has been factorized by ``zhetrf``.

    For full information please refer to the NAG Library document for
    f07mw

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07mwf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = PUDU^HP^T, where U is upper triangular.

        `uplo` = 'L'

            A = PLDL^HP^T, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    a : complex, array-like, shape (`n`, `n`)
        Details of the factorization of A, as returned by ``zhetrf``.

    ipiv : int, array-like, shape (`n`)
        Details of the interchanges and the block structure of D, as
        returned by ``zhetrf``.

    Returns
    -------
    a : complex, ndarray, shape (`n`, `n`)
        The factorization is overwritten by the n by n Hermitian matrix
        A^-1.

        If `uplo` = 'U', the upper triangle of A^-1 is stored in the
        upper triangular part of the array.

        If `uplo` = 'L', the lower triangle of A^-1 is stored in the
        lower triangular part of the array.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. D is
            singular and the inverse of A cannot be computed.

    References
    ----------
    Du Croz, J J and Higham, N J, 1992, `Stability of methods for matrix
    inversion`, IMA J. Numer. Anal. (12), 1--19
    """

    fun_name = 'naginterfaces.library.lapacklin.zhetri'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    try:
        _b_lapacklin.zhetri(
            uplo,
            n,
            _a,
            _ipiv,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _a_py = _a
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a'],
    )(
        _a_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _a_py

def zsysv(uplo, a, b):
    u"""
    Computes the solution to a complex symmetric system of linear
    equations.

    ``zsysv`` computes the solution to a complex system of linear
    equations

        AX = B,

    where A is an n by n symmetric matrix and X and B are n by r
    matrices.

    For full information please refer to the NAG Library document for
    f07nn

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07nnf.html

    Parameters
    ----------
    uplo : str, length 1
        If `uplo` = 'U', the upper triangle of A is stored.

        If `uplo` = 'L', the lower triangle of A is stored.

    a : complex, array-like, shape (n, n)
        The n by n symmetric matrix A.

    b : complex, array-like, shape (n, nrhs)
        The n by r right-hand side matrix B.

    Returns
    -------
    a : complex, ndarray, shape (n, n)
        If no exception or warning is raised, the block diagonal matrix
        D and the multipliers used to obtain the factor U or L from the
        factorization A = UDU^H or A = LDL^H as computed by ``zsytrf``.

    ipiv : int, ndarray, shape (n)
        Details of the interchanges and the block structure of D. More
        precisely,

            if `ipiv`[i-1] = k > 0, d_{ii} is a 1 by 1 pivot block and
            the ith row and column of A were interchanged with the kth
            row and column;

            if `uplo` = 'U' and `ipiv`[i-2] = `ipiv`[i-1] = -l < 0,
            [table omitted] is a 2 by 2 pivot block and the (i-1)th row
            and column of A were interchanged with the lth row and
            column;

            if `uplo` = 'L' and `ipiv`[i-1] = `ipiv`[i] = -m < 0, [table
            omitted] is a 2 by 2 pivot block and the (i+1)th row and
            column of A were interchanged with the mth row and column.

    b : complex, ndarray, shape (n, nrhs)
        If no exception or warning is raised, the n by r solution matrix
        X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. The
            factorization has been completed, but the block diagonal
            matrix D is exactly singular, so the solution could not be
            computed.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    fun_name = 'naginterfaces.library.lapacklin.zsysv'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
    )
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ipiv = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _b_lapacklin.zsysv(
            uplo,
            _n,
            _nrhs,
            _a,
            _ipiv,
            _b,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _a_py = _a
    _b_py = _b
    _ipiv_py = _ipiv
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'ipiv', 'b'],
    )(
        _a_py,
        _ipiv_py,
        _b_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def zsysvx(fact, uplo, a, af, ipiv, b):
    u"""
    Uses the diagonal pivoting factorization to compute the solution to
    a complex symmetric system of linear equations.

    ``zsysvx`` uses the diagonal pivoting factorization to compute the
    solution to a complex system of linear equations

        AX = B,

    where A is an n by n symmetric matrix and X and B are n by r
    matrices.
    Error bounds on the solution and a condition estimate are also
    provided.

    For full information please refer to the NAG Library document for
    f07np

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07npf.html

    Parameters
    ----------
    fact : str, length 1
        Specifies whether or not the factorized form of the matrix A has
        been supplied.

        `fact` = 'F'

            `af` and `ipiv` contain the factorized form of the matrix A.
            `af` and `ipiv` will not be modified.

        `fact` = 'N'

            The matrix A will be copied to `af` and factorized.

    uplo : str, length 1
        If `uplo` = 'U', the upper triangle of A is stored.

        If `uplo` = 'L', the lower triangle of A is stored.

    a : complex, array-like, shape (n, n)
        The n by n symmetric matrix A.

    af : complex, array-like, shape (n, n)
        If `fact` = 'F', `af` contains the block diagonal matrix D and
        the multipliers used to obtain the factor U or L from the
        factorization `a` = UDU^H or `a` = LDL^H as computed by
        ``zsytrf``.

    ipiv : int, array-like, shape (n)
        If `fact` = 'F', `ipiv` contains details of the interchanges and
        the block structure of D, as determined by ``zsytrf``.

            if `ipiv`[i-1] = k > 0, d_{ii} is a 1 by 1 pivot block and
            the ith row and column of A were interchanged with the kth
            row and column;

            if `uplo` = 'U' and `ipiv`[i-2] = `ipiv`[i-1] = -l < 0,
            [table omitted] is a 2 by 2 pivot block and the (i-1)th row
            and column of A were interchanged with the lth row and
            column;

            if `uplo` = 'L' and `ipiv`[i-1] = `ipiv`[i] = -m < 0, [table
            omitted] is a 2 by 2 pivot block and the (i+1)th row and
            column of A were interchanged with the mth row and column.

    b : complex, array-like, shape (n, nrhs)
        The n by r right-hand side matrix B.

    Returns
    -------
    af : complex, ndarray, shape (n, n)
        If `fact` = 'N', `af` returns the block diagonal matrix D and
        the multipliers used to obtain the factor U or L from the
        factorization `a` = UDU^H or `a` = LDL^H.

    ipiv : int, ndarray, shape (n)
        If `fact` = 'N', `ipiv` contains details of the interchanges and
        the block structure of D, as determined by ``zsytrf``, as
        described above.

    x : complex, ndarray, shape (n, nrhs)
        If the function exits successfully or `errno` = `n` + 1, the n
        by r solution matrix X.

    rcond : float
        The estimate of the reciprocal condition number of the matrix A.
        If `rcond` = 0.0, the matrix may be exactly singular. This
        condition is indicated by `errno` in 1 ... `n`. Otherwise, if
        `rcond` is less than the machine precision, the matrix is
        singular to working precision. This condition is indicated by
        `errno` = `n` + 1.

    ferr : float, ndarray, shape (nrhs)
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the forward error bound for each computed solution
        vector, such that ||x^_j-x_j||_infinity/||x_j||_infinity <=
        `ferr`[j-1] where x^_j is the jth column of the computed
        solution returned in the array `x` and x_j is the corresponding
        column of the exact solution X. The estimate is as reliable as
        the estimate for `rcond`, and is almost always a slight
        overestimate of the true error.

    berr : float, ndarray, shape (nrhs)
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the component-wise relative backward error of each
        computed solution vector x^_j (i.e., the smallest relative
        change in any element of A or B that makes x^_j an exact
        solution).

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `fact`.

            Constraint: `fact` = 'F' or 'N'.

        (`errno` -2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -3)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -4)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` (i > 0) and (i<=n))
            Element *<value>* of the diagonal is exactly zero. The
            factorization has been completed, but the factor D is
            exactly singular, so the solution and error bounds could not
            be computed. `rcond` = 0.0 is returned.

        (`errno` n+1)
            D is nonsingular, but `rcond` is less than machine
            precision, meaning that the matrix is singular to working
            precision. Nevertheless, the solution and error bounds are
            computed because there are a number of situations where the
            computed solution can be more accurate than the value of
            `rcond` would suggest.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    fun_name = 'naginterfaces.library.lapacklin.zsysvx'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
    )
    _af = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'af'},
        af,
        copy=True,
    )
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_a, 'a'), (_af, 'af'), (_b, 'b')),
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=True,
    )
    _x = _b_utils._EngineComplex128ArrayType.empty_ndarray(
        (_n_py, _nrhs_py),
        sorder=_m_sorder,
    )
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    try:
        _return_scalar = _b_lapacklin.zsysvx(
            fact,
            uplo,
            _n,
            _nrhs,
            _a,
            _af,
            _ipiv,
            _b,
            _x,
            _ferr,
            _berr,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _af_py = _af
    _ipiv_py = _ipiv
    _x_py = _x
    _rcond_py = _return_scalar
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['af', 'ipiv', 'x', 'rcond', 'ferr', 'berr'],
    )(
        _af_py,
        _ipiv_py,
        _x_py,
        _rcond_py,
        _ferr_py,
        _berr_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def zsytrf(uplo, a):
    u"""
    Bunch--Kaufman factorization of complex symmetric matrix.

    ``zsytrf`` computes the Bunch--Kaufman factorization of a complex
    symmetric matrix.

    For full information please refer to the NAG Library document for
    f07nr

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07nrf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored and how A is to be factorized.

        `uplo` = 'U'

            The upper triangular part of A is stored and A is factorized
            as PUDU^HP^T, where U is upper triangular.

        `uplo` = 'L'

            The lower triangular part of A is stored and A is factorized
            as PLDL^HP^T, where L is lower triangular.

    a : complex, array-like, shape (n, n)
        The n by n symmetric indefinite matrix A.

    Returns
    -------
    a : complex, ndarray, shape (n, n)
        The upper or lower triangle of A is overwritten by details of
        the block diagonal matrix D and the multipliers used to obtain
        the factor U or L as specified by `uplo`.

    ipiv : int, ndarray, shape (n)
        Details of the interchanges and the block structure of D. More
        precisely,

            if `ipiv`[i-1] = k > 0, d_{ii} is a 1 by 1 pivot block and
            the ith row and column of A were interchanged with the kth
            row and column;

            if `uplo` = 'U' and `ipiv`[i-2] = `ipiv`[i-1] = -l < 0,
            [table omitted] is a 2 by 2 pivot block and the (i-1)th row
            and column of A were interchanged with the lth row and
            column;

            if `uplo` = 'L' and `ipiv`[i-1] = `ipiv`[i] = -m < 0, [table
            omitted] is a 2 by 2 pivot block and the (i+1)th row and
            column of A were interchanged with the mth row and column.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. The
            factorization has been completed, but the block diagonal
            matrix D is exactly singular, and division by zero will
            occur if it is used to solve a system of equations.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.zsytrf'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _ipiv = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _b_lapacklin.zsytrf(
            uplo,
            _n,
            _a,
            _ipiv,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _a_py = _a
    _ipiv_py = _ipiv
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'ipiv'],
    )(
        _a_py,
        _ipiv_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def zsytrs(uplo, a, ipiv, b):
    u"""
    Solution of complex symmetric system of linear equations, multiple
    right-hand sides, matrix already factorized by ``zsytrf``.

    ``zsytrs`` solves a complex symmetric system of linear equations
    with multiple right-hand sides,

        AX = B,

    where A has been factorized by ``zsytrf``.

    For full information please refer to the NAG Library document for
    f07ns

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07nsf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = PUDU^HP^T, where U is upper triangular.

        `uplo` = 'L'

            A = PLDL^HP^T, where L is lower triangular.

    a : complex, array-like, shape (n, n)
        Details of the factorization of A, as returned by ``zsytrf``.

    ipiv : int, array-like, shape (n)
        Details of the interchanges and the block structure of D, as
        returned by ``zsytrf``.

    b : complex, array-like, shape (n, nrhs)
        The n by r right-hand side matrix B.

    Returns
    -------
    b : complex, ndarray, shape (n, nrhs)
        The n by r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.zsytrs'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
    )
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    _b_lapacklin.zsytrs(
        uplo,
        _n,
        _nrhs,
        _a,
        _ipiv,
        _b,
    )
    _b_py = _b
    return _b_py

def zsycon(uplo, a, ipiv, anorm):
    u"""
    Estimate condition number of complex symmetric matrix, matrix
    already factorized by ``zsytrf``.

    ``zsycon`` estimates the condition number of a complex symmetric
    matrix A, where A has been factorized by ``zsytrf``.

    For full information please refer to the NAG Library document for
    f07nu

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07nuf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = PUDU^HP^T, where U is upper triangular.

        `uplo` = 'L'

            A = PLDL^HP^T, where L is lower triangular.

    a : complex, array-like, shape (n, n)
        Details of the factorization of A, as returned by ``zsytrf``.

    ipiv : int, array-like, shape (n)
        Details of the interchanges and the block structure of D, as
        returned by ``zsytrf``.

    anorm : float
        The 1-norm of the **original** matrix A, which may be computed
        by calling ``blas.zlansy`` with its argument norm = '1'. `anorm`
        must be computed either **before** calling ``zsytrf`` or else
        from a **copy** of the original matrix A.

    Returns
    -------
    rcond : float
        An estimate of the reciprocal of the condition number of A.
        `rcond` is set to zero if exact singularity is detected or the
        estimate underflows. If `rcond` is less than machine precision,
        A is singular to working precision.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -6)
            On entry, error in parameter `anorm`.

            Constraint: `anorm` >= 0.0.

    References
    ----------
    Higham, N J, 1988, `FORTRAN codes for estimating the one-norm of a
    real or complex matrix, with applications to condition estimation`,
    ACM Trans. Math. Software (14), 381--396
    """

    fun_name = 'naginterfaces.library.lapacklin.zsycon'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    _return_scalar = _b_lapacklin.zsycon(
        uplo,
        _n,
        _a,
        _ipiv,
        anorm,
    )
    _rcond_py = _return_scalar
    return _rcond_py

def zsyrfs(uplo, n, a, af, ipiv, b, x):
    u"""
    Refined solution with error bounds of complex symmetric system of
    linear equations, multiple right-hand sides.

    ``zsyrfs`` returns error bounds for the solution of a complex
    symmetric system of linear equations with multiple right-hand sides,
    AX = B.
    It improves the solution by iterative refinement, in order to reduce
    the backward error as much as possible.

    For full information please refer to the NAG Library document for
    f07nv

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07nvf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored and how A is to be factorized.

        `uplo` = 'U'

            The upper triangular part of A is stored and A is factorized
            as PUDU^HP^T, where U is upper triangular.

        `uplo` = 'L'

            The lower triangular part of A is stored and A is factorized
            as PLDL^HP^T, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    a : complex, array-like, shape (`n`, `n`)
        The n by n original symmetric matrix A as supplied to
        ``zsytrf``.

    af : complex, array-like, shape (`n`, `n`)
        Details of the factorization of A, as returned by ``zsytrf``.

    ipiv : int, array-like, shape (`n`)
        Details of the interchanges and the block structure of D, as
        returned by ``zsytrf``.

    b : complex, array-like, shape (`n`, nrhs)
        The n by r right-hand side matrix B.

    x : complex, array-like, shape (`n`, nrhs)
        The n by r solution matrix X, as returned by ``zsytrs``.

    Returns
    -------
    x : complex, ndarray, shape (`n`, nrhs)
        The improved solution matrix X.

    ferr : float, ndarray, shape (nrhs)
        `ferr`[j-1] contains an estimated error bound for the jth
        solution vector, that is, the jth column of X, for j =
        1,2,...,r.

    berr : float, ndarray, shape (nrhs)
        `berr`[j-1] contains the component-wise backward error bound
        beta for the jth solution vector, that is, the jth column of X,
        for j = 1,2,...,r.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -3)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.zsyrfs'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
    )
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
    )
    _af = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'af'},
        af,
        copy=False,
    )
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
    )
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _b_lapacklin.zsyrfs(
        uplo,
        n,
        _nrhs,
        _a,
        _af,
        _ipiv,
        _b,
        _x,
        _ferr,
        _berr,
    )
    _x_py = _x
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'ferr', 'berr'],
    )(
        _x_py,
        _ferr_py,
        _berr_py,
    )
    return _return_data

def zsytri(uplo, a, ipiv):
    u"""
    Inverse of complex symmetric matrix, matrix already factorized by
    ``zsytrf``.

    ``zsytri`` computes the inverse of a complex symmetric matrix A,
    where A has been factorized by ``zsytrf``.

    For full information please refer to the NAG Library document for
    f07nw

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07nwf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = PUDU^HP^T, where U is upper triangular.

        `uplo` = 'L'

            A = PLDL^HP^T, where L is lower triangular.

    a : complex, array-like, shape (n, n)
        Details of the factorization of A, as returned by ``zsytrf``.

    ipiv : int, array-like, shape (n)
        Details of the interchanges and the block structure of D, as
        returned by ``zsytrf``.

    Returns
    -------
    a : complex, ndarray, shape (n, n)
        The factorization is overwritten by the n by n symmetric matrix
        A^-1.

        If `uplo` = 'U', the upper triangle of A^-1 is stored in the
        upper triangular part of the array.

        If `uplo` = 'L', the lower triangle of A^-1 is stored in the
        lower triangular part of the array.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. D is
            singular and the inverse of A cannot be computed.

    References
    ----------
    Du Croz, J J and Higham, N J, 1992, `Stability of methods for matrix
    inversion`, IMA J. Numer. Anal. (12), 1--19
    """

    fun_name = 'naginterfaces.library.lapacklin.zsytri'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    try:
        _b_lapacklin.zsytri(
            uplo,
            _n,
            _a,
            _ipiv,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _a_py = _a
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a'],
    )(
        _a_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _a_py

def dspsv(uplo, ap, b):
    u"""
    Computes the solution to a real symmetric system of linear
    equations, packed storage.

    ``dspsv`` computes the solution to a real system of linear equations

        AX = B,

    where A is an n by n symmetric matrix stored in packed format and X
    and B are n by r matrices.

    For full information please refer to the NAG Library document for
    f07pa

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07paf.html

    Parameters
    ----------
    uplo : str, length 1
        If `uplo` = 'U', the upper triangle of A is stored.

        If `uplo` = 'L', the lower triangle of A is stored.

    ap : float, array-like, shape (n*(n+1)/2)
        The n by n symmetric matrix A, packed by columns.

    b : float, array-like, shape (n, nrhs)
        The n by r right-hand side matrix B.

    Returns
    -------
    ap : float, ndarray, shape (n*(n+1)/2)
        The block diagonal matrix D and the multipliers used to obtain
        the factor U or L from the factorization A = UDU^T or A = LDL^T
        as computed by ``dsptrf``, stored as a packed triangular matrix
        in the same storage format as A.

    ipiv : int, ndarray, shape (n)
        Details of the interchanges and the block structure of D. More
        precisely,

            if `ipiv`[i-1] = k > 0, d_{ii} is a 1 by 1 pivot block and
            the ith row and column of A were interchanged with the kth
            row and column;

            if `uplo` = 'U' and `ipiv`[i-2] = `ipiv`[i-1] = -l < 0,
            [table omitted] is a 2 by 2 pivot block and the (i-1)th row
            and column of A were interchanged with the lth row and
            column;

            if `uplo` = 'L' and `ipiv`[i-1] = `ipiv`[i] = -m < 0, [table
            omitted] is a 2 by 2 pivot block and the (i+1)th row and
            column of A were interchanged with the mth row and column.

    b : float, ndarray, shape (n, nrhs)
        If no exception or warning is raised, the n by r solution matrix
        X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. The
            factorization has been completed, but the block diagonal
            matrix D is exactly singular, so the solution could not be
            computed.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    fun_name = 'naginterfaces.library.lapacklin.dspsv'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ipiv = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _b_lapacklin.dspsv(
            uplo,
            _n,
            _nrhs,
            _ap,
            _ipiv,
            _b,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _ap_py = _ap
    _b_py = _b
    _ipiv_py = _ipiv
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['ap', 'ipiv', 'b'],
    )(
        _ap_py,
        _ipiv_py,
        _b_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def dspsvx(fact, uplo, nrhs, ap, afp, ipiv, b):
    u"""
    Uses the diagonal pivoting factorization to compute the solution to
    a real symmetric system of linear equations, packed storage. Error
    bounds and a condition estimate are also computed.

    ``dspsvx`` uses the diagonal pivoting factorization

        A = UDU^T or A = LDL^T

    to compute the solution to a real system of linear equations

        AX = B,

    where A is an n by n symmetric matrix stored in packed format and X
    and B are n by r matrices.
    Error bounds on the solution and a condition estimate are also
    provided.

    For full information please refer to the NAG Library document for
    f07pb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07pbf.html

    Parameters
    ----------
    fact : str, length 1
        Specifies whether or not the factorized form of the matrix A has
        been supplied.

        `fact` = 'F'

            `afp` and `ipiv` contain the factorized form of the matrix
            A. `afp` and `ipiv` will not be modified.

        `fact` = 'N'

            The matrix A will be copied to `afp` and factorized.

    uplo : str, length 1
        If `uplo` = 'U', the upper triangle of A is stored.

        If `uplo` = 'L', the lower triangle of A is stored.

    nrhs : int
        r, the number of right-hand sides, i.e., the number of columns
        of the matrix B.

    ap : float, array-like, shape (n*(n+1)/2)
        The n by n symmetric matrix A, packed by columns.

    afp : float, array-like, shape (n*(n+1)/2)
        If `fact` = 'F', `afp` contains the block diagonal matrix D and
        the multipliers used to obtain the factor U or L from the
        factorization A = UDU^T or A = LDL^T as computed by ``dsptrf``,
        stored as a packed triangular matrix in the same storage format
        as A.

    ipiv : int, array-like, shape (n)
        If `fact` = 'F', `ipiv` contains details of the interchanges and
        the block structure of D, as determined by ``dsptrf``.

            if `ipiv`[i-1] = k > 0, d_{ii} is a 1 by 1 pivot block and
            the ith row and column of A were interchanged with the kth
            row and column;

            if `uplo` = 'U' and `ipiv`[i-2] = `ipiv`[i-1] = -l < 0,
            [table omitted] is a 2 by 2 pivot block and the (i-1)th row
            and column of A were interchanged with the lth row and
            column;

            if `uplo` = 'L' and `ipiv`[i-1] = `ipiv`[i] = -m < 0, [table
            omitted] is a 2 by 2 pivot block and the (i+1)th row and
            column of A were interchanged with the mth row and column.

    b : float, array-like, shape (n, `nrhs`)
        The n by r right-hand side matrix B.

    Returns
    -------
    afp : float, ndarray, shape (n*(n+1)/2)
        If `fact` = 'N', `afp` contains the block diagonal matrix D and
        the multipliers used to obtain the factor U or L from the
        factorization A = UDU^T or A = LDL^T as computed by ``dsptrf``,
        stored as a packed triangular matrix in the same storage format
        as A.

    ipiv : int, ndarray, shape (n)
        If `fact` = 'N', `ipiv` contains details of the interchanges and
        the block structure of D, as determined by ``dsptrf``, as
        described above.

    x : float, ndarray, shape (n, `nrhs`)
        If the function exits successfully or `errno` = `n` + 1, the n
        by r solution matrix X.

    rcond : float
        The estimate of the reciprocal condition number of the matrix A.
        If `rcond` = 0.0, the matrix may be exactly singular. This
        condition is indicated by `errno` in 1 ... `n`. Otherwise, if
        `rcond` is less than the machine precision, the matrix is
        singular to working precision. This condition is indicated by
        `errno` = `n` + 1.

    ferr : float, ndarray, shape (`nrhs`)
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the forward error bound for each computed solution
        vector, such that ||x^_j-x_j||_infinity/||x_j||_infinity <=
        `ferr`[j-1] where x^_j is the jth column of the computed
        solution returned in the array `x` and x_j is the corresponding
        column of the exact solution X. The estimate is as reliable as
        the estimate for `rcond`, and is almost always a slight
        overestimate of the true error.

    berr : float, ndarray, shape (`nrhs`)
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the component-wise relative backward error of each
        computed solution vector x^_j (i.e., the smallest relative
        change in any element of A or B that makes x^_j an exact
        solution).

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `fact`.

            Constraint: `fact` = 'F' or 'N'.

        (`errno` -2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -3)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -4)
            On entry, error in parameter `nrhs`.

            Constraint: `nrhs` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` (i > 0) and (i<=n))
            Element *<value>* of the diagonal is exactly zero. The
            factorization has been completed, but the factor D is
            exactly singular, so the solution and error bounds could not
            be computed. `rcond` = 0.0 is returned.

        (`errno` n+1)
            D is nonsingular, but `rcond` is less than machine
            precision, meaning that the matrix is singular to working
            precision. Nevertheless, the solution and error bounds are
            computed because there are a number of situations where the
            computed solution can be more accurate than the value of
            `rcond` would suggest.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    fun_name = 'naginterfaces.library.lapacklin.dspsvx'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_b, 'b'),),
    )
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=True,
    )
    _n_py = _b_utils._EngineIntArrayType.get_shape(_ipiv, {'fun_name': fun_name, 'entity': 'ipiv'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _afp = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'afp'},
        afp,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nrhs = _b_utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineIntScalarType.to_py(_nrhs)
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py, _nrhs_py),
        sorder=_m_sorder,
    )
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    try:
        _return_scalar = _b_lapacklin.dspsvx(
            fact,
            uplo,
            _n,
            _nrhs,
            _ap,
            _afp,
            _ipiv,
            _b,
            _x,
            _ferr,
            _berr,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _afp_py = _afp
    _ipiv_py = _ipiv
    _x_py = _x
    _rcond_py = _return_scalar
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['afp', 'ipiv', 'x', 'rcond', 'ferr', 'berr'],
    )(
        _afp_py,
        _ipiv_py,
        _x_py,
        _rcond_py,
        _ferr_py,
        _berr_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def dsptrf(uplo, n, ap):
    u"""
    Bunch--Kaufman factorization of real symmetric indefinite matrix,
    packed storage.

    ``dsptrf`` computes the Bunch--Kaufman factorization of a real
    symmetric indefinite matrix, using packed storage.

    For full information please refer to the NAG Library document for
    f07pd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07pdf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored and how A is to be factorized.

        `uplo` = 'U'

            The upper triangular part of A is stored and A is factorized
            as PUDU^TP^T, where U is upper triangular.

        `uplo` = 'L'

            The lower triangular part of A is stored and A is factorized
            as PLDL^TP^T, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    ap : float, array-like, shape (`n`*(`n`+1)/2)
        The n by n symmetric matrix A, packed by columns.

    Returns
    -------
    ap : float, ndarray, shape (`n`*(`n`+1)/2)
        A is overwritten by details of the block diagonal matrix D and
        the multipliers used to obtain the factor U or L as specified by
        `uplo`.

    ipiv : int, ndarray, shape (`n`)
        Details of the interchanges and the block structure of D. More
        precisely,

            if `ipiv`[i-1] = k > 0, d_{ii} is a 1 by 1 pivot block and
            the ith row and column of A were interchanged with the kth
            row and column;

            if `uplo` = 'U' and `ipiv`[i-2] = `ipiv`[i-1] = -l < 0,
            [table omitted] is a 2 by 2 pivot block and the (i-1)th row
            and column of A were interchanged with the lth row and
            column;

            if `uplo` = 'L' and `ipiv`[i-1] = `ipiv`[i] = -m < 0, [table
            omitted] is a 2 by 2 pivot block and the (i+1)th row and
            column of A were interchanged with the mth row and column.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. The
            factorization has been completed, but the block diagonal
            matrix D is exactly singular, and division by zero will
            occur if it is used to solve a system of equations.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.dsptrf'

    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _ipiv = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _b_lapacklin.dsptrf(
            uplo,
            _n,
            _ap,
            _ipiv,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _ap_py = _ap
    _ipiv_py = _ipiv
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['ap', 'ipiv'],
    )(
        _ap_py,
        _ipiv_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def dsptrs(uplo, n, ap, ipiv, b):
    u"""
    Solution of real symmetric indefinite system of linear equations,
    multiple right-hand sides, matrix already factorized by ``dsptrf``,
    packed storage.

    ``dsptrs`` solves a real symmetric indefinite system of linear
    equations with multiple right-hand sides,

        AX = B,

    where A has been factorized by ``dsptrf``, using packed storage.

    For full information please refer to the NAG Library document for
    f07pe

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07pef.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = PUDU^TP^T, where U is upper triangular.

        `uplo` = 'L'

            A = PLDL^TP^T, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    ap : float, array-like, shape (`n`*(`n`+1)/2)
        The factorization of A stored in packed form, as returned by
        ``dsptrf``.

    ipiv : int, array-like, shape (`n`)
        Details of the interchanges and the block structure of D, as
        returned by ``dsptrf``.

    b : float, array-like, shape (`n`, nrhs)
        The n by r right-hand side matrix B.

    Returns
    -------
    b : float, ndarray, shape (`n`, nrhs)
        The n by r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -3)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.dsptrs'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nrhs_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    _b_lapacklin.dsptrs(
        uplo,
        n,
        _nrhs,
        _ap,
        _ipiv,
        _b,
    )
    _b_py = _b
    return _b_py

def dspcon(uplo, n, ap, ipiv, anorm):
    u"""
    Estimate condition number of real symmetric indefinite matrix,
    matrix already factorized by ``dsptrf``, packed storage.

    ``dspcon`` estimates the condition number of a real symmetric
    indefinite matrix A, where A has been factorized by ``dsptrf``,
    using packed storage.

    For full information please refer to the NAG Library document for
    f07pg

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07pgf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = PUDU^TP^T, where U is upper triangular.

        `uplo` = 'L'

            A = PLDL^TP^T, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    ap : float, array-like, shape (`n`*(`n`+1)/2)
        The factorization of A stored in packed form, as returned by
        ``dsptrf``.

    ipiv : int, array-like, shape (`n`)
        Details of the interchanges and the block structure of D, as
        returned by ``dsptrf``.

    anorm : float
        The 1-norm of the **original** matrix A, which may be computed
        by calling ``blas.dlansp`` with its argument norm = '1'. `anorm`
        must be computed either **before** calling ``dsptrf`` or else
        from a **copy** of the original matrix A.

    Returns
    -------
    rcond : float
        An estimate of the reciprocal of the condition number of A.
        `rcond` is set to zero if exact singularity is detected or the
        estimate underflows. If `rcond` is less than machine precision,
        A is singular to working precision.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -5)
            On entry, error in parameter `anorm`.

            Constraint: `anorm` >= 0.0.

    References
    ----------
    Higham, N J, 1988, `FORTRAN codes for estimating the one-norm of a
    real or complex matrix, with applications to condition estimation`,
    ACM Trans. Math. Software (14), 381--396
    """

    fun_name = 'naginterfaces.library.lapacklin.dspcon'

    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    _return_scalar = _b_lapacklin.dspcon(
        uplo,
        n,
        _ap,
        _ipiv,
        anorm,
    )
    _rcond_py = _return_scalar
    return _rcond_py

def dsprfs(uplo, n, ap, afp, ipiv, b, x):
    u"""
    Refined solution with error bounds of real symmetric indefinite
    system of linear equations, multiple right-hand sides, packed
    storage.

    ``dsprfs`` returns error bounds for the solution of a real symmetric
    indefinite system of linear equations with multiple right-hand
    sides, AX = B, using packed storage.
    It improves the solution by iterative refinement, in order to reduce
    the backward error as much as possible.

    For full information please refer to the NAG Library document for
    f07ph

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07phf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored and how A is to be factorized.

        `uplo` = 'U'

            The upper triangular part of A is stored and A is factorized
            as PUDU^TP^T, where U is upper triangular.

        `uplo` = 'L'

            The lower triangular part of A is stored and A is factorized
            as PLDL^TP^T, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    ap : float, array-like, shape (`n`*(`n`+1)/2)
        The n by n original symmetric matrix A as supplied to
        ``dsptrf``.

    afp : float, array-like, shape (`n`*(`n`+1)/2)
        The factorization of A stored in packed form, as returned by
        ``dsptrf``.

    ipiv : int, array-like, shape (`n`)
        Details of the interchanges and the block structure of D, as
        returned by ``dsptrf``.

    b : float, array-like, shape (`n`, nrhs)
        The n by r right-hand side matrix B.

    x : float, array-like, shape (`n`, nrhs)
        The n by r solution matrix X, as returned by ``dsptrs``.

    Returns
    -------
    x : float, ndarray, shape (`n`, nrhs)
        The improved solution matrix X.

    ferr : float, ndarray, shape (nrhs)
        `ferr`[j-1] contains an estimated error bound for the jth
        solution vector, that is, the jth column of X, for j =
        1,2,...,r.

    berr : float, ndarray, shape (nrhs)
        `berr`[j-1] contains the component-wise backward error bound
        beta for the jth solution vector, that is, the jth column of X,
        for j = 1,2,...,r.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -3)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.dsprfs'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nrhs_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _afp = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'afp'},
        afp,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _b_lapacklin.dsprfs(
        uplo,
        n,
        _nrhs,
        _ap,
        _afp,
        _ipiv,
        _b,
        _x,
        _ferr,
        _berr,
    )
    _x_py = _x
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'ferr', 'berr'],
    )(
        _x_py,
        _ferr_py,
        _berr_py,
    )
    return _return_data

def dsptri(uplo, n, ap, ipiv):
    u"""
    Inverse of real symmetric indefinite matrix, matrix already
    factorized by ``dsptrf``, packed storage.

    ``dsptri`` computes the inverse of a real symmetric indefinite
    matrix A, where A has been factorized by ``dsptrf``, using packed
    storage.

    For full information please refer to the NAG Library document for
    f07pj

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07pjf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = PUDU^TP^T, where U is upper triangular.

        `uplo` = 'L'

            A = PLDL^TP^T, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    ap : float, array-like, shape (`n`*(`n`+1)/2)
        The factorization of A stored in packed form, as returned by
        ``dsptrf``.

    ipiv : int, array-like, shape (`n`)
        Details of the interchanges and the block structure of D, as
        returned by ``dsptrf``.

    Returns
    -------
    ap : float, ndarray, shape (`n`*(`n`+1)/2)
        The factorization is overwritten by the n by n matrix A^-1.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. D is
            singular and the inverse of A cannot be computed.

    References
    ----------
    Du Croz, J J and Higham, N J, 1992, `Stability of methods for matrix
    inversion`, IMA J. Numer. Anal. (12), 1--19
    """

    fun_name = 'naginterfaces.library.lapacklin.dsptri'

    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    try:
        _b_lapacklin.dsptri(
            uplo,
            n,
            _ap,
            _ipiv,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _ap_py = _ap
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['ap'],
    )(
        _ap_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _ap_py

def zhpsv(uplo, ap, b):
    u"""
    Computes the solution to a complex Hermitian system of linear
    equations, packed storage.

    ``zhpsv`` computes the solution to a complex system of linear
    equations

        AX = B,

    where A is an n by n Hermitian matrix stored in packed format and X
    and B are n by r matrices.

    For full information please refer to the NAG Library document for
    f07pn

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07pnf.html

    Parameters
    ----------
    uplo : str, length 1
        If `uplo` = 'U', the upper triangle of A is stored.

        If `uplo` = 'L', the lower triangle of A is stored.

    ap : complex, array-like, shape (n*(n+1)/2)
        The n by n Hermitian matrix A, packed by columns.

    b : complex, array-like, shape (n, nrhs)
        The n by r right-hand side matrix B.

    Returns
    -------
    ap : complex, ndarray, shape (n*(n+1)/2)
        The block diagonal matrix D and the multipliers used to obtain
        the factor U or L from the factorization A = UDU^H or A = LDL^H
        as computed by ``zhptrf``, stored as a packed triangular matrix
        in the same storage format as A.

    ipiv : int, ndarray, shape (n)
        Details of the interchanges and the block structure of D. More
        precisely,

            if `ipiv`[i-1] = k > 0, d_{ii} is a 1 by 1 pivot block and
            the ith row and column of A were interchanged with the kth
            row and column;

            if `uplo` = 'U' and `ipiv`[i-2] = `ipiv`[i-1] = -l < 0,
            [table omitted] is a 2 by 2 pivot block and the (i-1)th row
            and column of A were interchanged with the lth row and
            column;

            if `uplo` = 'L' and `ipiv`[i-1] = `ipiv`[i] = -m < 0, [table
            omitted] is a 2 by 2 pivot block and the (i+1)th row and
            column of A were interchanged with the mth row and column.

    b : complex, ndarray, shape (n, nrhs)
        If no exception or warning is raised, the n by r solution matrix
        X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. The
            factorization has been completed, but the block diagonal
            matrix D is exactly singular, so the solution could not be
            computed.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    fun_name = 'naginterfaces.library.lapacklin.zhpsv'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=True,
    )
    _ipiv = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _b_lapacklin.zhpsv(
            uplo,
            _n,
            _nrhs,
            _ap,
            _ipiv,
            _b,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _ap_py = _ap
    _b_py = _b
    _ipiv_py = _ipiv
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['ap', 'ipiv', 'b'],
    )(
        _ap_py,
        _ipiv_py,
        _b_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def zhpsvx(fact, uplo, nrhs, ap, afp, ipiv, b):
    u"""
    Uses the diagonal pivoting factorization to compute the solution to
    a complex, Hermitian, system of linear equations, error bounds and
    condition estimates. Packed storage.

    ``zhpsvx`` uses the diagonal pivoting factorization

        A = UDU^H or A = LDL^H

    to compute the solution to a complex system of linear equations

        AX = B,

    where A is an n by n Hermitian matrix stored in packed format and X
    and B are n by r matrices.
    Error bounds on the solution and a condition estimate are also
    provided.

    For full information please refer to the NAG Library document for
    f07pp

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07ppf.html

    Parameters
    ----------
    fact : str, length 1
        Specifies whether or not the factorized form of the matrix A has
        been supplied.

        `fact` = 'F'

            `afp` and `ipiv` contain the factorized form of the matrix
            A. `afp` and `ipiv` will not be modified.

        `fact` = 'N'

            The matrix A will be copied to `afp` and factorized.

    uplo : str, length 1
        If `uplo` = 'U', the upper triangle of A is stored.

        If `uplo` = 'L', the lower triangle of A is stored.

    nrhs : int
        r, the number of right-hand sides, i.e., the number of columns
        of the matrix B.

    ap : complex, array-like, shape (n*(n+1)/2)
        The n by n Hermitian matrix A, packed by columns.

    afp : complex, array-like, shape (n*(n+1)/2)
        If `fact` = 'F', `afp` contains the block diagonal matrix D and
        the multipliers used to obtain the factor U or L from the
        factorization A = UDU^H or A = LDL^H as computed by ``zhptrf``,
        stored as a packed triangular matrix in the same storage format
        as A.

    ipiv : int, array-like, shape (n)
        If `fact` = 'F', `ipiv` contains details of the interchanges and
        the block structure of D, as determined by ``zhptrf``.

            if `ipiv`[i-1] = k > 0, d_{ii} is a 1 by 1 pivot block and
            the ith row and column of A were interchanged with the kth
            row and column;

            if `uplo` = 'U' and `ipiv`[i-2] = `ipiv`[i-1] = -l < 0,
            [table omitted] is a 2 by 2 pivot block and the (i-1)th row
            and column of A were interchanged with the lth row and
            column;

            if `uplo` = 'L' and `ipiv`[i-1] = `ipiv`[i] = -m < 0, [table
            omitted] is a 2 by 2 pivot block and the (i+1)th row and
            column of A were interchanged with the mth row and column.

    b : complex, array-like, shape (n, `nrhs`)
        The n by r right-hand side matrix B.

    Returns
    -------
    afp : complex, ndarray, shape (n*(n+1)/2)
        If `fact` = 'N', `afp` contains the block diagonal matrix D and
        the multipliers used to obtain the factor U or L from the
        factorization A = UDU^H or A = LDL^H as computed by ``zhptrf``,
        stored as a packed triangular matrix in the same storage format
        as A.

    ipiv : int, ndarray, shape (n)
        If `fact` = 'N', `ipiv` contains details of the interchanges and
        the block structure of D, as determined by ``zhptrf``, as
        described above.

    x : complex, ndarray, shape (n, `nrhs`)
        If the function exits successfully or `errno` = `n` + 1, the n
        by r solution matrix X.

    rcond : float
        The estimate of the reciprocal condition number of the matrix A.
        If `rcond` = 0.0, the matrix may be exactly singular. This
        condition is indicated by `errno` in 1 ... `n`. Otherwise, if
        `rcond` is less than the machine precision, the matrix is
        singular to working precision. This condition is indicated by
        `errno` = `n` + 1.

    ferr : float, ndarray, shape (`nrhs`)
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the forward error bound for each computed solution
        vector, such that ||x^_j-x_j||_infinity/||x_j||_infinity <=
        `ferr`[j-1] where x^_j is the jth column of the computed
        solution returned in the array `x` and x_j is the corresponding
        column of the exact solution X. The estimate is as reliable as
        the estimate for `rcond`, and is almost always a slight
        overestimate of the true error.

    berr : float, ndarray, shape (`nrhs`)
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the component-wise relative backward error of each
        computed solution vector x^_j (i.e., the smallest relative
        change in any element of A or B that makes x^_j an exact
        solution).

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `fact`.

            Constraint: `fact` = 'F' or 'N'.

        (`errno` -2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -3)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -4)
            On entry, error in parameter `nrhs`.

            Constraint: `nrhs` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` (i > 0) and (i<=n))
            Element *<value>* of the diagonal is exactly zero. The
            factorization has been completed, but the factor D is
            exactly singular, so the solution and error bounds could not
            be computed. `rcond` = 0.0 is returned.

        (`errno` n+1)
            D is nonsingular, but `rcond` is less than machine
            precision, meaning that the matrix is singular to working
            precision. Nevertheless, the solution and error bounds are
            computed because there are a number of situations where the
            computed solution can be more accurate than the value of
            `rcond` would suggest.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    fun_name = 'naginterfaces.library.lapacklin.zhpsvx'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_b, 'b'),),
    )
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=True,
    )
    _n_py = _b_utils._EngineIntArrayType.get_shape(_ipiv, {'fun_name': fun_name, 'entity': 'ipiv'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=False,
    )
    _afp = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'afp'},
        afp,
        copy=True,
    )
    _nrhs = _b_utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineIntScalarType.to_py(_nrhs)
    _x = _b_utils._EngineComplex128ArrayType.empty_ndarray(
        (_n_py, _nrhs_py),
        sorder=_m_sorder,
    )
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    try:
        _return_scalar = _b_lapacklin.zhpsvx(
            fact,
            uplo,
            _n,
            _nrhs,
            _ap,
            _afp,
            _ipiv,
            _b,
            _x,
            _ferr,
            _berr,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _afp_py = _afp
    _ipiv_py = _ipiv
    _x_py = _x
    _rcond_py = _return_scalar
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['afp', 'ipiv', 'x', 'rcond', 'ferr', 'berr'],
    )(
        _afp_py,
        _ipiv_py,
        _x_py,
        _rcond_py,
        _ferr_py,
        _berr_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def zhptrf(uplo, n, ap):
    u"""
    Bunch--Kaufman factorization of complex Hermitian indefinite matrix,
    packed storage.

    ``zhptrf`` computes the Bunch--Kaufman factorization of a complex
    Hermitian indefinite matrix, using packed storage.

    For full information please refer to the NAG Library document for
    f07pr

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07prf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored and how A is to be factorized.

        `uplo` = 'U'

            The upper triangular part of A is stored and A is factorized
            as PUDU^HP^T, where U is upper triangular.

        `uplo` = 'L'

            The lower triangular part of A is stored and A is factorized
            as PLDL^HP^T, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    ap : complex, array-like, shape (`n`*(`n`+1)/2)
        The n by n Hermitian matrix A, packed by columns.

    Returns
    -------
    ap : complex, ndarray, shape (`n`*(`n`+1)/2)
        A is overwritten by details of the block diagonal matrix D and
        the multipliers used to obtain the factor U or L as specified by
        `uplo`.

    ipiv : int, ndarray, shape (`n`)
        Details of the interchanges and the block structure of D. More
        precisely,

            if `ipiv`[i-1] = k > 0, d_{ii} is a 1 by 1 pivot block and
            the ith row and column of A were interchanged with the kth
            row and column;

            if `uplo` = 'U' and `ipiv`[i-2] = `ipiv`[i-1] = -l < 0,
            [table omitted] is a 2 by 2 pivot block and the (i-1)th row
            and column of A were interchanged with the lth row and
            column;

            if `uplo` = 'L' and `ipiv`[i-1] = `ipiv`[i] = -m < 0, [table
            omitted] is a 2 by 2 pivot block and the (i+1)th row and
            column of A were interchanged with the mth row and column.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. The
            factorization has been completed, but the block diagonal
            matrix D is exactly singular, and division by zero will
            occur if it is used to solve a system of equations.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.zhptrf'

    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=True,
    )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _ipiv = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _b_lapacklin.zhptrf(
            uplo,
            _n,
            _ap,
            _ipiv,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _ap_py = _ap
    _ipiv_py = _ipiv
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['ap', 'ipiv'],
    )(
        _ap_py,
        _ipiv_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def zhptrs(uplo, n, ap, ipiv, b):
    u"""
    Solution of complex Hermitian indefinite system of linear equations,
    multiple right-hand sides, matrix already factorized by ``zhptrf``,
    packed storage.

    ``zhptrs`` solves a complex Hermitian indefinite system of linear
    equations with multiple right-hand sides,

        AX = B,

    where A has been factorized by ``zhptrf``, using packed storage.

    For full information please refer to the NAG Library document for
    f07ps

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07psf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = PUDU^HP^T, where U is upper triangular.

        `uplo` = 'L'

            A = PLDL^HP^T, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    ap : complex, array-like, shape (`n`*(`n`+1)/2)
        The factorization of A stored in packed form, as returned by
        ``zhptrf``.

    ipiv : int, array-like, shape (`n`)
        Details of the interchanges and the block structure of D, as
        returned by ``zhptrf``.

    b : complex, array-like, shape (`n`, nrhs)
        The n by r right-hand side matrix B.

    Returns
    -------
    b : complex, ndarray, shape (`n`, nrhs)
        The n by r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -3)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.zhptrs'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
    )
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=False,
    )
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    _b_lapacklin.zhptrs(
        uplo,
        n,
        _nrhs,
        _ap,
        _ipiv,
        _b,
    )
    _b_py = _b
    return _b_py

def zhpcon(uplo, ap, ipiv, anorm):
    u"""
    Estimate condition number of complex Hermitian indefinite matrix,
    matrix already factorized by ``zhptrf``, packed storage.

    ``zhpcon`` estimates the condition number of a complex Hermitian
    indefinite matrix A, where A has been factorized by ``zhptrf``,
    using packed storage.

    For full information please refer to the NAG Library document for
    f07pu

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07puf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = PUDU^HP^T, where U is upper triangular.

        `uplo` = 'L'

            A = PLDL^HP^T, where L is lower triangular.

    ap : complex, array-like, shape (n*(n+1)/2)
        The factorization of A stored in packed form, as returned by
        ``zhptrf``.

    ipiv : int, array-like, shape (n)
        Details of the interchanges and the block structure of D, as
        returned by ``zhptrf``.

    anorm : float
        The 1-norm of the **original** matrix A, which may be computed
        by calling ``blas.zlanhp`` with its argument norm = '1'. `anorm`
        must be computed either **before** calling ``zhptrf`` or else
        from a **copy** of the original matrix A.

    Returns
    -------
    rcond : float
        An estimate of the reciprocal of the condition number of A.
        `rcond` is set to zero if exact singularity is detected or the
        estimate underflows. If `rcond` is less than machine precision,
        A is singular to working precision.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -5)
            On entry, error in parameter `anorm`.

            Constraint: `anorm` >= 0.0.

    References
    ----------
    Higham, N J, 1988, `FORTRAN codes for estimating the one-norm of a
    real or complex matrix, with applications to condition estimation`,
    ACM Trans. Math. Software (14), 381--396
    """

    fun_name = 'naginterfaces.library.lapacklin.zhpcon'

    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    _n_py = _b_utils._EngineIntArrayType.get_shape(_ipiv, {'fun_name': fun_name, 'entity': 'ipiv'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=False,
    )
    _return_scalar = _b_lapacklin.zhpcon(
        uplo,
        _n,
        _ap,
        _ipiv,
        anorm,
    )
    _rcond_py = _return_scalar
    return _rcond_py

def zhprfs(uplo, n, ap, afp, ipiv, b, x):
    u"""
    Refined solution with error bounds of complex Hermitian indefinite
    system of linear equations, multiple right-hand sides, packed
    storage.

    ``zhprfs`` returns error bounds for the solution of a complex
    Hermitian indefinite system of linear equations with multiple
    right-hand sides, AX = B, using packed storage.
    It improves the solution by iterative refinement, in order to reduce
    the backward error as much as possible.

    For full information please refer to the NAG Library document for
    f07pv

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07pvf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored and how A is to be factorized.

        `uplo` = 'U'

            The upper triangular part of A is stored and A is factorized
            as PUDU^HP^T, where U is upper triangular.

        `uplo` = 'L'

            The lower triangular part of A is stored and A is factorized
            as PLDL^HP^T, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    ap : complex, array-like, shape (`n`*(`n`+1)/2)
        The n by n original Hermitian matrix A as supplied to
        ``zhptrf``.

    afp : complex, array-like, shape (`n`*(`n`+1)/2)
        The factorization of A stored in packed form, as returned by
        ``zhptrf``.

    ipiv : int, array-like, shape (`n`)
        Details of the interchanges and the block structure of D, as
        returned by ``zhptrf``.

    b : complex, array-like, shape (`n`, nrhs)
        The n by r right-hand side matrix B.

    x : complex, array-like, shape (`n`, nrhs)
        The n by r solution matrix X, as returned by ``zhptrs``.

    Returns
    -------
    x : complex, ndarray, shape (`n`, nrhs)
        The improved solution matrix X.

    ferr : float, ndarray, shape (nrhs)
        `ferr`[j-1] contains an estimated error bound for the jth
        solution vector, that is, the jth column of X, for j =
        1,2,...,r.

    berr : float, ndarray, shape (nrhs)
        `berr`[j-1] contains the component-wise backward error bound
        beta for the jth solution vector, that is, the jth column of X,
        for j = 1,2,...,r.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -3)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.zhprfs'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
    )
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=False,
    )
    _afp = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'afp'},
        afp,
        copy=False,
    )
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
    )
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _b_lapacklin.zhprfs(
        uplo,
        n,
        _nrhs,
        _ap,
        _afp,
        _ipiv,
        _b,
        _x,
        _ferr,
        _berr,
    )
    _x_py = _x
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'ferr', 'berr'],
    )(
        _x_py,
        _ferr_py,
        _berr_py,
    )
    return _return_data

def zhptri(uplo, n, ap, ipiv):
    u"""
    Inverse of complex Hermitian indefinite matrix, matrix already
    factorized by ``zhptrf``, packed storage.

    ``zhptri`` computes the inverse of a complex Hermitian indefinite
    matrix A, where A has been factorized by ``zhptrf``, using packed
    storage.

    For full information please refer to the NAG Library document for
    f07pw

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07pwf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = PUDU^HP^T, where U is upper triangular.

        `uplo` = 'L'

            A = PLDL^HP^T, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    ap : complex, array-like, shape (`n`*(`n`+1)/2)
        The factorization of A stored in packed form, as returned by
        ``zhptrf``.

    ipiv : int, array-like, shape (`n`)
        Details of the interchanges and the block structure of D, as
        returned by ``zhptrf``.

    Returns
    -------
    ap : complex, ndarray, shape (`n`*(`n`+1)/2)
        The factorization is overwritten by the n by n matrix A^-1.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. D is
            singular and the inverse of A cannot be computed.

    References
    ----------
    Du Croz, J J and Higham, N J, 1992, `Stability of methods for matrix
    inversion`, IMA J. Numer. Anal. (12), 1--19
    """

    fun_name = 'naginterfaces.library.lapacklin.zhptri'

    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=True,
    )
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    try:
        _b_lapacklin.zhptri(
            uplo,
            n,
            _ap,
            _ipiv,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _ap_py = _ap
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['ap'],
    )(
        _ap_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _ap_py

def zspsv(uplo, ap, b):
    u"""
    Computes the solution to a complex symmetric system of linear
    equations, packed storage.

    ``zspsv`` computes the solution to a complex system of linear
    equations

        AX = B,

    where A is an n by n symmetric matrix stored in packed format and X
    and B are n by r matrices.

    For full information please refer to the NAG Library document for
    f07qn

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07qnf.html

    Parameters
    ----------
    uplo : str, length 1
        If `uplo` = 'U', the upper triangle of A is stored.

        If `uplo` = 'L', the lower triangle of A is stored.

    ap : complex, array-like, shape (n*(n+1)/2)
        The n by n symmetric matrix A, packed by columns.

    b : complex, array-like, shape (n, nrhs)
        The n by r right-hand side matrix B.

    Returns
    -------
    ap : complex, ndarray, shape (n*(n+1)/2)
        The block diagonal matrix D and the multipliers used to obtain
        the factor U or L from the factorization A = UDU^H or A = LDL^H
        as computed by ``zsptrf``, stored as a packed triangular matrix
        in the same storage format as A.

    ipiv : int, ndarray, shape (n)
        Details of the interchanges and the block structure of D. More
        precisely,

            if `ipiv`[i-1] = k > 0, d_{ii} is a 1 by 1 pivot block and
            the ith row and column of A were interchanged with the kth
            row and column;

            if `uplo` = 'U' and `ipiv`[i-2] = `ipiv`[i-1] = -l < 0,
            [table omitted] is a 2 by 2 pivot block and the (i-1)th row
            and column of A were interchanged with the lth row and
            column;

            if `uplo` = 'L' and `ipiv`[i-1] = `ipiv`[i] = -m < 0, [table
            omitted] is a 2 by 2 pivot block and the (i+1)th row and
            column of A were interchanged with the mth row and column.

    b : complex, ndarray, shape (n, nrhs)
        If no exception or warning is raised, the n by r solution matrix
        X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -3)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. The
            factorization has been completed, but the block diagonal
            matrix D is exactly singular, so the solution could not be
            computed.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    fun_name = 'naginterfaces.library.lapacklin.zspsv'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=True,
    )
    _ipiv = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _b_lapacklin.zspsv(
            uplo,
            _n,
            _nrhs,
            _ap,
            _ipiv,
            _b,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _ap_py = _ap
    _b_py = _b
    _ipiv_py = _ipiv
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['ap', 'ipiv', 'b'],
    )(
        _ap_py,
        _ipiv_py,
        _b_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def zspsvx(fact, uplo, nrhs, ap, afp, ipiv, b):
    u"""
    Uses the diagonal pivoting factorization to compute the solution to
    a complex, symmetric, system of linear equations, error bounds and
    condition estimates. Packed storage.

    ``zspsvx`` uses the diagonal pivoting factorization

        A = UDU^T or A = LDL^T

    to compute the solution to a complex system of linear equations

        AX = B,

    where A is an n by n symmetric matrix stored in packed format and X
    and B are n by r matrices.
    Error bounds on the solution and a condition estimate are also
    provided.

    For full information please refer to the NAG Library document for
    f07qp

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07qpf.html

    Parameters
    ----------
    fact : str, length 1
        Specifies whether or not the factorized form of the matrix A has
        been supplied.

        `fact` = 'F'

            `afp` and `ipiv` contain the factorized form of the matrix
            A. `afp` and `ipiv` will not be modified.

        `fact` = 'N'

            The matrix A will be copied to `afp` and factorized.

    uplo : str, length 1
        If `uplo` = 'U', the upper triangle of A is stored.

        If `uplo` = 'L', the lower triangle of A is stored.

    nrhs : int
        r, the number of right-hand sides, i.e., the number of columns
        of the matrix B.

    ap : complex, array-like, shape (n*(n+1)/2)
        The n by n symmetric matrix A, packed by columns.

    afp : complex, array-like, shape (n*(n+1)/2)
        If `fact` = 'F', `afp` contains the block diagonal matrix D and
        the multipliers used to obtain the factor U or L from the
        factorization A = UDU^H or A = LDL^H as computed by ``zsptrf``,
        stored as a packed triangular matrix in the same storage format
        as A.

    ipiv : int, array-like, shape (n)
        If `fact` = 'F', `ipiv` contains details of the interchanges and
        the block structure of D, as determined by ``zsptrf``.

            if `ipiv`[i-1] = k > 0, d_{ii} is a 1 by 1 pivot block and
            the ith row and column of A were interchanged with the kth
            row and column;

            if `uplo` = 'U' and `ipiv`[i-2] = `ipiv`[i-1] = -l < 0,
            [table omitted] is a 2 by 2 pivot block and the (i-1)th row
            and column of A were interchanged with the lth row and
            column;

            if `uplo` = 'L' and `ipiv`[i-1] = `ipiv`[i] = -m < 0, [table
            omitted] is a 2 by 2 pivot block and the (i+1)th row and
            column of A were interchanged with the mth row and column.

    b : complex, array-like, shape (n, `nrhs`)
        The n by r right-hand side matrix B.

    Returns
    -------
    afp : complex, ndarray, shape (n*(n+1)/2)
        If `fact` = 'N', `afp` contains the block diagonal matrix D and
        the multipliers used to obtain the factor U or L from the
        factorization A = UDU^H or A = LDL^H as computed by ``zsptrf``,
        stored as a packed triangular matrix in the same storage format
        as A.

    ipiv : int, ndarray, shape (n)
        If `fact` = 'N', `ipiv` contains details of the interchanges and
        the block structure of D, as determined by ``zsptrf``, as
        described above.

    x : complex, ndarray, shape (n, `nrhs`)
        If the function exits successfully or `errno` = `n` + 1, the n
        by r solution matrix X.

    rcond : float
        The estimate of the reciprocal condition number of the matrix A.
        If `rcond` = 0.0, the matrix may be exactly singular. This
        condition is indicated by `errno` in 1 ... `n`. Otherwise, if
        `rcond` is less than the machine precision, the matrix is
        singular to working precision. This condition is indicated by
        `errno` = `n` + 1.

    ferr : float, ndarray, shape (`nrhs`)
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the forward error bound for each computed solution
        vector, such that ||x^_j-x_j||_infinity/||x_j||_infinity <=
        `ferr`[j-1] where x^_j is the jth column of the computed
        solution returned in the array `x` and x_j is the corresponding
        column of the exact solution X. The estimate is as reliable as
        the estimate for `rcond`, and is almost always a slight
        overestimate of the true error.

    berr : float, ndarray, shape (`nrhs`)
        If the function exits successfully or `errno` = `n` + 1, an
        estimate of the component-wise relative backward error of each
        computed solution vector x^_j (i.e., the smallest relative
        change in any element of A or B that makes x^_j an exact
        solution).

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `fact`.

            Constraint: `fact` = 'F' or 'N'.

        (`errno` -2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -3)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -4)
            On entry, error in parameter `nrhs`.

            Constraint: `nrhs` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` (i > 0) and (i<=n))
            Element *<value>* of the diagonal is exactly zero. The
            factorization has been completed, but the factor D is
            exactly singular, so the solution and error bounds could not
            be computed. `rcond` = 0.0 is returned.

        (`errno` n+1)
            D is nonsingular, but `rcond` is less than machine
            precision, meaning that the matrix is singular to working
            precision. Nevertheless, the solution and error bounds are
            computed because there are a number of situations where the
            computed solution can be more accurate than the value of
            `rcond` would suggest.

    References
    ----------
    Anderson, E, Bai, Z, Bischof, C, Blackford, S, Demmel, J, Dongarra,
    J J, Du Croz, J J, Greenbaum, A, Hammarling, S, McKenney, A and
    Sorensen, D, 1999, `LAPACK Users' Guide`, (3rd Edition), SIAM,
    Philadelphia, https://www.netlib.org/lapack/lug

    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Higham, N J, 2002, `Accuracy and Stability of Numerical Algorithms`,
    (2nd Edition), SIAM, Philadelphia
    """

    fun_name = 'naginterfaces.library.lapacklin.zspsvx'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_b, 'b'),),
    )
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=True,
    )
    _n_py = _b_utils._EngineIntArrayType.get_shape(_ipiv, {'fun_name': fun_name, 'entity': 'ipiv'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=False,
    )
    _afp = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'afp'},
        afp,
        copy=True,
    )
    _nrhs = _b_utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineIntScalarType.to_py(_nrhs)
    _x = _b_utils._EngineComplex128ArrayType.empty_ndarray(
        (_n_py, _nrhs_py),
        sorder=_m_sorder,
    )
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    try:
        _return_scalar = _b_lapacklin.zspsvx(
            fact,
            uplo,
            _n,
            _nrhs,
            _ap,
            _afp,
            _ipiv,
            _b,
            _x,
            _ferr,
            _berr,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _afp_py = _afp
    _ipiv_py = _ipiv
    _x_py = _x
    _rcond_py = _return_scalar
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['afp', 'ipiv', 'x', 'rcond', 'ferr', 'berr'],
    )(
        _afp_py,
        _ipiv_py,
        _x_py,
        _rcond_py,
        _ferr_py,
        _berr_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def zsptrf(uplo, n, ap):
    u"""
    Bunch--Kaufman factorization of complex symmetric matrix, packed
    storage.

    ``zsptrf`` computes the Bunch--Kaufman factorization of a complex
    symmetric matrix, using packed storage.

    For full information please refer to the NAG Library document for
    f07qr

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07qrf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored and how A is to be factorized.

        `uplo` = 'U'

            The upper triangular part of A is stored and A is factorized
            as PUDU^HP^T, where U is upper triangular.

        `uplo` = 'L'

            The lower triangular part of A is stored and A is factorized
            as PLDL^HP^T, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    ap : complex, array-like, shape (`n`*(`n`+1)/2)
        The n by n symmetric matrix A, packed by columns.

    Returns
    -------
    ap : complex, ndarray, shape (`n`*(`n`+1)/2)
        A is overwritten by details of the block diagonal matrix D and
        the multipliers used to obtain the factor U or L as specified by
        `uplo`.

    ipiv : int, ndarray, shape (`n`)
        Details of the interchanges and the block structure of D. More
        precisely,

            if `ipiv`[i-1] = k > 0, d_{ii} is a 1 by 1 pivot block and
            the ith row and column of A were interchanged with the kth
            row and column;

            if `uplo` = 'U' and `ipiv`[i-2] = `ipiv`[i-1] = -l < 0,
            [table omitted] is a 2 by 2 pivot block and the (i-1)th row
            and column of A were interchanged with the lth row and
            column;

            if `uplo` = 'L' and `ipiv`[i-1] = `ipiv`[i] = -m < 0, [table
            omitted] is a 2 by 2 pivot block and the (i+1)th row and
            column of A were interchanged with the mth row and column.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. The
            factorization has been completed, but the block diagonal
            matrix D is exactly singular, and division by zero will
            occur if it is used to solve a system of equations.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.zsptrf'

    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=True,
    )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _ipiv = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _b_lapacklin.zsptrf(
            uplo,
            _n,
            _ap,
            _ipiv,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _ap_py = _ap
    _ipiv_py = _ipiv
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['ap', 'ipiv'],
    )(
        _ap_py,
        _ipiv_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def zsptrs(uplo, n, ap, ipiv, b):
    u"""
    Solution of complex symmetric system of linear equations, multiple
    right-hand sides, matrix already factorized by ``zsptrf``, packed
    storage.

    ``zsptrs`` solves a complex symmetric system of linear equations
    with multiple right-hand sides,

        AX = B,

    where A has been factorized by ``zsptrf``, using packed storage.

    For full information please refer to the NAG Library document for
    f07qs

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07qsf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = PUDU^HP^T, where U is upper triangular.

        `uplo` = 'L'

            A = PLDL^HP^T, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    ap : complex, array-like, shape (`n`*(`n`+1)/2)
        The factorization of A stored in packed form, as returned by
        ``zsptrf``.

    ipiv : int, array-like, shape (`n`)
        Details of the interchanges and the block structure of D, as
        returned by ``zsptrf``.

    b : complex, array-like, shape (`n`, nrhs)
        The n by r right-hand side matrix B.

    Returns
    -------
    b : complex, ndarray, shape (`n`, nrhs)
        The n by r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -3)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.zsptrs'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
    )
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=False,
    )
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    _b_lapacklin.zsptrs(
        uplo,
        n,
        _nrhs,
        _ap,
        _ipiv,
        _b,
    )
    _b_py = _b
    return _b_py

def zspcon(uplo, ap, ipiv, anorm):
    u"""
    Estimate condition number of complex symmetric matrix, matrix
    already factorized by ``zsptrf``, packed storage.

    ``zspcon`` estimates the condition number of a complex symmetric
    matrix A, where A has been factorized by ``zsptrf``, using packed
    storage.

    For full information please refer to the NAG Library document for
    f07qu

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07quf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = PUDU^HP^T, where U is upper triangular.

        `uplo` = 'L'

            A = PLDL^HP^T, where L is lower triangular.

    ap : complex, array-like, shape (n*(n+1)/2)
        The factorization of A stored in packed form, as returned by
        ``zsptrf``.

    ipiv : int, array-like, shape (n)
        Details of the interchanges and the block structure of D, as
        returned by ``zsptrf``.

    anorm : float
        The 1-norm of the **original** matrix A, which may be computed
        by calling ``blas.zlansp`` with its argument norm = '1'. `anorm`
        must be computed either **before** calling ``zsptrf`` or else
        from a **copy** of the original matrix A.

    Returns
    -------
    rcond : float
        An estimate of the reciprocal of the condition number of A.
        `rcond` is set to zero if exact singularity is detected or the
        estimate underflows. If `rcond` is less than machine precision,
        A is singular to working precision.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -5)
            On entry, error in parameter `anorm`.

            Constraint: `anorm` >= 0.0.

    References
    ----------
    Higham, N J, 1988, `FORTRAN codes for estimating the one-norm of a
    real or complex matrix, with applications to condition estimation`,
    ACM Trans. Math. Software (14), 381--396
    """

    fun_name = 'naginterfaces.library.lapacklin.zspcon'

    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    _n_py = _b_utils._EngineIntArrayType.get_shape(_ipiv, {'fun_name': fun_name, 'entity': 'ipiv'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=False,
    )
    _return_scalar = _b_lapacklin.zspcon(
        uplo,
        _n,
        _ap,
        _ipiv,
        anorm,
    )
    _rcond_py = _return_scalar
    return _rcond_py

def zsprfs(uplo, n, ap, afp, ipiv, b, x):
    u"""
    Refined solution with error bounds of complex symmetric system of
    linear equations, multiple right-hand sides, packed storage.

    ``zsprfs`` returns error bounds for the solution of a complex
    symmetric system of linear equations with multiple right-hand sides,
    AX = B, using packed storage.
    It improves the solution by iterative refinement, in order to reduce
    the backward error as much as possible.

    For full information please refer to the NAG Library document for
    f07qv

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07qvf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored and how A is to be factorized.

        `uplo` = 'U'

            The upper triangular part of A is stored and A is factorized
            as PUDU^HP^T, where U is upper triangular.

        `uplo` = 'L'

            The lower triangular part of A is stored and A is factorized
            as PLDL^HP^T, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    ap : complex, array-like, shape (`n`*(`n`+1)/2)
        The n by n original symmetric matrix A as supplied to
        ``zsptrf``.

    afp : complex, array-like, shape (`n`*(`n`+1)/2)
        The factorization of A stored in packed form, as returned by
        ``zsptrf``.

    ipiv : int, array-like, shape (`n`)
        Details of the interchanges and the block structure of D, as
        returned by ``zsptrf``.

    b : complex, array-like, shape (`n`, nrhs)
        The n by r right-hand side matrix B.

    x : complex, array-like, shape (`n`, nrhs)
        The n by r solution matrix X, as returned by ``zsptrs``.

    Returns
    -------
    x : complex, ndarray, shape (`n`, nrhs)
        The improved solution matrix X.

    ferr : float, ndarray, shape (nrhs)
        `ferr`[j-1] contains an estimated error bound for the jth
        solution vector, that is, the jth column of X, for j =
        1,2,...,r.

    berr : float, ndarray, shape (nrhs)
        `berr`[j-1] contains the component-wise backward error bound
        beta for the jth solution vector, that is, the jth column of X,
        for j = 1,2,...,r.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -3)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.zsprfs'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
    )
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=False,
    )
    _afp = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'afp'},
        afp,
        copy=False,
    )
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
    )
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _b_lapacklin.zsprfs(
        uplo,
        n,
        _nrhs,
        _ap,
        _afp,
        _ipiv,
        _b,
        _x,
        _ferr,
        _berr,
    )
    _x_py = _x
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'ferr', 'berr'],
    )(
        _x_py,
        _ferr_py,
        _berr_py,
    )
    return _return_data

def zsptri(uplo, n, ap, ipiv):
    u"""
    Inverse of complex symmetric matrix, matrix already factorized by
    ``zsptrf``, packed storage.

    ``zsptri`` computes the inverse of a complex symmetric matrix A,
    where A has been factorized by ``zsptrf``, using packed storage.

    For full information please refer to the NAG Library document for
    f07qw

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07qwf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = PUDU^HP^T, where U is upper triangular.

        `uplo` = 'L'

            A = PLDL^HP^T, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    ap : complex, array-like, shape (`n`*(`n`+1)/2)
        The factorization of A stored in packed form, as returned by
        ``zsptrf``.

    ipiv : int, array-like, shape (`n`)
        Details of the interchanges and the block structure of D, as
        returned by ``zsptrf``.

    Returns
    -------
    ap : complex, ndarray, shape (`n`*(`n`+1)/2)
        The factorization is overwritten by the n by n matrix A^-1.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. D is
            singular and the inverse of A cannot be computed.

    References
    ----------
    Du Croz, J J and Higham, N J, 1992, `Stability of methods for matrix
    inversion`, IMA J. Numer. Anal. (12), 1--19
    """

    fun_name = 'naginterfaces.library.lapacklin.zsptri'

    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=True,
    )
    _ipiv = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ipiv'},
        ipiv,
        copy=False,
    )
    try:
        _b_lapacklin.zsptri(
            uplo,
            n,
            _ap,
            _ipiv,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _ap_py = _ap
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['ap'],
    )(
        _ap_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _ap_py

def dtrtrs(uplo, trans, diag, a, b):
    u"""
    Solution of real triangular system of linear equations, multiple
    right-hand sides.

    ``dtrtrs`` solves a real triangular system of linear equations with
    multiple right-hand sides, AX = B or A^TX = B.

    For full information please refer to the NAG Library document for
    f07te

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07tef.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

            A is upper triangular.

        `uplo` = 'L'

            A is lower triangular.

    trans : str, length 1
        Indicates the form of the equations.

        `trans` = 'N'

            The equations are of the form AX = B.

        `trans` = 'T' or 'C'

            The equations are of the form A^TX = B.

    diag : str, length 1
        Indicates whether A is a nonunit or unit triangular matrix.

        `diag` = 'N'

            A is a nonunit triangular matrix.

        `diag` = 'U'

            A is a unit triangular matrix; the diagonal elements are not
            referenced and are assumed to be 1.

    a : float, array-like, shape (n, n)
        The n by n triangular matrix A.

    b : float, array-like, shape (n, nrhs)
        The n by r right-hand side matrix B.

    Returns
    -------
    b : float, ndarray, shape (n, nrhs)
        The n by r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` -3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` -4)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -5)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. A is
            singular and the solution has not been computed.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Higham, N J, 1989, `The accuracy of solutions to triangular
    systems`, SIAM J. Numer. Anal. (26), 1252--1265
    """

    fun_name = 'naginterfaces.library.lapacklin.dtrtrs'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nrhs_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    try:
        _b_lapacklin.dtrtrs(
            uplo,
            trans,
            diag,
            _n,
            _nrhs,
            _a,
            _b,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _b_py = _b
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['b'],
    )(
        _b_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _b_py

def dtrcon(norm, uplo, diag, n, a):
    u"""
    Estimate condition number of real triangular matrix.

    ``dtrcon`` estimates the condition number of a real triangular
    matrix.

    For full information please refer to the NAG Library document for
    f07tg

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07tgf.html

    Parameters
    ----------
    norm : str, length 1
        Indicates whether kappa_1(A) or kappa_infinity (A) is estimated.

        `norm` = '1' or 'O'

            kappa_1(A) is estimated.

        `norm` = 'I'

            kappa_infinity (A) is estimated.

    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

            A is upper triangular.

        `uplo` = 'L'

            A is lower triangular.

    diag : str, length 1
        Indicates whether A is a nonunit or unit triangular matrix.

        `diag` = 'N'

            A is a nonunit triangular matrix.

        `diag` = 'U'

            A is a unit triangular matrix; the diagonal elements are not
            referenced and are assumed to be 1.

    n : int
        n, the order of the matrix A.

    a : float, array-like, shape (`n`, `n`)
        The n by n triangular matrix A.

    Returns
    -------
    rcond : float
        An estimate of the reciprocal of the condition number of A.
        `rcond` is set to zero if exact singularity is detected or if
        the estimate underflows. If `rcond` is less than machine
        precision, A is singular to working precision.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O' or 'I'.

        (`errno` -2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` -4)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    References
    ----------
    Higham, N J, 1988, `FORTRAN codes for estimating the one-norm of a
    real or complex matrix, with applications to condition estimation`,
    ACM Trans. Math. Software (14), 381--396
    """

    fun_name = 'naginterfaces.library.lapacklin.dtrcon'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _return_scalar = _b_lapacklin.dtrcon(
        norm,
        uplo,
        diag,
        n,
        _a,
    )
    _rcond_py = _return_scalar
    return _rcond_py

def dtrrfs(uplo, trans, diag, n, a, b, x):
    u"""
    Error bounds for solution of real triangular system of linear
    equations, multiple right-hand sides.

    ``dtrrfs`` returns error bounds for the solution of a real
    triangular system of linear equations with multiple right-hand
    sides, AX = B or A^TX = B.

    For full information please refer to the NAG Library document for
    f07th

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07thf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

            A is upper triangular.

        `uplo` = 'L'

            A is lower triangular.

    trans : str, length 1
        Indicates the form of the equations.

        `trans` = 'N'

            The equations are of the form AX = B.

        `trans` = 'T' or 'C'

            The equations are of the form A^TX = B.

    diag : str, length 1
        Indicates whether A is a nonunit or unit triangular matrix.

        `diag` = 'N'

            A is a nonunit triangular matrix.

        `diag` = 'U'

            A is a unit triangular matrix; the diagonal elements are not
            referenced and are assumed to be 1.

    n : int
        n, the order of the matrix A.

    a : float, array-like, shape (`n`, `n`)
        The n by n triangular matrix A.

    b : float, array-like, shape (`n`, nrhs)
        The n by r right-hand side matrix B.

    x : float, array-like, shape (`n`, nrhs)
        The n by r solution matrix X, as returned by ``dtrtrs``.

    Returns
    -------
    ferr : float, ndarray, shape (nrhs)
        `ferr`[j-1] contains an estimated error bound for the jth
        solution vector, that is, the jth column of X, for j =
        1,2,...,r.

    berr : float, ndarray, shape (nrhs)
        `berr`[j-1] contains the component-wise backward error bound
        beta for the jth solution vector, that is, the jth column of X,
        for j = 1,2,...,r.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` -3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` -4)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -5)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.dtrrfs'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nrhs_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _b_lapacklin.dtrrfs(
        uplo,
        trans,
        diag,
        n,
        _nrhs,
        _a,
        _b,
        _x,
        _ferr,
        _berr,
    )
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['ferr', 'berr'],
    )(
        _ferr_py,
        _berr_py,
    )
    return _return_data

def dtrtri(uplo, diag, a):
    u"""
    Inverse of real triangular matrix.

    ``dtrtri`` computes the inverse of a real triangular matrix.

    For full information please refer to the NAG Library document for
    f07tj

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07tjf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

            A is upper triangular.

        `uplo` = 'L'

            A is lower triangular.

    diag : str, length 1
        Indicates whether A is a nonunit or unit triangular matrix.

        `diag` = 'N'

            A is a nonunit triangular matrix.

        `diag` = 'U'

            A is a unit triangular matrix; the diagonal elements are not
            referenced and are assumed to be 1.

    a : float, array-like, shape (n, n)
        The n by n triangular matrix A.

    Returns
    -------
    a : float, ndarray, shape (n, n)
        A is overwritten by A^-1, using the same storage format as
        described above.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` -3)
            On entry, error in parameter n.

            Constraint: n >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. A is
            singular its inverse cannot be computed.

    References
    ----------
    Du Croz, J J and Higham, N J, 1992, `Stability of methods for matrix
    inversion`, IMA J. Numer. Anal. (12), 1--19
    """

    fun_name = 'naginterfaces.library.lapacklin.dtrtri'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    try:
        _b_lapacklin.dtrtri(
            uplo,
            diag,
            _n,
            _a,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _a_py = _a
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a'],
    )(
        _a_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _a_py

def ztrtrs(uplo, trans, diag, a, b):
    u"""
    Solution of complex triangular system of linear equations, multiple
    right-hand sides.

    ``ztrtrs`` solves a complex triangular system of linear equations
    with multiple right-hand sides, AX = B, A^TX = B or A^HX = B.

    For full information please refer to the NAG Library document for
    f07ts

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07tsf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

            A is upper triangular.

        `uplo` = 'L'

            A is lower triangular.

    trans : str, length 1
        Indicates the form of the equations.

        `trans` = 'N'

            The equations are of the form AX = B.

        `trans` = 'T'

            The equations are of the form A^TX = B.

        `trans` = 'C'

            The equations are of the form A^HX = B.

    diag : str, length 1
        Indicates whether A is a nonunit or unit triangular matrix.

        `diag` = 'N'

            A is a nonunit triangular matrix.

        `diag` = 'U'

            A is a unit triangular matrix; the diagonal elements are not
            referenced and are assumed to be 1.

    a : complex, array-like, shape (n, n)
        The n by n triangular matrix A.

    b : complex, array-like, shape (n, nrhs)
        The n by r right-hand side matrix B.

    Returns
    -------
    b : complex, ndarray, shape (n, nrhs)
        The n by r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` -3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` -4)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -5)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. A is
            singular and the solution has not been computed.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Higham, N J, 1989, `The accuracy of solutions to triangular
    systems`, SIAM J. Numer. Anal. (26), 1252--1265
    """

    fun_name = 'naginterfaces.library.lapacklin.ztrtrs'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
    )
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    try:
        _b_lapacklin.ztrtrs(
            uplo,
            trans,
            diag,
            _n,
            _nrhs,
            _a,
            _b,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _b_py = _b
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['b'],
    )(
        _b_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _b_py

def ztrcon(norm, uplo, diag, n, a):
    u"""
    Estimate condition number of complex triangular matrix.

    ``ztrcon`` estimates the condition number of a complex triangular
    matrix.

    For full information please refer to the NAG Library document for
    f07tu

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07tuf.html

    Parameters
    ----------
    norm : str, length 1
        Indicates whether kappa_1(A) or kappa_infinity (A) is estimated.

        `norm` = '1' or 'O'

            kappa_1(A) is estimated.

        `norm` = 'I'

            kappa_infinity (A) is estimated.

    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

            A is upper triangular.

        `uplo` = 'L'

            A is lower triangular.

    diag : str, length 1
        Indicates whether A is a nonunit or unit triangular matrix.

        `diag` = 'N'

            A is a nonunit triangular matrix.

        `diag` = 'U'

            A is a unit triangular matrix; the diagonal elements are not
            referenced and are assumed to be 1.

    n : int
        n, the order of the matrix A.

    a : complex, array-like, shape (`n`, `n`)
        The n by n triangular matrix A.

    Returns
    -------
    rcond : float
        An estimate of the reciprocal of the condition number of A.
        `rcond` is set to zero if exact singularity is detected or the
        estimate underflows. If `rcond` is less than machine precision,
        A is singular to working precision.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O' or 'I'.

        (`errno` -2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` -4)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    References
    ----------
    Higham, N J, 1988, `FORTRAN codes for estimating the one-norm of a
    real or complex matrix, with applications to condition estimation`,
    ACM Trans. Math. Software (14), 381--396
    """

    fun_name = 'naginterfaces.library.lapacklin.ztrcon'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
    )
    _return_scalar = _b_lapacklin.ztrcon(
        norm,
        uplo,
        diag,
        n,
        _a,
    )
    _rcond_py = _return_scalar
    return _rcond_py

def ztrrfs(uplo, trans, diag, n, a, b, x):
    u"""
    Error bounds for solution of complex triangular system of linear
    equations, multiple right-hand sides.

    ``ztrrfs`` returns error bounds for the solution of a complex
    triangular system of linear equations with multiple right-hand
    sides, AX = B, A^TX = B or A^HX = B.

    For full information please refer to the NAG Library document for
    f07tv

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07tvf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

            A is upper triangular.

        `uplo` = 'L'

            A is lower triangular.

    trans : str, length 1
        Indicates the form of the equations.

        `trans` = 'N'

            The equations are of the form AX = B.

        `trans` = 'T'

            The equations are of the form A^TX = B.

        `trans` = 'C'

            The equations are of the form A^HX = B.

    diag : str, length 1
        Indicates whether A is a nonunit or unit triangular matrix.

        `diag` = 'N'

            A is a nonunit triangular matrix.

        `diag` = 'U'

            A is a unit triangular matrix; the diagonal elements are not
            referenced and are assumed to be 1.

    n : int
        n, the order of the matrix A.

    a : complex, array-like, shape (`n`, `n`)
        The n by n triangular matrix A.

    b : complex, array-like, shape (`n`, nrhs)
        The n by r right-hand side matrix B.

    x : complex, array-like, shape (`n`, nrhs)
        The n by r solution matrix X, as returned by ``ztrtrs``.

    Returns
    -------
    ferr : float, ndarray, shape (nrhs)
        `ferr`[j-1] contains an estimated error bound for the jth
        solution vector, that is, the jth column of X, for j =
        1,2,...,r.

    berr : float, ndarray, shape (nrhs)
        `berr`[j-1] contains the component-wise backward error bound
        beta for the jth solution vector, that is, the jth column of X,
        for j = 1,2,...,r.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` -3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` -4)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -5)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.ztrrfs'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
    )
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
    )
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _b_lapacklin.ztrrfs(
        uplo,
        trans,
        diag,
        n,
        _nrhs,
        _a,
        _b,
        _x,
        _ferr,
        _berr,
    )
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['ferr', 'berr'],
    )(
        _ferr_py,
        _berr_py,
    )
    return _return_data

def ztrtri(uplo, diag, a):
    u"""
    Inverse of complex triangular matrix.

    ``ztrtri`` computes the inverse of a complex triangular matrix.

    For full information please refer to the NAG Library document for
    f07tw

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07twf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

            A is upper triangular.

        `uplo` = 'L'

            A is lower triangular.

    diag : str, length 1
        Indicates whether A is a nonunit or unit triangular matrix.

        `diag` = 'N'

            A is a nonunit triangular matrix.

        `diag` = 'U'

            A is a unit triangular matrix; the diagonal elements are not
            referenced and are assumed to be 1.

    a : complex, array-like, shape (n, n)
        The n by n triangular matrix A.

    Returns
    -------
    a : complex, ndarray, shape (n, n)
        A is overwritten by A^-1, using the same storage format as
        described above.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` -3)
            On entry, error in parameter n.

            Constraint: n >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. A is
            singular its inverse cannot be computed.

    References
    ----------
    Du Croz, J J and Higham, N J, 1992, `Stability of methods for matrix
    inversion`, IMA J. Numer. Anal. (12), 1--19
    """

    fun_name = 'naginterfaces.library.lapacklin.ztrtri'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    try:
        _b_lapacklin.ztrtri(
            uplo,
            diag,
            _n,
            _a,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _a_py = _a
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a'],
    )(
        _a_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _a_py

def dtptrs(uplo, trans, diag, n, ap, b):
    u"""
    Solution of real triangular system of linear equations, multiple
    right-hand sides, packed storage.

    ``dtptrs`` solves a real triangular system of linear equations with
    multiple right-hand sides, AX = B or A^TX = B, using packed storage.

    For full information please refer to the NAG Library document for
    f07ue

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07uef.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

            A is upper triangular.

        `uplo` = 'L'

            A is lower triangular.

    trans : str, length 1
        Indicates the form of the equations.

        `trans` = 'N'

            The equations are of the form AX = B.

        `trans` = 'T' or 'C'

            The equations are of the form A^TX = B.

    diag : str, length 1
        Indicates whether A is a nonunit or unit triangular matrix.

        `diag` = 'N'

            A is a nonunit triangular matrix.

        `diag` = 'U'

            A is a unit triangular matrix; the diagonal elements are not
            referenced and are assumed to be 1.

    n : int
        n, the order of the matrix A.

    ap : float, array-like, shape (`n`*(`n`+1)/2)
        The n by n triangular matrix A, packed by columns.

    b : float, array-like, shape (`n`, nrhs)
        The n by r right-hand side matrix B.

    Returns
    -------
    b : float, ndarray, shape (`n`, nrhs)
        The n by r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` -3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` -4)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -5)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. A is
            singular and the solution has not been computed.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Higham, N J, 1989, `The accuracy of solutions to triangular
    systems`, SIAM J. Numer. Anal. (26), 1252--1265
    """

    fun_name = 'naginterfaces.library.lapacklin.dtptrs'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nrhs_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    try:
        _b_lapacklin.dtptrs(
            uplo,
            trans,
            diag,
            n,
            _nrhs,
            _ap,
            _b,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _b_py = _b
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['b'],
    )(
        _b_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _b_py

def dtpcon(norm, uplo, diag, n, ap):
    u"""
    Estimate condition number of real triangular matrix, packed storage.

    ``dtpcon`` estimates the condition number of a real triangular
    matrix, using packed storage.

    For full information please refer to the NAG Library document for
    f07ug

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07ugf.html

    Parameters
    ----------
    norm : str, length 1
        Indicates whether kappa_1(A) or kappa_infinity (A) is estimated.

        `norm` = '1' or 'O'

            kappa_1(A) is estimated.

        `norm` = 'I'

            kappa_infinity (A) is estimated.

    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

            A is upper triangular.

        `uplo` = 'L'

            A is lower triangular.

    diag : str, length 1
        Indicates whether A is a nonunit or unit triangular matrix.

        `diag` = 'N'

            A is a nonunit triangular matrix.

        `diag` = 'U'

            A is a unit triangular matrix; the diagonal elements are not
            referenced and are assumed to be 1.

    n : int
        n, the order of the matrix A.

    ap : float, array-like, shape (`n`*(`n`+1)/2)
        The n by n triangular matrix A, packed by columns.

    Returns
    -------
    rcond : float
        An estimate of the reciprocal of the condition number of A.
        `rcond` is set to zero if exact singularity is detected or the
        estimate underflows. If `rcond` is less than machine precision,
        A is singular to working precision.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O' or 'I'.

        (`errno` -2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` -4)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    References
    ----------
    Higham, N J, 1988, `FORTRAN codes for estimating the one-norm of a
    real or complex matrix, with applications to condition estimation`,
    ACM Trans. Math. Software (14), 381--396
    """

    fun_name = 'naginterfaces.library.lapacklin.dtpcon'

    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _return_scalar = _b_lapacklin.dtpcon(
        norm,
        uplo,
        diag,
        n,
        _ap,
    )
    _rcond_py = _return_scalar
    return _rcond_py

def dtprfs(uplo, trans, diag, n, ap, b, x):
    u"""
    Error bounds for solution of real triangular system of linear
    equations, multiple right-hand sides, packed storage.

    ``dtprfs`` returns error bounds for the solution of a real
    triangular system of linear equations with multiple right-hand
    sides, AX = B or A^TX = B, using packed storage.

    For full information please refer to the NAG Library document for
    f07uh

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07uhf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

            A is upper triangular.

        `uplo` = 'L'

            A is lower triangular.

    trans : str, length 1
        Indicates the form of the equations.

        `trans` = 'N'

            The equations are of the form AX = B.

        `trans` = 'T' or 'C'

            The equations are of the form A^TX = B.

    diag : str, length 1
        Indicates whether A is a nonunit or unit triangular matrix.

        `diag` = 'N'

            A is a nonunit triangular matrix.

        `diag` = 'U'

            A is a unit triangular matrix; the diagonal elements are not
            referenced and are assumed to be 1.

    n : int
        n, the order of the matrix A.

    ap : float, array-like, shape (`n`*(`n`+1)/2)
        The n by n triangular matrix A, packed by columns.

    b : float, array-like, shape (`n`, nrhs)
        The n by r right-hand side matrix B.

    x : float, array-like, shape (`n`, nrhs)
        The n by r solution matrix X, as returned by ``dtptrs``.

    Returns
    -------
    ferr : float, ndarray, shape (nrhs)
        `ferr`[j-1] contains an estimated error bound for the jth
        solution vector, that is, the jth column of X, for j =
        1,2,...,r.

    berr : float, ndarray, shape (nrhs)
        `berr`[j-1] contains the component-wise backward error bound
        beta for the jth solution vector, that is, the jth column of X,
        for j = 1,2,...,r.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` -3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` -4)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -5)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.dtprfs'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nrhs_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _b_lapacklin.dtprfs(
        uplo,
        trans,
        diag,
        n,
        _nrhs,
        _ap,
        _b,
        _x,
        _ferr,
        _berr,
    )
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['ferr', 'berr'],
    )(
        _ferr_py,
        _berr_py,
    )
    return _return_data

def dtptri(uplo, diag, n, ap):
    u"""
    Inverse of real triangular matrix, packed storage.

    ``dtptri`` computes the inverse of a real triangular matrix, using
    packed storage.

    For full information please refer to the NAG Library document for
    f07uj

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07ujf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

            A is upper triangular.

        `uplo` = 'L'

            A is lower triangular.

    diag : str, length 1
        Indicates whether A is a nonunit or unit triangular matrix.

        `diag` = 'N'

            A is a nonunit triangular matrix.

        `diag` = 'U'

            A is a unit triangular matrix; the diagonal elements are not
            referenced and are assumed to be 1.

    n : int
        n, the order of the matrix A.

    ap : float, array-like, shape (`n`*(`n`+1)/2)
        The n by n triangular matrix A, packed by columns.

    Returns
    -------
    ap : float, ndarray, shape (`n`*(`n`+1)/2)
        A is overwritten by A^-1, using the same storage format as
        described above.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` -3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. A is
            singular its inverse cannot be computed.

    References
    ----------
    Du Croz, J J and Higham, N J, 1992, `Stability of methods for matrix
    inversion`, IMA J. Numer. Anal. (12), 1--19
    """

    fun_name = 'naginterfaces.library.lapacklin.dtptri'

    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    try:
        _b_lapacklin.dtptri(
            uplo,
            diag,
            n,
            _ap,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _ap_py = _ap
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['ap'],
    )(
        _ap_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _ap_py

def ztptrs(uplo, trans, diag, n, ap, b):
    u"""
    Solution of complex triangular system of linear equations, multiple
    right-hand sides, packed storage.

    ``ztptrs`` solves a complex triangular system of linear equations
    with multiple right-hand sides, AX = B, A^TX = B or A^HX = B, using
    packed storage.

    For full information please refer to the NAG Library document for
    f07us

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07usf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

            A is upper triangular.

        `uplo` = 'L'

            A is lower triangular.

    trans : str, length 1
        Indicates the form of the equations.

        `trans` = 'N'

            The equations are of the form AX = B.

        `trans` = 'T'

            The equations are of the form A^TX = B.

        `trans` = 'C'

            The equations are of the form A^HX = B.

    diag : str, length 1
        Indicates whether A is a nonunit or unit triangular matrix.

        `diag` = 'N'

            A is a nonunit triangular matrix.

        `diag` = 'U'

            A is a unit triangular matrix; the diagonal elements are not
            referenced and are assumed to be 1.

    n : int
        n, the order of the matrix A.

    ap : complex, array-like, shape (`n`*(`n`+1)/2)
        The n by n triangular matrix A, packed by columns.

    b : complex, array-like, shape (`n`, nrhs)
        The n by r right-hand side matrix B.

    Returns
    -------
    b : complex, ndarray, shape (`n`, nrhs)
        The n by r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` -3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` -4)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -5)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. A is
            singular and the solution has not been computed.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Higham, N J, 1989, `The accuracy of solutions to triangular
    systems`, SIAM J. Numer. Anal. (26), 1252--1265
    """

    fun_name = 'naginterfaces.library.lapacklin.ztptrs'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
    )
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=False,
    )
    try:
        _b_lapacklin.ztptrs(
            uplo,
            trans,
            diag,
            n,
            _nrhs,
            _ap,
            _b,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _b_py = _b
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['b'],
    )(
        _b_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _b_py

def ztpcon(norm, uplo, diag, n, ap):
    u"""
    Estimate condition number of complex triangular matrix, packed
    storage.

    ``ztpcon`` estimates the condition number of a complex triangular
    matrix, using packed storage.

    For full information please refer to the NAG Library document for
    f07uu

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07uuf.html

    Parameters
    ----------
    norm : str, length 1
        Indicates whether kappa_1(A) or kappa_infinity (A) is estimated.

        `norm` = '1' or 'O'

            kappa_1(A) is estimated.

        `norm` = 'I'

            kappa_infinity (A) is estimated.

    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

            A is upper triangular.

        `uplo` = 'L'

            A is lower triangular.

    diag : str, length 1
        Indicates whether A is a nonunit or unit triangular matrix.

        `diag` = 'N'

            A is a nonunit triangular matrix.

        `diag` = 'U'

            A is a unit triangular matrix; the diagonal elements are not
            referenced and are assumed to be 1.

    n : int
        n, the order of the matrix A.

    ap : complex, array-like, shape (`n`*(`n`+1)/2)
        The n by n triangular matrix A, packed by columns.

    Returns
    -------
    rcond : float
        An estimate of the reciprocal of the condition number of A.
        `rcond` is set to zero if exact singularity is detected or the
        estimate underflows. If `rcond` is less than machine precision,
        A is singular to working precision.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O' or 'I'.

        (`errno` -2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` -4)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    References
    ----------
    Higham, N J, 1988, `FORTRAN codes for estimating the one-norm of a
    real or complex matrix, with applications to condition estimation`,
    ACM Trans. Math. Software (14), 381--396
    """

    fun_name = 'naginterfaces.library.lapacklin.ztpcon'

    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=False,
    )
    _return_scalar = _b_lapacklin.ztpcon(
        norm,
        uplo,
        diag,
        n,
        _ap,
    )
    _rcond_py = _return_scalar
    return _rcond_py

def ztprfs(uplo, trans, diag, n, ap, b, x):
    u"""
    Error bounds for solution of complex triangular system of linear
    equations, multiple right-hand sides, packed storage.

    ``ztprfs`` returns error bounds for the solution of a complex
    triangular system of linear equations with multiple right-hand
    sides, AX = B, A^TX = B or A^HX = B, using packed storage.

    For full information please refer to the NAG Library document for
    f07uv

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07uvf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

            A is upper triangular.

        `uplo` = 'L'

            A is lower triangular.

    trans : str, length 1
        Indicates the form of the equations.

        `trans` = 'N'

            The equations are of the form AX = B.

        `trans` = 'T'

            The equations are of the form A^TX = B.

        `trans` = 'C'

            The equations are of the form A^HX = B.

    diag : str, length 1
        Indicates whether A is a nonunit or unit triangular matrix.

        `diag` = 'N'

            A is a nonunit triangular matrix.

        `diag` = 'U'

            A is a unit triangular matrix; the diagonal elements are not
            referenced and are assumed to be 1.

    n : int
        n, the order of the matrix A.

    ap : complex, array-like, shape (`n`*(`n`+1)/2)
        The n by n triangular matrix A, packed by columns.

    b : complex, array-like, shape (`n`, nrhs)
        The n by r right-hand side matrix B.

    x : complex, array-like, shape (`n`, nrhs)
        The n by r solution matrix X, as returned by ``ztptrs``.

    Returns
    -------
    ferr : float, ndarray, shape (nrhs)
        `ferr`[j-1] contains an estimated error bound for the jth
        solution vector, that is, the jth column of X, for j =
        1,2,...,r.

    berr : float, ndarray, shape (nrhs)
        `berr`[j-1] contains the component-wise backward error bound
        beta for the jth solution vector, that is, the jth column of X,
        for j = 1,2,...,r.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` -3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` -4)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -5)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.ztprfs'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
    )
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=False,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
    )
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _b_lapacklin.ztprfs(
        uplo,
        trans,
        diag,
        n,
        _nrhs,
        _ap,
        _b,
        _x,
        _ferr,
        _berr,
    )
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['ferr', 'berr'],
    )(
        _ferr_py,
        _berr_py,
    )
    return _return_data

def ztptri(uplo, diag, n, ap):
    u"""
    Inverse of complex triangular matrix, packed storage.

    ``ztptri`` computes the inverse of a complex triangular matrix,
    using packed storage.

    For full information please refer to the NAG Library document for
    f07uw

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07uwf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

            A is upper triangular.

        `uplo` = 'L'

            A is lower triangular.

    diag : str, length 1
        Indicates whether A is a nonunit or unit triangular matrix.

        `diag` = 'N'

            A is a nonunit triangular matrix.

        `diag` = 'U'

            A is a unit triangular matrix; the diagonal elements are not
            referenced and are assumed to be 1.

    n : int
        n, the order of the matrix A.

    ap : complex, array-like, shape (`n`*(`n`+1)/2)
        The n by n triangular matrix A, packed by columns.

    Returns
    -------
    ap : complex, ndarray, shape (`n`*(`n`+1)/2)
        A is overwritten by A^-1, using the same storage format as
        described above.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` -3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. A is
            singular its inverse cannot be computed.

    References
    ----------
    Du Croz, J J and Higham, N J, 1992, `Stability of methods for matrix
    inversion`, IMA J. Numer. Anal. (12), 1--19
    """

    fun_name = 'naginterfaces.library.lapacklin.ztptri'

    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=True,
    )
    try:
        _b_lapacklin.ztptri(
            uplo,
            diag,
            n,
            _ap,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _ap_py = _ap
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['ap'],
    )(
        _ap_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _ap_py

def dtbtrs(uplo, trans, diag, kd, ab, b):
    u"""
    Solution of real band triangular system of linear equations,
    multiple right-hand sides.

    ``dtbtrs`` solves a real triangular band system of linear equations
    with multiple right-hand sides, AX = B or A^TX = B.

    For full information please refer to the NAG Library document for
    f07ve

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07vef.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

            A is upper triangular.

        `uplo` = 'L'

            A is lower triangular.

    trans : str, length 1
        Indicates the form of the equations.

        `trans` = 'N'

            The equations are of the form AX = B.

        `trans` = 'T' or 'C'

            The equations are of the form A^TX = B.

    diag : str, length 1
        Indicates whether A is a nonunit or unit triangular matrix.

        `diag` = 'N'

            A is a nonunit triangular matrix.

        `diag` = 'U'

            A is a unit triangular matrix; the diagonal elements are not
            referenced and are assumed to be 1.

    kd : int
        k_d, the number of superdiagonals of the matrix A if `uplo` =
        'U', or the number of subdiagonals if `uplo` = 'L'.

    ab : float, array-like, shape (`kd`+1, n)
        The n by n triangular band matrix A.

    b : float, array-like, shape (n, nrhs)
        The n by r right-hand side matrix B.

    Returns
    -------
    b : float, ndarray, shape (n, nrhs)
        The n by r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` -3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` -4)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -5)
            On entry, error in parameter `kd`.

            Constraint: `kd` >= 0.

        (`errno` -6)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. A is
            singular and the solution has not been computed.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Higham, N J, 1989, `The accuracy of solutions to triangular
    systems`, SIAM J. Numer. Anal. (26), 1252--1265
    """

    fun_name = 'naginterfaces.library.lapacklin.dtbtrs'

    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nrhs_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    try:
        _b_lapacklin.dtbtrs(
            uplo,
            trans,
            diag,
            _n,
            kd,
            _nrhs,
            _ab,
            _b,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _b_py = _b
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['b'],
    )(
        _b_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _b_py

def dtbcon(norm, uplo, diag, kd, ab):
    u"""
    Estimate condition number of real band triangular matrix.

    ``dtbcon`` estimates the condition number of a real triangular band
    matrix.

    For full information please refer to the NAG Library document for
    f07vg

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07vgf.html

    Parameters
    ----------
    norm : str, length 1
        Indicates whether kappa_1(A) or kappa_infinity (A) is estimated.

        `norm` = '1' or 'O'

            kappa_1(A) is estimated.

        `norm` = 'I'

            kappa_infinity (A) is estimated.

    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

            A is upper triangular.

        `uplo` = 'L'

            A is lower triangular.

    diag : str, length 1
        Indicates whether A is a nonunit or unit triangular matrix.

        `diag` = 'N'

            A is a nonunit triangular matrix.

        `diag` = 'U'

            A is a unit triangular matrix; the diagonal elements are not
            referenced and are assumed to be 1.

    kd : int
        k_d, the number of superdiagonals of the matrix A if `uplo` =
        'U', or the number of subdiagonals if `uplo` = 'L'.

    ab : float, array-like, shape (`kd`+1, n)
        The n by n triangular band matrix A.

    Returns
    -------
    rcond : float
        An estimate of the reciprocal of the condition number of A.
        `rcond` is set to zero if exact singularity is detected or the
        estimate underflows. If `rcond` is less than machine precision,
        A is singular to working precision.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O' or 'I'.

        (`errno` -2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` -4)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -5)
            On entry, error in parameter `kd`.

            Constraint: `kd` >= 0.

    References
    ----------
    Higham, N J, 1988, `FORTRAN codes for estimating the one-norm of a
    real or complex matrix, with applications to condition estimation`,
    ACM Trans. Math. Software (14), 381--396
    """

    fun_name = 'naginterfaces.library.lapacklin.dtbcon'

    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _return_scalar = _b_lapacklin.dtbcon(
        norm,
        uplo,
        diag,
        _n,
        kd,
        _ab,
    )
    _rcond_py = _return_scalar
    return _rcond_py

def dtbrfs(uplo, trans, diag, kd, nrhs, ab, b, x):
    u"""
    Error bounds for solution of real band triangular system of linear
    equations, multiple right-hand sides.

    ``dtbrfs`` returns error bounds for the solution of a real
    triangular band system of linear equations with multiple right-hand
    sides, AX = B or A^TX = B.

    For full information please refer to the NAG Library document for
    f07vh

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07vhf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

            A is upper triangular.

        `uplo` = 'L'

            A is lower triangular.

    trans : str, length 1
        Indicates the form of the equations.

        `trans` = 'N'

            The equations are of the form AX = B.

        `trans` = 'T' or 'C'

            The equations are of the form A^TX = B.

    diag : str, length 1
        Indicates whether A is a nonunit or unit triangular matrix.

        `diag` = 'N'

            A is a nonunit triangular matrix.

        `diag` = 'U'

            A is a unit triangular matrix; the diagonal elements are not
            referenced and are assumed to be 1.

    kd : int
        k_d, the number of superdiagonals of the matrix A if `uplo` =
        'U', or the number of subdiagonals if `uplo` = 'L'.

    nrhs : int
        r, the number of right-hand sides.

    ab : float, array-like, shape (`kd`+1, n)
        The n by n triangular band matrix A.

    b : float, array-like, shape (n, `nrhs`)
        The n by r right-hand side matrix B.

    x : float, array-like, shape (n, `nrhs`)
        The n by r solution matrix X, as returned by ``dtbtrs``.

    Returns
    -------
    ferr : float, ndarray, shape (`nrhs`)
        `ferr`[j-1] contains an estimated error bound for the jth
        solution vector, that is, the jth column of X, for j =
        1,2,...,r.

    berr : float, ndarray, shape (`nrhs`)
        `berr`[j-1] contains the component-wise backward error bound
        beta for the jth solution vector, that is, the jth column of X,
        for j = 1,2,...,r.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` -3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` -4)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -5)
            On entry, error in parameter `kd`.

            Constraint: `kd` >= 0.

        (`errno` -6)
            On entry, error in parameter `nrhs`.

            Constraint: `nrhs` >= 0.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.dtbrfs'

    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nrhs = _b_utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineIntScalarType.to_py(_nrhs)
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _b_lapacklin.dtbrfs(
        uplo,
        trans,
        diag,
        _n,
        kd,
        _nrhs,
        _ab,
        _b,
        _x,
        _ferr,
        _berr,
    )
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['ferr', 'berr'],
    )(
        _ferr_py,
        _berr_py,
    )
    return _return_data

def ztbtrs(uplo, trans, diag, kd, ab, b):
    u"""
    Solution of complex band triangular system of linear equations,
    multiple right-hand sides.

    ``ztbtrs`` solves a complex triangular band system of linear
    equations with multiple right-hand sides, AX = B, A^TX = B or A^HX =
    B.

    For full information please refer to the NAG Library document for
    f07vs

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07vsf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

            A is upper triangular.

        `uplo` = 'L'

            A is lower triangular.

    trans : str, length 1
        Indicates the form of the equations.

        `trans` = 'N'

            The equations are of the form AX = B.

        `trans` = 'T'

            The equations are of the form A^TX = B.

        `trans` = 'C'

            The equations are of the form A^HX = B.

    diag : str, length 1
        Indicates whether A is a nonunit or unit triangular matrix.

        `diag` = 'N'

            A is a nonunit triangular matrix.

        `diag` = 'U'

            A is a unit triangular matrix; the diagonal elements are not
            referenced and are assumed to be 1.

    kd : int
        k_d, the number of superdiagonals of the matrix A if `uplo` =
        'U', or the number of subdiagonals if `uplo` = 'L'.

    ab : complex, array-like, shape (`kd`+1, n)
        The n by n triangular band matrix A.

    b : complex, array-like, shape (n, nrhs)
        The n by r right-hand side matrix B.

    Returns
    -------
    b : complex, ndarray, shape (n, nrhs)
        The n by r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` -3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` -4)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -5)
            On entry, error in parameter `kd`.

            Constraint: `kd` >= 0.

        (`errno` -6)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Element *<value>* of the diagonal is exactly zero. A is
            singular and the solution has not been computed.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Higham, N J, 1989, `The accuracy of solutions to triangular
    systems`, SIAM J. Numer. Anal. (26), 1252--1265
    """

    fun_name = 'naginterfaces.library.lapacklin.ztbtrs'

    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
    )
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    try:
        _b_lapacklin.ztbtrs(
            uplo,
            trans,
            diag,
            _n,
            kd,
            _nrhs,
            _ab,
            _b,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _b_py = _b
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['b'],
    )(
        _b_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _b_py

def ztbcon(norm, uplo, diag, kd, ab):
    u"""
    Estimate condition number of complex band triangular matrix.

    ``ztbcon`` estimates the condition number of a complex triangular
    band matrix.

    For full information please refer to the NAG Library document for
    f07vu

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07vuf.html

    Parameters
    ----------
    norm : str, length 1
        Indicates whether kappa_1(A) or kappa_infinity (A) is estimated.

        `norm` = '1' or 'O'

            kappa_1(A) is estimated.

        `norm` = 'I'

            kappa_infinity (A) is estimated.

    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

            A is upper triangular.

        `uplo` = 'L'

            A is lower triangular.

    diag : str, length 1
        Indicates whether A is a nonunit or unit triangular matrix.

        `diag` = 'N'

            A is a nonunit triangular matrix.

        `diag` = 'U'

            A is a unit triangular matrix; the diagonal elements are not
            referenced and are assumed to be 1.

    kd : int
        k_d, the number of superdiagonals of the matrix A if `uplo` =
        'U', or the number of subdiagonals if `uplo` = 'L'.

    ab : complex, array-like, shape (`kd`+1, n)
        The n by n triangular band matrix A.

    Returns
    -------
    rcond : float
        An estimate of the reciprocal of the condition number of A.
        `rcond` is set to zero if exact singularity is detected or the
        estimate underflows. If `rcond` is less than machine precision,
        A is singular to working precision.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `norm`.

            Constraint: `norm` = '1', 'O' or 'I'.

        (`errno` -2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` -4)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -5)
            On entry, error in parameter `kd`.

            Constraint: `kd` >= 0.

    References
    ----------
    Higham, N J, 1988, `FORTRAN codes for estimating the one-norm of a
    real or complex matrix, with applications to condition estimation`,
    ACM Trans. Math. Software (14), 381--396
    """

    fun_name = 'naginterfaces.library.lapacklin.ztbcon'

    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _return_scalar = _b_lapacklin.ztbcon(
        norm,
        uplo,
        diag,
        _n,
        kd,
        _ab,
    )
    _rcond_py = _return_scalar
    return _rcond_py

def ztbrfs(uplo, trans, diag, kd, nrhs, ab, b, x):
    u"""
    Error bounds for solution of complex band triangular system of
    linear equations, multiple right-hand sides.

    ``ztbrfs`` returns error bounds for the solution of a complex
    triangular band system of linear equations with multiple right-hand
    sides, AX = B, A^TX = B or A^HX = B.

    For full information please refer to the NAG Library document for
    f07vv

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07vvf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

            A is upper triangular.

        `uplo` = 'L'

            A is lower triangular.

    trans : str, length 1
        Indicates the form of the equations.

        `trans` = 'N'

            The equations are of the form AX = B.

        `trans` = 'T'

            The equations are of the form A^TX = B.

        `trans` = 'C'

            The equations are of the form A^HX = B.

    diag : str, length 1
        Indicates whether A is a nonunit or unit triangular matrix.

        `diag` = 'N'

            A is a nonunit triangular matrix.

        `diag` = 'U'

            A is a unit triangular matrix; the diagonal elements are not
            referenced and are assumed to be 1.

    kd : int
        k_d, the number of superdiagonals of the matrix A if `uplo` =
        'U', or the number of subdiagonals if `uplo` = 'L'.

    nrhs : int
        r, the number of right-hand sides.

    ab : complex, array-like, shape (`kd`+1, n)
        The n by n triangular band matrix A.

    b : complex, array-like, shape (n, `nrhs`)
        The n by r right-hand side matrix B.

    x : complex, array-like, shape (n, `nrhs`)
        The n by r solution matrix X, as returned by ``ztbtrs``.

    Returns
    -------
    ferr : float, ndarray, shape (`nrhs`)
        `ferr`[j-1] contains an estimated error bound for the jth
        solution vector, that is, the jth column of X, for j =
        1,2,...,r.

    berr : float, ndarray, shape (`nrhs`)
        `berr`[j-1] contains the component-wise backward error bound
        beta for the jth solution vector, that is, the jth column of X,
        for j = 1,2,...,r.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `trans`.

            Constraint: `trans` = 'N', 'T' or 'C'.

        (`errno` -3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` -4)
            On entry, error in parameter n.

            Constraint: n >= 0.

        (`errno` -5)
            On entry, error in parameter `kd`.

            Constraint: `kd` >= 0.

        (`errno` -6)
            On entry, error in parameter `nrhs`.

            Constraint: `nrhs` >= 0.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore
    """

    fun_name = 'naginterfaces.library.lapacklin.ztbrfs'

    _ab = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ab'},
        ab,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_ab, {'fun_name': fun_name, 'entity': 'ab'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
    )
    _nrhs = _b_utils._EngineIntScalarType(
        nrhs, {'fun_name': fun_name, 'entity': 'nrhs'},
    ).to_ctypes()
    _nrhs_py = _b_utils._EngineIntScalarType.to_py(_nrhs)
    _ferr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _berr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nrhs_py,),
    )
    _b_lapacklin.ztbrfs(
        uplo,
        trans,
        diag,
        _n,
        kd,
        _nrhs,
        _ab,
        _b,
        _x,
        _ferr,
        _berr,
    )
    _ferr_py = _ferr
    _berr_py = _berr
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['ferr', 'berr'],
    )(
        _ferr_py,
        _berr_py,
    )
    return _return_data

def dpftrf(transr, uplo, n, ar):
    u"""
    Cholesky factorization of real symmetric positive definite matrix,
    Rectangular Full Packed format.

    ``dpftrf`` computes the Cholesky factorization of a real symmetric
    positive definite matrix stored in Rectangular Full Packed (RFP)
    format.

    For full information please refer to the NAG Library document for
    f07wd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07wdf.html

    Parameters
    ----------
    transr : str, length 1
        Specifies whether the RFP representation of A is normal or
        transposed.

        `transr` = 'N'

            The matrix A is stored in normal RFP format.

        `transr` = 'T'

            The matrix A is stored in transposed RFP format.

    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

            The upper triangular part of A is stored, and A is
            factorized as U^TU, where U is upper triangular.

        `uplo` = 'L'

            The lower triangular part of A is stored, and A is
            factorized as LL^T, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    ar : float, array-like, shape (`n`*(`n`+1)/2)
        The upper or lower triangular part (as specified by `uplo`) of
        the n by n symmetric matrix A, in either normal or transposed
        RFP format (as specified by `transr`). The storage format is
        described in detail in [the F07 Introduction].

    Returns
    -------
    ar : float, ndarray, shape (`n`*(`n`+1)/2)
        If no exception or warning is raised, the factor U or L from the
        Cholesky factorization A = U^TU or A = LL^T, in the same storage
        format as A.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `transr`.

            Constraint: `transr` = 'N' or 'T'.

        (`errno` -2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            The leading minor of order *<value>* is not positive
            definite and the factorization could not be completed. Hence
            A itself is not positive definite. This may indicate an
            error in forming the matrix A. There is no function
            specifically designed to factorize a symmetric matrix stored
            in RFP format which is not positive definite; the matrix
            must be treated as a full symmetric matrix, by calling
            ``dsytrf``.

    References
    ----------
    Demmel, J W, 1989, `On floating-point errors in Cholesky`, LAPACK
    Working Note No. 14, University of Tennessee, Knoxville,
    https://www.netlib.org/lapack/lawnspdf/lawn14.pdf

    Gustavson, F G, Waśniewski, J, Dongarra, J J and Langou, J, 2010,
    `Rectangular full packed format for Cholesky's algorithm:
    factorization, solution, and inversion`, ACM Trans. Math. Software
    (37, 2)
    """

    fun_name = 'naginterfaces.library.lapacklin.dpftrf'

    _ar = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ar'},
        ar,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    try:
        _b_lapacklin.dpftrf(
            transr,
            uplo,
            n,
            _ar,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _ar_py = _ar
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['ar'],
    )(
        _ar_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _ar_py

def dpftrs(transr, uplo, n, ar, b):
    u"""
    Solution of real symmetric positive definite system of linear
    equations, multiple right-hand sides, coefficient matrix already
    factorized by ``dpftrf``, Rectangular Full Packed format.

    ``dpftrs`` solves a real symmetric positive definite system of
    linear equations with multiple right-hand sides,

        AX = B,

    using the Cholesky factorization computed by ``dpftrf`` stored in
    Rectangular Full Packed (RFP) format.

    For full information please refer to the NAG Library document for
    f07we

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07wef.html

    Parameters
    ----------
    transr : str, length 1
        Specifies whether the RFP representation of A is normal or
        transposed.

        `transr` = 'N'

            The matrix A is stored in normal RFP format.

        `transr` = 'T'

            The matrix A is stored in transposed RFP format.

    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = U^TU, where U is upper triangular.

        `uplo` = 'L'

            A = LL^T, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    ar : float, array-like, shape (`n`*(`n`+1)/2)
        The Cholesky factorization of A stored in RFP format, as
        returned by ``dpftrf``.

    b : float, array-like, shape (`n`, nrhs)
        The n by r right-hand side matrix B.

    Returns
    -------
    b : float, ndarray, shape (`n`, nrhs)
        The n by r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `transr`.

            Constraint: `transr` = 'N' or 'T'.

        (`errno` -2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -4)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    References
    ----------
    Gustavson, F G, Waśniewski, J, Dongarra, J J and Langou, J, 2010,
    `Rectangular full packed format for Cholesky's algorithm:
    factorization, solution, and inversion`, ACM Trans. Math. Software
    (37, 2)
    """

    fun_name = 'naginterfaces.library.lapacklin.dpftrs'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nrhs_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ar = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ar'},
        ar,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b_lapacklin.dpftrs(
        transr,
        uplo,
        n,
        _nrhs,
        _ar,
        _b,
    )
    _b_py = _b
    return _b_py

def dpftri(transr, uplo, n, ar):
    u"""
    Inverse of real symmetric positive definite matrix, matrix already
    factorized by ``dpftrf``, Rectangular Full Packed format.

    ``dpftri`` computes the inverse of a real symmetric positive
    definite matrix using the Cholesky factorization computed by
    ``dpftrf`` stored in Rectangular Full Packed (RFP) format.

    For full information please refer to the NAG Library document for
    f07wj

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07wjf.html

    Parameters
    ----------
    transr : str, length 1
        Specifies whether the RFP representation of A is normal or
        transposed.

        `transr` = 'N'

            The matrix A is stored in normal RFP format.

        `transr` = 'T'

            The matrix A is stored in transposed RFP format.

    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = U^TU, where U is upper triangular.

        `uplo` = 'L'

            A = LL^T, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    ar : float, array-like, shape (`n`*(`n`+1)/2)
        The Cholesky factorization of A stored in RFP format, as
        returned by ``dpftrf``.

    Returns
    -------
    ar : float, ndarray, shape (`n`*(`n`+1)/2)
        The factorization is overwritten by the n by n matrix A^-1
        stored in RFP format.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `transr`.

            Constraint: `transr` = 'N' or 'T'.

        (`errno` -2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            The leading minor of order *<value>* is not positive
            definite and the factorization could not be completed. Hence
            A itself is not positive definite. This may indicate an
            error in forming the matrix A. There is no function
            specifically designed to invert a symmetric matrix stored in
            RFP format which is not positive definite; the matrix must
            be treated as a full symmetric matrix, by calling
            ``dsytri``.

    References
    ----------
    Du Croz, J J and Higham, N J, 1992, `Stability of methods for matrix
    inversion`, IMA J. Numer. Anal. (12), 1--19

    Gustavson, F G, Waśniewski, J, Dongarra, J J and Langou, J, 2010,
    `Rectangular full packed format for Cholesky's algorithm:
    factorization, solution, and inversion`, ACM Trans. Math. Software
    (37, 2)
    """

    fun_name = 'naginterfaces.library.lapacklin.dpftri'

    _ar = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ar'},
        ar,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    try:
        _b_lapacklin.dpftri(
            transr,
            uplo,
            n,
            _ar,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _ar_py = _ar
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['ar'],
    )(
        _ar_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _ar_py

def dtftri(transr, uplo, diag, n, ar):
    u"""
    Inverse of real triangular matrix, Rectangular Full Packed format.

    ``dtftri`` computes the inverse of a real triangular matrix stored
    in Rectangular Full Packed (RFP) format.

    For full information please refer to the NAG Library document for
    f07wk

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07wkf.html

    Parameters
    ----------
    transr : str, length 1
        Specifies whether the RFP representation of A is normal or
        transposed.

        `transr` = 'N'

            The matrix A is stored in normal RFP format.

        `transr` = 'T'

            The matrix A is stored in transposed RFP format.

    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

            A is upper triangular.

        `uplo` = 'L'

            A is lower triangular.

    diag : str, length 1
        Indicates whether A is a nonunit or unit triangular matrix.

        `diag` = 'N'

            A is a nonunit triangular matrix.

        `diag` = 'U'

            A is a unit triangular matrix; the diagonal elements are not
            referenced and are assumed to be 1.

    n : int
        n, the order of the matrix A.

    ar : float, array-like, shape (`n`*(`n`+1)/2)
        The upper or lower triangular part (as specified by `uplo`) of
        the n by n symmetric matrix A, in either normal or transposed
        RFP format (as specified by `transr`). The storage format is
        described in detail in [the F07 Introduction].

    Returns
    -------
    ar : float, ndarray, shape (`n`*(`n`+1)/2)
        A is overwritten by A^-1, in the same storage format as A.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `transr`.

            Constraint: `transr` = 'N' or 'T'.

        (`errno` -2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` -4)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Diagonal element *<value>* of A is exactly zero. A is
            singular its inverse cannot be computed.

    References
    ----------
    Du Croz, J J and Higham, N J, 1992, `Stability of methods for matrix
    inversion`, IMA J. Numer. Anal. (12), 1--19

    Gustavson, F G, Waśniewski, J, Dongarra, J J and Langou, J, 2010,
    `Rectangular full packed format for Cholesky's algorithm:
    factorization, solution, and inversion`, ACM Trans. Math. Software
    (37, 2)
    """

    fun_name = 'naginterfaces.library.lapacklin.dtftri'

    _ar = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ar'},
        ar,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    try:
        _b_lapacklin.dtftri(
            transr,
            uplo,
            diag,
            n,
            _ar,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _ar_py = _ar
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['ar'],
    )(
        _ar_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _ar_py

def zpftrf(transr, uplo, n, ar):
    u"""
    Cholesky factorization of complex Hermitian positive definite
    matrix, Rectangular Full Packed format.

    ``zpftrf`` computes the Cholesky factorization of a complex
    Hermitian positive definite matrix stored in Rectangular Full Packed
    (RFP) format.

    For full information please refer to the NAG Library document for
    f07wr

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07wrf.html

    Parameters
    ----------
    transr : str, length 1
        Specifies whether the normal RFP representation of A or its
        conjugate transpose is stored.

        `transr` = 'N'

            The matrix A is stored in normal RFP format.

        `transr` = 'C'

            The conjugate transpose of the RFP representation of the
            matrix A is stored.

    uplo : str, length 1
        Specifies whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

            The upper triangular part of A is stored, and A is
            factorized as U^HU, where U is upper triangular.

        `uplo` = 'L'

            The lower triangular part of A is stored, and A is
            factorized as LL^H, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    ar : complex, array-like, shape (`n`*(`n`+1)/2)
        The upper or lower triangular part (as specified by `uplo`) of
        the n by n Hermitian matrix A, in either normal or transposed
        RFP format (as specified by `transr`). The storage format is
        described in detail in [the F07 Introduction].

    Returns
    -------
    ar : complex, ndarray, shape (`n`*(`n`+1)/2)
        If no exception or warning is raised, the factor U or L from the
        Cholesky factorization A = U^HU or A = LL^H, in the same storage
        format as A.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `transr`.

            Constraint: `transr` = 'N' or 'C'.

        (`errno` -2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` i > 0)
            The leading minor of order *<value>* is not positive
            definite and the factorization could not be completed. Hence
            A itself is not positive definite. This may indicate an
            error in forming the matrix A. There is no function
            specifically designed to factorize a Hermitian matrix stored
            in RFP format which is not positive definite; the matrix
            must be treated as a full Hermitian matrix, by calling
            ``zhetrf``.

    References
    ----------
    Demmel, J W, 1989, `On floating-point errors in Cholesky`, LAPACK
    Working Note No. 14, University of Tennessee, Knoxville,
    https://www.netlib.org/lapack/lawnspdf/lawn14.pdf

    Gustavson, F G, Waśniewski, J, Dongarra, J J and Langou, J, 2010,
    `Rectangular full packed format for Cholesky's algorithm:
    factorization, solution, and inversion`, ACM Trans. Math. Software
    (37, 2)
    """

    fun_name = 'naginterfaces.library.lapacklin.zpftrf'

    _ar = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ar'},
        ar,
        copy=True,
    )
    _b_lapacklin.zpftrf(
        transr,
        uplo,
        n,
        _ar,
    )
    _ar_py = _ar
    return _ar_py

def zpftrs(transr, uplo, n, ar, b):
    u"""
    Solution of complex Hermitian positive definite system of linear
    equations, multiple right-hand sides, coefficient matrix already
    factorized by ``zpftrf``, Rectangular Full Packed format.

    ``zpftrs`` solves a complex Hermitian positive definite system of
    linear equations with multiple right-hand sides,

        AX = B,

    using the Cholesky factorization computed by ``zpftrf`` stored in
    Rectangular Full Packed (RFP) format.

    For full information please refer to the NAG Library document for
    f07ws

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07wsf.html

    Parameters
    ----------
    transr : str, length 1
        Specifies whether the normal RFP representation of A or its
        conjugate transpose is stored.

        `transr` = 'N'

            The matrix A is stored in normal RFP format.

        `transr` = 'C'

            The conjugate transpose of the RFP representation of the
            matrix A is stored.

    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = U^HU, where U is upper triangular.

        `uplo` = 'L'

            A = LL^H, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    ar : complex, array-like, shape (`n`*(`n`+1)/2)
        The Cholesky factorization of A stored in RFP format, as
        returned by ``zpftrf``.

    b : complex, array-like, shape (`n`, nrhs)
        The n by r right-hand side matrix B.

    Returns
    -------
    b : complex, ndarray, shape (`n`, nrhs)
        The n by r solution matrix X.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `transr`.

            Constraint: `transr` = 'N' or 'C'.

        (`errno` -2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` -4)
            On entry, error in parameter nrhs.

            Constraint: nrhs >= 0.

    References
    ----------
    Gustavson, F G, Waśniewski, J, Dongarra, J J and Langou, J, 2010,
    `Rectangular full packed format for Cholesky's algorithm:
    factorization, solution, and inversion`, ACM Trans. Math. Software
    (37, 2)
    """

    fun_name = 'naginterfaces.library.lapacklin.zpftrs'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
    )
    _nrhs_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _nrhs = _b_utils._EngineIntScalarType(
        _nrhs_py, {'fun_name': fun_name, 'entity': '_nrhs'},
    ).to_ctypes()
    _ar = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ar'},
        ar,
        copy=False,
    )
    _b_lapacklin.zpftrs(
        transr,
        uplo,
        n,
        _nrhs,
        _ar,
        _b,
    )
    _b_py = _b
    return _b_py

def zpftri(transr, uplo, n, ar):
    u"""
    Inverse of complex Hermitian positive definite matrix, matrix
    already factorized by ``zpftrf``, Rectangular Full Packed format.

    ``zpftri`` computes the inverse of a complex Hermitian positive
    definite matrix using the Cholesky factorization computed by
    ``zpftrf`` stored in Rectangular Full Packed (RFP) format.

    For full information please refer to the NAG Library document for
    f07ww

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07wwf.html

    Parameters
    ----------
    transr : str, length 1
        Specifies whether the normal RFP representation of A or its
        conjugate transpose is stored.

        `transr` = 'N'

            The matrix A is stored in normal RFP format.

        `transr` = 'C'

            The conjugate transpose of the RFP representation of the
            matrix A is stored.

    uplo : str, length 1
        Specifies how A has been factorized.

        `uplo` = 'U'

            A = U^HU, where U is upper triangular.

        `uplo` = 'L'

            A = LL^H, where L is lower triangular.

    n : int
        n, the order of the matrix A.

    ar : complex, array-like, shape (`n`*(`n`+1)/2)
        The Cholesky factorization of A stored in RFP format, as
        returned by ``zpftrf``.

    Returns
    -------
    ar : complex, ndarray, shape (`n`*(`n`+1)/2)
        The factorization is overwritten by the n by n matrix A^-1
        stored in RFP format.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `transr`.

            Constraint: `transr` = 'N' or 'C'.

        (`errno` -2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

        (`errno` i > 0)
            The leading minor of order *<value>* is not positive
            definite and the factorization could not be completed. Hence
            A itself is not positive definite. This may indicate an
            error in forming the matrix A. There is no function
            specifically designed to invert a Hermitian matrix stored in
            RFP format which is not positive definite; the matrix must
            be treated as a full Hermitian matrix, by calling
            ``zhetri``.

    References
    ----------
    Du Croz, J J and Higham, N J, 1992, `Stability of methods for matrix
    inversion`, IMA J. Numer. Anal. (12), 1--19

    Gustavson, F G, Waśniewski, J, Dongarra, J J and Langou, J, 2010,
    `Rectangular full packed format for Cholesky's algorithm:
    factorization, solution, and inversion`, ACM Trans. Math. Software
    (37, 2)
    """

    fun_name = 'naginterfaces.library.lapacklin.zpftri'

    _ar = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ar'},
        ar,
        copy=True,
    )
    _b_lapacklin.zpftri(
        transr,
        uplo,
        n,
        _ar,
    )
    _ar_py = _ar
    return _ar_py

def ztftri(transr, uplo, diag, n, ar):
    u"""
    Inverse of complex triangular matrix, Rectangular Full Packed
    format.

    ``ztftri`` computes the inverse of a complex triangular matrix
    stored in Rectangular Full Packed (RFP) format.

    For full information please refer to the NAG Library document for
    f07wx

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f07/f07wxf.html

    Parameters
    ----------
    transr : str, length 1
        Specifies whether the normal RFP representation of A or its
        conjugate transpose is stored.

        `transr` = 'N'

            The matrix A is stored in normal RFP format.

        `transr` = 'C'

            The conjugate transpose of the RFP representation of the
            matrix A is stored.

    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

            A is upper triangular.

        `uplo` = 'L'

            A is lower triangular.

    diag : str, length 1
        Indicates whether A is a nonunit or unit triangular matrix.

        `diag` = 'N'

            A is a nonunit triangular matrix.

        `diag` = 'U'

            A is a unit triangular matrix; the diagonal elements are not
            referenced and are assumed to be 1.

    n : int
        n, the order of the matrix A.

    ar : complex, array-like, shape (`n`*(`n`+1)/2)
        The upper or lower triangular part (as specified by `uplo`) of
        the n by n Hermitian matrix A, in either normal or transposed
        RFP format (as specified by `transr`). The storage format is
        described in detail in [the F07 Introduction].

    Returns
    -------
    ar : complex, ndarray, shape (`n`*(`n`+1)/2)
        A is overwritten by A^-1, in the same storage format as A.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `transr`.

            Constraint: `transr` = 'N' or 'C'.

        (`errno` -2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -3)
            On entry, error in parameter `diag`.

            Constraint: `diag` = 'N' or 'U'.

        (`errno` -4)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i > 0)
            Diagonal element *<value>* of A is exactly zero. A is
            singular its inverse cannot be computed.

    References
    ----------
    Du Croz, J J and Higham, N J, 1992, `Stability of methods for matrix
    inversion`, IMA J. Numer. Anal. (12), 1--19

    Gustavson, F G, Waśniewski, J, Dongarra, J J and Langou, J, 2010,
    `Rectangular full packed format for Cholesky's algorithm:
    factorization, solution, and inversion`, ACM Trans. Math. Software
    (37, 2)
    """

    fun_name = 'naginterfaces.library.lapacklin.ztftri'

    _ar = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ar'},
        ar,
        copy=True,
    )
    try:
        _b_lapacklin.ztftri(
            transr,
            uplo,
            diag,
            n,
            _ar,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _ar_py = _ar
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['ar'],
    )(
        _ar_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _ar_py
