# -*- coding: utf-8 -*-
# pylint: disable=chained-comparison,invalid-name,line-too-long,protected-access,too-many-arguments,too-many-boolean-expressions,too-many-branches,too-many-lines,too-many-locals,too-many-statements,unused-import
u"""
Interfaces for the NAG Mark 27.0 `matop` Chapter.

``matop`` - Matrix Operations, Including Inversion

This module provides facilities for four types of problem:

(i) matrix inversion;

(#) matrix factorizations;

(#) matrix arithmetic and manipulation;

(#) matrix functions.

See [Background to the Problems] where these problems are discussed.

For full information please refer to the NAG Library document

https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01intro.html

See Also
--------
``naginterfaces.library.examples.matop`` :
    This subpackage contains examples for the ``matop`` module.
    See also the :ref:`library_matop_ex` subsection.
"""

# NAG Copyright 2017-2020.

import collections as _collections

from ..base import matop as _b_matop
from ..base import utils as _b_utils
from ..library import _utils

def real_symm_posdef_inv(a):
    u"""
    Inverse of real symmetric positive definite matrix using iterative
    refinement.

    ``real_symm_posdef_inv`` calculates the accurate inverse of a real
    symmetric positive definite matrix, using a Cholesky factorization
    and iterative refinement.

    For full information please refer to the NAG Library document for
    f01ab

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01abf.html

    Parameters
    ----------
    a : float, array-like, shape (n+1, n)
        The upper triangle of the n by n positive definite symmetric
        matrix A. The elements of the array below the diagonal need not
        be set.

    Returns
    -------
    a : float, ndarray, shape (n+1, n)
        The lower triangle of the inverse matrix X is stored in the
        elements of the array below the diagonal, in rows 2 to n+1;
        x_{ij} is stored in `a`[i,j-1] for i >= j. The upper triangle of
        the original matrix is unchanged.

    b : float, ndarray, shape (n, n)
        The lower triangle of the inverse matrix X, with x_{ij} stored
        in `b`[i-1,j-1], for i >= j.

    Raises
    ------
    NagValueError
        (`errno` 1)
            The matrix A is not positive definite, possibly due to
            rounding errors.

        (`errno` 2)
            The refinement process failed to converge. The matrix A is
            ill-conditioned.

        (`errno` 3)
            On entry, n = *<value>*.

            Constraint: n >= 1.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Wilkinson, J H and Reinsch, C, 1971, `Handbook for Automatic
    Computation II, Linear Algebra`, Springer--Verlag
    """

    fun_name = 'naginterfaces.library.matop.real_symm_posdef_inv'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_a, 'a'),),
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py, _n_py),
        sorder=_m_sorder,
    )
    _b_matop.real_symm_posdef_inv(
        _a,
        _n,
        _b,
    )
    _a_py = _a
    _b_py = _b
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'b'],
    )(
        _a_py,
        _b_py,
    )
    return _return_data

def real_symm_posdef_inv_noref(a):
    u"""
    Inverse of real symmetric positive definite matrix.

    ``real_symm_posdef_inv_noref`` calculates the approximate inverse of
    a real symmetric positive definite matrix, using a Cholesky
    factorization.

    For full information please refer to the NAG Library document for
    f01ad

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01adf.html

    Parameters
    ----------
    a : float, array-like, shape (n+1, n)
        The upper triangle of the n by n positive definite symmetric
        matrix A. The elements of the array below the diagonal need not
        be set.

    Returns
    -------
    a : float, ndarray, shape (n+1, n)
        The lower triangle of the inverse matrix X is stored in the
        elements of the array below the diagonal, in rows 2 to n+1;
        x_{ij} is stored in `a`[i,j-1] for i >= j. The upper triangle of
        the original matrix is unchanged.

    Raises
    ------
    NagValueError
        (`errno` 1)
            The matrix A is not positive definite, possibly due to
            rounding errors.

        (`errno` 2)
            On entry, n = *<value>*.

            Constraint: n >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Wilkinson, J H and Reinsch, C, 1971, `Handbook for Automatic
    Computation II, Linear Algebra`, Springer--Verlag
    """

    fun_name = 'naginterfaces.library.matop.real_symm_posdef_inv_noref'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b_matop.real_symm_posdef_inv_noref(
        _n,
        _a,
    )
    _a_py = _a
    return _a_py

def real_gen_pseudinv(t, a):
    u"""
    Pseudo-inverse and rank of real m by n matrix (m >= n).

    ``real_gen_pseudinv`` calculates the rank and pseudo-inverse of an m
    by n real matrix, m >= n, using a QR factorization with column
    interchanges.

    For full information please refer to the NAG Library document for
    f01bl

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01blf.html

    Parameters
    ----------
    t : float
        The tolerance used to decide when elements can be regarded as
        zero (see [Further Comments]).

    a : float, array-like, shape (m, n)
        The m by n rectangular matrix A.

    Returns
    -------
    a : float, ndarray, shape (m, n)
        The transpose of the pseudo-inverse of A.

    aijmax : float, ndarray, shape (n)
        `aijmax`[i-1] contains the element of largest modulus in the
        reduced matrix at the ith stage. If r < n, then only the first
        r+1 elements of `aijmax` have values assigned to them; the
        remaining elements are unused. The ratio
        `aijmax`[0]/`aijmax`[r-1] usually gives an indication of the
        condition number of the original matrix (see [Further
        Comments]).

    irank : int
        r, the rank of A as determined using the tolerance `t`.

    inc : int, ndarray, shape (n)
        The record of the column interchanges in the Householder
        factorization.

    Raises
    ------
    NagValueError
        (`errno` 1)
            Inverse not found. Incorrect `irank`.

        (`errno` 2)
            Invalid tolerance, `t` too large: `t` = *<value>*.

        (`errno` 2)
            Invalid tolerance, `t` < 0.0: `t` = *<value>*.

        (`errno` 3)
            On entry, m = *<value>* and n = *<value>*.

            Constraint: m >= n.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Peters, G and Wilkinson, J H, 1970, `The least squares problem and
    pseudo-inverses`, Comput. J. (13), 309--316

    Wilkinson, J H and Reinsch, C, 1971, `Handbook for Automatic
    Computation II, Linear Algebra`, Springer--Verlag
    """

    fun_name = 'naginterfaces.library.matop.real_gen_pseudinv'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _aijmax = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _inc = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n_py,),
    )
    _return_scalar = _b_matop.real_gen_pseudinv(
        _m,
        _n,
        t,
        _a,
        _aijmax,
        _inc,
    )
    _a_py = _a
    _aijmax_py = _aijmax
    _irank_py = _return_scalar
    _inc_py = _inc
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'aijmax', 'irank', 'inc'],
    )(
        _a_py,
        _aijmax_py,
        _irank_py,
        _inc_py,
    )
    return _return_data

def real_gen_sparse_lu(n, nz, warnlev, errlev, a, irn, icn, abort, pivot=0.1, lblock=True, grow=True, io_manager=None):
    u"""
    LU factorization of real sparse matrix.

    ``real_gen_sparse_lu`` factorizes a real sparse matrix.
    The function either forms the LU factorization of a permutation of
    the entire matrix, or, optionally, first permutes the matrix to
    block lower triangular form and then only factorizes the diagonal
    blocks.

    For full information please refer to the NAG Library document for
    f01br

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01brf.html

    Parameters
    ----------
    n : int
        n, the order of the matrix A.

    nz : int
        The number of nonzero elements in the matrix A.

    warnlev : int
        Setting `warnlev` = 0 disables output of warning messages during
        the process of the computation. Setting `warnlev` = 1 enables
        these messages.

    errlev : int
        Setting `errlev` = 0 disables output of error messages during
        the process of the computation. Setting `errlev` = 1 enables
        these messages.

    a : float, array-like, shape (licn)
        `a`[i-1], for i = 1,2,...,`nz`, must contain the nonzero
        elements of the sparse matrix A. They can be in any order since
        ``real_gen_sparse_lu`` will reorder them.

    irn : int, array-like, shape (lirn)
        `irn`[i-1], for i = 1,2,...,`nz`, must contain the row index of
        the nonzero element stored in `a`[i-1].

    icn : int, array-like, shape (licn)
        `icn`[i-1], for i = 1,2,...,`nz`, must contain the column index
        of the nonzero element stored in `a`[i-1].

    abort : bool, array-like, shape (4)
        If `abort`[0] = True, ``real_gen_sparse_lu`` will exit
        immediately on detecting a structural singularity (one that
        depends on the pattern of nonzeros) and return `errno` = 1;
        otherwise it will complete the factorization (see [Further
        Comments]).

        If `abort`[1] = True, ``real_gen_sparse_lu`` will exit
        immediately on detecting a numerical singularity (one that
        depends on the numerical values) and return `errno` = 2;
        otherwise it will complete the factorization (see [Further
        Comments]).

        If `abort`[2] = True, ``real_gen_sparse_lu`` will exit
        immediately (with `errno` = 5) when the arrays `a` and
        `comm`['icn'] are filled up by the previously factorized, active
        and unfactorized parts of the matrix; otherwise it continues so
        that better guidance on necessary array sizes can be given in
        `comm`['idisp'][5] and `comm`['idisp'][6], and will exit with
        errno in the range 4 to 6.

        Note that there is always an immediate error exit if the array
        `irn` is too small.

        If `abort`[3] = True, ``real_gen_sparse_lu`` exits immediately
        (with `errno` = 13) if it finds duplicate elements in the input
        matrix.

        If `abort`[3] = False, ``real_gen_sparse_lu`` proceeds using a
        value equal to the sum of the duplicate elements.

        In either case details of each duplicate element are output on
        the file object associated with the advisory I/O unit (see
        :class:`~naginterfaces.base.utils.FileObjManager`), unless
        suppressed by the value of errno on entry.

        Suggested value:

            `abort`[0] = True

            `abort`[1] = True

            `abort`[2] = False

            `abort`[3] = True

    pivot : float, optional
        Should have a value in the range 0.0 <= `pivot` <= 0.9999 and is
        used to control the choice of pivots. If `pivot` < 0.0, the
        value 0.0 is assumed, and if `pivot` > 0.9999, the value 0.9999
        is assumed. When searching a row for a pivot, any element is
        excluded which is less than `pivot` times the largest of those
        elements in the row available as pivots. Thus decreasing `pivot`
        biases the algorithm to maintaining sparsity at the expense of
        stability.

    lblock : bool, optional
        If `lblock` = True, the matrix is preordered into block lower
        triangular form before the LU factorization is performed;
        otherwise the entire matrix is factorized.

    grow : bool, optional
        If `grow` = True, then on exit `w`[0] contains an estimate (an
        upper bound) of the increase in size of elements encountered
        during the factorization. If the matrix is well-scaled (see
        [Further Comments]), then a high value for `w`[0] indicates that
        the LU factorization may be inaccurate and you should be wary of
        the results and perhaps increase the argument `pivot` for
        subsequent runs (see [Accuracy]).

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    a : float, ndarray, shape (licn)
        The nonzero elements in the LU factorization. The array must
        **not** be changed by you between a call of
        ``real_gen_sparse_lu`` and a call of
        ``linsys.real_sparse_fac_solve``.

    irn : int, ndarray, shape (lirn)
        `irn` is overwritten and is not needed for subsequent calls of
        ``real_gen_sparse_lu_reuse`` or
        ``linsys.real_sparse_fac_solve``.

    comm : dict, communication object
        Information about the factorization for subsequent communication
        to ``real_gen_sparse_lu_reuse`` or
        ``linsys.real_sparse_fac_solve``.

    w : float, ndarray, shape (`n`)
        If `grow` = True, `w`[0] contains an estimate (an upper bound)
        of the increase in size of elements encountered during the
        factorization (see `grow`); the rest of the array is used as
        workspace.

        If `grow` = False, the array is not used.

    Raises
    ------
    NagValueError
        (`errno` 1)
            Matrix is structurally singular -- decomposition aborted
            RANK = *<value>*.

        (`errno` 1)
            Matrix is structurally singular -- decomposition aborted.

        (`errno` 2)
            Matrix is numerically singular -- decomposition aborted.

        (`errno` 3)
            lirn too small. Decomposition aborted at stage *<value>* in
            block *<value>* with first row *<value>* and last row
            *<value>*. To continue set lirn to at least *<value>*.

        (`errno` 3)
            lirn too small. Decomposition aborted at stage *<value>* in
            block *<value>* with first row *<value>* and last row
            *<value>*.

        (`errno` 4)
            licn much too small. Decomposition aborted at stage
            *<value>* in block *<value>* with first row *<value>* and
            last row *<value>*. To continue set licn to at least
            *<value>*.

        (`errno` 4)
            licn much too small. Decomposition aborted at stage
            *<value>* in block *<value>* with first row *<value>* and
            last row *<value>*.

        (`errno` 5)
            licn too small. Decomposition aborted at stage *<value>* in
            block *<value>* with first row *<value>* and last row
            *<value>*.

        (`errno` 5)
            licn too small. Decomposition aborted at stage *<value>* in
            block *<value>* with first row *<value>* and last row
            *<value>*. To continue set licn to at least *<value>*.

        (`errno` 5)
            licn too small. For successful decomposition set licn to at
            least *<value>*.

        (`errno` 6)
            licn and lirn too small. Decomposition aborted at stage
            *<value>* in block *<value>* with first row *<value>* and
            last row *<value>*. To continue set lirn to at least
            *<value>*.

        (`errno` 6)
            licn and lirn too small. Decomposition aborted at stage
            *<value>* in block *<value>* with first row *<value>* and
            last row *<value>*.

        (`errno` 7)
            licn not big enough for permutation -- increase by
            *<value>*.

        (`errno` 8)
            On entry, `n` = *<value>*.

            Constraint: `n` > 0.

        (`errno` 9)
            On entry, `nz` = *<value>*.

            Constraint: `nz` > 0.

        (`errno` 10)
            On entry, licn = *<value>* and `nz` = *<value>*.

            Constraint: licn >= `nz`.

        (`errno` 11)
            On entry, lirn = *<value>* and `nz` = *<value>*.

            Constraint: lirn >= `nz`.

        (`errno` 12)
            On entry, `irn`[I-1] or `icn`[I-1] is out of range: I =
            *<value>*, `a`[I-1] = *<value>* `irn`[I-1] = *<value>*,
            `icn`[I-1] = *<value>*.

        (`errno` 13)
            On entry, duplicate elements found -- see advisory messages.

        (`errno` 14)
            On entry, `warnlev` = *<value>*.

            Constraint: `warnlev` = 0 or 1.

        (`errno` 15)
            On entry, `errlev` = *<value>*.

            Constraint: `errlev` = 0 or 1.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` -2)
            Matrix is numerically singular -- decomposition completed.

        (`errno` -1)
            Matrix is structurally singular -- decomposition completed.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Duff, I S, 1977, `MA28 -- a set of Fortran subroutines for sparse
    unsymmetric linear equations`, AERE Report R8730, HMSO
    """

    fun_name = 'naginterfaces.library.matop.real_gen_sparse_lu'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _licn_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=1)[0]
    _licn = _b_utils._EngineIntScalarType(
        _licn_py, {'fun_name': fun_name, 'entity': '_licn'},
    ).to_ctypes()
    _irn = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'irn'},
        irn,
        copy=True,
    )
    _lirn_py = _b_utils._EngineIntArrayType.get_shape(_irn, {'fun_name': fun_name, 'entity': 'irn'}, exp_rank=1)[0]
    _lirn = _b_utils._EngineIntScalarType(
        _lirn_py, {'fun_name': fun_name, 'entity': '_lirn'},
    ).to_ctypes()
    _icn = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'icn'},
        icn,
        copy=True,
    )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _w = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _abort = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'abort'},
        abort,
        copy=False,
    )
    try:
        _return_scalar = _b_matop.real_gen_sparse_lu(
            _n,
            nz,
            warnlev,
            errlev,
            _a,
            _licn,
            _irn,
            _lirn,
            _icn,
            pivot,
            _w,
            lblock,
            grow,
            _abort,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _a_py = _a
    _irn_py = _irn
    _comm_py = _return_scalar
    _w_py = _w
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'irn', 'comm', 'w'],
    )(
        _a_py,
        _irn_py,
        _comm_py,
        _w_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def real_gen_sparse_lu_reuse(warnlev, a, ivect, jvect, comm, grow, eta=0.0001, abort=True, io_manager=None):
    u"""
    LU factorization of real sparse matrix with known sparsity pattern.

    ``real_gen_sparse_lu_reuse`` factorizes a real sparse matrix using
    the pivotal sequence previously obtained by ``real_gen_sparse_lu``
    when a matrix of the same sparsity pattern was factorized.

    For full information please refer to the NAG Library document for
    f01bs

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01bsf.html

    Parameters
    ----------
    warnlev : int
        Setting `warnlev` = 0 disables output of warning messages during
        the process of the computation. Setting `warnlev` = 1 enables
        these messages.

    a : float, array-like, shape (licn)
        `a`[i-1], for i = 1,2,...,nz, must contain the nonzero elements
        of the sparse matrix A. They can be in any order since
        ``real_gen_sparse_lu_reuse`` will reorder them.

    ivect : int, array-like, shape (nz)
        `ivect`[i-1] and `jvect`[i-1], for i = 1,2,...,nz, must contain
        the row index and the column index respectively of the nonzero
        element stored in `a`[i-1].

    jvect : int, array-like, shape (nz)
        `ivect`[i-1] and `jvect`[i-1], for i = 1,2,...,nz, must contain
        the row index and the column index respectively of the nonzero
        element stored in `a`[i-1].

    comm : dict, communication object
        Communication structure.

        This argument must have been initialized by a prior call to
        ``real_gen_sparse_lu``.

    grow : bool
        If `grow` = True, then on exit `w`[0] contains an estimate (an
        upper bound) of the increase in size of elements encountered
        during the factorization. If the matrix is well-scaled (see
        [Further Comments]), then a high value for `w`[0] indicates that
        the LU factorization may be inaccurate and you should be wary of
        the results and perhaps increase the argument pivot for
        subsequent runs (see [Accuracy]).

    eta : float, optional
        The relative pivot threshold below which an error diagnostic is
        provoked and errno is set to `errno` = 7. If `eta` is greater
        than 1.0, then no check on pivot size is made.

    abort : bool, optional
        If `abort` = True, ``real_gen_sparse_lu_reuse`` exits
        immediately (with `errno` = 8) if it finds duplicate elements in
        the input matrix.

        If `abort` = False, ``real_gen_sparse_lu_reuse`` proceeds using
        a value equal to the sum of the duplicate elements.

        In either case details of each duplicate element are output on
        the file object associated with the advisory I/O unit (see
        :class:`~naginterfaces.base.utils.FileObjManager`), unless
        suppressed by the value of errno on entry.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    a : float, ndarray, shape (licn)
        The nonzero elements in the LU factorization. The array must
        **not** be changed by you between a call of
        ``real_gen_sparse_lu_reuse`` and a call of
        ``linsys.real_sparse_fac_solve``.

    w : float, ndarray, shape (n)
        If `grow` = True, `w`[0] contains an estimate (an upper bound)
        of the increase in size of elements encountered during the
        factorization (see `grow`); the rest of the array is used as
        workspace.

        If `grow` = False, the array is not used.

    rpmin : float
        If `eta` is less than 1.0, then `rpmin` gives the smallest ratio
        of the pivot to the largest element in the row of the
        corresponding upper triangular factor thus monitoring the
        stability of the factorization. If `rpmin` is very small it may
        be advisable to perform a new factorization using
        ``real_gen_sparse_lu``.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n > 0.

        (`errno` 2)
            On entry, nz = *<value>*.

            Constraint: nz > 0.

        (`errno` 3)
            On entry, licn = *<value>* and nz = *<value>*.

            Constraint: licn >= nz.

        (`errno` 4)
            On entry, irn[I-1] in ``real_gen_sparse_lu`` or
            `comm`['icn'][I-1] is out of range: I = *<value>*, `a`[I-1]
            = *<value>* irn[I-1] = *<value>* in ``real_gen_sparse_lu``,
            `comm`['icn'][I-1] = *<value>*.

        (`errno` 5)
            Nonzero element (*<value>*, *<value>*) was not in L/U
            pattern.

        (`errno` 5)
            Nonzero element (*<value>*, *<value>*) in zero off-diagonal
            block.

        (`errno` 6)
            Numerical singularity in row *<value>* - decomposition
            aborted.

        (`errno` 8)
            On entry, duplicate elements found -- see advisory messages.

        (`errno` 9)
            On entry, `warnlev` = *<value>*.

            Constraint: `warnlev` = 0 or 1.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 7)
            Subthreshold pivot in row *<value>* - decomposition
            completed.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Duff, I S, 1977, `MA28 -- a set of Fortran subroutines for sparse
    unsymmetric linear equations`, AERE Report R8730, HMSO
    """

    fun_name = 'naginterfaces.library.matop.real_gen_sparse_lu_reuse'

    _b_utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        req_keys=('icn', 'ikeep', 'idisp'),
    )
    _n = _b_utils._EngineIntScalarType(
        _b_utils._EngineIntArrayType.get_shape(comm['ikeep'], {'fun_name': fun_name, 'entity': 'comm[\'ikeep\']'}, exp_rank=1, validate=False)[0]//5, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _ivect = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ivect'},
        ivect,
        copy=False,
    )
    _nz_py = _b_utils._EngineIntArrayType.get_shape(_ivect, {'fun_name': fun_name, 'entity': 'ivect'}, exp_rank=1)[0]
    _nz = _b_utils._EngineIntScalarType(
        _nz_py, {'fun_name': fun_name, 'entity': '_nz'},
    ).to_ctypes()
    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _licn_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=1)[0]
    _licn = _b_utils._EngineIntScalarType(
        _licn_py, {'fun_name': fun_name, 'entity': '_licn'},
    ).to_ctypes()
    _jvect = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'jvect'},
        jvect,
        copy=False,
    )
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _w = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _return_scalar = _b_matop.real_gen_sparse_lu_reuse(
            _n,
            _nz,
            warnlev,
            _a,
            _licn,
            _ivect,
            _jvect,
            comm,
            _w,
            grow,
            eta,
            abort,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _a_py = _a
    _w_py = _w
    _rpmin_py = _return_scalar
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'w', 'rpmin'],
    )(
        _a_py,
        _w_py,
        _rpmin_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def real_symm_posdef_fac(k, a):
    u"""
    ULDL^TU^T factorization of real symmetric positive definite band
    matrix.

    ``real_symm_posdef_fac`` performs a ULDL^TU^T decomposition of a
    real symmetric positive definite band matrix.

    For full information please refer to the NAG Library document for
    f01bu

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01buf.html

    Parameters
    ----------
    k : int
        k, the change-over point in the decomposition.

    a : float, array-like, shape (m1, n)
        The upper triangle of the n by n symmetric band matrix A, with
        the diagonal of the matrix stored in the (m+1)th row of the
        array, and the m superdiagonals within the band stored in the
        first m rows of the array. Each column of the matrix is stored
        in the corresponding column of the array. For example, if n = 6
        and m = 2, the storage scheme is

            [table omitted]

        Elements in the top left corner of the array are not used. The
        matrix elements within the band can be assigned to the correct
        elements of the array using the following code:

        ::

          for j in range(A.shape[1]):
              for i in range(max(0, j+1-m1), j+1):
                  a[i-j+m1-1, j] = A[i, j]

    Returns
    -------
    a : float, ndarray, shape (m1, n)
        A is overwritten by the corresponding elements of L, D and U.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `k` = *<value>* and n = *<value>*.

            Constraint: `k` <= n.

        (`errno` 1)
            On entry, `k` = *<value>* and m1-1 = *<value>*.

            Constraint: `k` >= m1-1.

        (`errno` 2)
            The matrix A is not positive definite.

        (`errno` 3)
            The matrix A is not positive definite.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Wilkinson, J H, 1965, `The Algebraic Eigenvalue Problem`, Oxford
    University Press, Oxford

    Wilkinson, J H and Reinsch, C, 1971, `Handbook for Automatic
    Computation II, Linear Algebra`, Springer--Verlag
    """

    fun_name = 'naginterfaces.library.matop.real_symm_posdef_fac'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _m1_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _m1 = _b_utils._EngineIntScalarType(
        _m1_py, {'fun_name': fun_name, 'entity': '_m1'},
    ).to_ctypes()
    _b_matop.real_symm_posdef_fac(
        _n,
        _m1,
        k,
        _a,
    )
    _a_py = _a
    return _a_py

def real_symm_posdef_geneig(k, a, b):
    u"""
    Reduction to standard form, generalized real symmetric-definite
    banded eigenproblem.

    ``real_symm_posdef_geneig`` transforms the generalized
    symmetric-definite eigenproblem Ax = lambda`b`x to the equivalent
    standard eigenproblem Cy = lambda y, where A, `b` and C are
    symmetric band matrices and `b` is positive definite. `b` must have
    been decomposed by ``real_symm_posdef_fac``.

    For full information please refer to the NAG Library document for
    f01bv

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01bvf.html

    Parameters
    ----------
    k : int
        k, the change-over point in the transformations. It must be the
        same as the value used by ``real_symm_posdef_fac`` in the
        decomposition of B.

        Suggested value: the optimum value is the multiple of m_A
        nearest to n/2.

    a : float, array-like, shape (ma1, n)
        The upper triangle of the n by n symmetric band matrix A, with
        the diagonal of the matrix stored in the (m_A+1)th row of the
        array, and the m_A superdiagonals within the band stored in the
        first m_A rows of the array. Each column of the matrix is stored
        in the corresponding column of the array. For example, if n = 6
        and m_A = 2, the storage scheme is

            [table omitted]

        Elements in the top left corner of the array need not be set.
        The matrix elements within the band can be assigned to the
        correct elements of the array using the following code:

        ::

          for j in range(A.shape[1]):
              for i in range(max(0, j+1-ma1), j+1):
                  a[i-j+ma1-1, j] = A[i, j]

    b : float, array-like, shape (mb1, n)
        The elements of the decomposition of matrix B as returned by
        ``real_symm_posdef_fac``.

    Returns
    -------
    a : float, ndarray, shape (ma1, n)
        Is overwritten by the corresponding elements of C.

    b : float, ndarray, shape (mb1, n)
        The elements of `b` will have been permuted.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, mb1 = *<value>* and ma1 = *<value>*.

            Constraint: mb1 <= ma1.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Crawford, C R, 1973, `Reduction of a band-symmetric generalized
    eigenvalue problem`, Comm. ACM (16), 41--44
    """

    fun_name = 'naginterfaces.library.matop.real_symm_posdef_geneig'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _ma1_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _ma1 = _b_utils._EngineIntScalarType(
        _ma1_py, {'fun_name': fun_name, 'entity': '_ma1'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _mb1_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[0]
    _mb1 = _b_utils._EngineIntScalarType(
        _mb1_py, {'fun_name': fun_name, 'entity': '_mb1'},
    ).to_ctypes()
    _b_matop.real_symm_posdef_geneig(
        _n,
        _ma1,
        _mb1,
        k,
        _a,
        _b,
    )
    _a_py = _a
    _b_py = _b
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'b'],
    )(
        _a_py,
        _b_py,
    )
    return _return_data

def real_gen_matmul(b, c, opt):
    u"""
    Multiplication of real matrices.

    ``real_gen_matmul`` returns with the result of the multiplication of
    two matrices B and C in the matrix A, with the option to overwrite B
    or C.

    For full information please refer to the NAG Library document for
    f01ck

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01ckf.html

    Parameters
    ----------
    b : float, array-like, shape (n, m)
        The n by m matrix B.

    c : float, array-like, shape (m, p)
        The m by p matrix C.

    opt : int
        The value of `opt` determines which array is to contain the
        final result.

        `opt` = 1

            `a` must be distinct from `b` and `c` and, on exit, contains
            the result. `b` and `c` need not be distinct in this case.

        `opt` = 2

            `b` must be distinct from `c` and on exit, contains the
            result. `a` is not used in this case and need not be
            distinct from `b` or `c`.

        `opt` = 3

            `c` must be distinct from `b` and on exit, contains the
            result. `a` is not used in this case and need not be
            distinct from `b` or `c`.

    Returns
    -------
    a : float, ndarray, shape (n, p)
        If `opt` = 1, `a` contains the result of the matrix
        multiplication.

    b : float, ndarray, shape (n, m)
        If `opt` = 2, `b` contains the result of the multiplication.

    c : float, ndarray, shape (m, p)
        If `opt` = 3, `c` contains the result of the multiplication.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, p = *<value>*.

            Constraint: p >= 1.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 1)
            On entry, m = *<value>*.

            Constraint: m >= 1.

        (`errno` 2)
            On entry, m = *<value>* and p = *<value>*.

            Constraint: if `opt` = 2, m = p.

        (`errno` 3)
            On entry, n = *<value>* and m = *<value>*.

            Constraint: if `opt` = 3, n = m.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    fun_name = 'naginterfaces.library.matop.real_gen_matmul'

    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _c = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_b, 'b'), (_c, 'c')),
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _p_py = _b_utils._EngineFloat64ArrayType.get_shape(_c, {'fun_name': fun_name, 'entity': 'c'}, exp_rank=2)[1]
    _p = _b_utils._EngineIntScalarType(
        _p_py, {'fun_name': fun_name, 'entity': '_p'},
    ).to_ctypes()
    _a = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py, _p_py),
        sorder=_m_sorder,
    )
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _b_matop.real_gen_matmul(
        _a,
        _b,
        _c,
        _n,
        _p,
        _m,
        opt,
    )
    _b_py = _b
    _c_py = _c
    _a_py = _a
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'b', 'c'],
    )(
        _a_py,
        _b_py,
        _c_py,
    )
    return _return_data

def real_gen_trans_inplace(a, m, n):
    u"""
    Transposition of a real matrix.

    ``real_gen_trans_inplace`` transposes a rectangular matrix in-situ.

    For full information please refer to the NAG Library document for
    f01cr

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01crf.html

    Parameters
    ----------
    a : float, ndarray, shape (mn), modified in place
        `On entry`: the elements of the m by n matrix A, stored by
        columns.

        `On exit`: the elements of the transpose matrix, also stored by
        columns.

    m : int
        m, the number of rows of the matrix A.

    n : int
        n, the number of columns of the matrix A.

    Raises
    ------
    NagValueError
        (`errno` i < 0)
            A serious internal error has occurred in
            ``real_gen_trans_inplace``.

        (`errno` 1)
            On entry, `m` = *<value>*, `n` = *<value>* and mn =
            *<value>*.

            Constraint: mn = `m`*`n`.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Cate, E G and Twigg, D W, 1977, `Algorithm 513: Analysis of in-situ
    transposition`, ACM Trans. Math. Software (3), 104--110
    """

    fun_name = 'naginterfaces.library.matop.real_gen_trans_inplace'

    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
    )
    _mn_py = _b_utils._EngineFloat64ArrayType.get_shape(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=1)[0]
    _mn = _b_utils._EngineIntScalarType(
        _mn_py, {'fun_name': fun_name, 'entity': '_mn'},
    ).to_ctypes()
    _b_matop.real_gen_trans_inplace(
        a,
        m,
        n,
        _mn,
    )

def real_addsub(transa, transb, m, n, alpha, a, beta, b):
    u"""
    Sum or difference of two real matrices, optional scaling and
    transposition.

    ``real_addsub`` adds two `float` matrices, each one optionally
    transposed and multiplied by a scalar.

    For full information please refer to the NAG Library document for
    f01ct

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01ctf.html

    Parameters
    ----------
    transa : str, length 1
        `transa` and `transb` must specify whether or not the matrix A
        and the matrix B, respectively, are to be transposed before
        addition.

        `transa` or `transb` = 'N'

            The matrix will not be transposed.

        `transa` or `transb` = 'T' or 'C'

            The matrix will be transposed.

    transb : str, length 1
        `transa` and `transb` must specify whether or not the matrix A
        and the matrix B, respectively, are to be transposed before
        addition.

        `transa` or `transb` = 'N'

            The matrix will not be transposed.

        `transa` or `transb` = 'T' or 'C'

            The matrix will be transposed.

    m : int
        m, the number of rows of the matrices A and B or their
        transposes. Also the number of rows of the matrix C.

    n : int
        n, the number of columns of the matrices A and B or their
        transposes. Also the number of columns of the matrix C.

    alpha : float
        The scalar alpha, by which matrix A is multiplied before
        addition.

    a : float, array-like, shape (:, :)
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `transa`='N': `m`; otherwise: `n`.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `transa`='N': `n`; otherwise: `m`.

        If alpha = 0.0, the elements of array `a` need not be assigned.
        If alpha != 0.0, then if `transa` = 'N', the leading m by n part
        of `a` must contain the matrix A, otherwise the leading n by m
        part of `a` must contain the matrix A.

    beta : float
        The scalar beta, by which matrix B is multiplied before
        addition.

    b : float, array-like, shape (:, :)
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `transb`='N': `m`; otherwise: `n`.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `transb`='N': `n`; otherwise: `m`.

        If beta = 0.0, the elements of array `b` need not be assigned.
        If beta != 0.0, then if `transa` = 'N', the leading m by n part
        of `b` must contain the matrix B, otherwise the leading n by m
        part of `b` must contain the matrix B.

    Returns
    -------
    c : float, ndarray, shape (`m`, `n`)
        The elements of the m by n matrix C.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `transb` = *<value>*.

            Constraint: `transb` = 'N', 'T' or 'C'.

        (`errno` 1)
            On entry, `transa` = *<value>*.

            Constraint: `transa` = 'N', 'T' or 'C'.

        (`errno` 2)
            On entry, `m` = *<value>* and `n` = *<value>*.

            Constraint: `m` >= 0 and `n` >= 0.

        (`errno` 3)
            On entry, lda = *<value>* and `n` = *<value>*.

            Constraint: if `transa` = *<value>*, lda >= max(1, `n`).

        (`errno` 3)
            On entry, lda = *<value>* and `m` = *<value>*.

            Constraint: if `transa` = *<value>*, lda >= max(1, `m`).

        (`errno` 4)
            On entry, ldb = *<value>* and `n` = *<value>*.

            Constraint: if `transb` = *<value>*, ldb >= max(1, `n`).

        (`errno` 4)
            On entry, ldb = *<value>* and `m` = *<value>*.

            Constraint: if `transb` = *<value>*, ldb >= max(1, `m`).

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    fun_name = 'naginterfaces.library.matop.real_addsub'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_a, 'a'), (_b, 'b')),
    )
    _m = _b_utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = _b_utils._EngineIntScalarType.to_py(_m)
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _c = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_m_py, _n_py),
        sorder=_m_sorder,
    )
    _b_matop.real_addsub(
        transa,
        transb,
        _m,
        _n,
        alpha,
        _a,
        beta,
        _b,
        _c,
    )
    _c_py = _c
    return _c_py

def complex_addsub(transa, transb, m, n, alpha, a, beta, b):
    u"""
    Sum or difference of two complex matrices, optional scaling and
    transposition.

    ``complex_addsub`` adds two complex matrices, each one optionally
    transposed and multiplied by a scalar.

    For full information please refer to the NAG Library document for
    f01cw

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01cwf.html

    Parameters
    ----------
    transa : str, length 1
        `transa` and `transb` must specify whether or not the matrix A
        and the matrix B, respectively, are to be transposed before
        addition.

        `transa` or `transb` = 'N'

            The matrix will not be transposed.

        `transa` or `transb` = 'T'

            The matrix will be transposed.

        `transa` or `transb` = 'C'

            The matrix will be transposed and conjugated.

    transb : str, length 1
        `transa` and `transb` must specify whether or not the matrix A
        and the matrix B, respectively, are to be transposed before
        addition.

        `transa` or `transb` = 'N'

            The matrix will not be transposed.

        `transa` or `transb` = 'T'

            The matrix will be transposed.

        `transa` or `transb` = 'C'

            The matrix will be transposed and conjugated.

    m : int
        m, the number of rows of the matrices A and B or their
        transposes. Also the number of rows of the matrix C.

    n : int
        n, the number of columns of the matrices A and B or their
        transposes. Also the number of columns of the matrix C.

    alpha : complex
        The scalar alpha, by which matrix A is multiplied before
        addition.

    a : complex, array-like, shape (:, :)
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `transa`='N': `m`; otherwise: `n`.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `alpha`!=0 and `transa`='N': `n`; if
        `alpha`!=0 and `transa` in ('T', 'C'): `m`; if `alpha`=0: 1;
        otherwise: 0.

        The matrix A. If alpha = 0, the array `a` is not referenced.

    beta : complex
        The scalar beta, by which matrix B is multiplied before
        addition.

    b : complex, array-like, shape (:, :)
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `transb`='N': `m`; otherwise: `n`.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `beta`!=0 and `transb`='N': `n`; if
        `beta`!=0 and `transb` in ('T', 'C'): `m`; if `beta`=0: 1;
        otherwise: 0.

        The matrix B. If beta = 0, the array `b` is not referenced.

    Returns
    -------
    c : complex, ndarray, shape (`m`, `n`)
        The elements of the m by n matrix C.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `transb` = *<value>*.

            Constraint: `transb` = 'N', 'T' or 'C'.

        (`errno` 1)
            On entry, `transa` = *<value>*.

            Constraint: `transa` = 'N', 'T' or 'C'.

        (`errno` 2)
            On entry, `m` = *<value>* and `n` = *<value>*.

            Constraint: `m` >= 0 and `n` >= 0.

        (`errno` 3)
            On entry, lda = *<value>* and `n` = *<value>*.

            Constraint: if `transa` = *<value>*, lda >= max(1, `n`).

        (`errno` 3)
            On entry, lda = *<value>* and `m` = *<value>*.

            Constraint: if `transa` = *<value>*, lda >= max(1, `m`).

        (`errno` 4)
            On entry, ldb = *<value>* and `n` = *<value>*.

            Constraint: if `transb` = *<value>*, ldb >= max(1, `n`).

        (`errno` 4)
            On entry, ldb = *<value>* and `m` = *<value>*.

            Constraint: if `transb` = *<value>*, ldb >= max(1, `m`).

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    fun_name = 'naginterfaces.library.matop.complex_addsub'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
    )
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_a, 'a'), (_b, 'b')),
    )
    _m = _b_utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = _b_utils._EngineIntScalarType.to_py(_m)
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _c = _b_utils._EngineComplex128ArrayType.empty_ndarray(
        (_m_py, _n_py),
        sorder=_m_sorder,
    )
    _b_matop.complex_addsub(
        transa,
        transb,
        _m,
        _n,
        alpha,
        _a,
        beta,
        _b,
        _c,
    )
    _c_py = _c
    return _c_py

def real_tri_matmul_inplace(side, uplo, transa, alpha, a, b):
    u"""
    Matrix-matrix product, two real lower or upper triangular matrices.

    ``real_tri_matmul_inplace`` performs one of the matrix-matrix
    operations

        [table omitted]

    where A and B are real triangular matrices, and alpha is a real
    scalar.

    For full information please refer to the NAG Library document for
    f01dg

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01dgf.html

    Parameters
    ----------
    side : str, length 1
        Specifies whether B is operated on from the left or the right.

        `side` = 'L'

            B is pre-multiplied from the left.

        `side` = 'R'

            B is post-multiplied from the right.

    uplo : str, length 1
        Specifies whether A and B are upper or lower triangular.

        `uplo` = 'U'

            A and B are upper triangular.

        `uplo` = 'L'

            A and B are lower triangular.

    transa : str, length 1
        Specifies whether the operation involves A or A^T.

        `transa` = 'N'

            The operation involves A.

        `transa` = 'T' or 'C'

            The operation involves A^T.

    alpha : float
        The scalar alpha.

    a : float, array-like, shape (n, n)
        The n by n triangular matrix A.

    b : float, ndarray, shape (n, n), modified in place
        `On entry`: the n by n triangular matrix B.

        If `alpha` = 0, `b` need not be set.

        `On exit`: B is overwritten.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `side` = *<value>*.

            Constraint: `side` = 'L' or 'R'.

        (`errno` 2)
            On entry, `uplo` = *<value>*.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 3)
            On entry, `transa` = *<value>*.

            Constraint: `transa` = 'N', 'T' or 'C'.

        (`errno` 4)
            On entry, n = *<value>*.

            Constraint: n >= 0.
    """

    fun_name = 'naginterfaces.library.matop.real_tri_matmul_inplace'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
    )
    _b_matop.real_tri_matmul_inplace(
        side,
        uplo,
        transa,
        _n,
        alpha,
        _a,
        b,
    )

def complex_tri_matmul_inplace(side, uplo, transa, alpha, a, b):
    u"""
    Matrix-matrix product, two complex lower or upper triangular
    matrices.

    ``complex_tri_matmul_inplace`` performs one of the matrix-matrix
    operations

        [table omitted]

    where A and B are complex triangular matrices, and alpha is a
    complex scalar.

    For full information please refer to the NAG Library document for
    f01du

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01duf.html

    Parameters
    ----------
    side : str, length 1
        Specifies whether B is operated on from the left or the right.

        `side` = 'L'

            B is pre-multiplied from the left.

        `side` = 'R'

            B is post-multiplied from the right.

    uplo : str, length 1
        Specifies whether A and B are upper or lower triangular.

        `uplo` = 'U'

            A and B are upper triangular.

        `uplo` = 'L'

            A and B are lower triangular.

    transa : str, length 1
        Specifies whether the operation involves A, A^T or A^H.

        `transa` = 'N'

            The operation involves A.

        `transa` = 'T'

            The operation involves A^T.

        `transa` = 'C'

            The operation involves A^H.

    alpha : complex
        The scalar alpha.

    a : complex, array-like, shape (n, n)
        The n by n triangular matrix A.

    b : complex, ndarray, shape (n, n), modified in place
        `On entry`: the n by n triangular matrix B.

        If `alpha` = 0, `b` need not be set.

        `On exit`: B is overwritten.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `side` = *<value>*.

            Constraint: `side` = 'L' or 'R'.

        (`errno` 2)
            On entry, `uplo` = *<value>*.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` 3)
            On entry, `transa` = *<value>*.

            Constraint: `transa` = 'N', 'T' or 'C'.

        (`errno` 4)
            On entry, n = *<value>*.

            Constraint: n >= 0.
    """

    fun_name = 'naginterfaces.library.matop.complex_tri_matmul_inplace'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
    )
    _b_matop.complex_tri_matmul_inplace(
        side,
        uplo,
        transa,
        _n,
        alpha,
        _a,
        b,
    )

def real_gen_matrix_exp(a):
    u"""
    Real matrix exponential.

    ``real_gen_matrix_exp`` computes the matrix exponential, e^A, of a
    real n by n matrix A.

    For full information please refer to the NAG Library document for
    f01ec

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01ecf.html

    Parameters
    ----------
    a : float, array-like, shape (n, n)
        The n by n matrix A.

    Returns
    -------
    a : float, ndarray, shape (n, n)
        The n by n matrix exponential e^A.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, n = *<value>*.

            Constraint: n >= 0.

        (`errno` 1)
            The linear equations to be solved for the Padé approximant
            are singular; it is likely that this function has been
            called incorrectly.

        (`errno` 2)
            The linear equations to be solved are nearly singular and
            the Padé approximant probably has no correct figures; it is
            likely that this function has been called incorrectly.

        (`errno` 4)
            An unexpected internal error has occurred. Please contact
            NAG.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            The arithmetic precision is higher than that used for the
            Padé approximant computed matrix exponential.

    References
    ----------
    Al--Mohy, A H and Higham, N J, 2009, `A new scaling and squaring
    algorithm for the matrix exponential`, SIAM J. Matrix Anal. (31(3)),
    970--989

    Higham, N J, 2005, `The scaling and squaring method for the matrix
    exponential revisited`, SIAM J. Matrix Anal. Appl. (26(4)),
    1179--1193

    Higham, N J, 2008, `Functions of Matrices: Theory and Computation`,
    SIAM, Philadelphia, PA, USA

    Moler, C B and Van Loan, C F, 2003, `Nineteen dubious ways to
    compute the exponential of a matrix, twenty-five years later`, SIAM
    Rev. (45), 3--49
    """

    fun_name = 'naginterfaces.library.matop.real_gen_matrix_exp'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    try:
        _b_matop.real_gen_matrix_exp(
            _n,
            _a,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _a_py = _a
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a'],
    )(
        _a_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _a_py

def real_symm_matrix_exp(uplo, a):
    u"""
    Real symmetric matrix exponential.

    ``real_symm_matrix_exp`` computes the matrix exponential, e^A, of a
    real symmetric n by n matrix A.

    For full information please refer to the NAG Library document for
    f01ed

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01edf.html

    Parameters
    ----------
    uplo : str, length 1
        Indicates whether the upper or lower triangular part of A is
        stored.

        `uplo` = 'U'

            The upper triangular part of A is stored.

        `uplo` = 'L'

            The lower triangular part of A is stored.

    a : float, array-like, shape (n, n)
        The n by n symmetric matrix A.

    Returns
    -------
    a : float, ndarray, shape (n, n)
        If no exception or warning is raised, the upper or lower
        triangular part of the n by n matrix exponential, e^A.

    Raises
    ------
    NagValueError
        (`errno` -3)
            An internal error occurred when computing the spectral
            factorization. Please contact NAG.

        (`errno` -2)
            On entry, n = *<value>*.

            Constraint: n >= 0.

        (`errno` -1)
            On entry, `uplo` = *<value>*.

            Constraint: `uplo` = 'L' or 'U'.

        (`errno` i > 0)
            The computation of the spectral factorization failed to
            converge.

    References
    ----------
    Higham, N J, 2008, `Functions of Matrices: Theory and Computation`,
    SIAM, Philadelphia, PA, USA

    Moler, C B and Van Loan, C F, 2003, `Nineteen dubious ways to
    compute the exponential of a matrix, twenty-five years later`, SIAM
    Rev. (45), 3--49
    """

    fun_name = 'naginterfaces.library.matop.real_symm_matrix_exp'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b_matop.real_symm_matrix_exp(
        uplo,
        _n,
        _a,
    )
    _a_py = _a
    return _a_py

def real_symm_matrix_fun(uplo, a, f, data=None):
    u"""
    Function of a real symmetric matrix.

    ``real_symm_matrix_fun`` computes the matrix function, f(A), of a
    real symmetric n by n matrix A. f(A) must also be a real symmetric
    matrix.

    For full information please refer to the NAG Library document for
    f01ef

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01eff.html

    Parameters
    ----------
    uplo : str, length 1
        If `uplo` = 'U', the upper triangle of the matrix A is stored.

        If `uplo` = 'L', the lower triangle of the matrix A is stored.

    a : float, array-like, shape (n, n)
        The n by n symmetric matrix A.

    f : callable fx = f(x, data=None)
        The function `f` evaluates f(z_i) at a number of points z_i.

        Parameters
        ~~~~~~~~~~
        x : float, ndarray, shape (n)
            The n points x_1,x_2,...,x_n at which the function f is to
            be evaluated.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        fx : float, array-like, shape (n)
            The n function values. `fx`[i-1] should return the value
            f(x_i), for i = 1,2,...,n.

    data : arbitrary, optional
        User-communication data for callback functions.

    Returns
    -------
    a : float, ndarray, shape (n, n)
        If no exception or warning is raised, the upper or lower
        triangular part of the n by n matrix function, f(A).

    Raises
    ------
    NagValueError
        (`errno` -3)
            An internal error occurred when computing the spectral
            factorization. Please contact NAG.

        (`errno` -2)
            On entry, n = *<value>*.

            Constraint: n >= 0.

        (`errno` -1)
            On entry, `uplo` = *<value>*.

            Constraint: `uplo` = 'L' or 'U'.

        (`errno` i > 0)
            The computation of the spectral factorization failed to
            converge.

    Warns
    -----
    NagCallbackTerminateWarning
        (`errno` -6)
            Termination requested in `f`.

    References
    ----------
    Higham, N J, 2008, `Functions of Matrices: Theory and Computation`,
    SIAM, Philadelphia, PA, USA
    """

    fun_name = 'naginterfaces.library.matop.real_symm_matrix_fun'

    cb_f_name = fun_name + ':f'

    def cb_f(n, x, fx, data=None):
        """Wrapping user's f. Called as base f."""
        if data is not None:
            _fx_f = f(
                x,
                data,
            )
        else:
            _fx_f = f(
                x,
            )
        fx[:] = _utils._to_ndarray(
            {'fun_name': cb_f_name, 'returned': 'fx'},
            _fx_f,
            copy=False,
            exp_shape=(n,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b_utils._check_callable(f, {'fun_name': fun_name, 'entity': 'f'})
    _f = cb_f
    try:
        _b_matop.real_symm_matrix_fun(
            uplo,
            _n,
            _a,
            _f,
            data,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _a_py = _a
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a'],
    )(
        _a_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _a_py

def real_gen_matrix_log(a):
    u"""
    Real matrix logarithm.

    ``real_gen_matrix_log`` computes the principal matrix logarithm,
    log(A), of a real n by n matrix A, with no eigenvalues on the closed
    negative real line.

    For full information please refer to the NAG Library document for
    f01ej

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01ejf.html

    Parameters
    ----------
    a : float, array-like, shape (n, n)
        The n by n matrix A.

    Returns
    -------
    a : float, ndarray, shape (n, n)
        The n by n principal matrix logarithm, log(A).

    imnorm : float
        If the function has given a reliable answer then `imnorm` = 0.0.
        If `imnorm` differs from 0.0 by more than unit roundoff (as
        returned by ``machine.precision``) then the computed matrix
        logarithm is unreliable.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, n = *<value>*.

            Constraint: n >= 0.

        (`errno` 1)
            A is singular so the logarithm cannot be computed.

        (`errno` 2)
            A was found to have eigenvalues on the negative real line.
            The principal logarithm is not defined in this case.
            ``complex_gen_matrix_log`` can be used to find a complex
            non-principal logarithm.

        (`errno` 4)
            An unexpected internal error occurred. Please contact NAG.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            The arithmetic precision is higher than that used for the
            Padé approximant computed matrix logarithm.

    References
    ----------
    Al--Mohy, A H and Higham, N J, 2011, `Improved inverse scaling and
    squaring algorithms for the matrix logarithm`, SIAM J. Sci. Comput.
    (34(4)), C152--C169

    Al--Mohy, A H, Higham, N J and Relton, S D, 2012, `Computing the
    Fréchet derivative of the matrix logarithm and estimating the
    condition number`, SIAM J. Sci. Comput. (35(4)), C394--C410

    Higham, N J, 2008, `Functions of Matrices: Theory and Computation`,
    SIAM, Philadelphia, PA, USA
    """

    fun_name = 'naginterfaces.library.matop.real_gen_matrix_log'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    try:
        _return_scalar = _b_matop.real_gen_matrix_log(
            _n,
            _a,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _a_py = _a
    _imnorm_py = _return_scalar
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'imnorm'],
    )(
        _a_py,
        _imnorm_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def real_gen_matrix_fun_std(fun, a):
    u"""
    Exponential, sine, cosine, sinh or cosh of a real matrix
    (Schur--Parlett algorithm).

    ``real_gen_matrix_fun_std`` computes the matrix exponential, sine,
    cosine, sinh or cosh, of a real n by n matrix A using the
    Schur--Parlett algorithm.

    For full information please refer to the NAG Library document for
    f01ek

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01ekf.html

    Parameters
    ----------
    fun : str
        Indicates which matrix function will be computed.

        `fun` = 'EXP'

            The matrix exponential, e^A, will be computed.

        `fun` = 'SIN'

            The matrix sine, sin(A), will be computed.

        `fun` = 'COS'

            The matrix cosine, cos(A), will be computed.

        `fun` = 'SINH'

            The hyperbolic matrix sine, sinh(A), will be computed.

        `fun` = 'COSH'

            The hyperbolic matrix cosine, cosh(A), will be computed.

    a : float, array-like, shape (n, n)
        The n by n matrix A.

    Returns
    -------
    a : float, ndarray, shape (n, n)
        The n by n matrix, f(A).

    imnorm : float
        If A has complex eigenvalues, ``real_gen_matrix_fun_std`` will
        use complex arithmetic to compute the matrix function. The
        imaginary part is discarded at the end of the computation,
        because it will theoretically vanish. `imnorm` contains the
        1-norm of the imaginary part, which should be used to check that
        the routine has given a reliable answer.

        If A has real eigenvalues, ``real_gen_matrix_fun_std`` uses real
        arithmetic and `imnorm` = 0.

    Raises
    ------
    NagValueError
        (`errno` -2)
            Input argument number *<value>* is invalid.

        (`errno` -1)
            Input argument `fun` is invalid.

        (`errno` 1)
            A Taylor series failed to converge.

        (`errno` 2)
            An unexpected internal error occurred when evaluating the
            function at a point. Please contact NAG.

        (`errno` 3)
            There was an error whilst reordering the Schur form of A.

            **Note:** this failure should not occur and suggests that
            the function has been called incorrectly.

        (`errno` 4)
            The function was unable to compute the Schur decomposition
            of A.

            **Note:** this failure should not occur and suggests that
            the function has been called incorrectly.

        (`errno` 5)
            An unexpected internal error occurred. Please contact NAG.

        (`errno` 6)
            The linear equations to be solved are nearly singular and
            the Padé approximant used to compute the exponential may
            have no correct figures.

            **Note:** this failure should not occur and suggests that
            the function has been called incorrectly.

    References
    ----------
    Davies, P I and Higham, N J, 2003, `A Schur--Parlett algorithm for
    computing matrix functions`, SIAM J. Matrix Anal. Appl. (25(2)),
    464--485

    Higham, N J, 2008, `Functions of Matrices: Theory and Computation`,
    SIAM, Philadelphia, PA, USA
    """

    fun_name = 'naginterfaces.library.matop.real_gen_matrix_fun_std'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _return_scalar = _b_matop.real_gen_matrix_fun_std(
        fun,
        _n,
        _a,
    )
    _a_py = _a
    _imnorm_py = _return_scalar
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'imnorm'],
    )(
        _a_py,
        _imnorm_py,
    )
    return _return_data

def real_gen_matrix_fun_num(a, f, data=None):
    u"""
    Function of a real matrix (using numerical differentiation).

    ``real_gen_matrix_fun_num`` computes the matrix function, f(A), of a
    real n by n matrix A.
    Numerical differentiation is used to evaluate the derivatives of f
    when they are required.

    For full information please refer to the NAG Library document for
    f01el

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01elf.html

    Parameters
    ----------
    a : float, array-like, shape (n, n)
        The n by n matrix A.

    f : callable fz = f(z, data=None)
        The function `f` evaluates f(z_i) at a number of points z_i.

        Parameters
        ~~~~~~~~~~
        z : complex, ndarray, shape (nz)
            The n_z points z_1,z_2,...,z_n_z at which the function f is
            to be evaluated.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        fz : complex, array-like, shape (nz)
            The n_z function values. `fz`[i-1] should return the value
            f(z_i), for i = 1,2,...,n_z. If z_i lies on the real line,
            then so must f(z_i).

    data : arbitrary, optional
        User-communication data for callback functions.

    Returns
    -------
    a : float, ndarray, shape (n, n)
        The n by n matrix, f(A).

    imnorm : float
        If A has complex eigenvalues, ``real_gen_matrix_fun_num`` will
        use complex arithmetic to compute f(A). The imaginary part is
        discarded at the end of the computation, because it will
        theoretically vanish. `imnorm` contains the 1-norm of the
        imaginary part, which should be used to check that the routine
        has given a reliable answer.

        If `a` has real eigenvalues, ``real_gen_matrix_fun_num`` uses
        real arithmetic and `imnorm` = 0.

    Raises
    ------
    NagValueError
        (`errno` -1)
            Input argument number *<value>* is invalid.

        (`errno` 1)
            A Taylor series failed to converge after 40 terms. Further
            Taylor series coefficients can no longer reliably be
            obtained by numerical differentiation.

        (`errno` 3)
            There was an error whilst reordering the Schur form of A.

            **Note:** this failure should not occur and suggests that
            the function has been called incorrectly.

        (`errno` 4)
            The function was unable to compute the Schur decomposition
            of A.

            **Note:** this failure should not occur and suggests that
            the function has been called incorrectly.

        (`errno` 5)
            An unexpected internal error occurred. Please contact NAG.

    Warns
    -----
    NagCallbackTerminateWarning
        (`errno` 2)
            Termination requested in `f`.

    References
    ----------
    Davies, P I and Higham, N J, 2003, `A Schur--Parlett algorithm for
    computing matrix functions`, SIAM J. Matrix Anal. Appl. (25(2)),
    464--485

    Higham, N J, 2008, `Functions of Matrices: Theory and Computation`,
    SIAM, Philadelphia, PA, USA

    Lyness, J N and Moler, C B, 1967, `Numerical differentiation of
    analytic functions`, SIAM J. Numer. Anal. (4(2)), 202--210
    """

    fun_name = 'naginterfaces.library.matop.real_gen_matrix_fun_num'

    cb_f_name = fun_name + ':f'

    def cb_f(nz, z, fz, data=None):
        """Wrapping user's f. Called as base f."""
        if data is not None:
            _fz_f = f(
                z,
                data,
            )
        else:
            _fz_f = f(
                z,
            )
        fz[:] = _utils._to_ndarray(
            {'fun_name': cb_f_name, 'returned': 'fz'},
            _fz_f,
            copy=False,
            exp_shape=(nz,),
            exp_class=_b_utils._EngineComplex128ArrayType,
        )

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b_utils._check_callable(f, {'fun_name': fun_name, 'entity': 'f'})
    _f = cb_f
    try:
        _return_scalar = _b_matop.real_gen_matrix_fun_num(
            _n,
            _a,
            _f,
            data,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _a_py = _a
    _imnorm_py = _return_scalar
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'imnorm'],
    )(
        _a_py,
        _imnorm_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def real_gen_matrix_fun_usd(a, f, data=None):
    u"""
    Function of a real matrix (using user-supplied derivatives).

    ``real_gen_matrix_fun_usd`` computes the matrix function, f(A), of a
    real n by n matrix A, using analytical derivatives of f you have
    supplied.

    For full information please refer to the NAG Library document for
    f01em

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01emf.html

    Parameters
    ----------
    a : float, array-like, shape (n, n)
        The n by n matrix A.

    f : callable fz = f(m, z, data=None)
        Given an integer m, the function `f` evaluates f^(m)(z_i) at a
        number of points z_i.

        Parameters
        ~~~~~~~~~~
        m : int
            The order, m, of the derivative required.

            If `m` = 0, f(z_i) should be returned.

            For `m` > 0, f^(m)(z_i) should be returned.

        z : complex, ndarray, shape (nz)
            The n_z points z_1,z_2,...,z_n_z at which the function f is
            to be evaluated.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        fz : complex, array-like, shape (nz)
            The n_z function or derivative values. `fz`[i-1] should
            return the value f^(m)(z_i), for i = 1,2,...,n_z. If z_i
            lies on the real line, then so must f^(m)(z_i).

    data : arbitrary, optional
        User-communication data for callback functions.

    Returns
    -------
    a : float, ndarray, shape (n, n)
        The n by n matrix, f(A).

    imnorm : float
        If A has complex eigenvalues, ``real_gen_matrix_fun_usd`` will
        use complex arithmetic to compute f(A). The imaginary part is
        discarded at the end of the computation, because it will
        theoretically vanish. `imnorm` contains the 1-norm of the
        imaginary part, which should be used to check that the function
        has given a reliable answer.

        If A has real eigenvalues, ``real_gen_matrix_fun_usd`` uses real
        arithmetic and `imnorm` = 0.

    Raises
    ------
    NagValueError
        (`errno` -1)
            Input argument number *<value>* is invalid.

        (`errno` 1)
            A Taylor series failed to converge.

        (`errno` 3)
            There was an error whilst reordering the Schur form of A.

            **Note:** this failure should not occur and suggests that
            the function has been called incorrectly.

        (`errno` 4)
            The routine was unable to compute the Schur decomposition of
            A.

            **Note:** this failure should not occur and suggests that
            the function has been called incorrectly.

        (`errno` 5)
            An unexpected internal error occurred. Please contact NAG.

    Warns
    -----
    NagCallbackTerminateWarning
        (`errno` 2)
            Termination requested in `f`.

    References
    ----------
    Davies, P I and Higham, N J, 2003, `A Schur--Parlett algorithm for
    computing matrix functions`, SIAM J. Matrix Anal. Appl. (25(2)),
    464--485

    Higham, N J, 2008, `Functions of Matrices: Theory and Computation`,
    SIAM, Philadelphia, PA, USA

    See Also
    --------
    :meth:`naginterfaces.library.examples.matop.real_gen_matrix_fun_usd_ex.main`
    """

    fun_name = 'naginterfaces.library.matop.real_gen_matrix_fun_usd'

    cb_f_name = fun_name + ':f'

    def cb_f(m, nz, z, fz, data=None):
        """Wrapping user's f. Called as base f."""
        if data is not None:
            _fz_f = f(
                m,
                z,
                data,
            )
        else:
            _fz_f = f(
                m,
                z,
            )
        fz[:] = _utils._to_ndarray(
            {'fun_name': cb_f_name, 'returned': 'fz'},
            _fz_f,
            copy=False,
            exp_shape=(nz,),
            exp_class=_b_utils._EngineComplex128ArrayType,
        )

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b_utils._check_callable(f, {'fun_name': fun_name, 'entity': 'f'})
    _f = cb_f
    try:
        _return_scalar = _b_matop.real_gen_matrix_fun_usd(
            _n,
            _a,
            _f,
            data,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _a_py = _a
    _imnorm_py = _return_scalar
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'imnorm'],
    )(
        _a_py,
        _imnorm_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def real_gen_matrix_sqrt(a):
    u"""
    Real matrix square root.

    ``real_gen_matrix_sqrt`` computes the principal matrix square root,
    A^(1/2), of a real n by n matrix A.

    For full information please refer to the NAG Library document for
    f01en

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01enf.html

    Parameters
    ----------
    a : float, array-like, shape (n, n)
        The n by n matrix A.

    Returns
    -------
    a : float, ndarray, shape (n, n)
        Contains, if no exception or warning is raised, the n by n
        principal matrix square root, A^(1/2). Alternatively, if `errno`
        = 1, contains an n by n non-principal square root of A.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, n = *<value>*.

            Constraint: n >= 0.

        (`errno` 2)
            A has a defective vanishing eigenvalue. The square root
            cannot be found in this case.

        (`errno` 3)
            A has a negative real eigenvalue. The principal square root
            is not defined. ``complex_gen_matrix_sqrt`` can be used to
            return a complex, non-principal square root.

        (`errno` 4)
            An internal error occurred. It is likely that the function
            was called incorrectly.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            A has a semisimple vanishing eigenvalue. A non-principal
            square root is returned.

    References
    ----------
    Björck, Å and Hammarling, S, 1983, `A Schur method for the square
    root of a matrix`, Linear Algebra Appl. (52/53), 127--140

    Deadman, E, Higham, N J and Ralha, R, 2013, `Blocked Schur
    Algorithms for Computing the Matrix Square Root`, `Applied Parallel
    and Scientific Computing: 11th International Conference, (PARA 2012,
    Helsinki, Finland)`, P. Manninen and P. Öster, Eds, Lecture Notes in
    Computer Science (7782), 171--181, Springer--Verlag

    Higham, N J, 1987, `Computing real square roots of a real matrix`,
    Linear Algebra Appl. (88/89), 405--430

    Higham, N J, 2008, `Functions of Matrices: Theory and Computation`,
    SIAM, Philadelphia, PA, USA
    """

    fun_name = 'naginterfaces.library.matop.real_gen_matrix_sqrt'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    try:
        _b_matop.real_gen_matrix_sqrt(
            _n,
            _a,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _a_py = _a
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a'],
    )(
        _a_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _a_py

def real_tri_matrix_sqrt(a):
    u"""
    Real upper quasi-triangular matrix square root.

    ``real_tri_matrix_sqrt`` computes the principal matrix square root,
    A^(1/2), of a real upper quasi-triangular n by n matrix A.

    For full information please refer to the NAG Library document for
    f01ep

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01epf.html

    Parameters
    ----------
    a : float, array-like, shape (n, n)
        The n by n upper quasi-triangular matrix A.

    Returns
    -------
    a : float, ndarray, shape (n, n)
        The n by n principal matrix square root A^(1/2).

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, n = *<value>*.

            Constraint: n >= 0.

        (`errno` 1)
            A has negative or vanishing eigenvalues. The principal
            square root is not defined in this case.
            ``real_gen_matrix_sqrt`` or ``complex_gen_matrix_sqrt`` may
            be able to provide further information.

        (`errno` 2)
            An internal error occurred. It is likely that the function
            was called incorrectly.

    References
    ----------
    Björck, Å and Hammarling, S, 1983, `A Schur method for the square
    root of a matrix`, Linear Algebra Appl. (52/53), 127--140

    Deadman, E, Higham, N J and Ralha, R, 2013, `Blocked Schur
    Algorithms for Computing the Matrix Square Root`, `Applied Parallel
    and Scientific Computing: 11th International Conference, (PARA 2012,
    Helsinki, Finland)`, P. Manninen and P. Öster, Eds, Lecture Notes in
    Computer Science (7782), 171--181, Springer--Verlag

    Higham, N J, 1987, `Computing real square roots of a real matrix`,
    Linear Algebra Appl. (88/89), 405--430

    Higham, N J, 2008, `Functions of Matrices: Theory and Computation`,
    SIAM, Philadelphia, PA, USA
    """

    fun_name = 'naginterfaces.library.matop.real_tri_matrix_sqrt'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b_matop.real_tri_matrix_sqrt(
        _n,
        _a,
    )
    _a_py = _a
    return _a_py

def real_gen_matrix_pow(a, p):
    u"""
    General power of a real matrix.

    ``real_gen_matrix_pow`` computes the principal real power A^p, for
    arbitrary p, of a real n by n matrix A.

    For full information please refer to the NAG Library document for
    f01eq

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01eqf.html

    Parameters
    ----------
    a : float, array-like, shape (n, n)
        The n by n matrix A.

    p : float
        The required power of A.

    Returns
    -------
    a : float, ndarray, shape (n, n)
        The n by n matrix pth power, A^p.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, n = *<value>*.

            Constraint: n >= 0.

        (`errno` 1)
            A has eigenvalues on the negative real line. The principal
            pth power is not defined. ``complex_gen_matrix_pow`` can be
            used to find a complex, non-principal pth power.

        (`errno` 2)
            A is singular so the pth power cannot be computed.

        (`errno` 4)
            An unexpected internal error occurred. This failure should
            not occur and suggests that the function has been called
            incorrectly.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            A^p has been computed using an IEEE double precision Padé
            approximant, although the arithmetic precision is higher
            than IEEE double precision.

    References
    ----------
    Higham, N J, 2008, `Functions of Matrices: Theory and Computation`,
    SIAM, Philadelphia, PA, USA

    Higham, N J and Lin, L, 2011, `A Schur--Padé algorithm for
    fractional powers of a matrix`, SIAM J. Matrix Anal. Appl. (32(3)),
    1056--1078

    Higham, N J and Lin, L, 2013, `An improved Schur--Padé algorithm for
    fractional powers of a matrix and their Fréchet derivatives`, SIAM
    J. Matrix Anal. Appl. (34(3)), 1341--1360
    """

    fun_name = 'naginterfaces.library.matop.real_gen_matrix_pow'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    try:
        _b_matop.real_gen_matrix_pow(
            _n,
            _a,
            p,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _a_py = _a
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a'],
    )(
        _a_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _a_py

def complex_gen_matrix_exp(a):
    u"""
    Complex matrix exponential.

    ``complex_gen_matrix_exp`` computes the matrix exponential, e^A, of
    a complex n by n matrix A.

    For full information please refer to the NAG Library document for
    f01fc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01fcf.html

    Parameters
    ----------
    a : complex, array-like, shape (n, n)
        The n by n matrix A.

    Returns
    -------
    a : complex, ndarray, shape (n, n)
        With no exception or warning is raised, the n by n matrix
        exponential e^A.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, n = *<value>*.

            Constraint: n >= 0.

        (`errno` 1)
            The linear equations to be solved for the Padé approximant
            are singular; it is likely that this function has been
            called incorrectly.

        (`errno` 2)
            The linear equations to be solved are nearly singular and
            the Padé approximant probably has no correct figures; it is
            likely that this function has been called incorrectly.

        (`errno` 4)
            An unexpected internal error has occurred. Please contact
            NAG.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            The arithmetic precision is higher than that used for the
            Padé approximant computed matrix exponential.

    References
    ----------
    Al--Mohy, A H and Higham, N J, 2009, `A new scaling and squaring
    algorithm for the matrix exponential`, SIAM J. Matrix Anal. (31(3)),
    970--989

    Higham, N J, 2005, `The scaling and squaring method for the matrix
    exponential revisited`, SIAM J. Matrix Anal. Appl. (26(4)),
    1179--1193

    Higham, N J, 2008, `Functions of Matrices: Theory and Computation`,
    SIAM, Philadelphia, PA, USA

    Moler, C B and Van Loan, C F, 2003, `Nineteen dubious ways to
    compute the exponential of a matrix, twenty-five years later`, SIAM
    Rev. (45), 3--49
    """

    fun_name = 'naginterfaces.library.matop.complex_gen_matrix_exp'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    try:
        _b_matop.complex_gen_matrix_exp(
            _n,
            _a,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _a_py = _a
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a'],
    )(
        _a_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _a_py

def complex_herm_matrix_exp(uplo, a):
    u"""
    Complex Hermitian matrix exponential.

    ``complex_herm_matrix_exp`` computes the matrix exponential, e^A, of
    a complex Hermitian n by n matrix A.

    For full information please refer to the NAG Library document for
    f01fd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01fdf.html

    Parameters
    ----------
    uplo : str, length 1
        If `uplo` = 'U', the upper triangle of the matrix A is stored.

        If `uplo` = 'L', the lower triangle of the matrix A is stored.

    a : complex, array-like, shape (n, n)
        The n by n Hermitian matrix A.

    Returns
    -------
    a : complex, ndarray, shape (n, n)
        If no exception or warning is raised, the upper or lower
        triangular part of the n by n matrix exponential, e^A.

    Raises
    ------
    NagValueError
        (`errno` -3)
            An internal error occurred when computing the spectral
            factorization. Please contact NAG.

        (`errno` -2)
            On entry, n = *<value>*.

            Constraint: n >= 0.

        (`errno` -1)
            On entry, `uplo` = *<value>*.

            Constraint: `uplo` = 'L' or 'U'.

        (`errno` i > 0)
            The computation of the spectral factorization failed to
            converge.

    References
    ----------
    Higham, N J, 2005, `The scaling and squaring method for the matrix
    exponential revisited`, SIAM J. Matrix Anal. Appl. (26(4)),
    1179--1193

    Higham, N J, 2008, `Functions of Matrices: Theory and Computation`,
    SIAM, Philadelphia, PA, USA

    Moler, C B and Van Loan, C F, 2003, `Nineteen dubious ways to
    compute the exponential of a matrix, twenty-five years later`, SIAM
    Rev. (45), 3--49
    """

    fun_name = 'naginterfaces.library.matop.complex_herm_matrix_exp'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b_matop.complex_herm_matrix_exp(
        uplo,
        _n,
        _a,
    )
    _a_py = _a
    return _a_py

def complex_herm_matrix_fun(uplo, a, f, data=None):
    u"""
    Function of a complex Hermitian matrix.

    ``complex_herm_matrix_fun`` computes the matrix function, f(A), of a
    complex Hermitian n by n matrix A. f(A) must also be a complex
    Hermitian matrix.

    For full information please refer to the NAG Library document for
    f01ff

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01fff.html

    Parameters
    ----------
    uplo : str, length 1
        If `uplo` = 'U', the upper triangle of the matrix A is stored.

        If `uplo` = 'L', the lower triangle of the matrix A is stored.

    a : complex, array-like, shape (n, n)
        The n by n Hermitian matrix A.

    f : callable fx = f(x, data=None)
        The function `f` evaluates f(z_i) at a number of points z_i.

        Parameters
        ~~~~~~~~~~
        x : float, ndarray, shape (n)
            The n points x_1,x_2,...,x_n at which the function f is to
            be evaluated.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        fx : float, array-like, shape (n)
            The n function values. `fx`[i-1] should return the value
            f(x_i), for i = 1,2,...,n.

    data : arbitrary, optional
        User-communication data for callback functions.

    Returns
    -------
    a : complex, ndarray, shape (n, n)
        If no exception or warning is raised, the upper or lower
        triangular part of the n by n matrix function, f(A).

    Raises
    ------
    NagValueError
        (`errno` -3)
            An internal error occurred when computing the spectral
            factorization. Please contact NAG.

        (`errno` -2)
            On entry, n = *<value>*.

            Constraint: n >= 0.

        (`errno` -1)
            On entry, `uplo` = *<value>*.

            Constraint: `uplo` = 'L' or 'U'.

        (`errno` i > 0)
            The computation of the spectral factorization failed to
            converge.

    Warns
    -----
    NagCallbackTerminateWarning
        (`errno` -6)
            Termination requested in `f`.

    References
    ----------
    Higham, N J, 2008, `Functions of Matrices: Theory and Computation`,
    SIAM, Philadelphia, PA, USA
    """

    fun_name = 'naginterfaces.library.matop.complex_herm_matrix_fun'

    cb_f_name = fun_name + ':f'

    def cb_f(n, x, fx, data=None):
        """Wrapping user's f. Called as base f."""
        if data is not None:
            _fx_f = f(
                x,
                data,
            )
        else:
            _fx_f = f(
                x,
            )
        fx[:] = _utils._to_ndarray(
            {'fun_name': cb_f_name, 'returned': 'fx'},
            _fx_f,
            copy=False,
            exp_shape=(n,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b_utils._check_callable(f, {'fun_name': fun_name, 'entity': 'f'})
    _f = cb_f
    try:
        _b_matop.complex_herm_matrix_fun(
            uplo,
            _n,
            _a,
            _f,
            data,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _a_py = _a
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a'],
    )(
        _a_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _a_py

def complex_gen_matrix_log(a):
    u"""
    Complex matrix logarithm.

    ``complex_gen_matrix_log`` computes the principal matrix logarithm,
    log(A), of a complex n by n matrix A, with no eigenvalues on the
    closed negative real line.

    For full information please refer to the NAG Library document for
    f01fj

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01fjf.html

    Parameters
    ----------
    a : complex, array-like, shape (n, n)
        The n by n matrix A.

    Returns
    -------
    a : complex, ndarray, shape (n, n)
        The n by n principal matrix logarithm, log(A), unless `errno` =
        2, in which case a non-principal logarithm is returned.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, n = *<value>*.

            Constraint: n >= 0.

        (`errno` 1)
            A is singular so the logarithm cannot be computed.

        (`errno` 4)
            An unexpected internal error has occurred. Please contact
            NAG.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            A was found to have eigenvalues on the negative real line.
            The principal logarithm is not defined in this case, so a
            non-principal logarithm was returned.

        (`errno` 3)
            The arithmetic precision is higher than that used for the
            Padé approximant computed matrix logarithm.

    References
    ----------
    Al--Mohy, A H and Higham, N J, 2011, `Improved inverse scaling and
    squaring algorithms for the matrix logarithm`, SIAM J. Sci. Comput.
    (34(4)), C152--C169

    Higham, N J, 2008, `Functions of Matrices: Theory and Computation`,
    SIAM, Philadelphia, PA, USA
    """

    fun_name = 'naginterfaces.library.matop.complex_gen_matrix_log'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    try:
        _b_matop.complex_gen_matrix_log(
            _n,
            _a,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _a_py = _a
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a'],
    )(
        _a_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _a_py

def complex_gen_matrix_fun_std(fun, a):
    u"""
    Exponential, sine, cosine, sinh or cosh of a complex matrix
    (Schur--Parlett algorithm).

    ``complex_gen_matrix_fun_std`` computes the matrix exponential,
    sine, cosine, sinh or cosh, of a complex n by n matrix A using the
    Schur--Parlett algorithm.

    For full information please refer to the NAG Library document for
    f01fk

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01fkf.html

    Parameters
    ----------
    fun : str
        Indicates which matrix function will be computed.

        `fun` = 'EXP'

            The matrix exponential, e^A, will be computed.

        `fun` = 'SIN'

            The matrix sine, sin(A), will be computed.

        `fun` = 'COS'

            The matrix cosine, cos(A), will be computed.

        `fun` = 'SINH'

            The hyperbolic matrix sine, sinh(A), will be computed.

        `fun` = 'COSH'

            The hyperbolic matrix cosine, cosh(A), will be computed.

    a : complex, array-like, shape (n, n)
        The n by n matrix A.

    Returns
    -------
    a : complex, ndarray, shape (n, n)
        The n by n matrix, f(A).

    Raises
    ------
    NagValueError
        (`errno` -2)
            Input argument number *<value>* is invalid.

        (`errno` -1)
            Input argument `fun` is invalid.

        (`errno` 1)
            A Taylor series failed to converge.

        (`errno` 2)
            An unexpected internal error occurred when evaluating the
            function at a point. Please contact NAG.

        (`errno` 3)
            There was an error whilst reordering the Schur form of A.

            **Note:** this failure should not occur and suggests that
            the function has been called incorrectly.

        (`errno` 4)
            The function was unable to compute the Schur decomposition
            of A.

            **Note:** this failure should not occur and suggests that
            the function has been called incorrectly.

        (`errno` 5)
            An unexpected internal error occurred. Please contact NAG.

        (`errno` 6)
            The linear equations to be solved are nearly singular and
            the Padé approximant used to compute the exponential may
            have no correct figures.

            **Note:** this failure should not occur and suggests that
            the function has been called incorrectly.

    References
    ----------
    Davies, P I and Higham, N J, 2003, `A Schur--Parlett algorithm for
    computing matrix functions`, SIAM J. Matrix Anal. Appl. (25(2)),
    464--485

    Higham, N J, 2008, `Functions of Matrices: Theory and Computation`,
    SIAM, Philadelphia, PA, USA
    """

    fun_name = 'naginterfaces.library.matop.complex_gen_matrix_fun_std'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b_matop.complex_gen_matrix_fun_std(
        fun,
        _n,
        _a,
    )
    _a_py = _a
    return _a_py

def complex_gen_matrix_fun_num(a, f, data=None):
    u"""
    Function of a complex matrix (using numerical differentiation).

    ``complex_gen_matrix_fun_num`` computes the matrix function, f(A),
    of a complex n by n matrix A.
    Numerical differentiation is used to evaluate the derivatives of f
    when they are required.

    For full information please refer to the NAG Library document for
    f01fl

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01flf.html

    Parameters
    ----------
    a : complex, array-like, shape (n, n)
        The n by n matrix A.

    f : callable fz = f(z, data=None)
        The function `f` evaluates f(z_i) at a number of points z_i.

        Parameters
        ~~~~~~~~~~
        z : complex, ndarray, shape (nz)
            The n_z points z_1,z_2,...,z_n_z at which the function f is
            to be evaluated.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        fz : complex, array-like, shape (nz)
            The n_z function values. `fz`[i-1] should return the value
            f(z_i), for i = 1,2,...,n_z.

    data : arbitrary, optional
        User-communication data for callback functions.

    Returns
    -------
    a : complex, ndarray, shape (n, n)
        The n by n matrix, f(A).

    Raises
    ------
    NagValueError
        (`errno` -1)
            Input argument number *<value>* is invalid.

        (`errno` 1)
            A Taylor series failed to converge after 40 terms. Further
            Taylor series coefficients can no longer reliably be
            obtained by numerical differentiation.

        (`errno` 3)
            There was an error whilst reordering the Schur form of A.

            **Note:** this failure should not occur and suggests that
            the function has been called incorrectly.

        (`errno` 4)
            The function was unable to compute the Schur decomposition
            of A.

            **Note:** this failure should not occur and suggests that
            the function has been called incorrectly.

        (`errno` 5)
            An unexpected internal error occurred. Please contact NAG.

    Warns
    -----
    NagCallbackTerminateWarning
        (`errno` 2)
            Termination requested in `f`.

    References
    ----------
    Davies, P I and Higham, N J, 2003, `A Schur--Parlett algorithm for
    computing matrix functions`, SIAM J. Matrix Anal. Appl. (25(2)),
    464--485

    Higham, N J, 2008, `Functions of Matrices: Theory and Computation`,
    SIAM, Philadelphia, PA, USA

    Lyness, J N and Moler, C B, 1967, `Numerical differentiation of
    analytic functions`, SIAM J. Numer. Anal. (4(2)), 202--210
    """

    fun_name = 'naginterfaces.library.matop.complex_gen_matrix_fun_num'

    cb_f_name = fun_name + ':f'

    def cb_f(nz, z, fz, data=None):
        """Wrapping user's f. Called as base f."""
        if data is not None:
            _fz_f = f(
                z,
                data,
            )
        else:
            _fz_f = f(
                z,
            )
        fz[:] = _utils._to_ndarray(
            {'fun_name': cb_f_name, 'returned': 'fz'},
            _fz_f,
            copy=False,
            exp_shape=(nz,),
            exp_class=_b_utils._EngineComplex128ArrayType,
        )

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b_utils._check_callable(f, {'fun_name': fun_name, 'entity': 'f'})
    _f = cb_f
    try:
        _b_matop.complex_gen_matrix_fun_num(
            _n,
            _a,
            _f,
            data,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _a_py = _a
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a'],
    )(
        _a_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _a_py

def complex_gen_matrix_fun_usd(a, f, data=None):
    u"""
    Function of a complex matrix (using user-supplied derivatives).

    ``complex_gen_matrix_fun_usd`` computes the matrix function, f(A),
    of a complex n by n matrix A, using analytical derivatives of f you
    have supplied.

    For full information please refer to the NAG Library document for
    f01fm

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01fmf.html

    Parameters
    ----------
    a : complex, array-like, shape (n, n)
        The n by n matrix A.

    f : callable fz = f(m, z, data=None)
        Given an integer m, the function `f` evaluates f^(m)(z_i) at a
        number of points z_i.

        Parameters
        ~~~~~~~~~~
        m : int
            The order, m, of the derivative required.

            If `m` = 0, f(z_i) should be returned.

            For `m` > 0, f^(m)(z_i) should be returned.

        z : complex, ndarray, shape (nz)
            The n_z points z_1,z_2,...,z_n_z at which the function f is
            to be evaluated.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        fz : complex, array-like, shape (nz)
            The n_z function or derivative values. `fz`[i-1] should
            return the value f^(m)(z_i), for i = 1,2,...,n_z.

    data : arbitrary, optional
        User-communication data for callback functions.

    Returns
    -------
    a : complex, ndarray, shape (n, n)
        The n by n matrix, f(A).

    Raises
    ------
    NagValueError
        (`errno` -1)
            Input argument number *<value>* is invalid.

        (`errno` 1)
            A Taylor series failed to converge.

        (`errno` 3)
            There was an error whilst reordering the Schur form of A.

            **Note:** this failure should not occur and suggests that
            the function has been called incorrectly.

        (`errno` 4)
            The function was unable to compute the Schur decomposition
            of A.

            **Note:** this failure should not occur and suggests that
            the function has been called incorrectly.

        (`errno` 5)
            An unexpected internal error occurred. Please contact NAG.

    Warns
    -----
    NagCallbackTerminateWarning
        (`errno` 2)
            Termination requested in `f`.

    References
    ----------
    Davies, P I and Higham, N J, 2003, `A Schur--Parlett algorithm for
    computing matrix functions`, SIAM J. Matrix Anal. Appl. (25(2)),
    464--485

    Higham, N J, 2008, `Functions of Matrices: Theory and Computation`,
    SIAM, Philadelphia, PA, USA
    """

    fun_name = 'naginterfaces.library.matop.complex_gen_matrix_fun_usd'

    cb_f_name = fun_name + ':f'

    def cb_f(m, nz, z, fz, data=None):
        """Wrapping user's f. Called as base f."""
        if data is not None:
            _fz_f = f(
                m,
                z,
                data,
            )
        else:
            _fz_f = f(
                m,
                z,
            )
        fz[:] = _utils._to_ndarray(
            {'fun_name': cb_f_name, 'returned': 'fz'},
            _fz_f,
            copy=False,
            exp_shape=(nz,),
            exp_class=_b_utils._EngineComplex128ArrayType,
        )

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b_utils._check_callable(f, {'fun_name': fun_name, 'entity': 'f'})
    _f = cb_f
    try:
        _b_matop.complex_gen_matrix_fun_usd(
            _n,
            _a,
            _f,
            data,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _a_py = _a
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a'],
    )(
        _a_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _a_py

def complex_gen_matrix_sqrt(a):
    u"""
    Complex matrix square root.

    ``complex_gen_matrix_sqrt`` computes the principal matrix square
    root, A^(1/2), of a complex n by n matrix A.

    For full information please refer to the NAG Library document for
    f01fn

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01fnf.html

    Parameters
    ----------
    a : complex, array-like, shape (n, n)
        The n by n matrix A.

    Returns
    -------
    a : complex, ndarray, shape (n, n)
        Contains, if no exception or warning is raised, the n by n
        principal matrix square root, A^(1/2). Alternatively, if `errno`
        = 1, contains an n by n non-principal square root of A.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, n = *<value>*.

            Constraint: n >= 0.

        (`errno` 2)
            A has a defective vanishing eigenvalue. The square root
            cannot be found in this case.

        (`errno` 3)
            An internal error occurred. It is likely that the function
            was called incorrectly.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            A has a negative or semisimple vanishing eigenvalue. A
            non-principal square root is returned.

    References
    ----------
    Björck, Å and Hammarling, S, 1983, `A Schur method for the square
    root of a matrix`, Linear Algebra Appl. (52/53), 127--140

    Deadman, E, Higham, N J and Ralha, R, 2013, `Blocked Schur
    Algorithms for Computing the Matrix Square Root`, `Applied Parallel
    and Scientific Computing: 11th International Conference, (PARA 2012,
    Helsinki, Finland)`, P. Manninen and P. Öster, Eds, Lecture Notes in
    Computer Science (7782), 171--181, Springer--Verlag

    Higham, N J, 2008, `Functions of Matrices: Theory and Computation`,
    SIAM, Philadelphia, PA, USA
    """

    fun_name = 'naginterfaces.library.matop.complex_gen_matrix_sqrt'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    try:
        _b_matop.complex_gen_matrix_sqrt(
            _n,
            _a,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _a_py = _a
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a'],
    )(
        _a_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _a_py

def complex_tri_matrix_sqrt(a):
    u"""
    Complex upper triangular matrix square root.

    ``complex_tri_matrix_sqrt`` computes the principal matrix square
    root, A^(1/2), of a complex upper triangular n by n matrix A.

    For full information please refer to the NAG Library document for
    f01fp

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01fpf.html

    Parameters
    ----------
    a : complex, array-like, shape (n, n)
        The n by n upper triangular matrix A.

    Returns
    -------
    a : complex, ndarray, shape (n, n)
        Contains, if no exception or warning is raised, the n by n
        principal matrix square root, A^(1/2). Alternatively, if `errno`
        = 1, contains an n by n non-principal square root of A.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, n = *<value>*.

            Constraint: n >= 0.

        (`errno` 2)
            A has a defective vanishing eigenvalue. The square root
            cannot be found in this case.

        (`errno` 3)
            An internal error occurred. It is likely that the function
            was called incorrectly.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            A has negative or semisimple, vanishing eigenvalues. The
            principal square root is not defined in this case; a
            non-principal square root is returned.

    References
    ----------
    Björck, Å and Hammarling, S, 1983, `A Schur method for the square
    root of a matrix`, Linear Algebra Appl. (52/53), 127--140

    Deadman, E, Higham, N J and Ralha, R, 2013, `Blocked Schur
    Algorithms for Computing the Matrix Square Root`, `Applied Parallel
    and Scientific Computing: 11th International Conference, (PARA 2012,
    Helsinki, Finland)`, P. Manninen and P. Öster, Eds, Lecture Notes in
    Computer Science (7782), 171--181, Springer--Verlag

    Higham, N J, 2008, `Functions of Matrices: Theory and Computation`,
    SIAM, Philadelphia, PA, USA
    """

    fun_name = 'naginterfaces.library.matop.complex_tri_matrix_sqrt'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    try:
        _b_matop.complex_tri_matrix_sqrt(
            _n,
            _a,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _a_py = _a
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a'],
    )(
        _a_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _a_py

def complex_gen_matrix_pow(a, p):
    u"""
    General power of a complex matrix.

    ``complex_gen_matrix_pow`` computes an abitrary real power A^p of a
    complex n by n matrix A.

    For full information please refer to the NAG Library document for
    f01fq

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01fqf.html

    Parameters
    ----------
    a : complex, array-like, shape (n, n)
        The n by n matrix A.

    p : float
        The required power of A.

    Returns
    -------
    a : complex, ndarray, shape (n, n)
        If no exception or warning is raised, the n by n matrix pth
        power, A^p. Alternatively, if `errno` = 1, contains an n by n
        non-principal power of A.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, n = *<value>*.

            Constraint: n >= 0.

        (`errno` 2)
            A is singular so the pth power cannot be computed.

        (`errno` 4)
            An unexpected internal error occurred. This failure should
            not occur and suggests that the function has been called
            incorrectly.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            A has eigenvalues on the negative real line. The principal
            pth power is not defined so a non-principal power is
            returned.

        (`errno` 3)
            A^p has been computed using an IEEE double precision Padé
            approximant, although the arithmetic precision is higher
            than IEEE double precision.

    References
    ----------
    Higham, N J, 2008, `Functions of Matrices: Theory and Computation`,
    SIAM, Philadelphia, PA, USA

    Higham, N J and Lin, L, 2011, `A Schur--Padé algorithm for
    fractional powers of a matrix`, SIAM J. Matrix Anal. Appl. (32(3)),
    1056--1078

    Higham, N J and Lin, L, 2013, `An improved Schur--Padé algorithm for
    fractional powers of a matrix and their Fréchet derivatives`, SIAM
    J. Matrix Anal. Appl. (34(3)), 1341--1360
    """

    fun_name = 'naginterfaces.library.matop.complex_gen_matrix_pow'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    try:
        _b_matop.complex_gen_matrix_pow(
            _n,
            _a,
            p,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _a_py = _a
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a'],
    )(
        _a_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _a_py

def real_gen_matrix_actexp(a, b, t):
    u"""
    Action of a real matrix exponential on a real matrix.

    ``real_gen_matrix_actexp`` computes the action of the matrix
    exponential e^(tA), on the matrix B, where A is a real n by n
    matrix, B is a real n by m matrix and t is a real scalar.

    For full information please refer to the NAG Library document for
    f01ga

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01gaf.html

    Parameters
    ----------
    a : float, array-like, shape (n, n)
        The n by n matrix A.

    b : float, array-like, shape (n, m)
        The n by m matrix B.

    t : float
        The scalar t.

    Returns
    -------
    a : float, ndarray, shape (n, n)
        A is overwritten during the computation.

    b : float, ndarray, shape (n, m)
        The n by m matrix e^(tA)B.

    Raises
    ------
    NagValueError
        (`errno` -2)
            On entry, m = *<value>*.

            Constraint: m >= 0.

        (`errno` -1)
            On entry, n = *<value>*.

            Constraint: n >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            e^(tA)B has been computed using an IEEE double precision
            Taylor series, although the arithmetic precision is higher
            than IEEE double precision.

    References
    ----------
    Al--Mohy, A H and Higham, N J, 2011, `Computing the action of the
    matrix exponential, with an application to exponential integrators`,
    SIAM J. Sci. Statist. Comput. (33(2)), 488-511

    Higham, N J, 2008, `Functions of Matrices: Theory and Computation`,
    SIAM, Philadelphia, PA, USA
    """

    fun_name = 'naginterfaces.library.matop.real_gen_matrix_actexp'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    try:
        _b_matop.real_gen_matrix_actexp(
            _n,
            _m,
            _a,
            _b,
            t,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _a_py = _a
    _b_py = _b
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'b'],
    )(
        _a_py,
        _b_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def real_gen_matrix_actexp_rcomm(irevcm, b, t, b2, x, y, p, r, z, comm, tr=0.0):
    u"""
    Action of a real matrix exponential on a real matrix (reverse
    communication).

    ``real_gen_matrix_actexp_rcomm`` computes the action of the matrix
    exponential e^(tA), on the matrix B, where A is a real n by n
    matrix, B is a real n by m matrix and t is a real scalar.
    It uses reverse communication for evaluating matrix products, so
    that the matrix A is not accessed explicitly.

    For full information please refer to the NAG Library document for
    f01gb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01gbf.html

    Parameters
    ----------
    irevcm : int
        `On initial entry`: must be set to 0.

    b : float, ndarray, shape (n, m), modified in place
        `On initial entry`: the n by m matrix B.

        `On intermediate exit`: if `irevcm` = 1, contains the n by m
        matrix B.

        `On intermediate entry`: must not be changed.

        `On final exit`: the n by m matrix e^(tA)B.

    t : float
        The scalar t.

    b2 : float, ndarray, shape (n, m), modified in place
        `On initial entry`: need not be set.

        `On intermediate entry`: if `irevcm` = 1, must contain AB.

        `On final exit`: the array is undefined.

    x : float, ndarray, shape (n, 2), modified in place
        `On initial entry`: need not be set.

        `On intermediate exit`: if `irevcm` = 2, contains the current n
        by 2 matrix X.

        `On intermediate entry`: if `irevcm` = 3, must contain A^TY.

        `On final exit`: the array is undefined.

    y : float, ndarray, shape (n, 2), modified in place
        `On initial entry`: need not be set.

        `On intermediate exit`: if `irevcm` = 3, contains the current n
        by 2 matrix Y.

        `On intermediate entry`: if `irevcm` = 2, must contain AX.

        `On final exit`: the array is undefined.

    p : float, ndarray, shape (n), modified in place
        `On initial entry`: need not be set.

        `On intermediate entry`: if `irevcm` = 4, must contain Az.

        `On final exit`: the array is undefined.

    r : float, ndarray, shape (n), modified in place
        `On initial entry`: need not be set.

        `On intermediate entry`: if `irevcm` = 5, must contain A^Tz.

        `On final exit`: the array is undefined.

    z : float, ndarray, shape (n), modified in place
        `On initial entry`: need not be set.

        `On intermediate exit`: if `irevcm` = 4 or 5, contains the
        vector z.

        `On intermediate entry`: must not be changed.

        `On final exit`: the array is undefined.

    comm : dict, communication object, modified in place
        Communication structure.

        `On initial entry`: need not be set.

    tr : float, optional
        The trace of A. If this is not available then any number can be
        supplied (0.0 is a reasonable default); however, in the trivial
        case, n = 1, the result e^(`tr`t)B is immediately returned in
        the first row of B. See [Further Comments].

    Returns
    -------
    irevcm : int
        `On intermediate exit`: `irevcm` = 1, 2, 3, 4 or 5. The calling
        program must:

        (a) if `irevcm` = 1: evaluate B_2 = AB, where B_2 is an n by m
            matrix, and store the result in `b2`;

            if `irevcm` = 2: evaluate Y = AX, where X and Y are n by 2
            matrices, and store the result in `y`;

            if `irevcm` = 3: evaluate X = A^TY and store the result in
            `x`;

            if `irevcm` = 4: evaluate p = Az and store the result in
            `p`;

            if `irevcm` = 5: evaluate r = A^Tz and store the result in
            `r`.

        (#) call ``real_gen_matrix_actexp_rcomm`` again with all other
            parameters unchanged.

        `On final exit`: `irevcm` = 0.

    Raises
    ------
    NagValueError
        (`errno` -3)
            On initial entry, m = *<value>*.

            Constraint: m >= 0.

        (`errno` -2)
            On initial entry, n = *<value>*.

            Constraint: n >= 0.

        (`errno` -1)
            On intermediate re-entry, `irevcm` = *<value>*.

            Constraint: `irevcm` = 1, 2, 3, 4 or 5.

        (`errno` -1)
            On initial entry, `irevcm` = *<value>*.

            Constraint: `irevcm` = 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            e^(tA)B has been computed using an IEEE double precision
            Taylor series, although the arithmetic precision is higher
            than IEEE double precision.

    References
    ----------
    Al--Mohy, A H and Higham, N J, 2011, `Computing the action of the
    matrix exponential, with an application to exponential integrators`,
    SIAM J. Sci. Statist. Comput. (33(2)), 488-511

    Higham, N J, 2008, `Functions of Matrices: Theory and Computation`,
    SIAM, Philadelphia, PA, USA
    """

    fun_name = 'naginterfaces.library.matop.real_gen_matrix_actexp_rcomm'

    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'b2'},
        b2,
    )
    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
    )
    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
    )
    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'p'},
        p,
    )
    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'r'},
        r,
    )
    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'z'},
        z,
    )
    try:
        _return_scalar = _b_matop.real_gen_matrix_actexp_rcomm(
            irevcm,
            _n,
            _m,
            b,
            t,
            tr,
            b2,
            x,
            y,
            p,
            r,
            z,
            comm,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _irevcm_py = _return_scalar
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['irevcm'],
    )(
        _irevcm_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _irevcm_py

def complex_gen_matrix_actexp(a, b, t):
    u"""
    Action of a complex matrix exponential on a complex matrix.

    ``complex_gen_matrix_actexp`` computes the action of the matrix
    exponential e^(tA), on the matrix B, where A is a complex n by n
    matrix, B is a complex n by m matrix and t is a complex scalar.

    For full information please refer to the NAG Library document for
    f01ha

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01haf.html

    Parameters
    ----------
    a : complex, array-like, shape (n, n)
        The n by n matrix A.

    b : complex, array-like, shape (n, m)
        The n by m matrix B.

    t : complex
        The scalar t.

    Returns
    -------
    a : complex, ndarray, shape (n, n)
        A is overwritten during the computation.

    b : complex, ndarray, shape (n, m)
        The n by m matrix e^(tA)B.

    Raises
    ------
    NagValueError
        (`errno` -2)
            On entry, m = *<value>*.

            Constraint: m >= 0.

        (`errno` -1)
            On entry, n = *<value>*.

            Constraint: n >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            e^(tA)B has been computed using an IEEE double precision
            Taylor series, although the arithmetic precision is higher
            than IEEE double precision.

    References
    ----------
    Al--Mohy, A H and Higham, N J, 2011, `Computing the action of the
    matrix exponential, with an application to exponential integrators`,
    SIAM J. Sci. Statist. Comput. (33(2)), 488-511

    Higham, N J, 2008, `Functions of Matrices: Theory and Computation`,
    SIAM, Philadelphia, PA, USA
    """

    fun_name = 'naginterfaces.library.matop.complex_gen_matrix_actexp'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
    )
    _m_py = _b_utils._EngineComplex128ArrayType.get_shape(_b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    try:
        _b_matop.complex_gen_matrix_actexp(
            _n,
            _m,
            _a,
            _b,
            t,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _a_py = _a
    _b_py = _b
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'b'],
    )(
        _a_py,
        _b_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def complex_gen_matrix_actexp_rcomm(irevcm, b, t, b2, x, y, p, r, z, comm, tr=0.0):
    u"""
    Action of a complex matrix exponential on a complex matrix (reverse
    communication).

    ``complex_gen_matrix_actexp_rcomm`` computes the action of the
    matrix exponential e^(tA), on the matrix B, where A is a complex n
    by n matrix, B is a complex n by m matrix and t is a complex scalar.
    It uses reverse communication for evaluating matrix products, so
    that the matrix A is not accessed explicitly.

    For full information please refer to the NAG Library document for
    f01hb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01hbf.html

    Parameters
    ----------
    irevcm : int
        `On initial entry`: must be set to 0.

    b : complex, ndarray, shape (n, m), modified in place
        `On initial entry`: the n by m matrix B.

        `On intermediate exit`: if `irevcm` = 1, contains the n by m
        matrix B.

        `On intermediate entry`: must not be changed.

        `On final exit`: the n by m matrix e^(tA)B.

    t : complex
        The scalar t.

    b2 : complex, ndarray, shape (n, m), modified in place
        `On initial entry`: need not be set.

        `On intermediate entry`: if `irevcm` = 1, must contain AB.

        `On final exit`: the array is undefined.

    x : complex, ndarray, shape (n, 2), modified in place
        `On initial entry`: need not be set.

        `On intermediate exit`: if `irevcm` = 2, contains the current n
        by 2 matrix X.

        `On intermediate entry`: if `irevcm` = 3, must contain A^HY.

        `On final exit`: the array is undefined.

    y : complex, ndarray, shape (n, 2), modified in place
        `On initial entry`: need not be set.

        `On intermediate exit`: if `irevcm` = 3, contains the current n
        by 2 matrix Y.

        `On intermediate entry`: if `irevcm` = 2, must contain AX.

        `On final exit`: the array is undefined.

    p : complex, ndarray, shape (n), modified in place
        `On initial entry`: need not be set.

        `On intermediate entry`: if `irevcm` = 4, must contain Az.

        `On final exit`: the array is undefined.

    r : complex, ndarray, shape (n), modified in place
        `On initial entry`: need not be set.

        `On intermediate entry`: if `irevcm` = 5, must contain A^Hz.

        `On final exit`: the array is undefined.

    z : complex, ndarray, shape (n), modified in place
        `On initial entry`: need not be set.

        `On intermediate exit`: if `irevcm` = 4 or 5, contains the
        vector z.

        `On intermediate entry`: must not be changed.

        `On final exit`: the array is undefined.

    comm : dict, communication object, modified in place
        Communication structure.

        `On initial entry`: need not be set.

    tr : complex, optional
        The trace of A. If this is not available then any number can be
        supplied (0.0 is a reasonable default); however, in the trivial
        case, n = 1, the result e^(`tr`t)B is immediately returned in
        the first row of B. See [Further Comments].

    Returns
    -------
    irevcm : int
        `On intermediate exit`: `irevcm` = 1, 2, 3, 4 or 5. The calling
        program must:

        (a) if `irevcm` = 1: evaluate B_2 = AB, where B_2 is an n by m
            matrix, and store the result in `b2`;

            if `irevcm` = 2: evaluate Y = AX, where X and Y are n by 2
            matrices, and store the result in `y`;

            if `irevcm` = 3: evaluate X = A^HY and store the result in
            `x`;

            if `irevcm` = 4: evaluate p = Az and store the result in
            `p`;

            if `irevcm` = 5: evaluate r = A^Hz and store the result in
            `r`.

        (#) call ``complex_gen_matrix_actexp_rcomm`` again with all
            other parameters unchanged.

        `On final exit`: `irevcm` = 0.

    Raises
    ------
    NagValueError
        (`errno` -3)
            On initial entry, m = *<value>*.

            Constraint: m >= 0.

        (`errno` -2)
            On initial entry, n = *<value>*.

            Constraint: n >= 0.

        (`errno` -1)
            On intermediate re-entry, `irevcm` = *<value>*.

            Constraint: `irevcm` = 1, 2, 3, 4 or 5.

        (`errno` -1)
            On initial entry, `irevcm` = *<value>*.

            Constraint: `irevcm` = 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            e^(tA)B has been computed using an IEEE double precision
            Taylor series, although the arithmetic precision is higher
            than IEEE double precision.

    References
    ----------
    Al--Mohy, A H and Higham, N J, 2011, `Computing the action of the
    matrix exponential, with an application to exponential integrators`,
    SIAM J. Sci. Statist. Comput. (33(2)), 488-511

    Higham, N J, 2008, `Functions of Matrices: Theory and Computation`,
    SIAM, Philadelphia, PA, USA
    """

    fun_name = 'naginterfaces.library.matop.complex_gen_matrix_actexp_rcomm'

    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _m_py = _b_utils._EngineComplex128ArrayType.get_shape(b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2)[1]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'b2'},
        b2,
    )
    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
    )
    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
    )
    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'p'},
        p,
    )
    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'r'},
        r,
    )
    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'z'},
        z,
    )
    try:
        _return_scalar = _b_matop.complex_gen_matrix_actexp_rcomm(
            irevcm,
            _n,
            _m,
            b,
            t,
            tr,
            b2,
            x,
            y,
            p,
            r,
            z,
            comm,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _irevcm_py = _return_scalar
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['irevcm'],
    )(
        _irevcm_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _irevcm_py

def real_gen_matrix_cond_std(fun, a):
    u"""
    Condition number for the exponential, logarithm, sine, cosine, sinh
    or cosh of a real matrix.

    ``real_gen_matrix_cond_std`` computes an estimate of the absolute
    condition number of a matrix function f at a real n by n matrix A in
    the 1-norm, where f is either the exponential, logarithm, sine,
    cosine, hyperbolic sine (sinh) or hyperbolic cosine (cosh).
    The evaluation of the matrix function, f(A), is also returned.

    For full information please refer to the NAG Library document for
    f01ja

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01jaf.html

    Parameters
    ----------
    fun : str
        Indicates which matrix function will be used.

        `fun` = 'EXP'

            The matrix exponential, e^A, will be used.

        `fun` = 'SIN'

            The matrix sine, sin(A), will be used.

        `fun` = 'COS'

            The matrix cosine, cos(A), will be used.

        `fun` = 'SINH'

            The hyperbolic matrix sine, sinh(A), will be used.

        `fun` = 'COSH'

            The hyperbolic matrix cosine, cosh(A), will be used.

        `fun` = 'LOG'

            The matrix logarithm, log(A), will be used.

    a : float, array-like, shape (n, n)
        The n by n matrix A.

    Returns
    -------
    a : float, ndarray, shape (n, n)
        The n by n matrix, f(A).

    conda : float
        An estimate of the absolute condition number of f at A.

    norma : float
        The 1-norm of A.

    normfa : float
        The 1-norm of f(A).

    Raises
    ------
    NagValueError
        (`errno` -2)
            Input argument number *<value>* is invalid.

        (`errno` -1)
            Input argument number *<value>* is invalid.

        (`errno` 1)
            An internal error occurred when evaluating the matrix
            function f(A). Please contact NAG.

        (`errno` 2)
            An internal error occurred when estimating the norm of the
            Fréchet derivative of f at A. Please contact NAG.

    References
    ----------
    Higham, N J, 2008, `Functions of Matrices: Theory and Computation`,
    SIAM, Philadelphia, PA, USA
    """

    fun_name = 'naginterfaces.library.matop.real_gen_matrix_cond_std'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _return_scalars = _b_matop.real_gen_matrix_cond_std(
        fun,
        _n,
        _a,
    )
    _a_py = _a
    _conda_py = _return_scalars.conda
    _norma_py = _return_scalars.norma
    _normfa_py = _return_scalars.normfa
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'conda', 'norma', 'normfa'],
    )(
        _a_py,
        _conda_py,
        _norma_py,
        _normfa_py,
    )
    return _return_data

def real_gen_matrix_cond_num(a, f, data=None):
    u"""
    Condition number for a function of a real matrix (using numerical
    differentiation).

    ``real_gen_matrix_cond_num`` computes an estimate of the absolute
    condition number of a matrix function f at a real n by n matrix A in
    the 1-norm.
    Numerical differentiation is used to evaluate the derivatives of f
    when they are required.

    For full information please refer to the NAG Library document for
    f01jb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01jbf.html

    Parameters
    ----------
    a : float, array-like, shape (n, n)
        The n by n matrix A.

    f : callable fz = f(z, data=None)
        The function `f` evaluates f(z_i) at a number of points z_i.

        Parameters
        ~~~~~~~~~~
        z : complex, ndarray, shape (nz)
            The n_z points z_1,z_2,...,z_n_z at which the function f is
            to be evaluated.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        fz : complex, array-like, shape (nz)
            The n_z function values. `fz`[i-1] should return the value
            f(z_i), for i = 1,2,...,n_z. If z_i lies on the real line,
            then so must f(z_i).

    data : arbitrary, optional
        User-communication data for callback functions.

    Returns
    -------
    a : float, ndarray, shape (n, n)
        The n by n matrix, f(A).

    conda : float
        An estimate of the absolute condition number of f at A.

    norma : float
        The 1-norm of A.

    normfa : float
        The 1-norm of f(A).

    Raises
    ------
    NagValueError
        (`errno` -1)
            Input argument number *<value>* is invalid.

        (`errno` 1)
            An internal error occurred when estimating the norm of the
            Fréchet derivative of f at A. Please contact NAG.

        (`errno` 2)
            An internal error occurred when evaluating the matrix
            function f(A). You can investigate further by calling
            ``real_gen_matrix_fun_num`` with the matrix A and the
            function f.

    Warns
    -----
    NagCallbackTerminateWarning
        (`errno` 3)
            Termination requested in `f`.

    References
    ----------
    Higham, N J, 2008, `Functions of Matrices: Theory and Computation`,
    SIAM, Philadelphia, PA, USA
    """

    fun_name = 'naginterfaces.library.matop.real_gen_matrix_cond_num'

    cb_f_name = fun_name + ':f'

    def cb_f(nz, z, fz, data=None):
        """Wrapping user's f. Called as base f."""
        if data is not None:
            _fz_f = f(
                z,
                data,
            )
        else:
            _fz_f = f(
                z,
            )
        fz[:] = _utils._to_ndarray(
            {'fun_name': cb_f_name, 'returned': 'fz'},
            _fz_f,
            copy=False,
            exp_shape=(nz,),
            exp_class=_b_utils._EngineComplex128ArrayType,
        )

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b_utils._check_callable(f, {'fun_name': fun_name, 'entity': 'f'})
    _f = cb_f
    try:
        _return_scalars = _b_matop.real_gen_matrix_cond_num(
            _n,
            _a,
            _f,
            data,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _a_py = _a
    _conda_py = _return_scalars.conda
    _norma_py = _return_scalars.norma
    _normfa_py = _return_scalars.normfa
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'conda', 'norma', 'normfa'],
    )(
        _a_py,
        _conda_py,
        _norma_py,
        _normfa_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def real_gen_matrix_cond_usd(a, f, data=None):
    u"""
    Condition number for a function of a real matrix (using
    user-supplied derivatives).

    ``real_gen_matrix_cond_usd`` computes an estimate of the absolute
    condition number of a matrix function f at a real n by n matrix A in
    the 1-norm, using analytical derivatives of f you have supplied.

    For full information please refer to the NAG Library document for
    f01jc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01jcf.html

    Parameters
    ----------
    a : float, array-like, shape (n, n)
        The n by n matrix A.

    f : callable fz = f(m, z, data=None)
        Given an integer m, the function `f` evaluates f^(m)(z_i) at a
        number of points z_i.

        Parameters
        ~~~~~~~~~~
        m : int
            The order, m, of the derivative required.

            If `m` = 0, f(z_i) should be returned.

            For `m` > 0, f^(m)(z_i) should be returned.

        z : complex, ndarray, shape (nz)
            The n_z points z_1,z_2,...,z_n_z at which the function f is
            to be evaluated.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        fz : complex, array-like, shape (nz)
            The n_z function or derivative values. `fz`[i-1] should
            return the value f^(m)(z_i), for i = 1,2,...,n_z. If z_i
            lies on the real line, then so must f^(m)(z_i).

    data : arbitrary, optional
        User-communication data for callback functions.

    Returns
    -------
    a : float, ndarray, shape (n, n)
        The n by n matrix, f(A).

    conda : float
        An estimate of the absolute condition number of f at A.

    norma : float
        The 1-norm of A.

    normfa : float
        The 1-norm of f(A).

    Raises
    ------
    NagValueError
        (`errno` -1)
            Input argument number *<value>* is invalid.

        (`errno` 1)
            An internal error occurred when estimating the norm of the
            Fréchet derivative of f at A. Please contact NAG.

        (`errno` 2)
            An internal error occurred when evaluating the matrix
            function f(A). You can investigate further by calling
            ``real_gen_matrix_fun_usd`` with the matrix A and the
            function f.

    Warns
    -----
    NagCallbackTerminateWarning
        (`errno` 3)
            Termination requested in `f`.

    References
    ----------
    Higham, N J, 2008, `Functions of Matrices: Theory and Computation`,
    SIAM, Philadelphia, PA, USA
    """

    fun_name = 'naginterfaces.library.matop.real_gen_matrix_cond_usd'

    cb_f_name = fun_name + ':f'

    def cb_f(m, nz, z, fz, data=None):
        """Wrapping user's f. Called as base f."""
        if data is not None:
            _fz_f = f(
                m,
                z,
                data,
            )
        else:
            _fz_f = f(
                m,
                z,
            )
        fz[:] = _utils._to_ndarray(
            {'fun_name': cb_f_name, 'returned': 'fz'},
            _fz_f,
            copy=False,
            exp_shape=(nz,),
            exp_class=_b_utils._EngineComplex128ArrayType,
        )

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b_utils._check_callable(f, {'fun_name': fun_name, 'entity': 'f'})
    _f = cb_f
    try:
        _return_scalars = _b_matop.real_gen_matrix_cond_usd(
            _n,
            _a,
            _f,
            data,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _a_py = _a
    _conda_py = _return_scalars.conda
    _norma_py = _return_scalars.norma
    _normfa_py = _return_scalars.normfa
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'conda', 'norma', 'normfa'],
    )(
        _a_py,
        _conda_py,
        _norma_py,
        _normfa_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def real_gen_matrix_cond_sqrt(a):
    u"""
    Condition number for square root of real matrix.

    ``real_gen_matrix_cond_sqrt`` computes an estimate of the relative
    condition number kappa_A^(1/2) and a bound on the relative residual,
    in the Frobenius norm, for the square root of a real n by n matrix
    A.
    The principal square root, A^(1/2), of A is also returned.

    For full information please refer to the NAG Library document for
    f01jd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01jdf.html

    Parameters
    ----------
    a : float, array-like, shape (n, n)
        The n by n matrix A.

    Returns
    -------
    a : float, ndarray, shape (n, n)
        Contains, if no exception or warning is raised, the n by n
        principal matrix square root, A^(1/2). Alternatively, if `errno`
        = 1, contains an n by n non-principal square root of A.

    alpha : float
        An estimate of the stability of the relative residual for the
        computed principal (if no exception or warning is raised) or
        non-principal (if `errno` = 1) matrix square root, alpha.

    condsa : float
        An estimate of the relative condition number, in the Frobenius
        norm, of the principal (if no exception or warning is raised) or
        non-principal (if `errno` = 1) matrix square root at A,
        kappa_A^(1/2).

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, n = *<value>*.

            Constraint: n >= 0.

        (`errno` 2)
            A has a defective vanishing eigenvalue. The square root and
            condition number cannot be found in this case.

        (`errno` 3)
            A has a negative real eigenvalue. The principal square root
            is not defined. ``complex_gen_matrix_cond_sqrt`` can be used
            to return a complex, non-principal square root.

        (`errno` 4)
            An error occurred when computing the matrix square root.
            Consequently, `alpha` and `condsa` could not be computed. It
            is likely that the function was called incorrectly.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            A has a semisimple vanishing eigenvalue. A non-principal
            square root was returned.

        (`errno` 5)
            An error occurred when computing the condition number. The
            matrix square root was still returned but you should use
            ``real_gen_matrix_sqrt`` to check if it is the principal
            matrix square root.

    References
    ----------
    Björck, Å and Hammarling, S, 1983, `A Schur method for the square
    root of a matrix`, Linear Algebra Appl. (52/53), 127--140

    Deadman, E, Higham, N J and Ralha, R, 2013, `Blocked Schur
    Algorithms for Computing the Matrix Square Root`, `Applied Parallel
    and Scientific Computing: 11th International Conference, (PARA 2012,
    Helsinki, Finland)`, P. Manninen and P. Öster, Eds, Lecture Notes in
    Computer Science (7782), 171--181, Springer--Verlag

    Higham, N J, 1987, `Computing real square roots of a real matrix`,
    Linear Algebra Appl. (88/89), 405--430

    Higham, N J, 2008, `Functions of Matrices: Theory and Computation`,
    SIAM, Philadelphia, PA, USA
    """

    fun_name = 'naginterfaces.library.matop.real_gen_matrix_cond_sqrt'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    try:
        _return_scalars = _b_matop.real_gen_matrix_cond_sqrt(
            _n,
            _a,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _a_py = _a
    _alpha_py = _return_scalars.alpha
    _condsa_py = _return_scalars.condsa
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'alpha', 'condsa'],
    )(
        _a_py,
        _alpha_py,
        _condsa_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def real_gen_matrix_cond_pow(a, p):
    u"""
    Condition number for real matrix power.

    ``real_gen_matrix_cond_pow`` computes an estimate of the relative
    condition number kappa_A^p of the pth power (where p is real) of a
    real n by n matrix A, in the 1-norm.
    The principal matrix power A^p is also returned.

    For full information please refer to the NAG Library document for
    f01je

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01jef.html

    Parameters
    ----------
    a : float, array-like, shape (n, n)
        The n by n matrix A.

    p : float
        The required power of A.

    Returns
    -------
    a : float, ndarray, shape (n, n)
        The n by n principal matrix pth power, A^p.

    condpa : float
        If the function exits successfully or `errno` = 3, an estimate
        of the relative condition number of the matrix pth power,
        kappa_A^p. Alternatively, if `errno` = 4, the absolute condition
        number of the matrix pth power.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, n = *<value>*.

            Constraint: n >= 0.

        (`errno` 1)
            A has eigenvalues on the negative real line. The principal
            pth power is not defined in this case;
            ``complex_gen_matrix_cond_pow`` can be used to find a
            complex, non-principal pth power.

        (`errno` 2)
            A is singular so the pth power cannot be computed.

        (`errno` 4)
            The relative condition number is infinite. The absolute
            condition number was returned instead.

        (`errno` 5)
            An unexpected internal error occurred. This failure should
            not occur and suggests that the function has been called
            incorrectly.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            A^p has been computed using an IEEE double precision Padé
            approximant, although the arithmetic precision is higher
            than IEEE double precision.

    References
    ----------
    Higham, N J, 2008, `Functions of Matrices: Theory and Computation`,
    SIAM, Philadelphia, PA, USA

    Higham, N J and Lin, L, 2011, `A Schur--Padé algorithm for
    fractional powers of a matrix`, SIAM J. Matrix Anal. Appl. (32(3)),
    1056--1078

    Higham, N J and Lin, L, 2013, `An improved Schur--Padé algorithm for
    fractional powers of a matrix and their Fréchet derivatives`, SIAM
    J. Matrix Anal. Appl. (34(3)), 1341--1360
    """

    fun_name = 'naginterfaces.library.matop.real_gen_matrix_cond_pow'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    try:
        _return_scalar = _b_matop.real_gen_matrix_cond_pow(
            _n,
            _a,
            p,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _a_py = _a
    _condpa_py = _return_scalar
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'condpa'],
    )(
        _a_py,
        _condpa_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def real_gen_matrix_frcht_pow(a, e, p):
    u"""
    Fréchet derivative of real matrix power.

    ``real_gen_matrix_frcht_pow`` computes the Fréchet derivative L(A,
    E) of the pth power (where p is real) of the real n by n matrix A
    applied to the real n by n matrix E.
    The principal matrix power A^p is also returned.

    For full information please refer to the NAG Library document for
    f01jf

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01jff.html

    Parameters
    ----------
    a : float, array-like, shape (n, n)
        The n by n matrix A.

    e : float, array-like, shape (n, n)
        The n by n matrix E.

    p : float
        The required power of A.

    Returns
    -------
    a : float, ndarray, shape (n, n)
        The n by n principal matrix pth power, A^p.

    e : float, ndarray, shape (n, n)
        The Fréchet derivative L(A, E).

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, n = *<value>*.

            Constraint: n >= 0.

        (`errno` 1)
            A has eigenvalues on the negative real line. The principal
            pth power is not defined in this case;
            ``complex_gen_matrix_frcht_pow`` can be used to find a
            complex, non-principal pth power.

        (`errno` 2)
            A is singular so the pth power cannot be computed.

        (`errno` 4)
            An unexpected internal error occurred. This failure should
            not occur and suggests that the function has been called
            incorrectly.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            A^p has been computed using an IEEE double precision Padé
            approximant, although the arithmetic precision is higher
            than IEEE double precision.

    References
    ----------
    Higham, N J, 2008, `Functions of Matrices: Theory and Computation`,
    SIAM, Philadelphia, PA, USA

    Higham, N J and Lin, L, 2011, `A Schur--Padé algorithm for
    fractional powers of a matrix`, SIAM J. Matrix Anal. Appl. (32(3)),
    1056--1078

    Higham, N J and Lin, L, 2013, `An improved Schur--Padé algorithm for
    fractional powers of a matrix and their Fréchet derivatives`, SIAM
    J. Matrix Anal. Appl. (34(3)), 1341--1360
    """

    fun_name = 'naginterfaces.library.matop.real_gen_matrix_frcht_pow'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _e = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'e'},
        e,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    try:
        _b_matop.real_gen_matrix_frcht_pow(
            _n,
            _a,
            _e,
            p,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _a_py = _a
    _e_py = _e
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'e'],
    )(
        _a_py,
        _e_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def real_gen_matrix_cond_exp(a):
    u"""
    Condition number for real matrix exponential.

    ``real_gen_matrix_cond_exp`` computes an estimate of the relative
    condition number kappa_exp(A) of the exponential of a real n by n
    matrix A, in the 1-norm.
    The matrix exponential e^A is also returned.

    For full information please refer to the NAG Library document for
    f01jg

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01jgf.html

    Parameters
    ----------
    a : float, array-like, shape (n, n)
        The n by n matrix A.

    Returns
    -------
    a : float, ndarray, shape (n, n)
        The n by n matrix exponential e^A.

    condea : float
        An estimate of the relative condition number of the matrix
        exponential kappa_exp(A).

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, n = *<value>*.

            Constraint: n >= 0.

        (`errno` 1)
            The linear equations to be solved for the Padé approximant
            are singular; it is likely that this function has been
            called incorrectly.

        (`errno` 3)
            An unexpected internal error has occurred. Please contact
            NAG.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            The arithmetic precision is higher than that used for the
            Padé approximant computed matrix exponential.

    References
    ----------
    Al--Mohy, A H and Higham, N J, 2009, `A new scaling and squaring
    algorithm for the matrix exponential`, SIAM J. Matrix Anal. (31(3)),
    970--989

    Al--Mohy, A H and Higham, N J, 2009, `Computing the Fréchet
    derivative of the matrix exponential, with an application to
    condition number estimation`, SIAM J. Matrix Anal. Appl. (30(4)),
    1639--1657

    Higham, N J, 2008, `Functions of Matrices: Theory and Computation`,
    SIAM, Philadelphia, PA, USA

    Moler, C B and Van Loan, C F, 2003, `Nineteen dubious ways to
    compute the exponential of a matrix, twenty-five years later`, SIAM
    Rev. (45), 3--49
    """

    fun_name = 'naginterfaces.library.matop.real_gen_matrix_cond_exp'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    try:
        _return_scalar = _b_matop.real_gen_matrix_cond_exp(
            _n,
            _a,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _a_py = _a
    _condea_py = _return_scalar
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'condea'],
    )(
        _a_py,
        _condea_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def real_gen_matrix_frcht_exp(a, e):
    u"""
    Fréchet derivative of real matrix exponential.

    ``real_gen_matrix_frcht_exp`` computes the Fréchet derivative L(A,
    E) of the matrix exponential of a real n by n matrix A applied to
    the real n by n matrix E.
    The matrix exponential e^A is also returned.

    For full information please refer to the NAG Library document for
    f01jh

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01jhf.html

    Parameters
    ----------
    a : float, array-like, shape (n, n)
        The n by n matrix A.

    e : float, array-like, shape (n, n)
        The n by n matrix E

    Returns
    -------
    a : float, ndarray, shape (n, n)
        The n by n matrix exponential e^A.

    e : float, ndarray, shape (n, n)
        The Fréchet derivative L(A, E)

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, n = *<value>*.

            Constraint: n >= 0.

        (`errno` 1)
            The linear equations to be solved for the Padé approximant
            are singular; it is likely that this function has been
            called incorrectly.

        (`errno` 3)
            An unexpected internal error has occurred. Please contact
            NAG.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            The arithmetic precision is higher than that used for the
            Padé approximant computed matrix exponential.

    References
    ----------
    Al--Mohy, A H and Higham, N J, 2009, `A new scaling and squaring
    algorithm for the matrix exponential`, SIAM J. Matrix Anal. (31(3)),
    970--989

    Al--Mohy, A H and Higham, N J, 2009, `Computing the Fréchet
    derivative of the matrix exponential, with an application to
    condition number estimation`, SIAM J. Matrix Anal. Appl. (30(4)),
    1639--1657

    Higham, N J, 2008, `Functions of Matrices: Theory and Computation`,
    SIAM, Philadelphia, PA, USA

    Moler, C B and Van Loan, C F, 2003, `Nineteen dubious ways to
    compute the exponential of a matrix, twenty-five years later`, SIAM
    Rev. (45), 3--49
    """

    fun_name = 'naginterfaces.library.matop.real_gen_matrix_frcht_exp'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _e = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'e'},
        e,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    try:
        _b_matop.real_gen_matrix_frcht_exp(
            _n,
            _a,
            _e,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _a_py = _a
    _e_py = _e
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'e'],
    )(
        _a_py,
        _e_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def real_gen_matrix_cond_log(a):
    u"""
    Condition number for real matrix logarithm.

    ``real_gen_matrix_cond_log`` computes an estimate of the relative
    condition number kappa_log(A) of the logarithm of a real n by n
    matrix A, in the 1-norm.
    The principal matrix logarithm log(A) is also returned.

    For full information please refer to the NAG Library document for
    f01jj

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01jjf.html

    Parameters
    ----------
    a : float, array-like, shape (n, n)
        The n by n matrix A.

    Returns
    -------
    a : float, ndarray, shape (n, n)
        The n by n principal matrix logarithm, log(A).

    condla : float
        With the function exits successfully or `errno` = 3, an estimate
        of the relative condition number of the matrix logarithm,
        kappa_log(A). Alternatively, if `errno` = 4, contains the
        absolute condition number of the matrix logarithm.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, n = *<value>*.

            Constraint: n >= 0.

        (`errno` 1)
            A is singular so the logarithm cannot be computed.

        (`errno` 2)
            A has eigenvalues on the negative real line. The principal
            logarithm is not defined in this case;
            ``complex_gen_matrix_cond_log`` can be used to return a
            complex, non-principal log.

        (`errno` 5)
            An unexpected internal error occurred. This failure should
            not occur and suggests that the function has been called
            incorrectly.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            log(A) has been computed using an IEEE double precision Padé
            approximant, although the arithmetic precision is higher
            than IEEE double precision.

        (`errno` 4)
            The relative condition number is infinite. The absolute
            condition number was returned instead.

    References
    ----------
    Al--Mohy, A H and Higham, N J, 2011, `Improved inverse scaling and
    squaring algorithms for the matrix logarithm`, SIAM J. Sci. Comput.
    (34(4)), C152--C169

    Al--Mohy, A H, Higham, N J and Relton, S D, 2012, `Computing the
    Fréchet derivative of the matrix logarithm and estimating the
    condition number`, SIAM J. Sci. Comput. (35(4)), C394--C410

    Higham, N J, 2008, `Functions of Matrices: Theory and Computation`,
    SIAM, Philadelphia, PA, USA
    """

    fun_name = 'naginterfaces.library.matop.real_gen_matrix_cond_log'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    try:
        _return_scalar = _b_matop.real_gen_matrix_cond_log(
            _n,
            _a,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _a_py = _a
    _condla_py = _return_scalar
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'condla'],
    )(
        _a_py,
        _condla_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def real_gen_matrix_frcht_log(a, e):
    u"""
    Fréchet derivative of real matrix logarithm.

    ``real_gen_matrix_frcht_log`` computes the Fréchet derivative L(A,
    E) of the matrix logarithm of the real n by n matrix A applied to
    the real n by n matrix E.
    The principal matrix logarithm log(A) is also returned.

    For full information please refer to the NAG Library document for
    f01jk

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01jkf.html

    Parameters
    ----------
    a : float, array-like, shape (n, n)
        The n by n matrix A.

    e : float, array-like, shape (n, n)
        The n by n matrix E

    Returns
    -------
    a : float, ndarray, shape (n, n)
        The n by n principal matrix logarithm, log(A).

    e : float, ndarray, shape (n, n)
        The Fréchet derivative L(A, E)

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, n = *<value>*.

            Constraint: n >= 0.

        (`errno` 1)
            A is singular so the logarithm cannot be computed.

        (`errno` 2)
            A has eigenvalues on the negative real line. The principal
            logarithm is not defined in this case;
            ``complex_gen_matrix_frcht_log`` can be used to return a
            complex, non-principal log.

        (`errno` 4)
            An unexpected internal error occurred. This failure should
            not occur and suggests that the function has been called
            incorrectly.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            log(A) has been computed using an IEEE double precision Padé
            approximant, although the arithmetic precision is higher
            than IEEE double precision.

    References
    ----------
    Al--Mohy, A H and Higham, N J, 2011, `Improved inverse scaling and
    squaring algorithms for the matrix logarithm`, SIAM J. Sci. Comput.
    (34(4)), C152--C169

    Al--Mohy, A H, Higham, N J and Relton, S D, 2012, `Computing the
    Fréchet derivative of the matrix logarithm and estimating the
    condition number`, SIAM J. Sci. Comput. (35(4)), C394--C410

    Higham, N J, 2008, `Functions of Matrices: Theory and Computation`,
    SIAM, Philadelphia, PA, USA
    """

    fun_name = 'naginterfaces.library.matop.real_gen_matrix_frcht_log'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _e = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'e'},
        e,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    try:
        _b_matop.real_gen_matrix_frcht_log(
            _n,
            _a,
            _e,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _a_py = _a
    _e_py = _e
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'e'],
    )(
        _a_py,
        _e_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def complex_gen_matrix_cond_std(fun, a):
    u"""
    Condition number for the exponential, logarithm, sine, cosine, sinh
    or cosh of a complex matrix.

    ``complex_gen_matrix_cond_std`` computes an estimate of the absolute
    condition number of a matrix function f of a complex n by n matrix A
    in the 1-norm, where f is either the exponential, logarithm, sine,
    cosine, hyperbolic sine (sinh) or hyperbolic cosine (cosh).
    The evaluation of the matrix function, f(A), is also returned.

    For full information please refer to the NAG Library document for
    f01ka

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01kaf.html

    Parameters
    ----------
    fun : str
        Indicates which matrix function will be used.

        `fun` = 'EXP'

            The matrix exponential, e^A, will be used.

        `fun` = 'SIN'

            The matrix sine, sin(A), will be used.

        `fun` = 'COS'

            The matrix cosine, cos(A), will be used.

        `fun` = 'SINH'

            The hyperbolic matrix sine, sinh(A), will be used.

        `fun` = 'COSH'

            The hyperbolic matrix cosine, cosh(A), will be used.

        `fun` = 'LOG'

            The matrix logarithm, log(A), will be used.

    a : complex, array-like, shape (n, n)
        The n by n matrix A.

    Returns
    -------
    a : complex, ndarray, shape (n, n)
        The n by n matrix, f(A).

    conda : float
        An estimate of the absolute condition number of f at A.

    norma : float
        The 1-norm of A.

    normfa : float
        The 1-norm of f(A).

    Raises
    ------
    NagValueError
        (`errno` -2)
            Input argument number *<value>* is invalid.

        (`errno` -1)
            Input argument number *<value>* is invalid.

        (`errno` 1)
            An internal error occurred when estimating the norm of the
            Fréchet derivative of f at A. Please contact NAG.

        (`errno` 2)
            An internal error occurred when evaluating the matrix
            function f(A). You can investigate further by calling
            ``complex_gen_matrix_exp``, ``complex_gen_matrix_log`` or
            ``complex_gen_matrix_fun_std`` with the matrix A.

    References
    ----------
    Higham, N J, 2008, `Functions of Matrices: Theory and Computation`,
    SIAM, Philadelphia, PA, USA
    """

    fun_name = 'naginterfaces.library.matop.complex_gen_matrix_cond_std'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _return_scalars = _b_matop.complex_gen_matrix_cond_std(
        fun,
        _n,
        _a,
    )
    _a_py = _a
    _conda_py = _return_scalars.conda
    _norma_py = _return_scalars.norma
    _normfa_py = _return_scalars.normfa
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'conda', 'norma', 'normfa'],
    )(
        _a_py,
        _conda_py,
        _norma_py,
        _normfa_py,
    )
    return _return_data

def complex_gen_matrix_cond_num(a, f, data=None):
    u"""
    Condition number for a function of a complex matrix (using numerical
    differentiation).

    ``complex_gen_matrix_cond_num`` computes an estimate of the absolute
    condition number of a matrix function f of a complex n by n matrix A
    in the 1-norm.
    Numerical differentiation is used to evaluate the derivatives of f
    when they are required.

    For full information please refer to the NAG Library document for
    f01kb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01kbf.html

    Parameters
    ----------
    a : complex, array-like, shape (n, n)
        The n by n matrix A.

    f : callable fz = f(z, data=None)
        The function `f` evaluates f(z_i) at a number of points z_i.

        Parameters
        ~~~~~~~~~~
        z : complex, ndarray, shape (nz)
            The n_z points z_1,z_2,...,z_n_z at which the function f is
            to be evaluated.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        fz : complex, array-like, shape (nz)
            The n_z function values. `fz`[i-1] should return the value
            f(z_i), for i = 1,2,...,n_z.

    data : arbitrary, optional
        User-communication data for callback functions.

    Returns
    -------
    a : complex, ndarray, shape (n, n)
        The n by n matrix, f(A).

    conda : float
        An estimate of the absolute condition number of f at A.

    norma : float
        The 1-norm of A.

    normfa : float
        The 1-norm of f(A).

    Raises
    ------
    NagValueError
        (`errno` -1)
            Input argument number *<value>* is invalid.

        (`errno` 1)
            An internal error occurred when estimating the norm of the
            Fréchet derivative of f at A. Please contact NAG.

        (`errno` 2)
            An internal error occurred while evaluating the matrix
            function f(A). You can investigate further by calling
            ``complex_gen_matrix_fun_num`` with the matrix A and the
            function f.

    Warns
    -----
    NagCallbackTerminateWarning
        (`errno` 3)
            Termination requested in `f`.

    References
    ----------
    Higham, N J, 2008, `Functions of Matrices: Theory and Computation`,
    SIAM, Philadelphia, PA, USA
    """

    fun_name = 'naginterfaces.library.matop.complex_gen_matrix_cond_num'

    cb_f_name = fun_name + ':f'

    def cb_f(nz, z, fz, data=None):
        """Wrapping user's f. Called as base f."""
        if data is not None:
            _fz_f = f(
                z,
                data,
            )
        else:
            _fz_f = f(
                z,
            )
        fz[:] = _utils._to_ndarray(
            {'fun_name': cb_f_name, 'returned': 'fz'},
            _fz_f,
            copy=False,
            exp_shape=(nz,),
            exp_class=_b_utils._EngineComplex128ArrayType,
        )

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b_utils._check_callable(f, {'fun_name': fun_name, 'entity': 'f'})
    _f = cb_f
    try:
        _return_scalars = _b_matop.complex_gen_matrix_cond_num(
            _n,
            _a,
            _f,
            data,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _a_py = _a
    _conda_py = _return_scalars.conda
    _norma_py = _return_scalars.norma
    _normfa_py = _return_scalars.normfa
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'conda', 'norma', 'normfa'],
    )(
        _a_py,
        _conda_py,
        _norma_py,
        _normfa_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def complex_gen_matrix_cond_usd(a, f, data=None):
    u"""
    Condition number for a function of a complex matrix (using
    user-supplied derivatives).

    ``complex_gen_matrix_cond_usd`` computes an estimate of the absolute
    condition number of a matrix function f of a complex n by n matrix A
    in the 1-norm, using analytical derivatives of f you have supplied.

    For full information please refer to the NAG Library document for
    f01kc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01kcf.html

    Parameters
    ----------
    a : complex, array-like, shape (n, n)
        The n by n matrix A.

    f : callable fz = f(m, z, data=None)
        Given an integer m, the function `f` evaluates f^(m)(z_i) at a
        number of points z_i.

        Parameters
        ~~~~~~~~~~
        m : int
            The order, m, of the derivative required.

            If `m` = 0, f(z_i) should be returned.

            For `m` > 0, f^(m)(z_i) should be returned.

        z : complex, ndarray, shape (nz)
            The n_z points z_1,z_2,...,z_n_z at which the function f is
            to be evaluated.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        fz : complex, array-like, shape (nz)
            The n_z function or derivative values. `fz`[i-1] should
            return the value f^(m)(z_i), for i = 1,2,...,n_z.

    data : arbitrary, optional
        User-communication data for callback functions.

    Returns
    -------
    a : complex, ndarray, shape (n, n)
        The n by n matrix, f(A).

    conda : float
        An estimate of the absolute condition number of f at A.

    norma : float
        The 1-norm of A.

    normfa : float
        The 1-norm of f(A).

    Raises
    ------
    NagValueError
        (`errno` -1)
            Input argument number *<value>* is invalid.

        (`errno` 1)
            An internal error occurred when estimating the norm of the
            Fréchet derivative of f at A. Please contact NAG.

        (`errno` 2)
            An internal error occurred when evaluating the matrix
            function f(A). You can investigate further by calling
            ``complex_gen_matrix_fun_usd`` with the matrix A and the
            function f.

    Warns
    -----
    NagCallbackTerminateWarning
        (`errno` 3)
            Termination requested in `f`.

    References
    ----------
    Higham, N J, 2008, `Functions of Matrices: Theory and Computation`,
    SIAM, Philadelphia, PA, USA
    """

    fun_name = 'naginterfaces.library.matop.complex_gen_matrix_cond_usd'

    cb_f_name = fun_name + ':f'

    def cb_f(m, nz, z, fz, data=None):
        """Wrapping user's f. Called as base f."""
        if data is not None:
            _fz_f = f(
                m,
                z,
                data,
            )
        else:
            _fz_f = f(
                m,
                z,
            )
        fz[:] = _utils._to_ndarray(
            {'fun_name': cb_f_name, 'returned': 'fz'},
            _fz_f,
            copy=False,
            exp_shape=(nz,),
            exp_class=_b_utils._EngineComplex128ArrayType,
        )

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b_utils._check_callable(f, {'fun_name': fun_name, 'entity': 'f'})
    _f = cb_f
    try:
        _return_scalars = _b_matop.complex_gen_matrix_cond_usd(
            _n,
            _a,
            _f,
            data,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _a_py = _a
    _conda_py = _return_scalars.conda
    _norma_py = _return_scalars.norma
    _normfa_py = _return_scalars.normfa
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'conda', 'norma', 'normfa'],
    )(
        _a_py,
        _conda_py,
        _norma_py,
        _normfa_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def complex_gen_matrix_cond_sqrt(a):
    u"""
    Condition number for square root of complex matrix.

    ``complex_gen_matrix_cond_sqrt`` computes an estimate of the
    relative condition number, kappa_A^(1/2), and a bound on the
    relative residual, in the Frobenius norm, for the square root of a
    complex n by n matrix A.
    The principal square root, A^(1/2), of A is also returned.

    For full information please refer to the NAG Library document for
    f01kd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01kdf.html

    Parameters
    ----------
    a : complex, array-like, shape (n, n)
        The n by n matrix A.

    Returns
    -------
    a : complex, ndarray, shape (n, n)
        The n by n principal matrix square root A^(1/2). Alternatively,
        if `errno` = 1, contains an n by n non-principal square root of
        A.

    alpha : float
        An estimate of the stability of the relative residual for the
        computed principal (if no exception or warning is raised) or
        non-principal (if `errno` = 1) matrix square root, alpha.

    condsa : float
        An estimate of the relative condition number, in the Frobenius
        norm, of the principal (if no exception or warning is raised) or
        non-principal (if `errno` = 1) matrix square root at A,
        kappa_A^(1/2).

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, n = *<value>*.

            Constraint: n >= 0.

        (`errno` 2)
            A has a defective vanishing eigenvalue. The square root and
            condition number cannot be found in this case.

        (`errno` 3)
            An error occurred when computing the matrix square root.
            Consequently, `alpha` and `condsa` could not be computed. It
            is likely that the function was called incorrectly.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            A has a negative or semisimple vanishing eigenvalue. A
            non-principal square root was returned.

        (`errno` 4)
            An error occurred when computing the condition number. The
            matrix square root was still returned but you should use
            ``complex_gen_matrix_sqrt`` to check if it is the principal
            matrix square root.

    References
    ----------
    Björck, Å and Hammarling, S, 1983, `A Schur method for the square
    root of a matrix`, Linear Algebra Appl. (52/53), 127--140

    Deadman, E, Higham, N J and Ralha, R, 2013, `Blocked Schur
    Algorithms for Computing the Matrix Square Root`, `Applied Parallel
    and Scientific Computing: 11th International Conference, (PARA 2012,
    Helsinki, Finland)`, P. Manninen and P. Öster, Eds, Lecture Notes in
    Computer Science (7782), 171--181, Springer--Verlag

    Higham, N J, 1987, `Computing real square roots of a real matrix`,
    Linear Algebra Appl. (88/89), 405--430

    Higham, N J, 2008, `Functions of Matrices: Theory and Computation`,
    SIAM, Philadelphia, PA, USA
    """

    fun_name = 'naginterfaces.library.matop.complex_gen_matrix_cond_sqrt'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    try:
        _return_scalars = _b_matop.complex_gen_matrix_cond_sqrt(
            _n,
            _a,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _a_py = _a
    _alpha_py = _return_scalars.alpha
    _condsa_py = _return_scalars.condsa
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'alpha', 'condsa'],
    )(
        _a_py,
        _alpha_py,
        _condsa_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def complex_gen_matrix_cond_pow(a, p):
    u"""
    Condition number for complex matrix power.

    ``complex_gen_matrix_cond_pow`` computes an estimate of the relative
    condition number kappa_A^p of the pth power (where p is real) of a
    complex n by n matrix A, in the 1-norm.
    The principal matrix power A^p is also returned.

    For full information please refer to the NAG Library document for
    f01ke

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01kef.html

    Parameters
    ----------
    a : complex, array-like, shape (n, n)
        The n by n matrix A.

    p : float
        The required power of A.

    Returns
    -------
    a : complex, ndarray, shape (n, n)
        The n by n principal matrix pth power, A^p, unless `errno` = 1,
        in which case a non-principal pth power is returned.

    condpa : float
        If the function exits successfully or `errno` = 3, an estimate
        of the relative condition number of the matrix pth power,
        kappa_A^p. Alternatively, if `errno` = 4, the absolute condition
        number of the matrix pth power.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, n = *<value>*.

            Constraint: n >= 0.

        (`errno` 2)
            A is singular so the pth power cannot be computed.

        (`errno` 5)
            An unexpected internal error occurred. This failure should
            not occur and suggests that the function has been called
            incorrectly.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            A has eigenvalues on the negative real line. The principal
            pth power is not defined in this case, so a non-principal
            power was returned.

        (`errno` 3)
            A^p has been computed using an IEEE double precision Padé
            approximant, although the arithmetic precision is higher
            than IEEE double precision.

        (`errno` 4)
            The relative condition number is infinite. The absolute
            condition number was returned instead.

    References
    ----------
    Higham, N J, 2008, `Functions of Matrices: Theory and Computation`,
    SIAM, Philadelphia, PA, USA

    Higham, N J and Lin, L, 2011, `A Schur--Padé algorithm for
    fractional powers of a matrix`, SIAM J. Matrix Anal. Appl. (32(3)),
    1056--1078

    Higham, N J and Lin, L, 2013, `An improved Schur--Padé algorithm for
    fractional powers of a matrix and their Fréchet derivatives`, SIAM
    J. Matrix Anal. Appl. (34(3)), 1341--1360
    """

    fun_name = 'naginterfaces.library.matop.complex_gen_matrix_cond_pow'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    try:
        _return_scalar = _b_matop.complex_gen_matrix_cond_pow(
            _n,
            _a,
            p,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _a_py = _a
    _condpa_py = _return_scalar
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'condpa'],
    )(
        _a_py,
        _condpa_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def complex_gen_matrix_frcht_pow(a, e, p):
    u"""
    Fréchet derivative of complex matrix power.

    ``complex_gen_matrix_frcht_pow`` computes the Fréchet derivative
    L(A, E) of the pth power (where p is real) of the complex n by n
    matrix A applied to the complex n by n matrix E.
    The principal matrix power A^p is also returned.

    For full information please refer to the NAG Library document for
    f01kf

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01kff.html

    Parameters
    ----------
    a : complex, array-like, shape (n, n)
        The n by n matrix A.

    e : complex, array-like, shape (n, n)
        The n by n matrix E.

    p : float
        The required power of A.

    Returns
    -------
    a : complex, ndarray, shape (n, n)
        The n by n principal matrix pth power, A^p. Alternatively if
        `errno` = 1, a non-principal pth power is returned.

    e : complex, ndarray, shape (n, n)
        The Fréchet derivative L(A, E).

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, n = *<value>*.

            Constraint: n >= 0.

        (`errno` 2)
            A is singular so the pth power cannot be computed.

        (`errno` 4)
            An unexpected internal error occurred. This failure should
            not occur and suggests that the function has been called
            incorrectly.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            A has eigenvalues on the negative real line. The principal
            pth power is not defined in this case, so a non-principal
            power was returned.

        (`errno` 3)
            A^p has been computed using an IEEE double precision Padé
            approximant, although the arithmetic precision is higher
            than IEEE double precision.

    References
    ----------
    Higham, N J, 2008, `Functions of Matrices: Theory and Computation`,
    SIAM, Philadelphia, PA, USA

    Higham, N J and Lin, L, 2011, `A Schur--Padé algorithm for
    fractional powers of a matrix`, SIAM J. Matrix Anal. Appl. (32(3)),
    1056--1078

    Higham, N J and Lin, L, 2013, `An improved Schur--Padé algorithm for
    fractional powers of a matrix and their Fréchet derivatives`, SIAM
    J. Matrix Anal. Appl. (34(3)), 1341--1360
    """

    fun_name = 'naginterfaces.library.matop.complex_gen_matrix_frcht_pow'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _e = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'e'},
        e,
        copy=True,
    )
    try:
        _b_matop.complex_gen_matrix_frcht_pow(
            _n,
            _a,
            _e,
            p,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _a_py = _a
    _e_py = _e
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'e'],
    )(
        _a_py,
        _e_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def complex_gen_matrix_cond_exp(a):
    u"""
    Condition number for complex matrix exponential.

    ``complex_gen_matrix_cond_exp`` computes an estimate of the relative
    condition number kappa_exp(A) of the exponential of a complex n by n
    matrix A, in the 1-norm.
    The matrix exponential e^A is also returned.

    For full information please refer to the NAG Library document for
    f01kg

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01kgf.html

    Parameters
    ----------
    a : complex, array-like, shape (n, n)
        The n by n matrix A.

    Returns
    -------
    a : complex, ndarray, shape (n, n)
        The n by n matrix exponential e^A.

    condea : float
        An estimate of the relative condition number of the matrix
        exponential kappa_exp(A).

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, n = *<value>*.

            Constraint: n >= 0.

        (`errno` 1)
            The linear equations to be solved for the Padé approximant
            are singular; it is likely that this function has been
            called incorrectly.

        (`errno` 3)
            An unexpected internal error has occurred. Please contact
            NAG.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            The arithmetic precision is higher than that used for the
            Padé approximant computed matrix exponential.

    References
    ----------
    Al--Mohy, A H and Higham, N J, 2009, `A new scaling and squaring
    algorithm for the matrix exponential`, SIAM J. Matrix Anal. (31(3)),
    970--989

    Al--Mohy, A H and Higham, N J, 2009, `Computing the Fréchet
    derivative of the matrix exponential, with an application to
    condition number estimation`, SIAM J. Matrix Anal. Appl. (30(4)),
    1639--1657

    Higham, N J, 2008, `Functions of Matrices: Theory and Computation`,
    SIAM, Philadelphia, PA, USA

    Moler, C B and Van Loan, C F, 2003, `Nineteen dubious ways to
    compute the exponential of a matrix, twenty-five years later`, SIAM
    Rev. (45), 3--49
    """

    fun_name = 'naginterfaces.library.matop.complex_gen_matrix_cond_exp'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    try:
        _return_scalar = _b_matop.complex_gen_matrix_cond_exp(
            _n,
            _a,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _a_py = _a
    _condea_py = _return_scalar
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'condea'],
    )(
        _a_py,
        _condea_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def complex_gen_matrix_frcht_exp(a, e):
    u"""
    Fréchet derivative of complex matrix exponential.

    ``complex_gen_matrix_frcht_exp`` computes the Fréchet derivative
    L(A, E) of the matrix exponential of a complex n by n matrix A
    applied to the complex n by n matrix E.
    The matrix exponential e^A is also returned.

    For full information please refer to the NAG Library document for
    f01kh

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01khf.html

    Parameters
    ----------
    a : complex, array-like, shape (n, n)
        The n by n matrix A.

    e : complex, array-like, shape (n, n)
        The n by n matrix E

    Returns
    -------
    a : complex, ndarray, shape (n, n)
        The n by n matrix exponential e^A.

    e : complex, ndarray, shape (n, n)
        The Fréchet derivative L(A, E)

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, n = *<value>*.

            Constraint: n >= 0.

        (`errno` 1)
            The linear equations to be solved for the Padé approximant
            are singular; it is likely that this function has been
            called incorrectly.

        (`errno` 3)
            An unexpected internal error has occurred. Please contact
            NAG.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            The arithmetic precision is higher than that used for the
            Padé approximant computed matrix exponential.

    References
    ----------
    Al--Mohy, A H and Higham, N J, 2009, `A new scaling and squaring
    algorithm for the matrix exponential`, SIAM J. Matrix Anal. (31(3)),
    970--989

    Al--Mohy, A H and Higham, N J, 2009, `Computing the Fréchet
    derivative of the matrix exponential, with an application to
    condition number estimation`, SIAM J. Matrix Anal. Appl. (30(4)),
    1639--1657

    Higham, N J, 2008, `Functions of Matrices: Theory and Computation`,
    SIAM, Philadelphia, PA, USA

    Moler, C B and Van Loan, C F, 2003, `Nineteen dubious ways to
    compute the exponential of a matrix, twenty-five years later`, SIAM
    Rev. (45), 3--49
    """

    fun_name = 'naginterfaces.library.matop.complex_gen_matrix_frcht_exp'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _e = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'e'},
        e,
        copy=True,
    )
    try:
        _b_matop.complex_gen_matrix_frcht_exp(
            _n,
            _a,
            _e,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _a_py = _a
    _e_py = _e
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'e'],
    )(
        _a_py,
        _e_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def complex_gen_matrix_cond_log(a):
    u"""
    Condition number for complex matrix logarithm.

    ``complex_gen_matrix_cond_log`` computes an estimate of the relative
    condition number kappa_log(A) of the logarithm of a complex n by n
    matrix A, in the 1-norm.
    The principal matrix logarithm log(A) is also returned.

    For full information please refer to the NAG Library document for
    f01kj

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01kjf.html

    Parameters
    ----------
    a : complex, array-like, shape (n, n)
        The n by n matrix A.

    Returns
    -------
    a : complex, ndarray, shape (n, n)
        The n by n principal matrix logarithm, log(A). Alternatively, if
        `errno` = 2, a non-principal logarithm is returned.

    condla : float
        With the function exits successfully or `errno` = 2 or 3, an
        estimate of the relative condition number of the matrix
        logarithm, kappa_log(A). Alternatively, if `errno` = 4, contains
        the absolute condition number of the matrix logarithm.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, n = *<value>*.

            Constraint: n >= 0.

        (`errno` 1)
            A is singular so the logarithm cannot be computed.

        (`errno` 5)
            An unexpected internal error occurred. This failure should
            not occur and suggests that the function has been called
            incorrectly.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            A has eigenvalues on the negative real line. The principal
            logarithm is not defined in this case, so a non-principal
            logarithm was returned.

        (`errno` 3)
            log(A) has been computed using an IEEE double precision Padé
            approximant, although the arithmetic precision is higher
            than IEEE double precision.

        (`errno` 4)
            The relative condition number is infinite. The absolute
            condition number was returned instead.

    References
    ----------
    Al--Mohy, A H and Higham, N J, 2011, `Improved inverse scaling and
    squaring algorithms for the matrix logarithm`, SIAM J. Sci. Comput.
    (34(4)), C152--C169

    Al--Mohy, A H, Higham, N J and Relton, S D, 2012, `Computing the
    Fréchet derivative of the matrix logarithm and estimating the
    condition number`, SIAM J. Sci. Comput. (35(4)), C394--C410

    Higham, N J, 2008, `Functions of Matrices: Theory and Computation`,
    SIAM, Philadelphia, PA, USA
    """

    fun_name = 'naginterfaces.library.matop.complex_gen_matrix_cond_log'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    try:
        _return_scalar = _b_matop.complex_gen_matrix_cond_log(
            _n,
            _a,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _a_py = _a
    _condla_py = _return_scalar
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'condla'],
    )(
        _a_py,
        _condla_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def complex_gen_matrix_frcht_log(a, e):
    u"""
    Fréchet derivative of complex matrix logarithm.

    ``complex_gen_matrix_frcht_log`` computes the Fréchet derivative
    L(A, E) of the matrix logarithm of the complex n by n matrix A
    applied to the complex n by n matrix E.
    The principal matrix logarithm log(A) is also returned.

    For full information please refer to the NAG Library document for
    f01kk

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01kkf.html

    Parameters
    ----------
    a : complex, array-like, shape (n, n)
        The n by n matrix A.

    e : complex, array-like, shape (n, n)
        The n by n matrix E

    Returns
    -------
    a : complex, ndarray, shape (n, n)
        The n by n principal matrix logarithm, log(A). Alterntively, if
        `errno` = 2, a non-principal logarithm is returned.

    e : complex, ndarray, shape (n, n)
        With the function exits successfully or `errno` = 2 or 3, the
        Fréchet derivative L(A, E)

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, n = *<value>*.

            Constraint: n >= 0.

        (`errno` 1)
            A is singular so the logarithm cannot be computed.

        (`errno` 4)
            An unexpected internal error occurred. This failure should
            not occur and suggests that the function has been called
            incorrectly.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            A has eigenvalues on the negative real line. The principal
            logarithm is not defined in this case, so a non-principal
            logarithm was returned.

        (`errno` 3)
            log(A) has been computed using an IEEE double precision Padé
            approximant, although the arithmetic precision is higher
            than IEEE double precision.

    References
    ----------
    Al--Mohy, A H and Higham, N J, 2011, `Improved inverse scaling and
    squaring algorithms for the matrix logarithm`, SIAM J. Sci. Comput.
    (34(4)), C152--C169

    Al--Mohy, A H, Higham, N J and Relton, S D, 2012, `Computing the
    Fréchet derivative of the matrix logarithm and estimating the
    condition number`, SIAM J. Sci. Comput. (35(4)), C394--C410

    Higham, N J, 2008, `Functions of Matrices: Theory and Computation`,
    SIAM, Philadelphia, PA, USA
    """

    fun_name = 'naginterfaces.library.matop.complex_gen_matrix_frcht_log'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _e = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'e'},
        e,
        copy=True,
    )
    try:
        _b_matop.complex_gen_matrix_frcht_log(
            _n,
            _a,
            _e,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _a_py = _a
    _e_py = _e
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'e'],
    )(
        _a_py,
        _e_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def real_gen_tridiag_lu(a, lamda, b, c, tol):
    u"""
    LU factorization of real tridiagonal matrix.

    ``real_gen_tridiag_lu`` computes an LU factorization of a real
    tridiagonal matrix, using Gaussian elimination with partial
    pivoting.

    For full information please refer to the NAG Library document for
    f01le

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01lef.html

    Parameters
    ----------
    a : float, array-like, shape (n)
        The diagonal elements of T.

    lamda : float
        The scalar lambda. ``real_gen_tridiag_lu`` factorizes T-lambda
        I.

    b : float, array-like, shape (n)
        The superdiagonal elements of T, stored in `b`[1] to `b`[n-1];
        `b`[0] is not used.

    c : float, array-like, shape (n)
        The subdiagonal elements of T, stored in `c`[1] to `c`[n-1];
        `c`[0] is not used.

    tol : float
        A relative tolerance used to indicate whether or not the matrix
        (T-lambda I) is nearly singular. `tol` should normally be chosen
        as approximately the largest relative error in the elements of
        T. For example, if the elements of T are correct to about 4
        significant figures, then `tol` should be set to about
        5*10^(-4). See [Further Comments] for further details on how
        `tol` is used. If `tol` is supplied as less than epsilon, where
        epsilon is the machine precision, then the value epsilon is used
        in place of `tol`.

    Returns
    -------
    a : float, ndarray, shape (n)
        The diagonal elements of the upper triangular matrix U.

    b : float, ndarray, shape (n)
        The elements of the first superdiagonal of U, stored in `b`[1]
        to `b`[n-1].

    c : float, ndarray, shape (n)
        The subdiagonal elements of L, stored in `c`[1] to `c`[n-1].

    d : float, ndarray, shape (n)
        The elements of the second superdiagonal of U, stored in `d`[2]
        to `d`[n-1]; `d`[0] and `d`[1] are not used.

    ipiv : int, ndarray, shape (n)
        Details of the permutation matrix P. If an interchange occurred
        at the kth step of the elimination, then `ipiv`[k-1] = 1,
        otherwise `ipiv`[k-1] = 0. If a diagonal element of U is small,
        indicating that (T-lambda I) is nearly singular, then the
        element `ipiv`[n-1] is returned as positive. Otherwise
        `ipiv`[n-1] is returned as 0. See [Further Comments] for further
        details. If the application is such that it is important that
        (T-lambda I) is not nearly singular, then it is strongly
        recommended that `ipiv`[n-1] is inspected on return.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 1.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Wilkinson, J H, 1965, `The Algebraic Eigenvalue Problem`, Oxford
    University Press, Oxford

    Wilkinson, J H and Reinsch, C, 1971, `Handbook for Automatic
    Computation II, Linear Algebra`, Springer--Verlag
    """

    fun_name = 'naginterfaces.library.matop.real_gen_tridiag_lu'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _c = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _d = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _ipiv = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n_py,),
    )
    _b_matop.real_gen_tridiag_lu(
        _n,
        _a,
        lamda,
        _b,
        _c,
        tol,
        _d,
        _ipiv,
    )
    _a_py = _a
    _b_py = _b
    _c_py = _c
    _d_py = _d
    _ipiv_py = _ipiv
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'b', 'c', 'd', 'ipiv'],
    )(
        _a_py,
        _b_py,
        _c_py,
        _d_py,
        _ipiv_py,
    )
    return _return_data

def real_gen_blkdiag_lu(n, blkstr, a, tol):
    u"""
    LU factorization of real almost block diagonal matrix.

    ``real_gen_blkdiag_lu`` factorizes a real almost block diagonal
    matrix.

    For full information please refer to the NAG Library document for
    f01lh

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01lhf.html

    Parameters
    ----------
    n : int
        n, the order of the matrix A.

    blkstr : int, array-like, shape (3, nbloks)
        Information which describes the block structure of A as follows:

            `blkstr`[0,k-1] must contain the number of rows in the kth
            block, k = 1,2,...,nbloks;

            `blkstr`[1,k-1] must contain the number of columns in the
            kth block, k = 1,2,...,nbloks;

            `blkstr`[2,k-1] must contain the number of columns of
            overlap between the kth and (k+1)th blocks, k =
            1,2,...,nbloks-1. `blkstr`[2,nbloks-1] need not be set.

        The following conditions delimit the structure of A:

            `blkstr`[0,k-1],`blkstr`[1,k-1] > 0, k = 1,2,...,nbloks,

            `blkstr`[2,k-1] >= 0, k = 1,2,...,nbloks-1,

        (there must be at least one column and one row in each block and
        a non-negative number of columns of overlap);

            `blkstr`[2,k-2]+`blkstr`[2,k-1] <= `blkstr`[1,k-1], k =
            2,3,...,nbloks-1,

        (the total number of columns in overlaps in each block must not
        exceed the number of columns in that block);

            `blkstr`[1,0] >= `blkstr`[0,0],

            `blkstr`[1,0]+sum_k = 2^j[`blkstr`[1,k -1]-`blkstr`[2,k -1
            -1]] >= sum_k = 1^j`blkstr`[0,k -1], j = 2,3,...,nbloks-1,

            sum_k = 1^j[`blkstr`[1,k-1]-`blkstr`[2,k-1]] <= sum_k =
            1^j`blkstr`[0,k-1], j = 1,2,...,nbloks-1,

        (the index of the first column of the overlap between the jth
        and (j+1)th blocks must be <= the index of the last row of the
        jth block, and the index of the last column of overlap must be
        >= the index of the last row of the jth block);

            sum_k = 1^nbloks`blkstr`[0,k -1] = n,

            `blkstr`[1,0]+sum_k = 2^nbloks[`blkstr`[1,k -1]-`blkstr`[2,k
            -1 -1]] = nk,

        (both the number of rows and the number of columns of A must
        equal n).

    a : float, array-like, shape (lena)
        The elements of the almost block diagonal matrix stored block by
        block, with each block stored column by column. The sizes of the
        blocks and the overlaps are defined by the argument `blkstr`.

        If a_{rs} is the first element in the kth block, then an
        arbitrary element a_{ij} in the kth block must be stored in the
        array element:

            `a`[ p_k + (j-r) m_k + (i-s)+1 -1]

        where

            p_k = sum_l = 1^k-1`blkstr`[0,l-1]*`blkstr`[1,l-1]

        is the base address of the kth block, and

            m_k = `blkstr`[0,k-1]

        is the number of rows of the kth block.

        See [Further Comments] for comments on scaling.

    tol : float
        A relative tolerance to be used to indicate whether or not the
        matrix is singular. For a discussion on how `tol` is used see
        [Further Comments]. If `tol` is non-positive, `tol` is reset to
        10 epsilon, where epsilon is the machine precision.

    Returns
    -------
    a : float, ndarray, shape (lena)
        The factorized form of the matrix.

    pivot : int, ndarray, shape (`n`)
        Details of the interchanges.

    tol : float
        Unchanged unless `tol` <= 0.0 on entry, in which case it is set
        to 10*epsilon.

    kpivot : int
        If `errno` = 2, `kpivot` contains the value k, where k is the
        first position on the diagonal of the matrix A where too small a
        pivot was detected. Otherwise `kpivot` is set to 0.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, lena is too small. lena = *<value>*. Minimum
            possible dimension: *<value>*.

        (`errno` 1)
            On entry, the following equality does not hold:
            `blkstr`[1,0]+sum(`blkstr`[1,k-1]-`blkstr`[2,k-2]:k = 2,
            nbloks) = `n`.

        (`errno` 1)
            On entry, the following equality does not hold:
            sum(`blkstr`[0,k-1]:k = 1,nbloks) = `n`.

        (`errno` 1)
            On entry, K = *<value>* `blkstr`[2,K-1] = *<value>*
            `blkstr`[2,K-2] = *<value>* and `blkstr`[1,K-1] = *<value>*.

            Constraint: `blkstr`[2,K-1]+`blkstr`[2,K-2] >=
            `blkstr`[1,K-1].

        (`errno` 1)
            On entry, the following inequality was not satisfied for: J
            = *<value>*. sum(`blkstr`[1,k-1]-`blkstr`[2,k-1]:k = 1, J)
            <= sum(`blkstr`[0,k-1]:k = 1, J) <=
            `blkstr`[1,0]+sum(`blkstr`[1,k-1]-`blkstr`[2,k-2]:k = 2, J).

        (`errno` 1)
            On entry, K = *<value>*, `blkstr`[1,K-1] = *<value>*
            `blkstr`[2,K-1] = *<value>* and `blkstr`[0,K-1] = *<value>*.

            Constraint: `blkstr`[1,K-1]-`blkstr`[2,K-1] <=
            `blkstr`[0,K-1].

        (`errno` 1)
            On entry, K = *<value>*, `blkstr`[1,K-1] = *<value>* and
            `blkstr`[0,K-1] = *<value>*.

            Constraint: `blkstr`[1,K-1] >= `blkstr`[0,K-1].

        (`errno` 1)
            On entry, K = *<value>* and `blkstr`[2,K-1] = *<value>*.

            Constraint: `blkstr`[2,K-1] >= 0.

        (`errno` 1)
            On entry, K = *<value>* and `blkstr`[1,K-1] = *<value>*.

            Constraint: `blkstr`[1,K-1] >= 1.

        (`errno` 1)
            On entry, K = *<value>* and `blkstr`[0,K-1] = *<value>*.

            Constraint: `blkstr`[0,K-1] >= 1.

        (`errno` 1)
            On entry, `n` = *<value>* and nbloks = *<value>*.

            Constraint: `n` >= nbloks.

        (`errno` 1)
            On entry, nbloks = *<value>*.

            Constraint: nbloks >= 1.

        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            Factorization completed, but pivot in diagonal I was small:
            I = *<value>*.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Diaz, J C, Fairweather, G and Keast, P, 1983, `Fortran packages for
    solving certain almost block diagonal linear systems by modified
    alternate row and column elimination`, ACM Trans. Math. Software
    (9), 358--375
    """

    fun_name = 'naginterfaces.library.matop.real_gen_blkdiag_lu'

    _blkstr = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'blkstr'},
        blkstr,
        copy=False,
    )
    _nbloks_py = _b_utils._EngineIntArrayType.get_shape(_blkstr, {'fun_name': fun_name, 'entity': 'blkstr'}, exp_rank=2)[1]
    _nbloks = _b_utils._EngineIntScalarType(
        _nbloks_py, {'fun_name': fun_name, 'entity': '_nbloks'},
    ).to_ctypes()
    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lena_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=1)[0]
    _lena = _b_utils._EngineIntScalarType(
        _lena_py, {'fun_name': fun_name, 'entity': '_lena'},
    ).to_ctypes()
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _pivot = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _return_scalars = _b_matop.real_gen_blkdiag_lu(
            _n,
            _nbloks,
            _blkstr,
            _a,
            _lena,
            _pivot,
            tol,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _a_py = _a
    _tol_py = _return_scalars.tol
    _pivot_py = _pivot
    _kpivot_py = _return_scalars.kpivot
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'pivot', 'tol', 'kpivot'],
    )(
        _a_py,
        _pivot_py,
        _tol_py,
        _kpivot_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def real_vband_posdef_fac(a, nrow):
    u"""
    LDL^T factorization of real symmetric positive definite
    variable-bandwidth matrix.

    ``real_vband_posdef_fac`` computes the Cholesky factorization of a
    real symmetric positive definite variable-bandwidth matrix.

    For full information please refer to the NAG Library document for
    f01mc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01mcf.html

    Parameters
    ----------
    a : float, array-like, shape (lal)
        The elements within the envelope of the lower triangle of the
        positive definite symmetric matrix A, taken in row by row order.
        The matrix elements within the band can be assigned to the
        correct elements of the array using the following logic:

            set k = 0

            loop for i = 1 to n

                loop for j = i-`nrow`[i-1]+1 to i

                    set k = k+1

                    set `a`[k-1] = A_{ij}

        where A_{ij} is the element in the ith row and jth column of the
        matrix A.

        See also [Further Comments].

    nrow : int, array-like, shape (n)
        `nrow`[i-1] must contain the width of row i of the matrix A,
        i.e., the number of elements between the first (leftmost)
        nonzero element and the element on the diagonal, inclusive.

    Returns
    -------
    al : float, ndarray, shape (lal)
        The elements within the envelope of the lower triangular matrix
        L, taken in row by row order. The envelope of L is identical to
        that of the lower triangle of A. The unit diagonal elements of L
        are stored explicitly. See also [Further Comments].

    d : float, ndarray, shape (n)
        The diagonal elements of the diagonal matrix D. Note that the
        determinant of A is equal to the product of these diagonal
        elements. If the value of the determinant is required it should
        not be determined by forming the product explicitly, because of
        the possibility of overflow or underflow. The logarithm of the
        determinant may safely be formed from the sum of the logarithms
        of the diagonal elements.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, lal = *<value>* and `nrow`[0]+...+`nrow`[n-1] =
            *<value>*.

            Constraint: lal >= `nrow`[0]+...+`nrow`[n-1].

        (`errno` 1)
            On entry, I = *<value>* and `nrow`[I-1] = *<value>*.

            Constraint: `nrow`[I-1] >= 1 and `nrow`[I-1] <= I.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 2)
            A is not positive definite. Factorization abandoned.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            A is not positive definite. Factorization completed.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Jennings, A, 1966, `A compact storage scheme for the solution of
    symmetric linear simultaneous equations`, Comput. J. (9), 281--285

    Wilkinson, J H and Reinsch, C, 1971, `Handbook for Automatic
    Computation II, Linear Algebra`, Springer--Verlag
    """

    fun_name = 'naginterfaces.library.matop.real_vband_posdef_fac'

    _nrow = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'nrow'},
        nrow,
        copy=True,
    )
    _n_py = _b_utils._EngineIntArrayType.get_shape(_nrow, {'fun_name': fun_name, 'entity': 'nrow'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lal_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=1)[0]
    _lal = _b_utils._EngineIntScalarType(
        _lal_py, {'fun_name': fun_name, 'entity': '_lal'},
    ).to_ctypes()
    _al = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_lal_py,),
    )
    _d = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _b_matop.real_vband_posdef_fac(
            _n,
            _a,
            _lal,
            _nrow,
            _al,
            _d,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _al_py = _al
    _d_py = _d
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['al', 'd'],
    )(
        _al_py,
        _d_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def real_trapez_rq(m, a):
    u"""
    RQ factorization of real m by n upper trapezoidal matrix (m <= n).

    ``real_trapez_rq`` reduces the m by n (m <= n) real upper
    trapezoidal matrix A to upper triangular form by means of orthogonal
    transformations.

    For full information please refer to the NAG Library document for
    f01qg

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01qgf.html

    Parameters
    ----------
    m : int
        m, the number of rows of the matrix A.

        When `m` = 0 then an immediate return is effected.

    a : float, array-like, shape (`m`, n)
        The leading m by n upper trapezoidal part of the array `a` must
        contain the matrix to be factorized.

    Returns
    -------
    a : float, ndarray, shape (`m`, n)
        The m by m upper triangular part of `a` will contain the upper
        triangular matrix R, and the m by (n-m) upper trapezoidal part
        of `a` will contain details of the factorization as described in
        [Description].

    zeta : float, ndarray, shape (`m`)
        `zeta`[k-1] contains the scalar zeta_k for the (m-k+1)th
        transformation. If T_k = I then `zeta`[k-1] = 0.0, otherwise
        `zeta`[k-1] contains zeta_k as described in [Description] and
        zeta_k is always in the range (1.0,sqrt(2.0)).

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, n = *<value>* and `m` = *<value>*.

            Constraint: n >= `m`.

        (`errno` -1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Wilkinson, J H, 1965, `The Algebraic Eigenvalue Problem`, Oxford
    University Press, Oxford
    """

    fun_name = 'naginterfaces.library.matop.real_trapez_rq'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _m = _b_utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = _b_utils._EngineIntScalarType.to_py(_m)
    _zeta = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_m_py,),
    )
    _b_matop.real_trapez_rq(
        _m,
        _n,
        _a,
        _zeta,
    )
    _a_py = _a
    _zeta_py = _zeta
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'zeta'],
    )(
        _a_py,
        _zeta_py,
    )
    return _return_data

def real_gen_rq(m, a):
    u"""
    RQ factorization of real m by n matrix (m <= n).

    ``real_gen_rq`` finds the RQ factorization of the real m by n (m <=
    n) matrix A, so that A is reduced to upper triangular form by means
    of orthogonal transformations from the right.

    For full information please refer to the NAG Library document for
    f01qj

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01qjf.html

    Parameters
    ----------
    m : int
        m, the number of rows of the matrix A.

        When `m` = 0 then an immediate return is effected.

    a : float, array-like, shape (`m`, n)
        The leading m by n part of the array `a` must contain the matrix
        to be factorized.

    Returns
    -------
    a : float, ndarray, shape (`m`, n)
        The m by m upper triangular part of `a` will contain the upper
        triangular matrix R, and the m by m strictly lower triangular
        part of `a` and the m by (n-m) rectangular part of `a` to the
        right of the upper triangular part will contain details of the
        factorization as described in [Description].

    zeta : float, ndarray, shape (`m`)
        `zeta`[k-1] contains the scalar zeta_k for the (m-k+1)th
        transformation. If P_k = I then `zeta`[k-1] = 0.0, otherwise
        `zeta`[k-1] contains zeta_k as described in [Description] and
        zeta_k is always in the range (1.0,sqrt(2.0)).

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, n = *<value>* and `m` = *<value>*.

            Constraint: n >= `m`.

        (`errno` -1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Wilkinson, J H, 1965, `The Algebraic Eigenvalue Problem`, Oxford
    University Press, Oxford
    """

    fun_name = 'naginterfaces.library.matop.real_gen_rq'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _m = _b_utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = _b_utils._EngineIntScalarType.to_py(_m)
    _zeta = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_m_py,),
    )
    _b_matop.real_gen_rq(
        _m,
        _n,
        _a,
        _zeta,
    )
    _a_py = _a
    _zeta_py = _zeta
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'zeta'],
    )(
        _a_py,
        _zeta_py,
    )
    return _return_data

def real_gen_rq_formq(wheret, m, nrowp, a, zeta):
    u"""
    Operations with orthogonal matrices, form rows of Q, after RQ
    factorization by ``real_gen_rq``.

    ``real_gen_rq_formq`` returns the first l rows of the real n by n
    orthogonal matrix P^T, where P is given as the product of
    Householder transformation matrices.

    This function is intended for use following ``real_gen_rq``.

    For full information please refer to the NAG Library document for
    f01qk

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01qkf.html

    Parameters
    ----------
    wheret : str, length 1
        Indicates where the elements of zeta are to be found.

        `wheret` = 'I' (In `a`)

            The elements of zeta are in `a`.

        `wheret` = 'S' (Separate)

            The elements of zeta are separate from `a`, in `zeta`.

    m : int
        m, the number of rows of the matrix A.

    nrowp : int
        l, the required number of rows of P.

        If `nrowp` = 0, an immediate return is effected.

    a : float, array-like, shape (max(`m`, `nrowp`), n)
        The leading m by m strictly lower triangular part of the array
        `a`, and the m by (n-m) rectangular part of `a` with top
        left-hand corner at element `a`[0,`m`] must contain details of
        the matrix P. In addition, if `wheret` = 'I', the diagonal
        elements of `a` must contain the elements of zeta.

    zeta : float, array-like, shape (:)
        Note: the required length for this argument is determined as
        follows: if `wheret`='S': `m`; otherwise: 1.

        With `wheret` = 'S', the array `zeta` must contain the elements
        of zeta. If `zeta`[k-1] = 0.0 then P_k is assumed to be I,
        otherwise `zeta`[k-1] is assumed to contain zeta_k.

        When `wheret` = 'I', the array `zeta` is not referenced.

    Returns
    -------
    a : float, ndarray, shape (max(`m`, `nrowp`), n)
        The first `nrowp` rows of the array `a` are overwritten by the
        first `nrowp` rows of the n by n orthogonal matrix P^T.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, lda = *<value>*, `m` = *<value>* and `nrowp` =
            *<value>*.

            Constraint: lda >= max(`m`, `nrowp`).

        (`errno` -1)
            On entry, `nrowp` = *<value>* and n = *<value>*.

            Constraint: `nrowp` >= 0 and `nrowp` <= n.

        (`errno` -1)
            On entry, n = *<value>* and `m` = *<value>*.

            Constraint: n >= `m`.

        (`errno` -1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 0.

        (`errno` -1)
            On entry, `wheret` = *<value>*.

            Constraint: `wheret` = 'I' or 'S'.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Wilkinson, J H, 1965, `The Algebraic Eigenvalue Problem`, Oxford
    University Press, Oxford
    """

    fun_name = 'naginterfaces.library.matop.real_gen_rq_formq'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _zeta = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'zeta'},
        zeta,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b_matop.real_gen_rq_formq(
        wheret,
        m,
        _n,
        nrowp,
        _a,
        _zeta,
    )
    _a_py = _a
    return _a_py

def complex_trapez_rq(m, a):
    u"""
    RQ factorization of complex m by n upper trapezoidal matrix (m <=
    n).

    ``complex_trapez_rq`` reduces the complex m by n (m <= n) upper
    trapezoidal matrix A to upper triangular form by means of unitary
    transformations.

    For full information please refer to the NAG Library document for
    f01rg

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01rgf.html

    Parameters
    ----------
    m : int
        m, the number of rows of the matrix A.

        When `m` = 0 then an immediate return is effected.

    a : complex, array-like, shape (`m`, n)
        The leading m by n upper trapezoidal part of the array `a` must
        contain the matrix to be factorized.

    Returns
    -------
    a : complex, ndarray, shape (`m`, n)
        The m by m upper triangular part of `a` will contain the upper
        triangular matrix R, and the m by (n-m) upper trapezoidal part
        of `a` will contain details of the factorization as described in
        [Description].

    theta : complex, ndarray, shape (`m`)
        `theta`[k-1] contains the scalar theta_k for the (m-k+1)th
        transformation. If T_k = I then `theta`[k-1] = 0.0; if

            T_k = [table omitted], Re(alpha) < 0.0

        then `theta`[k-1] = alpha, otherwise `theta`[k-1] contains
        theta_k as described in [Description] and Re(theta_k) is always
        in the range (1.0,sqrt(2.0)).

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, n = *<value>* and `m` = *<value>*.

            Constraint: n >= `m`.

        (`errno` -1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Wilkinson, J H, 1965, `The Algebraic Eigenvalue Problem`, Oxford
    University Press, Oxford
    """

    fun_name = 'naginterfaces.library.matop.complex_trapez_rq'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _m = _b_utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = _b_utils._EngineIntScalarType.to_py(_m)
    _theta = _b_utils._EngineComplex128ArrayType.empty_ndarray(
        (_m_py,),
    )
    _b_matop.complex_trapez_rq(
        _m,
        _n,
        _a,
        _theta,
    )
    _a_py = _a
    _theta_py = _theta
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'theta'],
    )(
        _a_py,
        _theta_py,
    )
    return _return_data

def complex_gen_rq(m, a):
    u"""
    RQ factorization of complex m by n matrix (m <= n).

    ``complex_gen_rq`` finds the RQ factorization of the complex m by n
    (m <= n), matrix A, so that A is reduced to upper triangular form by
    means of unitary transformations from the right.

    For full information please refer to the NAG Library document for
    f01rj

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01rjf.html

    Parameters
    ----------
    m : int
        m, the number of rows of the matrix A.

        When `m` = 0 then an immediate return is effected.

    a : complex, array-like, shape (`m`, n)
        The leading m by n part of the array `a` must contain the matrix
        to be factorized.

    Returns
    -------
    a : complex, ndarray, shape (`m`, n)
        The m by m upper triangular part of `a` will contain the upper
        triangular matrix R, and the m by m strictly lower triangular
        part of `a` and the m by (n-m) rectangular part of `a` to the
        right of the upper triangular part will contain details of the
        factorization as described in [Description].

    theta : complex, ndarray, shape (`m`)
        `theta`[k-1] contains the scalar theta_k for the (m-k+1)th
        transformation. If P_k = I then `theta`[k-1] = 0.0; if

            T_k = [table omitted], Re(alpha) < 0.0

        then `theta`[k-1] = alpha, otherwise `theta`[k-1] contains
        theta_k as described in [Description] and theta_k is always in
        the range (1.0,sqrt(2.0)).

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, n = *<value>* and `m` = *<value>*.

            Constraint: n >= `m`.

        (`errno` -1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Wilkinson, J H, 1965, `The Algebraic Eigenvalue Problem`, Oxford
    University Press, Oxford
    """

    fun_name = 'naginterfaces.library.matop.complex_gen_rq'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _m = _b_utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = _b_utils._EngineIntScalarType.to_py(_m)
    _theta = _b_utils._EngineComplex128ArrayType.empty_ndarray(
        (_m_py,),
    )
    _b_matop.complex_gen_rq(
        _m,
        _n,
        _a,
        _theta,
    )
    _a_py = _a
    _theta_py = _theta
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'theta'],
    )(
        _a_py,
        _theta_py,
    )
    return _return_data

def complex_gen_rq_formq(wheret, m, nrowp, a, theta):
    u"""
    Operations with unitary matrices, form rows of Q, after RQ
    factorization by ``complex_gen_rq``.

    ``complex_gen_rq_formq`` returns the first l rows of the n by n
    unitary matrix P^H, where P is given as the product of Householder
    transformation matrices.

    This function is intended for use following ``complex_gen_rq``.

    For full information please refer to the NAG Library document for
    f01rk

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01rkf.html

    Parameters
    ----------
    wheret : str, length 1
        Indicates where the elements of theta are to be found.

        `wheret` = 'I' (In `a`)

            The elements of theta are in `a`.

        `wheret` = 'S' (Separate)

            The elements of theta are separate from `a`, in `theta`.

    m : int
        m, the number of rows of the matrix A.

    nrowp : int
        l, the required number of rows of P.

        If `nrowp` = 0, an immediate return is effected.

    a : complex, array-like, shape (max(`m`, `nrowp`), n)
        The leading m by m strictly lower triangular part of the array
        `a`, and the m by (n-m) rectangular part of `a` with top
        left-hand corner at element `a`[0,`m`] must contain details of
        the matrix P. In addition, if `wheret` = 'I', the diagonal
        elements of `a` must contain the elements of theta.

    theta : complex, array-like, shape (:)
        Note: the required length for this argument is determined as
        follows: if `wheret`='S': `m`; otherwise: 1.

        If `wheret` = 'S', the array `theta` must contain the elements
        of theta. If `theta`[k-1] = 0.0, P_k is assumed to be I, if
        `theta`[k-1] = alpha and Re(alpha) < 0.0, P_k is assumed to be
        of the form

            P_k = [table omitted],

        otherwise `theta`[k-1] is assumed to contain theta_k given by

            theta_k = (zeta_k,Im(gamma_k)).

        If `wheret` = 'I', the array `theta` is not referenced.

    Returns
    -------
    a : complex, ndarray, shape (max(`m`, `nrowp`), n)
        The first `nrowp` rows of the array `a` are overwritten by the
        first `nrowp` rows of the n by n unitary matrix P^H.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, lda = *<value>*, `m` = *<value>* and `nrowp` =
            *<value>*.

            Constraint: lda >= max(`m`, `nrowp`).

        (`errno` -1)
            On entry, `nrowp` = *<value>* and n = *<value>*.

            Constraint: `nrowp` >= 0 and `nrowp` <= n.

        (`errno` -1)
            On entry, n = *<value>* and `m` = *<value>*.

            Constraint: n >= `m`.

        (`errno` -1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 0.

        (`errno` -1)
            On entry, `wheret` = *<value>*.

            Constraint: `wheret` = 'I' or 'S'.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Golub, G H and Van Loan, C F, 1996, `Matrix Computations`, (3rd
    Edition), Johns Hopkins University Press, Baltimore

    Wilkinson, J H, 1965, `The Algebraic Eigenvalue Problem`, Oxford
    University Press, Oxford
    """

    fun_name = 'naginterfaces.library.matop.complex_gen_rq_formq'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _theta = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'theta'},
        theta,
        copy=False,
    )
    _b_matop.complex_gen_rq_formq(
        wheret,
        m,
        _n,
        nrowp,
        _a,
        _theta,
    )
    _a_py = _a
    return _a_py

def real_nmf(a, k=None, w=None, h=None, seed=None, errtol=0.0, maxit=0):
    u"""
    Non-negative matrix factorization of real non-negative matrix.

    ``real_nmf`` computes a non-negative matrix factorization for a real
    non-negative m by n matrix A.

    For full information please refer to the NAG Library document for
    f01sa

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01saf.html

    Parameters
    ----------
    a : float, array-like, shape (m, n)
        The m by n non-negative matrix A.

    k : None or int, optional
        Note: if this argument is None then a default value will be
        used, determined as follows: dimension 1 of `h`.

        k, the factorization rank.

        This argument is only referenced if you are not supplying
        initial iterates W and H.

        If you are providing initial iterates W and H then your supplied
        value of k will be ignored and it will be inferred from those
        arrays instead.

    w : None or float, array-like, shape (m, `k`), optional
            If `seed` <= 0, `w` should be set to an initial iterate for
            the non-negative matrix factor, W.

            If `seed` >= 1, `w` need not be set. ``real_nmf`` will
            generate a random initial iterate.

    h : None or float, array-like, shape (`k`, n), optional
            If `seed` <= 0, `h` should be set to an initial iterate for
            the non-negative matrix factor, H.

            If `seed` >= 1, `h` need not be set. ``real_nmf`` will
            generate a random initial iterate.

    seed : None or int, optional
        Note: if this argument is None then a default value will be
        used, determined as follows: if `w` is not None: 0; otherwise:
        1.

            If `seed` <= 0, the supplied values of W and H are used for
            the initial iterate.

            If `seed` >= 1, the value of `seed` is used to seed a random
            number generator for the initial iterates W and H. See
            [Generating Random Initial Iterates] for further details.

    errtol : float, optional
        The convergence tolerance for when the Hierarchical Alternating
        Least Squares iteration has reached a stationary point. If
        `errtol` <= 0.0, max(m, n)*sqrt(machine precision) is used.

    maxit : int, optional
        Specifies the maximum number of iterations to be used. If
        `maxit` <= 0, 200 is used.

    Returns
    -------
    w : float, ndarray, shape (m, `k`)
        The non-negative matrix factor, W.

    h : float, ndarray, shape (`k`, n)
        The non-negative matrix factor, H.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, m = *<value>*.

            Constraint: m >= 2.

        (`errno` 2)
            On entry, n = *<value>*.

            Constraint: n >= 2.

        (`errno` 3)
            On entry, `k` = *<value>*, m = *<value>* and n = *<value>*.

            Constraint: 1 <= `k` < min(m, n).

        (`errno` 8)
            An internal error occurred when generating initial values
            for `w` and `h`. Please contact NAG.

        (`errno` 9)
            On entry, one of more of the elements of `a`, `w` or `h`
            were negative.

        (`errno` 14)
            Both or neither of `w` and `h` must be provided.

        (`errno` 15)
            `h` or `k` must be provided.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 7)
            The function has failed to converge after *<value>*
            iterations. The factorization given by `w` and `h` may still
            be a good enough approximation to be useful. Alternatively
            an improved factorization may be obtained by increasing
            `maxit` or using different initial choices of `w` and `h`.

    References
    ----------
    Cichocki, A and Phan, A--H, 2009, `Fast local algorithms for large
    scale nonnegative matrix and tensor factorizations`, IEICE
    Transactions on Fundamentals of Electronics, Communications and
    Computer Sciences (E92--A), 708--721

    Cichocki, A, Zdunek, R and Amari, S--I, 2007, `Hierarchical ALS
    algorithms for nonnegative matrix and 3D tensor factorization`,
    Lecture Notes in Computer Science (4666), Springer, 169--176

    Ho, N--D, 2008, `Nonnegative matrix factorization algorithms and
    applications`, PhD Thesis, Univ. Catholique de Louvain
    """

    fun_name = 'naginterfaces.library.matop.real_nmf'

    if (w is None) and h is not None or w is not None and h is None:
        _b_utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                'Both or neither of w and h must be provided.',
            ],
            errcodes=(14,),
        ).manage()
    if h is None and k is None:
        _b_utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                'h or k must be provided.',
            ],
            errcodes=(15,),
        ).manage()
    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _w = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'w'},
        w,
        copy=True,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _h = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'h'},
        h,
        copy=True,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_a, 'a'), (_w, 'w'), (_h, 'h')),
    )
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _k = _b_utils._EngineFloat64ArrayType.get_shape(_h, {'fun_name': fun_name, 'entity': 'h'}, exp_rank=2, allow_none=True)[0] if k is None else k
    if w is not None:
        _seed = 0
    else:
        _seed = 1
    _seed = _seed if seed is None else seed
    _seed = _b_utils._EngineIntScalarType(
        _seed, {'fun_name': fun_name, 'entity': '_seed'},
    ).to_ctypes()
    _seed_py = _b_utils._EngineIntScalarType.to_py(_seed)
    if _seed_py <= 0:
        if w is None:
            raise _b_utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'w'},
                'must not be None when seed <= 0',
            )
    _k = _b_utils._EngineIntScalarType(
        _k, {'fun_name': fun_name, 'entity': '_k'},
    ).to_ctypes()
    _k_py = _b_utils._EngineIntScalarType.to_py(_k)
    if w is None:
        _w = _b_utils._EngineFloat64ArrayType.empty_ndarray(
            (_m_py, _k_py),
            sorder=_m_sorder,
        )
    if _seed_py <= 0:
        if h is None:
            raise _b_utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'h'},
                'must not be None when seed <= 0',
            )
    if h is None:
        _h = _b_utils._EngineFloat64ArrayType.empty_ndarray(
            (_k_py, _n_py),
            sorder=_m_sorder,
        )
    try:
        _b_matop.real_nmf(
            _m,
            _n,
            _k,
            _a,
            _w,
            _h,
            _seed,
            errtol,
            maxit,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _w_py = _w
    _h_py = _h
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['w', 'h'],
    )(
        _w_py,
        _h_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def real_nmf_rcomm(irevcm, w, h, ht, comm, seed=1, errtol=0.0):
    u"""
    Non-negative matrix factorization of real non-negative matrix
    (reverse communication).

    ``real_nmf_rcomm`` computes a non-negative matrix factorization for
    a real non-negative m by n matrix A.
    It uses reverse communication for evaluating matrix products, so
    that the matrix A is not accessed explicitly.

    For full information please refer to the NAG Library document for
    f01sb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01sbf.html

    Parameters
    ----------
    irevcm : int
        `On initial entry`: must be set to 0.

    w : float, ndarray, shape (m, k), modified in place
        `On initial entry`:

            if `seed` <= 0, `w` should be set to an initial iterate for
            the non-negative matrix factor, W.

            If `seed` >= 1, `w` need not be set. ``real_nmf_rcomm`` will
            generate a random initial iterate.

        `On intermediate exit`: if `irevcm` = 1 or 2, `w` contains the
        current iterate of the m*k non-negative matrix W.

        `On intermediate entry`:

            if `irevcm` = 3, `w` must contain AH^T, where H^T is stored
            in ht.

            If `irevcm` = 0, 1 or 2, `w` must not be changed.

        `On final exit`: `w` contains the m*k non-negative matrix W.

    h : float, ndarray, shape (k, n), modified in place
        `On initial entry`:

            if `seed` <= 0, `h` should be set to an initial iterate for
            the non-negative matrix factor, H.

            If `seed` >= 1, `h` need not be set. ``real_nmf_rcomm`` will
            generate a random initial iterate.

        `On intermediate exit`: if `irevcm` = 1, `h` contains the
        current iterate of the k*n non-negative matrix H.

        `On intermediate entry`: `h` must not be changed.

        `On final exit`: `h` contains the k*n non-negative matrix H.

    ht : float, ndarray, shape (n, k), modified in place
        `On initial entry`: `ht` need not be set.

        `On intermediate exit`: if `irevcm` = 3, `ht` contains the n*k
        non-negative matrix H^T, which is required in order to form
        AH^T.

        `On intermediate entry`: if `irevcm` = 2, `ht` must contain
        A^TW.

        If `irevcm` = 0, 1 or 3, `ht` must not be changed.

        `On final exit`: `ht` is undefined.

    comm : dict, communication object, modified in place
        Communication structure.

        `On initial entry`: need not be set.

    seed : int, optional
        `On initial entry`:

            if `seed` <= 0, the supplied values of W and H are used for
            the initial iterate.

            If `seed` >= 1, the value of `seed` is used to seed a random
            number generator for the initial iterates W and H. See
            [Generating Random Initial Iterates] for further details.

    errtol : float, optional
        The convergence tolerance for when the Hierarchical Alternating
        Least Squares iteration has reached a stationary point. If
        `errtol` <= 0.0, max(m, n)*sqrt(machine precision) is used.

    Returns
    -------
    irevcm : int
        `On intermediate exit`: specifies what action you must take
        before re-entering ``real_nmf_rcomm`` with `irevcm` unchanged.
        The value of `irevcm` should be interpreted as follows:

        `irevcm` = 1

            Indicates the start of a new iteration. No action is
            required by you, but `w` and `h` are available for printing,
            and a limit on the number of iterations can be applied.

        `irevcm` = 2

            Indicates that before re-entry to ``real_nmf_rcomm``, the
            product A^TW must be computed and stored in `ht`.

        `irevcm` = 3

            Indicates that before re-entry to ``real_nmf_rcomm``, the
            product AH^T must be computed and stored in `w`.

        `On final exit`: `irevcm` = 0.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On intermediate re-entry, `irevcm` = *<value>*.

            Constraint: `irevcm` = 1, 2 or 3.

        (`errno` 1)
            On initial entry, `irevcm` = *<value>*.

            Constraint: `irevcm` = 0.

        (`errno` 2)
            On entry, m = *<value>*.

            Constraint: m >= 2.

        (`errno` 3)
            On entry, n = *<value>*.

            Constraint: n >= 2.

        (`errno` 4)
            On entry, k = *<value>*, m = *<value>* and n = *<value>*.

            Constraint: 1 <= k < min(m, n).

        (`errno` 8)
            An internal error occurred when generating initial values
            for `w` and `h`. Please contact NAG.

        (`errno` 9)
            On entry, one of more of the elements of `w` or `h` were
            negative.

    References
    ----------
    Cichocki, A and Phan, A--H, 2009, `Fast local algorithms for large
    scale nonnegative matrix and tensor factorizations`, IEICE
    Transactions on Fundamentals of Electronics, Communications and
    Computer Sciences (E92--A), 708--721

    Cichocki, A, Zdunek, R and Amari, S--I, 2007, `Hierarchical ALS
    algorithms for nonnegative matrix and 3D tensor factorization`,
    Lecture Notes in Computer Science (4666), Springer, 169--176

    Ho, N--D, 2008, `Nonnegative matrix factorization algorithms and
    applications`, PhD Thesis, Univ. Catholique de Louvain
    """

    fun_name = 'naginterfaces.library.matop.real_nmf_rcomm'

    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'w'},
        w,
    )
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(w, {'fun_name': fun_name, 'entity': 'w'}, exp_rank=2)[0]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'h'},
        h,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(h, {'fun_name': fun_name, 'entity': 'h'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _k_py = _b_utils._EngineFloat64ArrayType.get_shape(w, {'fun_name': fun_name, 'entity': 'w'}, exp_rank=2)[1]
    _k = _b_utils._EngineIntScalarType(
        _k_py, {'fun_name': fun_name, 'entity': '_k'},
    ).to_ctypes()
    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'ht'},
        ht,
    )
    _return_scalar = _b_matop.real_nmf_rcomm(
        irevcm,
        _m,
        _n,
        _k,
        w,
        h,
        ht,
        seed,
        errtol,
        comm,
    )
    _irevcm_py = _return_scalar
    return _irevcm_py

def dtrttp(uplo, a):
    u"""
    Copies a real triangular matrix from full format to packed format.

    ``dtrttp`` copies a real triangular matrix, stored in a full format
    array, to a standard packed format array.

    For full information please refer to the NAG Library document for
    f01va

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01vaf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

            A is upper triangular.

        `uplo` = 'L'

            A is lower triangular.

    a : float, array-like, shape (n, n)
        The triangular matrix A.

    Returns
    -------
    ap : float, ndarray, shape (n*(n+1)/2)
        The n by n triangular matrix A, packed by columns.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.
    """

    fun_name = 'naginterfaces.library.matop.dtrttp'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _ap = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py*(_n_py+1)//2,),
    )
    _b_matop.dtrttp(
        uplo,
        _n,
        _a,
        _ap,
    )
    _ap_py = _ap
    return _ap_py

def ztrttp(uplo, a):
    u"""
    Copies a complex triangular matrix from full format to packed
    format.

    ``ztrttp`` copies a complex triangular matrix, stored in a full
    format array, to a packed format array.

    For full information please refer to the NAG Library document for
    f01vb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01vbf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

            A is upper triangular.

        `uplo` = 'L'

            A is lower triangular.

    a : complex, array-like, shape (n, n)
        The triangular matrix A.

    Returns
    -------
    ap : complex, ndarray, shape (n*(n+1)/2)
        The n by n triangular matrix A, packed by columns.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter n.

            Constraint: n >= 0.
    """

    fun_name = 'naginterfaces.library.matop.ztrttp'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _ap = _b_utils._EngineComplex128ArrayType.empty_ndarray(
        (_n_py*(_n_py+1)//2,),
    )
    _b_matop.ztrttp(
        uplo,
        _n,
        _a,
        _ap,
    )
    _ap_py = _ap
    return _ap_py

def dtpttr(uplo, n, ap):
    u"""
    Copies a real triangular matrix from packed format to full format.

    ``dtpttr`` unpacks a real triangular matrix, stored in a standard
    packed format array, to a full format array.

    For full information please refer to the NAG Library document for
    f01vc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01vcf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

            A is upper triangular.

        `uplo` = 'L'

            A is lower triangular.

    n : int
        n, the order of the matrix A.

    ap : float, array-like, shape (`n`*(`n`+1)/2)
        The n by n triangular matrix A, packed by columns.

    Returns
    -------
    a : float, ndarray, shape (`n`, `n`)
        The triangular matrix A.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.
    """

    fun_name = 'naginterfaces.library.matop.dtpttr'

    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _a = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py, _n_py),
        sorder=_b_utils._EngineData.default_sorder,
    )
    _b_matop.dtpttr(
        uplo,
        _n,
        _ap,
        _a,
    )
    _a_py = _a
    return _a_py

def ztpttr(uplo, n, ap):
    u"""
    Copies a complex triangular matrix from packed format to full
    format.

    ``ztpttr`` unpacks a complex triangular matrix, stored in a standard
    packed format array, to a full format array.

    For full information please refer to the NAG Library document for
    f01vd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01vdf.html

    Parameters
    ----------
    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

            A is upper triangular.

        `uplo` = 'L'

            A is lower triangular.

    n : int
        n, the order of the matrix A.

    ap : complex, array-like, shape (`n`*(`n`+1)/2)
        The n by n triangular matrix A, packed by columns.

    Returns
    -------
    a : complex, ndarray, shape (`n`, `n`)
        The triangular matrix A.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -2)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.
    """

    fun_name = 'naginterfaces.library.matop.ztpttr'

    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=False,
    )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _a = _b_utils._EngineComplex128ArrayType.empty_ndarray(
        (_n_py, _n_py),
        sorder=_b_utils._EngineData.default_sorder,
    )
    _b_matop.ztpttr(
        uplo,
        _n,
        _ap,
        _a,
    )
    _a_py = _a
    return _a_py

def dtrttf(transr, uplo, a):
    u"""
    Copies a real triangular matrix from full format to Rectangular Full
    Packed format.

    ``dtrttf`` copies a real triangular matrix, stored in a full format
    array, to a Rectangular Full Packed (RFP) format array.

    For full information please refer to the NAG Library document for
    f01ve

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01vef.html

    Parameters
    ----------
    transr : str, length 1
        Specifies whether the normal RFP representation of A or its
        transpose is stored.

        `transr` = 'N'

            The RFP representation of the matrix A is stored.

        `transr` = 'T'

            The transpose of the RFP representation of the matrix A is
            stored.

    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

            A is upper triangular.

        `uplo` = 'L'

            A is lower triangular.

    a : float, array-like, shape (n, n)
        The triangular matrix A.

    Returns
    -------
    ar : float, ndarray, shape (n*(n+1)/2)
        The upper or lower n by n triangular matrix A (as specified by
        `uplo`) in either normal or transposed RFP format (as specified
        by `transr`). The storage format is described in [the F07
        Introduction].

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `transr`.

            Constraint: `transr` = 'N' or 'T'.

        (`errno` -2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -3)
            On entry, error in parameter n.

            Constraint: n >= 0.

    References
    ----------
    Gustavson, F G, Waśniewski, J, Dongarra, J J and Langou, J, 2010,
    `Rectangular full packed format for Cholesky's algorithm:
    factorization, solution, and inversion`, ACM Trans. Math. Software
    (37, 2)
    """

    fun_name = 'naginterfaces.library.matop.dtrttf'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _ar = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py*(_n_py+1)//2,),
    )
    _b_matop.dtrttf(
        transr,
        uplo,
        _n,
        _a,
        _ar,
    )
    _ar_py = _ar
    return _ar_py

def ztrttf(transr, uplo, a):
    u"""
    Copies a complex triangular matrix from full format to Rectangular
    Full Packed format.

    ``ztrttf`` copies a complex triangular matrix, stored in a full
    format array, to a Rectangular Full Packed (RFP) format array.

    For full information please refer to the NAG Library document for
    f01vf

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01vff.html

    Parameters
    ----------
    transr : str, length 1
        Specifies whether the normal RFP representation of A or its
        conjugate transpose is stored.

        `transr` = 'N'

            The RFP representation of the matrix A is stored.

        `transr` = 'C'

            The conjugate transpose of the RFP representation of the
            matrix A is stored.

    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

            A is upper triangular.

        `uplo` = 'L'

            A is lower triangular.

    a : complex, array-like, shape (n, n)
        The triangular matrix A.

    Returns
    -------
    ar : complex, ndarray, shape (n*(n+1)/2)
        The upper or lower n by n triangular matrix A (as specified by
        `uplo`) in either normal or transposed RFP format (as specified
        by `transr`). The storage format is described in [the F07
        Introduction].

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `transr`.

            Constraint: `transr` = 'N' or 'C'.

        (`errno` -2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -3)
            On entry, error in parameter n.

            Constraint: n >= 0.

    References
    ----------
    Gustavson, F G, Waśniewski, J, Dongarra, J J and Langou, J, 2010,
    `Rectangular full packed format for Cholesky's algorithm:
    factorization, solution, and inversion`, ACM Trans. Math. Software
    (37, 2)
    """

    fun_name = 'naginterfaces.library.matop.ztrttf'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _ar = _b_utils._EngineComplex128ArrayType.empty_ndarray(
        (_n_py*(_n_py+1)//2,),
    )
    _b_matop.ztrttf(
        transr,
        uplo,
        _n,
        _a,
        _ar,
    )
    _ar_py = _ar
    return _ar_py

def dtfttr(transr, uplo, n, ar):
    u"""
    Copies a real triangular matrix from Rectangular Full Packed format
    to full format.

    ``dtfttr`` unpacks a real triangular matrix, stored in a Rectangular
    Full Packed (RFP) format array, to a full format array.

    For full information please refer to the NAG Library document for
    f01vg

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01vgf.html

    Parameters
    ----------
    transr : str, length 1
        Specifies whether the normal RFP representation of A or its
        transpose is stored.

        `transr` = 'N'

            The RFP representation of the matrix A is stored.

        `transr` = 'T'

            The transpose of the RFP representation of the matrix A is
            stored.

    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

            A is upper triangular.

        `uplo` = 'L'

            A is lower triangular.

    n : int
        n, the order of the matrix A.

    ar : float, array-like, shape (`n`*(`n`+1)/2)
        The upper or lower n by n triangular matrix A (as specified by
        `uplo`) in either normal or transposed RFP format (as specified
        by `transr`). The storage format is described in [the F07
        Introduction].

    Returns
    -------
    a : float, ndarray, shape (`n`, `n`)
        The triangular matrix A.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `transr`.

            Constraint: `transr` = 'N' or 'T'.

        (`errno` -2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    References
    ----------
    Gustavson, F G, Waśniewski, J, Dongarra, J J and Langou, J, 2010,
    `Rectangular full packed format for Cholesky's algorithm:
    factorization, solution, and inversion`, ACM Trans. Math. Software
    (37, 2)
    """

    fun_name = 'naginterfaces.library.matop.dtfttr'

    _ar = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ar'},
        ar,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _a = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py, _n_py),
        sorder=_b_utils._EngineData.default_sorder,
    )
    _b_matop.dtfttr(
        transr,
        uplo,
        _n,
        _ar,
        _a,
    )
    _a_py = _a
    return _a_py

def ztfttr(transr, uplo, n, ar):
    u"""
    Copies a complex triangular matrix from Rectangular Full Packed
    format to full format.

    ``ztfttr`` unpacks a complex triangular matrix, stored in a
    Rectangular Full Packed (RFP) format array, to a full format array.

    For full information please refer to the NAG Library document for
    f01vh

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01vhf.html

    Parameters
    ----------
    transr : str, length 1
        Specifies whether the normal RFP representation of A or its
        conjugate transpose is stored.

        `transr` = 'N'

            The RFP representation of the matrix A is stored.

        `transr` = 'C'

            The conjugate transpose of the RFP representation of the
            matrix A is stored.

    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

            A is upper triangular.

        `uplo` = 'L'

            A is lower triangular.

    n : int
        n, the order of the matrix A.

    ar : complex, array-like, shape (`n`*(`n`+1)/2)
        The upper or lower n by n triangular matrix A (as specified by
        `uplo`) in either normal or transposed RFP format (as specified
        by `transr`). The storage format is described in [the F07
        Introduction].

    Returns
    -------
    a : complex, ndarray, shape (`n`, `n`)
        The triangular matrix A.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `transr`.

            Constraint: `transr` = 'N' or 'C'.

        (`errno` -2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    References
    ----------
    Gustavson, F G, Waśniewski, J, Dongarra, J J and Langou, J, 2010,
    `Rectangular full packed format for Cholesky's algorithm:
    factorization, solution, and inversion`, ACM Trans. Math. Software
    (37, 2)
    """

    fun_name = 'naginterfaces.library.matop.ztfttr'

    _ar = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ar'},
        ar,
        copy=False,
    )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _a = _b_utils._EngineComplex128ArrayType.empty_ndarray(
        (_n_py, _n_py),
        sorder=_b_utils._EngineData.default_sorder,
    )
    _b_matop.ztfttr(
        transr,
        uplo,
        _n,
        _ar,
        _a,
    )
    _a_py = _a
    return _a_py

def dtpttf(transr, uplo, n, ap):
    u"""
    Copies a real triangular matrix from packed format to Rectangular
    Full Packed format.

    ``dtpttf`` copies a real triangular matrix, stored in a standard
    packed format array, to a Rectangular Full Packed (RFP) format
    array.

    For full information please refer to the NAG Library document for
    f01vj

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01vjf.html

    Parameters
    ----------
    transr : str, length 1
        Specifies whether the normal RFP representation of A or its
        transpose is stored.

        `transr` = 'N'

            The RFP representation of the matrix A is stored.

        `transr` = 'T'

            The transpose of the RFP representation of the matrix A is
            stored.

    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

            A is upper triangular.

        `uplo` = 'L'

            A is lower triangular.

    n : int
        n, the order of the matrix A.

    ap : float, array-like, shape (`n`*(`n`+1)/2)
        The n by n triangular matrix A, packed by columns.

    Returns
    -------
    ar : float, ndarray, shape (`n`*(`n`+1)/2)
        The upper or lower n by n triangular matrix A (as specified by
        `uplo`) in either normal or transposed RFP format (as specified
        by `transr`). The storage format is described in [the F07
        Introduction].

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `transr`.

            Constraint: `transr` = 'N' or 'T'.

        (`errno` -2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    References
    ----------
    Gustavson, F G, Waśniewski, J, Dongarra, J J and Langou, J, 2010,
    `Rectangular full packed format for Cholesky's algorithm:
    factorization, solution, and inversion`, ACM Trans. Math. Software
    (37, 2)
    """

    fun_name = 'naginterfaces.library.matop.dtpttf'

    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _ar = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py*(_n_py+1)//2,),
    )
    _b_matop.dtpttf(
        transr,
        uplo,
        _n,
        _ap,
        _ar,
    )
    _ar_py = _ar
    return _ar_py

def ztpttf(transr, uplo, n, ap):
    u"""
    Copies a complex triangular matrix from packed format to Rectangular
    Full Packed format.

    ``ztpttf`` copies a complex triangular matrix, stored in a standard
    packed format array, to a Rectangular Full Packed (RFP) format
    array.

    For full information please refer to the NAG Library document for
    f01vk

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01vkf.html

    Parameters
    ----------
    transr : str, length 1
        Specifies whether the normal RFP representation of A or its
        conjugate transpose is stored.

        `transr` = 'N'

            The RFP representation of the matrix A is stored.

        `transr` = 'C'

            The conjugate transpose of the RFP representation of the
            matrix A is stored.

    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

            A is upper triangular.

        `uplo` = 'L'

            A is lower triangular.

    n : int
        n, the order of the matrix A.

    ap : complex, array-like, shape (`n`*(`n`+1)/2)
        The n by n triangular matrix A, packed by columns.

    Returns
    -------
    ar : complex, ndarray, shape (`n`*(`n`+1)/2)
        The upper or lower n by n triangular matrix A (as specified by
        `uplo`) in either normal or transposed RFP format (as specified
        by `transr`). The storage format is described in [the F07
        Introduction].

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `transr`.

            Constraint: `transr` = 'N' or 'C'.

        (`errno` -2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    References
    ----------
    Gustavson, F G, Waśniewski, J, Dongarra, J J and Langou, J, 2010,
    `Rectangular full packed format for Cholesky's algorithm:
    factorization, solution, and inversion`, ACM Trans. Math. Software
    (37, 2)
    """

    fun_name = 'naginterfaces.library.matop.ztpttf'

    _ap = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ap'},
        ap,
        copy=False,
    )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _ar = _b_utils._EngineComplex128ArrayType.empty_ndarray(
        (_n_py*(_n_py+1)//2,),
    )
    _b_matop.ztpttf(
        transr,
        uplo,
        _n,
        _ap,
        _ar,
    )
    _ar_py = _ar
    return _ar_py

def dtfttp(transr, uplo, n, ar):
    u"""
    Copies a real triangular matrix from Rectangular Full Packed format
    to packed format.

    ``dtfttp`` copies a real triangular matrix, stored in a Rectangular
    Full Packed (RFP) format array, to a standard packed format array.

    For full information please refer to the NAG Library document for
    f01vl

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01vlf.html

    Parameters
    ----------
    transr : str, length 1
        Specifies whether the normal RFP representation of A or its
        transpose is stored.

        `transr` = 'N'

            The RFP representation of the matrix A is stored.

        `transr` = 'T'

            The transpose of the RFP representation of the matrix A is
            stored.

    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

            A is upper triangular.

        `uplo` = 'L'

            A is lower triangular.

    n : int
        n, the order of the matrix A.

    ar : float, array-like, shape (`n`*(`n`+1)/2)
        The upper or lower n by n triangular matrix A (as specified by
        `uplo`) in either normal or transposed RFP format (as specified
        by `transr`). The storage format is described in [the F07
        Introduction].

    Returns
    -------
    ap : float, ndarray, shape (`n`*(`n`+1)/2)
        The n by n triangular matrix A, packed by columns.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `transr`.

            Constraint: `transr` = 'N' or 'T'.

        (`errno` -2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    References
    ----------
    Gustavson, F G, Waśniewski, J, Dongarra, J J and Langou, J, 2010,
    `Rectangular full packed format for Cholesky's algorithm:
    factorization, solution, and inversion`, ACM Trans. Math. Software
    (37, 2)
    """

    fun_name = 'naginterfaces.library.matop.dtfttp'

    _ar = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ar'},
        ar,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _ap = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py*(_n_py+1)//2,),
    )
    _b_matop.dtfttp(
        transr,
        uplo,
        _n,
        _ar,
        _ap,
    )
    _ap_py = _ap
    return _ap_py

def ztfttp(transr, uplo, n, ar):
    u"""
    Copies a complex triangular matrix from Rectangular Full Packed
    format to packed format.

    ``ztfttp`` copies a complex triangular matrix, stored in a
    Rectangular Full Packed (RFP) format array, to a standard packed
    format array.

    For full information please refer to the NAG Library document for
    f01vm

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01vmf.html

    Parameters
    ----------
    transr : str, length 1
        Specifies whether the normal RFP representation of A or its
        conjugate transpose is stored.

        `transr` = 'N'

            The RFP representation of the matrix A is stored.

        `transr` = 'C'

            The conjugate transpose of the RFP representation of the
            matrix A is stored.

    uplo : str, length 1
        Specifies whether A is upper or lower triangular.

        `uplo` = 'U'

            A is upper triangular.

        `uplo` = 'L'

            A is lower triangular.

    n : int
        n, the order of the matrix A.

    ar : complex, array-like, shape (`n`*(`n`+1)/2)
        The upper or lower n by n triangular matrix A (as specified by
        `uplo`) in either normal or transposed RFP format (as specified
        by `transr`). The storage format is described in [the F07
        Introduction].

    Returns
    -------
    ap : complex, ndarray, shape (`n`*(`n`+1)/2)
        The n by n triangular matrix A, packed by columns.

    Raises
    ------
    NagValueError
        (`errno` -1)
            On entry, error in parameter `transr`.

            Constraint: `transr` = 'N' or 'C'.

        (`errno` -2)
            On entry, error in parameter `uplo`.

            Constraint: `uplo` = 'U' or 'L'.

        (`errno` -3)
            On entry, error in parameter `n`.

            Constraint: `n` >= 0.

    References
    ----------
    Gustavson, F G, Waśniewski, J, Dongarra, J J and Langou, J, 2010,
    `Rectangular full packed format for Cholesky's algorithm:
    factorization, solution, and inversion`, ACM Trans. Math. Software
    (37, 2)
    """

    fun_name = 'naginterfaces.library.matop.ztfttp'

    _ar = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ar'},
        ar,
        copy=False,
    )
    _n = _b_utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _ap = _b_utils._EngineComplex128ArrayType.empty_ndarray(
        (_n_py*(_n_py+1)//2,),
    )
    _b_matop.ztfttp(
        transr,
        uplo,
        _n,
        _ar,
        _ap,
    )
    _ap_py = _ap
    return _ap_py

def real_tri_pack(job, uplo, diag, a, b):
    u"""
    Convert real matrix between packed triangular and square storage
    formats.

    ``real_tri_pack`` copies a real triangular matrix stored in a packed
    one-dimensional array into an unpacked two-dimensional array, or
    vice versa.

    For full information please refer to the NAG Library document for
    f01za

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01zaf.html

    Parameters
    ----------
    job : str, length 1
        Specifies whether the triangular matrix is to be packed or
        unpacked.

        `job` = 'P' (Pack)

            The matrix is to be packed into array `b`.

        `job` = 'U' (Unpack)

            The matrix is to be unpacked into array `a`.

    uplo : str, length 1
        Specifies the type of the matrix to be copied

        `uplo` = 'L' (Lower)

            The matrix is lower triangular. In this case the packed
            vector holds, or will hold on exit, the matrix elements in
            the following order:
            (1,1),(2,1),...,(n,1),(2,2),(3,2),...,(n,2), etc..

        `uplo` = 'U' (Upper)

            The matrix is upper triangular. In this case the packed
            vector holds, or will hold on exit, the matrix elements in
            the following order: (1,1), (1,2), (2,2), (1,3), (2,3),
            (3,3), (1,4), etc..

    diag : str, length 1
        Must specify whether the diagonal elements of the matrix are to
        be copied.

        `diag` = 'B' (Blank)

            The diagonal elements of the matrix are not referenced and
            not copied.

        `diag` = 'U' (Unit diagonal)

            The diagonal elements of the matrix are not referenced, but
            are assumed all to be unity, and are copied as such.

        `diag` = 'N' (Non-unit diagonal)

            The diagonal elements of the matrix are referenced and
            copied.

    a : float, array-like, shape (n, n)
        If `job` = 'P', the leading n by n part of `a` must contain the
        matrix to be copied, stored in unpacked form, in the upper or
        lower triangle depending on argument `uplo`. The opposite
        triangle of `a` is not referenced and need not be assigned.

    b : float, array-like, shape ((n*(n+1))/2)
        If `job` = 'U', `b` must contain the triangular matrix packed by
        column.

    Returns
    -------
    a : float, ndarray, shape (n, n)
        If `job` = 'U', the leading n by n part of array `a` contains
        the copied matrix, stored in unpacked form, in the upper or
        lower triangle depending on argument `uplo`. The opposite
        triangle of `a` is not referenced.

    b : float, ndarray, shape ((n*(n+1))/2)
        If `job` = 'P', `b` contains the triangular matrix packed by
        column.

        Note that `b` must have space for the diagonal elements of the
        matrix, even if these are not stored.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `job` != 'P' or 'U': `job` = *<value>*.

        (`errno` 2)
            On entry, `uplo` != 'U' or 'L': `uplo` = *<value>*.

        (`errno` 3)
            On entry, `diag` != 'N', 'U' or 'B': `diag` = *<value>*.

        (`errno` 4)
            On entry, n = *<value>*.

            Constraint: n >= 1.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    fun_name = 'naginterfaces.library.matop.real_tri_pack'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b_matop.real_tri_pack(
        job,
        uplo,
        diag,
        _n,
        _a,
        _b,
    )
    _a_py = _a
    _b_py = _b
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'b'],
    )(
        _a_py,
        _b_py,
    )
    return _return_data

def complex_tri_pack(job, uplo, diag, a, b):
    u"""
    Convert complex matrix between packed triangular and square storage
    formats.

    ``complex_tri_pack`` copies a complex triangular matrix stored in a
    packed one-dimensional array into an unpacked two-dimensional array,
    or vice versa.

    For full information please refer to the NAG Library document for
    f01zb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01zbf.html

    Parameters
    ----------
    job : str, length 1
        Specifies whether the triangular matrix is to be packed or
        unpacked.

        `job` = 'P' (Pack)

            The matrix is to be packed into array `b`.

        `job` = 'U' (Unpack)

            The matrix is to be unpacked into array `a`.

    uplo : str, length 1
        Specifies the type of the matrix to be copied

        `uplo` = 'L' (Lower)

            The matrix is lower triangular. In this case the packed
            vector holds, or will hold on exit, the matrix elements in
            the following order:
            (1,1),(2,1),...,(n,1),(2,2),(3,2),...,(n,2), etc..

        `uplo` = 'U' (Upper)

            The matrix is upper triangular. In this case the packed
            vector holds, or will hold on exit, the matrix elements in
            the following order: (1,1), (1,2), (2,2), (1,3), (2,3),
            (3,3), (1,4), etc..

    diag : str, length 1
        Must specify whether the diagonal elements of the matrix are to
        be copied.

        `diag` = 'B' (Blank)

            The diagonal elements of the matrix are not referenced and
            not copied.

        `diag` = 'U' (Unit diagonal)

            The diagonal elements of the matrix are not referenced, but
            are assumed all to be unity, and are copied as such.

        `diag` = 'N' (Non-unit diagonal)

            The diagonal elements of the matrix are referenced and
            copied.

    a : complex, array-like, shape (n, n)
        If `job` = 'P', the leading n by n part of `a` must contain the
        matrix to be copied, stored in unpacked form, in the upper or
        lower triangle depending on argument `uplo`. The opposite
        triangle of `a` is not referenced and need not be assigned.

    b : complex, array-like, shape ((n*(n+1))/2)
        If `job` = 'U', `b` must contain the triangular matrix packed by
        column.

    Returns
    -------
    a : complex, ndarray, shape (n, n)
        If `job` = 'U', the leading n by n part of array `a` contains
        the copied matrix, stored in unpacked form, in the upper or
        lower triangle depending on argument `uplo`. The opposite
        triangle of `a` is not referenced.

    b : complex, ndarray, shape ((n*(n+1))/2)
        If `job` = 'P', `b` contains the triangular matrix packed by
        column.

        Note that `b` must have space for the diagonal elements of the
        matrix, even if these are not stored.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `job` != 'P' or 'U': `job` = *<value>*.

        (`errno` 2)
            On entry, `uplo` != 'U' or 'L': `uplo` = *<value>*.

        (`errno` 3)
            On entry, `diag` = *<value>*.

            Constraint: `diag` = 'N', 'U' or 'B'.

        (`errno` 4)
            On entry, n = *<value>*.

            Constraint: n >= 1.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    fun_name = 'naginterfaces.library.matop.complex_tri_pack'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
    )
    _b_matop.complex_tri_pack(
        job,
        uplo,
        diag,
        _n,
        _a,
        _b,
    )
    _a_py = _a
    _b_py = _b
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'b'],
    )(
        _a_py,
        _b_py,
    )
    return _return_data

def real_band_pack(job, kl, ku, a, b):
    u"""
    Convert real matrix between packed banded and rectangular storage
    formats.

    ``real_band_pack`` copies a real band matrix stored in a packed
    array into an unpacked array, or vice versa.

    For full information please refer to the NAG Library document for
    f01zc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01zcf.html

    Parameters
    ----------
    job : str, length 1
        Specifies whether the band matrix is to be packed or unpacked.

        `job` = 'P' (Pack)

            The band matrix is to be packed into array `b`.

        `job` = 'U' (Unpack)

            The band matrix is to be unpacked into array `a`.

    kl : int
        k_l, the number of subdiagonals of the band matrix.

    ku : int
        k_u, the number of superdiagonals of the band matrix.

    a : float, array-like, shape (m, n)
        If `job` = 'P', the leading m by n part of `a` must contain the
        band matrix stored in unpacked form. Elements of the array that
        lie outside the banded part of the matrix are not referenced and
        need not be assigned.

    b : float, array-like, shape ((`kl`+`ku`+1), min(m+`ku`, n))
        If `job` = 'U', `b` must contain the band matrix in packed form,
        in the leading (k_l+k_u+1) by min(m+k_u, n) part of the array.
        The matrix is packed column by column, with the leading diagonal
        of the matrix in row (k_u+1) of `b`, the first superdiagonal
        starting at position 2 in row k_u, the first subdiagonal
        starting at position 1 in row (k_u+2), and so on. Elements of
        `b` that are not needed to store the band matrix, for instance
        the leading k_u by k_u triangle, are not referenced and need not
        be assigned.

    Returns
    -------
    a : float, ndarray, shape (m, n)
        If `job` = 'U', the leading m by n part of `a` contains the band
        matrix stored in unpacked form. Elements of the leading m by n
        part of `a` that are not within the banded part of the matrix
        are assigned the value zero.

    b : float, ndarray, shape ((`kl`+`ku`+1), min(m+`ku`, n))
        If `job` = 'P', `b` contains the band matrix stored in packed
        form. Elements of `b` that are not needed to store the band
        matrix are not referenced.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `job` = *<value>*.

            Constraint: `job` = 'P' or 'U'.

        (`errno` 2)
            On entry, `kl` = *<value>*.

            Constraint: `kl` >= 0.

        (`errno` 3)
            On entry, `ku` = *<value>*.

            Constraint: `ku` >= 0.

        (`errno` 5)
            On entry, ldb = *<value>*, `kl` = *<value>* and `ku` =
            *<value>*.

            Constraint: ldb >= `kl`+`ku`+1.

        (`errno` 6)
            On entry, n = *<value>*.

            Constraint: n > 0.

        (`errno` 6)
            On entry, m = *<value>*.

            Constraint: m > 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    fun_name = 'naginterfaces.library.matop.real_band_pack'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b_matop.real_band_pack(
        job,
        _m,
        _n,
        kl,
        ku,
        _a,
        _b,
    )
    _a_py = _a
    _b_py = _b
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'b'],
    )(
        _a_py,
        _b_py,
    )
    return _return_data

def complex_band_pack(job, kl, ku, a, b):
    u"""
    Convert complex matrix between packed banded and rectangular storage
    formats.

    ``complex_band_pack`` copies a complex band matrix stored in a
    packed array into an unpacked array, or vice versa.

    For full information please refer to the NAG Library document for
    f01zd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/f01/f01zdf.html

    Parameters
    ----------
    job : str, length 1
        Specifies whether the band matrix is to be packed or unpacked.

        `job` = 'P' (Pack)

            The band matrix is to be packed into array `b`.

        `job` = 'U' (Unpack)

            The band matrix is to be unpacked into array `a`.

    kl : int
        k_l, the number of subdiagonals of the band matrix.

    ku : int
        k_u, the number of superdiagonals of the band matrix.

    a : complex, array-like, shape (m, n)
        If `job` = 'P', the leading m by n part of `a` must contain the
        band matrix stored in unpacked form. Elements of the array that
        lie outside the banded part of the matrix are not referenced and
        need not be assigned.

    b : complex, array-like, shape ((`kl`+`ku`+1), min(m+`ku`, n))
        If `job` = 'U', `b` must contain the band matrix in packed form,
        in the leading (k_l+k_u+1) by min(m+k_u, n) part of the array.
        The matrix is packed column by column, with the leading diagonal
        of the matrix in row (k_u+1) of `b`, the first superdiagonal
        starting at position 2 in row k_u, the first subdiagonal
        starting at position 1 in row (k_u+2), and so on. Elements of
        `b` that are not needed to store the band matrix, for instance
        the leading k_u by k_u triangle, are not referenced and need not
        be assigned.

    Returns
    -------
    a : complex, ndarray, shape (m, n)
        If `job` = 'U', the leading m by n part of `a` contains the band
        matrix stored in unpacked form. Elements of the leading m by n
        part of `a` that are not within the banded part of the matrix
        are assigned the value zero.

    b : complex, ndarray, shape ((`kl`+`ku`+1), min(m+`ku`, n))
        If `job` = 'P', `b` contains the band matrix stored in packed
        form. Elements of `b` that are not needed to store the band
        matrix are not referenced.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `job` = *<value>*.

            Constraint: `job` = 'P' or 'U'.

        (`errno` 2)
            On entry, `kl` = *<value>*.

            Constraint: `kl` >= 0.

        (`errno` 3)
            On entry, `ku` = *<value>*.

            Constraint: `ku` >= 0.

        (`errno` 5)
            On entry, ldb = *<value>*, `kl` = *<value>* and `ku` =
            *<value>*.

            Constraint: ldb >= `kl`+`ku`+1.

        (`errno` 6)
            On entry, n = *<value>*.

            Constraint: n > 0.

        (`errno` 6)
            On entry, m = *<value>*.

            Constraint: m > 0.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    fun_name = 'naginterfaces.library.matop.complex_band_pack'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
    )
    _m_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _n_py = _b_utils._EngineComplex128ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
    )
    _b_matop.complex_band_pack(
        job,
        _m,
        _n,
        kl,
        ku,
        _a,
        _b,
    )
    _a_py = _a
    _b_py = _b
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'b'],
    )(
        _a_py,
        _b_py,
    )
    return _return_data
