# -*- coding: utf-8 -*-
# pylint: disable=chained-comparison,invalid-name,line-too-long,protected-access,too-many-arguments,too-many-boolean-expressions,too-many-branches,too-many-lines,too-many-locals,too-many-statements,unused-import
# pylint: disable=simplifiable-if-statement
u"""
Interfaces for the NAG Mark 27.0 `opt` Chapter.

``opt`` - Minimizing or Maximizing a Function

This module provides functions for solving various mathematical
optimization problems by solvers based on local stopping criteria.
The main classes of problems covered in this module are:

    **Linear Programming (LP)** -- dense and sparse;

    **Quadratic Programming (QP)** -- convex and nonconvex, dense and
    sparse;

    **Nonlinear Programming (NLP)** -- dense and sparse, based on
    active-set SQP methods or interior point methods (IPM);

    **Second-order Cone Programming (SOCP)**;

    **Semidefinite Programming (SDP)** -- both linear matrix
    inequalities (LMI) and bilinear matrix inequalities (BMI);

    **Derivative-free Optimization (DFO)**;

    **Least Squares (LSQ)**, data fitting -- linear and nonlinear,
    constrained and unconstrained.

For a full overview of the functionality offered in this module, see
[the functionality index] or the Module Contents (submodule ``opt``).

See also other modules in the NAG Library relevant to optimization:

    submodule ``glopt`` contains functions to solve **global
    optimization** problems;

    submodule ``mip`` addresses problems arising in **operational
    research** and focuses on **Mixed Integer Programming (MIP)**;

    submodule ``lapacklin`` and submodule ``lapackeig`` include
    functions for linear algebra and in particular unconstrained linear
    least squares;

    submodule ``fit`` focuses on curve and surface fitting, in which
    linear data fitting in l_1 or l_infinity norm might be of interest.

For full information please refer to the NAG Library document

https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04intro.html

See Also
--------
``naginterfaces.library.examples.opt`` :
    This subpackage contains examples for the ``opt`` module.
    See also the :ref:`library_opt_ex` subsection.
"""

# NAG Copyright 2017-2020.

import collections as _collections

from ..base import opt as _b_opt
from ..base import utils as _b_utils
from ..library import _utils

def one_var_func(funct, e1, e2, a, b, maxcal, data=None):
    u"""
    Minimizes a function of one variable, using function values only.

    ``one_var_func`` searches for a minimum, in a given finite interval,
    of a continuous function of a single variable, using function values
    only.
    The method (based on quadratic interpolation) is intended for
    functions which have a continuous first derivative (although it will
    usually work if the derivative has occasional discontinuities).

    For full information please refer to the NAG Library document for
    e04ab

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04abf.html

    Parameters
    ----------
    funct : callable fc = funct(xc, data=None)
        You must supply this function to calculate the value of the
        function F(x) at any point x in [a,b].

        It should be tested separately before being used in conjunction
        with ``one_var_func``.

        Parameters
        ~~~~~~~~~~
        xc : float
            The point x at which the value of F is required.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        fc : float
            Must be set to the value of the function F at the current
            point x.

    e1 : float
        The relative accuracy to which the position of a minimum is
        required. (Note that, since `e1` is a relative tolerance, the
        scaling of x is automatically taken into account.)

        `e1` should be no smaller than 2 epsilon, and preferably not
        much less than sqrt(epsilon), where epsilon is the machine
        precision.

    e2 : float
        The absolute accuracy to which the position of a minimum is
        required. `e2` should be no smaller than 2 epsilon.

    a : float
        The lower bound a of the interval containing a minimum.

    b : float
        The upper bound b of the interval containing a minimum.

    maxcal : int
        The maximum number of calls of F(x) to be allowed.

    data : arbitrary, optional
        User-communication data for callback functions.

    Returns
    -------
    e1 : float
        If you set `e1` to 0.0 (or to any value less than epsilon), `e1`
        will be reset to the default value sqrt(epsilon) before starting
        the minimization process.

    e2 : float
        If you set `e2` to 0.0 (or to any value less than epsilon), `e2`
        will be reset to the default value sqrt(epsilon).

    a : float
        An improved lower bound on the position of the minimum.

    b : float
        An improved upper bound on the position of the minimum.

    maxcal : int
        The total number of times that `funct` was actually called.

    x : float
        The estimated position of the minimum.

    f : float
        The function value at the final point given in `x`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `maxcal` = *<value>*.

            Constraint: `maxcal` >= 3.

        (`errno` 1)
            On entry, `a` = *<value>*, `e2` = *<value>* and `b` =
            *<value>*.

            Constraint: `a`+`e2` < `b`.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            The maximum number of function calls, *<value>*, have been
            performed. This may have happened simply because `maxcal`
            was set too small for the particular problem, or may be due
            to a mistake in the user-supplied function `funct`. If no
            mistake can be found in `funct`, restart ``one_var_func``
            (preferably with the values of `a` and `b` given on exit
            from the previous call to ``one_var_func``).

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Gill, P E and Murray, W, 1973, `Safeguarded steplength algorithms
    for optimization using descent methods`, NPL Report NAC 37, National
    Physical Laboratory
    """

    def cb_funct(xc, data=None):
        """Wrapping user's funct. Called as base funct."""
        if data is not None:
            _fc_funct = funct(
                xc,
                data,
            )
        else:
            _fc_funct = funct(
                xc,
            )
        _fc_funct_base = _fc_funct
        return _fc_funct_base

    fun_name = 'naginterfaces.library.opt.one_var_func'

    _b_utils._check_callable(funct, {'fun_name': fun_name, 'entity': 'funct'})
    _funct = cb_funct
    try:
        _return_scalars = _b_opt.one_var_func(
            _funct,
            e1,
            e2,
            a,
            b,
            maxcal,
            data,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _e1_py = _return_scalars.e1
    _e2_py = _return_scalars.e2
    _a_py = _return_scalars.a
    _b_py = _return_scalars.b
    _maxcal_py = _return_scalars.maxcal
    _x_py = _return_scalars.x
    _f_py = _return_scalars.f
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['e1', 'e2', 'a', 'b', 'maxcal', 'x', 'f'],
    )(
        _e1_py,
        _e2_py,
        _a_py,
        _b_py,
        _maxcal_py,
        _x_py,
        _f_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def one_var_deriv(funct, e1, e2, a, b, maxcal, data=None):
    u"""
    Minimum, function of one variable, using first derivative.

    ``one_var_deriv`` searches for a minimum, in a given finite
    interval, of a continuous function of a single variable, using
    function and first derivative values.
    The method (based on cubic interpolation) is intended for functions
    which have a continuous first derivative (although it will usually
    work if the derivative has occasional discontinuities).

    For full information please refer to the NAG Library document for
    e04bb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04bbf.html

    Parameters
    ----------
    funct : callable (fc, gc) = funct(xc, data=None)
        You must supply this function to calculate the values of F(x)
        and dF/dx at any point x in [a,b].

        It should be tested separately before being used in conjunction
        with ``one_var_deriv``.

        Parameters
        ~~~~~~~~~~
        xc : float
            The point x at which the values of F and dF/dx are required.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        fc : float
            Must be set to the value of the function F at the current
            point x.

        gc : float
            Must be set to the value of the first derivative dF/dx at
            the current point x.

    e1 : float
        The relative accuracy to which the position of a minimum is
        required. (Note that, since `e1` is a relative tolerance, the
        scaling of x is automatically taken into account.)

        `e1` should be no smaller than 2 epsilon, and preferably not
        much less than sqrt(epsilon), where epsilon is the machine
        precision.

    e2 : float
        The absolute accuracy to which the position of a minimum is
        required. `e2` should be no smaller than 2 epsilon.

    a : float
        The lower bound a of the interval containing a minimum.

    b : float
        The upper bound b of the interval containing a minimum.

    maxcal : int
        The maximum number of calls of `funct` to be allowed.

    data : arbitrary, optional
        User-communication data for callback functions.

    Returns
    -------
    e1 : float
        If you set `e1` to 0.0 (or to any value less than epsilon), `e1`
        will be reset to the default value sqrt(epsilon) before starting
        the minimization process.

    e2 : float
        If you set `e2` to 0.0 (or to any value less than epsilon), `e2`
        will be reset to the default value sqrt(epsilon).

    a : float
        An improved lower bound on the position of the minimum.

    b : float
        An improved upper bound on the position of the minimum.

    maxcal : int
        The total number of times that `funct` was actually called.

    x : float
        The estimated position of the minimum.

    f : float
        The function value at the final point given in `x`.

    g : float
        The value of the first derivative at the final point in `x`.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            On entry, `a`+`e2` = *<value>* and `b` = *<value>*.

            Constraint: `a`+`e2` < `b`.

        (`errno` 1)
            On entry, `maxcal` = *<value>*.

            Constraint: `maxcal` >= 2.

        (`errno` 2)
            The maximum number of function calls, *<value>*, have been
            performed. This may have happened simply because `maxcal`
            was set too small for the particular problem, or may be due
            to a mistake in the user-supplied function `funct`. If no
            mistake can be found in `funct`, restart ``one_var_deriv``
            (preferably with the values of `a` and `b` given on exit
            from the previous call to ``one_var_deriv``).

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Gill, P E and Murray, W, 1973, `Safeguarded steplength algorithms
    for optimization using descent methods`, NPL Report NAC 37, National
    Physical Laboratory
    """

    def cb_funct(xc, data=None):
        """Wrapping user's funct. Called as base funct."""
        if data is not None:
            _fc_funct, _gc_funct = funct(
                xc,
                data,
            )
        else:
            _fc_funct, _gc_funct = funct(
                xc,
            )
        _fc_funct_base = _fc_funct
        _gc_funct_base = _gc_funct
        return (
            _fc_funct_base,
            _gc_funct_base,
        )

    fun_name = 'naginterfaces.library.opt.one_var_deriv'

    _b_utils._check_callable(funct, {'fun_name': fun_name, 'entity': 'funct'})
    _funct = cb_funct
    try:
        _return_scalars = _b_opt.one_var_deriv(
            _funct,
            e1,
            e2,
            a,
            b,
            maxcal,
            data,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _e1_py = _return_scalars.e1
    _e2_py = _return_scalars.e2
    _a_py = _return_scalars.a
    _b_py = _return_scalars.b
    _maxcal_py = _return_scalars.maxcal
    _x_py = _return_scalars.x
    _f_py = _return_scalars.f
    _g_py = _return_scalars.g
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['e1', 'e2', 'a', 'b', 'maxcal', 'x', 'f', 'g'],
    )(
        _e1_py,
        _e2_py,
        _a_py,
        _b_py,
        _maxcal_py,
        _x_py,
        _f_py,
        _g_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def uncon_simplex(x, tolf, tolx, funct, maxcal, monit=None, data=None):
    u"""
    Unconstrained minimum, Nelder--Mead simplex algorithm, using
    function values only.

    ``uncon_simplex`` minimizes a general function F(x) of n independent
    variables x = (x_1,x_2,...,x_n)^T by the Nelder and Mead simplex
    method (see Nelder and Mead (1965)).
    Derivatives of the function need not be supplied.

    For full information please refer to the NAG Library document for
    e04cb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04cbf.html

    Parameters
    ----------
    x : float, array-like, shape (n)
        A guess at the position of the minimum. Note that the problem
        should be scaled so that the values of the `x`[i-1] are of order
        unity.

    tolf : float
        The error tolerable in the function values, in the following
        sense. If f_i, for i = 1,2,...,n+1, are the individual function
        values at the vertices of the current simplex, and if f_m is the
        mean of these values, then you can request that
        ``uncon_simplex`` should terminate if

            sqrt(1/(n+1)sum_i = 1^n+1(f_i-f_m)^2) < `tolf`.

        You may specify `tolf` = 0 if you wish to use only the
        termination criterion (2) on the spatial values: see the
        description of `tolx`.

    tolx : float
        The error tolerable in the spatial values, in the following
        sense. If LV denotes the 'linearized' volume of the current
        simplex, and if LV_init denotes the 'linearized' volume of the
        initial simplex, then you can request that ``uncon_simplex``
        should terminate if

            LV/LV_init < `tolx`.

        You may specify `tolx` = 0 if you wish to use only the
        termination criterion (1) on function values: see the
        description of `tolf`.

    funct : callable fc = funct(xc, data=None)
        `funct` must evaluate the function F at a specified point.

        It should be tested separately before being used in conjunction
        with ``uncon_simplex``.

        Parameters
        ~~~~~~~~~~
        xc : float, ndarray, shape (n)
            The point at which the function value is required.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        fc : float
            The value of the function F at the current point x.

    maxcal : int
        The maximum number of function evaluations to be allowed.

    monit : None or callable monit(fmin, fmax, sim, ncall, serror,
        vratio, data=None), optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `monit` may be used to monitor the optimization process.

        It is invoked once every iteration.

        If no monitoring is required, `monit` may be None.

        Parameters
        ~~~~~~~~~~
        fmin : float
            The smallest function value in the current simplex.

        fmax : float
            The largest function value in the current simplex.

        sim : float, ndarray, shape (n+1, n)
            The n+1 position vectors of the current simplex.

        ncall : int
            The number of times that `funct` has been called so far.

        serror : float
            The current value of the standard deviation in function
            values used in termination test (1).

        vratio : float
            The current value of the linearized volume ratio used in
            termination test (2).

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

    data : arbitrary, optional
        User-communication data for callback functions.

    Returns
    -------
    x : float, ndarray, shape (n)
        The value of x corresponding to the function value in `f`.

    f : float
        The lowest function value found.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 1)
            On entry, `tolx` = 0.0 and `tolf` = *<value>*.

            Constraint: if `tolx` = 0.0 then `tolf` is greater than or
            equal to the machine precision.

        (`errno` 1)
            On entry, `tolf` = 0.0 and `tolx` = *<value>*.

            Constraint: if `tolf` = 0.0 then `tolx` is greater than or
            equal to the machine precision.

        (`errno` 1)
            On entry, `tolf` = *<value>* and `tolx` = *<value>*.

            Constraint: if `tolf` != 0.0 and `tolx` != 0.0 then both
            should be greater than or equal to the machine precision.

        (`errno` 1)
            On entry, `maxcal` = *<value>*.

            Constraint: `maxcal` >= 1.

    Warns
    -----
    NagAlgorithmicMajorWarning
        (`errno` 2)
            `maxcal` function evaluations have been completed without
            any other termination test passing. Check the coding of
            `funct` before increasing the value of `maxcal`.

    References
    ----------
    Nelder, J A and Mead, R, 1965, `A simplex method for function
    minimization`, Comput. J. (7), 308--313

    Parkinson, J M and Hutchinson, D, 1972, `An investigation into the
    efficiency of variants of the simplex method`, Numerical Methods for
    Nonlinear Optimization, (ed F A Lootsma), Academic Press

    Singer, S and Singer, S, 2004, `Efficient implementation of the
    Nelder--Mead search algorithm`, Appl. Num. Anal. Comp. Math. (1(3)),
    524--534

    See Also
    --------
    :meth:`naginterfaces.library.examples.opt.uncon_simplex_ex.main`
    """

    def cb_funct(n, xc, data=None):
        """Wrapping user's funct. Called as base funct."""
        if data is not None:
            _fc_funct = funct(
                xc,
                data,
            )
        else:
            _fc_funct = funct(
                xc,
            )
        _fc_funct_base = _fc_funct
        return _fc_funct_base

    def cb_monit(fmin, fmax, sim, n, ncall, serror, vratio, data=None):
        """Wrapping user's monit. Called as base monit."""
        if data is not None:
            monit(
                fmin,
                fmax,
                sim,
                ncall,
                serror,
                vratio,
                data,
            )
        else:
            monit(
                fmin,
                fmax,
                sim,
                ncall,
                serror,
                vratio,
            )

    fun_name = 'naginterfaces.library.opt.uncon_simplex'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b_utils._check_callable(funct, {'fun_name': fun_name, 'entity': 'funct'})
    _funct = cb_funct
    _b_utils._check_callable(monit, {'fun_name': fun_name, 'entity': 'monit'}, allow_none=True)
    _monit = None if monit is None else cb_monit
    try:
        _return_scalar = _b_opt.uncon_simplex(
            _n,
            _x,
            tolf,
            tolx,
            _funct,
            _monit,
            maxcal,
            data,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _x_py = _x
    _f_py = _return_scalar
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'f'],
    )(
        _x_py,
        _f_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

@_b_utils._withdrawal(
    'naginterfaces.library.opt.uncon_conjgrd_comp',
    mkwdrn=None,
    advice=[
        'Please use handle_solve_bounds_foas instead.',
        'See also https://www.nag.com/numeric/py/nagdoc_latest/replace.html',
    ]
)
def uncon_conjgrd_comp(objfun, x, comm, data=None, io_manager=None):
    u"""
    Unconstrained minimum, preconditioned conjugate gradient algorithm,
    using first derivatives (comprehensive).

    Note: this function uses optional algorithmic parameters, see also:
    ``uncon_conjgrd_option_file``, ``uncon_conjgrd_option_string``,
    ``nlp1_init``.

    ``uncon_conjgrd_comp`` minimizes an unconstrained nonlinear function
    of several variables using a pre-conditioned, limited memory
    quasi-Newton conjugate gradient method.
    First derivatives (or an 'acceptable' finite difference
    approximation to them) are required.
    It is intended for use on large scale problems.

    Note: ``uncon_conjgrd_comp`` is deprecated.
    Please use ``handle_solve_bounds_foas`` instead.
    See also https://www.nag.com/numeric/py/nagdoc_latest/replace.html

    For full information please refer to the NAG Library document for
    e04dg

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04dgf.html

    Parameters
    ----------
    objfun : callable (objf, objgrd) = objfun(mode, x, nstate,
        data=None)
        `objfun` must calculate the objective function F(x) and possibly
        its gradient as well for a specified n-element vector x.

        Parameters
        ~~~~~~~~~~
        mode : int
            Indicates which values must be assigned during each call of
            `objfun`. Only the following values need be assigned:

            `mode` = 0

                `objf`.

            `mode` = 2

                `objf` and `objgrd`.

        x : float, ndarray, shape (n)
            x, the vector of variables at which the objective function
            and its gradient are to be evaluated.

        nstate : int
            Will be 1 on the first call of `objfun` by
            ``uncon_conjgrd_comp``, and 0 for all subsequent calls.
            Thus, you may wish to test, `nstate` within `objfun` in
            order to perform certain calculations once only. For
            example, you may read data or initialize global variables
            when `nstate` = 1.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        objf : float
            The value of the objective function at x.

        objgrd : float, array-like, shape (n)
            If `mode` = 2, `objgrd`[i-1] must contain the value of
            dF/dx_i evaluated at x, for i = 1,2,...,n.

    x : float, array-like, shape (n)
        An initial estimate of the solution.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``nlp1_init``.

    data : arbitrary, optional
        User-communication data for callback functions.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    itera : int
        The total number of iterations performed.

    objf : float
        The value of the objective function at the final iterate.

    objgrd : float, ndarray, shape (n)
        The gradient of the objective function at the final iterate (or
        its finite difference approximation).

    x : float, ndarray, shape (n)
        The final estimate of the solution.

    Other Parameters
    ----------------
    'Defaults' : valueless
        This special keyword may be used to reset all options to their
        default values.

    'Estimated Optimal Function Value' : float
        This value of r specifies the user-supplied guess of the optimum
        objective function value F_est.
        This value is used to calculate an initial step length alpha_0
        (see [Algorithmic Details]).
        If the value of r is not specified (the default), then this has
        the effect of setting alpha_0 to unity.
        It should be noted that for badly scaled functions a unit step
        along the steepest descent direction will often compute the
        objective function at very large values of x.

    'Function Precision' : float
        Default = epsilon^0.9

        The argument defines epsilon_r, which is intended to be a
        measure of the accuracy with which the problem function F(x) can
        be computed.
        If r < epsilon or r >= 1, the default value is used.

        The value of epsilon_r should reflect the relative precision of
        1+|F(x)|; i.e., epsilon_r acts as a relative precision when |F|
        is large, and as an absolute precision when |F| is small.
        For example, if F(x) is typically of order 1000 and the first
        six significant digits are known to be correct, an appropriate
        value for epsilon_r would be 10^(-6).
        In contrast, if F(x) is typically of order 10^(-4) and the first
        six significant digits are known to be correct, an appropriate
        value for epsilon_r would be 10^(-10).
        The choice of epsilon_r can be quite complicated for badly
        scaled problems; see Module 8 of Gill `et al.` (1981) for a
        discussion of scaling techniques.
        The default value is appropriate for most simple functions that
        are computed with full accuracy.
        However when the accuracy of the computed function values is
        known to be significantly worse than full precision, the value
        of epsilon_r should be large enough so that no attempt will be
        made to distinguish between function values that differ by less
        than the error inherent in the calculation.

    'Iteration Limit' : int
        Default = max(50, 5n)

        The value of i specifies the maximum number of iterations
        allowed before termination.
        If i < 0, the default value is used.

        Problems whose Hessian matrices at the solution contain sets of
        clustered eigenvalues are likely to be minimized in
        significantly fewer than n iterations.
        Problems without this property may require anything between n
        and 5n iterations, with approximately 2n iterations being a
        common figure for moderately difficult problems.

    'Iters' : int
        Default = max(50, 5n)

        The value of i specifies the maximum number of iterations
        allowed before termination.
        If i < 0, the default value is used.

        Problems whose Hessian matrices at the solution contain sets of
        clustered eigenvalues are likely to be minimized in
        significantly fewer than n iterations.
        Problems without this property may require anything between n
        and 5n iterations, with approximately 2n iterations being a
        common figure for moderately difficult problems.

    'Itns' : int
        Default = max(50, 5n)

        The value of i specifies the maximum number of iterations
        allowed before termination.
        If i < 0, the default value is used.

        Problems whose Hessian matrices at the solution contain sets of
        clustered eigenvalues are likely to be minimized in
        significantly fewer than n iterations.
        Problems without this property may require anything between n
        and 5n iterations, with approximately 2n iterations being a
        common figure for moderately difficult problems.

    'Linesearch Tolerance' : float
        Default = 0.9

        The value r controls the accuracy with which the step alpha
        taken during each iteration approximates a minimum of the
        function along the search direction (the smaller the value of r,
        the more accurate the linesearch).
        The default value r = 0.9 requests an inaccurate search, and is
        appropriate for most problems.
        A more accurate search may be appropriate when it is desirable
        to reduce the number of iterations -- for example, if the
        objective function is cheap to evaluate.
        If r < 0 or r >= 1, the default value is used.

    'List' : valueless
        Option 'List' enables printing of each option specification as
        it is supplied. 'Nolist' suppresses this printing.

    'Nolist' : valueless
        Default for ``uncon_conjgrd_comp`` = 'Nolist'

        Option 'List' enables printing of each option specification as
        it is supplied. 'Nolist' suppresses this printing.

    'Maximum Step Length' : float
        Default = 10^20

        If r > 0, the maximum allowable step length for the linesearch
        is taken as min(1/machine.real_safe(), r/||p_k||).
        If r <= 0, the default value is used.

    'Optimality Tolerance' : float
        Default = epsilon_R^0.8

        The argument r specifies the accuracy to which you wish the
        final iterate to approximate a solution of the problem.
        Broadly speaking, r indicates the number of correct figures
        desired in the objective function at the solution.
        For example, if r is 10^(-6) and termination occurs with no
        exception or warning is raised (see [Parameters]), then the
        final point satisfies the termination criteria, where tau_F
        represents 'Optimality Tolerance'.
        If r < epsilon_r or r >= 1, the default value is used.
        If 'Optimality Tolerance' is chosen below a certain threshold,
        it will automatically be reset to another value.

    'Print Level' : int
        The value i controls the amount of printout produced by
        ``uncon_conjgrd_comp``, as indicated below.
        A detailed description of the printout is given in [Description
        of Printed Output] (summary output at each iteration and the
        final solution).

        [table omitted]

    'Start Objective Check at Variable' : int
        Default = 1

        These keywords take effect only if 'Verify Level' > 0.
        They may be used to control the verification of gradient
        elements computed by `objfun`.
        For example, if the first 30 elements of the objective gradient
        appeared to be correct in an earlier run, so that only element
        31 remains questionable, it is reasonable to specify 'Start
        Objective Check at Variable' = 31.
        If the first 30 variables appear linearly in the objective, so
        that the corresponding gradient elements are constant, the above
        choice would also be appropriate.

        If i_1 <= 0 or i_1 > max(1, min(n, i_2)), the default value is
        used.
        If i_2 <= 0 or i_2 > n, the default value is used.

    'Stop Objective Check at Variable' : int
        Default = n

        These keywords take effect only if 'Verify Level' > 0.
        They may be used to control the verification of gradient
        elements computed by `objfun`.
        For example, if the first 30 elements of the objective gradient
        appeared to be correct in an earlier run, so that only element
        31 remains questionable, it is reasonable to specify 'Start
        Objective Check at Variable' = 31.
        If the first 30 variables appear linearly in the objective, so
        that the corresponding gradient elements are constant, the above
        choice would also be appropriate.

        If i_1 <= 0 or i_1 > max(1, min(n, i_2)), the default value is
        used.
        If i_2 <= 0 or i_2 > n, the default value is used.

    'Verify Level' : int
        Default = 0

        These keywords refer to finite difference checks on the gradient
        elements computed by `objfun`.
        Gradients are verified at the user-supplied initial estimate of
        the solution.
        The possible choices for i are as follows:

        [table omitted]

        For example, the objective gradient will be verified if
        'Verify', 'Verify' = 'YES', 'Verify Gradients', 'Verify
        Objective Gradients' or 'Verify Level' = 1 is specified.

    'Verify' : valueless
        These keywords refer to finite difference checks on the gradient
        elements computed by `objfun`.
        Gradients are verified at the user-supplied initial estimate of
        the solution.
        The possible choices for i are as follows:

        [table omitted]

        For example, the objective gradient will be verified if
        'Verify', 'Verify' = 'YES', 'Verify Gradients', 'Verify
        Objective Gradients' or 'Verify Level' = 1 is specified.

    'Verify Gradients' : valueless
        These keywords refer to finite difference checks on the gradient
        elements computed by `objfun`.
        Gradients are verified at the user-supplied initial estimate of
        the solution.
        The possible choices for i are as follows:

        [table omitted]

        For example, the objective gradient will be verified if
        'Verify', 'Verify' = 'YES', 'Verify Gradients', 'Verify
        Objective Gradients' or 'Verify Level' = 1 is specified.

    'Verify Objective Gradients' : valueless
        These keywords refer to finite difference checks on the gradient
        elements computed by `objfun`.
        Gradients are verified at the user-supplied initial estimate of
        the solution.
        The possible choices for i are as follows:

        [table omitted]

        For example, the objective gradient will be verified if
        'Verify', 'Verify' = 'YES', 'Verify Gradients', 'Verify
        Objective Gradients' or 'Verify Level' = 1 is specified.

    Raises
    ------
    NagValueError
        (`errno` 8)
            Gradient at the starting point is too small.

        (`errno` 9)
            On entry, n = *<value>*.

            Constraint: n > 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 6)
            Current point cannot be improved upon.

        (`errno` 7)
            Large errors found in the derivatives.

    NagAlgorithmicMajorWarning
        (`errno` 3)
            Too many iterations.

        (`errno` 4)
            Computed upper bound on step length is too small.

    NagCallbackTerminateWarning
        (`errno` i < 0)
            User requested termination.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Gill, P E and Murray, W, 1979, `Conjugate-gradient methods for
    large-scale nonlinear optimization`, Technical Report SOL 79-15,
    Department of Operations Research, Stanford University

    Gill, P E, Murray, W and Wright, M H, 1981, `Practical
    Optimization`, Academic Press
    """

    fun_name = 'naginterfaces.library.opt.uncon_conjgrd_comp'

    cb_objfun_name = fun_name + ':objfun'

    def cb_objfun(mode, n, x, objgrd, nstate, data=None):
        """Wrapping user's objfun. Called as base objfun."""
        if data is not None:
            _objf_objfun, _objgrd_objfun = objfun(
                mode,
                x,
                nstate,
                data,
            )
        else:
            _objf_objfun, _objgrd_objfun = objfun(
                mode,
                x,
                nstate,
            )
        objgrd[:] = _utils._to_ndarray(
            {'fun_name': cb_objfun_name, 'returned': 'objgrd'},
            _objgrd_objfun,
            copy=False,
            exp_shape=(n,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _objf_objfun_base = _objf_objfun
        return _objf_objfun_base

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b_utils._check_callable(objfun, {'fun_name': fun_name, 'entity': 'objfun'})
    _objfun = cb_objfun
    _objgrd = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _return_scalars = _b_opt.uncon_conjgrd_comp(
            _n,
            _objfun,
            _objgrd,
            _x,
            comm,
            data,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _x_py = _x
    _itera_py = _return_scalars.itera
    _objf_py = _return_scalars.objf
    _objgrd_py = _objgrd
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['itera', 'objf', 'objgrd', 'x'],
    )(
        _itera_py,
        _objf_py,
        _objgrd_py,
        _x_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

@_b_utils._withdrawal(
    'naginterfaces.library.opt.uncon_conjgrd_option_file',
    mkwdrn=None,
)
def uncon_conjgrd_option_file(ioptns, comm, io_manager=None):
    u"""
    Supply option values for ``uncon_conjgrd_comp`` from external file.

    ``uncon_conjgrd_option_file`` may be used to supply options to
    ``uncon_conjgrd_comp`` from an external file.

    Note: ``uncon_conjgrd_option_file`` is deprecated.
    There is no suggested replacement for this routine.

    For full information please refer to the NAG Library document for
    e04dj

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04djf.html

    Parameters
    ----------
    ioptns : int
        The unit number (see
        :meth:`~naginterfaces.base.utils.FileObjManager.unit_from_fileobj`)
        of the options file to be read.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``nlp1_init``.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `ioptns` = *<value>*.

            Constraint: 0 <= `ioptns` <= 99.

        (`errno` 2)
            ``Begin`` was found, but end-of-file was found before
            ``End`` was found.

        (`errno` 3)
            End-of-file was found before ``Begin`` was found.

        (`errno` 5)
            One or more lines of the options file is invalid.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    _b_opt.uncon_conjgrd_option_file(
        ioptns,
        comm,
        io_manager,
    )

@_b_utils._withdrawal(
    'naginterfaces.library.opt.uncon_conjgrd_option_string',
    mkwdrn=None,
)
def uncon_conjgrd_option_string(optstr, comm, io_manager=None):
    u"""
    Supply option values to ``uncon_conjgrd_comp`` from a character
    string.

    ``uncon_conjgrd_option_string`` may be used to supply individual
    options to ``uncon_conjgrd_comp``.

    Note: ``uncon_conjgrd_option_string`` is deprecated.
    There is no suggested replacement for this routine.

    For full information please refer to the NAG Library document for
    e04dk

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04dkf.html

    Parameters
    ----------
    optstr : str
        A single valid option string (as described in [Description] and
        in [Other Parameters for uncon_conjgrd_comp]).

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``nlp1_init``.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Raises
    ------
    NagValueError
        (`errno` 5)
            The supplied option string is invalid. Supplied value was:
            *<value>*.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    _b_opt.uncon_conjgrd_option_string(
        optstr,
        comm,
        io_manager,
    )

def lsq_uncon_mod_func_comp(m, lsqfun, x, lsqmon=None, iprint=1, maxcal=None, eta=None, xtol=0.0, stepmx=100000.0, data=None):
    u"""
    Unconstrained minimum of a sum of squares, combined Gauss--Newton
    and modified Newton algorithm, using function values only
    (comprehensive).

    ``lsq_uncon_mod_func_comp`` is a comprehensive algorithm for finding
    an unconstrained minimum of a sum of squares of m nonlinear
    functions in n variables (m >= n).
    No derivatives are required.

    The function is intended for functions which have continuous first
    and second derivatives (although it will usually work even if the
    derivatives have occasional discontinuities).

    For full information please refer to the NAG Library document for
    e04fc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04fcf.html

    Parameters
    ----------
    m : int
        The number m of residuals, f_i(x), and the number n of
        variables, x_j.

    lsqfun : callable (iflag, fvec) = lsqfun(iflag, m, xc, data=None)
        `lsqfun` must calculate the vector of values f_i(x) at any point
        x. (However, if you do not wish to calculate the residuals at a
        particular x, there is the option of setting an argument to
        cause ``lsq_uncon_mod_func_comp`` to terminate immediately.)

        Parameters
        ~~~~~~~~~~
        iflag : int
            Has a non-negative value.

        m : int
            m, the numbers of residuals.

        xc : float, ndarray, shape (n)
            The point x at which the values of the f_i are required.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        iflag : int
            If `lsqfun` resets `iflag` to some negative number,
            ``lsq_uncon_mod_func_comp`` will terminate immediately, with
            errno set to your setting of `iflag`.

        fvec : float, array-like, shape (`m`)
            Unless `iflag` is reset to a negative number, `fvec`[i-1]
            must contain the value of f_i at the point x, for i =
            1,2,...,m.

    x : float, array-like, shape (n)
        `x`[j-1] must be set to a guess at the jth component of the
        position of the minimum, for j = 1,2,...,n.

    lsqmon : None or callable lsqmon(xc, fvec, fjac, s, igrade, niter,
        nf, data=None), optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        If `iprint` >= 0, you must supply `lsqmon` which is suitable for
        monitoring the minimization process. `lsqmon` must not change
        the values of any of its arguments.

        Parameters
        ~~~~~~~~~~
        xc : float, ndarray, shape (n)
            The coordinates of the current point x.

        fvec : float, ndarray, shape (m)
            The values of the residuals f_i at the current point x.

        fjac : float, ndarray, shape (m, n)
            `fjac`[i-1,j-1] contains the value of df_i/dx_j at the
            current point x, for j = 1,2,...,n, for i = 1,2,...,m.

        s : float, ndarray, shape (n)
            The singular values of the current approximation to the
            Jacobian matrix. Thus `s` may be useful as information about
            the structure of your problem.

        igrade : int
            ``lsq_uncon_mod_func_comp`` estimates the dimension of the
            subspace for which the Jacobian matrix can be used as a
            valid approximation to the curvature (see Gill and Murray
            (1978)). This estimate is called the grade of the Jacobian
            matrix, and `igrade` gives its current value.

        niter : int
            The number of iterations which have been performed in
            ``lsq_uncon_mod_func_comp``.

        nf : int
            The number of times that `lsqfun` has been called so far.
            (However, for intermediate calls of `lsqmon`, `nf` is
            calculated on the assumption that the latest linear search
            has been successful. If this is not the case, the n
            evaluations allowed for approximating the Jacobian at the
            new point will not in fact have been made. `nf` will be
            accurate at the final call of `lsqmon`.)

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

    iprint : int, optional
        The frequency with which `lsqmon` is to be called.

        If `iprint` > 0, `lsqmon` is called once every `iprint`
        iterations and just before exit from
        ``lsq_uncon_mod_func_comp``.

        If `iprint` = 0, `lsqmon` is just called at the final point.

        If `iprint` < 0, `lsqmon` is not called at all.

        `iprint` should normally be set to a small positive number.

    maxcal : None or int, optional
        Note: if this argument is None then a default value will be
        used, determined as follows: 400*n.

        The limit you set on the number of times that `lsqfun` may be
        called by ``lsq_uncon_mod_func_comp``. There will be an error
        exit (see [Raises/Warns]) after `maxcal` calls of `lsqfun`.

    eta : None or float, optional
        Note: if this argument is None then a default value will be
        used, determined as follows: if n = 1: 0.0; otherwise: 0.5.

        Every iteration of ``lsq_uncon_mod_func_comp`` involves a linear
        minimization, i.e., minimization of F(x^(k)+alpha^(k)p^(k)) with
        respect to alpha^(k).

        Specifies how accurately the linear minimizations are to be
        performed. The minimum with respect to alpha^(k) will be located
        more accurately for small values of `eta` (say, 0.01) than for
        large values (say, 0.9). Although accurate linear minimizations
        will generally reduce the number of iterations performed by
        ``lsq_uncon_mod_func_comp``, they will increase the number of
        calls of `lsqfun` made each iteration. On balance it is usually
        more efficient to perform a low accuracy minimization.

    xtol : float, optional
        The accuracy in x to which the solution is required.

        If x_true is the true value of x at the minimum, then x_sol, the
        estimated position before a normal exit, is such that

            ||x_sol-x_true|| < `xtol`*(1.0+||x_true||),

        where ||y|| = sqrt(sum_j = 1^ny_j^2).

        For example, if the elements of x_sol are not much larger than
        1.0 in modulus and if `xtol` = 1.0e-5, then x_sol is usually
        accurate to about five decimal places. (For further details see
        [Accuracy].)

    stepmx : float, optional
        An estimate of the Euclidean distance between the solution and
        the starting point supplied by you. (For maximum efficiency, a
        slight overestimate is preferable.) ``lsq_uncon_mod_func_comp``
        will ensure that, for each iteration,

            sum_j = 1^n(x_j^(k)-x_j^(k-1))^2 <= (`stepmx`)^2,

        where k is the iteration number. Thus, if the problem has more
        than one solution, ``lsq_uncon_mod_func_comp`` is most likely to
        find the one nearest to the starting point. On difficult
        problems, a realistic choice can prevent the sequence x^(k)
        entering a region where the problem is ill-behaved and can help
        avoid overflow in the evaluation of F(x). However, an
        underestimate of `stepmx` can lead to inefficiency.

    data : arbitrary, optional
        User-communication data for callback functions.

    Returns
    -------
    x : float, ndarray, shape (n)
        The final point x^(k). Thus, if no exception or warning is
        raised on exit, `x`[j-1] is the jth component of the estimated
        position of the minimum.

    fsumsq : float
        The value of F(x), the sum of squares of the residuals f_i(x),
        at the final point given in `x`.

    fvec : float, ndarray, shape (`m`)
        The value of the residual f_i(x) at the final point given in
        `x`, for i = 1,2,...,m.

    fjac : float, ndarray, shape (`m`, n)
        The estimate of the first derivative df_i/dx_j at the final
        point given in `x`, for j = 1,2,...,n, for i = 1,2,...,m.

    s : float, ndarray, shape (n)
        The singular values of the estimated Jacobian matrix at the
        final point. Thus `s` may be useful as information about the
        structure of your problem.

    v : float, ndarray, shape (n, n)
        The matrix V associated with the singular value decomposition

            J = USV^T

        of the estimated Jacobian matrix at the final point, stored by
        columns. This matrix may be useful for statistical purposes,
        since it is the matrix of orthonormalized eigenvectors of J^TJ.

    niter : int
        The number of iterations which have been performed in
        ``lsq_uncon_mod_func_comp``.

    nf : int
        The number of times that the residuals have been evaluated
        (i.e., number of calls of `lsqfun`).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `stepmx` = *<value>* and `xtol` = *<value>*.

            Constraint: `stepmx` >= `xtol`.

        (`errno` 1)
            On entry, `xtol` = *<value>*.

            Constraint: `xtol` >= 0.0.

        (`errno` 1)
            On entry, `eta` = *<value>*.

            Constraint: 0.0 <= `eta` < 1.0.

        (`errno` 1)
            On entry, `maxcal` = *<value>*.

            Constraint: `maxcal` >= 1.

        (`errno` 1)
            On entry, `m` = *<value>* and n = *<value>*.

            Constraint: `m` >= n.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 1.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i < 0)
            User requested termination by setting `iflag` negative in
            `lsqfun`.

        (`errno` 2)
            There have been `maxcal` = *<value>* calls to `lsqfun`.

        (`errno` 3)
            The conditions for a minimum have not all been satisfied,
            but a lower point could not be found.

        (`errno` 4)
            Failure in computing SVD of estimated Jacobian matrix.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Gill, P E and Murray, W, 1978, `Algorithms for the solution of the
    nonlinear least squares problem`, SIAM J. Numer. Anal. (15),
    977--992
    """

    fun_name = 'naginterfaces.library.opt.lsq_uncon_mod_func_comp'

    cb_lsqfun_name = fun_name + ':lsqfun'

    def cb_lsqfun(iflag, m, n, xc, fvec, data=None):
        """Wrapping user's lsqfun. Called as base lsqfun."""
        if data is not None:
            _iflag_lsqfun, _fvec_lsqfun = lsqfun(
                iflag,
                m,
                xc,
                data,
            )
        else:
            _iflag_lsqfun, _fvec_lsqfun = lsqfun(
                iflag,
                m,
                xc,
            )
        _iflag_lsqfun_base = _iflag_lsqfun
        fvec[:] = _utils._to_ndarray(
            {'fun_name': cb_lsqfun_name, 'returned': 'fvec'},
            _fvec_lsqfun,
            copy=False,
            exp_shape=(m,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        return _iflag_lsqfun_base

    def cb_lsqmon(m, n, xc, fvec, fjac, s, igrade, niter, nf, data=None):
        """Wrapping user's lsqmon. Called as base lsqmon."""
        if data is not None:
            lsqmon(
                xc,
                fvec,
                fjac,
                s,
                igrade,
                niter,
                nf,
                data,
            )
        else:
            lsqmon(
                xc,
                fvec,
                fjac,
                s,
                igrade,
                niter,
                nf,
            )

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b_utils._check_callable(lsqfun, {'fun_name': fun_name, 'entity': 'lsqfun'})
    _lsqfun = cb_lsqfun
    _b_utils._check_callable(lsqmon, {'fun_name': fun_name, 'entity': 'lsqmon'}, allow_none=True)
    _lsqmon = None if lsqmon is None else cb_lsqmon
    _maxcal = 400*_n_py if maxcal is None else maxcal
    if _n_py == 1:
        _eta = 0.0
    else:
        _eta = 0.5
    _eta = _eta if eta is None else eta
    _m = _b_utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = _b_utils._EngineIntScalarType.to_py(_m)
    _fvec = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_m_py,),
    )
    _fjac = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_m_py, _n_py),
        sorder=_b_utils._EngineData.default_sorder,
    )
    _s = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _v = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py, _n_py),
        sorder=_b_utils._EngineData.default_sorder,
    )
    try:
        _return_scalars = _b_opt.lsq_uncon_mod_func_comp(
            _m,
            _n,
            _lsqfun,
            _lsqmon,
            iprint,
            _maxcal,
            _eta,
            xtol,
            stepmx,
            _x,
            _fvec,
            _fjac,
            _s,
            _v,
            data,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _x_py = _x
    _fsumsq_py = _return_scalars.fsumsq
    _fvec_py = _fvec
    _fjac_py = _fjac
    _s_py = _s
    _v_py = _v
    _niter_py = _return_scalars.niter
    _nf_py = _return_scalars.nf
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'fsumsq', 'fvec', 'fjac', 's', 'v', 'niter', 'nf'],
    )(
        _x_py,
        _fsumsq_py,
        _fvec_py,
        _fjac_py,
        _s_py,
        _v_py,
        _niter_py,
        _nf_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def handle_solve_dfls(handle, objfun, x, nres, monit=None, data=None, io_manager=None):
    u"""
    Derivative-free (DFO) solver for a nonlinear least squares objective
    function with bounded variables.

    Note: this function uses optional algorithmic parameters, see also:
    ``handle_opt_set``, ``handle_opt_get``.

    ``handle_solve_dfls`` is a forward communication Derivative-free
    Optimization (DFO) solver from the NAG optimization modelling suite
    (DFLS) for small to medium-scale nonlinear least squares problems
    with bound constraints.

    For full information please refer to the NAG Library document for
    e04ff

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04fff.html

    Parameters
    ----------
    handle : Handle
        The handle to the problem. It needs to be initialized by
        ``handle_init`` and **must not** be changed before the call to
        ``handle_solve_dfls``.

    objfun : callable rx = objfun(x, nres, data=None)
        `objfun` must evaluate the value of the nonlinear residuals
        r_i(x) at a specified point x.

        Parameters
        ~~~~~~~~~~
        x : float, ndarray, shape (nvar)
            x, the vector of variable values at which the residuals r_i
            are to be evaluated.

        nres : int
            m_r, the number of residuals in the problem, as set during
            the initialization of the handle by ``handle_set_nlnls``.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        rx : float, array-like, shape (`nres`)
            The value of the residuals r_i(x) at x.

    x : float, array-like, shape (nvar)
        x_0, the initial estimates of the variables x.

    nres : int
        m_r, the number of residuals in the problem. It must be
        unchanged from the value set during the definition of the
        objective structure by ``handle_set_nlnls``.

    monit : None or callable monit(x, rinfo, stats, data=None), optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `monit` is provided to enable you to monitor the progress of the
        optimization and, if necessary, to halt the optimization
        process.

        If no monitoring is required, `monit` may be None.

        `monit` is called at the end of every i^th step where i is
        controlled by the option 'DFO Monitor Frequency' (default value
        0, `monit` is never called).

        Parameters
        ~~~~~~~~~~
        x : float, ndarray, shape (nvar)
            The current best point.

        rinfo : float, ndarray, shape (100)
            Best objective value computed and various indicators (the
            values are as described in the main argument `rinfo`).

        stats : float, ndarray, shape (100)
            Solver statistics at monitoring steps or at the end of the
            current iteration (the values are as described in the main
            argument `stats`).

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

    data : arbitrary, optional
        User-communication data for callback functions.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    x : float, ndarray, shape (nvar)
        The final values of the variables x.

    rx : float, ndarray, shape (`nres`)
        The values of the residuals at the final point given in `x`.

    rinfo : float, ndarray, shape (100)
        Optimal objective value and various indicators at monitoring
        steps or at the end of the final iteration. The measures are
        given in the table below:

        [table omitted]

    stats : float, ndarray, shape (100)
        Solver statistics at monitoring steps or at the end of the final
        iteration as given in the table below:

        [table omitted]

    Other Parameters
    ----------------
    'Defaults' : valueless
        This special keyword may be used to reset all options to their
        default values.
        Any value given with this keyword will be ignored.

    'DFLS Small Residuals Tol' : float
        Default = epsilon^0.75

        This option defines the tolerance on the value of the residuals.
        Namely, the solver declares convergence if

            f(x) = sum_i = 1^m_rr_i(x)^2 < 'DFLS Small Residuals Tol'.

        Constraint: 'DFLS Small Residuals Tol' > epsilon^2.

    'DFO Initial Interp Points' : str
        Default = 'Coordinate'

        Determines how the initial interpolation points are chosen.
        If 'DFO Initial Interp Points' = 'Coordinate', the interpolation
        points are chosen along the coordinate directions around the
        initial point.
        If 'DFO Initial Interp Points' = 'Random', the initial
        interpolation points are chosen along random orthogonal
        directions around the initial point.
        Set 'DFO Random Seed' to a positive value to fix the random seed
        and get reproducible results.

        Constraint: 'DFO Initial Interp Points' = 'Coordinate' or
        'Random'.

    'DFO Maximum Slow Steps' : int
        Default = 20

        If 'DFO Maximum Slow Steps' > 0, this argument defines the
        maximum number of consecutive slow iterations n_slow allowed.
        Set 'DFO Maximum Slow Steps' = 0 to deactivate the slow
        iteration detection.
        The algorithm can stop in two situations:

        (i) n_slow > 'DFO Maximum Slow Steps' and rho < 'DFO Trust
            Region Slow Tol' with `errno` = 50,

        (#) n_slow > 5*'DFO Maximum Slow Steps' with `errno` = 24.

        Constraint: 'DFO Maximum Slow Steps' >= 0.

    'DFO Max Objective Calls' : int
        Default = 500

        A limit on the number of objective function evaluations the
        solver is allowed to compute.
        If the limit is reached, the solver stops with `errno` = 21.

        Constraint: 'DFO Max Objective Calls' >= 1.

    'DFO Max Soft Restarts' : int
        Default = 5

        The maximum total number of soft restarts that can be performed
        if the objective function is declared as noisy ('DFO Noisy
        Problem' = 'YES').

        Constraint: 'DFO Max Soft Restarts' >= 1.

    'DFO Max Unsucc Soft Restarts' : int
        Default = 3

        The maximum number of consecutive unsuccessful soft restarts
        that can be performed if the objective function is declared as
        noisy ('DFO Noisy Problem' = 'YES').

        Constraint: 'DFO Max Unsucc Soft Restarts' >= 1.

    'DFO Monitor Frequency' : int
        Default = 0

        If 'DFO Monitor Frequency' > 0, `monit` will be called at the
        end of every ith step for monitoring purposes.

        Constraint: 'DFO Monitor Frequency' >= 0.

    'DFO Noise Level' : float
        Default = 0.0

        Indicates the noise level expected when evaluating the objective
        function if 'DFO Noisy Problem' = 'YES'.

        Constraint: 'DFO Noise Level' >= 0.0.

    'DFO Noisy Problem' : str
        Default = 'NO'

        Indicates if the function evaluations provided to the solver are
        noisy.
        If 'DFO Noisy Problem' = 'YES', some algorithmic features will
        be activated:

        (i) The trust region update becomes slower to reflect the
            decreased confidence in the objective values.

        (#) Soft restarts of the algorithm can be performed to ensure
            the algorithm did not get stuck because of the noise (see
            'DFO Max Soft Restarts', 'DFO Max Unsucc Soft Restarts',
            'DFO Number Soft Restarts Pts' to control the restart
            characteristics).

        (#) In addition, if 'DFO Noise Level' > 0.0, the solver will
            trigger a soft restart if all the function values are within
            the noise level.

    'DFO Number Initial Points' : int
        Default = 0

        The initial number of interpolation points in Y_0 (9) used to
        build the linear models of the residuals.
        If 'DFO Number Initial Points' = 0, the number of points is
        chosen to be equal to the total number of interpolation points
        set by 'DFO Number Interp Points'.

        If this parameter is chosen to be lower than the maximum set by
        'DFO Number Interp Points', the solver will progressively
        increase the number of interpolation points until it reaches
        that value.
        In this release, it is only possible to grow the interpolation
        set if 'DFO Number Interp Points' is set to the default value.

        Constraint: 'DFO Number Initial Points' >= 0.

        Consistency constraint, the solver stops with `errno` = 6 if not
        met:

            0 <= 'DFO Number Initial Points' <= 'DFO Number Interp
            Points'.

            If 'DFO Number Initial Points' < 'DFO Number Interp Points',
            'DFO Number Interp Points' must be set to the default value.

    'DFO Number Interp Points' : int
        Default = 0

        The maximum number of interpolation points in Y_k (9) used to
        build the linear models of the residuals.
        If 'DFO Number Interp Points' = 0, the number of points is
        chosen to be n_r+1 where n_r is the number of non-fixed
        variables.

        Constraint: 'DFO Number Interp Points' >= 0.

        Consistency constraint, the solver stops with `errno` = 6 if not
        met:

            n_r+1 <= 'DFO Number Interp Points' <= (n_r+1)*(n_r+2)/2.

    'DFO Number Soft Restarts Pts' : int
        Default = 3

        The number of interpolation points that are replaced during a
        soft restart.

        Constraint: 'DFO Number Soft Restarts Pts' >= 1.

    'DFO Print Frequency' : int
        Default = 1

        If 'DFO Print Frequency' > 0, the solver prints the iteration
        log to the appropriate units at the end of every ith step.

        Constraint: 'DFO Print Frequency' >= 0.

    'DFO Random Seed' : int
        Default = -1

        The random seed used to generate the random points used to build
        the initial model or build the underdetermined models when the
        interpolation set has not fully grown ('DFO Number Initial
        Points' < 'DFO Number Interp Points').
        If 'DFO Random Seed' < 0, the random seed will be based on
        values taken from the real-time clock, potentially resulting in
        the solver taking a different path each time it is run.
        Set it to a positive value to get fully reproducible runs.

        Constraint: 'DFO Print Frequency' >= -1.

    'DFO Starting Trust Region' : float
        Default = 0.1

        rho_beg, the initial trust region radius.
        This argument should be set to about one tenth of the greatest
        expected overall change to a variable: the initial quadratic
        model will be constructed by taking steps from the initial x of
        length rho_beg along each coordinate direction.
        The default value assumes that the variables have an order of
        magnitude 1.

        Constraint: 'DFO Starting Trust Region' > epsilon.

        Consistency constraints, the solver stops with `errno` = 5 if
        not met:

            'DFO Starting Trust Region' <= 'DFO Trust Region Tolerance'.

            'DFO Starting Trust Region' <= 1/2min_i(u_x(i)-l_x(i)).

    'DFO Trust Region Slow Tol' : float
        Default = epsilon^0.25

        The minimal acceptable trust region radius for the solution to
        be declared as acceptable.
        The solver stops if:

            n_slow > 'DFO Maximum Slow Steps' and rho_k < 'DFO Trust
            Region Slow Tol'.

        Constraint: 'DFO Trust Region Slow Tol' > epsilon.

        Consistency constraint, the solver stops with `errno` = 5 if not
        met:

            'DFO Trust Region Slow Tol' > 'DFO Trust Region Tolerance'.

    'DFO Trust Region Tolerance' : float
        Default = epsilon^0.37

        rho_end, the requested trust region radius.
        The algorithm declares convergence when the trust region radius
        reaches this limit.
        It should indicate the absolute accuracy that is required in the
        final values of the variables.

        Constraint: 'DFO Trust Region Tolerance' > epsilon.

        Consistency constraints, the solver stops with `errno` = 5 if
        not met:

            'DFO Starting Trust Region' > 'DFO Trust Region Tolerance'.

            'DFO Trust Region Slow Tol' > 'DFO Trust Region Tolerance'.

    'DFO Version' : str
        Default = 'Latest'

        At Mark 27, the underlying algorithm of ``handle_solve_dfls``
        underwent significant changes.
        This option allows you to continue using the Mark 26 version if
        it is set to ''26''.
        By default (recommended), the latest version of the code is
        called.

        Constraint: 'DFO Version' = 'Latest' or '26'.

    'Infinite Bound Size' : float
        Default = 10^20

        This defines the 'infinite' bound bigbnd in the definition of
        the problem constraints.
        Any upper bound greater than or equal to bigbnd will be regarded
        as + infinity (and similarly any lower bound less than or equal
        to -bigbnd will be regarded as - infinity).
        Note that a modification of this option does not influence
        constraints which have already been defined; only the
        constraints formulated after the change will be affected.

        Constraint: 'Infinite Bound Size' >= 1000.

    'Monitoring File' : int
        Default = -1

        If i >= 0, the unit number for the secondary (monitoring)
        output.
        If 'Monitoring File' = -1, no secondary output is provided.
        The information output to this unit is controlled by 'Monitoring
        Level'.

        Constraint: 'Monitoring File' >= -1.

    'Monitoring Level' : int
        Default = 4

        This argument sets the amount of information detail that will be
        printed by the solver to the secondary output.
        The meaning of the levels is the same as with 'Print Level'.

        Constraint: 0 <= 'Monitoring Level' <= 5.

    'Print File' : int
        Default = advisory message unit number

        If i >= 0, the unit number for the primary output of the solver.
        If 'Print File' = -1, the primary output is completely turned
        off independently of other settings. The default value is the
        advisory message unit number at the time of the options
        initialization, e.g., at the initialization of the handle. The
        information output to this unit is controlled by 'Print Level'.

        Constraint: 'Print File' >= -1.

    'Print Level' : int
        Default = 2

        This argument defines how detailed information should be printed
        by the solver to the primary and secondary output.

        [table omitted]

        Constraint: 0 <= 'Print Level' <= 5.

    'Print Options' : str
        Default = 'YES'

        If 'Print Options' = 'YES', a listing of options will be printed
        to the primary output and is always printed to the secondary
        output.

        Constraint: 'Print Options' = 'YES' or 'NO'.

    'Print Solution' : str
        Default = 'NO'

        If 'Print Solution' = 'YES', the solution will be printed to the
        primary and secondary output.

        Constraint: 'Print Solution' = 'YES' or 'NO'.

    'Stats Time' : str
        Default = 'NO'

        This argument turns on timings of various parts of the algorithm
        to give a better overview of where most of the time is spent.
        This might be helpful for a choice of different solving
        approaches.
        It is possible to choose between CPU and wall clock time.
        Choice 'YES' is equivalent to 'WALL CLOCK'.

        Constraint: 'Stats Time' = 'YES', 'NO', 'CPU' or 'WALL CLOCK'.

    'Time Limit' : float
        Default = 10^6

        A limit to the number of seconds that the solver can use to
        solve one problem.
        If during the convergence check this limit is exceeded, the
        solver will terminate with `errno` = 23.

        Constraint: 'Time Limit' > 0.

    Raises
    ------
    NagValueError
        (`errno` 1)
            `handle` has not been initialized.

        (`errno` 1)
            `handle` does not belong to the NAG optimization modelling
            suite, has not been initialized properly or is corrupted.

        (`errno` 1)
            `handle` has not been initialized properly or is corrupted.

        (`errno` 2)
            This solver does not support the model defined in the
            handle.

        (`errno` 2)
            The problem is already being solved.

        (`errno` 4)
            The information supplied does not match with that previously
            stored.

            On entry, nvar = *<value>* must match that given during
            initialization of the `handle`, i.e., *<value>*.

        (`errno` 4)
            The information supplied does not match with that previously
            stored.

            On entry, `nres` = *<value>* must match that given during
            the definition of the objective in the `handle`, i.e.,
            *<value>*.

        (`errno` 5)
            Inconsistent options 'DFO Trust Region Tolerance' rho_end
            and 'DFO Starting Trust Region' rho_beg.

            Constraint: rho_end < rho_beg.

            Use ``handle_opt_set`` to set compatible option values.

        (`errno` 5)
            Inconsistent options 'DFO Trust Region Tolerance' rho_end
            and 'DFO Trust Region Slow Tol' rho_tol.

            Constraint: rho_end < rho_tol.

            Use ``handle_opt_set`` to set compatible option values.

        (`errno` 5)
            Option 'DFO Starting Trust Region' rho_beg = *<value>*,
            l_x(i) = *<value>*, u_x(i) = *<value>* and i = *<value>*.

            Constraint: if l_x(i) != u_x(i) in coordinate i, then
            u_x(i)-l_x(i) >= 2*rho_beg.

            Use ``handle_opt_set`` to set compatible option values.

        (`errno` 6)
            There were n_r = *<value>* unequal bounds and the option
            'DFO Number Interp Points' npt = *<value>*.

            Constraint: n_r+1 <= npt <= (n_r+1)*(n_r+2)/2.

            Use ``handle_opt_set`` to set compatible option values.

        (`errno` 6)
            The number of initial interpolation points is greater than
            the maximum.

            Use 'DFO Number Interp Points' and 'DFO Number Initial
            Points' to control the number of interpolation points.

        (`errno` 6)
            Initial number of interpolation points ninit = *<value>*,
            total number of interpolation points npts = *<value>*,
            number of variables nvar = *<value>*.

            Constraint: growing interpolation set is only supported for
            linear models (npts = nvar+1).

            Use 'DFO Number Interp Points' and 'DFO Number Initial
            Points' to control the number of interpolation points.

        (`errno` 8)
            `nres` = 0. There are no residuals, the objective function
            is empty.

        (`errno` 8)
            Maximization is not possible for a nonlinear least squares
            problem.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 50)
            The problem was solved to an acceptable level after
            *<value>* consecutive slow iterations.

            Use the option 'DFO Maximum Slow Steps' to modify the
            maximum number of slow steps accepted.

    NagAlgorithmicMajorWarning
        (`errno` 17)
            Rescue failed: the trust region could not be reduced further
            after some function evaluation could not be provided. Check
            the specification of your objective and whether it needs
            rescaling. Try a different initial `x`.

        (`errno` 17)
            Some initial interpolation points were not provided. Rescue
            cannot be attempted at this stage.

            Check the specification of your objective and whether it
            needs rescaling. Try a different initial `x`.

        (`errno` 18)
            The predicted reduction in a trust region step was
            non-positive. Check your specification of `objfun` and
            whether the function needs rescaling. Try a different
            initial `x`.

        (`errno` 19)
            A rescue procedure has been called in order to correct
            damage from rounding errors when computing an update to a
            quadratic approximation of F, but no further progress could
            be made. Check your specification of `objfun` and whether
            the function needs rescaling. Try a different initial `x`.

        (`errno` 21)
            Maximum number of function evaluations exceeded.

        (`errno` 23)
            The solver terminated after the maximum time allowed was
            exceeded.

        (`errno` 24)
            No progress, the solver was stopped after *<value>*
            consecutive slow steps.

            Use the option 'DFO Maximum Slow Steps' to modify the
            maximum number of slow steps accepted.

    NagCallbackTerminateWarning
        (`errno` 20)
            User requested termination during a monitoring step.

        (`errno` 20)
            User requested termination after a call to the objective
            function.

    References
    ----------
    Cartis, C, Fiala, J, Marteau, B and Roberts, L, 2018, `Improving the
    Flexibility and Robustness of Model-Based Derivative-Free
    Optimization Solvers`, Technical Report, University of Oxford

    Cartis, C and Roberts, L, 2017, `A Derivative-Free Gauss-Newton
    Method`

    Conn, A R, Scheinberg, K and Vicente, L N, 2009, `Introduction to
    Derivative-Free Optimization, vol. 8 of MPS-SIAM Series on
    Optimization`, MPS/SIAM, Philadelphia

    Powell, M J D, 2009, `The BOBYQA algorithm for bound constrained
    optimization without derivatives`, Report DAMTP 2009/NA06,
    University of Cambridge,
    https://www.damtp.cam.ac.uk/user/na/NA_papers/NA2009_06.pdf

    Zhang, H, Conn, A R and Scheinberg, K, 2010, `A Derivative-Free
    Algorithm for Least-Squares Minimization`, SIAM J. Optim. (20(6)),
    3555--3576

    See Also
    --------
    :meth:`naginterfaces.library.examples.opt.handle_solve_dfls_ex.main`
    """

    fun_name = 'naginterfaces.library.opt.handle_solve_dfls'

    cb_objfun_name = fun_name + ':objfun'

    def cb_objfun(nvar, x, nres, rx, data=None):
        """Wrapping user's objfun. Called as base objfun."""
        if data is not None:
            _rx_objfun = objfun(
                x,
                nres,
                data,
            )
        else:
            _rx_objfun = objfun(
                x,
                nres,
            )
        rx[:] = _utils._to_ndarray(
            {'fun_name': cb_objfun_name, 'returned': 'rx'},
            _rx_objfun,
            copy=False,
            exp_shape=(nres,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )

    def cb_monit(nvar, x, rinfo, stats, data=None):
        """Wrapping user's monit. Called as base monit."""
        if data is not None:
            monit(
                x,
                rinfo,
                stats,
                data,
            )
        else:
            monit(
                x,
                rinfo,
                stats,
            )

    _b_utils._check_callable(objfun, {'fun_name': fun_name, 'entity': 'objfun'})
    _objfun = cb_objfun
    _b_utils._check_callable(monit, {'fun_name': fun_name, 'entity': 'monit'}, allow_none=True)
    _monit = None if monit is None else cb_monit
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nvar_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _nvar = _b_utils._EngineIntScalarType(
        _nvar_py, {'fun_name': fun_name, 'entity': '_nvar'},
    ).to_ctypes()
    _nres = _b_utils._EngineIntScalarType(
        nres, {'fun_name': fun_name, 'entity': 'nres'},
    ).to_ctypes()
    _nres_py = _b_utils._EngineIntScalarType.to_py(_nres)
    _rx = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_nres_py,),
    )
    _rinfo = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (100,),
    )
    _stats = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (100,),
    )
    try:
        _b_opt.handle_solve_dfls(
            handle,
            _objfun,
            _monit,
            _nvar,
            _x,
            _nres,
            _rx,
            _rinfo,
            _stats,
            data,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _x_py = _x
    _rx_py = _rx
    _rinfo_py = _rinfo
    _stats_py = _stats
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'rx', 'rinfo', 'stats'],
    )(
        _x_py,
        _rx_py,
        _rinfo_py,
        _stats_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def handle_solve_dfls_rcomm(handle, irevcm, x, rx, io_manager=None):
    u"""
    Reverse communication derivative-free (DFO) solver for a nonlinear
    least squares objective function with bounded variables.

    Note: this function uses optional algorithmic parameters, see also:
    ``handle_opt_set``, ``handle_opt_get``.

    ``handle_solve_dfls_rcomm`` is a reverse communication
    Derivative-free Optimization (DFO) solver from the NAG optimization
    modelling suite (DFLS) for small to medium-scale nonlinear least
    squares problems with bound constraints.

    For full information please refer to the NAG Library document for
    e04fg

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04fgf.html

    Parameters
    ----------
    handle : Handle
        The handle to the problem. It needs to be initialized by
        ``handle_init`` and **must not** be changed before the call to
        ``handle_solve_dfls_rcomm``.

    irevcm : int
        Does not need to be set on the first call of
        ``handle_solve_dfls_rcomm``. On subsequent calls, `irevcm` must
        be set to a positive integer if all the required function
        evaluations have been correctly provided in `rx`. Otherwise, if
        a problem occurred during a monitoring step or while providing
        objective values, it is possible to set it to a negative value:

        `irevcm` = -1

            If function evaluations were required, the solver will
            attempt a rescue procedure and request an alternative point.
            If no function were required (monitoring step), the solver
            will stop with `errno` = 20.

        `irevcm` <= -2

            The solver will cleanly exit and return the best available
            point as a well as the solve statistics.

    x : float, array-like, shape (nvar, maxeval)
        The first column contains x_0, the initial estimates of the
        variables x.

    rx : float, array-like, shape (nres, maxeval)
        Does not need to be set on the first call to
        ``handle_solve_dfls_rcomm``.

        If `irevcm` = 1 after the last call of
        ``handle_solve_dfls_rcomm``, the first `neval` columns must
        contain the residuals of the requested points.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    irevcm : int
        Indicates what action is to be performed before the next call to
        ``handle_solve_dfls_rcomm``.

        `irevcm` = 0

            Final exit of the solver.

        `irevcm` = 1

            `neval` objective evaluations are required.

        `irevcm` = 2

            Monitoring step, no evaluation is required, `x` and `rx`
            contain the best evaluation of the objective yet.

    neval : int
        Indicates the number of objective evaluations required for the
        next call of ``handle_solve_dfls_rcomm`` in `rx`. The
        coordinates of the points to evaluate are provided in the first
        `neval` columns of `x`.

    x : float, ndarray, shape (nvar, maxeval)
        If `irevcm` = 0 or 2, the first column contains the best
        computed estimate of the solution.

        If `irevcm` = 1, the first `neval` columns contain the
        coordinates of the points to evaluate.

    rx : float, ndarray, shape (nres, maxeval)
        If `irevcm` = 0 or 2, the first column contains the residuals of
        the best computed point.

    rinfo : float, ndarray, shape (100)
        Optimal objective value and various indicators at monitoring
        steps or at the end of the final iteration. The measures are
        given in the table below:

        [table omitted]

    stats : float, ndarray, shape (100)
        Solver statistics at monitoring steps or at the end of the final
        iteration as given in the table below:

        [table omitted]

    Other Parameters
    ----------------
    'Defaults' : valueless
        This special keyword may be used to reset all options to their
        default values.
        Any value given with this keyword will be ignored.

    'DFLS Small Residuals Tol' : float
        Default = epsilon^0.75

        This option defines the tolerance on the value of the residuals.
        Namely, the solver declares convergence if

            f(x) = sum_i = 1^m_rr_i(x)^2 < 'DFLS Small Residuals Tol'.

        Constraint: 'DFLS Small Residuals Tol' > epsilon^2.

    'DFO Initial Interp Points' : str
        Default = 'Coordinate'

        Determines how the initial interpolation points are chosen.
        If 'DFO Initial Interp Points' = 'Coordinate', the interpolation
        points are chosen along the coordinate directions around the
        initial point.
        If 'DFO Initial Interp Points' = 'Random', the initial
        interpolation points are chosen along random orthogonal
        directions around the initial point.
        Set 'DFO Random Seed' to a positive value to fix the random seed
        and get reproducible results.

        Constraint: 'DFO Initial Interp Points' = 'Coordinate' or
        'Random'.

    'DFO Maximum Slow Steps' : int
        Default = 20

        If 'DFO Maximum Slow Steps' > 0, this argument defines the
        maximum number of consecutive slow iterations n_slow allowed.
        Set 'DFO Maximum Slow Steps' = 0 to deactivate the slow
        iteration detection.
        The algorithm can stop in two situations:

        (i) n_slow > 'DFO Maximum Slow Steps' and rho < 'DFO Trust
            Region Slow Tol' with `errno` = 50,

        (#) n_slow > 5*'DFO Maximum Slow Steps' with `errno` = 24.

        Constraint: 'DFO Maximum Slow Steps' >= 0.

    'DFO Max Objective Calls' : int
        Default = 500

        A limit on the number of objective function evaluations the
        solver is allowed to compute.
        If the limit is reached, the solver stops with `errno` = 21.

        Constraint: 'DFO Max Objective Calls' >= 1.

    'DFO Max Soft Restarts' : int
        Default = 5

        The maximum total number of soft restarts that can be performed
        if the objective function is declared as noisy ('DFO Noisy
        Problem' = 'YES').

        Constraint: 'DFO Max Soft Restarts' >= 1.

    'DFO Max Unsucc Soft Restarts' : int
        Default = 3

        The maximum number of consecutive unsuccessful soft restarts
        that can be performed if the objective function is declared as
        noisy ('DFO Noisy Problem' = 'YES').

        Constraint: 'DFO Max Unsucc Soft Restarts' >= 1.

    'DFO Monitor Frequency' : int
        Default = 0

        If 'DFO Monitor Frequency' > 0, the solver will stop at the end
        of every ith step for monitoring purposes.
        ``handle_solve_dfls_rcomm`` needs to be called again to continue
        the optimization.

        Constraint: 'DFO Monitor Frequency' >= 0.

    'DFO Noise Level' : float
        Default = 0.0

        Indicates the noise level expected when evaluating the objective
        function.
        If 'DFO Noisy Problem' = 'YES'.

        Constraint: 'DFO Noise Level' >= 0.0.

    'DFO Noisy Problem' : str
        Default = 'NO'

        Indicates if the function evaluations provided to the solver are
        noisy.
        If 'DFO Noisy Problem' = 'YES', some algorithmic features will
        be activated:

        (i) The trust region update becomes slower to reflect the
            decreased confidence in the objective values.

        (#) Soft restarts of the algorithm can be performed to ensure
            the algorithm did not get stuck because of the noise (see
            'DFO Max Soft Restarts', 'DFO Max Unsucc Soft Restarts',
            'DFO Number Soft Restarts Pts' to control the restart
            characteristics).

        (#) In addition, if 'DFO Noise Level' > 0.0, the solver will
            trigger a soft restart if all the function values are within
            the noise level.

    'DFO Number Initial Points' : int
        Default = 0

        The initial number of interpolation points in Y_0 (1) used to
        build the linear models of the residuals.
        If 'DFO Number Initial Points' = 0, the number of points is
        chosen to be equal to the total number of interpolation points
        set by 'DFO Number Interp Points'.

        If this parameter is chosen to be lower than the maximum set by
        'DFO Number Interp Points', the solver will progressively
        increase the number of interpolation points until it reaches
        that value.
        In this release, it is only possible to grow the interpolation
        set if 'DFO Number Interp Points' is set to the default value.

        Constraint: 'DFO Number Initial Points' >= 0.

        Consistency constraints, the solver stops with `errno` = 6 if
        not met:

            0 <= 'DFO Number Initial Points' <= 'DFO Number Interp
            Points'.

            If 'DFO Number Initial Points' < 'DFO Number Interp Points',
            'DFO Number Interp Points' must be set to the default value.

    'DFO Number Interp Points' : int
        Default = 0

        The maximum number of interpolation points in Y_k (1) used to
        build the linear models of the residuals.
        If 'DFO Number Interp Points' = 0, the number of points is
        chosen to be n_r+1 where n_r is the number of non-fixed
        variables.

        Constraint: 'DFO Number Interp Points' >= 0.

        Consistency constraint, the solver stops with `errno` = 6 if not
        met:

            n_r+1 <= 'DFO Number Interp Points' <= (n_r+1)*(n_r+2)/2.

    'DFO Number Soft Restarts Pts' : int
        Default = 3

        The number of interpolation points that are replaced during a
        soft restart.

        Constraint: 'DFO Number Soft Restarts Pts' >= 1.

    'DFO Print Frequency' : int
        Default = 1

        If 'DFO Print Frequency' > 0, the solver prints the iteration
        log to the appropriate units at the end of every ith step.

        Constraint: 'DFO Print Frequency' >= 0.

    'DFO Random Seed' : int
        Default = -1

        The random seed used to generate the random points used to build
        the initial model or build the underdetermined models when the
        interpolation set has not fully grown ('DFO Number Initial
        Points' < 'DFO Number Interp Points').
        If 'DFO Random Seed' < 0, the random seed will be based on
        values taken from the real-time clock, potentially resulting in
        the solver taking a different path each time it is run.
        Set it to a positive value to get fully reproducible runs.

        Constraint: 'DFO Print Frequency' >= -1.

    'DFO Starting Trust Region' : float
        Default = 0.1

        rho_beg, the initial trust region radius.
        This argument should be set to about one tenth of the greatest
        expected overall change to a variable: the initial quadratic
        model will be constructed by taking steps from the initial x of
        length rho_beg along each coordinate direction.
        The default value assumes that the variables have an order of
        magnitude 1.

        Constraint: 'DFO Starting Trust Region' > epsilon.

        Consistency constraints, the solver stops with `errno` = 5 if
        not met:

            'DFO Starting Trust Region' <= 'DFO Trust Region Tolerance'.

            'DFO Starting Trust Region' <= 1/2min_i(u_x(i)-l_x(i)).

    'DFO Trust Region Slow Tol' : float
        Default = epsilon^0.25

        The minimal acceptable trust region radius for the solution to
        be declared as acceptable.
        The solver stops if:

            n_slow > 'DFO Maximum Slow Steps' and rho_k < 'DFO Trust
            Region Slow Tol'.

        Constraint: 'DFO Trust Region Slow Tol' > epsilon.

        Consistency constraint, the solver stops with `errno` = 5 if not
        met:

            'DFO Trust Region Slow Tol' > 'DFO Trust Region Tolerance'.

    'DFO Trust Region Tolerance' : float
        Default = epsilon^0.37

        rho_end, the requested trust region radius.
        The algorithm declares convergence when the trust region radius
        reaches this limit.
        It should indicate the absolute accuracy that is required in the
        final values of the variables.

        Constraint: 'DFO Trust Region Tolerance' > epsilon.

        Consistency constraints, the solver stops with `errno` = 5 if
        not met:

            'DFO Starting Trust Region' > 'DFO Trust Region Tolerance'.

            'DFO Trust Region Slow Tol' > 'DFO Trust Region Tolerance'.

    'Infinite Bound Size' : float
        Default = 10^20

        This defines the 'infinite' bound bigbnd in the definition of
        the problem constraints.
        Any upper bound greater than or equal to bigbnd will be regarded
        as + infinity (and similarly any lower bound less than or equal
        to -bigbnd will be regarded as - infinity).
        Note that a modification of this option does not influence
        constraints which have already been defined; only the
        constraints formulated after the change will be affected.

        Constraint: 'Infinite Bound Size' >= 1000.

    'Monitoring File' : int
        Default = -1

        If i >= 0, the unit number for the secondary (monitoring)
        output.
        If 'Monitoring File' = -1, no secondary output is provided.
        The information output to this unit is controlled by 'Monitoring
        Level'.

        Constraint: 'Monitoring File' >= -1.

    'Monitoring Level' : int
        Default = 4

        This argument sets the amount of information detail that will be
        printed by the solver to the secondary output.
        The meaning of the levels is the same as with 'Print Level'.

        Constraint: 0 <= 'Monitoring Level' <= 5.

    'Print File' : int
        Default = advisory message unit number

        If i >= 0, the unit number for the primary output of the solver.
        If 'Print File' = -1, the primary output is completely turned
        off independently of other settings. The default value is the
        advisory message unit number at the time of the options
        initialization, e.g., at the initialization of the handle. The
        information output to this unit is controlled by 'Print Level'.

        Constraint: 'Print File' >= -1.

    'Print Level' : int
        Default = 2

        This argument defines how detailed information should be printed
        by the solver to the primary and secondary output.

        [table omitted]

        Constraint: 0 <= 'Print Level' <= 5.

    'Print Options' : str
        Default = 'YES'

        If 'Print Options' = 'YES', a listing of options will be printed
        to the primary output and is always printed to the secondary
        output.

        Constraint: 'Print Options' = 'YES' or 'NO'.

    'Print Solution' : str
        Default = 'NO'

        If 'Print Solution' = 'YES', the solution will be printed to the
        primary and secondary output.

        Constraint: 'Print Solution' = 'YES' or 'NO'.

    'Stats Time' : str
        Default = 'NO'

        This argument turns on timings of various parts of the algorithm
        to give a better overview of where most of the time is spent.
        This might be helpful for a choice of different solving
        approaches.
        It is possible to choose between CPU and wall clock time.
        Choice 'YES' is equivalent to 'WALL CLOCK'.

        Constraint: 'Stats Time' = 'YES', 'NO', 'CPU' or 'WALL CLOCK'.

    'Time Limit' : float
        Default = 10^6

        A limit to the number of seconds that the solver can use to
        solve one problem.
        If during the convergence check this limit is exceeded, the
        solver will terminate with `errno` = 23.

        Constraint: 'Time Limit' > 0.

    Raises
    ------
    NagValueError
        (`errno` 1)
            `handle` has not been initialized.

        (`errno` 1)
            `handle` does not belong to the NAG optimization modelling
            suite, has not been initialized properly or is corrupted.

        (`errno` 1)
            `handle` has not been initialized properly or is corrupted.

        (`errno` 2)
            This solver does not support the model defined in the
            handle.

        (`errno` 2)
            The problem is already being solved.

        (`errno` 4)
            On entry, maxeval = *<value>*.

            Constraint: maxeval > 0.

        (`errno` 4)
            The information supplied does not match with that previously
            stored.

            On entry, nvar = *<value>* must match that given during
            initialization of the `handle`, i.e., *<value>*.

        (`errno` 4)
            The information supplied does not match with that previously
            stored.

            On entry, nres = *<value>* must match that given during the
            definition of the objective in the `handle`, i.e.,
            *<value>*.

        (`errno` 4)
            The information supplied does not match with that previously
            stored.

            On entry, maxeval = *<value>* must match that given during
            the first call of the function, i.e., *<value>*.

        (`errno` 5)
            Inconsistent options 'DFO Trust Region Tolerance' rho_end
            and 'DFO Starting Trust Region' rho_beg.

            Constraint: rho_end < rho_beg.

            Use ``handle_opt_set`` to set compatible option values.

        (`errno` 5)
            Inconsistent options 'DFO Trust Region Tolerance' rho_end
            and 'DFO Trust Region Slow Tol' rho_tol.

            Constraint: rho_end < rho_tol.

            Use ``handle_opt_set`` to set compatible option values.

        (`errno` 5)
            Option 'DFO Starting Trust Region' rho_beg = *<value>*,
            l_x(i) = *<value>*, u_x(i) = *<value>* and i = *<value>*.

            Constraint: if l_x(i) != u_x(i) in coordinate i, then
            u_x(i)-l_x(i) >= 2*rho_beg.

            Use ``handle_opt_set`` to set compatible option values.

        (`errno` 6)
            There were n_r = *<value>* unequal bounds and the option
            'DFO Number Interp Points' npt = *<value>*.

            Constraint: n_r+1 <= npt <= (n_r+1)*(n_r+2)/2.

            Use ``handle_opt_set`` to set compatible option values.

        (`errno` 6)
            The number of initial interpolation points is greater than
            the maximum.

            Use 'DFO Number Interp Points' and 'DFO Number Initial
            Points' to control the number of interpolation points.

        (`errno` 6)
            Initial number of interpolation points ninit = *<value>*,
            total number of interpolation points npts = *<value>*,
            number of variables nvar = *<value>*.

            Constraint: growing interpolation set is only supported for
            linear models (npts = nvar+1).

            Use 'DFO Number Interp Points' and 'DFO Number Initial
            Points' to control the number of interpolation points.

        (`errno` 8)
            nres = 0. There are no residuals, the objective function is
            empty.

        (`errno` 8)
            Maximization is not possible for a nonlinear least squares
            problem.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 50)
            The problem was solved to an acceptable level after
            *<value>* consecutive slow iterations.

            Use the option 'DFO Maximum Slow Steps' to modify the
            maximum number of slow steps accepted.

    NagAlgorithmicMajorWarning
        (`errno` 17)
            Rescue failed: the trust region could not be reduced further
            after some function evaluation could not be provided. Check
            the specification of your objective and whether it needs
            rescaling. Try a different initial `x`.

        (`errno` 17)
            Some initial interpolation points were not provided. Rescue
            cannot be attempted at this stage.

            Check the specification of your objective and whether it
            needs rescaling. Try a different initial `x`.

        (`errno` 18)
            The predicted reduction in a trust region step was
            non-positive. Check the specification of your objective and
            whether it needs rescaling. Try a different initial `x`.

        (`errno` 21)
            Maximum number of function evaluations exceeded.

        (`errno` 23)
            The solver terminated after the maximum time allowed was
            exceeded.

        (`errno` 24)
            No progress, the solver was stopped after *<value>*
            consecutive slow steps.

            Use the option 'DFO Maximum Slow Steps' to modify the
            maximum number of slow steps accepted.

    NagCallbackTerminateWarning
        (`errno` 20)
            User requested termination during a monitoring step.

        (`errno` 20)
            User requested termination during an objective evaluation
            step.

    References
    ----------
    Cartis, C, Fiala, J, Marteau, B and Roberts, L, 2018, `Improving the
    Flexibility and Robustness of Model-Based Derivative-Free
    Optimization Solvers`, Technical Report, University of Oxford

    Cartis, C and Roberts, L, 2017, `A Derivative-Free Gauss-Newton
    Method`

    Conn, A R, Scheinberg, K and Vicente, L N, 2009, `Introduction to
    Derivative-Free Optimization, vol. 8 of MPS-SIAM Series on
    Optimization`, MPS/SIAM, Philadelphia

    Powell, M J D, 2009, `The BOBYQA algorithm for bound constrained
    optimization without derivatives`, Report DAMTP 2009/NA06,
    University of Cambridge,
    https://www.damtp.cam.ac.uk/user/na/NA_papers/NA2009_06.pdf

    Zhang, H, Conn, A R and Scheinberg, K, 2010, `A Derivative-Free
    Algorithm for Least-Squares Minimization`, SIAM J. Optim. (20(6)),
    3555--3576
    """

    fun_name = 'naginterfaces.library.opt.handle_solve_dfls_rcomm'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _maxeval_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2)[1]
    _maxeval = _b_utils._EngineIntScalarType(
        _maxeval_py, {'fun_name': fun_name, 'entity': '_maxeval'},
    ).to_ctypes()
    _nvar_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2)[0]
    _nvar = _b_utils._EngineIntScalarType(
        _nvar_py, {'fun_name': fun_name, 'entity': '_nvar'},
    ).to_ctypes()
    _rx = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'rx'},
        rx,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nres_py = _b_utils._EngineFloat64ArrayType.get_shape(_rx, {'fun_name': fun_name, 'entity': 'rx'}, exp_rank=2)[0]
    _nres = _b_utils._EngineIntScalarType(
        _nres_py, {'fun_name': fun_name, 'entity': '_nres'},
    ).to_ctypes()
    _rinfo = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (100,),
    )
    _stats = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (100,),
    )
    try:
        _return_scalars = _b_opt.handle_solve_dfls_rcomm(
            handle,
            irevcm,
            _maxeval,
            _nvar,
            _x,
            _nres,
            _rx,
            _rinfo,
            _stats,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _irevcm_py = _return_scalars.irevcm
    _x_py = _x
    _rx_py = _rx
    _neval_py = _return_scalars.neval
    _rinfo_py = _rinfo
    _stats_py = _stats
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['irevcm', 'neval', 'x', 'rx', 'rinfo', 'stats'],
    )(
        _irevcm_py,
        _neval_py,
        _x_py,
        _rx_py,
        _rinfo_py,
        _stats_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def lsq_uncon_mod_func_easy(m, lsfun1, x, data=None):
    u"""
    Unconstrained minimum of a sum of squares, combined Gauss--Newton
    and modified Newton algorithm, using function values only
    (easy-to-use).

    ``lsq_uncon_mod_func_easy`` is an easy-to-use algorithm for finding
    an unconstrained minimum of a sum of squares of m nonlinear
    functions in n variables (m >= n).
    No derivatives are required.

    It is intended for functions which are continuous and which have
    continuous first and second derivatives (although it will usually
    work even if the derivatives have occasional discontinuities).

    For full information please refer to the NAG Library document for
    e04fy

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04fyf.html

    Parameters
    ----------
    m : int
        The number m of residuals, f_i(x), and the number n of
        variables, x_j.

    lsfun1 : callable fvec = lsfun1(m, xc, data=None)
        You must supply this function to calculate the vector of values
        f_i(x) at any point x.

        It should be tested separately before being used in conjunction
        with ``lsq_uncon_mod_func_easy`` (see [the E04 Introduction]).

        Parameters
        ~~~~~~~~~~
        m : int
            m, the numbers of residuals.

        xc : float, ndarray, shape (n)
            The point x at which the values of the f_i are required.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        fvec : float, array-like, shape (`m`)
            `fvec`[i-1] must contain the value of f_i at the point x,
            for i = 1,2,...,m.

    x : float, array-like, shape (n)
        `x`[j-1] must be set to a guess at the jth component of the
        position of the minimum, for j = 1,2,...,n.

    data : arbitrary, optional
        User-communication data for callback functions.

    Returns
    -------
    x : float, ndarray, shape (n)
        The lowest point found during the calculations. Thus, if no
        exception or warning is raised on exit, `x`[j-1] is the jth
        component of the position of the minimum.

    fsumsq : float
        The value of the sum of squares, F(x), corresponding to the
        final point stored in `x`.

    comm : dict, communication object
        Communication structure.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m` = *<value>* and n = *<value>*.

            Constraint: `m` >= n.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 2)
            There have been 400*n calls to `lsfun1`.

        (`errno` 4)
            Failure in computing SVD of estimated Jacobian matrix.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            The conditions for a minimum have not all been satisfied,
            but a lower point could not be found.

        (`errno` 5)
            It is probable that a local minimum has been found, but it
            cannot be guaranteed.

        (`errno` 6)
            It is possible that a local minimum has been found, but it
            cannot be guaranteed.

        (`errno` 7)
            It is unlikely that a local minimum has been found.

        (`errno` 8)
            It is very unlikely that a local minimum has been found.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Gill, P E and Murray, W, 1978, `Algorithms for the solution of the
    nonlinear least squares problem`, SIAM J. Numer. Anal. (15),
    977--992
    """

    fun_name = 'naginterfaces.library.opt.lsq_uncon_mod_func_easy'

    cb_lsfun1_name = fun_name + ':lsfun1'

    def cb_lsfun1(m, n, xc, fvec, data=None):
        """Wrapping user's lsfun1. Called as base lsfun1."""
        if data is not None:
            _fvec_lsfun1 = lsfun1(
                m,
                xc,
                data,
            )
        else:
            _fvec_lsfun1 = lsfun1(
                m,
                xc,
            )
        fvec[:] = _utils._to_ndarray(
            {'fun_name': cb_lsfun1_name, 'returned': 'fvec'},
            _fvec_lsfun1,
            copy=False,
            exp_shape=(m,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b_utils._check_callable(lsfun1, {'fun_name': fun_name, 'entity': 'lsfun1'})
    _lsfun1 = cb_lsfun1
    try:
        _return_scalars = _b_opt.lsq_uncon_mod_func_easy(
            m,
            _n,
            _lsfun1,
            _x,
            data,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _x_py = _x
    _fsumsq_py = _return_scalars.fsumsq
    _comm_py = _return_scalars.comm
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'fsumsq', 'comm'],
    )(
        _x_py,
        _fsumsq_py,
        _comm_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def lsq_uncon_quasi_deriv_comp(m, selct, lsqfun, maxcal, eta, xtol, x, lsqmon=None, iprint=1, stepmx=100000.0, data=None, spiked_sorder='C'):
    u"""
    Unconstrained minimum of a sum of squares, combined Gauss--Newton
    and quasi-Newton algorithm, using first derivatives (comprehensive).

    ``lsq_uncon_quasi_deriv_comp`` is a comprehensive quasi-Newton
    algorithm for finding an unconstrained minimum of a sum of squares
    of m nonlinear functions in n variables (m >= n).
    First derivatives are required.
    The function is intended for functions which have continuous first
    and second derivatives (although it will usually work even if the
    derivatives have occasional discontinuities).

    For full information please refer to the NAG Library document for
    e04gb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04gbf.html

    Parameters
    ----------
    m : int
        The number m of residuals, f_i(x), and the number n of
        variables, x_j.

    selct : int
        `selct` enables you to specify whether the linear minimizations
        (i.e., minimizations of F(x^(k)+alpha^(k)p^(k)) with respect to
        alpha^(k)) are to be performed by a function which just requires
        the evaluation of the f_i(x) (`selct` = 1), or by a function
        which also requires the first derivatives of the f_i(x) (`selct`
        = 2).

        It will often be possible to evaluate the first derivatives of
        the residuals in about the same amount of computer time that is
        required for the evaluation of the residuals themselves -- if
        this is so, then ``lsq_uncon_quasi_deriv_comp`` should be called
        with `selct` = 2.

        However, if the evaluation of the derivatives takes more than
        about 4 times as long as the evaluation of the residuals,
        `selct` = 1 will usually be preferable.

        If in doubt, use `selct` = 2 as it is slightly more robust.

    lsqfun : callable (iflag, fvec, fjac) = lsqfun(iflag, m, xc,
        data=None)
        `lsqfun` must calculate the vector of values f_i(x) and Jacobian
        matrix of first derivatives df_i/dx_j at any point x. (However,
        if you do not wish to calculate the residuals or first
        derivatives at a particular x, there is the option of setting an
        argument to cause ``lsq_uncon_quasi_deriv_comp`` to terminate
        immediately.)

        Parameters
        ~~~~~~~~~~
        iflag : int
            Will be set to 0, 1 or 2.

            `iflag` = 0

                Indicates that only the residuals need to be evaluated

            `iflag` = 1

                Indicates that only the Jacobian matrix needs to be
                evaluated

            `iflag` = 2

                Indicates that both the residuals and the Jacobian
                matrix must be calculated.

            If `selct` = 2, `lsqfun` will always be called with `iflag`
            set to 2.

        m : int
            m, the number of residuals.

        xc : float, ndarray, shape (n)
            The point x at which the values of the f_i and the df_i/dx_j
            are required.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        iflag : int
            If it is not possible to evaluate the f_i(x) or their first
            derivatives at the point given in `xc` (or if it is wished
            to stop the calculations for any other reason), you should
            reset `iflag` to some negative number and return control to
            ``lsq_uncon_quasi_deriv_comp``.
            ``lsq_uncon_quasi_deriv_comp`` will then terminate
            immediately, with errno set to your setting of `iflag`.

        fvec : float, array-like, shape (`m`)
            Unless `iflag` = 1 on entry, or `iflag` is reset to a
            negative number, `fvec`[i-1] must contain the value of f_i
            at the point x, for i = 1,2,...,m.

        fjac : float, array-like, shape (`m`, n)
            Unless `iflag` = 0 on entry, or `iflag` is reset to a
            negative number, `fjac`[i-1,j-1] must contain the value of
            df_i/dx_j at the point x, for j = 1,2,...,n, for i =
            1,2,...,m.

    maxcal : int
        Enables you to limit the number of times that `lsqfun` is called
        by ``lsq_uncon_quasi_deriv_comp``. There will be an error exit
        (see [Raises/Warns]) after `maxcal` calls of `lsqfun`.

    eta : float
        Every iteration of ``lsq_uncon_quasi_deriv_comp`` involves a
        linear minimization (i.e., minimization of
        F(x^(k)+alpha^(k)p^(k)) with respect to alpha^(k)). `eta`
        specifies how accurately these linear minimizations are to be
        performed. The minimum with respect to alpha^(k) will be located
        more accurately for small values of `eta` (say, 0.01) than for
        large values (say, 0.9).

        Although accurate linear minimizations will generally reduce the
        number of iterations performed by
        ``lsq_uncon_quasi_deriv_comp``, they will increase the number of
        calls of `lsqfun` made every iteration.

        On balance it is usually more efficient to perform a low
        accuracy minimization.

        Suggested value:

            `eta` = 0.9 if n > 1 and `selct` = 2,

            `eta` = 0.5 if n > 1 and `selct` = 1,

            `eta` = 0.0 if n = 1.

    xtol : float
        The accuracy in x to which the solution is required.

        If x_true is the true value of x at the minimum, then x_sol, the
        estimated position before a normal exit, is such that

            ||x_sol-x_true|| < `xtol`*(1.0+||x_true||),

        where ||y|| = sqrt(sum_j = 1^ny_j^2).

        For example, if the elements of x_sol are not much larger than
        1.0 in modulus and if `xtol` = 1.0e-5, then x_sol is usually
        accurate to about five decimal places. (For further details see
        [Accuracy].)

        If F(x) and the variables are scaled roughly as described in
        [Further Comments] and epsilon is the machine precision, then a
        setting of order `xtol` = sqrt(epsilon) will usually be
        appropriate.

        If `xtol` is set to 0.0 or some positive value less than 10
        epsilon, ``lsq_uncon_quasi_deriv_comp`` will use 10 epsilon
        instead of `xtol`, since 10 epsilon is probably the smallest
        reasonable setting.

    x : float, array-like, shape (n)
        `x`[j-1] must be set to a guess at the jth component of the
        position of the minimum, for j = 1,2,...,n.

    lsqmon : None or callable lsqmon(xc, fvec, fjac, s, igrade, niter,
        nf, data=None), optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        If `iprint` >= 0, you must supply `lsqmon` which is suitable for
        monitoring the minimization process. `lsqmon` must not change
        the values of any of its arguments.

        Parameters
        ~~~~~~~~~~
        xc : float, ndarray, shape (n)
            The coordinates of the current point x.

        fvec : float, ndarray, shape (m)
            The values of the residuals f_i at the current point x.

        fjac : float, ndarray, shape (m, n)
            `fjac`[i-1,j-1] contains the value of df_i/dx_j at the
            current point x, for j = 1,2,...,n, for i = 1,2,...,m.

        s : float, ndarray, shape (n)
            The singular values of the current Jacobian matrix. Thus `s`
            may be useful as information about the structure of your
            problem.

        igrade : int
            ``lsq_uncon_quasi_deriv_comp`` estimates the dimension of
            the subspace for which the Jacobian matrix can be used as a
            valid approximation to the curvature (see Gill and Murray
            (1978)). This estimate is called the grade of the Jacobian
            matrix, and `igrade` gives its current value.

        niter : int
            The number of iterations which have been performed in
            ``lsq_uncon_quasi_deriv_comp``.

        nf : int
            The number of evaluations of the residuals. (If `selct` = 2,
            `nf` is also the number of evaluations of the Jacobian
            matrix.)

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

    iprint : int, optional
        The frequency with which `lsqmon` is to be called.

        `iprint` > 0

            `lsqmon` is called once every `iprint` iterations and just
            before exit from ``lsq_uncon_quasi_deriv_comp``.

        `iprint` = 0

            `lsqmon` is just called at the final point.

        `iprint` < 0

            `lsqmon` is not called at all.

        `iprint` should normally be set to a small positive number.

    stepmx : float, optional
        An estimate of the Euclidean distance between the solution and
        the starting point supplied by you. (For maximum efficiency, a
        slight overestimate is preferable.)

        ``lsq_uncon_quasi_deriv_comp`` will ensure that, for each
        iteration,

            sum_j = 1^n(x_j^(k)-x_j^(k-1))^2 <= (`stepmx`)^2

        where k is the iteration number.

        Thus, if the problem has more than one solution,
        ``lsq_uncon_quasi_deriv_comp`` is most likely to find the one
        nearest to the starting point.

        On difficult problems, a realistic choice can prevent the
        sequence of x^(k) entering a region where the problem is
        ill-behaved and can help avoid overflow in the evaluation of
        F(x).

        However, an underestimate of `stepmx` can lead to inefficiency.

    data : arbitrary, optional
        User-communication data for callback functions.

    spiked_sorder : str, optional
        If `fjac` in `lsqfun` is spiked (i.e., has unit extent in all
        but one dimension, or has size 1), `spiked_sorder` selects the
        storage order to associate with it in the NAG Engine:

        spiked_sorder = 'C'
          row-major storage will be used;

        spiked_sorder = 'F'
          column-major storage will be used.

    Returns
    -------
    x : float, ndarray, shape (n)
        The final point x^(k). Thus, if no exception or warning is
        raised on exit, `x`[j-1] is the jth component of the estimated
        position of the minimum.

    fsumsq : float
        The value of F(x), the sum of squares of the residuals f_i(x),
        at the final point given in `x`.

    fvec : float, ndarray, shape (`m`)
        The value of the residual f_i(x) at the final point given in
        `x`, for i = 1,2,...,m.

    fjac : float, ndarray, shape (`m`, n)
        The value of the first derivative df_i/dx_j evaluated at the
        final point given in `x`, for j = 1,2,...,n, for i = 1,2,...,m.

    s : float, ndarray, shape (n)
        The singular values of the Jacobian matrix at the final point.
        Thus `s` may be useful as information about the structure of
        your problem.

    v : float, ndarray, shape (n, n)
        The matrix V associated with the singular value decomposition

            J = USV^T

        of the Jacobian matrix at the final point, stored by columns.
        This matrix may be useful for statistical purposes, since it is
        the matrix of orthonormalized eigenvectors of J^TJ.

    niter : int
        The number of iterations which have been performed in
        ``lsq_uncon_quasi_deriv_comp``.

    nf : int
        The number of times that the residuals have been evaluated
        (i.e., the number of calls of `lsqfun`). If `selct` = 2, `nf` is
        also the number of times that the Jacobian matrix has been
        evaluated.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `stepmx` = *<value>* and `xtol` = *<value>*.

            Constraint: `stepmx` >= `xtol`.

        (`errno` 1)
            On entry, `xtol` = *<value>*.

            Constraint: `xtol` >= 0.0.

        (`errno` 1)
            On entry, `eta` = *<value>*.

            Constraint: 0.0 <= `eta` < 1.0.

        (`errno` 1)
            On entry, `maxcal` = *<value>*.

            Constraint: `maxcal` >= 1.

        (`errno` 1)
            On entry, `m` = *<value>* and n = *<value>*.

            Constraint: `m` >= n.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 2)
            There have been `maxcal` = *<value>* calls to `lsqfun`.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i < 0)
            User requested termination by setting `iflag` negative in
            `lsqfun`.

        (`errno` 3)
            The conditions for a minimum have not all been satisfied,
            but a lower point could not be found.

    NagAlgorithmicMajorWarning
        (`errno` 4)
            Failure in computing SVD of Jacobian matrix.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Gill, P E and Murray, W, 1978, `Algorithms for the solution of the
    nonlinear least squares problem`, SIAM J. Numer. Anal. (15),
    977--992
    """

    fun_name = 'naginterfaces.library.opt.lsq_uncon_quasi_deriv_comp'

    cb_lsqfun_name = fun_name + ':lsqfun'

    def cb_lsqfun(iflag, m, n, xc, fvec, fjac, data=None):
        """Wrapping user's lsqfun. Called as base lsqfun."""
        if data is not None:
            _iflag_lsqfun, _fvec_lsqfun, _fjac_lsqfun = lsqfun(
                iflag,
                m,
                xc,
                data,
            )
        else:
            _iflag_lsqfun, _fvec_lsqfun, _fjac_lsqfun = lsqfun(
                iflag,
                m,
                xc,
            )
        _iflag_lsqfun_base = _iflag_lsqfun
        fvec[:] = _utils._to_ndarray(
            {'fun_name': cb_lsqfun_name, 'returned': 'fvec'},
            _fvec_lsqfun,
            copy=False,
            exp_shape=(m,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _fjac_lsqfun_np = _utils._to_ndarray(
            {'fun_name': cb_lsqfun_name, 'returned': 'fjac'},
            _fjac_lsqfun,
            copy=False,
            exp_shape=(m, n),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _utils._check_callback_sorder(
            {'fun_name': cb_lsqfun_name},
            ((_fjac_lsqfun_np, 'fjac'),),
            wrapper_sorder=_m_sorder,
            spiked_sorder=spiked_sorder,
        )
        fjac[:, :] = _fjac_lsqfun_np
        return _iflag_lsqfun_base

    def cb_lsqmon(m, n, xc, fvec, fjac, s, igrade, niter, nf, data=None):
        """Wrapping user's lsqmon. Called as base lsqmon."""
        if data is not None:
            lsqmon(
                xc,
                fvec,
                fjac,
                s,
                igrade,
                niter,
                nf,
                data,
            )
        else:
            lsqmon(
                xc,
                fvec,
                fjac,
                s,
                igrade,
                niter,
                nf,
            )

    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        (),
        spiked_sorder=spiked_sorder,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b_utils._check_callable(lsqfun, {'fun_name': fun_name, 'entity': 'lsqfun'})
    _lsqfun = cb_lsqfun
    _b_utils._check_callable(lsqmon, {'fun_name': fun_name, 'entity': 'lsqmon'}, allow_none=True)
    _lsqmon = None if lsqmon is None else cb_lsqmon
    _m = _b_utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = _b_utils._EngineIntScalarType.to_py(_m)
    _fvec = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_m_py,),
    )
    _fjac = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_m_py, _n_py),
        sorder=_b_utils._EngineData.default_sorder,
    )
    _s = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _v = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py, _n_py),
        sorder=_b_utils._EngineData.default_sorder,
    )
    try:
        _return_scalars = _b_opt.lsq_uncon_quasi_deriv_comp(
            _m,
            _n,
            selct,
            _lsqfun,
            _lsqmon,
            iprint,
            maxcal,
            eta,
            xtol,
            stepmx,
            _x,
            _fvec,
            _fjac,
            _s,
            _v,
            data,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _x_py = _x
    _fsumsq_py = _return_scalars.fsumsq
    _fvec_py = _fvec
    _fjac_py = _fjac
    _s_py = _s
    _v_py = _v
    _niter_py = _return_scalars.niter
    _nf_py = _return_scalars.nf
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'fsumsq', 'fvec', 'fjac', 's', 'v', 'niter', 'nf'],
    )(
        _x_py,
        _fsumsq_py,
        _fvec_py,
        _fjac_py,
        _s_py,
        _v_py,
        _niter_py,
        _nf_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def lsq_uncon_mod_deriv_comp(m, lsqfun, xtol, x, lsqmon=None, iprint=1, maxcal=None, eta=None, stepmx=100000.0, data=None, spiked_sorder='C'):
    u"""
    Unconstrained minimum of a sum of squares, combined Gauss--Newton
    and modified Newton algorithm, using first derivatives
    (comprehensive).

    ``lsq_uncon_mod_deriv_comp`` is a comprehensive modified
    Gauss--Newton algorithm for finding an unconstrained minimum of a
    sum of squares of m nonlinear functions in n variables (m >= n).
    First derivatives are required.

    The function is intended for functions which have continuous first
    and second derivatives (although it will usually work even if the
    derivatives have occasional discontinuities).

    For full information please refer to the NAG Library document for
    e04gd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04gdf.html

    Parameters
    ----------
    m : int
        The number m of residuals, f_i(x), and the number n of
        variables, x_j.

    lsqfun : callable (iflag, fvec, fjac) = lsqfun(iflag, m, xc,
        data=None)
        `lsqfun` must calculate the vector of values f_i(x) and Jacobian
        matrix of first derivatives df_i/dx_j at any point x. (However,
        if you do not wish to calculate the residuals or first
        derivatives at a particular x, there is the option of setting an
        argument to cause ``lsq_uncon_mod_deriv_comp`` to terminate
        immediately.)

        Parameters
        ~~~~~~~~~~
        iflag : int
            To `lsqfun`, `iflag` will be set to 1 or 2.

            `iflag` = 1

                Indicates that only the Jacobian matrix needs to be
                evaluated

            `iflag` = 2

                Indicates that both the residuals and the Jacobian
                matrix must be calculated

        m : int
            m, the numbers of residuals.

        xc : float, ndarray, shape (n)
            The point x at which the values of the f_i and the df_i/dx_j
            are required.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        iflag : int
            If it is not possible to evaluate the f_i(x) or their first
            derivatives at the point given in `xc` (or if it wished to
            stop the calculations for any other reason), you should
            reset `iflag` to some negative number and return control to
            ``lsq_uncon_mod_deriv_comp``. ``lsq_uncon_mod_deriv_comp``
            will then terminate immediately, with errno set to your
            setting of `iflag`.

        fvec : float, array-like, shape (`m`)
            Unless `iflag` = 1 on entry, or `iflag` is reset to a
            negative number, `fvec`[i-1] must contain the value of f_i
            at the point x, for i = 1,2,...,m.

        fjac : float, array-like, shape (`m`, n)
            Unless `iflag` is reset to a negative number,
            `fjac`[i-1,j-1] must contain the value of df_i/dx_j at the
            point x, for j = 1,2,...,n, for i = 1,2,...,m.

    xtol : float
        The accuracy in x to which the solution is required.

        If x_true is the true value of x at the minimum, then x_sol, the
        estimated position before a normal exit, is such that

            ||x_sol-x_true|| < `xtol`*(1.0+||x_true||)

        where ||y|| = sqrt(sum_j = 1^ny_j^2).

        For example, if the elements of x_sol are not much larger than
        1.0 in modulus and if `xtol` = 1.0e-5, then x_sol is usually
        accurate to about five decimal places. (For further details see
        [Accuracy].)

        If F(x) and the variables are scaled roughly as described in
        [Further Comments] and epsilon is the machine precision, then a
        setting of order `xtol` = sqrt(epsilon) will usually be
        appropriate.

        If `xtol` is set to 0.0 or some positive value less than 10
        epsilon, ``lsq_uncon_mod_deriv_comp`` will use 10 epsilon
        instead of `xtol`, since 10 epsilon is probably the smallest
        reasonable setting.

    x : float, array-like, shape (n)
        `x`[j-1] must be set to a guess at the jth component of the
        position of the minimum, for j = 1,2,...,n.

    lsqmon : None or callable lsqmon(xc, fvec, fjac, s, igrade, niter,
        nf, data=None), optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        If `iprint` >= 0, you must supply `lsqmon` which is suitable for
        monitoring the minimization process. `lsqmon` must not change
        the values of any of its arguments.

        Parameters
        ~~~~~~~~~~
        xc : float, ndarray, shape (n)
            The coordinates of the current point x.

        fvec : float, ndarray, shape (m)
            The values of the residuals f_i at the current point x.

        fjac : float, ndarray, shape (m, n)
            `fjac`[i-1,j-1] contains the value of df_i/dx_j at the
            current point x, for j = 1,2,...,n, for i = 1,2,...,m.

        s : float, ndarray, shape (n)
            The singular values of the current Jacobian matrix. Thus `s`
            may be useful as information about the structure of your
            problem. (If `iprint` > 0, `lsqmon` is called at the initial
            point before the singular values have been calculated. So
            the elements of `s` are set to zero for the first call of
            `lsqmon`.)

        igrade : int
            ``lsq_uncon_mod_deriv_comp`` estimates the dimension of the
            subspace for which the Jacobian matrix can be used as a
            valid approximation to the curvature (see Gill and Murray
            (1978)). This estimate is called the grade of the Jacobian
            matrix, and `igrade` gives its current value.

        niter : int
            The number of iterations which have been performed in
            ``lsq_uncon_mod_deriv_comp``.

        nf : int
            The number of times that `lsqfun` has been called so far
            with `iflag` = 2. (In addition to these calls monitored by
            `nf`, `lsqfun` is called not more than n times per iteration
            with `iflag` set to 1.)

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

    iprint : int, optional
        The frequency with which `lsqmon` is to be called.

        `iprint` > 0

            `lsqmon` is called once every `iprint` iterations and just
            before exit from ``lsq_uncon_mod_deriv_comp``.

        `iprint` = 0

            `lsqmon` is just called at the final point.

        `iprint` < 0

            `lsqmon` is not called at all.

        `iprint` should normally be set to a small positive number.

    maxcal : None or int, optional
        Note: if this argument is None then a default value will be
        used, determined as follows: 50*n.

        Enables you to limit the number of times that `lsqfun` is called
        by ``lsq_uncon_mod_deriv_comp``. There will be an error exit
        (see [Raises/Warns]) after `maxcal` evaluations of the residuals
        (i.e., calls of `lsqfun` with `iflag` set to 2). It should be
        borne in mind that, in addition to the calls of `lsqfun` which
        are limited directly by `maxcal`, there will be calls of
        `lsqfun` (with `iflag` set to 1) to evaluate only first
        derivatives.

    eta : None or float, optional
        Note: if this argument is None then a default value will be
        used, determined as follows: if n = 1: 0.0; otherwise: 0.5.

        Every iteration of ``lsq_uncon_mod_deriv_comp`` involves a
        linear minimization, i.e., minimization of
        F(x^(k)+alpha^(k)p^(k)) with respect to alpha^(k). `eta`
        specifies how accurately these linear minimizations are to be
        performed. The minimum with respect to alpha^(k) will be located
        more accurately for small values of `eta` (say, 0.01) than for
        large values (say, 0.9).

        Although accurate linear minimizations will generally reduce the
        number of iterations, they will tend to increase the number of
        calls of `lsqfun` (with `iflag` set to 2) needed for each linear
        minimization.

        On balance it is usually efficient to perform a low accuracy
        linear minimization.

    stepmx : float, optional
        An estimate of the Euclidean distance between the solution and
        the starting point supplied by you. (For maximum efficiency, a
        slight overestimate is preferable.) ``lsq_uncon_mod_deriv_comp``
        will ensure that, for each iteration,

            sum_j = 1^n(x_j^(k)-x_j^(k-1))^2 <= (`stepmx`)^2

        where k is the iteration number. Thus, if the problem has more
        than one solution, ``lsq_uncon_mod_deriv_comp`` is most likely
        to find the one nearest to the starting point. On difficult
        problems, a realistic choice can prevent the sequence of x^(k)
        entering a region where the problem is ill-behaved and can help
        avoid overflow in the evaluation of F(x). However, an
        underestimate of `stepmx` can lead to inefficiency.

    data : arbitrary, optional
        User-communication data for callback functions.

    spiked_sorder : str, optional
        If `fjac` in `lsqfun` is spiked (i.e., has unit extent in all
        but one dimension, or has size 1), `spiked_sorder` selects the
        storage order to associate with it in the NAG Engine:

        spiked_sorder = 'C'
          row-major storage will be used;

        spiked_sorder = 'F'
          column-major storage will be used.

    Returns
    -------
    x : float, ndarray, shape (n)
        The final point x^(k). Thus, if no exception or warning is
        raised on exit, `x`[j-1] is the jth component of the estimated
        position of the minimum.

    fsumsq : float
        The value of F(x), the sum of squares of the residuals f_i(x),
        at the final point given in `x`.

    fvec : float, ndarray, shape (`m`)
        The value of the residual f_i(x) at the final point given in
        `x`, for i = 1,2,...,m.

    fjac : float, ndarray, shape (`m`, n)
        The value of the first derivative df_i/dx_j evaluated at the
        final point given in `x`, for j = 1,2,...,n, for i = 1,2,...,m.

    s : float, ndarray, shape (n)
        The singular values of the Jacobian matrix at the final point.
        Thus `s` may be useful as information about the structure of
        your problem.

    v : float, ndarray, shape (n, n)
        The matrix V associated with the singular value decomposition

            J = USV^T

        of the Jacobian matrix at the final point, stored by columns.
        This matrix may be useful for statistical purposes, since it is
        the matrix of orthonormalized eigenvectors of J^TJ.

    niter : int
        The number of iterations which have been performed in
        ``lsq_uncon_mod_deriv_comp``.

    nf : int
        The number of times that the residuals have been evaluated
        (i.e., number of calls of `lsqfun` with `iflag` set to 2).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `stepmx` = *<value>* and `xtol` = *<value>*.

            Constraint: `stepmx` >= `xtol`.

        (`errno` 1)
            On entry, `xtol` = *<value>*.

            Constraint: `xtol` >= 0.0.

        (`errno` 1)
            On entry, `eta` = *<value>*.

            Constraint: 0.0 <= `eta` < 1.0.

        (`errno` 1)
            On entry, `maxcal` = *<value>*.

            Constraint: `maxcal` >= 1.

        (`errno` 1)
            On entry, `m` = *<value>* and n = *<value>*.

            Constraint: `m` >= n.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 2)
            There have been `maxcal` = *<value>* evaluations of the
            residuals.

        (`errno` 4)
            Failure in computing SVD of Jacobian matrix.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i < 0)
            User requested termination by setting `iflag` negative in
            `lsqfun`.

        (`errno` 3)
            The conditions for a minimum have not all been satisfied,
            but a lower point could not be found.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Gill, P E and Murray, W, 1978, `Algorithms for the solution of the
    nonlinear least squares problem`, SIAM J. Numer. Anal. (15),
    977--992
    """

    fun_name = 'naginterfaces.library.opt.lsq_uncon_mod_deriv_comp'

    cb_lsqfun_name = fun_name + ':lsqfun'

    def cb_lsqfun(iflag, m, n, xc, fvec, fjac, data=None):
        """Wrapping user's lsqfun. Called as base lsqfun."""
        if data is not None:
            _iflag_lsqfun, _fvec_lsqfun, _fjac_lsqfun = lsqfun(
                iflag,
                m,
                xc,
                data,
            )
        else:
            _iflag_lsqfun, _fvec_lsqfun, _fjac_lsqfun = lsqfun(
                iflag,
                m,
                xc,
            )
        _iflag_lsqfun_base = _iflag_lsqfun
        fvec[:] = _utils._to_ndarray(
            {'fun_name': cb_lsqfun_name, 'returned': 'fvec'},
            _fvec_lsqfun,
            copy=False,
            exp_shape=(m,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _fjac_lsqfun_np = _utils._to_ndarray(
            {'fun_name': cb_lsqfun_name, 'returned': 'fjac'},
            _fjac_lsqfun,
            copy=False,
            exp_shape=(m, n),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _utils._check_callback_sorder(
            {'fun_name': cb_lsqfun_name},
            ((_fjac_lsqfun_np, 'fjac'),),
            wrapper_sorder=_m_sorder,
            spiked_sorder=spiked_sorder,
        )
        fjac[:, :] = _fjac_lsqfun_np
        return _iflag_lsqfun_base

    def cb_lsqmon(m, n, xc, fvec, fjac, s, igrade, niter, nf, data=None):
        """Wrapping user's lsqmon. Called as base lsqmon."""
        if data is not None:
            lsqmon(
                xc,
                fvec,
                fjac,
                s,
                igrade,
                niter,
                nf,
                data,
            )
        else:
            lsqmon(
                xc,
                fvec,
                fjac,
                s,
                igrade,
                niter,
                nf,
            )

    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        (),
        spiked_sorder=spiked_sorder,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b_utils._check_callable(lsqfun, {'fun_name': fun_name, 'entity': 'lsqfun'})
    _lsqfun = cb_lsqfun
    _b_utils._check_callable(lsqmon, {'fun_name': fun_name, 'entity': 'lsqmon'}, allow_none=True)
    _lsqmon = None if lsqmon is None else cb_lsqmon
    _maxcal = 50*_n_py if maxcal is None else maxcal
    if _n_py == 1:
        _eta = 0.0
    else:
        _eta = 0.5
    _eta = _eta if eta is None else eta
    _m = _b_utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = _b_utils._EngineIntScalarType.to_py(_m)
    _fvec = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_m_py,),
    )
    _fjac = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_m_py, _n_py),
        sorder=_b_utils._EngineData.default_sorder,
    )
    _s = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _v = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py, _n_py),
        sorder=_b_utils._EngineData.default_sorder,
    )
    try:
        _return_scalars = _b_opt.lsq_uncon_mod_deriv_comp(
            _m,
            _n,
            _lsqfun,
            _lsqmon,
            iprint,
            _maxcal,
            _eta,
            xtol,
            stepmx,
            _x,
            _fvec,
            _fjac,
            _s,
            _v,
            data,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _x_py = _x
    _fsumsq_py = _return_scalars.fsumsq
    _fvec_py = _fvec
    _fjac_py = _fjac
    _s_py = _s
    _v_py = _v
    _niter_py = _return_scalars.niter
    _nf_py = _return_scalars.nf
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'fsumsq', 'fvec', 'fjac', 's', 'v', 'niter', 'nf'],
    )(
        _x_py,
        _fsumsq_py,
        _fvec_py,
        _fjac_py,
        _s_py,
        _v_py,
        _niter_py,
        _nf_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def lsq_uncon_quasi_deriv_easy(m, lsfun2, x, data=None, spiked_sorder='C'):
    u"""
    Unconstrained minimum of a sum of squares, combined Gauss--Newton
    and quasi-Newton algorithm, using first derivatives (easy-to-use).

    ``lsq_uncon_quasi_deriv_easy`` is an easy-to-use quasi-Newton
    algorithm for finding an unconstrained minimum of a sum of squares
    of m nonlinear functions in n variables (m >= n).
    First derivatives are required.

    It is intended for functions which are continuous and which have
    continuous first and second derivatives (although it will usually
    work even if the derivatives have occasional discontinuities).

    For full information please refer to the NAG Library document for
    e04gy

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04gyf.html

    Parameters
    ----------
    m : int
        The number m of residuals, f_i(x), and the number n of
        variables, x_j.

    lsfun2 : callable (fvec, fjac) = lsfun2(m, xc, data=None)
        You must supply this function to calculate the vector of values
        f_i(x) and the Jacobian matrix of first derivatives df_i/dx_j at
        any point x.

        It should be tested separately before being used in conjunction
        with ``lsq_uncon_quasi_deriv_easy`` (see [the E04
        Introduction]).

        Parameters
        ~~~~~~~~~~
        m : int
            m, the numbers of residuals.

        xc : float, ndarray, shape (n)
            The point x at which the values of the f_i and the df_i/dx_j
            are required.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        fvec : float, array-like, shape (`m`)
            `fvec`[i-1] must contain the value of f_i at the point x,
            for i = 1,2,...,m.

        fjac : float, array-like, shape (`m`, n)
            `fjac`[i-1,j-1] must contain the value of df_i/dx_j at the
            point x, for j = 1,2,...,n, for i = 1,2,...,m.

    x : float, array-like, shape (n)
        `x`[j-1] must be set to a guess at the jth component of the
        position of the minimum, for j = 1,2,...,n. The function checks
        the first derivatives calculated by `lsfun2` at the starting
        point and so is more likely to detect an error in your function
        if the initial `x`[j-1] are nonzero and mutually distinct.

    data : arbitrary, optional
        User-communication data for callback functions.

    spiked_sorder : str, optional
        If `fjac` in `lsfun2` is spiked (i.e., has unit extent in all
        but one dimension, or has size 1), `spiked_sorder` selects the
        storage order to associate with it in the NAG Engine:

        spiked_sorder = 'C'
          row-major storage will be used;

        spiked_sorder = 'F'
          column-major storage will be used.

    Returns
    -------
    x : float, ndarray, shape (n)
        The lowest point found during the calculations. Thus, if no
        exception or warning is raised on exit, `x`[j-1] is the jth
        component of the position of the minimum.

    fsumsq : float
        The value of the sum of squares, F(x), corresponding to the
        final point stored in `x`.

    comm : dict, communication object
        Communication structure.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m` = *<value>* and n = *<value>*.

            Constraint: `m` >= n.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 2)
            There have been 50*n calls to `lsfun2`.

        (`errno` 4)
            Failure in computing SVD of Jacobian matrix.

        (`errno` 9)
            It is very likely that you have made an error in forming the
            derivatives in `lsfun2`.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            The conditions for a minimum have not all been satisfied,
            but a lower point could not be found.

        (`errno` 5)
            It is probable that a local minimum has been found, but it
            cannot be guaranteed.

        (`errno` 6)
            It is possible that a local minimum has been found, but it
            cannot be guaranteed.

        (`errno` 7)
            It is unlikely that a local minimum has been found.

        (`errno` 8)
            It is very unlikely that a local minimum has been found.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Gill, P E and Murray, W, 1978, `Algorithms for the solution of the
    nonlinear least squares problem`, SIAM J. Numer. Anal. (15),
    977--992
    """

    fun_name = 'naginterfaces.library.opt.lsq_uncon_quasi_deriv_easy'

    cb_lsfun2_name = fun_name + ':lsfun2'

    def cb_lsfun2(m, n, xc, fvec, fjac, data=None):
        """Wrapping user's lsfun2. Called as base lsfun2."""
        if data is not None:
            _fvec_lsfun2, _fjac_lsfun2 = lsfun2(
                m,
                xc,
                data,
            )
        else:
            _fvec_lsfun2, _fjac_lsfun2 = lsfun2(
                m,
                xc,
            )
        fvec[:] = _utils._to_ndarray(
            {'fun_name': cb_lsfun2_name, 'returned': 'fvec'},
            _fvec_lsfun2,
            copy=False,
            exp_shape=(m,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _fjac_lsfun2_np = _utils._to_ndarray(
            {'fun_name': cb_lsfun2_name, 'returned': 'fjac'},
            _fjac_lsfun2,
            copy=False,
            exp_shape=(m, n),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _utils._check_callback_sorder(
            {'fun_name': cb_lsfun2_name},
            ((_fjac_lsfun2_np, 'fjac'),),
            wrapper_sorder=_m_sorder,
            spiked_sorder=spiked_sorder,
        )
        fjac[:, :] = _fjac_lsfun2_np

    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        (),
        spiked_sorder=spiked_sorder,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b_utils._check_callable(lsfun2, {'fun_name': fun_name, 'entity': 'lsfun2'})
    _lsfun2 = cb_lsfun2
    try:
        _return_scalars = _b_opt.lsq_uncon_quasi_deriv_easy(
            m,
            _n,
            _lsfun2,
            _x,
            data,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _x_py = _x
    _fsumsq_py = _return_scalars.fsumsq
    _comm_py = _return_scalars.comm
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'fsumsq', 'comm'],
    )(
        _x_py,
        _fsumsq_py,
        _comm_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def lsq_uncon_mod_deriv_easy(m, lsfun2, x, data=None, spiked_sorder='C'):
    u"""
    Unconstrained minimum of a sum of squares, combined Gauss--Newton
    and modified Newton algorithm, using first derivatives
    (easy-to-use).

    ``lsq_uncon_mod_deriv_easy`` is an easy-to-use modified
    Gauss--Newton algorithm for finding an unconstrained minimum of a
    sum of squares of m nonlinear functions in n variables (m >= n).
    First derivatives are required.

    It is intended for functions which are continuous and which have
    continuous first and second derivatives (although it will usually
    work even if the derivatives have occasional discontinuities).

    For full information please refer to the NAG Library document for
    e04gz

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04gzf.html

    Parameters
    ----------
    m : int
        The number m of residuals, f_i(x), and the number n of
        variables, x_j.

    lsfun2 : callable (fvec, fjac) = lsfun2(m, xc, data=None)
        You must supply this function to calculate the vector of values
        f_i(x) and the Jacobian matrix of first derivatives df_i/dx_j at
        any point x.

        It should be tested separately before being used in conjunction
        with ``lsq_uncon_mod_deriv_easy``.

        Parameters
        ~~~~~~~~~~
        m : int
            m, the numbers of residuals.

        xc : float, ndarray, shape (n)
            The point x at which the values of the f_i and the df_i/dx_j
            are required.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        fvec : float, array-like, shape (`m`)
            `fvec`[i-1] must be set to the value of f_i at the point x,
            for i = 1,2,...,m.

        fjac : float, array-like, shape (`m`, n)
            `fjac`[i-1,j-1] must be set to the value of df_i/dx_j at the
            point x, for j = 1,2,...,n, for i = 1,2,...,m.

    x : float, array-like, shape (n)
        `x`[j-1] must be set to a guess at the jth component of the
        position of the minimum, for j = 1,2,...,n. The function checks
        the first derivatives calculated by `lsfun2` at the starting
        point and so is more likely to detect any error in your
        functions if the initial `x`[j-1] are nonzero and mutually
        distinct.

    data : arbitrary, optional
        User-communication data for callback functions.

    spiked_sorder : str, optional
        If `fjac` in `lsfun2` is spiked (i.e., has unit extent in all
        but one dimension, or has size 1), `spiked_sorder` selects the
        storage order to associate with it in the NAG Engine:

        spiked_sorder = 'C'
          row-major storage will be used;

        spiked_sorder = 'F'
          column-major storage will be used.

    Returns
    -------
    x : float, ndarray, shape (n)
        The lowest point found during the calculations. Thus, if no
        exception or warning is raised on exit, `x`[j-1] is the jth
        component of the position of the minimum.

    fsumsq : float
        The value of the sum of squares, F(x), corresponding to the
        final point stored in `x`.

    comm : dict, communication object
        Communication structure.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m` = *<value>* and n = *<value>*.

            Constraint: `m` >= n.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 2)
            There have been 50*n calls to `lsfun2`.

        (`errno` 4)
            Failure in computing SVD of Jacobian matrix.

        (`errno` 9)
            It is very likely that you have made an error in forming the
            derivatives in `lsfun2`.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            The conditions for a minimum have not all been satisfied,
            but a lower point could not be found.

        (`errno` 5)
            It is probable that a local minimum has been found, but it
            cannot be guaranteed.

        (`errno` 6)
            It is possible that a local minimum has been found, but it
            cannot be guaranteed.

        (`errno` 7)
            It is unlikely that a local minimum has been found.

        (`errno` 8)
            It is very unlikely that a local minimum has been found.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Gill, P E and Murray, W, 1978, `Algorithms for the solution of the
    nonlinear least squares problem`, SIAM J. Numer. Anal. (15),
    977--992
    """

    fun_name = 'naginterfaces.library.opt.lsq_uncon_mod_deriv_easy'

    cb_lsfun2_name = fun_name + ':lsfun2'

    def cb_lsfun2(m, n, xc, fvec, fjac, data=None):
        """Wrapping user's lsfun2. Called as base lsfun2."""
        if data is not None:
            _fvec_lsfun2, _fjac_lsfun2 = lsfun2(
                m,
                xc,
                data,
            )
        else:
            _fvec_lsfun2, _fjac_lsfun2 = lsfun2(
                m,
                xc,
            )
        fvec[:] = _utils._to_ndarray(
            {'fun_name': cb_lsfun2_name, 'returned': 'fvec'},
            _fvec_lsfun2,
            copy=False,
            exp_shape=(m,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _fjac_lsfun2_np = _utils._to_ndarray(
            {'fun_name': cb_lsfun2_name, 'returned': 'fjac'},
            _fjac_lsfun2,
            copy=False,
            exp_shape=(m, n),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _utils._check_callback_sorder(
            {'fun_name': cb_lsfun2_name},
            ((_fjac_lsfun2_np, 'fjac'),),
            wrapper_sorder=_m_sorder,
            spiked_sorder=spiked_sorder,
        )
        fjac[:, :] = _fjac_lsfun2_np

    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        (),
        spiked_sorder=spiked_sorder,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b_utils._check_callable(lsfun2, {'fun_name': fun_name, 'entity': 'lsfun2'})
    _lsfun2 = cb_lsfun2
    try:
        _return_scalars = _b_opt.lsq_uncon_mod_deriv_easy(
            m,
            _n,
            _lsfun2,
            _x,
            data,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _x_py = _x
    _fsumsq_py = _return_scalars.fsumsq
    _comm_py = _return_scalars.comm
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'fsumsq', 'comm'],
    )(
        _x_py,
        _fsumsq_py,
        _comm_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def check_deriv(funct, x, data=None):
    u"""
    Check user's function for calculating first derivatives of function.

    ``check_deriv`` checks that a function for evaluating an objective
    function and its first derivatives produces derivative values which
    are consistent with the function values calculated.

    For full information please refer to the NAG Library document for
    e04hc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04hcf.html

    Parameters
    ----------
    funct : callable (iflag, fc, gc) = funct(iflag, xc, data=None)
        `funct` must evaluate the function and its first derivatives at
        a given point. (The minimization functions mentioned in
        [Description] gives you the option of resetting arguments of
        `funct` to cause the minimization process to terminate
        immediately. ``check_deriv`` will also terminate immediately,
        without finishing the checking process, if the argument in
        question is reset.)

        Parameters
        ~~~~~~~~~~
        iflag : int
            Will be set to 2.

        xc : float, ndarray, shape (n)
            The point x at which F and its derivatives are required.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        iflag : int
            If you reset `iflag` to a negative number in `funct` and
            return control to ``check_deriv``, ``check_deriv`` will
            terminate immediately with errno set to your setting of
            `iflag`.

        fc : float
            Unless `funct` resets `iflag`, `fc` must be set to the value
            of the function F at the current point x.

        gc : float, array-like, shape (n)
            Unless `funct` resets `iflag`, `gc`[j-1] must be set to the
            value of the first derivative dF/dx_j at the point x, for j
            = 1,2,...,n.

    x : float, array-like, shape (n)
        `x`[j-1], for j = 1,2,...,n, must be set to the coordinates of a
        suitable point at which to check the derivatives calculated by
        `funct`. 'Obvious' settings, such as 0.0 or 1.0, should not be
        used since, at such particular points, incorrect terms may take
        correct values (particularly zero), so that errors could go
        undetected. Similarly, it is preferable that no two elements of
        `x` should be the same.

    data : arbitrary, optional
        User-communication data for callback functions.

    Returns
    -------
    f : float
        Unless you set `iflag` negative in the first call of `funct`,
        `f` contains the value of the objective function F(x) at the
        point given by you in `x`.

    g : float, ndarray, shape (n)
        Unless you set `iflag` negative in the first call of `funct`,
        `g`[j-1] contains the value of the derivative dF/dx_j at the
        point given in `x`, as calculated by `funct`, for j = 1,2,...,n.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 1.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i < 0)
            User requested termination by setting `iflag` negative in
            `funct`.

        (`errno` 2)
            Large errors were found in the derivatives of F computed by
            `funct`.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    fun_name = 'naginterfaces.library.opt.check_deriv'

    cb_funct_name = fun_name + ':funct'

    def cb_funct(iflag, n, xc, gc, data=None):
        """Wrapping user's funct. Called as base funct."""
        if data is not None:
            _iflag_funct, _fc_funct, _gc_funct = funct(
                iflag,
                xc,
                data,
            )
        else:
            _iflag_funct, _fc_funct, _gc_funct = funct(
                iflag,
                xc,
            )
        _iflag_funct_base = _iflag_funct
        gc[:] = _utils._to_ndarray(
            {'fun_name': cb_funct_name, 'returned': 'gc'},
            _gc_funct,
            copy=False,
            exp_shape=(n,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _fc_funct_base = _fc_funct
        return (
            _iflag_funct_base,
            _fc_funct_base,
        )

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b_utils._check_callable(funct, {'fun_name': fun_name, 'entity': 'funct'})
    _funct = cb_funct
    _g = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _return_scalar = _b_opt.check_deriv(
            _n,
            _funct,
            _x,
            _g,
            data,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _f_py = _return_scalar
    _g_py = _g
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['f', 'g'],
    )(
        _f_py,
        _g_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def check_deriv2(funct, h, x, lh, data=None):
    u"""
    Check user's function for calculating second derivatives of
    function.

    ``check_deriv2`` checks that a function for calculating second
    derivatives of an objective function is consistent with a function
    for calculating the corresponding first derivatives.

    For full information please refer to the NAG Library document for
    e04hd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04hdf.html

    Parameters
    ----------
    funct : callable (iflag, fc, gc) = funct(iflag, xc, data=None)
        `funct` must evaluate the function and its first derivatives at
        a given point. (``bounds_mod_deriv2_comp`` gives you the option
        of resetting arguments of `funct` to cause the minimization
        process to terminate immediately. ``check_deriv2`` will also
        terminate immediately, without finishing the checking process,
        if the argument in question is reset.)

        Parameters
        ~~~~~~~~~~
        iflag : int
            To `funct`, `iflag` will be set to 2.

        xc : float, ndarray, shape (n)
            The point x at which the function and first derivatives are
            required.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        iflag : int
            If you set `iflag` to some negative number in `funct` and
            return control to ``check_deriv2``, ``check_deriv2`` will
            terminate immediately with errno set to your setting of
            `iflag`.

        fc : float
            Unless `funct` resets `iflag`, `fc` must be set to the value
            of the objective function F at the current point x.

        gc : float, array-like, shape (n)
            Unless `funct` resets `iflag`, `gc`[j-1] must be set to the
            value of the first derivative dF/dx_j at the point x, for j
            = 1,2,...,n.

    h : callable (iflag, fhesl, fhesd) = h(iflag, xc, lh, fhesd,
        data=None)
        `h` must evaluate the second derivatives of the function at a
        given point. (As with `funct`, an argument can be set to cause
        immediate termination.)

        Parameters
        ~~~~~~~~~~
        iflag : int
            Is set to a non-negative number.

        xc : float, ndarray, shape (n)
            The point x at which the second derivatives of F(x) are
            required.

        lh : int
            The length of the array `fhesl`.

        fhesd : float, ndarray, shape (n)
            Contains the value of dF/dx_j at the point x, for j =
            1,2,...,n. Functions written to take advantage of a similar
            feature of ``bounds_mod_deriv2_comp`` can be tested as they
            stand by ``check_deriv2``.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        iflag : int
            If `h` resets `iflag` to a negative number, ``check_deriv2``
            will terminate immediately with errno set to your setting of
            `iflag`.

        fhesl : float, array-like, shape (`lh`)
            Unless `iflag` is reset, `h` must place the strict lower
            triangle of the second derivative matrix of F (evaluated at
            the point x) in `fhesl`, stored by rows, i.e.,
            `fhesl`[(i-1)(i-2)/2+j-1] must be set to the value of
            d^2F/(dx_i dx_j) at the point x, for j = 1,2,...,i-1, for i
            = 2,3,...,n. (The upper triangle is not required because the
            matrix is symmetric.)

        fhesd : float, array-like, shape (n)
            Unless `iflag` is reset, `h` must place the diagonal
            elements of the second derivative matrix of F (evaluated at
            the point x) in `fhesd`, i.e., `fhesd`[j-1] must be set to
            the value of d^2F/dx_j^2 at the point x, for j = 1,2,...,n.

    x : float, array-like, shape (n)
        `x`[j-1], for j = 1,2,...,n, must contain the coordinates of a
        suitable point at which to check the derivatives calculated by
        `funct`. 'Obvious' settings, such as 0.0 or 1.0, should not be
        used since, at such particular points, incorrect terms may take
        correct values (particularly zero), so that errors could go
        undetected. Similarly, it is advisable that no two elements of
        `x` should be the same.

    lh : int
        The dimension of the array `hesl`.

    data : arbitrary, optional
        User-communication data for callback functions.

    Returns
    -------
    g : float, ndarray, shape (n)
        Unless you set `iflag` negative in the first call of `funct`,
        `g`[j-1] contains the value of the first derivative dF/dx_j at
        the point given in `x`, as calculated by `funct`, for j =
        1,2,...,n.

    hesl : float, ndarray, shape (`lh`)
        Unless you set `iflag` negative in `h`, `hesl` contains the
        strict lower triangle of the second derivative matrix of F, as
        evaluated by `h` at the point given in `x`, stored by rows.

    hesd : float, ndarray, shape (n)
        Unless you set `iflag` negative in `h`, `hesd` contains the
        diagonal elements of the second derivative matrix of F, as
        evaluated by `h` at the point given in `x`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 1)
            On entry, `lh` = *<value>* and n = *<value>*.

            Constraint: `lh` >= max(1, n*(n-1)/2).

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i < 0)
            User requested termination by setting `iflag` negative in
            `funct`.

        (`errno` 2)
            Large errors were found in the derivatives of F computed by
            `funct`.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    fun_name = 'naginterfaces.library.opt.check_deriv2'

    cb_funct_name = fun_name + ':funct'

    def cb_funct(iflag, n, xc, gc, data=None):
        """Wrapping user's funct. Called as base funct."""
        if data is not None:
            _iflag_funct, _fc_funct, _gc_funct = funct(
                iflag,
                xc,
                data,
            )
        else:
            _iflag_funct, _fc_funct, _gc_funct = funct(
                iflag,
                xc,
            )
        _iflag_funct_base = _iflag_funct
        gc[:] = _utils._to_ndarray(
            {'fun_name': cb_funct_name, 'returned': 'gc'},
            _gc_funct,
            copy=False,
            exp_shape=(n,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _fc_funct_base = _fc_funct
        return (
            _iflag_funct_base,
            _fc_funct_base,
        )

    cb_h_name = fun_name + ':h'

    def cb_h(iflag, n, xc, fhesl, lh, fhesd, data=None):
        """Wrapping user's h. Called as base h."""
        if data is not None:
            _iflag_h, _fhesl_h, _fhesd_h = h(
                iflag,
                xc,
                lh,
                fhesd,
                data,
            )
        else:
            _iflag_h, _fhesl_h, _fhesd_h = h(
                iflag,
                xc,
                lh,
                fhesd,
            )
        _iflag_h_base = _iflag_h
        fhesl[:] = _utils._to_ndarray(
            {'fun_name': cb_h_name, 'returned': 'fhesl'},
            _fhesl_h,
            copy=False,
            exp_shape=(lh,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        fhesd[:] = _utils._to_ndarray(
            {'fun_name': cb_h_name, 'returned': 'fhesd'},
            _fhesd_h,
            copy=False,
            exp_shape=(n,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        return _iflag_h_base

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b_utils._check_callable(funct, {'fun_name': fun_name, 'entity': 'funct'})
    _funct = cb_funct
    _b_utils._check_callable(h, {'fun_name': fun_name, 'entity': 'h'})
    _h = cb_h
    _g = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _lh = _b_utils._EngineIntScalarType(
        lh, {'fun_name': fun_name, 'entity': 'lh'},
    ).to_ctypes()
    _lh_py = _b_utils._EngineIntScalarType.to_py(_lh)
    _hesl = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_lh_py,),
    )
    _hesd = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _b_opt.check_deriv2(
            _n,
            _funct,
            _h,
            _x,
            _g,
            _hesl,
            _lh,
            _hesd,
            data,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _g_py = _g
    _hesl_py = _hesl
    _hesd_py = _hesd
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['g', 'hesl', 'hesd'],
    )(
        _g_py,
        _hesl_py,
        _hesd_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def lsq_uncon_mod_deriv2_comp(m, lsqfun, lsqhes, xtol, x, lsqmon=None, iprint=1, maxcal=None, eta=None, stepmx=100000.0, data=None, spiked_sorder='C'):
    u"""
    Unconstrained minimum of a sum of squares, combined Gauss--Newton
    and modified Newton algorithm, using second derivatives
    (comprehensive).

    ``lsq_uncon_mod_deriv2_comp`` is a comprehensive modified
    Gauss--Newton algorithm for finding an unconstrained minimum of a
    sum of squares of m nonlinear functions in n variables (m >= n).
    First and second derivatives are required.

    The function is intended for functions which have continuous first
    and second derivatives (although it will usually work even if the
    derivatives have occasional discontinuities).

    For full information please refer to the NAG Library document for
    e04he

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04hef.html

    Parameters
    ----------
    m : int
        The number m of residuals, f_i(x), and the number n of
        variables, x_j.

    lsqfun : callable (iflag, fvec, fjac) = lsqfun(iflag, m, xc,
        data=None)
        `lsqfun` must calculate the vector of values f_i(x) and Jacobian
        matrix of first derivatives df_i/dx_j at any point x. (However,
        if you do not wish to calculate the residuals or first
        derivatives at a particular x, there is the option of setting an
        argument to cause ``lsq_uncon_mod_deriv2_comp`` to terminate
        immediately.)

        Parameters
        ~~~~~~~~~~
        iflag : int
            To `lsqfun`, `iflag` will be set to 2.

        m : int
            m, the numbers of residuals.

        xc : float, ndarray, shape (n)
            The point x at which the values of the f_i and the df_i/dx_j
            are required.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        iflag : int
            If it is not possible to evaluate the f_i(x) or their first
            derivatives at the point given in `xc` (or if it wished to
            stop the calculations for any other reason), you should
            reset `iflag` to some negative number and return control to
            ``lsq_uncon_mod_deriv2_comp``. ``lsq_uncon_mod_deriv2_comp``
            will then terminate immediately, with errno set to your
            setting of `iflag`.

        fvec : float, array-like, shape (`m`)
            Unless `iflag` is reset to a negative number, `fvec`[i-1]
            must contain the value of f_i at the point x, for i =
            1,2,...,m.

        fjac : float, array-like, shape (`m`, n)
            Unless `iflag` is reset to a negative number,
            `fjac`[i-1,j-1] must contain the value of df_i/dx_j at the
            point x, for j = 1,2,...,n, for i = 1,2,...,m.

    lsqhes : callable (iflag, b) = lsqhes(iflag, fvec, xc, lb,
        data=None)
        `lsqhes` must calculate the elements of the symmetric matrix

            B(x) = sum_i = 1^mf_i(x)G_i(x),

        at any point x, where G_i(x) is the Hessian matrix of f_i(x).
        (As with `lsqfun`, there is the option of causing
        ``lsq_uncon_mod_deriv2_comp`` to terminate immediately.)

        Parameters
        ~~~~~~~~~~
        iflag : int
            Is set to a non-negative number.

        fvec : float, ndarray, shape (m)
            The value of the residual f_i at the point x, for i =
            1,2,...,m, so that the values of the f_i can be used in the
            calculation of the elements of `b`.

        xc : float, ndarray, shape (n)
            The point x at which the elements of `b` are to be
            evaluated.

        lb : int
            The length of the array `b`.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        iflag : int
            If `lsqhes` resets `iflag` to some negative number,
            ``lsq_uncon_mod_deriv2_comp`` will terminate immediately,
            with errno set to your setting of `iflag`.

        b : float, array-like, shape (`lb`)
            Unless `iflag` is reset to a negative number, `b` must
            contain the lower triangle of the matrix `b`[x-1], evaluated
            at the point x, stored by rows. (The upper triangle is not
            required because the matrix is symmetric.) More precisely,
            `b`[j(j-1)/2+k-1] must contain sum_i = 1^mf_i d^2f_i/(dx_j
            dx_k) evaluated at the point x, for k = 1,2,...,j, for j =
            1,2,...,n.

    xtol : float
        The accuracy in x to which the solution is required.

        If x_true is the true value of x at the minimum, then x_sol, the
        estimated position before a normal exit, is such that

            ||x_sol-x_true|| < `xtol`*(1.0+||x_true||),

        where ||y|| = sqrt(sum_j = 1^ny_j^2).

        For example, if the elements of x_sol are not much larger than
        1.0 in modulus and if `xtol` = 1.0e-5, then x_sol is usually
        accurate to about five decimal places. (For further details see
        [Accuracy].)

        If F(x) and the variables are scaled roughly as described in
        [Further Comments] and epsilon is the machine precision, then a
        setting of order `xtol` = sqrt(epsilon) will usually be
        appropriate.

        If `xtol` is set to 0.0 or some positive value less than 10
        epsilon, ``lsq_uncon_mod_deriv2_comp`` will use 10 epsilon
        instead of `xtol`, since 10 epsilon is probably the smallest
        reasonable setting.

    x : float, array-like, shape (n)
        `x`[j-1] must be set to a guess at the jth component of the
        position of the minimum, for j = 1,2,...,n.

    lsqmon : None or callable lsqmon(xc, fvec, fjac, s, igrade, niter,
        nf, data=None), optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        If `iprint` >= 0, you must supply `lsqmon` which is suitable for
        monitoring the minimization process. `lsqmon` must not change
        the values of any of its arguments.

        Parameters
        ~~~~~~~~~~
        xc : float, ndarray, shape (n)
            The coordinates of the current point x.

        fvec : float, ndarray, shape (m)
            The values of the residuals f_i at the current point x.

        fjac : float, ndarray, shape (m, n)
            `fjac`[i-1,j-1] contains the value of df_i/dx_j at the
            current point x, for j = 1,2,...,n, for i = 1,2,...,m.

        s : float, ndarray, shape (n)
            The singular values of the current Jacobian matrix. Thus `s`
            may be useful as information about the structure of your
            problem. (If `iprint` > 0, `lsqmon` is called at the initial
            point before the singular values have been calculated, so
            the elements of `s` are set to zero for the first call of
            `lsqmon`.)

        igrade : int
            ``lsq_uncon_mod_deriv2_comp`` estimates the dimension of the
            subspace for which the Jacobian matrix can be used as a
            valid approximation to the curvature (see Gill and Murray
            (1978)). This estimate is called the grade of the Jacobian
            matrix, and `igrade` gives its current value.

        niter : int
            The number of iterations which have been performed in
            ``lsq_uncon_mod_deriv2_comp``.

        nf : int
            The number of times that `lsqfun` has been called so far.
            Thus `nf` gives the number of evaluations of the residuals
            and the Jacobian matrix.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

    iprint : int, optional
        Specifies the frequency with which `lsqmon` is to be called.

        `iprint` > 0

            `lsqmon` is called once every `iprint` iterations and just
            before exit from ``lsq_uncon_mod_deriv2_comp``.

        `iprint` = 0

            `lsqmon` is just called at the final point.

        `iprint` < 0

            `lsqmon` is not called at all.

        `iprint` should normally be set to a small positive number.

    maxcal : None or int, optional
        Note: if this argument is None then a default value will be
        used, determined as follows: 50*n.

        This argument is present so as to enable you to limit the number
        of times that `lsqfun` is called by
        ``lsq_uncon_mod_deriv2_comp``. There will be an error exit (see
        [Raises/Warns]) after `maxcal` calls of `lsqfun`.

    eta : None or float, optional
        Note: if this argument is None then a default value will be
        used, determined as follows: if n = 1: 0.0; otherwise: 0.5.

        Every iteration of ``lsq_uncon_mod_deriv2_comp`` involves a
        linear minimization (i.e., minimization of
        F(x^(k)+alpha^(k)p^(k)) with respect to alpha^(k)). `eta` must
        lie in the range 0.0 <= `eta` < 1.0, and specifies how
        accurately these linear minimizations are to be performed. The
        minimum with respect to alpha^(k) will be located more
        accurately for small values of `eta` (say, 0.01) than for large
        values (say, 0.9).

        Although accurate linear minimizations will generally reduce the
        number of iterations performed by ``lsq_uncon_mod_deriv2_comp``,
        they will increase the number of calls of `lsqfun` made each
        iteration.

        On balance it is usually more efficient to perform a low
        accuracy minimization.

    stepmx : float, optional
        An estimate of the Euclidean distance between the solution and
        the starting point supplied by you. (For maximum efficiency, a
        slight overestimate is preferable.)

        ``lsq_uncon_mod_deriv2_comp`` will ensure that, for each
        iteration

            sum_j = 1^n(x_j^(k)-x_j^(k-1))^2 <= (`stepmx`)^2,

        where k is the iteration number.

        Thus, if the problem has more than one solution,
        ``lsq_uncon_mod_deriv2_comp`` is most likely to find the one
        nearest to the starting point.

        On difficult problems, a realistic choice can prevent the
        sequence of x^(k) entering a region where the problem is
        ill-behaved and can help avoid overflow in the evaluation of
        F(x).

        However, an underestimate of `stepmx` can lead to inefficiency.

    data : arbitrary, optional
        User-communication data for callback functions.

    spiked_sorder : str, optional
        If `fjac` in `lsqfun` is spiked (i.e., has unit extent in all
        but one dimension, or has size 1), `spiked_sorder` selects the
        storage order to associate with it in the NAG Engine:

        spiked_sorder = 'C'
          row-major storage will be used;

        spiked_sorder = 'F'
          column-major storage will be used.

    Returns
    -------
    x : float, ndarray, shape (n)
        The final point x^(k). Thus, if no exception or warning is
        raised on exit, `x`[j-1] is the jth component of the estimated
        position of the minimum.

    fsumsq : float
        The value of F(x), the sum of squares of the residuals f_i(x),
        at the final point given in `x`.

    fvec : float, ndarray, shape (`m`)
        The value of the residual f_i(x) at the final point given in
        `x`, for i = 1,2,...,m.

    fjac : float, ndarray, shape (`m`, n)
        The value of the first derivative df_i/dx_j evaluated at the
        final point given in `x`, for j = 1,2,...,n, for i = 1,2,...,m.

    s : float, ndarray, shape (n)
        The singular values of the Jacobian matrix at the final point.
        Thus `s` may be useful as information about the structure of
        your problem.

    v : float, ndarray, shape (n, n)
        The matrix V associated with the singular value decomposition

            J = USV^T

        of the Jacobian matrix at the final point, stored by columns.
        This matrix may be useful for statistical purposes, since it is
        the matrix of orthonormalized eigenvectors of J^TJ.

    niter : int
        The number of iterations which have been performed in
        ``lsq_uncon_mod_deriv2_comp``.

    nf : int
        The number of times that the residuals and Jacobian matrix have
        been evaluated (i.e., number of calls of `lsqfun`).

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `stepmx` = *<value>* and `xtol` = *<value>*.

            Constraint: `stepmx` >= `xtol`.

        (`errno` 1)
            On entry, `xtol` = *<value>*.

            Constraint: `xtol` >= 0.0.

        (`errno` 1)
            On entry, `eta` = *<value>*.

            Constraint: 0.0 <= `eta` < 1.0.

        (`errno` 1)
            On entry, `maxcal` = *<value>*.

            Constraint: `maxcal` >= 1.

        (`errno` 1)
            On entry, `m` = *<value>* and n = *<value>*.

            Constraint: `m` >= n.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 2)
            There have been `maxcal` = *<value>* calls to `lsqfun`.

        (`errno` 4)
            Failure in computing SVD of Jacobian matrix.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i < 0)
            User requested termination by setting `iflag` negative in
            `lsqfun` or `lsqhes`.

        (`errno` 3)
            The conditions for a minimum have not all been satisfied,
            but a lower point could not be found.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Gill, P E and Murray, W, 1978, `Algorithms for the solution of the
    nonlinear least squares problem`, SIAM J. Numer. Anal. (15),
    977--992
    """

    fun_name = 'naginterfaces.library.opt.lsq_uncon_mod_deriv2_comp'

    cb_lsqfun_name = fun_name + ':lsqfun'

    def cb_lsqfun(iflag, m, n, xc, fvec, fjac, data=None):
        """Wrapping user's lsqfun. Called as base lsqfun."""
        if data is not None:
            _iflag_lsqfun, _fvec_lsqfun, _fjac_lsqfun = lsqfun(
                iflag,
                m,
                xc,
                data,
            )
        else:
            _iflag_lsqfun, _fvec_lsqfun, _fjac_lsqfun = lsqfun(
                iflag,
                m,
                xc,
            )
        _iflag_lsqfun_base = _iflag_lsqfun
        fvec[:] = _utils._to_ndarray(
            {'fun_name': cb_lsqfun_name, 'returned': 'fvec'},
            _fvec_lsqfun,
            copy=False,
            exp_shape=(m,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _fjac_lsqfun_np = _utils._to_ndarray(
            {'fun_name': cb_lsqfun_name, 'returned': 'fjac'},
            _fjac_lsqfun,
            copy=False,
            exp_shape=(m, n),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _utils._check_callback_sorder(
            {'fun_name': cb_lsqfun_name},
            ((_fjac_lsqfun_np, 'fjac'),),
            wrapper_sorder=_m_sorder,
            spiked_sorder=spiked_sorder,
        )
        fjac[:, :] = _fjac_lsqfun_np
        return _iflag_lsqfun_base

    cb_lsqhes_name = fun_name + ':lsqhes'

    def cb_lsqhes(iflag, m, n, fvec, xc, b, lb, data=None):
        """Wrapping user's lsqhes. Called as base lsqhes."""
        if data is not None:
            _iflag_lsqhes, _b_lsqhes = lsqhes(
                iflag,
                fvec,
                xc,
                lb,
                data,
            )
        else:
            _iflag_lsqhes, _b_lsqhes = lsqhes(
                iflag,
                fvec,
                xc,
                lb,
            )
        _iflag_lsqhes_base = _iflag_lsqhes
        b[:] = _utils._to_ndarray(
            {'fun_name': cb_lsqhes_name, 'returned': 'b'},
            _b_lsqhes,
            copy=False,
            exp_shape=(lb,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        return _iflag_lsqhes_base

    def cb_lsqmon(m, n, xc, fvec, fjac, s, igrade, niter, nf, data=None):
        """Wrapping user's lsqmon. Called as base lsqmon."""
        if data is not None:
            lsqmon(
                xc,
                fvec,
                fjac,
                s,
                igrade,
                niter,
                nf,
                data,
            )
        else:
            lsqmon(
                xc,
                fvec,
                fjac,
                s,
                igrade,
                niter,
                nf,
            )

    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        (),
        spiked_sorder=spiked_sorder,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b_utils._check_callable(lsqfun, {'fun_name': fun_name, 'entity': 'lsqfun'})
    _lsqfun = cb_lsqfun
    _b_utils._check_callable(lsqhes, {'fun_name': fun_name, 'entity': 'lsqhes'})
    _lsqhes = cb_lsqhes
    _b_utils._check_callable(lsqmon, {'fun_name': fun_name, 'entity': 'lsqmon'}, allow_none=True)
    _lsqmon = None if lsqmon is None else cb_lsqmon
    _maxcal = 50*_n_py if maxcal is None else maxcal
    if _n_py == 1:
        _eta = 0.0
    else:
        _eta = 0.5
    _eta = _eta if eta is None else eta
    _m = _b_utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = _b_utils._EngineIntScalarType.to_py(_m)
    _fvec = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_m_py,),
    )
    _fjac = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_m_py, _n_py),
        sorder=_b_utils._EngineData.default_sorder,
    )
    _s = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _v = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py, _n_py),
        sorder=_b_utils._EngineData.default_sorder,
    )
    try:
        _return_scalars = _b_opt.lsq_uncon_mod_deriv2_comp(
            _m,
            _n,
            _lsqfun,
            _lsqhes,
            _lsqmon,
            iprint,
            _maxcal,
            _eta,
            xtol,
            stepmx,
            _x,
            _fvec,
            _fjac,
            _s,
            _v,
            data,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _x_py = _x
    _fsumsq_py = _return_scalars.fsumsq
    _fvec_py = _fvec
    _fjac_py = _fjac
    _s_py = _s
    _v_py = _v
    _niter_py = _return_scalars.niter
    _nf_py = _return_scalars.nf
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'fsumsq', 'fvec', 'fjac', 's', 'v', 'niter', 'nf'],
    )(
        _x_py,
        _fsumsq_py,
        _fvec_py,
        _fjac_py,
        _s_py,
        _v_py,
        _niter_py,
        _nf_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def lsq_uncon_mod_deriv2_easy(m, lsfun2, lshes2, x, data=None, spiked_sorder='C'):
    u"""
    Unconstrained minimum of a sum of squares, combined Gauss--Newton
    and modified Newton algorithm, using second derivatives
    (easy-to-use).

    ``lsq_uncon_mod_deriv2_easy`` is an easy-to-use modified
    Gauss--Newton algorithm for finding an unconstrained minimum of a
    sum of squares of m nonlinear functions in n variables (m >= n).
    First and second derivatives are required.

    It is intended for functions which are continuous and which have
    continuous first and second derivatives (although it will usually
    work even if the derivatives have occasional discontinuities).

    For full information please refer to the NAG Library document for
    e04hy

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04hyf.html

    Parameters
    ----------
    m : int
        The number m of residuals, f_i(x), and the number n of
        variables, x_j.

    lsfun2 : callable (fvec, fjac) = lsfun2(m, xc, data=None)
        You must supply this function to calculate the vector of values
        f_i(x) and the Jacobian matrix of first derivatives df_i/dx_j at
        any point x.

        It should be tested separately before being used in conjunction
        with ``lsq_uncon_mod_deriv2_easy`` (see [the E04 Introduction]).

        Parameters
        ~~~~~~~~~~
        m : int
            m, the numbers of residuals.

        xc : float, ndarray, shape (n)
            The point x at which the values of the f_i and the df_i/dx_j
            are required.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        fvec : float, array-like, shape (`m`)
            `fvec`[i-1] must be set to the value of f_i at the point x,
            for i = 1,2,...,m.

        fjac : float, array-like, shape (`m`, n)
            `fjac`[i-1,j-1] must be set to the value of df_i/dx_j at the
            point x, for j = 1,2,...,n, for i = 1,2,...,m.

    lshes2 : callable b = lshes2(fvec, xc, lb, data=None)
        You must supply this function to calculate the elements of the
        symmetric matrix

            B(x) = sum_i = 1^mf_i(x)G_i(x),

        at any point x, where G_i(x) is the Hessian matrix of f_i(x).

        It should be tested separately before being used in conjunction
        with ``lsq_uncon_mod_deriv2_easy`` (see [the E04 Introduction]).

        Parameters
        ~~~~~~~~~~
        fvec : float, ndarray, shape (m)
            The value of the residual f_i at the point x, for i =
            1,2,...,m, so that the values of the f_i can be used in the
            calculation of the elements of `b`.

        xc : float, ndarray, shape (n)
            The point x at which the elements of `b` are to be
            evaluated.

        lb : int
            The length of the array `b`.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        b : float, array-like, shape (`lb`)
            Must contain the lower triangle of the matrix B(x),
            evaluated at the point x, stored by rows. (The upper
            triangle is not required because the matrix is symmetric.)
            More precisely, `b`[j(j-1)/2+k-1] must contain sum_i =
            1^mf_i d^2f_i/(dx_j dx_k) evaluated at the point x, for k =
            1,2,...,j, for j = 1,2,...,n.

    x : float, array-like, shape (n)
        `x`[j-1] must be set to a guess at the jth component of the
        position of the minimum, for j = 1,2,...,n. The function checks
        `lsfun2` and `lshes2` at the starting point and so is more
        likely to detect any error in your functions if the initial
        `x`[j-1] are nonzero and mutually distinct.

    data : arbitrary, optional
        User-communication data for callback functions.

    spiked_sorder : str, optional
        If `fjac` in `lsfun2` is spiked (i.e., has unit extent in all
        but one dimension, or has size 1), `spiked_sorder` selects the
        storage order to associate with it in the NAG Engine:

        spiked_sorder = 'C'
          row-major storage will be used;

        spiked_sorder = 'F'
          column-major storage will be used.

    Returns
    -------
    x : float, ndarray, shape (n)
        The lowest point found during the calculations. Thus, if no
        exception or warning is raised on exit, `x`[j-1] is the jth
        component of the position of the minimum.

    fsumsq : float
        The value of the sum of squares, F(x), corresponding to the
        final point stored in `x`.

    comm : dict, communication object
        Communication structure.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m` = *<value>* and n = *<value>*.

            Constraint: `m` >= n.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 2)
            There have been 50*n calls to `lsfun2`.

        (`errno` 4)
            Failure in computing SVD of Jacobian matrix.

        (`errno` 9)
            It is very likely that you have made an error in forming the
            derivatives in `lsfun2`.

        (`errno` 10)
            It is very likely that you have made an error in setting up
            the array `b` in `lshes2`.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            The conditions for a minimum have not all been satisfied,
            but a lower point could not be found.

        (`errno` 5)
            It is probable that a local minimum has been found, but it
            cannot be guaranteed.

        (`errno` 6)
            It is possible that a local minimum has been found, but it
            cannot be guaranteed.

        (`errno` 7)
            It is unlikely that a local minimum has been found.

        (`errno` 8)
            It is very unlikely that a local minimum has been found.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Gill, P E and Murray, W, 1978, `Algorithms for the solution of the
    nonlinear least squares problem`, SIAM J. Numer. Anal. (15),
    977--992
    """

    fun_name = 'naginterfaces.library.opt.lsq_uncon_mod_deriv2_easy'

    cb_lsfun2_name = fun_name + ':lsfun2'

    def cb_lsfun2(m, n, xc, fvec, fjac, data=None):
        """Wrapping user's lsfun2. Called as base lsfun2."""
        if data is not None:
            _fvec_lsfun2, _fjac_lsfun2 = lsfun2(
                m,
                xc,
                data,
            )
        else:
            _fvec_lsfun2, _fjac_lsfun2 = lsfun2(
                m,
                xc,
            )
        fvec[:] = _utils._to_ndarray(
            {'fun_name': cb_lsfun2_name, 'returned': 'fvec'},
            _fvec_lsfun2,
            copy=False,
            exp_shape=(m,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _fjac_lsfun2_np = _utils._to_ndarray(
            {'fun_name': cb_lsfun2_name, 'returned': 'fjac'},
            _fjac_lsfun2,
            copy=False,
            exp_shape=(m, n),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _utils._check_callback_sorder(
            {'fun_name': cb_lsfun2_name},
            ((_fjac_lsfun2_np, 'fjac'),),
            wrapper_sorder=_m_sorder,
            spiked_sorder=spiked_sorder,
        )
        fjac[:, :] = _fjac_lsfun2_np

    cb_lshes2_name = fun_name + ':lshes2'

    def cb_lshes2(m, n, fvec, xc, b, lb, data=None):
        """Wrapping user's lshes2. Called as base lshes2."""
        if data is not None:
            _b_lshes2 = lshes2(
                fvec,
                xc,
                lb,
                data,
            )
        else:
            _b_lshes2 = lshes2(
                fvec,
                xc,
                lb,
            )
        b[:] = _utils._to_ndarray(
            {'fun_name': cb_lshes2_name, 'returned': 'b'},
            _b_lshes2,
            copy=False,
            exp_shape=(lb,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )

    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        (),
        spiked_sorder=spiked_sorder,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b_utils._check_callable(lsfun2, {'fun_name': fun_name, 'entity': 'lsfun2'})
    _lsfun2 = cb_lsfun2
    _b_utils._check_callable(lshes2, {'fun_name': fun_name, 'entity': 'lshes2'})
    _lshes2 = cb_lshes2
    try:
        _return_scalars = _b_opt.lsq_uncon_mod_deriv2_easy(
            m,
            _n,
            _lsfun2,
            _lshes2,
            _x,
            data,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _x_py = _x
    _fsumsq_py = _return_scalars.fsumsq
    _comm_py = _return_scalars.comm
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'fsumsq', 'comm'],
    )(
        _x_py,
        _fsumsq_py,
        _comm_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

@_b_utils._withdrawal(
    'naginterfaces.library.opt.bounds_bobyqa_func',
    mkwdrn=None,
    advice=[
        'Please use handle_solve_dfno and handle_solve_dfno_rcomm instead.',
        'See also https://www.nag.com/numeric/py/nagdoc_latest/replace.html',
    ]
)
def bounds_bobyqa_func(objfun, npt, x, bl, bu, rhobeg, rhoend, maxcal, monfun=None, data=None):
    u"""
    Bound constrained minimum, model-based algorithm, using function
    values only.

    ``bounds_bobyqa_func`` is an easy-to-use algorithm that uses methods
    of quadratic approximation to find a minimum of an objective
    function F over x in R^n, subject to fixed lower and upper bounds on
    the independent variables x_1,x_2,...,x_n.
    Derivatives of F are not required.

    The function is intended for functions that are continuous and that
    have continuous first and second derivatives (although it will
    usually work even if the derivatives have occasional
    discontinuities).
    Efficiency is maintained for large n.

    Note: ``bounds_bobyqa_func`` is deprecated.
    Please use ``handle_solve_dfno`` and ``handle_solve_dfno_rcomm``
    instead.
    See also https://www.nag.com/numeric/py/nagdoc_latest/replace.html

    For full information please refer to the NAG Library document for
    e04jc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04jcf.html

    Parameters
    ----------
    objfun : callable f = objfun(x, data=None)
        `objfun` must evaluate the objective function F at a specified
        vector x.

        Parameters
        ~~~~~~~~~~
        x : float, ndarray, shape (n)
            x, the vector at which the objective function is to be
            evaluated.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        f : float
            Must be set to the value of the objective function at x.

    npt : int
        m, the number of interpolation conditions imposed on the
        quadratic approximation at each iteration.

        Suggested value: `npt` = 2*n_r+1, where n_r denotes the number
        of non-fixed variables.

    x : float, array-like, shape (n)
        An estimate of the position of the minimum. If any component is
        out-of-bounds it is replaced internally by the bound it
        violates.

    bl : float, array-like, shape (n)
        `bl` must contain the fixed vector of lower bounds, l.

    bu : float, array-like, shape (n)
        `bu` must contain the fixed vector of upper bounds, u.

    rhobeg : float
        An initial lower bound on the value of the trust region radius.

        Suggested value: `rhobeg` should be about one tenth of the
        greatest expected overall change to a variable: the initial
        quadratic model will be constructed by taking steps from the
        initial `x` of length `rhobeg` along each coordinate direction.

    rhoend : float
        A final lower bound on the value of the trust region radius.

        Suggested value: `rhoend` should indicate the absolute accuracy
        that is required in the final values of the variables.

    maxcal : int
        The maximum permitted number of calls to `objfun`.

    monfun : None or callable monfun(nf, x, f, rho, data=None), optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `monfun` may be used to monitor the optimization process.

        It is invoked every time a new trust region radius is chosen.

        If no monitoring is required, `monfun` may be None.

        Parameters
        ~~~~~~~~~~
        nf : int
            The cumulative number of calls made to `objfun`.

        x : float, ndarray, shape (n)
            The current best point.

        f : float
            The value of `objfun` at `x`.

        rho : float
            A lower bound on the current trust region radius.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

    data : arbitrary, optional
        User-communication data for callback functions.

    Returns
    -------
    x : float, ndarray, shape (n)
        The lowest point found during the calculations. Thus, if no
        exception or warning is raised on exit, `x` is the position of
        the minimum.

    f : float
        The function value at the lowest point found (`x`).

    nf : int
        Unless `errno` = 1 or ``NagMemoryError`` on exit, the total
        number of calls made to `objfun`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            There were n_r = *<value>* unequal bounds and `npt` =
            *<value>* on entry.

            Constraint: n_r+2 <= `npt` <= (n_r+1)*(n_r+2)/2.

        (`errno` 1)
            There were n_r = *<value>* unequal bounds.

            Constraint: n_r >= 2.

        (`errno` 1)
            On entry, `rhobeg` = *<value>*, `bl`[i-1] = *<value>*,
            `bu`[i-1] = *<value>* and i = *<value>*.

            Constraint: if `bl`[i-1] != `bu`[i-1] in coordinate i,
            `bu`[i-1]-`bl`[i-1] >= 2*`rhobeg`.

        (`errno` 1)
            On entry, `rhobeg` = *<value>*.

            Constraint: `rhobeg` > 0.0.

        (`errno` 1)
            On entry, `rhoend` = *<value>*.

            Constraint: `rhoend` >= macheps, where macheps =
            machine.precision(), the machine precision.

        (`errno` 1)
            On entry, `rhobeg` = *<value>* and `rhoend` = *<value>*.

            Constraint: `rhoend` <= `rhobeg`.

        (`errno` 1)
            On entry, `maxcal` = *<value>*.

            Constraint: `maxcal` >= 1.

    Warns
    -----
    NagAlgorithmicMajorWarning
        (`errno` 2)
            The function evaluations limit was reached: `objfun` has
            been called `maxcal` times.

        (`errno` 3)
            The predicted reduction in a trust region step was
            non-positive. Check your specification of `objfun` and
            whether the function needs rescaling. Try a different
            initial `x`.

        (`errno` 4)
            A rescue procedure has been called in order to correct
            damage from rounding errors when computing an update to a
            quadratic approximation of F, but no further progess could
            be made. Check your specification of `objfun` and whether
            the function needs rescaling. Try a different initial `x`.

    NagCallbackTerminateWarning
        (`errno` 5)
            User-supplied objective function requested termination.

        (`errno` 5)
            User-supplied monitoring function requested termination.

    References
    ----------
    Powell, M J D, 2009, `The BOBYQA algorithm for bound constrained
    optimization without derivatives`, Report DAMTP 2009/NA06,
    University of Cambridge,
    https://www.damtp.cam.ac.uk/user/na/NA_papers/NA2009_06.pdf
    """

    def cb_objfun(n, x, data=None):
        """Wrapping user's objfun. Called as base objfun."""
        if data is not None:
            _f_objfun = objfun(
                x,
                data,
            )
        else:
            _f_objfun = objfun(
                x,
            )
        _f_objfun_base = _f_objfun
        return _f_objfun_base

    def cb_monfun(n, nf, x, f, rho, data=None):
        """Wrapping user's monfun. Called as base monfun."""
        if data is not None:
            monfun(
                nf,
                x,
                f,
                rho,
                data,
            )
        else:
            monfun(
                nf,
                x,
                f,
                rho,
            )

    fun_name = 'naginterfaces.library.opt.bounds_bobyqa_func'

    _b_utils._check_callable(objfun, {'fun_name': fun_name, 'entity': 'objfun'})
    _objfun = cb_objfun
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _bl = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'bl'},
        bl,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _bu = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'bu'},
        bu,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b_utils._check_callable(monfun, {'fun_name': fun_name, 'entity': 'monfun'}, allow_none=True)
    _monfun = None if monfun is None else cb_monfun
    try:
        _return_scalars = _b_opt.bounds_bobyqa_func(
            _objfun,
            _n,
            npt,
            _x,
            _bl,
            _bu,
            rhobeg,
            rhoend,
            _monfun,
            maxcal,
            data,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _x_py = _x
    _f_py = _return_scalars.f
    _nf_py = _return_scalars.nf
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'f', 'nf'],
    )(
        _x_py,
        _f_py,
        _nf_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def handle_solve_dfno(handle, objfun, x, monit=None, data=None, io_manager=None):
    u"""
    Direct communication derivative-free (DFO) solver for a nonlinear
    objective function with bounded variables.

    Note: this function uses optional algorithmic parameters, see also:
    ``handle_opt_set``, ``handle_opt_get``.

    ``handle_solve_dfno`` is a forward communication Derivative-free
    Optimization (DFO) solver from the NAG optimization modelling suite
    (DFNO) for small to medium-scale nonlinear problems with bound
    constraints.

    For full information please refer to the NAG Library document for
    e04jd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04jdf.html

    Parameters
    ----------
    handle : Handle
        The handle to the problem. It needs to be initialized by
        ``handle_init`` and **must not** be changed before the call to
        ``handle_solve_dfno``.

    objfun : callable (fx, inform) = objfun(x, inform, data=None)
        `objfun` calculates the value of the objective function f(x) at
        a specified point x.

        Parameters
        ~~~~~~~~~~
        x : float, ndarray, shape (nvar)
            x, the vector of variable values at which the objective
            function is to be evaluated.

        inform : int
            A non-negative value.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        fx : float
            The value of the objective function at x.

        inform : int
            May be used to indicate that the requested objective value
            could not be computed. Specifically, it can be set to a
            negative value:

            `inform` = -1

                The solver will attempt a rescue procedure and request
                an alternative point. If the rescue procedure fails, the
                solver will exit with `errno` = 17.

            `inform` = -2

                The solver will cleanly exit with `errno` = 20 and
                return the best available point as well as the solve
                statistics.

    x : float, array-like, shape (nvar)
        x_0, the initial estimates of the variables x.

    monit : None or callable monit(x, rinfo, stats, data=None), optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `monit` is provided to enable you to monitor the progress of the
        optimization.

        It is invoked at the end of every ith iteration where i is given
        by the 'DFO Monitor Frequency' (the default is 0, `monit` is not
        called).

        If no monitoring is required, `monit` may be None.

        Parameters
        ~~~~~~~~~~
        x : float, ndarray, shape (nvar)
            x, the vector of decision variables at the current
            iteration.

        rinfo : float, ndarray, shape (100)
            Error measures and various indicators at the end of the
            current iteration as described in the main argument `rinfo`.

        stats : float, ndarray, shape (100)
            Solver statistics at monitoring steps or at the end of the
            current iteration (the values are as described in the main
            argument `stats`).

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

    data : arbitrary, optional
        User-communication data for callback functions.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    x : float, ndarray, shape (nvar)
        The final values of the variables x.

    rinfo : float, ndarray, shape (100)
        Optimal objective value and various indicators at monitoring
        steps or at the end of the final iteration. The measures are
        given in the table below:

        [table omitted]

    stats : float, ndarray, shape (100)
        Solver statistics at monitoring steps or at the end of the final
        iteration as given in the table below:

        [table omitted]

    Other Parameters
    ----------------
    'Defaults' : valueless
        This special keyword may be used to reset all options to their
        default values.
        Any value given with this keyword will be ignored.

    'DFNO Detect Unbounded' : str
        Default = 'YES'

        The solver can try to detect whether the problem is unbounded.
        This option allows you to turn on or off the unboundedness
        detection heuristic.

        Constraint: 'DFNO Detect Unbounded' = 'YES' or 'NO'.

    'DFNO Objective Limit' : float
        Default = - infinity

        This option sets an additional convergence criterion.
        The solver will stop if it finds a point for which the function
        value is lower than this parameter.

    'DFO Maximum Slow Steps' : int
        Default = 20

        If 'DFO Maximum Slow Steps' > 0, this argument defines the
        maximum number of consecutive slow iterations n_slow allowed.
        Set 'DFO Maximum Slow Steps' = 0 to deactivate the slow
        iteration detection.
        The algorithm can stop in two situations:

        (i) n_slow > 'DFO Maximum Slow Steps' and rho < 'DFO Trust
            Region Slow Tol' with `errno` = 50,

        (#) n_slow > 5*'DFO Maximum Slow Steps' with `errno` = 24.

        Constraint: 'DFO Maximum Slow Steps' >= 0.

    'DFO Max Objective Calls' : int
        Default = 500

        A limit on the number of objective function evaluations the
        solver is allowed to compute.
        If the limit is reached, the solver stops with `errno` = 21.

        Constraint: 'DFO Max Objective Calls' >= 1.

    'DFO Max Soft Restarts' : int
        Default = 5

        The maximum total number of soft restarts that can be performed
        if the objective function is declared as noisy ('DFO Noisy
        Problem' = 'YES').

        Constraint: 'DFO Max Soft Restarts' >= 1.

    'DFO Max Unsucc Soft Restarts' : int
        Default = 3

        The maximum number of consecutive unsuccessful soft restarts
        that can be performed if the objective function is declared as
        noisy ('DFO Noisy Problem' = 'YES').

        Constraint: 'DFO Max Unsucc Soft Restarts' >= 1.

    'DFO Monitor Frequency' : int
        Default = 0

        If 'DFO Monitor Frequency' > 0, `monit` will be called at the
        end of every ith step for monitoring purposes.

        Constraint: 'DFO Monitor Frequency' >= 0.

    'DFO Noise Level' : float
        Default = 0.0

        Indicates the noise level expected when evaluating the objective
        function if 'DFO Noisy Problem' = 'YES'.

        Constraint: 'DFO Noise Level' >= 0.0.

    'DFO Noisy Problem' : str
        Default = 'NO'

        Indicates if the function evaluations provided to the solver are
        noisy.
        If 'DFO Noisy Problem' = 'YES', some algorithmic features will
        be activated:

        (i) The trust region update becomes slower to reflect the
            decreased confidence in the objective values.

        (#) Soft restarts of the algorithm can be performed to ensure
            the algorithm did not get stuck because of the noise (see
            'DFO Max Soft Restarts', 'DFO Max Unsucc Soft Restarts',
            'DFO Number Soft Restarts Pts' to control the restart
            characteristics).

        (#) In addition, if 'DFO Noise Level' > 0.0, the solver will
            trigger a soft restart if all the function values are within
            the noise level.

    'DFO Number Interp Points' : int
        Default = 0

        The maximum number of interpolation points in Y_k (9) used to
        build the linear models of the residuals.
        If 'DFO Number Interp Points' = 0, the number of points is
        chosen to be n_r+1 where n_r is the number of non-fixed
        variables.

        Constraint: 'DFO Number Interp Points' >= 0.

        Consistency constraint, the solver stops with `errno` = 6 if not
        met:

            n_r+1 <= 'DFO Number Interp Points' <= (n_r+1)*(n_r+2)/2.

    'DFO Number Soft Restarts Pts' : int
        Default = 3

        The number of interpolation points that are replaced during a
        soft restart.

        Constraint: 'DFO Number Soft Restarts Pts' >= 1.

    'DFO Print Frequency' : int
        Default = 1

        If 'DFO Print Frequency' > 0, the solver prints the iteration
        log to the appropriate units at the end of every ith step.

        Constraint: 'DFO Print Frequency' >= 0.

    'DFO Random Seed' : int
        Default = -1

        The random seed used to generate the random points used to build
        the initial model.
        If 'DFO Random Seed' < 0, the random seed will be based on
        values taken from the real-time clock, potentially resulting in
        the solver taking a different path each time it is run.
        Set it to a positive value to get fully reproducible runs.

        Constraint: 'DFO Print Frequency' >= -1.

    'DFO Starting Trust Region' : float
        Default = 0.1

        rho_beg, the initial trust region radius.
        This argument should be set to about one tenth of the greatest
        expected overall change to a variable: the initial quadratic
        model will be constructed by taking steps from the initial x of
        length rho_beg along each coordinate direction.
        The default value assumes that the variables have an order of
        magnitude 1.

        Constraint: 'DFO Starting Trust Region' > epsilon.

        Consistency constraints, the solver stops with `errno` = 5 if
        not met:

            'DFO Starting Trust Region' <= 'DFO Trust Region Tolerance'.

            'DFO Starting Trust Region' <= 1/2min_i(u_x(i)-l_x(i)).

    'DFO Trust Region Slow Tol' : float
        Default = epsilon^0.25

        The minimal acceptable trust region radius for the solution to
        be declared as acceptable.
        The solver stops if:

            n_slow > 'DFO Maximum Slow Steps' and rho_k < 'DFO Trust
            Region Slow Tol'.

        Constraint: 'DFO Trust Region Slow Tol' > epsilon.

        Consistency constraint, the solver stops with `errno` = 5 if not
        met:

            'DFO Trust Region Slow Tol' > 'DFO Trust Region Tolerance'.

    'DFO Trust Region Tolerance' : float
        Default = epsilon^0.37

        rho_end, the requested trust region radius.
        The algorithm declares convergence when the trust region radius
        reaches this limit.
        It should indicate the absolute accuracy that is required in the
        final values of the variables.

        Constraint: 'DFO Trust Region Tolerance' > epsilon.

        Consistency constraints, the solver stops with `errno` = 5 if
        not met:

            'DFO Starting Trust Region' > 'DFO Trust Region Tolerance'.

            'DFO Trust Region Slow Tol' > 'DFO Trust Region Tolerance'.

    'Infinite Bound Size' : float
        Default = 10^20

        This defines the 'infinite' bound bigbnd in the definition of
        the problem constraints.
        Any upper bound greater than or equal to bigbnd will be regarded
        as + infinity (and similarly any lower bound less than or equal
        to -bigbnd will be regarded as - infinity).
        Note that a modification of this option does not influence
        constraints which have already been defined; only the
        constraints formulated after the change will be affected.

        Constraint: 'Infinite Bound Size' >= 1000.

    'Monitoring File' : int
        Default = -1

        If i >= 0, the unit number for the secondary (monitoring)
        output.
        If 'Monitoring File' = -1, no secondary output is provided.
        The information output to this unit is controlled by 'Monitoring
        Level'.

        Constraint: 'Monitoring File' >= -1.

    'Monitoring Level' : int
        Default = 4

        This argument sets the amount of information detail that will be
        printed by the solver to the secondary output.
        The meaning of the levels is the same as with 'Print Level'.

        Constraint: 0 <= 'Monitoring Level' <= 5.

    'Print File' : int
        Default = advisory message unit number

        If i >= 0, the unit number for the primary output of the solver.
        If 'Print File' = -1, the primary output is completely turned
        off independently of other settings. The default value is the
        advisory message unit number at the time of the options
        initialization, e.g., at the initialization of the handle. The
        information output to this unit is controlled by 'Print Level'.

        Constraint: 'Print File' >= -1.

    'Print Level' : int
        Default = 2

        This argument defines how detailed information should be printed
        by the solver to the primary and secondary output.

        [table omitted]

        Constraint: 0 <= 'Print Level' <= 5.

    'Print Options' : str
        Default = 'YES'

        If 'Print Options' = 'YES', a listing of options will be printed
        to the primary output and is always printed to the secondary
        output.

        Constraint: 'Print Options' = 'YES' or 'NO'.

    'Print Solution' : str
        Default = 'NO'

        If 'Print Solution' = 'YES', the solution will be printed to the
        primary and secondary output.

        Constraint: 'Print Solution' = 'YES' or 'NO'.

    'Task' : str
        Default = 'MINIMIZE'

        This argument specifies the required direction of the
        optimization.
        If 'Task' = 'FEASIBLE POINT', the objective function (if set) is
        ignored and the algorithm stops as soon as a feasible point is
        found with respect to the given tolerance.

        Constraint: 'Task' = 'MINIMIZE', 'MAXIMIZE' or 'FEASIBLE POINT'.

    'Stats Time' : str
        Default = 'NO'

        This argument turns on timings of various parts of the algorithm
        to give a better overview of where most of the time is spent.
        This might be helpful for a choice of different solving
        approaches.
        It is possible to choose between CPU and wall clock time.
        Choice 'YES' is equivalent to 'WALL CLOCK'.

        Constraint: 'Stats Time' = 'YES', 'NO', 'CPU' or 'WALL CLOCK'.

    'Time Limit' : float
        Default = 10^6

        A limit to the number of seconds that the solver can use to
        solve one problem.
        If during the convergence check this limit is exceeded, the
        solver will terminate with `errno` = 23.

        Constraint: 'Time Limit' > 0.

    Raises
    ------
    NagValueError
        (`errno` 1)
            `handle` has not been initialized.

        (`errno` 1)
            `handle` does not belong to the NAG optimization modelling
            suite, has not been initialized properly or is corrupted.

        (`errno` 1)
            `handle` has not been initialized properly or is corrupted.

        (`errno` 2)
            This solver does not support the model defined in the
            handle.

        (`errno` 2)
            The problem is already being solved.

        (`errno` 4)
            The information supplied does not match with that previously
            stored.

            On entry, nvar = *<value>* must match that given during
            initialization of the `handle`, i.e., *<value>*.

        (`errno` 5)
            Inconsistent options 'DFO Trust Region Tolerance' rho_end
            and 'DFO Starting Trust Region' rho_beg.

            Constraint: rho_end < rho_beg.

            Use ``handle_opt_set`` to set compatible option values.

        (`errno` 5)
            Inconsistent options 'DFO Trust Region Tolerance' rho_end
            and 'DFO Trust Region Slow Tol' rho_tol.

            Constraint: rho_end < rho_tol.

            Use ``handle_opt_set`` to set compatible option values.

        (`errno` 5)
            Option 'DFO Starting Trust Region' rho_beg = *<value>*,
            l_x(i) = *<value>*, u_x(i) = *<value>* and i = *<value>*.

            Constraint: if l_x(i) != u_x(i) in coordinate i, then
            u_x(i)-l_x(i) >= 2*rho_beg.

            Use ``handle_opt_set`` to set compatible option values.

        (`errno` 6)
            There were n_r = *<value>* unequal bounds and the option
            'DFO Number Interp Points' npt = *<value>*.

            Constraint: n_r+1 <= npt <= (n_r+1)*(n_r+2)/2.

            Use ``handle_opt_set`` to set compatible option values.

        (`errno` 6)
            The number of initial interpolation points is different from
            the total set by 'DFO Number Interp Points'.

            Growing the interpolation set is not supported for this
            solver.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 50)
            The problem was solved to an acceptable level after
            *<value>* consecutive slow iterations.

            Use the option 'DFO Maximum Slow Steps' to modify the
            maximum number of slow steps accepted.

    NagAlgorithmicMajorWarning
        (`errno` 17)
            Rescue failed: the trust region could not be reduced further
            after some function evaluation could not be provided. Check
            the specification of your objective and whether it needs
            rescaling. Try a different initial `x`.

        (`errno` 17)
            Some initial interpolation points could not be provided.
            Rescue cannot be attempted at this stage.

            Check the specification of your objective and whether it
            needs rescaling. Try a different initial `x`.

        (`errno` 18)
            The predicted reduction in a trust region step was
            non-positive. Check the specification of your objective and
            whether it needs rescaling. Try a different initial `x`.

        (`errno` 19)
            The solver failed at the model building phase and the
            maximum number of restarts was reached. Check the
            specification of your objective and whether it needs
            rescaling. Try a different initial `x`.

        (`errno` 20)
            User requested termination after a call to the objective
            function.

        (`errno` 21)
            Maximum number of function evaluations exceeded.

        (`errno` 23)
            The solver terminated after the maximum time allowed was
            exceeded.

        (`errno` 24)
            No progress, the solver was stopped after *<value>*
            consecutive slow steps.

            Use the option 'DFO Maximum Slow Steps' to modify the
            maximum number of slow steps accepted.

        (`errno` 54)
            The problem seems to be unbounded.

            The unboundedness detection heuristic can be turned off with
            the option 'DFNO Detect Unbounded'.

    NagCallbackTerminateWarning
        (`errno` 20)
            User requested termination during a monitoring step.

    References
    ----------
    Cartis, C, Fiala, J, Marteau, B and Roberts, L, 2018, `Improving the
    Flexibility and Robustness of Model-Based Derivative-Free
    Optimization Solvers`, Technical Report, University of Oxford

    Conn, A R, Scheinberg, K and Vicente, L N, 2009, `Introduction to
    Derivative-Free Optimization, vol. 8 of MPS-SIAM Series on
    Optimization`, MPS/SIAM, Philadelphia

    Powell, M J D, 2009, `The BOBYQA algorithm for bound constrained
    optimization without derivatives`, Report DAMTP 2009/NA06,
    University of Cambridge,
    https://www.damtp.cam.ac.uk/user/na/NA_papers/NA2009_06.pdf

    See Also
    --------
    :meth:`naginterfaces.library.examples.opt.handle_solve_dfno_ex.main`
    """

    def cb_objfun(nvar, x, inform, data=None):
        """Wrapping user's objfun. Called as base objfun."""
        if data is not None:
            _fx_objfun, _inform_objfun = objfun(
                x,
                inform,
                data,
            )
        else:
            _fx_objfun, _inform_objfun = objfun(
                x,
                inform,
            )
        _inform_objfun_base = _inform_objfun
        _fx_objfun_base = _fx_objfun
        return (
            _fx_objfun_base,
            _inform_objfun_base,
        )

    def cb_monit(nvar, x, rinfo, stats, data=None):
        """Wrapping user's monit. Called as base monit."""
        if data is not None:
            monit(
                x,
                rinfo,
                stats,
                data,
            )
        else:
            monit(
                x,
                rinfo,
                stats,
            )

    fun_name = 'naginterfaces.library.opt.handle_solve_dfno'

    _b_utils._check_callable(objfun, {'fun_name': fun_name, 'entity': 'objfun'})
    _objfun = cb_objfun
    _b_utils._check_callable(monit, {'fun_name': fun_name, 'entity': 'monit'}, allow_none=True)
    _monit = None if monit is None else cb_monit
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nvar_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _nvar = _b_utils._EngineIntScalarType(
        _nvar_py, {'fun_name': fun_name, 'entity': '_nvar'},
    ).to_ctypes()
    _rinfo = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (100,),
    )
    _stats = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (100,),
    )
    try:
        _b_opt.handle_solve_dfno(
            handle,
            _objfun,
            _monit,
            _nvar,
            _x,
            _rinfo,
            _stats,
            data,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _x_py = _x
    _rinfo_py = _rinfo
    _stats_py = _stats
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'rinfo', 'stats'],
    )(
        _x_py,
        _rinfo_py,
        _stats_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def handle_solve_dfno_rcomm(handle, irevcm, x, f, io_manager=None):
    u"""
    Reverse communication derivative-free (DFO) solver for a nonlinear
    objective function with bounded variables.

    Note: this function uses optional algorithmic parameters, see also:
    ``handle_opt_set``, ``handle_opt_get``.

    ``handle_solve_dfno_rcomm`` is a reverse communication
    Derivative-free Optimization (DFO) solver from the NAG optimization
    modelling suite (DFNO) for small to medium-scale nonlinear problems
    with bound constraints.

    For full information please refer to the NAG Library document for
    e04je

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04jef.html

    Parameters
    ----------
    handle : Handle
        The handle to the problem. It needs to be initialized by
        ``handle_init`` and **must not** be changed before the call to
        ``handle_solve_dfno_rcomm``.

    irevcm : int
        Does not need to be set on the first call of
        ``handle_solve_dfno_rcomm``. On subsequent calls, `irevcm` must
        be set to a positive integer if all the required function
        evaluations have been correctly provided in `f`. Otherwise, if a
        problem occurred during a monitoring step or while providing
        objective values, it is possible to set it to a negative value:

        `irevcm` = -1

            If function evaluations were required, the solver will
            attempt a rescue procedure and request an alternative point.
            If no function were required (monitoring step), the solver
            will stop with `errno` = 20.

        `irevcm` <= -2

            The solver will cleanly exit and return the best available
            point as a well as the solve statistics.

    x : float, array-like, shape (nvar, maxeval)
        The first column contains x_0, the initial estimates of the
        variables x.

    f : float, array-like, shape (maxeval)
        Does not need to be set on the first call to
        ``handle_solve_dfno_rcomm``.

        If `irevcm` = 1 after the last call of
        ``handle_solve_dfno_rcomm``, the first `neval` elements must
        contain the function values of the requested points.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    irevcm : int
        Indicates what action is to be performed before the next call to
        ``handle_solve_dfno_rcomm``.

        `irevcm` = 0

            Final exit of the solver.

        `irevcm` = 1

            `neval` objective evaluations are required.

        `irevcm` = 2

            Monitoring step, no evaluation is required, `x` and `f`
            contain the best evaluation of the objective yet.

    neval : int
        Indicates the number of objective evaluations required for the
        next call of ``handle_solve_dfno_rcomm`` in `f`. The coordinates
        of the points to evaluate are provided in the first `neval`
        columns of `x`.

    x : float, ndarray, shape (nvar, maxeval)
        If `irevcm` = 0 or 2, the first column contains the best
        computed estimate of the solution.

        If `irevcm` = 1, the first `neval` columns contain the
        coordinates of the points to evaluate.

    f : float, ndarray, shape (maxeval)
        If `irevcm` = 0 or 2, the first element contains the function
        value of the best computed point.

    rinfo : float, ndarray, shape (100)
        Optimal objective value and various indicators at monitoring
        steps or at the end of the final iteration. The measures are
        given in the table below:

        [table omitted]

    stats : float, ndarray, shape (100)
        Solver statistics at monitoring steps or at the end of the final
        iteration as given in the table below:

        [table omitted]

    Other Parameters
    ----------------
    'Defaults' : valueless
        This special keyword may be used to reset all options to their
        default values.
        Any value given with this keyword will be ignored.

    'DFNO Detect Unbounded' : str
        Default = 'YES'

        The solver can try to detect whether the problem is unbounded.
        This option allows you to turn on or off the unboundedness
        detection heuristic.

        Constraint: 'DFNO Detect Unbounded' = 'YES' or 'NO'.

    'DFNO Objective Limit' : float
        Default = - infinity

        This option sets an additional convergence criterion.
        The solver will stop if it finds a point for which the function
        value is lower than this parameter.

    'DFO Maximum Slow Steps' : int
        Default = 20

        If 'DFO Maximum Slow Steps' > 0, this argument defines the
        maximum number of consecutive slow iterations n_slow allowed.
        Set 'DFO Maximum Slow Steps' = 0 to deactivate the slow
        iteration detection.
        The algorithm can stop in two situations:

        (i) n_slow > 'DFO Maximum Slow Steps' and rho < 'DFO Trust
            Region Slow Tol' with `errno` = 50,

        (#) n_slow > 5*'DFO Maximum Slow Steps' with `errno` = 24.

        Constraint: 'DFO Maximum Slow Steps' >= 0.

    'DFO Max Objective Calls' : int
        Default = 500

        A limit on the number of objective function evaluations the
        solver is allowed to compute.
        If the limit is reached, the solver stops with `errno` = 21.

        Constraint: 'DFO Max Objective Calls' >= 1.

    'DFO Max Soft Restarts' : int
        Default = 5

        The maximum total number of soft restarts that can be performed
        if the objective function is declared as noisy ('DFO Noisy
        Problem' = 'YES').

        Constraint: 'DFO Max Soft Restarts' >= 1.

    'DFO Max Unsucc Soft Restarts' : int
        Default = 3

        The maximum number of consecutive unsuccessful soft restarts
        that can be performed if the objective function is declared as
        noisy ('DFO Noisy Problem' = 'YES').

        Constraint: 'DFO Max Unsucc Soft Restarts' >= 1.

    'DFO Monitor Frequency' : int
        Default = 0

        If 'DFO Monitor Frequency' > 0, the solver will stop at the end
        of every ith step for monitoring purposes.
        ``handle_solve_dfno_rcomm`` needs to be called again to continue
        the optimization.

        Constraint: 'DFO Monitor Frequency' >= 0.

    'DFO Noise Level' : float
        Default = 0.0

        Indicates the noise level expected when evaluating the objective
        function.
        If 'DFO Noisy Problem' = 'YES'.

        Constraint: 'DFO Noise Level' >= 0.0.

    'DFO Noisy Problem' : str
        Default = 'NO'

        Indicates if the function evaluations provided to the solver are
        noisy.
        If 'DFO Noisy Problem' = 'YES', some algorithmic features will
        be activated:

        (i) The trust region update becomes slower to reflect the
            decreased confidence in the objective values.

        (#) Soft restarts of the algorithm can be performed to ensure
            the algorithm did not get stuck because of the noise (see
            'DFO Max Soft Restarts', 'DFO Max Unsucc Soft Restarts',
            'DFO Number Soft Restarts Pts' to control the restart
            characteristics).

        (#) In addition, if 'DFO Noise Level' > 0.0, the solver will
            trigger a soft restart if all the function values are within
            the noise level.

    'DFO Number Interp Points' : int
        Default = 0

        The maximum number of interpolation points in Y_k (9) used to
        build the linear models of the residuals.
        If 'DFO Number Interp Points' = 0, the number of points is
        chosen to be n_r+1 where n_r is the number of non-fixed
        variables.

        Constraint: 'DFO Number Interp Points' >= 0.

        Consistency constraint, the solver stops with `errno` = 6 if not
        met:

            n_r+2 <= 'DFO Number Interp Points' <= (n_r+1)*(n_r+2)/2.

    'DFO Number Soft Restarts Pts' : int
        Default = 3

        The number of interpolation points that are replaced during a
        soft restart.

        Constraint: 'DFO Number Soft Restarts Pts' >= 1.

    'DFO Print Frequency' : int
        Default = 1

        If 'DFO Print Frequency' > 0, the solver prints the iteration
        log to the appropriate units at the end of every ith step.

        Constraint: 'DFO Print Frequency' >= 0.

    'DFO Random Seed' : int
        Default = -1

        The random seed used to generate the random points used to build
        the initial model.
        If 'DFO Random Seed' < 0, the random seed will be based on
        values taken from the real-time clock, potentially resulting in
        the solver taking a different path each time it is run.
        Set it to a positive value to get fully reproducible runs.

        Constraint: 'DFO Print Frequency' >= -1.

    'DFO Starting Trust Region' : float
        Default = 0.1

        rho_beg, the initial trust region radius.
        This argument should be set to about one tenth of the greatest
        expected overall change to a variable: the initial quadratic
        model will be constructed by taking steps from the initial x of
        length rho_beg along each coordinate direction.
        The default value assumes that the variables have an order of
        magnitude 1.

        Constraint: 'DFO Starting Trust Region' > epsilon.

        Consistency constraints, the solver stops with `errno` = 5 if
        not met:

            'DFO Starting Trust Region' <= 'DFO Trust Region Tolerance'.

            'DFO Starting Trust Region' <= 1/2min_i(u_x(i)-l_x(i)).

    'DFO Trust Region Slow Tol' : float
        Default = epsilon^0.25

        The minimal acceptable trust region radius for the solution to
        be declared as acceptable.
        The solver stops if:

            n_slow > 'DFO Maximum Slow Steps' and rho_k < 'DFO Trust
            Region Slow Tol'.

        Constraint: 'DFO Trust Region Slow Tol' > epsilon.

        Consistency constraint, the solver stops with `errno` = 5 if not
        met:

            'DFO Trust Region Slow Tol' > 'DFO Trust Region Tolerance'.

    'DFO Trust Region Tolerance' : float
        Default = epsilon^0.37

        rho_end, the requested trust region radius.
        The algorithm declares convergence when the trust region radius
        reaches this limit.
        It should indicate the absolute accuracy that is required in the
        final values of the variables.

        Constraint: 'DFO Trust Region Tolerance' > epsilon.

        Consistency constraint, the solver stops with `errno` = 5 if not
        met:

            'DFO Starting Trust Region' > 'DFO Trust Region Tolerance'.

    'Infinite Bound Size' : float
        Default = 10^20

        This defines the 'infinite' bound bigbnd in the definition of
        the problem constraints.
        Any upper bound greater than or equal to bigbnd will be regarded
        as + infinity (and similarly any lower bound less than or equal
        to -bigbnd will be regarded as - infinity).
        Note that a modification of this option does not influence
        constraints which have already been defined; only the
        constraints formulated after the change will be affected.

        Constraint: 'Infinite Bound Size' >= 1000.

    'Monitoring File' : int
        Default = -1

        If i >= 0, the unit number for the secondary (monitoring)
        output.
        If 'Monitoring File' = -1, no secondary output is provided.
        The information output to this unit is controlled by 'Monitoring
        Level'.

        Constraint: 'Monitoring File' >= -1.

    'Monitoring Level' : int
        Default = 4

        This argument sets the amount of information detail that will be
        printed by the solver to the secondary output.
        The meaning of the levels is the same as with 'Print Level'.

        Constraint: 0 <= 'Monitoring Level' <= 5.

    'Print File' : int
        Default = advisory message unit number

        If i >= 0, the unit number for the primary output of the solver.
        If 'Print File' = -1, the primary output is completely turned
        off independently of other settings. The default value is the
        advisory message unit number at the time of the options
        initialization, e.g., at the initialization of the handle. The
        information output to this unit is controlled by 'Print Level'.

        Constraint: 'Print File' >= -1.

    'Print Level' : int
        Default = 2

        This argument defines how detailed information should be printed
        by the solver to the primary and secondary output.

        [table omitted]

        Constraint: 0 <= 'Print Level' <= 5.

    'Print Options' : str
        Default = 'YES'

        If 'Print Options' = 'YES', a listing of options will be printed
        to the primary output and is always printed to the secondary
        output.

        Constraint: 'Print Options' = 'YES' or 'NO'.

    'Print Solution' : str
        Default = 'NO'

        If 'Print Solution' = 'YES', the solution will be printed to the
        primary and secondary output.

        Constraint: 'Print Solution' = 'YES' or 'NO'.

    'Task' : str
        Default = 'MINIMIZE'

        This argument specifies the required direction of the
        optimization.
        If 'Task' = 'FEASIBLE POINT', the objective function (if set) is
        ignored and the algorithm stops as soon as a feasible point is
        found with respect to the given tolerance.

        Constraint: 'Task' = 'MINIMIZE', 'MAXIMIZE' or 'FEASIBLE POINT'.

    'Stats Time' : str
        Default = 'NO'

        This argument turns on timings of various parts of the algorithm
        to give a better overview of where most of the time is spent.
        This might be helpful for a choice of different solving
        approaches.
        It is possible to choose between CPU and wall clock time.
        Choice 'YES' is equivalent to 'WALL CLOCK'.

        Constraint: 'Stats Time' = 'YES', 'NO', 'CPU' or 'WALL CLOCK'.

    'Time Limit' : float
        Default = 10^6

        A limit to the number of seconds that the solver can use to
        solve one problem.
        If during the convergence check this limit is exceeded, the
        solver will terminate with `errno` = 23.

        Constraint: 'Time Limit' > 0.

    Raises
    ------
    NagValueError
        (`errno` 1)
            `handle` has not been initialized.

        (`errno` 1)
            `handle` does not belong to the NAG optimization modelling
            suite, has not been initialized properly or is corrupted.

        (`errno` 1)
            `handle` has not been initialized properly or is corrupted.

        (`errno` 2)
            This solver does not support the model defined in the
            handle.

        (`errno` 2)
            The problem is already being solved.

        (`errno` 4)
            On entry, maxeval = *<value>*.

            Constraint: maxeval > 0.

        (`errno` 4)
            The information supplied does not match with that previously
            stored.

            On entry, nvar = *<value>* must match that given during
            initialization of the `handle`, i.e., *<value>*.

        (`errno` 4)
            The information supplied does not match with that previously
            stored.

            On entry, maxeval = *<value>* must match that given during
            the first call of the function, i.e., *<value>*.

        (`errno` 5)
            Inconsistent options 'DFO Trust Region Tolerance' rho_end
            and 'DFO Starting Trust Region' rho_beg.

            Constraint: rho_end < rho_beg.

            Use ``handle_opt_set`` to set compatible option values.

        (`errno` 5)
            Inconsistent options 'DFO Trust Region Tolerance' rho_end
            and 'DFO Trust Region Slow Tol' rho_tol.

            Constraint: rho_end < rho_tol.

            Use ``handle_opt_set`` to set compatible option values.

        (`errno` 5)
            Option 'DFO Starting Trust Region' rho_beg = *<value>*,
            l_x(i) = *<value>*, u_x(i) = *<value>* and i = *<value>*.

            Constraint: if l_x(i) != u_x(i) in coordinate i, then
            u_x(i)-l_x(i) >= 2*rho_beg.

            Use ``handle_opt_set`` to set compatible option values.

        (`errno` 6)
            There were n_r = *<value>* unequal bounds and the option
            'DFO Number Interp Points' npt = *<value>*.

            Constraint: n_r+1 <= npt <= (n_r+1)*(n_r+2)/2.

            Use ``handle_opt_set`` to set compatible option values.

        (`errno` 6)
            The number of initial interpolation points is different from
            the total set by 'DFO Number Interp Points'.

            Growing the interpolation set is not supported for this
            solver.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 50)
            The problem was solved to an acceptable level after
            *<value>* consecutive slow iterations.

            Use the option 'DFO Maximum Slow Steps' to modify the
            maximum number of slow steps accepted.

    NagAlgorithmicMajorWarning
        (`errno` 17)
            Rescue failed: the trust region could not be reduced further
            after some function evaluation could not be provided. Check
            the specification of your objective and whether it needs
            rescaling. Try a different initial `x`.

        (`errno` 17)
            Some initial interpolation points could not be provided.
            Rescue cannot be attempted at this stage.

            Check the specification of your objective and whether it
            needs rescaling. Try a different initial `x`.

        (`errno` 18)
            The predicted reduction in a trust region step was
            non-positive. Check the specification of your objective and
            whether it needs rescaling. Try a different initial `x`.

        (`errno` 19)
            The solver failed at the model building phase and the
            maximum number of restarts was reached. Check the
            specification of your objective and whether it needs
            rescaling. Try a different initial `x`.

        (`errno` 21)
            Maximum number of function evaluations exceeded.

        (`errno` 23)
            The solver terminated after the maximum time allowed was
            exceeded.

        (`errno` 24)
            No progress, the solver was stopped after *<value>*
            consecutive slow steps.

            Use the option 'DFO Maximum Slow Steps' to modify the
            maximum number of slow steps accepted.

        (`errno` 54)
            The problem seems to be unbounded.

            The unboundedness detection heuristic can be turned off with
            the option 'DFNO Detect Unbounded'.

    NagCallbackTerminateWarning
        (`errno` 20)
            User requested termination during a monitoring step.

        (`errno` 20)
            User requested termination during an objective evaluation
            step.

    References
    ----------
    Cartis, C, Fiala, J, Marteau, B and Roberts, L, 2018, `Improving the
    Flexibility and Robustness of Model-Based Derivative-Free
    Optimization Solvers`, Technical Report, University of Oxford

    Conn, A R, Scheinberg, K and Vicente, L N, 2009, `Introduction to
    Derivative-Free Optimization, vol. 8 of MPS-SIAM Series on
    Optimization`, MPS/SIAM, Philadelphia

    Powell, M J D, 2009, `The BOBYQA algorithm for bound constrained
    optimization without derivatives`, Report DAMTP 2009/NA06,
    University of Cambridge,
    https://www.damtp.cam.ac.uk/user/na/NA_papers/NA2009_06.pdf
    """

    fun_name = 'naginterfaces.library.opt.handle_solve_dfno_rcomm'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _maxeval_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2)[1]
    _maxeval = _b_utils._EngineIntScalarType(
        _maxeval_py, {'fun_name': fun_name, 'entity': '_maxeval'},
    ).to_ctypes()
    _nvar_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2)[0]
    _nvar = _b_utils._EngineIntScalarType(
        _nvar_py, {'fun_name': fun_name, 'entity': '_nvar'},
    ).to_ctypes()
    _f = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'f'},
        f,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _rinfo = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (100,),
    )
    _stats = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (100,),
    )
    try:
        _return_scalars = _b_opt.handle_solve_dfno_rcomm(
            handle,
            irevcm,
            _maxeval,
            _nvar,
            _x,
            _f,
            _rinfo,
            _stats,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _irevcm_py = _return_scalars.irevcm
    _x_py = _x
    _f_py = _f
    _neval_py = _return_scalars.neval
    _rinfo_py = _rinfo
    _stats_py = _stats
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['irevcm', 'neval', 'x', 'f', 'rinfo', 'stats'],
    )(
        _irevcm_py,
        _neval_py,
        _x_py,
        _f_py,
        _rinfo_py,
        _stats_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def bounds_quasi_func_easy(ibound, funct1, bl, bu, x, liw=None, lw=None, data=None):
    u"""
    Bound constrained minimum, quasi-Newton algorithm, using function
    values only (easy-to-use).

    ``bounds_quasi_func_easy`` is an easy-to-use quasi-Newton algorithm
    for finding a minimum of a function F(x_1,x_2,...,x_n), subject to
    fixed upper and lower bounds of the independent variables
    x_1,x_2,...,x_n, using function values only.

    It is intended for functions which are continuous and which have
    continuous first and second derivatives (although it will usually
    work even if the derivatives have occasional discontinuities).

    For full information please refer to the NAG Library document for
    e04jy

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04jyf.html

    Parameters
    ----------
    ibound : int
        Indicates whether the facility for dealing with bounds of
        special forms is to be used.

        It must be set to one of the following values:

        `ibound` = 0

            If you are supplying all the l_j and u_j individually.

        `ibound` = 1

            If there are no bounds on any x_j.

        `ibound` = 2

            If all the bounds are of the form 0 <= x_j.

        `ibound` = 3

            If l_1 = l_2 = ... = l_n and u_1 = u_2 = ... = u_n.

    funct1 : callable fc = funct1(xc, data=None)
        You must supply `funct1` to calculate the value of the function
        F(x) at any point x.

        It should be tested separately before being used with
        ``bounds_quasi_func_easy`` (see [the E04 Introduction]).

        Parameters
        ~~~~~~~~~~
        xc : float, ndarray, shape (n)
            The point x at which the function value is required.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        fc : float
            The value of the function F at the current point x.

    bl : float, array-like, shape (n)
        The lower bounds l_j.

        If `ibound` is set to 0, you must set `bl`[j-1] to l_j, for j =
        1,2,...,n. (If a lower bound is not specified for a particular
        x_j, the corresponding `bl`[j-1] should be set to -10^6.)

        If `ibound` is set to 3, you must set `bl`[0] to l_1;
        ``bounds_quasi_func_easy`` will then set the remaining elements
        of `bl` equal to `bl`[0].

    bu : float, array-like, shape (n)
        The upper bounds u_j.

        If `ibound` is set to 0, you must set `bu`[j-1] to u_j, for j =
        1,2,...,n. (If an upper bound is not specified for a particular
        x_j, the corresponding `bu`[j-1] should be set to 10^6.)

        If `ibound` is set to 3, you must set `bu`[0] to u_1;
        ``bounds_quasi_func_easy`` will then set the remaining elements
        of `bu` equal to `bu`[0].

    x : float, array-like, shape (n)
        `x`[j-1] must be set to an estimate of the jth component of the
        position of the minimum, for j = 1,2,...,n.

    liw : None or int, optional
        Note: if this argument is None then a default value will be
        used, determined as follows: n+2.

        The dimension of the array `iw`.

    lw : None or int, optional
        Note: if this argument is None then a default value will be
        used, determined as follows: max(n*(n-1)/2+12*n, 13).

        The dimension of the array `w`.

    data : arbitrary, optional
        User-communication data for callback functions.

    Returns
    -------
    bl : float, ndarray, shape (n)
        The lower bounds actually used by ``bounds_quasi_func_easy``.

    bu : float, ndarray, shape (n)
        The upper bounds actually used by ``bounds_quasi_func_easy``.

    x : float, ndarray, shape (n)
        The lowest point found during the calculations. Thus, if no
        exception or warning is raised on exit, `x`[j-1] is the jth
        component of the position of the minimum.

    f : float
        The value of F(x) corresponding to the final point stored in
        `x`.

    iw : int, ndarray, shape (`liw`)
        If the function exits successfully or `errno` = 3 or 5, the
        first n elements of `iw` contain information about which
        variables are currently on their bounds and which are free.
        Specifically, if x_i is:

        -   fixed on its upper bound, `iw`[i-1] is -1;

        -   fixed on its lower bound, `iw`[i-1] is -2;

        -   effectively a constant (i.e., l_j = u_j), `iw`[i-1] is -3;

        -   free, `iw`[i-1] gives its position in the sequence of free
            variables.

        In addition, `iw`[n] contains the number of free variables
        (i.e., n_z).

        The rest of the array is used as workspace.

    w : float, ndarray, shape (`lw`)
        If the function exits successfully or `errno` = 3 or 5, `w`[i-1]
        contains a finite difference approximation to the ith element of
        the projected gradient vector g_z, for i = 1,2,...,n. In
        addition, `w`[n] contains an estimate of the condition number of
        the projected Hessian matrix (i.e., k). The rest of the array is
        used as workspace.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 1)
            On entry, `ibound` = *<value>*.

            Constraint: 0 <= `ibound` <= 3.

        (`errno` 1)
            On entry, `lw` = *<value>*.

            Constraint: `lw` >= *<value>*.

        (`errno` 1)
            On entry, `liw` = *<value>*.

            Constraint: `liw` >= *<value>*.

        (`errno` 1)
            On entry, either `ibound` = 0 and `bl`[j-1] > `bu`[j-1] for
            some j, or `ibound` = 3 and `bl`[0] > `bu`[0].

        (`errno` 2)
            There have been 400*n calls to `funct1`.

        (`errno` 4)
            An overflow occurred during computation.

        (`errno` 9)
            The modulus of a variable has become very large. There may
            be a mistake in `funct1`, your problem has no finite
            solution, or the problem needs rescaling.

        (`errno` 10)
            One of the forward differences was negative.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            The conditions for a minimum have not all been met but a
            lower point could not be found and the algorithm has failed.

        (`errno` 5)
            It is probable that a local minimum has been found, but it
            cannot be guaranteed.

        (`errno` 6)
            It is probable that a local minimum has been found, but it
            cannot be guaranteed.

        (`errno` 7)
            It is unlikely that a local minimum has been found.

        (`errno` 8)
            It is very unlikely that a local minimum has been found.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Gill, P E and Murray, W, 1976, `Minimization subject to bounds on
    the variables`, NPL Report NAC 72, National Physical Laboratory
    """

    def cb_funct1(n, xc, data=None):
        """Wrapping user's funct1. Called as base funct1."""
        if data is not None:
            _fc_funct1 = funct1(
                xc,
                data,
            )
        else:
            _fc_funct1 = funct1(
                xc,
            )
        _fc_funct1_base = _fc_funct1
        return _fc_funct1_base

    fun_name = 'naginterfaces.library.opt.bounds_quasi_func_easy'

    _bl = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'bl'},
        bl,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_bl, {'fun_name': fun_name, 'entity': 'bl'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b_utils._check_callable(funct1, {'fun_name': fun_name, 'entity': 'funct1'})
    _funct1 = cb_funct1
    _bu = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'bu'},
        bu,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _liw = _n_py+2 if liw is None else liw
    _liw = _b_utils._EngineIntScalarType(
        _liw, {'fun_name': fun_name, 'entity': '_liw'},
    ).to_ctypes()
    _liw_py = _b_utils._EngineIntScalarType.to_py(_liw)
    _iw = _b_utils._EngineIntArrayType.empty_ndarray(
        (_liw_py,),
    )
    _lw = max(_n_py*(_n_py-1)//2+12*_n_py, 13) if lw is None else lw
    _lw = _b_utils._EngineIntScalarType(
        _lw, {'fun_name': fun_name, 'entity': '_lw'},
    ).to_ctypes()
    _lw_py = _b_utils._EngineIntScalarType.to_py(_lw)
    _w = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_lw_py,),
    )
    try:
        _return_scalar = _b_opt.bounds_quasi_func_easy(
            _n,
            ibound,
            _funct1,
            _bl,
            _bu,
            _x,
            _iw,
            _liw,
            _w,
            _lw,
            data,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _bl_py = _bl
    _bu_py = _bu
    _x_py = _x
    _f_py = _return_scalar
    _iw_py = _iw
    _w_py = _w
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['bl', 'bu', 'x', 'f', 'iw', 'w'],
    )(
        _bl_py,
        _bu_py,
        _x_py,
        _f_py,
        _iw_py,
        _w_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def bounds_mod_deriv_comp(funct, monit, eta, ibound, bl, bu, x, lh, iprint=1, maxcal=None, xtol=0.0, delta=0.0, stepmx=100000.0, data=None, io_manager=None):
    u"""
    Bound constrained minimum, modified Newton algorithm, using first
    derivatives (comprehensive).

    ``bounds_mod_deriv_comp`` is a comprehensive modified Newton
    algorithm for finding:

    -   an unconstrained minimum of a function of several variables;

    -   a minimum of a function of several variables subject to fixed
        upper and/or lower bounds on the variables.

    First derivatives are required.
    The function is intended for functions which have continuous first
    and second derivatives (although it will usually work even if the
    derivatives have occasional discontinuities).

    For full information please refer to the NAG Library document for
    e04kd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04kdf.html

    Parameters
    ----------
    funct : callable (fc, gc) = funct(iflag, xc, data=None)
        `funct` must evaluate the function F(x) and its first
        derivatives dF/dx_j at a specified point. (However, if you do
        not wish to calculate F or its first derivatives at a particular
        x, there is the option of setting an argument to cause
        ``bounds_mod_deriv_comp`` to terminate immediately.)

        Parameters
        ~~~~~~~~~~
        iflag : int
            Will have been set to 1 or 2. The value 1 indicates that
            only the first derivatives of F need be supplied, and the
            value 2 indicates that both F itself and its first
            derivatives must be calculated.

        xc : float, ndarray, shape (n)
            The point x at which the dF/dx_j, or F and the dF/dx_j, are
            required.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        fc : float
            Unless `iflag` = 1 on entry, `funct` must set `fc` to the
            value of the objective function F at the current point x.

        gc : float, array-like, shape (n)
            `funct` must set `gc`[j-1] to the value of the first
            derivative dF/dx_j at the point x, for j = 1,2,...,n.

    monit : callable monit(xc, fc, gc, istate, gpjnrm, cond, posdef,
        niter, nf, data=None)
        If `iprint` >= 0, you must supply `monit` which is suitable for
        monitoring the minimization process. `monit` must not change the
        values of any of its arguments.

        If `iprint` < 0, a `monit` with the correct argument list must
        still be supplied, although it will not be called.

        Parameters
        ~~~~~~~~~~
        xc : float, ndarray, shape (n)
            The coordinates of the current point x.

        fc : float
            The value of F(x) at the current point x.

        gc : float, ndarray, shape (n)
            The value of dF/dx_j at the current point x, for j =
            1,2,...,n.

        istate : int, ndarray, shape (n)
            Information about which variables are currently fixed on
            their bounds and which are free.

            If `istate`[j-1] is negative, x_j is currently:

            -   fixed on its upper bound if `istate`[j-1] = -1

            -   fixed on its lower bound if `istate`[j-1] = -2

            -   effectively a constant (i.e., l_j = u_j) if
                `istate`[j-1] = -3

            If `istate`[j-1] is positive, its value gives the position
            of x_j in the sequence of free variables.

        gpjnrm : float
            The Euclidean norm of the current projected gradient vector
            g_z.

        cond : float
            The ratio of the largest to the smallest elements of the
            diagonal factor D of the approximated projected Hessian
            matrix. This quantity is usually a good estimate of the
            condition number of the projected Hessian matrix. (If no
            variables are currently free, `cond` is set to zero.)

        posdef : bool
            Specifies True or False according to whether or not the
            approximation to the second derivative matrix for the
            current subspace, H, is positive definite.

        niter : int
            The number of iterations (as outlined in [Description])
            which have been performed by ``bounds_mod_deriv_comp`` so
            far.

        nf : int
            The number of evaluations of F(x) so far, i.e., the number
            of calls of `funct` with `iflag` set to 2. Each such call of
            `funct` also calculates the first derivatives of F. (In
            addition to these calls monitored by `nf`, `funct` is called
            with `iflag` set to 1 not more than n times per iteration.)

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

    eta : float
        Every iteration of ``bounds_mod_deriv_comp`` involves a linear
        minimization (i.e., minimization of F(x+alpha p) with respect to
        alpha). `eta` specifies how accurately these linear
        minimizations are to be performed. The minimum with respect to
        alpha will be located more accurately for small values of `eta`
        (say, 0.01) than large values (say, 0.9).

        Although accurate linear minimizations will generally reduce the
        number of iterations (and hence the number of calls of `funct`
        to estimate the second derivatives), they will tend to increase
        the number of calls of `funct` needed for each linear
        minimization.

        On balance, it is usually more efficient to perform a low
        accuracy linear minimization when n is small and a high accuracy
        minimization when n is large.

        Suggested value:

            `eta` = 0.5 if 1 < n < 10

            `eta` = 0.1 if 10 <= n <= 20

            `eta` = 0.01 if n > 20

        **If n = 1, eta should be set to 0.0** (also when the problem is
        effectively one-dimensional even though n > 1; i.e., if for all
        except one of the variables the lower and upper bounds are
        equal).

    ibound : int
        Indicates whether the problem is unconstrained or bounded. If
        there are bounds on the variables, `ibound` can be used to
        indicate whether the facility for dealing with bounds of special
        forms is to be used. It must be set to one of the following
        values:

        `ibound` = 0

            If the variables are bounded and you are supplying all the
            l_j and u_j individually.

        `ibound` = 1

            If the problem is unconstrained.

        `ibound` = 2

            If the variables are bounded, but all the bounds are of the
            form 0 <= x_j.

        `ibound` = 3

            If all the variables are bounded, and l_1 = l_2 = ... = l_n
            and u_1 = u_2 = ... = u_n.

        `ibound` = 4

            If the problem is unconstrained. (The `ibound` = 4 option is
            provided for consistency with other functions. In
            ``bounds_mod_deriv_comp`` it produces the same effect as
            `ibound` = 1.)

    bl : float, array-like, shape (n)
        The fixed lower bounds l_j.

        If `ibound` is set to 0, you must set `bl`[j-1] to l_j, for j =
        1,2,...,n. (If a lower bound is not specified for any x_j, the
        corresponding `bl`[j-1] should be set to a large negative
        number, e.g., -10^6.)

        If `ibound` is set to 3, you must set `bl`[0] to l_1;
        ``bounds_mod_deriv_comp`` will then set the remaining elements
        of `bl` equal to `bl`[0].

        If `ibound` is set to 1, 2 or 4, `bl` will be initialized by
        ``bounds_mod_deriv_comp``.

    bu : float, array-like, shape (n)
        The fixed upper bounds u_j.

        If `ibound` is set to 0, you must set `bu`[j-1] to u_j, for j =
        1,2,...,n. (If an upper bound is not specified for any variable,
        the corresponding `bu`[j-1] should be set to a large positive
        number, e.g., 10^6.)

        If `ibound` is set to 3, you must set `bu`[0] to u_1;
        ``bounds_mod_deriv_comp`` will then set the remaining elements
        of `bu` equal to `bu`[0].

        If `ibound` is set to 1, 2 or 4, `bu` will be initialized by
        ``bounds_mod_deriv_comp``.

    x : float, array-like, shape (n)
        `x`[j-1] must be set to a guess at the jth component of the
        position of the minimum, for j = 1,2,...,n.

    lh : int
        The dimension of the array `hesl`.

    iprint : int, optional
        The frequency with which `monit` is to be called.

        `iprint` > 0

            `monit` is called once every `iprint` iterations and just
            before exit from ``bounds_mod_deriv_comp``.

        `iprint` = 0

            `monit` is just called at the final point.

        `iprint` < 0

            `monit` is not called at all.

        `iprint` should normally be set to a small positive number.

    maxcal : None or int, optional
        Note: if this argument is None then a default value will be
        used, determined as follows: 50*n.

        The maximum permitted number of evaluations of F(x), i.e., the
        maximum permitted number of calls of `funct` with `iflag` set to
        2. It should be borne in mind that, in addition to the calls of
        `funct` which are limited directly by `maxcal`, there will be
        calls of `funct` (with `iflag` set to 1) to evaluate only first
        derivatives.

    xtol : float, optional
        The accuracy in x to which the solution is required.

        If x_true is the true value of x at the minimum, then x_sol, the
        estimated position before a normal exit, is such that
        ||x_sol-x_true|| < `xtol`*(1.0+||x_true||) where ||y|| =
        sqrt(sum_j = 1^ny_j^2).

        For example, if the elements of x_sol are not much larger than
        1.0 in modulus, and if `xtol` is set to 10^(-5), then x_sol is
        usually accurate to about five decimal places. (For further
        details see [Accuracy].)

        If the problem is scaled as described in [Further Comments] and
        epsilon is the machine precision, then sqrt(epsilon) is probably
        the smallest reasonable choice for `xtol`.

        This is because, normally, to machine accuracy,
        F(x+sqrt(epsilon)e_j) = F(x), for any j where e_j is the jth
        column of the identity matrix.

        If you set `xtol` to 0.0 (or any positive value less than
        epsilon), ``bounds_mod_deriv_comp`` will use 10.0*sqrt(epsilon)
        instead of `xtol`.

    delta : float, optional
        The differencing interval to be used for approximating the
        second derivatives of F(x). Thus, for the finite difference
        approximations, the first derivatives of F(x) are evaluated at
        points which are `delta` apart. If epsilon is the machine
        precision, sqrt(epsilon) will usually be a suitable setting for
        `delta`. If you set `delta` to 0.0 (or to any positive value
        less than epsilon), ``bounds_mod_deriv_comp`` will automatically
        use sqrt(epsilon) as the differencing interval.

    stepmx : float, optional
        An estimate of the Euclidean distance between the solution and
        the starting point supplied by you. (For maximum efficiency a
        slight overestimate is preferable.)

        ``bounds_mod_deriv_comp`` will ensure that, for each iteration,

            sqrt(sum_j = 1^n[x_j^(k)-x_j^(k-1)]^2) <= `stepmx`,

        where k is the iteration number.

        Thus, if the problem has more than one solution,
        ``bounds_mod_deriv_comp`` is most likely to find the one nearest
        to the starting point.

        On difficult problems, a realistic choice can prevent the
        sequence of x^(k) entering a region where the problem is
        ill-behaved and can also help to avoid possible overflow in the
        evaluation of F(x).

        However, an underestimate of `stepmx` can lead to inefficiency.

    data : arbitrary, optional
        User-communication data for callback functions.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    bl : float, ndarray, shape (n)
        The lower bounds actually used by ``bounds_mod_deriv_comp``,
        e.g., if `ibound` = 2, `bl`[0] = `bl`[1] = ... = `bl`[n-1] =
        0.0.

    bu : float, ndarray, shape (n)
        The upper bounds actually used by ``bounds_mod_deriv_comp``,
        e.g., if `ibound` = 2, `bu`[0] = `bu`[1] = ... = `bu`[n-1] =
        10^6.

    x : float, ndarray, shape (n)
        The final point x^(k). Thus, if no exception or warning is
        raised on exit, `x`[j-1] is the jth component of the estimated
        position of the minimum.

    hesl : float, ndarray, shape (`lh`)
        During the determination of a direction p_z (see [Description]),
        H+E is decomposed into the product LDL^T, where L is a unit
        lower triangular matrix and D is a diagonal matrix. (The
        matrices H, E, L and D are all of dimension n_z, where n_z is
        the number of variables free from their bounds. H consists of
        those rows and columns of the full estimated second derivative
        matrix which relate to free variables. E is chosen so that H+E
        is positive definite.)

        `hesl` and `hesd` are used to store the factors L and D.

        The elements of the strict lower triangle of L are stored row by
        row in the first n_z(n_z-1)/2 positions of `hesl`.

        The diagonal elements of D are stored in the first n_z positions
        of `hesd`.

        In the last factorization before a normal exit, the matrix E
        will be zero, so that `hesl` and `hesd` will contain, on exit,
        the factors of the final estimated second derivative matrix H.

        The elements of `hesd` are useful for deciding whether to accept
        the results produced by ``bounds_mod_deriv_comp`` (see
        [Accuracy]).

    hesd : float, ndarray, shape (n)
        During the determination of a direction p_z (see [Description]),
        H+E is decomposed into the product LDL^T, where L is a unit
        lower triangular matrix and D is a diagonal matrix. (The
        matrices H, E, L and D are all of dimension n_z, where n_z is
        the number of variables free from their bounds. H consists of
        those rows and columns of the full estimated second derivative
        matrix which relate to free variables. E is chosen so that H+E
        is positive definite.)

        `hesl` and `hesd` are used to store the factors L and D.

        The elements of the strict lower triangle of L are stored row by
        row in the first n_z(n_z-1)/2 positions of `hesl`.

        The diagonal elements of D are stored in the first n_z positions
        of `hesd`.

        In the last factorization before a normal exit, the matrix E
        will be zero, so that `hesl` and `hesd` will contain, on exit,
        the factors of the final estimated second derivative matrix H.

        The elements of `hesd` are useful for deciding whether to accept
        the results produced by ``bounds_mod_deriv_comp`` (see
        [Accuracy]).

    istate : int, ndarray, shape (n)
        Information about which variables are currently on their bounds
        and which are free. If `istate`[j-1] is:

        -   equal to -1, x_j is fixed on its upper bound;

        -   equal to -2, x_j is fixed on its lower bound;

        -   equal to -3, x_j is effectively a constant (i.e., l_j =
            u_j);

        -   positive, `istate`[j-1] gives the position of x_j in the
            sequence of free variables.

    f : float
        The function value at the final point given in `x`.

    g : float, ndarray, shape (n)
        The first derivative vector corresponding to the final point
        given in `x`. The components of `g` corresponding to free
        variables should normally be close to zero.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `ibound` = 0 and `bl`[j-1] > `bu`[j-1] for some j.

        (`errno` 1)
            On entry, `ibound` = 3 and `bl`[0] > `bu`[0].

        (`errno` 1)
            On entry, `ibound` = *<value>*.

            Constraint: 0 <= `ibound` <= 4.

        (`errno` 1)
            On entry, `delta` = *<value>*.

            Constraint: `delta` >= 0.0.

        (`errno` 1)
            On entry, `maxcal` = *<value>*.

            Constraint: `maxcal` >= 1.

        (`errno` 1)
            On entry, `stepmx` = *<value>* and `xtol` = *<value>*.

            Constraint: `stepmx` >= `xtol`.

        (`errno` 1)
            On entry, `eta` = *<value>*.

            Constraint: 0.0 <= `eta` < 1.0.

        (`errno` 1)
            On entry, `xtol` = *<value>*.

            Constraint: `xtol` >= 0.0.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 1)
            On entry, `lh` = *<value>*.

            Constraint: `lh` >= *<value>*.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            There have been `maxcal` function evaluations.

        (`errno` 3)
            The conditions for a minimum have not all been satisfied,
            but a lower point could not be found.

        (`errno` 5)
            No further progress can be made.

    NagCallbackTerminateWarning
        (`errno` i < 0)
            User requested termination by setting `iflag` negative in
            `funct`.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Gill, P E and Murray, W, 1973, `Safeguarded steplength algorithms
    for optimization using descent methods`, NPL Report NAC 37, National
    Physical Laboratory

    Gill, P E and Murray, W, 1974, `Newton-type methods for
    unconstrained and linearly constrained optimization`, Math.
    Programming (7), 311--350

    Gill, P E and Murray, W, 1976, `Minimization subject to bounds on
    the variables`, NPL Report NAC 72, National Physical Laboratory
    """

    fun_name = 'naginterfaces.library.opt.bounds_mod_deriv_comp'

    cb_funct_name = fun_name + ':funct'

    def cb_funct(iflag, n, xc, gc, data=None):
        """Wrapping user's funct. Called as base funct."""
        if data is not None:
            _fc_funct, _gc_funct = funct(
                iflag,
                xc,
                data,
            )
        else:
            _fc_funct, _gc_funct = funct(
                iflag,
                xc,
            )
        gc[:] = _utils._to_ndarray(
            {'fun_name': cb_funct_name, 'returned': 'gc'},
            _gc_funct,
            copy=False,
            exp_shape=(n,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _fc_funct_base = _fc_funct
        return _fc_funct_base

    def cb_monit(n, xc, fc, gc, istate, gpjnrm, cond, posdef, niter, nf, data=None):
        """Wrapping user's monit. Called as base monit."""
        if data is not None:
            monit(
                xc,
                fc,
                gc,
                istate,
                gpjnrm,
                cond,
                posdef,
                niter,
                nf,
                data,
            )
        else:
            monit(
                xc,
                fc,
                gc,
                istate,
                gpjnrm,
                cond,
                posdef,
                niter,
                nf,
            )

    _bl = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'bl'},
        bl,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_bl, {'fun_name': fun_name, 'entity': 'bl'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b_utils._check_callable(funct, {'fun_name': fun_name, 'entity': 'funct'})
    _funct = cb_funct
    _b_utils._check_callable(monit, {'fun_name': fun_name, 'entity': 'monit'})
    _monit = cb_monit
    _maxcal = 50*_n_py if maxcal is None else maxcal
    _bu = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'bu'},
        bu,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lh = _b_utils._EngineIntScalarType(
        lh, {'fun_name': fun_name, 'entity': 'lh'},
    ).to_ctypes()
    _lh_py = _b_utils._EngineIntScalarType.to_py(_lh)
    _hesl = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_lh_py,),
    )
    _hesd = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _istate = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n_py,),
    )
    _g = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _return_scalar = _b_opt.bounds_mod_deriv_comp(
            _n,
            _funct,
            _monit,
            iprint,
            _maxcal,
            eta,
            xtol,
            delta,
            stepmx,
            ibound,
            _bl,
            _bu,
            _x,
            _hesl,
            _lh,
            _hesd,
            _istate,
            _g,
            data,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _bl_py = _bl
    _bu_py = _bu
    _x_py = _x
    _hesl_py = _hesl
    _hesd_py = _hesd
    _istate_py = _istate
    _f_py = _return_scalar
    _g_py = _g
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['bl', 'bu', 'x', 'hesl', 'hesd', 'istate', 'f', 'g'],
    )(
        _bl_py,
        _bu_py,
        _x_py,
        _hesl_py,
        _hesd_py,
        _istate_py,
        _f_py,
        _g_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def handle_solve_bounds_foas(handle, x, objfun=None, objgrd=None, monit=None, data=None, io_manager=None):
    u"""
    First-order active-set method for box constrained nonlinear
    optimization with low memory requirements.

    Note: this function uses optional algorithmic parameters, see also:
    ``handle_opt_set``, ``handle_opt_get``.

    ``handle_solve_bounds_foas`` is a solver from the NAG optimization
    modelling suite for bound-constrained large-scale nonlinear
    programming (NLP) problems.
    It is a first-order active-set method (FOAS) that has low memory
    requirements and thus is suitable for very large-scale problems.

    For full information please refer to the NAG Library document for
    e04kf

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04kff.html

    Parameters
    ----------
    handle : Handle
        The handle to the problem. It needs to be initialized by
        ``handle_init`` and the problem formulated by some of the
        functions ``handle_set_linobj``, ``handle_set_quadobj``,
        ``handle_set_nlnobj`` and ``handle_set_simplebounds``. It **must
        not** be changed between calls to the NAG optimization modelling
        suite.

    x : float, array-like, shape (nvar)
        x_0, the initial estimates of the variables x.

    objfun : None or callable fx = objfun(x, inform, data=None),
        optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `objfun` must calculate the value of the nonlinear objective
        function f(x) at a specified point x.

        If there is no nonlinear objective (e.g., ``handle_set_linobj``
        or ``handle_set_quadobj`` was called to define a linear or
        quadratic objective function), `objfun` will never be called by
        ``handle_solve_bounds_foas`` and may be None.

        Parameters
        ~~~~~~~~~~
        x : float, ndarray, shape (nvar)
            The vector x of variable values at which the objective
            function is to be evaluated.

        inform : int
            A non-negative value.

            In some cases, it is known beforehand that the evaluations
            of the objective function and its gradient are required at
            the same point x, in such cases, `inform` = 1.

            This may help to optimize your code in order to avoid
            recalculations of common quantities when evaluating both the
            objective function and gradient; the objective function is
            always evaluated before the objective gradient.

            This notification parameter may be safely ignored if such
            optimization is not required.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        fx : float
            The value of the objective function at x.

    objgrd : None or callable objgrd(x, fdx, inform, data=None),
        optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `objgrd` must calculate the values of the nonlinear objective
        function gradient df/dx at a specified point x.

        Every call to `objgrd` is preceded by a call to `objfun` at the
        same point, if this is known in advance, both functions will be
        notified via `inform` = 1.

        If there is no nonlinear objective (e.g., ``handle_set_linobj``
        or ``handle_set_quadobj`` was called to define a linear or
        quadratic objective function), `objgrd` will never be called by
        ``handle_solve_bounds_foas`` and `objgrd` may be None.

        If the option 'FOAS Estimate Derivatives' = 'YES', then after
        returning from `objgrd` the gradient vector is checked for
        missing entries which you have not supplied.

        Missing entries are estimated using the finite difference
        method, see option 'FOAS Estimate Derivatives' description for
        more details.

        Parameters
        ~~~~~~~~~~
        x : float, ndarray, shape (nvar)
            The vector x of variable values at which the objective
            function gradient is to be evaluated.

        fdx : float, ndarray, shape (nnzfd), to be modified in place
            `On entry`: the elements should only be assigned and not
            referenced.

            `On exit`: the values of the nonzero elements in the sparse
            gradient vector of the objective function, in the order
            specified by idxfd in a previous call to
            ``handle_set_nlnobj``. `fdx`[i-1] will store the gradient
            element df/dx_idxfd[i-1].

        inform : int
            A non-negative value.

            If `inform` = 1, then the previous call to `objfun` was also
            made at the same point x with `inform` = 1.

            This may help to optimize your code in order to avoid
            recalculations of common quantities when evaluating both the
            objective function and gradient.

            This notification parameter may be safely ignored if such
            optimization is not required.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

    monit : None or callable monit(x, rinfo, stats, data=None), optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `monit` is provided to enable you to monitor the progress of the
        optimization and optionally to terminate the solver early if
        necessary.

        It is invoked at the end of every ith iteration where i is given
        by the 'FOAS Monitor Frequency' (the default is 0, `monit` is
        not called).

        `monit` may be None.

        Parameters
        ~~~~~~~~~~
        x : float, ndarray, shape (nvar)
            The vector x of decision variables at the current iteration.

        rinfo : float, ndarray, shape (100)
            Error measures and various indicators at the end of the
            current iteration as described in `rinfo`.

        stats : float, ndarray, shape (100)
            Solver statistics at the end of the current iteration as
            described in `stats`.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

    data : arbitrary, optional
        User-communication data for callback functions.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    x : float, ndarray, shape (nvar)
        The final values of the variables x.

    rinfo : float, ndarray, shape (100)
        Error measures and various indicators at the end of the final
        iteration as given in the table below:

        [table omitted]

    stats : float, ndarray, shape (100)
        Solver statistics at the end of the final iteration as given in
        the table below:

        [table omitted]

    Other Parameters
    ----------------
    'Defaults' : valueless
        This special keyword may be used to reset all options to their
        default values.
        Any value given with this keyword will be ignored.

    'FOAS Estimate Derivatives' : str
        Default = 'NO'

        This option indicates whether to check for and estimate missing
        entries of the user-supplied gradient vector.
        Since the associated cost of estimating missing elements can be
        high, this option should only be used if strictly necessary.
        In general terms, if the gradient is not provided (in its
        entirety or for arbitrary points) potential degradation in the
        progress of the solver is to be expected.
        Depending on the complexity of the objective, the function may
        not achieve the desired optimality accuracy or even terminate
        with no possible further progress error `errno` = 24, it is
        advisable to increase the values of 'FOAS Stop Tolerance' and
        'FOAS Rel Stop Tolerance' when using this option.

        Missing elements from the gradient vector are estimated by
        finite-differences using the perturbation interval specified by
        the option 'FOAS Finite Diff Interval'.

        If 'FOAS Estimate Derivatives' = 'NO', the entries are not
        checked and all derivative elements need to be provided.

        Constraint: 'FOAS Estimate Derivatives' = 'YES' or 'NO'.

    'FOAS Finite Diff Interval' : float
        Default = sqrt(epsilon)

        Specifies the relative perturbation size used to estimate a
        derivative using the forward (or backward) finite-difference
        method.
        Setting the value too small or too big may lead
        ``handle_solve_bounds_foas`` to terminated with `errno` = 24 or
        25.

        Constraint: 10^-12 <= 'FOAS Finite Diff Interval' <= 10^-1.

    'FOAS Iteration Limit' : int
        Default = 10^7

        This argument sets the maximum number of iterations to be
        performed by ``handle_solve_bounds_foas``.
        Setting the option too low might lead to `errno` = 22.

        Constraint: 'FOAS Iteration Limit' >= 1.

    'FOAS Memory' : int
        Default = 11

        This argument specifies the maximum number of memory vectors to
        use in the LCG solver.

        Constraint: 0 <= 'FOAS Memory' <= 100.

    'FOAS Monitor Frequency' : int
        Default = 0

        This argument specifies the frequency on which to call the
        monitor function `monit`.
        If zero, the monitor function will not be called.

        Constraint: 'FOAS Monitor Frequency' >= 0.

    'FOAS Print Frequency' : int
        Default = '1'

        This argument specifies the frequency with which to print
        information regarding each iteration to 'Print File' and/or
        'Monitoring File'.
        By default, it will print information of every iteration.

        Constraint: 'FOAS Print Frequency' >= 1.

    'FOAS Progress Tolerance' : float
        Default = epsilon^(3/4)

        Specifies the tolerance for epsilon_prog (see [equation]) for
        which the function characterises a poor rate of progress given
        that it deems to be far from a solution.
        If this behaviour is persistent, then the function asserts that
        no substantial further progress can be achieved and the process
        is terminated with `errno` = 24.
        Setting a high tolerance can lead to misinterpret reasonable
        progress for unsatisfactory progress or even issue a premature
        stop, see [equation] in [Stopping Criteria].

        Constraint: 0 < 'FOAS Progress Tolerance' < 1.

    'FOAS Rel Stop Tolerance' : float
        Default = epsilon^(3/4)

        This argument sets the value of epsilon_rel which specifies the
        relative tolerance for the convergence measures in the stopping
        criteria, see [equation] and [equation] in [Stopping Criteria].

        Constraint: 0 <= epsilon_rel < 1.

    'FOAS Restart Factor' : float
        Default = 6.0

        This factor specifies the frequency nvar*'FOAS Restart Factor'
        with which the CG/LCG directions are replaced by the steepest
        descent direction (d_k = -g_k).
        Setting the value too small can potentially slow the convergence
        speed.

        Constraint: 'FOAS Restart Factor' >= 0.

    'FOAS Slow Tolerance' : float
        Default = epsilon^(1/8)

        Specifies the tolerance for epsilon_slow (see [equation]) for
        which the function characterises a slow rate of convergence.
        If this behaviour is deemed permanent, then the function asserts
        that no substantial improvement can be achieved and the process
        is terminated with `errno` = 50.
        Setting a large tolerance can lead to incorrectly identifying a
        sub-optimal solution, see [equation] in [Stopping Criteria].

        Constraint: 'FOAS Slow Tolerance' > 0.

    'FOAS Stop Tolerance' : float
        Default = max(10^(-6),sqrt(epsilon))

        This argument sets the value of epsilon_tol which specifies the
        tolerance for the convergence measures in the stopping criteria,
        see [equation] and [equation] in [Stopping Criteria].

        Constraint: 0 <= epsilon_tol < 1.

    'FOAS Tolerance Norm' : str
        Default = 'INFINITY'

        This argument specifies the norm used to measure some stopping
        metrics, such as optimality tolerances (see [Stopping
        Criteria]).
        It is possible to choose between 2-norm and ∞-norm.
        Solving problems using ∞-norm generally has lower computational
        costs than those based on 2-norm.

        Constraint: 'FOAS Tolerance Norm' = 'INFINITY' or 'TWO'.

    'Infinite Bound Size' : float
        Default = 10^20

        This defines the 'infinite' bound bigbnd in the definition of
        the problem constraints.
        Any upper bound greater than or equal to bigbnd will be regarded
        as + infinity (and similarly any lower bound less than or equal
        to -bigbnd will be regarded as - infinity).
        Note that a modification of this option does not influence
        constraints which have already been defined; only the
        constraints formulated after the change will be affected.

        Constraint: 'Infinite Bound Size' >= 1000.

    'Monitoring File' : int
        Default = -1

        If i >= 0, the unit number for the secondary (monitoring)
        output.
        If set to -1, no secondary output is provided.
        The following information is output to the unit:

        -   a listing of the options;

        -   problem statistics, the iteration log and the final status
            as set by 'Monitoring Level';

        -   the solution if set by 'Print Solution'.

        Constraint: 'Monitoring File' >= -1.

    'Monitoring Level' : int
        Default = 4

        This argument sets the amount of information detail that will be
        printed by the solver to the secondary output.
        The meaning of the levels is the same as 'Print Level'.

        Constraint: 0 <= 'Monitoring Level' <= 5.

    'Print File' : int
        Default = advisory message unit number

        If i >= 0, the unit number for the primary output of the solver.
        If 'Print File' = -1, the primary output is completely turned
        off independently of other settings. The default value is the
        advisory message unit number at the time of the option's
        initialization, e.g., at the initialization of the handle. The
        following information is output to the unit:

        -   a listing of options if set by 'Print Options';

        -   problem statistics, the iteration log and the final status
            from the solver as set by 'Print Level';

        -   the solution if set by 'Print Solution'.

        Constraint: 'Print File' >= -1.

    'Print Level' : int
        Default = 2

        This argument defines how detailed information should be printed
        by the solver to the primary output.

        [table omitted]

        Constraint: 0 <= 'Print Level' <= 5.

    'Print Options' : str
        Default = 'YES'

        If 'Print Options' = 'YES', a listing of options will be printed
        to the primary output.

        Constraint: 'Print Options' = 'YES' or 'NO'.

    'Print Solution' : str
        Default = 'NO'

        If 'Print Solution' = 'YES', the final values of the solution
        vector are printed on the primary and secondary outputs.

        Constraint: 'Print Solution' = 'YES' or 'NO'.

    'Stats Time' : str
        Default = 'NO'

        This argument allows you to turn on timings of various parts of
        the algorithm to give a better overview of where most of the
        time is spent.
        This might be helpful for a choice of different solving
        approaches.
        It is possible to choose between CPU and wall clock time.
        Choice 'YES' is equivalent to 'WALL CLOCK'.

        Constraint: 'Stats Time' = 'YES', 'NO', 'CPU' or 'WALL CLOCK'.

    'Task' : str
        Default = 'MINIMIZE'

        This argument specifies the required direction of the
        optimization.
        If 'Task' = 'FEASIBLE POINT', the objective function (if set) is
        ignored and the algorithm stops as soon as a feasible point is
        found.
        If no objective function is set, 'Task' reverts to 'FEASIBLE
        POINT' automatically.

        Constraint: 'Task' = 'MINIMIZE', 'MAXIMIZE' or 'FEASIBLE POINT'.

    'Time Limit' : float
        Default = 10^6

        This argument specifies a limit in seconds that the solver can
        use to solve one problem.
        If during the convergence check this limit is exceeded, the
        solver will terminate with `errno` = 23 error message.

        Constraint: 'Time Limit' > 0.

    'Verify Derivatives' : str
        Default = 'NO'

        This argument specifies whether the function should perform
        numerical checks on the consistency of the user-supplied
        gradient function.
        If any discrepancies are found, `errno` = 26 is returned.
        It is recommended that such checks are enabled when first
        developing the formulation of the problem, however, the
        verification process results in a significant increase of the
        number of the function evaluations and thus it shouldn't be used
        in production code.

        Constraint: 'Verify Derivatives' = 'YES' or 'NO'.

    Raises
    ------
    NagValueError
        (`errno` 1)
            `handle` has not been initialized.

        (`errno` 1)
            `handle` does not belong to the NAG optimization modelling
            suite, has not been initialized properly or is corrupted.

        (`errno` 1)
            `handle` has not been initialized properly or is corrupted.

        (`errno` 2)
            This solver does not support the model defined in the
            handle.

        (`errno` 2)
            The problem is already being solved.

        (`errno` 4)
            The information supplied does not match the value previously
            stored.

            On entry, nvar = *<value>* must match the value given during
            initialization of the `handle`, i.e., *<value>*.

        (`errno` 7)
            Please provide a proper `objfun` function.

        (`errno` 7)
            Please provide a proper `objgrd` function.

        (`errno` 21)
            The current starting point is unusable.

        (`errno` 26)
            User-provided gradient is likely to be incorrect.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 50)
            Problem was solved to an acceptable level; full accuracy was
            not achieved.

        (`errno` 54)
            The problem seems to be unbounded and the algorithm was
            stopped.

    NagAlgorithmicMajorWarning
        (`errno` 22)
            Maximum number of iterations reached.

        (`errno` 23)
            The solver terminated after the maximum time allowed was
            exhausted.

        (`errno` 24)
            The solver was terminated because no further progress could
            be achieved.

        (`errno` 25)
            Invalid number detected in user-supplied function and
            recovery failed.

        (`errno` 25)
            Invalid number detected in user-supplied gradient and
            recovery failed.

    NagCallbackTerminateWarning
        (`errno` 20)
            User requested termination during a monitoring step.

    References
    ----------
    Dai, Y-H and Kou, C-X, 2013, `A Nonlinear Conjugate Gradient
    Algorithm with an Optimal Property and an Improved Wolfe Line
    Search`, SIAM J. Optim. (23(1)), 296--320

    Gill, P E and Leonard, M W, 2003, `Limited-Memory Reduced-Hessian
    Methods for Large-Scale Unconstrained Optimization`, SIAM J. Optim.
    (14(2)), 380--401

    Hager, W W and Zhang, H, 2005, `A New Conjugate Gradient Method with
    Guaranteed Descent and an Efficient Line Search`, SIAM J. Optim.
    (16(1)), 170--192

    Hager, W W and Zhang, H, 2006, `Algorithm 851: CG DESCENT, a
    Conjugate Gradient Method with Guaranteed Descent`, ACM Trans. Math.
    Software (32(1)), 113--137

    Hager, W W and Zhang, H, 2006, `A New Active Set Algorithm for Box
    Constrained Optimization`, SIAM J. Optim. (17(2)), 525--557

    Hager, W W and Zhang, H, 2013, `The Limited Memory Conjugate
    Gradient Method`, SIAM J. Optim. (23(4)), 2150--2168

    Nocedal, J and Wright, S J, 2006, `Numerical Optimization`, (2nd
    Edition), Springer Series in Operations Research, Springer, New York

    See Also
    --------
    :meth:`naginterfaces.library.examples.opt.handle_solve_bounds_foas_ex.main`
    """

    def cb_objfun(nvar, x, inform, data=None):
        """Wrapping user's objfun. Called as base objfun."""
        if data is not None:
            _fx_objfun = objfun(
                x,
                inform,
                data,
            )
        else:
            _fx_objfun = objfun(
                x,
                inform,
            )
        _fx_objfun_base = _fx_objfun
        return _fx_objfun_base

    def cb_objgrd(nvar, x, nnzfd, fdx, inform, data=None):
        """Wrapping user's objgrd. Called as base objgrd."""
        if data is not None:
            objgrd(
                x,
                fdx,
                inform,
                data,
            )
        else:
            objgrd(
                x,
                fdx,
                inform,
            )

    def cb_monit(nvar, x, rinfo, stats, data=None):
        """Wrapping user's monit. Called as base monit."""
        if data is not None:
            monit(
                x,
                rinfo,
                stats,
                data,
            )
        else:
            monit(
                x,
                rinfo,
                stats,
            )

    fun_name = 'naginterfaces.library.opt.handle_solve_bounds_foas'

    _b_utils._check_callable(objfun, {'fun_name': fun_name, 'entity': 'objfun'}, allow_none=True)
    _objfun = None if objfun is None else cb_objfun
    _b_utils._check_callable(objgrd, {'fun_name': fun_name, 'entity': 'objgrd'}, allow_none=True)
    _objgrd = None if objgrd is None else cb_objgrd
    _b_utils._check_callable(monit, {'fun_name': fun_name, 'entity': 'monit'}, allow_none=True)
    _monit = None if monit is None else cb_monit
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nvar_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _nvar = _b_utils._EngineIntScalarType(
        _nvar_py, {'fun_name': fun_name, 'entity': '_nvar'},
    ).to_ctypes()
    _rinfo = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (100,),
    )
    _stats = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (100,),
    )
    try:
        _b_opt.handle_solve_bounds_foas(
            handle,
            _objfun,
            _objgrd,
            _monit,
            _nvar,
            _x,
            _rinfo,
            _stats,
            data,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _x_py = _x
    _rinfo_py = _rinfo
    _stats_py = _stats
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'rinfo', 'stats'],
    )(
        _x_py,
        _rinfo_py,
        _stats_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def bounds_quasi_deriv_easy(ibound, funct2, bl, bu, x, liw=None, lw=None, data=None):
    u"""
    Bound constrained minimum, quasi-Newton algorithm, using first
    derivatives (easy-to-use).

    ``bounds_quasi_deriv_easy`` is an easy-to-use quasi-Newton algorithm
    for finding a minimum of a function F(x_1,x_2,...,x_n), subject to
    fixed upper and lower bounds on the independent variables
    x_1,x_2,...,x_n, when first derivatives of F are available.

    It is intended for functions which are continuous and which have
    continuous first and second derivatives (although it will usually
    work even if the derivatives have occasional discontinuities).

    For full information please refer to the NAG Library document for
    e04ky

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04kyf.html

    Parameters
    ----------
    ibound : int
        Indicates whether the facility for dealing with bounds of
        special forms is to be used. It must be set to one of the
        following values:

        `ibound` = 0

            If you are supplying all the l_j and u_j individually.

        `ibound` = 1

            If there are no bounds on any x_j.

        `ibound` = 2

            If all the bounds are of the form 0 <= x_j.

        `ibound` = 3

            If l_1 = l_2 = ... = l_n and u_1 = u_2 = ... = u_n.

    funct2 : callable (fc, gc) = funct2(xc, data=None)
        You must supply `funct2` to calculate the values of the function
        F(x) and its first derivative dF/dx_j at any point x.

        It should be tested separately before being used in conjunction
        with ``bounds_quasi_deriv_easy`` (see [the E04 Introduction]).

        Parameters
        ~~~~~~~~~~
        xc : float, ndarray, shape (n)
            The point x at which the function and derivatives are
            required.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        fc : float
            The value of the function F at the current point x.

        gc : float, array-like, shape (n)
            `gc`[j-1] must be set to the value of the first derivative
            dF/dx_j at the point x, for j = 1,2,...,n.

    bl : float, array-like, shape (n)
        The lower bounds l_j.

        If `ibound` is set to 0, you must set `bl`[j-1] to l_j, for j =
        1,2,...,n. (If a lower bound is not specified for a particular
        x_j, the corresponding `bl`[j-1] should be set to -10^6.)

        If `ibound` is set to 3, you must set `bl`[0] to l_1;
        ``bounds_quasi_deriv_easy`` will then set the remaining elements
        of `bl` equal to `bl`[0].

    bu : float, array-like, shape (n)
        The upper bounds u_j.

        If `ibound` is set to 0, you must set `bu`[j-1] to u_j, for j =
        1,2,...,n. (If an upper bound is not specified for a particular
        x_j, the corresponding `bu`[j-1] should be set to 10^6.)

        If `ibound` is set to 3, you must set `bu`[0] to u_1;
        ``bounds_quasi_deriv_easy`` will then set the remaining elements
        of `bu` equal to `bu`[0].

    x : float, array-like, shape (n)
        `x`[j-1] must be set to a guess at the jth component of the
        position of the minimum, for j = 1,2,...,n. The function checks
        the gradient at the starting point, and is more likely to detect
        any error in your programming if the initial `x`[j-1] are
        nonzero and mutually distinct.

    liw : None or int, optional
        Note: if this argument is None then a default value will be
        used, determined as follows: n+2.

        The dimension of the array `iw`.

    lw : None or int, optional
        Note: if this argument is None then a default value will be
        used, determined as follows: max(10*n+n*(n-1)/2, 11).

        The dimension of the array `w`.

    data : arbitrary, optional
        User-communication data for callback functions.

    Returns
    -------
    bl : float, ndarray, shape (n)
        The lower bounds actually used by ``bounds_quasi_deriv_easy``.

    bu : float, ndarray, shape (n)
        The upper bounds actually used by ``bounds_quasi_deriv_easy``.

    x : float, ndarray, shape (n)
        The lowest point found during the calculations. Thus, if no
        exception or warning is raised on exit, `x`[j-1] is the jth
        component of the position of the minimum.

    f : float
        The value of F(x) corresponding to the final point stored in
        `x`.

    g : float, ndarray, shape (n)
        The value of dF/dx_j corresponding to the final point stored in
        `x`, for j = 1,2,...,n; the value of `g`[j-1] for variables not
        on a bound should normally be close to zero.

    iw : int, ndarray, shape (`liw`)
        If the function exits successfully or `errno` = 3 or 5, the
        first n elements of `iw` contain information about which
        variables are currently on their bounds and which are free.
        Specifically, if x_i is:

        -   fixed on its upper bound, `iw`[i-1] is -1;

        -   fixed on its lower bound, `iw`[i-1] is -2;

        -   effectively a constant (i.e., l_j = u_j), `iw`[i-1] is -3;

        -   free, `iw`[i-1] gives its position in the sequence of free
            variables.

        In addition, `iw`[n] contains the number of free variables
        (i.e., n_z).

        The rest of the array is used as workspace.

    w : float, ndarray, shape (`lw`)
        If the function exits successfully or `errno` = 3 or 5, `w`[i-1]
        contains the ith element of the projected gradient vector g_z,
        for i = 1,2,...,n. In addition, `w`[n] contains an estimate of
        the condition number of the projected Hessian matrix (i.e., k).
        The rest of the array is used as workspace.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `ibound` = 0 and `bl`[j-1] > `bu`[j-1] for some j.

        (`errno` 1)
            On entry, `ibound` = 3 and `bl`[0] > `bu`[0].

        (`errno` 1)
            On entry, `liw` = *<value>*.

            Constraint: `liw` >= *<value>*.

        (`errno` 1)
            On entry, `lw` = *<value>*.

            Constraint: `lw` >= *<value>*.

        (`errno` 1)
            On entry, `ibound` = *<value>*.

            Constraint: 0 <= `ibound` <= 3.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 2)
            There have been 100*n function evaluations.

        (`errno` 4)
            An overflow occurred during computation.

        (`errno` 9)
            The modulus of a variable has become very large. There may
            be a mistake in `funct2`, your problem has no finite
            solution, or the problem needs rescaling.

        (`errno` 10)
            It is very likely that you have made an error forming the
            gradient.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            The conditions for a minimum have not all been satisfied,
            but a lower point could not be found.

        (`errno` 5)
            It is probable that a local minimum has been found, but it
            cannot be guaranteed.

        (`errno` 6)
            It is possible that a local minimum has been found, but it
            cannot be guaranteed.

        (`errno` 7)
            It is unlikely that a local minimum has been found.

        (`errno` 8)
            It is very unlikely that a local minimum has been found.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Gill, P E and Murray, W, 1976, `Minimization subject to bounds on
    the variables`, NPL Report NAC 72, National Physical Laboratory
    """

    fun_name = 'naginterfaces.library.opt.bounds_quasi_deriv_easy'

    cb_funct2_name = fun_name + ':funct2'

    def cb_funct2(n, xc, gc, data=None):
        """Wrapping user's funct2. Called as base funct2."""
        if data is not None:
            _fc_funct2, _gc_funct2 = funct2(
                xc,
                data,
            )
        else:
            _fc_funct2, _gc_funct2 = funct2(
                xc,
            )
        gc[:] = _utils._to_ndarray(
            {'fun_name': cb_funct2_name, 'returned': 'gc'},
            _gc_funct2,
            copy=False,
            exp_shape=(n,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _fc_funct2_base = _fc_funct2
        return _fc_funct2_base

    _bl = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'bl'},
        bl,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_bl, {'fun_name': fun_name, 'entity': 'bl'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b_utils._check_callable(funct2, {'fun_name': fun_name, 'entity': 'funct2'})
    _funct2 = cb_funct2
    _bu = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'bu'},
        bu,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _g = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _liw = _n_py+2 if liw is None else liw
    _liw = _b_utils._EngineIntScalarType(
        _liw, {'fun_name': fun_name, 'entity': '_liw'},
    ).to_ctypes()
    _liw_py = _b_utils._EngineIntScalarType.to_py(_liw)
    _iw = _b_utils._EngineIntArrayType.empty_ndarray(
        (_liw_py,),
    )
    _lw = max(10*_n_py+_n_py*(_n_py-1)//2, 11) if lw is None else lw
    _lw = _b_utils._EngineIntScalarType(
        _lw, {'fun_name': fun_name, 'entity': '_lw'},
    ).to_ctypes()
    _lw_py = _b_utils._EngineIntScalarType.to_py(_lw)
    _w = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_lw_py,),
    )
    try:
        _return_scalar = _b_opt.bounds_quasi_deriv_easy(
            _n,
            ibound,
            _funct2,
            _bl,
            _bu,
            _x,
            _g,
            _iw,
            _liw,
            _w,
            _lw,
            data,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _bl_py = _bl
    _bu_py = _bu
    _x_py = _x
    _f_py = _return_scalar
    _g_py = _g
    _iw_py = _iw
    _w_py = _w
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['bl', 'bu', 'x', 'f', 'g', 'iw', 'w'],
    )(
        _bl_py,
        _bu_py,
        _x_py,
        _f_py,
        _g_py,
        _iw_py,
        _w_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def bounds_mod_deriv_easy(ibound, funct2, bl, bu, x, data=None):
    u"""
    Bound constrained minimum, modified Newton algorithm, using first
    derivatives (easy-to-use).

    ``bounds_mod_deriv_easy`` is an easy-to-use modified Newton
    algorithm for finding a minimum of a function F(x_1,x_2,...,x_n),
    subject to fixed upper and lower bounds on the independent variables
    x_1,x_2,...,x_n, when first derivatives of F are available.
    It is intended for functions which are continuous and which have
    continuous first and second derivatives (although it will usually
    work even if the derivatives have occasional discontinuities).

    For full information please refer to the NAG Library document for
    e04kz

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04kzf.html

    Parameters
    ----------
    ibound : int
        Indicates whether the facility for dealing with bounds of
        special forms is to be used. It must be set to one of the
        following values:

        `ibound` = 0

            If you are supplying all the l_j and u_j individually.

        `ibound` = 1

            If there are no bounds on any x_j.

        `ibound` = 2

            If all the bounds are of the form 0 <= x_j.

        `ibound` = 3

            If l_1 = l_2 = ... = l_n and u_1 = u_2 = ... = u_n.

    funct2 : callable (fc, gc) = funct2(xc, data=None)
        You must supply this function to calculate the values of the
        function F(x) and its first derivatives dF/dx_j at any point x.

        It should be tested separately before being used in conjunction
        with ``bounds_mod_deriv_easy`` (see submodule ``opt``).

        Parameters
        ~~~~~~~~~~
        xc : float, ndarray, shape (n)
            The point x at which the function and derivatives are
            required.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        fc : float
            The value of the function F at the current point x,

        gc : float, array-like, shape (n)
            `gc`[j-1] must be set to the value of the first derivative
            dF/dx_j at the point x, for j = 1,2,...,n.

    bl : float, array-like, shape (n)
        The lower bounds l_j.

        If `ibound` is set to 0, you must set `bl`[j-1] to l_j, for j =
        1,2,...,n. (If a lower bound is not specified for a particular
        x_j, the corresponding `bl`[j-1] should be set to -10^6.)

        If `ibound` is set to 3, you must set `bl`[0] to l_1;
        ``bounds_mod_deriv_easy`` will then set the remaining elements
        of `bl` equal to `bl`[0].

    bu : float, array-like, shape (n)
        The upper bounds u_j.

        If `ibound` is set to 0, you must set `bu`[j-1] to u_j, for j =
        1,2,...,n. (If an upper bound is not specified for a particular
        x_j, the corresponding `bu`[j-1] should be set to 10^6.)

        If `ibound` is set to 3, you must set `bu`[0] to u_1;
        ``bounds_mod_deriv_easy`` will then set the remaining elements
        of `bu` equal to `bu`[0].

    x : float, array-like, shape (n)
        `x`[j-1] must be set to a guess at the jth component of the
        position of the minimum, for j = 1,2,...,n. The function checks
        the gradient at the starting point, and is more likely to detect
        any error in your programming if the initial `x`[j-1] are
        nonzero and mutually distinct.

    data : arbitrary, optional
        User-communication data for callback functions.

    Returns
    -------
    bl : float, ndarray, shape (n)
        The lower bounds actually used by ``bounds_mod_deriv_easy``.

    bu : float, ndarray, shape (n)
        The upper bounds actually used by ``bounds_mod_deriv_easy``.

    x : float, ndarray, shape (n)
        The lowest point found during the calculations of the position
        of the minimum.

    f : float
        The value of F(x) corresponding to the final point stored in
        `x`.

    g : float, ndarray, shape (n)
        The value of dF/dx_j corresponding to the final point stored in
        `x`, for j = 1,2,...,n; the value of `g`[j-1] for variables not
        on a bound should normally be close to zero.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `ibound` = 0 and `bl`[j-1] > `bu`[j-1] for some j.

        (`errno` 1)
            On entry, `ibound` = 3 and `bl`[0] > `bu`[0].

        (`errno` 1)
            On entry, `ibound` = *<value>*.

            Constraint: 0 <= `ibound` <= 3.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 2)
            There have been 50*n function evaluations.

        (`errno` 9)
            The modulus of a variable has become very large. There may
            be a mistake in `funct2`, your problem has no finite
            solution, or the problem needs rescaling.

        (`errno` 10)
            It is very likely that you have made an error forming the
            gradient.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            The conditions for a minimum have not all been satisfied,
            but a lower point could not be found.

        (`errno` 5)
            It is probable that a local minimum has been found, but it
            cannot be guaranteed.

        (`errno` 6)
            It is possible that a local minimum has been found, but it
            cannot be guaranteed.

        (`errno` 7)
            It is unlikely that a local minimum has been found.

        (`errno` 8)
            It is very unlikely that a local minimum has been found.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Gill, P E and Murray, W, 1976, `Minimization subject to bounds on
    the variables`, NPL Report NAC 72, National Physical Laboratory
    """

    fun_name = 'naginterfaces.library.opt.bounds_mod_deriv_easy'

    cb_funct2_name = fun_name + ':funct2'

    def cb_funct2(n, xc, gc, data=None):
        """Wrapping user's funct2. Called as base funct2."""
        if data is not None:
            _fc_funct2, _gc_funct2 = funct2(
                xc,
                data,
            )
        else:
            _fc_funct2, _gc_funct2 = funct2(
                xc,
            )
        gc[:] = _utils._to_ndarray(
            {'fun_name': cb_funct2_name, 'returned': 'gc'},
            _gc_funct2,
            copy=False,
            exp_shape=(n,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _fc_funct2_base = _fc_funct2
        return _fc_funct2_base

    _bl = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'bl'},
        bl,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_bl, {'fun_name': fun_name, 'entity': 'bl'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b_utils._check_callable(funct2, {'fun_name': fun_name, 'entity': 'funct2'})
    _funct2 = cb_funct2
    _bu = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'bu'},
        bu,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _g = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _return_scalar = _b_opt.bounds_mod_deriv_easy(
            _n,
            ibound,
            _funct2,
            _bl,
            _bu,
            _x,
            _g,
            data,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _bl_py = _bl
    _bu_py = _bu
    _x_py = _x
    _f_py = _return_scalar
    _g_py = _g
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['bl', 'bu', 'x', 'f', 'g'],
    )(
        _bl_py,
        _bu_py,
        _x_py,
        _f_py,
        _g_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def bounds_mod_deriv2_comp(funct, h, monit, ibound, bl, bu, x, lh, iprint=1, maxcal=None, eta=None, xtol=0.0, stepmx=100000.0, data=None, io_manager=None):
    u"""
    Bound constrained minimum, modified Newton algorithm, using first
    and second derivatives (comprehensive).

    ``bounds_mod_deriv2_comp`` is a comprehensive modified Newton
    algorithm for finding:

        an unconstrained minimum of a function of several variables

        a minimum of a function of several variables subject to fixed
        upper and/or lower bounds on the variables.

    First and second derivatives are required.
    The function is intended for functions which have continuous first
    and second derivatives (although it will usually work even if the
    derivatives have occasional discontinuities).

    For full information please refer to the NAG Library document for
    e04lb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04lbf.html

    Parameters
    ----------
    funct : callable (fc, gc) = funct(xc, data=None)
        `funct` must evaluate the function F(x) and its first
        derivatives dF/dx_j at any point x.

        Parameters
        ~~~~~~~~~~
        xc : float, ndarray, shape (n)
            The point x at which F and the dF/dx_j are required.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        fc : float
            `funct` must set `fc` to the value of the objective function
            F at the current point x.

        gc : float, array-like, shape (n)
            `funct` must set `gc`[j-1] to the value of the first
            derivative dF/dx_j at the point x, for j = 1,2,...,n.

    h : callable (fhesl, fhesd) = h(xc, lh, fhesd, data=None)
        `h` must calculate the second derivatives of F at any point x.
        (As with `funct`, there is the option of causing
        ``bounds_mod_deriv2_comp`` to terminate immediately.)

        Parameters
        ~~~~~~~~~~
        xc : float, ndarray, shape (n)
            The point x at which the second derivatives of F are
            required.

        lh : int
            The length of the array `fhesl`.

        fhesd : float, ndarray, shape (n)
            The value of dF/dx_j at the point x, for j = 1,2,...,n.

            These values may be useful in the evaluation of the second
            derivatives.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        fhesl : float, array-like, shape (`lh`)
            `h` must place the strict lower triangle of the second
            derivative matrix of F (evaluated at the point x) in
            `fhesl`, stored by rows, i.e., set `fhesl`[ (i-1) (i-2) / 2
            +j -1] = d^2F/(dx_i dx_j)|_`xc`, for j = 1,2,...,i-1, for i
            = 2,3,...,n. (The upper triangle is not required because the
            matrix is symmetric.)

        fhesd : float, array-like, shape (n)
            `h` must place the diagonal elements of the second
            derivative matrix of F (evaluated at the point x) in
            `fhesd`, i.e., set `fhesd`[j-1] = d^2F/dx_j^2|_`xc`, j =
            1,2,...,n.

    monit : callable monit(xc, fc, gc, istate, gpjnrm, cond, posdef,
        niter, nf, data=None)
        If `iprint` >= 0, you must supply `monit` which is suitable for
        monitoring the minimization process. `monit` must not change the
        values of any of its arguments.

        If `iprint` < 0, a `monit` with the correct argument list should
        still be supplied, although it will not be called.

        Parameters
        ~~~~~~~~~~
        xc : float, ndarray, shape (n)
            The coordinates of the current point x.

        fc : float
            The value of F(x) at the current point x.

        gc : float, ndarray, shape (n)
            The value of dF/dx_j at the current point x, for j =
            1,2,...,n.

        istate : int, ndarray, shape (n)
            Information about which variables are currently fixed on
            their bounds and which are free.

            If `istate`[j-1] is negative, x_j is currently:

            -   fixed on its upper bound if `istate`[j-1] = -1;

            -   fixed on its lower bound if `istate`[j-1] = -2;

            -   effectively a constant (i.e., l_j = u_j) if
                `istate`[j-1] = -3.

            If `istate` is positive, its value gives the position of x_j
            in the sequence of free variables.

        gpjnrm : float
            The Euclidean norm of the projected gradient vector g_z.

        cond : float
            The ratio of the largest to the smallest elements of the
            diagonal factor D of the projected Hessian matrix (see
            specification of `h`). This quantity is usually a good
            estimate of the condition number of the projected Hessian
            matrix. (If no variables are currently free, `cond` is set
            to zero.)

        posdef : bool
            Is set True or False according to whether the second
            derivative matrix for the current subspace, H, is positive
            definite or not.

        niter : int
            The number of iterations (as outlined in [Description])
            which have been performed by ``bounds_mod_deriv2_comp`` so
            far.

        nf : int
            The number of times that `funct` has been called so far.
            Thus `nf` is the number of function and gradient evaluations
            made so far.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

    ibound : int
        Specifies whether the problem is unconstrained or bounded. If
        there are bounds on the variables, `ibound` can be used to
        indicate whether the facility for dealing with bounds of special
        forms is to be used. It must be set to one of the following
        values:

        `ibound` = 0

            If the variables are bounded and you are supplying all the
            l_j and u_j individually.

        `ibound` = 1

            If the problem is unconstrained.

        `ibound` = 2

            If the variables are bounded, but all the bounds are of the
            form 0 <= x_j.

        `ibound` = 3

            If all the variables are bounded, and l_1 = l_2 = ... = l_n
            and u_1 = u_2 = ... = u_n.

        `ibound` = 4

            If the problem is unconstrained. (The `ibound` = 4 option is
            provided purely for consistency with other functions. In
            ``bounds_mod_deriv2_comp`` it produces the same effect as
            `ibound` = 1.)

    bl : float, array-like, shape (n)
        The fixed lower bounds l_j.

        If `ibound` is set to 0, you must set `bl`[j-1] to l_j, for j =
        1,2,...,n. (If a lower bound is not specified for any x_j, the
        corresponding `bl`[j-1] should be set to a large negative
        number, e.g., -10^6.)

        If `ibound` is set to 3, you must set `bl`[0] to l_1;
        ``bounds_mod_deriv2_comp`` will then set the remaining elements
        of `bl` equal to `bl`[0].

        If `ibound` is set to 1, 2 or 4, `bl` will be initialized by
        ``bounds_mod_deriv2_comp``.

    bu : float, array-like, shape (n)
        The fixed upper bounds u_j.

        If `ibound` is set to 0, you must set `bu`[j-1] to u_j, for j =
        1,2,...,n. (If an upper bound is not specified for any variable,
        the corresponding `bu`[j-1] should be set to a large positive
        number, e.g., 10^6.)

        If `ibound` is set to 3, you must set `bu`[0] to u_1;
        ``bounds_mod_deriv2_comp`` will then set the remaining elements
        of `bu` equal to `bu`[0].

        If `ibound` is set to 1, 2 or 4, `bu` will then be initialized
        by ``bounds_mod_deriv2_comp``.

    x : float, array-like, shape (n)
        `x`[j-1] must be set to a guess at the jth component of the
        position of the minimum, for j = 1,2,...,n.

    lh : int
        The dimension of the array `hesl`.

    iprint : int, optional
        The frequency with which `monit` is to be called.

        `iprint` > 0

            `monit` is called once every `iprint` iterations and just
            before exit from ``bounds_mod_deriv2_comp``.

        `iprint` = 0

            `monit` is just called at the final point.

        `iprint` < 0

            `monit` is not called at all.

        `iprint` should normally be set to a small positive number.

    maxcal : None or int, optional
        Note: if this argument is None then a default value will be
        used, determined as follows: 50*n.

        The maximum permitted number of evaluations of F(x), i.e., the
        maximum permitted number of calls of `funct`.

    eta : None or float, optional
        Note: if this argument is None then a default value will be
        used, determined as follows: if n = 1: 0.0; otherwise: 0.9.

        Every iteration of ``bounds_mod_deriv2_comp`` involves a linear
        minimization (i.e., minimization of F(x+alpha p) with respect to
        alpha). `eta` specifies how accurately these linear
        minimizations are to be performed. The minimum with respect to
        alpha will be located more accurately for small values of `eta`
        (say, 0.01) than for large values (say, 0.9).

        Although accurate linear minimizations will generally reduce the
        number of iterations of ``bounds_mod_deriv2_comp``, this usually
        results in an increase in the number of function and gradient
        evaluations required for each iteration.

        On balance, it is usually more efficient to perform a low
        accuracy linear minimization.

    xtol : float, optional
        The accuracy in x to which the solution is required.

        If x_true is the true value of x at the minimum, then x_sol, the
        estimated position before a normal exit, is such that
        ||x_sol-x_true|| < `xtol`*(1.0+||x_true||), where ||y|| =
        sqrt(sum_j = 1^ny_j^2).

        For example, if the elements of x_sol are not much larger than
        1.0 in modulus, and if `xtol` is set to 10^(-5), then x_sol is
        usually accurate to about five decimal places. (For further
        details see [Accuracy].)

        If the problem is scaled roughly as described in [Further
        Comments] and epsilon is the machine precision, then
        sqrt(epsilon) is probably the smallest reasonable choice for
        `xtol`. (This is because, normally, to machine accuracy,
        F(x+sqrt(epsilon),e_j) = F(x) where e_j is any column of the
        identity matrix.)

        If you set `xtol` to 0.0 (or any positive value less than
        epsilon), ``bounds_mod_deriv2_comp`` will use 10.0*sqrt(epsilon)
        instead of `xtol`.

    stepmx : float, optional
        An estimate of the Euclidean distance between the solution and
        the starting point supplied by you. (For maximum efficiency a
        slight overestimate is preferable.)

        ``bounds_mod_deriv2_comp`` will ensure that, for each iteration,

            sqrt(sum_j = 1^n[x_j^(k)-x_j^(k-1)]^2) <= `stepmx`

        where k is the iteration number.

        Thus, if the problem has more than one solution,
        ``bounds_mod_deriv2_comp`` is most likely to find the one
        nearest to the starting point.

        On difficult problems, a realistic choice can prevent the
        sequence of x^(k) entering a region where the problem is
        ill-behaved and can also help to avoid possible overflow in the
        evaluation of F(x).

        However, an underestimate of `stepmx` can lead to inefficiency.

    data : arbitrary, optional
        User-communication data for callback functions.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    bl : float, ndarray, shape (n)
        The lower bounds actually used by ``bounds_mod_deriv2_comp``,
        e.g., if `ibound` = 2, `bl`[0] = `bl`[1] = ... = `bl`[n-1] =
        0.0.

    bu : float, ndarray, shape (n)
        The upper bounds actually used by ``bounds_mod_deriv2_comp``,
        e.g., if `ibound` = 2, `bu`[0] = `bu`[1] = ... = `bu`[n-1] =
        10^6.

    x : float, ndarray, shape (n)
        The final point x^(k). Thus, if no exception or warning is
        raised on exit, `x`[j-1] is the jth component of the estimated
        position of the minimum.

    hesl : float, ndarray, shape (`lh`)
        During the determination of a direction p_z (see [Description]),
        H+E is decomposed into the product LDL^T, where L is a unit
        lower triangular matrix and D is a diagonal matrix. (The
        matrices H, E, L and D are all of dimension n_z, where n_z is
        the number of variables free from their bounds. H consists of
        those rows and columns of the full estimated second derivative
        matrix which relate to free variables. E is chosen so that H+E
        is positive definite.)

        `hesl` and `hesd` are used to store the factors L and D.

        The elements of the strict lower triangle of L are stored row by
        row in the first n_z(n_z-1)/2 positions of `hesl`.

        The diagonal elements of D are stored in the first n_z positions
        of `hesd`.

        In the last factorization before a normal exit, the matrix E
        will be zero, so that `hesl` and `hesd` will contain, on exit,
        the factors of the final estimated second derivative matrix H.

        The elements of `hesd` are useful for deciding whether to accept
        the results produced by ``bounds_mod_deriv2_comp`` (see
        [Accuracy]).

    hesd : float, ndarray, shape (n)
        During the determination of a direction p_z (see [Description]),
        H+E is decomposed into the product LDL^T, where L is a unit
        lower triangular matrix and D is a diagonal matrix. (The
        matrices H, E, L and D are all of dimension n_z, where n_z is
        the number of variables free from their bounds. H consists of
        those rows and columns of the full second derivative matrix
        which relate to free variables. E is chosen so that H+E is
        positive definite.)

        `hesl` and `hesd` are used to store the factors L and D.

        The elements of the strict lower triangle of L are stored row by
        row in the first n_z(n_z-1)/2 positions of `hesl`.

        The diagonal elements of D are stored in the first n_z positions
        of `hesd`.

        In the last factorization before a normal exit, the matrix E
        will be zero, so that `hesl` and `hesd` will contain, on exit,
        the factors of the final second derivative matrix H.

        The elements of `hesd` are useful for deciding whether to accept
        the result produced by ``bounds_mod_deriv2_comp`` (see
        [Accuracy]).

    istate : int, ndarray, shape (n)
        Information about which variables are currently on their bounds
        and which are free. If `istate`[j-1] is:

        -   equal to -1, x_j is fixed on its upper bound;

        -   equal to -2, x_j is fixed on its lower bound;

        -   equal to -3, x_j is effectively a constant (i.e., l_j =
            u_j);

        -   positive, `istate`[j-1] gives the position of x_j in the
            sequence of free variables.

    f : float
        The function value at the final point given in `x`.

    g : float, ndarray, shape (n)
        The first derivative vector corresponding to the final point
        given in `x`. The components of `g` corresponding to free
        variables should normally be close to zero.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `ibound` = 0 and `bl`[j-1] > `bu`[j-1] for some j.

        (`errno` 1)
            On entry, `ibound` = 3 and `bl`[0] > `bu`[0].

        (`errno` 1)
            On entry, `ibound` = *<value>*.

            Constraint: 0 <= `ibound` <= 4.

        (`errno` 1)
            On entry, `maxcal` = *<value>*.

            Constraint: `maxcal` >= 1.

        (`errno` 1)
            On entry, `stepmx` = *<value>* and `xtol` = *<value>*.

            Constraint: `stepmx` >= `xtol`.

        (`errno` 1)
            On entry, `eta` = *<value>*.

            Constraint: 0.0 <= `eta` < 1.0.

        (`errno` 1)
            On entry, `xtol` = *<value>*.

            Constraint: `xtol` >= 0.0.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 1)
            On entry, `lh` = *<value>*.

            Constraint: `lh` >= *<value>*.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            There have been `maxcal` function evaluations.

        (`errno` 3)
            The conditions for a minimum have not all been satisfied,
            but a lower point could not be found.

        (`errno` 5)
            No further progress can be made.

    NagCallbackTerminateWarning
        (`errno` i < 0)
            User requested termination by setting `iflag` negative in
            `funct` or `h`.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Gill, P E and Murray, W, 1973, `Safeguarded steplength algorithms
    for optimization using descent methods`, NPL Report NAC 37, National
    Physical Laboratory

    Gill, P E and Murray, W, 1974, `Newton-type methods for
    unconstrained and linearly constrained optimization`, Math.
    Programming (7), 311--350

    Gill, P E and Murray, W, 1976, `Minimization subject to bounds on
    the variables`, NPL Report NAC 72, National Physical Laboratory
    """

    fun_name = 'naginterfaces.library.opt.bounds_mod_deriv2_comp'

    cb_funct_name = fun_name + ':funct'

    def cb_funct(n, xc, gc, data=None):
        """Wrapping user's funct. Called as base funct."""
        if data is not None:
            _fc_funct, _gc_funct = funct(
                xc,
                data,
            )
        else:
            _fc_funct, _gc_funct = funct(
                xc,
            )
        gc[:] = _utils._to_ndarray(
            {'fun_name': cb_funct_name, 'returned': 'gc'},
            _gc_funct,
            copy=False,
            exp_shape=(n,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _fc_funct_base = _fc_funct
        return _fc_funct_base

    cb_h_name = fun_name + ':h'

    def cb_h(n, xc, fhesl, lh, fhesd, data=None):
        """Wrapping user's h. Called as base h."""
        if data is not None:
            _fhesl_h, _fhesd_h = h(
                xc,
                lh,
                fhesd,
                data,
            )
        else:
            _fhesl_h, _fhesd_h = h(
                xc,
                lh,
                fhesd,
            )
        fhesl[:] = _utils._to_ndarray(
            {'fun_name': cb_h_name, 'returned': 'fhesl'},
            _fhesl_h,
            copy=False,
            exp_shape=(lh,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        fhesd[:] = _utils._to_ndarray(
            {'fun_name': cb_h_name, 'returned': 'fhesd'},
            _fhesd_h,
            copy=False,
            exp_shape=(n,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )

    def cb_monit(n, xc, fc, gc, istate, gpjnrm, cond, posdef, niter, nf, data=None):
        """Wrapping user's monit. Called as base monit."""
        if data is not None:
            monit(
                xc,
                fc,
                gc,
                istate,
                gpjnrm,
                cond,
                posdef,
                niter,
                nf,
                data,
            )
        else:
            monit(
                xc,
                fc,
                gc,
                istate,
                gpjnrm,
                cond,
                posdef,
                niter,
                nf,
            )

    _bl = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'bl'},
        bl,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_bl, {'fun_name': fun_name, 'entity': 'bl'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b_utils._check_callable(funct, {'fun_name': fun_name, 'entity': 'funct'})
    _funct = cb_funct
    _b_utils._check_callable(h, {'fun_name': fun_name, 'entity': 'h'})
    _h = cb_h
    _b_utils._check_callable(monit, {'fun_name': fun_name, 'entity': 'monit'})
    _monit = cb_monit
    _maxcal = 50*_n_py if maxcal is None else maxcal
    if _n_py == 1:
        _eta = 0.0
    else:
        _eta = 0.9
    _eta = _eta if eta is None else eta
    _bu = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'bu'},
        bu,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _lh = _b_utils._EngineIntScalarType(
        lh, {'fun_name': fun_name, 'entity': 'lh'},
    ).to_ctypes()
    _lh_py = _b_utils._EngineIntScalarType.to_py(_lh)
    _hesl = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_lh_py,),
    )
    _hesd = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _istate = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n_py,),
    )
    _g = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _return_scalar = _b_opt.bounds_mod_deriv2_comp(
            _n,
            _funct,
            _h,
            _monit,
            iprint,
            _maxcal,
            _eta,
            xtol,
            stepmx,
            ibound,
            _bl,
            _bu,
            _x,
            _hesl,
            _lh,
            _hesd,
            _istate,
            _g,
            data,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _bl_py = _bl
    _bu_py = _bu
    _x_py = _x
    _hesl_py = _hesl
    _hesd_py = _hesd
    _istate_py = _istate
    _f_py = _return_scalar
    _g_py = _g
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['bl', 'bu', 'x', 'hesl', 'hesd', 'istate', 'f', 'g'],
    )(
        _bl_py,
        _bu_py,
        _x_py,
        _hesl_py,
        _hesd_py,
        _istate_py,
        _f_py,
        _g_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def bounds_mod_deriv2_easy(ibound, funct2, hess2, bl, bu, x, data=None):
    u"""
    Bound constrained minimum, modified Newton algorithm, using first
    and second derivatives (easy-to-use).

    ``bounds_mod_deriv2_easy`` is an easy-to-use modified-Newton
    algorithm for finding a minimum of a function, F(x_1,x_2,...,x_n)
    subject to fixed upper and lower bounds on the independent
    variables, x_1,x_2,...,x_n when first and second derivatives of F
    are available.
    It is intended for functions which are continuous and which have
    continuous first and second derivatives (although it will usually
    work even if the derivatives have occasional discontinuities).

    For full information please refer to the NAG Library document for
    e04ly

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04lyf.html

    Parameters
    ----------
    ibound : int
        Indicates whether the facility for dealing with bounds of
        special forms is to be used. It must be set to one of the
        following values:

        `ibound` = 0

            If you are supplying all the l_j and u_j individually.

        `ibound` = 1

            If there are no bounds on any x_j.

        `ibound` = 2

            If all the bounds are of the form 0 <= x_j.

        `ibound` = 3

            If l_1 = l_2 = ... = l_n and u_1 = u_2 = ... = u_n.

    funct2 : callable (fc, gc) = funct2(xc, data=None)
        You must supply this function to calculate the values of the
        function F(x) and its first derivatives dF/dx_j at any point x.

        It should be tested separately before being used in conjunction
        with ``bounds_mod_deriv2_easy`` (see [the E04 Introduction]).

        Parameters
        ~~~~~~~~~~
        xc : float, ndarray, shape (n)
            The point x at which the function and its derivatives are
            required.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        fc : float
            The value of the function F at the current point x.

        gc : float, array-like, shape (n)
            `gc`[j-1] must be set to the value of the first derivative
            dF/dx_j at the point x, for j = 1,2,...,n.

    hess2 : callable (heslc, hesdc) = hess2(xc, lh, data=None)
        You must supply this function to evaluate the elements H_{ij} =
        d^2F/(dx_i dx_j) of the matrix of second derivatives of F(x) at
        any point x.

        It should be tested separately before being used in conjunction
        with ``bounds_mod_deriv2_easy`` (see [the E04 Introduction]).

        Parameters
        ~~~~~~~~~~
        xc : float, ndarray, shape (n)
            The point x at which the derivatives are required.

        lh : int
            The length of the array `heslc`.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        heslc : float, array-like, shape (`lh`)
            `hess2` must place the strict lower triangle of the second
            derivative matrix H in `heslc`, stored by rows, i.e., set
            `heslc`[(i-1)(i-2)/2+j-1] = d^2F/(dx_i dx_j), for j =
            1,2,...,i-1, for i = 2,3,...,n. (The upper triangle is not
            required because the matrix is symmetric.)

        hesdc : float, array-like, shape (n)
            Must contain the diagonal elements of the second derivative
            matrix, i.e., set `hesdc`[j-1] = d^2F/dx_j^2, for j =
            1,2,...,n.

    bl : float, array-like, shape (n)
        The lower bounds l_j.

        If `ibound` is set to 0, `bl`[j-1] must be set to l_j, for j =
        1,2,...,n. (If a lower bound is not specified for any x_j, the
        corresponding `bl`[j-1] should be set to -10^6.)

        If `ibound` is set to 3, you must set `bl`[0] to l_1;
        ``bounds_mod_deriv2_easy`` will then set the remaining elements
        of `bl` equal to `bl`[0].

    bu : float, array-like, shape (n)
        The upper bounds u_j.

        If `ibound` is set to 0, `bu`[j-1] must be set to u_j, for j =
        1,2,...,n. (If an upper bound is not specified for any x_j the
        corresponding `bu`[j-1] should be set to 10^6.)

        If `ibound` is set to 3, you must set `bu`[0] to u_1;
        ``bounds_mod_deriv2_easy`` will then set the remaining elements
        of `bu` equal to `bu`[0].

    x : float, array-like, shape (n)
        `x`[j-1] must be set to a guess at the jth component of the
        position of the minimum, for j = 1,2,...,n. The function checks
        the gradient and the Hessian matrix at the starting point, and
        is more likely to detect any error in your programming if the
        initial `x`[j-1] are nonzero and mutually distinct.

    data : arbitrary, optional
        User-communication data for callback functions.

    Returns
    -------
    bl : float, ndarray, shape (n)
        The lower bounds actually used by ``bounds_mod_deriv2_easy``.

    bu : float, ndarray, shape (n)
        The upper bounds actually used by ``bounds_mod_deriv2_easy``.

    x : float, ndarray, shape (n)
        The lowest point found during the calculations. Thus, if no
        exception or warning is raised on exit, `x`[j-1] is the jth
        component of the position of the minimum.

    f : float
        The value of F(x) corresponding to the final point stored in
        `x`.

    g : float, ndarray, shape (n)
        The value of dF/dx_j corresponding to the final point stored in
        `x`, for j = 1,2,...,n; the value of `g`[j-1] for variables not
        on a bound should normally be close to zero.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `ibound` = 0 and `bl`[j-1] > `bu`[j-1] for some j.

        (`errno` 1)
            On entry, `ibound` = 3 and `bl`[0] > `bu`[0].

        (`errno` 1)
            On entry, `ibound` = *<value>*.

            Constraint: 0 <= `ibound` <= 3.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 2)
            There have been 50*n function evaluations.

        (`errno` 9)
            The modulus of a variable has become very large. There may
            be a mistake in your supplied functions, your problem has no
            finite solution, or the problem needs rescaling.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            The conditions for a minimum have not all been satisfied,
            but a lower point could not be found.

        (`errno` 5)
            It is probable that a local minimum has been found, but it
            cannot be guaranteed.

        (`errno` 6)
            It is possible that a local minimum has been found, but it
            cannot be guaranteed.

        (`errno` 7)
            It is unlikely that a local minimum has been found.

        (`errno` 8)
            It is very unlikely that a local minimum has been found.

        (`errno` 10)
            It is very likely that you have made an error forming the
            gradient.

        (`errno` 11)
            It is very likely that you have made an error forming the
            2nd derivatives.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Gill, P E and Murray, W, 1976, `Minimization subject to bounds on
    the variables`, NPL Report NAC 72, National Physical Laboratory
    """

    fun_name = 'naginterfaces.library.opt.bounds_mod_deriv2_easy'

    cb_funct2_name = fun_name + ':funct2'

    def cb_funct2(n, xc, gc, data=None):
        """Wrapping user's funct2. Called as base funct2."""
        if data is not None:
            _fc_funct2, _gc_funct2 = funct2(
                xc,
                data,
            )
        else:
            _fc_funct2, _gc_funct2 = funct2(
                xc,
            )
        gc[:] = _utils._to_ndarray(
            {'fun_name': cb_funct2_name, 'returned': 'gc'},
            _gc_funct2,
            copy=False,
            exp_shape=(n,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _fc_funct2_base = _fc_funct2
        return _fc_funct2_base

    cb_hess2_name = fun_name + ':hess2'

    def cb_hess2(n, xc, heslc, lh, hesdc, data=None):
        """Wrapping user's hess2. Called as base hess2."""
        if data is not None:
            _heslc_hess2, _hesdc_hess2 = hess2(
                xc,
                lh,
                data,
            )
        else:
            _heslc_hess2, _hesdc_hess2 = hess2(
                xc,
                lh,
            )
        heslc[:] = _utils._to_ndarray(
            {'fun_name': cb_hess2_name, 'returned': 'heslc'},
            _heslc_hess2,
            copy=False,
            exp_shape=(lh,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        hesdc[:] = _utils._to_ndarray(
            {'fun_name': cb_hess2_name, 'returned': 'hesdc'},
            _hesdc_hess2,
            copy=False,
            exp_shape=(n,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )

    _bl = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'bl'},
        bl,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_bl, {'fun_name': fun_name, 'entity': 'bl'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b_utils._check_callable(funct2, {'fun_name': fun_name, 'entity': 'funct2'})
    _funct2 = cb_funct2
    _b_utils._check_callable(hess2, {'fun_name': fun_name, 'entity': 'hess2'})
    _hess2 = cb_hess2
    _bu = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'bu'},
        bu,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _g = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _return_scalar = _b_opt.bounds_mod_deriv2_easy(
            _n,
            ibound,
            _funct2,
            _hess2,
            _bl,
            _bu,
            _x,
            _g,
            data,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _bl_py = _bl
    _bu_py = _bu
    _x_py = _x
    _f_py = _return_scalar
    _g_py = _g
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['bl', 'bu', 'x', 'f', 'g'],
    )(
        _bl_py,
        _bu_py,
        _x_py,
        _f_py,
        _g_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def lp_solve(nclin, a, bl, bu, cvec, istate, x, comm, io_manager=None):
    u"""
    Linear programming (LP), dense, active-set method.

    Note: this function uses optional algorithmic parameters, see also:
    ``lp_option_file``, ``lp_option_string``, ``nlp1_init``.

    ``lp_solve`` solves general linear programming problems.
    It is not intended for large sparse problems.

    For full information please refer to the NAG Library document for
    e04mf

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04mff.html

    Parameters
    ----------
    nclin : int
        m_L, the number of general linear constraints.

    a : float, array-like, shape (`nclin`, :)
        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `nclin`>0: n; if `nclin`=0: 1;
        otherwise: 0.

        The ith row of `a` must contain the coefficients of the ith
        general linear constraint, for i = 1,2,...,m_L.

        If `nclin` = 0, `a` is not referenced.

    bl : float, array-like, shape (n+`nclin`)
        `bl` must contain the lower bounds for all the constraints

    bu : float, array-like, shape (n+`nclin`)
        `bu` must contain the upper bounds for all the constraints

    cvec : float, array-like, shape (:)
        Note: the required length for this argument is determined as
        follows: if `comm`['iwsav'][43]!=1: n; otherwise: 1.

        The coefficients of the objective function when the problem is
        of type LP.

        If the problem is of type FP, `cvec` is not referenced.

    istate : int, array-like, shape (n+`nclin`)
        Need not be set if the (default) option 'Cold Start' is used.

        If the option 'Warm Start' has been chosen, `istate` specifies
        the desired status of the constraints at the start of the
        feasibility phase.

        More precisely, the first n elements of `istate` refer to the
        upper and lower bounds on the variables, and the next m_L
        elements refer to the general linear constraints (if any).

        Possible values for `istate`[j-1] are as follows:

        [table omitted]

        The values -2, -1 and 4 are also acceptable but will be reset to
        zero by the function.

        If ``lp_solve`` has been called previously with the same values
        of n and `nclin`, `istate` already contains satisfactory
        information. (See also the description of the option 'Warm
        Start'.) The function also adjusts (if necessary) the values
        supplied in `x` to be consistent with `istate`.

    x : float, array-like, shape (n)
        An initial estimate of the solution.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``nlp1_init``.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    istate : int, ndarray, shape (n+`nclin`)
        The status of the constraints in the working set at the point
        returned in `x`. The significance of each possible value of
        `istate`[j-1] is as follows:

        [table omitted]

    x : float, ndarray, shape (n)
        The point at which ``lp_solve`` terminated. If the function
        exits successfully or `errno` = 1 or 4, `x` contains an estimate
        of the solution.

    itera : int
        The total number of iterations performed.

    obj : float
        The value of the objective function at x if x is feasible, or
        the sum of infeasibiliites at x otherwise. If the problem is of
        type FP and x is feasible, `obj` is set to zero.

    ax : float, ndarray, shape (max(1, `nclin`))
        The final values of the linear constraints Ax.

        If `nclin` = 0, `ax` is not referenced.

    clamda : float, ndarray, shape (n+`nclin`)
        The values of the Lagrange multipliers for each constraint with
        respect to the current working set. The first n elements contain
        the multipliers for the bound constraints on the variables, and
        the next m_L elements contain the multipliers for the general
        linear constraints (if any). If `istate`[j-1] = 0 (i.e.,
        constraint j is not in the working set), `clamda`[j-1] is zero.
        If x is optimal, `clamda`[j-1] should be non-negative if
        `istate`[j-1] = 1, non-positive if `istate`[j-1] = 2 and zero if
        `istate`[j-1] = 4.

    Other Parameters
    ----------------
    'Check Frequency' : int
        Default = 50

        Every ith iteration, a numerical test is made to see if the
        current solution x satisfies the constraints in the working set.
        If the largest residual of the constraints in the working set is
        judged to be too large, the current working set is refactorized
        and the variables are recomputed to satisfy the constraints more
        accurately.
        If i <= 0, the default value is used.

    'Cold Start' : valueless
        Default

        This option specifies how the initial working set is chosen.
        With a 'Cold Start', ``lp_solve`` chooses the initial working
        set based on the values of the variables and constraints at the
        initial point.
        Broadly speaking, the initial working set will include equality
        constraints and bounds or inequality constraints that violate or
        'nearly' satisfy their bounds (to within 'Crash Tolerance').

        With a 'Warm Start', you must provide a valid definition of
        every element of the array `istate`. ``lp_solve`` will override
        your specification of `istate` if necessary, so that a poor
        choice of the working set will not cause a fatal error.
        For instance, any elements of `istate` which are set to -2, -1
        or 4 will be reset to zero, as will any elements which are set
        to 3 when the corresponding elements of `bl` and `bu` are not
        equal.
        A warm start will be advantageous if a good estimate of the
        initial working set is available -- for example, when
        ``lp_solve`` is called repeatedly to solve related problems.

    'Warm Start' : valueless
        This option specifies how the initial working set is chosen.
        With a 'Cold Start', ``lp_solve`` chooses the initial working
        set based on the values of the variables and constraints at the
        initial point.
        Broadly speaking, the initial working set will include equality
        constraints and bounds or inequality constraints that violate or
        'nearly' satisfy their bounds (to within 'Crash Tolerance').

        With a 'Warm Start', you must provide a valid definition of
        every element of the array `istate`. ``lp_solve`` will override
        your specification of `istate` if necessary, so that a poor
        choice of the working set will not cause a fatal error.
        For instance, any elements of `istate` which are set to -2, -1
        or 4 will be reset to zero, as will any elements which are set
        to 3 when the corresponding elements of `bl` and `bu` are not
        equal.
        A warm start will be advantageous if a good estimate of the
        initial working set is available -- for example, when
        ``lp_solve`` is called repeatedly to solve related problems.

    'Crash Tolerance' : float
        Default = 0.01

        This value is used in conjunction with the option 'Cold Start'
        (the default value) when ``lp_solve`` selects an initial working
        set.
        If 0 <= r <= 1, the initial working set will include (if
        possible) bounds or general inequality constraints that lie
        within r of their bounds.
        In particular, a constraint of the form a_j^Tx >= l will be
        included in the initial working set if |a_j^Tx-l| <= r(1+|l|).
        If r < 0 or r > 1, the default value is used.

    'Defaults' : valueless
        This special keyword may be used to reset all options to their
        default values.

    'Expand Frequency' : int
        Default = 5

        This option is part of an anti-cycling procedure designed to
        guarantee progress even on highly degenerate problems.

        The strategy is to force a positive step at every iteration, at
        the expense of violating the constraints by a small amount.
        Suppose that the value of the option 'Feasibility Tolerance' is
        delta.
        Over a period of i iterations, the feasibility tolerance
        actually used by ``lp_solve`` (i.e., the `working` feasibility
        tolerance) increases from 0.5 delta to delta (in steps of 0.5
        delta/i).

        At certain stages the following 'resetting procedure' is used to
        remove constraint infeasibilities.
        First, all variables whose upper or lower bounds are in the
        working set are moved exactly onto their bounds.
        A count is kept of the number of nontrivial adjustments made.
        If the count is positive, iterative refinement is used to give
        variables that satisfy the working set to (essentially) machine
        precision.
        Finally, the working feasibility tolerance is reinitialized to
        0.5 delta.

        If a problem requires more than i iterations, the resetting
        procedure is invoked and a new cycle of i iterations is started
        with i incremented by 10. (The decision to resume the
        feasibility phase or optimality phase is based on comparing any
        constraint infeasibilities with delta.)

        The resetting procedure is also invoked when ``lp_solve``
        reaches an apparently optimal, infeasible or unbounded solution,
        unless this situation has already occurred twice.
        If any nontrivial adjustments are made, iterations are
        continued.

        If i <= 0, the default value is used.
        If i >= 9999999, no anti-cycling procedure is invoked.

    'Feasibility Tolerance' : float
        Default = sqrt(epsilon)

        If r >= epsilon, r defines the maximum acceptable `absolute`
        violation in each constraint at a 'feasible' point.
        For example, if the variables and the coefficients in the
        general constraints are of order unity, and the latter are
        correct to about 6 decimal digits, it would be appropriate to
        specify r as 10^(-6).
        If 0 <= r < epsilon, the default value is used.

        ``lp_solve`` attempts to find a feasible solution before
        optimizing the objective function.
        If the sum of infeasibilities cannot be reduced to zero, the
        option 'Minimum Sum of Infeasibilities' can be used to find the
        minimum value of the sum.
        Let Sinf be the corresponding sum of infeasibilities.
        If Sinf is quite small, it may be appropriate to raise r by a
        factor of 10 or 100.
        Otherwise, some error in the data should be suspected.

        Note that a 'feasible solution' is a solution that satisfies the
        current constraints to within the tolerance r.

    'Infinite Bound Size' : float
        Default = 10^20

        If r > 0, r defines the 'infinite' bound bigbnd in the
        definition of the problem constraints.
        Any upper bound greater than or equal to bigbnd will be regarded
        as + infinity (and similarly any lower bound less than or equal
        to -bigbnd will be regarded as - infinity).
        If r < 0, the default value is used.

    'Infinite Step Size' : float
        Default = max(bigbnd, 10^20)

        If r > 0, r specifies the magnitude of the change in variables
        that will be considered a step to an unbounded solution. (Note
        that an unbounded solution can occur only when the problem is of
        type LP.) If the change in x during an iteration would exceed
        the value of r, the objective function is considered to be
        unbounded below in the feasible region.
        If r <= 0, the default value is used.

    'Iteration Limit' : int
        Default = max(50, 5*(n+m_L))

        The value of i specifies the maximum number of iterations
        allowed before termination.
        With i = 0 and 'Print Level' > 0, the workspace needed will be
        computed and printed, but no iterations will be performed.
        If i < 0, the default value is used.

    'Iters' : int
        Default = max(50, 5*(n+m_L))

        The value of i specifies the maximum number of iterations
        allowed before termination.
        With i = 0 and 'Print Level' > 0, the workspace needed will be
        computed and printed, but no iterations will be performed.
        If i < 0, the default value is used.

    'Itns' : int
        Default = max(50, 5*(n+m_L))

        The value of i specifies the maximum number of iterations
        allowed before termination.
        With i = 0 and 'Print Level' > 0, the workspace needed will be
        computed and printed, but no iterations will be performed.
        If i < 0, the default value is used.

    'List' : valueless
        Option 'List' enables printing of each option specification as
        it is supplied. 'Nolist' suppresses this printing.

    'Nolist' : valueless
        Default for ``lp_solve`` = 'Nolist'

        Option 'List' enables printing of each option specification as
        it is supplied. 'Nolist' suppresses this printing.

    'Minimum Sum of Infeasibilities' : str
        Default = NO

        If no feasible point exists for the constraints, this option is
        used to control whether or not ``lp_solve`` will calculate a
        point that minimizes the constraint violations.
        If 'Minimum Sum of Infeasibilities' = 'NO', ``lp_solve`` will
        terminate as soon as it is evident that no feasible point exists
        for the constraints.
        The final point will generally not be the point at which the sum
        of infeasibilities is minimized.
        If 'Minimum Sum of Infeasibilities' = 'YES', ``lp_solve`` will
        continue until the sum of infeasibilities is minimized.

    'Monitoring File' : int
        Default = -1

        If i >= 0 and 'Print Level' >= 5, monitoring information
        produced by ``lp_solve`` at every iteration is sent to a file
        with logical unit number i.
        If i < 0 and/or 'Print Level' < 5, no monitoring information is
        produced.

    'Optimality Tolerance' : float
        Default = epsilon^0.8

        If r >= epsilon, r defines the tolerance used to determine if
        the bounds and general constraints have the right 'sign' for the
        solution to be judged to be optimal.

        If 0 <= r < epsilon, the default value is used.

    'Print Level' : int
        The value of i controls the amount of printout produced by
        ``lp_solve``, as indicated below.
        A detailed description of the printed output is given in
        [Further Comments] (summary output at each iteration and the
        final solution) and [Monitoring Information] (monitoring
        information at each iteration).

        The following printout is sent to the file object associated
        with the advisory I/O unit (see
        :class:`~naginterfaces.base.utils.FileObjManager`):

        [table omitted]

        The following printout is sent to the unit number given by the
        option 'Monitoring File':

        [table omitted]

        If 'Print Level' >= 5 and the unit number defined by the option
        'Monitoring File' is the advisory unit number, the summary
        output for each major iteration is suppressed.

    'Problem Type' : str
        Default = LP

        This option specifies the type of objective function to be
        minimized during the optimality phase.
        The following is the optional keyword and the dimensions of the
        array that must be specified in order to define the objective
        function:

        [table omitted]

        For problems of type FP, the objective function is omitted and
        `cvec` is not referenced.
        The following keywords are also acceptable.

        [table omitted]

    Raises
    ------
    NagValueError
        (`errno` 4)
            Too many iterations.

        (`errno` 6)
            On entry, n = *<value>*.

            Constraint: n > 0.

        (`errno` 6)
            On entry, `nclin` = *<value>*.

            Constraint: `nclin` >= 0.

        (`errno` 6)
            On entry, the equal bounds on *<value>* are infinite,
            because `bl`[*<value>*] = beta and `bu`[*<value>*] = beta,
            but abs(beta) >= bigbnd: beta = *<value>* and bigbnd =
            *<value>*.

        (`errno` 6)
            On entry, the bounds on *<value>* are inconsistent:
            `bl`[*<value>*] = *<value>* and `bu`[*<value>*] = *<value>*.

        (`errno` 6)
            On entry with a Warm Start, `istate`[*<value>*] = *<value>*.

        (`errno` 6)
            On entry with a Cold Start, `istate`[*<value>*] = *<value>*.

        (`errno` 6)
            On entry, the equal bounds on variable *<value>* are
            infinite, because `bl`[*<value>*] = beta and `bu`[*<value>*]
            = beta, but abs(beta) >= bigbnd: beta = *<value>* and bigbnd
            = *<value>*.

        (`errno` 6)
            On entry, the equal bounds on linear constraint *<value>*
            are infinite, because `bl`[*<value>*] = beta and
            `bu`[*<value>*] = beta, but abs(beta) >= bigbnd: beta =
            *<value>* and bigbnd = *<value>*.

        (`errno` 6)
            On entry, the equal bounds on nonlinear constraint *<value>*
            are infinite, because `bl`[*<value>*] = beta and
            `bu`[*<value>*] = beta, but abs(beta) >= bigbnd: beta =
            *<value>* and bigbnd = *<value>*.

        (`errno` 6)
            On entry, the bounds on variable *<value>* are inconsistent:
            `bl`[*<value>*] = *<value>* and `bu`[*<value>*] = *<value>*.

        (`errno` 6)
            On entry, the bounds on linear constraint *<value>* are
            inconsistent: `bl`[*<value>*] = *<value>* and
            `bu`[*<value>*] = *<value>*.

        (`errno` 6)
            On entry, the bounds on nonlinear constraint *<value>* are
            inconsistent: `bl`[*<value>*] = *<value>* and
            `bu`[*<value>*] = *<value>*.

        (`errno` 7)
            'Problem Type' not recognized. Problem abandoned.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            Weak *<value>* solution.

        (`errno` 2)
            *<value>* solution is unbounded.

        (`errno` 3)
            No feasible point for the linear constraints.

        (`errno` 3)
            Cannot satisfy the working-set constraints to the accuracy
            requested.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Gill, P E, Hammarling, S, Murray, W, Saunders, M A and Wright, M H,
    1986, `Users' guide for LSSOL (Version 1.0)`, Report SOL 86-1,
    Department of Operations Research, Stanford University

    Gill, P E and Murray, W, 1978, `Numerically stable methods for
    quadratic programming`, Math. Programming (14), 349--372

    Gill, P E, Murray, W, Saunders, M A and Wright, M H, 1984,
    `Procedures for optimization problems with a mixture of bounds and
    general linear constraints`, ACM Trans. Math. Software (10),
    282--298

    Gill, P E, Murray, W, Saunders, M A and Wright, M H, 1989, `A
    practical anti-cycling procedure for linearly constrained
    optimization`, Math. Programming (45), 437--474

    Gill, P E, Murray, W, Saunders, M A and Wright, M H, 1991,
    `Inertia-controlling methods for general quadratic programming`,
    SIAM Rev. (33), 1--36

    Gill, P E, Murray, W and Wright, M H, 1981, `Practical
    Optimization`, Academic Press
    """

    fun_name = 'naginterfaces.library.opt.lp_solve'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _bl = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'bl'},
        bl,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _bu = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'bu'},
        bu,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _cvec = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'cvec'},
        cvec,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _istate = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'istate'},
        istate,
        copy=True,
    )
    _nclin = _b_utils._EngineIntScalarType(
        nclin, {'fun_name': fun_name, 'entity': 'nclin'},
    ).to_ctypes()
    _nclin_py = _b_utils._EngineIntScalarType.to_py(_nclin)
    _ax = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (max(1, _nclin_py),),
    )
    _clamda = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py+_nclin_py,),
    )
    try:
        _return_scalars = _b_opt.lp_solve(
            _n,
            _nclin,
            _a,
            _bl,
            _bu,
            _cvec,
            _istate,
            _x,
            _ax,
            _clamda,
            comm,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _istate_py = _istate
    _x_py = _x
    _itera_py = _return_scalars.itera
    _obj_py = _return_scalars.obj
    _ax_py = _ax
    _clamda_py = _clamda
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['istate', 'x', 'itera', 'obj', 'ax', 'clamda'],
    )(
        _istate_py,
        _x_py,
        _itera_py,
        _obj_py,
        _ax_py,
        _clamda_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def lp_option_file(ioptns, comm, io_manager=None):
    u"""
    Supply option values for ``lp_solve`` from external file.

    ``lp_option_file`` may be used to supply options to ``lp_solve``
    from an external file.

    For full information please refer to the NAG Library document for
    e04mg

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04mgf.html

    Parameters
    ----------
    ioptns : int
        The unit number (see
        :meth:`~naginterfaces.base.utils.FileObjManager.unit_from_fileobj`)
        of the options file to be read.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``nlp1_init``.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Raises
    ------
    NagValueError
        (`errno` 1)
            `ioptns` is not in the range [0,99].

        (`errno` 1)
            On entry, `ioptns` = *<value>*.

            Constraint: 0 <= `ioptns` <= 99.

        (`errno` 2)
            ``Begin`` was found, but end-of-file was found before
            ``End`` was found.

        (`errno` 3)
            End-of-file was found before ``Begin`` was found.

        (`errno` 5)
            One or more lines of the options file is invalid.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    _b_opt.lp_option_file(
        ioptns,
        comm,
        io_manager,
    )

def lp_option_string(optstr, comm, io_manager=None):
    u"""
    Supply option values to ``lp_solve`` from a character string.

    ``lp_option_string`` may be used to supply individual options to
    ``lp_solve``.

    For full information please refer to the NAG Library document for
    e04mh

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04mhf.html

    Parameters
    ----------
    optstr : str
        A single valid option string (as described in [Description] and
        in [Other Parameters for lp_solve]).

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``nlp1_init``.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Raises
    ------
    NagValueError
        (`errno` 5)
            The supplied option string is invalid. Supplied value was:
            *<value>*.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    _b_opt.lp_option_string(
        optstr,
        comm,
        io_manager,
    )

def handle_solve_lp_ipm(handle, x=None, u=None, monit=None, data=None, io_manager=None):
    u"""
    Linear programming (LP), sparse, interior point method (IPM).

    Note: this function uses optional algorithmic parameters, see also:
    ``handle_opt_set``, ``handle_opt_get``.

    ``handle_solve_lp_ipm`` is a solver from the NAG optimization
    modelling suite for large-scale linear programming (LP) problems
    based on an interior point method (IPM).

    For full information please refer to the NAG Library document for
    e04mt

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04mtf.html

    Parameters
    ----------
    handle : Handle
        The handle to the problem. It needs to be initialized by
        ``handle_init`` and the problem formulated by some of the
        functions ``handle_set_linobj``, ``handle_set_quadobj``,
        ``handle_set_simplebounds`` and ``handle_set_linconstr``. It
        **must not** be changed between calls to the NAG optimization
        modelling suite.

    x : None or float, array-like, shape (nvar), optional
        The input of `x` is reserved for future releases of the NAG
        Library and it is ignored at the moment.

    u : None or float, array-like, shape (nnzu), optional
        Note: if nnzu > 0, `u` holds Lagrange multipliers (dual
        variables) for the bound constraints and linear constraints. If
        nnzu = 0, `u` will not be referenced.

        The input of `u` is reserved for future releases of the NAG
        Library and it is ignored at the moment.

    monit : None or callable monit(handle, rinfo, stats, data=None),
        optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `monit` is provided to enable you to monitor the progress of the
        optimization.

        It is invoked at the end of every ith iteration where i is given
        by the option 'LPIPM Monitor Frequency' (the default is 0,
        `monit` is not called).

        Parameters
        ~~~~~~~~~~
        handle : Handle
            The handle to the problem as provided on entry to
            ``handle_solve_lp_ipm``. It may be used to query the model
            during the solve, and extract the current approximation of
            the solution by ``handle_set_get_real``.

        rinfo : float, ndarray, shape (100)
            Error measures and various indicators at the end of the
            current iteration as described in `rinfo`.

        stats : float, ndarray, shape (100)
            Solver statistics at the end of the current iteration as
            described in `stats`, however, elements 2, 3, 5, 9, 10, 11
            and 15 refer to the quantities in the last iteration rather
            than accumulated over all iterations through the whole
            algorithm run.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

    data : arbitrary, optional
        User-communication data for callback functions.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    x : float, ndarray, shape (nvar)
        The final values of the variables x.

    u : float, ndarray, shape (nnzu)
        The final values of the variables u.

    rinfo : float, ndarray, shape (100)
        Error measures and various indicators of the algorithm (see
        [Algorithmic Details] for details) as given in the table below:

        [table omitted]

    stats : float, ndarray, shape (100)
        Solver statistics as given in the table below. Note that time
        statistics are provided only if 'Stats Time' is set (the default
        is 'NO'), the measured time is returned in seconds.

        [table omitted]

    Other Parameters
    ----------------
    'Defaults' : valueless
        This special keyword may be used to reset all options to their
        default values.
        Any value given with this keyword will be ignored.

    'Infinite Bound Size' : float
        Default = 10^20

        This defines the 'infinite' bound bigbnd in the definition of
        the problem constraints.
        Any upper bound greater than or equal to bigbnd will be regarded
        as + infinity (and similarly any lower bound less than or equal
        to -bigbnd will be regarded as - infinity).
        Note that a modification of this option does not influence
        constraints which have already been defined; only the
        constraints formulated after the change will be affected.

        Constraint: 'Infinite Bound Size' >= 1000.

    'LP Presolve' : str
        Default = 'FULL'

        This argument allows you to reduce the level of presolving of
        the problem or turn it off completely.
        If the presolver is turned off, the solver will try to handle
        the problem you have given.
        In such a case, the presence of fixed variables or linear
        dependencies in the constraint matrix can cause numerical
        instabilities to occur.
        In normal circumstances, it is recommended to use the full
        presolve which is the default.

        Constraint: 'LP Presolve' = 'FULL', 'BASIC' or 'NO'.

    'LPIPM Algorithm' : str
        Default = 'PRIMAL-DUAL'

        As described in [Algorithmic Details], ``handle_solve_lp_ipm``
        implements the infeasible Primal-Dual algorithm, see [The
        Infeasible-interior-point Primal-Dual Algorithm], and the
        homogeneous Self-Dual algorithm, see [Homogeneous Self-Dual
        Algorithm].
        This argument controls which one to use.

        Constraint: 'LPIPM Algorithm' = 'PRIMAL-DUAL', 'PD', 'SELF-DUAL'
        or 'SD'.

    'LPIPM Centrality Correctors' : int
        Default = 6

        This argument controls the number of centrality correctors (see
        [Weighted Multiple Centrality Correctors]) used at each
        iteration.
        Each corrector step attempts to improve the current iterate for
        the price of additional solve(s) of the factorized system matrix
        in order to reduce the total number of iterations.
        Therefore, it trades the additional solves of the system with
        the number of factorizations.
        The more expensive the factorization is with respect to the
        solve, the more corrector steps should be allowed.

        If i > 0, the maximum number of corrector steps will be computed
        by timing heuristics (the ratio between the times of the
        factorization and the solve in the first iteration) but will not
        be greater than i.
        The number computed by the heuristic can be recovered after the
        solve or during a monitoring step in `stats`.
        This might cause non-repeatable results.

        If i < 0, the maximum number of corrector steps will be set to
        |i|.

        If it is set to 0, no additional centrality correctors will be
        used and the algorithm reverts to Mehrotra's
        predictor-corrector.

    'LPIPM Iteration Limit' : int
        Default = 100

        The maximum number of iterations to be performed by
        ``handle_solve_lp_ipm``.
        Setting the option too low might lead to `errno` = 22.

        Constraint: 'LPIPM Iteration Limit' >= 1.

    'LPIPM Scaling' : str
        Default = 'ARITHMETIC'

        This argument controls the type of scaling to be applied on the
        constraint matrix A before solving the problem.
        More precisely, the scaling procedure will try to find diagonal
        matrices D_1 and D_2 such that the values in D_1 AD_2 are of a
        similar order of magnitude.
        The solver is less likely to run into numerical difficulties
        when the constraint matrix is well scaled.

        Constraint: 'LPIPM Scaling' = 'ARITHMETIC', 'GEOMETRIC' or
        'NONE'.

    'LPIPM Monitor Frequency' : int
        Default = 0

        This argument defines the frequency of how often function
        `monit` is called.
        If i > 0, the solver calls `monit` at the end of every ith
        iteration.
        If it is set to 0, the function is not called at all.

        Constraint: 'LPIPM Monitor Frequency' >= 0.

    'LPIPM Stop Tolerance' : float
        Default = sqrt(epsilon)

        This argument sets the value epsilon_1 which is the tolerance
        for the convergence measures in the stopping criteria, see
        [Stopping Criteria].

        Constraint: 'LPIPM Stop Tolerance' > epsilon.

    'LPIPM Stop Tolerance 2' : float
        Default = epsilon^0.6

        This argument sets the additional tolerance epsilon_2 used in
        the stopping criteria for the Self-Dual algorithm, see [Stopping
        Criteria].

        Constraint: 'LPIPM Stop Tolerance 2' > epsilon.

    'LPIPM System Formulation' : str
        Default = 'AUTO'

        As described in [Solving the KKT System],
        ``handle_solve_lp_ipm`` can internally work either with the
        normal equations formulation [equation] or with the augmented
        system [equation].
        A brief discussion of advantages and disadvantages is presented
        in [Solving the KKT System].
        Option 'AUTO' leaves the decision to the solver based on the
        structure of the constraints and it is the recommended option.
        This will typically lead to the normal equations formulation
        unless there are many dense columns or the system is
        significantly cheaper to factorize as the augmented system.
        Note that in some cases even if 'LPIPM System Formulation' =
        'NORMAL EQUATIONS' the solver might switch the formulation
        through the computation to the augmented system due to numerical
        instabilities or computational cost.

        Constraint: 'LPIPM System Formulation' = 'AUTO', 'AUGMENTED
        SYSTEM', 'AS', 'NORMAL EQUATIONS' or 'NE'.

    'Monitoring File' : int
        Default = -1

        If i >= 0, the unit number for the secondary (monitoring)
        output.
        If set to -1, no secondary output is provided.
        The following information is output to the unit:

        -   a listing of the options if set by 'Print Options';

        -   problem statistics, the iteration log and the final status
            as set by 'Monitoring Level';

        -   the solution if set by 'Print Solution'.

        Constraint: 'Monitoring File' >= -1.

    'Monitoring Level' : int
        Default = 4

        This argument sets the amount of information detail that will be
        printed by the solver to the secondary output.
        The meaning of the levels is the same as with 'Print Level'.

        Constraint: 0 <= 'Monitoring Level' <= 5.

    'Print File' : int
        Default = advisory message unit number

        If i >= 0, the unit number for the primary output of the solver.
        If 'Print File' = -1, the primary output is completely turned
        off independently of other settings. The default value is the
        advisory message unit number at the time of the options
        initialization, e.g., at the initialization of the handle. The
        following information is output to the unit:

        -   a listing of options if set by 'Print Options';

        -   problem statistics, the iteration log and the final status
            from the solver as set by 'Print Level';

        -   the solution if set by 'Print Solution'.

        Constraint: 'Print File' >= -1.

    'Print Level' : int
        Default = 2

        This argument defines how detailed information should be printed
        by the solver to the primary output.

        [table omitted]

        Constraint: 0 <= 'Print Level' <= 5.

    'Print Options' : str
        Default = 'YES'

        If 'Print Options' = 'YES', a listing of options will be printed
        to the primary and secondary output.

        Constraint: 'Print Options' = 'YES' or 'NO'.

    'Print Solution' : str
        Default = 'NO'

        If 'Print Solution' = 'X', the final values of the primal
        variables are printed on the primary and secondary outputs.

        If 'Print Solution' = 'YES' or 'ALL', in addition to the primal
        variables, the final values of the dual variables are printed on
        the primary and secondary outputs.

        Constraint: 'Print Solution' = 'YES', 'NO', 'X' or 'ALL'.

    'Stats Time' : str
        Default = 'NO'

        This argument allows you to turn on timings of various parts of
        the algorithm to give a better overview of where most of the
        time is spent.
        This might be helpful for a choice of different solving
        approaches.
        It is possible to choose between CPU and wall clock time.
        Choice 'YES' is equivalent to 'WALL CLOCK'.

        Constraint: 'Stats Time' = 'YES', 'NO', 'CPU' or 'WALL CLOCK'.

    'Task' : str
        Default = 'MINIMIZE'

        This argument specifies the required direction of the
        optimization.
        If 'Task' = 'FEASIBLE POINT', the objective function (if set) is
        ignored and the algorithm stops as soon as a feasible point is
        found with respect to the given tolerance.
        If no objective function is set, 'Task' reverts to 'FEASIBLE
        POINT' automatically.

        Constraint: 'Task' = 'MINIMIZE', 'MAXIMIZE' or 'FEASIBLE POINT'.

    Raises
    ------
    NagValueError
        (`errno` 1)
            `handle` has not been initialized.

        (`errno` 1)
            `handle` does not belong to the NAG optimization modelling
            suite, has not been initialized properly or is corrupted.

        (`errno` 1)
            `handle` has not been initialized properly or is corrupted.

        (`errno` 2)
            This solver does not support the model defined in the
            handle.

        (`errno` 2)
            The problem is already being solved.

        (`errno` 4)
            On entry, nvar = *<value>*, expected value = *<value>*.

            Constraint: nvar must match the value given during
            initialization of `handle`.

        (`errno` 5)
            On entry, nnzu = *<value>*.

            nnzu does not match the size of the Lagrangian multipliers
            for constraints.

            The correct value is either 0 or *<value>*.

        (`errno` 5)
            On entry, nnzu = *<value>*.

            nnzu does not match the size of the Lagrangian multipliers
            for constraints.

            The correct value is 0 for no constraints.

        (`errno` 51)
            The problem was found to be primal infeasible.

        (`errno` 52)
            The problem was found to be dual infeasible.

        (`errno` 53)
            The problem seems to be primal or dual infeasible, the
            algorithm was stopped.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 50)
            Suboptimal solution.

    NagAlgorithmicMajorWarning
        (`errno` 22)
            Maximum number of iterations exceeded.

        (`errno` 24)
            No progress, stopping early.

    NagCallbackTerminateWarning
        (`errno` 20)
            User requested termination during a monitoring step.

    References
    ----------
    Andersen, E D, Gondzio, J, Mészáros, C and Xu, X, 1996,
    `Implementation of interior point methods for large scale linear
    programming`, HEC/Université de Genève

    Colombo, M and Gondzio, J, 2008, `Further development of multiple
    centrality correctors for interior point methods`, Computational
    Optimization and Algorithms (41(3)), 277--305

    Goldfard, D and Scheinberg, K, 2004, `A product-form Cholesky
    factorization method for handling dense columns in interior point
    methods for linear programming`, Mathematical Programming (99(1)),
    1--34

    Gondzio, J, 1996, `Multiple centrality corrections in a primal-dual
    method for linear programming`, Computational Optimization and
    Algorithms (6(2)), 137--156

    Gondzio, J, 2012, `Interior point methods 25 years later`, European
    Journal of Operations Research (218(3)), 587--601

    Hogg, J D and Scott, J A, 2011, `HSL MA97: a bit-compatible
    multifrontal code for sparse symmetric systems`, RAL Technical
    Report. RAL-TR-2011-024

    HSL, 2011, `A collection of Fortran codes for large-scale scientific
    computation`, http://www.hsl.rl.ac.uk/

    Karypis, G and Kumar, V, 1998, `A fast and high quality multilevel
    scheme for partitioning irregular graphs`, SIAM J. Sci. Comput.
    (20(1)), 359--392

    Mészáros, C, 1996, `The efficient implementation of interior point
    methods for linear programming and their applications`, PhD Thesis,
    Eötvös Loránd University of Science, Budapest

    Nocedal, J and Wright, S J, 2006, `Numerical Optimization`, (2nd
    Edition), Springer Series in Operations Research, Springer, New York

    Wright, S W, 1997, `Primal-dual interior point methods`, SIAM,
    Philadelphia

    Xu, X, Hung, P-F and Ye, Y, 1996, `A simplified homogeneous and
    self-dual linear programming algorithm and its implementation`,
    Annals of Operations Research (62(1)), 151--171

    See Also
    --------
    :meth:`naginterfaces.library.examples.opt.handle_solve_lp_ipm_ex.main`
    """

    def cb_monit(handle, rinfo, stats, data=None):
        """Wrapping user's monit. Called as base monit."""
        if data is not None:
            monit(
                handle,
                rinfo,
                stats,
                data,
            )
        else:
            monit(
                handle,
                rinfo,
                stats,
            )

    _nvar, _nnzu, _nnzuc, _nnzua = _b_utils._e04ptan(handle)
    _nvar_py = _b_utils._EngineIntScalarType.to_py(_nvar)
    _vl_e04pta_nvar_py = _nvar_py
    fun_name = 'naginterfaces.library.opt.handle_solve_lp_ipm'

    _nvar_py = _b_utils._EngineIntScalarType.to_py(_nvar)
    if x is None:
        _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
            (_nvar_py,),
        )
    else:
        _x = _utils._to_ndarray(
            {'fun_name': fun_name, 'entity': 'x'},
            x,
            copy=True,
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
    _nnzu_py = _b_utils._EngineIntScalarType.to_py(_nnzu)
    _vl_e04pta_nnzu_py = _nnzu_py
    _nnzu_py = _b_utils._EngineIntScalarType.to_py(_nnzu)
    if u is None:
        _u = _b_utils._EngineFloat64ArrayType.empty_ndarray(
            (_nnzu_py,),
        )
    else:
        _u = _utils._to_ndarray(
            {'fun_name': fun_name, 'entity': 'u'},
            u,
            copy=True,
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
    _rinfo = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (100,),
    )
    _stats = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (100,),
    )
    _b_utils._check_callable(monit, {'fun_name': fun_name, 'entity': 'monit'}, allow_none=True)
    _monit = None if monit is None else cb_monit
    try:
        _b_opt.handle_solve_lp_ipm(
            handle,
            _nvar,
            _x,
            _nnzu,
            _u,
            _rinfo,
            _stats,
            _monit,
            data,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _x_py = _x
    _u_py = _u
    _rinfo_py = _rinfo
    _stats_py = _stats
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'u', 'rinfo', 'stats'],
    )(
        _x_py,
        _u_py,
        _rinfo_py,
        _stats_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def miqp_mps_write(outfile, n, m, ncolh, iobj, bl, bu, pnames, crname, minmax, idxc=None, c=None, a=None, irowa=None, iccola=None, h=None, irowh=None, iccolh=None, intvar=None, io_manager=None):
    u"""
    Write MPS data file defining LP, QP, MILP or MIQP problem.

    ``miqp_mps_write`` writes data for sparse linear programming, mixed
    integer linear programming, quadratic programming or mixed integer
    quadratic programming problems to a file in MPS format.

    For full information please refer to the NAG Library document for
    e04mw

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04mwf.html

    Parameters
    ----------
    outfile : int
        The ID of the file to store the problem data as associated by a
        call to the ``FileObjManager`` method
        :meth:`~naginterfaces.base.utils.FileObjManager.unit_from_fileobj`.

    n : int
        n, the number of variables in the problem.

    m : int
        m, the number of constraints in the problem. This is the number
        of rows in the linear constraint matrix A, including the free
        row (if any; see `iobj`).

    ncolh : int
        The number of leading nonzero columns of the Hessian matrix H.

        If `ncolh` = 0, the quadratic term H of the objective function
        is considered zero (e.g., LP problems), and arrays `h`, `irowh`
        and `iccolh` will not be referenced.

    iobj : int
        If `iobj` > 0, row `iobj` of A is a free row containing the
        nonzero coefficients of the linear terms of the objective
        function. In this case nnzc is set to 0.

        If `iobj` = 0, there is no free row in A, and the linear terms
        might be supplied in array `c`.

    bl : float, array-like, shape (`n`+`m`)
        `bl` and `bu` contains the lower bounds l and the upper bounds
        u, respectively.

        The first `n` elements refer to the bounds for the variables x
        and the rest to the bounds for the linear constraints (including
        the objective row `iobj` if present).

        To specify a nonexistent lower bound (i.e., l_j = -inf), set
        `bl`[j-1] <= -10^20; to specify a nonexistent upper bound, set
        `bu`[j-1] >= 10^20.

    bu : float, array-like, shape (`n`+`m`)
        `bl` and `bu` contains the lower bounds l and the upper bounds
        u, respectively.

        The first `n` elements refer to the bounds for the variables x
        and the rest to the bounds for the linear constraints (including
        the objective row `iobj` if present).

        To specify a nonexistent lower bound (i.e., l_j = -inf), set
        `bl`[j-1] <= -10^20; to specify a nonexistent upper bound, set
        `bu`[j-1] >= 10^20.

    pnames : str, length 8, array-like, shape (5)
        A set of names associated with the MPSX form of the problem.

        The names can be composed only from 'printable' characters
        (ASCII codes between 32 and 126).

        If any of the names are blank, the default name is used.

        `pnames`[0]

            Contains the name of the problem.

        `pnames`[1]

            Contains the name of the objective row if the objective is
            provided in `c` instead of `iobj` and all names `crname` are
            given. The name must be nonempty and unique. In all other
            cases `pnames`[1] is not used.

        `pnames`[2]

            Contains the name of the RHS set.

        `pnames`[3]

            Contains the name of the RANGE.

        `pnames`[4]

            Contains the name of the BOUNDS.

    crname : str, length 8, array-like, shape (nname)
        The names of all the variables and constraints in the problem in
        that order.

        The names can be composed only from 'printable' characters and
        must be unique.

    minmax : int
        `minmax` defines the direction of optimization problem.

        `minmax` = -1

            Minimization.

        `minmax` = 1

            Maximization.

    idxc : None or int, array-like, shape (nnzc), optional
        The nonzero elements of sparse vector c. `idxc`[i-1] must
        contain the index of `c`[i-1] in the vector, for i =
        1,2,...,nnzc.

        The elements are stored in ascending order.

    c : None or float, array-like, shape (nnzc), optional
        The nonzero elements of sparse vector c. `idxc`[i-1] must
        contain the index of `c`[i-1] in the vector, for i =
        1,2,...,nnzc.

        The elements are stored in ascending order.

    a : None or float, array-like, shape (nnza), optional
        The nonzero elements of matrix A in compressed column storage
        (see [the F11 Introduction]). Arrays `irowa` and `a` store the
        row indices and the values of the nonzero elements,
        respectively. The elements are sorted by columns and within each
        column in nondecreasing order. Duplicate entries are not
        allowed. `iccola` contains the (one-based) indices to the
        beginning of each column in `a` and `irowa`.

        If nnza = 0, `a` and `irowa` are not referenced.

    irowa : None or int, array-like, shape (nnza), optional
        The nonzero elements of matrix A in compressed column storage
        (see [the F11 Introduction]). Arrays `irowa` and `a` store the
        row indices and the values of the nonzero elements,
        respectively. The elements are sorted by columns and within each
        column in nondecreasing order. Duplicate entries are not
        allowed. `iccola` contains the (one-based) indices to the
        beginning of each column in `a` and `irowa`.

        If nnza = 0, `a` and `irowa` are not referenced.

    iccola : None or int, array-like, shape (:), optional
        Note: the required length for this argument is determined as
        follows: if nnza>0: `n`+1; otherwise: 0.

        The nonzero elements of matrix A in compressed column storage
        (see [the F11 Introduction]). Arrays `irowa` and `a` store the
        row indices and the values of the nonzero elements,
        respectively. The elements are sorted by columns and within each
        column in nondecreasing order. Duplicate entries are not
        allowed. `iccola` contains the (one-based) indices to the
        beginning of each column in `a` and `irowa`.

        If nnza = 0, `a` and `irowa` are not referenced.

    h : None or float, array-like, shape (nnzh), optional
        The nonzero elements of the Hessian matrix H in compressed
        column storage (see [the F11 Introduction]). The Hessian matrix,
        H, is symmetric and its elements are stored in a lower
        triangular matrix.

        Arrays `irowh` and `h` store the row indices and the values of
        the nonzero elements, respectively.

        The elements are sorted by columns and within each column in
        nondecreasing order.

        Duplicate entries are not allowed. `iccolh` contains the
        (one-based) indices to the beginning of each column in `h` and
        `irowh`.

        If `ncolh` = 0, `h` is not referenced.

    irowh : None or int, array-like, shape (nnzh), optional
        The nonzero elements of the Hessian matrix H in compressed
        column storage (see [the F11 Introduction]). The Hessian matrix,
        H, is symmetric and its elements are stored in a lower
        triangular matrix.

        Arrays `irowh` and `h` store the row indices and the values of
        the nonzero elements, respectively.

        The elements are sorted by columns and within each column in
        nondecreasing order.

        Duplicate entries are not allowed. `iccolh` contains the
        (one-based) indices to the beginning of each column in `h` and
        `irowh`.

        If `ncolh` = 0, `irowh` is not referenced.

    iccolh : None or int, array-like, shape (:), optional
        Note: the required length for this argument is determined as
        follows: if `ncolh`>0: `ncolh`+1; otherwise: 0.

        The nonzero elements of the Hessian matrix H in compressed
        column storage (see [the F11 Introduction]). The Hessian matrix,
        H, is symmetric and its elements are stored in a lower
        triangular matrix.

        Arrays `irowh` and `h` store the row indices and the values of
        the nonzero elements, respectively.

        The elements are sorted by columns and within each column in
        nondecreasing order.

        Duplicate entries are not allowed. `iccolh` contains the
        (one-based) indices to the beginning of each column in `h` and
        `irowh`.

        If `ncolh` = 0, `iccolh` is not referenced.

    intvar : None or int, array-like, shape (lintvar), optional
        `intvar` contains the indices k of variables x_k which are
        defined as integers. Duplicate indices are not allowed.

        If lintvar = 0, `intvar` is not referenced.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `outfile` = *<value>*.

            Constraint: `outfile` >= 0.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 2)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 0.

        (`errno` 3)
            On entry, nnzc = *<value>*.

            Constraint: nnzc >= 0.

        (`errno` 3)
            On entry, nnza = *<value>*.

            Constraint: nnza >= 0.

        (`errno` 3)
            On entry, lintvar = *<value>*.

            Constraint: lintvar >= 0.

        (`errno` 3)
            On entry, nname = *<value>*, `n` = *<value>* and `m` =
            *<value>*.

            Constraint: nname = 0 or `n`+`m`.

        (`errno` 4)
            On entry, `ncolh` = *<value>* and `n` = *<value>*.

            Constraint: 0 <= `ncolh` <= `n`.

        (`errno` 4)
            On entry, `ncolh` = *<value>* and nnzh = *<value>*.

            Constraint: if `ncolh` > 0, nnzh > 0.

        (`errno` 4)
            On entry, `ncolh` = *<value>* and nnzh = *<value>*.

            Constraint: if `ncolh` = 0, nnzh = 0.

        (`errno` 5)
            On entry, j = *<value>*, `idxc`[j-1] = *<value>* and `n` =
            *<value>*.

            Constraint: 1 <= `idxc`[j-1] <= `n`.

        (`errno` 5)
            On entry, j = *<value>*, `idxc`[j-1] = *<value>* and
            `idxc`[j] = *<value>*.

            Constraint: `idxc`[j-1] < `idxc`[j].

        (`errno` 6)
            On entry, `minmax` = *<value>*.

            Constraint: `minmax` = -1 or 1.

        (`errno` 7)
            On entry, `iobj` = *<value>* and `m` = *<value>*.

            Constraint: 0 <= `iobj` <= `m`.

        (`errno` 7)
            On entry, `iobj` = *<value>* and nnzc = *<value>*.

            Constraint: at most one of `iobj` or nnzc may be nonzero.

        (`errno` 8)
            On entry, j = *<value>*, `bl`[j-1] = *<value>* and `bu`[j-1]
            = *<value>* are incorrect.

            Constraint: `bl`[j-1] <= `bu`[j-1].

        (`errno` 8)
            On entry, j = *<value>* and `bl`[j-1] = *<value>*, `bl`[j-1]
            is incorrect.

            Constraint: `bl`[j-1] < 1e+20.

        (`errno` 8)
            On entry, j = *<value>* and `bu`[j-1] = *<value>*, `bu`[j-1]
            is incorrect.

            Constraint: `bu`[j-1] > -1e+20.

        (`errno` 8)
            On entry, `iobj` = *<value>*, `bl`[j-1] = *<value>* and
            `bu`[j-1] = *<value>*, if `iobj` > 0 the bounds must be
            infinite.

            Constraints: `bl`[j-1] <= -1e+20, `bu`[j-1] >= 1e+20.

        (`errno` 8)
            On entry, j = *<value>*, `bl`[j-1] = *<value>* and `bu`[j-1]
            = *<value>*, the integer variable j requires at least one
            bound finite.

            Constraint: at least one of the following conditions must be
            met for integer variable j: `bl`[j-1] > -1e+20, `bu`[j-1] <
            1e+20.

        (`errno` 9)
            On entry, `crname`[j-1] for j = *<value>* is incorrect.

            Constraint: the names in `crname` must consist only of
            printable characters.

        (`errno` 9)
            On entry, `pnames`[j-1] for j = *<value>* is incorrect.

            Constraint: the names in `pnames` must consist only of
            printable characters.

        (`errno` 9)
            On entry, `crname`[j-1] for j = *<value>* has been already
            used.

            Constraint: the names in `crname` must be unique.

        (`errno` 9)
            The name specified in `pnames`[1] is empty or has been
            already used among row names.

            Constraint: the names in `pnames`[1] must be unique and
            nonempty if `crname` is provided and nnzc > 0.

        (`errno` 10)
            On entry, j = *<value>*, `intvar`[j-1] = *<value>* and
            lintvar = *<value>*.

            Constraint: 1 <= `intvar`[j-1] <= lintvar.

        (`errno` 10)
            On entry, `intvar`[*<value>*] = `intvar`[*<value>*] =
            *<value>*.

            Constraint: all entries in `intvar` must be unique.

        (`errno` 11)
            On entry, i = *<value>*, `irowa`[i-1] = *<value>* and `m` =
            *<value>*.

            Constraint: 1 <= `irowa`[i-1] <= `m`.

        (`errno` 11)
            On entry, more than one element of `a` has row index
            *<value>* and column index *<value>*.

            Constraint: each element of `a` must have a unique row and
            column index.

        (`errno` 12)
            On entry, j = *<value>*, `iccola`[j-1] = *<value>* and
            `iccola`[j] = *<value>*, the values of `iccola` must be
            nondecreasing.

            Constraint: `iccola`[j-1] <= `iccola`[j].

        (`errno` 12)
            On entry, `iccola`[0] = *<value>*.

            Constraint: `iccola`[0] = 1.

        (`errno` 12)
            On entry, `iccola`[`n`] = *<value>* and nnza = *<value>*.

            Constraint: `iccola`[`n`] = nnza+1.

        (`errno` 13)
            On entry, j = *<value>*, i = *<value>*, `ncolh` = *<value>*
            and `irowh`[i-1] = *<value>*

            Constraint: j <= `irowh`[i-1] <= `ncolh` (within the lower
            triangle).

        (`errno` 13)
            On entry, more than one element of `h` has row index
            *<value>* and column index *<value>*.

            Constraint: each element of `h` must have a unique row and
            column index.

        (`errno` 14)
            On entry, j = *<value>*, `iccolh`[j-1] = *<value>* and
            `iccolh`[j] = *<value>*, the values of `iccolh` must be
            nondecreasing.

            Constraint: `iccolh`[j-1] <= `iccolh`[j].

        (`errno` 14)
            On entry, `iccolh`[0] = *<value>*.

            Constraint: `iccolh`[0] = 1.

        (`errno` 14)
            On entry, `iccolh`[`ncolh`] = *<value>* and nnzh =
            *<value>*.

            Constraint: `iccolh`[`ncolh`] = nnzh+1.

        (`errno` 15)
            An error occurred when writing to file.

    References
    ----------
    IBM, 1971, `MPSX -- Mathematical programming system`, Program Number
    5734 XM4, IBM Trade Corporation, New York
    """

    fun_name = 'naginterfaces.library.opt.miqp_mps_write'

    _idxc = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'idxc'},
        idxc,
        copy=False,
        allow_none=True,
    )
    _nnzc = _b_utils._EngineIntScalarType(
        0 if idxc is None else _b_utils._EngineIntArrayType.get_shape(_idxc, {'fun_name': fun_name, 'entity': 'idxc'}, exp_rank=1, allow_none=True)[0], {'fun_name': fun_name, 'entity': '_nnzc'},
    ).to_ctypes()
    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nnza = _b_utils._EngineIntScalarType(
        0 if a is None else _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=1, allow_none=True)[0], {'fun_name': fun_name, 'entity': '_nnza'},
    ).to_ctypes()
    _h = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'h'},
        h,
        copy=False,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    if h is None:
        _nnzh = 0
        _nnzh_py = _nnzh
    else:
        _nnzh_py = _b_utils._EngineFloat64ArrayType.get_shape(_h, {'fun_name': fun_name, 'entity': 'h'}, exp_rank=1, allow_none=True)[0]
        _nnzh = _b_utils._EngineIntScalarType(
            _nnzh_py, {'fun_name': fun_name, 'entity': '_nnzh'},
        ).to_ctypes()
    _intvar = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'intvar'},
        intvar,
        copy=False,
        allow_none=True,
    )
    _lintvar = _b_utils._EngineIntScalarType(
        0 if intvar is None else _b_utils._EngineIntArrayType.get_shape(_intvar, {'fun_name': fun_name, 'entity': 'intvar'}, exp_rank=1, allow_none=True)[0], {'fun_name': fun_name, 'entity': '_lintvar'},
    ).to_ctypes()
    _c = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
        copy=False,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _irowa = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'irowa'},
        irowa,
        copy=False,
        allow_none=True,
    )
    _iccola = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'iccola'},
        iccola,
        copy=False,
        allow_none=True,
    )
    _bl = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'bl'},
        bl,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _bu = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'bu'},
        bu,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _pnames = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'pnames'},
        pnames,
        copy=False,
    )
    _crname = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'crname'},
        crname,
        copy=False,
    )
    _nname_py = _b_utils._EngineCharArrayType.get_shape(_crname, {'fun_name': fun_name, 'entity': 'crname'}, exp_rank=1)[0]
    _nname = _b_utils._EngineIntScalarType(
        _nname_py, {'fun_name': fun_name, 'entity': '_nname'},
    ).to_ctypes()
    _irowh = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'irowh'},
        irowh,
        copy=False,
        allow_none=True,
    )
    _iccolh = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'iccolh'},
        iccolh,
        copy=False,
        allow_none=True,
    )
    _b_opt.miqp_mps_write(
        outfile,
        n,
        m,
        _nnzc,
        _nnza,
        ncolh,
        _nnzh,
        _lintvar,
        _idxc,
        _c,
        iobj,
        _a,
        _irowa,
        _iccola,
        _bl,
        _bu,
        _pnames,
        _nname,
        _crname,
        _h,
        _irowh,
        _iccolh,
        minmax,
        _intvar,
        io_manager,
    )

def miqp_mps_read(infile, maxn, maxm, maxnnz, maxncolh, maxnnzh, maxlintvar, pnames, mpslst=0, io_manager=None):
    u"""
    Read MPS data file defining LP, QP, MILP or MIQP problem.

    ``miqp_mps_read`` reads data for sparse linear programming, mixed
    integer linear programming, quadratic programming or mixed integer
    quadratic programming problems from an external file which is in
    standard or compatible MPS input format.

    For full information please refer to the NAG Library document for
    e04mx

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04mxf.html

    Parameters
    ----------
    infile : int
        The ID of the MPSX data file to be read as returned by a call to
        the ``FileObjManager`` method
        :meth:`~naginterfaces.base.utils.FileObjManager.unit_from_fileobj`.

    maxn : int
        An upper limit for the number of variables in the problem.

        If `maxn` < 1, ``miqp_mps_read`` will start in query mode (see
        [Query Mode]).

    maxm : int
        An upper limit for the number of general linear constraints
        (including the objective row) in the problem.

        If `maxm` < 1, ``miqp_mps_read`` will start in query mode (see
        [Query Mode]).

    maxnnz : int
        An upper limit for the number of nonzeros (including the
        objective row) in the problem.

        If `maxnnz` < 1, ``miqp_mps_read`` will start in query mode (see
        [Query Mode]).

    maxncolh : int
        An upper limit for the dimension of the matrix H.

        If `maxncolh` < 0, ``miqp_mps_read`` will start in query mode
        (see [Query Mode]).

    maxnnzh : int
        An upper limit for the number of nonzeros of the matrix H.

        If `maxnnzh` < 0, ``miqp_mps_read`` will start in query mode
        (see [Query Mode]).

    maxlintvar : int
        If `maxlintvar` >= 0, an upper limit for the number of integer
        variables.

        If `maxlintvar` < 0, ``miqp_mps_read`` will treat all integer
        variables in the file as continuous variables.

    pnames : str, length 8, array-like, shape (5)
        A set of names associated with the MPSX form of the problem.

        `pnames`[0]

            Must either contain the name of the problem or be blank.

        `pnames`[1]

            Must either be blank or contain the name of the objective
            row (in which case it overrides OBJNAME section and the
            default choice of the first objective free row).

        `pnames`[2]

            Must either contain the name of the RHS set to be used or be
            blank (in which case the first RHS set is used).

        `pnames`[3]

            Must either contain the name of the RANGE set to be used or
            be blank (in which case the first RANGE set (if any) is
            used).

        `pnames`[4]

            Must either contain the name of the BOUNDS set to be used or
            be blank (in which case the first BOUNDS set (if any) is
            used).

    mpslst : int, optional
        If `mpslst` != 0, summary messages are sent to the file object
        associated with the advisory I/O unit (see
        :class:`~naginterfaces.base.utils.FileObjManager`) as
        ``miqp_mps_read`` reads through the data file. This can be
        useful for debugging the file. If `mpslst` = 0, then no summary
        is produced.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    n : int
        If ``miqp_mps_read`` was run in query mode (see [Query Mode]),
        or returned with `errno` = 2, an upper estimate of the number of
        variables of the problem. Otherwise, n, the actual number of
        variables in the problem.

    m : int
        If ``miqp_mps_read`` was run in query mode (see [Query Mode]),
        or returned with `errno` = 2, an upper estimate of the number of
        general linear constraints in the problem (including the
        objective row). Otherwise m, the actual number of general linear
        constraints of the problem.

    nnz : int
        If ``miqp_mps_read`` was run in query mode (see [Query Mode]),
        or returned with `errno` = 2, an upper estimate of the number of
        nonzeros in the problem (including the objective row). Otherwise
        the actual number of nonzeros in the problem (including the
        objective row).

    ncolh : int
        If ``miqp_mps_read`` was run in query mode (see [Query Mode]),
        or returned with `errno` = 2, an upper estimate of the value of
        ncolh required by ``qpconvex1_sparse_solve`` and
        ``qpconvex2_sparse_solve``. In this context `ncolh` is the
        number of leading nonzero columns of the Hessian matrix H.
        Otherwise, the actual dimension of the matrix H.

    nnzh : int
        If ``miqp_mps_read`` was run in query mode (see [Query Mode]),
        or returned with `errno` = 2, an upper estimate of the number of
        nonzeros of the matrix H. Otherwise, the actual number of
        nonzeros of the matrix H.

    lintvar : int
        If on entry `maxlintvar` < 0, all integer variables are treated
        as continuous and `lintvar` = -1.

        If ``miqp_mps_read`` was run in query mode (see [Query Mode]),
        or returned with `errno` = 2, an upper estimate of the number of
        integer variables of the problem.

        Otherwise, the actual number of integer variables of the
        problem.

    iobj : int
        If `iobj` > 0, row `iobj` of A is a free row containing the
        nonzero coefficients of the vector c.

        If `iobj` = 0, the coefficients of c are assumed to be zero.

        If ``miqp_mps_read`` is run in query mode (see [Query Mode])
        `iobj` is not referenced.

    a : float, ndarray, shape (`maxnnz`)
        The nonzero elements of A, ordered by increasing column index.

        If ``miqp_mps_read`` is run in query mode (see [Query Mode]),
        `a` is not referenced.

    irowa : int, ndarray, shape (`maxnnz`)
        The row indices of the nonzero elements stored in `a`.

        If ``miqp_mps_read`` is run in query mode (see [Query Mode]),
        `irowa` is not referenced.

    iccola : int, ndarray, shape (:)
        A set of pointers to the beginning of each column of A. More
        precisely, `iccola`[i-1] contains the index in `a` of the start
        of the ith column, for i = 1,2,...,`n`. Note that `iccola`[0] =
        1 and `iccola`[`n`] = `nnz`+1.

        If ``miqp_mps_read`` is run in query mode (see [Query Mode]),
        `iccola` is not referenced.

    bl : float, ndarray, shape (`maxn`+`maxm`)
        `bl` contains the vector l (the lower bounds) for all the
        variables and constraints

    bu : float, ndarray, shape (`maxn`+`maxm`)
        `bu` contains the vector u (the upper bounds) for all the
        variables and constraints

    pnames : str, length 8, ndarray, shape (5)
        A set of names associated with the problem as defined in the
        MPSX data file as follows:

        `pnames`[0]

            Contains the name of the problem (or blank if none).

        `pnames`[1]

            Contains the name of the objective row (or blank if none).

        `pnames`[2]

            Contains the name of the RHS set (or blank if none).

        `pnames`[3]

            Contains the name of the RANGE set (or blank if none).

        `pnames`[4]

            Contains the name of the BOUNDS set (or blank if none).

        If ``miqp_mps_read`` is run in query mode (see [Query Mode]),
        `pnames` is not referenced.

    nname : int
        n+m, the total number of variables and constraints in the
        problem (including the objective row).

        If ``miqp_mps_read`` was run in query mode (see [Query Mode]),
        or returned with `errno` = 2, `nname` is not set.

    crname : str, length 8, ndarray, shape (`maxn`+`maxm`)
        The MPS names of all the variables and constraints in the
        problem in the following order. The first `n` elements contain
        the MPS names for the variables and the next `m` elements
        contain the MPS names for the objective row and general linear
        constraints (if any). Note that the MPS name for the objective
        row is stored in `crname`[`n`+`iobj`-1].

        If ``miqp_mps_read`` is run in query mode (see [Query Mode]),
        `crname` is not referenced.

    h : float, ndarray, shape (`maxnnzh`)
        The `nnzh` nonzero elements of H, arranged by increasing column
        index.

        If ``miqp_mps_read`` is run in query mode (see [Query Mode]),
        `h` is not referenced.

    irowh : int, ndarray, shape (`maxnnzh`)
        The `nnzh` row indices of the elements stored in H.

        If ``miqp_mps_read`` is run in query mode (see [Query Mode]),
        `irowh` is not referenced.

    iccolh : int, ndarray, shape (:)
        A set of pointers to the beginning of each column of H. More
        precisely, `iccolh`[i-1] contains the index in H of the start of
        the ith column, for i = 1,2,...,`ncolh`. Note that `iccolh`[0] =
        1 and `iccolh`[`ncolh`] = `nnzh`+1.

        If ``miqp_mps_read`` is run in query mode (see [Query Mode]),
        `iccolh` is not referenced.

    minmax : int
        `minmax` defines the direction of the optimization as read from
        the MPS file. By default the function assumes the objective
        function should be minimized and will return `minmax` = -1. If
        the function discovers in the OBJSENSE section that the
        objective function should be maximized it will return `minmax` =
        1. If the function discovers that there is neither the linear
        objective term c (the objective row) nor the Hessian matrix H,
        the problem is considered as a feasible point problem and
        `minmax` = 0 is returned.

        If ``miqp_mps_read`` was run in query mode (see [Query Mode]),
        or returned with `errno` = 2, `minmax` is not set.

    intvar : int, ndarray, shape (`maxlintvar`)
        If `maxlintvar` > 0 on entry, `intvar` contains pointers to the
        columns that are defined as integer variables. More precisely,
        `intvar`[i-1] = k, where k is the index of a column that is
        defined as an integer variable, for i = 1,2,...,`lintvar`.

        If `maxlintvar` <= 0 on entry, or ``miqp_mps_read`` was run in
        query mode (see [Query Mode]), or it returned with `errno` = 2,
        `intvar` is not set.

    Raises
    ------
    NagValueError
        (`errno` 3)
            Line *<value>*: Incorrect ordering of indicator lines.

            OBJNAME indicator line found after ROWS indicator line.

        (`errno` 4)
            Line *<value>*: Incorrect ordering of indicator lines.

            COLUMNS indicator line found before ROWS indicator line.

        (`errno` 5)
            Line *<value>*: Incorrect ordering of indicator lines.

            RHS indicator line found before COLUMNS indicator line.

        (`errno` 6)
            Line *<value>*: Incorrect ordering of indicator lines.

            RANGES indicator line found before RHS indicator line.

        (`errno` 7)
            Line *<value>*: Incorrect ordering of indicator lines.

            BOUNDS indicator line found before COLUMNS indicator line.

        (`errno` 8)
            Line *<value>*: Incorrect ordering of indicator lines.

            QUADOBJ indicator line found before BOUNDS indicator line.

        (`errno` 9)
            Line *<value>*: Incorrect ordering of indicator lines.

            QUADOBJ indicator line found before COLUMNS indicator line.

        (`errno` 10)
            Line *<value>*: Unknown indicator line *<value>*.

        (`errno` 12)
            Line *<value>*: Indicator line '*<value>*' has been found
            more than once in the MPS file.

        (`errno` 13)
            End of file found before ENDATA indicator line.

        (`errno` 14)
            No indicator line found in file. It may be an empty file.

        (`errno` 15)
            At least one mandatory section not found in MPS file.

        (`errno` 16)
            Line *<value>*: An illegal line was detected in *<value>*
            section.

            This is neither a comment nor a valid data line.

        (`errno` 17)
            Line *<value>*: Unknown inequality key *<value>* in ROWS
            section.

            Expected '``N``', '``G``', '``L``' or '``E``'.

        (`errno` 18)
            Line *<value>*: Empty ROWS section.

            Neither the objective row nor the constraints were defined.

        (`errno` 19)
            The supplied name, in `pnames`[1] or in OBJNAME, of the
            objective row was not found among the free rows in the ROWS
            section.

        (`errno` 20)
            The supplied name, in `pnames`[4], of the BOUNDS set to be
            used was not found in the BOUNDS section.

        (`errno` 21)
            The supplied name, in `pnames`[2], of the RHS set to be used
            was not found in the RHS section.

        (`errno` 22)
            The supplied name, in `pnames`[3], of the RANGES set to be
            used was not found in the RANGES section.

        (`errno` 23)
            Line *<value>*: Illegal row name.

            Row names must consist of printable characters only.

        (`errno` 24)
            Line *<value>*: Illegal column name.

            Column names must consist of printable characters only.

        (`errno` 25)
            Line *<value>*: Row name *<value>* has been defined more
            than once in the ROWS section.

        (`errno` 26)
            Line *<value>*: Column '*<value>*' has been defined more
            than once in the COLUMNS section. Column definitions must be
            continuous. (See [COLUMNS Section]).

        (`errno` 27)
            Line *<value>*: Found ``'INTORG'`` marker within
            ``'INTORG'`` to ``'INTEND'`` range.

        (`errno` 28)
            Line *<value>*: Found ``'INTEND'`` marker without previous
            marker being ``'INTORG'``.

        (`errno` 29)
            Line *<value>*: Found ``'INTORG'`` but not ``'INTEND'``
            before the end of the COLUMNS section.

        (`errno` 30)
            Line *<value>*: Illegal marker type *<value>*.

            Should be either ``'INTORG'`` or ``'INTEND'``.

        (`errno` 31)
            Line *<value>*: Unknown row name *<value>* in *<value>*
            section.

            All row names must be specified in the ROWS section.

        (`errno` 32)
            Line *<value>*: Unknown column name *<value>* in *<value>*
            section.

            All column names must be specified in the COLUMNS section.

        (`errno` 33)
            Inconsistent bounds for row '*<value>*'.

        (`errno` 33)
            Inconsistent bounds for column '*<value>*'.

        (`errno` 33)
            Line *<value>*: Unknown bound type *<value>* in BOUNDS
            section.

        (`errno` 34)
            More than one nonzero of A has row name '*<value>*' and
            column name '*<value>*' in the COLUMNS section.

        (`errno` 35)
            Field *<value>* did not contain a number (see
            [Description]).

        (`errno` 36)
            On entry, `infile` = *<value>*.

            Constraint: `infile` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            Warning: MPS file not strictly fixed format, although the
            problem was read anyway. The data may have been read
            incorrectly. You should set `mpslst` = 1 and repeat the call
            to ``miqp_mps_read`` for more details.

        (`errno` 2)
            At least one of `maxm`, `maxn`, `maxnnz`, `maxnnzh` or
            `maxncolh` is too small. You must provide `maxm` >=
            *<value>*, `maxn` >= *<value>*, `maxnnz` >= *<value>*,
            `maxnnzh` >= *<value>*, `maxncolh` >= *<value>*.

        (`errno` 2)
            At least one of `maxm`, `maxn`, `maxnnz`, `maxnnzh`,
            `maxncolh` or `maxlintvar` is too small. You must provide
            `maxm` >= *<value>*, `maxn` >= *<value>*, `maxnnz` >=
            *<value>*, `maxnnzh` >= *<value>*, `maxncolh` >= *<value>*,
            `maxlintvar` >= *<value>*.

    References
    ----------
    IBM, 1971, `MPSX -- Mathematical programming system`, Program Number
    5734 XM4, IBM Trade Corporation, New York
    """

    fun_name = 'naginterfaces.library.opt.miqp_mps_read'

    _maxnnz = _b_utils._EngineIntScalarType(
        maxnnz, {'fun_name': fun_name, 'entity': 'maxnnz'},
    ).to_ctypes()
    _maxnnz_py = _b_utils._EngineIntScalarType.to_py(_maxnnz)
    _a = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_maxnnz_py,),
    )
    _irowa = _b_utils._EngineIntArrayType.empty_ndarray(
        (_maxnnz_py,),
    )
    _maxn = _b_utils._EngineIntScalarType(
        maxn, {'fun_name': fun_name, 'entity': 'maxn'},
    ).to_ctypes()
    _maxn_py = _b_utils._EngineIntScalarType.to_py(_maxn)
    if _maxn_py > 0:
        _iccola_dim1 = _maxn_py+1
    else:
        _iccola_dim1 = 0
    exp_shape_iccola = (_iccola_dim1,)
    _iccola = _b_utils._EngineIntArrayType.empty_ndarray(
        exp_shape_iccola,
    )
    _maxm = _b_utils._EngineIntScalarType(
        maxm, {'fun_name': fun_name, 'entity': 'maxm'},
    ).to_ctypes()
    _maxm_py = _b_utils._EngineIntScalarType.to_py(_maxm)
    _bl = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_maxn_py+_maxm_py,),
    )
    _bu = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_maxn_py+_maxm_py,),
    )
    _pnames = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'pnames'},
        pnames,
        copy=True,
    )
    _crname = _b_utils._EngineCharArrayType.empty_ndarray(
        (_maxn_py+_maxm_py,),
        chrlen=8,
    )
    _maxnnzh = _b_utils._EngineIntScalarType(
        maxnnzh, {'fun_name': fun_name, 'entity': 'maxnnzh'},
    ).to_ctypes()
    _maxnnzh_py = _b_utils._EngineIntScalarType.to_py(_maxnnzh)
    _h = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_maxnnzh_py,),
    )
    _irowh = _b_utils._EngineIntArrayType.empty_ndarray(
        (_maxnnzh_py,),
    )
    _maxncolh = _b_utils._EngineIntScalarType(
        maxncolh, {'fun_name': fun_name, 'entity': 'maxncolh'},
    ).to_ctypes()
    _maxncolh_py = _b_utils._EngineIntScalarType.to_py(_maxncolh)
    if _maxncolh_py > 0:
        _iccolh_dim1 = _maxncolh_py+1
    else:
        _iccolh_dim1 = 0
    exp_shape_iccolh = (_iccolh_dim1,)
    _iccolh = _b_utils._EngineIntArrayType.empty_ndarray(
        exp_shape_iccolh,
    )
    _maxlintvar = _b_utils._EngineIntScalarType(
        maxlintvar, {'fun_name': fun_name, 'entity': 'maxlintvar'},
    ).to_ctypes()
    _maxlintvar_py = _b_utils._EngineIntScalarType.to_py(_maxlintvar)
    _intvar = _b_utils._EngineIntArrayType.empty_ndarray(
        (_maxlintvar_py,),
    )
    try:
        _return_scalars = _b_opt.miqp_mps_read(
            infile,
            _maxn,
            _maxm,
            _maxnnz,
            _maxncolh,
            _maxnnzh,
            _maxlintvar,
            mpslst,
            _a,
            _irowa,
            _iccola,
            _bl,
            _bu,
            _pnames,
            _crname,
            _h,
            _irowh,
            _iccolh,
            _intvar,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _pnames_py = _pnames
    _n_py = _return_scalars.n
    _m_py = _return_scalars.m
    _nnz_py = _return_scalars.nnz
    _ncolh_py = _return_scalars.ncolh
    _nnzh_py = _return_scalars.nnzh
    _lintvar_py = _return_scalars.lintvar
    _iobj_py = _return_scalars.iobj
    _a_py = _a
    _irowa_py = _irowa
    _iccola_py = _iccola
    _bl_py = _bl
    _bu_py = _bu
    _nname_py = _return_scalars.nname
    _crname_py = _crname
    _h_py = _h
    _irowh_py = _irowh
    _iccolh_py = _iccolh
    _minmax_py = _return_scalars.minmax
    _intvar_py = _intvar
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['n', 'm', 'nnz', 'ncolh', 'nnzh', 'lintvar', 'iobj', 'a', 'irowa', 'iccola', 'bl', 'bu', 'pnames', 'nname', 'crname', 'h', 'irowh', 'iccolh', 'minmax', 'intvar'],
    )(
        _n_py,
        _m_py,
        _nnz_py,
        _ncolh_py,
        _nnzh_py,
        _lintvar_py,
        _iobj_py,
        _a_py,
        _irowa_py,
        _iccola_py,
        _bl_py,
        _bu_py,
        _pnames_py,
        _nname_py,
        _crname_py,
        _h_py,
        _irowh_py,
        _iccolh_py,
        _minmax_py,
        _intvar_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

@_b_utils._withdrawal(
    'naginterfaces.library.opt.qpconvex1_sparse_mps',
    mkwdrn=None,
    advice=[
        'Please use miqp_mps_read instead.',
        'See also https://www.nag.com/numeric/py/nagdoc_latest/replace.html',
    ]
)
def qpconvex1_sparse_mps(infile, maxn, maxm, maxnnz, xbldef, xbudef, mpslst, names, io_manager=None):
    u"""
    Read MPS data file defining LP or QP problem, deprecated.

    ``qpconvex1_sparse_mps`` reads data for a sparse linear programming
    or quadratic programming problem from an external file which is in
    standard or compatible MPSX input format.

    Note: ``qpconvex1_sparse_mps`` is deprecated.
    Please use ``miqp_mps_read`` instead.
    See also https://www.nag.com/numeric/py/nagdoc_latest/replace.html

    For full information please refer to the NAG Library document for
    e04mz

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04mzf.html

    Parameters
    ----------
    infile : int
        The unit number (see
        :meth:`~naginterfaces.base.utils.FileObjManager.unit_from_fileobj`)
        associated with the MPSX data file.

    maxn : int
        An upper limit for the number of variables in the problem.

    maxm : int
        An upper limit for the number of constraints (including the
        objective row) in the problem.

    maxnnz : int
        An upper limit for the number of nonzeros (including the
        objective row) in the problem.

    xbldef : float
        The default lower bound to be used for the variables in the
        problem when none is specified in the BOUNDS section of the MPSX
        data file. For a standard LP or QP problem `xbldef` would
        normally be set to zero.

    xbudef : float
        The default upper bound to be used for the variables in the
        problem when none is specified in the BOUNDS section of the MPSX
        data file. For a standard LP or QP problem `xbudef` would
        normally be set to 'infinity' (i.e., `xbudef` >= 10^20).

    mpslst : bool
        If `mpslst` = True, a listing of the input data is sent to the
        file object associated with the advisory I/O unit (see
        :class:`~naginterfaces.base.utils.FileObjManager`). This can be
        useful for debugging the MPSX data file. If `mpslst` = False, no
        listing is produced.

    names : str, length 8, array-like, shape (5)
        A set of names associated with the MPSX form of the problem.

        `names`[0]

            Must contain either the name of the problem or be blank.

        `names`[1]

            Must contain either the name of the objective row or be
            blank (in which case the first objective free row is used).

        `names`[2]

            Must contain either the name of the RHS set to be used or be
            blank (in which case the first RHS set is used).

        `names`[3]

            Must contain either the name of the RANGE set to be used or
            be blank (in which case the first RANGE set (if any) is
            used).

        `names`[4]

            Must contain either the name of the BOUNDS set to be used or
            be blank (in which case the first BOUNDS set (if any) is
            used).

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    n : int
        n, the actual number of variables in the problem.

    m : int
        m, the actual number of general linear constraints in the
        problem (including the objective row).

    nnz : int
        The actual number of nonzeros in the problem (including the
        objective row).

    iobj : int
        If `iobj` > 0, row `iobj` of A is a free row containing the
        nonzero coefficients of the vector c.

        If `iobj` = 0, the coefficients of c are assumed to be zero.

        If `iobj` = -1, no such row was found and the function
        terminates with `errno` = 4 or 5 (see [Raises/Warns]).

    ncolh : int
        `ncolh` = 0. For QP problems, `ncolh` is the number of leading
        nonzero columns of the Hessian matrix H and must therefore be
        set > 0 before calling ``qpconvex1_sparse_solve``.

    a : float, ndarray, shape (`maxnnz`)
        The nonzero elements of A, ordered by increasing column index.

    ha : int, ndarray, shape (`maxnnz`)
        The row indices of the nonzero elements stored in `a`.

    ka : int, ndarray, shape (`maxn`+1)
        A set of pointers to the beginning of each column of A. More
        precisely, `ka`[i-1] contains the index in `a` of the start of
        the ith column, for i = 1,2,...,`n`. Note that `ka`[0] = 1 and
        `ka`[`n`] = `nnz`+1.

    bl : float, ndarray, shape (`maxn`+`maxm`)
        `bl` contains the vector l (the lower bounds) for all the
        variables and constraints

    bu : float, ndarray, shape (`maxn`+`maxm`)
        `bu` contains the vector u (the upper bounds) for all the
        variables and constraints

    start : str, length 1
        `start` = 'C' and an internal Crash procedure will be used by
        ``qpconvex1_sparse_solve`` to choose an initial basis.

    names : str, length 8, ndarray, shape (5)
        A set of names associated with the problem as defined in the
        MPSX data file as follows:

        `names`[0]

            Contains the name of the problem (or blank if none).

        `names`[1]

            Contains the name of the objective row (or blank if none).

        `names`[2]

            Contains the name of the RHS set (or blank if none).

        `names`[3]

            Contains the name of the RANGE set (or blank if none).

        `names`[4]

            Contains the name of the BOUNDS set (or blank if none).

    nname : int
        n+m, the total number of variables and constraints in the
        problem.

    crname : str, length 8, ndarray, shape (`maxn`+`maxm`)
        The MPSX names of all the variables and constraints in the
        problem in the following order. The first `n` elements contain
        the MPSX names for the variables and the next `m` elements
        contain the MPSX names for the objective row and general linear
        constraints (if any). Note that the MPSX name for the objective
        row is stored in `crname`[`n`+`iobj`-1].

    xs : float, ndarray, shape (`maxn`+`maxm`)
        A set of initial values for the variables and constraints in the
        problem. More precisely, `xs`[j-1] = min(max(0.0, `bl`[j-1]),
        `bu`[j-1]), for j = 1,2,...,`nname`.

    istate : int, ndarray, shape (`maxn`+`maxm`)
        A set of initial states for the variables and constraints in the
        problem. More precisely, `istate`[j-1] = 1 if `xs`[j-1] =
        `bu`[j-1] and 0 otherwise, for j = 1,2,...,`nname`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            Too many rows. Limit is *<value>*, but the actual number
            required is *<value>*.

        (`errno` 2)
            Too many columns. Limit is *<value>*, but the actual number
            required is *<value>*.

        (`errno` 3)
            Too many nonzeros. Limit is *<value>*, but the actual number
            required is *<value>*.

        (`errno` 4)
            No objective row found.

        (`errno` 5)
            Objective row name *<value>* is not defined in the ROWS
            section.

        (`errno` 6)
            No rows specified.

        (`errno` 7)
            Illegal constraint type *<value>*.

        (`errno` 8)
            Row name with leading blank or non-alphanumeric character:
            *<value>*.

        (`errno` 9)
            Column name with leading blank or non-alphanumeric
            character: *<value>*.

        (`errno` 10)
            Illegal bound type *<value>*.

        (`errno` 11)
            Column name *<value>* is not defined in the COLUMNS section.

        (`errno` 12)
            The last line must be the ENDATA indicator line.

        (`errno` 13)
            Line *<value>* is not a comment nor a valid line.

        (`errno` 14)
            Row name *<value>* is not defined in the ROWS section.

        (`errno` 15)
            No columns specified.

        (`errno` 16)
            BOUNDS name *<value>* was not found.

        (`errno` 16)
            RANGES name *<value>* was not found.

        (`errno` 16)
            RHS name *<value>* was not found.

        (`errno` 17)
            On entry, `xbldef` = *<value>* and `xbudef` = *<value>*.

            Constraint: `xbldef` <= `xbudef`.

        (`errno` 17)
            On entry, `maxnnz` = *<value>*.

            Constraint: `maxnnz` >= 1.

        (`errno` 17)
            On entry, `maxm` = *<value>*.

            Constraint: `maxm` >= 1.

        (`errno` 17)
            On entry, `maxn` = *<value>*.

            Constraint: `maxn` >= 1.

        (`errno` 17)
            On entry, `infile` = *<value>*.

            Constraint: 0 <= `infile` <= 99.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    IBM, 1971, `MPSX -- Mathematical programming system`, Program Number
    5734 XM4, IBM Trade Corporation, New York
    """

    fun_name = 'naginterfaces.library.opt.qpconvex1_sparse_mps'

    _maxnnz = _b_utils._EngineIntScalarType(
        maxnnz, {'fun_name': fun_name, 'entity': 'maxnnz'},
    ).to_ctypes()
    _maxnnz_py = _b_utils._EngineIntScalarType.to_py(_maxnnz)
    _a = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_maxnnz_py,),
    )
    _ha = _b_utils._EngineIntArrayType.empty_ndarray(
        (_maxnnz_py,),
    )
    _maxn = _b_utils._EngineIntScalarType(
        maxn, {'fun_name': fun_name, 'entity': 'maxn'},
    ).to_ctypes()
    _maxn_py = _b_utils._EngineIntScalarType.to_py(_maxn)
    _ka = _b_utils._EngineIntArrayType.empty_ndarray(
        (_maxn_py+1,),
    )
    _maxm = _b_utils._EngineIntScalarType(
        maxm, {'fun_name': fun_name, 'entity': 'maxm'},
    ).to_ctypes()
    _maxm_py = _b_utils._EngineIntScalarType.to_py(_maxm)
    _bl = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_maxn_py+_maxm_py,),
    )
    _bu = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_maxn_py+_maxm_py,),
    )
    _names = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'names'},
        names,
        copy=True,
    )
    _crname = _b_utils._EngineCharArrayType.empty_ndarray(
        (_maxn_py+_maxm_py,),
        chrlen=8,
    )
    _xs = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_maxn_py+_maxm_py,),
    )
    _istate = _b_utils._EngineIntArrayType.empty_ndarray(
        (_maxn_py+_maxm_py,),
    )
    _return_scalars = _b_opt.qpconvex1_sparse_mps(
        infile,
        _maxn,
        _maxm,
        _maxnnz,
        xbldef,
        xbudef,
        mpslst,
        _a,
        _ha,
        _ka,
        _bl,
        _bu,
        _names,
        _crname,
        _xs,
        _istate,
        io_manager,
    )
    _names_py = _names
    _n_py = _return_scalars.n
    _m_py = _return_scalars.m
    _nnz_py = _return_scalars.nnz
    _iobj_py = _return_scalars.iobj
    _ncolh_py = _return_scalars.ncolh
    _a_py = _a
    _ha_py = _ha
    _ka_py = _ka
    _bl_py = _bl
    _bu_py = _bu
    _start_py = _return_scalars.start
    _nname_py = _return_scalars.nname
    _crname_py = _crname
    _xs_py = _xs
    _istate_py = _istate
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['n', 'm', 'nnz', 'iobj', 'ncolh', 'a', 'ha', 'ka', 'bl', 'bu', 'start', 'names', 'nname', 'crname', 'xs', 'istate'],
    )(
        _n_py,
        _m_py,
        _nnz_py,
        _iobj_py,
        _ncolh_py,
        _a_py,
        _ha_py,
        _ka_py,
        _bl_py,
        _bu_py,
        _start_py,
        _names_py,
        _nname_py,
        _crname_py,
        _xs_py,
        _istate_py,
    )
    return _return_data

def lsq_lincon_solve(c, bl, bu, cvec, istate, kx, x, a, b, comm, io_manager=None):
    u"""
    Linear programming (LP) convex quadratic programming (QP) or
    linearly-constrained linear least squares problem, dense.

    Note: this function uses optional algorithmic parameters, see also:
    ``lsq_lincon_option_file``, ``lsq_lincon_option_string``,
    ``nlp1_init``.

    ``lsq_lincon_solve`` solves linearly constrained linear least
    squares problems and convex quadratic programming problems.
    It is not intended for large sparse problems.

    For full information please refer to the NAG Library document for
    e04nc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04ncf.html

    Parameters
    ----------
    c : float, array-like, shape (nclin, :)
        Note: the required extent for this argument in dimension 2 is
        determined as follows: if nclin>0: n; otherwise: 1.

        The ith row of `c` must contain the coefficients of the ith
        general constraint, for i = 1,2,...,nclin.

        If nclin = 0, `c` is not referenced.

    bl : float, array-like, shape (n+nclin)
        `bl` must contain the lower bounds for all the constraints

    bu : float, array-like, shape (n+nclin)
        `bu` must contain the upper bounds for all the constraints

    cvec : float, array-like, shape (:)
        Note: the required length for this argument is determined as
        follows: if `comm`['iwsav'][402] in (2, 4, 6, 8, 10): n;
        otherwise: 1.

        The coefficients of the explicit linear term of the objective
        function.

        If the problem is of type FP, QP1, QP3, LS1 (the default) or
        LS3, `cvec` is not referenced.

    istate : int, array-like, shape (n+nclin)
        Need not be set if the (default) option 'Cold Start' is used.

        If the option 'Warm Start' has been chosen, `istate` specifies
        the desired status of the constraints at the start of the
        feasibility phase.

        More precisely, the first n elements of `istate` refer to the
        upper and lower bounds on the variables, and the next n_L
        elements refer to the general linear constraints (if any).

        Possible values for `istate`[j-1] are as follows:

        [table omitted]

        The values -2, -1 and 4 are also acceptable but will be reset to
        zero by the function.

        If ``lsq_lincon_solve`` has been called previously with the same
        values of n and nclin, `istate` already contains satisfactory
        information. (See also the description of the option 'Warm
        Start'.) The function also adjusts (if necessary) the values
        supplied in `x` to be consistent with `istate`.

    kx : int, array-like, shape (n)
        Need not be initialized for problems of type FP, LP, QP1, QP2,
        LS1 (the default) or LS2.

        For problems QP3, QP4, LS3 or LS4, `kx` must specify the order
        of the columns of the matrix A with respect to the ordering of
        `x`.

        Thus if column j of A is the column associated with the variable
        x_i then `kx`[j-1] = i.

    x : float, array-like, shape (n)
        An initial estimate of the solution.

        Note: that it may be best to avoid the choice `x` = 0.0.

    a : float, array-like, shape (m, :)
        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `comm`['iwsav'][402]>2 or
        `comm`['iwsav'][402]<=0: n; otherwise: 1.

        The matrix A.

    b : float, array-like, shape (:)
        Note: the required length for this argument is determined as
        follows: if `comm`['iwsav'][402]<1 or `comm`['iwsav'][402]>6: m;
        otherwise: 1.

        The m elements of the vector of observations.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``nlp1_init``.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    istate : int, ndarray, shape (n+nclin)
        The status of the constraints in the working set at the point
        returned in `x`. The significance of each possible value of
        `istate`[j-1] is as follows:

        [table omitted]

    kx : int, ndarray, shape (n)
        Defines the order of the columns of `a` with respect to the
        ordering of `x`, as described above.

    x : float, ndarray, shape (n)
        The point at which ``lsq_lincon_solve`` terminated. If the
        function exits successfully or `errno` = 1 or 4, `x` contains an
        estimate of the solution.

    a : float, ndarray, shape (m, :)
        The matrix A.

    b : float, ndarray, shape (:)
        The transformed residual vector of equation (0) (see [Main
        Iteration]).

        If the problem is of type FP, LP, QP1, QP2, QP3 or QP4, `b` is
        not referenced.

    itera : int
        The total number of iterations performed.

    obj : float
        The value of the objective function at x if x is feasible, or
        the sum of infeasibiliites at x otherwise. If the problem is of
        type FP and x is feasible, `obj` is set to zero.

    clamda : float, ndarray, shape (n+nclin)
        The values of the Lagrange multipliers for each constraint with
        respect to the current working set. The first n elements contain
        the multipliers for the bound constraints on the variables, and
        the next n_L elements contain the multipliers for the general
        linear constraints (if any). If `istate`[j-1] = 0 (i.e.,
        constraint j is not in the working set), `clamda`[j-1] is zero.
        If x is optimal, `clamda`[j-1] should be non-negative if
        `istate`[j-1] = 1, non-positive if `istate`[j-1] = 2 and zero if
        `istate`[j-1] = 4.

    Other Parameters
    ----------------
    'Cold Start' : valueless
        Default

        This option specifies how the initial working set is chosen.
        With a 'Cold Start', ``lsq_lincon_solve`` chooses the initial
        working set based on the values of the variables and constraints
        at the initial point.
        Broadly speaking, the initial working set will include equality
        constraints and bounds or inequality constraints that violate or
        'nearly' satisfy their bounds (to within 'Crash Tolerance').

        With a 'Warm Start', you must provide a valid definition of
        every element of the array `istate`. ``lsq_lincon_solve`` will
        override your specification of `istate` if necessary, so that a
        poor choice of the working set will not cause a fatal error.
        For instance, any elements of `istate` which are set to -2, -1
        or 4 will be reset to zero, as will any elements which are set
        to 3 when the corresponding elements of `bl` and `bu` are not
        equal.
        A warm start will be advantageous if a good estimate of the
        initial working set is available -- for example, when
        ``lsq_lincon_solve`` is called repeatedly to solve related
        problems.

    'Warm Start' : valueless
        This option specifies how the initial working set is chosen.
        With a 'Cold Start', ``lsq_lincon_solve`` chooses the initial
        working set based on the values of the variables and constraints
        at the initial point.
        Broadly speaking, the initial working set will include equality
        constraints and bounds or inequality constraints that violate or
        'nearly' satisfy their bounds (to within 'Crash Tolerance').

        With a 'Warm Start', you must provide a valid definition of
        every element of the array `istate`. ``lsq_lincon_solve`` will
        override your specification of `istate` if necessary, so that a
        poor choice of the working set will not cause a fatal error.
        For instance, any elements of `istate` which are set to -2, -1
        or 4 will be reset to zero, as will any elements which are set
        to 3 when the corresponding elements of `bl` and `bu` are not
        equal.
        A warm start will be advantageous if a good estimate of the
        initial working set is available -- for example, when
        ``lsq_lincon_solve`` is called repeatedly to solve related
        problems.

    'Crash Tolerance' : float
        Default = 0.01

        This value is used in conjunction with the option 'Cold Start'
        (the default value) when ``lsq_lincon_solve`` selects an initial
        working set.
        If 0 <= r <= 1, the initial working set will include (if
        possible) bounds or general inequality constraints that lie
        within r of their bounds.
        In particular, a constraint of the form c_j^Tx >= l will be
        included in the initial working set if |c_j^Tx-l| <= r(1+|l|).
        If r < 0 or r > 1, the default value is used.

    'Defaults' : valueless
        This special keyword may be used to reset all options to their
        default values.

    'Feasibility Phase Iteration Limit' : int
        Default = max(50, 5*(n+n_L))

        The scalars i_1 and i_2 specify the maximum number of iterations
        allowed in the feasibility and optimality phases. Option
        'Optimality Phase Iteration Limit' is equivalent to option
        'Iteration Limit'.
        Setting i_2 = 0 and 'Print Level' > 0 means that the workspace
        needed will be computed and printed, but no iterations will be
        performed.
        If i_1 < 0 or i_2 < 0, the default value is used.

    'Optimality Phase Iteration Limit' : int
        Default = max(50, 5*(n+n_L))

        The scalars i_1 and i_2 specify the maximum number of iterations
        allowed in the feasibility and optimality phases. Option
        'Optimality Phase Iteration Limit' is equivalent to option
        'Iteration Limit'.
        Setting i_2 = 0 and 'Print Level' > 0 means that the workspace
        needed will be computed and printed, but no iterations will be
        performed.
        If i_1 < 0 or i_2 < 0, the default value is used.

    'Feasibility Tolerance' : float
        Default = sqrt(epsilon)

        If r > epsilon, r defines the maximum acceptable `absolute`
        violation in each constraint at a 'feasible' point.
        For example, if the variables and the coefficients in the
        general constraints are of order unity, and the latter are
        correct to about 6 decimal digits, it would be appropriate to
        specify r as 10^(-6).
        If 0 <= r < epsilon, the default value is used.

        Note that a 'feasible solution' is a solution that satisfies the
        current constraints to within the tolerance r.

    'Hessian' : str
        Default = NO

        This option controls the contents of the upper triangular matrix
        R (see the description of `a` in [Parameters]).
        ``lsq_lincon_solve`` works exclusively with the transformed and
        reordered matrix H_Q (8), and hence extra computation is
        required to form the Hessian itself.
        If 'Hessian' = 'NO', `a` contains the Cholesky factor of the
        matrix H_Q with columns ordered as indicated by `kx` (see
        [Parameters]).
        If 'Hessian' = 'YES', `a` contains the Cholesky factor of the
        matrix H, with columns ordered as indicated by `kx`.

    'Infinite Bound Size' : float
        Default = 10^20

        If r > 0, r defines the 'infinite' bound bigbnd in the
        definition of the problem constraints.
        Any upper bound greater than or equal to bigbnd will be regarded
        as + infinity (and similarly any lower bound less than or equal
        to -bigbnd will be regarded as - infinity).
        If r < 0, the default value is used.

    'Infinite Step Size' : float
        Default = max(bigbnd, 10^20)

        If r > 0, r specifies the magnitude of the change in variables
        that will be considered a step to an unbounded solution. (Note
        that an unbounded solution can occur only when the Hessian is
        singular and the objective contains an explicit linear term.) If
        the change in x during an iteration would exceed the value of r,
        the objective function is considered to be unbounded below in
        the feasible region.
        If r <= 0, the default value is used.

    'Iteration Limit' : int
        Default = max(50, 5*(n+n_L))

        See option 'Feasibility Phase Iteration Limit'.

    'Iters' : int
        Default = max(50, 5*(n+n_L))

        See option 'Feasibility Phase Iteration Limit'.

    'Itns' : int
        Default = max(50, 5*(n+n_L))

        See option 'Feasibility Phase Iteration Limit'.

    'List' : valueless
        Option 'List' enables printing of each option specification as
        it is supplied. 'Nolist' suppresses this printing.

    'Nolist' : valueless
        Default for ``lsq_lincon_solve`` = 'Nolist'

        Option 'List' enables printing of each option specification as
        it is supplied. 'Nolist' suppresses this printing.

    'Monitoring File' : int
        Default = -1

        If i >= 0 and 'Print Level' >= 5, monitoring information
        produced by ``lsq_lincon_solve`` at every iteration is sent to a
        file with logical unit number i.
        If i < 0 and/or 'Print Level' < 5, no monitoring information is
        produced.

    'Print Level' : int
        The value of i controls the amount of printout produced by
        ``lsq_lincon_solve``, as indicated below.
        A detailed description of the printed output is given in
        [Further Comments] (summary output at each iteration and the
        final solution) and [Monitoring Information] (monitoring
        information at each iteration).

        The following printout is sent to the file object associated
        with the advisory I/O unit (see
        :class:`~naginterfaces.base.utils.FileObjManager`):

        [table omitted]

        The following printout is sent to the unit number given by the
        option 'Monitoring File':

        [table omitted]

        If 'Print Level' >= 5 and the unit number defined by the option
        'Monitoring File' is the advisory unit number, the summary
        output for each major iteration is suppressed.

    'Problem Type' : str
        Default = LS1

        This option specifies the type of objective function to be
        minimized during the optimality phase.
        The following are the nine optional keywords and the dimensions
        of the arrays that must be specified in order to define the
        objective function:

        [table omitted]

        For problems of type FP, the objective function is omitted and
        `a`, `b` and `cvec` are not referenced.

        The following keywords are also acceptable.

        [table omitted]

        In addition, the keywords LS and LSQ are equivalent to the
        default option LS1, and the keyword QP is equivalent to the
        option QP2.

        If A = 0, i.e., the objective function is purely linear, the
        efficiency of ``lsq_lincon_solve`` may be increased by
        specifying a as LP.

    'Rank Tolerance' : float
        Default = 100 epsilon or 10sqrt(epsilon) (see below)

        Note that this option does not apply to problems of type FP or
        LP.

        The default value of r depends on the problem type.
        If A occurs as a least squares matrix, as it does in problem
        types QP1, LS1 and LS3, then the default value of r is 100
        epsilon.
        In all other cases, A is treated as the 'square root' of the
        Hessian matrix H and r has the default value 10sqrt(epsilon).

        This argument enables you to control the estimate of the
        triangular factor R_1 (see [Main Iteration]).
        If rho_i denotes the function rho_i =
        max{|R_11|,|R_22|,...,|R_{ii}|}, the rank of R is defined to be
        smallest index `i` such that |R_{i+1,i+1}| <= r|rho_{i+1}|.
        If r <= 0, the default value is used.

    Raises
    ------
    NagValueError
        (`errno` 4)
            Too many iterations.

        (`errno` 5)
            Too many iterations without changing x.

        (`errno` 6)
            On entry, n = *<value>*.

            Constraint: n > 0.

        (`errno` 6)
            On entry, nclin = *<value>*.

            Constraint: nclin >= 0.

        (`errno` 6)
            On entry, m = *<value>*.

            Constraint: m > 0.

        (`errno` 6)
            On entry, `kx` has not been supplied as a valid permutation.

        (`errno` 6)
            On entry, the equal bounds on *<value>* are infinite,
            because `bl`[*<value>*] = beta and `bu`[*<value>*] = beta,
            but abs(beta) >= bigbnd: beta = *<value>* and bigbnd =
            *<value>*.

        (`errno` 6)
            On entry, the bounds on *<value>* are inconsistent:
            `bl`[*<value>*] = *<value>* and `bu`[*<value>*] = *<value>*.

        (`errno` 6)
            On entry with a Warm Start, `istate`[*<value>*] = *<value>*.

        (`errno` 6)
            On entry, the equal bounds on variable *<value>* are
            infinite, because `bl`[*<value>*] = beta and `bu`[*<value>*]
            = beta, but abs(beta) >= bigbnd: beta = *<value>* and bigbnd
            = *<value>*.

        (`errno` 6)
            On entry, the equal bounds on linear constraint *<value>*
            are infinite, because `bl`[*<value>*] = beta and
            `bu`[*<value>*] = beta, but abs(beta) >= bigbnd: beta =
            *<value>* and bigbnd = *<value>*.

        (`errno` 6)
            On entry, the equal bounds on nonlinear constraint *<value>*
            are infinite, because `bl`[*<value>*] = beta and
            `bu`[*<value>*] = beta, but abs(beta) >= bigbnd: beta =
            *<value>* and bigbnd = *<value>*.

        (`errno` 6)
            On entry, the bounds on variable *<value>* are inconsistent:
            `bl`[*<value>*] = *<value>* and `bu`[*<value>*] = *<value>*.

        (`errno` 6)
            On entry, the bounds on linear constraint *<value>* are
            inconsistent: `bl`[*<value>*] = *<value>* and
            `bu`[*<value>*] = *<value>*.

        (`errno` 6)
            On entry, the bounds on nonlinear constraint *<value>* are
            inconsistent: `bl`[*<value>*] = *<value>* and
            `bu`[*<value>*] = *<value>*.

        (`errno` 7)
            The problem to be solved is of type QP1 or QP2, but the
            Hessian matrix supplied in `a` is not positive semidefinite.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            Weak *<value>* solution.

        (`errno` 2)
            *<value>* solution is unbounded.

        (`errno` 3)
            Cannot satisfy the linear constraints.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Gill, P E, Hammarling, S, Murray, W, Saunders, M A and Wright, M H,
    1986, `Users' guide for LSSOL (Version 1.0)`, Report SOL 86-1,
    Department of Operations Research, Stanford University

    Gill, P E, Murray, W, Saunders, M A and Wright, M H, 1984,
    `Procedures for optimization problems with a mixture of bounds and
    general linear constraints`, ACM Trans. Math. Software (10),
    282--298

    Gill, P E, Murray, W and Wright, M H, 1981, `Practical
    Optimization`, Academic Press

    Stoer, J, 1971, `On the numerical solution of constrained least
    squares problems`, SIAM J. Numer. Anal. (8), 382--411
    """

    fun_name = 'naginterfaces.library.opt.lsq_lincon_solve'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _kx = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'kx'},
        kx,
        copy=True,
    )
    _n_py = _b_utils._EngineIntArrayType.get_shape(_kx, {'fun_name': fun_name, 'entity': 'kx'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _c = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nclin_py = _b_utils._EngineFloat64ArrayType.get_shape(_c, {'fun_name': fun_name, 'entity': 'c'}, exp_rank=2)[0]
    _nclin = _b_utils._EngineIntScalarType(
        _nclin_py, {'fun_name': fun_name, 'entity': '_nclin'},
    ).to_ctypes()
    _bl = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'bl'},
        bl,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _bu = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'bu'},
        bu,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _cvec = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'cvec'},
        cvec,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _istate = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'istate'},
        istate,
        copy=True,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _clamda = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py+_nclin_py,),
    )
    try:
        _return_scalars = _b_opt.lsq_lincon_solve(
            _m,
            _n,
            _nclin,
            _c,
            _bl,
            _bu,
            _cvec,
            _istate,
            _kx,
            _x,
            _a,
            _b,
            _clamda,
            comm,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _istate_py = _istate
    _kx_py = _kx
    _x_py = _x
    _a_py = _a
    _b_py = _b
    _itera_py = _return_scalars.itera
    _obj_py = _return_scalars.obj
    _clamda_py = _clamda
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['istate', 'kx', 'x', 'a', 'b', 'itera', 'obj', 'clamda'],
    )(
        _istate_py,
        _kx_py,
        _x_py,
        _a_py,
        _b_py,
        _itera_py,
        _obj_py,
        _clamda_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def lsq_lincon_option_file(ioptns, comm, io_manager=None):
    u"""
    Supply option values for ``lsq_lincon_solve`` from external file.

    ``lsq_lincon_option_file`` may be used to supply options to
    ``lsq_lincon_solve`` from an external file.

    For full information please refer to the NAG Library document for
    e04nd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04ndf.html

    Parameters
    ----------
    ioptns : int
        The unit number (see
        :meth:`~naginterfaces.base.utils.FileObjManager.unit_from_fileobj`)
        of the options file to be read.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``nlp1_init``.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Raises
    ------
    NagValueError
        (`errno` 1)
            `ioptns` is not in the range [0,99].

        (`errno` 1)
            On entry, `ioptns` = *<value>*.

            Constraint: 0 <= `ioptns` <= 99.

        (`errno` 2)
            ``Begin`` was found, but end-of-file was found before
            ``End`` was found.

        (`errno` 3)
            End-of-file was found before ``Begin`` was found.

        (`errno` 5)
            One or more lines of the options file is invalid.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    _b_opt.lsq_lincon_option_file(
        ioptns,
        comm,
        io_manager,
    )

def lsq_lincon_option_string(optstr, comm, io_manager=None):
    u"""
    Supply option values to ``lsq_lincon_solve`` from a character
    string.

    ``lsq_lincon_option_string`` may be used to supply individual
    options to ``lsq_lincon_solve``.

    For full information please refer to the NAG Library document for
    e04ne

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04nef.html

    Parameters
    ----------
    optstr : str
        A single valid option string (as described in [Description] and
        in [Other Parameters for lsq_lincon_solve]).

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``nlp1_init``.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Raises
    ------
    NagValueError
        (`errno` 5)
            The supplied option string is invalid. Supplied value was:
            *<value>*.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    _b_opt.lsq_lincon_option_string(
        optstr,
        comm,
        io_manager,
    )

def qp_dense_solve(nclin, a, bl, bu, cvec, h, istate, x, comm, qphess=None, data=None, io_manager=None):
    u"""
    General (possibly non-convex) quadratic programming (QP), dense,
    active-set method.

    Note: this function uses optional algorithmic parameters, see also:
    ``qp_dense_option_file``, ``qp_dense_option_string``, ``nlp1_init``.

    ``qp_dense_solve`` solves general quadratic programming problems.
    It is not intended for large sparse problems.

    For full information please refer to the NAG Library document for
    e04nf

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04nff.html

    Parameters
    ----------
    nclin : int
        m_L, the number of general linear constraints.

    a : float, array-like, shape (`nclin`, :)
        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `nclin`>0: n; if `nclin`=0: 1;
        otherwise: 0.

        The ith row of `a` must contain the coefficients of the ith
        general linear constraint, for i = 1,2,...,m_L.

        If `nclin` = 0, `a` is not referenced.

    bl : float, array-like, shape (n+`nclin`)
        `bl` must contain the lower bounds for all the constraints

    bu : float, array-like, shape (n+`nclin`)
        `bu` must contain the upper bounds for all the constraints

    cvec : float, array-like, shape (:)
        Note: the required length for this argument is determined as
        follows: if `comm`['iwsav'][364]<1 or `comm`['iwsav'][364] in
        (2, 4, 6): n; otherwise: 1.

        The coefficients of the explicit linear term of the objective
        function when the problem is of type LP, QP2 (the default) and
        QP4.

        If the problem is of type FP, QP1, or QP3, `cvec` is not
        referenced.

    h : float, array-like, shape (:, :)
        May be used to store the quadratic term H of the QP objective
        function if desired. In some cases, you need not use `h` to
        store H explicitly (see the specification of function `qphess`).
        The elements of `h` are referenced only by function `qphess`.
        The number of rows of H is denoted by m, whose default value is
        n. (The option 'Hessian Rows' may be used to specify a value of
        m < n.)

        If the default version of `qphess` is used and the problem is of
        type QP1 or QP2 (the default), the first m rows and columns of
        `h` must contain the leading m by m rows and columns of the
        symmetric Hessian matrix H.

        Only the diagonal and upper triangular elements of the leading m
        rows and columns of `h` are referenced.

        The remaining elements need not be assigned.

        If the default version of `qphess` is used and the problem is of
        type QP3 or QP4, the first m rows of `h` must contain an m by n
        upper trapezoidal factor of the symmetric Hessian matrix H^TH.

        The factor need not be of full rank, i.e., some of the diagonal
        elements may be zero.

        However, as a general rule, the larger the dimension of the
        leading nonsingular sub-matrix of `h`, the fewer iterations will
        be required.

        Elements outside the upper trapezoidal part of the first m rows
        of `h` need not be assigned.

        If a non-default version of `qphess` is supplied, then in some
        cases it may be desirable to use a one-dimensional array to
        transmit data to `qphess`. `h` is then declared as an ldh by 1
        array, where ldh >= n*(n+1)/2.

        In other situations, it may be desirable to compute Hx or H^THx
        without accessing `h` -- for example, if H or H^TH is sparse or
        has special structure.

        The arguments `h` and ldh may then refer to any convenient
        array.

        If the problem is of type FP or LP, `h` is not referenced.

    istate : int, array-like, shape (n+`nclin`)
        Need not be set if the (default) option 'Cold Start' is used.

        If the option 'Warm Start' has been chosen, `istate` specifies
        the desired status of the constraints at the start of the
        feasibility phase.

        More precisely, the first n elements of `istate` refer to the
        upper and lower bounds on the variables, and the next m_L
        elements refer to the general linear constraints (if any).

        Possible values for `istate`[j-1] are as follows:

        [table omitted]

        The values -2, -1 and 4 are also acceptable but will be reset to
        zero by the function.

        If ``qp_dense_solve`` has been called previously with the same
        values of n and `nclin`, `istate` already contains satisfactory
        information. (See also the description of the option 'Warm
        Start'.) The function also adjusts (if necessary) the values
        supplied in `x` to be consistent with `istate`.

    x : float, array-like, shape (n)
        An initial estimate of the solution.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``nlp1_init``.

    qphess : None or callable hx = qphess(jthcol, h, x, data=None),
        optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        In general, you need not provide a version of `qphess`.

        However, the algorithm of ``qp_dense_solve`` requires only the
        product of H or H^TH and a vector x; and in some cases you may
        obtain increased efficiency by providing a version of `qphess`
        that avoids the need to define the elements of the matrices H or
        H^TH explicitly.

        `qphess` is not referenced if the problem is of type FP or LP,
        in which case `qphess` may be None.

        Parameters
        ~~~~~~~~~~
        jthcol : int
            Specifies whether or not the vector x is a column of the
            identity matrix.

            `jthcol` = j > 0

                The vector x is the jth column of the identity matrix,
                and hence Hx or H^THx is the jth column of H or H^TH,
                respectively. This may in some cases require very little
                computation and `qphess` may be coded to take advantage
                of this. However special code is not necessary because x
                is always stored explicitly in the array `x`.

            `jthcol` = 0

                x has no special form.

        h : float, ndarray, shape (:, :)
            See: `h`

        x : float, ndarray, shape (n)
            The vector x.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        hx : float, array-like, shape (n)
            The product Hx if the problem is of type QP1 or QP2 (the
            default), or the product H^THx if the problem is of type QP3
            or QP4.

    data : arbitrary, optional
        User-communication data for callback functions.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    istate : int, ndarray, shape (n+`nclin`)
        The status of the constraints in the working set at the point
        returned in `x`. The significance of each possible value of
        `istate`[j-1] is as follows:

        [table omitted]

    x : float, ndarray, shape (n)
        The point at which ``qp_dense_solve`` terminated. If the
        function exits successfully or `errno` = 1 or 4, `x` contains an
        estimate of the solution.

    itera : int
        The total number of iterations performed.

    obj : float
        The value of the objective function at x if x is feasible, or
        the sum of infeasibilities at x otherwise. If the problem is of
        type FP and x is feasible, `obj` is set to zero.

    ax : float, ndarray, shape (max(1, `nclin`))
        The final values of the linear constraints Ax.

        If `nclin` = 0, `ax` is not referenced.

    clamda : float, ndarray, shape (n+`nclin`)
        The values of the Lagrange multipliers for each constraint with
        respect to the current working set. The first n elements contain
        the multipliers for the bound constraints on the variables, and
        the next m_L elements contain the multipliers for the general
        linear constraints (if any). If `istate`[j-1] = 0 (i.e.,
        constraint j is not in the working set), `clamda`[j-1] is zero.
        If x is optimal, `clamda`[j-1] should be non-negative if
        `istate`[j-1] = 1, non-positive if `istate`[j-1] = 2 and zero if
        `istate`[j-1] = 4.

    Other Parameters
    ----------------
    'Check Frequency' : float
        Default = 50

        Every ith iteration, a numerical test is made to see if the
        current solution x satisfies the constraints in the working set.
        If the largest residual of the constraints in the working set is
        judged to be too large, the current working set is refactorized
        and the variables are recomputed to satisfy the constraints more
        accurately.
        If i <= 0, the default value is used.

    'Cold Start' : valueless
        Default

        This option specifies how the initial working set is chosen.
        With a 'Cold Start', ``qp_dense_solve`` chooses the initial
        working set based on the values of the variables and constraints
        at the initial point.
        Broadly speaking, the initial working set will include equality
        constraints and bounds or inequality constraints that violate or
        'nearly' satisfy their bounds (to within 'Crash Tolerance').

        With a 'Warm Start', you must provide a valid definition of
        every element of the array `istate`. ``qp_dense_solve`` will
        override your specification of `istate` if necessary, so that a
        poor choice of the working set will not cause a fatal error.
        For instance, any elements of `istate` which are set to -2, -1
        or 4 will be reset to zero, as will any elements which are set
        to 3 when the corresponding elements of `bl` and `bu` are not
        equal.
        A warm start will be advantageous if a good estimate of the
        initial working set is available -- for example, when
        ``qp_dense_solve`` is called repeatedly to solve related
        problems.

    'Warm Start' : valueless
        This option specifies how the initial working set is chosen.
        With a 'Cold Start', ``qp_dense_solve`` chooses the initial
        working set based on the values of the variables and constraints
        at the initial point.
        Broadly speaking, the initial working set will include equality
        constraints and bounds or inequality constraints that violate or
        'nearly' satisfy their bounds (to within 'Crash Tolerance').

        With a 'Warm Start', you must provide a valid definition of
        every element of the array `istate`. ``qp_dense_solve`` will
        override your specification of `istate` if necessary, so that a
        poor choice of the working set will not cause a fatal error.
        For instance, any elements of `istate` which are set to -2, -1
        or 4 will be reset to zero, as will any elements which are set
        to 3 when the corresponding elements of `bl` and `bu` are not
        equal.
        A warm start will be advantageous if a good estimate of the
        initial working set is available -- for example, when
        ``qp_dense_solve`` is called repeatedly to solve related
        problems.

    'Crash Tolerance' : float
        Default = 0.01

        This value is used in conjunction with the option 'Cold Start'
        (the default value) when ``qp_dense_solve`` selects an initial
        working set.
        If 0 <= r <= 1, the initial working set will include (if
        possible) bounds or general inequality constraints that lie
        within r of their bounds.
        In particular, a constraint of the form a_j^Tx >= l will be
        included in the initial working set if |a_j^Tx-l| <= r(1+|l|).
        If r < 0 or r > 1, the default value is used.

    'Defaults' : valueless
        This special keyword may be used to reset all options to their
        default values.

    'Expand Frequency' : int
        Default = 5

        This option is part of an anti-cycling procedure designed to
        guarantee progress even on highly degenerate problems.

        The strategy is to force a positive step at every iteration, at
        the expense of violating the constraints by a small amount.
        Suppose that the value of the option 'Feasibility Tolerance' is
        delta.
        Over a period of i iterations, the feasibility tolerance
        actually used by ``qp_dense_solve`` (i.e., the `working`
        feasibility tolerance) increases from 0.5 delta to delta (in
        steps of 0.5 delta/i).

        At certain stages the following 'resetting procedure' is used to
        remove constraint infeasibilities.
        First, all variables whose upper or lower bounds are in the
        working set are moved exactly onto their bounds.
        A count is kept of the number of nontrivial adjustments made.
        If the count is positive, iterative refinement is used to give
        variables that satisfy the working set to (essentially) machine
        precision.
        Finally, the working feasibility tolerance is reinitialized to
        0.5 delta.

        If a problem requires more than i iterations, the resetting
        procedure is invoked and a new cycle of i iterations is started
        with i incremented by 10. (The decision to resume the
        feasibility phase or optimality phase is based on comparing any
        constraint infeasibilities with delta.)

        The resetting procedure is also invoked when ``qp_dense_solve``
        reaches an apparently optimal, infeasible or unbounded solution,
        unless this situation has already occurred twice.
        If any nontrivial adjustments are made, iterations are
        continued.

        If i <= 0, the default value is used.
        If i >= 9999999, no anti-cycling procedure is invoked.

    'Feasibility Phase Iteration Limit' : int
        Default = max(50, 5*(n+m_L))

        For problems of type FP, the scalar i_1 specifies the maximum
        number of iterations allowed before temination.
        Setting i_1 = 0 and 'Print Level' > 0 means that the workspace
        needed will be computed and printed, but no iterations will be
        performed.

        For problems of type LP, the maximum number of iterations
        allowed before temination is taken as max(i_1, i_2).
        Setting i_1 = 0, i_2 = 0 and 'Print Level' > 0 means that the
        workspace needed will be computed and printed, but no iterations
        will be performed.

        For problems of type QP, the scalars i_1 and i_2 specify the
        maximum number of iterations allowed in the feasibility and
        optimality phases. 'Optimality Phase Iteration Limit' is
        equivalent to 'Iteration Limit'.
        Setting i_1 = 0 and 'Print Level' > 0 means that the workspace
        needed will be computed and printed, but no iterations will be
        performed.

        If i_1 < 0 or i_2 < 0, the default value is used.

    'Optimality Phase Iteration Limit' : int
        Default = max(50, 5*(n+m_L))

        For problems of type FP, the scalar i_1 specifies the maximum
        number of iterations allowed before temination.
        Setting i_1 = 0 and 'Print Level' > 0 means that the workspace
        needed will be computed and printed, but no iterations will be
        performed.

        For problems of type LP, the maximum number of iterations
        allowed before temination is taken as max(i_1, i_2).
        Setting i_1 = 0, i_2 = 0 and 'Print Level' > 0 means that the
        workspace needed will be computed and printed, but no iterations
        will be performed.

        For problems of type QP, the scalars i_1 and i_2 specify the
        maximum number of iterations allowed in the feasibility and
        optimality phases. 'Optimality Phase Iteration Limit' is
        equivalent to 'Iteration Limit'.
        Setting i_1 = 0 and 'Print Level' > 0 means that the workspace
        needed will be computed and printed, but no iterations will be
        performed.

        If i_1 < 0 or i_2 < 0, the default value is used.

    'Feasibility Tolerance' : float
        Default = sqrt(epsilon)

        If r >= epsilon, r defines the maximum acceptable `absolute`
        violation in each constraint at a 'feasible' point.
        For example, if the variables and the coefficients in the
        general constraints are of order unity, and the latter are
        correct to about 6 decimal digits, it would be appropriate to
        specify r as 10^(-6).
        If 0 <= r < epsilon, the default value is used.

        ``qp_dense_solve`` attempts to find a feasible solution before
        optimizing the objective function.
        If the sum of infeasibilities cannot be reduced to zero, the
        option 'Minimum Sum of Infeasibilities' can be used to find the
        minimum value of the sum.
        Let Sinf be the corresponding sum of infeasibilities.
        If Sinf is quite small, it may be appropriate to raise r by a
        factor of 10 or 100.
        Otherwise, some error in the data should be suspected.

        Note that a 'feasible solution' is a solution that satisfies the
        current constraints to within the tolerance r.

    'Hessian Rows' : int
        Default = n

        Note that this option does not apply to problems of type FP or
        LP.

        This specifies m, the number of rows of the Hessian matrix H.
        The default value of m is n, the number of variables of the
        problem.

        If the problem is of type QP, then m will usually be n, the
        number of variables.
        However, a value of m less than n is appropriate for QP3 or QP4
        if H is an upper trapezoidal matrix with m rows.
        Similarly, m may be used to define the dimension of a leading
        block of nonzeros in the Hessian matrices of QP1 or QP2.
        In this case the last n-m rows and columns of H are assumed to
        be zero.
        In the QP case m should not be greater than n; if it is, the
        last m-n rows of H are ignored.

        If i < 0 or i > n, the default value is used.

    'Infinite Bound Size' : float
        Default = 10^20

        If r > 0, r defines the 'infinite' bound bigbnd in the
        definition of the problem constraints.
        Any upper bound greater than or equal to bigbnd will be regarded
        as + infinity (and similarly any lower bound less than or equal
        to -bigbnd will be regarded as - infinity).
        If r < 0, the default value is used.

    'Infinite Step Size' : float
        Default = max(bigbnd, 10^20)

        If r > 0, r specifies the magnitude of the change in variables
        that will be considered a step to an unbounded solution. (Note
        that an unbounded solution can occur only when the Hessian is
        not positive definite.) If the change in x during an iteration
        would exceed the value of r, then the objective function is
        considered to be unbounded below in the feasible region.
        If r <= 0, the default value is used.

    'Iteration Limit' : int
        Default = max(50, 5*(n+m_L))

        See option 'Feasibility Phase Iteration Limit'.

    'Iters' : int
        Default = max(50, 5*(n+m_L))

        See option 'Feasibility Phase Iteration Limit'.

    'Itns' : int
        Default = max(50, 5*(n+m_L))

        See option 'Feasibility Phase Iteration Limit'.

    'List' : valueless
        Option 'List' enables printing of each option specification as
        it is supplied. 'Nolist' suppresses this printing.

    'Nolist' : valueless
        Default for ``qp_dense_solve`` = 'Nolist'

        Option 'List' enables printing of each option specification as
        it is supplied. 'Nolist' suppresses this printing.

    'Maximum Degrees of Freedom' : int
        Default = n

        Note that this option does not apply to problems of type FP or
        LP.

        This places a limit on the storage allocated for the triangular
        factor R of the reduced Hessian H_R.
        Ideally, i should be set slightly larger than the value of n_R
        expected at the solution.
        It need not be larger than m_n+1, where m_n is the number of
        variables that appear nonlinearly in the quadratic objective
        function.
        For many problems it can be much smaller than m_n.

        For quadratic problems, a minimizer may lie on any number of
        constraints, so that n_R may vary between 1 and n.
        The default value of i is therefore the number of variables n.
        If 'Hessian Rows' m is specified, the default value of i is the
        same number, m.

    'Minimum Sum of Infeasibilities' : str
        Default = NO

        If no feasible point exists for the constraints, then this
        option is used to control whether or not ``qp_dense_solve`` will
        calculate a point that minimizes the constraint violations.
        If 'Minimum Sum of Infeasibilities' = 'NO', ``qp_dense_solve``
        will terminate as soon as it is evident that no feasible point
        exists for the constraints.
        The final point will generally not be the point at which the sum
        of infeasibilities is minimized.
        If 'Minimum Sum of Infeasibilities' = 'YES', ``qp_dense_solve``
        will continue until the sum of infeasibilities is minimized.

    'Monitoring File' : int
        Default = -1

        If i > 6 and 'Print Level' >= 5, monitoring information produced
        by ``qp_dense_solve`` at every iteration is sent to a file with
        logical unit number i.

        If i < 0 and/or 'Print Level' < 5, no monitoring information is
        produced.

    'Optimality Tolerance' : float
        Default = epsilon^0.5

        If r >= epsilon, r defines the tolerance used to determine if
        the bounds and general constraints have the right 'sign' for the
        solution to be judged to be optimal.

        If 0 <= r < epsilon, the default value is used.

    'Print Level' : int
        The value of i controls the amount of printout produced by
        ``qp_dense_solve``, as indicated below.
        A detailed description of the printed output is given in
        [Further Comments] (summary output at each iteration and the
        final solution) and [Monitoring Information] (monitoring
        information at each iteration).
        If i < 0, the default value is used.

        The following printout is sent to the file object associated
        with the advisory I/O unit (see
        :class:`~naginterfaces.base.utils.FileObjManager`):

        [table omitted]

        The following printout is sent to the unit number given by the
        option 'Monitoring File':

        [table omitted]

        If 'Print Level' >= 5 and the unit number defined by the option
        'Monitoring File' is the advisory unit number, the summary
        output is suppressed.

    'Problem Type' : str
        Default = QP2

        This option specifies the type of objective function to be
        minimized during the optimality phase.
        The following are the five optional keywords and the dimensions
        of the arrays that must be specified in order to define the
        objective function:

        [table omitted]

        For problems of type FP the objective function is omitted and
        neither `h` nor `cvec` are referenced.

        The following keywords are also acceptable.

        [table omitted]

        In addition, the keyword QP is equivalent to the default option
        QP2.

        If H = 0 (i.e., the objective function is purely linear), the
        efficiency of ``qp_dense_solve`` may be increased by specifying
        a as LP.

    'Rank Tolerance' : float
        Default = 100 epsilon

        Note that this option does not apply to problems of type FP or
        LP.

        This option enables you to control the condition number of the
        triangular factor R (see [Algorithmic Details]).
        If rho_i denotes the function rho_i =
        max{|R_11|,|R_22|,...,|R_{ii}|}, the dimension of R is defined
        to be smallest index i such that |R_{i+1,i+1}| <=
        sqrt(r)|rho_{i+1}|.
        If r <= 0, the default value is used.

    Raises
    ------
    NagValueError
        (`errno` 4)
            Too many iterations.

        (`errno` 5)
            Reduced Hessian exceeds assigned dimension.

        (`errno` 6)
            On entry, n = *<value>*.

            Constraint: n > 0.

        (`errno` 6)
            On entry, `nclin` = *<value>*.

            Constraint: `nclin` >= 0.

        (`errno` 6)
            On entry, the equal bounds on *<value>* are infinite,
            because `bl`[*<value>*] = beta and `bu`[*<value>*] = beta,
            but abs(beta) >= bigbnd: beta = *<value>* and bigbnd =
            *<value>*.

        (`errno` 6)
            On entry, the bounds on *<value>* are inconsistent:
            `bl`[*<value>*] = *<value>* and `bu`[*<value>*] = *<value>*.

        (`errno` 6)
            On entry with a Warm Start, `istate`[*<value>*] = *<value>*.

        (`errno` 6)
            On entry with a Cold Start, `istate`[*<value>*] = *<value>*.

        (`errno` 6)
            On entry, the equal bounds on variable *<value>* are
            infinite, because `bl`[*<value>*] = beta and `bu`[*<value>*]
            = beta, but abs(beta) >= bigbnd: beta = *<value>* and bigbnd
            = *<value>*.

        (`errno` 6)
            On entry, the equal bounds on linear constraint *<value>*
            are infinite, because `bl`[*<value>*] = beta and
            `bu`[*<value>*] = beta, but abs(beta) >= bigbnd: beta =
            *<value>* and bigbnd = *<value>*.

        (`errno` 6)
            On entry, the equal bounds on nonlinear constraint *<value>*
            are infinite, because `bl`[*<value>*] = beta and
            `bu`[*<value>*] = beta, but abs(beta) >= bigbnd: beta =
            *<value>* and bigbnd = *<value>*.

        (`errno` 6)
            On entry, the bounds on variable *<value>* are inconsistent:
            `bl`[*<value>*] = *<value>* and `bu`[*<value>*] = *<value>*.

        (`errno` 6)
            On entry, the bounds on linear constraint *<value>* are
            inconsistent: `bl`[*<value>*] = *<value>* and
            `bu`[*<value>*] = *<value>*.

        (`errno` 6)
            On entry, the bounds on nonlinear constraint *<value>* are
            inconsistent: `bl`[*<value>*] = *<value>* and
            `bu`[*<value>*] = *<value>*.

        (`errno` 7)
            'Problem Type' not recognized. Problem abandoned.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            Iterations terminated at a dead point.

        (`errno` 1)
            Optimal solution is not unique.

        (`errno` 2)
            *<value>* solution is unbounded.

        (`errno` 3)
            No feasible point for the linear constraints.

        (`errno` 3)
            Cannot satisfy the working set constraints to the accuracy
            requested.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Gill, P E, Hammarling, S, Murray, W, Saunders, M A and Wright, M H,
    1986, `Users' guide for LSSOL (Version 1.0)`, Report SOL 86-1,
    Department of Operations Research, Stanford University

    Gill, P E and Murray, W, 1978, `Numerically stable methods for
    quadratic programming`, Math. Programming (14), 349--372

    Gill, P E, Murray, W, Saunders, M A and Wright, M H, 1984,
    `Procedures for optimization problems with a mixture of bounds and
    general linear constraints`, ACM Trans. Math. Software (10),
    282--298

    Gill, P E, Murray, W, Saunders, M A and Wright, M H, 1989, `A
    practical anti-cycling procedure for linearly constrained
    optimization`, Math. Programming (45), 437--474

    Gill, P E, Murray, W, Saunders, M A and Wright, M H, 1991,
    `Inertia-controlling methods for general quadratic programming`,
    SIAM Rev. (33), 1--36

    Gill, P E, Murray, W and Wright, M H, 1981, `Practical
    Optimization`, Academic Press

    Pardalos, P M and Schnitger, G, 1988, `Checking local optimality in
    constrained quadratic programming is NP-hard`, Operations Research
    Letters (7), 33--35
    """

    fun_name = 'naginterfaces.library.opt.qp_dense_solve'

    cb_qphess_name = fun_name + ':qphess'

    def cb_qphess(n, jthcol, h, x, hx, data=None):
        """Wrapping user's qphess. Called as base qphess."""
        if data is not None:
            _hx_qphess = qphess(
                jthcol,
                h,
                x,
                data,
            )
        else:
            _hx_qphess = qphess(
                jthcol,
                h,
                x,
            )
        hx[:] = _utils._to_ndarray(
            {'fun_name': cb_qphess_name, 'returned': 'hx'},
            _hx_qphess,
            copy=False,
            exp_shape=(n,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _bl = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'bl'},
        bl,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _bu = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'bu'},
        bu,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _cvec = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'cvec'},
        cvec,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _h = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'h'},
        h,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b_utils._check_callable(qphess, {'fun_name': fun_name, 'entity': 'qphess'}, allow_none=True)
    _qphess = None if qphess is None else cb_qphess
    _istate = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'istate'},
        istate,
        copy=True,
    )
    _nclin = _b_utils._EngineIntScalarType(
        nclin, {'fun_name': fun_name, 'entity': 'nclin'},
    ).to_ctypes()
    _nclin_py = _b_utils._EngineIntScalarType.to_py(_nclin)
    _ax = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (max(1, _nclin_py),),
    )
    _clamda = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py+_nclin_py,),
    )
    try:
        _return_scalars = _b_opt.qp_dense_solve(
            _n,
            _nclin,
            _a,
            _bl,
            _bu,
            _cvec,
            _h,
            _qphess,
            _istate,
            _x,
            _ax,
            _clamda,
            comm,
            data,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _istate_py = _istate
    _x_py = _x
    _itera_py = _return_scalars.itera
    _obj_py = _return_scalars.obj
    _ax_py = _ax
    _clamda_py = _clamda
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['istate', 'x', 'itera', 'obj', 'ax', 'clamda'],
    )(
        _istate_py,
        _x_py,
        _itera_py,
        _obj_py,
        _ax_py,
        _clamda_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def qp_dense_option_file(ioptns, comm, io_manager=None):
    u"""
    Supply option values for ``qp_dense_solve`` from external file.

    ``qp_dense_option_file`` may be used to supply options to
    ``qp_dense_solve`` from an external file.

    For full information please refer to the NAG Library document for
    e04ng

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04ngf.html

    Parameters
    ----------
    ioptns : int
        The unit number (see
        :meth:`~naginterfaces.base.utils.FileObjManager.unit_from_fileobj`)
        of the options file to be read.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``nlp1_init``.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Raises
    ------
    NagValueError
        (`errno` 1)
            `ioptns` is not in the range [0,99].

        (`errno` 1)
            On entry, `ioptns` = *<value>*.

            Constraint: 0 <= `ioptns` <= 99.

        (`errno` 2)
            ``Begin`` was found, but end-of-file was found before
            ``End`` was found.

        (`errno` 3)
            End-of-file was found before ``Begin`` was found.

        (`errno` 5)
            One or more lines of the options file is invalid.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    _b_opt.qp_dense_option_file(
        ioptns,
        comm,
        io_manager,
    )

def qp_dense_option_string(optstr, comm, io_manager=None):
    u"""
    Supply option values to ``qp_dense_solve`` from a character string.

    ``qp_dense_option_string`` may be used to supply individual options
    to ``qp_dense_solve``.

    For full information please refer to the NAG Library document for
    e04nh

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04nhf.html

    Parameters
    ----------
    optstr : str
        A single valid option string (as described in [Description] and
        in [Other Parameters for qp_dense_solve]).

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``nlp1_init``.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Raises
    ------
    NagValueError
        (`errno` 5)
            The supplied option string is invalid. Supplied value was:
            *<value>*.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    _b_opt.qp_dense_option_string(
        optstr,
        comm,
        io_manager,
    )

def qpconvex1_sparse_solve(m, iobj, ncolh, a, ha, ka, bl, bu, start, names, crname, ns, xs, istate, leniz, lenz, comm, qphx=None, data=None, io_manager=None):
    u"""
    Linear programming (LP) or convex quadratic programming (QP),
    sparse, active-set method.

    Note: this function uses optional algorithmic parameters, see also:
    ``qpconvex1_sparse_option_file``,
    ``qpconvex1_sparse_option_string``, ``nlp1_init``.

    ``qpconvex1_sparse_solve`` solves sparse linear programming or
    convex quadratic programming problems.

    For full information please refer to the NAG Library document for
    e04nk

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04nkf.html

    Parameters
    ----------
    m : int
        m, the number of general linear constraints (or slacks). This is
        the number of rows in A, including the free row (if any; see
        `iobj`).

    iobj : int
        If `iobj` > 0, row `iobj` of A is a free row containing the
        nonzero elements of the vector c appearing in the linear
        objective term c^Tx.

        If `iobj` = 0, there is no free row, i.e., the problem is either
        an FP problem (in which case `iobj` must be set to zero), or a
        QP problem with c = 0.

    ncolh : int
        n_H, the number of leading nonzero columns of the Hessian matrix
        H. For FP and LP problems, `ncolh` must be set to zero.

    a : float, array-like, shape (nnz)
        The nonzero elements of A, ordered by increasing column index.
        Note that elements with the same row and column indices are not
        allowed.

    ha : int, array-like, shape (nnz)
        `ha`[i-1] must contain the row index of the nonzero element
        stored in `a`[i-1], for i = 1,2,...,nnz. Note that the row
        indices for a column may be supplied in any order.

    ka : int, array-like, shape (n+1)
        `ka`[j-1] must contain the index in `a` of the start of the jth
        column, for j = 1,2,...,n. `ka`[n] must be set to nnz+1. To
        specify the jth column as empty, set `ka`[j-1] = `ka`[j]. As a
        consequence `ka`[0] is always 1.

    bl : float, array-like, shape (n+`m`)
        l, the lower bounds for all the variables and general
        constraints, in the following order. The first n elements of
        `bl` must contain the bounds on the variables x, and the next
        `m` elements the bounds for the general linear constraints Ax
        (or slacks s) and the free row (if any). To specify a
        nonexistent lower bound (i.e., l_j = - infinity), set `bl`[j-1]
        <= -bigbnd, where bigbnd is the value of the option 'Infinite
        Bound Size'. To specify the jth constraint as an `equality`, set
        `bl`[j-1] = `bu`[j-1] = beta, say, where |beta| < bigbnd. Note
        that the lower bound corresponding to the free row must be set
        to - infinity and stored in `bl`[n+`iobj`-1].

    bu : float, array-like, shape (n+`m`)
        u, the upper bounds for all the variables and general
        constraints, in the following order. The first n elements of
        `bu` must contain the bounds on the variables x, and the next
        `m` elements the bounds for the general linear constraints Ax
        (or slacks s) and the free row (if any). To specify a
        nonexistent upper bound (i.e., u_j = + infinity), set `bu`[j-1]
        >= bigbnd. Note that the upper bound corresponding to the free
        row must be set to + infinity and stored in `bu`[n+`iobj`-1].

    start : str, length 1
        Indicates how a starting basis is to be obtained.

        `start` = 'C'

            An internal Crash procedure will be used to choose an
            initial basis matrix B.

        `start` = 'W'

            A basis is already defined in `istate` (probably from a
            previous call).

    names : str, length 8, array-like, shape (5)
        A set of names associated with the so-called MPSX form of the
        problem, as follows:

        `names`[0]

            Must contain the name for the problem (or be blank).

        `names`[1]

            Must contain the name for the free row (or be blank).

        `names`[2]

            Must contain the name for the constraint right-hand side (or
            be blank).

        `names`[3]

            Must contain the name for the ranges (or be blank).

        `names`[4]

            Must contain the name for the bounds (or be blank).

        (These names are used in the monitoring file output; see
        [Monitoring Information].)

    crname : str, length 8, array-like, shape (nname)
        The optional column and row names, respectively.

        If nname = 1, `crname` is not referenced and the printed output
        will use default names for the columns and rows.

        If nname = n+`m`, the first n elements must contain the names
        for the columns and the next `m` elements must contain the names
        for the rows.

        Note that the name for the free row (if any) must be stored in
        `crname`[n+`iobj`-1].

    ns : int
        n_S, the number of superbasics. For QP problems, `ns` need not
        be specified if `start` = 'C', but must retain its value from a
        previous call when `start` = 'W'. For FP and LP problems, `ns`
        need not be initialized.

    xs : float, array-like, shape (n+`m`)
        The initial values of the variables and slacks (x,s). (See the
        description for `istate`.)

    istate : int, array-like, shape (n+`m`)
        If `start` = 'C', the first n elements of `istate` and `xs` must
        specify the initial states and values, respectively, of the
        variables x. (The slacks s need not be initialized.) An internal
        Crash procedure is then used to select an initial basis matrix
        B. The initial basis matrix will be triangular (neglecting
        certain small elements in each column). It is chosen from
        various rows and columns of (A -I). Possible values for
        `istate`[j-1] are as follows:

        [table omitted]

        If nothing special is known about the problem, or there is no
        wish to provide special information, you may set `istate`[j-1] =
        0 and `xs`[j-1] = 0.0, for j = 1,2,...,n.

        All variables will then be eligible for the initial basis.

        Less trivially, to say that the jth variable will probably be
        equal to one of its bounds, set `istate`[j-1] = 4 and `xs`[j-1]
        = `bl`[j-1] or `istate`[j-1] = 5 and `xs`[j-1] = `bu`[j-1] as
        appropriate.

        Following the Crash procedure, variables for which `istate`[j-1]
        = 2 are made superbasic.

        Other variables not selected for the basis are then made
        nonbasic at the value `xs`[j-1] if `bl`[j-1] <= `xs`[j-1] <=
        `bu`[j-1], or at the value `bl`[j-1] or `bu`[j-1] closest to
        `xs`[j-1].

        If `start` = 'W', `istate` and `xs` must specify the initial
        states and values, respectively, of the variables and slacks
        (x,s).

        If ``qpconvex1_sparse_solve`` has been called previously with
        the same values of n and `m`, `istate` already contains
        satisfactory information.

    leniz : int
        The dimension of the internal workspace array iz.

    lenz : int
        The dimension of the internal workspace array z.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``nlp1_init``.

    qphx : None or callable hx = qphx(nstate, x, data=None), optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        For QP problems, you must supply a version of `qphx` to compute
        the matrix product Hx.

        If H has zero rows and columns, it is most efficient to order
        the variables x = (y z)^T so that

            Hx = [table omitted](y;z) = (H_1 y;0),

        where the nonlinear variables y appear first as shown.

        For FP and LP problems, `qphx` will never be called by
        ``qpconvex1_sparse_solve`` and hence `qphx` may be None.

        Parameters
        ~~~~~~~~~~
        nstate : int
            If `nstate` = 1, ``qpconvex1_sparse_solve`` is calling
            `qphx` for the first time. This argument setting allows you
            to save computation time if certain data must be read or
            calculated only once.

            If `nstate` >= 2, ``qpconvex1_sparse_solve`` is calling
            `qphx` for the last time.

            This argument setting allows you to perform some additional
            computation on the final solution.

            In general, the last call to `qphx` is made with `nstate` =
            2+errno (see [Raises/Warns]).

            Otherwise, `nstate` = 0.

        x : float, ndarray, shape (ncolh)
            The first `ncolh` elements of the vector x.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        hx : float, array-like, shape (ncolh)
            The product Hx.

    data : arbitrary, optional
        User-communication data for callback functions.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    ns : int
        The final number of superbasics. This will be zero for FP and LP
        problems.

    xs : float, ndarray, shape (n+`m`)
        The final values of the variables and slacks (x,s).

    istate : int, ndarray, shape (n+`m`)
        The final states of the variables and slacks (x,s). The
        significance of each possible value of `istate`[j-1] is as
        follows:

        [table omitted]

        If `ninf` = 0, basic and superbasic variables may be outside
        their bounds by as much as the value of the option 'Feasibility
        Tolerance'.

        Note that unless the 'Scale Option' = 0 is specified, the option
        'Feasibility Tolerance' applies to the variables of the scaled
        problem.

        In this case, the variables of the original problem may be as
        much as 0.1 outside their bounds, but this is unlikely unless
        the problem is very badly scaled.

        Very occasionally some nonbasic variables may be outside their
        bounds by as much as the option 'Feasibility Tolerance', and
        there may be some nonbasic variables for which `xs`[j-1] lies
        strictly between its bounds.

        If `ninf` > 0, some basic and superbasic variables may be
        outside their bounds by an arbitrary amount (bounded by `sinf`
        if 'Scale Option' = 0).

    miniz : int
        The minimum value of `leniz` required to start solving the
        problem. If `errno` = 12, ``qpconvex1_sparse_solve`` may be
        called again with `leniz` suitably larger than `miniz`. (The
        bigger the better, since it is not certain how much workspace
        the basis factors need.)

    minz : int
        The minimum value of `lenz` required to start solving the
        problem. If `errno` = 13, ``qpconvex1_sparse_solve`` may be
        called again with `lenz` suitably larger than `minz`. (The
        bigger the better, since it is not certain how much workspace
        the basis factors need.)

    ninf : int
        The number of infeasibilities. This will be zero if the function
        exits successfully or `errno` = 1.

    sinf : float
        The sum of infeasibilities. This will be zero if `ninf` = 0.
        (Note that ``qpconvex1_sparse_solve`` does `not` attempt to
        compute the minimum value of `sinf` if `errno` = 3.)

    obj : float
        The value of the objective function.

        If `ninf` = 0, `obj` includes the quadratic objective term
        1/2x^THx (if any).

        If `ninf` > 0, `obj` is just the linear objective term c^Tx (if
        any).

        For FP problems, `obj` is set to zero.

    clamda : float, ndarray, shape (n+`m`)
        A set of Lagrange multipliers for the bounds on the variables
        and the general constraints. More precisely, the first n
        elements contain the multipliers (`reduced costs`) for the
        bounds on the variables, and the next `m` elements contain the
        multipliers (`shadow prices`) for the general linear
        constraints.

    Other Parameters
    ----------------
    'Check Frequency' : int
        Default = 60

        Every ith iteration after the most recent basis factorization, a
        numerical test is made to see if the current solution (x,s)
        satisfies the linear constraints Ax-s = 0.
        If the largest element of the residual vector r = Ax-s is judged
        to be too large, the current basis is refactorized and the basic
        variables recomputed to satisfy the constraints more accurately.
        If i < 0, the default value is used.
        If i = 0, the value i = 99999999 is used and effectively no
        checks are made.

    'Crash Option' : int
        Default = 2

        Note that this option does not apply when `start` = 'W' (see
        [Parameters]).

        If `start` = 'C', an internal Crash procedure is used to select
        an initial basis from various rows and columns of the constraint
        matrix (A -I).
        The value of i determines which rows and columns are initially
        eligible for the basis, and how many times the Crash procedure
        is called.
        If i = 0, the all-slack basis B = -I is chosen.
        If i = 1, the Crash procedure is called once (looking for a
        triangular basis in all rows and columns of the linear
        constraint matrix A).
        If i = 2, the Crash procedure is called twice (looking at any
        `equality` constraints first followed by any `inequality`
        constraints).
        If i < 0 or i > 2, the default value is used.

        If i = 1 or 2, certain slacks on inequality rows are selected
        for the basis first. (If i = 2, numerical values are used to
        exclude slacks that are close to a bound.) The Crash procedure
        then makes several passes through the columns of A, searching
        for a basis matrix that is essentially triangular.
        A column is assigned to 'pivot' on a particular row if the
        column contains a suitably large element in a row that has not
        yet been assigned. (The pivot elements ultimately form the
        diagonals of the triangular basis.) For remaining unassigned
        rows, slack variables are inserted to complete the basis.

    'Crash Tolerance' : float
        Default = 0.1

        This value allows the Crash procedure to ignore certain 'small'
        nonzero elements in the constraint matrix A while searching for
        a triangular basis.
        For each column of A, if a_max is the largest element in the
        column, other nonzeros in that column are ignored if they are
        less than (or equal to) a_max*r.

        When r > 0, the basis obtained by the Crash procedure may not be
        strictly triangular, but it is likely to be nonsingular and
        almost triangular.
        The intention is to obtain a starting basis with more column
        variables and fewer (arbitrary) slacks.
        A feasible solution may be reached earlier for some problems.
        If r < 0 or r >= 1, the default value is used.

    'Defaults' : valueless
        This special keyword may be used to reset all options to their
        default values.

    'Expand Frequency' : int
        Default = 10000

        This option is part of an anti-cycling procedure (see
        [Miscellaneous]) designed to allow progress even on highly
        degenerate problems.

        For LP problems, the strategy is to force a positive step at
        every iteration, at the expense of violating the constraints by
        a small amount.
        Suppose that the value of the option 'Feasibility Tolerance' is
        delta.
        Over a period of i iterations, the feasibility tolerance
        actually used by ``qpconvex1_sparse_solve`` (i.e., the `working`
        feasibility tolerance) increases from 0.5 delta to delta (in
        steps of 0.5 delta/i).

        For QP problems, the same procedure is used for iterations in
        which there is only one superbasic variable. (Cycling can only
        occur when the current solution is at a vertex of the feasible
        region.) Thus, zero steps are allowed if there is more than one
        superbasic variable, but otherwise positive steps are enforced.

        Increasing the value of i helps reduce the number of slightly
        infeasible nonbasic basic variables (most of which are
        eliminated during the resetting procedure).
        However, it also diminishes the freedom to choose a large pivot
        element (see option 'Pivot Tolerance').

        If i < 0, the default value is used.
        If i = 0, the value i = 99999999 is used and effectively no
        anti-cycling procedure is invoked.

    'Factorization Frequency' : int
        Default = 100

        If i > 0, at most i basis changes will occur between
        factorizations of the basis matrix.
        For LP problems, the basis factors are usually updated at every
        iteration.
        For QP problems, fewer basis updates will occur as the solution
        is approached.
        The number of iterations between basis factorizations will
        therefore increase.
        During these iterations a test is made regularly according to
        the value of option 'Check Frequency' to ensure that the linear
        constraints Ax-s = 0 are satisfied.
        If necessary, the basis will be refactorized before the limit of
        i updates is reached.
        If i <= 0, the default value is used.

    'Feasibility Tolerance' : float
        Default = max(10^(-6), sqrt(epsilon))

        If r >= epsilon, r defines the maximum acceptable `absolute`
        violation in each constraint at a 'feasible' point (including
        slack variables).
        For example, if the variables and the coefficients in the linear
        constraints are of order unity, and the latter are correct to
        about five decimal digits, it would be appropriate to specify r
        as 10^(-5).
        If r < epsilon, the default value is used.

        ``qpconvex1_sparse_solve`` attempts to find a feasible solution
        before optimizing the objective function.
        If the sum of infeasibilities cannot be reduced to zero, the
        problem is assumed to be `infeasible`.
        Let Sinf be the corresponding sum of infeasibilities.
        If Sinf is quite small, it may be appropriate to raise r by a
        factor of 10 or 100.
        Otherwise, some error in the data should be suspected.
        Note that the function does `not` attempt to find the minimum
        value of Sinf.

        If the constraints and variables have been scaled (see 'Scale
        Option'), then feasibility is defined in terms of the scaled
        problem (since it is more likely to be meaningful).

    'Infinite Bound Size' : float
        Default = 10^20

        If r > 0, r defines the 'infinite' bound bigbnd in the
        definition of the problem constraints.
        Any upper bound greater than or equal to bigbnd will be regarded
        as + infinity (and similarly any lower bound less than or equal
        to -bigbnd will be regarded as - infinity).
        If r <= 0, the default value is used.

    'Infinite Step Size' : float
        Default = max(bigbnd, 10^20)

        If r > 0, r specifies the magnitude of the change in variables
        that will be considered a step to an unbounded solution. (Note
        that an unbounded solution can occur only when the Hessian is
        not positive definite.) If the change in x during an iteration
        would exceed the value of r, the objective function is
        considered to be unbounded below in the feasible region.
        If r <= 0, the default value is used.

    'Iteration Limit' : int
        Default = max(50, 5*(n+m))

        The value of i specifies the maximum number of iterations
        allowed before termination.
        Setting i = 0 and 'Print Level' > 0 means that the workspace
        needed to start solving the problem will be computed and
        printed, but no iterations will be performed.
        If i < 0, the default value is used.

    'Iters' : int
        Default = max(50, 5*(n+m))

        The value of i specifies the maximum number of iterations
        allowed before termination.
        Setting i = 0 and 'Print Level' > 0 means that the workspace
        needed to start solving the problem will be computed and
        printed, but no iterations will be performed.
        If i < 0, the default value is used.

    'Itns' : int
        Default = max(50, 5*(n+m))

        The value of i specifies the maximum number of iterations
        allowed before termination.
        Setting i = 0 and 'Print Level' > 0 means that the workspace
        needed to start solving the problem will be computed and
        printed, but no iterations will be performed.
        If i < 0, the default value is used.

    'List' : valueless
        Option 'List' enables printing of each option specification as
        it is supplied. 'Nolist' suppresses this printing.

    'Nolist' : valueless
        Default for ``qpconvex1_sparse_solve`` = 'Nolist'

        Option 'List' enables printing of each option specification as
        it is supplied. 'Nolist' suppresses this printing.

    'LU Factor Tolerance' : float
        Default = 100.0

        The values of r_1 and r_2 affect the stability and sparsity of
        the basis factorization B = LU, during refactorization and
        updates respectively.
        The lower triangular matrix L is a product of matrices of the
        form

            [table omitted]

        where the multipliers mu will satisfy |mu| <= r_i.
        The default values of r_1 and r_2 usually strike a good
        compromise between stability and sparsity.
        For large and relatively dense problems, setting r_1 and r_2 to
        25 (say) may give a marked improvement in sparsity without
        impairing stability to a serious degree.

        Note that for band matrices it may be necessary to set r_1 in
        the range 1 <= r_1 < 2 in order to achieve stability.
        If r_1 < 1 or r_2 < 1, the default value is used.

    'LU Update Tolerance' : float
        Default = 10.0

        The values of r_1 and r_2 affect the stability and sparsity of
        the basis factorization B = LU, during refactorization and
        updates respectively.
        The lower triangular matrix L is a product of matrices of the
        form

            [table omitted]

        where the multipliers mu will satisfy |mu| <= r_i.
        The default values of r_1 and r_2 usually strike a good
        compromise between stability and sparsity.
        For large and relatively dense problems, setting r_1 and r_2 to
        25 (say) may give a marked improvement in sparsity without
        impairing stability to a serious degree.

        Note that for band matrices it may be necessary to set r_1 in
        the range 1 <= r_1 < 2 in order to achieve stability.
        If r_1 < 1 or r_2 < 1, the default value is used.

    'LU Singularity Tolerance' : float
        Default = epsilon^0.67

        If r > 0, r defines the singularity tolerance used to guard
        against ill-conditioned basis matrices.
        Whenever the basis is refactorized, the diagonal elements of U
        are tested as follows.
        If |u_{jj}| <= r or |u_{jj}| < r*max_i|u_{ij}|, the jth column
        of the basis is replaced by the corresponding slack variable.
        If r <= 0, the default value is used.

    'Minimize' : valueless
        Default

        This option specifies the required direction of the
        optimization.
        It applies to both linear and nonlinear terms (if any) in the
        objective function.
        Note that if two problems are the same except that one minimizes
        f(x) and the other maximizes -f(x), their solutions will be the
        same but the signs of the dual variables pi_i and the reduced
        gradients d_j (see [Main Iteration]) will be reversed.

    'Maximize' : valueless
        This option specifies the required direction of the
        optimization.
        It applies to both linear and nonlinear terms (if any) in the
        objective function.
        Note that if two problems are the same except that one minimizes
        f(x) and the other maximizes -f(x), their solutions will be the
        same but the signs of the dual variables pi_i and the reduced
        gradients d_j (see [Main Iteration]) will be reversed.

    'Monitoring File' : int
        Default = -1

        If i >= 0 and 'Print Level' > 0 (see 'Print Level'), monitoring
        information produced by ``qpconvex1_sparse_solve`` is sent to a
        file with logical unit number i.
        If i < 0 and/or 'Print Level' = 0, the default value is used and
        hence no monitoring information is produced.

    'Optimality Tolerance' : float
        Default = max(10^(-6), sqrt(epsilon))

        If r >= epsilon, r is used to judge the size of the reduced
        gradients d_j = g_j-pi^Ta_j.
        By definition, the reduced gradients for basic variables are
        always zero.
        Optimality is declared if the reduced gradients for any nonbasic
        variables at their lower or upper bounds satisfy -r*max(1,
        ||pi||) <= d_j <= r*max(1, ||pi||), and if |d_j| <= r*max(1,
        ||pi||) for any superbasic variables.
        If r < epsilon, the default value is used.

    'Partial Price' : int
        Default = 10

        Note that this option does not apply to QP problems.

        This option is recommended for large FP or LP problems that have
        significantly more variables than constraints (i.e., n≫m).
        It reduces the work required for each pricing operation (i.e.,
        when a nonbasic variable is selected to enter the basis).
        If i = 1, all columns of the constraint matrix (A -I) are
        searched.
        If i > 1, A and -I are partitioned to give i roughly equal
        segments A_j,K_j, for j = 1,2,...,p (modulo p).
        If the previous pricing search was successful on
        A_{j-1},K_{j-1}, the next search begins on the segments A_j,K_j.
        If a reduced gradient is found that is larger than some dynamic
        tolerance, the variable with the largest such reduced gradient
        (of appropriate sign) is selected to enter the basis.
        If nothing is found, the search continues on the next segments
        A_{j+1},K_{j+1}, and so on.
        If i <= 0, the default value is used.

    'Pivot Tolerance' : float
        Default = epsilon^0.67

        If r > 0, r is used to prevent columns entering the basis if
        they would cause the basis to become almost singular.
        If r <= 0, the default value is used.

    'Print Level' : int
        The value of i controls the amount of printout produced by
        ``qpconvex1_sparse_solve``, as indicated below.
        A detailed description of the printed output is given in
        [Further Comments] (summary output at each iteration and the
        final solution) and [Monitoring Information] (monitoring
        information at each iteration).
        Note that the summary output will not exceed 80 characters per
        line and that the monitoring information will not exceed 120
        characters per line.
        If i < 0, the default value is used.

        The following printout is sent to the file object associated
        with the advisory I/O unit (see
        :class:`~naginterfaces.base.utils.FileObjManager`):

        [table omitted]

        The following printout is sent to the logical unit number
        defined by the option 'Monitoring File':

        [table omitted]

        If 'Print Level' > 0 and the unit number defined by option
        'Monitoring File' is the advisory unit number, then the summary
        output is suppressed.

    'Rank Tolerance' : float
        Default = 100 epsilon

        See above.

    'Scale Option' : int
        Default = 2

        This option enables you to scale the variables and constraints
        using an iterative procedure due to Fourer (1982), which
        attempts to compute row scales r_i and column scales c_j such
        that the scaled matrix coefficients a__{ij} = a_{ij}*(c_j/r_i)
        are as close as possible to unity.
        This may improve the overall efficiency on some problems. (The
        lower and upper bounds on the variables and slacks for the
        scaled problem are redefined as l__j = l_j/c_j and u__j =
        u_j/c_j respectively, where c_j = r_{j-n} if j > n.)

        If i = 0, no scaling is performed.
        If i = 1, all rows and columns of the constraint matrix A are
        scaled.
        If i = 2, an additional scaling is performed that may be helpful
        when the solution x is large; it takes into account columns of
        (A -I) that are fixed or have positive lower bounds or negative
        upper bounds.
        If i < 0 or i > 2, the default value is used.

    'Scale Tolerance' : float
        Default = 0.9

        Note that this option does not apply when 'Scale Option' = 0.

        If 0 < r < 1, r is used to control the number of scaling passes
        to be made through the constraint matrix A.
        At least 3 (and at most 10) passes will be made.
        More precisely, let a_p denote the largest column ratio (i.e.,
        'biggest' element/'smallest' element in some sense) after the
        pth scaling pass through A.
        The scaling procedure is terminated if a_p >= a_{p-1}*r for some
        p >= 3.
        Thus, increasing the value of r from 0.9 to 0.99 (say) will
        probably increase the number of passes through A.

        If r <= 0 or r >= 1, the default value is used.

    'Superbasics Limit' : int
        Default = min(n_H+1, n)

        Note that this option does not apply to FP or LP problems.

        The value of i specifies 'how nonlinear' you expect the QP
        problem to be.
        If i <= 0, the default value is used.

    Raises
    ------
    NagValueError
        (`errno` 4)
            Too many iterations.

        (`errno` 5)
            Reduced Hessian matrix Z^T*H*Z exceeds its assigned
            dimension.

        (`errno` 6)
            Hessian matrix H appears to be indefinite.

        (`errno` 7)
            On entry, `start` = *<value>* and `istate`[*<value>*] =
            *<value>*.

            Constraint: if `start` = *<value>* then `istate`[i] <= 3 for
            all i.

        (`errno` 7)
            On entry, `start` = *<value>* and `istate`[*<value>*] =
            *<value>*.

            Constraint: if `start` = *<value>* then `istate`[i] <= 5 for
            all i.

        (`errno` 7)
            On entry, `start` = *<value>* and `istate`[*<value>*] =
            *<value>*.

            Constraint: if `start` = *<value>* then `istate`[i] >= 0 for
            all i.

        (`errno` 7)
            On entry, the bounds on *<value>* are inconsistent:
            `bl`[*<value>*] = *<value>* and `bu`[*<value>*] = *<value>*.

        (`errno` 7)
            On entry, the equal bounds on *<value>* are infinite,
            because `bl`[*<value>*] = beta and `bu`[*<value>*] = beta,
            but abs(beta) >= bigbnd: beta = *<value>* and bigbnd =
            *<value>*.

        (`errno` 7)
            On entry, duplicate element found in row *<value>*, column
            *<value>*.

        (`errno` 7)
            On entry, `lenz` = *<value>*.

            Constraint: `lenz` >= 1.

        (`errno` 7)
            On entry, `leniz` = *<value>*.

            Constraint: `leniz` >= 1.

        (`errno` 7)
            On entry, `m` = *<value>*, `ka`[j] = *<value>* and `ka`[j-1]
            = *<value>*, for j = *<value>*.

            Constraint: 0 <= `ka`[j+1]-`ka`[j] <= `m` for all j.

        (`errno` 7)
            On entry, `ka`[*<value>*] = *<value>*.

            Constraint: `ka`[j] >= 1 for all j.

        (`errno` 7)
            On entry, n+1 = *<value>*, `ka`[n] = *<value>* and nnz+1 =
            *<value>*.

            Constraint: `ka`[n] = nnz+1.

        (`errno` 7)
            On entry, `ka`[0] = *<value>*.

            Constraint: `ka`[0] = 1.

        (`errno` 7)
            On entry, `m` = *<value>* and `ha`[*<value>*] = *<value>*.

            Constraint: 1 <= `ha`[j] <= `m` for all j.

        (`errno` 7)
            On entry, nname = *<value>*, n = *<value>* and `m` =
            *<value>*.

            Constraint: nname = 1 or nname = n+`m`.

        (`errno` 7)
            On entry, n = *<value>*, `iobj` = *<value>*,
            `bu`[n+`iobj`-1] = *<value>* and bigbnd = *<value>*.

            Constraint: `bu`[n+`iobj`-1] >= bigbnd.

        (`errno` 7)
            On entry, n = *<value>*, `iobj` = *<value>*,
            `bl`[n+`iobj`-1] = *<value>* and bigbnd = *<value>*.

            Constraint: `bl`[n+`iobj`-1] <= -bigbnd.

        (`errno` 7)
            On entry, `iobj` = *<value>* and `m` = *<value>*.

            Constraint: 0 <= `iobj` <= `m`.

        (`errno` 7)
            On entry, `ncolh` = *<value>* and n = *<value>*.

            Constraint: 0 <= `ncolh` <= n.

        (`errno` 7)
            On entry, nnz = *<value>*, n = *<value>* and `m` =
            *<value>*.

            Constraint: 1 <= nnz <= n*`m`.

        (`errno` 7)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 7)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 7)
            On entry, `start` = *<value>*.

            Constraint: `start` = 'C' or 'W'.

        (`errno` 8)
            Cannot satisfy the general constraints.

        (`errno` 9)
            Not enough integer workspace for the basis factors.

        (`errno` 10)
            Not enough real workspace for the basis factors.

        (`errno` 11)
            The basis is singular after 15 factorization attempts.

        (`errno` 21)
            Error in basis package. Please contact NAG.

        (`errno` 32)
            System error. Wrong number of basic variables. Please
            contact NAG.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            Weak *<value>* solution found.

        (`errno` 2)
            *<value>* problem is unbounded (or badly scaled).

        (`errno` 3)
            *<value>* problem is infeasible.

        (`errno` 12)
            Not enough integer workspace to start solving the problem.

        (`errno` 13)
            Not enough real workspace to start solving the problem.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Fourer, R, 1982, `Solving staircase linear programs by the simplex
    method`, Math. Programming (23), 274--313

    Gill, P E and Murray, W, 1978, `Numerically stable methods for
    quadratic programming`, Math. Programming (14), 349--372

    Gill, P E, Murray, W and Saunders, M A, 2002, `SNOPT: An SQP
    Algorithm for Large-scale Constrained Optimization` (12), 979--1006,
    SIAM J. Optim.

    Gill, P E, Murray, W, Saunders, M A and Wright, M H, 1987,
    `Maintaining LU factors of a general sparse matrix`, Linear Algebra
    and its Applics. (88/89), 239--270

    Gill, P E, Murray, W, Saunders, M A and Wright, M H, 1989, `A
    practical anti-cycling procedure for linearly constrained
    optimization`, Math. Programming (45), 437--474

    Gill, P E, Murray, W, Saunders, M A and Wright, M H, 1991,
    `Inertia-controlling methods for general quadratic programming`,
    SIAM Rev. (33), 1--36

    Hall, J A J and McKinnon, K I M, 1996, `The simplest examples where
    the simplex method cycles and conditions where EXPAND fails to
    prevent cycling`, Report MS, 96--100, Department of Mathematics and
    Statistics, University of Edinburgh

    Murtagh, B A and Saunders, M A, 1995, `MINOS 5.4 users' guide`,
    Report SOL 83-20R, Department of Operations Research, Stanford
    University
    """

    fun_name = 'naginterfaces.library.opt.qpconvex1_sparse_solve'

    cb_qphx_name = fun_name + ':qphx'

    def cb_qphx(nstate, ncolh, x, hx, data=None):
        """Wrapping user's qphx. Called as base qphx."""
        if data is not None:
            _hx_qphx = qphx(
                nstate,
                x,
                data,
            )
        else:
            _hx_qphx = qphx(
                nstate,
                x,
            )
        hx[:] = _utils._to_ndarray(
            {'fun_name': cb_qphx_name, 'returned': 'hx'},
            _hx_qphx,
            copy=False,
            exp_shape=(ncolh,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )

    _ka = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ka'},
        ka,
        copy=False,
    )
    _n = _b_utils._EngineIntScalarType(
        _b_utils._EngineIntArrayType.get_shape(_ka, {'fun_name': fun_name, 'entity': 'ka'}, exp_rank=1)[0]-1, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nnz_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=1)[0]
    _nnz = _b_utils._EngineIntScalarType(
        _nnz_py, {'fun_name': fun_name, 'entity': '_nnz'},
    ).to_ctypes()
    _b_utils._check_callable(qphx, {'fun_name': fun_name, 'entity': 'qphx'}, allow_none=True)
    _qphx = None if qphx is None else cb_qphx
    _ha = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ha'},
        ha,
        copy=False,
    )
    _bl = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'bl'},
        bl,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _bu = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'bu'},
        bu,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _names = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'names'},
        names,
        copy=False,
    )
    _crname = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'crname'},
        crname,
        copy=False,
    )
    _nname_py = _b_utils._EngineCharArrayType.get_shape(_crname, {'fun_name': fun_name, 'entity': 'crname'}, exp_rank=1)[0]
    _nname = _b_utils._EngineIntScalarType(
        _nname_py, {'fun_name': fun_name, 'entity': '_nname'},
    ).to_ctypes()
    _xs = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xs'},
        xs,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _istate = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'istate'},
        istate,
        copy=True,
    )
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _m = _b_utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = _b_utils._EngineIntScalarType.to_py(_m)
    _clamda = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py+_m_py,),
    )
    try:
        _return_scalars = _b_opt.qpconvex1_sparse_solve(
            _n,
            _m,
            _nnz,
            iobj,
            ncolh,
            _qphx,
            _a,
            _ha,
            _ka,
            _bl,
            _bu,
            start,
            _names,
            _nname,
            _crname,
            ns,
            _xs,
            _istate,
            _clamda,
            leniz,
            lenz,
            comm,
            data,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _ns_py = _return_scalars.ns
    _xs_py = _xs
    _istate_py = _istate
    _miniz_py = _return_scalars.miniz
    _minz_py = _return_scalars.minz
    _ninf_py = _return_scalars.ninf
    _sinf_py = _return_scalars.sinf
    _obj_py = _return_scalars.obj
    _clamda_py = _clamda
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['ns', 'xs', 'istate', 'miniz', 'minz', 'ninf', 'sinf', 'obj', 'clamda'],
    )(
        _ns_py,
        _xs_py,
        _istate_py,
        _miniz_py,
        _minz_py,
        _ninf_py,
        _sinf_py,
        _obj_py,
        _clamda_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def qpconvex1_sparse_option_file(ioptns, comm, io_manager=None):
    u"""
    Supply option values for ``qpconvex1_sparse_solve`` from external
    file.

    ``qpconvex1_sparse_option_file`` may be used to supply options to
    ``qpconvex1_sparse_solve`` from an external file.

    For full information please refer to the NAG Library document for
    e04nl

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04nlf.html

    Parameters
    ----------
    ioptns : int
        The unit number (see
        :meth:`~naginterfaces.base.utils.FileObjManager.unit_from_fileobj`)
        of the options file to be read.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``nlp1_init``.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Raises
    ------
    NagValueError
        (`errno` 1)
            `ioptns` is not in the range [0,99].

        (`errno` 1)
            On entry, `ioptns` = *<value>*.

            Constraint: 0 <= `ioptns` <= 99.

        (`errno` 2)
            ``Begin`` was found, but end-of-file was found before
            ``End`` was found.

        (`errno` 3)
            End-of-file was found before ``Begin`` was found.

        (`errno` 5)
            One or more lines of the options file is invalid.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    _b_opt.qpconvex1_sparse_option_file(
        ioptns,
        comm,
        io_manager,
    )

def qpconvex1_sparse_option_string(optstr, comm, io_manager=None):
    u"""
    Supply option values to ``qpconvex1_sparse_solve`` from a character
    string.

    ``qpconvex1_sparse_option_string`` may be used to supply individual
    options to ``qpconvex1_sparse_solve``.

    For full information please refer to the NAG Library document for
    e04nm

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04nmf.html

    Parameters
    ----------
    optstr : str
        A single valid option string (as described in [Description] and
        in [Other Parameters for qpconvex1_sparse_solve]).

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``nlp1_init``.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Raises
    ------
    NagValueError
        (`errno` 5)
            The supplied option string is invalid. Supplied value was:
            *<value>*.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    _b_opt.qpconvex1_sparse_option_string(
        optstr,
        comm,
        io_manager,
    )

def qpconvex2_sparse_init(io_manager=None):
    u"""
    Initialization function for ``qpconvex2_sparse_solve``.

    ``qpconvex2_sparse_init`` is used to initialize the function
    ``qpconvex2_sparse_solve``.

    For full information please refer to the NAG Library document for
    e04np

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04npf.html

    Parameters
    ----------
    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    comm : dict, communication object
        Communication structure.

    See Also
    --------
    :meth:`naginterfaces.library.examples.opt.qpconvex2_sparse_solve_ex.main`
    """

    _return_scalar = _b_opt.qpconvex2_sparse_init(
        io_manager,
    )
    _comm_py = _return_scalar
    return _comm_py

def qpconvex2_sparse_solve(start, m, ncolh, iobj, objadd, prob, acol, inda, loca, bl, bu, names, helast, hs, x, ns, comm, qphx=None, c=None, data=None, io_manager=None):
    u"""
    Linear programming (LP) or convex quadratic programming (QP),
    sparse, active-set method, recommended.

    Note: this function uses optional algorithmic parameters, see also:
    ``qpconvex2_sparse_option_file``,
    ``qpconvex2_sparse_option_string``,
    ``qpconvex2_sparse_option_integer_set``,
    ``qpconvex2_sparse_option_double_set``, ``qpconvex2_sparse_init``,
    ``qpconvex2_sparse_option_integer_get``,
    ``qpconvex2_sparse_option_double_get``.

    ``qpconvex2_sparse_solve`` solves sparse linear programming or
    convex quadratic programming problems.
    The initialization function ``qpconvex2_sparse_init`` **must** have
    been called before calling ``qpconvex2_sparse_solve``.

    For full information please refer to the NAG Library document for
    e04nq

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04nqf.html

    Parameters
    ----------
    start : str, length 1
        Indicates how a starting basis (and certain other items) will be
        obtained.

        `start` = 'C'

            Requests that an internal Crash procedure be used to choose
            an initial basis, unless a Basis file is provided via
            options 'Old Basis File', 'Insert File' or 'Load File'.

        `start` = 'B'

            Is the same as `start` = 'C' but is more meaningful when a
            Basis file is given.

        `start` = 'W'

            Means that a basis is already defined in `hs` and a start
            point is already defined in `x` (probably from an earlier
            call).

    m : int
        m, the number of general linear constraints (or slacks). This is
        the number of rows in the linear constraint matrix A, including
        the free row (if any; see `iobj`). Note that A must have at
        least one row. If your problem has no constraints, or only upper
        or lower bounds on the variables, then you must include a dummy
        row with sufficiently wide upper and lower bounds (see also
        `acol`, `inda` and `loca`).

    ncolh : int
        n_H, the number of leading nonzero columns of the Hessian matrix
        H. For FP and LP problems, `ncolh` must be set to zero.

        The first `ncolh` elements of x belong to variables
        corresponding to the nonzero block of the QP Hessian.

    iobj : int
        If `iobj` > 0, row `iobj` of A is a free row containing the
        nonzero elements of the vector c appearing in the linear
        objective term c^Tx.

        If `iobj` = 0, there is no free row, and the linear objective
        vector should be supplied in array `c`.

    objadd : float
        The constant q, to be added to the objective for printing
        purposes. Typically `objadd` = 0.0e0.

    prob : str, length 8
        The name for the problem. It is used in the printed solution and
        in some functions that output Basis files. A blank name may be
        used.

    acol : float, array-like, shape (ne)
        The nonzero elements of A, ordered by increasing column index.
        Note that all elements must be assigned a value in the calling
        program.

    inda : int, array-like, shape (ne)
        `inda`[i-1] must contain the row index of the nonzero element
        stored in `acol`[i-1], for i = 1,2,...,ne. Thus a pair of values
        (`acol`[i-1],`inda`[i-1]) contains a matrix element and its
        corresponding row index.

        Note that the row indices for a column may be supplied in any
        order.

    loca : int, array-like, shape (n+1)
        `loca`[j-1] must contain the index in `acol` and `inda` of the
        start of the jth column, for j = 1,2,...,n. Thus for j = 1:n,
        the entries of column j are held in `acol`[k-1:l] and their
        corresponding row indices are in `inda`[k-1:l], where k =
        `loca`[j-1] and l = `loca`[j]-1. To specify the jth column as
        empty, set `loca`[j-1] = `loca`[j]. Note that the first and last
        elements of `loca` must be `loca`[0] = 1 and `loca`[n] = ne+1.
        If your problem has no constraints, or just bounds on the
        variables, you may include a dummy 'free' row with a single
        (zero) element by setting ne = 1, `acol`[0] = 0.0, `inda`[0] =
        1, `loca`[0] = 1, and `loca`[j-1] = 2, for j = 2:n+1. This row
        is made 'free' by setting its bounds to be `bl`[n] = -bigbnd and
        `bu`[n] = bigbnd, where bigbnd is the value of the option
        'Infinite Bound Size'.

    bl : float, array-like, shape (n+`m`)
        l, the lower bounds for all the variables and general
        constraints, in the following order. The first n elements of
        `bl` must contain the bounds on the variables x, and the next
        `m` elements the bounds for the general linear constraints Ax
        (which, equivalently, are the bounds for the slacks, s) and the
        free row (if any). To fix the jth variable, set `bl`[j-1] =
        `bu`[j-1] = beta, say, where |beta| < bigbnd. To specify a
        nonexistent lower bound (i.e., l_j = - infinity), set `bl`[j-1]
        <= -bigbnd. Here, bigbnd is the value of the option 'Infinite
        Bound Size'. To specify the jth constraint as an `equality`, set
        `bl`[n+j-1] = `bu`[n+j-1] = beta, say, where |beta| < bigbnd.
        Note that the lower bound corresponding to the free row must be
        set to - infinity and stored in `bl`[n+`iobj`-1].

    bu : float, array-like, shape (n+`m`)
        u, the upper bounds for all the variables and general
        constraints, in the following order. The first n elements of
        `bu` must contain the bounds on the variables x, and the next
        `m` elements the bounds for the general linear constraints Ax
        (which, equivalently, are the bounds for the slacks, s) and the
        free row (if any). To specify a nonexistent upper bound (i.e.,
        u_j = + infinity), set `bu`[j-1] >= bigbnd. Note that the upper
        bound corresponding to the free row must be set to + infinity
        and stored in `bu`[n+`iobj`-1].

    names : str, length 8, array-like, shape (nname)
        The optional column and row names, respectively.

        If nname = 1, `names` is not referenced and the printed output
        will use default names for the columns and rows.

        If nname = n+`m`, the first n elements must contain the names
        for the columns and the next `m` elements must contain the names
        for the rows.

        Note that the name for the free row (if any) must be stored in
        `names`[n+`iobj`-1].

    helast : int, array-like, shape (n+`m`)
        Defines which variables are to be treated as being elastic in
        elastic mode. The allowed values of `helast` are:

        [table omitted]

        `helast` need not be assigned if option 'Elastic Mode' = 0.

    hs : int, array-like, shape (n+`m`)
        If `start` = 'C' or 'B', and a Basis file of some sort is to be
        input (see the description of the options 'Old Basis File',
        'Insert File' or 'Load File'), then `hs` and `x` need not be set
        at all.

        If `start` = 'C' or 'B' and there is no Basis file, the first n
        elements of `hs` and `x` must specify the initial states and
        values, respectively, of the variables x. (The slacks s need not
        be initialized.) An internal Crash procedure is then used to
        select an initial basis matrix B.

        The initial basis matrix will be triangular (neglecting certain
        small elements in each column).

        It is chosen from various rows and columns of (A -I).

        Possible values for `hs`[j-1] are as follows:

        [table omitted]

        If nothing special is known about the problem, or there is no
        wish to provide special information, you may set `hs`[j-1] = 0
        and `x`[j-1] = 0.0, for j = 1,2,...,n.

        All variables will then be eligible for the initial basis.

        Less trivially, to say that the jth variable will probably be
        equal to one of its bounds, set `hs`[j-1] = 4 and `x`[j-1] =
        `bl`[j-1] or `hs`[j-1] = 5 and `x`[j-1] = `bu`[j-1] as
        appropriate.

        Following the Crash procedure, variables for which `hs`[j-1] = 2
        are made superbasic.

        Other variables not selected for the basis are then made
        nonbasic at the value `x`[j-1] if `bl`[j-1] <= `x`[j-1] <=
        `bu`[j-1], or at the value `bl`[j-1] or `bu`[j-1] closest to
        `x`[j-1].

        If `start` = 'W', `hs` and `x` must specify the initial states
        and values, respectively, of the variables and slacks (x,s).

        If ``qpconvex2_sparse_solve`` has been called previously with
        the same values of n and `m`, `hs` already contains satisfactory
        information.

    x : float, array-like, shape (n+`m`)
        The initial values of the variables x, and, if `start` = 'W',
        the slacks s, i.e., (x,s). (See the description for argument
        `hs`.)

    ns : int
        n_S, the number of superbasics. For QP problems, `ns` need not
        be specified if `start` = 'C', but must retain its value from a
        previous call when `start` = 'W'. For FP and LP problems, `ns`
        need not be initialized.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``qpconvex2_sparse_init``.

    qphx : None or callable hx = qphx(x, nstate, data=None), optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        For QP problems, you must supply a version of `qphx` to compute
        the matrix product Hx for a given vector x.

        If H has rows and columns of zeros, it is most efficient to
        order x so that the nonlinear variables appear first.

        For example, if x = (y,z)^T and only y enters the objective
        quadratically, then

            Hx = [table omitted](y;z) = (H_1 y;0).

        In this case, `ncolh` should be the dimension of y, and `qphx`
        should compute H_1 y.

        For FP and LP problems, `qphx` will never be called by
        ``qpconvex2_sparse_solve`` and hence `qphx` may be None.

        Parameters
        ~~~~~~~~~~
        x : float, ndarray, shape (ncolh)
            The first `ncolh` elements of the vector x.

        nstate : int
            Allows you to save computation time if certain data must be
            read or calculated only once. To preserve this data for a
            subsequent calculation place it in one of `cuser`, `ruser`
            or `iuser`.

            `nstate` = 1

                ``qpconvex2_sparse_solve`` is calling `qphx` for the
                first time.

            `nstate` = 0

                There is nothing special about the current call of
                `qphx`.

            `nstate` >= 2

                ``qpconvex2_sparse_solve`` is calling `qphx` for the
                last time. This argument setting allows you to perform
                some additional computation on the final solution.

                `nstate` = 2

                    The current x is optimal.

                `nstate` = 3

                    The problem appears to be infeasible.

                `nstate` = 4

                    The problem appears to be unbounded.

                `nstate` = 5

                    The iterations limit was reached.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        hx : float, array-like, shape (ncolh)
            The product Hx. If `ncolh` is less than the input argument
            n, Hx is really the product H_1 y in [equation].

    c : None or float, array-like, shape (max(1, lenc)), optional
        Contains the explicit objective vector c (if any). If the
        problem is of type FP, or if lenc = 0, `c` is not referenced.
        (In that case, `c` may be dimensioned (1), or it could be any
        convenient array.)

    data : arbitrary, optional
        User-communication data for callback functions.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    hs : int, ndarray, shape (n+`m`)
        The final states of the variables and slacks (x,s). The
        significance of each possible value of `hs`[j-1] is as follows:

        [table omitted]

        If `ninf` = 0, basic and superbasic variables may be outside
        their bounds by as much as the value of the option 'Feasibility
        Tolerance'.

        Note that unless the option 'Scale Option' = 0 is specified, the
        option 'Feasibility Tolerance' applies to the variables of the
        scaled problem.

        In this case, the variables of the original problem may be as
        much as 0.1 outside their bounds, but this is unlikely unless
        the problem is very badly scaled.

        Very occasionally some nonbasic variables may be outside their
        bounds by as much as the option 'Feasibility Tolerance', and
        there may be some nonbasic variables for which `x`[j-1] lies
        strictly between its bounds.

        If `ninf` > 0, some basic and superbasic variables may be
        outside their bounds by an arbitrary amount (bounded by `sinf`
        if 'Scale Option' = 0).

    x : float, ndarray, shape (n+`m`)
        The final values of the variables and slacks (x,s).

    pi : float, ndarray, shape (`m`)
        Contains the dual variables pi (a set of Lagrange multipliers
        (shadow prices) for the general constraints).

    rc : float, ndarray, shape (n+`m`)
        Contains the reduced costs, g-(A -I)^T pi. The vector g is the
        gradient of the objective if `x` is feasible, otherwise it is
        the gradient of the Phase 1 objective. In the former case, g(i)
        = 0, for i = n+1:`m`, hence `rc`[n+1:`m`-1] = pi.

    ns : int
        The final number of superbasics. This will be zero for FP and LP
        problems.

    ninf : int
        The number of infeasibilities.

    sinf : float
        The sum of the scaled infeasibilities. This will be zero if
        `ninf` = 0, and is most meaningful when 'Scale Option' = 0.

    obj : float
        The value of the objective function.

        If `ninf` = 0, `obj` includes the quadratic objective term
        1/2x^THx (if any).

        If `ninf` > 0, `obj` is just the linear objective term c^Tx (if
        any).

        For FP problems, `obj` is set to zero.

        Note that `obj` does not include contributions from the constant
        term `objadd` or the objective row, if any.

    Other Parameters
    ----------------
    'Check Frequency' : int
        Default = 60

        Every ith iteration after the most recent basis factorization, a
        numerical test is made to see if the current solution (x,s)
        satisfies the linear constraints Ax-s = 0.
        If the largest element of the residual vector r = Ax-s is judged
        to be too large, the current basis is refactorized and the basic
        variables recomputed to satisfy the constraints more accurately.
        If i <= 0, the value i = 99999999 is used and effectively no
        checks are made.

        'Check Frequency' = 1 is useful for debugging purposes, but
        otherwise this option should not be needed.

    'Crash Option' : int
        Default = 3

        Note that these options do not apply when `start` = 'W' (see
        [Parameters]).

        If `start` = 'C', an internal Crash procedure is used to select
        an initial basis from various rows and columns of the constraint
        matrix (A -I).
        The value of i determines which rows and columns of A are
        initially eligible for the basis, and how many times the Crash
        procedure is called.
        Columns of -I are used to pad the basis where necessary.

        [table omitted]

        If i >= 1, certain slacks on inequality rows are selected for
        the basis first. (If i >= 2, numerical values are used to
        exclude slacks that are close to a bound.) The Crash procedure
        then makes several passes through the columns of A, searching
        for a basis matrix that is essentially triangular.
        A column is assigned to 'pivot' on a particular row if the
        column contains a suitably large element in a row that has not
        yet been assigned. (The pivot elements ultimately form the
        diagonals of the triangular basis.) For remaining unassigned
        rows, slack variables are inserted to complete the basis.

        The 'Crash Tolerance' allows the Crash procedure to ignore
        certain 'small' nonzero elements in each column of A.
        If a_max is the largest element in column j, other nonzeros
        a_{ij} in the column are ignored if |a_{ij}| <= a_max*r. (To be
        meaningful, r should be in the range 0 <= r < 1.)

        When r > 0.0, the basis obtained by the Crash procedure may not
        be strictly triangular, but it is likely to be nonsingular and
        almost triangular.
        The intention is to obtain a starting basis containing more
        columns of A and fewer (arbitrary) slacks.
        A feasible solution may be reached sooner on some problems.

        For example, suppose the first m columns of A form the matrix
        shown under 'LU Factor Tolerance'; i.e., a tridiagonal matrix
        with entries -1, 4, -1.
        To help the Crash procedure choose all m columns for the initial
        basis, we would specify a 'Crash Tolerance' of r for some value
        of r > 0.5.

    'Crash Tolerance' : float
        Default = 0.1

        Note that these options do not apply when `start` = 'W' (see
        [Parameters]).

        If `start` = 'C', an internal Crash procedure is used to select
        an initial basis from various rows and columns of the constraint
        matrix (A -I).
        The value of i determines which rows and columns of A are
        initially eligible for the basis, and how many times the Crash
        procedure is called.
        Columns of -I are used to pad the basis where necessary.

        [table omitted]

        If i >= 1, certain slacks on inequality rows are selected for
        the basis first. (If i >= 2, numerical values are used to
        exclude slacks that are close to a bound.) The Crash procedure
        then makes several passes through the columns of A, searching
        for a basis matrix that is essentially triangular.
        A column is assigned to 'pivot' on a particular row if the
        column contains a suitably large element in a row that has not
        yet been assigned. (The pivot elements ultimately form the
        diagonals of the triangular basis.) For remaining unassigned
        rows, slack variables are inserted to complete the basis.

        The 'Crash Tolerance' allows the Crash procedure to ignore
        certain 'small' nonzero elements in each column of A.
        If a_max is the largest element in column j, other nonzeros
        a_{ij} in the column are ignored if |a_{ij}| <= a_max*r. (To be
        meaningful, r should be in the range 0 <= r < 1.)

        When r > 0.0, the basis obtained by the Crash procedure may not
        be strictly triangular, but it is likely to be nonsingular and
        almost triangular.
        The intention is to obtain a starting basis containing more
        columns of A and fewer (arbitrary) slacks.
        A feasible solution may be reached sooner on some problems.

        For example, suppose the first m columns of A form the matrix
        shown under 'LU Factor Tolerance'; i.e., a tridiagonal matrix
        with entries -1, 4, -1.
        To help the Crash procedure choose all m columns for the initial
        basis, we would specify a 'Crash Tolerance' of r for some value
        of r > 0.5.

    'Defaults' : valueless
        This special keyword may be used to reset all options to their
        default values.

    'Dump File' : int
        Default = 0

        Options 'Dump File' and 'Load File' are similar to options
        'Punch File' and 'Insert File', but they record solution
        information in a manner that is more direct and more easily
        modified.
        A full description of information recorded in options 'Dump
        File' and 'Load File' is given in Gill `et al.` (2005a).

        If i_1 > 0, the last solution obtained will be output to the
        file with unit number i.

        If i_2 > 0, the 'Load File' containing basis information will be
        read. The file will usually have been output previously as a
        'Dump File'.
        The file will not be accessed if options 'Old Basis File' or
        'Insert File' are specified.

    'Load File' : int
        Default = 0

        Options 'Dump File' and 'Load File' are similar to options
        'Punch File' and 'Insert File', but they record solution
        information in a manner that is more direct and more easily
        modified.
        A full description of information recorded in options 'Dump
        File' and 'Load File' is given in Gill `et al.` (2005a).

        If i_1 > 0, the last solution obtained will be output to the
        file with unit number i.

        If i_2 > 0, the 'Load File' containing basis information will be
        read. The file will usually have been output previously as a
        'Dump File'.
        The file will not be accessed if options 'Old Basis File' or
        'Insert File' are specified.

    'Elastic Mode' : int
        Default = 1

        This argument determines if (and when) elastic mode is to be
        started.
        Three elastic modes are available as follows:

        [table omitted]

    'Elastic Objective' : int
        Default = 1

        This determines the form of the composite objective f(x)+gamma
        sum_j(v_j+w_j) in Phase 2 (gamma).
        Three types of composite objectives are available.

        [table omitted]

    'Elastic Weight' : float
        Default = 1.0

        This defines the value of gamma in the composite objective in
        Phase 2 (gamma).

        At each iteration of elastic mode, the composite objective is
        defined to be

            minimize sigma f(x)+gamma (sum of infeasibilities);

        where sigma = 1 for 'Minimize', sigma = -1 for 'Maximize', and
        f(x) is the quadratic objective.

        Note that the effect of gamma is `not` disabled once a feasible
        point is obtained.

    'Expand Frequency' : int
        Default = 10000

        This option is part of an anti-cycling procedure (see
        [Miscellaneous]) designed to allow progress even on highly
        degenerate problems.

        The strategy is to force a positive step at every iteration, at
        the expense of violating the constraints by a small amount.
        Suppose that the value of the option 'Feasibility Tolerance' is
        delta.
        Over a period of i iterations, the feasibility tolerance
        actually used by ``qpconvex2_sparse_solve`` (i.e., the `working`
        feasibility tolerance) increases from 0.5 delta to delta (in
        steps of 0.5 delta/i).

        Increasing the value of i helps reduce the number of slightly
        infeasible nonbasic variables (most of which are eliminated
        during the resetting procedure).
        However, it also diminishes the freedom to choose a large pivot
        element (see the description of the option 'Pivot Tolerance').

        If i <= 0, the value i = 99999999 is used and effectively no
        anti-cycling procedure is invoked.

    'Factorization Frequency' : int
        Default = 100*(LP) or 50*(QP)

        If i > 0, at most i basis changes will occur between
        factorizations of the basis matrix.

        For LP problems, the basis factors are usually updated at every
        iteration.
        Higher values of i may be more efficient on problems that are
        extremely sparse and well scaled.

        For QP problems, fewer basis updates will occur as the solution
        is approached.
        The number of iterations between basis factorizations will
        therefore increase.
        During these iterations a test is made regularly according to
        the value of option 'Check Frequency' to ensure that the linear
        constraints Ax-s = 0 are satisfied.
        Occasionally, the basis will be refactorized before the limit of
        i updates is reached.
        If i <= 0, the default value is used.

    'Feasibility Tolerance' : float
        Default = max{10^(-6),sqrt(epsilon)}

        A `feasible problem` is one in which all variables satisfy their
        upper and lower bounds to within the absolute tolerance r. (This
        includes slack variables.
        Hence, the general constraints are also satisfied to within r.)

        ``qpconvex2_sparse_solve`` attempts to find a feasible solution
        before optimizing the objective function.
        If the sum of infeasibilities cannot be reduced to zero, the
        problem is assumed to be `infeasible`.
        Let sInf be the corresponding sum of infeasibilities.
        If sInf is quite small, it may be appropriate to raise r by a
        factor of 10 or 100.
        Otherwise, some error in the data should be suspected.

        Note that if sInf is not small and you have not asked
        ``qpconvex2_sparse_solve`` to minimize the violations of the
        elastic variables (i.e., you have not specified 'Elastic
        Objective' = 2), there may be other points that have a
        `significantly smaller sum of infeasibilities`.
        ``qpconvex2_sparse_solve`` will not attempt to find the solution
        that minimizes the sum unless 'Elastic Objective' = 2.

        If the constraints and variables have been scaled (see the
        description of the option 'Scale Option'), then feasibility is
        defined in terms of the scaled problem (since it is more likely
        to be meaningful).

    'Infinite Bound Size' : float
        Default = 10^20

        If r >= 0, r defines the 'infinite' bound infbnd in the
        definition of the problem constraints.
        Any upper bound greater than or equal to infbnd will be regarded
        as + infinity (and similarly any lower bound less than or equal
        to -infbnd will be regarded as - infinity).
        If r < 0, the default value is used.

    'Iterations Limit' : int
        Default = max{10000,10max{m,n}}

        The value of i specifies the maximum number of iterations
        allowed before termination.
        Setting i = 0 and 'Print Level' > 0 means that: the workspace
        needed to start solving the problem will be computed and
        printed; and feasibility and optimality will be checked.
        No iterations will be performed.
        If i < 0, the default value is used.

    'LU Density Tolerance' : float
        Default = 0.6

        The density tolerance r_1 is used during LU factorization of the
        basis matrix.
        Columns of L and rows of U are formed one at a time, and the
        remaining rows and columns of the basis are altered
        appropriately.
        At any stage, if the density of the remaining matrix exceeds
        r_1, the Markowitz strategy for choosing pivots is terminated.
        The remaining matrix is factored by a dense LU procedure.
        Raising the density tolerance towards 1.0 may give slightly
        sparser LU factors, with a slight increase in factorization
        time.

        If r_2 > 0, r_2 defines the singularity tolerance used to guard
        against ill-conditioned basis matrices.
        After B is refactorized, the diagonal elements of U are tested
        as follows.
        If |u_{jj}| <= r_2 or |u_{jj}| < r_2 max_i|u_{ij}|, the jth
        column of the basis is replaced by the corresponding slack
        variable.
        If r_2 <= 0, the default value is used.

    'LU Singularity Tolerance' : float
        Default = epsilon^(2/3)

        The density tolerance r_1 is used during LU factorization of the
        basis matrix.
        Columns of L and rows of U are formed one at a time, and the
        remaining rows and columns of the basis are altered
        appropriately.
        At any stage, if the density of the remaining matrix exceeds
        r_1, the Markowitz strategy for choosing pivots is terminated.
        The remaining matrix is factored by a dense LU procedure.
        Raising the density tolerance towards 1.0 may give slightly
        sparser LU factors, with a slight increase in factorization
        time.

        If r_2 > 0, r_2 defines the singularity tolerance used to guard
        against ill-conditioned basis matrices.
        After B is refactorized, the diagonal elements of U are tested
        as follows.
        If |u_{jj}| <= r_2 or |u_{jj}| < r_2 max_i|u_{ij}|, the jth
        column of the basis is replaced by the corresponding slack
        variable.
        If r_2 <= 0, the default value is used.

    'LU Factor Tolerance' : float
        Default = 100.0

        The values of r_1 and r_2 affect the stability and sparsity of
        the basis factorization B = LU, during refactorization and
        updates respectively.
        The lower triangular matrix L is a product of matrices of the
        form

            [table omitted]

        where the multipliers mu will satisfy |mu| <= r_i.
        The default values of r_1 and r_2 usually strike a good
        compromise between stability and sparsity.
        They must satisfy r_1, r_2 >= 1.0.

        For large and relatively dense problems, r_1 = 10.0 or 5.0 (say)
        may give a useful improvement in stability without impairing
        sparsity to a serious degree.

        For certain very regular structures (e.g., band matrices) it may
        be necessary to reduce r_1 and/or r_2 in order to achieve
        stability.
        For example, if the columns of A include a sub-matrix of the
        form

            [table omitted],

        one should set both r_1 and r_2 to values in the range 1.0 <=
        r_i < 4.0.

    'LU Update Tolerance' : float
        Default = 10.0

        The values of r_1 and r_2 affect the stability and sparsity of
        the basis factorization B = LU, during refactorization and
        updates respectively.
        The lower triangular matrix L is a product of matrices of the
        form

            [table omitted]

        where the multipliers mu will satisfy |mu| <= r_i.
        The default values of r_1 and r_2 usually strike a good
        compromise between stability and sparsity.
        They must satisfy r_1, r_2 >= 1.0.

        For large and relatively dense problems, r_1 = 10.0 or 5.0 (say)
        may give a useful improvement in stability without impairing
        sparsity to a serious degree.

        For certain very regular structures (e.g., band matrices) it may
        be necessary to reduce r_1 and/or r_2 in order to achieve
        stability.
        For example, if the columns of A include a sub-matrix of the
        form

            [table omitted],

        one should set both r_1 and r_2 to values in the range 1.0 <=
        r_i < 4.0.

    'LU Partial Pivoting' : valueless
        Default

        The LU factorization implements a Markowitz-type search for
        pivots that locally minimize the fill-in subject to a threshold
        pivoting stability criterion.
        The default option is to use threshold partial pivoting.
        The options 'LU Complete Pivoting' and 'LU Rook Pivoting' are
        more expensive but more stable and better at revealing rank, as
        long as the 'LU Factor Tolerance' is not too large (say < 2.0).

    'LU Complete Pivoting' : valueless
        The LU factorization implements a Markowitz-type search for
        pivots that locally minimize the fill-in subject to a threshold
        pivoting stability criterion.
        The default option is to use threshold partial pivoting.
        The options 'LU Complete Pivoting' and 'LU Rook Pivoting' are
        more expensive but more stable and better at revealing rank, as
        long as the 'LU Factor Tolerance' is not too large (say < 2.0).

    'LU Rook Pivoting' : valueless
        The LU factorization implements a Markowitz-type search for
        pivots that locally minimize the fill-in subject to a threshold
        pivoting stability criterion.
        The default option is to use threshold partial pivoting.
        The options 'LU Complete Pivoting' and 'LU Rook Pivoting' are
        more expensive but more stable and better at revealing rank, as
        long as the 'LU Factor Tolerance' is not too large (say < 2.0).

    'Minimize' : valueless
        Default

        This option specifies the required direction of the
        optimization.
        It applies to both linear and nonlinear terms (if any) in the
        objective function.
        Note that if two problems are the same except that one minimizes
        f(x) and the other maximizes -f(x), their solutions will be the
        same but the signs of the dual variables pi_i and the reduced
        gradients d_j (see [Main Iteration]) will be reversed.

        The option 'Feasible Point' means 'ignore the objective
        function, while finding a feasible point for the linear
        constraints'.
        It can be used to check that the constraints are feasible
        without altering the call to ``qpconvex2_sparse_solve``.

    'Maximize' : valueless
        This option specifies the required direction of the
        optimization.
        It applies to both linear and nonlinear terms (if any) in the
        objective function.
        Note that if two problems are the same except that one minimizes
        f(x) and the other maximizes -f(x), their solutions will be the
        same but the signs of the dual variables pi_i and the reduced
        gradients d_j (see [Main Iteration]) will be reversed.

        The option 'Feasible Point' means 'ignore the objective
        function, while finding a feasible point for the linear
        constraints'.
        It can be used to check that the constraints are feasible
        without altering the call to ``qpconvex2_sparse_solve``.

    'Feasible Point' : valueless
        This option specifies the required direction of the
        optimization.
        It applies to both linear and nonlinear terms (if any) in the
        objective function.
        Note that if two problems are the same except that one minimizes
        f(x) and the other maximizes -f(x), their solutions will be the
        same but the signs of the dual variables pi_i and the reduced
        gradients d_j (see [Main Iteration]) will be reversed.

        The option 'Feasible Point' means 'ignore the objective
        function, while finding a feasible point for the linear
        constraints'.
        It can be used to check that the constraints are feasible
        without altering the call to ``qpconvex2_sparse_solve``.

    'New Basis File' : int
        Default = 0

        Options 'New Basis File' and 'Backup Basis File' are sometimes
        referred to as `basis maps`.
        They contain the most compact representation of the state of
        each variable.
        They are intended for restarting the solution of a problem at a
        point that was reached by an earlier run.
        For nontrivial problems, it is advisable to save basis maps at
        the end of a run, in order to restart the run if necessary.

        If i_1 > 0, a basis map will be saved on file i_1 every i_3th
        iteration, where i_3 is the 'Save Frequency'.
        The first record of the file will contain the word
        ``PROCEEDING`` if the run is still in progress.
        A basis map will also be saved at the end of a run, with some
        other word indicating the final solution status.

        Use of i_2 > 0 is intended as a safeguard against losing the
        results of a long run.
        Suppose that a 'New Basis File' is being saved every 100 ('Save
        Frequency') iterations, and that ``qpconvex2_sparse_solve`` is
        about to save such a basis at iteration 2000.
        It is conceivable that the run may be interrupted during the
        next few milliseconds (in the middle of the save).
        In this case the Basis file will be corrupted and the run will
        have been essentially wasted.

        To eliminate this risk, both a 'New Basis File' and a 'Backup
        Basis File' may be specified.

        The current basis will then be saved every 100 iterations, first
        on 'New Basis File' and then immediately on 'Backup Basis File'.
        If the run is interrupted at iteration 2000 during the save on
        'New Basis File', there will still be a usable basis on 'Backup
        Basis File' (corresponding to iteration 1900).

        Note that a new basis will be saved in 'New Basis File' at the
        end of a run if it terminates normally, but it will not be saved
        in 'Backup Basis File'.
        In the above example, if an optimum solution is found at
        iteration 2050 (or if the iteration limit is 2050), the final
        basis on 'New Basis File' will correspond to iteration 2050, but
        the last basis saved on 'Backup Basis File' will be the one for
        iteration 2000.

        A full description of information recorded in 'New Basis File'
        and 'Backup Basis File' is given in Gill `et al.` (2005a).

    'Backup Basis File' : int
        Default = 0

        Options 'New Basis File' and 'Backup Basis File' are sometimes
        referred to as `basis maps`.
        They contain the most compact representation of the state of
        each variable.
        They are intended for restarting the solution of a problem at a
        point that was reached by an earlier run.
        For nontrivial problems, it is advisable to save basis maps at
        the end of a run, in order to restart the run if necessary.

        If i_1 > 0, a basis map will be saved on file i_1 every i_3th
        iteration, where i_3 is the 'Save Frequency'.
        The first record of the file will contain the word
        ``PROCEEDING`` if the run is still in progress.
        A basis map will also be saved at the end of a run, with some
        other word indicating the final solution status.

        Use of i_2 > 0 is intended as a safeguard against losing the
        results of a long run.
        Suppose that a 'New Basis File' is being saved every 100 ('Save
        Frequency') iterations, and that ``qpconvex2_sparse_solve`` is
        about to save such a basis at iteration 2000.
        It is conceivable that the run may be interrupted during the
        next few milliseconds (in the middle of the save).
        In this case the Basis file will be corrupted and the run will
        have been essentially wasted.

        To eliminate this risk, both a 'New Basis File' and a 'Backup
        Basis File' may be specified.

        The current basis will then be saved every 100 iterations, first
        on 'New Basis File' and then immediately on 'Backup Basis File'.
        If the run is interrupted at iteration 2000 during the save on
        'New Basis File', there will still be a usable basis on 'Backup
        Basis File' (corresponding to iteration 1900).

        Note that a new basis will be saved in 'New Basis File' at the
        end of a run if it terminates normally, but it will not be saved
        in 'Backup Basis File'.
        In the above example, if an optimum solution is found at
        iteration 2050 (or if the iteration limit is 2050), the final
        basis on 'New Basis File' will correspond to iteration 2050, but
        the last basis saved on 'Backup Basis File' will be the one for
        iteration 2000.

        A full description of information recorded in 'New Basis File'
        and 'Backup Basis File' is given in Gill `et al.` (2005a).

    'Save Frequency' : int
        Default = 100

        Options 'New Basis File' and 'Backup Basis File' are sometimes
        referred to as `basis maps`.
        They contain the most compact representation of the state of
        each variable.
        They are intended for restarting the solution of a problem at a
        point that was reached by an earlier run.
        For nontrivial problems, it is advisable to save basis maps at
        the end of a run, in order to restart the run if necessary.

        If i_1 > 0, a basis map will be saved on file i_1 every i_3th
        iteration, where i_3 is the 'Save Frequency'.
        The first record of the file will contain the word
        ``PROCEEDING`` if the run is still in progress.
        A basis map will also be saved at the end of a run, with some
        other word indicating the final solution status.

        Use of i_2 > 0 is intended as a safeguard against losing the
        results of a long run.
        Suppose that a 'New Basis File' is being saved every 100 ('Save
        Frequency') iterations, and that ``qpconvex2_sparse_solve`` is
        about to save such a basis at iteration 2000.
        It is conceivable that the run may be interrupted during the
        next few milliseconds (in the middle of the save).
        In this case the Basis file will be corrupted and the run will
        have been essentially wasted.

        To eliminate this risk, both a 'New Basis File' and a 'Backup
        Basis File' may be specified.

        The current basis will then be saved every 100 iterations, first
        on 'New Basis File' and then immediately on 'Backup Basis File'.
        If the run is interrupted at iteration 2000 during the save on
        'New Basis File', there will still be a usable basis on 'Backup
        Basis File' (corresponding to iteration 1900).

        Note that a new basis will be saved in 'New Basis File' at the
        end of a run if it terminates normally, but it will not be saved
        in 'Backup Basis File'.
        In the above example, if an optimum solution is found at
        iteration 2050 (or if the iteration limit is 2050), the final
        basis on 'New Basis File' will correspond to iteration 2050, but
        the last basis saved on 'Backup Basis File' will be the one for
        iteration 2000.

        A full description of information recorded in 'New Basis File'
        and 'Backup Basis File' is given in Gill `et al.` (2005a).

    'Nolist' : valueless
        Default

        Option 'List' enables printing of each option specification as
        it is supplied. 'Nolist' suppresses this printing.

    'List' : valueless
        Option 'List' enables printing of each option specification as
        it is supplied. 'Nolist' suppresses this printing.

    'Old Basis File' : int
        Default = 0

        If i > 0, the basis maps information will be obtained from this
        file. The file will usually have been output previously as a
        'New Basis File' or 'Backup Basis File'.
        A full description of information recorded in 'New Basis File'
        and 'Backup Basis File' is given in Gill `et al.` (2005a).

        The file will not be acceptable if the number of rows or columns
        in the problem has been altered.

    'Optimality Tolerance' : float
        Default = max{10^(-6),sqrt(epsilon)}

        This is used to judge the size of the reduced gradients d_j =
        g_j-a_j^T pi, where g_j is the jth component of the gradient,
        a_j is the associated column of the constraint matrix (A -I),
        and pi is the set of dual variables.

        By construction, the reduced gradients for basic variables are
        always zero.
        The problem will be declared optimal if the reduced gradients
        for nonbasic variables at their lower or upper bounds satisfy

            d_j/||pi|| >= -r or d_j/||pi|| <= r

        respectively, and if |d_j|/||pi|| <= r for superbasic variables.

        In the above tests, ||pi|| is a measure of the size of the dual
        variables.
        It is included to make the tests independent of a scale factor
        on the objective function.
        The quantity ||pi|| actually used is defined by

            ||pi|| = max(sigma/sqrt(m),1), where sigma = sum_i =
            1^m|pi_i|,

        so that only large scale factors are allowed for.

        If the objective is scaled down to be very `small`, the
        optimality test reduces to comparing d_j against 0.01r.

    'Partial Price' : int
        Default = 10*(LP) or 1*(QP)

        This option is recommended for large FP or LP problems that have
        significantly more variables than constraints (i.e., n≫m).
        It reduces the work required for each pricing operation (i.e.,
        when a nonbasic variable is selected to enter the basis).
        If i = 1, all columns of the constraint matrix (A -I) are
        searched.
        If i > 1, A and I are partitioned to give i roughly equal
        segments A_j,I_j, for j = 1,2,...,i (modulo i).
        If the previous pricing search was successful on
        A_{j-1},I_{j-1}, the next search begins on the segments A_j and
        I_j.
        If a reduced gradient is found that is larger than some dynamic
        tolerance, the variable with the largest such reduced gradient
        (of appropriate sign) is selected to enter the basis.
        If nothing is found, the search continues on the next segments
        A_{j+1},I_{j+1}, and so on.
        If i <= 0, the default value is used.

    'Pivot Tolerance' : float
        Default = epsilon^(2/3)

        Broadly speaking, the pivot tolerance is used to prevent columns
        entering the basis if they would cause the basis to become
        almost singular.

        When x changes to x+alpha p for some search direction p, a
        'ratio test' determines which component of x reaches an upper or
        lower bound first.
        The corresponding element of p is called the pivot element.
        Elements of p are ignored (and therefore cannot be pivot
        elements) if they are smaller than the pivot tolerance r.

        It is common for two or more variables to reach a bound at
        essentially the same time.
        In such cases, the option 'Feasibility Tolerance' (say t)
        provides some freedom to maximize the pivot element and thereby
        improve numerical stability.
        Excessively small values of t should therefore not be specified.
        To a lesser extent, the option 'Expand Frequency' (say f) also
        provides some freedom to maximize the pivot element.
        Excessively `large` values of f should therefore not be
        specified.

    'Print File' : int
        Default = 0

        If i > 0, the following information is output to i during the
        solution of each problem:

        -   a listing of the options;

        -   some statistics about the problem;

        -   the amount of storage available for the LU factorization of
            the basis matrix;

        -   notes about the initial basis resulting from a Crash
            procedure or a Basis file;

        -   the iteration log;

        -   basis factorization statistics;

        -   the exit errno condition and some statistics about the
            solution obtained;

        -   the printed solution, if requested.

        The last four items are described in [Further Comments] and
        [Monitoring Information].
        Further brief output may be directed to the 'Summary File'.

    'Print Frequency' : int
        Default = 100

        If i > 0, one line of the iteration log will be printed every
        ith iteration.
        A value such as i = 10 is suggested for those interested only in
        the final solution.
        If i <= 0, the value of i = 99999999 is used and effectively no
        checks are made.

    'Print Level' : int
        Default = 1

        This controls the amount of printing produced by
        ``qpconvex2_sparse_solve`` as follows.

        [table omitted]

    'Punch File' : int
        Default = 0

        These files provide compatibility with commercial mathematical
        programming systems.
        The 'Punch File' from a previous run may be used as an 'Insert
        File' for a later run on the same problem.
        A full description of information recorded in 'Insert File' and
        'Punch File' is given in Gill `et al.` (2005a).

        If i_1 > 0, the final solution obtained will be output to file
        i_1. For linear programs, this format is compatible with various
        commercial systems.

        If i_2 > 0, the 'Insert File' containing basis information will
        be read.
        The file will usually have been output previously as a 'Punch
        File'.
        The file will not be accessed if 'Old Basis File' is specified.

    'Insert File' : int
        Default = 0

        These files provide compatibility with commercial mathematical
        programming systems.
        The 'Punch File' from a previous run may be used as an 'Insert
        File' for a later run on the same problem.
        A full description of information recorded in 'Insert File' and
        'Punch File' is given in Gill `et al.` (2005a).

        If i_1 > 0, the final solution obtained will be output to file
        i_1. For linear programs, this format is compatible with various
        commercial systems.

        If i_2 > 0, the 'Insert File' containing basis information will
        be read.
        The file will usually have been output previously as a 'Punch
        File'.
        The file will not be accessed if 'Old Basis File' is specified.

    'QPSolver Cholesky' : valueless
        Default

        Specifies the active-set algorithm used to solve the quadratic
        program in Phase 2 (gamma). 'QPSolver Cholesky' holds the full
        Cholesky factor R of the reduced Hessian Z^THZ.
        As the QP iterations proceed, the dimension of R changes with
        the number of superbasic variables.
        If the number of superbasic variables needs to increase beyond
        the value of 'Reduced Hessian Dimension', the reduced Hessian
        cannot be stored and the solver switches to 'QPSolver CG'.
        The Cholesky solver is reactivated if the number of superbasics
        stabilizes at a value less than 'Reduced Hessian Dimension'.

        'QPSolver QN' solves the QP using a quasi-Newton method.
        In this case, R is the factor of a quasi-Newton approximate
        Hessian.

        'QPSolver CG' uses an active-set method similar to 'QPSolver
        QN', but uses the conjugate-gradient method to solve all systems
        involving the reduced Hessian.

        The Cholesky QP solver is the most robust, but may require a
        significant amount of computation if there are many superbasics.

        The quasi-Newton QP solver does not require computation of the
        exact R at the start of Phase 2 (gamma).
        It may be appropriate when the number of superbasics is large
        but relatively few iterations are needed to reach a solution
        (e.g., if ``qpconvex2_sparse_solve`` is called with a Warm
        Start).

        The conjugate-gradient QP solver is appropriate for problems
        with many degrees of freedom (say, more than 2000 superbasics).

    'QPSolver CG' : valueless
        Specifies the active-set algorithm used to solve the quadratic
        program in Phase 2 (gamma). 'QPSolver Cholesky' holds the full
        Cholesky factor R of the reduced Hessian Z^THZ.
        As the QP iterations proceed, the dimension of R changes with
        the number of superbasic variables.
        If the number of superbasic variables needs to increase beyond
        the value of 'Reduced Hessian Dimension', the reduced Hessian
        cannot be stored and the solver switches to 'QPSolver CG'.
        The Cholesky solver is reactivated if the number of superbasics
        stabilizes at a value less than 'Reduced Hessian Dimension'.

        'QPSolver QN' solves the QP using a quasi-Newton method.
        In this case, R is the factor of a quasi-Newton approximate
        Hessian.

        'QPSolver CG' uses an active-set method similar to 'QPSolver
        QN', but uses the conjugate-gradient method to solve all systems
        involving the reduced Hessian.

        The Cholesky QP solver is the most robust, but may require a
        significant amount of computation if there are many superbasics.

        The quasi-Newton QP solver does not require computation of the
        exact R at the start of Phase 2 (gamma).
        It may be appropriate when the number of superbasics is large
        but relatively few iterations are needed to reach a solution
        (e.g., if ``qpconvex2_sparse_solve`` is called with a Warm
        Start).

        The conjugate-gradient QP solver is appropriate for problems
        with many degrees of freedom (say, more than 2000 superbasics).

    'QPSolver QN' : valueless
        Specifies the active-set algorithm used to solve the quadratic
        program in Phase 2 (gamma). 'QPSolver Cholesky' holds the full
        Cholesky factor R of the reduced Hessian Z^THZ.
        As the QP iterations proceed, the dimension of R changes with
        the number of superbasic variables.
        If the number of superbasic variables needs to increase beyond
        the value of 'Reduced Hessian Dimension', the reduced Hessian
        cannot be stored and the solver switches to 'QPSolver CG'.
        The Cholesky solver is reactivated if the number of superbasics
        stabilizes at a value less than 'Reduced Hessian Dimension'.

        'QPSolver QN' solves the QP using a quasi-Newton method.
        In this case, R is the factor of a quasi-Newton approximate
        Hessian.

        'QPSolver CG' uses an active-set method similar to 'QPSolver
        QN', but uses the conjugate-gradient method to solve all systems
        involving the reduced Hessian.

        The Cholesky QP solver is the most robust, but may require a
        significant amount of computation if there are many superbasics.

        The quasi-Newton QP solver does not require computation of the
        exact R at the start of Phase 2 (gamma).
        It may be appropriate when the number of superbasics is large
        but relatively few iterations are needed to reach a solution
        (e.g., if ``qpconvex2_sparse_solve`` is called with a Warm
        Start).

        The conjugate-gradient QP solver is appropriate for problems
        with many degrees of freedom (say, more than 2000 superbasics).

    'Reduced Hessian Dimension' : int
        Default = 1*(LP) or min(2000,n_H+1,n)(QP)

        This specifies that an i by i triangular matrix R (to define the
        reduced Hessian according to R^TR = Z^THZ). is to be available
        for use by the Cholesky QP solver.

    'Scale Option' : int
        Default = 2

        Three scale options are available as follows:

        [table omitted]

        Option 'Scale Tolerance' affects how many passes might be needed
        through the constraint matrix.
        On each pass, the scaling procedure computes the ratio of the
        largest and smallest nonzero coefficients in each column:

            rho_j = max_j|a_{ij}|/min_i|a_{ij}| (a_{ij} != 0).

        If max_jrho_j is less than r times its previous value, another
        scaling pass is performed to adjust the row and column scales.
        Raising r from 0.9 to 0.99 (say) usually increases the number of
        scaling passes through A.
        At most 10 passes are made.
        The value of r should lie in the range 0 < r < 1.

        'Scale Print' causes the row scales r(i) and column scales c(j)
        to be printed to 'Print File', if 'System Information Yes' has
        been specified.
        The scaled matrix coefficients are a__{ij} = a_{ij}c(j)/r(i),
        and the scaled bounds on the variables and slacks are l__j =
        l_j/c(j), u__j = u_j/c(j), where c(j) = r(j-n) if j > n.

    'Scale Tolerance' : float
        Default = 0.9

        Three scale options are available as follows:

        [table omitted]

        Option 'Scale Tolerance' affects how many passes might be needed
        through the constraint matrix.
        On each pass, the scaling procedure computes the ratio of the
        largest and smallest nonzero coefficients in each column:

            rho_j = max_j|a_{ij}|/min_i|a_{ij}| (a_{ij} != 0).

        If max_jrho_j is less than r times its previous value, another
        scaling pass is performed to adjust the row and column scales.
        Raising r from 0.9 to 0.99 (say) usually increases the number of
        scaling passes through A.
        At most 10 passes are made.
        The value of r should lie in the range 0 < r < 1.

        'Scale Print' causes the row scales r(i) and column scales c(j)
        to be printed to 'Print File', if 'System Information Yes' has
        been specified.
        The scaled matrix coefficients are a__{ij} = a_{ij}c(j)/r(i),
        and the scaled bounds on the variables and slacks are l__j =
        l_j/c(j), u__j = u_j/c(j), where c(j) = r(j-n) if j > n.

    'Scale Print' : valueless
        Three scale options are available as follows:

        [table omitted]

        Option 'Scale Tolerance' affects how many passes might be needed
        through the constraint matrix.
        On each pass, the scaling procedure computes the ratio of the
        largest and smallest nonzero coefficients in each column:

            rho_j = max_j|a_{ij}|/min_i|a_{ij}| (a_{ij} != 0).

        If max_jrho_j is less than r times its previous value, another
        scaling pass is performed to adjust the row and column scales.
        Raising r from 0.9 to 0.99 (say) usually increases the number of
        scaling passes through A.
        At most 10 passes are made.
        The value of r should lie in the range 0 < r < 1.

        'Scale Print' causes the row scales r(i) and column scales c(j)
        to be printed to 'Print File', if 'System Information Yes' has
        been specified.
        The scaled matrix coefficients are a__{ij} = a_{ij}c(j)/r(i),
        and the scaled bounds on the variables and slacks are l__j =
        l_j/c(j), u__j = u_j/c(j), where c(j) = r(j-n) if j > n.

    'Solution Yes' : valueless
        Default

        This option determines if the final obtained solution is to be
        output to the 'Print File'.
        Note that the 'Solution File' option operates independently.

    'Solution No' : valueless
        This option determines if the final obtained solution is to be
        output to the 'Print File'.
        Note that the 'Solution File' option operates independently.

    'Solution File' : int
        Default = 0

        If i > 0, the final solution will be output to file i (whether
        optimal or not).

        To see more significant digits in the printed solution, it will
        sometimes be useful to make i refer to the system 'Print File'.

    'Summary File' : int
        Default = 0

        If i_1 > 0, the 'Summary File' is output to file i_1, including
        a line of the iteration log every i_2th iteration. In an
        interactive environment, it is useful to direct this output to
        the terminal, to allow a run to be monitored online. (If
        something looks wrong, the run can be manually terminated.)
        Further details are given in [Monitoring Information].
        If i_2 <= 0, the value of i_2 = 99999999 is used and effectively
        no checks are made.

    'Summary Frequency' : int
        Default = 100

        If i_1 > 0, the 'Summary File' is output to file i_1, including
        a line of the iteration log every i_2th iteration. In an
        interactive environment, it is useful to direct this output to
        the terminal, to allow a run to be monitored online. (If
        something looks wrong, the run can be manually terminated.)
        Further details are given in [Monitoring Information].
        If i_2 <= 0, the value of i_2 = 99999999 is used and effectively
        no checks are made.

    'Superbasics Limit' : int
        Default = 1*(LP) or min{n_H+1,n}(QP)

        This places a limit on the storage allocated for superbasic
        variables.
        Ideally, i should be set slightly larger than the 'number of
        degrees of freedom' expected at an optimal solution.

        For linear programs, an optimum is normally a basic solution
        with no degrees of freedom. (The number of variables lying
        strictly between their bounds is no more than m, the number of
        general constraints.) The default value of i is therefore 1.

        For quadratic problems, the number of degrees of freedom is
        often called the 'number of independent variables'.
        Normally, i need not be greater than n_H+1, where n_H is the
        number of leading nonzero columns of H.
        For many problems, i may be considerably smaller than n_H.
        This will save storage if n_H is very large.

    'Suppress Parameters' : valueless
        Normally ``qpconvex2_sparse_solve`` prints the options file as
        it is being read, and then prints a complete list of the
        available keywords and their final values.
        The option 'Suppress Parameters' tells
        ``qpconvex2_sparse_solve`` not to print the full list.

    'System Information No' : valueless
        Default

        This option prints additional information on the progress of
        major and minor iterations, and Crash statistics.
        See [Monitoring Information].

    'System Information Yes' : valueless
        This option prints additional information on the progress of
        major and minor iterations, and Crash statistics.
        See [Monitoring Information].

    'Timing Level' : int
        Default = 0

        If i > 0, some timing information will be output to the Print
        file, if 'Print File' > 0.

    'Unbounded Step Size' : float
        Default = infbnd

        If r > 0, r specifies the magnitude of the change in variables
        that will be considered a step to an unbounded solution. (Note
        that an unbounded solution can occur only when the Hessian is
        not positive definite.) If the change in x during an iteration
        would exceed the value of r, the objective function is
        considered to be unbounded below in the feasible region.
        If r <= 0, the default value is used.
        See 'Infinite Bound Size' for the definition of infbnd.

    Raises
    ------
    NagValueError
        (`errno` 1)
            The initialization function ``qpconvex2_sparse_init`` has
            not been called.

        (`errno` 2)
            On entry, `start` = *<value>*.

            Constraint: `start` = 'B', 'C' or 'W'.

        (`errno` 2)
            On entry, ne = *<value>*, n = *<value>* and `m` = *<value>*.

            Constraint: 1 <= ne <= n*`m`.

        (`errno` 2)
            On entry, lenc = *<value>* and n = *<value>*.

            Constraint: 0 <= lenc <= n.

        (`errno` 2)
            An error has occurred in the basis package, perhaps
            indicating incorrect setup of arrays `inda` and `loca`. Set
            the option 'Print File' and examine the output carefully for
            further information.

        (`errno` 2)
            On entry, `ncolh` = *<value>* and n = *<value>*.

            Constraint: 0 <= `ncolh` <= n.

        (`errno` 2)
            On entry, nname = *<value>*, n = *<value>* and `m` =
            *<value>*.

            Constraint: nname = 1 or n+`m`.

        (`errno` 2)
            On entry, `iobj` = *<value>* and `m` = *<value>*.

            Constraint: 0 <= `iobj` <= `m`.

        (`errno` 2)
            On entry, `loca`[0] = *<value>*, `loca`[*<value>*] =
            *<value>*, ne = *<value>*.

            Constraint: `loca`[0] = 1 or `loca`[*<value>*] = ne+1.

        (`errno` 2)
            On entry, row index *<value>* in `inda`[*<value>*] is
            outside the range 1 to `m` = *<value>*.

        (`errno` 2)
            On entry, ne is not equal to the number of nonzeros in
            `acol`. ne = *<value>*, nonzeros in `acol` = *<value>*.

        (`errno` 2)
            On entry, bounds `bl` and `bu` for *<value>* are equal and
            infinite: `bl` = `bu` = *<value>* and infbnd = *<value>*.

        (`errno` 2)
            On entry, bounds for *<value>* are inconsistent. `bl` =
            *<value>* and `bu` = *<value>*.

        (`errno` 2)
            On entry, bounds `bl` and `bu` for *<value>**<value>* are
            equal and infinite. `bl` = `bu` = *<value>* and infbnd =
            *<value>*.

        (`errno` 2)
            On entry, bounds for *<value>**<value>* are inconsistent.
            `bl` = *<value>* and `bu` = *<value>*.

        (`errno` 2)
            Basis file dimensions do not match this problem.

        (`errno` 2)
            On entry, `m` = *<value>*.

            Constraint: `m` >= *<value>*.

        (`errno` 2)
            On entry, n = *<value>*.

            Constraint: n >= *<value>*.

        (`errno` 12)
            Internal memory allocation failed when attempting to obtain
            workspace sizes *<value>*, *<value>* and *<value>*. Please
            contact NAG.

        (`errno` 13)
            Internal memory allocation was insufficient. Please contact
            NAG.

        (`errno` 14)
            An error has occurred in the basis package, perhaps
            indicating incorrect setup of arrays `inda` and `loca`. Set
            the option 'Print File' and examine the output carefully for
            further information.

        (`errno` 15)
            An unexpected error has occurred. Set the option 'Print
            File' and examine the output carefully for further
            information.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 4)
            Weak solution found -- the solution is not unique.

        (`errno` 5)
            The linear constraints appear to be infeasible.

        (`errno` 5)
            The problem appears to be infeasible. The linear equality
            constraints could not be satisfied.

        (`errno` 5)
            The problem appears to be infeasible. Nonlinear
            infeasibilites have been minimized.

        (`errno` 5)
            The problem appears to be infeasible. Infeasibilites have
            been minimized.

        (`errno` 6)
            The problem appears to be unbounded. The objective function
            is unbounded.

        (`errno` 6)
            The problem appears to be unbounded. The constraint
            violation limit has been reached.

    NagAlgorithmicMajorWarning
        (`errno` 3)
            The requested accuracy could not be achieved.

        (`errno` 7)
            Iteration limit reached.

        (`errno` 7)
            Major iteration limit reached.

        (`errno` 8)
            The value of the option 'Superbasics Limit' is too small.

        (`errno` 9)
            The basis is singular after several attempts to factorize it
            (and add slacks where necessary).

        (`errno` 10)
            Numerical difficulties have been encountered and no further
            progress can be made.

        (`errno` 11)
            Error in `qphx`: the QP Hessian is indefinite.

    References
    ----------
    Fourer, R, 1982, `Solving staircase linear programs by the simplex
    method`, Math. Programming (23), 274--313

    Gill, P E and Murray, W, 1978, `Numerically stable methods for
    quadratic programming`, Math. Programming (14), 349--372

    Gill, P E, Murray, W and Saunders, M A, 1995, `User's guide for
    QPOPT 1.0: a Fortran package for quadratic programming`, Report SOL
    95-4, Department of Operations Research, Stanford University

    Gill, P E, Murray, W and Saunders, M A, 2005, `Users' guide for
    SQOPT 7: a Fortran package for large-scale linear and quadratic
    programming`, Report NA 05-1, Department of Mathematics, University
    of California, San Diego,
    https://www.ccom.ucsd.edu/~peg/papers/sqdoc7.pdf

    Gill, P E, Murray, W and Saunders, M A, 2005, `Users' guide for
    SNOPT 7.1: a Fortran package for large-scale linear nonlinear
    programming`, Report NA 05-2, Department of Mathematics, University
    of California, San Diego,
    https://www.ccom.ucsd.edu/~peg/papers/sndoc7.pdf

    Gill, P E, Murray, W, Saunders, M A and Wright, M H, 1987,
    `Maintaining LU factors of a general sparse matrix`, Linear Algebra
    and its Applics. (88/89), 239--270

    Gill, P E, Murray, W, Saunders, M A and Wright, M H, 1989, `A
    practical anti-cycling procedure for linearly constrained
    optimization`, Math. Programming (45), 437--474

    Gill, P E, Murray, W, Saunders, M A and Wright, M H, 1991,
    `Inertia-controlling methods for general quadratic programming`,
    SIAM Rev. (33), 1--36

    Hall, J A J and McKinnon, K I M, 1996, `The simplest examples where
    the simplex method cycles and conditions where EXPAND fails to
    prevent cycling`, Report MS, 96--100, Department of Mathematics and
    Statistics, University of Edinburgh

    See Also
    --------
    :meth:`naginterfaces.library.examples.opt.qpconvex2_sparse_solve_ex.main`
    """

    fun_name = 'naginterfaces.library.opt.qpconvex2_sparse_solve'

    cb_qphx_name = fun_name + ':qphx'

    def cb_qphx(ncolh, x, hx, nstate, data=None):
        """Wrapping user's qphx. Called as base qphx."""
        if data is not None:
            _hx_qphx = qphx(
                x,
                nstate,
                data,
            )
        else:
            _hx_qphx = qphx(
                x,
                nstate,
            )
        hx[:] = _utils._to_ndarray(
            {'fun_name': cb_qphx_name, 'returned': 'hx'},
            _hx_qphx,
            copy=False,
            exp_shape=(ncolh,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )

    _b_utils._check_callable(qphx, {'fun_name': fun_name, 'entity': 'qphx'}, allow_none=True)
    _qphx = None if qphx is None else cb_qphx
    _loca = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'loca'},
        loca,
        copy=False,
    )
    _n = _b_utils._EngineIntScalarType(
        _b_utils._EngineIntArrayType.get_shape(_loca, {'fun_name': fun_name, 'entity': 'loca'}, exp_rank=1)[0]-1, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _acol = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'acol'},
        acol,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ne_py = _b_utils._EngineFloat64ArrayType.get_shape(_acol, {'fun_name': fun_name, 'entity': 'acol'}, exp_rank=1)[0]
    _ne = _b_utils._EngineIntScalarType(
        _ne_py, {'fun_name': fun_name, 'entity': '_ne'},
    ).to_ctypes()
    _names = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'names'},
        names,
        copy=False,
    )
    _nname_py = _b_utils._EngineCharArrayType.get_shape(_names, {'fun_name': fun_name, 'entity': 'names'}, exp_rank=1)[0]
    _nname = _b_utils._EngineIntScalarType(
        _nname_py, {'fun_name': fun_name, 'entity': '_nname'},
    ).to_ctypes()
    _c = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
        copy=True,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    if c is None:
        _lenc = 0
        _lenc_py = _lenc
    else:
        _lenc_py = _b_utils._EngineFloat64ArrayType.get_shape(_c, {'fun_name': fun_name, 'entity': 'c'}, exp_rank=1, allow_none=True)[0]
        _lenc = _b_utils._EngineIntScalarType(
            _lenc_py, {'fun_name': fun_name, 'entity': '_lenc'},
        ).to_ctypes()
    _inda = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'inda'},
        inda,
        copy=False,
    )
    _bl = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'bl'},
        bl,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _bu = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'bu'},
        bu,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _iobj = _b_utils._EngineIntScalarType(
        iobj, {'fun_name': fun_name, 'entity': 'iobj'},
    ).to_ctypes()
    _iobj_py = _b_utils._EngineIntScalarType.to_py(_iobj)
    if _iobj_py == 0:
        if c is None:
            raise _b_utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'c'},
                'must not be None when iobj = 0',
            )
    if c is None:
        _c = _b_utils._EngineFloat64ArrayType.empty_ndarray(
            (max(1, _lenc_py),),
        )
    _helast = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'helast'},
        helast,
        copy=False,
    )
    _hs = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'hs'},
        hs,
        copy=True,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m = _b_utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = _b_utils._EngineIntScalarType.to_py(_m)
    _pi = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_m_py,),
    )
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _rc = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py+_m_py,),
    )
    try:
        _return_scalars = _b_opt.qpconvex2_sparse_solve(
            start,
            _qphx,
            _m,
            _n,
            _ne,
            _nname,
            _lenc,
            ncolh,
            _iobj,
            objadd,
            prob,
            _acol,
            _inda,
            _loca,
            _bl,
            _bu,
            _c,
            _names,
            _helast,
            _hs,
            _x,
            _pi,
            _rc,
            ns,
            comm,
            data,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _hs_py = _hs
    _x_py = _x
    _ns_py = _return_scalars.ns
    _pi_py = _pi
    _rc_py = _rc
    _ninf_py = _return_scalars.ninf
    _sinf_py = _return_scalars.sinf
    _obj_py = _return_scalars.obj
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['hs', 'x', 'pi', 'rc', 'ns', 'ninf', 'sinf', 'obj'],
    )(
        _hs_py,
        _x_py,
        _pi_py,
        _rc_py,
        _ns_py,
        _ninf_py,
        _sinf_py,
        _obj_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def qpconvex2_sparse_option_file(ispecs, comm, io_manager=None):
    u"""
    Supply option values for ``qpconvex2_sparse_solve`` from external
    file.

    ``qpconvex2_sparse_option_file`` may be used to supply options to
    ``qpconvex2_sparse_solve`` from an external file.
    The initialization function ``qpconvex2_sparse_init`` **must** have
    been called before calling ``qpconvex2_sparse_option_file``.

    For full information please refer to the NAG Library document for
    e04nr

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04nrf.html

    Parameters
    ----------
    ispecs : int
        The unit number (see
        :meth:`~naginterfaces.base.utils.FileObjManager.unit_from_fileobj`)
        of the option file to be read.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``qpconvex2_sparse_init``.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Raises
    ------
    NagValueError
        (`errno` 1)
            The initialization function ``qpconvex2_sparse_init`` has
            not been called.

        (`errno` 2)
            At least one line of the options file is invalid.

        (`errno` 2)
            Could not read options file on unit `ispecs` = *<value>*.
    """

    _b_opt.qpconvex2_sparse_option_file(
        ispecs,
        comm,
        io_manager,
    )

def qpconvex2_sparse_option_string(string, comm, io_manager=None):
    u"""
    Set a single option for ``qpconvex2_sparse_solve`` from a character
    string.

    ``qpconvex2_sparse_option_string`` may be used to supply individual
    options to ``qpconvex2_sparse_solve``.
    The initialization function ``qpconvex2_sparse_init`` **must** have
    been called before calling ``qpconvex2_sparse_option_string``.

    For full information please refer to the NAG Library document for
    e04ns

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04nsf.html

    Parameters
    ----------
    string : str
        A single valid option string (see [Description] in
        ``qpconvex2_sparse_option_string`` and [Other Parameters for
        qpconvex2_sparse_solve]).

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``qpconvex2_sparse_init``.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Raises
    ------
    NagValueError
        (`errno` 1)
            The initialization function ``qpconvex2_sparse_init`` has
            not been called.

        (`errno` 2)
            The supplied option is invalid. Check that the keywords are
            neither ambiguous nor misspelt. The option string is
            *<value>*.
    """

    _b_opt.qpconvex2_sparse_option_string(
        string,
        comm,
        io_manager,
    )

def qpconvex2_sparse_option_integer_set(string, ivalue, comm, io_manager=None):
    u"""
    Set a single option for ``qpconvex2_sparse_solve`` from an integer
    argument.

    ``qpconvex2_sparse_option_integer_set`` may be used to supply
    individual `int` options to ``qpconvex2_sparse_solve``.
    The initialization function ``qpconvex2_sparse_init`` **must** have
    been called before calling ``qpconvex2_sparse_option_integer_set``.

    For full information please refer to the NAG Library document for
    e04nt

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04ntf.html

    Parameters
    ----------
    string : str
        A single valid keyword of an `int` option (as described in
        [Other Parameters for qpconvex2_sparse_solve]).

    ivalue : int
        An `int` value associated with the keyword in `string`.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``qpconvex2_sparse_init``.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Raises
    ------
    NagValueError
        (`errno` 1)
            The initialization function ``qpconvex2_sparse_init`` has
            not been called.

        (`errno` 2)
            The supplied option is invalid. Check that the keywords are
            neither ambiguous nor misspelt. The option string is
            *<value>* and `ivalue` = *<value>*.
    """

    _b_opt.qpconvex2_sparse_option_integer_set(
        string,
        ivalue,
        comm,
        io_manager,
    )

def qpconvex2_sparse_option_double_set(string, rvalue, comm, io_manager=None):
    u"""
    Set a single option for ``qpconvex2_sparse_solve`` from a real
    argument.

    ``qpconvex2_sparse_option_double_set`` may be used to supply
    individual `float` options to ``qpconvex2_sparse_solve``.
    The initialization function ``qpconvex2_sparse_init`` **must** have
    been called before calling ``qpconvex2_sparse_option_double_set``.

    For full information please refer to the NAG Library document for
    e04nu

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04nuf.html

    Parameters
    ----------
    string : str
        A single valid keyword of a `float` option (as described in
        [Other Parameters for qpconvex2_sparse_solve]).

    rvalue : float
        The value associated with the keyword in `string`.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``qpconvex2_sparse_init``.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Raises
    ------
    NagValueError
        (`errno` 1)
            The initialization function ``qpconvex2_sparse_init`` has
            not been called.

        (`errno` 2)
            The supplied option is invalid. Check that the keywords are
            neither ambiguous nor misspelt. The option string is
            *<value>* and `rvalue` = *<value>*.
    """

    _b_opt.qpconvex2_sparse_option_double_set(
        string,
        rvalue,
        comm,
        io_manager,
    )

def qpconvex2_sparse_option_integer_get(string, comm, io_manager=None):
    u"""
    Get the setting of an integer valued option of
    ``qpconvex2_sparse_solve``.

    ``qpconvex2_sparse_option_integer_get`` is used to get the value of
    an `int` option. ``qpconvex2_sparse_option_integer_get`` can be used
    before or after calling ``qpconvex2_sparse_solve``.

    For full information please refer to the NAG Library document for
    e04nx

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04nxf.html

    Parameters
    ----------
    string : str
        A single valid keyword of an `int` option (as described in
        [Other Parameters for qpconvex2_sparse_solve]).

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``qpconvex2_sparse_init``.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    ivalue : int
        The `int` value associated with the keyword in `string`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            The initialization function ``qpconvex2_sparse_init`` has
            not been called.

        (`errno` 2)
            The supplied option string is invalid. Check that the
            keywords are neither ambiguous nor misspelt. The option
            string is *<value>*.
    """

    _return_scalar = _b_opt.qpconvex2_sparse_option_integer_get(
        string,
        comm,
        io_manager,
    )
    _ivalue_py = _return_scalar
    return _ivalue_py

def qpconvex2_sparse_option_double_get(string, comm, io_manager=None):
    u"""
    Get the setting of a real valued option of
    ``qpconvex2_sparse_solve``.

    ``qpconvex2_sparse_option_double_get`` is used to get the value of a
    `float` option. ``qpconvex2_sparse_option_double_get`` can be used
    before or after calling ``qpconvex2_sparse_solve``.

    For full information please refer to the NAG Library document for
    e04ny

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04nyf.html

    Parameters
    ----------
    string : str
        A single valid keyword of a `float` option (as described in
        [Other Parameters for qpconvex2_sparse_solve]).

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``qpconvex2_sparse_init``.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    rvalue : float
        The `float` value associated with the keyword in `string`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            The initialization function ``qpconvex2_sparse_init`` has
            not been called.

        (`errno` 2)
            The supplied option is invalid. Check that the keywords are
            neither ambiguous nor misspelt. The option string is
            *<value>*.
    """

    _return_scalar = _b_opt.qpconvex2_sparse_option_double_get(
        string,
        comm,
        io_manager,
    )
    _rvalue_py = _return_scalar
    return _rvalue_py

def bnd_lin_lsq(a, b, bl, bu, itype=1, tol=0.0):
    u"""
    Computes the least squares solution to a set of linear equations
    subject to fixed upper and lower bounds on the variables. An option
    is provided to return a minimal length solution if a solution is not
    unique.

    ``bnd_lin_lsq`` solves a linear least squares problem subject to
    fixed lower and upper bounds on the variables.

    For full information please refer to the NAG Library document for
    e04pc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04pcf.html

    Parameters
    ----------
    a : float, array-like, shape (m, n)
        The m by n matrix A.

    b : float, array-like, shape (m)
        The right-hand side vector b.

    bl : float, array-like, shape (n)
        `bl`[i-1] and `bu`[i-1] must specify the lower and upper bounds,
        l_i and u_i respectively, to be imposed on the solution vector
        x_i.

    bu : float, array-like, shape (n)
        `bl`[i-1] and `bu`[i-1] must specify the lower and upper bounds,
        l_i and u_i respectively, to be imposed on the solution vector
        x_i.

    itype : int, optional
        Provides the choice of returning a regularized solution if the
        matrix is not of full rank.

        `itype` = 0

            Specifies that a regularized solution is to be computed.

        `itype` = 1

            Specifies that no regularization is to take place.

    tol : float, optional
        `tol` specifies a parameter used to determine the relative
        linear dependence of a column vector for a variable moved from
        its initial value. It determines the computational rank of the
        matrix. Increasing its value from sqrt(machine precision) will
        increase the likelihood of additional elements of x being set to
        zero. It may be worth experimenting with increasing values of
        `tol` to determine whether the nature of the solution, x,
        changes significantly. In practice a value of sqrt(machine
        precision) is recommended (see ``machine.precision``).

        If on entry `tol` < sqrt(machine precision), sqrt(machine
        precision) is used.

    Returns
    -------
    a : float, ndarray, shape (m, n)
        If `itype` = 1, `a` contains the product matrix Q*A, where Q is
        an m by m orthogonal matrix generated by ``bnd_lin_lsq``;
        otherwise `a` is unchanged.

    b : float, ndarray, shape (m)
        If `itype` = 1, the product of Q times the original vector b,
        where Q is as described in argument `a`; otherwise `b` is
        unchanged.

    x : float, ndarray, shape (n)
        The solution vector x.

    rnorm : float
        The Euclidean norm of the residual vector b-Ax.

    nfree : int
        Indicates the number of components of the solution vector that
        are not at one of the constraints.

    w : float, ndarray, shape (n)
        Contains the dual solution vector. The magnitude of `w`[i-1]
        gives a measure of the improvement in the objective value if the
        corresponding bound were to be relaxed so that x_i could take
        different values.

        A value of `w`[i-1] equal to the special value -999.0 is
        indicative of the matrix A not having full rank.

        It is only likely to occur when `itype` = 1.

        However a matrix may have less than full rank without `w`[i-1]
        being set to -999.0.

        If `itype` = 1, then the values contained in `w` (other than
        those set to -999.0) may be unreliable; the corresponding values
        in `indx` may likewise be unreliable.

        If you have any doubts set `itype` = 0.

        Otherwise the values of `w`[i-1] have the following meaning:

        `w`[i-1] = 0

            if x_i is unconstrained.

        `w`[i-1] < 0

            if x_i is constrained by its lower bound.

        `w`[i-1] > 0

            if x_i is constrained by its upper bound.

        `w`[i-1]

            may be any value if l_i = u_i.

    indx : int, ndarray, shape (n)
        The contents of this array describe the components of the
        solution vector as follows:

        `indx`[i-1], for i = 1,2,...,`nfree`

            These elements of the solution have not hit a constraint;
            i.e., `w`[i-1] = 0.

        `indx`[i-1], for i = `nfree`+1,...,k

            These elements of the solution have been constrained by
            either the lower or upper bound.

        `indx`[i-1], for i = k+1,...,n

            These elements of the solution are fixed by the bounds;
            i.e., `bl`[i-1] = `bu`[i-1].

        Here k is determined from `nfree` and the number of fixed
        components. (Often the latter will be 0, so k will be
        n-`nfree`.)

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 0.

        (`errno` 1)
            On entry, m = *<value>*.

            Constraint: m >= 0.

        (`errno` 1)
            On entry, when i = *<value>*, `bl`[i-1] = *<value>* and
            `bu`[i-1] = *<value>*.

            Constraint: `bl`[i-1] <= `bu`[i-1].

        (`errno` 1)
            On entry, m = *<value>* and lda = *<value>*.

            Constraint: lda >= m.

    Warns
    -----
    NagAlgorithmicMajorWarning
        (`errno` 2)
            The function failed to converge in 3*n iterations. This is
            not expected. Please contact NAG.

    References
    ----------
    Lawson, C L and Hanson, R J, 1974, `Solving Least Squares Problems`,
    Prentice--Hall
    """

    fun_name = 'naginterfaces.library.opt.bnd_lin_lsq'

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[1]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _bl = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'bl'},
        bl,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _bu = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'bu'},
        bu,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _w = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _indx = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _return_scalars = _b_opt.bnd_lin_lsq(
            itype,
            _m,
            _n,
            _a,
            _b,
            _bl,
            _bu,
            tol,
            _x,
            _w,
            _indx,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _a_py = _a
    _b_py = _b
    _x_py = _x
    _rnorm_py = _return_scalars.rnorm
    _nfree_py = _return_scalars.nfree
    _w_py = _w
    _indx_py = _indx
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'b', 'x', 'rnorm', 'nfree', 'w', 'indx'],
    )(
        _a_py,
        _b_py,
        _x_py,
        _rnorm_py,
        _nfree_py,
        _w_py,
        _indx_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def handle_solve_socp_ipm(handle, x=None, u=None, uc=None, monit=None, data=None, io_manager=None):
    u"""
    Solve second-order cone programming (SOCP) and other convex related
    problems, such as, quadratically constrained quadratic programming
    (QCQP), quadratic programming (QP), sparse, interior point method
    (IPM).

    Note: this function uses optional algorithmic parameters, see also:
    ``handle_opt_set``, ``handle_opt_get``.

    ``handle_solve_socp_ipm`` is a solver from the NAG optimization
    modelling suite for large-scale second-order cone programming (SOCP)
    problems.
    It is based on an interior point method (IPM).

    For full information please refer to the NAG Library document for
    e04pt

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04ptf.html

    Parameters
    ----------
    handle : Handle
        The handle to the problem. It needs to be initialized by
        ``handle_init`` and the problem formulated by some of the
        functions ``handle_set_group``, ``handle_set_linobj``,
        ``handle_set_quadobj``, ``handle_set_simplebounds`` and
        ``handle_set_linconstr``. It **must not** be changed between
        calls to the NAG optimization modelling suite.

    x : None or float, array-like, shape (nvar), optional
        The input of `x` is reserved for future releases of the NAG
        Library and it is ignored at the moment.

    u : None or float, array-like, shape (nnzu), optional
        Note: if nnzu > 0, `u` holds Lagrange multipliers (dual
        variables) for the bound constraints and linear constraints. If
        nnzu = 0, `u` will not be referenced.

        The input of `u` is reserved for future releases of the NAG
        Library and it is ignored at the moment.

    uc : None or float, array-like, shape (nnzuc), optional
        Note: if nnzuc > 0, `uc` holds Lagrange multipliers (dual
        variables) for second-order cones as defined by
        ``handle_set_group``. If nnzuc = 0, `uc` will not be referenced.

        The input of `uc` is reserved for future releases of the NAG
        Library and it is ignored at the moment.

    monit : None or callable monit(handle, rinfo, stats, data=None),
        optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `monit` is provided to enable you to monitor the progress of the
        optimization.

        It is invoked at the end of every ith iteration where i is given
        by the option 'SOCP Monitor Frequency' (the default is 0,
        `monit` is not called).

        Parameters
        ~~~~~~~~~~
        handle : Handle
            The handle to the problem. It needs to be initialized by
            ``handle_init`` and **must not** be changed between calls to
            the NAG optimization modelling suite.

        rinfo : float, ndarray, shape (100)
            Error measures and various indicators at the end of the
            current iteration as described in `rinfo`.

        stats : float, ndarray, shape (100)
            Solver statistics at the end of the current iteration as
            described in `stats`, however, elements 3, 5, 9, 10 and 15
            refer to the quantities in the last iteration rather than
            accumulated over all iterations through the whole algorithm
            run.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

    data : arbitrary, optional
        User-communication data for callback functions.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    x : float, ndarray, shape (nvar)
        The final values of the variables x.

    u : float, ndarray, shape (nnzu)
        The final values of the variables u.

    uc : float, ndarray, shape (nnzuc)
        The final values of the variables uc.

    rinfo : float, ndarray, shape (100)
        Error measures and various indicators of the algorithm (see
        [Algorithmic Details] for details) as given in the table below:

        [table omitted]

    stats : float, ndarray, shape (100)
        Solver statistics as given in the table below. Note that times
        are measured in seconds, see option 'Stats Time'.

        [table omitted]

    Other Parameters
    ----------------
    'Defaults' : valueless
        This special keyword may be used to reset all options to their
        default values.
        Any argument value given with this keyword will be ignored.

    'Infinite Bound Size' : float
        Default = 10^20

        This defines the 'infinite' bound bigbnd in the definition of
        the problem constraints.
        Any upper bound greater than or equal to bigbnd will be regarded
        as + infinity (and similarly any lower bound less than or equal
        to -bigbnd will be regarded as - infinity).
        Note that a modification of this option does not influence
        constraints which have already been defined; only the
        constraints formulated after the change will be affected.

        Constraint: 'Infinite Bound Size' >= 1000.

    'Monitoring File' : int
        Default = -1

        If i >= 0, the unit number for the secondary (monitoring)
        output.
        If set to -1, no secondary output is provided.
        The following information is output to the unit:

        -   a listing of the options;

        -   problem statistics, the iteration log, and the final status
            as set by 'Monitoring Level';

        -   the solution if set by 'Print Solution'.

        Constraint: 'Monitoring File' >= -1.

    'Monitoring Level' : int
        Default = 4

        This argument sets the amount of information detail that will be
        printed by the solver to the secondary output.
        The meaning of the levels is the same as with 'Print Level'.

        Constraint: 0 <= 'Monitoring Level' <= 5.

    'Print File' : int
        Default = advisory message unit number

        If i >= 0, the unit number for the primary output of the solver.
        If 'Print File' = -1, the primary output is completely turned
        off independently of other settings. The default value is the
        advisory message unit number at the time of the options
        initialization, e.g., at the initialization of the handle. The
        following information is output to the unit:

        -   a listing of options if set by 'Print Options';

        -   problem statistics, the iteration log, and the final status
            from the solver as set by 'Print Level';

        -   the solution if set by 'Print Solution'.

        Constraint: 'Print File' >= -1.

    'Print Level' : int
        Default = 2

        This argument defines how detailed information should be printed
        by the solver to the primary output.

        [table omitted]

        Constraint: 0 <= 'Print Level' <= 5.

    'Print Options' : str
        Default = 'YES'

        If 'Print Options' = 'YES', a listing of options will be printed
        to the primary output.

        Constraint: 'Print Options' = 'YES' or 'NO'.

    'Print Solution' : str
        Default = 'NO'

        If 'Print Solution' = 'X', the final values of the primal
        variables are printed on the primary and secondary outputs.

        If 'Print Solution' = 'YES' or 'ALL', in addition to the primal
        variables, the final values of the dual variables are printed on
        the primary and secondary outputs.

        Constraint: 'Print Solution' = 'YES', 'NO', 'X' or 'ALL'.

    'SOCP Iteration Limit' : int
        Default = 100

        The maximum number of iterations to be performed by
        ``handle_solve_socp_ipm``.
        Setting the option too low might lead to `errno` = 22.

        Constraint: 'SOCP Iteration Limit' >= 1.

    'SOCP Monitor Frequency' : int
        Default = 0

        This argument defines the frequency of how often function
        `monit` is called.
        If i > 0, the solver calls `monit` at the end of every ith
        iteration.
        If it is set to 0, the function is not called at all.

        Constraint: 'SOCP Monitor Frequency' >= 0.

    'SOCP Presolve' : str
        Default = 'FULL'

        This argument allows you to reduce the level of presolving of
        the problem or turn it off completely.
        If the presolver is turned off, the solver will try to handle
        the problem as given by you.
        In such a case, the presence of fixed variables or linear
        dependencies in the constraint matrix can cause numerical
        instabilities to occur.
        In normal circumstances, it is recommended to use the full
        presolve which is the default.

        Constraint: 'SOCP Presolve' = 'FULL', 'BASIC' or 'NO'.

    'SOCP Scaling' : str
        Default = 'NONE'

        This argument controls the type of scaling to be applied on the
        constraint matrix A before solving the problem.
        More precisely, the scaling procedure will try to find diagonal
        matrices D_1 and D_2 such that the values in D_1 AD_2 are of a
        similar order of magnitude.
        The solver is less likely to run into numerical difficulties
        when the constraint matrix is well scaled.

        Constraint: 'SOCP Scaling' = 'ARITHMETIC', 'GEOMETRIC' or
        'NONE'.

    'SOCP Stop Tolerance' : float
        Default = sqrt(epsilon)

        This argument sets the value epsilon_1 which is the tolerance
        for the convergence measures in the stopping criteria, see
        [Stopping Criteria].

        Constraint: 'SOCP Stop Tolerance' > epsilon.

    'SOCP Stop Tolerance 2' : float
        Default = sqrt(epsilon)

        This argument sets the additional tolerance epsilon_2 used in
        the stopping criteria, see [Stopping Criteria].

        Constraint: 'SOCP Stop Tolerance 2' > epsilon.

    'SOCP System Formulation' : str
        Default = 'AUTO'

        As described in [Solving the KKT System],
        ``handle_solve_socp_ipm`` can internally work either with the
        normal equations formulation [equation] or with the augmented
        system [equation] and [equation].
        A brief discussion of advantages and disadvantages is presented
        in [equation].
        Setting the option value to 'AUTO' leaves the decision to the
        solver based on the structure of the constraints and it is the
        recommended value.
        This will typically lead to the normal equations formulation
        unless there are many dense columns or the system is
        significantly cheaper to factorize as the augmented system.
        Note that in some cases even if 'SOCP System Formulation' =
        'NORMAL EQUATIONS' the solver might switch the formulation
        through the computation to the augmented system due to numerical
        instabilities or computational cost.

        Constraint: 'SOCP System Formulation' = 'AUTO', 'AUGMENTED
        SYSTEM', 'AS', 'NORMAL EQUATIONS' or 'NE'.

    'Stats Time' : str
        Default = 'NO'

        This argument allows you to turn on timings of various parts of
        the algorithm to give a better overview of where most of the
        time is spent.
        This might be helpful for a choice of different solving
        approaches.
        It is possible to choose between CPU and wall clock time.
        Choice 'YES' is equivalent to 'WALL CLOCK'.

        Constraint: 'Stats Time' = 'YES', 'NO', 'CPU' or 'WALL CLOCK'.

    'Task' : str
        Default = 'MINIMIZE'

        This argument specifies the required direction of the
        optimization.
        If 'Task' = 'FEASIBLE POINT', the objective function (if set) is
        ignored and the algorithm stops as soon as a feasible point is
        found with respect to the given tolerance.
        If no objective function is set, 'Task' reverts to 'FEASIBLE
        POINT' automatically.

        Constraint: 'Task' = 'MINIMIZE', 'MAXIMIZE' or 'FEASIBLE POINT'.

    'Time Limit' : float
        Default = 10^6

        A limit to the number of seconds that the solver can use to
        solve one problem.
        If during the convergence check this limit is exceeded, the
        solver will terminate with `errno` = 23.

        Constraint: 'Time Limit' > 0.

    Raises
    ------
    NagValueError
        (`errno` 1)
            `handle` has not been initialized.

        (`errno` 1)
            `handle` does not belong to the NAG optimization modelling
            suite, has not been initialized properly or is corrupted.

        (`errno` 1)
            `handle` has not been initialized properly or is corrupted.

        (`errno` 2)
            This solver does not support the model defined in the
            handle.

        (`errno` 2)
            The problem is already being solved.

        (`errno` 4)
            On entry, nvar = *<value>*, expected value = *<value>*.

            Constraint: nvar must match the value given during
            initialization of `handle`.

        (`errno` 5)
            On entry, nnzu = *<value>*.

            nnzu does not match the size of the Lagrangian multipliers
            for constraints.

            The correct value is either 0 or *<value>*.

        (`errno` 5)
            On entry, nnzu = *<value>*.

            nnzu does not match the size of the Lagrangian multipliers
            for constraints.

            The correct value is 0 for no constraints.

        (`errno` 5)
            On entry, nnzuc = *<value>*.

            nnzuc does not match the size of the Lagrangian multipliers
            for second-order cone constraints.

            The correct value is either 0 or *<value>*.

        (`errno` 5)
            On entry, nnzuc = *<value>*.

            nnzuc does not match the size of the Lagrangian multipliers
            for second-order cone constraints.

            nnzuc = 0 when there are no second-order cone constraints.

        (`errno` 51)
            The problem was found to be primal infeasible.

        (`errno` 52)
            The problem was found to be dual infeasible.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 50)
            Suboptimal solution.

    NagAlgorithmicMajorWarning
        (`errno` 22)
            Maximum number of iterations exceeded.

        (`errno` 23)
            The solver terminated after the maximum time allowed was
            exceeded.

        (`errno` 24)
            No progress, stopping early.

    NagCallbackTerminateWarning
        (`errno` 20)
            User requested termination during a monitoring step.

    References
    ----------
    Alizadeh, F and Goldfarb, D, 2003, `Second-order cone programming`,
    Mathematical programming (95(1)), 3--51

    Andersen, E D, Roos, C and Terlaky, T, 2003, `On implementing a
    primal-dual interior-point method for conic quadratic optimization`,
    Mathematical programming (95(2)), 249--277

    Goldfarb, D and Scheinberg, K, 2005, `Product-form Cholesky
    factorization in interior point methods for second-order cone
    programming`, Mathematical programming (103(1)), 153--179

    Goldman, A J and Tucker, A W, 1956, `Theory of linear programming`,
    Linear inequalities and related systems (38), 53--97

    Hogg, J D and Scott, J A, 2011, `HSL MA97: a bit-compatible
    multifrontal code for sparse symmetric systems`, RAL Technical
    Report. RAL-TR-2011-024

    HSL, 2011, `A collection of Fortran codes for large-scale scientific
    computation`, http://www.hsl.rl.ac.uk/

    Karypis, G and Kumar, V, 1998, `A fast and high quality multilevel
    scheme for partitioning irregular graphs`, SIAM J. Sci. Comput.
    (20(1)), 359--392

    Lobo, M S, Vandenberghe, L, Boyd, S and Levret, H, 1998,
    `Applications of second-order cone programming`, Linear Algebra and
    its Applications (284(1-3)), 193--228

    Lustig, I J, Marsten, R E and Shanno, D F, 1992, `On implementing
    Mehrotra's predictor--corrector interior-point method for linear
    programming`, SIAM J. Optim. (2(3)), 435--449

    Mehrotra, S, 1992, `On the implementation of a primal-dual interior
    point method`, SIAM J. Optim. (2), 575--601

    Nesterov, Y E and Todd, M J, 1997, `Self-scaled barriers and
    interior-point methods for convex programming`, Mathematics of
    Operations research (22(1)), 1--42

    Nesterov, Y E and Todd, M J, 1998, `Primal-dual interior-point
    methods for self-scaled cones`, SIAM J. Optim. (8(2)), 324--364

    Nocedal, J and Wright, S J, 2006, `Numerical Optimization`, (2nd
    Edition), Springer Series in Operations Research, Springer, New York

    Sturm, J F, 2002, `Implementation of Interior Point Methods for
    Mixed Semidefinite and Second Order Cone Optimization Problems`,
    Optimization Methods and Software (17(6)), 151--171

    Xu, X, Hung, P-F and Ye, Y, 1996, `A simplified homogeneous and
    self-dual linear programming algorithm and its implementation`,
    Annals of Operations Research (62(1)), 151--171

    See Also
    --------
    :meth:`naginterfaces.library.examples.opt.handle_solve_socp_ipm_ex.main`
    """

    def cb_monit(handle, rinfo, stats, data=None):
        """Wrapping user's monit. Called as base monit."""
        if data is not None:
            monit(
                handle,
                rinfo,
                stats,
                data,
            )
        else:
            monit(
                handle,
                rinfo,
                stats,
            )

    _nvar, _nnzu, _nnzuc, _nnzua = _b_utils._e04ptan(handle)
    _nvar_py = _b_utils._EngineIntScalarType.to_py(_nvar)
    _vl_e04pta_nvar_py = _nvar_py
    fun_name = 'naginterfaces.library.opt.handle_solve_socp_ipm'

    _nvar_py = _b_utils._EngineIntScalarType.to_py(_nvar)
    if x is None:
        _x = _b_utils._EngineFloat64ArrayType.empty_ndarray(
            (_nvar_py,),
        )
    else:
        _x = _utils._to_ndarray(
            {'fun_name': fun_name, 'entity': 'x'},
            x,
            copy=True,
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
    _nnzu_py = _b_utils._EngineIntScalarType.to_py(_nnzu)
    _vl_e04pta_nnzu_py = _nnzu_py
    _nnzu_py = _b_utils._EngineIntScalarType.to_py(_nnzu)
    if u is None:
        _u = _b_utils._EngineFloat64ArrayType.empty_ndarray(
            (_nnzu_py,),
        )
    else:
        _u = _utils._to_ndarray(
            {'fun_name': fun_name, 'entity': 'u'},
            u,
            copy=True,
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
    _nnzuc_py = _b_utils._EngineIntScalarType.to_py(_nnzuc)
    _vl_e04pta_nnzuc_py = _nnzuc_py
    _nnzuc_py = _b_utils._EngineIntScalarType.to_py(_nnzuc)
    if uc is None:
        _uc = _b_utils._EngineFloat64ArrayType.empty_ndarray(
            (_nnzuc_py,),
        )
    else:
        _uc = _utils._to_ndarray(
            {'fun_name': fun_name, 'entity': 'uc'},
            uc,
            copy=True,
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
    _rinfo = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (100,),
    )
    _stats = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (100,),
    )
    _b_utils._check_callable(monit, {'fun_name': fun_name, 'entity': 'monit'}, allow_none=True)
    _monit = None if monit is None else cb_monit
    try:
        _b_opt.handle_solve_socp_ipm(
            handle,
            _nvar,
            _x,
            _nnzu,
            _u,
            _nnzuc,
            _uc,
            _rinfo,
            _stats,
            _monit,
            data,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _x_py = _x
    _u_py = _u
    _uc_py = _uc
    _rinfo_py = _rinfo
    _stats_py = _stats
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'u', 'uc', 'rinfo', 'stats'],
    )(
        _x_py,
        _u_py,
        _uc_py,
        _rinfo_py,
        _stats_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def handle_init(nvar):
    u"""
    Initialization of a handle for the NAG optimization modelling suite
    for problems, such as, linear programming (LP), quadratic
    programming (QP), nonlinear programming (NLP), least squares (LSQ)
    problems, linear semidefinite programming (SDP) or SDP with bilinear
    matrix inequalities (BMI-SDP).

    ``handle_init`` initializes a data structure for the NAG
    optimization modelling suite for problems such as, linear
    programming (LP), quadratic programming (QP), nonlinear programming
    (NLP), least squares (LSQ) problems, second-order cone programming
    (SOCP), linear semidefinite programming (SDP) and semidefinite
    programming with bilinear matrix inequalities (BMI-SDP).

    For full information please refer to the NAG Library document for
    e04ra

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04raf.html

    Parameters
    ----------
    nvar : int
        n, the number of decision variables in the problem.

    Returns
    -------
    handle : Handle
        Holds a handle to the internal data structure where an empty
        problem with `nvar` variables is defined.

    Raises
    ------
    NagValueError
        (`errno` 6)
            On entry, `nvar` = *<value>*.

            Constraint: `nvar` > 0.

    See Also
    --------
    :meth:`naginterfaces.library.examples.opt.handle_solve_bounds_foas_ex.main`
    :meth:`naginterfaces.library.examples.opt.handle_solve_dfls_ex.main`
    :meth:`naginterfaces.library.examples.opt.handle_solve_dfno_ex.main`
    :meth:`naginterfaces.library.examples.opt.handle_solve_ipopt_ex.main`
    :meth:`naginterfaces.library.examples.opt.handle_solve_lp_ipm_ex.main`
    :meth:`naginterfaces.library.examples.opt.handle_solve_pennon_bmi_ex.main`
    :meth:`naginterfaces.library.examples.opt.handle_solve_pennon_lmi_ex.main`
    :meth:`naginterfaces.library.examples.opt.handle_solve_socp_ipm_ex.main`
    """

    _return_scalar = _b_opt.handle_init(
        nvar,
    )
    _handle_py = _return_scalar
    return _handle_py

def handle_set_group(handle, gtype, idgroup, group=None):
    u"""
    Define a set of variables which form a second-order cone to a
    problem initialized by ``handle_init``.

    ``handle_set_group`` is a part of the NAG optimization modelling
    suite and modifies a model by either adding a new, or replacing or
    deleting an existing, quadratic or rotated quadratic cone
    constraint.

    For full information please refer to the NAG Library document for
    e04rb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04rbf.html

    Parameters
    ----------
    handle : Handle
        The handle to the problem. It needs to be initialized by
        ``handle_init`` and **must not** be changed before the call to
        ``handle_set_group``.

    gtype : str
        The type of the cone constraint, case insensitive.

        `gtype` = 'QUAD' or 'Q'

            The group defines a quadratic cone.

        `gtype` = 'RQUAD' or 'R'

            The group defines a rotated quadratic cone.

    idgroup : int
        `idgroup` = 0

            A new cone constraint is created.

        `idgroup` > 0

            i, the ID number of the existing constraint to be deleted or
            replaced.

    group : None or int, array-like, shape (lgroup), optional
        G^i, the indices of the variables in the constraint. If lgroup =
        0, `group` is not referenced.

    Returns
    -------
    idgroup : int
        If `idgroup` = 0 on entry, the ID number of the new cone
        constraint is returned. By definition, this is the number of the
        cone constraints already defined plus one. Otherwise, `idgroup`
        stays unchanged.

    Raises
    ------
    NagValueError
        (`errno` 1)
            `handle` has not been initialized.

        (`errno` 1)
            `handle` does not belong to the NAG optimization modelling
            suite, has not been initialized properly or is corrupted.

        (`errno` 1)
            `handle` has not been initialized properly or is corrupted.

        (`errno` 2)
            The problem cannot be modified in this phase any more, the
            solver has already been called.

        (`errno` 3)
            On entry, variable with index i = *<value>* has been defined
            in a cone in a previous call to this function.

            Constraint: each variable may be defined in one cone
            constraint at most.

        (`errno` 4)
            On entry, `idgroup` = *<value>*.

            The given `idgroup` does not match with any cone constraint
            already defined.

        (`errno` 5)
            On entry, lgroup = *<value>*.

            Constraint: lgroup >= 0.

        (`errno` 5)
            On entry, `gtype` = *<value>* and lgroup = *<value>*.

            Constraint: if `gtype` = 'QUAD' or 'Q', lgroup = 0 or lgroup
            >= 2.

        (`errno` 5)
            On entry, `gtype` = *<value>* and lgroup = *<value>*.

            Constraint: if `gtype` = 'RQUAD' or 'R', lgroup = 0 or
            lgroup >= 3.

        (`errno` 6)
            On entry, `gtype` = *<value>*.

            Constraint: `gtype` = 'QUAD', 'Q', 'RQUAD' or 'R'.

        (`errno` 7)
            On entry, `idgroup` = *<value>*.

            Constraint: `idgroup` >= 0.

        (`errno` 8)
            On entry, k = *<value>*, `group`[k-1] = *<value>* and n =
            *<value>*.

            Constraint: 1 <= `group`[k-1] <= n.

        (`errno` 9)
            On entry, `group`[i-1] = `group`[j-1] = *<value>* for i =
            *<value>* and j = *<value>*.

            Constraint: elements in `group` cannot repeat.

    See Also
    --------
    :meth:`naginterfaces.library.examples.opt.handle_solve_socp_ipm_ex.main`
    """

    fun_name = 'naginterfaces.library.opt.handle_set_group'

    _group = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'group'},
        group,
        copy=False,
        allow_none=True,
    )
    if group is None:
        _lgroup = 0
        _lgroup_py = _lgroup
    else:
        _lgroup_py = _b_utils._EngineIntArrayType.get_shape(_group, {'fun_name': fun_name, 'entity': 'group'}, exp_rank=1, allow_none=True)[0]
        _lgroup = _b_utils._EngineIntScalarType(
            _lgroup_py, {'fun_name': fun_name, 'entity': '_lgroup'},
        ).to_ctypes()
    _return_scalar = _b_opt.handle_set_group(
        handle,
        gtype,
        _lgroup,
        _group,
        idgroup,
    )
    _idgroup_py = _return_scalar
    return _idgroup_py

def sdp_read_sdpa(infile, maxnvar, maxnblk, maxnnz, filelst=0, io_manager=None):
    u"""
    A reader of sparse SDPA data files for linear SDP problems.

    ``sdp_read_sdpa`` reads in a linear semidefinite programming problem
    (SDP) from a file in sparse SDPA format and returns it in the form
    which is usable by functions ``handle_init`` (initialization),
    ``handle_set_linobj`` (linear objective function),
    ``handle_set_linmatineq`` (linear matrix constraints),
    ``handle_solve_pennon`` (solver) and ``handle_free`` (deallocation)
    from the NAG optimization modelling suite.

    For full information please refer to the NAG Library document for
    e04rd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04rdf.html

    Parameters
    ----------
    infile : int
        The unit number (see
        :meth:`~naginterfaces.base.utils.FileObjManager.unit_from_fileobj`)
        associated with the sparse SDPA data file. **Note:** that the
        file needs to be opened in read mode.

    maxnvar : int
        The upper limit for the number of variables in the problem. If
        it is set to zero, `cvec` and `nnza` will not be referenced.

    maxnblk : int
        The upper limit for the number of matrix constraints (i.e., the
        number of diagonal blocks within the matrix). If it is set to
        zero, `blksizea` will not be referenced.

    maxnnz : int
        The upper limit on the sum of nonzeros in all matrices A_i^k,
        for k = 1,2,...,`nblk`, for i = 0,1,...,`nvar`. If it is set to
        zero, `irowa`, `icola` and `a` will not be referenced.

    filelst : int, optional
        If `filelst` != 0, a listing of the input data is sent to the
        file object associated with the advisory I/O unit (see
        :class:`~naginterfaces.base.utils.FileObjManager`). This can be
        useful for debugging the data file.

        If `filelst` = 0, no listing is produced.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    nvar : int
        The actual number of the variables n, matrix constraints m_A and
        number of nonzeros of the problem in the file. This also
        indicates the exact memory needed in `cvec`, `nnza`, `irowa`,
        `icola`, `a` and `blksizea`.

    nblk : int
        The actual number of the variables n, matrix constraints m_A and
        number of nonzeros of the problem in the file. This also
        indicates the exact memory needed in `cvec`, `nnza`, `irowa`,
        `icola`, `a` and `blksizea`.

    nnz : int
        The actual number of the variables n, matrix constraints m_A and
        number of nonzeros of the problem in the file. This also
        indicates the exact memory needed in `cvec`, `nnza`, `irowa`,
        `icola`, `a` and `blksizea`.

    cvec : float, ndarray, shape (`maxnvar`)
        `cvec`[i-1], for i = 1,2,...,`nvar`, stores the dense vector c
        of the linear objective function.

    nnza : int, ndarray, shape (`maxnvar`+1)
        `nnza`[i], for i = 0,1,...,`nvar`, stores the number of nonzero
        elements in matrices A_i.

    irowa : int, ndarray, shape (`maxnnz`)
        `irowa`, `icola` and `a` store the nonzeros in the upper
        triangle of matrices A_i, for i = 0,1,...,`nvar`, in the
        coordinate storage, i.e., `irowa`[j-1] are one-based row
        indices, `icola`[j-1] are one-based column indices and `a`[j-1]
        are the values of the nonzero elements, for j = 1,2,...,`nnz`.
        See [Further Comments].

    icola : int, ndarray, shape (`maxnnz`)
        `irowa`, `icola` and `a` store the nonzeros in the upper
        triangle of matrices A_i, for i = 0,1,...,`nvar`, in the
        coordinate storage, i.e., `irowa`[j-1] are one-based row
        indices, `icola`[j-1] are one-based column indices and `a`[j-1]
        are the values of the nonzero elements, for j = 1,2,...,`nnz`.
        See [Further Comments].

    a : float, ndarray, shape (`maxnnz`)
        `irowa`, `icola` and `a` store the nonzeros in the upper
        triangle of matrices A_i, for i = 0,1,...,`nvar`, in the
        coordinate storage, i.e., `irowa`[j-1] are one-based row
        indices, `icola`[j-1] are one-based column indices and `a`[j-1]
        are the values of the nonzero elements, for j = 1,2,...,`nnz`.
        See [Further Comments].

    blksizea : int, ndarray, shape (`maxnblk`)
        `blksizea`[k-1], for k = 1,2,...,`nblk`, stores the sizes of the
        diagonal blocks in matrices A_i from the top to the bottom.

    Raises
    ------
    NagValueError
        (`errno` 2)
            The token on line *<value>* at position *<value>* to
            *<value>* was not recognized as a valid integer.

        (`errno` 3)
            The token on line *<value>* at position *<value>* to
            *<value>* was not recognized as a valid real number.

        (`errno` 4)
            The token on line *<value>* starting at position *<value>*
            was too long and was not recognized.

        (`errno` 5)
            An invalid number of variables was given on line *<value>*.

            The number stated there is *<value>* and needs to be at
            least 1.

        (`errno` 6)
            An invalid number of blocks was given on line *<value>*.

            The number stated there is *<value>* and needs to be at
            least 1.

        (`errno` 7)
            An invalid size of the block number *<value>* was given on
            line *<value>*.

            The number stated there is *<value>* and needs to be
            nonzero.

        (`errno` 8)
            Not enough data was given on line *<value>* specifying block
            sizes.

            Expected *<value>* tokens but found only *<value>*.

        (`errno` 9)
            Not enough data was given on line *<value>* specifying the
            objective function.

            Expected *<value>* tokens but found only *<value>*.

        (`errno` 10)
            Not enough data was given on line *<value>* specifying
            nonzero matrix elements.

            Expected *<value>* tokens but found only *<value>*.

        (`errno` 11)
            Invalid structural data found on line *<value>*.

            The given matrix number is out of bounds. Its value
            *<value>* must be between *<value>* and *<value>*
            (inclusive).

        (`errno` 12)
            Invalid structural data found on line *<value>*.

            The given block number is out of bounds. Its value *<value>*
            must be between *<value>* and *<value>* (inclusive).

        (`errno` 13)
            Invalid structural data found on line *<value>*.

            The given row index is out of bounds, it must respect the
            size of the block. Its value *<value>* must be between
            *<value>* and *<value>* (inclusive).

        (`errno` 14)
            Invalid structural data found on line *<value>*.

            The given column index is out of bounds, it must respect the
            size of the block. Its value *<value>* must be between
            *<value>* and *<value>* (inclusive).

        (`errno` 15)
            Invalid structural data found on line *<value>*.

            The specified nonzero element is not in the upper triangle.

            The row index is *<value>* and column index is *<value>*.

        (`errno` 16)
            Invalid structural data found on line *<value>*.

            The specified element belongs to a diagonal block but is not
            diagonal.

            The row index is *<value>* and column index is *<value>*.

        (`errno` 17)
            An entry in the constraints with matno = *<value>*, blkno =
            *<value>*, row index *<value>* and column index *<value>*
            was defined more than once. All entries need to be unique.

        (`errno` 18)
            A premature end of the input stream. The part defining the
            number of variables was not found.

        (`errno` 18)
            A premature end of the input stream. The part defining the
            number of blocks was not found.

        (`errno` 18)
            A premature end of the input stream. The part defining the
            dimensions of the blocks was not found.

        (`errno` 18)
            A premature end of the input stream. The part defining the
            objective function was not found.

        (`errno` 18)
            A premature end of the input stream. The part defining the
            nonzero entries was not found.

        (`errno` 19)
            The input stream seems to be empty. No data was read.

        (`errno` 20)
            Reading from the stream caused an unknown error on line
            *<value>*.

        (`errno` 21)
            On entry, `infile` = *<value>*.

            Constraint: `infile` >= 0.

        (`errno` 21)
            On entry, `maxnvar` = *<value>*.

            Constraint: `maxnvar` >= 0.

        (`errno` 21)
            On entry, `maxnblk` = *<value>*.

            Constraint: `maxnblk` >= 0.

        (`errno` 21)
            On entry, `maxnnz` = *<value>*.

            Constraint: `maxnnz` >= 0.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            At least one of `maxnvar`, `maxnblk` or `maxnnz` is too
            small.

            `maxnvar` should be at least *<value>*, was *<value>*.

            `maxnblk` should be at least *<value>*, was *<value>*.

            `maxnnz` should be at least *<value>*, was *<value>*.

    References
    ----------
    Borchers, B, 1999, `SDPLIB 1.2, A Library of semidefinite
    programming test problems`, Optimization Methods and Software
    (11(1)), 683--690, http://euler.nmt.edu/~brian/sdplib/

    Fujisawa, K, Kojima, M and Nakata, K, 1998, `SDPA (Semidefinite
    Programming Algorithm) User's Manual`, Technical Report B-308,
    Department of Mathematical and Computing Sciences, Tokyo Institute
    of Technology.
    """

    fun_name = 'naginterfaces.library.opt.sdp_read_sdpa'

    _maxnvar = _b_utils._EngineIntScalarType(
        maxnvar, {'fun_name': fun_name, 'entity': 'maxnvar'},
    ).to_ctypes()
    _maxnvar_py = _b_utils._EngineIntScalarType.to_py(_maxnvar)
    _cvec = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_maxnvar_py,),
    )
    _nnza = _b_utils._EngineIntArrayType.empty_ndarray(
        (_maxnvar_py+1,),
    )
    _maxnnz = _b_utils._EngineIntScalarType(
        maxnnz, {'fun_name': fun_name, 'entity': 'maxnnz'},
    ).to_ctypes()
    _maxnnz_py = _b_utils._EngineIntScalarType.to_py(_maxnnz)
    _irowa = _b_utils._EngineIntArrayType.empty_ndarray(
        (_maxnnz_py,),
    )
    _icola = _b_utils._EngineIntArrayType.empty_ndarray(
        (_maxnnz_py,),
    )
    _a = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_maxnnz_py,),
    )
    _maxnblk = _b_utils._EngineIntScalarType(
        maxnblk, {'fun_name': fun_name, 'entity': 'maxnblk'},
    ).to_ctypes()
    _maxnblk_py = _b_utils._EngineIntScalarType.to_py(_maxnblk)
    _blksizea = _b_utils._EngineIntArrayType.empty_ndarray(
        (_maxnblk_py,),
    )
    try:
        _return_scalars = _b_opt.sdp_read_sdpa(
            infile,
            _maxnvar,
            _maxnblk,
            _maxnnz,
            filelst,
            _cvec,
            _nnza,
            _irowa,
            _icola,
            _a,
            _blksizea,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _nvar_py = _return_scalars.nvar
    _nblk_py = _return_scalars.nblk
    _nnz_py = _return_scalars.nnz
    _cvec_py = _cvec
    _nnza_py = _nnza
    _irowa_py = _irowa
    _icola_py = _icola
    _a_py = _a
    _blksizea_py = _blksizea
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['nvar', 'nblk', 'nnz', 'cvec', 'nnza', 'irowa', 'icola', 'a', 'blksizea'],
    )(
        _nvar_py,
        _nblk_py,
        _nnz_py,
        _cvec_py,
        _nnza_py,
        _irowa_py,
        _icola_py,
        _a_py,
        _blksizea_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def handle_set_linobj(handle, cvec):
    u"""
    Define a linear objective function to a problem initialized by
    ``handle_init``.

    ``handle_set_linobj`` is a part of the NAG optimization modelling
    suite and defines the linear objective function of the problem.

    For full information please refer to the NAG Library document for
    e04re

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04ref.html

    Parameters
    ----------
    handle : Handle
        The handle to the problem. It needs to be initialized by
        ``handle_init`` and **must not** be changed before the call to
        ``handle_set_linobj``.

    cvec : float, array-like, shape (nvar)
        The dense vector c of the objective function.

    Raises
    ------
    NagValueError
        (`errno` 1)
            `handle` has not been initialized.

        (`errno` 1)
            `handle` does not belong to the NAG optimization modelling
            suite, has not been initialized properly or is corrupted.

        (`errno` 1)
            `handle` has not been initialized properly or is corrupted.

        (`errno` 2)
            The problem cannot be modified in this phase any more, the
            solver has already been called.

        (`errno` 3)
            The objective function has already been defined.

        (`errno` 4)
            On entry, nvar = *<value>*, expected value = *<value>*.

            Constraint: nvar must match the value given during
            initialization of `handle`.

    See Also
    --------
    :meth:`naginterfaces.library.examples.opt.handle_solve_pennon_bmi_ex.main`
    :meth:`naginterfaces.library.examples.opt.handle_solve_socp_ipm_ex.main`
    """

    fun_name = 'naginterfaces.library.opt.handle_set_linobj'

    _cvec = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'cvec'},
        cvec,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nvar_py = _b_utils._EngineFloat64ArrayType.get_shape(_cvec, {'fun_name': fun_name, 'entity': 'cvec'}, exp_rank=1)[0]
    _nvar = _b_utils._EngineIntScalarType(
        _nvar_py, {'fun_name': fun_name, 'entity': '_nvar'},
    ).to_ctypes()
    _b_opt.handle_set_linobj(
        handle,
        _nvar,
        _cvec,
    )

def handle_set_quadobj(handle, idxc=None, c=None, irowh=None, icolh=None, h=None):
    u"""
    Define a linear or a quadratic objective function to a problem
    initialized by ``handle_init``.

    ``handle_set_quadobj`` is a part of the NAG optimization modelling
    suite and defines the linear or the quadratic objective function of
    the problem.

    For full information please refer to the NAG Library document for
    e04rf

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04rff.html

    Parameters
    ----------
    handle : Handle
        The handle to the problem. It needs to be initialized by
        ``handle_init`` and **must not** be changed before the call to
        ``handle_set_quadobj``.

    idxc : None or int, array-like, shape (nnzc), optional
        The nonzero elements of the sparse vector c. `idxc`[i-1] must
        contain the index of `c`[i-1] in the vector, for i =
        1,2,...,nnzc. The elements are stored in ascending order. Note
        that n, the number of variables in the problem, was set in nvar
        during the initialization of the handle by ``handle_init``.

    c : None or float, array-like, shape (nnzc), optional
        The nonzero elements of the sparse vector c. `idxc`[i-1] must
        contain the index of `c`[i-1] in the vector, for i =
        1,2,...,nnzc. The elements are stored in ascending order. Note
        that n, the number of variables in the problem, was set in nvar
        during the initialization of the handle by ``handle_init``.

    irowh : None or int, array-like, shape (nnzh), optional
        Arrays `irowh`, `icolh` and `h` store the nonzeros of the upper
        triangle of the matrix H in coordinate storage (CS) format (see
        [the F11 Introduction]). `irowh` specifies one-based row
        indices, `icolh` specifies one-based column indices and `h`
        specifies the values of the nonzero elements in such a way that
        h_{ij} = `h`[l-1] where i = `irowh`[l-1], j = `icolh`[l-1], for
        l = 1,2,...,nnzh. No particular order is expected, but elements
        should not repeat.

    icolh : None or int, array-like, shape (nnzh), optional
        Arrays `irowh`, `icolh` and `h` store the nonzeros of the upper
        triangle of the matrix H in coordinate storage (CS) format (see
        [the F11 Introduction]). `irowh` specifies one-based row
        indices, `icolh` specifies one-based column indices and `h`
        specifies the values of the nonzero elements in such a way that
        h_{ij} = `h`[l-1] where i = `irowh`[l-1], j = `icolh`[l-1], for
        l = 1,2,...,nnzh. No particular order is expected, but elements
        should not repeat.

    h : None or float, array-like, shape (nnzh), optional
        Arrays `irowh`, `icolh` and `h` store the nonzeros of the upper
        triangle of the matrix H in coordinate storage (CS) format (see
        [the F11 Introduction]). `irowh` specifies one-based row
        indices, `icolh` specifies one-based column indices and `h`
        specifies the values of the nonzero elements in such a way that
        h_{ij} = `h`[l-1] where i = `irowh`[l-1], j = `icolh`[l-1], for
        l = 1,2,...,nnzh. No particular order is expected, but elements
        should not repeat.

    Raises
    ------
    NagValueError
        (`errno` 1)
            `handle` has not been initialized.

        (`errno` 1)
            `handle` does not belong to the NAG optimization modelling
            suite, has not been initialized properly or is corrupted.

        (`errno` 1)
            `handle` has not been initialized properly or is corrupted.

        (`errno` 2)
            The problem cannot be modified in this phase any more, the
            solver has already been called.

        (`errno` 3)
            The objective function has already been defined.

        (`errno` 6)
            On entry, nnzh = *<value>*.

            Constraint: nnzh >= 0.

        (`errno` 6)
            On entry, nnzc = *<value>*.

            Constraint: nnzc >= 0.

        (`errno` 7)
            On entry, i = *<value>*, `idxc`[i-1] = *<value>* and
            `idxc`[i] = *<value>*.

            Constraint: `idxc`[i-1] < `idxc`[i] (ascending order).

        (`errno` 7)
            On entry, i = *<value>*, `idxc`[i-1] = *<value>* and n =
            *<value>*.

            Constraint: 1 <= `idxc`[i-1] <= n.

        (`errno` 8)
            On entry, i = *<value>*, `irowh`[i-1] = *<value>* and n =
            *<value>*.

            Constraint: 1 <= `irowh`[i-1] <= n.

        (`errno` 8)
            On entry, i = *<value>*, `icolh`[i-1] = *<value>* and n =
            *<value>*.

            Constraint: 1 <= `icolh`[i-1] <= n.

        (`errno` 8)
            On entry, i = *<value>*, `irowh`[i-1] = *<value>* and
            `icolh`[i-1] = *<value>*.

            Constraint: `irowh`[i-1] <= `icolh`[i-1] (elements within
            the upper triangle).

        (`errno` 8)
            On entry, more than one element of `h` has row index
            *<value>* and column index *<value>*.

            Constraint: each element of `h` must have a unique row and
            column index.

    See Also
    --------
    :meth:`naginterfaces.library.examples.opt.handle_solve_lp_ipm_ex.main`
    :meth:`naginterfaces.library.examples.opt.handle_solve_pennon_lmi_ex.main`
    """

    fun_name = 'naginterfaces.library.opt.handle_set_quadobj'

    _idxc = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'idxc'},
        idxc,
        copy=False,
        allow_none=True,
    )
    _nnzc = _b_utils._EngineIntScalarType(
        0 if idxc is None else _b_utils._EngineIntArrayType.get_shape(_idxc, {'fun_name': fun_name, 'entity': 'idxc'}, exp_rank=1, allow_none=True)[0], {'fun_name': fun_name, 'entity': '_nnzc'},
    ).to_ctypes()
    _c = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
        copy=False,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _irowh = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'irowh'},
        irowh,
        copy=False,
        allow_none=True,
    )
    _nnzh = _b_utils._EngineIntScalarType(
        0 if irowh is None else _b_utils._EngineIntArrayType.get_shape(_irowh, {'fun_name': fun_name, 'entity': 'irowh'}, exp_rank=1, allow_none=True)[0], {'fun_name': fun_name, 'entity': '_nnzh'},
    ).to_ctypes()
    _icolh = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'icolh'},
        icolh,
        copy=False,
        allow_none=True,
    )
    _h = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'h'},
        h,
        copy=False,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b_opt.handle_set_quadobj(
        handle,
        _nnzc,
        _idxc,
        _c,
        _nnzh,
        _irowh,
        _icolh,
        _h,
    )

def handle_set_nlnobj(handle, idxfd=None):
    u"""
    Define a nonlinear objective function to a problem initialized by
    ``handle_init``.

    ``handle_set_nlnobj`` is a part of the NAG optimization modelling
    suite and declares the objective function of the problem as a
    nonlinear function with a particular gradient sparsity structure.

    For full information please refer to the NAG Library document for
    e04rg

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04rgf.html

    Parameters
    ----------
    handle : Handle
        The handle to the problem. It needs to be initialized by
        ``handle_init`` and **must not** be changed before the call to
        ``handle_set_nlnobj``.

    idxfd : None or int, array-like, shape (nnzfd), optional
        The one-based indices of the nonzero elements of the sparse
        gradient vector. The indices must be stored in ascending order.
        Note that n, the number of decision variables in the problem,
        was set in nvar during the initialization of the handle by
        ``handle_init``.

        If nnzfd = 0, the objective is assumed to be zero and the array
        `idxfd` will not be referenced.

    Raises
    ------
    NagValueError
        (`errno` 1)
            `handle` has not been initialized.

        (`errno` 1)
            `handle` does not belong to the NAG optimization modelling
            suite, has not been initialized properly or is corrupted.

        (`errno` 1)
            `handle` has not been initialized properly or is corrupted.

        (`errno` 2)
            The problem cannot be modified in this phase any more, the
            solver has already been called.

        (`errno` 2)
            The Hessians of nonlinear functions have already been
            defined, a nonlinear objective cannot be added.

        (`errno` 3)
            The objective function has already been defined.

        (`errno` 6)
            On entry, nnzfd = *<value>*.

            Constraint: nnzfd >= 0.

        (`errno` 7)
            On entry, i = *<value>*, `idxfd`[i-1] = *<value>* and
            `idxfd`[i] = *<value>*.

            Constraint: `idxfd`[i-1] < `idxfd`[i] (ascending order).

        (`errno` 7)
            On entry, i = *<value>*, `idxfd`[i-1] = *<value>* and n =
            *<value>*.

            Constraint: 1 <= `idxfd`[i-1] <= n.

    See Also
    --------
    :meth:`naginterfaces.library.examples.opt.handle_solve_bounds_foas_ex.main`
    :meth:`naginterfaces.library.examples.opt.handle_solve_dfno_ex.main`
    :meth:`naginterfaces.library.examples.opt.handle_solve_ipopt_ex.main`
    """

    fun_name = 'naginterfaces.library.opt.handle_set_nlnobj'

    _idxfd = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'idxfd'},
        idxfd,
        copy=False,
        allow_none=True,
    )
    _nnzfd = _b_utils._EngineIntScalarType(
        0 if idxfd is None else _b_utils._EngineIntArrayType.get_shape(_idxfd, {'fun_name': fun_name, 'entity': 'idxfd'}, exp_rank=1, allow_none=True)[0], {'fun_name': fun_name, 'entity': '_nnzfd'},
    ).to_ctypes()
    _b_opt.handle_set_nlnobj(
        handle,
        _nnzfd,
        _idxfd,
    )

def handle_set_simplebounds(handle, bl, bu):
    u"""
    Define bounds of variables of a problem initialized by
    ``handle_init``.

    ``handle_set_simplebounds`` is a part of the NAG optimization
    modelling suite and defines bounds on the variables of the problem.

    For full information please refer to the NAG Library document for
    e04rh

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04rhf.html

    Parameters
    ----------
    handle : Handle
        The handle to the problem. It needs to be initialized by
        ``handle_init`` and **must not** be changed before the call to
        ``handle_set_simplebounds``.

    bl : float, array-like, shape (nvar)
        `bl` defines the lower bounds on the variables

    bu : float, array-like, shape (nvar)
        `bu` defines the upper bounds on the variables

    Raises
    ------
    NagValueError
        (`errno` 1)
            `handle` has not been initialized.

        (`errno` 1)
            `handle` does not belong to the NAG optimization modelling
            suite, has not been initialized properly or is corrupted.

        (`errno` 1)
            `handle` has not been initialized properly or is corrupted.

        (`errno` 2)
            The problem cannot be modified in this phase any more, the
            solver has already been called.

        (`errno` 3)
            Variable bounds have already been defined.

        (`errno` 4)
            On entry, nvar = *<value>*, expected value = *<value>*.

            Constraint: nvar must match the value given during
            initialization of `handle`.

        (`errno` 10)
            On entry, j = *<value>*, `bl`[j-1] = *<value>* and `bu`[j-1]
            = *<value>*.

            Constraint: `bl`[j-1] <= `bu`[j-1].

        (`errno` 10)
            On entry, j = *<value>*, `bl`[j-1] = *<value>*, bigbnd =
            *<value>*.

            Constraint: `bl`[j-1] < bigbnd.

        (`errno` 10)
            On entry, j = *<value>*, `bu`[j-1] = *<value>*, bigbnd =
            *<value>*.

            Constraint: `bu`[j-1] > -bigbnd.

    References
    ----------
    Candes, E and Recht, B, 2009, `Exact matrix completion via convex
    optimization`, Foundations of Computation Mathematics (Volume 9),
    717--772

    See Also
    --------
    :meth:`naginterfaces.library.examples.opt.handle_solve_bounds_foas_ex.main`
    :meth:`naginterfaces.library.examples.opt.handle_solve_dfls_ex.main`
    :meth:`naginterfaces.library.examples.opt.handle_solve_dfno_ex.main`
    :meth:`naginterfaces.library.examples.opt.handle_solve_ipopt_ex.main`
    :meth:`naginterfaces.library.examples.opt.handle_solve_lp_ipm_ex.main`
    :meth:`naginterfaces.library.examples.opt.handle_solve_socp_ipm_ex.main`
    """

    fun_name = 'naginterfaces.library.opt.handle_set_simplebounds'

    _bl = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'bl'},
        bl,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nvar_py = _b_utils._EngineFloat64ArrayType.get_shape(_bl, {'fun_name': fun_name, 'entity': 'bl'}, exp_rank=1)[0]
    _nvar = _b_utils._EngineIntScalarType(
        _nvar_py, {'fun_name': fun_name, 'entity': '_nvar'},
    ).to_ctypes()
    _bu = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'bu'},
        bu,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b_opt.handle_set_simplebounds(
        handle,
        _nvar,
        _bl,
        _bu,
    )

def handle_set_linconstr(handle, bl=None, bu=None, irowb=None, icolb=None, b=None, idlc=0):
    u"""
    Define a block of linear constraints to a problem initialized by
    ``handle_init``.

    ``handle_set_linconstr`` is a part of the NAG optimization modelling
    suite and defines the block of linear constraints of the problem.

    For full information please refer to the NAG Library document for
    e04rj

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04rjf.html

    Parameters
    ----------
    handle : Handle
        The handle to the problem. It needs to be initialized by
        ``handle_init`` and **must not** be changed before the call to
        ``handle_set_linconstr``.

    bl : None or float, array-like, shape (nclin), optional
        `bl` defines the lower bounds of the linear constraints, l_B

    bu : None or float, array-like, shape (nclin), optional
        `bu` defines the upper bounds of the linear constraints, u_B

    irowb : None or int, array-like, shape (nnzb), optional
        `irowb` specifies one-based row indices of the nnzb nonzeros of
        the sparse matrix B in coordinate storage (CS) format.

    icolb : None or int, array-like, shape (nnzb), optional
        `icolb` specifies one-based column indices of the nnzb nonzeros
        of the sparse matrix B in coordinate storage (CS) format.

    b : None or float, array-like, shape (nnzb), optional
        `b` stores the nnzb nonzeros of the sparse matrix B in
        coordinate storage (CS) format.

    idlc : int, optional
        Note: `idlc` is reserved for future releases of the NAG Library.

        If `idlc` = 0, new linear constraints are added to the problem
        definition. This is the only value allowed at the moment.

    Returns
    -------
    idlc : int
        The number of the last linear constraint added, thus nclin.

    Raises
    ------
    NagValueError
        (`errno` 1)
            `handle` has not been initialized.

        (`errno` 1)
            `handle` does not belong to the NAG optimization modelling
            suite, has not been initialized properly or is corrupted.

        (`errno` 1)
            `handle` has not been initialized properly or is corrupted.

        (`errno` 2)
            The problem cannot be modified in this phase any more, the
            solver has already been called.

        (`errno` 3)
            A set of linear constraints has already been defined.

        (`errno` 4)
            On entry, `idlc` = *<value>*.

            Constraint: `idlc` = 0.

        (`errno` 6)
            On entry, nnzb = *<value>*.

            Constraint: nnzb > 0.

        (`errno` 6)
            On entry, nclin = *<value>*.

            Constraint: nclin >= 0.

        (`errno` 8)
            On entry, i = *<value>*, `irowb`[i-1] = *<value>* and nclin
            = *<value>*.

            Constraint: 1 <= `irowb`[i-1] <= nclin.

        (`errno` 8)
            On entry, i = *<value>*, `icolb`[i-1] = *<value>* and n =
            *<value>*.

            Constraint: 1 <= `icolb`[i-1] <= n.

        (`errno` 8)
            On entry, more than one element of `b` has row index
            *<value>* and column index *<value>*.

            Constraint: each element of `b` must have a unique row and
            column index.

        (`errno` 10)
            On entry, j = *<value>*, `bl`[j-1] = *<value>* and `bu`[j-1]
            = *<value>*.

            Constraint: `bl`[j-1] <= `bu`[j-1].

        (`errno` 10)
            On entry, j = *<value>*, `bl`[j-1] = *<value>*, bigbnd =
            *<value>*.

            Constraint: `bl`[j-1] < bigbnd.

        (`errno` 10)
            On entry, j = *<value>*, `bu`[j-1] = *<value>*, bigbnd =
            *<value>*.

            Constraint: `bu`[j-1] > -bigbnd.

    See Also
    --------
    :meth:`naginterfaces.library.examples.opt.handle_solve_ipopt_ex.main`
    :meth:`naginterfaces.library.examples.opt.handle_solve_lp_ipm_ex.main`
    :meth:`naginterfaces.library.examples.opt.handle_solve_socp_ipm_ex.main`
    """

    fun_name = 'naginterfaces.library.opt.handle_set_linconstr'

    _bl = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'bl'},
        bl,
        copy=False,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nclin = _b_utils._EngineIntScalarType(
        0 if bl is None else _b_utils._EngineFloat64ArrayType.get_shape(_bl, {'fun_name': fun_name, 'entity': 'bl'}, exp_rank=1, allow_none=True)[0], {'fun_name': fun_name, 'entity': '_nclin'},
    ).to_ctypes()
    _bu = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'bu'},
        bu,
        copy=False,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _irowb = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'irowb'},
        irowb,
        copy=False,
        allow_none=True,
    )
    if irowb is None:
        _nnzb = 0
        _nnzb_py = _nnzb
    else:
        _nnzb_py = _b_utils._EngineIntArrayType.get_shape(_irowb, {'fun_name': fun_name, 'entity': 'irowb'}, exp_rank=1, allow_none=True)[0]
        _nnzb = _b_utils._EngineIntScalarType(
            _nnzb_py, {'fun_name': fun_name, 'entity': '_nnzb'},
        ).to_ctypes()
    _icolb = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'icolb'},
        icolb,
        copy=False,
        allow_none=True,
    )
    _b = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'b'},
        b,
        copy=False,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _return_scalar = _b_opt.handle_set_linconstr(
        handle,
        _nclin,
        _bl,
        _bu,
        _nnzb,
        _irowb,
        _icolb,
        _b,
        idlc,
    )
    _idlc_py = _return_scalar
    return _idlc_py

def handle_set_nlnconstr(handle, bl=None, bu=None, irowgd=None, icolgd=None):
    u"""
    Define a block of nonlinear constraints to a problem initialized by
    ``handle_init``.

    ``handle_set_nlnconstr`` is a part of the NAG optimization modelling
    suite and defines the number of nonlinear constraints of the problem
    as well as the sparsity structure of their first derivatives.

    For full information please refer to the NAG Library document for
    e04rk

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04rkf.html

    Parameters
    ----------
    handle : Handle
        The handle to the problem. It needs to be initialized by
        ``handle_init`` and **must not** be changed before the call to
        ``handle_set_nlnconstr``.

    bl : None or float, array-like, shape (ncnln), optional
        `bl` and `bu` define lower and upper bounds of the nonlinear
        constraints, l_g and u_g, respectively. To define the jth
        constraint as equality, set `bl`[j-1] = `bu`[j-1] = beta, where
        |beta| < bigbnd. To specify a nonexistent lower bound (i.e., l_j
        = - infinity), set `bl`[j-1] <= -bigbnd; to specify a
        nonexistent upper bound, set `bu`[j-1] >= bigbnd.

    bu : None or float, array-like, shape (ncnln), optional
        `bl` and `bu` define lower and upper bounds of the nonlinear
        constraints, l_g and u_g, respectively. To define the jth
        constraint as equality, set `bl`[j-1] = `bu`[j-1] = beta, where
        |beta| < bigbnd. To specify a nonexistent lower bound (i.e., l_j
        = - infinity), set `bl`[j-1] <= -bigbnd; to specify a
        nonexistent upper bound, set `bu`[j-1] >= bigbnd.

    irowgd : None or int, array-like, shape (nnzgd), optional
        Arrays `irowgd` and `icolgd` store the sparsity structure
        (pattern) of the Jacobian matrix as nnzgd nonzeros in coordinate
        storage (CS) format (see [the F11 Introduction]). The matrix has
        dimensions ncnln*n. `irowgd` specifies one-based row indices and
        `icolgd` specifies one-based column indices. No particular order
        of elements is expected, but elements should not repeat and the
        same order should be used when the Jacobian is evaluated for the
        solver, e.g., the value of dg_i/dx_j where i = `irowgd`[l-1] and
        j = `icolgd`[l-1] should be stored in gdx[l-1] in congrd in
        ``handle_solve_ipopt``, for l = 1,2,...,nnzgd.

    icolgd : None or int, array-like, shape (nnzgd), optional
        Arrays `irowgd` and `icolgd` store the sparsity structure
        (pattern) of the Jacobian matrix as nnzgd nonzeros in coordinate
        storage (CS) format (see [the F11 Introduction]). The matrix has
        dimensions ncnln*n. `irowgd` specifies one-based row indices and
        `icolgd` specifies one-based column indices. No particular order
        of elements is expected, but elements should not repeat and the
        same order should be used when the Jacobian is evaluated for the
        solver, e.g., the value of dg_i/dx_j where i = `irowgd`[l-1] and
        j = `icolgd`[l-1] should be stored in gdx[l-1] in congrd in
        ``handle_solve_ipopt``, for l = 1,2,...,nnzgd.

    Raises
    ------
    NagValueError
        (`errno` 1)
            `handle` has not been initialized.

        (`errno` 1)
            `handle` does not belong to the NAG optimization modelling
            suite, has not been initialized properly or is corrupted.

        (`errno` 1)
            `handle` has not been initialized properly or is corrupted.

        (`errno` 2)
            The problem cannot be modified in this phase any more, the
            solver has already been called.

        (`errno` 2)
            The Hessian of the nonlinear objective has already been
            defined, nonlinear constraints cannot be added.

        (`errno` 3)
            A set of nonlinear constraints has already been defined.

        (`errno` 6)
            On entry, nnzgd = *<value>*.

            Constraint: nnzgd > 0.

        (`errno` 6)
            On entry, ncnln = *<value>*.

            Constraint: ncnln >= 0.

        (`errno` 8)
            On entry, i = *<value>*, `irowgd`[i-1] = *<value>* and ncnln
            = *<value>*.

            Constraint: 1 <= `irowgd`[i-1] <= ncnln.

        (`errno` 8)
            On entry, i = *<value>*, `icolgd`[i-1] = *<value>* and n =
            *<value>*.

            Constraint: 1 <= `icolgd`[i-1] <= n.

        (`errno` 8)
            On entry, more than one element of structural Jacobian
            matrix has row index *<value>* and column index *<value>*.

            Constraint: each element of structural Jacobian matrix must
            have a unique row and column index.

        (`errno` 10)
            On entry, j = *<value>*, `bl`[j-1] = *<value>* and `bu`[j-1]
            = *<value>*.

            Constraint: `bl`[j-1] <= `bu`[j-1].

        (`errno` 10)
            On entry, j = *<value>*, `bl`[j-1] = *<value>*, bigbnd =
            *<value>*.

            Constraint: `bl`[j-1] < bigbnd.

        (`errno` 10)
            On entry, j = *<value>*, `bu`[j-1] = *<value>*, bigbnd =
            *<value>*.

            Constraint: `bu`[j-1] > -bigbnd.

    See Also
    --------
    :meth:`naginterfaces.library.examples.opt.handle_solve_ipopt_ex.main`
    """

    fun_name = 'naginterfaces.library.opt.handle_set_nlnconstr'

    _bl = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'bl'},
        bl,
        copy=False,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ncnln = _b_utils._EngineIntScalarType(
        0 if bl is None else _b_utils._EngineFloat64ArrayType.get_shape(_bl, {'fun_name': fun_name, 'entity': 'bl'}, exp_rank=1, allow_none=True)[0], {'fun_name': fun_name, 'entity': '_ncnln'},
    ).to_ctypes()
    _bu = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'bu'},
        bu,
        copy=False,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _irowgd = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'irowgd'},
        irowgd,
        copy=False,
        allow_none=True,
    )
    if irowgd is None:
        _nnzgd = 0
        _nnzgd_py = _nnzgd
    else:
        _nnzgd_py = _b_utils._EngineIntArrayType.get_shape(_irowgd, {'fun_name': fun_name, 'entity': 'irowgd'}, exp_rank=1, allow_none=True)[0]
        _nnzgd = _b_utils._EngineIntScalarType(
            _nnzgd_py, {'fun_name': fun_name, 'entity': '_nnzgd'},
        ).to_ctypes()
    _icolgd = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'icolgd'},
        icolgd,
        copy=False,
        allow_none=True,
    )
    _b_opt.handle_set_nlnconstr(
        handle,
        _ncnln,
        _bl,
        _bu,
        _nnzgd,
        _irowgd,
        _icolgd,
    )

def handle_set_nlnhess(handle, idf, irowh=None, icolh=None):
    u"""
    Define a structure of Hessian of the objective, constraints or the
    Lagrangian to a problem initialized by ``handle_init``.

    ``handle_set_nlnhess`` is a part of the NAG optimization modelling
    suite and defines the structure of the Hessians of the nonlinear
    objective and constraints, on assumption that they are present in
    the problem.
    Alternatively, it may be used to define the Hessian of the
    Lagrangian.

    For full information please refer to the NAG Library document for
    e04rl

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04rlf.html

    Parameters
    ----------
    handle : Handle
        The handle to the problem. It needs to be initialized by
        ``handle_init`` and **must not** be changed before the call to
        ``handle_set_nlnhess``.

    idf : int
        Specifies the quantities for which a sparsity structure is
        provided in nnzh, `irowh` and `icolh`.

        `idf` = -1

            The sparsity structure of the Hessian of the Lagrangian is
            provided.

        `idf` = 0

            The sparsity structure of the Hessian of the objective
            function is provided.

        `idf` > 0

            The sparsity structure of the Hessian of the `idf`th
            constraint function is provided.

        The value of `idf` will also determine how an NLP solver will
        call the user-supplied functions that evaluate these nonzeros at
        particular points of the decision variable space, i.e., whether
        the solver will expect the nonzero values of the objective and
        constraint Hessians in separate calls or merged in the
        Lagrangian Hessian, in one call.

        See, for example, hess of ``handle_solve_ipopt``.

    irowh : None or int, array-like, shape (nnzh), optional
        Arrays `irowh` and `icolh` store the nonzeros of the upper
        triangle of the matrix H in coordinate storage (CS) format (see
        [the F11 Introduction]). `irowh` specifies one-based row
        indices, `icolh` specifies one-based column indices and
        specifies the values of the nonzero elements in such a way that
        h_{ij} = H[l-1] where i = `irowh`[l-1] and j = `icolh`[l-1], for
        l = 1,2,...,nnzh. No particular order is expected, but elements
        should not repeat.

    icolh : None or int, array-like, shape (nnzh), optional
        Arrays `irowh` and `icolh` store the nonzeros of the upper
        triangle of the matrix H in coordinate storage (CS) format (see
        [the F11 Introduction]). `irowh` specifies one-based row
        indices, `icolh` specifies one-based column indices and
        specifies the values of the nonzero elements in such a way that
        h_{ij} = H[l-1] where i = `irowh`[l-1] and j = `icolh`[l-1], for
        l = 1,2,...,nnzh. No particular order is expected, but elements
        should not repeat.

    Raises
    ------
    NagValueError
        (`errno` 1)
            `handle` has not been initialized.

        (`errno` 1)
            `handle` does not belong to the NAG optimization modelling
            suite, has not been initialized properly or is corrupted.

        (`errno` 1)
            `handle` has not been initialized properly or is corrupted.

        (`errno` 2)
            The problem cannot be modified in this phase any more, the
            solver has already been called.

        (`errno` 2)
            Neither nonlinear objective nor nonlinear constraints are
            present. The structure of the Hessian cannot be defined.

        (`errno` 2)
            No nonlinear objective has been defined, its Hessian cannot
            be set.

        (`errno` 3)
            The structure of the Hessian of the Lagrangian has already
            been defined.

        (`errno` 3)
            The structure of the individual Hessians has already been
            defined, the Hessian of the Lagrangian cannot be defined.

        (`errno` 3)
            On entry, `idf` = *<value>*.

            The structure of the Hessian of nonlinear function linked to
            the given `idf` has already been defined.

        (`errno` 6)
            On entry, nnzh = *<value>*.

            Constraint: nnzh > 0.

        (`errno` 7)
            On entry, `idf` = *<value>*.

            Constraint: *<value>* <= `idf` <= *<value>*.

        (`errno` 8)
            On entry, i = *<value>*, `irowh`[i-1] = *<value>* and n =
            *<value>*.

            Constraint: 1 <= `irowh`[i-1] <= n.

        (`errno` 8)
            On entry, i = *<value>*, `icolh`[i-1] = *<value>* and n =
            *<value>*.

            Constraint: 1 <= `icolh`[i-1] <= n.

        (`errno` 8)
            On entry, i = *<value>*, `irowh`[i-1] = *<value>* and
            `icolh`[i-1] = *<value>*.

            Constraint: `irowh`[i-1] <= `icolh`[i-1] (elements within
            the upper triangle).

        (`errno` 8)
            On entry, more than one element of structural matrix H has
            row index *<value>* and column index *<value>*.

            Constraint: each element of structural matrix H must have a
            unique row and column index.

    See Also
    --------
    :meth:`naginterfaces.library.examples.opt.handle_solve_ipopt_ex.main`
    """

    fun_name = 'naginterfaces.library.opt.handle_set_nlnhess'

    _irowh = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'irowh'},
        irowh,
        copy=False,
        allow_none=True,
    )
    if irowh is None:
        _nnzh = 0
        _nnzh_py = _nnzh
    else:
        _nnzh_py = _b_utils._EngineIntArrayType.get_shape(_irowh, {'fun_name': fun_name, 'entity': 'irowh'}, exp_rank=1, allow_none=True)[0]
        _nnzh = _b_utils._EngineIntScalarType(
            _nnzh_py, {'fun_name': fun_name, 'entity': '_nnzh'},
        ).to_ctypes()
    _icolh = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'icolh'},
        icolh,
        copy=False,
        allow_none=True,
    )
    _b_opt.handle_set_nlnhess(
        handle,
        idf,
        _nnzh,
        _irowh,
        _icolh,
    )

def handle_set_nlnls(handle, nres, irowrd=None, icolrd=None):
    u"""
    Define a nonlinear least squares objective function for a problem
    initialized by ``handle_init``.

    ``handle_set_nlnls`` is a part of the NAG optimization modelling
    suite and defines the number of residuals in a sum of squares
    objective function (nonlinear least squares problems) and,
    optionally, the sparsity structure of their first derivatives.

    For full information please refer to the NAG Library document for
    e04rm

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04rmf.html

    Parameters
    ----------
    handle : Handle
        The handle to the problem. It needs to be initialized by
        ``handle_init`` and **must not** be changed before the call to
        ``handle_set_nlnls``.

    nres : int
        m_r, the number of residuals in the objective function.

        If `nres` = 0, no objective function will be defined and
        `irowrd` and `icolrd` will not be referenced.

    irowrd : None or int, array-like, shape (nnzrd), optional
        Arrays `irowrd` and `icolrd` store the sparsity structure
        (pattern) of the first derivative matrix as nnzrd nonzeros in
        coordinate storage (CS) format (see [the F11 Introduction]). The
        matrix has dimensions n*m_r. `irowrd` specifies one-based row
        indices and `icolrd` specifies one-based column indices. No
        particular order of elements is expected, but elements should
        not repeat and the same order should be used when the first
        derivative matrix is evaluated for the solver.

        If `irowrd` and `icolrd` are both None then the first derivative
        matrix is considered dense and `irowrd` and `icolrd` will not be
        referenced.

        The ordering is assumed to be column-wise, namely the function
        will behave as if nnzrd = n*m_r and the vectors `irowrd` and
        `icolrd` filled as:

            `irowrd` = (1,2,…,n,1,2,…,n,…,1,2,…,n);

            `icolrd` = (1,1,…,1,2,2,…,2,…,m_r,m_r,…,m_r).

    icolrd : None or int, array-like, shape (nnzrd), optional
        Arrays `irowrd` and `icolrd` store the sparsity structure
        (pattern) of the first derivative matrix as nnzrd nonzeros in
        coordinate storage (CS) format (see [the F11 Introduction]). The
        matrix has dimensions n*m_r. `irowrd` specifies one-based row
        indices and `icolrd` specifies one-based column indices. No
        particular order of elements is expected, but elements should
        not repeat and the same order should be used when the first
        derivative matrix is evaluated for the solver.

        If `irowrd` and `icolrd` are both None then the first derivative
        matrix is considered dense and `irowrd` and `icolrd` will not be
        referenced.

        The ordering is assumed to be column-wise, namely the function
        will behave as if nnzrd = n*m_r and the vectors `irowrd` and
        `icolrd` filled as:

            `irowrd` = (1,2,…,n,1,2,…,n,…,1,2,…,n);

            `icolrd` = (1,1,…,1,2,2,…,2,…,m_r,m_r,…,m_r).

    Raises
    ------
    NagValueError
        (`errno` 1)
            `handle` has not been initialized.

        (`errno` 1)
            `handle` does not belong to the NAG optimization modelling
            suite, has not been initialized properly or is corrupted.

        (`errno` 1)
            `handle` has not been initialized properly or is corrupted.

        (`errno` 2)
            The problem cannot be modified in this phase any more, the
            solver has already been called.

        (`errno` 2)
            The Hessians of nonlinear functions have already been
            defined, a nonlinear objective cannot be added.

        (`errno` 3)
            The objective function has already been defined.

        (`errno` 6)
            On entry, `nres` = *<value>*.

            Constraint: `nres` >= 0.

        (`errno` 6)
            On entry, nnzrd = *<value>*.

            Constraint: nnzrd > 0.

        (`errno` 8)
            On entry, i = *<value>*, `irowrd`[i-1] = *<value>* and n =
            *<value>*.

            Constraint: 1 <= `irowrd`[i-1] <= n.

        (`errno` 8)
            On entry, i = *<value>*, `icolrd`[i-1] = *<value>* and
            `nres` = *<value>*.

            Constraint: 1 <= `icolrd`[i-1] <= `nres`.

        (`errno` 8)
            On entry, more than one element of first derivative matrix
            has row index *<value>* and column index *<value>*.

            Constraint: each element of first derivative matrix must
            have a unique row and column index.

    See Also
    --------
    :meth:`naginterfaces.library.examples.opt.handle_solve_dfls_ex.main`
    """

    fun_name = 'naginterfaces.library.opt.handle_set_nlnls'

    if (irowrd is None) != (icolrd is None):
        _b_utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                'All of the following must be provided if one is provided:',
                'irowrd, icolrd.',
            ],
            errcodes=(9,),
        ).manage()
    _irowrd = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'irowrd'},
        irowrd,
        copy=False,
        allow_none=True,
    )
    if irowrd is None:
        _nnzrd = 0
        _nnzrd_py = _nnzrd
    else:
        _nnzrd_py = _b_utils._EngineIntArrayType.get_shape(_irowrd, {'fun_name': fun_name, 'entity': 'irowrd'}, exp_rank=1, allow_none=True)[0]
        _nnzrd = _b_utils._EngineIntScalarType(
            _nnzrd_py, {'fun_name': fun_name, 'entity': '_nnzrd'},
        ).to_ctypes()
    _icolrd = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'icolrd'},
        icolrd,
        copy=False,
        allow_none=True,
    )
    _b_opt.handle_set_nlnls(
        handle,
        nres,
        _nnzrd,
        _irowrd,
        _icolrd,
    )

def handle_set_linmatineq(handle, dima, nnza, irowa, icola, a, blksizea=None, idblk=0):
    u"""
    Add one or more linear matrix inequality constraints to a problem
    initialized by ``handle_init``.

    ``handle_set_linmatineq`` is a part of the NAG optimization
    modelling suite and defines one or more linear matrix constraints of
    the problem.

    For full information please refer to the NAG Library document for
    e04rn

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04rnf.html

    Parameters
    ----------
    handle : Handle
        The handle to the problem. It needs to be initialized by
        ``handle_init`` and **must not** be changed before the call to
        ``handle_set_linmatineq``.

    dima : int
        d, the dimension of the matrices A_i, for i = 0,1,...,nvar.

    nnza : int, array-like, shape (nvar+1)
        `nnza`[i], for i = 0,1,...,nvar, gives the number of nonzero
        elements in the upper triangle of matrix A_i. To define A_i as a
        zero matrix, set `nnza`[i] = 0. However, there must be at least
        one matrix with at least one nonzero.

    irowa : int, array-like, shape (nnzasum)
        Nonzero elements in upper triangle of matrices A_i stored in
        coordinate storage. The first `nnza`[0] elements belong to A_0,
        the following `nnza`[1] elements belong to A_1, etc. See
        explanation above.

    icola : int, array-like, shape (nnzasum)
        Nonzero elements in upper triangle of matrices A_i stored in
        coordinate storage. The first `nnza`[0] elements belong to A_0,
        the following `nnza`[1] elements belong to A_1, etc. See
        explanation above.

    a : float, array-like, shape (nnzasum)
        Nonzero elements in upper triangle of matrices A_i stored in
        coordinate storage. The first `nnza`[0] elements belong to A_0,
        the following `nnza`[1] elements belong to A_1, etc. See
        explanation above.

    blksizea : None or int, array-like, shape (nblk), optional
        If nblk > 1, sizes d_k of the diagonal blocks.

        If nblk = 1, `blksizea` is not referenced.

    idblk : int, optional
        If `idblk` = 0, new matrix inequalities are created. This is the
        only value allowed at the moment; nonzero values are reserved
        for future releases of the NAG Library.

    Returns
    -------
    idblk : int
        The number of the last matrix inequality added. By definition,
        it is the number of the matrix inequalities already defined plus
        nblk.

    Raises
    ------
    NagValueError
        (`errno` 1)
            `handle` has not been initialized.

        (`errno` 1)
            `handle` does not belong to the NAG optimization modelling
            suite, has not been initialized properly or is corrupted.

        (`errno` 1)
            `handle` has not been initialized properly or is corrupted.

        (`errno` 2)
            The problem cannot be modified in this phase any more, the
            solver has already been called.

        (`errno` 4)
            On entry, nvar = *<value>*, expected value = *<value>*.

            Constraint: nvar must match the value given during
            initialization of `handle`.

        (`errno` 4)
            On entry, `idblk` = *<value>*.

            Constraint: `idblk` = 0.

        (`errno` 6)
            On entry, `dima` = *<value>*.

            Constraint: `dima` > 0.

        (`errno` 6)
            On entry, nnzasum = *<value>* and sum(`nnza`) = *<value>*.

            Constraint: nnzasum >= sum(`nnza`).

        (`errno` 6)
            On entry, i = *<value>* and `nnza`[i-1] = *<value>*.

            Constraint: `nnza`[i-1] >= 0.

        (`errno` 6)
            On entry, sum(`nnza`) = *<value>*.

            Constraint: sum(`nnza`) >= 1.

        (`errno` 7)
            On entry, nblk = *<value>*.

            Constraint: nblk > 0.

        (`errno` 7)
            On entry, i = *<value>* and `blksizea`[i-1] = *<value>*.

            Constraint: `blksizea`[i-1] >= 1.

        (`errno` 7)
            On entry, `dima` = *<value>* and sum(`blksizea`) =
            *<value>*.

            Constraint: sum(`blksizea`) = `dima`.

        (`errno` 8)
            An error occurred in matrix A_i, i = *<value>* (counting
            indices 1...nvar+1).

            On entry, j = *<value>*, `irowa`[j-1] = *<value>* and `dima`
            = *<value>*.

            Constraint: 1 <= `irowa`[j-1] <= `dima`.

        (`errno` 8)
            An error occurred in matrix A_i, i = *<value>* (counting
            indices 1...nvar+1).

            On entry, j = *<value>*, `icola`[j-1] = *<value>* and `dima`
            = *<value>*.

            Constraint: 1 <= `icola`[j-1] <= `dima`.

        (`errno` 8)
            An error occurred in matrix A_i, i = *<value>* (counting
            indices 1...nvar+1).

            On entry, j = *<value>*, `irowa`[j-1] = *<value>* and
            `icola`[j-1] = *<value>*.

            Constraint: `irowa`[j-1] <= `icola`[j-1] (elements within
            the upper triangle).

        (`errno` 8)
            An error occurred in matrix A_i, i = *<value>* (counting
            indices 1...nvar+1).

            On entry, more than one element of A_i has row index
            *<value>* and column index *<value>*.

            Constraint: each element of A_i must have a unique row and
            column index.

        (`errno` 8)
            An error occurred in matrix A_i, i = *<value>* (counting
            indices 1...nvar+1).

            On entry, j = *<value>*, `irowa`[j-1] = *<value>* and
            `icola`[j-1] = *<value>*. Maximum column index in this row
            given by the block structure defined by `blksizea` is
            *<value>*.

            Constraint: all elements of A_i must respect the block
            structure given by `blksizea`.

    See Also
    --------
    :meth:`naginterfaces.library.examples.opt.handle_solve_pennon_bmi_ex.main`
    :meth:`naginterfaces.library.examples.opt.handle_solve_pennon_lmi_ex.main`
    """

    fun_name = 'naginterfaces.library.opt.handle_set_linmatineq'

    _nnza = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'nnza'},
        nnza,
        copy=False,
    )
    _nvar = _b_utils._EngineIntScalarType(
        _b_utils._EngineIntArrayType.get_shape(_nnza, {'fun_name': fun_name, 'entity': 'nnza'}, exp_rank=1)[0]-1, {'fun_name': fun_name, 'entity': '_nvar'},
    ).to_ctypes()
    _irowa = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'irowa'},
        irowa,
        copy=False,
    )
    _nnzasum_py = _b_utils._EngineIntArrayType.get_shape(_irowa, {'fun_name': fun_name, 'entity': 'irowa'}, exp_rank=1)[0]
    _nnzasum = _b_utils._EngineIntScalarType(
        _nnzasum_py, {'fun_name': fun_name, 'entity': '_nnzasum'},
    ).to_ctypes()
    _icola = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'icola'},
        icola,
        copy=False,
    )
    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _blksizea = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'blksizea'},
        blksizea,
        copy=False,
        allow_none=True,
    )
    _nblk = _b_utils._EngineIntScalarType(
        1 if blksizea is None else _b_utils._EngineIntArrayType.get_shape(_blksizea, {'fun_name': fun_name, 'entity': 'blksizea'}, exp_rank=1, allow_none=True)[0], {'fun_name': fun_name, 'entity': '_nblk'},
    ).to_ctypes()
    _return_scalar = _b_opt.handle_set_linmatineq(
        handle,
        _nvar,
        dima,
        _nnza,
        _nnzasum,
        _irowa,
        _icola,
        _a,
        _nblk,
        _blksizea,
        idblk,
    )
    _idblk_py = _return_scalar
    return _idblk_py

def handle_set_quadmatineq(handle, qi, qj, dimq, nnzq, irowq, icolq, q, idblk):
    u"""
    Define bilinear matrix terms to a problem initialized by
    ``handle_init``.

    ``handle_set_quadmatineq`` is a part of the NAG optimization
    modelling suite and defines bilinear matrix terms either in a new
    matrix constraint or adds them to an existing linear matrix
    inequality.

    For full information please refer to the NAG Library document for
    e04rp

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04rpf.html

    Parameters
    ----------
    handle : Handle
        The handle to the problem. It needs to be initialized by
        ``handle_init`` and **must not** be changed before the call to
        ``handle_set_quadmatineq``.

    qi : int, array-like, shape (nq)
        The index pairs i,j of the nonzero matrices Q_{ij} in any order.

    qj : int, array-like, shape (nq)
        The index pairs i,j of the nonzero matrices Q_{ij} in any order.

    dimq : int
        d, the dimension of matrices Q_{ij}.

    nnzq : int, array-like, shape (nq)
        The numbers of nonzero elements in the upper triangles of Q_{ij}
        matrices.

    irowq : int, array-like, shape (nnzqsum)
        The nonzero elements of the upper triangles of matrices Q_{ij}
        stored in coordinate storage format. The first `nnzq`[0]
        elements belong to the first Q_{i_1 j_1}, the following
        `nnzq`[1] to Q_{i_2 j_2}, etc.

    icolq : int, array-like, shape (nnzqsum)
        The nonzero elements of the upper triangles of matrices Q_{ij}
        stored in coordinate storage format. The first `nnzq`[0]
        elements belong to the first Q_{i_1 j_1}, the following
        `nnzq`[1] to Q_{i_2 j_2}, etc.

    q : float, array-like, shape (nnzqsum)
        The nonzero elements of the upper triangles of matrices Q_{ij}
        stored in coordinate storage format. The first `nnzq`[0]
        elements belong to the first Q_{i_1 j_1}, the following
        `nnzq`[1] to Q_{i_2 j_2}, etc.

    idblk : int
        If `idblk` = 0, a new matrix constraint is created; otherwise
        `idblk` = k > 0, the number of the existing linear matrix
        constraint to be expanded with the bilinear terms.

    Returns
    -------
    idblk : int
        If `idblk` = 0 on entry, the number of the new matrix constraint
        is returned. By definition, it is the number of the matrix
        inequalities already defined plus one. Otherwise, `idblk` > 0
        stays unchanged.

    Raises
    ------
    NagValueError
        (`errno` 1)
            `handle` has not been initialized.

        (`errno` 1)
            `handle` does not belong to the NAG optimization modelling
            suite, has not been initialized properly or is corrupted.

        (`errno` 1)
            `handle` has not been initialized properly or is corrupted.

        (`errno` 2)
            The problem cannot be modified in this phase any more, the
            solver has already been called.

        (`errno` 3)
            On entry, `idblk` = *<value>*.

            Bilinear terms of the matrix inequality block with the given
            `idblk` have already been defined.

        (`errno` 4)
            On entry, `idblk` = *<value>*.

            The given `idblk` refers to a nonexistent matrix inequality
            block.

            No matrix inequalities have been added yet.

        (`errno` 4)
            On entry, `idblk` = *<value>*.

            The given `idblk` does not match with any existing matrix
            inequality block.

            The maximum `idblk` is currently *<value>*.

        (`errno` 5)
            On entry, `dimq` = *<value>*, `idblk` = *<value>*.

            The correct dimension of the given `idblk` is *<value>*.

            Constraint: `dimq` must match the dimension of the block
            supplied earlier.

        (`errno` 6)
            On entry, `idblk` = *<value>*.

            Constraint: `idblk` >= 0.

        (`errno` 6)
            On entry, nq = *<value>*.

            Constraint: nq > 0.

        (`errno` 6)
            On entry, `dimq` = *<value>*.

            Constraint: `dimq` > 0.

        (`errno` 6)
            On entry, nnzqsum = *<value>* and sum(`nnzq`) = *<value>*.

            Constraint: nnzqsum >= sum(`nnzq`).

        (`errno` 6)
            On entry, i = *<value>* and `nnzq`[i-1] = *<value>*.

            Constraint: `nnzq`[i-1] > 0.

        (`errno` 8)
            On entry, an error occurred in matrix Q_{{ij}} of index k =
            *<value>*, `qi`[k-1] = *<value>*, `qj`[k-1] = *<value>*.

            For j = *<value>*, `irowq`[j-1] = *<value>* and `dimq` =
            *<value>*.

            Constraint: 1 <= `irowq`[j-1] <= `dimq`.

        (`errno` 8)
            On entry, an error occurred in matrix Q_{{ij}} of index k =
            *<value>*, `qi`[k-1] = *<value>*, `qj`[k-1] = *<value>*.

            For j = *<value>*, `icolq`[j-1] = *<value>* and `dimq` =
            *<value>*.

            Constraint: 1 <= `icolq`[j-1] <= `dimq`.

        (`errno` 8)
            On entry, an error occurred in matrix Q_{{ij}} of index k =
            *<value>*, `qi`[k-1] = *<value>*, `qj`[k-1] = *<value>*.

            For j = *<value>*, `irowq`[j-1] = *<value>* and `icolq`[j-1]
            = *<value>*.

            Constraint: `irowq`[j-1] <= `icolq`[j-1] (elements within
            the upper triangle).

        (`errno` 8)
            On entry, an error occurred in matrix Q_{{ij}} of index k =
            *<value>*, `qi`[k-1] = *<value>*, `qj`[k-1] = *<value>*.

            More than one element of Q_{{ij}} has row index *<value>*
            and column index *<value>*.

            Constraint: each element of Q_{{ij}} must have a unique row
            and column index.

        (`errno` 9)
            On entry, k = *<value>*, `qi`[k-1] = *<value>* and n =
            *<value>*.

            Constraint: 1 <= `qi`[k-1] <= n.

        (`errno` 9)
            On entry, k = *<value>*, `qj`[k-1] = *<value>* and n =
            *<value>*.

            Constraint: 1 <= `qj`[k-1] <= n.

        (`errno` 9)
            On entry, index pair with `qi` = *<value>* and `qj` =
            *<value>* repeats.

            Constraint: each index pair `qi`, `qj` must be unique.

    References
    ----------
    Syrmos, V L, Abdallah, C T, Dorato, P and Grigoriadis, K, 1997,
    `Static output feedback -- a survey`, Journal Automatica (Journal of
    IFAC) (Volume 33) (2), 125--137

    See Also
    --------
    :meth:`naginterfaces.library.examples.opt.handle_solve_pennon_bmi_ex.main`
    """

    fun_name = 'naginterfaces.library.opt.handle_set_quadmatineq'

    _qi = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'qi'},
        qi,
        copy=False,
    )
    _nq_py = _b_utils._EngineIntArrayType.get_shape(_qi, {'fun_name': fun_name, 'entity': 'qi'}, exp_rank=1)[0]
    _nq = _b_utils._EngineIntScalarType(
        _nq_py, {'fun_name': fun_name, 'entity': '_nq'},
    ).to_ctypes()
    _qj = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'qj'},
        qj,
        copy=False,
    )
    _nnzq = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'nnzq'},
        nnzq,
        copy=False,
    )
    _irowq = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'irowq'},
        irowq,
        copy=False,
    )
    _nnzqsum_py = _b_utils._EngineIntArrayType.get_shape(_irowq, {'fun_name': fun_name, 'entity': 'irowq'}, exp_rank=1)[0]
    _nnzqsum = _b_utils._EngineIntScalarType(
        _nnzqsum_py, {'fun_name': fun_name, 'entity': '_nnzqsum'},
    ).to_ctypes()
    _icolq = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'icolq'},
        icolq,
        copy=False,
    )
    _q = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'q'},
        q,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _return_scalar = _b_opt.handle_set_quadmatineq(
        handle,
        _nq,
        _qi,
        _qj,
        dimq,
        _nnzq,
        _nnzqsum,
        _irowq,
        _icolq,
        _q,
        idblk,
    )
    _idblk_py = _return_scalar
    return _idblk_py

def handle_set_get_real(handle, cmdstr, rarr=None):
    u"""
    Insert or extract data relating to a problem initialized by
    ``handle_init``.

    ``handle_set_get_real`` is a part of the NAG optimization modelling
    suite.
    It allows you to read or write a piece of information to the problem
    stored in the handle.
    For example, it may be used to extract the current approximation of
    the solution during a monitoring step.

    For full information please refer to the NAG Library document for
    e04rx

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04rxf.html

    Parameters
    ----------
    handle : Handle
        The handle to the problem. It needs to be initialized by
        ``handle_init`` and **must not** be changed between calls to the
        NAG optimization modelling suite.

    cmdstr : str
        A string which identifies the item within the handle to be read
        or written. The string is case insensitive and space tolerant.

    rarr : None or float, array-like, shape (lrarr), optional
        If `rarr` is not None (write mode), `rarr` must contain the
        information to be written to the handle.

    Returns
    -------
    lrarr : int
        In write mode and when the size of the input `rarr` does not
        match the item identified by `cmdstr`, `lrarr` is the correct
        expected size. Otherwise, it is merely the actual size of the
        relevant `rarr`.

    rarr : None or float, ndarray, shape (lrarr)
        If `rarr` was None on entry (read mode), `rarr` contains the
        information requested by `cmdstr`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            `handle` has not been initialized.

        (`errno` 1)
            `handle` does not belong to the NAG optimization modelling
            suite, has not been initialized properly or is corrupted.

        (`errno` 1)
            `handle` has not been initialized properly or is corrupted.

        (`errno` 3)
            The provided `cmdstr` is not recognised.

        (`errno` 4)
            Reading mode is not supported for the given `cmdstr`.

        (`errno` 5)
            Writing mode is not supported for the given `cmdstr`.

        (`errno` 6)
            The request cannot be processed at this phase.

        (`errno` 6)
            The requested information is not available.

        (`errno` 7)
            The request cannot be processed by the current solver.
    """

    if rarr is None:
        _ioflag = 1
        _ioflag_py = _ioflag
    else:
        _ioflag = 0
        _ioflag_py = _ioflag
    fun_name = 'naginterfaces.library.opt.handle_set_get_real'

    _rarr = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'rarr'},
        rarr,
        copy=True,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    if rarr is None:
        _lrarr = _b_utils._EngineIntScalarType.c_type(-1)
        _lrarr_py = _b_utils._EngineIntScalarType.to_py(_lrarr)
    else:
        _lrarr_py = _b_utils._EngineFloat64ArrayType.get_shape(_rarr, {'fun_name': fun_name, 'entity': 'rarr'}, exp_rank=1, allow_none=True)[0]
        _lrarr = _b_utils._EngineIntScalarType(
            _lrarr_py, {'fun_name': fun_name, 'entity': '_lrarr'},
        ).to_ctypes()
    if rarr is None:
        _rarr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
            (_lrarr_py,),
        )
    _return_scalar = _b_opt.handle_set_get_real(
        handle,
        cmdstr,
        _ioflag,
        _lrarr,
        _rarr,
    )
    if rarr is None:
        _lrarr_py = _return_scalar
        _lrarr = _b_utils._EngineIntScalarType(
            _lrarr_py, {'fun_name': fun_name, 'entity': 'lrarr'},
        ).to_ctypes()
        _rarr = _b_utils._EngineFloat64ArrayType.empty_ndarray(
            (_lrarr_py,),
        )
        _return_scalar = _b_opt.handle_set_get_real(
            handle,
            cmdstr,
            _ioflag,
            _lrarr,
            _rarr,
        )
    if _ioflag_py != 0:
        _rarr_py = _rarr
    else:
        _rarr_py = None
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['lrarr', 'rarr'],
    )(
        _lrarr_py,
        _rarr_py,
    )
    return _return_data

def handle_print(handle, nout, cmdstr, io_manager=None):
    u"""
    Print information about a problem handle initialized by
    ``handle_init``.

    ``handle_print`` is a part of the NAG optimization modelling suite.
    It allows you to print information about the problem, stored as a
    handle, such as which parts have already been defined or details of
    the matrix constraints.

    For full information please refer to the NAG Library document for
    e04ry

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04ryf.html

    Parameters
    ----------
    handle : Handle
        The handle to the problem. It needs to be initialized by
        ``handle_init`` and **must not** be changed before the call to
        ``handle_print``.

    nout : int
        The unit number (see
        :meth:`~naginterfaces.base.utils.FileObjManager.unit_from_fileobj`)
        which identifies the file to be written to.

    cmdstr : str
        A command string which contains one or more keywords which
        identify the piece of information about the handle to be
        printed. Keywords are case-insensitive and space tolerant.
        Multiple keywords in `cmdstr` must be separated by commas or
        semicolons.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Raises
    ------
    NagValueError
        (`errno` 1)
            `handle` has not been initialized.

        (`errno` 1)
            `handle` does not belong to the NAG optimization modelling
            suite, has not been initialized properly or is corrupted.

        (`errno` 1)
            `handle` has not been initialized properly or is corrupted.

        (`errno` 6)
            On entry, `nout` = *<value>*.

            Constraint: `nout` >= 0.

        (`errno` 7)
            Keyword number *<value>* is not recognized.

        (`errno` 7)
            Keyword number *<value>* is not recognized, it is too long.

        (`errno` 7)
            `cmdstr` does not contain any keywords or is empty.

        (`errno` 199)
            An error occurred when writing to output.
    """

    _b_opt.handle_print(
        handle,
        nout,
        cmdstr,
        io_manager,
    )

def handle_free(handle):
    u"""
    Destroy the problem handle initialized by ``handle_init`` and
    deallocate all the memory used.

    ``handle_free`` is a part of the NAG optimization modelling suite.
    It is used to deallocate the memory used within the problem handle
    and to destroy the handle itself.

    For full information please refer to the NAG Library document for
    e04rz

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04rzf.html

    Parameters
    ----------
    handle : Handle, modified in place
        `On entry`: the handle to the problem. It needs to be
        initialized (e.g., by ``handle_init``) and **must not** be
        changed between calls to the NAG optimization modelling suite.

        `On exit`: the handle is destroyed and set to a null Handle.

    Raises
    ------
    NagValueError
        (`errno` 1)
            `handle` does not belong to the NAG optimization modelling
            suite, has not been initialized properly or is corrupted.

        (`errno` 1)
            `handle` has not been initialized properly or is corrupted.

    See Also
    --------
    :meth:`naginterfaces.library.examples.opt.handle_solve_bounds_foas_ex.main`
    :meth:`naginterfaces.library.examples.opt.handle_solve_dfls_ex.main`
    :meth:`naginterfaces.library.examples.opt.handle_solve_dfno_ex.main`
    :meth:`naginterfaces.library.examples.opt.handle_solve_ipopt_ex.main`
    :meth:`naginterfaces.library.examples.opt.handle_solve_lp_ipm_ex.main`
    :meth:`naginterfaces.library.examples.opt.handle_solve_pennon_bmi_ex.main`
    :meth:`naginterfaces.library.examples.opt.handle_solve_pennon_lmi_ex.main`
    :meth:`naginterfaces.library.examples.opt.handle_solve_socp_ipm_ex.main`
    """

    _b_opt.handle_free(
        handle,
    )

def handle_read_file(file, ftype):
    u"""
    Load a problem from a file to a new handle for the NAG optimization
    modelling suite; supported formats: extended MPS, SDPA.

    ``handle_read_file`` initializes a data structure for the NAG
    optimization modelling suite from a data file for problems such as,
    linear programming (LP), quadratic programming (QP), second-order
    cone programming (SOCP), or linear semidefinite programming (SDP).

    For full information please refer to the NAG Library document for
    e04sa

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04saf.html

    Parameters
    ----------
    file : str
        The name of the file to be opened.

    ftype : str
        The expected file format of the input file, case insensitive.

        `ftype` = 'MPS' or 'M'

            The input file is in MPS format.

        `ftype` = 'SDPA' or 'S'

            The input file is in SDPA format.

    Returns
    -------
    handle : Handle
        Holds a handle to the internal data structure. It **must not**
        be changed between calls to the NAG optimization modelling
        suite.

    pinfo : int, ndarray, shape (100)
        Problem sizes and statistics as given in the table below:

        [table omitted]

        nvar, nnzu, nnzuc and nnzua define the problem size and are
        required by other routines such as ``handle_solve_lp_ipm``,
        ``handle_solve_socp_ipm`` and ``handle_solve_pennon`` in the
        optimization suite.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `ftype` = *<value>*.

            Constraint: `ftype` = 'MPS', 'M', 'SDPA' or 'S'.

        (`errno` 2)
            Cannot open file *<value>* for reading.

        (`errno` 3)
            Cannot close file.

        (`errno` 4)
            The input file seems to be empty. No data was read.

        (`errno` 5)
            Reading from the file caused an unknown error on line
            *<value>*.

        (`errno` 50)
            The token on line *<value>* at position *<value>* to
            *<value>* was not recognized as a valid integer.

        (`errno` 51)
            The token on line *<value>* at position *<value>* to
            *<value>* was not recognized as a valid real number.

        (`errno` 52)
            The token on line *<value>* starting at position *<value>*
            was too long and was not recognized.

        (`errno` 53)
            An invalid number of variables was given on line *<value>*.

            The number stated there is *<value>* and needs to be at
            least 1.

        (`errno` 54)
            An invalid number of blocks was given on line *<value>*.

            The number stated there is *<value>* and needs to be at
            least 1.

        (`errno` 55)
            An invalid size of the block number *<value>* was given on
            line *<value>*.

            The number stated there is *<value>* and needs to be
            nonzero.

        (`errno` 56)
            Not enough data was given on line *<value>* specifying block
            sizes.

            Expected *<value>* tokens but found only *<value>*.

        (`errno` 57)
            Not enough data was given on line *<value>* specifying the
            objective function.

            Expected *<value>* tokens but found only *<value>*.

        (`errno` 58)
            Not enough data was given on line *<value>* specifying
            nonzero matrix elements.

            Expected *<value>* tokens but found only *<value>*.

        (`errno` 59)
            Invalid structural data found on line *<value>*.

            The given matrix number is out of bounds. Its value
            *<value>* must be between *<value>* and *<value>*
            (inclusive).

        (`errno` 60)
            Invalid structural data found on line *<value>*.

            The given block number is out of bounds. Its value *<value>*
            must be between *<value>* and *<value>* (inclusive).

        (`errno` 61)
            Invalid structural data found on line *<value>*.

            The given row index is out of bounds, it must respect the
            size of the block. Its value *<value>* must be between
            *<value>* and *<value>* (inclusive).

        (`errno` 62)
            Invalid structural data found on line *<value>*.

            The given column index is out of bounds, it must respect the
            size of the block. Its value *<value>* must be between
            *<value>* and *<value>* (inclusive).

        (`errno` 63)
            Invalid structural data found on line *<value>*.

            The specified nonzero element is not in the upper triangle.

            The row index is *<value>* and column index is *<value>*.

        (`errno` 64)
            Invalid structural data found on line *<value>*.

            The specified element belongs to a diagonal block but is not
            diagonal.

            The row index is *<value>* and column index is *<value>*.

        (`errno` 65)
            An entry in the constraints with matno = *<value>*, blkno =
            *<value>*, row index *<value>* and column index *<value>*
            was defined more than once. All entries need to be unique.

        (`errno` 66)
            A premature end of the input stream. The part defining the
            number of variables was not found.

        (`errno` 66)
            A premature end of the input stream. The part defining the
            number of blocks was not found.

        (`errno` 66)
            A premature end of the input stream. The part defining the
            dimensions of the blocks was not found.

        (`errno` 66)
            A premature end of the input stream. The part defining the
            objective function was not found.

        (`errno` 66)
            A premature end of the input stream. The part defining the
            nonzero entries was not found.

        (`errno` 100)
            Line *<value>*: Incorrect ordering of indicator lines.

            OBJNAME indicator line found after ROWS indicator line.

        (`errno` 101)
            Line *<value>*: Incorrect ordering of indicator lines.

            COLUMNS indicator line found before ROWS indicator line.

        (`errno` 102)
            Line *<value>*: Incorrect ordering of indicator lines.

            RHS indicator line found before COLUMNS indicator line.

        (`errno` 103)
            Line *<value>*: Incorrect ordering of indicator lines.

            RANGES indicator line found before RHS indicator line.

        (`errno` 104)
            Line *<value>*: Incorrect ordering of indicator lines.

            BOUNDS indicator line found before COLUMNS indicator line.

        (`errno` 105)
            Line *<value>*: Incorrect ordering of indicator lines.

            QUADOBJ indicator line found before BOUNDS indicator line.

        (`errno` 106)
            Line *<value>*: Incorrect ordering of indicator lines.

            QUADOBJ indicator line found before COLUMNS indicator line.

        (`errno` 107)
            Line *<value>*: Incorrect ordering of indicator lines.

            CSECTION indicator line found before COLUMNS indicator line.

        (`errno` 108)
            Line *<value>*: Unknown indicator line *<value>*.

        (`errno` 109)
            Line *<value>*: Indicator line '*<value>*' has been found
            more than once in the MPS file.

        (`errno` 110)
            End of file found before ENDATA indicator line.

        (`errno` 111)
            At least one mandatory section not found in MPS file.

        (`errno` 112)
            Line *<value>*: An illegal line was detected in *<value>*
            section.

            This is neither a comment nor a valid data line.

        (`errno` 113)
            Line *<value>*: Unknown inequality key *<value>* in ROWS
            section.

            Expected '``N``', '``G``', '``L``' or '``E``'.

        (`errno` 114)
            Line *<value>*: Empty ROWS section.

            Neither the objective row nor the constraints were defined.

        (`errno` 115)
            The supplied name in OBJNAME of the objective row was not
            found among the free rows in the ROWS section.

        (`errno` 116)
            Line *<value>*: Illegal row name.

            Row names must consist of printable characters only.

        (`errno` 117)
            Line *<value>*: Illegal column name.

            Column names must consist of printable characters only.

        (`errno` 118)
            Line *<value>*: Row name *<value>* has been defined more
            than once in the ROWS section.

        (`errno` 119)
            Line *<value>*: Column '*<value>*' has been defined more
            than once in the COLUMNS section. Column definitions must be
            continuous. (See [COLUMNS Section]).

        (`errno` 120)
            Line *<value>*: Found ``'INTORG'`` marker within
            ``'INTORG'`` to ``'INTEND'`` range.

        (`errno` 121)
            Line *<value>*: Found ``'INTEND'`` marker without previous
            marker being ``'INTORG'``.

        (`errno` 122)
            Line *<value>*: Found ``'INTORG'`` but not ``'INTEND'``
            before the end of the COLUMNS section.

        (`errno` 123)
            Line *<value>*: Illegal marker type *<value>*.

            Should be either ``'INTORG'`` or ``'INTEND'``.

        (`errno` 124)
            Line *<value>*: Unknown row name *<value>* in *<value>*
            section.

            All row names must be specified in the ROWS section.

        (`errno` 125)
            Line *<value>*: Unknown column name *<value>* in *<value>*
            section.

            All column names must be specified in the COLUMNS section.

        (`errno` 126)
            Inconsistent bounds for row '*<value>*'.

        (`errno` 126)
            Inconsistent bounds for column '*<value>*'.

        (`errno` 126)
            Line *<value>*: Unknown bound type *<value>* in BOUNDS
            section.

        (`errno` 127)
            More than one nonzero of A has row name '*<value>*' and
            column name '*<value>*' in the COLUMNS section.

        (`errno` 128)
            Field *<value>* did not contain a number (see [MPS input
            format]).

        (`errno` 129)
            Both quadratic objective and cone constraints found, not
            supported.

        (`errno` 130)
            Line *<value>*: Second-order cone *<value>* should have at
            least 2 variables.

        (`errno` 130)
            Line *<value>*: Rotated second-order cone *<value>* should
            have at least 3 variables.

        (`errno` 131)
            Line *<value>*: Unknown cone type *<value>*.

        (`errno` 132)
            Line *<value>*: Illegal cone name.

            Cone names must consist of printable characters only.

        (`errno` 133)
            Line *<value>*: Cone name *<value>* has been defined more
            than once.

        (`errno` 134)
            Line *<value>*: Column name *<value>* has been defined more
            than once in cone constraint.

    References
    ----------
    Borchers, B, 1999, `SDPLIB 1.2, A Library of semidefinite
    programming test problems`, Optimization Methods and Software
    (11(1)), 683--690, http://euler.nmt.edu/~brian/sdplib/

    Fujisawa, K, Kojima, M and Nakata, K, 1998, `SDPA (Semidefinite
    Programming Algorithm) User's Manual`, Technical Report B-308,
    Department of Mathematical and Computing Sciences, Tokyo Institute
    of Technology.

    IBM, 1971, `MPSX -- Mathematical programming system`, Program Number
    5734 XM4, IBM Trade Corporation, New York
    """

    _pinfo = _b_utils._EngineIntArrayType.empty_ndarray(
        (100,),
    )
    _return_scalar = _b_opt.handle_read_file(
        file,
        ftype,
        _pinfo,
    )
    _handle_py = _return_scalar
    _pinfo_py = _pinfo
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['handle', 'pinfo'],
    )(
        _handle_py,
        _pinfo_py,
    )
    return _return_data

def handle_solve_ipopt(handle, x, nnzu, objfun=None, objgrd=None, confun=None, congrd=None, hess=None, monit=None, data=None, io_manager=None):
    u"""
    Run an interior point solver on a sparse nonlinear programming
    problem (NLP) initialized by ``handle_init`` and defined by other
    functions from the suite.

    Note: this function uses optional algorithmic parameters, see also:
    ``handle_opt_set``, ``handle_opt_get``.

    ``handle_solve_ipopt``, an interior point method optimization
    solver, based on the IPOPT software package, is a solver for the NAG
    optimization modelling suite and is suitable for large scale
    nonlinear programming (NLP) problems.

    For full information please refer to the NAG Library document for
    e04st

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04stf.html

    Parameters
    ----------
    handle : Handle
        The handle to the problem. It needs to be initialized by
        ``handle_init`` and the problem formulated by some of the
        functions ``handle_set_linobj``, ``handle_set_quadobj``,
        ``handle_set_nlnobj``, ``handle_set_simplebounds``,
        ``handle_set_linconstr``, ``handle_set_nlnconstr`` and
        ``handle_set_nlnhess``. It **must not** be changed between calls
        to the NAG optimization modelling suite.

    x : float, array-like, shape (nvar)
        x^0, the initial estimates of the variables x.

    nnzu : int
        The number of Lagrange multipliers that are to be returned in
        array `u`.

        If `nnzu` = 0, `u` will not be referenced; otherwise it needs to
        match the dimension q as explained in [Structure of the Lagrange
        Multipliers].

    objfun : None or callable fx = objfun(x, data=None), optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `objfun` must calculate the value of the nonlinear objective
        function f(x) at a specified value of the n-element vector of x
        variables.

        If there is no nonlinear objective (e.g., ``handle_set_linobj``
        or ``handle_set_quadobj`` was called to define a linear or
        quadratic objective function), `objfun` will never be called by
        ``handle_solve_ipopt`` and `objfun` may be None.

        Parameters
        ~~~~~~~~~~
        x : float, ndarray, shape (nvar)
            The vector x of variable values at which the objective
            function is to be evaluated.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        fx : float
            The value of the objective function at x.

    objgrd : None or callable fdx = objgrd(x, nnzfd, data=None),
        optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `objgrd` must calculate the values of the nonlinear objective
        function gradients df/dx at a specified value of the n-element
        vector of x variables.

        If there is no nonlinear objective (e.g., ``handle_set_linobj``
        or ``handle_set_quadobj`` was called to define a linear or
        quadratic objective function), `objgrd` will never be called by
        ``handle_solve_ipopt`` and `objgrd` may be None.

        Parameters
        ~~~~~~~~~~
        x : float, ndarray, shape (nvar)
            The vector x of variable values at which the objective
            function gradient is to be evaluated.

        nnzfd : int
            The number of nonzero elements in the sparse gradient vector
            of the objective function, as was set in a previous call to
            ``handle_set_nlnobj``.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        fdx : float, array-like, shape (`nnzfd`)
            The values of the nonzero elements in the sparse gradient
            vector of the objective function, in the order specified by
            idxfd in a previous call to ``handle_set_nlnobj``.
            `fdx`[i-1] will be the gradient df/dx_idxfd[i-1].

    confun : None or callable gx = confun(x, ncnln, data=None), optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `confun` must calculate the values of the m_g-element vector
        g_i(x) of nonlinear constraint functions at a specified value of
        the n-element vector of x variables.

        If no nonlinear constraints were registered in this `handle`,
        `confun` will never be called by ``handle_solve_ipopt`` and
        `confun` may be None.

        Parameters
        ~~~~~~~~~~
        x : float, ndarray, shape (nvar)
            The vector x of variable values at which the constraint
            functions are to be evaluated.

        ncnln : int
            m_g, the number of nonlinear constraints, as specified in an
            earlier call to ``handle_set_nlnconstr``.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        gx : float, array-like, shape (`ncnln`)
            The m_g values of the nonlinear constraint functions at x.

    congrd : None or callable gdx = congrd(x, nnzgd, data=None),
        optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `congrd` must calculate the nonzero values of the sparse
        Jacobian of the nonlinear constraint functions dg_i/dx at a
        specified value of the n-element vector of x variables.

        If there are no nonlinear constraints (e.g.,
        ``handle_set_nlnconstr`` was never called with the same handle
        or it was called with ncnln = 0), `congrd` will never be called
        by ``handle_solve_ipopt`` and `congrd` may be None.

        Parameters
        ~~~~~~~~~~
        x : float, ndarray, shape (nvar)
            The vector x of variable values at which the Jacobian of the
            constraint functions is to be evaluated.

        nnzgd : int
            Is the number of nonzero elements in the sparse Jacobian of
            the constraint functions, as was set in a previous call to
            ``handle_set_nlnconstr``.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        gdx : float, array-like, shape (`nnzgd`)
            The nonzero values of the Jacobian of the nonlinear
            constraints, in the order specified by irowgd and icolgd in
            an earlier call to ``handle_set_nlnconstr``. `gdx`[i-1] will
            be the gradient dg_irowgd[i-1]/dx_icolgd[i-1].

    hess : None or callable hx = hess(x, idf, sigma, lamda, nnzh,
        data=None), optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `hess` must calculate the nonzero values of one of a set of
        second derivative quantities:

            the Hessian of the Lagrangian function sigma*nabla^2f+sum_i
            = 1^m_g lambda_i*nabla^2g_i

            the Hessian of the objective function nabla^2f

            the Hessian of the constraint functions nabla^2g_i

        The value of argument `idf` determines which one of these is to
        be computed and this, in turn, is determined by earlier calls to
        ``handle_set_nlnhess``, when the nonzero sparsity structure of
        these Hessians was registered.

        Please note that it is not possible to only supply a subset of
        the Hessians (see `errno` = 6).

        If there were no calls to ``handle_set_nlnhess``, `hess` will
        never be called by ``handle_solve_ipopt`` and `hess` may be
        None. In this case, the Hessian of the Lagrangian will be
        approximated by a limited-memory quasi-Newton method (L-BFGS).

        Parameters
        ~~~~~~~~~~
        x : float, ndarray, shape (nvar)
            The vector x of variable values at which the Hessian
            functions are to be evaluated.

        idf : int
            Specifies the quantities to be computed in `hx`.

            `idf` = -1

                The values of the Hessian of the Lagrangian will be
                computed in `hx`. This will be the case if
                ``handle_set_nlnhess`` has been called with `idf` of the
                same value.

            `idf` = 0

                The values of the Hessian of the objective function will
                be computed in `hx`. This will be the case if
                ``handle_set_nlnhess`` has been called with `idf` of the
                same value.

            `idf` > 0

                The values of the Hessian of the `idf`th constraint
                function will be computed in `hx`. This will be the case
                if ``handle_set_nlnhess`` has been called with `idf` of
                the same value.

        sigma : float
            If `idf` = -1, the value of the sigma quantity in the
            definition of the Hessian of the Lagrangian. Otherwise,
            `sigma` should not be referenced.

        lamda : float, ndarray, shape (ncnln)
            If `idf` = -1, the values of the lambda_i quantities in the
            definition of the Hessian of the Lagrangian. Otherwise,
            `lamda` should not be referenced.

        nnzh : int
            The number of nonzero elements in the Hessian to be
            computed.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        hx : float, array-like, shape (`nnzh`)
            The nonzero values of the requested Hessian evaluated at x.
            For each value of `idf`, the ordering of nonzeros must
            follow the sparsity structure registered in the `handle` by
            earlier calls to ``handle_set_nlnhess`` through the
            arguments irowh and icolh.

    monit : None or callable monit(x, u, rinfo, stats, data=None),
        optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `monit` is provided to enable you to monitor the progress of the
        optimization.

        `monit` may be None.

        Parameters
        ~~~~~~~~~~
        x : float, ndarray, shape (nvar)
            x^i, the values of the decision variables x at the current
            iteration.

        u : float, ndarray, shape (nnzu)
            If `nnzu` > 0, `u` holds the values at the current iteration
            of Lagrange multipliers (dual variables) for the
            constraints. See [Structure of the Lagrange Multipliers] for
            layout information.

        rinfo : float, ndarray, shape (32)
            Error measures and various indicators at the end of the
            current iteration as described in [Description of the
            Printed Output].

        stats : float, ndarray, shape (32)
            Solver statistics at the end of the current iteration as
            described in [Description of the Printed Output].

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

    data : arbitrary, optional
        User-communication data for callback functions.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    x : float, ndarray, shape (nvar)
        The final values of the variables x.

    u : None or float, ndarray, shape (`nnzu`)
        The final value of Lagrange multipliers z,lambda.

    rinfo : float, ndarray, shape (32)
        Error measures and various indicators at the end of the final
        iteration as given in the table below:

        [table omitted]

    stats : float, ndarray, shape (32)
        Solver statistics at the end of the final iteration as given in
        the table below:

        [table omitted]

    Other Parameters
    ----------------
    'Defaults' : valueless
        This special keyword may be used to reset all options to their
        default values.
        Any value given with this keyword will be ignored.

    'Hessian Mode' : str
        Default = 'AUTO'

        This argument specifies whether the Hessian will be
        user-supplied (in `hx`) or approximated by
        ``handle_solve_ipopt`` using a limited-memory quasi-Newton
        L-BFGS method.
        In the 'AUTO' setting, if no Hessian structure has been
        registered in the problem with a call to ``handle_set_nlnhess``,
        and there are explicitly nonlinear user-supplied functions, then
        the Hessian will be approximated.
        Otherwise `hess` will be called if and only if any of
        ``handle_set_nlnobj`` or ``handle_set_nlnconstr`` have been used
        to define the problem.
        Approximating the Hessian is likely to require more iterations
        to achieve convergence but will reduce the time spent in
        user-supplied functions.

        Constraint: 'Hessian Mode' = 'AUTO', 'EXACT' or 'APPROXIMATE'.

    'Infinite Bound Size' : float
        Default = 10^20

        This defines the 'infinite' bound bigbnd in the definition of
        the problem constraints.
        Any upper bound greater than or equal to bigbnd will be regarded
        as + infinity (and similarly any lower bound less than or equal
        to -bigbnd will be regarded as - infinity).
        Note that a modification of this option does not influence
        constraints which have already been defined; only the
        constraints formulated after the change will be affected.

        It also serves as a limit for the objective function to be
        considered unbounded (`errno` = 54).

        Constraint: 'Infinite Bound Size' >= 1000.

    'Monitoring File' : int
        Default = -1

        If i >= 0, the unit number for the secondary (monitoring)
        output.
        If set to -1, no secondary output is provided.
        The information output to this unit is controlled by 'Monitoring
        Level'.

        Constraint: 'Monitoring File' >= -1.

    'Monitoring Level' : int
        Default = 4

        This argument sets the amount of information detail that will be
        printed by the solver to the secondary output.
        The meaning of the levels is the same as with 'Print Level'.

        Constraint: 0 <= 'Monitoring Level' <= 5.

    'Matrix Ordering' : str
        Default = 'AUTO'

        This argument specifies the ordering to be used by the internal
        sparse linear algebra solver.
        It affects the number of nonzeros in the factorized matrix and
        thus influences the cost per iteration.

        'Matrix Ordering' = 'AUTO'
            A heuristic is used to choose automatically between METIS
            and AMD orderings.

        'Matrix Ordering' = 'BEST'
            Both AMD and METIS orderings are computed at the beginning
            of the solve and the one with the fewest nonzeros in the
            factorized matrix is selected.

        'Matrix Ordering' = 'AMD'
            An approximate minimum degree (AMD) ordering is used.

        'Matrix Ordering' = 'METIS'
            METIS ordering is used.

        Constraint: 'Matrix Ordering' = 'AUTO', 'BEST', 'AMD' or
        'METIS'.

    'Outer Iteration Limit' : int
        Default = 100

        The maximum number of iterations to be performed by
        ``handle_solve_ipopt``.
        Setting the option too low might lead to `errno` = 22.

        Constraint: 'Outer Iteration Limit' >= 0.

    'Print File' : int
        Default = advisory message unit number

        If i >= 0, the unit number for the primary output of the solver.
        If 'Print File' = -1, the primary output is completely turned
        off independently of other settings. The default value is the
        advisory message unit number at the time of the options
        initialization, e.g., at the initialization of the handle. The
        information output to this unit is controlled by 'Print Level'.

        Constraint: 'Print File' >= -1.

    'Print Level' : int
        Default = 2

        This argument defines how detailed information should be printed
        by the solver to the primary output.

        [table omitted]

        Constraint: 0 <= 'Print Level' <= 5.

    'Stats Time' : str
        Default = 'NO'

        This argument allows you to turn on timings of various parts of
        the algorithm to give a better overview of where most of the
        time is spent.
        This might be helpful for a choice of different solving
        approaches.

        Constraint: 'Stats Time' = 'YES' or 'NO'.

    'Stop Tolerance 1' : float
        Default = max(10^(-6), sqrt(epsilon))

        This option sets the value epsilon_tol which is used for
        optimality and complementarity tests from KKT conditions See
        [Stopping Criteria].

        Constraint: 'Stop Tolerance 1' > epsilon.

    'Task' : str
        Default = 'MINIMIZE'

        This argument specifies the required direction of the
        optimization.
        If 'Task' = 'FEASIBLE POINT', the objective function (if set) is
        ignored and the algorithm stops as soon as a feasible point is
        found with respect to the given tolerance.
        If no objective function is set, 'Task' reverts to 'FEASIBLE
        POINT' automatically.

        Constraint: 'Task' = 'MINIMIZE', 'MAXIMIZE' or 'FEASIBLE POINT'.

    'Time Limit' : float
        Default = 10^6

        A limit to the number of seconds that the solver can use to
        solve one problem.
        If during the convergence check this limit is exceeded, the
        solver will terminate with a corresponding error message.

        Constraint: 'Time Limit' > 0.

    'Verify Derivatives' : str
        Default = 'NO'

        This argument specifies whether the function should perform
        numerical checks on the consistency of the user-supplied
        functions.
        It is recommended that such checks are enabled when first
        developing the formulation of the problem, however, the
        derivative check results in a significant increase of the number
        of the function evaluations and thus it shouldn't be used in
        production code.

        Constraint: 'Verify Derivatives' = 'YES' or 'NO'.

    Raises
    ------
    NagValueError
        (`errno` -199)
            ``handle_solve_ipopt`` is not available in this
            implementation.

        (`errno` 1)
            `handle` has not been initialized.

        (`errno` 1)
            `handle` does not belong to the NAG optimization modelling
            suite, has not been initialized properly or is corrupted.

        (`errno` 1)
            `handle` has not been initialized properly or is corrupted.

        (`errno` 2)
            The problem is already being solved.

        (`errno` 2)
            This solver does not support the model defined in the
            handle.

        (`errno` 3)
            A different solver from the suite has already been used.
            Initialize a new `handle` using ``handle_init``.

        (`errno` 4)
            The information supplied does not match with that previously
            stored.

            On entry, nvar = *<value>* must match that given during
            initialization of the `handle`, i.e., *<value>*.

        (`errno` 5)
            On entry, `nnzu` = *<value>*.

            Constraint: `nnzu` = *<value>* or 0.

        (`errno` 5)
            On entry, `nnzu` = *<value>*.

            Constraint: no constraints present, so `nnzu` must be 0.

        (`errno` 6)
            On entry, a nonlinear objective function has been defined
            but no objective Hessian sparsity structure has been defined
            through ``handle_set_nlnhess``.

        (`errno` 6)
            On entry, a nonlinear constraint function has been defined
            but no constraint Hessian sparsity structure has been
            defined through ``handle_set_nlnhess``, for constraint
            number *<value>*.

        (`errno` 7)
            Either change the option 'Hessian Mode' or provide a proper
            `hess` function.

        (`errno` 7)
            Please provide a proper `objfun` function.

        (`errno` 7)
            Please provide a proper `objgrd` function.

        (`errno` 7)
            Please provide a proper `confun` function.

        (`errno` 7)
            Please provide a proper `congrd` function.

        (`errno` 23)
            The solver terminated with not enough degrees of freedom.

        (`errno` 23)
            The solver terminated due to an invalid problem definition.

        (`errno` 23)
            The solver terminated due to an invalid option.

        (`errno` 51)
            The solver detected an infeasible problem.

        (`errno` 54)
            The solver terminated due to diverging iterates.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 22)
            Maximum number of iterations exceeded.

        (`errno` 50)
            The solver reports NLP solved to acceptable level.

    NagAlgorithmicMajorWarning
        (`errno` 20)
            User requested termination during a monitoring step.

        (`errno` 23)
            The solver terminated after failure in the restoration
            phase.

        (`errno` 23)
            The solver terminated after an error in the step
            computation.

        (`errno` 23)
            The solver terminated after the maximum time allowed was
            exceeded.

        (`errno` 24)
            The solver terminated after the search direction became too
            small.

        (`errno` 25)
            Invalid number detected in user function.

    References
    ----------
    Byrd, R H, Gilbert, J Ch and Nocedal, J, 2000, `A trust region
    method based on interior point techniques for nonlinear
    programming`, Mathematical Programming (89), 149--185

    Byrd, R H, Liu, G and Nocedal, J, 1997, `On the local behavior of an
    interior point method for nonlinear programming`, Numerical
    Analysis, (eds D F Griffiths and D J Higham), Addison--Wesley

    Conn, A R, Gould, N I M, Orban, D and Toint, Ph L, 2000, `A
    primal-dual trust-region algorithm for non-convex nonlinear
    programming`, Mathematical Programming (87 (2)), 215--249

    Conn, A R, Gould, N I M and Toint, Ph L, 2000, `Trust Region
    Methods`, SIAM, Philadephia

    Fiacco, A V and McCormick, G P, 1990, `Nonlinear Programming:
    Sequential Unconstrained Minimization Techniques`, SIAM,
    Philadelphia

    Gould, N I M, Orban, D, Sartenaer, A and Toint, Ph L, 2001,
    `Superlinear convergence of primal-dual interior point algorithms
    for nonlinear programming`, SIAM Journal on Optimization (11 (4)),
    974--1002

    Hock, W and Schittkowski, K, 1981, `Test Examples for Nonlinear
    Programming Codes. Lecture Notes in Economics and Mathematical
    Systems` (187), Springer--Verlag

    Hogg, J D and Scott, J A, 2011, `HSL MA97: a bit-compatible
    multifrontal code for sparse symmetric systems`, RAL Technical
    Report. RAL-TR-2011-024

    Wächter, A and Biegler, L T, 2006, `On the implementation of a
    primal-dual interior point filter line search algorithm for
    large-scale nonlinear programming`, Mathematical Programming
    (106(1)), 25--57

    Williams, P and Lang, B, 2013, `A framework for the MR^3 Algorithm:
    theory and implementation`, SIAM J. Sci. Comput. (35), 740--766

    Yamashita, H, 1998, `A globally convergent primal-dual
    interior-point method for constrained optimization`, Optimization
    Methods and Software (10), 443--469

    See Also
    --------
    :meth:`naginterfaces.library.examples.opt.handle_solve_ipopt_ex.main`
    """

    def cb_objfun(nvar, x, data=None):
        """Wrapping user's objfun. Called as base objfun."""
        if data is not None:
            _fx_objfun = objfun(
                x,
                data,
            )
        else:
            _fx_objfun = objfun(
                x,
            )
        _fx_objfun_base = _fx_objfun
        return _fx_objfun_base

    fun_name = 'naginterfaces.library.opt.handle_solve_ipopt'

    cb_objgrd_name = fun_name + ':objgrd'

    def cb_objgrd(nvar, x, nnzfd, fdx, data=None):
        """Wrapping user's objgrd. Called as base objgrd."""
        if data is not None:
            _fdx_objgrd = objgrd(
                x,
                nnzfd,
                data,
            )
        else:
            _fdx_objgrd = objgrd(
                x,
                nnzfd,
            )
        fdx[:] = _utils._to_ndarray(
            {'fun_name': cb_objgrd_name, 'returned': 'fdx'},
            _fdx_objgrd,
            copy=False,
            exp_shape=(nnzfd,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )

    cb_confun_name = fun_name + ':confun'

    def cb_confun(nvar, x, ncnln, gx, data=None):
        """Wrapping user's confun. Called as base confun."""
        if data is not None:
            _gx_confun = confun(
                x,
                ncnln,
                data,
            )
        else:
            _gx_confun = confun(
                x,
                ncnln,
            )
        gx[:] = _utils._to_ndarray(
            {'fun_name': cb_confun_name, 'returned': 'gx'},
            _gx_confun,
            copy=False,
            exp_shape=(ncnln,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )

    cb_congrd_name = fun_name + ':congrd'

    def cb_congrd(nvar, x, nnzgd, gdx, data=None):
        """Wrapping user's congrd. Called as base congrd."""
        if data is not None:
            _gdx_congrd = congrd(
                x,
                nnzgd,
                data,
            )
        else:
            _gdx_congrd = congrd(
                x,
                nnzgd,
            )
        gdx[:] = _utils._to_ndarray(
            {'fun_name': cb_congrd_name, 'returned': 'gdx'},
            _gdx_congrd,
            copy=False,
            exp_shape=(nnzgd,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )

    cb_hess_name = fun_name + ':hess'

    def cb_hess(nvar, x, ncnln, idf, sigma, lamda, nnzh, hx, data=None):
        """Wrapping user's hess. Called as base hess."""
        if data is not None:
            _hx_hess = hess(
                x,
                idf,
                sigma,
                lamda,
                nnzh,
                data,
            )
        else:
            _hx_hess = hess(
                x,
                idf,
                sigma,
                lamda,
                nnzh,
            )
        hx[:] = _utils._to_ndarray(
            {'fun_name': cb_hess_name, 'returned': 'hx'},
            _hx_hess,
            copy=False,
            exp_shape=(nnzh,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )

    def cb_monit(nvar, x, nnzu, u, rinfo, stats, data=None):
        """Wrapping user's monit. Called as base monit."""
        if data is not None:
            monit(
                x,
                u,
                rinfo,
                stats,
                data,
            )
        else:
            monit(
                x,
                u,
                rinfo,
                stats,
            )

    _b_utils._check_callable(objfun, {'fun_name': fun_name, 'entity': 'objfun'}, allow_none=True)
    _objfun = None if objfun is None else cb_objfun
    _b_utils._check_callable(objgrd, {'fun_name': fun_name, 'entity': 'objgrd'}, allow_none=True)
    _objgrd = None if objgrd is None else cb_objgrd
    _b_utils._check_callable(confun, {'fun_name': fun_name, 'entity': 'confun'}, allow_none=True)
    _confun = None if confun is None else cb_confun
    _b_utils._check_callable(congrd, {'fun_name': fun_name, 'entity': 'congrd'}, allow_none=True)
    _congrd = None if congrd is None else cb_congrd
    _b_utils._check_callable(hess, {'fun_name': fun_name, 'entity': 'hess'}, allow_none=True)
    _hess = None if hess is None else cb_hess
    _b_utils._check_callable(monit, {'fun_name': fun_name, 'entity': 'monit'}, allow_none=True)
    _monit = None if monit is None else cb_monit
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nvar_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _nvar = _b_utils._EngineIntScalarType(
        _nvar_py, {'fun_name': fun_name, 'entity': '_nvar'},
    ).to_ctypes()
    _nnzu = _b_utils._EngineIntScalarType(
        nnzu, {'fun_name': fun_name, 'entity': 'nnzu'},
    ).to_ctypes()
    _nnzu_py = _b_utils._EngineIntScalarType.to_py(_nnzu)
    if _nnzu_py > 0:
        _u = _b_utils._EngineFloat64ArrayType.empty_ndarray(
            (_nnzu_py,),
        )
    else:
        _u = None
    _rinfo = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (32,),
    )
    _stats = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (32,),
    )
    try:
        _b_opt.handle_solve_ipopt(
            handle,
            _objfun,
            _objgrd,
            _confun,
            _congrd,
            _hess,
            _monit,
            _nvar,
            _x,
            _nnzu,
            _u,
            _rinfo,
            _stats,
            data,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _x_py = _x
    if _nnzu_py > 0:
        _u_py = _u
    else:
        _u_py = None
    _rinfo_py = _rinfo
    _stats_py = _stats
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'u', 'rinfo', 'stats'],
    )(
        _x_py,
        _u_py,
        _rinfo_py,
        _stats_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def handle_solve_pennon(handle, x, inform, u=None, uc=None, ua=None, io_manager=None):
    u"""
    Run the Pennon solver on a compatible problem initialized by
    ``handle_init`` and defined by other functions from the suite, such
    as, semidefinite programming (SDP) and SDP with bilinear matrix
    inequalities (BMI).

    Note: this function uses optional algorithmic parameters, see also:
    ``handle_opt_set``, ``handle_opt_get``.

    ``handle_solve_pennon`` is a solver from the NAG optimization
    modelling suite for problems such as, quadratic programming (QP),
    linear semidefinite programming (SDP) and semidefinite programming
    with bilinear matrix inequalities (BMI-SDP).

    For full information please refer to the NAG Library document for
    e04sv

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04svf.html

    Parameters
    ----------
    handle : Handle
        The handle to the problem. It needs to be initialized by
        ``handle_init`` and **must not** be changed before the call to
        ``handle_solve_pennon``.

    x : float, array-like, shape (nvar)
        Note: intermediate stops take place only if 'Monitor Frequency'
        > 0.

        If 'Initial X' = 'USER' (the default), x^0, the initial estimate
        of the variables x, otherwise `x` need not be set.

        `On intermediate entry`: the input is ignored.

    inform : int
        Note: intermediate stops take place only if 'Monitor Frequency'
        > 0.

        `On initial entry`: no effect.

        `On intermediate entry`: if set to 0, solving the current
        problem is terminated and the function returns `errno` = 20;
        otherwise the function continues.

    u : None or float, array-like, shape (nnzu), optional
        Note: intermediate stops take place only if 'Monitor Frequency'
        > 0.

        Note: if nnzu > 0, `u` holds Lagrangian multipliers (dual
        variables) for (standard) constraints, i.e., simple bounds
        defined by ``handle_set_simplebounds`` and a set of m_B linear
        constraints defined by ``handle_set_linconstr``. Either their
        initial estimates, intermediate approximations or final values,
        see [Structure of the Lagrangian Multipliers].

        Note: if nnzu = 0, `u` will not be referenced.

        If 'Initial U' = 'USER' (the default is 'AUTOMATIC'), u^0, the
        initial estimate of the Lagrangian multipliers u, otherwise `u`
        need not be set.

        `On intermediate entry`: the input is ignored.

    uc : None or float, array-like, shape (nnzuc), optional
        `uc` is reserved for future releases of the NAG Library which
        will allow definition of second-order cone constraints.

        It is not referenced at the moment.

    ua : None or float, array-like, shape (nnzua), optional
        Note: intermediate stops take place only if 'Monitor Frequency'
        > 0.

        If nnzua > 0, `ua` holds the Lagrangian multipliers for matrix
        constraints defined by ``handle_set_linmatineq`` and
        ``handle_set_quadmatineq``, see [Structure of the Lagrangian
        Multipliers].

        If nnzua = 0, `ua` will not be referenced.

        If 'Initial U' = 'USER' (the default is 'AUTOMATIC'), U^0, the
        initial estimate of the matrix Lagrangian multipliers U,
        otherwise `ua` need not be set.

        `On intermediate entry`: the input is ignored.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    x : float, ndarray, shape (nvar)
        `On intermediate exit`: the value of the variables x at the end
        of the current outer iteration.

        `On final exit`: the final value of the variables x.

    u : float, ndarray, shape (nnzu)
        `On intermediate exit`: the estimate of the multipliers u at the
        end of the current outer iteration.The final value of
        multipliers u.

    uc : float, ndarray, shape (nnzuc)
        `uc` is reserved for future releases of the NAG Library which
        will allow definition of second-order cone constraints.

        It is not referenced at the moment.

    ua : float, ndarray, shape (nnzua)
        `On intermediate exit`: the estimate of the matrix multipliers U
        at the end of the outer iteration.

        `On final exit`: the final estimate of the multipliers U.

    rinfo : float, ndarray, shape (32)
        `On intermediate or final entry`: error measures and various
        indicators (see [Algorithmic Details] for details) at the end of
        the current (or final) outer iteration as given in the table
        below:

        [table omitted]

    stats : float, ndarray, shape (32)
        `On intermediate or final exit`: solver statistics at the end of
        the current (or final) outer iteration as given in the table
        below. Note that time statistics is provided only if 'Stats
        Time' is set (the default is 'NO'), the measured time is
        returned in seconds.

        [table omitted]

    inform : int
        Note: intermediate stops take place only if 'Monitor Frequency'
        > 0.

        `On intermediate exit`: `inform` = 1.

        `On final exit`: `inform` = 0.

    Other Parameters
    ----------------
    'Defaults' : valueless
        This special keyword may be used to reset all options to their
        default values.
        Any value given with this keyword will be ignored.

    'DIMACS Measures' : str
        Default = 'CHECK'

        If the problem is a linear semidefinite programming problem,
        this argument specifies if DIMACS error measures (see [Stopping
        Criteria]) should be computed and/or checked.
        In other cases, this option reverts to 'NO' automatically.

        Constraint: 'DIMACS Measures' = 'COMPUTE', 'CHECK' or 'NO'.

    'Hessian Density' : str
        Default = 'AUTO'

        This option guides the solver on how the Hessian matrix of
        augmented Lagrangian F(x,u,v,U,p,P) should be built.
        Option 'AUTO' leaves the decision to the solver and it is the
        recommended option.
        Setting it to 'DENSE' bypasses the autodetection and the Hessian
        is always built as a dense matrix.
        Option 'SPARSE' instructs the solver to use a sparse storage and
        factorization of the matrix if possible.

        Constraint: 'Hessian Density' = 'AUTO', 'DENSE' or 'SPARSE'

    'Infinite Bound Size' : float
        Default = 10^20

        This defines the 'infinite' bound bigbnd in the definition of
        the problem constraints.
        Any upper bound greater than or equal to bigbnd will be regarded
        as + infinity (and similarly any lower bound less than or equal
        to -bigbnd will be regarded as - infinity).
        Note that a modification of this option does not influence
        constraints which have already been defined; only the
        constraints formulated after the change will be affected.

        Constraint: 'Infinite Bound Size' >= 1000.

    'Initial P' : str
        Default = 'AUTOMATIC'

        This option defines the choice of the penalty options p^0, P^0,
        see [Algorithm 1].

        'Initial P' = 'AUTOMATIC'
            The penalty options are chosen automatically as set by
            option 'Init Value P', 'Init Value Pmat' and subject to
            automatic scaling. Note that P^0 might be increased so that
            the penalty function Phi_P() is defined for all matrix
            constraints at the starting point.

        'Initial P' = 'KEEP PREVIOUS'
            The penalty options are kept from the previous run of the
            solver if possible. If not, this options reverts to
            'AUTOMATIC'. Note that even if the matrix penalty options
            are the same as in the previous run, they are still subject
            to a possible increase so that the penalty function Phi_P()
            is well defined at the starting point.

        Constraint: 'Initial P' = 'AUTOMATIC' or 'KEEP PREVIOUS'.

    'Initial U' : str
        Default = 'AUTOMATIC'

        This argument guides the solver on which initial Lagrangian
        multipliers are to be used.

        'Initial U' = 'AUTOMATIC'
            The Lagrangian multipliers are chosen automatically as set
            by automatic scaling.

        'Initial U' = 'USER'
            The values of arrays `u` and `ua` (if provided) are used as
            the initial Lagrangian multipliers subject to automatic
            adjustments. If one or the other array is not provided, the
            choice for missing data is as in 'AUTOMATIC'.

        'Initial U' = 'KEEP PREVIOUS'
            The Lagrangian multipliers are kept from the previous run of
            the solver. If this option is set for the first run or
            options change the approach of the solver, the choice
            automatically reverts to 'AUTOMATIC'. This might be useful
            if the solver is hot started, for example, to achieve higher
            precision of the solution.

        Constraint: 'Initial U' = 'AUTOMATIC', 'USER' or 'KEEP
        PREVIOUS'.

    'Initial X' : str
        Default = 'USER'

        This argument guides which starting point x^0 is to be used.

        'Initial X' = 'AUTOMATIC'
            The starting point is chosen automatically so that it
            satisfies simple bounds on the variables or as a zero
            vector. Input of argument `x` is ignored.

        'Initial X' = 'USER'
            Initial values of argument `x` are used as a starting point.

        Constraint: 'Initial X' = 'AUTOMATIC' or 'USER'.

    'Init Value P' : float
        Default = 1.0

        This argument defines the value p^0, the initial penalty option
        for (standard) inequalities.
        A low value of the penalty causes the solution of the inner
        problem to be closer to the feasible region and thus to the
        desirable result.
        However, it also increases ill-conditioning of the system.
        It is not advisable to set the penalty too low unless a good
        starting point is provided.

        Constraint: fourthroot(epsilon) <= 'Init Value P' <= 10^4.

    'Init Value Pmat' : float
        Default = 1.0

        The value of this option suggests P^0, the initial penalty
        option for matrix inequalities.
        It is similar to 'Init Value P' (and the same advice applies),
        however, P^0 gets increased automatically if the matrix
        constraints are more infeasible than the actual penalty option.

        Constraint: fourthroot(epsilon) <= 'Init Value Pmat' <= 10^4.

    'Inner Iteration Limit' : int
        Default = 100

        The maximum number of the inner iterations (Newton steps) to be
        performed by [Algorithm 2] in each outer iteration.
        Setting the option too low might lead to `errno` = 23.
        Values higher than 100 are unlikely to improve convergence.

        Constraint: 'Inner Iteration Limit' > 0.

    'Inner Stop Criteria' : str
        Default = 'HEURISTIC'

        The precision alpha for the solution of the inner subproblem is
        determined in [Algorithm 1] and under typical circumstances
        [Algorithm 2] is expected to reach this precision within the
        given 'Inner Iteration Limit'.
        If any problems are detected and 'Inner Stop Criteria' =
        'HEURISTIC', [Algorithm 2] is allowed to stop before reaching
        the requested precision or the 'Inner Iteration Limit'.
        This usually saves many unfruitful iterations and the solver may
        recover in the following iterations.
        If you suspect that the heuristic problem detection is not
        suitable for your problem, setting 'Inner Stop Criteria' =
        'STRICT' disallows such behaviour.

        Constraint: 'Inner Stop Criteria' = 'HEURISTIC' or 'STRICT'.

    'Inner Stop Tolerance' : float
        Default = 10^(-2)

        This option sets the required precision alpha^0 for the first
        inner problem solved by [Algorithm 2].
        The precison of the solution of the inner problem does not need
        to be very high in the first outer iterations and it is
        automatically adjusted through the outer iterations to reach the
        optimality limit epsilon_2 in the last one.

        Setting alpha^0 too restrictive (too low) causes an increase of
        the number of inner iterations needed in the first outer
        iterations and might lead to `errno` = 23.
        In certain cases it might be helpful to use a more relaxed
        (higher) alpha^0 and increase 'P Update Speed' which should
        reduce the number of inner iterations needed at the beginning of
        the computation in exchange for a possibly higher number of the
        outer iterations.

        Constraint: epsilon < 'Inner Stop Tolerance' <= 10^3.

    'Linesearch Mode' : str
        Default = 'AUTO'

        This controls the step size selection in [Algorithm 2].
        If 'Linesearch Mode' = 'FULLSTEP' (the default for linear
        problems), unit steps are taken where possible and the step
        shortening takes place only to avoid undefined regions for the
        matrix penalty function Phi_P() (see (6)).
        This may be used for linear problems but it is not recommended
        for any nonlinear ones.
        If 'Linesearch Mode' = 'ARMIJO', Armijo backtracking linesearch
        is used instead which is a fairly basic linesearch.
        If 'Linesearch Mode' = 'GOLDSTEIN', a cubic safe guarded
        linesearch based on Goldstein condition is employed, this is the
        recommended (and default) choice for nonlinear problems.

        Constraint: 'Linesearch Mode' = 'AUTO', 'FULLSTEP', 'ARMIJO' or
        'GOLDSTEIN'.

    'List' : str
        Default = 'NO'

        This argument may be set to 'YES' if you wish to turn on
        printing of each option specification as it is supplied.

        Constraint: 'List' = 'YES' or 'NO'

    'Monitor Frequency' : int
        Default = 0

        If 'Monitor Frequency' > 0, the solver returns to you at the end
        of every ith outer iteration.
        During these intermediate exits, the current point `x` and
        Lagrangian multipliers `u`, `ua` (if requested) are provided as
        well as the statistics and error measures (`rinfo`, `stats`).
        Argument `inform` helps to distinguish between intermediate and
        final exits and also allows immediate termination.

        If 'Monitor Frequency' = 0, the solver stops only once on the
        final point and no intermediate exits are made.

        Constraint: 'Monitor Frequency' >= 0.

    'Monitoring File' : int
        Default = -1

        If i >= 0, the unit number for the secondary (monitoring)
        output.
        If set to -1, no secondary output is provided.
        The following information is output to the unit:

        -   a listing of the options;

        -   problem statistics, the iteration log and the final status
            as set by 'Monitoring Level'.

        Constraint: 'Monitoring File' >= -1.

    'Monitoring Level' : int
        Default = 4

        This argument sets the amount of information detail that will be
        printed by the solver to the secondary output.
        The meaning of the levels is the same as with 'Print Level'.

        Constraint: 0 <= 'Monitoring Level' <= 5.

    'Outer Iteration Limit' : int
        Default = 100

        The maximum number of the outer iterations to be performed by
        [Algorithm 1].
        If 'Outer Iteration Limit' = 0, no iteration is performed, only
        quantities needed in the stopping criteria are computed and
        returned in `rinfo`.
        This might be useful in connection with 'Initial X' = 'USER' and
        'Initial U' = 'USER' to check optimality of the given point.
        However, note that the rules for possible modifications of the
        starting point still apply, see `u` and `ua`.
        Setting the option too low might lead to `errno` = 22.

        Constraint: 'Outer Iteration Limit' >= 0.

    'P Min' : float
        Default = sqrt(epsilon)

        This controls p_min, the lowest possible penalty value p used
        for (standard) inequalities.
        In general, very small values of the penalty options cause
        ill-conditioning which might lead to numerical difficulties.
        On the other hand, very high p_min prevents the algorithm from
        reaching the requested accuracy on the feasibility.
        Under normal circumstances, the default value is recommended.

        Constraint: epsilon <= 'P Min' <= 10^(-2).

    'Pmat Min' : float
        Default = sqrt(epsilon)

        This is an equivalent of 'P Min' for the minimal matrix penalty
        option P_min.
        The same advice applies.

        Constraint: epsilon <= 'Pmat Min' <= 10^(-2).

    'Preference' : str
        Default = 'SPEED'

        This option affects how contributions from the matrix
        constraints (6) to the system Hessian matrix are computed.
        The default option of 'Preference' = 'SPEED' should be suitable
        in most cases.
        However, dealing with matrix constraints of a very high
        dimension may cause noticable memory overhead and switching to
        'Preference' = 'MEMORY' may be required.

        Constraint: 'Preference' = 'SPEED' or 'MEMORY'.

    'Presolve Block Detect' : str
        Default = 'YES'

        If 'Presolve Block Detect' = 'YES', the matrix constraints are
        checked during preprocessoring to determine if they can be split
        into smaller independent ones, thus speeding up the solver.

        Constraint: 'Presolve Block Detect' = 'YES' or 'NO'.

    'Print File' : int
        Default = advisory message unit number

        If i >= 0, the unit number for the primary output of the solver.
        If 'Print File' = -1, the primary output is completely turned
        off independently of other settings. The default value is the
        advisory message unit number at the time of the options
        initialization, e.g., at the initialization of the handle. The
        following information is output to the unit:

        -   a listing of options if set by 'Print Options';

        -   problem statistics, the iteration log and the final status
            from the solver as set by 'Print Level'.

        Constraint: 'Print File' >= -1.

    'Print Level' : int
        Default = 2

        This argument defines how detailed information should be printed
        by the solver to the primary output.

        [table omitted]

        Constraint: 0 <= 'Print Level' <= 5.

    'Print Options' : str
        Default = 'YES'

        If 'Print Options' = 'YES', a listing of options will be printed
        to the primary output.

        Constraint: 'Print Options' = 'YES' or 'NO'.

    'P Update Speed' : int
        Default = 12

        This option affects the rate at which the penalty options p,P
        are updated ([Algorithm 1], step (3)) and thus indirectly
        influences the overall number of outer iterations.
        Its value can be interpretted as the typical number of outer
        iterations needed to get from the initial penalty values p^0,
        P^0 half-way to the p_min and P_min.
        Values smaller than 3 causes a very agressive penalty update
        strategy which might lead to the increased number of inner
        iterations and possibly to numerical difficulties.
        On the other hand, values higher than 15 produce a relatively
        conservative approach which leads to a higher number of the
        outer iterations.

        If the solver encounters difficulties on your problem, a higher
        value might help.
        If your problem is working fine, setting a lower value might
        increase the speed.

        Constraint: 1 <= 'P Update Speed' <= 100.

    'Stats Time' : str
        Default = 'NO'

        This argument turns on timings of various parts of the algorithm
        to give a better overview of where most of the time is spent.
        This might be helpful for a choice of different solving
        approaches.
        It is possible to choose between CPU and wall clock time.
        Choice 'YES' is equivalent to 'WALL CLOCK'.

        Constraint: 'Stats Time' = 'YES', 'NO', 'CPU' or 'WALL CLOCK'.

    'Stop Criteria' : str
        Default = 'SOFT'

        If 'Stop Criteria' = 'SOFT', the solver is allowed to stop
        prematurely with a suboptimal solution, `errno` = 50, if it
        predicts that a better estimate of the solution cannot be
        reached.
        This is the recommended option.

        Constraint: 'Stop Criteria' = 'SOFT' or 'STRICT'.

    'Stop Tolerance 1' : float
        Default = max(10^(-6), sqrt(epsilon))

        This option defines epsilon_1 used as a tolerance for the
        relative duality gap (0) and the relative precision (1), see
        [Stopping Criteria].

        Constraint: 'Stop Tolerance 1' > epsilon.

    'Stop Tolerance 2' : float
        Default = max(10^(-7), sqrt(epsilon))

        This option sets the value epsilon_2 which is used for
        optimality (2) and complementarity (4) tests from KKT conditions
        or if 'DIMACS Measures' = 'Check' for all DIMACS error measures
        instead.
        See [Stopping Criteria].

        Constraint: 'Stop Tolerance 2' > epsilon.

    'Stop Tolerance Feasibility' : float
        Default = max(10^(-7), sqrt(epsilon))

        This argument places an acceptance limit on the feasibility of
        the solution (3), epsilon_feas.
        See [Stopping Criteria].

        Constraint: 'Stop Tolerance Feasibility' > epsilon.

    'Task' : str
        Default = 'MINIMIZE'

        This argument specifies the required direction of the
        optimization.
        If 'Task' = 'FEASIBLE POINT', the objective function (if set) is
        ignored and the algorithm stops as soon as a feasible point is
        found with respect to the given tolerance.
        If no objective function was set, 'Task' reverts to 'FEASIBLE
        POINT' automatically.

        Constraint: 'Task' = 'MINIMIZE', 'MAXIMIZE' or 'FEASIBLE POINT'.

    'Transform Constraints' : str
        Default = 'AUTO'

        This argument controls how equality constraints are treated by
        the solver.
        If 'Transform Constraints' = 'EQUALITIES', all equality
        constraints h_k(x) = 0 from (4) are treated as two inequalities
        h_k(x) <= 0 and h_k(x) >= 0, see [Solution of the inner
        problem].
        This is the default and the only option in this release for
        equality constrained problems.

        Constraint: 'Transform Constraints' = 'AUTO', 'NO' or
        'EQUALITIES'.

    'U Update Restriction' : float
        Default = 0.5

        This defines the value mu_g giving the bounds on the updates of
        Lagrangian multipliers for (standard) inequalities between the
        outer iterations.
        Values close to 1 limit the changes of the multipliers and serve
        as a kind of smoothing, lower values allow more significant
        changes.

        Based on numerical experience, big variation in the multipliers
        may lead to a large number of iterations in the subsequent step
        and might disturb the convergence due to ill-conditioning.

        It might be worth experimenting with the value on your
        particular problem.
        Mid range values are recommended over the more extremal ones.

        Constraint: epsilon < 'U Update Restriction' < 1.

    'Umat Update Restriction' : float
        Default = 0.3

        This is an equivalent of 'U Update Restriction' for matrix
        constraints, denoted as mu_A in [Overview].
        The advice above applies equally.

        Constraint: epsilon < 'Umat Update Restriction' < 1.

    Raises
    ------
    NagValueError
        (`errno` 1)
            `handle` has not been initialized.

        (`errno` 1)
            `handle` does not belong to the NAG optimization modelling
            suite, has not been initialized properly or is corrupted.

        (`errno` 1)
            `handle` has not been initialized properly or is corrupted.

        (`errno` 2)
            This solver does not support the model defined in the
            handle.

        (`errno` 2)
            This solver does not support fixed variables.

        (`errno` 3)
            A different solver from the suite has already been used.

        (`errno` 4)
            On entry, nvar = *<value>*, expected value = *<value>*.

            Constraint: nvar must match the value given during
            initialization of `handle`.

        (`errno` 5)
            On entry, nnzu = *<value>*.

            nnzu does not match the size of the Lagrangian multipliers
            for (standard) constraints.

            nnzu = 0 or *<value>*.

        (`errno` 5)
            On entry, nnzu = *<value>*.

            nnzu does not match the size of the Lagrangian multipliers
            for (standard) constraints.

            nnzu = 0 when there are no (standard) constraints.

        (`errno` 5)
            On entry, nnzua = *<value>*.

            nnzua does not match the size of the Lagrangian multipliers
            for matrix constraints.

            nnzua = 0 or *<value>*.

        (`errno` 5)
            On entry, nnzua = *<value>*.

            nnzua does not match the size of the Lagrangian multipliers
            for matrix constraints.

            nnzua = 0 when there are no matrix constraints.

        (`errno` 5)
            On entry, nnzuc = *<value>*.

            nnzuc does not match the size of the Lagrangian multipliers
            for second-order cone constraints.

            nnzuc = 0 when there are no second-order cone constraints.

        (`errno` 21)
            The current starting point is unusable.

        (`errno` 23)
            The inner subproblem could not be solved to the required
            accuracy.

            Inner iteration limit has been reached.

        (`errno` 23)
            The inner subproblem could not be solved to the required
            accuracy.

            Limited progress in the inner subproblem triggered a stop
            (heuristic inner stop criteria).

        (`errno` 23)
            The inner subproblem could not be solved to the required
            accuracy.

            Line search or another internal component failed.

        (`errno` 51)
            The problem was found to be infeasible during preprocessing.

        (`errno` 52)
            The problem was found unbounded during preprocessing.

        (`errno` 53)
            The problem seems to be infeasible, the algorithm was
            stopped.

        (`errno` 54)
            The problem seems to be unbounded, the algorithm was
            stopped.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 50)
            The algorithm converged to a suboptimal solution.

            The full accuracy was not achieved. The solution should
            still be usable.

    NagAlgorithmicMajorWarning
        (`errno` 20)
            User requested termination during a monitoring step.

        (`errno` 22)
            Outer iteration limit has been reached.

            The requested accuracy is not achieved.

        (`errno` 24)
            Unable to make progress, the algorithm was stopped.

    References
    ----------
    Ben--Tal, A and Zibulevsky, M, 1997, `Penalty/barrier multiplier
    methods for convex programming problems`, SIAM Journal on
    Optimization (7), 347--366

    Fujisawa, K, Kojima, M, Nakata, K, 1997, `Exploiting sparsity in
    primal-dual interior-point method for semidefinite programming`,
    Math. Programming (79), 235--253

    Hogg, J D and Scott, J A, 2011, `HSL MA97: a bit-compatible
    multifrontal code for sparse symmetric systems`, RAL Technical
    Report. RAL-TR-2011-024

    Kočvara, M and Stingl, M, 2003, `PENNON -- a code for convex
    nonlinear and semidefinite programming`, Optimization Methods and
    Software (18(3)), 317--333

    Kočvara, M and Stingl, M, 2007, `On the solution of large-scale SDP
    problems by the modified barrier method using iterative solvers`,
    Math. Programming (Series B) (109(2--3)), 413--444

    Mittelmann, H D, 2003, `An independent benchmarking of SDP and SOCP
    solvers`, Math. Programming (95), 407--430

    Stingl, M, 2006, `On the Solution of Nonlinear Semidefinite Programs
    by Augmented Lagrangian Methods, PhD thesis`, Institute of Applied
    Mathematics II, Friedrich--Alexander University of
    Erlangen--Nuremberg

    See Also
    --------
    :meth:`naginterfaces.library.examples.opt.handle_solve_pennon_bmi_ex.main`
    :meth:`naginterfaces.library.examples.opt.handle_solve_pennon_lmi_ex.main`
    """

    _nvar, _nnzu, _nnzuc, _nnzua = _b_utils._e04ptan(handle)
    _nvar_py = _b_utils._EngineIntScalarType.to_py(_nvar)
    _vl_e04pta_nvar_py = _nvar_py
    fun_name = 'naginterfaces.library.opt.handle_solve_pennon'

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nnzu_py = _b_utils._EngineIntScalarType.to_py(_nnzu)
    _vl_e04pta_nnzu_py = _nnzu_py
    _nnzu_py = _b_utils._EngineIntScalarType.to_py(_nnzu)
    if u is None:
        _u = _b_utils._EngineFloat64ArrayType.empty_ndarray(
            (_nnzu_py,),
        )
    else:
        _u = _utils._to_ndarray(
            {'fun_name': fun_name, 'entity': 'u'},
            u,
            copy=True,
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
    _nnzuc_py = _b_utils._EngineIntScalarType.to_py(_nnzuc)
    _vl_e04pta_nnzuc_py = _nnzuc_py
    _nnzuc_py = _b_utils._EngineIntScalarType.to_py(_nnzuc)
    if uc is None:
        _uc = _b_utils._EngineFloat64ArrayType.empty_ndarray(
            (_nnzuc_py,),
        )
    else:
        _uc = _utils._to_ndarray(
            {'fun_name': fun_name, 'entity': 'uc'},
            uc,
            copy=True,
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
    _nnzua_py = _b_utils._EngineIntScalarType.to_py(_nnzua)
    _vl_e04pta_nnzua_py = _nnzua_py
    _nnzua_py = _b_utils._EngineIntScalarType.to_py(_nnzua)
    if ua is None:
        _ua = _b_utils._EngineFloat64ArrayType.empty_ndarray(
            (_nnzua_py,),
        )
    else:
        _ua = _utils._to_ndarray(
            {'fun_name': fun_name, 'entity': 'ua'},
            ua,
            copy=True,
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
    _rinfo = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (32,),
    )
    _stats = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (32,),
    )
    try:
        _return_scalar = _b_opt.handle_solve_pennon(
            handle,
            _nvar,
            _x,
            _nnzu,
            _u,
            _nnzuc,
            _uc,
            _nnzua,
            _ua,
            _rinfo,
            _stats,
            inform,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalar = _the_exc.return_data[0]
    _x_py = _x
    _inform_py = _return_scalar
    _u_py = _u
    _uc_py = _uc
    _ua_py = _ua
    _rinfo_py = _rinfo
    _stats_py = _stats
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'u', 'uc', 'ua', 'rinfo', 'stats', 'inform'],
    )(
        _x_py,
        _u_py,
        _uc_py,
        _ua_py,
        _rinfo_py,
        _stats_py,
        _inform_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def nlp1_solve(a, bl, bu, objfun, x, comm, confun=None, istate=None, cjac=None, clamda=None, r=None, data=None, io_manager=None, spiked_sorder='C'):
    u"""
    Nonlinear programming (NLP), dense, active-set SQP method, using
    function values and optionally first derivatives, recommended.

    Note: this function uses optional algorithmic parameters, see also:
    ``nlp1_option_file``, ``nlp1_option_string``, ``nlp1_init``.

    ``nlp1_solve`` is designed to minimize an arbitrary smooth function
    subject to constraints (which may include simple bounds on the
    variables, linear constraints and smooth nonlinear constraints)
    using a sequential quadratic programming (SQP) method.
    As many first derivatives as possible should be supplied by you; any
    unspecified derivatives are approximated by finite differences.
    It is not intended for large sparse problems.

    ``nlp1_solve`` may also be used for unconstrained, bound-constrained
    and linearly constrained optimization.

    ``nlp1_solve`` uses **forward** **communication** for evaluating the
    objective function, the nonlinear constraint functions, and any of
    their derivatives.

    For full information please refer to the NAG Library document for
    e04uc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04ucf.html

    Parameters
    ----------
    a : float, array-like, shape (nclin, :)
        Note: the required extent for this argument in dimension 2 is
        determined as follows: if nclin>0: n; otherwise: 1.

        The ith row of `a` contains the ith row of the matrix A_L of
        general linear constraints in (1). That is, the ith row contains
        the coefficients of the ith general linear constraint, for i =
        1,2,...,nclin.

    bl : float, array-like, shape (n+nclin+ncnln)
        `bl` must contain the lower bounds for all the constraints

    bu : float, array-like, shape (n+nclin+ncnln)
        `bu` must contain the upper bounds for all the constraints

    objfun : callable (objf, objgrd) = objfun(mode, x, objgrd, nstate,
        data=None)
        `objfun` must calculate the objective function F(x) and
        (optionally) its gradient g(x) = dF/dx for a specified n-vector
        x.

        Parameters
        ~~~~~~~~~~
        mode : int
            Indicates which values must be assigned during each call of
            `objfun`. Only the following values need be assigned:

            `mode` = 0

                `objf`.

            `mode` = 1

                All available elements of `objgrd`.

            `mode` = 2

                `objf` and all available elements of `objgrd`.

        x : float, ndarray, shape (n)
            x, the vector of variables at which the objective function
            and/or all available elements of its gradient are to be
            evaluated.

        objgrd : float, ndarray, shape (n)
            The elements of `objgrd` are set to special values which
            enable ``nlp1_solve`` to detect whether they are changed by
            `objfun`.

        nstate : int
            If `nstate` = 1, then ``nlp1_solve`` is calling `objfun` for
            the first time. This argument setting allows you to save
            computation time if certain data must be read or calculated
            only once.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        objf : float
            If `mode` = 0 or 2, `objf` must be set to the value of the
            objective function at x.

        objgrd : float, array-like, shape (n)
            If `mode` = 1 or 2, `objgrd` must return the available
            elements of the gradient evaluated at x.

    x : float, array-like, shape (n)
        An initial estimate of the solution.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``nlp1_init``.

    confun : None or callable (c, cjac) = confun(mode, needc, x, cjac,
        nstate, data=None), optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `confun` must calculate the vector c(x) of nonlinear constraint
        functions and (optionally) its Jacobian (= dc/dx) for a
        specified n-element vector x.

        If there are no nonlinear constraints (i.e., ncnln = 0),
        `confun` will never be called by ``nlp1_solve`` and `confun` may
        be None. If there are nonlinear constraints, the first call to
        `confun` will occur before the first call to `objfun`.

        Parameters
        ~~~~~~~~~~
        mode : int
            Indicates which values must be assigned during each call of
            `confun`. Only the following values need be assigned, for
            each value of i such that `needc`[i-1] > 0:

            `mode` = 0

                `c`[i-1].

            `mode` = 1

                All available elements in the ith row of `cjac`.

            `mode` = 2

                `c`[i-1] and all available elements in the ith row of
                `cjac`.

        needc : int, ndarray, shape (ncnln)
            The indices of the elements of `c` and/or `cjac` that must
            be evaluated by `confun`. If `needc`[i-1] > 0, the ith
            element of `c` and/or the available elements of the ith row
            of `cjac` (see argument `mode`) must be evaluated at x.

        x : float, ndarray, shape (n)
            x, the vector of variables at which the constraint functions
            and/or the available elements of the constraint Jacobian are
            to be evaluated.

        cjac : float, ndarray, shape (max(1, ncnln), n)
            Unless 'Derivative Level' = 2 or 3, the elements of `cjac`
            are set to special values which enable ``nlp1_solve`` to
            detect whether they are changed by `confun`.

        nstate : int
            If `nstate` = 1, then ``nlp1_solve`` is calling `confun` for
            the first time. This argument setting allows you to save
            computation time if certain data must be read or calculated
            only once.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        c : float, array-like, shape (ncnln)
            If `needc`[i-1] > 0 and `mode` = 0 or 2, `c`[i-1] must
            contain the value of the ith constraint at x. The remaining
            elements of `c`, corresponding to the non-positive elements
            of `needc`, are ignored.

        cjac : float, array-like, shape (max(1, ncnln), n)
            If `needc`[i-1] > 0 and `mode` = 1 or 2, the ith row of
            `cjac` must contain the available elements of the vector
            nabla c_i given by

                nabla c_i = (dc_i/dx_1,dc_i/dx_2,...,dc_i/dx_n)^T,

            where dc_i/dx_j is the partial derivative of the ith
            constraint with respect to the jth variable, evaluated at
            the point x. See also the argument `nstate`. The remaining
            rows of `cjac`, corresponding to non-positive elements of
            `needc`, are ignored.

            If all elements of the constraint Jacobian are known (i.e.,
            'Derivative Level' = 2 or 3), any constant elements may be
            assigned to `cjac` one time only at the start of the
            optimization.

            An element of `cjac` that is not subsequently assigned in
            `confun` will retain its initial value throughout.

            Constant elements may be loaded into `cjac` either before
            the call to ``nlp1_solve`` or during the first call to
            `confun` (signalled by the value `nstate` = 1).

            The ability to preload constants is useful when many
            Jacobian elements are identically zero, in which case `cjac`
            may be initialized to zero and nonzero elements may be reset
            by `confun`.

            Note that constant nonzero elements do affect the values of
            the constraints.

            Thus, if `cjac`[i-1,j-1] is set to a constant value, it need
            not be reset in subsequent calls to `confun`, but the value
            `cjac`[i-1,j-1]*`x`[j-1] must nonetheless be added to
            `c`[i-1].

            For example, if `cjac`[0,0] = 2 and `cjac`[0,1] = -5, then
            the term 2*`x`[0]-5*`x`[1] must be included in the
            definition of `c`[0].

            It must be emphasized that, if 'Derivative Level' = 0 or 1,
            unassigned elements of `cjac` are not treated as constant;
            they are estimated by finite differences, at nontrivial
            expense.

            If you do not supply a value for the option 'Difference
            Interval', an interval for each element of x is computed
            automatically at the start of the optimization.

            The automatic procedure can usually identify constant
            elements of `cjac`, which are then computed once only by
            finite differences.

    istate : None or int, array-like, shape (n+nclin+ncnln), optional
        Need not be set if the (default) option 'Cold Start' is used.

        If the option 'Warm Start' has been chosen, the elements of
        `istate` corresponding to the bounds and linear constraints
        define the initial working set for the procedure that finds a
        feasible point for the linear constraints and bounds.

        The active set at the conclusion of this procedure and the
        elements of `istate` corresponding to nonlinear constraints then
        define the initial working set for the first QP subproblem.

        More precisely, the first n elements of `istate` refer to the
        upper and lower bounds on the variables, the next n_L elements
        refer to the upper and lower bounds on A_L x, and the next n_N
        elements refer to the upper and lower bounds on c(x).

        Possible values for `istate`[j-1] are as follows:

        [table omitted]

        The values -2, -1 and 4 are also acceptable but will be modified
        by the function.

        If ``nlp1_solve`` has been called previously with the same
        values of n, nclin and ncnln, `istate` already contains
        satisfactory information.

        The function also adjusts (if necessary) the values supplied in
        `x` to be consistent with `istate`.

    cjac : None or float, array-like, shape (ncnln, :), optional
        Note: the required extent for this argument in dimension 2 is
        determined as follows: if ncnln>0: n; otherwise: 1.

        In general, `cjac` need not be initialized before the call to
        ``nlp1_solve``. However, if 'Derivative Level' = 2 or 3, you may
        optionally set the constant elements of `cjac` (see argument
        `nstate` in the description of `confun`). Such constant elements
        need not be re-assigned on subsequent calls to `confun`.

    clamda : None or float, array-like, shape (n+nclin+ncnln), optional
        Need not be set if the (default) option 'Cold Start' is used.

        If the option 'Warm Start' has been chosen, `clamda`[j-1] must
        contain a multiplier estimate for each nonlinear constraint with
        a sign that matches the status of the constraint specified by
        the `istate` array, for j = n+nclin+1,...,n+nclin+ncnln.

        The remaining elements need not be set.

        Note that if the jth constraint is defined as 'inactive' by the
        initial value of `istate` array (i.e., `istate`[j-1] = 0),
        `clamda`[j-1] should be zero; if the jth constraint is an
        inequality active at its lower bound (i.e., `istate`[j-1] = 1),
        `clamda`[j-1] should be non-negative; if the jth constraint is
        an inequality active at its upper bound (i.e., `istate`[j-1] =
        2), `clamda`[j-1] should be non-positive.

        If necessary, the function will modify `clamda` to match these
        rules.

    r : None or float, array-like, shape (n, n), optional
        Need not be initialized if the (default) option 'Cold Start' is
        used.

        If the option 'Warm Start' has been chosen, `r` must contain the
        upper triangular Cholesky factor R of the initial approximation
        of the Hessian of the Lagrangian function, with the variables in
        the natural order.

        Elements not in the upper triangular part of `r` are assumed to
        be zero and need not be assigned.

    data : arbitrary, optional
        User-communication data for callback functions.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    spiked_sorder : str, optional
        If `a` and `r` are spiked (i.e., have unit extent in all but one
        dimension, or have size 1), `spiked_sorder` selects the storage
        order to associate with them in the NAG Engine:

        spiked_sorder = 'C'
          row-major storage will be used;

        spiked_sorder = 'F'
          column-major storage will be used.

        Two-dimensional arrays returned from callback functions in this
        routine must then use the same storage order.

    Returns
    -------
    itera : int
        The number of major iterations performed.

    istate : int, ndarray, shape (n+nclin+ncnln)
        The status of the constraints in the QP working set at the point
        returned in `x`. The significance of each possible value of
        `istate`[j-1] is as follows:

        [table omitted]

    c : None or float, ndarray, shape (max(1, ncnln))
        If ncnln > 0, `c`[i-1] contains the value of the ith nonlinear
        constraint function c_i at the final iterate, for i =
        1,2,...,ncnln.

        If ncnln = 0, `c` is returned as None.

    cjac : None or float, ndarray, shape (ncnln, :)
        If ncnln > 0, `cjac` contains the Jacobian matrix of the
        nonlinear constraint functions at the final iterate, i.e.,
        `cjac`[i-1,j-1] contains the partial derivative of the ith
        constraint function with respect to the jth variable, for j =
        1,2,...,n, for i = 1,2,...,ncnln. (See the discussion of
        argument `cjac` under `confun`.)

        If ncnln = 0, `cjac` is returned as None.

    clamda : float, ndarray, shape (n+nclin+ncnln)
        The values of the QP multipliers from the last QP subproblem.
        `clamda`[j-1] should be non-negative if `istate`[j-1] = 1 and
        non-positive if `istate`[j-1] = 2.

    objf : float
        The value of the objective function at the final iterate.

    objgrd : float, ndarray, shape (n)
        The gradient of the objective function at the final iterate (or
        its finite difference approximation).

    r : float, ndarray, shape (n, n)
        If 'Hessian' = 'NO', `r` contains the upper triangular Cholesky
        factor R of Q^TH~Q, an estimate of the transformed and reordered
        Hessian of the Lagrangian at x (see (6) in [Overview]). If
        'Hessian' = 'YES', `r` contains the upper triangular Cholesky
        factor R of H, the approximate (untransformed) Hessian of the
        Lagrangian, with the variables in the natural order.

    x : float, ndarray, shape (n)
        The final estimate of the solution.

    Other Parameters
    ----------------
    'Central Difference Interval' : float
        Default values are computed

        If the algorithm switches to central differences because the
        forward-difference approximation is not sufficiently accurate,
        the value of r is used as the difference interval for every
        element of x.
        The switch to central differences is indicated by C at the end
        of each line of intermediate printout produced by the major
        iterations (see [Further Comments]).
        The use of finite differences is discussed further under the
        option 'Difference Interval'.

        If you supply a value for this option, a small value between 0.0
        and 1.0 is appropriate.

    'Cold Start' : valueless
        Default

        This option controls the specification of the initial working
        set in both the procedure for finding a feasible point for the
        linear constraints and bounds and in the first QP subproblem
        thereafter.
        With a 'Cold Start', the first working set is chosen by
        ``nlp1_solve`` based on the values of the variables and
        constraints at the initial point.
        Broadly speaking, the initial working set will include equality
        constraints and bounds or inequality constraints that violate or
        'nearly' satisfy their bounds (to within 'Crash Tolerance').

        With a 'Warm Start', you must set the `istate` array and define
        `clamda` and `r` as discussed in [Parameters]. `istate` values
        associated with bounds and linear constraints determine the
        initial working set of the procedure to find a feasible point
        with respect to the bounds and linear constraints. `istate`
        values associated with nonlinear constraints determine the
        initial working set of the first QP subproblem after such a
        feasible point has been found. ``nlp1_solve`` will override your
        specification of `istate` if necessary, so that a poor choice of
        the working set will not cause a fatal error.
        For instance, any elements of `istate` which are set to -2, -1
        or 4 will be reset to zero, as will any elements which are set
        to 3 when the corresponding elements of `bl` and `bu` are not
        equal.
        A warm start will be advantageous if a good estimate of the
        initial working set is available -- for example, when
        ``nlp1_solve`` is called repeatedly to solve related problems.

    'Warm Start' : valueless
        This option controls the specification of the initial working
        set in both the procedure for finding a feasible point for the
        linear constraints and bounds and in the first QP subproblem
        thereafter.
        With a 'Cold Start', the first working set is chosen by
        ``nlp1_solve`` based on the values of the variables and
        constraints at the initial point.
        Broadly speaking, the initial working set will include equality
        constraints and bounds or inequality constraints that violate or
        'nearly' satisfy their bounds (to within 'Crash Tolerance').

        With a 'Warm Start', you must set the `istate` array and define
        `clamda` and `r` as discussed in [Parameters]. `istate` values
        associated with bounds and linear constraints determine the
        initial working set of the procedure to find a feasible point
        with respect to the bounds and linear constraints. `istate`
        values associated with nonlinear constraints determine the
        initial working set of the first QP subproblem after such a
        feasible point has been found. ``nlp1_solve`` will override your
        specification of `istate` if necessary, so that a poor choice of
        the working set will not cause a fatal error.
        For instance, any elements of `istate` which are set to -2, -1
        or 4 will be reset to zero, as will any elements which are set
        to 3 when the corresponding elements of `bl` and `bu` are not
        equal.
        A warm start will be advantageous if a good estimate of the
        initial working set is available -- for example, when
        ``nlp1_solve`` is called repeatedly to solve related problems.

    'Crash Tolerance' : float
        Default = 0.01

        This value is used in conjunction with the option 'Cold Start'
        (the default value) when ``nlp1_solve`` selects an initial
        working set.
        If 0 <= r <= 1, the initial working set will include (if
        possible) bounds or general inequality constraints that lie
        within r of their bounds.
        In particular, a constraint of the form a_j^Tx >= l will be
        included in the initial working set if |a_j^Tx-l| <= r(1+|l|).
        If r < 0 or r > 1, the default value is used.

    'Defaults' : valueless
        This special keyword may be used to reset all options to their
        default values.

    'Derivative Level' : int
        Default = 3

        This argument indicates which derivatives are provided in
        functions `objfun` and `confun`.
        The possible choices for i are the following.

        [table omitted]

        The value i = 3 should be used whenever possible, since
        ``nlp1_solve`` is more reliable (and will usually be more
        efficient) when all derivatives are exact.

        If i = 0 or 2, ``nlp1_solve`` will estimate the unspecified
        elements of the objective gradient, using finite differences.
        The computation of finite difference approximations usually
        increases the total run-time, since a call to `objfun` is
        required for each unspecified element.
        Furthermore, less accuracy can be attained in the solution (see
        Module 8 of Gill `et al.` (1981), for a discussion of limiting
        accuracy).

        If i = 0 or 1, ``nlp1_solve`` will approximate unspecified
        elements of the constraint Jacobian.
        One call to `confun` is needed for each variable for which
        partial derivatives are not available.
        For example, if the Jacobian has the form

            [table omitted]

        where '*' indicates an element provided by you and '?' indicates
        an unspecified element, ``nlp1_solve`` will call `confun` twice:
        once to estimate the missing element in column 2, and again to
        estimate the two missing elements in column 3. (Since columns 1
        and 4 are known, they require no calls to `confun`.)

        At times, central differences are used rather than forward
        differences, in which case twice as many calls to `objfun` and
        `confun` are needed. (The switch to central differences is not
        under your control.)

        If i < 0 or i > 3, the default value is used.

    'Difference Interval' : float
        Default values are computed

        This option defines an interval used to estimate derivatives by
        finite differences in the following circumstances:

        (a) For verifying the objective and/or constraint gradients (see
            the description of the option 'Verify').

        (#) For estimating unspecified elements of the objective
            gradient or the constraint Jacobian.

        In general, a derivative with respect to the jth variable is
        approximated using the interval delta_j, where delta_j =
        r(1+|x^_j|), with x^ the first point feasible with respect to
        the bounds and linear constraints.
        If the functions are well scaled, the resulting derivative
        approximation should be accurate to O(r).
        See Gill `et al.` (1981) for a discussion of the accuracy in
        finite difference approximations.

        If a difference interval is not specified by you, a finite
        difference interval will be computed automatically for each
        variable by a procedure that requires up to six calls of
        `confun` and `objfun` for each element.
        This option is recommended if the function is badly scaled or
        you wish to have ``nlp1_solve`` determine constant elements in
        the objective and constraint gradients (see the descriptions of
        `confun` and `objfun` in [Parameters]).

        If you supply a value for this option, a small value between 0.0
        and 1.0 is appropriate.

    'Feasibility Tolerance' : float
        Default = sqrt(epsilon)

        The scalar r defines the maximum acceptable `absolute`
        violations in linear and nonlinear constraints at a 'feasible'
        point; i.e., a constraint is considered satisfied if its
        violation does not exceed r.
        If r < epsilon or r >= 1, the default value is used.
        Using this keyword sets both options 'Linear Feasibility
        Tolerance' and 'Nonlinear Feasibility Tolerance' to r, if
        epsilon <= r < 1. (Additional details are given under the
        descriptions of these options.)

    'Function Precision' : float
        Default = epsilon^0.9

        This argument defines epsilon_r, which is intended to be a
        measure of the accuracy with which the problem functions F(x)
        and c(x) can be computed.
        If r < epsilon or r >= 1, the default value is used.

        The value of epsilon_r should reflect the relative precision of
        1+|F(x)|; i.e., epsilon_r acts as a relative precision when |F|
        is large and as an absolute precision when |F| is small.
        For example, if F(x) is typically of order 1000 and the first
        six significant digits are known to be correct, an appropriate
        value for epsilon_r would be 10^(-6).
        In contrast, if F(x) is typically of order 10^(-4) and the first
        six significant digits are known to be correct, an appropriate
        value for epsilon_r would be 10^(-10).
        The choice of epsilon_r can be quite complicated for badly
        scaled problems; see Module 8 of Gill `et al.` (1981) for a
        discussion of scaling techniques.
        The default value is appropriate for most simple functions that
        are computed with full accuracy.
        However, when the accuracy of the computed function values is
        known to be significantly worse than full precision, the value
        of epsilon_r should be large enough so that ``nlp1_solve`` will
        not attempt to distinguish between function values that differ
        by less than the error inherent in the calculation.

    'Hessian' : str
        Default = NO

        This option controls the contents of the upper triangular matrix
        R (see [Parameters]). ``nlp1_solve`` works exclusively with the
        `transformed and reordered` Hessian H_Q (6), and hence extra
        computation is required to form the Hessian itself.
        If 'Hessian' = 'NO', `r` contains the Cholesky factor of the
        transformed and reordered Hessian.
        If 'Hessian' = 'YES', the Cholesky factor of the approximate
        Hessian itself is formed and stored in `r`.
        You should select 'Hessian' = 'YES' if a 'Warm Start' will be
        used for the next call to ``nlp1_solve``.

    'Infinite Bound Size' : float
        Default = 10^20

        If r > 0, r defines the 'infinite' bound bigbnd in the
        definition of the problem constraints.
        Any upper bound greater than or equal to bigbnd will be regarded
        as + infinity (and similarly any lower bound less than or equal
        to -bigbnd will be regarded as - infinity).
        If r < 0, the default value is used.

    'Infinite Step Size' : float
        Default = max(bigbnd, 10^20)

        If r > 0, r specifies the magnitude of the change in variables
        that is treated as a step to an unbounded solution.
        If the change in x during an iteration would exceed the value of
        r, the objective function is considered to be unbounded below in
        the feasible region.
        If r <= 0, the default value is used.

    'Line Search Tolerance' : float
        Default = 0.9

        The value r (0 <= r < 1) controls the accuracy with which the
        step alpha taken during each iteration approximates a minimum of
        the merit function along the search direction (the smaller the
        value of r, the more accurate the linesearch).
        The default value r = 0.9 requests an inaccurate search and is
        appropriate for most problems, particularly those with any
        nonlinear constraints.

        If there are no nonlinear constraints, a more accurate search
        may be appropriate when it is desirable to reduce the number of
        major iterations -- for example, if the objective function is
        cheap to evaluate, or if a substantial number of derivatives are
        unspecified.
        If r < 0 or r >= 1, the default value is used.

    'Linear Feasibility Tolerance' : float
        Default = sqrt(epsilon)

        The default value of r_2 is epsilon^0.33 if 'Derivative Level' =
        0 or 1, and sqrt(epsilon) otherwise.

        The scalars r_1 and r_2 define the maximum acceptable `absolute`
        violations in linear and nonlinear constraints at a 'feasible'
        point; i.e., a linear constraint is considered satisfied if its
        violation does not exceed r_1, and similarly for a nonlinear
        constraint and r_2.
        If r_m < epsilon or r_m >= 1, the default value is used, for m =
        1,2,...,2.

        On entry to ``nlp1_solve``, an iterative procedure is executed
        in order to find a point that satisfies the linear constraints
        and bounds on the variables to within the tolerance r_1.
        All subsequent iterates will satisfy the linear constraints to
        within the same tolerance (unless r_1 is comparable to the
        finite difference interval).

        For nonlinear constraints, the feasibility tolerance r_2 defines
        the largest constraint violation that is acceptable at an
        optimal point.
        Since nonlinear constraints are generally not satisfied until
        the final iterate, the value of option 'Nonlinear Feasibility
        Tolerance' acts as a partial termination criterion for the
        iterative sequence generated by ``nlp1_solve`` (see the
        discussion of option 'Optimality Tolerance').

        These tolerances should reflect the precision of the
        corresponding constraints.
        For example, if the variables and the coefficients in the linear
        constraints are of order unity, and the latter are correct to
        about 6 decimal digits, it would be appropriate to specify r_1
        as 10^(-6).

    'Nonlinear Feasibility Tolerance' : float
        Default = epsilon^0.33 or sqrt(epsilon)

        The default value of r_2 is epsilon^0.33 if 'Derivative Level' =
        0 or 1, and sqrt(epsilon) otherwise.

        The scalars r_1 and r_2 define the maximum acceptable `absolute`
        violations in linear and nonlinear constraints at a 'feasible'
        point; i.e., a linear constraint is considered satisfied if its
        violation does not exceed r_1, and similarly for a nonlinear
        constraint and r_2.
        If r_m < epsilon or r_m >= 1, the default value is used, for m =
        1,2,...,2.

        On entry to ``nlp1_solve``, an iterative procedure is executed
        in order to find a point that satisfies the linear constraints
        and bounds on the variables to within the tolerance r_1.
        All subsequent iterates will satisfy the linear constraints to
        within the same tolerance (unless r_1 is comparable to the
        finite difference interval).

        For nonlinear constraints, the feasibility tolerance r_2 defines
        the largest constraint violation that is acceptable at an
        optimal point.
        Since nonlinear constraints are generally not satisfied until
        the final iterate, the value of option 'Nonlinear Feasibility
        Tolerance' acts as a partial termination criterion for the
        iterative sequence generated by ``nlp1_solve`` (see the
        discussion of option 'Optimality Tolerance').

        These tolerances should reflect the precision of the
        corresponding constraints.
        For example, if the variables and the coefficients in the linear
        constraints are of order unity, and the latter are correct to
        about 6 decimal digits, it would be appropriate to specify r_1
        as 10^(-6).

    'List' : valueless
        Option 'List' enables printing of each option specification as
        it is supplied. 'Nolist' suppresses this printing.

    'Nolist' : valueless
        Default for ``nlp1_solve``

        Option 'List' enables printing of each option specification as
        it is supplied. 'Nolist' suppresses this printing.

    'Major Iteration Limit' : int
        Default = max(50, 3*(n+n_L)+10*n_N)

        The value of i specifies the maximum number of major iterations
        allowed before termination.
        Setting i = 0 and 'Major Print Level' > 0 means that the
        workspace needed will be computed and printed, but no iterations
        will be performed.
        If i < 0, the default value is used.

    'Iteration Limit' : int
        Default = max(50, 3*(n+n_L)+10*n_N)

        The value of i specifies the maximum number of major iterations
        allowed before termination.
        Setting i = 0 and 'Major Print Level' > 0 means that the
        workspace needed will be computed and printed, but no iterations
        will be performed.
        If i < 0, the default value is used.

    'Iters' : int
        Default = max(50, 3*(n+n_L)+10*n_N)

        The value of i specifies the maximum number of major iterations
        allowed before termination.
        Setting i = 0 and 'Major Print Level' > 0 means that the
        workspace needed will be computed and printed, but no iterations
        will be performed.
        If i < 0, the default value is used.

    'Itns' : int
        Default = max(50, 3*(n+n_L)+10*n_N)

        The value of i specifies the maximum number of major iterations
        allowed before termination.
        Setting i = 0 and 'Major Print Level' > 0 means that the
        workspace needed will be computed and printed, but no iterations
        will be performed.
        If i < 0, the default value is used.

    'Major Print Level' : int
        The value of i controls the amount of printout produced by the
        major iterations of ``nlp1_solve``, as indicated below.
        A detailed description of the printed output is given in
        [Further Comments] (summary output at each major iteration and
        the final solution) and [Monitoring Information] (monitoring
        information at each major iteration). (See also the description
        of the option 'Minor Print Level'.)

        The following printout is sent to the file object associated
        with the advisory I/O unit (see
        :class:`~naginterfaces.base.utils.FileObjManager`):

        [table omitted]

        The following printout is sent to the unit number given by the
        option 'Monitoring File':

        [table omitted]

        If 'Major Print Level' >= 5 and the unit number defined by the
        option 'Monitoring File' is the advisory unit number, the
        summary output for each major iteration is suppressed.

    'Print Level' : int
        Default for ``nlp1_solve`` = 0

        The value of i controls the amount of printout produced by the
        major iterations of ``nlp1_solve``, as indicated below.
        A detailed description of the printed output is given in
        [Further Comments] (summary output at each major iteration and
        the final solution) and [Monitoring Information] (monitoring
        information at each major iteration). (See also the description
        of the option 'Minor Print Level'.)

        The following printout is sent to the file object associated
        with the advisory I/O unit (see
        :class:`~naginterfaces.base.utils.FileObjManager`):

        [table omitted]

        The following printout is sent to the unit number given by the
        option 'Monitoring File':

        [table omitted]

        If 'Major Print Level' >= 5 and the unit number defined by the
        option 'Monitoring File' is the advisory unit number, the
        summary output for each major iteration is suppressed.

    'Minor Iteration Limit' : int
        Default = max(50, 3*(n+n_L+n_N))

        The value of i specifies the maximum number of iterations for
        finding a feasible point with respect to the bounds and linear
        constraints (if any).
        The value of i also specifies the maximum number of minor
        iterations for the optimality phase of each QP subproblem.
        If i <= 0, the default value is used.

    'Minor Print Level' : int
        Default = 0

        The value of i controls the amount of printout produced by the
        minor iterations of ``nlp1_solve`` (i.e., the iterations of the
        quadratic programming algorithm), as indicated below.
        A detailed description of the printed output is given in
        [Further Comments] (summary output at each minor iteration and
        the final QP solution) and [Monitoring Information] (monitoring
        information at each minor iteration). (See also the description
        of the option 'Major Print Level'.)

        The following printout is sent to the file object associated
        with the advisory I/O unit (see
        :class:`~naginterfaces.base.utils.FileObjManager`):

        [table omitted]

        The following printout is sent to the unit number given by the
        option 'Monitoring File':

        [table omitted]

        If 'Major Print Level' >= 5 and the unit number defined by the
        option 'Monitoring File' is the advisory unit number, the
        summary output for each major iteration is suppressed.

    'Monitoring File' : int
        Default = -1

        If i >= 0 and 'Major Print Level' >= 5 or i >= 0 and 'Minor
        Print Level' >= 5, monitoring information produced by
        ``nlp1_solve`` at every iteration is sent to a file with logical
        unit number i.
        If i < 0 and/or 'Major Print Level' < 5 and 'Minor Print Level'
        < 5, no monitoring information is produced.

    'Optimality Tolerance' : float
        Default = epsilon_r^0.8

        The argument r (epsilon_r <= r < 1) specifies the accuracy to
        which you wish the final iterate to approximate a solution of
        the problem.
        Broadly speaking, r indicates the number of correct figures
        desired in the objective function at the solution.
        For example, if r is 10^(-6) and ``nlp1_solve`` terminates
        successfully, the final value of F should have approximately six
        correct figures.
        If r < epsilon_r or r >= 1, the default value is used.

        ``nlp1_solve`` will terminate successfully if the iterative
        sequence of x values is judged to have converged and the final
        point satisfies the first-order Kuhn--Tucker conditions (see
        [Overview]).
        The sequence of iterates is considered to have converged at x if

            alpha||p|| <= sqrt(r)(1+||x||),

        where p is the search direction and alpha the step length from
        (3).
        An iterate is considered to satisfy the first-order conditions
        for a minimum if

            || Z^Tg_FR|| <= sqrt(r)(1+max(1+|F(x)|,||g_FR||))

        and

            |res_j| <= ftol for all j,

        where Z^Tg_FR is the projected gradient (see [Overview]), g_FR
        is the gradient of F(x) with respect to the free variables,
        res_j is the violation of the jth active nonlinear constraint,
        and ftol is the 'Nonlinear Feasibility Tolerance'.

    'Start Objective Check At Variable' : int
        Default = 1

        These keywords take effect only if 'Verify Level' > 0.
        They may be used to control the verification of gradient
        elements computed by `objfun` and/or Jacobian elements computed
        by `confun`.
        For example, if the first 30 elements of the objective gradient
        appeared to be correct in an earlier run, so that only element
        31 remains questionable, it is reasonable to specify 'Start
        Objective Check At Variable' = 31.
        If the first 30 variables appear linearly in the objective, so
        that the corresponding gradient elements are constant, the above
        choice would also be appropriate.

        If i_{2*m-1} <= 0 or i_{2*m-1} > min(n, i_{2m}), the default
        value is used, for m = 1,2,...,2.
        If i_{2*m} <= 0 or i_{2*m} > n, the default value is used, for m
        = 1,2,...,2.

    'Stop Objective Check At Variable' : int
        Default = n

        These keywords take effect only if 'Verify Level' > 0.
        They may be used to control the verification of gradient
        elements computed by `objfun` and/or Jacobian elements computed
        by `confun`.
        For example, if the first 30 elements of the objective gradient
        appeared to be correct in an earlier run, so that only element
        31 remains questionable, it is reasonable to specify 'Start
        Objective Check At Variable' = 31.
        If the first 30 variables appear linearly in the objective, so
        that the corresponding gradient elements are constant, the above
        choice would also be appropriate.

        If i_{2*m-1} <= 0 or i_{2*m-1} > min(n, i_{2m}), the default
        value is used, for m = 1,2,...,2.
        If i_{2*m} <= 0 or i_{2*m} > n, the default value is used, for m
        = 1,2,...,2.

    'Start Constraint Check At Variable' : int
        Default = 1

        These keywords take effect only if 'Verify Level' > 0.
        They may be used to control the verification of gradient
        elements computed by `objfun` and/or Jacobian elements computed
        by `confun`.
        For example, if the first 30 elements of the objective gradient
        appeared to be correct in an earlier run, so that only element
        31 remains questionable, it is reasonable to specify 'Start
        Objective Check At Variable' = 31.
        If the first 30 variables appear linearly in the objective, so
        that the corresponding gradient elements are constant, the above
        choice would also be appropriate.

        If i_{2*m-1} <= 0 or i_{2*m-1} > min(n, i_{2m}), the default
        value is used, for m = 1,2,...,2.
        If i_{2*m} <= 0 or i_{2*m} > n, the default value is used, for m
        = 1,2,...,2.

    'Stop Constraint Check At Variable' : int
        Default = n

        These keywords take effect only if 'Verify Level' > 0.
        They may be used to control the verification of gradient
        elements computed by `objfun` and/or Jacobian elements computed
        by `confun`.
        For example, if the first 30 elements of the objective gradient
        appeared to be correct in an earlier run, so that only element
        31 remains questionable, it is reasonable to specify 'Start
        Objective Check At Variable' = 31.
        If the first 30 variables appear linearly in the objective, so
        that the corresponding gradient elements are constant, the above
        choice would also be appropriate.

        If i_{2*m-1} <= 0 or i_{2*m-1} > min(n, i_{2m}), the default
        value is used, for m = 1,2,...,2.
        If i_{2*m} <= 0 or i_{2*m} > n, the default value is used, for m
        = 1,2,...,2.

    'Step Limit' : float
        Default = 2.0

        If r > 0,r specifies the maximum change in variables at the
        first step of the linesearch.
        In some cases, such as F(x) = ae^(bx) or F(x) = ax^b, even a
        moderate change in the elements of x can lead to floating-point
        overflow.
        The argument r is therefore used to encourage evaluation of the
        problem functions at meaningful points.
        Given any major iterate x, the first point x~ at which F and c
        are evaluated during the linesearch is restricted so that

            ||x~-x||_2 <= r(1+||x||_2).

        The linesearch may go on and evaluate F and c at points further
        from x if this will result in a lower value of the merit
        function (indicated by L at the end of each line of output
        produced by the major iterations; see [Further Comments]).
        If L is printed for most of the iterations, r should be set to a
        larger value.

        Wherever possible, upper and lower bounds on x should be used to
        prevent evaluation of nonlinear functions at wild values.
        The default value 'Step Limit' = 2.0 should not affect progress
        on well-behaved functions, but values such as 0.1 or 0.01 may be
        helpful when rapidly varying functions are present.
        If a small value of 'Step Limit' is selected, a good starting
        point may be required.
        An important application is to the class of nonlinear least
        squares problems.
        If r <= 0, the default value is used.

    'Verify Level' : int
        Default = 0

        These keywords refer to finite difference checks on the gradient
        elements computed by `objfun` and `confun`.
        The possible choices for i are as follows:

        [table omitted]

        It is possible to specify 'Verify Level' = 0 to 3 in several
        ways.
        For example, the nonlinear objective gradient (if any) will be
        verified if either 'Verify Objective Gradients' or 'Verify
        Level' = 1 is specified.
        The constraint gradients will be verified if 'Verify' = 'YES' or
        'Verify Level' = 2 or 'Verify' is specified.
        Similarly, the objective and the constraint gradients will be
        verified if 'Verify' = 'YES' or 'Verify Level' = 3 or 'Verify'
        is specified.

        If 0 <= i <= 3, gradients will be verified at the first point
        that satisfies the linear constraints and bounds.

        If i = 0, only a 'cheap' test will be performed, requiring one
        call to `objfun` and (if appropriate) one call to `confun`.

        If 1 <= i <= 3, a more reliable (but more expensive) check will
        be made on individual gradient elements, within the ranges
        specified by the 'Start Constraint Check At Variable' and 'Stop
        Constraint Check At Variable' keywords.
        A result of the form OK or BAD? is printed by ``nlp1_solve`` to
        indicate whether or not each element appears to be correct.
        If a gradient element is determined to be extremely poor (i.e.,
        if it appears to have no significant digits of accuracy at all),
        then ``nlp1_solve`` will also exit with an error indicator in
        argument errno.

        If 10 <= i <= 13, the action is the same as for i-10, except
        that it will take place at the user-specified initial value of
        x.

        If i < -1 or 4 <= i <= 9 or i > 13, the default value is used.

        We suggest that 'Verify Level' = 3 be used whenever a new
        function function is being developed.

    'Verify' : valueless
        These keywords refer to finite difference checks on the gradient
        elements computed by `objfun` and `confun`.
        The possible choices for i are as follows:

        [table omitted]

        It is possible to specify 'Verify Level' = 0 to 3 in several
        ways.
        For example, the nonlinear objective gradient (if any) will be
        verified if either 'Verify Objective Gradients' or 'Verify
        Level' = 1 is specified.
        The constraint gradients will be verified if 'Verify' = 'YES' or
        'Verify Level' = 2 or 'Verify' is specified.
        Similarly, the objective and the constraint gradients will be
        verified if 'Verify' = 'YES' or 'Verify Level' = 3 or 'Verify'
        is specified.

        If 0 <= i <= 3, gradients will be verified at the first point
        that satisfies the linear constraints and bounds.

        If i = 0, only a 'cheap' test will be performed, requiring one
        call to `objfun` and (if appropriate) one call to `confun`.

        If 1 <= i <= 3, a more reliable (but more expensive) check will
        be made on individual gradient elements, within the ranges
        specified by the 'Start Constraint Check At Variable' and 'Stop
        Constraint Check At Variable' keywords.
        A result of the form OK or BAD? is printed by ``nlp1_solve`` to
        indicate whether or not each element appears to be correct.
        If a gradient element is determined to be extremely poor (i.e.,
        if it appears to have no significant digits of accuracy at all),
        then ``nlp1_solve`` will also exit with an error indicator in
        argument errno.

        If 10 <= i <= 13, the action is the same as for i-10, except
        that it will take place at the user-specified initial value of
        x.

        If i < -1 or 4 <= i <= 9 or i > 13, the default value is used.

        We suggest that 'Verify Level' = 3 be used whenever a new
        function function is being developed.

    'Verify Constraint Gradients' : valueless
        These keywords refer to finite difference checks on the gradient
        elements computed by `objfun` and `confun`.
        The possible choices for i are as follows:

        [table omitted]

        It is possible to specify 'Verify Level' = 0 to 3 in several
        ways.
        For example, the nonlinear objective gradient (if any) will be
        verified if either 'Verify Objective Gradients' or 'Verify
        Level' = 1 is specified.
        The constraint gradients will be verified if 'Verify' = 'YES' or
        'Verify Level' = 2 or 'Verify' is specified.
        Similarly, the objective and the constraint gradients will be
        verified if 'Verify' = 'YES' or 'Verify Level' = 3 or 'Verify'
        is specified.

        If 0 <= i <= 3, gradients will be verified at the first point
        that satisfies the linear constraints and bounds.

        If i = 0, only a 'cheap' test will be performed, requiring one
        call to `objfun` and (if appropriate) one call to `confun`.

        If 1 <= i <= 3, a more reliable (but more expensive) check will
        be made on individual gradient elements, within the ranges
        specified by the 'Start Constraint Check At Variable' and 'Stop
        Constraint Check At Variable' keywords.
        A result of the form OK or BAD? is printed by ``nlp1_solve`` to
        indicate whether or not each element appears to be correct.
        If a gradient element is determined to be extremely poor (i.e.,
        if it appears to have no significant digits of accuracy at all),
        then ``nlp1_solve`` will also exit with an error indicator in
        argument errno.

        If 10 <= i <= 13, the action is the same as for i-10, except
        that it will take place at the user-specified initial value of
        x.

        If i < -1 or 4 <= i <= 9 or i > 13, the default value is used.

        We suggest that 'Verify Level' = 3 be used whenever a new
        function function is being developed.

    'Verify Gradients' : valueless
        These keywords refer to finite difference checks on the gradient
        elements computed by `objfun` and `confun`.
        The possible choices for i are as follows:

        [table omitted]

        It is possible to specify 'Verify Level' = 0 to 3 in several
        ways.
        For example, the nonlinear objective gradient (if any) will be
        verified if either 'Verify Objective Gradients' or 'Verify
        Level' = 1 is specified.
        The constraint gradients will be verified if 'Verify' = 'YES' or
        'Verify Level' = 2 or 'Verify' is specified.
        Similarly, the objective and the constraint gradients will be
        verified if 'Verify' = 'YES' or 'Verify Level' = 3 or 'Verify'
        is specified.

        If 0 <= i <= 3, gradients will be verified at the first point
        that satisfies the linear constraints and bounds.

        If i = 0, only a 'cheap' test will be performed, requiring one
        call to `objfun` and (if appropriate) one call to `confun`.

        If 1 <= i <= 3, a more reliable (but more expensive) check will
        be made on individual gradient elements, within the ranges
        specified by the 'Start Constraint Check At Variable' and 'Stop
        Constraint Check At Variable' keywords.
        A result of the form OK or BAD? is printed by ``nlp1_solve`` to
        indicate whether or not each element appears to be correct.
        If a gradient element is determined to be extremely poor (i.e.,
        if it appears to have no significant digits of accuracy at all),
        then ``nlp1_solve`` will also exit with an error indicator in
        argument errno.

        If 10 <= i <= 13, the action is the same as for i-10, except
        that it will take place at the user-specified initial value of
        x.

        If i < -1 or 4 <= i <= 9 or i > 13, the default value is used.

        We suggest that 'Verify Level' = 3 be used whenever a new
        function function is being developed.

    'Verify Objective Gradients' : valueless
        These keywords refer to finite difference checks on the gradient
        elements computed by `objfun` and `confun`.
        The possible choices for i are as follows:

        [table omitted]

        It is possible to specify 'Verify Level' = 0 to 3 in several
        ways.
        For example, the nonlinear objective gradient (if any) will be
        verified if either 'Verify Objective Gradients' or 'Verify
        Level' = 1 is specified.
        The constraint gradients will be verified if 'Verify' = 'YES' or
        'Verify Level' = 2 or 'Verify' is specified.
        Similarly, the objective and the constraint gradients will be
        verified if 'Verify' = 'YES' or 'Verify Level' = 3 or 'Verify'
        is specified.

        If 0 <= i <= 3, gradients will be verified at the first point
        that satisfies the linear constraints and bounds.

        If i = 0, only a 'cheap' test will be performed, requiring one
        call to `objfun` and (if appropriate) one call to `confun`.

        If 1 <= i <= 3, a more reliable (but more expensive) check will
        be made on individual gradient elements, within the ranges
        specified by the 'Start Constraint Check At Variable' and 'Stop
        Constraint Check At Variable' keywords.
        A result of the form OK or BAD? is printed by ``nlp1_solve`` to
        indicate whether or not each element appears to be correct.
        If a gradient element is determined to be extremely poor (i.e.,
        if it appears to have no significant digits of accuracy at all),
        then ``nlp1_solve`` will also exit with an error indicator in
        argument errno.

        If 10 <= i <= 13, the action is the same as for i-10, except
        that it will take place at the user-specified initial value of
        x.

        If i < -1 or 4 <= i <= 9 or i > 13, the default value is used.

        We suggest that 'Verify Level' = 3 be used whenever a new
        function function is being developed.

    Raises
    ------
    NagValueError
        (`errno` 7)
            Large errors found in the derivatives.

        (`errno` 9)
            On entry, n = *<value>*.

            Constraint: n > 0.

        (`errno` 9)
            On entry, nclin = *<value>*.

            Constraint: nclin >= 0.

        (`errno` 9)
            On entry, ncnln = *<value>*.

            Constraint: ncnln >= 0.

        (`errno` 9)
            Not enough workspace to solve problem. Workspace provided is
            iwork[*<value>*] and work[*<value>*]. To solve problem we
            need iwork[*<value>*] and work[*<value>*].

        (`errno` 9)
            On entry, the equal bounds on *<value>* are infinite,
            because `bl`[*<value>*] = beta and `bu`[*<value>*] = beta,
            but abs(beta) >= bigbnd: beta = *<value>* and bigbnd =
            *<value>*.

        (`errno` 9)
            On entry, the bounds on *<value>* are inconsistent:
            `bl`[*<value>*] = *<value>* and `bu`[*<value>*] = *<value>*.

        (`errno` 9)
            On entry with a Warm Start, `istate`[*<value>*] = *<value>*.

        (`errno` 9)
            On entry, the equal bounds on variable *<value>* are
            infinite, because `bl`[*<value>*] = beta and `bu`[*<value>*]
            = beta, but abs(beta) >= bigbnd: beta = *<value>* and bigbnd
            = *<value>*.

        (`errno` 9)
            On entry, the equal bounds on linear constraint *<value>*
            are infinite, because `bl`[*<value>*] = beta and
            `bu`[*<value>*] = beta, but abs(beta) >= bigbnd: beta =
            *<value>* and bigbnd = *<value>*.

        (`errno` 9)
            On entry, the equal bounds on nonlinear constraint *<value>*
            are infinite, because `bl`[*<value>*] = beta and
            `bu`[*<value>*] = beta, but abs(beta) >= bigbnd: beta =
            *<value>* and bigbnd = *<value>*.

        (`errno` 9)
            On entry, the bounds on variable *<value>* are inconsistent:
            `bl`[*<value>*] = *<value>* and `bu`[*<value>*] = *<value>*.

        (`errno` 9)
            On entry, the bounds on linear constraint *<value>* are
            inconsistent: `bl`[*<value>*] = *<value>* and
            `bu`[*<value>*] = *<value>*.

        (`errno` 9)
            On entry, the bounds on nonlinear constraint *<value>* are
            inconsistent: `bl`[*<value>*] = *<value>* and
            `bu`[*<value>*] = *<value>*.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            Optimal solution found, but requested accuracy not achieved.

        (`errno` 2)
            No feasible point for the linear constraints.

        (`errno` 3)
            No feasible point for the nonlinear constraints.

        (`errno` 6)
            Current point cannot be improved upon.

    NagAlgorithmicMajorWarning
        (`errno` 4)
            Too many major iterations.

    NagCallbackTerminateWarning
        (`errno` i < 0)
            User requested termination by setting `mode` negative in
            `objfun` or `confun`.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Dennis, J E Jr and Moré, J J, 1977, `Quasi-Newton methods,
    motivation and theory`, SIAM Rev. (19), 46--89

    Dennis, J E Jr and Schnabel, R B, 1981, `A new derivation of
    symmetric positive-definite secant updates`, nonlinear programming,
    (eds O L Mangasarian, R R Meyer and S M Robinson) (4), 167--199,
    Academic Press

    Dennis, J E Jr and Schnabel, R B, 1983, `Numerical Methods for
    Unconstrained Optimization and Nonlinear Equations`, Prentice--Hall

    Fletcher, R, 1987, `Practical Methods of Optimization`, (2nd
    Edition), Wiley

    Gill, P E, Hammarling, S, Murray, W, Saunders, M A and Wright, M H,
    1986, `Users' guide for LSSOL (Version 1.0)`, Report SOL 86-1,
    Department of Operations Research, Stanford University

    Gill, P E, Murray, W, Saunders, M A and Wright, M H, 1984, `Users'
    guide for SOL/QPSOL version 3.2`, Report SOL, 84--5, Department of
    Operations Research, Stanford University

    Gill, P E, Murray, W, Saunders, M A and Wright, M H, 1984,
    `Procedures for optimization problems with a mixture of bounds and
    general linear constraints`, ACM Trans. Math. Software (10),
    282--298

    Gill, P E, Murray, W, Saunders, M A and Wright, M H, 1986, `Some
    theoretical properties of an augmented Lagrangian merit function`,
    Report SOL, 86--6R, Department of Operations Research, Stanford
    University

    Gill, P E, Murray, W, Saunders, M A and Wright, M H, 1986, `Users'
    guide for NPSOL (Version 4.0): a Fortran package for nonlinear
    programming`, Report SOL 86-2, Department of Operations Research,
    Stanford University

    Gill, P E, Murray, W and Wright, M H, 1981, `Practical
    Optimization`, Academic Press

    Hock, W and Schittkowski, K, 1981, `Test Examples for Nonlinear
    Programming Codes. Lecture Notes in Economics and Mathematical
    Systems` (187), Springer--Verlag

    Powell, M J D, 1974, `Introduction to constrained optimization`,
    Numerical Methods for Constrained Optimization, (eds P E Gill and W
    Murray), 1--28, Academic Press

    Powell, M J D, 1983, `Variable metric methods in constrained
    optimization`, Mathematical Programming: the State of the Art, (eds
    A Bachem, M Grötschel and B Korte), 288--311, Springer--Verlag

    See Also
    --------
    :meth:`naginterfaces.library.examples.opt.nlp1_solve_ex.main`
    """

    fun_name = 'naginterfaces.library.opt.nlp1_solve'

    cb_confun_name = fun_name + ':confun'

    def cb_confun(mode, ncnln, n, needc, x, c, cjac, nstate, data=None):
        """Wrapping user's confun. Called as base confun."""
        if data is not None:
            _c_confun, _cjac_confun = confun(
                mode,
                needc,
                x,
                cjac,
                nstate,
                data,
            )
        else:
            _c_confun, _cjac_confun = confun(
                mode,
                needc,
                x,
                cjac,
                nstate,
            )
        c[:] = _utils._to_ndarray(
            {'fun_name': cb_confun_name, 'returned': 'c'},
            _c_confun,
            copy=False,
            exp_shape=(ncnln,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _cjac_confun_np = _utils._to_ndarray(
            {'fun_name': cb_confun_name, 'returned': 'cjac'},
            _cjac_confun,
            copy=False,
            exp_shape=(max(1, ncnln), n),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _utils._check_callback_sorder(
            {'fun_name': cb_confun_name},
            ((_cjac_confun_np, 'cjac'),),
            wrapper_sorder=_m_sorder,
            spiked_sorder=spiked_sorder,
        )
        cjac[:, :] = _cjac_confun_np

    cb_objfun_name = fun_name + ':objfun'

    def cb_objfun(mode, n, x, objgrd, nstate, data=None):
        """Wrapping user's objfun. Called as base objfun."""
        if data is not None:
            _objf_objfun, _objgrd_objfun = objfun(
                mode,
                x,
                objgrd,
                nstate,
                data,
            )
        else:
            _objf_objfun, _objgrd_objfun = objfun(
                mode,
                x,
                objgrd,
                nstate,
            )
        objgrd[:] = _utils._to_ndarray(
            {'fun_name': cb_objfun_name, 'returned': 'objgrd'},
            _objgrd_objfun,
            copy=False,
            exp_shape=(n,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _objf_objfun_base = _objf_objfun
        return _objf_objfun_base

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _cjac = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'cjac'},
        cjac,
        copy=True,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _r = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'r'},
        r,
        copy=True,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_a, 'a'), (_cjac, 'cjac'), (_r, 'r')),
        spiked_sorder=spiked_sorder,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _nclin_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _nclin = _b_utils._EngineIntScalarType(
        _nclin_py, {'fun_name': fun_name, 'entity': '_nclin'},
    ).to_ctypes()
    _bl = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'bl'},
        bl,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _ncnln = _b_utils._EngineIntScalarType(
        _b_utils._EngineFloat64ArrayType.get_shape(_bl, {'fun_name': fun_name, 'entity': 'bl'}, exp_rank=1)[0]-_n_py-_nclin_py, {'fun_name': fun_name, 'entity': '_ncnln'},
    ).to_ctypes()
    _bu = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'bu'},
        bu,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b_utils._check_callable(confun, {'fun_name': fun_name, 'entity': 'confun'}, allow_none=True)
    _confun = None if confun is None else cb_confun
    _b_utils._check_callable(objfun, {'fun_name': fun_name, 'entity': 'objfun'})
    _objfun = cb_objfun
    _b_utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        req_keys=('lwsav', 'iwsav', 'rwsav'),
    )
    if _b_utils._EngineIntArrayType.get_element(comm['iwsav'], {'fun_name': fun_name, 'entity': 'comm[\'iwsav\']'}, (400,), validate=False) == 1:
        if istate is None:
            raise _b_utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'istate'},
                'must not be None when the option \'Warm Start\' has been chosen.',
            )
    _ncnln_py = _b_utils._EngineIntScalarType.to_py(_ncnln)
    if istate is None:
        _istate = _b_utils._EngineIntArrayType.empty_ndarray(
            (_n_py+_nclin_py+_ncnln_py,),
        )
    else:
        _istate = _utils._to_ndarray(
            {'fun_name': fun_name, 'entity': 'istate'},
            istate,
            copy=True,
        )
    if _ncnln_py != 0:
        _c = _b_utils._EngineFloat64ArrayType.empty_ndarray(
            (max(1, _ncnln_py),),
        )
    else:
        _c = None
    if (_b_utils._EngineIntArrayType.get_element(comm['iwsav'], {'fun_name': fun_name, 'entity': 'comm[\'iwsav\']'}, (151,), validate=False) == 2 or _b_utils._EngineIntArrayType.get_element(comm['iwsav'], {'fun_name': fun_name, 'entity': 'comm[\'iwsav\']'}, (151,), validate=False) == 3) and _ncnln_py != 0:
        if cjac is None:
            raise _b_utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'cjac'},
                'must not be None when \'Derivative Level\' = 2 or 3 and ncnln != 0.',
            )
    if cjac is None:
        if _ncnln_py > 0:
            _cjac_dim2 = _n_py
        else:
            _cjac_dim2 = 1
        exp_shape_cjac = (_ncnln_py, _cjac_dim2)
        _cjac = _b_utils._EngineFloat64ArrayType.empty_ndarray(
            exp_shape_cjac,
            sorder=_m_sorder,
        )
    if _b_utils._EngineIntArrayType.get_element(comm['iwsav'], {'fun_name': fun_name, 'entity': 'comm[\'iwsav\']'}, (400,), validate=False) == 1:
        if clamda is None:
            raise _b_utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'clamda'},
                'must not be None when the option \'Warm Start\' has been chosen.',
            )
    if clamda is None:
        _clamda = _b_utils._EngineFloat64ArrayType.empty_ndarray(
            (_n_py+_nclin_py+_ncnln_py,),
        )
    else:
        _clamda = _utils._to_ndarray(
            {'fun_name': fun_name, 'entity': 'clamda'},
            clamda,
            copy=True,
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
    _objgrd = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    if _b_utils._EngineIntArrayType.get_element(comm['iwsav'], {'fun_name': fun_name, 'entity': 'comm[\'iwsav\']'}, (400,), validate=False) == 1:
        if r is None:
            raise _b_utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'r'},
                'must not be None when the option \'Warm Start\' has been chosen.',
            )
    if r is None:
        _r = _b_utils._EngineFloat64ArrayType.empty_ndarray(
            (_n_py, _n_py),
            sorder=_m_sorder,
        )
    try:
        _return_scalars = _b_opt.nlp1_solve(
            _n,
            _nclin,
            _ncnln,
            _a,
            _bl,
            _bu,
            _confun,
            _objfun,
            _istate,
            _c,
            _cjac,
            _clamda,
            _objgrd,
            _r,
            _x,
            comm,
            data,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _x_py = _x
    _istate_py = _istate
    if _ncnln_py != 0:
        _cjac_py = _cjac
    else:
        _cjac_py = None
    _clamda_py = _clamda
    _r_py = _r
    _itera_py = _return_scalars.itera
    if _ncnln_py != 0:
        _c_py = _c
    else:
        _c_py = None
    _objf_py = _return_scalars.objf
    _objgrd_py = _objgrd
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['itera', 'istate', 'c', 'cjac', 'clamda', 'objf', 'objgrd', 'r', 'x'],
    )(
        _itera_py,
        _istate_py,
        _c_py,
        _cjac_py,
        _clamda_py,
        _objf_py,
        _objgrd_py,
        _r_py,
        _x_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def nlp1_option_file(ioptns, comm, io_manager=None):
    u"""
    Supply option values for ``nlp1_solve`` or ``nlp1_rcomm`` from
    external file.

    ``nlp1_option_file`` may be used to supply options to ``nlp1_solve``
    from an external file.

    ``nlp1_option_file`` can also be used to supply options to
    ``nlp1_rcomm``.

    For full information please refer to the NAG Library document for
    e04ud

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04udf.html

    Parameters
    ----------
    ioptns : int
        The unit number (see
        :meth:`~naginterfaces.base.utils.FileObjManager.unit_from_fileobj`)
        of the options file to be read.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``nlp1_init``.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `ioptns` = *<value>*.

            Constraint: 0 <= `ioptns` <= 99.

        (`errno` 2)
            ``Begin`` was found, but end-of-file was found before
            ``End`` was found.

        (`errno` 3)
            End-of-file was found before ``Begin`` was found.

        (`errno` 5)
            One or more lines of the options file is invalid.
    """

    _b_opt.nlp1_option_file(
        ioptns,
        comm,
        io_manager,
    )

def nlp1_option_string(optstr, comm, io_manager=None):
    u"""
    Supply option values to ``nlp1_solve`` or ``nlp1_rcomm`` from a
    character string.

    ``nlp1_option_string`` may be used to supply individual options to
    ``nlp1_solve``

    ``nlp1_option_string`` can also be used to supply individual options
    to ``nlp1_rcomm``.

    For full information please refer to the NAG Library document for
    e04ue

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04uef.html

    Parameters
    ----------
    optstr : str
        A single valid option string (as described in [Description] and
        in [Other Parameters for nlp1_solve]).

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``nlp1_init``.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Raises
    ------
    NagValueError
        (`errno` 5)
            The supplied option string is invalid. Supplied value was:
            *<value>*.

    See Also
    --------
    :meth:`naginterfaces.library.examples.opt.nlp1_solve_ex.main`
    """

    _b_opt.nlp1_option_string(
        optstr,
        comm,
        io_manager,
    )

def nlp1_rcomm(irevcm, nclin, a, bl, bu, itera, istate, c, cjac, clamda, objf, objgrd, r, x, comm, io_manager=None):
    u"""
    Nonlinear programming (NLP), dense, active-set, SQP method, using
    function values and optionally first derivatives (reverse
    communication, comprehensive).

    Note: this function uses optional algorithmic parameters, see also:
    ``nlp1_option_file``, ``nlp1_option_string``, ``nlp1_init``.

    ``nlp1_rcomm`` is designed to minimize an arbitrary smooth function
    subject to constraints (which may include simple bounds on the
    variables, linear constraints and smooth nonlinear constraints)
    using a sequential quadratic programming (SQP) method.
    You should supply as many first derivatives as possible; any
    unspecified derivatives are approximated by finite differences.
    It is not intended for large sparse problems.

    ``nlp1_rcomm`` may also be used for unconstrained, bound-constrained
    and linearly constrained optimization.

    ``nlp1_rcomm`` uses **reverse communication** for evaluating the
    objective function, the nonlinear constraint functions and any of
    their derivatives.

    For full information please refer to the NAG Library document for
    e04uf

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04uff.html

    Parameters
    ----------
    irevcm : int
        `On initial entry`: must be set to 0.

        `On intermediate entry`: **must remain unchanged**, unless you
        wish to terminate the solution to the current problem. In this
        case `irevcm` may be set to a negative value and then
        ``nlp1_rcomm`` will take a final exit with errno set to this
        value of `irevcm`.

    nclin : int
        `On initial entry`: n_L, the number of general linear
        constraints.

    a : float, array-like, shape (`nclin`, :)
        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `nclin`>0: n; otherwise: 1.

        `On initial entry`: the ith row of the matrix A_L of general
        linear constraints in (1) must be stored in `a`[i-1,j-1], for j
        = 1,2,...,n, for i = 1,2,...,`nclin`. That is, the ith row
        contains the coefficients of the ith general linear constraint,
        for i = 1,2,...,`nclin`.

        If `nclin` = 0, the array `a` is not referenced.

    bl : float, array-like, shape (n+`nclin`+ncnln)
        On initial entry `bl` must contain the lower bounds for all the
        constraints

    bu : float, array-like, shape (n+`nclin`+ncnln)
        On initial entry `bu` must contain the upper bounds for all the
        constraints

    itera : int
        `On intermediate entry`: must remain unchanged from a previous
        call to ``nlp1_rcomm``.

    istate : int, ndarray, shape (n+`nclin`+ncnln), modified in place
        `On initial entry`: need not be set if the (default) option
        'Cold Start' is used.

        If the option 'Warm Start' has been chosen, the elements of
        `istate` corresponding to the bounds and linear constraints
        define the initial working set for the procedure that finds a
        feasible point for the linear constraints and bounds.

        The active set at the conclusion of this procedure and the
        elements of `istate` corresponding to nonlinear constraints then
        define the initial working set for the first QP subproblem.

        More precisely, the first n elements of `istate` refer to the
        upper and lower bounds on the variables, the next n_L elements
        refer to the upper and lower bounds on A_L x, and the next n_N
        elements refer to the upper and lower bounds on c(x).

        Possible values for `istate`[j-1] are as follows:

        [table omitted]

        The values -2, -1 and 4 are also acceptable but will be modified
        by the function.

        If ``nlp1_rcomm`` has been called previously with the same
        values of n, `nclin` and ncnln, `istate` already contains
        satisfactory information. (See also the description of the
        option 'Warm Start'.) The function also adjusts (if necessary)
        the values supplied in `x` to be consistent with `istate`.

        `On final exit`: the status of the constraints in the QP working
        set at the point returned in `x`. The significance of each
        possible value of `istate`[j-1] is as follows:

        [table omitted]

    c : float, ndarray, shape (ncnln), modified in place
        `On initial entry`: need not be set.

        `On intermediate entry`: if `irevcm` = 4 or 6 and `needc`[i-1] >
        0, `c`[i-1] must contain the value of the ith constraint at x.
        The remaining elements of `c`, corresponding to the non-positive
        elements of `needc`, are ignored.

        `On final exit`: if ncnln > 0, `c`[i-1] contains the value of
        the ith nonlinear constraint function c_i at the final iterate,
        for i = 1,2,...,ncnln.

        If ncnln = 0, the array `c` is not referenced.

    cjac : float, ndarray, shape (ncnln, :), modified in place
        Note: the required extent for this argument in dimension 2 is
        determined as follows: if ncnln>0: n; otherwise: 1.

        `On initial entry`: in general, `cjac` need not be initialized
        before the call to ``nlp1_rcomm``. However, if the option
        'Derivative Level' = 2 or 3, you may optionally set the constant
        elements of `cjac`. Such constant elements need not be
        re-assigned on subsequent intermediate exits.

        If all elements of the constraint Jacobian are known (i.e.,
        'Derivative Level' = 2 or 3), any constant elements may be
        assigned to `cjac` one time only at the start of the
        optimization.

        An element of `cjac` that is not subsequently assigned during an
        intermediate exit will retain its initial value throughout.

        Constant elements may be loaded into `cjac` either before the
        call to ``nlp1_rcomm`` or during the first intermediate exit.

        The ability to preload constants is useful when many Jacobian
        elements are identically zero, in which case `cjac` may be
        initialized to zero and nonzero elements may be reset during
        intermediate exits.

        `On intermediate entry`: if `irevcm` = 5 or 6 and `needc`[i-1] >
        0, the ith row of `cjac` must contain the available elements of
        the vector nabla c_i given by

            nabla c_i = (dc_i/dx_1,dc_i/dx_2,...,dc_i/dx_n)^T,

        where dc_i/dx_j is the partial derivative of the ith constraint
        with respect to the jth variable, evaluated at the point x. The
        remaining rows of `cjac`, corresponding to non-positive elements
        of `needc`, are ignored. The ith row of the Jacobian should be
        stored in elements `cjac`[i-1,j-1], for j = 1,2,...,n, for i =
        1,2,...,ncnln.

        Note that constant nonzero elements do affect the values of the
        constraints.

        Thus, if `cjac`[i-1,j-1] is set to a constant value, it need not
        be reset during subsequent intermediate exits, but the value
        `cjac`[i-1,j-1]*`x`[j-1] must nonetheless be added to `c`[i-1].

        For example, if `cjac`[0,0] = 2 and `cjac`[0,1] = -5, the term
        2*`x`[0]-5*`x`[1] must be included in the definition of `c`[0].

        It must be emphasized that, if 'Derivative Level' = 0 or 1,
        unassigned elements of `cjac` are not treated as constant; they
        are estimated by finite differences, at nontrivial expense.

        If you do not supply a value for the option 'Difference
        Interval', an interval for each element of x is computed
        automatically at the start of the optimization.

        The automatic procedure can usually identify constant elements
        of `cjac`, which are then computed once only by finite
        differences.

        See also the description of the option 'Verify'.

        `On final exit`: if ncnln > 0, `cjac` contains the Jacobian
        matrix of the nonlinear constraint functions at the final
        iterate, i.e., `cjac`[i-1,j-1] contains the partial derivative
        of the ith constraint function with respect to the jth variable,
        for j = 1,2,...,n, for i = 1,2,...,ncnln.

        If ncnln = 0, the array `cjac` is not referenced.

    clamda : float, ndarray, shape (n+`nclin`+ncnln), modified in place
        `On initial entry`: need not be set if the (default) option
        'Cold Start' is used.

        If the option 'Warm Start' has been chosen, `clamda`[j-1] must
        contain a multiplier estimate for each nonlinear constraint with
        a sign that matches the status of the constraint specified by
        the `istate` array, for j = n+`nclin`+1,...,n+`nclin`+ncnln.

        The remaining elements need not be set.

        Note that if the jth constraint is defined as 'inactive' by the
        initial value of the `istate` array (i.e. `istate`[j-1] = 0),
        `clamda`[j-1] should be zero; if the jth constraint is an
        inequality active at its lower bound (i.e. `istate`[j-1] = 1),
        `clamda`[j-1] should be non-negative; if the jth constraint is
        an inequality active at its upper bound (i.e. `istate`[j-1] =
        2), `clamda`[j-1] should be non-positive.

        If necessary, the function will modify `clamda` to match these
        rules.

        `On final exit`: the values of the QP multipliers from the last
        QP subproblem. `clamda`[j-1] should be non-negative if
        `istate`[j-1] = 1 and non-positive if `istate`[j-1] = 2.

    objf : float
        `On initial entry`: need not be set.

        `On intermediate entry`: if `irevcm` = 1 or 3, `objf` must be
        set to the value of the objective function at x.

    objgrd : float, ndarray, shape (n), modified in place
        `On initial entry`: need not be set.

        `On intermediate entry`: if `irevcm` = 2 or 3, `objgrd` must
        contain the available elements of the gradient evaluated at x.

        See also the description of the option 'Verify'.

        `On final exit`: the gradient of the objective function at the
        final iterate (or its finite difference approximation).

    r : float, ndarray, shape (n, n), modified in place
        `On initial entry`: need not be initialized if the (default)
        option 'Cold Start' is used.

        If the option 'Warm Start' has been chosen, `r` must contain the
        upper triangular Cholesky factor R of the initial approximation
        of the Hessian of the Lagrangian function, with the variables in
        the natural order.

        Elements not in the upper triangular part of `r` are assumed to
        be zero and need not be assigned.

        `On final exit`: if 'Hessian' = 'NO', `r` contains the upper
        triangular Cholesky factor R of Q^TH~Q, an estimate of the
        transformed and reordered Hessian of the Lagrangian at x (see
        (6) in [Overview]).

        If 'Hessian' = 'YES', `r` contains the upper triangular Cholesky
        factor R of H, the approximate (untransformed) Hessian of the
        Lagrangian, with the variables in the natural order.

    x : float, ndarray, shape (n), modified in place
        `On initial entry`: an initial estimate of the solution.

        `On intermediate exit`: the point x at which the objective
        function, constraint functions or their derivatives are to be
        evaluated.

        `On final exit`: the final estimate of the solution.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``nlp1_init``.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    irevcm : int
        `On intermediate exit`: specifies what values the calling
        program must assign to arguments of ``nlp1_rcomm`` before
        re-entering the function.

        `irevcm` = 1

            Set `objf` to the value of the objective function F(x).

        `irevcm` = 2

            Set `objgrd`[j-1] to the value dF/dx_j if available, for j =
            1,2,...,n.

        `irevcm` = 3

            Set `objf` and `objgrd`[j-1] as for `irevcm` = 1 and
            `irevcm` = 2.

        `irevcm` = 4

            Set `c`[i-1] to the value of the constraint function c_i(x),
            for each i such that `needc`[i-1] > 0.

        `irevcm` = 5

            Set `cjac`[i-1,j-1] to the value dc_i/dx_j if available, for
            each i such that `needc`[i-1] > 0 and j = 1,2,...,n.

        `irevcm` = 6

            Set `c`[i-1] and `cjac`[i-1,j-1] as for `irevcm` = 4 and
            `irevcm` = 5.

        `On final exit`: `irevcm` = 0.

    itera : int
        `On final exit`: the number of major iterations performed.

    objf : float
        `On final exit`: the value of the objective function at the
        final iterate.

    needc : int, ndarray, shape (max(1, ncnln))
        `On intermediate exit`: if `irevcm` >= 4, `needc` specifies the
        indices of the elements of `c` and/or `cjac` that must be
        assigned. If `needc`[i-1] > 0, the ith element of `c` and/or the
        available elements of the ith row of `cjac` must be evaluated at
        x.

    Other Parameters
    ----------------
    'Central Difference Interval' : float
        Default values are computed

        If the algorithm switches to central differences because the
        forward-difference approximation is not sufficiently accurate,
        the value of r is used as the difference interval for every
        element of x.
        The switch to central differences is indicated by C at the end
        of each line of intermediate printout produced by the major
        iterations (see [Further Comments]).
        The use of finite differences is discussed further under the
        option 'Difference Interval'.

        If you supply a value for this option, a small value between 0.0
        and 1.0 is appropriate.

    'Cold Start' : valueless
        Default

        This option controls the specification of the initial working
        set in both the procedure for finding a feasible point for the
        linear constraints and bounds and in the first QP subproblem
        thereafter.
        With a 'Cold Start', the first working set is chosen by
        ``nlp1_rcomm`` based on the values of the variables and
        constraints at the initial point.
        Broadly speaking, the initial working set will include equality
        constraints and bounds or inequality constraints that violate or
        'nearly' satisfy their bounds (to within 'Crash Tolerance').

        With a 'Warm Start', you must set the `istate` array and define
        `clamda` and `r` as discussed in [Parameters]. `istate` values
        associated with bounds and linear constraints determine the
        initial working set of the procedure to find a feasible point
        with respect to the bounds and linear constraints. `istate`
        values associated with nonlinear constraints determine the
        initial working set of the first QP subproblem after such a
        feasible point has been found. ``nlp1_rcomm`` will override your
        specification of `istate` if necessary, so that a poor choice of
        the working set will not cause a fatal error.
        For instance, any elements of `istate` which are set to -2, -1
        or 4 will be reset to zero, as will any elements which are set
        to 3 when the corresponding elements of `bl` and `bu` are not
        equal.
        A warm start will be advantageous if a good estimate of the
        initial working set is available -- for example, when
        ``nlp1_rcomm`` is called repeatedly to solve related problems.

    'Warm Start' : valueless
        This option controls the specification of the initial working
        set in both the procedure for finding a feasible point for the
        linear constraints and bounds and in the first QP subproblem
        thereafter.
        With a 'Cold Start', the first working set is chosen by
        ``nlp1_rcomm`` based on the values of the variables and
        constraints at the initial point.
        Broadly speaking, the initial working set will include equality
        constraints and bounds or inequality constraints that violate or
        'nearly' satisfy their bounds (to within 'Crash Tolerance').

        With a 'Warm Start', you must set the `istate` array and define
        `clamda` and `r` as discussed in [Parameters]. `istate` values
        associated with bounds and linear constraints determine the
        initial working set of the procedure to find a feasible point
        with respect to the bounds and linear constraints. `istate`
        values associated with nonlinear constraints determine the
        initial working set of the first QP subproblem after such a
        feasible point has been found. ``nlp1_rcomm`` will override your
        specification of `istate` if necessary, so that a poor choice of
        the working set will not cause a fatal error.
        For instance, any elements of `istate` which are set to -2, -1
        or 4 will be reset to zero, as will any elements which are set
        to 3 when the corresponding elements of `bl` and `bu` are not
        equal.
        A warm start will be advantageous if a good estimate of the
        initial working set is available -- for example, when
        ``nlp1_rcomm`` is called repeatedly to solve related problems.

    'Crash Tolerance' : float
        Default = 0.01

        This value is used in conjunction with the option 'Cold Start'
        (the default value) when ``nlp1_rcomm`` selects an initial
        working set.
        If 0 <= r <= 1, the initial working set will include (if
        possible) bounds or general inequality constraints that lie
        within r of their bounds.
        In particular, a constraint of the form a_j^Tx >= l will be
        included in the initial working set if |a_j^Tx-l| <= r(1+|l|).
        If r < 0 or r > 1, the default value is used.

    'Defaults' : valueless
        This special keyword may be used to reset all options to their
        default values.

    'Derivative Level' : int
        Default = 3

        This argument indicates which derivatives are provided during
        intermediate exits.
        The possible choices for i are the following.

        [table omitted]

        The value i = 3 should be used whenever possible, since
        ``nlp1_rcomm`` is more reliable (and will usually be more
        efficient) when all derivatives are exact.

        If i = 0 or 2, ``nlp1_rcomm`` will estimate the unspecified
        elements of the objective gradient, using finite differences.
        The computation of finite difference approximations usually
        increases the total run-time, since an intermediate exit to the
        calling program is required for each unspecified element.
        Furthermore, less accuracy can be attained in the solution (see
        Module 8 of Gill `et al.` (1981), for a discussion of limiting
        accuracy).

        If i = 0 or 1, ``nlp1_rcomm`` will approximate unspecified
        elements of the constraint Jacobian.
        One intermediate exit is needed for each variable for which
        partial derivatives are not available.
        For example, if the Jacobian has the form

            [table omitted]

        where '*' indicates an element provided and '?' indicates an
        unspecified element, ``nlp1_rcomm`` will make an intermediate
        exit to the calling program twice: once to estimate the missing
        element in column 2, and again to estimate the two missing
        elements in column 3. (Since columns 1 and 4 are known, they
        require no intermediate exits for information.)

        At times, central differences are used rather than forward
        differences, in which case twice as many intermediate exits are
        needed. (The switch to central differences is not under your
        control.)

        If i < 0 or i > 3, the default value is used.

    'Difference Interval' : float
        Default values are computed

        This option defines an interval used to estimate derivatives by
        finite differences in the following circumstances:

        (a) For verifying the objective and/or constraint gradients (see
            the description of the option 'Verify').

        (#) For estimating unspecified elements of the objective
            gradient or the constraint Jacobian.

        In general, a derivative with respect to the jth variable is
        approximated using the interval delta_j, where delta_j =
        r(1+|x^_j|), with x^ the first point feasible with respect to
        the bounds and linear constraints.
        If the functions are well scaled, then the resulting derivative
        approximation should be accurate to O(r).
        See Gill `et al.` (1981) for a discussion of the accuracy in
        finite difference approximations.

        If a difference interval is not specified, then a finite
        difference interval will be computed automatically for each
        variable by a procedure that requires up to six intermediate
        exits for each element.
        This option is recommended if the function is badly scaled or
        you wish to have ``nlp1_rcomm`` determine constant elements in
        the objective and constraint gradients.

        If you supply a value for this option, a small value between 0.0
        and 1.0 is appropriate.

    'Feasibility Tolerance' : float
        Default = sqrt(epsilon)

        The scalar r defines the maximum acceptable `absolute`
        violations in linear and nonlinear constraints at a 'feasible'
        point; i.e., a constraint is considered satisfied if its
        violation does not exceed r.
        If r < epsilon or r >= 1, the default value is used.
        Using this keyword sets both options 'Linear Feasibility
        Tolerance' and 'Nonlinear Feasibility Tolerance' to r, if
        epsilon <= r < 1. (Additional details are given under the
        descriptions of these options.)

    'Function Precision' : float
        Default = epsilon^0.9

        This argument defines epsilon_r, which is intended to be a
        measure of the accuracy with which the problem functions F(x)
        and c(x) can be computed.
        If r < epsilon or r >= 1, the default value is used.

        The value of epsilon_r should reflect the relative precision of
        1+|F(x)|; i.e., epsilon_r acts as a relative precision when |F|
        is large and as an absolute precision when |F| is small.
        For example, if F(x) is typically of order 1000 and the first
        six significant digits are known to be correct, an appropriate
        value for epsilon_r would be 10^(-6).
        In contrast, if F(x) is typically of order 10^(-4) and the first
        six significant digits are known to be correct, an appropriate
        value for epsilon_r would be 10^(-10).
        The choice of epsilon_r can be quite complicated for badly
        scaled problems; see Module 8 of Gill `et al.` (1981) for a
        discussion of scaling techniques.
        The default value is appropriate for most simple functions that
        are computed with full accuracy.
        However, when the accuracy of the computed function values is
        known to be significantly worse than full precision, the value
        of epsilon_r should be large enough so that ``nlp1_rcomm`` will
        not attempt to distinguish between function values that differ
        by less than the error inherent in the calculation.

    'Hessian' : valueless
        Default = 'NO'

        This option controls the contents of the upper triangular matrix
        R (see [Parameters]). ``nlp1_rcomm`` works exclusively with the
        `transformed and reordered` Hessian H_Q (6), and hence extra
        computation is required to form the Hessian itself.
        If 'Hessian' = 'NO', `r` contains the Cholesky factor of the
        transformed and reordered Hessian.
        If 'Hessian' = 'YES', the Cholesky factor of the approximate
        Hessian itself is formed and stored in `r`.
        You should select 'Hessian' = 'YES' if a 'Warm Start' will be
        used for the next call to ``nlp1_rcomm``.

    'Infinite Bound Size' : float
        Default = 10^20

        If r > 0, r defines the 'infinite' bound bigbnd in the
        definition of the problem constraints.
        Any upper bound greater than or equal to bigbnd will be regarded
        as + infinity (and similarly any lower bound less than or equal
        to -bigbnd will be regarded as - infinity).
        If r < 0, the default value is used.

    'Infinite Step Size' : float
        Default = max(bigbnd, 10^20)

        If r > 0, r specifies the magnitude of the change in variables
        that is treated as a step to an unbounded solution.
        If the change in x during an iteration would exceed the value of
        r, the objective function is considered to be unbounded below in
        the feasible region.
        If r <= 0, the default value is used.

    'Line Search Tolerance' : float
        Default = 0.9

        The value r (0 <= r < 1) controls the accuracy with which the
        step alpha taken during each iteration approximates a minimum of
        the merit function along the search direction (the smaller the
        value of r, the more accurate the linesearch).
        The default value r = 0.9 requests an inaccurate search and is
        appropriate for most problems, particularly those with any
        nonlinear constraints.

        If there are no nonlinear constraints, a more accurate search
        may be appropriate when it is desirable to reduce the number of
        major iterations -- for example, if the objective function is
        cheap to evaluate, or if a substantial number of derivatives are
        unspecified.
        If r < 0 or r >= 1, the default value is used.

    'Linear Feasibility Tolerance' : float
        Default = sqrt(epsilon)

        The default value of r_2 is epsilon^0.33 if 'Derivative Level' =
        0 or 1, and sqrt(epsilon) otherwise.

        The scalars r_1 and r_2 define the maximum acceptable `absolute`
        violations in linear and nonlinear constraints at a 'feasible'
        point; i.e., a linear constraint is considered satisfied if its
        violation does not exceed r_1.
        Similarly a nonlinear constraint is considered satisfied if its
        violation does not exceed r_2.
        If r_m < epsilon or r_m >= 1, the default value is used, for m =
        1,2,...,2.

        On entry to ``nlp1_rcomm``, an iterative procedure is executed
        in order to find a point that satisfies the linear constraints
        and bounds on the variables to within the tolerance r_1.
        All subsequent iterates will satisfy the linear constraints to
        within the same tolerance (unless r_1 is comparable to the
        finite difference interval).

        For nonlinear constraints, the feasibility tolerance r_2 defines
        the largest constraint violation that is acceptable at an
        optimal point.
        Since nonlinear constraints are generally not satisfied until
        the final iterate, the value of option 'Nonlinear Feasibility
        Tolerance' acts as a partial termination criterion for the
        iterative sequence generated by ``nlp1_rcomm`` (see the
        discussion of option 'Optimality Tolerance').

        These tolerances should reflect the precision of the
        corresponding constraints.
        For example, if the variables and the coefficients in the linear
        constraints are of order unity, and the latter are correct to
        about 6 decimal digits, it would be appropriate to specify r_1
        as 10^(-6).

    'Nonlinear Feasibility Tolerance' : float
        Default = epsilon^0.33 or sqrt(epsilon)

        The default value of r_2 is epsilon^0.33 if 'Derivative Level' =
        0 or 1, and sqrt(epsilon) otherwise.

        The scalars r_1 and r_2 define the maximum acceptable `absolute`
        violations in linear and nonlinear constraints at a 'feasible'
        point; i.e., a linear constraint is considered satisfied if its
        violation does not exceed r_1.
        Similarly a nonlinear constraint is considered satisfied if its
        violation does not exceed r_2.
        If r_m < epsilon or r_m >= 1, the default value is used, for m =
        1,2,...,2.

        On entry to ``nlp1_rcomm``, an iterative procedure is executed
        in order to find a point that satisfies the linear constraints
        and bounds on the variables to within the tolerance r_1.
        All subsequent iterates will satisfy the linear constraints to
        within the same tolerance (unless r_1 is comparable to the
        finite difference interval).

        For nonlinear constraints, the feasibility tolerance r_2 defines
        the largest constraint violation that is acceptable at an
        optimal point.
        Since nonlinear constraints are generally not satisfied until
        the final iterate, the value of option 'Nonlinear Feasibility
        Tolerance' acts as a partial termination criterion for the
        iterative sequence generated by ``nlp1_rcomm`` (see the
        discussion of option 'Optimality Tolerance').

        These tolerances should reflect the precision of the
        corresponding constraints.
        For example, if the variables and the coefficients in the linear
        constraints are of order unity, and the latter are correct to
        about 6 decimal digits, it would be appropriate to specify r_1
        as 10^(-6).

    'List' : valueless
        Option 'List' enables printing of each option specification as
        it is supplied. 'Nolist' suppresses this printing.

    'Nolist' : valueless
        Default for ``nlp1_rcomm``

        Option 'List' enables printing of each option specification as
        it is supplied. 'Nolist' suppresses this printing.

    'Major Iteration Limit' : int
        Default = max(50, 3*(n+n_L)+10n_N)

        The value of i specifies the maximum number of major iterations
        allowed before termination.
        Setting i = 0 and 'Major Print Level' > 0 means that the
        workspace needed will be computed and printed, but no iterations
        will be performed.
        If i < 0, the default value is used.

    'Iteration Limit' : int
        Default = max(50, 3*(n+n_L)+10n_N)

        The value of i specifies the maximum number of major iterations
        allowed before termination.
        Setting i = 0 and 'Major Print Level' > 0 means that the
        workspace needed will be computed and printed, but no iterations
        will be performed.
        If i < 0, the default value is used.

    'Iters' : int
        Default = max(50, 3*(n+n_L)+10n_N)

        The value of i specifies the maximum number of major iterations
        allowed before termination.
        Setting i = 0 and 'Major Print Level' > 0 means that the
        workspace needed will be computed and printed, but no iterations
        will be performed.
        If i < 0, the default value is used.

    'Itns' : int
        Default = max(50, 3*(n+n_L)+10n_N)

        The value of i specifies the maximum number of major iterations
        allowed before termination.
        Setting i = 0 and 'Major Print Level' > 0 means that the
        workspace needed will be computed and printed, but no iterations
        will be performed.
        If i < 0, the default value is used.

    'Major Print Level' : int
        The value of i controls the amount of printout produced by the
        major iterations of ``nlp1_rcomm``, as indicated below.
        A detailed description of the printed output is given in
        [Further Comments] (summary output at each major iteration and
        the final solution) and [Monitoring Information] (monitoring
        information at each major iteration). (See also the description
        of the option 'Minor Print Level'.)

        The following printout is sent to the file object associated
        with the advisory I/O unit (see
        :class:`~naginterfaces.base.utils.FileObjManager`):

        [table omitted]

        The following printout is sent to the unit number given by the
        option 'Monitoring File':

        [table omitted]

        If 'Major Print Level' >= 5 and the unit number defined by the
        option 'Monitoring File' is the advisory unit number, the
        summary output for each major iteration is suppressed.

    'Print Level' : int
        Default for ``nlp1_rcomm`` = 0

        The value of i controls the amount of printout produced by the
        major iterations of ``nlp1_rcomm``, as indicated below.
        A detailed description of the printed output is given in
        [Further Comments] (summary output at each major iteration and
        the final solution) and [Monitoring Information] (monitoring
        information at each major iteration). (See also the description
        of the option 'Minor Print Level'.)

        The following printout is sent to the file object associated
        with the advisory I/O unit (see
        :class:`~naginterfaces.base.utils.FileObjManager`):

        [table omitted]

        The following printout is sent to the unit number given by the
        option 'Monitoring File':

        [table omitted]

        If 'Major Print Level' >= 5 and the unit number defined by the
        option 'Monitoring File' is the advisory unit number, the
        summary output for each major iteration is suppressed.

    'Minor Iteration Limit' : int
        Default = max(50, 3*(n+n_L+n_N))

        The value of i specifies the maximum number of iterations for
        finding a feasible point with respect to the bounds and linear
        constraints (if any).
        The value of i also specifies the maximum number of minor
        iterations for the optimality phase of each QP subproblem.
        If i <= 0, the default value is used.

    'Minor Print Level' : int
        Default = 0

        The value of i controls the amount of printout produced by the
        minor iterations of ``nlp1_rcomm`` (i.e., the iterations of the
        quadratic programming algorithm), as indicated below.
        A detailed description of the printed output is given in
        [Further Comments] (summary output at each minor iteration and
        the final QP solution) and [Monitoring Information]) (monitoring
        information at each minor iteration). (See also the description
        of the option 'Major Print Level'.)

        The following printout is sent to the file object associated
        with the advisory I/O unit (see
        :class:`~naginterfaces.base.utils.FileObjManager`):

        [table omitted]

        The following printout is sent to the unit number given by the
        option 'Monitoring File':

        [table omitted]

        If 'Major Print Level' >= 5 and the unit number defined by the
        option 'Monitoring File' is the advisory unit number, the
        summary output for each major iteration is suppressed.

    'Monitoring File' : int
        Default = -1

        If i >= 0 and 'Major Print Level' >= 5 or i >= 0 and 'Minor
        Print Level' >= 5, monitoring information produced by
        ``nlp1_rcomm`` at every iteration is sent to a file with logical
        unit number i. If i < 0 and/or 'Major Print Level' < 5 and
        'Minor Print Level' < 5, no monitoring information is produced.

    'Optimality Tolerance' : float
        Default = epsilon_r^0.8

        The argument r (epsilon_r <= r < 1) specifies the accuracy to
        which you wish the final iterate to approximate a solution of
        the problem.
        Broadly speaking, r indicates the number of correct figures
        desired in the objective function at the solution.
        For example, if r is 10^(-6) and ``nlp1_rcomm`` terminates
        successfully, the final value of F should have approximately six
        correct figures.
        If r < epsilon_r or r >= 1, the default value is used.

        ``nlp1_rcomm`` will terminate successfully if the iterative
        sequence of x values is judged to have converged and the final
        point satisfies the first-order Kuhn--Tucker conditions (see
        [Overview]).
        The sequence of iterates is considered to have converged at x if

            alpha||p|| <= sqrt(r)(1+||x||),

        where p is the search direction and alpha the step length from
        (3).
        An iterate is considered to satisfy the first-order conditions
        for a minimum if

            || Z^Tg_FR|| <= sqrt(r)(1+max(1+|F(x)|,||g_FR||))

        and

            |res_j| <= ftol for all j,

        where Z^Tg_FR is the projected gradient (see [Overview]), g_FR
        is the gradient of F(x) with respect to the free variables,
        res_j is the violation of the jth active nonlinear constraint,
        and ftol is the 'Nonlinear Feasibility Tolerance'.

    'Start Objective Check At Variable' : int
        Default = 1

        These keywords take effect only if 'Verify Level' > 0.
        They may be used to control the verification of gradient
        elements and/or Jacobian elements computed by the calling
        program during intermediate exits.
        For example, if the first 30 elements of the objective gradient
        appeared to be correct in an earlier run, so that only element
        31 remains questionable, it is reasonable to specify 'Start
        Objective Check At Variable' = 31.
        If the first 30 variables appear linearly in the objective, so
        that the corresponding gradient elements are constant, the above
        choice would also be appropriate.

        If i_{2*m-1} <= 0 or i_{2m-1} > min(n, i_{2m}), the default
        value is used, for m = 1,2,...,2.
        If i_{2*m} <= 0 or i_{2*m} > n, the default value is used, for m
        = 1,2,...,2.

    'Stop Objective Check At Variable' : int
        Default = n

        These keywords take effect only if 'Verify Level' > 0.
        They may be used to control the verification of gradient
        elements and/or Jacobian elements computed by the calling
        program during intermediate exits.
        For example, if the first 30 elements of the objective gradient
        appeared to be correct in an earlier run, so that only element
        31 remains questionable, it is reasonable to specify 'Start
        Objective Check At Variable' = 31.
        If the first 30 variables appear linearly in the objective, so
        that the corresponding gradient elements are constant, the above
        choice would also be appropriate.

        If i_{2*m-1} <= 0 or i_{2m-1} > min(n, i_{2m}), the default
        value is used, for m = 1,2,...,2.
        If i_{2*m} <= 0 or i_{2*m} > n, the default value is used, for m
        = 1,2,...,2.

    'Start Constraint Check At Variable' : int
        Default = 1

        These keywords take effect only if 'Verify Level' > 0.
        They may be used to control the verification of gradient
        elements and/or Jacobian elements computed by the calling
        program during intermediate exits.
        For example, if the first 30 elements of the objective gradient
        appeared to be correct in an earlier run, so that only element
        31 remains questionable, it is reasonable to specify 'Start
        Objective Check At Variable' = 31.
        If the first 30 variables appear linearly in the objective, so
        that the corresponding gradient elements are constant, the above
        choice would also be appropriate.

        If i_{2*m-1} <= 0 or i_{2m-1} > min(n, i_{2m}), the default
        value is used, for m = 1,2,...,2.
        If i_{2*m} <= 0 or i_{2*m} > n, the default value is used, for m
        = 1,2,...,2.

    'Stop Constraint Check At Variable' : int
        Default = n

        These keywords take effect only if 'Verify Level' > 0.
        They may be used to control the verification of gradient
        elements and/or Jacobian elements computed by the calling
        program during intermediate exits.
        For example, if the first 30 elements of the objective gradient
        appeared to be correct in an earlier run, so that only element
        31 remains questionable, it is reasonable to specify 'Start
        Objective Check At Variable' = 31.
        If the first 30 variables appear linearly in the objective, so
        that the corresponding gradient elements are constant, the above
        choice would also be appropriate.

        If i_{2*m-1} <= 0 or i_{2m-1} > min(n, i_{2m}), the default
        value is used, for m = 1,2,...,2.
        If i_{2*m} <= 0 or i_{2*m} > n, the default value is used, for m
        = 1,2,...,2.

    'Step Limit' : float
        Default = 2.0

        If r > 0,r specifies the maximum change in variables at the
        first step of the linesearch.
        In some cases, such as F(x) = ae^(bx) or F(x) = ax^b, even a
        moderate change in the elements of x can lead to floating-point
        overflow.
        The argument r is therefore used to encourage evaluation of the
        problem functions at meaningful points.
        Given any major iterate x, the first point x~ at which F and c
        are evaluated during the linesearch is restricted so that

            ||x~-x||_2 <= r(1+||x||_2).

        The linesearch may go on and evaluate F and c at points further
        from x if this will result in a lower value of the merit
        function (indicated by L at the end of each line of output
        produced by the major iterations; see [Further Comments]).
        If L is printed for most of the iterations, r should be set to a
        larger value.

        Wherever possible, upper and lower bounds on x should be used to
        prevent evaluation of nonlinear functions at wild values.
        The default value 'Step Limit' = 2.0 should not affect progress
        on well-behaved functions, but values such as 0.1 or 0.01 may be
        helpful when rapidly varying functions are present.
        If a small value of 'Step Limit' is selected, then a good
        starting point may be required.
        An important application is to the class of nonlinear least
        squares problems.
        If r <= 0, the default value is used.

    'Verify Level' : int
        Default = 0

        These keywords refer to finite difference checks on the gradient
        elements computed by the calling program during intermediate
        exits. (Unspecified gradient elements are not checked.) The
        possible choices for i are as follows:

        [table omitted]

        It is possible to specify 'Verify Level' = 0 to 3 in several
        ways.
        For example, the objective gradient will be verified if
        'Verify', 'Verify' = 'YES', 'Verify Gradients', 'Verify
        Objective Gradients' or 'Verify Level' = 1 is specified.
        The constraint gradients will be verified if 'Verify' = 'YES' or
        'Verify Level' = 2 or 'Verify' is specified.
        Similarly, the objective and the constraint gradients will be
        verified if 'Verify' = 'YES' or 'Verify Level' = 3 or 'Verify'
        is specified.

        If 0 <= i <= 3, gradients will be verified at the first point
        that satisfies the linear constraints and bounds.

        If i = 0, only a 'cheap' test will be performed, requiring one
        intermediate exit for the objective function gradients and (if
        appropriate) one intermediate exit for the partial derivatives
        of the constraints.

        If 1 <= i <= 3, a more reliable (but more expensive) check will
        be made on individual gradient elements, within the ranges
        specified by the 'Start Objective Check At Variable' and 'Stop
        Objective Check At Variable' keywords.
        A result of the form OK or BAD? is printed by ``nlp1_rcomm`` to
        indicate whether or not each element appears to be correct.
        If a gradient element is determined to be extremely poor (i.e.,
        if it appears to have no significant digits of accuracy at all),
        then ``nlp1_rcomm`` will also exit with an error indicator in
        argument errno.

        If 10 <= i <= 13, the action is the same as for i < 10, except
        that it will take place at the user-specified initial value of
        x.

        If i < -1 or 4 <= i <= 9 or i > 13, the default value is used.

        We suggest that 'Verify Level' = 3 be used whenever a new
        calling program is being developed.

    'Verify' : valueless
        These keywords refer to finite difference checks on the gradient
        elements computed by the calling program during intermediate
        exits. (Unspecified gradient elements are not checked.) The
        possible choices for i are as follows:

        [table omitted]

        It is possible to specify 'Verify Level' = 0 to 3 in several
        ways.
        For example, the objective gradient will be verified if
        'Verify', 'Verify' = 'YES', 'Verify Gradients', 'Verify
        Objective Gradients' or 'Verify Level' = 1 is specified.
        The constraint gradients will be verified if 'Verify' = 'YES' or
        'Verify Level' = 2 or 'Verify' is specified.
        Similarly, the objective and the constraint gradients will be
        verified if 'Verify' = 'YES' or 'Verify Level' = 3 or 'Verify'
        is specified.

        If 0 <= i <= 3, gradients will be verified at the first point
        that satisfies the linear constraints and bounds.

        If i = 0, only a 'cheap' test will be performed, requiring one
        intermediate exit for the objective function gradients and (if
        appropriate) one intermediate exit for the partial derivatives
        of the constraints.

        If 1 <= i <= 3, a more reliable (but more expensive) check will
        be made on individual gradient elements, within the ranges
        specified by the 'Start Objective Check At Variable' and 'Stop
        Objective Check At Variable' keywords.
        A result of the form OK or BAD? is printed by ``nlp1_rcomm`` to
        indicate whether or not each element appears to be correct.
        If a gradient element is determined to be extremely poor (i.e.,
        if it appears to have no significant digits of accuracy at all),
        then ``nlp1_rcomm`` will also exit with an error indicator in
        argument errno.

        If 10 <= i <= 13, the action is the same as for i < 10, except
        that it will take place at the user-specified initial value of
        x.

        If i < -1 or 4 <= i <= 9 or i > 13, the default value is used.

        We suggest that 'Verify Level' = 3 be used whenever a new
        calling program is being developed.

    'Verify Constraint Gradients' : valueless
        These keywords refer to finite difference checks on the gradient
        elements computed by the calling program during intermediate
        exits. (Unspecified gradient elements are not checked.) The
        possible choices for i are as follows:

        [table omitted]

        It is possible to specify 'Verify Level' = 0 to 3 in several
        ways.
        For example, the objective gradient will be verified if
        'Verify', 'Verify' = 'YES', 'Verify Gradients', 'Verify
        Objective Gradients' or 'Verify Level' = 1 is specified.
        The constraint gradients will be verified if 'Verify' = 'YES' or
        'Verify Level' = 2 or 'Verify' is specified.
        Similarly, the objective and the constraint gradients will be
        verified if 'Verify' = 'YES' or 'Verify Level' = 3 or 'Verify'
        is specified.

        If 0 <= i <= 3, gradients will be verified at the first point
        that satisfies the linear constraints and bounds.

        If i = 0, only a 'cheap' test will be performed, requiring one
        intermediate exit for the objective function gradients and (if
        appropriate) one intermediate exit for the partial derivatives
        of the constraints.

        If 1 <= i <= 3, a more reliable (but more expensive) check will
        be made on individual gradient elements, within the ranges
        specified by the 'Start Objective Check At Variable' and 'Stop
        Objective Check At Variable' keywords.
        A result of the form OK or BAD? is printed by ``nlp1_rcomm`` to
        indicate whether or not each element appears to be correct.
        If a gradient element is determined to be extremely poor (i.e.,
        if it appears to have no significant digits of accuracy at all),
        then ``nlp1_rcomm`` will also exit with an error indicator in
        argument errno.

        If 10 <= i <= 13, the action is the same as for i < 10, except
        that it will take place at the user-specified initial value of
        x.

        If i < -1 or 4 <= i <= 9 or i > 13, the default value is used.

        We suggest that 'Verify Level' = 3 be used whenever a new
        calling program is being developed.

    'Verify Gradients' : valueless
        These keywords refer to finite difference checks on the gradient
        elements computed by the calling program during intermediate
        exits. (Unspecified gradient elements are not checked.) The
        possible choices for i are as follows:

        [table omitted]

        It is possible to specify 'Verify Level' = 0 to 3 in several
        ways.
        For example, the objective gradient will be verified if
        'Verify', 'Verify' = 'YES', 'Verify Gradients', 'Verify
        Objective Gradients' or 'Verify Level' = 1 is specified.
        The constraint gradients will be verified if 'Verify' = 'YES' or
        'Verify Level' = 2 or 'Verify' is specified.
        Similarly, the objective and the constraint gradients will be
        verified if 'Verify' = 'YES' or 'Verify Level' = 3 or 'Verify'
        is specified.

        If 0 <= i <= 3, gradients will be verified at the first point
        that satisfies the linear constraints and bounds.

        If i = 0, only a 'cheap' test will be performed, requiring one
        intermediate exit for the objective function gradients and (if
        appropriate) one intermediate exit for the partial derivatives
        of the constraints.

        If 1 <= i <= 3, a more reliable (but more expensive) check will
        be made on individual gradient elements, within the ranges
        specified by the 'Start Objective Check At Variable' and 'Stop
        Objective Check At Variable' keywords.
        A result of the form OK or BAD? is printed by ``nlp1_rcomm`` to
        indicate whether or not each element appears to be correct.
        If a gradient element is determined to be extremely poor (i.e.,
        if it appears to have no significant digits of accuracy at all),
        then ``nlp1_rcomm`` will also exit with an error indicator in
        argument errno.

        If 10 <= i <= 13, the action is the same as for i < 10, except
        that it will take place at the user-specified initial value of
        x.

        If i < -1 or 4 <= i <= 9 or i > 13, the default value is used.

        We suggest that 'Verify Level' = 3 be used whenever a new
        calling program is being developed.

    'Verify Objective Gradients' : valueless
        These keywords refer to finite difference checks on the gradient
        elements computed by the calling program during intermediate
        exits. (Unspecified gradient elements are not checked.) The
        possible choices for i are as follows:

        [table omitted]

        It is possible to specify 'Verify Level' = 0 to 3 in several
        ways.
        For example, the objective gradient will be verified if
        'Verify', 'Verify' = 'YES', 'Verify Gradients', 'Verify
        Objective Gradients' or 'Verify Level' = 1 is specified.
        The constraint gradients will be verified if 'Verify' = 'YES' or
        'Verify Level' = 2 or 'Verify' is specified.
        Similarly, the objective and the constraint gradients will be
        verified if 'Verify' = 'YES' or 'Verify Level' = 3 or 'Verify'
        is specified.

        If 0 <= i <= 3, gradients will be verified at the first point
        that satisfies the linear constraints and bounds.

        If i = 0, only a 'cheap' test will be performed, requiring one
        intermediate exit for the objective function gradients and (if
        appropriate) one intermediate exit for the partial derivatives
        of the constraints.

        If 1 <= i <= 3, a more reliable (but more expensive) check will
        be made on individual gradient elements, within the ranges
        specified by the 'Start Objective Check At Variable' and 'Stop
        Objective Check At Variable' keywords.
        A result of the form OK or BAD? is printed by ``nlp1_rcomm`` to
        indicate whether or not each element appears to be correct.
        If a gradient element is determined to be extremely poor (i.e.,
        if it appears to have no significant digits of accuracy at all),
        then ``nlp1_rcomm`` will also exit with an error indicator in
        argument errno.

        If 10 <= i <= 13, the action is the same as for i < 10, except
        that it will take place at the user-specified initial value of
        x.

        If i < -1 or 4 <= i <= 9 or i > 13, the default value is used.

        We suggest that 'Verify Level' = 3 be used whenever a new
        calling program is being developed.

    Raises
    ------
    NagValueError
        (`errno` 7)
            Large errors found in the derivatives.

        (`errno` 9)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 9)
            On entry, `nclin` = *<value>*.

            Constraint: `nclin` >= 0.

        (`errno` 9)
            On entry, ncnln = *<value>*.

            Constraint: ncnln >= 0.

        (`errno` 9)
            On entry, the equal bounds on *<value>* are infinite,
            because `bl`[*<value>*] = beta and `bu`[*<value>*] = beta,
            but abs(beta) >= bigbnd: beta = *<value>* and bigbnd =
            *<value>*.

        (`errno` 9)
            On entry, the bounds on *<value>* are inconsistent:
            `bl`[*<value>*] = *<value>* and `bu`[*<value>*] = *<value>*.

        (`errno` 9)
            On entry with a Warm Start, `istate`[*<value>*] = *<value>*.

        (`errno` 9)
            On entry, `irevcm` = *<value>*.

            Constraint: 6 >= `irevcm` >= 0.

        (`errno` 9)
            On entry, the equal bounds on variable *<value>* are
            infinite, because `bl`[*<value>*] = beta and `bu`[*<value>*]
            = beta, but abs(beta) >= bigbnd: beta = *<value>* and bigbnd
            = *<value>*.

        (`errno` 9)
            On entry, the equal bounds on linear constraint *<value>*
            are infinite, because `bl`[*<value>*] = beta and
            `bu`[*<value>*] = beta, but abs(beta) >= bigbnd: beta =
            *<value>* and bigbnd = *<value>*.

        (`errno` 9)
            On entry, the equal bounds on nonlinear constraint *<value>*
            are infinite, because `bl`[*<value>*] = beta and
            `bu`[*<value>*] = beta, but abs(beta) >= bigbnd: beta =
            *<value>* and bigbnd = *<value>*.

        (`errno` 9)
            On entry, the bounds on variable *<value>* are inconsistent:
            `bl`[*<value>*] = *<value>* and `bu`[*<value>*] = *<value>*.

        (`errno` 9)
            On entry, the bounds on linear constraint *<value>* are
            inconsistent: `bl`[*<value>*] = *<value>* and
            `bu`[*<value>*] = *<value>*.

        (`errno` 9)
            On entry, the bounds on nonlinear constraint *<value>* are
            inconsistent: `bl`[*<value>*] = *<value>* and
            `bu`[*<value>*] = *<value>*.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i < 0)
            User requested termination by setting `irevcm` negative
            during an intermediate exit.

        (`errno` 1)
            Optimal solution found, but requested accuracy not achieved.

        (`errno` 2)
            No feasible point for the linear constraints.

        (`errno` 3)
            No feasible point for the nonlinear constraints.

        (`errno` 6)
            Current point cannot be improved upon.

    NagAlgorithmicMajorWarning
        (`errno` 4)
            Too many major iterations.

    References
    ----------
    Dennis, J E Jr and Moré, J J, 1977, `Quasi-Newton methods,
    motivation and theory`, SIAM Rev. (19), 46--89

    Dennis, J E Jr and Schnabel, R B, 1981, `A new derivation of
    symmetric positive-definite secant updates`, nonlinear programming,
    (eds O L Mangasarian, R R Meyer and S M Robinson) (4), 167--199,
    Academic Press

    Dennis, J E Jr and Schnabel, R B, 1983, `Numerical Methods for
    Unconstrained Optimization and Nonlinear Equations`, Prentice--Hall

    Fletcher, R, 1987, `Practical Methods of Optimization`, (2nd
    Edition), Wiley

    Gill, P E, Hammarling, S, Murray, W, Saunders, M A and Wright, M H,
    1986, `Users' guide for LSSOL (Version 1.0)`, Report SOL 86-1,
    Department of Operations Research, Stanford University

    Gill, P E, Murray, W, Saunders, M A and Wright, M H, 1984, `Users'
    guide for SOL/QPSOL version 3.2`, Report SOL, 84--5, Department of
    Operations Research, Stanford University

    Gill, P E, Murray, W, Saunders, M A and Wright, M H, 1984,
    `Procedures for optimization problems with a mixture of bounds and
    general linear constraints`, ACM Trans. Math. Software (10),
    282--298

    Gill, P E, Murray, W, Saunders, M A and Wright, M H, 1986, `Some
    theoretical properties of an augmented Lagrangian merit function`,
    Report SOL, 86--6R, Department of Operations Research, Stanford
    University

    Gill, P E, Murray, W, Saunders, M A and Wright, M H, 1986, `Users'
    guide for NPSOL (Version 4.0): a Fortran package for nonlinear
    programming`, Report SOL 86-2, Department of Operations Research,
    Stanford University

    Gill, P E, Murray, W and Wright, M H, 1981, `Practical
    Optimization`, Academic Press

    Hock, W and Schittkowski, K, 1981, `Test Examples for Nonlinear
    Programming Codes. Lecture Notes in Economics and Mathematical
    Systems` (187), Springer--Verlag

    Murtagh, B A and Saunders, M A, 1983, `MINOS 5.0 user's guide`,
    Report SOL 83-20, Department of Operations Research, Stanford
    University

    Powell, M J D, 1974, `Introduction to constrained optimization`,
    Numerical Methods for Constrained Optimization, (eds P E Gill and W
    Murray), 1--28, Academic Press

    Powell, M J D, 1983, `Variable metric methods in constrained
    optimization`, Mathematical Programming: the State of the Art, (eds
    A Bachem, M Grötschel and B Korte), 288--311, Springer--Verlag

    See Also
    --------
    :meth:`naginterfaces.library.examples.opt.nlp1_rcomm_ex.main`
    """

    fun_name = 'naginterfaces.library.opt.nlp1_rcomm'

    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'objgrd'},
        objgrd,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(objgrd, {'fun_name': fun_name, 'entity': 'objgrd'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'c'},
        c,
    )
    _ncnln_py = _b_utils._EngineFloat64ArrayType.get_shape(c, {'fun_name': fun_name, 'entity': 'c'}, exp_rank=1)[0]
    _ncnln = _b_utils._EngineIntScalarType(
        _ncnln_py, {'fun_name': fun_name, 'entity': '_ncnln'},
    ).to_ctypes()
    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _bl = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'bl'},
        bl,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _bu = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'bu'},
        bu,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'istate'},
        istate,
    )
    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'cjac'},
        cjac,
    )
    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'clamda'},
        clamda,
    )
    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'r'},
        r,
    )
    _utils._check_is_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
    )
    _needc = _b_utils._EngineIntArrayType.empty_ndarray(
        (max(1, _ncnln_py),),
    )
    try:
        _return_scalars = _b_opt.nlp1_rcomm(
            irevcm,
            _n,
            nclin,
            _ncnln,
            _a,
            _bl,
            _bu,
            itera,
            istate,
            c,
            cjac,
            clamda,
            objf,
            objgrd,
            r,
            x,
            _needc,
            comm,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _irevcm_py = _return_scalars.irevcm
    _itera_py = _return_scalars.itera
    _objf_py = _return_scalars.objf
    _needc_py = _needc
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['irevcm', 'itera', 'objf', 'needc'],
    )(
        _irevcm_py,
        _itera_py,
        _objf_py,
        _needc_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def nlp1_sparse_solve(m, ncnln, nonln, njnln, iobj, a, ha, ka, bl, bu, start, names, ns, xs, istate, clamda, comm, confun=None, objfun=None, leniz=None, lenz=500, data=None, io_manager=None):
    u"""
    Nonlinear programming (NLP), sparse, active-set SQP method, using
    function values and optionally first derivatives.

    Note: this function uses optional algorithmic parameters, see also:
    ``nlp1_sparse_option_file``, ``nlp1_sparse_option_string``,
    ``nlp1_init``.

    ``nlp1_sparse_solve`` solves sparse nonlinear programming problems.

    For full information please refer to the NAG Library document for
    e04ug

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04ugf.html

    Parameters
    ----------
    m : int
        m, the number of general constraints (or slacks). This is the
        number of rows in A, including the free row (if any; see
        `iobj`). Note that A must contain at least one row. If your
        problem has no constraints, or only upper and lower bounds on
        the variables, then you must include a dummy 'free' row
        consisting of a single (zero) element subject to 'infinite'
        upper and lower bounds. Further details can be found under the
        descriptions for `iobj`, nnz, `a`, `ha`, `ka`, `bl` and `bu`.

    ncnln : int
        n_N, the number of nonlinear constraints.

    nonln : int
        n'_1, the number of nonlinear objective variables. If the
        objective function is nonlinear, the leading n'_1 columns of A
        belong to the nonlinear objective variables. (See also the
        description for `njnln`.)

    njnln : int
        n_1^'', the number of nonlinear Jacobian variables. If there are
        any nonlinear constraints, the leading n_1^'' columns of A
        belong to the nonlinear Jacobian variables. If n'_1 > 0 and
        n_1^'' > 0, the nonlinear objective and Jacobian variables
        overlap. The total number of nonlinear variables is given by n_=
        max(n'_1, n_1^'').

    iobj : int
        If `iobj` > `ncnln`, row `iobj` of A is a free row containing
        the nonzero elements of the linear part of the objective
        function.

        `iobj` = 0

            There is no free row.

        `iobj` = -1

            There is a dummy 'free' row.

    a : float, array-like, shape (nnz)
        The nonzero elements of the Jacobian matrix A, ordered by
        increasing column index. Since the constraint Jacobian matrix
        J(x^('')) must always appear in the top left-hand corner of A,
        those elements in a column associated with any nonlinear
        constraints must come before any elements belonging to the
        linear constraint matrix G and the free row (if any; see
        `iobj`).

        In general, `a` is partitioned into a nonlinear part and a
        linear part corresponding to the nonlinear variables and linear
        variables in the problem.

        Elements in the nonlinear part may be set to any value (e.g.,
        zero) because they are initialized at the first point that
        satisfies the linear constraints and the upper and lower bounds.

        If 'Derivative Level' = 2 or 3, the nonlinear part may also be
        used to store any constant Jacobian elements.

        Note that if `confun` does not define the constant Jacobian
        element `fjac`[i-1], then the missing value will be obtained
        directly from `a`[j] for some j >= i.

        If 'Derivative Level' = 0 or 1, unassigned elements of `fjac`
        are `not` treated as constant; they are estimated by finite
        differences, at nontrivial expense.

        The linear part must contain the nonzero elements of G and the
        free row (if any).

        If `iobj` = -1, set `a`[0] = 0.

        Elements with the same row and column indices are not allowed.
        (See also the descriptions for `ha` and `ka`.)

    ha : int, array-like, shape (nnz)
        `ha`[i-1] must contain the row index of the nonzero element
        stored in `a`[i-1], for i = 1,2,...,nnz. The row indices for a
        column may be supplied in any order subject to the condition
        that those elements in a column associated with any nonlinear
        constraints must appear before those elements associated with
        any linear constraints (including the free row, if any). Note
        that `confun` must define the Jacobian elements in the same
        order. If `iobj` = -1, set `ha`[0] = 1.

    ka : int, array-like, shape (n+1)
        `ka`[j-1] must contain the index in `a` of the start of the jth
        column, for j = 1,2,...,n. To specify the jth column as empty,
        set `ka`[j-1] = `ka`[j]. Note that the first and last elements
        of `ka` must be such that `ka`[0] = 1 and `ka`[n] = nnz+1. If
        `iobj` = -1, set `ka`[j-1] = 2, for j = 2,3,...,n.

    bl : float, array-like, shape (n+`m`)
        l, the lower bounds for all the variables and general
        constraints, in the following order. The first n elements of
        `bl` must contain the bounds on the variables x, the next
        `ncnln` elements the bounds for the nonlinear constraints F(x)
        (if any) and the next (`m`-`ncnln`) elements the bounds for the
        linear constraints Gx and the free row (if any). To specify a
        nonexistent lower bound (i.e., l_j = - infinity), set `bl`[j-1]
        <= -bigbnd. To specify the jth constraint as an `equality`, set
        `bl`[j-1] = `bu`[j-1] = beta, say, where |beta| < bigbnd. If
        `iobj` = -1, set `bl`[n+abs(`iobj`)-1] <= -bigbnd.

    bu : float, array-like, shape (n+`m`)
        u, the upper bounds for all the variables and general
        constraints, in the following order. The first n elements of
        `bu` must contain the bounds on the variables x, the next
        `ncnln` elements the bounds for the nonlinear constraints F(x)
        (if any) and the next (`m`-`ncnln`) elements the bounds for the
        linear constraints Gx and the free row (if any). To specify a
        nonexistent upper bound (i.e., u_j = + infinity), set `bu`[j-1]
        >= bigbnd. To specify the jth constraint as an `equality`, set
        `bu`[j-1] = `bl`[j-1] = beta, say, where |beta| < bigbnd. If
        `iobj` = -1, set `bu`[n+abs(`iobj`)-1] >= bigbnd.

    start : str, length 1
        Indicates how a starting basis is to be obtained.

        `start` = 'C'

            An internal Crash procedure will be used to choose an
            initial basis.

        `start` = 'W'

            A basis is already defined in `istate` and `ns` (probably
            from a previous call).

    names : str, length 8, array-like, shape (nname)
        Specifies the column and row names to be used in the printed
        output.

        If nname = 1, `names` is not referenced and the printed output
        will use default names for the columns and rows.

        If nname = n+`m`, the first n elements must contain the names
        for the columns, the next `ncnln` elements must contain the
        names for the nonlinear rows (if any) and the next (`m`-`ncnln`)
        elements must contain the names for the linear rows (if any) to
        be used in the printed output.

        Note that the name for the free row or dummy 'free' row must be
        stored in `names`[n+abs(`iobj`)-1].

    ns : int
        n_S, the number of superbasics. It need not be specified if
        `start` = 'C', but must retain its value from a previous call
        when `start` = 'W'.

    xs : float, array-like, shape (n+`m`)
        The initial values of the variables and slacks (x,s). (See the
        description for `istate`.)

    istate : int, array-like, shape (n+`m`)
        If `start` = 'C', the first n elements of `istate` and `xs` must
        specify the initial states and values, respectively, of the
        variables x. (The slacks s need not be initialized.) An internal
        Crash procedure is then used to select an initial basis matrix
        B. The initial basis matrix will be triangular (neglecting
        certain small elements in each column). It is chosen from
        various rows and columns of (A -I). Possible values for
        `istate`[j-1] are as follows:

        [table omitted]

        If nothing special is known about the problem, or there is no
        wish to provide special information, you may set `istate`[j-1] =
        0 and `xs`[j-1] = 0.0, for j = 1,2,...,n.

        All variables will then be eligible for the initial basis.

        Less trivially, to say that the jth variable will probably be
        equal to one of its bounds, set `istate`[j-1] = 4 and `xs`[j-1]
        = `bl`[j-1] or `istate`[j-1] = 5 and `xs`[j-1] = `bu`[j-1] as
        appropriate.

        Following the Crash procedure, variables for which `istate`[j-1]
        = 2 are made superbasic.

        Other variables not selected for the basis are then made
        nonbasic at the value `xs`[j-1] if `bl`[j-1] <= `xs`[j-1] <=
        `bu`[j-1], or at the value `bl`[j-1] or `bu`[j-1] closest to
        `xs`[j-1].

        If `start` = 'W', `istate` and `xs` must specify the initial
        states and values, respectively, of the variables and slacks
        (x,s).

        If the function has been called previously with the same values
        of n and `m`, `istate` already contains satisfactory
        information.

    clamda : float, array-like, shape (n+`m`)
        If `ncnln` > 0, `clamda`[j-1] must contain a Lagrange multiplier
        estimate for the jth nonlinear constraint F_j(x), for j =
        n+1,...,n+`ncnln`. If nothing special is known about the
        problem, or there is no wish to provide special information, you
        may set `clamda`[j-1] = 0.0. The remaining elements need not be
        set.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``nlp1_init``.

    confun : None or callable (mode, f, fjac) = confun(mode, ncnln, x,
        fjac, nstate, data=None), optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `confun` must calculate the vector F(x) of nonlinear constraint
        functions and (optionally) its Jacobian (= dF/dx) for a
        specified n_1^'' (<= n) element vector x.

        If there are no nonlinear constraints (i.e., `ncnln` = 0),
        `confun` will never be called by ``nlp1_sparse_solve`` and
        `confun` may be None. If there are nonlinear constraints, the
        first call to `confun` will occur before the first call to
        `objfun`.

        Parameters
        ~~~~~~~~~~
        mode : int
            Indicates which values must be assigned during each call of
            `confun`. Only the following values need be assigned:

            `mode` = 0

                `f`.

            `mode` = 1

                All available elements of `fjac`.

            `mode` = 2

                `f` and all available elements of `fjac`.

        ncnln : int
            n_N, the number of nonlinear constraints. These must be the
            first `ncnln` constraints in the problem.

        x : float, ndarray, shape (njnln)
            x, the vector of nonlinear Jacobian variables at which the
            nonlinear constraint functions and/or the available elements
            of the constraint Jacobian are to be evaluated.

        fjac : float, ndarray, shape (nnzjac)
            The elements of `fjac` are set to special values which
            enable ``nlp1_sparse_solve`` to detect whether they are
            changed by `confun`.

        nstate : int
            If `nstate` = 1, then ``nlp1_sparse_solve`` is calling
            `confun` for the first time. This argument setting allows
            you to save computation time if certain data must be read or
            calculated only once.

            If `nstate` >= 2, ``nlp1_sparse_solve`` is calling `confun`
            for the last time.

            This argument setting allows you to perform some additional
            computation on the final solution.

            In general, the last call to `confun` is made with `nstate`
            = 2+errno (see [Raises/Warns]).

            Otherwise, `nstate` = 0.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        mode : int
            You may set to a negative value as follows:

            `mode` <= -2

                The solution to the current problem is terminated and in
                this case ``nlp1_sparse_solve`` will terminate with
                errno set to `mode`.

            `mode` = -1

                The nonlinear constraint functions cannot be calculated
                at the current x. ``nlp1_sparse_solve`` will then
                terminate with `errno` = -1 unless this occurs during
                the linesearch; in this case, the linesearch will
                shorten the step and try again.

        f : float, array-like, shape (`ncnln`)
            If `mode` = 0 or 2, `f`[i-1] must contain the value of the
            ith nonlinear constraint function at x.

        fjac : float, array-like, shape (nnzjac)
            If `mode` = 1 or 2, `fjac` must return the available
            elements of the constraint Jacobian evaluated at x. These
            elements must be stored in exactly the same positions as
            implied by the definitions of the arrays `a`, `ha` and `ka`.
            If option 'Derivative Level' = 2 or 3, the value of any
            constant Jacobian element not defined by `confun` will be
            obtained directly from `a`. Note that the function does not
            perform any internal checks for consistency (except
            indirectly via the option 'Verify Level'), so great care is
            essential.

    objfun : None or callable (mode, objf, objgrd) = objfun(mode, x,
        objgrd, nstate, data=None), optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `objfun` must calculate the nonlinear part of the objective
        function f(x) and (optionally) its gradient (= df/dx) for a
        specified n'_1 (<= n) element vector x.

        If there are no nonlinear objective variables (i.e., `nonln` =
        0), `objfun` will never be called by ``nlp1_sparse_solve`` and
        `objfun` may be None.

        Parameters
        ~~~~~~~~~~
        mode : int
            Indicates which values must be assigned during each call of
            `objfun`. Only the following values need be assigned:

            `mode` = 0

                `objf`.

            `mode` = 1

                All available elements of `objgrd`.

            `mode` = 2

                `objf` and all available elements of `objgrd`.

        x : float, ndarray, shape (nonln)
            x, the vector of nonlinear variables at which the nonlinear
            part of the objective function and/or all available elements
            of its gradient are to be evaluated.

        objgrd : float, ndarray, shape (nonln)
            The elements of `objgrd` are set to special values which
            enable ``nlp1_sparse_solve`` to detect whether they are
            changed by `objfun`.

        nstate : int
            If `nstate` = 1, ``nlp1_sparse_solve`` is calling `objfun`
            for the first time. This argument setting allows you to save
            computation time if certain data must be read or calculated
            only once.

            If `nstate` >= 2, ``nlp1_sparse_solve`` is calling `objfun`
            for the last time.

            This argument setting allows you to perform some additional
            computation on the final solution.

            In general, the last call to `objfun` is made with `nstate`
            = 2+errno (see [Raises/Warns]).

            Otherwise, `nstate` = 0.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        mode : int
            You may set to a negative value as follows:

            `mode` <= -2

                The solution to the current problem is terminated and in
                this case ``nlp1_sparse_solve`` will terminate with
                errno set to `mode`.

            `mode` = -1

                The nonlinear part of the objective function cannot be
                calculated at the current x. ``nlp1_sparse_solve`` will
                then terminate with `errno` = -1 unless this occurs
                during the linesearch; in this case, the linesearch will
                shorten the step and try again.

        objf : float
            If `mode` = 0 or 2, `objf` must be set to the value of the
            objective function at x.

        objgrd : float, array-like, shape (nonln)
            If `mode` = 1 or 2, `objgrd` must return the available
            elements of the gradient evaluated at x.

    leniz : None or int, optional
        Note: if this argument is None then a default value will be
        used, determined as follows: max(500, n+`m`).

        The dimension of the internal workspace array iz.

    lenz : int, optional
        The dimension of the internal workspace array z.

    data : arbitrary, optional
        User-communication data for callback functions.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    a : float, ndarray, shape (nnz)
        Elements in the nonlinear part corresponding to nonlinear
        Jacobian variables are overwritten.

    ns : int
        The final number of superbasics.

    xs : float, ndarray, shape (n+`m`)
        The final values of the variables and slacks (x,s).

    istate : int, ndarray, shape (n+`m`)
        The final states of the variables and slacks (x,s). The
        significance of each possible value of `istate`[j-1] is as
        follows:

        [table omitted]

        If `ninf` = 0, basic and superbasic variables may be outside
        their bounds by as much as the value of the option 'Minor
        Feasibility Tolerance'.

        Note that if scaling is specified, the option 'Minor Feasibility
        Tolerance' applies to the variables of the `scaled` problem.

        In this case, the variables of the original problem may be as
        much as 0.1 outside their bounds, but this is unlikely unless
        the problem is very badly scaled.

        Very occasionally some nonbasic variables may be outside their
        bounds by as much as the option 'Minor Feasibility Tolerance'
        and there may be some nonbasic variables for which `xs`[j-1]
        lies strictly between its bounds.

        If `ninf` > 0, some basic and superbasic variables may be
        outside their bounds by an arbitrary amount (bounded by `sinf`
        if scaling was not used).

    clamda : float, ndarray, shape (n+`m`)
        A set of Lagrange multipliers for the bounds on the variables
        (`reduced costs`) and the general constraints (`shadow costs`).
        More precisely, the first n elements contain the multipliers for
        the bounds on the variables, the next `ncnln` elements contain
        the multipliers for the nonlinear constraints F(x) (if any) and
        the next (`m`-`ncnln`) elements contain the multipliers for the
        linear constraints Gx and the free row (if any).

    miniz : int
        The minimum value of `leniz` required to start solving the
        problem. If `errno` = 12, ``nlp1_sparse_solve`` may be called
        again with `leniz` suitably larger than `miniz`. (The bigger the
        better, since it is not certain how much workspace the basis
        factors need.)

    minz : int
        The minimum value of `lenz` required to start solving the
        problem. If `errno` = 13, ``nlp1_sparse_solve`` may be called
        again with `lenz` suitably larger than `minz`. (The bigger the
        better, since it is not certain how much workspace the basis
        factors need.)

    ninf : int
        The number of constraints that lie outside their bounds by more
        than the value of the option 'Minor Feasibility Tolerance'.

        If the `linear` constraints are infeasible, the sum of the
        infeasibilities of the linear constraints is minimized subject
        to the upper and lower bounds being satisfied.

        In this case, `ninf` contains the number of elements of Gx that
        lie outside their upper or lower bounds.

        Note that the nonlinear constraints are not evaluated.

        Otherwise, the sum of the infeasibilities of the `nonlinear`
        constraints is minimized subject to the linear constraints and
        the upper and lower bounds being satisfied.

        In this case, `ninf` contains the number of elements of F(x)
        that lie outside their upper or lower bounds.

    sinf : float
        The sum of the infeasibilities of constraints that lie outside
        their bounds by more than the value of the option 'Minor
        Feasibility Tolerance'.

    obj : float
        The value of the objective function.

    Other Parameters
    ----------------
    'Central Difference Interval' : float
        Default = cubreroot('Function Precision')

        Note that this option does not apply when 'Derivative Level' =
        3.

        The value of r is used near an optimal solution in order to
        obtain more accurate (but more expensive) estimates of
        gradients.
        This requires twice as many function evaluations as compared to
        using forward differences (see option 'Forward Difference
        Interval').
        The interval used for the jth variable is h_j = r(1+|x_j|).
        The resulting gradient estimates should be accurate to O(r^2),
        unless the functions are badly scaled.
        The switch to central differences is indicated by c at the end
        of each line of intermediate printout produced by the major
        iterations (see [Major Iteration Printout]).
        See Gill `et al.` (1981) for a discussion of the accuracy in
        finite difference approximations.

        If r <= 0, the default value is used.

    'Check Frequency' : int
        Default = 60

        Every ith minor iteration after the most recent basis
        factorization, a numerical test is made to see if the current
        solution (x,s) satisfies the general linear constraints
        (including any linearized nonlinear constraints).
        The constraints are of the form Ax-s = b, where s is the set of
        slack variables.
        If the largest element of the residual vector r = b-Ax+s is
        judged to be too large, the current basis is refactorized and
        the basic variables recomputed to satisfy the general
        constraints more accurately.

        If i < 0, the default value is used.
        If i = 0, the value i = 99999999 is used and effectively no
        checks are made.

    'Crash Option' : int
        Default = 0 or 3

        The default value of i is 0 if there are any nonlinear
        constraints and 3 otherwise.
        Note that this option does not apply when `start` = 'W' (see
        [Parameters]).

        If `start` = 'C', an internal Crash procedure is used to select
        an initial basis from various rows and columns of the constraint
        matrix (A -I).
        The value of i determines which rows and columns of A are
        initially eligible for the basis and how many times the Crash
        procedure is called.
        Columns of -I are used to pad the basis where necessary.
        The possible choices for i are the following.

        [table omitted]

        If i < 0 or i > 3, the default value is used.

        If i >= 1, certain slacks on inequality rows are selected for
        the basis first. (If i >= 2, numerical values are used to
        exclude slacks that are close to a bound.) The Crash procedure
        then makes several passes through the columns of A, searching
        for a basis matrix that is essentially triangular.
        A column is assigned to 'pivot' on a particular row if the
        column contains a suitably large element in a row that has not
        yet been assigned. (The pivot elements ultimately form the
        diagonals of the triangular basis.) For remaining unassigned
        rows, slack variables are inserted to complete the basis.

    'Crash Tolerance' : float
        Default = 0.1

        The value r (0 <= r < 1) allows the Crash procedure to ignore
        certain 'small' nonzero elements in the columns of A while
        searching for a triangular basis.
        If a_max is the largest element in the jth column, other
        nonzeros a_{ij} in the column are ignored if |a_{ij}| <=
        a_max*r.

        When r > 0, the basis obtained by the Crash procedure may not be
        strictly triangular, but it is likely to be nonsingular and
        almost triangular.
        The intention is to obtain a starting basis containing more
        columns of A and fewer (arbitrary) slacks.
        A feasible solution may be reached earlier on some problems.

        If r < 0 or r >= 1, the default value is used.

    'Defaults' : valueless
        This special keyword may be used to reset all options to their
        default values.

    'Derivative Level' : int
        Default = 3

        This argument indicates which nonlinear function gradients are
        provided in functions `objfun` and `confun`.
        The possible choices for i are the following.

        [table omitted]

        The default value i = 3 should be used whenever possible.
        It is the most reliable and will usually be the most efficient.

        If i = 0 or 2, ``nlp1_sparse_solve`` will `estimate` the
        unspecified elements of the objective gradient, using finite
        differences.
        This may simplify the coding of `objfun`.
        However, the computation of finite difference approximations
        usually increases the total run-time substantially (since a call
        to `objfun` is required for each unspecified element) and there
        is less assurance that an acceptable solution will be found.

        If i = 0 or 1, ``nlp1_sparse_solve`` will approximate
        unspecified elements of the constraint Jacobian.
        For each column of the Jacobian, one call to `confun` is needed
        to estimate all unspecified elements in that column (if any).
        For example, if the sparsity pattern of the Jacobian has the
        form

            [table omitted]

        where '*' indicates an element provided and '?' indicates an
        unspecified element, ``nlp1_sparse_solve`` will call `confun`
        twice: once to estimate the missing element in column 2 and
        again to estimate the two missing elements in column 3. (Since
        columns 1 and 4 are known, they require no calls to `confun`.)

        At times, central differences are used rather than forward
        differences, in which case twice as many calls to `objfun` and
        `confun` are needed. (The switch to central differences is not
        under your control.)

        If i < 0 or i > 3, the default value is used.

    'Derivative Linesearch' : valueless
        Default

        At each major iteration, a linesearch is used to improve the
        value of the Lagrangian merit function [equation].
        The default linesearch uses safeguarded cubic interpolation and
        requires both function and gradient values in order to compute
        estimates of the step alpha_k.
        If some analytic derivatives are not provided or option
        'Nonderivative Linesearch' is specified, a linesearch based upon
        safeguarded quadratic interpolation (which does not require the
        evaluation or approximation of any gradients) is used instead.

        A nonderivative linesearch can be slightly less robust on
        difficult problems and it is recommended that the default be
        used if the functions and their derivatives can be computed at
        approximately the same cost.
        If the gradients are very expensive to compute relative to the
        functions however, a nonderivative linesearch may result in a
        significant decrease in the total run-time.

        If option 'Nonderivative Linesearch' is selected,
        ``nlp1_sparse_solve`` signals the evaluation of the linesearch
        by calling `objfun` and `confun` with `mode` = 0.
        Once the linesearch is complete, the nonlinear functions are
        re-evaluated with `mode` = 2.
        If the potential savings offered by a nonderivative linesearch
        are to be fully realised, it is essential that `objfun` and
        `confun` be coded so that no derivatives are computed when
        `mode` = 0.

    'Nonderivative Linesearch' : valueless
        At each major iteration, a linesearch is used to improve the
        value of the Lagrangian merit function [equation].
        The default linesearch uses safeguarded cubic interpolation and
        requires both function and gradient values in order to compute
        estimates of the step alpha_k.
        If some analytic derivatives are not provided or option
        'Nonderivative Linesearch' is specified, a linesearch based upon
        safeguarded quadratic interpolation (which does not require the
        evaluation or approximation of any gradients) is used instead.

        A nonderivative linesearch can be slightly less robust on
        difficult problems and it is recommended that the default be
        used if the functions and their derivatives can be computed at
        approximately the same cost.
        If the gradients are very expensive to compute relative to the
        functions however, a nonderivative linesearch may result in a
        significant decrease in the total run-time.

        If option 'Nonderivative Linesearch' is selected,
        ``nlp1_sparse_solve`` signals the evaluation of the linesearch
        by calling `objfun` and `confun` with `mode` = 0.
        Once the linesearch is complete, the nonlinear functions are
        re-evaluated with `mode` = 2.
        If the potential savings offered by a nonderivative linesearch
        are to be fully realised, it is essential that `objfun` and
        `confun` be coded so that no derivatives are computed when
        `mode` = 0.

    'Elastic Weight' : float
        Default = 1.0 or 100.0

        The default value of r is 100.0 if there are any nonlinear
        constraints and 1.0 otherwise.

        This option defines the initial weight gamma associated with
        problem [equation].

        At any given major iteration k, elastic mode is entered if the
        QP subproblem is infeasible or the QP dual variables (Lagrange
        multipliers) are larger in magnitude than r*(1+||g(x_k)||_2),
        where g is the objective gradient.
        In either case, the QP subproblem is resolved in elastic mode
        with gamma = r*(1+||g(x_k)||_2).

        Thereafter, gamma is increased (subject to a maximum allowable
        value) at any point that is optimal for problem [equation], but
        not feasible for problem (1).
        After the pth increase, gamma = r*10^p*(1+||g(x_k_1)||_2), where
        x_k_1 is the iterate at which gamma was first needed.

        If r < 0, the default value is used.

    'Expand Frequency' : int
        Default = 10000

        This option is part of the EXPAND anti-cycling procedure due to
        Gill `et al.` (1989), which is designed to make progress even on
        highly degenerate problems.

        For linear models, the strategy is to force a positive step at
        every iteration, at the expense of violating the constraints by
        a small amount.
        Suppose that the value of option 'Minor Feasibility Tolerance'
        is delta.
        Over a period of i iterations, the feasibility tolerance
        actually used by ``nlp1_sparse_solve`` (i.e., the `working`
        feasibility tolerance) increases from 0.5 delta to delta (in
        steps of 0.5 delta/i).

        For nonlinear models, the same procedure is used for iterations
        in which there is only one superbasic variable. (Cycling can
        only occur when the current solution is at a vertex of the
        feasible region.) Thus, zero steps are allowed if there is more
        than one superbasic variable, but otherwise positive steps are
        enforced.

        Increasing the value of i helps reduce the number of slightly
        infeasible nonbasic basic variables (most of which are
        eliminated during the resetting procedure).
        However, it also diminishes the freedom to choose a large pivot
        element (see option 'Pivot Tolerance').

        If i < 0, the default value is used.
        If i = 0, the value i = 99999999 is used and effectively no
        anti-cycling procedure is invoked.

    'Factorization Frequency' : int
        Default = 50 or 100

        The default value of i is 50 if there are any nonlinear
        constraints and 100 otherwise.

        If i > 0, at most i basis changes will occur between
        factorizations of the basis matrix.

        For linear problems, the basis factors are usually updated at
        every iteration.
        The default value i = 100 is reasonable for typical problems,
        particularly those that are extremely sparse and well-scaled.

        When the objective function is nonlinear, fewer basis updates
        will occur as the solution is approached.
        The number of iterations between basis factorizations will
        therefore increase.
        During these iterations a test is made regularly according to
        the value of option 'Check Frequency' to ensure that the general
        constraints are satisfied.
        If necessary, the basis will be refactorized before the limit of
        i updates is reached.

        If i <= 0, the default value is used.

    'Infeasible Exit' : valueless
        Default

        Note that this option is ignored if the value of option 'Major
        Iteration Limit' is exceeded, or the linear constraints are
        infeasible.

        If termination is about to occur at a point that does not
        satisfy the nonlinear constraints and option 'Feasible Exit' is
        selected, this option requests that additional iterations be
        performed in order to find a feasible point (if any) for the
        nonlinear constraints.
        This involves solving a feasible point problem in which the
        objective function is omitted.

        Otherwise, this option requests no additional iterations be
        performed.

    'Feasible Exit' : valueless
        Note that this option is ignored if the value of option 'Major
        Iteration Limit' is exceeded, or the linear constraints are
        infeasible.

        If termination is about to occur at a point that does not
        satisfy the nonlinear constraints and option 'Feasible Exit' is
        selected, this option requests that additional iterations be
        performed in order to find a feasible point (if any) for the
        nonlinear constraints.
        This involves solving a feasible point problem in which the
        objective function is omitted.

        Otherwise, this option requests no additional iterations be
        performed.

    'Minimize' : valueless
        Default

        If option 'Feasible Point' is selected, this option attempts to
        find a feasible point (if any) for the nonlinear constraints by
        omitting the objective function.
        It can also be used to check whether the nonlinear constraints
        are feasible.

        Otherwise, this option specifies the required direction of the
        optimization.
        It applies to both linear and nonlinear terms (if any) in the
        objective function.
        Note that if two problems are the same except that one minimizes
        f(x) and the other maximizes -f(x), their solutions will be the
        same but the signs of the dual variables pi_i and the reduced
        gradients d_j will be reversed.

    'Maximize' : valueless
        If option 'Feasible Point' is selected, this option attempts to
        find a feasible point (if any) for the nonlinear constraints by
        omitting the objective function.
        It can also be used to check whether the nonlinear constraints
        are feasible.

        Otherwise, this option specifies the required direction of the
        optimization.
        It applies to both linear and nonlinear terms (if any) in the
        objective function.
        Note that if two problems are the same except that one minimizes
        f(x) and the other maximizes -f(x), their solutions will be the
        same but the signs of the dual variables pi_i and the reduced
        gradients d_j will be reversed.

    'Feasible Point' : valueless
        If option 'Feasible Point' is selected, this option attempts to
        find a feasible point (if any) for the nonlinear constraints by
        omitting the objective function.
        It can also be used to check whether the nonlinear constraints
        are feasible.

        Otherwise, this option specifies the required direction of the
        optimization.
        It applies to both linear and nonlinear terms (if any) in the
        objective function.
        Note that if two problems are the same except that one minimizes
        f(x) and the other maximizes -f(x), their solutions will be the
        same but the signs of the dual variables pi_i and the reduced
        gradients d_j will be reversed.

    'Forward Difference Interval' : float
        Default = sqrt('Function Precision')

        This option defines an interval used to estimate derivatives by
        forward differences in the following circumstances:

        (a) For verifying the objective and/or constraint gradients (see
            the description of the option 'Verify Level').

        (#) For estimating unspecified elements of the objective
            gradient and/or the constraint Jacobian.

        A derivative with respect to x_j is estimated by perturbing that
        element of x to the value x_j+r(1+|x_j|) and then evaluating
        f(x) and/or F(x) (as appropriate) at the perturbed point.
        The resulting gradient estimates should be accurate to O(r),
        unless the functions are badly scaled.
        Judicious alteration of r may sometimes lead to greater
        accuracy.
        See Gill `et al.` (1981) for a discussion of the accuracy in
        finite difference approximations.

        If r <= 0, the default value is used.

    'Function Precision' : float
        Default = epsilon^0.8

        This argument defines the `relative function precision`
        epsilon_r, which is intended to be a measure of the relative
        accuracy with which the nonlinear functions can be computed.
        For example, if f(x) (or F_i(x)) is computed as 1000.56789 for
        some relevant x and the first 6 significant digits are known to
        be correct, then the appropriate value for epsilon_r would be
        10^(-6).

        Ideally the functions f(x) or F_i(x) should have magnitude of
        order 1.
        If all functions are substantially `less` than 1 in magnitude,
        epsilon_r should be the `absolute` precision.
        For example, if f(x) (or F_i(x)) is computed as
        1.23456789*10^(-4) for some relevant x and the first 6
        significant digits are known to be correct, then the appropriate
        value for epsilon_r would be 10^(-10).

        The choice of epsilon_r can be quite complicated for badly
        scaled problems; see Module 8 of Gill `et al.` (1981) for a
        discussion of scaling techniques.
        The default value is appropriate for most simple functions that
        are computed with full accuracy.

        In some cases the function values will be the result of
        extensive computation, possibly involving an iterative procedure
        that can provide few digits of precision at reasonable cost.
        Specifying an appropriate value of r may therefore lead to
        savings, by allowing the linesearch procedure to terminate when
        the difference between function values along the search
        direction becomes as small as the absolute error in the values.

        If r < epsilon or r >= 1, the default value is used.

    'Hessian Frequency' : int
        Default = 99999999

        This option forces the approximate Hessian formed from i BFGS
        updates to be reset to the identity matrix upon completion of a
        major iteration.
        It is intended to be used in conjunction with option 'Hessian
        Full Memory'.

        If i <= 0, the default value is used and effectively no resets
        occur.

    'Hessian Full Memory' : valueless
        Default when n_< 75

        These options specify the method for storing and updating the
        quasi-Newton approximation to the Hessian of the Lagrangian
        function.

        If 'Hessian Full Memory' is specified, the approximate Hessian
        is treated as a dense matrix and BFGS quasi-Newton updates are
        applied explicitly.
        This is most efficient when the total number of nonlinear
        variables is not too large (say, n_< 75).
        In this case, the storage requirement is fixed and you can
        expect 1-step Q-superlinear convergence to the solution.

        'Hessian Limited Memory' should only be specified when n_ is
        very large.
        In this case a limited memory procedure is used to update a
        diagonal Hessian approximation H_r a limited number of times.
        (Updates are accumulated as a list of vector pairs.
        They are discarded at regular intervals after H_r has been reset
        to their diagonal.)

        Note that if 'Hessian Frequency' = 20 is used in conjunction
        with 'Hessian Full Memory', the effect will be similar to using
        'Hessian Limited Memory' in conjunction with 'Hessian Updates' =
        20, except that the latter will retain the current diagonal
        during resets.

    'Hessian Limited Memory' : valueless
        Default when n_>= 75

        These options specify the method for storing and updating the
        quasi-Newton approximation to the Hessian of the Lagrangian
        function.

        If 'Hessian Full Memory' is specified, the approximate Hessian
        is treated as a dense matrix and BFGS quasi-Newton updates are
        applied explicitly.
        This is most efficient when the total number of nonlinear
        variables is not too large (say, n_< 75).
        In this case, the storage requirement is fixed and you can
        expect 1-step Q-superlinear convergence to the solution.

        'Hessian Limited Memory' should only be specified when n_ is
        very large.
        In this case a limited memory procedure is used to update a
        diagonal Hessian approximation H_r a limited number of times.
        (Updates are accumulated as a list of vector pairs.
        They are discarded at regular intervals after H_r has been reset
        to their diagonal.)

        Note that if 'Hessian Frequency' = 20 is used in conjunction
        with 'Hessian Full Memory', the effect will be similar to using
        'Hessian Limited Memory' in conjunction with 'Hessian Updates' =
        20, except that the latter will retain the current diagonal
        during resets.

    'Hessian Updates' : int
        Default = 20 or 99999999

        The default value of i is 20 when 'Hessian Limited Memory' is in
        effect and 99999999 when 'Hessian Full Memory' is in effect, in
        which case no updates are performed.

        If 'Hessian Limited Memory' is in effect, this option defines
        the maximum number of pairs of Hessian update vectors that are
        to be used to define the quasi-Newton approximate Hessian.
        Once the limit of i updates is reached, all but the diagonal
        elements of the accumulated updates are discarded and the
        process starts again.
        Broadly speaking, the more updates that are stored, the better
        the quality of the approximate Hessian.
        On the other hand, the more vectors that are stored, the greater
        the cost of each QP iteration.

        The default value of i is likely to give a robust algorithm
        without significant expense, but faster convergence may be
        obtained with far fewer updates (e.g., i = 5).

        If i < 0, the default value is used.

    'Infinite Bound Size' : float
        Default = 10^20

        If r > 0, r defines the 'infinite' bound bigbnd in the
        definition of the problem constraints.
        Any upper bound greater than or equal to bigbnd will be regarded
        as + infinity (and similarly any lower bound less than or equal
        to -bigbnd will be regarded as - infinity).

        If r <= 0, the default value is used.

    'Iteration Limit' : int
        Default = 10000

        The value of i specifies the maximum number of minor iterations
        allowed (i.e., iterations of the simplex method or the QP
        algorithm), summed over all major iterations. (See also the
        description of the options 'Major Iteration Limit' and 'Minor
        Iteration Limit'.)

        If i < 0, the default value is used.

    'Linesearch Tolerance' : float
        Default = 0.9

        This option controls the accuracy with which a step length will
        be located along the direction of search at each iteration.
        At the start of each linesearch a target directional derivative
        for the Lagrangian merit function is identified.
        The value of r therefore determines the accuracy to which this
        target value is approximated.

        The default value r = 0.9 requests an inaccurate search and is
        appropriate for most problems, particularly those with any
        nonlinear constraints.

        If the nonlinear functions are cheap to evaluate, a more
        accurate search may be appropriate; try r = 0.1,0.01 or 0.001.
        The number of major iterations required to solve the problem
        might decrease.

        If the nonlinear functions are expensive to evaluate, a less
        accurate search may be appropriate.
        If 'Derivative Level' = 3, try r = 0.99. (The number of major
        iterations required to solve the problem might increase, but the
        total number of function evaluations may decrease enough to
        compensate.)

        If 'Derivative Level' < 3, a moderately accurate search may be
        appropriate; try r = 0.5.
        Each search will (typically) require only 1-5 function values,
        but many function calls will then be needed to estimate the
        missing gradients for the next iteration.

        If r < 0 or r >= 1, the default value is used.

    'List' : valueless
        Option 'List' enables printing of each option specification as
        it is supplied. 'Nolist' suppresses this printing.

    'Nolist' : valueless
        Default for ``nlp1_sparse_solve`` = 'Nolist'

        Option 'List' enables printing of each option specification as
        it is supplied. 'Nolist' suppresses this printing.

    'LU Density Tolerance' : float
        Default = 0.6

        If r_1 > 0, r_1 defines the density tolerance used during the LU
        factorization of the basis matrix.
        Columns of L and rows of U are formed one at a time and the
        remaining rows and columns of the basis are altered
        appropriately.
        At any stage, if the density of the remaining matrix exceeds
        r_1, the Markowitz strategy for choosing pivots is terminated.
        The remaining matrix is then factorized using a dense LU
        procedure.
        Increasing the value of r_1 towards unity may give slightly
        sparser LU factors, with a slight increase in factorization
        time.
        If r_1 <= 0, the default value is used.

        If r_2 > 0, r_2 defines the singularity tolerance used to guard
        against ill-conditioned basis matrices.
        Whenever the basis is refactorized, the diagonal elements of U
        are tested as follows.
        If |u_{jj}| <= r_2 or |u_{jj}| < r_2*max_i|u_{ij}|, the jth
        column of the basis is replaced by the corresponding slack
        variable.
        This is most likely to occur when `start` = 'W' (see
        [Parameters]), or at the start of a major iteration.
        If r_2 <= 0, the default value is used.

        In some cases, the Jacobian matrix may converge to values that
        make the basis exactly singular (e.g., a whole row of the
        Jacobian matrix could be zero at an optimal solution).
        Before exact singularity occurs, the basis could become very
        ill-conditioned and the optimization could progress very slowly
        (if at all).
        Setting r_2 = 0.00001 (say) may therefore help cause a judicious
        change of basis in such situations.

    'LU Singularity Tolerance' : float
        Default = epsilon^0.67

        If r_1 > 0, r_1 defines the density tolerance used during the LU
        factorization of the basis matrix.
        Columns of L and rows of U are formed one at a time and the
        remaining rows and columns of the basis are altered
        appropriately.
        At any stage, if the density of the remaining matrix exceeds
        r_1, the Markowitz strategy for choosing pivots is terminated.
        The remaining matrix is then factorized using a dense LU
        procedure.
        Increasing the value of r_1 towards unity may give slightly
        sparser LU factors, with a slight increase in factorization
        time.
        If r_1 <= 0, the default value is used.

        If r_2 > 0, r_2 defines the singularity tolerance used to guard
        against ill-conditioned basis matrices.
        Whenever the basis is refactorized, the diagonal elements of U
        are tested as follows.
        If |u_{jj}| <= r_2 or |u_{jj}| < r_2*max_i|u_{ij}|, the jth
        column of the basis is replaced by the corresponding slack
        variable.
        This is most likely to occur when `start` = 'W' (see
        [Parameters]), or at the start of a major iteration.
        If r_2 <= 0, the default value is used.

        In some cases, the Jacobian matrix may converge to values that
        make the basis exactly singular (e.g., a whole row of the
        Jacobian matrix could be zero at an optimal solution).
        Before exact singularity occurs, the basis could become very
        ill-conditioned and the optimization could progress very slowly
        (if at all).
        Setting r_2 = 0.00001 (say) may therefore help cause a judicious
        change of basis in such situations.

    'LU Factor Tolerance' : float
        Default = 5.0 or 100.0

        The default value of r_1 is 5.0 if there are any nonlinear
        constraints and 100.0 otherwise.
        The default value of r_2 is 5.0 if there are any nonlinear
        constraints and 10.0 otherwise.

        If r_1 >= 1 and r_2 >= 1, the values of r_1 and r_2 affect the
        stability and sparsity of the basis factorization B = LU, during
        refactorization and updating, respectively.
        The lower triangular matrix L is a product of matrices of the
        form

            [table omitted],

        where the multipliers mu satisfy |mu| <= r_i.
        Smaller values of r_i favour stability, while larger values
        favour sparsity.
        The default values of r_1 and r_2 usually strike a good
        compromise.
        For large and relatively dense problems, setting r_1 = 10.0 or
        5.0 (say) may give a marked improvement in sparsity without
        impairing stability to a serious degree.
        Note that for problems involving band matrices, it may be
        necessary to reduce r_1 and/or r_2 in order to achieve
        stability.

        If r_1 < 1 or r_2 < 1, the appropriate default value is used.

    'LU Update Tolerance' : float
        Default = 5.0 or 10.0

        The default value of r_1 is 5.0 if there are any nonlinear
        constraints and 100.0 otherwise.
        The default value of r_2 is 5.0 if there are any nonlinear
        constraints and 10.0 otherwise.

        If r_1 >= 1 and r_2 >= 1, the values of r_1 and r_2 affect the
        stability and sparsity of the basis factorization B = LU, during
        refactorization and updating, respectively.
        The lower triangular matrix L is a product of matrices of the
        form

            [table omitted],

        where the multipliers mu satisfy |mu| <= r_i.
        Smaller values of r_i favour stability, while larger values
        favour sparsity.
        The default values of r_1 and r_2 usually strike a good
        compromise.
        For large and relatively dense problems, setting r_1 = 10.0 or
        5.0 (say) may give a marked improvement in sparsity without
        impairing stability to a serious degree.
        Note that for problems involving band matrices, it may be
        necessary to reduce r_1 and/or r_2 in order to achieve
        stability.

        If r_1 < 1 or r_2 < 1, the appropriate default value is used.

    'Major Feasibility Tolerance' : float
        Default = sqrt(epsilon)

        This option specifies how accurately the nonlinear constraints
        should be satisfied.
        The default value is appropriate when the linear and nonlinear
        constraints contain data to approximately that accuracy.
        A larger value may be appropriate if some of the problem
        functions are known to be of low accuracy.

        Let `rowerr` be defined as the maximum nonlinear constraint
        violation normalized by the size of the solution.
        It is required to satisfy

            rowerr = max_i(viol_i)/||(x,s)|| <= r,

        where viol_i is the violation of the ith nonlinear constraint.

        If r <= epsilon, the default value is used.

    'Major Iteration Limit' : int
        Default = 1000

        The value of i specifies the maximum number of major iterations
        allowed before termination.
        It is intended to guard against an excessive number of
        linearizations of the nonlinear constraints.
        Setting i = 0 and 'Major Print Level' > 0 means that the
        objective and constraint gradients will be checked if 'Verify
        Level' > 0 and the workspace needed to start solving the problem
        will be computed and printed, but no iterations will be
        performed.

        If i < 0, the default value is used.

    'Major Optimality Tolerance' : float
        Default = sqrt(epsilon)

        This option specifies the final accuracy of the dual variables.
        If ``nlp1_sparse_solve`` terminates with no exception or warning
        is raised, a primal and dual solution (x,s,pi) will have been
        computed such that

            maxgap = max_j(gap_j)/(||pi||) <= r,

        where gap_j is an estimate of the complementarity gap for the
        jth variable and ||pi|| is a measure of the size of the QP dual
        variables (or Lagrange multipliers) given by

            ||pi|| = max(sigma/(sqrt(m)), 1), where sigma = sum_i =
            1^m|pi_i|.

        It is included to make the tests independent of a scale factor
        on the objective function.
        Specifically, gap_j is computed from the final QP solution using
        the reduced gradients d_j = g_j-pi^Ta_j, where g_j is the jth
        element of the objective gradient and a_j is the associated
        column of the constraint matrix (A -I):

            gap_j = {[table omitted]

        If r <= 0, the default value is used.

    'Optimality Tolerance' : float
        Default = sqrt(epsilon)

        This option specifies the final accuracy of the dual variables.
        If ``nlp1_sparse_solve`` terminates with no exception or warning
        is raised, a primal and dual solution (x,s,pi) will have been
        computed such that

            maxgap = max_j(gap_j)/(||pi||) <= r,

        where gap_j is an estimate of the complementarity gap for the
        jth variable and ||pi|| is a measure of the size of the QP dual
        variables (or Lagrange multipliers) given by

            ||pi|| = max(sigma/(sqrt(m)), 1), where sigma = sum_i =
            1^m|pi_i|.

        It is included to make the tests independent of a scale factor
        on the objective function.
        Specifically, gap_j is computed from the final QP solution using
        the reduced gradients d_j = g_j-pi^Ta_j, where g_j is the jth
        element of the objective gradient and a_j is the associated
        column of the constraint matrix (A -I):

            gap_j = {[table omitted]

        If r <= 0, the default value is used.

    'Major Print Level' : int
        The value of i controls the amount of printout produced by the
        major iterations of ``nlp1_sparse_solve``, as indicated below.
        A detailed description of the printed output is given in [Major
        Iteration Printout] (summary output at each major iteration and
        the final solution) and [Monitoring Information] (monitoring
        information at each major iteration). (See also the description
        of the option 'Minor Print Level'.)

        The following printout is sent to the file object associated
        with the advisory I/O unit (see
        :class:`~naginterfaces.base.utils.FileObjManager`):

        [table omitted]

        The following printout is sent to the unit number given by the
        option 'Monitoring File':

        [table omitted]

        If 'Major Print Level' >= 5 and the unit number defined by the
        option 'Monitoring File' is the advisory unit number, the
        summary output for each major iteration is suppressed.

    'Print Level' : int
        The value of i controls the amount of printout produced by the
        major iterations of ``nlp1_sparse_solve``, as indicated below.
        A detailed description of the printed output is given in [Major
        Iteration Printout] (summary output at each major iteration and
        the final solution) and [Monitoring Information] (monitoring
        information at each major iteration). (See also the description
        of the option 'Minor Print Level'.)

        The following printout is sent to the file object associated
        with the advisory I/O unit (see
        :class:`~naginterfaces.base.utils.FileObjManager`):

        [table omitted]

        The following printout is sent to the unit number given by the
        option 'Monitoring File':

        [table omitted]

        If 'Major Print Level' >= 5 and the unit number defined by the
        option 'Monitoring File' is the advisory unit number, the
        summary output for each major iteration is suppressed.

    'Major Step Limit' : float
        Default = 2.0

        If r > 0,r limits the change in x during a linesearch.
        It applies to all nonlinear problems once a 'feasible solution'
        or 'feasible subproblem' has been found.

        A linesearch determines a step alpha in the interval 0 < alpha
        <= beta, where beta = 1 if there are any nonlinear constraints,
        or the step to the nearest upper or lower bound on x if all the
        constraints are linear.
        Normally, the first step attempted is alpha_1 = min(1, beta).

        In some cases, such as f(x) = ae^(bx) or f(x) = ax^b, even a
        moderate change in the elements of x can lead to floating-point
        overflow.
        The argument r is therefore used to define a step limit beta_
        given by

            beta_= r(1+||x||_2)/||p||_2,

        where p is the search direction and the first evaluation of f(x)
        is made at the (potentially) smaller step length alpha_1 =
        min(1, beta_, beta).

        Wherever possible, upper and lower bounds on x should be used to
        prevent evaluation of nonlinear functions at meaningless points.
        The default value r = 2.0 should not affect progress on
        well-behaved functions, but values such as r = 0.1 or 0.01 may
        be helpful when rapidly varying functions are present.
        If a small value of r is selected, a 'good' starting point may
        be required.
        An important application is to the class of nonlinear least
        squares problems.

        If r <= 0, the default value is used.

    'Minor Feasibility Tolerance' : float
        Default = sqrt(epsilon)

        This option attempts to ensure that all variables eventually
        satisfy their upper and lower bounds to within the tolerance r.
        Since this includes slack variables, general linear constraints
        should also be satisfied to within r.
        Note that feasibility with respect to nonlinear constraints is
        judged by the value of option 'Major Feasibility Tolerance' and
        not by r.

        If the bounds and linear constraints cannot be satisfied to
        within r, the problem is declared `infeasible`.
        Let Sinf be the corresponding sum of infeasibilities.
        If Sinf is quite small, it may be appropriate to raise r by a
        factor of 10 or 100.
        Otherwise, some error in the data should be suspected.

        If 'Scale Option' >= 1, feasibility is defined in terms of the
        `scaled` problem (since it is more likely to be meaningful).

        Nonlinear functions will only be evaluated at points that
        satisfy the bounds and linear constraints.
        If there are regions where a function is undefined, every effort
        should be made to eliminate these regions from the problem.
        For example, if f(x_1,x_2) = sqrt(x_1)+log(x_2), it is essential
        to place lower bounds on both x_1 and x_2.
        If the value r = 10^(-6) is used, the bounds x_1 >= 10^(-5) and
        x_2 >= 10^(-4) might be appropriate. (The log singularity is
        more serious; in general, you should attempt to keep x as far
        away from singularities as possible.)

        In reality, r is used as a feasibility tolerance for satisfying
        the bounds on x and s in each QP subproblem.
        If the sum of infeasibilities cannot be reduced to zero, the QP
        subproblem is declared infeasible and the function is, then in
        `elastic mode` thereafter (with only the linearized nonlinear
        constraints defined to be elastic). (See also the description of
        'Elastic Weight'.)

        If r <= epsilon, the default value is used.

    'Feasibility Tolerance' : float
        Default = sqrt(epsilon)

        This option attempts to ensure that all variables eventually
        satisfy their upper and lower bounds to within the tolerance r.
        Since this includes slack variables, general linear constraints
        should also be satisfied to within r.
        Note that feasibility with respect to nonlinear constraints is
        judged by the value of option 'Major Feasibility Tolerance' and
        not by r.

        If the bounds and linear constraints cannot be satisfied to
        within r, the problem is declared `infeasible`.
        Let Sinf be the corresponding sum of infeasibilities.
        If Sinf is quite small, it may be appropriate to raise r by a
        factor of 10 or 100.
        Otherwise, some error in the data should be suspected.

        If 'Scale Option' >= 1, feasibility is defined in terms of the
        `scaled` problem (since it is more likely to be meaningful).

        Nonlinear functions will only be evaluated at points that
        satisfy the bounds and linear constraints.
        If there are regions where a function is undefined, every effort
        should be made to eliminate these regions from the problem.
        For example, if f(x_1,x_2) = sqrt(x_1)+log(x_2), it is essential
        to place lower bounds on both x_1 and x_2.
        If the value r = 10^(-6) is used, the bounds x_1 >= 10^(-5) and
        x_2 >= 10^(-4) might be appropriate. (The log singularity is
        more serious; in general, you should attempt to keep x as far
        away from singularities as possible.)

        In reality, r is used as a feasibility tolerance for satisfying
        the bounds on x and s in each QP subproblem.
        If the sum of infeasibilities cannot be reduced to zero, the QP
        subproblem is declared infeasible and the function is, then in
        `elastic mode` thereafter (with only the linearized nonlinear
        constraints defined to be elastic). (See also the description of
        'Elastic Weight'.)

        If r <= epsilon, the default value is used.

    'Minor Iteration Limit' : int
        Default = 500

        The value of i specifies the maximum number of iterations
        allowed between successive linearizations of the nonlinear
        constraints.
        A value in the range 10 <= i <= 50 prevents excessive effort
        being expended on early major iterations, but allows later QP
        subproblems to be solved to completion.
        Note that an extra m minor iterations are allowed if the first
        QP subproblem to be solved starts with the all-slack basis B =
        I. (See the description of the option 'Crash Option'.)

        In general, it is unsafe to specify values as small as i = 1 or
        2 (because even when an optimal solution has been reached, a few
        minor iterations may be needed for the corresponding QP
        subproblem to be recognized as optimal).

        If i <= 0, the default value is used.

    'Minor Optimality Tolerance' : float
        Default = sqrt(epsilon)

        This option is used to judge optimality for each QP subproblem.
        Let the QP reduced gradients be d_j = g_j-pi^Ta_j, where g_j is
        the jth element of the QP gradient, a_j is the associated column
        of the QP constraint matrix and pi is the set of QP dual
        variables.

        By construction, the reduced gradients for basic variables are
        always zero.
        The QP subproblem will be declared optimal if the reduced
        gradients for nonbasic variables at their upper or lower bounds
        satisfy

            (d_j)/(||pi||) >= -r or (d_j)/(||pi||) <= r

        respectively, and if |d_j|/(||pi||) <= r for superbasic
        variables.

        Note that ||pi|| is a measure of the size of the dual variables.
        It is included to make the tests independent of a scale factor
        on the objective function. (The value of ||pi|| actually used is
        defined in the description for option 'Major Optimality
        Tolerance'.)

        If the objective is scaled down to be very `small`, the
        optimality test reduces to comparing d_j against r.

        If r <= 0, the default value is used.

    'Minor Print Level' : int
        Default = 0

        The value of i controls the amount of printout produced by the
        minor iterations of ``nlp1_sparse_solve`` (i.e., the iterations
        of the quadratic programming algorithm), as indicated below.
        A detailed description of the printed output is given in [Minor
        Iteration Printout] (summary output at each minor iteration) and
        [Monitoring Information] (monitoring information at each minor
        iteration). (See also the description of the option 'Major Print
        Level'.)

        The following printout is sent to the file object associated
        with the advisory I/O unit (see
        :class:`~naginterfaces.base.utils.FileObjManager`):

        [table omitted]

        The following printout is sent to the unit number given by the
        option 'Monitoring File':

        [table omitted]

        If 'Major Print Level' >= 5 and the unit number defined by the
        option 'Monitoring File' is the advisory unit number, the
        summary output for each major iteration is suppressed.

    'Monitoring File' : int
        Default = -1

        If i >= 0 and 'Major Print Level' >= 5 or i >= 0 and 'Minor
        Print Level' >= 1, then monitoring information is produced by
        ``nlp1_sparse_solve`` at every iteration is sent to a file with
        logical unit number i.
        If i < 0 and/or 'Major Print Level' < 5 and 'Minor Print Level'
        < 1, then no monitoring information is produced.

    'Partial Price' : int
        Default = 1 or 10

        The default value of i is 1 if there are any nonlinear
        constraints and 10 otherwise.

        This option is recommended for large problems that have
        significantly more variables than constraints (i.e., n≫m).
        It reduces the work required for each 'pricing' operation (i.e.,
        when a nonbasic variable is selected to become superbasic).
        The possible choices for i are the following.

        [table omitted]

        If i <= 0, the default value is used.

    'Pivot Tolerance' : float
        Default = epsilon^0.67

        If r > 0, r is used during the solution of QP subproblems to
        prevent columns entering the basis if they would cause the basis
        to become almost singular.

        When x changes to x+alpha p for some specified search direction
        p, a 'ratio test' is used to determine which element of x
        reaches an upper or lower bound first.
        The corresponding element of p is called the `pivot element`.
        Elements of p are ignored (and therefore cannot be pivot
        elements) if they are smaller than r.

        It is common in practice for two (or more) variables to reach a
        bound at essentially the same time.
        In such cases, the 'Minor Feasibility Tolerance' provides some
        freedom to maximize the pivot element and thereby improve
        numerical stability.
        Excessively `small` values of 'Minor Feasibility Tolerance'
        should therefore not be specified.
        To a lesser extent, the 'Expand Frequency' also provides some
        freedom to maximize the pivot element.
        Excessively `large` values of 'Expand Frequency' should
        therefore not be specified.

        If r <= 0, the default value is used.

    'Scale Option' : int
        Default = 1 or 2

        The default value of i is 1 if there are any nonlinear
        constraints and 2 otherwise.

        This option enables you to scale the variables and constraints
        using an iterative procedure due to Fourer (1982), which
        attempts to compute row scales r_i and column scales c_j such
        that the scaled matrix coefficients a__{ij} = a_{ij}*(c_j/r_i)
        are as close as possible to unity. (The lower and upper bounds
        on the variables and slacks for the scaled problem are redefined
        as l__j = l_j/c_j and u__j = u_j/c_j respectively, where c_j =
        r_{j-n} if j > n.) The possible choices for i are the following.

        [table omitted]

        If there are any nonlinear constraints present, the scale
        factors depend on the Jacobian at the first point that satisfies
        the linear constraints and the upper and lower bounds.
        The setting i = 2 should therefore be used only if a 'good'
        starting point is available and the problem is not highly
        nonlinear.

        If i < 0 or i > 2, the default value is used.

    'Scale Tolerance' : float
        Default = 0.9

        Note that this option does not apply when 'Scale Option' = 0.

        The value r (0 < r < 1) is used to control the number of scaling
        passes to be made through the constraint matrix A.
        At least 3 (and at most 10) passes will be made.
        More precisely, let a_p denote the largest column ratio (i.e.,
        'biggest' element/'smallest' element in some sense) after the
        pth scaling pass through A.
        The scaling procedure is terminated if a_p >= a_{p-1}*r for some
        p >= 3.
        Thus, increasing the value of r from 0.9 to 0.99 (say) will
        probably increase the number of passes through A.

        If r <= 0 or r >= 1, the default value is used.

    'Start Objective Check At Column' : int
        Default = 1

        These keywords take effect only if 'Verify Level' > 0.
        They may be used to control the verification of gradient
        elements computed by `objfun` and/or Jacobian elements computed
        by `confun`.
        For example, if the first 30 elements of the objective gradient
        appeared to be correct in an earlier run, so that only element
        31 remains questionable, then it is reasonable to specify 'Start
        Objective Check At Column' = 31.
        Similarly for columns of the Jacobian.
        If the first 30 variables occur nonlinearly in the constraints
        but the remaining variables are nonlinear only in the objective,
        then `objfun` must set the first 30 elements of the array
        `objgrd` to zero, but these hardly need to be verified.
        Again it is reasonable to specify 'Start Objective Check At
        Column' = 31.

        If i_2 <= 0 or i_2 > n'_1, the default value is used.

        If i_1 <= 0 or i_1 > min(n'_1, i_2), the default value is used.

        If i_4 <= 0 or i_4 > n_1^'', the default value is used.

        If i_3 <= 0 or i_3 > min(n_1^'', i_4), the default value is
        used.

    'Stop Objective Check At Column' : int
        Default = n'_1

        These keywords take effect only if 'Verify Level' > 0.
        They may be used to control the verification of gradient
        elements computed by `objfun` and/or Jacobian elements computed
        by `confun`.
        For example, if the first 30 elements of the objective gradient
        appeared to be correct in an earlier run, so that only element
        31 remains questionable, then it is reasonable to specify 'Start
        Objective Check At Column' = 31.
        Similarly for columns of the Jacobian.
        If the first 30 variables occur nonlinearly in the constraints
        but the remaining variables are nonlinear only in the objective,
        then `objfun` must set the first 30 elements of the array
        `objgrd` to zero, but these hardly need to be verified.
        Again it is reasonable to specify 'Start Objective Check At
        Column' = 31.

        If i_2 <= 0 or i_2 > n'_1, the default value is used.

        If i_1 <= 0 or i_1 > min(n'_1, i_2), the default value is used.

        If i_4 <= 0 or i_4 > n_1^'', the default value is used.

        If i_3 <= 0 or i_3 > min(n_1^'', i_4), the default value is
        used.

    'Start Constraint Check At Column' : int
        Default = 1

        These keywords take effect only if 'Verify Level' > 0.
        They may be used to control the verification of gradient
        elements computed by `objfun` and/or Jacobian elements computed
        by `confun`.
        For example, if the first 30 elements of the objective gradient
        appeared to be correct in an earlier run, so that only element
        31 remains questionable, then it is reasonable to specify 'Start
        Objective Check At Column' = 31.
        Similarly for columns of the Jacobian.
        If the first 30 variables occur nonlinearly in the constraints
        but the remaining variables are nonlinear only in the objective,
        then `objfun` must set the first 30 elements of the array
        `objgrd` to zero, but these hardly need to be verified.
        Again it is reasonable to specify 'Start Objective Check At
        Column' = 31.

        If i_2 <= 0 or i_2 > n'_1, the default value is used.

        If i_1 <= 0 or i_1 > min(n'_1, i_2), the default value is used.

        If i_4 <= 0 or i_4 > n_1^'', the default value is used.

        If i_3 <= 0 or i_3 > min(n_1^'', i_4), the default value is
        used.

    'Stop Constraint Check At Column' : int
        Default = n_1^''

        These keywords take effect only if 'Verify Level' > 0.
        They may be used to control the verification of gradient
        elements computed by `objfun` and/or Jacobian elements computed
        by `confun`.
        For example, if the first 30 elements of the objective gradient
        appeared to be correct in an earlier run, so that only element
        31 remains questionable, then it is reasonable to specify 'Start
        Objective Check At Column' = 31.
        Similarly for columns of the Jacobian.
        If the first 30 variables occur nonlinearly in the constraints
        but the remaining variables are nonlinear only in the objective,
        then `objfun` must set the first 30 elements of the array
        `objgrd` to zero, but these hardly need to be verified.
        Again it is reasonable to specify 'Start Objective Check At
        Column' = 31.

        If i_2 <= 0 or i_2 > n'_1, the default value is used.

        If i_1 <= 0 or i_1 > min(n'_1, i_2), the default value is used.

        If i_4 <= 0 or i_4 > n_1^'', the default value is used.

        If i_3 <= 0 or i_3 > min(n_1^'', i_4), the default value is
        used.

    'Superbasics Limit' : int
        Default = min(500, n_+1)

        Note that this option does not apply to linear problems.

        It places a limit on the storage allocated for superbasic
        variables.
        Ideally, the value of i should be set slightly larger than the
        'number of degrees of freedom' expected at the solution.

        For nonlinear problems, the number of degrees of freedom is
        often called the 'number of independent variables'.
        Normally, the value of i need not be greater than n_+1, but for
        many problems it may be considerably smaller. (This will save
        storage if n_ is very large.)

        If i <= 0, the default value is used.

    'Unbounded Objective' : float
        Default = 10^15

        These options are intended to detect unboundedness in nonlinear
        problems.
        During the linesearch, the objective function f is evaluated at
        points of the form x+alpha p, where x and p are fixed and alpha
        varies.
        If |f| exceeds r_1 or alpha exceeds r_2, the iterations are
        terminated and the function returns with `errno` = 3.

        If singularities are present, unboundedness in f(x) may manifest
        itself by a floating-point overflow during the evaluation of
        f(x+alpha p), before the test against r_1 can be made.

        Unboundedness in x is best avoided by placing finite upper and
        lower bounds on the variables.

        If r_1 <= 0 or r_2 <= 0, the appropriate default value is used.

    'Unbounded Step Size' : float
        Default = max(bigbnd,10^20)

        These options are intended to detect unboundedness in nonlinear
        problems.
        During the linesearch, the objective function f is evaluated at
        points of the form x+alpha p, where x and p are fixed and alpha
        varies.
        If |f| exceeds r_1 or alpha exceeds r_2, the iterations are
        terminated and the function returns with `errno` = 3.

        If singularities are present, unboundedness in f(x) may manifest
        itself by a floating-point overflow during the evaluation of
        f(x+alpha p), before the test against r_1 can be made.

        Unboundedness in x is best avoided by placing finite upper and
        lower bounds on the variables.

        If r_1 <= 0 or r_2 <= 0, the appropriate default value is used.

    'Verify Level' : int
        Default = 0

        This option refers to finite difference checks on the gradient
        elements computed by `objfun` and `confun`.
        Gradients are verified at the first point that satisfies the
        linear constraints and the upper and lower bounds.
        Unspecified gradient elements are not checked and hence they
        result in no overhead.
        The possible choices for i are the following.

        [table omitted]

        The value i = 3 should be used whenever a new function function
        is being developed.
        The 'Start Objective Check At Column' and 'Stop Objective Check
        At Column' keywords may be used to limit the number of nonlinear
        variables to be checked.

        If i < -1 or i > 3, the default value is used.

    'Violation Limit' : float
        Default = 10.0

        This option defines an absolute limit on the magnitude of the
        maximum constraint violation after the linesearch.
        Upon completion of the linesearch, the new iterate x_{k+1}
        satisfies the condition

            v_i(x_{k+1}) <= r*max(1, v_i(x_0)),

        where x_0 is the point at which the nonlinear constraints are
        first evaluated and v_i(x) is the ith nonlinear constraint
        violation v_i(x) = max(0, l_i-F_i(x), F_i(x)-u_i).

        The effect of the violation limit is to restrict the iterates to
        lie in an `expanded` feasible region whose size depends on the
        magnitude of r.
        This makes it possible to keep the iterates within a region
        where the objective function is expected to be well-defined and
        bounded below (or above in the case of maximization).
        If the objective function is bounded below (or above in the case
        of maximization) for all values of the variables, then r may be
        any large positive value.

        If r <= 0, the default value is used.

    Raises
    ------
    NagValueError
        (`errno` 7)
            On entry, `start` = *<value>*.

            Constraint: `start` = 'C' or 'W'.

        (`errno` 7)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 7)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 7)
            On entry, nnz = *<value>*, n = *<value>* and `m` =
            *<value>*.

            Constraint: 1 <= nnz <= n*`m`.

        (`errno` 7)
            On entry, `ncnln` = *<value>* and `m` = *<value>*.

            Constraint: 0 <= `ncnln` <= `m`.

        (`errno` 7)
            On entry, `nonln` = *<value>* and n = *<value>*.

            Constraint: 0 <= `nonln` <= n.

        (`errno` 7)
            On entry, `ncnln` = 0 and `njnln` = *<value>*.

            Constraint: if `ncnln` = 0 then `nonln` = 0.

        (`errno` 7)
            On entry, `njnln` = *<value>*, n = *<value>* and `ncnln` =
            *<value>*.

            Constraint: if `ncnln` > 0 then 1 <= `njnln` <= n.

        (`errno` 7)
            On entry, `iobj` = *<value>*.

            Constraint: `iobj` >= -1.

        (`errno` 7)
            On entry, `iobj` = *<value>*, `ncnln` = *<value>* and `m` =
            *<value>*.

            Constraint: `ncnln` < `iobj` <= `m`.

        (`errno` 7)
            On entry, nname = *<value>*, n = *<value>* and `m` =
            *<value>*.

            Constraint: nname = 1 or nname = n+`m`.

        (`errno` 7)
            On entry, `leniz` = *<value>*, n = *<value>* and `m` =
            *<value>*.

            Constraint: `leniz` >= max(500, `m`+n).

        (`errno` 7)
            On entry, `lenz` = *<value>*.

            Constraint: `lenz` >= 500.

        (`errno` 7)
            On entry, `m` = *<value>* and `ha`[*<value>*] = *<value>*.

            Constraint: 1 <= `ha`[j] <= `m` for all j.

        (`errno` 7)
            On entry, `ka`[0] = *<value>*.

            Constraint: `ka`[0] = 1.

        (`errno` 7)
            On entry, n+1 = *<value>*, `ka`[n] = *<value>* and nnz+1 =
            *<value>*.

            Constraint: `ka`[n] = nnz+1.

        (`errno` 7)
            On entry, `ka`[*<value>*] = *<value>*.

            Constraint: `ka`[j] >= 1 for all j.

        (`errno` 7)
            On entry, `m` = *<value>*, `ka`[j] = *<value>* and `ka`[j-1]
            = *<value>*, for j = *<value>*.

            Constraint: 0 <= `ka`[j+1]-`ka`[j] <= `m` for all j.

        (`errno` 7)
            On entry, duplicate element found in row *<value>*, column
            *<value>*.

        (`errno` 7)
            On entry, `start` = 'C' and `istate`[*<value>*] = *<value>*.

            Constraint: if `start` = 'C' then `istate`[i] >= 0 for all
            i.

        (`errno` 7)
            On entry, `start` = 'C' and `istate`[*<value>*] = *<value>*.

            Constraint: if `start` = 'C' then `istate`[i] <= 5 for all
            i.

        (`errno` 7)
            On entry, `start` = 'W' and `istate`[*<value>*] = *<value>*.

            Constraint: if `start` = 'W' then `istate`[i] >= 0 for all
            i.

        (`errno` 7)
            On entry, `start` = 'W' and `istate`[*<value>*] = *<value>*.

            Constraint: if `start` = 'W' then `istate`[i] <= 3 for all
            i.

        (`errno` 7)
            On entry, n = *<value>*, `iobj` = *<value>*,
            `bl`[n+abs(`iobj`)-1] = *<value>* and bigbnd = *<value>*.

            Constraint: `bl`[n+abs(`iobj`)-1] <= -bigbnd.

        (`errno` 7)
            On entry, n = *<value>*, `iobj` = *<value>*,
            `bu`[n+abs(`iobj`)-1] = *<value>* and bigbnd = *<value>*.

            Constraint: `bu`[n+abs(`iobj`)-1] >= bigbnd.

        (`errno` 7)
            On entry, the equal bounds on *<value>* are infinite,
            because `bl`[*<value>*] = beta and `bu`[*<value>*] = beta,
            but abs(beta) >= bigbnd: beta = *<value>* and bigbnd =
            *<value>*.

        (`errno` 7)
            On entry, the bounds on *<value>* are inconsistent:
            `bl`[*<value>*] = *<value>* and `bu`[*<value>*] = *<value>*.

        (`errno` 8)
            Function `objfun` appears to be giving incorrect gradients.

        (`errno` 9)
            Function `confun` appears to be giving incorrect gradients.

        (`errno` 11)
            Numerical error in trying to satisfy the linear constraints.

        (`errno` 12)
            Not enough integer workspace for the basis factors.

        (`errno` 13)
            Not enough real workspace for the basis factors.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` -1)
            Constraint and objective values could not be calculated.

        (`errno` i < 0)
            User requested termination by setting `mode` negative in
            `objfun` or `confun`.

        (`errno` 1)
            Infeasible problem, nonlinear infeasibilities minimized.

        (`errno` 1)
            No feasible point for the nonlinear constraints.

        (`errno` 1)
            No feasible point for the linear constraints.

        (`errno` 2)
            The problem is unbounded (or badly scaled).

        (`errno` 3)
            Violation Limit exceeded. The problem may be unbounded.

        (`errno` 5)
            Feasible solution, but requested accuracy could not be
            achieved.

        (`errno` 10)
            Current point cannot be improved upon.

        (`errno` 14)
            The basis is singular after 15 factorization attempts.

        (`errno` 15)
            Not enough integer workspace to start solving the problem.

        (`errno` 16)
            Not enough real workspace to start solving the problem.

    NagAlgorithmicMajorWarning
        (`errno` 4)
            Major Iteration Limit exceeded.

        (`errno` 4)
            Minor Iteration Limit exceeded.

        (`errno` 4)
            Iteration Limit exceeded.

        (`errno` 6)
            The value of the option 'Superbasics Limit' is too small.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Conn, A R, 1973, `Constrained optimization using a nondifferentiable
    penalty function`, SIAM J. Numer. Anal. (10), 760--779

    Eldersveld, S K, 1991, `Large-scale sequential quadratic programming
    algorithms`, PhD Thesis, Department of Operations Research, Stanford
    University, Stanford

    Fletcher, R, 1984, `An l_1 penalty method for nonlinear
    constraints`, Numerical Optimization 1984, (eds P T Boggs, R H Byrd
    and R B Schnabel), 26--40, SIAM Philadelphia

    Fourer, R, 1982, `Solving staircase linear programs by the simplex
    method`, Math. Programming (23), 274--313

    Gill, P E, Murray, W and Saunders, M A, 2002, `SNOPT: An SQP
    Algorithm for Large-scale Constrained Optimization` (12), 979--1006,
    SIAM J. Optim.

    Gill, P E, Murray, W, Saunders, M A and Wright, M H, 1986, `Users'
    guide for NPSOL (Version 4.0): a Fortran package for nonlinear
    programming`, Report SOL 86-2, Department of Operations Research,
    Stanford University

    Gill, P E, Murray, W, Saunders, M A and Wright, M H, 1989, `A
    practical anti-cycling procedure for linearly constrained
    optimization`, Math. Programming (45), 437--474

    Gill, P E, Murray, W, Saunders, M A and Wright, M H, 1992, `Some
    theoretical properties of an augmented Lagrangian merit function`,
    Advances in Optimization and Parallel Computing, (ed P M Pardalos),
    101--128, North Holland

    Gill, P E, Murray, W and Wright, M H, 1981, `Practical
    Optimization`, Academic Press

    Hock, W and Schittkowski, K, 1981, `Test Examples for Nonlinear
    Programming Codes. Lecture Notes in Economics and Mathematical
    Systems` (187), Springer--Verlag

    Murtagh, B A and Saunders, M A, 1995, `MINOS 5.4 users' guide`,
    Report SOL 83-20R, Department of Operations Research, Stanford
    University

    Ortega, J M and Rheinboldt, W C, 1970, `Iterative Solution of
    Nonlinear Equations in Several Variables`, Academic Press

    Powell, M J D, 1974, `Introduction to constrained optimization`,
    Numerical Methods for Constrained Optimization, (eds P E Gill and W
    Murray), 1--28, Academic Press

    See Also
    --------
    :meth:`naginterfaces.library.examples.opt.nlp1_sparse_solve_ex.main`
    """

    fun_name = 'naginterfaces.library.opt.nlp1_sparse_solve'

    cb_confun_name = fun_name + ':confun'

    def cb_confun(mode, ncnln, njnln, nnzjac, x, f, fjac, nstate, data=None):
        """Wrapping user's confun. Called as base confun."""
        if data is not None:
            _mode_confun, _f_confun, _fjac_confun = confun(
                mode,
                ncnln,
                x,
                fjac,
                nstate,
                data,
            )
        else:
            _mode_confun, _f_confun, _fjac_confun = confun(
                mode,
                ncnln,
                x,
                fjac,
                nstate,
            )
        _mode_confun_base = _mode_confun
        f[:] = _utils._to_ndarray(
            {'fun_name': cb_confun_name, 'returned': 'f'},
            _f_confun,
            copy=False,
            exp_shape=(ncnln,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        fjac[:] = _utils._to_ndarray(
            {'fun_name': cb_confun_name, 'returned': 'fjac'},
            _fjac_confun,
            copy=False,
            exp_shape=(nnzjac,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        return _mode_confun_base

    cb_objfun_name = fun_name + ':objfun'

    def cb_objfun(mode, nonln, x, objgrd, nstate, data=None):
        """Wrapping user's objfun. Called as base objfun."""
        if data is not None:
            _mode_objfun, _objf_objfun, _objgrd_objfun = objfun(
                mode,
                x,
                objgrd,
                nstate,
                data,
            )
        else:
            _mode_objfun, _objf_objfun, _objgrd_objfun = objfun(
                mode,
                x,
                objgrd,
                nstate,
            )
        _mode_objfun_base = _mode_objfun
        objgrd[:] = _utils._to_ndarray(
            {'fun_name': cb_objfun_name, 'returned': 'objgrd'},
            _objgrd_objfun,
            copy=False,
            exp_shape=(nonln,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _objf_objfun_base = _objf_objfun
        return (
            _mode_objfun_base,
            _objf_objfun_base,
        )

    _b_utils._check_callable(confun, {'fun_name': fun_name, 'entity': 'confun'}, allow_none=True)
    _confun = None if confun is None else cb_confun
    _b_utils._check_callable(objfun, {'fun_name': fun_name, 'entity': 'objfun'}, allow_none=True)
    _objfun = None if objfun is None else cb_objfun
    _ka = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ka'},
        ka,
        copy=False,
    )
    _n = _b_utils._EngineIntScalarType(
        _b_utils._EngineIntArrayType.get_shape(_ka, {'fun_name': fun_name, 'entity': 'ka'}, exp_rank=1)[0]-1, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nnz_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=1)[0]
    _nnz = _b_utils._EngineIntScalarType(
        _nnz_py, {'fun_name': fun_name, 'entity': '_nnz'},
    ).to_ctypes()
    _ha = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ha'},
        ha,
        copy=False,
    )
    _bl = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'bl'},
        bl,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _bu = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'bu'},
        bu,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _names = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'names'},
        names,
        copy=False,
    )
    _nname_py = _b_utils._EngineCharArrayType.get_shape(_names, {'fun_name': fun_name, 'entity': 'names'}, exp_rank=1)[0]
    _nname = _b_utils._EngineIntScalarType(
        _nname_py, {'fun_name': fun_name, 'entity': '_nname'},
    ).to_ctypes()
    _xs = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xs'},
        xs,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _istate = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'istate'},
        istate,
        copy=True,
    )
    _clamda = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'clamda'},
        clamda,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineIntScalarType.to_py(_n)
    _m = _b_utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = _b_utils._EngineIntScalarType.to_py(_m)
    _leniz = max(500, _n_py+_m_py) if leniz is None else leniz
    try:
        _return_scalars = _b_opt.nlp1_sparse_solve(
            _confun,
            _objfun,
            _n,
            _m,
            ncnln,
            nonln,
            njnln,
            iobj,
            _nnz,
            _a,
            _ha,
            _ka,
            _bl,
            _bu,
            start,
            _nname,
            _names,
            ns,
            _xs,
            _istate,
            _clamda,
            _leniz,
            lenz,
            comm,
            data,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _a_py = _a
    _ns_py = _return_scalars.ns
    _xs_py = _xs
    _istate_py = _istate
    _clamda_py = _clamda
    _miniz_py = _return_scalars.miniz
    _minz_py = _return_scalars.minz
    _ninf_py = _return_scalars.ninf
    _sinf_py = _return_scalars.sinf
    _obj_py = _return_scalars.obj
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['a', 'ns', 'xs', 'istate', 'clamda', 'miniz', 'minz', 'ninf', 'sinf', 'obj'],
    )(
        _a_py,
        _ns_py,
        _xs_py,
        _istate_py,
        _clamda_py,
        _miniz_py,
        _minz_py,
        _ninf_py,
        _sinf_py,
        _obj_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def nlp1_sparse_option_file(ioptns, comm, io_manager=None):
    u"""
    Supply option values for ``nlp1_sparse_solve`` from external file.

    ``nlp1_sparse_option_file`` may be used to supply options to
    ``nlp1_sparse_solve`` from an external file.

    For full information please refer to the NAG Library document for
    e04uh

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04uhf.html

    Parameters
    ----------
    ioptns : int
        The unit number (see
        :meth:`~naginterfaces.base.utils.FileObjManager.unit_from_fileobj`)
        of the options file to be read.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``nlp1_init``.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Raises
    ------
    NagValueError
        (`errno` 1)
            `ioptns` is not in the range [0,99].

        (`errno` 1)
            On entry, `ioptns` = *<value>*.

            Constraint: 0 <= `ioptns` <= 99.

        (`errno` 2)
            ``Begin`` was found, but end-of-file was found before
            ``End`` was found.

        (`errno` 3)
            End-of-file was found before ``Begin`` was found.

        (`errno` 5)
            One or more lines of the options file is invalid.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    References
    ----------
    Hock, W and Schittkowski, K, 1981, `Test Examples for Nonlinear
    Programming Codes. Lecture Notes in Economics and Mathematical
    Systems` (187), Springer--Verlag
    """

    _b_opt.nlp1_sparse_option_file(
        ioptns,
        comm,
        io_manager,
    )

def nlp1_sparse_option_string(optstr, comm, io_manager=None):
    u"""
    Supply option values to ``nlp1_sparse_solve`` from a character
    string.

    ``nlp1_sparse_option_string`` may be used to supply individual
    options to ``nlp1_sparse_solve``.

    For full information please refer to the NAG Library document for
    e04uj

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04ujf.html

    Parameters
    ----------
    optstr : str
        A single valid option string (as described in [Description] and
        in [Other Parameters for nlp1_sparse_solve]).

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``nlp1_init``.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Raises
    ------
    NagValueError
        (`errno` 5)
            The supplied option string is invalid. Supplied value was:
            *<value>*.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    _b_opt.nlp1_sparse_option_string(
        optstr,
        comm,
        io_manager,
    )

def lsq_gencon_deriv_option_file(ioptns, comm, io_manager=None):
    u"""
    Supply option values for ``lsq_gencon_deriv`` from external file.

    ``lsq_gencon_deriv_option_file`` may be used to supply options to
    ``lsq_gencon_deriv`` from an external file.

    For full information please refer to the NAG Library document for
    e04uq

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04uqf.html

    Parameters
    ----------
    ioptns : int
        The unit number (see
        :meth:`~naginterfaces.base.utils.FileObjManager.unit_from_fileobj`)
        of the options file to be read.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``nlp1_init``.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Raises
    ------
    NagValueError
        (`errno` 1)
            `ioptns` is not in the range [0,99].

        (`errno` 1)
            On entry, `ioptns` = *<value>*.

            Constraint: 0 <= `ioptns` <= 99.

        (`errno` 2)
            ``Begin`` was found, but end-of-file was found before
            ``End`` was found.

        (`errno` 3)
            End-of-file was found before ``Begin`` was found.

        (`errno` 5)
            One or more lines of the options file is invalid.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    _b_opt.lsq_gencon_deriv_option_file(
        ioptns,
        comm,
        io_manager,
    )

def lsq_gencon_deriv_option_string(optstr, comm, io_manager=None):
    u"""
    Supply option values to ``lsq_gencon_deriv`` from a character
    string.

    ``lsq_gencon_deriv_option_string`` may be used to supply individual
    options to ``lsq_gencon_deriv``.

    For full information please refer to the NAG Library document for
    e04ur

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04urf.html

    Parameters
    ----------
    optstr : str
        A single valid option string (as described in [Description] and
        in [Other Parameters for lsq_gencon_deriv]).

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``nlp1_init``.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Raises
    ------
    NagValueError
        (`errno` 5)
            The supplied option string is invalid. Supplied value was:
            *<value>*.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    _b_opt.lsq_gencon_deriv_option_string(
        optstr,
        comm,
        io_manager,
    )

def lsq_gencon_deriv(bl, bu, y, objfun, x, comm, a=None, confun=None, istate=None, cjac=None, fjac=None, clamda=None, r=None, data=None, io_manager=None):
    u"""
    Minimum of a sum of squares, nonlinear constraints, dense,
    active-set SQP method, using function values and optionally first
    derivatives.

    Note: this function uses optional algorithmic parameters, see also:
    ``lsq_gencon_deriv_option_file``,
    ``lsq_gencon_deriv_option_string``, ``nlp1_init``.

    ``lsq_gencon_deriv`` is designed to minimize an arbitrary smooth sum
    of squares function subject to constraints (which may include simple
    bounds on the variables, linear constraints and smooth nonlinear
    constraints) using a sequential quadratic programming (SQP) method.
    As many first derivatives as possible should be supplied by you; any
    unspecified derivatives are approximated by finite differences.
    See the description of the option 'Derivative Level', in [Other
    Parameters].
    It is not intended for large sparse problems.

    ``lsq_gencon_deriv`` may also be used for unconstrained,
    bound-constrained and linearly constrained optimization.

    For full information please refer to the NAG Library document for
    e04us

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04usf.html

    Parameters
    ----------
    bl : float, array-like, shape (n+nclin+ncnln)
        `bl` must contain the lower bounds for all the constraints.

    bu : float, array-like, shape (n+nclin+ncnln)
        `bu` must contain the upper bounds for all the constraints.

    y : float, array-like, shape (m)
        The coefficients of the constant vector y of the objective
        function.

    objfun : callable f = objfun(mode, needfi, x, fjac, nstate,
        data=None)
        `objfun` must calculate either the ith element of the vector
        f(x) = (f_1(x),f_2(x),...,f_m(x))^T or all m elements of f(x)
        and (optionally) its Jacobian (= df/dx) for a specified
        n-element vector x.

        Parameters
        ~~~~~~~~~~
        mode : int
            Indicates which values must be assigned during each call of
            `objfun`. Only the following values need be assigned:

            `mode` = 0 and `needfi` = i, where i > 0

                `f`[i-1].

            `mode` = 0 and `needfi` < 0

                `f`.

            `mode` = 1 and `needfi` < 0

                All available elements of `fjac`.

            `mode` = 2 and `needfi` < 0

                `f` and all available elements of `fjac`.

        needfi : int
            If `needfi` = i > 0, only the ith element of f(x) needs to
            be evaluated at x; the remaining elements need not be set.
            This can result in significant computational savings when
            m≫n.

        x : float, ndarray, shape (n)
            x, the vector of variables at which f(x) and/or all
            available elements of its Jacobian are to be evaluated.

        fjac : float, ndarray, shape (m, n), to be modified in place
            `On entry`: is set to a special value.

            `On exit`: if `mode` = 1 or 2 and `needfi` < 0, the ith row
            of `fjac` must contain the available elements of the vector
            nabla f_i given by

                nabla f_i = (df_i/dx_1,df_i/dx_2,...,df_i/dx_n)^T,

            evaluated at the point x. See also the argument `nstate`.

        nstate : int
            If `nstate` = 1, then ``lsq_gencon_deriv`` is calling
            `objfun` for the first time. This argument setting allows
            you to save computation time if certain data must be read or
            calculated only once.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        f : float, array-like, shape (m)
            If `mode` = 0 and `needfi` = i > 0, `f`[i-1] must contain
            the value of f_i at x.

            If `mode` = 0 or 2 and `needfi` < 0, `f`[i-1] must contain
            the value of f_i at x, for i = 1,2,...,m.

    x : float, array-like, shape (n)
        An initial estimate of the solution.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``nlp1_init``.

    a : None or float, array-like, shape (nclin, :), optional
        Note: the required extent for this argument in dimension 2 is
        determined as follows: if nclin>0: n; otherwise: 1.

        The ith row of `a` contains the ith row of the matrix A_L of
        general linear constraints in (1). That is, the ith row contains
        the coefficients of the ith general linear constraint, for i =
        1,2,...,nclin.

    confun : None or callable c = confun(mode, needc, x, cjac, nstate,
        data=None), optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `confun` must calculate the vector c(x) of nonlinear constraint
        functions and (optionally) its Jacobian (= dc/dx) for a
        specified n-element vector x.

        If there are no nonlinear constraints (i.e., ncnln = 0),
        `confun` will never be called by ``lsq_gencon_deriv`` and
        `confun` may be None. If there are nonlinear constraints, the
        first call to `confun` will occur before the first call to
        `objfun`.

        Parameters
        ~~~~~~~~~~
        mode : int
            Indicates which values must be assigned during each call of
            `confun`. Only the following values need be assigned, for
            each value of i such that `needc`[i-1] > 0:

            `mode` = 0

                `c`[i-1].

            `mode` = 1

                All available elements in the ith row of `cjac`.

            `mode` = 2

                `c`[i-1] and all available elements in the ith row of
                `cjac`.

        needc : int, ndarray, shape (ncnln)
            The indices of the elements of `c` and/or `cjac` that must
            be evaluated by `confun`. If `needc`[i-1] > 0, the ith
            element of `c` and/or the available elements of the ith row
            of `cjac` (see argument `mode`) must be evaluated at x.

        x : float, ndarray, shape (n)
            x, the vector of variables at which the constraint functions
            and/or all available elements of the constraint Jacobian are
            to be evaluated.

        cjac : float, ndarray, shape (max(1, ncnln), n), to be modified
        in place
            `On entry`: is set to a special value.

            `On exit`: if `needc`[i-1] > 0 and `mode` = 1 or 2, the ith
            row of `cjac` must contain the available elements of the
            vector nabla c_i given by

                nabla c_i = (dc_i/dx_1,dc_i/dx_2,...,dc_i/dx_n)^T,

            where dc_i/dx_j is the partial derivative of the ith
            constraint with respect to the jth variable, evaluated at
            the point x. See also the argument `nstate`. The remaining
            rows of `cjac`, corresponding to non-positive elements of
            `needc`, are ignored.

            If all elements of the constraint Jacobian are known (i.e.,
            'Derivative Level' = 2 or 3), any constant elements may be
            assigned to `cjac` one time only at the start of the
            optimization.

            An element of `cjac` that is not subsequently assigned in
            `confun` will retain its initial value throughout.

            Constant elements may be loaded into `cjac` either before
            the call to ``lsq_gencon_deriv`` or during the first call to
            `confun` (signalled by the value `nstate` = 1).

            The ability to preload constants is useful when many
            Jacobian elements are identically zero, in which case `cjac`
            may be initialized to zero and nonzero elements may be reset
            by `confun`.

            Note that constant nonzero elements do affect the values of
            the constraints.

            Thus, if `cjac`[i-1,j-1] is set to a constant value, it need
            not be reset in subsequent calls to `confun`, but the value
            `cjac`[i-1,j-1]*`x`[j-1] must nonetheless be added to
            `c`[i-1].

            For example, if `cjac`[0,0] = 2 and `cjac`[0,1] = -5, the
            term 2*`x`[0]-5*`x`[1] must be included in the definition of
            `c`[0].

            It must be emphasized that, if 'Derivative Level' = 0 or 1,
            unassigned elements of `cjac` are not treated as constant;
            they are estimated by finite differences, at nontrivial
            expense.

            If you do not supply a value for the option 'Difference
            Interval', an interval for each element of x is computed
            automatically at the start of the optimization.

            The automatic procedure can usually identify constant
            elements of `cjac`, which are then computed once only by
            finite differences.

        nstate : int
            If `nstate` = 1, then ``lsq_gencon_deriv`` is calling
            `confun` for the first time. This argument setting allows
            you to save computation time if certain data must be read or
            calculated only once.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        c : float, array-like, shape (ncnln)
            If `needc`[i-1] > 0 and `mode` = 0 or 2, `c`[i-1] must
            contain the value of the ith constraint at x. The remaining
            elements of `c`, corresponding to the non-positive elements
            of `needc`, are ignored.

    istate : None or int, array-like, shape (n+nclin+ncnln), optional
        Need not be set if the (default) option 'Cold Start' is used.

        If the option 'Warm Start' has been chosen, the elements of
        `istate` corresponding to the bounds and linear constraints
        define the initial working set for the procedure that finds a
        feasible point for the linear constraints and bounds.

        The active set at the conclusion of this procedure and the
        elements of `istate` corresponding to nonlinear constraints then
        define the initial working set for the first QP subproblem.

        More precisely, the first n elements of `istate` refer to the
        upper and lower bounds on the variables, the next n_L elements
        refer to the upper and lower bounds on A_L x, and the next n_N
        elements refer to the upper and lower bounds on c(x).

        Possible values for `istate`[j-1] are as follows:

        [table omitted]

        The values -2, -1 and 4 are also acceptable but will be modified
        by the function.

        If ``lsq_gencon_deriv`` has been called previously with the same
        values of n, nclin and ncnln, `istate` already contains
        satisfactory information. (See also the description of the
        option 'Warm Start'.) The function also adjusts (if necessary)
        the values supplied in `x` to be consistent with `istate`.

    cjac : None or float, array-like, shape (ncnln, :), optional
        Note: the required extent for this argument in dimension 2 is
        determined as follows: if ncnln>0: n; otherwise: 1.

        In general, `cjac` need not be initialized before the call to
        ``lsq_gencon_deriv``. However, if 'Derivative Level' = 3, you
        may optionally set the constant elements of `cjac` (see argument
        `nstate` in the description of `confun`). Such constant elements
        need not be re-assigned on subsequent calls to `confun`.

    fjac : None or float, array-like, shape (m, n), optional
        In general, `fjac` need not be initialized before the call to
        ``lsq_gencon_deriv``. However, if 'Derivative Level' = 3, you
        may optionally set the constant elements of `fjac` (see argument
        `nstate` in the description of `objfun`). Such constant elements
        need not be re-assigned on subsequent calls to `objfun`.

    clamda : None or float, array-like, shape (n+nclin+ncnln), optional
        Need not be set if the (default) option 'Cold Start' is used.

        If the option 'Warm Start' has been chosen, `clamda`[j-1] must
        contain a multiplier estimate for each nonlinear constraint with
        a sign that matches the status of the constraint specified by
        the `istate` array, for j = n+nclin+1,...,n+nclin+ncnln.

        The remaining elements need not be set.

        Note that if the jth constraint is defined as 'inactive' by the
        initial value of the `istate` array (i.e., `istate`[j-1] = 0),
        `clamda`[j-1] should be zero; if the jth constraint is an
        inequality active at its lower bound (i.e., `istate`[j-1] = 1),
        `clamda`[j-1] should be non-negative; if the jth constraint is
        an inequality active at its upper bound (i.e., `istate`[j-1] =
        2, `clamda`[j-1] should be non-positive.

        If necessary, the function will modify `clamda` to match these
        rules.

    r : None or float, array-like, shape (n, n), optional
        Need not be initialized if the (default) option 'Cold Start' is
        used.

        If the option 'Warm Start' has been chosen, `r` must contain the
        upper triangular Cholesky factor R of the initial approximation
        of the Hessian of the Lagrangian function, with the variables in
        the natural order.

        Elements not in the upper triangular part of `r` are assumed to
        be zero and need not be assigned.

    data : arbitrary, optional
        User-communication data for callback functions.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    itera : int
        The number of major iterations performed.

    istate : int, ndarray, shape (n+nclin+ncnln)
        The status of the constraints in the QP working set at the point
        returned in `x`. The significance of each possible value of
        `istate`[j-1] is as follows:

        [table omitted]

    c : None or float, ndarray, shape (ncnln)
        If ncnln > 0, `c`[i-1] contains the value of the ith nonlinear
        constraint function c_i at the final iterate, for i =
        1,2,...,ncnln.

        If ncnln = 0, `c` is returned as None.

    cjac : None or float, ndarray, shape (ncnln, :)
        If ncnln > 0, `cjac` contains the Jacobian matrix of the
        nonlinear constraint functions at the final iterate, i.e.,
        `cjac`[i-1,j-1] contains the partial derivative of the ith
        constraint function with respect to the jth variable, for j =
        1,2,...,n, for i = 1,2,...,ncnln. (See the discussion of
        argument `cjac` under `confun`.)

        If ncnln = 0, `cjac` is returned as None.

    f : float, ndarray, shape (m)
        `f`[i-1] contains the value of the ith function f_i at the final
        iterate, for i = 1,2,...,m.

    fjac : float, ndarray, shape (m, n)
        The Jacobian matrix of the functions f_1,f_2,...,f_m at the
        final iterate, i.e., `fjac`[i-1,j-1] contains the partial
        derivative of the ith function with respect to the jth variable,
        for j = 1,2,...,n, for i = 1,2,...,m. (See also the discussion
        of argument `fjac` under `objfun`.)

    clamda : float, ndarray, shape (n+nclin+ncnln)
        The values of the QP multipliers from the last QP subproblem.
        `clamda`[j-1] should be non-negative if `istate`[j-1] = 1 and
        non-positive if `istate`[j-1] = 2.

    objf : float
        The value of the objective function at the final iterate.

    r : float, ndarray, shape (n, n)
        If 'Hessian' = 'NO', `r` contains the upper triangular Cholesky
        factor R of Q^TH~Q, an estimate of the transformed and reordered
        Hessian of the Lagrangian at x (see (6)). If 'Hessian' = 'YES',
        `r` contains the upper triangular Cholesky factor R of H, the
        approximate (untransformed) Hessian of the Lagrangian, with the
        variables in the natural order.

    x : float, ndarray, shape (n)
        The final estimate of the solution.

    Other Parameters
    ----------------
    'Central Difference Interval' : float
        Default values are computed

        If the algorithm switches to central differences because the
        forward-difference approximation is not sufficiently accurate,
        the value of r is used as the difference interval for every
        element of x.
        The switch to central differences is indicated by C at the end
        of each line of intermediate printout produced by the major
        iterations (see [Further Comments]).
        The use of finite differences is discussed further under the
        option 'Difference Interval'.

        If you supply a value for this option, a small value between 0.0
        and 1.0 is appropriate.

    'Cold Start' : valueless
        Default

        This option controls the specification of the initial working
        set in both the procedure for finding a feasible point for the
        linear constraints and bounds, and in the first QP subproblem
        thereafter.
        With a 'Cold Start', the first working set is chosen by
        ``lsq_gencon_deriv`` based on the values of the variables and
        constraints at the initial point.
        Broadly speaking, the initial working set will include equality
        constraints and bounds or inequality constraints that violate or
        'nearly' satisfy their bounds (to within 'Crash Tolerance').

        With a 'Warm Start', you must set the `istate` array and define
        `clamda` and `r` as discussed in [Parameters]. `istate` values
        associated with bounds and linear constraints determine the
        initial working set of the procedure to find a feasible point
        with respect to the bounds and linear constraints. `istate`
        values associated with nonlinear constraints determine the
        initial working set of the first QP subproblem after such a
        feasible point has been found. ``lsq_gencon_deriv`` will
        override your specification of `istate` if necessary, so that a
        poor choice of the working set will not cause a fatal error.
        For instance, any elements of `istate` which are set to -2, -1
        or 4 will be reset to zero, as will any elements which are set
        to 3 when the corresponding elements of `bl` and `bu` are not
        equal.
        A 'Warm Start' will be advantageous if a good estimate of the
        initial working set is available -- for example, when
        ``lsq_gencon_deriv`` is called repeatedly to solve related
        problems.

    'Warm Start' : valueless
        This option controls the specification of the initial working
        set in both the procedure for finding a feasible point for the
        linear constraints and bounds, and in the first QP subproblem
        thereafter.
        With a 'Cold Start', the first working set is chosen by
        ``lsq_gencon_deriv`` based on the values of the variables and
        constraints at the initial point.
        Broadly speaking, the initial working set will include equality
        constraints and bounds or inequality constraints that violate or
        'nearly' satisfy their bounds (to within 'Crash Tolerance').

        With a 'Warm Start', you must set the `istate` array and define
        `clamda` and `r` as discussed in [Parameters]. `istate` values
        associated with bounds and linear constraints determine the
        initial working set of the procedure to find a feasible point
        with respect to the bounds and linear constraints. `istate`
        values associated with nonlinear constraints determine the
        initial working set of the first QP subproblem after such a
        feasible point has been found. ``lsq_gencon_deriv`` will
        override your specification of `istate` if necessary, so that a
        poor choice of the working set will not cause a fatal error.
        For instance, any elements of `istate` which are set to -2, -1
        or 4 will be reset to zero, as will any elements which are set
        to 3 when the corresponding elements of `bl` and `bu` are not
        equal.
        A 'Warm Start' will be advantageous if a good estimate of the
        initial working set is available -- for example, when
        ``lsq_gencon_deriv`` is called repeatedly to solve related
        problems.

    'Crash Tolerance' : float
        Default = 0.01

        This value is used in conjunction with the option 'Cold Start'
        (the default value) when ``lsq_gencon_deriv`` selects an initial
        working set.
        If 0 <= r <= 1, the initial working set will include (if
        possible) bounds or general inequality constraints that lie
        within r of their bounds.
        In particular, a constraint of the form a_j^Tx >= l will be
        included in the initial working set if |a_j^Tx-l| <= r(1+|l|).
        If r < 0 or r > 1, the default value is used.

    'Defaults' : valueless
        This special keyword may be used to reset all options to their
        default values.

    'Derivative Level' : int
        Default = 3

        This argument indicates which derivatives are provided in
        functions `objfun` and `confun`.
        The possible choices for i are the following.

        [table omitted]

        The value i = 3 should be used whenever possible, since
        ``lsq_gencon_deriv`` is more reliable (and will usually be more
        efficient) when all derivatives are exact.

        If i = 0 or 2, ``lsq_gencon_deriv`` will approximate unspecified
        elements of the objective Jacobian, using finite differences.
        The computation of finite difference approximations usually
        increases the total run-time, since a call to `objfun` is
        required for each unspecified element.
        Furthermore, less accuracy can be attained in the solution (see
        Module 8 of Gill `et al.` (1981), for a discussion of limiting
        accuracy).

        If i = 0 or 1, ``lsq_gencon_deriv`` will approximate unspecified
        elements of the constraint Jacobian.
        One call to `confun` is needed for each variable for which
        partial derivatives are not available.
        For example, if the constraint Jacobian has the form

            [table omitted]

        where '*' indicates an element provided by you and '?' indicates
        an unspecified element, ``lsq_gencon_deriv`` will call `confun`
        twice: once to estimate the missing element in column 2, and
        again to estimate the two missing elements in column 3. (Since
        columns 1 and 4 are known, they require no calls to `confun`.)

        At times, central differences are used rather than forward
        differences, in which case twice as many calls to `objfun` and
        `confun` are needed. (The switch to central differences is not
        under your control.)

        If i < 0 or i > 3, the default value is used.

    'Difference Interval' : float
        Default values are computed

        This option defines an interval used to estimate derivatives by
        finite differences in the following circumstances:

        (a) For verifying the objective and/or constraint gradients (see
            the description of the option 'Verify').

        (#) For estimating unspecified elements of the objective and/or
            constraint Jacobian matrix.

        In general, a derivative with respect to the jth variable is
        approximated using the interval delta_j, where delta_j =
        r(1+|x^_j|), with x^ the first point feasible with respect to
        the bounds and linear constraints.
        If the functions are well scaled, the resulting derivative
        approximation should be accurate to O(r).
        See Gill `et al.` (1981) for a discussion of the accuracy in
        finite difference approximations.

        If a difference interval is not specified, a finite difference
        interval will be computed automatically for each variable by a
        procedure that requires up to six calls of `confun` and `objfun`
        for each element.
        This option is recommended if the function is badly scaled or
        you wish to have ``lsq_gencon_deriv`` determine constant
        elements in the objective and constraint gradients (see the
        descriptions of `confun` and `objfun` in [Parameters]).

        If you supply a value for this option, a small value between 0.0
        and 1.0 is appropriate.

    'Feasibility Tolerance' : float
        Default = sqrt(epsilon)

        The scalar r defines the maximum acceptable `absolute`
        violations in linear and nonlinear constraints at a 'feasible'
        point; i.e., a constraint is considered satisfied if its
        violation does not exceed r.
        If r < epsilon or r >= 1, the default value is used.
        Using this keyword sets both options 'Linear Feasibility
        Tolerance' and 'Nonlinear Feasibility Tolerance' to r, if
        epsilon <= r < 1. (Additional details are given under the
        descriptions of these options.)

    'Function Precision' : float
        Default = epsilon^0.9

        This argument defines epsilon_r, which is intended to be a
        measure of the accuracy with which the problem functions F(x)
        and c(x) can be computed.
        If r < epsilon or r >= 1, the default value is used.

        The value of epsilon_r should reflect the relative precision of
        1+|F(x)|; i.e., epsilon_r acts as a relative precision when |F|
        is large and as an absolute precision when |F| is small.
        For example, if F(x) is typically of order 1000 and the first
        six significant digits are known to be correct, an appropriate
        value for epsilon_r would be 10^(-6).
        In contrast, if F(x) is typically of order 10^(-4) and the first
        six significant digits are known to be correct, an appropriate
        value for epsilon_r would be 10^(-10).
        The choice of epsilon_r can be quite complicated for badly
        scaled problems; see Module 8 of Gill `et al.` (1981) for a
        discussion of scaling techniques.
        The default value is appropriate for most simple functions that
        are computed with full accuracy.
        However, when the accuracy of the computed function values is
        known to be significantly worse than full precision, the value
        of epsilon_r should be large enough so that ``lsq_gencon_deriv``
        will not attempt to distinguish between function values that
        differ by less than the error inherent in the calculation.

    'Hessian' : str
        Default = NO

        This option controls the contents of the upper triangular matrix
        R (see [Parameters]). ``lsq_gencon_deriv`` works exclusively
        with the `transformed and reordered` Hessian H_Q, and hence
        extra computation is required to form the Hessian itself.
        If 'Hessian' = 'NO', `r` contains the Cholesky factor of the
        transformed and reordered Hessian.
        If 'Hessian' = 'YES', the Cholesky factor of the approximate
        Hessian itself is formed and stored in `r`.
        You should select 'Hessian' = 'YES' if a 'Warm Start' will be
        used for the next call to ``lsq_gencon_deriv``.

    'Infinite Bound Size' : float
        Default = 10^20

        If r > 0, r defines the 'infinite' bound bigbnd in the
        definition of the problem constraints.
        Any upper bound greater than or equal to bigbnd will be regarded
        as + infinity (and similarly any lower bound less than or equal
        to -bigbnd will be regarded as - infinity).
        If r < 0, the default value is used.

    'Infinite Step Size' : float
        Default = max(bigbnd, 10^20)

        If r > 0, r specifies the magnitude of the change in variables
        that is treated as a step to an unbounded solution.
        If the change in x during an iteration would exceed the value of
        r, the objective function is considered to be unbounded below in
        the feasible region.
        If r <= 0, the default value is used.

    'JTJ Initial Hessian' : valueless
        Default

        This option controls the initial value of the upper triangular
        matrix R.
        If J denotes the objective Jacobian matrix nabla f(x), J^TJ is
        often a good approximation to the objective Hessian matrix
        nabla^2F(x) (see also option 'Reset Frequency').

    'Unit Initial Hessian' : valueless
        This option controls the initial value of the upper triangular
        matrix R.
        If J denotes the objective Jacobian matrix nabla f(x), J^TJ is
        often a good approximation to the objective Hessian matrix
        nabla^2F(x) (see also option 'Reset Frequency').

    'Line Search Tolerance' : float
        Default = 0.9

        The value r (0 <= r < 1) controls the accuracy with which the
        step alpha taken during each iteration approximates a minimum of
        the merit function along the search direction (the smaller the
        value of r, the more accurate the linesearch).
        The default value r = 0.9 requests an inaccurate search and is
        appropriate for most problems, particularly those with any
        nonlinear constraints.

        If there are no nonlinear constraints, a more accurate search
        may be appropriate when it is desirable to reduce the number of
        major iterations -- for example, if the objective function is
        cheap to evaluate, or if a substantial number of derivatives are
        unspecified.
        If r < 0 or r >= 1, the default value is used.

    'Linear Feasibility Tolerance' : float
        Default = sqrt(epsilon)

        The default value of r_2 is epsilon^0.33 if 'Derivative Level' =
        0 or 1, and sqrt(epsilon) otherwise.

        The scalars r_1 and r_2 define the maximum acceptable `absolute`
        violations in linear and nonlinear constraints at a 'feasible'
        point; i.e., a linear constraint is considered satisfied if its
        violation does not exceed r_1, and similarly for a nonlinear
        constraint and r_2.
        If r_m < epsilon or r_m >= 1, the default value is used, for m =
        1,2,...,2.

        On entry to ``lsq_gencon_deriv``, an iterative procedure is
        executed in order to find a point that satisfies the linear
        constraints and bounds on the variables to within the tolerance
        r_1.
        All subsequent iterates will satisfy the linear constraints to
        within the same tolerance (unless r_1 is comparable to the
        finite difference interval).

        For nonlinear constraints, the feasibility tolerance r_2 defines
        the largest constraint violation that is acceptable at an
        optimal point.
        Since nonlinear constraints are generally not satisfied until
        the final iterate, the value of option 'Nonlinear Feasibility
        Tolerance' acts as a partial termination criterion for the
        iterative sequence generated by ``lsq_gencon_deriv`` (see also
        option 'Optimality Tolerance').

        These tolerances should reflect the precision of the
        corresponding constraints.
        For example, if the variables and the coefficients in the linear
        constraints are of order unity, and the latter are correct to
        about 6 decimal digits, it would be appropriate to specify r_1
        as 10^(-6).

    'Nonlinear Feasibility Tolerance' : float
        Default = epsilon^0.33 or sqrt(epsilon)

        The default value of r_2 is epsilon^0.33 if 'Derivative Level' =
        0 or 1, and sqrt(epsilon) otherwise.

        The scalars r_1 and r_2 define the maximum acceptable `absolute`
        violations in linear and nonlinear constraints at a 'feasible'
        point; i.e., a linear constraint is considered satisfied if its
        violation does not exceed r_1, and similarly for a nonlinear
        constraint and r_2.
        If r_m < epsilon or r_m >= 1, the default value is used, for m =
        1,2,...,2.

        On entry to ``lsq_gencon_deriv``, an iterative procedure is
        executed in order to find a point that satisfies the linear
        constraints and bounds on the variables to within the tolerance
        r_1.
        All subsequent iterates will satisfy the linear constraints to
        within the same tolerance (unless r_1 is comparable to the
        finite difference interval).

        For nonlinear constraints, the feasibility tolerance r_2 defines
        the largest constraint violation that is acceptable at an
        optimal point.
        Since nonlinear constraints are generally not satisfied until
        the final iterate, the value of option 'Nonlinear Feasibility
        Tolerance' acts as a partial termination criterion for the
        iterative sequence generated by ``lsq_gencon_deriv`` (see also
        option 'Optimality Tolerance').

        These tolerances should reflect the precision of the
        corresponding constraints.
        For example, if the variables and the coefficients in the linear
        constraints are of order unity, and the latter are correct to
        about 6 decimal digits, it would be appropriate to specify r_1
        as 10^(-6).

    'List' : valueless
        Option 'List' enables printing of each option specification as
        it is supplied. 'Nolist' suppresses this printing.

    'Nolist' : valueless
        Default for ``lsq_gencon_deriv`` = 'Nolist'

        Option 'List' enables printing of each option specification as
        it is supplied. 'Nolist' suppresses this printing.

    'Major Iteration Limit' : int
        Default = max(50, 3*(n+n_L)+10n_N)

        The value of i specifies the maximum number of major iterations
        allowed before termination.
        Setting i = 0 and 'Major Print Level' > 0 means that the
        workspace needed will be computed and printed, but no iterations
        will be performed.
        If i < 0, the default value is used.

    'Iteration Limit' : int
        Default = max(50, 3*(n+n_L)+10n_N)

        The value of i specifies the maximum number of major iterations
        allowed before termination.
        Setting i = 0 and 'Major Print Level' > 0 means that the
        workspace needed will be computed and printed, but no iterations
        will be performed.
        If i < 0, the default value is used.

    'Iters' : int
        Default = max(50, 3*(n+n_L)+10n_N)

        The value of i specifies the maximum number of major iterations
        allowed before termination.
        Setting i = 0 and 'Major Print Level' > 0 means that the
        workspace needed will be computed and printed, but no iterations
        will be performed.
        If i < 0, the default value is used.

    'Itns' : int
        Default = max(50, 3*(n+n_L)+10n_N)

        The value of i specifies the maximum number of major iterations
        allowed before termination.
        Setting i = 0 and 'Major Print Level' > 0 means that the
        workspace needed will be computed and printed, but no iterations
        will be performed.
        If i < 0, the default value is used.

    'Major Print Level' : int
        The value of i controls the amount of printout produced by the
        major iterations of ``lsq_gencon_deriv``, as indicated below.
        A detailed description of the printed output is given in
        [Further Comments] (summary output at each major iteration and
        the final solution) and [Monitoring Information] (monitoring
        information at each major iteration). (See also the description
        of the option 'Minor Print Level'.)

        The following printout is sent to the file object associated
        with the advisory I/O unit (see
        :class:`~naginterfaces.base.utils.FileObjManager`):

        [table omitted]

        The following printout is sent to the unit number given by the
        option 'Monitoring File':

        [table omitted]

        If 'Major Print Level' >= 5 and the unit number defined by the
        option 'Monitoring File' is the advisory unit number, the
        summary output for each major iteration is suppressed.

    'Print Level' : int
        Default for ``lsq_gencon_deriv`` = 0

        The value of i controls the amount of printout produced by the
        major iterations of ``lsq_gencon_deriv``, as indicated below.
        A detailed description of the printed output is given in
        [Further Comments] (summary output at each major iteration and
        the final solution) and [Monitoring Information] (monitoring
        information at each major iteration). (See also the description
        of the option 'Minor Print Level'.)

        The following printout is sent to the file object associated
        with the advisory I/O unit (see
        :class:`~naginterfaces.base.utils.FileObjManager`):

        [table omitted]

        The following printout is sent to the unit number given by the
        option 'Monitoring File':

        [table omitted]

        If 'Major Print Level' >= 5 and the unit number defined by the
        option 'Monitoring File' is the advisory unit number, the
        summary output for each major iteration is suppressed.

    'Minor Iteration Limit' : int
        Default = max(50, 3*(n+n_L+n_N))

        The value of i specifies the maximum number of iterations for
        finding a feasible point with respect to the bounds and linear
        constraints (if any).
        The value of i also specifies the maximum number of minor
        iterations for the optimality phase of each QP subproblem.
        If i <= 0, the default value is used.

    'Minor Print Level' : int
        Default = 0

        The value of i controls the amount of printout produced by the
        minor iterations of ``lsq_gencon_deriv`` (i.e., the iterations
        of the quadratic programming algorithm), as indicated below.
        A detailed description of the printed output is given in
        [Further Comments] (summary output at each minor iteration and
        the final QP solution) and [Monitoring Information] (monitoring
        information at each minor iteration). (See also the description
        of the option 'Major Print Level'.)

        The following printout is sent to the file object associated
        with the advisory I/O unit (see
        :class:`~naginterfaces.base.utils.FileObjManager`):

        [table omitted]

        The following printout is sent to the unit number given by the
        option 'Monitoring File':

        [table omitted]

        If 'Major Print Level' >= 5 and the unit number defined by the
        option 'Monitoring File' is the advisory unit number, the
        summary output for each major iteration is suppressed.

    'Monitoring File' : int
        Default = -1

        If i >= 0 and 'Major Print Level' >= 5 or i >= 0 and 'Minor
        Print Level' >= 5, monitoring information produced by
        ``lsq_gencon_deriv`` at every iteration is sent to a file with
        logical unit number i.
        If i < 0 and/or 'Major Print Level' < 5 and 'Minor Print Level'
        < 5, no monitoring information is produced.

    'Optimality Tolerance' : float
        Default = epsilon_R^0.8

        The argument r (epsilon_R <= r < 1) specifies the accuracy to
        which you wish the final iterate to approximate a solution of
        the problem.
        Broadly speaking, r indicates the number of correct figures
        desired in the objective function at the solution.
        For example, if r is 10^(-6) and ``lsq_gencon_deriv`` terminates
        successfully, the final value of F should have approximately six
        correct figures.
        If r < epsilon_R or r >= 1, the default value is used.

        ``lsq_gencon_deriv`` will terminate successfully if the
        iterative sequence of x values is judged to have converged and
        the final point satisfies the first-order Kuhn--Tucker
        conditions (see [Overview]).
        The sequence of iterates is considered to have converged at x if

            alpha||p|| <= sqrt(r)(1+||x||),

        where p is the search direction and alpha the step length.
        An iterate is considered to satisfy the first-order conditions
        for a minimum if

            || Z^Tg_FR|| <= sqrt(r)(1+max(1+|F(x)|,||g_FR||))

        and

            |res_j| <= ftol for all j,

        where Z^Tg_FR is the projected gradient, g_FR is the gradient of
        F(x) with respect to the free variables, res_j is the violation
        of the jth active nonlinear constraint, and ftol is the
        'Nonlinear Feasibility Tolerance'.

    'Reset Frequency' : int
        Default = 2

        If i > 0, this argument allows you to reset the approximate
        Hessian matrix to J^TJ every i iterations, where J is the
        objective Jacobian matrix nabla f(x) (see also the description
        of the option 'JTJ Initial Hessian').

        At any point where there are no nonlinear constraints active and
        the values of f are small in magnitude compared to the norm of
        J, J^TJ will be a good approximation to the objective Hessian
        nabla^2F(x).
        Under these circumstances, frequent resetting can significantly
        improve the convergence rate of ``lsq_gencon_deriv``.

        Resetting is suppressed at any iteration during which there are
        nonlinear constraints active.

        If i <= 0, the default value is used.

    'Start Objective Check At Variable' : int
        Default = 1

        These keywords take effect only if 'Verify Level' > 0.
        They may be used to control the verification of Jacobian
        elements computed by functions `objfun` and `confun`.
        For example, if the first 30 columns of the objective Jacobian
        appeared to be correct in an earlier run, so that only column 31
        remains questionable, it is reasonable to specify 'Start
        Objective Check At Variable' = 31.
        If the first 30 variables appear linearly in the subfunctions,
        so that the corresponding Jacobian elements are constant, the
        above choice would also be appropriate.

        If i_{2*m-1} <= 0 or i_{2*m-1} > min(n, i_{2m}), the default
        value is used, for m = 1,2,...,2.
        If i_{2*m} <= 0 or i_{2*m} > n, the default value is used, for m
        = 1,2,...,2.

    'Stop Objective Check At Variable' : int
        Default = n

        These keywords take effect only if 'Verify Level' > 0.
        They may be used to control the verification of Jacobian
        elements computed by functions `objfun` and `confun`.
        For example, if the first 30 columns of the objective Jacobian
        appeared to be correct in an earlier run, so that only column 31
        remains questionable, it is reasonable to specify 'Start
        Objective Check At Variable' = 31.
        If the first 30 variables appear linearly in the subfunctions,
        so that the corresponding Jacobian elements are constant, the
        above choice would also be appropriate.

        If i_{2*m-1} <= 0 or i_{2*m-1} > min(n, i_{2m}), the default
        value is used, for m = 1,2,...,2.
        If i_{2*m} <= 0 or i_{2*m} > n, the default value is used, for m
        = 1,2,...,2.

    'Start Constraint Check At Variable' : int
        Default = 1

        These keywords take effect only if 'Verify Level' > 0.
        They may be used to control the verification of Jacobian
        elements computed by functions `objfun` and `confun`.
        For example, if the first 30 columns of the objective Jacobian
        appeared to be correct in an earlier run, so that only column 31
        remains questionable, it is reasonable to specify 'Start
        Objective Check At Variable' = 31.
        If the first 30 variables appear linearly in the subfunctions,
        so that the corresponding Jacobian elements are constant, the
        above choice would also be appropriate.

        If i_{2*m-1} <= 0 or i_{2*m-1} > min(n, i_{2m}), the default
        value is used, for m = 1,2,...,2.
        If i_{2*m} <= 0 or i_{2*m} > n, the default value is used, for m
        = 1,2,...,2.

    'Stop Constraint Check At Variable' : int
        Default = n

        These keywords take effect only if 'Verify Level' > 0.
        They may be used to control the verification of Jacobian
        elements computed by functions `objfun` and `confun`.
        For example, if the first 30 columns of the objective Jacobian
        appeared to be correct in an earlier run, so that only column 31
        remains questionable, it is reasonable to specify 'Start
        Objective Check At Variable' = 31.
        If the first 30 variables appear linearly in the subfunctions,
        so that the corresponding Jacobian elements are constant, the
        above choice would also be appropriate.

        If i_{2*m-1} <= 0 or i_{2*m-1} > min(n, i_{2m}), the default
        value is used, for m = 1,2,...,2.
        If i_{2*m} <= 0 or i_{2*m} > n, the default value is used, for m
        = 1,2,...,2.

    'Step Limit' : float
        Default = 2.0

        If r > 0,r specifies the maximum change in variables at the
        first step of the linesearch.
        In some cases, such as F(x) = ae^(bx) or F(x) = ax^b, even a
        moderate change in the elements of x can lead to floating-point
        overflow.
        The argument r is therefore used to encourage evaluation of the
        problem functions at meaningful points.
        Given any major iterate x, the first point x~ at which F and c
        are evaluated during the linesearch is restricted so that

            ||x~-x||_2 <= r(1+||x||_2).

        The linesearch may go on and evaluate F and c at points further
        from x if this will result in a lower value of the merit
        function (indicated by L at the end of each line of output
        produced by the major iterations; see [Further Comments]).
        If L is printed for most of the iterations, r should be set to a
        larger value.

        Wherever possible, upper and lower bounds on x should be used to
        prevent evaluation of nonlinear functions at wild values.
        The default value 'Step Limit' = 2.0 should not affect progress
        on well-behaved functions, but values such as 0.1 or 0.01 may be
        helpful when rapidly varying functions are present.
        If a small value of 'Step Limit' is selected, a good starting
        point may be required.
        An important application is to the class of nonlinear least
        squares problems.
        If r <= 0, the default value is used.

    'Verify Level' : int
        Default = 0

        These keywords refer to finite difference checks on the gradient
        elements computed by `objfun` and `confun`. (Unspecified
        gradient elements are not checked.) The possible choices for i
        are the following:

        [table omitted]

        It is possible to specify 'Verify Level' = 0 to 3 in several
        ways.
        For example, the nonlinear objective gradient (if any) will be
        verified if either 'Verify Objective Gradients' or 'Verify
        Level' = 1 is specified.
        Similarly, the objective and the constraint gradients will be
        verified if 'Verify' = 'YES' or 'Verify Level' = 3 or 'Verify'
        is specified.

        If i = -1, no checking will be performed.

        If 0 <= i <= 3, gradients will be verified at the first point
        that satisfies the linear constraints and bounds.

        If i = 0, only a 'cheap' test will be performed, requiring one
        call to `objfun` and (if appropriate) one call to `confun`.

        If 1 <= i <= 3, a more reliable (but more expensive) check will
        be made on individual gradient elements, within the ranges
        specified by the 'Start Objective Check At Variable' and 'Stop
        Objective Check At Variable' keywords.
        A result of the form OK or BAD? is printed by
        ``lsq_gencon_deriv`` to indicate whether or not each element
        appears to be correct.
        If a gradient element is determined to be extremely poor (i.e.,
        if it appears to have no significant digits of accuracy at all),
        then ``lsq_gencon_deriv`` will also exit with an error indicator
        in argument errno.

        If 10 <= i <= 13, the action is the same as for i-10, except
        that it will take place at the user-specified initial value of
        x.

        If i < -1 or 4 <= i <= 9 or i > 13, the default value is used.

        We suggest that 'Verify Level' = 3 be used whenever a new
        function function is being developed.

    'Verify' : valueless
        These keywords refer to finite difference checks on the gradient
        elements computed by `objfun` and `confun`. (Unspecified
        gradient elements are not checked.) The possible choices for i
        are the following:

        [table omitted]

        It is possible to specify 'Verify Level' = 0 to 3 in several
        ways.
        For example, the nonlinear objective gradient (if any) will be
        verified if either 'Verify Objective Gradients' or 'Verify
        Level' = 1 is specified.
        Similarly, the objective and the constraint gradients will be
        verified if 'Verify' = 'YES' or 'Verify Level' = 3 or 'Verify'
        is specified.

        If i = -1, no checking will be performed.

        If 0 <= i <= 3, gradients will be verified at the first point
        that satisfies the linear constraints and bounds.

        If i = 0, only a 'cheap' test will be performed, requiring one
        call to `objfun` and (if appropriate) one call to `confun`.

        If 1 <= i <= 3, a more reliable (but more expensive) check will
        be made on individual gradient elements, within the ranges
        specified by the 'Start Objective Check At Variable' and 'Stop
        Objective Check At Variable' keywords.
        A result of the form OK or BAD? is printed by
        ``lsq_gencon_deriv`` to indicate whether or not each element
        appears to be correct.
        If a gradient element is determined to be extremely poor (i.e.,
        if it appears to have no significant digits of accuracy at all),
        then ``lsq_gencon_deriv`` will also exit with an error indicator
        in argument errno.

        If 10 <= i <= 13, the action is the same as for i-10, except
        that it will take place at the user-specified initial value of
        x.

        If i < -1 or 4 <= i <= 9 or i > 13, the default value is used.

        We suggest that 'Verify Level' = 3 be used whenever a new
        function function is being developed.

    'Verify Constraint Gradients' : valueless
        These keywords refer to finite difference checks on the gradient
        elements computed by `objfun` and `confun`. (Unspecified
        gradient elements are not checked.) The possible choices for i
        are the following:

        [table omitted]

        It is possible to specify 'Verify Level' = 0 to 3 in several
        ways.
        For example, the nonlinear objective gradient (if any) will be
        verified if either 'Verify Objective Gradients' or 'Verify
        Level' = 1 is specified.
        Similarly, the objective and the constraint gradients will be
        verified if 'Verify' = 'YES' or 'Verify Level' = 3 or 'Verify'
        is specified.

        If i = -1, no checking will be performed.

        If 0 <= i <= 3, gradients will be verified at the first point
        that satisfies the linear constraints and bounds.

        If i = 0, only a 'cheap' test will be performed, requiring one
        call to `objfun` and (if appropriate) one call to `confun`.

        If 1 <= i <= 3, a more reliable (but more expensive) check will
        be made on individual gradient elements, within the ranges
        specified by the 'Start Objective Check At Variable' and 'Stop
        Objective Check At Variable' keywords.
        A result of the form OK or BAD? is printed by
        ``lsq_gencon_deriv`` to indicate whether or not each element
        appears to be correct.
        If a gradient element is determined to be extremely poor (i.e.,
        if it appears to have no significant digits of accuracy at all),
        then ``lsq_gencon_deriv`` will also exit with an error indicator
        in argument errno.

        If 10 <= i <= 13, the action is the same as for i-10, except
        that it will take place at the user-specified initial value of
        x.

        If i < -1 or 4 <= i <= 9 or i > 13, the default value is used.

        We suggest that 'Verify Level' = 3 be used whenever a new
        function function is being developed.

    'Verify Gradients' : valueless
        These keywords refer to finite difference checks on the gradient
        elements computed by `objfun` and `confun`. (Unspecified
        gradient elements are not checked.) The possible choices for i
        are the following:

        [table omitted]

        It is possible to specify 'Verify Level' = 0 to 3 in several
        ways.
        For example, the nonlinear objective gradient (if any) will be
        verified if either 'Verify Objective Gradients' or 'Verify
        Level' = 1 is specified.
        Similarly, the objective and the constraint gradients will be
        verified if 'Verify' = 'YES' or 'Verify Level' = 3 or 'Verify'
        is specified.

        If i = -1, no checking will be performed.

        If 0 <= i <= 3, gradients will be verified at the first point
        that satisfies the linear constraints and bounds.

        If i = 0, only a 'cheap' test will be performed, requiring one
        call to `objfun` and (if appropriate) one call to `confun`.

        If 1 <= i <= 3, a more reliable (but more expensive) check will
        be made on individual gradient elements, within the ranges
        specified by the 'Start Objective Check At Variable' and 'Stop
        Objective Check At Variable' keywords.
        A result of the form OK or BAD? is printed by
        ``lsq_gencon_deriv`` to indicate whether or not each element
        appears to be correct.
        If a gradient element is determined to be extremely poor (i.e.,
        if it appears to have no significant digits of accuracy at all),
        then ``lsq_gencon_deriv`` will also exit with an error indicator
        in argument errno.

        If 10 <= i <= 13, the action is the same as for i-10, except
        that it will take place at the user-specified initial value of
        x.

        If i < -1 or 4 <= i <= 9 or i > 13, the default value is used.

        We suggest that 'Verify Level' = 3 be used whenever a new
        function function is being developed.

    'Verify Objective Gradients' : valueless
        These keywords refer to finite difference checks on the gradient
        elements computed by `objfun` and `confun`. (Unspecified
        gradient elements are not checked.) The possible choices for i
        are the following:

        [table omitted]

        It is possible to specify 'Verify Level' = 0 to 3 in several
        ways.
        For example, the nonlinear objective gradient (if any) will be
        verified if either 'Verify Objective Gradients' or 'Verify
        Level' = 1 is specified.
        Similarly, the objective and the constraint gradients will be
        verified if 'Verify' = 'YES' or 'Verify Level' = 3 or 'Verify'
        is specified.

        If i = -1, no checking will be performed.

        If 0 <= i <= 3, gradients will be verified at the first point
        that satisfies the linear constraints and bounds.

        If i = 0, only a 'cheap' test will be performed, requiring one
        call to `objfun` and (if appropriate) one call to `confun`.

        If 1 <= i <= 3, a more reliable (but more expensive) check will
        be made on individual gradient elements, within the ranges
        specified by the 'Start Objective Check At Variable' and 'Stop
        Objective Check At Variable' keywords.
        A result of the form OK or BAD? is printed by
        ``lsq_gencon_deriv`` to indicate whether or not each element
        appears to be correct.
        If a gradient element is determined to be extremely poor (i.e.,
        if it appears to have no significant digits of accuracy at all),
        then ``lsq_gencon_deriv`` will also exit with an error indicator
        in argument errno.

        If 10 <= i <= 13, the action is the same as for i-10, except
        that it will take place at the user-specified initial value of
        x.

        If i < -1 or 4 <= i <= 9 or i > 13, the default value is used.

        We suggest that 'Verify Level' = 3 be used whenever a new
        function function is being developed.

    Raises
    ------
    NagValueError
        (`errno` 9)
            On entry, m = *<value>*.

            Constraint: m >= 1.

        (`errno` 9)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 9)
            On entry, nclin = *<value>*.

            Constraint: nclin >= 0.

        (`errno` 9)
            On entry, ncnln = *<value>*.

            Constraint: ncnln >= 0.

        (`errno` 9)
            On entry, the equal bounds on *<value>* are infinite,
            because `bl`[*<value>*] = beta and `bu`[*<value>*] = beta,
            but abs(beta) >= bigbnd: beta = *<value>* and bigbnd =
            *<value>*.

        (`errno` 9)
            On entry, the bounds on *<value>* are inconsistent:
            `bl`[*<value>*] = *<value>* and `bu`[*<value>*] = *<value>*.

        (`errno` 9)
            On entry with a Warm Start, `istate`[*<value>*] = *<value>*.

        (`errno` 9)
            On entry, the equal bounds on variable *<value>* are
            infinite, because `bl`[*<value>*] = beta and `bu`[*<value>*]
            = beta, but abs(beta) >= bigbnd: beta = *<value>* and bigbnd
            = *<value>*.

        (`errno` 9)
            On entry, the equal bounds on linear constraint *<value>*
            are infinite, because `bl`[*<value>*] = beta and
            `bu`[*<value>*] = beta, but abs(beta) >= bigbnd: beta =
            *<value>* and bigbnd = *<value>*.

        (`errno` 9)
            On entry, the equal bounds on nonlinear constraint *<value>*
            are infinite, because `bl`[*<value>*] = beta and
            `bu`[*<value>*] = beta, but abs(beta) >= bigbnd: beta =
            *<value>* and bigbnd = *<value>*.

        (`errno` 9)
            On entry, the bounds on variable *<value>* are inconsistent:
            `bl`[*<value>*] = *<value>* and `bu`[*<value>*] = *<value>*.

        (`errno` 9)
            On entry, the bounds on linear constraint *<value>* are
            inconsistent: `bl`[*<value>*] = *<value>* and
            `bu`[*<value>*] = *<value>*.

        (`errno` 9)
            On entry, the bounds on nonlinear constraint *<value>* are
            inconsistent: `bl`[*<value>*] = *<value>* and
            `bu`[*<value>*] = *<value>*.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 1)
            Optimal solution found, but requested accuracy not achieved.

        (`errno` 2)
            No feasible point for the linear constraints.

        (`errno` 3)
            No feasible point for the nonlinear constraints.

        (`errno` 4)
            Too many major iterations.

        (`errno` 6)
            Current point cannot be improved upon.

        (`errno` 7)
            Large errors found in the derivatives.

    NagCallbackTerminateWarning
        (`errno` i < 0)
            User requested termination by setting `mode` negative in
            `objfun` or `confun`.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Gill, P E, Murray, W and Wright, M H, 1981, `Practical
    Optimization`, Academic Press

    Hock, W and Schittkowski, K, 1981, `Test Examples for Nonlinear
    Programming Codes. Lecture Notes in Economics and Mathematical
    Systems` (187), Springer--Verlag

    See Also
    --------
    :meth:`naginterfaces.library.examples.opt.lsq_gencon_deriv_ex.main`
    """

    fun_name = 'naginterfaces.library.opt.lsq_gencon_deriv'

    cb_confun_name = fun_name + ':confun'

    def cb_confun(mode, ncnln, n, needc, x, c, cjac, nstate, data=None):
        """Wrapping user's confun. Called as base confun."""
        if data is not None:
            _c_confun = confun(
                mode,
                needc,
                x,
                cjac,
                nstate,
                data,
            )
        else:
            _c_confun = confun(
                mode,
                needc,
                x,
                cjac,
                nstate,
            )
        c[:] = _utils._to_ndarray(
            {'fun_name': cb_confun_name, 'returned': 'c'},
            _c_confun,
            copy=False,
            exp_shape=(ncnln,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )

    cb_objfun_name = fun_name + ':objfun'

    def cb_objfun(mode, m, n, needfi, x, f, fjac, nstate, data=None):
        """Wrapping user's objfun. Called as base objfun."""
        if data is not None:
            _f_objfun = objfun(
                mode,
                needfi,
                x,
                fjac,
                nstate,
                data,
            )
        else:
            _f_objfun = objfun(
                mode,
                needfi,
                x,
                fjac,
                nstate,
            )
        f[:] = _utils._to_ndarray(
            {'fun_name': cb_objfun_name, 'returned': 'f'},
            _f_objfun,
            copy=False,
            exp_shape=(m,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _cjac = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'cjac'},
        cjac,
        copy=True,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _fjac = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'fjac'},
        fjac,
        copy=True,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _r = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'r'},
        r,
        copy=True,
        allow_none=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_a, 'a'), (_cjac, 'cjac'), (_fjac, 'fjac'), (_r, 'r')),
    )
    _y = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'y'},
        y,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_py = _b_utils._EngineFloat64ArrayType.get_shape(_y, {'fun_name': fun_name, 'entity': 'y'}, exp_rank=1)[0]
    _m = _b_utils._EngineIntScalarType(
        _m_py, {'fun_name': fun_name, 'entity': '_m'},
    ).to_ctypes()
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    if a is None:
        _a = None
    _nclin = _b_utils._EngineIntScalarType(
        0 if a is None else _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, allow_none=True)[0], {'fun_name': fun_name, 'entity': '_nclin'},
    ).to_ctypes()
    _bl = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'bl'},
        bl,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nclin_py = _b_utils._EngineIntScalarType.to_py(_nclin)
    _ncnln = _b_utils._EngineIntScalarType(
        _b_utils._EngineFloat64ArrayType.get_shape(_bl, {'fun_name': fun_name, 'entity': 'bl'}, exp_rank=1)[0]-_n_py-_nclin_py, {'fun_name': fun_name, 'entity': '_ncnln'},
    ).to_ctypes()
    _bu = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'bu'},
        bu,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b_utils._check_callable(confun, {'fun_name': fun_name, 'entity': 'confun'}, allow_none=True)
    _confun = None if confun is None else cb_confun
    _b_utils._check_callable(objfun, {'fun_name': fun_name, 'entity': 'objfun'})
    _objfun = cb_objfun
    _b_utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        req_keys=('lwsav', 'iwsav', 'rwsav'),
    )
    _vl_start_py = _b_utils._EngineIntArrayType.get_element(comm['iwsav'], {'fun_name': fun_name, 'entity': 'comm[\'iwsav\']'}, (400,), validate=False)
    if _vl_start_py in (1, 2):
        _vl_warm_start_py = True
    else:
        _vl_warm_start_py = False
    if _vl_warm_start_py:
        if istate is None:
            raise _b_utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'istate'},
                'must not be None when the option \'Warm Start\' has been chosen.',
            )
    _ncnln_py = _b_utils._EngineIntScalarType.to_py(_ncnln)
    if istate is None:
        _istate = _b_utils._EngineIntArrayType.empty_ndarray(
            (_n_py+_nclin_py+_ncnln_py,),
        )
    else:
        _istate = _utils._to_ndarray(
            {'fun_name': fun_name, 'entity': 'istate'},
            istate,
            copy=True,
        )
    if _ncnln_py != 0:
        _c = _b_utils._EngineFloat64ArrayType.empty_ndarray(
            (_ncnln_py,),
        )
    else:
        _c = None
    if _b_utils._EngineIntArrayType.get_element(comm['iwsav'], {'fun_name': fun_name, 'entity': 'comm[\'iwsav\']'}, (151,), validate=False) == 3 and _ncnln_py != 0:
        if cjac is None:
            raise _b_utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'cjac'},
                'must not be None when \'Derivative Level\' = 3 and ncnln != 0.',
            )
    if cjac is None:
        if _ncnln_py > 0:
            _cjac_dim2 = _n_py
        else:
            _cjac_dim2 = 1
        exp_shape_cjac = (_ncnln_py, _cjac_dim2)
        _cjac = _b_utils._EngineFloat64ArrayType.empty_ndarray(
            exp_shape_cjac,
            sorder=_m_sorder,
        )
    _f = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_m_py,),
    )
    if fjac is None:
        _fjac = _b_utils._EngineFloat64ArrayType.empty_ndarray(
            (_m_py, _n_py),
            sorder=_m_sorder,
        )
    if _vl_warm_start_py:
        if clamda is None:
            raise _b_utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'clamda'},
                'must not be None when the option \'Warm Start\' has been chosen.',
            )
    if clamda is None:
        _clamda = _b_utils._EngineFloat64ArrayType.empty_ndarray(
            (_n_py+_nclin_py+_ncnln_py,),
        )
    else:
        _clamda = _utils._to_ndarray(
            {'fun_name': fun_name, 'entity': 'clamda'},
            clamda,
            copy=True,
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
    if _vl_warm_start_py:
        if r is None:
            raise _b_utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'r'},
                'must not be None when the option \'Warm Start\' has been chosen.',
            )
    if r is None:
        _r = _b_utils._EngineFloat64ArrayType.empty_ndarray(
            (_n_py, _n_py),
            sorder=_m_sorder,
        )
    try:
        _return_scalars = _b_opt.lsq_gencon_deriv(
            _m,
            _n,
            _nclin,
            _ncnln,
            _a,
            _bl,
            _bu,
            _y,
            _confun,
            _objfun,
            _istate,
            _c,
            _cjac,
            _f,
            _fjac,
            _clamda,
            _r,
            _x,
            comm,
            data,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _x_py = _x
    _istate_py = _istate
    if _ncnln_py != 0:
        _cjac_py = _cjac
    else:
        _cjac_py = None
    _fjac_py = _fjac
    _clamda_py = _clamda
    _r_py = _r
    _itera_py = _return_scalars.itera
    if _ncnln_py != 0:
        _c_py = _c
    else:
        _c_py = None
    _f_py = _f
    _objf_py = _return_scalars.objf
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['itera', 'istate', 'c', 'cjac', 'f', 'fjac', 'clamda', 'objf', 'r', 'x'],
    )(
        _itera_py,
        _istate_py,
        _c_py,
        _cjac_py,
        _f_py,
        _fjac_py,
        _clamda_py,
        _objf_py,
        _r_py,
        _x_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def nlp2_sparse_init():
    u"""
    Initialization function for ``nlp2_sparse_solve``.

    ``nlp2_sparse_init`` is used to initialize the function
    ``nlp2_sparse_solve``.

    For full information please refer to the NAG Library document for
    e04vg

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04vgf.html

    Returns
    -------
    comm : dict, communication object
        Communication structure.

    See Also
    --------
    :meth:`naginterfaces.library.examples.opt.nlp2_sparse_solve_ex.main`
    """

    _return_scalar = _b_opt.nlp2_sparse_init(
    )
    _comm_py = _return_scalar
    return _comm_py

def nlp2_sparse_solve(start, objadd, objrow, prob, usrfun, iafun, javar, nea, a, igfun, jgvar, neg, xlow, xupp, xnames, flow, fupp, fnames, x, xstate, f, fstate, fmul, ns, comm, data=None, io_manager=None):
    u"""
    Nonlinear programming (NLP), sparse, active-set SQP method, using
    function values and optionally first derivatives, recommended.

    Note: this function uses optional algorithmic parameters, see also:
    ``nlp2_sparse_option_file``, ``nlp2_sparse_option_string``,
    ``nlp2_sparse_option_integer_set``,
    ``nlp2_sparse_option_double_set``, ``nlp2_sparse_init``,
    ``nlp2_sparse_jacobian``.

    ``nlp2_sparse_solve`` solves sparse linear and nonlinear programming
    problems.

    For full information please refer to the NAG Library document for
    e04vh

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04vhf.html

    Parameters
    ----------
    start : int
        Indicates how a starting point is to be obtained.

        `start` = 0

            Requests that the Crash procedure be used, unless a Basis
            file is provided via options 'Old Basis File', 'Insert File'
            or 'Load File'.

        `start` = 1

            Is the same as `start` = 0 but is more meaningful when a
            Basis file is given.

        `start` = 2

            Means that `xstate` and `fstate` define a valid starting
            point (probably from an earlier call, though not
            necessarily).

    objadd : float
        Is a constant that will be added to the objective row F_obj for
        printing purposes. Typically, `objadd` = 0.0e+0.

    objrow : int
        Says which row of F(x) is to act as the objective function. If
        there is no such row, set `objrow` = 0. Then
        ``nlp2_sparse_solve`` will seek a feasible point such that l_F
        <= F(x) <= u_F and l_x <= x <= u_x.

    prob : str, length 8
        Is an 8-character name for the problem. `prob` is used in the
        printed solution and in some functions that output Basis files.
        A blank name may be used.

    usrfun : callable (status, f, g) = usrfun(status, x, needf, f,
        needg, g, data=None)
        `usrfun` must define the nonlinear portion f(x) of the problem
        functions F(x) = f(x)+Ax, along with its gradient elements
        G_{ij}(x) = df_i(x)/dx_j. (A dummy function is needed even if f
        = 0 and all functions are linear.)

        In general, `usrfun` should return all function and gradient
        values on every entry except perhaps the last.

        This provides maximum reliability and corresponds to the default
        option setting, 'Derivative Option' = 1.

        The elements of G(x) are stored in the array `g`(1:`leng`) in
        the order specified by the input arrays `igfun` and `jgvar`.

        In practice it is often convenient `not` to code gradients.
        ``nlp2_sparse_solve`` is able to estimate them by finite
        differences, using a call to `usrfun` for each variable x_j for
        which some df_i(x)/dx_j needs to be estimated.

        However, this reduces the reliability of the optimization
        algorithm, and it can be very expensive if there are many such
        variables x_j.

        As a compromise, ``nlp2_sparse_solve`` allows you to code `as
        many gradients as you like`.

        This option is implemented as follows.

        Just before `usrfun` is called, each element of the derivative
        array `g` is initialized to a specific value.

        On exit, any element retaining that value must be estimated by
        finite differences.

        Some rules of thumb follow:

        (i) for maximum reliability, compute all gradients;

        (#) if the gradients are expensive to compute, specify option
            'Nonderivative Linesearch' and use the value of the input
            argument `needg` to avoid computing them on certain entries.
            (There is no need to compute gradients if `needg` = 0 on
            entry to `usrfun`.);

        (#) if not all gradients are known, you must specify 'Derivative
            Option' = 0. You should still compute as many gradients as
            you can. (It often happens that some of them are constant or
            zero.);

        (#) again, if the known gradients are expensive, don't compute
            them if `needg` = 0 on entry to `usrfun`;

        (#) use the input argument `status` to test for special actions
            on the first or last entries;

        (#) while `usrfun` is being developed, use the option 'Verify
            Level' to check the computation of gradients that are
            supposedly known;

        (#) `usrfun` is not called until the linear constraints and
            bounds on x are satisfied. This helps confine x to regions
            where the functions f_i(x) are likely to be defined.
            However, be aware of the option 'Minor Feasibility
            Tolerance' if the functions have singularities on the
            constraint boundaries;

        (#) set `status` = -1 if some of the functions are undefined.
            The linesearch will shorten the step and try again;

        (#) set `status` <= -2 if you want ``nlp2_sparse_solve`` to
            stop.

        Parameters
        ~~~~~~~~~~
        status : int
            Indicates the first and last calls to `usrfun`.

            `status` = 0

                There is nothing special about the current call to
                `usrfun`.

            `status` = 1

                ``nlp2_sparse_solve`` is calling your function for the
                `first` time. You may wish to do something special such
                as read data from a file.

            `status` >= 2

                ``nlp2_sparse_solve`` is calling your function for the
                `last` time. This argument setting allows you to perform
                some additional computation on the final solution.

                `status` = 2

                    The current `x` is `optimal`.

                `status` = 3

                    The problem appears to be infeasible.

                `status` = 4

                    The problem appears to be unbounded.

                `status` = 5

                    An iterations limit was reached.

            If the functions are expensive to evaluate, it may be
            desirable to do nothing on the last call.

        x : float, ndarray, shape (n)
            The variables x at which the problem functions are to be
            calculated. The array x must not be altered.

        needf : int
            Indicates whether `f` must be assigned during this call of
            `usrfun`.

            `needf` = 0

                `f` is not required and is ignored.

            `needf` > 0

                The components of f(x) corresponding to the nonlinear
                part of F(x) must be calculated and assigned to `f`.

                If F_i(x) is linear and completely defined by the ith
                row of A, A'_i, the associated value f_i(x) is ignored
                and need not be assigned.

                However, if F_i(x) has a nonlinear portion f_i(x) that
                happens to be zero at x, it is still necessary to set
                f_i(x) = 0.

                If the linear part A'_i of a nonlinear F_i(x) is
                provided using the arrays `iafun`, `javar` and `a`, it
                must not be computed again as part of f_i(x).

            To simplify the code, you may ignore the value of `needf`
            and compute f(x) on every entry to `usrfun`.

            `needf` may also be ignored with 'Derivative Linesearch' and
            'Derivative Option' = 1.

            In this case, `needf` is always 1, and `f` must always be
            assigned.

        f : float, ndarray, shape (nf)
            Concerns the calculation of f(x).

        needg : int
            Indicates whether `g` must be assigned during this call of
            `usrfun`.

            `needg` = 0

                `g` is not required and is ignored.

            `needg` > 0

                The partial derivatives of f(x) must be calculated and
                assigned to `g`. The value of `g`[k-1] should be
                df_i(x)/dx_j, where i = `igfun`[k-1], j = `jgvar`[k-1]
                and k = 1,2,...,leng.

        g : float, ndarray, shape (leng)
            Concerns the calculations of the derivatives of the function
            f(x).

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        status : int
            May be used to indicate that you are unable to evaluate f or
            its gradients at the current x. (For example, the problem
            functions may not be defined there.)

            During the linesearch, f(x) is evaluated at points x =
            x_k+alpha p_k for various step lengths alpha, where f(x_k)
            has already been evaluated satisfactorily.

            For any such x, if you set `status` = -1,
            ``nlp2_sparse_solve`` will reduce alpha and evaluate f again
            (closer to x_k, where f(x_k) is more likely to be defined).

            If for some reason you wish to terminate the current
            problem, set `status` <= -2.

        f : float, array-like, shape (nf)
            `f` contains the computed functions f(x) (except perhaps if
            `needf` = 0).

        g : float, array-like, shape (leng)
            Contains the computed derivatives G(x) (unless `needg` = 0).

            These derivative elements must be stored in `g` in exactly
            the same positions as implied by the definitions of arrays
            `igfun` and `jgvar`.

            There is no internal check for consistency (except
            indirectly via the option 'Verify Level'), so great care is
            essential.

    iafun : int, array-like, shape (lena)
        `iafun` and `javar` define the coordinates (i,j) of the nonzero
        elements of the linear part A of the function F(x) = f(x)+Ax.

        In particular, `nea` triples
        (`iafun`[k-1],`javar`[k-1],`a`[k-1]) define the row and column
        indices i = `iafun`[k-1] and j = `javar`[k-1] of the element
        A_{ij} = `a`[k-1].

        The coordinates may define the elements of A in any order.

    javar : int, array-like, shape (lena)
        `iafun` and `javar` define the coordinates (i,j) of the nonzero
        elements of the linear part A of the function F(x) = f(x)+Ax.

        In particular, `nea` triples
        (`iafun`[k-1],`javar`[k-1],`a`[k-1]) define the row and column
        indices i = `iafun`[k-1] and j = `javar`[k-1] of the element
        A_{ij} = `a`[k-1].

        The coordinates may define the elements of A in any order.

    nea : int
        Is the number of nonzero entries in A such that F(x) = f(x)+Ax.

    a : float, array-like, shape (lena)
        Defines the values A_{ij} of the nonzero elements of the linear
        part A of the function F(x) = f(x)+Ax.

        In particular, `nea` triples
        (`iafun`[k-1],`javar`[k-1],`a`[k-1]) define the row and column
        indices i = `iafun`[k-1] and j = `javar`[k-1] of the element
        A_{ij} = `a`[k-1].

        The coordinates may define the elements of A in any order.

    igfun : int, array-like, shape (leng)
        `igfun` and `jgvar` define the coordinates (i,j) of the nonzero
        elements of G, the nonlinear part of the derivative J(x) =
        G(x)+A of the function F(x) = f(x)+Ax. ``nlp2_sparse_jacobian``
        may be used to define these two arrays.

        The coordinates can define the elements of G in any order.

        However, `usrfun` must define the actual elements of `g` in
        exactly the same order as defined by the coordinates
        (`igfun`,`jgvar`).

    jgvar : int, array-like, shape (leng)
        `igfun` and `jgvar` define the coordinates (i,j) of the nonzero
        elements of G, the nonlinear part of the derivative J(x) =
        G(x)+A of the function F(x) = f(x)+Ax. ``nlp2_sparse_jacobian``
        may be used to define these two arrays.

        The coordinates can define the elements of G in any order.

        However, `usrfun` must define the actual elements of `g` in
        exactly the same order as defined by the coordinates
        (`igfun`,`jgvar`).

    neg : int
        The number of nonzero entries in G.

    xlow : float, array-like, shape (n)
        Contain the lower and upper bounds l_x and u_x on the variables
        x.

        To specify a nonexistent lower bound [l_x]_j = - infinity, set
        `xlow`[j-1] <= -bigbnd, where bigbnd is the option 'Infinite
        Bound Size'.

        To specify a nonexistent upper bound [u_x]_j = infinity, set
        `xupp`[j-1] >= bigbnd.

        To fix the jth variable at x_j = beta, where |beta| < bigbnd,
        set `xlow`[j-1] = `xupp`[j-1] = beta.

    xupp : float, array-like, shape (n)
        Contain the lower and upper bounds l_x and u_x on the variables
        x.

        To specify a nonexistent lower bound [l_x]_j = - infinity, set
        `xlow`[j-1] <= -bigbnd, where bigbnd is the option 'Infinite
        Bound Size'.

        To specify a nonexistent upper bound [u_x]_j = infinity, set
        `xupp`[j-1] >= bigbnd.

        To fix the jth variable at x_j = beta, where |beta| < bigbnd,
        set `xlow`[j-1] = `xupp`[j-1] = beta.

    xnames : str, length 8, array-like, shape (nxname)
        The optional names for the variables.

        If nxname = 1, `xnames` is not referenced and default names will
        be used for output.

        If nxname = n, `xnames`[j-1] should contain the 8-character name
        of the jth variable.

    flow : float, array-like, shape (nf)
        Contain the lower and upper bounds l_F and u_F on F(x).

        To specify a nonexistent lower bound [l_F]_i = - infinity, set
        `flow`[i-1] <= -bigbnd.

        For a nonexistent upper bound [u_F]_i = infinity, set
        `fupp`[i-1] >= bigbnd.

        To make the ith constraint an `equality` at F_i = beta, where
        |beta| < bigbnd, set `flow`[i-1] = `fupp`[i-1] = beta.

    fupp : float, array-like, shape (nf)
        Contain the lower and upper bounds l_F and u_F on F(x).

        To specify a nonexistent lower bound [l_F]_i = - infinity, set
        `flow`[i-1] <= -bigbnd.

        For a nonexistent upper bound [u_F]_i = infinity, set
        `fupp`[i-1] >= bigbnd.

        To make the ith constraint an `equality` at F_i = beta, where
        |beta| < bigbnd, set `flow`[i-1] = `fupp`[i-1] = beta.

    fnames : str, length 8, array-like, shape (nfname)
        The optional names for the problem functions.

        If nfname = 1, `fnames` is not referenced and default names will
        be used for output.

        If nfname = nf, `fnames`[i-1] should contain the 8-character
        name of the ith row of F.

    x : float, array-like, shape (n)
        An initial estimate of the variables x. See the following
        description of `xstate`.

    xstate : int, array-like, shape (n)
        The initial state for each variable x.

        If `start` = 0 or 1 and no basis information is provided (the
        options 'Old Basis File', 'Insert File' and 'Load File' are all
        set to 0; the default) `x` and `xstate` must be defined.

        If nothing special is known about the problem, or if there is no
        wish to provide special information, you may set `x`[j-1] = 0.0,
        `xstate`[j-1] = 0, for all j = 1,2,...,n.

        If you set `x`[j-1] = `xlow`[j-1] set `xstate`[j-1] = 4; if you
        set `x`[j-1] = `xupp`[j-1], then set `xstate`[j-1] = 5.

        In this case a Crash procedure is used to select an initial
        basis.

        If `start` = 0 or 1 and basis information is provided (at least
        one of the options 'Old Basis File', 'Insert File' and 'Load
        File' is nonzero) `x` and `xstate` need not be set.

        If `start` = 2 (Warm Start), `x` and `xstate` must be set
        (probably from a previous call).

        In this case `xstate`[j-1] must be 0, 1, 2 or 3, for j =
        1,2,...,n.

    f : float, array-like, shape (nf)
        An initial value for the problem functions F. See the following
        description of `fstate`.

    fstate : int, array-like, shape (nf)
        The initial state for the problem functions F.

        If `start` = 0 or 1 and no basis information is provided (the
        options 'Old Basis File', 'Insert File' and 'Load File' are all
        set to 0; the default, `f` and `fstate` must be defined.

        If nothing special is known about the problem, or if there is no
        wish to provide special information, you may set `f`[i-1] = 0.0,
        `fstate`[i-1] = 0, for all i = 1,2,...,nf.

        Less trivially, to say that the optimal value of function
        `f`[i-1] will probably be equal to one of its bounds, set
        `f`[i-1] = `flow`[i-1] and `fstate`[i-1] = 4 or `f`[i-1] =
        `fupp`[i-1] and `fstate`[i-1] = 5 as appropriate.

        In this case a Crash procedure is used to select an initial
        basis.

        If `start` = 0 or 1 and basis information is provided (at least
        one of the options 'Old Basis File', 'Insert File' and 'Load
        File' is nonzero), `f` and `fstate` need not be set.

        If `start` = 2 (Warm Start), `f` and `fstate` must be set
        (probably from a previous call).

        In this case `fstate`[i-1] must be 0, 1, 2 or 3, for i =
        1,2,...,nf.

    fmul : float, array-like, shape (nf)
        An estimate of gamma, the vector of Lagrange multipliers (shadow
        prices) for the constraints l_F <= F(x) <= u_F. All nf
        components must be defined. If nothing is known about gamma, set
        `fmul`[i-1] = 0.0, for i = 1,2,...,nf. For warm start use the
        values from a previous call.

    ns : int
        The number of superbasic variables. `ns` need not be specified
        for cold starts, but should retain its value from a previous
        call when warm start is used.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``nlp2_sparse_init``.

    data : arbitrary, optional
        User-communication data for callback functions.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    x : float, ndarray, shape (n)
        The final values of the variable x.

    xstate : int, ndarray, shape (n)
        The final state of the variables.

        [table omitted]

        Basic and superbasic variables may be outside their bounds by as
        much as the option 'Minor Feasibility Tolerance'.

        Note that if scaling is specified, the feasibility tolerance
        applies to the variables of the `scaled` problem.

        In this case, the variables of the original problem may be as
        much as 0.1 outside their bounds, but this is unlikely unless
        the problem is very badly scaled.

        Check the value of Primal infeasibility output to the unit
        number (see
        :meth:`~naginterfaces.base.utils.FileObjManager.unit_from_fileobj`)
        associated with the option 'Print File'.

        Very occasionally some nonbasic variables may be outside their
        bounds by as much as the option 'Minor Feasibility Tolerance',
        and there may be some nonbasics for which `x`[j-1] lies strictly
        between its bounds.

        If `ninf` > 0, some basic and superbasic variables may be
        outside their bounds by an arbitrary amount (bounded by `sinf`
        if scaling was not used).

    xmul : float, ndarray, shape (n)
        The vector of the dual variables (Lagrange multipliers) for the
        simple bounds l_x <= x <= u_x.

    f : float, ndarray, shape (nf)
        The final values for the problem functions F (the values F at
        the final point `x`).

    fstate : int, ndarray, shape (nf)
        The final state of the variables. The elements of `fstate` have
        the following meaning:

        [table omitted]

        Basic and superbasic slack variables may lead to the
        corresponding functions being outside their bounds by as much as
        the option 'Minor Feasibility Tolerance'.

        Very occasionally some functions may be outside their bounds by
        as much as the option 'Minor Feasibility Tolerance', and there
        may be some nonbasics for which `f`[i-1] lies strictly between
        its bounds.

        If `ninf` > 0, some basic and superbasic variables may be
        outside their bounds by an arbitrary amount (bounded by `sinf`
        if scaling was not used).

    fmul : float, ndarray, shape (nf)
        The vector of the dual variables (Lagrange multipliers) for the
        general constraints l_F <= F(x) <= u_F

    ns : int
        The final number of superbasic variables.

    ninf : int
        Are the number and the sum of the infeasibilities of constraints
        that lie outside one of their bounds by more than the option
        'Minor Feasibility Tolerance' `before the solution is unscaled`.

        If any `linear` constraints are infeasible, x minimizes the sum
        of the infeasibilities of the linear constraints subject to the
        upper and lower bounds being satisfied.

        In this case `ninf` gives the number of variables and linear
        constraints lying outside their upper or lower bounds.

        The nonlinear constraints are not evaluated.

        Otherwise, x minimizes the sum of infeasibilities of the
        `nonlinear` constraints subject to the linear constraints and
        upper and lower bounds being satisfied.

        In this case `ninf` gives the number of components of F(x) lying
        outside their bounds by more than the option 'Minor Feasibility
        Tolerance'.

        Again this is `before the solution is unscaled`.

    sinf : float
        Are the number and the sum of the infeasibilities of constraints
        that lie outside one of their bounds by more than the option
        'Minor Feasibility Tolerance' `before the solution is unscaled`.

        If any `linear` constraints are infeasible, x minimizes the sum
        of the infeasibilities of the linear constraints subject to the
        upper and lower bounds being satisfied.

        In this case `ninf` gives the number of variables and linear
        constraints lying outside their upper or lower bounds.

        The nonlinear constraints are not evaluated.

        Otherwise, x minimizes the sum of infeasibilities of the
        `nonlinear` constraints subject to the linear constraints and
        upper and lower bounds being satisfied.

        In this case `ninf` gives the number of components of F(x) lying
        outside their bounds by more than the option 'Minor Feasibility
        Tolerance'.

        Again this is `before the solution is unscaled`.

    Other Parameters
    ----------------
    'Central Difference Interval' : float
        Default = epsilon_r^1/3

        When 'Derivative Option' = 0, the central-difference interval r
        is used near an optimal solution to obtain more accurate (but
        more expensive) estimates of gradients.
        Twice as many function evaluations are required compared to
        forward differencing.
        The interval used for the jth variable is h_j = r(1+|x_j|).
        The resulting derivative estimates should be accurate to O(r^2),
        unless the functions are badly scaled.

        If you supply a value for this option, a small value between 0.0
        and 1.0 is appropriate.

    'Check Frequency' : int
        Default = 60

        Every ith minor iteration after the most recent basis
        factorization, a numerical test is made to see if the current
        solution x satisfies the general linear constraints (the linear
        constraints and the linearized nonlinear constraints, if any).
        The constraints are of the form Ax-s = b, where s is the set of
        slack variables.
        To perform the numerical test, the residual vector r = b-Ax+s is
        computed.
        If the largest component of r is judged to be too large, the
        current basis is refactorized and the basic variables are
        recomputed to satisfy the general constraints more accurately.
        If i <= 0, the value of i = 99999999 is used and effectively no
        checks are made.

        'Check Frequency' = 1 is useful for debugging purposes, but
        otherwise this option should not be needed.

    'Crash Option' : int
        Default = 3

        Except on restarts, an internal Crash procedure is used to
        select an initial basis from certain rows and columns of the
        constraint matrix (A -I).
        The 'Crash Option' i determines which rows and columns of A are
        eligible initially, and how many times the Crash procedure is
        called.
        Columns of -I are used to pad the basis where necessary.

        [table omitted]

        If i >= 1, certain slacks on inequality rows are selected for
        the basis first. (If i >= 2, numerical values are used to
        exclude slacks that are close to a bound).
        The Crash procedure then makes several passes through the
        columns of A, searching for a basis matrix that is essentially
        triangular.
        A column is assigned to 'pivot' on a particular row if the
        column contains a suitably large element in a row that has not
        yet been assigned. (The pivot elements ultimately form the
        diagonals of the triangular basis.) For remaining unassigned
        rows, slack variables are inserted to complete the basis.

        The 'Crash Tolerance' r allows the starting Crash procedure to
        ignore certain 'small' nonzeros in each column of A.
        If a_max is the largest element in column j, other nonzeros of
        a_{ij} in the columns are ignored if |a_{ij}| <= a_max*r. (To be
        meaningful, r should be in the range 0 <= r < 1.)

        When r > 0.0, the basis obtained by the Crash procedure may not
        be strictly triangular, but it is likely to be nonsingular and
        almost triangular.
        The intention is to obtain a starting basis containing more
        columns of A and fewer (arbitrary) slacks.
        A feasible solution may be reached sooner on some problems.

        For example, suppose the first m columns of A form the matrix
        shown under 'LU Factor Tolerance'; i.e., a tridiagonal matrix
        with entries -1, 4, -1.
        To help the Crash procedure choose all m columns for the initial
        basis, we would specify a 'Crash Tolerance' of r for some value
        of r > 0.5.

    'Crash Tolerance' : float
        Default = 0.1

        Except on restarts, an internal Crash procedure is used to
        select an initial basis from certain rows and columns of the
        constraint matrix (A -I).
        The 'Crash Option' i determines which rows and columns of A are
        eligible initially, and how many times the Crash procedure is
        called.
        Columns of -I are used to pad the basis where necessary.

        [table omitted]

        If i >= 1, certain slacks on inequality rows are selected for
        the basis first. (If i >= 2, numerical values are used to
        exclude slacks that are close to a bound).
        The Crash procedure then makes several passes through the
        columns of A, searching for a basis matrix that is essentially
        triangular.
        A column is assigned to 'pivot' on a particular row if the
        column contains a suitably large element in a row that has not
        yet been assigned. (The pivot elements ultimately form the
        diagonals of the triangular basis.) For remaining unassigned
        rows, slack variables are inserted to complete the basis.

        The 'Crash Tolerance' r allows the starting Crash procedure to
        ignore certain 'small' nonzeros in each column of A.
        If a_max is the largest element in column j, other nonzeros of
        a_{ij} in the columns are ignored if |a_{ij}| <= a_max*r. (To be
        meaningful, r should be in the range 0 <= r < 1.)

        When r > 0.0, the basis obtained by the Crash procedure may not
        be strictly triangular, but it is likely to be nonsingular and
        almost triangular.
        The intention is to obtain a starting basis containing more
        columns of A and fewer (arbitrary) slacks.
        A feasible solution may be reached sooner on some problems.

        For example, suppose the first m columns of A form the matrix
        shown under 'LU Factor Tolerance'; i.e., a tridiagonal matrix
        with entries -1, 4, -1.
        To help the Crash procedure choose all m columns for the initial
        basis, we would specify a 'Crash Tolerance' of r for some value
        of r > 0.5.

    'Defaults' : valueless
        This special keyword may be used to reset all options to their
        default values.

    'Derivative Option' : int
        Default = 1

        Option 'Derivative Option' specifies which nonlinear function
        gradients are known analytically and will be supplied to
        ``nlp2_sparse_solve`` by `usrfun`.

        [table omitted]

        The value i = 1 should be used whenever possible.
        It is the most reliable and will usually be the most efficient.

        If i = 0, ``nlp2_sparse_solve`` will `estimate` the missing
        components of G(x) using finite differences.
        This may simplify the coding of `usrfun`.
        However, it could increase the total run-time substantially
        (since a special call to `usrfun` is required for each column of
        the Jacobian that has a missing element), and there is less
        assurance that an acceptable solution will be located.
        If the nonlinear variables are not well scaled, it may be
        necessary to specify a nonstandard option 'Difference Interval'.

        For each column of the Jacobian, one call to `usrfun` is needed
        to estimate all missing elements in that column, if any.

        At times, central differences are used rather than forward
        differences.
        Twice as many calls to `usrfun` are needed. (This is not under
        your control.)

    'Derivative Linesearch' : valueless
        Default

        At each major iteration a linesearch is used to improve the
        merit function. Option 'Derivative Linesearch' uses safeguarded
        cubic interpolation and requires both function and gradient
        values to compute estimates of the step alpha_k.
        If some analytic derivatives are not provided, or option
        'Nonderivative Linesearch' is specified, ``nlp2_sparse_solve``
        employs a linesearch based upon safeguarded quadratic
        interpolation, which does not require gradient evaluations.

        A nonderivative linesearch can be slightly less robust on
        difficult problems, and it is recommended that the default be
        used if the functions and derivatives can be computed at
        approximately the same cost.
        If the gradients are very expensive relative to the functions, a
        nonderivative linesearch may give a significant decrease in
        computation time.

        If 'Nonderivative Linesearch' is selected, ``nlp2_sparse_solve``
        signals the evaluation of the linesearch by calling `usrfun`
        with `needg` = 0.
        Once the linesearch is completed, the problem functions are
        called again with `needf` = 0 and `needg` = 0.
        If the potential saving provided by a nonderivative linesearch
        is to be realised, it is essential that `usrfun` be coded so
        that derivatives are not computed when `needg` = 0.

    'Nonderivative Linesearch' : valueless
        At each major iteration a linesearch is used to improve the
        merit function. Option 'Derivative Linesearch' uses safeguarded
        cubic interpolation and requires both function and gradient
        values to compute estimates of the step alpha_k.
        If some analytic derivatives are not provided, or option
        'Nonderivative Linesearch' is specified, ``nlp2_sparse_solve``
        employs a linesearch based upon safeguarded quadratic
        interpolation, which does not require gradient evaluations.

        A nonderivative linesearch can be slightly less robust on
        difficult problems, and it is recommended that the default be
        used if the functions and derivatives can be computed at
        approximately the same cost.
        If the gradients are very expensive relative to the functions, a
        nonderivative linesearch may give a significant decrease in
        computation time.

        If 'Nonderivative Linesearch' is selected, ``nlp2_sparse_solve``
        signals the evaluation of the linesearch by calling `usrfun`
        with `needg` = 0.
        Once the linesearch is completed, the problem functions are
        called again with `needf` = 0 and `needg` = 0.
        If the potential saving provided by a nonderivative linesearch
        is to be realised, it is essential that `usrfun` be coded so
        that derivatives are not computed when `needg` = 0.

    'Difference Interval' : float
        Default = sqrt(epsilon_r)

        This alters the interval r used to estimate gradients by forward
        differences.
        It does so in the following circumstances:

        -   in the interval ('cheap') phase of verifying the problem
            derivatives;

        -   for verifying the problem derivatives;

        -   for estimating missing derivatives.

        In all cases, a derivative with respect to x_j is estimated by
        perturbing that component of x to the value x_j+r(1+|x_j|), and
        then evaluating F_obj(x) or f(x) at the perturbed point.
        The resulting gradient estimates should be accurate to O(r)
        unless the functions are badly scaled.
        Judicious alteration of r may sometimes lead to greater
        accuracy.

        If you supply a value for this option, a small value between 0.0
        and 1.0 is appropriate.

    'Dump File' : int
        Default = 0

        Options 'Dump File' and 'Load File' are similar to options
        'Punch File' and 'Insert File', but they record solution
        information in a manner that is more direct and more easily
        modified.
        A full description of information recorded in options 'Dump
        File' and 'Load File' is given in Gill `et al.` (2005a).

        If i_1 > 0, the last solution obtained will be output to the
        file with unit number i_1.

        If i_2 > 0, the 'Load File', containing basis information, will
        be read.
        The file will usually have been output previously as a 'Dump
        File'.
        The file will not be accessed if options 'Old Basis File' or
        'Insert File' are specified.

    'Load File' : int
        Default = 0

        Options 'Dump File' and 'Load File' are similar to options
        'Punch File' and 'Insert File', but they record solution
        information in a manner that is more direct and more easily
        modified.
        A full description of information recorded in options 'Dump
        File' and 'Load File' is given in Gill `et al.` (2005a).

        If i_1 > 0, the last solution obtained will be output to the
        file with unit number i_1.

        If i_2 > 0, the 'Load File', containing basis information, will
        be read.
        The file will usually have been output previously as a 'Dump
        File'.
        The file will not be accessed if options 'Old Basis File' or
        'Insert File' are specified.

    'Elastic Weight' : float
        Default = 10^4

        This keyword determines the initial weight gamma associated with
        the problem (1) (see [Treatment of Constraint Infeasibilities]).

        At major iteration k, if elastic mode has not yet started, a
        scale factor sigma_k = 1+||g(x_k)||_infinity is defined from the
        current objective gradient.
        Elastic mode is then started if the QP subproblem is infeasible,
        or the QP dual variables are larger in magnitude than sigma_k r.
        The QP is resolved in elastic mode with gamma = sigma_k r.

        Thereafter, major iterations continue in elastic mode until they
        converge to a point that is optimal for (1) (see [Treatment of
        Constraint Infeasibilities]).
        If the point is feasible for equation (1) (v = w = 0), it is
        declared locally optimal.
        Otherwise, gamma is increased by a factor of 10 and major
        iterations continue.
        If gamma has already reached a maximum allowable value, equation
        (1) is declared locally infeasible.

    'Expand Frequency' : int
        Default = 10000

        This option is part of the anti-cycling procedure designed to
        make progress even on highly degenerate problems.

        For linear models, the strategy is to force a positive step at
        every iteration, at the expense of violating the bounds on the
        variables by a small amount.
        Suppose that the option 'Minor Feasibility Tolerance' is delta.
        Over a period of i iterations, the tolerance actually used by
        ``nlp2_sparse_solve`` increases from 0.5 delta to delta (in
        steps of 0.5 delta/i).

        For nonlinear models, the same procedure is used for iterations
        in which there is only one superbasic variable. (Cycling can
        occur only when the current solution is at a vertex of the
        feasible region.) Thus, zero steps are allowed if there is more
        than one superbasic variable, but otherwise positive steps are
        enforced.

        Increasing i helps reduce the number of slightly infeasible
        nonbasic variables (most of which are eliminated during a
        resetting procedure).
        However, it also diminishes the freedom to choose a large pivot
        element (see option 'Pivot Tolerance').

    'Factorization Frequency' : int
        Default = 50

        At most i basis changes will occur between factorizations of the
        basis matrix.

        With linear programs, the basis factors are usually updated
        every iteration.
        The default i is reasonable for typical problems.
        Higher values up to i = 100 (say) may be more efficient on
        well-scaled problems.

        When the objective function is nonlinear, fewer basis updates
        will occur as an optimum is approached.
        The number of iterations between basis factorizations will
        therefore increase.
        During these iterations a test is made regularly (according to
        the option 'Check Frequency') to ensure that the general
        constraints are satisfied.
        If necessary the basis will be refactorized before the limit of
        i updates is reached.

    'Function Precision' : float
        Default = epsilon^0.8

        The `relative function precision` epsilon_r is intended to be a
        measure of the relative accuracy with which the nonlinear
        functions can be computed.
        For example, if f(x) is computed as 1000.56789 for some relevant
        x and if the first 6 significant digits are known to be correct,
        the appropriate value for epsilon_r would be 1.0e-6.

        Ideally the functions f_i(x) should have magnitude of order 1.
        If all functions are substantially `less` than 1 in magnitude,
        epsilon_r should be the `absolute` precision.
        For example, if f(x) = 1.23456789e-4 at some point and if the
        first 6 significant digits are known to be correct, the
        appropriate value for epsilon_r would be 1.0e-10.)

        The default value of epsilon_r is appropriate for simple
        analytic functions.

        In some cases the function values will be the result of
        extensive computation, possibly involving a costly iterative
        procedure that can provide few digits of precision.
        Specifying an appropriate 'Function Precision' may lead to
        savings, by allowing the linesearch procedure to terminate when
        the difference between function values along the search
        direction becomes as small as the absolute error in the values.

    'Hessian Full Memory' : valueless
        Default if n_1 <= 75

        These options select the method for storing and updating the
        approximate Hessian. (``nlp2_sparse_solve`` uses a quasi-Newton
        approximation to the Hessian of the Lagrangian.
        A BFGS update is applied after each major iteration.)

        If 'Hessian Full Memory' is specified, the approximate Hessian
        is treated as a dense matrix and the BFGS updates are applied
        explicitly.
        This option is most efficient when the number of variables n is
        not too large (say, less than 75).
        In this case, the storage requirement is fixed and one can
        expect 1-step Q-superlinear convergence to the solution.

        'Hessian Limited Memory' should be used on problems where n is
        very large.
        In this case a limited-memory procedure is used to update a
        diagonal Hessian approximation H_r a limited number of times.
        (Updates are accumulated as a list of vector pairs.
        They are discarded at regular intervals after H_r has been reset
        to their diagonal.)

    'Hessian Limited Memory' : valueless
        Default if n_1 > 75

        These options select the method for storing and updating the
        approximate Hessian. (``nlp2_sparse_solve`` uses a quasi-Newton
        approximation to the Hessian of the Lagrangian.
        A BFGS update is applied after each major iteration.)

        If 'Hessian Full Memory' is specified, the approximate Hessian
        is treated as a dense matrix and the BFGS updates are applied
        explicitly.
        This option is most efficient when the number of variables n is
        not too large (say, less than 75).
        In this case, the storage requirement is fixed and one can
        expect 1-step Q-superlinear convergence to the solution.

        'Hessian Limited Memory' should be used on problems where n is
        very large.
        In this case a limited-memory procedure is used to update a
        diagonal Hessian approximation H_r a limited number of times.
        (Updates are accumulated as a list of vector pairs.
        They are discarded at regular intervals after H_r has been reset
        to their diagonal.)

    'Hessian Frequency' : int
        Default = 99999999

        If option 'Hessian Full Memory' is in effect and i BFGS updates
        have already been carried out, the Hessian approximation is
        reset to the identity matrix. (For certain problems, occasional
        resets may improve convergence, but in general they should not
        be necessary.)

        'Hessian Full Memory' and 'Hessian Frequency' = 10 have a
        similar effect to 'Hessian Limited Memory' and 'Hessian Updates'
        = 10 (except that the latter retains the current diagonal during
        resets).

    'Hessian Updates' : int
        Default = 'Hessian Frequency' if 'Hessian Full Memory', 10
        otherwise

        If option 'Hessian Limited Memory' is in effect and i BFGS
        updates have already been carried out, all but the diagonal
        elements of the accumulated updates are discarded and the
        updating process starts again.

        Broadly speaking, the more updates stored, the better the
        quality of the approximate Hessian.
        However, the more vectors stored, the greater the cost of each
        QP iteration.
        The default value is likely to give a robust algorithm without
        significant expense, but faster convergence can sometimes be
        obtained with significantly fewer updates (e.g., i = 5).

    'Infinite Bound Size' : float
        Default = 10^20

        If r >= 0, r defines the 'infinite' bound bigbnd in the
        definition of the problem constraints.
        Any upper bound greater than or equal to bigbnd will be regarded
        as + infinity (and similarly any lower bound less than or equal
        to -bigbnd will be regarded as - infinity).
        If r < 0, the default value is used.

    'Iterations Limit' : int
        Default = max(10000,10max(n,nf))

        The value of i specifies the maximum number of minor iterations
        allowed (i.e., iterations of the simplex method or the QP
        algorithm), summed over all major iterations. (See also the
        description of the option 'Minor Iterations Limit'.)

    'Linesearch Tolerance' : float
        Default = 0.9

        This tolerance, r, controls the accuracy with which a step
        length will be located along the direction of search each
        iteration.
        At the start of each linesearch a target directional derivative
        for the merit function is identified.
        This argument determines the accuracy to which this target value
        is approximated, and it must be a value in the range 0.0 <= r <=
        1.0.

        The default value r = 0.9 requests just moderate accuracy in the
        linesearch.

        If the nonlinear functions are cheap to evaluate, a more
        accurate search may be appropriate; try r = 0.1, 0.01 or 0.001.

        If the nonlinear functions are expensive to evaluate, a less
        accurate search may be appropriate. `If all gradients are
        known`, try r = 0.99. (The number of major iterations might
        increase, but the total number of function evaluations may
        decrease enough to compensate.)

        If not all gradients are known, a moderately accurate search
        remains appropriate.
        Each search will require only 1--5 function values (typically),
        but many function calls will then be needed to estimate missing
        gradients for the next iteration.

    'LU Density Tolerance' : float
        Default = 0.6

        The density tolerance, r_1, is used during LU factorization of
        the basis matrix B.
        Columns of L and rows of U are formed one at a time, and the
        remaining rows and columns of the basis are altered
        appropriately.
        At any stage, if the density of the remaining matrix exceeds
        r_1, the Markowitz strategy for choosing pivots is terminated,
        and the remaining matrix is factored by a dense LU procedure.
        Raising the density tolerance towards 1.0 may give slightly
        sparser LU factors, with a slight increase in factorization
        time.

        The singularity tolerance, r_2, helps guard against
        ill-conditioned basis matrices.
        After B is refactorized, the diagonal elements of U are tested
        as follows: if |u_{jj}| <= r_2 or |u_{jj}| < r_2 max_i|u_{ij}|,
        the jth column of the basis is replaced by the corresponding
        slack variable. (This is most likely to occur after a restart.)

    'LU Singularity Tolerance' : float
        Default = epsilon^(2/3)

        The density tolerance, r_1, is used during LU factorization of
        the basis matrix B.
        Columns of L and rows of U are formed one at a time, and the
        remaining rows and columns of the basis are altered
        appropriately.
        At any stage, if the density of the remaining matrix exceeds
        r_1, the Markowitz strategy for choosing pivots is terminated,
        and the remaining matrix is factored by a dense LU procedure.
        Raising the density tolerance towards 1.0 may give slightly
        sparser LU factors, with a slight increase in factorization
        time.

        The singularity tolerance, r_2, helps guard against
        ill-conditioned basis matrices.
        After B is refactorized, the diagonal elements of U are tested
        as follows: if |u_{jj}| <= r_2 or |u_{jj}| < r_2 max_i|u_{ij}|,
        the jth column of the basis is replaced by the corresponding
        slack variable. (This is most likely to occur after a restart.)

    'LU Factor Tolerance' : float
        Default = 3.99

        The values of r_1 and r_2 affect the stability of the basis
        factorization B = LU, during refactorization and updates
        respectively.
        The lower triangular matrix L is a product of matrices of the
        form

            [table omitted]

        where the multipliers mu will satisfy |mu| <= r_i.
        The default values of r_1 and r_2 usually strike a good
        compromise between stability and sparsity.
        They must satisfy r_1, r_2 >= 1.0.

        For large and relatively dense problems, r_1 = 10.0 or 5.0 (say)
        may give a useful improvement in stability without impairing
        sparsity to a serious degree.

        For certain very regular structures (e.g., band matrices) it may
        be necessary to reduce r_1 and/or r_2 in order to achieve
        stability.
        For example, if the columns of A include a sub-matrix of the
        form

            [table omitted],

        one should set both r_1 and r_2 to values in the range 1.0 <=
        r_i < 4.0.

    'LU Update Tolerance' : float
        Default = 3.99

        The values of r_1 and r_2 affect the stability of the basis
        factorization B = LU, during refactorization and updates
        respectively.
        The lower triangular matrix L is a product of matrices of the
        form

            [table omitted]

        where the multipliers mu will satisfy |mu| <= r_i.
        The default values of r_1 and r_2 usually strike a good
        compromise between stability and sparsity.
        They must satisfy r_1, r_2 >= 1.0.

        For large and relatively dense problems, r_1 = 10.0 or 5.0 (say)
        may give a useful improvement in stability without impairing
        sparsity to a serious degree.

        For certain very regular structures (e.g., band matrices) it may
        be necessary to reduce r_1 and/or r_2 in order to achieve
        stability.
        For example, if the columns of A include a sub-matrix of the
        form

            [table omitted],

        one should set both r_1 and r_2 to values in the range 1.0 <=
        r_i < 4.0.

    'LU Partial Pivoting' : valueless
        Default

        The LU factorization implements a Markowitz-type search for
        pivots that locally minimize the fill-in subject to a threshold
        pivoting stability criterion.
        The default option is to use threshhold partial pivoting.
        The options 'LU Rook Pivoting' and 'LU Complete Pivoting' are
        more expensive than partial pivoting but are more stable and
        better at revealing rank, as long as 'LU Factor Tolerance' is
        not too large (say < 2.0).
        When numerical difficulties are encountered,
        ``nlp2_sparse_solve`` automatically reduces the LU tolerance
        towards 1.0 and switches (if necessary) to rook or complete
        pivoting, before reverting to the default or specified options
        at the next refactorization (with 'System Information Yes',
        relevant messages are output to the 'Print File').

    'LU Complete Pivoting' : valueless
        The LU factorization implements a Markowitz-type search for
        pivots that locally minimize the fill-in subject to a threshold
        pivoting stability criterion.
        The default option is to use threshhold partial pivoting.
        The options 'LU Rook Pivoting' and 'LU Complete Pivoting' are
        more expensive than partial pivoting but are more stable and
        better at revealing rank, as long as 'LU Factor Tolerance' is
        not too large (say < 2.0).
        When numerical difficulties are encountered,
        ``nlp2_sparse_solve`` automatically reduces the LU tolerance
        towards 1.0 and switches (if necessary) to rook or complete
        pivoting, before reverting to the default or specified options
        at the next refactorization (with 'System Information Yes',
        relevant messages are output to the 'Print File').

    'LU Rook Pivoting' : valueless
        The LU factorization implements a Markowitz-type search for
        pivots that locally minimize the fill-in subject to a threshold
        pivoting stability criterion.
        The default option is to use threshhold partial pivoting.
        The options 'LU Rook Pivoting' and 'LU Complete Pivoting' are
        more expensive than partial pivoting but are more stable and
        better at revealing rank, as long as 'LU Factor Tolerance' is
        not too large (say < 2.0).
        When numerical difficulties are encountered,
        ``nlp2_sparse_solve`` automatically reduces the LU tolerance
        towards 1.0 and switches (if necessary) to rook or complete
        pivoting, before reverting to the default or specified options
        at the next refactorization (with 'System Information Yes',
        relevant messages are output to the 'Print File').

    'Major Feasibility Tolerance' : float
        Default = max(10^(-6),sqrt(epsilon))

        This tolerance, r, specifies how accurately the nonlinear
        constraints should be satisfied.
        The default value is appropriate when the linear and nonlinear
        constraints contain data to about that accuracy.

        Let v_max be the maximum nonlinear constraint violation,
        normalized by the size of the solution, which is required to
        satisfy

            v_max = max_i v_i / ||x|| <= r,

        where v_i is the violation of the ith nonlinear constraint, for
        i = 1,2,...,nf.

        In the major iteration log (see [Minor Iteration Log]), v_max
        appears as the quantity labelled 'Feasible'.
        If some of the problem functions are known to be of low
        accuracy, a larger 'Major Feasibility Tolerance' may be
        appropriate.

    'Major Optimality Tolerance' : float
        Default = 2max(10^(-6),sqrt(epsilon))

        This tolerance, r, specifies the final accuracy of the dual
        variables.
        On successful termination, ``nlp2_sparse_solve`` will have
        computed a solution (x,s,pi) such that

            c_max = max_j c_j / ||pi|| <= r,

        where c_j is an estimate of the complementarity slackness for
        variable j, for j = 1,2,...,n+nf.
        The values c_i are computed from the final QP solution using the
        reduced gradients d_j = g_j-pi^Ta_j (where g_j is the jth
        component of the objective gradient, a_j is the associated
        column of the constraint matrix (A -I), and pi is the set of QP
        dual variables):

            c_j = {[table omitted])

        In the 'Print File', c_max appears as the quantity labelled
        'Optimal'.

    'Major Iterations Limit' : int
        Default = max(1000,3max(n,nf))

        This is the maximum number of major iterations allowed.
        It is intended to guard against an excessive number of
        linearizations of the constraints.
        If i = 0, optimality and feasibility are checked.

    'Major Print Level' : int
        Default = 1

        This controls the amount of output to the options 'Print File'
        and 'Summary File' at each major iteration. 'Major Print Level'
        = 0 suppresses most output, except for error messages. 'Major
        Print Level' = 1 gives normal output for linear and nonlinear
        problems, and 'Major Print Level' = 11 gives additional details
        of the Jacobian factorization that commences each major
        iteration.

        In general, the value being specified may be thought of as a
        binary number of the form

            'Major Print Level' JFDXbs

        where each letter stands for a digit that is either 0 or 1 as
        follows:

        [table omitted]

        To obtain output of any items JFDXbs, set the corresponding
        digit to 1, otherwise to 0.

        If J = 1, the Jacobian matrix will be output column-wise at the
        start of each major iteration.
        Column j will be preceded by the value of the corresponding
        variable x_j and a key to indicate whether the variable is
        basic, superbasic or nonbasic. (Hence if J = 1, there is no
        reason to specify X = 1 unless the objective contains more
        nonlinear variables than the Jacobian.) A typical line of output
        is

        ::

          3 1.250000e+01 BS 1 1.00000e+00 4 2.00000e+00

        which would mean that x_3 is basic at value 12.5, and the third
        column of the Jacobian has elements of 1.0 and 2.0 in rows 1 and
        4.

    'Major Step Limit' : float
        Default = 2.0

        This argument limits the change in x during a linesearch.
        It applies to all nonlinear problems, once a 'feasible solution'
        or 'feasible subproblem' has been found.

        (1) A linesearch determines a step alpha over the range 0 <
            alpha <= beta, where beta is 1 if there are nonlinear
            constraints or is the step to the nearest upper or lower
            bound on x if all the constraints are linear. Normally, the
            first step length tried is alpha_1 = min(1,beta).

        (#) In some cases, such as f(x) = ae^(bx) or f(x) = ax^b, even a
            moderate change in the components of x can lead to
            floating-point overflow. The argument r is therefore used to
            define a limit beta_= r(1+||x||)/||p|| (where p is the
            search direction), and the first evaluation of f(x) is at
            the potentially smaller step length alpha_1 =
            min(1,beta_,beta).

        (#) Wherever possible, upper and lower bounds on x should be
            used to prevent evaluation of nonlinear functions at
            meaningless points. The option 'Major Step Limit' provides
            an additional safeguard. The default value r = 2.0 should
            not affect progress on well behaved problems, but setting r
            = 0.1 or 0.01 may be helpful when rapidly varying functions
            are present. A 'good' starting point may be required. An
            important application is to the class of nonlinear least
            squares problems.

        (#) In cases where several local optima exist, specifying a
            small value for r may help locate an optimum near the
            starting point.

    'Minimize' : valueless
        Default

        The keywords 'Minimize' and 'Maximize' specify the required
        direction of optimization.
        It applies to both linear and nonlinear terms in the objective.

        The keyword 'Feasible Point' means 'Ignore the objective
        function, while finding a feasible point for the linear and
        nonlinear constraints'.
        It can be used to check that the nonlinear constraints are
        feasible without altering the call to ``nlp2_sparse_solve``.

    'Maximize' : valueless
        The keywords 'Minimize' and 'Maximize' specify the required
        direction of optimization.
        It applies to both linear and nonlinear terms in the objective.

        The keyword 'Feasible Point' means 'Ignore the objective
        function, while finding a feasible point for the linear and
        nonlinear constraints'.
        It can be used to check that the nonlinear constraints are
        feasible without altering the call to ``nlp2_sparse_solve``.

    'Feasible Point' : valueless
        The keywords 'Minimize' and 'Maximize' specify the required
        direction of optimization.
        It applies to both linear and nonlinear terms in the objective.

        The keyword 'Feasible Point' means 'Ignore the objective
        function, while finding a feasible point for the linear and
        nonlinear constraints'.
        It can be used to check that the nonlinear constraints are
        feasible without altering the call to ``nlp2_sparse_solve``.

    'Minor Feasibility Tolerance' : float
        Default = max(10^(-6),sqrt(epsilon))

        ``nlp2_sparse_solve`` tries to ensure that all variables
        eventually satisfy their upper and lower bounds to within this
        tolerance, r.
        This includes slack variables.
        Hence, general linear constraints should also be satisfied to
        within r.

        Feasibility with respect to nonlinear constraints is judged by
        the option 'Major Feasibility Tolerance' (not by r).

        If the bounds and linear constraints cannot be satisfied to
        within r, the problem is declared `infeasible`.
        If `sinf` is quite small, it may be appropriate to raise r by a
        factor of 10 or 100.
        Otherwise, some error in the data should be suspected.

        Nonlinear functions will be evaluated only at points that
        satisfy the bounds and linear constraints.
        If there are regions where a function is undefined, every
        attempt should be made to eliminate these regions from the
        problem.

        For example, if f(x) = sqrt(x_1)+log(x_2), it is essential to
        place lower bounds on both variables.
        If r = 1.0e-6, the bounds x_1 >= 10^(-5) and x_2 >= 10^(-4)
        might be appropriate. (The log singularity is more serious.
        In general, keep x as far away from singularities as possible.)

        If 'Scale Option' >= 1, feasibility is defined in terms of the
        `scaled` problem (since it is then more likely to be
        meaningful).

        In reality, ``nlp2_sparse_solve`` uses r as a feasibility
        tolerance for satisfying the bounds on x and s in each QP
        subproblem.
        If the sum of infeasibilities cannot be reduced to zero, the QP
        subproblem is declared infeasible. ``nlp2_sparse_solve`` is then
        in `elastic mode` thereafter (with only the linearized nonlinear
        constraints defined to be elastic).
        See the description of the option 'Elastic Weight'.

    'Feasibility Tolerance' : float
        Default = max{10^(-6),sqrt(epsilon)}

        ``nlp2_sparse_solve`` tries to ensure that all variables
        eventually satisfy their upper and lower bounds to within this
        tolerance, r.
        This includes slack variables.
        Hence, general linear constraints should also be satisfied to
        within r.

        Feasibility with respect to nonlinear constraints is judged by
        the option 'Major Feasibility Tolerance' (not by r).

        If the bounds and linear constraints cannot be satisfied to
        within r, the problem is declared `infeasible`.
        If `sinf` is quite small, it may be appropriate to raise r by a
        factor of 10 or 100.
        Otherwise, some error in the data should be suspected.

        Nonlinear functions will be evaluated only at points that
        satisfy the bounds and linear constraints.
        If there are regions where a function is undefined, every
        attempt should be made to eliminate these regions from the
        problem.

        For example, if f(x) = sqrt(x_1)+log(x_2), it is essential to
        place lower bounds on both variables.
        If r = 1.0e-6, the bounds x_1 >= 10^(-5) and x_2 >= 10^(-4)
        might be appropriate. (The log singularity is more serious.
        In general, keep x as far away from singularities as possible.)

        If 'Scale Option' >= 1, feasibility is defined in terms of the
        `scaled` problem (since it is then more likely to be
        meaningful).

        In reality, ``nlp2_sparse_solve`` uses r as a feasibility
        tolerance for satisfying the bounds on x and s in each QP
        subproblem.
        If the sum of infeasibilities cannot be reduced to zero, the QP
        subproblem is declared infeasible. ``nlp2_sparse_solve`` is then
        in `elastic mode` thereafter (with only the linearized nonlinear
        constraints defined to be elastic).
        See the description of the option 'Elastic Weight'.

    'Minor Iterations Limit' : int
        Default = 500

        If the number of minor iterations for the optimality phase of
        the QP subproblem exceeds i, then all nonbasic QP variables that
        have not yet moved are frozen at their current values and the
        reduced QP is solved to optimality.

        Note that more than i minor iterations may be necessary to solve
        the reduced QP to optimality.
        These extra iterations are necessary to ensure that the
        terminated point gives a suitable direction for the linesearch.

        In the major iteration log (see [Minor Iteration Log]) a t at
        the end of a line indicates that the corresponding QP was
        artificially terminated using the limit i.

        Compare with the option 'Iterations Limit', which defines an
        independent `absolute` limit on the `total` number of minor
        iterations (summed over all QP subproblems).

    'Minor Print Level' : int
        Default = 1

        This controls the amount of output to the 'Print File' and
        'Summary File' during solution of the QP subproblems.
        The value of i has the following effect:

        [table omitted]

    'New Basis File' : int
        Default = 0

        'New Basis File' and 'Backup Basis File' are sometimes referred
        to as basis maps.
        They contain the most compact representation of the state of
        each variable.
        They are intended for restarting the solution of a problem at a
        point that was reached by an earlier run.
        For nontrivial problems, it is advisable to save basis maps at
        the end of a run, in order to restart the run if necessary.

        If i_1 > 0, a basis map will be saved in the file associated
        with unit i_1 every i_3th iteration.
        The first record of the file will contain the word
        ``PROCEEDING`` if the run is still in progress.
        A basis map will also be saved at the end of a run, with some
        other word indicating the final solution status.

        Use of i_2 > 0 is intended as a safeguard against losing the
        results of a long run.
        Suppose that a 'New Basis File' is being saved every 100 ('Save
        Frequency') iterations, and that ``nlp2_sparse_solve`` is about
        to save such a basis at iteration 2000.
        It is conceivable that the run may be interrupted during the
        next few milliseconds (in the middle of the save).
        In this case the Basis file will be corrupted and the run will
        have been essentially wasted.

        To eliminate this risk, both a 'New Basis File' and a 'Backup
        Basis File' may be specified.
        The following would be suitable for the above example:

        ::

          Backup Basis File 11
          New Basis File 12

        The current basis will then be saved every 100 iterations, first
        in the file associated with unit 12 and then immediately in the
        file associated with unit 11.
        If the run is interrupted at iteration 2000 during the save in
        the file associated with unit 12, there will still be a usable
        basis in the file associated with unit 11 (corresponding to
        iteration 1900).

        Note that a new basis will be saved in 'New Basis File' at the
        end of a run if it terminates normally, but it will not be saved
        in 'Backup Basis File'.
        In the above example, if an optimum solution is found at
        iteration 2050 (or if the iteration limit is 2050), the final
        basis in the file associated with unit 12 will correspond to
        iteration 2050, but the last basis saved in the file associated
        with unit 11 will be the one for iteration 2000.

        A full description of information recorded in 'New Basis File'
        and 'Backup Basis File' is given in Gill `et al.` (2005a).

    'Backup Basis File' : int
        Default = 0

        'New Basis File' and 'Backup Basis File' are sometimes referred
        to as basis maps.
        They contain the most compact representation of the state of
        each variable.
        They are intended for restarting the solution of a problem at a
        point that was reached by an earlier run.
        For nontrivial problems, it is advisable to save basis maps at
        the end of a run, in order to restart the run if necessary.

        If i_1 > 0, a basis map will be saved in the file associated
        with unit i_1 every i_3th iteration.
        The first record of the file will contain the word
        ``PROCEEDING`` if the run is still in progress.
        A basis map will also be saved at the end of a run, with some
        other word indicating the final solution status.

        Use of i_2 > 0 is intended as a safeguard against losing the
        results of a long run.
        Suppose that a 'New Basis File' is being saved every 100 ('Save
        Frequency') iterations, and that ``nlp2_sparse_solve`` is about
        to save such a basis at iteration 2000.
        It is conceivable that the run may be interrupted during the
        next few milliseconds (in the middle of the save).
        In this case the Basis file will be corrupted and the run will
        have been essentially wasted.

        To eliminate this risk, both a 'New Basis File' and a 'Backup
        Basis File' may be specified.
        The following would be suitable for the above example:

        ::

          Backup Basis File 11
          New Basis File 12

        The current basis will then be saved every 100 iterations, first
        in the file associated with unit 12 and then immediately in the
        file associated with unit 11.
        If the run is interrupted at iteration 2000 during the save in
        the file associated with unit 12, there will still be a usable
        basis in the file associated with unit 11 (corresponding to
        iteration 1900).

        Note that a new basis will be saved in 'New Basis File' at the
        end of a run if it terminates normally, but it will not be saved
        in 'Backup Basis File'.
        In the above example, if an optimum solution is found at
        iteration 2050 (or if the iteration limit is 2050), the final
        basis in the file associated with unit 12 will correspond to
        iteration 2050, but the last basis saved in the file associated
        with unit 11 will be the one for iteration 2000.

        A full description of information recorded in 'New Basis File'
        and 'Backup Basis File' is given in Gill `et al.` (2005a).

    'Save Frequency' : int
        Default = 100

        'New Basis File' and 'Backup Basis File' are sometimes referred
        to as basis maps.
        They contain the most compact representation of the state of
        each variable.
        They are intended for restarting the solution of a problem at a
        point that was reached by an earlier run.
        For nontrivial problems, it is advisable to save basis maps at
        the end of a run, in order to restart the run if necessary.

        If i_1 > 0, a basis map will be saved in the file associated
        with unit i_1 every i_3th iteration.
        The first record of the file will contain the word
        ``PROCEEDING`` if the run is still in progress.
        A basis map will also be saved at the end of a run, with some
        other word indicating the final solution status.

        Use of i_2 > 0 is intended as a safeguard against losing the
        results of a long run.
        Suppose that a 'New Basis File' is being saved every 100 ('Save
        Frequency') iterations, and that ``nlp2_sparse_solve`` is about
        to save such a basis at iteration 2000.
        It is conceivable that the run may be interrupted during the
        next few milliseconds (in the middle of the save).
        In this case the Basis file will be corrupted and the run will
        have been essentially wasted.

        To eliminate this risk, both a 'New Basis File' and a 'Backup
        Basis File' may be specified.
        The following would be suitable for the above example:

        ::

          Backup Basis File 11
          New Basis File 12

        The current basis will then be saved every 100 iterations, first
        in the file associated with unit 12 and then immediately in the
        file associated with unit 11.
        If the run is interrupted at iteration 2000 during the save in
        the file associated with unit 12, there will still be a usable
        basis in the file associated with unit 11 (corresponding to
        iteration 1900).

        Note that a new basis will be saved in 'New Basis File' at the
        end of a run if it terminates normally, but it will not be saved
        in 'Backup Basis File'.
        In the above example, if an optimum solution is found at
        iteration 2050 (or if the iteration limit is 2050), the final
        basis in the file associated with unit 12 will correspond to
        iteration 2050, but the last basis saved in the file associated
        with unit 11 will be the one for iteration 2000.

        A full description of information recorded in 'New Basis File'
        and 'Backup Basis File' is given in Gill `et al.` (2005a).

    'New Superbasics Limit' : int
        Default = 99

        This option causes early termination of the QP subproblems if
        the number of free variables has increased significantly since
        the first feasible point.
        If the number of new superbasics is greater than i, the nonbasic
        variables that have not yet moved are frozen and the resulting
        smaller QP is solved to optimality.

        In the major iteration log (see [Major Iteration Log]), a t at
        the end of a line indicates that the QP was terminated early in
        this way.

    'Nolist' : valueless
        Default

        Option 'List' enables printing of each option specification as
        it is supplied. 'Nolist' suppresses this printing.

    'List' : valueless
        Option 'List' enables printing of each option specification as
        it is supplied. 'Nolist' suppresses this printing.

    'Old Basis File' : int
        Default = 0

        If i > 0, the basis maps information will be obtained from this
        file. The file will usually have been output previously as a
        'New Basis File' or 'Backup Basis File'.
        A full description of information recorded in 'New Basis File'
        and 'Backup Basis File' is given in Gill `et al.` (2005a).

        The file will not be acceptable if the number of rows or columns
        in the problem has been altered.

    'Partial Price' : int
        Default = 1

        This argument is recommended for large problems that have
        significantly more variables than constraints.
        It reduces the work required for each 'pricing' operation (where
        a nonbasic variable is selected to become superbasic).
        When i = 1, all columns of the constraint matrix (A -I) are
        searched.
        Otherwise, A and I are partitioned to give i roughly equal
        segments A_j and I_j, for j = 1,2,...,i.
        If the previous pricing search was successful on A_{j-1} and
        I_{j-1}, the next search begins on the segments A_j and I_j.
        (All subscripts here are modulo i.) If a reduced gradient is
        found that is larger than some dynamic tolerance, the variable
        with the largest such reduced gradient (of appropriate sign) is
        selected to become superbasic.
        If nothing is found, the search continues on the next segments
        A_{j+1} and I_{j+1}, and so on.

        For time-stage models having t time periods, 'Partial Price' t
        (or t/2 or t/3) may be appropriate.

    'Pivot Tolerance' : float
        Default = epsilon^(2/3)

        During the solution of QP subproblems, the pivot tolerance is
        used to prevent columns entering the basis if they would cause
        the basis to become almost singular.

        When x changes to x+alpha p for some search direction p, a
        'ratio test' determines which component of x reaches an upper or
        lower bound first.
        The corresponding element of p is called the pivot element.
        Elements of p are ignored (and therefore cannot be pivot
        elements) if they are smaller than the pivot tolerance r.

        It is common for two or more variables to reach a bound at
        essentially the same time.
        In such cases, the 'Minor Feasibility Tolerance' (say, t)
        provides some freedom to maximize the pivot element and thereby
        improve numerical stability.
        Excessively small values of t should therefore not be specified.
        To a lesser extent, the 'Expand Frequency' (say, f) also
        provides some freedom to maximize the pivot element.
        Excessively `large` values of f should therefore not be
        specified.

    'Print File' : int
        Default = 0

        If i > 0, the following information is output to a file
        associated with unit i during the solution of each problem:

        -   a listing of the options;

        -   some statistics about the problem;

        -   the amount of storage available for the LU factorization of
            the basis matrix;

        -   notes about the initial basis resulting from a Crash
            procedure or a Basis file;

        -   the iteration log;

        -   basis factorization statistics;

        -   the exit errno condition and some statistics about the
            solution obtained;

        -   the printed solution, if requested.

        These items are described in [Further Comments] and [Monitoring
        Information].
        Further brief output may be directed to the 'Summary File'.

    'Print Frequency' : int
        Default = 100

        If i > 0, one line of the iteration log will be printed every
        ith iteration.
        A value such as i = 10 is suggested for those interested only in
        the final solution.
        If i <= 0, the value of i = 99999999 is used and effectively no
        checks are made.

    'Proximal Point Method' : int
        Default = 1

        i = 1 or 2 specifies minimization of ||x-x_0||_1 or
        1/2||x-x_0||_2^2 when the starting point x_0 is changed to
        satisfy the linear constraints (where x_0 refers to nonlinear
        variables).

    'Punch File' : int
        Default = 0

        The 'Punch File' from a previous run may be used as an 'Insert
        File' for a later run on the same problem.
        A full description of information recorded in 'Insert File' and
        'Punch File' is given in Gill `et al.` (2005a).

        If i_1 > 0, the final solution obtained will be output to the
        file. For linear programs, this format is compatible with
        various commercial systems.

        If i_2 > 0 the 'Insert File' containing basis information will
        be read from unit i_2. The file will usually have been output
        previously as a 'Punch File'.
        The file will not be accessed if 'Old Basis File' is specified.

    'Insert File' : int
        Default = 0

        The 'Punch File' from a previous run may be used as an 'Insert
        File' for a later run on the same problem.
        A full description of information recorded in 'Insert File' and
        'Punch File' is given in Gill `et al.` (2005a).

        If i_1 > 0, the final solution obtained will be output to the
        file. For linear programs, this format is compatible with
        various commercial systems.

        If i_2 > 0 the 'Insert File' containing basis information will
        be read from unit i_2. The file will usually have been output
        previously as a 'Punch File'.
        The file will not be accessed if 'Old Basis File' is specified.

    'Scale Option' : int
        Default = 0

        Three scale options are available as follows:

        [table omitted]

        Option 'Scale Tolerance' affects how many passes might be needed
        through the constraint matrix.
        On each pass, the scaling procedure computes the ratio of the
        largest and smallest nonzero coefficients in each column:

            rho_j = max_j|a_{ij}|/min_i|a_{ij}| (a_{ij} != 0).

        If max_jrho_j is less than r times its previous value, another
        scaling pass is performed to adjust the row and column scales.
        Raising r from 0.9 to 0.99 (say) usually increases the number of
        scaling passes through A.
        At most 10 passes are made.
        The value of r should lie in the range 0 < r < 1.

        'Scale Print' causes the row scales r(i) and column scales c(j)
        to be printed to 'Print File', if 'System Information Yes' has
        been specified.
        The scaled matrix coefficients are a__{ij} = a_{ij}c(j)/r(i),
        and the scaled bounds on the variables and slacks are l__j =
        l_j/c(j), u__j = u_j/c(j), where c(j) = r(j-n) if j > n.

    'Scale Tolerance' : float
        Default = 0.9

        Three scale options are available as follows:

        [table omitted]

        Option 'Scale Tolerance' affects how many passes might be needed
        through the constraint matrix.
        On each pass, the scaling procedure computes the ratio of the
        largest and smallest nonzero coefficients in each column:

            rho_j = max_j|a_{ij}|/min_i|a_{ij}| (a_{ij} != 0).

        If max_jrho_j is less than r times its previous value, another
        scaling pass is performed to adjust the row and column scales.
        Raising r from 0.9 to 0.99 (say) usually increases the number of
        scaling passes through A.
        At most 10 passes are made.
        The value of r should lie in the range 0 < r < 1.

        'Scale Print' causes the row scales r(i) and column scales c(j)
        to be printed to 'Print File', if 'System Information Yes' has
        been specified.
        The scaled matrix coefficients are a__{ij} = a_{ij}c(j)/r(i),
        and the scaled bounds on the variables and slacks are l__j =
        l_j/c(j), u__j = u_j/c(j), where c(j) = r(j-n) if j > n.

    'Scale Print' : valueless
        Three scale options are available as follows:

        [table omitted]

        Option 'Scale Tolerance' affects how many passes might be needed
        through the constraint matrix.
        On each pass, the scaling procedure computes the ratio of the
        largest and smallest nonzero coefficients in each column:

            rho_j = max_j|a_{ij}|/min_i|a_{ij}| (a_{ij} != 0).

        If max_jrho_j is less than r times its previous value, another
        scaling pass is performed to adjust the row and column scales.
        Raising r from 0.9 to 0.99 (say) usually increases the number of
        scaling passes through A.
        At most 10 passes are made.
        The value of r should lie in the range 0 < r < 1.

        'Scale Print' causes the row scales r(i) and column scales c(j)
        to be printed to 'Print File', if 'System Information Yes' has
        been specified.
        The scaled matrix coefficients are a__{ij} = a_{ij}c(j)/r(i),
        and the scaled bounds on the variables and slacks are l__j =
        l_j/c(j), u__j = u_j/c(j), where c(j) = r(j-n) if j > n.

    'Solution File' : int
        Default = 0

        If i > 0, the final solution will be output to file i (whether
        optimal or not).
        All numbers are printed in ``1pe16.6`` format.

        To see more significant digits in the printed solution, it will
        sometimes be useful to make i refer to 'Print File'.

    'Summary File' : int
        Default = 0

        If i_1 > 0, a brief log will be output to the file associated
        with unit i_1, including one line of information every i_2th
        iteration.
        In an interactive environment, it is useful to direct this
        output to the terminal, to allow a run to be monitored online.
        (If something looks wrong, the run can be manually terminated.)
        Further details are given in [The Summary File].

    'Summary Frequency' : int
        Default = 100

        If i_1 > 0, a brief log will be output to the file associated
        with unit i_1, including one line of information every i_2th
        iteration.
        In an interactive environment, it is useful to direct this
        output to the terminal, to allow a run to be monitored online.
        (If something looks wrong, the run can be manually terminated.)
        Further details are given in [The Summary File].

    'Superbasics Limit' : int
        Default = n_1

        This option places a limit on the storage allocated for
        superbasic variables.
        Ideally, i should be set slightly larger than the 'number of
        degrees of freedom' expected at an optimal solution.

        For nonlinear problems, the number of degrees of freedom is
        often called the 'number of independent variables'.
        Normally, i need not be greater than n+1, where n_1 is the
        number of nonlinear variables.
        For many problems, i may be considerably smaller than n.
        This will save storage if n is very large.

    'Suppress Parameters' : valueless
        Normally ``nlp2_sparse_solve`` prints the options file as it is
        being read, and then prints a complete list of the available
        keywords and their final values.
        The option 'Suppress Parameters' tells ``nlp2_sparse_solve`` not
        to print the full list.

    'System Information No' : valueless
        Default

        This option prints additional information on the progress of
        major and minor iterations, and Crash statistics.
        See [Monitoring Information].

    'System Information Yes' : valueless
        This option prints additional information on the progress of
        major and minor iterations, and Crash statistics.
        See [Monitoring Information].

    'Timing Level' : int
        Default = 0

        If i > 0, some timing information will be output to the Print
        file, if 'Print File' > 0.

    'Unbounded Objective' : float
        Default = 1.0e+15

        These arguments are intended to detect unboundedness in
        nonlinear problems.
        During a linesearch, F_obj is evaluated at points of the form
        x+alpha p, where x and p are fixed and alpha varies.
        If |F_obj| exceeds r_1 or alpha exceeds r_2, iterations are
        terminated with the exit message `errno` = 5.

        If singularities are present, unboundedness in F_obj(x) may be
        manifested by a floating-point overflow (during the evaluation
        of F_obj(x+alpha p)), before the test against r_1 can be made.

        Unboundedness in x is best avoided by placing finite upper and
        lower bounds on the variables.

    'Unbounded Step Size' : float
        Default = infbnd

        These arguments are intended to detect unboundedness in
        nonlinear problems.
        During a linesearch, F_obj is evaluated at points of the form
        x+alpha p, where x and p are fixed and alpha varies.
        If |F_obj| exceeds r_1 or alpha exceeds r_2, iterations are
        terminated with the exit message `errno` = 5.

        If singularities are present, unboundedness in F_obj(x) may be
        manifested by a floating-point overflow (during the evaluation
        of F_obj(x+alpha p)), before the test against r_1 can be made.

        Unboundedness in x is best avoided by placing finite upper and
        lower bounds on the variables.

    'Verify Level' : int
        Default = 0

        This option refers to finite difference checks on the
        derivatives computed by the user-supplied functions.
        Derivatives are checked at the first point that satisfies all
        bounds and linear constraints.

        [table omitted]

        'Verify Level' = 3 should be specified whenever a new `usrfun`
        is being developed.

    'Violation Limit' : float
        Default = 1.0e+6

        This keyword defines an absolute limit on the magnitude of the
        maximum constraint violation, r, after the linesearch.
        On completion of the linesearch, the new iterate x_{k+1}
        satisfies the condition

            v_i(x_{k+1}) <= r max(1,v_i(x_0)),

        where x_0 is the point at which the nonlinear constraints are
        first evaluated and v_i(x) is the ith nonlinear constraint
        violation v_i(x) = max(0,l_i-f_i(x),f_i(x)-u_i).

        The effect of this violation limit is to restrict the iterates
        to lie in an `expanded` feasible region whose size depends on
        the magnitude of r.
        This makes it possible to keep the iterates within a region
        where the objective is expected to be well-defined and bounded
        below.
        If the objective is bounded below for all values of the
        variables, r may be any large positive value.

    Raises
    ------
    NagValueError
        (`errno` 1)
            The initialization function ``nlp2_sparse_init`` has not
            been called.

        (`errno` 2)
            On entry, one but not both of nxname and nfname is equal to
            1. nxname = *<value>* and nfname = *<value>*.

        (`errno` 2)
            On entry, `objrow` = *<value>* and nf = *<value>*.

            Constraint: 0 <= `objrow` <= nf.

        (`errno` 2)
            On entry, bounds `xlow` and `xupp` for *<value>* are equal
            and infinite. `xlow` = `xupp` = *<value>* and infbnd =
            *<value>*.

        (`errno` 2)
            On entry, bounds for *<value>* are inconsistent. `xlow` =
            *<value>* and `xupp` = *<value>*.

        (`errno` 2)
            On entry, bounds `xlow` and `xupp` for variable *<value>*
            are equal and infinite. `xlow` = `xupp` = *<value>* and
            infbnd = *<value>*.

        (`errno` 2)
            On entry, bounds for variable *<value>* are inconsistent.
            `xlow` = *<value>* and `xupp` = *<value>*.

        (`errno` 2)
            On entry, bounds `flow` and `fupp` for *<value>* are equal
            and infinite. `flow` = `fupp` = *<value>* and infbnd =
            *<value>*.

        (`errno` 2)
            On entry, bounds for *<value>* are inconsistent. `flow` =
            *<value>* and `fupp` = *<value>*.

        (`errno` 2)
            On entry, bounds `flow` and `fupp` for variable *<value>*
            are equal and infinite. `flow` = `fupp` = *<value>* and
            infbnd = *<value>*.

        (`errno` 2)
            On entry, bounds for variable *<value>* are inconsistent.
            `flow` = *<value>* and `fupp` = *<value>*.

        (`errno` 2)
            Basis file dimensions do not match this problem.

        (`errno` 2)
            Array element `igfun`[*<value>*] = *<value>* is out of range
            1 to nf = *<value>*, or array element `jgvar`[*<value>*] =
            *<value>* is out of range 1 to n = *<value>*.

        (`errno` 2)
            On entry, lena = *<value>*.

            Constraint: lena >= 1.

        (`errno` 2)
            On entry, leng = *<value>*.

            Constraint: leng >= 1.

        (`errno` 2)
            On entry, nf = *<value>*.

            Constraint: nf >= 1.

        (`errno` 2)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 2)
            On entry, `start` = *<value>*.

            Constraint: `start` = 0, 1 or 2.

        (`errno` 2)
            On entry, `nea` = *<value>*.

            Constraint: `nea` >= 0.

        (`errno` 2)
            On entry, `neg` = *<value>*.

            Constraint: `neg` >= 0.

        (`errno` 2)
            On entry, `nea` = *<value>*, n = *<value>* and nf =
            *<value>*.

            Constraint: `nea` <= n*nf.

        (`errno` 2)
            On entry, `neg` = *<value>*, n = *<value>* and nf =
            *<value>*.

            Constraint: `neg` <= n*nf.

        (`errno` 2)
            On entry, nxname = *<value>* and n = *<value>*.

            Constraint: nxname = 1 or n.

        (`errno` 2)
            On entry, nfname = *<value>* and nf = *<value>*.

            Constraint: nfname = 1 or nf.

        (`errno` 8)
            User-supplied function computes incorrect objective
            derivatives.

        (`errno` 8)
            User-supplied function computes incorrect constraint
            derivatives.

        (`errno` 11)
            Internal error: memory allocation failed when attempting to
            allocate workspace sizes *<value>*, *<value>* and *<value>*.
            Please contact NAG.

        (`errno` 12)
            Internal memory allocation was insufficient. Please contact
            NAG.

        (`errno` 13)
            An error has occurred in the basis package. Check that
            arrays `iafun`, `javar`, `igfun` and `jgvar` contain values
            in the appropriate ranges and do not define duplicate
            elements of `a` or `g`. Set the option 'Print File' and
            examine the output carefully for further information.

        (`errno` 14)
            An unexpected error has occurred. Set the option 'Print
            File' and examine the output carefully for further
            information.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            The requested accuracy could not be achieved.

        (`errno` 4)
            The linear constraints appear to be infeasible.

        (`errno` 4)
            The problem appears to be infeasible. The linear equality
            constraints could not be satisfied.

        (`errno` 4)
            The problem appears to be infeasible. Nonlinear
            infeasibilites have been minimized.

        (`errno` 4)
            The problem appears to be infeasible. Infeasibilites have
            been minimized.

        (`errno` 5)
            The problem appears to be unbounded. The objective function
            is unbounded.

        (`errno` 5)
            The problem appears to be unbounded. The constraint
            violation limit has been reached.

        (`errno` 10)
            User-supplied function requested termination.

    NagAlgorithmicMajorWarning
        (`errno` 6)
            Iteration limit reached.

        (`errno` 6)
            Major iteration limit reached.

        (`errno` 6)
            The value of the option 'Superbasics Limit' is too small.

        (`errno` 7)
            Numerical difficulties have been encountered and no further
            progress can be made.

        (`errno` 9)
            User-supplied function is undefined at the first feasible
            point.

        (`errno` 9)
            User-supplied function is undefined at the initial point.

        (`errno` 9)
            Unable to proceed into undefined region of user-supplied
            function.

    References
    ----------
    Dantzig, G B, 1963, `Linear Programming and Extensions`, Princeton
    University Press

    Eldersveld, S K, 1991, `Large-scale sequential quadratic programming
    algorithms`, PhD Thesis, Department of Operations Research, Stanford
    University, Stanford

    Fourer, R, 1982, `Solving staircase linear programs by the simplex
    method`, Math. Programming (23), 274--313

    Gill, P E, Murray, W and Saunders, M A, 2002, `SNOPT: An SQP
    Algorithm for Large-scale Constrained Optimization` (12), 979--1006,
    SIAM J. Optim.

    Gill, P E, Murray, W and Saunders, M A, 2005, `Users' guide for
    SQOPT 7: a Fortran package for large-scale linear and quadratic
    programming`, Report NA 05-1, Department of Mathematics, University
    of California, San Diego,
    https://www.ccom.ucsd.edu/~peg/papers/sqdoc7.pdf

    Gill, P E, Murray, W and Saunders, M A, 2005, `Users' guide for
    SNOPT 7.1: a Fortran package for large-scale linear nonlinear
    programming`, Report NA 05-2, Department of Mathematics, University
    of California, San Diego,
    https://www.ccom.ucsd.edu/~peg/papers/sndoc7.pdf

    Gill, P E, Murray, W, Saunders, M A and Wright, M H, 1986, `Users'
    guide for NPSOL (Version 4.0): a Fortran package for nonlinear
    programming`, Report SOL 86-2, Department of Operations Research,
    Stanford University

    Gill, P E, Murray, W, Saunders, M A and Wright, M H, 1987,
    `Maintaining LU factors of a general sparse matrix`, Linear Algebra
    and its Applics. (88/89), 239--270

    Gill, P E, Murray, W, Saunders, M A and Wright, M H, 1992, `Some
    theoretical properties of an augmented Lagrangian merit function`,
    Advances in Optimization and Parallel Computing, (ed P M Pardalos),
    101--128, North Holland

    Hock, W and Schittkowski, K, 1981, `Test Examples for Nonlinear
    Programming Codes. Lecture Notes in Economics and Mathematical
    Systems` (187), Springer--Verlag

    Murtagh, B A and Saunders, M A, 1978, `Large-scale linearly
    constrained optimization`, 14, 41--72, Math. Programming

    Murtagh, B A and Saunders, M A, 1982, `A projected Lagrangian
    algorithm and its implementation for sparse nonlinear constraints`,
    Math. Program. Stud. (16), 84--118

    Murtagh, B A and Saunders, M A, 1995, `MINOS 5.4 users' guide`,
    Report SOL 83-20R, Department of Operations Research, Stanford
    University

    See Also
    --------
    :meth:`naginterfaces.library.examples.opt.nlp2_sparse_solve_ex.main`
    """

    fun_name = 'naginterfaces.library.opt.nlp2_sparse_solve'

    cb_usrfun_name = fun_name + ':usrfun'

    def cb_usrfun(status, n, x, needf, nf, f, needg, leng, g, data=None):
        """Wrapping user's usrfun. Called as base usrfun."""
        if data is not None:
            _status_usrfun, _f_usrfun, _g_usrfun = usrfun(
                status,
                x,
                needf,
                f,
                needg,
                g,
                data,
            )
        else:
            _status_usrfun, _f_usrfun, _g_usrfun = usrfun(
                status,
                x,
                needf,
                f,
                needg,
                g,
            )
        _status_usrfun_base = _status_usrfun
        f[:] = _utils._to_ndarray(
            {'fun_name': cb_usrfun_name, 'returned': 'f'},
            _f_usrfun,
            copy=False,
            exp_shape=(nf,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        g[:] = _utils._to_ndarray(
            {'fun_name': cb_usrfun_name, 'returned': 'g'},
            _g_usrfun,
            copy=False,
            exp_shape=(leng,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        return _status_usrfun_base

    _flow = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'flow'},
        flow,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _nf_py = _b_utils._EngineFloat64ArrayType.get_shape(_flow, {'fun_name': fun_name, 'entity': 'flow'}, exp_rank=1)[0]
    _nf = _b_utils._EngineIntScalarType(
        _nf_py, {'fun_name': fun_name, 'entity': '_nf'},
    ).to_ctypes()
    _xlow = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xlow'},
        xlow,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_xlow, {'fun_name': fun_name, 'entity': 'xlow'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _xnames = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xnames'},
        xnames,
        copy=False,
    )
    _nxname_py = _b_utils._EngineCharArrayType.get_shape(_xnames, {'fun_name': fun_name, 'entity': 'xnames'}, exp_rank=1)[0]
    _nxname = _b_utils._EngineIntScalarType(
        _nxname_py, {'fun_name': fun_name, 'entity': '_nxname'},
    ).to_ctypes()
    _fnames = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'fnames'},
        fnames,
        copy=False,
    )
    _nfname_py = _b_utils._EngineCharArrayType.get_shape(_fnames, {'fun_name': fun_name, 'entity': 'fnames'}, exp_rank=1)[0]
    _nfname = _b_utils._EngineIntScalarType(
        _nfname_py, {'fun_name': fun_name, 'entity': '_nfname'},
    ).to_ctypes()
    _b_utils._check_callable(usrfun, {'fun_name': fun_name, 'entity': 'usrfun'})
    _usrfun = cb_usrfun
    _iafun = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'iafun'},
        iafun,
        copy=False,
    )
    _lena_py = _b_utils._EngineIntArrayType.get_shape(_iafun, {'fun_name': fun_name, 'entity': 'iafun'}, exp_rank=1)[0]
    _lena = _b_utils._EngineIntScalarType(
        _lena_py, {'fun_name': fun_name, 'entity': '_lena'},
    ).to_ctypes()
    _javar = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'javar'},
        javar,
        copy=False,
    )
    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _igfun = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'igfun'},
        igfun,
        copy=False,
    )
    _leng_py = _b_utils._EngineIntArrayType.get_shape(_igfun, {'fun_name': fun_name, 'entity': 'igfun'}, exp_rank=1)[0]
    _leng = _b_utils._EngineIntScalarType(
        _leng_py, {'fun_name': fun_name, 'entity': '_leng'},
    ).to_ctypes()
    _jgvar = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'jgvar'},
        jgvar,
        copy=False,
    )
    _xupp = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xupp'},
        xupp,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _fupp = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'fupp'},
        fupp,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _xstate = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xstate'},
        xstate,
        copy=True,
    )
    _xmul = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _f = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'f'},
        f,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _fstate = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'fstate'},
        fstate,
        copy=True,
    )
    _fmul = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'fmul'},
        fmul,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    try:
        _return_scalars = _b_opt.nlp2_sparse_solve(
            start,
            _nf,
            _n,
            _nxname,
            _nfname,
            objadd,
            objrow,
            prob,
            _usrfun,
            _iafun,
            _javar,
            _lena,
            nea,
            _a,
            _igfun,
            _jgvar,
            _leng,
            neg,
            _xlow,
            _xupp,
            _xnames,
            _flow,
            _fupp,
            _fnames,
            _x,
            _xstate,
            _xmul,
            _f,
            _fstate,
            _fmul,
            ns,
            comm,
            data,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _x_py = _x
    _xstate_py = _xstate
    _f_py = _f
    _fstate_py = _fstate
    _fmul_py = _fmul
    _ns_py = _return_scalars.ns
    _xmul_py = _xmul
    _ninf_py = _return_scalars.ninf
    _sinf_py = _return_scalars.sinf
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['x', 'xstate', 'xmul', 'f', 'fstate', 'fmul', 'ns', 'ninf', 'sinf'],
    )(
        _x_py,
        _xstate_py,
        _xmul_py,
        _f_py,
        _fstate_py,
        _fmul_py,
        _ns_py,
        _ninf_py,
        _sinf_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def nlp2_sparse_jacobian(nf, usrfun, lena, leng, x, xlow, xupp, comm, data=None, io_manager=None):
    u"""
    Determine the pattern of nonzeros in the Jacobian matrix for
    ``nlp2_sparse_solve``.

    ``nlp2_sparse_jacobian`` may be used before ``nlp2_sparse_solve`` to
    determine the sparsity pattern for the Jacobian.

    For full information please refer to the NAG Library document for
    e04vj

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04vjf.html

    Parameters
    ----------
    nf : int
        nf, the number of problem functions in F(x), including the
        objective function (if any) and the linear and nonlinear
        constraints. Simple upper and lower bounds on x can be defined
        using the arguments `xlow` and `xupp` and should not be included
        in F.

    usrfun : callable (status, f, g) = usrfun(status, x, needf, f,
        needg, g, data=None)
        `usrfun` must define the problem functions F(x).

        This function is passed to ``nlp2_sparse_jacobian`` as the
        external argument `usrfun`.

        Parameters
        ~~~~~~~~~~
        status : int
            Indicates the first call to `usrfun`.

            `status` = 0

                There is nothing special about the current call to
                `usrfun`.

            `status` = 1

                ``nlp2_sparse_jacobian`` is calling your function for
                the `first` time. Some data may need to be input or
                computed and saved.

        x : float, ndarray, shape (n)
            The variables x at which the problem functions are to be
            calculated. The array x must not be altered.

        needf : int
            Indicates if `f` must be assigned during the call to
            `usrfun` (see `f`).

        f : float, ndarray, shape (nf)
            This will be set by ``nlp2_sparse_jacobian``.

        needg : int
            ``nlp2_sparse_jacobian`` will call `usrfun` with `needg` = 0
            to indicate that `g` is not required.

        g : float, ndarray, shape (leng)
            Concerns the calculations of the derivatives of the function
            f(x).

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        status : int
            May be used to indicate that you are unable to evaluate F at
            the current x. (For example, the problem functions may not
            be defined there).

            ``nlp2_sparse_jacobian`` evaluates F(x) at random
            perturbation of the initial point x, say x_p.

            If the functions cannot be evaluated at x_p, you can set
            `status` = -1, ``nlp2_sparse_jacobian`` will use another
            random perturbation.

            If for some reason you wish to terminate the current
            problem, set `status` <= -2.

        f : float, array-like, shape (nf)
            The computed F(x) according to the setting of `needf`.

            If `needf` = 0, `f` is not required and is ignored.

            If `needf` > 0, the components of F(x) must be calculated
            and assigned to `f`. ``nlp2_sparse_jacobian`` will always
            call `usrfun` with `needf` > 0.

            To simplify the code, you may ignore the value of `needf`
            and compute F(x) on every entry to `usrfun`.

        g : float, array-like, shape (leng)
            ``nlp2_sparse_jacobian`` will always call `usrfun` with
            `needg` = 0: `g` is not required to be set on exit but must
            be declared correctly.

    lena : int
        `lena` should be an `overestimate` of the number of elements in
        the linear part of the Jacobian.

    leng : int
        `leng` should be an `overestimate` of the number of elements in
        the nonlinear part of the Jacobian.

    x : float, array-like, shape (n)
        An initial estimate of the variables x. The contents of x will
        be used by ``nlp2_sparse_jacobian`` in the call of `usrfun`, and
        so each element of `x` should be within the bounds given by
        `xlow` and `xupp`.

    xlow : float, array-like, shape (n)
        Contain the lower and upper bounds l_x and u_x on the variables
        x.

        To specify a nonexistent lower bound [l_x]_j = - infinity, set
        `xlow`[j-1] <= -bigbnd, where bigbnd is the option 'Infinite
        Bound Size'.

        To specify a nonexistent upper bound `xupp`[j-1] >= bigbnd.

        To fix the jth variable (say, x_j = beta, where |beta| <
        bigbnd), set `xlow`[j-1] = `xupp`[j-1] = beta.

    xupp : float, array-like, shape (n)
        Contain the lower and upper bounds l_x and u_x on the variables
        x.

        To specify a nonexistent lower bound [l_x]_j = - infinity, set
        `xlow`[j-1] <= -bigbnd, where bigbnd is the option 'Infinite
        Bound Size'.

        To specify a nonexistent upper bound `xupp`[j-1] >= bigbnd.

        To fix the jth variable (say, x_j = beta, where |beta| <
        bigbnd), set `xlow`[j-1] = `xupp`[j-1] = beta.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``nlp2_sparse_init``.

    data : arbitrary, optional
        User-communication data for callback functions.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    iafun : int, ndarray, shape (`lena`)
        Define the i coordinates of the nonzero elements of the linear
        part A of the function F(x) = f(x)+Ax.

    javar : int, ndarray, shape (`lena`)
        Define the j coordinates of the nonzero elements of the linear
        part A of the function F(x) = f(x)+Ax.

    nea : int
        Is the number of nonzero entries in A such that F(x) = f(x)+Ax.

    a : float, ndarray, shape (`lena`)
        Define the coordinates (i,j) and values A_{ij} of the nonzero
        elements of the linear part A of the function F(x) = f(x)+Ax.

        In particular, `nea` triples
        (`iafun`[k-1],`javar`[k-1],`a`[k-1]) define the row and column
        indices i = `iafun`[k-1] and j = `javar`[k-1] of the element
        A_{ij} = `a`[k-1].

    igfun : int, ndarray, shape (`leng`)
        Define the coordinates (i,j) of the nonzero elements of G, the
        nonlinear part of the derivatives J(x) = G(x)+A of the function
        F(x) = f(x)+Ax.

    jgvar : int, ndarray, shape (`leng`)
        Define the coordinates (i,j) of the nonzero elements of G, the
        nonlinear part of the derivatives J(x) = G(x)+A of the function
        F(x) = f(x)+Ax.

    neg : int
        The number of nonzero entries in G.

    Raises
    ------
    NagValueError
        (`errno` 1)
            The initialization function ``nlp2_sparse_init`` has not
            been called.

        (`errno` 2)
            On entry, `lena` = *<value>*.

            Constraint: `lena` >= 1.

        (`errno` 2)
            On entry, `leng` = *<value>*.

            Constraint: `leng` >= 1.

        (`errno` 3)
            User-supplied function `usrfun` indicates that functions are
            undefined near given point `x`.

        (`errno` 5)
            Either `lena` or `leng` is too small. Increase both of them
            and corresponding array sizes. `lena` = *<value>* and `leng`
            = *<value>*.

        (`errno` 6)
            Cannot estimate Jacobian structure at given point `x`.

        (`errno` 7)
            Internal error: memory allocation failed when attempting to
            allocate workspace sizes *<value>*, *<value>* and *<value>*.
            Please contact NAG.

        (`errno` 8)
            Internal memory allocation was insufficient. Please contact
            NAG.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 4)
            User-supplied function `usrfun` requested termination.

    References
    ----------
    Hock, W and Schittkowski, K, 1981, `Test Examples for Nonlinear
    Programming Codes. Lecture Notes in Economics and Mathematical
    Systems` (187), Springer--Verlag
    """

    fun_name = 'naginterfaces.library.opt.nlp2_sparse_jacobian'

    cb_usrfun_name = fun_name + ':usrfun'

    def cb_usrfun(status, n, x, needf, nf, f, needg, leng, g, data=None):
        """Wrapping user's usrfun. Called as base usrfun."""
        if data is not None:
            _status_usrfun, _f_usrfun, _g_usrfun = usrfun(
                status,
                x,
                needf,
                f,
                needg,
                g,
                data,
            )
        else:
            _status_usrfun, _f_usrfun, _g_usrfun = usrfun(
                status,
                x,
                needf,
                f,
                needg,
                g,
            )
        _status_usrfun_base = _status_usrfun
        f[:] = _utils._to_ndarray(
            {'fun_name': cb_usrfun_name, 'returned': 'f'},
            _f_usrfun,
            copy=False,
            exp_shape=(nf,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        g[:] = _utils._to_ndarray(
            {'fun_name': cb_usrfun_name, 'returned': 'g'},
            _g_usrfun,
            copy=False,
            exp_shape=(leng,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        return _status_usrfun_base

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b_utils._check_callable(usrfun, {'fun_name': fun_name, 'entity': 'usrfun'})
    _usrfun = cb_usrfun
    _lena = _b_utils._EngineIntScalarType(
        lena, {'fun_name': fun_name, 'entity': 'lena'},
    ).to_ctypes()
    _lena_py = _b_utils._EngineIntScalarType.to_py(_lena)
    _iafun = _b_utils._EngineIntArrayType.empty_ndarray(
        (_lena_py,),
    )
    _javar = _b_utils._EngineIntArrayType.empty_ndarray(
        (_lena_py,),
    )
    _a = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_lena_py,),
    )
    _leng = _b_utils._EngineIntScalarType(
        leng, {'fun_name': fun_name, 'entity': 'leng'},
    ).to_ctypes()
    _leng_py = _b_utils._EngineIntScalarType.to_py(_leng)
    _igfun = _b_utils._EngineIntArrayType.empty_ndarray(
        (_leng_py,),
    )
    _jgvar = _b_utils._EngineIntArrayType.empty_ndarray(
        (_leng_py,),
    )
    _xlow = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xlow'},
        xlow,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _xupp = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'xupp'},
        xupp,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    try:
        _return_scalars = _b_opt.nlp2_sparse_jacobian(
            nf,
            _n,
            _usrfun,
            _iafun,
            _javar,
            _lena,
            _a,
            _igfun,
            _jgvar,
            _leng,
            _x,
            _xlow,
            _xupp,
            comm,
            data,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _iafun_py = _iafun
    _javar_py = _javar
    _nea_py = _return_scalars.nea
    _a_py = _a
    _igfun_py = _igfun
    _jgvar_py = _jgvar
    _neg_py = _return_scalars.neg
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['iafun', 'javar', 'nea', 'a', 'igfun', 'jgvar', 'neg'],
    )(
        _iafun_py,
        _javar_py,
        _nea_py,
        _a_py,
        _igfun_py,
        _jgvar_py,
        _neg_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def nlp2_sparse_option_file(ispecs, comm, io_manager=None):
    u"""
    Supply option values for ``nlp2_sparse_solve`` from external file.

    ``nlp2_sparse_option_file`` may be used to supply options to
    ``nlp2_sparse_solve`` from an external file.
    The initialization function ``nlp2_sparse_init`` **must** have been
    called before calling ``nlp2_sparse_option_file``.

    For full information please refer to the NAG Library document for
    e04vk

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04vkf.html

    Parameters
    ----------
    ispecs : int
        The unit number (see
        :meth:`~naginterfaces.base.utils.FileObjManager.unit_from_fileobj`)
        of the option file to be read.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``nlp2_sparse_init``.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Raises
    ------
    NagValueError
        (`errno` 1)
            The initialization function ``nlp2_sparse_init`` has not
            been called.

        (`errno` 2)
            At least one line of the options file is invalid.

        (`errno` 2)
            Could not read options file on unit `ispecs` = *<value>*.
    """

    _b_opt.nlp2_sparse_option_file(
        ispecs,
        comm,
        io_manager,
    )

def nlp2_sparse_option_string(string, comm, io_manager=None):
    u"""
    Set a single option for ``nlp2_sparse_solve`` from a character
    string.

    ``nlp2_sparse_option_string`` may be used to supply individual
    options to ``nlp2_sparse_solve``.
    The initialization function ``nlp2_sparse_init`` **must** have been
    called before calling ``nlp2_sparse_option_string``.

    For full information please refer to the NAG Library document for
    e04vl

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04vlf.html

    Parameters
    ----------
    string : str
        A single valid option string (see [Description] in
        ``nlp2_sparse_option_string`` and [Other Parameters for
        nlp2_sparse_solve]).

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``nlp2_sparse_init``.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Raises
    ------
    NagValueError
        (`errno` 1)
            The initialization function ``nlp2_sparse_init`` has not
            been called.

        (`errno` 2)
            The supplied option is invalid. Check that the keywords are
            neither ambiguous nor misspelt. The option string is
            *<value>*.
    """

    _b_opt.nlp2_sparse_option_string(
        string,
        comm,
        io_manager,
    )

def nlp2_sparse_option_integer_set(string, ivalue, comm, io_manager=None):
    u"""
    Set a single option for ``nlp2_sparse_solve`` from an integer
    argument.

    ``nlp2_sparse_option_integer_set`` may be used to supply individual
    `int` options to ``nlp2_sparse_solve``.
    The initialization function ``nlp2_sparse_init`` **must** have been
    called before calling ``nlp2_sparse_option_integer_set``.

    For full information please refer to the NAG Library document for
    e04vm

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04vmf.html

    Parameters
    ----------
    string : str
        A single valid keyword of an `int` option (as described in
        [Other Parameters for nlp2_sparse_solve]).

    ivalue : int
        An `int` value associated with the keyword in `string`.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``nlp2_sparse_init``.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Raises
    ------
    NagValueError
        (`errno` 1)
            The initialization function ``nlp2_sparse_init`` has not
            been called.

        (`errno` 2)
            The supplied option is invalid. Check that the keywords are
            neither ambiguous nor misspelt. The option string is
            *<value>* and `ivalue` = *<value>*.
    """

    _b_opt.nlp2_sparse_option_integer_set(
        string,
        ivalue,
        comm,
        io_manager,
    )

def nlp2_sparse_option_double_set(string, rvalue, comm, io_manager=None):
    u"""
    Set a single option for ``nlp2_sparse_solve`` from a real argument.

    ``nlp2_sparse_option_double_set`` may be used to supply individual
    `float` options to ``nlp2_sparse_solve``.
    The initialization function ``nlp2_sparse_init`` **must** have been
    called before calling ``nlp2_sparse_option_double_set``.

    For full information please refer to the NAG Library document for
    e04vn

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04vnf.html

    Parameters
    ----------
    string : str
        A single valid keyword of a `float` option (as described in
        [Other Parameters for nlp2_sparse_solve]).

    rvalue : float
        The value associated with the keyword in `string`.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``nlp2_sparse_init``.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Raises
    ------
    NagValueError
        (`errno` 1)
            The initialization function ``nlp2_sparse_init`` has not
            been called.

        (`errno` 2)
            The supplied option is invalid. Check that the keywords are
            neither ambiguous nor misspelt. The option string is
            *<value>* and `rvalue` = *<value>*.
    """

    _b_opt.nlp2_sparse_option_double_set(
        string,
        rvalue,
        comm,
        io_manager,
    )

def nlp2_sparse_option_integer_get(string, comm, io_manager=None):
    u"""
    Get the setting of an integer valued option of
    ``nlp2_sparse_solve``.

    ``nlp2_sparse_option_integer_get`` is used to get the value of an
    `int` option. ``nlp2_sparse_option_integer_get`` can be used before
    or after calling ``nlp2_sparse_solve``.

    For full information please refer to the NAG Library document for
    e04vr

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04vrf.html

    Parameters
    ----------
    string : str
        A single valid keyword of an `int` option (as described in
        [Other Parameters for nlp2_sparse_solve]).

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``nlp2_sparse_init``.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    ivalue : int
        The `int` value associated with the keyword in `string`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            The initialization function ``nlp2_sparse_init`` has not
            been called.

        (`errno` 2)
            The supplied option string is invalid. Check that the
            keywords are neither ambiguous nor misspelt. The option
            string is *<value>*.
    """

    _return_scalar = _b_opt.nlp2_sparse_option_integer_get(
        string,
        comm,
        io_manager,
    )
    _ivalue_py = _return_scalar
    return _ivalue_py

def nlp2_sparse_option_double_get(string, comm, io_manager=None):
    u"""
    Get the setting of a real valued option of ``nlp2_sparse_solve``.

    ``nlp2_sparse_option_double_get`` is used to get the value of a
    `float` option. ``nlp2_sparse_option_double_get`` can be used before
    or after calling ``nlp2_sparse_solve``.

    For full information please refer to the NAG Library document for
    e04vs

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04vsf.html

    Parameters
    ----------
    string : str
        A single valid keyword of a `float` option (as described in
        [Other Parameters for nlp2_sparse_solve]).

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``nlp2_sparse_init``.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    rvalue : float
        The `float` value associated with the keyword in `string`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            The initialization function ``nlp2_sparse_init`` has not
            been called.

        (`errno` 2)
            The supplied option string is invalid. Check that the
            keywords are neither ambiguous nor misspelt. The option
            string is *<value>*.
    """

    _return_scalar = _b_opt.nlp2_sparse_option_double_get(
        string,
        comm,
        io_manager,
    )
    _rvalue_py = _return_scalar
    return _rvalue_py

def nlp1_init(rname):
    u"""
    Initialization function for ``uncon_conjgrd_comp``, ``lp_solve``,
    ``lsq_lincon_solve``, ``qp_dense_solve``,
    ``qpconvex1_sparse_solve``, ``nlp1_solve``, ``nlp1_rcomm``,
    ``nlp1_sparse_solve`` and ``lsq_gencon_deriv``.

    ``nlp1_init`` is used to initialize functions
    ``uncon_conjgrd_comp``, ``lp_solve``, ``lsq_lincon_solve``,
    ``qp_dense_solve``, ``qpconvex1_sparse_solve``, ``nlp1_solve``,
    ``nlp1_rcomm``, ``nlp1_sparse_solve`` and ``lsq_gencon_deriv``.

    For full information please refer to the NAG Library document for
    e04wb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04wbf.html

    Parameters
    ----------
    rname : str
        The name of the function to be initialized.

    Returns
    -------
    comm : dict, communication object
        Communication structure.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `rname` = *<value>* was not a valid function name.

    See Also
    --------
    :meth:`naginterfaces.library.examples.mip.ilp_dense_ex.main`
    :meth:`naginterfaces.library.examples.opt.lsq_gencon_deriv_ex.main`
    :meth:`naginterfaces.library.examples.opt.nlp1_rcomm_ex.main`
    :meth:`naginterfaces.library.examples.opt.nlp1_solve_ex.main`
    :meth:`naginterfaces.library.examples.opt.nlp1_sparse_solve_ex.main`
    """

    _return_scalar = _b_opt.nlp1_init(
        rname,
    )
    _comm_py = _return_scalar
    return _comm_py

def nlp2_init(io_manager=None):
    u"""
    Initialization function for ``nlp2_solve``.

    ``nlp2_init`` is used to initialize the function ``nlp2_solve``.

    For full information please refer to the NAG Library document for
    e04wc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04wcf.html

    Parameters
    ----------
    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    comm : dict, communication object
        Communication structure.

    See Also
    --------
    :meth:`naginterfaces.library.examples.opt.nlp2_solve_ex.main`
    """

    _return_scalar = _b_opt.nlp2_init(
        io_manager,
    )
    _comm_py = _return_scalar
    return _comm_py

def nlp2_solve(a, bl, bu, objfun, istate, ccon, cjac, clamda, h, x, comm, confun=None, data=None, io_manager=None, spiked_sorder='C'):
    u"""
    Nonlinear programming (NLP), dense, active-set SQP method, using
    function values and optionally first derivatives.

    Note: this function uses optional algorithmic parameters, see also:
    ``nlp2_option_file``, ``nlp2_option_string``,
    ``nlp2_option_integer_set``, ``nlp2_option_double_set``,
    ``nlp2_init``.

    ``nlp2_solve`` is designed to minimize an arbitrary smooth function
    subject to constraints (which may include simple bounds on the
    variables, linear constraints and smooth nonlinear constraints)
    using a sequential quadratic programming (SQP) method.
    As many first derivatives as possible should be supplied by you; any
    unspecified derivatives are approximated by finite differences.
    It is not intended for large sparse problems.

    ``nlp2_solve`` may also be used for unconstrained, bound-constrained
    and linearly constrained optimization.

    ``nlp2_solve`` uses **forward** **communication** for evaluating the
    objective function, the nonlinear constraint functions, and any of
    their derivatives.

    The initialization function ``nlp2_init`` **must** have been called
    before to calling ``nlp2_solve``.

    For full information please refer to the NAG Library document for
    e04wd

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04wdf.html

    Parameters
    ----------
    a : float, array-like, shape (nclin, :)
        Note: the required extent for this argument in dimension 2 is
        determined as follows: if nclin>0: n; otherwise: 1.

        The ith row of `a` contains the ith row of the matrix A_L of
        general linear constraints in (1). That is, the ith row contains
        the coefficients of the ith general linear constraint, for i =
        1,2,...,nclin.

    bl : float, array-like, shape (n+nclin+ncnln)
        `bl` must contain the lower bounds for all the constraints

    bu : float, array-like, shape (n+nclin+ncnln)
        `bu` must contain the upper bounds for all the constraints

    objfun : callable (mode, objf, grad) = objfun(mode, x, grad, nstate,
        data=None)
        `objfun` must calculate the objective function F(x) and
        (optionally) its gradient g(x) = dF/dx for a specified n-vector
        x.

        Parameters
        ~~~~~~~~~~
        mode : int
            Is set by ``nlp2_solve`` to indicate which values must be
            assigned during each call of `objfun`. Only the following
            values need be assigned:

            `mode` = 0

                `objf`.

            `mode` = 1

                All available elements of `grad`.

            `mode` = 2

                `objf` and all available elements of `grad`.

        x : float, ndarray, shape (n)
            x, the vector of variables at which the objective function
            and/or all available elements of its gradient are to be
            evaluated.

        grad : float, ndarray, shape (n)
            The elements of `grad` are set to special values.

        nstate : int
            If `nstate` = 1, then ``nlp2_solve`` is calling `objfun` for
            the first time. This argument setting allows you to save
            computation time if certain data must be read or calculated
            only once.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        mode : int
            May be used to indicate that you are unable or unwilling to
            evaluate the objective function at the current x.

            During the linesearch, the function is evaluated at points
            of the form x = x_k+alpha p_k after they have already been
            evaluated satisfactorily at x_k.

            For any such x, if you set `mode` = -1, ``nlp2_solve`` will
            reduce alpha and evaluate the functions again (closer to
            x_k, where they are more likely to be defined).

            If for some reason you wish to terminate the current
            problem, set `mode` < -1.

        objf : float
            If `mode` = 0 or 2, `objf` must be set to the value of the
            objective function at x.

        grad : float, array-like, shape (n)
            If `mode` = 1 or 2, `grad` must return the available
            elements of the gradient evaluated at x, i.e., `grad`[i-1]
            contains the partial derivative dF/dx_i.

    istate : int, array-like, shape (n+nclin+ncnln)
        Is an integer array that need not be initialized if
        ``nlp2_solve`` is called with the 'Cold Start' option (the
        default).

        If option 'Warm Start' has been chosen, every element of
        `istate` must be set.

        If ``nlp2_solve`` has just been called on a problem with the
        same dimensions, `istate` already contains valid values.

        Otherwise, `istate`[j-1] should indicate whether either of the
        constraints r_j(x) >= l_j or r_j(x) <= u_j is expected to be
        active at a solution of (1).

        The ordering of `istate` is the same as for `bl`, `bu` and r(x),
        i.e., the first n components of `istate` refer to the upper and
        lower bounds on the variables, the next nclin refer to the
        bounds on A_L x, and the last ncnln refer to the bounds on c(x).

        Possible values of `istate`[i-1] follow:

        [table omitted]

        The values 1, 2 or 3 all have the same effect when `bl`[j-1] =
        `bu`[j-1].

        If necessary, ``nlp2_solve`` will override your specification of
        `istate`, so that a poor choice will not cause the algorithm to
        fail.

    ccon : float, array-like, shape (max(1, ncnln))
        `ccon` need not be initialized if the (default) option 'Cold
        Start' is used.

        For a 'Warm Start', and if ncnln > 0, `ccon` contains values of
        the nonlinear constraint functions c_i, for i = 1,2,...,ncnln,
        calculated in a previous call to ``nlp2_solve``.

    cjac : float, array-like, shape (ncnln, :)
        Note: the required extent for this argument in dimension 2 is
        determined as follows: if ncnln>0: n; otherwise: 1.

        In general, `cjac` need not be initialized before the call to
        ``nlp2_solve``. However, if 'Derivative Level' = 2 or 3, any
        constant elements of `cjac` may be initialized. Such elements
        need not be reassigned on subsequent calls to `confun`.

    clamda : float, array-like, shape (n+nclin+ncnln)
        Need not be set if the (default) option 'Cold Start' is used.

        If the option 'Warm Start' has been chosen, `clamda`[j-1] must
        contain a multiplier estimate for each nonlinear constraint,
        with a sign that matches the status of the constraint specified
        by the `istate` array, for j = n+nclin+1,...,n+nclin+ncnln.

        The remaining elements need not be set.

        If the jth constraint is defined as 'inactive' by the initial
        value of the `istate` array (i.e., `istate`[j-1] = 0),
        `clamda`[j-1] should be zero; if the jth constraint is an
        inequality active at its lower bound (i.e., `istate`[j-1] = 1),
        `clamda`[j-1] should be non-negative; if the jth constraint is
        an inequality active at its upper bound (i.e., `istate`[j-1] =
        2), `clamda`[j-1] should be non-positive.

        If necessary, the function will modify `clamda` to match these
        rules.

    h : float, array-like, shape (n, n)
        `h` need not be initialized if the (default) option 'Cold Start'
        is used, and will be set to the identity.

        If the option 'Warm Start' has been chosen, `h` provides the
        initial approximation of the Hessian of the Lagrangian, i.e.,
        `h`[i-1,j-1]≈d^2L(x,lambda)/(dx_i dx_j), where L(x,lambda) =
        F(x)- c(x)^T lambda and lambda is an estimate of the Lagrange
        multipliers. `h` must be a positive definite matrix.

    x : float, array-like, shape (n)
        An initial estimate of the solution.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``nlp2_init``.

    confun : None or callable (mode, ccon, cjac) = confun(mode, needc,
        x, cjac, nstate, data=None), optional
        Note: if this argument is None then a NAG-supplied facility will
        be used.

        `confun` must calculate the vector c(x) of nonlinear constraint
        functions and (optionally) its Jacobian, dc/dx, for a specified
        n-vector x.

        If there are no nonlinear constraints (i.e., ncnln = 0),
        ``nlp2_solve`` will never call `confun`, so it may be None. If
        there are nonlinear constraints, the first call to `confun` will
        occur before the first call to `objfun`.

        If all constraint gradients (Jacobian elements) are known (i.e.,
        'Derivative Level' = 2 or 3), any `constant` elements may be
        assigned to `cjac` once only at the start of the optimization.

        An element of `cjac` that is not subsequently assigned in
        `confun` will retain its initial value throughout.

        Constant elements may be loaded in `cjac` during the first call
        to `confun` (signalled by the value of `nstate` = 1).

        The ability to preload constants is useful when many Jacobian
        elements are identically zero, in which case `cjac` may be
        initialized to zero and nonzero elements may be reset by
        `confun`.

        It must be emphasized that, if 'Derivative Level' < 2,
        unassigned elements of `cjac` are `not` treated as constant;
        they are estimated by finite differences, at nontrivial expense.

        Parameters
        ~~~~~~~~~~
        mode : int
            Is set by ``nlp2_solve`` to indicate which values must be
            assigned during each call of `confun`. Only the following
            values need be assigned, for each value of i such that
            `needc`[i-1] > 0:

            `mode` = 0

                The components of `ccon` corresponding to positive
                values in `needc` must be set. Other components and the
                array `cjac` are ignored.

            `mode` = 1

                The known components of the rows of `cjac` corresponding
                to positive values in `needc` must be set. Other rows of
                `cjac` and the array `ccon` will be ignored.

            `mode` = 2

                Only the elements of `ccon` corresponding to positive
                values of `needc` need to be set (and similarly for the
                known components of the rows of `cjac`).

        needc : int, ndarray, shape (ncnln)
            The indices of the elements of `ccon` and/or `cjac` that
            must be evaluated by `confun`. If `needc`[i-1] > 0, the ith
            element of `ccon` and/or the available elements of the ith
            row of `cjac` (see argument `mode`) must be evaluated at x.

        x : float, ndarray, shape (n)
            x, the vector of variables at which the constraint functions
            and/or the available elements of the constraint Jacobian are
            to be evaluated.

        cjac : float, ndarray, shape (max(1, ncnln), n)
            The elements of `cjac` are set to special values that enable
            ``nlp2_solve`` to detect whether they are reset by `confun`.

        nstate : int
            If `nstate` = 1, then ``nlp2_solve`` is calling `confun` for
            the first time. This argument setting allows you to save
            computation time if certain data must be read or calculated
            only once.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        mode : int
            May be used to indicate that you are unable or unwilling to
            evaluate the constraint functions at the current x.

            During the linesearch, the constraint functions are
            evaluated at points of the form x = x_k+alpha p_k after they
            have already been evaluated satisfactorily at x_k.

            At any such alpha, if you set `mode` = -1, ``nlp2_solve``
            will evaluate the functions at some point closer to x_k
            (where they are more likely to be defined).

            If for some reason you wish to terminate the current
            problem, set `mode` < -1.

        ccon : float, array-like, shape (max(1, ncnln))
            If `needc`[i-1] > 0 and `mode` = 0 or 2, `ccon`[i-1] must
            contain the value of the ith constraint at x. The remaining
            elements of `ccon`, corresponding to the non-positive
            elements of `needc`, are ignored.

        cjac : float, array-like, shape (max(1, ncnln), n)
            If `needc`[i-1] > 0 and `mode` = 1 or 2, the ith row of
            `cjac` must contain the available elements of the vector
            nabla c_i given by

                nabla c_i = (dc_i/dx_1,dc_i/dx_2,...,dc_i/dx_n)^T,

            where dc_i/dx_j is the partial derivative of the ith
            constraint with respect to the jth variable, evaluated at
            the point x. See also the argument `nstate`. The remaining
            rows of `cjac`, corresponding to non-positive elements of
            `needc`, are ignored.

            If all elements of the constraint Jacobian are known (i.e.,
            'Derivative Level' = 2 or 3), any constant elements may be
            assigned to `cjac` one time only at the start of the
            optimization.

            An element of `cjac` that is not subsequently assigned in
            `confun` will retain its initial value throughout.

            Constant elements may be loaded into `cjac` during the first
            call to `confun` (signalled by the value `nstate` = 1).

            The ability to preload constants is useful when many
            Jacobian elements are identically zero, in which case `cjac`
            may be initialized to zero and nonzero elements may be reset
            by `confun`.

            Note that constant nonzero elements do affect the values of
            the constraints.

            Thus, if `cjac`[i-1,j-1] is set to a constant value, it need
            not be reset in subsequent calls to `confun`, but the value
            `cjac`[i-1,j-1]*`x`[j-1] must nonetheless be added to
            `ccon`[i-1].

            For example, if `cjac`[0,0] = 2 and `cjac`[0,1] = -5, then
            the term 2*`x`[0]-5*`x`[1] must be included in the
            definition of `ccon`[0].

            It must be emphasized that, if 'Derivative Level' = 0 or 1,
            unassigned elements of `cjac` are not treated as constant;
            they are estimated by finite differences, at nontrivial
            expense.

            If you do not supply a value for the option 'Difference
            Interval', an interval for each element of x is computed
            automatically at the start of the optimization.

            The automatic procedure can usually identify constant
            elements of `cjac`, which are then computed once only by
            finite differences.

    data : arbitrary, optional
        User-communication data for callback functions.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    spiked_sorder : str, optional
        If `a` and `h` are spiked (i.e., have unit extent in all but one
        dimension, or have size 1), `spiked_sorder` selects the storage
        order to associate with them in the NAG Engine:

        spiked_sorder = 'C'
          row-major storage will be used;

        spiked_sorder = 'F'
          column-major storage will be used.

        Two-dimensional arrays returned from callback functions in this
        routine must then use the same storage order.

    Returns
    -------
    majits : int
        The number of major iterations performed.

    istate : int, ndarray, shape (n+nclin+ncnln)
        Describes the status of the constraints l <= r(x) <= u. For the
        jth lower or upper bound, j = 1,2,...,n+nclin+ncnln, the
        possible values of `istate`[j-1] are as follows (see Figure
        [label omitted]). delta is the appropriate feasibility
        tolerance.

        [table omitted]

        These values of `istate` are labelled in the printed solution
        according to Table [label omitted].

        [table omitted]

    ccon : float, ndarray, shape (max(1, ncnln))
        If ncnln > 0, `ccon`[i-1] contains the value of the ith
        nonlinear constraint function c_i at the final iterate, for i =
        1,2,...,ncnln.

        If ncnln = 0, the array `ccon` is not referenced.

    cjac : float, ndarray, shape (ncnln, :)
        If ncnln > 0, `cjac` contains the Jacobian matrix of the
        nonlinear constraint functions at the final iterate, i.e.,
        `cjac`[i-1,j-1] contains the partial derivative of the ith
        constraint function with respect to the jth variable, for j =
        1,2,...,n, for i = 1,2,...,ncnln. (See the discussion of
        argument `cjac` under `confun`.)

        If ncnln = 0, the array `cjac` is not referenced.

    clamda : float, ndarray, shape (n+nclin+ncnln)
        The values of the QP multipliers from the last QP subproblem.
        `clamda`[j-1] should be non-negative if `istate`[j-1] = 1 and
        non-positive if `istate`[j-1] = 2.

    objf : float
        The value of the objective function at the final iterate.

    grad : float, ndarray, shape (n)
        The gradient of the objective function (or its finite difference
        approximation) at the final iterate.

    h : float, ndarray, shape (n, n)
        Contains the Hessian of the Lagrangian at the final estimate x.

    x : float, ndarray, shape (n)
        The final estimate of the solution.

    Other Parameters
    ----------------
    'Central Difference Interval' : float
        Default = epsilon_r^1/3

        When 'Derivative Level' < 3, the central-difference interval r
        is used near an optimal solution to obtain more accurate (but
        more expensive) estimates of gradients.
        Twice as many function evaluations are required compared to
        forward differencing.
        The interval used for the jth variable is h_j = r(1+|x_j|).
        The resulting derivative estimates should be accurate to O(r^2),
        unless the functions are badly scaled.

        If you supply a value for this option, a small value between 0.0
        and 1.0 is appropriate.

    'Check Frequency' : int
        Default = 60

        Every ith minor iteration after the most recent basis
        factorization, a numerical test is made to see if the current
        solution x satisfies the general linear constraints (the linear
        constraints and the linearized nonlinear constraints, if any).
        The constraints are of the form Ax-s = b, where s is the set of
        slack variables.
        To perform the numerical test, the residual vector r = b-Ax+s is
        computed.
        If the largest component of r is judged to be too large, the
        current basis is refactorized and the basic variables are
        recomputed to satisfy the general constraints more accurately.
        If i <= 0, the value of i = 99999999 is used and effectively no
        checks are made.

        'Check Frequency' = 1 is useful for debugging purposes, but
        otherwise this option should not be needed.

    'Cold Start' : valueless
        Default

        This option controls the specification of the initial working
        set in the procedure for finding a feasible point for the linear
        constraints and bounds and in the first QP subproblem
        thereafter.
        With a 'Cold Start', the first working set is chosen by
        ``nlp2_solve`` based on the values of the variables and
        constraints at the initial point.
        Broadly speaking, the initial working set will include equality
        constraints and bounds or inequality constraints that violate or
        'nearly' satisfy their bounds (to within 'Crash Tolerance').

        With a 'Warm Start', you must set the `istate` array and define
        `clamda` and `h` as discussed in [Parameters]. `istate` values
        associated with bounds and linear constraints determine the
        initial working set of the procedure to find a feasible point
        with respect to the bounds and linear constraints. `istate`
        values associated with nonlinear constraints determine the
        initial working set of the first QP subproblem after such a
        feasible point has been found. ``nlp2_solve`` will override your
        specification of `istate` if necessary, so that a poor choice of
        the working set will not cause a fatal error.
        For instance, any elements of `istate` which are set to -2, -1
        or 4 will be reset to zero, as will any elements which are set
        to 3 when the corresponding elements of `bl` and `bu` are not
        equal.
        A warm start will be advantageous if a good estimate of the
        initial working set is available -- for example, when
        ``nlp2_solve`` is called repeatedly to solve related problems.

    'Warm Start' : valueless
        This option controls the specification of the initial working
        set in the procedure for finding a feasible point for the linear
        constraints and bounds and in the first QP subproblem
        thereafter.
        With a 'Cold Start', the first working set is chosen by
        ``nlp2_solve`` based on the values of the variables and
        constraints at the initial point.
        Broadly speaking, the initial working set will include equality
        constraints and bounds or inequality constraints that violate or
        'nearly' satisfy their bounds (to within 'Crash Tolerance').

        With a 'Warm Start', you must set the `istate` array and define
        `clamda` and `h` as discussed in [Parameters]. `istate` values
        associated with bounds and linear constraints determine the
        initial working set of the procedure to find a feasible point
        with respect to the bounds and linear constraints. `istate`
        values associated with nonlinear constraints determine the
        initial working set of the first QP subproblem after such a
        feasible point has been found. ``nlp2_solve`` will override your
        specification of `istate` if necessary, so that a poor choice of
        the working set will not cause a fatal error.
        For instance, any elements of `istate` which are set to -2, -1
        or 4 will be reset to zero, as will any elements which are set
        to 3 when the corresponding elements of `bl` and `bu` are not
        equal.
        A warm start will be advantageous if a good estimate of the
        initial working set is available -- for example, when
        ``nlp2_solve`` is called repeatedly to solve related problems.

    'Crash Option' : int
        Default = 3

        If a 'Cold Start' is specified, an internal Crash procedure is
        used to select an initial basis from certain rows and columns of
        the constraint matrix (A -I).
        The option 'Crash Option' i determines which rows and columns of
        A are eligible initially, and how many times the Crash procedure
        is called.
        Columns of -I are used to pad the basis where necessary.

        [table omitted]

        If i >= 1, certain slacks on inequality rows are selected for
        the basis first. (If i >= 2, numerical values are used to
        exclude slacks that are close to a bound).
        The Crash procedure then makes several passes through the
        columns of A, searching for a basis matrix that is essentially
        triangular.
        A column is assigned to 'pivot' on a particular row if the
        column contains a suitably large element in a row that has not
        yet been assigned. (The pivot elements ultimately form the
        diagonals of the triangular basis.) For remaining unassigned
        rows, slack variables are inserted to complete the basis.

        The 'Crash Tolerance' r allows the starting Crash procedure to
        ignore certain 'small' nonzeros in each column of A.
        If a_max is the largest element in column j, other nonzeros of
        a_{ij} in the columns are ignored if |a_{ij}| <= a_max*r. (To be
        meaningful, r must be in the range 0 <= r < 1.)

        When r > 0.0, the basis obtained by the Crash procedure may not
        be strictly triangular, but it is likely to be nonsingular and
        almost triangular.
        The intention is to obtain a starting basis containing more
        columns of A and fewer (arbitrary) slacks.
        A feasible solution may be reached sooner on some problems.

        For example, suppose the first m columns of A form the matrix
        shown under 'LU Factor Tolerance'; i.e., a tridiagonal matrix
        with entries -1, 4, -1.
        To help the Crash procedure choose all m columns for the initial
        basis, we would specify a 'Crash Tolerance' of r for some value
        of r > 0.5.

    'Crash Tolerance' : float
        Default = 0.1

        If a 'Cold Start' is specified, an internal Crash procedure is
        used to select an initial basis from certain rows and columns of
        the constraint matrix (A -I).
        The option 'Crash Option' i determines which rows and columns of
        A are eligible initially, and how many times the Crash procedure
        is called.
        Columns of -I are used to pad the basis where necessary.

        [table omitted]

        If i >= 1, certain slacks on inequality rows are selected for
        the basis first. (If i >= 2, numerical values are used to
        exclude slacks that are close to a bound).
        The Crash procedure then makes several passes through the
        columns of A, searching for a basis matrix that is essentially
        triangular.
        A column is assigned to 'pivot' on a particular row if the
        column contains a suitably large element in a row that has not
        yet been assigned. (The pivot elements ultimately form the
        diagonals of the triangular basis.) For remaining unassigned
        rows, slack variables are inserted to complete the basis.

        The 'Crash Tolerance' r allows the starting Crash procedure to
        ignore certain 'small' nonzeros in each column of A.
        If a_max is the largest element in column j, other nonzeros of
        a_{ij} in the columns are ignored if |a_{ij}| <= a_max*r. (To be
        meaningful, r must be in the range 0 <= r < 1.)

        When r > 0.0, the basis obtained by the Crash procedure may not
        be strictly triangular, but it is likely to be nonsingular and
        almost triangular.
        The intention is to obtain a starting basis containing more
        columns of A and fewer (arbitrary) slacks.
        A feasible solution may be reached sooner on some problems.

        For example, suppose the first m columns of A form the matrix
        shown under 'LU Factor Tolerance'; i.e., a tridiagonal matrix
        with entries -1, 4, -1.
        To help the Crash procedure choose all m columns for the initial
        basis, we would specify a 'Crash Tolerance' of r for some value
        of r > 0.5.

    'Defaults' : valueless
        This special keyword may be used to reset all options to their
        default values.

    'Derivative Level' : int
        Default = 3

        Option 'Derivative Level' specifies which nonlinear function
        gradients are known analytically and will be supplied to
        ``nlp2_solve`` by functions `objfun` and `confun`.

        [table omitted]

        The value i = 3 should be used whenever possible.
        It is the most reliable and will usually be the most efficient.

        If i = 0 or 2, ``nlp2_solve`` will `estimate` the missing
        components of the objective gradient, using finite differences.
        This may simplify the coding of `objfun`.
        However, it could increase the total run-time substantially
        (since a special call to `objfun` is required for each missing
        element), and there is less assurance that an acceptable
        solution will be located.
        If the nonlinear variables are not well scaled, it may be
        necessary to specify a non-default option 'Difference Interval'.

        If i = 0 or 1, ``nlp2_solve`` will estimate missing elements of
        the Jacobian.
        For each column of the Jacobian, one call to `confun` is needed
        to estimate all missing elements in that column, if any.

        At times, central differences are used rather than forward
        differences. (This is not under your control.)

    'Derivative Linesearch' : valueless
        Default

        At each major iteration a linesearch is used to improve the
        merit function. Option 'Derivative Linesearch' uses safeguarded
        cubic interpolation and requires both function and gradient
        values to compute estimates of the step alpha_k.
        If some analytic derivatives are not provided, or option
        'Nonderivative Linesearch' is specified, ``nlp2_solve`` employs
        a linesearch based upon safeguarded quadratic interpolation,
        which does not require gradient evaluations.

        A nonderivative linesearch can be slightly less robust on
        difficult problems, and it is recommended that the default be
        used if the functions and derivatives can be computed at
        approximately the same cost.
        If the gradients are very expensive relative to the functions, a
        nonderivative linesearch may give a significant decrease in
        computation time.

        If 'Nonderivative Linesearch' is selected, ``nlp2_solve``
        signals the evaluation of the linesearch by calling `objfun`
        with `mode` = 0.
        If the potential saving provided by a nonderivative linesearch
        is to be realised, it is essential that `objfun` be coded so
        that derivatives are not computed when `mode` = 0.

    'Nonderivative Linesearch' : valueless
        At each major iteration a linesearch is used to improve the
        merit function. Option 'Derivative Linesearch' uses safeguarded
        cubic interpolation and requires both function and gradient
        values to compute estimates of the step alpha_k.
        If some analytic derivatives are not provided, or option
        'Nonderivative Linesearch' is specified, ``nlp2_solve`` employs
        a linesearch based upon safeguarded quadratic interpolation,
        which does not require gradient evaluations.

        A nonderivative linesearch can be slightly less robust on
        difficult problems, and it is recommended that the default be
        used if the functions and derivatives can be computed at
        approximately the same cost.
        If the gradients are very expensive relative to the functions, a
        nonderivative linesearch may give a significant decrease in
        computation time.

        If 'Nonderivative Linesearch' is selected, ``nlp2_solve``
        signals the evaluation of the linesearch by calling `objfun`
        with `mode` = 0.
        If the potential saving provided by a nonderivative linesearch
        is to be realised, it is essential that `objfun` be coded so
        that derivatives are not computed when `mode` = 0.

    'Difference Interval' : float
        Default = sqrt(epsilon_r)

        This alters the interval r used to estimate gradients by forward
        differences.
        It does so in the following circumstances:

        -   in the interval ('cheap') phase of verifying the problem
            derivatives;

        -   for verifying the problem derivatives;

        -   for estimating missing derivatives.

        In all cases, a derivative with respect to x_j is estimated by
        perturbing that component of x to the value x_j+r(1+|x_j|), and
        then evaluating F(x) or c(x) at the perturbed point.
        The resulting gradient estimates should be accurate to O(r)
        unless the functions are badly scaled.
        Judicious alteration of r may sometimes lead to greater
        accuracy.

        If you supply a value for this option, a small value between 0.0
        and 1.0 is appropriate.

    'Dump File' : int
        Default = 0

        Options 'Dump File' and 'Load File' are similar to options
        'Punch File' and 'Insert File', but they record solution
        information in a manner that is more direct and more easily
        modified.
        A full description of information recorded in options 'Dump
        File' and 'Load File' is given in Gill `et al.` (2005a).

        If i_1 > 0, the last solution obtained will be output to the
        file with unit number i_1.

        If i_2 > 0, the 'Load File', containing basis information, will
        be read.
        The file will usually have been output previously as a 'Dump
        File'.
        The file will not be accessed if options 'Old Basis File' or
        'Insert File' are specified.

    'Load File' : int
        Default = 0

        Options 'Dump File' and 'Load File' are similar to options
        'Punch File' and 'Insert File', but they record solution
        information in a manner that is more direct and more easily
        modified.
        A full description of information recorded in options 'Dump
        File' and 'Load File' is given in Gill `et al.` (2005a).

        If i_1 > 0, the last solution obtained will be output to the
        file with unit number i_1.

        If i_2 > 0, the 'Load File', containing basis information, will
        be read.
        The file will usually have been output previously as a 'Dump
        File'.
        The file will not be accessed if options 'Old Basis File' or
        'Insert File' are specified.

    'Elastic Weight' : float
        Default = 10^4

        This keyword determines the initial weight gamma associated with
        the problem (1) (see [Treatment of Constraint Infeasibilities]).

        At major iteration k, if elastic mode has not yet started, a
        scale factor sigma_k = 1+||g(x_k)||_infinity is defined from the
        current objective gradient.
        Elastic mode is then started if the QP subproblem is infeasible,
        or the QP dual variables are larger in magnitude than sigma_k r.
        The QP is resolved in elastic mode with gamma = sigma_k r.

        Thereafter, major iterations continue in elastic mode until they
        converge to a point that is optimal for (1) (see [Treatment of
        Constraint Infeasibilities]).
        If the point is feasible for equation (1) (v = w = 0), it is
        declared locally optimal.
        Otherwise, gamma is increased by a factor of 10 and major
        iterations continue.
        If gamma has already reached a maximum allowable value, equation
        (1) is declared locally infeasible.

    'Expand Frequency' : int
        Default = 10000

        This option is part of the anti-cycling procedure designed to
        make progress even on highly degenerate problems.

        For linear models, the strategy is to force a positive step at
        every iteration, at the expense of violating the bounds on the
        variables by a small amount.
        Suppose that the option 'Minor Feasibility Tolerance' is delta.
        Over a period of i iterations, the tolerance actually used by
        ``nlp2_solve`` increases from 0.5 delta to delta (in steps of
        0.5 delta/i).

        For nonlinear models, the same procedure is used for iterations
        in which there is only one superbasic variable. (Cycling can
        occur only when the current solution is at a vertex of the
        feasible region.) Thus, zero steps are allowed if there is more
        than one superbasic variable, but otherwise positive steps are
        enforced.

        Increasing i helps reduce the number of slightly infeasible
        nonbasic variables (most of which are eliminated during a
        resetting procedure).
        However, it also diminishes the freedom to choose a large pivot
        element (see option 'Pivot Tolerance').

    'Factorization Frequency' : int
        Default = 50

        At most i basis changes will occur between factorizations of the
        basis matrix.

        With linear programs, the basis factors are usually updated
        every iteration.
        The default i is reasonable for typical problems.
        Higher values up to i = 100 (say) may be more efficient on
        well-scaled problems.

        When the objective function is nonlinear, fewer basis updates
        will occur as an optimum is approached.
        The number of iterations between basis factorizations will
        therefore increase.
        During these iterations a test is made regularly (according to
        the option 'Check Frequency') to ensure that the general
        constraints are satisfied.
        If necessary the basis will be refactorized before the limit of
        i updates is reached.

    'Function Precision' : float
        Default = epsilon^0.8

        The `relative function precision` epsilon_r is intended to be a
        measure of the relative accuracy with which the functions can be
        computed.
        For example, if F(x) is computed as 1000.56789 for some relevant
        x and if the first 6 significant digits are known to be correct,
        the appropriate value for epsilon_r would be 1.0e-6.

        (Ideally the functions F(x) or c_i(x) should have magnitude of
        order 1.
        If all functions are substantially `less` than 1 in magnitude,
        epsilon_r should be the `absolute` precision.
        For example, if F(x) = 1.23456789e-4 at some point and if the
        first 6 significant digits are known to be correct, the
        appropriate value for epsilon_r would be 1.0e-10.)

        The default value of epsilon_r is appropriate for simple
        analytic functions.

        In some cases the function values will be the result of
        extensive computation, possibly involving a costly iterative
        procedure that can provide few digits of precision.
        Specifying an appropriate 'Function Precision' may lead to
        savings, by allowing the linesearch procedure to terminate when
        the difference between function values along the search
        direction becomes as small as the absolute error in the values.

    'Hessian Full Memory' : valueless
        Default if n <= 75

        These options select the method for storing and updating the
        approximate Hessian. (``nlp2_solve`` uses a quasi-Newton
        approximation to the Hessian of the Lagrangian.
        A BFGS update is applied after each major iteration.)

        If 'Hessian Full Memory' is specified, the approximate Hessian
        is treated as a dense matrix and the BFGS updates are applied
        explicitly.
        This option is most efficient when the number of variables n is
        not too large (say, less than 75).
        In this case, the storage requirement is fixed and one can
        expect 1-step Q-superlinear convergence to the solution.

        'Hessian Limited Memory' should be used on problems where n is
        very large.
        In this case a limited-memory procedure is used to update a
        diagonal Hessian approximation H_r a limited number of times.
        (Updates are accumulated as a list of vector pairs.
        They are discarded at regular intervals after H_r has been reset
        to their diagonal.)

    'Hessian Limited Memory' : valueless
        Default if n > 75

        These options select the method for storing and updating the
        approximate Hessian. (``nlp2_solve`` uses a quasi-Newton
        approximation to the Hessian of the Lagrangian.
        A BFGS update is applied after each major iteration.)

        If 'Hessian Full Memory' is specified, the approximate Hessian
        is treated as a dense matrix and the BFGS updates are applied
        explicitly.
        This option is most efficient when the number of variables n is
        not too large (say, less than 75).
        In this case, the storage requirement is fixed and one can
        expect 1-step Q-superlinear convergence to the solution.

        'Hessian Limited Memory' should be used on problems where n is
        very large.
        In this case a limited-memory procedure is used to update a
        diagonal Hessian approximation H_r a limited number of times.
        (Updates are accumulated as a list of vector pairs.
        They are discarded at regular intervals after H_r has been reset
        to their diagonal.)

    'Hessian Frequency' : int
        Default = 99999999

        If option 'Hessian Full Memory' is in effect and i BFGS updates
        have already been carried out, the Hessian approximation is
        reset to the identity matrix. (For certain problems, occasional
        resets may improve convergence, but in general they should not
        be necessary.)

        'Hessian Full Memory' and 'Hessian Frequency' = 10 have a
        similar effect to 'Hessian Limited Memory' and 'Hessian Updates'
        = 10 (except that the latter retains the current diagonal during
        resets).

    'Hessian Updates' : int
        Default = 'Hessian Frequency' if 'Hessian Full Memory', 10
        otherwise

        If option 'Hessian Limited Memory' is in effect and i BFGS
        updates have already been carried out, all but the diagonal
        elements of the accumulated updates are discarded and the
        updating process starts again.

        Broadly speaking, the more updates stored, the better the
        quality of the approximate Hessian.
        However, the more vectors stored, the greater the cost of each
        QP iteration.
        The default value is likely to give a robust algorithm without
        significant expense, but faster convergence can sometimes be
        obtained with significantly fewer updates (e.g., i = 5).

    'Infinite Bound Size' : float
        Default = 10^20

        If r > 0, r defines the 'infinite' bound bigbnd in the
        definition of the problem constraints.
        Any upper bound greater than or equal to bigbnd will be regarded
        as + infinity (and similarly any lower bound less than or equal
        to -bigbnd will be regarded as - infinity).
        If r < 0, the default value is used.

    'Iterations Limit' : int
        Default = max(10000,10max(n,n_L+n_N))

        The value of i specifies the maximum number of minor iterations
        allowed (i.e., iterations of the simplex method or the QP
        algorithm), summed over all major iterations. (See also the
        description of the option 'Minor Iterations Limit'.)

    'Linesearch Tolerance' : float
        Default = 0.9

        This tolerance, r, controls the accuracy with which a step
        length will be located along the direction of search each
        iteration.
        At the start of each linesearch a target directional derivative
        for the merit function is identified.
        This argument determines the accuracy to which this target value
        is approximated, and it must be a value in the range 0.0 <= r <=
        1.0.

        The default value r = 0.9 requests just moderate accuracy in the
        linesearch.

        If the nonlinear functions are cheap to evaluate, a more
        accurate search may be appropriate; try r = 0.1, 0.01 or 0.001.

        If the nonlinear functions are expensive to evaluate, a less
        accurate search may be appropriate. `If all gradients are
        known`, try r = 0.99. (The number of major iterations might
        increase, but the total number of function evaluations may
        decrease enough to compensate.)

        If not all gradients are known, a moderately accurate search
        remains appropriate.
        Each search will require only 1--5 function values (typically),
        but many function calls will then be needed to estimate missing
        gradients for the next iteration.

    'Nolist' : valueless
        Default

        Option 'List' enables printing of each option specification as
        it is supplied. 'Nolist' suppresses this printing.

    'List' : valueless
        Option 'List' enables printing of each option specification as
        it is supplied. 'Nolist' suppresses this printing.

    'LU Density Tolerance' : float
        Default = 0.6

        The density tolerance, r_1, is used during LU factorization of
        the basis matrix B.
        Columns of L and rows of U are formed one at a time, and the
        remaining rows and columns of the basis are altered
        appropriately.
        At any stage, if the density of the remaining matrix exceeds
        r_1, the Markowitz strategy for choosing pivots is terminated,
        and the remaining matrix is factored by a dense LU procedure.
        Raising the density tolerance towards 1.0 may give slightly
        sparser LU factors, with a slight increase in factorization
        time.

        The singularity tolerance, r_2, helps guard against
        ill-conditioned basis matrices.
        After B is refactorized, the diagonal elements of U are tested
        as follows: if |u_{jj}| <= r_2 or |u_{jj}| < r_2 max_i|u_{ij}|,
        the jth column of the basis is replaced by the corresponding
        slack variable. (This is most likely to occur after a restart.)

    'LU Singularity Tolerance' : float
        Default = epsilon^(2/3)

        The density tolerance, r_1, is used during LU factorization of
        the basis matrix B.
        Columns of L and rows of U are formed one at a time, and the
        remaining rows and columns of the basis are altered
        appropriately.
        At any stage, if the density of the remaining matrix exceeds
        r_1, the Markowitz strategy for choosing pivots is terminated,
        and the remaining matrix is factored by a dense LU procedure.
        Raising the density tolerance towards 1.0 may give slightly
        sparser LU factors, with a slight increase in factorization
        time.

        The singularity tolerance, r_2, helps guard against
        ill-conditioned basis matrices.
        After B is refactorized, the diagonal elements of U are tested
        as follows: if |u_{jj}| <= r_2 or |u_{jj}| < r_2 max_i|u_{ij}|,
        the jth column of the basis is replaced by the corresponding
        slack variable. (This is most likely to occur after a restart.)

    'LU Factor Tolerance' : float
        Default = 1.10

        The values of r_1 and r_2 affect the stability of the basis
        factorization B = LU, during refactorization and updates
        respectively.
        The lower triangular matrix L is a product of matrices of the
        form

            [table omitted]

        where the multipliers mu will satisfy |mu| <= r_i.
        The default values of r_1 and r_2 usually strike a good
        compromise between stability and sparsity.
        They must satisfy r_1, r_2 >= 1.0.

        For large and relatively dense problems, r_1 = 10.0 or 5.0 (say)
        may give a useful improvement in stability without impairing
        sparsity to a serious degree.

        For certain very regular structures (e.g., band matrices) it may
        be necessary to reduce r_1 and/or r_2 in order to achieve
        stability.
        For example, if the columns of A include a sub-matrix of the
        form

            [table omitted],

        one should set both r_1 and r_2 to values in the range 1.0 <=
        r_i < 4.0.

    'LU Update Tolerance' : float
        Default = 1.10

        The values of r_1 and r_2 affect the stability of the basis
        factorization B = LU, during refactorization and updates
        respectively.
        The lower triangular matrix L is a product of matrices of the
        form

            [table omitted]

        where the multipliers mu will satisfy |mu| <= r_i.
        The default values of r_1 and r_2 usually strike a good
        compromise between stability and sparsity.
        They must satisfy r_1, r_2 >= 1.0.

        For large and relatively dense problems, r_1 = 10.0 or 5.0 (say)
        may give a useful improvement in stability without impairing
        sparsity to a serious degree.

        For certain very regular structures (e.g., band matrices) it may
        be necessary to reduce r_1 and/or r_2 in order to achieve
        stability.
        For example, if the columns of A include a sub-matrix of the
        form

            [table omitted],

        one should set both r_1 and r_2 to values in the range 1.0 <=
        r_i < 4.0.

    'LU Partial Pivoting' : valueless
        Default

        The LU factorization implements a Markowitz-type search for
        pivots that locally minimize the fill-in subject to a threshold
        pivoting stability criterion.
        The default option is to use threshhold partial pivoting.
        The options 'LU Rook Pivoting' and 'LU Complete Pivoting' are
        more expensive than partial pivoting but are more stable and
        better at revealing rank, as long as 'LU Factor Tolerance' is
        not too large (say < 2.0).
        When numerical difficulties are encountered, ``nlp2_solve``
        automatically reduces the LU tolerance towards 1.0 and switches
        (if necessary) to rook or complete pivoting, before reverting to
        the default or specified options at the next refactorization
        (with 'System Information Yes', relevant messages are output to
        the 'Print File').

    'LU Complete Pivoting' : valueless
        The LU factorization implements a Markowitz-type search for
        pivots that locally minimize the fill-in subject to a threshold
        pivoting stability criterion.
        The default option is to use threshhold partial pivoting.
        The options 'LU Rook Pivoting' and 'LU Complete Pivoting' are
        more expensive than partial pivoting but are more stable and
        better at revealing rank, as long as 'LU Factor Tolerance' is
        not too large (say < 2.0).
        When numerical difficulties are encountered, ``nlp2_solve``
        automatically reduces the LU tolerance towards 1.0 and switches
        (if necessary) to rook or complete pivoting, before reverting to
        the default or specified options at the next refactorization
        (with 'System Information Yes', relevant messages are output to
        the 'Print File').

    'LU Rook Pivoting' : valueless
        The LU factorization implements a Markowitz-type search for
        pivots that locally minimize the fill-in subject to a threshold
        pivoting stability criterion.
        The default option is to use threshhold partial pivoting.
        The options 'LU Rook Pivoting' and 'LU Complete Pivoting' are
        more expensive than partial pivoting but are more stable and
        better at revealing rank, as long as 'LU Factor Tolerance' is
        not too large (say < 2.0).
        When numerical difficulties are encountered, ``nlp2_solve``
        automatically reduces the LU tolerance towards 1.0 and switches
        (if necessary) to rook or complete pivoting, before reverting to
        the default or specified options at the next refactorization
        (with 'System Information Yes', relevant messages are output to
        the 'Print File').

    'Major Feasibility Tolerance' : float
        Default = max(10^(-6),sqrt(epsilon))

        This tolerance, r, specifies how accurately the nonlinear
        constraints should be satisfied.
        The default value is appropriate when the linear and nonlinear
        constraints contain data to about that accuracy.

        Let v_max be the maximum nonlinear constraint violation,
        normalized by the size of the solution, which is required to
        satisfy

            v_max/ = max_i v_i/||x|| <= r,

        where v_i is the violation of the ith nonlinear constraint (i =
        1:n_L).

        In the major iteration log (see [Minor Iteration Log], v_max
        appears as the quantity labelled 'Feasible'.
        If some of the problem functions are known to be of low
        accuracy, a larger 'Major Feasibility Tolerance' may be
        appropriate.

    'Major Optimality Tolerance' : float
        Default = 2max(10^(-6),sqrt(epsilon))

        This tolerance, r, specifies the final accuracy of the dual
        variables.
        On successful termination, ``nlp2_solve`` will have computed a
        solution (x,s,pi) such that

            c_max = max_j c_j/||pi|| <= r,

        where c_j is an estimate of the complementarity slackness for
        variable j where j = 1:n+n_L+n_N.
        The values c_i are computed from the final QP solution using the
        reduced gradients d_j = g_j-pi^Ta_j (where g_j is the jth
        component of the objective gradient, a_j is the associated
        column of the constraint matrix (A -I), and pi is the set of QP
        dual variables):

            c_j = {[table omitted])

        In the 'Print File', c_max appears as the quantity labelled
        'Optimal'.

    'Major Iterations Limit' : int
        Default = max(1000,3max(n,n_L+n_N))

        This is the maximum number of major iterations allowed.
        It is intended to guard against an excessive number of
        linearizations of the constraints.
        If i = 0, optimality and feasibility are checked.

    'Major Print Level' : int
        Default = 000001

        This controls the amount of output to the options 'Print File'
        and 'Summary File' at each major iteration. 'Major Print Level'
        = 0 suppresses most output, except for error messages. 'Major
        Print Level' = 1 gives normal output for linear and nonlinear
        problems, and 'Major Print Level' = 11 gives additional details
        of the Jacobian factorization that commences each major
        iteration.

        In general, the value being specified may be thought of as a
        binary number of the form

            'Major Print Level' JFDXbs

        where each letter stands for a digit that is either 0 or 1 as
        follows:

        [table omitted]

        To obtain output of any items JFDXbs, set the corresponding
        digit to 1, otherwise to 0.

        If J = 1, the Jacobian matrix will be output column-wise at the
        start of each major iteration.
        Column j will be preceded by the value of the corresponding
        variable x_j and a key to indicate whether the variable is
        basic, superbasic or nonbasic. (Hence if J = 1, there is no
        reason to specify X = 1 unless the objective contains more
        nonlinear variables than the Jacobian.) A typical line of output
        is

        ::

          3 1.250000e+01 BS 1 1.00000e+00 4 2.00000e+00

        which would mean that x_3 is basic at value 12.5, and the third
        column of the Jacobian has elements of 1.0 and 2.0 in rows 1 and
        4.

    'Major Step Limit' : float
        Default = 2.0

        This argument limits the change in x during a linesearch.
        It applies to all nonlinear problems, once a 'feasible solution'
        or 'feasible subproblem' has been found.
        A linesearch determines a step alpha over the range 0 < alpha <=
        beta, where beta is 1 if there are nonlinear constraints, or is
        the step to the nearest upper or lower bound on x if all the
        constraints are linear.
        Normally, the first step length tried is alpha_1 = min(1,beta).

        (1) In some cases, such as f(x) = ae^(bx) or f(x) = ax^b, even a
            moderate change in the components of x can lead to
            floating-point overflow. The argument r is therefore used to
            define a limit beta_= r(1+||x||)/||p|| (where p is the
            search direction), and the first evaluation of f(x) is at
            the potentially smaller step length alpha_1 =
            min(1,beta_,beta).

        (#) Wherever possible, upper and lower bounds on x should be
            used to prevent evaluation of nonlinear functions at
            meaningless points. The option 'Major Step Limit' provides
            an additional safeguard. The default value r = 2.0 should
            not affect progress on well behaved problems, but setting r
            = 0.1 or 0.01 may be helpful when rapidly varying functions
            are present. A 'good' starting point may be required. An
            important application is to the class of nonlinear least
            squares problems.

        (#) In cases where several local optima exist, specifying a
            small value for r may help locate an optimum near the
            starting point.

    'Minimize' : valueless
        Default

        The keywords 'Minimize' and 'Maximize' specify the required
        direction of optimization.
        It applies to both linear and nonlinear terms in the objective.

        The keyword 'Feasible Point' means 'Ignore the objective
        function, while finding a feasible point for the linear and
        nonlinear constraints'.
        It can be used to check that the nonlinear constraints are
        feasible without altering the call to ``nlp2_solve``.

    'Maximize' : valueless
        The keywords 'Minimize' and 'Maximize' specify the required
        direction of optimization.
        It applies to both linear and nonlinear terms in the objective.

        The keyword 'Feasible Point' means 'Ignore the objective
        function, while finding a feasible point for the linear and
        nonlinear constraints'.
        It can be used to check that the nonlinear constraints are
        feasible without altering the call to ``nlp2_solve``.

    'Feasible Point' : valueless
        The keywords 'Minimize' and 'Maximize' specify the required
        direction of optimization.
        It applies to both linear and nonlinear terms in the objective.

        The keyword 'Feasible Point' means 'Ignore the objective
        function, while finding a feasible point for the linear and
        nonlinear constraints'.
        It can be used to check that the nonlinear constraints are
        feasible without altering the call to ``nlp2_solve``.

    'Minor Feasibility Tolerance' : valueless
        ``nlp2_solve`` tries to ensure that all variables eventually
        satisfy their upper and lower bounds to within this tolerance,
        r.
        This includes slack variables.
        Hence, general linear constraints should also be satisfied to
        within r.

        Feasibility with respect to nonlinear constraints is judged by
        the option 'Major Feasibility Tolerance' (not by r).

        If the bounds and linear constraints cannot be satisfied to
        within r, the problem is declared `infeasible`.
        If the corresponding sum of infeasibilities is quite small, it
        may be appropriate to raise r by a factor of 10 or 100.
        Otherwise, some error in the data should be suspected.

        Nonlinear functions will be evaluated only at points that
        satisfy the bounds and linear constraints.
        If there are regions where a function is undefined, every
        attempt should be made to eliminate these regions from the
        problem.

        For example, if f(x) = sqrt(x_1)+log(x_2), it is essential to
        place lower bounds on both variables.
        If r = 1.0e-6, the bounds x_1 >= 10^(-5) and x_2 >= 10^(-4)
        might be appropriate. (The log singularity is more serious.
        In general, keep x as far away from singularities as possible.)

        If 'Scale Option' >= 1, feasibility is defined in terms of the
        `scaled` problem (since it is then more likely to be
        meaningful).

        In reality, ``nlp2_solve`` uses r as a feasibility tolerance for
        satisfying the bounds on x and s in each QP subproblem.
        If the sum of infeasibilities cannot be reduced to zero, the QP
        subproblem is declared infeasible. ``nlp2_solve`` is then in
        `elastic mode` thereafter (with only the linearized nonlinear
        constraints defined to be elastic).
        See the description of the option 'Elastic Weight'.

    'Feasibility Tolerance' : float
        Default = max{10^(-6),sqrt(epsilon)}

        ``nlp2_solve`` tries to ensure that all variables eventually
        satisfy their upper and lower bounds to within this tolerance,
        r.
        This includes slack variables.
        Hence, general linear constraints should also be satisfied to
        within r.

        Feasibility with respect to nonlinear constraints is judged by
        the option 'Major Feasibility Tolerance' (not by r).

        If the bounds and linear constraints cannot be satisfied to
        within r, the problem is declared `infeasible`.
        If the corresponding sum of infeasibilities is quite small, it
        may be appropriate to raise r by a factor of 10 or 100.
        Otherwise, some error in the data should be suspected.

        Nonlinear functions will be evaluated only at points that
        satisfy the bounds and linear constraints.
        If there are regions where a function is undefined, every
        attempt should be made to eliminate these regions from the
        problem.

        For example, if f(x) = sqrt(x_1)+log(x_2), it is essential to
        place lower bounds on both variables.
        If r = 1.0e-6, the bounds x_1 >= 10^(-5) and x_2 >= 10^(-4)
        might be appropriate. (The log singularity is more serious.
        In general, keep x as far away from singularities as possible.)

        If 'Scale Option' >= 1, feasibility is defined in terms of the
        `scaled` problem (since it is then more likely to be
        meaningful).

        In reality, ``nlp2_solve`` uses r as a feasibility tolerance for
        satisfying the bounds on x and s in each QP subproblem.
        If the sum of infeasibilities cannot be reduced to zero, the QP
        subproblem is declared infeasible. ``nlp2_solve`` is then in
        `elastic mode` thereafter (with only the linearized nonlinear
        constraints defined to be elastic).
        See the description of the option 'Elastic Weight'.

    'Minor Iterations Limit' : int
        Default = 500

        If the number of minor iterations for the optimality phase of
        the QP subproblem exceeds i, then all nonbasic QP variables that
        have not yet moved are frozen at their current values and the
        reduced QP is solved to optimality.

        Note that more than i minor iterations may be necessary to solve
        the reduced QP to optimality.
        These extra iterations are necessary to ensure that the
        terminated point gives a suitable direction for the linesearch.

        In the major iteration log (see [Minor Iteration Log]) a t at
        the end of a line indicates that the corresponding QP was
        artificially terminated using the limit i.

        Compare with the option 'Iterations Limit', which defines an
        independent `absolute` limit on the `total` number of minor
        iterations (summed over all QP subproblems).

    'Minor Print Level' : int
        Default = 1

        This controls the amount of output to the 'Print File' and the
        'Summary File' during solution of the QP subproblems.
        The value of i has the following effect:

        [table omitted]

    'New Basis File' : int
        Default = 0

        'New Basis File' and 'Backup Basis File' are sometimes referred
        to as basis maps.
        They contain the most compact representation of the state of
        each variable.
        They are intended for restarting the solution of a problem at a
        point that was reached by an earlier run.
        For nontrivial problems, it is advisable to save basis maps at
        the end of a run, in order to restart the run if necessary.

        If i_1 > 0, a basis map will be saved on the file associated
        with unit i_1 every i_3th iteration.
        The first record of the file will contain the word
        ``PROCEEDING`` if the run is still in progress.
        A basis map will also be saved at the end of a run, with some
        other word indicating the final solution status.

        Use of i_2 > 0 is intended as a safeguard against losing the
        results of a long run.
        Suppose that a 'New Basis File' is being saved every 100 ('Save
        Frequency') iterations, and that ``nlp2_solve`` is about to save
        such a basis at iteration 2000.
        It is conceivable that the run may be interrupted during the
        next few milliseconds (in the middle of the save).
        In this case the Basis file will be corrupted and the run will
        have been essentially wasted.

        To eliminate this risk, both a 'New Basis File' and a 'Backup
        Basis File' may be specified. The following would be suitable
        for the above example:

        ::

           Backup Basis File 11
           New Basis File 12

        The current basis will then be saved every 100 iterations, first
        on the file associated with unit 12 and then immediately on the
        file associated with unit 11. If the run is interrupted at
        iteration 2000 during the save on the file associated with unit
        12, there will still be a usable basis on the file associated
        with unit 11 (corresponding to iteration 1900).

        Note that a new basis will be saved in 'New Basis File' at the
        end of a run if it terminates normally, but it will not be saved
        in 'Backup Basis File'.
        In the above example, if an optimum solution is found at
        iteration 2050 (or if the iteration limit is 2050), the final
        basis in the file associated with unit 12 will correspond to
        iteration 2050, but the last basis saved in the file associated
        with unit 11 will be the one for iteration 2000.

        A full description of information recorded in 'New Basis File'
        and 'Backup Basis File' is given in Gill `et al.` (2005a).

    'Backup Basis File' : int
        Default = 0

        'New Basis File' and 'Backup Basis File' are sometimes referred
        to as basis maps.
        They contain the most compact representation of the state of
        each variable.
        They are intended for restarting the solution of a problem at a
        point that was reached by an earlier run.
        For nontrivial problems, it is advisable to save basis maps at
        the end of a run, in order to restart the run if necessary.

        If i_1 > 0, a basis map will be saved on the file associated
        with unit i_1 every i_3th iteration.
        The first record of the file will contain the word
        ``PROCEEDING`` if the run is still in progress.
        A basis map will also be saved at the end of a run, with some
        other word indicating the final solution status.

        Use of i_2 > 0 is intended as a safeguard against losing the
        results of a long run.
        Suppose that a 'New Basis File' is being saved every 100 ('Save
        Frequency') iterations, and that ``nlp2_solve`` is about to save
        such a basis at iteration 2000.
        It is conceivable that the run may be interrupted during the
        next few milliseconds (in the middle of the save).
        In this case the Basis file will be corrupted and the run will
        have been essentially wasted.

        To eliminate this risk, both a 'New Basis File' and a 'Backup
        Basis File' may be specified. The following would be suitable
        for the above example:

        ::

           Backup Basis File 11
           New Basis File 12

        The current basis will then be saved every 100 iterations, first
        on the file associated with unit 12 and then immediately on the
        file associated with unit 11. If the run is interrupted at
        iteration 2000 during the save on the file associated with unit
        12, there will still be a usable basis on the file associated
        with unit 11 (corresponding to iteration 1900).

        Note that a new basis will be saved in 'New Basis File' at the
        end of a run if it terminates normally, but it will not be saved
        in 'Backup Basis File'.
        In the above example, if an optimum solution is found at
        iteration 2050 (or if the iteration limit is 2050), the final
        basis in the file associated with unit 12 will correspond to
        iteration 2050, but the last basis saved in the file associated
        with unit 11 will be the one for iteration 2000.

        A full description of information recorded in 'New Basis File'
        and 'Backup Basis File' is given in Gill `et al.` (2005a).

    'Save Frequency' : int
        Default = 100

        'New Basis File' and 'Backup Basis File' are sometimes referred
        to as basis maps.
        They contain the most compact representation of the state of
        each variable.
        They are intended for restarting the solution of a problem at a
        point that was reached by an earlier run.
        For nontrivial problems, it is advisable to save basis maps at
        the end of a run, in order to restart the run if necessary.

        If i_1 > 0, a basis map will be saved on the file associated
        with unit i_1 every i_3th iteration.
        The first record of the file will contain the word
        ``PROCEEDING`` if the run is still in progress.
        A basis map will also be saved at the end of a run, with some
        other word indicating the final solution status.

        Use of i_2 > 0 is intended as a safeguard against losing the
        results of a long run.
        Suppose that a 'New Basis File' is being saved every 100 ('Save
        Frequency') iterations, and that ``nlp2_solve`` is about to save
        such a basis at iteration 2000.
        It is conceivable that the run may be interrupted during the
        next few milliseconds (in the middle of the save).
        In this case the Basis file will be corrupted and the run will
        have been essentially wasted.

        To eliminate this risk, both a 'New Basis File' and a 'Backup
        Basis File' may be specified. The following would be suitable
        for the above example:

        ::

           Backup Basis File 11
           New Basis File 12

        The current basis will then be saved every 100 iterations, first
        on the file associated with unit 12 and then immediately on the
        file associated with unit 11. If the run is interrupted at
        iteration 2000 during the save on the file associated with unit
        12, there will still be a usable basis on the file associated
        with unit 11 (corresponding to iteration 1900).

        Note that a new basis will be saved in 'New Basis File' at the
        end of a run if it terminates normally, but it will not be saved
        in 'Backup Basis File'.
        In the above example, if an optimum solution is found at
        iteration 2050 (or if the iteration limit is 2050), the final
        basis in the file associated with unit 12 will correspond to
        iteration 2050, but the last basis saved in the file associated
        with unit 11 will be the one for iteration 2000.

        A full description of information recorded in 'New Basis File'
        and 'Backup Basis File' is given in Gill `et al.` (2005a).

    'New Superbasics Limit' : int
        Default = 99

        This option causes early termination of the QP subproblems if
        the number of free variables has increased significantly since
        the first feasible point.
        If the number of new superbasics is greater than i, the nonbasic
        variables that have not yet moved are frozen and the resulting
        smaller QP is solved to optimality.

        In the major iteration log (see [Major Iteration Log]), a t at
        the end of a line indicates that the QP was terminated early in
        this way.

    'Old Basis File' : int
        Default = 0

        If i > 0, the basis maps information will be obtained from this
        file. The file will usually have been output previously as a
        'New Basis File' or 'Backup Basis File'.
        A full description of information recorded in 'New Basis File'
        and 'Backup Basis File' is given in Gill `et al.` (2005a).

        The file will not be acceptable if the number of rows or columns
        in the problem has been altered.

    'Partial Price' : int
        Default = 1

        This argument is recommended for large problems that have
        significantly more variables than constraints.
        It reduces the work required for each 'pricing' operation (where
        a nonbasic variable is selected to become superbasic).
        When i = 1, all columns of the constraint matrix (A -I) are
        searched.
        Otherwise, A and I are partitioned to give i roughly equal
        segments A_j and I_j, for j = 1,2,...,i.
        If the previous pricing search was successful on A_{j-1} and
        I_{j-1}, the next search begins on the segments A_j and I_j.
        (All subscripts here are modulo i.) If a reduced gradient is
        found that is larger than some dynamic tolerance, the variable
        with the largest such reduced gradient (of appropriate sign) is
        selected to become superbasic.
        If nothing is found, the search continues on the next segments
        A_{j+1} and I_{j+1}, and so on.

        For time-stage models having t time periods, 'Partial Price' t
        (or t/2 or t/3) may be appropriate.

    'Pivot Tolerance' : float
        Default = epsilon^(2/3)

        During the solution of QP subproblems, the pivot tolerance is
        used to prevent columns entering the basis if they would cause
        the basis to become almost singular.

        When x changes to x+alpha p for some search direction p, a
        'ratio test' determines which component of x reaches an upper or
        lower bound first.
        The corresponding element of p is called the pivot element.
        Elements of p are ignored (and therefore cannot be pivot
        elements) if they are smaller than the pivot tolerance r.

        It is common for two or more variables to reach a bound at
        essentially the same time.
        In such cases, the 'Minor Feasibility Tolerance' (say, t)
        provides some freedom to maximize the pivot element and thereby
        improve numerical stability.
        Excessively small values of t should therefore not be specified.
        To a lesser extent, the 'Expand Frequency' (say, f) also
        provides some freedom to maximize the pivot element.
        Excessively `large` values of f should therefore not be
        specified.

    'Print File' : int
        Default = 0

        If i > 0, the following information is output to a file
        associated with unit i during the solution of each problem:

        -   a listing of the options;

        -   some statistics about the problem;

        -   the amount of storage available for the LU factorization of
            the basis matrix;

        -   notes about the initial basis resulting from a Crash
            procedure or a Basis file;

        -   the iteration log;

        -   basis factorization statistics;

        -   the exit errno condition and some statistics about the
            solution obtained;

        -   the printed solution, if requested.

        These items are described in [Further Comments] and [Monitoring
        Information].
        Further brief output may be directed to the 'Summary File'.

    'Print Frequency' : int
        Default = 100

        If i > 0, one line of the iteration log will be printed every
        ith iteration.
        A value such as i = 10 is suggested for those interested only in
        the final solution.
        If i <= 0, the value of i = 99999999 is used and effectively no
        checks are made.

    'Proximal Point Method' : int
        Default = 1

        i = 1 or 2 specifies minimization of ||x-x_0||_1 or
        1/2||x-x_0||_2^2 when the starting point x_0 is changed to
        satisfy the linear constraints (where x_0 refers to nonlinear
        variables).

    'Punch File' : int
        Default = 0

        The 'Punch File' from a previous run may be used as an 'Insert
        File' for a later run on the same problem.
        A full description of information recorded in 'Insert File' and
        'Punch File' is given in Gill `et al.` (2005a).

        If i_1 > 0, the final solution obtained will be output to the
        file. For linear programs, this format is compatible with
        various commercial systems.

        If i_2 > 0 the 'Insert File' containing basis information will
        be read from unit i_2. The file will usually have been output
        previously as a 'Punch File'.
        The file will not be accessed if 'Old Basis File' is specified.

    'Insert File' : int
        Default = 0

        The 'Punch File' from a previous run may be used as an 'Insert
        File' for a later run on the same problem.
        A full description of information recorded in 'Insert File' and
        'Punch File' is given in Gill `et al.` (2005a).

        If i_1 > 0, the final solution obtained will be output to the
        file. For linear programs, this format is compatible with
        various commercial systems.

        If i_2 > 0 the 'Insert File' containing basis information will
        be read from unit i_2. The file will usually have been output
        previously as a 'Punch File'.
        The file will not be accessed if 'Old Basis File' is specified.

    'QPSolver Cholesky' : valueless
        Default

        Specifies the active-set algorithm used to solve subproblem (1)
        (see [Treatment of Constraint Infeasibilities]). 'QPSolver
        Cholesky' holds the full Cholesky factor R of the reduced
        Hessian Z^THZ.
        As the QP iterations proceed, the dimension of R changes with
        the number of superbasic variables.
        If the number of superbasic variables needs to increase beyond
        the value of 'Reduced Hessian Dimension', the reduced Hessian
        cannot be stored and the solver switches to 'QPSolver CG'.
        The Cholesky solver is reactivated if the number of superbasics
        stabilizes at a value less than 'Reduced Hessian Dimension'.

        'QPSolver QN' solves the QP using a quasi-Newton method.
        In this case, R is the factor of a quasi-Newton approximate
        Hessian.

        'QPSolver CG' uses an active-set method similar to 'QPSolver
        QN', but uses the conjugate-gradient method to solve all systems
        involving the reduced Hessian.

        The Cholesky QP solver is the most robust, but may require a
        significant amount of computation if there are many superbasics.

        The quasi-Newton QP solver does not require computation of the
        exact R at the start of the subproblem in (1).
        It may be appropriate when the number of superbasics is large
        but relatively few iterations are needed to reach a solution
        (e.g., if ``nlp2_solve`` is called with a Warm Start).

        The conjugate-gradient QP solver is appropriate for problems
        with many degrees of freedom (say, more than 2000 superbasics).

    'QPSolver CG' : valueless
        Specifies the active-set algorithm used to solve subproblem (1)
        (see [Treatment of Constraint Infeasibilities]). 'QPSolver
        Cholesky' holds the full Cholesky factor R of the reduced
        Hessian Z^THZ.
        As the QP iterations proceed, the dimension of R changes with
        the number of superbasic variables.
        If the number of superbasic variables needs to increase beyond
        the value of 'Reduced Hessian Dimension', the reduced Hessian
        cannot be stored and the solver switches to 'QPSolver CG'.
        The Cholesky solver is reactivated if the number of superbasics
        stabilizes at a value less than 'Reduced Hessian Dimension'.

        'QPSolver QN' solves the QP using a quasi-Newton method.
        In this case, R is the factor of a quasi-Newton approximate
        Hessian.

        'QPSolver CG' uses an active-set method similar to 'QPSolver
        QN', but uses the conjugate-gradient method to solve all systems
        involving the reduced Hessian.

        The Cholesky QP solver is the most robust, but may require a
        significant amount of computation if there are many superbasics.

        The quasi-Newton QP solver does not require computation of the
        exact R at the start of the subproblem in (1).
        It may be appropriate when the number of superbasics is large
        but relatively few iterations are needed to reach a solution
        (e.g., if ``nlp2_solve`` is called with a Warm Start).

        The conjugate-gradient QP solver is appropriate for problems
        with many degrees of freedom (say, more than 2000 superbasics).

    'QPSolver QN' : valueless
        Specifies the active-set algorithm used to solve subproblem (1)
        (see [Treatment of Constraint Infeasibilities]). 'QPSolver
        Cholesky' holds the full Cholesky factor R of the reduced
        Hessian Z^THZ.
        As the QP iterations proceed, the dimension of R changes with
        the number of superbasic variables.
        If the number of superbasic variables needs to increase beyond
        the value of 'Reduced Hessian Dimension', the reduced Hessian
        cannot be stored and the solver switches to 'QPSolver CG'.
        The Cholesky solver is reactivated if the number of superbasics
        stabilizes at a value less than 'Reduced Hessian Dimension'.

        'QPSolver QN' solves the QP using a quasi-Newton method.
        In this case, R is the factor of a quasi-Newton approximate
        Hessian.

        'QPSolver CG' uses an active-set method similar to 'QPSolver
        QN', but uses the conjugate-gradient method to solve all systems
        involving the reduced Hessian.

        The Cholesky QP solver is the most robust, but may require a
        significant amount of computation if there are many superbasics.

        The quasi-Newton QP solver does not require computation of the
        exact R at the start of the subproblem in (1).
        It may be appropriate when the number of superbasics is large
        but relatively few iterations are needed to reach a solution
        (e.g., if ``nlp2_solve`` is called with a Warm Start).

        The conjugate-gradient QP solver is appropriate for problems
        with many degrees of freedom (say, more than 2000 superbasics).

    'Reduced Hessian Dimension' : int
        Default = min(2000,n)

        This specifies that an i by i triangular matrix R (to define the
        reduced Hessian according to R^TR = Z^THZ) is to be available
        for use by the Cholesky QP solver.

    'Scale Option' : int
        Default = 0

        Three scale options are available as follows:

        [table omitted]

        Option 'Scale Tolerance' affects how many passes might be needed
        through the constraint matrix.
        On each pass, the scaling procedure computes the ratio of the
        largest and smallest nonzero coefficients in each column:

            rho_j = max_j|a_{ij}|/min_i|a_{ij}| (a_{ij} != 0).

        If max_jrho_j is less than r times its previous value, another
        scaling pass is performed to adjust the row and column scales.
        Raising r from 0.9 to 0.99 (say) usually increases the number of
        scaling passes through A.
        At most 10 passes are made.
        The value of r should lie in the range 0 < r < 1.

        'Scale Print' causes the row scales r(i) and column scales c(j)
        to be printed to 'Print File', if 'System Information Yes' has
        been specified.
        The scaled matrix coefficients are a__{ij} = a_{ij}c(j)/r(i),
        and the scaled bounds on the variables and slacks are l__j =
        l_j/c(j), u__j = u_j/c(j), where c(j) = r(j-n) if j > n.

    'Scale Tolerance' : float
        Default = 0.9

        Three scale options are available as follows:

        [table omitted]

        Option 'Scale Tolerance' affects how many passes might be needed
        through the constraint matrix.
        On each pass, the scaling procedure computes the ratio of the
        largest and smallest nonzero coefficients in each column:

            rho_j = max_j|a_{ij}|/min_i|a_{ij}| (a_{ij} != 0).

        If max_jrho_j is less than r times its previous value, another
        scaling pass is performed to adjust the row and column scales.
        Raising r from 0.9 to 0.99 (say) usually increases the number of
        scaling passes through A.
        At most 10 passes are made.
        The value of r should lie in the range 0 < r < 1.

        'Scale Print' causes the row scales r(i) and column scales c(j)
        to be printed to 'Print File', if 'System Information Yes' has
        been specified.
        The scaled matrix coefficients are a__{ij} = a_{ij}c(j)/r(i),
        and the scaled bounds on the variables and slacks are l__j =
        l_j/c(j), u__j = u_j/c(j), where c(j) = r(j-n) if j > n.

    'Scale Print' : valueless
        Three scale options are available as follows:

        [table omitted]

        Option 'Scale Tolerance' affects how many passes might be needed
        through the constraint matrix.
        On each pass, the scaling procedure computes the ratio of the
        largest and smallest nonzero coefficients in each column:

            rho_j = max_j|a_{ij}|/min_i|a_{ij}| (a_{ij} != 0).

        If max_jrho_j is less than r times its previous value, another
        scaling pass is performed to adjust the row and column scales.
        Raising r from 0.9 to 0.99 (say) usually increases the number of
        scaling passes through A.
        At most 10 passes are made.
        The value of r should lie in the range 0 < r < 1.

        'Scale Print' causes the row scales r(i) and column scales c(j)
        to be printed to 'Print File', if 'System Information Yes' has
        been specified.
        The scaled matrix coefficients are a__{ij} = a_{ij}c(j)/r(i),
        and the scaled bounds on the variables and slacks are l__j =
        l_j/c(j), u__j = u_j/c(j), where c(j) = r(j-n) if j > n.

    'Solution File' : int
        Default = 0

        If i > 0, the final solution will be output to file i (whether
        optimal or not).
        All numbers are printed in ``1pe16.6`` format.

        To see more significant digits in the printed solution, it will
        sometimes be useful to make i refer to 'Print File'.

    'Start Objective Check At Variable' : int
        Default = 1

        These keywords take effect only if 'Verify Level' > 0.
        They may be used to contol the verification of gradient elements
        computed by function `objfun` and/or Jacobian elements computed
        by function `confun`.
        For eample, if the first 30 elements of the objective gradient
        appeared to be correct in an earlier run, so that only element
        31 remains questionable, it is reasonable to specify 'Start
        Objective Check At Variable' = 31.
        If the first 30 variables appear linearly in the objective, so
        that the corresponding gradient elements are constant, the above
        choice would also be appropriate.

    'Stop Objective Check At Variable' : int
        Default = n

        These keywords take effect only if 'Verify Level' > 0.
        They may be used to contol the verification of gradient elements
        computed by function `objfun` and/or Jacobian elements computed
        by function `confun`.
        For eample, if the first 30 elements of the objective gradient
        appeared to be correct in an earlier run, so that only element
        31 remains questionable, it is reasonable to specify 'Start
        Objective Check At Variable' = 31.
        If the first 30 variables appear linearly in the objective, so
        that the corresponding gradient elements are constant, the above
        choice would also be appropriate.

    'Start Constraint Check At Variable' : int
        Default = 1

        These keywords take effect only if 'Verify Level' > 0.
        They may be used to contol the verification of gradient elements
        computed by function `objfun` and/or Jacobian elements computed
        by function `confun`.
        For eample, if the first 30 elements of the objective gradient
        appeared to be correct in an earlier run, so that only element
        31 remains questionable, it is reasonable to specify 'Start
        Objective Check At Variable' = 31.
        If the first 30 variables appear linearly in the objective, so
        that the corresponding gradient elements are constant, the above
        choice would also be appropriate.

    'Stop Constraint Check At Variable' : int
        Default = n

        These keywords take effect only if 'Verify Level' > 0.
        They may be used to contol the verification of gradient elements
        computed by function `objfun` and/or Jacobian elements computed
        by function `confun`.
        For eample, if the first 30 elements of the objective gradient
        appeared to be correct in an earlier run, so that only element
        31 remains questionable, it is reasonable to specify 'Start
        Objective Check At Variable' = 31.
        If the first 30 variables appear linearly in the objective, so
        that the corresponding gradient elements are constant, the above
        choice would also be appropriate.

    'Summary File' : int
        Default = 0

        If i_1 > 0, a brief log will be output to the file associated
        with unit i_1, including one line of information every i_2th
        iteration.
        In an interactive environment, it is useful to direct this
        output to the terminal, to allow a run to be monitored online.
        (If something looks wrong, the run can be manually terminated.)
        Further details are given in [The Summary File].

    'Summary Frequency' : int
        Default = 100

        If i_1 > 0, a brief log will be output to the file associated
        with unit i_1, including one line of information every i_2th
        iteration.
        In an interactive environment, it is useful to direct this
        output to the terminal, to allow a run to be monitored online.
        (If something looks wrong, the run can be manually terminated.)
        Further details are given in [The Summary File].

    'Superbasics Limit' : int
        Default = n

        This option places a limit on the storage allocated for
        superbasic variables.
        Ideally, i should be set slightly larger than the 'number of
        degrees of freedom' expected at an optimal solution.

        For nonlinear problems, the number of degrees of freedom is
        often called the 'number of independent variables'.
        Normally, i need not be greater than n_N+1, where n_N is the
        number of nonlinear variables.
        For many problems, i may be considerably smaller than n_N.
        This will save storage if n_N is very large.

    'Suppress Parameters' : valueless
        Normally ``nlp2_solve`` prints the options file as it is being
        read, and then prints a complete list of the available keywords
        and their final values.
        The option 'Suppress Parameters' tells ``nlp2_solve`` not to
        print the full list.

    'System Information No' : valueless
        Default

        This option prints additional information on the progress of
        major and minor iterations, and Crash statistics.
        See [Monitoring Information].

    'System Information Yes' : valueless
        This option prints additional information on the progress of
        major and minor iterations, and Crash statistics.
        See [Monitoring Information].

    'Timing Level' : int
        Default = 0

        If i > 0, some timing information will be output to the Print
        file, if 'Print File' > 0.

    'Unbounded Objective' : float
        Default = 1.0e+15

        These arguments are intended to detect unboundedness in
        nonlinear problems.
        During a linesearch, F is evaluated at points of the form
        x+alpha p, where x and p are fixed and alpha varies.
        If |F| exceeds r_1 or alpha exceeds r_2, iterations are
        terminated with the exit message `errno` = 5.

        If singularities are present, unboundedness in F(x) may be
        manifested by a floating-point overflow (during the evaluation
        of F(x+alpha p)), before the test against r_1 can be made.

        Unboundedness in x is best avoided by placing finite upper and
        lower bounds on the variables.

    'Unbounded Step Size' : float
        Default = bigbnd

        These arguments are intended to detect unboundedness in
        nonlinear problems.
        During a linesearch, F is evaluated at points of the form
        x+alpha p, where x and p are fixed and alpha varies.
        If |F| exceeds r_1 or alpha exceeds r_2, iterations are
        terminated with the exit message `errno` = 5.

        If singularities are present, unboundedness in F(x) may be
        manifested by a floating-point overflow (during the evaluation
        of F(x+alpha p)), before the test against r_1 can be made.

        Unboundedness in x is best avoided by placing finite upper and
        lower bounds on the variables.

    'Verify Level' : int
        Default = 0

        This option refers to finite difference checks on the
        derivatives computed by the user-supplied functions.
        Derivatives are checked at the first point that satisfies all
        bounds and linear constraints.

        [table omitted]

        'Verify Level' = 3 should be specified whenever a new function
        function is being developed.
        Missing derivatives are not checked, so they result in no
        overhead.

    'Violation Limit' : float
        Default = 1.0e+6

        This keyword defines an absolute limit on the magnitude of the
        maximum constraint violation, r, after the linesearch.
        On completion of the linesearch, the new iterate x_{k+1}
        satisfies the condition

            v_i(x_{k+1}) <= r max(1,v_i(x_0)),

        where x_0 is the point at which the nonlinear constraints are
        first evaluated and v_i(x) is the ith nonlinear constraint
        violation v_i(x) = max(0,l_i-c_i(x),c_i(x)-u_i).

        The effect of this violation limit is to restrict the iterates
        to lie in an `expanded` feasible region whose size depends on
        the magnitude of r.
        This makes it possible to keep the iterates within a region
        where the objective is expected to be well defined and bounded
        below.
        If the obective is bounded below for all values of the
        variables, r may be any large positive value.

    Raises
    ------
    NagValueError
        (`errno` 1)
            The initialization function ``nlp2_init`` has not been
            called.

        (`errno` 2)
            On entry, bounds `bl` and `bu` for *<value>* are equal and
            infinite. `bl` = `bu` = *<value>* and bigbnd = *<value>*.

        (`errno` 2)
            On entry, bounds for *<value>* are inconsistent. `bl` =
            *<value>* and `bu` = *<value>*.

        (`errno` 2)
            On entry, bounds `bl` and `bu` for *<value>**<value>* are
            equal and infinite. `bl` = `bu` = *<value>* and bigbnd =
            *<value>*.

        (`errno` 2)
            On entry, bounds for *<value>**<value>* are inconsistent.
            `bl` = *<value>* and `bu` = *<value>*.

        (`errno` 2)
            Basis file dimensions do not match this problem.

        (`errno` 2)
            On entry, n = *<value>*.

            Constraint: n >= *<value>*.

        (`errno` 2)
            On entry, nclin = *<value>*.

            Constraint: nclin >= *<value>*.

        (`errno` 2)
            On entry, ncnln = *<value>*.

            Constraint: ncnln >= *<value>*.

        (`errno` 8)
            User-supplied function computes incorrect objective
            derivatives.

        (`errno` 8)
            User-supplied function computes incorrect constraint
            derivatives.

        (`errno` 11)
            Internal error: memory allocation failed when attempting to
            allocate workspace sizes *<value>* and *<value>*. Please
            contact NAG.

        (`errno` 12)
            Internal memory allocation was insufficient. Please contact
            NAG.

        (`errno` 13)
            An error has occurred in the basis package, perhaps
            indicating incorrect setup of arrays. Set the option 'Print
            File' and examine the output carefully for further
            information.

        (`errno` 14)
            An unexpected error has occurred. Set the option 'Print
            File' and examine the output carefully for further
            information.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 3)
            The requested accuracy could not be achieved.

        (`errno` 4)
            The linear constraints appear to be infeasible.

        (`errno` 4)
            The problem appears to be infeasible. The linear equality
            constraints could not be satisfied.

        (`errno` 4)
            The problem appears to be infeasible. Nonlinear
            infeasibilites have been minimized.

        (`errno` 4)
            The problem appears to be infeasible. Infeasibilites have
            been minimized.

        (`errno` 5)
            The problem appears to be unbounded. The objective function
            is unbounded.

        (`errno` 5)
            The problem appears to be unbounded. The constraint
            violation limit has been reached.

        (`errno` 7)
            Numerical difficulties have been encountered and no further
            progress can be made.

        (`errno` 10)
            User-supplied constraint function requested termination.

        (`errno` 10)
            User-supplied objective function requested termination.

    NagAlgorithmicMajorWarning
        (`errno` 6)
            Iteration limit reached.

        (`errno` 6)
            Major iteration limit reached.

        (`errno` 6)
            The value of the option 'Superbasics Limit' is too small.

        (`errno` 9)
            User-supplied function is undefined at the first feasible
            point.

        (`errno` 9)
            User-supplied function is undefined at the initial point.

        (`errno` 9)
            Unable to proceed into undefined region of user-supplied
            function.

    References
    ----------
    Eldersveld, S K, 1991, `Large-scale sequential quadratic programming
    algorithms`, PhD Thesis, Department of Operations Research, Stanford
    University, Stanford

    Fourer, R, 1982, `Solving staircase linear programs by the simplex
    method`, Math. Programming (23), 274--313

    Gill, P E, Murray, W and Saunders, M A, 2002, `SNOPT: An SQP
    Algorithm for Large-scale Constrained Optimization` (12), 979--1006,
    SIAM J. Optim.

    Gill, P E, Murray, W and Saunders, M A, 2005, `Users' guide for
    SQOPT 7: a Fortran package for large-scale linear and quadratic
    programming`, Report NA 05-1, Department of Mathematics, University
    of California, San Diego,
    https://www.ccom.ucsd.edu/~peg/papers/sqdoc7.pdf

    Gill, P E, Murray, W and Saunders, M A, 2005, `Users' guide for
    SNOPT 7.1: a Fortran package for large-scale linear nonlinear
    programming`, Report NA 05-2, Department of Mathematics, University
    of California, San Diego,
    https://www.ccom.ucsd.edu/~peg/papers/sndoc7.pdf

    Gill, P E, Murray, W, Saunders, M A and Wright, M H, 1986, `Users'
    guide for NPSOL (Version 4.0): a Fortran package for nonlinear
    programming`, Report SOL 86-2, Department of Operations Research,
    Stanford University

    Gill, P E, Murray, W, Saunders, M A and Wright, M H, 1992, `Some
    theoretical properties of an augmented Lagrangian merit function`,
    Advances in Optimization and Parallel Computing, (ed P M Pardalos),
    101--128, North Holland

    Gill, P E, Murray, W and Wright, M H, 1981, `Practical
    Optimization`, Academic Press

    Hock, W and Schittkowski, K, 1981, `Test Examples for Nonlinear
    Programming Codes. Lecture Notes in Economics and Mathematical
    Systems` (187), Springer--Verlag

    See Also
    --------
    :meth:`naginterfaces.library.examples.opt.nlp2_solve_ex.main`
    """

    fun_name = 'naginterfaces.library.opt.nlp2_solve'

    cb_confun_name = fun_name + ':confun'

    def cb_confun(mode, ncnln, n, needc, x, ccon, cjac, nstate, data=None):
        """Wrapping user's confun. Called as base confun."""
        if data is not None:
            _mode_confun, _ccon_confun, _cjac_confun = confun(
                mode,
                needc,
                x,
                cjac,
                nstate,
                data,
            )
        else:
            _mode_confun, _ccon_confun, _cjac_confun = confun(
                mode,
                needc,
                x,
                cjac,
                nstate,
            )
        _mode_confun_base = _mode_confun
        ccon[:] = _utils._to_ndarray(
            {'fun_name': cb_confun_name, 'returned': 'ccon'},
            _ccon_confun,
            copy=False,
            exp_shape=(max(1, ncnln),),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _cjac_confun_np = _utils._to_ndarray(
            {'fun_name': cb_confun_name, 'returned': 'cjac'},
            _cjac_confun,
            copy=False,
            exp_shape=(max(1, ncnln), n),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _utils._check_callback_sorder(
            {'fun_name': cb_confun_name},
            ((_cjac_confun_np, 'cjac'),),
            wrapper_sorder=_m_sorder,
            spiked_sorder=spiked_sorder,
        )
        cjac[:, :] = _cjac_confun_np
        return _mode_confun_base

    cb_objfun_name = fun_name + ':objfun'

    def cb_objfun(mode, n, x, grad, nstate, data=None):
        """Wrapping user's objfun. Called as base objfun."""
        if data is not None:
            _mode_objfun, _objf_objfun, _grad_objfun = objfun(
                mode,
                x,
                grad,
                nstate,
                data,
            )
        else:
            _mode_objfun, _objf_objfun, _grad_objfun = objfun(
                mode,
                x,
                grad,
                nstate,
            )
        _mode_objfun_base = _mode_objfun
        grad[:] = _utils._to_ndarray(
            {'fun_name': cb_objfun_name, 'returned': 'grad'},
            _grad_objfun,
            copy=False,
            exp_shape=(n,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _objf_objfun_base = _objf_objfun
        return (
            _mode_objfun_base,
            _objf_objfun_base,
        )

    _a = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'a'},
        a,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _cjac = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'cjac'},
        cjac,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _h = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'h'},
        h,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((_a, 'a'), (_cjac, 'cjac'), (_h, 'h')),
        spiked_sorder=spiked_sorder,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_h, {'fun_name': fun_name, 'entity': 'h'}, exp_rank=2)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _nclin_py = _b_utils._EngineFloat64ArrayType.get_shape(_a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2)[0]
    _nclin = _b_utils._EngineIntScalarType(
        _nclin_py, {'fun_name': fun_name, 'entity': '_nclin'},
    ).to_ctypes()
    _ncnln_py = _b_utils._EngineFloat64ArrayType.get_shape(_cjac, {'fun_name': fun_name, 'entity': 'cjac'}, exp_rank=2)[0]
    _ncnln = _b_utils._EngineIntScalarType(
        _ncnln_py, {'fun_name': fun_name, 'entity': '_ncnln'},
    ).to_ctypes()
    _bl = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'bl'},
        bl,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _bu = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'bu'},
        bu,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _b_utils._check_callable(confun, {'fun_name': fun_name, 'entity': 'confun'}, allow_none=True)
    _confun = None if confun is None else cb_confun
    _b_utils._check_callable(objfun, {'fun_name': fun_name, 'entity': 'objfun'})
    _objfun = cb_objfun
    _istate = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'istate'},
        istate,
        copy=True,
    )
    _ccon = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'ccon'},
        ccon,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _clamda = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'clamda'},
        clamda,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _grad = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    try:
        _return_scalars = _b_opt.nlp2_solve(
            _n,
            _nclin,
            _ncnln,
            _a,
            _bl,
            _bu,
            _confun,
            _objfun,
            _istate,
            _ccon,
            _cjac,
            _clamda,
            _grad,
            _h,
            _x,
            comm,
            data,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _istate_py = _istate
    _ccon_py = _ccon
    _cjac_py = _cjac
    _clamda_py = _clamda
    _h_py = _h
    _x_py = _x
    _majits_py = _return_scalars.majits
    _objf_py = _return_scalars.objf
    _grad_py = _grad
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['majits', 'istate', 'ccon', 'cjac', 'clamda', 'objf', 'grad', 'h', 'x'],
    )(
        _majits_py,
        _istate_py,
        _ccon_py,
        _cjac_py,
        _clamda_py,
        _objf_py,
        _grad_py,
        _h_py,
        _x_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def nlp2_option_file(ispecs, comm, io_manager=None):
    u"""
    Supply option values for ``nlp2_solve`` from external file.

    ``nlp2_option_file`` may be used to supply options to ``nlp2_solve``
    from an external file.
    The initialization function ``nlp2_init`` **must** have been called
    before calling ``nlp2_option_file``.

    For full information please refer to the NAG Library document for
    e04we

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04wef.html

    Parameters
    ----------
    ispecs : int
        The unit number (see
        :meth:`~naginterfaces.base.utils.FileObjManager.unit_from_fileobj`)
        of the option file to be read.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``nlp2_init``.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Raises
    ------
    NagValueError
        (`errno` 1)
            The initialization function ``nlp2_init`` has not been
            called.

        (`errno` 2)
            At least one line of the options file is invalid.

        (`errno` 2)
            Could not read options file on unit `ispecs` = *<value>*.

    References
    ----------
    Hock, W and Schittkowski, K, 1981, `Test Examples for Nonlinear
    Programming Codes. Lecture Notes in Economics and Mathematical
    Systems` (187), Springer--Verlag
    """

    _b_opt.nlp2_option_file(
        ispecs,
        comm,
        io_manager,
    )

def nlp2_option_string(string, comm, io_manager=None):
    u"""
    Set a single option for ``nlp2_solve`` from a character string.

    ``nlp2_option_string`` may be used to supply individual options to
    ``nlp2_solve``.
    The initialization function ``nlp2_init`` **must** have been called
    before calling ``nlp2_option_string``.

    For full information please refer to the NAG Library document for
    e04wf

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04wff.html

    Parameters
    ----------
    string : str
        A single valid option string (see [Description] in
        ``nlp2_option_string`` and [Other Parameters for nlp2_solve]).

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``nlp2_init``.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Raises
    ------
    NagValueError
        (`errno` 1)
            The initialization function ``nlp2_init`` has not been
            called.

        (`errno` 2)
            The supplied option is invalid. Check that the keywords are
            neither ambiguous nor misspelt. The option string is
            *<value>*.
    """

    _b_opt.nlp2_option_string(
        string,
        comm,
        io_manager,
    )

def nlp2_option_integer_set(string, ivalue, comm, io_manager=None):
    u"""
    Set a single option for ``nlp2_solve`` from an integer argument.

    ``nlp2_option_integer_set`` may be used to supply individual `int`
    options to ``nlp2_solve``.
    The initialization function ``nlp2_init`` **must** have been called
    before calling ``nlp2_option_integer_set``.

    For full information please refer to the NAG Library document for
    e04wg

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04wgf.html

    Parameters
    ----------
    string : str
        A single valid keyword of an `int` option (as described in
        [Other Parameters for nlp2_solve]).

    ivalue : int
        An `int` value associated with the keyword in `string`.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``nlp2_init``.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Raises
    ------
    NagValueError
        (`errno` 1)
            The initialization function ``nlp2_init`` has not been
            called.

        (`errno` 2)
            The supplied option is invalid. Check that the keywords are
            neither ambiguous nor misspelt. The option string is
            *<value>* and `ivalue` = *<value>*.
    """

    _b_opt.nlp2_option_integer_set(
        string,
        ivalue,
        comm,
        io_manager,
    )

def nlp2_option_double_set(string, rvalue, comm, io_manager=None):
    u"""
    Set a single option for ``nlp2_solve`` from a real argument.

    ``nlp2_option_double_set`` may be used to supply individual `float`
    options to ``nlp2_solve``.
    The initialization function ``nlp2_init`` **must** have been called
    before calling ``nlp2_option_double_set``.

    For full information please refer to the NAG Library document for
    e04wh

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04whf.html

    Parameters
    ----------
    string : str
        A single valid keyword of a `float` option (as described in
        [Other Parameters for nlp2_solve]).

    rvalue : float
        The value associated with the keyword in `string`.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``nlp2_init``.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Raises
    ------
    NagValueError
        (`errno` 1)
            The initialization function ``nlp2_init`` has not been
            called.

        (`errno` 2)
            The supplied option is invalid. Check that the keywords are
            neither ambiguous nor misspelt. The option string is
            *<value>* and `rvalue` = *<value>*.
    """

    _b_opt.nlp2_option_double_set(
        string,
        rvalue,
        comm,
        io_manager,
    )

def nlp2_option_integer_get(string, comm, io_manager=None):
    u"""
    Get the setting of an integer valued option of ``nlp2_solve``.

    ``nlp2_option_integer_get`` is used to get the value of an `int`
    option. ``nlp2_option_integer_get`` can be used before or after
    calling ``nlp2_solve``.

    For full information please refer to the NAG Library document for
    e04wk

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04wkf.html

    Parameters
    ----------
    string : str
        A single valid keyword of an `int` option (as described in
        [Other Parameters for nlp2_solve]).

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``nlp2_init``.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    ivalue : int
        The `int` value associated with the keyword in `string`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            The initialization function ``nlp2_init`` has not been
            called.

        (`errno` 2)
            The supplied option string is invalid. Check that the
            keywords are neither ambiguous nor misspelt. The option
            string is *<value>*.
    """

    _return_scalar = _b_opt.nlp2_option_integer_get(
        string,
        comm,
        io_manager,
    )
    _ivalue_py = _return_scalar
    return _ivalue_py

def nlp2_option_double_get(string, comm, io_manager=None):
    u"""
    Get the setting of a real valued option of ``nlp2_solve``.

    ``nlp2_option_double_get`` is used to get the value of a `float`
    option. ``nlp2_option_double_get`` can be used before or after
    calling ``nlp2_solve``.

    For full information please refer to the NAG Library document for
    e04wl

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04wlf.html

    Parameters
    ----------
    string : str
        A single valid keyword of a `float` option (as described in
        [Other Parameters for nlp2_solve]).

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``nlp2_init``.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    rvalue : float
        The `float` value associated with the keyword in `string`.

    Raises
    ------
    NagValueError
        (`errno` 1)
            The initialization function ``nlp2_init`` has not been
            called.

        (`errno` 2)
            The supplied option string is invalid. Check that the
            keywords are neither ambiguous nor misspelt. The option
            string is *<value>*.
    """

    _return_scalar = _b_opt.nlp2_option_double_get(
        string,
        comm,
        io_manager,
    )
    _rvalue_py = _return_scalar
    return _rvalue_py

def estimate_deriv(msglvl, epsrf, x, mode, objfun, hforw, data=None, io_manager=None):
    u"""
    Estimate (using numerical differentiation) gradient and/or Hessian
    of a function.

    ``estimate_deriv`` computes an approximation to the gradient vector
    and/or the Hessian matrix for use in conjunction with, or following
    the use of an optimization function (such as ``nlp1_rcomm``).

    For full information please refer to the NAG Library document for
    e04xa

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04xaf.html

    Parameters
    ----------
    msglvl : int
        Must indicate the amount of intermediate output desired (see
        [Further Comments] for a description of the printed output). All
        output is written on the file object associated with the
        advisory I/O unit (see
        :class:`~naginterfaces.base.utils.FileObjManager`).

        [table omitted]

    epsrf : float
        Must define e_R, which is intended to be a measure of the
        accuracy with which the problem function F can be computed. The
        value of e_R should reflect the relative precision of 1+|F(x)|,
        i.e., acts as a relative precision when |F| is large, and as an
        absolute precision when |F| is small. For example, if F(x) is
        typically of order 1000 and the first six significant digits are
        known to be correct, an appropriate value for e_R would be
        1.0e-6.

        A discussion of `epsrf` is given in Module 8 of Gill `et al.`
        (1981).

        If `epsrf` is either too small or too large on entry a warning
        will be printed if `msglvl` = 1, the argument `iwarn` set to the
        appropriate value on exit and ``estimate_deriv`` will use a
        default value of e_M^0.9, where e_M is the machine precision.

        If `epsrf` <= 0.0 on entry, then ``estimate_deriv`` will use the
        default value internally.

        The default value will be appropriate for most simple functions
        that are computed with full accuracy.

    x : float, array-like, shape (n)
        The point x at which the derivatives are to be computed.

    mode : int
        Indicates which derivatives are required.

        `mode` = 0

            The gradient and Hessian diagonal values having supplied the
            objective function via `objfun`.

        `mode` = 1

            The Hessian matrix having supplied both the objective
            function and gradients via `objfun`.

        `mode` = 2

            The gradient values and Hessian matrix having supplied the
            objective function via `objfun`.

    objfun : callable (objf, objgrd) = objfun(mode, x, nstate,
        data=None)
        If `mode` = 0 or 2, `objfun` must calculate the objective
        function; otherwise if `mode` = 1, `objfun` must calculate the
        objective function and the gradients.

        Parameters
        ~~~~~~~~~~
        mode : int
            `mode` indicates which argument values within `objfun` need
            to be set.

            To `objfun`, `mode` is always set to the value that you set
            it to before the call to ``estimate_deriv``.

        x : float, ndarray, shape (n)
            The point x at which the objective function (and gradients
            if `mode` = 1) is to be evaluated.

        nstate : int
            Will be set to 1 on the first call of `objfun` by
            ``estimate_deriv``, and is 0 for all subsequent calls. Thus,
            if you wish, `nstate` may be tested within `objfun` in order
            to perform certain calculations once only. For example you
            may read data.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        objf : float
            Must be set to the value of the objective function.

        objgrd : float, array-like, shape (n)
            If `mode` = 1, `objgrd`[j-1] must contain the value of the
            first derivative with respect to x.

            If `mode` != 1, `objgrd` need not be set.

    hforw : float, array-like, shape (n)
        The initial trial interval for computing the appropriate partial
        derivative to the jth variable.

        If `hforw`[j-1] <= 0.0, the initial trial interval is computed
        by ``estimate_deriv`` (see [Description]).

    data : arbitrary, optional
        User-communication data for callback functions.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Returns
    -------
    mode : int
        Is changed **only** if you set `mode` negative in `objfun`,
        i.e., you have requested termination of ``estimate_deriv``.

    hforw : float, ndarray, shape (n)
        `hforw`[j-1] is the best interval found for computing a
        forward-difference approximation to the appropriate partial
        derivative for the jth variable.

    objf : float
        The value of the objective function evaluated at the input
        vector in `x`.

    objgrd : float, ndarray, shape (n)
        If `mode` = 0 or 2, `objgrd`[j-1] contains the best estimate of
        the first partial derivative for the jth variable.

        If `mode` = 1, `objgrd`[j-1] contains the first partial
        derivative for the jth variable evaluated at the input vector in
        `x`.

    hcntrl : float, ndarray, shape (n)
        `hcntrl`[j-1] is the best interval found for computing a
        central-difference approximation to the appropriate partial
        derivative for the jth variable.

    h : float, ndarray, shape (n, :)
        If `mode` = 0, the estimated Hessian diagonal elements are
        contained in the first column of this array.

        If `mode` = 1 or 2, the estimated Hessian matrix is contained in
        the leading n by n part of this array.

    iwarn : int
        `iwarn` = 0 on successful exit.

        If the value of `epsrf` on entry is too small or too large then
        `iwarn` is set to 1 or 2 respectively on exit and the default
        value for `epsrf` is used within ``estimate_deriv``.

        If `msglvl` > 0 then warnings will be printed if `epsrf` is too
        small or too large.

    info : int, ndarray, shape (n)
        `info`[j-1] represents diagnostic information on variable j as
        follows:

        `info`[j-1] = 1

            The appropriate function appears to be constant.
            `hforw`[i-1] is set to the initial trial interval value (see
            [Description]) corresponding to a well-scaled problem and
            Error est. in the printed output is set to zero. This value
            occurs when the estimated relative condition error in the
            first derivative approximation is unacceptably large for
            every value of the finite difference interval. If this
            happens when the function is not constant the initial
            interval may be too small; in this case, it may be
            worthwhile to rerun ``estimate_deriv`` with larger initial
            trial interval values supplied in `hforw` (see
            [Description]). This error may also occur if the function
            evaluation includes an inordinately large constant term or
            if `epsrf` is too large.

        `info`[j-1] = 2

            The appropriate function appears to be linear or odd.
            `hforw`[i-1] is set to the smallest interval with acceptable
            bounds on the relative condition error in the forward - and
            backward-difference estimates. In this case, the estimated
            relative condition error in the second derivative
            approximation remained large for every trial interval, but
            the estimated error in the first derivative approximation
            was acceptable for at least one interval. If the function is
            not linear or odd the relative condition error in the second
            derivative may be decreasing very slowly, it may be
            worthwhile to rerun ``estimate_deriv`` with larger initial
            trial interval values supplied in `hforw` (see
            [Description]).

        `info`[j-1] = 3

            The second derivative of the appropriate function appears to
            be so large that it cannot be reliably estimated (i.e., near
            a singularity). `hforw`[i-1] is set to the smallest trial
            interval.

            This value occurs when the relative condition error estimate
            in the second derivative remained very small for every trial
            interval.

            If the second derivative is not large the relative condition
            error in the second derivative may be increasing very
            slowly.

            It may be worthwhile to rerun ``estimate_deriv`` with
            smaller initial trial interval values supplied in `hforw`
            (see [Description]).

            This error may also occur when the given value of `epsrf` is
            not a good estimate of a bound on the absolute error in the
            appropriate function (i.e., `epsrf` is too small).

        `info`[j-1] = 4

            The algorithm terminated with an apparently acceptable
            estimate of the second derivative. However the
            forward-difference estimates of the appropriate first
            derivatives (computed with the final estimate of the
            'optimal' forward-difference interval) and the central
            difference estimates (computed with the interval used to
            compute the final estimate of the second derivative) do not
            agree to half a decimal place. The usual reason that the
            forward - and central-difference estimates fail to agree is
            that the first derivative is small.

            If the first derivative is not small, it may be helpful to
            execute the procedure at a different point.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `mode` = *<value>*.

            Constraint: 0 <= `mode` <= 2.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 1.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            One or more variables have a nonzero `info` value.

    NagCallbackTerminateWarning
        (`errno` i < 0)
            User requested termination by setting `mode` negative in
            `objfun`.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Gill, P E, Murray, W, Saunders, M A and Wright, M H, 1983,
    `Documentation for FDCALC and FDCORE`, Technical Report SOL, 83--6,
    Stanford University

    Gill, P E, Murray, W, Saunders, M A and Wright, M H, 1983,
    `Computing forward-difference intervals for numerical optimization`,
    SIAM J. Sci. Statist. Comput. (4), 310--321

    Gill, P E, Murray, W and Wright, M H, 1981, `Practical
    Optimization`, Academic Press
    """

    fun_name = 'naginterfaces.library.opt.estimate_deriv'

    cb_objfun_name = fun_name + ':objfun'

    def cb_objfun(mode, n, x, objgrd, nstate, data=None):
        """Wrapping user's objfun. Called as base objfun."""
        if data is not None:
            _objf_objfun, _objgrd_objfun = objfun(
                mode,
                x,
                nstate,
                data,
            )
        else:
            _objf_objfun, _objgrd_objfun = objfun(
                mode,
                x,
                nstate,
            )
        objgrd[:] = _utils._to_ndarray(
            {'fun_name': cb_objfun_name, 'returned': 'objgrd'},
            _objgrd_objfun,
            copy=False,
            exp_shape=(n,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _objf_objfun_base = _objf_objfun
        return _objf_objfun_base

    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b_utils._check_callable(objfun, {'fun_name': fun_name, 'entity': 'objfun'})
    _objfun = cb_objfun
    _hforw = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'hforw'},
        hforw,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _objgrd = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _hcntrl = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    _mode = _b_utils._EngineIntScalarType(
        mode, {'fun_name': fun_name, 'entity': 'mode'},
    ).to_ctypes()
    _mode_py = _b_utils._EngineIntScalarType.to_py(_mode)
    if _mode_py == 0:
        _h_dim2 = 1
    elif _mode_py in (1, 2):
        _h_dim2 = _n_py
    else:
        _h_dim2 = 0
    exp_shape_h = (_n_py, _h_dim2)
    _h = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        exp_shape_h,
        sorder=_b_utils._EngineData.default_sorder,
    )
    _info = _b_utils._EngineIntArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _return_scalars = _b_opt.estimate_deriv(
            msglvl,
            _n,
            epsrf,
            _x,
            _mode,
            _objfun,
            _hforw,
            _objgrd,
            _hcntrl,
            _h,
            _info,
            data,
            io_manager,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    _mode_py = _return_scalars.mode
    _hforw_py = _hforw
    _objf_py = _return_scalars.objf
    _objgrd_py = _objgrd
    _hcntrl_py = _hcntrl
    if _mode_py == 0:
        _h_dim2 = 1
    elif _mode_py in (1, 2):
        _h_dim2 = _n_py
    else:
        _h_dim2 = 0
    _h_py = _h[:, :_h_dim2]
    _iwarn_py = _return_scalars.iwarn
    _info_py = _info
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['mode', 'hforw', 'objf', 'objgrd', 'hcntrl', 'h', 'iwarn', 'info'],
    )(
        _mode_py,
        _hforw_py,
        _objf_py,
        _objgrd_py,
        _hcntrl_py,
        _h_py,
        _iwarn_py,
        _info_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def lsq_check_deriv(m, lsqfun, x, data=None, spiked_sorder='C'):
    u"""
    Check user's function for calculating Jacobian of first derivatives.

    ``lsq_check_deriv`` checks that a user-supplied function for
    evaluating a vector of functions and the matrix of their first
    derivatives produces derivative values which are consistent with the
    function values calculated.

    For full information please refer to the NAG Library document for
    e04ya

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04yaf.html

    Parameters
    ----------
    m : int
        The number m of residuals, f_i(x), and the number n of
        variables, x_j.

    lsqfun : callable (iflag, fvec, fjac) = lsqfun(iflag, m, xc,
        data=None)
        `lsqfun` must calculate the vector of values f_i(x) and their
        first derivatives df_i/dx_j at any point x. (The minimization
        functions mentioned in [Description] give you the option of
        resetting an argument to terminate immediately.
        ``lsq_check_deriv`` will also terminate immediately, without
        finishing the checking process, if the argument in question is
        reset.)

        Parameters
        ~~~~~~~~~~
        iflag : int
            To `lsqfun`, `iflag` will be set to 2.

        m : int
            The numbers m of residuals.

        xc : float, ndarray, shape (n)
            x, the point at which the values of the f_i and the
            df_i/dx_j are required.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        iflag : int
            If you reset `iflag` to some negative number in `lsqfun` and
            return control to ``lsq_check_deriv``, the function will
            terminate immediately with errno set to your setting of
            `iflag`.

        fvec : float, array-like, shape (`m`)
            Unless `iflag` is reset to a negative number, `fvec`[i-1]
            must contain the value of f_i at the point x, for i =
            1,2,...,m.

        fjac : float, array-like, shape (`m`, n)
            Unless `iflag` is reset to a negative number,
            `fjac`[i-1,j-1] must contain the value of df_i/dx_j at the
            point x, for j = 1,2,...,n, for i = 1,2,...,m.

    x : float, array-like, shape (n)
        `x`[j-1], for j = 1,2,...,n, must be set to the coordinates of a
        suitable point at which to check the derivatives calculated by
        `lsqfun`. 'Obvious' settings, such as 0 or 1, should not be used
        since, at such particular points, incorrect terms may take
        correct values (particularly zero), so that errors can go
        undetected. For a similar reason, it is preferable that no two
        elements of `x` should have the same value.

    data : arbitrary, optional
        User-communication data for callback functions.

    spiked_sorder : str, optional
        If `fjac` in `lsqfun` is spiked (i.e., has unit extent in all
        but one dimension, or has size 1), `spiked_sorder` selects the
        storage order to associate with it in the NAG Engine:

        spiked_sorder = 'C'
          row-major storage will be used;

        spiked_sorder = 'F'
          column-major storage will be used.

    Returns
    -------
    fvec : float, ndarray, shape (`m`)
        Unless you set `iflag` negative in the first call of `lsqfun`,
        `fvec`[i-1] contains the value of f_i at the point supplied by
        you in `x`, for i = 1,2,...,m.

    fjac : float, ndarray, shape (`m`, n)
        Unless you set `iflag` negative in the first call of `lsqfun`,
        `fjac`[i-1,j-1] contains the value of the first derivative
        df_i/dx_j at the point given in `x`, as calculated by `lsqfun`,
        for j = 1,2,...,n, for i = 1,2,...,m.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 1)
            On entry, `m` = *<value>* and n = *<value>*.

            Constraint: `m` >= n.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i < 0)
            User requested termination by setting `iflag` negative in
            `lsqfun`.

        (`errno` 2)
            It is very likely that you have made an error in forming the
            derivatives in `lsqfun`.

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.
    """

    fun_name = 'naginterfaces.library.opt.lsq_check_deriv'

    cb_lsqfun_name = fun_name + ':lsqfun'

    def cb_lsqfun(iflag, m, n, xc, fvec, fjac, data=None):
        """Wrapping user's lsqfun. Called as base lsqfun."""
        if data is not None:
            _iflag_lsqfun, _fvec_lsqfun, _fjac_lsqfun = lsqfun(
                iflag,
                m,
                xc,
                data,
            )
        else:
            _iflag_lsqfun, _fvec_lsqfun, _fjac_lsqfun = lsqfun(
                iflag,
                m,
                xc,
            )
        _iflag_lsqfun_base = _iflag_lsqfun
        fvec[:] = _utils._to_ndarray(
            {'fun_name': cb_lsqfun_name, 'returned': 'fvec'},
            _fvec_lsqfun,
            copy=False,
            exp_shape=(m,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _fjac_lsqfun_np = _utils._to_ndarray(
            {'fun_name': cb_lsqfun_name, 'returned': 'fjac'},
            _fjac_lsqfun,
            copy=False,
            exp_shape=(m, n),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _utils._check_callback_sorder(
            {'fun_name': cb_lsqfun_name},
            ((_fjac_lsqfun_np, 'fjac'),),
            wrapper_sorder=_m_sorder,
            spiked_sorder=spiked_sorder,
        )
        fjac[:, :] = _fjac_lsqfun_np
        return _iflag_lsqfun_base

    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        (),
        spiked_sorder=spiked_sorder,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b_utils._check_callable(lsqfun, {'fun_name': fun_name, 'entity': 'lsqfun'})
    _lsqfun = cb_lsqfun
    _m = _b_utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = _b_utils._EngineIntScalarType.to_py(_m)
    _fvec = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_m_py,),
    )
    _fjac = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_m_py, _n_py),
        sorder=_b_utils._EngineData.default_sorder,
    )
    try:
        _b_opt.lsq_check_deriv(
            _m,
            _n,
            _lsqfun,
            _x,
            _fvec,
            _fjac,
            data,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _fvec_py = _fvec
    _fjac_py = _fjac
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['fvec', 'fjac'],
    )(
        _fvec_py,
        _fjac_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def lsq_check_hessian(m, lsqfun, lsqhes, x, lb, data=None, spiked_sorder='C'):
    u"""
    Check user's function for calculating Hessian of a sum of squares.

    ``lsq_check_hessian`` checks that a user-supplied function for
    evaluating the second derivative term of the Hessian matrix of a sum
    of squares is consistent with a user-supplied function for
    calculating the corresponding first derivatives.

    For full information please refer to the NAG Library document for
    e04yb

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04ybf.html

    Parameters
    ----------
    m : int
        The number m of residuals, f_i(x), and the number n of
        variables, x_j.

    lsqfun : callable (iflag, fvec, fjac) = lsqfun(iflag, m, xc,
        data=None)
        `lsqfun` must calculate the vector of values f_i(x) and their
        first derivatives df_i/dx_j at any point x.
        (``lsq_uncon_mod_deriv2_comp`` gives you the option of resetting
        arguments of `lsqfun` to cause the minimization process to
        terminate immediately. ``lsq_check_hessian`` will also terminate
        immediately, without finishing the checking process, if the
        argument in question is reset.)

        Parameters
        ~~~~~~~~~~
        iflag : int
            To `lsqfun`, `iflag` will be set to 2.

        m : int
            The numbers m of residuals.

        xc : float, ndarray, shape (n)
            The point x at which the values of the f_i and the df_i/dx_j
            are required.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        iflag : int
            If you reset `iflag` to some negative number in `lsqfun` and
            return control to ``lsq_check_hessian``, the function will
            terminate immediately with errno set to your setting of
            `iflag`.

        fvec : float, array-like, shape (`m`)
            Unless `iflag` is reset to a negative number, `fvec`[i-1]
            must contain the value of f_i at the point x, for i =
            1,2,...,m.

        fjac : float, array-like, shape (`m`, n)
            Unless `iflag` is reset to a negative number,
            `fjac`[i-1,j-1] must contain the value of df_i/dx_j at the
            point x, for j = 1,2,...,n, for i = 1,2,...,m.

    lsqhes : callable (iflag, b) = lsqhes(iflag, fvec, xc, lb,
        data=None)
        `lsqhes` must calculate the elements of the symmetric matrix

            B(x) = sum_i = 1^mf_i(x)G_i(x),

        at any point x, where G_i(x) is the Hessian matrix of f_i(x).
        (As with `lsqfun`, an argument can be set to cause immediate
        termination.)

        Parameters
        ~~~~~~~~~~
        iflag : int
            Is set to a non-negative number.

        fvec : float, ndarray, shape (m)
            The value of the residual f_i at the point x, for i =
            1,2,...,m, so that the values of the f_i can be used in the
            calculation of the elements of `b`.

        xc : float, ndarray, shape (n)
            The point x at which the elements of `b` are to be
            evaluated.

        lb : int
            Gives the length of the array `b`.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        iflag : int
            If `lsqhes` resets `iflag` to some negative number,
            ``lsq_check_hessian`` will terminate immediately, with errno
            set to your setting of `iflag`.

        b : float, array-like, shape (`lb`)
            Unless `iflag` is reset to a negative number `b` must
            contain the lower triangle of the matrix B(x), evaluated at
            the point in `xc`, stored by rows. (The upper triangle is
            not needed because the matrix is symmetric.) More precisely,
            `b`[j(j-1)/2+k-1] must contain sum_i = 1^mf_i d^2f_i/(dx_j
            dx_k) evaluated at the point x, for k = 1,2,...,j, for j =
            1,2,...,n.

    x : float, array-like, shape (n)
        `x`[j-1], for j = 1,2,...,n, must be set to the coordinates of a
        suitable point at which to check the b_{jk} calculated by
        `lsqhes`. 'Obvious' settings, such as 0 or 1, should not be used
        since, at such particular points, incorrect terms may take
        correct values (particularly zero), so that errors could go
        undetected. For a similar reason, it is preferable that no two
        elements of `x` should have the same value.

    lb : int
        The dimension of the array `b`.

    data : arbitrary, optional
        User-communication data for callback functions.

    spiked_sorder : str, optional
        If `fjac` in `lsqfun` is spiked (i.e., has unit extent in all
        but one dimension, or has size 1), `spiked_sorder` selects the
        storage order to associate with it in the NAG Engine:

        spiked_sorder = 'C'
          row-major storage will be used;

        spiked_sorder = 'F'
          column-major storage will be used.

    Returns
    -------
    fvec : float, ndarray, shape (`m`)
        Unless you set `iflag` negative in the first call of `lsqfun`,
        `fvec`[i-1] contains the value of f_i at the point supplied by
        you in `x`, for i = 1,2,...,m.

    fjac : float, ndarray, shape (`m`, n)
        Unless you set `iflag` negative in the first call of `lsqfun`,
        `fjac`[i-1,j-1] contains the value of the first derivative
        df_i/dx_j at the point given in `x`, as calculated by `lsqfun`,
        for j = 1,2,...,n, for i = 1,2,...,m.

    b : float, ndarray, shape (`lb`)
        Unless you set `iflag` negative in `lsqhes`, `b`[j*(j-1)/2+k-1]
        contains the value of b_{jk} at the point given in `x` as
        calculated by `lsqhes`, for k = 1,2,...,j, for j = 1,2,...,n.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m` = *<value>* and n = *<value>*.

            Constraint: `m` >= n.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 1)
            On entry, `lb` = *<value>* and (n+1)*n/2 = *<value>*.

            Constraint: `lb` >= (n+1)*n/2.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` i < 0)
            User requested termination by setting `iflag` negative in
            `lsqfun` or `lsqhes`.

        (`errno` 2)
            It is very likely that you have made an error in setting up
            `b` in `lsqhes`.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.
    """

    fun_name = 'naginterfaces.library.opt.lsq_check_hessian'

    cb_lsqfun_name = fun_name + ':lsqfun'

    def cb_lsqfun(iflag, m, n, xc, fvec, fjac, data=None):
        """Wrapping user's lsqfun. Called as base lsqfun."""
        if data is not None:
            _iflag_lsqfun, _fvec_lsqfun, _fjac_lsqfun = lsqfun(
                iflag,
                m,
                xc,
                data,
            )
        else:
            _iflag_lsqfun, _fvec_lsqfun, _fjac_lsqfun = lsqfun(
                iflag,
                m,
                xc,
            )
        _iflag_lsqfun_base = _iflag_lsqfun
        fvec[:] = _utils._to_ndarray(
            {'fun_name': cb_lsqfun_name, 'returned': 'fvec'},
            _fvec_lsqfun,
            copy=False,
            exp_shape=(m,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _fjac_lsqfun_np = _utils._to_ndarray(
            {'fun_name': cb_lsqfun_name, 'returned': 'fjac'},
            _fjac_lsqfun,
            copy=False,
            exp_shape=(m, n),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        _utils._check_callback_sorder(
            {'fun_name': cb_lsqfun_name},
            ((_fjac_lsqfun_np, 'fjac'),),
            wrapper_sorder=_m_sorder,
            spiked_sorder=spiked_sorder,
        )
        fjac[:, :] = _fjac_lsqfun_np
        return _iflag_lsqfun_base

    cb_lsqhes_name = fun_name + ':lsqhes'

    def cb_lsqhes(iflag, m, n, fvec, xc, b, lb, data=None):
        """Wrapping user's lsqhes. Called as base lsqhes."""
        if data is not None:
            _iflag_lsqhes, _b_lsqhes = lsqhes(
                iflag,
                fvec,
                xc,
                lb,
                data,
            )
        else:
            _iflag_lsqhes, _b_lsqhes = lsqhes(
                iflag,
                fvec,
                xc,
                lb,
            )
        _iflag_lsqhes_base = _iflag_lsqhes
        b[:] = _utils._to_ndarray(
            {'fun_name': cb_lsqhes_name, 'returned': 'b'},
            _b_lsqhes,
            copy=False,
            exp_shape=(lb,),
            exp_class=_b_utils._EngineFloat64ArrayType,
        )
        return _iflag_lsqhes_base

    _m_sorder = _b_utils._get_mutual_sorder(
        {'fun_name': fun_name},
        (),
        spiked_sorder=spiked_sorder,
    )
    _x = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'x'},
        x,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _b_utils._check_callable(lsqfun, {'fun_name': fun_name, 'entity': 'lsqfun'})
    _lsqfun = cb_lsqfun
    _b_utils._check_callable(lsqhes, {'fun_name': fun_name, 'entity': 'lsqhes'})
    _lsqhes = cb_lsqhes
    _m = _b_utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = _b_utils._EngineIntScalarType.to_py(_m)
    _fvec = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_m_py,),
    )
    _fjac = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_m_py, _n_py),
        sorder=_b_utils._EngineData.default_sorder,
    )
    _lb = _b_utils._EngineIntScalarType(
        lb, {'fun_name': fun_name, 'entity': 'lb'},
    ).to_ctypes()
    _lb_py = _b_utils._EngineIntScalarType.to_py(_lb)
    _b = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_lb_py,),
    )
    try:
        _b_opt.lsq_check_hessian(
            _m,
            _n,
            _lsqfun,
            _lsqhes,
            _x,
            _fvec,
            _fjac,
            _b,
            _lb,
            data,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _fvec_py = _fvec
    _fjac_py = _fjac
    _b_py = _b
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['fvec', 'fjac', 'b'],
    )(
        _fvec_py,
        _fjac_py,
        _b_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def lsq_uncon_covariance(job, m, fsumsq, s, v):
    u"""
    Covariance matrix for nonlinear least squares problem
    (unconstrained).

    ``lsq_uncon_covariance`` returns estimates of elements of the
    variance-covariance matrix of the estimated regression coefficients
    for a nonlinear least squares problem.
    The estimates are derived from the Jacobian of the function f(x) at
    the solution.

    This function may be used following any one of the nonlinear least
    squares functions ``lsq_uncon_mod_func_comp``,
    ``lsq_uncon_quasi_deriv_comp``, ``lsq_uncon_mod_deriv_comp`` or
    ``lsq_uncon_mod_deriv2_comp``.

    For full information please refer to the NAG Library document for
    e04yc

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04ycf.html

    Parameters
    ----------
    job : int
        Which elements of C are returned as follows:

        `job` = -1

            The n by n symmetric matrix C is returned.

        `job` = 0

            The diagonal elements of C are returned.

        `job` > 0

            The elements of column `job` of C are returned.

    m : int
        The number m of observations (residuals f_i(x)).

    fsumsq : float
        The sum of squares of the residuals, F(x_), at the solution x_,
        as returned by the nonlinear least squares function.

    s : float, array-like, shape (n)
        The n singular values of the Jacobian as returned by the
        nonlinear least squares function. See [Description] for
        information on supplying `s` following one of the easy-to-use
        functions.

    v : float, array-like, shape (n, n)
        The n by n right-hand orthogonal matrix (the right singular
        vectors) of J as returned by the nonlinear least squares
        function. See [Description] for information on supplying `v`
        following one of the easy-to-use functions.

    Returns
    -------
    v : float, ndarray, shape (n, n)
        If `job` >= 0, `v` is unchanged.

        If `job` = -1, the leading n by n part of `v` is overwritten by
        the n by n matrix C.

        When ``lsq_uncon_covariance`` is called with `job` = -1
        following an easy-to-use function this means that C is returned,
        column by column, in the n^2 elements of w given by
        w[NV-1],w[NV],...,w[NV+n^2-2]. (See [Description] for the
        definition of NV.)

    cj : float, ndarray, shape (n)
        If `job` = 0, `cj` returns the n diagonal elements of C.

        If `job` = j > 0, `cj` returns the n elements of the jth column
        of C.

        If `job` = -1, `cj` is not referenced.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `job` = *<value>*.

            Constraint: `job` >= -1.

        (`errno` 1)
            On entry, n = *<value>*.

            Constraint: n >= 1.

        (`errno` 1)
            On entry, `job` = *<value>* and n = *<value>*.

            Constraint: `job` <= n.

        (`errno` 1)
            On entry, `m` = *<value>* and n = *<value>*.

            Constraint: `m` >= n.

        (`errno` 1)
            On entry, `fsumsq` = *<value>*.

            Constraint: `fsumsq` >= 0.0.

        (`errno` 1)
            On entry, `job` = -1, ldv = *<value>* and n = *<value>*.

            Constaint: if `job` = -1, ldv >= n.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 2)
            The singular values are all zero, so that at the solution
            the Jacobian matrix has rank 0.

        (`errno` i > 2)
            At the solution the Jacobian matrix contains linear, or near
            linear, dependencies amongst its columns. The required
            elements of C have still been computed based upon J having
            an assumed rank errno-2 = *<value>*. The rank is computed by
            regarding as zero singular values SV(j) that are not larger
            than 10*epsilon*SV(1), where epsilon is the machine
            precision (see ``machine.precision``).

    Notes
    -----
    In the NAG Library the traditional C interface for this routine
    uses a different algorithmic base.
    Please contact NAG if you have any questions about compatibility.

    References
    ----------
    Bard, Y, 1974, `Nonlinear Parameter Estimation`, Academic Press

    Wolberg, J R, 1967, `Prediction Analysis`, Van Nostrand
    """

    fun_name = 'naginterfaces.library.opt.lsq_uncon_covariance'

    _s = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 's'},
        s,
        copy=False,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _n_py = _b_utils._EngineFloat64ArrayType.get_shape(_s, {'fun_name': fun_name, 'entity': 's'}, exp_rank=1)[0]
    _n = _b_utils._EngineIntScalarType(
        _n_py, {'fun_name': fun_name, 'entity': '_n'},
    ).to_ctypes()
    _v = _utils._to_ndarray(
        {'fun_name': fun_name, 'entity': 'v'},
        v,
        copy=True,
        exp_class=_b_utils._EngineFloat64ArrayType,
    )
    _cj = _b_utils._EngineFloat64ArrayType.empty_ndarray(
        (_n_py,),
    )
    try:
        _b_opt.lsq_uncon_covariance(
            job,
            m,
            _n,
            fsumsq,
            _s,
            _v,
            _cj,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
    _v_py = _v
    _cj_py = _cj
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['v', 'cj'],
    )(
        _v_py,
        _cj_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _return_data

def handle_opt_set(handle, optstr):
    u"""
    Option setting routine for the solvers from the NAG optimization
    modelling suite.

    ``handle_opt_set`` is an option setting function for all solvers
    from the NAG optimization modelling suite.
    It can set a single option or reset all of them to their default.

    For full information please refer to the NAG Library document for
    e04zm

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04zmf.html

    Parameters
    ----------
    handle : Handle
        The handle to the problem. It needs to be initialized by
        ``handle_init`` and **must not** be changed before the call to
        ``handle_opt_set``.

    optstr : str
        A string identifying the option and its value to be set.

        Defaults

            Resets all options to their default values.

        Option = optval

            See the documentation of the particular solver for details
            of valid values for option and optval. The equals sign (=)
            delimiter must be used to separate the option from its
            optval value.

        Option = Default

            Resets the given option back to its default value.

        `optstr` is case insensitive.

        Each token in the option and optval component must be separated
        by at least one space.

    Raises
    ------
    NagValueError
        (`errno` 1)
            `handle` has not been initialized.

        (`errno` 1)
            `handle` does not belong to the NAG optimization modelling
            suite, has not been initialized properly or is corrupted.

        (`errno` 1)
            `handle` has not been initialized properly or is corrupted.

        (`errno` 2)
            The options cannot be modified in this phase.

        (`errno` 11)
            On entry, the option supplied in `optstr` was not
            recognized: `optstr` = *<value>*.

        (`errno` 12)
            On entry, the expected delimiter '=' was not found in
            `optstr`: `optstr` = *<value>*.

        (`errno` 13)
            On entry, could not convert the specified optval to an
            integer: optval = *<value>*.

        (`errno` 13)
            On entry, could not convert the specified optval to a real:
            optval = *<value>*.

        (`errno` 15)
            On entry, the optval supplied for the integer option is not
            valid.

            option = *<value>*, optval = *<value>*.

            Constraint: optval >= *<value>*.

        (`errno` 15)
            On entry, the optval supplied for the integer option is not
            valid.

            option = *<value>*, optval = *<value>*.

            Constraint: optval > *<value>*.

        (`errno` 15)
            On entry, the optval supplied for the integer option is not
            valid.

            option = *<value>*, optval = *<value>*.

            Constraint: optval <= *<value>*.

        (`errno` 15)
            On entry, the optval supplied for the integer option is not
            valid.

            option = *<value>*, optval = *<value>*.

            Constraint: optval < *<value>*.

        (`errno` 16)
            On entry, the optval supplied for the real option is not
            valid.

            option = *<value>*, optval = *<value>*.

            Constraint: optval >= *<value>*.

        (`errno` 16)
            On entry, the optval supplied for the real option is not
            valid.

            option = *<value>*, optval = *<value>*.

            Constraint: optval > *<value>*.

        (`errno` 16)
            On entry, the optval supplied for the real option is not
            valid.

            option = *<value>*, optval = *<value>*.

            Constraint: optval <= *<value>*.

        (`errno` 16)
            On entry, the optval supplied for the real option is not
            valid.

            option = *<value>*, optval = *<value>*.

            Constraint: optval < *<value>*.

        (`errno` 17)
            On entry, the optval supplied for the character option is
            not valid.

            option = *<value>*, optval = *<value>*.

    See Also
    --------
    :meth:`naginterfaces.library.examples.opt.handle_solve_bounds_foas_ex.main`
    :meth:`naginterfaces.library.examples.opt.handle_solve_dfls_ex.main`
    :meth:`naginterfaces.library.examples.opt.handle_solve_dfno_ex.main`
    :meth:`naginterfaces.library.examples.opt.handle_solve_ipopt_ex.main`
    :meth:`naginterfaces.library.examples.opt.handle_solve_lp_ipm_ex.main`
    :meth:`naginterfaces.library.examples.opt.handle_solve_pennon_bmi_ex.main`
    :meth:`naginterfaces.library.examples.opt.handle_solve_pennon_lmi_ex.main`
    :meth:`naginterfaces.library.examples.opt.handle_solve_socp_ipm_ex.main`
    """

    _b_opt.handle_opt_set(
        handle,
        optstr,
    )

def handle_opt_get(handle, optstr):
    u"""
    Option getting routine for the solvers from the NAG optimization
    modelling suite.

    ``handle_opt_get`` is an option getting function for all solvers
    from the NAG optimization modelling suite.
    It is used to query the value of options.

    For full information please refer to the NAG Library document for
    e04zn

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04znf.html

    Parameters
    ----------
    handle : Handle
        The handle to the problem. It needs to be initialized by
        ``handle_init`` and **must not** be changed before the call to
        ``handle_opt_get``.

    optstr : str
        A string identifying the option whose current value is required.
        See the documentation of the particular solver for information
        on valid options.

    Returns
    -------
    optvalue : dict
        The option-value ``dict``, with the following keys:

        ``'value'`` : float, int or str
            The value of the requested option.

        ``'annotation'`` : None or str
            Possible additional information about the option value.

    Raises
    ------
    NagValueError
        (`errno` 1)
            `handle` has not been initialized.

        (`errno` 1)
            `handle` does not belong to the NAG optimization modelling
            suite, has not been initialized properly or is corrupted.

        (`errno` 1)
            `handle` has not been initialized properly or is corrupted.

        (`errno` 11)
            On entry, the option supplied in `optstr` was not
            recognized: `optstr` = *<value>*.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 41)
            On entry, `optstr` indicates a character option, but cvalue
            is too short to hold the stored value. The returned value
            will be truncated.
    """

    try:
        _return_scalars = _b_opt.handle_opt_get(
            handle,
            optstr,
        )
        _reraise = False
    except _b_utils.NagAlgorithmicWarning as _exc:
        _the_exc = _exc
        _reraise = True
        _return_scalars = _the_exc.return_data
    fun_name = 'naginterfaces.library.opt.handle_opt_get'

    _optype_py = _return_scalars.optype
    _ivalue_py = _return_scalars.ivalue
    _rvalue_py = _return_scalars.rvalue
    _cvalue_py = _return_scalars.cvalue
    if _optype_py == 1:
        _optvalue_py = {}
        _optvalue_py['value'] = _ivalue_py
        _optvalue_py['annotation'] = None
    elif _optype_py == 2:
        _optvalue_py = {}
        _optvalue_py['value'] = _rvalue_py
        _optvalue_py['annotation'] = None
    elif _optype_py == 3:
        _optvalue_py = {}
        _optvalue_py['value'] = _cvalue_py
        _optvalue_py['annotation'] = None
    else:
        raise _b_utils.NagException(
            {'fun_name': fun_name},
            'inconceivable option configuration',
        )
    _return_data = _collections.namedtuple(
        'nag_return_data',
        ['optvalue'],
    )(
        _optvalue_py,
    )
    if _reraise:
        _the_exc.return_data = _return_data
        raise _the_exc
    return _optvalue_py

def handle_opt_set_file(handle, infile, io_manager=None):
    u"""
    Option setting routine for the solvers from the NAG optimization
    modelling suite from external file.

    ``handle_opt_set_file`` is an option setting function for all
    solvers from the NAG optimization modelling suite.
    It can set one or more options from an external file.

    For full information please refer to the NAG Library document for
    e04zp

    https://www.nag.com/numeric/nl/nagdoc_27/flhtml/e04/e04zpf.html

    Parameters
    ----------
    handle : Handle
        The handle to the problem. It needs to be initialized by
        ``handle_init`` and **must not** be changed before the call to
        ``handle_opt_set_file``.

    infile : int
        The file identifier associated with the argument data file.
        **Note:** that the file needs to be opened in read mode.

    io_manager : FileObjManager, optional
        Manager for I/O in this routine.

    Raises
    ------
    NagValueError
        (`errno` 1)
            `handle` has not been initialized.

        (`errno` 1)
            `handle` does not belong to the NAG optimization modelling
            suite, has not been initialized properly or is corrupted.

        (`errno` 1)
            `handle` has not been initialized properly or is corrupted.

        (`errno` 2)
            The options cannot be modified in this phase.

        (`errno` 3)
            On entry, `infile` = *<value>*.

            Constraint: `infile` >= 0.

        (`errno` 4)
            End-of-file or read error detected before BEGIN was found.

        (`errno` 4)
            BEGIN found, but end-of-file found before END.

        (`errno` 11)
            The option provided on line *<value>* was not recognized.

        (`errno` 12)
            The expected delimiter '=' was not found on line *<value>*.

        (`errno` 13)
            The optval provided on line *<value>* could not be converted
            to the expected numerical type.

        (`errno` 15)
            The optval provided on line *<value>* for the integer option
            is out of bounds.

        (`errno` 16)
            The optval provided on line *<value>* for the real option is
            out of bounds.

        (`errno` 17)
            The optval provided on line *<value>* for the character
            option is not valid.
    """

    _b_opt.handle_opt_set_file(
        handle,
        infile,
        io_manager,
    )
